/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * LDoms Domain manager audit implementation.
 *
 * This API writes Solaris BSM Audit records to track requests sent to
 * the domain manager.  Information written includes the requestor (source),
 * request, and status (success or failure).
 *
 * This is implemented using public, stable APIs documented in man pages
 * and header files, including bsm/{audit_record.h,audit.h,libbsm.h}
 *
 * Note: this file requires these libraries:
 *	-lbsm  -lsocket  -lnsl
 *
 */

#ifdef __GNUC__
#define	__EXTENSIONS__		/* for NGROUPS_MAX when compiling with gcc */
#endif

#include <stdlib.h>		/* *alloc(), free() */
#include <string.h>		/* str*() functions */
#include <pwd.h>		/* getpwuid_r() */
#include <ucred.h>		/* ucred_t, getpeerucred() */
#include <errno.h>		/* errno */
#include <stdarg.h>		/* for va_* stdarg(3EXT) variable arguments */

/* For BSD sockets; requires -lsocket  -lnsl */
#include <sys/socket.h>		/* sockaddr */
#include <netinet/in.h>		/* sockaddr_in, sockaddr_in6 */
#include <netdb.h>		/* getaddrinfo() */

/* For BSM audit; requires -lbsm -lsocket -lnsl */
#include <sys/param.h>		/* sizes and limits; for BSM audit */
#include <bsm/audit.h>		/* BSM audit structures */
#include <bsm/audit_record.h>	/* defines audit tokens, au_to_* macros */
#include <bsm/libbsm.h>		/* auditon() */
#include <bsm/audit_uevents.h>	/* AUE_* event defines */

#ifdef USE_PRIVATE_BSD_AUDIT_APIS
#include <bsm/adt_event.h>	/* ADT_* defines */
#endif

#include "audit_ldom.h"

/* Convert addresses between generic sockets, IPv4, and IPv6 sockets */
union sockaddr_u
{
	struct sockaddr		sa;
	struct sockaddr_in	sai;
	struct sockaddr_in6	sai6;
};

/* Constants */
#define	PWBUF_SIZE	512

#define	MAX_COMMAND_TOKEN_ARGS	64

/*
 * Audit event code for LDoms (see /etc/security/audit_event)
 * Should eventually be in /usr/include/bsm/audit_uevents.h
 */
#ifndef AUE_LDOMS
#define	AUE_LDOMS	40700
#endif



/*
 * au_tid_addr_to_au_tid_addr64()
 *
 * Convert a au_tid_addr type to a au_tid_addr64().
 *
 * The difference is the former has 14/18-bit major,minor numbers in a
 * long (dev_t) and the latter 32-bit numbers in a structure.
 * See header bsm/audit.h.
 *
 * Input:
 * au_tid_addr		Variable to be converted
 *
 * Output:
 * au_tid64_addr_ptr	Pointer to variable to contain the converted variable
 */
static void
au_tid_addr_to_au_tid_addr64(au_tid_addr_t au_tid_addr,
	au_tid64_addr_t *au_tid64_addr_ptr)
{
	/* Sanity check: */
	if (au_tid64_addr_ptr == NULL) {
		return;
	}

	(void) memset(au_tid64_addr_ptr, 0, sizeof (au_tid64_addr_t));

	/* Copy structure members one-by-one: */
	au_tid64_addr_ptr->at_port.at_major =
	    ((au_tid_addr.at_port >> 18) & 0x3ffff);
	au_tid64_addr_ptr->at_port.at_minor = au_tid_addr.at_port & 0x3fff;
	au_tid64_addr_ptr->at_type = au_tid_addr.at_type;
	au_tid64_addr_ptr->at_addr[0] = au_tid_addr.at_addr[0];
	au_tid64_addr_ptr->at_addr[1] = au_tid_addr.at_addr[1];
	au_tid64_addr_ptr->at_addr[2] = au_tid_addr.at_addr[2];
	au_tid64_addr_ptr->at_addr[3] = au_tid_addr.at_addr[3];
}


/*
 * au_tid_addr64_to_au_tid_addr()
 *
 * Convert a au_tid_addr64 type to a au_tid_addr().
 *
 * The difference is the former has 32-bit major, minor device numbers in a
 * structure, and the latter has 14/18-bit major,minor numbers in a long.
 * See header bsm/audit.h.
 *
 * Input:
 * au_tid64_addr	Variable to be converted
 *
 * Output:
 * au_tid_addr_ptr	Pointer to variable to contain the converted variable
 */
static void
au_tid_addr64_to_au_tid_addr(au_tid64_addr_t au_tid64_addr,
	au_tid_addr_t *au_tid_addr_ptr)
{
	/* Sanity check: */
	if (au_tid_addr_ptr == NULL) {
		return;
	}

	(void) memset(au_tid_addr_ptr, 0, sizeof (au_tid64_addr_t));

	/* Copy structure members one-by-one: */
	au_tid_addr_ptr->at_port = (au_tid64_addr.at_port.at_major << 18) +
	    au_tid64_addr.at_port.at_minor;
	au_tid_addr_ptr->at_type = au_tid64_addr.at_type;
	au_tid_addr_ptr->at_addr[0] = au_tid64_addr.at_addr[0];
	au_tid_addr_ptr->at_addr[1] = au_tid64_addr.at_addr[1];
	au_tid_addr_ptr->at_addr[2] = au_tid64_addr.at_addr[2];
	au_tid_addr_ptr->at_addr[3] = au_tid64_addr.at_addr[3];
}


#ifdef USE_PRIVATE_BSD_AUDIT_APIS

/*
 * BSD Audit project-private APIs
 *
 * This kernel APIs, although useful, have not matured and are project private.
 * Currently, the kernel interfaces are documented only in
 * kernel source file $SRC/usr/src/lib/libbsm/common/adt_ucred.h and are not
 * available in header /usr/include/ucred.h
 * They are emulated below using underlying stable public API auditon(2).
 * Use these kernel APIs if and when they ever mature and become stable public.
 */
extern  au_id_t ucred_getauid(const ucred_t *uc);
extern  au_asid_t ucred_getasid(const ucred_t *uc);
extern  const au_tid64_addr_t *ucred_getatid(const ucred_t *uc);


#else /* use only public APIs */

/*
 * Emulation APIs
 *
 * These APIs emulate project-private BSM Audit kernel calls by the same name.
 * The kernel calls can't be used because they have not matured and are
 * project private.  They are emulated below using underlying stable public
 * API auditon(2).
 */


/*
 * ucred_getauid()
 *
 * Emulate the private, unstable interface by the same name with the stable
 * auditon(2) API.
 *
 * Note: if or when the kernel function ucred_getauid() ever becomes stable,
 * simply replace this function with the kernel function.
 * Currently, the kernel interface is documented only in
 * kernel source file $SRC/usr/src/lib/libbsm/common/adt_ucred.h and is not
 * available in header /usr/include/ucred.h
 *
 * Note2: If BSM auditing is not enabled, this call fails with EINVAL.
 * No error should be printed or logged if this occurs.
 *
 * Input:
 * uc	Pointer to credential (from getpeerucred())
 *
 * Return:
 * Audit ID number. If none or on failure, return AU_NOAUDITID.
 */
static	au_id_t
ucred_getauid(const ucred_t *uc) {
	pid_t			peer_pid = (pid_t)-1;
	struct auditpinfo_addr	ap;

	/* Sanity check */
	if (uc == NULL) {
		errno = EINVAL;
		return ((au_id_t)AU_NOAUDITID);
	}

	/* Get PID, to lookup Audit ID */
	peer_pid = ucred_getpid(uc);
	if (peer_pid == (pid_t)-1) {
		return ((au_id_t)AU_NOAUDITID);
	}

	/* Lookup the Audit ID with the stable auditon(2) API: */
	(void) memset(&ap, 0, sizeof (struct auditpinfo_addr));
	ap.ap_pid = peer_pid;
	if (auditon(A_GETPINFO_ADDR, (caddr_t)&ap, sizeof (ap)) == -1) {
		return ((au_id_t)AU_NOAUDITID);
	}

	/* Extract and return the Audit ID information: */
	return (ap.ap_auid);
}


/*
 * ucred_getasid()
 *
 * Emulate the private, unstable interface by the same name with the stable
 * auditon(2) API.
 *
 * Note: if or when the kernel function ucred_getasid() ever becomes stable,
 * simply replace this function with the kernel function.
 * Currently, the kernel interface is documented only in
 * kernel source file $SRC/usr/src/lib/libbsm/common/adt_ucred.h and is not
 * available in header /usr/include/ucred.h
 *
 * Note2: If BSM auditing is not enabled, this call fails with EINVAL.
 * No error should be printed or logged if this occurs.
 *
 * Input:
 * uc	Pointer to credential (from getpeerucred())
 *
 * Return:
 * Audit session ID number. If none or on failure, return -1
 */
static au_asid_t
ucred_getasid(const ucred_t *uc) {
	pid_t			peer_pid = (pid_t)-1;
	struct auditpinfo_addr	ap;

	/* Sanity check */
	if (uc == NULL) {
		errno = EINVAL;
		return ((au_asid_t)-1);
	}

	/* Get PID, to lookup Audit ID */
	peer_pid = ucred_getpid(uc);
	if (peer_pid == (pid_t)-1) {
		return ((au_asid_t)-1);
	}

	/* Lookup the Audit ID with the stable auditon(2) API: */
	(void) memset(&ap, 0, sizeof (struct auditpinfo_addr));
	ap.ap_pid = peer_pid;
	if (auditon(A_GETPINFO_ADDR, (caddr_t)&ap, sizeof (ap)) == -1) {
		return ((au_asid_t)-1);
	}

	/* Extract and return the Audit Session ID information: */
	return (ap.ap_asid);
}


/*
 * ucred_getatid()
 *
 * Emulate the private, unstable interface by the same name with the stable
 * auditon(2) API.
 *
 * Note: if or when the kernel function ucred_getatid() ever becomes stable,
 * simply replace this function with the kernel function.
 * Currently, the kernel interface is documented only in
 * kernel source file $SRC/usr/src/lib/libbsm/common/adt_ucred.h and is not
 * available in header /usr/include/ucred.h
 *
 * Note2: If BSM auditing is not enabled, this call fails with EINVAL.
 * No error should be printed or logged if this occurs.
 *
 * Input:
 * uc	Pointer to credential (from getpeerucred())
 *
 * Return:
 * Audit terminal ID number. If none or on failure, return NULL.
 * The TID consists of port (major & minor #), address type (4 or 16 bytes
 * for IPv4/IPv6), and address).
 */
static const au_tid64_addr_t *
ucred_getatid(const ucred_t *uc) {
	pid_t			peer_pid = (pid_t)-1;
	struct auditpinfo_addr	ap;
	static au_tid64_addr_t	au_tid64_addr;

	/* Sanity check */
	if (uc == NULL) {
		errno = EINVAL;
		return (NULL);
	}

	/* Get PID, to lookup Audit ID */
	peer_pid = ucred_getpid(uc);
	if (peer_pid == (pid_t)-1) {
		return (NULL);
	}

	/* Lookup the Audit ID with the stable auditon(2) API: */
	(void) memset(&ap, 0, sizeof (struct auditpinfo_addr));
	ap.ap_pid = peer_pid;
	if (auditon(A_GETPINFO_ADDR, (caddr_t)&ap, sizeof (ap)) == -1) {
		/* Extract the Audit ID information */
		return (NULL);
	}

	/*
	 * Extract and return the Audit Terminal ID information,
	 * after converting to an au_tid64_addr_t used by BSM Audit records.
	 */
	au_tid_addr_to_au_tid_addr64(ap.ap_termid, &au_tid64_addr);

	return (&au_tid64_addr);
}

#endif


/*
 * sockaddr_to_au_tid_addr()
 *
 * Translate a sockaddr, IPv4 or IPv6, to an au_tid_addr_t.
 * au_tid_addr_t is used to create an expanded subject audit token.
 *
 * Note: * sockaddr is defined in socket.h, au_tid_addr_t is in bsm/audit.h.
 *
 * Parameters:
 * sa		sockaddr structure pointer. May contain IPv4 or IPv6 info.
 *		Contains socket peer's address
 * au_tid_addr	pointer to audit structure to contain the port and address
 *		info from parameters sa and port.
 *
 * Return:
 * 0 on success, non-zero on failure
 */
static int
sockaddr_to_au_tid_addr(struct sockaddr *sa, au_tid_addr_t *au_tid_addr) {
	struct sockaddr_in	*sai = NULL;
	struct sockaddr_in6	*sai6 = NULL;

	/* Sanity checks */
	if ((sa == NULL) || (au_tid_addr == NULL)) {
		return (-1);
	}

	(void) memset(au_tid_addr, 0, sizeof (au_tid_addr_t));

	switch (sa->sa_family) {
	case AF_INET: /* copy over IPv4 address from sockaddr_in */
		sai = (struct sockaddr_in *)sa;
		au_tid_addr->at_type = AU_IPv4;
		au_tid_addr->at_addr[0] = sai->sin_addr._S_un._S_addr;
		break;

	case AF_INET6: /* copy over IPv6 address from sockaddr_in6 */
		sai6 = (struct sockaddr_in6 *)sa;
		au_tid_addr->at_type = AU_IPv6;
		/* copy IPv6 address in struct in6_addr: */
		au_tid_addr->at_addr[0] = sai6->sin6_addr._S6_un._S6_u32[0];
		au_tid_addr->at_addr[1] = sai6->sin6_addr._S6_un._S6_u32[1];
		au_tid_addr->at_addr[2] = sai6->sin6_addr._S6_un._S6_u32[2];
		au_tid_addr->at_addr[3] = sai6->sin6_addr._S6_un._S6_u32[3];
		break;

	default: /* unknown address */
		break;
	}

	return (0);
}


/*
 * is_auditing_enabled()
 *
 * Return 1 (true) if BSM auditing is enabled, otherwise return 0 (false).
 * Similar to the CLI: /usr/sbin/auditconfig -getcond
 *
 * Note: auditon(2)'s second parameter must be an integer, not a long, as
 * incorrectly documented in the man page (otherwise it won't work for some
 * 64-bit models where the size of long isn't the size of int).
 * See doc bug 6411356.
 */
static int
is_auditing_enabled()
{
	int	audit_cond = AUC_UNSET;
	int	rv;

	/* See if auditing is enabled */
	rv = auditon(A_GETCOND, (caddr_t)&audit_cond, sizeof (audit_cond));
	if (rv == -1) {
		/* BSM not installed or loaded */
		return (0);
	} else {
		return (audit_cond == AUC_AUDITING);
	}
}


/*
 * open_audit_stream()
 *
 * Open an audit stream for writing tokens.
 * Opening the stream also writes a header token.
 * Also write an extended subject token (which handles borh IPv4/IPv6)
 *
 * Parameter:
 * sock_fd	File number of socket connected to peer.  Used to get uid and
 *		other information on peer.  Use -1 if not known.
 *
 * Return value:
 * The return value from au_open(), which is the Audit Descriptor (ad) number,
 * or NULL_AUDIT_HANDLE on failure.
 */
static audit_handle_t
open_audit_stream(int sock_fd)
{
	audit_handle_t		ad = NULL_AUDIT_HANDLE;
	int			rc = -1;
	ucred_t			*ucredp = NULL;
	int			saved_errno = 0;

	/* Sanity check */
	if (sock_fd < -1) {
		return (NULL_AUDIT_HANDLE);
	}

	/*
	 * Open the audit stream
	 */
#ifdef USE_PRIVATE_BSD_AUDIT_APIS
	rc = adt_start_session(&ad, NULL, 0);
#ifdef AUDIT_DEBUG
	if (rc == -1) {
		perror("au_open");
		printf("\nBSM adt_start_session: %s failed: %s\n",
		    __func__, strerror(errno));
	}
#endif

#else /* public interface */

	/* Open the audit stream (implicitly writes a header token) */
	ad = au_open();
	/* printf("au_open returned audit descriptor %d\n", ad); */
#ifdef AUDIT_DEBUG
	if (ad == -1) {
		perror("au_open");
		printf("\nBSM au_open: %s failed to open: %s\n",
		    __func__, strerror(errno));
	}
#endif
#endif


	/*
	 * Write an extended subject record
	 */
#ifdef USE_PRIVATE_BSD_AUDIT_APIS

	if (sock_fd >= 0) { /* extract info on socket peer */
		/* Get info on the peer on the other side of the socket */
		rc = getpeerucred(sock_fd, &ucredp);
		if (rc == 0) { /* local socket peer */
			rc = adt_set_from_ucred(ad, ucredp, ADT_NEW);
			ucred_free(ucredp);

		} else { /* remote */
			adt_termid_t	*termid = NULL;
			rc = adt_load_termid(sock_fd, &termid);
			if (rc == 0) { /* local socket peer */
				rc = adt_set_user(ad, (uid_t)-1, (gid_t)-1,
				    (uid_t)-1, (gid_t)-1, termid, ADT_NEW);
			}
		}
	}


#else /* public interface */
	uid_t			peer_euid = (uid_t)AU_NOAUDITID;
	gid_t			peer_egid = (gid_t)AU_NOAUDITID;
	uid_t			peer_ruid = (uid_t)AU_NOAUDITID;
	gid_t			peer_rgid = (gid_t)AU_NOAUDITID;
	pid_t			peer_pid = (pid_t)-1;
	au_id_t			peer_auid = (au_id_t)AU_NOAUDITID;
	au_asid_t		peer_asid = (au_asid_t)-1;
	const au_tid64_addr_t	*peer_atid64_ptr;
	au_tid_addr_t		peer_atid;
	union sockaddr_u	sa;
	socklen_t		sockaddr_len = sizeof (sa);

	(void) memset(&peer_atid, 0, sizeof (peer_atid));

	/* printf("write_ex_subject_token(ad %d, fd %d)\n", ad, sock_fd); */

	if (sock_fd >= 0) { /* extract info on socket peer */
		/* Get info on the peer on the other side of the socket */
		rc = getpeerucred(sock_fd, &ucredp);
		if (rc == 0) { /* local socket peer */

			/* Extract user/group info for the local socket peer */
			peer_euid = ucred_geteuid(ucredp);
			peer_egid = ucred_getegid(ucredp);
			peer_ruid = ucred_getruid(ucredp);
			peer_rgid = ucred_getrgid(ucredp);
			peer_pid = ucred_getpid(ucredp);
			peer_auid = ucred_getauid(ucredp);
			peer_asid = ucred_getasid(ucredp);

			peer_atid64_ptr = ucred_getatid(ucredp);
			if (peer_atid64_ptr != NULL) {
				au_tid_addr64_to_au_tid_addr(*peer_atid64_ptr,
				    &peer_atid);
			}

			ucred_free(ucredp);

		} else if (errno == EINVAL) { /* remote socket peer */

			/* Get remote peer's IP address */
			(void) memset(&sa, 0, sizeof (sa));
			if (getpeername(sock_fd, &sa.sa, &sockaddr_len) == 0) {
				/* convert to a type supported by audit */
				sockaddr_to_au_tid_addr(&sa.sa, &peer_atid);
			}
		}
	}

	/* Write the extended subject record */
	rc = au_write(ad, au_to_subject_ex(peer_auid, peer_euid, peer_egid,
	    peer_ruid, peer_rgid, peer_pid, peer_asid, &peer_atid));
	saved_errno = errno;

	if (rc < 0) {
#ifdef AUDIT_DEBUG
		perror("au_write(au_to_subject())");
		printf("\nBSM au_write(au_to_subject):"
		    " %s failed to write record: %s\n",
		    __func__, strerror(errno));
#endif
		close_audit_stream(ad, saved_errno);
	}
#endif
	return ((rc < 0) ? NULL_AUDIT_HANDLE : ad);
}


/*
 * close_audit_stream()
 *
 * Write a return token that records the status codes
 * and close the audit stream.
 *
 * Parameters:
 * ad		Audit stream descriptor number or handle
 *		If NULL_AUDIT_HANDLE, return silently
 * status	Program Status value to save (0=OK, else error status code)
 *
 * Return:
 * 0 on success, non-zero on failure
 */
int
close_audit_stream(audit_handle_t ad, int status)
{
	int		rc = -1;

	if (!is_auditing_enabled()) {
		return (0); /* silent no-op if auditing disabled */
	}

	if (ad == NULL_AUDIT_HANDLE) {
		return (0); /* silent no-op if no audit record created */
	}

#ifdef USE_PRIVATE_BSD_AUDIT_APIS

	/* Close the audit descriptor: */
	rc = adt_end_session(ad);

#ifdef AUDIT_DEBUG
	if (rc < 0) {
		perror("adt_end_session");
		printf("\nBSM audit: %s failed to end session: %s\n",
		    __func__, strerror(errno));
	}
#endif

#else /* public interface */

	/* Write a "return" audit record */
	rc = au_write(ad, au_to_return32((status == 0 ? 0 : -1), status));

#ifdef AUDIT_DEBUG
	if (rc < 0) {
		perror("au_write(au_to_return32)");
		printf("\nBSM au_write(au_to_return32):"
		    " %s failed to write record: %s\n",
		    __func__, strerror(errno));
	}
#endif

	/* Close the audit descriptor: */
	rc = au_close(ad, AU_TO_WRITE, AUE_LDOMS);
#ifdef AUDIT_DEBUG
	if (rc < 0) {
		perror("au_close");
		printf("\nBSM audit: %s failed to close record: %s\n",
		    __func__, strerror(errno));
	}
#endif
#endif

	return (rc);
}



/*
 * write_command_token_argv()
 *
 * Write the cmd token from argv
 *
 * Parameters:
 * ad		Audit stream descriptor number or handle
 * argc		Number of arguments in argv
 * argv		Array of strings listing arguments, if any, terminated by NULL.
 *		argv[0] is the command name (e.g., "ldominfo")
 *		NULL if no arguments.
 *
 * Return:
 * 0 on success, non-zero on failure
 */
static int
write_command_token_argv(audit_handle_t ad, int argc, char *argv[])
{
	int	rc = -1;

	rc = au_write(ad, au_to_cmd(argc, argv, NULL));
#ifdef AUDIT_DEBUG
	if (rc < 0) {
		perror("au_write(au_to_cmd)");
		printf("\nBSM au_write(au_to_cmd):"
		    " %s failed to write record: %s\n",
		    __func__, strerror(errno));
	}
#endif
	return (rc);
}

/*
 * audit_ldm_request()
 *
 * Audit a "generic" LDom request (that is, the request only passes the
 * domain name and nothing else).
 *
 * Note: close_audit_stream() must be called later with the ad (returned by
 * this function) and the exit codes, when known.
 *
 * Input:
 * sock_fd	File number of socket connected to peer.  Used to get uid and
 *		other information on peer.  Use -1 if not known.
 * argc		Count of arguments, including "request"
 * cmd		String containing the command (request) name
 * 		(e.g., "ldm" or "ldom*")
 * ...		One or more strings containing parameters (optional)
 *
 * Return:
 * audit descriptor number if known, else NULL_AUDIT_HANDLE (failure)
 */
audit_handle_t
audit_ldm_request(int sock_fd, int argc, char *cmd, ...)
{
	va_list		ap;
	char		*argv_buffer[MAX_COMMAND_TOKEN_ARGS];
	char		**argvp = argv_buffer;
	void		*allocated_memory = NULL;
	audit_handle_t	ad = NULL_AUDIT_HANDLE;
	int		i = 0;

	/* Sanity checks */
	if ((sock_fd < -1) || !is_auditing_enabled() || (argc < 1)) {
		return (NULL_AUDIT_HANDLE);
	}

	/* Initialize array to hold arguments */
	if (argc <= MAX_COMMAND_TOKEN_ARGS) {
		(void) memset(argv_buffer, 0, sizeof (argv_buffer));
	} else { /* allocate memory */
		argvp = allocated_memory = calloc(argc, sizeof (char *));
		if (argvp == NULL) { /* no memory: truncate args */
			argc = MAX_COMMAND_TOKEN_ARGS;
			argvp = argv_buffer;
			(void) memset(argv_buffer, 0, sizeof (argv_buffer));
		}
	}

	/* First argument is always the command request */
	argvp[0] = cmd;

	/* Remaining argument are parameters, if any */
	if (argc > 1) { /* copy arguments */
		va_start(ap, cmd);
		for (i = 1; i <= argc - 1; ++i) {
			argvp[i] = va_arg(ap, char *);
		}
		va_end(ap);
	}

	/* Write tokens */
	ad = open_audit_stream(sock_fd);
	if (ad != NULL_AUDIT_HANDLE) {
		(void) write_command_token_argv(ad, argc, argvp);
	}

	/* Cleanup and return */
	if (allocated_memory != NULL) {
		free(allocated_memory);
	}

	return (ad);
}



/*
 * audit_ldm_request_argv()
 *
 * Audit a "generic" LDom request (that is, the request only passes the
 * domain name and nothing else).
 *
 * Note: close_audit_stream() must be called later with the ad (returned by
 * this function) and the exit codes, when known.
 *
 * Input:
 * sock_fd	File number of socket connected to peer.  Used to get uid and
 *		other information on peer.  Use -1 if not known.
 * argc		Count of arguments, including argv[0] (request)
 * argv		Array of strings containing the command (request) name in
 *		argv[0] and zero or more arguments in argv[1] . . .
 *
 * Return:
 * audit descriptor number if known, else NULL_AUDIT_HANDLE (failure)
 */
audit_handle_t
audit_ldm_request_argv(int sock_fd, int argc, char *argv[])
{
	audit_handle_t		ad = NULL_AUDIT_HANDLE;

	/* Sanity checks */
	if ((sock_fd < -1) || !is_auditing_enabled() || (argc < 1)) {
		return (NULL_AUDIT_HANDLE);
	}

	/* Write tokens */
	ad = open_audit_stream(sock_fd);
	if (ad != NULL_AUDIT_HANDLE) {
		(void) write_command_token_argv(ad, argc, argv);
	}

	/* Return */
	return (ad);
}


/*
 * audit_ldm_request_cmd_argv()
 *
 * Audit a "generic" LDom request (that is, the request only passes the
 * domain name and nothing else).
 *
 * Note: close_audit_stream() must be called later with the ad (returned by
 * this function) and the exit codes, when known.
 *
 * Input:
 * sock_fd	File number of socket connected to peer.  Used to get uid and
 *		other information on peer.  Use -1 if not known.
 * cmd		String containing the command (request) name
 * 		(e.g., "ldm" or "ldom*")
 * argc		Count of arguments, excluding argv[0] (request)
 * argv		Array of strings containing zero or more arguments
 *		in argv[0] . . .
 *
 * Return:
 * audit descriptor number if known, else NULL_AUDIT_HANDLE (failure)
 */
audit_handle_t
audit_ldm_request_cmd_argv(int sock_fd, char *cmd, int argc, char *argv[])
{
	char			*argv_buffer[MAX_COMMAND_TOKEN_ARGS];
	char			**argvp = argv_buffer;
	void			*allocated_memory = NULL;
	int			i = 0;
	audit_handle_t		ad = NULL_AUDIT_HANDLE;

	/* Sanity checks */
	if ((sock_fd < -1) || !is_auditing_enabled() || (argc < 0)) {
		return (NULL_AUDIT_HANDLE);
	}

	/* Correct argc count to include command */
	++argc;

	/* Initialize array to hold arguments */
	if (argc <= MAX_COMMAND_TOKEN_ARGS) {
		(void) memset(argv_buffer, 0, sizeof (argv_buffer));
	} else { /* allocate memory */
		argvp = allocated_memory = calloc(argc, sizeof (char *));
		if (argvp == NULL) { /* no memory: truncate args */
			argc = MAX_COMMAND_TOKEN_ARGS;
			(void) memset(argv_buffer, 0, sizeof (argv_buffer));
			argvp = argv_buffer;
		}
	}

	/* First argument is always the command request */
	argvp[0] = cmd;

	/* Remaining argument are parameters, if any */
	if (argc > 1) { /* copy arguments */
		for (i = 1; i < argc; ++i) {
			argvp[i] = argv[i - 1];
		}
	}

	/* Write tokens */
	ad = open_audit_stream(sock_fd);
	if (ad != NULL_AUDIT_HANDLE) {
		(void) write_command_token_argv(ad, argc, argvp);
	}

	/* Cleanup and return */
	if (allocated_memory != NULL) {
		free(allocated_memory);
	}

	return (ad);
}
