/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)auth_ldom.c	1.18	09/05/07 SMI"

/*
 * LDoms Domain manager authorization checking.
 *
 * These functions check LDoms authorization to perform Zeus commands.
 * The mechanism used is Solaris authorizations.  Each command needs a
 * read or write authorization and the local client process requesting
 * the command is verified to have that authorization.
 *
 * On failure, functions return an error status that should be checked
 * by the caller.  On errors the caller should generate an error message
 * and NOT perform the request.
 *
 * Authorizations available are read or write permission.
 * A client request via CLI must have the appropriate authorization
 * from file /etc/security/auth_attr (added during Zeus install):
 * solaris.ldoms.read:::View LDom configuration::
 * solaris.ldoms.write:::Modify LDom configuration::
 *
 * Specific users are authorized with usermod(1M). To add an authorization
 * (to /etc/user_attr) type a command similar to this (when user NOT
 * logged in):
 *	usermod -A solaris.ldoms.read johndoe
 *
 * Note: this file requires these libraries in LOADLIBES (for chkauthattr()):
 *	-lsecdb  -lsocket  -lnsl
 *
 * Note 2: for a standalone unit test see tests/auth_ldom_test.c.
 *
 */

/* System headers */
#include <sys/types.h>		/* uid_t */
#include <string.h>		/* memset() */
#include <pwd.h>		/* getpw*() */
#include <priv.h>		/* privileges  */
#include <auth_attr.h>		/* chkauthattr() */
#include <secdb.h>		/* chkauthattr() */
#include <ucred.h>		/* getpeerucred() */
#include <limits.h>		/* PASS_MAX */

/* Project headers */
#include "auth_ldom.h"		/* auth_check_fd() */
#include "clp.h"		/* ldm_cmd_num_t */
#include "errors.h"		/* MSG_ERR_XXX */
#include "ldom.h"		/* Needed by zprintf.h */



/*
 * socket_peer_euid()
 *
 * Return the effective UID (EUID) of the socket peer.
 * If none, return -1.
 *
 * Parameters:
 * sock_fd	The socket fd of a locally-connected socket (mapped to a pid)
 *
 * Returns:
 * EUID if OK
 * -1 on failure or unknown EUID (passed on from ucred_geteuid()).
 */
uid_t
socket_peer_euid(int sock_fd)
{
	int		rc;
	uid_t		peer_euid;
	ucred_t		*ucredp = NULL;

	/* Get info on the peer on the other side of the socket */
	rc = getpeerucred(sock_fd, &ucredp);
	if (rc == -1) {
		/* If errno is EINVAL, it's probably a non-local socket peer */
		return (-1);
	}

	/* Extract effective UID (EUID) info for the socket peer process */
	peer_euid = ucred_geteuid(ucredp);
	ucred_free(ucredp);

	/* Return EUID */
	return (peer_euid);
}

/*
 * auth_check_xml_username()
 *
 * Check LDom authorization (either read or write), given a user account.
 * Called by auth_check_xml_euid.
 *
 * Note: for "ldm *"-style commands (e.g., "ldm create").
 *
 * Parameters:
 * username	The name of a user account to check authorization
 * xml_cmd	Zeus XML command string to check
 *
 * Returns:
 * 0 if OK (authorized), 1 on authorization failure.
 */
int
auth_check_xml_username(char *username, char *xml_cmd)
{
	/* Sanity check: */
	if ((username == NULL) || (username[0] == '\0') || (xml_cmd == NULL) ||
	    (xml_cmd[0] == '\0')) {
		return (1); /* error (bad parameter) */
	}

	/* DO authorization check: */
	if (strncmp(xml_cmd, "list", strlen("list")) == 0) {
		/* list*: a "read" command */
		return ((chkauthattr("solaris.ldoms.read", username) != 1) &&
		    (chkauthattr("solaris.ldoms.write", username) != 1));

	} else { /* everything else are "write" commands */
		return (chkauthattr("solaris.ldoms.write", username) != 1);
	}
}


/*
 * auth_check_xml_euid()
 *
 * Check LDom authorization (either read or write), given a EUID.
 * Called by auth_check_xml_fd and may be called externally.
 *
 * Note: for "ldm *"-style commands (e.g., "ldm create").
 *
 * Parameters:
 * euid		The effective UID of a user account to check authorization
 * xml_cmd	Zeus XML command string to check
 *
 * Returns:
 * 0 if OK (authorized), 1 on authorization failure.
 */
int
auth_check_xml_euid(uid_t euid, char *xml_cmd)
{
	struct passwd	*passwdp = NULL;
	char		*username = NULL;

	/* Sanity check: */
	if ((xml_cmd == NULL) || (xml_cmd[0] == '\0')) {
		return (1); /* error (bad parameter) */
	}

	/* If EUID is -1, then it's unknown, so fail */
	if (euid == -1) {
		return (1);
	}

	/* Map EUID to user name */
	passwdp = getpwuid(euid);
	if (passwdp == NULL) { /* lookup failed */
		return (1);
	}
	username = passwdp->pw_name;

	/* DO authorization check: */
	return (auth_check_xml_username(username, xml_cmd));
}


/*
 * auth_check_xml_fd()
 *
 * Check LDom authorization (either read or write), given a fd of a socket.
 * The socket fd is mapped to a pid (and should not be used for remote
 * connections).
 *
 * Note: for "ldm *"-style commands (e.g., "ldm create").
 *
 * Parameters:
 * sock_fd	The socket fd of a locally-connected socket (mapped to a pid)
 * xml_cmd	Zeus XML command string to check
 *
 * Returns:
 * 0 if OK (authorized), 1 on authorization failure.
 */
int
auth_check_xml_fd(int sock_fd, char *xml_cmd)
{
	uid_t		peer_euid;

	/* Sanity check: */
	if ((xml_cmd == NULL) || (xml_cmd[0] == '\0')) {
		return (1); /* error (bad parameter) */
	}

	peer_euid = socket_peer_euid(sock_fd);
	if (peer_euid == -1) { /* unknown EUID */
		return (1);
	}

	/* DO authorization check: */
	return (auth_check_xml_euid(peer_euid, xml_cmd));
}


/*
 * auth_check_username()
 *
 * Check LDom authorization (either read or write), given a user account.
 * Called by auth_check_euid.
 *
 * Parameters:
 * username	The name of a user account to check authorization
 * cmd		Zeus command to check
 *
 * Returns:
 * 0 if OK (authorized), 1 on authorization failure.
 */
static int
auth_check_username(char *username, action_verb_t cmd)
{
	/* Sanity check: */
	if ((username == NULL) || (username[0] == '\0')) {
		return (1); /* error (bad parameter) */
	}

	/* DO authorization check: */
	switch (cmd) {
	case CLI_VERB_LIST:
		return ((chkauthattr("solaris.ldoms.read", username) != 1) &&
		    (chkauthattr("solaris.ldoms.write", username) != 1));

	case CLI_VERB_ADD:
	case CLI_VERB_SET:
	case CLI_VERB_REMOVE:
	case CLI_VERB_START:
	case CLI_VERB_STOP:
	case CLI_VERB_BIND:
	case CLI_VERB_UNBIND:
	case CLI_VERB_PANIC:
	case CLI_VERB_MIGRATE:
	case CLI_VERB_CANCEL:
	default:
		return (chkauthattr("solaris.ldoms.write", username) != 1);
	}
}

/*
 * auth_check_euid()
 *
 * Check LDom authorization (either read or write), given a EUID.
 * Called by auth_check_fd() and may be called externally.
 *
 * Parameters:
 * euid		The effective UID of a user account to check authorization
 * cmd		Zeus command to check
 *
 * Returns:
 * 0 if OK (authorized), 1 on authorization failure.
 */
int
auth_check_euid(uid_t euid, action_verb_t cmd)
{
	struct passwd	*passwdp = NULL;
	char		*username = NULL;

	/* If EUID is -1, then it's unknown, so fail */
	if (euid == -1) {
		return (1);
	}

	/* Map EUID to user name */
	passwdp = getpwuid(euid);
	if (passwdp == NULL) { /* lookup failed */
		return (1);
	}
	username = passwdp->pw_name;

	/* DO authorization check: */
	return (auth_check_username(username, cmd));
}

/*
 * auth_check_fd()
 *
 * Check LDom authorization (either read or write), given a fd of a socket.
 * The socket fd is mapped to a pid (and should not be used for remote
 * connections).
 *
 * Parameters:
 * sock_fd	The socket fd of a locally-connected socket (mapped to a pid)
 * cmd		Zeus command to check
 *
 * Returns:
 * 0 if OK (authorized), 1 on authorization failure.
 */
int
auth_check_fd(int sock_fd, action_verb_t cmd)
{
	uid_t		peer_euid;

	peer_euid = socket_peer_euid(sock_fd);
	if (peer_euid == -1) { /* unknown EUID */
		return (1);
	}

	/* DO authorization check: */
	return (auth_check_euid(peer_euid, cmd));
}

/*
 * Least privilege helper functions
 */

void
ldm_priv_init(void)
{
	priv_set_t *privset;

	/*
	 * Add our privileges and remove unneeded 'basic' privileges from the
	 * permitted set.
	 */
	if ((privset = priv_str_to_set("basic", ",", NULL)) == NULL)
		printf(MSG_ERR_PRIV);

	(void) priv_addset(privset, PRIV_FILE_DAC_SEARCH);
	(void) priv_addset(privset, PRIV_FILE_DAC_READ);
	(void) priv_delset(privset, PRIV_PROC_EXEC);
	(void) priv_delset(privset, PRIV_PROC_FORK);
	(void) priv_delset(privset, PRIV_PROC_INFO);
	(void) priv_delset(privset, PRIV_PROC_SESSION);
	priv_inverse(privset);
	if (setppriv(PRIV_OFF, PRIV_PERMITTED, privset) == -1)
		printf(MSG_ERR_PRIV);
	priv_freeset(privset);

	/*
	 * Clear the Inheritable and Limit sets.
	 */
	if ((privset = priv_allocset()) == NULL) {
		printf(MSG_ERR_PRIV);
		return;
	}
	priv_emptyset(privset);
	if (setppriv(PRIV_SET, PRIV_INHERITABLE, privset) == -1 ||
	    setppriv(PRIV_SET, PRIV_LIMIT, privset) == -1)
		printf(MSG_ERR_PRIV);
	priv_freeset(privset);

	/*
	 * Turn off the file_dac_search, file_dac_read privileges until needed.
	 */
	(void) priv_set(PRIV_OFF, PRIV_EFFECTIVE, PRIV_FILE_DAC_SEARCH, NULL);
	(void) priv_set(PRIV_OFF, PRIV_EFFECTIVE, PRIV_FILE_DAC_READ, NULL);
}

void
priv_on(void)
{
	int	rv;

	rv = priv_set(PRIV_ON, PRIV_EFFECTIVE, PRIV_FILE_DAC_READ, NULL);
	if (rv == -1) {
		printf(MSG_ERR_PRIV_SET, "enable", "FILE_DAC_READ");
		exit(1);
	}
	rv = priv_set(PRIV_ON, PRIV_EFFECTIVE, PRIV_FILE_DAC_SEARCH, NULL);
	if (rv == -1) {
		printf(MSG_ERR_PRIV_SET, "enable", "FILE_DAC_SEARCH");
		exit(1);
	}
}

void
priv_off(void)
{
	int	rv;

	rv = priv_set(PRIV_OFF, PRIV_EFFECTIVE, PRIV_FILE_DAC_SEARCH, NULL);
	if (rv == -1) {
		printf(MSG_ERR_PRIV_SET, "disable", "FILE_DAC_READ");
		exit(1);
	}

	rv = priv_set(PRIV_OFF, PRIV_EFFECTIVE, PRIV_FILE_DAC_READ, NULL);
	if (rv == -1) {
		printf(MSG_ERR_PRIV_SET, "disable", "FILE_DAC_SEARCH");
		exit(1);
	}
}
