/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <errno.h>
#include <alloca.h>
#include <unistd.h>
#include <time.h>
#include <fcntl.h>
#include <dirent.h>

#include "allocate.h"
#include "basics.h"
#include "errors.h"
#include "fatal.h"
#include "response.h"
#include "ldom.h"
#include "zprintf.h"
#include "option.h"
#include "mdstore.h"
#include "sequencer.h"
#include "support.h"
#include "ds_impl.h"
#include "autosave.h"

/*
 * The autosave files are saved under the directory:
 *	/var/opt/SUNWldm/autosave-<config_name>
 *
 *	The files in the directory are of the form:
 *	  md-<guest_name>.bin -	Guest domain MD
 *	  hv-md.bin -           Hypervisor MD
 *	  sp-pri.bin -		SP PRI MD
 *	  sp-timestamp -	Timestamp of the sp config
 *	  timestamp -		Timestamp of the last update (should be the
 *				max timestamp value in all the MD files).
 *
 *	Each MD file has a Autosave MD header which includes various information
 *	about the MD, such as physical address, MD type, timestamp, etc.  A
 *	checksum on the data segment is also maintained.
 */

/*
 * Macros to generate file names.
 */
#define	AUTOSAVE_ROOTDIR	zeus.admdir
#define	AUTOSAVE_FILETYPE	"bin"
#define	AUTOSAVE_MDPREFIX	"autosave-"
#define	AUTOSAVE_MDPREFIXLEN	9
#define	AUTOSAVE_GPREFIX	"md-"
#define	AUTOSAVE_GPREFIXLEN	3
#define	AUTOSAVE_HVMD_NAME	"hv-md"
#define	AUTOSAVE_SPPRI_NAME	"sp-pri"
#define	AUTOSAVE_HVMD_FILE	AUTOSAVE_HVMD_NAME "." AUTOSAVE_FILETYPE
#define	AUTOSAVE_SPPRIMD_FILE	AUTOSAVE_SPPRI_NAME "." AUTOSAVE_FILETYPE
#define	AUTOSAVE_PRIMARY_FILE	AUTOSAVE_GPREFIX "primary" "." AUTOSAVE_FILETYPE
#define	AUTOSAVE_TS_FILE	"timestamp"
#define	AUTOSAVE_SP_TS_FILE	"sp-timestamp"

#define	AUTOSAVE_MDSETDIR(mdset_name)	zsprintf("%s/%s%s", AUTOSAVE_ROOTDIR, \
	AUTOSAVE_MDPREFIX, mdset_name)
#define	AUTOSAVE_FGPREFIX(mdset_name)	zsprintf("%s/%s%s/%s", \
	AUTOSAVE_ROOTDIR, AUTOSAVE_MDPREFIX, mdset_name, AUTOSAVE_GPREFIX)
#define	AUTOSAVE_FNAME(mdset_name, dpname)	zsprintf("%s/%s%s/%s", \
	AUTOSAVE_ROOTDIR, AUTOSAVE_MDPREFIX, mdset_name, dpname)
#define	AUTOSAVE_FILENAME(mdset_name, prefix, postfix)	\
	zsprintf("%s/%s%s/%s%s.%s", AUTOSAVE_ROOTDIR, AUTOSAVE_MDPREFIX, \
	mdset_name, prefix, postfix, AUTOSAVE_FILETYPE)
#define	AUTOSAVE_JOINNAME(dname, dp)	zsprintf("%s/%s", dname, dp)
#define	AUTOSAVE_TIMESTAMP_FILE(mdset_name)	zsprintf("%s/%s%s/%s", \
	AUTOSAVE_ROOTDIR, AUTOSAVE_MDPREFIX, mdset_name, AUTOSAVE_TS_FILE)
#define	AUTOSAVE_SP_TIMESTAMP_FILE(mdset_name)	zsprintf("%s/%s%s/%s", \
	AUTOSAVE_ROOTDIR, AUTOSAVE_MDPREFIX, mdset_name, AUTOSAVE_SP_TS_FILE)

static int autosave_create_file(char *mdset_name, char *fname);

/* State variables */
static bool_t autosave_doing_rm_autosave_spconfig = false;
static bool_t autosave_do_autorecovery = false;
static bool_t autosave_mdstore_isready = false;
static bool_t autosave_allow_updates_flag = false;


/*
 * Header structure for each MD file.
 */
#define	AFM_MAGIC		0xc1f2afde
#define	AFM_VERSION(maj, min)	(((maj)<<16)|(min))
#define	AFM_MAX_DOMAIN_NAME	MAXLEN_OBJECT_NAME	/* Domain name */

typedef struct {
	uint32_t	afm_magic;	/* AFM_MAGIC */
	uint32_t	afm_version;	/* AFM version # */
	uint32_t	afm_mdtype;	/* MD type */
	uint32_t	afm_size;	/* MD size */
	uint64_t	afm_timestamp;	/* MD timestamp */
	uint64_t	afm_chksum;	/* MD chksum */
	uint64_t	afm_mdaddr;	/* MD addr */
	uint8_t		afm_name[AFM_MAX_DOMAIN_NAME];
} autosave_mdhdr_t;

/*
 * Test if spconfig name is an internal one or not.
 */
#define	SPCONFIG_IS_INTERNAL(name)	(name[0] == '_')

static const char default_sp_config[] = "_Default_SP_Config";
static const char delayed_reconfig_backup[] = "_Delayed_Reconfig_Backup";

static char *
autosave_get_spconfig(void)
{
	if (zeus.current_spconfig != NULL)
		return (zeus.current_spconfig);

	return ((char *)default_sp_config);
}

/*
 * Free an array of ptrs to strings.
 */
void
autosave_free_list(char **listp, int nlist)
{
	int i;

	if (listp == NULL)
		return;

	for (i = 0; i < nlist; i++) {
		Xfree(listp[i]);
	}
	Xfree(listp);
}

/*
 * Get the "mdstore" module function 'funcname'
 */
void *
mdstore_get_function(char *funcname, response_t *resp)
{
	static int msg_out = 0;
	void *function;

	function = ds_channel_service_routine(ds_sp_channel(), "mdstore",
	    funcname);

	/* Allow re-outputting error message if communication was restored. */
	if (function != NULL && msg_out == 1)
		msg_out = 0;

	if (function == NULL && msg_out == 0) {
		rprintf(resp, MSG_ERR_AUTOSAVE_MDSTORE_OPERATION);
		msg_out = 1;
	}
	return (function);
}

/*
 * We call mdstore:mdstore("remove") in both the 'ldm rm-spconfig <cfg>'
 * and the 'ldm add-spconfig -r <cfg> [<newcfg>] cases.  At the
 * end of mdstore:mstore("remove") we call autosave_remove_spconfig_update.
 * In the first case, we want to remove the corresponding autosave and in
 * the second we don't.  This flag indicates that we're doing the
 * mdstore:mdstore("remove") call from the second case (within
 * autosave_remove_spconfig).
 */

/*
 * Remove an spconfig.
 * This is needed because when we do an "ldm add-spconfig -r <cfg> [<newcfg>]"
 * if the spconfig to be downloaded already exists on the SP, the download would
 * fail.
 */
void
autosave_remove_spconfig(char *mdset_name, response_t *resp)
{
	typedef int (*mdstore_f)(char *, char *, response_t *, xmlNodePtr);
	mdstore_f mdstore;

	mdstore = (mdstore_f)mdstore_get_function("mdstore", resp);
	if (mdstore != NULL) {
		autosave_doing_rm_autosave_spconfig = true;
		(void) (*mdstore)("remove", mdset_name, NULL, NULL);
		autosave_doing_rm_autosave_spconfig = false;
	}
}

static uint64_t
autosave_md_chksum(uint8_t *buf, int size)
{
	uint64_t chksum = 0;
	uint64_t last;
	int ns = size / sizeof (uint64_t);
	uint64_t *dp = (uint64_t *)buf;
	int i;

	for (i = 0; i < ns; i++)
		chksum += *dp++;
	size -= ns * sizeof (uint64_t);
	if (size) {
		last = 0;
		memcpy(&last, dp, sizeof (uint64_t) - size);
		chksum += last;
	}
	return (chksum);
}

/*
 * Open, check/read the AFM header, then read the data of an MD file.
 */
static int
autosave_read_md_file(char *filename, int *mdtypep, uint64_t *mdaddrp,
    uint8_t **bufp, int *sizep)
{
	autosave_mdhdr_t afm;
	int size;
	uint8_t *buf;
	int ret;
	uint64_t chksum;
	int fd;

	if ((fd = open(filename, O_RDONLY)) < 0) {
		DBG(zprintf("%s: %s: Can't open\n", __func__, filename));
		return (-1);
	}
	if (full_read(fd, (uint8_t *)&afm, sizeof (afm)) != sizeof (afm)) {
		DBG(zprintf("%s: %s: Can't read header\n", __func__, filename));
		close(fd);
		return (-1);
	}
	if (afm.afm_magic != AFM_MAGIC) {
		DBG(zprintf("%s: %s: Bad magic number\n", __func__, filename));
		close(fd);
		return (-1);
	}
	if (afm.afm_version != AFM_VERSION(1, 0)) {
		DBG(zprintf("%s: %s: Bad version number 0x%x\n", __func__,
		    filename, afm.afm_version));
		close(fd);
		return (-1);
	}

	DBG(zprintf("%s: %s: type: %d size: %d mdaddr: 0x%"PRIx64
	    " chksum: 0x%"PRIx64" name: '%s'\n", __func__, filename,
	    afm.afm_mdtype, afm.afm_size, afm.afm_mdaddr, afm.afm_chksum,
	    afm.afm_name));

	size = afm.afm_size;
	buf = Xmalloc(size);
	ret = full_read(fd, buf, size);
	if (ret != size) {
		DBG(zprintf("%s: %s: Bad buf read\n", __func__, filename));
		Xfree(buf);
		close(fd);
		return (-1);
	}
	chksum = autosave_md_chksum(buf, size);
	if (afm.afm_chksum != chksum) {
		DBG(zprintf("%s: %s: chksum fail is: 0x%"PRIx64" sb: "
		    "0x%"PRIx64"\n", __func__, filename, chksum,
		    afm.afm_chksum));
		close(fd);
		Xfree(buf);
		return (-1);
	}
	*mdtypep = afm.afm_mdtype;
	*mdaddrp = afm.afm_mdaddr;
	*bufp = buf;
	*sizep = size;
	return (0);
}

static int
autosave_exists(char *mdset_name)
{
	struct stat statb;
	char *dirname = AUTOSAVE_MDSETDIR(mdset_name);
	int exists;

	exists = (stat(dirname, &statb) == 0 && (statb.st_mode & S_IFDIR) != 0);
	Xfree(dirname);
	return (exists);
}

/*
 * For a given autosave config (mdset_name), return a list of all the MD
 * files corresponding to that config.  Make sure that the config is
 * complete by checking for a 'primary' domain MD, an HV MD and an SP PRI MD.
 */
static int
autosave_get_mdset_mds(char *mdset_name, int *nmdsp, char ***filenamesp)
{
	char *dirname;
	char **filenames = NULL;
	char *filename;
	int nmds;
	DIR *dirp;
	struct dirent *dp;
	int err = 0;
	uint64_t mdaddr;
	int mdtype;
	uint8_t *buf;
	int size;
	bool_t have_hv = false;
	bool_t have_sp_pri = false;
	bool_t have_primary = false;

	if (!autosave_exists(mdset_name)) {
		DBG(zprintf("%s: '%s' does not exist\n", __func__, mdset_name));
		return (-1);
	}

	dirname = AUTOSAVE_MDSETDIR(mdset_name);

	/*
	 * scan through the directory and collect file names as we go.
	 */
	if ((dirp = opendir(dirname)) == NULL) {
		DBG(zprintf("%s: can't opendir: %s\n", __func__, dirname));
		Xfree(dirname);
		return (-1);
	}

	nmds = 0;

	while ((dp = readdir(dirp)) != NULL) {
		if (strcmp(dp->d_name, ".") == 0 ||
		    strcmp(dp->d_name, "..") == 0)
			continue;
		if ((strncmp(dp->d_name, AUTOSAVE_GPREFIX,
		    AUTOSAVE_GPREFIXLEN) == 0 ||
		    strcmp(dp->d_name, AUTOSAVE_HVMD_FILE) == 0 ||
		    strcmp(dp->d_name, AUTOSAVE_SPPRIMD_FILE) == 0)) {
			filename = AUTOSAVE_JOINNAME(dirname, dp->d_name);

			/*
			 * We read the autosave file here to make sure it's
			 * not corrupt.
			 */
			if ((err = autosave_read_md_file(filename, &mdtype,
			    &mdaddr, &buf, &size)) != 0) {
				warning("Autosave config '%s': corrupted file: "
				    "'%s'", mdset_name, filename);
				Xfree(filename);
				break;
			}
			Xfree(buf);
			nmds++;
			if (filenamesp != NULL) {
				filenames = Xrealloc(filenames,
				    nmds * sizeof (char *));
				filenames[nmds - 1] = filename;
			}
			if (strcmp(dp->d_name, AUTOSAVE_HVMD_FILE) == 0)
				have_hv = true;
			else if (strcmp(dp->d_name, AUTOSAVE_SPPRIMD_FILE) == 0)
				have_sp_pri = true;
			else if (strcmp(dp->d_name, AUTOSAVE_PRIMARY_FILE) == 0)
				have_primary = true;
		}
	}

	closedir(dirp);

	Xfree(dirname);

	if (err != 0 || nmds == 0 || !have_hv || !have_sp_pri ||
	    !have_primary) {
		if (err == 0 && nmds != 0) {
			if (!have_hv) {
				warning("Autosave config '%s' missing HV MD",
				    mdset_name);
			} else if (!have_sp_pri) {
				warning("Autosave config '%s' missing SP PRI "
				    "MD", mdset_name);
			} else if (!have_primary) {
				warning("Autosave config '%s' missing primary "
				    "MD", mdset_name);
			}
		}
		autosave_free_list(filenames, nmds);
		return (-1);
	}
	if (filenamesp != NULL)
		*filenamesp = filenames;
	*nmdsp = nmds;
	return (0);
}

/*
 * Collect all the valid autosave config names in a list.  This scans
 * /var/opt/SUNWldm looking for "autosave-<name>" directories.
 */
int
autosave_get_mdsets(char *mdset_name, int *nmdsetsp, char ***mdset_namesp)
{
	char **mdset_names = NULL;
	int nmdsets = 0;
	DIR *dirp;
	struct dirent *dp;
	char *tmdset_name;
	int nmds;

	/*
	 * scan through the directory and collect file names as we go.
	 */
	if ((dirp = opendir(AUTOSAVE_ROOTDIR)) == NULL) {
		DBG(zprintf("%s: can't opendir: %s\n", __func__,
		    AUTOSAVE_ROOTDIR));
		return (-1);
	}

	while ((dp = readdir(dirp)) != NULL) {
		if (strcmp(dp->d_name, ".") == 0 ||
		    strcmp(dp->d_name, "..") == 0)
			continue;
		if (strncmp(dp->d_name, AUTOSAVE_MDPREFIX,
		    AUTOSAVE_MDPREFIXLEN) == 0 &&
		    strlen(dp->d_name) > AUTOSAVE_MDPREFIXLEN) {
			tmdset_name = dp->d_name + AUTOSAVE_MDPREFIXLEN;
			/*
			 * Suppress the display of internal autosave
			 * names.
			 */
			if (SPCONFIG_IS_INTERNAL(tmdset_name))
				continue;
			if (mdset_name == NULL ||
			    strcmp(mdset_name, tmdset_name) == 0) {
				if (autosave_get_mdset_mds(tmdset_name, &nmds,
				    NULL) != 0)
					continue;
				nmdsets += 1;
				mdset_names = Xrealloc(mdset_names,
				    nmdsets * sizeof (char *));
				mdset_names[nmdsets - 1] = Xstrdup(tmdset_name);
			}
		}
	}

	closedir(dirp);

	*nmdsetsp = nmdsets;
	*mdset_namesp = mdset_names;
	return (0);
}

static int
autosave_clean_dir(char *mdset_name, bool_t keep_sp_timestamp)
{
	DIR *dirp;
	struct dirent *dp;
	char *dirname;
	char *filename;
	int err = 0;

	dirname = AUTOSAVE_MDSETDIR(mdset_name);

	/*
	 * scan through the directory and unlink files as we go.  If
	 * we can't unlink something, we return an error.
	 */
	if ((dirp = opendir(dirname)) == NULL) {
		DBG(zprintf("%s: can't opendir: %s\n", __func__, dirname));
		Xfree(dirname);
		return (-1);
	}

	while (err == 0 && (dp = readdir(dirp)) != NULL) {
		if (strcmp(dp->d_name, ".") == 0 ||
		    strcmp(dp->d_name, "..") == 0)
			continue;
		if (keep_sp_timestamp &&
		    strcmp(dp->d_name, AUTOSAVE_SP_TS_FILE) == 0)
			continue;
		filename = AUTOSAVE_JOINNAME(dirname, dp->d_name);
		if ((err = unlink(filename)) != 0) {
			DBG(zprintf("%s: can't unlink: %s\n", __func__,
			    filename));
		}
		Xfree(filename);
	}
	closedir(dirp);
	Xfree(dirname);
	return (err);
}

/*
 * remove a directory and everything underneath it.
 */
int
autosave_rmdir(char *mdset_name, bool_t keep_sp_timestamp)
{
	struct stat statb;
	char *dirname;
	int ret;

	dirname = AUTOSAVE_MDSETDIR(mdset_name);

	/*
	 * Check if directory exists, if it doesn't, we're done.  If we
	 * get some other error from stat, we've got a problem.
	 */
	if (stat(dirname, &statb) != 0) {
		if (errno == ENOENT)
			return (1);
		DBG(zprintf("%s: can't stat: %s\n", __func__, dirname));
		Xfree(dirname);
		return (-1);
	}

	/*
	 * not a directory, simply unlink it.
	 */
	if ((statb.st_mode & S_IFDIR) == 0) {
		if (unlink(dirname) != 0) {
			DBG(zprintf("%s: can't unlink: %s\n", __func__,
			    dirname));
			Xfree(dirname);
			return (-1);
		}
		Xfree(dirname);
		return (0);
	}

	autosave_clean_dir(mdset_name, keep_sp_timestamp);

	/*
	 * Try to delete the directory.  If the SP timestamp is left
	 * intact from keep_sp_timestamp, rmdir will fail with the
	 * EEXIST error, so ignore the error in that case.
	 */
	ret = rmdir(dirname);
	if (keep_sp_timestamp == true && ret == -1 && errno == EEXIST) {
		ret = 0;
	} else if (ret != 0) {
		DBG(zprintf("%s: can't rmdir: %s\n", __func__, dirname));
	}
	Xfree(dirname);
	return (ret);
}

/*
 * copy a directory and everything underneath it.
 */
static int
autosave_copy_dir(char *mdset_name, char *new_mdset_name)
{
	char *from_dirname;
	char *to_dirname;
	char *from_filename;
	char *to_filename;
	int from_fd;
	int to_fd;
	DIR *dirp;
	struct dirent *dp;
	uint8_t buf[BUFSIZ];
	int n;
	int err = 0;

	/*
	 * Check if autosave exists.
	 */
	if (!autosave_exists(mdset_name)) {
		DBG(zprintf("%s: autosave '%s' doesn't exist\n", __func__,
		    mdset_name));
		return (-1);
	}

	autosave_clean_dir(new_mdset_name, false);

	from_dirname = AUTOSAVE_MDSETDIR(mdset_name);

	/*
	 * scan through the directory and copy files as we go.
	 */
	if ((dirp = opendir(from_dirname)) == NULL) {
		DBG(zprintf("%s: can't opendir: %s\n", __func__, from_dirname));
		Xfree(from_dirname);
		return (-1);
	}

	to_dirname = AUTOSAVE_MDSETDIR(new_mdset_name);

	while ((dp = readdir(dirp)) != NULL) {
		if (strcmp(dp->d_name, ".") == 0 ||
		    strcmp(dp->d_name, "..") == 0)
			continue;
		from_filename = AUTOSAVE_JOINNAME(from_dirname, dp->d_name);
		to_filename = AUTOSAVE_JOINNAME(to_dirname, dp->d_name);
		from_fd = open(from_filename, O_RDONLY);
		to_fd = autosave_create_file(new_mdset_name, to_filename);
		if (from_fd >= 0 && to_fd >= 0) {
			while ((n = full_read(from_fd, buf, sizeof (buf))) > 0)
				if (full_write(to_fd, buf, n) != n)
					err = -1;
			if (n < 0)
				err = -1;
		} else
			err = -1;
		close(from_fd);
		close(to_fd);
		Xfree(from_filename);
		Xfree(to_filename);
	}

	closedir(dirp);

	Xfree(from_dirname);
	Xfree(to_dirname);

	return (err);
}

static int
autosave_rename_dir(char *old_mdset_name, char *new_mdset_name)
{
	char *olddir = AUTOSAVE_MDSETDIR(old_mdset_name);
	char *newdir = AUTOSAVE_MDSETDIR(new_mdset_name);
	int err;

	err = rename(olddir, newdir);
	if (err) {
		DBG(zprintf("Can't rename autosave directory '%s' to '%s'\n",
		    olddir, newdir));
	}
	Xfree(olddir);
	Xfree(newdir);
	return (err);
}

bool_t
autosave_rename_default(char *mdset_name, response_t *resp)
{
	if (!autosave_exists((char *)default_sp_config)) {
		/*
		 * The default SP config autosave does not
		 * exist, which can happen if a previous rm-spconfig
		 * was done and no intervening config change has
		 * happened.  Return true so that we update the
		 * SP timestamp file.
		 */
		return (true);
	}

	(void) autosave_rmdir(mdset_name, false);

	if (autosave_rename_dir((char *)default_sp_config, mdset_name) != 0) {
		rprintf(resp, MSG_ERR_AUTOSAVE_CANT_RENAME, default_sp_config,
		    mdset_name, default_sp_config);
		return (false);
	}

	return (true);
}

/*
 * Download an autosave MD to the SP.
 * This requires that we get the mdstore module "download_md" function.
 */
static int
autosave_download_md(char *mdset_name, char *new_mdset_name, char *filename,
    response_t *resp)
{
	int mdtype;
	uint64_t mdaddr;
	uint8_t *buf;
	int size;
	int err = 0;
	typedef int (*download_md_t)(void *, uint32_t, uint32_t, uint64_t,
	    response_t *);
	download_md_t download_md;
	md_t *mdp;

	download_md = (download_md_t)mdstore_get_function("download_md", resp);
	if (download_md == NULL) {
		return (-1);
	}

	if (autosave_read_md_file(filename, &mdtype, &mdaddr, &buf, &size) != 0)
		return (-1);

	/*
	 * If we're downloading the autosave to a new spconfig name, then
	 * we need to fixup the HV MD's "current_spconfig" property in the
	 * copy we download to the MD.
	 */
	if (mdtype == HV_MD_TYPE && strcmp(mdset_name, new_mdset_name) != 0) {
		mdp = md_scan_bin(buf, size);
		Xfree(buf);
		if (mdp == NULL) {
			warning("Failed reading MD from %s", filename);
			return (-1);
		}

		md_update_string_property(mdp, md_root_node(mdp),
		    "current_spconfig", new_mdset_name);

		size = md_gen_bin(mdp, &buf);
		md_fini(mdp);
	}

	DBG(zprintf("%s: %s: downloading: type: 0x%x size: %d mdaddr: 0x%"
	    PRIx64"\n", __func__, filename, mdtype, size, mdaddr));

	err = (*download_md)(buf, mdtype, size, mdaddr, resp);
	Xfree(buf);
	if (err) {
		warning("%s: %s: download_md mdtype: 0x%x size: %d "
		    "mdaddr: %"PRIx64"\n", __func__, filename, mdtype, size,
		    mdaddr);
	}
	return (err);
}

int
autosave_download_mdset(char *mdset_name, char *new_mdset_name,
    response_t *resp)
{
	char **filenames;
	char *guest_prefix;
	int guest_prefixlen;
	int err;
	int nmds;
	int i;
	int hvmd_id = -1;
	int sppri_id = -1;
	typedef int (*bracket_transfer_t)(int, char *, int, response_t *);
	bracket_transfer_t bracket_transfer;
	char *orig_spconfig = NULL;
	bool_t orig_is_next_spconfig;

	/*
	 * If the mdset is an internal name, don't download it.
	 */
	if (SPCONFIG_IS_INTERNAL(mdset_name))
		return (-1);

	if (!autosave_exists(mdset_name)) {
		rprintf(resp, MSG_ERR_AUTOSAVE_NO_SUCH_CONFIG, mdset_name);
		return (-1);
	}

	/*
	 * Do this first to ensure that we can communicate to the SP.
	 */
	bracket_transfer = (bracket_transfer_t)
	    mdstore_get_function("bracket_transfer", resp);
	if (bracket_transfer == NULL) {
		return (-1);
	}

	/*
	 * Get the autosave file names, autosave_get_mdset_mds also
	 * verifies the config, so we only need to do minimal checking
	 * here.
	 */
	if (autosave_get_mdset_mds(mdset_name, &nmds, &filenames) != 0) {
		rprintf(resp, MSG_ERR_AUTOSAVE_CONFIG_INVALID, mdset_name);
		return (-1);
	}

	guest_prefix = AUTOSAVE_FGPREFIX(mdset_name);
	guest_prefixlen = strlen(guest_prefix);

	/*
	 * Check filename array for HV MD, SP PRI and Guests
	 */
	for (i = 0; i < nmds; i++) {
		if (strstr(filenames[i], AUTOSAVE_HVMD_FILE) != 0)
			hvmd_id = i;
		else if (strstr(filenames[i], AUTOSAVE_SPPRIMD_FILE) != 0)
			sppri_id = i;
		else if (strncmp(filenames[i], guest_prefix,
		    guest_prefixlen) != 0) {
			err = -1;
			DBG(zprintf("%s: Invalid filename: '%s'\n", __func__,
			    filenames[i]));
			goto err_handler;
		}
	}

	if (hvmd_id < 0 || sppri_id < 0) {
		rprintf(resp, MSG_ERR_AUTOSAVE_CONFIG_INVALID, mdset_name);
		goto err_handler;
	}

	DBG(zprintf("%s: Downloading %s to %s nmds: %d:\n", __func__,
	    mdset_name, new_mdset_name, nmds));

	if (zeus.current_spconfig != NULL) {
		orig_spconfig = Xstrdup(zeus.current_spconfig);
		orig_is_next_spconfig = zeus.is_next_spconfig;
	}

	/*
	 * Remove mdset for the 'ldm add-spconfig -r <cfg>' case.
	 */
	if (strcmp(mdset_name, new_mdset_name) == 0) {
		autosave_remove_spconfig(mdset_name, resp);
	}

	if ((err = (*bracket_transfer)(MDSET_BEGIN_REQUEST, new_mdset_name,
	    nmds, resp)) != 0) {
		DBG(zprintf("%s: bracket_transfer: %s nmds: %d failed\n",
		    __func__, new_mdset_name, nmds));
		goto err_handler;
	}

	/*
	 * Download Guest MDs
	 */
	for (i = 0; i < nmds; i++) {
		if (i == hvmd_id || i == sppri_id)
			continue;

		if ((err = autosave_download_md(mdset_name, new_mdset_name,
		    filenames[i], resp)) != 0) {
			goto err_handler;
		}
	}

	/*
	 * Download HV MD
	 */
	if ((err = autosave_download_md(mdset_name, new_mdset_name,
	    filenames[hvmd_id], resp)) != 0) {
		goto err_handler;
	}

	/*
	 * Downwload SP PRI
	 */
	if ((err = autosave_download_md(mdset_name, new_mdset_name,
	    filenames[sppri_id], resp)) != 0) {
		goto err_handler;
	}

	err = (*bracket_transfer)(MDSET_END_REQUEST, NULL, nmds, resp);

	/*
	 * If we've downloaded to a new_mdset_name, then we
	 * copy the original mdset autosave to the new_mdset_name
	 * autosave.
	 */
	if (strcmp(mdset_name, new_mdset_name) != 0) {
		if (autosave_copy_dir(mdset_name, new_mdset_name) < 0)
			rprintf(resp, MSG_ERR_AUTOSAVE_CANT_COPY,
			    new_mdset_name, mdset_name, new_mdset_name);
	}

	/*
	 * Update the SP timestamp of the config we just downloaded.
	 */
	(void) autosave_set_spconfig_timestamp(new_mdset_name);

	/*
	 * Restore the original mdset as the "current" spconfig.
	 */
	if (orig_spconfig != NULL) {
		if (zeus.current_spconfig != NULL)
			Xfree(zeus.current_spconfig);
		zeus.current_spconfig = Xstrdup(orig_spconfig);
		zeus.is_next_spconfig = orig_is_next_spconfig;
	}

err_handler:

	Xfree(guest_prefix);

	if (orig_spconfig != NULL)
		Xfree(orig_spconfig);

	autosave_free_list(filenames, nmds);
	return (err);
}

/*
 * Create a file under the /var/opt/SUNWldm/autosave-<mdset_name> directory.
 * Create the directory if it doesn't already exist.
 */
static int
autosave_create_file(char *mdset_name, char *fname)
{
	char *dirname = AUTOSAVE_MDSETDIR(mdset_name);
	struct stat statb;
	int fd;

	if (stat(dirname, &statb) != 0) {
		if (mkdir(dirname, 0700) != 0) {
			warning("Can't mkdir '%s'\n", dirname);
		}
	}
	Xfree(dirname);
	if ((fd = open(fname, O_CREAT|O_TRUNC|O_RDWR, 0600)) < 0) {
		warning("Can't create '%s'\n", fname);
	}
	return (fd);
}

/*
 * Wrapper to return the "current" timestamp, use time(2) for now.
 */
static uint64_t
autosave_curtimestamp()
{
	return (time(NULL));
}

/*
 * Get a timestamp value from a timestamp file.
 */
static int
autosave_get_timestamp(char *filename, uint64_t *timestampp)
{
	int fd;
	uint64_t timestamp;

	if ((fd = open(filename, O_RDONLY)) < 0) {
		return (-1);
	}
	if (full_read(fd, (uint8_t *)&timestamp,
	    sizeof (timestamp)) != sizeof (timestamp)) {
		close(fd);
		return (-1);
	}
	close(fd);
	*timestampp = timestamp;
	return (0);
}

/*
 * Set a timestamp value in a timestamp file, creating the file if necessary.
 * Will create the /var/opt/SUNWldm/autosave-<mdset_name> directory if it
 * doesn't exit.
 */
static int
autosave_set_timestamp(char *mdset_name, char *filename, uint64_t timestamp)
{
	int fd;

	if ((fd = autosave_create_file(mdset_name, filename)) < 0) {
		return (-1);
	}

	if (full_write(fd, (uint8_t *)&timestamp,
	    sizeof (timestamp)) != sizeof (timestamp)) {
		close(fd);
		return (-1);
	}
	close(fd);
	return (0);
}

/*
 * Set the autosave timestamp for this mdset_name.
 */
static int
autosave_set_autosave_timestamp(char *mdset_name, uint64_t timestamp)
{
	char *filename = AUTOSAVE_TIMESTAMP_FILE(mdset_name);
	int ret;

	ret = autosave_set_timestamp(mdset_name, filename, timestamp);
	Xfree(filename);
	return (ret);
}

/*
 * Get the autosave timestamp for the given mdset_name.
 */
static int
autosave_get_autosave_timestamp(char *mdset_name, uint64_t *timestampp)
{
	char *filename = AUTOSAVE_TIMESTAMP_FILE(mdset_name);
	int ret;

	ret = autosave_get_timestamp(filename, timestampp);
	Xfree(filename);
	return (ret);
}

/*
 * Handle 'ldm rm-spconfig <cfg>'
 * If the user removes an spconfig, delete its corresponding autosave also.
 * This is called via mdstore:mdstore("remove").
 */
void
autosave_remove_spconfig_update(char *mdset_name)
{
	if (autosave_doing_rm_autosave_spconfig == false) {
		(void) autosave_rmdir(mdset_name, false);
	}
}

/*
 * Interface for setting the current timestamp for an spconfig.
 */
int
autosave_set_spconfig_timestamp(char *mdset_name)
{
	char *filename = AUTOSAVE_SP_TIMESTAMP_FILE(mdset_name);
	uint64_t timestamp;
	int ret;

	timestamp = autosave_curtimestamp();
	ret = autosave_set_timestamp(mdset_name, filename, timestamp);
	Xfree(filename);
	return (ret);
}

/*
 * Interface for gettng the current timestamp for an spconfig.
 * If the timestamp doesn't exist, it is created with the current time.
 */
static int
autosave_get_spconfig_timestamp(char *mdset_name, uint64_t *timestampp)
{
	char *filename = AUTOSAVE_SP_TIMESTAMP_FILE(mdset_name);
	int ret;

	ret = autosave_get_timestamp(filename, timestampp);
	if (ret) {
		/* getting timestamp failed, try to update it to now. */
		ret = autosave_set_spconfig_timestamp(mdset_name);
		if (ret == 0) {
			ret = autosave_get_timestamp(filename, timestampp);
		}
	}
	Xfree(filename);
	return (ret);
}

/*
 * Generic generate autosave MD file.
 */
static int
autosave_gen_md(char *mdset_name, char *fname, char *md_name, int mdtype,
    uint64_t mdaddr, uint8_t *buf, int size)
{
	autosave_mdhdr_t afm;
	int fd;

	if ((fd = autosave_create_file(mdset_name, fname)) < 0) {
		return (-1);
	}

	/*
	 * Write header
	 */
	afm.afm_magic = AFM_MAGIC;
	afm.afm_version = AFM_VERSION(1, 0);
	afm.afm_mdtype = mdtype;
	afm.afm_mdaddr = mdaddr;
	afm.afm_size = size;
	afm.afm_timestamp = autosave_curtimestamp();
	afm.afm_chksum = autosave_md_chksum(buf, size);
	strncpy((char *)afm.afm_name, md_name, AFM_MAX_DOMAIN_NAME);
	if (full_write(fd, (uint8_t *)&afm, sizeof (afm)) != sizeof (afm)) {
		DBG(zprintf("%s: %s: Bad header write\n", __func__, fname));
		return (-1);
	}

	/*
	 * Write MD
	 */
	if (full_write(fd, buf, size) != size) {
		DBG(zprintf("%s: %s: Bad buf write size %d\n", __func__,
		    fname, size));
		return (-1);
	}

	close(fd);

	/*
	 * Only update the autosave timestamp if we're updating the HV MD.
	 * We always update the guest MD's followed by the HV MD.
	 */
	if (mdtype == HV_MD_TYPE) {
		autosave_set_autosave_timestamp(mdset_name, afm.afm_timestamp);
	}

	DBG(zprintf("%s: %s: %s: type: 0x%x size: %d mdaddr: %"PRIx64
	    " chksum: 0x%"PRIx64"\n", __func__, fname, md_name, afm.afm_mdtype,
	    afm.afm_size, afm.afm_mdaddr, afm.afm_chksum));
	return (0);
}

/*
 * Interface for generating guest MD autosave file.
 */
void
autosave_gen_guest_md(ldom_t *ldp, uint64_t mdaddr, uint8_t *buf, int size)
{
	char *filename;
	int mdtype;
	int err;
	char *mdset_name = autosave_get_spconfig();

	filename = AUTOSAVE_FILENAME(mdset_name, AUTOSAVE_GPREFIX, ldp->namep);

	mdtype = IS_CONTROL_DOMAIN(ldp) ? CTL_DOM_MD_TYPE : 0;

	/* Does domain own any I/O? */
	if (CHAIN_LENGTH(ldp->bound.iodevices) > 0)
		mdtype |= SVC_DOM_MD_TYPE;

	err = autosave_gen_md(mdset_name, filename, ldp->namep, mdtype, mdaddr,
	    buf, size);

	if (err) {
		warning("Autosave config '%s' can't generate guest MD file: "
		    "'%s'", mdset_name, filename);
	}

	Xfree(filename);
}

/*
 * Interface for generating HV MD autosave file.
 */
void
autosave_gen_hv_md(uint64_t mdaddr, uint8_t *buf, int size)
{
	char *filename;
	int err;
	char *mdset_name = autosave_get_spconfig();

	filename = AUTOSAVE_FILENAME(mdset_name, AUTOSAVE_HVMD_NAME, "");

	err = autosave_gen_md(mdset_name, filename, AUTOSAVE_HVMD_NAME,
	    HV_MD_TYPE, mdaddr, buf, size);

	if (err) {
		warning("Autosave config '%s' can't generate HV MD file: '%s'",
		    mdset_name, filename);
		Xfree(filename);
		return;
	}

	Xfree(filename);
}

/*
 * Interface for generating SP PRI MD autosave file.
 */
void
autosave_gen_sp_pri_md(uint8_t *buf, int size)
{
	char *filename;
	int err;
	char *mdset_name = autosave_get_spconfig();

	filename = AUTOSAVE_FILENAME(mdset_name, AUTOSAVE_SPPRI_NAME, "");

	err = autosave_gen_md(mdset_name, filename, AUTOSAVE_SPPRI_NAME,
	    PRI_TYPE, 0, buf, size);

	if (err) {
		warning("Autosave config '%s' can't generate SP PRI MD file: "
		    "'%s'", mdset_name, filename);
	}

	Xfree(filename);
}

void
autosave_update_autosave()
{
	typedef int (*mdstore_download_f)(char *, response_t *, bool_t, bool_t);
	mdstore_download_f mdstore_download;
	response_t *resp;
	char *resp_text;
	char *mdset_name = autosave_get_spconfig();

	/*
	 * Don't allow updates in startup.
	 */
	if (autosave_allow_updates_flag == false)
		return;

	autosave_clean_dir(mdset_name, true);

	init_response(&resp);
	mdstore_download = (mdstore_download_f)
	    mdstore_get_function("mdstore_download", resp);
	if (mdstore_download != NULL) {
		if ((*mdstore_download)(mdset_name, resp, false, true) == 0) {
			(void) autosave_set_autosave_timestamp(mdset_name,
			    autosave_curtimestamp());
		}
	}
	if ((resp_text = get_response_text(resp)) != NULL) {
		if (strlen(resp_text) > 0) {
			warning(resp_text);
		}
		Xfree(resp_text);
	}
	fini_response(resp);
}

/*
 * Returns true if autosave config is newer than sp config.
 */
bool_t
autosave_autosave_is_newer(char *mdset_name)
{
	uint64_t sp_timestamp;
	uint64_t auto_timestamp;

	if (autosave_get_spconfig_timestamp(mdset_name, &sp_timestamp) != 0) {
		return (false);
	}

	if (autosave_get_autosave_timestamp(mdset_name, &auto_timestamp) != 0) {
		/*
		 * autosave may not exist, so we quietly return.
		 */
		return (false);
	}

	return (auto_timestamp > sp_timestamp);
}

/*
 * The autorecovery_policy is one of the following providing the autosave config
 * corresponding to the current SP config is newer than the current SP
 * config:
 *  AUTORECOVERY_POLICY_LOG	= log a warning message.
 *  AUTORECOVERY_POLICY_NOTIFY	= return an appropriate message when
 *				the user tries the first CLI command or
 *				return an XML event.
 *  AUTORECOVERY_POLICY_RECOVER	= attempt to download the autosave
 *				config to the SP config on poweron.
 */
/*
 * Do the CLI/XML query for AUTORECOVERY_POLICY_NOTIFY.
 * Returns 1 if there's a warning message, 0 otherwise.
 * FIXME - XML support not implemented yet.
 */
/* ARGSUSED */
int
autosave_check_spconfig(response_t *resp, xmlNodePtr data)
{
	static int warn_out = 0;
	char *mdset_name = autosave_get_spconfig();

	if (option.autorecovery_policy != AUTORECOVERY_POLICY_NOTIFY ||
	    warn_out != 0)
		return (0);

	if (autosave_autosave_is_newer(mdset_name)) {
		if (resp) {
			warn_out = 1;
			rprintf(resp, MSG_ERR_AUTOSAVE_NOTIFY, mdset_name);
		}
	}
	return (warn_out);
}

/*
 * Function to perform AUTORECOVERY_POLICY_RECOVER policy, download the
 * autosave if it is newer.  This is called from the sequencer queue,
 * since some functions under autosave_download_mdset require being run
 * via the sequencer.
 */
static void
autosave_autorecovery()
{
	int err;
	response_t *resp;
	char *resp_text;
	char *mdset_name = autosave_get_spconfig();
	char *mdset_copy;

	/*
	 * Check for autosave being newer than current SP config.
	 */
	if (autosave_autosave_is_newer(mdset_name) == false)
		return;

	/*
	 * autosave_download_mdset will call mdstore:mdstore("remove")
	 * which will free zeus.current_spconfig, so mdset_name must be
	 * a copy of zeus.current_spconfig, not a pointer to it.
	 */
	mdset_copy = Xstrdup(mdset_name);
	init_response(&resp);
	err = autosave_download_mdset(mdset_copy, mdset_copy, resp);
	if ((resp_text = get_response_text(resp)) != NULL) {
		if (strlen(resp_text) > 0) {
			warning(resp_text);
		}
		Xfree(resp_text);
	}
	fini_response(resp);
	if (err) {
		warning("Autosave config '%s' recovery to SP config "
		    "failed\n", mdset_copy);
	} else {
		zprintf("Autosave config '%s' recovered to SP config\n"
		    "A powercycle is required for the updated SP "
		    "config to take effect\n",
		    mdset_copy);
	}
	Xfree(mdset_copy);
}

/*
 * The mdstore service may not be ready when we enter autosave_poweron,
 * so we have to deal with asynchronous notification when it is
 * registered.  If we try to access mdstore before it's registered,
 * it wedges.  We use these sequencer functions to access the state
 * variables so that there is no potential timing window in accessing them.
 */
/*
 * Set that the mdstore service is ready and enqueue to autorecovery
 * function if it was flagged.
 */
/* ARGSUSED */
static void
autosave_set_mdstore_ready(void *arg)
{
	autosave_mdstore_isready = true;
	if (autosave_do_autorecovery) {
		/*
		 * Ensure that autorecovery function is called at
		 * most once.
		 */
		autosave_do_autorecovery = false;
		autosave_autorecovery();
	}
}

/*
 * Set that we need to do the autorecovery function, and if mdstore
 * is already ready, enqueue the autorecovery function.
 */
/* ARGSUSED */
static void
autosave_set_autorecovery_ready(void *arg)
{
	autosave_do_autorecovery = true;
	if (autosave_mdstore_isready) {
		/*
		 * Ensure that autorecovery function is called at
		 * most once.
		 */
		autosave_do_autorecovery = false;
		autosave_autorecovery();
	}
}

/*
 * Enqueue a request to set mdstore to ready.
 */
void
autosave_notify_mdstore_ready()
{
	seq_enqueue_work(autosave_set_mdstore_ready, NULL, false);
}

/*
 * Interface for poweron processing.
 */
void
autosave_poweron(void)
{
	char *mdset_name = autosave_get_spconfig();

	/*
	 * Check for autosave being newer than current SP config.
	 */
	if (autosave_autosave_is_newer(mdset_name) == false)
		return;

	/*
	 * Check for autosave being internal name.
	 */
	if (SPCONFIG_IS_INTERNAL(mdset_name))
		return;

	switch (option.autorecovery_policy) {

	default:
		ASSERT(0);
		/* FALLTHROUGH */
	case AUTORECOVERY_POLICY_LOG:
		warning("Autosave config '%s' is newer than SP config\n",
		    mdset_name);
		break;

	case AUTORECOVERY_POLICY_NOTIFY:
		/* Wait until user issues CLI/XML command */
		break;

	case AUTORECOVERY_POLICY_RECOVER:
		seq_enqueue_work(autosave_set_autorecovery_ready, NULL, false);
		break;
	}
}

/* ARGSUSED */
static void
autosave_set_allow_updates(void *arg)
{
	autosave_allow_updates_flag = true;
}

void
autosave_allow_updates()
{
	seq_enqueue_work(autosave_set_allow_updates, NULL, false);
}


/*
 * Finishing a delayed reconfig, remove the delayed reconfig
 * backup directory.
 */
void
autosave_finish_delayed_reconfig()
{
	(void) autosave_rmdir((char *)delayed_reconfig_backup, false);
}

/*
 * Initiating a delayed reconfig, save the current autosave in
 * the delayed reconfig backup config.
 */
void
autosave_initiate_delayed_reconfig()
{
	char *mdset_name = autosave_get_spconfig();

	if (autosave_copy_dir(mdset_name, (char *)delayed_reconfig_backup) < 0)
		warning("Autosave can't copy '%s' to '%s'",
		    mdset_name, (char *)delayed_reconfig_backup);
}

/*
 * Cancelling a delayed reconfig, restore the delayed reconfig
 * backup config to the current config and remove it.
 */
void
autosave_cancel_delayed_reconfig()
{
	char *mdset_name = autosave_get_spconfig();

	if (autosave_copy_dir((char *)delayed_reconfig_backup, mdset_name) < 0)
		warning("Autosave can't copy '%s' to '%s'",
		    (char *)delayed_reconfig_backup, mdset_name);

	(void) autosave_rmdir((char *)delayed_reconfig_backup, false);
}
