/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_BACKEND_H_
#define	_BACKEND_H_

#ifdef __cplusplus
extern "C" {
#endif

#include "ldom.h"
#include "hvctl.h"

/*
 * Definitions for all Zeus hypervisor targets & backends
 *
 * A target is the specific hypervisor (usually related to a chip or chip
 * family that zeus needs to manage.
 *
 * A backend is the set of methods required to communicate with the specified
 * target. Typically it owns the methods used to deliver messages and receive
 * responses from a platform's hypervisor and/or service processor.
 */
/*
 * To indicate whether the HV-MD is a complete MD or a reconfiguration
 * MD
 */
typedef enum {
	HVMD,
	RECONF_HVMD
} hvmd_type_t;

/*
 * init			Called to initialize backend
 * start		Called to start backend, e.g. make contact with HV
 * process_option	Process backend specific options
 * get_pri		Get physical resource inventory from platform
 * init_pri		Initialize PRI (only used in simulation)
 * get_hv_state		Get current HV configuration
 * get_guest_md		Upload guest MD from HV
 * aug_guestmd		Augment the Guest MD for backend
 * dl_guestmd		Download a new MD for a guest
 * aug_hvmd		Augment the HV MD for backend
 * dl_hvmd		Download a new MD for the hypervisor
 * parse_hv_guest	Parse guest node in HV MD
 * aug_hv_guest		Augment guest node in HV MD
 * config_guests	(Re)Configure guests based on active list & new HV MD
 * start_new_guests	Kick off / boot new guests after configure
 * stop_guest		Capture the cpus of a guest
 * get_guest_state	Get the state of a guest
 * banner		Display any relevent version or info about backend
 * get_guest_utilization Get guest utilization
 * get_vcpu_state	Get state of virtual CPU
 * panic_guest		Panic a guest
 * read_mem		Read a range of physical memory
 * write_mem		Write to a range of physical memory
 * scrub_mem		Clear a range of physical memory
 * fini			Close backend ready for zeus exit
 */
typedef struct {
	char		*namep;
	char		*descriptionp;
	void		(*init)(void);
	void		(*start)(void);
	bool_t		(*process_option)(int *ip, int argc, char **argv);
	md_t		*(*get_pri)(void);
	void		(*init_pri)(md_t *pri);
	void		(*get_hv_state)(void);
	md_t		*(*get_guest_md)(ldom_t *ldp);
	void		(*aug_guestmd)(ldom_t *ldp, md_t *mdp);
	void		(*dl_guestmd)(ldom_t *ldp, uint64_t paddr,
				uint8_t *bufp, int size);
	void		(*aug_hvmd)(md_t *mdp);
	void		(*dl_hvmd)(uint64_t paddr, uint8_t *bufp,
				int size, hvmd_type_t type);
	void		(*parse_hv_guest)(ldom_t *ldp,
				md_node_t *guest, md_t *hvmd);
	void		(*aug_hv_guest)(ldom_t *ldp,
				md_node_t *guest, md_t *hvmd);
	void		(*config_guests)(config_flag_t flag);
	void		(*start_new_guests)(void);
	void		(*stop_guest)(ldom_t *ldp);
	int		(*get_guest_state)(ldom_t *ldp, int which,
				ldom_hv_state_t *state);
	void		(*banner)(response_t *);
	int		(*get_guest_utilization)(ldom_t *ldp,
				rs_guest_util_t *util);
	int		(*get_vcpu_state)(resid_t *vcpu,
				rs_vcpu_state_t *vstate);
	int		(*get_mem_state)(resid_t *mem,
				rs_mem_state_t *mstate);
	int		(*panic_guest)(ldom_t *ldp, response_t *resp);
	int		(*read_mem)(uint64_t pa, uint8_t *bufp, int len);
	int		(*write_mem)(uint64_t pa, uint8_t *bufp, int len);
	int64_t		(*scrub_mem)(uint64_t pa, uint64_t len);
	int		(*mem_dr_verify)(ldom_t *ldp, uint64_t paddr,
				uint64_t raddr, uint64_t size);
	void		(*fini)(void);
	uint64_t	magic;
} backend_t;

#define	BACKEND_MAGIC	((uint64_t)0x4261636b656e6400LL)	/* "Backend" */

extern backend_t	backend;

typedef void		be_instance_t;

extern bool_t		backend_select(char *);
extern void		backend_list(void);

#define	PRI_TEMPLATES	"pri-template"
#define	PRI_FILE	"pri-md.bin"
#define	HVMD_FILE	"1up-hvmd.bin"

/*
 * We define the target in the same file as the backend because they are
 * currently both always used in the same places.  Split this if this
 * situation ever changes.
 */

/*
 * init			Called to initialize target
 * process_option 	Handle target specific command line option
 * complete_setup 	Any remaining actions before reading HV state
 * gen_hvmd		Generate target specific hypervisor MD.
 * parse_hvmd		Parse the HVMD retrieved from a running HV
 * fini			Close target ready for zeus exit
 */

typedef struct {
	char		*namep;
	char		*descriptionp;
	void		(*init)(void);
	bool_t		(*process_option)(int *ip, int argc, char **argv);
	void		(*complete_setup)(void);
	md_t		*(*gen_hvmd)(bool_t include_pm_status);
	void		(*parse_hvmd)(md_t *);
	void		(*fini)(void);
	uint64_t	magic;
} target_t;

#define	TARGET_MAGIC	((uint64_t)0x4154617267657400LL)	/* "ATarget" */

extern target_t		target;

typedef void		tgt_instance_t;

extern bool_t		target_select(char *);
extern void		target_list(void);

#ifdef __cplusplus
}
#endif

#endif	/* _BACKEND_H_ */
