/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_BASICS_H_
#define	_BASICS_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <inttypes.h>
#include <sys/types.h>
#include <sys/byteorder.h>
#include <stdarg.h>
#include <string.h>

#define	INT_STR_LEN	16    /* enough for string representation of an int */
#define	LONG_STR_LEN	24    /* enough for string representation of a long */
#define	UINT64_STR_LEN	24    /* enough for string representation of a uint64 */

typedef enum {
	false = 0,
	true = 1
} bool_t;

extern uint_t debug_level;

#if ZDBG
#include <assert.h>

#define	ASSERT(_s)	assert(_s)

/*
 * Zeus Debugging Statements
 *
 * Statements enclosed in the DBG1, ..., DBG8 macros are executed iff the
 * corresponding n'th bit in debug_level variable is set.
 *
 * Statements enclosed in the DBG macro are executed if any of the bits 0-5
 * are set (i.e. does not include DBG7 or DBG8)
 *
 * The DBG7 macro is used to enable the memset calls which zero out memory
 * to avoid spurious dbx memcheck warnings. DBG7 is used so that these calls
 * can be enabled without enabling any debug printing.
 *
 * The DBG8 macro is used to enable saving of various machine state to files
 * during a migration. This is kept separate from other debugging macros as
 * these files can be very large and take a large amount of time to generate.
 *
 * The run time command line option, -g <debug_level>, can be used to set the
 * debug level. To enable memset for dbx access check debugging specify
 * "-g 0x40". To enable all debug output specify "-g 0x3f".
 *
 * The convention for the debug levels is:
 *
 *   DBG1: Domain Services
 *   DBG2: Function Tracing
 *   DBG3: Detailed Debug Output
 *   DBG4: Hypervisor Control
 *   DBG5: The Really Verbose Stuff
 *   DBG6: Used by Power Management
 *   DBG7: memset() Calls As Described Above
 *   DBG8: Save Migration State to Files
 *   DBG9: DRM policy selected
 */

#define	DBG(_s)		if (debug_level & 0x3f) { do { _s; } while (0); }
#define	DBG1(_s)	if (debug_level & 0x01) { do { _s; } while (0); }
#define	DBG2(_s)	if (debug_level & 0x02) { do { _s; } while (0); }
#define	DBG3(_s)	if (debug_level & 0x04) { do { _s; } while (0); }
#define	DBG4(_s)	if (debug_level & 0x08) { do { _s; } while (0); }
#define	DBG5(_s)	if (debug_level & 0x10) { do { _s; } while (0); }
#define	DBG6(_s)	if (debug_level & 0x20) { do { _s; } while (0); }
#define	DBG7(_s)	if (debug_level & 0x40) { do { _s; } while (0); }
#define	DBG8(_s)	if (debug_level & 0x80) { do { _s; } while (0); }
#define	DBG9(_s)	if (debug_level & 0x100) { do { _s; } while (0); }

#define	DBG_BLOCK	if (debug_level & 0x3f)
#define	DBG8_BLOCK	if (debug_level & 0x80)

#else	/* ZDBG */

#define	ASSERT(_s)

#define	DBG(_s)
#define	DBG1(_s)
#define	DBG2(_s)
#define	DBG3(_s)
#define	DBG4(_s)
#define	DBG5(_s)
#define	DBG6(_s)
#define	DBG7(_s)
#define	DBG8(_s)
#define	DBG9(_s)

#define	DBG_BLOCK	if (0)
#define	DBG8_BLOCK	if (0)

#endif	/* ZDBG */

/* FUTURE: Ideally this should go away */
#define	SANITY(_s)	do { _s; } while (0)

#if defined(_BIG_ENDIAN)
#define	hton8(_x)	((uint8_t)(_x))
#define	hton16(_x)	((uint16_t)(_x))
#define	hton32(_x)	((uint32_t)(_x))
#define	hton64(_x)	((uint64_t)(_x))
#define	ntoh8(_x)	((uint8_t)(_x))
#define	ntoh16(_x)	((uint16_t)(_x))
#define	ntoh32(_x)	((uint32_t)(_x))
#define	ntoh64(_x)	((uint64_t)(_x))
#else
#define	hton8(_x)	((uint8_t)(_x))
#define	hton16(_x)	BSWAP_16((uint16_t)(_x))
#define	hton32(_x)	BSWAP_32((uint32_t)(_x))
#define	hton64(_x)	BSWAP_64((uint64_t)(_x))
#define	ntoh8(_x)	((uint8_t)(_x))
#define	ntoh16(_x)	BSWAP_16((uint16_t)(_x))
#define	ntoh32(_x)	BSWAP_32((uint32_t)(_x))
#define	ntoh64(_x)	BSWAP_64((uint64_t)(_x))
#endif

#ifdef __cplusplus
}
#endif

#endif /* _BASICS_H_ */
