/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * This is the backend which runs on real HW and talks
 * to the HV over the hvctl channel.
 */

#include <sys/types.h>
#include <sys/sysmacros.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <dlfcn.h>
#include <errno.h>
#include <libintl.h>
#include <netdb.h>
#include <stdlib.h>
#include <thread.h>
#include <synch.h>

#include "basics.h"
#include "errors.h"
#include "fatal.h"
#include "allocate.h"
#include "mdesc.h"
#include "mdesc_impl.h"
#include "ldom.h"
#include "backend.h"
#include "database.h"
#include "zpri.h"
#include "restart.h"
#include "support.h"
#include "pri.h"
#include "pri_svc.h"
#include "zprintf.h"
#include "hvctl.h"
#include "sequencer.h"
#include "frags.h"
#include "vcpu.h"
#include "vldc.h"
#include "vio.h"
#include "pm/pm_shared.h"

/* routines used in the backend_t ops structure */
static void	hv_init(void);
static void	hv_start(void);
static bool_t	hv_process_option(int *ip, int argc, char **argv);
static md_t	*hv_get_pri(void);
static void	hv_init_pri(md_t *pri);
static void	hv_get_hv_state(void);
static md_t	*hv_get_guest_md(ldom_t *ldp);
static void	hv_aug_guestmd(ldom_t *ldp, md_t *mdp);
static void	hv_dl_guestmd(ldom_t *ldp, uint64_t paddr,
			uint8_t *bufp, int size);
static void	hv_aug_hvmd(md_t *mdp);
static void	hv_dl_hvmd(uint64_t paddr, uint8_t *bufp,
			int size, hvmd_type_t md_type);
static void	hv_parse_hv_guest(ldom_t *ldp, md_node_t *guest, md_t *hvmd);
static void	hv_aug_hv_guest(ldom_t *ldp, md_node_t *guest, md_t *hvmd);
static void	hv_config_guests(config_flag_t flag);
static void	hv_start_new_guests(void);
static void	hv_stop_guest(ldom_t *ldp);
int	hv_get_guest_state(ldom_t *ldp, int which,
			ldom_hv_state_t *state);
bool_t	is_hvctl_ver_supported(uint64_t maj, uint64_t min);

static void	hv_banner(response_t *resp);
static int	hv_get_guest_utilization(ldom_t *ldp, rs_guest_util_t *util);
static int	hv_get_vcpu_state(resid_t *vcpu, rs_vcpu_state_t *vstate);
static int	hv_get_mem_state(resid_t *mem, rs_mem_state_t *mstate);
static int	hv_panic_guest(ldom_t *ldp, response_t *resp);
static int	hv_read_mem(uint64_t pa, uint8_t *bufp, int len);
static int	hv_write_mem(uint64_t pa, uint8_t *bufp, int len);
static int64_t	hv_scrub_mem(uint64_t pa, uint64_t len);
static int	hv_mem_dr_verify(ldom_t *ldp, uint64_t paddr,
			uint64_t raddr, uint64_t size);
static void	hv_fini(void);

/* internal routines */
static void	grab_hv_properties(md_t *hvmd);
static void	open_hvctl(void);
static void	hvctl_negotiate(void);

#if EMULATE_MDFILES
#define	DIO_HV_MD		"dump-dio-hv.md"
#define	DIO_HV_MD_RECONF	"dump-dio-hv-reconf.md"
#endif

backend_t hv_be = {
	"hv",
	"Backend for LDom Manager talking to hypervisor",
	hv_init,
	hv_start,
	hv_process_option,
	hv_get_pri,
	hv_init_pri,
	hv_get_hv_state,
	hv_get_guest_md,
	hv_aug_guestmd,
	hv_dl_guestmd,
	hv_aug_hvmd,
	hv_dl_hvmd,
	hv_parse_hv_guest,
	hv_aug_hv_guest,
	hv_config_guests,
	hv_start_new_guests,
	hv_stop_guest,
	hv_get_guest_state,
	hv_banner,
	hv_get_guest_utilization,
	hv_get_vcpu_state,
	hv_get_mem_state,
	hv_panic_guest,
	hv_read_mem,
	hv_write_mem,
	hv_scrub_mem,
	hv_mem_dr_verify,
	hv_fini,
	BACKEND_MAGIC
};

#define	MAX_MEM_DR_VERIFY_RETRIES	128
#define	MAX_READ_REPLY_RETRIES		10

int get_pri_method = GET_PRI_INVAL;

static hvctl_msg_t *hvreply = NULL;
static mutex_t hvlock = DEFAULTMUTEX;
static cond_t hvcond = DEFAULTCV;
static cond_t hvreply_taken = DEFAULTCV;
static mutex_t physmemlock = DEFAULTMUTEX;

/*LINTED*/
static hvctl_msg_t hverr;		/* to hold error message for debug */
static bool_t pm_hvreply = false;
static cond_t pm_hvcond = DEFAULTCV;
static hvctl_status_t pm_hvstatus;
static uint16_t pm_send_seqnum;
static hvctl_msg_t *pm_hvrp = NULL;
static bound_mfrag_t *pm_frag = NULL;		/* physmem for PM usage */
static int hvpm_buf_select = 0;
static uint64_t hvpm_bufp[HVPM_BUF_COUNT];

static void init_hv_cmd(hvctl_msg_t *cmd_msg, int cmd);
static hvctl_msg_t *do_hv_cmd(hvctl_msg_t *cmd);
static void *hvctl_poll(void *ignored);
static void *pri_poll(void *ignored);

/*
 * Backend functions.
 */

static void
hv_init(void)
{
	DBG(zprintf("%s:\n", __func__));

	hvctl_state.fd = -1;
	hvctl_state.hvctl_chan_state = HVCTL_down;
	hvctl_state.major_ver = -1;
	hvctl_state.minor_ver = -1;
	hvctl_state.version_set = false;
	hvctl_state.zeus_seqn = 0;
	hvctl_state.hv_seqn = 0;
}

static void
hv_start(void)
{
	DBG(zprintf("%s:\n", __func__));

	/* initiate contact with HV */
	open_hvctl();

	if (thr_create(NULL, NULL, hvctl_poll, NULL, THR_DAEMON, NULL) != 0)
		efatal("Could not create HVCTL poll thread");

	hvctl_negotiate();

	/* start PRI poll thread */
	if (thr_create(NULL, NULL, pri_poll, NULL, THR_DAEMON, NULL) != 0)
		efatal("Could not create PRI poll thread");
}

static void
hv_banner(response_t *resp)
{
	switch (hvctl_state.hvctl_chan_state) {
	case HVCTL_down:
		rprintf(resp, gettext("\tUnable to contact hypervisor\n"));
		break;
	case HVCTL_challenge:
		rprintf(resp, gettext("\tContacting hypervisor\n"));
		break;
	case HVCTL_up:
		/* TRANSLATION_NOTE 'v' stands for version */
		rprintf(resp, gettext(
		    "\tHypervisor control protocol v %d.%d\n"),
		    hvctl_state.major_ver, hvctl_state.minor_ver);
		/* TRANSLATION_NOTE 'v' stands for version */
		rprintf(resp, gettext(
		    "\tUsing Hypervisor MD v %"PRIu64".%"PRIu64"\n"),
		    zeus.hvmd_vers_major, zeus.hvmd_vers_minor);
		break;
	}
}

/*
 * FIXME: This code is run to open up the HVCTL channel. However, if we lose a
 * packet or some other problem along the way this backend will not try and
 * re-open the channel .. this needs to get fixed or the hypervisor will stop
 * talking to us in the event of a channel error.
 */
static void
hvctl_negotiate(void)
{
	hvctl_msg_t cmd;
	hvctl_msg_t *rp;
	uint64_t key;
	hvctl_status_t cc;
	uint64_t major, minor;

	/*
	 * Exchange hello message with HV to negotiate versions and to
	 * handshake keys in order to open all hvctl commands.
	 *
	 * Currently we negotiate and support only one version number - either
	 * the HV does it or we bail.
	 */
	init_hv_cmd(&cmd, HVctl_op_hello);
	cmd.msg.hello.major = hton64(HVCTL_VERSION_MAJOR_NUMBER);
	cmd.msg.hello.minor = hton64(HVCTL_VERSION_MINOR_NUMBER);

	/*
	 * The response is likely either a fail, or wrong version. Either way
	 * we kill Zeus - the exit should cause the driver to reset the
	 * channel, and the HV to re-init the hvctl connection.
	 */
	rp = do_hv_cmd(&cmd);

	cc = ntoh16(rp->hdr.status);
	if (cc != HVctl_st_ok) {
		major = ntoh64(rp->msg.hello.major);
		minor = ntoh64(rp->msg.hello.minor);
		hverr = *rp;
		if (hvctl_state.version_set) {
			if (hvctl_state.major_ver != major &&
			    hvctl_state.minor_ver != minor)
				fatal("hvctl has suddenly changed versions !");
		}
		/*
		 * HV didn't like the major version number we sent, and is
		 * offering no others.
		 * FIXME: Return a version negotiation failure to HV before
		 * giving up.
		 */
		fatal("Hello to hypervisor failed\nVersion %"PRId64".%"PRId64
		    " returned", major, minor);
	}


	/*
	 * HV returns version numbers in different places for success
	 * vs. failure to maintain back compatibility with old ldmd which
	 * did not expect version numbers on success.
	 */
	major = ntoh64(rp->msg.clnge.major);
	minor = ntoh64(rp->msg.clnge.minor);

	/*
	 * This is a simple challenge algorithm whose sole purpose is to
	 * prevent inadvertent access to this channel (e.g. so things like
	 * "cat /dev/hvctl" don't work).  It is NOT meant to provide
	 * any kind of real security; the control domain needs to be treated
	 * as a privileged domain, with the appropriate access controls in
	 * place.
	 */
	hvctl_state.hvctl_chan_state = HVCTL_challenge;

	key = ntoh64(rp->msg.clnge.code);
	key ^= HVCTL_HV_CHALLENGE_K;
	key ^= HVCTL_ZEUS_CHALLENGE_K;

	/*
	 * Version checks:
	 *
	 * For now, we only support one major version.
	 * We support all minor version variants, though.  These currently
	 * include:
	 * 0: baseline
	 * 1: addition of HVctl_op_guest_only_reconf &
	 *    HVctl_op_guest_only_delayed_reconf; support for simple rollback
	 *    (i.e. not canceling delayed reconfig) on parse failure.
	 */
	if (major == 0) {
		/* No valid version info recv'd from HV; assume version 1.0 */
		major = 1;
		minor = 0;
	}

	if (major != HVCTL_VERSION_MAJOR_NUMBER)
		/*
		 * FIXME: Return a version negotiation failure to HV before
		 * giving up.
		 */
		fatal("hvctl version mismatch: supported major ver %" PRId64
		    ";\nreceived major ver %" PRId64 "\n",
		    HVCTL_VERSION_MAJOR_NUMBER, major);
	hvctl_state.version_set = true;
	hvctl_state.major_ver = HVCTL_VERSION_MAJOR_NUMBER;
	if (minor < HVCTL_VERSION_MINOR_NUMBER)
		hvctl_state.minor_ver = minor;
	else
		hvctl_state.minor_ver = HVCTL_VERSION_MINOR_NUMBER;

	Xfree(rp);
	DBG(zprintf("Contact with HV established\n"
	    "Version negotiated = %d.%d\n", hvctl_state.major_ver,
	    hvctl_state.minor_ver));

	/*
	 * OK - we've been given a challenge .. must respond
	 * to it in order to open up the channel.
	 */

	init_hv_cmd(&cmd, HVctl_op_response);
	cmd.msg.clnge.code = hton64(key);

	rp = do_hv_cmd(&cmd);

	cc = ntoh16(rp->hdr.status);
	if (cc != HVctl_st_ok) {
		hverr = *rp;
		fatal("Failed the key challenge - giving up");
	}

	Xfree(rp);
	DBG(zprintf("HV unlocked\n"));

	hvctl_state.hvctl_chan_state = HVCTL_up;
}

static char *
hvctl_op_to_str(hvctl_op_t op)
{
	static char s[16];

	switch (op) {
	case HVctl_op_hello:
		return ("HVctl_op_hello");
	case HVctl_op_challenge:
		return ("HVctl_op_challenge");
	case HVctl_op_response:
		return ("HVctl_op_response");
	case HVctl_op_get_hvconfig:
		return ("HVctl_op_get_hvconfig");
	case HVctl_op_reconfigure:
		return ("HVctl_op_reconfigure");
	case HVctl_op_guest_start:
		return ("HVctl_op_guest_start");
	case HVctl_op_guest_stop:
		return ("HVctl_op_guest_stop");
	case HVctl_op_guest_delayed_reconf:
		return ("HVctl_op_guest_delayed_reconf");
	case HVctl_op_guest_suspend:
		return ("HVctl_op_guest_suspend");
	case HVctl_op_guest_resume:
		return ("HVctl_op_guest_resume");
	case HVctl_op_guest_panic:
		return ("Hvctl_op_guest_panic");
	case HVctl_op_get_res_stat:
		return ("HVctl_op_get_res_stat");
	case HVctl_op_new_res_stat:
		return ("HVctl_op_new_res_stat");
	case HVctl_op_cancel_reconf:
		return ("HVctl_op_cancel_reconf");
	case HVctl_op_guest_only_reconf:
		return ("HVctl_op_guest_only_reconf");
	case HVctl_op_guest_only_delayed_reconf:
		return ("HVctl_op_guest_only_delayed_reconf");
	case HVctl_op_pm:
		return ("HVctl_op_pm");
	case HVctl_op_state_save:
		return ("HVctl_op_state_save");
	case HVctl_op_state_restore:
		return ("HVctl_op_state_restore");
	case HVctl_op_api_group_query:
		return ("HVctl_op_api_group_query");
	case HVctl_op_scrub_mem:
		return ("HVctl_op_scrub_mem");
	case HVctl_op_mem_dr_verify:
		return ("HVctl_op_mem_dr_verify");
	default:
		break;
	}

	snprintf(s, sizeof (s), "%d", op);
	return (s);
}

/*
 * Issue a reconfigure operation to the hypervisor.
 * For a delayed reconfiguration guestid is a valid
 * guest resource identifier otherwise it is
 * INVALID_GUESTID.
 */
static int
do_hv_reconfig(uint64_t hvmd_addr, uint32_t guestid, bool_t guest_only)
{
	hvctl_msg_t cmd;
	hvctl_msg_t *rp;
	hvctl_op_t op;
	hvctl_status_t cc;
	int rc = 0;

	DBG(zprintf("%s: guestid=%d, guest_only=%s...\n", __func__, guestid,
	    (guest_only) ? "true" : "false"));

#if !EMULATE_MDFILES
	/* Cancel delayed reconfig? */
	if (guestid == INVALID_GUESTID && hvmd_addr == NULL) {
		op = HVctl_op_cancel_reconf;
		DBG(zprintf("Sending HVctl_op_cancel_reconf\n"));
		init_hv_cmd(&cmd, op);
		rp = do_hv_cmd(&cmd);

		cc = ntoh16(rp->hdr.status);
		if (cc != HVctl_st_ok) {
			warning("No delayed reconfig in progress");
		}
		Xfree(rp);
		return (rc);
	} else if (guest_only) {
		/* Guest-only reconfig: delayed or not? */
		if (guestid == INVALID_GUESTID) {
			op = HVctl_op_guest_only_reconf;
			DBG(zprintf("Sending HVctl_op_guest_only_reconf\n"));
		} else {
			op = HVctl_op_guest_only_delayed_reconf;
			DBG(zprintf(
			    "Sending HVctl_op_guest_only_delayed_reconf\n"));
		}
		init_hv_cmd(&cmd, op);
		cmd.msg.reconfig.guestid = hton32(guestid);
		cmd.msg.reconfig.hvmdp = hton64(hvmd_addr);
	} else {
		if (guestid == INVALID_GUESTID) {
			op = HVctl_op_reconfigure;
			DBG(zprintf("Sending HVctl_op_reconfigure, hvmdp "
			    "0x%llx\n", hvmd_addr));
		} else {
			op = HVctl_op_guest_delayed_reconf;
			DBG(zprintf("Sending HVctl_op_guest_delayed_reconf, "
			    "hvmdp 0x%llx\n", hvmd_addr));
		}
		init_hv_cmd(&cmd, op);
		cmd.msg.reconfig.guestid = hton32(guestid);
		cmd.msg.reconfig.hvmdp = hton64(hvmd_addr);
	}

	rp = do_hv_cmd(&cmd);

	cc = ntoh16(rp->hdr.status);
	DBG(zprintf("Got response status %u to op %s\n", (int)cc,
	    hvctl_op_to_str(ntoh16(rp->hdr.op))));
	if (cc != HVctl_st_ok) {
		DBG(zprintf("\thvmdp   0x%llx\n",
		    ntoh64(rp->msg.rcfail.hvmdp)));
		DBG(zprintf("\tres     %u\n",
		    ntoh32(rp->msg.rcfail.res)));
		DBG(zprintf("\tcode    %u\n",
		    ntoh32(rp->msg.rcfail.code)));
		DBG(zprintf("\tnodeidx 0x%x\n",
		    ntoh32(rp->msg.rcfail.nodeidx)));
		DBG(zprintf("\tresid   0x%x\n",
		    ntoh32(rp->msg.rcfail.resid)));
		hverr = *rp;
		rc = cc;
	}

	Xfree(rp);
#endif
	return (rc);
}

/*
 * Issue a scrub memory operation to the hypervisor.
 */
static int64_t
hv_scrub_mem(uint64_t pa, uint64_t len)
{
	hvctl_msg_t	cmd;
	hvctl_msg_t	*rp;
	hvctl_status_t	cc;
	int64_t		rc = -1;

	init_hv_cmd(&cmd, HVctl_op_scrub_mem);
	cmd.msg.scrub_mem.hv_membase = hton64(pa);
	cmd.msg.scrub_mem.hv_memsize = hton64(len);

	rp = do_hv_cmd(&cmd);
	cc = ntoh16(rp->hdr.status);

	if (cc == HVctl_st_ok)
		rc = ntoh64(rp->msg.scrub_mem.hv_memsize);

	Xfree(rp);
	return (rc);
}

/*
 * Issue a guest operation to the hypervisor.
 */
static int
do_hv_guest_op(ldom_t *ldp, uint16_t op)
{
	hvctl_msg_t cmd;
	hvctl_msg_t *rp;
	hvctl_status_t cc;
	int rc = 0;

	init_hv_cmd(&cmd, op);
	cmd.msg.guestop.guestid = hton32(ldp->res_id->id);

	DBG(zprintf("Sending %s for guest %s (gid %"PRId64",resid %"PRId64")\n",
	    hvctl_op_to_str(op), ldp->namep, ldp->idp->id, ldp->res_id->id));

	rp = do_hv_cmd(&cmd);

	cc = ntoh16(rp->hdr.status);
	DBG(zprintf("Got response status %u to %s op\n", (int)cc,
	    hvctl_op_to_str(op)));
	if (cc != HVctl_st_ok) {
		DBG(zprintf("\tcode %u\n",
		    ntoh32(rp->msg.guestop.code)));
		hverr = *rp;
		rc = cc;
	}

	Xfree(rp);
	return (rc);
}

/*
 * Issue a mem dr verify operation to the hypervisor.
 */
static int
hv_mem_dr_verify(ldom_t *ldp, uint64_t paddr, uint64_t raddr,
    uint64_t size)
{
	hvctl_msg_t cmd;
	hvctl_msg_t *rp;
	hvctl_status_t cc;
	int rc = 0;
	uint32_t cookie = 0;
	int retries = 0;

	init_hv_cmd(&cmd, HVctl_op_mem_dr_verify);
	cmd.msg.mem_dr_verify.paddr = hton64(paddr);
	cmd.msg.mem_dr_verify.raddr = hton64(raddr);
	cmd.msg.mem_dr_verify.size = hton64(size);
	cmd.msg.mem_dr_verify.guestid = hton32(ldp->res_id->id);

	while (retries < MAX_MEM_DR_VERIFY_RETRIES) {
		cmd.msg.mem_dr_verify.cookie = hton32(cookie);
		rp = do_hv_cmd(&cmd);
		cc = ntoh16(rp->hdr.status);
		if (cc != HVctl_st_again)
			break;
		cookie = ntoh32(rp->msg.mem_dr_verify.cookie);
		Xfree(rp);
		retries++;
	}

	Xfree(rp);

	rc = cc;

	if (cc != HVctl_st_ok) {
		/*
		 * verify failed. The domain is not valid. Panic the domain
		 */
		zprintf("%s paddr=0x%"PRIx64" size=0x%"PRIx64
		    " raddr=0x%"PRIx64" guest = %s"
		    " status =%"PRIu16" cookie=%"PRIu32"\n",
		    __func__,  paddr, size, raddr, ldp->namep, cc,
		    cookie);
		cc = do_hv_guest_op(ldp, HVctl_op_guest_panic);
		if (cc != HVctl_st_ok)
			/* panic domain failed, cannot recover */
			fatal("HV guest panic operation failed (%d)",
			    cc);
	}

	return (rc);
}

/*
 * Issue a get resource status operation to the hypervisor.
 * Returns a pointer to the reply from the hypervisor in 'stat'.
 */
static int
do_hv_get_res_stat(uint32_t res, uint32_t resid, uint32_t infoid,
    hvctl_msg_t **stat)
{
	hvctl_msg_t cmd;
	hvctl_msg_t *rp;
	hvctl_status_t cc;

	init_hv_cmd(&cmd, HVctl_op_get_res_stat);
	cmd.msg.resstat.res = hton32(res);
	cmd.msg.resstat.resid = hton32(resid);
	cmd.msg.resstat.infoid = hton32(infoid);

	DBG5(zprintf("Sending HVctl_op_get_res_stat, res 0x%x, resid 0x%x, "
	    "infoid 0x%x\n", res, resid, infoid));

	rp = do_hv_cmd(&cmd);

	cc = ntoh16(rp->hdr.status);
	DBG5(zprintf("Got resp status %u to HVctl_op_get_res_stat\n", (int)cc));

	*stat = rp;
	return (cc);
}

/*
 * hv_init_pm_buf -  Allocate and set up toggle physical memory buffers for PM
 */
static int
hv_init_pm_buf()
{
	ASSERT(pm_frag == NULL);
	pm_frag = frag_alloc(NULL, FRAG_SIZE);
	if (pm_frag == NULL) {
		DBG(zprintf("%s: lack of phys mem, frag_alloc failed\n",
		    __func__));
		return (-1);
	} else {
		hvpm_buf_select = 0;
		hvpm_bufp[0] = pm_frag->paddr;
		hvpm_bufp[1] = pm_frag->paddr + HVPM_BUF_SIZE;
		return (0);
	}
}

/*
 *  hv_get_pm_buf - return alternative (toggled) physical memory buffer address
 *  to caller. Only called by do_hv_pm_state().
 */
static uint64_t
hv_get_pm_buf()
{
	hvpm_buf_select = (hvpm_buf_select % HVPM_BUF_COUNT);
	return (hvpm_bufp[hvpm_buf_select++]);
}

/*
 * do_hv_pm_state() - backend to HVctl_op_pm_state operation.
 * called by ldm_get/set_power_states(), or debug call directly.
 */
int
do_hv_pm_state(hvctl_pm_state_dir_t dir, int count, char **cookie,
	uint16_t *states, hvctl_status_t *status)
{
	hvctl_msg_t	cmd;
	hvctl_msg_t	*rp;
	hvctl_status_t	cc;
	timestruc_t	to;
	int 		i, len;
	hvctl_pm_cookie_info_t	*bufp, *rbufp = NULL;
	hvctl_pm_state_t	*pm_state_p;
	uint64_t	paddr;
	int		ret;

	/* allocate reusable physical memory */
	if (pm_frag == NULL) {
		if (hv_init_pm_buf() != 0)
			return (-1);
	}

	init_hv_cmd(&cmd, HVctl_op_pm);
	cmd.msg.pm.pm_op = HVctl_pm_op_state;

	len = count * sizeof (hvctl_pm_cookie_info_t);
	bufp = Xcalloc(len, uint8_t);

	DBG(zprintf("%s: %s: \n  ", __func__,
	    (dir == HVctl_pm_state_set) ? "HVctl_pm_state_set" :
	    "HVctl_pm_state_get"));
	for (i = 0; i < count; i++) {
		strcpy(bufp[i].pm_cookie, cookie[i]);
		if (dir == HVctl_pm_state_set) {
			bufp[i].pm_res_state = states[i];
			DBG(zprintf("%s -> %d ", cookie[i], states[i]));
		}
	}
	DBG(zprintf("\n"));

	paddr = hv_get_pm_buf();
	(void) hv_write_mem(paddr, (uint8_t *)bufp, len);
	pm_state_p = &cmd.msg.pm.pm_op_data.pm_state;
	pm_state_p->pm_state_dir = hton16(dir);
	pm_state_p->pm_cookie_addr = hton64(paddr);
	pm_state_p->pm_num_cookies = hton16(count);

	/*
	 * write to hvstate.fd and get response
	 */
	rp = do_hv_cmd(&cmd);
	pm_send_seqnum = cmd.hdr.seqn;	/* record send cmd seq. num */

	/*
	 * process response to HVctl_op_pm
	 */
	cc = ntoh16(rp->hdr.status);
	switch (cc) {
	case HVctl_st_ok:

		/* read per resource reply */
		rbufp = Xcalloc(len, uint8_t);
		(void) hv_read_mem(paddr, (uint8_t *)rbufp, len);
		for (i = 0; i < count; i++) {
			status[i] = rbufp[i].pm_res_status;
			if (dir == HVctl_pm_state_get)
				states[i] = rbufp[i].pm_res_state;
		}
		Xfree(rbufp);
		break;

	case HVctl_st_async:

		/* wait for async reply */
		DBG(zprintf("%s: wait for async reply, send_seqn=0x%X\n",
		    __func__, pm_send_seqnum));
		mutex_lock(&hvlock);
		to.tv_sec = time(NULL) + HV_PM_ASYNC_TIMEOUT;
		to.tv_nsec = 0;
		while (pm_hvreply != true) {
			if (cond_timedwait(&pm_hvcond, &hvlock, &to)
			    == ETIME) {
				/* Timed out */
				break;
			}
		}

		/* got async message */
		if (pm_hvreply == true) {
			pm_hvreply = false;
			cc = pm_hvstatus;
			rbufp = Xcalloc(len, uint8_t);
			(void) hv_read_mem(paddr, (uint8_t *)rbufp, len);
			for (i = 0; i < count; i++) {
				status[i] = rbufp[i].pm_res_status;
				if (dir == HVctl_pm_state_get)
					states[i] = rbufp[i].pm_res_state;
			}
			Xfree(pm_hvrp);
			Xfree(rbufp);
		} else {
			/*
			 * timedout while waiting for async msg, force the
			 * old seq num to be discarded
			 */
			DBG(zprintf("%s: Timedout on async reply for "
			    "seqnun %d\n", __func__, pm_send_seqnum));

			pm_send_seqnum--;
			cc = HVctl_st_timedout;

		}
		mutex_unlock(&hvlock);
		break;

	case HVctl_st_timedout:
		DBG(zprintf("%s: Received HVctl_st_timedout reply on seqn %d\n",
		    __func__, pm_send_seqnum));
		hverr = *rp;
		break;

	case HVctl_st_einval:
		DBG(zprintf("%s: Received HVctl_st_einval reply on seqn %d\n",
		    __func__, pm_send_seqnum));
		hverr = *rp;
		break;

	case HVctl_st_eillegal:
		DBG(zprintf("%s: Received HVctl_st_eillegal reply on seqn %d\n",
		    __func__, pm_send_seqnum));
		hverr = *rp;
		break;

	default:
		DBG(zprintf("%s: Unrecognized HV reply(cc=%d) on seqn %d\n",
		    __func__, (int)cc, pm_send_seqnum));
		hverr = *rp;
	}
	ret = cc;

pm_done:
	Xfree(bufp);
	Xfree(rp);

	return (ret);
}

/*
 * Power Management asynchronous response message handler
 *
 * The PM async message is caused by HVctl_pm_op_state operation
 * from a PM thread holding the sequencer lock, there is
 * one outstanding HVctl_pm_op_state operation going on
 * at a time, the PM thread waits for the results from HV.
 */
static void
do_hv_pm_async(hvctl_msg_t *msg)
{
	hvctl_pm_async_resp_t	*aresp;
	uint16_t	async_ret_seqn;

	if (ntoh16(msg->msg.pm.pm_op) != HVctl_pm_op_async_resp) {
		DBG(zprintf("%s: unexpected pm_op(%d)\n",
		    __func__, (int)msg->msg.pm.pm_op));
		Xfree(msg);
		return;
	}

	aresp = &msg->msg.pm.pm_op_data.pm_async_resp;
	async_ret_seqn = ntoh16(aresp->pm_aresp_seqnum);
	if (pm_send_seqnum != async_ret_seqn) {
		DBG(zprintf("%s: async_ret_seqn(0x%X) does not match "
		    "pm_send_seqnum(0x%X)\n",  __func__, async_ret_seqn,
		    pm_send_seqnum));
		Xfree(msg);
		return;
	}

	DBG(zprintf("%s: received async seqn(0x%X), vBSC status(%d)\n",
	    __func__, async_ret_seqn, ntoh16(aresp->pm_aresp_status)));

	/* wake up the PM thread on sequencer */
	mutex_lock(&hvlock);
	if (pm_hvreply != false)
		fatal("Unexpected Power Management async HV message");
	pm_hvreply = true;
	pm_hvstatus = ntoh16(msg->hdr.status);
	pm_hvrp = msg;
	cond_signal(&pm_hvcond);
	mutex_unlock(&hvlock);
}

/*
 * Check whether HV stack supports platform PM
 */
int
do_hv_pm_version(uint16_t *major, uint16_t *minor)
{
	hvctl_msg_t	cmd;
	hvctl_msg_t	*rp;
	hvctl_pm_version_t	*pm_vers_p;
	int		ret;

	/* Check HVctl major and minor version */
	if ((hvctl_state.major_ver < HVCTL_PM_MAJOR_VERSION) ||
	    ((hvctl_state.major_ver == HVCTL_PM_MAJOR_VERSION) &&
	    (hvctl_state.minor_ver < HVCTL_PM_MINOR_VERSION))) {
		DBG(zprintf("%s: HVctl(version %d.%d) does not support PM.\n",
		    __func__, hvctl_state.major_ver, hvctl_state.minor_ver));
		return (HVctl_st_enotsupp);
	}

	/*
	 * check whether HV supports the pm_cookie format given in
	 * current platform PRI
	 */
	init_hv_cmd(&cmd, HVctl_op_pm);
	cmd.msg.pm.pm_op = HVctl_pm_op_version;

	rp = do_hv_cmd(&cmd);
	ret = -1;
	if (ntoh16(rp->msg.pm.pm_op) != HVctl_pm_op_version) {
		DBG(zprintf("%s: unexpected response from pm_op(%d)\n",
		    __func__, ntoh16(rp->msg.pm.pm_op)));
	} else if ((ret = ntoh16(rp->hdr.status)) != HVctl_st_ok) {
		DBG(zprintf("%s: op failed, status(%d)\n",
		    __func__, ntoh16(rp->hdr.status)));
	} else {
		pm_vers_p = &rp->msg.pm.pm_op_data.pm_version;
		*major = ntoh16(pm_vers_p->pm_version_major);
		*minor = ntoh16(pm_vers_p->pm_version_minor);
		DBG(zprintf("%s: Read HV pm-version major(%d), minor(%d)\n",
		    __func__, *major, *minor));
	}

	Xfree(rp);
	return (ret);
}

/*
 * Backend specific command line options.
 */

/*ARGSUSED0*/
static bool_t
hv_process_option(int *ip, int argc, char **argv)
{
	return (false);
}


mutex_t pri_lock = DEFAULTMUTEX;
cond_t pri_cond = DEFAULTCV;
md_t *initial_pri = NULL;

static md_t *
hv_get_pri(void)
{
	timestruc_t tout;
	int err = 0;
	int retries;

	/* wait for pri poll thread to start */
	mutex_lock(&pri_lock);
	while (get_pri_method == GET_PRI_INVAL)
		cond_wait(&pri_cond, &pri_lock);
	mutex_unlock(&pri_lock);

again:
	if (get_pri_method == GET_PRI_DS) {
		/*
		 * Libpri is not available for obtaining the PRI so do phase
		 * one initialization of domain services which will start the
		 * pri domain service to get the PRI.
		 */
		ds_init1();
	}

	retries = INITIAL_PRI_RETRIES;

	do {
		tout.tv_sec = time(NULL) + INITIAL_PRI_TIMEOUT;
		tout.tv_nsec = 0;

		mutex_lock(&pri_lock);
		while (initial_pri == NULL) {
			err = cond_timedwait(&pri_cond, &pri_lock, &tout);
			if (err == 0 || err == EINTR)
				continue;
			else
				break;
		}
		mutex_unlock(&pri_lock);

		if (initial_pri != NULL) {
			return (initial_pri);
		} else {
			warning("Error attempting to obtain the PRI from the "
			    "system controller: %s", strerror(err));
			if ((get_pri_method == GET_PRI_DS) &&
			    (ds_reset_channel(ds_sp_channel()) != 0))
				fatal("Resetting SP domain services channel");
		}
	} while (--retries > 0);

	if (get_pri_method == GET_PRI_LIBPRI) {
		/*
		 * Getting the PRI via libpri failed so now try using the pri
		 * domain service.
		 */
		warning("Could not get the PRI via libpri.\nFalling back to "
		    "use the pri domain service");
		get_pri_method = GET_PRI_DS;
		goto again;
	}

	fatal("Could not obtain the PRI from the system controller");
	/*NOTREACHED*/
}

/*ARGSUSED0*/
static void
hv_init_pri(md_t *pri_md)
{
}

md_t *
read_md(uint64_t md_paddr)
{
	md_header_t	*md_hp;
	uint8_t		*bufp;
	md_t		*mdp;
	int		total_size;

	md_hp = Xcalloc(1, md_header_t);

	hv_read_mem(md_paddr, (uint8_t *)md_hp, MD_HEADER_SIZE);

	total_size = md_scan_bin_size(md_hp);

	bufp = Xrealloc(md_hp, total_size);
	SANITY(md_hp = NULL);

	hv_read_mem(md_paddr + MD_HEADER_SIZE, bufp + MD_HEADER_SIZE,
	    total_size - MD_HEADER_SIZE);

	DBG(dump_to_file(bufp, total_size, "dump-memory.%llx.md", md_paddr));

	mdp = md_scan_bin(bufp, total_size);
	if (mdp == NULL)
		fatal("failed reading MD from HV");

	Xfree(bufp);
	return (mdp);
}

static md_t *
hv_get_guest_md(ldom_t *ldp)
{
	md_t *mdp;
	bound_mfrag_t *active_md_fp;

#if EMULATE_MDFILES
	char mdname[256];
	snprintf(mdname, sizeof (mdname), "dump-dio-%s.md", ldp->namep);
	mdp = get_md_from_file(mdname);
	if (mdp != NULL)
		return (mdp);
#endif

	if (zeus.delayed_reconfig_ldom == ldp) {
		ASSERT(ldp->flag.gen_new_md == true);
		active_md_fp = ldp->frags.next_md_frag;
	} else {
		ASSERT(ldp->flag.gen_new_md == false);
		active_md_fp = ldp->frags.committed_md_frag;
	}

	ASSERT(active_md_fp != NULL);
	mdp = read_md(active_md_fp->paddr);

	return (mdp);
}

static void
ensure_hvmd_consistency(md_t **hvmd, uint64_t *hvmd_pa)
{
	uint64_t new_mdpa;
	int rc;

#if EMULATE_MDFILES
	*hvmd = NULL;
	*hvmd = get_md_from_file(DIO_HV_MD);
	if (*hvmd != NULL)
		return;
#else

	*hvmd = read_md(*hvmd_pa);

	if (md_get_prop_val(*hvmd, md_root_node(*hvmd), "new_mdpa",
	    &new_mdpa) == 0) {
		/*
		 * Reconfiguration sequence was not completed.  Recover by
		 * reconfiguring with the HV MD at PA 'new_mdpa'.
		 */
		DBG(zprintf("Incomplete reconfiguration detected, completing "
		    "with hvmd @ 0x%llx\n", new_mdpa));

		if ((rc = do_hv_reconfig(new_mdpa, INVALID_GUESTID, false)) !=
		    0)
			fatal("(%d) Reconfiguring the HV. The HV is in an "
			    "inconsistent state and a system reset is "
			    "required", rc);

		md_fini(*hvmd);
		*hvmd = read_md(new_mdpa);
		*hvmd_pa = new_mdpa;
	}
#endif
}

/*
 * If we're restarting in a delayed reconfig scenario, the HV MD that we
 * parsed was the pending one.  This function captures the last committed guest
 * MD frag of the target of the delayed reconfig (from the current HV MD)
 */
static void
assign_committed_md_frags(uint64_t md_fragp)
{
	md_t *mdp;
	md_node_t **guestlistp;
	int i, numguests;
	uint64_t guestmdpa;
	char *name;
	ldom_t *ldp = NULL;

	mdp = read_md(md_fragp);
	guestlistp = Xmalloc((int)sizeof (md_node_t *) * md_node_count(mdp));
	numguests = md_scan_dag(mdp, NULL, md_find_name(mdp, "guest"),
	    md_find_name(mdp, "fwd"), guestlistp);

	for (i = 0; i < numguests; i++) {
		md_node_t *guestp = guestlistp[i];

		/*
		 * Locate the guest MD
		 */
		if (md_get_prop_val(mdp, guestp, "mdpa", &guestmdpa) != 0)
			fatal("Guest node of HV MD missing mdpa property");

		if (md_get_prop_str(mdp, guestp, "name", &name) != 0)
			fatal("Guest node of HV MD missing name property");

		ldp = find_ldom_by_name(name);

		if (ldp == NULL)
			fatal("Delayed reconfig: no LDom named %s found "
			    "in current config", name);

		/*
		 * Only the guest under delayed_reconfig will have a different
		 * current guest MD that needs its own frag assigned.
		 */
		if (zeus.delayed_reconfig_ldom != ldp)
			continue;

		/*
		 * It's possible that the committed & next guest MDs are the
		 * same, in which case we've already assigned the frag at
		 * guestmdpa, causing this to harmlessly return NULL.
		 */
		ldp->frags.committed_md_frag = frag_assign(ldp, guestmdpa);

		DBG(zprintf("\tCurrent guest MD frag at 0x%llx, gen %llu\n",
		    guestmdpa, ldp->props.md_generation_num));

		break;
	}
	md_fini(mdp);
	Xfree(guestlistp);
}

hvctl_msg_t *
get_hvconfig(void)
{
	hvctl_msg_t cmd;

	init_hv_cmd(&cmd, HVctl_op_get_hvconfig);
	return (do_hv_cmd(&cmd));
}

static void
hv_get_hv_state(void)
{
	uint64_t	membase;
	uint64_t	memsize;
	uint64_t	hvmdpa;
	uint64_t	current_hvmdpa;
	uint64_t	del_reconf_hvmdpa;
	uint32_t	del_reconf_gid;
	md_t		*mdp;
	hvctl_msg_t	*rp;
	int		cc;

	/*
	 * This is where Zeus goes to the HV - reads back its MD, and
	 * discovers where all the existing LDoms are situated, as well as
	 * where the hypervisor is located.
	 */

	rp = get_hvconfig();

	/*
	 * FIXME: request - response types should be
	 * handled via a send state machine to ensure
	 * we actually got the response we expected, and that
	 * other async messages are filtered out.
	 */
	ASSERT(ntoh16(rp->hdr.op) == HVctl_op_get_hvconfig);

	cc = ntoh16(rp->hdr.status);
	if (cc != 0)
		fatal("Request for hypervisor config failed - code 0x%x", cc);

	membase = ntoh64(rp->msg.hvcnf.hv_membase);
	memsize = ntoh64(rp->msg.hvcnf.hv_memsize);
	current_hvmdpa = ntoh64(rp->msg.hvcnf.hvmdp);
	del_reconf_hvmdpa = ntoh64(rp->msg.hvcnf.del_reconf_hvmdp);
	del_reconf_gid = (int)ntoh32(rp->msg.hvcnf.del_reconf_gid);
	if (del_reconf_gid != -1) {
		/* zeus.delayed_reconfig_ldom set as zeus.ldoms is populated */
		zeus.delayed_reconfig_gid = del_reconf_gid;

		/*
		 * In hvctl 1.0, we need to lock out any subsequent delayed
		 * reconfig operations.
		 */
		if (hvctl_state.minor_ver < 1) {
			warning("Restart: disallowing subsequent operations "
			    "for delayed reconfig on gid %d", del_reconf_gid);
			zeus.delayed_reconfig_type = DEL_RECONFIG_LOCKED;
		} else {
			zeus.delayed_reconfig_type = DEL_RECONFIG_OTHER;
		}

		/* Set up to parse the delayed config, not the current one */
		hvmdpa = del_reconf_hvmdpa;
	} else {
		hvmdpa = current_hvmdpa;
	}
	Xfree(rp);

	DBG(zprintf("HV info:\tmembase=0x%llx\tmemsize=0x%llx\thvmd@ 0x%llx\n",
	    membase, memsize, hvmdpa));

	/*
	 * Check HV MD to see if the LDom manager died before a two step
	 * reconfiguration sequence (i.e. unbind or CPU DR) could be completed,
	 * and if so perform recovery.
	 */
	ensure_hvmd_consistency(&mdp, &hvmdpa);

	DBG(dump_md_to_file(mdp, "hvmd-in"));

	/*
	 * Now initialize state from the HV MD.
	 */
	target.parse_hvmd(mdp);

	/* Do we need to handle legacy HV MD properties? */
	if (zeus.hvmd_vers_major == 0 ||
	    (zeus.hvmd_vers_major == 1 && zeus.hvmd_vers_minor < 1)) {
		grab_hv_properties(mdp);
	}

	/* Once we parse the MDs, use zeus.delayed_reconfig_ldom */
	zeus.delayed_reconfig_gid = INVALID_GUESTID;

	/* allocate the hypervisor's reserved memory pool */
	zeus.hv_mblockp = pull_bmblock(membase, memsize);
	if (zeus.hv_mblockp == NULL)
		fatal("HV memory binding is not available from the "
		    "memory described by the PRI");

	/* protect the HVMD(s) in frag(s) */
	if (in_delayed_reconfig(NULL)) {
		zeus.current_hvmd_fp = hv_md_assign(current_hvmdpa);
		ASSERT(zeus.current_hvmd_fp != NULL);
		zeus.next_hvmd_fp = hv_md_assign(hvmdpa);
		ASSERT(zeus.next_hvmd_fp != NULL);

		/*
		 * Since target.parse_hvmd() handled the pending config,
		 * we still need to assign the frags for any committed
		 * guest MDs (as they might be different).
		 */
		assign_committed_md_frags(current_hvmdpa);
	} else {
		zeus.current_hvmd_fp = hv_md_assign(hvmdpa);
		ASSERT(zeus.current_hvmd_fp != NULL);
	}

	/* save away HV MD */
	if (zeus.hv_mdp != NULL)
		md_fini(zeus.hv_mdp);
	zeus.hv_mdp = mdp;
}

/*
 * This is for legacy use only on systems that don't provide this info
 * directly to the HV via a mini-MD from the SP, but instead include them
 * in the HV MD itself (whereby we must propagate them to new HV MDs that
 * we generate).
 */
static void
grab_hv_properties(md_t *hvmd)
{
	md_node_t *root;

	root = md_root_node(hvmd);

	if (md_get_prop_val(hvmd, root, "tod-frequency",
	    &pri.tod_frequency) < 0)
		pri.tod_frequency = 0;
	if (md_get_prop_val(hvmd, root, "tod", &pri.tod) < 0)
		pri.tod = 0;
	/*
	 * FIXME: these are simply propagated to every HV MD.  Instead, the LDOM
	 * Manager should not even know about them; they should be messaged
	 * from the SP to HV directly at start of day, and removed from the MD.
	 */
	if (md_get_prop_val(hvmd, root, "erpt-pa", &pri.erpt_pa) < 0)
		pri.erpt_pa = 0;
	if (md_get_prop_val(hvmd, root, "erpt-size", &pri.erpt_size) < 0)
		pri.erpt_size = 0;
}

/*
 * This callback provides the opportunity to add backend specific features to
 * a guest MD It is mostly never used, but here just in case.
 */
/*ARGSUSED0*/
static void
hv_aug_guestmd(ldom_t *ldp, md_t *mdp)
{
}

static void
hv_dl_guestmd(ldom_t *ldp, uint64_t paddr, uint8_t *bufp, int mdsize)
{
	static int seq = 0;

	DBG(zprintf("downloaded new guest MD @ 0x%" PRIx64 " (%d), guest %s, "
	    "gid %" PRId64 ", res_id %" PRId64 "\n", paddr, seq, ldp->namep,
	    ldp->idp->id, ldp->res_id->id));
	DBG(dump_to_file(bufp, mdsize, "dump-guest.%02d-%s.md",
	    seq, ldp->namep));
	DBG(if (++seq == 20) seq = 0);

#if EMULATE_MDFILES
	dump_to_file(bufp, mdsize, "dump-dio-%s.md", ldp->namep);

#else
	/* write the guest MD into its memory location */
	hv_write_mem(paddr, bufp, mdsize);
#endif
}

static void
add_local_tick_properties(ldom_t *ldp, md_t *hvmd, md_node_t *guest)
{
	int		i;
	pri_core_t	**cores;
	int		ncores;
	pri_cpu_t	*cpu;
	uint64_t	*ids;

	ASSERT(ldp->flag.local_tick_sync == true);

	/* add the property to the HV MD */
	md_update_value_property(hvmd, guest, HVMD_PROP_TICK_SYNC,
	    GUEST_TICK_SYNC_LOCAL);

	/*
	 * Create an array of physical strand IDs that
	 * represent the reserved cores for the domain.
	 * It is only necessary to save one ID for each
	 * core in use.
	 */
	ncores = get_cores_in_use(ldp, &cores);

	ids = Xmalloc(ncores * sizeof (*ids));

	for (i = 0; i < ncores; i++) {
		/* just use the first strand on the core */
		cpu = CHAIN_START(cores[i]->cpus);
		ASSERT(cpu != NULL);

		ids[i] = cpu->phys_id;
	}

	/* add the array to the MD */
	md_update_data_property(hvmd, guest, HVMD_PROP_MIG_RSVD,
	    (ncores * sizeof (*ids)), (uint8_t *)ids);

	Xfree(cores);
}

static void
add_mem_dr_props(ldom_t *ldp, md_t *hvmd, md_node_t *guest)
{
	ASSERT(ldp->mem_dr_state.inprogress == true);

	/* add the flag to the HV MD */
	md_update_value_property(hvmd, guest, HVMD_PROP_MEMDR_IN_PROGRESS, 1);
}

static void
add_suspended_props(ldom_t *ldp, md_t *hvmd, md_node_t *guest)
{
	char	*mig_prop;

	ASSERT(ldp->mig_state.suspended == true);

	mig_prop = (ldp->flag.migration_target) ?
	    HVMD_PROP_MIG_TGT : HVMD_PROP_MIG_SRC;

	/* add the flag to the HV MD */
	md_update_value_property(hvmd, guest, mig_prop, 1);

	if (ldp->mig_state.bind_fp != NULL) {
		md_update_value_property(hvmd, guest, HVMD_PROP_MIG_MDPA,
		    ldp->mig_state.bind_fp->paddr);
	}
}

/*
 * Opportunity for this backend to augment the HV MD with additional
 * information, prior to the binary form being created, and it's memory
 * space allocated.
 *
 * Currently, the modifications that need to be made is a scan
 * of all the domains to see if any has been migrated and is running
 * in LOCAL tick sync mode or/and if any memory dr has been in progress.
 * In those cases, provide this information to the HV.
 */
static void
hv_aug_hvmd(md_t *hvmd)
{
	md_node_t	**guests;
	int		nguests;
	ldom_t		*ldp;
	char		*name;

	/* get all the guest nodes in the MD */
	guests = Xmalloc((int)sizeof (md_node_t *) * md_node_count(hvmd));
	nguests = md_scan_dag(hvmd, NULL, md_find_name(hvmd, "guest"),
	    md_find_name(hvmd, "fwd"), guests);

	/*
	 * Walk the list of guests and see if any migration and memory DR
	 * related properties need to be added to the HV MD.
	 */
	for (int i = 0; i < nguests; i++) {

		if (md_get_prop_str(hvmd, guests[i], "name", &name) != 0) {
			DBG(zprintf("%s: no name prop for guest\n", __func__));
			continue;
		}

		CHAIN_ITER(zeus.ldoms, ldp, link) {

			/* check if this is the domain of interest */
			if (strcmp(name, ldp->namep) != 0)
				continue;

			/* check if the domain is in local tick sync mode */
			if (ldp->flag.local_tick_sync)
				add_local_tick_properties(ldp, hvmd, guests[i]);

			/* check if a migration is in progress */
			if (ldp->mig_state.suspended)
				add_suspended_props(ldp, hvmd, guests[i]);

			/* check if a mem dr is in progress */
			if (ldp->mem_dr_state.inprogress == true)
				add_mem_dr_props(ldp, hvmd, guests[i]);
		}
	}

	Xfree(guests);
}

/* ARGSUSED1 */
static void
hv_dl_hvmd(uint64_t paddr, uint8_t *bufp, int buflen, hvmd_type_t md_type)
{
	static int seq = 0;

	DBG(zprintf("downloaded new HV MD @ 0x%llx (%d)\n", paddr, seq));
#if EMULATE_MDFILES
	if (md_type == HVMD) {
		dump_to_file(bufp, buflen, DIO_HV_MD);
	} else {
		dump_to_file(bufp, buflen, DIO_HV_MD_RECONF);
	}
#else
	DBG(dump_to_file(bufp, buflen, "dump-hv.%02d.md", seq));
	DBG(if (++seq == 20) seq = 0);

	hv_write_mem(paddr, bufp, buflen);
#endif

	zeus.domaining_enabled = true;
}

static void
grab_guest_boot_image(md_t *mdp, md_node_t *guestp, ldom_t *ldp)
{
	uint64_t rombase;

	if (md_get_prop_val(mdp, guestp, "rombase", &rombase))
		fatal("Guest has no rombase");

	ldp->boot_imagep = pri_get_boot_image(rombase);

	if (ldp->boot_imagep == NULL)
		fatal("Guest has no bootable image at 0x%llx", rombase);
}

static void
hv_parse_hv_guest(ldom_t *ldp, md_node_t *guest, md_t *hvmd)
{
	grab_guest_boot_image(hvmd, guest, ldp);
}

static void
hv_aug_hv_guest(ldom_t *ldp, md_node_t *guest, md_t *hvmd)
{
	if (ldp->boot_imagep == NULL)
		return;

	md_add_value_property(hvmd, guest, "rombase",
	    ldp->boot_imagep->imgcp->pa_base + ldp->boot_imagep->offset);

	md_add_value_property(hvmd, guest, "romsize", ldp->boot_imagep->size);
}

static void
hv_config_guests(config_flag_t flag)
{
	int rc;
	uint64_t paddr;
	uint32_t guestid;
	bool_t guest_only = false;
	bool_t warm_start_on_failure = false;

	/* Config, reconfig or delayed reconfig? */
	switch (flag) {
	case CONFIG:
		guestid = INVALID_GUESTID;
		ASSERT(zeus.next_hvmd_fp != NULL);
		paddr = zeus.next_hvmd_fp->paddr;
		break;
	case GUEST_CONFIG:
		if (hvctl_state.minor_ver > 0) {
			guestid = INVALID_GUESTID;
			guest_only = true;
			ASSERT(zeus.next_hvmd_fp != NULL);
			paddr = zeus.next_hvmd_fp->paddr;
		} else {
			fatal("Unsupported config_flag %d for HV "
			    "version %d.%d\n", (int)flag, hvctl_state.major_ver,
			    hvctl_state.minor_ver);
		}
		break;
	case DELAYED_RECONFIG:
		/* Mark the LDom as having its new config committed */
		zeus.delayed_reconfig_ldom->flag.config_committed = true;

		/*
		 * If HVctl_op_guest_reconf() not available, use old broken
		 * methodology of cancelling and rebuilding the delayed reconfig
		 */
		if (hvctl_state.minor_ver > 0) {
			ASSERT(zeus.delayed_reconfig_ldom != NULL);
			guestid = zeus.delayed_reconfig_ldom->idp->id;
			ASSERT(zeus.reconf_fp != NULL);
			paddr = zeus.reconf_fp->paddr;
			guest_only = true;
			warm_start_on_failure = true;
			DBG(zprintf("Initiating delayed reconfig for LDom %s\n",
			    zeus.delayed_reconfig_ldom->namep));
			break;
		} else {
			/* FALLTHROUGH */
		}
	case RECONFIG:
		guestid = INVALID_GUESTID;
		ASSERT(zeus.reconf_fp != NULL);
		paddr = zeus.reconf_fp->paddr;
		break;
	case DELAYED_CONFIG:
		ASSERT(zeus.delayed_reconfig_ldom != NULL);
		guestid = zeus.delayed_reconfig_ldom->idp->id;
		ASSERT(zeus.next_hvmd_fp != NULL);
		paddr = zeus.next_hvmd_fp->paddr;
		/* Mark the LDom as having its new config delayed */
		zeus.delayed_reconfig_ldom->flag.config_committed = false;
		DBG(zprintf("Initiating delayed config for LDom %s\n",
		    zeus.delayed_reconfig_ldom->namep));
		warm_start_on_failure = true;
		break;
	case CANCEL_DELAYED_CONFIG:
		paddr = NULL;
		guestid = INVALID_GUESTID;
		break;
	default:
		fatal("Unknown config_flag_t %d\n", (int)flag);
		/* NOTREACHED */
	}

	/* now configure the guests */
	if ((rc = do_hv_reconfig(paddr, guestid, guest_only)) != 0) {
		if (warm_start_on_failure == true) {
			/*
			 * The idea here is to support a warm start which can
			 * resync Zeus to the current state of the system,
			 * allowing the code to handle a parse error returned
			 * from HV or a delayed reconfig that was cancelled
			 * (whether by Zeus or HV), even if intermediate guest
			 * MD changes have been committed.  Having saved each
			 * most recent such intermediate guest MD's frag (in
			 * ldp->frags.committed_md_frag) enables us to do this.
			 */
			zeus.do_warm_start = true;
			fatal("(%d) Reconfiguring the HV (FIXME: do warmstart)",
			    rc);
		} else {
			fatal("(%d) Reconfiguring the HV", rc);
		}
	}
}

/*
 * Call into hypervisor to start a guest.
 */
static void
start_guest(ldom_t *ldp)
{
	int rc;

	if ((rc = do_hv_guest_op(ldp, HVctl_op_guest_start)) == 0) {
		ldp->state = STARTING;
	} else if (rc < 0) {
		/*
		 * Unrecoverable error, abort with a fatal error. SMF
		 * will restart the LDom manager and the state will be
		 * re-determined from the HV.
		 */
		fatal("HV guest start operation failed (%d)", rc);
	} else {
		/*
		 * Recoverable error. Issue a warning and leave the guest in
		 * the bound state.
		 */
		warning("HV guest start operation failed (%d)", rc);
	}
}

static void
hv_start_new_guests(void)
{
	ldom_t *ldp;

	DBG(zprintf("%s:\n", __func__));

	CHAIN_ITER(zeus.inprogress, ldp, inprog) {
		ASSERT(ldp->state == BOUND);
		start_guest(ldp);
	}
}

static void
hv_stop_guest(ldom_t *ldp)
{
	int rc;

	if ((rc = do_hv_guest_op(ldp, HVctl_op_guest_stop)) < 0) {
		/*
		 * Unrecoverable error, abort with a fatal error. SMF will
		 * restart the LDom manager and the state will be
		 * re-determined from the HV.
		 */
		fatal("HV guest stop operation failed (%d)", rc);
	} else if (rc > 0) {
		if (rc == HVctl_st_stop_failed) {
			/*
			 * Stop failed because LDom was resetting (rebooting)
			 * so set it back to active.
			 */
			ldp->state = ACTIVE;
		} else {
			/*
			 * Recoverable error. Issue a warning and leave the
			 * guest state unchanged.
			 */
			warning("HV guest stop operation failed (%d)", rc);
		}
	}
}

static void
hv_fini(void)
{
}

int
hv_get_guest_state(ldom_t *ldp, int which, ldom_hv_state_t *state)
{
	hvctl_msg_t *stat;
	rs_guest_state_t *gs;
	rs_guest_soft_state_t *gss;
	int cc;

	if (which & LDOM_HV_STATE) {
		cc = do_hv_get_res_stat(HVctl_res_guest, ldp->res_id->id,
		    HVctl_info_guest_state, &stat);

		if (cc == HVctl_st_ok) {
			gs = (rs_guest_state_t *)stat->msg.resstat.data;
			state->state = ntoh64(gs->state);
			DBG(zprintf("%s HV guest state %s\n", ldp->namep,
			    hv_state_str(state->state)));
			Xfree(stat);
		} else {
			fatal("(%d) Getting LDom state from the HV", cc);
		}
	}

	if (which & LDOM_SOFT_STATE) {
		cc = do_hv_get_res_stat(HVctl_res_guest, ldp->res_id->id,
		    HVctl_info_guest_soft_state, &stat);

		if (cc == HVctl_st_ok) {
			gss = (rs_guest_soft_state_t *)stat->msg.resstat.data;
			state->soft = ntoh8(gss->soft_state);
			memcpy(state->soft_str, gss->soft_state_str,
			    SOFT_STATE_SIZE);
			state->soft_update = 0;
			Xfree(stat);
		} else {
			fatal("(%d) Getting LDom state from the HV", cc);
		}
	}

	if (ldp->flag.memdr_in_progress && state->state != GUEST_STATE_NORMAL) {
		/*
		 * guest domain is no longer active, if a memory dr was
		 * in progress, that memory dr is no longer valid. When
		 * domain reboots, it reads MD, therefore, no
		 * need to do any sync between zeus and the domain.
		 */
		ldp->mem_dr_state.inprogress = false;
		ldp->flag.locked = false;
		seq_update_guest_mds(ldp);
	}

	return (0);
}

static int
hv_get_guest_utilization(ldom_t *ldp, rs_guest_util_t *vutil)
{
	hvctl_msg_t *stat;
	rs_guest_util_t *hvutil;
	int cc;

	cc = do_hv_get_res_stat(HVctl_res_guest, ldp->res_id->id,
	    HVctl_info_guest_utilisation, &stat);

	if (cc == HVctl_st_ok) {
		hvutil = (rs_guest_util_t *)stat->msg.resstat.data;
		vutil->lifespan = ntoh64(hvutil->lifespan);
		vutil->wallclock_delta = ntoh64(hvutil->wallclock_delta);
		vutil->active_delta = ntoh64(hvutil->active_delta);
		vutil->stopped_cycles = ntoh64(hvutil->stopped_cycles);
		vutil->yielded_cycles = ntoh64(hvutil->yielded_cycles);
	} else {
		fatal("(%d) Getting LDom utilization from the HV", cc);
	}

	Xfree(stat);
	return (0);
}

static int
hv_get_vcpu_state(resid_t *vcpu, rs_vcpu_state_t *vstate)
{
	hvctl_msg_t *stat;
	rs_vcpu_state_t *hvstate;
	int cc;

	cc = do_hv_get_res_stat(HVctl_res_vcpu, vcpu->id,
	    HVctl_info_vcpu_state, &stat);

	if (cc == HVctl_st_ok) {
		hvstate = (rs_vcpu_state_t *)stat->msg.resstat.data;
		vstate->state = hvstate->state;
		vstate->lifespan = ntoh64(hvstate->lifespan);
		vstate->wallclock_delta = ntoh64(hvstate->wallclock_delta);
		vstate->active_delta = ntoh64(hvstate->active_delta);
		vstate->yielded_cycles = ntoh64(hvstate->yielded_cycles);
	} else {
		fatal("(%d) Getting VCPU state from the HV", cc);
	}

	Xfree(stat);
	return (0);
}

static int
hv_get_mem_state(resid_t *mem, rs_mem_state_t *mstate)
{
	hvctl_msg_t *stat;
	rs_mem_state_t *hvstate;
	int rc = -1;
	int cc;

	if (!hv_is_mem_util_supported())
		return (rc);

	cc = do_hv_get_res_stat(HVctl_res_memory, mem->id, HVctl_info_mem_util,
	    &stat);
	if (cc == HVctl_st_ok) {
		hvstate = (rs_mem_state_t *)stat->msg.resstat.data;
		mstate->perf_cntr = ntoh64(hvstate->perf_cntr);
		mstate->sample_time = ntoh64(hvstate->sample_time);
		rc = 0;
	} else {
		warning("(%d) Getting mem state from the HV", cc);
	}

	Xfree(stat);
	return (rc);
}

static int
hv_panic_guest(ldom_t *ldp, response_t *resp)
{
	int rc;

	if ((rc = do_hv_guest_op(ldp, HVctl_op_guest_panic)) < 0) {
		/*
		 * Unrecoverable error, abort with a fatal error. SMF
		 * will restart the LDom manager and the state will be
		 * re-determined from the HV.
		 */
		fatal("HV guest panic operation failed (%d)", rc);
	} else if (rc > 0) {
		/*
		 * Recoverable error, issue a warning.
		 */
		warning("HV guest panic operation failed (%d)", rc);
		if (resp != NULL) {
			if (rc == HVctl_st_eillegal)
				rprintf(resp,
				    gettext("LDom %s is not active\n"),
				    ldp->namep);
			else
				rprintf(resp, gettext("LDom %s is busy\n"),
				    ldp->namep);
		}

		return (-1);
	}

	return (0);
}

static void
open_hvctl(void)
{
	vldc_opt_op_t op;
	char pathname[MAXPATHLEN];
	int nbytes;

	/* FIXME: instances #'s should be configurable */
	nbytes = snprintf(pathname, MAXPATHLEN,
	    "/devices/%s@%x/%s@%x/%s@%d:%s", VDEVS_NAME, VDEVS_CFG_HANDLE,
	    CHANNELS_NAME, CHANNELS_CFG_HANDLE, VLDC_MD_NAME, VLDC_DEV_INST,
	    HVCTL_MD_NAME);

	if ((nbytes < 0) || (nbytes >= MAXPATHLEN)) {
		fatal("open_hvctl: unable to generate HVCTL VLDC pathname");
	}

	hvctl_state.fd = open(pathname, O_RDWR);
	if (hvctl_state.fd == -1) {
		if (errno == EBUSY) {
			sleep(1);
			fprintf(stderr, "fatal error: %s: Unable to open VLDC "
			    "instance %s for contacting hypervisor\n",
			    strerror(errno), pathname);
			exit(EXIT_RESTART);
		} else {
			efatal("Unable to open VLDC instance %s "
			    "for contacting hypervisor", pathname);
		}
	}

	op.op_sel = VLDC_OP_SET;
	op.opt_sel = VLDC_OPT_MODE;
	op.opt_val = LDC_RAW;

	if (ioctl(hvctl_state.fd, VLDC_IOCTL_OPT_OP, &op) == -1)
		efatal("Unable to set HVCTL LDC channel to LDC_RAW mode");
}

/*
 * Contact the hypervisor to read the specified range of physical
 * memory into the provided buffer.
 *
 * N.B. All vldc ioctl reads and writes share a single buffer within
 * the driver. For that reason, a locking scheme is required in order
 * prevent multiple operations from using the buffer at the same time.
 */
static int
hv_read_mem(uint64_t pa, uint8_t *bufp, int len)
{
	vldc_data_t data;

	data.src_vaddr = (uint64_t)bufp;
	data.dst_paddr = pa;
	data.length = len;

	DBG7(memset(bufp, 0xf, len));

	mutex_lock(&physmemlock);

	if (ioctl(hvctl_state.fd, VLDC_IOCTL_READ_PA, &data) == -1)
		efatal("ioctl VLDC_IOCTL_READ_PA failed");

	mutex_unlock(&physmemlock);

	return (len);
}

/*
 * Contact the hypervisor to write the contents of the specified
 * buffer into physical memory.
 *
 * N.B. All vldc ioctl reads and writes share a single buffer within
 * the driver. For that reason, a locking scheme is required in order
 * prevent multiple operations from using the buffer at the same time.
 */
static int
hv_write_mem(uint64_t pa, uint8_t *bufp, int len)
{
	vldc_data_t data;

	data.src_vaddr = (uint64_t)bufp;
	data.dst_paddr = pa;
	data.length = len;

	mutex_lock(&physmemlock);

	if (ioctl(hvctl_state.fd, VLDC_IOCTL_WRITE_PA, &data) == -1)
		efatal("ioctl VLDC_IOCTL_WRITE_PA failed");

	mutex_unlock(&physmemlock);

	return (len);
}

static void
init_hv_cmd(hvctl_msg_t *cmd_msg, int cmd)
{
	memset(cmd_msg, 0, HVCTL_MSG_MAXLEN);

	cmd_msg->hdr.chksum = hton16(0);		/* FIXME: unused */
	cmd_msg->hdr.op = hton16(cmd);
}

/*
 * Send a command to the hypervisor.
 * Since it is possible to issue HV commands from multiple threads
 * there is a need for locking
 */
static hvctl_msg_t *
do_hv_cmd(hvctl_msg_t *cmd)
{
	hvctl_msg_t *replyp;
	hvctl_op_t rop;
	hvctl_op_t op;
	timespec_t to;
	int rv;

	mutex_lock(&hvlock);
	op = ntoh16(cmd->hdr.op);
	cmd->hdr.seqn = hton16(hvctl_state.zeus_seqn++);

	rv = full_write(hvctl_state.fd, (uint8_t *)cmd, HVCTL_MSG_MAXLEN);
	if (rv < 0)
		fatal("Failed to send command to HV");
	if (rv != HVCTL_MSG_MAXLEN) {
		fatal("Command send of %d bytes came up short as %d",
		    HVCTL_MSG_MAXLEN, rv);
	}

	/*
	 * Wait for the hypervisor's response to a command.
	 */
	while (true) {
		if (hvreply != NULL) {
			rop = ntoh16(hvreply->hdr.op);
			if (op == rop) {
				break;
			}

			/* check special case for op_hello */
			if (op == HVctl_op_hello && rop ==
			    HVctl_op_challenge)
				break;
			/* response is not for us. continue to wait */
			DBG(zprintf("Got response for %s waiting for %s\n",
			    hvctl_op_to_str(rop), hvctl_op_to_str(op)));
		}

		to.tv_sec = time(NULL) + HV_ASYNC_TIMEOUT;
		to.tv_nsec = 0;
		rv = cond_timedwait(&hvcond, &hvlock, &to);
		if (rv == ETIME && hvreply == NULL) {
			zprintf("timeout waiting for op %s\n",
			    hvctl_op_to_str(op));
			fatal("waiting for hv response timeout\n");
		}
	}

	/* expected op in response */
	replyp = hvreply;
	hvreply = NULL;
	cond_signal(&hvreply_taken);
	mutex_unlock(&hvlock);

	return (replyp);
}

/*
 * Callback for this backend to handle an asynchronous hypervisor message
 * within the sequencer.
 */
static int
hv_async_handler(void *data)
{
	hvctl_msg_t *msg = data;
	hvctl_msg_t *stat = NULL;
	uint32_t res, resid, infoid;
	ldom_t *ldp;
	int cc;
	int rc = 0;

	res = ntoh32(msg->msg.resstat.res);

	switch (res) {
	case HVctl_res_guest:
		ASSERT_IN_SEQUENCER();
		resid = ntoh32(msg->msg.resstat.resid);
		infoid = ntoh32(msg->msg.resstat.infoid);

		if (infoid >= HVctl_info_guest_max) {
			rc = EINVAL;
			goto out;
		}

		if (infoid != HVctl_info_guest_soft_state) {
			cc = do_hv_get_res_stat(res, resid, infoid, &stat);
			if (cc == -1) {
				rc = EIO;
				goto out;
			} else if (cc != 0) {
				rc = EINVAL;
				goto out;
			}
		}

		ldp = find_ldom_by_resid(resid);
		if (ldp == NULL) {
			rc = EINVAL;
			goto out;
		}

		switch (infoid) {

		case HVctl_info_guest_state: {
			rs_guest_state_t *s;
			s = (rs_guest_state_t *)stat->msg.resstat.data;
			seq_ldom_state_change(ldp, NULL, ntoh64(s->state));
			break;
		}
		case HVctl_info_guest_soft_state: {
			ldp->hv_state.soft_update = 1;
			break;
		}
		case HVctl_info_guest_tod: {
			rs_guest_tod_t *t;
			t = (rs_guest_tod_t *)stat->msg.resstat.data;
			ldp->props.tod_offset = ntoh64(t->tod);
			break;
		}
		case HVctl_info_guest_api_groups:
			break;
		default:
			rc = EINVAL;
			break;
		}
		break;

	default:
		rc = EINVAL;
		break;
	}

out:
	Xfree(stat);
	Xfree(data);

	return (rc);
}

/*
 * Handle an asynchronous message from the hypervisor by queueing it up
 * to be processed in the sequencer.
 */
static void
handle_async_hv_msg(hvctl_msg_t *msg)
{
	int res;

	res = ntoh32(msg->msg.resstat.res);

	switch (res) {
	case HVctl_res_guest:
		/*
		 * Queue up to be dealt with within the sequencer.
		 * The message will be freed by the handler.
		 */
		seq_enqueue_async_hv_msg(hv_async_handler, msg);
		break;

	default:
		warning("Got unexpected async resource status message, res %d",
		    res);
		Xfree(msg);
		break;
	}
}

/*
 * Reset the hvctl channel.
 */
static void
reset_hvctl(void)
{
	close(hvctl_state.fd);
	open_hvctl();
	hvctl_negotiate();
}

/*
 * Main loop to poll for messages from the hypervisor.
 * This is run in its own thread.
 *
 * Messages from the hypervisor may be synchronous or asynchronous.
 * A synchronous message is passed to the thread that is waiting for it in
 * recv_hv_response().
 *
 * The HVctl_res_guest type async messages are queued on the sequencer to
 * to be handled later, while the power management asynchronous messages
 * are processed immediately on this thread.
 */
/*ARGSUSED0*/
static void *
hvctl_poll(void *unused)
{
	hvctl_msg_t *replyp;
	pollfd_t pollfd;
	int r;
	timestruc_t	to;
	int max_retries;

	pollfd.fd = hvctl_state.fd;
	pollfd.events = POLLIN;

	while (1) {
		hvctl_op_t op;

		replyp = Xmalloc(HVCTL_MSG_MAXLEN);

		if (poll(&pollfd, 1, -1) <= 0) {
			ewarning("Error polling the hvctl channel, resetting "
			    "the channel");
			reset_hvctl();
			pollfd.fd = hvctl_state.fd;
		}

		r = full_read(hvctl_state.fd, (uint8_t *)replyp,
		    HVCTL_MSG_MAXLEN);
		if (r < 0 || r != HVCTL_MSG_MAXLEN) {
			if (r < 0)
				ewarning("Error reading message from the hvctl "
				    "channel, resetting the channel");
			else
				warning("Short read of %d bytes (expected %d) "
				    "on hvctl channel, resetting the channel",
				    r, HVCTL_MSG_MAXLEN);
			reset_hvctl();
			pollfd.fd = hvctl_state.fd;
		}

		/* FIXME Verify checksum */

		op = ntoh16(replyp->hdr.op);

		switch (op) {
		case HVctl_op_new_res_stat:
			handle_async_hv_msg(replyp);
			break;

		case HVctl_op_pm:
			if (ntoh16(replyp->msg.pm.pm_op) ==
			    HVctl_pm_op_async_resp) {
				do_hv_pm_async(replyp);
				break;
			}
			/* FALLTHRU */

		default:
			/*
			 * Wake up the thread waiting for the reply.
			 */
			mutex_lock(&hvlock);
			to.tv_sec = time(NULL) + HV_ASYNC_TIMEOUT;
			to.tv_nsec = 0;
			max_retries = 0;
			while (hvreply != NULL &&
			    max_retries < MAX_READ_REPLY_RETRIES) {
				/*
				 * previous hvreply has not read yet.
				 * There should be a thread waiting to process
				 * the message.
				 */
				if (cond_timedwait(&hvreply_taken, &hvlock,
				    &to) == ETIME)
					max_retries++;
			}
			if (hvreply != NULL)
				fatal("Unread HV reply");

			hvreply = replyp;
			cond_broadcast(&hvcond);
			mutex_unlock(&hvlock);
			break;
		}
	}

	/* NOTREACHED */
}

/*
 * libpri allocator and deallocator
 */
static void *
libpri_alloc(size_t size)
{
	return (malloc(size));
}

/*ARGSUSED1*/
static void
libpri_free(void *bufp, size_t size)
{
	free(bufp);
}

static void
deliver_initial_pri(void *buf, ssize_t size)
{
	md_t *pri;

	pri = md_scan_bin(buf, size);
	if (pri == NULL)
		fatal("Received invalid PRI");
	libpri_free(buf, size);

	mutex_lock(&pri_lock);
	/*
	 * Check just in case the pri domain service is also running and
	 * delivered the PRI first. For now all PRIs except the first are
	 * discarded.
	 */
	if (initial_pri == NULL) {
		initial_pri = pri;
		cond_signal(&pri_cond);
	} else {
		md_fini(pri);
	}
	mutex_unlock(&pri_lock);
}

/*
 * Prototypes and pointers to functions needed from libpri.
 */
typedef int (*pri_init_pt)(void);
typedef void (*pri_fini_pt)(void);
typedef ssize_t (*pri_get_pt)(uint8_t, uint64_t *, uint64_t **,
    void *(*)(size_t), void (*)(void *, size_t));

static pri_init_pt pri_init_f = NULL;
static pri_fini_pt pri_fini_f = NULL;
static pri_get_pt pri_get_f = NULL;

/*
 * Try to dynamically load libpri.
 */
static void
check_for_libpri(void)
{
	void *dl_hdl;

	if ((dl_hdl = dlopen("libpri.so.1", RTLD_NOW)) == NULL) {
		warning("libpri.so.1 could not be loaded: %s", dlerror());
		return;
	}

	pri_init_f = (pri_init_pt)dlsym(dl_hdl, "pri_init");
	pri_fini_f = (pri_fini_pt)dlsym(dl_hdl, "pri_fini");
	pri_get_f = (pri_get_pt)dlsym(dl_hdl, "pri_get");

	if (pri_init_f == NULL ||pri_fini_f == NULL || pri_get_f == NULL)
		fatal("libpri does not provide one of pri_init(), pri_fini() "
		    "or pri_get()");
}

static void
notify_poll_started(int method)
{
	mutex_lock(&pri_lock);
	get_pri_method = method;
	cond_signal(&pri_cond);
	mutex_unlock(&pri_lock);
}

/*
 * Poll for PRI updates.
 * This is run in its own thread.
 */
/*ARGSUSED0*/
static void *
pri_poll(void *unused)
{
	uint64_t token;
	uint64_t *buf;
	ssize_t size;
	bool_t have_initial_pri = false;

	check_for_libpri();

	if (pri_init_f == NULL) {
		warning("libpri is not available. Falling back to pri domain "
		    "service to get the PRI");
		notify_poll_started(GET_PRI_DS);
		return (NULL);
	}

	notify_poll_started(GET_PRI_LIBPRI);

	if (pri_init_f() != 0) {
		ewarning("libpri initialization failed");
		return (NULL);
	}

	size = pri_get_f(PRI_GET, &token, &buf, libpri_alloc, libpri_free);
	if (size < 0) {
		ewarning("Initial pri_get() failed");
		return (NULL);
	}

	if (size > 0) {
		DBG(zprintf("PRI_GET received PRI, size=%d, token=%" PRIx64"\n",
		    size, token));
		DBG(dump_to_file(buf, size, "dump-ds-pri.bin"));

		deliver_initial_pri(buf, size);
		have_initial_pri = true;
	}

	/* loop polling for PRI updates */
	while (1) {
		size = pri_get_f(PRI_WAITGET, &token, &buf, libpri_alloc,
		    libpri_free);
		if (size < 0) {
			ewarning("PRI_WAITGET failed");

			/* Close & reopen connection to libpri */
			pri_fini_f();
			sleep(1);
			if (pri_init_f() != 0) {
				/*
				 * Print warning but keep going; PRI_WAITGET
				 * will return EBADF and we'll try again.
				 * FIXME: we should give up at some point.
				 */
				ewarning("libpri initialization failed");
			}
		} else if (size == 0) {
			warning("PRI_WAITGET unexpectedly returned 0");
		} else {
			DBG(zprintf("Received a PRI update, size=%d, "
			    "token=%" PRIx64 "\n", size, token));

			if (have_initial_pri) {
				/* for now just drop update on the floor */
				libpri_free(buf, size);
			} else {
				deliver_initial_pri(buf, size);
				have_initial_pri = true;
			}
		}
	}
	/*NOTREACHED*/
}

/*
 * Domain Migration routines
 */
static int
hv_state_save(ldom_t *ldp, uint64_t paddr, state_restype_t restype,
    uint32_t resid)
{
	int		rv = 0;
	hvctl_msg_t	cmd;
	hvctl_msg_t	*rp;
	int		cc;

	/* check that the firmware supports this */
	if (is_hvctl_ver_supported(HVCTL_MIG_MAJOR_VERSION,
	    HVCTL_MIG_MINOR_VERSION) == false) {
		warning("Negotiated HVCTL version does not match that "
		    "needed for migration (%" PRId64 ".%" PRId64")\n",
		    HVCTL_MIG_MAJOR_VERSION, HVCTL_MIG_MINOR_VERSION);
		return (HVctl_st_enotsupp);
	}

	init_hv_cmd(&cmd, HVctl_op_state_save);
	cmd.msg.save.statepa = hton64(paddr);
	cmd.msg.save.guestid = hton32(ldp->res_id->id);
	cmd.msg.save.restype = hton32(restype);
	cmd.msg.save.resid = hton32(resid);

	rp = do_hv_cmd(&cmd);

	cc = ntoh16(rp->hdr.status);
	DBG(zprintf("Got response status %u to %s op\n", cc,
	    "HVctl_op_state_save"));
	if (cc != HVctl_st_ok) {
		hverr = *rp;
		rv = cc;
	}

	Xfree(rp);
	return (rv);
}

static int
hv_state_restore(ldom_t *ldp, uint64_t paddr, state_restype_t restype,
    uint32_t resid)
{
	int		rv = 0;
	hvctl_msg_t	cmd;
	hvctl_msg_t	*rp;
	int		cc;

	/* check that the firmware supports this */
	if (is_hvctl_ver_supported(HVCTL_MIG_MAJOR_VERSION,
	    HVCTL_MIG_MINOR_VERSION) == false) {
		warning("Negotiated HVCTL version does not match that"
		    "needed for migration (%" PRId64 ".%" PRId64")\n",
		    HVCTL_MIG_MAJOR_VERSION, HVCTL_MIG_MINOR_VERSION);
		return (HVctl_st_enotsupp);
	}

	init_hv_cmd(&cmd, HVctl_op_state_restore);
	cmd.msg.save.statepa = hton64(paddr);
	cmd.msg.save.guestid = hton32(ldp->res_id->id);
	cmd.msg.save.restype = hton32(restype);
	cmd.msg.save.resid = hton32(resid);

	rp = do_hv_cmd(&cmd);
	cc = ntoh16(rp->hdr.status);
	DBG(zprintf("Got response status %u to %s op\n", cc,
	    "HVctl_op_state_restore"));
	if (cc != HVctl_st_ok) {
		hverr = *rp;
		rv = cc;
	}

	Xfree(rp);
	return (rv);
}

/*
 * This function requests the versions of the API Groups negotiated
 * with the HV
 *
 * Parameters:
 * 	resid		- resource id of the domain
 *	versionpp	- pointer to a pointer to an array of API Group versions
 *
 * Returns:
 *	rv == 0		- No versions available
 *	rv > 0		- numer of API Group versions
 *
 * Note: it is the responsibility of the caller to free the memory allocated
 * to versionpp.
 */
static int
hv_get_negotiated_api_groups(int resid, hv_api_group_version_t **versionpp)
{
	rs_guest_api_groups_t	*data;
	hvctl_msg_t		*stat;
	uint32_t	num_api_groups;
	uint64_t	base;
	int		amt, cc, len;

	cc = do_hv_get_res_stat(HVctl_res_guest, resid,
	    HVctl_info_guest_api_groups, &stat);

	if (cc != HVctl_st_ok) {
		DBG(zprintf("HVctl command failed (error=%d)\n", cc));
		return (0);
	}

	data = (rs_guest_api_groups_t *)stat->msg.resstat.data;
	ASSERT(data != NULL);

	num_api_groups = ntoh32(data->group_size);
	base = ntoh64(data->group_table);
	len = num_api_groups * (int)sizeof (hv_api_group_version_t);
	DBG5(zprintf("base:0x%" PRIx64 " len:%d\n", base, len));
	DBG(zprintf("API Group versions (%d) for guest %d\n",
	    num_api_groups, resid));

	if (num_api_groups == 0) {
		DBG(zprintf("No API Group version information "
		    "returned for guest %d\nIt is probably inactive "
		    "or non-existant\n", resid));
		return (0);
	}

	*versionpp = Xcalloc(num_api_groups, hv_api_group_version_t);
	amt = backend.read_mem(base, (uint8_t *)*versionpp, len);
	if (amt != len)
		DBG(zprintf("Note: only read %d of %d bytes\n", amt, len));

	return (num_api_groups);
}

/*
 * This function queries the HV asking if the versions of the API Groups
 * specified are supported.
 *
 * Parameters:
 * 	resid		- resource id of the domain
 *	num_api_groups	- number of entries in the array
 *	versionpp	- pointer to an array of API Group versions
 *
 * Returns:
 *	rv == 0		- Success (HVctl_st_ok)
 *	rv > 0		- error from HV
 *	rv < 0		- internal Zeus error
 */
/* ARGSUSED0 */
static int
hv_query_negotiated_api_groups(int resid, uint32_t num_api_groups,
    hv_api_group_version_t **versionpp)
{
	bound_mfrag_t	*frag;
	hvctl_msg_t	cmd;
	hvctl_msg_t	*rp;
	int		cc, len, amt;

	/* check that the firmware supports this */
	if (is_hvctl_ver_supported(HVCTL_MIG_MAJOR_VERSION,
	    HVCTL_MIG_MINOR_VERSION) == false) {
		warning("Negotiated HVCTL version does not match that"
		    "needed for migration (%" PRId64 ".%" PRId64")\n",
		    HVCTL_MIG_MAJOR_VERSION, HVCTL_MIG_MINOR_VERSION);
		return (HVctl_st_enotsupp);
	}

	len = num_api_groups * (int)sizeof (hv_api_group_version_t);
	if (num_api_groups == 0) {
		DBG(zprintf("No API Group versions specified\n"));
		return (ERR_GENERIC);
	}
	DBG(zprintf("Number of API Groups = %d\n", num_api_groups));

	if (*versionpp == NULL) {
		DBG(zprintf("No API Group version information found\n"));
		return (ERR_GENERIC);
	}

	frag = frag_alloc(NULL, len);
	ASSERT(frag != NULL);
	amt = backend.write_mem(frag->paddr, (uint8_t *)*versionpp, len);
	if (amt != len) {
		DBG(zprintf("Note: only wrote %d of %d bytes\n", amt, len));
		frag_free(frag);
		return (ERR_GENERIC);
	}

	/* populate the request and send it off to the HV */
	init_hv_cmd(&cmd, HVctl_op_api_group_query);
	cmd.msg.api_query.group_size = hton32(num_api_groups);
	cmd.msg.api_query.group_table = hton64(frag->paddr);

	rp = do_hv_cmd(&cmd);
	cc = ntoh16(rp->hdr.status);
	DBG(zprintf("HVctl_op_api_group_query return %d\n", cc));

	frag_free(frag);
	Xfree(rp);

	return (cc);
}

static int
hv_get_tick_sync_mode(ldom_t *ldp, uint32_t *mode)
{
	int			cc;
	rs_guest_tick_sync_t	*data;
	hvctl_msg_t		*stat;

	/* check that the firmware supports this */
	if (is_hvctl_ver_supported(HVCTL_MIG_MAJOR_VERSION,
	    HVCTL_MIG_MINOR_VERSION) == false) {
		warning("Negotiated HVCTL version does not match that"
		    "needed for migration (%" PRId64 ".%" PRId64")\n",
		    HVCTL_MIG_MAJOR_VERSION, HVCTL_MIG_MINOR_VERSION);
		return (HVctl_st_enotsupp);
	}

	cc = do_hv_get_res_stat(HVctl_res_guest, ldp->res_id->id,
	    HVctl_info_guest_tick_sync, &stat);

	if (cc != HVctl_st_ok) {
		DBG(zprintf("HVctl command failed (error=%d)\n", cc));
		return (-1);
	}

	data = (rs_guest_tick_sync_t *)stat->msg.resstat.data;
	ASSERT(data != NULL);

	*mode = ntoh32(data->tick_sync_mode);

	return (0);
}

bool_t
hv_is_scrub_mem_supported(void)
{
	return (is_hvctl_ver_supported(HVCTL_SCRUB_MEM_MAJOR_VERSION,
	    HVCTL_SCRUB_MEM_MINOR_VERSION));
}

bool_t
hv_is_mem_dr_supported(void)
{
	return (is_hvctl_ver_supported(HVCTL_MEM_DR_MAJOR_VERSION,
	    HVCTL_MEM_DR_MINOR_VERSION));
}

bool_t
hv_is_migration_supported(void)
{
	if ((hvctl_state.major_ver == HVCTL_MIG_MAJOR_VERSION &&
	    hvctl_state.minor_ver >= HVCTL_MIG_MINOR_VERSION) ||
	    (hvctl_state.major_ver > HVCTL_MIG_MAJOR_VERSION))
		return (true);

	return (false);
}

bool_t
hv_is_mem_util_supported(void)
{
	return (is_hvctl_ver_supported(HVCTL_MEM_UTIL_MAJOR_VERSION,
	    HVCTL_MEM_UTIL_MINOR_VERSION));
}

/*
 * Utility function to check if the HV supports a particular HVCTL
 * version.
 */
bool_t
is_hvctl_ver_supported(uint64_t maj, uint64_t min)
{
	if ((hvctl_state.major_ver == maj && hvctl_state.minor_ver >= min) ||
	    (hvctl_state.major_ver > maj))
		return (true);

	return (false);
}

/*
 * XXXMIG - These bypass the regular backend interface. They
 * should probably be folded into the real way of doing things.
 */

int
call_hv_get_tick_sync_mode(ldom_t *ldp, uint32_t *mode)
{
	return (hv_get_tick_sync_mode(ldp, mode));
}

int
call_hv_get_api_groups(uint32_t resid, hv_api_group_version_t **versionpp)
{
	return (hv_get_negotiated_api_groups(resid, versionpp));
}

int
call_hv_query_api_groups(uint32_t resid, uint32_t num_groups,
    hv_api_group_version_t **versionpp)
{
	return (hv_query_negotiated_api_groups(resid, num_groups, versionpp));
}

int
call_hv_state_save(ldom_t *ldp, uint64_t paddr, state_restype_t restype,
    uint32_t resid)
{
	return (hv_state_save(ldp, paddr, restype, resid));
}

int
call_hv_state_restore(ldom_t *ldp, uint64_t paddr, state_restype_t restype,
    uint32_t resid)
{
	return (hv_state_restore(ldp, paddr, restype, resid));
}

int
call_hv_guest_suspend(ldom_t *ldp)
{
	return (do_hv_guest_op(ldp, HVctl_op_guest_suspend));
}

int
call_hv_guest_resume(ldom_t *ldp)
{
	return (do_hv_guest_op(ldp, HVctl_op_guest_resume));
}

#if ZEUS_DEBUG_CLI

/*
 * Some external entry points for calling routines from debug CLI.
 */

int
call_hv_get_res_stat(uint32_t res, uint32_t resid, uint32_t infoid,
    hvctl_msg_t **stat)
{
	return (do_hv_get_res_stat(res, resid, infoid, stat));
}
#endif /* ZEUS_DEBUG_CLI */
