/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <stdlib.h>

#include "basics.h"
#include "fatal.h"
#include "allocate.h"
#include "mdesc.h"
#include "mdesc_impl.h"
#include "ldom.h"
#include "backend.h"
#include "option.h"
#include "zpri.h"
#include "sequencer.h"
#include "support.h"
#include "zprintf.h"
#include "frags.h"

/*
 * Definitions for Legion config file and HV MD generation.
 */
#define	ROM_BASE	(0xfff0000000ull)
#define	TOD_BASE	(0xfff0c1fff8ull)
#define	TOD_SIZE	(0x8ull)
#define	HVUART_BASE	(0xfff0c2c000ull)
#define	DUMBSERIAL_SIZE	(0x50ull)
#define	UART_SIZE	DUMBSERIAL_SIZE
#define	HV_SIZE		(62ull*1024ull*1024ull)
#define	HV_BASE		(2ull*1024ull*1024ull)

/*
 * Per-partition stuff.
 */
#define	UART_BASE	(0x1f10002000ull)
#define	UART_INC	(0x2000ull)
#define	NV_BASE		(0x1f11000000ull)
#define	NV_SIZE		(0x2000ull)
#define	NV_INC		(0x4000ull)
#define	DISK_BASE	(0x1f40000000ull)
#define	DISK_INC	(0x40000000ull)
#define	DISK_SIZE	(0x20100000ull)    /* 513 MB */

/*
 * State held specifically for this backend
 */

static struct LEGION_BE {
	uint64_t	xdc_bridge_base;
	uint64_t	xdc_bridge_size;
	char		*xdc_bridge_addrp;
	int		max_strands;
	int		min_guests;
	int		flag_trim;
} legion_state;

/*
 * Backend methods for driving a stand-alone legion
 * configuration etc.
 */

static void	legion_init(void);
static void	legion_start(void);
static bool_t	legion_process_option(int *ip, int argc, char **argv);
static md_t	*legion_get_pri(void);
static void	legion_init_pri(md_t *pri);
static void	legion_get_hv_state(void);
static md_t	*legion_get_guest_md(ldom_t *ldp);
static void	legion_aug_guestmd(ldom_t *ldp, md_t *mdp);
static void	legion_dl_guestmd(ldom_t *ldp, uint64_t paddr,
			uint8_t *bufp, int size);
static void	legion_aug_hvmd(md_t *mdp);
static void	legion_dl_hvmd(uint64_t paddr, uint8_t *bufp,
			int size, hvmd_type_t md_type);
static void	legion_parse_hv_guest(ldom_t *ldp,
			md_node_t *guest, md_t *hvmd);
static void	legion_aug_hv_guest(ldom_t *ldp,
			md_node_t *guest, md_t *hvmd);
static void	legion_config_guests(config_flag_t flag);
static void	legion_start_new_guests(void);
static void	legion_stop_guest(ldom_t *ldp);
static int	legion_get_guest_state(ldom_t *ldp, int which,
			ldom_hv_state_t *state);
static void	legion_banner(response_t *resp);
static int	legion_get_guest_utilization(ldom_t *ldp,
			rs_guest_util_t *util);
static int	legion_get_vcpu_state(resid_t *vcpu, rs_vcpu_state_t *vstate);
static int	legion_get_mem_state(resid_t *mem, rs_mem_state_t *mstate);
static int	legion_panic_guest(ldom_t *ldp, response_t *resp);
static int	legion_read_mem(uint64_t pa, uint8_t *bufp, int len);
static int	legion_write_mem(uint64_t pa, uint8_t *bufp, int len);
static int64_t	legion_scrub_mem(uint64_t pa, uint64_t len);
static int	legion_mem_dr_verify(ldom_t *ldp, uint64_t paddr,
			uint64_t raddr, uint64_t size);
static void	legion_fini(void);

static int gen_legion_conf(void);

#define	LEG_PREFIX	"legion"

backend_t legion_be = {
	"legion",
	"Backend for driving a stand-alone legion\n"
"    options:\n"
"\t--maxcpus <n>	- max number of cpu strands in legion.conf\n"
"\t--minguests <n>	- minimum number of guests in legion.conf\n"
"\t--trim		- reduce number of strands & consoles etc. to minimum\n"
"\t--ldccons	- force use of LDC based console",
	legion_init,
	legion_start,
	legion_process_option,
	legion_get_pri,
	legion_init_pri,
	legion_get_hv_state,
	legion_get_guest_md,
	legion_aug_guestmd,
	legion_dl_guestmd,
	legion_aug_hvmd,
	legion_dl_hvmd,
	legion_parse_hv_guest,
	legion_aug_hv_guest,
	legion_config_guests,
	legion_start_new_guests,
	legion_stop_guest,
	legion_get_guest_state,
	legion_banner,
	legion_get_guest_utilization,
	legion_get_vcpu_state,
	legion_get_mem_state,
	legion_panic_guest,
	legion_read_mem,
	legion_write_mem,
	legion_scrub_mem,
	legion_mem_dr_verify,
	legion_fini,
	BACKEND_MAGIC
};

static int
legion_mem_dr_verify(ldom_t *ldp, uint64_t paddr, uint64_t raddr, uint64_t size)
{
	return (0);
}

static void
legion_init(void)
{
	DBG(zprintf("legion_init()\n"));

	legion_state.max_strands = 0;
	legion_state.min_guests = 0;
	legion_state.flag_trim = false;
	zeus.ldc_console_config = false;
	zeus.allow_ctrl_domain_reconfig = true;
	zeus.simulating = true;
	option.in_place = true;
}

static void
legion_start(void)
{
	DBG(zprintf("legion_start()\n"));

	/* This routine initially left empty */
}

static void
legion_banner(response_t *resp)
{
	rprintf(resp, "Legion backend - banner info required\n");
}


/*
 * Backend specific command line options
 */
static bool_t
legion_process_option(int *ip, int argc, char **argv)
{
	int i = *ip;
	if (i < argc-1 && strcmp(argv[i], "--maxcpus") == 0) {
		legion_state.max_strands = atoi(argv[++i]);
		DBG(zprintf("legion_process_option: --maxcpus %d\n",
		    legion_state.max_strands));
		*ip = i;
		return (true);
	} else
	if (i < argc-1 && strcmp(argv[i], "--minguests") == 0) {
		legion_state.min_guests = atoi(argv[++i]);
		DBG(zprintf("legion_process_option: --minguests %d\n",
		    legion_state.min_guests));
		*ip = i;
		return (true);
	} else
	if (i < argc && strcmp(argv[i], "--trim") == 0) {
		legion_state.flag_trim = true;
		DBG(zprintf("legion_process_option: --trim\n"));
		*ip = i;
		return (true);
	}
	if (i < argc && strcmp(argv[i], "--ldccons") == 0) {
		zeus.ldc_console_config = true;
		DBG(zprintf("legion_process_option: --ldccons\n"));
		*ip = i;
		return (true);
	}
	return (false);
}

static md_t *
legion_get_pri(void)
{
	md_t *md;
	char *path;

	if (option.pri_file != NULL) {
		if (strchr(option.pri_file, '/') == NULL)
			path = zsprintf("%s/%s", PRI_TEMPLATES,
			    option.pri_file);
		else
			path = Xstrdup(option.pri_file);
	} else {
		path = zsprintf("%s/%s", PRI_TEMPLATES, PRI_FILE);
	}

	md = read_md_from_file(path);

	DBG(zprintf("PRI MD read from %s\n", path));
	Xfree(path);

	return (md);
}

static void
legion_init_pri(md_t *pri_md)
{
	md_node_t **nodelistp;
	char *namep;
	int i, num;

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(pri_md));

	/*
	 * If the PRI specifies an XDC bridge, we record what we need about
	 * it, and remember to generate the proper legion and MD files related
	 * to it.
	 */
	num = md_scan_dag(pri_md, NULL, md_find_name(pri_md, "xdc_bridge"),
	    md_find_name(pri_md, "fwd"), nodelistp);

	legion_state.xdc_bridge_addrp = NULL;

	/* if there is no XDC bridge, then we're done */
	if (num < 1) {
		Xfree(nodelistp);
		return;
	}

	if (num > 1)
		fatal("Multiple XDC bridges in the PRI");

	if (md_get_prop_val(pri_md, nodelistp[0],
	    "base", &legion_state.xdc_bridge_base))
		fatal("No base address in PRI xdc_bridge node");

	if (md_get_prop_val(pri_md, nodelistp[0], "size",
	    &legion_state.xdc_bridge_size))
		fatal("No size in PRI xdc_bridge node");

	if (md_get_prop_str(pri_md, nodelistp[0], "port", &namep))
		fatal("No contact port name for XDC bridge");

	legion_state.xdc_bridge_addrp = Xstrdup(namep);

	DBG(zprintf("XDC bridge @ 0x%"PRIx64" (size 0x%"PRIx64") - %s\n",
	    legion_state.xdc_bridge_base, legion_state.xdc_bridge_size,
	    legion_state.xdc_bridge_addrp));

	/*
	 * Limit the number of strands to the max requested if at all.
	 */
	if (legion_state.max_strands > 0) {
		num = md_scan_dag(pri_md, NULL, md_find_name(pri_md, "cpu"),
		    md_find_name(pri_md, "fwd"), nodelistp);

		for (i = legion_state.max_strands; i < num; i++)
			md_delete_node(pri_md, nodelistp[i]);

		md_scour(pri_md, md_find_name(pri_md, "back"));
	}

	Xfree(nodelistp);
}

static void
md_write(char *fnamep, uint8_t *bufp, int size)
{
	int fd;

	fd = open(fnamep, O_CREAT|O_RDWR|O_TRUNC, 0666);
	full_write(fd, bufp, size);
	close(fd);
}

/*
 * Normally we get this from the HV but since there's none active we
 * hard code it.
 */
static void
get_hv_mem(bool_t have_hv_md)
{
	mblock_t *bmbp;

	if (!have_hv_md) {
		/*
		 * Create a frag area from FRAG_SIZE up to the HV base. Memory
		 * from 0 to FRAG_SIZE is left for a red-zone.
		 */
		ASSERT(HV_BASE > FRAG_SIZE);
		bmbp = pull_bmblock(FRAG_SIZE, HV_BASE - FRAG_SIZE);
		if (bmbp == NULL)
			fatal("Unable to pull mblock");

		frag_free_mblock(bmbp);
	}

	/*
	 * Allocate memory for HV.
	 */
	zeus.hv_mblockp = mem_bind_phys_explicit(HV_BASE, HV_SIZE);
	if (zeus.hv_mblockp == NULL)
		fatal("No memory for HV");

	/* allocate HV memory for HV MD */
	zeus.current_hvmd_fp = hv_md_alloc();
}

static void
legion_get_hv_state(void)
{
	char		*fname;
	md_t		*mdp;
	bool_t		have_hv_md;
	int		fd;

	DBG(zprintf("legion_get_hv_state()\n"));

	/*
	 * Ordinarily this is where Zeus would go to the HV - read back its
	 * MD, and discover where all the existing LDoms are situated, as well
	 * as where the hypervisor is located.
	 *
	 * Since we've no HV to talk to, this function relies on the HV MD
	 * file created by the Legion BE to have previously been created.  If
	 * it's not there, we simply assume a machine that hasn't been booted
	 * yet, and apply some needed defaults.
	 */

	fname = zsprintf("%s-hv.md", LEG_PREFIX);
	fd = open(fname, O_RDONLY);
	if (fd == -1) {
		free(fname);
		/* try hardcoded HV MD from template */
		fname = zsprintf("%s/%s", PRI_TEMPLATES, HVMD_FILE);
		fd = open(fname, O_RDONLY);
	}
	if (fd != -1) {
		close(fd);
		mdp = read_md_from_file(fname);
		target.parse_hvmd(mdp);
		zeus.hv_mdp = mdp;
		have_hv_md = true;
		DBG(zprintf("HV MD read from %s\n", fname));
	} else {
		have_hv_md = false;
		DBG(zprintf("HV MD file not found\n"));
	}

	free(fname);

	get_hv_mem(have_hv_md);
}

static md_t *
legion_get_guest_md(ldom_t *ldp)
{
	char		*namep;
	int		size;
	int		len;
	md_t		*guest_mdp;

	size = strlen(ldp->namep) + sizeof (LEG_PREFIX) + 12;
	namep = alloca(size);
	len = snprintf(namep, size, "%s-guest-%s.md", LEG_PREFIX, ldp->namep);
	ASSERT(len < size);

	guest_mdp = read_md_from_file(namep);

	DBG(zprintf("Guest MD read from %s\n", namep));

	return (guest_mdp);
}

/*
 * This callback provides the opportunity to add backend specific features to
 * a guest MD It is mostly never used, but here just in case.
 */
static void
legion_aug_guestmd(ldom_t *ldp, md_t *mdp)
{
	md_node_t *vdevsp;
	md_node_t *vdiskp;
	md_node_t **nodelistp;
	int num;

	DBG(zprintf("legion_aug_guestmd()\n"));

	/*
	 * For the moment we stuff in the hypervisor supported virtual disk
	 * device nodes into the MD.  Eventually this is to be supplanted by a
	 * proper device module.
	 */

	nodelistp = Xmalloc(md_node_count(mdp) * sizeof (md_node_t *));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "virtual-devices"),
	    md_find_name(mdp, "fwd"), nodelistp);

	ASSERT(num == 1);

	vdevsp = nodelistp[0];

	vdiskp = md_link_new_node(mdp, "virtual-device",
	    vdevsp, "fwd", "back");

	(void) md_add_string_property(mdp, vdiskp, "name", "disk");
	(void) md_add_string_property(mdp, vdiskp, "fcode-driver-name",
	    "disk-virtual-device");

	/* FIXME: What is "my-space" for ? */
	(void) md_add_value_property(mdp, vdiskp, "my-space", 0);
	(void) md_add_value_property(mdp, vdiskp, "intr", 0);
	(void) md_add_value_property(mdp, vdiskp, "ino", 0);
	(void) md_add_value_property(mdp, vdiskp, "cfg-handle", 0xf0);

	Xfree(nodelistp);
}

static void
legion_dl_guestmd(ldom_t *ldp, uint64_t paddr, uint8_t *bufp, int mdsize)
{
	char *namep;
	int size;
	int len;

	DBG(zprintf("legion_dl_guestmd()\n"));

	/* Dump new guest MD to file for use by Legion */
	size = strlen(ldp->namep) + sizeof (LEG_PREFIX) + 12;
	namep = alloca(size);
	len = snprintf(namep, size, "%s-guest-%s.md", LEG_PREFIX, ldp->namep);
	ASSERT(len < size);	/* did it fit ? */

	md_write(namep, bufp, mdsize);
}

/*
 * Opportunity for this backend to augment the HV MD with additional
 * information, prior to the binary form being created, and it's memory space
 * allocated.
 */
static void
legion_aug_hvmd(md_t *mdp)
{
	md_node_t *rootp;
	md_node_t *guestsp;
	md_node_t **nodelistp;
	int i, num;

	DBG(zprintf("legion_aug_hvmd()\n"));

	/*
	 * Stuff specific to Legion ..
	 */

	rootp = md_root_node(mdp);

	md_add_value_property(mdp, rootp, "hvuart", HVUART_BASE);
	md_add_value_property(mdp, rootp, "tod", TOD_BASE);

	/*
	 * If the PRI indicated an XDC bridge, configure it for hvcontrol.
	 */

	if (legion_state.xdc_bridge_addrp != NULL) {
		md_node_t *bridgep;

		bridgep = md_link_new_node(mdp, "ldc_bridge", rootp,
		    "fwd", "back");
		md_add_value_property(mdp, bridgep, "ldcb_base",
		    legion_state.xdc_bridge_base);

		/*
		 * FIXME: Any in-hypervisor endpoints that the bridge
		 * terminates get defined here. For the moment zeus
		 * cant configure this.
		 */
	}

	/*
	 * Extras for each guest
	 */
	nodelistp = Xmalloc(sizeof (md_node_t *)*md_node_count(mdp));

	/*
	 * Ugh .. HV MD uses in-consistent link names
	 * not a DAG. This needs to be cleaned up.
	 */
	num = md_scan_dag(mdp, NULL, md_find_string(mdp, "guests", NULL),
	    md_find_string(mdp, "fwd", NULL), nodelistp);

	if (num != 1)
		fatal("legion_aug_hvmd: A single guests node is expected");
	guestsp = nodelistp[0];

	num = md_scan_dag(mdp, guestsp, md_find_string(mdp, "guest", NULL),
	    md_find_string(mdp, "fwd", NULL), nodelistp);

	/*
	 * For each guest find it's ID and create an NVRAM and UART The GID is
	 * the same as the ldp->idp->id, but since we're scanning from the MD
	 * and not the LDom list this is faster and easier
	 */
	for (i = 0; i < num; i++) {
		uint64_t gid;
		md_node_t *guestp;

		guestp = nodelistp[i];

		if (md_get_prop_val(mdp, guestp, "gid", &gid) != 0)
			fatal("Guest %d is missing gid property", i);

		md_add_value_property(mdp, guestp, "uartbase",
		    UART_BASE + UART_INC * gid);
		md_add_value_property(mdp, guestp, "nvbase",
		    NV_BASE + NV_INC * gid);
		md_add_value_property(mdp, guestp, "nvsize",
		    NV_SIZE);
		md_add_value_property(mdp, guestp, "diskpa",
		    DISK_BASE + DISK_INC * gid);
	}

	Xfree(nodelistp);
}

static void
legion_dl_hvmd(uint64_t paddr, uint8_t *bufp, int buflen,
	hvmd_type_t md_type)
{
	char *namep;
	int size;
	int len;

	/* Dump new HV MD to file for use by Legion */
	DBG(zprintf("legion_dl_hvmd()\n"));

	/*
	 * allocate sufficient space to hold the file-name
	 * string of the legion hv MD files: "legion-hv.md"
	 * and "legion-re-hv.md". A length of 10 should hold
	 * the longer prefix "-re-hv.md"
	 */
	size = sizeof (LEG_PREFIX) + 10;
	namep = alloca(size);
	if (md_type == HVMD)
		len = snprintf(namep, size, "%s-hv.md", LEG_PREFIX);
	else
		len = snprintf(namep, size, "%s-re-hv.md", LEG_PREFIX);
	ASSERT(len < size);	/* did it fit ? */

	md_write(namep, bufp, buflen);
}

static void
grab_guest_boot_image(md_t *mdp, md_node_t *guestp, ldom_t *ldp)
{
	uint64_t rombase;

	if (md_get_prop_val(mdp, guestp, "rombase", &rombase))
		fatal("Guest has no rombase");

	ldp->boot_imagep = pri_get_boot_image(rombase);

	if (ldp->boot_imagep == NULL)
		fatal("Guest has no bootable image at 0x%lx", rombase);
}

static void
legion_parse_hv_guest(ldom_t *ldp, md_node_t *guest, md_t *hvmd)
{
	grab_guest_boot_image(hvmd, guest, ldp);
}

static void
legion_aug_hv_guest(ldom_t *ldp, md_node_t *guest, md_t *hvmd)
{
	if (ldp->boot_imagep == NULL)
		return;

	md_add_value_property(hvmd, guest, "rombase",
	    ldp->boot_imagep->imgcp->pa_base + ldp->boot_imagep->offset);

	md_add_value_property(hvmd, guest, "romsize", ldp->boot_imagep->size);
}

static void
legion_config_guests(config_flag_t flag)
{
	zprintf("==> %s()...\n", __func__);

	if (flag == GUEST_CONFIG)
		return;

	/* This is the most appropriate time to generate Legion config file */
	if (gen_legion_conf() < 0)
		zprintf("Unable to generate Legion config file\n");
}

static void
legion_start_new_guests(void)
{
	ldom_t *ldp;

	DBG(zprintf("legion_start_new_guests()\n"));

	CHAIN_ITER(zeus.inprogress, ldp, inprog) {
		ASSERT(ldp->state == BOUND);
		seq_ldom_state_change(ldp, NULL, GUEST_STATE_NORMAL);
		ldp->state = ACTIVE;
		legion_get_guest_state(ldp, LDOM_SOFT_STATE, &ldp->hv_state);
	}
}

static void
legion_stop_guest(ldom_t *ldp)
{
	DBG(zprintf("legion_stop_guest - %s\n", ldp->namep));
	seq_ldom_state_change(ldp, NULL, GUEST_STATE_STOPPED);
	ldp->state = BOUND;
	legion_get_guest_state(ldp, LDOM_SOFT_STATE, &ldp->hv_state);
}

static int
legion_get_guest_state(ldom_t *ldp, int which, ldom_hv_state_t *state)
{
	DBG(zprintf("legion_get_guest_state - %s\n", ldp->namep));
	if (which & LDOM_HV_STATE) {
		if (ldp->state >= BOUND)
			state->state = GUEST_STATE_NORMAL;
		else if (ldp->state == BOUND)
			state->state = GUEST_STATE_STOPPED;
		else
			state->state = GUEST_STATE_UNCONFIGURED;
	}

	if (which & LDOM_SOFT_STATE) {
		if (ldp->state == ACTIVE) {
			state->soft = SIS_NORMAL;
			strlcpy(state->soft_str, "Solaris running",
			    SOFT_STATE_SIZE);
		} else {
			state->soft = SIS_TRANSITION;
			memset(state->soft_str, 0, SOFT_STATE_SIZE);
		}
	}

	return (0);
}

static int
legion_get_guest_utilization(ldom_t *ldp, rs_guest_util_t *util)
{
	return (ENOTSUP);
}

static int
legion_get_vcpu_state(resid_t *vcpu, rs_vcpu_state_t *vstate)
{
	return (ENOTSUP);
}

static int
legion_get_mem_state(resid_t *mem, rs_mem_state_t *mstate)
{
	return (ENOTSUP);
}

static int
legion_panic_guest(ldom_t *ldp, response_t *resp)
{
	if (resp)
		rprintf(resp, gettext("The panic command is not supported\n"));

	return (ENOTSUP);
}

/*ARGSUSED*/
static int
legion_read_mem(uint64_t pa, uint8_t *bufp, int len)
{
	return (len);
}

/*ARGSUSED*/
static int
legion_write_mem(uint64_t pa, uint8_t *bufp, int len)
{
	return (len);
}

/*ARGSUSED*/
static int64_t
legion_scrub_mem(uint64_t pa, uint64_t len)
{
	return (len);
}

static void
legion_fini(void)
{
	DBG(zprintf("legion_fini()\n"));
}

static void
conf_gen_roms(FILE *fp)
{
	pri_img_cont_t	*icp;
	pri_image_t	*hv_romp;
	pri_image_t	*rst_romp;
	pri_image_t	*pip;
	bound_mfrag_t	*hvfragp;

	hv_romp = NULL;
	rst_romp = NULL;

	for (pip = CHAIN_START(pri.image_list);
	    pip != NULL; pip = CHAIN_NEXT(pip->link)) {

		if (strcmp(pip->namep, "Hypervisor") == 0) {
			if (NULL != hv_romp)
				fatal("More than one hypervisor "
				    "ROM image in PRI");
			hv_romp = pip;
		} else if (strcmp(pip->namep, "ResetConfig") == 0) {
			if (NULL != rst_romp)
				fatal("More than one reset/config "
				    "ROM image in PRI");
			rst_romp = pip;
		}
	}

	if (NULL == rst_romp)
		fatal("be_legion: Reset/Config ROM image not found in PRI");
	if (NULL == hv_romp)
		fatal("be_legion: Hypervisor ROM image not found in PRI");

	hvfragp = zeus.next_hvmd_fp != NULL ?
	    zeus.next_hvmd_fp : zeus.current_hvmd_fp;

	ASSERT(hvfragp != NULL);

	/*
	 * Configure ROM images
	 */
	fprintf(fp, "\t\t\t/////////////////////////////////////////\n");
	fprintf(fp, "\t\t\t// ROM images                          //\n");
	fprintf(fp, "\t\t\t/////////////////////////////////////////\n");

	for (icp = CHAIN_START(pri.container_list);
	    icp != NULL; icp = CHAIN_NEXT(icp->link)) {

		fprintf(fp, "\t\t\tdevice \"memory\" 0x%"PRIx64" + 0x%"PRIx64
		    "{\n", icp->pa_base, icp->size);
		fprintf(fp, "\t\t\t\trom;\n");

		for (pip = CHAIN_START(pri.image_list);
		    pip != NULL; pip = CHAIN_NEXT(pip->link)) {

			if (pip->imgcp != icp)
				continue;

			fprintf(fp, "\t\t\t\tload +0x%"PRIx64
			    " \"%s-%s.bin\";\n",
			    pip->offset, LEG_PREFIX, pip->namep);

			/*
			 * Reset ROM image requires an additional
			 * content overlay
			 */

			if (pip == rst_romp) {
				fprintf(fp, "\t\t\t\tcontent +0x%"PRIx64" 64 "
				    "{\n", pip->offset+0x100);
				fprintf(fp, "\t\t\t\t\t0x%"PRIx64" 0x%"PRIx64
				    " 0x%"PRIx64" 0x%"PRIx64"\n\t\t\t\t}\n",
				    zeus.hv_mblockp->paddr,
				    zeus.hv_mblockp->size, hvfragp->paddr,
				    hv_romp->offset + hv_romp->imgcp->pa_base);
			}
		}

		fprintf(fp, "\t\t\t}\n");
	}
}

static int
gen_legion_conf(void)
{
	int i;
	uint64_t num, numcpus, numcores, numvthreads;
	uint64_t numglobals, numwindows, maxtl;
	uint64_t num_itlb_entries, num_dtlb_entries;
	FILE *fp;
	char legion_file[256];
	md_t *mdp;
	md_node_t **nodelistp;
	md_node_t *cpunodep;
	uint64_t sysclkfreq;
	uint64_t clkfreq;
	uint64_t rstv;
	uint64_t uart, nvbase, diskpa;
	ldom_t *ldp;
	pri_mblock_t *pmp;
	int sqr, xscale, yscale;
	int nconsoles;
	resid_t *idp;

	sprintf(legion_file, "%s.conf", LEG_PREFIX);

	if ((fp = fopen(legion_file, "w")) == NULL) {
		perror("open");
		return (-1);
	}

	/* Output boilerplate */
	fprintf(fp, "/*\n"
" * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.\n"
" * Use is subject to license terms.\n"
" *\n"
" */\n"
"\n"
"// Legion config file - automatically generated\n\n");

	/*
	 * Search MD for necessary info (it's not in the PRI, since Zeus
	 * doesn't care about this stuff internally, it just passes it on).
	 */
	mdp = pri.guest_template;

	nodelistp = Xmalloc(sizeof (md_node_t) * md_node_count(mdp));

	/*
	 * Get System clock frequency
	 */

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "platform"),
	    md_find_name(mdp, "fwd"), nodelistp);

	if (num == 0)
		fatal("No platform node found");
	if (num > 1)
		fatal("More than one platform node found");

	if (md_get_prop_val(mdp, nodelistp[0],
	    "stick-frequency", &sysclkfreq) < 0)
		fatal("Missing platform stick-frequency");

	/*
	 * Get CPU-type info (e.g. clock freq, nwins)
	 */
	numcpus = md_scan_dag(mdp, NULL, md_find_name(mdp, "cpu"),
	    md_find_name(mdp, "fwd"), nodelistp);

	if (numcpus == 0)
		fatal("No CPUs found");

	/* Just use first cpu node returned to get info */
	cpunodep = nodelistp[0];

	if (md_get_prop_val(mdp, cpunodep, "clock-frequency", &clkfreq) < 0)
		fatal("Missing CPU clock-frequency");

	if (md_get_prop_val(mdp, cpunodep, "nwins", &numwindows) < 0)
		fatal("Missing CPU nwins");

	/*
	 * To get # cores, count # L1$'s (I or D) that share an L2$.
	 * While we're at it, we can determine #vthreads by counting
	 * # cpus sharing an L1$.
	 */
	num = md_scan_dag(mdp, cpunodep, md_find_name(mdp, "cache"),
	    md_find_name(mdp, "fwd"), nodelistp);

	if (num == 0)
		fatal("No caches found");

	numcores = numvthreads = 0;

	for (i = 0; i < num; ++i) {
		md_node_t *nodep;
		uint64_t level;

		nodep = nodelistp[i];
		if (md_get_prop_val(mdp, nodep, "level", &level) < 0)
			fatal("No level specified for cache");

		if ((level == 1) && (numvthreads == 0)) {
			numvthreads = md_scan_dag(mdp, nodep,
			    md_find_name(mdp, "cpu"),
			    md_find_name(mdp, "back"), NULL);
			if (numvthreads == 0)
				fatal("No CPUs share L1$");
		}
		if ((level == 2) && (numcores == 0)) {
			numcores = md_scan_dag(mdp, nodep,
			    md_find_name(mdp, "cache"),
			    md_find_name(mdp, "back"), NULL);
			if (numcores == 0)
				fatal("No L1$'s share the L2$");
			/*
			 * Rather than having to count I or D caches, assume an
			 * equal # of each.
			 */
			numcores /= 2;
		}
		if ((numcores > 0) && (numvthreads > 0))
			break;
	}
	if (numcores == 0)
		fatal("No L2$ found");
	if (numvthreads == 0)
		fatal("No L1$ found");

	/*
	 * There are some things Legion needs that we can't get from the MD.
	 */
	numglobals = 4;
	rstv = ROM_BASE + 0x20;
	maxtl = 7;

	/*
	 * For Legion we fake up based on the largest CPU pid we find ..
	 * This saves us having to run too many host threads ..
	 */

	if (legion_state.flag_trim) {
		int maxpcpu;
		pri_cpu_t *pcpup;

		maxpcpu = 0;
		CHAIN_ITER(pri.cpu_list, pcpup, link) {
			if ((CHAIN_LENGTH(pcpup->bound.vcpus) > 0) &&
			    (maxpcpu < pcpup->phys_id))
				maxpcpu = pcpup->phys_id;
		}
		maxpcpu += 1;	/* conv largest id to count */

		if (maxpcpu <= 0)
			fatal("Internal error no strands assigned");

		numcores = (maxpcpu + numvthreads - 1) / numvthreads;
		if (numcores == 1)
			numvthreads = maxpcpu;
	}

	/* Open system, domain nodes */
	fprintf(fp, "system \"ldm\" {\n"
	    "\tdomain {\n"
	    "\t\tsysclkfreq 0x%lx;\n"
	    "\n"
	    "\t\tprocessor \"niagara\" {\n"
	    "\t\t\tclkfreq 0x%lx;\n"
	    "\t\t\tcores %d;\n"
	    "\t\t\tvthreads %d;\n"
	    "\t\t\tnglobals %d;\n"
	    "\t\t\tnwins %d;\n"
	    "\t\t\trstv 0x%lx;\n"
	    "\t\t\tmaxtl %d;\n",
	    sysclkfreq, clkfreq, numcores, numvthreads,
	    numglobals, numwindows, rstv, maxtl);

	/* Add TLBs info */
	num = md_scan_dag(mdp, cpunodep, md_find_name(mdp, "tlb"),
	    md_find_name(mdp, "fwd"), nodelistp);

	num_itlb_entries = num_dtlb_entries = 0;

	for (i = 0; i < num; ++i) {
		md_node_t *nodep;
		uint8_t *type;
		int len;

		nodep = nodelistp[i];

		if (md_get_prop_data(mdp, nodep, "type", &type, &len) < 0)
			fatal("TLB has no type info");

		if ((strcmp((char *)type, "instn") == 0) &&
		    (num_itlb_entries == 0)) {
			if (md_get_prop_val(mdp, nodep, "entries",
			    &num_itlb_entries) < 0)
				fatal("No entries info in TLB");
		}
		if ((strcmp((char *)type, "data") == 0) &&
		    (num_dtlb_entries == 0)) {
			if (md_get_prop_val(mdp, nodep, "entries",
			    &num_dtlb_entries) < 0)
				fatal("No entries info in TLB");
		}
		if ((num_itlb_entries > 0) && (num_dtlb_entries > 0))
			break;
	}

	if (num_itlb_entries == 0)
		fatal("No iTLB found");
	if (num_dtlb_entries == 0)
		fatal("No dTLB found");

	fprintf(fp, "\t\t\titlb {\n\t\t\t\tentries %d;\n\t\t\t}\n",
	    num_itlb_entries);

	fprintf(fp, "\t\t\tdtlb {\n\t\t\t\tentries %d;\n\t\t\t}\n",
	    num_dtlb_entries);

	/* close processor */
	fprintf(fp, "\t\t}\n\n");

	/* open addressmap */
	fprintf(fp, "\t\taddressmap {\n");

	for (pmp = CHAIN_START(pri.memory_list);
	    pmp != NULL; pmp = CHAIN_NEXT(pmp->link)) {

		fprintf(fp, "\t\t\tdevice \"memory\" 0x%"PRIx64" +"
		    " 0x%"PRIx64" {\n", pmp->pa_base, pmp->size);

		/*
		 * Search for any images that need to be loaded into
		 * this range.
		 */
		for (ldp = CHAIN_START(zeus.ldoms); ldp != NULL;
		    ldp = CHAIN_NEXT(ldp->link)) {
			bound_mfrag_t *mfp;

			/* Skip unbound LDoms */
			if (ldp->state == INACTIVE)
				continue;

			/* There must be a guest MD */
			if (ldp->flag.gen_new_md) {
				mfp = ldp->frags.next_md_frag;
			} else {
				mfp = ldp->frags.committed_md_frag;
				ASSERT(ldp->frags.next_md_frag == NULL);
			}

			ASSERT(mfp != NULL);
			if (mfp->pmblockp == pmp) {
				fprintf(fp, "\t\t\t\tload "
				    "+0x%"PRIx64" \"%s-guest-%s.md\";\n",
				    mfp->paddr - pmp->pa_base, LEG_PREFIX,
				    ldp->namep);
			}
		}

		/*
		 * This function is called before the HV MD is updated.
		 */
		ASSERT(zeus.next_hvmd_fp != NULL);
		if (zeus.next_hvmd_fp->pmblockp == pmp) {
			fprintf(fp, "\t\t\t\tload +0x%"PRIx64" \"%s-hv.md\";\n",
			    zeus.next_hvmd_fp->paddr - pmp->pa_base,
			    LEG_PREFIX);
		}

		fprintf(fp, "\t\t\t}\n");
	}

	/*
	 * Create the ROMs described in the PRI.
	 */

	conf_gen_roms(fp);

	/* TOD */
	fprintf(fp,
	    "\t\t\t// TOD (XXX hack)\n"
	    "\t\t\tdevice \"dumbtod\" 0x%lx +0x%lx;\n\n",
	    TOD_BASE, TOD_SIZE);

	/* serial out */
	fprintf(fp,
	    "\t\t\t// serial out\n"
	    "\t\t\tdevice \"dumbserial\" 0x%lx +0x%lx;\n\n",
	    HVUART_BASE, DUMBSERIAL_SIZE);

	/* LDC Bridge device */
	if (legion_state.xdc_bridge_addrp != NULL) {
		fprintf(fp, "\t\t\t// XDC Bridge device\n"
		    "\t\t\tdevice \"xdcbridge\" 0x%lx +0x%lx {\n",
		    legion_state.xdc_bridge_base,
		    legion_state.xdc_bridge_size);

		fprintf(fp, "\t\t\t\t// trace csr put "
		    "get dgram interrupt io;\n");
		fprintf(fp, "\t\t\t\tservice hvctl tcp \"%s\" listener;\n",
		    legion_state.xdc_bridge_addrp);
		fprintf(fp, "\t\t\t\tlink 2047 hvctl;\n\t\t\t}\n\n");
	}

	/* Compute total number of consoles */
	nconsoles = CHAIN_LENGTH(zeus.ldoms);
	if (nconsoles < legion_state.min_guests)
		nconsoles = legion_state.min_guests;

	for (sqr = 1; (sqr * sqr) < nconsoles; sqr++) /* LINTED */
		;

	xscale = 1024 / sqr;
	yscale = 800 / sqr;

	/* Now generate the LDom-specific Legion info */
	for (i = 0, ldp = CHAIN_START(zeus.ldoms); ldp != NULL;
	    ++i, ldp = CHAIN_NEXT(ldp->link)) {

		/* Skip unbound LDoms */
		if (ldp->state == INACTIVE)
			continue;

		/* increment addresses for next guest */
		uart = UART_BASE + UART_INC*ldp->idp->id;
		nvbase = NV_BASE + NV_INC*ldp->idp->id;
		diskpa = DISK_BASE + DISK_INC*ldp->idp->id;

		fprintf(fp,
		    "\t\t\t/////////////////////////////////////////\n"
		    "\t\t\t// Partition %d\n"
		    "\t\t\t/////////////////////////////////////////\n"
		    "\n", i+1);

		/* Console */
		fprintf(fp,
		    "\t\t\t// console\n"
		    "\t\t\tdevice \"dumbserial\" 0x%"PRIx64" +0x%"PRIx64" {\n"
		    "\t\t\t\txterm \"/pkg/X11/bin/xterm -l -lf guest%d.log "
		    "-bg black -fg yellow -geometry 80x25+%d+%d "
		    "-T \'%s %d\' -e netcons %%s %%d &\""
		    ";\n"
		    "\t\t\t}\n\n", uart, UART_SIZE, i + 1,
		    50 + xscale * (i % sqr),
		    30 + yscale * (i / sqr), ldp->namep, i);

#if 1 /* { FIXME: Remove the NVRAM stuff eventually */
		/* nvram */
		fprintf(fp,
		    "\t\t\t// nvram\n"
		    "\t\t\tdevice \"memory\" 0x%"PRIx64" +0x%"PRIx64" {\n"
		    "\t\t\t\t// private;\n"
		    "\t\t\t\tload +0x0 \"%s-nvram-%s.bin\";\n"
		    "\t\t\t}\n\n", nvbase, NV_SIZE, LEG_PREFIX, ldp->namep);
#endif /* } */

		/* Disk */
		/* FIXME: Should get disk image name from config file? */
		fprintf(fp,
		    "\t\t\t// disk\n"
		    "\t\t\tdevice \"memory\" 0x%"PRIx64" +0x%"PRIx64" {\n"
		    "\t\t\t\tvirtual_disk;\n"
		    "\t\t\t\tload s0 rom \"%s-disk-%s.img\";\n"
		    "\t\t\t}\n\n",
		    diskpa, DISK_SIZE, LEG_PREFIX, ldp->namep);

	}

	/* Any additional guests required ? */
	/* If so select the next PIDs to be allocated */

	idp = CHAIN_START(zeus.free_ldom_ids.free_list);
	for (; i < legion_state.min_guests; i++) {

		/* increment addresses for next guest */
		uart = UART_BASE + UART_INC * idp->id;
		nvbase = NV_BASE + NV_INC * idp->id;
		diskpa = DISK_BASE + DISK_INC * idp->id;

		fprintf(fp,
		    "\t\t\t// console\n"
		    "\t\t\tdevice \"dumbserial\" 0x%lx +0x%lx {\n"
		    "\t\t\t\txterm \"/pkg/X11/bin/xterm -l -lf guest%d.log "
		    "-bg black -fg yellow -geometry 80x25+%d+%d "
		    "-T \'%s %d\' -e netcons %%s %%d &\""
		    ";\n"
		    "\t\t\t}\n\n",
		    uart, UART_SIZE, i + 1, 50 + xscale * (i % sqr),
		    30 + yscale * (i / sqr), "Guest Console", i);

#if 1 /* { FIXME: Remove the NVRAM stuff eventually */
		/* nvram */
		fprintf(fp,
		    "\t\t\t// nvram\n"
		    "\t\t\tdevice \"memory\" 0x%lx +0x%lx {\n"
		    "\t\t\t\t// private;\n"
		    "\t\t\t\tload +0x0 \"%s-nvram-%02d.bin\";\n"
		    "\t\t\t}\n\n", nvbase, NV_SIZE, LEG_PREFIX, i);
#endif /* } */

		/* Disk */
		/* FIXME: Should get disk image name from config file? */
		fprintf(fp,
		    "\t\t\t// disk\n"
		    "\t\t\tdevice \"memory\" 0x%lx +0x%lx {\n"
		    "\t\t\t\tvirtual_disk;\n"
		    "\t\t\t\tload s0 rom \"%s-disk-%02d.img\";\n"
		    "\t\t\t}\n\n", diskpa, DISK_SIZE, LEG_PREFIX, i);

		idp = CHAIN_NEXT(idp->link);
	}

	/* close everything up */
	fprintf(fp, "\t\t}\n\t}\n}\n");

	fclose(fp);
	Xfree(nodelistp);

	return (0);
}
