/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <stdlib.h>

#include "basics.h"
#include "fatal.h"
#include <mdesc.h>
#include <mdesc_impl.h>
#include "ldom.h"
#include "backend.h"
#include "allocate.h"
#include "option.h"
#include "zpri.h"
#include "support.h"
#include "zprintf.h"

/*
 * Backend for Zeus running in an LDom.
 */

static void	testldom_init(void);
static void	testldom_start(void);
static bool_t	testldom_process_option(int *ip, int argc, char **argv);
static md_t	*testldom_get_pri(void);
static void	testldom_init_pri(md_t *pri);
static void	testldom_get_hv_state(void);
static md_t	*testldom_get_guest_md(ldom_t *ldp);
static void	testldom_aug_guestmd(ldom_t *ldp, md_t *mdp);
static void	testldom_dl_guestmd(ldom_t *, uint64_t paddr,
			uint8_t *bufp, int size);
static void	testldom_aug_hvmd(md_t *);
static void	testldom_dl_hvmd(uint64_t paddr, uint8_t *bufp,
			int size, hvmd_type_t md_type);
static void	testldom_parse_hv_guest(ldom_t *ldp,
			md_node_t *guest, md_t *hvmd);
static void	testldom_aug_hv_guest(ldom_t *ldp,
			md_node_t *guest, md_t *hvmd);
static void	testldom_config_guests(config_flag_t flag);
static void	testldom_start_new_guests(void);
static void	testldom_stop_guest(ldom_t *ldp);
static int	testldom_get_guest_state(ldom_t *ldp, int which,
			ldom_hv_state_t *state);
static void	testldom_banner(response_t *resp);
static int	testldom_get_guest_utilization(ldom_t *ldp,
			rs_guest_util_t *util);
static int	testldom_get_vcpu_state(resid_t *vcpu, rs_vcpu_state_t *vstate);
static int	testldom_get_mem_state(resid_t *mem, rs_mem_state_t *mstate);
static int	testldom_panic_guest(ldom_t *ldp, response_t *resp);
static int	testldom_read_mem(uint64_t pa, uint8_t *bufp, int len);
static int	testldom_write_mem(uint64_t pa, uint8_t *bufp, int len);
static int64_t	testldom_scrub_mem(uint64_t pa, uint64_t len);
static int	testldom_mem_dr_verify(ldom_t *ldp, uint64_t paddr,
			uint64_t raddr, uint64_t size);
static void	testldom_fini(void);

backend_t testldom_be = {
	"testldom",
	"Test backend for simulating LDom Manager running on an LDom",
	testldom_init,
	testldom_start,
	testldom_process_option,
	testldom_get_pri,
	testldom_init_pri,
	testldom_get_hv_state,
	testldom_get_guest_md,
	testldom_aug_guestmd,
	testldom_dl_guestmd,
	testldom_aug_hvmd,
	testldom_dl_hvmd,
	testldom_parse_hv_guest,
	testldom_aug_hv_guest,
	testldom_config_guests,
	testldom_start_new_guests,
	testldom_stop_guest,
	testldom_get_guest_state,
	testldom_banner,
	testldom_get_guest_utilization,
	testldom_get_vcpu_state,
	testldom_get_mem_state,
	testldom_panic_guest,
	testldom_read_mem,
	testldom_write_mem,
	testldom_scrub_mem,
	testldom_mem_dr_verify,
	testldom_fini,
	BACKEND_MAGIC
};

extern backend_t legion_be;

static int
testldom_mem_dr_verify(ldom_t *ldp, uint64_t paddr, uint64_t raddr,
    uint64_t size)
{
	return (0);
}

static void
testldom_init(void)
{
	DBG(zprintf("testldom_init()\n"));

	zeus.allow_ctrl_domain_reconfig = true;
	zeus.simulating = true;
	option.in_place = true;
}

static void
testldom_start(void)
{
	DBG(zprintf("testldom_start()\n"));

	/* This routine initially left empty */
}

static void
testldom_banner(response_t *resp)
{
	rprintf(resp, "testldom backend banner info\n");
}

/*
 * Backend specific command line options.
 */

static bool_t
testldom_process_option(int *ip, int argc, char **argv)
{
	return (legion_be.process_option(ip, argc, argv));
}

static md_t *
testldom_get_pri(void)
{
	/* delegate to the Legion backend */
	return (legion_be.get_pri());
}

static void
testldom_init_pri(md_t *pri)
{
	/* delegate to the Legion backend */
	legion_be.init_pri(pri);
}

static md_t *
testldom_get_guest_md(ldom_t *ldp)
{
	/* delegate to Legion backend */
	return (legion_be.get_guest_md(ldp));
}

static void
testldom_get_hv_state(void)
{
	DBG(zprintf("testldom_get_hv_state()\n"));

	/* delegate to Legion backend */
	legion_be.get_hv_state();
}

/*
 * This callback provides the opportunity to add backend specific features to
 * a guest MD It is mostly never used, but here just in case.
 */

static void
testldom_aug_guestmd(ldom_t *ldp, md_t *mdp)
{
}

static void
testldom_dl_guestmd(ldom_t *ldp, uint64_t paddr, uint8_t *bufp, int mdsize)
{
	DBG(zprintf("testldom_dl_guestmd()\n"));

	legion_be.dl_guestmd(ldp, paddr, bufp, mdsize);
}

static void
testldom_aug_hvmd(md_t *mdp)
{
	DBG(zprintf("testldom_aug_hvmd()\n"));
}

static void
testldom_dl_hvmd(uint64_t paddr, uint8_t *bufp,
	int size, hvmd_type_t md_type)
{
	DBG(zprintf("testldom_dl_hvmd()\n"));

	legion_be.dl_hvmd(paddr, bufp, size, md_type);
}

static void
testldom_parse_hv_guest(ldom_t *ldp, md_node_t *guest, md_t *hvmd)
{
	DBG(zprintf("testldom_parse_hv_guest\n"));
}

static void
testldom_aug_hv_guest(ldom_t *ldp, md_node_t *guest, md_t *hvmd)
{
	DBG(zprintf("testldom_aug_hv_guest\n"));
}

static void
testldom_config_guests(config_flag_t flag)
{
	DBG(zprintf("testldom_config_guests(%d)\n", flag));
}

static void
testldom_start_new_guests(void)
{
	ldom_t *ldp;

	DBG(zprintf("testldom_start_new_guests()\n"));

	CHAIN_ITER(zeus.inprogress, ldp, inprog) {
		ASSERT(ldp->state == BOUND);
		ldp->state = ACTIVE;
	}
}

static void
testldom_stop_guest(ldom_t *ldp)
{
	DBG(zprintf("testldom_stop_guest - %s\n", ldp->namep));
	ldp->state = BOUND;
}

static int
testldom_get_guest_state(ldom_t *ldp, int which, ldom_hv_state_t *state)
{
	DBG(zprintf("testldom_get_guest_state - %s\n", ldp->namep));
	return (legion_be.get_guest_state(ldp, which, state));
}

static int
testldom_get_guest_utilization(ldom_t *ldp, rs_guest_util_t *util)
{
	return (ENOTSUP);
}

static int
testldom_get_vcpu_state(resid_t *vcpu, rs_vcpu_state_t *vstate)
{
	return (ENOTSUP);
}

static int
testldom_get_mem_state(resid_t *mem, rs_mem_state_t *mstate)
{
	return (ENOTSUP);
}

static int
testldom_panic_guest(ldom_t *ldp, response_t *resp)
{
	if (resp)
		rprintf(resp, gettext("The panic command is not supported\n"));

	return (ENOTSUP);
}

/*ARGSUSED*/
static int
testldom_read_mem(uint64_t pa, uint8_t *bufp, int len)
{
	return (len);
}

/*ARGSUSED*/
static int
testldom_write_mem(uint64_t pa, uint8_t *bufp, int len)
{
	return (len);
}

/*ARGSUSED*/
static int64_t
testldom_scrub_mem(uint64_t pa, uint64_t len)
{
	return (len);
}

static void
testldom_fini(void)
{
	DBG(zprintf("testldom_fini()\n"));
}
