/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_CHAIN_H_
#define	_CHAIN_H_

#ifdef __cplusplus
extern "C" {
#endif

#include "fatal.h"

/*
 * Simple chain support.
 *
 * start and end take the form: struct.chain.startp
 *
 * links take the form:	struct.link.nextp
 */

#define	DEFAULTCHAIN	{ NULL, NULL, 0 }

#define	CHAIN(_type, _chain)						\
	struct {							\
		_type	*startp;					\
		_type	*endp;						\
		int	count;						\
	} _chain

#define	CHAIN_LINK(_type, _link)					\
	struct {							\
		_type	* prevp;					\
		_type	* nextp;					\
	} _link

#define	CHAIN_ITER(_chain, _itv, _link)					\
	for ((_itv) = CHAIN_START(_chain); (_itv) != NULL;		\
		(_itv) = CHAIN_NEXT((_itv)->_link))

#define	CHAIN_REVERSE_ITER(_chain, _itv, _link)				\
	for ((_itv) = CHAIN_END(_chain); (_itv) != NULL;		\
		(_itv) = CHAIN_PREV((_itv)->_link))

#define	CHAIN_ITER_VOL(_chain, _itv, _nitv, _link)			\
	for ((_itv) = CHAIN_START(_chain),				\
	    (_nitv) = (_itv) ? CHAIN_NEXT((_itv)->_link) : NULL;	\
	    (_itv) != NULL;						\
	    (_itv) = (_nitv),						\
	    (_nitv) = (_itv) ? CHAIN_NEXT((_itv)->_link) : NULL)	\

#define	CHAIN_DOUBLE_ITER(_ch1, _itv1, _link1, _link2, _itv2, _link3)	\
	CHAIN_ITER(_ch1, _itv1, _link1)					\
		CHAIN_ITER(_itv1->_link2, _itv2, _link3)

#define	CHAIN_INIT(_name)						\
	(_name).startp = NULL;						\
	(_name).endp = NULL;						\
	(_name).count = 0;

#define	CHAIN_START(_name)	((_name).startp)
#define	CHAIN_END(_name)	((_name).endp)
#define	CHAIN_LENGTH(_name)	((_name).count)
#define	CHAIN_NEXT(_chain_link)	((_chain_link).nextp)
#define	CHAIN_PREV(_chain_link)	((_chain_link).prevp)

#define	CHAIN_LINK_INIT(_chain_link)					\
	do {								\
		(_chain_link).nextp =					\
		(_chain_link).prevp = NULL;				\
	} while (0)

/*
 * Add node _nodep to the end of _chain via _link.
 * The node's link does not need to be initialized before calling this.
 */
#define	CHAIN_ADD(_chain, _nodep, _link)				\
	do {								\
		if ((_chain).startp == NULL) {				\
			(_chain).startp = (_nodep);			\
		} else {						\
			(_chain).endp->_link.nextp = (_nodep);		\
		}							\
		(_nodep)->_link.prevp = (_chain).endp;			\
		(_chain).endp = (_nodep);				\
		(_nodep)->_link.nextp = NULL;				\
		(_chain).count ++;					\
	} while (0)

#define	CHAIN_ADD_SANITY(_chain, _nodep, _link, _temp)			\
	do {								\
			CHAIN_ITER(_chain, _temp, _link) {		\
				if (_temp == _nodep)			\
					fatal("Adding dup node!!");	\
			}						\
			CHAIN_ADD(_chain, _nodep, _link);		\
	} while (0)

#define	CHAIN_ADD_FIRST(_chain, _nodep, _link)				\
	do {								\
		if ((_chain).startp == NULL) {				\
			(_chain).startp = (_nodep);			\
			(_chain).endp = (_nodep);			\
			CHAIN_LINK_INIT((_nodep)->_link);		\
		} else {						\
			(_nodep)->_link.prevp = NULL;			\
			(_nodep)->_link.nextp = (_chain).startp;	\
			(_nodep)->_link.nextp->_link.prevp = (_nodep);	\
			(_chain).startp = (_nodep);			\
		}							\
		(_chain).count ++;					\
	} while (0)

#define	CHAIN_INSERT_AFTER(_chain, _here, _nodep, _link)		\
	do {								\
		if ((_chain).startp == NULL || (_here) == NULL) {	\
			CHAIN_ADD_FIRST(_chain, _nodep, _link);		\
		} else {						\
			(_nodep)->_link.prevp = (_here);		\
			(_nodep)->_link.nextp = (_here)->_link.nextp;	\
			if ((_here)->_link.nextp != NULL)		\
				(_here)->_link.nextp->_link.prevp = (_nodep); \
			(_here)->_link.nextp = (_nodep);		\
			if ((_chain).endp == (_here))			\
				(_chain).endp = (_nodep);		\
			(_chain).count ++;				\
		}							\
	} while (0)

#define	CHAIN_UNLINK(_chain, _nodep, _link)				\
	do {								\
		if ((_chain).startp == (_nodep))			\
			(_chain).startp = (_nodep)->_link.nextp;	\
		else							\
			(_nodep)->_link.prevp->_link.nextp =		\
				(_nodep)->_link.nextp;			\
		if ((_chain).endp == (_nodep))				\
			(_chain).endp = (_nodep)->_link.prevp;		\
		else							\
			(_nodep)->_link.nextp->_link.prevp =		\
				(_nodep)->_link.prevp;			\
		(_chain).count --;					\
		SANITY((_nodep)->_link.prevp = NULL);			\
		SANITY((_nodep)->_link.nextp = NULL);			\
	} while (0)

/*
 * Add node _nodep into sorted chain _chain via _link.
 * _itv and _prev are variables used to iterate through the chain and their
 * values are clobbered. _cmp is the sort function used to order the list.
 * The node's link does not need to be initialized before calling this.
 */
#define	CHAIN_ADD_SORTED(_chain, _nodep, _itv, _prev, _link, _cmp)	\
	do {								\
		(_prev) = NULL;						\
		CHAIN_ITER((_chain), (_itv), _link) {			\
			if ((_cmp)((_nodep), (_itv)) < 0)		\
				break;					\
			(_prev) = (_itv);				\
		}							\
									\
		CHAIN_INSERT_AFTER((_chain), (_prev), (_nodep), _link);	\
	} while (0)

#ifdef __cplusplus
}
#endif

#endif	/* _CHAIN_H_ */
