/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <port.h>
#include <thread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

#include "allocate.h"
#include "basics.h"
#include "clp.h"
#include "cli_conn.h"
#include "fatal.h"
#include "option.h"
#include "sequencer.h"
#include "support.h"
#include "zprintf.h"
#include "ldm_protocol.h"
#include "auth_ldom.h"
#include "control_class.h"
#include "mig_class.h"

/*
 * supported versions tables
 */
static prot_version_t ldm_sunmc_protocol_versions[] = {
	{ 1, 0 }
};

static prot_version_t ldm_control_protocol_versions[] = {
	{ 1, 0 }
};

static prot_version_t ldm_cli_protocol_versions[] = {
	{ 2, 0 },
};

static prot_version_t ldm_mig_protocol_versions[] = {
	{ 1, 1 },
	{ 1, 0 }
};

/*
 * Any commands longer than this will be summarily rejected.
 */
#define	MAX_CMD_LEN	(4*(1024 * 1024))

static int evport;		/* event port for client IO events */

/* forward declarations */
static void *write_msg(cli_conn_t *conn);
static void ldmd_sig_handler(int sig, siginfo_t *info, void *p);

static void
close_conn(cli_conn_t *conn)
{
	if (conn == NULL)
		return;

	(void) close(conn->fd);
	Xfree(conn->buf);
	Xfree(conn->cookie);
	Xfree(conn);
}

void
protocol_error(cli_conn_t *conn, uint32_t errcode)
{
	ldm_hdr_t *hdr;
	ldm_error_t *err;
	int msg_len;

	if (conn->buf != NULL)
		free(conn->buf);

	msg_len = sizeof (ldm_hdr_t) + sizeof (ldm_error_t);
	conn->buf = Xmalloc(msg_len);

	hdr = (ldm_hdr_t *)conn->buf;
	hdr->msg_type = hton32(LDM_ERROR);
	hdr->payload_len = hton32(sizeof (ldm_error_t));
	err = (ldm_error_t *)(conn->buf + sizeof (ldm_hdr_t));
	err->error_code = hton32(errcode);

	conn->state = CONN_ERR;
	conn->bufp = conn->buf;
	conn->left = msg_len;
	conn->callback = write_msg;

	if (port_associate(evport, PORT_SOURCE_FD, conn->fd, POLLOUT, conn) < 0)
		efatal("port_associate");
}

static void
client_handle_init_req(cli_conn_t *conn)
{
	ldm_hdr_t *hdr;
	ldm_init_req_t *init_req;
	ldm_init_ack_t *init_ack;
	ldm_init_nack_t *init_nack;
	uint16_t req_major, req_minor;
	uint16_t major;
	uint16_t minor = 0;
	int len, nvers;
	prot_version_t *ldm_protocol_versions;

	ASSERT(conn != NULL);

	if (conn->hdr.payload_len != sizeof (ldm_init_req_t)) {
		protocol_error(conn, LDM_ERROR_INVALID);
		return;
	}

	ASSERT(conn->buf != NULL);
	init_req = (ldm_init_req_t *)conn->buf;
	conn->class = ntoh32(init_req->client_class);
	req_major = ntoh16(init_req->major_vers);
	req_minor = ntoh16(init_req->minor_vers);

	Xfree(conn->buf);
	conn->buf = NULL;

	switch (conn->class) {
	case LDM_CLASS_CLI:
		ldm_protocol_versions = ldm_cli_protocol_versions;
		nvers = NUM_VERS(ldm_cli_protocol_versions);
		break;
	case LDM_CLASS_CONTROL:
		ldm_protocol_versions = ldm_control_protocol_versions;
		nvers = NUM_VERS(ldm_control_protocol_versions);
		break;
	case LDM_CLASS_SUNMC:
		ldm_protocol_versions = ldm_sunmc_protocol_versions;
		nvers = NUM_VERS(ldm_sunmc_protocol_versions);
		break;
	case LDM_CLASS_MIG:
		ldm_protocol_versions = ldm_mig_protocol_versions;
		nvers = NUM_VERS(ldm_mig_protocol_versions);
		break;
	default:
		protocol_error(conn, LDM_ERROR_INVALID);
		DBG(zprintf("%s: invalid class\n", __func__));
		return;
	}

	if (negotiate_version(nvers, ldm_protocol_versions, req_major, &major,
	    &minor)) {
		conn->state = CONN_UP;
		len = sizeof (ldm_hdr_t) + sizeof (ldm_init_ack_t);
		conn->buf = Xmalloc(len);
		hdr = (ldm_hdr_t *)conn->buf;
		hdr->msg_type = hton32(LDM_INIT_ACK);
		hdr->payload_len = hton32(sizeof (ldm_init_ack_t));
		init_ack = (ldm_init_ack_t *)(conn->buf + sizeof (ldm_hdr_t));
		if (minor < req_minor)
			init_ack->minor_vers = hton16(minor);
		else
			init_ack->minor_vers = hton16(req_minor);
	} else {
		len = sizeof (ldm_hdr_t) + sizeof (ldm_init_nack_t);
		conn->buf = Xmalloc(len);
		hdr = (ldm_hdr_t *)conn->buf;
		hdr->msg_type = hton32(LDM_INIT_NACK);
		hdr->payload_len = hton32(sizeof (ldm_init_nack_t));
		init_nack = (ldm_init_nack_t *)(conn->buf + sizeof (ldm_hdr_t));
		init_nack->major_vers = hton16(major);
	}

	conn->left = len;
	conn->bufp = conn->buf;
	conn->callback = write_msg;
	if (port_associate(evport, PORT_SOURCE_FD, conn->fd, POLLOUT, conn) < 0)
		efatal("port_associate");
}

/*
 * handle_op_req() is for those classes that use an ldm_op_req_t msg_type. It
 * resets conn->bufp to point to the rq_data field in the ldm_op_req_t
 * structure.
 */
static int
handle_op_req(cli_conn_t *conn)
{
	ldm_op_req_t *req;

	if (conn->hdr.payload_len < sizeof (*req)) {
		protocol_error(conn, LDM_ERROR_INVALID);
		return (-1);
	}

	req = (ldm_op_req_t *)conn->buf;
	conn->reqid = req->rq_id;	/* sent back as-is so don't convert */

	/* set bufp to point to the start of the request data */
	conn->bufp = (char *)req->rq_data;

	return (0);
}

static void
cli_req(cli_conn_t *conn)
{
	/* switch here in case the CLI adds additional message types */
	switch (conn->hdr.msg_type) {
	case LDM_OP_REQ:
		if (handle_op_req(conn) != 0)
			return;
		switch (conn->class) {
		case LDM_CLASS_CLI:
			seq_enqueue_cmd(conn, SEQ_CLI_CMD);
			break;
		default:
			protocol_error(conn, LDM_ERROR_UNEXPECTED);
			DBG(zprintf("cli_req: invalid class %d\n",
			    conn->class));
		}
		break;
	default:
		protocol_error(conn, LDM_ERROR_UNEXPECTED);
		DBG(zprintf("cli_req: invalid op type %d\n",
		    conn->hdr.msg_type));
	}
}

static void
control_req(cli_conn_t *conn)
{
	/* switch here in case control class adds additional message types */
	switch (conn->hdr.msg_type) {
	case LDM_CNTRL_PRI_REQ:
		seq_enqueue_cmd(conn, SEQ_PRI_REQ);
		break;
	case LDM_CNTRL_XML_REQ:
		seq_enqueue_cmd(conn, SEQ_XML_CMD);
		break;
	default:
		protocol_error(conn, LDM_ERROR_UNEXPECTED);
		DBG(zprintf("control_req: invalid op type %d\n",
		    conn->hdr.msg_type));
	}
}

static void
sunmc_req(cli_conn_t *conn)
{
	/* switch here in case SUNMC adds additional message types */
	switch (conn->hdr.msg_type) {
	case LDM_OP_REQ:
		if (handle_op_req(conn) != 0)
			return;
		handle_sunmc_client(conn);
		break;
	default:
		protocol_error(conn, LDM_ERROR_UNEXPECTED);
		DBG(zprintf("xml_req: invalid op type %d\n",
		    conn->hdr.msg_type));
	}
}

static void
handle_msg(cli_conn_t *conn)
{
	/*
	 * If this is a init_req message then we need to call
	 * client_handle_init_req() so that the connection can be setup.
	 */
	if (conn->hdr.msg_type == LDM_INIT_REQ) {
		client_handle_init_req(conn);
		return;
	}

	if (conn->state != CONN_UP) {
		protocol_error(conn, LDM_ERROR_UNEXPECTED);
		return;
	}

	/*
	 * If this is not an init_req message then goto the connection class
	 * routine.
	 */
	switch (conn->class) {
	case LDM_CLASS_CLI:
		init_response(&conn->resp);
		cli_req(conn);
		break;
	case LDM_CLASS_CONTROL:
		init_response(&conn->resp);
		control_req(conn);
		break;
	case LDM_CLASS_SUNMC:
		init_response(&conn->resp);
		sunmc_req(conn);
		break;
	default:
		protocol_error(conn, LDM_ERROR_UNEXPECTED);
		DBG(zprintf(
		    "handle_msg: bad class = %d\n", conn->class));
		return;
	}
}

/*
 * Connection callback to read a client message body.
 */
static void *
read_msg(cli_conn_t *conn)
{
	int count;

	count = read(conn->fd, conn->bufp, conn->left);

	if (count <= 0) {
		/* broken connection */
		close_conn(conn);
	} else if (count < conn->left) {
		/* partial read, wait for more to read */
		conn->left -= count;
		conn->bufp += count;

		if (port_associate(evport, PORT_SOURCE_FD, conn->fd,
		    POLLIN, conn) < 0)
			efatal("port_associate");
	} else {
		handle_msg(conn);
	}

	return (NULL);
}

/*
 * Connection callback to read a CLI command header.
 */
static void *
read_msg_header(cli_conn_t *conn)
{
	int count;

	count = read(conn->fd, conn->bufp, conn->left);

	if (count <= 0) {
		/* connection broken */
		close_conn(conn);
		return (NULL);
	}

	if (count < conn->left) {
		/* partial read, wait for more to read */
		conn->left -= count;
		conn->bufp += count;
	} else {
		/* all of header has been read */
		conn->hdr.msg_type = ntoh32(conn->hdr.msg_type);
		conn->hdr.payload_len = ntoh32(conn->hdr.payload_len);

		if (conn->hdr.payload_len == 0 ||
		    conn->hdr.payload_len > MAX_CMD_LEN) {
			protocol_error(conn, LDM_ERROR_INVALID);
			return (NULL);
		}

		/* now read in the command */
		conn->buf = Xmalloc(conn->hdr.payload_len);
		conn->bufp = conn->buf;
		conn->left = conn->hdr.payload_len;
		conn->callback = read_msg;
	}

	if (port_associate(evport, PORT_SOURCE_FD, conn->fd, POLLIN, conn) < 0)
		efatal("port_associate");

	return (NULL);
}

/*
 * Connection callback to write a message to the client.
 */
static void *
write_msg(cli_conn_t *conn)
{
	int count;

	count = write(conn->fd, conn->bufp, conn->left);

	if (count <= 0) {
		/* broken connection */
		close_conn(conn);
	} else if (count < conn->left) {
		/* partial write - wait to write more */
		conn->left -= count;
		conn->bufp += count;
		if (port_associate(evport, PORT_SOURCE_FD, conn->fd,
		    POLLOUT, conn) < 0)
			efatal("port_associate");
	} else {
		/*
		 * Done. If the connection incurred a prior protcol error shut
		 * it down, otherwise free the response and message buffers if
		 * any, and prepare to read next message from the client.
		 */
		if (conn->state == CONN_ERR) {
			close_conn(conn);
		} else {
			/*
			 * Free buf which is also set to bufp and the response
			 * memory address (either from SunMC or the CLI
			 */
			if (conn->buf != NULL) {
				Xfree(conn->buf);
				conn->buf = NULL;
			}

			conn->callback = read_msg_header;
			conn->bufp = (char *)&conn->hdr;
			conn->left = sizeof (ldm_hdr_t);

			/* poll for data available from the client */
			if (port_associate(evport, PORT_SOURCE_FD, conn->fd,
			    POLLIN, conn) < 0)
				efatal("port_associate");
		}
	}

	return (NULL);
}

/*
 * This write response is for all other classes other than the CLI. It assumes
 * that the caller is sending in a buffer complete with header.
 */
void
client_write_buffer(cli_conn_t *conn, void *bufp)
{
	ldm_hdr_t	*hdr;
	response_t	*resp;

	hdr = bufp;
	resp = conn->resp;

	conn->left = sizeof (ldm_hdr_t) + hdr->payload_len;
	conn->buf = conn->bufp = bufp;
	conn->callback = write_msg;

	Xfree(resp->bufp);
	Xfree(resp);

	if (port_associate(evport, PORT_SOURCE_FD, conn->fd, POLLOUT,
	    conn) < 0)
		efatal("port_associate");
}

#if ZDBG
/*
 * Print a comma separated list of major minor pairs
 */
static void
dump_version_list(response_t *resp, prot_version_t *versions, int nversions)
{
	int i;

	for (i = 0; i < nversions; i++) {
		rprintf(resp, "v%d.%d",
		    versions[i].major_number,
		    versions[i].minor_number);
		if (i < nversions - 1)
			rprintf(resp, ", ");
	}
	rprintf(resp, "\n");
}

/*
 * Dump out the latest versions of the protocols
 * called by display_banner() when in debug mode
 */
void
dump_protocol_versions(response_t *resp)
{
	ASSERT(resp != NULL);

	rprintf(resp, "\nLDMD Internal Protocol Versions:\n");

	rprintf(resp, "    CLI:\t");
	dump_version_list(resp, ldm_cli_protocol_versions,
	    NUM_VERS(ldm_cli_protocol_versions));

	rprintf(resp, "    CONTROL:\t");
	dump_version_list(resp, ldm_control_protocol_versions,
	    NUM_VERS(ldm_control_protocol_versions));

	rprintf(resp, "    SUNMC:\t");
	dump_version_list(resp, ldm_sunmc_protocol_versions,
	    NUM_VERS(ldm_sunmc_protocol_versions));

	rprintf(resp, "    MIGRATION:\t");
	dump_version_list(resp, ldm_mig_protocol_versions,
	    NUM_VERS(ldm_mig_protocol_versions));
}

static void
dump_cli_response(cli_conn_t *conn)
{
	char timestamp[16];
	time_t t;

	(void) time(&t);
	(void) cftime(timestamp, "%T", &t);
	zprintf("<== [%s] 0x%p <- %d\n\n",
	    timestamp, (void *)conn, conn->resp->ret);
	(void) fsync(fileno(stderr));
}
#endif

void
client_write_response(cli_conn_t *conn)
{
	ldm_hdr_t *hdr;
	ldm_op_reply_t *reply;
	response_t *resp;
	cli_reply_data_t *data_info;
	int msg_len, text_len;

	ASSERT(conn != NULL);
	resp = conn->resp;

	text_len = resp->length - sizeof (ldm_hdr_t) - sizeof (ldm_op_reply_t) -
	    sizeof (data_info->length) - sizeof (data_info->condition_code) -
	    sizeof (data_info->text_length);

	if (text_len > 0) {
		/* Need to account for NULL terminator in response string */
		resp->length++;
		text_len++;
	}
	msg_len = resp->length;

	hdr = (ldm_hdr_t *)resp->bufp;
	hdr->msg_type = hton32(LDM_OP_REPLY);
	resp->length -= sizeof (ldm_hdr_t);
	hdr->payload_len = hton32(resp->length);
	reply = (ldm_op_reply_t *)(resp->bufp + sizeof (ldm_hdr_t));
	resp->length -= sizeof (ldm_op_reply_t);
	reply->rp_len = hton32(resp->length);
	reply->rp_reqid = conn->reqid;

	/*
	 * The response format is:
	 * <4 bytes length of what follows = 4+4+N>
	 * <4 bytes condition code>
	 * <4 bytes text block length = N>
	 * <N bytes of text block>
	 *
	 * The text block is already in the buffer. We just need to fill in
	 * the first three fields at the head of the buffer before sending the
	 * response.
	 */

	data_info = (cli_reply_data_t *)reply->rp_data;

	/* resp->length is the length of the header plus the text block */
	data_info->length = hton32(resp->length - 4);
	data_info->condition_code = hton32(resp->ret);
	data_info->text_length = hton32(text_len);

	conn->left = msg_len;
	conn->buf = conn->bufp = resp->bufp;
	conn->callback = write_msg;

	DBG(dump_cli_response(conn));

	Xfree(resp);

	if (port_associate(evport, PORT_SOURCE_FD, conn->fd, POLLOUT, conn) < 0)
		efatal("port_associate");
}

/*
 * Callback to accept a new client connection.
 */
void *
new_client(cli_conn_t *conn)
{
	int alen = 0;
	cli_conn_t *cli_conn;
	int client_fd;

	client_fd = accept(conn->fd, NULL, &alen);
	if (client_fd < 0) {
		ewarning("accept");
	} else {
		cli_conn = Xmalloc(sizeof (cli_conn_t));
		cli_conn->fd = client_fd;
		cli_conn->euid = socket_peer_euid(client_fd);
		cli_conn->conntime = time(NULL);
		cli_conn->buf = NULL;
		cli_conn->state = CONN_INIT;
		cli_conn->callout = NULL;
		cli_conn->xmpp_msg = false;
		CHAIN_INIT(cli_conn->waiters);
		cli_conn->cookie = NULL;

		/* prepare to read the message header from the new client */
		cli_conn->callback = read_msg_header;
		cli_conn->bufp = (char *)&cli_conn->hdr;
		cli_conn->left = sizeof (ldm_hdr_t);

		/* poll on data available from the new client */
		if (port_associate(evport, PORT_SOURCE_FD, cli_conn->fd,
		    POLLIN, cli_conn) < 0)
			efatal("port_associate");
	}

	/* reset to poll for new clients */
	if (port_associate(evport, PORT_SOURCE_FD, conn->fd, POLLIN, conn) < 0)
		efatal("port_associate");

	return (NULL);
}

/*
 * Main loop to handle IO with CLI clients.
 * This is run in its own thread.
 */
/* ARGSUSED */
void *
client_loop(void *unused)
{
	struct sockaddr_in myaddr;
	cli_conn_t *acc_conn, *conn;
	port_event_t ev;
	int sock;
	int on = 1;
	struct sigaction sigact;
	sigset_t sig_set;

	/*
	 * Install the signal handler here so that the sequencer thread
	 * continues to run in the event of a signal. Right now the signals
	 * are used to take care of PM related issues.
	 */
	sigact.sa_handler = ldmd_sig_handler;
	sigact.sa_flags = SA_RESTART | SA_SIGINFO;
	(void) sigemptyset(&sigact.sa_mask);

	/* Register signal handler */
	if (sigaction(SIGINT, &sigact, (struct sigaction *)NULL) < 0) {
		efatal("Warning: sigaction SIGINT");
	}
	if (sigaction(SIGTERM, &sigact, (struct sigaction *)NULL) < 0) {
		efatal("Warning: sigaction SIGTERM");
	}

	/*
	 * We also have to unblock SIGINT and SIGTERM as they were blocked by
	 * the main thread and all child threads created by that main thread
	 * inherit this.  This means that this thread is the only thread which
	 * can handle SIGINT and SIGTERM signals.
	 */
	sigemptyset(&sig_set);
	sigaddset(&sig_set, SIGINT);
	sigaddset(&sig_set, SIGTERM);
	thr_sigsetmask(SIG_UNBLOCK, &sig_set, NULL);

	DBG7((void) memset(&ev, 0, sizeof (ev)));

	if ((evport = port_create()) < 0)
		efatal("port_create");

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		efatal("socket");

	if (setsockopt(sock, SOL_SOCKET,
	    SO_REUSEADDR, (char *)&on, sizeof (on)) < 0)
		efatal("turning on REUSEADDR");

	myaddr.sin_family = AF_INET;
	myaddr.sin_port = htons(option.port);
	myaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);

	if (bind(sock, (struct sockaddr *)&myaddr, sizeof (myaddr)) < 0)
		efatal("bind");

	(void) listen(sock, 3);

	DBG(zprintf("Listening on port %d\n", option.port));

	/* create a connection which handles accepting new clients */
	acc_conn = Xmalloc(sizeof (cli_conn_t));
	acc_conn->fd = sock;
	acc_conn->callback = new_client;

	if (port_associate(evport, PORT_SOURCE_FD, sock, POLLIN, acc_conn) < 0)
		efatal("port_associate");

	/* loop - waiting for and handling IO events */
	while (1) {
		if (port_get(evport, &ev, NULL) < 0)
			efatal("port_get");

		if (ev.portev_source == PORT_SOURCE_FD) {
			conn = ev.portev_user;
			(void) conn->callback(conn);
		} else {
			fatal("Unexpected port event source");
		}
	}

	/*NOTREACHED*/
}

/* ARGSUSED1 */
static void
ldmd_sig_handler(int sig, siginfo_t *info, void *p)
{

	DBG(zprintf("Received sig=%d\n", sig));

	if (option.allow_pm) {
		seq_enqueue_work(pm_stop, NULL, true);
	}

	exit(0);
}
