/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <libintl.h>

#include "basics.h"
#include "allocate.h"
#include "clp.h"
#include "variables.h"
#include "mdesc_impl.h"
#include "ldom.h"
#include "zpri.h"
#include "fatal.h"
#include "zprintf.h"
#include "mac_addrs.h"
#include "frags.h"
#include "vcpu.h"
#include "vmau.h"
#include "memory.h"
#include "vio.h"
#include "master.h"
#include "pcie.h"

/*
 * This is a forceful binding of a specific memory range, from the free memory
 * table.  We return NULL if the memory was not available, or already
 * allocated.
 */
mblock_t *
pull_bmblock(uint64_t basepa, uint64_t size)
{
	mblock_t *mbp;

	ASSERT(size > 0);

	if ((mbp = mem_bind_phys_explicit(basepa, size)) == NULL)
		DBG(zprintf("physical memory 0x%"PRIx64" + 0x%"PRIx64" "
		    "not available.\n", basepa, size));

	return (mbp);
}

/* FIXME: provide CLI to explicitly set boot image */
bool_t
bind_boot_options(ldom_t *ldp, response_t *resp)
{
	pri_image_t *pip;

	if (CHAIN_LENGTH(pri.image_list) > 0) {
		/* find a bootable image and select it for this LDom */
		CHAIN_ITER(pri.image_list, pip, link) {
			if (pip->guest_bootable)
				break;
		}

		if (pip != NULL) {
			ldp->boot_imagep = pip;
		} else {
			if (resp != NULL)
				rprintf(resp,
				    gettext("No bootable image for LDom %s\n"),
				    ldp->namep);
			return (false);
		}
	}

	return (true);
}

/*
 * Check whether a requested IO device could be found in PRI.
 */
bool_t
match_iodev(char *iodev_p, pri_iodevice_t *pio_p)
{
	return ((strcmp(pio_p->name, iodev_p) == 0) ||
	    (pio_p->cli_name && (strcmp(pio_p->cli_name, iodev_p)) == 0));
}

/*
 * Find the pointer to the PRI device with a matching name.
 */
pri_iodevice_t *
find_matching_io(char *iodev_p, int *state)
{
	pri_iodevice_t *pio_p;

	if (state)
		*state = IODEV_NOT_PRESENT;

	CHAIN_ITER(pri.iodevice_list, pio_p, link) {
		if (match_iodev(iodev_p, pio_p)) {
			if (state)
				*state = pio_p->free ?
				    IODEV_UNBOUND : IODEV_BOUND;
			return (pio_p);
		}
	}

	return (NULL);
}

/*
 * Create the structure for a root complex that has been bound
 * to a domain.
 */
bound_iodevice_t *
build_bound_iodevice(pri_iodevice_t *pio,
    resid_t *id, bool_t virt, bool_t io_mmu_bypass)
{
	bound_iodevice_t *bio;

	bio = Xcalloc(1, bound_iodevice_t);
	CHAIN_INIT(bio->dev_list);
	bio->name = Xstrdup(pio->name);
	bio->phys_iodev = pio;
	bio->virtual_pcie_bus = virt;
	bio->id = id;
	bio->no_pcie_reset = pio->no_pcie_reset;
	bio->io_mmu_bypass = io_mmu_bypass;
	return (bio);
}

/*
 * Bind the pcie devices to the target domain using the constraints.
 */
bool_t
bind_pciedevs(ldom_t *ldp, response_t *resp)
{
	pri_iodevice_t *pio;
	pri_pcie_dev_t *pdev;
	req_iodevice_t *rio;
	req_pcie_dev_t *rdev;
	bound_iodevice_t *bio;
	bound_pcie_dev_t *bdev;
	resid_t *resid;

	CHAIN_ITER(ldp->required.io_pcie_devs, rdev, link) {
		if (rdev->bound)
			continue;

		pdev = pcie_find_pri_pciedev(rdev->cli_name, &pio);
		if ((pdev == NULL) || !pdev->free) {
			rprintf(resp, gettext("No free matching pcie device "
			    "for LDom %s, name %s\n"),
			    ldp->namep, rdev->cli_name);
			return (false);
		}

		bio = pcie_biodevice_exists(ldp, pio->cfg_handle);
		if (bio == NULL) {
			bound_iodevice_t *new_bio;

			/* Create a virtual bound_iodevice_t for target ldom */
			resid = alloc_res(pio->res_id_pool);
			new_bio = build_bound_iodevice(pio, resid, true, false);
			CHAIN_ADD(ldp->bound.iodevices, new_bio, ldom);

			/* Create and link a requirement for it. */
			rio = Xcalloc(1, req_iodevice_t);
			rio->name = pio->name;
			rio->bound = true;
			CHAIN_ADD(pio->bound_ios, new_bio, pio);
			new_bio->req_iop = rio;
			bio = new_bio;
		}

		bdev = pcie_build_bound_pciedev(pdev);
		bdev->req_pcie_dev = rdev;
		CHAIN_ADD(bio->dev_list, bdev, ldom);
		CHAIN_ADD(pdev->bound_devs, bdev, pio);
		pcie_realloc_res(bio);
		pdev->free = false;
		rdev->bound = true;
	}

	return (true);
}

/*
 * Bind all the pcie buses using the constraints.
 */
bool_t
bind_io(ldom_t *ldp, response_t *resp)
{
	pri_iodevice_t *pio_p;
	req_iodevice_t *rio_p;
	resid_t *resid;

	CHAIN_ITER(ldp->required.io_devices, rio_p, link) {
		/* For each requirement, attempt to bind */
		if (rio_p->bound)
			continue;
		CHAIN_ITER(pri.iodevice_list, pio_p, link) {
			/* Initial substring match */
			if (pio_p->free && match_iodev(rio_p->name, pio_p)) {
				bound_iodevice_t *bio_p;

				if (rio_p->io_mmu_bypass && !pio_p->is_pciex) {
					rprintf(resp, gettext(
					    "IO MMU bypass not applicable for "
					    "device type of %s\n"),
					    pio_p->device_type);
					return (false);
				}

				resid = alloc_res(pio_p->res_id_pool);
				if (resid == NULL) {
					rprintf(resp, "%s\n", gettext(
					    "Out of available device IDs"));
					return (false);
				}
				bio_p = build_bound_iodevice(
				    pio_p, resid, false, rio_p->io_mmu_bypass);

				if (pio_p->is_pciex) {
					/*
					 * If we are binding an entire bus
					 * then we need to assign all to it.
					 */
					pcie_assign_all_devs(bio_p);
					pcie_update_reqs(ldp, bio_p);
				}

				bio_p->req_iop = rio_p;
				CHAIN_ADD(ldp->bound.iodevices, bio_p, ldom);
				CHAIN_ADD(pio_p->bound_ios, bio_p, pio);
				pio_p->free = false;
				break;
			}
		}
		if (pio_p == NULL) {
			/*
			 * TRANSLATION_NOTE
			 * second %s is the name of the physical I/O device
			 */
			rprintf(resp, gettext(
			    "No free matching I/O device for LDom %s, "
			    "name %s\n"), ldp->namep, rio_p->name);
			return (false);
		} else
			rio_p->bound = true;
	}
	return (true);

}

/*
 * Forcefully bind specific iodevice.
 */
bound_iodevice_t *
pull_biodevice(char *device_type, uint64_t cfghdl, uint64_t res_id,
    bool_t io_mmu_bypass)
{
	pri_iodevice_t *pio_p;
	bound_iodevice_t *bio_p;
	resid_t *rid;

	CHAIN_ITER(pri.iodevice_list, pio_p, link) {

		/* Find specified device */
		if (strcmp(pio_p->device_type, device_type) == 0 &&
		    cfghdl == pio_p->cfg_handle) {

			if (false == pio_p->free) {
				zprintf("Attempt to force bind an already "
				    "allocated iodevice\n");
				return (NULL);
			}
			rid = pull_res_id(pio_p->res_id_pool, res_id);
			if (rid == NULL) {
				fatal("Device res_id %"PRId64" already taken",
				    res_id);
			}
			bio_p = build_bound_iodevice(pio_p,
			    rid, false, io_mmu_bypass);
			CHAIN_ADD(pio_p->bound_ios, bio_p, pio);
			pio_p->free = false;
			return (bio_p);
		}
	}
	zprintf("No matching I/O device %s at cfg_handle %"PRId64"\n",
	    device_type, cfghdl);
	return (NULL);
}

/*
 * Forcefully bind specific iodevice (root complex).  The pcie props are
 * passed in a temporary bound_iodevice_t.
 */
bound_iodevice_t *
pull_pcie_biodevice(char *device_type, uint64_t cfghdl,
    uint64_t res_id, bound_iodevice_t *tbiop, bool_t assign_all)
{
	pri_iodevice_t *pio_p;
	bound_iodevice_t *bio_p;
	bound_pcie_dev_t *bdev;
	pri_pcie_dev_t *pdev;
	resid_t *rid;

	CHAIN_ITER(pri.iodevice_list, pio_p, link) {

		/* Find specified device */
		if (strcmp(pio_p->device_type, device_type) == 0 &&
		    cfghdl == pio_p->cfg_handle) {
			if (!tbiop->virtual_pcie_bus && (!pio_p->free)) {
				zprintf("Attempt to force bind an already "
				    "allocated iodevice\n");
				return (NULL);
			}
			break;
		}
	}
	if (pio_p == NULL) {
		zprintf("No matching I/O device %s at cfg_handle %"PRId64"\n",
		    device_type, cfghdl);
		return (NULL);
	}
	if ((rid = pull_res_id(pio_p->res_id_pool, res_id)) == NULL)
		fatal("Device res_id %"PRId64" already taken", res_id);
	bio_p = build_bound_iodevice(pio_p,
	    rid, tbiop->virtual_pcie_bus, tbiop->io_mmu_bypass);

	if (assign_all) {
		pcie_assign_all_devs(bio_p);
	} else {

		while (CHAIN_LENGTH(tbiop->dev_list) > 0) {
			bdev = CHAIN_START(tbiop->dev_list);
			CHAIN_UNLINK(tbiop->dev_list, bdev, ldom);
			CHAIN_ITER(pio_p->dev_list, pdev, link) {
				if (strcmp(bdev->skel_path,
				    pdev->skel_path) == 0) {
					ASSERT(pdev->free);
					bdev->phys_pcie_dev = pdev;
					bdev->cli_name =
					    Xstrdup(pdev->cli_name);
					CHAIN_ADD(bio_p->dev_list, bdev, ldom);
					CHAIN_ADD(pdev->bound_devs, bdev, pio);
					pdev->free = false;
					break;
				}
			}
			ASSERT(pdev != NULL);
		}
		pcie_alloc_fixed(bio_p, &tbiop->busres);
	}
	CHAIN_ADD(pio_p->bound_ios, bio_p, pio);
	if (!bio_p->virtual_pcie_bus)
		pio_p->free = false;
	return (bio_p);
}

void
undo_partial_binding(ldom_t *ldp)
{
	unbind_an_ldom(ldp);
	vio_unbind(ldp, NULL);

	free_all_staged_ldcs(ldp);

	/*
	 * Clear all flags, no need for MD update and DS notifications now.
	 */
	seq_clear_ldom_flags();
}

/*
 * Unbind a pcie device and unlink it from the parent iodevice
 * (root complex) structure.
 */
void
unbind_pciedev(ldom_t *ldp, bound_iodevice_t *bio, bound_pcie_dev_t *bdev)
{
	pri_pcie_dev_t *pdev;
	req_pcie_dev_t *rdev;

	pdev = bdev->phys_pcie_dev;
	rdev = bdev->req_pcie_dev;
	CHAIN_UNLINK(bio->dev_list, bdev, ldom);
	CHAIN_UNLINK(pdev->bound_devs, bdev, pio);

	/* Free the MSI, EQ and DVMA resources. */
	pcie_realloc_res(bio);
	pdev->free = true;
	rdev->bound = false;
	pcie_destroy_bound_pciedev(bdev);
}

/*
 * Walk the chain of bound pcie device structures under the parent
 * root complex and unbind each device.
 */
static void
unbind_pciedevs(ldom_t *ldp, bound_iodevice_t *bio_p)
{
	bound_pcie_dev_t *bdev;

	/*
	 * Walk the dev_list and unbind pcie_devs
	 */
	while ((bdev = CHAIN_START(bio_p->dev_list)) != NULL) {
		unbind_pciedev(ldp, bio_p, bdev);
	}

}

/*
 * This function creates a new binding for an LDom.  It moves the LDOM state
 * to BOUND, which means the sequencer will generate a hypervisor MD update
 * for this LDom - this update then results in the LDom moving to the STOPPED
 * or ACTIVE state.
 *
 * Returns false if binding fails
 */
bool_t
bind_an_ldom(ldom_t *ldp, response_t *resp)
{
	ASSERT(INACTIVE == ldp->state);

	/*
	 * Allocate an id!
	 */

	ldp->idp = alloc_res(&zeus.free_ldom_ids);
	if (ldp->idp == NULL) {
		rprintf(resp, gettext("No more active LDoms allowed\n"));
		return (false);
	}

	ldp->res_id = alloc_res(&zeus.free_guest_ids);
	if (ldp->res_id == NULL) {
		rprintf(resp, gettext("No more LDom slots available\n"));
		return (false);
	}

	DBG(zprintf("Binding LDom %s, gid %"PRId64", res_id %"PRId64"\n",
	    ldp->namep, ldp->idp->id, ldp->res_id->id));

	/*
	 * Reset the generation number for the guests MD
	 * Auto pre-increment means the first number seen
	 * in an MD is 1.
	 */
	ldp->props.md_generation_num = 0;

	/*
	 * Allocate physical CPU resources from the CPU requirements.
	 */
	cpu_adm_t cpu_adm = {
		.type = CPU_ADM_NUM,
		.num = ldp->required.num_vcpus
		/* CSTYLED */
	};

	if (cons_bind_vcpus(ldp, &cpu_adm, resp) == false)
		goto bind_err;

	/*
	 * Once the physical CPUs have been bound, check if a MAU
	 * has been requested and then allocate physical MAU
	 * resources from the MAU requirements.
	 */
	mau_adm_t mau_adm = {
		.type = MAU_ADM_NUM,
		.num = ldp->required.num_vmaus
		/* CSTYLED */
	};

	if (cons_bind_vmaus(ldp, &mau_adm, resp) == false)
		goto bind_err;

	/*
	 * Allocate memory from the requirements.
	 * This is much harder, since we may shrink a free block
	 * rather than simply re-assigning it.
	 */
	if (!mem_bind_virt(ldp, ldp->required.size_mem, resp)) {
		rprintf(resp, gettext(
		    "Could not bind requested memory for LDom %s\n"),
		    ldp->namep);
		goto bind_err;
	} else {
		/*
		 * Set required memory to match the bound memory. A different
		 * amount may have been bound because of alignment or size
		 * constraints.
		 */
		ldp->required.size_mem = ldom_size_bound_mem(ldp);
	}

	if (bind_boot_options(ldp, resp) == false)
		goto bind_err;

	if (bind_io(ldp, resp) == false)
		goto bind_err;

	if (bind_pciedevs(ldp, resp) == false)
		goto bind_err;

	if (vio_bind(ldp, resp) != 0)
		goto bind_err;

	if (bind_dependency(ldp, resp) == false)
		goto bind_err;

	return (true);

bind_err:
	undo_partial_binding(ldp);

	return (false);
}

/*
 * Simply free the allocated data associated with an ldom_t.
 */

void
free_ldom(ldom_t *ldp)
{
	req_iodevice_t *ior, *ior_next;

	destroy_res_pool(&ldp->guest_vcpu_ids);
	destroy_res_pool(&ldp->guest_vmau_ids);
	destroy_res_pool(&ldp->guest_free_ldcs);
	destroy_res_pool(&ldp->free_hvcfg_handles);
	destroy_res_pool(&ldp->free_ds_ports);
#if VDEV_CFG_HANDLE
	destroy_res_pool(&ldp->free_cfg_handles);
#endif

	CHAIN_ITER_VOL(ldp->required.io_devices, ior, ior_next, link) {
		Xfree(ior->name);
		Xfree(ior);
	}

	/* bindings and VIO structs have already been freed */

	Xfree(ldp->namep);

	if (ldp->devalias_md != NULL)
		md_fini(ldp->devalias_md);
	if (ldp->ds_sp_md != NULL)
		md_fini(ldp->ds_sp_md);

	free_keys(&ldp->keys);
	free_variables(&ldp->variables);
	free_masters(ldp);
	free_policies(ldp);

	if (ldp->props.mac_address != ILLEGAL_MAC_ADDR)
		add_used_mac(ldp->props.mac_address);

	mutex_lock(&zeus.mlock);
	CHAIN_UNLINK(zeus.ldoms, ldp, link);
	mutex_unlock(&zeus.mlock);

	Zfree(ldp->mig_info.peer_name);
	if (ldp->mig_state.bind_mdp != NULL)
		md_fini(ldp->mig_state.bind_mdp);

	if (ldp->mig_state.bind_fp != NULL) {
		frag_free(ldp->mig_state.bind_fp);
		ldp->mig_state.bind_fp = NULL;
	}

	Xfree(ldp);
}

void
unbind_io(ldom_t *ldp, bound_iodevice_t *bio_p)
{
	pri_iodevice_t *pio_p;
	req_iodevice_t *rio_p;

	pio_p = bio_p->phys_iodev;
	rio_p = bio_p->req_iop;

	if (pio_p->is_pciex) {
		if (pcie_is_rc_owner(ldp, bio_p)) {
			pcie_free_res(bio_p);
			pcie_unassign_all_devs(ldp, bio_p);
			/*
			 * FIXME: Subdivided I/O needs something more
			 * robust than a single free flag.
			 */
			pio_p->free = true;
		} else {
			unbind_pciedevs(ldp, bio_p);
		}
	} else {
		pio_p->free = true;
	}

	CHAIN_UNLINK(ldp->bound.iodevices, bio_p, ldom);
	CHAIN_UNLINK(pio_p->bound_ios, bio_p, pio);

	if (rio_p != NULL) {
		rio_p->bound = false;
	}
	destroy_bound_iodevice(bio_p);
}


/*
 * Clean up the memory associated with a bound iodevice (root complex).
 */
void
destroy_bound_iodevice(bound_iodevice_t *bio)
{
	ASSERT(bio);
	if (bio->name != NULL)
		Xfree(bio->name);

	if (bio->id != NULL)
		free_res(bio->id);

	pcie_destroy_bound_res(bio);
	Xfree(bio);
}

bool_t
unbind_an_ldom(ldom_t *ldp)
{
	bound_iodevice_t *bio_p;
	guest_sp_ldc_t *ldc, *ldc_next;

	ASSERT(ldp->idp != NULL);

	free_res(ldp->idp);
	SANITY(ldp->idp = NULL);

	free_res(ldp->res_id);

	mau_unbind_all(ldp);

	cpu_unbind_all(ldp);

	mem_unbind_all(ldp);

	/* Free all guest MD frags */
	if (ldp->frags.committed_md_frag != NULL) {
		frag_free(ldp->frags.committed_md_frag);
		ldp->frags.committed_md_frag = NULL;
	}

	if (ldp->frags.parsed_md_frag != NULL) {
		frag_free(ldp->frags.parsed_md_frag);
		ldp->frags.parsed_md_frag = NULL;
	}
	if (ldp->frags.next_md_frag != NULL) {
		frag_free(ldp->frags.next_md_frag);
		ldp->frags.next_md_frag = NULL;
	}

	/*
	 * Walk bound IO list for this LDom, unlinking from both
	 * ldom's bound & pri_iodevice's pio chains
	 */
	for (bio_p = CHAIN_START(ldp->bound.iodevices);
	    bio_p != NULL; /* nothing */) {
		bound_iodevice_t *next_bio_p = CHAIN_NEXT(bio_p->ldom);

		/*
		 * Unbind IO devices and associated PCIe devices.
		 */
		unbind_io(ldp, bio_p);

		bio_p = next_bio_p;
	}

	/*
	 * Free all SP LDC channels.
	 */
	CHAIN_ITER_VOL(ldp->sp_ldcs, ldc, ldc_next, link) {
		free_res(ldc->channel->chan);
		free_res(ldc->channel->id);
		Xfree(ldc->channel);
		Xfree(ldc);
	}

	return (true);
}
