/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <string.h>

#include "allocate.h"
#include "database.h"
#include "dispctl.h"
#include "errors.h"
#include "fatal.h"
#include "ldom.h"
#include "ldcs.h"
#include "vio.h"

static bool_t is_port_avail(vcc_t *vcc, response_t *resp, int tcp_port);
#if LIVE_GRP_REBIND
int console_update_binding(ldom_t *ldp, response_t *resp, char *new_service,
	char *new_group, int new_port, int mod, int bindmod);
static int can_bind_new_console(ldom_t *ldp, response_t *resp,
    char *new_service, char *new_group, int new_port);
#endif

/*
 * Returns true if one of more of the console parameters has been set.
 */
bool_t
is_console_constraints_set(ldom_t *ldp)
{
	if ((ldp->console.service_name != NULL) ||
	    (ldp->console.group_name != NULL) ||
	    (!IS_VAL_AUTO(ldp->console.tcp_port)))
		return (true);
	else
		return (false);
}

bool_t
is_console_set(ldom_t *ldp)
{
	if (is_console_constraints_set(ldp) || (ldp->state >= BOUND))
		return (true);
	else
		return (false);
}

console_t *
console_new(ldom_t *ldp, char *service, char *group, int tcp_port)
{
	ldp->console.service_name = (service != NULL) ? Xstrdup(service) : NULL;
	ldp->console.group_name = (group != NULL) ? Xstrdup(group) : NULL;
	ldp->console.tcp_port = tcp_port;

	return (&ldp->console);
}

static void
destroy_console_binding(console_t *console)
{
	if (console->binding != NULL) {
		/* destroy server side if it exists */
		if (console->binding->vcc_cons != NULL)
			vcc_destory_console(console->binding->vcc_cons);

		free_res(console->binding->id);
		ldcs_free(&console->binding->ldcs);
		Xfree(console->binding);
		console->binding = NULL;
	}
}

/*
 * Called when a domain is deleted.
 */
void
console_delete(ldom_t *ldp)
{
	ASSERT(ldp->console.binding == NULL);

	Xfree(ldp->console.service_name);
	Xfree(ldp->console.group_name);
}

/*
 * Check a console client can be bound to a particular vcc server. In case the
 * server isn't specified, we need to iterate over all bound vcc servers
 * available and see if any vcc can meet the console constraint
 */
static bool_t
meet_vcc_const(ldom_t *ldp, response_t *resp, vcc_t *vcc,
    char *group_name, int new_port)
{
	console_group_t *ng_p, *npg_p;
	vcc_binding_t *binding = vcc->binding;

	if ((vcc == NULL) || (vcc->ldp->state < BOUND)) {
		return (false);
	}

	/*
	 * When binding a new console here is the algorithm:
	 * (1) Neither group nor port specified:
	 * This is easy, just assign to any existing console group. If none
	 * exists, create one with LDom's name
	 * (2) Only group specified: we can always meet this provided:
	 *	group already exists -OR-
	 *	Group doesn't exist, but at least one port is available in range
	 * (3) Only port specified: we can meet this if port within vcc's range
	 * If no group exists using this port, create a new group with the
	 * LDom name.
	 * (4) Both group and port specified: Check if a console group
	 * exists with the same group or port. If exists, they must match.
	 */
	if (group_name == NULL && IS_VAL_AUTO(new_port)) {
		return (true);
	} else if (group_name != NULL && IS_VAL_AUTO(new_port)) {
		/*
		 * Either the group should exist, or at least one free tcp
		 * port should be available to create a new group
		 */
		ng_p = console_group_find_by_name(vcc, group_name);
		if (ng_p != NULL || (available_res_ids(&binding->tcp_pool) >
		    0)) {
			return (true);
		} else {
			rprintf(resp, gettext("No free tcp port available in "
			    "vcc's port range\n"));
			return (false);
		}
	} else if (group_name == NULL && (!IS_VAL_AUTO(new_port))) {
		if (PORT_IN_RANGE(new_port, vcc)) {
			return (true);
		} else {
			rprintf(resp, gettext("The tcp port %d is "
			    "outside vcc's port range\n"), new_port);
			return (false);
		}
	} else {
		/*
		 * Both group and port are specified for the console
		 */
		ng_p = console_group_find_by_name(vcc, group_name);
		npg_p = console_group_find_by_port(vcc, new_port);
		if (ng_p != NULL) {
			if (ng_p->tcp_port->id != new_port) {
				rprintf(resp, gettext("The tcp port %d is "
				    "already in use by another console group "
				    "%s\n"), new_port, ng_p->group_name);
				return (false);
			} else {
				return (true);
			}
		} else if (npg_p != NULL) {
			/*
			 * Since the previous condition wasn't met, it must be
			 * that the port is allocated, but the group isn't, so
			 * this must be a console group with a different
			 * group name
			 */
			rprintf(resp, gettext("The tcp port %d is "
			    "already in use by another console group "
			    "%s\n"), new_port, npg_p->group_name);
			return (false);
		} else {
			if (is_port_avail(vcc, resp, new_port) == false)
				return (false);
			else
				return (true);
		}
	}
}

/*
 * If the name of the console vcc service is not specified in a console
 * specification, we need to find out the vcc service which can best meet the
 * console constraint by going through all available vcc servers available
 */
static vcc_t *
find_vcc_to_bind(ldom_t *ldp, response_t *resp, char *service_name,
    char *group_name, int tcp_port)
{
	extern CHAIN(vcc_t, vcc_services);
	vcc_t *vcc;
	if (service_name == NULL) {
		CHAIN_ITER(vcc_services, vcc, link) {
			if ((vcc->ldp->state >= BOUND) &&
			    (meet_vcc_const(ldp, resp, vcc,
			    group_name, tcp_port) == true))
				return (vcc);
		}
		/*
		 * fallthrough
		 */
		rprintf(resp, gettext("Didn't find a suitable vcc service "
		    "in a bound service domain to bind guest %s console\n"),
		    ldp->namep);

	} else {
		vcc = vcc_find_by_name(service_name);
		if (vcc == NULL || vcc->ldp->state < BOUND) {
			rprintf(resp, gettext("The virtual console "
			    "concentrator service %s not exist or is not "
			    "bound.\n"), service_name);
		} else if (meet_vcc_const(ldp, resp, vcc, group_name,
		    tcp_port) == false) {
			rprintf(resp, gettext("The virtual console concentrator"
			    "service %s does not satisfy the constraints of "
			    "guest \"%s\"'s virtual console\n"), service_name,
			    ldp->namep);
		} else {
			return (vcc);
		}
	}

	return (NULL);
}

static int
init_console_binding(console_t *console, uint64_t cons_id, uint64_t target_type,
    uint64_t uartbase)
{
	console_binding_t *binding;
	resid_t *id;

	if (cons_id == NEW_INSTANCE) {
		if ((id = alloc_res(&zeus.free_console_ids)) == NULL)
			/*
			 * Console ids are effectively unlimited hence this
			 * can only fail if something is badly broken.
			 */
			fatal("No free console ids");
	} else {
		if ((id = pull_res_id(&zeus.free_console_ids, cons_id)) == NULL)
			return (ERR_CONSOLE_DUP_ID);
	}

	binding = Xmalloc(sizeof (console_binding_t));

	CHAIN_INIT(binding->ldcs.ldcs);
	binding->id = id;
	binding->sp_cons = (target_type == TYPE_SP) ? true : false;
	binding->uartbase = uartbase;
	binding->vcc_cons = NULL;

	console->binding = binding;
	return (0);
}

static console_group_t *
console_find_group(ldom_t *ldp, vcc_t *vcc, char *group, int port)
{
	console_group_t *group_p = NULL;
	char *group_s = (group == NULL) ? ldp->namep : group;

	if (!IS_VAL_AUTO(port) && !PORT_IN_RANGE(port, vcc))
		return (NULL);

	/*
	 * If group is not specified, but port is, check if a group
	 * already exists with the port
	 */
	if (group == NULL && (!IS_VAL_AUTO(port)))
		group_p = console_group_find_by_port(vcc, port);

	if (group_p == NULL)
		group_p = vcc_get_console_group(vcc, group_s, port);
	else
		group_p->refcnt++;

	return (group_p);
}

/*
 * Bind a domain's console.
 */
int
console_bind(ldom_t *ldp, response_t *resp)
{
	vcc_t *vcc;
	vcc_console_t *vcons;
	console_group_t *group;
	char *group_name = ldp->console.group_name;
	int rc;

	ASSERT(ldp->console.binding == NULL);

	if (!is_console_set(ldp) && (!zeus.ldc_console_config ||
	    IS_CONTROL_DOMAIN(ldp)))
		/* not using LDC consoles, nothing to do */
		return (0);

	vcc = find_vcc_to_bind(ldp, resp, ldp->console.service_name,
	    group_name, ldp->console.tcp_port);

	if ((vcc == NULL) || (vcc->binding == NULL)) {
		return (ERR_NO_SERVICE);
	}

	if (vcc->ldp->flag.locked) {
		rprintf(resp, gettext("Cannot bind console because that would "
		    "modify locked LDom %s\n"), vcc->ldp->namep);
		return (ERR_LDOM_LOCKED);
	}


	group = console_find_group(ldp, vcc, group_name, ldp->console.tcp_port);

	if (group == NULL) {
		rprintf(resp, gettext("Could not get console group %s\n"),
		    group_name == NULL ? "unspecified" : group_name);
		return (ERR_CONSOLE_NO_GROUP);
	}

	(void) init_console_binding(&ldp->console, NEW_INSTANCE, TYPE_GUEST, 0);

	/* create and link to the server side */
	vcons = vcc_new_console(vcc, group, NEW_CONSOLE_PORT);
	vcons->console = &ldp->console;
	ldp->console.binding->vcc_cons = vcons;

	/* allocate the console and server LDC channels */
	rc = ldcs_chan_alloc(&ldp->console.binding->ldcs,
	    &ldp->guest_free_ldcs);
	if (rc == 0)
		rc = ldcs_chan_alloc(&vcons->ldcs, &vcc->ldp->guest_free_ldcs);
	if (rc != 0) {
		destroy_console_binding(&ldp->console);
		rprintf(resp, gettext("Insufficient free LDC channels to bind "
		    "console on LDom %s\n"), ldp->namep);
		return (rc);
	}

	if (vcc->ldp != ldp) {
		vcc->ldp->bound.num_clients++;
		vcc->ldp->flag.gen_new_md = true;
		vcc->ldp->flag.md_notify = true;
	}

	return (0);
}

void
console_unbind(ldom_t *ldp)
{
	console_binding_t *binding = ldp->console.binding;
	vcc_console_t *vcons;

	if (binding == NULL)
		return;

	ASSERT(!binding->sp_cons && binding->uartbase == 0);

	vcons = binding->vcc_cons;

	if (vcons->vcc->ldp != ldp) {
		vcons->vcc->ldp->bound.num_clients--;
		vcons->vcc->ldp->flag.gen_new_md = true;
		vcons->vcc->ldp->flag.md_notify = true;
	}

	vcc_unbind_console(vcons);	/* unbind on server side */

	ldcs_stage(&binding->ldcs);
	free_res(binding->id);
	Xfree(binding);
	ldp->console.binding = NULL;
}

static int
get_console_ldc_endpoint(md_t *hv_mdp, md_node_t *hv_guestp,
    md_node_t **console, md_node_t **ldc_endpoint, uint64_t *uart_base)
{
	md_node_t **nodes;
	int num;

	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(hv_mdp));
	num = md_scan_dag(hv_mdp, hv_guestp, md_find_name(hv_mdp, "console"),
	    md_find_name(hv_mdp, "fwd"), nodes);
	if (num > 1)
		fatal("Guest has multiple consoles in HV MD");

	if (num == 1) {
		*console = nodes[0];
		*uart_base = 0;

		/* no LDC endpoint if UART console */
		if (md_get_prop_val(hv_mdp, *console, "uartbase",
		    uart_base) == 0) {
			Xfree(nodes);
			return (-1);
		}

		num = md_scan_dag(hv_mdp, *console,
		    md_find_name(hv_mdp, "ldc_endpoint"),
		    md_find_name(hv_mdp, "fwd"), nodes);
		if (num != 1)
			fatal("Invalid console node in HV MD");
		*ldc_endpoint = nodes[0];

		Xfree(nodes);
		return (0);
	} else {
		Xfree(nodes);
		return (-1);
	}
}

/*
 * The console is not a regular vdevice node in guest MD. It's identified by
 * virtue of console node in the HV MD.
 */
void
console_add_from_md(ldom_t *ldp, md_t *hv_mdp, md_node_t *hv_guestp)
{
	md_node_t *console, *ldc_endp;
	uint64_t cons_id, ldc_id, ldc_resid;
	uint64_t target_type;
	uint64_t uart_base = 0;

	if (get_console_ldc_endpoint(hv_mdp, hv_guestp, &console,
	    &ldc_endp, &uart_base) == -1) {
		if (uart_base == 0) {
			warning("Unable to create console client for guest %s",
			    ldp->namep);
			return;
		}
	} else {
		if (md_get_prop_val(hv_mdp, ldc_endp, "channel",
		    &ldc_id) != 0)
			fatal("Invalid console LDC endpoint node in HV MD");

		if (md_get_prop_val(hv_mdp, ldc_endp, "resource_id",
		    &ldc_resid) != 0)
			fatal("Invalid console LDC endpoint node in HV MD");

		if (md_get_prop_val(hv_mdp, ldc_endp, "target_type",
		    &target_type) != 0)
			fatal("Invalid console LDC endpoint node in HV MD");
	}

	if (md_get_prop_val(hv_mdp, console, "resource_id", &cons_id) != 0)
		fatal("Invalid console node in HV MD");

	/*
	 * The migration recovery code will handle resurrecting the console
	 * binding, so there is nothing more to do at this point.
	 */
	if (ldp->flag.migration_source)
		return;

	/*
	 * Partially initialize the console. At this point, the group name is
	 * unknown and we don't know which vcc server to bind to. This will be
	 * determined when finalizing the vio bindings after parsing all the
	 * MDs.
	 */
	init_console_binding(&ldp->console, cons_id, target_type, uart_base);

	if (uart_base == 0) {
		ldc_pull_ids(ldp, &ldp->guest_free_ldcs, 1, &ldc_id,
		    &ldp->console.binding->ldcs);
	}
}

static md_node_t *
add_console_node(md_t *hv_md, md_node_t *hv_guest, ldom_t *ldp)
{
	md_node_t **nodes;
	md_node_t *console_node;
	int num;

	nodes = Xmalloc(md_node_count(hv_md) * sizeof (md_node_t *));
	num = md_scan_dag(hv_md, NULL, md_find_name(hv_md, "consoles"),
	    md_find_name(hv_md, "fwd"), nodes);
	ASSERT(num == 1);

	console_node = md_link_new_node(hv_md, "console", nodes[0],
	    "fwd", "back");

	md_add_arc_property(hv_md, hv_guest, "fwd", console_node);
	md_add_arc_property(hv_md, console_node, "back", hv_guest);
	md_add_value_property(hv_md, console_node, "resource_id",
	    ldp->console.binding->id->id);
	md_add_value_property(hv_md, console_node, "ino", LDC_CONSOLE_INO);

	Xfree(nodes);

	return (console_node);
}

void
console_gen_ldc_endpoints(ldom_t *ldp, md_t *hv_md, md_node_t *hv_guest,
    md_node_t *ldc_endpoints)
{
	console_binding_t *binding = ldp->console.binding;
	ldc_chan_t *cons_chan = NULL;
	md_node_t *cons_endpoint, *console_node;

	if (binding == NULL)
		return;	/* this can be the case in legion backend for example */

	if (binding->uartbase == 0) {
		ASSERT(CHAIN_LENGTH(binding->ldcs.ldcs) == 1);
		cons_chan = CHAIN_START(binding->ldcs.ldcs);
	}

	/*
	 * Add the console LDC endpoint.  Check whether the LDC channel
	 * needs to connect to a vcc instance in a guest or to the SP.
	 */
	if (binding->sp_cons) {
		sp_guest_ldc_t *sp_chan;
		/*
		 * Console LDC connects to SP.  Look through previously saved
		 * SP channel list to find target channel.
		 */
		sp_chan = sp_guest_ldc_find_by_target(ldp->res_id->id,
		    cons_chan->chan->id);
		if (sp_chan == NULL) {
			warning("No SP LDC found for guest LDC ID %"PRIu64"",
			    cons_chan->chan->id);
			return;
		}

		cons_endpoint = vio_gen_ldc_endpoint_node(hv_md, hv_guest,
		    ldc_endpoints, TYPE_GUEST, TYPE_SP, cons_chan, -1,
		    sp_chan->channel);
	} else if (binding->uartbase == 0) {
		/*
		 * Console connects to a vcc port.
		 */
		cons_endpoint = vio_gen_ldc_endpoint_node(hv_md, hv_guest,
		    ldc_endpoints, TYPE_GUEST, TYPE_GUEST, cons_chan,
		    binding->vcc_cons->vcc->ldp->idp->id,
		    CHAIN_START(binding->vcc_cons->ldcs.ldcs));
	}

	/* add console node for the guest to HV MD */
	console_node = add_console_node(hv_md, hv_guest, ldp);

	if (binding->uartbase == 0) {
		/* Set console specific properties in channel endpoint node */
		md_add_value_property(hv_md, cons_endpoint, "private_svc",
		    LDC_CONSOLE_SVC);
		md_update_value_property(hv_md, cons_endpoint, "tx-ino",
		    LDC_CONSOLE_INO);
		md_update_value_property(hv_md, cons_endpoint, "rx-ino",
		    LDC_CONSOLE_INO);


		/* link channel endpoint with console node */
		md_add_arc_property(hv_md, console_node, "fwd", cons_endpoint);
		md_add_arc_property(hv_md, cons_endpoint, "back", console_node);
	} else {
		md_add_value_property(hv_md, console_node, "uartbase",
		    binding->uartbase);

		/*
		 * Add the uartbase to the guest node as well
		 */
		md_add_value_property(hv_md, hv_guest, "uartbase",
		    binding->uartbase);
	}
}

char *
console_print(ldom_t *ldp)
{
	static char consstr[INT_STR_LEN];

	if (ldp->console.binding == NULL)
		return ("");

	if (ldp->console.binding->sp_cons)
		return ("SP");
	else if (ldp->console.binding->uartbase != 0)
		return ("UART");

	sprintf(consstr, "%u",
	    (int)ldp->console.binding->vcc_cons->group->tcp_port->id);

	return (consstr);
}

void
console_list(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	console_binding_t *binding = ldp->console.binding;
	char *service;

	if (dctl->oformat && !dctl->oconsole)
		return;
	if (dctl->constraints && (binding != NULL) && (binding->sp_cons ||
	    binding->uartbase != 0))
		return;
	if ((ldp->state == INACTIVE) && !is_console_set(ldp))
		return;

	/*
	 * TRANSLATION_NOTE
	 * The following are column headers in tables with fixed column widths.
	 * WIDTH  HEADER	MEANING
	 * 16	  "NAME"	name of client
	 * 27	  "SERVICE"	name of the service the client is using
	 * 6	  "PORT"	console port
	 * 3	  "LDC"		Logical Domain Channel number
	 */

	if (dctl->parseable) {
		rprintf(resp, "%s", dctl_label("vcons", dctl));
	} else {
		rprintf(resp, "%s\n", dctl_label("vcons", dctl));
		if (dctl->bindings && dctl->extended)
			rprintf(resp, "    "FNAME" "FBNDNAME" "FPORT" %s\n",
			    gettext("NAME"), gettext("SERVICE"),
			    gettext("PORT"), gettext("LDC"));
		else
			rprintf(resp, "    "FNAME" "FBNDNAME" %s\n",
			    gettext("NAME"), gettext("SERVICE"),
			    gettext("PORT"));
	}

	if ((binding != NULL) && binding->sp_cons) {
		if (dctl->parseable)
			rprintf(resp, "|type=SP\n");
		else
			rprintf(resp, "    SP\n\n");
		return;
	} else if ((binding != NULL) && binding->uartbase != 0) {
		if (dctl->parseable)
			rprintf(resp, "|type=UART\n");
		else
			rprintf(resp, "    UART\n\n");
		return;
	}

	if ((binding != NULL) && !dctl->constraints) {
		service = zsprintf("%s@%s", binding->vcc_cons->vcc->service,
		    binding->vcc_cons->vcc->ldp->namep);

		if (dctl->parseable) {
			rprintf(resp, "|group=%s|service=%s|port=%"PRIu64"",
			    binding->vcc_cons->group->group_name, service,
			    binding->vcc_cons->group->tcp_port->id);
		} else {
			rprintf(resp, "    "FNAME" "FBNDNAME" "FPORTd,
			    binding->vcc_cons->group->group_name, service,
			    binding->vcc_cons->group->tcp_port->id);
		}
		Xfree(service);

		if (dctl->bindings && dctl->extended) {
			if (dctl->parseable)
				rprintf(resp, "|ldc=");
			else
				rprintf(resp, " ");
			ldcs_print(&binding->ldcs, resp);
		}
	} else if (is_console_set(ldp)) {
		char tcp_port[INT_STR_LEN];
		char *gname;

		if (ldp->console.service_name != NULL)
			service = ldp->console.service_name;
		else
			service = "";

		if (ldp->console.group_name != NULL)
			gname = ldp->console.group_name;
		else
			gname = "";

		if (ldp->console.tcp_port > 0)
			sprintf(tcp_port, "%"PRId32"", ldp->console.tcp_port);
		else
			(void) strcpy(tcp_port, "");

		if (dctl->parseable) {
			rprintf(resp, "|group=%s|service=%s|port=%s",
			    gname, service, tcp_port);
		} else {
			rprintf(resp, "    "FNAME" "FBNDNAME" "FPORT,
			    gname, service, tcp_port);
		}
	} else if (zeus.simulating) {
		rprintf(resp, "%sUART",
		    dctl->parseable ? "|type=" : "    ");
	}

	rprintf(resp, "\n");
	if (!dctl->parseable)
		rprintf(resp, "\n");
}

int
link_console(ldom_t *ldp)
{
	ldc_epdb_t *ep;
	ldc_chan_t *ldc;
	vcc_console_t *vcons;
	char *db_service, *db_group;
	int db_port;
	bool_t ret;
	bool_t console_set = false;

	if ((ldp->console.binding == NULL) || ldp->console.binding->sp_cons ||
	    ldp->console.binding->uartbase != 0)
		return (0);

	if ((ldc = CHAIN_START(ldp->console.binding->ldcs.ldcs)) == NULL)
		return (0);	/* not a LDC console */

	if ((ep = ldc_epdb_find(ldp->idp->id, ldc->chan->id)) != NULL)
		vcons = vcc_console_find_by_ldcid(ep->peer->ldp, ep->peer->ldc);
	if ((ep == NULL) || (vcons == NULL))
		fatal("Could not find matching vcc port for console on LDom %s",
		    ldp->namep);

	ldp->console.binding->vcc_cons = vcons;
	vcons->console = &ldp->console;

	if (vcons->vcc->ldp != ldp)
		vcons->vcc->ldp->bound.num_clients++;

	/*
	 * The constraints need to be populated from the old database entries
	 * if any. returns true if the guest node was present in old db. The
	 * flag console_set is set to true if there were properties present in
	 * the console node, i.e. a console was automatically assigned.
	 */
	ret = database_get_console(ldp, &db_port, &db_group,
	    &db_service, &console_set);
	if (ret == true && console_set == true) {
		/*
		 * A console constraint was previously present
		 */
		ldp->console.service_name = db_service;
		ldp->console.group_name = db_group;
		ldp->console.tcp_port = db_port;
	} else {
		/*
		 * Either the old DB wasn't present, or the console for
		 * the guest was auto assigned, hence no values in old DB
		 */
		ldp->console.service_name = Xstrdup(vcons->vcc->service);
		ldp->console.group_name = Xstrdup(vcons->group->group_name);
		ldp->console.tcp_port = vcons->group->tcp_port->id;
	}

	/*
	 * Set the constraint in the DB in case there was a console
	 * node in the DB and values were set underneath, or if the
	 * constraint DB isn't found.
	 */
	if (ret == false || console_set == true)
		database_ldom_add_console_client(ldp, ldp->console.service_name,
		    ldp->console.group_name, ldp->console.tcp_port);

	return (0);
}

int
console_bind_check(ldom_t *ldp, response_t *resp)
{
	vcc_t *vcc;

	if (ldp->console.service_name == NULL) {
		if (!is_console_set(ldp) && !zeus.ldc_console_config)
			/* not using ldc consoles, nothing to do */
			return (0);
		/*
		 * No server name specified, bind to the first VCC server on
		 * control domain.
		 */
		vcc = CHAIN_START(zeus.hvctl_ldom_p->vccs);
	} else {
		vcc = vcc_find_by_name(ldp->console.service_name);
	}

	if ((vcc == NULL) || ((vcc->ldp != ldp) && (vcc->binding == NULL))) {
		rprintf(resp, gettext("The virtual console concentrator "
		    "service for the virtual console,\ndoes not exist "
		    "or is not bound.\n"));
		return (ERR_NO_SERVICE);
	}

	ldp->work.num_ldcs++;
	vcc->ldp->work.num_ldcs++;

	return (0);
}

static void
update_cons(ldom_t *ldp, int mod, char *new_service, char *new_group,
    int new_port)
{
	if (mod&CONSOLE_SERVICE) {
		Xfree(ldp->console.service_name);
		ldp->console.service_name = Xstrdup(new_service);
	}

	if (mod&CONSOLE_GROUP) {
		Xfree(ldp->console.group_name);
		ldp->console.group_name = (new_group == NULL ||
		    IS_EMPTY_STR(new_group)) ?  NULL : Xstrdup(new_group);
	}

	if (mod&CONSOLE_PORT) {
		ldp->console.tcp_port = new_port;
	}
}

/*
 * service= is not accepted, no default. Once a vcc service have been
 * specified, it cannnot be unset.
 * group= implies choose any group available
 * port= implies any port available
 * service can be either NULL or a valid string
 * group can be NULL (no change) or "" (set default) or a valid string
 * port can be -1 (no change), 0 (set default) or a valid tcp port
 */
int
console_modify(ldom_t *ldp, char *new_service, char *new_group, int new_port,
	response_t *resp)
{
	console_t *cons;
	console_binding_t *binding = ldp->console.binding;
	vcc_t *vcc;
	int mod = 0;
	int bindmod = 0;

	char *service = ldp->console.service_name;
	char *group = ldp->console.group_name;
	int port = ldp->console.tcp_port;
	char *bound_service = NULL;
	char *bound_group = NULL;
	int bound_port = _AUTO_VAL_;

	char *svc = NULL;
	char *grp = NULL;
	int nport = _AUTO_VAL_;

	/*
	 * Check if console is not already available for this LDom
	 */
	if (is_console_set(ldp) == false) {
		port = (new_port == NEW_INSTANCE) ?
		    _AUTO_VAL_ : new_port;

		/*
		 * a guest can't be bound without a console
		 */
		if (ldp->state >= BOUND) {
			rprintf(resp, gettext("Console not found in bound "
			    "guest %s\n"), ldp->namep);
			return (ERR_NO_CLIENT);
		}

		cons = console_new(ldp, new_service, new_group, port);

		if (cons == NULL) {
			rprintf(resp, gettext("Failed to add console\n"));
			return (ERR_NO_CLIENT);
		}

		database_ldom_add_console_client(ldp, new_service, new_group,
		    port);

		return (0);
	}

	/*
	 * Process the changes in the console constraints
	 */
	if (new_service) {
		if (service == NULL)
			mod |= CONSOLE_SERVICE;
		else if (strcmp(service, new_service) != 0)
			mod |= CONSOLE_SERVICE;
	}

	if (new_group) {
		if (IS_EMPTY_STR(new_group)) {
			if (group != NULL)
				mod |= CONSOLE_GROUP;
		} else {
			if (group == NULL)
				mod |= CONSOLE_GROUP;
			else if (strcmp(group, new_group) != 0)
				mod |= CONSOLE_GROUP;
		}
	}

	if (new_port != NEW_INSTANCE) {
		if (new_port != port)
			mod |= CONSOLE_PORT;
	}

	/*
	 * check if the existing binding is affected
	 */
	if (ldp->state >= BOUND) {
		vcc = binding->vcc_cons->vcc;

		bound_service = vcc->service;
		bound_group = binding->vcc_cons->group->group_name;
		bound_port = binding->vcc_cons->group->tcp_port->id;

		if (new_service && !IS_EMPTY_STR(new_service)) {
			if (strcmp(new_service, bound_service))
				bindmod |= CONSOLE_SERVICE;
		}

		if (new_group && !IS_EMPTY_STR(new_group)) {
			if (strcmp(new_group, bound_group))
				bindmod |= CONSOLE_GROUP;
		}

		if (new_port != NEW_INSTANCE && !IS_VAL_AUTO(new_port)) {
			if (bound_port != new_port)
				bindmod |= CONSOLE_PORT;
		}
	}

	if (mod == 0) {
		rprintf(resp, gettext("No changes made to console for guest "
		    "%s\n"), ldp->namep);
		return (0);
	} else {
		svc = (mod & CONSOLE_SERVICE) ? new_service :
		    ldp->console.service_name;
		grp = (mod & CONSOLE_GROUP) ? (IS_EMPTY_STR(new_group) ?
		    NULL : new_group) : ldp->console.group_name;
		nport = (mod & CONSOLE_PORT) ? new_port : ldp->console.tcp_port;

		if (ldp->state < BOUND || bindmod == 0) {
			goto ret;
		}
	}

	if (ldp->state >= BOUND) {
#if !LIVE_GRP_REBIND
		rprintf(resp, gettext("Rebinding console group is supported "
		    "only when the domain is in unbound state\n"));
		return (ERR_LDOM_IS_BOUND);
#else
		/*
		 * TODO:
		 * This feature is currently disabled because the vcc/vntsd
		 * in the service domain does not yet support live
		 * re-assignment of group and/or port. If this feature ever
		 * gets enabled in the service domain, this part of code could
		 * be enabled after doing the necessary verification.
		 *
		 * Requested changes affect bindings. Before breaking up the
		 * existing binding, ensure that the console can be
		 * successfully rebound, otherwise we would be left with an
		 * LDom with non-usable console
		 */
		if (bindmod & CONSOLE_SERVICE) {
			/*
			 * Binding to a new vcc service. Since binding to a
			 * new service requires tearing down and re-creating
			 * the ldc connection, we want to make sure this is
			 * possible
			 */
			int rc = 0;
			vcc_t *new_vcc = find_vcc_to_bind(ldp, resp, svc, grp,
			    nport);

			if (new_vcc == NULL) {
				rprintf(resp, gettext("Unable to bind to vcc "
				    "service %s\n"), svc);
				return (false);
			} else if (new_vcc->ldp->state < BOUND) {
				rprintf(resp, gettext("The service named %s "
				    "belongs to the guest %s which is "
				    "unbound\n"), svc, new_vcc->ldp->namep);
				return (false);
			}

			if (can_bind_new_console(ldp, resp, new_vcc->service,
			    grp, nport) == false)
				return (false);
		}

		/*
		 * For re-bind within the same vcc, we save the bound
		 * group and port info, release current group and
		 * attempt to re-bind the new group/port. If re-bind
		 * fails, we just bind back to the saved group and port
		 */

		if ((mod & CONSOLE_PORT) && !is_port_avail(vcc, resp, nport))
			return (false);

		if ((rc = console_update_binding(ldp, resp, svc,
		    grp, nport, mod, bindmod)) != 0) {

			rprintf(resp, gettext("Failed to modify console for "
			    "guest %s\n"), ldp->namep);
			/*
			 * Just re-attach the old group and port
			 */
			binding->vcc_cons->group = console_find_group(ldp, vcc,
			    bound_group, bound_port);

			ASSERT(binding->vcc_cons->group != NULL);

			return (rc);
		}

		if (zeus.mode == ZEUS_MGMT_MODE) {
			seq_config_now_or_later(ldp, DEL_RECONFIG_OTHER, resp);
			/* No delayed reconfig in progress */
			if (bindmod & CONSOLE_SERVICE) {
				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_OTHER, resp);
			} else {
				seq_download_mds_do_config(vcc->ldp);
			}
		}
#endif
	}

ret:
	update_cons(ldp, mod, svc, grp, nport);
	database_ldom_modify_console_client(ldp, mod, svc, grp, nport);

	return (0);
}

int
console_mig_add_bindings(ldom_t *ldp, md_t *mdp)
{
	int			i;
	md_node_t		**vdevs;
	int			nvdevs;
	md_node_t		*cons;
	char			*name;
	struct ldc_chan		*chan;
	uint64_t		ldcid;
	console_binding_t	*binding;

	/*
	 * Lookup the LDC for the console.
	 */
	binding = ldp->console.binding;
	if (binding == NULL) {
		DBG(zprintf("Console not bound for LDom %s\n", ldp->namep));
		return (-1);
	}

	if (CHAIN_LENGTH(binding->ldcs.ldcs) != 1) {
		DBG(zprintf("Console does not have exactly one LDC\n"));
		return (-1);
	}

	chan = CHAIN_START(binding->ldcs.ldcs);
	ldcid = chan->chan->id;

	/*
	 * Scan through the MD for all virtual device nodes
	 */
	vdevs = Xmalloc(md_node_count(mdp) * sizeof (md_node_t *));
	nvdevs = md_scan_dag(mdp, NULL, md_find_name(mdp, VDEV_NAME),
	    md_find_name(mdp, "fwd"), vdevs);

	ASSERT(nvdevs > 0);

	for (i = 0; i < nvdevs; i++) {
		if (md_get_prop_str(mdp, vdevs[i], "name", &name) < 0) {
			DBG(zprintf("virtual-device node has no name prop\n"));
			Xfree(vdevs);
			return (-1);
		}

		if (strcmp(name, "console") != 0)
			continue;

		/* found the console node */
		cons = vdevs[i];

		/* add the service name */
		if (ldp->console.service_name != NULL) {
			md_add_string_property(mdp, cons, "service-name",
			    ldp->console.service_name);
		}

		/* add the group name */
		if (ldp->console.group_name != NULL) {
			md_add_string_property(mdp, cons, "group-name",
			    ldp->console.group_name);
		}

		/* add the ldc channel */
		/* XXXMIG - make this a real endpoint node? */
		md_add_value_property(mdp, cons, "ldc-id", ldcid);

		/* only one console */
		break;
	}

	Xfree(vdevs);

	if (i == nvdevs) {
		DBG(zprintf("No console node in MD for LDom %s\n", ldp->namep));
		return (-1);
	}

	return (0);
}

/*
 * XXXMIG - For the initial release of migration, there is no support
 * for console devices that have a manually assigned group or port.
 * All migrated domains are given a console connected to any available
 * port on the first vcc on the control domain, using the domain name
 * as the group name. This function assumes that a check has already
 * been performed to ensure that a valid vcc device exists with an
 * available console port.
 */
int
console_mig_bind(ldom_t *ldp, md_t *mdp, md_t *hvmdp, response_t *resp)
{
	int		i;
	md_node_t	**vdevs;
	int		nvdevs;
	uint64_t	ldcid;
	char		*group_name;
	vcc_t		*vcc;
	console_group_t	*group;
	vcc_console_t	*vcons;
	int		rc;
	char		*name;
	md_node_t	*cons;

	DBG(zprintf("%s\n", __func__));

	vdevs = Xmalloc(md_node_count(mdp) * sizeof (md_node_t *));
	nvdevs = md_scan_dag(mdp, NULL, md_find_name(mdp, VDEV_NAME),
	    md_find_name(mdp, "fwd"), vdevs);

	ASSERT(nvdevs > 0);

	for (i = 0; i < nvdevs; i++) {
		if (md_get_prop_str(mdp, vdevs[i], "name", &name) < 0) {
			DBG(zprintf("virtual-device node has no name prop\n"));
			Xfree(vdevs);
			return (-1);
		}

		if (strcmp(name, "console") != 0)
			continue;

		/* found the console node */
		cons = vdevs[i];

		/* XXXMIG - do something with service-name, group-name? */

		/* get the ldc channel */
		if (md_get_prop_val(mdp, cons, "ldc-id", &ldcid) != 0) {
			DBG(zprintf("no ldc-id property for console\n"));
			Xfree(vdevs);
			return (-1);
		}
		break;
	}

	Xfree(vdevs);

	if (i == nvdevs) {
		DBG(zprintf("No console node in MD for LDom %s\n", ldp->namep));
		return (-1);
	}

	/* XXXMIG - just use the first vcc for now */
	vcc = CHAIN_START(zeus.hvctl_ldom_p->vccs);
	if ((vcc == NULL) || (vcc->binding == NULL)) {
		rprintf(resp, MSG_ERR_MIG_NO_TGT_VCC);
		return (-1);
	}

	/*
	 * XXXMIG - Use the default group name and an automatically
	 * assigned port number for now.
	 */
	group_name = ldp->namep;
	ldp->console.tcp_port = _AUTO_VAL_;

	group = vcc_get_console_group(vcc, group_name, ldp->console.tcp_port);
	if (group == NULL) {
		rprintf(resp, "Unable to bind console for LDom %s\n",
		    ldp->namep);
		return (-1);
	}

	(void) init_console_binding(&ldp->console, NEW_INSTANCE, TYPE_GUEST, 0);

	/* create and link to the server side */
	vcons = vcc_new_console(vcc, group, NEW_CONSOLE_PORT);
	vcons->console = &ldp->console;
	ldp->console.binding->vcc_cons = vcons;

	/* allocate the console and server LDC channels */
	rc = ldc_mig_pull_ids(ldp, hvmdp, &ldp->guest_free_ldcs, 1, &ldcid,
	    &ldp->console.binding->ldcs);

	if (rc == 0) {
		rc = ldc_mig_get_and_pull_peer_ids(ldp, hvmdp,
		    &ldp->console.binding->ldcs, &vcons->ldcs,
		    &vcc->ldp->guest_free_ldcs);
	}
	if (rc != 0) {
		destroy_console_binding(&ldp->console);
		DBG(zprintf("Insufficient free LDC channels to bind "
		    "console on LDom %s\n", ldp->namep));
		return (rc);
	}

	if (vcc->ldp != ldp) {
		vcc->ldp->bound.num_clients++;
		vcc->ldp->flag.gen_new_md = true;
		vcc->ldp->flag.md_notify = true;
	}

	return (0);
}

void
console_print_bindings(ldom_t *ldp, response_t *resp)
{
	char			idstr[UINT64_STR_LEN];
	ldc_chan_t		*ldcp;
	console_binding_t	*bindp;

	bindp = ldp->console.binding;

	(void) snprintf(idstr, UINT64_STR_LEN, "console%d", bindp->id->id);
	rprintf(resp, "  %-11s", idstr);

	CHAIN_ITER(bindp->ldcs.ldcs, ldcp, link) {
		rprintf(resp, "%2"PRId64" ", ldcp->chan->id);
	}

	rprintf(resp, "\n");
}

/*
 * Check whether the specified port is available in the server's
 * port range and not used by another group
 */
static bool_t
is_port_avail(vcc_t *vcc, response_t *resp, int tcp_port)
{
	if (IS_VAL_AUTO(tcp_port)) {
		if (available_res_ids(&vcc->binding->tcp_pool) <= 0) {
			if (resp)
				rprintf(resp, gettext("No free tcp port is "
				    "available in this vcc\n"));
			return (false);
		} else {
			return (true);
		}
	} else if (!PORT_IN_RANGE(tcp_port, vcc)) {
		if (resp != NULL)
			rprintf(resp, gettext("Tcp port %ld is outside vcc "
			    "server's port range\n"), tcp_port);
		return (false);
	} else {
		return (true);
	}
}

#if LIVE_GRP_REBIND
/*
 * TODO
 * The LIVE_GRP_REBIND code needs to be validated once support is available in
 * vcc/vntsd
 */
static bool_t
validate_port(vcc_t *vcc, response_t *resp, char *group_name, int tcp_port)
{
	console_group_t *group = NULL;

	/*
	 * If we are trying to set to default tcp_port and need a
	 * new group, but no new tcp_port exists, return false
	 */
	if (is_port_avail(vcc, resp, tcp_port) == false)
		return (false);

	if (IS_VAL_AUTO(tcp_port))
		return (true);

	/*
	 * Check if a group already owns this tcp_port
	 */
	group = console_group_find_by_port(vcc, tcp_port);

	if (group == NULL)
		return (true);

	if (group_name && strcmp(group->group_name, group_name)) {
		if (resp != NULL)
			rprintf(resp, gettext("Tcp port %ld is already used by "
			    "group %s in the same vcc server\n"), tcp_port,
			    group->group_name);
		return (false);
	}

	return (true);
}

/*
 * Check if a bound console could be bound to a new service. The
 * console client should have been created with the console attributes
 * This may be invoked to check whether an unbound console could
 * be bound to a new service -OR- whether a currently bound
 * console could be bound to a different vcc service
 */
static int
can_bind_new_console(ldom_t *ldp, response_t *resp,
	char *new_service, char *new_group, int new_port)
{

	vcc_t *vcc;
	char *group_name, *grp_pstr;
	int port;
	console_group_t *ng_p = NULL;
	console_group_t *npg_p = NULL;

	ASSERT(new_service != NULL);

	vcc = vcc_find_by_name(new_service);

	if (vcc == NULL) {
		rprintf(resp, gettext("The vcc service named %s "
		    "does not exist\n"), new_service);
		return (false);
	} else if (vcc->ldp->state < BOUND) {
		rprintf(resp, gettext("The service named %s belongs to "
		    "the service domain %s which is inactive\n"),
		    new_service, vcc->ldp->namep);
		return (false);
	}

	port = new_port;
	group_name = new_group;
	grp_pstr = (group_name == NULL) ? "Unspecified" : group_name;

	if (group_name)
		ng_p = console_group_find_by_name(vcc, group_name);
	if (!IS_VAL_AUTO(port))
		npg_p = console_group_find_by_port(vcc, port);

	if (ng_p != NULL) {
		if (IS_VAL_AUTO(port) || ng_p->tcp_port->id == port) {
			return (true);
		} else {
			rprintf(resp, gettext("Cannot specify port %ld "
			    "for group %s because it is bound to port "
			    "%ld\n"), port, grp_pstr,
			    ng_p->tcp_port->id);
			return (false);
		}
	} else if (npg_p != NULL) {
		if (group_name == NULL || strcmp(npg_p->group_name,
		    group_name) == 0) {
			return (true);
		} else {
			if (port > 0)
				rprintf(resp, gettext("Cannot specify group %s "
				    "for port %ld because it is bound to group "
				    "%s\n"), grp_pstr, port, npg_p->group_name);
			return (false);
		}
	} else {
		return (validate_port(vcc, resp, group_name, port));
	}
}

/*
 * Allow the console client in a guest to change the service
 * and the group and also specify the port number. If the
 * port number is 0, an available port is chosen.
 * Return 0 if changes took effect
 * This routine is invoked only when a valid console client
 * has already been created for this guest
 * Valid values are:
 *
 * service_name/group_name
 * 	NULL - No Change
 * 	_AUTO_STR_ - Set to auto selection
 * 	<valid string> - Set to a valid service or group name
 * port
 * 	-1 = No Change
 * 	0 - Auto Selection
 * 	<valid port> - Set to a valid TCP port
 */
int
console_update_binding(ldom_t *ldp, response_t *resp, char *new_service,
	char *new_group, int new_port, int mod, int bindmod)
{
	console_binding_t *binding = ldp->console.binding;

	if ((bindmod & CONSOLE_SERVICE) == CONSOLE_SERVICE) {
		console_unbind(ldp);
		update_cons(ldp, mod, new_service, new_group, new_port);
		if (console_bind(ldp, resp) == false)
			return (ERR_GENERIC);
	} else if ((bindmod & (CONSOLE_GROUP|CONSOLE_PORT)) != 0) {
		/*
		 * only the group and/or port under the same vcc service
		 * being modified. The console remains connected to the
		 * original vcc using the same ldc, only the way vcc logically
		 * separates the console from others is affected.
		 *
		 * TODO: This Needs to be tested with solaris once live group
		 * re-binding is supported.
		 */
		vcc_console_t *vcons = binding->vcc_cons;
		vcc_t *vcc = vcons->vcc;

		/*
		 * If requested changes affect current console binding
		 * return the group and get a new group
		 */
		release_console_group(vcons->group);

		if ((vcons->group = console_find_group(ldp, vcc, new_group,
		    new_port)) == NULL) {
			rprintf(resp, gettext("Could not get console group for "
			    "LDom %s, the group and/or port may be in\nuse "
			    "by other consoles\n"), ldp->namep);
			return (ERR_CONSOLE_NO_GROUP);
		}

		vcc->ldp->flag.gen_new_md = true;
		vcc->ldp->flag.md_notify = true;
	}

	return (0);
}
#endif /* LIVE_GRP_REBIND */
