/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2008 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)core_md.c	1.7	08/11/04 SMI"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <alloca.h>
#include <errno.h>
#include <sys/processor.h>

#include "fatal.h"
#include "basics.h"
#include "allocate.h"
#include "clp.h"
#include "mdesc.h"
#include "mdesc_impl.h"
#include "ldom.h"
#include "zpri.h"
#include "backend.h"
#include "zprintf.h"
#include "core_md.h"

static void
add_md_version(md_t *core_mdp, md_node_t *rootnode)
{
	static uint64_t update_no = 0;

	md_add_value_property(core_mdp, rootnode, "md_update_no", update_no);

	update_no++;
}

static void
add_ldoms_version(md_t *core_mdp, md_node_t *rootnode)
{
	uint64_t	major_version = 1;
	uint64_t	minor_version = 0;

	md_add_value_property(core_mdp, rootnode, "ldm_version_major",
	    major_version);
	md_add_value_property(core_mdp, rootnode, "ldm_version_minor",
	    minor_version);
}

static void
add_cpu_op_pm_status(md_t *core_mdp, int num_nodes, md_node_t *rootnode)
{
	md_node_t	**cpulistp;
	int		ncpus, x;

	cpulistp = Xmalloc(md_node_count(core_mdp) * sizeof (md_node_t *));

	/*
	 * Starting at the root node, scan the "fwd" dag for
	 * all the cpus in this description.
	 */
	ncpus = md_scan_dag(core_mdp, rootnode,
	    md_find_name(core_mdp, "cpu"),
	    md_find_name(core_mdp, "fwd"), cpulistp);

	/*
	 * For each cpu node add the property "op_status" to show if the CPU
	 * is ONLINE or OFFLINE. "op_status" initial state will always be
	 * ONLINE.  Similarly, "pm_status" is initially marked as normal.
	 */
	for (x = 0; x < ncpus; x++) {
		md_add_value_property(core_mdp, cpulistp[x], CORE_MD_OP_STATUS,
		    (uint64_t)P_ONLINE);
		md_add_value_property(core_mdp, cpulistp[x], CORE_MD_PM_STATUS,
		    (uint64_t)PM_NORMAL);
	}

	Xfree(cpulistp);
}

/*
 * augment_core_md() aguments the core MD with additional properties that
 * external programs need to continue operations.
 */
void
augment_core_md(md_t *core_md)
{
	int		num_nodes;
	md_node_t	*rootnode;

	rootnode = md_root_node(core_md);

	num_nodes = md_node_count(core_md);

	/* Add "op_status" & "pm_status" properties to each CPU node */
	add_cpu_op_pm_status(core_md, num_nodes, rootnode);

	/* Add the ldoms version number to the root node */
	add_ldoms_version(core_md, rootnode);

	/* Add the MD number to the root number */
	add_md_version(core_md, rootnode);
}

int
set_core_md_cpu_status(uint64_t pid, char *status_prop, uint64_t cpu_status)
{
	md_node_t	*rootnode;
	int		num_nodes, ncpus, x;
	md_node_t	**cpulistp;
	uint64_t	id;

	/* Sanity check parameters */
	if (strcmp(status_prop, CORE_MD_OP_STATUS) == 0) {
		if (cpu_status != P_ONLINE && cpu_status != P_OFFLINE)
			return (-1);
	} else if (strcmp(status_prop, CORE_MD_PM_STATUS) == 0) {
		if (cpu_status != PM_NORMAL && cpu_status != PM_MANAGED)
			return (-1);
	} else {
		return (-1);
	}

	rootnode = md_root_node(pri.core_mdp);
	num_nodes = md_node_count(pri.core_mdp);

	cpulistp = Xcalloc(num_nodes, md_node_t *);

	/*
	 * Starting at the root node, scan the "fwd" dag for
	 * all the cpus in this description.
	 */
	ncpus = md_scan_dag(pri.core_mdp, rootnode,
	    md_find_name(pri.core_mdp, "cpu"),
	    md_find_name(pri.core_mdp, "fwd"), cpulistp);

	/*
	 * For each cpu node check the id to see if this is the node we want.
	 * If so then update the op_status property and return it.
	 */
	for (x = 0; x < ncpus; x++) {
		id = pri_cpu_phys_id(pri.core_mdp, cpulistp[x]);
		if (id == pid) {
			md_update_value_property(pri.core_mdp, cpulistp[x],
			    status_prop, cpu_status);
			break;
		}
	}

	Xfree(cpulistp);

	if (x == ncpus)
		return (-1);

	return (0);
}

int
get_core_md_cpu_status(uint64_t pid, char *status_prop)
{
	md_node_t	*rootnode;
	int		num_nodes, ncpus, x;
	md_node_t	**cpulistp;
	uint64_t	id, status, cpu_status;

	rootnode = md_root_node(pri.core_mdp);
	num_nodes = md_node_count(pri.core_mdp);

	cpulistp = Xcalloc(num_nodes, md_node_t *);

	/*
	 * Starting at the root node, scan the "fwd" dag for
	 * all the cpus in this description.
	 */
	ncpus = md_scan_dag(pri.core_mdp, rootnode,
	    md_find_name(pri.core_mdp, "cpu"),
	    md_find_name(pri.core_mdp, "fwd"), cpulistp);

	/*
	 * For each cpu node check the id to see if this is the node we want.
	 * If so then update the op_status property and return it.
	 */
	for (x = 0; x < ncpus; x++) {
		id = pri_cpu_phys_id(pri.core_mdp, cpulistp[x]);
		if (id == pid) {
			status = md_get_prop_val(pri.core_mdp, cpulistp[x],
			    status_prop, &cpu_status);
			break;
		}
	}

	Xfree(cpulistp);

	if (status != 0 || x == ncpus)
		return (-1);

	return (cpu_status);
}
