/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <values.h>
#include <libintl.h>

#include "allocate.h"
#include "basics.h"
#include "clp.h"
#include "core_md.h"
#include "database.h"
#include "dr_cpu.h"
#include "ds.h"
#include "errors.h"
#include "fatal.h"
#include "ldom.h"
#include "mac_addrs.h"
#include "mdesc_impl.h"
#include "mig.h"
#include "pm_events.h"
#include "option.h"
#include "sequencer.h"
#include "variables.h"
#include "vcpu.h"
#include "vmau.h"
#include "zpri.h"
#include "zprintf.h"
#include "autosave.h"

vcpu_t *
vid_to_vcpu(ldom_t *ldp, int vid)
{
	vcpu_t *vcpu;

	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		if (vcpu->vid == vid)
			break;
	}

	return (vcpu);
}

vcpu_t *
pid_to_vcpu(ldom_t *ldp, int pid)
{
	vcpu_t *vcpu;

	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		if (vcpu->cpu->phys_id == pid)
			break;
	}

	return (vcpu);
}

pri_cpu_t *
pid_to_pcpu(int pid)
{
	pri_cpu_t *pcpu;

	CHAIN_ITER(pri.cpu_list, pcpu, link) {
		if (pcpu->phys_id == pid)
			break;
	}

	return (pcpu);
}

bool_t
pid_to_vid(int pid, ldom_t **ldp, int *vid) {
	vcpu_t *vcpu;

	CHAIN_ITER(zeus.ldoms, *ldp, link) {
		if ((vcpu = pid_to_vcpu(*ldp, pid)) != NULL) {
			*vid = vcpu->vid;
			return (true);
		}
	}

	return (false);
}

static resid_t *
get_mmu_partid(ldom_t *ldp, pri_core_t *core, int mmu_partid)
{
	vcpu_t *vcpu;
	resid_t *partid;

	/* look for other vcpu bindings for the ldom on the core */
	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		if (vcpu->cpu->core == core)
			break;
	}

	if (vcpu == NULL) {
		/*
		 * No existing vcpu bindings for the ldom on the core.
		 * Either allocate a new one or the requested one.
		 */
		if (mmu_partid == NEW_MMU_PARTID)
			partid = alloc_res(&core->mmu_free_partids);
		else
			partid = pull_res_id(&core->mmu_free_partids,
			    mmu_partid);

		if (partid == NULL)
			fatal("Failed to allocate MMU partition ID %d",
			    mmu_partid);
	} else {
		/*
		 * There are existing vcpu bindings for the ldom on the core.
		 * Check that a either a new MMU partid is being requested or
		 * that the same MMU partid is being requested.
		 */
		if (!((mmu_partid == NEW_MMU_PARTID) ||
		    (mmu_partid == vcpu->partid->id)))
			fatal("Inconsistent MMU partition IDs");

		partid = vcpu->partid;
	}

	return (partid);
}

static void
core_free_mmu_partid(pri_core_t *core, resid_t *partid)
{
	pri_cpu_t *cpu;
	vcpu_t *vcpu;

	/* check if any bindings on the core are using the partid */
	CHAIN_DOUBLE_ITER(core->cpus, cpu, core_link,
	    bound.vcpus, vcpu, pcpu) {
		if (vcpu->partid->id == partid->id)
			return;
	}

	free_res(partid);
}

static int
vid_cmp(vcpu_t *new, vcpu_t *current)
{
	if (new->vid < current->vid)
		return (-1);
	else if (new->vid > current->vid)
		return (1);
	else
		return (0);
}

static int
pid_cmp(vcpu_t *new, vcpu_t *current)
{
	if (new->cpu->phys_id < current->cpu->phys_id)
		return (-1);
	else if (new->cpu->phys_id > current->cpu->phys_id)
		return (1);
	else
		return (0);
}

/*
 * Bind a virtual CPU to a specific physical CPU.
 */
bool_t
cpu_bind(pri_cpu_t *pcpu, ldom_t *ldp, int vid, int mmu_partid, int fraction)
{
	pri_core_t *core = pcpu->core;
	vcpu_t *vcpu, *tmp1, *tmp2;
	resid_t *vcpuid;

	DBG(zprintf("Attempting to bind %s cpu %d to pstrand %d\n",
	    ldp->namep, vid, pcpu->phys_id));

	if (fraction > pcpu->bound.fraction_free)
		return (false);

	/* the vcpu's resource ID is actually stored in the pcpu struct */
	if (pcpu->res_id == NULL) {
		pcpu->res_id = alloc_res(&zeus.free_vcpu_ids);
		if (pcpu->res_id == NULL)
			fatal("Out of VCPU resource IDs");
	}

	if (vid == NEW_VCPUID)
		vcpuid = alloc_res(&ldp->guest_vcpu_ids);
	else
		vcpuid = pull_res_id(&ldp->guest_vcpu_ids, vid);

	if (vcpuid == NULL) {
		DBG(zprintf("Unable to obtain vcpu %d\n", vid));
		return (false);
	}

	vcpu = Xmalloc(sizeof (*vcpu));

	vcpu->vcpuid = vcpuid;
	vcpu->vid = vcpu->vcpuid->id;
	vcpu->partid = get_mmu_partid(ldp, core, mmu_partid);
	vcpu->fraction = fraction;
	vcpu->cpu = pcpu;
	vcpu->ldp = ldp;
	vcpu->vmau = NULL;
	vcpu->flag_isconfigd = false;
	vcpu->util_idx = 0;
	(void) memset(vcpu->util_table, 0, sizeof (vcpu->util_table));

	pcpu->bound.fraction_free -= fraction;

	CHAIN_ADD_SORTED(ldp->bound.vcpus, vcpu, tmp1, tmp2, ldom, vid_cmp);
	CHAIN_ADD(pcpu->bound.vcpus, vcpu, pcpu);

	return (true);
}

int
ldom_num_bound_vcpus(ldom_t *ldp)
{
	return (CHAIN_LENGTH(ldp->bound.vcpus));
}

int
ldom_num_rsvd_vcpus(ldom_t *ldp)
{
	return (CHAIN_LENGTH(ldp->mig_rsvd.vcpus));
}

int
ldom_num_faulted_vcpus(ldom_t *ldp)
{
	md_node_t	*rootnode;
	int		num_nodes, ncpus, x;
	md_node_t	**cpulistp;
	uint64_t	id, status, cpu_status;
	int		num = 0;
	vcpu_t		*vcpu;

	rootnode = md_root_node(pri.core_mdp);
	num_nodes = md_node_count(pri.core_mdp);

	cpulistp = Xcalloc(num_nodes, md_node_t *);

	/*
	 * Starting at the root node, scan the "fwd" dag for
	 * all the cpus in this description.
	 */
	ncpus = md_scan_dag(pri.core_mdp, rootnode,
	    md_find_name(pri.core_mdp, "cpu"),
	    md_find_name(pri.core_mdp, "fwd"), cpulistp);

	/*
	 * For each vcpu in the given ldom_t, try to find the matching cpu.
	 * If found, get the cpu status and increment the count if it's faulted.
	 */
	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		for (x = 0; x < ncpus; x++) {
			id = pri_cpu_phys_id(pri.core_mdp, cpulistp[x]);
			if (id == vcpu->cpu->phys_id) {
				status = md_get_prop_val(pri.core_mdp,
				    cpulistp[x], CORE_MD_OP_STATUS,
				    &cpu_status);

				if ((status == 0) && (cpu_status != P_ONLINE))
					num++;

				break;
			}
		}
	}

	Xfree(cpulistp);

	return (num);
}

int
num_bound_vcpus(void)
{
	pri_cpu_t *pcpu;
	int nbound = 0;

	CHAIN_ITER(pri.cpu_list, pcpu, link) {
		nbound += CHAIN_LENGTH(pcpu->bound.vcpus);
	}

	return (nbound);
}

int
num_free_vcpus(void)
{
	pri_cpu_t *pcpu;
	int nfree = 0;

	CHAIN_ITER(pri.cpu_list, pcpu, link) {
		nfree += pcpu->bound.fraction_free;
	}

	return (nfree);
}

static bool_t
bind_num_vcpus(ldom_t *ldp, int num_to_bind, response_t *resp)
{
	int		i = 0;
	pri_cpu_t	*pcpu;
	pm_resid_t	*resids = NULL;
	char		*res_name = CPU_RES_NAME;
	ldm_err_t	pm_result = { 0 };
	bool_t		ret = true;
	int		num_free;
	vcpu_t		*vcpu = NULL;
	uint32_t	*vcpuids = NULL;

	num_free = num_free_vcpus();

	if (num_free < num_to_bind) {
		rprintf(resp, gettext(
		    "Insufficient VCPU resources to bind LDom %s\n"),
		    ldp->namep);
		return (false);
	}

	if (pm_cpu_events_on) {
		resids = Xmalloc(sizeof (pm_resid_t) * num_to_bind);
		vcpuids = Xmalloc(sizeof (uint32_t) * num_to_bind);
	}

	pcpu = CHAIN_START(pri.cpu_list);
	while (num_to_bind > 0 && pcpu != NULL) {
		if (pcpu->bound.fraction_free > 0) {
			if (cpu_bind(pcpu, ldp, NEW_VCPUID, NEW_MMU_PARTID,
			    PRI_CPU_FRAC_DENOM)) {
				if (resids != NULL) {
					resids[i].pi_res_type = PM_RES_CPU;
					resids[i].pi_res_name = res_name;
					resids[i].u.cpu.pi_id_val =
					    pcpu->phys_id;
					vcpu = pid_to_vcpu(ldp, pcpu->phys_id);
					vcpuids[i] = vcpu->vid;
				}
				num_to_bind--;
				i++;
			} else {
				/* error */
				break;
			}
		} else {
			pcpu = CHAIN_NEXT(pcpu->link);
		}
	}

	if (num_to_bind > 0)
		fatal("Binding failed after finding sufficient CPUs to bind");

	if (pm_cpu_events_on) {
		send_pm_cpu_event(ldp, ADD_RESOURCE_PRE, resids, i, &pm_result);

		if (pm_result.ldme_errs != LDME_OP_EOK) {
			rprintf(resp, gettext("Power Management Failure\n"));
			cleanup_partial_vcpu_add(ldp, num_to_bind, vcpuids);
			ret = false;
		}
		pm_free_errors(&pm_result);
		Xfree(vcpuids);
		Xfree(resids);
	}

	return (ret);
}

static bool_t
bind_explicit_vcpus(ldom_t *ldp, cpu_adm_t *cpu_adm, response_t *resp)
{
	int		i;
	pri_cpu_t	*pcpu;
	char		*res_name = CPU_RES_NAME;
	pm_resid_t	*resids = NULL;
	ldm_err_t	pm_result = { 0 };
	bool_t		ret = true;
	int		pid;
	vcpu_t		*vcpu = NULL;
	uint32_t	*vcpuids = NULL;

	ASSERT(cpu_adm->type == CPU_ADM_IDS);

	for (i = 0; i < cpu_adm->num; i++) {
		pid = cpu_adm->ids[i];
		pcpu = pid_to_pcpu(pid);
		if (pcpu == NULL) {
			rprintf(resp, gettext("No physical CPU ID %d\n"), pid);
			return (false);
		}

		if (pcpu->bound.fraction_free == 0) {
			rprintf(resp,
			    gettext("Physical CPU ID %d is not free\n"), pid);
			return (false);
		}
	}

	if (pm_cpu_events_on) {
		resids = Xmalloc(sizeof (pm_resid_t) * cpu_adm->num);
		vcpuids = Xmalloc(sizeof (uint32_t) * cpu_adm->num);
	}

	for (i = 0; i < cpu_adm->num; i++) {
		pid = cpu_adm->ids[i];
		pcpu = pid_to_pcpu(pid);
		if (!cpu_bind(pcpu, ldp, NEW_VCPUID, NEW_MMU_PARTID,
		    PRI_CPU_FRAC_DENOM))
			fatal("Binding failed after finding CPU to bind");
		if (resids != NULL) {
			resids[i].pi_res_type = PM_RES_CPU;
			resids[i].pi_res_name = res_name;
			resids[i].u.cpu.pi_id_val = pcpu->phys_id;
			vcpu = pid_to_vcpu(ldp, pcpu->phys_id);
			vcpuids[i] = vcpu->vid;
		}
	}

	if (pm_cpu_events_on) {
		send_pm_cpu_event(ldp, ADD_RESOURCE_PRE, resids, i, &pm_result);

		if (pm_result.ldme_errs != LDME_OP_EOK) {
			rprintf(resp, gettext("Power Management Failure\n"));
			cleanup_partial_vcpu_add(ldp, cpu_adm->num, vcpuids);
			ret = false;
		}
		pm_free_errors(&pm_result);
		Xfree(resids);
		Xfree(vcpuids);
	}

	return (ret);
}

/*
 * Using the constraints allocate and bind the VCPUs for the given LDom.
 */
bool_t
cons_bind_vcpus(ldom_t *ldp, cpu_adm_t *cpu_adm, response_t *resp)
{
	if (cpu_adm->num == 0) {
		rprintf(resp, gettext("No VCPUs requested for LDom %s\n"),
		    ldp->namep);
		return (false);
	}

	if (cpu_adm->type == CPU_ADM_NUM)
		return (bind_num_vcpus(ldp, cpu_adm->num, resp));
	else
		return (bind_explicit_vcpus(ldp, cpu_adm, resp));
}

static void
cpu_unbind(vcpu_t *vcpu, bool_t reserved)
{
	ldom_t *ldp = vcpu->ldp;
	pri_cpu_t *pcpu = vcpu->cpu;
	pri_core_t *core = pcpu->core;

	pcpu->bound.fraction_free += vcpu->fraction;
	ASSERT(pcpu->bound.fraction_free <= PRI_CPU_FRAC_DENOM);

	/* XXXMIG - This flag business is nasty. Patch up ASAP */
	if (reserved)
		CHAIN_UNLINK(ldp->mig_rsvd.vcpus, vcpu, ldom);
	else
		CHAIN_UNLINK(ldp->bound.vcpus, vcpu, ldom);

	CHAIN_UNLINK(pcpu->bound.vcpus, vcpu, pcpu);

	core_free_mmu_partid(core, vcpu->partid);

	/*
	 * If the bound cpu is linked in the mau chain, remove
	 * from that chain.
	 */
	if (vcpu->vmau)
		CHAIN_UNLINK(vcpu->vmau->vcpus, vcpu, mau);

	if (vcpu->vcpuid)
		free_res(vcpu->vcpuid);		/* guest vcpu id */

	/*
	 * Need to make sure the core_md status for vcpu is set back to
	 * PM_NORMAL.
	 */
	set_core_md_cpu_status(pcpu->phys_id, CORE_MD_PM_STATUS,
	    PM_NORMAL);

	/*
	 * We do not free the vcpu's resource ID (which was stored in the pcpu
	 * struct to which it was bound).  This keeps this resource ID tied to
	 * the physical strand to which this vcpu had been bound for all future
	 * bindings of this strand, regardless of vcpu.
	 *
	 * FIXME: It's unclear whether this is the desired behavior.  Add the
	 * following lines to properly free the resource ID if needed:
	 *
	 * free_res(pcpu->res_id);
	 * pcpu->res_id = NULL;
	 */

	Xfree(vcpu);
}

void
cpu_unbind_all(ldom_t *ldp)
{
	vcpu_t *vcpu;
	vcpu_t *next;

	CHAIN_ITER_VOL(ldp->bound.vcpus, vcpu, next, ldom) {
		cpu_unbind(vcpu, false);
	}

	CHAIN_ITER_VOL(ldp->mig_rsvd.vcpus, vcpu, next, ldom) {
		cpu_unbind(vcpu, true);
	}
}

static bool_t
vcpu_in_list(uint32_t vid, int num, uint32_t *vcpu_list)
{
	int i;
	for (i = 0; i < num; i++)
		if (vid == vcpu_list[i])
			return (true);

	return (false);
}

/*
 * Unbind the specified vcpus from the guest. We go through the
 * bound vcpus in the guest, and unbind the ones in the list.
 */
void
cleanup_partial_vcpu_add(ldom_t *ldp, int num, uint32_t *vcpu_ids)
{
	vcpu_t *vcpu, *nextvcpu;

	CHAIN_ITER_VOL(ldp->bound.vcpus, vcpu, nextvcpu, ldom) {
		if (vcpu_in_list(vcpu->vid, num, vcpu_ids) == true) {
			DBG(zprintf("%s: unbinding vcpu %d\n",
			    __func__, vcpu->vid));
			cpu_unbind(vcpu, false);
		}
	}
}

static bool_t check_cpu_mau_sharing(ldom_t *ldp, vcpu_t **vcpu, int n_cpus);

static int
dr_add_cpus(ldom_t *ldp, int num, response_t *resp)
{
	int ret;

	if (zeus.mode == ZEUS_CONFIG_MODE)
		return (0);

	if ((ret = seq_add_vcpus(ldp, resp)) == CPU_DR_OP_FAILED) {
		ldp->required.num_vcpus -= num;
		rprintf(resp, gettext(
		    "Unable to add the requested VCPU(s) to LDom %s\n"),
		    ldp->namep);
	} else if (ret == CPU_DR_OP_NO_RESPONSE) {
		rprintf(resp, gettext(
		    "VCPU(s) are considered allocated to LDom, but might not "
		    "be available\nto the guest OS\n"));
	} else if ((ret != 0) && (ret != num)) {
		/* reduce required by amount that couldn't be allocated */
		ldp->required.num_vcpus -= (num - ret);
		rprintf(resp,
		    gettext("Only %d VCPUs could be added to LDom %s\n"),
		    ret, ldp->namep);
	}

	return (ret);
}

int
check_cpu_dr_capable(ldom_t *ldp)
{
	int bound_maus = ldom_num_bound_maus(ldp);

	ASSERT(ldp->state == ACTIVE);

	/* check if the domain is using a local version of tick */
	if (!MIG_INPROGRESS(ldp) && ldp->flag.local_tick_sync)
		return (ERR_CPU_DR_MIG_DISABLED);

	if (ds_lookup_svc(ldp->ds_chan, DR_CPU_SVC) == NULL)
		return (ERR_CPU_DR_NOTSUP);

	if (ds_lookup_svc(ldp->ds_chan, MD_UPDATE_SVC) == NULL)
		return (ERR_MD_UPDATE_NOTSUP);

	/*
	 * Check for crypto units in the domain. The presence of crypto
	 * units could affect the ability to add or remove CPUs.
	 */
	if (bound_maus > 0) {

		/* check if the user has turned off crypto DR manually */
		if (!option.allow_crypto_dr)
			return (ERR_CPU_DR_BOUND_MAUS);

		/*
		 * Crypto DR is enabled, so if there are crypto devices
		 * in the domain, make sure that the necessary domain
		 * services are registered. This is an indication that
		 * the domain can handle vcpus associated with a crypto
		 * unit coming and going. Orphaning a crypto unit is not
		 * allowed, but is caught later in the operation.
		 */
		if (check_crypto_dr_services(ldp) != 0)
			return (ERR_CPU_DR_BOUND_MAUS);
	}

	return (CPU_DR_OP_SUCC);
}

int
cli_check_cpu_dr_capable(ldom_t *ldp, response_t *resp)
{
	int rc;

	if ((rc = check_cpu_dr_capable(ldp)) != 0) {

		switch (rc) {
		case ERR_CPU_DR_BOUND_MAUS:
			rprintf(resp, gettext("VCPU DR disabled because "
			    "LDom %s has crypto units bound\n"), ldp->namep);
			break;

		case ERR_CPU_DR_NOTSUP:
			rprintf(resp, gettext("LDom %s is unable to dynamically"
			    " reconfigure VCPUs.  Please verify\nthe guest "
			    "operating system is running and supports VCPU DR."
			    "\n"), ldp->namep);
			break;

		case ERR_MD_UPDATE_NOTSUP:
			rprintf(resp, gettext("LDom %s does not support "
			    "runtime re-configuration\n"), ldp->namep);
			break;

		case ERR_CPU_DR_MIG_DISABLED:
			rprintf(resp, gettext("VCPU DR disabled because "
			    "LDom %s has been migrated.\nRebooting the "
			    "domain will remove this restriction\n"),
			    ldp->namep);
			break;

		default:
			FATAL_UNEXPECTED_ERROR(rc);
			break;
		}
	}

	return (rc);
}

bool_t
cpu_add(ldom_t *ldp, cpu_adm_t *cpu_adm, response_t *resp)
{
	int ret = 0;

	DBG(zprintf("%s: %"PRIu64"\n", __func__, cpu_adm->num));

	if (ldp->state == STARTING || ldp->state == STOPPING) {
		rprintf(resp, gettext("Cannot add VCPUs to LDom %s "
		    "while it is currently %s.\n"), ldp->namep,
		    ldom_state_str(ldp->state));
		return (false);
	}

	if (ldp->state >= BOUND) {
		if (ldp->state == ACTIVE && zeus.mode != ZEUS_CONFIG_MODE &&
		    in_delayed_reconfig(NULL) == false) {
			if (cli_check_cpu_dr_capable(ldp, resp) != 0)
				return (false);
		}

		if (cons_bind_vcpus(ldp, cpu_adm, resp) == false)
			return (false);

		ldp->required.num_vcpus += cpu_adm->num;

		/*
		 * For each of the just added vcpus, if there is an mau bound
		 * to this guest on the same core, we want to bind the vcpu to
		 * the mau.
		 * Rather than figure out which vcpus just got added, just
		 * rebind them all.
		 */
		vmau_t *vmau;
		CHAIN_ITER(ldp->bound.vmaus, vmau, ldom) {
			mau_link_vcpus(ldp, vmau);
		}

		if (ldp->state != BOUND) {
			/*
			 * If we're in a delayed reconfig state, we don't do
			 * CPU DR, instead we delay this operation as well.
			 */
			if (in_delayed_reconfig(ldp) == true &&
			    zeus.mode == ZEUS_MGMT_MODE) {
				/*
				 * Handle as a delayed reconfig operation.
				 */
				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_ADD, resp);
			} else {
				ret = dr_add_cpus(ldp, cpu_adm->num, resp);
			}
		} else {
			/* BOUND domain - we can reconfigure now */
			seq_download_mds_do_config(ldp);
		}
	} else {
		ldp->required.num_vcpus += cpu_adm->num;
	}

	if (ret != CPU_DR_OP_FAILED) {
		database_ldom_set_cpus(ldp, ldp->required.num_vcpus);
	}

	return (ret < 0 ? false : true);
}

static int
dr_rm_cpus(ldom_t *ldp, vcpu_t **vcpus, int *num, bool_t force,
    response_t *resp)
{
	uint32_t	*op_stats;
	int		*pm_stats;
	vcpu_t		**vcpus_to_dr;
	int		i, num_to_remove, rc;
	int		ret = CPU_DR_OP_SUCC;

	/* allocate an array of 64-bit status to hold DR status */
	op_stats = Xcalloc(*num, uint64_t);
	pm_stats = Xcalloc(*num, uint64_t);
	vcpus_to_dr = Xcalloc(*num, vcpu_t *);
	num_to_remove = 0;
	/*
	 * Create new list, that excludes vcpus bound to strands
	 * that are currently being power managed (and thus have
	 * already been unconfigured from the guest).
	 */
	for (i = 0, num_to_remove = 0; i < *num; i++) {
		pm_stats[i] = get_core_md_cpu_status(vcpus[i]->cpu->phys_id,
		    CORE_MD_PM_STATUS);
		if (pm_stats[i] == PM_NORMAL) {
			vcpus_to_dr[num_to_remove] = vcpus[i];
			num_to_remove++;
		} else {
			DBG(zprintf("Skipping DR of vcpu %d at index %d\n",
			    vcpus[i]->vid, i));
		}
	}

	if (zeus.mode == ZEUS_CONFIG_MODE || num_to_remove == 0)
		rc = num_to_remove;
	else
		rc = seq_rm_vcpus(ldp, vcpus_to_dr, num_to_remove, force,
		    op_stats, resp);

	if (rc == CPU_DR_OP_FAILED) {
		rprintf(resp, gettext("Unable to remove the requested "
		    "VCPU(s) from LDom %s\n"), ldp->namep);
		*num = 0;
		ret = ERR_CPU_DR_OP_FAILED;

	} else if (rc == CPU_DR_OP_NO_RESPONSE) {
		rprintf(resp, gettext(
		    "VCPU(s) will remain allocated to LDom, but might not be "
		    "available\nto the guest OS\n"));
		*num = 0;
		ret = ERR_CPU_DR_OP_NO_RESPONSE;

	} else if (rc < 0) {
		FATAL_UNEXPECTED_ERROR(rc);

	} else {
		int dr_idx, num_removed;

		num_removed = dr_idx = 0;
		for (i = 0; i < *num; i++) {
			/*
			 * Check whether this index corresponds to a vcpu that
			 * was requested to be unconfigured or one that was
			 * skipped (because its strand is being power managed).
			 * If the former, check the status of the unconfigure
			 * and unbind & count it only if successful.  If the
			 * latter, set the strand's PM status to PM_NORMAL,
			 * as that's the correct state for unallocated cpus.
			 */
			if (pm_stats[i] == PM_MANAGED) {
				DBG(zprintf("Updating PM_STATUS of vcpu"
				    " %d at index %d\n", vcpus[i]->vid, i));
				set_core_md_cpu_status(vcpus[i]->cpu->phys_id,
				    CORE_MD_PM_STATUS, PM_NORMAL);
				cpu_unbind(vcpus[i], false);
				num_removed++;
			} else {
				if (op_stats[dr_idx] == DR_CPU_RES_OK) {
					DBG(zprintf("DR out vcpu%d @ idx %d\n",
					    vcpus_to_dr[dr_idx]->vid, dr_idx));
					cpu_unbind(vcpus_to_dr[dr_idx], false);
					num_removed++;
				}
				dr_idx++;
			}
		}

		ASSERT(num_to_remove == dr_idx);

		seq_rm_vcpus_cleanup(ldp);

		if (num_removed == 0) {
			rprintf(resp, gettext("LDom %s didn't allow "
			    "removal of any VCPUs\n"), ldp->namep);
			ret = ERR_CPU_DR_OP_FAILED;
		} else if (num_removed != *num) {
			rprintf(resp, gettext("LDom %s permitted "
			    "removal of only %d VCPUs\n"),
			    ldp->namep, num_removed);
			ret = ERR_CPU_DR_PARTIAL;
		}

		*num = num_removed;

		if (num_removed > 0)
			autosave_update_autosave();
	}

	Xfree(vcpus_to_dr);
	Xfree(pm_stats);
	Xfree(op_stats);
	return (ret);
}

static int
get_vcpus_to_delete(ldom_t *ldp, cpu_adm_t *cpu_adm, vcpu_t **vcpus,
    response_t *resp)
{
	int	i;
	int	num_bound;
	vcpu_t	*vcpu;
	vcpu_t	*skip_vcpu = NULL;

	if (cpu_adm->type == CPU_ADM_NUM) {
		num_bound = ldom_num_bound_vcpus(ldp);

		if (cpu_adm->num == num_bound) {
			rprintf(resp, gettext("You cannot delete all the bound "
			    "VCPUs in an LDom\n"));
			return (ERR_CPU_DELETE_ALL);
		} else if (cpu_adm->num > num_bound) {
			rprintf(resp, gettext(
			    "LDom %s does not have %"PRIu64" VCPUs bound\n"),
			    ldp->namep, cpu_adm->num);
			return (ERR_CPU_DELETE_NUM_BOUND);
		}

		/*
		 * For now, just pull them off the list from highest to
		 * lowest vcpuid. We must also ensure that at least one
		 * non power-managed strand is left configured, else we
		 * could end up attempting to DR out all the remaining
		 * configured vcpus.
		 *
		 * To accomplish this, walk the bound vcpus twice. On
		 * the first pass, select the lowest numbered non power
		 * managed vcpu to be protected. On the second pass,
		 * select the vcpus to remove, skipping the vcpu selected
		 * in the first pass if necessary.
		 */
		CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
			if (get_core_md_cpu_status(vcpu->cpu->phys_id,
			    CORE_MD_PM_STATUS) == PM_NORMAL) {
				skip_vcpu = vcpu;
				break;
			}
		}

		/* must be at least one non power managed vcpu */
		ASSERT(skip_vcpu != NULL);

		i = 0;
		vcpu = CHAIN_END(ldp->bound.vcpus);
		while ((vcpu != NULL) && (i < cpu_adm->num)) {

			/* avoid the pre-selected vcpu */
			if (vcpu != skip_vcpu)
				vcpus[i++] = vcpu;

			vcpu = CHAIN_PREV(vcpu->ldom);
		}

		ASSERT(i == cpu_adm->num);

	} else {
		int pid;

		ASSERT(cpu_adm->type == CPU_ADM_IDS);

		for (i = 0; i < cpu_adm->num; i++) {
			pid = cpu_adm->ids[i];
			vcpus[i] = pid_to_vcpu(ldp, pid);
			if (vcpus[i] == NULL) {
				rprintf(resp, gettext(
				    "CPU pid %d is not bound in LDom %s\n"),
				    pid, ldp->namep);
				return (ERR_CPU_NOT_BOUND);
			}
		}
	}

	return (0);
}

bool_t
cpu_del(ldom_t *ldp, cpu_adm_t *cpu_adm, response_t *resp, bool_t force)
{
	vcpu_t **vcpus;
	int num = cpu_adm->num;
	int i;
	pm_resid_t *resids = NULL;
	char *res_name = CPU_RES_NAME;
	ldm_err_t pm_result = { 0 };
	pm_domain_event_info_t *event;

	DBG(zprintf("%s: %d\n", __func__, num));

	if (ldp->state == STARTING || ldp->state == STOPPING) {
		rprintf(resp, gettext("Cannot remove VCPUs from LDom %s "
		    "while it is currently %s.\n"), ldp->namep,
		    ldom_state_str(ldp->state));
		return (false);
	}

	switch (ldp->state) {
	case BOUND:
	case ACTIVE:
		if (ldp->state == ACTIVE && zeus.mode != ZEUS_CONFIG_MODE &&
		    in_delayed_reconfig(NULL) == false &&
		    cli_check_cpu_dr_capable(ldp, resp) != 0) {
			return (false);
		}

		vcpus = Xmalloc(num * sizeof (vcpu_t *));
		if (get_vcpus_to_delete(ldp, cpu_adm, vcpus, resp) != 0) {
			Xfree(vcpus);
			return (false);
		}

		/*
		 * Since crypto units can only be referenced by their
		 * association with a vcpu, operations that remove the
		 * last such association must be prevented.
		 */
		if (!check_cpu_mau_sharing(ldp, vcpus, num)) {
			/*
			 * At this point, it has been verified that the domain
			 * has crypto DR support. So, alert the user to remove
			 * the crypto unit before proceeding.
			 */
			rprintf(resp, gettext("Removing %d CPU(s) would leave "
			    "orphaned crypto unit(s).\nRemove one or more "
			    "crypto units and retry the operation.\n"), num);

			return (false);
		}

		if (pm_cpu_events_on)
			resids = Xmalloc(sizeof (pm_resid_t) * num);

		if (ldp->state == ACTIVE) {
			/*
			 * If we're in a delayed reconfig state, we don't do
			 * CPU DR, instead we delay this operation as well.
			 */
			if (in_delayed_reconfig(ldp) == true &&
			    zeus.mode == ZEUS_MGMT_MODE) {
				/*
				 * Handle as a delayed reconfig operation.
				 * First call unbind, then initiate the delayed
				 * reconfig.
				 */
				for (i = 0; i < num; i++) {
					if (resids != NULL) {
						resids[i].pi_res_type =
						    PM_RES_CPU;
						resids[i].pi_res_name =
						    res_name;
						resids[i].u.cpu.pi_id_val =
						    vcpus[i]->cpu->phys_id;
					}
					cpu_unbind(vcpus[i], false);
				}

				if (pm_cpu_events_on) {
					ldp->pm_vcpu_list = resids;
					ldp->pm_vcpu_count = num;
				} else {
					ldp->pm_vcpu_count = 0;
				}

				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_REMOVE_PHYS, resp);
			} else {
				if (dr_rm_cpus(ldp, vcpus, &num, force, resp)) {
					/* Failed to remove any */
					if (num == 0) {
						Xfree(vcpus);
						return (false);
					}
				}
			}
		} else {
			/* BOUND domain - unbind & reconfigure now */
			for (i = 0; i < num; i++) {
				if (resids != NULL) {
					resids[i].pi_res_type = PM_RES_CPU;
					resids[i].pi_res_name = res_name;
					resids[i].u.cpu.pi_id_val =
					    vcpus[i]->cpu->phys_id;
				}
				cpu_unbind(vcpus[i], false);
			}

			seq_download_mds_do_config(ldp);
			if (pm_cpu_events_on) {
				CHAIN_ITER(pm_domain_list.domain_event_list,
				    event, link) {
					if ((event->event_type ==
					    PM_RES_EVENT) ||
					    (event->res_type == PM_RES_CPU)) {
						break;
					}
				}

				pm_domain_config_change_notify(event->handle,
				    (srcid_t)ldp, REM_RESOURCE_POST,
				    PM_RES_CPU, resids, num, &pm_result);

				/* Free per resource status data */
				pm_free_errors(&pm_result);
				Xfree(resids);
			}
		}

		/* Complete or partial success */
		ldp->required.num_vcpus -= num;

		Xfree(vcpus);
		break;

	case INACTIVE:
		if (num > ldp->required.num_vcpus) {
			rprintf(resp,
			    gettext("You do not have %d VCPUs requested\n"),
			    num);
			return (false);
		}

		if ((ldp->required.num_vcpus -= num) == 0)
			rprintf(resp, gettext("Warning: setting required "
			    "number of VCPUs to zero\n"));
		break;
	default:
		abort();	/* Internal error */
	}

	database_ldom_set_cpus(ldp, ldp->required.num_vcpus);

	return (true);
}

bool_t
cpu_gen_hvmd_nodes(ldom_t *ldp, md_t *mdp, md_node_t *cpusp, md_node_t *guestp,
    bool_t include_pm_status)
{
	vcpu_t		*vcpu;
	md_node_t	*rootnode;
	md_node_t	**cpulistp;
	uint64_t	num_nodes, ncpus;

	rootnode = md_root_node(pri.core_mdp);
	num_nodes = md_node_count(pri.core_mdp);

	cpulistp = Xcalloc(num_nodes, md_node_t *);

	/*
	 * Starting at the root node, scan the "fwd" dag for
	 * all the cpus in this description.
	 */
	ncpus = md_scan_dag(pri.core_mdp, rootnode,
	    md_find_name(pri.core_mdp, "cpu"),
	    md_find_name(pri.core_mdp, "fwd"), cpulistp);

	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		pri_cpu_t *pcpu = vcpu->cpu;
		md_node_t *cpunodep;
		uint64_t fma_status, pm_status;
		int x;

		cpunodep = md_link_new_node(mdp, "cpu", cpusp, "fwd", "back");
		md_add_arc_property(mdp, guestp, "fwd", cpunodep);
		md_add_arc_property(mdp, cpunodep, "back", guestp);

		md_add_value_property(mdp, cpunodep, "pid", pcpu->phys_id);
		md_add_value_property(mdp, cpunodep, "vid", vcpu->vid);

		md_add_value_property(mdp, cpunodep, "resource_id",
		    pcpu->res_id->id);

		md_add_value_property(mdp, cpunodep, "gid", ldp->res_id->id);

		/* Find core_md node corresponding to strand */
		for (x = 0; x < ncpus; ++x) {
			uint64_t pid, status;

			pid = pri_cpu_phys_id(pri.core_mdp, cpulistp[x]);
			if (pid == pcpu->phys_id) {
				status = md_get_prop_val(pri.core_mdp,
				    cpulistp[x], CORE_MD_OP_STATUS,
				    &fma_status);
				if (status == 0)
					md_add_value_property(mdp, cpunodep,
					    "fma_status", fma_status);

				if (include_pm_status) {
					status = md_get_prop_val(pri.core_mdp,
					    cpulistp[x], CORE_MD_PM_STATUS,
					    &pm_status);
					if (status == 0)
						md_add_value_property(mdp,
						    cpunodep, "pm_status",
						    pm_status);
				}
				break;
			}
		}
		if (x == ncpus)
			warning("%s: no PM or FMA status available for domain "
			    "%s vcpu %d (pcpu %d)", __func__, ldp->namep,
			    vcpu->vid, pcpu->phys_id);

		/*
		 * Add partid property for proper IDMMU_PARTITION_ID
		 * register programming.
		 */
		md_add_value_property(mdp, cpunodep, "partid",
		    vcpu->partid->id);

		/*
		 * If the cpu has access to a vmau, create a link.
		 */
		if (vcpu->vmau != NULL) {
			md_add_arc_property(mdp, cpunodep, "fwd",
			    vcpu->vmau->temp.mau_node);
			md_add_arc_property(mdp, vcpu->vmau->temp.mau_node,
			    "back", cpunodep);

			/* set up cwq link if there is a cwq */
			if (vcpu->vmau->temp.cwq_node != NULL) {
				md_add_arc_property(mdp, cpunodep, "fwd",
				    vcpu->vmau->temp.cwq_node);
				md_add_arc_property(mdp,
				    vcpu->vmau->temp.cwq_node, "back",
				    cpunodep);
			}
		}
	}

	Xfree(cpulistp);

	return (true);
}

/*
 * Given a physical mau which is bound in a guest, stitch all the
 * vcpus together which are sharing the same mau.
 */
void
mau_link_vcpus(ldom_t *ldp, vmau_t *vmau)
{
	vcpu_t *vcpu;

	CHAIN_INIT(vmau->vcpus);

	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		if (vmau->mau == vcpu->cpu->maup) {
			vcpu->vmau = vmau;
			CHAIN_ADD(vmau->vcpus, vcpu, mau);
		}
	}
}

/*
 * Data structure used during cpu removal to check
 * how maus are shared by the cpus and enforce the
 * constraints
 */
typedef struct {
	vmau_t *vmau;
	int count;
} mau_cpu_t;

/*
 * Before we allow these cpus to be removed check whether the cpu is the last
 * cpu shared by the maus already bound to the guest.
 */
static bool_t
check_cpu_mau_sharing(ldom_t *ldp, vcpu_t **vcpus, int n_cpus)
{
	mau_cpu_t *mau_map;
	int num_bound_maus, n_maus;
	int i, ret;

	num_bound_maus = ldom_num_bound_maus(ldp);
	if (num_bound_maus == 0)
		return (true);

	mau_map = Xmalloc(num_bound_maus * sizeof (mau_cpu_t));

	/*
	 * For each cpu being removed, add the mau being shared
	 * to the list.
	 */
	for (n_maus = 0, i = 0; i < n_cpus; i++) {
		if (vcpus[i]->vmau != NULL) {
			int match_found = 0;
			for (int j = 0; j < n_maus; j++) {
				if (vcpus[i]->vmau == mau_map[j].vmau) {
					match_found = 1;
					mau_map[j].count++;
					break;
				}
			}
			if (!match_found) {
				mau_map[n_maus].vmau = vcpus[i]->vmau;
				mau_map[n_maus].count = 1;
				n_maus++;
			}
		}
	}

	/*
	 * For each bound mau in the newly created list, check if
	 * the count is greater than the number of cpus sharing it.
	 */
	ret = true;
	for (i = 0; i < n_maus; i++) {
		if (mau_map[i].count >= CHAIN_LENGTH(mau_map[i].vmau->vcpus)) {
			ret = false;
			break;
		}
	}

	Xfree(mau_map);
	return (ret);
}

bool_t
check_full_core_allocation(int num)
{
	pri_core_t	*core;
	pri_cpu_t	*cpu;
	int		nbound = 0;
	int		nfree = 0;

	DBG(zprintf("%s:\n", __func__));

	CHAIN_ITER(pri.core_list, core, link) {
		DBG(zprintf("  core%d: ", core->idx));

		nbound = 0;

		CHAIN_ITER(core->cpus, cpu, core_link) {
			if (CHAIN_LENGTH(cpu->bound.vcpus) != 0) {
				DBG(zprintf("%2d ", cpu->phys_id));
				nbound++;
			}
		}
		DBG(zprintf("\n"));

		if (nbound == 0)
			nfree += CHAIN_LENGTH(core->cpus);

		/* check if there are enough free strands */
		if (nfree >= num)
			return (true);
	}

	/* not enough free strands on full cores */
	return (false);
}

static bool_t
cpu_reserve(ldom_t *ldp, int pid, resid_t *partid)
{
	pri_cpu_t	*pcpu;
	vcpu_t		*vcpu;
	vcpu_t		*tmp1;
	vcpu_t		*tmp2;

	/* get the pcpu struct */
	if ((pcpu = pid_to_pcpu(pid)) == NULL) {
		DBG(zprintf("%s: invalid physical cpu id %d\n", __func__, pid));
		return (false);
	}

	/* check that the strand is available */
	if (CHAIN_LENGTH(pcpu->bound.vcpus) != 0) {
		DBG(zprintf("%s: strand %d is not free\n", __func__, pid));
		return (false);
	}

	/* create and fill in a dummy vcpu struct */
	vcpu = Xmalloc(sizeof (*vcpu));

	vcpu->vcpuid = NULL;
	vcpu->vid = NEW_VCPUID;
	vcpu->fraction = PRI_CPU_FRAC_DENOM;
	vcpu->partid = partid;
	vcpu->cpu = pcpu;
	vcpu->ldp = ldp;
	vcpu->vmau = NULL;
	vcpu->flag_isconfigd = false;
	vcpu->util_idx = 0;
	(void) memset(vcpu->util_table, 0, sizeof (vcpu->util_table));

	pcpu->bound.fraction_free -= vcpu->fraction;

	/* put the vcpu on the domain's reserved list */
	CHAIN_ADD_SORTED(ldp->mig_rsvd.vcpus, vcpu, tmp1, tmp2, ldom, pid_cmp);
	CHAIN_ADD(pcpu->bound.vcpus, vcpu, pcpu);

	return (true);
}

/* ARGSUSED2 */
bool_t
cpu_reserve_core(ldom_t *ldp, pri_core_t *core, response_t *resp)
{
	resid_t		*partid = NULL;
	pri_cpu_t	*cpu;
	vcpu_t		*vcpu;

	partid = get_mmu_partid(ldp, core, NEW_MMU_PARTID);

	/*
	 * Reserve all unbound strands on the core. Strands that are
	 * already reserved have also been marked as bound.
	 */
	CHAIN_ITER(core->cpus, cpu, core_link) {

		/* skip this strand if it is already bound or reserved */
		if (CHAIN_LENGTH(cpu->bound.vcpus) != 0) {

			/* verify all bindings are consistent */
			DBG_BLOCK {
				CHAIN_ITER(cpu->bound.vcpus, vcpu, pcpu)
					ASSERT(ldp == vcpu->ldp);
			}
			continue;
		}

		DBG(zprintf("reserving strand %d for LDom %s\n",
		    cpu->phys_id, ldp->namep));

		/* reserve this strand */
		if (!cpu_reserve(ldp, cpu->phys_id, partid)) {
			DBG(zprintf("reservation of strand %d failed\n",
			    cpu->phys_id));
			return (false);
		}
	}

	return (true);
}

void
cpu_release_reserved(ldom_t *ldp)
{
	vcpu_t	*vcpu;
	vcpu_t	*nvcpu;

	/*
	 * XXXMIG - Resetting the tick sync mode is not CPU specific,
	 * but rather migration specific. Maybe tease this apart to
	 * a more logical arrangement.
	 */
	ldp->flag.local_tick_sync = false;

	CHAIN_ITER_VOL(ldp->mig_rsvd.vcpus, vcpu, nvcpu, ldom) {
		cpu_unbind(vcpu, true);
	}
}

static vcpu_t *
cpu_bind_reserved(ldom_t *ldp, int vid)
{
	vcpu_t	*vcpu;
	vcpu_t	*tmp1;
	vcpu_t	*tmp2;
	resid_t	*vcpuid;

	/* pull a vcpu off the reserved list */
	vcpu = CHAIN_START(ldp->mig_rsvd.vcpus);
	CHAIN_UNLINK(ldp->mig_rsvd.vcpus, vcpu, ldom);

	ASSERT(vcpu->vid == NEW_VCPUID);

	/* allocate the required ID for this vcpu */
	if (vid == NEW_VCPUID)
		vcpuid = alloc_res(&ldp->guest_vcpu_ids);
	else
		vcpuid = pull_res_id(&ldp->guest_vcpu_ids, vid);

	if (vcpuid == NULL) {
		DBG(zprintf("Unable to obtain vcpu %d\n", vid));
		return (NULL);
	}

	vcpu->vcpuid = vcpuid;
	vcpu->vid = vcpuid->id;

	/* allocate the resource ID for the vcpu */
	if (vcpu->cpu->res_id == NULL) {
		vcpu->cpu->res_id = alloc_res(&zeus.free_vcpu_ids);
		if (vcpu->cpu->res_id == NULL)
			fatal("Out of VCPU resource IDs");
	}

	DBG(zprintf("%s: binding vcpu %d to strand %d for LDom %s\n",
	    __func__, vcpu->vid, vcpu->cpu->phys_id, ldp->namep));

	/* add the vcpu to the bound list */
	CHAIN_ADD_SORTED(ldp->bound.vcpus, vcpu, tmp1, tmp2, ldom, vid_cmp);

	return (vcpu);
}

/*
 * For a migration that requires the domain to be reduced to one vcpu
 * before transport, select that one vcpu to bind in the newly created
 * domain. By agreement with the source, this vcpu will be the lowest
 * numbered vcpu that has not been power managed. Power managed vcpus
 * will have the 'pm_status' prop in their MD nodes set to PM_MANAGED.
 */
static int
cpu_mig_select_bind_vcpu(md_t *mdp, md_node_t **cpus, int ncpus)
{
	int		i;
	int		ret;
	uint64_t	id;
	uint64_t	status;
	int		vcpuid = MAXINT;

	ASSERT(ncpus != 0);

	for (i = 0; i < ncpus; i++) {
		if (md_get_prop_val(mdp, cpus[i], "id", &id) != 0) {
			DBG(zprintf("missing id property in vcpu MD node\n"));
			return (-1);
		}

		/* get the PM status for this CPU */
		ret = md_get_prop_val(mdp, cpus[i], CORE_MD_PM_STATUS, &status);

		/* skip nodes that are being managed by PM */
		if ((ret == 0) && (status == PM_MANAGED))
			continue;

		if (id < vcpuid)
			vcpuid = (int)id;
	}

	/* there must be one non power managed vcpu */
	ASSERT(vcpuid != MAXINT);

	DBG(zprintf("%s: bind target (lowest non-pm vcpu): %d\n",
	    __func__, vcpuid));

	return (vcpuid);
}

bool_t
cpu_mig_bind(ldom_t *ldp, md_t *mdp, response_t *resp)
{
	int		i, j;
	uint64_t	id;
	int		ncpus;
	md_node_t	**nodes = NULL;
	int		*cpuids = NULL;
	int		bind_vcpu;
	bool_t		result = false;
	pri_core_t	*core;
	int		needed;
	bool_t		core_free;
	pri_cpu_t	*cpu;
	vcpu_t		*vcpup;
	pm_resid_t	resid;
	ldm_err_t	pm_result = { 0 };

	/*
	 * Scan the MD for all the vcpu nodes. From this, generate
	 * a list of vcpu ids that need to be created for this
	 * domain.
	 */
	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));
	ncpus = md_scan_dag(mdp, NULL, md_find_name(mdp, "cpu"),
	    md_find_name(mdp, "fwd"), nodes);

	if (ncpus == 0) {
		DBG(zprintf("No vcpus found in MD for LDom %s\n", ldp->namep));
		rprintf(resp, gettext("Invalid CPU constraints for "
		    "LDom %s\n"), ldp->namep);
		goto done;
	}

	cpuids = Xmalloc(ncpus * sizeof (uint64_t));

	for (i = 0, j = 0; i < ncpus; i++) {
		if (md_get_prop_val(mdp, nodes[i], "id", &id) != 0) {
			DBG(zprintf("missing id property in vcpu MD node\n"));
			rprintf(resp, gettext("Invalid CPU constraints for "
			    "LDom %s\n"), ldp->namep);
			goto done;
		}

		cpuids[j++] = (int)id;
	}

	/* sort the vcpu array in ascending order */
	sort_int_array(cpuids, ncpus);

	/*
	 * A migration must reserve full cores for the target domain.
	 * Scan the list of cores, looking for ones that have no vcpus
	 * bound to any of the strands. Reserve the core by putting
	 * all strands on the reserved vcpu list of the domain.
	 */
	needed = ncpus;
	i = 0;
	CHAIN_ITER(pri.core_list, core, link) {
		core_free = true;

		CHAIN_ITER(core->cpus, cpu, core_link) {
			if (CHAIN_LENGTH(cpu->bound.vcpus) != 0) {
				core_free = false;
				break;
			}
		}

		/* skip if core is not free */
		if (!core_free)
			continue;

		/* reserve all the strands on the core */
		if (!cpu_reserve_core(ldp, core, resp)) {
			DBG(zprintf("failed to reserve core %d\n", core->idx));
			cpu_release_reserved(ldp);
			goto done;
		}

		needed -= CHAIN_LENGTH(core->cpus);

		/* check if we have enough cores */
		if (needed <= 0)
			break;
	}

	if (needed > 0) {
		rprintf(resp, gettext("Not enough free cores "
		    "available for migration\n"));
		cpu_release_reserved(ldp);
		goto done;
	}

	/* select the vcpu to bind */
	bind_vcpu = cpu_mig_select_bind_vcpu(mdp, nodes, ncpus);

	ASSERT(bind_vcpu >= 0);

	if ((vcpup = cpu_bind_reserved(ldp, bind_vcpu)) == NULL) {
		DBG(zprintf("Unable to bind requested vcpu %d\n", bind_vcpu));
		goto done;
	}

	result = true;

	if (pm_cpu_events_on) {
		ASSERT(vcpup != NULL);

		/* fill in the resource structure for PM */
		resid.pi_res_type = PM_RES_CPU;
		resid.pi_res_name = CPU_RES_NAME;
		resid.u.cpu.pi_id_val = vcpup->cpu->phys_id;

		send_pm_cpu_event(ldp, ADD_RESOURCE_PRE, &resid, 1, &pm_result);

		if (pm_result.ldme_errs != LDME_OP_EOK) {
			rprintf(resp, gettext("Power management failure while "
			    "preparing CPUs for migrated domain\n"));
			result = false;
		}
		pm_free_errors(&pm_result);
	}

done:
	if (nodes != NULL)
		Xfree(nodes);

	if (cpuids != NULL)
		Xfree(cpuids);

	return (result);
}

/*
 * Create a list of core structures that are in use by a given
 * domain. Memory for the array must be deallocated by the caller.
 */
int
get_cores_in_use(ldom_t *ldp, pri_core_t ***cores)
{
	int		i;
	vcpu_t		*vcpu;
	pri_core_t	*core;
	int		ncores;

	/* allocate some space to store the info */
	*cores = Xmalloc(sizeof (pri_core_t *) *
	    (CHAIN_LENGTH(ldp->bound.vcpus) +
	    CHAIN_LENGTH(ldp->mig_rsvd.vcpus)));

	ncores = 0;

	/* add all the cores for bound vcpus */
	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		core = vcpu->cpu->core;

		/* check if the core is already recorded as in use */
		for (i = 0; i < ncores; i++) {
			if (core->idx == (*cores)[i]->idx)
				break;
		}

		if (i == ncores) {
			/* record this new core as in use */
			(*cores)[ncores++] = core;
		}
	}

	/* add all the cores for reserved strands */
	CHAIN_ITER(ldp->mig_rsvd.vcpus, vcpu, ldom) {
		core = vcpu->cpu->core;

		/* check if the core is already recorded as in use */
		for (i = 0; i < ncores; i++) {
			if (core->idx == (*cores)[i]->idx)
				break;
		}

		if (i == ncores) {
			/* record this new core as in use */
			(*cores)[ncores++] = core;
		}
	}

	return (ncores);
}

/*
 * Function called by the migration code which shrinks a domain down to
 * a single vcpu so that it can be suspended.
 *
 * Return Value:
 * 	0		- Success
 *	non-zero	- Failure
 */
int
cpu_mig_reduce_to_one(ldom_t *ldp, response_t *resp)
{
	int		i;
	vcpu_t		**vcpus;
	pri_cpu_t	**pcpus;
	pri_core_t	**cores;
	int		ncores = 0;
	int		num;
	int		num_bound;
	int		rc;
	int		add_count;

	ASSERT(ldp->state == ACTIVE);
	ASSERT(in_delayed_reconfig(NULL) == false);

	num_bound = ldom_num_bound_vcpus(ldp);

	if (num_bound == 1)
		return (0);

	if (zeus.simulating)
		return (0);

	if ((rc = check_cpu_dr_capable(ldp)) != 0)
		return (rc);

	num = num_bound - 1;

	/*
	 * If the domain is using its own copy of tick, it is necessary
	 * to keep track of all the cores in use and make sure that they
	 * are all reserved until the domain is destroyed.
	 */
	if (ldp->flag.local_tick_sync)
		ncores = get_cores_in_use(ldp, &cores);

	vcpus = Xmalloc(num * sizeof (vcpu_t *));
	pcpus = Xmalloc(num * sizeof (pri_cpu_t *));
	cpu_adm_t cpu_adm = { .type = CPU_ADM_NUM, .num = num };

	if ((rc = get_vcpus_to_delete(ldp, &cpu_adm, vcpus, resp)) != 0)
		FATAL_UNEXPECTED_ERROR(rc);

	/*
	 * Before we DR out the cpus we need to save the list of physical
	 * cpus backing the vcpus currently bound by this domain because
	 * we may need to add them back in later and the 'vcpus' array is
	 * freed in cpu_unbind() which is called from dr_rm_cpus()
	 */
	for (i = 0; i < cpu_adm.num; i++) {
		ASSERT(vcpus[i]->cpu != NULL);
		pcpus[i] = vcpus[i]->cpu;
		DBG(zprintf("%s (%d) vid %d pid %d\n",
		    __func__, i, vcpus[i]->vid, pcpus[i]->phys_id));
	}

	rc = dr_rm_cpus(ldp, vcpus, &num, false, resp);
	Xfree(vcpus);

	/*
	 * A partial success is of no use as a migration requires that we
	 * have only one vcpu in the domain prior to migrating it.
	 * We attempt to add back in all the vcpus that were removed
	 */
	if (rc == ERR_CPU_DR_PARTIAL) {
		cpu_adm_t *cpu_adm_add =
		    Xmalloc(sizeof (cpu_adm_t) + num * sizeof (int));

		/*
		 * Loop through the array of vcpus originally marked for
		 * deletion and check for the ones which were actually
		 * removed. Add these to the array of ids which will be
		 * passed to cpu_add() [so that we have the same physical
		 * cpus backing the vcpus as before the migration]
		 *
		 * The vcpus were selected for removal in get_vcpus_to_delete()
		 * in the order of highest to lowest vcpuid, so we need to add
		 * them back in the same order so that each vcpu is backed by
		 * the same physical strand as before.
		 */
		add_count = 0;
		for (i = cpu_adm.num - 1; i >= 0; i--) {
			if (pcpus[i]->bound.fraction_free) {
				cpu_adm_add->ids[add_count] =
				    pcpus[i]->phys_id;
				add_count++;
			}
		}
		ASSERT(num < num_bound - 1);
		ASSERT(add_count == num);

		cpu_adm_add->type = CPU_ADM_IDS;
		cpu_adm_add->num = add_count;

		if (cpu_add(ldp, cpu_adm_add, resp) != true)
			rprintf(resp, MSG_ERR_MIG_CPU_REDUCED);

		/*
		 * The migration functions to add and remove CPUs during
		 * migration do not modify the number of CPUs in the constraint
		 * (because they do not call cpu_add()/cpu_del() directly).
		 * However while handling the failure case above, calling
		 * cpu_add() to restore the CPUs does increase the number of
		 * CPUs in the constraint and we need to account for that here
		 * and reduce the constraint by the number of CPUs added in
		 * cpu_add().
		 *
		 * cpu_add() adds the increased amount to the constraint
		 * (ldp->required.num_vcpus) and dr_add_cpus() reduces the
		 * amount if seq_add_vcpus() returns CPU_DR_OP_FAILED or
		 * the amount that couldn't be allocated. Since seq_add_vcpus()
		 * returns the number of cpus added in case of both full or
		 * partial success, the amount is still added to to the
		 * constraint for the partial CPU DR case.
		 */
		ldp->required.num_vcpus -= add_count;

		Xfree(cpu_adm_add);
	}

	/*
	 * If the domain is using its own copy of tick, reserve any
	 * cores that were in use by the domain at the start of the
	 * operation.
	 */
	if (ldp->flag.local_tick_sync) {
		for (i = 0; i < ncores; i++) {
			/* reserve all the strands on the core */
			if (!cpu_reserve_core(ldp, cores[i], resp)) {
				DBG(zprintf("failed to reserve core %d\n",
				    cores[i]->idx));
			}
		}
	}

	Xfree(pcpus);

	return (rc);
}

bool_t
cpu_mig_add_remaining(ldom_t *ldp, int num_req, response_t *resp)
{
	int		count = 0;
	int		num_needed;
	vcpu_t		*vcpup;
	pm_resid_t	*resids = NULL;
	ldm_err_t	pm_result = { 0 };

	ASSERT(num_req >= 1);

	num_needed = (num_req - 1);

	if (num_req == ldom_num_bound_vcpus(ldp)) {
		/* nothing to do */
		return (true);
	}

	if (check_cpu_dr_capable(ldp) != 0) {
		DBG(zprintf("%s: CPU DR domain service not available\n",
		    __func__));
		return (false);
	}

	DBG(zprintf("%s: need to increase bound vcpus from %d to %d\n",
	    __func__, ldom_num_bound_vcpus(ldp), num_req));

	/* there should already be reserved strands */
	ASSERT(num_needed <= CHAIN_LENGTH(ldp->mig_rsvd.vcpus));

	/* allocate some space for the PM event data */
	if (pm_cpu_events_on)
		resids = Xmalloc(sizeof (pm_resid_t) * num_needed);

	for (count = 0; count < num_needed; count++) {

		/* bind a reserved strand to any available vcpu */
		if ((vcpup = cpu_bind_reserved(ldp, NEW_VCPUID)) == NULL) {
			DBG(zprintf("%s: Unable to bind vcpu\n", __func__));
			Xfree(resids);
			return (false);
		}

		/* fill in information for the PM event */
		if (pm_cpu_events_on) {
			resids[count].pi_res_type = PM_RES_CPU;
			resids[count].pi_res_name = CPU_RES_NAME;
			resids[count].u.cpu.pi_id_val = vcpup->cpu->phys_id;
		}
	}

	ASSERT(count == num_needed);

	/* send the PM event to ensure strands have power */
	if (pm_cpu_events_on) {
		send_pm_cpu_event(ldp, ADD_RESOURCE_PRE, resids,
		    num_needed, &pm_result);

		Xfree(resids);

		if (pm_result.ldme_errs != LDME_OP_EOK) {
			rprintf(resp, gettext("Power management failure while "
			    "preparing CPUs for migrated domain\n"));
			pm_free_errors(&pm_result);
			return (false);
		}
		pm_free_errors(&pm_result);
	}

	/* add the CPUs to the domain */
	if (dr_add_cpus(ldp, num_needed, resp) != num_needed)
		return (false);

	return (true);
}

bool_t
cpu_mig_restore(ldom_t *ldp, int num_req, response_t *resp)
{
	int		count;
	int		num_needed;
	cpu_adm_t	cpu_adm;
	int		rv;

	/* one vcpu is already bound */
	num_needed = num_req - 1;

	if (num_needed == 0)
		return (true);

	DBG(zprintf("%s: need to increase bound vcpus from %d to %d\n",
	    __func__, ldom_num_bound_vcpus(ldp), num_req));

	if ((rv = check_cpu_dr_capable(ldp)) != 0) {
		DBG(zprintf("%s: CPU DR domain service not available (%d)\n",
		    __func__, rv));
		return (false);
	}

	/*
	 * If the guest is using its own value of tick, all CPUs must
	 * be bound from the pool of reserved strands. These already have
	 * their tick synchronized. Otherwise, it does not matter which
	 * strands are selected.
	 */
	if (ldp->flag.local_tick_sync) {
		DBG(zprintf("binding from reserved strands...\n"));

		/* there should already be reserved strands */
		ASSERT(num_needed <= CHAIN_LENGTH(ldp->mig_rsvd.vcpus));

		for (count = 0; count < num_needed; count++) {

			/* bind a reserved strand to any available vcpu */
			if (cpu_bind_reserved(ldp, NEW_VCPUID) == NULL) {
				DBG(zprintf("%s: Unable to bind vcpu\n",
				    __func__));
				return (false);
			}
		}
	} else {
		DBG(zprintf("binding from any available strands...\n"));

		cpu_adm.type = CPU_ADM_NUM;
		cpu_adm.num = num_needed;

		if (!cons_bind_vcpus(ldp, &cpu_adm, resp)) {
			DBG(zprintf("%s: unable to restore original CPU "
			    "configuration\n", __func__));
			return (false);
		}
	}

	if (dr_add_cpus(ldp, num_needed, resp) != num_needed)
		return (false);

	return (true);
}

#if ZDBG
/*
 * Scan through an MD looking for a CPU node that corresponds to
 * the vcpu id passed in. Currently this is only used for debug
 * sanity checks since this method of doing a full scan for a
 * single vcpu id is rarely the most efficient way to do things.
 */
static md_node_t *
lookup_vcpu_in_md(md_t *mdp, uint64_t vid)
{
	int		idx;
	md_node_t	*rv;
	md_node_t	**vcpus;
	int		nvcpus;
	uint64_t	id;

	/*
	 * Lookup all cpu nodes in the MD.
	 */
	vcpus = Xmalloc(md_node_count(mdp) * sizeof (md_node_t *));
	nvcpus = md_scan_dag(mdp, NULL, md_find_name(mdp, "cpu"),
	    md_find_name(mdp, "fwd"), vcpus);

	/* search all vcpu nodes for the one of interest */
	for (idx = 0; idx < nvcpus; idx++) {
		md_get_prop_val(mdp, vcpus[idx], "id", &id);

		/* check for a match */
		if (id == vid)
			break;
	}

	rv = (idx == nvcpus) ? NULL : vcpus[idx];

	Xfree(vcpus);

	return (rv);
}
#endif /* ZDBG */

/*
 * The guest MD is used as the source of binding information for a
 * migration. However, if a CPU has been power managed, it will be
 * removed from the guest MD. This function adds back any CPUs that
 * have been power managed to the guest MD. This ensures that the
 * target domain ends up with the same number of CPUs as the source.
 */
bool_t
cpu_mig_add_pm_bindings(ldom_t *ldp, md_t *mdp)
{
	vcpu_t		*vcpup;
	uint64_t	pid;
	md_node_t	**nodes;
	int		nnodes;
	int		status;
	md_node_t	*cpus_node;
	md_node_t	*nodep;

	/* sequencer protects against PM updates */
	ASSERT_IN_SEQUENCER();

	DBG(zprintf("%s\n", __func__));

	/*
	 * Scan the MD, looking for the 'cpus' node. This will
	 * be the parent for any power managed CPU nodes that
	 * get added.
	 */
	nodes = Xcalloc(md_node_count(mdp), md_node_t *);
	nnodes = md_scan_dag(mdp, NULL, md_find_name(mdp, "cpus"),
	    md_find_name(mdp, "fwd"), nodes);

	/* should only be one 'cpus' node */
	ASSERT(nnodes == 1);

	cpus_node = nodes[0];

	/*
	 * Walk all bound vcpus checking if each has been power
	 * managed. If so, it will not be present in the guest MD
	 * and a placeholder CPU node should be added to reflect
	 * the true CPU bindings to the target.
	 */
	CHAIN_ITER(ldp->bound.vcpus, vcpup, ldom) {
		pid = vcpup->cpu->phys_id;

		status = get_core_md_cpu_status(pid, CORE_MD_PM_STATUS);
		if (status != PM_MANAGED)
			continue;

		/*
		 * A paranoid check to make sure that the vcpu
		 * is not already present in the MD.
		 */
		DBG(ASSERT(lookup_vcpu_in_md(mdp, vcpup->vid) == NULL));

		DBG(zprintf("  adding powered off vcpu %d to bindings MD\n",
		    vcpup->vid));

		/*
		 * Create the new cpu node and link it off the 'cpus'
		 * node. For the purposes of migration, only a couple
		 * pieces of information are required. The 'id' and
		 * the PM status, used to quickly identify CPUs being
		 * power managed.
		 */
		nodep = md_link_new_node(mdp, "cpu", cpus_node, "fwd", "back");
		md_add_value_property(mdp, nodep, "id", vcpup->vid);
		md_add_value_property(mdp, nodep, "pm_status", status);
	}

	return (true);
}
