/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Domain manager database implementation.
 *
 * The database is stored and operated on as an XML document.  When the
 * database is first loaded from persistent storage it is checked for validity
 * against the database schema (see database_schema.c) and if invalid it is
 * discarded. When doing database updates the implementation assumes that the
 * database is valid and updates maintain this invariant. This simplifies many
 * operations, as the existence and order of many elements can be assumed from
 * the schema definition. This also means however that care must be taken if
 * the schema is to be modified.
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>

#include "allocate.h"
#include "basics.h"
#include "variables.h"
#include "fatal.h"
#include "ldom.h"
#include "option.h"
#include "zprintf.h"
#include "xml_common.h"
#include "support.h"
#include "vio.h"
#include "database_internal.h"
#include "master.h"
#include "pcie.h"

#define	XML_VERSION	((xmlChar *)"1.0") /* version of XML used */

static char *filename = NULL;		/* database filename */
static char *tmpfilename = NULL;	/* temporary database filename */
static xmlDocPtr dbase = NULL;		/* current database */
static xmlDocPtr old_dbase = NULL;	/* old database */

/*
 * Supported database versions:
 *
 * Major versions-
 * 0: N/A
 * 1: Pre-1.0 release
 * 2: LDoms 1.0
 * 3: LDoms 1.1
 *
 * N.B.  When/if there is a major version 4, the change will also
 * delete the OPTIONS_NODE support introduced in version 3.0.
 * See the comment regarding OPTIONS_NODE in database_internal.h.
 */

/* Versions this code supports */
static prot_version_t dbase_versions[] =
{
	{3, 9},		/* adds SDIO PCIEDEV related nodes */
	{3, 8},		/* adds universally unique id (UUID) to each ldom */
	{3, 7},		/* adds DRM policies and stats */
	{3, 6},		/* adds id in vdisk, vsw, & vnet nodes */
	{3, 5},		/* adds linkprop in network nodes */
	{3, 4},		/* adds failure-policy and master */
	{3, 3},		/* adds mtu in network nodes (vsw & vnet) */
	{3, 2},		/* adds HOSTID_NODE in ldom, hybrid IO, mpgroup */
	{3, 1},		/* VLAN support, pvid and vid nodes */
	{3, 0},		/* no instance nums; adds OPTIONS_NODE VSW_MODE_NODE */
	{2, 3},		/* adds TIMEOUT_NODE in vdisk */
	{2, 2},		/* adds IO_NODE and MAC_ADDRESS_NODE in ldom */
	{2, 1},		/* adds VOLUME_OPTS_NODE and VCONS_PORT_NODE */
	{2, 0},		/* LDoms 1.0 */
};

/* Version we produce */
#define	LDOM_DATABASE_VERSION	((xmlChar *)"3.9")

/* Version of existing database.  Updated if a valid one is found */
static uint16_t db_major = 3;
static uint16_t db_minor = 9;

/*
 * Returns the database xmlDocPtr for external files wishing to parse the
 * document.
 */
xmlDocPtr
get_dbase(void)
{
	return (dbase);
}

/*
 * Get the major & minor numbers from the DB.  If non-NULL ver passed in,
 * caller is responsible for doing the xmlFree().
 */
static bool_t
get_dbase_version(xmlDocPtr db, xmlChar **ver, uint16_t *major, uint16_t *minor)
{
	xmlChar *version;
	xmlNodePtr root;

	root = xmlDocGetRootElement(db);

	version = xmlGetProp(root, VERSION_ATTR);

	if (version == NULL) {
		warning("Missing \"%s\" attribute in the database",
		    (char *)VERSION_ATTR);
		return (false);
	}

	if (sscanf((char *)version, "%hu.%hu", major, minor) != 2) {
		warning("Invalid \"%s\" version attribute \"%s\" in the "
		    "database", (char *)VERSION_ATTR, (char *)version);
		xmlFree(version);
		return (false);
	}

	if (ver != NULL)
		*ver = version;
	else
		xmlFree(version);

	return (true);
}

/*
 * Search for an child node under 'parent' node of a given
 * type and return it if it's property prop_name matches string
 * 'prop_value', otherwise return NULL. This can be used to
 * find a particular instance of a server (e.g. a vds_instance)
 * node, or a particular vds volume.
 */
static xmlNodePtr
find_node_by_prop_val(xmlNodePtr parent, const xmlChar *prop_name,
    const xmlChar *prop_value)
{
	xmlNodePtr node, propNode;
	xmlChar *content;
	bool_t found = false;

	/*
	 * For each sibling of the parent node, find the node containing
	 * the property and check if the value matches.
	 */
	for (node = parent->xmlChildrenNode; node != NULL; node = node->next) {
		if ((propNode = xml_find_subnode(node, prop_name)) != NULL) {
			content = xmlNodeGetContent(propNode);
			if (xmlStrcmp(content, prop_value) == 0)
				found = true;
			xmlFree(content);

			if (found == true)
				return (node);
		}
	}

	return (NULL);
}

/*
 * Find and return a node under node 'ldom' which has type 'node_name' and
 * property named 'prop_name' with a string value 'prop_val'. Returns NULL
 * if no such node exists.  This function can be used to search for a
 * specific resource type within an LDom. E.g. find_ldom_node_by_prop(ldom,
 * "vds", "service_name", "myvds") will search an instance named "myvds"
 * of a virtual disk service.
 */
static xmlNodePtr
find_node_by_prop(xmlNodePtr parent, const xmlChar *node_name,
    xmlChar *prop_name, xmlChar *prop_val)
{
	xmlNodePtr node;

	if ((node = xml_find_subnode(parent, node_name)) != NULL)
		return (find_node_by_prop_val(node, prop_name, prop_val));

	return (NULL);
}

/*
 * Find and return a node under node 'ldom' which has name 'name' and instance
 * 'instance'. Returns NULL if no such node exists.  This function can be used
 * to search for a specific resource instance within an LDom.
 * E.g. find_ldom_node_instance(ldom, "vds", "3") will search for instance 3
 * of a virtual disk service.
 */
xmlNodePtr
find_ldom_node_instance(xmlNodePtr ldom, const xmlChar *name, int instance)
{
	char instance_str[INT_STR_LEN];

	snprintf(instance_str, sizeof (instance_str), "%d", instance);

	return (find_node_by_prop(ldom, name, INSTANCE_NODE,
	    (xmlChar *)instance_str));
}

/*
 * This routine finds the property node under a given instance node
 * e.g. <ldom, "vds", 1, "service_name"> -or- <ldom, "vds", 1,
 * "vds_volumes">
 */
static xmlNodePtr
find_prop_node_by_instance(xmlNodePtr ldom, const char *node_type,
    int instance, const xmlChar *prop_name)
{
	xmlNodePtr node;

	node = find_ldom_node_instance(ldom, (xmlChar *)node_type, instance);
	if (node == NULL)
		return (NULL);

	return (xml_find_subnode(node, prop_name));
}

/*
 * This routine finds the property node under a given node
 * with matching property value
 */
static xmlNodePtr
find_prop_node_by_prop_val(xmlNodePtr ldom, xmlChar *node_type,
    xmlChar *match_prop, xmlChar *match_val, xmlChar *prop_name)
{
	xmlNodePtr node;

	node = find_node_by_prop(ldom, node_type, match_prop, match_val);
	if (node == NULL)
		return (NULL);

	return (xml_find_subnode(node, prop_name));
}

/*
 * Returns the name of the LDom represented in 'ldom_node'.
 */
static xmlChar *
ldom_get_name(xmlNodePtr ldom_node)
{
	return (xmlNodeGetContent(ldom_node->xmlChildrenNode));
}

/*
 * Returns true iff the LDom represented in 'ldom_node' has name 'name'.
 */
static bool_t
ldom_is_named(xmlNodePtr ldom_node, const xmlChar *name)
{
	xmlChar *content;
	bool_t is_named = false;

	content = ldom_get_name(ldom_node);
	if (xmlStrcmp(content, name) == 0)
		is_named = true;
	xmlFree(content);

	return (is_named);
}

/*
 * Set up a link from LDom 'ldp' to its representation 'ldom_node' in the
 * XML database.
 */
static void
ldom_link(ldom_t *ldp, xmlNodePtr ldom_node)
{
	struct ldom_database *dbp;

	dbp = Xmalloc(sizeof (struct ldom_database));

	dbp->ldom_node = ldom_node;
	ldp->dbp = dbp;
}

/*
 * Create a new node in the XML database representing LDom 'ldp'.
 */
void
database_ldom_create(ldom_t *ldp)
{
	xmlNodePtr root, ldom_node;

	assert(dbase != NULL);

	root = xmlDocGetRootElement(dbase);

	ldom_node = xmlNewChild(root, NULL, LDOM_NODE, NULL);
	xmlNewChild(ldom_node, NULL, LDOM_NAME_NODE, (xmlChar *)ldp->namep);
	xmlNewChild(ldom_node, NULL, CPU_NODE, NULL);
	xmlNewChild(ldom_node, NULL, MAU_NODE, NULL);
	xmlNewChild(ldom_node, NULL, MEMORY_NODE, NULL);
	xmlNewChild(ldom_node, NULL, IO_NODE, NULL);
	xmlNewChild(ldom_node, NULL, MAC_ADDRESS_NODE, NULL);
	xmlNewChild(ldom_node, NULL, VSW_NODE, NULL);
	xmlNewChild(ldom_node, NULL, VNET_NODE, NULL);
	xmlNewChild(ldom_node, NULL, VDS_NODE, NULL);
	xmlNewChild(ldom_node, NULL, DISK_NODE, NULL);
	xmlNewChild(ldom_node, NULL, VCC_NODE, NULL);
	xmlNewChild(ldom_node, NULL, CONSOLE_NODE, NULL);
	xmlNewChild(ldom_node, NULL, VDPCS_NODE, NULL);
	xmlNewChild(ldom_node, NULL, VDPCC_NODE, NULL);
	xmlNewChild(ldom_node, NULL, VARIABLES_NODE, NULL);
	if (db_major == 3)
		/* see comments in database_internal.h for an explanation */
		xmlNewChild(ldom_node, NULL, OPTIONS_NODE, NULL);
	xmlNewChild(ldom_node, NULL, HOSTID_NODE, NULL);
	xmlNewChild(ldom_node, NULL, FAILURE_POLICY_NODE, NULL);
	xmlNewChild(ldom_node, NULL, MASTER_NODE, NULL);
	xmlNewChild(ldom_node, NULL, POLICY_NODE, NULL);
	xmlNewChild(ldom_node, NULL, UUID_NODE, NULL);
	xmlNewChild(ldom_node, NULL, PCIEDEV_NODE, NULL);

	ldom_link(ldp, ldom_node);
}

/*
 * Returns the node in the XML database which represents the LDom
 * named 'name'.
 */
xmlNodePtr
find_ldom(xmlDocPtr db, const xmlChar *name)
{
	xmlNodePtr root, node;

	root = xmlDocGetRootElement(db);
	node = root->xmlChildrenNode;
	while (node != NULL) {
		if (xmlStrcmp(node->name, LDOM_NODE) == 0)
			if (ldom_is_named(node, name))
				break;
		node = node->next;
	}

	return (node);
}

/*
 * Add the CPU requirements in database node 'cpu_node' to the LDom 'ldp'.
 */
static void
from_db_add_cpu(xmlNodePtr cpu_node, ldom_t *ldp)
{
	if (cpu_node->xmlChildrenNode == NULL)
		return;

	ldp->required.num_vcpus =
	    xml_get_long_content(cpu_node->xmlChildrenNode);

	database_ldom_set_cpus(ldp, ldp->required.num_vcpus);
}

/*
 * Add the MAU requirements in database node 'mau_node' to the LDom 'ldp'.
 */
static void
from_db_add_mau(xmlNodePtr mau_node, ldom_t *ldp)
{
	if (mau_node->xmlChildrenNode == NULL)
		return;

	ldp->required.num_vmaus =
	    xml_get_long_content(mau_node->xmlChildrenNode);

	database_ldom_set_maus(ldp, ldp->required.num_vmaus);
}

/*
 * Add the memory requirements in database node 'mem_node' to the LDom 'ldp'.
 */
static void
from_db_add_memory(xmlNodePtr mem_node, ldom_t *ldp)
{
	if (mem_node->xmlChildrenNode == NULL)
		return;

	ldp->required.size_mem =
	    xml_get_uint64_content(mem_node->xmlChildrenNode);

	database_ldom_set_memory(ldp, ldp->required.size_mem);
}

/*
 * Add the MAC address in database node 'macaddr_node' to the LDom 'ldp'.
 */
static void
from_db_add_mac_addr(xmlNodePtr macaddr_node, ldom_t *ldp)
{
	uint64_t mac_addr;

	mac_addr = xml_get_uint64_content(macaddr_node);

	ldp->props.mac_address = mac_addr;

	database_ldom_set_mac_addr(ldp, mac_addr);
}

/*
 * Add 'hostid' node to the LDom 'ldp'.
 */
static void
from_db_add_hostid(xmlNodePtr hostid_node, ldom_t *ldp)
{
	uint64_t hostid;

	hostid = xml_get_uint64_content(hostid_node);

	ldp->props.hostid = hostid;

	database_ldom_set_hostid(ldp, hostid);
}

/*
 * Add 'failure-policy' node to the LDOM 'ldp'
 */
static void
from_db_add_failure_policy(xmlNodePtr failure_policy_node, ldom_t *ldp)
{
	uint64_t failure_policy = xml_get_uint64_content(failure_policy_node);

	ldp->props.failure_policy = failure_policy;
	database_ldom_set_failure_policy(ldp, failure_policy);
}

/*
 * Add 'master' node to the LDOM 'ldp'
 */
static void
from_db_add_master(xmlNodePtr master_node, ldom_t *ldp)
{
	char		*field;
	char		*master_list;
	ldom_t		*master_ldp;
	master_t	*master;
	int		i = 0;

	if (master_node == NULL || ldp == NULL)
		return;

	master_list = (char *)xmlNodeGetContent(master_node);
	if ((field = strtok(master_list, ",")) != NULL)
		free_masters(ldp);

	for (i = 0; i < pri.max_guest_dependencies && field != NULL; ++i) {
		if ((master_ldp = find_ldom_by_name(field)) != NULL) {
			master = Xcalloc(1, master_t);
			master->ldp = master_ldp;
			CHAIN_ADD(ldp->props.masters, master, link);
		}
		field = strtok(NULL, ",");
	}

	database_ldom_set_master(ldp);
}

/*
 * Add 'uuid' node to the LDOM 'ldp'
 */
static void
from_db_add_uuid(xmlNodePtr uuid_node, ldom_t *ldp)
{
	char *uuid = NULL;

	uuid = (char *)xmlNodeGetContent(uuid_node);
	if (uuid == NULL)
		return;

	strncpy(ldp->props.uuid, uuid, sizeof (ldp->props.uuid));
	xmlFree(uuid);
	database_ldom_set_uuid(ldp);
	DBG(zprintf("%s: LDom %s restored uuid=%s from db\n",
	    __func__, ldp->namep, ldp->props.uuid));
}

/*
 * Add the variables defined in database node 'variables_node' to the LDom
 * 'ldp'.
 */
static void
from_db_add_variables(xmlNodePtr variables_node, ldom_t *ldp)
{
	xmlNodePtr var_node, node;
	xmlChar *name, *value;

	var_node = variables_node->xmlChildrenNode;
	while (var_node != NULL) {
		node = var_node->xmlChildrenNode;
		name = xmlNodeGetContent(node);
		node = node->next;
		value = xmlNodeGetContent(node);

		add_ldom_variable(ldp, (char *)name, (char *)value);
		database_ldom_add_variable(ldp, (char *)name, (char *)value);

		xmlFree(name);
		xmlFree(value);

		var_node = var_node->next;
	}
}

/*
 * Add the disk client requirement in database node 'disk_node' to the LDom
 * 'ldp'.
 */
static void
from_db_add_disk(xmlNodePtr disk_node, ldom_t *ldp)
{
	xmlNodePtr node;
	xmlChar *vdisk_name;
	xmlChar *service_name;
	xmlChar *vol_name;
	int timeout = VDISK_DEFAULT_TIMEOUT;
	int id = NO_NEW_ID;

	node = disk_node->xmlChildrenNode;
	if (db_major < 3)
		node = node->next;	/* skip instance */
	vdisk_name = xmlNodeGetContent(node);
	node = node->next;
	service_name = xmlNodeGetContent(node);
	node = node->next;
	vol_name = xmlNodeGetContent(node);
	if ((db_major == 2 && db_minor >= 3) || (db_major > 2)) {
		node = node->next;
		timeout = xml_get_int_content(node);
	}

	if ((db_major == 3 && db_minor >= 6) || (db_major > 3)) {
		node = node->next;
		id = xml_get_int_content(node);
	}

	if (vdisk_new(ldp, (char *)vdisk_name, (char *)service_name,
	    (char *)vol_name, timeout, id) != NULL)
		database_ldom_add_disk_client(ldp, (char *)vdisk_name,
		    (char *)service_name, (char *)vol_name, timeout, id);

	xmlFree(vdisk_name);
	xmlFree(service_name);
	xmlFree(vol_name);
}

/*
 * Add a vds volume from a vds_volume node.
 */
static void
from_db_add_volume_to_vds(xmlNodePtr volume_node, vds_t *vds)
{
	xmlNodePtr node;
	xmlChar *blockdev = NULL;
	xmlChar *vol_name = NULL;
	xmlChar *options = NULL;
	xmlChar *mpgcontent = NULL;

	node = volume_node->xmlChildrenNode;
	if (db_major < 3)
		node = node->next;	/* skip minor */

	blockdev = xmlNodeGetContent(node);
	node = node->next;
	vol_name = xmlNodeGetContent(node);

	if (db_major > 2 || db_minor >= 1) {
		node = node->next;
		options = xmlNodeGetContent(node);

		if (options && IS_EMPTY_STR((char *)options)) {
			xmlFree(options);
			options = NULL;
		}
	}

	if (((db_major == 3) && (db_minor >= 2)) || (db_major > 3)) {
		node = node->next;
		mpgcontent = xmlNodeGetContent(node);

		if (mpgcontent && IS_EMPTY_STR((char *)mpgcontent)) {
			xmlFree(mpgcontent);
			mpgcontent = NULL;
		}
	}

	/*
	 * Add the volume only if it's not already in Zeus inventory.
	 */
	if (vds_find_volume_by_name(vds, (char *)vol_name) == NULL) {
		vds_add_volume(vds, (char *)vol_name, (char *)blockdev,
		    (char *)options, (char *)mpgcontent);
		database_ldom_add_disk_service_vol(vds->ldp, vds->service,
		    (char *)vol_name, (char *)blockdev, (char *)options,
		    (char *)mpgcontent);
	}

	xmlFree(blockdev);
	xmlFree(vol_name);
	xmlFree(options);
	xmlFree(mpgcontent);
}

/*
 * Add the disk service requirement in database node 'vds_node' to the LDom
 * 'ldp'.
 */
static void
from_db_add_vds(xmlNodePtr vds_node, ldom_t *ldp)
{
	xmlNodePtr node;
	xmlNodePtr volumes_node;
	xmlChar *service_name = NULL;
	vds_t *vds;

	node = vds_node->xmlChildrenNode;
	if (db_major < 3)
		node = node->next;	/* skip instance */

	service_name = xmlNodeGetContent(node);

	vds = vds_new(ldp, (char *)service_name);
	if (vds != NULL)
		database_ldom_add_disk_service(ldp, (char *)service_name);

	xmlFree(service_name);

	/*
	 * Add the vds volumes.
	 */
	volumes_node = node->next;
	node = volumes_node->xmlChildrenNode;
	while (node != NULL) {
		from_db_add_volume_to_vds(node, vds);
		node = node->next;
	}
}

/*
 * Add the network client requirement in database node 'net_node' to the LDom
 * 'ldp'.
 */
static void
from_db_add_network(xmlNodePtr net_node, ldom_t *ldp)
{
	xmlNodePtr node;
	uint64_t mac_address;
	long mtu;	/* FIXME - should these be uint64_t ? */
	uint64_t phys_state = 0;
	int num_vlans;
	int pvid = DEFAULT_VLAN_ID;
	int *vlans = NULL;
	xmlChar *vnet_name = NULL;
	xmlChar *service_name = NULL;
	xmlChar *vnet_mode = NULL;
	xmlChar *vids = NULL;
	int id = NO_NEW_ID;

	node = net_node->xmlChildrenNode;
	if (db_major < 3)
		node = node->next;	/* skip instance */
	vnet_name = xmlNodeGetContent(node);
	node = node->next;
	service_name = xmlNodeGetContent(node);
	node = node->next;
	mac_address = xml_get_uint64_content(node);
	node = node->next;
	mtu = xml_get_long_content(node);
	if ((db_major == 3 && db_minor >= 1) || (db_major > 3)) {
		node = node->next;
		vnet_mode = xmlNodeGetContent(node);
		node = node->next;
		pvid = xml_get_long_content(node);
		node = node->next;
		vids = xmlNodeGetContent(node);

		if (vnet_mode && IS_EMPTY_STR((char *)vnet_mode)) {
			xmlFree(vnet_mode);
			vnet_mode = NULL;
		}

		if (vids && IS_EMPTY_STR((char *)vids)) {
			xmlFree(vids);
			vids = NULL;
		}
		vlans = parse_num_list((char *)vids, &num_vlans);
	}
	if ((db_major == 3 && db_minor >= 5) || (db_major > 3)) {
		/*
		 * A patch containing v3.6 of the database was released without
		 * v3.5 support. Consequently, a v3.5+ database might not
		 * contain a 'linkprop' node. Explicitly verify it is present
		 * before assigning a database value.
		 */
		ASSERT(node->next != NULL);
		if (strcmp((char *)node->next->name,
		    (char *)LINKPROP_NODE) == 0) {
			node = node->next;
			phys_state = xml_get_uint64_content(node);
		}
	}
	if ((db_major == 3 && db_minor >= 6) || (db_major > 3)) {
		node = node->next;
		id = xml_get_int_content(node);
	}

	if (vnet_new(ldp, (char *)vnet_name, (char *)service_name,
	    mac_address, mtu, phys_state, (char *)vnet_mode, pvid, num_vlans,
	    vlans, id) != NULL) {
		database_ldom_add_net_client(ldp, (char *)vnet_name,
		    (char *)service_name, mac_address, mtu, phys_state,
		    (char *)vnet_mode, pvid, (char *)vids, id);
	} else {
		Xfree(vlans);
	}

	xmlFree(vids);
	xmlFree(vnet_name);
	xmlFree(service_name);
	xmlFree(vnet_mode);
}

/*
 * Add the pcie requirement in database node 'pcie_node' to the LDom 'ldp'.
 */
static void
from_db_add_pciedev(xmlNodePtr pciedev_node, ldom_t *ldp)
{
	xmlNodePtr node;
	xmlChar *pciedev_name = NULL;

	node = pciedev_node->xmlChildrenNode;
	pciedev_name = xmlNodeGetContent(node);

	/*
	 * Add only if it's not already in Zeus inventory
	 */
	if (find_node_by_prop(pciedev_node, PCIEDEV_INSTANCE_NODE,
	    PCIEDEV_NAME_NODE, pciedev_name) == NULL) {

		/* Add IO to constraint list for LDom */
		(void) add_pciedev_constraint(ldp, (char *)pciedev_name);

		database_ldom_add_pcie_device(ldp, (char *)pciedev_name);
	}

	xmlFree(pciedev_name);
}

/*
 * Add the io requirement in database node 'io_node' to the LDom 'ldp'.
 */
static void
from_db_add_io(xmlNodePtr io_node, ldom_t *ldp)
{
	xmlNodePtr node;
	xmlChar *io_name = NULL;
	int bypass_mode;
	bool_t flag;
	bool_t bypass = false;

	node = io_node->xmlChildrenNode;
	io_name = xmlNodeGetContent(node);
	node = node->next;
	bypass_mode = xml_get_int_content(node);
	if (bypass_mode == 1)
		bypass = true;

	/*
	 * Add only if it's not already in Zeus inventory
	 */
	if (find_node_by_prop(io_node, IO_INSTANCE_NODE, IO_NAME_NODE,
	    io_name) == NULL)
		flag = true;
	else
		flag = false;

	if (flag == true)
		/* Add IO to constraint list for LDom */
		(void) add_io_constraint(ldp, (char *)io_name, bypass);

	/*
	 * Add the node to the new db
	 */
	if (flag == true) {
		database_ldom_add_iodevice(ldp, (char *)io_name, bypass);
	}

	xmlFree(io_name);
}

/*
 * Add the network service requirement in database node 'vsw_node' to the LDom
 * 'ldp'.
 */
static void
from_db_add_vsw(xmlNodePtr vsw_node, ldom_t *ldp)
{
	xmlNodePtr node;
	xmlChar *service_name = NULL;
	xmlChar *phys_dev = NULL;
	xmlChar *mode = NULL;
	uint64_t mac_address;
	int default_vid = DEFAULT_VLAN_ID;
	int pvid = default_vid;
	xmlChar *vids = NULL;
	int *vlans = NULL;
	int num_vlans = 0;
	int mtu = VIO_ETHERMTU;
	uint64_t phys_state = 0;
	int id = NO_NEW_ID;

	node = vsw_node->xmlChildrenNode;
	if (db_major < 3)
		node = node->next;	/* skip instance */
	service_name = xmlNodeGetContent(node);
	node = node->next;
	if (db_major < 3)
		node = node->next;	/* skip sharing node */
	mac_address = xml_get_uint64_content(node);
	node = node->next;
	phys_dev = xmlNodeGetContent(node);
	/*
	 * Check for no physical device. In older DB versions " " was used to
	 * represent this so check for both this and empty string.
	 */
	if (phys_dev && (IS_EMPTY_STR((char *)phys_dev) ||
	    ((char *)phys_dev)[0] == ' ')) {
		xmlFree(phys_dev);
		phys_dev = NULL;
	}
	if (db_major >= 3) {
		node = node->next;
		mode = xmlNodeGetContent(node);
		if (mode && IS_EMPTY_STR((char *)mode)) {
			xmlFree(mode);
			mode = NULL;
		}
	}

	if ((db_major == 3 && db_minor >= 1) || (db_major > 3)) {
		node = node->next;
		default_vid = xml_get_long_content(node);
		node = node->next;
		pvid = xml_get_long_content(node);
		node = node->next;
		vids = xmlNodeGetContent(node);

		if (vids && IS_EMPTY_STR((char *)vids)) {
			xmlFree(vids);
			vids = NULL;
		}
		vlans = parse_num_list((char *)vids, &num_vlans);
	}

	if ((db_major == 3 && db_minor >= 3) || (db_major > 3)) {
		node = node->next;
		mtu = xml_get_long_content(node);
	}

	if ((db_major == 3 && db_minor >= 5) || (db_major > 3)) {
		/*
		 * A patch containing v3.6 of the database was released without
		 * v3.5 support. Consequently, a v3.5+ database might not
		 * contain a 'linkprop' node. Explicitly verify it is present
		 * before assigning a database value.
		 */
		ASSERT(node->next != NULL);
		if (strcmp((char *)node->next->name,
		    (char *)LINKPROP_NODE) == 0) {
			node = node->next;
			phys_state = xml_get_uint64_content(node);
		}
	}

	if ((db_major == 3 && db_minor >= 6) || (db_major > 3)) {
		node = node->next;
		id = xml_get_int_content(node);
	}

	if (vsw_new(ldp, (char *)service_name, (char *)phys_dev, mac_address,
	    (char *)mode, default_vid, pvid, num_vlans, vlans, mtu,
	    phys_state, id) != NULL) {
		database_ldom_add_net_service(ldp, (char *)service_name,
		    mac_address, (char *)phys_dev, phys_state, (char *)mode,
		    default_vid, pvid, (char *)vids, mtu, id);
	} else {
		Xfree(vlans);
	}

	xmlFree(vids);
	xmlFree(service_name);
	xmlFree(phys_dev);
	xmlFree(mode);
}

/*
 * Add the console service requirement in database node 'vcc_node' to the LDom
 * 'ldp'.
 */
static void
from_db_add_vcc(xmlNodePtr vcc_node, ldom_t *ldp)
{
	xmlNodePtr node;
	xmlChar *service_name = NULL;
	int min_port, max_port;

	node = vcc_node->xmlChildrenNode;
	if (db_major < 3)
		node = node->next;	/* skip instance */
	service_name = xmlNodeGetContent(node);
	node = node->next;
	min_port = xml_get_int_content(node);
	node = node->next;
	max_port = xml_get_int_content(node);

	if (vcc_new(ldp, (char *)service_name, min_port, max_port) != NULL)
		database_ldom_add_console_service(ldp, (char *)service_name,
		    min_port, max_port);

	xmlFree(service_name);
}

/*
 * Add the console client requirement in database node 'cons_node' to the LDom
 * 'ldp'.
 */
static void
from_db_add_console(xmlNodePtr cons_node, ldom_t *ldp)
{
	xmlNodePtr node;
	xmlChar *service_name = NULL;
	xmlChar *group_name = NULL;
	int port = _AUTO_VAL_;

	node = cons_node->xmlChildrenNode;
	if (db_major < 3)
		node = node->next;	/* skip instance */
	service_name = xmlNodeGetContent(node);
	node = node->next;
	group_name = xmlNodeGetContent(node);

	if (db_major > 2 || db_minor >= 1) {
		node = node->next;
		port = xml_get_int_content(node);
	}

	if (service_name && IS_EMPTY_STR((char *)service_name)) {
		xmlFree(service_name);
		service_name = NULL;
	}

	if (group_name && IS_EMPTY_STR((char *)group_name)) {
		xmlFree(group_name);
		group_name = NULL;
	}

	if (console_new(ldp, (char *)service_name, (char *)group_name, port) !=
	    NULL)
		database_ldom_add_console_client(ldp, (char *)service_name,
		    (char *)group_name, port);

	xmlFree(service_name);
	xmlFree(group_name);
}

/*
 * Add the VDPCS channel server
 */
static void
from_db_add_vdpc_server(xmlNodePtr vdpcs_node, ldom_t *ldp)
{
	xmlNodePtr node;
	xmlChar *service_name;

	node = vdpcs_node->xmlChildrenNode;
	if (db_major < 3)
		node = node->next;	/* skip instance */
	service_name = xmlNodeGetContent(node);

	if (vdpcs_new(ldp, (char *)service_name) != NULL)
		database_ldom_add_vdpc_service(ldp, (char *)service_name);

	xmlFree(service_name);
}

/*
 * Add the VDPCS channel client
 */
static void
from_db_add_vdpc_client(xmlNodePtr vdpc_node, ldom_t *ldp)
{
	xmlNodePtr node;
	xmlChar *vdpcc_name = NULL;
	xmlChar *service_name = NULL;

	node = vdpc_node->xmlChildrenNode;
	if (db_major < 3)
		node = node->next;	/* skip instance */
	vdpcc_name = xmlNodeGetContent(node);
	node = node->next;
	service_name = xmlNodeGetContent(node);
	node = node->next;

	if (vdpcc_new(ldp, (char *)vdpcc_name, (char *)service_name) != NULL)
		database_ldom_add_vdpc_client(ldp, (char *)vdpcc_name,
		    (char *)service_name);

	xmlFree(vdpcc_name);
	xmlFree(service_name);
}

typedef void add_instance_f(xmlNodePtr, ldom_t *);

/*
 * Add DRM policies
 */
static void
from_db_add_policy(xmlNodePtr pinst_node, ldom_t *ldp)
{
	policy_t	*polp;
	xmlNodePtr	node;
	xmlChar		*content;

	polp = Xcalloc(1, policy_t);
	polp->ldp = ldp;

	node = pinst_node->xmlChildrenNode;
	content = xmlNodeGetContent(node);
	polp->namep = Xstrdup((char *)content);
	xmlFree(content);

	node = node->next;
	content = xmlNodeGetContent(node);
	if (strcmp((char *)content, "on") == 0)
		polp->enable = true;
	else
		polp->enable = false;
	xmlFree(content);

	node = node->next;
	polp->priority = xml_get_int_content(node);

	node = node->next;
	polp->vcpu_min = xml_get_int_content(node);

	node = node->next;
	polp->vcpu_max = xml_get_int_content(node);

	node = node->next;
	polp->util_lower = xml_get_int_content(node);

	node = node->next;
	polp->util_upper = xml_get_int_content(node);

	node = node->next;
	polp->tod_begin = xml_get_int_content(node);

	node = node->next;
	polp->tod_end = xml_get_int_content(node);

	node = node->next;
	polp->sample_rate = xml_get_int_content(node);

	node = node->next;
	polp->elastic_margin = xml_get_int_content(node);

	node = node->next;
	polp->attack = xml_get_int_content(node);

	node = node->next;
	polp->decay = xml_get_int_content(node);

	ldp_set_policy(ldp, polp);
	database_ldom_set_policy(ldp, polp);
	database_save();
}

static void
from_db_add_instances(xmlNodePtr instances, ldom_t *ldp, add_instance_f addfn)
{
	xmlNodePtr node;

	node = instances->xmlChildrenNode;
	while (node != NULL) {
		addfn(node, ldp);
		node = node->next;
	}
}

/*
 * Add all the requirements in database node 'ldom_node' to the LDom 'ldp'.
 */
static void
from_db_add_to_manager(xmlNodePtr ldom_node, ldom_t *ldp)
{
	xmlNodePtr node;

	DBG(zprintf("Adding LDom %s from database\n", ldp->namep));

	/* skip name node */
	node = ldom_node->xmlChildrenNode->next;
	from_db_add_cpu(node, ldp);
	node = node->next;
	from_db_add_mau(node, ldp);
	node = node->next;
	from_db_add_memory(node, ldp);
	node = node->next;
	if (db_major > 2 || db_minor >= 2) {
		from_db_add_instances(node, ldp, from_db_add_io);
		node = node->next;
		from_db_add_mac_addr(node, ldp);
		node = node->next;
	}

	from_db_add_instances(node, ldp, from_db_add_vsw);
	node = node->next;
	from_db_add_instances(node, ldp, from_db_add_network);
	node = node->next;
	from_db_add_instances(node, ldp, from_db_add_vds);
	node = node->next;
	from_db_add_instances(node, ldp, from_db_add_disk);
	node = node->next;
	from_db_add_instances(node, ldp, from_db_add_vcc);
	node = node->next;
	from_db_add_instances(node, ldp, from_db_add_console);
	node = node->next;
	if (db_major < 3)
		/* skip vldcc node */
		node = node->next;
	from_db_add_instances(node, ldp, from_db_add_vdpc_server);
	node = node->next;
	from_db_add_instances(node, ldp, from_db_add_vdpc_client);
	node = node->next;
	from_db_add_variables(node, ldp);
	if (db_major == 3)
		/* skip options node */
		/* see comments in database_internal.h for an explanation */
		node = node->next;
	if ((db_major == 3 && db_minor >= 2) || (db_major > 3)) {
		node = node->next;
		from_db_add_hostid(node, ldp);
	}
	if ((db_major == 3 && db_minor >= 4) || (db_major > 3)) {
		node = node->next;
		from_db_add_failure_policy(node, ldp);
		node = node->next;
		from_db_add_master(node, ldp);
	}
	if ((db_major == 3 && db_minor >= 7) || (db_major > 3)) {
		node = node->next;
		from_db_add_instances(node, ldp, from_db_add_policy);
	}
	if (DB_SUPPORTS_UUID()) {
		node = node->next;
		from_db_add_uuid(node, ldp);
	}
	if (DB_SUPPORTS_SDIO()) {
		node = node->next;
		from_db_add_instances(node, ldp, from_db_add_pciedev);
	}
}

#define	DBFILE	"ldom-db.xml"

static void
make_database_filename(void)
{
	if (filename == NULL) {
		filename = zsprintf("%s/%s", zeus.admdir, DBFILE);
		tmpfilename = zsprintf("%s.tmp", filename);
	}
}

static bool_t
old_dbase_exists(void)
{
	int fd;

	fd = open(filename, O_RDWR);
	if (fd == -1) {
		if (errno == ENOENT)
			return (false);
		else
			fatal("Error opening database file \"%s\"",
			    filename);
			/* NOTREACHED */
	} else {
		close(fd);
		return (true);
	}
}

/*
 * get the node ptr from the old dbase (not yet deleted)
 */
static xmlNodePtr
ldom_node_in_old_dbase(ldom_t *ldp)
{
	if (old_dbase != NULL)
		return (find_ldom(old_dbase, (xmlChar *)ldp->namep));
	else
		return (NULL);
}

/*
 * Find the property node from the old database with a particular
 * node type having a matching property, e.g. find the mtu node
 * from the particular node of type VNET_NODE in ldom 'ldg1' whose
 * VNET_NAME node matches value 'vnet0' would use the following
 * arguments:
 * node_type = VNET_NODE
 * match_prop = VNET_NAME_NODE
 * match_val = "vnet0"
 * prop_name = MTU_NODE
 */
static xmlNodePtr
find_prop_node_in_old_dbase(ldom_t *ldp, xmlChar *node_type,
    xmlChar *match_prop, char *match_val, xmlChar *prop_name)
{
	xmlNodePtr ldom_node;

	if ((ldom_node = ldom_node_in_old_dbase(ldp)) == NULL) {
		return (NULL);
	} else {
		return (find_prop_node_by_prop_val(ldom_node, node_type,
		    match_prop, (xmlChar *)match_val, (xmlChar *)prop_name));
	}
}

static bool_t
supported_dbase_version(xmlDocPtr db)
{
	xmlChar *version;
	uint16_t major, minor, sup_major, sup_minor;
	bool_t supported = false;

	if (get_dbase_version(db, &version, &major, &minor) == false)
		return (false);

	DBG(zprintf("Current database version %d.%d\n", major, minor));

	if (negotiate_version(NUM_VERS(dbase_versions),
	    dbase_versions, major, &sup_major, &sup_minor)) {
		DBG(zprintf("Negotiated database version %d.%d\n",
		    sup_major, sup_minor));
		supported = true;
	}

	if (!supported)
		warning("Unsupported database version %s", (char *)version);
	xmlFree(version);

	return (supported);
}

static void
create_database(void)
{
	xmlNodePtr root;

	dbase = xmlNewDoc(XML_VERSION);
	root = xmlNewDocNode(dbase, NULL, LDOM_DATABASE_NODE, NULL);
	xmlDocSetRootElement(dbase, root);
	xmlNewProp(root, VERSION_ATTR, (xmlChar *)LDOM_DATABASE_VERSION);
}


static void
database_backup(xmlDocPtr db)
{
	char *backup_filename;

	assert(filename != NULL);
	backup_filename = zsprintf("%s.bak", filename);

	if (xmlSaveFormatFile(backup_filename, db, true) == -1)
		warning("Database could not be backed up");
	else
		warning("Database backed up to file %s", backup_filename);

	Xfree(backup_filename);
}

/*
 * Create a new empty database and optionally load the old database.
 */
int
database_init(void)
{
	LIBXML_TEST_VERSION;		/* calls xmlInitParser() */

	make_database_filename();

	create_database();

	if (option.load_database) {
		if (old_dbase_exists()) {
			xmlKeepBlanksDefault(0);
			old_dbase = xmlParseFile(filename);
			if (old_dbase == NULL) {
				warning("Old database is not well-formed. "
				    "Ignoring old database");
			} else {
				if (!supported_dbase_version(old_dbase)) {
					database_backup(old_dbase);
					xmlFreeDoc(old_dbase);
					old_dbase = NULL;
				} else if (option.validate_xml &&
				    !validate_database(old_dbase)) {
					warning("Old database is not valid. "
					    "Ignoring old database");
					database_backup(old_dbase);
					xmlFreeDoc(old_dbase);
					old_dbase = NULL;
				}
			}
			/* Override DB version globals if old_dbase valid */
			if (old_dbase) {
				if (get_dbase_version(old_dbase, NULL,
				    &db_major, &db_minor) == false)
					/*
					 * Shouldn't get here; version
					 * has already been validated.
					 */
					fatal("old_dbase with bad version");
			}
		}
	}

	return (0);
}

/*ARGSUSED1*/
static void
from_db_add_to_vds(xmlNodePtr vds_node, ldom_t *ldp)
{
	xmlNodePtr node;
	xmlNodePtr volumes_node;
	xmlChar *service_name;
	vds_t *vds;

	node = vds_node->xmlChildrenNode;
	if (db_major < 3) {
		service_name = xmlNodeGetContent(node->next);
		volumes_node = node->next->next;
	} else {
		/* no instance node */
		service_name = xmlNodeGetContent(node);
		volumes_node = node->next;
	}

	vds = vds_find_by_name((char *)service_name);
	if (vds != NULL) {
		node = volumes_node->xmlChildrenNode;
		while (node != NULL) {
			from_db_add_volume_to_vds(node, vds);
			node = node->next;
		}
	}

	xmlFree(service_name);
}

/*
 * Currently all we do here is add the vds devices for any vds instance which
 * is in both the MD and the database.
 */
static void
from_db_add_vio_services(xmlNodePtr ldom_node, ldom_t *ldp)
{
	xmlNodePtr vds_node;

	if ((vds_node = xml_find_subnode(ldom_node, VDS_NODE)) != NULL)
		from_db_add_instances(vds_node, ldp, from_db_add_to_vds);
}

static void
from_db_add_drm_policies(xmlNodePtr ldom_node, ldom_t *ldp)
{
	xmlNodePtr pnode;

	if ((pnode = xml_find_subnode(ldom_node, POLICY_NODE)) != NULL)
		from_db_add_instances(pnode, ldp, from_db_add_policy);
}


/*
 * Add all LDoms which are in the old database but are not in the HV MDs, to
 * the domain manager and to the current database.
 */
int
database_add_unbound_ldoms(void)
{
	ldom_t		*ldp;
	xmlNodePtr	oldroot;
	xmlNodePtr	node;
	xmlChar		*name;

	if (old_dbase == NULL)
		return (0);

	oldroot = xmlDocGetRootElement(old_dbase);
	node = oldroot->xmlChildrenNode;

	while (node != NULL) {

		if (xmlStrcmp(node->name, LDOM_NODE) == 0) {
			/* lookup the domain */
			name = ldom_get_name(node);
			assert(name != NULL);
			ldp = find_ldom_by_name((char *)name);

			if (ldp == NULL) {
				ldp = new_ldom((char *)name);
				from_db_add_to_manager(node, ldp);

				/*
				 * If there is no control domain then
				 * make this the control domain.
				 */
				new_ldom_vldc(ldp, NULL);
			} else {
				from_db_add_vio_services(node, ldp);
				from_db_add_drm_policies(node, ldp);
			}
			xmlFree(name);
		}
		node = node->next;
	}

	database_save();
	return (0);
}

/*
 * Now that unbound domains are available, finish initialization of masters.
 */
int
database_add_masters(void)
{
	ldom_t		*ldp;
	xmlNodePtr	oldroot;
	xmlNodePtr	node;
	xmlChar		*name;
	xmlNodePtr	master_node;

	if (old_dbase == NULL)
		return (0);

	oldroot = xmlDocGetRootElement(old_dbase);
	node = oldroot->xmlChildrenNode;

	while (node != NULL) {
		if (xmlStrcmp(node->name, LDOM_NODE) == 0) {
			/* lookup the domain */
			name = ldom_get_name(node);
			assert(name != NULL);
			ldp = find_ldom_by_name((char *)name);
			xmlFree(name);

			master_node = xml_find_subnode(node, MASTER_NODE);
			from_db_add_master(master_node, ldp);
		}
		node = node->next;
	}

	/* clean up old database */
	xmlFreeDoc(old_dbase);
	old_dbase = NULL;

	database_save();
	return (0);
}

/*
 * Now that unbound domains are available, finish initialization of UUID.
 */
int
database_ldom_get_uuid(ldom_t *ldp)
{
	xmlNodePtr	oldroot;
	xmlNodePtr	node;
	xmlChar		*name;
	xmlNodePtr	uuid_node;

	if (old_dbase == NULL)
		return (0);

	oldroot = xmlDocGetRootElement(old_dbase);

	for (node = oldroot->xmlChildrenNode; node != NULL; node = node->next) {
		if (xmlStrcmp(node->name, LDOM_NODE) == 0) {
			/* lookup the domain */
			name = ldom_get_name(node);
			assert(name != NULL);
			if (strcmp(ldp->namep, (char *)name) == 0) {
				xmlFree(name);
				uuid_node = xml_find_subnode(node, UUID_NODE);
				if (uuid_node == NULL)
					return (0);
				from_db_add_uuid(uuid_node, ldp);
				return (1);
			} else {
				xmlFree(name);
			}
		}
	}

	return (0);
}

/*
 * Save database to permanent storage.
 */
int
database_save(void)
{
	assert(dbase != NULL);
	assert(filename != NULL);

	/* write to temporary file */
	if (xmlSaveFormatFile(tmpfilename, dbase, true) == -1) {
		warning("Database could not be saved to file");
		return (-1);
	}

	/* rename over real database file */
	if (rename(tmpfilename, filename) == -1) {
		ewarning("Database could not be saved to file");
		return (-1);
	}

	return (0);
}

/*
 * Remove the definition for LDom 'ldp' from the database and clean up the
 * link to it.
 */
void
database_ldom_destroy(ldom_t *ldp)
{
	xmlNodePtr ldom_node;

	ldom_node = find_ldom(dbase, (xmlChar *)ldp->namep);
	if (ldom_node != NULL) {
		assert((ldp->dbp != NULL) &&
		    (ldom_node == ldp->dbp->ldom_node));

		xmlUnlinkNode(ldom_node);
		xmlFreeNode(ldom_node);
		database_save();
		Xfree(ldp->dbp);
	}
}

/*
 * Remove a specified policy instance from the database and clean up the
 * link to it.
 */
void
database_ldom_destroy_policy_inst(ldom_t *ldp, char *policy_name)
{
	xmlNodePtr	policy_node;
	xmlNodePtr	pinst_node;
	xmlNodePtr	node;
	xmlChar		*name;

	policy_node = xml_find_subnode(ldp->dbp->ldom_node, POLICY_NODE);
	if (policy_node == NULL)
		return;

	pinst_node = policy_node->xmlChildrenNode;
	while (pinst_node != NULL) {
		node =  xml_find_subnode(pinst_node, POLICY_NAME_NODE);
		if (node == NULL)
			return;
		name = xmlNodeGetContent(node);
		if (strcmp((char *)name, policy_name) == 0) {
			xmlUnlinkNode(pinst_node);
			xmlFreeNode(pinst_node);
			database_save();
			return;
		}
		pinst_node = pinst_node->next;
	}
}

/*
 * Set in the database the CPU requirements for LDom 'ldp'.
 */
int
database_ldom_set_cpus(ldom_t *ldp, int ncpus)
{
	xmlNodePtr cpu_node;
	char ncpus_str[INT_STR_LEN];

	snprintf(ncpus_str, sizeof (ncpus_str), "%d", ncpus);

	cpu_node = xml_find_subnode(ldp->dbp->ldom_node, CPU_NODE);

	if (cpu_node->xmlChildrenNode != NULL)
		xmlNodeSetContent(cpu_node->xmlChildrenNode,
		    (xmlChar *)ncpus_str);
	else
		xmlNewChild(cpu_node, NULL, NUMBER_NODE, (xmlChar *)ncpus_str);

	return (0);
}

/*
 * Set in the database the MAU requirements for LDom 'ldp'.
 */
int
database_ldom_set_maus(ldom_t *ldp, int nmaus)
{
	xmlNodePtr mau_node;
	char nmaus_str[INT_STR_LEN];

	snprintf(nmaus_str, sizeof (nmaus_str), "%d", nmaus);

	mau_node = xml_find_subnode(ldp->dbp->ldom_node, MAU_NODE);

	ASSERT(mau_node != NULL);

	if (mau_node->xmlChildrenNode != NULL)
		xmlNodeSetContent(mau_node->xmlChildrenNode,
		    (xmlChar *)nmaus_str);
	else
		xmlNewChild(mau_node, NULL, NUMBER_NODE, (xmlChar *)nmaus_str);

	return (0);
}

/*
 * Set in the database the MAC address requirements for LDom 'ldp'.
 */
int
database_ldom_set_mac_addr(ldom_t *ldp, uint64_t mac_address)
{
	xmlNodePtr macaddr_node;
	char macaddr_str[MAC_STR_LEN];

	snprintf(macaddr_str, sizeof (macaddr_str),
	    "0x%012"PRIx64"", mac_address);

	macaddr_node = xml_find_subnode(ldp->dbp->ldom_node, MAC_ADDRESS_NODE);

	ASSERT(macaddr_node != NULL);

	xmlNodeSetContent(macaddr_node, (xmlChar *)macaddr_str);

	return (0);
}

/*
 * Set in the database the hostid requirement for LDom 'ldp'.
 */
int
database_ldom_set_hostid(ldom_t *ldp, uint64_t hostid)
{
	xmlNodePtr hostid_node;
	char hostid_str[UINT64_STR_LEN + 2]; /* Add +2 for "0x" */

	/* Need to have the 0x for strtoull to work later */
	snprintf(hostid_str, sizeof (hostid_str), "0x%"PRIx64, hostid);

	hostid_node = xml_find_subnode(ldp->dbp->ldom_node, HOSTID_NODE);

	ASSERT(hostid_node != NULL);

	xmlNodeSetContent(hostid_node, (xmlChar *)hostid_str);

	return (0);
}

/*
 * Set in the database the failure-policy requirements for LDom 'ldp'.
 */
int
database_ldom_set_failure_policy(ldom_t *ldp, uint64_t failure_policy)
{
	xmlNodePtr failure_policy_node;
	char failure_policy_str[UINT64_STR_LEN];

	snprintf(failure_policy_str, sizeof (failure_policy_str), "%"PRIx64,
	    failure_policy);
	failure_policy_node = xml_find_subnode(ldp->dbp->ldom_node,
	    FAILURE_POLICY_NODE);
	ASSERT(failure_policy_node != NULL);

	xmlNodeSetContent(failure_policy_node, (xmlChar *)failure_policy_str);
	return (0);
}

/*
 * Set in the database the master requirements for LDom 'ldp'.
 */
int
database_ldom_set_master(ldom_t *ldp)
{
	xmlNodePtr	master_node;
	char		*masters_str;

	if (validate_masters(ldp) == false)
		return (1);

	masters_str = get_masters_string(ldp);

	master_node = xml_find_subnode(ldp->dbp->ldom_node, MASTER_NODE);
	xmlNodeSetContent(master_node, (xmlChar *)masters_str);

	Zfree(masters_str);
	return (0);
}

/*
 * Set in the database the uuid requirements for LDom 'ldp'.
 */
int
database_ldom_set_uuid(ldom_t *ldp)
{
	xmlNodePtr uuid_node = NULL;

	uuid_node = xml_find_subnode(ldp->dbp->ldom_node, UUID_NODE);
	ASSERT(uuid_node != NULL);

	xmlNodeSetContent(uuid_node, (xmlChar *)ldp->props.uuid);
	return (0);
}

/*
 * Set in the database the memory requirements for LDom 'ldp'.
 */
int
database_ldom_set_memory(ldom_t *ldp, uint64_t memsize)
{
	xmlNodePtr mem_node;
	char memsize_str[UINT64_STR_LEN];

	snprintf(memsize_str, sizeof (memsize_str), "%"PRIu64"", memsize);

	mem_node = xml_find_subnode(ldp->dbp->ldom_node, MEMORY_NODE);

	if (mem_node->xmlChildrenNode != NULL)
		xmlNodeSetContent(mem_node->xmlChildrenNode,
		    (xmlChar *)memsize_str);
	else
		xmlNewChild(mem_node, NULL, SIZE_NODE, (xmlChar *)memsize_str);

	return (0);
}

/*
 * Set in the database the physical I/O bus nexus requirements for LDom 'ldp'.
 */
int
database_ldom_add_iodevice(ldom_t *ldp, char *name, bool_t bypass)
{
	char int_str[INT_STR_LEN];
	xmlNodePtr io_node, iodevice_node;
	int bypass_mode = 0;

	io_node = xml_find_subnode(ldp->dbp->ldom_node, IO_NODE);

	iodevice_node = xmlNewChild(io_node, NULL, IO_INSTANCE_NODE, NULL);

	xmlNewChild(iodevice_node, NULL, IO_NAME_NODE, (xmlChar *)name);

	if (bypass)
		bypass_mode = 1;
	snprintf(int_str, sizeof (int_str), "%d", bypass_mode);
	xmlNewChild(iodevice_node, NULL, IO_BYPASS_NODE, (xmlChar *)int_str);

	return (0);
}

/*
 * Set in the database the pcie device requirements for LDom 'ldp'.
 */
int
database_ldom_add_pcie_device(ldom_t *ldp, char *name)
{
	xmlNodePtr pciedev_node;
	xmlNodePtr pciedev_instance;

	ASSERT(name != NULL);
	pciedev_node = xml_find_subnode(ldp->dbp->ldom_node, PCIEDEV_NODE);

	pciedev_instance =
	    xmlNewChild(pciedev_node, NULL, PCIEDEV_INSTANCE_NODE, NULL);
	xmlNewChild(pciedev_instance, NULL, PCIEDEV_NAME_NODE, (xmlChar *)name);

	return (0);
}

/*
 * Add to the database a console service for LDom 'ldp'.
 */
int
database_ldom_add_console_service(ldom_t *ldp,
    char *service_name, int min_port, int max_port)
{
	xmlNodePtr console_servers, vcc_node;
	char int_str[INT_STR_LEN];

	console_servers = xml_find_subnode(ldp->dbp->ldom_node, VCC_NODE);

	vcc_node = xmlNewChild(console_servers, NULL, VCC_INSTANCE_NODE, NULL);

	xmlNewChild(vcc_node, NULL, SERVICE_NAME_NODE, (xmlChar *)service_name);

	snprintf(int_str, sizeof (int_str), "%ld", min_port);
	xmlNewChild(vcc_node, NULL, MIN_PORT_NODE, (xmlChar *)int_str);

	snprintf(int_str, sizeof (int_str), "%ld", max_port);
	xmlNewChild(vcc_node, NULL, MAX_PORT_NODE, (xmlChar *)int_str);

	return (0);
}

/*
 * Modify the specified console service in LDom 'ldp'.
 */
int
database_ldom_modify_console_service(ldom_t *ldp, char *service_name,
    int min_port, int max_port)
{
	xmlNodePtr vcc_node;

	/*
	 * Find the vcc node by the service name.
	 */
	vcc_node = find_node_by_prop(ldp->dbp->ldom_node, VCC_NODE,
	    SERVICE_NAME_NODE, (xmlChar *)service_name);

	if (min_port > 0) {
		xmlNodePtr min_port_node =
		    xml_find_subnode(vcc_node, MIN_PORT_NODE);

		xml_set_int_content(min_port_node, min_port);
	}

	if (max_port > 0) {
		xmlNodePtr max_port_node =
		    xml_find_subnode(vcc_node, MAX_PORT_NODE);

		xml_set_int_content(max_port_node, max_port);
	}

	return (0);
}

/*
 * Get a guest's single console instance if any.
 */
static xmlNodePtr
get_console_instance(ldom_t *ldp)
{
	xmlNodePtr node;

	node = xml_find_subnode(ldp->dbp->ldom_node, CONSOLE_NODE);
	if (node != NULL)
		node = xml_find_subnode(node, CONSOLE_INSTANCE_NODE);

	return (node);
}

/*
 * Add to the database a console client for LDom 'ldp'. If
 * port is NEW_INSTANCE for a console, save as 0 in DB
 */
int
database_ldom_add_console_client(ldom_t *ldp, char *service_name, char *group,
    int port)
{
	xmlNodePtr console_clients, cons_node;
	char port_str[INT_STR_LEN];

	console_clients = xml_find_subnode(ldp->dbp->ldom_node, CONSOLE_NODE);

	/*
	 * If a node already exists with default values, just modify it
	 */
	if (xml_find_subnode(console_clients, CONSOLE_INSTANCE_NODE) != NULL) {
		vcons_args_t mod = CONSOLE_SERVICE | CONSOLE_GROUP |
		    CONSOLE_PORT;

		return (database_ldom_modify_console_client(ldp, mod,
		    service_name, group, port));
	}

	cons_node = xmlNewChild(console_clients, NULL, CONSOLE_INSTANCE_NODE,
	    NULL);

	if (service_name && IS_EMPTY_STR(service_name))
		service_name = NULL;

	xmlNewChild(cons_node, NULL, SERVICE_NAME_NODE,
	    (xmlChar *)service_name);

	if (group && IS_EMPTY_STR(group))
		group = NULL;

	xmlNewChild(cons_node, NULL, GROUP_NODE, (xmlChar *)group);

	if (port == NEW_INSTANCE)
		port = 0;

	snprintf(port_str, sizeof (port_str), "%d", port);
	xmlNewChild(cons_node, NULL, VCONS_PORT_NODE, (xmlChar *)port_str);

	return (0);
}

/*
 * Modify the console client in LDom 'ldp'.
 */
int
database_ldom_modify_console_client(ldom_t *ldp, vcons_args_t mod,
    char *new_service_name, char *new_group_name, int tcp_port)
{
	xmlNodePtr vcons_node;

	vcons_node = get_console_instance(ldp);

	/*
	 * If no console instance exists yet, create one.
	 */
	if (vcons_node == NULL)
		return (database_ldom_add_console_client(ldp, new_service_name,
		    new_group_name, tcp_port));

	/*
	 * For each valid property value specified, find the sub node and
	 * update it.
	 */
	if (mod & CONSOLE_SERVICE) {
		xmlNodePtr service_name_node =
		    xml_find_subnode(vcons_node, SERVICE_NAME_NODE);

		ASSERT(service_name_node != NULL);

		if (new_service_name == NULL ||
		    IS_EMPTY_STR(new_service_name))
			new_service_name = NULL;

		xmlNodeSetContent(service_name_node,
		    (xmlChar *)new_service_name);
	}

	if (mod & CONSOLE_GROUP) {
		xmlNodePtr group_name_node =
		    xml_find_subnode(vcons_node, GROUP_NODE);

		ASSERT(group_name_node != NULL);

		if (new_group_name == NULL ||
		    IS_EMPTY_STR(new_group_name))
			new_group_name = NULL;

		xmlNodeSetContent(group_name_node, (xmlChar *)new_group_name);
	}

	/*
	 * If port is changed, save the change. A value of 0 means automatic
	 * port selection.
	 */
	if (mod & CONSOLE_PORT) {
		xmlNodePtr port_node =
		    xml_find_subnode(vcons_node, VCONS_PORT_NODE);
		xml_set_int_content(port_node, tcp_port);
	}

	return (0);
}

/*
 * Add to the database a disk service for LDom 'ldp'.
 */
int
database_ldom_add_disk_service(ldom_t *ldp, char *service_name)
{
	xmlNodePtr disk_servers, vds_node;

	disk_servers = xml_find_subnode(ldp->dbp->ldom_node, VDS_NODE);

	vds_node = xmlNewChild(disk_servers, NULL, VDS_INSTANCE_NODE, NULL);

	xmlNewChild(vds_node, NULL, SERVICE_NAME_NODE, (xmlChar *)service_name);

	/*
	 * Create an empty "vds_volumes" node
	 */
	(void) xmlNewChild(vds_node, NULL, VDS_VOLUMES_NODE, NULL);

	return (0);
}

/*
 * Add a disk volume to a given vds service. For this we need to
 * find the vds_instance node for the given ldom and instance. Then
 * locate the vds_volumes node and add a child node underneath.
 */
int
database_ldom_add_disk_service_vol(ldom_t *ldp, char *service, char *vol_name,
    char *block_dev, char *options, char *mpgname)
{
	xmlNodePtr node, vol_node;

	/*
	 * Find the vds_volumes node in vds node with matching service name
	 */
	node = find_prop_node_by_prop_val(ldp->dbp->ldom_node, VDS_NODE,
	    SERVICE_NAME_NODE, (xmlChar *)service, VDS_VOLUMES_NODE);
	if (node == NULL)
		return (-1);

	/*
	 * Create a new "vds_volume" node under "vds_volumes".
	 */
	vol_node = xmlNewChild(node, NULL, VDS_VOLUME_NODE, NULL);

	/*
	 * Add the property nodes under the "vds_volume" node.
	 */
	xmlNewChild(vol_node, NULL, BLOCKDEV_NODE, (xmlChar *)block_dev);
	xmlNewChild(vol_node, NULL, VOLUME_NAME_NODE, (xmlChar *)vol_name);
	xmlNewChild(vol_node, NULL, VOLUME_OPTS_NODE, (xmlChar *)options);
	xmlNewChild(vol_node, NULL, VOLUME_MPGROUP_NODE, (xmlChar *)mpgname);

	return (0);
}

/*
 * Modify options of a disk volume for LDom 'ldp'
 */
int
database_ldom_modify_vdsdev_props(ldom_t *ldp, char *service_name,
	char *vdsdev_name, char *options, char *mpgname)
{
	xmlNodePtr service_node;
	xmlNodePtr vdsdev_node;
	xmlNodePtr options_node;
	xmlNodePtr mpgroup_node;

	ASSERT(ldp != NULL);

	service_node = find_prop_node_by_prop_val(ldp->dbp->ldom_node, VDS_NODE,
	    SERVICE_NAME_NODE, (xmlChar *)service_name, VDS_VOLUMES_NODE);

	vdsdev_node = find_node_by_prop_val(service_node, VOLUME_NAME_NODE,
	    (const xmlChar *)vdsdev_name);

	if (db_major > 2 || db_minor > 1) {
		options_node = xml_find_subnode(vdsdev_node, VOLUME_OPTS_NODE);
		xmlNodeSetContent(options_node, (xmlChar *)options);
	} else {
		xmlNewChild(vdsdev_node, NULL, VOLUME_OPTS_NODE,
		    (xmlChar *)options);
	}

	if (((db_major == 3) && (db_minor >= 2)) || (db_major > 3)) {
		mpgroup_node = xml_find_subnode(vdsdev_node,
		    VOLUME_MPGROUP_NODE);
		xmlNodeSetContent(mpgroup_node, (xmlChar *)mpgname);
	} else {
		xmlNewChild(vdsdev_node, NULL, VOLUME_MPGROUP_NODE,
		    (xmlChar *)mpgname);
	}

	return (0);
}

/*
 * Add to the database a VDPCS channel service for LDom 'ldp'.
 */
int
database_ldom_add_vdpc_service(ldom_t *ldp, char *service_name)
{
	xmlNodePtr vdpc_servers, vdpcs_node;

	vdpc_servers = xml_find_subnode(ldp->dbp->ldom_node, VDPCS_NODE);

	vdpcs_node = xmlNewChild(vdpc_servers, NULL, VDPCS_INSTANCE_NODE, NULL);

	xmlNewChild(vdpcs_node, NULL, SERVICE_NAME_NODE,
	    (xmlChar *)service_name);

	return (0);
}

/*
 * Add to the database a VDPCS channel client for LDom 'ldp'.
 */
int
database_ldom_add_vdpc_client(ldom_t *ldp, char *vdpc_name,
    char *service_name)
{
	xmlNodePtr vdpc_clients, vdpc_node;

	vdpc_clients = xml_find_subnode(ldp->dbp->ldom_node, VDPCC_NODE);

	vdpc_node = xmlNewChild(vdpc_clients, NULL, VDPCC_INSTANCE_NODE, NULL);

	xmlNewChild(vdpc_node, NULL, VDPCC_NAME_NODE, (xmlChar *)vdpc_name);
	xmlNewChild(vdpc_node, NULL, SERVICE_NAME_NODE,
	    (xmlChar *)service_name);

	return (0);
}

/*
 * Add or modify a policy instance for LDom 'ldp'.
 */
int
database_ldom_set_policy(ldom_t *ldp, policy_t *polp)
{
	xmlNodePtr	policy_node;
	xmlNodePtr	pinst_node;
	xmlNodePtr	node;
	xmlChar		*name;
	bool_t		newChild = true;
	char		int_str[INT_STR_LEN];

	policy_node = xml_find_subnode(ldp->dbp->ldom_node, POLICY_NODE);
	ASSERT(policy_node != NULL);

	/* Look for a matching policy instance */
	pinst_node = policy_node->xmlChildrenNode;
	while (pinst_node != NULL) {
		node =  xml_find_subnode(pinst_node, POLICY_NAME_NODE);
		ASSERT(node != NULL);
		name = xmlNodeGetContent(node);
		if (strcmp((char *)name, polp->namep) == 0) {
			newChild = false;
			break;
		}
		pinst_node = pinst_node->next;
	}

	if (newChild == true) {
		pinst_node = xmlNewChild(policy_node, NULL,
		    POLICY_INSTANCE_NODE, NULL);
		xmlNewChild(pinst_node, NULL, POLICY_NAME_NODE,
		    (xmlChar *)polp->namep);
		xmlNewChild(pinst_node, NULL, POLICY_ENABLE_NODE,
		    (polp->enable == true) ? (xmlChar *)"on" :
		    (xmlChar *)"off");
		snprintf(int_str, sizeof (int_str), "%d", polp->priority);
		xmlNewChild(pinst_node, NULL, POLICY_PRIORITY_NODE,
		    (xmlChar *)int_str);
		snprintf(int_str, sizeof (int_str), "%d", polp->vcpu_min);
		xmlNewChild(pinst_node, NULL, POLICY_VCPU_MIN_NODE,
		    (xmlChar *)int_str);
		snprintf(int_str, sizeof (int_str), "%d", polp->vcpu_max);
		xmlNewChild(pinst_node, NULL, POLICY_VCPU_MAX_NODE,
		    (xmlChar *)int_str);
		snprintf(int_str, sizeof (int_str), "%d", polp->util_lower);
		xmlNewChild(pinst_node, NULL, POLICY_UTIL_LOWER_NODE,
		    (xmlChar *)int_str);
		snprintf(int_str, sizeof (int_str), "%d", polp->util_upper);
		xmlNewChild(pinst_node, NULL, POLICY_UTIL_UPPER_NODE,
		    (xmlChar *)int_str);
		snprintf(int_str, sizeof (int_str), "%d", polp->tod_begin);
		xmlNewChild(pinst_node, NULL, POLICY_TOD_BEGIN_NODE,
		    (xmlChar *)int_str);
		snprintf(int_str, sizeof (int_str), "%d", polp->tod_end);
		xmlNewChild(pinst_node, NULL, POLICY_TOD_END_NODE,
		    (xmlChar *)int_str);
		snprintf(int_str, sizeof (int_str), "%d", polp->sample_rate);
		xmlNewChild(pinst_node, NULL, POLICY_SAMPLE_RATE_NODE,
		    (xmlChar *)int_str);
		snprintf(int_str, sizeof (int_str), "%d", polp->elastic_margin);
		xmlNewChild(pinst_node, NULL, POLICY_ELASTIC_MARGIN_NODE,
		    (xmlChar *)int_str);
		snprintf(int_str, sizeof (int_str), "%d", polp->attack);
		xmlNewChild(pinst_node, NULL, POLICY_ATTACK_NODE,
		    (xmlChar *)int_str);
		snprintf(int_str, sizeof (int_str), "%d", polp->decay);
		xmlNewChild(pinst_node, NULL, POLICY_DECAY_NODE,
		    (xmlChar *)int_str);
	} else {
		node = pinst_node->xmlChildrenNode;
		xmlNodeSetContent(node, (xmlChar *)polp->namep);
		node = node->next;
		xmlNodeSetContent(node, (polp->enable == true) ?
		    (xmlChar *)"on" : (xmlChar *)"off");
		node = node->next;
		xml_set_int_content(node, polp->priority);
		node = node->next;
		xml_set_int_content(node, polp->vcpu_min);
		node = node->next;
		xml_set_int_content(node, polp->vcpu_max);
		node = node->next;
		xml_set_int_content(node, polp->util_lower);
		node = node->next;
		xml_set_int_content(node, polp->util_upper);
		node = node->next;
		xml_set_int_content(node, polp->tod_begin);
		node = node->next;
		xml_set_int_content(node, polp->tod_end);
		node = node->next;
		xml_set_int_content(node, polp->sample_rate);
		node = node->next;
		xml_set_int_content(node, polp->elastic_margin);
		node = node->next;
		xml_set_int_content(node, polp->attack);
		node = node->next;
		xml_set_int_content(node, polp->decay);
	}

	return (0);
}

/*
 * Add to the database a disk client for LDom 'ldp'.
 */
int
database_ldom_add_disk_client(ldom_t *ldp, char *vdisk_name,
    char *service_name, char *vol_name, int timeout, int id)
{
	xmlNodePtr disk_clients, disk_node;
	char int_str[INT_STR_LEN];

	disk_clients = xml_find_subnode(ldp->dbp->ldom_node, DISK_NODE);

	disk_node = xmlNewChild(disk_clients, NULL, DISK_INSTANCE_NODE, NULL);

	xmlNewChild(disk_node, NULL, VDISK_NAME_NODE, (xmlChar *)vdisk_name);
	xmlNewChild(disk_node, NULL, SERVICE_NAME_NODE,
	    (xmlChar *)service_name);
	xmlNewChild(disk_node, NULL, VOLUME_NAME_NODE, (xmlChar *)vol_name);
	snprintf(int_str, sizeof (int_str), "%d", timeout);
	xmlNewChild(disk_node, NULL, TIMEOUT_NODE, (xmlChar *)int_str);
	snprintf(int_str, sizeof (int_str), "%d", id);
	xmlNewChild(disk_node, NULL, ID_NODE, (xmlChar *)int_str);

	return (0);
}

/*
 * Modify a vdisk client for LDom 'ldp'
 */
int
database_ldom_modify_disk_client(ldom_t *ldp, char *vdisk_name,
    char *service_name, char *volume_name, int timeout)
{
	xmlNodePtr disk_node;
	char int_str[INT_STR_LEN];

	ASSERT(ldp != NULL);

	disk_node = find_node_by_prop(ldp->dbp->ldom_node, DISK_NODE,
	    VDISK_NAME_NODE, (unsigned char *)vdisk_name);

	if (service_name != NULL) {
		xmlNodePtr service_node =
		    xml_find_subnode(disk_node, SERVICE_NAME_NODE);
		xmlNodeSetContent(service_node, (xmlChar *)service_name);
	}

	if (volume_name != NULL) {
		xmlNodePtr volume_node =
		    xml_find_subnode(disk_node, VOLUME_NAME_NODE);
		xmlNodeSetContent(volume_node, (xmlChar *)volume_name);
	}

	if (timeout != VDISK_NO_NEW_TIMEOUT) {
		snprintf(int_str, sizeof (int_str), "%d", timeout);
		if (db_major > 2 || db_minor > 2) {
			xmlNodePtr timeout_node = xml_find_subnode(disk_node,
			    TIMEOUT_NODE);
			xmlNodeSetContent(timeout_node, (xmlChar *)int_str);
		} else {
			xmlNewChild(disk_node, NULL, TIMEOUT_NODE,
			    (xmlChar *)int_str);
		}
	}

	return (0);
}

/*
 * Add to the database a network service for LDom 'ldp'.
 */
int
database_ldom_add_net_service(ldom_t *ldp, char *service_name,
    mac_addr_t mac_address, char *dev_path, uint64_t phys_state, char *mode,
    int default_vid, int pvid, char *vlans, int mtu, int id)
{
	xmlNodePtr net_servers, vsw_node;
	char mac_str[MAC_STR_LEN];
	char default_vid_str[INT_STR_LEN];
	char pvid_str[INT_STR_LEN];
	char mtu_str[INT_STR_LEN];
	char linkprop_str[INT_STR_LEN];
	char id_str[INT_STR_LEN];

	net_servers = xml_find_subnode(ldp->dbp->ldom_node, VSW_NODE);

	vsw_node = xmlNewChild(net_servers, NULL, VSW_INSTANCE_NODE, NULL);

	xmlNewChild(vsw_node, NULL, SERVICE_NAME_NODE, (xmlChar *)service_name);

	snprintf(mac_str, sizeof (mac_str), "0x%012"PRIx64"", mac_address);
	xmlNewChild(vsw_node, NULL, MAC_ADDRESS_NODE, (xmlChar *)mac_str);

	xmlNewChild(vsw_node, NULL, DEV_PATH_NODE, (xmlChar *)dev_path);

	xmlNewChild(vsw_node, NULL, VSW_MODE_NODE, (xmlChar *)mode);

	snprintf(default_vid_str, sizeof (default_vid_str), "%d", default_vid);
	snprintf(pvid_str, sizeof (pvid_str), "%d", pvid);

	xmlNewChild(vsw_node, NULL, VSW_DEF_VLAN_NODE,
	    (xmlChar *)default_vid_str);
	xmlNewChild(vsw_node, NULL, VSW_PORT_VLAN_NODE, (xmlChar *)pvid_str);
	xmlNewChild(vsw_node, NULL, VSW_VLANS_NODE, (xmlChar *)vlans);

	ASSERT(mtu == VIO_NONEMTU || MTU_IN_RANGE(mtu));
	snprintf(mtu_str, sizeof (mtu_str), "%d", mtu);
	xmlNewChild(vsw_node, NULL, VSW_MTU_NODE, (xmlChar *)mtu_str);

	ASSERT(phys_state == 0 || phys_state == 0x1);
	if (phys_state != 0 && phys_state != 0x1)
		phys_state = 0;
	snprintf(linkprop_str, sizeof (linkprop_str), "%llu", phys_state);
	xmlNewChild(vsw_node, NULL, LINKPROP_NODE, (xmlChar *)linkprop_str);

	snprintf(id_str, sizeof (id_str), "%d", id);
	xmlNewChild(vsw_node, NULL, ID_NODE, (xmlChar *)id_str);

	return (0);
}

/*
 * Modify the parameters of an existing vsw node
 */
int
database_ldom_modify_net_service(ldom_t *ldp, int mod,
    char *service_name, mac_addr_t mac_address, char *dev_path,
    uint64_t phys_state, char *mode, int pvid, int num_vlans, char *vlans,
    int mtu)
{
	xmlNodePtr vsw_node;

	/*
	 * Find the vsw node by the service name.
	 */
	vsw_node = find_node_by_prop(ldp->dbp->ldom_node, VSW_NODE,
	    SERVICE_NAME_NODE, (xmlChar *)service_name);

	/*
	 * Now for each valid property value specified, find the sub node and
	 * update it.
	 */
	if (mod & VSW_MAC) {
		char mac_str[MAC_STR_LEN];
		xmlNodePtr mac_addr_node =
		    xml_find_subnode(vsw_node, MAC_ADDRESS_NODE);

		snprintf(mac_str, sizeof (mac_str),
		    "0x%012"PRIx64"", mac_address);
		xmlNodeSetContent(mac_addr_node, (xmlChar *)mac_str);
	}

	if (mod & VSW_NET_DEV) {
		xmlNodePtr dev_path_node =
		    xml_find_subnode(vsw_node, DEV_PATH_NODE);
		xmlNodeSetContent(dev_path_node, (xmlChar *)dev_path);
	}

	if (mod & VSW_MODE) {
		if (db_major > 2) {
			xmlNodePtr mode_node = xml_find_subnode(vsw_node,
			    VSW_MODE_NODE);
			xmlNodeSetContent(mode_node, (xmlChar *)mode);
		} else {
			xmlNewChild(vsw_node, NULL, VSW_MODE_NODE,
			    (xmlChar *)mode);
		}
	}

	if (mod & VSW_PVID) {
		char pvid_str[INT_STR_LEN];
		snprintf(pvid_str, sizeof (pvid_str), "%d", pvid);
		if ((db_major == 3 && db_minor >= 1) || (db_major > 3)) {
			xmlNodePtr pvid_node = xml_find_subnode(vsw_node,
			    VSW_PORT_VLAN_NODE);
			xmlNodeSetContent(pvid_node, (xmlChar *)pvid_str);
		} else {
			xmlNewChild(vsw_node, NULL, VSW_PORT_VLAN_NODE,
			    (xmlChar *)pvid_str);
		}
	}

	if (mod & VSW_VID) {
		int nvlans;
		int *vlan_list = parse_num_list(vlans, &nvlans);

		ASSERT(vid_check(NULL, num_vlans, vlan_list) == 0);
		Xfree(vlan_list);
		if ((db_major == 3 && db_minor >= 1) || (db_major > 3)) {
			xmlNodePtr vid_node = xml_find_subnode(vsw_node,
			    VSW_VLANS_NODE);
			xmlNodeSetContent(vid_node, (xmlChar *)vlans);
		} else {
			xmlNewChild(vsw_node, NULL, VSW_VLANS_NODE,
			    (xmlChar *)vlans);
		}
	}

	if (mod & VSW_MTU) {
		char mtu_str[INT_STR_LEN];
		ASSERT(mtu == VIO_NONEMTU || MTU_IN_RANGE(mtu));
		snprintf(mtu_str, sizeof (mtu_str), "%d", mtu);
		if ((db_major == 3 && db_minor >= 3) || (db_major > 3)) {
			xmlNodePtr mtu_node = xml_find_subnode(vsw_node,
			    VSW_MTU_NODE);
			xmlNodeSetContent(mtu_node, (xmlChar *)mtu_str);
		} else {
			xmlNewChild(vsw_node, NULL, VSW_MTU_NODE,
			    (xmlChar *)mtu_str);
		}
	}

	if (mod & VSW_PSTATE) {
		char linkprop_str[INT_STR_LEN];
		ASSERT(phys_state == 0 || phys_state == 0x1);
		if (phys_state != 0 && phys_state != 0x1)
			phys_state = 0;
		snprintf(linkprop_str, sizeof (linkprop_str), "%llu",
		    phys_state);
		if ((db_major == 3 && db_minor >= 5) || (db_major > 3)) {
			xmlNodePtr link_node = xml_find_subnode(vsw_node,
			    LINKPROP_NODE);
			xmlNodeSetContent(link_node, (xmlChar *)linkprop_str);
		} else {
			xmlNewChild(vsw_node, NULL, LINKPROP_NODE,
			    (xmlChar *)linkprop_str);
		}
	}

	return (0);
}

/*
 * Add to the database a network client for LDom 'ldp'.
 */
int
database_ldom_add_net_client(ldom_t *ldp, char *vnet_name, char *service_name,
    mac_addr_t mac_address, int mtu, uint64_t phys_state, char *vnet_mode,
    int pvid, char *vlans, int id)
{
	xmlNodePtr net_clients, net_node;
	char s[UINT64_STR_LEN];
	char pvid_str[INT_STR_LEN];
	char linkprop_str[INT_STR_LEN];
	char id_str[INT_STR_LEN];

	net_clients = xml_find_subnode(ldp->dbp->ldom_node, VNET_NODE);

	net_node = xmlNewChild(net_clients, NULL, VNET_INSTANCE_NODE, NULL);

	xmlNewChild(net_node, NULL, VNET_NAME_NODE, (xmlChar *)vnet_name);
	xmlNewChild(net_node, NULL, SERVICE_NAME_NODE, (xmlChar *)service_name);

	snprintf(s, sizeof (s), "0x%012"PRIx64"", mac_address);
	xmlNewChild(net_node, NULL, MAC_ADDRESS_NODE, (xmlChar *)s);

	ASSERT(mtu == VIO_NONEMTU || MTU_IN_RANGE(mtu));
	snprintf(s, sizeof (s), "%ld", mtu);
	xmlNewChild(net_node, NULL, MTU_NODE, (xmlChar *)s);

	xmlNewChild(net_node, NULL, VNET_MODE_NODE, (xmlChar *)vnet_mode);

	snprintf(pvid_str, sizeof (pvid_str), "%d", pvid);
	xmlNewChild(net_node, NULL, VNET_PORT_VLAN_NODE, (xmlChar *)pvid_str);

	xmlNewChild(net_node, NULL, VNET_VLANS_NODE, (xmlChar *)vlans);

	ASSERT(phys_state == 0 || phys_state == 0x1);
	if (phys_state != 0 && phys_state != 0x1)
		phys_state = 0;
	snprintf(linkprop_str, sizeof (linkprop_str), "%llu", phys_state);
	xmlNewChild(net_node, NULL, LINKPROP_NODE, (xmlChar *)linkprop_str);

	snprintf(id_str, sizeof (id_str), "%d", id);
	xmlNewChild(net_node, NULL, ID_NODE, (xmlChar *)id_str);

	return (0);
}

/*
 * Modify the parameters of an existing vnet node.
 */
int
database_ldom_modify_net_client(ldom_t *ldp, int mod, char *client_name,
    char *service_name, mac_addr_t mac_address, int mtu, uint64_t phys_state,
    char *vnet_mode, int pvid, int num_vlans, char *vlans)
{
	xmlNodePtr vnet_node;

	/*
	 * Find the vnet node by the service name
	 */
	vnet_node = find_node_by_prop(ldp->dbp->ldom_node, VNET_NODE,
	    VNET_NAME_NODE, (xmlChar *)client_name);

	/*
	 * Now for each valid property value specified, find the sub node
	 * and update it
	 */
	if (mod & VNET_SERVICE) {
		xmlNodePtr service_name_node =
		    xml_find_subnode(vnet_node, SERVICE_NAME_NODE);

		xmlNodeSetContent(service_name_node, (xmlChar *)service_name);
	}

	if (mod & VNET_MAC) {
		char mac_str[MAC_STR_LEN];
		xmlNodePtr mac_addr_node =
		    xml_find_subnode(vnet_node, MAC_ADDRESS_NODE);

		snprintf(mac_str, sizeof (mac_str),
		    "0x%012"PRIx64"", mac_address);

		xmlNodeSetContent(mac_addr_node, (xmlChar *)mac_str);
	}

	if (mod & VNET_MTU) {
		char mtu_str[INT_STR_LEN];
		ASSERT(mtu == VIO_NONEMTU || MTU_IN_RANGE(mtu));
		snprintf(mtu_str, sizeof (mtu_str), "%d", mtu);
		if ((db_major == 3 && db_minor >= 2) || (db_major > 3)) {
			xmlNodePtr mtu_node = xml_find_subnode(vnet_node,
			    MTU_NODE);
			xmlNodeSetContent(mtu_node, (xmlChar *)mtu_str);
		} else {
			xmlNewChild(vnet_node, NULL, MTU_NODE,
			    (xmlChar *)mtu_str);
		}
	}

	if (mod & VNET_MODE) {
		if ((db_major == 3 && db_minor >= 2) || (db_major > 3)) {
			xmlNodePtr mode_node = xml_find_subnode(vnet_node,
			    VNET_MODE_NODE);
			ASSERT(mode_node != NULL);
			xmlNodeSetContent(mode_node, (xmlChar *)vnet_mode);
		} else {
			xmlNewChild(vnet_node, NULL, VNET_MODE_NODE,
			    (xmlChar *)vnet_mode);
		}
	}

	if (mod & VNET_PVID) {
		char pvid_str[INT_STR_LEN];
		snprintf(pvid_str, sizeof (pvid_str), "%d", pvid);
		if ((db_major == 3 && db_minor >= 1) || (db_major > 3)) {
			xmlNodePtr pvid_node = xml_find_subnode(vnet_node,
			    VNET_PORT_VLAN_NODE);
			xmlNodeSetContent(pvid_node, (xmlChar *)pvid_str);
		} else {
			xmlNewChild(vnet_node, NULL, VNET_PORT_VLAN_NODE,
			    (xmlChar *)pvid_str);
		}
	}

	if (mod & VNET_VID) {
		if ((db_major == 3 && db_minor >= 1) || (db_major > 3)) {
			xmlNodePtr pvid_node = xml_find_subnode(vnet_node,
			    VNET_VLANS_NODE);
			xmlNodeSetContent(pvid_node, (xmlChar *)vlans);
		} else {
			xmlNewChild(vnet_node, NULL, VNET_VLANS_NODE,
			    (xmlChar *)vlans);
		}
	}

	if (mod & VNET_PSTATE) {
		char linkprop_str[INT_STR_LEN];
		ASSERT(phys_state == 0 || phys_state == 0x1);
		if (phys_state != 0 && phys_state != 0x1)
			phys_state = 0;
		snprintf(linkprop_str, sizeof (linkprop_str), "%llu",
		    phys_state);
		if ((db_major == 3 && db_minor >= 5) || (db_major > 3)) {
			xmlNodePtr link_node = xml_find_subnode(vnet_node,
			    LINKPROP_NODE);
			xmlNodeSetContent(link_node, (xmlChar *)linkprop_str);
		} else {
			xmlNewChild(vnet_node, NULL, LINKPROP_NODE,
			    (xmlChar *)linkprop_str);
		}
	}

	return (0);
}

static bool_t
delete_node_by_prop(ldom_t *ldp, const xmlChar *type, xmlChar *prop_name,
    xmlChar *prop_val)
{
	xmlNodePtr node;

	node = find_node_by_prop(ldp->dbp->ldom_node, type,
	    prop_name, prop_val);

	if (node != NULL) {
		xmlUnlinkNode(node);
		xmlFreeNode(node);
		return (true);
	}

	return (false);
}

/*
 * Delete the console client from the database node for LDom 'ldp'.
 */
bool_t
database_ldom_del_console_client(ldom_t *ldp)
{
	xmlNodePtr node;

	node = get_console_instance(ldp);
	if (node != NULL) {
		xmlUnlinkNode(node);
		xmlFreeNode(node);
		return (true);
	}

	return (false);
}

/*
 * Delete the console service with instance 'instance' from the database node
 * for LDom 'ldp'.
 */
bool_t
database_ldom_del_console_service(ldom_t *ldp, char *service_name)
{
	return (delete_node_by_prop(ldp, VCC_NODE, SERVICE_NAME_NODE,
	    (xmlChar *)service_name));
}

/*
 * Delete the disk client with instance 'instance' from the database node
 * for LDom 'ldp'.
 */
bool_t
database_ldom_del_disk_client(ldom_t *ldp, char *client_name)
{
	return (delete_node_by_prop(ldp, DISK_NODE, VDISK_NAME_NODE,
	    (xmlChar *)client_name));
}

/*
 * Delete the disk service with instance 'instance' from the database node
 * for LDom 'ldp'.
 */
bool_t
database_ldom_del_disk_service(ldom_t *ldp, char *service_name)
{
	return (delete_node_by_prop(ldp, VDS_NODE, SERVICE_NAME_NODE,
	    (xmlChar *)service_name));
}

bool_t
database_ldom_del_disk_service_block(ldom_t *ldp, char *service_name,
    char *vol_name)
{
	xmlNodePtr volumes_node, node;

	/*
	 * Find the vds_volumes node in vds node with matching service name
	 */
	volumes_node = find_prop_node_by_prop_val(ldp->dbp->ldom_node, VDS_NODE,
	    SERVICE_NAME_NODE, (xmlChar *)service_name, VDS_VOLUMES_NODE);

	if (volumes_node == NULL)
		return (false);

	/*
	 * Find the volume with matching volume name
	 */
	node = find_node_by_prop_val(volumes_node, VOLUME_NAME_NODE,
	    (const xmlChar *)vol_name);

	if (node != NULL) {
		xmlUnlinkNode(node);
		xmlFreeNode(node);
		return (true);
	}

	return (false);
}

/*
 * Delete the network client with instance 'instance' from the database node
 * for LDom 'ldp'.
 */
bool_t
database_ldom_del_net_client(ldom_t *ldp, char *client_name)
{
	return (delete_node_by_prop(ldp, VNET_NODE, VNET_NAME_NODE,
	    (xmlChar *)client_name));
}

/*
 * Delete the network service with instance 'instance' from the database node
 * for LDom 'ldp'.
 */
bool_t
database_ldom_del_net_service(ldom_t *ldp, char *service_name)
{
	return (delete_node_by_prop(ldp, VSW_NODE, SERVICE_NAME_NODE,
	    (xmlChar *)service_name));
}

/*
 * Delete the VDPCS channel service from the database node
 * for LDom 'ldp'.
 */
bool_t
database_ldom_del_vdpc_service(ldom_t *ldp, char *service_name)
{
	return (delete_node_by_prop(ldp, VDPCS_NODE, SERVICE_NAME_NODE,
	    (xmlChar *)service_name));
}

/*
 * Delete the VDPCS channel client from the database node
 * for LDom 'ldp'.
 */
bool_t
database_ldom_del_vdpc_client(ldom_t *ldp, char *client_name)
{
	return (delete_node_by_prop(ldp, VDPCC_NODE, VDPCC_NAME_NODE,
	    (xmlChar *)client_name));
}

bool_t
database_ldom_del_iodevice(ldom_t *ldp, char *iodevice)
{
	return (delete_node_by_prop(ldp, IO_NODE, IO_NAME_NODE,
	    (xmlChar *)iodevice));
}

/*
 * Remove from the database the pcie device requirements for LDom 'ldp'.
 */
bool_t
database_ldom_del_pcie_device(ldom_t *ldp, char *name)
{
	ASSERT(name != NULL);
	return (delete_node_by_prop(ldp,
	    PCIEDEV_NODE, PCIEDEV_NAME_NODE, (xmlChar *)name));
}

/*
 * Returns true iff the pair node 'pair_node' has name 'name'.
 */
static bool_t
pair_is_named(xmlNodePtr pair_node, xmlChar *name)
{
	xmlChar *content;
	bool_t found;

	content = xmlNodeGetContent(pair_node->xmlChildrenNode);
	if (xmlStrcmp(content, name) == 0)
		found = true;
	else
		found = false;
	xmlFree(content);

	return (found);
}

/*
 * Find and return the nvpair node with name 'name' within the
 * pairs in 'pairs_node'. Returns NULL if there is no pair of
 * that name.
 */
static xmlNodePtr
find_pair(xmlNodePtr pairs_node, xmlChar *name)
{
	xmlNodePtr pair_node;

	pair_node = pairs_node->xmlChildrenNode;
	while (pair_node != NULL) {
		if (pair_is_named(pair_node, name))
			break;
		pair_node = pair_node->next;
	}

	return (pair_node);
}

/*
 * Add <'name', 'value'> pair to the database representation 'pairs_node' of a
 * set of pairs. If there is already a pair defined with the same name its
 * value is replaced with the new value. 'Tag' is the XML node tag for
 * a pair.
 */
static bool_t
add_pair(xmlNodePtr pairs_node, xmlChar *tag, xmlChar *name, xmlChar *value)
{
	xmlNodePtr pair_node;

	pair_node = find_pair(pairs_node, name);

	if (pair_node == NULL) {
		pair_node = xmlNewChild(pairs_node, NULL, tag, NULL);
		xmlNewChild(pair_node, NULL, NAME_NODE, name);
		xmlNewChild(pair_node, NULL, VALUE_NODE, value);
	} else {
		/*
		 * pair already exists, replace the value with the new value
		 */
		xmlNodeSetContent(pair_node->xmlChildrenNode->next, value);
	}

	return (true);
}

/*
 * Delete a pair with name 'name' from the database representation of
 * a set of <name,value> pairs.
 */
static bool_t
delete_pair(xmlNodePtr pairs_node, xmlChar *name)
{
	xmlNodePtr pair_node;

	pair_node = find_pair(pairs_node, name);
	if (pair_node == NULL)
		return (false);

	xmlUnlinkNode(pair_node);
	xmlFreeNode(pair_node);

	return (true);
}

/*
 * Add a variable <'name', 'value'> pair to the database representation of
 * the variables for LDom 'ldp'. If there is already a variable defined with
 * the same name its value is replaced with the new value.
 */
bool_t
database_ldom_add_variable(ldom_t *ldp, char *name, char *value)
{
	xmlNodePtr variables_node;

	if ((name != NULL) && (strcmp(name, "security-password") == 0))
		return (true);

	variables_node = xml_find_subnode(ldp->dbp->ldom_node, VARIABLES_NODE);

	return (add_pair(variables_node, VAR_NODE, (xmlChar *)name,
	    (xmlChar *)value));
}

/*
 * Delete a single variable 'name' from the database representation of
 * the variables for LDom 'ldp'.
 */
bool_t
database_ldom_del_variable(ldom_t *ldp, char *name)
{
	xmlNodePtr variables_node;

	variables_node = xml_find_subnode(ldp->dbp->ldom_node, VARIABLES_NODE);

	return (delete_pair(variables_node, (xmlChar *)name));
}

/*
 * Get from the old database the properties for the console server instance
 * 'instance'. Returns true if the instance could be found in the database,
 * otherwise false.
 */
bool_t
database_get_vcc_properties(ldom_t *ldp, char *service_name,
    uint64_t *min_port, uint64_t *max_port)
{
	xmlNodePtr ldom_node, vcc_node, node;

	if (old_dbase == NULL)
		return (false);
	ldom_node = find_ldom(old_dbase, (xmlChar *)ldp->namep);
	if (ldom_node == NULL)
		return (false);
	vcc_node = find_node_by_prop(ldom_node, VCC_NODE,
	    SERVICE_NAME_NODE, (xmlChar *)service_name);
	if (vcc_node == NULL)
		return (false);

	/*
	 * If we have a vcc node it must have min_port and max_port
	 * elements since the old database is valid.
	 */
	node = vcc_node->xmlChildrenNode;
	while (node != NULL) {
		if (xmlStrcmp(node->name, MIN_PORT_NODE) == 0) {
			*min_port = (uint64_t)xml_get_long_content(node);
		} else if (xmlStrcmp(node->name, MAX_PORT_NODE) == 0) {
			*max_port = (uint64_t)xml_get_long_content(node);
		}

		node = node->next;
	}

	return (true);
}

/*
 * Given the cfg-handle (instance) of a vio service, return a pointer to the
 * node with the specified prop name. This function is used when Zeus
 * resurrects the state from HV. The service_name is not part of MD, so Zeus
 * needs to get it from the DB by correlating the cfg-handle.
 */
static char *
database_get_prop_value(ldom_t *ldp, xmlChar *type, int instance,
    xmlChar *prop_name)
{
	xmlNodePtr ldom_node, prop_node;
	char *prop_value;
	xmlChar *content = NULL;

	ldom_node = ldom_node_in_old_dbase(ldp);

	if (ldom_node == NULL)
		return (NULL);

	prop_node = find_prop_node_by_instance(ldom_node, (char *)type,
	    instance, prop_name);
	if (prop_node != NULL) {
		if ((content = xmlNodeGetContent(prop_node)) == NULL)
			return (NULL);

		prop_value = Xstrdup((char *)content);
		xmlFree(content);

		return (prop_value);
	}

	return (NULL);
}

/*
 * Get a service name from the DB. Returns NULL if a server of the
 * required type and instance number cannot be found.
 * The caller is responsible for freeing the returned value.
 */
char *
get_service_name_from_db(ldom_t *ldp, char *service_type, int instance)
{
	char *name = database_get_prop_value(ldp, (xmlChar *)service_type,
	    instance, SERVICE_NAME_NODE);

	return (name != NULL ? Xstrdup(name) : NULL);
}

/*
 * The caller is responsible for freeing the returned value.
 */
char *
get_vnet_name_from_db(ldom_t *ldp, int instance)
{
	char *name = database_get_prop_value(ldp, VNET_NODE, instance,
	    VNET_NAME_NODE);

	return (name != NULL ? Xstrdup(name) : NULL);
}

/*
 * The caller is responsible for freeing the returned value.
 */
char *
get_vdisk_name_from_db(ldom_t *ldp, int instance)
{
	char *name = database_get_prop_value(ldp, DISK_NODE, instance,
	    VDISK_NAME_NODE);

	return (name != NULL ? Xstrdup(name) : NULL);
}
/*
 * The caller is responsible for freeing the returned value.
 */
bool_t
get_vdisk_props_from_db(ldom_t *ldp, char *vdisk_name,
    char **vds_name, char **vdsdev_name, int *id)
{
	xmlNodePtr ldom, vdisk_node, prop_node;
	xmlChar *content = NULL;
	*id = NO_NEW_ID;

	if (vds_name)
		*vds_name = NULL;
	if (vdsdev_name)
		*vdsdev_name = NULL;

	if ((ldom = ldom_node_in_old_dbase(ldp)) == NULL)
		return (false);

	/*
	 * Locate the "disk_instance" node for the specified service.
	 */
	if ((vdisk_node = find_node_by_prop(ldom, DISK_NODE, VDISK_NAME_NODE,
	    (xmlChar *)vdisk_name)) == NULL)
		return (false);

	if (vds_name && (prop_node = xml_find_subnode(vdisk_node,
	    SERVICE_NAME_NODE)) != NULL) {
		if ((content = xmlNodeGetContent(prop_node)) != NULL) {
			*vds_name = Xstrdup((char *)content);
			xmlFree(content);
		}
	}

	if (vdsdev_name && (prop_node = xml_find_subnode(vdisk_node,
	    VOLUME_NAME_NODE)) != NULL) {
		if ((content = xmlNodeGetContent(prop_node)) != NULL) {
			*vdsdev_name = Xstrdup((char *)content);
			xmlFree(content);
		}
	}

	if ((prop_node = xml_find_subnode(vdisk_node, ID_NODE)) != NULL) {
		*id = xml_get_int_content(prop_node);
	}

	return (true);
}

/*
 * Get the vsw properties from the DB
 */
bool_t
get_vsw_props_from_db(ldom_t *ldp, char *vsw_name, int *mtu)
{
	xmlNodePtr mtu_node = NULL;

	*mtu = -1;

	if ((db_major == 3 && db_minor >= 3) || (db_major > 3)) {
		mtu_node = find_prop_node_in_old_dbase(ldp,
		    VSW_NODE, SERVICE_NAME_NODE, vsw_name, VSW_MTU_NODE);
		if (mtu_node == NULL) {
			return (false);
		}
		*mtu = xml_get_long_content(mtu_node);

		return (true);
	} else {
		return (false);
	}
}

/*
 * Get the vsw properties from the DB
 */
bool_t
get_vsw_linkprop_from_db(ldom_t *ldp, char *vsw_name, uint64_t *db_pstate)
{
	xmlNodePtr link_node = NULL;

	*db_pstate = ILLEGAL_LNKP;

	if ((db_major == 3 && db_minor >= 5) || (db_major > 3)) {
		link_node = find_prop_node_in_old_dbase(ldp,
		    VSW_NODE, SERVICE_NAME_NODE, vsw_name, LINKPROP_NODE);
		if (link_node == NULL) {
			return (false);
		}
		*db_pstate = xml_get_uint64_content(link_node);

		return (true);
	} else {
		return (false);
	}
}

/*
 * Get the vsw id property from the DB
 */
bool_t
get_vsw_id_from_db(ldom_t *ldp, char *vsw_name, int *id)
{
	xmlNodePtr id_node = NULL;

	*id = NO_NEW_ID;

	if ((db_major == 3 && db_minor >= 6) || (db_major > 3)) {
		id_node = find_prop_node_in_old_dbase(ldp,
		    VSW_NODE, SERVICE_NAME_NODE, vsw_name, ID_NODE);
		if (id_node == NULL) {
			return (false);
		}
		*id = xml_get_int_content(id_node);

		return (true);
	} else {
		return (false);
	}
}


/*
 * Get the vnet properties from the DB
 */
bool_t
get_vnet_props_from_db(ldom_t *ldp, char *vnet_name, int *mtu)
{
	xmlNodePtr mtu_node = NULL;

	*mtu = -1;

	if ((db_major == 3 && db_minor >= 3) || (db_major > 3)) {
		mtu_node = find_prop_node_in_old_dbase(ldp,
		    VNET_NODE, VNET_NAME_NODE, vnet_name, MTU_NODE);
		if (mtu_node == NULL) {
			return (false);
		}
		*mtu = xml_get_long_content(mtu_node);

		return (true);
	} else {
		return (false);
	}
}

/*
 * Get the vnet properties from the DB
 */
bool_t
get_vnet_linkprop_from_db(ldom_t *ldp, char *vnet_name, uint64_t *db_pstate)
{
	xmlNodePtr link_node = NULL;

	*db_pstate = ILLEGAL_LNKP;

	if ((db_major == 3 && db_minor >= 5) || (db_major > 3)) {
		link_node = find_prop_node_in_old_dbase(ldp,
		    VNET_NODE, VNET_NAME_NODE, vnet_name, LINKPROP_NODE);
		if (link_node == NULL) {
			return (false);
		}
		*db_pstate = xml_get_uint64_content(link_node);

		return (true);
	} else {
		return (false);
	}
}

/*
 * Get the vnet id property from the DB
 */
bool_t
get_vnet_id_from_db(ldom_t *ldp, char *vnet_name, int *id)
{
	xmlNodePtr id_node = NULL;

	*id = NO_NEW_ID;

	if ((db_major == 3 && db_minor >= 6) || (db_major > 3)) {
		id_node = find_prop_node_in_old_dbase(ldp,
		    VNET_NODE, VNET_NAME_NODE, vnet_name, ID_NODE);
		if (id_node == NULL) {
			return (false);
		}
		*id = xml_get_int_content(id_node);

		return (true);
	} else {
		return (false);
	}
}

/*
 * Get the reqd num of resources from old database, if not present
 * return 0 (applicable for cpu and mau)
 */
int
database_ldom_get_cpu(ldom_t *ldp)
{
	xmlNodePtr node, ldom_node;

	if ((ldom_node = ldom_node_in_old_dbase(ldp)) == NULL)
		return (-1);

	/*
	 * Locate the node
	 */
	if ((node = xml_find_subnode(ldom_node, CPU_NODE)) == NULL)
		return (-1);

	/*
	 * Set the reqd amt from the DB
	 */
	from_db_add_cpu(node, ldp);

	return (0);
}

int
database_ldom_get_mau(ldom_t *ldp)
{
	xmlNodePtr node, ldom_node;

	if ((ldom_node = ldom_node_in_old_dbase(ldp)) == NULL)
		return (-1);

	/*
	 * Locate the node
	 */
	if ((node = xml_find_subnode(ldom_node, MAU_NODE)) == NULL)
		return (-1);

	/*
	 * Set the reqd amt from the DB
	 */
	from_db_add_mau(node, ldp);

	return (0);
}

/*
 * Get the port and group from the old DB if exists
 * returns true is console node is found, false otherwise.
 * Once an LDom is created, a console node would exist and
 * console_set is set to true and returns true.
 */
bool_t
database_get_console(ldom_t *ldp, int *port, char **group, char **service,
	bool_t *console_set)
{
	xmlNodePtr node, ldom_node, cons_node;
	xmlChar *service_name = NULL;
	xmlChar *group_name = NULL;

	if (port)
		*port = _AUTO_VAL_;
	if (service)
		*service = NULL;
	if (group)
		*group = NULL;
	if (console_set)
		*console_set = false;

	if ((ldom_node = ldom_node_in_old_dbase(ldp)) == NULL)
		return (false);

	cons_node = xml_find_subnode(ldom_node, CONSOLE_NODE);
	if (cons_node == NULL) {
		return (false);
	} else {
		/*
		 * If a console node was present, assume the console
		 * was set with or without any constraint.
		 */
		*console_set = true;
		if (cons_node->xmlChildrenNode == NULL) {
			/*
			 * No constraints specified
			 */
			return (true);
		} else {
			node = cons_node->xmlChildrenNode->xmlChildrenNode;
		}
	}

	if (db_major < 3)
		node = node->next;	/* skip instance */

	if (service) {
		service_name = xmlNodeGetContent(node);
		if (!IS_EMPTY_STR((char *)service_name))
			*service = Xstrdup((char *)service_name);
	}
	node = node->next;
	group_name = xmlNodeGetContent(node);
	if (group) {
		if (!IS_EMPTY_STR((char *)group_name))
			*group = Xstrdup((char *)group_name);
	}

	if (db_major > 2 || (db_major == 2 && db_minor >= 1)) {
		node = node->next;
		*port = xml_get_int_content(node);
	}

	xmlFree(service_name);
	xmlFree(group_name);

	return (true);
}

/*
 * Get the lowest positive free id (such as diskid, networkid, or switchid)
 * from the database. Return 0 if no existing id is found for the given ldom.
 * It can potentially wrap and will only return 0 from then on once it does.
 *
 * E.g. use DISK_NODE as the node_name to get the lowest diskid to use.
 */
int
database_ldom_get_lowest_id(ldom_t *ldp, xmlChar *node_name)
{
	int id = 0;
	char id_str[INT_STR_LEN];

	for (id = 0; id >= 0; id++) {
		snprintf(id_str, sizeof (id_str), "%d", id);
		if (find_node_by_prop(ldp->dbp->ldom_node, node_name,
		    ID_NODE, (xmlChar *)id_str) == NULL)
			break;
	}

	return ((id <= 0) ? 0 : id);
}

/*
 * Check if the given id (such as diskid, networkid, switchid) is already
 * in use. Return true if the id is found.
 *
 * E.g. use DISK_NODE as the node_name to check the diskid for the vdisks
 * in the database.
 */
bool_t
database_ldom_check_dup_id(ldom_t *ldp, xmlChar *node_name, int id)
{
	char id_str[INT_STR_LEN];

	snprintf(id_str, sizeof (id_str), "%d", id);

	/*
	 * Locate the node with the same id
	 */
	if (find_node_by_prop(ldp->dbp->ldom_node, node_name,
	    ID_NODE, (xmlChar *)id_str) == NULL)
		return (false);
	else
		return (true);

}
