/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Module implementing the 'dr_cpu' DS protocol
 */

#include <strings.h>
#include <stddef.h>

#include "basics.h"
#include "dr_cpu.h"
#include "allocate.h"
#include "zprintf.h"
#include "ds.h"
#include "ds_impl.h"
#include "ds_registry.h"
#include "support.h"
#include "fatal.h"
#include "core_md.h"
#include "vcpu.h"
#include "pm_events.h"

static uint64_t dr_cpu_seq = 0;

static ds_ver_t dr_cpu_vers[] = { {1, 1}, {1, 0} };
#define	DR_CPU_NVERS  (sizeof (dr_cpu_vers) / sizeof (dr_cpu_vers[0]))

static ds_capability_t dr_cpu_cap = {
	DR_CPU_SVC,	/* svc_id */
	dr_cpu_vers,	/* vers */
	DR_CPU_NVERS	/* nvers */
};

static ds_ops_t dr_cpu_libops = {
	dslib_register_handler,		/* ds_reg_cb */
	dslib_unregister_handler,	/* ds_unreg_cb */
	dslib_data_handler,		/* ds_data_cb */
	DR_CPU_SVC			/* cb_arg */
};

/*
 * Info needed to invoke ds_clnt_reg/ds_svc_reg when loading plugins
 * There could be multiple services in the same info structure
 */
static libds_info_t ds_cpu_info[] = {
	{
		DS_CLIENT,
		&dr_cpu_cap,
		&dr_cpu_libops
	}
};

static int
dr_cpu_send_request(ldom_t *ldp, uint_t cmd, uint_t ncpuid, uint32_t *cpuidp,
    uint_t flags, ds_msg_t *resp)
{
	size_t		msg_len, data_len;
	ds_req_t	*msg;
	ds_msg_t	ds_msg;
	ds_svc_t	*svcp;
	dr_cpu_msg_t	*dr_cpu_reqp, *dr_cpu_resp;
	uint32_t	resp_type;
	int		idx;
	int		rv = CPU_DR_OP_SUCC;

	svcp = ds_lookup_svc(ldp->ds_chan, DR_CPU_SVC);
	if ((svcp == NULL) || (svcp->hdl == DS_SVC_HDL_INVALID)) {
		DBG(zprintf("%s: %s %s not yet registered\n",
		    __func__, ldp->namep, DR_CPU_SVC));
		return (CPU_DR_OP_FAILED);
	}

	data_len = sizeof (dr_cpu_hdr_t) + ncpuid * sizeof (ds_vcpu_id_t);
	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) + data_len;

	msg = Xmalloc(msg_len);
	DBG7(memset(msg, 0, msg_len));

	ds_fill_data_header(msg, svcp->hdl, data_len);

	/*
	 * Fill in the CPU DR specific message
	 */
	dr_cpu_reqp = (dr_cpu_msg_t *)DS_DATA_BUF(msg);
	dr_cpu_reqp->hdr.msg_type = hton32(cmd);
	dr_cpu_reqp->hdr.num_records = hton32(ncpuid);
	dr_cpu_reqp->hdr.req_num = hton64(dr_cpu_seq++);

	/* copy the cpuids into the message */
	for (idx = 0; idx < ncpuid; idx++) {
		dr_cpu_reqp->dr_cpu_payload.req_cpus[idx] = hton32(cpuidp[idx]);
	}

	ds_msg.buf = msg;
	ds_msg.len = msg_len;
	ds_msg.seq = dr_cpu_reqp->hdr.req_num;

	DBG(zprintf("%s: %s <0x%"PRIx64"> sending dr-cpu request %"PRId64"\n",
	    __func__, ldp->namep, svcp->hdl, ds_msg.seq));

	ds_send_data_msg(ldp->ds_chan, &ds_msg, DR_DS_TIMEOUT, resp);

	Xfree(msg);

	/* a response is expected */
	if (resp->buf == NULL) {
		DBG(zprintf("%s: no response "
		    "received from configure request\n", __func__));
		rv = CPU_DR_OP_NO_RESPONSE;
		goto done;
	}

	dr_cpu_resp = (dr_cpu_msg_t *)((char *)resp->buf +
	    sizeof (ds_data_handle_t));

	/* sanity check the response */
	resp_type = ntoh32(dr_cpu_resp->hdr.msg_type);
	if (resp_type != DR_CPU_OK && resp_type != DR_CPU_ERROR) {
		DBG(zprintf("%s: Unrecognized "
		    "response to dr-cpu request\n", __func__));
		Xfree(resp->buf);
		resp->buf = NULL;
		rv = CPU_DR_OP_NO_RESPONSE;
	}

done:
	return (rv);
}

/*
 * Return the error string associated with the 'n' th cpu in the message.
 * If no string is retuned by the guest for this cpu, return NULL
 */
static char *
dr_cpu_err_string(ds_msg_t *resp, int n)
{
	dr_cpu_msg_t	*dr_cpu_resp =
	    (dr_cpu_msg_t *)((char *)resp->buf + sizeof (ds_data_handle_t));
	int msglen = resp->len - sizeof (ds_data_handle_t);
	int num_records = ntoh32(dr_cpu_resp->hdr.num_records);
	int str_start = offsetof(dr_cpu_msg_t, dr_cpu_payload.resp_cpu_stats) +
	    num_records * sizeof (dr_cpu_stat_t);

	dr_cpu_stat_t *stat = &(dr_cpu_resp->dr_cpu_payload.resp_cpu_stats[n]);
	uint32_t string_off = ntoh32(stat->string_off);
	char *str;

	if (string_off == 0)
		return (NULL);

	/*
	 * Ensure that the string offset is within the message range
	 */
	if (string_off >= msglen || string_off < str_start) {
		warning("Invalid string_off in dr cpu response message "
		    "for the %s th cpu\n", n);
		return (NULL);
	}

	str = (char *)dr_cpu_resp + string_off;

	/*
	 * If there is at least one error string, the last character in
	 * the CPU response message must be 0
	 */
	if (str[msglen - string_off - 1] != '\0') {
		warning("CPU error string not NULL terminated");
		return (NULL);
	}

	if (strlen(str) == 0)
		return (NULL);

	/*
	 * Return the pointer into the offset pointed to by string_off
	 */
	return (str);
}

int
dr_cpu_configure(ldom_t *ldp, uint_t ncpuid, uint32_t *cpuidp,
	bool_t force, uint32_t *op_stats, response_t *rsp)
{
	ds_msg_t	resp;
	uint_t		flags = 0;
	int		rv;
	int		idx;
	dr_cpu_msg_t	*dr_cpu_resp;
	ds_svc_t	*svcp;

	DBG(zprintf("%s...\n", __func__));

	rv = dr_cpu_send_request(ldp, DR_CPU_CONFIGURE, ncpuid, cpuidp,
	    flags, &resp);
	if (rv != CPU_DR_OP_SUCC) {
		DBG(zprintf("%s: error "
		    "sending configure request (%d)\n", __func__, rv));
		if (rv == CPU_DR_OP_NO_RESPONSE) {
			rprintf(rsp, gettext("Request to add cpu(s) sent, but "
			    "no valid response received\n"));
		}
		return (rv);
	}

	/*
	 * If rv == 0, there should be a valid response
	 */
	dr_cpu_resp = (dr_cpu_msg_t *)((char *)resp.buf +
	    sizeof (ds_data_handle_t));

	if ((svcp = ds_lookup_svc(ldp->ds_chan, DR_CPU_SVC)) == NULL) {
		DBG(zprintf("%s...svc lookup failed\n", __func__));
		return (CPU_DR_OP_FAILED);
	}

	/*
	 * If the msg_type is DR_CPU_ERROR, the entire operation
	 * failed, and the payload consists of an error string.
	 *
	 */
	if (VERSION_GTEQ(svcp->version, 1, 1) &&
	    (dr_cpu_resp->hdr.msg_type == DR_CPU_ERROR)) {
		if (dr_cpu_resp->dr_cpu_payload.err_msg != NULL)
			(void) rputs_safe(rsp,
			    dr_cpu_resp->dr_cpu_payload.err_msg,
			    DR_CPU_MAX_ERR_SIZE);
		DBG(zprintf("%s...command failed\n", __func__));
		return (CPU_DR_OP_FAILED);
	}

	/*
	 * If the msg_type is not DR_CPU_OK, the guest probably
	 * did not process the message at all, no additional
	 * payload is to be expected
	 */
	if (dr_cpu_resp->hdr.msg_type != DR_CPU_OK) {
		DBG(zprintf("%s...malformed message\n", __func__));
		Xfree(resp.buf);
		rv = CPU_DR_OP_FAILED;
		return (rv);
	}

	rv = CPU_DR_OP_SUCC;

	for (idx = 0; idx < ntoh32(dr_cpu_resp->hdr.num_records); idx++) {
		dr_cpu_stat_t	*stat;
		uint32_t	cpuid;
		uint32_t	result;

		/* unpack the status message for this CPU */
		stat = &(dr_cpu_resp->dr_cpu_payload.resp_cpu_stats[idx]);
		cpuid = ntoh32(stat->cpuid);
		result = ntoh32(stat->result);

		if (op_stats != NULL)
			op_stats[idx] = result;

		if (result != DR_CPU_RES_OK) {
			char *errstr;
			rprintf(rsp, "Addition of cpu %d failed, error: ",
			    cpuid);
			rv = CPU_DR_OP_PARTIAL;
			errstr = dr_cpu_err_string(&resp, idx);
			(void) rputs_safe(rsp, errstr, DR_CPU_MAX_ERR_SIZE);

		} else {
			DBG(zprintf("%s: cpu %d configured\n",
			    __func__, cpuid));
		}
	}

	if (resp.buf)
		Xfree(resp.buf);

	return (rv);
}

int
dr_cpu_unconfigure(ldom_t *ldp, uint_t ncpuid, uint32_t *cpuidp,
	bool_t force, uint32_t *op_stats, response_t *rsp)
{
	ds_msg_t	resp;
	uint_t		flags = 0;
	int		rv;
	int		idx;
	dr_cpu_msg_t	*dr_cpu_resp;
	ds_svc_t	*svcp;
	uint_t		dr_op =
	    (force) ?  DR_CPU_FORCE_UNCONFIG : DR_CPU_UNCONFIGURE;

	DBG(zprintf("%s:...\n", __func__));

	rv = dr_cpu_send_request(ldp, dr_op, ncpuid, cpuidp, flags, &resp);

	if (rv != CPU_DR_OP_SUCC) {
		DBG(zprintf("%s: error sending "
		    "unconfigure request (%d)\n", __func__, rv));
		if (rv == CPU_DR_OP_NO_RESPONSE) {
			rprintf(rsp, gettext("Request to remove cpu(s) sent, "
			    "but no valid response received\n"));
		}
		return (rv);
	}

	/*
	 * If rv == 0, there should be a valid response
	 */
	dr_cpu_resp = (dr_cpu_msg_t *)((char *)resp.buf +
	    sizeof (ds_data_handle_t));

	if ((svcp = ds_lookup_svc(ldp->ds_chan, DR_CPU_SVC)) == NULL) {
		DBG(zprintf("%s...svc lookup failed\n", __func__));
		return (CPU_DR_OP_FAILED);
	}

	/*
	 * If the msg_type is DR_CPU_ERROR, the entire operation
	 * failed, and the payload consists of an error string.
	 */
	if (VERSION_GTEQ(svcp->version, 1, 1) &&
	    (dr_cpu_resp->hdr.msg_type == DR_CPU_ERROR)) {
		if (dr_cpu_resp->dr_cpu_payload.err_msg != NULL)
			(void) rputs_safe(rsp,
			    dr_cpu_resp->dr_cpu_payload.err_msg,
			    DR_CPU_MAX_ERR_SIZE);
		DBG(zprintf("%s...command failed\n", __func__));
		return (CPU_DR_OP_FAILED);
	}

	/*
	 * If the msg_type is not DR_CPU_OK, the guest probably
	 * did not process the message at all, no additional
	 * payload is to be expected
	 */
	if (dr_cpu_resp->hdr.msg_type != DR_CPU_OK) {
		DBG(zprintf("%s...request failed or "
		    "malformed message\n", __func__));
		rv = CPU_DR_OP_FAILED;
		Xfree(resp.buf);
		return (rv);
	}

	rv = CPU_DR_OP_SUCC;

	for (idx = 0; idx < ntoh32(dr_cpu_resp->hdr.num_records); idx++) {
		dr_cpu_stat_t	*stat;
		uint32_t	cpuid;
		uint32_t	result;
		vcpu_t		*vcpu;

		/* unpack the status message for this CPU */
		stat = &(dr_cpu_resp->dr_cpu_payload.resp_cpu_stats[idx]);
		cpuid = ntoh32(stat->cpuid);
		result = ntoh32(stat->result);

		vcpu = vid_to_vcpu(ldp, cpuid);
		if (vcpu == NULL) {
			warning("Couldn't determine vcpu for vid");
		}

		if (op_stats != NULL)
			op_stats[idx] = result;

		if (result != DR_CPU_RES_OK) {
			char *errstr;
			rprintf(rsp, "Removal of cpu %d failed, error: ",
			    cpuid);
			rv = CPU_DR_OP_PARTIAL;
			errstr = dr_cpu_err_string(&resp, idx);
			(void) rputs_safe(rsp, errstr, DR_CPU_MAX_ERR_SIZE);
		} else {
			DBG(zprintf("%s: cpu %d unconfigured\n",
			    __func__, cpuid));
		}
	}

	if (resp.buf)
		Xfree(resp.buf);

	DBG(zprintf("%s: returned %d\n", __func__, rv));

	return (rv);
}

/*
 * Sends a notification to PM that the cpu DR registration is happening.
 */
void
dr_cpu_reg_handler(ds_svc_t *svc)
{
	pm_domain_event_info_t	*event;
	ldm_err_t		pm_result = { 0 };

	if (pm_domain_dr_reg_event_on) {
		CHAIN_ITER(pm_domain_list.domain_event_list, event, link) {
			if (event->event_type == PM_DOMAIN_DR_REG_EVENT)
				break;
		}

		pm_domain_config_change_notify(event->handle,
		    svc->chan->ldp, GUEST_DOMAIN_DR_REG,
		    0, NULL, 0, &pm_result);

		pm_free_errors(&pm_result);
	}
}

void
dr_cpu_data_handler(ds_svc_t *svc, void *buf, size_t msglen)
{
	ds_msg_tgt_t	tgt;
	ds_msg_t	resp;
	dr_cpu_msg_t	*dr_cpu_resp;
	uint32_t	resp_type;

	if (msglen < sizeof (ds_data_handle_t) + sizeof (dr_cpu_hdr_t)) {
		DBG(zprintf("%s: dropping short dr-cpu response of %d bytes, "
		    "expected >= %d\n", __func__, msglen,
		    sizeof (ds_data_handle_t) + sizeof (dr_cpu_hdr_t)));
		Xfree(buf);
		return;
	}

	dr_cpu_resp = (dr_cpu_msg_t *)((char *)buf + sizeof (ds_data_handle_t));

	/* sanity check the response */
	resp_type = ntoh32(dr_cpu_resp->hdr.msg_type);
	if (resp_type != DR_CPU_OK && resp_type != DR_CPU_ERROR) {
		DBG(zprintf("%s: dropping "
		    "invalid dr-cpu response\n", __func__));
		Xfree(buf);
		return;
	}

	tgt.chan = svc->chan;
	tgt.ds_hdl = svc->hdl;

	resp.buf = buf;
	resp.len = msglen;
	resp.seq = ntoh64(dr_cpu_resp->hdr.req_num);

	/*
	 * Check the pending queue for a matching message and
	 * signal the waiting thread
	 */
	if (ds_check_resp(&tgt, &resp) == DS_MSG_CLAIMED) {
		/* message already handled */
		return;
	}

	DBG(zprintf("%s: dropping dr-cpu response %llu\n", __func__, resp.seq));
	Xfree(buf);
}

/*
 * The versions supported
 */
static prot_version_t dr_cpu_versions[] =
{
	{1, 1}, {1, 0}
};

static svc_ops_t dr_cpu_ops = {
	dr_cpu_reg_handler,		/* reg_handler */
	NULL,				/* unreg_handler */
	dr_cpu_data_handler,		/* data_handler */
	NULL,				/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	int i;
	svc_info_t *info;

	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 2;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(DS_CHAN_NAME);
	info->chan_names[1] = Xstrdup(LIBDS_CHAN_NAME);
	info->num_svcs = 1;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(DR_CPU_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = 0;
	info->num_vers = NUM_VERS(dr_cpu_versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	memcpy(info->versions, dr_cpu_versions,
	    info->num_vers * sizeof (prot_version_t));

	/*
	 * copy the info structure for each service
	 */
	info->libds_svcs = Xmalloc(info->num_svcs * sizeof (libds_info_t *));
	for (i = 0; i < info->num_svcs; i++) {
		info->libds_svcs[i] = cp_dsinfo(&ds_cpu_info[i], info->svcs[i]);
	}

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, DR_CPU_SVC) == 0)
		return (&dr_cpu_ops);
	else
		return (NULL);
}
