/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)dr_crypto.c	1.4	09/02/09 SMI"

/*
 * Module implementing the set of domain services required for
 * DR of crypto units. This encompasses the ability to add and
 * remove MAU and CWQ devices.
 */

#include <strings.h>
#include <stddef.h>

#include "basics.h"
#include "vcpu.h"
#include "dr_crypto.h"
#include "allocate.h"
#include "zprintf.h"
#include "ds.h"
#include "ds_impl.h"
#include "ds_registry.h"
#include "support.h"
#include "fatal.h"
#include "vmau.h"

static uint64_t dr_crypto_cwq_seq = 0;
static uint64_t dr_crypto_mau_seq = 0;

static ds_ver_t dr_crypto_vers[] = { {1, 0} };
#define	DR_CRYPTO_NVERS  (sizeof (dr_crypto_vers) / sizeof (dr_crypto_vers[0]))

static ds_capability_t dr_crypto_cwq_cap = {
	DR_CRYPTO_CWQ_SVC,	/* svc_id */
	dr_crypto_vers,	/* vers */
	DR_CRYPTO_NVERS	/* nvers */
};

static ds_ops_t dr_crypto_cwq_libops = {
	dslib_register_handler,		/* ds_reg_cb */
	dslib_unregister_handler,	/* ds_unreg_cb */
	dslib_data_handler,		/* ds_data_cb */
	DR_CRYPTO_CWQ_SVC		/* cb_arg */
};

static ds_capability_t dr_crypto_mau_cap = {
	DR_CRYPTO_MAU_SVC,	/* svc_id */
	dr_crypto_vers,	/* vers */
	DR_CRYPTO_NVERS	/* nvers */
};

static ds_ops_t dr_crypto_mau_libops = {
	dslib_register_handler,		/* ds_reg_cb */
	dslib_unregister_handler,	/* ds_unreg_cb */
	dslib_data_handler,		/* ds_data_cb */
	DR_CRYPTO_MAU_SVC		/* cb_arg */
};


/*
 * Info needed to invoke ds_clnt_reg/ds_svc_reg when loading plugins
 * There could be multiple services in the same info structure
 */
static libds_info_t ds_crypto_info[] = {
	{
		DS_CLIENT,
		&dr_crypto_cwq_cap,
		&dr_crypto_cwq_libops
	},
	{
		DS_CLIENT,
		&dr_crypto_mau_cap,
		&dr_crypto_mau_libops
	}
};

static int
dr_crypto_send_request(ldom_t *ldp, uint_t cmd, uint_t ncpuid,
    uint32_t *cpuidp, ds_msg_t *resp, crypto_dev_type_t type)
{
	size_t		msg_len;
	size_t		data_len;
	ds_req_t	*msg;
	ds_msg_t	ds_msg;
	ds_svc_t	*svcp;
	dr_crypto_msg_t	*dr_crypto_reqp;
	dr_crypto_msg_t *dr_crypto_resp;
	uint32_t	resp_type;
	int		idx;
	char		*svc_name;
	uint64_t	*seqp;

	switch (type) {
	case CRYPTO_DEV_TYPE_CWQ:
		svc_name = DR_CRYPTO_CWQ_SVC;
		seqp = &dr_crypto_cwq_seq;
		break;

	case CRYPTO_DEV_TYPE_MAU:
		svc_name = DR_CRYPTO_MAU_SVC;
		seqp = &dr_crypto_mau_seq;
		break;

	default:
		DBG(zprintf("Invalid crypto device type: %d\n", type));
		return (CRYPTO_DR_OP_FAILED);
	}

	/* get the requested service structure */
	svcp = ds_lookup_svc(ldp->ds_chan, svc_name);
	if ((svcp == NULL) || (svcp->hdl == DS_SVC_HDL_INVALID)) {
		DBG(zprintf("Service %s is not registered for domain %s\n",
		    svc_name, ldp->namep));
		return (CRYPTO_DR_OP_NO_SVC);
	}

	data_len = sizeof (dr_crypto_hdr_t) + ncpuid * sizeof (ds_crypto_id_t);
	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) + data_len;

	msg = Xmalloc(msg_len);
	DBG7(memset(msg, 0, msg_len));

	ds_fill_data_header(msg, svcp->hdl, data_len);

	/*
	 * Fill in the Crypto DR specific message
	 */
	dr_crypto_reqp = (dr_crypto_msg_t *)DS_DATA_BUF(msg);
	dr_crypto_reqp->hdr.msg_type = hton32(cmd);
	dr_crypto_reqp->hdr.num_records = hton32(ncpuid);
	dr_crypto_reqp->hdr.req_num = hton64((*seqp)++);

	/* copy in the cpuids, each of which represents a crypto unit */
	for (idx = 0; idx < ncpuid; idx++) {
		dr_crypto_reqp->dr_crypto_payload.req_cpus[idx] =
		    hton32(cpuidp[idx]);
	}

	ds_msg.buf = msg;
	ds_msg.len = msg_len;
	ds_msg.seq = dr_crypto_reqp->hdr.req_num;

	DBG(zprintf("%s <0x%"PRIx64"> sending %s request %"PRId64"\n",
	    ldp->namep, svcp->hdl, svc_name, ds_msg.seq));

	ds_send_data_msg(ldp->ds_chan, &ds_msg, DR_DS_TIMEOUT, resp);

	Xfree(msg);

	/* a response is expected */
	if (resp->buf == NULL) {
		DBG(zprintf("no response received from request\n"));
		return (CRYPTO_DR_OP_NO_RESPONSE);
	}

	dr_crypto_resp = (dr_crypto_msg_t *)((char *)resp->buf +
	    sizeof (ds_data_handle_t));

	/* sanity check the response */
	resp_type = ntoh32(dr_crypto_resp->hdr.msg_type);
	if (resp_type != DR_CRYPTO_OK && resp_type != DR_CRYPTO_ERROR) {
		DBG(zprintf("Unrecognized response to %s request\n", svc_name));
		Xfree(resp->buf);
		resp->buf = NULL;
		return (CRYPTO_DR_OP_NO_RESPONSE);
	}

	return (CRYPTO_DR_OP_SUCC);
}

static int
dr_crypto_process_response(ldom_t *ldp, ds_msg_t *ds_resp, uint32_t *op_stats,
    int op, response_t *resp)
{
	int			rv;
	int			idx;
	dr_crypto_msg_t		*dr_crypto_resp;
	dr_crypto_stat_t	*stat;
	vcpu_t			*vcpu;
	uint32_t		cpuid;
	uint16_t		result;
	int			nrec;
	int			nfail;

	DBG(zprintf("%s...\n", __func__));

	dr_crypto_resp = (dr_crypto_msg_t *)((char *)ds_resp->buf +
	    sizeof (ds_data_handle_t));

	/*
	 * If the msg_type is not DR_CRYPTO_OK, the guest probably
	 * did not process the message at all and no additional
	 * payload is to be expected.
	 */
	if (dr_crypto_resp->hdr.msg_type != DR_CRYPTO_OK) {
		DBG(zprintf("%s: complete failure message received from "
		    "the domain\n", __func__));
		Xfree(ds_resp->buf);
		rv = CRYPTO_DR_OP_FAILED;
		return (rv);
	}

	nrec = ntoh32(dr_crypto_resp->hdr.num_records);
	nfail = 0;
	rv = CRYPTO_DR_OP_SUCC;

	for (idx = 0; idx < nrec; idx++) {

		/* unpack the status message for this CPU */
		stat = &(dr_crypto_resp->dr_crypto_payload.resp_cpu_stats[idx]);
		cpuid = ntoh64(stat->cpuid);
		result = ntoh16(stat->result);
		vcpu = vid_to_vcpu(ldp, cpuid);
		ASSERT(vcpu->vmau != NULL);

		if (op_stats != NULL)
			op_stats[idx] = result;

		if (result != DR_CRYPTO_RES_OK) {
			rprintf(resp, "%s of crypto unit %"PRIu64" failed\n",
			    (op == DR_CRYPTO_CONFIGURE) ? "Addition" :
			    "Removal", vcpu->vmau->vid);
			rv = CRYPTO_DR_OP_PARTIAL;
			nfail++;
		} else {
			DBG(zprintf("%s of crypto unit %"PRIu64" successful\n",
			    (op == DR_CRYPTO_CONFIGURE) ? "Addition" :
			    "Removal", vcpu->vmau->vid));
		}
	}

	/*
	 * Fail the entire operation if a failure was reported for
	 * all the requested crypto units.
	 */
	if (nfail == nrec)
		rv = CRYPTO_DR_OP_FAILED;

	if (ds_resp->buf)
		Xfree(ds_resp->buf);

	return (rv);
}

static char *
dr_crypto_op_name(char op)
{
	char *op_name;

	switch (op) {
	case DR_CRYPTO_CONFIGURE:
		op_name = "configure";
		break;
	case DR_CRYPTO_UNCONFIGURE:
		op_name = "unconfigure";
		break;
	case DR_CRYPTO_FORCE_UNCONFIG:
		op_name = "forced unconfigure";
		break;
	default:
		op_name = "UNKNOWN";
		break;
	}

	return (op_name);
}

/*
 * Perform a crypto DR operation against the crypto device linked to
 * the specified vcpu.
 *
 * The current implementation of crypto DR does not utilize the ability
 * to sequence more than one crypto device in a single request. This
 * simplifies error handling when a crypto device consists of both a MAU
 * and CWQ.
 */
static int
dr_crypto_op(uint_t dr_op, ldom_t *ldp, uint32_t cpuid, response_t *resp)
{
	ds_msg_t	ds_resp;
	int		rv_m;
	int		rv_c;
	uint32_t	op_stat;

	DBG(zprintf("%s...\n", __func__));

	/*
	 * Sequence the operation against the MAU.
	 */
	rv_m = dr_crypto_send_request(ldp, dr_op, 1, &cpuid, &ds_resp,
	    CRYPTO_DEV_TYPE_MAU);

	if ((rv_m != CRYPTO_DR_OP_SUCC) && (rv_m != CRYPTO_DR_OP_NO_SVC)) {
		DBG(zprintf("error sending MAU %s request (%d)\n",
		    dr_crypto_op_name(dr_op), rv_m));

		return (rv_m);
	}

	/*
	 * If the MAU domain service was not available, there is no
	 * response to process. However, the overall operation must
	 * continue in case there is a CWQ device present.
	 */
	if (rv_m != CRYPTO_DR_OP_NO_SVC) {
		rv_m = dr_crypto_process_response(ldp, &ds_resp,
		    &op_stat, dr_op, resp);

		/* check if the message could not be processed */
		if (rv_m != CRYPTO_DR_OP_SUCC)
			return (rv_m);

		/* check if the operation failed */
		if (op_stat != DR_CRYPTO_RES_OK)
			return (CRYPTO_DR_OP_FAILED);
	}

	/* check for a CWQ device */
	if (!pri.have_cwq)
		return (rv_m);

	/*
	 * At this point, the MAU operation has succeeded (or was not
	 * necessary) and there is a CWQ to process as well.
	 */
	rv_c = dr_crypto_send_request(ldp, dr_op, 1, &cpuid, &ds_resp,
	    CRYPTO_DEV_TYPE_CWQ);

	if (rv_c != CRYPTO_DR_OP_SUCC) {
		DBG(zprintf("error sending CWQ %s request (%d)\n",
		    dr_crypto_op_name(dr_op), rv_c));

		return (rv_c);
	}

	/* process the response */
	rv_c = dr_crypto_process_response(ldp, &ds_resp, &op_stat, dr_op, resp);
	return (rv_c);
}

int
dr_crypto_configure(ldom_t *ldp, uint_t ncpuid, uint32_t *cpuidp,
    bool_t force, uint32_t *op_stats, response_t *resp)
{
	int	rv;
	int	idx;
	int	ret;

	DBG(zprintf("%s...\n", __func__));

	ret = CRYPTO_DR_OP_SUCC;

	for (idx = 0; idx < ncpuid; idx++) {

		rv = dr_crypto_op(DR_CRYPTO_CONFIGURE, ldp, cpuidp[idx], resp);

		switch (rv) {
		case CRYPTO_DR_OP_SUCC:
			op_stats[idx] = DR_CRYPTO_RES_OK;
			break;

		case CRYPTO_DR_OP_FAILED:
			op_stats[idx] = DR_CRYPTO_RES_FAILURE;
			ret = CRYPTO_DR_OP_PARTIAL;
			break;

		case CRYPTO_DR_OP_NO_RESPONSE:
		case CRYPTO_DR_OP_NO_SVC:
			/* fail immediately */
			return (rv);

		default:
			DBG(zprintf("unexpected error code: %d\n", rv));
			return (rv);
		}
	}

	return (ret);
}

int
dr_crypto_unconfigure(ldom_t *ldp, uint_t ncpuid, uint32_t *cpuidp,
	bool_t force, uint32_t *op_stats, response_t *resp)
{
	int	rv;
	int	idx;
	uint_t	dr_op;
	int	ret;

	DBG(zprintf("%s...\n", __func__));

	dr_op = (force) ? DR_CRYPTO_FORCE_UNCONFIG : DR_CRYPTO_UNCONFIGURE;
	ret = CRYPTO_DR_OP_SUCC;

	for (idx = 0; idx < ncpuid; idx++) {

		rv = dr_crypto_op(dr_op, ldp, cpuidp[idx], resp);

		switch (rv) {
		case CRYPTO_DR_OP_SUCC:
			op_stats[idx] = DR_CRYPTO_RES_OK;
			break;

		case CRYPTO_DR_OP_FAILED:
			op_stats[idx] = DR_CRYPTO_RES_FAILURE;
			ret = CRYPTO_DR_OP_PARTIAL;
			break;

		case CRYPTO_DR_OP_NO_RESPONSE:
		case CRYPTO_DR_OP_NO_SVC:
			/* fail immediately */
			return (rv);

		default:
			DBG(zprintf("unexpected error code: %d\n", rv));
			return (rv);
		}
	}

	return (ret);
}

/*
 * The protocol for the MAU and CWQ services are identical. So,
 * this data handler function is used for both domain services.
 */
void
dr_crypto_data_handler(ds_svc_t *svc, void *buf, size_t msglen)
{
	ds_msg_tgt_t	tgt;
	ds_msg_t	resp;
	dr_crypto_msg_t	*dr_crypto_resp;
	uint32_t	resp_type;

	if (msglen < sizeof (ds_data_handle_t) + sizeof (dr_crypto_hdr_t)) {
		DBG(zprintf("%s: dropping short dr-crypto response of %d "
		    "bytes, expected >= %d\n", __func__, msglen,
		    sizeof (ds_data_handle_t) + sizeof (dr_crypto_hdr_t)));
		Xfree(buf);
		return;
	}

	dr_crypto_resp = (dr_crypto_msg_t *)((char *)buf +
	    sizeof (ds_data_handle_t));

	/* sanity check the response */
	resp_type = ntoh32(dr_crypto_resp->hdr.msg_type);
	if (resp_type != DR_CRYPTO_OK && resp_type != DR_CRYPTO_ERROR) {
		DBG(zprintf("%s: dropping invalid dr-crypto response\n",
		    __func__));
		Xfree(buf);
		return;
	}

	tgt.chan = svc->chan;
	tgt.ds_hdl = svc->hdl;

	resp.buf = buf;
	resp.len = msglen;
	resp.seq = ntoh64(dr_crypto_resp->hdr.req_num);

	/*
	 * Check the pending queue for a matching message and
	 * signal the waiting thread
	 */
	if (ds_check_resp(&tgt, &resp) == DS_MSG_CLAIMED) {
		/* message already handled */
		return;
	}

	DBG(zprintf("%s: dropping dr-crypto response %"PRIu64"\n",
	    __func__, resp.seq));

	Xfree(buf);
}

/*
 * The versions supported
 */
static prot_version_t dr_crypto_versions[] =
{
	{ 1, 0 }
};

/*
 * Both the MAU and CWQ services are identical. So, they
 * can use the same ops structure.
 */
static svc_ops_t dr_crypto_ops = {
	NULL,				/* reg_handler */
	NULL,				/* unreg_handler */
	dr_crypto_data_handler,		/* data_handler */
	NULL,				/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	int i;
	svc_info_t *info;

	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 2;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(DS_CHAN_NAME);
	info->chan_names[1] = Xstrdup(LIBDS_CHAN_NAME);
	info->num_svcs = 2;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(DR_CRYPTO_MAU_SVC);
	info->svcs[1] = Xstrdup(DR_CRYPTO_CWQ_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = 0;
	info->props[1] = 0;
	info->num_vers = NUM_VERS(dr_crypto_versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	memcpy(info->versions, dr_crypto_versions,
	    info->num_vers * sizeof (prot_version_t));

	/*
	 * copy the info structure for each service
	 */
	info->libds_svcs = Xmalloc(info->num_svcs * sizeof (libds_info_t *));
	for (i = 0; i < info->num_svcs; i++)
		info->libds_svcs[i] = cp_dsinfo(&ds_crypto_info[i],
		    info->svcs[i]);

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, DR_CRYPTO_MAU_SVC) == 0) {
		return (&dr_crypto_ops);
	} else if (strcmp(svc_name, DR_CRYPTO_CWQ_SVC) == 0) {
		return (&dr_crypto_ops);
	} else {
		return (NULL);
	}
}
