/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Module implementing the 'dr_mem' DS protocol
 */

#include <sys/types.h>
#include <strings.h>
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <values.h>
#include <sys/sysmacros.h>
#include <sys/time_impl.h>
#include <sys/errno.h>
#include <time.h>

#include "basics.h"
#include "chain.h"
#include "errors.h"
#include "ldom.h"
#include "allocate.h"
#include "database.h"
#include "zprintf.h"
#include "fatal.h"
#include "support.h"
#include "sequencer.h"

#include "ds.h"
#include "ds_impl.h"
#include "ds_registry.h"
#include "dr_mem.h"
#include "xml_event.h"


#define	DR_MEM_DS_CFG_TIMEOUT	(-1)	    /* infinite timeout for cfg/uncfg */
#define	DR_MEM_DS_TIMEOUT	(60)	    /* timeout for non-cfg ops */


#define	DR_MEM_PROG_QUERY_TIME	    2

/*
 * When waiting for an unconfigure response from domain, every 2 seconds, we
 * send a progress query to domain for unconfigured bytes done within the mblk
 * this information is displayed as % of total completion in ldomconfigure.c
 */
typedef struct {
	mutex_t	    mp;
	cond_t	    cvp;		/* signal for request done */
	int	    done;		/* flag for request is done/cancelled */
	ldom_t	    *ldp;
	uint64_t    collected;		/* unconfigured bytes in mblk */
	uint64_t    total;		/* total unconfigured bytes */
	uint64_t    size;		/* requested bytes */
} status_query_t;

static uint64_t dr_mem_seq = 0;

static ds_ver_t dr_mem_vers[] = { {1, 1}, {1, 0} };
#define	DR_MEM_NVERS  (sizeof (dr_mem_vers) / sizeof (dr_mem_vers[0]))

static ds_capability_t dr_mem_cap = {
	DR_MEM_SVC,	/* svc_id */
	dr_mem_vers,	/* vers */
	DR_MEM_NVERS	/* nvers */
};

static ds_ops_t dr_mem_libops = {
	dslib_register_handler,		/* ds_reg_cb */
	dslib_unregister_handler,	/* ds_unreg_cb */
	dslib_data_handler,		/* ds_data_cb */
	DR_MEM_SVC			/* cb_arg */
};

/*
 * Info needed to invoke ds_clnt_reg/ds_svc_reg when loading plugins
 * There could be multiple services in the same info structure.
 */
static libds_info_t ds_mem_info[] = {
	{
		DS_CLIENT,
		&dr_mem_cap,
		&dr_mem_libops
	}
};

#define	DR_MEM_RESP(ds_resp)	((dr_mem_msg_t *)((char *)(ds_resp)->buf + \
	    sizeof (ds_data_handle_t)))

/*
 * Translate domain's message
 */
static int
get_resp_msg_type(int msg_type, ldom_t *ldp)
{
	int rv;

	switch (msg_type) {
	case DR_MEM_OK:
		rv = MEM_DR_OP_SUCC;
		break;

	case DR_MEM_ERROR:
		/*
		 * Domain thinks the request is malformed and
		 * operation is not attempted.
		 */
		DBG(zprintf("%s rejected the mem-dr request as a "
		    "malformed request\n", ldp->namep));
		rv = ERR_MEM_DR_OP_FAILED;
		break;

	default:
		/* garbage in response */
		DBG(zprintf("%s: invalid response for mem-dr"
		    " request\n", ldp->namep));
		rv = ERR_MEM_DR_OP_FAILED;
		break;
	}

	return (rv);
}

static int
dr_mem_send_request(ldom_t *ldp, uint32_t cmd, uint_t nmblk, dr_mblk_t *mblkp,
    ds_msg_t *ds_resp)
{
	size_t		msg_len, data_len;
	ds_req_t	*msg;
	ds_msg_t	ds_msg;
	ds_svc_t	*svcp;
	dr_mem_msg_t	*dr_mem_reqp, *dr_mem_resp;
	uint32_t	resp_type;
	int		idx;
	int		rv;

	svcp = ds_lookup_svc(ldp->ds_chan, DR_MEM_SVC);
	if ((svcp == NULL) || (svcp->hdl == DS_SVC_HDL_INVALID)) {
		DBG(zprintf("%s: %s is not registered\n", ldp->namep,
		    DR_MEM_SVC));
		return (ERR_MEM_DR_OP_FAILED);
	}

	data_len = sizeof (dr_mem_hdr_t) + nmblk * sizeof (dr_mblk_t);
	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) + data_len;

	msg = Xmalloc(msg_len);
	DBG7(memset(msg, 0, msg_len));

	ds_fill_data_header(msg, svcp->hdl, data_len);

	/* fill in the memory DR specific message */
	dr_mem_reqp = (dr_mem_msg_t *)DS_DATA_BUF(msg);
	dr_mem_reqp->hdr.msg_type = hton32(cmd);
	dr_mem_reqp->hdr.msg_arg = hton32(nmblk);
	dr_mem_reqp->hdr.req_num = hton64(dr_mem_seq++);

	DBG(zprintf("%s <0x%"PRIx64"> sending %s request '%c', seq=%"
	    PRId64"\n", ldp->namep, svcp->hdl, DR_MEM_SVC,
	    (char)(dr_mem_reqp->hdr.msg_type), dr_mem_reqp->hdr.req_num));

	/* copy the mblks into the message */
	for (idx = 0; idx < nmblk; idx++) {
		DBG(zprintf("mblk 0x%"PRIx64".%"PRIx64,
		    mblkp[idx].addr, mblkp[idx].size));
		dr_mem_reqp->dr_mem_payload.req_mblks[idx].addr =
		    hton64(mblkp[idx].addr);
		dr_mem_reqp->dr_mem_payload.req_mblks[idx].size =
		    hton64(mblkp[idx].size);
	}

	DBG(zprintf("\n"));

	ds_msg.buf = msg;
	ds_msg.len = msg_len;
	ds_msg.seq = dr_mem_reqp->hdr.req_num;

	/*
	 * Memory configure and unconfigure use infinite timeout since
	 * both operation could take a long time.
	 * Other operations such as query memory list and cancel a memory
	 * unconfigure use normal ds timeout.
	 */
	if ((cmd == DR_MEM_QUERY && nmblk == 0) || cmd == DR_MEM_UNCONF_CANCEL)
		/* query memory map */
		ds_send_data_msg(ldp->ds_chan, &ds_msg, DR_MEM_DS_TIMEOUT,
		    ds_resp);
	else
		ds_send_data_msg(ldp->ds_chan, &ds_msg, DR_MEM_DS_CFG_TIMEOUT,
		    ds_resp);

	Xfree(msg);

	/* expect a response */
	if (ds_resp->buf == NULL) {
		DBG(zprintf("No response received from %s for memory dr "
		    "cmd=%d\n", ldp->namep, cmd));
		return (ERR_MEM_DR_OP_FAILED);
	}

	ASSERT(ds_resp->buf != NULL);
	dr_mem_resp = DR_MEM_RESP(ds_resp);

	/*
	 * If the msg_type is not DR_MEM_OK, the guest probably
	 * did not process the message at all, no additional
	 * payload is to be expected.
	 */

	if ((rv = get_resp_msg_type(dr_mem_resp->hdr.msg_type, ldp)) !=
	    MEM_DR_OP_SUCC) {
		Xfree(ds_resp->buf);
		return (rv);
	}
	return (MEM_DR_OP_SUCC);
}

/*
 * Return the error string associated with the 'n'th mblk in the message.
 * If no string is retuned by the guest for this mblk, return NULL.
 */
static char *
dr_mem_err_string(ds_msg_t *ds_resp, int n)
{
	dr_mem_msg_t	*dr_mem_resp;
	int		msglen;
	int		msg_arg;
	int		str_start;
	dr_mem_stat_t	*stat;
	uint32_t	string_off;
	char *str;

	dr_mem_resp = DR_MEM_RESP(ds_resp);
	msglen = ds_resp->len - sizeof (ds_data_handle_t);
	msg_arg = ntoh32(dr_mem_resp->hdr.msg_arg);
	str_start = offsetof(dr_mem_msg_t, dr_mem_payload.resp_mem_stats) +
	    msg_arg * sizeof (dr_mem_stat_t);
	stat = &(dr_mem_resp->dr_mem_payload.resp_mem_stats[n]);

	string_off = ntoh32(stat->string_off);
	if (string_off == 0)
		return (NULL);

	/* ensure that the string offset is within the message range */
	if (string_off >= msglen || string_off < str_start) {
		warning("Invalid string_off in dr mem response message "
		    "for the %d th mblk\n", n);
		return (NULL);
	}

	str = (char *)dr_mem_resp + string_off;

	/*
	 * If there is at least one error string, the last character in
	 * the memory response message must be 0.
	 */
	if (str[msglen - string_off - 1] != '\0') {
		warning("memory error string not NULL terminated");
		return (NULL);
	}

	if (strlen(str) == 0)
		return (NULL);

	/* return the pointer into the offset pointerd to by string_off */
	return (str);
}

/*
 * Obtain the progress information from domain
 */
static int
dr_mem_ustat(status_query_t *statp)
{
	ds_msg_t	ds_resp;
	int		rv;
	dr_mem_msg_t	*dr_mem_resp;
	dr_mem_ustat_t	*stat;
	mem_dr_state_t	*state;
	time_t		t;
	uint64_t	count;
	ldom_t *ldp = statp->ldp;

	DBG(zprintf("dr_mem_ustat...\n"));

	rv = dr_mem_send_request(ldp, DR_MEM_UNCONF_STATUS, 0, NULL,
	    &ds_resp);

	if (rv != MEM_DR_OP_SUCC) {
		return (rv);
	}

	ASSERT(ds_resp.buf != NULL);
	dr_mem_resp = DR_MEM_RESP(&ds_resp);
	if (ntoh32(dr_mem_resp->hdr.msg_arg) == 0) {
		/* there is no valid data */
		return (MEM_DR_OP_SUCC);
	}

	state = &(ldp->mem_dr_state);

	/* unpack the status message for this mblk */
	stat = &(dr_mem_resp->dr_mem_payload.resp_mem_ustat);

	mutex_lock(&statp->mp);

	statp->collected = ntoh64(stat->collected);
	count = statp->total + statp->collected;
	state->progress = (count * 100) / statp->size;

	mutex_unlock(&statp->mp);

	DBG(zprintf("type='%c' arg=%"PRId32" tot=0x%"PRIx64" mgd=0x%"PRIx64
	    " col=0x%"PRIx64" total=0x%"PRIx64"MB \n",
	    (char)ntoh32(dr_mem_resp->hdr.msg_type),
	    ntoh32(dr_mem_resp->hdr.msg_arg), ntoh64(stat->total),
	    ntoh64(stat->managed), ntoh64(stat->collected),
	    BYTES2MB(statp->total)));
	ASSERT(state->progress <= 100);

	Xfree(ds_resp.buf);

	/* send progess event */
	event_send_progress_event(MEMDR_PROGRESS, ldp, XML_ONGOING);
	return (MEM_DR_OP_SUCC);
}

/*
 * Thread to query progress from domain periodically
 * For each unconfigure request, this thread periodically query the progress.
 * The thread will exit when either the request is done or the request is
 * canceled.
 */
static void *
dr_mem_do_ucfg_stat(void *args)
{
	timespec_t to;
	status_query_t *statp = (status_query_t *)args;
	int rv;

	DBG(zprintf("dr_mem_query_stat thread....\n"));

	mutex_lock(&statp->mp);
	/*
	 * The unconfig_mblk or dr_mem_cancel sets statp->done when
	 * the unconfigure request is done or canceled.
	 * If an unconfigure request is in progress,
	 * every DR_MEM_PROG_QUERY_TIME (2) seconds, a query progress
	 * will be sent to the domain.
	 */
	while (!statp->done) {
		to.tv_sec = DR_MEM_PROG_QUERY_TIME;
		to.tv_nsec = 0;
		rv = cond_reltimedwait(&statp->cvp, &statp->mp, &to);
		if (rv != ETIME) {
			/*
			 * Wake up because of signal (cvp), the request is
			 * either done or canceled.
			 */
			if (!statp->done) {
				DBG(zprintf("Unknown condition_timedwait"
				    " rv =%"PRId16"\n", rv));
			}
			continue;
		}

		mutex_unlock(&statp->mp);

		rv  = dr_mem_ustat(statp);
		if (rv == ERR_MEM_DR_OP_FAILED) {
			thr_exit(NULL);
		}

		mutex_lock(&statp->mp);
	}

	mutex_unlock(&statp->mp);

	return (NULL);

}

/*
 * Initialize and start progress query thread
 */
static status_query_t *
dr_mem_status_query_init(ldom_t *ldp, uint64_t size, thread_t *tid)
{
	status_query_t *statp = Xmalloc(sizeof (status_query_t));

	mutex_init(&statp->mp, USYNC_THREAD, NULL);
	cond_init(&statp->cvp, USYNC_THREAD, NULL);
	statp->done = false;
	statp->ldp = ldp;
	statp->collected = 0;
	statp->total = 0;
	statp->size = size;
	ldp->mem_dr_state.progress = 0;
	ldp->mem_dr_state.statp = statp;
	ldp->mem_dr_state.cancel = false;

	if (thr_create(NULL, NULL, dr_mem_do_ucfg_stat, (void *)statp,
	    NULL, tid) != 0)
		fatal("Could not create dr_mem_ustat thread");

	return (statp);
}

static void
dr_mem_status_query_fini(status_query_t *statp, thread_t tid)
{
	mutex_lock(&statp->mp);
	statp->done = true;
	cond_signal(&statp->cvp);
	mutex_unlock(&statp->mp);
	(void) thr_join(tid, NULL, NULL);
	mutex_destroy(&statp->mp);
	cond_destroy(&statp->cvp);
	Xfree(statp);
}

/*
 * Add memory to a domain
 */
int
dr_mem_configure(ldom_t *ldp, uint_t nmblk, dr_mblk_t *mblkp,
    uint64_t *op_stats)
{
	ds_msg_t	ds_resp;
	int		rv;
	int		idx;
	dr_mem_msg_t	*dr_mem_resp;
	status_query_t	*statp;
	uint64_t	failed_sz = 0;
	int		mblk_added = 0;
	dr_mem_stat_t	*stat;
	uint32_t	result;
	uint32_t	status;
	uint64_t	addr;
	uint64_t	size;
	char		*errstr;

	DBG(zprintf("dr_mem_configure...\n"));

	ASSERT(op_stats);
	rv = dr_mem_send_request(ldp, DR_MEM_CONFIGURE, nmblk, mblkp, &ds_resp);

	if (rv != MEM_DR_OP_SUCC) {
		return (rv);
	}

	ASSERT(ds_resp.buf != NULL);
	dr_mem_resp = DR_MEM_RESP(&ds_resp);

	/* unpack status messages for each added mblk */
	for (idx = 0; idx < ntoh32(dr_mem_resp->hdr.msg_arg); idx++) {

		/* unpack the status message for this mblk */
		stat = &(dr_mem_resp->dr_mem_payload.resp_mem_stats[idx]);
		result = ntoh32(stat->result);
		status = ntoh32(stat->status);


		DBG(zprintf("dr_mem_config: mblk 0x%"PRIx64" %"PRId64
		    " MB configured\n", mblkp[idx].addr,
		    BYTES2MB(mblkp[idx].size)));

		switch (result) {
		case DR_MEM_RES_OK:

			switch (status) {
			case DR_MEM_STAT_CONFIGURED:
				op_stats[idx] = DR_MEM_RES_OK;
				mblk_added++;
				break;

			case DR_MEM_STAT_UNCONFIGURED:
				op_stats[idx] = status;
				break;

			case DR_MEM_STAT_NOT_PRESENT:
			default:
				/* unexpected status - zeus bug */
				goto fatal;

			}

			break;

		case DR_MEM_RES_FAILURE:
		case DR_MEM_RES_BLOCKED:
		case DR_MEM_RES_ERESOURCE:
		{
			/* domain site errors */
			uint64_t	addr;
			uint64_t	size;
			char *errstr;

			op_stats[idx] = result;
			addr = ntoh64(stat->addr);
			size = ntoh64(stat->size);
			errstr = dr_mem_err_string(&ds_resp, idx);

			DBG(zprintf("%s Add mblk 0x%"PRIx64" %"PRIu64
			    "MB failed, error (%d): %s returned addr=%llx "
			    "size=%llx\n",
			    ldp->namep, mblkp[idx].addr,
			    BYTES2MB(mblkp[idx].size),
			    result, errstr != NULL ?  errstr : "unknown",
			    addr, size));
			ldp->mem_dr_state.status = result;
		}
		break;

		case DR_MEM_RES_ENOWORK:
		default:
fatal:
			addr = ntoh64(stat->addr);
			/*
			 * Uunexpected response from domain -
			 * zeus internal bug
			 */
			addr = ntoh64(stat->addr);
			size = ntoh64(stat->size);
			errstr = dr_mem_err_string(&ds_resp, idx);

			zprintf("Add mblk 0x%"PRIx64" %"PRIu64"MB failed,"
			    " error (%d): %s returned addr=%llx size=%llx\n",
			    mblkp[idx].addr, BYTES2MB(mblkp[idx].size),
			    result, errstr != NULL ?  errstr : "unknown",
			    addr, size);
			fatal("unexpected error no work %d\n", result);

			break;

		}
	}

	Xfree(ds_resp.buf);

	if (mblk_added == 0)
		rv = ERR_MEM_DR_OP_FAILED;
	else if (mblk_added < idx)
		rv = ERR_MEM_DR_OP_PARTIAL;

	DBG(zprintf("dr_mem_config.... returns %d\n", rv));

	return (rv);
}

/*
 * Query mblk information
 * This routine is used in two ways:
 * 1. When there is a non-reloc memory error, query where non-reloc
 * memory is located within the mblk.
 * 2. During initialization, if there is a mem dr flag in MD, query domain's
 * memory list and sync zeus memory list with domain's memory list.
 */
static int
dr_mem_query(ldom_t *ldp, uint_t *nmblk, dr_mblk_t *mblkp,
	dr_mem_query_t *op_stats)
{
	ds_msg_t	ds_resp;
	int		rv;
	int		idx;
	dr_mem_msg_t	*dr_mem_resp;

	DBG(zprintf("dr_mem_query...\n"));

	rv = dr_mem_send_request(ldp, DR_MEM_QUERY, *nmblk, mblkp, &ds_resp);
	if (rv != MEM_DR_OP_SUCC) {
		return (rv);
	}

	ASSERT(ds_resp.buf != NULL);
	dr_mem_resp = DR_MEM_RESP(&ds_resp);

	/* unpack the status */
	*nmblk = ntoh32(dr_mem_resp->hdr.msg_arg);
	for (idx = 0; idx < *nmblk; idx++) {
		dr_mem_query_t	*statp, *queryp;

		/* unpack the status message for this mblk */
		queryp = &(dr_mem_resp->dr_mem_payload.resp_mem_query[idx]);
		statp = &op_stats[idx];

		statp->addr = ntoh64(queryp->addr);
		statp->size = ntoh64(queryp->size);
		statp->perm_size = ntoh64(queryp->perm_size);
		statp->first_perm_addr = ntoh64(queryp->first_perm_addr);
		statp->last_perm_addr = ntoh64(queryp->last_perm_addr);

	}

	Xfree(ds_resp.buf);

	return (MEM_DR_OP_SUCC);
}

/*
 * Cancel a memory dr request
 */
int
dr_mem_cancel(ldom_t *ldp)
{
	ds_msg_t	ds_resp;
	int		rv;

	DBG(zprintf("dr_mem_cancel...\n"));

	if (ldp->mem_dr_state.statp == NULL || !ldp->flag.memdr_in_progress ||
	    ldp->state != ACTIVE) {
		return (ERR_MEM_DR_OP_FAILED);
	}

	ldp->mem_dr_state.cancel = true;
	rv = dr_mem_send_request(ldp, DR_MEM_UNCONF_CANCEL, 0, NULL, &ds_resp);

	if (rv != MEM_DR_OP_SUCC)
		return (rv);

	Xfree(ds_resp.buf);

	return (MEM_DR_OP_SUCC);
}

/*
 * Request domain's memory list
 * The domain's memory list is a list (*mblk) of (raddr, size) and #
 * of pairs is returned as *nmblk.
 */
int
dr_get_mem_map(ldom_t *ldp, uint_t *nmblk, dr_mblk_t **mblk)
{
	int		rv;
	ds_msg_t	ds_resp;
	dr_mblk_t	mp;
	dr_mem_query_t *query;
	mblock_t    *mblkp;
	uint_t	    max_nmblks;

	DBG(zprintf("dr_get_mem_map... %s\n", ldp->namep));
	/*
	 * # of mblk in a domain should always <= zeus (HV) knows about.
	 * For add, max # of mblks that a domain adds is the # of mblks
	 * that zeus sends to it. For delete, domain at most did not
	 * unconfigure any mblk and therefore has same mblks as zeus'.
	 */
	max_nmblks = 0;

	CHAIN_ITER(ldp->bound.memory, mblkp, ldom) {
		max_nmblks++;
	}
	ASSERT(max_nmblks > 0);

	mp.addr = 0;
	mp.size = 0;
	query = Xcalloc(max_nmblks, dr_mem_query_t);
	*nmblk = 1;
	rv = dr_mem_query(ldp, nmblk, &mp, query);

	if (rv != MEM_DR_OP_SUCC) {
		Xfree(query);
		return (rv);
	}

	*mblk  = Xcalloc(*nmblk, dr_mblk_t);
	for (int i = 0; i < (*nmblk); i++) {
		(*mblk)[i].addr = query[i].addr;
		(*mblk)[i].size = query[i].size;
	}

	Xfree(query);

	return (MEM_DR_OP_SUCC);
}

#if OBP_RA_ERRATA

/*
 * Check if a mblk has non-reloc memory.
 * This function is only called for OBP special case.
 */
static bool_t
has_perm_mem(ldom_t *ldp, uint64_t raddr, uint64_t size)
{
	dr_mem_query_t *query;
	uint_t nmblk;
	dr_mblk_t mblk;
	uint64_t perm_size;
	int rv;


	mblk.addr = P2ROUNDUP(raddr, DR_MEM_MINALIGN);
	mblk.size = P2ALIGN(size, DR_MEM_MINALIGN);

	/* query */
	query = Xcalloc(1, dr_mem_query_t);
	nmblk = 1;
	rv = dr_mem_query(ldp, &nmblk, &mblk, query);

	perm_size = query->perm_size;
	Xfree(query);

	if (rv == MEM_DR_OP_SUCC && query->perm_size > 0)
		return (true);

	return (false);
}

#endif

/*
 * Unconfigure a mblk that contains non-reloc memory
 * When a mblk contains non-reloc memory, we need to find out the location
 * of the reloc memory within the mblk and then send unconfigure request
 * with (addr, size) of reloc memory.
 *
 * Step 1. Query the location of no-reloc memory within a mblk.
 * Step 2. Analyze the response. There are 4 possible cases:
 *	1. Relocatable memory is at lower end only.
 *	2. Relocatable memory is at upper end only
 *	3. Relocatable memory is at both lower and upper ends
 *	4. Relocatable memory is at the middle of the mblk
 * Step 3. Return true with reloc span(s) for case 1,2,3. return false
 *	for case 4.
 *
 * params:
 * raddr, size - mblk contains non-reloc memory
 * req_size - remaining size of memory to be unconfigured
 * span - (addr, size) the first span of relocatable memory
 * u_span - (addr, size) upper end of reloc memory
 *
 * return:
 * true - found reloc memory within the mblk
 * false - no reloc memory can be unconfigured within the mblk
 */
static bool_t
unconfig_nreloc_mblk(ldom_t *ldp, uint64_t raddr, uint64_t size,
    uint64_t req_size, dr_mblk_t *span, dr_mblk_t *u_span)
{
	int rv;
	uint64_t avail, blk_mem_end, nonreloc;
	uint64_t reloc_start, reloc_end, reloc_sz;
	dr_mem_query_t *query;
	uint_t nmblk;

	/* addr and size must be 256M aligned */
	span->addr = P2ROUNDUP(raddr, DR_MEM_MINALIGN);
	span->size = P2ALIGN(size - (span->addr - raddr), MEM_MINSIZE);

	/* default to no 2nd span */
	u_span->size = 0;

	/* query mblk information */
	query = Xcalloc(1, dr_mem_query_t);
	nmblk = 1;
	rv = dr_mem_query(ldp, &nmblk, span, query);

	if (rv != MEM_DR_OP_SUCC) {
		Xfree(query);
		return (false);
	}

	/*
	 * If there is any relocatable memory at one end, or both
	 * ends of the span, try to unconfigure it; otherwise skip this mblk.
	 */

	DBG(zprintf("query_mem: perm_size=0x%"PRIx64" first_perm_addr=0x%"
	    PRIx64" last_perm_addr=0x%"PRIx64" nonreloc=0x%"PRIx64
	    " \n", query->perm_size,
	    query->first_perm_addr, query->last_perm_addr, nonreloc));

	nonreloc = query->last_perm_addr + 1 - query->first_perm_addr;
	if (nonreloc == 1) {
		/* no relocatable memory. Skip it */
		return (false);
	}

	if (nonreloc > query->perm_size) {
		/*
		 * additional removable memory between the first_perm_addr
		 * and the last_perm_addr
		 * FIXME this case is handled as if all memory in between
		 * first_perm_addr and the last_perm_addr is permanent
		 * memory.
		 */
		DBG(zprintf("%s: There are both permanent memory (%"
		    PRId64") and non-permanent memory (%"PRId64") between "
		    "addr %"PRIx64" and %"PRIx64
		    " . None of those non-permanent "
		    "memory is used for removing memory.\n", ldp->namep,
		    (query->perm_size), (nonreloc - query->perm_size),
		    query->first_perm_addr, query->last_perm_addr));
	}

	/* aligned to 256M */
	avail = P2ALIGN(span->size - nonreloc, MEM_MINSIZE);

	blk_mem_end = span->addr + span->size;

	DBG(zprintf("blk_mem=0x%"PRIx64".0x%"PRIx64"MB \n", span->addr,
	    span->size));

	reloc_start = reloc_end = reloc_sz = 0;
	if (avail < MEM_MINSIZE) {

		/* no relocatable pages available */
		DBG(zprintf("avail is less than MEM_MINSIZE.\n"));
		Xfree(query);
		return (false);
	}

	if (blk_mem_end - 1 == query->last_perm_addr) {

		/* reloc memory at lower end of the mblk */
		DBG(zprintf("removable memory at lower end of the mblk\n"));
		reloc_start = span->addr;
		reloc_end = P2ALIGN(blk_mem_end - nonreloc, DR_MEM_MINALIGN);
		reloc_sz = reloc_end - reloc_start;

		Xfree(query);

		span->size = MIN(reloc_sz, req_size);

		return (true);

	}

	if (span->addr == query->first_perm_addr) {
		/*
		 * There is removable memory at
		 * the upper end of this span.  This could
		 * happen for Solaris if the top part
		 * of the span was allocated before
		 * the cage expanded there.
		 */
		DBG(zprintf("removable memory  at upper end of the span\n"));

		reloc_start = span->addr + nonreloc;
		reloc_sz = avail;
		span->addr = P2ROUNDUP(reloc_start, DR_MEM_MINALIGN);
		if (reloc_sz > req_size) {
			span->addr += reloc_sz - req_size;
			reloc_sz = req_size;
		}
		span->size = reloc_sz;

		Xfree(query);

		return (true);
	}

	/* permanent memory span in the middle */
	DBG(zprintf("removable memory  at both end of the mblk\n"));

	reloc_start = span->addr;
	reloc_end  = P2ALIGN(query->first_perm_addr, DR_MEM_MINALIGN);
	reloc_sz = reloc_end - reloc_start;

	if (reloc_sz >= MEM_MINSIZE) {
		/* fill in lower span */
		span->addr = reloc_start;
		span->size = MIN(reloc_sz, req_size);
		if (span->size >= MEM_MINSIZE) {
			/*
			 * keep the lower end and fill in upper end
			 * to u_span
			 */
			req_size -= span->size;
			if (req_size <= 0) {
				Xfree(query);
				return (true);
			}
			span = u_span;
		}
	}

	/* fill in upper span */
	reloc_start = P2ROUNDUP(query->last_perm_addr + 1, DR_MEM_MINALIGN);
	reloc_end = P2ALIGN(blk_mem_end, DR_MEM_MINALIGN);
	reloc_sz = reloc_end - reloc_start;

	Xfree(query);
	if (reloc_sz < MEM_MINSIZE) {
		if (span != u_span)
			/* both removable span are too small */
			return (false);
		else
			/* upper span is too small */
			return (true);
	}

	/* both lower and upper span have removable memory */
	span->addr = reloc_start;
	span->size = MIN(reloc_sz, req_size);
	if (reloc_sz > req_size)
		span->addr += reloc_sz - req_size;

	return (true);
}

/*
 * Unconfigure one mblk
 * First, we send an unconfigure (addr, size) request to domain.
 * If the request is successful, return true and req_size is adjusted.
 * If the mblk contains non-reloc memory, call unconfig_nreloc_mblk.
 *    If there is reloc span(s) within mblk, call unconfig_mblk recursively to
 *    unconfigure the span(s).
 *
 * If response from domain is EREFUSED or ENOTVIABLE, retry the request with
 *    a smaller size. A side effect of reducing the size is that
 *    the unconfigured memory span can end up in the middle of the original
 *    mblk.For example, suppose the request to unconfigure mblk is retried
 *    with mblk1 and mblk2, where mblk1 is the upper subset of mblk and
 *    mblk2 is the lower subset. if the lower end of mblk1 is actually
 *    unconfigured, this would place the unconfigure span in the middle of
 *    the original mblk. unbind_span in memory.c handles this case.
 *    We can reduce size down to the MINSIZE(256M) by recursively calling
 *    unconfig_mblk. Any mblk is possibly split into 2 smaller mblks. Therefore,
 *    an extra mblk structure is allocated.
 *
 * All other responses from domain return false.
 * During the operation, if request is cancelled, return false.
 * If *req_size = 0, we are done for the entire unconfigure request.
 *
 * param:
 *	raddr, size - mblk
 *	*mblkp, *nmblk - list of mblks successfully unconfigured
 *	*req_size - remaining size to be unconfigured
 *	*statp - progress info
 * return:
 *	true - continue unconfigure memory process
 *	false - stop unconfigure memory process - either done with
 *	requested unconfigure size or non-recoverable error(s).
 */
static bool_t
unconfig_mblk(ldom_t *ldp, uint64_t raddr, uint64_t size,
    dr_mblk_t *mblkp, int *nmblk, int max_nmblks,
    uint64_t *req_size, int *rv,
    status_query_t *statp)
{
	dr_mem_msg_t	*dr_mem_resp;
	ds_msg_t	ds_resp;
	uint16_t	result;
	dr_mem_stat_t	*stat;
	char		*errstr;
	dr_mblk_t	u_span;


	/* addr, size must be aligned with 256M */
	mblkp[*nmblk].addr = P2ROUNDUP(raddr, DR_MEM_MINALIGN);
	size = MIN(P2ALIGN(size -
	    (mblkp[*nmblk].addr - raddr), MEM_MINSIZE), *req_size);

	if (size == 0)
		return (true);

	mblkp[*nmblk].size = size;
	u_span.size = 0;

send:
	DBG(zprintf("req:mp=0x%"PRIx64".0x%"PRIx64" raddr %llx size=%llx\n",
	    mblkp[*nmblk].addr,
	    mblkp[*nmblk].size, raddr, size));

	*rv = dr_mem_send_request(ldp, DR_MEM_UNCONFIGURE, 1, mblkp + (*nmblk),
	    &ds_resp);

	if (*rv != MEM_DR_OP_SUCC) {
		return (false);
	}

	dr_mem_resp = DR_MEM_RESP(&ds_resp);

	/* unpack the status message for this mblk */
	stat = dr_mem_resp->dr_mem_payload.resp_mem_stats;
	result = ntoh16(stat->result);
	Xfree(ds_resp.buf);
	size = mblkp[*nmblk].size;

	switch (result) {
	case DR_MEM_RES_OK:
		/* unconfigure is done for this mblk */
		(*nmblk)++;
		(*req_size) -= size;

		/* update the progress */
		mutex_lock(&statp->mp);

		statp->total += size;
		ldp->mem_dr_state.progress =
		    (statp->total * 100) / statp->size;

		DBG(zprintf("%s remain=%"PRId64" MB mblk sz=%"PRId64
		    " prog = %"PRId16"\n", ldp->namep, BYTES2MB(*req_size),
		    BYTES2MB(size), ldp->mem_dr_state.progress));

		/* send progess event */
		event_send_progress_event(MEMDR_PROGRESS, ldp,
		    XML_ONGOING);

		mutex_unlock(&statp->mp);

		if (*req_size == 0)
			/* memory dr request is successful */
			return (false);

		if (*nmblk == max_nmblks) {
			/* we reach max nmblks allocated for mblkp */
			zprintf("%s allocated nmblks %"PRId16" unconfig %"
			    PRId16"\n", ldp->namep, max_nmblks, *nmblk);
			return (false);
		}
		break;

	/* unconfigure this mblk has error */
	case DR_MEM_RES_ECANCELLED:	    /* operation cancelled */
		*rv = ERR_MEM_DR_OP_CANCELLED;
		return (false);

	case DR_MEM_RES_EREFUSED:	    /* operation refused */
	case DR_MEM_RES_ENOTVIABLE:	    /* VM viability test failed */
		/*
		 * Retry the request with a smaller size. Split the request
		 * into two smaller requests, each half of the original size.
		 * Continue the splitting until the request size is viable or
		 * reaches MINSIZE.
		 */

		if (size > MEM_MINSIZE) {
			uint64_t half_size;
			uint64_t align_raddr;

			align_raddr  = mblkp[*nmblk].addr;
			half_size = mblkp[*nmblk].size / 2;

			if (half_size >= MEM_MINSIZE) {
				/* unconfigure lower half */
				if (unconfig_mblk(ldp, align_raddr,
				    half_size, mblkp, nmblk, max_nmblks,
				    req_size, rv, statp) == false) {
					/* don't continue unconfigure */
					return (false);
				}

				/* unconfigure upper half */
				if (unconfig_mblk(ldp, align_raddr + half_size,
				    half_size, mblkp, nmblk, max_nmblks,
				    req_size, rv, statp) == false) {
					/* don't continue unconfigure */
					return (false);
				}
			}

		}
		break;

	case DR_MEM_RES_EPERM:	/* noreloc error */
		/* find reloc memory within mblk */
		if (unconfig_nreloc_mblk(ldp, raddr, size, *req_size,
		    &mblkp[*nmblk], &u_span) == true) {
			/* found reloc span(s) */
			size = mblkp[*nmblk].size;
			ASSERT(size % MEM_MINSIZE == 0);
			/*
			 * either found reloc memory at one end or both
			 * ends of mblk. We unconfigure the lower end first
			 * and unconfigure upper span after.
			 */
			goto send;
		}

		/* no reloc span  - continue to next mblk */
		return (true);

	/* unconfigure on this mblk failed - continue to next mblk */
	case DR_MEM_RES_EBUSY:
		return (true);

	/* domain side non-recoverable errors */
	case DR_MEM_RES_BLOCKED:
	case DR_MEM_RES_ERESOURCE:
	case DR_MEM_RES_FAILURE:
	case DR_MEM_RES_NOT_IN_MD:  /* MD Download failed */
	case DR_MEM_RES_EFAULT:	  /* domain knows memory, cannot access to it */
		*rv = ERR_MEM_DR_OP_DOMAIN_FAILED;
		/* abort memory dr */
		return (false);

	/* zeus internal errors */
	case DR_MEM_RES_ENOWORK: /*  memory no pages to unconfigure */
	case DR_MEM_RES_ESPAN:	 /*  memory already in use */
	case DR_MEM_RES_EDUP:	 /*  memory span duplicate */
	case DR_MEM_RES_EINVAL:	 /*  invalid argument */
	default:
		/*
		 * These errors indicate logic error(s) or bug(s) in zeus
		 * that should never happen.
		 * We just want to print out information of memory list
		 * for debug purpose.
		 * At this point, most likely, the memory list/allocation for
		 * the domain is corrupted.
		 */

		errstr = dr_mem_err_string(&ds_resp, 0);
		zprintf("unexpected from domain %s (%"PRIx16"): %s\n",
		    ldp->namep, result, errstr != NULL ? errstr : "unknown");

		{
			mblock_t *tmpmblk;
			CHAIN_ITER(ldp->bound.memory, tmpmblk, ldom) {
				zprintf("paddr=%"PRIx64" raddr=%"PRIx64
				    " size=%"PRIx64"\n",
				    tmpmblk->paddr, tmpmblk->raddr,
				    tmpmblk->size);
			}
		}

		zprintf("req:mp=0x%"PRIx64".0x%"PRIx64" raddr %"PRIx64
		    " rv=%"PRId16"\n", mblkp[*nmblk].addr,
		    mblkp[*nmblk].size, raddr, *rv);
		fatal("Zeus - internal error mem-dr!");
	}

	/* check if there is a upper span to be unconfigured */
	if (u_span.size > 0) {
		mblkp[*nmblk].addr = u_span.addr;
		mblkp[*nmblk].size = u_span.size;
		u_span.size = 0;
		size = mblkp[*nmblk].size;
		goto send;
	}

	/* more memory mblk needs to be unconfigured */
	return (true);
}

/*
 * Request to remove memory
 * Before sending requests to domain, we first set up a progress query thread,
 * since remove memory could take a long time.
 * We send one mblk each time to domain for unconfigure until either all mblks
 * have been sent or requested size of memory has been unconfigured.
 * params:
 *  req_size - memory size to be unconfigured
 *  mblkpp, nmblkp - returns with unconfigured spans
 * returns:
 *   ok	 - request was successful
 *   partial - less than requested size of memory was unconfigured
 *   failure - no  requested memory was unconfigured.
 */
int
dr_mem_unconfigure(ldom_t *ldp, uint64_t req_size, uint_t *nmblkp,
	void *mblkpp)
{
	int		rv;
	int		nmblk = 0;
	int		max_nmblk = 0;
	uint64_t	size;
	dr_mem_stat_t	*stat;
	dr_mblk_t	*mblkp;
	mblock_t	*bmp;
	status_query_t	*statp;
	uint64_t	saved_req_size = req_size;
	uint64_t	step = 0;
	thread_t	tid;
	uint64_t	send_raddr;
	uint64_t	send_size;


	DBG(zprintf("dr_mem_unconfigure...\n"));

	/* start progress query thread */
	statp = dr_mem_status_query_init(ldp, req_size, &tid);

	/* max number of span to be unconfigured */
	max_nmblk = ldom_size_bound_mem(ldp) / MEM_MINSIZE;

	/*
	 * a mbp may have upper and lower segments to be unconfigured
	 */
	mblkp = Xcalloc(max_nmblk, dr_mblk_t);
	ASSERT(mblkp != NULL);

	CHAIN_REVERSE_ITER(ldp->bound.memory, bmp, ldom) {

		if (ldp->mem_dr_state.cancel) {
			/* user requested to cancel mem dr */
			rv = ERR_MEM_DR_OP_CANCELLED;
			break;
		}

		/* initialize the progress for this mblk */
		mutex_lock(&statp->mp);
		statp->collected = 0;
		mutex_unlock(&statp->mp);

		send_raddr = bmp->raddr;
		send_size = bmp->size;

#if OBP_RA_ERRATA
		/*
		 * Don't remove the entire mblk if it's RA address is
		 * pri.max_pagesize. (as set during bind).
		 * Do a query to find if there is non-reloc memory in the mblk.
		 * If the mblk has non-reloc memory, don't need to do
		 * anything since the mblk wouldn't be removed in its entirety.
		 * Othersize, keep the lowest page of the mblk to avoid
		 * the entire mblk to be removed.
		 */
		if (send_raddr == pri.max_pagesize &&
		    has_perm_mem(ldp, send_raddr, send_size) == false) {
			ASSERT(send_raddr >> zeus.ra2pa_shift == 0);
			/* no perm memory in segment 0 */
			send_raddr += pri.max_pagesize;
			send_size -= pri.max_pagesize;
		}
#endif

		/* unconfigure this mblk */
		if (unconfig_mblk(ldp, send_raddr, send_size, mblkp,
		    &nmblk, max_nmblk, &req_size, &rv, statp) == false) {
			/* non recoverable error */
			break;
		}

		if (req_size == 0)
			break;
	}

	/* clean up the progress query thread */
	dr_mem_status_query_fini(statp, tid);

	if (nmblkp)
		*nmblkp = nmblk;
	if (mblkp)
		*((dr_mblk_t **)mblkpp) = mblkp;

	if (req_size == saved_req_size) {
		ldp->mem_dr_state.size = 0;
		if (rv != ERR_MEM_DR_OP_CANCELLED &&
		    rv != ERR_MEM_DR_OP_DOMAIN_FAILED) {
			rv = ERR_MEM_DR_OP_FAILED;
		}
	} else if (req_size > 0) {
		if (rv != ERR_MEM_DR_OP_CANCELLED &&
		    rv != ERR_MEM_DR_OP_DOMAIN_FAILED) {
			rv = ERR_MEM_DR_OP_PARTIAL;
		}
		if (rv == ERR_MEM_DR_OP_CANCELLED) {
			DBG(zprintf("%"PRId64"MB of requested %"PRId64
			    "MB  memory has been removed from %s before the "
			    "cancel request\n",
			    BYTES2MB(saved_req_size - req_size),
			    BYTES2MB(saved_req_size), ldp->namep));
		} else {
			if (rv == MEM_DR_OP_SUCC)
				rv = ERR_MEM_DR_OP_PARTIAL;
			DBG(zprintf("%"PRId64"MB of requested %"PRId64
			    "MB memory has been removed from %s\n",
			    BYTES2MB(saved_req_size - req_size),
			    BYTES2MB(saved_req_size),
			    ldp->namep));
		}
		ldp->mem_dr_state.size = saved_req_size - req_size;
	} else
		ldp->mem_dr_state.size = saved_req_size;


	DBG(zprintf("dr_mem_unconfigure: returned %"PRId16"\n", rv));
	return (rv);
}

void
dr_mem_data_handler(ds_svc_t *svc, void *buf, size_t msglen)
{
	ds_msg_tgt_t	tgt;
	ds_msg_t	ds_resp;
	dr_mem_msg_t	*dr_mem_resp;
	uint32_t	resp_type;

	if (msglen < sizeof (ds_data_handle_t) + sizeof (dr_mem_hdr_t)) {
		DBG(zprintf("%s: dropping short %s response of %"PRId16
		    " bytes, expected >= %"PRId16"\n", __func__, DR_MEM_SVC,
		    msglen, sizeof (ds_data_handle_t) + sizeof (dr_mem_hdr_t)));
		Xfree(buf);
		return;
	}

	tgt.chan = svc->chan;
	tgt.ds_hdl = svc->hdl;

	ds_resp.buf = buf;
	ds_resp.len = msglen;

	dr_mem_resp = DR_MEM_RESP(&ds_resp);
	ds_resp.seq = ntoh64(dr_mem_resp->hdr.req_num);

	/* sanity check the response */
	resp_type = ntoh32(dr_mem_resp->hdr.msg_type);
	if (resp_type != DR_MEM_OK && resp_type != DR_MEM_ERROR) {
		DBG(zprintf("%s: dropping "
		    "invalid dr-cpu response\n", __func__));
		Xfree(buf);
		return;
	}

	/*
	 * Check the pending queue for a matching message and
	 * signal the waiting thread.
	 */
	if (ds_check_resp(&tgt, &ds_resp) == DS_MSG_CLAIMED) {
		/* message already handled */
		return;
	}

	zprintf("%s: dropping %s response %"PRId64" %c\n",
	    __func__, DR_MEM_SVC, ds_resp.seq, (char)resp_type);
	Xfree(buf);
}

static void
dr_mem_reg_handler(ds_svc_t *svc)
{
	ldom_t *ldp = svc->chan->ldp;
	if (ldp->mem_dr_state.inprogress == true) {
		/*
		 * In the middle of mem dr, we lost the contact with
		 * the domain. This could happen either the domain rebooted
		 * or ds service or dr-mem driver gone offline. In the later
		 * case, we need to sync memory list with domain.
		 */
		seq_enqueue_work(mem_sync_with_domain, NULL, false);
	}
}
/*
 * The versions supported
 */
static prot_version_t dr_mem_versions[] =
{
	{1, 0}
};

static svc_ops_t dr_mem_ops = {
	dr_mem_reg_handler,		/* reg_handler */
	NULL,				/* unreg_handler */
	dr_mem_data_handler,		/* data_handler */
	NULL,				/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	svc_info_t *info;
	int i;

	info = Xmalloc(sizeof (*info));
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 2;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(DS_CHAN_NAME);
	info->chan_names[1] = Xstrdup(LIBDS_CHAN_NAME);

	info->num_svcs = 1;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(DR_MEM_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = 0;
	info->num_vers = NUM_VERS(dr_mem_versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	memcpy(info->versions, dr_mem_versions,
	    info->num_vers * sizeof (prot_version_t));

	/* copy the info structure for each service */
	info->libds_svcs = Xmalloc(info->num_svcs * sizeof (libds_info_t *));
	for (i = 0; i < info->num_svcs; i++) {
		info->libds_svcs[i] = cp_dsinfo(&ds_mem_info[i], info->svcs[i]);
	}

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, DR_MEM_SVC) == 0)
		return (&dr_mem_ops);
	else
		return (NULL);
}
