/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef _DR_MEM_H
#define	_DR_MEM_H

/*
 * MEM DR Control Protocol
 */
#include <sys/types.h>
#include "ldom.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Common message header
 */
typedef struct {
	uint32_t	msg_type;	/* CONFIG/UNCONFIG etc. */
	uint32_t	msg_arg;
	uint64_t	req_num;
} dr_mem_hdr_t;

/*
 * A mem-dr message is followed by a list of one or
 * more mblks.
 */
typedef struct {
	uint64_t	addr;	/* mblk base address */
	uint64_t	size;	/* mblk size */
} dr_mblk_t;

/*
 * dr-mem status
 */
typedef struct {
	uint64_t	addr;
	uint64_t	size;
	uint32_t	result;
	uint32_t	status;
	uint32_t	string_off;
} dr_mem_stat_t;

typedef struct {
	uint64_t	addr;		/* mblk address in bytes */
	uint64_t	size;		/* mblk size in bytes */
	uint64_t	managed;	/* amount managed by VM */
	uint64_t	perm_size;	/* amount of permanent mem in mblk */
	uint64_t	first_perm_addr;    /* First permanent RA in mblk */
	uint64_t	last_perm_addr;	/* last Permanent RA in mblk */
} dr_mem_query_t;

typedef struct {
	uint64_t	total;		/* total size */
	uint64_t	managed;	/* managed size */
	uint64_t	collected;	/* done when == managed */
} dr_mem_ustat_t;

/*
 * mem-dr message type
 */
typedef struct {
	dr_mem_hdr_t		hdr;
	union {
		dr_mblk_t	req_mblks[1];
		dr_mem_stat_t	resp_mem_stats[1];
		dr_mem_query_t	resp_mem_query[1];
		dr_mem_ustat_t	resp_mem_ustat;
		dr_mblk_t	resp_mem_map[1];
	} dr_mem_payload;
} dr_mem_msg_t;

typedef struct {
	ldom_t	    *ldp;
	uint64_t    size;
	void	   *client;
} mem_dr_args_t;

#define	MEM_DR_OP_SUCC			0x0
/*
 * Memory DR request mesage types
 */

#define	DR_MEM_CONFIGURE	(0x00004d43)    /* 'MC' */
#define	DR_MEM_UNCONFIGURE	(0x00004d55)    /* 'MU' */
#define	DR_MEM_UNCONF_STATUS	(0x00004d53)    /* 'MS' */
#define	DR_MEM_UNCONF_CANCEL	(0x00004d4e)    /* 'MN' */
#define	DR_MEM_QUERY		(0x00004d51)    /* 'MQ' */

/*
 * Response message types
 */
#define	DR_MEM_OK		('o')
#define	DR_MEM_ERROR		('e')

/*
 * Result codes for DR_MEM_OK
 */
#define	DR_MEM_RES_OK			0x0	/* operation succeeded */
#define	DR_MEM_RES_FAILURE		0x1	/* operation failed */
#define	DR_MEM_RES_BLOCKED		0x2	/* operation was blocked */
#define	DR_MEM_RES_NOT_IN_MD		0x3	/* memory not defined in MD */
#define	DR_MEM_RES_ESPAN		0x4	/* memory already in use */
#define	DR_MEM_RES_EFAULT		0x5	/* memory access test failed */
#define	DR_MEM_RES_ERESOURCE		0x6	/* resource not available */
#define	DR_MEM_RES_EPERM		0x7	/* permanenet pages in span */
#define	DR_MEM_RES_EBUSY		0x8	/* memory span busy */
#define	DR_MEM_RES_ENOTVIABLE		0x9	/* VM viability test failed */
#define	DR_MEM_RES_ENOWORK		0xA	/* no pages to unconfigure */
#define	DR_MEM_RES_ECANCELLED		0xB	/* operation cancelled */
#define	DR_MEM_RES_EREFUSED		0xC	/* operation refused */
#define	DR_MEM_RES_EDUP			0xD	/* memory span duplicate */
#define	DR_MEM_RES_EINVAL		0xE	/* invalid argument */

/*
 * Status codes for DR_MEM_OK
 */
#define	DR_MEM_STAT_NOT_PRESENT		0x0
#define	DR_MEM_STAT_UNCONFIGURED	0x1
#define	DR_MEM_STAT_CONFIGURED		0x2

/*
 * Return codes for memory config and unconfig operations
 */

#define	DR_MEM_MAX_MBLK			16
#define	DR_MEM_MINALIGN	    MEM_MINSIZE	/* minimum DR addr/size align */
#define	DR_MEM_MAXSIZE	(0x400000000)	/* max add/del req to domain (16G) */


/*
 * No additional payload is returned when msg_type
 * is DR_MEM_ERROR. This implies a malformed request
 * message and no memory dr related operation was
 * attempted in the guest end
 */

typedef int (*dr_mem_cfg_f)(ldom_t *, uint_t, dr_mblk_t *, uint64_t *);
typedef int (*dr_mem_ufg_f)(ldom_t *, uint64_t, uint_t *, void *);

typedef int (*dr_mem_cancel_f)(ldom_t *);
typedef int (*dr_get_mem_map_f)(ldom_t *, uint_t *, dr_mblk_t **);

#ifdef __cplusplus
}
#endif

#endif /* _DR_MEM_H */
