/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)dr_vio.c	1.4	09/02/09 SMI"

/*
 * Module implementing the 'dr-vio' DS protocol
 */

#include <strings.h>
#include <stddef.h>

#include "basics.h"
#include "dr.h"
#include "dr_vio.h"
#include "errors.h"
#include "allocate.h"
#include "zprintf.h"
#include "ds.h"
#include "ds_impl.h"
#include "ds_registry.h"
#include "support.h"
#include "fatal.h"

static void dr_vio_data_handler(ds_svc_t *svc, void *buf, size_t msglen);

static uint64_t dr_vio_seq = 0;

static ds_ver_t dr_vio_vers[] = { {1, 0} };
#define	DR_VIO_NVERS  (sizeof (dr_vio_vers) / sizeof (dr_vio_vers[0]))

static ds_capability_t dr_vio_cap = {
	DR_VIO_SVC,	/* svc_id */
	dr_vio_vers,	/* vers */
	DR_VIO_NVERS	/* nvers */
};

static ds_ops_t dr_vio_libops = {
	dslib_register_handler,		/* ds_reg_cb */
	dslib_unregister_handler,	/* ds_unreg_cb */
	dslib_data_handler,		/* ds_data_cb */
	DR_VIO_SVC			/* cb_arg */
};

/*
 * Info needed to invoke ds_clnt_reg/ds_svc_reg when loading plugins
 * There could be multiple services in the same info structure
 */
static libds_info_t dr_vio_info[] = {
	{
		DS_CLIENT,
		&dr_vio_cap,
		&dr_vio_libops
	}
};

/*
 * Note: It is the caller's reponsibility to free ds_rsp->buf
 *
 * Returns:
 *	Success:		DR_OP_SUCC
 *	Solaris Failure:	DR_OP_FAILED or DR_OP_NO_RESPONSE
 *	Zeus Failure:		ERR_VIO_DR_yyy
 */
static int
dr_vio_send_request(ldom_t *ldp, uint_t cmd, uint64_t dev_id, char *name,
    ds_msg_t *ds_rsp, response_t *resp)
{
	size_t		msg_len, data_len;
	ds_req_t	*msg;
	ds_msg_t	ds_msg;
	ds_svc_t	*svcp;
	dr_vio_req_t	*dr_vio_reqp;
	dr_vio_res_t	*dr_vio_resp;
	uint32_t	result;
	uint32_t	status;
	int		rv;

	svcp = ds_lookup_svc(ldp->ds_chan, DR_VIO_SVC);
	if ((svcp == NULL) || (svcp->hdl == DS_SVC_HDL_INVALID)) {
		DBG(zprintf("%s: %s not yet registered\n",
		    ldp->namep, DR_VIO_SVC));
		return (ERR_DR_VIO_NOTSUP);
	}

	if (name == NULL) {
		DBG(zprintf("Invalid <name> specified\n"));
		return (ERR_DR_VIO_INVALID_ARG);
	}

	data_len = sizeof (dr_vio_req_t) + strlen(name) + 1;
	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) + data_len;

	msg = Xmalloc((int)msg_len);
	DBG7((void) memset(msg, 0, msg_len));

	ds_fill_data_header(msg, svcp->hdl, (int)data_len);

	/*
	 * Fill in the I/O DR specific message
	 */
	dr_vio_reqp = (dr_vio_req_t *)DS_DATA_BUF(msg);
	dr_vio_reqp->req_num = hton64(dr_vio_seq++);
	dr_vio_reqp->dev_id = hton64(dev_id);
	dr_vio_reqp->msg_type = hton32(cmd);
	(void) strlcpy(dr_vio_reqp->name, name, DR_DEV_TYPE_MAX_LEN);

	ds_msg.buf = msg;
	ds_msg.len = (int)msg_len;
	ds_msg.seq = dr_vio_reqp->req_num;

	DBG(zprintf("%s: <0x%"PRIx64"> sending '%s' request 0x%"PRIx64"\n",
	    ldp->namep, svcp->hdl, DR_VIO_SVC, ds_msg.seq));

	ds_send_data_msg(ldp->ds_chan, &ds_msg, DR_DS_TIMEOUT, ds_rsp);

	Xfree(msg);

	/* a response is expected */
	if (ds_rsp->buf == NULL) {
		DBG(zprintf("%s: no response received from configure "
		    "request\n", __func__));
		return (ERR_DR_VIO_NO_RESPONSE);
	}

	dr_vio_resp =
	    (dr_vio_res_t *)((char *)ds_rsp->buf + sizeof (ds_data_handle_t));

	result	= ntoh32(dr_vio_resp->result);
	status	= ntoh32(dr_vio_resp->status);

	/* sanity check the response */
	switch (result) {
	case DR_VIO_RES_OK:
		/* success */
		rv = 0;
		break;

	case DR_VIO_RES_FAILURE:
		if ((cmd == DR_VIO_CONFIGURE) &&
		    (status == DR_VIO_STAT_UNCONFIGURED)) {
			rv = ERR_DR_VIO_ROLLBACK;
		} else {
			rv = ERR_DR_VIO_FAILED;
		}
		break;

	case DR_VIO_RES_BLOCKED:
		rv = ERR_DR_VIO_IN_USE;
		break;

	case DR_VIO_RES_NOT_IN_MD:
		rv = ERR_DR_VIO_NOT_IN_MD;
		break;

	default:
		DBG(zprintf("Unrecognized response (%d) to '%s' request\n",
		    result, DR_VIO_SVC));
		rv = ERR_DR_VIO_UNK_RESPONSE;
	}

	DBG(zprintf("%s: '%s' request returned result=%d:status=%d\n",
	    __func__, DR_VIO_SVC, rv, status));

	/*
	 * If this is a known error and there is a valid 'reason' string
	 * returned we dump it into the response to be printed by
	 * the CLI.
	 */
	if ((rv != DR_VIO_RES_OK) && (rv != ERR_DR_VIO_UNK_RESPONSE)) {
		char *safe_msg = Xstrldup(dr_vio_resp->reason,
		    DR_VIO_MAX_ERR_SIZE);

		warning("VIO operation on LDom %s failed with reason '%s'",
		    ldp->namep, safe_msg);
		if ((dr_vio_resp->reason != NULL) &&
		    (strcmp(dr_vio_resp->reason, "\0") != 0)) {
			rprintf(resp, MSG_ERR_VIO_DR_REASON);
			rprintf(resp, "%s", safe_msg);
		}

		Xfree(safe_msg);
	}

	return (rv);
}


static int
dr_vio_do_reconfigure(ldom_t *ldp, uint32_t dr_op, uint64_t dev_id, char *name,
    response_t *resp)
{
	ds_msg_t	ds_rsp;
	int		rv;

	DBG(zprintf("%s...\n", __func__));

	rv = dr_vio_send_request(ldp, dr_op, dev_id, name, &ds_rsp, resp);
	if (rv != DR_OP_SUCC) {
		DBG(zprintf("error sending VIO request (%d)\n", rv));
		if (rv == ERR_DR_VIO_NO_RESPONSE) {
			rprintf(resp, gettext("VIO %s request sent, but "
			    "no valid response received\n"),
			    (dr_op == DR_VIO_CONFIGURE) ?
			    "configure" : "unconfigure");
		}
	}

	if (ds_rsp.buf)
		Xfree(ds_rsp.buf);

	return (rv);
}

/*ARGSUSED1*/
int
dr_vio_configure(ldom_t *ldp, uint64_t dev_id, char *name,
	bool_t force, response_t *rsp)
{
	int		rv;

	DBG(zprintf("%s...\n", __func__));

	rv = dr_vio_do_reconfigure(ldp, DR_VIO_CONFIGURE, dev_id, name, rsp);

	return (rv);
}

int
dr_vio_unconfigure(ldom_t *ldp, uint64_t dev_id, char *name,
	bool_t force, response_t *rsp)
{
	int	rv;
	uint_t	dr_op = (force) ? DR_VIO_FORCE_UNCONFIG : DR_VIO_UNCONFIGURE;

	DBG(zprintf("%s...\n", __func__));

	rv = dr_vio_do_reconfigure(ldp, dr_op, dev_id, name, rsp);

	return (rv);
}

static void
dr_vio_data_handler(ds_svc_t *svc, void *buf, size_t msglen)
{
	ds_msg_tgt_t	tgt;
	ds_msg_t	resp;
	dr_vio_res_t	*dr_vio_resp;
	uint32_t	resp_type;

	if (msglen < (sizeof (ds_data_handle_t) + sizeof (dr_vio_res_t))) {
		DBG(zprintf("%s: dropping short dr-vio response of "
		    "0x%x bytes, expected >= 0x%lx\n", __func__, (int)msglen,
		    (sizeof (ds_data_handle_t) + sizeof (dr_vio_res_t))));
		Xfree(buf);
		return;
	}

	dr_vio_resp = (dr_vio_res_t *)((char *)buf + sizeof (ds_data_handle_t));

	/* sanity check the response */
	resp_type = ntoh32(dr_vio_resp->result);
	switch (resp_type) {
	case DR_VIO_RES_OK:
	case DR_VIO_RES_FAILURE:
	case DR_VIO_RES_BLOCKED:
	case DR_VIO_RES_NOT_IN_MD:
		break;

	default:
		DBG(zprintf("%s: dropping invalid dr-vio response (%d)\n",
		    __func__, resp_type));
		Xfree(buf);
		return;
	}

	tgt.chan = svc->chan;
	tgt.ds_hdl = svc->hdl;

	resp.buf = buf;
	resp.len = msglen;
	resp.seq = ntoh64(dr_vio_resp->req_num);

	/*
	 * Check the pending queue for a matching message and
	 * signal the waiting thread
	 */
	if (ds_check_resp(&tgt, &resp) == DS_MSG_CLAIMED) {
		/* message already handled */
		return;
	}

	DBG(zprintf("%s: dropping 'dr-vio' response 0x%"PRIx64"\n",
	    __func__, resp.seq));

	Xfree(buf);
}

/*
 * The versions supported
 */
static prot_version_t dr_vio_versions[] =
{
	{1, 0}
};

static svc_ops_t dr_vio_ops = {
	NULL,				/* reg_handler */
	NULL,				/* unreg_handler */
	dr_vio_data_handler,		/* data_handler */
	NULL,				/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	int i;
	svc_info_t *info;

	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 2;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(DS_CHAN_NAME);
	info->chan_names[1] = Xstrdup(LIBDS_CHAN_NAME);
	info->num_svcs = 1;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(DR_VIO_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = 0;
	info->num_vers = NUM_VERS(dr_vio_versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	(void) memcpy(info->versions, dr_vio_versions,
	    info->num_vers * sizeof (prot_version_t));

	/*
	 * copy the info structure for each service
	 */
	info->libds_svcs = Xmalloc(info->num_svcs * sizeof (libds_info_t *));
	for (i = 0; i < info->num_svcs; i++)
		info->libds_svcs[i] = cp_dsinfo(&dr_vio_info[i], info->svcs[i]);

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, DR_VIO_SVC) == 0)
		return (&dr_vio_ops);
	else
		return (NULL);
}
