/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)ds.c	1.58	09/06/22 SMI"

/*
 * Domain Services
 */

#include <sys/sysmacros.h>
#include <sys/types.h>
#include <stdio.h>
#include <stddef.h>
#include <strings.h>
#include <stdlib.h>
#include <errno.h>
#include <dlfcn.h>

#include <netinet/in.h>
#include <inttypes.h>

#include "ldom.h"
#include "fatal.h"
#include "allocate.h"
#include "zprintf.h"
#include "libmgr.h"
#include "mdesc_impl.h"
#include "ds.h"
#include "ds_impl.h"
#include "sequencer.h"
#include "svcs.h"
#include "zpri.h"
#include "vio.h"

/*
 * Incoming message handling functions.
 */
static int handle_init_req(ds_channel_t *chan, void *buf, size_t len);
static int handle_init_ack(ds_channel_t *chan, void *buf, size_t len);
static int handle_init_nack(ds_channel_t *chan, void *buf, size_t len);
static int handle_reg_req(ds_channel_t *chan, void *buf, size_t len);
static int handle_reg_ack(ds_channel_t *chan, void *buf, size_t len);
static int handle_reg_nack(ds_channel_t *chan, void *buf, size_t len);
static int handle_unreg(ds_channel_t *chan, void *buf, size_t len);
static int handle_unreg_ack(ds_channel_t *chan, void *buf, size_t len);
static int handle_unreg_nack(ds_channel_t *chan, void *buf, size_t len);
static int handle_data(ds_channel_t *chan, void *buf, size_t len);
static int handle_nack(ds_channel_t *chan, void *buf, size_t len);

static ds_channel_t *new_channel(char *name, channel_plugins_t *chan_plugins);
static ds_channel_t *find_channel_by_name(char *name);
static ds_svc_t *find_channel_svc_by_name(ds_channel_t *chan, char *name);
static ds_svc_t *find_channel_svc_by_hdl(ds_channel_t *chan, ds_handle_t hdl);
static void *msg_handler(void *ignored);

/*
 * Message handler lookup table (v1.0 only for now). Future
 * versions can add their own lookup table.
 */
typedef int (*msg_handler_t)(ds_channel_t *chan, void *buf, size_t len);

static const msg_handler_t msg_handlers[] = {
	handle_init_req,		/* DS_INIT_REQ */
	handle_init_ack,		/* DS_INIT_ACK */
	handle_init_nack,		/* DS_INIT_NACK */
	handle_reg_req,			/* DS_REG_REQ */
	handle_reg_ack,			/* DS_REG_ACK */
	handle_reg_nack,		/* DS_REG_NACK */
	handle_unreg,			/* DS_UNREG */
	handle_unreg_ack,		/* DS_UNREG_ACK */
	handle_unreg_nack,		/* DS_UNREG_NACK */
	handle_data,			/* DS_DATA */
	handle_nack			/* DS_NACK */
};

static CHAIN(ds_channel_t, channels) = DEFAULTCHAIN;
static CHAIN(dslib_svc_aggr_t, dslib_svcs) = DEFAULTCHAIN;
static mutex_t dslib_aglock = DEFAULTMUTEX;

/*
 * The DS protocol versions supported.
 */
static prot_version_t ds_versions[] = {
	{ 1, 0 }
};

static int num_ds_versions = sizeof (ds_versions) / sizeof (ds_versions[0]);

extern CHAIN(channel_plugins_t, channel_plugins);

static bool_t ds_do_pruned_pri = false;

/*
 * Find out if pruned PRI should be supported
 * FIXME when PRI versioning is available (CR 6729956), should use pri version
 */
static void
check_pruned_pri_support(void)
{
	int		num, i;
	md_node_t	**nodelist;
	char		*name;
	char		*pri_ver;
	md_t		*mdp;

	/* check if there is a valid pri */
	if (pri.pri == NULL)
		return;

	mdp = md_clone(pri.pri);

	nodelist = Xmalloc(md_node_count(mdp) * sizeof (md_node_t *));
	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "root"),
	    md_find_name(mdp, "fwd"), nodelist);

	if (num != 1)
		fatal("pri does not have signleton root node");

	if (md_get_prop_str(mdp, nodelist[0], "pri-version", &pri_ver)
	    != 0) {
		num = md_scan_dag(mdp, NULL, md_find_name(mdp, "component"),
		    md_find_name(mdp, "fwd"), nodelist);

		for (i = 0; i < num; i++) {

			if (md_get_prop_str(mdp, nodelist[i], "topo-hc-name",
			    &name) != 0) {
				continue;
			}

			if (strcmp(name, "pciexrc") == 0 ||
			    strcmp(name, "niu") == 0) {
				ds_do_pruned_pri = true;
				break;
			}
		}
	} else {
		ds_do_pruned_pri = true;
	}

	Xfree(nodelist);
}

/*
 * Clean up and destroy a service structure.
 * Unloads the service's plugin library.
 */
static void
destroy_svc(ds_svc_t *ds)
{
	mutex_destroy(&ds->lock);
	if (ds->plugin_dlhdl != NULL)
		unload_library(ds->plugin_dlhdl);
	Xfree(ds);
}

void
ds_svc_incref(ds_svc_t *ds)
{
	mutex_lock(&ds->lock);
	ds->refcnt++;
	mutex_unlock(&ds->lock);
}

void
ds_svc_decref(ds_svc_t *ds, uint_t count)
{
	int refcount;

	mutex_lock(&ds->lock);
	ds->refcnt -= count;
	refcount = ds->refcnt;
	ASSERT(refcount >= 0);
	mutex_unlock(&ds->lock);

	if (refcount == 0) {
		/*
		 * The reference count only becomes zero once the service has
		 * been unregistered and all of its outstanding service
		 * requests have been deleted or handled. Once this happens
		 * the service can be safely destroyed.
		 */
		destroy_svc(ds);
	}
}

/*
 * Compare two protocol versions in descending order.
 */
static int
compare_prot_version(const void *a, const void *b)
{
	prot_version_t *ver_a = (prot_version_t *)a;
	prot_version_t *ver_b = (prot_version_t *)b;

	if (ver_a->major_number > ver_b->major_number)
		return (-1);
	else if (ver_a->major_number < ver_b->major_number)
		return (1);

	if (ver_a->minor_number > ver_b->minor_number)
		return (-1);
	else if (ver_a->minor_number < ver_b->minor_number)
		return (1);
	else
		return (0);
}

/*
 * Returns the DS channel to the SP if it exists else NULL.
 */
ds_channel_t *
ds_sp_channel(void)
{
	return (find_channel_by_name(SP_CHAN_NAME));
}

/*
 * Returns true iff the channel is a LDom domain services channel.
 */
bool_t
is_ds_channel(ds_channel_t *chan)
{
	if (strncmp(chan->name, VLDC_DS_SVC_PREFIX,
	    strlen(VLDC_DS_SVC_PREFIX)) == 0)
		return (true);
	else
		return (false);
}

bool_t
is_libds_channel(ds_channel_t *chan)
{
	if (strncmp(chan->name, LIBDS_SVC_PREFIX,
	    strlen(LIBDS_SVC_PREFIX)) == 0)
		return (true);
	else
		return (false);
}


/*
 * First phase initialization of domain services.
 */
void
ds_init1(void)
{
	channel_plugins_t *chan_plugins;
	ds_channel_t *chan;

	check_pruned_pri_support();

	if (thr_create(NULL, NULL, msg_handler, NULL, THR_DAEMON, NULL) != 0)
		efatal("Could not create DS message handler thread");

	ds_comm_init();

	/* sort the DS protocol versions supported in descending order */
	qsort(ds_versions, num_ds_versions, sizeof (prot_version_t),
	    compare_prot_version);

	/*
	 * Create all channels which are needed by any plugged in services,
	 * except for Domain Service channels to logical domains. These are
	 * created when a domain is registered for services via
	 * ds_register_domain().
	 */
	CHAIN_ITER(channel_plugins, chan_plugins, link) {
		if (strcmp(chan_plugins->chan_name, DS_CHAN_NAME) != 0 &&
		    strcmp(chan_plugins->chan_name, LIBDS_CHAN_NAME) != 0) {
			chan = find_channel_by_name(chan_plugins->chan_name);
			if (chan == NULL)
				chan = new_channel(chan_plugins->chan_name,
				    chan_plugins);
		}
	}

	/*
	 * Open the SP channel which is needed for obtaining the PRI.  Once
	 * the PRI is received the rest of the channels will be opened in the
	 * second phase of domain service initialization.
	 */
	CHAIN_ITER(channels, chan, link) {
		if (strcmp(chan->name, SP_CHAN_NAME) == 0) {
			/*
			 * If the channel open fails we take no further
			 * action. Later attempts to use services which may
			 * have been provided over this channel will fail and
			 * action may be taken there.
			 */
			(void) ds_open_channel(chan);
		}
	}
}

/*
 * Second phase domain service initialization. Open all channels except the
 * already opened SP channel.
 */
void
ds_init2(void)
{
	ds_channel_t *chan;

	CHAIN_ITER(channels, chan, link) {
		if (strcmp(chan->name, SP_CHAN_NAME) != 0) {
			/*
			 * If the channel open fails we take no further
			 * action. Later attempts to use services which may
			 * have been provided over this channel will fail and
			 * action may be taken there.
			 */
			(void) ds_open_channel(chan);
		}
	}
}

/*
 * Perform channel state initialization which is common across initial
 * channel creation and channel reset.
 */
static void
reset_channel_state(ds_channel_t *chan)
{
	chan->state = CHAN_DOWN;
	chan->unregister = false;
	chan->ldp = NULL;
	chan->fd = -1;
	chan->buf = NULL;
	chan->reg_svc = NULL;
}

/*
 * Create and initialize a new domain services channel structure for a channel
 * named 'name' with plugged in service 'plugins'.
 */
static ds_channel_t *
new_channel(char *name, channel_plugins_t *plugins)
{
	ds_channel_t *chan;

	chan = Xcalloc(1, ds_channel_t);

	chan->name = Xstrdup(name);
	chan->plugins = plugins;
	chan->versions = ds_versions;
	chan->num_vers = num_ds_versions;
	chan->usecnt = 0;
	chan->msgcnt = 0;
	CHAIN_INIT(chan->ds_svcs);
	mutex_init(&chan->lock, USYNC_THREAD, NULL);
	cond_init(&chan->cond, USYNC_THREAD, NULL);
	CHAIN_INIT(chan->pending.msgs);
	mutex_init(&chan->pending.lock, USYNC_THREAD, NULL);
	cond_init(&chan->pending.cond, USYNC_THREAD, NULL);

	if (strcmp(chan->name, FMA_CHAN_NAME) == 0) {
		chan->props = CHAN_OUTGOING;
	} else {
		chan->props = 0;
	}

	/* do initialization common to first time and reset */
	reset_channel_state(chan);

	/* link this new channel into the list of all created channels */
	CHAIN_ADD(channels, chan, link);

	return (chan);
}

/*
 * Free up a channel structure.
 */
void
ds_free_channel(ds_channel_t *chan)
{
	DBG(zprintf("%s: %s 0x%lx\n", __func__, chan->name, chan));

	CHAIN_UNLINK(channels, chan, link);

	mutex_destroy(&chan->pending.lock);
	cond_destroy(&chan->pending.cond);
	mutex_destroy(&chan->lock);
	cond_destroy(&chan->cond);
	Xfree(chan->name);
	Xfree(chan);
}

/*
 * Register a domain for DS communication. If an LDom has the vldc device path
 * present for DS communication, we use DS_CHAN, otherwise fallback to
 * LIBDS_CHAN. This routine is invoked after an LDom is bound, or if coming up
 * from an existing state, after the DS is linked to the DS server.
 */
int
ds_register_domain(ldom_t *ldp, bool_t open_chan)
{
	channel_plugins_t *ds_plugins;
	ds_channel_t *chan;
	char *name;

	/*
	 * The domain service can be either a VLDC channel or a libds channel,
	 * but not both
	 */
	if (ldp->ds_chan != NULL) {
		/* already registered, nothing to do */
		DBG(zprintf("%s: DS channel named %s exists\n",
		    __func__, ldp->ds_chan->name));
		return (0);
	}

	/*
	 * For any newly created domain, we would set up the libds channel.
	 * But when zeus comes up from an existing config already setup with a
	 * VLDC channel or DS communication we can't use libds
	 */
	ds_plugins = find_channel_plugins_by_name(LIBDS_CHAN_NAME);
	if (ds_plugins != NULL && check_for_libds(ldp) == true) {
		name = zsprintf("%s%s", LIBDS_SVC_PREFIX, ldp->namep);
		DBG(zprintf("%s: creating libds channel - %s\n",
		    __func__, name));
	} else {
		ds_plugins = find_channel_plugins_by_name(DS_CHAN_NAME);
		if (ds_plugins != NULL && check_for_ds_chan(ldp) == true) {
			name = zsprintf("%s%s", VLDC_DS_SVC_PREFIX, ldp->namep);
			DBG(zprintf("%s: creating ds channel - %s\n",
			    __func__, name));
		} else {
			fatal("No plugins found for LDom domain services");
		}
	}

	/*
	 * Create either a DS or a libds channel. Each channel gets a unique
	 * channel name, "ldom-<guestname>" or "libds-<guestname>"
	 */
	chan = new_channel(name, ds_plugins);
	free(name);

	chan->ldp = ldp;
	ldp->ds_chan = chan;

	/*
	 * Open the channel for communication.  If the open fails we
	 * take no further action. Later attempts to use services which may
	 * have been provided over this channel will fail and action may be
	 * taken there.
	 */
	if (open_chan)
		(void) ds_open_channel(chan);

	return (0);
}

/*
 * Unregister a domain for domain services. This is called when a LDom is
 * being unbound.  This routine must only be called from within the sequencer
 * to ensure that no pending actions that could lead to DS messages being sent
 * can occur concurrently with deregistration.
 */
void
ds_unregister_domain(ldom_t *ldp)
{
	ds_svc_t *ds = NULL;
	ds_svc_t *nds;
	dslib_svc_aggr_t *aggr = NULL;
	ds_channel_t *chan = ldp->ds_chan;

	ASSERT_IN_SEQUENCER();

	if (chan == NULL)
		return;

	if (is_libds_channel(chan)) {
		mutex_lock(&chan->lock);
		/*
		 * Mark all services not already being unregistered as part of
		 * the individual service unregister callback.
		 */
		CHAIN_ITER_VOL(chan->ds_svcs, ds, nds, link) {
			mutex_lock(&ds->lock);
			if (ds->unreg != SVC_UNREG_HDL) {
				ds->unreg = SVC_UNREG_LDOM;
				mutex_unlock(&ds->lock);
				channel_error_notify_pending(chan, ds->hdl);
				aggr = ds->svc_aggr;
				mutex_lock(&aggr->lock);
				CHAIN_UNLINK(aggr->dslib_receivers, ds,
				    aggr_link);
				mutex_unlock(&aggr->lock);
				CHAIN_UNLINK(chan->ds_svcs, ds, link);
				(void) libds_ops.ds_unreg_hdl(ds->hdl);
				seq_dequeue_ds_svc(ds, false);
				ds_svc_decref(ds, 1);
			} else {
				mutex_unlock(&ds->lock);
			}
		}

		/*
		 * Wait for any individual unregistration on the handle to
		 * finish
		 */
		while (CHAIN_LENGTH(chan->ds_svcs) > 0)
			cond_wait(&chan->cond, &chan->lock);

		mutex_unlock(&chan->lock);
	}

	ds_unregister_channel(chan);
	ds_free_channel(chan);
	ldp->ds_chan = NULL;
}

int
ds_reset_channel(ds_channel_t *chan)
{
	if (is_ds_channel(chan) || is_libds_channel(chan))
		return (EINVAL);

	ds_unregister_channel(chan);
	reset_channel_state(chan);
	return (ds_open_channel(chan));
}

/*
 * Create a new domain service descriptor for service 'svc' on channel 'chan'.
 */
int
new_ds_svc(ds_channel_t *chan, svc_t *svc, ds_svc_t **ds)
{
	ds_svc_t *d;

	d = find_channel_svc_by_name(chan, svc->name);
	if (d == NULL) {
		d = Xcalloc(1, ds_svc_t);
		d->name = svc->name;
		d->chan = chan;
		d->version.major_number = INVALID_VERSION;
		d->version.minor_number = INVALID_VERSION;
		d->hdl = DS_SVC_HDL_INVALID;
		d->state = SVC_INACTIVE;
		d->svc = svc;
		d->ver_idx = 0;  /* start registration from highest version */
		d->refcnt = 1;
		mutex_init(&d->lock, USYNC_THREAD, NULL);

		*ds = d;
		return (0);
	} else {
		*ds = d;
		return (EEXIST);		/* already registered */
	}
}

/*
 * Add a newly registered service to a channel. This is called when a peer has
 * successfully registered service 'svc' with the LDom manager with the handle
 * 'hdl' and the version given by 'major' and 'minor'.
 */
ds_svc_t *
add_registered_svc(ds_channel_t *chan, svc_t *svc, ds_handle_t hdl,
    uint16_t major, uint16_t minor, int *status)
{
	ds_svc_t *ds;
	int ret;

	DBG(zprintf("%s: Service %s on channel %s with handle %"PRIx64"\n",
	    __func__, svc->name, chan->name, hdl));

	ASSERT(major > 0);

	ret = new_ds_svc(chan, svc, &ds);

	if (status != NULL)
		*status = ret;

	if (ret == EEXIST && ds != NULL) {
		warning("Service %s already on channel %s with handle "
		    "0x%"PRIx64", replacing with handle 0x%"PRIx64"\n",
		    svc->name, chan->name, ds->hdl, hdl);
	}

	if (ds != NULL) {
		ds->version.major_number = major;
		ds->version.minor_number = minor;
		ds->hdl = hdl;
	} else {
		warning("%s failed: Service %s on channel %s with handle "
		    "0x%"PRIx64"\n", __func__, svc->name, chan->name, hdl);
		return (NULL);
	}

	if (ret == 0) {
		DBG(zprintf("Adding service %s handle 0x%"PRIx64" to channel "
		    "%s\n", svc->name, hdl, chan->name));
		load_svc_plugin(ds);

		mutex_lock(&chan->lock);
		CHAIN_ADD(chan->ds_svcs, ds, link);
		mutex_unlock(&chan->lock);
	}

	return (ds);
}

/*
 * Find a channel by name on the list of all channels.
 */
static ds_channel_t *
find_channel_by_name(char *name)
{
	ds_channel_t *chan;

	CHAIN_ITER(channels, chan, link) {
		if (strcmp(chan->name, name) == 0)
			return (chan);
	}

	return (NULL);
}

/*
 * Find a domain service on channel 'chan' with name 'name'.
 */
static ds_svc_t *
find_channel_svc_by_name(ds_channel_t *chan, char *name)
{
	ds_svc_t *ds;

	mutex_lock(&chan->lock);
	CHAIN_ITER(chan->ds_svcs, ds, link) {
		if (strcmp(ds->name, name) == 0)
			break;
	}
	mutex_unlock(&chan->lock);

	return (ds);
}

/*
 * Find a domain service on channel 'chan' with handle 'hdl'.
 */
static ds_svc_t *
find_channel_svc_by_hdl(ds_channel_t *chan, ds_handle_t hdl)
{
	ds_svc_t *ds;

	mutex_lock(&chan->lock);
	CHAIN_ITER(chan->ds_svcs, ds, link) {
		if (ds->hdl == hdl)
			break;
	}
	mutex_unlock(&chan->lock);

	return (ds);
}

/*
 * Among all channels, find the service with the specified handle
 */
static ds_svc_t *
find_channel_svc(ds_handle_t hdl)
{
	ds_channel_t *chan;
	ds_svc_t *ds = NULL;

	CHAIN_ITER(channels, chan, link) {
		if ((ds = find_channel_svc_by_hdl(chan, hdl)) != NULL)
			break;
	}

	return (ds);
}

/*
 * Send a domain services initialization request on channel 'chan'.
 */
void
ds_channel_init_req(ds_channel_t *chan)
{
	ds_hdr_t	*hdr;
	ds_init_req_t	*init_req;
	size_t		nbytes;
	prot_version_t	*ver;

	ver = &chan->versions[chan->ver_idx]; /* version to request */

	nbytes = sizeof (ds_hdr_t) + sizeof (ds_init_req_t);
	hdr = Xmalloc(nbytes);

	hdr->msg_type = hton32(DS_INIT_REQ);
	hdr->payload_len = hton32(sizeof (ds_init_req_t));

	init_req = (ds_init_req_t *)(hdr + 1);
	init_req->major_vers = hton16(ver->major_number);
	init_req->minor_vers = hton16(ver->minor_number);

	DBG(zprintf("%s: >CHAN_INIT_REQ version %d.%d, chan %s\n", __func__,
	    ver->major_number, ver->minor_number, chan->name));

	mutex_lock(&chan->lock);
	ASSERT(chan->state == CHAN_OPEN);
	chan->state = CHAN_INIT_REQ;
	mutex_unlock(&chan->lock);
	if (ds_chan_ready_read(chan) == 0)
		ds_chan_send_msg(chan, hdr, nbytes);

	Xfree(hdr);
}

/*
 * Send a domain services data nack message on channel 'chan' with handle
 * 'bad_hdl'.
 */
static void
send_data_nack(ds_channel_t *chan, ds_handle_t bad_hdl)
{
	caddr_t		msg;
	size_t		msglen;
	ds_hdr_t	*hdr;
	ds_data_nack_t	*nack;

	msglen = sizeof (ds_hdr_t) + sizeof (ds_data_nack_t);
	msg = Xmalloc(msglen);

	hdr = (ds_hdr_t *)msg;
	hdr->msg_type = hton32(DS_NACK);
	hdr->payload_len = hton32(sizeof (ds_data_nack_t));

	nack = (ds_data_nack_t *)(msg + sizeof (ds_hdr_t));
	nack->svc_handle = hton64(bad_hdl);
	nack->result = hton64(DS_INV_HDL);

	/* send the message */
	ds_chan_send_msg(chan, msg, msglen);
	Xfree(msg);
}

/*
 * Handle an incoming domain services data message on channel 'chan'.  The
 * message buffer does not include the domain services protocol header.
 */
static int
handle_data(ds_channel_t *chan, void *buf, size_t len)
{
	ds_data_handle_t	*data = buf;
	ds_svc_t		*ds;
	uint64_t		svc_handle;

	if (len < sizeof (ds_data_handle_t)) {
		Xfree(buf);
		return (-1);		/* data message is too short */
	}

	svc_handle = ntoh64(data->svc_handle);

	DBG(zprintf("%s: <DS_DATA len %u, chan %s, hdl 0x%"PRIx64"\n", __func__,
	    len - sizeof (ds_data_handle_t), chan->name, svc_handle));

	/* lookup the service the data message is for */
	if ((ds = find_channel_svc_by_hdl(chan, svc_handle)) == NULL) {
		DBG(zprintf("%s: no service for handle 0x%"PRIx64" on channel "
		    "%s\n", __func__, svc_handle, chan->name));
		send_data_nack(chan, svc_handle);
		Xfree(buf);
		return (0);
	}

	DBG(zprintf("%s: message for service %s on channel %s\n",
	    __func__, ds->name, chan->name));

	/*
	 * Dispatch this message to the service for handling. The service is
	 * responsible for freeing it.
	 */
	ds->ops->data_handler(ds, buf, len);

	return (0);
}


ldom_t *
find_ldom_by_dhdl(uint64_t id)
{
	ldom_t *ldp = NULL;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->idp != NULL && ldp->idp->id == id)
			break;
	}

	return (ldp);
}

/*
 * Given the ds_hdl_t, find the recipient of the message (the channel
 * and the service). Even if this handler is invoked, it's possible that
 * the registration is on a vldc port because the clnt_reg()/svc_reg()
 * were done before a REG_REQ was sent out on the vldc port.
 */
int
find_libds_chan_svc(ds_hdl_t hdl, ds_domain_hdl_t dhdl, ds_cb_arg_t arg,
    svc_t **svcp, ds_channel_t **chanp)
{
	char *svc_id = (char *)arg;
	ds_channel_t *chan;
	char ldom_name[MAXLEN_OBJECT_NAME], *chan_name;
	ldom_t *ldp;
	svc_t *svc;

	if (dhdl == -1 && libds_ops.ds_domain_lookup(hdl, &dhdl) != 0) {
		DBG(zprintf("%s: ds_domain_lookup failed on hdl 0x%"PRIx64", "
		    "svc %s\n", __func__, hdl, svc_id));
		return (-1);
	}

	if (libds_ops.ds_dom_hdl_to_name(dhdl, ldom_name,
	    MAXLEN_OBJECT_NAME) != 0) {
		DBG(zprintf("Domain handle 0x%"PRIx64" not valid\n", dhdl));
		ldp = find_ldom_by_dhdl(dhdl);
	} else {
		ldp = find_ldom_by_name(ldom_name);
	}

	if (ldp == NULL) {
		DBG(zprintf("%s: find_ldom_by_dhdl: handle 0x%"PRIx64" - Domain"
		    " handle 0x%"PRIx64" not valid\n", __func__, hdl, dhdl));
		return (-1);
	}
	if (ldp->state <= BOUND) {
		DBG(zprintf("%s: Invalid Domain\n", __func__));
		return (-1);
	}

	/*
	 * Check if zeus is using a different channel
	 */
	if (ldp->ds_chan != NULL && !is_libds_channel(ldp->ds_chan)) {
		DBG(zprintf("%s: Ignoring libds registration, LDom  = %s,"
		    "hdl 0x%"PRIx64", not libds channel\n", __func__,
		    ldp->namep, hdl));
		return (-1);
	}

	chan_name = zsprintf("%s%s", LIBDS_SVC_PREFIX, ldp->namep);

	if ((chan = find_channel_by_name(chan_name)) == NULL) {
		Xfree(chan_name);
		return (-1);
	} else {
		*chanp = chan;
	}

	Xfree(chan_name);

	svc = find_plugged_svc_by_name(chan->plugins, svc_id);

	if (svc == NULL) {
		DBG(zprintf("%s: Unidentfied service %s\n", __func__, svc_id));
		return (-1);
	} else {
		*svcp = svc;
	}

	return (0);
}

/*
 * Handle an incoming domain services data message on channel 'chan'.  The
 * message buffer does not include the domain services protocol header.
 * Currently nothing is done with data nacks.
 */
static int
handle_nack(ds_channel_t *chan, void *buf, size_t len)
{
	ds_data_nack_t *nack = buf;

	DBG(zprintf("%s: <DS_NACK result 0x%"PRIx64", chan %s, hdl "
	    "0x%"PRIx64"\n", __func__, ntoh64(nack->result), chan->name,
	    ntoh64(nack->svc_handle)));
	Xfree(buf);

	return (0);
}

static void
send_unreg_ack(ds_channel_t *chan, uint64_t svc_handle)
{
	ds_unreg_ack_t	*ack;
	ds_hdr_t	*hdr;
	char		*msg;
	size_t		msglen;

	DBG(zprintf("%s: >DS_UNREG_ACK chan %s, hdl 0x%"PRIx64"\n",
	    __func__, chan->name, svc_handle));

	msglen = sizeof (ds_hdr_t) + sizeof (ds_unreg_ack_t);
	msg = Xmalloc(msglen);

	hdr = (ds_hdr_t *)msg;
	hdr->msg_type = hton32(DS_UNREG_ACK);
	hdr->payload_len = hton32(sizeof (ds_unreg_ack_t));

	ack = (ds_unreg_ack_t *)(msg + sizeof (ds_hdr_t));
	ack->svc_handle = hton64(svc_handle);

	ds_chan_send_msg(chan, msg, msglen);
	Xfree(msg);
}

static void
send_unreg_nack(ds_channel_t *chan, uint64_t svc_handle)
{
	ds_unreg_nack_t	*nack;
	ds_hdr_t	*hdr;
	char		*msg;
	size_t		msglen;

	DBG(zprintf("%s: >DS_UNREG_NACK chan %s, hdl 0x%"PRIx64"\n",
	    __func__, chan->name, svc_handle));

	msglen = sizeof (ds_hdr_t) + sizeof (ds_unreg_nack_t);
	msg = Xmalloc(msglen);

	hdr = (ds_hdr_t *)msg;
	hdr->msg_type = hton32(DS_UNREG_NACK);
	hdr->payload_len = hton32(sizeof (ds_unreg_nack_t));

	nack = (ds_unreg_nack_t *)(msg + sizeof (ds_hdr_t));
	nack->svc_handle = hton64(svc_handle);
	nack->status = hton64(DS_INV_HDL);

	ds_chan_send_msg(chan, msg, msglen);
	Xfree(msg);
}

/*
 * Handle an incoming service unregistration message on channel 'chan'.  The
 * message buffer does not include the domain services protocol header.
 */
static int
handle_unreg(ds_channel_t *chan, void *buf, size_t len)
{
	ds_unreg_req_t	*req = buf;
	ds_svc_t	*ds;
	uint64_t	svc_handle;

	if (len != sizeof (ds_unreg_req_t)) {
		Xfree(buf);
		return (-1);		/* message length invalid */
	}

	svc_handle = ntoh64(req->svc_handle);

	DBG(zprintf("%s: <DS_UNREG chan %s, hdl 0x%"PRIx64"\n",
	    __func__, chan->name, svc_handle));

	/* lookup the service the data message is for */
	if ((ds = find_channel_svc_by_hdl(chan, svc_handle)) == NULL) {
		DBG(zprintf("%s: no service for handle 0x%"PRIx64" on "
		    "channel %s\n", __func__, svc_handle, chan->name));
		send_unreg_nack(chan, svc_handle);
	} else {
		/*
		 * Remove service from list of active services. If the client
		 * is misbehaving and sending more messages for the service
		 * after this unregister request then they will be dropped
		 * since the service lookup will now fail.
		 */
		mutex_lock(&chan->lock);
		CHAIN_UNLINK(chan->ds_svcs, ds, link);
		mutex_unlock(&chan->lock);

		/* dequeue any pending sequencer requests for the service */
		seq_dequeue_ds_svc(ds, true);

		/*
		 * There may be a request for this service being handled
		 * concurrently in the sequencer, hence we just decrement the
		 * reference count for the service. The storage for the
		 * service will be destroyed when the reference count reaches
		 * zero.
		 */
		ds_svc_decref(ds, 1);

		send_unreg_ack(chan, svc_handle);
	}

	Xfree(buf);
	return (0);
}

/*
 * Handle an incoming domain services unregistration ack message on channel
 * 'chan'.  The message buffer does not include the domain services protocol
 * header.  Currently nothing is done here as the LDom manager does not send
 * unregistration requests.
 */
static int
handle_unreg_ack(ds_channel_t *chan, void *buf, size_t len)
{
	ds_unreg_ack_t	*ack = buf;

	DBG(zprintf("%s: <DS_UNREG_ACK chan %s, hdl=0x%"PRIx64"\n", __func__,
	    chan->name, ntoh64(ack->svc_handle)));
	Xfree(buf);

	return (0);
}

/*
 * Handle an incoming domain services unregistration ack message on channel
 * 'chan'.  The message buffer does not include the domain services protocol
 * header.  Currently nothing is done here as the LDom manager does not send
 * unregistration requests.
 */
static int
handle_unreg_nack(ds_channel_t *chan, void *buf, size_t len)
{
	ds_unreg_nack_t	*nack = buf;

	DBG(zprintf("%s: <DS_UNREG_NACK chan %s, hdl=0x%"PRIx64"\n", __func__,
	    chan->name, ntoh64(nack->svc_handle)));
	Xfree(buf);

	return (0);
}

/*
 * Send a service registration request for domain service 'ds'.
 */
static void
send_reg_req(ds_svc_t *ds)
{
	size_t		idlen;
	ds_hdr_t	*hdr;
	caddr_t		msg;
	size_t		msglen;
	ds_reg_req_t	*req;
	prot_version_t	*ver;

	DBG(zprintf("%s, svc: %s, guest %s\n", __func__, ds->name,
	    ds->chan->ldp == NULL ? "N/A" : ds->chan->ldp->namep));

	ASSERT(ds->state == SVC_INACTIVE);

	idlen = strlen(ds->name);
	msglen = sizeof (ds_hdr_t) + sizeof (ds_reg_req_t) + idlen;
	msg = Xmalloc(msglen);

	hdr = (ds_hdr_t *)msg;
	hdr->msg_type = hton32(DS_REG_REQ);
	hdr->payload_len = sizeof (ds_reg_req_t) + idlen;

	req = (ds_reg_req_t *)(msg + sizeof (ds_hdr_t));
	req->svc_handle = hton64(ds->hdl);

	DBG(zprintf("%s: %s -hdl =  0x%"PRIx64"\n", __func__, ds->name,
	    req->svc_handle));
	ver = ds->svc->versions[ds->ver_idx]; /* version being requested */
	req->major_vers = hton16(ver->major_number);
	req->minor_vers = hton16(ver->minor_number);

	/* copy in the service id (aka name) */
	strcpy(req->svc_id, ds->name);

	ds->state = SVC_REG_PENDING;

	DBG(zprintf("%s: >DS_REG_REQ svc %s version %d.%d, chan %s, "
	    "hdl 0x%"PRIx64"\n", __func__, ds->name, ver->major_number,
	    ver->minor_number, ds->chan->name, ds->hdl));

	ds_chan_send_msg(ds->chan, msg, msglen);
	Xfree(msg);
}

/*
 * Create the next outgoing domain service on channel 'chan' and start its
 * registration.
 */
void
ds_channel_reg_req(ds_channel_t *chan)
{
	ds_svc_t *ds;

	if (chan->reg_svc == NULL) {
		DBG(zprintf("%s: No svc: %s@%s\n", __func__, chan->name,
		    chan->ldp == NULL ? "N/A" : chan->ldp->namep));
		return;			/* all done */
	}

	DBG(zprintf("%s: %s@%s, svc %s\n", __func__, chan->name,
	    chan->ldp == NULL ? "N/A" : chan->ldp->namep, chan->reg_svc->name));

	if (new_ds_svc(chan, chan->reg_svc, &ds) == 0) {
		ds->ver_idx = 0;	/* start from highest version */
		ds->hdl = chan->reg_svc->hdl_base++; /* bump up handle */
		mutex_lock(&chan->lock);
		CHAIN_ADD(chan->ds_svcs, ds, link);
		mutex_unlock(&chan->lock);
		send_reg_req(ds);
	} else {
		warning("Service %s already on channel %s with handle "
		    "0x%"PRIx64"", chan->reg_svc->name, chan->name, ds->hdl);
	}

	DBG(zprintf("%s: %s@%s, svc %s - reg sent\n", __func__, chan->name,
	    chan->ldp == NULL ? "N/A" : chan->ldp->namep, chan->reg_svc->name));

	/* advance to next service to register */
	chan->reg_svc = CHAIN_NEXT(chan->reg_svc->outlink);
}

/*
 * Send a service registration nack.
 */
static void
send_reg_nack(ds_channel_t *chan, uint64_t svc_handle, uint64_t status,
    uint16_t major)
{
	ds_reg_nack_t	*nack;
	ds_hdr_t	*hdr;
	char		*msg;
	size_t		msglen;

	DBG(zprintf("%s: >DS_REG_NACK status %"PRIu64", major %u, chan %s, "
	    "hdl 0x%"PRIx64"\n", __func__, status, major, chan->name,
	    svc_handle));

	msglen = sizeof (ds_hdr_t) + sizeof (ds_reg_nack_t);
	msg = Xmalloc(msglen);

	hdr = (ds_hdr_t *)msg;
	hdr->msg_type = hton32(DS_REG_NACK);
	hdr->payload_len = hton32(sizeof (ds_reg_nack_t));

	nack = (ds_reg_nack_t *)(msg + sizeof (ds_hdr_t));
	nack->svc_handle = hton64(svc_handle);
	nack->status = hton64(status);
	nack->major_vers = hton16(major);

	ds_chan_send_msg(chan, msg, msglen);
	Xfree(msg);
}

/*
 * Send a service registration ack.
 */
static void
send_reg_ack(ds_channel_t *chan, uint64_t svc_handle, uint16_t minor)
{
	ds_reg_ack_t	*ack;
	ds_hdr_t	*hdr;
	char		*msg;
	size_t		msglen;

	DBG(zprintf("%s: >DS_REG_ACK minor %u, chan %s, hdl 0x%"PRIx64"\n",
	    __func__, minor, chan->name, svc_handle));

	msglen = sizeof (ds_hdr_t) + sizeof (ds_reg_ack_t);
	msg = Xmalloc(msglen);

	hdr = (ds_hdr_t *)msg;
	hdr->msg_type = hton32(DS_REG_ACK);
	hdr->payload_len = hton32(sizeof (ds_reg_ack_t));

	ack = (ds_reg_ack_t *)(msg + sizeof (ds_hdr_t));
	ack->svc_handle = hton64(svc_handle);
	ack->minor_vers = hton16(minor);

	ds_chan_send_msg(chan, msg, msglen);
	Xfree(msg);
}

/*
 * Handle a incoming service registration request.
 */
static int
handle_reg_req(ds_channel_t *chan, void *buf, size_t len)
{
	ds_reg_req_t *req;
	uint64_t svc_handle;
	uint16_t req_major, req_minor;
	int namelen;
	bool_t match;
	uint16_t new_major, new_minor;
	ds_svc_t *ds;
	svc_t *svc;

	/*
	 * Make sure we have all the mandatory fields.
	 */

	if (len <= offsetof(ds_reg_req_t, svc_id)) {
		Xfree(buf);
		return (-1);		/* request is too short */
	}

	req = buf;
	svc_handle = ntoh64(req->svc_handle);
	req_major = ntoh16(req->major_vers);
	req_minor = ntoh16(req->minor_vers);

	if (((char *)buf)[len-1] != '\0') {
		Xfree(buf);
		return (-1);		/* service must be NUL terminated */
	}

	namelen = len - offsetof(ds_reg_req_t, svc_id);	/* includes NUL */
	if (namelen <= 1 || namelen > DS_MAX_SVCID_LEN) {
		Xfree(buf);
		return (-1);		/* service id too short or too long */
	}

	DBG(zprintf("%s: <DS_REQ_REQ, svc %s, chan %s, hdl 0x%"PRIx64"\n",
	    __func__, req->svc_id, chan->name, svc_handle));

	/*
	 * The request appears to be sane, now do the real handling.
	 */

	if (find_channel_svc_by_name(chan, req->svc_id) != NULL) {
		/* service already registered */
		send_reg_nack(chan, svc_handle, DS_REG_DUP, 0);
		Xfree(buf);
		return (0);
	}

	DBG(zprintf("%s: <DS_REQ_REQ, svc %s, chan %s, hdl 0x%"PRIx64"\n",
	    __func__, req->svc_id, chan->name, svc_handle));

	svc = find_plugged_svc_by_name(chan->plugins, req->svc_id);
	if (svc == NULL || (strcmp(svc->name, PRI_SVC) == 0 &&
	    ((IS_CONTROL_DOMAIN(chan->ldp) && get_pri_method != GET_PRI_DS) ||
	    !ds_do_pruned_pri))) {
		/*
		 * The service is not supported or this is the pri service and
		 * the PRI is being obtained via a method other than domain
		 * services. Either way, nack the registration request.
		 */
		send_reg_nack(chan, svc_handle, DS_REG_VER_NACK, 0);
		Xfree(buf);
		return (0);
	}

	/*
	 * Check version info. ACK only if the major numbers exactly
	 * match. The service entity can retry with a new minor
	 * based on the response sent as part of the NACK.
	 */
	match = negotiate_version(svc->num_vers, *svc->versions,
	    req_major, &new_major, &new_minor);

	if (match) {
		ds = add_registered_svc(chan, svc, svc_handle, req_major,
		    MIN(new_minor, req_minor), NULL);

		send_reg_ack(chan, svc_handle, new_minor);

		/*
		 * Notify the service that registration is complete.  This is
		 * done after sending the registration ack since the service
		 * registration handler may send data messages and these must
		 * come after the ack.
		 */
		if (ds && ds->ops->reg_handler)
			ds->ops->reg_handler(ds);
	} else {
		send_reg_nack(chan, svc_handle, DS_REG_VER_NACK, new_major);
	}

	Xfree(buf);
	return (0);
}

/*
 * Handle the acknowledgment to a service registration request.
 */
static int
handle_reg_ack(ds_channel_t *chan, void *buf, size_t len)
{
	ds_reg_ack_t	*ack = buf;
	prot_version_t	*ver;
	ds_svc_t	*ds;

	ack->svc_handle = ntoh64(ack->svc_handle);

	DBG(zprintf("%s: <DS_REG_ACK chan %s, hdl=0x%"PRIx64"\n",
	    __func__, chan->name, ack->svc_handle));

	/* lookup appropriate client */
	if ((ds = find_channel_svc_by_hdl(chan, ack->svc_handle)) == NULL) {
		DBG(zprintf("%s: invalid handle 0x%"PRIx64"\n", __func__,
		    ack->svc_handle));
		Xfree(buf);
		return (0);
	}

	/* make sure the message makes sense */
	if (ds->state != SVC_REG_PENDING) {
		DBG(zprintf("%s: service %s in invalid state %d\n",
		    __func__, ds->name, ds->state));
		Xfree(buf);
		return (0);
	}

	ver = &chan->versions[chan->ver_idx];
	ack->minor_vers = ntoh16(ack->minor_vers);

	/* major version has been agreed upon */
	ds->version.major_number = ver->major_number;

	if (ack->minor_vers >= ver->minor_number) {
		/*
		 * Use the minor version specified in the
		 * original request.
		 */
		ds->version.minor_number = ver->minor_number;
	} else {
		/*
		 * Use the lower minor version returned in
		 * the ack. By definition, all lower minor
		 * versions must be supported.
		 */
		ds->version.minor_number = ack->minor_vers;
	}

	ds->state = SVC_ACTIVE;

	load_svc_plugin(ds);

	DBG(zprintf("%s: %s version %d.%d ready, hdl 0x%"PRIx64"\n", __func__,
	    ds->name, ds->version.major_number,
	    ds->version.minor_number, ds->hdl));

	/* notify the service that registration is complete */
	if (ds->ops->reg_handler)
		ds->ops->reg_handler(ds);

	/* start registration for next service on the channel */
	ds_channel_reg_req(chan);

	Xfree(buf);
	return (0);
}

/*
 * Increment a service's handle. This is used to get a new
 * handle when a service registration is nacked and a new
 * registration request is sent.
 */
static void
inc_svc_handle(ds_svc_t *ds)
{
	uint32_t hdl_low;

	hdl_low = (uint32_t)(ds->hdl & 0xffffffffUL) + 1;
	ds->hdl = (ds->hdl & 0xffffffff00000000ULL) | hdl_low;
}

static int
handle_reg_nack(ds_channel_t *chan, void *buf, size_t len)
{
	ds_reg_nack_t	*nack = buf;
	ds_svc_t	*ds;
	int		idx;

	nack->svc_handle = ntoh64(nack->svc_handle);

	DBG(zprintf("%s: <DS_REG_NACK chan %s, hdl=0x%"PRIx64"\n",
	    __func__, chan->name, nack->svc_handle));

	/* lookup appropriate client */
	if ((ds = find_channel_svc_by_hdl(chan, nack->svc_handle)) == NULL) {
		DBG(zprintf("%s: invalid handle 0x%"PRIx64"\n", __func__,
		    nack->svc_handle));
		Xfree(buf);
		return (0);
	}

	/* make sure the message makes sense */
	if (ds->state != SVC_REG_PENDING) {
		DBG(zprintf("%s: service %s in invalid state %d\n",
		    __func__, ds->name, ds->state));
		Xfree(buf);
		return (0);
	}

	nack->major_vers = ntoh16(nack->major_vers);

	/*
	 * A major version of zero indicates that the
	 * service is not supported at all.
	 */
	if (nack->major_vers == 0) {
		DBG(zprintf("%s: %s not supported\n", __func__, ds->name));
		ds->state = SVC_INACTIVE;
		Xfree(buf);
		return (0);
	}

	DBG(zprintf("%s: hdl=0x%"PRIx64", nack=%d.x\n", __func__,
	    nack->svc_handle, nack->major_vers));

	/*
	 * Walk the version list for the service, looking for
	 * a major version that is as close to the requested
	 * major version as possible.
	 */
	for (idx = ds->ver_idx; idx < ds->svc->num_vers; idx++) {
		if (ds->svc->versions[idx]->major_number <=
		    nack->major_vers) {
			/* found a version to try */
			break;
		}
	}

	if (idx == ds->svc->num_vers) {
		/* no supported version */
		DBG(zprintf("%s: %s major version %d not supported\n", __func__,
		    ds->name, nack->major_vers));
		ds->state = SVC_INACTIVE;
		Xfree(buf);
		return (0);
	}

	/* start the handshake again with a new handle */
	inc_svc_handle(ds);
	ds->state = SVC_INACTIVE;
	ds->ver_idx = idx;

	send_reg_req(ds);

	Xfree(buf);
	return (0);
}

static int
handle_init_nack(ds_channel_t *chan, void *buf, size_t len)
{
	ds_init_nack_t	*nack = buf;
	prot_version_t	*ver;
	int		idx;

	mutex_lock(&chan->lock);
	if (chan->state != CHAN_INIT_REQ) {
		mutex_unlock(&chan->lock);
		Xfree(buf);
		return (-1);
	}
	mutex_unlock(&chan->lock);

	nack->major_vers = ntoh16(nack->major_vers);

	ver = &chan->versions[chan->ver_idx];

	DBG(zprintf("%s: <DS_INIT_NACK req=v%d.%d, nack=v%d.x, chan %s\n",
	    __func__, ver->major_number, ver->minor_number, nack->major_vers,
	    chan->name));

	if (nack->major_vers == 0) {
		/* no supported protocol version */
		DBG(zprintf("%s: service not supported\n", __func__));
		Xfree(buf);
		return (0);
	}

	/*
	 * Walk the version list, looking for a major version
	 * that is as close to the requested major version as
	 * possible.
	 */
	for (idx = chan->ver_idx; idx < chan->num_vers; idx++) {
		if (chan->versions[idx].major_number <= nack->major_vers) {
			/* found a version to try */
			chan->ver_idx = idx;
			ds_channel_init_req(chan);
			Xfree(buf);
			return (0);
		}
	}

	/* no supported version */
	DBG(zprintf("%s: service version %d not supported\n",
	    __func__, nack->major_vers));

	Xfree(buf);
	return (0);
}

static int
handle_init_ack(ds_channel_t *chan, void *buf, size_t len)
{
	ds_init_ack_t	*ack = buf;
	prot_version_t	*ver;

	mutex_lock(&chan->lock);
	if (chan->state == CHAN_INIT_REQ) {
		chan->state = CHAN_UP;
	} else {
		mutex_unlock(&chan->lock);
		Xfree(buf);
		return (-1);
	}
	mutex_unlock(&chan->lock);

	ver = &chan->versions[chan->ver_idx];

	/* agreed upon a major version */
	chan->version.major_number = ver->major_number;

	/*
	 * For the minor version use the lower of the requested and the
	 * returned minor version. Note that when a minor version is
	 * supported, then by definition all lower minor are also supported.
	 */
	ack->minor_vers = ntoh16(ack->minor_vers);
	chan->version.minor_number = MIN(ver->minor_number, ack->minor_vers);

	DBG(zprintf("%s: <DS_INIT_ACK v%d.%d, chan %s\n", __func__,
	    chan->version.major_number, chan->version.minor_number,
	    chan->name));

	/* start service registration */
	ds_channel_reg_req(chan);

	Xfree(buf);
	return (0);
}

static void
send_init_ack(ds_channel_t *chan, uint16_t minor)
{
	ds_hdr_t	*hdr;
	ds_init_ack_t	*ack;
	char		*msg;
	size_t		msglen;

	DBG(zprintf("%s: >DS_INIT_ACK minor %u, chan %s\n",
	    __func__, minor, chan->name));

	msglen = sizeof (ds_hdr_t) + sizeof (ds_init_ack_t);
	msg = Xmalloc(msglen);

	hdr = (ds_hdr_t *)msg;
	hdr->msg_type = hton32(DS_INIT_ACK);
	hdr->payload_len = hton32(sizeof (ds_init_ack_t));

	ack = (ds_init_ack_t *)(msg + sizeof (ds_hdr_t));
	ack->minor_vers = hton16(minor);

	ds_chan_send_msg(chan, msg, msglen);
	Xfree(msg);
}

static void
send_init_nack(ds_channel_t *chan, uint16_t major)
{
	ds_hdr_t	*hdr;
	ds_init_nack_t	*nack;
	char		*msg;
	size_t		msglen;

	DBG(zprintf("%s: >DS_INIT_NACK major %u, chan %s\n",
	    __func__, major, chan->name));

	msglen = sizeof (ds_hdr_t) + sizeof (ds_init_nack_t);
	msg = Xmalloc(msglen);

	hdr = (ds_hdr_t *)msg;
	hdr->msg_type = hton32(DS_INIT_NACK);
	hdr->payload_len = hton32(sizeof (ds_init_nack_t));

	nack = (ds_init_nack_t *)(msg + sizeof (ds_hdr_t));
	nack->major_vers = hton16(major);

	ds_chan_send_msg(chan, msg, msglen);
	Xfree(msg);
}

static int
handle_init_req(ds_channel_t *chan, void *buf, size_t len)
{
	ds_init_req_t	*req;
	uint16_t req_minor, req_major;
	uint16_t new_minor, new_major;
	bool_t match;

	DBG(zprintf("%s: <DS_INIT_REQ chan %s\n", __func__, chan->name));

	if (chan->props & CHAN_OUTGOING) {
		/*
		 * On outgoing channels version negotiation is always
		 * initiated by the LDom Manager. Any attempt by a remote
		 * party to initiate the handshake gets nack'd with a major
		 * number equal to zero. This indicates that no version is
		 * supported since an init request is not expected.
		 */
		send_init_nack(chan, 0);
	} else {
		if (len != sizeof (ds_init_req_t)) {
			Xfree(buf);
			return (-1);		/* request length invalid */
		}

		req = buf;
		req_major = ntoh16(req->major_vers);
		req_minor = ntoh16(req->minor_vers);

		match = negotiate_version(chan->num_vers, chan->versions,
		    req_major, &new_major, &new_minor);

		if (match) {
			chan->version.major_number = req_major;
			chan->version.minor_number = MIN(new_minor, req_minor);
			send_init_ack(chan, chan->version.minor_number);
		} else {
			send_init_nack(chan, new_major);
		}
	}

	Xfree(buf);
	return (0);
}

void
ds_fill_data_header(void *resp, uint64_t svc_handle, uint32_t data_len)
{
	ds_resp_t *ms_resp = resp;

	ms_resp->hdr.msg_type = hton32(DS_DATA);
	ms_resp->hdr.payload_len = hton32(sizeof (ds_data_handle_t) + data_len);
	ms_resp->payload.data.datahdr.svc_handle = hton64(svc_handle);
}

/*
 * Queue of messages to be handled.
 */
static struct msg_queue ms_queue = {
	DEFAULTMUTEX,
	DEFAULTCV,
	DEFAULTCHAIN
};

/*
 * Must be called with chan->lock held
 */
void
ds_channel_free_queued(ds_channel_t *chan)
{
	ms_qmsg_t *qmsg, *next;

	ASSERT(MUTEX_HELD(&chan->lock));

	mutex_lock(&ms_queue.lock);

	for (qmsg = CHAIN_START(ms_queue.msgs); qmsg != NULL; qmsg = next) {
		next = CHAIN_NEXT(qmsg->link);
		if (qmsg->chan == chan) {
			CHAIN_UNLINK(ms_queue.msgs, qmsg, link);
			Xfree(qmsg->msg);
			Xfree(qmsg);
			chan->msgcnt--;
			ASSERT(chan->msgcnt >= 0);
			if (chan->msgcnt == 0)
				cond_broadcast(&chan->cond);
		}
	}

	mutex_unlock(&ms_queue.lock);
}

/*
 * Message handler thread. This processes incoming domain services messages
 * according to their type.
 */
static void *
msg_handler(void *ignored)
{
	ms_qmsg_t *qmsg;
	uint32_t msg_type;
	uint32_t payload_len;

	while (1) {
		/*
		 * Wait for a message to handle.
		 */
		mutex_lock(&ms_queue.lock);
		while ((qmsg = CHAIN_START(ms_queue.msgs)) == NULL)
			cond_wait(&ms_queue.cond, &ms_queue.lock);
		CHAIN_UNLINK(ms_queue.msgs, qmsg, link);
		mutex_unlock(&ms_queue.lock);

		msg_type = ntoh32(qmsg->hdr.msg_type);
		payload_len = ntoh32(qmsg->hdr.payload_len);

		/*
		 * Handle the message if it has a valid type, otherwise it is
		 * just dropped on the floor. The handler takes care of
		 * freeing the message.
		 */
		if (msg_type <= DS_NACK) {
			(*msg_handlers[msg_type])(qmsg->chan, qmsg->msg,
			    payload_len);
		} else {
			Xfree(qmsg->msg);
		}

		mutex_lock(&qmsg->chan->lock);
		qmsg->chan->msgcnt--;
		ASSERT(qmsg->chan->msgcnt >= 0);
		if (qmsg->chan->msgcnt == 0)
			cond_broadcast(&qmsg->chan->cond);
		mutex_unlock(&qmsg->chan->lock);

		Xfree(qmsg);
	}

	/*NOTREACHED*/
}

/*
 * Find a routine in a service plugin for a given channel.
 */
void *
ds_service_routine(ds_channel_t *chan, char *service, char *function)
{
	ds_svc_t *ds;

	if (chan == NULL)
		return (NULL);

	ds = find_channel_svc_by_name(chan, service);
	if (ds == NULL || ds->plugin_dlhdl == NULL)
		return (NULL);
	else
		return (dlsym(ds->plugin_dlhdl, function));
}

/*
 * Find a routine in a service plugin for a channel.
 */
void *
ds_channel_service_routine(ds_channel_t *chan, char *service, char *function)
{
	ds_svc_t *ds;

	if (chan != NULL) {
		ds = find_channel_svc_by_name(chan, service);
		if (ds == NULL || ds->plugin_dlhdl == NULL)
			return (NULL);
		else
			return (dlsym(ds->plugin_dlhdl, function));
	} else {
		return (NULL);
	}
}

/*
 * Queue up a domain service message for handling by the message handler.
 * This is always called from the domain services I/O thread. This routine
 * must be invoked with chan->lock held
 */
void
ds_enqueue_msg(ds_channel_t *chan, ds_hdr_t *hdr, void *msg)
{
	ms_qmsg_t *qmsg;

	qmsg = Xmalloc(sizeof (*qmsg));
	qmsg->chan = chan;
	qmsg->hdr = *hdr;
	qmsg->msg = msg;

	ASSERT(MUTEX_HELD(&chan->lock));

	mutex_lock(&ms_queue.lock);
	chan->msgcnt++;
	CHAIN_ADD(ms_queue.msgs, qmsg, link);
	cond_signal(&ms_queue.cond);
	mutex_unlock(&ms_queue.lock);
}

ds_svc_t *
ds_lookup_svc(ds_channel_t *chan, char *name)
{
	if (chan == NULL)
		return (NULL);
	else
		return (find_channel_svc_by_name(chan, name));
}

/*
 * Send the specified data message to the specified target. The message must
 * already be in network byte order. The request argument is the complete
 * message, not just the data portion. This allows us to send the complete
 * message without any intermediate copy.
 *
 * If the timeout value is -1, wait indefinitely for a response. If the
 * timeout value is 0, return immediately without sleeping. Otherwise,
 * wait for the specified number of seconds for a response.
 *
 * This must not be called from the domain services IO handler since it may
 * block waiting for a response which the IO handler must read.
 */
int
ds_send_data_msg(ds_channel_t *chan, ds_msg_t *msg, int tout, ds_msg_t *resp)
{
	ds_pend_msg_t	*pend;
	timestruc_t	to;
	ds_req_t	*ds_req = msg->buf;
	int		err, rv;
	ds_svc_t	*ds = find_channel_svc(DS_DATA_SVC_HANDLE(ds_req));

	/*
	 * If the service has been unregistered already, return
	 */
	if (ds == NULL || chan == NULL)
		return (EINVAL);

	if (resp != NULL)
		/* ensure response buffer is NULL when error is returned */
		resp->buf = NULL;

	if ((msg->buf == NULL) || (msg->len == 0) ||
	    (DS_SIZE(ds_req) != msg->len) ||
	    (DS_REQ_TYPE(ds_req) != DS_DATA))
		return (EINVAL);

	if (DS_SIZE(ds_req) != msg->len)
		return (EINVAL);

	/* check for a message that is too large */
	if ((chan->ldp != NULL) && (msg->len > DS_MAX_MSG_SIZE))
		return (EMSGSIZE);

	mutex_lock(&chan->lock);
	if (chan->state != CHAN_UP) {
		rv = ENOLINK;
	} else {
		rv = 0;
		chan->usecnt++;
	}
	mutex_unlock(&chan->lock);

	if (rv != 0)
		return (rv);

	/* record the message being sent */
	if (tout != 0) {
		pend = Xcalloc(1, ds_pend_msg_t);
		pend->ds_handle = DS_DATA_SVC_HANDLE(ds_req);
		pend->seq = msg->seq;
		pend->status = DS_MSG_PENDING;

		pend->len = 0;
		pend->msg = NULL;

		mutex_lock(&chan->pending.lock);
		CHAIN_ADD(chan->pending.msgs, pend, link);
		mutex_unlock(&chan->pending.lock);
	}

	/*
	 * The entire data is already framed in the message buffer, hence we
	 * don't have to copy anything here.
	 */
	err = ds_chan_send_msg(chan, msg->buf, msg->len);
	if (err != 0) {
		DBG(zprintf("%s: error sending msg (%d bytes) on chan %s: %s\n",
		    __func__, msg->len, chan->name, strerror(errno)));

		if (tout != 0) {
			mutex_lock(&chan->pending.lock);
			CHAIN_UNLINK(chan->pending.msgs, pend, link);
			mutex_unlock(&chan->pending.lock);
			Xfree(pend);
		}
		rv = EIO;
		goto out;
	}

	/* do not wait if the timeout is zero */
	if (tout == 0) {
		rv = 0;
		goto out;
	}

	DBG(zprintf("%s: waiting for response to msg %"PRId64", chan %s, "
	    "service %s , hdl 0x%"PRIx64"\n", __func__, pend->seq, chan->name,
	    ds->name, pend->ds_handle));

	/*
	 * Until the response is received, wait here. Once the callback is
	 * invoked, this should wake up, provided the right response is
	 * received.
	 */
	if (tout != -1) {
		to.tv_sec = time(NULL) + tout;
		to.tv_nsec = 0;
	}

	mutex_lock(&chan->pending.lock);

	while (pend->status == DS_MSG_PENDING) {
		if (tout != -1)
			err = cond_timedwait(&chan->pending.cond,
			    &chan->pending.lock, &to);
		else
			err = cond_wait(&chan->pending.cond,
			    &chan->pending.lock);

		if (err == 0 || err == EINTR)
			continue;
		else
			break;
	}

	/* whether successful or not unlink the pending entry */
	CHAIN_UNLINK(chan->pending.msgs, pend, link);

	mutex_unlock(&chan->pending.lock);

	if (pend->status == DS_MSG_COMPLETE) {
		/*
		 * The received message if any is in a newly allocated buffer
		 * pend->msg.
		 */
		if (resp != NULL) {
			/* ok if incoming message is NULL */
			resp->buf = pend->msg;
			resp->len = pend->len;
		} else {
			/* caller is not interested in the message data */
			if (pend->msg != NULL)
				Xfree(pend->msg);
		}

		rv = 0;
	} else if (pend->status == DS_MSG_ERROR) {
		rv = EIO;
	} else {
		/* timed out or some other error */
		rv = err;
	}

	DBG(zprintf("%s <%"PRIx64"> response to %"PRId64", %s\n", chan->name,
	    pend->ds_handle, pend->seq, (rv == 0) ? "ok" : strerror(err)));

	Xfree(pend);
out:
	mutex_lock(&chan->lock);
	chan->usecnt--;
	ASSERT(chan->usecnt >= 0);
	if (chan->usecnt == 0)
		cond_broadcast(&chan->cond);
	mutex_unlock(&chan->lock);

	return (rv);
}

/*
 * Check the specified response from the specified target to see if there are
 * any threads blocked waiting for the response. If there is a waiting thread,
 * copy out the data and notify the waiting thread that the response is ready.
 */
int
ds_check_resp(ds_msg_tgt_t *tgt, ds_msg_t *resp)
{
	ds_pend_msg_t *pend;

	if ((tgt == NULL) || (resp == NULL)) {
		/* return an error */
		return (DS_MSG_UNCLAIMED);
	}

	/* walk the pending requests looking for the specified one */
	CHAIN_ITER(tgt->chan->pending.msgs, pend, link) {
		if ((pend->ds_handle == tgt->ds_hdl) &&
		    (pend->seq == resp->seq)) {
			/* found it */
			break;
		}
	}

	if (pend == NULL) {
		/*
		 * A matching request for this response was not in the list,
		 * so it is an asynchronous message that must be handled by
		 * the caller.
		 */
		return (DS_MSG_UNCLAIMED);
	}

	/*
	 * Matching request found for this response.
	 */

	mutex_lock(&tgt->chan->pending.lock);

	/* fill in the message information */
	pend->status = DS_MSG_COMPLETE;
	pend->len = resp->len;
	if (resp->len > 0) {
		/* memory previously allocated in ds_msg_handler() */
		pend->msg = resp->buf;
	}

	/* signal that the response has been received */
	cond_broadcast(&tgt->chan->pending.cond);

	mutex_unlock(&tgt->chan->pending.lock);

	return (DS_MSG_CLAIMED);
}

dslib_svc_aggr_t *
find_dslib_svc_aggr(char *svc_id)
{
	dslib_svc_aggr_t *aggr = NULL;

	mutex_lock(&dslib_aglock);
	CHAIN_ITER(dslib_svcs, aggr, link) {
		if (strcmp(aggr->svc_id, svc_id) == 0)
			break;
	}
	mutex_unlock(&dslib_aglock);

	return (aggr);
}

/*
 * Create a dslib service aggregator if not present
 */
dslib_svc_aggr_t *
get_dslib_svc_aggr(char *svc_id)
{
	dslib_svc_aggr_t *aggr;

	aggr  = find_dslib_svc_aggr(svc_id);
	if (aggr == NULL) {
		aggr = Xcalloc(1, dslib_svc_aggr_t);
		aggr->svc_id = Xstrdup(svc_id);
		mutex_init(&aggr->lock, USYNC_THREAD, NULL);
		CHAIN_LINK_INIT(aggr->link);
		CHAIN_INIT(aggr->dslib_receivers);

		mutex_lock(&dslib_aglock);
		CHAIN_ADD(dslib_svcs, aggr, link);
		mutex_unlock(&dslib_aglock);
	}

	return (aggr);
}

/*
 * Generic registration, unregistration and data handlers for dslib
 * When the registration handler is invoked, we add the service to
 * the channel for the LDom. Unlike the vldc handler, the libds
 * registration handler is invoked *after* a successful version
 * negotiation.
 */
void
dslib_register_handler(ds_hdl_t hdl, ds_cb_arg_t arg, ds_ver_t *ver,
    ds_domain_hdl_t dhdl)
{
	ds_svc_t *ds;
	dslib_svc_aggr_t *aggr;
	svc_t *svc;
	ds_channel_t *chan;
	char *svc_id = (char *)arg;
	int ret = 0;

	DBG(zprintf("%s: hdl = 0x%"PRIx64", dhdl = 0x%"PRIx64", service = %s\n",
	    __func__, hdl, dhdl, svc_id));

	/*
	 * If the channel in use on the LDom isn't libds, just ignore
	 */
	if (find_libds_chan_svc(hdl, dhdl, arg, &svc, &chan) != 0) {
		DBG(zprintf("%s: failed: hdl = 0x%"PRIx64", dhdl = "
		    "0x%"PRIx64"\n", __func__, hdl, dhdl));
		return;
	}

	ds = add_registered_svc(chan, svc, hdl, ver->major, ver->minor, &ret);

	if (ds == NULL) {
		DBG(zprintf("%s: add_registered_svc() failed: hdl = 0x%"PRIx64""
		    ", dhdl = 0x%"PRIx64", svc = %s\n", __func__, hdl,
		    dhdl, svc_id));
		return;
	}

	if (ds->ops->reg_handler) {
		ds->ops->reg_handler(ds);
	}

	/*
	 * Since there is only one callback for a particular service,
	 * keep all individual instances of the services in a list
	 * If the registration only replaced a handle on the already
	 * registered ds service, the following won't be needed.
	 */
	aggr = get_dslib_svc_aggr(svc_id);
	if (ret == 0) {
		mutex_lock(&aggr->lock);
		CHAIN_ADD(aggr->dslib_receivers, ds, aggr_link);
		mutex_unlock(&aggr->lock);
		ds->svc_aggr = aggr;
	}

	DBG(zprintf("%s: Registered: hdl = 0x%"PRIx64", "
	    "dhdl = 0x%"PRIx64", svc = %s\n", __func__, hdl, dhdl, svc_id));
}

static ds_svc_t *
dslib_remove_hdl(ds_hdl_t hdl, dslib_svc_aggr_t *aggr, char *svc_id)
{
	ds_svc_t *ds = NULL;
	ds_svc_t *nds;
	ldom_t *ldp;

	/*
	 * First remove the service from the aggregator list. The service can
	 * only be identified by searching for the handle in the aggregator
	 * list
	 */
	mutex_lock(&aggr->lock);
	CHAIN_ITER_VOL(aggr->dslib_receivers, ds, nds, aggr_link) {
		mutex_lock(&ds->lock);
		if (ds->hdl == hdl) {
			/*
			 * Proceed only if there isn't already a domain
			 * unregistration in progress
			 */
			if (ds->unreg != SVC_UNREG_LDOM) {
				ds->unreg = SVC_UNREG_HDL;
				mutex_unlock(&ds->lock);
				channel_error_notify_pending(ds->chan, hdl);
				CHAIN_UNLINK(aggr->dslib_receivers, ds,
				    aggr_link);
				break;
			} else {
				mutex_unlock(&ds->lock);
				mutex_unlock(&aggr->lock);
				return (ds);
			}

		} else {
			mutex_unlock(&ds->lock);
		}
	}
	mutex_unlock(&aggr->lock);

	/*
	 * Now remove the service from the channel
	 */
	if (ds != NULL) {
		mutex_lock(&ds->chan->lock);
		ldp = ds->chan->ldp;
		DBG(zprintf("Unregistering handle 0x%"PRIx64", "
		    "service %s, ldom %s\n", hdl, svc_id, ldp->namep));
		CHAIN_UNLINK(ds->chan->ds_svcs, ds, link);
		seq_dequeue_ds_svc(ds, false);
		ds_svc_decref(ds, 1);
		cond_broadcast(&ds->chan->cond); /* wake up domain unregister */
		mutex_unlock(&ds->chan->lock);
	}

	return (ds);
}

/*
 * When a service is unregistered, the following is invoked. At this
 * time the handle is **INVALID** for libds calls, so we need to
 * handle it appropriately
 */
void
dslib_unregister_handler(ds_hdl_t hdl, ds_cb_arg_t arg)
{
	dslib_svc_aggr_t *aggr;
	char *svc_id = (char *)arg;

	DBG(zprintf("%s: handle 0x%"PRIx64", svc_id = %s\n",
	    __func__, hdl, svc_id));

	/*
	 * Can't use find_libds_chan_svc() because the handle is invalid now
	 *
	 * Remove the service instance from the aggregator pool
	 */
	aggr = find_dslib_svc_aggr(svc_id);
	if (aggr == NULL) {
		DBG(zprintf("%s: find_dslib_svc_aggr(): handle 0x%"PRIx64", "
		    "service %s: service not registered\n", __func__,
		    hdl, svc_id));
		return;
	}

	if (dslib_remove_hdl(hdl, aggr, svc_id) == NULL) {
		DBG(zprintf("%s: Didn't find handle 0x%"PRIx64", service %s\n",
		    __func__, hdl, svc_id));
	}
}

#if ZDBG
void
dump_libds_handle(ds_hdl_t hdl)
{
	ds_svc_t *ds = find_channel_svc(hdl);
	char *svc_id;

	if (ds == NULL) {
		DBG(zprintf("%s: find_libds_chan_svc(): handle 0x%"PRIx64" not "
		    "registered\n", __func__, hdl));
		return;
	}

	if (ds->hdl == hdl) {
		ldom_t *ldp = ds->chan->ldp;
		svc_id = ds->name;
		DBG(zprintf("%s: Handle 0x%"PRIx64", service %s, ldom %s\n",
		    __func__, hdl, svc_id, ldp->namep));
	} else {
		DBG(zprintf("%s: Handle 0x%"PRIx64" no match\n",
		    __func__, hdl));
	}
}
#endif

extern char *libds_chan_create_rsp(ds_hdr_t *hdr, ds_hdl_t hdl,
    char *buf, size_t buflen);
/*
 * msg received on libds channel. enqueue the message. Find the relevant libds
 * channel and then create a message and copy the content to the messgge.
 */
void
dslib_data_handler(ds_hdl_t hdl, ds_cb_arg_t arg, void *buf, size_t buflen)
{
	dslib_svc_aggr_t *aggr;
	char *svc_id = (char *)arg;
	ds_channel_t *chan;
	svc_t *svc;

	DBG(zprintf("%s: svc_id = %s\n", __func__, svc_id));

	if (find_libds_chan_svc(hdl, -1, arg, &svc, &chan) != 0) {
		DBG(zprintf("%s: find_libds_chan_svc(): handle 0x%"PRIx64", "
		    "service %s: service not registered\n", __func__,
		    hdl, svc_id));
		return;
	}

	if (find_channel_svc_by_hdl(chan, hdl) == NULL) {
		DBG(zprintf("%s: find_channel_svc_by_hdl(): handle "
		    "0x%"PRIx64", service %s: service not registered\n",
		    __func__, hdl, svc_id));
		return;
	}

	aggr = find_dslib_svc_aggr(svc_id);
	if (aggr == NULL) {
		DBG(zprintf("%s: find_dslib_svc_aggr(): handle 0x%"PRIx64", "
		    "service %s: service not registered\n", __func__,
		    hdl, svc_id));
		return;
	}

	/*
	 * Once the data handler is invoked, it needs to queue up the received
	 * message on the libds channel
	 */
	mutex_lock(&chan->lock);
	chan->buf = libds_chan_create_rsp(&chan->hdr, hdl, buf, buflen);
	ds_enqueue_msg(chan, &chan->hdr, chan->buf);
	chan->buf = NULL;		/* message handling frees the buffer */
	mutex_unlock(&chan->lock);
}
