/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_DS_H
#define	_DS_H

#pragma ident	"@(#)ds.h	1.27	09/05/20 SMI"

#ifdef	__cplusplus
extern "C" {
#endif

#include <inttypes.h>
#include "libds.h"

#define	DR_DS_TIMEOUT	(60)	/* DR timeout value - Units: seconds */

#define	DS_DEFAULT_MTU	0x1000	/* matches VLDC default mtu size of 4K */

typedef struct libds_flag {
	int supported:1;	/* true if libds available */
	int config:1;		/* true if md indicated libds */
	int functional:1;	/* true if supported and boot config had vlds */
	int convert_libds:1;	/* true if vldc ports need conversion */
	int convert_vldc:1;	/* true if need reverse conversion */
} libds_flag_t;

extern libds_flag_t libds_cap;

/*
 * DS message header.
 */
typedef struct ds_hdr {
	uint32_t	msg_type;	/* message type */
	uint32_t	payload_len;	/* payload length */
} ds_hdr_t;

/*
 * Fixed message types.
 */
#define	DS_INIT_REQ	0x0		/* initiate connection */
#define	DS_INIT_ACK	0x1		/* initiation acknowledgement */
#define	DS_INIT_NACK	0x2		/* initiation failed */

/*
 * Content of message sent from domain to initialize the
 * communication channel.
 */
typedef struct {
	uint16_t	major_vers;
	uint16_t	minor_vers;
} ds_init_req_t;

/*
 * If zeus accepts the initialization request, it sends
 * the minor version which is the minimum of the two
 * sides.
 */
typedef struct {
	uint16_t	minor_vers;
} ds_init_ack_t;

/*
 * Zeus will refuse the initialization request if the major
 * version of the domain side is larger, and sends the
 * major number that zeus can actually support - the domain
 * may re-negotiate again with a lower major number
 */
typedef struct {
	uint16_t	major_vers;
} ds_init_nack_t;

/*
 * Protocol Version 1.0 message types.
 */

#define	DS_REG_REQ	0x3		/* register a service */

typedef struct ds_reg_req {
	uint64_t	svc_handle;
	uint16_t	major_vers;
	uint16_t	minor_vers;
	char		svc_id[1];
} ds_reg_req_t;

#define	DS_REG_ACK	0x4		/* register acknowledgement */

typedef struct ds_reg_ack {
	uint64_t	svc_handle;
	uint16_t	minor_vers;
} ds_reg_ack_t;

#define	DS_REG_NACK	0x5		/* register failed */

typedef struct ds_reg_nack {
	uint64_t	svc_handle;
	uint64_t	status;
	uint16_t	major_vers;
} ds_reg_nack_t;

#define	DS_UNREG	0x6		/* unregister a service */

typedef struct ds_unreg_req {
	uint64_t	svc_handle;
} ds_unreg_req_t;

#define	DS_UNREG_ACK	0x7		/* unregister acknowledgement */

typedef struct ds_unreg_ack {
	uint64_t	svc_handle;
} ds_unreg_ack_t;

#define	DS_UNREG_NACK	0x8		/* unregister failed */

typedef struct ds_unreg_nack {
	uint64_t	svc_handle;
	uint64_t	status;
} ds_unreg_nack_t;

#define	DS_DATA		0x9		/* data message */

typedef struct ds_data_handle {
	uint64_t	svc_handle;
} ds_data_handle_t;

typedef struct ds_data_msg {
	ds_data_handle_t datahdr;
	uint8_t	bytes[1];
} ds_data_msg_t;

#define	DS_NACK		0xa		/* data error */

typedef struct ds_data_nack {
	uint64_t	svc_handle;	/* service handle sent in data msg */
	uint64_t	result;		/* reason for failure */
} ds_data_nack_t;

/*
 * Status codes.
 */
#define	DS_OK		0x0	/* success */
#define	DS_REG_VER_NACK	0x1	/* version not supported */
#define	DS_REG_DUP	0x2	/* duplicate registration request */
#define	DS_INV_HDL	0x3
#define	DS_TYPE_UNKNOWN	0x4
#define	DS_SVC_NOT_SUPP	0x5	/* service entity is not supported */

/*
 * Request type messages.
 */
typedef union {
	ds_init_req_t	init_req;
	ds_reg_req_t 	reg_req;
	ds_unreg_req_t	unreg_req;
	ds_data_msg_t	data;
} ds_req_payload_t;

typedef struct {
	ds_hdr_t		hdr;
	ds_req_payload_t	payload;
} ds_req_t;

/*
 * Response type messages.
 */
typedef union {
	ds_init_ack_t init_ack;
	ds_init_nack_t init_nack;
	ds_reg_ack_t reg_ack;
	ds_reg_nack_t reg_nack;
	ds_unreg_ack_t unreg_ack;
	ds_unreg_nack_t unreg_nack;
	ds_data_nack_t data_nack;
	ds_data_msg_t data;
} ds_resp_payload_t;

typedef struct {
	ds_hdr_t hdr;
	ds_resp_payload_t payload;
} ds_resp_t;

void dslib_register_handler(ds_hdl_t hdl, ds_cb_arg_t arg, ds_ver_t *ver,
    ds_domain_hdl_t dhdl);
void dslib_unregister_handler(ds_hdl_t hdl, ds_cb_arg_t arg);
void dslib_data_handler(ds_hdl_t hdl, ds_cb_arg_t arg, void *buf,
    size_t buflen);
void initialize_libds(void);

#ifdef	__cplusplus
}
#endif

#endif	/* _DS_H */
