/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)ds_comm.c	1.67	09/06/22 SMI"

/*
 * Low level communication module between higher level Domain Services
 * modules in Zeus and the Domain Services on a guest LDom.
 *
 * It isolates the communication method from the rest of Zeus,
 * making it invisible where Zeus is running. e.g. the low
 * level communication used may be different if Zeus is running
 * inside an LDom as opposed to running on a service processor.
 */

#include <sys/param.h>			/* MAXPATHLEN */
#include <sys/time.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <port.h>
#include <string.h>
#include <thread.h>
#include <unistd.h>

#include "allocate.h"
#include "basics.h"
#include "chain.h"
#include "fatal.h"
#include "ds.h"
#include "ds_impl.h"
#include "zprintf.h"

static int comm_port;			/* port for DS communication events */

static void read_msg_header(ds_channel_t *chan);
static int libds_chan_send_msg(ds_channel_t *chan, void *msg, size_t len);

extern int open_ldc_channel(ds_channel_t *chan);
extern int close_ldc_channel(ds_channel_t *chan);

/*
 * Callback to handle a channel coming up.
 */
static void
channel_up(struct ds_channel *chan)
{
	DBG(zprintf("Channel %s up event, guest %s\n", chan->name,
	    chan->ldp == NULL ? "N/A" : chan->ldp->namep));

	if (is_libds_channel(chan))
		return;

	if (chan->props & CHAN_OUTGOING) {
		/* start channel version negotiation */
		ds_channel_init_req(chan);
	} else {
		/*
		 * There can't be any messages to process yet and we are in
		 * the IO thread so the state cannot be externally
		 * changed, hence no lock is needed.
		 */
		ASSERT(chan->state == CHAN_OPEN);
		chan->state = CHAN_UP;
		if (ds_chan_ready_read(chan) == 0) {
			/* start registration of any outgoing services */
			while (chan->reg_svc != NULL)
				ds_channel_reg_req(chan);
		}
	}
}

/*
 * Error out all pending messages on a channel and notify any
 * threads waiting on them. If hdl is not DS_INVALID_HDL, only
 * error out the pending message matching the handle.
 */
void
channel_error_notify_pending(ds_channel_t *chan, ds_hdl_t hdl)
{
	ds_pend_msg_t *pend = NULL;

	mutex_lock(&chan->pending.lock);

	CHAIN_ITER(chan->pending.msgs, pend, link) {
		if (pend->status == DS_MSG_PENDING && hdl == DS_INVALID_HDL) {
			pend->status = DS_MSG_ERROR;
		} else if (pend->ds_handle == hdl) {
			pend->status = DS_MSG_ERROR;
			break;
		}
	}

	if (hdl == DS_INVALID_HDL || pend != NULL)
		cond_broadcast(&chan->pending.cond);
	mutex_unlock(&chan->pending.lock);
}

/*
 * Free all the services on a channel.
 */
static void
channel_free_svcs(ds_channel_t *chan)
{
	ds_svc_t *ds, *next;

	mutex_lock(&chan->lock);
	for (ds = CHAIN_START(chan->ds_svcs); ds != NULL; ds = next) {
		next = CHAIN_NEXT(ds->link);
		ds_svc_decref(ds, 1);
	}

	CHAIN_INIT(chan->ds_svcs);
	mutex_unlock(&chan->lock);
}

static void
close_channel(ds_channel_t *chan)
{
	DBG(zprintf("%s: %s, state %s, fd %d\n", __func__,
	    chan->name, chan_state_str(chan->state), chan->fd));

	mutex_lock(&chan->lock);
	if (chan->state == CHAN_RESET) {
		/* already closed */
		mutex_unlock(&chan->lock);
		return;
	} else {
		chan->state = CHAN_RESET;
	}

	if (close_ldc_channel(chan) == -1)
		efatal("Closing channel %s", chan->name);

	/* free any incoming messages queued up for handling */
	ds_channel_free_queued(chan);

	mutex_unlock(&chan->lock);

	/*
	 * Notify anyone waiting for a response on this channel that one won't
	 * be coming.
	 */
	channel_error_notify_pending(chan, DS_INVALID_HDL);

	/* wait for all messages to be handled */
	mutex_lock(&chan->lock);
	while (chan->msgcnt > 0)
		cond_wait(&chan->cond, &chan->lock);
	ASSERT(chan->msgcnt == 0);
	mutex_unlock(&chan->lock);

	/* free any queued up sequencer tasks for the channel */
	seq_dequeue_ds_chan(chan);

	/* wait for all sequencer tasks to be done */
	mutex_lock(&chan->lock);
	while (chan->usecnt > 0)
		cond_wait(&chan->cond, &chan->lock);
	ASSERT(chan->usecnt == 0);
	mutex_unlock(&chan->lock);

	/* free channel resources */
	channel_free_svcs(chan);
	if (chan->buf != NULL) {
		Xfree(chan->buf);
		chan->buf = NULL;
	}
}

static void
reset_channel(ds_channel_t *chan, bool_t alerted)
{
	DBG(zprintf("%s: %s, state %s, alerted %d\n", __func__, chan->name,
	    chan_state_str(chan->state), alerted));

	close_channel(chan);

	ASSERT(CHAIN_LENGTH(chan->pending.msgs) == 0);

	mutex_lock(&chan->lock);
	if (chan->unregister) {
		/*
		 * The channel is being unregistered. If an alert for the
		 * channel has been received then mark it as down and signal
		 * the sequencer thread which will be waiting for it to be
		 * closed. Otherwise leave it marked as in reset and when a
		 * alert for it is received it will marked as down.
		 */
		if (alerted) {
			chan->state = CHAN_DOWN;
			cond_broadcast(&chan->cond);
		}

		mutex_unlock(&chan->lock);
		return;
	}

	mutex_unlock(&chan->lock);

	/*
	 * This is a reset because of an error and not because the channel is
	 * being unregistered, hence re-open the channel.
	 *
	 * If the open fails we take no further action. Subsequent attempts to
	 * use services which would have been provided over the channel will
	 * fail and action may be taken there.
	 */
	(void) ds_open_channel(chan);
}

/*
 * Enable polling for read input on a channel.
 * Returns -1 if an error occurred and the channel was reset
 * otherwise 0.
 */
static int
enable_read(ds_channel_t *chan)
{
	if (port_associate(chan->evport, PORT_SOURCE_FD, chan->fd,
	    POLLIN|POLLHUP, chan) < 0) {
		ewarning("port_associate");
		reset_channel(chan, false);
		return (-1);
	}

	return (0);
}

/*
 * Read data on a channel.
 */
static int
chan_read(ds_channel_t *chan)
{
	int r;

	while (chan->left > 0) {
		r = read(chan->fd, chan->bufp, chan->left);
		if (r < 0) {
			if (errno != EAGAIN) {
				DBG(zprintf("read error on %s: %s\n",
				    chan->name, strerror(errno)));
				reset_channel(chan, false);
			} else {
				enable_read(chan);
			}
			return (-1);
		} else if (r == 0) {
			enable_read(chan);
			return (-1);
		} else {
			chan->bufp += r;
			chan->left -= r;
		}
	}

	return (0);
}

/*
 * Ready a channel to receive messages.
 * Returns -1 if an error occurred and the channel was reset
 * otherwise 0.
 */
int
ds_chan_ready_read(ds_channel_t *chan)
{
	/* set up to read a message header */
	chan->bufp = (char *)&chan->hdr;
	chan->left = sizeof (chan->hdr);
	chan->callback = read_msg_header;

	return (enable_read(chan));
}

static void
read_msg_payload(ds_channel_t *chan)
{
	if (chan_read(chan) == -1)
		return;

	mutex_lock(&chan->lock);
	ds_enqueue_msg(chan, &chan->hdr, chan->buf);
	mutex_unlock(&chan->lock);
	chan->buf = NULL;		/* message handling frees the buffer */

	(void) ds_chan_ready_read(chan); /* set up to read next message */
}

static void
read_msg_header(ds_channel_t *chan)
{
	if (chan_read(chan) == -1)
		return;

	chan->hdr.msg_type = ntoh32(chan->hdr.msg_type);
	chan->hdr.payload_len = ntoh32(chan->hdr.payload_len);
	if (chan->hdr.payload_len > DS_MAX_MSG_SIZE) {
		warning("Invalid message header on channel %s: "
		    "length %d is too big", chan->name, chan->hdr.payload_len);
		reset_channel(chan, false);
		return;
	}

	chan->buf = Xmalloc(chan->hdr.payload_len);
	chan->bufp = chan->buf;
	chan->left = chan->hdr.payload_len;
	chan->callback = read_msg_payload;

	read_msg_payload(chan);
}

int
ds_chan_send_msg(ds_channel_t *chan, void *msg, size_t len)
{
	struct timespec delay;
	hrtime_t start, tout;
	char *msgp;
	int attempts;	/* number of consecutive write calls with no progress */
	int r;

	if (is_libds_channel(chan))
		return (libds_chan_send_msg(chan, msg, len));

	DBG(zprintf("%s: chan = %s, msg_len = %d\n", __func__,
	    chan->name, len));

	tout = chan->send_latency + (len / 1024) * chan->send_inverse_bw;
	tout *= 1000000LL;		/* millisecs to nanosecs */

	attempts = 0;
	msgp = msg;
	start = gethrtime();

	while (len > 0) {
		r = write(chan->fd, msgp, len);
		if (r <= 0) {
			if ((r < 0) && errno != EAGAIN) {
				return (-1);
			}

			if (++attempts < chan->send_retries)
				continue;
			else
				attempts = 0;

			if ((gethrtime() - start) > tout) {
				errno = ETIME;
				return (-1);
			}

			delay.tv_sec = 0;
			delay.tv_nsec = (long)chan->send_delay * 1000000L;
			(void) nanosleep(&delay, NULL);
		} else {
			attempts = 0;
			msgp += r;
			len -= r;
		}
	}

	return (0);
}

/*
 * Send a message using libds ds_send_msg() interface. Only send
 * the data portion of the message to the handle embedded in the
 * message header
 */
static int
libds_chan_send_msg(ds_channel_t *chan, void *msg, size_t len)
{
	ds_req_t	*ds_req = msg;
	ds_hdl_t ds_handle = DS_DATA_SVC_HANDLE(ds_req);
	boolean_t isready;
	char *buf = DS_BUF(msg);
	int buflen = DS_DATA_BUF_SZ(len);

	struct timespec delay;
	hrtime_t start, tout;
	int attempts;	/* number of consecutive write calls with no progress */
	int r;

	DBG(zprintf("%s: chan = %s, msg_len = %d\n", __func__,
	    chan->name, len));

	attempts = 0;
	start = gethrtime();
	tout = chan->send_latency + (len / 1024) * chan->send_inverse_bw;
	tout *= (NANOSEC / MILLISEC);		/* millisecs to nanosecs */

	while (len > 0) {
		if ((r = libds_ops.ds_isready(ds_handle, &isready)) != 0) {
			DBG(zprintf("%s: 0x%"PRIx64", chan = %s, ds_isready() "
			    "failed: %d\n", __func__, ds_handle,
			    chan->name, r));
			DBG(dump_libds_handle((ds_hdl_t)ds_handle));
			return (-1);
		}

		if (isready == 0) {
			/*
			 * Handle not ready yet, retry calling ds_isready()
			 * again
			 */
			if (++attempts < chan->send_retries)
				continue;
			else
				attempts = 0;

			/*
			 * Pause for a while and then call ds_isready()
			 */
			if ((gethrtime() - start) > tout) {
				errno = ETIME;
				DBG(zprintf("%s: chan = %s timed out\n",
				    __func__, chan->name));
				return (-1);
			}

			delay.tv_sec = 0;
			delay.tv_nsec = (long)chan->send_delay *
			    (NANOSEC / MILLISEC);
			(void) nanosleep(&delay, NULL);
		} else {
			attempts = 0;
			len = 0;
			if ((r = libds_ops.ds_send_msg(ds_handle, buf,
			    buflen)) != 0) {
				DBG(zprintf("%s: 0x%"PRIx64", chan = %s, "
				    "ds_send_msg() failed, ret = %d\n",
				    __func__, ds_handle, chan->name, r));
				DBG(dump_libds_handle((ds_hdl_t)ds_handle));
				return (r);
			}
		}
	}

	return (0);
}

/*
 * Given the ds_hdl, response buffer and length, formulate a message
 * in vldc format
 */
char *
libds_chan_create_rsp(ds_hdr_t *hdr, ds_hdl_t hdl, char *buf, size_t buflen)
{
	int datalen = sizeof (ds_data_handle_t) + buflen;
	    /* length of ds_data_msg_t */
	char *databuf = Xcalloc(datalen, char);
	ds_resp_payload_t *resp = (ds_resp_payload_t *)databuf;

	resp->data.datahdr.svc_handle = hdl;
	memcpy(resp->data.bytes, buf, buflen);
	hdr->msg_type = DS_DATA;
	hdr->payload_len = datalen;

	return (databuf);
}

int
ds_open_channel(ds_channel_t *chan)
{
	/* set first service to register if any */
	chan->reg_svc = CHAIN_START(chan->plugins->outgoing);

	if (chan->props & CHAN_OUTGOING) {
		/*
		 * Start channel initialization from the highest version once
		 * the channel comes up.
		 */
		chan->ver_idx = 0;
	}

	if (libds_cap.supported == true && is_libds_channel(chan)) {
		/*
		 * setup the channel parameters anyway
		 */
		chan->state = CHAN_UP;
		chan->send_latency = 100; 	/* 100 milliseconds */
		chan->send_inverse_bw = 0;
		chan->send_delay = 10;		/* 10 milliseconds */
		chan->send_retries = 5;
		return (0);
	}

	if ((strcmp(chan->name, SP_CHAN_NAME) == 0) ||
	    (strcmp(chan->name, FMA_CHAN_NAME) == 0) || is_ds_channel(chan)) {
		if (open_ldc_channel(chan) == -1)
			return (-1);
	} else {
		return (-1);
	}

	/*
	 * Set up the channel for callback for non-libds channel
	 */
	chan->evport = comm_port;
	chan->callback = channel_up;
	chan->state = CHAN_OPEN;

	DBG(zprintf("%s: %s, fd %d\n", __func__, chan->name, chan->fd));

	if (port_associate(chan->evport, PORT_SOURCE_FD, chan->fd,
	    POLLOUT|POLLHUP, chan) < 0)
		efatal("port_associate");

	return (0);
}

void
ds_unregister_channel(ds_channel_t *chan)
{
	DBG(zprintf("%s: %s 0x%lx, state %s\n", __func__, chan->name, chan,
	    chan_state_str(chan->state)));

	mutex_lock(&chan->lock);

	chan->unregister = true;

	if (is_libds_channel(chan)) {
		mutex_unlock(&chan->lock);
		reset_channel(chan, true);
		return;
	} else if (chan->state != CHAN_DOWN) {
		/* send port alert which will trigger a reset */
		port_alert(chan->evport, PORT_ALERT_SET, 1, chan);
		while (chan->state != CHAN_DOWN)
			cond_wait(&chan->cond, &chan->lock);
	}
	mutex_unlock(&chan->lock);
}

char *
chan_state_str(int state)
{
	switch (state) {
	case CHAN_DOWN:		return ("down");
	case CHAN_RESET: 	return ("reset");
	case CHAN_OPEN:		return ("open");
	case CHAN_INIT_REQ:	return ("init_req");
	case CHAN_UP:		return ("up");
	default:		break;
	}

	return ("unknown");
}

/*
 * I/O thread.
 */
static void *
comm_io(void *unused)
{
	port_event_t ev;
	ds_channel_t *chan;

	DBG7(memset(&ev, 0, sizeof (ev)));

	DBG(zprintf("\nStarted Domain Services Poll thread ...\n"));

	while (1) {
		if (port_get(comm_port, &ev, NULL) < 0)
			efatal("port_get");

		if (ev.portev_source == PORT_SOURCE_FD) {
			chan = ev.portev_user;

			if (ev.portev_events & POLLHUP) {
				DBG(zprintf("POLLHUP on %s\n", chan->name));
				reset_channel(chan, false);
			} else {
				chan->callback(chan);
			}

		} else if (ev.portev_source == PORT_SOURCE_ALERT) {
			chan = ev.portev_user;

			DBG(zprintf("port alert for %s\n", chan->name));

			port_dissociate(chan->evport, PORT_SOURCE_FD, chan->fd);
			port_alert(chan->evport, PORT_ALERT_SET, 0, NULL);

			reset_channel(chan, true);
		} else {
			break;
		}
	}

	return (NULL);
}

void
ds_comm_init(void)
{
	if ((comm_port = port_create()) < 0)
		efatal("port_create");

	if (thr_create(NULL, NULL, comm_io, NULL, THR_DAEMON, NULL) != 0)
		efatal("Could not create DS I/O thread");
}
