/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)ds_comm_ldc.c	1.11	09/05/21 SMI"

#include <sys/time.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <port.h>
#include <string.h>
#include <thread.h>
#include <unistd.h>

#include "allocate.h"
#include "basics.h"
#include "fatal.h"
#include "ldom.h"
#include "ds.h"
#include "ds_impl.h"
#include "zprintf.h"
#include "vio.h"
#include "vldc.h"

int
close_ldc_channel(ds_channel_t *chan)
{
	if (chan->fd == -1)
		return (0);

	if (close(chan->fd) == 0 || errno == ENOLINK) {
		chan->fd = -1;
		return (0);
	} else {
		return (-1);
	}
}

char *
vldc_chan_by_name(char *chan_name, char *ldom_name)
{
	char *devname, *vdevname, *portname;
	int instance = 0;

	if (strncmp(chan_name, DS_CHAN_NAME,
	    strlen(DS_CHAN_NAME)) == 0) {
		portname = zsprintf("%s%s", VLDC_DS_SVC_PREFIX, ldom_name);
		vdevname = Xstrdup(VLDC_MD_NAME);
	} else if (strcmp(chan_name, SP_CHAN_NAME) == 0) {
		portname = Xstrdup(SP_VLDC_DEV_NAME);
		instance = SP_VLDC_DEV_INST;
		vdevname = Xstrdup(VLDC_MD_NAME);
	} else if (strcmp(chan_name, FMA_CHAN_NAME) == 0) {
		portname = Xstrdup(VLDC_FMACTL_SVC_NAME);
		vdevname = Xstrdup(VLDC_MD_NAME);
	} else {
		return (NULL);
	}

	devname = zsprintf("/devices/%s@%x/%s@%x/%s@%x:%s", VDEVS_NAME,
	    VDEVS_CFG_HANDLE, CHANNELS_NAME, CHANNELS_CFG_HANDLE,
	    vdevname, instance, portname);

	Xfree(portname);
	Xfree(vdevname);

	return (devname);
}

bool_t
vldc_node_exists(char *devname)
{
	struct stat statbuf;

	/*
	 * Check if the device exists
	 */
	if (stat(devname, &statbuf) != 0) {
		return (false);
	} else {
		DBG(zprintf("%s: Found device %s\n", __func__, devname));
	}

	return (true);
}

char *
check_vldc_chan_by_name(char *chan_name, char *ldom_name)
{
	struct stat statbuf;

	char *devname = vldc_chan_by_name(chan_name, ldom_name);

	/*
	 * Check if the device exists
	 */
	if (stat(devname, &statbuf) != 0) {
		Xfree(devname);
		return (NULL);
	}

	DBG(zprintf("%s: Found device %s\n", __func__, devname));

	return (devname);
}

/*
 * Get the vldc dev for DS communication to an LDom
 */
static char *
get_channel_devname(ds_channel_t *chan)
{
	char *ldom_name = (chan->ldp == NULL) ? NULL : chan->ldp->namep;

	return (check_vldc_chan_by_name(chan->plugins->chan_name, ldom_name));
}

static int
do_ioctl(int fd, int request, vldc_opt_op_t *op)
{
	struct timespec delay;
	int i;

	delay.tv_sec = 0;
	delay.tv_nsec = 10 * 1000000L;	/* 10 milliseconds */

	for (i = 0; i < 10; i++) {
		if (ioctl(fd, request, op) != -1)
			return (0);

		if (errno != EAGAIN)
			return (-1);

		(void) nanosleep(&delay, NULL);
	}

	errno = ETIMEDOUT;
	return (-1);
}

/*
 * Given the channel name, e.g. "fmactl" or "spds",
 * check if the channel exists
 */
bool_t
check_vldc_chan_pres(char *chan_name)
{
	bool_t ret = false;
	char *devname = check_vldc_chan_by_name(chan_name, NULL);

	if (devname != NULL)
		ret = true;
	Xfree(devname);

	return (ret);
}

/*
 * Check if the ds channel is present
 */
bool_t
check_for_ds_chan(ldom_t *ldp)
{
	char *devname;
	ds_port_t *ds_port = CHAIN_START(ldp->ds_ports);

	/*
	 * Is the other end a vldc port ?
	 */
	if (ds_port->vldc_port == NULL ||
	    ds_port->vldc_port->peer.ds != ds_port)
		return (false);

	devname = vldc_chan_by_name(DS_CHAN_NAME, ldp->namep);

	if (vldc_node_exists(devname) == false) {
		warning("%s: vldc ports are setup for DS communication, but "
		    "vldc device node %s isn't found\n, DS communication "
		    "will not be available\n", __func__, devname);
	}

	Xfree(devname);

	return (true);
}

/*
 * Check if an LDom uses libds for DS communication
 */
bool_t
check_for_libds(ldom_t *ldp)
{
	ds_port_t *ds_port = CHAIN_START(ldp->ds_ports);

	/*
	 * For primary domain, the DS port is attached to a vldc-port
	 * which isn't used with libds enabled. In that case we just
	 * check if the vlds node exists, the use-libds flag is already
	 * setup to check whether the ds library can be initialized.
	 */
	if (IS_CONTROL_DOMAIN(ldp)) {
		if (libds_cap.functional)
			return (true);
		else
			return (false);
	}

	/*
	 * Check if the vldc end of the port exist
	 */
	if (ds_port != NULL) {
		vldc_port_t	*vldc_port = ds_port->vldc_port;
		if (vldc_port != NULL && vldc_port->peer.ds == ds_port)
			return (false);
	} else {
		return (false);
	}

	/*
	 * If the ds port to ds port connection is set up, libds is the
	 * only possible mode to communicate
	 */
	if (ds_port->peer_ds_port->peer_ds_port != ds_port)
		return (false);

	if (libds_cap.config && !libds_cap.functional) {
		warning("%s: ds ports are setup, but no libds support "
		    "is available, DS will be disabled\n", __func__);
	}

	return (true);
}


int
open_ldc_channel(ds_channel_t *chan)
{
	char *chandev;
	vldc_opt_op_t op;
	int fd;

	if (strcmp(chan->name, SP_CHAN_NAME) == 0) {
		/* communication with SP */
		chan->send_latency = 10000; 	/* 10 seconds */
		chan->send_inverse_bw = 100; 	/* 100 milliseconds/KB */
		chan->send_delay = 10;		/* 10 milliseconds */
		chan->send_retries = 5;
	} else if (strcmp(chan->name, FMA_CHAN_NAME) == 0) {
		/* communication with FMD */
		chan->send_latency = 1000; 	/* 1 second */
		chan->send_inverse_bw = 0;
		chan->send_delay = 10;		/* 10 milliseconds */
		chan->send_retries = 5;
	} else if (is_ds_channel(chan)) {
		/* communication with LDom */
		chan->send_latency = 100; 	/* 100 milliseconds */
		chan->send_inverse_bw = 0;
		chan->send_delay = 10;		/* 10 milliseconds */
		chan->send_retries = 5;
	} else {
		return (-1);
	}

	if ((chandev = get_channel_devname(chan)) == NULL) {
		warning("%s: Could not get device name for channel %s",
		    __func__, chan->name);
		return (-1);
	}

	if ((fd = open(chandev, O_RDWR)) == -1)
		ewarning("%s: Open of %s failed", __func__, chandev);
	Xfree(chandev);
	if (fd == -1)
		return (-1);

	op.op_sel = VLDC_OP_SET;
	op.opt_sel = VLDC_OPT_MODE;
	op.opt_val = LDC_MODE_STREAM;

	if (do_ioctl(fd, VLDC_IOCTL_OPT_OP, &op) != 0) {
		ewarning("%s: Unable to set LDC channel %s to LDC_MODE_STREAM "
		    "mode", __func__, chan->name);
		close(fd);
		return (-1);
	}

	chan->fd = fd;
	return (0);
}
