/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Module implementing the 'agent-dio' DS protocol
 */

#include <strings.h>
#include <stddef.h>
#include <errno.h>

#include "basics.h"
#include "allocate.h"
#include "zprintf.h"
#include "ds.h"
#include "ds_impl.h"
#include "ds_registry.h"
#include "support.h"
#include "fatal.h"
#include "core_md.h"
#include "ldma.h"
#include "pcie.h"

#ifdef ZDBG
#define	DIO_MDFILE	DIO_SVC
#endif

#define	DIO_SVC_TIMEOUT	(60)	/* timeout value - Units: seconds */


static uint64_t dsdio_seq = 0;

static ds_ver_t dsdio_vers[] = { {1, 0} };
#define	DSDIO_NVERS  (sizeof (dsdio_vers) / sizeof (ds_ver_t))

static ds_capability_t dsdio_cap = {
	DIO_SVC,		/* svc_id */
	dsdio_vers,	/* vers */
	DSDIO_NVERS	/* nvers */
};

static ds_ops_t dsdio_libops = {
	dslib_register_handler,		/* ds_reg_cb */
	dslib_unregister_handler,	/* ds_unreg_cb */
	dslib_data_handler,		/* ds_data_cb */
	DIO_SVC				/* cb_arg */
};

static libds_info_t dsdio_info[] = {
	{
		DS_CLIENT,
		&dsdio_cap,
		&dsdio_libops
	}
};

static void *
dsdio_getinfo_thread(void *arg)
{
	ldom_t *ldp = arg;

	if (ldp->state == ACTIVE) {
		DBG2(zprintf("%s: before calling %s\n", __func__, DIO_SVC))
		pcie_get_devinfo(ldp);
		DBG(zprintf("%s: after calling %s\n", __func__, DIO_SVC));
	}
	return (NULL);
}

void
dsdio_reg_handler(struct ds_svc *svc)
{
	ldom_t *ldp = svc->chan->ldp;

	DBG(zprintf("%s: %s reg handler called for %s\n",
	    __func__, DIO_SVC, ldp->namep));

	/* For now, let's query the control domain only */
	if (!IS_CONTROL_DOMAIN(ldp)) {
		return;
	}

	/*
	 * Direct I/O agent is online, start a thread to obtain the
	 * latest device info asynchronously and update the info
	 * accordingly.
	 */
	if (thr_create(NULL, NULL, dsdio_getinfo_thread, ldp,
	    THR_DAEMON, NULL) != 0) {
		warning("Could not create %s getinfo thread", DIO_SVC);
	}
}

static int
dsdio_send_request(ldom_t *ldp, uint32_t msg_type, ds_msg_t *resp)
{
	ds_msg_t		ds_msg;
	ds_req_t		*msg;
	ldma_message_header_t	*dp_reqp;
	ldma_message_t		*dp_resp;
	size_t			msg_len, data_len;
	ds_svc_t		*svcp;
	int			rv = 0;

	svcp = ds_lookup_svc(ldp->ds_chan, DIO_SVC);
	if ((svcp == NULL) || (svcp->hdl == DS_SVC_HDL_INVALID)) {
		DBG(zprintf("%s: %s %s not yet registered\n",
		    __func__, ldp->namep, DIO_SVC));
		return (EIO);
	}

	data_len = DSDIO_HDR_SIZE;
	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) + data_len;

	msg = Xcalloc(msg_len, uint8_t);

	ds_fill_data_header(msg, svcp->hdl, data_len);

	/* Fill MSGDIO_PCIDEV_INFO */
	dp_reqp = (ldma_message_header_t *)DS_DATA_BUF(msg);

	dp_reqp->msg_type = hton32(msg_type);
	dp_reqp->msg_num = hton64(dsdio_seq++);
	dp_reqp->msg_info = 0;

	ds_msg.buf = msg;
	ds_msg.len = msg_len;
	ds_msg.seq = dp_reqp->msg_num;

	DBG(zprintf("%s: %s <0x%"PRIx64"> sending pci device info "
	    "request %"PRId64"\n", __func__, ldp->namep,
	    svcp->hdl, ds_msg.seq));

	ds_send_data_msg(ldp->ds_chan, &ds_msg, DIO_SVC_TIMEOUT, resp);


	/* a response is expected */
	if (resp->buf == NULL) {
		DBG(zprintf("%s: no response "
		    "received from pci devinfo request\n", __func__));
		rv = EIO;
		goto done;
	}

	dp_resp = (ldma_message_t *)((char *)resp->buf +
	    sizeof (ds_data_handle_t));

	/* sanity check the response */
	if ((dp_reqp->msg_num != dp_resp->msg_hdr.msg_num) ||
	    (dp_resp->msg_hdr.msg_type != LDMA_MSG_RESULT)) {
		DBG(zprintf("%s: Failure "
		    "response for request num=<0x%"PRIx64"> err=%d\n", __func__,
		    dp_reqp->msg_num, dp_resp->msg_hdr.msg_info));
		Xfree(resp->buf);
		resp->buf = NULL;
		rv = EBADMSG;
	}

done:
	Xfree(msg);
	return (rv);
}

int
dsdio_get_devinfo(ldom_t *ldp, void **bufpp, uint32_t *size)
{
	ds_msg_t	resp;
	int		rv;
	int		hlen;
	ldma_message_t	*dp_resp;

	DBG(zprintf("%s...\n", __func__));

	rv = dsdio_send_request(ldp, MSGDIO_PCIDEV_INFO, &resp);
	if (rv != 0) {
		DBG(zprintf("%s: error "
		    "sending MSGDIO_PCIDEV_INFO request (%d)\n", __func__, rv));
		return (rv);
	}

	dp_resp = (ldma_message_t *)((char *)resp.buf +
	    sizeof (ds_data_handle_t));

	hlen = sizeof (ds_data_handle_t) + DSDIO_HDR_SIZE;

	if ((resp.len <= hlen) || (dp_resp->msg_hdr.msg_info >
	    (resp.len - hlen))) {
		DBG(zprintf("%s: malformed message len=%d expected=%d\n",
		    __func__, resp.len, dp_resp->msg_hdr.msg_info));
		Xfree(resp.buf);
		return (ENXIO);
	}
	*bufpp = Xmalloc(dp_resp->msg_hdr.msg_info);
	*size = dp_resp->msg_hdr.msg_info;
	memcpy(*bufpp, LDMA_HDR2DATA(dp_resp), dp_resp->msg_hdr.msg_info);

	DBG(dump_to_file(LDMA_HDR2DATA(dp_resp), dp_resp->msg_hdr.msg_info,
	    "dump-%s.bin", DIO_MDFILE));

	Xfree(resp.buf);
	return (0);
}


void
dsdio_data_handler(ds_svc_t *svc, void *buf, size_t msglen)
{
	ds_msg_tgt_t	tgt;
	ds_msg_t	resp;
	ldma_message_t	*dp_resp;

	if (msglen < sizeof (ds_data_handle_t) + DSDIO_HDR_SIZE) {
		DBG(zprintf("%s: dropping invalid response of %d bytes, "
		    "expected >= %d\n", __func__, msglen,
		    sizeof (ds_data_handle_t) + DSDIO_HDR_SIZE));
		Xfree(buf);
		return;
	}

	dp_resp = (ldma_message_t *)((char *)buf + sizeof (ds_data_handle_t));

	tgt.chan = svc->chan;
	tgt.ds_hdl = svc->hdl;

	resp.buf = buf;
	resp.len = msglen;
	resp.seq = ntoh64(dp_resp->msg_hdr.msg_num);

	/*
	 * Check the pending queue for a matching message and
	 * signal the waiting thread
	 */
	if (ds_check_resp(&tgt, &resp) == DS_MSG_CLAIMED) {
		/* message claimed */
		return;
	}
	DBG(zprintf("%s: dropping %s response %"PRIu64"\n",
	    __func__, DIO_SVC, resp.seq));
	Xfree(buf);
}

/*
 * The versions supported
 */
static prot_version_t dsdio_versions[] = { {1, 0} };

static svc_ops_t dsdio_ops = {
	dsdio_reg_handler,		/* reg_handler */
	NULL,				/* unreg_handler */
	dsdio_data_handler,		/* data_handler */
	NULL,				/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	int i;
	svc_info_t *info;

	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 2;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(DS_CHAN_NAME);
	info->chan_names[1] = Xstrdup(LIBDS_CHAN_NAME);
	info->num_svcs = 1;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(DIO_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = 0;
	info->num_vers = NUM_VERS(dsdio_versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	memcpy(info->versions, dsdio_versions,
	    info->num_vers * sizeof (prot_version_t));

	/*
	 * copy the info structure for each service
	 */
	info->libds_svcs = Xmalloc(info->num_svcs * sizeof (libds_info_t *));
	for (i = 0; i < info->num_svcs; i++) {
		info->libds_svcs[i] = cp_dsinfo(&dsdio_info[i], info->svcs[i]);
	}

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, DIO_SVC) == 0)
		return (&dsdio_ops);
	else
		return (NULL);
}
