/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_DS_IMPL_H
#define	_DS_IMPL_H

#pragma ident	"@(#)ds_impl.h	1.13	09/06/22 SMI"

#ifdef	__cplusplus
extern "C" {
#endif

#include "ds.h"
#include "ldom.h"
#include "support.h"
#include "svcs.h"
#include "ds_registry.h"

/*
 * DS service handle.
 */
typedef uint64_t ds_handle_t;

#define	DS_SVC_HDL_INVALID	0x0	/* invalid domain service handle */
#define	DS_MAX_HDLS	256

/*
 * Internal states for a domain service.
 */
typedef enum {
	SVC_INACTIVE,			/* svc not registered */
	SVC_REG_PENDING,		/* register message sent */
	SVC_ACTIVE			/* register message acknowledged */
} ds_state_t;

typedef enum {
	SVC_UNREG_NONE	= 0,		/* Not marked for unregistration */
	SVC_UNREG_HDL	= 1,		/* unregistration of handle pending */
	SVC_UNREG_LDOM	= 2		/* unregistration of domain pending */
} ds_unreg_t;

typedef struct dslib_svc_aggr dslib_svc_aggr_t;
/*
 * DS service descriptor.
 */
typedef struct ds_svc {
	char		*name;
	ds_handle_t	hdl;
	ds_state_t	state;
	ds_unreg_t	unreg;		/* unregistration state */
	prot_version_t	version;	/* agreed upon version */
	int		ver_idx;	/* protocol version negotiation index */
	struct ds_channel *chan;	/* channel service is attached to */
	svc_ops_t	*ops;
	svc_t		*svc;
	void		*plugin_dlhdl;	/* linker handle to services plugin */
	int		refcnt;
	mutex_t		lock;		/* protects refcnt */
	CHAIN_LINK(struct ds_svc, link); /* links active services on channel */
	CHAIN_LINK(struct ds_svc, aggr_link);
	dslib_svc_aggr_t *svc_aggr;	/* service aggregatror */
} ds_svc_t;

/*
 * The service aggregator is a collection of identical services
 * being provided to different domains, presumably with different
 * versions. The reason we keep this is because there is a single
 * callback handler catering to all the individual instances of
 * the service to different LDoms.
 */
struct dslib_svc_aggr {
	char *svc_id;
	mutex_t	lock;
	CHAIN(ds_svc_t, dslib_receivers);
	CHAIN_LINK(dslib_svc_aggr_t, link);
};

/*
 * Conservative upper bound on the size of a domain services message.
 */
#define	DS_MAX_MSG_SIZE		(64*1024*1024)

#define	DS_MAX_SVCID_LEN	(50)

#define	DS_NO_BLOCK		(0)
#define	DS_BLOCK_FOREVER	(-1)

#define	DS_MSG_CLAIMED		1
#define	DS_MSG_UNCLAIMED	2

struct ds_channel;

/*
 * The target of a message.
 */
typedef struct {
	struct ds_channel *chan;
	ds_handle_t	ds_hdl;
} ds_msg_tgt_t;

/*
 * A message buffer.
 */
typedef struct {
	void		*buf;
	int		len;
	uint64_t	seq;
} ds_msg_t;

/*
 * For enqueuing a DS request on the sequencer queue.
 */
typedef struct {
	ds_svc_t	*svc;
	uint8_t		*msg;
	uint16_t	len;
} ds_seq_req_t;

typedef enum {
	DS_MSG_COMPLETE = 0,
	DS_MSG_ERROR = 1,
	DS_MSG_PENDING = 2
} ds_pend_stat_t;

/*
 * Details about a DS message sent to an LDom for which
 * a response is expected. Only zeus initiated request
 * messages are kept in a queue so that the response
 * from the guests can be correlated.
 */
typedef struct ds_pend_msg {
	ds_handle_t	ds_handle;	/* handle of destination */
	uint64_t	seq;		/* seq number of the command */
	ds_pend_stat_t	status;		/* completion status */
	uint8_t		*msg;		/* message body */
	uint16_t	len;		/* message length */
	CHAIN_LINK(struct ds_pend_msg, link);
} ds_pend_msg_t;

typedef enum {
	CHAN_DOWN,			/* channel down */
	CHAN_RESET,			/* channel has been reset */
	CHAN_OPEN,			/* channel open waiting for up event */
	CHAN_INIT_REQ,			/* initialization handshake sent */
	CHAN_UP,			/* channel up */
} chan_state_t;

typedef struct ds_channel {
	char		*name;		/* channel name */
	ldom_t		*ldp;
	chan_state_t	state;		/* channel state */
	int		props;		/* channel properties */
	bool_t		unregister;	/* flag in process of unregistering */
	CHAIN_LINK(struct ds_channel, link); /* next channel */
	CHAIN(ds_svc_t, ds_svcs);	/* active services (excl dslib) */
	svc_t		*reg_svc;	/* next service to register */
	channel_plugins_t *plugins;
	prot_version_t	*versions;	/* protocol versions supported */
	int		num_vers;	/* # of protocol versions supported */
	int		ver_idx;	/* protocol version negotiation index */
	prot_version_t	version;	/* protocol version in use */
	int		usecnt;		/* in use count */
	int		msgcnt;		/* number of outstanding messages */
	struct {
		mutex_t lock;
		cond_t	cond;
		CHAIN(ds_pend_msg_t, msgs);
	} pending;
	mutex_t		lock;		/* protects state, counts, svcs list */
	cond_t		cond;
	/* callback to invoke when event occurs */
	void		(*callback)(struct ds_channel *);
	int		evport;		/* channel event port */
	int		fd;		/* channel file descriptor */
	uint32_t	len;		/* length of data transfer */
	uint32_t	left;		/* # of bytes left to transfer */
	char		*buf;		/* data buffer */
	char		*bufp;		/* current position in data buffer */
	ds_hdr_t	hdr;		/* space for a message header */
	int		send_latency;	/* send timeout latency (millisecs) */
	int		send_inverse_bw; /* inverse bandwidth (millisecs/KB) */
	int		send_delay;	/* send delay (millisecs) */
	int		send_retries;	/* send retries before delay */
} ds_channel_t;

#define	CHAN_OUTGOING	0x1

#define	SP_CHAN_NAME		"spds"
#define	FMA_CHAN_NAME		"fmactl"
#define	DS_CHAN_NAME		"ds"
#define	LIBDS_CHAN_NAME		"libds"

/*
 * For enqueuing a message on the manager services queue.
 */
typedef struct ms_qmsg {
	CHAIN_LINK(struct ms_qmsg, link);	/* next in queue */
	ds_hdr_t	hdr;			/* message header */
	ds_channel_t	*chan;			/* channel message came in on */
	void		*msg;			/* the message sans header */
} ms_qmsg_t;

/*
 * Manager services queue.
 */
struct msg_queue {
	mutex_t	lock;			/* protects add/remove from queue */
	cond_t	cond;			/* used to wait on queue not empty */
	CHAIN(ms_qmsg_t, msgs);		/* queue of msgs */
};

/*
 * Message access and size macros.
 */
#define	DS_SIZE(x)		(sizeof (ds_hdr_t) + (x)->hdr.payload_len)
#define	DS_DATA_BUF_SZ(x)	\
	((x) - sizeof (ds_hdr_t) - sizeof (ds_data_handle_t))

#define	DS_MSG_LEN(t) 		(sizeof (ds_hdr_t) + sizeof (t))
#define	DS_BUF(m)		((char *)(m) + DS_MSG_LEN(ds_data_handle_t))

#define	DS_REQ_TYPE(x)		(((ds_req_t *)(x))->hdr.msg_type)
#define	DS_REQ_PAY_LEN(x)	(((ds_req_t *)(x))->hdr.payload_len)
#define	DS_DATA_REQ(x)		(((ds_req_t *)(x))->payload.data)
#define	DS_DATA_SVC_HANDLE(x)	(DS_DATA_REQ(x).datahdr.svc_handle)
#define	DS_DATA_BUF(x)		(DS_DATA_REQ(x).bytes)

/*
 * Prototypes.
 */
void ds_init1(void);
void ds_init2(void);
void ds_comm_init(void);

int ds_register_domain(ldom_t *ldp, bool_t open_chan);
void ds_unregister_domain(ldom_t *ldp);
void ds_unregister_channel(ds_channel_t *chan);

int ds_open_channel(ds_channel_t *chan);
void ds_close_channel(ds_channel_t *chan);
int ds_reset_channel(ds_channel_t *chan);
void ds_free_channel(ds_channel_t *chan);
bool_t is_ds_channel(ds_channel_t *chan);
bool_t is_libds_channel(ds_channel_t *chan);
ds_channel_t *ds_sp_channel(void);
char *check_vldc_chan_by_name(char *chan_name, char *ldom_name);
bool_t check_vldc_chan_pres(char *chan_name);
bool_t check_for_libds(ldom_t *ldp);
bool_t check_for_ds_chan(ldom_t *ldp);

int ds_send_data_msg(ds_channel_t *chan, ds_msg_t *msgp, int tout,
    ds_msg_t *resp);
int ds_chan_send_msg(ds_channel_t *chan, void *msg, size_t len);
int ds_check_resp(ds_msg_tgt_t *tgtp, ds_msg_t *resp);
void ds_fill_data_header(void *reqp, uint64_t svc_handle, uint32_t data_len);

void ds_enqueue_msg(ds_channel_t *chan, ds_hdr_t *hdr, void *msg);
void ds_channel_free_queued(ds_channel_t *chan);
void ds_channel_reg_req(ds_channel_t *chan);
void ds_channel_init_req(ds_channel_t *chan);
int ds_chan_ready_read(ds_channel_t *chan);

void *ds_service_routine(ds_channel_t *chan, char *service, char *function);
void *ds_channel_service_routine(ds_channel_t *chan, char *service,
    char *function);

ds_svc_t *ds_add_svc(ds_channel_t *chan, char *name, ds_handle_t hdl,
    uint16_t major, uint16_t minor);
void ds_rem_svc(ds_channel_t *chan, ds_handle_t hdl);
void ds_svc_incref(ds_svc_t *ds);
void ds_svc_decref(ds_svc_t *ds, uint_t count);
ds_svc_t *ds_lookup_svc(ds_channel_t *chan, char *name);

typedef int (*update_md_f)(ldom_t *);

char *chan_state_str(int state);
void channel_error_notify_pending(ds_channel_t *chan, ds_hdl_t hdl);
#if ZDBG
extern void dump_libds_handle(ds_hdl_t hdl);
#endif

#ifdef	__cplusplus
}
#endif

#endif	/* _DS_IMPL_H */
