/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)fma_cpu_common.c	1.13	09/06/22 SMI"

/*
 * Common routines for the 'fma-cpu' and 'fma-phys-cpu-service' domain
 * services.
 */

#include <sys/processor.h>
#include <sys/types.h>
#include <alloca.h>
#include <errno.h>
#include <string.h>

#include "ldom.h"
#include "zpri.h"
#include "allocate.h"
#include "zprintf.h"
#include "ds.h"
#include "ds_impl.h"
#include "ds_registry.h"
#include "support.h"
#include "fma_cpu.h"
#include "fatal.h"
#include "svcs.h"
#include "vcpu.h"
#include "pm_events.h"
#include "core_md.h"
#include "sequencer.h"

static int
get_md_op_status(processorid_t pid)
{
	int		status, fma_status;

	status = get_core_md_cpu_status(pid, CORE_MD_OP_STATUS);

	if (status == -1) {
		/* pid is not found or has no FMA status */
		return (status);
	}

	/*
	 * Must switch from P_ONLINE/OFFLINE definitions to FMA_CPU_STAT
	 * definitions.
	 */
	switch (status) {
	case P_ONLINE:
		fma_status = FMA_CPU_STAT_ONLINE;
		break;
	case P_OFFLINE:
		fma_status = FMA_CPU_STAT_OFFLINE;
		break;
	}

	return (fma_status);
}

static int
set_md_op_status(processorid_t pid, int cpu_status)
{
	int		status, md_status;

	/*
	 * Must switch from FMA_CPU_STAT definitions to P_ONLINE/OFFLINE
	 * definitions.
	 */
	switch (cpu_status) {
	case (FMA_CPU_STAT_ONLINE):
		md_status = P_ONLINE;
		break;
	case (FMA_CPU_STAT_OFFLINE):
		md_status = P_OFFLINE;
		break;
	}

	status = set_core_md_cpu_status((uint64_t)pid, CORE_MD_OP_STATUS,
	    md_status);

	/* ensure status is persisted (in HV MD) */
	seq_update_hvmd();

	return (status);
}

typedef int (*ds_set_p_status_f)(ldom_t *, processorid_t vid,
    fma_cpu_service_req_t *, fma_cpu_resp_t *);

int
fma_cpu_svc_set_p_status(fma_cpu_service_req_t *req, fma_cpu_resp_t *resp)
{
	ldom_t *ldp;
	processorid_t vid;
	uint32_t cpu_id;
	ds_set_p_status_f ds_set_p_status;
	pm_resid_t resid;
	char *res_name = "cpu";

	cpu_id = ntoh32(req->cpu_id);

	if (pid_to_vid(cpu_id, &ldp, &vid)) {

		if (ldp == NULL) {
			DBG(zprintf("%s: CPU %u not assigned to a LDom\n",
			    __func__, cpu_id));
			return (-1);
		}

		DBG(zprintf("%s: %s, pid %d, vid %d\n", __func__, ldp->namep,
		    cpu_id, vid));

		if ((pm_fma_events_on) &&
		    (req->msg_type == FMA_CPU_REQ_ONLINE)) {
			resid.pi_res_type = PM_RES_CPU;
			resid.pi_res_name = res_name;
			resid.u.cpu.pi_id_val = req->cpu_id;
			pm_fma_event_notify((void *)ldp, resid,
			    RESOURCE_UNRETIRE_EVENT);
		}

		ds_set_p_status = (ds_set_p_status_f)ds_service_routine(
		    ldp->ds_chan, "fma-cpu-service", "fma_cpu_set_p_status");
		if (ds_set_p_status == NULL) {
			warning("%s: Can't find fma_cpu_set_p_status routine",
			    __func__);
			return (-1);
		}

		if (ds_set_p_status(ldp, vid, req, resp) != 0)
			return (-1);
		if ((pm_fma_events_on) &&
		    (req->msg_type == FMA_CPU_REQ_OFFLINE)) {
			resid.pi_res_type = PM_RES_CPU;
			resid.pi_res_name = res_name;
			resid.u.cpu.pi_id_val = req->cpu_id;

			pm_fma_event_notify((void *)ldp, resid,
			    RESOURCE_RETIRE_EVENT);
		}
	} else {
		/* fill in the old MD status */
		resp->result = hton32(FMA_CPU_RESP_OK);
		resp->status = hton32(get_md_op_status(cpu_id));
	}

	switch (ntoh32(req->msg_type)) {
	case (FMA_CPU_REQ_ONLINE):
		set_md_op_status(cpu_id, FMA_CPU_STAT_ONLINE);
		break;
	case (FMA_CPU_REQ_OFFLINE):
		set_md_op_status(cpu_id, FMA_CPU_STAT_OFFLINE);
		break;
	}

	return (0);
}

typedef int (*ds_get_p_status_f)(ldom_t *, processorid_t vid,
    fma_cpu_service_req_t *, fma_cpu_resp_t *);

int
fma_cpu_svc_get_p_status(fma_cpu_service_req_t *req, fma_cpu_resp_t *resp)
{
	ldom_t *ldp;
	processorid_t vid;
	uint32_t cpu_id;
	uint32_t cpu_md_status, cpu_status;
	ds_get_p_status_f ds_get_p_status;
	vcpu_t *vcpu;
	static bool_t warn_msg = false;

	cpu_id = ntoh32(req->cpu_id);

	/* get_md_op_status returns status in FMA_CPU_STAT format */
	cpu_md_status = get_md_op_status(cpu_id);

	/* This shouldn't happen but just in case */
	if (cpu_md_status == -1) {
		return (-1);
	}

	/*
	 * FIXME: pid_to_vid() should really include this flag_isconfigd check
	 */
	if (pid_to_vid(cpu_id, &ldp, &vid) &&
	    ((vcpu = vid_to_vcpu(ldp, vid)) != NULL) &&
	    vcpu->flag_isconfigd) {

		if (ldp == NULL) {
			DBG(zprintf("%s: CPU %u not assigned to a LDom\n",
			    __func__, cpu_id));
			return (-1);
		}

		DBG(zprintf("%s: %s, pid %d, vid %d\n",
		    __func__, ldp->namep, cpu_id, vid));

		/*
		 * If allocated cpu is marked as being power-managed,
		 * return status from the core MD.
		 */
		if (get_core_md_cpu_status(cpu_id, CORE_MD_PM_STATUS) ==
		    PM_MANAGED) {
			resp->result = hton32(FMA_CPU_RESP_OK);
			resp->status = hton32(cpu_md_status);
			return (0);
		}

		ds_get_p_status = (ds_get_p_status_f)ds_service_routine(
		    ldp->ds_chan, "fma-cpu-service", "fma_cpu_get_p_status");
		if (ds_get_p_status == NULL) {
			/* Only display this message once in the LDoms log */
			if (!warn_msg) {
				warning("%s: FMA cpu operations to some domains"
				    " may not be not available", __func__);
				warn_msg = true;
			}
			return (-1);
		}

		if (ds_get_p_status(ldp, vid, req, resp) != 0)
			return (-1);

		/*
		 * Double check to see if the MD status is the same as what
		 * the actual CPU status is. If they are different then update
		 * the MD.
		 */
		cpu_status = ntoh32(resp->status);
		if (cpu_status != cpu_md_status)
			set_md_op_status(cpu_id, cpu_status);
	} else {
		/* Use the MD status for unallocated resources */
		resp->result = hton32(FMA_CPU_RESP_OK);
		resp->status = hton32(cpu_md_status);
	}

	return (0);
}
