/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)fma_cpu_svc.c	1.9	09/02/09 SMI"

/*
 * Module implementing the 'fma-phys-cpu-service' LDom manager service.
 */

#include <sys/processor.h>
#include <sys/types.h>
#include <alloca.h>
#include <errno.h>
#include <string.h>

#include "ldom.h"
#include "zpri.h"
#include "allocate.h"
#include "zprintf.h"
#include "ds.h"
#include "ds_impl.h"
#include "ds_registry.h"
#include "support.h"
#include "fma_cpu.h"
#include "fatal.h"
#include "sequencer.h"
#include "svcs.h"

static void
cpu_data_handler(ds_svc_t *svc, void *msg, size_t msglen)
{
	ds_seq_req_t *dsreq;

	dsreq = Xmalloc(sizeof (*dsreq));
	dsreq->svc = svc;
	dsreq->msg = msg;
	dsreq->len = msglen;

	seq_enqueue_ds_req(dsreq);
}

static void
cpu_seq_handler(ds_svc_t *svc, void *msg, size_t msglen)
{
	fma_cpu_service_req_t *cpu_req;
	fma_cpu_resp_t *cpu_resp;
	void *resp;
	size_t resp_len, data_len;
	int rv;

	if (msglen != sizeof (ds_data_handle_t) +
	    sizeof (fma_cpu_service_req_t)) {
		return;
	}

	cpu_req = (fma_cpu_service_req_t *)((char *)msg +
	    sizeof (ds_data_handle_t));

	data_len = sizeof (fma_cpu_resp_t);
	resp_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) + data_len;
	resp = Xmalloc(resp_len);
	cpu_resp = (fma_cpu_resp_t *)((char *)resp + sizeof (ds_hdr_t) +
	    sizeof (ds_data_handle_t));

	if (ntoh32(cpu_req->msg_type) == FMA_CPU_REQ_STATUS) {
		rv = fma_cpu_svc_get_p_status(cpu_req, cpu_resp);
	} else {
		rv = fma_cpu_svc_set_p_status(cpu_req, cpu_resp);
	}

	if (rv == 0) {
		ds_fill_data_header(resp, svc->hdl, data_len);
		cpu_resp->req_num = cpu_req->req_num;

		ds_chan_send_msg(svc->chan, resp, resp_len);
	} else {
		/*
		 * Status call failed. No response is sent and
		 * the client is left to time out.
		 */
		DBG(zprintf("%s: bad status, rv %d\n", __func__, rv));
	}

	Xfree(resp);
}

static prot_version_t versions[] = {
	{ 1, 0 }
};

static svc_ops_t fma_cpu_service_ops = {
	NULL,				/* reg_handler */
	NULL,				/* unreg_handler */
	cpu_data_handler,		/* data_handler */
	cpu_seq_handler,		/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	svc_info_t *info;

	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 2;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));

	info->chan_names[0] = Xstrdup(FMA_CHAN_NAME);
	info->chan_names[1] = Xstrdup(SP_CHAN_NAME);

	info->num_svcs = 1;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(FMA_PHYS_CPU_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = SVC_OUTGOING;
	info->num_vers = NUM_VERS(versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	memcpy(info->versions, versions,
	    info->num_vers * sizeof (prot_version_t));

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, FMA_PHYS_CPU_SVC) == 0)
		return (&fma_cpu_service_ops);
	else
		return (NULL);
}
