/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)fma_io_domain.c	1.6	09/03/05 SMI"

/*
 * Module implementing the 'fma-io-domain-service' DS protocol
 */

#include <sys/types.h>
#include <sys/processor.h>
#include <alloca.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>

#include "ldom.h"
#include "zpri.h"
#include "allocate.h"
#include "zprintf.h"
#include "ds.h"
#include "ds_impl.h"
#include "ds_registry.h"
#include "fma_io_domain.h"
#include "support.h"
#include "fatal.h"
#include "sequencer.h"
#include "svcs.h"

static ldom_t *
get_ldom_by_rc_name(char *io_rc_name)
{
	bound_iodevice_t *biodevs_p;
	ldom_t *ldp;

	CHAIN_ITER(zeus.ldoms, ldp, link) {

		CHAIN_ITER(ldp->bound.iodevices, biodevs_p, ldom) {
			if (biodevs_p->phys_iodev->cli_name != NULL) {
				if (strcmp(io_rc_name, biodevs_p->name) == 0)
					return (ldp);
			}
		}
	}

	return (NULL);
}

static pri_iodevice_t *
get_io_device_by_name(char *io_rc_name)
{
	pri_iodevice_t *pio_p;

	CHAIN_ITER(pri.iodevice_list, pio_p, link) {
		if (pio_p->name != NULL) {
			if (strcmp(pio_p->name, io_rc_name) == 0)
				return (pio_p);
		}

	}

	return (NULL);
}

static void *
form_resp(uint8_t io_rv, uint64_t req_num, uint64_t virt_rsrc_addr,
    ldom_t *ldp)
{
	void *resp;
	fma_io_resp_t *io_resp;
	int resp_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) +
	    sizeof (fma_io_resp_t);

	if (ldp != NULL)
		resp_len += strlen(ldp->namep) + 1;

	resp = Xmalloc(resp_len);

	io_resp = (fma_io_resp_t *)((char *)resp + sizeof (ds_hdr_t) +
	    sizeof (ds_data_handle_t));

	io_resp->req_num = hton64(req_num);
	io_resp->result = hton32(io_rv);
	io_resp->virt_rsrc_address = virt_rsrc_addr;

	if (ldp != NULL) {
		io_resp->dom_id = ldp->idp->id;
		memcpy((void *)((char *)io_resp + sizeof (fma_io_resp_t)),
		    ldp->namep, strlen(ldp->namep) + 1);
	} else {
		io_resp->dom_id = ILLEGAL_DOM_ID;
	}

	return (resp);
}

static void
fma_io_svc_get_ldom(fma_io_service_req_t *io_req, void **resp, size_t *data_len)
{

	pri_iodevice_t *pio_p = NULL;
	char io_rc_name[20];
	ldom_t *ldp;
	int result = FMA_IO_RESP_ILLEGAL;

	*data_len = sizeof (fma_io_resp_t);

	switch (ntoh32(io_req->msg_type)) {

	case FMA_PCI_REQ_LDOM:
		sprintf(io_rc_name, "pci@%"PRIx64,
		    ntoh64(io_req->rsrc_address));
		break;

	case FMA_NIU_REQ_LDOM:
		sprintf(io_rc_name, "niu@%"PRIx64,
		    ntoh64(io_req->rsrc_address));
		break;

	default:
		/* unknown msg type */
		*resp = form_resp(FMA_IO_RESP_FAILURE,
		    ntoh64(io_req->req_num), ntoh64(io_req->rsrc_address),
		    NULL);
		return;
	}

	pio_p = get_io_device_by_name(io_rc_name);
	ldp = get_ldom_by_rc_name(io_rc_name);

	if (pio_p !=  NULL) {
		if (pio_p->free) {
			result = FMA_IO_RESP_UNASSIGNED;
		} else {
			if (ldp == NULL) {
				/* not on bound list */
				result = FMA_IO_RESP_UNASSIGNED;
			} else {
				if (ldp->state >= BOUND)
					result = FMA_IO_RESP_OK;
				else
					result = FMA_IO_RESP_UNASSIGNED;

				*data_len += strlen(ldp->namep)+1;
			}
		}
	} else if (zeus.mode == ZEUS_CONFIG_MODE) {
		result = FMA_IO_RESP_FAILURE;
	} else {
		result = FMA_IO_RESP_ILLEGAL;
	}

	*resp = form_resp(result, ntoh64(io_req->req_num),
	    ntoh64(io_req->rsrc_address), ldp);
}

static void
io_data_handler(ds_svc_t *svc, void *msg, size_t msglen)
{
	fma_io_service_req_t *io_req;
	void *resp;
	size_t resp_len, data_len;

	if (msglen != sizeof (ds_data_handle_t) +
	    sizeof (fma_io_service_req_t)) {
		return;
	}

	io_req = (fma_io_service_req_t *)((char *)msg +
	    sizeof (ds_data_handle_t));

	fma_io_svc_get_ldom(io_req, &resp, &data_len);

	resp_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) +
	    data_len;

	ds_fill_data_header(resp, svc->hdl, data_len);
	ds_chan_send_msg(svc->chan, resp, resp_len);

	Xfree(resp);
}

static prot_version_t versions[] = {
	{ 1, 0 }
};

static svc_ops_t fma_io_service_ops = {
	NULL,				/* reg_handler */
	NULL,				/* unreg_handler */
	io_data_handler, 		/* data_handler */
	NULL,				/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	svc_info_t *info;
	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 1;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(FMA_CHAN_NAME);
	info->num_svcs = 1;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(FMA_IO_DOMAIN_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = SVC_OUTGOING;
	info->num_vers = NUM_VERS(versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	memcpy(info->versions, versions,
	    info->num_vers * sizeof (prot_version_t));

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, FMA_IO_DOMAIN_SVC) == 0)
		return (&fma_io_service_ops);
	else {
		return (NULL);
	}
}
