/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)fma_mem.c	1.12	09/02/09 SMI"

/*
 * Module implementing the 'fma-mem-service' DS protocol
 */

#include <strings.h>
#include <sys/types.h>
#include <errno.h>

#include "ldom.h"
#include "allocate.h"
#include "zprintf.h"
#include "ds.h"
#include "ds_impl.h"
#include "ds_registry.h"
#include "fma_mem.h"

#define	FMA_MEM_DS_TIMEOUT	(10)

static uint32_t	fma_mem_seqno = 0;

static ds_ver_t fma_mem_vers[] = { {1, 0} };
#define	FMA_MEM_NVERS  (sizeof (fma_mem_vers) / sizeof (fma_mem_vers[0]))

static ds_capability_t fma_mem_cap = {
	FMA_MEM_SVC,	/* svc_id */
	fma_mem_vers,	/* vers */
	FMA_MEM_NVERS	/* nvers */
};

static ds_ops_t fma_mem_libops = {
	dslib_register_handler,		/* ds_reg_cb */
	dslib_unregister_handler,	/* ds_unreg_cb */
	dslib_data_handler,		/* ds_data_cb */
	FMA_MEM_SVC			/* cb_arg */
};

/*
 * Info needed to invoke ds_clnt_reg/ds_svc_reg when loading plugins
 * There could be multiple services in the same info structure
 */
static libds_info_t fma_mem_info[] = {
	{
		DS_CLIENT,
		&fma_mem_cap,
		&fma_mem_libops
	}
};

void
fma_mem_data_handler(ds_svc_t *svc, void *buf, size_t msglen)
{
	ds_msg_tgt_t		tgt;
	ds_msg_t		resp;
	fma_mem_resp_t		*mem_msg;

	if (msglen != sizeof (ds_data_handle_t) + sizeof (fma_mem_resp_t)) {
		zprintf("Invalid length fma-mem response message, %d bytes, "
		    "expected %d\n", msglen,
		    sizeof (ds_data_handle_t) + sizeof (fma_mem_resp_t));
		Xfree(buf);
		return;
	}

	tgt.chan = svc->chan;
	tgt.ds_hdl = svc->hdl;

	mem_msg = (fma_mem_resp_t *)((char *)buf + sizeof (ds_data_handle_t));

	resp.buf = buf;
	resp.len = msglen;
	resp.seq = ntoh64(mem_msg->req_num);

	if (ds_check_resp(&tgt, &resp) == DS_MSG_CLAIMED) {
		/* message already handled */
		return;
	}

	zprintf("asynchronous fma-mem message: not handled yet\n");
	Xfree(buf);
}

int
fma_mem_page_retire(ldom_t *ldp, uint64_t real_addr,
    fma_mem_service_req_t *send, fma_mem_resp_t *resp)
{
	ds_msg_t		msg;
	ds_msg_t		ds_msg_resp;
	ds_svc_t		*svcp;
	fma_mem_service_req_t	*mem_reqp;
	fma_mem_resp_t		*mem_resp;
	ds_req_t 		*ds_reqp;
	size_t 			msg_len;
	int 			rv;

	svcp = ds_lookup_svc(ldp->ds_chan, "fma-mem-service");
	if ((svcp == NULL) || (svcp->hdl == DS_SVC_HDL_INVALID)) {
		DBG(zprintf("fma-mem-service not yet registered for %s\n",
		    ldp->namep));
		return (-1);
	}

	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) +
	    sizeof (fma_mem_service_req_t);
	ds_reqp = Xmalloc(msg_len);
	DBG7(memset(ds_reqp, 0, msg_len));

	ds_fill_data_header(ds_reqp, svcp->hdl, sizeof (fma_mem_service_req_t));

	mem_reqp = (fma_mem_service_req_t *)DS_DATA_BUF(ds_reqp);

	mem_reqp->req_num = hton64(fma_mem_seqno++);
	mem_reqp->msg_type = send->msg_type;
	mem_reqp->addr = hton64(real_addr);
	mem_reqp->length = send->length;

	msg.buf = ds_reqp;
	msg.len = msg_len;
	msg.seq = mem_reqp->req_num;

	rv = ds_send_data_msg(ldp->ds_chan, &msg, FMA_MEM_DS_TIMEOUT,
	    &ds_msg_resp);
	Xfree(ds_reqp);

	if (rv == 0) {
		mem_resp = (fma_mem_resp_t *)((char *)ds_msg_resp.buf +
		    sizeof (ds_data_handle_t));

		resp->result = mem_resp->result;
		resp->status = mem_resp->status;
		resp->res_addr = mem_resp->res_addr;
		resp->res_length = mem_resp->res_length;

		Xfree(ds_msg_resp.buf);
	}

	return (rv);
}

/*
 * The versions supported
 */
static prot_version_t fma_mem_versions[] =
{
	{1, 0}
};

static svc_ops_t fma_mem_ops = {
	NULL,				/* reg_handler */
	NULL,				/* unreg_handler */
	fma_mem_data_handler,		/* data_handler */
	NULL,				/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	int i;
	svc_info_t *info;

	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 2;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(DS_CHAN_NAME);
	info->chan_names[1] = Xstrdup(LIBDS_CHAN_NAME);
	info->num_svcs = 1;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(FMA_MEM_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = 0;
	info->num_vers = NUM_VERS(fma_mem_versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	memcpy(info->versions, fma_mem_versions,
	    info->num_vers * sizeof (prot_version_t));

	/*
	 * copy the info structure for each service
	 */
	info->libds_svcs = Xmalloc(info->num_svcs * sizeof (libds_info_t *));
	for (i = 0; i < info->num_svcs; i++)
		info->libds_svcs[i] = cp_dsinfo(&fma_mem_info[i],
		    info->svcs[i]);

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, FMA_MEM_SVC) == 0)
		return (&fma_mem_ops);
	else
		return (NULL);
}
