/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)fma_mem_svc.c	1.9	09/02/09 SMI"

/*
 * Module implementing the 'fma-phys-mem-service' LDom manager service.
 */

#include <strings.h>
#include <sys/types.h>
#include <alloca.h>
#include <errno.h>

#include "ldom.h"
#include "allocate.h"
#include "zprintf.h"
#include "ds.h"
#include "ds_impl.h"
#include "ds_registry.h"
#include "sequencer.h"
#include "support.h"
#include "fma_mem.h"
#include "fatal.h"
#include "basics.h"
#include "memory.h"

static uint64_t
ra_to_pa(ldom_t *ldp, uint64_t ra)
{
	mblock_t *mblk;

	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		if ((ra >= mblk->raddr) && (ra < (mblk->raddr + mblk->size)))
			return (mblk->paddr + (ra - mblk->raddr));
	}

	return (0);
}

/* FIXME: assumes get_binding returns a single bound_mblock */
/*
 * check_binding() takes in a memory resource binding, determines if the
 * physical address falls within this binding, and then fills in the domains
 * real address to send to the domain service.
 */
static bool_t
check_binding(mblock_t *mblk, uint64_t pa, uint64_t *ra)
{
	if ((pa >= mblk->paddr) && (pa < mblk->paddr + mblk->size)) {
		*ra = mblk->raddr + (pa - mblk->paddr);
		return (true);
	}

	return (false);
}

/* FIXME Need to update this to handle memory block */
static bool_t
is_page_assigned_ldom(uint64_t pa, uint64_t *ra, ldom_t **ldp)
{
	mblock_t *mblk;

	CHAIN_ITER(zeus.ldoms, *ldp, link) {
		CHAIN_ITER((*ldp)->bound.memory, mblk, ldom) {
			if (check_binding(mblk, pa, ra))
				return (true);
		}
	}

	return (false);
}

typedef int (*ds_page_retire_f)(ldom_t *, uint64_t, fma_mem_service_req_t *,
    fma_mem_resp_t *);

static int
page_retire(fma_mem_service_req_t *send, fma_mem_resp_t *resp)
{
	int rv = -1;
	ldom_t *ldp;
	ds_page_retire_f ds_page_retire;
	uint64_t phys_addr, real_addr;

	phys_addr = ntoh64(send->addr);

	if (is_page_assigned_ldom(phys_addr, &real_addr, &ldp)) {
		if (ldp == NULL) {
			/* FIXME handle error */
			DBG(zprintf("%s: Page address 0x%llx not assigned to "
			    "a LDom\n", __func__, phys_addr));
			return (-1);
		}

		ds_page_retire = (ds_page_retire_f)ds_service_routine(
		    ldp->ds_chan, "fma-mem-service", "fma_mem_page_retire");
		if (ds_page_retire == NULL) {
			warning("%s: Can't find fma_mem_page_retire routine",
			    __func__);
			return (-1);
		}

		rv = ds_page_retire(ldp, real_addr, send, resp);
		if (rv != 0) {
			return (rv);
		} else {
			DBG(zprintf("resp address = 0x%llx\n",
			    resp->res_addr));
			/* convert from domain real addr to physical addr */
			phys_addr = ra_to_pa(ldp, resp->res_addr);
			/*
			 * If phys_addr is 0 from this call then the call to
			 * the domain service got an Illegal address sent back.
			 * This really shouldn't happen.
			 */
			ASSERT(phys_addr != 0);
			resp->res_addr = phys_addr;
			return (rv);
		}
	} else {
		resp->result = hton32(FMA_MEM_RESP_FAILURE);
		resp->status = hton32(FMA_MEM_STAT_ILLEGAL);
		resp->res_addr = send->addr;
		resp->res_length = send->length;

		return (0);
	}
}

static void
mem_data_handler(ds_svc_t *svc, void *msg, size_t msglen)
{
	ds_seq_req_t *dsreq;

	dsreq = Xmalloc(sizeof (*dsreq));
	dsreq->svc = svc;
	dsreq->msg = msg;
	dsreq->len = msglen;

	seq_enqueue_ds_req(dsreq);
}

static void
mem_seq_handler(ds_svc_t *svc, void *msg, size_t msglen)
{
	fma_mem_service_req_t *mem_req;
	fma_mem_resp_t *mem_resp;
	void *resp;
	size_t resp_len, data_len;
	int rv;

	if (msglen != sizeof (ds_data_handle_t) +
	    sizeof (fma_mem_service_req_t)) {
		return;
	}

	mem_req = (fma_mem_service_req_t *)((char *)msg +
	    sizeof (ds_data_handle_t));

	data_len = sizeof (fma_mem_resp_t);
	resp_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) + data_len;
	resp = Xmalloc(resp_len);
	mem_resp = (fma_mem_resp_t *)((char *)resp + sizeof (ds_hdr_t) +
	    sizeof (ds_data_handle_t));

	rv = page_retire(mem_req, mem_resp);

	if (rv == 0) {
		ds_fill_data_header(resp, svc->hdl, data_len);
		mem_resp->req_num = mem_req->req_num;

		ds_chan_send_msg(svc->chan, resp, resp_len);
	} else {
		/*
		 * Page retire failed. No response is sent and
		 * the client is left to time out.
		 */
		DBG(zprintf("%s: page retire failed, rv %d\n", __func__, rv));
	}

	Xfree(resp);
}

static prot_version_t versions[] = {
	{ 1, 0 }
};

static svc_ops_t fma_mem_service_ops = {
	NULL,				/* reg_handler */
	NULL,				/* unreg_handler */
	mem_data_handler,		/* data_handler */
	mem_seq_handler,		/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	svc_info_t *info;

	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 2;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(FMA_CHAN_NAME);
	info->chan_names[1] = Xstrdup(SP_CHAN_NAME);

	info->num_svcs = 1;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(FMA_PHYS_MEM_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = SVC_OUTGOING;
	info->num_vers = NUM_VERS(versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	memcpy(info->versions, versions,
	    info->num_vers * sizeof (prot_version_t));

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, "fma-phys-mem-service") == 0)
		return (&fma_mem_service_ops);
	else
		return (NULL);
}
