/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>

#include "allocate.h"
#include "ldom.h"
#include "frags.h"
#include "fatal.h"
#include "zprintf.h"
#include "ldomconfigure.h"
#include "mdesc.h"
#include "mdesc_impl.h"

/*
 * Number of concurrently allocated MDs needed to handle all reconfiguration
 * scenarios.
 */
#define	NUM_HV_MDS	4

int max_hv_md_size = FRAG_SIZE;

static bool_t have_hv_md_mblock = false;

static uint64_t hv_md_mblock_base;
static uint64_t hv_md_mblock_size;

static bound_mfrag_t hv_md_frags[NUM_HV_MDS];

bound_mfrag_t *
frag_alloc(ldom_t *ldp, int size)
{
	bound_mfrag_t *frag, *minfrag;

	if (CHAIN_LENGTH(zeus.free_frags) == 0) {
		/* no free frags, grab a new chunk for fragments */
		mblock_t *fmp;

		if ((fmp = mem_bind_phys(MEM_MINSIZE)) == NULL)
			fatal("Out of free memory for fragments");

		frag_free_mblock(fmp);
	}

	/*
	 * Find the frag with the lowest address. We do this because we need
	 * frags low in memory for boot HV and primary MDs. We should probably
	 * keep the frag list sorted for efficiency.
	 */

	minfrag = CHAIN_START(zeus.free_frags);
	CHAIN_ITER(zeus.free_frags, frag, link) {
		if (frag->paddr < minfrag->paddr)
			minfrag = frag;
	}

	CHAIN_UNLINK(zeus.free_frags, minfrag, link);
	CHAIN_LINK_INIT(minfrag->link);
	minfrag->ldomp = ldp;
	minfrag->ref_cnt = 1;
	DBG(zprintf("Allocating frag @ 0x%llx\n", minfrag->paddr));
	return (minfrag);
}

bound_mfrag_t *
frag_assign(ldom_t *ldp, uint64_t pa)
{
	bound_mfrag_t *frag;

	/* Find the matching frag - and hook it out */
	CHAIN_ITER(zeus.free_frags, frag, link) {
		if (frag->paddr == pa) {
			CHAIN_UNLINK(zeus.free_frags, frag, link);
			CHAIN_LINK_INIT(frag->link);
			frag->ldomp = ldp;
			ASSERT(frag->ref_cnt == 0);
			frag->ref_cnt = 1;
			DBG(zprintf("Assigning frag @ 0x%llx\n", frag->paddr));
			return (frag);
		}
	}

	return (NULL);
}

/*
 * Just bump the reference count & return the pointer passed in
 */
bound_mfrag_t *
frag_clone(bound_mfrag_t *orig_frag)
{
	orig_frag->ref_cnt++;
	return (orig_frag);
}

static void
frag_free_internal(bound_mfrag_t *frag)
{

	ASSERT(frag->ref_cnt > 0);
	if (--frag->ref_cnt == 0) {
		CHAIN_LINK_INIT(frag->link);
		CHAIN_LINK_INIT(frag->ldom);
		frag->ldomp = NULL;
#if ZDBG
		bound_mfrag_t *temp;
		CHAIN_ADD_SANITY(zeus.free_frags, frag, link, temp);
#else
		CHAIN_ADD(zeus.free_frags, frag, link);
#endif
	}
}

/*
 * Simply return the fragment to the free list.
 */
void
frag_free(bound_mfrag_t *frag)
{
	if (frag != NULL) {
		DBG(zprintf("Freeing frag @ 0x%llx\n", frag->paddr));
		frag_free_internal(frag);
	}
}

/*
 * Assign a block of memory to the frag free list.
 */
void
frag_free_mblock(mblock_t *mblk)
{
	bound_mfrag_t *frag;
	uint64_t pbase, psize;

	if (mblk == NULL)
		return;			/* OK to free a null pointer */
	CHAIN_LINK_INIT(mblk->ldom);

	/* All frag allocations are saved/restored in HV MD */
	CHAIN_ADD(zeus.frag_mblocks, mblk, ldom);

	pbase = mblk->paddr;
	psize = mblk->size;

	/* Check alignment and size */
	ASSERT((pbase & (FRAG_SIZE-1)) == 0);
	ASSERT((psize & (FRAG_SIZE-1)) == 0);

	/* Add fragments to free list */
	while (psize > 0) {
		frag = Xcalloc(1, bound_mfrag_t);
		frag->paddr = pbase;
		frag->size = FRAG_SIZE;
		frag->pmblockp = mblk->pmblk;
		frag->ref_cnt = 1; /* so frag_free_internal really frees it */

		frag_free_internal(frag);

		pbase += FRAG_SIZE;
		psize -= FRAG_SIZE;
	}
}

void
hv_md_alloc_init(uint64_t baseaddr, uint64_t size, uint64_t md_maxsize)
{
	int i;

	if (baseaddr % 64 != 0)
		fatal("HV MD mblock base address 0x%llx is not 64 byte aligned",
		    baseaddr);
	if (md_maxsize % 64 != 0)
		fatal("HV MD mblock md_maxsize 0x%llx is not a 64 byte "
		    "multiple", md_maxsize);
	if (NUM_HV_MDS * md_maxsize > size)
		fatal("HV MD mblock size 0x%llx is too small to contain %d "
		    "HV MDs", size, NUM_HV_MDS);

	for (i = 0; i < NUM_HV_MDS; i++) {
		hv_md_frags[i].paddr = baseaddr + i * md_maxsize;
		hv_md_frags[i].ref_cnt = 0;
	}

	have_hv_md_mblock = true;
	hv_md_mblock_base = baseaddr;
	hv_md_mblock_size = size;

	max_hv_md_size = md_maxsize;
}

bound_mfrag_t *
hv_md_alloc(void)
{
	int i;

	if (!have_hv_md_mblock)
		return (frag_alloc(NULL, FRAG_SIZE));

	for (i = 0; i < NUM_HV_MDS; i++) {
		if (hv_md_frags[i].ref_cnt == 0) {
			hv_md_frags[i].ref_cnt = 1;
			DBG(zprintf("Allocating hvmd @ 0x%llx\n",
			    hv_md_frags[i].paddr));
			return (&hv_md_frags[i]);
		}
	}

	fatal("Out of free memory for HV MDs");
	/*NOTREACHED*/
}

bound_mfrag_t *
hv_md_assign(uint64_t pa)
{
	int i;

	if (!have_hv_md_mblock)
		return (frag_assign(NULL, pa));

	for (i = 0; i < NUM_HV_MDS; i++) {
		if (hv_md_frags[i].paddr == pa && hv_md_frags[i].ref_cnt == 0) {
			hv_md_frags[i].ref_cnt = 1;
			DBG(zprintf("Assigning hvmd @ 0x%llx\n",
			    hv_md_frags[i].paddr));
			return (&hv_md_frags[i]);
		}
	}

	return (NULL);
}

void
hv_md_free(bound_mfrag_t *hvfrag)
{
	if (have_hv_md_mblock) {
		ASSERT(hvfrag->ref_cnt == 1);
		hvfrag->ref_cnt = 0;
		DBG(zprintf("Freeing hvmd @ 0x%llx\n", hvfrag->paddr));
	} else {
		frag_free(hvfrag);
	}
}

void
hv_md_mblock_gen_node(md_t *hvmd)
{
	md_node_t *mblock;

	if (have_hv_md_mblock) {
		mblock = md_link_new_node(hvmd, "hvmd_mblock",
		    md_root_node(hvmd), "fwd", "back");

		md_add_value_property(hvmd, mblock, "base", hv_md_mblock_base);
		md_add_value_property(hvmd, mblock, "size", hv_md_mblock_size);
		md_add_value_property(hvmd, mblock, "md_maxsize",
		    max_hv_md_size);
	}
}
