/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2008 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)keys.c	1.1	08/09/19 SMI"

#include <sys/types.h>
#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "basics.h"
#include "chain.h"
#include "keys.h"
#include "mdesc_impl.h"
#include "ldom.h"
#include "clp.h"
#include "allocate.h"
#include "database.h"
#include "zprintf.h"
#include "zpri.h"
#include "response.h"
#include "fatal.h"
#include "ldomconfigure.h"

void
init_keys(ldom_keys_t *keys)
{
	nvlist_init(keys);
}

void
free_keys(ldom_keys_t *keys)
{
	nvlist_clear(keys);
}

/*
 * Generate the key store description for LDom 'ldp' in the machine
 * description 'mdp'.
 */
void
gen_guest_md_keys(ldom_t *ldp, md_t *mdp)
{
	ldom_key_t *key;
	md_node_t *keystore, **nodelistp;
	int i, num;

	if (ldp->keys.len == 0)
		return;			/* no keys defined */

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "keystore"),
	    md_find_name(mdp, "fwd"), nodelistp);

	Xfree(nodelistp);

	assert(num == 0);

	keystore = md_link_new_node(mdp, "keystore",
	    md_root_node(mdp), "fwd", "back");

	key = ldp->keys.pairs;
	for (i = 0; i < ldp->keys.len; i++, key++)
		md_add_string_property(mdp, keystore, key->name, key->value);
}

bool_t
set_keys_from_properties(ldom_t *ldp, md_node_t *node)
{
	md_prop_t *propp;
	char *val, *name;
	bool_t update = false;

	for (propp = CHAIN_START(node->prop_list); propp != NULL;
	    propp = CHAIN_NEXT(propp->link)) {
		if (propp->type == MDET_PROP_STR) {
			name = propp->sp->strp;
			val = (char *)propp->d.dbp->datap;
			if (strcmp(name, "_delete") == 0) {
				if (nvlist_delete_pair(&ldp->keys, val))
					update = true;
			} else {
				if (nvlist_add_pair(&ldp->keys, name, val))
					update = true;
			}
		}
	}

	return (update);
}

/*
 * Restore the key store for LDom 'ldp' from the machine description
 * 'mdp'.
 */
void
restore_guest_keys(ldom_t *ldp, md_t *mdp)
{
	md_node_t **nodelistp;
	int num;

	DBG(zprintf("Restoring keys for guest %s\n", ldp->namep));

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "keystore"),
	    md_find_name(mdp, "fwd"), nodelistp);

	if (num == 1)
		set_keys_from_properties(ldp, nodelistp[0]);

	Xfree(nodelistp);
}

/*
 * Add a key <'name','value'> pair to the key store represented by
 * resource 'handle'.
 */
bool_t
add_ldom_key(ldom_t *ldp, char *name, char *value)
{
	return (nvlist_add_pair(&ldp->keys, name, value));
}

/*
 * Delete a key 'name' from the key store represented by
 * resource 'handle'.
 */
bool_t
del_ldom_key(ldom_t *ldp, char *name)
{
	return (nvlist_delete_pair(&ldp->keys, name));
}

/*
 * Lookup a key 'name' from the key store represented by
 * resource 'handle'.
 */
char *
get_ldom_key(ldom_t *ldp, char *name)
{
	ldom_key_t *key = nvlist_find(&ldp->keys, name);
	return ((key != NULL) ? key->value : NULL);
}
