/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)keystore.c	1.4	09/02/09 SMI"

#include <strings.h>
#include <sys/types.h>

#include "ldom.h"
#include "allocate.h"
#include "zprintf.h"
#include "ds.h"
#include "ds_registry.h"
#include "sequencer.h"
#include "database.h"
#include "keys.h"

/*
 * The Keystore Domain Service - FWARC 2006/523
 *
 * This DS provides the ability to manipulate the keystore. Only two
 * operations are supported, setting a key and deleting a key. Setting a
 * key that is not present in the store creates it.
 */

static ds_ver_t keystore_vers[] = { {1, 0} };
#define	KEYSTORE_NVERS  (sizeof (keystore_vers) / sizeof (keystore_vers[0]))

static ds_capability_t keystore_cap = {
	KEYSTORE_SVC,	/* svc_id */
	keystore_vers,	/* vers */
	KEYSTORE_NVERS	/* nvers */
};

static ds_ops_t keystore_libops = {
	dslib_register_handler,		/* ds_reg_cb */
	dslib_unregister_handler,	/* ds_unreg_cb */
	dslib_data_handler,		/* ds_data_cb */
	KEYSTORE_SVC			/* cb_arg */
};

/*
 * Info needed to invoke ds_clnt_reg/ds_svc_reg when loading plugins
 * There could be multiple services in the same info structure
 */
static libds_info_t keystore_info[] = {
	{
		DS_CLIENT,
		&keystore_cap,
		&keystore_libops
	}
};

typedef enum {
	KEYSTORE_SET_REQ,
	KEYSTORE_DELETE_REQ,
	KEYSTORE_SET_RESP,
	KEYSTORE_DELETE_RESP,
	KEYSTORE_UPDATES_REQ,
	KEYSTORE_UPDATES_RESP,
	KEYSTORE_NULL
} keystore_cmd_t;

typedef struct {
	keystore_cmd_t cmd;
} keystore_hdr_t;

typedef struct {
	char	name_and_value[1];
} keystore_set_req_t;

typedef struct {
	char	name[1];
} keystore_delete_req_t;

typedef enum {
	KEYSTORE_SUCCESS,
	KEYSTORE_NO_SPACE,
	KEYSTORE_INVALID_NAME,
	KEYSTORE_INVALID_VAL,
	KEYSTORE_NOT_PRESENT
} keystore_status_t;

typedef struct {
	keystore_status_t status;
} keystore_resp_t;

typedef struct {
	keystore_hdr_t keys_hdr;
	union {
		keystore_set_req_t	keys_set;
		keystore_delete_req_t	keys_delete;
		keystore_resp_t		keys_resp;
	} u;
} keystore_msg_t;

typedef struct {
	uint32_t	size;
} keystore_updates_resp_t;

#define	keystore_cmd		keys_hdr.cmd
#define	keystore_set		u.keys_set
#define	keystore_delete		u.keys_delete
#define	keystore_status		u.keys_resp.status

#define	KEYSTORE_MAX_NAME_LEN	32	/* excluding NUL */
#define	KEYSTORE_MAX_VAL_LEN	64	/* excluding NUL */

#define	KEYSTORE_MAX_SET_REQ_LEN				\
	(KEYSTORE_MAX_NAME_LEN + KEYSTORE_MAX_VAL_LEN + 2	\
	    + sizeof (keystore_hdr_t))

#define	KEYSTORE_MAX_DEL_REQ_LEN				\
	(KEYSTORE_MAX_NAME_LEN + 1 + sizeof (keystore_hdr_t))

struct nvlist keys;

static bool_t
keystore_is_str_valid(char *str, int maxlen)
{
	int	len;

	if (str == NULL) {
		DBG(zprintf("%s: invalid string: NULL\n", __func__));
		return (false);
	}

	len = strlen(str);

	if (len <= 0) {
		DBG(zprintf("%s: invalid string: too short\n", __func__));
		return (false);
	}

	if (len > maxlen) {
		DBG(zprintf("%s: invalid string: too long\n", __func__));
		return (false);
	}

	return (true);
}

static keystore_status_t
keystore_set_key(ldom_t *ldp, keystore_msg_t *keys_reqp, uint16_t len)
{
	char *name, *value;

	/* sanity check message */
	if (len > KEYSTORE_MAX_SET_REQ_LEN) {
		DBG(zprintf("%s: message too long: %d bytes, %d bytes max\n",
		    __func__, len, KEYSTORE_MAX_SET_REQ_LEN));
		return (KEYSTORE_INVALID_NAME);
	}

	name = (char *)((char *)keys_reqp + sizeof (keystore_hdr_t));

	if (!keystore_is_str_valid(name, KEYSTORE_MAX_NAME_LEN)) {
		DBG(zprintf("%s: invalid 'name' parameter\n", __func__));
		return (KEYSTORE_INVALID_NAME);
	}

	value = name + strlen(name) + 1;

	if (!keystore_is_str_valid(value, KEYSTORE_MAX_VAL_LEN)) {
		DBG(zprintf("%s: invalid 'value' parameter\n", __func__));
		return (KEYSTORE_INVALID_VAL);
	}

	DBG(zprintf("%s: name=%s, value=%s\n", __func__, name, value));

	/* update key store */
	if (add_ldom_key(ldp, name, value)) {

		/* update machine description */
		if (ldp->state != INACTIVE) {
			/* mark config as different from SP */
			zeus.is_next_spconfig = false;
			seq_update_vars(ldp, true);
		}
	}

	return (KEYSTORE_SUCCESS);
}

static keystore_status_t
keystore_del_key(ldom_t *ldp, keystore_msg_t *keys_reqp, uint16_t len)
{
	char *name;

	/* sanity check message */
	if (len > KEYSTORE_MAX_DEL_REQ_LEN) {
		DBG(zprintf("%s: message too long: %d bytes, %d bytes max\n",
		    __func__, len, KEYSTORE_MAX_DEL_REQ_LEN));
		return (KEYSTORE_INVALID_NAME);
	}

	name = (char *)((char *)keys_reqp + sizeof (keystore_hdr_t));

	if (!keystore_is_str_valid(name, KEYSTORE_MAX_NAME_LEN)) {
		DBG(zprintf("%s: invalid 'name' parameter\n", __func__));
		return (KEYSTORE_INVALID_NAME);
	}

	DBG(zprintf("%s: name=%s\n", __func__, name));

	/* check that the key is present */
	if (get_ldom_key(ldp, name) == NULL) {
		DBG(zprintf("%s: key not present\n", __func__));
		return (KEYSTORE_NOT_PRESENT);
	}

	/* update variable store */
	(void) del_ldom_key(ldp, name);

	/*
	 * Update machine description.
	 */
	if (ldp->state != INACTIVE) {
		/* mark config as different from SP */
		zeus.is_next_spconfig = false;
		seq_update_vars(ldp, true);
	}

	return (KEYSTORE_SUCCESS);
}

void
keystore_data_handler(ds_svc_t *svc, void *msg, size_t msglen)
{
	ds_seq_req_t *dsreq;

	dsreq = Xmalloc(sizeof (*dsreq));
	dsreq->svc = svc;
	dsreq->msg = msg;
	dsreq->len = msglen;

	seq_enqueue_ds_req(dsreq);
}

void
keystore_seq_handler(ds_svc_t *svc, void *msg, size_t msglen)
{
	int			cmd;
	int			len;
	ds_msg_t		ds_msg;
	keystore_msg_t		*ks_req;
	keystore_msg_t		*ks_resp;
	keystore_status_t	status;
	ds_resp_t		*ds_resp;
	size_t			resp_len, resp_data_len;

	ks_req = (keystore_msg_t *)((char *)msg + sizeof (ds_data_handle_t));
	cmd = ntoh32(ks_req->keystore_cmd);

	/*
	 * Prime a response.
	 */
	resp_data_len = sizeof (keystore_hdr_t) + sizeof (keystore_resp_t);
	resp_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) +
	    resp_data_len;
	ds_resp = Xmalloc(resp_len);
	DBG7(memset(ds_resp, 0, resp_len));

	ds_fill_data_header(ds_resp, svc->hdl, resp_data_len);

	ks_resp = (keystore_msg_t *)((char *)ds_resp + sizeof (ds_hdr_t) +
	    sizeof (ds_data_handle_t));

	ds_msg.buf = ds_resp;
	ds_msg.len = resp_len;
	ds_msg.seq = 0;

	len = msglen - sizeof (ds_data_handle_t);
	switch (cmd) {
	case KEYSTORE_SET_REQ:

		status = keystore_set_key(svc->chan->ldp, ks_req, len);

		ks_resp->keystore_cmd = hton32(KEYSTORE_SET_RESP);
		ks_resp->keystore_status = hton32(status);
		break;

	case KEYSTORE_DELETE_REQ:

		status = keystore_del_key(svc->chan->ldp, ks_req, len);

		ks_resp->keystore_cmd = hton32(KEYSTORE_DELETE_RESP);
		ks_resp->keystore_status = hton32(status);
		break;

	case KEYSTORE_SET_RESP:
	case KEYSTORE_DELETE_RESP:
	case KEYSTORE_UPDATES_REQ:
	case KEYSTORE_UPDATES_RESP:
	case KEYSTORE_NULL:

		zprintf("Unsupported keystore command: %d\n", cmd);
		ks_resp->keystore_status = hton32(KEYSTORE_INVALID_NAME);
		break;

	default:
		zprintf("Unknown keystore command: %d\n", cmd);
		ks_resp->keystore_status = hton32(KEYSTORE_INVALID_NAME);
		break;
	}

	/* send reply, don't wait for a response */
	(void) ds_send_data_msg(svc->chan, &ds_msg, DS_NO_BLOCK, NULL);
	Xfree(ds_resp);
}

/*
 * The versions supported
 */
static prot_version_t keystore_versions[] =
{
	{1, 0}
};

static svc_ops_t keystore_ops = {
	NULL,				/* reg_handler */
	NULL,				/* unreg_handler */
	keystore_data_handler,		/* data_handler */
	keystore_seq_handler,		/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	int i;
	svc_info_t *info;

	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 2;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(DS_CHAN_NAME);
	info->chan_names[1] = Xstrdup(LIBDS_CHAN_NAME);
	info->num_svcs = 1;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(KEYSTORE_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = 0;
	info->num_vers = NUM_VERS(keystore_versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	memcpy(info->versions, keystore_versions,
	    info->num_vers * sizeof (prot_version_t));

	/*
	 * copy the info structure for each service
	 */
	info->libds_svcs = Xmalloc(info->num_svcs * sizeof (libds_info_t *));
	for (i = 0; i < info->num_svcs; i++)
		info->libds_svcs[i] = cp_dsinfo(&keystore_info[i],
		    info->svcs[i]);

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, KEYSTORE_SVC) == 0)
		return (&keystore_ops);
	else
		return (NULL);
}
