/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)ldcs.c	1.5	09/02/09 SMI"

#include "ldmd_types.h"
#include "ldom.h"
#include "allocate.h"
#include "chain.h"
#include "errors.h"
#include "fatal.h"
#include "ldcs.h"
#include "resource_id.h"
#include "response.h"
#include "support.h"
#include "vio.h"

/*
 * Allocate LDC from a pool.
 */
int
ldcs_chan_alloc(ldcs_t *ldcs, res_pool_t *ldc_pool)
{
	ldc_chan_t *ldc;

	ldc = Xmalloc(sizeof (ldc_chan_t));

	ldc->id = alloc_res(&zeus.free_ldc_ids);
	if (ldc->id == NULL) {
		free(ldc);
		return (ERR_LDC_NO_IDS);
	}

	ldc->chan = alloc_res(ldc_pool);
	if (ldc->chan == NULL) {
		free_res(ldc->id);
		free(ldc);
		return (ERR_LDC_NO_CHANS);
	}

	CHAIN_ADD(ldcs->ldcs, ldc, link);

	return (0);
}

static void
free_ldc(ldc_chan_t *ldc, bool_t staged)
{
	if (staged)
		stage_res(ldc->chan);
	else
		free_res(ldc->chan);

	free_res(ldc->id);
	Xfree(ldc);
}

static void
internal_ldcs_free(ldcs_t *ldcs, bool_t staged)
{
	ldc_chan_t *ldc, *next;

	for (ldc = CHAIN_START(ldcs->ldcs); ldc != NULL; ldc = next) {
		next = CHAIN_NEXT(ldc->link);
		free_ldc(ldc, staged);
	}

	CHAIN_INIT(ldcs->ldcs);
}

/*
 * Free a list of LDC channels.
 */
void
ldcs_free(ldcs_t *ldcs)
{
	internal_ldcs_free(ldcs, false);
}

/*
 * Stage a list of LDC channels for later freeing.
 */
void
ldcs_stage(ldcs_t *ldcs)
{
	internal_ldcs_free(ldcs, true);
}

ldc_chan_t *
ldcs_search(ldcs_t *ldcs, uint64_t ldc_id)
{
	ldc_chan_t *ldc;

	CHAIN_ITER(ldcs->ldcs, ldc, link) {
		if (ldc->chan->id == ldc_id)
			break;
	}

	return (ldc);
}

/*
 * This is called while reconstructing the bindings from the MDs and is
 * only called once for each endpoint. As each endpoint is referenced,
 * the resource ID is cleared out of the database, allowing the database
 * cleanup to release any unclaimed IDs.
 *
 * This is necessary because a domain recovering from an aborted migration
 * does not use the database to get its resource IDs. Thus, the IDs must
 * be free to be claimed directly by the migration recovery code.
 */
static resid_t *
get_res_id(ldom_t *ldp, uint64_t id)
{
	ldc_epdb_t	*ep;
	resid_t		*res_id;

	ep = ldc_epdb_find(ldp->idp->id, id);
	if (ep != NULL) {
		DBG5(zprintf("%s: found: gid=%"PRId64", ldc=%"PRId64", "
		    "id=%"PRId64"\n", __func__, ep->gid, ep->ldc, ep->id->id));

		res_id = ep->id;
		ep->id = NULL;
		return (res_id);
	}

	DBG5(zprintf("%s: NOT found: gid=%"PRId64", id=%"PRId64"\n",
	    __func__, ldp->idp->id, id));

	return (NULL);
}

/*
 * This is only called when constructing state from MDs.  A failure means we
 * have inconsistent/corrupt MDs and there is really nothing to be done other
 * than to abort. Assumes that the LDC endpoint DB has been constructed.
 */
void
ldc_pull_ids(ldom_t *ldp, res_pool_t *ldc_pool, int num_ldc_ids,
    uint64_t *ldc_ids, ldcs_t *ldcs)
{
	ldc_chan_t *ldc;

	for (int i = 0; i < num_ldc_ids; i++) {
		ldc = Xmalloc(sizeof (ldc_chan_t));
		ldc->chan = pull_res_id(ldc_pool, ldc_ids[i]);
		ldc->id = get_res_id(ldp, ldc_ids[i]);

		if ((ldc->chan == NULL) || (ldc->id == NULL))
			fatal("Failed to allocate a LDC channel while parsing "
			    "MDs.\nThe machine state is corrupt. To recover "
			    "reset with a known good configuration.");

		CHAIN_ADD(ldcs->ldcs, ldc, link);
	}
}

static resid_t *
get_and_pull_res_id(ldom_t *ldp, md_t *hvmdp, res_pool_t *pool, uint64_t ldcid)
{
	int		i;
	resid_t		*id = NULL;
	md_node_t	**chans;
	int		nchans;
	md_node_t	*guest;
	uint64_t	chan;
	uint64_t	res_id;

	if (hvmdp == NULL)
		return (alloc_res(pool));

	if ((guest = find_guest_md_node(hvmdp, ldp)) == NULL) {
		DBG(zprintf("%s: guest %s not found\n", __func__, ldp->namep));
		goto done;
	}

	/*
	 * Search the MD for all the LDC endpoints for the guest.
	 */
	chans = Xmalloc(sizeof (*chans) * md_node_count(hvmdp));
	nchans = md_scan_dag(hvmdp, guest, md_find_name(hvmdp, "ldc_endpoint"),
	    md_find_name(hvmdp, "fwd"), chans);

	for (i = 0; i < nchans; i++) {
		if (md_get_prop_val(hvmdp, chans[i], "channel", &chan)) {
			DBG(zprintf("%s: endpoint has no 'channel' "
			    "property\n", __func__));
			goto done;
		}

		if (chan != ldcid)
			continue;

		if (md_get_prop_val(hvmdp, chans[i], "resource_id", &res_id)) {
			DBG(zprintf("%s: endpoint has no 'resource_id' "
			    "property\n", __func__));
			goto done;
		}

		/* found it, pull the resource id */
		id = pull_res_id(pool, res_id);
		break;
	}

done:
	Xfree(chans);
	return (id);
}

/*
 * This is a version of ldc_pull_ids() that doesn't try to get
 * the same global ldc ID for an endpoint. It will preserve the
 * channel ID, but not the global ldc ID.
 *
 * XXXMIG - any chance to share some code with ldc_pull_ids()?
 */
int
ldc_mig_pull_ids(ldom_t *ldp, md_t *hvmdp, res_pool_t *ldc_pool,
    int num_ldc_ids, uint64_t *ldc_ids, ldcs_t *ldcs)
{
	ldc_chan_t *ldc;

	for (int i = 0; i < num_ldc_ids; i++) {
		ldc = Xmalloc(sizeof (ldc_chan_t));
		ldc->chan = pull_res_id(ldc_pool, ldc_ids[i]);
		ldc->id = get_and_pull_res_id(ldp, hvmdp, &zeus.free_ldc_ids,
		    ldc_ids[i]);

		if (ldc->chan == NULL) {
			DBG(zprintf("Unable to allocate the LDC %"PRId64"\n",
			    ldc_ids[i]));
			return (-1);
		}

		if (ldc->id == NULL) {
			DBG(zprintf("No available LDC channels left in the "
			    "system\n"));
			return (-1);
		}

		CHAIN_ADD(ldcs->ldcs, ldc, link);
	}

	return (0);
}

/*
 * Does not assume that an endpoint database has been setup.
 */
int
ldc_mig_get_and_pull_ids(ldom_t *ldp, md_t *hvmdp, res_pool_t *ldc_pool,
    md_t *md, md_node_t *node, ldcs_t *ldcs)
{
	uint64_t	*ids;
	int		num_ids;
	int		rv;

	get_ldc_ids(md, node, &num_ids, &ids);
	rv = ldc_mig_pull_ids(ldp, hvmdp, ldc_pool, num_ids, ids, ldcs);
	Xfree(ids);

	return (rv);
}

/*
 * Determine all the LDC ids for a channel from a MD.
 */
void
get_ldc_ids(md_t *md, md_node_t *node, int *num_ldc_ids, uint64_t **ldc_ids)
{
	md_node_t **ldc_nodes;
	int i;

	ldc_nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(md));

	*num_ldc_ids = md_scan_dag(md, node, md_find_name(md, CHANNEL_ENDPOINT),
	    md_find_name(md, "fwd"), ldc_nodes);

	*ldc_ids = Xmalloc(*num_ldc_ids * sizeof (uint64_t));

	for (i = 0; i < *num_ldc_ids; i++) {
		if (md_get_prop_val(md, ldc_nodes[i], "id",
		    &(*ldc_ids)[i]) != 0)
			fatal("channel-endpoint has no id property");
	}

	Xfree(ldc_nodes);
}

/*
 * Determine all the LDC ids for a channel from a MD and allocate them from
 * the given pool. Assumes that the LDC endpoint DB has been constructed.
 */
void
ldc_get_and_pull_ids(ldom_t *ldp, res_pool_t *ldc_pool, md_t *md,
    md_node_t *node, ldcs_t *ldcs)
{
	uint64_t *ids;
	int num_ids;

	get_ldc_ids(md, node, &num_ids, &ids);
	ldc_pull_ids(ldp, ldc_pool, num_ids, ids, ldcs);
	Xfree(ids);
}

void
ldcs_print(const ldcs_t *ldcs, response_t *resp)
{
	ldc_chan_t *ldc;

	CHAIN_ITER(ldcs->ldcs, ldc, link) {
		if (CHAIN_NEXT(ldc->link) != NULL)
			rprintf(resp, "0x%"PRIx64",", ldc->chan->id);
		else
			rprintf(resp, "0x%"PRIx64"", ldc->chan->id);
	}
}

static md_node_t *
lookup_peer_ldc_endpoint(md_t *hvmdp, md_node_t *guest, uint64_t chan_id)
{
	int		i;
	md_node_t	**chans;
	int		nchans;
	md_node_t	*peer = NULL;
	uint64_t	gid;
	uint64_t	tgt_chan;
	uint64_t	tgt_type;
	uint64_t	tgt_gid;

	/* lookup the gid of the local guest */
	if (md_get_prop_val(hvmdp, guest, "gid", &gid) != 0) {
		DBG(zprintf("%s: guest node does not have a gid property\n",
		    __func__));
		return (NULL);
	}

	/* lookup all the LDC endpoint nodes */
	chans = Xmalloc(sizeof (md_node_t *) * md_node_count(hvmdp));
	nchans = md_scan_dag(hvmdp, NULL, md_find_name(hvmdp, "ldc_endpoint"),
	    md_find_name(hvmdp, "fwd"), chans);

	/*
	 * Walk the endpoint nodes, checking each one to see if it
	 * is the peer of our local endpoint. This can be determined
	 * if the 'target_channel' property matches the specified
	 * LDC ID, and the 'target_guest' matches the local GID.
	 */
	for (i = 0; i < nchans; i++) {

		/* get the LDC ID of the peer endpoint */
		if (md_get_prop_val(hvmdp, chans[i], "target_channel",
		    &tgt_chan) != 0) {
			DBG(zprintf("%s: endpoint does not have a "
			    "target_channel property\n", __func__));
			goto done;
		}

		if (tgt_chan != chan_id)
			continue;

		/* get the type of the peer endpoint */
		if (md_get_prop_val(hvmdp, chans[i], "target_type",
		    &tgt_type) != 0) {
			DBG(zprintf("%s: endpoint does not have a "
			    "target_channel property\n", __func__));
			goto done;
		}

		/* only interested in guest endpoints - ASSERT? */
		if (tgt_type != TYPE_GUEST)
			continue;

		if (md_get_prop_val(hvmdp, chans[i], "target_guest",
		    &tgt_gid) != 0) {
			DBG(zprintf("%s: endpoint does not have a "
			    "target_guest property\n", __func__));
			goto done;
		}

		if (tgt_gid == gid) {
			peer = chans[i];
			break;
		}
	}

done:
	Xfree(chans);
	return (peer);
}

/*
 * Given a list of LDC endpoints, assign channel and resource IDs
 * for all peer endpoints. If no HV MD is specified, these IDs can
 * just be allocated and assigned directly from whatever is available
 * in the pool. Otherwise, more care must be taken to ensure that
 * the same mappings in the HV MD are used.
 */
int
ldc_mig_get_and_pull_peer_ids(ldom_t *ldp, md_t *hvmdp, ldcs_t *src_ldcs,
    ldcs_t *peer_ldcs, res_pool_t *peer_pool)
{
	int		rc = -1;
	md_node_t	*src_guest = NULL;
	md_node_t	*peer;
	ldc_chan_t	*src_ldc;
	ldc_chan_t	*ldc;
	uint64_t	src_id;
	uint64_t	res_id;
	uint64_t	chan_id;

	/*
	 * If no HV MD is specified, the peer mapping is not
	 * critical, so just allocate the next available IDs
	 * in the pool.
	 */
	if (hvmdp == NULL)
		return (ldcs_chan_alloc(peer_ldcs, peer_pool));

	/*
	 * Lookup the source LDom in the HV MD.
	 */
	if ((src_guest = find_guest_md_node(hvmdp, ldp)) == NULL) {
		DBG(zprintf("%s: guest %s not found\n", __func__,
		    ldp->namep));
		goto done;
	}

	/*
	 * Walk the incoming list of LDC endpoints. For each one,
	 * lookup the peer in the HV MD and pull the LDC ID and
	 * resource ID.
	 */
	CHAIN_ITER(src_ldcs->ldcs, src_ldc, link) {
		src_id = src_ldc->chan->id;

		peer = lookup_peer_ldc_endpoint(hvmdp, src_guest, src_id);
		if (peer == NULL) {
			DBG(zprintf("%s: unable to find peer endpoint for "
			    "chan %"PRId64" in LDom %s\n", __func__, src_id,
			    ldp->namep));
			goto done;
		}

		if (md_get_prop_val(hvmdp, peer, "channel", &chan_id) != 0) {
			DBG(zprintf("%s: peer of LDC %"PRId64" has no "
			    "channel property\n", __func__, src_id));
			goto done;
		}

		if (md_get_prop_val(hvmdp, peer, "resource_id", &res_id) != 0) {
			DBG(zprintf("%s: peero fo LDC %"PRId64" has no "
			    "resource_id property\n", __func__, src_id));
			goto done;
		}

		DBG(zprintf("%s: linking %s:%"PRId64" -> id=%"PRId64"\n",
		    __func__, ldp->namep, src_id, chan_id));

		ldc = Xmalloc(sizeof (ldc_chan_t));
		ldc->chan = pull_res_id(peer_pool, chan_id);
		ldc->id = pull_res_id(&zeus.free_ldc_ids, res_id);

		if (ldc->chan == NULL) {
			DBG(zprintf("Unable to allocate the LDC %"PRId64"\n",
			    chan_id));
			goto done;
		}

		if (ldc->id == NULL) {
			DBG(zprintf("Unable to allocate the global LDC "
			    "%"PRId64"\n", res_id));
			goto done;
		}

		CHAIN_ADD(peer_ldcs->ldcs, ldc, link);
	}

	rc = 0;

done:
	return (rc);
}
