/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <libintl.h>
#include <locale.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>		/* geteuid */

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "auth_ldom.h"
#include "basics.h"
#include "clp.h"
#include "cli_conn.h"
#include "control_class.h"
#include "database_internal.h"
#include "errors.h"
#include "ldm_protocol.h"
#include "ldom.h"
#include "mig_common.h"
#include "option.h"
#include "xml_cntrl_common.h"
#include "xml_common.h"
#include "xml_parse.h"
#include "xml_v3_defs.h"
#include "xmpp_common.h"
#include "xmpp_client.h"
#include "ssl_connection.h"

#if !defined(TEXT_DOMAIN)	/* should be defined by cc -D */
#define	TEXT_DOMAIN "SYS_TEST"	/* use this only if it isn't */
#endif

/*
 * This variable is unused in ldm but the definition is needed to permit
 * the linking of zprintf.o
 * FIXME: We could create ldmd.h which defines a #define to allow for optional
 * compilation - zprintf.c could then use this definition and there'd be no
 * need for this variable. That needs ldm and ldmd to be built in different
 * directories
 */
struct ldmd_options option;

/* Needed by DBG() macros */
uint_t debug_level = 0;

/*
 * These variables store the command line arguments so that the signal handler
 * can work out if it's a long running command which needs to be cancelled.
 */
static char *cmd_name;
static char *ldom_name;

enum {
	LOCAL_CONNECTION,
};

typedef struct chan_desc chan_desc_t;

struct chan_desc {
	ssize_t (*write)(chan_desc_t *chan, void *buf, ssize_t len);
	ssize_t (*read)(chan_desc_t *chan, void *buf, ssize_t len);
	int sock;
	int	seqno;
};

static char *prog;			/* program name */

/*
 * getopt parsing
 */
extern char *optarg;
extern int optind;

/*
 * prototypes
 */
static chan_desc_t *create_connection(int sock, int type, int class, char *usr);
static int get_cli_response(chan_desc_t *chan);
static bool_t is_migrate_command(int argc, char **argv);
static bool_t is_memdr_command(char *cmd);
static bool_t is_valid_memdr_cmd(int argc, char **argv);
static int do_migration(int argc, char **argv);
static int do_memdr(int argc, char **argv);

static int get_xml_response(chan_desc_t *chan);

static ssize_t
sockread(chan_desc_t *chan, void *buf, ssize_t len)
{
	return (recv(chan->sock, buf, len, MSG_WAITALL));
}

static ssize_t
sockwrite(chan_desc_t *chan, void *buf, ssize_t len)
{
	return (send(chan->sock, buf, len, 0));
}

static ssize_t
ldm_chan_read(chan_desc_t *chan, void *buf, ssize_t len)
{
	return (chan->read(chan, buf, len));
}

static ssize_t
ldm_chan_write(chan_desc_t *chan, void *buf, ssize_t len)
{
	return (chan->write(chan, buf, len));
}

/*
 * Prepend the CWD to the file for the -i argument if found.  This is done
 * currently for ldm start, bind, create. NOTE!!! If we ever create a command
 * that does not take a file argument after -i we need to revisit this code.
 */
static void
fix_file_opts(int argc, char **argv, char **file_path)
{
	char path[MAXPATHLEN];
	char *newarg;
	int lastopt, size, optlen, c;

	lastopt = optind;
	while ((c = getopt(argc, argv, ":i:(input)")) != EOF) {
		switch (c) {
		case 'i':
			if (optarg[0] == '-')
				break;

			if (optarg[0] == '/') {
				*file_path = optarg;
				break;
			}

			if (getcwd(path, sizeof (path)) == NULL) {
				fprintf(stderr, "%s: %s\n", prog,
				    strerror(errno));
				exit(1);
			}

			if (optind > lastopt + 1) {
				/* form is "-i arg" or "--input arg" */
				size = strlen(path) + strlen(optarg) + 2;
				newarg = Xmalloc(size);
				snprintf(newarg, size, "%s/%s", path, optarg);
				*file_path = newarg;
			} else if (optind > lastopt) {
				/* form is "-iarg" or "--input=arg" */
				optlen = optarg - argv[lastopt];
				size = optlen + strlen(path) +
				    strlen(optarg) + 2;
				newarg = Xmalloc(size);
				snprintf(newarg, optlen + 1,
				    "%s", argv[lastopt]);
				snprintf(newarg + optlen, size - optlen,
				    "%s/%s", path, optarg);
				*file_path = newarg;
			}
			break;
		case ':':
			/* missing argument to -i or -o, bail out */
			return;
		case '?':
		default:
			/* keep processing through other options */
			break;
		}

		lastopt = optind;
	}
}

static int
get_textblocks(zeus_cmd_arg_t *cmd_arg, char **argv, int argc)
{
	int i;
	int len = 0;

	for (i = 0; i < argc; i++) {
		cmd_arg[i].arg_type = TEXTBLOCK;
		cmd_arg[i].arg_len = strlen(argv[i]) + 1;
		cmd_arg[i].arg = Xstrdup(argv[i]);
		len += cmd_arg[i].arg_len + 2 * (int)sizeof (uint32_t);
	}

	return (len);
}

static void
serialized_cmd_prep(int cmd_argc, char **cmd_argv, zeus_cmd_arg_t **cmd_arg,
    int *cmd_len, int *msg_len)
{
	int args_len = 0;

	if (cmd_argc > 0) {
		*cmd_arg = Xmalloc(cmd_argc * (int)sizeof (zeus_cmd_arg_t));
		args_len = get_textblocks(*cmd_arg, cmd_argv, cmd_argc);
	} else {
		*cmd_arg = NULL;
	}

	*cmd_len = (int)sizeof (zeus_cmd_msg_t) + args_len;
	*msg_len = *cmd_len + (int)(sizeof (ldm_hdr_t) + sizeof (ldm_op_req_t));
}

static void
serialize_cmd(char *buf, int argc, zeus_cmd_arg_t *args, int id, int cmd_len)
{
	char *p;
	zeus_cmd_msg_t *msg;
	ldm_hdr_t *hdr;
	ldm_op_req_t *req;
	int i;

	/* set LDM protocol headers */
	hdr = (ldm_hdr_t *)buf;
	hdr->msg_type = hton32(LDM_OP_REQ);
	hdr->payload_len = hton32(sizeof (ldm_op_req_t) + cmd_len);
	req = (ldm_op_req_t *)(buf + sizeof (ldm_hdr_t));
	req->rq_len = hton32(cmd_len);
	req->rq_id = id;

	/*
	 * Append the byte-order converted command header to the buffer
	 * leaving any required prefix space before it.
	 */
	msg = (zeus_cmd_msg_t *)req->rq_data;
	msg->len = hton32(cmd_len);
	msg->argc = hton32(argc);

	/* if there were no args specified to the ldm command, return now */
	if (args == NULL)
		return;

	/* append byte-ordering converted arguments to buffer */
	p = (char *)msg + sizeof (zeus_cmd_msg_t);
	for (i = 0; i < argc; i++) {
		if (args[i].arg_type == OPTION) {
			uint32_t typearg[2];

			typearg[0] = hton32(args[i].arg_type);
			typearg[1] = hton32(*(uint32_t *)args[i].arg);
			memcpy(p, &typearg, 2 * sizeof (uint32_t));
			p += 2 * sizeof (uint32_t);
		} else if (args[i].arg_type == TEXTBLOCK) {
			uint32_t typelen[2];

			typelen[0] = hton32(args[i].arg_type);
			typelen[1] = hton32(args[i].arg_len);
			memcpy(p, &typelen, 2 * sizeof (uint32_t));
			p += 2 * sizeof (uint32_t);
			memcpy(p, args[i].arg, args[i].arg_len);
			p += args[i].arg_len;
		}
	}
}

static void
send_cli_cmd(int cmd_argc, char **cmd_argv, chan_desc_t *chan)
{
	zeus_cmd_arg_t	*cmd_arg = NULL;
	char		*msg;
	int		msg_len, cmd_len, i;

	serialized_cmd_prep(cmd_argc, cmd_argv, &cmd_arg, &cmd_len, &msg_len);
	msg = Xcalloc(msg_len, char);
	serialize_cmd(msg, cmd_argc, cmd_arg, chan->seqno, cmd_len);

	for (i = 0; i < cmd_argc; i++) {
		DBG2(zprintf("cmd[%d] = %s\n", i, cmd_argv[i]));
	}
	if (ldm_chan_write(chan, msg, msg_len) == -1) {
		fprintf(stderr, "%s: %s\n", gettext("Unable to send message"),
		    strerror(errno));
		exit(1);
	}

	free(msg);
}

static int
send_xml_cmd(char *file_path, chan_desc_t *chan, char **cmd_argv)
{
	char		*xml_buf;
	xmlChar		*action;
	int		xml_buf_len;
	char		*pkt_buf;
	int		pkt_buf_len;
	int		status;
	xmlDocPtr	doc;
	xmlNodePtr	root, cmd, data;

	if ((strncmp(cmd_argv[0], "add-dom", 7) == 0) ||
	    (strcmp(cmd_argv[0], "create") == 0)) {
		action = XML_ADD_DOMAIN;
	} else if ((strncmp(cmd_argv[0], "set-dom", 7) == 0) ||
	    (strcmp(cmd_argv[0], "modify") == 0)) {
		action = XML_SET_DOMAIN;
	} else if (strncmp(cmd_argv[0], "bind", 4) == 0) {
		action = XML_BIND_DOMAIN;
	} else if (strncmp(cmd_argv[0], "start", 5) == 0) {
		action = XML_START_DOMAIN;
	} else {
		printf("-i not applicable with given command %s\n",
		    cmd_argv[0]);
		return (-1);
	}

	doc = xmlReadFile(file_path, NULL, 0);
	if (doc == NULL) {
		fprintf(stderr, gettext("Invalid XML input document %s\n"),
		    file_path);
		exit(1);
	}

	root = xmlDocGetRootElement(doc);
	cmd = xmlNewChild(root, NULL, XML_CMD, NULL);
	xmlNewChild(cmd, NULL, XML_ACTION, action);

	data = xml_find_subnode(root, XML_DATA);
	while (data != NULL) {
		xmlUnlinkNode(data);
		xmlAddChild(cmd, data);

		data = xml_find_subnode(root, XML_DATA);
	}

	(void) xmlDocDumpMemory(doc, (xmlChar **)&xml_buf,
	    &xml_buf_len);

	pkt_buf_len = create_pkt_buf(&pkt_buf, xml_buf, xml_buf_len,
	    LDM_CNTRL_XML_REQ);

	status = ldm_chan_write(chan, pkt_buf, pkt_buf_len);
	if (status == -1) {
		fprintf(stderr, "%s: %s\n", gettext("Unable to send message"),
		    strerror(errno));
		exit(1);
	}

	return (0);
}

static void
xmlerr(void *ctx, const char *fmt, ...)
{
	char buf[256];
	va_list ap;

	DBG_BLOCK {
		va_start(ap, fmt);
		vsnprintf(buf, sizeof (buf), fmt, ap);
		va_end(ap);
		fprintf(stderr, "%s: %s\n", (char *)ctx, buf);
	}
}

/* ARGSUSED1 */
static void
ldm_sig_handler(int sig, siginfo_t *info, void *p)
{
	chan_desc_t	*chan;
	int		rv, i;
	int		num_cancel_cmd_args = 3;
	char		**argv = Xcalloc(num_cancel_cmd_args, char *);
	int		sock;

	DBG2(zprintf("Received sig=%d\n", sig));

	/*
	 * If information about the pending operation has not been
	 * saved, either it is not a long running command, or the
	 * options to the long running command have not been parsed
	 * yet. In either case, there is nothing to cancel, so the
	 * command can exit.
	 */
	if ((cmd_name == NULL) || (ldom_name == NULL))
		exit(0);

	DBG(zprintf("%s: cmd '%s' currently running\n", __func__, cmd_name));

	/*
	 * Check to see what type of command was run. If it is one of
	 * the long running commands which can be cancelled, send a
	 * 'cancel-op' request to ldmd. Otherwise, there is nothing
	 * more to do.
	 */
	if ((strcmp(cmd_name, "migrate") == 0) ||
	    (strcmp(cmd_name, "migrate-domain") == 0)) {
		/*
		 * We set the operation to be cancelled to 'migration'
		 * and get the LDom name from the original request.
		 */
		argv[1] = strdup("migration");	/* operation type */
		argv[2] = strdup(ldom_name);	/* LDom name */

		DBG(zprintf("%s: mig LDom name %s\n", __func__, argv[2]));
	} else if (is_memdr_command(cmd_name)) {
		/*
		 * We set the operation to be cancelled to 'memdr'
		 * and get the LDom name from the original request.
		 */
		argv[1] = strdup("memdr");	/* operation type */
		argv[2] = strdup(ldom_name);	/* LDom name */
		DBG(zprintf("%s: mem dr LDom name %s\n", __func__, argv[2]));
	} else {
		exit(0);
	}

	sock = connect_local(DEFAULT_PORT);
	if (sock < 0)
		exit(ENOTCONN);

	/*
	 * Connect to the LDom manager, send the CLI command.
	 */
	chan = create_connection(sock, LOCAL_CONNECTION, LDM_CLASS_CLI, NULL);
	if (chan == NULL) {
		/* create_connection generates error msg */
		exit(1);
	}

	/* Fill in the first argument to denote a cancel operation */
	argv[0] = strdup("cancel-op");

	DBG(zprintf("sig handler sending CLI cmd to cancel operation\n"));
	send_cli_cmd(num_cancel_cmd_args, argv, chan);

	rv = get_cli_response(chan);

	for (i = 0; i < num_cancel_cmd_args; i++) {
		free(argv[i]);
	}

	exit(rv == 0 ? 0 : 1);
}

int
main(int argc, char **argv)
{
	struct sigaction sigact;
	char		*file_path = NULL;
	chan_desc_t	*chan;
	int		ret;

	/* setup privileges */
	ldm_priv_init();

	/* internationalization */
	(void) setlocale(LC_MESSAGES, "");
	(void) textdomain(TEXT_DOMAIN);

	prog = strrchr(argv[0], '/');
	prog = (prog == NULL) ? argv[0] : prog++;

	argc--;
	argv++;

	/* Initialize signal handler */
	sigact.sa_handler = ldm_sig_handler;
	sigact.sa_flags = SA_RESTART;
	(void) sigemptyset(&sigact.sa_mask);

	/* Register signal handler */
	if (sigaction(SIGINT, &sigact, (struct sigaction *)NULL) < 0) {
		perror("Warning: sigaction SIGINT");
		return (1);
	}
	if (sigaction(SIGTERM, &sigact, (struct sigaction *)NULL) < 0) {
		perror("Warning: sigaction SIGTERM");
		return (1);
	}

	/* the migrate-domain command requires special handling */
	if (is_migrate_command(argc, argv))
		return (do_migration(argc, argv));

	if (is_memdr_command(argv[0]) == true) {
		return (do_memdr(argc, argv));
	}

	int sock = connect_local(DEFAULT_PORT);
	if (sock < 0)
		return (1);

	fix_file_opts(argc, argv, &file_path);
	if (file_path == NULL) {
		/*
		 * Connect to the LDom manager, send the CLI command.
		 */
		chan = create_connection(sock, LOCAL_CONNECTION, LDM_CLASS_CLI,
		    NULL);
		if (chan == NULL) {
			/* create_connection generates error msg */
			exit(1);
		}
		send_cli_cmd(argc, argv, chan);

		if (get_cli_response(chan) == 0)
			return (0);
		else
			return (1);
	} else {
		/*
		 * Connect to the LDom manager, send the XML command.
		 */
		chan = create_connection(sock, LOCAL_CONNECTION,
		    LDM_CLASS_CONTROL, NULL);
		if (chan == NULL) {
			/* create_connection generates error msg */
			exit(1);
		}

		LIBXML_TEST_VERSION;		/* calls xmlInitParser() */

		/* avoid excessive libxml parse error spew */
		xmlSetGenericErrorFunc("xml", xmlerr);

		ret = send_xml_cmd(file_path, chan, argv);
		if (ret == -1)
			return (1);

		if (get_xml_response(chan) == 0)
			return (0);
		else
			return (1);
	}
}

static chan_desc_t *
open_local_connection(int sock)
{
	chan_desc_t *chan;

	chan = Xmalloc(sizeof (chan_desc_t));

	chan->read = sockread;
	chan->write = sockwrite;
	chan->sock = sock;
	chan->seqno = 1;

	return (chan);
}

/*ARGSUSED3*/
chan_desc_t *
create_connection(int sock, int conn_type, int class, char *user)
{
	chan_desc_t *chan;
	int error, msg_len;
	ldm_hdr_t *hdr, hdr_buf;
	ldm_init_req_t *init_req;
	char *buf;

	/* Only type of connection supported, programming error otherwise */
	ASSERT(conn_type == LOCAL_CONNECTION);

	chan = open_local_connection(sock);
	if (chan == NULL)
		return (NULL);

	msg_len = sizeof (ldm_hdr_t) + sizeof (ldm_init_req_t);
	buf = Xcalloc(msg_len, char);

	hdr = (ldm_hdr_t *)buf;
	hdr->msg_type = hton32(LDM_INIT_REQ);
	hdr->payload_len = hton32(sizeof (ldm_init_req_t));
	init_req = (ldm_init_req_t *)(buf + sizeof (ldm_hdr_t));
	init_req->client_class = hton32(class);
	if (class == LDM_CLASS_CONTROL) {
		init_req->major_vers = hton16(LDM_CONTROL_MAJOR_VER);
		init_req->minor_vers = hton16(LDM_CONTROL_MINOR_VER);
	} else {
		ASSERT(class == LDM_CLASS_CLI);
		init_req->major_vers = hton16(LDM_CLI_MAJOR_VER);
		init_req->minor_vers = hton16(LDM_CLI_MINOR_VER);
	}

	DBG(zprintf("%s: Negotiating class:%d ver:(%d.%d)\n",
	    __func__, class, init_req->major_vers, init_req->minor_vers));

	error = ldm_chan_write(chan, buf, msg_len);
	if (error == -1) {
		fprintf(stderr, "%s: %s\n",
		    gettext("Failed to send version negotiation to logical "
		    "domain manager"), strerror(errno));
		return (NULL);
	}

	free(buf);

	error = ldm_chan_read(chan, &hdr_buf, sizeof (ldm_hdr_t));
	if (error == -1) {
		fprintf(stderr, "%s: %s\n",
		    gettext("Failed to receive version negotiation response "
		    "from logical domain manager"), strerror(errno));
		return (NULL);
	}

	hdr_buf.msg_type = ntoh32(hdr_buf.msg_type);
	hdr_buf.payload_len = ntoh32(hdr_buf.payload_len);

	switch (hdr_buf.msg_type) {
	case LDM_INIT_ACK: {
		ldm_init_ack_t ack;
		error = ldm_chan_read(chan, &ack, sizeof (ack));
		if (error == -1) {
			fprintf(stderr, "%s: %s\n",
			    gettext("Failed to receive LDM_INIT_ACK msg"),
			    strerror(errno));
			return (NULL);
		}
		break;
	}
	case LDM_INIT_NACK: {
		ldm_init_nack_t nack;
		error = ldm_chan_read(chan, &nack, sizeof (nack));
		if (error == -1) {
			fprintf(stderr, "%s: %s\n",
			    gettext("Failed to receive LDM_INIT_NACK msg"),
			    strerror(errno));
			return (NULL);
		}
		fprintf(stderr, "%s\n", gettext("Version mismatch"));
		exit(1);
		/*NOTREACHED*/
	}
	case LDM_ERROR: {
		ldm_error_t err;
		error = ldm_chan_read(chan, &err, sizeof (err));
		if (error == -1) {
			fprintf(stderr, "%s: %s\n",
			    gettext("Failed to receive LDM_ERROR msg"),
			    strerror(errno));
			return (NULL);
		}
		fprintf(stderr, "%s\n", gettext("Protocol error"));
		exit(1);
	}
	default:
		fprintf(stderr, "%s\n", gettext("Unknown response msg type"));
		exit(1);
	}

	return (chan);
}

static void
get_response(chan_desc_t *chan, int *type, char **resp, int *resp_len)
{
	ldm_hdr_t hdr;
	char *buf;
	int status;

	status = ldm_chan_read(chan, &hdr, sizeof (ldm_hdr_t));
	if (status < 0) {
		/* error or LDom manager exited */
		fprintf(stderr, "%s: %s\n",
		    gettext("Receive failed"), status == 0 ?
		    gettext("logical domain manager not responding") :
		    strerror(errno));
		exit(1);
	}

	hdr.msg_type = ntoh32(hdr.msg_type);
	hdr.payload_len = ntoh32(hdr.payload_len);

	if (hdr.msg_type == LDM_OP_REPLY) {
		if (hdr.payload_len < sizeof (ldm_op_reply_t)) {
			fprintf(stderr, gettext("Invalid response\n"));
			exit(1);
		}
	} else {
		fprintf(stderr, gettext("Invalid response\n"));
		exit(1);
	}

	buf = Xmalloc(hdr.payload_len);

	/* receive the actual response body from the LDom Manager */
	status = ldm_chan_read(chan, buf, hdr.payload_len);
	if (status == -1) {
		fprintf(stderr, "%s: %s\n",
		    gettext("Receive message failed"), strerror(errno));
		exit(1);
	}

	if (hdr.msg_type == LDM_OP_REPLY) {
		ldm_op_reply_t *reply = (ldm_op_reply_t *)buf;
		if (reply->rp_reqid != chan->seqno) {
			fprintf(stderr, gettext("Invalid response id\n"));
			exit(1);
		}
	}

	chan->seqno++;

	*type = hdr.msg_type;
	*resp = buf;
	*resp_len = hdr.payload_len;
}

static int
handle_cli_response(char *buf)
{
	cli_reply_data_t *info_resp;
	int cc, text_len;

	info_resp = (cli_reply_data_t *)(buf + sizeof (ldm_op_reply_t));

	text_len = ntoh32(info_resp->text_length);
	if (text_len > 0)
		printf("%s", info_resp->text);

	cc = ntoh32(info_resp->condition_code);
	return (cc);
}

static int
get_cli_response(chan_desc_t *chan)
{
	char *buf = NULL;
	int buf_len, cc, type = LDM_ERROR;

	get_response(chan, &type, &buf, &buf_len);
	if (type != LDM_OP_REPLY) {
		fprintf(stderr, gettext("Invalid response\n"));
		exit(1);
	}

	cc = handle_cli_response(buf);
	free(buf);

	return (cc);
}

/*
 * If a request failed, there should be error messages. If a request
 * succeeded, there may or may not be a warning message. We assume that
 * if there are warning messages, those warning messages can be at document,
 * CMD, or/and DATA sections of the response. For a successful request, We parse
 * all sections of the response to print out warning messages. However,
 * for a failed request, we parse the response to the first error
 * message is encountered.
 */
static int
handle_xml_resp(char *resp_buf, int buf_len)
{
	xmlDocPtr	xml_output;
	xmlNodePtr	root, data, resp, status, resp_msg, cmd;
	char		*status_str, *resp_msg_str;
	bool_t		xml_parse_error = true;

	xml_output = xmlParseMemory((const char *)resp_buf, buf_len);
	if (xml_output == NULL) {
		fprintf(stderr, gettext("Invalid XML response document\n"));
		exit(1);
	}
	DBG_BLOCK {
		if (xmlSaveFormatFile("output.xml", xml_output, true) == -1) {
			printf("Unable to save XML output file");
		}
	}

	/*
	 * Check to see if the XML document has a message.  If it does and it
	 * is at the document level then print it out regardless of the status.
	 */
	root = xmlDocGetRootElement(xml_output);
	if (root == NULL)
		goto done;
	resp = xml_find_subnode(root, XML_RESPONSE);
	if (resp == NULL)
		goto done;

	status = xml_find_subnode(resp, XML_STATUS);
	if (status == NULL)
		goto done;

	status_str = (char *)xmlNodeGetContent(status);
	if (status_str == NULL)
		goto done;
	resp_msg = xml_find_subnode(resp, XML_RESP_MSG);
	if (resp_msg != NULL) {
		resp_msg_str = (char *)xmlNodeGetContent(resp_msg);
		printf("%s\n", resp_msg_str);
		if (strcmp(status_str, XML_FAILURE) == 0) {
			/* nothing more can be done */
			DBG(zprintf("XML document failure\n"));
			return (1);
		}
		/* continue to see if there is any other warning msg */
	}

	cmd = xml_find_subnode(root, XML_CMD);
	if (cmd == NULL)
		goto done;
	resp = xml_find_subnode(cmd, XML_RESPONSE);
	if (resp == NULL)
		goto done;
	status = xml_find_subnode(resp, XML_STATUS);
	if (status == NULL)
		goto done;
	status_str = (char *)xmlNodeGetContent(status);
	if (status_str == NULL)
		goto done;

	/* If we have a resp_msg in the CMD response then we print it out */
	resp_msg = xml_find_subnode(resp, XML_RESP_MSG);
	if (resp_msg != NULL) {
		resp_msg_str = (char *)xmlNodeGetContent(resp_msg);
		printf("%s\n", resp_msg_str);
		if (strcmp(status_str, XML_FAILURE) == 0) {
			/*
			 * Don't need to check anything else as we didn't
			 * get there. Nothing more can be done.
			 */
			DBG(zprintf("XML_CMD failure\n"));
			return (1);
		}
		/* continue to see if there is any other warning msg */
	}

	data = xml_find_subnode(cmd, XML_DATA);
	while ((data != NULL) &&
	    (strcmp((char *)data->name, (char *)XML_DATA) == 0)) {
		DBG(zprintf("XML_DATA message\n"));

		resp = xml_find_subnode(data, XML_RESPONSE);
		if (resp == NULL)
			goto next;
		DBG(zprintf("XML_DATA response\n"));

		status = xml_find_subnode(resp, XML_STATUS);
		if (status == NULL)
			goto next;
		DBG(zprintf("XML_DATA status\n"));

		status_str = (char *)xmlNodeGetContent(status);
		if (status_str == NULL)
			goto next;
		DBG(zprintf("XML_DATA status str\n"));

		resp_msg = xml_find_subnode(resp, XML_RESP_MSG);
		if (resp_msg != NULL) {
			resp_msg_str = (char *)xmlNodeGetContent(resp_msg);
			if (resp_msg_str != NULL) {
				DBG(zprintf("XML_DATA response str\n"));
				printf("%s\n", resp_msg_str);
			} else {
				DBG(zprintf("NULL XML response msg\n"));
			}

			if (strcmp(status_str, XML_FAILURE) == 0)
				return (1);
		}

next:
		data = xml_get_next_ele_node(data);
	}

	/*
	 * We parsed the XML document correctly.
	 */
	xml_parse_error = false;

done:
	/* We encountered an error parsing the XML */
	if (xml_parse_error) {
		fprintf(stderr, gettext("Invalid XML response document\n"));
		return (1);
	}

	if (strcmp(status_str, XML_FAILURE) == 0)
		return (1);

	return (0);
}

static int
get_xml_response(chan_desc_t *chan)
{
	ldm_op_req_t	*req;
	ldm_hdr_t	hdr;
	char		*resp_buf;
	int		status;

	status = ldm_chan_read(chan, &hdr, sizeof (ldm_hdr_t));
	if (status < 0) {
		/* error or LDom manager exited */
		fprintf(stderr, "%s: %s\n",
		    gettext("Receive failed"), status == 0 ?
		    gettext("logical domain manager not responding") :
		    strerror(errno));
		exit(1);
	}

	hdr.msg_type = ntoh32(hdr.msg_type);

	if (hdr.msg_type != LDM_CNTRL_XML_RESP) {
		fprintf(stderr, "%s\n",
		    gettext("Unexpected response message type"));
		exit(1);
	}

	hdr.payload_len = ntoh32(hdr.payload_len);

	if (hdr.payload_len == 0)
		return (0);

	resp_buf = Xmalloc(hdr.payload_len);

	/* receive the actual response body from Zeus */
	status = ldm_chan_read(chan, resp_buf, hdr.payload_len);
	if (status == -1) {
		fprintf(stderr, "%s: %s\n", gettext("Receive message failed"),
		    strerror(errno));
		exit(1);
	}

	/* pull out the buffer with the XML and send it for parsing */
	req = (ldm_op_req_t *)resp_buf;
	status = handle_xml_resp((char *)req->rq_data, req->rq_len);

	Xfree(resp_buf);

	return (status);
}

/*
 * LDom migration support.
 */

/*
 * Print the usage message for the Domain Migration command. The string
 * passed in contains command used by the user (either "migrate-domain"
 * or the alias "migrate")
 */
static void
migrate_usage(char *cmdname)
{
	ASSERT(cmdname != NULL);
	printf("\n%s:\n", gettext("Usage"));
	printf("\tldm %s [-f] [-n] [-p <password_file>] <source_ldom>\n"
	    "\t\t[<user>@]<target_host>[:<target_ldom>]\n\n", cmdname);
}

/*
 * Check to see if this is a 'migrate-domain' request (or its
 * alias 'migrate')
 */
static bool_t
is_migrate_command(int argc, char **argv)
{
	if ((argc < 1) ||
	    ((strcmp(argv[0], "migrate") != 0) &&
	    (strcmp(argv[0], "migrate-domain")) != 0))
		return (false);
	else
		return (true);
}

/*
 * Check to see if this is a memory dr cli.
 */
static bool_t
is_memdr_command(char *cmd)
{
	if (cmd == NULL)
		return (false);

	if ((strncmp(cmd, "remove-mem", 10) != 0) &&
	    (strncmp(cmd, "rm-mem", 6) != 0) &&
	    (strncmp(cmd, "rm-memory", 9) != 0) &&
	    (strncmp(cmd, "remove-memory", 13) != 0) &&
	    (strncmp(cmd, "set-mem", 7) != 0) &&
	    (strncmp(cmd, "set-memory", 10) != 0) &&
	    (strncmp(cmd, "add-mem", 7) != 0) &&
	    (strncmp(cmd, "add-memory", 10) != 0))
		return (false);
	else
		return (true);
}

/*
 * Print the usage message for the memory DR command. The string
 * passed in contains command used by the user (remove-mem, set-mem",
 * "add-mem" or the alias of those cli)
 */
static void
memdr_usage(char *cmdname)
{
	ASSERT(cmdname != NULL);

	printf("\n%s:\n", gettext("Usage"));
	if (strncmp(cmdname, "remove-mem", 10) == 0 ||
	    strncmp(cmdname, "rm-mem", 6) == 0 ||
	    strncmp(cmdname, "rm-memory", 9) == 0 ||
	    strncmp(cmdname, "remove-memory", 13) == 0)
		printf("\tldm %s <number>[GMK] <ldom>\n", cmdname);

	else if (strncmp(cmdname, "set-mem", 7) == 0 ||
	    strncmp(cmdname, "set-memory", 10) == 0)
		printf("\tldm %s [--auto-adj] <number>[GMK] <ldom>\n", cmdname);

	else
		printf("\tldm %s (--unaligned primary | <number>[GMK] <ldom>)"
		    "\n", cmdname);
}


/*
 * Check syntax of memory DR. Valid syntaxes are:
 * ldm add-mem/set-mem/rm-mem <size> <ldom>
 * ldm add_mem --unassigned primary
 * ldm set-mem --auto-adj <size> <ldom>
 * This function is called after we know the cli is one of
 * "add-mem/set-mem/rm-mem" or their alias.
 * A memory DR request is sent via xmpp interface. Unlike ldomcli interface,
 * the xmpp interface does not return the "usage"  for the return value
 * of "CLI_RESULT_USAGE". Because of this limitation, we parse the request
 * to identify any case of "CLI_RESULT_USAGE and return false. The caller
 * will print out usage message.
 */
static bool_t
is_valid_memdr_cmd(int argc, char **argv)
{
	if (argc < 3)
		return (false);

	/* help */
	if (strcmp(argv[1], "-?") == 0 || strcmp(argv[1], "--help") == 0)
		return (false);

	/* ldm add-mem --unaligned primary */
	if ((strncmp(argv[0], "add-mem", 7) == 0 ||
	    strncmp(argv[0], "add-memory", 10) == 0) &&
	    (strncmp(argv[1], "--unaligned", 11) == 0)) {
		if (strncmp(argv[2], SOD_CONTROL_DOMAIN,
		    strlen(SOD_CONTROL_DOMAIN)) == 0)
			return (true);
		return (false);
	}

	/* ldm set-mem --auto-adj */
	if ((strncmp(argv[0], "set-mem", 7) == 0) ||
	    (strncmp(argv[0], "set-memory", 10) == 0)) {
		if (strncmp(argv[1], "--auto-adj", 10) == 0) {
			if (argc != 4)
				return (false);
			else
				return (true);
		}
	}

	/* ldm add-mem/set-mem/rm-mem <size> <ldom> */

	if (argc != 3)
		return (false);

	return (true);
}

xmlNodePtr
create_memdr_XML(int argc, char **argv)
{
	xmlNodePtr	root, cmd, data, envelope;
	xmlNodePtr	ldom_content, mem;
	xmlChar		*act;
	bool_t		auto_adj_op = false;



	if ((strncmp(argv[0], "remove-mem", 10) == 0) ||
	    (strncmp(argv[0], "rm-mem", 6) == 0)) {
		act = XML_REMOVE_MEM;

	} else if ((strncmp(argv[0], "set-mem", 7) == 0) ||
	    (strncmp(argv[0], "set-memory", 10) == 0)) {
		act = XML_SET_MEM;
		if (argc == 4)
			auto_adj_op = true;

	} else {
		/* add-mem */
		act = XML_ADD_MEM;
	}

	root = xmlNewNode(NULL, XML_LDM_INTERFACE);
	xmlNewProp(root, VERSION_ATTR, LDOM_INTERFACE_VERSION);
	xmlNewProp(root, XSI, XML_SCHEMA);
	xmlNewProp(root, NONAMESPACE, LDMD_SCHEMA);
	xmlNewProp(root, OVF, ENVELOPE_SCHEMA);
	xmlNewProp(root, RASD, RASD_SCHEMA);
	xmlNewProp(root, VSSD, VSSD_SCHEMA);
	xmlNewProp(root, GPROP, GPROP_SCHEMA);
	xmlNewProp(root, BIND, BIND_SCHEMA);
	cmd = xmlNewChild(root, NULL, XML_CMD, NULL);
	xmlNewChild(cmd, NULL, XML_ACTION, act);
	data = xmlNewChild(cmd, NULL, XML_DATA, NULL);
	xmlNewProp(data, VERSION_ATTR, LDOM_V3_DATA_VERSION);

	envelope = xmlNewChild(data, NULL, ENVELOPE_NODE, NULL);
	xmlNewChild(envelope, NULL, REFERENCES_NODE, NULL);

	ldom_content = xmlNewChild(envelope, NULL, CONTENT_NODE, NULL);
	xmlNewProp(ldom_content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);

	/*
	 * argv[0] has either set-mem/add-mem/rm-mem
	 * argv[1] has "--auto-adj" if auto_adj_op is true. Otherwise, it has
	 * the size of the memory.
	 * argv[2] has the size of the memory if auto_adj_op is true.
	 * Otherwise, it has the ldom's name.
	 * argv[3] has the ldom's name if auto_adj_op is true. Otherwise, it
	 * should not exist.
	 */
	if (auto_adj_op == false)
		xmlNewProp(ldom_content, ID_ATTR, (xmlChar *)argv[2]);
	else
		xmlNewProp(ldom_content, ID_ATTR, (xmlChar *)argv[3]);

	mem = create_v3_section(ldom_content, VIRTUAL_HARDWARE_TYPE,
	    MEMORY_NODE);
	if (auto_adj_op == false) {
		xmlNewChild(mem, NULL, ALLOCATION_NODE_OUT, (xmlChar *)argv[1]);
	} else {
		xmlNewChild(mem, NULL, ALLOCATION_NODE_OUT, (xmlChar *)argv[2]);
		xmlNewChild(mem, NULL, XML_AUTO_ADJ, (xmlChar *)argv[1]);
	}

	return (root);
}

static int
do_memdr(int argc, char **argv)
{
	int		xmpp_sock;
	xmlNodePtr	root;
	char		*result;

	if (is_valid_memdr_cmd(argc, argv) == false) {
		memdr_usage(argv[0]);
		return (1);
	}

	cmd_name = strdup(argv[0]);
	ldom_name = strdup(argv[argc-1]);

	xmpp_sock = connect_local(XMPP_DEFAULT_PORT);

	if (xmpp_sock < 0) {
		fprintf(stderr, "%s\n", gettext("XMPP connection failed"));
		return (1);
	}

	root = create_memdr_XML(argc, argv);
	if (root == NULL)
		return (1);

	result = ldm_send_async_command(xmpp_sock, root);
	if (result == NULL) {
		printf("Failed to send memory dr command\n");
		exit(1);
	}

	return (handle_xml_resp(result, strlen(result)));
}

/*
 * Verify that the supplied hostname <host> does not match the hostname of
 * the local system.  We want to block an attempt to migrate a domain to the
 * same host (i.e. source == target)
 */
static bool_t
verify_target_hostname(char *host)
{
	char src_hostname[MAXHOSTNAMELEN];
	char *local_hostname = src_hostname;

	ASSERT(host != NULL);

	if (gethostname(local_hostname, MAXHOSTNAMELEN) != 0) {
		printf(MSG_ERR_MIG_FAILED_SRC_HOSTNAME);
		return (false);
	}

	if (strcmp(host, local_hostname) == 0) {
		printf(MSG_ERR_MIG_IDENTICAL_TARGET_SOURCE, host);
		return (false);
	}

	/*
	 * Note: if support is added in the future to allow the IP address
	 * of the target host to be specified instead of the hostname, we'd
	 * need to check that here too
	 */

	return (true);
}

/*
 * Obtain the password for the target system interactively from the
 * user. A prompt is displayed on the terminal and the command waits
 * until a password is entered. Memory is allocated for the buffer
 * used to hold the password string. The caller must deallocate this
 * memory once the password has been processed.
 */
static bool_t
get_target_password_from_user(char **passwd)
{
	char *pass;

	pass = getpassphrase("Target Password: ");

	if (pass == NULL)
		return (false);

	*passwd = strdup(pass);
	return (true);
}

/*
 * Obtain the password for the target system from a user specified
 * file. The file must have the following properties:
 *
 *     o The first line of the file must contain the password
 *     o All lines following the first line will be ignored
 *     o The password must be plain text
 *     o The password must not exceed 256 characters
 *     o A trailing newline will be ignored
 *
 * In addition, to promote the practice of securing the password file
 * adequately, the following will result in a non-fatal warning:
 *
 *     o The file is accessible by someone other than the owner
 *     o The file is not owned by root
 *
 * Memory is allocated for the buffer used to hold the password string.
 * The caller must deallocate this memory once the password has been
 * processed.
 */
static bool_t
get_target_password_from_file(mig_arg_t *args, char **passwd)
{
	int		i;
	int		fd;
	size_t		len;
	char		*buf;
	struct stat	stat;
	bool_t		rv = false;

	ASSERT(args->passfile != NULL);

	/*
	 * Open the password file for reading.
	 */
	if ((fd = open(args->passfile, O_RDONLY)) < 0) {
		fprintf(stderr, "%s: %s\n", gettext("Unable to open specified "
		    "password file"), strerror(errno));
		return (false);
	}

	/*
	 * Check the basic properties of the file.
	 */
	if (fstat(fd, &stat) < 0) {
		fprintf(stderr, "%s: %s\n", gettext("Unable to stat specified "
		    "password file"), strerror(errno));
		goto done;
	}

	/* should be a regular file */
	if (!S_ISREG(stat.st_mode)) {
		fprintf(stderr, gettext("Specified password file is not a "
		    "plain text file\n"));
		goto done;
	}

	/* warn about access by someone other than the owner */
	if ((stat.st_mode & S_IAMB) & ~S_IRWXU) {
		fprintf(stderr, gettext("warning: users other than the owner "
		    "can access specified password file\n"));
	}

	/* warn about a file not owned by root */
	if (stat.st_uid != 0) {
		fprintf(stderr, gettext("warning: specified password file is "
		    "not owned by root\n"));
	}

	buf = Xmalloc(MAXPASSWD + 1);

	/* read the password from the first line of the file */
	if ((len = read(fd, buf, (MAXPASSWD + 1))) < 0) {
		fprintf(stderr, "%s: %s\n", gettext("Error reading specified "
		    "password file"), strerror(errno));
		Xfree(buf);
		goto done;
	}

	/* check for an empty file */
	if (len == 0) {
		fprintf(stderr, gettext("Specified password file is empty\n"));
		Xfree(buf);
		goto done;
	}

	/*
	 * Process the password.
	 */
	for (i = 0; i < (MAXPASSWD + 1); i++) {

		/* only interested in the first line */
		if (buf[i] == '\n')
			break;
	}

	/* check for a password that is too long */
	if (i == (MAXPASSWD + 1)) {
		fprintf(stderr, gettext("Password cannot be longer than %d "
		    "characters\n"), MAXPASSWD);
		Xfree(buf);
		goto done;

	}

	buf[i] = '\0';

	*passwd = buf;
	rv = true;

done:
	close(fd);
	return (rv);
}

/*
 * Obtain the password for the target system. This handles the password
 * being read from the user interactively and also directly from a file.
 */
static bool_t
get_target_password(mig_arg_t *args, char **passwd)
{
	bool_t	rv;

	if (args->passfile == NULL)
		return (get_target_password_from_user(passwd));

	priv_on();
	rv = get_target_password_from_file(args, passwd);
	priv_off();

	return (rv);
}

xmlNodePtr
create_migrate_XML(mig_arg_t *args)
{
	xmlNodePtr	root, cmd, data, envelope;
	xmlNodePtr	ldom_content, new_ldom_content, info_node;
	char		*new_ldom;
	char		*out;
	char		*password;
	struct passwd	*passwdp;

	root = xmlNewNode(NULL, XML_LDM_INTERFACE);
	xmlNewProp(root, VERSION_ATTR, LDOM_INTERFACE_VERSION);
	xml_add_ldm_interface_properties(root);

	cmd = xmlNewChild(root, NULL, XML_CMD, NULL);
	xmlNewChild(cmd, NULL, XML_ACTION, XML_MIGRATE_DOMAIN);
	data = xmlNewChild(cmd, NULL, XML_DATA, NULL);
	xmlNewProp(data, VERSION_ATTR, LDOM_V3_DATA_VERSION);

	envelope = xmlNewChild(data, NULL, ENVELOPE_NODE, NULL);
	xmlNewChild(envelope, NULL, REFERENCES_NODE, NULL);

	ldom_content = xmlNewChild(envelope, NULL, CONTENT_NODE, NULL);
	xmlNewProp(ldom_content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
	xmlNewProp(ldom_content, ID_ATTR, (xmlChar *)args->src_ldom);

	new_ldom_content = xmlNewChild(envelope, NULL, CONTENT_NODE, NULL);
	xmlNewProp(new_ldom_content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);

	new_ldom = (args->tgt_ldom != NULL) ? args->tgt_ldom : args->src_ldom;
	xmlNewProp(new_ldom_content, ID_ATTR, (xmlChar *)new_ldom);

	info_node = create_v3_section(new_ldom_content, RESOURCE_SECTION_TYPE,
	    XML_LDM_INFO);
	create_v3_property(info_node, XML_TARGET, (xmlChar *)args->host);

	/* lookup up the user if not specified */
	if (args->user == NULL) {
		passwdp = getpwuid(getuid());
		if (passwdp == NULL) {
			fprintf(stderr, "User name lookup failed\n");
			exit(1);
		}
		args->user = passwdp->pw_name;
	}

	ASSERT(args->user != NULL);

	if (!encode_string_base64(args->user, strlen(args->user), &out)) {
		fprintf(stderr, "Failed to encode user name\n");
		exit(1);
	}
	create_v3_property(info_node, XML_USERNAME, (xmlChar *)out);
	Xfree(out);

	if (!get_target_password(args, &password))
		exit(1);

	if (!encode_string_base64(password, strlen(password), &out)) {
		fprintf(stderr, "Failed to encode password\n");
		exit(1);
	}
	create_v3_property(info_node, XML_PASSWORD, (xmlChar *)out);
	Xfree(out);
	Xfree(password);

	if (args->dry_run == true) {
		create_v3_property(info_node, XML_DRY_RUN, NULL);
	}

	if (args->force == true) {
		create_v3_property(info_node, XML_FORCE, NULL);
	}

	return (root);
}

/*
 * Initiate a migration operation. A migration is handled separate
 * from other operations because it uses the XML interface over XMPP
 * to communicate with the LDom Manager. Returns 0 if the migration
 * is successful and 1 otherwise.
 */
static int
do_migration(int argc, char **argv)
{
	mig_arg_t	args = { 0 };
	int		rc;
	int		rv = 1;
	int		xmpp_sock;
	xmlNodePtr	root;
	char		*result;	/* string containing XML doc */

	rc = migrate_parse_options(argc - 1, argv + 1, &args);
	if (rc == -2) {
		migrate_usage(argv[0]);
		return (1);
	} else if (rc != 0) {
		fprintf(stderr, "%s", strerror(errno));
		return (1);
	}

	/*
	 * Check that this is not an attempt to migrate a domain to the
	 * same host (i.e. source == target)
	 */
	if (args.host == NULL || verify_target_hostname(args.host) == false)
		goto done;

	xmpp_sock = connect_local(XMPP_DEFAULT_PORT);
	if (xmpp_sock < 0) {
		fprintf(stderr, "%s\n", gettext("XMPP connection failed"));
		goto done;
	}

	root = create_migrate_XML(&args);

	/*
	 * Save the command name and source domain name. This will be
	 * used if the operation is canceled by a signal.
	 */
	cmd_name = argv[0];
	ldom_name = strdup(args.src_ldom);

	result = ldm_send_async_command(xmpp_sock, root);
	if (result == NULL) {
		printf(MSG_ERR_FAIL_SEND_CMD, argv[0]);
		goto done;
	}

	rv = handle_xml_resp(result, strlen(result));

done:
	Zfree(args.passfile);
	Zfree(args.src_ldom);
	Zfree(args.user);
	Zfree(args.host);
	Zfree(args.tgt_ldom);

	return (rv);
}

/* MIGXXX duplicates func in xml_event.o but linking that in is non-trivial */
/*ARGSUSED0*/
int
event_rm_all_event(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client,	int cmd_num)
{
	return (0);
}
