/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_LDOM_H_
#define	_LDOM_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <thread.h>
#include <synch.h>
#include <sys/time.h>
#include <uuid/uuid.h>

#include "ldmd_types.h"

#include "basics.h"
#include "chain.h"
#include "dr.h"
#include "hvctl.h"
#include "resource_id.h"
#include "mdesc_impl.h"
#include "response.h"
#include "keys.h"
#include "variables.h"
#include "console.h"
#include "memory.h"
#include "mig.h"
#include "pm_events.h"
#include "pm/pm_shared.h"
#include "master.h"
#include "policy.h"
#include "pcie.h"

#define	NEW_INSTANCE	(-1)
#define	NO_NEW_ID	(-1)

#define	DEFAULT_PORT		5311
#define	XMPP_DEFAULT_PORT	6482
#define	MIG_DEFAULT_PORT	8101	/* IANA registered port */
#define	MIG_LEGACY_PORT		4983	/* for backwards compatibility */

#define	HOSTID_PREFIX		0x84000000
#define	UNASSIGNED_HOSTID	0xFFFFFFFF

#define	GUEST_UTIL_HIST_TIME	(10000)	/* utilization history - in ms */
#define	GUEST_UTIL_POLL_INTV	(500)	/* utilization poll interval - in ms */
#define	GUEST_UTIL_TAB_NROWS	(GUEST_UTIL_HIST_TIME / GUEST_UTIL_POLL_INTV)

#define	DERIVED_HOSTID(mac_addr)    \
	((HOSTID_PREFIX & 0xff000000)|(mac_addr & 0x00ffffff))
#define	IS_DERIVED_HOSTID_ASSIGNED(mac_addr, hostid)	\
	((hostid == DERIVED_HOSTID(mac_addr)) || (hostid == UNASSIGNED_HOSTID))

/*
 * Get the index of the next most recent entry in the utilization history table
 * from the given index.
 */
#define	GET_PREV_GUEST_UTIL_IDX(i)			\
	(i == 0) ? (GUEST_UTIL_TAB_NROWS - 1) : i - 1

/*
 * Increment the index of the guest's utilization history table and add the
 * utilization value and time to the table.
 */
#define	ADD_GUEST_UTIL(ldp, val, t) {					       \
	ldp->stats.hist_idx = (ldp->stats.hist_idx + 1) % GUEST_UTIL_TAB_NROWS;\
	ldp->stats.hist_table[ldp->stats.hist_idx].value = val;		       \
	ldp->stats.hist_table[ldp->stats.hist_idx].time = t;		       \
}

/*
 * LDom states.
 */
typedef enum {
	INACTIVE	= 0x001, /* rule set exist but nothing else */
	BINDING		= 0x002, /* assigning resources */
	UNBINDING	= 0x004, /* unassigning resources */
	BOUND		= 0x008, /* configured in HV */
	STARTING	= 0x010, /* start requested, waiting for ack from HV */
	STOPPING	= 0x020, /* stop requested, waiting for ack from HV */
	SUSPENDING	= 0x040,
	RESUMING	= 0x080,
	SUSPENDED	= 0x100,
	ACTIVE		= 0x200, /* actively running */
} ldom_state_t;

/*
 * Zeus operating mode.
 */
typedef enum {
	ZEUS_CONFIG_MODE = 0,
	ZEUS_MGMT_MODE = 1
} zeus_mode_t;

/*
 * config_guests flags
 */
typedef enum {
	CONFIG,				/* Normal reconfig */
	GUEST_CONFIG,			/* Guest-only reconfig */
	RECONFIG,			/* Two-step (guest then HV) reconfig */
	DELAYED_CONFIG,			/* Delayed reconfig */
	DELAYED_RECONFIG,		/* Two-step delayed reconfig */
	CANCEL_DELAYED_CONFIG,		/* Cancel in-progress delayed reconf */
} config_flag_t;

/*
 * Name of start of day control domain - needs to be in sync with the MD
 * vbsc builds.
 */
#define	SOD_CONTROL_DOMAIN "primary"

/*
 * Be careful in using the macro in early stage of zeus startup
 */
#define	CONTROL_DOMAIN		(zeus.hvctl_ldom_p)

/* Determine if <ldp> (type: ldom_t *) is the control domain */
#define	IS_CONTROL_DOMAIN(ldp)	((ldp) == zeus.hvctl_ldom_p)

/* Setting/Checking flag to indicate if a domain is locked for modification */
#define	LOCK_LDOM(ldp)		((ldp)->flag.locked = true)
#define	UNLOCK_LDOM(ldp)	((ldp)->flag.locked = false)
#define	IS_LDOM_LOCKED(ldp)	((ldp)->flag.locked == true)

/*
 * Delayed reconfig types
 */
typedef enum {
	DEL_RECONFIG_UNKNOWN = 0x00,
	DEL_RECONFIG_ADD = 0x01,
	DEL_RECONFIG_REMOVE = 0x02,
	DEL_RECONFIG_REMOVE_PHYS = 0x4,
	DEL_RECONFIG_REMOVE_2PHASE = 0x8,
	DEL_RECONFIG_STOP = 0x10,
	DEL_RECONFIG_OTHER = 0x20,
	DEL_RECONFIG_LOCKED = 0x100
} del_reconfig_type_t;

#define	INVALID_GUESTID	0xffffffffu

enum {
	LDC_RAW,
	LDC_UNRELIABLE,
	LDC_RELIABLE,
	LDC_MODE_STREAM
};

struct bound_iodevice {
	char		*name;
	resid_t		*id;
	bool_t		no_pcie_reset;
	bool_t		io_mmu_bypass;
	CHAIN_LINK(bound_iodevice_t, ldom);
	CHAIN_LINK(bound_iodevice_t, pio);
	pri_iodevice_t	*phys_iodev;
	req_iodevice_t	*req_iop;	/* req. ptr that caused this binding */
	bool_t		virtual_pcie_bus; /* physical/virtual root complex? */
	pcie_bus_res_t	busres;		/* MSIs, DVMA ranges, etc. */
	md_node_t	*hvmd_iodev;	/* used in HV MD generation */
	CHAIN(bound_pcie_dev_t, dev_list); /* loanable devs for this pcie bus */
};

struct bound_pcie_dev {
	char		*cli_name;		/* like pci_0_0_c */
	char		*skel_path;		/* like /@400/@0/@c */
	resid_t		*id;			/* resource id */
	CHAIN_LINK(bound_pcie_dev_t, ldom); /* next/prev bound device in ldom */
	CHAIN_LINK(bound_pcie_dev_t, pio);  /* next/prev binding for phys dev */
	pri_pcie_dev_t	*phys_pcie_dev;	/* to corresponding phys device (PRI) */
	req_pcie_dev_t	*req_pcie_dev;	/* to corresponding constraint */
};

/*
 * Memory fragments are not quite memory blocks They are smaller blocks of
 * memory allocated from a fragment pool maintained by zeus, and assigned to
 * hypervisor activities such as containment for machine descriptions, LDC
 * maptables and scratch areas.
 */

#define	FRAG_SIZE	(512 * 1024)	/* sized to fit largest guest MD */
#define	MEM_MINSIZE	(256 * 1024 * 1024)	/* MIN memory size in bytes */

struct bound_mfrag {
	uint64_t	paddr;
	uint64_t	size;
	pri_mblock_t	*pmblockp;		/* allocated from this mblock */
	ldom_t		*ldomp;			/* assigned to this ldomp */
	int		ref_cnt;
	CHAIN_LINK(bound_mfrag_t, ldom);	/* bound to same ldom */
	CHAIN_LINK(bound_mfrag_t, link);	/* free list or other */
};

/*
 * LRI (LDom Requirements Info)
 */

/*
 * LDOM requirements.
 * At some point this requirement database
 * needs a more sophisticated structure.
 * FIXME:
 * Right now this is extremely simplistic, and we'll
 * update with a real rule set once the constraint engine
 * is properly implemented.
 */

struct req_iodevice {
	char		*name;
	CHAIN_LINK(req_iodevice_t, link);
	bool_t		bound;
	bool_t		io_mmu_bypass;
};

struct req_pcie_dev {
	/*
	 * Either the device path or pseudonym,
	 * e.g., either pci@400/pci@0/pci@c or PCIE1.
	 */
	char		*cli_name;
	bool_t		bound;		/* true if bound to an ldom */
	CHAIN_LINK(req_pcie_dev_t, link); /* prev/next constraint */
};

/*
 * Sun4v guest state as known in the hypervisor.
 */
typedef struct ldom_hv_state {
	uint64_t	state;
	uint8_t		soft;
	uint8_t		soft_update;	/* true if soft state needs updating */
	char		soft_str[SOFT_STATE_SIZE];
} ldom_hv_state_t;

enum {
	LDOM_HV_STATE = 1,
	LDOM_SOFT_STATE = 2
};

/*
 * Entry in the utilization history table
 */
typedef struct guest_util {
	double		value;		/* utilization value */
	hrtime_t	time;		/* time util value was obtained */
} guest_util_t;

/*
 * Guest utilization statistics
 */
typedef struct guest_stats {
	uint64_t	uptime;		/* time running, in seconds */
	double		cpu;		/* CPU utilization */
	int		hist_idx;	/* most recent entry in util history */
	guest_util_t	hist_table[GUEST_UTIL_TAB_NROWS];  /* history table */
} guest_stats_t;

/*
 * State of a memory DR operation as well as cache of (addr, size) pairs for
 * HV scrubbing.
 */
typedef struct mem_dr_state {
	bool_t		is_mem_add;	/* add/del operation */
	bool_t		inprogress;	/* MD bit for mem dr */
	bool_t		cancel;		/* cancel requested */
	uint64_t	size;		/* requested mem dr size in bytes */
	uint32_t	progress;	/* % of progress made so far */
	void		*statp;		/* status/progress on a mblk */
	int		status;		/* status */
	char		unit;		/* unit of requested size(M,G) */
	thread_t	tid;		/* thread for mem DR thread */
	struct	{			/* mblks to be scrubbed */
		int	nmblks;		/* nmblks to be scrubbed */
		int	max_nmblks;	/* max # of (addr, size) allocated */
		void	*unbind_mblks;  /* address, size pairs */
	} scrub;
	xmlChar		*event;		/* for event notificatin */
	uint64_t	auto_adj_size;	/* auto adj size */
} mem_dr_state_t;

/*
 * Basic LDOM structure.
 * Used as the binding for the requirements and resource
 * assignments made for the LDOM if BOUND and ACTIVE.
 * Also holds the on-going state for the LDOM.
 *
 * This structure can be entirely re-build from the
 * Zeus requirements database, and the HV MD retrieved from
 * an active HV.
 */

struct ldom {
	char		*namep;
	ldom_state_t	state;
	ldom_hv_state_t hv_state;
	guest_stats_t	stats;
	resid_t		*idp;		/* guest ID */
	resid_t		*res_id;	/* index into HV guest struct */

	pm_resid_t	*pm_vcpu_list;	/* resid list in case of reconfig */
	int		pm_vcpu_count;	/* number of cpus in resid list */

	struct ldom_database *dbp;

	struct {
		int md_notify;		/* MD update notification needed */
		int gen_new_md:1;	/* generate a new MD for this LDom */
		int locked:1;
		int sp_ldc_cons:1;
		int config_committed:1;	/* config has been committed to HV */
		int migration_source:1;
		int migration_target:1;
		int migration_error:1;
		int local_tick_sync:1;
		int memdr_in_progress:1;
	} flag;

	res_pool_t	guest_vcpu_ids;		/* guest specific vcpu ids */
	res_pool_t	guest_vmau_ids;		/* guest specific vmau ids */
	res_pool_t	guest_free_ldcs;	/* available LDCs */
	res_pool_t	free_ds_ports;		/* DS ports */
	res_pool_t	free_hvcfg_handles;	/* for devs not in MD */
#if VDEV_CFG_HANDLE
	res_pool_t	free_cfg_handles;	/* cfg_handles for devices */
#endif
	struct ds_channel *ds_chan;		/* used for DS communication */
	CHAIN(guest_sp_ldc_t, sp_ldcs);		/* LDC channels to SP */

	CHAIN_LINK(ldom_t, link);	/* all LDoms chain */
	CHAIN_LINK(ldom_t, inprog);	/* in progress chain for sequencer */

	CHAIN(struct waiter, waiters);	/* clients waiting on LDOm event */

	ldom_keys_t	keys;
	variables_t	variables;

	struct {
		uint64_t md_generation_num;	/* gen number of guest MD */
		uint64_t mem_generation_id;	/* memory generation ID */
		uint64_t mac_address;
		uint64_t tod_offset;
		uint64_t diagpriv;
		uint64_t perfctraccess;
		uint64_t perfctrhtaccess;
		uint64_t rngctlaccessible;
		uint64_t hostid;		/* cli specified hostid */
		uint64_t ldc_mapin_base;
		uint64_t ldc_mapin_size;
		uint64_t failure_policy;	/* "guestcontrol" policy */
		CHAIN(master_t, masters);	/* "guestdependency" */
		char uuid[UUID_PRINTABLE_STRING_LENGTH];	/* "uuid" */
	} props;

	struct {
		bound_mfrag_t	*committed_md_frag;
		bound_mfrag_t	*parsed_md_frag;
		bound_mfrag_t	*next_md_frag;
	} frags;

	struct {
		bool_t	flag_gen_new_md;
		uint64_t committed_pa;
		uint64_t parsed_pa;
		uint64_t next_pa;
	} restore;

	md_t		*devalias_md;
	md_t		*ds_sp_md;
	pri_image_t	*boot_imagep;

	CHAIN(ds_port_t, ds_ports);
	console_t 	console;

	CHAIN(vcc_t, vccs);

	CHAIN(vsw_t, vsws);
	CHAIN(vnet_t, vnets);

	CHAIN(vds_t, vdss);
	CHAIN(vdisk_t, vdisks);

	CHAIN(vdpcs_t, vdpcss);
	CHAIN(vdpcc_t, vdpccs);

	CHAIN(vldc_t, vldcs);
	CHAIN(vldcc_t, vldccs);

	struct {
		int	num_services;
		int	num_clients;
		CHAIN(struct vcpu, vcpus);
		CHAIN(struct vmau, vmaus);
		CHAIN(struct mblock, memory);
		/*
		 * List of committed memory bindings.  Used solely to guarantee
		 * mutual exclusivity of the newly generated mblk_resid against
		 * the last committed set.
		 */
		mem_blocks_t committed_memory;
		CHAIN(bound_iodevice_t,	iodevices);
#if !VDEV_CFG_HANDLE
		res_pool_t vcc_cfg_handles;
		res_pool_t vldc_cfg_handles;
		res_pool_t vldcc_cfg_handles;
		res_pool_t vsw_cfg_handles;
		res_pool_t vnet_cfg_handles;
		res_pool_t vds_cfg_handles;
		res_pool_t vdisk_cfg_handles;
		res_pool_t vdpcs_cfg_handles;
		res_pool_t vdpcc_cfg_handles;
#endif
	} bound;

	struct {
		CHAIN(struct vcpu, vcpus);
		CHAIN(struct vmau, vmaus);
	} susp;

	struct {
		int	num_vcpus;
		int	num_vmaus;
		uint64_t size_mem;
		CHAIN(req_iodevice_t, io_devices);
		CHAIN(req_pcie_dev_t, io_pcie_devs);
	} required;

	/* scratch work area used in checking binding LDC requirements */
	struct {
		int	num_ldcs;
	} work;

	ldm_resutils_t	pm_cpu_resutils;	/* CPU resource util for PM */

	mig_cache_t	mig_info;	/* info about a migration in progress */
	mig_state_t	mig_state;	/* state of a migration in progress */

	/* reserved vcpus for full core migration restriction */
	struct {
		CHAIN(struct vcpu, vcpus);
	} mig_rsvd;

	policy_chain_t  pchain;		/* list of DRM policies */
	mem_dr_state_t	mem_dr_state;	/* info about memory DR operation */
};

/* Main struct for all zeus' internal state. */

typedef struct {
	CHAIN(ldom_t,	ldoms);
	/* list of LDoms currently being sequenced */
	CHAIN(ldom_t,	inprogress);

	CHAIN(bound_mfrag_t, free_frags); 	/* list of free mem fragments */
	CHAIN(struct mblock, frag_mblocks);	/* list of frag mem mblocks */
	CHAIN(vds_t, vds_services);	/* list of all virtual disk servers */
	CHAIN(mpgroup_t, mpgroups);	/* list of disk multipath groups */
	CHAIN(struct mblock, rsvd_mblocks);	/* list of reserved mblocks */

	/*
	 * All top-level HV resource structures need a resource ID assigned
	 * for each instance.  These are the pools from which they're
	 * allocated.
	 */
	res_pool_t	free_vcpu_ids;
	res_pool_t	free_mblk_ids;
	res_pool_t	free_mau_ids;
	res_pool_t	free_pcidev_ids;
	res_pool_t	free_netdev_ids;
	res_pool_t	free_ldom_ids;
	res_pool_t	free_ldc_ids;
	res_pool_t	free_guest_ids;
	res_pool_t	free_console_ids;

	res_pool_t	free_sp_ldcs;
	/* currently a pool size of 1, since HVCTL is the only one */
	res_pool_t	free_hv_ldc_ids;

	ldom_t		*hvctl_ldom_p;	/* control domain, NULL if hvctl ext. */
	vldc_t		*ds_server;
	bound_mfrag_t	*current_hvmd_fp;
	bound_mfrag_t	*next_hvmd_fp;
	bound_mfrag_t	*reconf_fp;	/* for two-step reconfig */
	bound_mfrag_t	*prev_del_reconf_hvmd_fp; /* in case of HV rollback */
	struct {
		uint64_t current_pa;
		uint64_t next_pa;
		uint64_t reconf_pa;
	} restore;
	struct mblock	*hv_mblockp;	/* location of HV in main mem */

	/* true by default except for Legion backend */
	bool_t		ldc_console_config;

	mutex_t		mlock;		/* for now, one big honkin' lock */
	md_t		*hv_mdp;	/* ldmd copy of HV MD */
	uint64_t	hvmd_vers_major;	/* From "content-version" */
	uint64_t	hvmd_vers_minor;	/* From "content-version" */
	bool_t		allow_ctrl_domain_reconfig;
	bool_t		do_warm_start;
	zeus_mode_t	mode;		/* operating mode */
	bool_t		simulating;

	/* Static Direct I/O flag */
	bool_t		sdio_enabled;	/* true if Static Direct IO enabled */

	/*
	 * For a delayed reconfig, we need to keep track of the LDom
	 * being reconfigured, as well as pointers to the delayed MDs.
	 */
	ldom_t		*delayed_reconfig_ldom;
	uint32_t	delayed_reconfig_gid;
	md_t		*delayed_hvmd;
	del_reconfig_type_t delayed_reconfig_type;
	char		*admdir;	/* database and core admin directory */
	char		*pkgdir;	/* package directory */
	bool_t		domaining_enabled; /* flag if domaining is enabled */
	bool_t		is_next_spconfig;  /* Config same as selected on SP? */
	char		*current_spconfig; /* Which config is current */

	/* Configuration parameters */
	uint64_t	ra2pa_shift;	/* First bit of RA denoting segment # */
	uint64_t	max_ra_bits;	/* # RA bits supported by platform */
	uint64_t	guest_max_mblocks;	/* Number supported mblocks */
	char		*current_cookie;
	int		cookie_time;
	/*
	 * Future: This will need to be revisited if/when we add support for
	 * more than one migration at a time on a machine.
	 */
	char		*local_ip;	/* IP address of the local machine */
	char		*remote_ip;	/* Source IP used by the target */
} zeus_t;

extern zeus_t zeus;

bool_t db_init(void);

ldom_t	*new_ldom(char *namep);
char	*ldom_state_str(ldom_state_t);
void	free_ldom(ldom_t *ldp);

/*
 * Basic LDom ops.
 */
ldom_t *find_ldom_by_name(char *namep);
ldom_t *find_ldom_by_gid(uint64_t id);
ldom_t *find_ldom_by_resid(uint64_t id);

/*
 * Binding and memory operations.
 */
struct mblock *pull_bmblock(uint64_t basepa, uint64_t size);
bound_iodevice_t *pull_biodevice(char *device_type, uint64_t cfghdl,
    uint64_t res_id, bool_t io_mmu_bypass);
bound_iodevice_t *
pull_pcie_biodevice(char *device_type, uint64_t cfghdl,
    uint64_t res_id, bound_iodevice_t *tbiop, bool_t assign_all_res);
void destroy_bound_iodevice(bound_iodevice_t *bio);

bool_t cons_bind_vcpus(ldom_t *, struct cpu_adm *cpu_adm, response_t *);
bool_t cons_bind_vmaus(ldom_t *, struct mau_adm *mau_adm, response_t *);


/*
 * Status of an iodevice in PRI.
 */
typedef enum {
	IODEV_NOT_PRESENT,
	IODEV_BOUND,
	IODEV_UNBOUND
} iodev_st_t;

bool_t cpu_bind(pri_cpu_t *pcpup, ldom_t *ldp, int vid, int mmu_partid,
    int fraction);
bool_t match_iodev(char *iodev_p, pri_iodevice_t *pio_p);
pri_iodevice_t *find_matching_io(char *iodev_p, int *state);
void mau_link_vcpus(ldom_t *ldp, struct vmau *vmau);
struct vmau *mau_bound_to_cpu(ldom_t *ldp, uint32_t cpuvid);
bool_t	bind_io(ldom_t *, response_t *);
void	unbind_io(ldom_t *, bound_iodevice_t *);
bool_t	bind_pciedevs(ldom_t *ldp, response_t *resp);
void unbind_pciedev(ldom_t *ldp, bound_iodevice_t *bio, bound_pcie_dev_t *bdev);
bool_t	bind_an_ldom(ldom_t *, response_t *);
bool_t	unbind_an_ldom(ldom_t *);
req_iodevice_t *add_io_constraint(ldom_t *ldp, char *name,
    bool_t io_mmu_bypass);
req_pcie_dev_t *add_pciedev_constraint(ldom_t *ldp, char *name);
bool_t	ldom_io_reset_allowed(ldom_t *ldp);

/*
 * Delayed reconfiguration
 */
bool_t config_mod_allowed(ldom_t *ldp, del_reconfig_type_t type,
    dr_op_type_t dr_type, response_t *resp);
void initiate_or_update_delayed_reconfig(ldom_t *ldp, del_reconfig_type_t type,
    response_t *resp);
void seq_config_now_or_later(ldom_t *ldp, del_reconfig_type_t type,
    response_t *resp);
bool_t in_delayed_reconfig(ldom_t *ldp);
void cancel_delayed_reconfig(ldom_t *ldp, struct cli_conn *client,
    response_t *resp);

/*
 * Sequencer operations.
 */
void seq_bind_new(void);
void seq_send_md_update_notifs(void);
void seq_start_bound(void);
void seq_stop_bound(bool_t force);
void seq_unbind_ldoms(void);
void seq_rm_ldoms(void);
void seq_download_mds_do_config(ldom_t *ldp);
void seq_rm_reconfigure(ldom_t *ldp);
int seq_vio_dr_add(ldom_t *ldp, char *type, uint64_t dev_id, response_t *resp);
int seq_vio_dr_rm(ldom_t *ldp, char *type, uint64_t dev_id, response_t *resp);
void seq_update_vars(ldom_t *ldp, bool_t notify_guest);
void seq_dequeue_ds_svc(struct ds_svc *svcp, bool_t locking_reqd);
void seq_dequeue_ds_chan(struct ds_channel *chan);
md_t *seq_gen_guest_md(ldom_t *ldp, bool_t ignore_pm);
char *hv_state_str(uint64_t hv_state);
void seq_delayed_reconfig_finish(ldom_t *ldp);
void free_all_staged_ldcs(ldom_t *ldom);
void seq_clear_ldom_flags(void);

/*
 * DR operations
 */
/* vcpu DR operations */
int seq_add_vcpus(ldom_t *ldp, response_t *resp);
void cleanup_partial_vcpu_add(ldom_t *ldp, int num, uint32_t *vcpu_ids);
int seq_rm_vcpus(ldom_t *ldp, struct vcpu **vcpus, int num, bool_t force,
    uint32_t *op_stats, response_t *resp);
void seq_rm_vcpus_cleanup(ldom_t *ldp);

/* crypto DR operations */
int seq_add_cryptos(ldom_t *ldp, response_t *resp);
void cleanup_partial_crypto_add(ldom_t *ldp, int num, struct vmau **vmaus);
int seq_rm_cryptos(ldom_t *ldp, struct vmau **vmaus, int num,
    uint32_t *op_stats, response_t *resp);
void seq_rm_cryptos_cleanup(ldom_t *ldp);
int seq_reconfig_cryptos(ldom_t *ldp, struct vmau **vmaus, int num,
    response_t *resp);

/* vio DR operations */
int seq_vio_dr_add(ldom_t *ldp, char *type, uint64_t dev_id, response_t *resp);
int seq_vio_dr_rm(ldom_t *ldp, char *type, uint64_t dev_id, response_t *resp);

#ifdef __cplusplus
}
#endif

#endif /* _LDOM_H_ */
