/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines and tables specific to the 'ldm cancel-reconf' and
 * 'ldm cancel-operation' commands.
 */

#include <string.h>

#include <libintl.h>

#include "basics.h"
#include "cli_conn.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "errors.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "mig.h"
#include "response.h"


cli_res_cmds_t	cli_reconf_cmds = {
	.resinfop = "reconf",
	.cancel_op =	{ cli_cancel_reconf, "<ldom>" },
};

cli_res_cmds_t	cli_cancel_cmds = {
	.resinfop = "operation",
	.cancel_op =	{ cli_cancel_operation,
		"(migration | reconf | memdr) <ldom>" },
};


/* List or cancel a delayed reconfiguration */
cli_result_t
cli_cancel_reconf(int argc, char **argv, cli_conn_t *client)
{
	response_t *resp = client->resp;
	ldom_t *ldp = NULL;

	if (argc == 0) {
		rprintf(resp, gettext("No LDom(s) specified\n"));
		return (CLI_RESULT_USAGE);
	}

	if ((ldp = ldm_get_ldp(resp, argv[--argc])) == NULL)
		return (CLI_RESULT_FAIL);

	if (check_locked_ldom(ldp, resp))
		return (CLI_RESULT_FAIL);

	if (zeus.delayed_reconfig_ldom == NULL) {
		rprintf(resp, gettext(
		    "No delayed reconfiguration in process\n"));
		return (CLI_RESULT_FAIL);
	}

	if (zeus.delayed_reconfig_ldom != ldp) {
		rprintf(resp, gettext(
		    "LDom %s not in process of a delayed reconfiguration\n"),
		    ldp->namep);
		return (CLI_RESULT_FAIL);
	}

	/*
	 * Can cancel only if not locked & no VIO remove
	 * operations have occurred.
	 */
	if (zeus.delayed_reconfig_type == DEL_RECONFIG_LOCKED) {
		rprintf(resp, gettext(
"All further delayed reconfig operations have been locked out until LDom\n"
"%s reboots.\n"), zeus.delayed_reconfig_ldom->namep);
		return (CLI_RESULT_FAIL);
	}

	if (zeus.delayed_reconfig_type & DEL_RECONFIG_REMOVE_2PHASE) {
		rprintf(resp,
"Cannot cancel this delayed reconfig because a 2-phase remove operation\n"
"was recorded\n");
		return (CLI_RESULT_FAIL);
	}

	/*
	 * There is no easy way to resync zeus' internal state
	 * after cancelling a delayed reconfig, so the
	 * cancel_delayed_reconfig() routine calls exit() when it's
	 * done.  SMF will then restart us, and all will be
	 * back in good order.
	 */
	cancel_delayed_reconfig(ldp, client, resp);
	/*NOTREACHED*/
	zeus.delayed_reconfig_ldom = NULL;
	return (CLI_RESULT_OK);
}

/*
 * Cancel a domain migration
 */

static cli_result_t
cancel_migration(int argc, char **argv, cli_conn_t *client)
{
	response_t *resp = client->resp;
	ldom_t *ldp = NULL;

	ASSERT(argc == 2);

	if ((ldp = ldm_get_ldp(resp, argv[--argc])) == NULL)
		return (CLI_RESULT_FAIL);

	/* Check that it isn't a request to cancel a migration on the target */
	if (ldp->flag.migration_target) {
		rprintf(resp, MSG_ERR_MIG_CANNOT_CANCEL_ON_TARGET, ldp->namep);
		return (CLI_RESULT_FAIL);
	}

	/* We need to check if any sort of a migration is in progress */
	if ((ldp->flag.migration_source | ldp->flag.migration_error) == 0) {
		rprintf(resp, MSG_ERR_MIG_NOT_IN_PROGRESS, ldp->namep);
		return (CLI_RESULT_FAIL);
	}

	ASSERT(in_delayed_reconfig(NULL) == false);

	/* notify the migration thread that it should finish up */
	flag_migration_for_cancellation(ldp);

	/*
	 * Although the cancellation is asynchronous, it is expected
	 * to be close to instantaneous so it's OK to print that the
	 * migration has been cancelled
	 */
	rprintf(resp, MSG_ERR_MIG_CANCEL_COMPLETED, ldp->namep);

	return (CLI_RESULT_OK);
}

/*
 * Cancel a mem DR operation
 */

static cli_result_t
cancel_memdr(int argc, char **argv, cli_conn_t *client)
{
	ldom_t *ldp;
	response_t *resp = client->resp;

	if ((ldp = ldm_get_ldp(resp, argv[--argc])) == NULL)
		return (CLI_RESULT_USAGE);
	if (!ldp->flag.memdr_in_progress) {
		rprintf(resp, MSG_ERR_MEM_DR_NOT_IN_PROGRESS, ldp->namep);
		return (CLI_RESULT_FAIL);
	}
	if (!mem_cancel_dr(ldp, resp))
		return (CLI_RESULT_FAIL);
	return (CLI_RESULT_OK);
}

cli_result_t
cli_cancel_operation(int argc, char **argv, cli_conn_t *client)
{
	response_t *resp = client->resp;

	if (argc != 2) {
		rprintf(resp, gettext("Improper arguments supplied\n"));
		return (CLI_RESULT_USAGE);
	}

	/*
	 * Check if the first argument is one of 'migration' or
	 * 'reconf' and deal with it accordingly.
	 */
	if (strcmp(argv[0], "migration") == 0)
		return (cancel_migration(argc, argv, client));
	else if (strcmp(argv[0], "reconf") == 0)
		return (cli_cancel_reconf(argc, argv, client));
	else if (strcmp(argv[0], "memdr") == 0)
		return (cancel_memdr(argc, argv, client));
	else
		return (CLI_RESULT_USAGE);
}
