/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Various configuration error checking utilities shared by multiple
 * CLI commands.
 */

#include <stddef.h>

#include <libintl.h>

#include "basics.h"
#include "cli_util.h"
#include "dr_cpu.h"
#include "dr_crypto.h"
#include "vcpu.h"
#include "vio.h"
#include "vmau.h"
#include "errors.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "response.h"
#include "dr_mem.h"

static bool_t delayed_reconfig_allow(
    ldom_t *ldp, del_reconfig_type_t type, response_t *resp);

static void
reconfig_mode_change_warning(response_t *resp)
{
	if (zeus.delayed_reconfig_ldom != NULL) {
		rprintf(resp, "%s\n", LDM_SEPARATOR);
		rprintf(resp, gettext(
"Notice: LDom %s is in the process of a delayed reconfiguration.\n"
"Any changes made to %s will only take effect after it reboots.\n"),
		    zeus.delayed_reconfig_ldom->namep,
		    zeus.delayed_reconfig_ldom->namep);
		rprintf(resp, "%s\n", LDM_SEPARATOR);
	}
}

void
config_mode_change_warning(response_t *resp)
{
	if (zeus.mode != ZEUS_CONFIG_MODE) {
		reconfig_mode_change_warning(resp);
		return;
	}

	rprintf(resp, "%s\n%s\n%s\n", LDM_SEPARATOR, gettext(
"Notice: the LDom Manager is running in configuration mode. Any configuration\n"
"changes made will only take effect after the machine configuration is\n"
"downloaded to the system controller and the host is reset."),
	    LDM_SEPARATOR);
}

void
config_mode_info_warning(response_t *resp, dispctl_t *dctl)
{
	if (zeus.mode != ZEUS_CONFIG_MODE)
		return;

	if ((dctl != NULL) && dctl->parseable)
		return;

	rprintf(resp, "%s\n%s\n%s\n", LDM_SEPARATOR, gettext(
"Notice: the LDom Manager is running in configuration mode. Configuration and\n"
"resource information is displayed for the configuration under construction;\n"
"not the current active configuration. The configuration being constructed\n"
"will only take effect after it is downloaded to the system controller and\n"
"the host is reset."),
	    LDM_SEPARATOR);
}

static bool_t
check_dr_op_capable(ldom_t *ldp, dr_op_type_t type, response_t *resp)
{
	bool_t	dr_capable = false;	/* set to true if DR is allowed */
	bool_t	msg_printed = false;	/* set to true if callee prints msg */

	switch (type) {
	case DR_OP_TYPE_CPU:
		if (cli_check_cpu_dr_capable(ldp, resp) == CPU_DR_OP_SUCC)
			dr_capable = true;
		msg_printed = true;
		break;

	case DR_OP_TYPE_CRYPTO:
		if (cli_check_crypto_dr_capable(ldp, resp) == CRYPTO_DR_OP_SUCC)
			dr_capable = true;
		msg_printed = true;
		break;

	case DR_OP_TYPE_VIO:
		if (vio_check_dr_capable(ldp, NULL) == true)
			dr_capable = true;
		break;

	case DR_OP_TYPE_VAR:
		if (in_delayed_reconfig(NULL) == false)
			dr_capable = true;
		break;

	case DR_OP_TYPE_MEM:
		if (cli_check_mem_dr_capable(ldp, resp) == MEM_DR_OP_SUCC)
			dr_capable = true;
		msg_printed = true;
		break;

	case DR_OP_TYPE_PIO:
	case DR_OP_TYPE_NDPS:
	case DR_OP_TYPE_VIO_SET:
		break;			/* DR not yet implemented */

	case DR_OP_TYPE_VIO_MD_NOP:
	case DR_OP_TYPE_DOMAIN:
		dr_capable = true;	/* not a DR operation */
		break;

	default:
		ASSERT(0);
	}

	/*
	 * If the DR operation would fail and we haven't already printed an
	 * error message in the function that did the actual check, we print
	 * a generic message now.
	 */
	if (dr_capable == false && msg_printed == false)
		rprintf(resp, MSG_ERR_DO_OP_BOUND_OR_INACTIVE);

	return (dr_capable);
}

/*
 * Determine whether a particular configuration operation is allowed, based on
 * a number of criteria:
 * 	- Is a migration in progress
 * 	- Will this result in a dynamic or delayed reconfiguration
 * 	- Is this a control domain
 */
bool_t
config_mod_allowed(ldom_t *ldp, del_reconfig_type_t type,
    dr_op_type_t dr_type, response_t *resp)
{
	ldom_t	*tmpldp;

	config_mode_change_warning(resp);

	/*
	 * If any domain on the machine other than the one specified by 'ldp'
	 * is undergoing a migration then a reconfig must be blocked as
	 * the MDs could get out of sync.
	 *
	 * Note: We skip 'ldp' because the migration thread ends up in this
	 * function as it iterates though the devices while it binds the domain
	 * on the target during the migration. All other requests to modify
	 * 'ldp' will fail because the domain is locked so it's safe to skip
	 * it in this case.
	 */
	CHAIN_ITER(zeus.ldoms, tmpldp, link) {
		if (tmpldp == ldp)
			continue;

		if (MIG_INPROGRESS(tmpldp) == true) {
			rprintf(resp, MSG_ERR_OP_NOT_ALLOWED, ldp->namep);
			rprintf(resp, MSG_ERR_MIG_IN_PROGRESS_ON_LDOM,
			    tmpldp->namep);
			return (false);
		}
	}

	/*
	 * No modifications are permitted on a locked LDom except that a domain
	 * can be stopped while a memory DR operation is in progress.
	 */
	if (ldp->flag.memdr_in_progress == true && type == DEL_RECONFIG_STOP)
		return (true);

	if (check_locked_ldom(ldp, resp) == true)
		return (false);
	/*
	 * If the domain is not the control domain, delayed reconfiguration
	 * is not allowed. If a delayed reconfiguration is in progress,
	 * the request is rejected. Otherwise, the domain must be either
	 * bound/inactive or support Dynamic Reconfiguration (DR).
	 *
	 * NOTE: We would like to remove delayed config support for the
	 * control domain also but since we cannot put it in the bound
	 * (or inactive) state, we cannot remove that support until
	 * all resources (memory, physical I/O, etc) support DR.
	 */
	if (IS_CONTROL_DOMAIN(ldp) == false) {
		if (in_delayed_reconfig(NULL) == true) {
			rprintf(resp, MSG_ERR_DEL_RECONF_NOTSUP, ldp->namep);
			return (false);
		}

		/*
		 * Check to see whether a Dynamic Reconfiguration (DR)
		 * operation on a domain is supported.
		 */
		if (ldp->state == ACTIVE &&
		    check_dr_op_capable(ldp, dr_type, resp) == false) {
			return (false);
		}

		return (true);
	}

	/*
	 * Check if a delayed reconfiguration operation is allowed. This is
	 * just to check it's feasible, the operation may subsequently be
	 * carried out as a DR operation if the control domain supports it.
	 */
	if (delayed_reconfig_allow(ldp, type, resp) == false)
		return (false);

	return (true);
}

/*
 * Determine whether a particular delayed reconfiguration is allowed, based on
 * whether one is already in progress & the LDom involved.
 */
static bool_t
delayed_reconfig_allow(ldom_t *ldp, del_reconfig_type_t type, response_t *resp)
{
	ldom_t	*tmpldp;

	/*
	 * If any domain on the machine other than the one specified by 'ldp'
	 * is undergoing a migration then a reconfig must be blocked as
	 * the MDs could get out of sync.
	 *
	 * Note: We skip 'ldp' because the migration thread ends up in this
	 * function as it iterates though the devices while it binds the domain
	 * on the target during the migration. All other requests to modify
	 * 'ldp' will fail because the domain is locked so it's safe to skip
	 * it in this case.
	 */
	CHAIN_ITER(zeus.ldoms, tmpldp, link) {
		if (tmpldp == ldp)
			continue;

		if (MIG_INPROGRESS(tmpldp) == true) {
			rprintf(resp, MSG_ERR_OP_NOT_ALLOWED, ldp->namep);
			rprintf(resp, MSG_ERR_MIG_IN_PROGRESS_ON_LDOM,
			    tmpldp->namep);
			return (false);
		}
	}

	/*
	 * The caller will usually have already checked to see if an LDom is
	 * locked but we check again just in case.
	 */
	if (check_locked_ldom(ldp, resp) == true)
		return (false);

	if (type == DEL_RECONFIG_STOP && (zeus.delayed_reconfig_ldom == ldp ||
	    zeus.delayed_reconfig_ldom == NULL))
		return (true);

	if (zeus.delayed_reconfig_type == DEL_RECONFIG_LOCKED) {
		rprintf(resp, gettext(
"All further delayed reconfig operations have been locked out until LDom\n"
"%s reboots.\n"), zeus.delayed_reconfig_ldom->namep);
		return (false);
	}

	if (zeus.delayed_reconfig_ldom == NULL)
		return (true);

	/*
	 * While any LDom is in the process of a delayed reconfiguration,
	 * disallow any modification to any other LDom, or to all LDoms (i.e.
	 * via a '-a' switch).
	 */
	if (ldp == NULL || zeus.delayed_reconfig_ldom != ldp) {
		if (resp != NULL)
			rprintf(resp, gettext(
"Operation on other LDom(s) not allowed while LDom %s is in the process\n"
"of a delayed reconfiguration\n"),
			    zeus.delayed_reconfig_ldom->namep);
		return (false);
	}

	if (hvctl_state.minor_ver < 1) {
		/*
		 * After a DEL_RECONFIG_REMOVE_PHYS, no further
		 * DEL_RECONFIG_REMOVE_2PHASE operations can be allowed.
		 */
		if (type == DEL_RECONFIG_REMOVE_2PHASE &&
		    (zeus.delayed_reconfig_type & DEL_RECONFIG_REMOVE_PHYS)) {
			if (resp != NULL)
				rprintf(resp, gettext(
"This remove operation not allowed as part of this delayed reconfiguration\n"
"because a previous physical hardware removal was requested.\n"));
			return (false);
		}
	}
	return (true);
}
