/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines and tables specific to operations on console
 * resources 'vconscon' and 'vconsole'.
 */

#include <string.h>
#include <stdlib.h>

#include <libintl.h>

#include "allocate.h"
#include "basics.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "console.h"
#include "database.h"
#include "database_internal.h"
#include "dr.h"
#include "ldmd_types.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "response.h"
#include "vcc.h"
#include "vio.h"
#include "xml_cntrl_common.h"
#include "xml_event.h"


cli_res_cmds_t	cli_vcc_cmds = {
	.resinfop = "vconscon",
	.add_op =	{
		cli_vcc_add,
		"port-range=<x>-<y> <vcc_name> <ldom>" },
	.set_op	=	{
		cli_vcc_set,
		"port-range=<x>-<y> <vcc_name>" },
	.remove_op =	{
		cli_vcc_remove,
		"[-f] <vcc_name>" },
};

cli_res_cmds_t	cli_vcons_cmds = {
	.resinfop = "vconsole",
	.set_op	=	{ cli_vcons_set,
	    "[port=[<port-num>]] [group=<group>] [service=<vcc_server>] <ldom>"
	},
};


/*
 * Parse a numeric range, fail if its not in the form X-Y
 * or if X>Y or if X or Y is out of the range of min & max.
 *
 * Returns false on a failure.
 */
static bool_t
parse_range(char *argp, int min, int max, int *min_retp, int *max_retp)
{
	char *basep, *esp;
	long long val;
	int	first;
	int	last;

	val = strtoll(argp, &esp, 10);
	if (esp == argp || val < (long long)min || val > (long long)max)
		return (false);

	first = (int)val;

	if (esp[0] != '-')
		return (false);

	basep = esp+1;

	val = strtoll(basep, &esp, 10);
	if (esp == argp || val < (long long)min || val > (long long)max)
		return (false);

	last = (int)val;

	if (esp[0] != '\0')
		return (false);

	if (first > last)
		return (false);

	*min_retp = first;
	*max_retp = last;

	return (true);
}

/*
 * CLI commands for virtual console concentrator (vcc)
 */

/*
 * Add a vcc to a domain
 * port-range=<x>-<y> <vcc_name> <ldom>
 */
cli_result_t
cli_vcc_add(int argc, char **argvp, cli_conn_t *clientp)
{
	int i;
	char	*vcc_namep;
	char	*ldom_namep;
	ldom_t	*ldp;
	response_t *resp = clientp->resp;
	int min_port = -1;
	int max_port;

	if (argc != 3)
		return (CLI_RESULT_USAGE);

	/* Get the port-range */
	for (i = 0; i < argc-2; i++) {
		char *namep;
		char *valuep;

		namep = strtok(argvp[i], "=");
		valuep = strtok(NULL, "=");

		if (strcmp(namep, "port-range") == 0) {
			if (valuep == NULL || !parse_range(valuep,
			    MIN_CONSOLE_TCP_PORT, MAX_CONSOLE_TCP_PORT,
			    &min_port, &max_port)) {
				/*
				 * TRANSLATION_NOTE
				 * Don't translate 'port-range'
				 */
				rprintf(resp, gettext("Invalid port-range\n"));
				return (CLI_RESULT_FAIL);
			}
		} else {
			/* print error and usage */
			if (valuep != NULL)
				rprintf(resp, gettext(
				    "Unknown property: %s\n"), namep);
			return (CLI_RESULT_USAGE);
		}
	}

	if ((argc-i) != 2)
		return (CLI_RESULT_USAGE);

	if (min_port == -1) {
		/* TRANSLATION_NOTE Don't translate 'port-range=' */
		rprintf(resp, gettext(
		    "The port-range= property must be specified.\n"));
		return (CLI_RESULT_USAGE);
	}

	vcc_namep = argvp[i];
	if (!is_valid_object_name(resp, vcc_namep)) {
		rprintf(resp, gettext("Illegal name for VCC service: %s\n"),
		    vcc_namep);
		return (CLI_RESULT_FAIL);
	}

	ldom_namep = argvp[i+1];

	if ((ldp = ldm_get_ldp(resp, ldom_namep)) == NULL)
		return (CLI_RESULT_FAIL);

	if (!config_mod_allowed(ldp, DEL_RECONFIG_ADD, DR_OP_TYPE_VIO, resp))
		return (CLI_RESULT_FAIL);

	if (vcc_add(ldp, vcc_namep, min_port, max_port, resp) != 0)
		return (CLI_RESULT_FAIL);

	database_save();
	event_send_res_event(XML_ADD_VCONSCON, ldp, NULL, NULL, clientp);
	return (CLI_RESULT_OK);
}

/*
 * Modify a vcc in a domain
 * port-range=<x>-<y> <vcc_name>
 */
cli_result_t
cli_vcc_set(int argc, char **argvp, cli_conn_t *clientp)
{
	vcc_t *vcc;
	char *vcc_namep;
	response_t *resp = clientp->resp;
	int min_port = -1;
	int max_port;
	int i;

	if (argc <= 1)
		return (CLI_RESULT_USAGE);

	/* optional arguments processed first */
	for (i = 0; i < argc-1; i++) {
		char *namep;
		char *valuep;

		namep = strtok(argvp[i], "=");
		valuep = strtok(NULL, "=");

		if (strcmp(namep, "port-range") == 0) {
			if (valuep == NULL || !parse_range(valuep,
			    MIN_CONSOLE_TCP_PORT, MAX_CONSOLE_TCP_PORT,
			    &min_port, &max_port)) {
				/*
				 * TRANSLATION_NOTE
				 * Don't translate 'port-range='
				 */
				rprintf(resp, gettext("Invalid port-range\n"));
				return (CLI_RESULT_FAIL);
			}
		} else {
			/* print error and usage */
			if (valuep != NULL)
				rprintf(resp,
				    gettext("Unknown property: %s\n"), namep);
			return (CLI_RESULT_USAGE);
		}
	}

	if ((argc-i) != 1)
		return (CLI_RESULT_USAGE);

	if (min_port == -1) {
		/* TRANSLATION_NOTE Don't translate 'port-range=' */
		rprintf(resp,
		    gettext("The port-range= property must be specified.\n"));
		return (CLI_RESULT_USAGE);
	}

	vcc_namep = argvp[i];
	if (!is_valid_object_name(resp, vcc_namep)) {
		rprintf(resp,
		    gettext("Illegal name for VCC service: %s\n"), vcc_namep);
		return (CLI_RESULT_FAIL);
	}

	if ((vcc = vcc_find_by_name(vcc_namep)) == NULL) {
		rprintf(resp, gettext("vcc service \"%s\" not found\n"),
		    vcc_namep);
		return (CLI_RESULT_FAIL);
	}

	/*
	 * Modifying a vcc is not a VIO DR operation (we don't modify MDs)
	 * so we only need to pass DR_OP_TYPE_VIO_MD_NOP
	 */
	if (config_mod_allowed(vcc->ldp, DEL_RECONFIG_OTHER,
	    DR_OP_TYPE_VIO_MD_NOP, resp) == false)
		return (CLI_RESULT_FAIL);

	if (vcc_modify(vcc, min_port, max_port, resp) != 0)
		return (CLI_RESULT_FAIL);

	database_save();
	event_send_res_event(XML_SET_VCONSCON, NULL, vcc_namep, VCC_NODE,
	    clientp);

	return (CLI_RESULT_OK);
}

cli_result_t
cli_vcc_remove(int argc, char **argv, cli_conn_t *client)
{
	response_t *resp = client->resp;
	bool_t force = false;
	int i;
	char *vcc_name;
	vcc_t *vcc;

	/* extract meaningful flags */
	for (i = 0; i < argc && argv[i][0] == '-'; i++) {
		char *argp = argv[i];
		if (strcmp(argp, "-f") == 0 ||
		    strcmp(argp, "--force") == 0) {
			force = true;
		} else if (strcmp(argp, "--") == 0) {
			break;
		} else {
			rprintf(resp, gettext("Unknown flag %s\n"), argp);
			return (CLI_RESULT_USAGE);
		}
	}

	/* service name should be last arg in list */

	if (i != argc - 1)
		return (CLI_RESULT_USAGE);

	vcc_name = argv[i];

	if (!is_valid_object_name(resp, vcc_name)) {
		rprintf(resp, gettext("%s is not a valid VCC service name\n"),
		    vcc_name);
		return (CLI_RESULT_FAIL);
	}

	if ((vcc = vcc_find_by_name(vcc_name)) == NULL) {
		rprintf(resp, gettext("VCC service \"%s\" not found\n"),
		    vcc_name);
		return (CLI_RESULT_FAIL);
	}

	if (config_mod_allowed(vcc->ldp, DEL_RECONFIG_REMOVE,
	    DR_OP_TYPE_VIO, resp) == false)
		return (CLI_RESULT_FAIL);

	if (vcc_remove(vcc, force, resp) != 0) {
		rprintf(resp, gettext(
		    "Failed to remove virtual console concentrator\n"));
		return (CLI_RESULT_FAIL);
	}

	database_save();

	event_send_res_event(XML_REMOVE_VCONSCON, NULL, vcc_name, VCC_NODE,
	    client);
	return (CLI_RESULT_OK);
}

/*
 * CLI command for virtual console (vcons)
 */

cli_result_t
cli_vcons_set(int argc, char **argvp, cli_conn_t *clientp)
{
	int i;
	char *group = NULL;
	char *service = NULL;
	char *ldom_namep;
	response_t *resp = clientp->resp;
	ldom_t *ldp;
	int port = NEW_INSTANCE;

	/*
	 * There are no options, and at least one param required
	 * [port=<port>] [group=<group>] [service=<vcc>]
	 */

	if (argc < 2 || argc > 4)
		return (CLI_RESULT_USAGE);

	ldom_namep = argvp[argc - 1];

	if ((ldp = ldm_get_ldp(resp, ldom_namep)) == NULL)
		return (CLI_RESULT_FAIL);

	/* optional arguments processed first */
	for (i = 0; i < argc-1; i++) {
		char *namep;
		char *valuep;

		namep = strtok(argvp[i], "=");
		valuep = strtok(NULL, "=");

		if (strcmp(namep, "port") == 0) {
			if (valuep == NULL) {
				if (is_console_set(ldp) == false) {
					rprintf(resp, gettext("No non-default "
					    "console constraints exist, cannot "
					    "specify empty port constraint\n"));
					return (CLI_RESULT_FAIL);
				}
				port = _AUTO_VAL_;
			} else {
				char *esp;
				port = strtoll(valuep, &esp, 10);

				if (esp == valuep || *esp != '\0' ||
				    ((port != 0) &&
				    ((port < MIN_CONSOLE_TCP_PORT) ||
				    (port > MAX_CONSOLE_TCP_PORT)))) {
					rprintf(resp,
					    gettext("Port out of range\n"));
					return (CLI_RESULT_FAIL);
				}
			}
		} else if (strcmp(namep, "service") == 0) {
			service = valuep;
			if (service == NULL) {
				rprintf(resp, gettext("Cannot specify empty "
				    "virtual console concentrator (vcc) "
				    "constraint\n"));
				return (CLI_RESULT_USAGE);
			} else if (!is_valid_object_name(resp, service)) {
				rprintf(resp, gettext("Illegal console "
				    "concentrator (vcc) name: %s\n"), service);
				return (CLI_RESULT_FAIL);
			}
		} else if (strcmp(namep, "group") == 0) {
			group = valuep;
			if (group == NULL) {
				if (is_console_set(ldp) == false) {
					rprintf(resp, gettext("No non-default "
					    "console constraints exist, cannot "
					    "specify empty group "
					    "constraint\n"));
					return (CLI_RESULT_FAIL);
				}
				group = _AUTO_STR_;

			} else if (!is_valid_object_name(resp, group)) {
				rprintf(resp, gettext("Illegal group name: "
				    "%s\n"), group);
				return (CLI_RESULT_FAIL);
			}
		} else {
			rprintf(resp, gettext("Illegal option: %s\n"), namep);
			return (CLI_RESULT_USAGE);
		}
	}

	if (service == NULL && group == NULL && port == NEW_INSTANCE) {
		rprintf(resp, gettext("Invalid console specification\n"));
		return (CLI_RESULT_USAGE);
	}

#if LIVE_GRP_REBIND
	if (config_mod_allowed(ldp, DEL_RECONFIG_OTHER,
	    DR_OP_TYPE_VIO_MD_NOP, resp) == false)
		return (CLI_RESULT_FAIL);
#else
	config_mode_change_warning(resp);

	if (check_locked_ldom(ldp, resp))
		return (CLI_RESULT_FAIL);
#endif

	if (console_modify(ldp, service, group, port, resp) == 0) {
		event_send_res_event(XML_SET_VCONSOLE, ldp, NULL, NULL,
		    clientp);
		database_save();
		return (CLI_RESULT_OK);
	} else {
		return (CLI_RESULT_FAIL);
	}
}
