/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines and tables for operations on the 'vmau' resource
 * type.
 */

#include <stdint.h>
#include <string.h>

#include <libintl.h>

#include "allocate.h"
#include "basics.h"
#include "cli_conn.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "database.h"
#include "dr.h"
#include "fatal.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "option.h"
#include "response.h"
#include "support.h"
#include "vmau.h"
#include "xml_cntrl_common.h"
#include "xml_event.h"


cli_res_cmds_t	cli_crypto_cmds = {
	.resinfop = "crypto",
	.add_op =	{ cli_mau_add,	"<number> <ldom>" },
	.set_op =	{ cli_mau_set,	"<number> <ldom>" },
	.remove_op =	{ cli_mau_remove,	"<number> <ldom>" },
};


static int
parse_mau_m_option(char *arg, mau_adm_t **mau_adm)
{
	int n;

	if (!option.allow_dash_p)
		return (-1);

	n = get_comma_sep_list_len(arg);

	*mau_adm = Xmalloc(sizeof (**mau_adm) + n * sizeof (int));
	(*mau_adm)->type = MAU_ADM_IDS;
	(*mau_adm)->num = n;

	if (get_comma_sep_list(arg, (*mau_adm)->ids, n) == -1)
		return (-1);

	return (0);
}

static cli_result_t
mau_adm(int argc, char **argv, cli_conn_t *client, int action)
{
	response_t *resp = client->resp;
	mau_adm_t *mau_adm = NULL;
	ldom_t *ldp;
	uint64_t curr, amt;
	int i;
	int rv = CLI_RESULT_OK;
	xmlChar	*event;

	if (argc == 0)
		return (CLI_RESULT_USAGE);

	if ((ldp = ldm_get_ldp(resp, argv[--argc])) == NULL)
		return (CLI_RESULT_FAIL);

	for (i = 0; (i < argc) && (argv[i][0] == '-') && (rv == CLI_RESULT_OK);
	    i++) {
		if (strcmp(argv[i], "-m") == 0) {
			if (action == ACT_SET) {
				rprintf(resp, gettext("Invalid option\n"));
				rv = CLI_RESULT_USAGE;
				break;
			}
			if (i < (argc - 1)) {
				if (parse_mau_m_option(argv[i+1],
				    &mau_adm) != 0) {
					rv = CLI_RESULT_USAGE;
				}
				i++;
			} else {
				rprintf(resp, gettext("Missing argument to -m "
				    "option\n"));
				rv = CLI_RESULT_USAGE;
			}
		} else if (strcmp(argv[i], "-?") == 0 ||
		    strcmp(argv[i], "--help") == 0) {
			rv = CLI_RESULT_USAGE;
		} else if (strcmp(argv[i], "--") == 0) {
			i++;
			break;
		} else {
			rv = CLI_RESULT_USAGE;
		}
	}

	if (rv != CLI_RESULT_OK)
		goto done;

	if (mau_adm == NULL) {
		/* no -i option, there should be a single number arg left */
		if ((i != argc-1) || !parse_num(resp, &amt, argv[i])) {
			rv = CLI_RESULT_USAGE;
			goto done;
		}
	} else if (i != argc) {
		rv = CLI_RESULT_USAGE;
		goto done;
	}

	switch (action) {
	case ACT_SET:
		event = XML_SET_MAU;
		/*
		 * FIXME: workaround since mau required and bound are not
		 * staying in sync
		 */
		curr = (ldp->state < BOUND) ?
		    ldp->required.num_vmaus : ldom_num_bound_maus(ldp);

		if (add_res(curr, &amt)) {
			/* need to add */
			goto add_resource;
		} else {
			/* need to delete */
			goto remove_resource;
		}
		/*FALLTHROUGH*/

	case ACT_REMOVE:
		event = XML_REMOVE_MAU;
remove_resource:
		if (config_mod_allowed(ldp, DEL_RECONFIG_REMOVE_PHYS,
		    DR_OP_TYPE_CRYPTO, resp) == false) {
			rv = CLI_RESULT_FAIL;
			goto done;
		}

		if (mau_adm == NULL) {
			if (amt == 0)
				/* requested removing nothing, so we're done! */
				goto done;
			mau_adm = Xmalloc(sizeof (*mau_adm));
			mau_adm->type = MAU_ADM_NUM;
			mau_adm->num = amt;
		}

		if (!mau_del(ldp, mau_adm, resp)) {
			if (action == ACT_REMOVE)
				rprintf(resp,
				    gettext("Resource removal failed\n"));
			else
				rprintf(resp,
				    gettext("Resource modification failed\n"));

			rv = CLI_RESULT_FAIL;
			goto done;
		}
		break;

	case ACT_ADD:
		event = XML_ADD_MAU;
add_resource:
		if (config_mod_allowed(ldp, DEL_RECONFIG_ADD,
		    DR_OP_TYPE_CRYPTO, resp) == false) {
			rv = CLI_RESULT_FAIL;
			goto done;
		}

		if (mau_adm == NULL) {
			if (amt == 0)
				/* requested removing nothing, so we're done! */
				goto done;
			mau_adm = Xmalloc(sizeof (*mau_adm));
			mau_adm->type = MAU_ADM_NUM;
			mau_adm->num = amt;
		}

		if (!mau_add(ldp, mau_adm, resp)) {
			if (action == ACT_ADD)
				rprintf(resp,
				    gettext("Resource addition failed\n"));
			else
				rprintf(resp,
				    gettext("Resource modification failed\n"));

			rv = CLI_RESULT_FAIL;
			goto done;
		}
		break;

	default:
		fatal("%s: unknown action %d\n", __func__, action);
	}

done:
	if (rv == CLI_RESULT_OK) {
		event_send_res_event(event, ldp, NULL, NULL, client);
		database_save();
	}
	Xfree(mau_adm);

	return (rv);
}

cli_result_t
cli_mau_add(int argc, char **argv, cli_conn_t *client)
{
	return (mau_adm(argc, argv, client, ACT_ADD));
}

cli_result_t
cli_mau_remove(int argc, char **argv, cli_conn_t *client)
{
	return (mau_adm(argc, argv, client, ACT_REMOVE));
}

cli_result_t
cli_mau_set(int argc, char **argv, cli_conn_t *client)
{
	return (mau_adm(argc, argv, client, ACT_SET));
}
