/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <alloca.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/param.h>

#include "allocate.h"
#include "basics.h"
#include "backend.h"
#include "clp.h"
#include "response.h"
#include "hvctl.h"
#include "ldom.h"
#include "mig_common.h"
#include "sequencer.h"
#include "backend.h"
#include "ds.h"
#include "zprintf.h"
#include "fatal.h"
#include "cli_conn.h"
#include "zpri.h"
#include "core_md.h"
#include "pm/pm_shared.h"
#include "vcpu.h"
#include "frags.h"
#include "vio.h"

#define	CLI_SUCCESS	true
#define	CLI_FAILURE	false

static ldom_t *parse_ldom_arg(char *arg, response_t *resp);

typedef int (*mdstore_f)(char *, char *, response_t *);

/*
 * Frag for migration state MDs.
 */
extern bound_mfrag_t *mig_frag;

static bool_t
match(char *a, char *b, int minlen)
{
	int len;

	len = strlen(a);
	if (len < minlen)
		len = minlen;

	return (strncmp(a, b, len) == 0 ? true : false);
}

static bool_t
do_mdstore(int argc, void **argv, response_t *resp)
{
	mdstore_f mdstore;
	int r;

	if (argc == 0) {
		rprintf(resp, "Usage: ldm debug mdstore dump\n");
		return (CLI_FAILURE);
	}

	if (!match(argv[0], "dump", 1)) {
		rprintf(resp, "Invalid action \"%s\"\n", argv[0]);
		return (CLI_FAILURE);
	}

	mdstore = (mdstore_f)ds_channel_service_routine(ds_sp_channel(),
	    "mdstore", "mdstore");
	if (mdstore == NULL) {
		rprintf(resp, "Could not load mdstore routine\n");
		return (CLI_FAILURE);
	}

	r = (*mdstore)("dump", NULL, resp);

	return (r == 0 ? CLI_SUCCESS : CLI_FAILURE);
}

static bool_t
do_core_md(int argc, void **argv, response_t *resp)
{
	int pid;
	int val;

	if (argc == 0) {
		rprintf(resp, "Usage: ldm debug core_md [get|set] <cpu pid> "
		    "[<value>]\n");
		return (CLI_FAILURE);
	}

	if (match(argv[0], "get", 1)) {
		if (argc != 2) {
			rprintf(resp, "Usage: ldm debug core_md get "
			    "<cpu pid>\n");
			return (CLI_FAILURE);
		}

		pid = atoi(argv[1]);

		if ((val = get_core_md_cpu_status(pid, CORE_MD_PM_STATUS)) <
		    0) {
			rprintf(resp, "Invalid cpu pid %d\n", pid);
			return (CLI_FAILURE);
		}

		rprintf(resp, "PM_STATUS of cpu pid %d is %d\n", pid, val);
		if ((val = get_core_md_cpu_status(pid, CORE_MD_OP_STATUS)) <
		    0) {
			rprintf(resp, "Invalid cpu pid %d\n", pid);
			return (CLI_FAILURE);
		}

		rprintf(resp, "OP_STATUS of cpu pid %d is %d\n", pid, val);
	} else if (match(argv[0], "set", 1)) {
		if (argc != 3) {
			rprintf(resp, "Usage: ldm debug core_md set "
			    "<cpu pid> <value>\n");
			return (CLI_FAILURE);
		}

		pid = atoi(argv[1]);
		val = atoi(argv[2]);

		if (set_core_md_cpu_status(pid, CORE_MD_PM_STATUS, val) < 0) {
			rprintf(resp, "Invalid value %d\n", val);
			return (CLI_FAILURE);
		}
	} else {
		rprintf(resp, "Invalid action \"%s\"\n", argv[0]);
		return (CLI_FAILURE);
	}
	return (CLI_SUCCESS);
}

static bool_t
do_cpu_reconf(int argc, void **argv, response_t *resp)
{
	int rv, status[1], *stats;
	pm_config_op_t op;
	uint32_t pid;

	if (argc != 2) {
		rprintf(resp, "Usage: ldm debug cpu_reconf <cpu pid> in|out\n");
		return (CLI_FAILURE);
	}

	if (strcmp(argv[1], "in") == 0) {
		op = PCO_ADD_RESOURCE;
	} else if (strcmp(argv[1], "out") == 0) {
		op = PCO_REMOVE_RESOURCE;
	} else {
		rprintf(resp, "Invalid op \"%s\"\n", argv[1]);
		return (CLI_FAILURE);
	}

	status[0] = 0;
	stats = &status[0];
	pid = atoi(argv[0]);
	rv = seq_reconf_pcpus(op, 1, &pid, stats, 0);

	rprintf(resp, "seq_reconf_pcpus() returned %d (%d)\n", rv, status[0]);

	return (CLI_SUCCESS);
}

static bool_t
do_echo(int argc, void **argv, response_t *resp)
{
	for (int i = 0; i < argc - 1; i++)
		zprintf("%s ", (char *)argv[i]);
	if (argc > 0)
		zprintf("%s\n", (char *)argv[argc-1]);

	return (CLI_SUCCESS);
}

static bool_t
do_mode(int argc, void **argv, response_t *resp)
{
	if (argc == 0) {
		rprintf(resp, "LDom Manager is in %s mode\n",
		    zeus.mode == ZEUS_CONFIG_MODE ?
		    "configuration" : "management");
	} else if (match(argv[0], "configuration", 1)) {
		if (zeus.mode == ZEUS_CONFIG_MODE) {
			rprintf(resp,
			    "LDom Manager is already in configuration mode\n");
		} else {
			zeus.mode = ZEUS_CONFIG_MODE;
			rprintf(resp,
			    "LDom Manager is now in configuration mode\n");
		}
	} else if (match(argv[0], "management", 1)) {
		if (zeus.mode == ZEUS_MGMT_MODE) {
			rprintf(resp,
			    "LDom Manager is already in management mode\n");
		} else {
			exit(0);
		}
	} else {
		rprintf(resp, "Invalid mode \"%s\"\n", argv[0]);
		rprintf(resp, "Usage: ldm debug mode [config | management]\n");
		return (CLI_FAILURE);
	}

	return (CLI_SUCCESS);
}

static void
do_exit(void)
{
	exit(2);
}

#define	ALIGN		8
#define	ALIGN_MASK	((uint64_t)(ALIGN-1))

static void
hex_dump(uint64_t base, uint8_t *ptr, int len, response_t *resp)
{
	uint64_t addr;
	uint64_t lim;

	lim = base + len;

	for (addr = base & ~ALIGN_MASK; addr < lim; ) {
		char buf[ALIGN+1];
		int ch;
		int i;

		rprintf(resp, "%08"PRIx64" : ", addr);
		buf[ALIGN] = '\0';

		for (i = 0; i < ALIGN; i++) {
			if (addr < base || addr >= lim) {
				rprintf(resp, " ..");
				ch = ' ';
			} else {
				ch = ptr[addr - base];
				rprintf(resp, " %02x", ch);
			}
			buf[i] = (ch < ' ' || ch > 126) ? '.' : ch;
			addr++;
		}
		rprintf(resp, " : %s\n", buf);
	}
}

#undef ALIGN
#undef ALIGN_MASK

#define	CHUNKLEN	(64 * 1024 * 1024)

static bool_t
mapped_mem_read(uint64_t base, uint64_t len, response_t *resp)
{
	void *pa;
	size_t chunklen;
	off_t offset;
	int amt, fd;

	fd = open("/tmp/memdump", O_RDWR|O_CREAT|O_TRUNC, 0600);
	if (fd == -1) {
		rprintf(resp, "Failed to open /tmp/memdump: %s\n",
		    strerror(errno));
		return (false);
	}

	if (ftruncate(fd, len) == -1) {
		rprintf(resp, "Failed to set size of file: %s\n",
		    strerror(errno));
		return (false);
	}

	offset = 0;
	while (len > 0) {
		if (len > CHUNKLEN)
			chunklen = CHUNKLEN;
		else
			chunklen = len;
		pa = mmap(0, chunklen, PROT_WRITE, MAP_SHARED, fd, offset);
		amt = backend.read_mem(base, pa, chunklen);
		if (amt != chunklen)
			rprintf(resp, "Warning: amount read was 0x%"PRIx64"\n",
			    amt);
		munmap(pa, chunklen);
		len -= amt;
		base += amt;
		offset += amt;
	}

	close(fd);
	return (true);
}

extern void generate_size_string(uint64_t bytes, char **str, size_t len);
extern void generate_time_string(uint64_t ns, char **str, size_t len);
extern bool_t do_read_mem_to_file(char *file, uint64_t addr, uint64_t len);

static bool_t
do_rdmem(int argc, void **argv, response_t *resp)
{
	uint64_t base, len;
	uint8_t *bufp;
	int amt;

	if (argc < 2) {
		rprintf(resp, "Usage: ldm debug rdmem <address> <length> "
		    "[file]\n");
		return (false);
	}

	errno = 0;
	base = strtoull((char *)argv[0], NULL, 0);
	if (errno != 0) {
		rprintf(resp, "Base address %s : %s",
		    (char *)argv[0], strerror(errno));
		return (false);
	}
	errno = 0;
	len = strtoull((char *)argv[1], NULL, 0);
	if (errno != 0) {
		rprintf(resp, "Length %s : %s",
		    (char *)argv[0], strerror(errno));
		return (false);
	}

	rprintf(resp, "Reading %"PRId64" bytes from address 0x%"PRIx64"\n",
	    len, base);

	if (argc == 2 && len > 8*1024*1024)
		return (mapped_mem_read(base, len, resp));

	if (argc == 2 && len > 8192) {
		rprintf(resp, "Sorry maximum read size of 8K bytes\n");
		return (false);
	}

	if (argc == 2) {
		bufp = malloc(len);
		amt = backend.read_mem(base, bufp, len);
		if (amt != len)
			rprintf(resp, "Warning: amount read was 0x%"PRIx64"\n",
			    amt);
		hex_dump(base, bufp, amt, resp);
	} else {
		char fname[MAXPATHLEN];

		snprintf(fname, MAXPATHLEN, "/tmp/%s", argv[2]);

		do_read_mem_to_file(fname, base, len);
	}

	return (true);
}

extern bool_t do_write_file_to_addr(char *file, uint64_t addr);

static bool_t
do_wrmem(int argc, void **argv, response_t *resp)
{
	uint64_t base;

	if (argc < 2) {
		rprintf(resp, "Usage: ldm debug wrmem <address> <file>\n");
		return (false);
	}

	/* read in the base address */
	errno = 0;
	base = strtoull((char *)argv[0], NULL, 0);
	if (errno != 0) {
		rprintf(resp, "Base address %s : %s",
		    (char *)argv[0], strerror(errno));
		return (false);
	}

	return (do_write_file_to_addr((char *)argv[1], base));
}

extern void store_resource_state(ldom_t *ldp, state_restype_t restype,
    int resid, uint64_t paddr, uint64_t size);
extern size_t save_resource_state(ldom_t *ldp, state_restype_t rtype,
    int resid);

static bool_t
do_save_ldc_state(ldom_t *ldp)
{
	md_t		*mdp;
	md_node_t	**gnodesp;
	md_node_t	**lnodesp;
	int		nldcs;
	int		nguests;
	uint64_t	id;
	uint64_t	chan;
	uint32_t	gid;
	int		i, j;
	size_t		mdsz;

	gid = ldp->res_id->id;

	/*
	 * Scan the current HV MD looking for all endpoints that
	 * belong to the specified guest.
	 */
	mdp = zeus.hv_mdp;
	gnodesp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));
	lnodesp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	nguests = md_scan_dag(mdp, md_root_node(mdp),
	    md_find_name(mdp, "guest"),
	    md_find_name(mdp, "fwd"), gnodesp);

	/* walk the list of guests in the system */
	for (i = 0; i < nguests; i++) {

		if (md_get_prop_val(mdp, gnodesp[i], "gid", &id) != 0)
			fatal("guest node has no 'gid' property");

		/* check if this is the specified guest */
		if (gid != id)
			continue;

		nldcs = md_scan_dag(mdp, gnodesp[i],
		    md_find_name(mdp, "ldc_endpoint"),
		    md_find_name(mdp, "fwd"), lnodesp);

		/* walk the list of ldcs in this guest */
		for (j = 0; j < nldcs; j++) {
			if (md_get_prop_val(mdp, lnodesp[j], "channel", &chan)
			    != 0) {
				fatal("ldc node has no 'channel' property");
			}

			/*
			 * XXX - is it necessary to check the type of
			 * channel? For example, does it make sense to
			 * migrate SP channels? Does it even make sense
			 * to migrate a domain that has SP channels?
			 */

			/* save the state of this ldc */
			DBG(zprintf("Saving state for ldc %"PRId64":\n", chan));

			mdsz = save_resource_state(ldp, STATE_RES_LDC, chan);
			if (mdsz == -1)
				return (false);

			store_resource_state(ldp, STATE_RES_LDC, chan,
			    mig_frag->paddr, mdsz);

			DBG(zprintf("\n"));
		}
	}

	Xfree(gnodesp);
	Xfree(lnodesp);

	return (true);
}

static bool_t
do_state_save(int argc, void **argv, response_t *resp)
{
	int		i;
	ldom_t		*ldp;
	uint64_t	filesize;
	state_restype_t	restype = STATE_RES_GUEST;
	int		resid = -1;
	vcpu_t		*vcpup;
	bool_t		do_guests = true;
	bool_t		do_vcpus = false;
	bool_t		do_ldcs = false;
	static char	*usage_str =
	    "Usage: ldm debug guest-save [-a ] [-v] [-l] (<ldom> | <ldom id>)";

	if (argc < 1) {
		rprintf(resp, "%s\n", usage_str);
		return (false);
	}

	for (i = 0; i < argc; i++) {

		/* check if the current arg is an option */
		if (*(char *)argv[i] != '-') {
			/* done processing options */
			break;
		}

		if (match(argv[i], "-v", 2)) {
			do_vcpus = true;
			do_guests = false;
			continue;
		}

		if (match(argv[i], "-l", 2)) {
			do_ldcs = true;
			do_guests = false;
			continue;
		}

		if (match(argv[i], "-a", 2)) {
			do_ldcs = true;
			do_vcpus = true;
			do_guests = true;
			continue;
		}

		/* unknown option */
		rprintf(resp, "invalid option '%s'\n%s\n", argv[i], usage_str);
		return (false);
	}

	/* require exactly 1 more argument */
	if (i != (argc - 1)) {
		/* no LDom argument specified */
		rprintf(resp, "%s\n", usage_str);
		return (false);
	}

	/* get the LDom argument */
	if ((ldp = parse_ldom_arg((char *)argv[i], resp)) == NULL)
		return (false);

	/* make sure the guest is active */
	if (ldp->state == INACTIVE) {
		rprintf(resp, "LDom %s is inactive\n", (char *)argv[i]);
		return (false);
	}

	/*
	 * Save the guest state.
	 */
	if (do_guests) {
		resid = ldp->res_id->id;
		restype = STATE_RES_GUEST;

		rprintf(resp, "Saving state for guest '%s':\n", ldp->namep);

		/* save the resource state */
		filesize = save_resource_state(ldp, restype, resid);
		if (filesize == -1)
			return (false);

		/* dump the state to a file */
		store_resource_state(ldp, restype, resid, mig_frag->paddr,
		    filesize);

		rprintf(resp, "\n");
	}

	/*
	 * Save the vcpu state.
	 */
	if (do_vcpus) {
		/* XXX - only support for guests with one vcpu */
		if (ldp->required.num_vcpus != 1) {
			rprintf(resp, "Operation only supported for domains "
			    "with exactly one vcpu\n");
			return (false);
		}

		vcpup = CHAIN_START(ldp->bound.vcpus);
		if (vcpup == NULL) {
			rprintf(resp, "LDom %s does not have any bound "
			    "vcpus\n", ldp->namep);
			return (false);
		}

		resid = vcpup->vid;
		restype = STATE_RES_VCPU;

		rprintf(resp, "Saving state for vcpu %d:\n", resid);

		/* save the resource state */
		filesize = save_resource_state(ldp, restype, resid);
		if (filesize == -1)
			return (false);

		/* dump the state to a file */
		store_resource_state(ldp, restype, resid, mig_frag->paddr,
		    filesize);

		rprintf(resp, "\n");
	}

	/*
	 * Save the ldc state.
	 */
	if (do_ldcs) {
		return (do_save_ldc_state(ldp));
	}

	return (true);
}

extern int call_hv_state_restore(ldom_t *ldp, uint64_t paddr,
    state_restype_t restype, uint32_t resid);

static bool_t
do_state_restore(int argc, void **argv, response_t *resp)
{
	ldom_t		*ldp;
	char		*end;
	int		cc;
	char		*file;
	uint64_t	filesize;
	int		fh;
	struct stat	status;
	int		i;
	state_restype_t restype = STATE_RES_GUEST;
	int		resid = -1;
	vcpu_t		*vcpup;
	static char	*usage_str =
	    "Usage: ldm debug guest-restore [-v | (-l <ldc id>)] "
	    "<restore-md-file> (<ldom> | <ldom id>)";

	if (argc < 2) {
		rprintf(resp, "%s\n", usage_str);
		return (false);
	}

	for (i = 0; i < argc; i++) {

		/* check if the current arg is an option */
		if (*(char *)argv[i] != '-') {
			/* done processing options */
			break;
		}

		if (match(argv[i], "-v", 2)) {
			restype = STATE_RES_VCPU;
			continue;
		}

		if (match(argv[i], "-l", 2)) {
			/* user wants to get the state of an ldc */
			if (++i == argc) {
				rprintf(resp, "-l option requires an ldc id "
				    "argument\n%s\n", usage_str);
				return (false);
			}

			if (restype != STATE_RES_GUEST) {
				rprintf(resp, "The '%s' option has already "
				    "been specified\n%s\n",
				    (restype == STATE_RES_VCPU) ?
				    "-v" : "-l", usage_str);

				return (false);
			}

			restype = STATE_RES_LDC;

			/* get the ldc id */
			resid = strtoll((char *)argv[i], &end, 0);
			if (end == NULL || *end != '\0') {
				rprintf(resp, "invalid ldc id '%s'\n%s\n",
				    argv[i], usage_str);
				return (false);
			}

			/* XXX - check that the ldc id is valid? */

			rprintf(resp, "Restoring state for ldc %d:\n", resid);

		} else {
			rprintf(resp, "invalid option '%s'\n%s\n",
			    argv[i], usage_str);
			return (false);
		}
	}

	/* require exactly 2 more arguments */
	if (i != (argc - 2)) {
		/* no MD file and LDom arguments specified */
		rprintf(resp, "%s\n", usage_str);
		return (false);
	}

	/* get the LDom argument */
	if ((ldp = parse_ldom_arg((char *)argv[i + 1], resp)) == NULL)
		return (false);

	if (restype == STATE_RES_GUEST) {
		resid = ldp->res_id->id;
		rprintf(resp, "Restoring state for guest '%s':\n", ldp->namep);
	}

	if (restype == STATE_RES_VCPU) {
		/* XXX - only support for guests with one vcpu */
		if (ldp->required.num_vcpus != 1) {
			rprintf(resp, "Operation only supported for domains "
			    "with exactly one vcpu\n");
			return (false);
		}

		vcpup = CHAIN_START(ldp->bound.vcpus);
		if (vcpup == NULL) {
			rprintf(resp, "LDom %s does not have any bound vcpus\n",
			    ldp->namep);
			return (false);
		}

		resid = vcpup->vid;
		rprintf(resp, "Restoring state for vcpu %d:\n", resid);
	}

	/*
	 * Allocate a frag and write the MD file to that location
	 * in memory.
	 */
	file = (char *)argv[i];
	fh = open(file, O_RDONLY);
	if (fh < 0) {
		rprintf(resp, "Opening %s : %s\n", file, strerror(errno));
		return (false);
	}

	if (fstat(fh, &status) < 0) {
		rprintf(resp, "Stat'ing %s : %s\n", file, strerror(errno));
		return (false);
	}
	filesize = status.st_size;
	close(fh);

	/* free any old frag before we start */
	if (mig_frag != NULL)
		frag_free(mig_frag);

	mig_frag = frag_alloc(ldp, filesize);
	if (mig_frag == NULL) {
		rprintf(resp, "Unable to allocate frag for state MD\n");
		return (false);
	}

	if (do_write_file_to_addr(file, mig_frag->paddr) == false)
		return (false);

	/* issue the HV call to restore the state */
	cc = call_hv_state_restore(ldp, mig_frag->paddr, restype, resid);
	if (cc != HVctl_st_ok) {
		rprintf(resp, "HVctl command failed: %d\n", cc);
		return (false);
	}

	return (true);
}

/*
 * Get the resource ID of a domain from a string supplied
 *
 * Returns:
 *	resid >= 0	- valid resource ID
 *	resid == -1	- error
 */
static int
get_resid(char *cmdlnstr, response_t *resp)
{
	ldom_t *ldp;
	char	*end;
	int	resid;

	/*
	 * FIXME: no verification of resid provided on command line is done,
	 * just when domain name is provided. Unimportant on a debug CLI...
	 */
	resid = strtoll(cmdlnstr, &end, 0);
	if (end == NULL || *end != '\0') {
		ldp = find_ldom_by_name(cmdlnstr);
		if (ldp == NULL) {
			rprintf(resp, "LDom %s not found\n", cmdlnstr);
			return (-1);
		}
		if (ldp->state == INACTIVE) {
			rprintf(resp, "LDom %s is inactive\n", cmdlnstr);
			return (-1);
		}
		if (ldp->res_id == NULL) {
			rprintf(resp, "LDom %s has no resource id\n", cmdlnstr);
			return (-1);
		}

		resid = ldp->res_id->id;
	}

	return (resid);
}

extern int call_hv_get_res_stat(uint32_t, uint32_t, uint32_t,
    hvctl_msg_t **stat);

static bool_t
do_state(hvctl_guest_info_t infoid, int argc, void **argv, response_t *resp)
{
	hvctl_msg_t *stat;
	int cc, resid;

	if (argc != 1) {
		rprintf(resp, "Missing LDom name or id\n");
		rprintf(resp, "Usage: ldm debug "
		    "[state | softstate | tod | apivers] "
		    "(<ldom> | <ldom id>)\n");
		return (false);
	}

	resid = get_resid((char *)argv[0], resp);
	if (resid == -1)
		return (false);

	cc = call_hv_get_res_stat(HVctl_res_guest, resid, infoid, &stat);

	if (cc != HVctl_st_ok) {
		rprintf(resp, "HVctl command failed (error=%d)\n", cc);
		return (false);
	}

	switch (infoid) {
	case HVctl_info_guest_soft_state:
		rprintf(resp, "guest %d soft state: 0x%x\n", resid,
		    ((rs_guest_soft_state_t *)stat->msg.resstat.data)->
		    soft_state);
		rprintf(resp, "guest %d soft state str: %s\n", resid,
		    ((rs_guest_soft_state_t *)stat->msg.resstat.data)->
		    soft_state_str);
		break;
	case HVctl_info_guest_state: {
		uint64_t state;
		state = ((rs_guest_state_t *)stat->msg.resstat.data)->state;
		rprintf(resp, "guest %d state: %s (0x%"PRIx64")\n", resid,
		    hv_state_str(state), state);
		break;
	}
	case HVctl_info_guest_tod:
		rprintf(resp, "guest %d tod: 0x%"PRIx64"\n", resid,
		    ((rs_guest_tod_t *)stat->msg.resstat.data)->tod);
		break;
	default:
		break;
	}

	return (true);
}


static bool_t
get_api_groups(int argc, void **argv, response_t *resp)
{
	hv_api_group_version_t	*versionp = NULL;
	uint32_t	num_api_groups;
	int		rv, resid;

	if (argc != 1) {
		rprintf(resp, "Missing LDom name or id\n");
		rprintf(resp, "Usage: ldm debug apivers (<ldom>|<ldom id>)\n");
		return (false);
	}

	resid = get_resid((char *)argv[0], resp);
	if (resid == -1)
		return (false);

	/* request the API group versions from the HV */
	num_api_groups = call_hv_get_api_groups(resid, &versionp);
	rprintf(resp, "Got API versions for %d groups\n", num_api_groups);

	if (num_api_groups == 0 || versionp == NULL)
		return (false);

	DUMP_API_GROUPS(num_api_groups, versionp);

	/* reflect back to HV and get the HV to check that they are valid */
	rv = call_hv_query_api_groups(resid, num_api_groups, &versionp);
	rprintf(resp, "Reflecting API Groups to HV returned %d\n", rv);

	/* Modify table, send to HV and check that it is invalid */
	versionp->version.minor_ver = 9;
	rv = call_hv_query_api_groups(resid, num_api_groups, &versionp);
	rprintf(resp, "Modified API Group sent to HV returned %d\n", rv);

	Xfree(versionp);

	return (true);
}

static int
get_cpu_pid(resid_t *resid)
{
	pri_cpu_t *pcpup;

	CHAIN_ITER(pri.cpu_list, pcpup, link) {
		if (pcpup->res_id != NULL &&
		    pcpup->res_id->id == resid->id) {
			return (pcpup->phys_id);
		}
	}

	return (-1);
}

/* from hypervisor.h */
#define	CPU_STATE_INVALID	0x0
#define	CPU_STATE_STOPPED	0x1	/* cpu not started */
#define	CPU_STATE_RUNNING	0x2	/* cpu running guest code */
#define	CPU_STATE_ERROR		0x3	/* cpu is in the error state */
#define	CPU_STATE_SUSPENDED	0x4 	/* cpu in suspend loop */
#define	CPU_STATE_UNCONFIGURED	0x5 	/* cpu unconfigured from guest */

static void
one_vcpu_state(int id, bool_t util, response_t *resp)
{
	rs_vcpu_state_t vstate;
	char *s;
	resid_t resid;

	resid.id = id;

	if (backend.get_vcpu_state(&resid, &vstate) == 0) {
		switch (vstate.state) {
		case CPU_STATE_INVALID:		s = "INVALID"; break;
		case CPU_STATE_STOPPED:		s = "STOPPED"; break;
		case CPU_STATE_RUNNING:		s = "RUNNING"; break;
		case CPU_STATE_ERROR:		s = "ERROR"; break;
		case CPU_STATE_SUSPENDED:	s = "SUSPENDED"; break;
		case CPU_STATE_UNCONFIGURED:	s = "UNCONFIGURED"; break;
		default:			s = "UNKNOWN"; break;
		}

		rprintf(resp, "cpu resid %2d pid %2d: %s\n", id,
		    get_cpu_pid(&resid), s);
		if (util) {
			rprintf(resp, "  lifespan     %"PRIu64"\n",
			    vstate.lifespan);
			rprintf(resp, "  wall delta   %"PRIu64"\n",
			    vstate.wallclock_delta);
			rprintf(resp, "  active delta %"PRIu64"\n",
			    vstate.active_delta);
			rprintf(resp, "  yielded      %"PRIu64"\n",
			    vstate.yielded_cycles);
		}
	} else {
		rprintf(resp, "cpu resid %2d:  error getting state\n", id);
	}
}

static bool_t
do_vcpu_state(int argc, void **argv, response_t *resp)
{
	int i, id;
	bool_t util = false;

	if (argc > 0 && strcmp(argv[0], "-u") == 0) {
		util = true;
		argc--;
		argv++;
	}

	if (argc == 0) {
		for (i = 0; i < 32; i++)
			one_vcpu_state(i, util, resp);
	} else {
		for (i = 0; i < argc; i++) {
			errno = 0;
			id = strtoull((char *)(argv[i]), NULL, 0);
			if (errno == 0)
				one_vcpu_state(id, util, resp);
		}
	}

	return (true);
}

extern hvctl_msg_t *get_hvconfig(void);

static int
do_hvconfig(response_t *resp)
{
	hvctl_msg_t *rp;

	if (strcmp(backend.namep, "ldom") == 0) {
		rp = get_hvconfig();

		rprintf(resp, "mem base      0x%"PRIx64"\n",
		    ntoh64(rp->msg.hvcnf.hv_membase));
		rprintf(resp, "mem size      0x%"PRIx64"\n",
		    ntoh64(rp->msg.hvcnf.hv_memsize));
		rprintf(resp, "hvmd pa       0x%"PRIx64"\n",
		    ntoh64(rp->msg.hvcnf.hvmdp));
		rprintf(resp, "delreconf pa  0x%"PRIx64"\n",
		    ntoh64(rp->msg.hvcnf.del_reconf_hvmdp));
		rprintf(resp, "delreconf gid %d\n",
		    (int)ntoh32(rp->msg.hvcnf.del_reconf_gid));

		Xfree(rp);
	}

	return (0);
}

extern int cli_set_cpu_states(int count, uint32_t *cpuids,
	int dir, int do_rcfg);
extern int cli_get_cpu_states(int count, uint32_t *cpuids);
extern int cli_get_power_all(void);

/*
 * Power Management debug
 */
static bool_t
do_pmi_get_all(response_t *resp)
{
	bool_t	ret;

	rprintf(resp, "Update and dump all resources' power states:\n");
	ret = (cli_get_power_all() == 0) ? true : false;
	rprintf(resp, "    %s\n", (ret ? "Successful" : "Failed"));

	return (ret);
}

static bool_t
do_pm_state(hvctl_pm_state_dir_t dir, int argc, void **argv, response_t *resp)
{
	int	res_count, i;
	uint32_t	*cpuids = NULL;
	uint16_t	new_state;
	bool_t	ret = true;

	if (argc < 2)
		goto do_syntax;

	res_count = atoi((char *)argv[0]);
	if (argc < (res_count + 1))
		goto do_syntax;

	argv++;
	cpuids = Xmalloc(res_count * sizeof (uint32_t));
	for (i = 0; i < res_count; i++)
		cpuids[i] = atoi((char *)argv[i]);

	switch (dir) {
	case HVctl_pm_state_set:
		if (argc != (res_count + 2))
			goto do_syntax;

		/*
		 * Valid new_states ought to be 0 for parking the strand,
		 * 1 for unpark the strand
		 */
		new_state = atoi((char *)argv[i]);
		if ((new_state != 0) && (new_state != 1)) {
			rprintf(resp, "Invalid new_state %d\n", new_state);
			ret = false;
			goto done;
		} else
			rprintf(resp, "Set %d strands to new power state: %d\n",
			    res_count, new_state);

		ret = cli_set_cpu_states(res_count, cpuids, new_state, 0);
		if (ret == 0) {
			rprintf(resp, "Successful\n");
			ret = true;
		} else {
			rprintf(resp, "Failed, cc(%d)\n", ret);
			ret = false;
		}
		break;

	case HVctl_pm_state_get:
		rprintf(resp, "Get %d strands' current power state: \n",
		    res_count);
		if (cli_get_cpu_states(res_count, cpuids) == 0) {
			rprintf(resp, "Successful, see log for current "
			    "states\n");
			ret = true;
		} else {
			rprintf(resp, "Failed\n");
			ret = false;
		}
		break;

	default:
		goto do_syntax;
	}

	goto done;

do_syntax:
	rprintf(resp, "Usage:\n\t");
	rprintf(resp, "ldm debug pm_set_state <num_of_cpus> <pcpu-id lists> "
	    "<new_state_index> \nOR\n\t"
	    "ldm debug pm_get_state <num_of_cpus> <pcpu-id-list>\n");
	ret = false;
done:
	if (cpuids)
		Xfree(cpuids);
	return (ret);
}

static bool_t
do_lock(bool_t lock, int argc, void **argv, response_t *resp)
{
	ldom_t *ldp;
	char *end;
	uint64_t gid;

	if (argc != 1) {
		rprintf(resp, "Missing LDom name or id\n");
		rprintf(resp, "Usage: ldm debug [lock] (<ldom> | <ldom id>)\n");
		return (false);
	}

	gid = strtoull((char *)argv[0], &end, 0);
	if (end == NULL || *end != '\0') {
		ldp = find_ldom_by_name((char *)argv[0]);
		if (ldp == NULL) {
			rprintf(resp, "LDom %s not found\n", (char *)argv[0]);
			return (false);
		}
	} else {
		ldp = find_ldom_by_gid(gid);
		if (ldp == NULL) {
			rprintf(resp, "LDom id %" PRId64 " not found\n", gid);
			return (false);
		}
	}

	ldp->flag.locked = lock;

	return (true);
}

static void
resource_usage(ldom_t *ldp, response_t *resp)
{
	rprintf(resp, "LDom %s\n", ldp->namep);
	rprintf(resp, "LDom %s\n", ldp->namep);
	rprintf(resp, "  num clients\t\t  %d\n", ldp->bound.num_clients);
	rprintf(resp, "  num services\t\t  %d\n", ldp->bound.num_services);
	rprintf(resp, "  io devices\t\t  %d\n",
	    CHAIN_LENGTH(ldp->bound.iodevices));
	rprintf(resp, "  used ldc ids            %"PRIu64"\n",
	    used_res_ids(&ldp->guest_free_ldcs));
	rprintf(resp, "  used ds ports           %"PRIu64"\n",
	    used_res_ids(&ldp->free_ds_ports));
#if VDEV_CFG_HANDLE
	rprintf(resp, "  used cfg handles        %"PRIu64"\n",
	    used_res_ids(&ldp->free_cfg_handles));
#else
	if (CHAIN_LENGTH(ldp->vccs) > 0)
		rprintf(resp, "  used vcc cfg handles    %"PRIu64"\n",
		    used_res_ids(&ldp->bound.vcc_cfg_handles));
	if (CHAIN_LENGTH(ldp->vldcs) > 0)
		rprintf(resp, "  used vldc cfg handles   %"PRIu64"\n",
		    used_res_ids(&ldp->bound.vldc_cfg_handles));
	if (CHAIN_LENGTH(ldp->vldccs) > 0)
		rprintf(resp, "  used vldcc cfg handles  %"PRIu64"\n",
		    used_res_ids(&ldp->bound.vldcc_cfg_handles));
	if (CHAIN_LENGTH(ldp->vsws) > 0)
		rprintf(resp, "  used vsw cfg handles    %"PRIu64"\n",
		    used_res_ids(&ldp->bound.vsw_cfg_handles));
	if (CHAIN_LENGTH(ldp->vnets) > 0)
		rprintf(resp, "  used vnet cfg handles   %"PRIu64"\n",
		    used_res_ids(&ldp->bound.vnet_cfg_handles));
	if (CHAIN_LENGTH(ldp->vdss) > 0)
		rprintf(resp, "  used vds cfg handles    %"PRIu64"\n",
		    used_res_ids(&ldp->bound.vds_cfg_handles));
	if (CHAIN_LENGTH(ldp->vdisks) > 0)
		rprintf(resp, "  used vdisk cfg handles  %"PRIu64"\n",
		    used_res_ids(&ldp->bound.vdisk_cfg_handles));
	if (CHAIN_LENGTH(ldp->vdpcss) > 0)
		rprintf(resp, "  used vdpcs cfg handles  %"PRIu64"\n",
		    used_res_ids(&ldp->bound.vdpcs_cfg_handles));
	if (CHAIN_LENGTH(ldp->vdpccs) > 0)
		rprintf(resp, "  used vpcc cfg handles   %"PRIu64"\n",
		    used_res_ids(&ldp->bound.vdpcc_cfg_handles));
#endif
}

static bool_t
do_resource_usage(int argc, void **argv, response_t *resp)
{
	ldom_t *ldp;
	char *end;
	uint64_t gid;

	if (argc < 1) {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			resource_usage(ldp, resp);
		}

		return (true);
	}

	for (int i = 0; i < argc; i++) {
		gid = strtoull((char *)argv[i], &end, 0);
		if (end == NULL || *end != '\0') {
			ldp = find_ldom_by_name((char *)argv[i]);
			if (ldp == NULL) {
				rprintf(resp, "LDom %s not found\n",
				    (char *)argv[i]);
				continue;
			}
		} else {
			ldp = find_ldom_by_gid(gid);
			if (ldp == NULL) {
				rprintf(resp, "LDom id %" PRId64
				    " not found\n", gid);
				continue;
			}
		}

		resource_usage(ldp, resp);
	}

	return (true);
}

static uint64_t inject_flags;
static mutex_t inject_lock = DEFAULTMUTEX;

int
inject_error(uint64_t error)
{
	int hit;

	mutex_lock(&inject_lock);
	if (inject_flags & error) {
		hit = 1;
		/* clear the injection state */
		inject_flags &= ~error;
	} else {
		hit = 0;
	}
	mutex_unlock(&inject_lock);

	return (hit);
}

static bool_t
do_inject(int argc, void **argv, response_t *resp)
{
	uint64_t flags;

	if (argc < 1) {
		rprintf(resp, "Usage: ldm debug inject <flags>\n");
		return (false);
	}

	errno = 0;
	flags = strtoull((char *)argv[0], NULL, 0);
	if (errno != 0)
		return (false);

	mutex_lock(&inject_lock);
	inject_flags = flags;
	mutex_unlock(&inject_lock);

	return (true);
}

extern int call_hv_guest_suspend(ldom_t *ldp);
extern int call_hv_guest_resume(ldom_t *ldp);

static bool_t
do_suspend(int argc, void **argv, response_t *resp)
{
	int		i;
	ldom_t		*ldp;
	char		*end;
	uint64_t	gid;
	int		cc;
	bool_t		do_isolate = true;
	static char	*usage_str =
	    "Usage: ldm debug suspend [-n] (<ldom> | <ldom id>)";

	if (argc == 0) {
		rprintf(resp, "%s\n", usage_str);
		return (false);
	}

	for (i = 0; i < argc; i++) {

		/* check if the current arg is an option */
		if (*(char *)argv[i] != '-') {
			/* done processing options */
			break;
		}

		if (match(argv[i], "-n", 2)) {
			/* user wants to skip the suspend */
			do_isolate = false;

		} else {
			rprintf(resp, "invalid option '%s'\n%s\n",
			    argv[i], usage_str);
			return (false);
		}
	}

	if (i == argc) {
		/* no LDom argument specified */
		rprintf(resp, "%s\n", usage_str);
		return (false);
	}

	/* get the ID for the specified guest */
	gid = strtoll((char *)argv[i], &end, 0);
	if (end == NULL || *end != '\0') {
		ldp = find_ldom_by_name((char *)argv[i]);
		if (ldp == NULL) {
			rprintf(resp, "LDom %s not found\n", (char *)argv[i]);
			return (false);
		}
	} else {
		ldp = find_ldom_by_gid(gid);
		if (ldp == NULL) {
			rprintf(resp, "LDom id %" PRId64 " not found\n", gid);
			return (false);
		}
	}

	/* It only makes sense to suspend an active domain */
	if (ldp->state != ACTIVE) {
		rprintf(resp, "LDom %s is %s\n", ldp->namep,
		    ldom_state_str(ldp->state));
		return (false);
	}

	ldp->state = SUSPENDING;

	/* XXX - only support for guests with one vcpu */
	if (ldp->required.num_vcpus != 1) {
		rprintf(resp, "Operation only supported for domains with "
		    "exactly one vcpu\n");
		return (false);
	}

	if (do_isolate) {
		rprintf(resp, "isolating LDom %s...\n", ldp->namep);
		seq_isolate_ldom(ldp, resp);
	} else {
		rprintf(resp, "skipping isolate phase...\n");
	}

	rprintf(resp, "suspending LDom %s...\n", ldp->namep);
	cc = call_hv_guest_suspend(ldp);

	if (cc != HVctl_st_ok) {
		rprintf(resp, "HVctl command failed\n");
		return (false);
	}

	ldp->state = SUSPENDED;

	return (true);
}

static ldom_t *
parse_ldom_arg(char *arg, response_t *resp)
{
	ldom_t		*ldp;
	uint64_t	gid;
	char		*end;

	/* get the ID for the specified guest */
	gid = strtoll(arg, &end, 0);
	if (end == NULL || *end != '\0') {
		ldp = find_ldom_by_name(arg);
		if (ldp == NULL) {
			rprintf(resp, "LDom %s not found\n", arg);
			return (NULL);
		}
	} else {
		ldp = find_ldom_by_gid(gid);
		if (ldp == NULL) {
			rprintf(resp, "LDom %s not found\n", arg);
			return (NULL);
		}
	}

	return (ldp);
}

static bool_t
do_isolate(int argc, void **argv, response_t *resp)
{
	ldom_t		*ldp;
	md_t		*mdp;
	static char	*usage_str =
	    "Usage: ldm debug isolate (<ldom> | <ldom id>)";

	if (argc == 0) {
		rprintf(resp, "%s\n", usage_str);
		return (false);
	}

	/* get the LDom argument */
	if ((ldp = parse_ldom_arg((char *)argv[0], resp)) == NULL)
		return (false);

	if (ldp->state == INACTIVE) {
		rprintf(resp, "LDom %s is inactive\n", ldp->namep);
		return (false);
	}

	/* XXX - only support for guests with one vcpu */
	if (ldp->required.num_vcpus != 1) {
		rprintf(resp, "Operation only supported for domains with "
		    "exactly one vcpu\n");
		return (false);
	}

	/* generate a migration binding guest MD */
	mdp = backend.get_guest_md(ldp);

	/* patch up any power managed CPUs */
	(void) cpu_mig_add_pm_bindings(ldp, mdp);

	/* augment the MD with extra VIO binding info */
	(void) vio_mig_add_bindings(ldp, mdp);

	if (ldp->mig_state.bind_mdp != NULL)
		md_fini(ldp->mig_state.bind_mdp);
	ldp->mig_state.bind_mdp = mdp;

	rprintf(resp, "isolating LDom %s...\n", ldp->namep);
	seq_isolate_ldom(ldp, resp);

	return (true);
}

static bool_t
do_restore(int argc, void **argv, response_t *resp)
{
	ldom_t		*ldp;
	static char	*usage_str =
	    "Usage: ldm debug restore (<ldom> | <ldom id>)";

	if (argc == 0) {
		rprintf(resp, "%s\n", usage_str);
		return (false);
	}

	/* get the LDom argument */
	if ((ldp = parse_ldom_arg((char *)argv[0], resp)) == NULL)
		return (false);

	if (ldp->state == INACTIVE) {
		rprintf(resp, "LDom %s is inactive\n", ldp->namep);
		return (false);
	}

	/* XXX - only support for guests with one vcpu */
	if (ldp->required.num_vcpus != 1) {
		rprintf(resp, "Operation only supported for domains with "
		    "exactly one vcpu\n");
		return (false);
	}

	rprintf(resp, "restoring LDom %s...\n", ldp->namep);
	seq_restore_ldom(ldp, resp);

	return (true);
}

static bool_t
do_resume(int argc, void **argv, response_t *resp)
{
	int		i;
	ldom_t		*ldp;
	int		cc;
	vcpu_t		*vcpup;
	bool_t		do_restore = true;
	static char	*usage_str =
	    "Usage: ldm debug resume [-n] (<ldom> | <ldom id>)";

	if (argc == 0) {
		rprintf(resp, "%s\n", usage_str);
		return (false);
	}

	for (i = 0; i < argc; i++) {

		/* check if the current arg is an option */
		if (*(char *)argv[i] != '-') {
			/* done processing options */
			break;
		}

		if (match(argv[i], "-n", 2)) {
			/* user wants to skip the resume */
			do_restore = false;
		} else {
			rprintf(resp, "invalid option '%s'\n%s\n",
			    argv[i], usage_str);
			return (false);
		}
	}

	if (i == argc) {
		/* no LDom argument specified */
		rprintf(resp, "%s\n", usage_str);
		return (false);
	}

	/* get the LDom argument */
	if ((ldp = parse_ldom_arg((char *)argv[i], resp)) == NULL)
		return (false);

	if (ldp->state == INACTIVE) {
		rprintf(resp, "LDom %s is inactive\n", ldp->namep);
		return (false);
	}

	/* XXX - only support for guests with one vcpu */
	if (ldp->required.num_vcpus != 1) {
		rprintf(resp, "Operation only supported for domains with "
		    "exactly one vcpu\n");
		return (false);
	}

	vcpup = CHAIN_START(ldp->bound.vcpus);
	if (vcpup == NULL) {
		rprintf(resp, "LDom %s does not have any bound vcpus\n",
		    ldp->namep);
		return (false);
	}

	rprintf(resp, "resuming LDom %s...\n", ldp->namep);
	cc = call_hv_guest_resume(ldp);

	if (cc != HVctl_st_ok) {
		rprintf(resp, "HVctl command failed\n");
		return (false);
	}

	ldp->state = ACTIVE;

	/* check if the user wants to skip the restore */
	if (!do_restore) {
		rprintf(resp, "skipping restore phase...\n");
		return (true);
	}

	rprintf(resp, "restoring LDom %s...\n", ldp->namep);
	seq_restore_ldom(ldp, resp);

	return (true);
}

static bool_t
do_dumpmem(int argc, void **argv, response_t *resp)
{
	ldom_t		*ldp;
	mblock_t	*mbp;
	char		fname[MAXPATHLEN];

	if (argc != 1) {
		rprintf(resp, "Usage: ldm debug dumpmem (<ldom> | "
		    "<ldom id>)\n");
		return (false);
	}

	/* get the LDom argument */
	if ((ldp = parse_ldom_arg((char *)argv[0], resp)) == NULL)
		return (false);

	if (ldp->state == INACTIVE) {
		rprintf(resp, "LDom %s is inactive\n", ldp->namep);
		return (false);
	}

	/*
	 * Walk the list of mblocks. For each mblock, read
	 * its contents and save it out to disk.
	 */
	CHAIN_ITER(ldp->bound.memory, mbp, ldom) {

		rprintf(resp, "Saving mblock @ 0x%"PRIx64" (%"PRId64" bytes)\n",
		    mbp->paddr, mbp->size);

		snprintf(fname, MAXPATHLEN, "/export/%s-mem-0x%"PRIx64".bin",
		    ldp->namep, mbp->paddr);

		do_read_mem_to_file(fname, mbp->paddr, mbp->size);
	}

	return (true);
}

static void
print_ldc_endpoint(md_t *mdp, md_node_t *node, response_t *resp)
{
	uint64_t chan;
	uint64_t type;
	uint64_t tgt;
	uint64_t tchan;

	if (md_get_prop_val(mdp, node, "channel", &chan) != 0) {
		fatal("ldc node has no 'channel' property");
	}

	rprintf(resp, "  ldc channel %2"PRId64" ->", chan);

	if (md_get_prop_val(mdp, node, "target_type", &type) != 0) {
		fatal("ldc node has no 'target_type' property");
	}

	switch (type) {
	case TYPE_GUEST:
		if (md_get_prop_val(mdp, node, "target_guest", &tgt) != 0) {
			fatal("ldc node has no 'target_guest' property");
		}

		if (md_get_prop_val(mdp, node, "target_channel", &tchan) != 0) {
			fatal("ldc node has no 'target_channel' property");
		}

		rprintf(resp, " target: gid=%"PRId64", chan=%"PRId64"\n",
		    tgt, tchan);
		break;

	case TYPE_HV:
		rprintf(resp, " target: HV\n");
		break;

	case TYPE_SP:
		rprintf(resp, " target: SP\n");
		break;

	default:
		rprintf(resp, " target: UNKNOWN\n");
		break;
	}
}

static bool_t
do_ldc_usage(int argc, void **argv, response_t *resp)
{
	int		i, j;
	ldom_t		*ldp;
	uint64_t	gid;
	md_t		*mdp;
	md_node_t	**gnodesp;
	md_node_t	**lnodesp;
	uint64_t	id;
	int		nldcs;
	int		nguests;

	if (argc != 1) {
		rprintf(resp, "Usage: ldm debug ldc-usage (<ldom> | "
		    "<ldom id>)\n");
		return (false);
	}

	/* get the LDom argument */
	if ((ldp = parse_ldom_arg((char *)argv[0], resp)) == NULL)
		return (false);

	gid = ldp->res_id->id;

	/*
	 * Scan the current HV MD and print out the LDC usage.
	 */
	mdp = zeus.hv_mdp;
	gnodesp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));
	lnodesp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	nguests = md_scan_dag(mdp, md_root_node(mdp),
	    md_find_name(mdp, "guest"),
	    md_find_name(mdp, "fwd"), gnodesp);

	for (i = 0; i < nguests; i++) {
		if (md_get_prop_val(mdp, gnodesp[i], "gid", &id) != 0)
			fatal("guest node has no 'gid' property");

		if (gid != id)
			continue;

		rprintf(resp, "LDom '%s'\n", ldp->namep);

		nldcs = md_scan_dag(mdp, gnodesp[i],
		    md_find_name(mdp, "ldc_endpoint"),
		    md_find_name(mdp, "fwd"), lnodesp);

		for (j = 0; j < nldcs; j++) {
			print_ldc_endpoint(mdp, lnodesp[j], resp);
		}
		rprintf(resp, "\n");
	}

	Xfree(gnodesp);
	Xfree(lnodesp);

	/* XXX - hack vio usage in here for now */
	vio_print_bindings(ldp, resp);

	return (true);
}

bool_t
ldm_debug(int argc, void **argv, cli_conn_t *client)
{
	response_t *resp = client->resp;

	if (argc >= 1) {
		if (match((char *)argv[0], "apivers", 2))
			return (get_api_groups(argc-1, argv+1, resp));
		if (match((char *)argv[0], "echo", 2))
			return (do_echo(argc-1, argv+1, resp));
		if (match((char *)argv[0], "mode", 2))
			return (do_mode(argc-1, argv+1, resp));
		if (match((char *)argv[0], "mdstore", 2))
			return (do_mdstore(argc-1, argv+1, resp));
		if (match((char *)argv[0], "rdmem", 2))
			return (do_rdmem(argc-1, argv+1, resp));
		if (match((char *)argv[0], "wrmem", 2))
			return (do_wrmem(argc-1, argv+1, resp));
		if (match((char *)argv[0], "exit", 4))
			do_exit();
		if (match((char *)argv[0], "softstate", 2))
			return (do_state(HVctl_info_guest_soft_state,
			    argc-1, argv+1, resp));
		if (match((char *)argv[0], "state", 2))
			return (do_state(HVctl_info_guest_state,
			    argc-1, argv+1, resp));
		if (match((char *)argv[0], "cpu", 2))
			return (do_vcpu_state(argc-1, argv+1, resp));
		if (match((char *)argv[0], "tod", 2))
			return (do_state(HVctl_info_guest_tod,
			    argc-1, argv+1, resp));
		if (match((char *)argv[0], "inval", 2))
			return (do_state(HVctl_info_guest_max,
			    argc-1, argv+1, resp));
		if (match((char *)argv[0], "hvconfig", 3))
			return (do_hvconfig(resp));
		if (match((char *)argv[0], "pmi_get_all", 9))
			return (do_pmi_get_all(resp));
		if (match((char *)argv[0], "pm_get_state", 6))
			return (do_pm_state(HVctl_pm_state_get,
			    argc-1, argv+1, resp));
		if (match((char *)argv[0], "pm_set_state", 6))
			return (do_pm_state(HVctl_pm_state_set,
			    argc-1, argv+1, resp));
		if (match((char *)argv[0], "lock", 2))
			return (do_lock(true, argc-1, argv+1, resp));
		if (match((char *)argv[0], "unlock", 2))
			return (do_lock(false, argc-1, argv+1, resp));
		if (match((char *)argv[0], "res-usage", 2))
			return (do_resource_usage(argc-1, argv+1, resp));
		if (match((char *)argv[0], "inject", 3))
			return (do_inject(argc-1, argv+1, resp));
		if (match((char *)argv[0], "core_md", 4))
			return (do_core_md(argc-1, argv+1, resp));
		if (match((char *)argv[0], "cpu_reconf", 4))
			return (do_cpu_reconf(argc-1, argv+1, resp));
		if (match((char *)argv[0], "dumpmem", 2))
			return (do_dumpmem(argc-1, argv+1, resp));
		if (match((char *)argv[0], "ldc-usage", 3))
			return (do_ldc_usage(argc-1, argv+1, resp));
		if (match((char *)argv[0], "isolate", 4))
			return (do_isolate(argc-1, argv+1, resp));
		if (match((char *)argv[0], "restore", 4))
			return (do_restore(argc-1, argv+1, resp));
		if (match((char *)argv[0], "suspend", 4))
			return (do_suspend(argc-1, argv+1, resp));
		if (match((char *)argv[0], "state-save", 7))
			return (do_state_save(argc-1, argv+1, resp));
		if (match((char *)argv[0], "state-restore", 7))
			return (do_state_restore(argc-1, argv+1, resp));
		if (match((char *)argv[0], "resume", 4))
			return (do_resume(argc-1, argv+1, resp));
	}

	rprintf(resp, "Available subcommands are:\n\n");
	rprintf(resp, "    apivers\n");
	rprintf(resp, "    echo\n");
	rprintf(resp, "    mode\n");
	rprintf(resp, "    mdstore\n");
	rprintf(resp, "    rdmem\n");
	rprintf(resp, "    wrmem\n");
	rprintf(resp, "    exit\n");
	rprintf(resp, "    (soft)state\n");
	rprintf(resp, "    state\n");
	rprintf(resp, "    cpu\n");
	rprintf(resp, "    tod\n");
	rprintf(resp, "    inval\n");
	rprintf(resp, "    hvconfig\n");
	rprintf(resp, "    pmi_get_all\n");
	rprintf(resp, "    pm_get_state\n");
	rprintf(resp, "    pm_set_state\n");
	rprintf(resp, "    (un)lock\n");
	rprintf(resp, "    res-usage\n");
	rprintf(resp, "    inject\n");
	rprintf(resp, "    core_md\n");
	rprintf(resp, "    cpu_reconf\n");
	rprintf(resp, "    dumpmem\n");
	rprintf(resp, "    ldc-usage\n");
	rprintf(resp, "    isolate\n");
	rprintf(resp, "    suspend\n");
	rprintf(resp, "    state-save\n");
	rprintf(resp, "    state-restore\n");
	rprintf(resp, "    resume\n");
	rprintf(resp, "    restore\n");

	return (CLI_FAILURE);
}
