/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines and tables for operations on 'domain' resources,
 * except for the 'migrate-domain' operation.  'migrate-domain' is
 * kept separate to keep the size of this file more manageable.
 */

#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>

#include <libintl.h>

#include "allocate.h"
#include "backend.h"
#include "basics.h"
#include "chain.h"
#include "cli_conn.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "database.h"
#include "dr.h"
#include "ds_impl.h"
#include "dispctl.h"
#include "errors.h"
#include "fatal.h"
#include "hvctl.h"
#include "ldom.h"
#include "mac_addrs.h"
#include "mac_multicast.h"
#include "master.h"
#include "mig.h"
#include "mig_be.h"
#include "mig_common.h"
#include "pm_events.h"
#include "response.h"
#include "sequencer.h"
#include "vsw.h"
#include "vnet.h"
#include "waiter.h"
#include "xml_cntrl_common.h"
#include "xml_event.h"
#include "zpri.h"

static bool_t stop_one_ldom(ldom_t *ldp, bool_t force, bool_t all,
    response_t *resp);
static bool_t start_one_ldom(ldom_t *ldp, bool_t all, cli_conn_t *client);
static bool_t bind_one_ldom(ldom_t *ldp, response_t *resp);
static bool_t unbind_one_ldom(ldom_t *ldp, response_t *resp);


cli_res_cmds_t	cli_domain_cmds = {
	.resinfop = "domain",
	.add_op =	{ cli_ldom_create,
		"(-i <file> | [mac-addr=<num>] [hostid=<num>]\n"
		"\t\t[failure-policy=<ignore|stop|reset|panic>]\n"
		"\t\t[master=<master_ldom1,...,master_ldom4>] <ldom>...)" },
	.set_op =	{ cli_ldom_set,
		"(-i <file> | [mac-addr=<num>] [hostid=<num>]\n"
		"\t\t[failure-policy=<ignore|stop|reset|panic>]\n "
		"\t\t[master=<master_ldom1,...,master_ldom4>] <ldom>)" },
	.remove_op =	{ cli_ldom_destroy,
		"(-a | <ldom>...)" },
	.list_op =	{ cli_ldom_list,
		"[-e] [-l] [-o <format>] [-p] [<ldom>...]\n\t\t"
		"'format' is one or more of:\n\t\tconsole,cpu,crypto,disk,"
		"domain,memory,network,physio,\n\t\tresmgmt,serial,status" },
	.start_op =	{ cli_ldom_start,
		"(-a | -i <file> | <ldom>...)" },
	.stop_op =	{ cli_ldom_stop,
		"[-f] (-a | <ldom>...)"	},
	.bind_op =	{ cli_ldom_bind,
		"(-i <file> | <ldom>)"	},
	.unbind_op =	{ cli_ldom_unbind,	"<ldom>" },
	.panic_op =	{ cli_ldom_panic,	"<ldom>" },
	.migrate_op =	{ cli_ldom_migrate,
		"[-f] [-n] [-p <password_file>] <source_ldom>\n"
		"\t\t[<user>@]<target_host>[:<target_ldom>]" },
};


/*
 * Returns true iff LDom 'ldp' is providing services for any other active
 * (state is other than INACTIVE) LDom.
 */
static bool_t
has_active_clients(ldom_t *ldp)
{
	return (ldp->bound.num_clients > 0);
}


cli_result_t
cli_ldom_list(int argc, char **argvp, cli_conn_t *clientp)
{
	response_t *resp = clientp->resp;
	int i = 0;
	char *arg;
	char *field;
	int start;
	ldom_t *ldp;
	dispctl_t dctl;

	dctl_init(&dctl);
	dctl.list = 1;

	/*
	 * Extract meaningful flags.
	 * LDom names should be last arg in list.
	 */
	for (i = 0; i < argc && argvp[i][0] == '-'; i++) {
		if (strcmp(argvp[i], "-l") == 0 ||
		    strcmp(argvp[i], "--long") == 0) {
			dctl.longf = 1;
		} else if (strcmp(argvp[i], "-p") == 0 ||
		    strcmp(argvp[i], "--parseable") == 0) {
			dctl.parseable = 1;
		} else if (strcmp(argvp[i], "-e") == 0 ||
		    strcmp(argvp[i], "--extended") == 0) {
			dctl.extended = 1;
			dctl.longf = 1;
		} else if (strcmp(argvp[i], "-o") == 0 ||
		    strcmp(argvp[i], "--oformat") == 0) {
			dctl.extended = 1;
			dctl.longf = 1;
			dctl.oformat = 1;
			if (i + 1 >= argc) {
				rprintf(resp, gettext("Missing format "
				    "argument\n"));
				return (CLI_RESULT_USAGE);
			}
			arg = argvp[++i];
			field = strtok(arg, ",");
			do {
				if (strcmp(field, "console") == 0)
					dctl.oconsole = 1;
				else if (strcmp(field, "cpu") == 0)
					dctl.ocpu = 1;
				else if (strcmp(field, "crypto") == 0)
					dctl.ocrypto = 1;
				else if (strcmp(field, "disk") == 0)
					dctl.odisk = 1;
				else if (strcmp(field, "domain") == 0)
					dctl.odomain = 1;
				else if (strcmp(field, "memory") == 0)
					dctl.omemory = 1;
				else if (strcmp(field, "network") == 0)
					dctl.onetwork = 1;
				else if (strcmp(field, "physio") == 0)
					dctl.ophysio = 1;
				else if (strncmp(field, "resmgmt", 2) == 0)
					dctl.oresmgmt = 1;
				else if (strcmp(field, "serial") == 0)
					dctl.oserial = 1;
				else if (strncmp(field, "status", 2) == 0)
					dctl.ostatus = 1;
				else {
					rprintf(resp, gettext("Unknown format "
					    "argument: %s\n"), field);
					return (CLI_RESULT_USAGE);
				}
				field = strtok(NULL, ",");
			} while (field != NULL);
		} else if (strcmp(argvp[i], "-?") == 0 ||
		    strcmp(argvp[i], "--help") == 0) {
			return (CLI_RESULT_USAGE);
		} else if (strcmp(argvp[i], "--") == 0) {
			i++;
			break;
		} else {
			return (CLI_RESULT_USAGE);
		}
	}

	config_mode_info_warning(resp, &dctl);
	output_version(resp, &dctl);

	if (i == argc) {
		/* first list domains which are not inactive */
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			if (ldp->state < BOUND)
				continue;
			update_statistics(ldp, dctl.longf ? true : false);
			if (dctl.longf) {
				print_bsc(ldp, resp, &dctl);
			} else {
				printldom(ldp, resp, &dctl);
				rprintf(resp, "\n");
			}
		}
		/* now list inactive domains */
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			if (ldp->state >= BOUND)
				continue;
			if (dctl.longf) {
				print_bsc(ldp, resp, &dctl);
			} else {
				printldom(ldp, resp, &dctl);
				rprintf(resp, "\n");
			}
		}

		return (CLI_RESULT_OK);
	}

	start = i;

	/* two pass walk through listed LDoms */
	for (i = start; i < argc; ++i) {
		if ((ldp = ldm_get_ldp(resp, argvp[i])) == NULL) {
			return (CLI_RESULT_FAIL);
		}
	}
	/* walk through the LDoms specified on the command line */
	for (i = start; i < argc; ++i) {
		ldp = ldm_get_ldp(resp, argvp[i]);

		update_statistics(ldp, dctl.longf ? true : false);
		if (dctl.longf) {
			print_bsc(ldp, resp, &dctl);
		} else {
			printldom(ldp, resp, &dctl);
			rprintf(resp, "\n");
		}
	}

	return (CLI_RESULT_OK);
}

cli_result_t
cli_ldom_panic(int argc, char **argvp, cli_conn_t *clientp)
{
	response_t *resp = clientp->resp;
	ldom_t *ldp;
	int rc;

	if (argc != 1)
		return (CLI_RESULT_USAGE);

	if ((ldp = ldm_get_ldp(resp, argvp[0])) == NULL)
		return (CLI_RESULT_FAIL);

	if (IS_CONTROL_DOMAIN(ldp) == true) {
		rprintf(resp,
		    gettext("Panic of the control domain is not allowed\n"));
		return (CLI_RESULT_FAIL);
	}

	if (check_locked_ldom(ldp, resp))
		return (CLI_RESULT_FAIL);

	if (ldp->state > BOUND) {
		if (ldom_io_reset_allowed(ldp)) {
			rc = backend.panic_guest(ldp, resp);
			return (rc == 0 ? CLI_RESULT_OK : CLI_RESULT_FAIL);
		} else {
			rprintf(resp, "Panic of domain %s in not allowed "
			    "because it contains an\nIO device which may not "
			    "be reset\n", ldp->namep);
			return (CLI_RESULT_FAIL);
		}
	} else {
		rprintf(resp, gettext("LDom %s is not active\n"), ldp->namep);
		return (CLI_RESULT_FAIL);
	}
}

cli_result_t
cli_ldom_create(int argc, char **argv, cli_conn_t *client)
{
	mac_addr_t mac_address = ILLEGAL_MAC_ADDR;
	char *value = NULL;
	char *end = NULL;
	int i = 0;
	response_t *resp = client->resp;
	ldom_t *ldp;
	bool_t one_ldom = false;
	bool_t set_master = false;
	uint64_t hostid = UNASSIGNED_HOSTID;
	uint64_t failure_policy = INVALID_FAILURE_POLICY;
	ldom_t *master_ldp;
	mlist_t *ml = NULL;
	mlist_t *mlist = NULL;
	cli_result_t result = CLI_RESULT_OK;

	if (argc < 1)
		return (CLI_RESULT_USAGE);

	while (i < argc) {
		/* check for optional mac-addr arg */
		if (strncmp(argv[i], "mac-addr=", strlen("mac-addr=")) == 0) {
			/* TRANSLATION_NOTE Don't translate 'mac-addr' */
			if (mac_address != ILLEGAL_MAC_ADDR) {
				rprintf(resp, gettext("Only one mac-addr can "
				    "be specified for a domain\n"));
				result = CLI_RESULT_USAGE;
				goto done;
			}

			value = strchr(argv[i], '=');
			if (*(++value) == '\0' ||
			    !atomac(value, &mac_address)) {
				rprintf(resp, gettext("Invalid mac-addr\n"));
				result = CLI_RESULT_FAIL;
				goto done;
			}

			/*
			 * Can MAC addr be allocated?
			 */
			if (IS_PUBLIC_MAC_ADDR(mac_address) &&
			    is_macaddr_in_use(mac_address,
			    PERFORM_LOCAL_MAC_CHECK, resp)) {
				result = CLI_RESULT_FAIL;
				goto done;
			}

			one_ldom = true;
			++i;
			continue;
		}

		/* check for optional hostid arg */
		if (strncmp(argv[i], "hostid=", strlen("hostid=")) == 0) {
			/* TRANSLATION_NOTE Don't translate 'hostid' */
			if (hostid != UNASSIGNED_HOSTID) {
				rprintf(resp, gettext("Only one hostid can "
				    "be specified for a domain\n"));
				result = CLI_RESULT_USAGE;
				goto done;
			}

			value = strchr(argv[i], '=');
			if (*(++value) == '\0' ||
			    ((hostid = strtoll(value, &end, 16)) == 0) ||
			    (end != NULL && *end != '\0') ||
			    (hostid > UINT_MAX) ||
			    (hostid == UNASSIGNED_HOSTID)) {
				rprintf(resp, gettext("Invalid hostid=%s\n"),
				    value);
				result = CLI_RESULT_FAIL;
				goto done;
			}

			one_ldom = true;
			++i;
			continue;
		}

		/* check for optional failure-policy arg */
		if (strncmp(argv[i], "failure-policy=",
		    strlen("failure-policy=")) == 0) {
			if (!LDOM_DEPENDENCIES_SUPPORTED() &&
			    failure_policy == INVALID_FAILURE_POLICY) {
				value = strchr(argv[i], '=');
				if (!get_failure_policy(++value,
				    &failure_policy) || failure_policy !=
				    FAILURE_POLICY_IGNORE) {
					rprintf(resp, gettext("System Firmware "
					    "version does not support failure "
					    "policies,\n%s specification will "
					    "be ignored.\n"), argv[i]);
					failure_policy = FAILURE_POLICY_IGNORE;
				}
				++i;
				continue;
			}

			/* TRANSLATION_NOTE Don't translate 'failure-policy' */
			if (failure_policy != INVALID_FAILURE_POLICY) {
				rprintf(resp, gettext("Only one failure policy "
				    "can be specified for a domain\n"));
				result = CLI_RESULT_USAGE;
				goto done;
			}

			value = strchr(argv[i], '=');
			if (*(++value) == '\0' ||
			    !get_failure_policy(value, &failure_policy)) {
				rprintf(resp, gettext("Invalid failure policy: "
				    "%s\n"), value);
				result = CLI_RESULT_FAIL;
				goto done;
			}

			++i;
			continue;
		}

		/* check for optional master arg */
		if (strncmp(argv[i], "master=",
		    strlen("master=")) == 0) {
			char *field;
			int j;

			if (!LDOM_DEPENDENCIES_SUPPORTED() &&
			    set_master == false) {
				rprintf(resp, gettext("System Firmware version "
				    "does not support domain dependencies,\n"
				    "%s specification will be ignored.\n"),
				    argv[i]);
				set_master = true;
				++i;
				continue;
			}

			/* TRANSLATION_NOTE Don't translate 'master' */
			if (set_master) {
				rprintf(resp, gettext("Only one set of masters "
				    "can be specified for a domain\n"));
				result = CLI_RESULT_USAGE;
				goto done;
			}
			set_master = true;

			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				rprintf(resp, gettext("Missing master "
				    "value\n"));
				result = CLI_RESULT_FAIL;
				goto done;
			}

			field = strtok(value, ",");
			for (j = 0; field != NULL &&
			    j < pri.max_guest_dependencies; ++j) {
				master_ldp = ldm_get_ldp(NULL, field);
				if (master_ldp == NULL) {
					rprintf(resp, gettext("master LDom "
					    "\"%s\" was not found\n"), field);
					result = CLI_RESULT_FAIL;
					goto done;
				} else {
					if (ml == NULL) {
						ml = Xcalloc(1, mlist_t);
						mlist = ml;
					} else {
						ml->next = Xcalloc(1, mlist_t);
						ml = ml->next;
					}
					ml->master = master_ldp;
					ml->next = NULL;
				}
				field = strtok(NULL, ",");
			}
			if (field != NULL && j == pri.max_guest_dependencies) {
				rprintf(resp, gettext("Only %d masters can be "
				    "specified per domain\n"),
				    pri.max_guest_dependencies);
				result = CLI_RESULT_FAIL;
				goto done;
			}
			if (unique_masters(mlist, resp) == false) {
				result = CLI_RESULT_FAIL;
				goto done;
			}

			++i;
			continue;
		}

		break;
	}

	if (mac_address == ILLEGAL_MAC_ADDR) {
		/* Make sure auto MAC allocation is available */
		if (!mac_socket_up) {
			/* Try to set up the multicast socket again */
			if (!start_mac_auto_allocate()) {
				rprintf(resp,
				    MSG_ERR_MAC_SOCKET_CREATION_FAILED);
				rprintf(resp, MSG_ERR_MAC_SET_MANUALLY);
				result = CLI_RESULT_FAIL;
				goto done;
			}
		}
	}

	/* default failure policy is to ignore everything */
	if (failure_policy == INVALID_FAILURE_POLICY)
		failure_policy = FAILURE_POLICY_IGNORE;

	/* Verify exactly one arg (ldom) left */
	if (one_ldom && (i != argc - 1)) {
		result = CLI_RESULT_USAGE;
		goto done;
	}

	/*
	 * Check that all the LDoms to be created do not exist.
	 */
	for (int j = i; j < argc; j++) {
		if (!is_valid_object_name(resp, argv[j])) {
			rprintf(resp, gettext("Illegal domain name %s\n"),
			    argv[j]);
			result = CLI_RESULT_FAIL;
			goto done;
		}

		if (find_ldom_by_name(argv[j]) != NULL) {
			rprintf(resp, gettext("LDom %s already exists\n"),
			    argv[j]);
			result = CLI_RESULT_FAIL;
			goto done;
		}
	}

	/*
	 * Create all the LDoms.
	 */
	for (int j = i; j < argc; j++) {
		int rv;
		vsw_mac_t  *vsw_nodep = NULL;

		/*
		 * We need to create a vsw node pointer so that the MAC
		 * address can be searched for collisions.
		 */
		rv = create_vsw_mac(&vsw_nodep, argv[j], argv[j], VSW_ROUTED,
		    mac_address);
		/*
		 * The only time we hit this error is if the Automatic MAC
		 * generation failed. This should ONLY occur for automatically
		 * generated MAC addresses (either send in or requested).
		 */
		if (rv != 0) {
			rprintf(resp, gettext("Automatic MAC allocation "
			    "failed.\nPlease set the MAC address for "
			    "the domain\n"));
			result = CLI_RESULT_FAIL;
			goto done;
		}

		(void) ldom_create(argv[j], vsw_nodep->mac_addr, hostid,
		    failure_policy, mlist);
		ldp = ldm_get_ldp(resp, argv[j]);
		if (ldp != NULL)
			event_send_dom_event(XML_ADD_DOMAIN, ldp, client);
	}

done:
	free_mlist(mlist);
	return (result);
}

static bool_t
is_hostid_in_use(uint64_t hostid)
{
	ldom_t  *ldp;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->props.hostid == hostid) {
			return (true);
		}
	}
	return (false);
}

cli_result_t
cli_ldom_set(int argc, char **argv, cli_conn_t *client)
{
	response_t	*resp = client->resp;
	char		*value = NULL;
	char		*end = NULL;
	ldom_t		*ldp = NULL;
	mac_addr_t	mac_address = ILLEGAL_MAC_ADDR;
	uint64_t	hostid = UNASSIGNED_HOSTID;
	uint64_t	failure_policy = INVALID_FAILURE_POLICY;
	bool_t		set_master = false;
	ldom_t		*master_ldp = NULL;
	mlist_t		*ml = NULL;
	mlist_t		*mlist = NULL;
	int		i = 0;
	int		j = 0;
	cli_result_t	result = CLI_RESULT_OK;

	if (argc < 1)
		return (CLI_RESULT_USAGE);

	if ((ldp = ldm_get_ldp(resp, argv[argc - 1])) == NULL)
		return (CLI_RESULT_USAGE);

	if (check_locked_ldom(ldp, resp))
		return (CLI_RESULT_FAIL);

	if (config_mod_allowed(ldp, DEL_RECONFIG_OTHER,
	    DR_OP_TYPE_DOMAIN, resp) == false)
		return (CLI_RESULT_FAIL);

	for (i = 0; i < argc - 1; ++i) {
		if (strncmp(argv[i], "mac-addr=", strlen("mac-addr=")) == 0) {
			/* TRANSLATION_NOTE Don't translate 'mac-addr' */
			if (mac_address != ILLEGAL_MAC_ADDR) {
				rprintf(resp, gettext("Only one mac-addr can "
				    "be specified for a domain\n"));
				result = CLI_RESULT_USAGE;
				goto done;
			}

			value = strchr(argv[i], '=');
			if (*(++value) == '\0' ||
			    !atomac(value, &mac_address)) {
				rprintf(resp, gettext("Invalid mac-addr\n"));
				result = CLI_RESULT_FAIL;
				goto done;
			}

			/*
			 * Can MAC addr be allocated?
			 */
			if (IS_PUBLIC_MAC_ADDR(mac_address) &&
			    is_macaddr_in_use(mac_address,
			    PERFORM_LOCAL_MAC_CHECK, resp)) {
					result = CLI_RESULT_FAIL;
					goto done;
			}

		} else if (strncmp(argv[i], "hostid=",
		    strlen("hostid=")) == 0) {
			/* TRANSLATION_NOTE Don't translate 'hostid' */
			if (hostid != UNASSIGNED_HOSTID) {
				rprintf(resp, gettext("Only one hostid can "
				    "be specified for a domain\n"));
				result = CLI_RESULT_USAGE;
				goto done;
			}

			value = strchr(argv[i], '=');
			if (*(++value) == '\0' ||
			    ((hostid = strtoll(value, &end, 16)) == 0) ||
			    (end != NULL && *end != '\0') ||
			    (hostid > UINT_MAX) ||
			    (hostid == UNASSIGNED_HOSTID)) {
				rprintf(resp, gettext("Invalid hostid=%s\n"),
				    value);
				result = CLI_RESULT_FAIL;
				goto done;
			}
		/*
		 * Handle failure-policy, aka guestcontrol
		 */
		} else if (strncmp(argv[i], "failure-policy=",
		    strlen("failure-policy=")) == 0) {
			if (!LDOM_DEPENDENCIES_SUPPORTED()) {
				rprintf(resp, MSG_ERR_GUEST_DEP_BAD_FIRMWARE);
				result = CLI_RESULT_FAIL;
				goto done;
			}

			if (failure_policy != INVALID_FAILURE_POLICY) {
				rprintf(resp, gettext("Only one failure policy "
				    "can be specified for a domain\n"));
				result = CLI_RESULT_USAGE;
				goto done;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0' ||
			    !get_failure_policy(value, &failure_policy)) {
				rprintf(resp, gettext("Invalid failure policy: "
				    "%s\n"), value);
				result = CLI_RESULT_FAIL;
				goto done;
			}
		/*
		 * Handle master, aka guestdependency
		 */
		} else if (strncmp(argv[i], "master=",
		    strlen("master=")) == 0) {
			char *field;

			if (!LDOM_DEPENDENCIES_SUPPORTED()) {
				rprintf(resp, MSG_ERR_GUEST_DEP_BAD_FIRMWARE);
				result = CLI_RESULT_FAIL;
				goto done;
			}

			if (IS_CONTROL_DOMAIN(ldp)) {
				rprintf(resp, gettext("Control LDom \"%s\" "
				    "cannot be a slave domain\n"), ldp->namep);
				result = CLI_RESULT_FAIL;
				goto done;
			}

			if (set_master == true) {
				rprintf(resp, gettext("Only one set of masters "
				    "can be specified for a domain\n"));
				result = CLI_RESULT_USAGE;
				goto done;
			}

			set_master = true;
			value = strchr(argv[i], '=');
			if (*(++value) == '\0')
				continue;

			field = strtok(value, ",");
			for (j = 0; field != NULL &&
			    j < pri.max_guest_dependencies; ++j) {

				master_ldp = ldm_get_ldp(NULL, field);
				if (master_ldp == NULL) {
					rprintf(resp, gettext("master LDom "
					    "\"%s\" was not found\n"), field);
					field = strtok(NULL, ",");
					continue;
				}

				if (ldp->state >= BOUND &&
				    master_ldp->state < BOUND) {
					rprintf(resp, gettext("LDom \"%s\" is "
					    "bound and requires LDom \"%s\" "
					    "be bound\n"), ldp->namep,
					    master_ldp->namep);
					result = CLI_RESULT_FAIL;
					goto done;
				} else {
					if (ml == NULL) {
						ml = Xcalloc(1, mlist_t);
						mlist = ml;
					} else {
						ml->next = Xcalloc(1, mlist_t);
						ml = ml->next;
					}
					ml->namep = ldp->namep;
					ml->master = master_ldp;
					ml->next = NULL;
				}
				field = strtok(NULL, ",");
			}
			if (field != NULL && j == pri.max_guest_dependencies) {
				rprintf(resp, gettext("Only %d masters can be "
				    "specified per domain\n"),
				    pri.max_guest_dependencies);
				result = CLI_RESULT_FAIL;
				goto done;
			}
			if (unique_masters(mlist, resp) == false) {
				result = CLI_RESULT_FAIL;
				goto done;
			}

		/*
		 * Handle all other arguments
		 */
		} else {
			rprintf(resp, gettext("Invalid argument '%s'\n"),
			    argv[i]);
			result = CLI_RESULT_FAIL;
			goto done;
		}
	}

	int rv = 0;
	vsw_mac_t  *vsw_nodep = NULL;

	if (mac_address != ILLEGAL_MAC_ADDR) {
		if (ldp->state > BOUND) {
			rprintf(resp, gettext(
			    "LDom %s must be stopped before MAC "
			    "address can be modified\n"), ldp->namep);
			result = CLI_RESULT_FAIL;
			goto done;
		}

		vsw_nodep = get_vsw_node(argv[i]);
		if (vsw_nodep == NULL) {
			rprintf(resp, gettext("Virtual switch service \"%s\" "
			    "cannot be found\n"), argv[i]);
			result = CLI_RESULT_FAIL;
			goto done;
		}

		/*
		 * We need to update a vsw node pointer
		 * so that the MAC address can be searched
		 * for collisions.
		 */
		rv = update_vsw_mac(vsw_nodep, argv[i],
		    mac_address, VSW_ROUTED);

		/*
		 * The only time we hit this error is if
		 * the Automatic MAC generation failed.
		 * This should ONLY occur for automatically
		 * generated MAC addresses (either send in or
		 * requested).
		 */
		if (rv != 0) {
			rprintf(resp, gettext("Automatic MAC allocation "
			    "failed.\nPlease set the MAC "
			    "address for the domain\n"));
			result = CLI_RESULT_FAIL;
			goto done;
		}

		ldp->props.mac_address = mac_address;
		database_ldom_set_mac_addr(ldp, mac_address);
	}

	if (hostid != UNASSIGNED_HOSTID) {
		if (ldp->state > BOUND) {
			rprintf(resp, gettext(
			    "LDom %s must be stopped before the hostid "
			    "can be modified\n"), ldp->namep);
			result = CLI_RESULT_FAIL;
			goto done;
		}

		if (is_hostid_in_use(hostid)) {
			rprintf(resp,
			    gettext("Hostid 0x%"PRIx64" is already in use\n"),
			    hostid);
			result = CLI_RESULT_FAIL;
			goto done;
		}
		ldp->props.hostid = hostid;
		database_ldom_set_hostid(ldp, hostid);
	}

	if (failure_policy != INVALID_FAILURE_POLICY) {
		ldp->props.failure_policy = failure_policy;
		database_ldom_set_failure_policy(ldp, failure_policy);
	}

	if (set_master == true) {
		if (dependency_cycle(ldp, mlist, resp) == true) {
			result = CLI_RESULT_FAIL;
			goto done;
		}
		mlist2master(ldp, mlist);
		database_ldom_set_master(ldp);
	}

	if (ldp->state >= BOUND)
		seq_config_now_or_later(ldp, DEL_RECONFIG_OTHER, resp);
	database_save();

done:
	free_mlist(mlist);
	return (result);
}

cli_result_t
cli_ldom_unbind(int argc, char **argv, cli_conn_t *client)
{
	ldom_t *ldp;
	response_t *resp = client->resp;

	if (argc != 1)
		return (CLI_RESULT_USAGE);

	if ((ldp = ldm_get_ldp(resp, (char *)argv[0])) == NULL)
		return (CLI_RESULT_FAIL);

	if (config_mod_allowed(ldp, DEL_RECONFIG_REMOVE_2PHASE,
	    DR_OP_TYPE_DOMAIN, resp) == false)
		return (CLI_RESULT_FAIL);

	if (!unbind_one_ldom(ldp, resp))
		return (CLI_RESULT_FAIL);

	seq_unbind_ldoms();
	cleanse_progress_chain();
	event_send_dom_event(XML_UNBIND, ldp, client);
	return (CLI_RESULT_OK);
}

cli_result_t
cli_ldom_stop(int argc, char **argv, cli_conn_t *client)
{
	int i;
	cli_result_t result = CLI_RESULT_OK;
	ldom_t *ldp;
	bool_t force = false;
	bool_t stop_all = false;
	response_t *resp = client->resp;
	ldm_err_t pm_result = { 0 };	/* Not an array because currently */
					/* only PM_RES_CPU resource type is */
					/* affected by a domain state change. */

	if (zeus.mode == ZEUS_CONFIG_MODE) {
		rprintf(resp, gettext(
		    "LDoms cannot be stopped when the LDom Manager "
		    "is in configuration mode.\n"));
		return (CLI_RESULT_FAIL);
	}

	for (i = 0; i < argc && argv[i][0] == '-'; i++) {
		if (strcmp(argv[i], "-f") == 0 ||
		    strcmp(argv[i], "--force") == 0) {
			force = true;
		} else if (strcmp(argv[i], "-a") == 0 ||
		    strcmp(argv[i], "--all") == 0) {
			stop_all = true;
		} else if (strcmp(argv[i], "-?") == 0 ||
		    strcmp(argv[i], "--help") == 0) {
			return (CLI_RESULT_USAGE);
		} else if (strcmp(argv[i], "--") == 0) {
			i++;
			break;
		} else {
			return (CLI_RESULT_USAGE);
		}
	}

	ASSERT(CHAIN_LENGTH(zeus.inprogress) == 0);

	/* First check for -a arg */
	if (stop_all) {
		if (i != argc) {
			rprintf(resp, gettext(
			    "Can't use --all and specify an LDom\n"));
			return (CLI_RESULT_FAIL);
		}

		/* Walk through all LDoms */
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			if (!config_mod_allowed(ldp, DEL_RECONFIG_STOP,
			    DR_OP_TYPE_DOMAIN, resp)) {
				result = CLI_RESULT_FAIL;
			} else {
				if (!stop_one_ldom(ldp, force, true, resp))
					result = CLI_RESULT_FAIL;
			}
		}
	} else {
		if (i == argc) {
			rprintf(resp, gettext("No LDom(s) specified\n"));
			return (CLI_RESULT_USAGE);
		}

		/* Walk through list of LDoms from command line */
		for (; i < argc; i++) {
			ldp = ldm_get_ldp(resp, (char *)argv[i]);
			if (ldp == NULL) {
				cleanse_progress_chain();
				return (CLI_RESULT_FAIL);
			}

			if (!config_mod_allowed(ldp, DEL_RECONFIG_STOP,
			    DR_OP_TYPE_DOMAIN, resp)) {
				result = CLI_RESULT_FAIL;
				break;
			}

			if (!stop_one_ldom(ldp, force, false, resp)) {
				result = CLI_RESULT_FAIL;
				break;
			}
		}
	}

	if (CHAIN_LENGTH(zeus.inprogress) > 0) {
		seq_stop_bound(force);
		/*
		 * Report any failures and set up to wait for LDoms which have
		 * not yet stopped.
		 */
		CHAIN_ITER(zeus.inprogress, ldp, inprog) {
			if (ldp->state == BOUND) {
				mem_dr_cleanup(ldp);
				rprintf(resp, gettext("LDom %s stopped\n"),
				    ldp->namep);

				event_send_dom_event(XML_STOP, ldp, client);

				/*
				 * Send a notification that a domain
				 * is down, which will result in its
				 * being removed from the list of
				 * booting guest domains.
				 */
				if (pm_domain_dr_reg_event_on) {
					send_pm_dom_event(ldp,
					    GUEST_DOMAIN_DOWN, NULL);
				}

				/* power down resources */
				if (pm_domain_events_on) {
					send_pm_dom_event(ldp, STOP_DOMAIN_POST,
					    &pm_result);
					/*
					 * FIXME: check for errors by checking
					 * pm_result.ldme_errs and traversing
					 * pm_result.ldme_private.
					 */
					pm_free_errors(&pm_result);
				}
			} else if (ldp->state == STOPPING) {
				mem_dr_cleanup(ldp);
				backend.get_guest_state(ldp, LDOM_HV_STATE,
				    &ldp->hv_state);

				if (ldp->hv_state.state ==
				    GUEST_STATE_STOPPED) {
					seq_ldom_state_change(ldp, client,
					    GUEST_STATE_STOPPED);
					rprintf(resp, gettext(
					    "LDom %s stopped\n"),
					    ldp->namep);
				} else if ((ldp->hv_state.state ==
				    GUEST_STATE_RESETTING)) {
					ldp->state = ACTIVE;
					rprintf(resp, gettext(
					    "LDom %s resetting, stop failed\n"),
					    ldp->namep);
					result = CLI_RESULT_FAIL;
				} else {
					add_waiter(ldp, client, BOUND,
					    LDM_STOP);
				}
			} else {
				rprintf(resp, gettext(
				    "LDom %s stop notification failed"),
				    ldp->namep);
				DBG(rprintf(resp, " (state=%s)",
				    ldom_state_str(ldp->state)));
				rprintf(resp, "\n");
				result = CLI_RESULT_FAIL;
			}
		}

		cleanse_progress_chain();
	}

	return (result);
}

cli_result_t
cli_ldom_bind(int argc, char **argv, cli_conn_t *client)
{
	ldom_t *ldp;
	response_t *resp = client->resp;

	if (argc != 1)
		return (CLI_RESULT_USAGE);

	ASSERT(CHAIN_LENGTH(zeus.inprogress) == 0);

	if ((ldp = ldm_get_ldp(resp, argv[0])) == NULL)
		return (CLI_RESULT_FAIL);

	if (config_mod_allowed(ldp, DEL_RECONFIG_OTHER,
	    DR_OP_TYPE_DOMAIN, resp) == false)
		return (CLI_RESULT_FAIL);

	if (!bind_one_ldom(ldp, resp))
		return (CLI_RESULT_FAIL);

	event_send_dom_event(XML_BIND_DOMAIN, ldp, client);

	/* notify HV & effected services of new binding */
	seq_bind_new();
	seq_send_md_update_notifs();

	/* register newly bound guests for Domain Services */
	CHAIN_ITER(zeus.inprogress, ldp, inprog) {
		ds_register_domain(ldp, true);
	}

	cleanse_progress_chain();
	return (CLI_RESULT_OK);
}

/*
 * Kick off sequencer to start the listed LDoms.
 */
cli_result_t
cli_ldom_start(int argc, char **argv, cli_conn_t *client)
{
	int i;
	bool_t start_all = false;
	ldom_t *ldp;
	cli_result_t result = CLI_RESULT_OK;
	response_t *resp = client->resp;

	if (zeus.mode == ZEUS_CONFIG_MODE) {
		rprintf(resp, gettext(
		    "LDoms cannot be started when the LDom Manager "
		    "is in configuration mode.\n"));
		return (CLI_RESULT_FAIL);
	}

	ASSERT(CHAIN_LENGTH(zeus.inprogress) == 0);

	/*
	 * parse the argument stream
	 */

	if (argc < 1)
		return (CLI_RESULT_USAGE);

	for (i = 0; i < argc && argv[i][0] == '-'; i++) {
		if (strcmp(argv[i], "-a") == 0 ||
		    strcmp(argv[i], "--all") == 0) {
			start_all = true;
		} else if (strcmp(argv[i], "-?") == 0 ||
		    strcmp(argv[i], "--help") == 0) {
			return (CLI_RESULT_USAGE);
		} else if (strcmp(argv[i], "--") == 0) {
			i++;
			break;
		} else {
			return (CLI_RESULT_USAGE);
		}
	}

	/* First check for -a option */
	if (start_all == true) {
		if (i != argc) {
			rprintf(resp, gettext(
			    "Can't use --all and specify an LDom\n"));
			return (CLI_RESULT_FAIL);
		}

		/* Walk through entire ldoms_reqs */
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			if (!start_one_ldom(ldp, true, client))
				result = CLI_RESULT_FAIL;
		}
	} else {
		/* Walk through list of LDoms */
		for (; i < argc; i++) {
			ldp = ldm_get_ldp(resp, (char *)argv[i]);
			if (ldp == NULL) {
				result = CLI_RESULT_FAIL;
			} else {
				if (!start_one_ldom(ldp, false, client)) {
					result = CLI_RESULT_FAIL;
				}
			}
		}
	}

	if (CHAIN_LENGTH(zeus.inprogress) > 0) {
		/* bring guests on line */
		seq_start_bound();

		/* check on guest status and record success/failure */
		CHAIN_ITER(zeus.inprogress, ldp, inprog) {
			if (ldp->state == STARTING) {
				backend.get_guest_state(ldp, LDOM_HV_STATE,
				    &ldp->hv_state);
				if (ldp->hv_state.state == GUEST_STATE_NORMAL ||
				    ldp->hv_state.state ==
				    GUEST_STATE_RESETTING) {
					seq_ldom_state_change(ldp, client,
					    GUEST_STATE_NORMAL);
					rprintf(resp, gettext(
					    "LDom %s started\n"),
					    ldp->namep);
				} else {
					add_waiter(ldp, client, ACTIVE|BOUND,
					    LDM_START);
				}
			} else if (ldp->state != ACTIVE) {
				/*
				 * The backend failed to initiate the guest
				 * start. Check the guest state in the
				 * hypervisor in case it became active,
				 * perhaps via a reboot.
				 */
				backend.get_guest_state(ldp, LDOM_HV_STATE,
				    &ldp->hv_state);
				if (ldp->hv_state.state == GUEST_STATE_NORMAL ||
				    ldp->hv_state.state ==
				    GUEST_STATE_RESETTING) {
					seq_ldom_state_change(ldp, client,
					    GUEST_STATE_NORMAL);
					rprintf(resp, gettext(
					    "LDom %s started\n"),
					    ldp->namep);
				} else {
					rprintf(resp, gettext(
					    "LDom %s start failed, retry "
					    "the operation\n"), ldp->namep);
				}
			} else {
				ASSERT(ldp->state == ACTIVE);
				rprintf(resp, gettext(
				    "LDom %s started\n"), ldp->namep);
				event_send_dom_event(XML_START_DOMAIN, ldp,
				    client);
			}
		}

		cleanse_progress_chain();
	}

	return (result);
}

cli_result_t
cli_ldom_destroy(int argc, char **argv, cli_conn_t *client)
{
	response_t *resp = client->resp;
	bool_t delete_all = false;
	ldom_t *ldp, *nextldp;
	int i;

	config_mode_change_warning(resp);

	for (i = 0; i < argc && argv[i][0] == '-'; i++) {
		if (strcmp(argv[i], "-a") == 0 ||
		    strcmp(argv[i], "--all") == 0) {
			delete_all = true;
		} else if (strcmp(argv[i], "-?") == 0 ||
		    strcmp(argv[i], "--help") == 0) {
			return (CLI_RESULT_USAGE);
		} else if (strcmp(argv[i], "--") == 0) {
			i++;
			break;
		}
	}

	ASSERT(CHAIN_LENGTH(zeus.inprogress) == 0);
	SANITY(CHAIN_INIT(zeus.inprogress));

	if (delete_all == true) {
		if (i != argc) {
			rprintf(resp, gettext(
			    "Can't use --all and specify an LDom\n"));
			return (CLI_RESULT_FAIL);
		}
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			if (IS_CONTROL_DOMAIN(ldp) == false)
				CHAIN_ADD(zeus.inprogress, ldp, inprog);
		}
	} else {
		if (i == argc)
			return (CLI_RESULT_USAGE);

		for (; i < argc; i++) {
			if ((ldp = ldm_get_ldp(resp, argv[i])) == NULL) {
				goto fail;
			}
			if (IS_CONTROL_DOMAIN(ldp) == true &&
			    zeus.allow_ctrl_domain_reconfig == false) {
				rprintf(resp, gettext(
				    "LDom %s cannot be deleted because "
				    "it is the control domain\n"), ldp->namep);
				goto fail;
			}

			CHAIN_ADD(zeus.inprogress, ldp, inprog);
		}
	}

	CHAIN_ITER(zeus.inprogress, ldp, inprog) {
		if (ldp->state != INACTIVE) {
			rprintf(resp, gettext(
			    "LDom %s must be %s before it can be deleted\n"),
			    ldp->namep, (ldp->state > BOUND) ?
			    "stopped and unbound" : "unbound");
			goto fail;
		}
	}

	for (ldp = CHAIN_START(zeus.inprogress); ldp != NULL; ldp = nextldp) {
		nextldp = CHAIN_NEXT(ldp->inprog);
		/* unlink from inprogress list now, while ldp still valid */
		CHAIN_UNLINK(zeus.inprogress, ldp, inprog);
		if (!destroy_one_ldom(ldp, resp, client)) {
			goto error_fail;
		}
	}

	CHAIN_INIT(zeus.inprogress);
	return (CLI_RESULT_OK);

error_fail:
	rprintf(resp, gettext("Unable to delete LDom %s\n"), ldp->namep);

fail:
	cleanse_progress_chain();
	return (CLI_RESULT_FAIL);
}

/*
 * These functions are used to start and stop LDOMs by ensuring resources are
 * bound, and then invoking the sequencer to get the HV to do the job.
 *
 * Returns true on success
 */
static bool_t
unbind_one_ldom(ldom_t *ldp, response_t *resp)
{
	if (check_locked_ldom(ldp, resp))
		return (false);

	switch (ldp->state) {
	case BOUND:
		if (IS_CONTROL_DOMAIN(ldp) == true) {
			rprintf(resp, gettext(
			    "LDom %s cannot be unbound because it is "
			    "the control domain\n"), ldp->namep);
			return (false);
		}
		if (has_active_clients(ldp)) {
			rprintf(resp, gettext(
			    "LDom %s cannot be unbound because it has "
			    "active clients\n"), ldp->namep);
			return (false);
		}
		if (has_bound_slaves(ldp, resp)) {
			rprintf(resp, gettext("LDom %s cannot be unbound "
			    "because it has at least one bound dependent "
			    "LDom\n"), ldp->namep);
			return (false);
		}
		if (has_io_dependencies(ldp)) {
			rprintf(resp, gettext("LDom %s cannot be "
			    "unbound because it has at least one bound domain "
			    "dependent on its I/O resources.\n"), ldp->namep);
			return (false);
		}
		ldp->state = UNBINDING;
		CHAIN_ADD(zeus.inprogress, ldp, inprog);
		break;
	case INACTIVE:
	case STARTING:
	case STOPPING:
	case SUSPENDING:
	case RESUMING:
	case SUSPENDED:
	case ACTIVE:
		rprintf(resp, gettext(
		    "LDom %s cannot be unbound because it is %s\n"),
		    ldp->namep, ldom_state_str(ldp->state));
		return (false);
	default:
		fatal("%s: Internal error LDom %s in state (%d) %s", __func__,
		    ldp->namep, (int)ldp->state, ldom_state_str(ldp->state));
		return (false);
	}

	return (true);
}

static bool_t
stop_one_ldom(ldom_t *ldp, bool_t force, bool_t all, response_t *resp)
{
	/*
	 * No modifications are permitted on a locked LDom except that
	 * a domain can be stopped while a memory DR operation is
	 * in progress.
	 */
	if (ldp->flag.memdr_in_progress == false &&
	    check_locked_ldom(ldp, resp)) {
		return (false);
	}

	switch (ldp->state) {
	case ACTIVE:
	case SUSPENDED:
		if (IS_CONTROL_DOMAIN(ldp) == true &&
		    zeus.allow_ctrl_domain_reconfig == false) {
			if (all) {
				return (true);
			} else {
				rprintf(resp, gettext(
				    "LDom %s cannot be stopped "
				    "because it is the control domain\n"),
				    ldp->namep);
				return (false);
			}
		}

		CHAIN_ADD(zeus.inprogress, ldp, inprog);
		break;
	case STARTING:
	case STOPPING:
	case SUSPENDING:
	case RESUMING:
		if (!force) {
			rprintf(resp, gettext(
			    "LDom %s is already %s, use -f "
			    "to force it to the STOPPED state\n"),
			    ldp->namep, ldom_state_str(ldp->state));
			return (false);
		}
		CHAIN_ADD(zeus.inprogress, ldp, inprog);
		break;
	case INACTIVE:
		if (all) {
			return (true);
		} else {
			rprintf(resp, gettext(
			    "LDom %s cannot be stopped because it is %s\n"),
			    ldp->namep, ldom_state_str(ldp->state));
			return (false);
		}
	case BOUND:
		if (all) {
			return (true);
		} else {
			rprintf(resp, gettext("LDom %s is already stopped\n"),
			    ldp->namep);
			return (false);
		}
	default:
		fatal("%s: Internal error LDom %s in state (%d) %s", __func__,
		    ldp->namep, (int)ldp->state, ldom_state_str(ldp->state));
	}

	return (true);
}

static bool_t
bind_one_ldom(ldom_t *ldp, response_t *resp)
{
	if (check_locked_ldom(ldp, resp))
		return (false);

	switch (ldp->state) {
	case ACTIVE:
	case BOUND:
	case SUSPENDED:
		rprintf(resp, gettext(
		    "LDom %s cannot be bound because it is already %s\n"),
		    ldp->namep, ldom_state_str(ldp->state));
		return (false);

	case STARTING:
	case STOPPING:
	case SUSPENDING:
	case RESUMING:
		rprintf(resp, gettext(
		    "LDom %s cannot be bound because it is %s\n"),
		    ldp->namep, ldom_state_str(ldp->state));
		return (false);

	case INACTIVE:
		if (!bind_an_ldom(ldp, resp))
			return (false);

		/*
		 * Binding was successful. Flag that a new MD must be
		 * generated for the LDom.
		 */
		ldp->flag.gen_new_md = true;
		ldp->state = BINDING;
		CHAIN_ADD(zeus.inprogress, ldp, inprog);
		break;

	default:
		fatal("%s: Internal error LDom %s in state (%d) %s", __func__,
		    ldp->namep, (int)ldp->state, ldom_state_str(ldp->state));
	}

	return (true);
}

static bool_t
start_one_ldom(ldom_t *ldp, bool_t all, cli_conn_t *client)
{
	mac_addr_t	ldom_mac, vsw_mac, vnet_mac; /* for collision check */
	vsw_t		*vsw;
	vnet_t		*vnet;

	ldm_err_t pm_result = { 0 };	/* Not an array because currently */
					/* only PM_RES_CPU resource type is */
					/* affected by a domain state change. */

	if (config_mod_allowed(ldp, DEL_RECONFIG_OTHER,
	    DR_OP_TYPE_DOMAIN, client->resp) == false)
		return (false);

	/*
	 * Run a MAC address collision check.
	 *
	 * MAC address collision check is already done with the "ldm create",
	 * but run the collision check here for the "ldm start" to find
	 * conflicts when using different networks between the create and
	 * start time. The MAC address collision check here will skip
	 * the local MAC check and only check the MAC collision on a different
	 * system.
	 */
	/* collision check for domain MAC address */
	ldom_mac = ldp->props.mac_address;
	if (is_macaddr_in_use(ldom_mac, SKIP_LOCAL_MAC_CHECK, client->resp))
		return (false);

	/* collision check for vsw MAC address */
	CHAIN_ITER(ldp->vsws, vsw, ldom) {
		vsw_mac = vsw->mac;
		if (is_macaddr_in_use(vsw_mac, SKIP_LOCAL_MAC_CHECK,
		    client->resp))
			return (false);
	}

	/* collision check for vnet MAC address */
	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		vnet_mac = vnet->mac;
		if (is_macaddr_in_use(vnet_mac, SKIP_LOCAL_MAC_CHECK,
		    client->resp))
			return (false);
	}

	switch (ldp->state) {
	case ACTIVE:
	case STARTING:
		if (all) {
			return (true);
		} else {
			rprintf(client->resp, gettext("LDom %s is %s\n"),
			    ldp->namep, ldom_state_str(ldp->state));
			return (false);
		}
	case INACTIVE:
		if (all) {
			return (true);
		} else {
			rprintf(client->resp, gettext(
			    "LDom %s is not bound\n"), ldp->namep);
			return (false);
		}
	case STOPPING:
		/* check if guest has actually stopped */
		backend.get_guest_state(ldp, LDOM_HV_STATE, &ldp->hv_state);
		if (ldp->hv_state.state == GUEST_STATE_STOPPED) {
			seq_ldom_state_change(ldp, client, GUEST_STATE_STOPPED);
			break;
		}
		/* guest is still not stopped */
		/* FALLTHRU */

	case SUSPENDING:
	case RESUMING:
	case SUSPENDED:
		/* a suspended domain can only be resumed or stopped */
		if (all) {
			return (true);
		} else {
			rprintf(client->resp, gettext(
			    "LDom %s cannot be started because it is %s\n"),
			    ldp->namep, ldom_state_str(ldp->state));
			return (false);
		}
	case BOUND:
		break;
	default:
		fatal("%s: Internal error LDom %s in state (%d) %s", __func__,
		    ldp->namep, (int)ldp->state, ldom_state_str(ldp->state));
	}

	if (pm_domain_events_on) {
		send_pm_dom_event(ldp, START_DOMAIN_PRE, &pm_result);

		if (pm_result.ldme_errs != LDME_OP_EOK) {
			rprintf(client->resp,
			    gettext("Power Mgmt failure %s\n"), ldp->namep);
			pm_free_errors(&pm_result);
			return (false);
		}

		/*
		 * FIXME: check for errors by checking pm_result.ldme_errs
		 * and traversing pm_result.ldme_private.
		 */
		pm_free_errors(&pm_result);
	}

	CHAIN_ADD(zeus.inprogress, ldp, inprog);
	return (true);
}
