/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines and tables for the 'add-io' and 'rm-io' operations.
 */

#include <string.h>

#include <libintl.h>

#include "allocate.h"
#include "basics.h"
#include "chain.h"
#include "cli_conn.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "database.h"
#include "dr.h"
#include "ldmd_types.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "response.h"
#include "sequencer.h"
#include "xml_cntrl_common.h"
#include "xml_event.h"
#include "errors.h"
#include "pcie.h"


typedef enum {
	IO_UNKNOWN_TOKEN = -1,
	IO_BUS_TOKEN	= 0,
	IO_DEVICE_TOKEN	= 1
} io_token_t;

cli_res_cmds_t	cli_io_cmds = {
	.resinfop = "io",
	.add_op =	{ cli_io_add,	"[bypass=on] <bus/device> <ldom>" },
	.remove_op =	{ cli_io_remove,	"<bus/device> <ldom>"	},
	.list_op =	{ cli_io_list,	"[-l] [-p] " }
};

/* ARGSUSED2 */
static bool_t
check_dup_io_name(ldom_t *ldp, char *name, response_t *resp)
{
	req_iodevice_t		*riop;
	pri_iodevice_t		*piop;
	pri_iodevice_t		*pp;
	bound_iodevice_t	*biop;
	req_pcie_dev_t		*rdev;
	pri_pcie_dev_t		*pdev;
	pri_pcie_dev_t		*rpdev;
	bound_pcie_dev_t	*bdev;

	/* Check against pcie buses (root complexes) first. */
	if (ldp->state < BOUND) {
		/* names */
		CHAIN_ITER(ldp->required.io_devices, riop, link) {
			if (strcmp(riop->name, name) == 0)
				return (true);
		}
		/* pseudonyms */
		if (piop = find_matching_io(name, NULL)) {
			CHAIN_ITER(ldp->required.io_devices, riop, link) {
				pp = find_matching_io(riop->name, NULL);
				if (pp == piop)
					return (true);
			}
		}
	} else {
		CHAIN_ITER(ldp->bound.iodevices, biop, ldom) {
			if (strcmp(biop->name, name) == 0 ||
			    match_iodev(name, biop->phys_iodev))
				return (true);
		}
	}

	/*
	 * 'name' did not match any pcie bus, so now check it as the name
	 * of a single loanable I/O device (including pseudonyms).
	 */
	if (ldp->state < BOUND) {
		/* names */
		CHAIN_ITER(ldp->required.io_pcie_devs, rdev, link) {
			if (strcmp(rdev->cli_name, name) == 0)
				return (true);
		}
		/* pseudonyms */
		if (pdev = pcie_find_matching_dev(name)) {
			CHAIN_ITER(ldp->required.io_pcie_devs, rdev, link) {
				rpdev = pcie_find_matching_dev(rdev->cli_name);
				if (pdev == rpdev)
					return (true);
			}
		}
	} else {
		CHAIN_ITER(ldp->bound.iodevices, biop, ldom) {
			CHAIN_ITER(biop->dev_list, bdev, ldom) {
				if (strcmp(bdev->cli_name, name) == 0 ||
				    pcie_match_pciedev(name,
				    bdev->phys_pcie_dev))
					return (true);
			}
		}
	}

	return (false);
}

/*
 * Use composition of string to determine the type of io token
 * (CLI arg for add-io, rm-io).  The token can refer to either
 * a pcie bus or a removable pcie device.  Either of those can be
 * in a path format or a pseudonym format.
 */
static io_token_t
io_token_type(char *s)
{
	pri_iodevice_t *pio;

	/*
	 * Search the list of pcie buses first trying for a match
	 * of the argument 's' with either the absolute name or
	 * the pseudonym of each bus.
	 */
	CHAIN_ITER(pri.iodevice_list, pio, link) {
		if ((strcmp(pio->name, s) == 0) ||
		    (strcmp(pio->cli_name, s)) == 0)
			return (IO_BUS_TOKEN);
	}

	/*
	 * No point in searching the pcie devices separately.  If we
	 * didn't find a match (which is not an error) we classify
	 * the argument as a DEVICE.
	 */
	return (IO_DEVICE_TOKEN);
}

static cli_result_t
ldm_add_bus(ldom_t *ldp,
    char *name, bool_t bypass, response_t *resp, cli_conn_t *clientp)
{
	req_iodevice_t *iodevs_p;

	iodevs_p = add_io_constraint(ldp, name, bypass);

	if (ldp->state >= BOUND) {
		if (bind_io(ldp, resp) == false) {
			CHAIN_UNLINK(ldp->required.io_devices, iodevs_p, link);
			Xfree(iodevs_p->name);
			Xfree(iodevs_p);
			return (CLI_RESULT_FAIL);
		}

		if (zeus.mode == ZEUS_MGMT_MODE) {
			if (ldp->state != BOUND) {
				/*
				 * Handle as a delayed reconfig operation.
				 * This is temporary, until PIO DR is directly
				 * supported.
				 */
				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_ADD, resp);
			} else {
				/* Do reconfig now */
				seq_download_mds_do_config(ldp);
			}
		}
	}

	database_ldom_add_iodevice(ldp, name, bypass);
	event_send_res_event(XML_ADD_IO, ldp, NULL, NULL, clientp);

	database_save();
	return (CLI_RESULT_OK);
}

/*
 * Add a loanable physical device to a domain.
 */
static cli_result_t
ldm_add_pciedev(ldom_t *ldp,
    char *name, bool_t bypass, response_t *resp, cli_conn_t *clientp)
{
	req_pcie_dev_t *rdev;

	if (bypass) {
		rprintf(resp,
		    gettext("`bybass' option invalid for pcie device"));
		return (CLI_RESULT_USAGE);
	}

	rdev = add_pciedev_constraint(ldp, name);

	if (ldp->state >= BOUND) {
		if (bind_pciedevs(ldp, resp) == false) {
			CHAIN_UNLINK(ldp->required.io_pcie_devs, rdev, link);
			Xfree(rdev->cli_name);
			Xfree(rdev);
			return (CLI_RESULT_FAIL);
		}

		if (ldp->state != BOUND) {
			/*
			 * Handle as a delayed reconfig operation.
			 */
			initiate_or_update_delayed_reconfig(ldp,
			    DEL_RECONFIG_ADD, resp);
		} else {
			/* Do reconfig now */
			seq_download_mds_do_config(ldp);
		}
	}

	database_ldom_add_pcie_device(ldp, name);
	event_send_res_event(XML_ADD_IO, ldp, NULL, NULL, clientp);

	database_save();
	return (CLI_RESULT_OK);
}

cli_result_t
cli_io_add(int argc, char **argvp, cli_conn_t *clientp)
{
	ldom_t *ldp;
	char *name;
	char *value;
	bool_t io_mmu_bypass = false;
	response_t *resp = clientp->resp;
	int i = 0;

	if (argc == 0)
		return (CLI_RESULT_USAGE);

	/* Optional arguments processed first */
	for (i = 0; i < argc; i++) {
		name = argvp[i];
		value = strchr(argvp[i], '=');
		if (value != NULL) {
			*value = '\0';	/* overwrite the name/value separator */
			value++;
		} else {
			/* End of optional args */
			break;
		}

		if (strcmp(name, "bypass") == 0) {
			/* TRANSLATION_NOTE Don't translate 'bypass' */
			if (strcmp(value, "on") != 0) {
				rprintf(resp, gettext("Invalid bypass mode\n"));
				return (CLI_RESULT_FAIL);
			}
			io_mmu_bypass = true;
		} else {
			return (CLI_RESULT_USAGE);
		}
	}

	if (argc - i != 2)
		return (CLI_RESULT_USAGE);

	name = argvp[i];
	if (valid_name(resp, name) == false) {
		rprintf(resp, gettext(
		    "%s is not a valid I/O resource specifier\n"), name);
		return (CLI_RESULT_FAIL);
	}

	if ((ldp = ldm_get_ldp(resp, (char *)argvp[++i])) == NULL) {
		return (CLI_RESULT_FAIL);
	}

	if (!IS_CONTROL_DOMAIN(ldp) && ldp->state == ACTIVE) {
		rprintf(resp, gettext("The device cannot be added "
		    "to a domain which is in the active state.\n"));
		return (CLI_RESULT_FAIL);
	}

	if (!config_mod_allowed(ldp, DEL_RECONFIG_ADD, DR_OP_TYPE_PIO, resp))
		return (CLI_RESULT_FAIL);

	if (check_dup_io_name(ldp, name, resp)) {
		rprintf(resp, gettext("IO %s \"%s\" already exists\n"),
		    (ldp->state < BOUND) ? "constraint" : "resource", name);
		return (CLI_RESULT_FAIL);
	}

	switch (io_token_type(name)) {
	case IO_BUS_TOKEN:
		return (ldm_add_bus(ldp, name, io_mmu_bypass, resp, clientp));
	case IO_DEVICE_TOKEN:
	default:
		return (ldm_add_pciedev(ldp,
		    name, io_mmu_bypass, resp, clientp));
	}
}

static cli_result_t
ldm_rm_pci_bus(ldom_t *ldp, char *name, response_t *resp, cli_conn_t *clientp)
{
	req_iodevice_t *riop;
	bound_iodevice_t *biop;

	if (ldp->state >= BOUND) {
		CHAIN_ITER(ldp->bound.iodevices, biop, ldom) {
			/* Skip virtual_pcie_bus nodes */
			if (biop->virtual_pcie_bus)
				continue;
			if ((strcmp(biop->name, name) == 0) ||
			    (strcmp(biop->req_iop->name, name) == 0) ||
			    find_matching_io(name, NULL) == biop->phys_iodev) {
				break;
			}
		}

		if (biop == NULL) {
			rprintf(resp, gettext("No matching I/O found.\n"));
			return (CLI_RESULT_FAIL);
		}

		if (!pcie_is_rc_removable(ldp, biop)) {
			rprintf(resp,
			    gettext("Cannot remove the IO bus since one or "
			    "more of its devices has been removed.\n"));
			return (CLI_RESULT_FAIL);
		}

		riop = biop->req_iop;
		unbind_io(ldp, biop);

		if (zeus.mode == ZEUS_MGMT_MODE) {
			if (ldp->state != BOUND) {
				/*
				 * Handle as a delayed reconfig operation.
				 * This is temporary, until PIO DR is directly
				 * supported.
				 */
				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_REMOVE_PHYS, resp);
			} else {
				/* Do reconfig now */
				seq_download_mds_do_config(ldp);
			}
		}
	} else {
		CHAIN_ITER(ldp->required.io_devices, riop, link) {
			pri_iodevice_t *piop;
			if ((strcmp(riop->name, name) == 0) ||
			    (((piop = find_matching_io(riop->name, NULL))
			    != NULL) && match_iodev(name, piop)))
				break;
		}

		if (riop == NULL) {
			rprintf(resp, gettext("No matching I/O found.\n"));
			return (CLI_RESULT_FAIL);
		}
	}

	CHAIN_UNLINK(ldp->required.io_devices, riop, link);
	Xfree(riop->name);
	Xfree(riop);

	database_ldom_del_iodevice(ldp, name);
	event_send_res_event(XML_REMOVE_IO, ldp, NULL, NULL, clientp);
	database_save();

	return (CLI_RESULT_OK);
}

static cli_result_t
ldm_rm_pcie_device(ldom_t *ldp,
    char *name, response_t *resp, cli_conn_t *clientp)
{
	req_pcie_dev_t *rdev;
	bound_pcie_dev_t *bdev;
	bound_iodevice_t *bio;

	if (ldp->state >= BOUND) {
		if ((bdev = pcie_find_bound_pciedev(ldp,
		    name, &bio)) == NULL) {
			rprintf(resp, gettext("No matching I/O found.\n"));
			return (CLI_RESULT_FAIL);
		}

		if (!pcie_is_device_removable(ldp, bdev->phys_pcie_dev)) {
			rprintf(resp, gettext("The %s "
			    "device is not removable\n"), bdev->cli_name);
			return (CLI_RESULT_FAIL);
		}

		rdev = bdev->req_pcie_dev;
		unbind_pciedev(ldp, bio, bdev);

		/*
		 * If we have removed the last shareable device from
		 * a non-fabric owner, unlink and free the virtual
		 * bound_iodevice_t we created in this domain.
		 */
		if (bio->virtual_pcie_bus &&
		    (CHAIN_START(bio->dev_list) == NULL)) {
			CHAIN_UNLINK(ldp->bound.iodevices, bio, ldom);
			destroy_bound_iodevice(bio);
		}

		if (ldp->state != BOUND) {
			/*
			 * Handle as a delayed reconfig operation.
			 */
			initiate_or_update_delayed_reconfig(ldp,
			    DEL_RECONFIG_REMOVE_PHYS, resp);
		} else {
			/* Do reconfig now */
			seq_download_mds_do_config(ldp);
		}
	} else { /* not bound, so just search for the constraint */
		CHAIN_ITER(ldp->required.io_pcie_devs, rdev, link) {
			pri_pcie_dev_t *pdev;

			if ((strcmp(rdev->cli_name, name) == 0) || (((pdev =
			    pcie_find_matching_dev(rdev->cli_name)) != NULL) &&
			    pcie_match_pciedev(name, pdev)))
				break;
		}

		if (rdev == NULL) {
			rprintf(resp, gettext(
			    "Cannot find matching pcie device."));
			return (CLI_RESULT_FAIL);


		}
	}

	ASSERT(rdev != NULL);
	CHAIN_UNLINK(ldp->required.io_pcie_devs, rdev, link);
	pcie_destroy_req_pciedev(rdev);

	(void) database_ldom_del_pcie_device(ldp, name);
	event_send_res_event(XML_REMOVE_IO, ldp, NULL, NULL, clientp);
	database_save();

	return (CLI_RESULT_OK);
}

cli_result_t
cli_io_remove(int argc, char **argv, cli_conn_t *client)
{
	response_t *resp = client->resp;
	ldom_t *ldp;
	char *name;

	if (argc != 2)
		return (CLI_RESULT_USAGE);

	name = argv[0];
	if (valid_name(resp, name) == false) {
		rprintf(resp, gettext(
		    "%s is not a valid I/O resource specifier\n"), name);
		return (CLI_RESULT_FAIL);
	}

	if ((ldp = ldm_get_ldp(resp, (char *)argv[1])) == NULL)
		return (CLI_RESULT_FAIL);

	if (!IS_CONTROL_DOMAIN(ldp) && ldp->state > BOUND) {
		rprintf(resp, MSG_ERR_DO_OP_BOUND_OR_INACTIVE);
		return (CLI_RESULT_FAIL);
	}

	if (!config_mod_allowed(ldp,
	    DEL_RECONFIG_REMOVE_PHYS, DR_OP_TYPE_PIO, resp))
		return (CLI_RESULT_FAIL);

	switch (io_token_type(name)) {
	case IO_BUS_TOKEN:
		return (ldm_rm_pci_bus(ldp, name, resp, client));
	case IO_DEVICE_TOKEN:
		return (ldm_rm_pcie_device(ldp, name, resp, client));
	case IO_UNKNOWN_TOKEN:
	default:
		ASSERT(0);
		return (CLI_RESULT_FAIL);
	}
}

/*
 * Add a pcie device constraint.
 *
 * Unlike bound pcie devices, we don't know where this one will go
 * yet, or if it even exists.  So, keep such requests in a flat list under
 * the target ldom.  If it finally gets bound, it will be added to the
 * dev_list of the appropriate bound_iodevice_t.
 */
req_pcie_dev_t *
add_pciedev_constraint(ldom_t *ldp, char *name)
{
	req_pcie_dev_t *rdev;

	rdev = pcie_build_req_pciedev(name, false);
	CHAIN_ADD(ldp->required.io_pcie_devs, rdev, link);

	return (rdev);
}

cli_result_t
cli_io_list(int argc, char **argv, cli_conn_t *client)
{
	dispctl_t dctl;
	int i;
	char *fmt;
	char *sub_fmt;
	response_t *resp = client->resp;

	dctl_init(&dctl);

	for (i = 0; i < argc; i++) {
		if (strcmp(argv[i], "-l") == 0 ||
		    strcmp(argv[i], "--long") == 0) {
			dctl.longf = 1;
		} else if (strcmp(argv[i], "-p") == 0 ||
		    strcmp(argv[i], "--parseable") == 0) {
			dctl.parseable = 1;
		} else {
			return (CLI_RESULT_USAGE);
		}
	}

	if (!dctl.parseable) {
		fmt = "%-16s%-16s%-16s\n";
		rprintf(resp, fmt, "IO", "PSEUDONYM", "DOMAIN");
		rprintf(resp, fmt, "--", "---------", "------");
	} else {
		fmt = "|dev=%s|alias=%s|domain=%s\n";
	}

	pcie_display_io(fmt, resp);

	if (!zeus.sdio_enabled)
		return (CLI_RESULT_OK);

	if (!dctl.parseable) {
		rprintf(resp, "\n");
		fmt = pcie_get_listio_format();
		sub_fmt = dctl.longf ? "\t%s\n" : NULL;
		rprintf(resp, fmt,
		    "PCIE", "PSEUDONYM", "STATUS", "DOMAIN");
		rprintf(resp, fmt,
		    "----", "---------", "------", "------");
	} else {
		fmt = "|dev=%s|alias=%s|status=%s|domain=%s\n";
		sub_fmt = dctl.longf ? "|subdev=%s\n" : NULL;
	}

	pcie_display_devs(fmt, sub_fmt, resp);
	return (CLI_RESULT_OK);
}
