/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Routines and tables for the top-most level of CLI processing.
 */

#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <alloca.h>
#include <libintl.h>
#include <time.h>
#include <unistd.h>

#include "allocate.h"
#include "audit_ldom.h"
#include "auth_ldom.h"
#include "autosave.h"
#include "backend.h"
#include "basics.h"
#include "cli_conn.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "clp.h"
#include "fatal.h"
#include "ldm_protocol.h"
#include "ldomconfigure.h"
#include "response.h"
#include "zpri.h"
#include "zprintf.h"


static void
cleanup_args(int argc, char **argv)
{
	int i;

	if (argv != NULL) {
		for (i = 0; i < argc; i++)
			if (argv[i] != NULL)
				Xfree(argv[i]);
		Xfree(argv);
	}
}

static uint32_t
get_uint32(void *p)
{
	uint32_t value;

	/* can't use ntoh32 directly since p may not be suitably aligned */
	memcpy(&value, p, sizeof (uint32_t));
	return (ntoh32(value));
}

static bool_t
deserialize_cmd(caddr_t buf, int *argc, char ***argv, response_t *resp)
{
	zeus_cmd_msg_t *msg;
	char **cmd_argv = NULL;
	uint32_t cmd_argc;
	caddr_t p;
	int i, arg_len, arg_type;
	bool_t ok = true;

	/* pull command header from beginning of buffer */
	msg = (zeus_cmd_msg_t *)buf;
	cmd_argc = ntoh32(msg->argc);

	if (cmd_argc >= MAXARGC) {
		ok = false;
		goto done;
	}

	if (cmd_argc > 0)
		cmd_argv = Xcalloc(cmd_argc, char *);

	/* pull arguments out of buffer */
	p = buf + sizeof (*msg);
	for (i = 0; i < cmd_argc; i++) {
		arg_type = get_uint32(p);
		p += sizeof (uint32_t);

		switch (arg_type) {

		case TEXTBLOCK:
			arg_len = get_uint32(p);
			p += sizeof (uint32_t);
			cmd_argv[i] = Xmalloc(arg_len);
			memcpy(cmd_argv[i], p, arg_len);
			p += arg_len;
			break;

		default:
			rprintf(resp, gettext("Invalid argument type %d\n"),
			    arg_type);
			ok = false;
			goto done;
		}
	}

done:
	if (ok) {
		*argc = cmd_argc;
		*argv = cmd_argv;
	} else {
		cleanup_args(cmd_argc, cmd_argv);
	}

	return (ok);
}

#if ZDBG
static void
dump_cli_command(cli_conn_t *client, char *cmd, int argc, char **argv)
{
	char timestamp[16];
	time_t t;
	int i;

	time(&t);
	cftime(timestamp, "%T", &t);
	zprintf_lock();
	zprintf("[%s] 0x%p -> %s ", timestamp, (void *)client, cmd);
	for (i = 0; i < argc; i++)
		zprintf("%s ", argv[i]);
	zprintf("\n");
	zprintf_unlock();
	fsync(fileno(stderr));
}
#endif	/* ZDBG */

typedef struct CLI_VERB cli_verb_t;
typedef struct CLI_RES cli_res_t;

struct CLI_RES {
	char		*namep;
	cli_res_cmds_t	*cmdsp;
};

struct CLI_VERB {
	char	*namep;
	action_verb_t	cmd;
	bool_t	implied_domain;
	bool_t	is_alias;
};

cli_verb_t	cli_verb[] = {
	{ "add",	CLI_VERB_ADD,		false,	false },
	{ "set",	CLI_VERB_SET,		false,	false },
	{ "remove",	CLI_VERB_REMOVE,	false,	false },
	{ "rm",		CLI_VERB_REMOVE,	false,	true },
	{ "list",	CLI_VERB_LIST,		false,	false },
	{ "ls",		CLI_VERB_LIST,		false,	true },
	{ "start",	CLI_VERB_START,		true,	false },
	{ "stop",	CLI_VERB_STOP,		true,	false },
	{ "bind",	CLI_VERB_BIND,		true,	false },
	{ "unbind",	CLI_VERB_UNBIND,	true,	false },
	{ "panic",	CLI_VERB_PANIC,		true,	false },
	{ "migrate",	CLI_VERB_MIGRATE,	true,	false },
	{ "cancel",	CLI_VERB_CANCEL,	false,	false },
	{ NULL, -1, 0 },
};


static cli_res_t	cli_res[] = {
	{ "bindings",	&cli_bindings_cmds },
	{ "services",	&cli_services_cmds },
	{ "constraints", &cli_constraints_cmds },
	{ "devices",	&cli_devices_cmds },
	{ "domain",	&cli_domain_cmds },
	{ "dom",	&cli_domain_cmds },
	{ "io",		&cli_io_cmds },
	{ "crypto",	&cli_crypto_cmds },
	{ "mau",	&cli_crypto_cmds },
	{ "memory",	&cli_memory_cmds },
	{ "mem",	&cli_memory_cmds },
	{ "operation",	&cli_cancel_cmds },
	{ "policy",	&cli_policy_cmds },
	{ "reconf",	&cli_reconf_cmds },
	{ "spconfig",	&cli_spconfig_cmds },
	{ "config",	&cli_spconfig_cmds },
	{ "variable",	&cli_var_cmds },
	{ "var",	&cli_var_cmds },
	{ "vconscon",	&cli_vcc_cmds },
	{ "vcc",	&cli_vcc_cmds },
	{ "vconsole",	&cli_vcons_cmds },
	{ "vcons",	&cli_vcons_cmds },
	{ "vcpu",	&cli_vcpu_cmds },
	{ "vdisk",	&cli_vdisk_cmds },
	{ "vdiskserver", &cli_vdiskserver_cmds },
	{ "vds",	&cli_vdiskserver_cmds },
	{ "vdpcc",	&cli_vdpcc_cmds },
	{ "ndpsldcc",	&cli_vdpcc_cmds },
	{ "vdpcs",	&cli_vdpcs_cmds },
	{ "ndpsldcs",	&cli_vdpcs_cmds },
	{ "vdiskserverdevice", &cli_vds_device_cmds },
	{ "vdsdev",	&cli_vds_device_cmds },
	{ "vnet",	&cli_vnet_cmds },
	{ "vswitch",	&cli_vswitch_cmds },
	{ "vsw",	&cli_vswitch_cmds },
	{ NULL, NULL },
};

typedef struct {
	char	*aliasp;
	char	*cmdp;
} cli_alias_t;

static cli_alias_t	cli_alias[] = {
	{ "cancel-op",		"cancel-operation" },
	{ "create",		"add-domain" },
	{ "modify",		"set-domain" },
	{ "destroy",		"remove-domain"	},
	{ "remove-reconf",	"cancel-reconf"	},
	{ NULL, NULL }
};


static action_op_t *
cli_object_op(cli_res_cmds_t *cmdsp, int cmd)
{
	action_op_t	*op;
	op = NULL;
	switch (cmd) {
	case	CLI_VERB_ADD:		op = &cmdsp->add_op;	break;
	case	CLI_VERB_SET:		op = &cmdsp->set_op;	break;
	case	CLI_VERB_REMOVE:	op = &cmdsp->remove_op;	break;
	case	CLI_VERB_LIST:		op = &cmdsp->list_op;	break;
	case	CLI_VERB_START:		op = &cmdsp->start_op;	break;
	case	CLI_VERB_STOP:		op = &cmdsp->stop_op;	break;
	case	CLI_VERB_BIND:		op = &cmdsp->bind_op;	break;
	case	CLI_VERB_UNBIND:	op = &cmdsp->unbind_op;	break;
	case	CLI_VERB_PANIC:		op = &cmdsp->panic_op;	break;
	case	CLI_VERB_MIGRATE:	op = &cmdsp->migrate_op; break;
	case	CLI_VERB_CANCEL:	op = &cmdsp->cancel_op;	break;
	default:
		fatal("Illegal internal command value");
	}

	return (op);
}

static cli_res_t *
cli_find_resource_by_name(char *namep)
{
	int i;

	for (i = 0; cli_res[i].namep != NULL; i++) {
		if (strcmp(cli_res[i].namep, namep) == 0) {
			return (&(cli_res[i]));
		}
	}
	return (NULL);
}

/*
 * Break a command into the relevant verb and resource
 * or return false if its not recognised.
 */

static bool_t
cli_break_command(char *cmd_strp, cli_verb_t **verbpp,
    cli_res_cmds_t **cmdspp, action_op_t **opp)
{
	char *vstrp, *rstrp;
	cli_verb_t *verbp;
	cli_res_t *resourcep;
	action_op_t *op;
	int i;

	/* Firstly look for any alias substitutions */

	for (i = 0; cli_alias[i].aliasp != NULL; i++) {
		if (strcmp(cmd_strp, cli_alias[i].aliasp) == 0) {
			DBG(zprintf("cli alias: %s -> %s\n", cmd_strp,
			    cli_alias[i].cmdp));
			cmd_strp = cli_alias[i].cmdp;
			break;
		}
	}

	/* Now break the cmd into the verb and resource pair */
	vstrp = alloca(strlen(cmd_strp)+1);
	strcpy(vstrp, cmd_strp);

	/* Find the resource for the command or a null string */
	rstrp = strchr(vstrp, '-');
	if (rstrp != NULL) {
		*rstrp++ = '\0';
		if (*rstrp == '\0')
			rstrp = NULL;
	}

	/* Now match the verb to the appropriate method */
	for (verbp = cli_verb; verbp->namep != NULL; verbp++) {
		if (strcmp(verbp->namep, vstrp) == 0)
			goto match;
	}

	goto unknown;

match:;

	/*
	 * Verbs either apply to a resource, in which case the next
	 * operation is to try and figure out which resource it is so we can
	 * call the function.
	 * Alternatively the verb implicitly applies to a domain.
	 * Finally, list applies not just to resources but to other
	 * concepts like bindings or free devices etc.
	 */

	if (rstrp == NULL) {
		/* check for the list case */
		if (verbp->cmd == CLI_VERB_LIST ||
		    verbp->implied_domain) {
			resourcep = cli_find_resource_by_name("domain");
			goto full_match;
		}
		goto unknown;
	}

	resourcep = cli_find_resource_by_name(rstrp);
	if (resourcep == NULL)
		goto unknown;

full_match:;
		/* finally check to see if the verb has a func */
	op = cli_object_op(resourcep->cmdsp, verbp->cmd);

	if (op->funcp == NULL)
		goto unknown;

	*verbpp = verbp;
	*cmdspp = resourcep->cmdsp;
	*opp = op;
	return (true);

unknown:;
	*verbpp = NULL;
	*cmdspp = NULL;
	*opp = NULL;
	return (false);
}

static void
cli_object_help(cli_res_cmds_t *cmdsp, response_t *resp)
{
	action_op_t	*op;
	int i;

	for (i = 0; cli_verb[i].namep != NULL; i++) {
		if (!cli_verb[i].is_alias) {
			op = cli_object_op(cmdsp, cli_verb[i].cmd);
			if (op->funcp != NULL) {
				/* FIXME: does this need i18n? */
				rprintf(resp, "\t%s-%s %s\n",
				    cli_verb[i].namep, cmdsp->resinfop,
				    op->helpstrp);
			}
		}
	}
}

static void
display_aliases(response_t *resp)
{
	int	i, j;
	int	maxlen;

	/* XXX: This only needs to happen as aliases are added */
	maxlen = 0;
	for (i = 0; cli_alias[i].aliasp != NULL; i++) {
		int len;
		len = strlen(cli_alias[i].aliasp);
		if (len > maxlen)
			maxlen = len;
	}
	for (i = 0; cli_verb[i].namep != NULL; i++) {
		int len;
		if (cli_verb[i].is_alias || cli_verb[i].implied_domain) {
			len = strlen(cli_verb[i].namep);
			if (len > maxlen)
				maxlen = len;
		}
	}

	maxlen++;
	rprintf(resp, "\n%s\n", gettext("Verb aliases:"));
	rprintf(resp, "\t%2$-*1$s %3$s\n", maxlen, gettext("Alias"),
	    gettext("Verb"));
	rprintf(resp, "\t%2$-*1$s %3$s\n", maxlen, "-----", "-------");
	for (i = 0, j = 0; cli_verb[i].namep != NULL; i++) {
		if (cli_verb[i].is_alias) {
			rprintf(resp, "\t%2$-*1$s %3$s\n", maxlen,
			    cli_verb[i].namep, cli_verb[j].namep);
		} else {
			/* j points to most recent non-alias */
			j = i;
		}
	}

	rprintf(resp, "\n%s\n", gettext("Command aliases:"));
	rprintf(resp, "\t%2$-*1$s %3$s\n", maxlen, gettext("Alias"),
	    gettext("Command"));
	rprintf(resp, "\t%2$-*1$s %3$s\n", maxlen, "-----", "-------");

	for (i = 0; cli_alias[i].aliasp != NULL; i++) {
		rprintf(resp, "\t%2$-*1$s %3$s\n", maxlen,
		    cli_alias[i].aliasp, cli_alias[i].cmdp);
	}
	for (i = 0; cli_verb[i].namep != NULL; i++) {
		if (cli_verb[i].implied_domain)
			rprintf(resp, "\t%2$-*1$s %2$s-domain\n", maxlen,
			    cli_verb[i].namep);
	}
}

static void
cli_help(int argc, char *argv[], response_t *resp)
{
	int i, j, idx;
	cli_verb_t *verbp;
	cli_res_cmds_t *cmdsp;
	action_op_t *op;

	if (argc <= 1 || argv == NULL)
		goto usage;

	/*
	 * This is the tricky part ...
	 * Maybe a command was specified in which case dump the help for all the
	 * resources the command acts on.
	 * Maybe a resource was specified in which case dump the help for all
	 * the commands that can act on that resource.
	 * If we get a command and a resource then just output that help.
	 */

	idx = 1;

	/* Start by looking for a command in argv[idx] */

	if (cli_break_command(argv[idx], &verbp, &cmdsp, &op)) {
		rprintf(resp, "\n%s\n\tldm %s %s\n\n", gettext("Usage:"),
		    argv[idx], op->helpstrp);
		return;
	}

	rprintf(resp, gettext("Unknown command %s\n"), argv[idx]);

usage:;
	rprintf(resp, "\n%s\n", gettext("Usage:"));

	/* TRANSLATION_NOTE Don't change word order */
	rprintf(resp, "\tldm [--help] %s\n\tldm -V\n\n",
	    gettext("command [options] [properties] operands"));
	rprintf(resp, gettext("Options:\n\t-V\tDisplay version "
	    "information\n\n"));

	rprintf(resp, gettext("Command(s) for each resource "
	    "(aliases in parens):\n"));

	for (i = 0; cli_res[i].namep != NULL; i = j) {
		j = i + 1;
		rprintf(resp, "\n    %s", cli_res[i].namep);
		if (cli_res[i].cmdsp == cli_res[i+1].cmdsp) {

			rprintf(resp, "\t(");
			for (; cli_res[j].namep != NULL &&
			    cli_res[j].cmdsp == cli_res[i].cmdsp; j++) {
				rprintf(resp, " %s", cli_res[j].namep);
			}
			rprintf(resp, " )");
		}
		rprintf(resp, "\n");

		cli_object_help(cli_res[i].cmdsp, resp);
	}

	display_aliases(resp);
}

/*
 * This function takes the first argument and attempts to
 * break it into a verb-resource pair. The appropriate method in the
 * resource will then be invoked to perform the command.
 * For example:
 *	ldm add-vcpu 5 domainX
 * The verb is the add method, and the object type to manipulate
 * is a vcpu .. acting on the domainX subject.
 */
static int
cli_parser(int argc, char *argv[], uid_t euid, cli_conn_t *client)
{
	/*LINTED*/
	uid_t		zeus_euid;
	cli_res_cmds_t	*cmdsp;
	action_op_t	*op;
	response_t 	*resp = client->resp;
	char		*cmd_strp;
	cli_verb_t	*verbp;

	zeus_euid = geteuid();

	if (argc == 0)
		return (-1);	/* no args passed, don't process */

	resp->ret = LDOM_GENERAL_FAILURE;	/* fail by default */

	cmd_strp = argv[0];

	if (autosave_check_spconfig(resp, NULL)) {
		resp->ret = LDOM_GENERAL_FAILURE;
		return (-1);
	}

#if ZEUS_DEBUG_CLI
	if (strcmp(cmd_strp, "debug") == 0) {
		/* the debug CLI is handled as a special case */
		DBG(dump_cli_command(client, cmd_strp, argc - 1, argv + 1));
		resp->ret = ldm_debug(argc - 1, (void **)(argv + 1), client) ?
		    LDOM_OK : LDOM_GENERAL_FAILURE;
		return (-1);
	}
#endif
	if (!cli_break_command(cmd_strp, &verbp, &cmdsp, &op)) {
		/* TRANSLATION_NOTE Don't translate '--help' */
		rprintf(resp, gettext(
		    "Unknown command %s; use --help option for list of "
		    "available commands\n"), cmd_strp);
		return (-1);
	}

	/* Authorization check if not simulating */
	if (zeus_euid == 0) {
		if (auth_check_euid(euid, verbp->cmd) != 0) {
			resp->ret = LDOM_SECURITY_FAILURE;
			rprintf(resp, gettext("Authorization failed\n"));
			return (-1);
		}
	}

	DBG(dump_cli_command(client, cmd_strp, argc - 1, argv + 1));

	switch (op->funcp(argc - 1, &(argv[1]), client)) {
	case CLI_RESULT_OK:
		resp->ret = LDOM_OK;
		return (0);
	case CLI_RESULT_FAIL:
		resp->ret = LDOM_GENERAL_FAILURE;
		return (-1);
	case CLI_RESULT_USAGE:
		rprintf(resp, "\n%s\n\tldm %s %s\n\n", gettext("Usage:"),
		    cmd_strp, op->helpstrp);
		resp->ret = LDOM_GENERAL_FAILURE;
		return (-1);
	}

	return (-1);
}

/*
 * LDom Manager banner
 */

static void
display_banner(response_t *resp)
{
	rprintf(resp, "\nLogical Domain Manager (v "LDM_VERSION")\n");
	backend.banner(resp);

	rprintf(resp, "\n");
	pri_display_firmware(resp);

#if ZDBG
	rprintf(resp, "\nDefines:\n");
	rprintf(resp, "    CONFIG_MODE=%d\n", CONFIG_MODE);
	rprintf(resp, "    BIND_CHECK=%d\n", BIND_CHECK);

	/* display the client protocol versions (CLI, CONTROL, MIG etc) */
	dump_protocol_versions(resp);
#endif
}

/*
 * CLI command parser
 */

int
ldm_cli_command(cli_conn_t *client)
{
	int		argc, rc;
	char		**argv;
	uid_t		euid = socket_peer_euid(client->fd);
	response_t	*resp = client->resp;

	if (!deserialize_cmd(client->bufp, &argc, &argv, resp)) {
		resp->ret = -1;
		return (0);
	}

	/* Note: all exits from this point *must* go through cleanup. */
	resp->ret = 0;
	if (argc == 0) {
		cli_help(0, NULL, resp);
		goto cleanup;
	}

	if (strcmp(argv[0], "-V") == 0 || strcmp(argv[0], "--version") == 0) {
		display_banner(resp);
		goto cleanup;
	}

	if (strcmp(argv[0], "-?") == 0 || strcmp(argv[0], "--help") == 0) {
		cli_help(argc, argv, resp);
		goto cleanup;
	}

	/* start audit (must call close_audit_stream() before returning) */
	client->audit_handle = audit_ldm_request_argv(client->fd, argc, argv);

	rc = cli_parser(argc, argv, euid, client);

	if ((client->class == LDM_CLASS_MIG) && (rc == 0)) {
		/*
		 * If this is a migration class CLI command which has
		 * sucessfully run so far (cli_parser() returned 0) then it
		 * will continue processing in the background. The value 1 is
		 * returned to indicate to the caller that the command is not
		 * finished.  Once the command finishes in the background it
		 * will close the audit stream.
		 */
		cleanup_args(argc, argv);
		return (1);
	}

	close_audit_stream(client->audit_handle, resp->ret);

cleanup:
	cleanup_args(argc, argv);
	return (0);
}
