/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines and tables for the 'list-devices', 'list-bindings',
 * 'list-services', and 'list-constraints' operations.
 */

#include <string.h>

#include <libintl.h>

#include "basics.h"
#include "chain.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "core_md.h"
#include "dispctl.h"
#include "errors.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "response.h"
#include "support.h"
#include "vio.h"
#include "xml_output.h"
#include "zpri.h"


cli_res_cmds_t	cli_devices_cmds = {
	.resinfop = "devices",
	.list_op =	{
		cli_devices_list,
		"[-a] [-p] [cpu] [crypto|mau] [memory] [io]"
	},
};

cli_res_cmds_t	cli_bindings_cmds = {
	.resinfop = "bindings",
	.list_op =	{
		cli_bindings_list,
		"[-e] [-p] [<ldom>...]"
	},
};

cli_res_cmds_t	cli_services_cmds = {
	.resinfop = "services",
	.list_op =	{
		cli_services_list,
		"[-e] [-p] [<ldom>...]"
	},
};

cli_res_cmds_t	cli_constraints_cmds = {
	.resinfop = "constraints",
	.list_op =	{
		cli_constraints_list,
		"([-x] | [-e] [-p]) [<ldom>...]"
	},
};


/*
 * Constraint listing
 * Arguments take two forms:
 * 1.	-x | --xml <ldom>
 * 2.	<ldom>*
 */

cli_result_t
cli_constraints_list(int argc, char **argvp, cli_conn_t *clientp)
{
	int i;
	ldom_t *ldp;
	response_t *resp = clientp->resp;
	dispctl_t dctl;

	dctl_init(&dctl);
	dctl.constraints = 1;

	/*
	 * We look for the special XML case separately
	 */

	if (argc >= 1 && (strcmp(argvp[0], "-x") == 0 ||
	    strcmp(argvp[0], "--xml") == 0)) {
		for (i = 1; i < argc && argvp[i][0] == '-'; i++) {
			if (strcmp(argvp[i], "-p") == 0 ||
			    strcmp(argvp[i], "--parseable") == 0) {
				return (CLI_RESULT_USAGE);
			} else if (strcmp(argvp[i], "-e") == 0 ||
			    strcmp(argvp[i], "--extended") == 0) {
				return (CLI_RESULT_USAGE);
			}
		}
		xml_output_constraints(argc - 1, argvp + 1, resp);
		return (CLI_RESULT_OK);
	}

	for (i = 0; i < argc && argvp[i][0] == '-'; i++) {
		if (strcmp(argvp[i], "-p") == 0 ||
		    strcmp(argvp[i], "--parseable") == 0) {
			dctl.parseable = 1;
		} else if (strcmp(argvp[i], "-e") == 0 ||
		    strcmp(argvp[i], "--extended") == 0) {
			dctl.extended = 1;
		} else if (strcmp(argvp[i], "-?") == 0 ||
		    strcmp(argvp[i], "--help") == 0) {
			return (CLI_RESULT_USAGE);
		} else if (strcmp(argvp[i], "--") == 0) {
			i++;
			break;
		} else {
			return (CLI_RESULT_USAGE);
		}
	}

	config_mode_info_warning(resp, &dctl);
	output_version(resp, &dctl);

	if (i == argc) {
		/* first list domains which are not inactive */
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			if (ldp->state >= BOUND)
				print_bsc(ldp, resp, &dctl);
		}
		/* now list inactive domains */
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			if (ldp->state < BOUND)
				print_bsc(ldp, resp, &dctl);
		}
	} else {
		/* walk through the LDoms specified on the command line */
		for (; i < argc; i++) {
			if ((ldp = ldm_get_ldp(resp, argvp[i])) == NULL)
				return (CLI_RESULT_FAIL);
			print_bsc(ldp, resp, &dctl);
		}
	}

	return (CLI_RESULT_OK);
}

static cli_result_t
list_bindings_or_services(int argc, char **argv, response_t *resp,
    dispctl_t *dctl)
{
	int i;
	ldom_t *ldp;

	/*
	 * parse command line
	 */
	for (i = 0; i < argc && argv[i][0] == '-'; i++) {
		if (strcmp(argv[i], "-p") == 0 ||
		    strcmp(argv[i], "--parseable") == 0) {
			dctl->parseable = 1;
		} else if (strcmp(argv[i], "-e") == 0 ||
		    strcmp(argv[i], "--extended") == 0) {
			dctl->extended = 1;
			dctl->longf = 1;
		} else if (strcmp(argv[i], "-?") == 0 ||
		    strcmp(argv[i], "--help") == 0) {
			return (CLI_RESULT_USAGE);
		} else if (strcmp(argv[i], "--") == 0) {
			i++;
			break;
		} else {
			return (CLI_RESULT_USAGE);
		}
	}

	config_mode_info_warning(resp, dctl);
	output_version(resp, dctl);

	if (i == argc) {
		if (dctl->services) {
			print_group_vio(resp, dctl);
			return (CLI_RESULT_OK);
		}
		/* first list domains which are not inactive */
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			if (ldp->state >= BOUND) {
				update_statistics(ldp, true);
				print_bsc(ldp, resp, dctl);
			}
		}
		/* now list inactive domains */
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			if (ldp->state < BOUND)
				print_bsc(ldp, resp, dctl);
		}
	} else {
		/* walk through the LDoms specified on the command line */
		for (; i < argc; i++) {
			if ((ldp = ldm_get_ldp(resp, (char *)argv[i])) == NULL)
				return (CLI_RESULT_FAIL);
			update_statistics(ldp, true);
			print_bsc(ldp, resp, dctl);
		}
	}

	return (CLI_RESULT_OK);
}

cli_result_t
cli_bindings_list(int argc, char **argv,  cli_conn_t *client)
{
	dispctl_t dctl;

	dctl_init(&dctl);
	dctl.bindings = 1;
	dctl.longf = 1;

	return (list_bindings_or_services(argc, argv, client->resp, &dctl));
}

cli_result_t
cli_services_list(int argc, char **argv,  cli_conn_t *client)
{
	dispctl_t dctl;

	dctl_init(&dctl);
	dctl.services = 1;
	dctl.longf = 1;

	return (list_bindings_or_services(argc, argv, client->resp, &dctl));
}

static void
ldm_show_cpu(response_t *resp, dispctl_t *dctl)
{
	pri_cpu_t *pcpu_p;
	char *fmt;
	int frac_free;
	bool_t display = false;

	mark_response(resp);

	if (dctl->parseable) {
		rprintf(resp, "%s\n", dctl_label("vcpu", dctl));
		if (dctl->all)
			fmt = "|pid=%d|free=%d|pm=%s\n";
		else
			fmt = "|pid=%d|free=%d\n";
	} else {
		rprintf(resp, "%s\n", dctl_label(gettext("vcpu"), dctl));
		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in a table with
		 * fixed column widths.
		 * WIDTH  HEADER	MEANING
		 * 7	  "PID"		physical processor id
		 * 7	  "%FREE"	percentage of processor unallocated
		 *	  "PM"		Is resource being power managed?
		 */
		if (dctl->all) {
			rprintf(resp, "    %-7s %-7s %s\n", gettext("PID"),
			    gettext("%FREE"), gettext("PM"));
			fmt = "    %-7d %-7d %s\n";
		} else {
			/*
			 * No PM column if only reporting on
			 * unallocated resources.
			 */
			rprintf(resp, "    %-7s %s\n", gettext("PID"),
			    gettext("%FREE"));
			fmt = "    %-7d %d\n";
		}
	}

	CHAIN_ITER(pri.cpu_list, pcpu_p, link) {
		frac_free = pcpu_p->bound.fraction_free;
		if (dctl->all) {
			char *pm_str;

			/*
			 * Internal PM state not valid on
			 * unallocated resources.
			 */
			if (frac_free == PRI_CPU_FRAC_DENOM) {
				if (dctl->parseable)
					pm_str = "";
				else
					pm_str = "---";
			} else {
				pm_str = (get_core_md_cpu_status(pcpu_p->
				    phys_id, CORE_MD_PM_STATUS) ==
				    PM_MANAGED ? MSG_YES_I18N : MSG_NO_I18N);
			}

			rprintf(resp, fmt, pcpu_p->phys_id,
			    frac_free * 100 / PRI_CPU_FRAC_DENOM, pm_str);

			display = true;
		} else if (frac_free > 0) {
			rprintf(resp, fmt, pcpu_p->phys_id,
			    frac_free * 100 / PRI_CPU_FRAC_DENOM);

			display = true;
		}
	}

	if (display) {
		if (!dctl->parseable)
			rprintf(resp, "\n");
	} else {
		reset_response(resp);
	}
}

static void
ldm_show_mau(response_t *resp, dispctl_t *dctl)
{
	pri_mau_t *mau;
	ldom_t *ldp;
	char *maus, *fmt, *fmtb;
	bool_t display = false;

	mark_response(resp);

	if (dctl->parseable) {
		rprintf(resp, "%s\n", dctl_label("mau", dctl));
		fmt = "|id=%d|cpuset=%s\n";
		fmtb = "|id=%d|cpuset=%s|bound=%s\n";
	} else {
		rprintf(resp, "%s\n", dctl_label(gettext("mau"), dctl));
		fmt = "    %-6d %s\n";
		fmtb = "    %-6d %-39s %s\n";
		if (dctl->all)
			rprintf(resp, "    %-6s %-39s %s\n",
			    gettext("ID"), gettext("CPUSET"), gettext("BOUND"));
		else
			rprintf(resp, "    %-6s %s\n",
			    gettext("ID"), gettext("CPUSET"));
	}

	CHAIN_ITER(pri.mau_list, mau, link) {
		ldp = get_mau_ldom(mau); 	/* NULL if MAU is not bound */
		if (dctl->all || ldp == NULL) {
			maus = dctl->parseable ? mau->parse_repr : mau->repr;
			if (dctl->all && ldp != NULL)
				rprintf(resp, fmtb, mau->mau_id, maus,
				    ldp->namep);
			else
				rprintf(resp, fmt, mau->mau_id, maus);
			display = true;
		}
	}

	if (display) {
		if (!dctl->parseable)
			rprintf(resp, "\n");
	} else {
		reset_response(resp);
	}
}

static int
bmblk_pa_cmp(mblock_t *a, mblock_t *b)
{
	if (a->paddr < b->paddr)
		return (-1);
	else if (a->paddr == b->paddr)
		return (0);
	else
		return (1);
}

static void
ldm_show_mem(response_t *resp, dispctl_t *dctl)
{
	ldom_t *ldp;
	mblock_t *mblk, *iter, *prev;
	char *size, *fmt, *bound;
	CHAIN(mblock_t, memory);

	CHAIN_INIT(memory);

	/* add free memory blocks to display list */
	CHAIN_ITER(free_memory.blocks, mblk, pblk) {
		mblk->is_free = true;
		CHAIN_ADD_SORTED(memory, mblk, iter, prev, blk, bmblk_pa_cmp);
	}

	if (dctl->all) {
		/* add memory blocks used by system to display list */
		CHAIN_ITER(system_memory.blocks, mblk, sys) {
			mblk->is_free = false;
			CHAIN_ADD_SORTED(memory, mblk, iter, prev, blk,
			    bmblk_pa_cmp);
		}
		/* add memory blocks bound to ldoms to display list */
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
				mblk->is_free = false;
				CHAIN_ADD_SORTED(memory, mblk, iter, prev, blk,
				    bmblk_pa_cmp);
			}
		}
	}

	if (CHAIN_LENGTH(memory) == 0)
		return;			/* nothing to display */

	if (dctl->parseable) {
		rprintf(resp, "%s\n", dctl_label("memory", dctl));
		fmt = "|pa=0x%"PRIx64"|size=%"PRIu64"";
	} else {
		rprintf(resp, "%s\n", dctl_label(gettext("memory"), dctl));
		fmt = "    0x%-18"PRIx64" %-15s";
		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in a table with
		 * fixed column widths.
		 * WIDTH  HEADER	MEANING
		 * 20	  "PA"		physical memory address
		 * 15	  "SIZE"	amount of memory
		 *	  "BOUND"	the domain the memory is allocated
		 *			(bound) to
		 */
		if (dctl->all)
			rprintf(resp, "    %-20s %-15s %s\n",
			    gettext("PA"), gettext("SIZE"), gettext("BOUND"));
		else
			rprintf(resp, "    %-20s %s\n", gettext("PA"),
			    gettext("SIZE"));
	}

	CHAIN_ITER(memory, mblk, blk) {
		if (dctl->parseable) {
			rprintf(resp, fmt, mblk->paddr, mblk->size);
		} else {
			size = conv_size(mblk->size);
			rprintf(resp, fmt, mblk->paddr, size);
			Xfree(size);
		}
		if (mblk->is_free) {
			rprintf(resp, "\n");
		} else {
			if (mblk->ldp == NULL)
				bound = "_sys_";
			else
				bound = mblk->ldp->namep;

			if (dctl->parseable)
				rprintf(resp, "|bound=%s\n", bound);
			else
				rprintf(resp, " %s\n", bound);
		}
	}

	if (!dctl->parseable)
		rprintf(resp, "\n");
}

static void
ldm_show_io(response_t *resp, dispctl_t *dctl)
{
	pri_iodevice_t *pio_p;
	pri_pcie_dev_t *pdev;
	char *fmt, *fmtb, *yes;
	bool_t display = false;

	mark_response(resp);

	if (dctl->parseable) {
		rprintf(resp, "%s\n", dctl_label("IO", dctl));
		fmt = "|dev=%s|alias=%s";
		fmtb = "|dev=%s|alias=%s|bound=%s";
		yes = MSG_YES;
	} else {
		rprintf(resp, "%s\n", dctl_label(gettext("IO"), dctl));
		fmt = "    "FIO" %s";
		fmtb = "    "FIO" "FIO" %-7s";
		yes = MSG_YES_I18N;
		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in a table
		 * with fixed column widths.
		 * WIDTH  HEADER	MEANING
		 * 16	  "DEVICE"	name of the IO device
		 * 16	  "PSEUDONYM"	alternate name for the device
		 * 7	  "BOUND"	whether device is bound to a domain
		 *	  "OPTIONS"	device options
		 */
		if (dctl->all)
			rprintf(resp, "    "FIO" "FIO" %-7s %s\n",
			    gettext("DEVICE"), gettext("PSEUDONYM"),
			    gettext("BOUND"), gettext("OPTIONS"));
		else
			rprintf(resp, "    "FIO" %s\n",
			    gettext("DEVICE"), gettext("PSEUDONYM"));
	}

	CHAIN_ITER(pri.iodevice_list, pio_p, link) {
		char *pseudonym;

		if (pio_p->cli_name == NULL)
			pseudonym = "";
		else
			pseudonym = pio_p->cli_name;

		if (pio_p->free) {
			if (dctl->all)
				rprintf(resp, fmtb, pio_p->name, pseudonym,
				    "", "");
			else
				rprintf(resp, fmt, pio_p->name, pseudonym);
			rprintf(resp, "\n");
			display = true;

		} else if (dctl->all) {
			rprintf(resp, fmtb, pio_p->name, pseudonym, yes);
			if (CHAIN_START(pio_p->bound_ios)->io_mmu_bypass) {
				if (dctl->parseable)
					rprintf(resp, "|opt=bypass=on");
				else
					rprintf(resp, " bypass=on");
			}
			rprintf(resp, "\n");
			display = true;
		}
	}

	if (zeus.sdio_enabled) {
		CHAIN_ITER(pri.iodevice_list, pio_p, link) {
			CHAIN_ITER(pio_p->dev_list, pdev, link) {
				if (pdev->free) {
					if (dctl->all)
						rprintf(resp, fmtb, pdev->name,
						    pdev->cli_name, "");
					else
						rprintf(resp, fmt, pdev->name,
						    pdev->cli_name);
					rprintf(resp, "\n");
					display = true;
				} else if (dctl->all) {
					rprintf(resp, fmtb,
					    pdev->name, pdev->cli_name, yes);
					rprintf(resp, "\n");
					display = true;
				}
			}
		}
	}

	if (display) {
		if (!dctl->parseable)
			rprintf(resp, "\n");
	} else {
		reset_response(resp);
	}
}

cli_result_t
cli_devices_list(int argc, char **argv, cli_conn_t *client)
{
	dispctl_t dctl;
	bool_t show_cpu = false;
	bool_t show_mau = false;
	bool_t show_mem = false;
	bool_t show_io = false;
	int i;
	response_t *resp = client->resp;

	dctl_init(&dctl);

	/*
	 * parse command line
	 */
	for (i = 0; i < argc && argv[i][0] == '-'; i++) {
		if (strcmp(argv[i], "-p") == 0 ||
		    strcmp(argv[i], "--parseable") == 0) {
			dctl.parseable = 1;
		} else if (strcmp(argv[i], "-a") == 0 ||
		    strcmp(argv[i], "--all") == 0) {
			dctl.all = 1;
		} else if (strcmp(argv[i], "-?") == 0 ||
		    strcmp(argv[i], "--help") == 0) {
			return (CLI_RESULT_USAGE);
		} else if (strcmp(argv[i], "--") == 0) {
			break;
		} else {
			return (CLI_RESULT_USAGE);
		}
	}

	if (i == argc) {
		/* no devices specified hence show all devices */
		show_cpu = true;
		show_mau = true;
		show_mem = true;
		show_io = true;
	} else {
		while (i < argc) {
			if (strcmp(argv[i], "cpu") == 0 ||
			    strcmp(argv[i], "vcpu") == 0) {
				show_cpu = true;
			} else if (strcmp(argv[i], "crypto") == 0 ||
			    strcmp(argv[i], "mau") == 0) {
				show_mau = true;
			} else if (strcmp(argv[i], "mem") == 0 ||
			    strcmp(argv[i], "memory") == 0) {
				show_mem = true;
			} else if (strcmp(argv[i], "io") == 0) {
				show_io = true;
			} else {
				rprintf(resp, gettext("Unknown device type: "
				    "%s\n"), argv[i]);
				return (CLI_RESULT_FAIL);
			}
			i++;
		}
	}

	config_mode_info_warning(resp, &dctl);
	output_version(resp, &dctl);

	if (show_cpu)
		ldm_show_cpu(resp, &dctl);

	if (show_mau)
		ldm_show_mau(resp, &dctl);

	if (show_mem)
		ldm_show_mem(resp, &dctl);

	if (show_io)
		ldm_show_io(resp, &dctl);

	return (CLI_RESULT_OK);
}
