/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines and tables for operations on the 'memory' resource
 * type.
 */

#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>

#include <libintl.h>

#include "allocate.h"
#include "basics.h"
#include "cli_conn.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "database.h"
#include "dr.h"
#include "fatal.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "option.h"
#include "response.h"
#include "support.h"
#include "memory.h"
#include "xml_cntrl_common.h"
#include "xml_event.h"
#include "errors.h"
#include "dr_mem.h"


cli_res_cmds_t	cli_memory_cmds = {
	.resinfop = "memory",
	.add_op =	{ cli_mem_add, "(--unaligned primary | <number>[GMK]) "
		"<ldom>" },
	.set_op =	{ cli_mem_set,	"[--auto-adj] <number>[GMK] <ldom>" },
	.remove_op =	{ cli_mem_remove, "<number>[GMK] <ldom>" },
};

/*
 * Routine for add, delete or modify of memory.
 */
static cli_result_t
mem_adm(int argc, char **argv, cli_conn_t *client, int action)
{
	response_t *resp = client->resp;
	ldom_t *ldp;
	uint64_t curr, amt;
	xmlChar *event;
	bool_t add_mem;
	bool_t auto_adj = false;
	char	*arg_size;
	uint64_t amount;

	if (argc == 0)
		return (CLI_RESULT_USAGE);

	if (strcmp(argv[0], "-?") == 0 ||
	    strcmp(argv[0], "--help") == 0)
		return (CLI_RESULT_USAGE);

	if (action == ACT_SET && strcmp(argv[0], "--auto-adj") == 0) {
		if (argc != 3)
			return (CLI_RESULT_USAGE);
		arg_size = argv[1];
		auto_adj = true;
	} else {
		arg_size = argv[0];
	}

	if ((ldp = ldm_get_ldp(resp, argv[--argc])) == NULL)
		return (CLI_RESULT_FAIL);


	if (argc != 1 && auto_adj == false)
		/* should only be a number left */
		return (CLI_RESULT_USAGE);

	if (!parse_mem_size(resp, &amt, arg_size, &(ldp->mem_dr_state.unit)))
		return (CLI_RESULT_FAIL);

	ldp->mem_dr_state.auto_adj_size = 0;

	switch (action) {
	case ACT_SET:
		event = XML_SET_MEM;
		curr = (ldp->state < BOUND) ?
		    ldp->required.size_mem : ldom_size_bound_mem(ldp);

		add_mem = add_res(curr, &amt);

		/* check alignment */
		amount = amt % MEM_MINSIZE;
		if (amount != 0) {
			/* not aligned */
			if (auto_adj == true) {
				/*
				 * auto adjustment - increase if add-mem
				 * decrease if remove-mem
				 */
				if (add_mem) {
					amount = MEM_MINSIZE - amount;
					amt += amount;
				} else {
					amt -= amount;
				}

				ldp->mem_dr_state.auto_adj_size = amount;

				ASSERT(amt % MEM_MINSIZE == 0);
			} else {
				amount = amt - amount;
				if (BYTES2MB(amount) == 0)
					amount = MEM_MINSIZE;

				rprintf(resp, MSG_ERR_MEM_DR_SET_MEM_ALIGN,
				    argv[0], (add_mem ? "add" : "remove"),
				    BYTES2MB(amt), BYTES2MB(MEM_MINSIZE),
				    (add_mem ? "add" : "remove"),
				    BYTES2MB(amount), BYTES2MB(MEM_MINSIZE));

				return (CLI_RESULT_FAIL);
			}
		}

		if (add_mem) {
			/* need to add */
			goto add_resource;
		} else {
			/* need to delete */
			goto remove_resource;
		}

	case ACT_REMOVE:
		event = XML_REMOVE_MEM;
remove_resource:
		if (config_mod_allowed(ldp, DEL_RECONFIG_REMOVE_PHYS,
		    DR_OP_TYPE_MEM, resp) == false)
			return (CLI_RESULT_FAIL);

		if (amt == 0)
			/* requested removing nothing, so we're done! */
			return (CLI_RESULT_OK);

		ldp->mem_dr_state.event = event;
		if (!mem_del(ldp, amt, client)) {
			if (action == ACT_REMOVE)
				rprintf(resp, MSG_ERR_MEM_DR_REMOVE_FAILED);
			else
				rprintf(resp, MSG_ERR_MEM_DR_MODIFY_FAILED);

			return (CLI_RESULT_FAIL);
		}
		break;

	case ACT_ADD:
		event = XML_ADD_MEM;
add_resource:
		if (config_mod_allowed(ldp, DEL_RECONFIG_ADD,
		    DR_OP_TYPE_MEM, resp) == false)
			return (CLI_RESULT_FAIL);

		if (amt == 0)
			/* requested adding nothing, so we're done! */
			return (CLI_RESULT_OK);

		ldp->mem_dr_state.event = event;
		if (!mem_add(ldp, amt, client)) {
			if (action == ACT_ADD)
				rprintf(resp, MSG_ERR_MEM_DR_ADD_FAILED);
			else
				rprintf(resp, MSG_ERR_MEM_DR_MODIFY_FAILED);

			return (CLI_RESULT_FAIL);
		}
		break;

	default:
		fatal("%s: unknown action %d\n", __func__, action);
	}

	/*
	 * Send event notification and save database only if the domain is
	 * inactive or in delayed reconfig. If domain is active and not in
	 * delayed reconfig, there will be a memory DR thread in progress.
	 * Both sending event notification and saving database will be done in
	 * that thread.
	 */
	if (ldp->state != ACTIVE || in_delayed_reconfig(NULL) == true) {

		/* print out message if size of memory was adjusted */
		if (ldp->mem_dr_state.auto_adj_size > 0) {
			rprintf(resp, MSG_WARN_MEMORY_CONSTRAINS, ldp->namep,
			    BYTES2MB(amount));
		}

		event_send_res_event(event, ldp, NULL, NULL, client);
		database_save();
	}

	return (CLI_RESULT_OK);
}

cli_result_t
cli_mem_add(int argc, char **argv, cli_conn_t *client)
{
	if (argc == 0)
		return (CLI_RESULT_USAGE);

	/*
	 * Assign unaligned memory to primary domain
	 */
	if (strcmp(argv[0], "--unaligned") == 0) {
		ldom_t *ldp;

		if (argc != 2)
			return (CLI_RESULT_USAGE);

		if ((ldp = ldm_get_ldp(client->resp, argv[1])) == NULL)
			return (CLI_RESULT_FAIL);

		if (IS_CONTROL_DOMAIN(ldp) != true) {
			rprintf(client->resp,
			    MSG_ERR_MEM_UNALIGN_MEM_CONTROL_DOMAIN);
			return (CLI_RESULT_USAGE);
		}

		if (mem_reassign_unaligned(client->resp))
			return (CLI_RESULT_OK);

		return (CLI_RESULT_FAIL);
	}

	return (mem_adm(argc, argv, client, ACT_ADD));
}

cli_result_t
cli_mem_remove(int argc, char **argv, cli_conn_t *client)
{
	return (mem_adm(argc, argv, client, ACT_REMOVE));
}

cli_result_t
cli_mem_set(int argc, char **argv, cli_conn_t *client)
{
	return (mem_adm(argc, argv, client, ACT_SET));
}
