/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines for the 'ldm migrate-domain' operation.  This command
 * nominally belongs with other 'domain' operations, but is separate
 * to keep the size of that file manageable.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <thread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "allocate.h"
#include "cli_conn.h"
#include "cli_util.h"
#include "errors.h"
#include "fatal.h"
#include "ldm_protocol.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "mig.h"
#include "mig_be.h"
#include "mig_class.h"
#include "mig_common.h"
#include "option.h"
#include "response.h"
#include "ssl_connection.h"
#include "xmpp_common.h"
#include "zprintf.h"

void mig_process_cancel(mig_seq_args_t *mig);
bool_t check_for_cancel(ldom_t *ldp);

static int
mig_negotiate_version(mig_conn_t *mconn, uint16_t major, uint16_t minor)
{
	int msg_len;
	ldm_hdr_t *hdr;
	ldm_init_req_t *init_req;
	ldm_init_ack_t *init_ack;
	char *buf;

	if (mconn == NULL)
		return (-1);

	msg_len = sizeof (ldm_hdr_t) + sizeof (ldm_init_req_t);
	buf = Xmalloc(msg_len);

	hdr = (ldm_hdr_t *)buf;
	hdr->msg_type = hton32(LDM_INIT_REQ);
	hdr->payload_len = hton32(sizeof (ldm_init_req_t));
	init_req = (ldm_init_req_t *)(buf + sizeof (ldm_hdr_t));
	init_req->client_class = hton32(LDM_CLASS_MIG);
	init_req->major_vers = hton16(major);
	init_req->minor_vers = hton16(minor);

	DBG(zprintf("%s: Negotiating ver (%d, %d)\n", __func__, major, minor));

	if (mig_write_msg(mconn, buf, msg_len, MIG_HANDSHAKE_TIMEOUT) == -1) {
		ewarning("Failed to send version negotiation to logical "
		    "domain manager");
		Xfree(buf);
		mig_close_conn(mconn);
		return (-1);
	}

	Xfree(buf);

	if (mig_read_msg(mconn, MIG_HANDSHAKE_TIMEOUT) == -1) {
		ewarning("Failed to receive version negotiation response "
		    "from logical domain manager");
		mig_close_conn(mconn);
		return (-1);
	}

	switch (mconn->hdr.msg_type) {

	case LDM_INIT_ACK:
		DBG(zprintf("%s: got LDM_INIT_ACK\n", __func__));
		init_ack = (ldm_init_ack_t *)mconn->read.buf;
		mconn->major_vers = major;
		mconn->minor_vers = ntoh16(init_ack->minor_vers);
		DBG(zprintf("%s: Negotiated ver (%d, %d)\n", __func__,
		    mconn->major_vers, mconn->minor_vers));
		ASSERT(mig_check_version(major, mconn->minor_vers));
		break;

	case LDM_INIT_NACK:
		DBG(zprintf("%s: got LDM_INIT_NACK\n", __func__));
		mig_close_conn(mconn);
		return (-1);

	case LDM_ERROR:
		DBG(zprintf("%s: got LDM_ERROR\n", __func__));
		mig_close_conn(mconn);
		return (-1);

	default:
		warning("%s: Unknown response message type (%d)", __func__,
		    mconn->hdr.msg_type);
		mig_close_conn(mconn);
		return (-1);
	}

	Zfree(mconn->read.buf);

	return (0);
}

static bool_t
mig_secure_client_conn(mig_conn_t *mconn)
{
	int	rv;

#if ZDBG
	/* nothing to do if connection is not secure */
	if (!option.secure_migration)
		return (true);
#endif

	mconn->ssl = SSL_new_f(server_ssl_ctx);
	if (mconn->ssl == NULL) {
		DBG(zprintf("Failure in SSL_new for migration\n"));
		return (false);
	}

	rv = SSL_set_fd_f(mconn->ssl, mconn->fd);
	if (rv == 0) {
		DBG(zprintf("Failed to set the file descriptor in "
		    "the new SSL handle\n"));
		return (false);
	}

	rv = SSL_accept_f(mconn->ssl);
	if (rv != 1) {
		DBG_BLOCK {
			int		ssl_err;
			unsigned long	err_err;

			ssl_err = SSL_get_error_f(mconn->ssl, rv);
			err_err = ERR_get_error_f();

			zprintf("SSL ACCEPT failed, ssl_err=%d %s\n", ssl_err,
			    ERR_error_string_f(err_err, NULL));
		}
		return (false);
	}

	DBG(zprintf("Outgoing migration request secured\n"));

	return (true);
}

static mig_conn_t *
connect_to_peer_ldmd(char *host, int port)
{
	mig_conn_t	*mconn;
	int		sock;
	int		rv;

	DBG(zprintf("Connecting to peer ldmd on port %d\n", port));

	sock = connect_to_host_port(host, port, XMPP_CONNECT_TIMEOUT);
	if (sock == -1) {
		DBG(zprintf("Unable to establish connection\n"));
		return (NULL);
	}

	mconn = mig_new_connection(sock);

	if (!mig_secure_client_conn(mconn)) {
		mig_close_conn(mconn);
		return (NULL);
	}

	set_sock_nonblocking(mconn->fd);

	rv = mig_negotiate_version(mconn, LDM_MIG_MAJOR_VER, LDM_MIG_MINOR_VER);
	if (rv != 0) {
		mig_close_conn(mconn);
		return (NULL);
	}

	return (mconn);
}

static mig_conn_t *
establish_connection(char *host, response_t *resp)
{
	mig_conn_t	*mconn;
	int		port = MIG_DEFAULT_PORT;

	DBG(zprintf("\nEstablishing connection with host %s...\n", host));

	/* attempt to connect to the standard migration port */
	mconn = connect_to_peer_ldmd(host, port);

	/*
	 * If the standard port is not up, try the legacy port.
	 * This will cover the case where the target is running
	 * an older version of the LDom manager. This legacy
	 * support should be removed at some point.
	 */
	if (mconn == NULL) {
		DBG(zprintf("WARNING: unable to connect on standard "
		    "port, attempting legacy port...\n"));
		port = MIG_LEGACY_PORT;
		mconn = connect_to_peer_ldmd(host, port);
	}

	/* there really is no one listening */
	if (mconn == NULL) {
		DBG(zprintf("Unable to establish connection\n\n"));

		rprintf(resp, MSG_ERR_MIG_CONN_FAILED, host);
		rprintf(resp, "\n");
		return (NULL);
	}

	DBG(zprintf("Connection established on port %d\n\n", port));

	return (mconn);
}

static int
verify_cookie(mig_seq_args_t *args, void *cookie)
{
	int		rc;
	int		msg_len;
	ldm_hdr_t	*hdr;
	mig_msg_t	*mig_msg;
	uint32_t	status;
	char		*buf;
	mig_conn_t	*mconn = args->mconn;
	response_t	*resp = args->resp;

	if (cookie == NULL) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, gettext("Received empty cookie from target "
		    "LDoms Manager\n"));
		return (-1);
	}

	msg_len = MIG_MSG_HDR_LEN + MIG_COOKIE_LEN;
	buf = Xmalloc(msg_len);

	hdr = (ldm_hdr_t *)buf;
	hdr->msg_type = hton32(LDM_MIG_MSG);
	hdr->payload_len = hton32(sizeof (mig_msg_t) + MIG_COOKIE_LEN);

	mig_msg = (mig_msg_t *)(buf + sizeof (ldm_hdr_t));
	mig_msg->type = hton32(MIG_AUTH_COOKIE_VERIFY);
	mig_msg->len = hton32(MIG_COOKIE_LEN);
	mig_msg->magic = hton32(MIG_MAGIC);
	strcpy(mig_msg->data, (char *)cookie);

	if (mig_write_msg(mconn, buf, msg_len, MIG_HANDSHAKE_TIMEOUT) == -1) {
		args->status = MIG_STATUS_FAILURE;
		ewarning("Failed to send cookie verification request to "
		    "logical domain manager");
		Xfree(buf);
		return (-1);
	}

	Xfree(buf);

	if (mig_read_mig_msg(mconn, NULL, MIG_MAX_MSG_LEN,
	    MIG_HANDSHAKE_TIMEOUT) == -1) {
		args->status = MIG_STATUS_FAILURE;
		ewarning("Failed to receive cookie verification response "
		    "from logical domain manager");
		return (-1);
	}

	mig_msg = (mig_msg_t *)mconn->read.buf;
	status = hton32(mig_msg->status);
	if (status == MIG_STATUS_SUCCESS) {
		rc = 0;
	} else if (status == MIG_STATUS_FAILURE) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, MSG_ERR_MIG_AUTH_FAILED);
		rc = -1;
	} else {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, gettext("Invalid response to remote "
		    "authorization request\n"));
		rc = -1;
	}

	Zfree(mconn->read.buf);

	return (rc);
}

static int
migrate_init(mig_seq_args_t *args)
{
	mig_msg_init_info_t *msg_init;
	mig_msg_init_info_v1_1_t *msg_init_v1_1;
	mig_msg_t	*mig_msg;
	char		src_hostname[MAXHOSTNAMELEN];
	int		msg_len;
	int		vlen;
	uint8_t		*bufp;
	char		*tgt_namep;
	size_t		namelen;
	mig_conn_t	*mconn = args->mconn;
	ldom_t		*ldp = args->ldp;

	tgt_namep = (args->tgt_ldom == NULL) ? ldp->namep : args->tgt_ldom;
	namelen = strlen(tgt_namep) + 1;

	if (gethostname(src_hostname, MAXHOSTNAMELEN) != 0) {
		ewarning("Failed to gethostname()");
		return (-1);
	}

	/* calculate the max variable data length */
	if (is_mig_init_extended(mconn)) {
		vlen = (int)(sizeof (mig_msg_init_info_t) + namelen);
	} else if (args->force == true) {
		vlen = (int)(sizeof (mig_msg_init_info_v1_1_t) + namelen);
	} else {
		/* prevent loss of extended data */
		rprintf(args->client->resp, gettext("The target machine does "
		    "not support UUID. Use -f to migrate\nthe domain and "
		    "discard the UUID property.\n"));
		return (-1);
	}
	msg_len = MIG_MSG_HDR_LEN + vlen;

	bufp = init_mig_msg(MIG_INIT_INFO, vlen, MIG_STATUS_SUCCESS);

	/* fill in the message specific information */
	if (is_mig_init_extended(mconn)) {
		msg_init = (mig_msg_init_info_t *)(bufp + MIG_MSG_HDR_LEN);
		msg_init->dryrun = hton32(args->dry_run);
		msg_init->final_state = hton32(ldp->state);
		memcpy(msg_init->src_hostname, src_hostname,
		    strlen(src_hostname) + 1);
		memcpy(msg_init->tgt_name, tgt_namep, namelen);
		memcpy(msg_init->uuid, args->ldp->props.uuid,
		    UUID_PRINTABLE_STRING_LENGTH);
	} else {
		ASSERT(args->force == true);
		msg_init_v1_1 = (mig_msg_init_info_v1_1_t *)(bufp +
		    MIG_MSG_HDR_LEN);
		msg_init_v1_1->dryrun = hton32(args->dry_run);
		msg_init_v1_1->final_state = hton32(ldp->state);
		memcpy(msg_init_v1_1->src_hostname, src_hostname,
		    strlen(src_hostname) + 1);
		memcpy(msg_init_v1_1->tgt_name, tgt_namep, namelen);
	}

	/* debugging output */
	mig_msg = (mig_msg_t *)(bufp + sizeof (ldm_hdr_t));
	DUMP_MIG_MSG(mig_msg, "OUT");

	/* send the message */
	if (mig_write_msg(mconn, bufp, msg_len, MIG_HANDSHAKE_TIMEOUT) == -1) {
		ewarning("Failed to send migration init message");
		Xfree(bufp);
		return (-1);
	}

	Xfree(bufp);

	return (0);
}

/*
 * Trigger the local feasibility check on the target before we attempt
 * to migrate a domain across.
 */
static bool_t
do_local_feasibility_check_on_tgt(mig_seq_args_t *args)
{
	if (send_mig_msg_resp(args, MIG_CHECK_INIT, MIG_STATUS_SUCCESS) != 0)
		return (false);

	/* wait for response from target */
	if (recv_mig_msg_resp(args, MIG_CHECK_INIT) == false)
		return (false);

	return (true);
}

static bool_t
init_source_mig_cache(mig_seq_args_t *args, char *src_ldom, char *host)
{
	ldom_t *ldp;

	ASSERT(args != NULL);
	ASSERT(src_ldom != NULL);
	ASSERT(host != NULL);

	ldp = args->ldp;
	ldp->mig_info.total_mem = 0;
	ldp->mig_info.migrated_mem = 0;
	ldp->mig_info.progress = 0;	/* initialise to 0% */

	/*
	 * verify if the target name was explicitly specified (if not the
	 * guest will be called the same name on the new host).
	 *
	 * now that we have the host and the target_name values
	 * we generate the name to be displayed in the status output.
	 *
	 */
	if (args->tgt_ldom == NULL) {
		args->tgt_ldom = Xstrdup(src_ldom);
		ldp->mig_info.peer_name = Xstrdup(host);
	} else {
		if (!is_valid_object_name(args->resp, args->tgt_ldom)) {
			rprintf(args->resp, MSG_ERR_ILLEGAL_LDOM_NAME,
			    args->tgt_ldom);
			return (false);
		}

		/*
		 * XXXMIG - we could detect if a rename is necessary
		 * and not add the tgt_ldom if we wanted.
		 */
		ldp->mig_info.peer_name =
		    Xmalloc(strlen(host) + 1 + strlen(args->tgt_ldom) + 1);
		sprintf(ldp->mig_info.peer_name, "%s:%s", host,
		    args->tgt_ldom);
	}

	return (true);
}

/*
 * Lookup and cache the IP address of the interface over which the
 * migration is occurring. This information is added to MAC address
 * collision responses so that the target can filter collisions from
 * the source domain.
 */
static bool_t
get_local_ip_addr(mig_seq_args_t *args)
{
	char			*ip;
	struct sockaddr_in	*sin;
	struct sockaddr		sock_info;
	socklen_t		sock_info_len = sizeof (struct sockaddr);
	mig_conn_t		*mconn = args->mconn;

	/* lookup the local IP address for this connection */
	if (getsockname(mconn->fd, &sock_info, &sock_info_len) != 0) {
		DBG(zprintf("Failed the getsockname() call\n"));
		return (false);
	}

	if (sock_info_len == 0) {
		ewarning("Failed to get IP address of migration connection");
		return (false);
	}

	/* convert the IP address into a string */
	sin = (struct sockaddr_in *)&sock_info;
	ip = Xmalloc(INET_ADDRSTRLEN);
	if (inet_ntop(AF_INET, &sin->sin_addr, ip, INET_ADDRSTRLEN) == NULL) {
		DBG(zprintf("%s: inet_ntop() failed: %s\n", __func__,
		    strerror(errno)));
		Xfree(ip);
		return (false);
	}

	/* save the IP for MAC collision responses */
	zeus.local_ip = ip;

	DBG(zprintf("Local IP address: %s\n", zeus.local_ip));

	return (true);
}

static mig_src_ops_t *
mig_src_ops_init(mig_seq_args_t *args)
{
	/* check for cold migration */
	if (IS_MIG_COLD(args))
		return (&mig_src_cold_ops);

	return (&mig_src_warm_ops);
}

cli_result_t
cli_ldom_migrate(int argc, char **argv, cli_conn_t *client)
{
	int		rc;
	mig_seq_args_t	*args = NULL;
	mig_src_ops_t	*ops;
	char		*resp_str;
	mig_arg_t	cli_args = { 0 };
	bool_t		do_not_unlock_ldom = false;

	/* parse the command line arguments */
	rc = migrate_parse_options(argc, argv, &cli_args);

	if (rc == -2) {
		return (CLI_RESULT_USAGE);
	} else if (rc != 0) {
		efatal("%s", __func__);
	}

	ASSERT(cli_args.passfile == NULL);
	ASSERT(cli_args.src_ldom != NULL);
	ASSERT(cli_args.host != NULL);

	/*
	 * Initialize a structure that holds all information
	 * needed to sequence the migration in progress.
	 */
	args = Xcalloc(1, mig_seq_args_t);
	init_response(&args->resp);
	args->status = MIG_STATUS_SUCCESS;
	args->dry_run = cli_args.dry_run;
	args->force = cli_args.force;
	args->client = client;
	args->tgt_ldom = Xstrdup(cli_args.tgt_ldom);
	args->peer_hostname = Xstrdup(cli_args.host);

	/* lookup the source ldom */
	args->ldp = ldm_get_ldp(args->resp, cli_args.src_ldom);
	if (args->ldp == NULL)
		goto fail;

	/*
	 * Check that the domain is not already locked, e.g. because it is
	 * currently being migrated onto the machine. If it is locked, we
	 * need to set a flag to indicate to the cleanup code that the LDom
	 * wasn't locked by this operation and the lock shouldn't be cleared
	 * at this time.
	 */
	if (check_locked_ldom(args->ldp, args->resp)) {
		do_not_unlock_ldom = true;
		goto fail;
	}

	/*
	 * Now lock the LDom so that the config can't be changed while we
	 * check to see if a migration is feasible, etc.
	 */
	LOCK_LDOM(args->ldp);

	/* ensure the LDom is in a valid state before migrating */
	if (args->ldp->state != INACTIVE && args->ldp->state != BOUND &&
	    args->ldp->state != ACTIVE) {
		rprintf(args->resp, MSG_ERR_MIG_INVALID_STATE,
		    args->ldp->namep, ldom_state_str(args->ldp->state));
		goto fail;
	}
	args->final_state = args->ldp->state;

	/* initialize mig cache information for the domain */
	if (!init_source_mig_cache(args, cli_args.src_ldom, cli_args.host))
		goto fail;

	args->ldp->flag.migration_source = true;

	/* migration backend determination */
	ops = mig_src_ops_init(args);

	ASSERT(ops != NULL);

	/* perform preliminary checks */
	if (ops->mig_src_init(args) == false)
		goto fail;

	/* Connect to target LDom and verify the cookie */
	args->mconn = establish_connection(cli_args.host, args->resp);
	if (args->mconn == NULL)
		goto fail;

	if (get_local_ip_addr(args) == false)
		goto fail;

	/* send over the cookie to authenticate the user */
	if (verify_cookie(args, client->cookie) != 0)
		goto fail;

	/* send over some preliminary information */
	if (migrate_init(args) != 0)
		goto fail;

	/* verify that the initial feasibility checks succeeded on the tgt */
	if (do_local_feasibility_check_on_tgt(args) == false)
		goto fail;

	/* send static configuration */
	if (ops->mig_src_send_config(args) == false)
		goto fail;

	/* send domain bindings */
	if (ops->mig_src_send_bindings(args) == false)
		goto fail;

	/* perform feasibility checks */
	if (ops->mig_src_feasibility_check(args) == false)
		goto fail;

	/* ensure bind succeeded on target */
	if (ops->mig_src_bind_check(args) == false)
		goto fail;

	/* send run time state */
	if (ops->mig_src_send_state(args) == false)
		goto fail;

	rc = CLI_RESULT_OK;
	goto cleanup;

fail:
	/* failure paths don't go through migration_complete() */
	if (args->ldp != NULL && do_not_unlock_ldom == false)
		args->ldp->flag.migration_source = false;

	/* add any messages into the resp sent to ldm */
	resp_str = get_response_text(args->resp);
	if (resp_str != NULL) {
		rprintf(client->resp, "%s", resp_str);
		Xfree(resp_str);
	}

	if (!IS_MIG_DRY_RUN(args))
		rprintf(client->resp, MSG_ERR_MIG_FAILED, cli_args.src_ldom);

	rc = CLI_RESULT_FAIL;

cleanup:
	/*
	 * If there was a failure, if this is a dry run, or this is
	 * a cold migration, we are finishing the operation. In that
	 * case the connection to the target machine is no longer needed
	 * so we can close it as well as unlocking the LDom.
	 * Otherwise, it is still needed by the background processing thread.
	 */
	if ((rc == CLI_RESULT_FAIL) || IS_MIG_DRY_RUN(args) ||
	    IS_MIG_COLD(args)) {
		if (args->mconn != NULL)
			mig_close_conn(args->mconn);

		if (args->ldp != NULL && do_not_unlock_ldom == false) {
			UNLOCK_LDOM(args->ldp);
		}

		fini_response(args->resp);
		free_mig_seq_args(args);
	}

	Zfree(cli_args.src_ldom);
	Zfree(cli_args.user);
	Zfree(cli_args.host);
	Zfree(cli_args.tgt_ldom);

	return (rc);
}
