/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI utilities for various kinds of parsing operations.
 */

#define	__XPG4_CHAR_CLASS__	/* i18n for isalnum() */
#include <ctype.h>
#include <errno.h>
#include <stdint.h>
#include <string.h>

#include <libintl.h>

#include "allocate.h"
#include "basics.h"
#include "cli_conn.h"
#include "cli_util.h"
#include "clp.h"
#include "dispctl.h"
#include "errors.h"
#include "ldmd_types.h"
#include "ldom.h"
#include "response.h"
#include "support.h"


bool_t
valid_name(response_t *resp, char *name)
{

	/*
	 * The caller needs to check for NULL if needed.
	 */
	if (name == NULL)
		return (false);

	if (strlen(name) > MAXLEN_VALID_NAME) {
		rprintf(resp, MSG_ERR_NAME_TOO_LONG, MAXLEN_VALID_NAME);
		return (false);
	}

	/* check that the first char is a letter, number, or '/' */
	if (!isalnum(*name) && *name != '/') {
		rprintf(resp, MSG_ERR_NAME_ALPHANUM);
		return (false);
	}

	/*
	 * check that the remaining char are letters, numbers, and
	 * punctuation.
	 */
	for (; *name; name++) {
		if (!isalnum(*name) && !ispunct(*name))
			return (false);
	}

	return (true);
}

/*
 * Check the syntactic validity of a variable name. Currently this is very
 * simplistic as no restrictions are placed on variable names other than that
 * be non-empty strings.
 */
bool_t
is_valid_variable(response_t *resp, char *name)
{
	ASSERT(name != NULL);
	if (name == NULL)
		return (false);
	if (strlen(name) < 1)
		return (false);

	/* validate charaters in name */
	return (valid_name(resp, name));
}

bool_t
is_valid_device_name(response_t *resp, char *namep)
{
	char *s;
	int ch;

	if (namep == NULL || namep[0] == '\0')
		return (true);

	if (strlen(namep) > MAXLEN_DEVICE_NAME) {
		rprintf(resp, MSG_ERR_NAME_TOO_LONG, MAXLEN_DEVICE_NAME);
		return (false);
	}

	for (s = namep; (ch = *s) != '\0'; s++)
		if (!isalnum(ch) && !ispunct(ch))
			return (false);

	return (true);
}

bool_t
is_valid_object_name(response_t *resp, char *namep)
{
	char *s;
	int ch;

	if (namep == NULL || namep[0] == '\0') {
		rprintf(resp, MSG_ERR_NAME_TOO_SHORT);
		return (false);
	}

	if (strlen(namep) > MAXLEN_OBJECT_NAME) {
		rprintf(resp, MSG_ERR_NAME_TOO_LONG, MAXLEN_OBJECT_NAME);
		return (false);
	}

	/* First character must be alpha-numeric */
	if (!isalnum(namep[0])) {
		rprintf(resp, MSG_ERR_NAME_ALPHANUM);
		return (false);
	}

	/* Remaining can be al-num or other printables */
	for (s = namep; (ch = *s) != '\0'; s++)
		if (!isalnum(ch) && strchr("-_+#.:;~()", ch) == NULL)
			return (false);

	return (true);
}

ldom_t *
ldm_get_ldp(response_t *resp, char *name)
{
	ldom_t	*ldp = NULL;

	/* Check to see if a name was passed */
	if (name == NULL) {
		if (resp != NULL)
			rprintf(resp, gettext("No LDom name was found\n"));
		return (NULL);
	}

	/* First see if the name is valid */
	if (!is_valid_object_name(resp, name)) {
		if (resp != NULL)
			rprintf(resp, gettext("%s is not a valid LDom name\n"),
			    name);
		return (NULL);
	}

	/* Now see if it exists */
	ldp = find_ldom_by_name(name);
	if (ldp == NULL && resp != NULL)
		rprintf(resp, gettext("LDom \"%s\" was not found\n"), name);

	return (ldp);
}

bool_t
parse_mem_size(response_t *resp, uint64_t *mem_size, char *value, char *unit)
{
	char *argp;
	char *endp = NULL;
	uint64_t size;
	/* Set the max size in Gb for uint_64 */

	errno = 0;
	argp = value;

	if (!isalnum(*value)) {
		rprintf(resp, gettext(
		    "%s is not a valid memory size\n"), value);
		return (false);
	}

	size = strtoull(argp, &endp, 0);

	/* endp points to string start on error */
	if (endp == argp || errno != 0)
		goto memerror;

	*unit = ' ';
	switch (strlen(endp)) {
	case 0: break;
	case 1:
		*unit = endp[0];
		switch (endp[0]) {
		/*
		 * Check that the given value with the specified unit
		 * does not overflow 63 bits so that it can be stored
		 * into an uint64_t variable.
		 */
		case 'G':
		case 'g':
			if (size >= (1LL << 32))
				goto memerror;
			size <<= 30;
			break;
		case 'M':
		case 'm':
			if (size >= (1LL << 42))
				goto memerror;
			size <<= 20;
			break;
		case 'K':
		case 'k':
			if (size >= (1LL << 52))
				goto memerror;
			size <<= 10;
			break;
		default:
			goto memerror;
		}
		break;
	default:
		goto memerror;
	}

	*mem_size = size;
	return (true);

memerror:
	rprintf(resp, gettext("Invalid memory size %s\n"), value);
	return (false);
}

bool_t
parse_num(response_t *resp, uint64_t *num, char *value)
{
	uint64_t r;
	char *err;

	errno = 0;

	/*
	 * validate the characters received.
	 */
	if (valid_name(resp, value) == false) {
		rprintf(resp, gettext(
		    "%s is not a valid number\n"), value);
		return (false);
	}

	r = strtoll(value, &err, 0);

	if (*err != '\0' || errno != 0) {
		rprintf(resp, gettext("Invalid amount specified: %s\n"),
		    value);
		return (false);
	}

	*num = r;
	return (true);
}

bool_t
add_res(uint64_t curr, uint64_t *amt)
{
	if (curr > *amt) {
		/* need to do a remove, get the diff and return false */
		*amt = curr - *amt;
		return (false);
	} else {
		/* need to do an add, get the diff and return true */
		*amt -= curr;
		return (true);
	}
}

int
get_comma_sep_list_len(char *slist)
{
	int n = 1;

	while ((slist = strchr(slist, ',')) != NULL) {
		n++;
		slist++;
	}

	return (n);
}

static bool_t
check_no_duplicate(int *list, int n)
{
	int *checklist;
	int i;

	checklist = Xmalloc(n * sizeof (int));
	memcpy(checklist, list, n * sizeof (int));

	sort_int_array(list, n);

	for (i = 1; i < n; i++) {
		if (checklist[i-1] == checklist[i])
			break;
	}

	Xfree(checklist);
	return (i < n ? false : true);
}

int
get_comma_sep_list(char *slist, int *list, int n)
{
	char *s;
	int i;

	slist = Xstrdup(slist);

	s = strtok(slist, ",");
	for (i = 0; (i < n) && (s != NULL); i++) {
		char *err;

		errno = 0;
		list[i] = strtoll(s, &err, 0);

		if (*err != NULL || errno != 0)
			return (-1);

		s = strtok(NULL, ",");
	}

	Xfree(slist);

	if (i != n || s != NULL)
		return (-1);

	if (!check_no_duplicate(list, n))
		return (-1);

	return (0);
}
