/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines and tables for operations on the 'policy' resource
 * type.
 */

#include <string.h>
#include <stdint.h>
#include <stdlib.h>

#include <libintl.h>
#include <macros.h>
#include <tzfile.h>

#include "allocate.h"
#include "basics.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "database.h"
#include "dr.h"
#include "errors.h"
#include "ldmd_types.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "pm/pm_shared.h"
#include "policy.h"
#include "response.h"
#include "tod.h"
#include "vcpu.h"
#include "vio.h"
#include "xml_event.h"


cli_res_cmds_t	cli_policy_cmds = {
	.resinfop = "policy",
	.add_op =	{ cli_policy_add,
		"[enable=yes|no] [priority=<value>]\n"
		"\t\t[attack=<value>] [decay=<value>] "
		"[elastic-margin=<value>]\n\t\t[sample-rate=<value>] "
		"[tod-begin=<hh>:<mm>[:<ss>]]\n\t\t[tod-end=<hh>:<mm>[:<ss>]] "
		"[util-lower=<percent>]\n\t\t[util-upper=<percent>] "
		"[vcpu-min=<value>]\n\t\t[vcpu-max=<value>] "
		"name=<policy-name> <ldom>..."
		},
	.set_op	=	{ cli_policy_set,
		"[enable=[yes|no]] [priority=[<value>]]\n"
		"\t\t[attack=[<value>]] [decay=[<value>]] "
		"[elastic-margin=[<value>]]\n\t\t[sample-rate=[<value>]] "
		"[tod-begin=[<hh>:<mm>:<ss>]]\n\t\t[tod-end=[<hh>:<mm>:<ss>]] "
		"[util-lower=[<percent>]]\n\t\t[util-upper=[<percent>]] "
		"[vcpu-min=[<value>]]\n\t\t[vcpu-max=[<value>]] "
		"name=<policy-name> <ldom>..."
		},
	.remove_op = 	{ cli_policy_remove,
		"[name=]<policy-name>... <ldom>" },
};


/*
 * Handles add-policy and set-policy
 */
static cli_result_t
policy_adm(int argc, char **argv, cli_conn_t *client, int action)
{
	response_t	*resp = client->resp;
	ldom_t		*ldp = NULL;
	bool_t		enable = false;
	bool_t		set_enable = false;
	int		priority = INVALID_POLICY_PRIORITY;
	int		vcpu_min = INVALID_POLICY_CPU;
	int		vcpu_max = INVALID_POLICY_CPU;
	int		util_upper = INVALID_POLICY_UTIL;
	int		util_lower = INVALID_POLICY_UTIL;
	uint64_t	tod_begin = INVALID_POLICY_TOD;
	uint64_t	tod_end = INVALID_POLICY_TOD;
	int		sample_rate = INVALID_POLICY_SAMPLE_RATE;
	int		elastic_margin = INVALID_POLICY_ELASTIC_MARGIN;
	int		attack = INVALID_POLICY_ATTACK;
	int		decay = INVALID_POLICY_DECAY;
	char		*policy_name = NULL;
	policy_t	*polp = NULL;
	char		*value = NULL;
	int		nfree_vcpus = 0;
	int		nbound_vcpus = 0;
	int		total_vcpus = 0;
	char		*eos = NULL;
	int		i = 0;
	int		j = 0;

	if (argc == 0)
		return (CLI_RESULT_USAGE);

	nfree_vcpus = num_free_vcpus();
	nbound_vcpus = num_bound_vcpus();
	total_vcpus = nfree_vcpus + nbound_vcpus;

	for (i = 0; i < argc; ) {
		if (strcmp(argv[i], "-?") == 0 ||
		    strcmp(argv[i], "--help") == 0) {
			goto exit_usage;
		}

		if (strncmp(argv[i], "name=", strlen("name=")) == 0) {
			if (policy_name != NULL) {
				rprintf(resp, gettext("Only one policy-name "
				    "can be specified for each policy\n"));
				goto exit_failed;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				rprintf(resp, gettext("Invalid policy name\n"));
				goto exit_failed;
			}
			if (valid_name(resp, value) == false) {
				rprintf(resp, gettext("\"%s\" is not a valid "
				    "policy name\n"), value);
				goto exit_failed;
			}
			policy_name = Xstrdup(value);
			++i;
			continue;
		}

		if (strncmp(argv[i], "enable=", strlen("enable=")) == 0) {
			if (set_enable == true) {
				rprintf(resp, gettext("Only one enable property"
				    " can be specified for each policy\n"));
				goto exit_failed;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				if (action == ACT_ADD) {
					rprintf(resp, gettext("Invalid enable "
					    "property\n"));
					goto exit_failed;
				}
				enable = false;
			} else if (strcmp(value, "yes") == 0) {
				/* PM must be in performance mode */
				if (!pm_migration_is_possible()) {
					rprintf(resp,
					    MSG_ERR_DRM_DISABLED_BY_PM);
					goto exit_failed;
				}
				enable = true;
			} else if (strcmp(value, "no") == 0) {
				enable = false;
			} else {
				rprintf(resp, gettext("Invalid enable property,"
				    " please specify \"yes\" or \"no\"\n"));
				goto exit_failed;
			}
			set_enable = true;
			++i;
			continue;
		}

		if (strncmp(argv[i], "priority=", strlen("priority=")) == 0) {
			if (priority != INVALID_POLICY_PRIORITY) {
				rprintf(resp, gettext("Only one priority can "
				    "be specified for each policy\n"));
				goto exit_failed;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				if (action == ACT_ADD) {
					rprintf(resp, gettext("Invalid "
					    "priority\n"));
					goto exit_failed;
				}
				priority = POLICY_DEFAULT_PRIORITY;
			} else {
				priority = strtol(value, &eos, 10);
				if (eos == NULL || *eos != '\0') {
					rprintf(resp, gettext("priority must "
					    "be an integer\n"));
					goto exit_failed;
				}
				if (priority < 1 || priority > 9999) {
					rprintf(resp, gettext("priority=%d "
					    "cannot be less than 1 nor greater "
					    "than 9999\n"), priority);
					goto exit_failed;
				}
			}
			++i;
			continue;
		}

		if (strncmp(argv[i], "vcpu-min=", strlen("vcpu-min=")) == 0) {
			if (vcpu_min != INVALID_POLICY_CPU) {
				rprintf(resp, gettext("Only one vcpu-min can "
				    "be specified for each policy\n"));
				goto exit_failed;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				if (action == ACT_ADD) {
					rprintf(resp, gettext("Invalid "
					    "vcpu-min\n"));
					goto exit_failed;
				}
				vcpu_min = POLICY_DEFAULT_VCPU_MIN;
			} else {
				vcpu_min = strtol(value, &eos, 10);
				if (eos == NULL || *eos != '\0') {
					rprintf(resp, gettext("vcpu_min must "
					    "be an integer\n"));
					goto exit_failed;
				}
				if (vcpu_min < POLICY_MIN_CPU) {
					rprintf(resp, gettext("vcpu-min "
					    "cannot be less than %d\n"),
					    POLICY_MIN_CPU);
					goto exit_failed;
				} else if (vcpu_min > total_vcpus - 2) {
					rprintf(resp, gettext("vcpu_min "
					    "cannot be greater than the "
					    "total number of vcpus minus "
					    "2 (%d)\n"), total_vcpus - 2);
					goto exit_failed;
				}
				if (vcpu_max != INVALID_POLICY_CPU &&
				    vcpu_min >= vcpu_max &&
				    vcpu_max != POLICY_UNLIMITED_CPU) {
					rprintf(resp, gettext("vcpu-min=%d "
					    "cannot be greater than or equal "
					    "to vcpu-max=%d\n"), vcpu_min,
					    vcpu_max);
					goto exit_failed;
				}
			}
			++i;
			continue;
		}

		if (strncmp(argv[i], "vcpu-max=", strlen("vcpu-max=")) == 0) {
			if (vcpu_max != INVALID_POLICY_CPU) {
				rprintf(resp, gettext("Only one vcpu-max can "
				    "be specified for each policy\n"));
				goto exit_failed;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				if (action == ACT_ADD) {
					rprintf(resp, gettext("Invalid "
					    "vcpu-max\n"));
					goto exit_failed;
				}
				vcpu_max = POLICY_UNLIMITED_CPU;
			} else {
				vcpu_max = strtol(value, &eos, 10);
				if (eos == NULL || *eos != '\0') {
					rprintf(resp, gettext("vcpu_max must "
					    "be an integer\n"));
					goto exit_failed;
				}
				if (vcpu_max < POLICY_MIN_CPU + 1) {
					rprintf(resp, gettext("vcpu-max "
					    "cannot be less than %d\n"),
					    POLICY_MIN_CPU + 1);
					goto exit_failed;
				} else if (vcpu_max > total_vcpus - 1) {
					rprintf(resp, gettext("vcpu_max "
					    "cannot be greater than the "
					    "total number of vcpus minus "
					    "1 (%d)\n"), total_vcpus - 1);
					goto exit_failed;
				}
				if (vcpu_min != INVALID_POLICY_CPU &&
				    vcpu_min >= vcpu_max &&
				    vcpu_max != POLICY_UNLIMITED_CPU) {
					rprintf(resp, gettext("vcpu-min=%d "
					    "cannot be greater than or equal "
					    "to vcpu-max=%d\n"), vcpu_min,
					    vcpu_max);
					goto exit_failed;
				}
			}
			++i;
			continue;
		}

		if (strncmp(argv[i], "util-lower=",
		    strlen("util-lower=")) == 0) {
			if (util_lower != INVALID_POLICY_UTIL) {
				rprintf(resp, gettext("Only one util-lower "
				    "can be specified for each policy\n"));
				goto exit_failed;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				if (action == ACT_ADD) {
					rprintf(resp, gettext("Invalid "
					    "util-lower\n"));
					goto exit_failed;
				}
				util_lower = POLICY_DEFAULT_UTIL_LOWER;
			} else {
				util_lower = strtol(value, &eos, 10);
				if (eos == NULL || *eos != '\0') {
					rprintf(resp, gettext("util_lower must "
					    "be an integer\n"));
					goto exit_failed;
				}
				if (1 > util_lower || util_lower > 98) {
					rprintf(resp, gettext("util-lower "
					    "is a percentage [1-98]\n"));
					goto exit_failed;
				}
				if (util_upper != INVALID_POLICY_UTIL &&
				    util_lower >= util_upper) {
					rprintf(resp, gettext("util-lower=%d "
					    "cannot be greater than or equal "
					    "to util-upper=%d\n"), util_lower,
					    util_upper);
					goto exit_failed;
				}
			}
			++i;
			continue;
		}

		if (strncmp(argv[i], "util-upper=",
		    strlen("util-upper=")) == 0) {
			if (util_upper != INVALID_POLICY_UTIL) {
				rprintf(resp, gettext("Only one util-upper "
				    "can be specified for each policy\n"));
				goto exit_failed;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				if (action == ACT_ADD) {
					rprintf(resp, gettext("Invalid "
					    "util-upper\n"));
					goto exit_failed;
				}
				util_upper = POLICY_DEFAULT_UTIL_UPPER;
			} else {
				util_upper = strtol(value, &eos, 10);
				if (eos == NULL || *eos != '\0') {
					rprintf(resp, gettext("util_upper must "
					    "be an integer\n"));
					goto exit_failed;
				}
				if (2 > util_upper || util_upper > 99) {
					rprintf(resp, gettext("util-upper "
					    "is a percentage [2-99]\n"));
					goto exit_failed;
				}
				if (util_lower != INVALID_POLICY_UTIL &&
				    util_lower >= util_upper) {
					rprintf(resp, gettext("util-lower=%d "
					    "cannot be greater than or equal "
					    "to util-upper=%d\n"), util_lower,
					    util_upper);
					goto exit_failed;
				}
			}
			++i;
			continue;
		}

		if (strncmp(argv[i], "tod-begin=", strlen("tod-begin=")) == 0) {
			if (tod_begin != INVALID_POLICY_TOD) {
				rprintf(resp, gettext("Only one tod-begin "
				    "can be specified for each policy\n"));
				goto exit_failed;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				if (action == ACT_ADD) {
					rprintf(resp, gettext("Invalid time of"
					    " day, please use tod-begin=<hh>:"
					    "<mm>:[<ss>]\n"));
					goto exit_failed;
				}
				tod_begin = POLICY_START_OF_DAY;
			} else {
				tod_begin = str2tod(value, resp);
				if (tod_begin == INVALID_POLICY_TOD) {
					rprintf(resp, gettext("Invalid time of"
					    " day, please use tod-begin=<hh>:"
					    "<mm>:[<ss>]\n"));
					goto exit_failed;
				}
				if (tod_end != INVALID_POLICY_TOD &&
				    tod_begin >= tod_end) {
					BLOCK_ERR_BAD_TOD;
					goto exit_failed;
				}
			}
			++i;
			continue;
		}

		if (strncmp(argv[i], "tod-end=", strlen("tod-end=")) == 0) {
			if (tod_end != INVALID_POLICY_TOD) {
				rprintf(resp, gettext("Only one tod-end "
				    "can be specified for each policy\n"));
				goto exit_failed;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				if (action == ACT_ADD) {
					rprintf(resp, gettext("Invalid time "
					    "of day, please use tod-end=<hh>:"
					    "<mm>:[<ss>]\n"));
					goto exit_failed;
				}
				tod_end = POLICY_END_OF_DAY;
			} else {
				tod_end = str2tod(value, resp);
				if (tod_end == INVALID_POLICY_TOD) {
					rprintf(resp, gettext("Invalid time "
					    "of day, please use tod-end=<hh>:"
					    "<mm>:[<ss>]\n"));
					goto exit_failed;
				}
				if (tod_begin != INVALID_POLICY_TOD &&
				    tod_begin >= tod_end) {
					BLOCK_ERR_BAD_TOD;
					goto exit_failed;
				}
			}
			++i;
			continue;
		}

		if (strncmp(argv[i], "sample-rate=",
		    strlen("sample-rate=")) == 0) {
			if (sample_rate != INVALID_POLICY_SAMPLE_RATE) {
				rprintf(resp, gettext("Only one sample-rate "
				    "can be specified for each policy\n"));
				goto exit_failed;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				if (action == ACT_ADD) {
					rprintf(resp, gettext("Invalid "
					    "sample-rate\n"));
					goto exit_failed;
				}
				sample_rate = POLICY_DEFAULT_SAMPLE_RATE;
			} else {
				sample_rate = strtol(value, &eos, 10);
				if (eos == NULL || *eos != '\0') {
					rprintf(resp, gettext("sample_rate "
					    "must be an integer\n"));
					goto exit_failed;
				}
				if (sample_rate < 1 || sample_rate > 9999) {
					rprintf(resp, gettext("sample-rate=%d "
					    "cannot be less than 1 nor greater "
					    "than 9999\n"), sample_rate);
					goto exit_failed;
				}
			}
			++i;
			continue;
		}

		if (strncmp(argv[i], "elastic-margin=",

		    strlen("elastic-margin=")) == 0) {
			if (elastic_margin != INVALID_POLICY_ELASTIC_MARGIN) {
				rprintf(resp, gettext("Only one elastic-margin "
				    "can be specified for each policy\n"));
				goto exit_failed;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				if (action == ACT_ADD) {
					rprintf(resp, gettext("Invalid "
					    "elastic-margin\n"));
					goto exit_failed;
				}
				elastic_margin = POLICY_DEFAULT_ELASTIC_MARGIN;
			} else {
				elastic_margin = strtol(value, &eos, 10);
				if (eos == NULL || *eos != '\0') {
					rprintf(resp, gettext("elastic_margin "
					    "must be an integer\n"));
					goto exit_failed;
				}
				if (elastic_margin < 0 ||
				    elastic_margin > 100) {
					rprintf(resp, gettext("elastic-margin "
					    "is a percentage [0-100]\n"));
					goto exit_failed;
				}
			}
			++i;
			continue;
		}

		if (strncmp(argv[i], "attack=", strlen("attack=")) == 0) {
			if (attack != INVALID_POLICY_ATTACK) {
				rprintf(resp, gettext("Only one attack "
				    "can be specified for each policy\n"));
				goto exit_failed;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				if (action == ACT_ADD) {
					rprintf(resp, gettext("Invalid "
					    "attack\n"));
					goto exit_failed;
				}
				attack = POLICY_UNLIMITED_ATTACK;
			} else {
				attack = strtol(value, &eos, 10);
				if (eos == NULL || *eos != '\0') {
					rprintf(resp, gettext("attack must "
					    "be an integer\n"));
					goto exit_failed;
				}
				if (attack < 1) {
					rprintf(resp, gettext("attack cannot "
					    "be less than 1\n"));
					goto exit_failed;
				} else if (attack > total_vcpus - 1) {
					rprintf(resp, gettext("attack "
					    "cannot be greater than the "
					    "total number of vcpus minus "
					    "1 (%d)\n"), total_vcpus - 1);
					goto exit_failed;
				}
			}
			++i;
			continue;
		}

		if (strncmp(argv[i], "decay=", strlen("decay=")) == 0) {
			if (decay != INVALID_POLICY_DECAY) {
				rprintf(resp, gettext("Only one decay "
				    "can be specified for each policy\n"));
				goto exit_failed;
			}
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				if (action == ACT_ADD) {
					rprintf(resp, gettext("Invalid "
					    "decay\n"));
					goto exit_failed;
				}
				decay = POLICY_DEFAULT_DECAY;
			} else {
				decay = strtol(value, &eos, 10);
				if (eos == NULL || *eos != '\0') {
					rprintf(resp, gettext("decay must "
					    "be an integer\n"));
					goto exit_failed;
				}
				if (decay < 1) {
					rprintf(resp, gettext("decay "
					    "cannot be less than 1\n"), decay);
					goto exit_failed;
				} else if (decay > total_vcpus - 1) {
					rprintf(resp, gettext("decay "
					    "cannot be greater than the "
					    "total number of vcpus minus "
					    "1 (%d)\n"), total_vcpus - 1);
					goto exit_failed;
				}
			}
			++i;
			continue;
		}

		/*
		 * Everything else must a domain
		 */
		break;
	}

	if (i == argc) {
		rprintf(resp, gettext("At least one domain must be "
		    "specified\n"));
		goto exit_usage;
	}

	value = strchr(argv[i], '=');
	if (value != NULL) {
		rprintf(resp, gettext("\"%s\" does not match any property\n"),
		    argv[i]);
		goto exit_usage;
	}

	if (policy_name == NULL) {
		rprintf(resp, gettext("\"%s\" appears to be a domain "
		    "specification, but at least one\npolicy name must "
		    "be specified first.\n"), argv[i]);
		goto exit_usage;
	}

	/*
	 * Set default values if needed (skip vcpu_max and attack for now)
	 */
	if (action == ACT_ADD) {
		if (priority == INVALID_POLICY_PRIORITY)
			priority = POLICY_DEFAULT_PRIORITY;
		if (vcpu_min == INVALID_POLICY_CPU)
			vcpu_min = POLICY_DEFAULT_VCPU_MIN;
		if (vcpu_max != INVALID_POLICY_CPU && vcpu_min >= vcpu_max &&
		    vcpu_max != POLICY_UNLIMITED_CPU) {
			rprintf(resp, gettext("vcpu-min=%d cannot be greater "
			    "than or equal to vcpu-max=%d\n"), vcpu_min,
			    vcpu_max);
			goto exit_failed;
		}
		if (util_lower == INVALID_POLICY_UTIL)
			util_lower = POLICY_DEFAULT_UTIL_LOWER;
		if (util_upper == INVALID_POLICY_UTIL) {
			util_upper = max((util_lower + 1),
			    POLICY_DEFAULT_UTIL_UPPER);
		} else if (util_lower >= util_upper) {
			rprintf(resp, gettext("util-lower=%d cannot be greater "
			    "than or equal to util-upper=%d\n"), util_lower,
			    util_upper);
			goto exit_failed;
		}
		if (tod_begin == INVALID_POLICY_TOD)
			tod_begin = POLICY_START_OF_DAY;
		if (tod_end == INVALID_POLICY_TOD)
			tod_end = POLICY_END_OF_DAY;
		if (tod_begin >= tod_end) {
			BLOCK_ERR_BAD_TOD;
			goto exit_failed;
		}
		if (sample_rate == INVALID_POLICY_SAMPLE_RATE)
			sample_rate = POLICY_DEFAULT_SAMPLE_RATE;
		if (elastic_margin == INVALID_POLICY_ELASTIC_MARGIN)
			elastic_margin = POLICY_DEFAULT_ELASTIC_MARGIN;
		if (decay == INVALID_POLICY_DECAY)
			decay = POLICY_DEFAULT_DECAY;
	}

	/*
	 * First pass to make sure all domains are valid, available for
	 * modification, and if ACT_SET contains the specified policy or
	 * if ACT_ADD does not already contain the policy.
	 */
	for (j = i; j < argc; ++j) {
		if ((ldp = ldm_get_ldp(resp, argv[j])) == NULL)
			goto exit_failed;
		if (!config_mod_allowed(ldp, DEL_RECONFIG_OTHER,
		    DR_OP_TYPE_DOMAIN, resp))
			goto exit_failed;

		mutex_lock(&ldp->pchain.lock);
		polp = find_policy_by_name(ldp, policy_name);
		switch (action) {
		case ACT_ADD:
			if (polp != NULL) {
				mutex_unlock(&ldp->pchain.lock);
				rprintf(resp, gettext("LDom \"%s\" already "
				    "contains policy \"%s\"\n"), ldp->namep,
				    policy_name);
				goto exit_failed;
			}
			break;
		case ACT_SET:
			if (polp == NULL) {
				mutex_unlock(&ldp->pchain.lock);
				rprintf(resp, gettext("LDom \"%s\" does not "
				    "contain policy \"%s\"\n"), ldp->namep,
				    policy_name);
				goto exit_failed;
			}

			if (vcpu_min != INVALID_POLICY_CPU &&
			    vcpu_max == INVALID_POLICY_CPU) {
				if (polp->vcpu_max != POLICY_UNLIMITED_CPU &&
				    vcpu_min >= polp->vcpu_max) {
					rprintf(resp, gettext("LDom \"%s\" "
					    "vcpu-min=%d cannot be greater "
					    "than or equal to vcpu-max=%d\n"),
					    ldp->namep, vcpu_min,
					    polp->vcpu_max);
					mutex_unlock(&ldp->pchain.lock);
					goto exit_failed;
				}
			}

			if (vcpu_max != INVALID_POLICY_CPU &&
			    vcpu_min == INVALID_POLICY_CPU) {
				if (vcpu_max != POLICY_UNLIMITED_CPU &&
				    vcpu_max <= polp->vcpu_min) {
					rprintf(resp, gettext("LDom \"%s\" "
					    "vcpu-min=%d cannot be greater "
					    "than or equal to vcpu-max=%d\n"),
					    ldp->namep, polp->vcpu_min,
					    vcpu_max);
					mutex_unlock(&ldp->pchain.lock);
					goto exit_failed;
				}
			}

			if (util_lower != INVALID_POLICY_UTIL &&
			    util_upper == INVALID_POLICY_UTIL) {
				if (util_lower >= polp->util_upper) {
					rprintf(resp, gettext("LDom \"%s\" "
					    "util-lower=%d cannot be greater "
					    "than or equal to util-upper=%d\n"),
					    ldp->namep, util_lower,
					    polp->util_upper);
					mutex_unlock(&ldp->pchain.lock);
					goto exit_failed;
				}
			}

			if (util_upper != INVALID_POLICY_UTIL &&
			    util_lower == INVALID_POLICY_UTIL) {
				if (polp->util_lower >= util_upper) {
					rprintf(resp, gettext("LDom \"%s\" "
					    "util-lower=%d cannot be greater "
					    "than or equal to util-upper=%d\n"),
					    ldp->namep, polp->util_lower,
					    util_upper);
					mutex_unlock(&ldp->pchain.lock);
					goto exit_failed;
				}
			}

			if (tod_begin != INVALID_POLICY_TOD &&
			    tod_end == INVALID_POLICY_TOD) {
				if (tod_begin >= polp->tod_end) {
					BLOCK_ERR_BAD_TOD;
					mutex_unlock(&ldp->pchain.lock);
					goto exit_failed;
				}
			}

			if (tod_end != INVALID_POLICY_TOD &&
			    tod_begin == INVALID_POLICY_TOD) {
				if (polp->tod_begin >= tod_end) {
					BLOCK_ERR_BAD_TOD;
					mutex_unlock(&ldp->pchain.lock);
					goto exit_failed;
				}
			}

			break;

		default:
			fatal("%s: invalid action=%d\n", __func__, action);
		}
		mutex_unlock(&ldp->pchain.lock);
	}

	/*
	 * Second pass to add or set policy in all domains.
	 * The last remaining default is set here.
	 */
	for (j = i; j < argc; ++j) {
		ldp = find_ldom_by_name(argv[j]);
		ASSERT(ldp != NULL);
		mutex_lock(&ldp->pchain.lock);

		if (action == ACT_ADD) {		/* ACT_ADD */
			if (CHAIN_LENGTH(ldp->pchain.policy) >
			    MAX_NUM_POLICIES) {
				rprintf(resp, gettext("LDom \"%s\" has reached "
				    "its maximum limit of %d policies\n"),
				    MAX_NUM_POLICIES);
				mutex_unlock(&ldp->pchain.lock);
				continue;
			}
			polp = Xcalloc(1, policy_t);
			polp->ldp = ldp;
			polp->namep = Xstrdup(policy_name);
			polp->tm_stamp = 0;
			CHAIN_LINK_INIT(polp->link);
			CHAIN_ADD(ldp->pchain.policy, polp, link);
			/*
			 * vcpu_max and attack are finally addressed here
			 */
			if (IS_CONTROL_DOMAIN(ldp) || (ldp->state >= BOUND &&
			    is_providing_services(ldp))) {
				/* Primary or service domains */
				if (vcpu_max == INVALID_POLICY_CPU)
					vcpu_max = POLICY_UNLIMITED_CPU;
				if (attack == INVALID_POLICY_ATTACK)
					attack = POLICY_UNLIMITED_ATTACK;
			} else {
				/* Guest domains */
				if (vcpu_max == INVALID_POLICY_CPU)
					vcpu_max = POLICY_DEFAULT_VCPU_MAX;
				if (attack == INVALID_POLICY_ATTACK)
					attack = POLICY_DEFAULT_ATTACK;
			}
		} else {				/* ACT_SET */
			polp = find_policy_by_name(ldp, policy_name);
		}

		if (set_enable || action == ACT_ADD)
			polp->enable = enable;
		if (priority != INVALID_POLICY_PRIORITY)
			polp->priority = priority;
		if (vcpu_min != INVALID_POLICY_CPU)
			polp->vcpu_min = vcpu_min;
		if (vcpu_max != INVALID_POLICY_CPU)
			polp->vcpu_max = vcpu_max;
		if (util_lower != INVALID_POLICY_CPU)
			polp->util_lower = util_lower;
		if (util_upper != INVALID_POLICY_CPU)
			polp->util_upper = util_upper;
		if (tod_begin != INVALID_POLICY_TOD)
			polp->tod_begin = tod_begin;
		if (tod_end != INVALID_POLICY_TOD)
			polp->tod_end = tod_end;
		if (sample_rate != INVALID_POLICY_SAMPLE_RATE)
			polp->sample_rate = sample_rate;
		if (elastic_margin != INVALID_POLICY_ELASTIC_MARGIN)
			polp->elastic_margin = elastic_margin;
		if (attack != INVALID_POLICY_ATTACK)
			polp->attack = attack;
		if (decay != INVALID_POLICY_DECAY)
			polp->decay = decay;

		mutex_unlock(&ldp->pchain.lock);
		database_ldom_set_policy(ldp, polp);
	}

	database_save();
	Xfree(policy_name);
	return (CLI_RESULT_OK);

exit_usage:
	Xfree(policy_name);
	return (CLI_RESULT_USAGE);

exit_failed:
	Xfree(policy_name);
	return (CLI_RESULT_FAIL);
}

cli_result_t
cli_policy_add(int argc, char **argv, cli_conn_t *client)
{
	return (policy_adm(argc, argv, client, ACT_ADD));
}

cli_result_t
cli_policy_set(int argc, char **argv, cli_conn_t *client)
{
	return (policy_adm(argc, argv, client, ACT_SET));
}

cli_result_t
cli_policy_remove(int argc, char **argv, cli_conn_t *client)
{
	response_t	*resp = client->resp;
	ldom_t		*ldp;
	policy_t	*polp;
	char		*value = NULL;
	char		*policy_name;
	int		i;

	if (argc < 2)
		return (CLI_RESULT_USAGE);

	if ((ldp = ldm_get_ldp(resp, argv[argc - 1])) == NULL)
		return (CLI_RESULT_USAGE);

	if (!config_mod_allowed(ldp, DEL_RECONFIG_OTHER,
	    DR_OP_TYPE_DOMAIN, resp))
		return (CLI_RESULT_FAIL);

	/* First pass to make sure all policies can be found */
	for (i = 0; i < argc - 1; ++i) {
		/* Forgive a verbose policy name specification */
		if (strncmp(argv[i], "name=", strlen("name=")) == 0) {
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				rprintf(resp, gettext("Invalid policy-name\n"));
				return (CLI_RESULT_FAIL);
			}
			policy_name = value;
		} else {
			policy_name = argv[i];
		}
		mutex_lock(&ldp->pchain.lock);
		polp = find_policy_by_name(ldp, policy_name);
		if (polp == NULL) {
			rprintf(resp, gettext("LDom \"%s\" does not contain "
			    "policy \"%s\"\n"), ldp->namep, policy_name);
			mutex_unlock(&ldp->pchain.lock);
			return (CLI_RESULT_FAIL);
		}
		mutex_unlock(&ldp->pchain.lock);
	}

	for (i = 0; i < argc - 1; ++i) {
		/* Forgive a verbose policy name specification */
		if (strncmp(argv[i], "name=", strlen("name=")) == 0) {
			value = strchr(argv[i], '=');
			if (*(++value) == '\0') {
				rprintf(resp, gettext("Invalid policy-name\n"));
				return (CLI_RESULT_FAIL);
			}
			policy_name = value;
		} else {
			policy_name = argv[i];
		}
		mutex_lock(&ldp->pchain.lock);
		destroy_one_policy(ldp, policy_name);
		mutex_unlock(&ldp->pchain.lock);
	}

	return (CLI_RESULT_OK);
}
