/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI utilities for printing LDom content for various list operations
 * including "ldm ls" and "ldm ls -l".
 */

#include <limits.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>

#include <libintl.h>
#include <alloca.h>

#include "allocate.h"
#include "basics.h"
#include "chain.h"
#include "cli_util.h"
#include "core_md.h"
#include "dispctl.h"
#include "errors.h"
#include "fatal.h"
#include "hvctl.h"
#include "ldmd_types.h"
#include "ldom.h"
#include "mac_addrs.h"
#include "mac_multicast.h"
#include "memory.h"
#include "nvlist.h"
#include "response.h"
#include "support.h"
#include "variables.h"
#include "vcpu.h"
#include "vio.h"
#include "vmau.h"
#include "xml_cntrl_common.h"


void
output_version(response_t *resp, dispctl_t *dctl)
{
	if (dctl->parseable)
		rprintf(resp, "VERSION %s\n", PARSEABLE_OUTPUT_VERSION);
}

bool_t
print_ldom_variable_list(ldom_t *ldp, response_t *resp, int nvars,
    char **listvars)
{
	variable_t *var;
	int i;

	if (nvars == 0) {
		print_ldom_variables(ldp, resp, NULL);
		return (true);
	}

	for (i = 0; i < nvars; i++) {
		char *s;

		/* verify variable name is valid */
		if (!is_valid_variable(resp, listvars[i])) {
			rprintf(resp,
			    gettext("\"%s\" is not a valid variable name\n"),
			    listvars[i]);
			return (false);
		}

		var = nvlist_find(&ldp->variables, listvars[i]);
		if (var == NULL) {
			rprintf(resp, gettext("%s not found\n"), listvars[i]);
			return (false);
		}
		if (var->name == NULL)
			continue;
		if ((var->value != NULL) &&
		    (strcmp(var->name, "security-password") != 0)) {
			s = var->value;
		} else {
			s = "";
		}
		rprintf(resp, "%s=%s\n", var->name, s);
	}

	return (true);
}

void
print_ldom_variables(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	variables_t *vars;
	char *fmt;

	vars = &ldp->variables;
	if (vars->len == 0)
		return;

	if (dctl != NULL) {
		if (dctl->oformat && !dctl->odomain)
			return;
		if (dctl->parseable) {
			rprintf(resp, "%s\n", dctl_label("variables", dctl));
			fmt = "|%s=%s\n";
		} else {
			rprintf(resp, "%s\n", dctl_label(gettext("variables"),
			    dctl));
			fmt = "    %s=%s\n";
		}
	} else {
		fmt = "%s=%s\n";
	}

	nvlist_print(vars, fmt, resp);

	if ((dctl != NULL) && !dctl->parseable)
		rprintf(resp, "\n");
}


/*
 * Translate LDom state value to a string for listing.
 */
static char *
ldom_state_list(ldom_state_t state)
{
	char *statep;

	/*
	 * TRANSLATION_NOTE
	 * inactive, bound, active, and unknown are part of a fixed column
	 * table and must be less than 11 columns.
	 */
	switch (state) {
	case INACTIVE:
	case BINDING:
		statep = gettext("inactive");
		break;
	case BOUND:
	case UNBINDING:
		statep = gettext("bound");
		break;
	case SUSPENDING:
		statep = gettext("suspending");
		break;
	case RESUMING:
		statep = gettext("resuming");
		break;
	case SUSPENDED:
		statep = gettext("suspended");
		break;
	case STARTING:
	case STOPPING:
	case ACTIVE:
		statep = gettext("active");
		break;
	default:
		statep = gettext("unknown");
		break;
	}

	return (statep);
}

/*
 * The positions of the flags in the short form ldm listing.
 * If the order of the flags is changed make sure to change
 * ldom_flags_long() as well.
 */
enum flagname {
	START_STOP	= 0,
	SYS_IND		= 1,
	DEL_RECONF	= 2,
	CTRL_DOM	= 3,
	SVC_DOM		= 4,
	MIGRATION	= 5,
};

/*
 * Returns the short form string representation of the list output state flags
 * for LDom 'ldp'. This routine uses static storage hence it is not reentrant.
 */
char *
ldom_flags(ldom_t *ldp)
{
	static char flags[9];

	memset(flags, 0, sizeof (flags));

	/* control domain */
	if (IS_CONTROL_DOMAIN(ldp) == true)
		flags[CTRL_DOM] = 'c';
	else
		flags[CTRL_DOM] = '-';

	/* service domain */
	if (ldp->state >= BOUND && is_providing_services(ldp))
		flags[SVC_DOM] = 'v';
	else
		flags[SVC_DOM] = '-';

	/* delayed reconfiguration */
	if (zeus.delayed_reconfig_ldom == ldp)
		flags[DEL_RECONF] = 'd';
	else if (ldp->flag.memdr_in_progress)
		flags[DEL_RECONF] = 'r';
	else
		flags[DEL_RECONF] = '-';

	/* system indicator status */
	flags[SYS_IND] = '-';
	if (ldp->state > BOUND) {
		if (ldp->hv_state.soft == SIS_NORMAL)
			flags[SYS_IND] = 'n';
		else if (ldp->hv_state.soft == SIS_TRANSITION)
			flags[SYS_IND] = 't';
	}

	/* starting or stopping */
	if (ldp->state == STARTING || ldp->state == STOPPING)
		flags[START_STOP] = 's';
	else
		flags[START_STOP] = '-';

	/* migration */
	if (ldp->flag.migration_error)
		flags[MIGRATION] = 'e';
	else if (ldp->flag.migration_source)
		flags[MIGRATION] = 's';
	else if (ldp->flag.migration_target)
		flags[MIGRATION] = 't';
	else
		flags[MIGRATION] = '-';

	return (flags);
}

/*
 * Concatenate 'flag' onto the end of string 's' with a comma separator if
 * 'first' is false. 'size' is the size of storage already preallocated for
 * 's'.
 * FIXME - this is not i18n friendly
 */
/*ARGSUSED2*/
static void
concat(char *s, char *flag, int size, dispctl_t *dctl)
{
	int len;

	len = strlen(s);
	if (len == 0)
		strlcpy(s, flag, size);
	else
		snprintf(s + len, size - len, ",%s", flag);
}

static char *
delayed_reconfig_type_to_str(del_reconfig_type_t type, dispctl_t *dctl)
{
	static char s[32];

	strcpy(s, "");

	if (type == DEL_RECONFIG_LOCKED) {
		concat(s, "locked", sizeof (s), dctl);
	}
	if (type & DEL_RECONFIG_ADD) {
		concat(s, "add", sizeof (s), dctl);
	}
	if (type & DEL_RECONFIG_REMOVE) {
		concat(s, "remove", sizeof (s), dctl);
	}
	if (type & DEL_RECONFIG_REMOVE_2PHASE) {
		concat(s, "remove(2-phase)", sizeof (s), dctl);
	}
	if (type & DEL_RECONFIG_REMOVE_PHYS) {
		concat(s, "remove(phys)", sizeof (s), dctl);
	}
	if (type & DEL_RECONFIG_OTHER) {
		concat(s, "modify", sizeof (s), dctl);
	}
	if (type & DEL_RECONFIG_STOP) {
		concat(s, "stop", sizeof (s), dctl);
	}
	if (type & DEL_RECONFIG_UNKNOWN) {
		concat(s, "unknown", sizeof (s), dctl);
	}
	return (s);
}

/*
 * Returns the long form string representation of the state flags for
 * LDom 'ldp'. This routine uses static storage hence it is not reentrant.
 */
static char *
ldom_flags_long(ldom_t *ldp, dispctl_t *dctl)
{
	static char flags[128];
	char *flg;

	strcpy(flags, "");

	/* starting or stopping */
	if (ldp->state == STARTING || ldp->state == STOPPING) {
		if (ldp->state == STARTING)
			flg = "starting";
		else
			flg = "stopping";

		concat(flags, flg, sizeof (flags), dctl);
	}

	/* system indicator status */
	if (ldp->state > BOUND) {
		if (ldp->hv_state.soft == SIS_NORMAL) {
			flg = "normal";
		} else if (ldp->hv_state.soft == SIS_TRANSITION) {
			flg = "transition";
		} else {
			flg = "unknown";
			warning("Unknown soft state %"PRIu8" for LDom %s",
			    ldp->hv_state.soft, ldp->namep);
		}

		concat(flags, flg, sizeof (flags), dctl);
	}

	/* delayed reconfiguration */
	if (zeus.delayed_reconfig_ldom == ldp) {
		concat(flags, "delayed(", sizeof (flags), dctl);
		strlcat(flags,
		    delayed_reconfig_type_to_str(zeus.delayed_reconfig_type,
		    dctl), sizeof (flags));
		strlcat(flags, ")", sizeof (flags));
	} else if (ldp->flag.memdr_in_progress) {
		concat(flags, "memory-dr", sizeof (flags), dctl);
	}

	/* control domain */
	if (IS_CONTROL_DOMAIN(ldp) == true)
		concat(flags, "control", sizeof (flags), dctl);

	/* service domain */
	if (ldp->state >= BOUND && is_providing_services(ldp))
		concat(flags, "vio-service", sizeof (flags), dctl);

	/* migration */
	if (ldp->flag.migration_error)
		concat(flags, "migration(error)", sizeof (flags), dctl);
	else if (ldp->flag.migration_source)
		concat(flags, "migration(source)", sizeof (flags), dctl);
	else if (ldp->flag.migration_target)
		concat(flags, "migration(target)", sizeof (flags), dctl);

	return (flags);
}

/*
 * Returns a string representation of the current uptime of LDom
 * 'ldp'. This routine uses static storage hence it is not reentrant.
 */
static char *
ldom_uptime(ldom_t *ldp, dispctl_t *dctl)
{
	uint64_t uptime, days, hours, mins;
	static char uptimestr[64];

	if (ldp->state <= BOUND)
		return ("");

	if (dctl->parseable) {
		snprintf(uptimestr, sizeof (uptimestr), "%"PRIu64"",
		    ldp->stats.uptime);
	} else {
		uptime = ldp->stats.uptime + 30;	/* rounding */
		days = uptime / (60 * 60 * 24);
		uptime %= (60 * 60 * 24);
		hours = uptime / (60 * 60);
		uptime %= (60 * 60);
		mins = uptime / 60;

		strcpy(uptimestr, "");

		if (days > 0)
			sprintf(uptimestr, "%"PRIu64"d ", days);
		if (hours > 0)
			sprintf(uptimestr + strlen(uptimestr),
			    "%"PRIu64"h ", hours);
		if ((days < 100) && (mins > 0))
			sprintf(uptimestr + strlen(uptimestr),
			    "%"PRIu64"m", mins);

		if (strlen(uptimestr) == 0)
			sprintf(uptimestr, "%"PRIu64"s", ldp->stats.uptime);
	}

	return (uptimestr);
}

/*
 * Returns a string representation of the guest utilization for domain
 * 'ldp'. This routine uses static storage hence it is not reentrant.
 */
static char *
ldom_guest_util(ldom_t *ldp, dispctl_t *dctl)
{
	static char utilstr[8];

	if (ldp->state > BOUND) {
		if (dctl->parseable) {
			snprintf(utilstr, sizeof (utilstr), "%.1f",
			    ldp->stats.cpu);
		} else {
			if (ldp->stats.cpu >= 10.0)
				snprintf(utilstr, sizeof (utilstr), "%.0f%%",
				    ldp->stats.cpu);
			else
				snprintf(utilstr, sizeof (utilstr), "%.1f%%",
				    ldp->stats.cpu);
		}
	} else {
		strcpy(utilstr, "");
	}

	return (utilstr);
}

/*
 * Returns a string representation of the number of CPUs bound to LDom
 * 'ldp'. This routine uses static storage hence it is not reentrant.
 */
static char *
get_num_cpu(ldom_t *ldp, dispctl_t *dctl)
{
	int ncpu;
	static char ncpustr[INT_STR_LEN];

	if (ldp->state < BOUND || dctl->constraints)
		ncpu = ldp->required.num_vcpus;
	else
		ncpu = ldom_num_bound_vcpus(ldp);

	if (ncpu > 0)
		snprintf(ncpustr, sizeof (ncpustr), "%d", ncpu);
	else
		strcpy(ncpustr, "");

	return (ncpustr);
}

/*
 * Returns a string representation of the current amount of memory bound to
 * LDom 'ldp'. This routine uses static storage hence it is not reentrant.
 */
static char *
get_total_mem(ldom_t *ldp, dispctl_t *dctl)
{
	uint64_t mem;
	static char memstr[UINT64_STR_LEN+2];

	if (ldp->state < BOUND || dctl->constraints)
		mem = ldp->required.size_mem;
	else
		mem = ldom_size_bound_mem(ldp);

	if (mem == 0) {
		strcpy(memstr, "");
	} else if (dctl->parseable) {
		snprintf(memstr, sizeof (memstr), "%"PRIu64"", mem);
	} else {
		char *s = conv_size(mem);
		strlcpy(memstr, s, sizeof (memstr));
		Xfree(s);
	}

	return (memstr);
}

static void
print_list_header(response_t *resp, dispctl_t *dctl)
{
	if (dctl->parseable)
		return;

	/*
	 * TRANSLATION_NOTE
	 * The following are part of a table that has fixed column widths
	 * width	label		meaning
	 * 17		"NAME"		LDom name
	 * 11		"STATE"		State of the LDom (active, bound, etc)
	 * 8		"FLAGS"		single character description of LDom,
	 *				that says if its starting, stopping,
	 *				control domain, service domain, etc.
	 * 6		"UTIL"		percent utilization
	 * 8		"CONS"		console port
	 * 6		"VCPU"		number vcpus
	 * 9		"MEMORY"	amount of memory
	 * 15		"UPTIME"	amount of time LDom has been running
	 */
	rprintf(resp, "%-17s", gettext("NAME"));
	if (dctl->oformat && !dctl->odomain) {
		rprintf(resp, "\n");
		return;
	}
	rprintf(resp, "%-11s", gettext("STATE"));
	rprintf(resp, "%-8s", gettext("FLAGS"));
	if (dctl->oformat) {
		rprintf(resp, "%-6s\n", gettext("UTIL"));
	} else {
		rprintf(resp, "%-8s", gettext("CONS"));
		rprintf(resp, "%-6s", gettext("VCPU"));
		rprintf(resp, "%-9s", gettext("MEMORY"));
		rprintf(resp, "%-6s", gettext("UTIL"));
		rprintf(resp, "%s\n", gettext("UPTIME"));
	}
}

void
printldom(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	if (dctl->first || dctl->longf)
		print_list_header(resp, dctl);

	if (dctl->parseable) {
		rprintf(resp, "%s|", dctl_label("domain", dctl));
		rprintf(resp, "name=%s|", ldp->namep);
		if (dctl->oformat && !dctl->odomain)
			goto done;
		rprintf(resp, "state=%s|", ldom_state_list(ldp->state));
		rprintf(resp, "flags=%s|",
		    dctl->longf ? ldom_flags_long(ldp, dctl) : ldom_flags(ldp));
		if (dctl->oformat) {
			rprintf(resp, "util=%s", ldom_guest_util(ldp, dctl));
			goto done;
		} else {
			rprintf(resp, "cons=%s|", console_print(ldp));
			rprintf(resp, "ncpu=%s|", get_num_cpu(ldp, dctl));
			rprintf(resp, "mem=%s|", get_total_mem(ldp, dctl));
			rprintf(resp, "util=%s|", ldom_guest_util(ldp, dctl));
			rprintf(resp, "uptime=%s", ldom_uptime(ldp, dctl));
		}
	} else {
		rprintf(resp, "%-16s ", ldp->namep);
		if (dctl->oformat && !dctl->odomain)
			goto done;
		rprintf(resp, "%-11s", ldom_state_list(ldp->state));
		rprintf(resp, "%-8s", ldom_flags(ldp));
		if (dctl->oformat) {
			rprintf(resp, "%4s  ", ldom_guest_util(ldp, dctl));
			goto done;
		} else {
			rprintf(resp, "%-8s", console_print(ldp));
			rprintf(resp, "%-6s", get_num_cpu(ldp, dctl));
			rprintf(resp, "%-9s", get_total_mem(ldp, dctl));
			rprintf(resp, "%4s  ", ldom_guest_util(ldp, dctl));
			rprintf(resp, "%s", ldom_uptime(ldp, dctl));
		}
	}

done:
	dctl->first = 0;	/* at least one domain has been output */
}

static void
printldom_long(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	printldom(ldp, resp, dctl);
	if (!dctl->bindings && ldp->state > BOUND) {
		if (dctl->parseable) {
			if (dctl->oformat && !dctl->odomain)
				goto done;
			rprintf(resp, "|softstate=%s", ldp->hv_state.soft_str);
		} else {
			/*
			 * If the domain is in delayed reconfig then display
			 * the long format flags which contain extra
			 * information about the type of delayed reconfig.
			 */
			if (zeus.delayed_reconfig_ldom == ldp)
				rprintf(resp, "\n\n%s\n%s", gettext("FLAGS"),
				    ldom_flags_long(ldp, dctl));
			if (strlen(ldp->hv_state.soft_str) > 0) {
				if (dctl->oformat && !dctl->odomain)
					goto done;
				rprintf(resp, "\n\n%s\n%s",
				    gettext("SOFTSTATE"),
				    ldp->hv_state.soft_str);
			}
		}
	}
done:
	rprintf(resp, "\n");
	if (!dctl->parseable)
		rprintf(resp, "\n");
}

static void
print_migration_status(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	char	*peer = ldp->mig_info.peer_name;

	rprintf(resp, "%s\n", dctl_label("status", dctl));

	if (dctl->parseable) {
		rprintf(resp, "|op=%s", gettext("migration"));
		rprintf(resp, "|progress=%d", ldp->mig_info.progress);

		/*
		 * Always print whether error occurred or not (unlike
		 * human-readable case which only prints when error occurs)
		 */
		rprintf(resp, "|error=%s",
		    ldp->flag.migration_error ?  MSG_YES :  MSG_NO);

		if (ldp->flag.migration_source)
			rprintf(resp, "|target=%s\n", peer != NULL ? peer : "");
		else
			rprintf(resp, "|source=%s\n", peer != NULL ? peer : "");
	} else {
		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in a table with
		 * fixed column widths.
		 * WIDTH  HEADER	MEANING
		 * 9	  "OPERATION"	Operation type in progress
		 * 8	  "PROGRESS"	% progress
		 * 5	  "ERROR"	optional column if error ocurred
		 *
		 * 6	  "SOURCE"	machine on which migration was initiated
		 *	    or
		 * 6	  "TARGET"	machine the domain is being migrated to
		 */
		rprintf(resp, "    %-9s\t%-8s\t",
		    gettext("OPERATION"), gettext("PROGRESS"));
		if (ldp->flag.migration_error)
			rprintf(resp, "%-5s\t", gettext("ERROR"));

		/*
		 * On the source machine, show the target. On the target
		 * machine, show the source.
		 */
		if (ldp->flag.migration_source)
			rprintf(resp, "%-6s\n", gettext("TARGET"));
		else
			rprintf(resp, "%-6s\n", gettext("SOURCE"));

		rprintf(resp, "    %-9s\t%2d%%     \t",
		    gettext("migration"), ldp->mig_info.progress);

		/* conditionaly print column if there is an error */
		if (ldp->flag.migration_error)
			rprintf(resp, "%-5s\t", MSG_YES_I18N);
		rprintf(resp, "%s\n\n", peer != NULL ? peer : "");
	}
}

static void
print_mem_dr_status(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{

	if (!ldp->flag.memdr_in_progress)
		return;

	rprintf(resp, "%s", dctl_label("status", dctl));

	if (dctl->parseable) {
		if (ldp->flag.memdr_in_progress) {
			if (ldp->mem_dr_state.is_mem_add) {
				rprintf(resp, "|reconf=%s|progress=\n",
				    gettext("mem-add"));
			} else {
				rprintf(resp, "|reconf=%s|progress=%"PRId16
				    "\n", gettext("mem-del"),
				    ldp->mem_dr_state.progress);
			}
		} else {
			rprintf(resp, "|reconf=|progress=\n");
		}

	} else {
		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in a table with
		 * fixed column widths.
		 * WIDTH  HEADER	MEANING
		 * 9	  "OPERATION"	Operation type in progress
		 * 8	  "PROGRESS"	% progress
		 */
		if (ldp->mem_dr_state.is_mem_add) {
			rprintf(resp, "\n    %-14s\n", gettext("OPERATION"));

			rprintf(resp, "    %-14s\n", gettext("mem-add"));
		} else {
			rprintf(resp, "\n    %-14s\t%-8s\n",
			    gettext("OPERATION"), gettext("PROGRESS"));

			rprintf(resp, "    %-14s\t%2d%%     \n",
			    gettext("mem-del"), ldp->mem_dr_state.progress);
		}
	}
}

static void
print_status(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	if (dctl->oformat && !dctl->ostatus)
		return;

	if ((ldp->flag.migration_source | ldp->flag.migration_target) != 0)
		print_migration_status(ldp, resp, dctl);

	if (ldp->flag.memdr_in_progress == true)
		print_mem_dr_status(ldp, resp, dctl);

}

static void
print_uuid(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	if (dctl->oformat && !dctl->odomain)
		return;

	rprintf(resp, "%s", dctl_label("uuid", dctl));
	if (dctl->parseable)
		rprintf(resp, "|uuid=%s\n", ldp->props.uuid);
	else
		rprintf(resp, "\n    %s\n\n", ldp->props.uuid);
}

static void
print_mac(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	char *macaddr = NULL;

	if (dctl->oformat && !dctl->onetwork)
		return;

	/*
	 * If the mac address was configured, show it on list constraints
	 */
	if (dctl->constraints) {
		if (is_mac_allocated(ldp->props.mac_address) == false) {
			macaddr = alloca(sizeof (char) * UINT64_STR_LEN);
			mactoa(ldp->props.mac_address, macaddr);
		}
	}

	/*
	 * If list bindings or list -l and LDom is ACTIVE or BOUND
	 * get its mac address
	 */
	if (dctl->bindings || dctl->longf) {
		if ((ldp->state == ACTIVE) || (ldp->state == BOUND)) {
			macaddr = alloca(sizeof (char) * UINT64_STR_LEN);
			mactoa(ldp->props.mac_address, macaddr);
		}
	}

	/*
	 * If we've retrieved a mac address, display it
	 */
	if (macaddr != NULL) {
		rprintf(resp, "%s", dctl_label("mac", dctl));
		if (dctl->parseable)
			rprintf(resp, "|mac-addr=%s\n", macaddr);
		else
			rprintf(resp, "\n    %s\n\n", macaddr);
	}
}

static void
print_hostid(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	char *hostid = NULL;

	if (dctl->oformat && !dctl->odomain)
		return;

	/*
	 * If the hostid was configured, show it on list constraints
	 */
	if (dctl->constraints) {
		if (ldp->props.hostid != UNASSIGNED_HOSTID) {
			hostid = alloca(sizeof (char) * UINT64_STR_LEN);
			sprintf(hostid, "0x%"PRIx64, ldp->props.hostid);
		}
	}

	/*
	 * If list bindings or list -l and LDom is ACTIVE or BOUND
	 * get its hostid
	 */
	if (dctl->bindings || dctl->longf) {
		if ((ldp->state == ACTIVE) || (ldp->state == BOUND)) {
			uint64_t h_id;
			hostid = alloca(sizeof (char) * UINT64_STR_LEN);
			if (ldp->props.hostid != UNASSIGNED_HOSTID)
				h_id = ldp->props.hostid;
			else
				h_id = DERIVED_HOSTID(ldp->props.mac_address);
			sprintf(hostid, "0x%"PRIx64, h_id);
		}
	}

	/*
	 * If we've retrieved a hostid, display it
	 */
	if (hostid != NULL) {
		rprintf(resp, "%s", dctl_label("hostid", dctl));
		if (dctl->parseable)
			rprintf(resp, "|hostid=%s\n", hostid);
		else
			rprintf(resp, "\n    %s\n\n", hostid);

	}
}

static void
print_a_vcpu(vcpu_t *vcpu, bool_t active, dispctl_t *dctl, response_t *resp)
{
	char util[8];
	char *fmt;
	pm_status_t pm_status;

	pm_status = get_core_md_cpu_status(vcpu->cpu->phys_id,
	    CORE_MD_PM_STATUS);

	if (pm_status == PM_MANAGED) {
		if (dctl->parseable)
			strcpy(util, "");
		else
			strcpy(util, "----");
	} else {
		if (active) {
			if (vcpu->util >= 10.0)
				snprintf(util, sizeof (util), "%3.0f%%",
				    vcpu->util);
			else
				snprintf(util, sizeof (util), "%3.1f%%",
				    vcpu->util);
		} else {
			if (dctl->parseable)
				strcpy(util, "0");
			else
				strcpy(util, " ");
		}

	}

	if (dctl->parseable) {
		fmt = "|vid=%d|pid=%d|util=%s|strand=%d\n";
	} else {
		fmt = "    %-6d %-6d %4s %5d%%\n";
	}

	rprintf(resp, fmt, vcpu->vid, vcpu->cpu->phys_id, util,
	    vcpu->fraction * 100 / PRI_CPU_FRAC_DENOM);
}

static void
print_cpu(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	vcpu_t *vcpu;
	bool_t active;
	int ncpu;

	if (dctl->oformat && !dctl->ocpu)
		return;

	if (dctl->constraints) {
		ncpu = ldp->required.num_vcpus;
		if (ncpu == 0)
			return;
		rprintf(resp, "%s", dctl_label("vcpu", dctl));
		if (dctl->parseable)
			rprintf(resp, "|count=%d\n", ncpu);
		else
			rprintf(resp, "\n    %s\n    %d\n",
			    gettext("COUNT"), ncpu);

	} else if (dctl->list || dctl->bindings) {
		ncpu = ldom_num_bound_vcpus(ldp);
		if (ncpu == 0)
			return;
		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in a table with
		 * fixed column widths.
		 * WIDTH  HEADER	MEANING
		 * 6	  "VID"		virtual CPU id
		 * 6	  "PID"		physical CPU id
		 * 4	  "UTIL"	percent CPU utilization
		 * 6	  "STRAND"	percent of strand allocated
		 */
		rprintf(resp, "%s\n", dctl_label("vcpu", dctl));
		if (!dctl->parseable)
			rprintf(resp, "    %-6s %-6s %4s %6s\n",
			    gettext("VID"), gettext("PID"),
			    gettext("UTIL"), gettext("STRAND"));

		active = ldp->state > BOUND ? true : false;

		CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
			print_a_vcpu(vcpu, active, dctl, resp);
		}
	}

	if (!dctl->parseable)
		rprintf(resp, "\n");
}

static void
print_ldom_mau(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	vmau_t *vmau;
	int nmau;

	if (dctl->oformat && !dctl->ocrypto)
		return;

	if ((dctl->list && ldp->state < BOUND) || dctl->constraints) {
		nmau = ldp->required.num_vmaus;
		if (nmau > 0) {
			rprintf(resp, "%s", dctl_label("mau", dctl));
			if (dctl->parseable)
				rprintf(resp, "|count=%d\n", nmau);
			else
				rprintf(resp, "\n    %s\n    %d\n\n",
				    gettext("COUNT"), nmau);
		}
		return;
	}

	if (dctl->list || dctl->bindings) {
		nmau = ldom_num_bound_maus(ldp);
		if (nmau == 0)
			return;
		rprintf(resp, "%s\n", dctl_label("mau", dctl));
		if (!dctl->parseable)
			rprintf(resp, "    %-6s %s\n",
			    gettext("ID"), gettext("CPUSET"));

		CHAIN_ITER(ldp->bound.vmaus, vmau, ldom) {
			if (dctl->parseable)
				rprintf(resp, "|id=%d|cpuset=%s\n",
				    vmau->mau->mau_id, vmau->mau->parse_repr);
			else
				rprintf(resp, "    %-6d %s\n",
				    vmau->mau->mau_id, vmau->mau->repr);
		}

		if (!dctl->parseable)
			rprintf(resp, "\n");
	}
}

static void
print_a_memblk(mblock_t *mblk, response_t *resp, dispctl_t *dctl)
{
	char *sizep;

	if (dctl->parseable) {
		rprintf(resp, "|ra=0x%"PRIx64"|pa=0x%"PRIx64"|size=%"PRIu64"\n",
		    mblk->raddr, mblk->paddr, mblk->size);
	} else {
		rprintf(resp, "    0x%-14"PRIx64" 0x%-14"PRIx64" ",
		    mblk->raddr, mblk->paddr);
		sizep = conv_size(mblk->size);
		rprintf(resp, "%s\n", sizep);
		Xfree(sizep);
	}
}

static int
bmblk_ra_cmp(mblock_t *a, mblock_t *b)
{
	if (a->raddr < b->raddr)
		return (-1);
	else if (a->raddr == b->raddr)
		return (0);
	else
		return (1);
}

static void
print_mem(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	uint64_t size;
	char *sizep;
	mblock_t *mblk, *iter, *prev;
	CHAIN(mblock_t, memory);

	if (dctl->oformat && !dctl->omemory)
		return;

	if (dctl->constraints) {
		size = ldp->required.size_mem;
		if (size == 0)
			return;
		if (dctl->parseable) {
			rprintf(resp, "%s", dctl_label("memory", dctl));
			rprintf(resp, "|size=%"PRIu64"\n", size);
		} else {
			rprintf(resp, "%s",
			    dctl_label(gettext("memory"), dctl));
			sizep = conv_size(size);
			/* TRANSLATION_NOTE SIZE refers to amount of memory */
			rprintf(resp, "\n    %s\n    %s\n\n",
			    gettext("SIZE"), sizep);
			Xfree(sizep);
		}
		return;
	}

	if (dctl->list || dctl->bindings) {
		size = ldom_size_bound_mem(ldp);
		if (size == 0)
			return;
		if (dctl->parseable) {
			rprintf(resp, "%s\n", dctl_label("memory", dctl));
		} else {
			rprintf(resp, "%s\n",
			    dctl_label(gettext("memory"), dctl));
			/*
			 * TRANSLATION_NOTE
			 * The following are column headers in a table with
			 * fixed column widths.
			 * WIDTH  HEADER	MEANING
			 * 16	  "RA"		real memory address
			 * 16	  "PA"		physical memory address
			 * 16	  "SIZE"	amount of memory
			 */
			rprintf(resp, "    %-16s %-16s %-16s\n",
			    gettext("RA"), gettext("PA"), gettext("SIZE"));
		}

		CHAIN_INIT(memory);

		CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
			CHAIN_ADD_SORTED(memory, mblk, iter, prev, blk,
			    bmblk_ra_cmp);
		}

		CHAIN_ITER(memory, mblk, blk) {
			print_a_memblk(mblk, resp, dctl);
		}

		if (!dctl->parseable)
			rprintf(resp, "\n");
	}
}

static void
print_io(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	req_iodevice_t *iodevs_p;
	req_pcie_dev_t *rdev;
	bound_iodevice_t *biodevs_p;
	pri_pcie_dev_t *pdev;
	bound_pcie_dev_t *bdev;
	int ndevs;
	char *fmt, *fmt2;

	/*
	 * Don't print IO when listing services nor when
	 * it should be ignored with (-o) output format
	 */
	if (dctl->services || (dctl->oformat && !dctl->ophysio))
		return;

	if (dctl->parseable) {
		fmt = "|dev=%s|alias=";
		fmt2 = "|dev=%s|alias=%s";
	} else {
		fmt = "    "FIO;
		fmt2 = "    "FIO" "FIO;
	}

	if ((dctl->list && ldp->state < BOUND) || dctl->constraints) {
		ndevs = CHAIN_LENGTH(ldp->required.io_devices) +
		    CHAIN_LENGTH(ldp->required.io_pcie_devs);
		if (ndevs > 0) {
			if (dctl->parseable) {
				rprintf(resp, "%s\n", dctl_label("IO", dctl));
			} else {
				rprintf(resp, "%s\n",
				    dctl_label(gettext("IO"), dctl));
				/*
				 * TRANSLATION_NOTE
				 * The following are column headers in a table
				 * with fixed column widths.
				 * WIDTH  HEADER	MEANING
				 * 16	  "DEVICE"	name of IO device
				 *	  "OPTIONS"	IO device options
				 */
				rprintf(resp, "    "FIO" %s\n",
				    gettext("DEVICE"), gettext("OPTIONS"));
			}

			CHAIN_ITER(ldp->required.io_devices, iodevs_p, link) {
				rprintf(resp, fmt, iodevs_p->name);
				if (iodevs_p->io_mmu_bypass) {
					if (dctl->parseable)
						rprintf(resp,
						    "|opt=bypass=on");
					else
						rprintf(resp, gettext(
						    " bypass=on"));
				}
				rprintf(resp, "\n");
			}

			CHAIN_ITER(ldp->required.io_pcie_devs, rdev, link) {
				rprintf(resp, fmt, rdev->cli_name);
				rprintf(resp, "\n");
			}

			if (!dctl->parseable)
				rprintf(resp, "\n");
		}
	} else if (dctl->list || dctl->bindings) {
		ndevs = CHAIN_LENGTH(ldp->bound.iodevices);
		if (ndevs > 0) {
			if (dctl->parseable) {
				rprintf(resp, "%s\n", dctl_label("IO", dctl));
			} else {
				rprintf(resp, "%s\n",
				    dctl_label(gettext("IO"), dctl));
				/*
				 * TRANSLATION_NOTE
				 * The following are column headers in a table
				 * with fixed column widths.
				 * WIDTH  HEADER	MEANING
				 * 16	  "DEVICE"	name of IO device
				 * 16	  "PSEUDONYM"	alternate IO device name
				 *	  "OPTIONS"	IO device options
				 */
				rprintf(resp, "    "FIO" "FIO" %s\n",
				    gettext("DEVICE"), gettext("PSEUDONYM"),
				    gettext("OPTIONS"));
			}

			CHAIN_ITER(ldp->bound.iodevices, biodevs_p, ldom) {
				pri_iodevice_t *pio_p = biodevs_p->phys_iodev;

				if (biodevs_p->virtual_pcie_bus)
					continue;
				if (biodevs_p->phys_iodev->cli_name == NULL)
					rprintf(resp, fmt, pio_p->name);
				else
					rprintf(resp, fmt2,
					    pio_p->name, pio_p->cli_name);
				if (biodevs_p->io_mmu_bypass) {
					if (dctl->parseable)
						rprintf(resp,
						    "|opt=bypass=on");
					else
						rprintf(resp, gettext(
						    " bypass=on"));
				}
				rprintf(resp, "\n");
			}

			if (!zeus.sdio_enabled) {
				if (!dctl->parseable)
					rprintf(resp, "\n");
				return;
			}

			CHAIN_ITER(ldp->bound.iodevices, biodevs_p, ldom) {
				CHAIN_ITER(biodevs_p->dev_list, bdev, ldom) {
					pdev = bdev->phys_pcie_dev;
					rprintf(resp, fmt2,
					    pdev->name, pdev->cli_name);
					rprintf(resp, "\n");
				}
			}

			if (!dctl->parseable)
				rprintf(resp, "\n");
		}
	}
}

void
print_bsc(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	/* *almost* always print the header and short info */
	if (!dctl->services) {
		if (!dctl->first && !dctl->parseable)
			rprintf(resp, "%s\n", LDM_SEPARATOR);
		if (!dctl->constraints) {
			printldom_long(ldp, resp, dctl);
		} else {
			if (dctl->parseable) {
				rprintf(resp, "%s", dctl_label("domain", dctl));
				rprintf(resp, "|name=%s\n", ldp->namep);
			} else {
				rprintf(resp, "%s",
				    dctl_label(gettext("domain"), dctl));
				rprintf(resp, "\n%s\n\n", ldp->namep);
			}
		}
		print_status(ldp, resp, dctl);
		print_uuid(ldp, resp, dctl);
		print_mac(ldp, resp, dctl);
		print_hostid(ldp, resp, dctl);
		print_failure_policy(ldp, resp, dctl);
		print_master(ldp, resp, dctl);
		print_cpu(ldp, resp, dctl);
		print_ldom_mau(ldp, resp, dctl);
		print_mem(ldp, resp, dctl);
		print_resmgmt(ldp, resp, dctl);
		print_ldom_variables(ldp, resp, dctl);
	}

	/* now finish with IO and VIO */
	print_io(ldp, resp, dctl);
	print_vio(ldp, resp, dctl);

	dctl->first = 0;	/* at least one domain has been output */
}
