/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines and tables for operations on the 'spconfig' resource
 * type.
 */

#include <stddef.h>
#include <string.h>

#include <libxml/tree.h>

#include "autosave.h"
#include "basics.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "ds_impl.h"
#include "errors.h"
#include "ldomconfigure.h"
#include "response.h"
#include "xml_cntrl_common.h"
#include "xml_event.h"


cli_res_cmds_t	cli_spconfig_cmds = {
	.resinfop = "spconfig",
	.add_op =	{ cli_spconfig_add,
		"[-r <autosave>] <config_name>" },
	.set_op	=	{ cli_spconfig_set,	"<config_name>" },
	.remove_op = 	{ cli_spconfig_remove,	"[-r] <config_name>" },
	.list_op =	{ cli_spconfig_list,	"[-r [<autosave>]]" },
};


/*
 * CLI interface for "ldm list-spconfig -r [<mdset_name>]"
 * FIXME - XML support not implemented yet.
 */
static int
autosave_list_cli(char *mdset_name, response_t *resp, xmlNodePtr data)
{
	int i;
	char **mdset_names;
	int nmdsets;
	int ret = CLI_RESULT_OK;

	if (mdset_name != NULL && valid_name(resp, mdset_name) == false) {
		rprintf(resp, MSG_ERR_AUTOSAVE_CONFIG_NAME_INVALID,
		    mdset_name);
		return (CLI_RESULT_FAIL);
	}

	if (autosave_get_mdsets(mdset_name, &nmdsets, &mdset_names) != 0)
		return (CLI_RESULT_FAIL);

	if (data == NULL) {
		for (i = 0; i < nmdsets; i++) {
			rprintf(resp, "%s", mdset_names[i]);
			if (autosave_autosave_is_newer(mdset_names[i]))
				rprintf(resp, " [newer]");
			rprintf(resp, "\n");
		}
		if (nmdsets == 0) {
			if (mdset_name != NULL) {
				rprintf(resp, MSG_ERR_AUTOSAVE_NO_SUCH_CONFIG,
				    mdset_name);
			} else {
				rprintf(resp, MSG_ERR_AUTOSAVE_NO_CONFIGS);
			}
			ret = CLI_RESULT_FAIL;
		}
	}
	autosave_free_list(mdset_names, nmdsets);
	return (ret);
}

/*
 * CLI interface for "ldm add-spconfig -r <mdset_name> [<new_mdset_name>]"
 */
static int
autosave_add_cli(char *mdset_name, char *new_mdset_name, response_t *resp)
{
	int err;

	if (valid_name(resp, mdset_name) == false) {
		rprintf(resp, MSG_ERR_AUTOSAVE_CONFIG_NAME_INVALID,
		    mdset_name);
		return (CLI_RESULT_FAIL);
	}

	if (valid_name(resp, new_mdset_name) == false) {
		rprintf(resp, MSG_ERR_AUTOSAVE_CONFIG_NAME_INVALID,
		    new_mdset_name);
		return (CLI_RESULT_FAIL);
	}

	err = autosave_download_mdset(mdset_name, new_mdset_name, resp);

	return (err == 0 ? CLI_RESULT_OK : CLI_RESULT_FAIL);
}

/*
 * CLI interface for "ldm remove-spconfig -r <mdset_name>"
 */
static int
autosave_remove_cli(char *mdset_name, response_t *resp)
{
	int err;

	if (valid_name(resp, mdset_name) == false) {
		rprintf(resp, MSG_ERR_AUTOSAVE_CONFIG_NAME_INVALID,
		    mdset_name);
		return (CLI_RESULT_FAIL);
	}

	err = autosave_rmdir(mdset_name, true);
	if (err < 0) {
		rprintf(resp, MSG_ERR_AUTOSAVE_CANT_REMOVE, mdset_name);
	} else if (err == 1) {
		rprintf(resp, MSG_ERR_AUTOSAVE_NO_SUCH_CONFIG, mdset_name);
	}
	return (err);
}


typedef int (*mdstore_f)(char *, char *, response_t *, xmlNodePtr);

static cli_result_t
spconfig(int argc, char **argv, char *op, cli_conn_t *client)
{
	mdstore_f mdstore;
	char *setname = NULL;
	char *newname = NULL;
	int ret;
	response_t *resp = client->resp;

	if (strcmp(op, "list") == 0) {
		if (argc == 0) {
			setname = NULL;
		} else if ((argc == 1 || argc == 2) &&
		    strcmp(argv[0], "-r") == 0) {
			if (argc == 2)
				setname = argv[1];
			return (autosave_list_cli(setname, resp, NULL));
		} else {
			return (CLI_RESULT_USAGE);
		}
	} else if (strcmp(op, "download") == 0) {
		if ((argc == 2 || argc == 3) && strcmp(argv[0], "-r") == 0) {
			setname = argv[1];
			if (argc == 3)
				newname = argv[2];
			else
				newname = setname;
			return (autosave_add_cli(setname, newname, resp));
		} else if (argc == 1) {
			if (strcmp(argv[0], "-r") == 0) {
				return (CLI_RESULT_USAGE);
			}
			setname = argv[0];
		} else {
			return (CLI_RESULT_USAGE);
		}
	} else if (strcmp(op, "remove") == 0) {
		if (argc == 2 && strcmp(argv[0], "-r") == 0) {
			return (autosave_remove_cli(argv[1], resp));
		} else if (argc == 1) {
			if (strcmp(argv[0], "-r") == 0) {
				return (CLI_RESULT_USAGE);
			}
			setname = argv[0];
		} else {
			return (CLI_RESULT_USAGE);
		}
	} else {
		/* "set" subcommand */
		if (argc == 1) {
			setname = argv[0];
		} else {
			return (CLI_RESULT_USAGE);
		}
	}

	if (setname != NULL) {
		/* validate characters in set name */
		if (valid_name(resp, setname) == false) {
			rprintf(resp, gettext(
			    "%s is not a valid set name\n"), setname);
			return (CLI_RESULT_FAIL);
		}
	}

	mdstore = (mdstore_f)ds_channel_service_routine(ds_sp_channel(),
	    "mdstore", "mdstore");
	if (mdstore == NULL) {
		rprintf(resp, gettext("The requested operation could not be "
		    "performed because the communication\nchannel between the "
		    "LDom Manager and the system controller is down.\n"
		    "To recover, restart the LDom Manager.\n"));
		return (CLI_RESULT_FAIL);
	}

	ret = (*mdstore)(op, setname, resp, NULL);

	if (ret == 0) {
		if (strcmp(op, "download") == 0) {
			event_send_hardware_event(XML_ADD_SPCONFIG, NULL,
			    setname, client);
		} else if (strcmp(op, "remove") == 0) {
			event_send_hardware_event(XML_REMOVE_SPCONFIG, NULL,
			    setname, client);
		} else if (strcmp(op, "select") == 0) {
			event_send_hardware_event(XML_SET_SPCONFIG, NULL,
			    setname, client);
		}
	}

	return (ret);
}

cli_result_t
cli_spconfig_add(int argc, char **argv, cli_conn_t *client)
{
	return (spconfig(argc, argv, "download", client));
}

cli_result_t
cli_spconfig_remove(int argc, char **argv, cli_conn_t *client)
{
	return (spconfig(argc, argv, "remove", client));
}

cli_result_t
cli_spconfig_set(int argc, char **argv, cli_conn_t *client)
{
	return (spconfig(argc, argv, "select", client));
}

cli_result_t
cli_spconfig_list(int argc, char **argv, cli_conn_t *client)
{
	return (spconfig(argc, argv, "list", client));
}
