/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines and tables for operations on the 'variable' resource
 * type.
 */

#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdint.h>

#include <libintl.h>
#include <libxml/xmlstring.h>

#include "allocate.h"
#include "basics.h"
#include "cli_conn.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "database.h"
#include "dispctl.h"
#include "dr.h"
#include "errors.h"
#include "ldmd_types.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "nvlist.h"
#include "response.h"
#include "sequencer.h"
#include "variables.h"
#include "xml_cntrl_common.h"
#include "xml_event.h"


cli_res_cmds_t	cli_var_cmds = {
	.resinfop = "variable",
	.add_op =	{ cli_var_set,	"<var_name>=<value>... <ldom>" },
	.set_op	=	{ cli_var_set,	"<var_name>=<value>... <ldom>" },
	.remove_op = 	{ cli_var_remove,	"<var_name>... <ldom>" },
	.list_op =	{ cli_var_list,	"[<var_name>...] <ldom>" },
};


static bool_t
var_add(ldom_t *ldp, int nvars, char **vars, cli_conn_t *client)
{
	char *name, *value, *current_value;
	int i;
	bool_t update = false;
	xmlChar *event;
	response_t *resp = client->resp;

	for (i = 0; i < nvars; i++) {
		name = vars[i];
		value = strchr(vars[i], '=');
		if (value != NULL) {
			*value = '\0';	/* overwrite the name/value separator */
			value++;
		}

		if (is_valid_variable(resp, name) == false) {
			rprintf(resp,
			    gettext("\"%s\" is not a valid variable name\n"),
			    name);
			return (false);
		}

		if (value != NULL &&
		    (uint_t)strlen(value) > MAXLEN_VARIABLE_VALUE) {
			rprintf(resp, MSG_ERR_VAR_VALUE_TOO_LONG,
			    MAXLEN_VARIABLE_VALUE);
			return (false);
		}

		current_value = get_ldom_variable(ldp, name);
		if (current_value == NULL) {
			event = XML_ADD_VAR;
		} else {
			event = XML_SET_VAR;
		}
		/*
		 * Variables can be NULL or any value at all so no value
		 * validation is done.
		 */
		if (nvlist_add_pair(&ldp->variables, name, value)) {
			update = true;
			database_ldom_add_variable(ldp, name, value);
			event_send_res_event(event, ldp, NULL, NULL, client);
		}
	}

	if (update) {
		if (ldp->state != INACTIVE) {
			/* mark config as different from SP */
			zeus.is_next_spconfig = false;
			seq_update_vars(ldp, true);
		}
		database_save();
	}

	return (true);
}

static bool_t
var_del(ldom_t *ldp, int nvars, char **vars, cli_conn_t *client)
{
	bool_t update = false;
	int i;
	response_t *resp = client->resp;

	for (i = 0; i < nvars; i++) {
		if (is_valid_variable(resp, vars[i]) == false) {
			rprintf(resp,
			    gettext("\"%s\" is not a valid variable name\n"),
			    vars[i]);
			break;
		}

		if (del_ldom_variable(ldp, vars[i])) {
			update = true;
			database_ldom_del_variable(ldp, vars[i]);
			event_send_res_event(XML_REMOVE_VAR, ldp, NULL, NULL,
			    client);
		} else {
			rprintf(resp, gettext("%s not found\n"), vars[i]);
			break;
		}
	}

	if (i < nvars) {
		return (false);		/* broke out of loop owing to error */
	} else {
		if (update) {
			if (ldp->state != INACTIVE) {
				/* mark config as different from SP */
				zeus.is_next_spconfig = false;
				seq_update_vars(ldp, true);
			}
			database_save();
		}

		return (true);
	}
}

/*
 * ldm add/set-var <var_name>=<value>... <ldom>
 */

cli_result_t
cli_var_set(int argc, char **argvp, cli_conn_t *clientp)
{
	response_t *resp = clientp->resp;
	ldom_t *ldp = NULL;
	char *ldom_namep;

	if (argc < 2)
		return (CLI_RESULT_USAGE);

	ldom_namep = argvp[argc-1];

	if ((ldp = ldm_get_ldp(resp, ldom_namep)) == NULL)
		return (CLI_RESULT_FAIL);

	if (!config_mod_allowed(ldp, DEL_RECONFIG_OTHER, DR_OP_TYPE_VAR, resp))
		return (CLI_RESULT_FAIL);

	if (!var_add(ldp, argc-1, argvp, clientp))
		return (CLI_RESULT_FAIL);

	return (CLI_RESULT_OK);
}

cli_result_t
cli_var_remove(int argc, char **argvp, cli_conn_t *clientp)
{
	response_t *resp = clientp->resp;
	ldom_t *ldp = NULL;
	char *ldom_namep;

	if (argc < 2)
		return (CLI_RESULT_USAGE);

	ldom_namep = argvp[argc-1];

	if ((ldp = ldm_get_ldp(resp, ldom_namep)) == NULL)
		return (CLI_RESULT_FAIL);

	if (!config_mod_allowed(ldp, DEL_RECONFIG_OTHER, DR_OP_TYPE_VAR, resp))
		return (CLI_RESULT_FAIL);

	if (!var_del(ldp, argc-1, argvp, clientp))
		return (CLI_RESULT_FAIL);

	return (CLI_RESULT_OK);
}

cli_result_t
cli_var_list(int argc, char **argvp, cli_conn_t *clientp)
{
	response_t *resp = clientp->resp;
	ldom_t *ldp;
	char *ldom_namep;

	if (argc < 1)
		return (CLI_RESULT_USAGE);

	ldom_namep = argvp[argc-1];

	if ((ldp = ldm_get_ldp(resp, ldom_namep)) == NULL)
		return (CLI_RESULT_FAIL);

	config_mode_info_warning(resp, NULL);

	if (argc == 1) {
		print_ldom_variables(ldp, resp, NULL);
		return (CLI_RESULT_OK);
	}

	if (!print_ldom_variable_list(ldp, resp, argc-1, argvp))
		return (CLI_RESULT_FAIL);

	return (CLI_RESULT_OK);
}
