/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines and tables for operations on 'vdiskserver' and
 * 'vdiskserverdevice' resources.
 */

#include <string.h>
#include <stdlib.h>

#include <libintl.h>

#include "allocate.h"
#include "basics.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "database.h"
#include "database_internal.h"
#include "dr.h"
#include "ldmd_types.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "response.h"
#include "vdisk.h"
#include "vds.h"
#include "xml_cntrl_common.h"
#include "xml_event.h"


cli_res_cmds_t	cli_vdiskserver_cmds = {
	.resinfop = "vdiskserver",
	.add_op =	{ cli_vds_add, "<service_name> <ldom>" },
	.remove_op =	{ cli_vds_remove, "[-f] <service_name>" },
};

/* FIXME: generate list of options for vdsdev help message automatically */
cli_res_cmds_t	cli_vds_device_cmds = {
	.resinfop = "vdiskserverdevice",
	.add_op =	{
		cli_vdsdev_add,
		"[-f] [options={ro,slice,excl}] [mpgroup=<mpgroup>]\n\t\t"
		"<backend> <volume_name>@<service_name>"
		},
	.set_op =	{
		cli_vdsdev_set,
		"[-f] [options=[{ro,slice,excl}]] [mpgroup=[<mpgroup>]]\n\t\t"
		"<volume_name>@<service_name>"
		},
	.remove_op =	{
		cli_vdsdev_remove,
		"[-f] <volume_name>@<service_name>"
		},
};

cli_res_cmds_t	cli_vdisk_cmds = {
	.resinfop = "vdisk",
	.add_op	= 	{ cli_vdisk_add,
			"[timeout=<seconds>] [id=<diskid>]\n\t\t<disk_name> "
			"<volume_name>@<service_name> <ldom>" },
	.set_op	= 	{ cli_vdisk_set,
			"[timeout=<seconds>] "
			"[volume=<volume_name>@<service_name>] "
			"<disk_name> <ldom>" },
	.remove_op =	{ cli_vdisk_remove,
		"[-f] <disk_name> <ldom>" },
};


/*
 * Virtual disk server CLI processing
 */

cli_result_t
cli_vds_add(int argc, char **argvp, cli_conn_t *clientp)
{
	char *vds_name;
	char *ldom_namep;
	response_t *resp = clientp->resp;
	ldom_t	*ldp;

	if (argc != 2)
		return (CLI_RESULT_USAGE);

	vds_name = argvp[0];
	ldom_namep = argvp[1];

	if (!is_valid_object_name(resp, vds_name)) {
		rprintf(resp, gettext("%s is not a valid service name\n"),
		    vds_name);
		return (CLI_RESULT_FAIL);
	}

	if ((ldp = ldm_get_ldp(resp, ldom_namep)) == NULL)
		return (CLI_RESULT_FAIL);

	if (!config_mod_allowed(ldp, DEL_RECONFIG_ADD, DR_OP_TYPE_VIO, resp))
		return (CLI_RESULT_FAIL);

	if (vds_add(ldp, vds_name, resp) != 0)
		return (CLI_RESULT_FAIL);

	event_send_res_event(XML_ADD_VDS, ldp, NULL, NULL, clientp);
	database_save();
	return (CLI_RESULT_OK);
}

cli_result_t
cli_vds_remove(int argc, char **argvp, cli_conn_t *client)
{
	response_t *resp = client->resp;
	bool_t force = false;
	int i;
	char *vds_name;
	vds_t *vds;
	ldom_t	*ldp;

	/* extract meaningful flags */
	for (i = 0; i < argc && argvp[i][0] == '-'; i++) {
		char *argp = argvp[i];
		if (strcmp(argp, "-f") == 0 ||
		    strcmp(argp, "--force") == 0) {
			force = true;
		} else
		if (strcmp(argp, "--") == 0) {
			break;
		} else {
			rprintf(resp, gettext("Unknown flag %s\n"), argp);
			return (CLI_RESULT_USAGE);
		}
	}

	/* service name should be last arg in list */

	if (i != argc-1)
		return (CLI_RESULT_USAGE);

	vds_name = argvp[i];

	if (!is_valid_object_name(resp, vds_name)) {
		rprintf(resp, gettext("%s is not a valid VDS service name\n"),
		    vds_name);
		return (CLI_RESULT_FAIL);
	}

	if ((vds = vds_find_by_name(vds_name)) == NULL) {
		rprintf(resp,
		    gettext("Virtual disk server %s cannot be found\n"),
		    vds_name);
		return (CLI_RESULT_FAIL);
	}
	ldp = vds->ldp;

	if (!config_mod_allowed(ldp, DEL_RECONFIG_REMOVE, DR_OP_TYPE_VIO, resp))
		return (CLI_RESULT_FAIL);

	if (vds_remove(vds, force, resp) != 0) {
		rprintf(resp,
		    gettext("Failed to remove virtual disk server\n"));
		return (CLI_RESULT_FAIL);
	}

	event_send_res_event(XML_REMOVE_VDS, NULL, vds_name, VDS_NODE, client);
	database_save();

	return (CLI_RESULT_OK);
}

static cli_result_t
process_vdsdev_opts(int argc, char **argvp, response_t *resp, char **options,
    bool_t *force, char **mpgname)
{
	char *name, *value;
	int i;
	int start = 0;

	*force = false;
	/* extract meaningful flags */
	for (i = 0; i < argc && argvp[i][0] == '-'; i++) {
		char *argp = argvp[i];
		if (strcmp(argp, "-f") == 0 || strcmp(argp, "--force") == 0) {
			*force = true;
			start = 1;
		} else if (strcmp(argp, "--") == 0) {
			break;
		} else {
			rprintf(resp, gettext("Unknown flag %s\n"), argp);
			return (CLI_RESULT_USAGE);
		}
	}

	for (i = start; i < argc; i++) {
		name = argvp[i];
		value = strchr(argvp[i], '=');
		if (value != NULL) {
			*value = '\0';	/* overwrite the name/value separator */
			value++;
		}

		if (strcmp(name, "options") == 0) {
			/* Check for valid options */
			char *opt, *opts;

			if (value == NULL) {
				/* allow option= */
				*options = ""; /* Empty string */
			} else {
				opts = Xstrdup(value);
				opt = strtok(opts, ",");

				while (opt != NULL) {
					int j = 0;

					while (vdsdev_opts[j] != NULL) {
						if (strcmp(opt, vdsdev_opts[j])
						    == 0)
							break;
						j++;
					}
					if (vdsdev_opts[j] == NULL) {
						rprintf(resp,
						    gettext("%s not a valid "
						    "option\n"), opt);
						Xfree(opts);
						return (CLI_RESULT_FAIL);
					}
					opt = strtok(NULL, ",");
				}
				Xfree(opts);
				*options = value;
			}
		} else if (strcmp(name, "mpgroup") == 0) {
			if (value == NULL || value[0] == '\0') {
				/* allow mpgroup= */
				*mpgname = ""; /* Empty string */
			} else {
				if (!is_valid_object_name(resp, value)) {
					rprintf(resp, gettext("%s is not a "
					    "valid mpgroup name\n"), value);
					return (CLI_RESULT_FAIL);
				}
				*mpgname = value;
			}
		} else {
			return (CLI_RESULT_USAGE);
		}
	}
	return (CLI_RESULT_OK);
}

/*
 * Add a device to a vds service.
 * usage: [options=<opts>] [mpgroup=<group_name>] <backend>
 * <volume_name>@<service_name>
 */
cli_result_t
cli_vdsdev_add(int argc, char **argvp, cli_conn_t *clientp)
{
	char *devicep;
	char *volumep;
	char *servicep;
	cli_result_t rc;
	response_t *resp = clientp->resp;
	char *options = NULL;
	char *mpgroup = NULL;
	bool_t force = false;

	if (argc < 2 || argc > 4)
		return (CLI_RESULT_USAGE);

	/* Optional arguments processed first */
	rc = process_vdsdev_opts(argc-2, argvp, resp, &options, &force,
	    &mpgroup);
	if (rc != CLI_RESULT_OK)
		return (rc);

	/*
	 * options cannot be empty when adding a vdsdev
	 */
	if (options != NULL && options[0] == '\0') {
		rprintf(resp, gettext("options cannot be empty\n"));
		return (CLI_RESULT_USAGE);
	}

	devicep = argvp[argc-2];
	if (!is_valid_device_name(resp, devicep)) {
		rprintf(resp, gettext("%s is not a valid device/file name\n"),
		    devicep);
		return (CLI_RESULT_FAIL);
	}

	volumep = argvp[argc-1];
	servicep = strchr(volumep, '@');
	if (servicep == NULL || servicep == volumep || servicep[1] == '\0')
		return (CLI_RESULT_USAGE);
	*servicep++ = '\0';

	if (!is_valid_object_name(resp, volumep)) {
		rprintf(resp, gettext("Illegal volume name: %s\n"), volumep);
		return (CLI_RESULT_FAIL);
	}
	if (!is_valid_object_name(resp, servicep)) {
		rprintf(resp, gettext(
		    "Illegal disk server name: %s\n"), servicep);
		return (CLI_RESULT_FAIL);
	}

	vds_t *vds = vds_find_by_name(servicep);

	if (vds == NULL) {
		rprintf(resp, gettext("Disk server \"%s\" not found\n"),
		    servicep);
		return (CLI_RESULT_FAIL);
	}

	/*
	 * Adding a vdsdev is not a VIO DR operation (we don't modify MDs)
	 * so we only need to pass DR_OP_TYPE_VIO_MD_NOP
	 */
	if (config_mod_allowed(vds->ldp, DEL_RECONFIG_ADD,
	    DR_OP_TYPE_VIO_MD_NOP, resp) == false)
		return (CLI_RESULT_FAIL);

	if (vds_add_dev(vds, volumep, devicep, options, mpgroup, servicep,
	    force, resp) != 0)
		return (CLI_RESULT_FAIL);

	event_send_res_event(XML_ADD_VDSDEV, NULL, servicep, VDS_VOLUME_NODE,
	    clientp);

	database_save();
	return (CLI_RESULT_OK);
}

/*
 * ldm set-vdsdev [options=<opts>] <[mpgroup=<group_name>]
 * volume_name>@<service_name>
 */
cli_result_t
cli_vdsdev_set(int argc, char **argvp, cli_conn_t *clientp)
{
	char *volumep;
	char *servicep;
	response_t *resp = clientp->resp;
	char *options = NULL;
	char *mpgroup = NULL;
	int rc = 0;
	bool_t force = false;

	if (argc != 2)
	if (argc < 2 || argc > 3)
		return (CLI_RESULT_USAGE);

	rc = process_vdsdev_opts(argc-1, argvp, resp, &options, &force,
	    &mpgroup);
	if (rc != CLI_RESULT_OK)
		return (rc);

	volumep = argvp[argc-1];
	servicep = strchr(volumep, '@');
	if (servicep == NULL || servicep == volumep || servicep[1] == '\0')
		return (CLI_RESULT_USAGE);

	*servicep++ = '\0';

	if (!is_valid_object_name(resp, volumep)) {
		rprintf(resp, gettext("Illegal volume name: %s\n"), volumep);
		return (CLI_RESULT_FAIL);
	}

	if (!is_valid_object_name(resp, servicep)) {
		rprintf(resp, gettext(
		    "Illegal disk server name: %s\n"), servicep);
		return (CLI_RESULT_FAIL);
	}

	vds_t *vds = vds_find_by_name(servicep);
	if (vds == NULL) {
		rprintf(resp, gettext("Disk server \"%s\" not found\n"),
		    servicep);
		return (CLI_RESULT_FAIL);
	}

	/*
	 * Modifying a vdsdev is not a VIO DR operation (we don't modify MDs)
	 * so we only need to pass DR_OP_TYPE_VIO_MD_NOP
	 */
	if (config_mod_allowed(vds->ldp, DEL_RECONFIG_REMOVE,
	    DR_OP_TYPE_VIO_MD_NOP, resp) == false)
		return (CLI_RESULT_FAIL);

	if (vds_modify_dev(vds, volumep, options, mpgroup, force, resp) != 0)
		return (CLI_RESULT_FAIL);

	event_send_res_event(XML_SET_VDSDEV, NULL, servicep, VDS_VOLUME_NODE,
	    clientp);
	database_save();

	return (CLI_RESULT_OK);
}

/*
 * Remove a device from a vds service.
 * usage: <-f|--force> <volume_name>@<service_name>
 */
cli_result_t
cli_vdsdev_remove(int argc, char **argvp, cli_conn_t *client)
{
	response_t *resp = client->resp;
	int i = 0;
	char	*servicep;
	char	*volumep;
	bool_t force = false;

	/* extract meaningful flags */
	for (i = 0; i < argc && argvp[i][0] == '-'; i++) {
		char *argp = argvp[i];
		if (strcmp(argp, "-f") == 0 || strcmp(argp, "--force") == 0) {
			force = true;
		} else if (strcmp(argp, "--") == 0) {
			break;
		} else {
			rprintf(resp, gettext("Unknown flag %s\n"), argp);
			return (CLI_RESULT_USAGE);
		}
	}

	/* service name should be last arg in list */

	if (i != argc-1)
		return (CLI_RESULT_USAGE);

	volumep = argvp[i];
	servicep = strchr(volumep, '@');
	if (servicep == NULL || servicep == volumep || servicep[1] == '\0')
		return (CLI_RESULT_USAGE);
	*servicep++ = '\0';

	if (!is_valid_object_name(resp, volumep)) {
		rprintf(resp, gettext("Illegal volume name: %s\n"), volumep);
		return (CLI_RESULT_FAIL);
	}
	if (!is_valid_object_name(resp, servicep)) {
		rprintf(resp, gettext("Illegal disk server name: %s\n"),
		    servicep);
		return (CLI_RESULT_FAIL);
	}

	vds_t *vds = vds_find_by_name(servicep);

	if (vds == NULL) {
		rprintf(resp, gettext("Disk server \"%s\" not found\n"),
		    servicep);
		return (CLI_RESULT_FAIL);
	}

	/*
	 * Removing a vdsdev is not a VIO DR operation (we don't modify MDs)
	 * so we only need to pass DR_OP_TYPE_VIO_MD_NOP
	 */
	if (config_mod_allowed(vds->ldp, DEL_RECONFIG_REMOVE,
	    DR_OP_TYPE_VIO_MD_NOP, resp) == false)
		return (CLI_RESULT_FAIL);

	if (vds_remove_dev(vds, volumep, force, resp) != 0) {
		rprintf(resp, gettext("Failed removing volume %s from %s\n"),
		    volumep, servicep);
		return (CLI_RESULT_FAIL);
	}

	event_send_res_event(XML_REMOVE_VDSDEV, NULL, servicep,
	    VDS_VOLUME_NODE, client);

	database_save();

	return (CLI_RESULT_OK);
}

static cli_result_t
parse_volume_service(char *volumep, char **servicep, response_t *resp)
{
	*servicep = strchr(volumep, '@');
	if (*servicep == NULL || *servicep == volumep)
		return (CLI_RESULT_USAGE);
	*(*servicep)++ = '\0';

	if (!is_valid_object_name(resp, volumep)) {
		rprintf(resp, gettext("Illegal volume name: %s\n"), volumep);
		return (CLI_RESULT_FAIL);
	}

	if (!is_valid_object_name(resp, *servicep)) {
		rprintf(resp, gettext("%s is not a valid VDS service name\n"),
		    *servicep);
		return (CLI_RESULT_FAIL);
	}

	return (CLI_RESULT_OK);
}

/*
 * Add a vdisk volume to a domain
 * [<timeout=<seconds>] [id=<diskid>] <disk_name> <volume@service> <ldom>
 */

cli_result_t
cli_vdisk_add(int argc, char **argvp, cli_conn_t *clientp)
{
	char *diskp;
	char *volumep;
	char *servicep;
	char *ldom_namep;
	int timeout = VDISK_DEFAULT_TIMEOUT;
	int diskid = NO_NEW_ID;
	int i = 0;
	response_t *resp = clientp->resp;
	ldom_t	*ldp;
	cli_result_t rc;


	if (argc > 5 || argc < 3)
		return (CLI_RESULT_USAGE);

	/* Optional arguments processed first */
	for (i = 0; i < argc - 3; i++) {
		char *namep;
		char *valuep;

		namep = strtok(argvp[i], "=");
		valuep = strtok(NULL, "=");

		if (strcmp(namep, "timeout") == 0) {
			if (valuep == NULL) {
				/* allow the case of timeout= */
				timeout = VDISK_DEFAULT_TIMEOUT;
			} else {
				timeout = atoi(valuep);
				if (timeout < 0) {
					return (CLI_RESULT_USAGE);
				}
			}
		} else if (strcmp(namep, "id") == 0) {
			char *esp;
			if (valuep == NULL) {
				rprintf(resp, gettext("id must be "
				    "specified\n"));
				return (CLI_RESULT_FAIL);
			}
			diskid = strtoll(valuep, &esp, 10);
			if ((esp == valuep) || (esp[0] != '\0')) {
				rprintf(resp, gettext("Incorrect "
				    "id specification\n"));
				return (CLI_RESULT_FAIL);
			}
			if (diskid < 0)
				return (CLI_RESULT_USAGE);
		} else {
			/* print error and usage */
			if (valuep != NULL)
				rprintf(resp, gettext(
				    "Unknown property: %s\n"), namep);
				return (CLI_RESULT_USAGE);
		}
	}

	diskp = argvp[i++];
	if (!is_valid_object_name(resp, diskp)) {
		rprintf(resp, gettext("%s is not a valid vdisk name\n"), diskp);
		return (CLI_RESULT_FAIL);
	}

	volumep = argvp[i++];
	rc = parse_volume_service(volumep, &servicep, resp);
	if (rc != CLI_RESULT_OK)
		return (rc);

	ldom_namep = argvp[i++];

	if ((ldp = ldm_get_ldp(resp, ldom_namep)) == NULL)
		return (CLI_RESULT_FAIL);

	if (!config_mod_allowed(ldp, DEL_RECONFIG_ADD, DR_OP_TYPE_VIO, resp))
		return (CLI_RESULT_FAIL);

	if (vdisk_add(ldp, diskp, servicep, volumep, timeout, diskid,
	    resp) != 0)
		return (CLI_RESULT_FAIL);

	event_send_res_event(XML_ADD_VDISK, ldp, NULL, NULL, clientp);
	database_save();
	return (CLI_RESULT_OK);
}

/*
 *	ldm set-vdisk [timeout=<seconds>][volume=<volume_name>@<service_name>]
 *	<disk_name><ldom>
 */
cli_result_t
cli_vdisk_set(int argc, char **argvp, cli_conn_t *clientp)
{
	char *diskp = NULL;
	int timeout = VDISK_NO_NEW_TIMEOUT;
	int i;
	char *volumep = NULL;
	char *servicep = NULL;
	response_t *resp = clientp->resp;
	ldom_t	*ldp;
	cli_result_t rc;

	if (argc > 4 || argc < 3)
		return (CLI_RESULT_USAGE);

	/* Optional arguments processed first */
	for (i = 0; i < argc-2; i++) {
		char *namep;
		char *valuep;

		namep = strtok(argvp[i], "=");
		valuep = strtok(NULL, "=");

		if (strcmp(namep, "timeout") == 0) {
			if (valuep == NULL) {
				/* allow the case of timeout= */
				timeout = VDISK_DEFAULT_TIMEOUT;
			} else {
				timeout = atoi(valuep);
				if (timeout < 0) {
					return (CLI_RESULT_USAGE);
				}
			}
		} else if (strcmp(namep, "volume") == 0) {
			if (valuep == NULL)
				return (CLI_RESULT_USAGE);

			volumep = valuep;
			rc = parse_volume_service(volumep, &servicep, resp);
			if (rc != CLI_RESULT_OK)
				return (rc);
		} else {
			/* print error and usage */
			if (valuep != NULL)
				rprintf(resp, gettext(
				    "Unknown property: %s\n"), namep);
			return (CLI_RESULT_USAGE);
		}
	}

	diskp = argvp[i++];
	if (!is_valid_object_name(resp, diskp)) {
		rprintf(resp, gettext("Illegal name for VDISK: %s\n"), diskp);
		return (CLI_RESULT_FAIL);
	}

	if ((ldp = ldm_get_ldp(resp, argvp[i])) == NULL)
		return (CLI_RESULT_FAIL);

	if (config_mod_allowed(ldp, DEL_RECONFIG_OTHER,
	    DR_OP_TYPE_VIO_SET, resp) == false)
		return (CLI_RESULT_FAIL);

	if (vdisk_modify(ldp, diskp, servicep, volumep, timeout, resp) != 0) {
		return (CLI_RESULT_FAIL);
	}

	event_send_res_event(XML_SET_VDISK, ldp, NULL, NULL, clientp);
	database_save();

	return (CLI_RESULT_OK);
}

/*
 * Delete a disk from a domain
 * [-f|--force] <disk_name> <ldom>
 */
cli_result_t
cli_vdisk_remove(int argc, char **argvp, cli_conn_t *clientp)
{
	char	*diskp;
	char	*ldom_namep;
	bool_t force = false;
	response_t *resp = clientp->resp;
	int	i;
	ldom_t	*ldp;

	/* extract meaningful flags */
	for (i = 0; i < argc && argvp[i][0] == '-'; i++) {
		char *argp = argvp[i];
		if (strcmp(argp, "-f") == 0 ||
		    strcmp(argp, "--force") == 0) {
			force = true;
		} else
		if (strcmp(argp, "--") == 0) {
			break;
		} else {
			rprintf(resp, gettext("Unknown flag %s\n"), argp);
			return (CLI_RESULT_USAGE);
		}
	}

	/* service & ldom names should be last arg in list */

	if (i != argc-2)
		return (CLI_RESULT_USAGE);

	diskp = argvp[i];
	ldom_namep = argvp[i+1];

	if (!is_valid_object_name(resp, diskp)) {
		rprintf(resp, gettext("%s is not a valid vdisk name\n"),
		    diskp);
		return (CLI_RESULT_FAIL);
	}

	if ((ldp = ldm_get_ldp(resp, ldom_namep)) == NULL)
		return (CLI_RESULT_FAIL);

	if (config_mod_allowed(ldp, DEL_RECONFIG_REMOVE_2PHASE, DR_OP_TYPE_VIO,
	    resp) == false)
		return (CLI_RESULT_FAIL);

	if (vdisk_remove(ldp, diskp, force, resp) != 0) {
		rprintf(resp, gettext("Failed to remove vdisk instance\n"));
		return (CLI_RESULT_FAIL);
	}

	event_send_res_event(XML_REMOVE_VDISK, ldp, NULL, NULL, clientp);
	database_save();
	return (CLI_RESULT_OK);
}
