/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines and tables for operations on Netra Data Plane
 * client and server resources.
 */

#include <string.h>
#include <stdlib.h>

#include <libintl.h>

#include "allocate.h"
#include "basics.h"
#include "cli_conn.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "database.h"
#include "database_internal.h"
#include "dr.h"
#include "ldmd_types.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "response.h"
#include "vdpcs.h"
#include "vdpcc.h"
#include "xml_cntrl_common.h"
#include "xml_event.h"


cli_res_cmds_t	cli_vdpcs_cmds = {
	.resinfop = "vdpcs",
	.add_op	= 	{ cli_vdpcs_add,
		"<vdpcs_name> <ldom>" },
	.remove_op =	{ cli_vdpcs_remove,
		"[-f] <vdpcs_name>" },
};

cli_res_cmds_t	cli_vdpcc_cmds = {
	.resinfop = "vdpcc",
	.add_op	= 	{ cli_vdpcc_add,
		"<vdpcc_name> <service_name> <ldom>" },
	.remove_op =	{ cli_vdpcc_remove,
		"[-f] <vdpcc_name> <ldom>" },
};

/*
 * Netra Data Plane server CLI processing
 */

cli_result_t
cli_vdpcs_add(int argc, char **argvp, cli_conn_t *clientp)
{
	char *servicep;
	char *ldom_namep;
	response_t *resp = clientp->resp;
	ldom_t	*ldp;

	if (argc != 2)
		return (CLI_RESULT_USAGE);

	servicep = argvp[0];
	ldom_namep = argvp[1];

	if (!is_valid_object_name(resp, servicep)) {
		rprintf(resp, gettext("%s is not a valid service name\n"),
		    servicep);
		return (CLI_RESULT_FAIL);
	}

	if ((ldp = ldm_get_ldp(resp, ldom_namep)) == NULL)
		return (CLI_RESULT_FAIL);

	if (!config_mod_allowed(ldp, DEL_RECONFIG_ADD, DR_OP_TYPE_NDPS, resp))
		return (CLI_RESULT_FAIL);

	if (vdpcs_add(ldp, servicep, resp) != 0) {
		rprintf(resp, gettext(
		    "Unable to add vdpcs server \"%s\" to LDom %s\n"),
		    servicep, ldom_namep);
		return (CLI_RESULT_FAIL);
	}

	event_send_res_event(XML_ADD_VDPCS, ldp, NULL, NULL, clientp);
	database_save();
	return (CLI_RESULT_OK);
}

cli_result_t
cli_vdpcs_remove(int argc, char **argvp, cli_conn_t *client)
{
	response_t *resp = client->resp;
	int i;
	char	*servicep;
	bool_t force = false;
	vdpcs_t *vdpcs;

	/* extract meaningful flags */
	for (i = 0; i < argc && argvp[i][0] == '-'; i++) {
		char *argp = argvp[i];
		if (strcmp(argp, "-f") == 0 ||
		    strcmp(argp, "--force") == 0) {
			force = true;
		} else if (strcmp(argp, "--") == 0) {
			break;
		} else {
			rprintf(resp, gettext("Unknown flag %s\n"), argp);
			return (CLI_RESULT_USAGE);
		}
	}

	/* service name should be last arg in list */

	if (i != argc-1)
		return (CLI_RESULT_USAGE);

	servicep = argvp[i];

	if (!is_valid_object_name(resp, servicep)) {
		rprintf(resp, gettext("%s is not a valid vdpcs service name\n"),
		    servicep);
		return (CLI_RESULT_FAIL);
	}

	if ((vdpcs = vdpcs_find_by_name(servicep)) == NULL) {
		rprintf(resp, gettext("vdpcs service \"%s\" not found\n"),
		    servicep);
		return (CLI_RESULT_FAIL);
	}

	if (!config_mod_allowed(vdpcs->ldp, DEL_RECONFIG_REMOVE_2PHASE,
	    DR_OP_TYPE_NDPS, resp))
		return (CLI_RESULT_FAIL);

	if (vdpcs_remove(vdpcs, force, resp) != 0) {
		rprintf(resp, gettext("Failed to remove vdpcs server\n"));
		return (CLI_RESULT_FAIL);
	}

	event_send_res_event(XML_REMOVE_VDPCS, NULL, servicep, VDPCS_NODE,
	    client);
	database_save();

	return (CLI_RESULT_OK);
}

/*
 * Netra Data Plane client CLI processing
 */

cli_result_t
cli_vdpcc_add(int argc, char **argvp, cli_conn_t *clientp)
{
	char *netp;
	char *servicep;
	char *ldom_namep;
	response_t *resp = clientp->resp;
	ldom_t	*ldp;

	/* There are no options, and three params required */

	if (argc != 3)
		return (CLI_RESULT_USAGE);

	netp = argvp[0];
	if (!is_valid_object_name(resp, netp)) {
		rprintf(resp, gettext("%s is not a valid vdpcc name\n"),
		    netp);
		return (CLI_RESULT_FAIL);
	}

	servicep = argvp[1];
	if (!is_valid_object_name(resp, servicep)) {
		rprintf(resp, gettext("%s is not a valid vdpcs service name\n"),
		    servicep);
		return (CLI_RESULT_FAIL);
	}

	ldom_namep = argvp[2];

	if ((ldp = ldm_get_ldp(resp, ldom_namep)) == NULL)
		return (CLI_RESULT_FAIL);

	if (!config_mod_allowed(ldp, DEL_RECONFIG_ADD, DR_OP_TYPE_NDPS, resp))
		return (CLI_RESULT_FAIL);

	if (vdpcc_add(ldp, netp, servicep, resp) != 0)
		return (CLI_RESULT_FAIL);

	event_send_res_event(XML_ADD_VDPCC, ldp, NULL, NULL, clientp);
	database_save();
	return (CLI_RESULT_OK);
}

cli_result_t
cli_vdpcc_remove(int argc, char **argvp, cli_conn_t *clientp)
{
	char	*netp;
	char	*ldom_namep;
	bool_t force = false;
	response_t *resp = clientp->resp;
	int	i;
	ldom_t	*ldp;

	/* extract meaningful flags */
	for (i = 0; i < argc && argvp[i][0] == '-'; i++) {
		char *argp = argvp[i];
		if (strcmp(argp, "-f") == 0 ||
		    strcmp(argp, "--force") == 0) {
			force = true;
		} else if (strcmp(argp, "--") == 0) {
			break;
		} else {
			rprintf(resp, gettext("Unknown flag %s\n"), argp);
			return (CLI_RESULT_USAGE);
		}
	}

	/* service & ldom names should be last arg in list */

	if (i != argc-2)
		return (CLI_RESULT_USAGE);

	netp = argvp[i];
	ldom_namep = argvp[i+1];

	if (!is_valid_object_name(resp, netp)) {
		rprintf(resp, gettext("%s is not a valid vdpcc name\n"), netp);
		return (CLI_RESULT_FAIL);
	}

	if ((ldp = ldm_get_ldp(resp, ldom_namep)) == NULL)
		return (CLI_RESULT_FAIL);

	if (!config_mod_allowed(ldp, DEL_RECONFIG_REMOVE_2PHASE,
	    DR_OP_TYPE_NDPS, resp))
		return (CLI_RESULT_FAIL);

	if (vdpcc_remove(ldp, netp, force, resp) != 0) {
		rprintf(resp, gettext("Failed to remove vdpcc instance\n"));
		return (CLI_RESULT_FAIL);
	}

	event_send_res_event(XML_REMOVE_VDPCC, ldp, NULL, NULL, clientp);

	return (CLI_RESULT_OK);
}
