/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * CLI routines and tables for operations on 'vswitch' and 'vnet'
 * resources.
 */

#include <string.h>
#include <stdlib.h>
#include <stdint.h>

#include <libintl.h>

#include "allocate.h"
#include "basics.h"
#include "cli_conn.h"
#include "cli_res_cmds.h"
#include "cli_util.h"
#include "database_internal.h"
#include "errors.h"
#include "ldmd_types.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "mac_addrs.h"
#include "mac_multicast.h"
#include "response.h"
#include "vio.h"
#include "vnet.h"
#include "vsw.h"
#include "xml_cntrl_common.h"
#include "xml_event.h"


cli_res_cmds_t	cli_vswitch_cmds = {
	.resinfop = "vswitch",
	.add_op =	{ cli_vswitch_add,
		"[default-vlan-id=<vid>] [pvid=<pvid>] [vid=<vid1,vid2,...>]"
		"\n\t\t[mac-addr=<num>] [net-dev=<device>] "
		"[linkprop=phys-state]\n\t\t[mode=<mode>] [mtu=<mtu>] "
		"[id=<switchid>] <vswitch_name> <ldom>" },
	.set_op =	{ cli_vswitch_set,
		"[pvid=[<pvid>]] [vid=[<vid1,vid2,...>]] [mac-addr=<num>]\n\t\t"
		"[net-dev=[<device>]] [mode=[<mode>]] [mtu=[<mtu>]]\n\t\t"
		"[linkprop=[phys-state]] <vswitch_name>" },
	.remove_op =	{ cli_vswitch_remove, "[-f] <vswitch_name>"},
};

cli_res_cmds_t	cli_vnet_cmds = {
	.resinfop = "vnet",
	.add_op =	{
		cli_vnet_add,
		"[mac-addr=<num>] [mode=hybrid] [pvid=<pvid>]\n\t\t"
		"[vid=<vid1,vid2,...>] [mtu=<mtu>] [linkprop=phys-state]"
		"\n\t\t[id=<networkid>] <if_name> <vswitch_name> <ldom>"
	},
	.set_op =	{
		cli_vnet_set,
		"[mac-addr=<num>] [mode=[hybrid]] [pvid=[<pvid>]]\n\t\t"
		"[vid=[<vid1,vid2,...>]] [mtu=[<mtu>]] [linkprop=[phys-state]]"
		"\n\t\t[vswitch=<vswitch_name>] <if_name> <ldom>"
	},
	.remove_op =	{
	    cli_vnet_remove, "[-f] <if_name> <ldom>"
	},
};


/*
 * Virtual switch CLI processing.
 */

/*
 * Check if selected vsw mode(s) is valid.
 */
static bool_t
is_vsw_mode(char *cli_str, response_t *resp)
{
	char *opt = NULL;
	char *opts = NULL;

	if (cli_str == NULL) {
		/* mode= */
		return (true);
	}

	opts = Xstrdup(cli_str);
	opt = strtok(opts, ",");
	while (opt != NULL) {
		int i = 0;
		while (vsw_mode_options[i].cli != NULL) {
			if (strcmp(opt, vsw_mode_options[i].cli) == 0)
				break;
			i++;
		}

		if (vsw_mode_options[i].cli == NULL) {
			rprintf(resp, gettext("%s not a valid option "
			    "for mode\n"), opt);
			Xfree(opts);
			return (false);
		}

		opt = strtok(NULL, ",");
	}

	Xfree(opts);
	return (true);
}

/*
 * ldm add-vswitch [mac-addr=<num>] [net-dev=<device>] [mode=<mode>]
 *	[linkprop=phys-state] [default-vlan-id=<def-pvid>] [pvid=<pvid>]
 *	[vid=<',' separated vlan-ids>] [mtu=<mtu>] [id=<switchid>]
 *	<vswitch_name> <ldom>
 */

cli_result_t
cli_vswitch_add(int argc, char **argvp, cli_conn_t *clientp)
{
	int i;
	char		*vsw_name = NULL;
	mac_addr_t	mac_address = ILLEGAL_MAC_ADDR;
	char		*sw_mode;		/* vestigial */
	char		*net_devicep = NULL;
	ldom_t		*ldp;
	response_t 	*resp = clientp->resp;
	vsw_mac_t	*vsw_nodep = NULL;
	char		*mode = NULL;
	int		rv;
	int		default_vid = -1;
	int		pvid = -1;
	char		*vids = NULL;
	int		mtu = VIO_NOARGSUP;
	uint64_t	phys_state = 0;
	int		switchid = NO_NEW_ID;

	/* Optional arguments processed first */
	for (i = 0; i < argc; i++) {
		char *namep;
		char *valuep;

		namep = strtok(argvp[i], "=");
		valuep = strtok(NULL, "=");

		if (strcmp(namep, "mac-addr") == 0) {
			/* TRANSLATION_NOTE Don't translate 'mac-addr' */
			if (valuep == NULL || !atomac(valuep, &mac_address)) {
				rprintf(resp, gettext("Invalid mac-addr\n"));
				return (CLI_RESULT_FAIL);
			}

			/*
			 * Can MAC addr be allocated?
			 */
			if (IS_PUBLIC_MAC_ADDR(mac_address)) {
				if (is_macaddr_in_use(mac_address,
				    PERFORM_LOCAL_MAC_CHECK, resp)) {
					return (CLI_RESULT_FAIL);
				}
			}
		} else if (strcmp(namep, "net-dev") == 0) {
			/* validate net device */
			if (!is_valid_device_name(resp, valuep)) {
				rprintf(resp, gettext(
				    "Invalid device name %s\n"), valuep);
				return (CLI_RESULT_FAIL);
			}
			if (valuep != NULL)
				net_devicep = valuep;
		} else if (strcmp(namep, "mode") == 0) {
			if (!is_vsw_mode(valuep, resp))
				return (CLI_RESULT_FAIL);
			mode = valuep;
		} else if (strcmp(namep, "default-vlan-id") == 0) {
			char *esp;
			if (valuep == NULL) {
				rprintf(resp, gettext("default-vlan-id must be "
				    "specified\n"));
				return (CLI_RESULT_FAIL);
			}
			default_vid = strtoll(valuep, &esp, 10);
			if ((esp == valuep) || (esp[0] != '\0')) {
				rprintf(resp, gettext("Incorrect "
				    "default-vlan-id specification\n"));
				return (CLI_RESULT_FAIL);
			} else if (!VID_IN_RANGE(default_vid)) {
				rprintf(resp, gettext("default-vlan-id must be "
				    "between %d and %d\n"), VID_MIN, VID_MAX);
				return (CLI_RESULT_FAIL);
			}
		} else if (strcmp(namep, "pvid") == 0) {
			char *esp;
			if (valuep == NULL) {
				rprintf(resp, gettext("pvid must be "
				    "specified\n"));
				return (CLI_RESULT_FAIL);
			}
			pvid = strtoll(valuep, &esp, 10);
			if ((esp == valuep) || (esp[0] != '\0')) {
				rprintf(resp, gettext("Incorrect pvid "
				    "specification\n"));
				return (CLI_RESULT_FAIL);
			} else if (!VID_IN_RANGE(pvid)) {
				rprintf(resp, gettext("port VLAN ID must be "
				    "between %d and %d\n"), VID_MIN, VID_MAX);
				return (CLI_RESULT_FAIL);
			}
		} else if (strcmp(namep, "vid") == 0) {
			vids = valuep;
			if (valuep == NULL) {
				rprintf(resp, gettext("vid must be "
				    "specified\n"));
				return (CLI_RESULT_FAIL);
			}
		} else if (strcmp(namep, "mtu") == 0) {
			char *esp;
			if (valuep == NULL) {
				rprintf(resp, gettext("Missing MTU value\n"));
				return (CLI_RESULT_FAIL);
			}
			mtu = strtoll(valuep, &esp, 10);
			if ((esp == valuep) || (esp[0] != '\0')) {
				rprintf(resp, gettext("Incorrect "
				    "MTU specification\n"));
				return (CLI_RESULT_FAIL);
			} else if (!MTU_IN_RANGE(mtu)) {
				rprintf(resp, gettext("MTU must be between %d "
				    "and %d\n"), ETHERMTU_MIN, ETHERMTU_MAX);
				return (CLI_RESULT_FAIL);
			}
		} else if (strcmp(namep, "linkprop") == 0) {
			if (valuep == NULL) {
				rprintf(resp, gettext("Missing linkprop "
				    "value\n"));
				return (CLI_RESULT_FAIL);
			} else if (strcmp(valuep, "phys-state") == 0) {
				phys_state = 0x1;
			} else {
				rprintf(resp, MSG_ERR_LINKPROP_INVALID_VAL);
				return (CLI_RESULT_FAIL);
			}
		} else if (strcmp(namep, "id") == 0) {
			char *esp;
			if (valuep == NULL) {
				rprintf(resp, gettext("id must be "
				    "specified\n"));
				return (CLI_RESULT_FAIL);
			}
			switchid = strtoll(valuep, &esp, 10);
			if ((esp == valuep) || (esp[0] != '\0')) {
				rprintf(resp, gettext("Incorrect "
				    "id specification\n"));
				return (CLI_RESULT_FAIL);
			}
			if (switchid < 0)
				return (CLI_RESULT_USAGE);
		} else {
			/* argument contains "=" token */
			if (valuep != NULL) {
				rprintf(resp, gettext(
				    "Unknown property: %s\n"), namep);
				return (CLI_RESULT_USAGE);
			}
			/* otherwise continue with remaining args */
			break;
		}
	}

	if ((argc - i) <= 0) {
		rprintf(resp, "Virtual switch service and LDom are missing\n");
		return (CLI_RESULT_USAGE);
	} else if ((argc - i) == 1) {
		if (!is_valid_object_name(resp, argvp[i])) {
			rprintf(resp, "Illegal name for virtual switch service "
			    "or ldom: %s\n", argvp[i]);
			return (CLI_RESULT_USAGE);
		}
		if (find_ldom_by_name(argvp[i]) != NULL)
			rprintf(resp, "Virtual switch service is missing\n");
		else
			rprintf(resp, "LDom is missing or \"%s\" was not "
			    "found\n", argvp[i]);
		return (CLI_RESULT_USAGE);
	} else if ((argc - i) > 2) {
		rprintf(resp, "Too many arguments\n");
		return (CLI_RESULT_USAGE);
	}

	/* exactly 2 arguments left to process */
	vsw_name = argvp[i];
	if (!is_valid_object_name(resp, vsw_name)) {
		rprintf(resp, gettext("Illegal name for virtual switch "
		    "service: %s\n"), vsw_name);
		return (CLI_RESULT_FAIL);
	}

	if ((ldp = ldm_get_ldp(resp, argvp[i+1])) == NULL)
		return (CLI_RESULT_FAIL);

	if (net_devicep == NULL && phys_state != 0 && ldp->state >= BOUND) {
		rprintf(resp, MSG_ERR_LINKPROP_NO_NET_DEV, vsw_name);
		return (CLI_RESULT_FAIL);
	}

	if (!config_mod_allowed(ldp, DEL_RECONFIG_ADD, DR_OP_TYPE_VIO, resp))
		return (CLI_RESULT_FAIL);

	/*
	 * If the user did not give us a MAC address and the sockets failed
	 * to be created when LDMD started we cannot allow automatic MAC
	 * allocation.
	 */
	if ((mac_address == ILLEGAL_MAC_ADDR) && (!mac_socket_up)) {
		/* Try to set up the multicast socket again */
		if (!start_mac_auto_allocate()) {
			rprintf(resp, MSG_ERR_MAC_SOCKET_CREATION_FAILED);
			rprintf(resp, MSG_ERR_CHECK_NETWORK);
			return (CLI_RESULT_FAIL);
		}
	}

	/*
	 * If we are not given a net-dev then the mode should automatically
	 * be set to "routed" mode.  If we ARE given a net-dev but NOT given
	 * a mode then default mode is "promisc". We need to set the mode so
	 * that the MAC allocation code can check the mode later.
	 */
	if (net_devicep == NULL || net_devicep[0] == '\0')
		sw_mode = VSW_ROUTED;
	else
		sw_mode = VSW_PROMISC;

	vsw_nodep = get_vsw_node(vsw_name);
	if (vsw_nodep == NULL) {
		/*
		 * There is no place holder vsw in the list so create a new
		 * vsw node in the mac address list.  Then get the mac address
		 * assigned to this vsw node.
		 */
		rv = create_vsw_mac(&vsw_nodep, ldp->namep, vsw_name, sw_mode,
		    mac_address);
	} else {
		/*
		 * We have a place holder vsw node.  Update the MAC address
		 * for this node.
		 */
		rv = update_vsw_mac(vsw_nodep, ldp->namep, mac_address,
		    sw_mode);
	}

	if (rv != 0) {
		rprintf(resp, gettext("Automatic MAC allocation failed.\n"
		    "Please set the MAC address for vswitch %s\n"),
		    vsw_name);
		return (CLI_RESULT_FAIL);
	}

	rv = vsw_add(ldp, vsw_name, net_devicep, vsw_nodep->mac_addr, mode,
	    default_vid, pvid, vids, mtu, phys_state, switchid, resp);
	if (rv != 0) {
		if (rv != ERR_DR_VIO_STILL_ALLOCATED) {
			remove_vsw_mac(vsw_name);
		}
		return (CLI_RESULT_FAIL);
	}

	database_save();

	event_send_res_event(XML_ADD_VSW, ldp, NULL, NULL, clientp);
	DBG(print_mac_list());
	return (CLI_RESULT_OK);
}

/*
 * ldm set-vswitch [mac-addr=<num>] [net-dev=[<device>]] [mode=[<mode>]]
 *	[linkprop=[phys-state]] [pvid=[<port-vlan-id>]] [vid=[<v1,v2,...>]]
 *	[mtu=[<mtu>]] <vswitch_name>
 */

cli_result_t
cli_vswitch_set(int argc, char **argvp, cli_conn_t *clientp)
{
	char		*vsw_name;
	vsw_t		*vsw;
	mac_addr_t	mac_address = ILLEGAL_MAC_ADDR;
	char		*net_devicep = NULL;
	response_t 	*resp = clientp->resp;
	vsw_mac_t	*vsw_nodep = NULL;
	char		*mode = NULL;
	int		pvid = -1;
	int		mtu = VIO_NOARGSUP;
	uint64_t	phys_state = ILLEGAL_LNKP;
	int		i;
	char		*vids = NULL;
	char		*sw_mode;

	/* Optional arguments processed first */
	for (i = 0; i < argc; i++) {
		char *namep;
		char *valuep;

		namep = strtok(argvp[i], "=");
		valuep = strtok(NULL, "=");

		if (strcmp(namep, "mac-addr") == 0) {
			if (valuep == NULL || !atomac(valuep, &mac_address)) {
				rprintf(resp,
				    gettext("Invalid MAC address\n"));
				return (CLI_RESULT_FAIL);
			}

			/*
			 * Can MAC addr be allocated?
			 */
			if (IS_PUBLIC_MAC_ADDR(mac_address)) {
				if (is_macaddr_in_use(mac_address,
				    PERFORM_LOCAL_MAC_CHECK, resp)) {
					return (CLI_RESULT_FAIL);
				}
			}
		} else if (strcmp(namep, "net-dev") == 0) {
			/* validate net device */
			if (!is_valid_device_name(resp, valuep)) {
				rprintf(resp, gettext(
				    "Invalid device name %s\n"), valuep);
				return (CLI_RESULT_FAIL);
			}
			if (valuep != NULL)
				net_devicep = valuep;
			else
				net_devicep = "";
		} else if (strcmp(namep, "mode") == 0) {
			if (!is_vsw_mode(valuep, resp))
				return (CLI_RESULT_FAIL);
			if (valuep != NULL)
				mode = valuep;
			else
				mode = "";
		} else if (strcmp(namep, "pvid") == 0) {
			if (valuep == NULL) {
				pvid = 0;
			} else {
				char *esp;
				pvid = strtoll(valuep, &esp, 10);
				if ((esp == valuep) || (esp[0] != '\0')) {
					rprintf(resp, gettext("Incorrect pvid "
					    "specification\n"));
					return (CLI_RESULT_FAIL);
				} else if (!VID_IN_RANGE(pvid)) {
					rprintf(resp, gettext("port-vlan-id "
					    "must be between %d and %d\n"),
					    VID_MIN, VID_MAX);
					return (CLI_RESULT_FAIL);
				}
			}
		} else if (strcmp(namep, "vid") == 0) {
			vids = valuep;
			if (valuep == NULL)
				vids = _AUTO_STR_;
		} else if (strcmp(namep, "mtu") == 0) {
			if (valuep == NULL) {
				mtu = VIO_NONEMTU;
			} else {
				char *esp;
				mtu = strtoll(valuep, &esp, 10);
				if ((esp == valuep) || (esp[0] != '\0')) {
					rprintf(resp, gettext("Incorrect MTU "
					    "specification\n"));
					return (CLI_RESULT_FAIL);
				} else if (!MTU_IN_RANGE(mtu)) {
					rprintf(resp, gettext("MTU must be "
					    "between %d and %d\n"),
					    ETHERMTU_MIN, ETHERMTU_MAX);
					return (CLI_RESULT_FAIL);
				}
			}
		} else if (strcmp(namep, "linkprop") == 0) {
			if (valuep == NULL) {
				phys_state = 0;
			} else if (strcmp(valuep, "phys-state") == 0) {
				phys_state = 0x1;
			} else {
				rprintf(resp, MSG_ERR_LINKPROP_INVALID_VAL);
				return (CLI_RESULT_FAIL);
			}
		} else {
			/* argument contains "=" token */
			if (valuep != NULL) {
				rprintf(resp, gettext(
				    "Unknown property: %s\n"), namep);
				return (CLI_RESULT_USAGE);
			}
			/* otherwise continue with remaining args */
			break;
		}
	}

	if ((argc - i) <= 0) {
		rprintf(resp, "Virtual switch service is missing\n");
		return (CLI_RESULT_USAGE);
	} else if ((argc - i) > 1) {
		rprintf(resp, "Too many arguments\n");
		return (CLI_RESULT_USAGE);
	}

	/* exactly one argument left to process */
	vsw_name = argvp[i];
	if (!is_valid_object_name(resp, vsw_name)) {
		rprintf(resp, gettext("Illegal name for virtual switch "
		    "service: %s\n"), vsw_name);
		return (CLI_RESULT_FAIL);
	}

	if ((vsw = vsw_find_by_name(vsw_name)) == NULL) {
		rprintf(resp,
		    gettext("Virtual switch service \"%s\" cannot be found\n"),
		    vsw_name);
		return (CLI_RESULT_FAIL);
	}

	if (net_devicep == NULL) {
		if (vsw->phys_dev == NULL && phys_state == 0x1 &&
		    vsw->ldp->state >= BOUND) {
			rprintf(resp, MSG_ERR_LINKPROP_NO_NET_DEV, vsw_name);
			return (CLI_RESULT_FAIL);
		}
	}

	if (check_locked_ldom(vsw->ldp, resp))
		return (CLI_RESULT_FAIL);

	vsw_nodep = get_vsw_node(vsw_name);
	if (vsw_nodep == NULL) {
		rprintf(resp, gettext("Virtual switch service \"%s\" cannot "
		    "be found\n"), vsw_name);
		return (CLI_RESULT_FAIL);
	}

	config_mode_change_warning(resp);

	if (vsw_set(vsw, net_devicep, mac_address, mode, pvid, vids, mtu,
	    phys_state, resp) != 0)
		return (CLI_RESULT_FAIL);

	event_send_res_event(XML_SET_VSW, NULL, vsw_name, VSW_NODE, clientp);

	database_save();

	if (net_devicep == NULL)
		sw_mode = VSW_ROUTED;
	else
		sw_mode = VSW_PROMISC;

	update_vsw_mac(vsw_nodep, NULL, mac_address, sw_mode);
	DBG(print_mac_list());

	return (CLI_RESULT_OK);
}

cli_result_t
cli_vswitch_remove(int argc, char **argvp, cli_conn_t *clientp)
{
	response_t *resp = clientp->resp;
	bool_t force = false;
	int i = 0;
	char *vsw_name;
	vsw_t *vsw;

	if (argc == 0)
		return (CLI_RESULT_USAGE);

	if (argvp[i][0] == '-') {
		if (strcmp(argvp[i], "-f") == 0 ||
		    strcmp(argvp[i], "--force") == 0) {
			force = true;
			i++;
		} else {
			return (CLI_RESULT_USAGE);
		}
	}

	if (argc - i != 1)
		return (CLI_RESULT_USAGE);

	vsw_name = argvp[i];

	if (!is_valid_object_name(resp, vsw_name)) {
		rprintf(resp, gettext("Illegal vswitch name\n"));
		return (CLI_RESULT_FAIL);
	}

	if ((vsw = vsw_find_by_name(vsw_name)) == NULL) {
		rprintf(resp,
		    gettext("Virtual switch %s cannot be found\n"),
		    vsw_name);
		return (CLI_RESULT_FAIL);
	}

	if (config_mod_allowed(vsw->ldp, DEL_RECONFIG_REMOVE,
	    DR_OP_TYPE_VIO, resp) == false)
		return (CLI_RESULT_FAIL);

	if (vsw_remove(vsw, force, resp) != 0) {
		rprintf(resp,
		    gettext("Failed to remove virtual switch\n"));
		return (CLI_RESULT_FAIL);
	}

	event_send_res_event(XML_REMOVE_VSW, NULL, vsw_name, VSW_NODE,
	    clientp);

	database_save();
	remove_vsw_mac(vsw_name);

	DBG3(print_mac_list());

	return (CLI_RESULT_OK);
}

/*
 * Virtual net interface CLI processing
 */

/*
 *	ldm add-vnet [mac-addr=<num> | mode=<mode>] [mode=hybrid] [pvid=<pvid>]
 *		[vid=<vid1,vid2,...>] [mtu=<mtu>] [linkprop=phys-state]
 *		[id=<networkid>] <if_name> <vswitch_name> <ldom>
 */

cli_result_t
cli_vnet_add(int argc, char **argvp, cli_conn_t *clientp)
{
	int i;
	char		*servicep = NULL;
	mac_addr_t	mac_address = ILLEGAL_MAC_ADDR;
	char		*vnet_namep = NULL;
	ldom_t		*ldp;
	response_t 	*resp = clientp->resp;
	vsw_mac_t	*vsw_nodep = NULL;
	vnet_mac_t	*vnet_nodep = NULL;
	int		rv;
	char		*vnet_mode = NULL;
	int		pvid = DEFAULT_VLAN_ID;
	char		*vids = NULL;
	int		mtu = VIO_NOARGSUP;
	uint64_t	phys_state = 0;
	int		networkid = NO_NEW_ID;
	vsw_t		*vsw;

	/* Optional arguments processed first */
	for (i = 0; i < argc-3; i++) {
		char *namep;
		char *valuep;

		namep = strtok(argvp[i], "=");
		valuep = strtok(NULL, "=");

		if (strcmp(namep, "mac-addr") == 0) {
			if (valuep == NULL || !atomac(valuep, &mac_address)) {
				rprintf(resp, gettext("Invalid mac-addr\n"));
				return (CLI_RESULT_FAIL);
			}

			/*
			 * Can MAC addr be allocated?
			 */
			if (IS_PUBLIC_MAC_ADDR(mac_address)) {
				if (is_macaddr_in_use(mac_address,
				    PERFORM_LOCAL_MAC_CHECK, resp)) {
					return (CLI_RESULT_FAIL);
				}
			}
		} else if (strcmp(namep, "mode") == 0) {
			vnet_mode = valuep;
			if (vnet_mode == NULL)
				return (CLI_RESULT_USAGE);
			if (strncmp(vnet_mode,
			    VIO_MODE_HYBRID, strlen(VIO_MODE_HYBRID)) != 0) {
				rprintf(resp, gettext("Invalid mode: %s\n"),
				    vnet_mode);
				return (CLI_RESULT_FAIL);
			}
		} else if (strcmp(namep, "pvid") == 0) {
			char *esp;
			if (valuep == NULL) {
				rprintf(resp, gettext("pvid must be "
				    "specified\n"));
				return (CLI_RESULT_FAIL);
			}
			pvid = strtoll(valuep, &esp, 10);
			if ((esp == valuep) || (esp[0] != '\0')) {
				rprintf(resp, gettext("Incorrect pvid "
				    "specification\n"));
				return (CLI_RESULT_FAIL);
			} else if (!VID_IN_RANGE(pvid)) {
				rprintf(resp, gettext("port-vlan-id "
				    "must be between %d and %d\n"),
				    VID_MIN, VID_MAX);
				return (CLI_RESULT_FAIL);
			}
		} else if (strcmp(namep, "vid") == 0) {
			vids = valuep;
			if (valuep == NULL) {
				rprintf(resp, gettext("vid must be "
				    "specified\n"));
				return (CLI_RESULT_FAIL);
			}
		} else if (strcmp(namep, "mtu") == 0) {
			char *esp;
			if (valuep == NULL) {
				rprintf(resp, gettext("Missing MTU value\n"));
				return (CLI_RESULT_FAIL);
			}
			mtu = strtoll(valuep, &esp, 10);
			if ((esp == valuep) || (esp[0] != '\0')) {
				rprintf(resp, gettext("Incorrect "
				    "MTU specification\n"));
				return (CLI_RESULT_FAIL);
			} else if (!MTU_IN_RANGE(mtu)) {
				rprintf(resp, gettext("MTU must be between %d "
				    "and %d\n"), ETHERMTU_MIN, ETHERMTU_MAX);
				return (CLI_RESULT_FAIL);
			}
		} else if (strcmp(namep, "linkprop") == 0) {
			if (valuep == NULL) {
				rprintf(resp, gettext("Missing linkprop "
				    "value\n"));
				return (CLI_RESULT_FAIL);
			} else if (strcmp(valuep, "phys-state") == 0) {
				phys_state = 0x1;
			} else {
				rprintf(resp, MSG_ERR_LINKPROP_INVALID_VAL);
				return (CLI_RESULT_FAIL);
			}
		} else if (strcmp(namep, "id") == 0) {
			char *esp;
			if (valuep == NULL) {
				rprintf(resp, gettext("id must be "
				    "specified\n"));
				return (CLI_RESULT_FAIL);
			}
			networkid = strtoll(valuep, &esp, 10);
			if ((esp == valuep) || (esp[0] != '\0')) {
				rprintf(resp, gettext("Incorrect "
				    "id specification\n"));
				return (CLI_RESULT_FAIL);
			}
			if (networkid < 0)
				return (CLI_RESULT_USAGE);
		} else {
			/* print error and usage */
			if (valuep != NULL)
				rprintf(resp, gettext(
				    "Unknown property: %s\n"), namep);
			return (CLI_RESULT_USAGE);
		}
	}

	if ((argc-i) != 3)
		return (CLI_RESULT_USAGE);

	vnet_namep = argvp[i];
	if (!is_valid_object_name(resp, vnet_namep)) {
		rprintf(resp, gettext(
		    "Illegal name for VNET interface: %s\n"), vnet_namep);
		return (CLI_RESULT_FAIL);
	}

	servicep = argvp[i+1];
	if (!is_valid_object_name(resp, servicep)) {
		rprintf(resp, gettext("Illegal name for vswitch: %s\n"),
		    servicep);
		return (CLI_RESULT_FAIL);
	}

	if ((ldp = ldm_get_ldp(resp, argvp[i+2])) == NULL)
		return (CLI_RESULT_FAIL);

	if (check_locked_ldom(ldp, resp))
		return (CLI_RESULT_FAIL);

	/*
	 * If the user did not give us a MAC address and the sockets failed
	 * to be created when LDMD started we cannot allow automatic MAC
	 * allocation.
	 */
	if ((mac_address == ILLEGAL_MAC_ADDR) && (!mac_socket_up)) {
		/* Try to set up the multicast socket again */
		if (!start_mac_auto_allocate()) {
			rprintf(resp, MSG_ERR_MAC_SOCKET_CREATION_FAILED);
			rprintf(resp, MSG_ERR_CHECK_NETWORK);
			return (CLI_RESULT_FAIL);
		}
	}

	if (!config_mod_allowed(ldp, DEL_RECONFIG_ADD, DR_OP_TYPE_VIO, resp))
		return (CLI_RESULT_FAIL);

	vsw_nodep = get_vsw_node(servicep);
	/*
	 * If there is no corresponding vswitch in existence, create_vnet_node()
	 * will create a stub entry.  We need a reference to that stub in case
	 * we get an error in the backend and need to rollback & delete the
	 * vnet node (& the stub vsw node).  That's why we pass in &vsw_nodep.
	 */
	rv = create_vnet_node(&vnet_nodep, &vsw_nodep, ldp->namep, servicep,
	    vnet_namep, NULL, mac_address);
	if (rv != 0) {
		rprintf(resp, gettext(
		    "Automatic MAC allocation failed.  Please "
		    "set the vnet MAC address manually\n"));
		DBG(print_mac_list());
		/* We return false here so that the above message gets shown */
		return (CLI_RESULT_FAIL);
	}

	if (phys_state != 0) {
		vsw = vsw_find_by_name(servicep);
		if (vsw == NULL) {
			rprintf(resp, gettext("Virtual switch \"%s\" was not "
			    "found\n"), servicep);
			return (CLI_RESULT_FAIL);
		} else if (vsw->phys_dev == NULL && ldp->state >= BOUND) {
			rprintf(resp, MSG_ERR_LINKPROP_NO_NET_DEV, servicep);
			return (CLI_RESULT_FAIL);
		}
	}

	/*
	 * Call backend routine to process
	 */
	rv = vnet_add(ldp, vnet_namep, servicep, vnet_nodep->mac_addr,
	    vnet_mode, pvid, vids, mtu, phys_state, networkid, resp);
	if (rv  != 0) {
		if (rv != ERR_DR_VIO_STILL_ALLOCATED) {
			/*
			 * Delete vnet node & vnet's entry in vsw node.  We
			 * can't utilize remove_vnet_mac() here, because it's
			 * possible we failed because vnet_namep already
			 * exists.  We therefore need to explicitly prune any
			 * empty vsw_nodep here.
			 */
			delete_vnet(vsw_nodep, vnet_nodep);
			remove_empty_vsw(vsw_nodep);
		}

		return (CLI_RESULT_FAIL);
	}

	event_send_res_event(XML_ADD_VNET, ldp, NULL, NULL, clientp);

	database_save();

	DBG(print_mac_list());

	return (CLI_RESULT_OK);
}


/*
 *	ldm set-vnet [mac-addr=<num>] [mode=[hybrid]] [pvid=[<pvid>]]
 *		[vid=[<vid1,vid2,...>]] [mtu=[<mtu>]] [linkprop=[phys-state]]
 *		[vswitch=<vswitch_name>] <if_name> <ldom>
 */

cli_result_t
cli_vnet_set(int argc, char **argvp, cli_conn_t *clientp)
{
	mac_addr_t	mac_address = ILLEGAL_MAC_ADDR, old_mac_addr;
	char		*switch_namep = NULL, *old_switch_namep;
	char		*vnet_namep = NULL;
	ldom_t		*ldp;
	response_t 	*resp = clientp->resp;
	vnet_mac_t	*vnet_nodep = NULL;
	vnet_t		*vnet;
	int		i, rv;
	char		*vnet_mode = NULL;
	int		pvid = -1;
	char		*vids = NULL;
	int		mtu = VIO_NOARGSUP;
	uint64_t	phys_state = ILLEGAL_LNKP;
	vsw_t		*vsw;

	if (argc <= 2)
		return (CLI_RESULT_USAGE);

	/* Optional arguments processed first */
	for (i = 0; i < argc-2; i++) {
		char *namep;
		char *valuep;

		namep = strtok(argvp[i], "=");
		valuep = strtok(NULL, "=");

		if (strcmp(namep, "mac-addr") == 0) {
			if (valuep == NULL || !atomac(valuep, &mac_address)) {
				rprintf(resp, gettext("Invalid mac-addr\n"));
				return (CLI_RESULT_FAIL);
			}

			/*
			 * Can MAC addr be allocated?
			 */
			if (IS_PUBLIC_MAC_ADDR(mac_address)) {
				if (is_macaddr_in_use(mac_address,
				    PERFORM_LOCAL_MAC_CHECK, resp)) {
					return (CLI_RESULT_FAIL);
				}
			}
		} else if (strcmp(namep, "vswitch") == 0) {
			if (!is_valid_object_name(resp, valuep)) {
				rprintf(resp, gettext(
				    "Invalid name for vswitch: %s\n"), valuep);
				return (CLI_RESULT_FAIL);
			}
			switch_namep = valuep;
		} else if (strcmp(namep, "mode") == 0) {
			vnet_mode = valuep;
			if ((vnet_mode != NULL) && (strncmp(vnet_mode,
			    VIO_MODE_HYBRID, strlen(VIO_MODE_HYBRID)) != 0)) {
				rprintf(resp, gettext("Invalid mode: %s\n"),
				    vnet_mode);
				return (CLI_RESULT_FAIL);
			}
			if (vnet_mode == NULL)
				vnet_mode = "";
		} else if (strcmp(namep, "pvid") == 0) {
			if (valuep == NULL) {
				pvid = 0;
			} else {
				char *esp;
				pvid = strtoll(valuep, &esp, 10);
				if ((esp == valuep) || (esp[0] != '\0')) {
					rprintf(resp, gettext("Incorrect pvid "
					    "specification\n"));
					return (CLI_RESULT_FAIL);
				} else if (!VID_IN_RANGE(pvid)) {
					rprintf(resp, gettext("port-vlan-id "
					    "must be between %d and %d\n"),
					    VID_MIN, VID_MAX);
					return (CLI_RESULT_FAIL);
				}
			}
		} else if (strcmp(namep, "vid") == 0) {
			vids = valuep;
			if (valuep == NULL)
				vids = _AUTO_STR_;
		} else if (strcmp(namep, "mtu") == 0) {
			if (valuep == NULL) {
				mtu = VIO_NONEMTU;
			} else {
				char *esp;
				mtu = strtoll(valuep, &esp, 10);
				if ((esp == valuep) || (esp[0] != '\0')) {
					rprintf(resp, gettext("Incorrect MTU "
					    "specification\n"));
					return (CLI_RESULT_FAIL);
				} else if (!MTU_IN_RANGE(mtu)) {
					rprintf(resp, gettext("MTU must be "
					    "between %d and %d\n"),
					    ETHERMTU_MIN, ETHERMTU_MAX);
					return (CLI_RESULT_FAIL);
				}
			}
		} else if (strcmp(namep, "linkprop") == 0) {
			if (valuep == NULL) {
				phys_state = 0;
			} else if (strcmp(valuep, "phys-state") == 0) {
				phys_state = 0x1;
			} else {
				rprintf(resp, MSG_ERR_LINKPROP_INVALID_VAL);
				return (CLI_RESULT_FAIL);
			}
		} else {
			/* print error and usage */
			if (valuep != NULL)
				rprintf(resp, gettext(
				    "Unknown property: %s\n"), namep);
			return (CLI_RESULT_USAGE);
		}
	}

	if ((argc-i) != 2)
		return (CLI_RESULT_USAGE);

	vnet_namep = argvp[i];
	if (!is_valid_object_name(resp, vnet_namep)) {
		rprintf(resp, gettext("Illegal name for VNET interface: %s\n"),
		    vnet_namep);
		return (CLI_RESULT_FAIL);
	}

	if ((ldp = ldm_get_ldp(resp, argvp[i+1])) == NULL)
		return (CLI_RESULT_FAIL);

	/*
	 * At this stage, we don't know if this operation would trigger
	 * a delayed reconfig operation so we make the DR check a no-op
	 * (using DR_OP_TYPE_VIO_MD_NOP) and call config_mod_allowed()
	 * again in vnet_modify().
	 */
	if (config_mod_allowed(ldp, DEL_RECONFIG_OTHER,
	    DR_OP_TYPE_VIO_MD_NOP, resp) == false)
		return (CLI_RESULT_FAIL);

	vnet_nodep = get_vnet_node(ldp->namep, vnet_namep);
	if (vnet_nodep == NULL) {
		rprintf(resp, gettext(
		    "Network interface named %s not found in LDom %s\n"),
		    vnet_namep, ldp->namep);
		return (CLI_RESULT_FAIL);
	}

	if ((vnet = vnet_find_by_name(ldp, vnet_namep)) == NULL) {
		rprintf(resp, gettext("%s named %s not found\n"),
		    vnet_attrs.cli_name, vnet_namep);
		return (CLI_RESULT_FAIL);
	}

	old_mac_addr = vnet_nodep->mac_addr;
	old_switch_namep = Xstrdup(vnet_nodep->service_name);

	rv = update_vnet_node(&vnet_nodep, switch_namep, mac_address, NULL);
	if (rv != 0) {
		rprintf(resp, gettext("Automatic MAC allocation failed.\n"));
		DBG(print_mac_list());
		Xfree(old_switch_namep);
		return (CLI_RESULT_FAIL);
	}

	if (phys_state == 0x1 && ldp->state >= BOUND) {
		if (switch_namep != NULL)
			vsw = vsw_find_by_name(switch_namep);
		else
			vsw = vsw_find_by_name(vnet->service);
		if (vsw == NULL || vsw->phys_dev == NULL) {
			rprintf(resp, MSG_ERR_LINKPROP_NO_NET_DEV,
			    (switch_namep == NULL) ? vnet->service :
			    switch_namep);
			return (CLI_RESULT_FAIL);
		}
	}

	if (vnet_modify(vnet, switch_namep, vnet_nodep->mac_addr, vnet_mode,
	    pvid, vids, mtu, phys_state, resp) != 0) {
		update_vnet_node(&vnet_nodep, old_switch_namep, old_mac_addr,
		    NULL);
		Xfree(old_switch_namep);
		return (CLI_RESULT_FAIL);
	}

	database_save();

	DBG(print_mac_list());

	event_send_res_event(XML_SET_VNET, ldp, NULL, NULL, clientp);

	Xfree(old_switch_namep);
	return (CLI_RESULT_OK);
}


/*
 *	ldm remove-vnet [-f | --force] <if_name> <ldom>
 */

cli_result_t
cli_vnet_remove(int argc, char **argvp, cli_conn_t *clientp)
{
	response_t *resp = clientp->resp;
	int i = 0;
	bool_t force = false;
	char *vnet_namep, *ldom_namep;
	ldom_t *ldp = NULL;
	vnet_t *vnet;

	if (argc == 0)
		return (CLI_RESULT_USAGE);

	if (argvp[i][0] == '-') {
		if (strcmp(argvp[i], "-f") == 0 ||
		    strcmp(argvp[i], "--force") == 0) {
			force = true;
			i++;
		} else {
			return (CLI_RESULT_USAGE);
		}
	}

	if (argc - i != 2)
		return (CLI_RESULT_USAGE);

	vnet_namep = argvp[i];
	if (!is_valid_object_name(resp, argvp[i])) {
		rprintf(resp, gettext("Illegal VNET name\n"));
		return (CLI_RESULT_FAIL);
	}
	ldom_namep = argvp[i+1];
	if (!is_valid_object_name(resp, argvp[i])) {
		rprintf(resp, gettext("Illegal LDom name: %s\n"), ldom_namep);
		return (CLI_RESULT_FAIL);
	}

	if ((ldp = ldm_get_ldp(resp, ldom_namep)) == NULL)
		return (CLI_RESULT_FAIL);

	if (!config_mod_allowed(ldp, DEL_RECONFIG_REMOVE_2PHASE,
	    DR_OP_TYPE_VIO, resp))
		return (CLI_RESULT_FAIL);

	if ((vnet = vnet_find_by_name(ldp, vnet_namep)) == NULL) {
		rprintf(resp, gettext("%s named %s not found\n"),
		    vnet_attrs.cli_name, vnet_namep);
		return (CLI_RESULT_FAIL);
	}

	if (vnet_remove(vnet, force, resp) != 0) {
		rprintf(resp, gettext("Failed to remove VNET instance\n"));
		return (CLI_RESULT_FAIL);
	}

	event_send_res_event(XML_REMOVE_VNET, ldp, NULL, NULL, clientp);
	database_save();

	remove_vnet_mac(ldp->namep, vnet_namep);
	DBG3(print_mac_list());

	return (CLI_RESULT_OK);
}
