/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <string.h>

#include <thread.h>
#include <synch.h>
#include <libintl.h>
#include <sys/time.h>

#include "allocate.h"
#include "basics.h"
#include "backend.h"
#include "chain.h"
#include "database.h"
#include "frags.h"
#include "hvctl.h"
#include "keys.h"
#include "ldmd_types.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "mac_addrs.h"
#include "master.h"
#include "resource_id.h"
#include "response.h"
#include "uuid.h"
#include "variables.h"
#include "vcpu.h"
#include "vio.h"
#include "vldc.h"
#include "sequencer.h"
#include "xml_event.h"
#include "xml_cntrl_common.h"
#include "zpri.h"

/*
 * Maximum vcpus in each guest. This is the number of
 * physical strands multiplied by the sub-strand scheduling
 * granularity, which we define as 1 (until this is derived
 * from HV)
 */
#define	VCPU_SCHED_GRAN	(1)
#define	MAX_GUEST_VCPUS (VCPU_SCHED_GRAN * (pri.max_guest_vids))

/*
 * Maximum vmaus in each guest. MAUS are currently allocated
 * as a whole.
 */
#define	VMAU_SCHED_GRAN	(1)
#define	MAX_GUEST_VMAUS (VMAU_SCHED_GRAN * (pri.num_maus))


zeus_t zeus;

ldom_t *
find_ldom_by_name(char *namep)
{
	ldom_t *ldp;

	if (namep == NULL)
		return (NULL);

	ldp = CHAIN_START(zeus.ldoms);
	while (ldp != NULL && strcmp(ldp->namep, namep) != 0)
		ldp = CHAIN_NEXT(ldp->link);

	return (ldp);
}

ldom_t *
find_ldom_by_gid(uint64_t id)
{
	ldom_t *ldp;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state != INACTIVE && ldp->idp != NULL &&
		    ldp->idp->id == id)
			break;
	}

	return (ldp);
}

ldom_t *
find_ldom_by_resid(uint64_t id)
{
	ldom_t *ldp;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state != INACTIVE && ldp->res_id != NULL &&
		    ldp->res_id->id == id)
			break;
	}

	return (ldp);
}

bool_t
check_locked_ldom(ldom_t *ldp, response_t *resp)
{
	if (IS_LDOM_LOCKED(ldp) == true) {
		rprintf(resp, "LDom %s is locked against modification\n",
		    ldp->namep);
		return (true);
	} else {
		return (false);
	}
}

/*
 * Physical I/O
 */

req_iodevice_t *
add_io_constraint(ldom_t *ldp, char *name, bool_t io_mmu_bypass)
{
	req_iodevice_t *iodevs_p;

	iodevs_p = Xcalloc(1, req_iodevice_t);
	iodevs_p->name = Xstrdup(name);
	iodevs_p->bound = false;
	iodevs_p->io_mmu_bypass = io_mmu_bypass;
	CHAIN_ADD(ldp->required.io_devices, iodevs_p, link);
	return (iodevs_p);
}

/*
 * Comparison function for sorting ldoms by name. The primary domain always
 * comes first and then the others alphabetically.
 */
static int
ldom_name_cmp(ldom_t *a, ldom_t *b)
{
	if (strcmp(a->namep, b->namep) == 0)
		return (0);
	if (strcmp(a->namep, "primary") == 0)
		return (-1);
	if (strcmp(b->namep, "primary") == 0)
		return (1);

	return (strcmp(a->namep, b->namep));
}

ldom_t *
new_ldom(char *namep)
{
	ldom_t *ldp, *iter, *prev;
	int max_ds_ports = (strcmp(namep, SOD_CONTROL_DOMAIN) == 0) ?
	    MAX_DS_PORTS : 1; /* only primary needs a lot of DS ports */

	ldp = Xmalloc(sizeof (ldom_t));

	ldp->namep = Xstrdup(namep);
	ldp->idp = NULL;
	ldp->res_id = NULL;

	/* initialize HV state */
	ldp->hv_state.state = GUEST_STATE_UNCONFIGURED;
	ldp->hv_state.soft = SIS_TRANSITION;
	ldp->hv_state.soft_update = 1;
	memset(ldp->hv_state.soft_str, 0, SOFT_STATE_SIZE);

	CHAIN_INIT(ldp->required.io_devices);
	CHAIN_INIT(ldp->required.io_pcie_devs);
	ldp->required.num_vcpus = 0;
	ldp->required.num_vmaus = 0;
	ldp->required.size_mem = 0;

	/* initialize bindings */
	ldp->props.md_generation_num = 0;
	ldp->props.mem_generation_id = 0;
	ldp->props.mac_address = ILLEGAL_MAC_ADDR;
	ldp->props.hostid = UNASSIGNED_HOSTID;
	ldp->props.tod_offset = 0;
	ldp->props.diagpriv = 0;
	ldp->props.perfctraccess = 0;
	ldp->props.perfctrhtaccess = 0;
	ldp->props.rngctlaccessible = 0;

	/*
	 * Use final RA segment for LDC shared memory. Use actual ra2pa shift
	 * values so base & size are consistent between sparse-memory and non
	 * sparse-memory systems.
	 */
	ldp->props.ldc_mapin_base = (zeus.guest_max_mblocks) <<
	    zeus.ra2pa_shift;
	ldp->props.ldc_mapin_size = 1ULL << pri.ra2pa_shift;

	ldp->bound.num_services = 0;
	ldp->bound.num_clients = 0;
	CHAIN_INIT(ldp->bound.vcpus);
	CHAIN_INIT(ldp->bound.vmaus);
	CHAIN_INIT(ldp->bound.memory);
	CHAIN_INIT(ldp->bound.committed_memory.blocks);
	CHAIN_INIT(ldp->bound.iodevices);
	init_keys(&ldp->keys);
	init_variables(&ldp->variables);
	ldp->devalias_md = NULL;
	ldp->ds_sp_md = NULL;

	CHAIN_INIT(ldp->susp.vcpus);
	CHAIN_INIT(ldp->susp.vmaus);

	/*
	 * Need to capture memory allocation restrictions placed on us
	 * by the boot image before we try to allocate any memory to
	 * the LDom.  So we setup an initial default boot image here.
	 * FIXME: provide CLI for user to select a different boot image.
	 */
	if (bind_boot_options(ldp, NULL) == false)
		ldp->boot_imagep = NULL;

	/* Initialize links */
	CHAIN_LINK_INIT(ldp->link);
	CHAIN_LINK_INIT(ldp->inprog);

	/* Initialize statistics */
	ldp->stats.uptime = 0;
	ldp->stats.cpu = 0;
	ldp->stats.hist_idx = 0;
	(void) memset(ldp->stats.hist_table, 0, sizeof (ldp->stats.hist_table));

	/* Initialize empty container */
	ldp->state = INACTIVE;
	ldp->flag.md_notify = false;
	ldp->flag.gen_new_md = false;
	ldp->flag.locked = false;
	ldp->flag.sp_ldc_cons = false;
	ldp->flag.config_committed = true;
	ldp->flag.migration_source = false;
	ldp->flag.migration_target = false;
	ldp->flag.migration_error = false;
	ldp->flag.local_tick_sync = false;
	ldp->flag.memdr_in_progress = false;
	ldp->mem_dr_state.tid = NULL;
	create_res_pool(&ldp->guest_vcpu_ids, 0, MAX_GUEST_VCPUS - 1,
	    Res_sorted);
	create_res_pool(&ldp->guest_vmau_ids, 0, MAX_GUEST_VMAUS - 1,
	    Res_sorted);
	create_res_pool(&ldp->guest_free_ldcs, 0, pri.max_guest_ldcs - 1,
	    Res_sorted);
	ldp->ds_chan = NULL;
	CHAIN_INIT(ldp->sp_ldcs);
	CHAIN_INIT(ldp->waiters);
	CHAIN_INIT(ldp->vccs);
	CHAIN_INIT(ldp->vsws);
	CHAIN_INIT(ldp->vnets);
	CHAIN_INIT(ldp->vdss);
	CHAIN_INIT(ldp->vdisks);
	CHAIN_INIT(ldp->vdpcss);
	CHAIN_INIT(ldp->vdpccs);
	CHAIN_INIT(ldp->vldcs);
	CHAIN_INIT(ldp->vldccs);
	CHAIN_INIT(ldp->ds_ports);
	ldp->console.ldp = ldp;
	ldp->console.service_name = NULL;
	ldp->console.group_name = NULL;
	ldp->console.tcp_port = _AUTO_VAL_;
	ldp->console.binding = NULL;

	/* Initialize I/O related structures */
	create_res_pool(&ldp->free_hvcfg_handles, 0, IDENT_MAX, Res_sorted);
	create_res_pool(&ldp->free_ds_ports, 0, max_ds_ports, Res_sorted);
#if VDEV_CFG_HANDLE
	create_res_pool(&ldp->free_cfg_handles, 0, IDENT_MAX, Res_sorted);
#endif

	/* Initialize bound memory fragments */
	ldp->frags.committed_md_frag = NULL;
	ldp->frags.parsed_md_frag = NULL;
	ldp->frags.next_md_frag = NULL;
	ldp->restore.committed_pa = INVALID_FRAG_PA;
	ldp->restore.parsed_pa = INVALID_FRAG_PA;
	ldp->restore.next_pa = INVALID_FRAG_PA;

	/* Initialize PM fields */
	ldp->pm_cpu_resutils.ldmr_type = PM_RES_CPU;
	ldp->pm_cpu_resutils.ldmr_time = 0;
	ldp->pm_cpu_resutils.ldmr_util_dir = PM_UTIL_WITHIN;
	ldp->pm_vcpu_count = 0;

	/* Initialize migration information */
	ldp->mig_info.peer_name = NULL;
	ldp->mig_state.cancel = false;
	ldp->mig_state.suspended = false;
	ldp->mig_state.bind_mdp = NULL;
	ldp->mig_state.bind_fp = NULL;
	CHAIN_INIT(ldp->mig_rsvd.vcpus);

	/* Initialize guestdependency */
	ldp->props.failure_policy = INVALID_FAILURE_POLICY;
	CHAIN_INIT(ldp->props.masters);

	database_ldom_create(ldp);

	/* debug aid */
	name_ldom_res_pool(&ldp->guest_free_ldcs, ldp, "guest_free_ldcs");
	name_ldom_res_pool(&ldp->guest_vcpu_ids, ldp, "guest_vcpu_ids");
	name_ldom_res_pool(&ldp->guest_vmau_ids, ldp, "guest_vmau_ids");

	mutex_lock(&zeus.mlock);
	CHAIN_ADD_SORTED(zeus.ldoms, ldp, iter, prev, link, ldom_name_cmp);
	mutex_unlock(&zeus.mlock);

	/* Initialize DRM policy */
	CHAIN_INIT(ldp->pchain.policy);
	mutex_init(&ldp->pchain.lock, USYNC_THREAD, NULL);

	return (ldp);
}

/*
 * Assumes name and MAC address already validated.
 */
ldom_t *
ldom_create(char *name, mac_addr_t mac_address, uint64_t hostid,
    uint64_t failure_policy, mlist_t *mlist)
{
	ldom_t		*ldp;

	ldp = new_ldom(name);
	ldp->props.mac_address = mac_address;
	ldp->props.hostid = hostid;
	ldp->props.failure_policy = failure_policy;
	mlist2master(ldp, mlist);

	/* create vldc & domain_service if relevant */
	new_ldom_vldc(ldp, NULL);

	/* add uuid for new ldoms */
	new_ldom_uuid(ldp);

	database_ldom_set_mac_addr(ldp, mac_address);
	database_ldom_set_hostid(ldp, hostid);
	database_ldom_set_failure_policy(ldp, failure_policy);
	database_ldom_set_master(ldp);
	database_ldom_set_uuid(ldp);
	database_save();

	return (ldp);
}

/*
 * Translate LDom state value to a string.
 */
char *
ldom_state_str(ldom_state_t state)
{
	char *statep;

	switch (state) {
	case INACTIVE:
		statep = gettext("inactive");
		break;
	case BINDING:
		statep = gettext("binding");
		break;
	case UNBINDING:
		statep = gettext("unbinding");
		break;
	case BOUND:
		statep = gettext("bound");
		break;
	case STARTING:
		statep = gettext("starting");
		break;
	case STOPPING:
		statep = gettext("stopping");
		break;
	case SUSPENDING:
		statep = gettext("suspending");
		break;
	case RESUMING:
		statep = gettext("resuming");
		break;
	case SUSPENDED:
		statep = gettext("suspended");
		break;
	case ACTIVE:
		statep = gettext("active");
		break;
	default:
		statep = gettext("unknown");
		break;
	}

	return (statep);
}

/*
 * Returns true iff it is allowed to reset all IO devices in domain 'ldp'.
 */
bool_t
ldom_io_reset_allowed(ldom_t *ldp)
{
	bound_iodevice_t *bip;

	for (bip = CHAIN_START(ldp->bound.iodevices); bip != NULL;
	    bip = CHAIN_NEXT(bip->ldom)) {
		/*
		 * Check both the new per device no_pcie_reset property and
		 * the old global property.
		 */
		if (bip->no_pcie_reset == true || pri.no_pcie_reset == true)
			return (false);
	}

	/*
	 * Either there are no IO devices in the domain or all IO devices in
	 * the domain are resettable.
	 */
	return (true);
}

/*
 * Arguments:
 *	ldp	- pointer to domain's internal structure
 *	resp	- UNUSED; if used in future, will have to handle case
 *		  where it is NULL (if caller has no resp available).
 */
/*ARGSUSED1*/
bool_t
destroy_one_ldom(ldom_t *ldp, response_t *resp, cli_conn_t *client)
{
	if (check_locked_ldom(ldp, resp))
		return (false);

	/* free up VIO client/server structures */
	vio_delete(ldp);

	/*
	 * If we're destroying the LDom running Zeus (i.e. we're about
	 * to shoot ourselves), at least try to be orderly about it.
	 */
	if (zeus.hvctl_ldom_p != NULL && IS_CONTROL_DOMAIN(ldp) == true)
		zeus.hvctl_ldom_p = NULL;

	database_ldom_destroy(ldp);
	remove_ldom_macs(ldp->namep);
	invalidate_slaves(ldp);
	event_send_dom_event(XML_DELETE, ldp, client);
	free_ldom(ldp);
	return (true);
}

void
cleanse_progress_chain(void)
{
	ldom_t *ldp, *nldp;

	for (ldp = CHAIN_START(zeus.inprogress); ldp != NULL; ldp = nldp) {
		nldp = CHAIN_NEXT(ldp->inprog);
		CHAIN_LINK_INIT(ldp->inprog);
	}

	CHAIN_INIT(zeus.inprogress);
}

double
avg_guest_util(ldom_t *ldp, hrtime_t current_time)
{
	int i = ldp->stats.hist_idx;
	int num_values = 0;
	double total_util = 0;
	hrtime_t util_time;

	/*
	 * Loop through the utilization history table to total the values.
	 * Only include the values that are within GUEST_UTIL_HIST_TIME.
	 * It's possible, though very unlikely, that no values are within
	 * GUEST_UTIL_HIST_TIME.  In this case, just use the most recent value
	 * for total_util.
	 */
	do {
		total_util += ldp->stats.hist_table[i].value;

		num_values++;

		/* move to the next most recent entry in the history table */
		i = GET_PREV_GUEST_UTIL_IDX(i);

		util_time = ldp->stats.hist_table[i].time;

	} while (((current_time - util_time) < GUEST_UTIL_HIST_TIME) &&
	    (num_values < GUEST_UTIL_TAB_NROWS));

	return (total_util / num_values);
}

double
avg_cpu_util(vcpu_t *vcpu, hrtime_t current_time)
{
	int i = vcpu->util_idx;
	int num_values = 0;
	double total_util = 0;
	hrtime_t util_time;

	/*
	 * Loop through the utilization history table to total the values.
	 * Only include the values that are within VCPU_UTIL_HIST_TIME.
	 * It's possible, though very unlikely, that no values are within
	 * VCPU_UTIL_HIST_TIME.  In this case, just use the most recent value
	 * for total_util.
	 */
	do {
		ASSERT(i <= VCPU_UTIL_TAB_NROWS);
		total_util += vcpu->util_table[i].value;

		num_values++;

		/* move to the next most recent entry in the history table */
		if (i == 0)
			i = VCPU_UTIL_TAB_NROWS - 1;
		else
			i--;

		util_time = vcpu->util_table[i].time;

	} while (((current_time - util_time) < VCPU_UTIL_HIST_TIME) &&
	    (num_values < VCPU_UTIL_TAB_NROWS));

	return (total_util / num_values);
}

void
update_statistics(ldom_t *ldp, bool_t do_cpus)
{
	hrtime_t t;
	vcpu_t *vcpu;

	ASSERT_IN_SEQUENCER();

	if (ldp->state <= BOUND)
		return;

	t = gethrtime();

	if (do_cpus) {
		CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
			vcpu->util = avg_cpu_util(vcpu, t);
		}
	}

	ldp->stats.cpu = avg_guest_util(ldp, t);

	if (ldp->hv_state.soft_update)
		backend.get_guest_state(ldp, LDOM_SOFT_STATE, &ldp->hv_state);
}
