/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <dirent.h>
#include <dlfcn.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include "allocate.h"
#include "chain.h"
#include "ds.h"
#include "ds_impl.h"
#include "fatal.h"
#include "ldom.h"
#include "libmgr.h"
#include "support.h"
#include "svcs.h"
#include "zprintf.h"

#define	SVCDIR "lib/ds"

typedef svc_info_t *(*svc_info_f)(void);
typedef svc_ops_t *(*svc_ops_f)(char *);

libds_ops_t libds_ops;

libds_flag_t libds_cap;

struct libds_plugins {
	svc_info_t *info;
	CHAIN_LINK(struct libds_plugins, link);
};

CHAIN(svc_t, all_services) = DEFAULTCHAIN;
CHAIN(channel_plugins_t, channel_plugins) = DEFAULTCHAIN;
CHAIN(struct libds_plugins, libds_list) = DEFAULTCHAIN;

static uint64_t svc_handles_base;		/* base for service handles */
static void *dl_hdl;

void
check_libds_support(void)
{
	if ((dl_hdl = dlopen("libds.so.1", RTLD_NOW)) == NULL) {
		DBG(zprintf("libds.so.1 could not be loaded: Userland domain "
		    "service library isn't available\n"));
		return;
	}

	libds_ops.ds_init = (ds_init_f)dlsym(dl_hdl, "ds_init");
	libds_ops.ds_fini = (ds_fini_f)dlsym(dl_hdl, "ds_fini");
	libds_ops.ds_svc_reg = (ds_svc_reg_f)dlsym(dl_hdl, "ds_svc_reg");
	libds_ops.ds_clnt_reg = (ds_clnt_reg_f)dlsym(dl_hdl, "ds_clnt_reg");
	libds_ops.ds_isready = (ds_isready_f)dlsym(dl_hdl, "ds_isready");
	libds_ops.ds_hdl_lookup = (ds_hdl_lookup_f)dlsym(dl_hdl,
	    "ds_hdl_lookup");
	libds_ops.ds_send_msg = (ds_send_msg_f)dlsym(dl_hdl, "ds_send_msg");
	libds_ops.ds_recv_msg = (ds_recv_msg_f)dlsym(dl_hdl, "ds_recv_msg");
	libds_ops.ds_unreg_svc = (ds_unreg_svc_f)dlsym(dl_hdl, "ds_unreg_svc");
	libds_ops.ds_unreg_hdl = (ds_unreg_hdl_f)dlsym(dl_hdl, "ds_unreg_hdl");
	libds_ops.ds_dom_name_to_hdl = (ds_dom_name_to_hdl_f)dlsym(dl_hdl,
	    "ds_dom_name_to_hdl");
	libds_ops.ds_dom_hdl_to_name = (ds_dom_hdl_to_name_f)dlsym(dl_hdl,
	    "ds_dom_hdl_to_name");
	libds_ops.ds_domain_lookup = (ds_domain_lookup_f)dlsym(dl_hdl,
	    "ds_domain_lookup");

	if (libds_ops.ds_init == NULL || libds_ops.ds_svc_reg == NULL ||
	    libds_ops.ds_clnt_reg == NULL)
		fatal("libds does not provide one of ds_init(), ds_svc_reg() "
		    "or ds_clnt_reg()");


	DBG(zprintf("libds detected, may be used for DS communication\n"));
	libds_cap.supported = true;
}

static void
add_svc_to_channel_plugin(channel_plugins_t *plugins, svc_t *svc)
{
	CHAIN_ADD(plugins->plugged_svcs, svc, chanlink);
	if (svc->props & SVC_OUTGOING)
		CHAIN_ADD(plugins->outgoing, svc, outlink);
}

void
free_dsinfo(libds_info_t *dsinfo)
{
	free(dsinfo->cap->vers);
	free(dsinfo->cap->svc_id);
	free(dsinfo->cap);
	free(dsinfo->ds_ops);
	free(dsinfo);
}

static int
compare_prot_version_ptr(const void *a, const void *b)
{
	prot_version_t *ver_a = *(prot_version_t **)a;
	prot_version_t *ver_b = *(prot_version_t **)b;

	if (ver_a->major_number > ver_b->major_number)
		return (-1);
	else if (ver_a->major_number < ver_b->major_number)
		return (1);

	if (ver_a->minor_number > ver_b->minor_number)
		return (-1);
	else if (ver_a->minor_number < ver_b->minor_number)
		return (1);
	else
		return (0);
}

static void
sort_protocol_versions(svc_t *svc)
{
	svc_plugin_info_t *plugin_info;
	prot_version_t **p;
	int i, num_vers;

	num_vers = 0;
	CHAIN_ITER(svc->plugins, plugin_info, link) {
		num_vers += plugin_info->num_vers;
	}

	svc->versions = Xmalloc(num_vers * sizeof (prot_version_t *));
	svc->num_vers = num_vers;

	p = svc->versions;
	CHAIN_ITER(svc->plugins, plugin_info, link) {
		for (i = 0; i < plugin_info->num_vers; i++)
			*p++ = &plugin_info->versions[i];
	}

	qsort(svc->versions, num_vers, sizeof (prot_version_t *),
	    compare_prot_version_ptr);
}

svc_t *
find_plugged_svc_by_name(channel_plugins_t *channel_plugin, char *svc_name)
{
	svc_t *svc = NULL;

	CHAIN_ITER(channel_plugin->plugged_svcs, svc, chanlink) {
		if (strcmp(svc->name, svc_name) == 0) {
			break;
		}
	}

	return (svc);
}

channel_plugins_t *
find_channel_plugins_by_name(char *chan_name)
{
	channel_plugins_t *plugins;

	CHAIN_ITER(channel_plugins, plugins, link) {
		if (strcmp(plugins->chan_name, chan_name) == 0)
			return (plugins);
	}

	return (NULL);
}

static channel_plugins_t *
get_channel_plugin(char *chan_name)
{
	channel_plugins_t *plugins;

	plugins = find_channel_plugins_by_name(chan_name);
	if (plugins == NULL) {
		plugins = Xcalloc(1, channel_plugins_t);
		plugins->chan_name = Xstrdup(chan_name);
		CHAIN_INIT(plugins->plugged_svcs);
		CHAIN_INIT(plugins->outgoing);
		CHAIN_ADD(channel_plugins, plugins, link);
	}

	return (plugins);
}

int
load_svc_plugin(ds_svc_t *ds)
{
	svc_plugin_info_t *plugin_info;
	svc_ops_f opsfn;
	int i;

	CHAIN_ITER(ds->svc->plugins, plugin_info, link) {
		for (i = 0; i < plugin_info->num_vers; i++) {
			if (ds->version.major_number ==
			    plugin_info->versions[i].major_number) {
				ds->plugin_dlhdl =
				    load_library(plugin_info->filename);
				opsfn = (svc_ops_f)dlsym(ds->plugin_dlhdl,
				    "ops");
				ds->ops = opsfn(ds->svc->name);
			}
		}
	}

	return (0);
}

static int
cross_check(svc_t *svc, svc_info_t *info, int svcidx)
{
	svc_plugin_info_t *plugged;
	int i, j;

	if (svc->props != info->props[svcidx]) {
		warning("Plugins for service \"%s\" have inconsistent props",
		    svc->name);
		return (-1);
	}
	for (i = 0; i < info->num_chans; i++) {
		if (strcmp(svc->chan_name, info->chan_names[i]) == 0)
			break;
	}

	if (i == info->num_chans) {
		warning("Plugins for service \"%s\" have inconsistent channel "
		    "names", svc->name);
		return (-1);
	}

	/*
	 * Check that major numbers supported in info for the service are not
	 * supported in any other plugin files.
	 */
	CHAIN_ITER(svc->plugins, plugged, link) {
		for (i = 0; i < plugged->num_vers; i++) {
			for (j = 0; j < info->num_vers; j++) {
				if (plugged->versions[i].major_number ==
				    info->versions[j].major_number) {
					warning("Major version %u of service %s"
					    " is supported in multiple files",
					    info->versions[j].major_number,
					    svc->name);
					return (-1);
				}
			}
		}
	}

	return (0);
}

#if ZDBG
static void
dump_plugin_info(svc_t *svc, svc_plugin_info_t *plugin_info)
{
	int i;

	DBG(zprintf("Added to channel %s, service %s, versions ",
	    svc->chan_name, svc->name));
	for (i = 0; i < plugin_info->num_vers; i++)
		DBG(zprintf("%d.%d ", plugin_info->versions[i].major_number,
		    plugin_info->versions[i].minor_number));
	DBG(zprintf("\n"));
}
#endif

static char *ds_channels[] = {
    DS_CHAN_NAME,
    LIBDS_CHAN_NAME,
    SP_CHAN_NAME,
    FMA_CHAN_NAME,
    NULL
};

/*
 * Validate that the channels specified in the plugin are known and there
 * aren't any duplicates
 */
static bool_t
validate_plugin(svc_info_t *info)
{
	int i = 0;
	int val = 0;
	char **chans = info->chan_names;
	int nchans = info->num_chans;
	char **svcs = info->svcs;
	int nsvcs = info->num_svcs;

	/*
	 * validate channel name array
	 */
	if (chans == NULL || nchans <= 0) {
		DBG(zprintf("%s: channels incorrectly specified in plugin\n",
		    __func__));
		return (false);
	}

	/*
	 * validate services name array
	 */
	if (svcs == NULL || nsvcs <= 0) {
		DBG(zprintf("%s: Services incorrectly specified in plugin\n",
		    __func__));
		return (false);
	}

	for (i = 0; i < nchans; i++) {
		int j = 0;

		while (ds_channels[j] != NULL) {
			if (strcmp(chans[i], ds_channels[j]) == 0)
				break;
			j++;
		}
		if (ds_channels[j] == NULL) {
			DBG(zprintf("%s: Invalid channel specified: %s\n",
			    __func__, chans[i]));
			return (false);

		}

		if (val & (1 << j)) {
			DBG(zprintf("%s: Duplicate channel specified: %s\n",
			    __func__, chans[i]));
			return (false);
		} else {
			val |= (1 << j);
		}
	}

	return (true);
}

/*
 * A plugin can provide multiple services from the same library. We
 * create a svc_t structure for each individual service described
 * in the plugin.
 */
static void
add_svc_plugin_info(svc_info_t *info, char *plugin_filename)
{
	svc_t *svc;
	svc_plugin_info_t *plugin_info;
	int i, j;
	channel_plugins_t *channel_plugin;
	char *chan_name;

	for (i = 0; i < info->num_chans; i++) {
		chan_name = info->chan_names[i];
		channel_plugin = get_channel_plugin(chan_name);
		for (j = 0; j < info->num_svcs; j++) {
			svc = find_plugged_svc_by_name(channel_plugin,
			    info->svcs[j]);
			if (svc == NULL) {
				/*
				 * Create a separate instance of the service
				 * for each channel type the service can be
				 * active on
				 */
				svc = Xmalloc(sizeof (*svc));
				svc->name = Xstrdup(info->svcs[j]);
				svc->chan_name = Xstrdup(info->chan_names[i]);
				svc->hdl_base = (svc_handles_base +=
				    (1ULL << 32));
				svc->props = info->props[j];
				svc->num_vers = 0;
				svc->versions = NULL;
				CHAIN_INIT(svc->plugins);
				CHAIN_LINK_INIT(svc->chanlink);
				CHAIN_LINK_INIT(svc->outlink);
				CHAIN_ADD(all_services, svc, link);
				add_svc_to_channel_plugin(channel_plugin, svc);
			} else {
				if (cross_check(svc, info, j) == -1)
					return;
			}
			plugin_info = Xmalloc(sizeof (*plugin_info));
			plugin_info->filename = Xstrdup(plugin_filename);
			plugin_info->num_vers = info->num_vers;
			plugin_info->versions = Xcalloc(info->num_vers,
			    prot_version_t);
			memcpy(plugin_info->versions, info->versions,
			    info->num_vers * sizeof (prot_version_t));
			CHAIN_ADD(svc->plugins, plugin_info, link);

			DBG(dump_plugin_info(svc, plugin_info));
		}
	}
}

static void
free_svc_info(svc_info_t *info)
{
	int i;

	for (i = 0; i < info->num_svcs; i++) {
		free(info->svcs[i]);
		if (info->libds_svcs != NULL)
			free_dsinfo(info->libds_svcs[i]);
	}
	free(info->svcs);
	free(info->props);
	free(info->versions);
	for (i = 0; i < info->num_chans; i++)
		free(info->chan_names[i]);
	free(info->chan_names);
	free(info);
}

int
get_plugged_services(void)
{
	char *svcdir;
	char *plugin_filename;
	svc_info_t *info;
	struct dirent *dp;
	DIR *dirp;
	void *dl_hdl;
	svc_info_f infofn;
	svc_t *svc;

	/* initialize base for service handles */
	srand(time(NULL));
	svc_handles_base = (((uint64_t)rand()) << 32) + 1;

	svcdir = zsprintf("%s/%s", zeus.pkgdir, SVCDIR);

	if ((dirp = opendir(svcdir)) == 0) {
		ewarning("Could not open services plugin directory \"%s\"",
		    svcdir);
		free(svcdir);
		return (0);
	}

	while (1) {
		errno = 0;
		if ((dp = readdir(dirp)) == NULL)
			break;

		if (strcmp(dp->d_name, ".") == 0 ||
		    strcmp(dp->d_name, "..") == 0)
			continue;

		plugin_filename = zsprintf("%s/%s", SVCDIR, dp->d_name);
		ASSERT(plugin_filename != NULL);
		dl_hdl = load_library(plugin_filename);
		if (dl_hdl == NULL) {
			warning("Unable to open service plugin \"%s\": %s",
			    dp->d_name, dlerror());
			free(plugin_filename);
			continue;
		}

		infofn = (svc_info_f)dlsym(dl_hdl, "info");
		if (infofn == NULL || dlsym(dl_hdl, "ops") == NULL) {
			warning("Service plugin \"%s\" is missing a required "
			    "interface", dp->d_name);
			unload_library(dl_hdl);
			free(plugin_filename);
			continue;
		}

		info = infofn();
		unload_library(dl_hdl);

		if (info->versions != NULL) {
		/*
		 * Validate channels specified in plugin
		 */
		if (validate_plugin(info) == false)
			return (1);

			add_svc_plugin_info(info, plugin_filename);
		} else {
			warning("Service plugin \"%s\" does not support any "
			    "protocol versions", dp->d_name);
		}

		/*
		 * If not a libds dependent service, skip the following
		 */
		if (info->libds_svcs != NULL) {
			struct libds_plugins *pl;
			pl = Xcalloc(1, struct libds_plugins);
			pl->info = info;
			CHAIN_ADD(libds_list, pl, link);
		}
		free(plugin_filename);
	}

	if (errno != 0)
		ewarning("Failure reading service plugin directory \"%s\"",
		    svcdir);

	closedir(dirp);
	free(svcdir);

	CHAIN_ITER(all_services, svc, link) {
		sort_protocol_versions(svc);
	}

	return (0);
}

/*
 * Copy the entire libds_info_t structure into zeus memory
 */
libds_info_t *
cp_dsinfo(libds_info_t *dsinfo, char *svc_id)
{
	libds_info_t *ifo = Xcalloc(1, libds_info_t);

	memcpy(ifo, dsinfo, sizeof (libds_info_t));
	ifo->ds_ops = Xcalloc(1, ds_ops_t);
	memcpy(ifo->ds_ops, dsinfo->ds_ops, sizeof (ds_ops_t));
	ifo->ds_ops->cb_arg = Xstrdup(svc_id);
	ifo->cap = Xcalloc(1, ds_capability_t);
	memcpy(ifo->cap, dsinfo->cap, sizeof (ds_capability_t));
	ifo->cap->svc_id = Xstrdup(svc_id);
	ifo->cap->vers = Xcalloc(ifo->cap->nvers, ds_ver_t);
	memcpy(ifo->cap->vers, dsinfo->cap->vers,
	    ifo->cap->nvers * sizeof (ds_ver_t));

	return (ifo);
}

static void
register_libds_plugin_svc(svc_info_t *info, int idx)
{
	ds_capability_t *cap = info->libds_svcs[idx]->cap;
	ds_ops_t *ops = info->libds_svcs[idx]->ds_ops;
	ds_hdl_t hdls[DS_MAX_HDLS];
	boolean_t is_client = info->libds_svcs[idx]->ds_type;
	uint_t nhdls = 0;

	/*
	 * If certain services/clients are already registered, manually
	 * call the registration handler; ideally this shouldn't happen.
	 */
	if ((libds_ops.ds_hdl_lookup(cap->svc_id, is_client, hdls, DS_MAX_HDLS,
	    &nhdls) == 0) && (nhdls > 0)) {
		int j;
		for (j = 0; j < nhdls; j++) {
			ds_domain_hdl_t dhdl;
			ds_ver_t *ver = &cap->vers[0];

			if (libds_ops.ds_domain_lookup(hdls[j], &dhdl) == 0)
				dslib_register_handler(hdls[j], cap->svc_id,
				    &ver[0], dhdl);
		}
		return;
	}

	/*
	 * For now, only register the clients, the fma
	 * services are on the fmactl channel, not on DS
	 */
	if (info->libds_svcs[idx]->ds_type == DS_CLIENT) {
		libds_ops.ds_unreg_svc(cap->svc_id, 1);
		libds_ops.ds_clnt_reg(cap, ops);
	} else if (check_vldc_chan_pres(FMA_CHAN_NAME) == false) {
		/*
		 * Fma service, but not on FMA channel
		 */
		libds_ops.ds_unreg_svc(cap->svc_id, 0);
		libds_ops.ds_svc_reg(cap, ops);
	}
}

void
initialize_libds(void)
{
	/*
	 * If the config in MD indicates no vlds, return right away
	 */
	if (libds_cap.config == false) {
		libds_cap.functional = false;
	} else if (libds_ops.ds_init() == 0) {
		/*
		 * The config to support libds may be present in MD, attempt
		 * to initialize the library.
		 */
		DBG(zprintf("%s: ds_init() returns success, libds may be used "
		    "for DS communication\n", __func__));
		libds_cap.functional = true;
	} else {
		/*
		 * The current MD config may have a vlds node, but ds_init()
		 * would still fail if the boot-config didn't have a vlds node
		 * to begin with in control domain, and in this case we revert
		 * to vldc based communication
		 */
		DBG(zprintf("%s: ds_init() returns failure, libds will not be "
		    "used for DS communication\n", __func__));
		libds_cap.functional = false;
	}
}

/*
 * Register the libds plugins at the end
 */
void
register_libds_plugins(void)
{
	int i;
	struct libds_plugins *pl, *npl;

	CHAIN_ITER_VOL(libds_list, pl, npl, link) {
		svc_info_t *info = pl->info;
		for (i = 0; i < info->num_svcs; i++)
			register_libds_plugin_svc(info, i);
		CHAIN_UNLINK(libds_list, pl, link);
		free_svc_info(info);
	}
}

/*
 * For each libds service registered with zeus, call ds_unreg_svc(). Call
 * ds_fini() once done
 */
void
unregister_libds_services(void)
{
	svc_t *svc;
	boolean_t is_client;

	CHAIN_ITER(all_services, svc, link) {
		is_client = (svc->props & SVC_OUTGOING) ? 0 : 1;
		if (strcmp(svc->chan_name, LIBDS_CHAN_NAME) == 0) {
			libds_ops.ds_unreg_svc(svc->name, is_client);
		}
	}

	libds_ops.ds_fini();
}

svc_t *
find_plugged_svc_on_chan_by_name(char *chan_name, char *svc_name)
{
	channel_plugins_t *plugins;
	svc_t *svc;

	CHAIN_DOUBLE_ITER(channel_plugins, plugins, link, plugged_svcs,
	    svc, link) {
		if (strcmp(plugins->chan_name, chan_name) == 0 &&
		    strcmp(svc->name, svc_name) == 0)
			return (svc);
	}

	return (NULL);
}
