/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "basics.h"
#include "support.h"
#include "zprintf.h"
#include "fatal.h"
#include "chain.h"
#include "resource_id.h"
#include "allocate.h"
#include "database.h"
#include "database_internal.h"
#include "xml_common.h"
#include "mac_addrs.h"
#include "mac_multicast.h"
#include "vsw.h"

#define	XML_VERSION	((xmlChar *)"1.0") /* version of XML used */

#define	MAC_ADDR_FILE		"mac_allocation.xml"

#define	MAC_ALLOC_VERSION	((xmlChar *)"1.0")

static prot_version_t mac_alloc_versions[] = {
	{ 1, 0 }
};

static xmlDocPtr used_mac_doc = NULL;
static char *xml_mac_filename = NULL;

static void populate_mac_list(void);
static void populate_used_mac_list(void);
static int32_t urandom_num(void);
static int32_t rand48_num(void);
static void print_used_list(void);

static void
create_mac_alloc_xml()
{
	xmlNodePtr	root;

	used_mac_doc = xmlNewDoc(XML_VERSION);
	root =  xmlNewDocNode(used_mac_doc, NULL, USED_MAC_LIST_NODE, NULL);
	xmlNewProp(root, VERSION_ATTR, MAC_ALLOC_VERSION);
	xmlDocSetRootElement(used_mac_doc, root);

	if (xmlSaveFormatFile(xml_mac_filename, used_mac_doc, true) < 0) {
		warning("Unable to preserve used mac address database %s",
		    xml_mac_filename);
	}
}

/*
 * Initializes the private MAC address pool, the MAC address list, and then
 * populates this list with and VSWs or VNETs already created on domains.
 */
void
mac_init(void)
{
	hrtime_t	t;
	xmlNodePtr	root;
	int		fd;
	xmlChar		*version;
	uint16_t 	major, minor, sup_major, sup_minor;

	create_res_pool(&private_mac_pool, PRIVATE_MAC_START, PRIVATE_MAC_END,
	    Res_none);
	name_res_pool(&private_mac_pool, "private macs");

	CHAIN_INIT(mac_list.vsw_list);
	CHAIN_INIT(used_list.mac_list);

	populate_mac_list();

	t = gethrtime();
	srand48(t);

	xml_mac_filename = zsprintf("%s/%s", zeus.admdir, MAC_ADDR_FILE);

	fd = open(xml_mac_filename, O_RDWR);
	if (fd == -1) {
		if (errno != ENOENT) {
			DBG(zprintf("Open failed. errno = %d Creating MAC XML "
			    "file\n", errno));
		}
	} else {
		close(fd);
		used_mac_doc = xmlParseFile(xml_mac_filename);
		if (used_mac_doc != NULL) {
			root = xmlDocGetRootElement(used_mac_doc);
			version = xmlGetProp(root, VERSION_ATTR);
			if (version == NULL) {
				DBG(zprintf("No version, rewriting MAC XML\n"));
			} else if (sscanf((char *)version, "%hu.%hu", &major,
			    &minor) != 2) {
				xmlFree(version);
				DBG(zprintf("sscanf version fails re-writing "
				    "MAC XML\n"));
			} else if (!negotiate_version(
			    NUM_VERS(mac_alloc_versions), mac_alloc_versions,
			    major, &sup_major, &sup_minor)) {
				xmlFree(version);
				DBG(zprintf("version # wrong re-writing MAC "
				    "XML\n"));
			} else {
				populate_used_mac_list();
				xmlFree(version);
				goto got_file;
			}

			xmlFreeDoc(used_mac_doc);
		} else {
			DBG(zprintf("Stat succeded but xmlParseFile failed\n"));
		}
	}

	create_mac_alloc_xml();

got_file:;

	DBG3(print_mac_list());
}

static void
populate_used_mac_list()
{
	xmlNodePtr	root, mac_node;
	mac_addr_t	node_content;
	used_mac_t	*used_mac;

	root = xmlDocGetRootElement(used_mac_doc);
	mac_node = xml_find_subnode(root, MAC_ADDRESS_NODE);

	while (mac_node != NULL) {
		node_content = xml_get_uint64_content(mac_node);

		used_mac = Xmalloc(sizeof (used_mac_t));

		used_mac->mac_addr = node_content;

		CHAIN_ADD(used_list.mac_list, used_mac, link);

		mac_node = xml_get_next_ele_node(mac_node);
	}

	DBG3(print_used_list());
}


/*
 * Given a VSWs service name it returns you the vsw_mac_t pointer in the
 * MAC address list. If a VSW is not present with that name it returns NULL.
 */
vsw_mac_t *
get_vsw_node(char *service_name)
{
	vsw_mac_t	*vsw_node;

	CHAIN_ITER(mac_list.vsw_list, vsw_node, link) {
		if (strcmp(service_name, vsw_node->service_name) == 0)
			return (vsw_node);
	}

	return (NULL);
}

/*
 * Given a VNETs name and ldom it is attached to it returns you the vnet_mac_t
 * pointer in the  MAC address list. If a VNET is not present with that name
 * and ldom combination it returns NULL.
 */
vnet_mac_t *
get_vnet_node(char *ldm_name, char *vnet_name)
{
	vsw_mac_t	*vsw_node;
	vnet_mac_t	*vnet_node;

	CHAIN_DOUBLE_ITER(mac_list.vsw_list, vsw_node, link, vnet_list,
	    vnet_node, link) {
		if (strcmp(vnet_node->vnet_name, vnet_name) == 0) {
			/*
			 * Make sure that this is the right vnet in the right
			 * ldom.
			 */
			if (strcmp(vnet_node->ldm_name, ldm_name) != 0)
				continue;
			return (vnet_node);
		}
	}

	return (NULL);
}

/*
 * Used to check to see if a VSW node has VNETs with allocated private MAC
 * addresses.  If there is a VNET referencing the VSW which has an allocated
 * private MAC address it returns the pointer to the vnet_mac_t. If not then
 * it return a NULL.
 */
vnet_mac_t *
private_macs_allocated(vsw_mac_t *vsw_node)
{
	vnet_mac_t	*vnet_node;
	mac_addr_t	mac_address;

	CHAIN_ITER(vsw_node->vnet_list, vnet_node, link) {
		mac_address = vnet_node->mac_addr;
		if ((mac_address >= PRIVATE_MAC_START) &&
		    (mac_address <= PRIVATE_MAC_END)) {
			return (vnet_node);
		}
	}
	return (NULL);
}

/*
 * Checks to see if a given MAC address is in the range of the allocated MAC
 * addresses, both private and public.  If it is withing the range we return
 * true and if not then we return false.
 */
bool_t
is_mac_allocated(mac_addr_t mac_address)
{

	if ((mac_address >= PRIVATE_MAC_START) &&
	    (mac_address <= PRIVATE_MAC_END)) {
		return (true);
	}

	if ((mac_address >= PUBLIC_MAC_START) &&
	    (mac_address <= PUBLIC_MAC_END)) {
		return (true);
	}

	return (false);
}

static int32_t
urandom_num()
{
	int		fd;
	int		rv;
	uint32_t	number;

	fd = open("/dev/urandom", O_RDONLY);
	if (fd == -1)
		return (-1);

	rv = read(fd, &number, sizeof (number));
	if (rv == -1) {
		close(fd);
		return (-1);
	}

	close(fd);

	return (number);
}

static int32_t
rand48_num()
{
	long		number;

	number = lrand48();

	return (number);
}

/*
 * This routine returns a usable public MAC address.
 */
uint64_t
get_usable_mac()
{
	uint32_t	number, random_number;
	mac_addr_t	mac = ILLEGAL_MAC_ADDR;
	bool_t		mac_found = false;
	time_t		start, now;
	int		diff = 0;
	used_mac_t	*used_mac;

	time(&start);

	while ((mac_found != true) && (diff < OBTAIN_MAC_TIMEOUT)) {
		/*
		 * First get the last removed MAC address on our "used list"
		 * If that address has been taken then move on to getting the
		 * MAC address randomly.
		 */
		if (CHAIN_LENGTH(used_list.mac_list) != 0) {
			used_mac = CHAIN_END(used_list.mac_list);
			mac = used_mac->mac_addr;
			free_used_mac(used_mac);
		} else {
			/*
			 * Try to get a random number using /dev/urandom if
			 * that fails then goto the backup method lrand48()
			 */
			number = urandom_num();
			if (number == -1)
				number = rand48_num();
			random_number = number & NO_OF_MACS;
			mac = PUBLIC_MAC_START + random_number;
		}

		mac_found = is_macaddr_free(mac, PERFORM_LOCAL_MAC_CHECK);
		time(&now);
		diff = now - start;
	}

	return (mac_found ? mac : ILLEGAL_MAC_ADDR);
}

/*
 * Helper routine to remove VNETs from a VSW node.
 */
static void
remove_ldoms_vnets(vsw_mac_t *vsw_node, char *ldm_name)
{
	vnet_mac_t	*vnet_node, *next_vnet_node;

	if (CHAIN_LENGTH(vsw_node->vnet_list) != 0) {
		CHAIN_ITER_VOL(vsw_node->vnet_list, vnet_node,
		    next_vnet_node, link) {
			if (strcmp(vnet_node->ldm_name, ldm_name) == 0) {
				delete_vnet(vsw_node, vnet_node);
			}
		}
	}
}

/*
 * This routine removes all VNETs and VSWs from a given ldom.
 */
void
remove_ldom_macs(char *ldm_name)
{
	vsw_mac_t	*vsw_node, *next_vsw_node;

	CHAIN_ITER_VOL(mac_list.vsw_list, vsw_node, next_vsw_node, link) {
		remove_ldoms_vnets(vsw_node, ldm_name);

		if ((vsw_node->ldm_name == NULL) ||
		    strcmp(vsw_node->ldm_name, ldm_name) == 0) {
			if (CHAIN_LENGTH(vsw_node->vnet_list) == 0) {
				delete_vsw(vsw_node);
			} else {
				blank_out_vsw(vsw_node);
			}
		}
	}

	DBG3(print_mac_list());
}

/*
 * This routine is called on LDMD startup.  It populates out mac_list first
 * with VSW nodes and then with VNETs nodes based on what is in the database
 * file.
 */
static void
populate_mac_list(void)
{
	xmlDocPtr	dbase;
	xmlNodePtr	root, ldom, node;
	vsw_mac_t	*vsw_node;
	mac_addr_t	mac_addr;
	char		*ldm_name;

	dbase = get_dbase();

	root = xmlDocGetRootElement(dbase);
	if (root == NULL) {
		zprintf("Root element in dbase NULL\n");
		return;
	}

	ldom = xml_find_subnode(root, LDOM_NODE);
	while (ldom != NULL) {
		/* Get the ldom name */
		node = xml_find_subnode(ldom, LDOM_NAME_NODE);
		ldm_name = (char *)xmlNodeGetContent(node);

		/*
		 * Need to add in a vsw node for the MAC address of
		 * the domain
		 */
		node = xml_find_subnode(ldom, MAC_ADDRESS_NODE);
		mac_addr = xml_get_uint64_content(node);

		create_vsw_mac(&vsw_node, ldm_name, ldm_name, VSW_ROUTED,
		    mac_addr);
		xmlFree(ldm_name);
		ldom = xml_get_next_ele_node(ldom);
	}

	/* reset ldom node */
	ldom = xml_find_subnode(root, LDOM_NODE);

	/* Search for and add all vsw nodes */
	populate_vsw(ldom);

	/*
	 * After searching for vsw in populate_vsw() the xmlNodePtr for ldom
	 * is now no longer at the first ldom in the database. Reset that
	 * pointer here.
	 */
	ldom = xml_find_subnode(root, LDOM_NODE);

	/* Search for and add all vnet nodes */
	populate_vnet(ldom);
}

static void
print_used_list(void)
{
	used_mac_t	*used_mac;

	CHAIN_ITER(used_list.mac_list, used_mac, link) {
		zprintf("Used MAC address: 0x%"PRIx64"\n", used_mac->mac_addr);
	}
}

#if ZDBG
/*
 * This is a DEBUG routine which prints out the VSW and VNET node in the
 * mac_list.
 */
void
print_mac_list(void)
{
	vsw_mac_t	*vsw_node;
	vnet_mac_t	*vnet_node;

	zprintf_lock();

	CHAIN_ITER(mac_list.vsw_list, vsw_node, link) {
		zprintf("\n");
		zprintf("VSW NODE\n");
		zprintf("name = %s\n", vsw_node->service_name);

		switch (vsw_node->mode) {
		case ROUTED_MODE:
			zprintf("vsw mode = ROUTED_MODE\n");
			break;
		case NON_ROUTED_MODE:
			zprintf("vsw mode = NON_ROUTED_MODE\n");
			break;
		}

		zprintf("mac address = 0x%"PRIx64"\n", vsw_node->mac_addr);

		if (vsw_node->ldm_name != NULL)
			zprintf("ldm name = %s\n", vsw_node->ldm_name);
		else
			zprintf("ldm name = NULL\n");

		CHAIN_ITER(vsw_node->vnet_list, vnet_node, link) {
			zprintf("\tVNET NODE\n");
			zprintf("\tname = %s\n", vnet_node->vnet_name);

			switch (vnet_node->mode) {
			case ROUTED_MODE:
				zprintf("\tvnet mode = ROUTED_MODE\n");
				break;
			case NON_ROUTED_MODE:
				zprintf("\tvnet mode = NON_ROUTED_MODE\n");
				break;
			}

			zprintf("\tmac address = 0x%"PRIx64"\n",
			    vnet_node->mac_addr);
			if (vnet_node->macaddr_res == NULL)
				zprintf("\tmac address res == NULL\n");
			else
				zprintf("\tmac address res == ALLOCATED\n");

			zprintf("\tldm name = %s\n", vnet_node->ldm_name);
			zprintf("\tservice name = %s\n",
			    vnet_node->service_name);
		}
		zprintf("\n");
	}

	zprintf_unlock();
}
#endif	/* ZDBG */

xmlNodePtr
find_mac_xml_node(mac_addr_t mac_addr, xmlNodePtr root)
{
	xmlNodePtr	mac_node;
	mac_addr_t	node_content;

	mac_node = xml_find_subnode(root, MAC_ADDRESS_NODE);
	if (mac_node == NULL)
		return (NULL);

	node_content = xml_get_uint64_content(mac_node);

	while (node_content != mac_addr) {
		mac_node = xml_get_next_ele_node(mac_node);
		if (mac_node == NULL)
			break;
		node_content = xml_get_uint64_content(mac_node);
	}

	return (mac_node);
}

/*
 * This routine adds a public MAC address onto our "used list" so we can
 * reuse these addresses.  This will help DHCP servers not run out of IP
 * addresses.
 */
void
add_used_mac(mac_addr_t mac_addr)
{
	used_mac_t	*used_mac;
	xmlNodePtr	root;
	xmlChar		mac[20];

	/* If the address is not allocated do not add it to the list */
	if (is_mac_allocated(mac_addr) == false)
		return;

	/* If the address is not a public address do not add it to the list */
	if ((mac_addr >= PRIVATE_MAC_START) &&
	    (mac_addr <= PRIVATE_MAC_END)) {
		return;
	}

	used_mac = Xmalloc(sizeof (used_mac_t));

	used_mac->mac_addr = mac_addr;

	CHAIN_ADD(used_list.mac_list, used_mac, link);

	root = xmlDocGetRootElement(used_mac_doc);
	snprintf((char *)mac, sizeof (mac), "0x%"PRIx64"", mac_addr);
	xmlNewChild(root, NULL, MAC_ADDRESS_NODE, mac);

	if (xmlSaveFormatFile(xml_mac_filename, used_mac_doc, true) < 0) {
		warning("Unable to preserve used mac address database %s",
		    xml_mac_filename);
	}
}

/*
 * This routine removes a public MAC address from our "used list". This will
 * be done if 1)we are reusing the MAC address on the same system or
 * 2)another ldmd called in with a MAC address on our list.
 */
void
free_used_mac(used_mac_t *used_mac)
{
	xmlNodePtr	root, mac_node;

	root = xmlDocGetRootElement(used_mac_doc);
	mac_node = find_mac_xml_node(used_mac->mac_addr, root);
	if (mac_node != NULL) {
		xmlUnlinkNode(mac_node);
		xmlFreeNode(mac_node);

		if (xmlSaveFormatFile(xml_mac_filename, used_mac_doc, true) < 0)
			warning("Unable to preserve used mac address "
			    "database %s", xml_mac_filename);
	}

	CHAIN_UNLINK(used_list.mac_list, used_mac, link);

	Xfree(used_mac);
}

/*
 * This routine checks the MAC address "used list" to see if a given MAC
 * address is in our list. If it is it return the pointer to that use_mac_t
 * object. If not it returns NULL.
 */
used_mac_t *
is_mac_used(mac_addr_t mac_addr)
{
	used_mac_t 	*used_mac;

	CHAIN_ITER(used_list.mac_list, used_mac, link) {
		if (used_mac->mac_addr == mac_addr)
			return (used_mac);
	}

	return (NULL);
}

/*
 * Function to convert a raw MAC address into ASCII xx:xx:xx .. form
 * Requires a target buffer of at least 18 bytes size.
 */
void
mactoa(mac_addr_t mac_addr, char *bufp)
{
	uint64_t mac = mac_addr;
	if (mac_addr == ILLEGAL_MAC_ADDR ||
	    (((uint64_t)mac_addr) & ~0xffffffffffffULL) != 0ULL) {
		strcpy(bufp, "Illegal MAC");
		return;
	}

	sprintf(bufp, "%02x:%02x:%02x:%02x:%02x:%02x",
	    (int)((mac >> 40) & 0xff),
	    (int)((mac >> 32) & 0xff),
	    (int)((mac >> 24) & 0xff),
	    (int)((mac >> 16) & 0xff),
	    (int)((mac >> 8) & 0xff),
	    (int)((mac >> 0) & 0xff));
}

/*
 * Simple conversion of 1 or 2 characters representing
 * a hex value into an int.
 */
static bool_t
atobyte(char *strp, uint8_t *valp)
{
	uint8_t val;

	if (strp[0] >= '0' && strp[0] <= '9') {
		val = strp[0]-'0';
	} else
	if (strp[0] >= 'a' && strp[0] <= 'f') {
		val = strp[0]-'a'+10;
	} else
	if (strp[0] >= 'A' && strp[0] <= 'F') {
		val = strp[0]-'A'+10;
	} else
		return (false);

	if (strp[1] != '\0') {
		val <<= 4;
		if (strp[1] >= '0' && strp[1] <= '9') {
			val |= strp[1]-'0';
		} else
		if (strp[1] >= 'a' && strp[1] <= 'f') {
			val |= strp[1]-'a'+10;
		} else
		if (strp[1] >= 'A' && strp[1] <= 'F') {
			val |= strp[1]-'A'+10;
		} else
			return (false);

		if (strp[2] != '\0')
			return (false);
	}

	*valp = val;
	return (true);
}

/*
 * Would be nice to be able to use strtok, except that it is broken
 * and skips over :: as a valid (empty) value. So we do it the hard way with
 * strchr instead.
 */
bool_t
atomac(char *bufp, mac_addr_t *mac_addressp)
{
	uint8_t	b[6];
	char	*strp, *nextp;
	uint64_t	mac;
	int	i;

	if (bufp == NULL)
		return (false);

	strp = bufp;
	nextp = strchr(strp, ':');
	if (nextp) *nextp++ = '\0';

	for (i = 0; i < 6 && strp; i++) {
		if (!atobyte(strp, &b[i]))
			return (false);
		strp = nextp;
		if (strp) {
			nextp = strchr(strp, ':');
			if (nextp) *nextp++ = '\0';
		}
	}

	/* Too few or too many ':' chars? */
	if (i < 6 || nextp != NULL)
		return (false);

	mac = b[0];
	for (i = 1; i < 6; i++) {
		mac <<= 8;
		mac |= b[i];
	}

	*mac_addressp = (mac_addr_t)mac;
	return (true);
}
