/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <thread.h>
#include <synch.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "basics.h"
#include "allocate.h"
#include "zprintf.h"
#include "option.h"
#include "mac_addrs.h"
#include "mac_multicast.h"
#include "ldom.h"
#include "response.h"
#include "errors.h"

/* Socket Descriptors */
static int mac_sock_in, mac_sock_out;

/* Sockets */
static struct sockaddr_in	sock_r, sock_w;

/* Mutex and condition variable */
static mutex_t macaddr_mutex;
static cond_t macaddr_cond;

/* mac_lookup response packet */
static uint64_t	current_req;

bool_t
start_mac_auto_allocate(void)
{
	/* check if there is any work to do */
	if (mac_socket_up)
		return (true);

	mac_socket_up = macaddr_sock_create();

	if (mac_socket_up) {
		if (thr_create(NULL, NULL, macaddr_loop, NULL, THR_DAEMON,
		    NULL) != 0) {
			DBG(zprintf("Could not create macaddr_loop thread"));
			mac_socket_up = false;
		}
	}

	return (mac_socket_up);
}

/*
 * This routine creates our read IN socket and our write OUT socket. It also
 * initializes our mutex and condition variable.
 */
bool_t
macaddr_sock_create(void)
{
	struct ip_mreq		mreq;
	uint8_t			ttl;
	uint32_t		on = 1;
	int			err;

	/* Initialize the read socket */
	memset(&sock_r, 0, sizeof (sock_r));
	sock_r.sin_family = AF_INET;
	sock_r.sin_port = MAC_MULTI_PORT;
	sock_r.sin_addr.s_addr = htonl(INADDR_ANY);

	/* Create incoming socket */
	if ((mac_sock_in = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
		DBG(zprintf("mac_sock_in socket create fail\n"));
		return (false);
	}

	/* Set re-user addr. */

	err = setsockopt(mac_sock_in, SOL_SOCKET, SO_REUSEADDR,
	    &on, sizeof (on));
	if (err < 0) {
		DBG(zprintf("mac_sock_in REUSEADDR failed: errno = %s\n",
		    strerror(errno)));
	}

	/* Bind the read socket */
	if (bind(mac_sock_in, (struct sockaddr *)&sock_r,
	    sizeof (sock_r)) < 0) {
		DBG(zprintf("mac_sock_in socket bind failed\n"));
		close(mac_sock_in);
		return (false);
	}

	/* Initialize the ip_mreq  for reception */
	mreq.imr_multiaddr.s_addr = inet_addr(MAC_MULTI_GROUP);
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);

	/* Set read socket to be a member of the multicast group */
	if (setsockopt(mac_sock_in, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq,
	    sizeof (mreq)) < 0) {
		DBG(zprintf("mac_sock_in socket setsockopt MEMBERSHIP "
		    "failed\n"));
		close(mac_sock_in);
		return (false);
	}

	/* Initialize the write socket */
	memset(&sock_w, 0, sizeof (sock_w));
	sock_w.sin_family = AF_INET;
	sock_w.sin_port = MAC_MULTI_PORT;
	sock_w.sin_addr.s_addr = inet_addr(MAC_MULTI_GROUP);

	/* Create outgoing socket */
	if ((mac_sock_out = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
		DBG(zprintf("mac_sock_out socket create fail\n"));
		close(mac_sock_in);
		return (false);
	}

	if (setsockopt(mac_sock_out, SOL_SOCKET, SO_REUSEADDR,
	    &on, sizeof (on)) < 0) {
		DBG(zprintf("mac_sock_in REUSEADDR failed\n"));
	}

	if (option.mac_ttl > 255)
		ttl = 255;
	else
		ttl = (uint8_t)option.mac_ttl;

	DBG(zprintf("Setting MAC socket TTL = %d\n", ttl));

	/* Set write socket for correct time to live */
	if (setsockopt(mac_sock_out, IPPROTO_IP, IP_MULTICAST_TTL, &ttl,
	    sizeof (ttl)) < 0) {
		DBG(zprintf("mac_sock_out socket setsockopt TTL failed\n"));
		close(mac_sock_in);
		close(mac_sock_out);
		return (false);
	}

	/* Initialize our Mutex and Condition variable */
	mutex_init(&macaddr_mutex, USYNC_THREAD, NULL);
	cond_init(&macaddr_cond, USYNC_THREAD, NULL);

	return (true);
}

/*
 * This routine checks our mac_list to see if a given MAC address is already
 * allocated on this system (true) or not (false).
 */
static bool_t
is_macaddr_allocated_on_system(uint64_t mac_addr)
{
	vsw_mac_t	*vsw_node;
	vnet_mac_t	*vnet_node;

	CHAIN_ITER(mac_list.vsw_list, vsw_node, link) {
		if (vsw_node->mac_addr == mac_addr)
			return (true);
		CHAIN_ITER(vsw_node->vnet_list, vnet_node, link) {
			if (vnet_node->mac_addr == mac_addr)
				return (true);
		}
	}

	return (false);
}

/*
 * Handle a send message. Check if the MAC address is already allocated on
 * this system, and if so, send a response message.
 */
static void
handle_mac_send_msg(multicast_msg_t *multicast_msg)
{
	uint32_t	msg_type = multicast_msg->msg_type;
	mac_lookup_t	mac_lookup = multicast_msg->lookup;
	uint64_t	mac_addr = ntoh64(mac_lookup.mac_addr);
	used_mac_t	*used_mac;
	multicast_msg_t	resp_msg;
	char		*local_ip;
	bool_t		local_mac_collision_check;

	ASSERT(msg_type == SEND_MSG);

	/* lock the list while we are checking it */
	(void) mutex_lock(&macaddr_mutex);

	DBG3(zprintf("%s: incoming MAC 0x%llx\n", __func__, mac_addr));

	/*
	 * First check to see if the MAC address is on our "used
	 * list". If so, we pull it off and return because it is
	 * obviously not being used on this system.
	 */
	if ((used_mac = is_mac_used(mac_addr)) != NULL) {
		DBG(zprintf("MAC address was used, removing from our list\n"));
		free_used_mac(used_mac);
		goto done;
	}

	/*
	 * Check the local MAC address collision at domain create time
	 * and skip at domain start time.
	 */
	local_mac_collision_check = mac_lookup.local_mac_collision_check;
	if (LOCAL_MAC_ALLOCATED(mac_addr) == false)
		goto done;

	/*
	 * The MAC address is in use on this system. Generate a
	 * response message and send it out.
	 */
	DBG(zprintf("MAC address found in list, sending response\n"));

	resp_msg.magic_no = hton32(MAC_MULTI_MAGIC_NO);
	resp_msg.version_no = hton32(MAC_MULTI_VERSION);
	resp_msg.msg_type = hton32(RESPONSE_MSG);
	resp_msg.lookup.mac_addr = hton64(mac_addr);

	/* kick off the MAC collision thread */
	if (!mac_socket_up)
		start_mac_auto_allocate();

	/*
	 * The body of the message includes an IP address used when
	 * a migration is in progress. If this machine is the source
	 * of a migration, the IP will be for the interface connected
	 * to the peer ldmd. Otherwise, it will be all zeros.
	 */
	local_ip = (zeus.local_ip != NULL) ? zeus.local_ip : "0.0.0.0";
	strcpy(resp_msg.lookup.source_ip, local_ip);

	(void) sendto(mac_sock_out, &resp_msg, sizeof (resp_msg), 0,
	    (struct sockaddr *)&sock_w, sizeof (sock_w));

done:
	(void) mutex_unlock(&macaddr_mutex);
}

/*
 * Handle a response message. Check if the MAC address in the incoming
 * message is the one that is being checked. If so, wake up the thread
 * blocked waiting for a response. Otherwise, there is no match and the
 * message can be dropped. It is still possible that other hosts will
 * detect a MAC collision and send their own response.
 */
static void
handle_mac_response_msg(multicast_msg_t *multicast_msg)
{
	uint32_t	msg_type = multicast_msg->msg_type;
	mac_lookup_t	mac_lookup = multicast_msg->lookup;
	uint64_t	mac_addr = ntoh64(mac_lookup.mac_addr);

	ASSERT(msg_type == RESPONSE_MSG);

	/* lock the list while we are checking it */
	(void) mutex_lock(&macaddr_mutex);

	DBG3(zprintf("%s: incoming MAC 0x%llx\n", __func__, mac_addr));

	/*
	 * If the MAC address in the message is not the MAC
	 * being actively checked, there is no collision so
	 * there is nothing to do.
	 */
	if (mac_addr != current_req)
		goto done;

	/*
	 * There is a collision, but if there is a migration
	 * in progress, more checking is required. If this
	 * machine is the target, the source IP address will
	 * have been saved when the migration started. If this
	 * matches the IP address in the incoming message, the
	 * collision is with the source domain, so it can be
	 * ignored.
	 */
	if (zeus.remote_ip != NULL) {
		if (strcmp(zeus.remote_ip, mac_lookup.source_ip) == 0) {
			DBG(zprintf("Duplicate from a migrating domain\n"));
			goto done;
		}
	}

	/* MAC address really is in use */
	DBG(zprintf("%s: MAC address already allocated, retry\n", __func__));
	cond_signal(&macaddr_cond);

done:
	(void) mutex_unlock(&macaddr_mutex);
}

/*
 * This routine is started as a new thread. It simply listens to the read IN
 * socket, checks to make sure the message in is for a LDMD and then handles
 * the message.
 */
/* ARGSUSED0 */
void *
macaddr_loop(void *unused)
{
	multicast_msg_t		*multicast_msg;
	int			msg_size = sizeof (multicast_msg_t);
	int			count;
	int			sock_r_len = sizeof (sock_r);
	uint32_t		magic_no, version_no, msg_type;

	multicast_msg = Xmalloc(msg_size);
	while (1) {
		count = recvfrom(mac_sock_in, multicast_msg, msg_size, 0,
		    (struct sockaddr *)&sock_r, &sock_r_len);

		if (count < sizeof (mac_lookup_t)) {
			continue;
		}

		magic_no = ntoh32(multicast_msg->magic_no);
		version_no = ntoh32(multicast_msg->version_no);
		msg_type = ntoh32(multicast_msg->msg_type);

		/*
		 * Check to see if this is a message for us. The magic number
		 * should check out for us to proceed.
		 */
		if (magic_no != MAC_MULTI_MAGIC_NO) {
			DBG(zprintf("Magic Number did not match(%d)\n",
			    magic_no));
			continue;
		}
		/*
		 * Check to make sure the incoming packet can be handled by
		 * LDMD.
		 */
		if (version_no > MAC_MULTI_VERSION) {
			DBG(zprintf("MAC allocation version number is not"
			    "correct (%d)\n", version_no));
			continue;
		}

		switch (msg_type) {

		case SEND_MSG:
			handle_mac_send_msg(multicast_msg);
			break;

		case RESPONSE_MSG:
			handle_mac_response_msg(multicast_msg);
			break;

		case LDMD_DISC_SEND:
			handle_ldmd_discovery(&multicast_msg->discovery,
			    mac_sock_out, sock_w);
			break;
		default:
			/* catch unsupported message types */
			DBG(fatal("Unsupported message type: %d", msg_type));
			break;
		}
	}

	/*NOTREACHED*/
}

/*
 * This routine checks to see if a given MAC address is already allocated
 * either on this system OR on a different system. current_req is set here
 * and checked in handle_mac_msg() above.  This routine should not be called
 * multiple times because of the sequencer. If this changes we need to make
 * sure current_req is safe guarded.
 *
 * If the boolean flag 'local_mac_collision_check' is true, it will
 * perform the local MAC collision check (at domain create time).
 * Otherwise, it will skip the local MAC collision check (at domain start
 * time).
 */
bool_t
is_macaddr_free(uint64_t mac_addr, bool_t local_mac_collision_check)
{
	multicast_msg_t	mac_lookup;
	timestruc_t	timeout;
	int		err, count;
	int		sock_w_len = sizeof (sock_w);

	(void) mutex_lock(&macaddr_mutex);

	/*
	 * First check our own system.
	 *
	 * Check the local MAC address collision at domain create time
	 * and skip at domain start time.
	 */
	if (LOCAL_MAC_ALLOCATED(mac_addr)) {
		mutex_unlock(&macaddr_mutex);
		return (false);
	}

	/* Now check outside of this system */
	mac_lookup.magic_no = hton32(MAC_MULTI_MAGIC_NO);
	mac_lookup.version_no = hton32(MAC_MULTI_VERSION);
	mac_lookup.msg_type = hton32(SEND_MSG);
	mac_lookup.lookup.mac_addr = hton64(mac_addr);
	mac_lookup.lookup.local_mac_collision_check =
	    local_mac_collision_check;
	current_req = mac_addr;

	DBG(zprintf("%s: Testing MAC 0x%"PRIx64" to see if allocated\n",
	    __func__, mac_addr));

	count = sendto(mac_sock_out, &mac_lookup, sizeof (mac_lookup), 0,
	    (struct sockaddr *)&sock_w, sock_w_len);
	if (count != sizeof (mac_lookup)) {
		DBG(zprintf("%s: Count is wrong on send\n", __func__));
		mutex_unlock(&macaddr_mutex);
		return (false);
	} else {
		DBG3(zprintf("%s: sendto count == %d\n", __func__, count));
	}

	timeout.tv_sec = MAC_MULTI_TIMEOUT;
	timeout.tv_nsec = 0;

	err = cond_reltimedwait(&macaddr_cond, &macaddr_mutex, &timeout);
	/*
	 * If we timeout here then no other LDMDs are using the MAC
	 * address requested.
	 */
	if (err == ETIME) {
		current_req = 0;
		mutex_unlock(&macaddr_mutex);
		DBG(zprintf("%s: TIMEOUT: %"PRIx64" is NOT allocated\n",
		    __func__, mac_addr));
		return (true);
	}

	/*
	 * If we get here then we have recieved a message back from another
	 * LDMD saying that they have already allocated that MAC address.
	 */
	current_req = 0;
	mutex_unlock(&macaddr_mutex);
	DBG(zprintf("%s: %"PRIx64" IS allocated try again\n",
	    __func__, mac_addr));

	return (false);
}

/*
 * If a given MAC address is in the range of MAC auto allocation pool, this
 * routine checks to see if the MAC address is already allocated either on
 * this system OR on a different system. If a given MAC address is NOT in
 * the range of MAC auto allocation pool, this routine checks to see if the
 * MAC address is aready allocated on this system only.
 *
 * If the boolean flag 'local_mac_collision_check' is true, it will
 * perform the local MAC collision check (at domain create time).
 * Otherwise, it will skip the local MAC collision check (at domain start
 * time).
 */
bool_t
is_macaddr_in_use(mac_addr_t mac_addr, bool_t local_mac_collision_check,
    response_t *resp)
{
	used_mac_t	*used_mac;
	char		mac[MAC_STR_LEN];

	/*
	 * Is MAC addr in MAC auto allocation range?
	 */
	if (is_mac_allocated(mac_addr)) {

		/* make sure auto MAC allocation is available */
		if (!mac_socket_up) {
			/* try to set up the multicast socket again */
			if (!start_mac_auto_allocate()) {
				rprintf(resp,
				    MSG_ERR_MAC_SOCKET_CREATION_FAILED);
				if (local_mac_collision_check) {
					rprintf(resp,
					    MSG_ERR_MAC_SET_MANUALLY);
				} else {
					/* at domain start time */
					rprintf(resp,
					    MSG_ERR_MAC_REMOTE_COLLISION);
				}
				return (false);
			}
		}

		/*
		 * Is MAC addr free?
		 */
		if (!is_macaddr_free(mac_addr, local_mac_collision_check))
			goto mac_in_use;

		/*
		 * if a mac addr was used by a vnet/vsw and then the vnet/vsw
		 * was deleted, the mac addr would be saved in
		 * the used mac addr file. In that case, we need to
		 * remove the mac addr from used mac addr file.
		 */
		used_mac = is_mac_used(mac_addr);
		if (used_mac != NULL)
			free_used_mac(used_mac);
	} else {
		/*
		 * Is MAC addr already allocated in this system?
		 *
		 * Check the local MAC address collision at domain create time
		 * and skip at domain start time.
		 */
		if (LOCAL_MAC_ALLOCATED(mac_addr)) {
			goto mac_in_use;
		}
	}

	return (false);

mac_in_use:
	mactoa(mac_addr, mac);
	rprintf(resp, gettext("MAC address %s is already in use \n"), mac);

	return (true);
}
