/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2008 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)mac_vnet.c	1.10	08/09/24 SMI"

#include <string.h>

#include "basics.h"
#include "zprintf.h"
#include "allocate.h"
#include "database_internal.h"
#include "xml_common.h"
#include "mac_addrs.h"

/*
 * This routine is called from ldomconfigure.c: ldm_add_vnet.  It creates a
 * VNET node and then adds it into the mac_list under its VSW node. If there
 * is not already a VSW node created for the service_name given it creates a
 * basic VSW node which is a place holder node.
 */
int
create_vnet_node(vnet_mac_t **sent_vnet, vsw_mac_t **vsw_node, char *ldm_name,
    char *service_name, char *vnet_name, char *mode, mac_addr_t mac_addr)
{
	vnet_mac_t	*vnet_node;
	int		int_mode;

	if (mode != NULL) {
		if (strcmp(mode, "private") == 0)
			int_mode = ROUTED_MODE;
		else
			int_mode = NON_ROUTED_MODE;
	} else
		int_mode = NON_ROUTED_MODE;

	if (*vsw_node == NULL) {
		/* We need to create a place holder VSW */
		*vsw_node = create_basic_vsw_node(NULL, service_name,
		    ROUTED_MODE);
	}

	vnet_node = create_vnet(*vsw_node, ldm_name, vnet_name, int_mode);

	if (mac_addr == ILLEGAL_MAC_ADDR) {
		if (int_mode == ROUTED_MODE) {
			vnet_node->macaddr_res =
			    alloc_res(&private_mac_pool);
			vnet_node->mac_addr = vnet_node->macaddr_res->id;
		} else {
			vnet_node->macaddr_res = NULL;
			vnet_node->mac_addr = get_usable_mac();
			if (vnet_node->mac_addr == ILLEGAL_MAC_ADDR) {
				return (-1);
			}
		}
	} else {
		vnet_node->macaddr_res = NULL;
		vnet_node->mac_addr = mac_addr;
	}

	*sent_vnet = vnet_node;

	return (0);
}

/*
 * This routine is called from ldomconfigure.c: ldm_mod_vnet. It modifies an
 * already create VNET node.
 */
int
update_vnet_node(vnet_mac_t **sent_vnet, char *new_service, mac_addr_t mac_addr,
    char *mode)
{
	vsw_mac_t	*vsw_node;
	vnet_mac_t	*vnet_node = *sent_vnet;
	vnet_mac_t	*new_vnet;
	uint64_t	old_mac = ILLEGAL_MAC_ADDR;
	uint64_t	new_mac;
	int		old_mode;
	char		*ldm_name = NULL;
	char		*vnet_name = NULL;

	/*
	 * We can only change the mode OR the MAC addr never both.
	 * ldomconfigure.c makes sure of this.
	 */
	if (mode != NULL) {
		if (strcmp(mode, "private") == 0) {
			vnet_node->mode = ROUTED_MODE;
			vnet_node->macaddr_res =
			    alloc_res(&private_mac_pool);
			vnet_node->mac_addr = vnet_node->macaddr_res->id;
		} else {
			new_mac =  get_usable_mac();
			if (new_mac == ILLEGAL_MAC_ADDR) {
				return (-1);
			}
			vnet_node->mac_addr = new_mac;
			vnet_node->mode = NON_ROUTED_MODE;
			free_res(vnet_node->macaddr_res);
			vnet_node->macaddr_res = NULL;
		}
	}

	if (mac_addr != ILLEGAL_MAC_ADDR) {
		/*
		 * If the MAC was previously auto-allocated we need to add it
		 * to our used MAC list.
		 */
		if (is_mac_allocated(vnet_node->mac_addr)) {
			add_used_mac(vnet_node->mac_addr);
		}
		vnet_node->mac_addr = mac_addr;
	}

	/*
	 * If we are changing the service node for the VNET then we save most
	 * the information from current VNET node, remove that node, create a
	 * NEW node under the new service, and then fill in the saved
	 * information.
	 */
	if ((new_service != NULL) && (strcmp(vnet_node->service_name,
	    new_service) != 0)) {
		old_mac = vnet_node->mac_addr;
		old_mode = vnet_node->mode;

		ldm_name = Xmalloc(strlen(vnet_node->ldm_name) + 1);
		strcpy(ldm_name, vnet_node->ldm_name);

		vnet_name = Xmalloc(strlen(vnet_node->vnet_name) + 1);
		strcpy(vnet_name, vnet_node->vnet_name);

		remove_vnet_mac(ldm_name, vnet_name);

		vsw_node = get_vsw_node(new_service);
		(void) create_vnet_node(&new_vnet, &vsw_node, ldm_name,
		    new_service, vnet_name, "public", 0x1);

		new_vnet->mode = old_mode;
		new_vnet->mac_addr = old_mac;

		if ((old_mac >= PRIVATE_MAC_START) &&
		    (old_mac <= PRIVATE_MAC_END)) {
			new_vnet->macaddr_res = pull_res_id(&private_mac_pool,
			    old_mac);
		}
		*sent_vnet = new_vnet;
		Xfree(ldm_name);
		Xfree(vnet_name);
		return (0);
	}

	return (0);
}

void
remove_empty_vsw(vsw_mac_t *vsw_node)
{
	/*
	 * If the vnet list is empty and the vsw ldm_name ==
	 * NULL then that means that the vsw is a place holder node.
	 * We need to remove it from the mac_list.
	 */
	if ((CHAIN_LENGTH(vsw_node->vnet_list) == 0) &&
	    (vsw_node->ldm_name == NULL)) {
		CHAIN_UNLINK(mac_list.vsw_list, vsw_node, link);
		Xfree(vsw_node->service_name);
		Xfree(vsw_node);
	}
}

/*
 * This routine is called from ldomconfigure.c: ldm_remove_io. Given a ldom
 * name and VNET node name it removes a VNET node from the VSW it is attached
 * to. If that VSW no longer has VNETs  attached to it and is a place holder
 * this routine also removes that VSW node.
 */
void
remove_vnet_mac(char *ldm_name, char *vnet_name)
{
	vsw_mac_t	*vsw_node, *next_vsw_node;
	vnet_mac_t	*vnet_node, *next_vnet_node;

	CHAIN_ITER_VOL(mac_list.vsw_list, vsw_node, next_vsw_node, link) {
		CHAIN_ITER_VOL(vsw_node->vnet_list, vnet_node,
		    next_vnet_node, link) {
			if (strcmp(vnet_node->vnet_name, vnet_name) == 0) {
				/*
				 * Make sure that this is the right
				 * vnet in the right ldom.
				 */
				if (strcmp(vnet_node->ldm_name, ldm_name) != 0)
					continue;

				delete_vnet(vsw_node, vnet_node);
			}
		}
		remove_empty_vsw(vsw_node);
	}
}

/*
 * This routine is used to create a basic VNET node and add it to a VSWs
 * vnet_list.  The MAC address and the resource pointer are blanked out.
 */
vnet_mac_t *
create_vnet(vsw_mac_t *vsw_node, char *ldm_name, char *vnet_name, int mode)
{
	vnet_mac_t	*vnet_node;

	vnet_node = Xmalloc(sizeof (vnet_mac_t));

	vnet_node->ldm_name = Xmalloc(strlen(ldm_name) + 1);
	strcpy(vnet_node->ldm_name, ldm_name);

	vnet_node->service_name = Xmalloc(strlen(vsw_node->service_name) + 1);
	strcpy(vnet_node->service_name, vsw_node->service_name);

	vnet_node->vnet_name = Xmalloc(strlen(vnet_name) + 1);
	strcpy(vnet_node->vnet_name, vnet_name);

	vnet_node->mode = mode;

	vnet_node->macaddr_res = NULL;
	vnet_node->mac_addr = ILLEGAL_MAC_ADDR;

	CHAIN_ADD(vsw_node->vnet_list, vnet_node, link);

	return (vnet_node);
}

/*
 * This routine is the basic code to actually remove a VNET from a list and
 * clean up memory.
 */
void
delete_vnet(vsw_mac_t *vsw_node, vnet_mac_t *vnet_node)
{
	/*
	 * If the address is a private address it should go back into the
	 * resource pool. If it is a public address add it to our "used list"
	 */
	if (vnet_node->macaddr_res != NULL)
		free_res(vnet_node->macaddr_res);
	else
		add_used_mac(vnet_node->mac_addr);

	CHAIN_UNLINK(vsw_node->vnet_list, vnet_node, link);
	Xfree(vnet_node->ldm_name);
	Xfree(vnet_node->vnet_name);
	Xfree(vnet_node->service_name);
	Xfree(vnet_node);
}

/*
 * This routine parses the database and create VNET nodes (and place holder
 * VSWs if needed) for each VNET in the database.
 */
void
populate_vnet(xmlNodePtr ldom)
{
	xmlNodePtr	vnet_node, node;
	char		*ldm_name, *service_name, *vnet_name;
	vsw_mac_t	*vsw;
	vnet_mac_t	*vnet;
	mac_addr_t	mac_addr;

	/* Search for and add all vnets */
	while (ldom != NULL) {
		node = xml_find_subnode(ldom, LDOM_NAME_NODE);
		ldm_name = (char *)xmlNodeGetContent(node);

		node = xml_find_subnode(ldom, VNET_NODE);
		vnet_node = xml_find_subnode(node, VNET_INSTANCE_NODE);

		while (vnet_node != NULL) {
			node = xml_find_subnode(vnet_node, SERVICE_NAME_NODE);
			service_name = (char *)xmlNodeGetContent(node);

			vsw = get_vsw_node(service_name);
			if (vsw == NULL) {
				vsw = create_basic_vsw_node(NULL, service_name,
				    ROUTED_MODE);
			}
			xmlFree(service_name);

			node = xml_find_subnode(vnet_node, VNET_NAME_NODE);
			vnet_name = (char *)xmlNodeGetContent(node);

			vnet = create_vnet(vsw, ldm_name, vnet_name,
			    NON_ROUTED_MODE);
			xmlFree(vnet_name);

			node = xml_find_subnode(vnet_node, MAC_ADDRESS_NODE);
			mac_addr = xml_get_uint64_content(node);

			if ((mac_addr >= PRIVATE_MAC_START) &&
			    (mac_addr <= PRIVATE_MAC_END)) {
				vnet->macaddr_res =
				    pull_res_id(&private_mac_pool, mac_addr);
				vnet->mac_addr = vnet->macaddr_res->id;
				vnet->mode = ROUTED_MODE;
			} else {
				vnet->macaddr_res = NULL;
				vnet->mac_addr = mac_addr;
			}


			vnet_node = vnet_node->next;
		}

		xmlFree(ldm_name);
		ldom = xml_get_next_ele_node(ldom);
	}
}
