/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2008 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)mac_vsw.c	1.9	08/09/24 SMI"

#include <string.h>
#include <sys/types.h>

#include "basics.h"
#include "allocate.h"
#include "chain.h"
#include "database.h"
#include "database_internal.h"
#include "xml_common.h"
#include "mac_addrs.h"
#include "vsw.h"

static void create_vsw_inst(xmlNodePtr inst_node, char *ldm_name);

/*
 * This routine is called from ldomconfigure.c: ldm_add_vsw.  This creates a
 * basic VSW node, adds it to the mac_list, and then adds the MAC address to
 * the node.
 */
int
create_vsw_mac(vsw_mac_t **sent_vsw, char *ldm_name, char *service_name,
    char *mode, mac_addr_t mac_address)
{
	vsw_mac_t 	*vsw_node;
	int		int_mode;

	if (strcmp(mode, "routed") == 0)
		int_mode = ROUTED_MODE;
	else
		int_mode = NON_ROUTED_MODE;

	vsw_node = create_basic_vsw_node(ldm_name, service_name, int_mode);

	if (mac_address != ILLEGAL_MAC_ADDR) {
		vsw_node->mac_addr = mac_address;
	} else {
		vsw_node->mac_addr = get_usable_mac();
		if (vsw_node->mac_addr == ILLEGAL_MAC_ADDR)
			return (-1);
	}

	*sent_vsw = vsw_node;
	return (0);
}

/*
 * This routine modifies the VSWs mode or MAC address when called from
 * ldm_mod_vsw.  And fills in the ldm name and MAC address to a place holder
 * VSW node if called from ldm_add_vsw.
 */
int
update_vsw_mac(vsw_mac_t *vsw_node, char *ldm_name, mac_addr_t mac_address,
    char *mode)
{
	if (mode != NULL) {
		if (strcmp(mode, "routed") == 0)
			vsw_node->mode = ROUTED_MODE;
		else
			vsw_node->mode = NON_ROUTED_MODE;
	}

	/*
	 * If a call comes in with the ldm_name filled in then we are updating
	 * this from ldm_add_vsw because it was a place holder. We need to
	 * update the name AND the MAC address. Then simply return.
	 */
	if (ldm_name != NULL) {
		vsw_node->ldm_name = Xstrdup(ldm_name);
		if (mac_address != ILLEGAL_MAC_ADDR) {
			vsw_node->mac_addr = mac_address;
		} else {
			vsw_node->mac_addr = get_usable_mac();
			if (vsw_node->mac_addr == ILLEGAL_MAC_ADDR) {
				return (-1);
			}
		}
		return (0);
	}

	/* Change the MAC address ONLY if they have supplied one. */
	if (mac_address != ILLEGAL_MAC_ADDR) {
		/*
		 * If the MAC was previously auto-allocated we need to add it
		 * to our used MAC list.
		 */
		if (is_mac_allocated(vsw_node->mac_addr)) {
			add_used_mac(vsw_node->mac_addr);
		}
		vsw_node->mac_addr = mac_address;
	}
	return (0);
}

/*
 * This routine, given the name of a VSW, serches the mac_list, finds the VSW
 * and then either deletes it completely if there are no VNETs referencing the
 * VSW or blanks it out (making it a place holder) if there are VNETs still
 * referencing the VSW.
 */
void
remove_vsw_mac(char *service_name)
{
	vsw_mac_t	*vsw_node, *next_vsw_node;

	CHAIN_ITER_VOL(mac_list.vsw_list, vsw_node, next_vsw_node, link) {
		if (strcmp(vsw_node->service_name, service_name) == 0) {
			/*
			 * If the vnet list is 0 then remove it from the
			 * mac_list.
			 */
			if (CHAIN_LENGTH(vsw_node->vnet_list) == 0) {
				delete_vsw(vsw_node);
			} else {
				blank_out_vsw(vsw_node);
			}
		}
	}
}

/*
 * This routine creates a basic VSW node which doesn't have it's MAC address
 * filled in OR its ldm_name if the VSW is a place holder node.
 */
vsw_mac_t *
create_basic_vsw_node(char *ldm_name, char *service_name, int mode)
{
	vsw_mac_t 	*vsw_node;

	vsw_node = Xmalloc(sizeof (vsw_mac_t));

	vsw_node->mode = mode;

	vsw_node->mac_addr = ILLEGAL_MAC_ADDR;

	if (ldm_name != NULL) {
		vsw_node->ldm_name = Xmalloc(strlen(ldm_name) + 1);
		strcpy(vsw_node->ldm_name, ldm_name);
	} else
		vsw_node->ldm_name = NULL;

	vsw_node->service_name = Xmalloc(strlen(service_name) + 1);
	strcpy(vsw_node->service_name, service_name);

	CHAIN_INIT(vsw_node->vnet_list);

	CHAIN_ADD(mac_list.vsw_list, vsw_node, link);

	return (vsw_node);
}

/*
 * This routine parses the database for any instances of VSW nodes. It then
 * calls create_vsw_inst() to populate the mac_list.
 */
void
populate_vsw(xmlNodePtr ldom)
{
	xmlNodePtr	vsw_node, node;
	char		*ldm_name;

	while (ldom != NULL) {
		node = xml_find_subnode(ldom, LDOM_NAME_NODE);
		ldm_name = (char *)xmlNodeGetContent(node);

		node = xml_find_subnode(ldom, VSW_NODE);
		vsw_node = xml_find_subnode(node, VSW_INSTANCE_NODE);

		while (vsw_node != NULL) {
			create_vsw_inst(vsw_node, ldm_name);
			vsw_node = vsw_node->next;
		}

		xmlFree(ldm_name);
		ldom = xml_get_next_ele_node(ldom);
	}
}

/*
 * For each VSW node instance found in the database this routine creates a
 * VSW in the mac_list.
 */
static void
create_vsw_inst(xmlNodePtr inst_node, char *ldm_name)
{
	xmlNodePtr	prop_node;
	char		*service_name, *policies, *dev_path;
	mac_addr_t	mac_addr;
	vsw_mac_t	*vsw_node;

	/* Get the service_name */
	prop_node = xml_find_subnode(inst_node, SERVICE_NAME_NODE);
	service_name = (char *)xmlNodeGetContent(prop_node);

	/* Get the mac_address */
	prop_node = xml_find_subnode(inst_node, MAC_ADDRESS_NODE);
	mac_addr = xml_get_uint64_content(prop_node);

	/* Get the mode */
	prop_node = xml_find_subnode(inst_node, DEV_PATH_NODE);
	dev_path = (char *)xmlNodeGetContent(prop_node);

	if (strcmp(dev_path, " ") == 0)
		policies = VSW_ROUTED;
	else
		policies = VSW_PROMISC;

	create_vsw_mac(&vsw_node, ldm_name, service_name, policies, mac_addr);

	xmlFree(service_name);
	xmlFree(dev_path);
}

/*
 * This routine unlinks a VSW node from the mac_list, frees up memory, and
 * deletes the vsw_node.
 */
void
delete_vsw(vsw_mac_t *vsw_node)
{
	add_used_mac(vsw_node->mac_addr);

	CHAIN_UNLINK(mac_list.vsw_list, vsw_node, link);

	if (vsw_node->ldm_name != NULL)
		Xfree(vsw_node->ldm_name);

	Xfree(vsw_node->service_name);
	Xfree(vsw_node);
}

/*
 * blank_out_vsw() is called then we delete the VSW but still have VNETs
 * attached to them.
 */
void
blank_out_vsw(vsw_mac_t *vsw_node)
{
	add_used_mac(vsw_node->mac_addr);

	vsw_node->mac_addr = ILLEGAL_MAC_ADDR;
	vsw_node->mode = ROUTED_MODE;

	Xfree(vsw_node->ldm_name);
	vsw_node->ldm_name = NULL;
}
