/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <limits.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <libgen.h>
#include <netinet/in.h>
#include <errno.h>
#include <thread.h>
#include <synch.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/corectl.h>
#include <libintl.h>	/* i18n - initialization */

#include "allocate.h"
#include "basics.h"
#include "database.h"
#include "fatal.h"
#include "mdesc_impl.h"
#include "chain.h"
#include "clp.h"
#include "ldom.h"
#include "zpri.h"
#include "backend.h"
#include "ds.h"
#include "ds_impl.h"
#include "option.h"
#include "restart.h"
#include "svcs.h"
#include "zprintf.h"
#include "mac_addrs.h"
#include "mac_multicast.h"
#include "frags.h"
#include "mig.h"
#include "vio.h"
#include "xml_event.h"
#include "utilization.h"
#include "pm/pm_shared.h"
#include "sequencer.h"
#include "ssl_connection.h"
#include "xml_parse.h"
#include "autosave.h"
#include "mig.h"
#include "uuid.h"

struct ldmd_options option;
uint_t debug_level = 0;			/* debug disabled by default */

#if !defined(TEXT_DOMAIN)	/* should be defined by cc -D */
#define	TEXT_DOMAIN "SYS_TEST"	/* use this only if it isn't  */
#endif

static void usage(char *progname);
static void process_opts(int argc, char **argv);

extern void init_service_libs(void);
extern void *client_loop(void *unused);
extern void *xmpp_loop(void *unused);
extern void *sequence(void *unused);

/*
 * Initialize base LDom data structures.
 */
static void
zeus_init(zeus_t *zeusp)
{
	CHAIN_INIT(zeusp->ldoms);
	CHAIN_INIT(zeusp->inprogress);
	CHAIN_INIT(zeusp->free_frags);
	CHAIN_INIT(zeusp->frag_mblocks);
	CHAIN_INIT(zeusp->vds_services);
	CHAIN_INIT(zeusp->mpgroups);
	CHAIN_INIT(zeusp->rsvd_mblocks);
	zeusp->next_hvmd_fp = NULL;
	zeusp->current_hvmd_fp = NULL;
	zeusp->restore.next_pa = INVALID_FRAG_PA;
	zeusp->restore.current_pa = INVALID_FRAG_PA;
	zeusp->hv_mblockp = NULL;
	zeusp->hvctl_ldom_p = NULL;
	zeusp->ds_server = NULL;
	zeusp->ldc_console_config = true;
	zeusp->hv_mdp = NULL;
	zeusp->allow_ctrl_domain_reconfig = false;
	zeusp->do_warm_start = false;
	zeusp->mode = ZEUS_MGMT_MODE;
	zeusp->simulating = false;
	zeusp->reconf_fp = NULL;
	zeusp->prev_del_reconf_hvmd_fp = NULL;
	zeusp->restore.reconf_pa = INVALID_FRAG_PA;
	zeusp->delayed_reconfig_ldom = NULL;
	zeusp->delayed_reconfig_gid = INVALID_GUESTID;
	zeusp->delayed_reconfig_type = DEL_RECONFIG_UNKNOWN;
	zeusp->delayed_hvmd = NULL;
	zeusp->admdir = NULL;
	zeusp->pkgdir = NULL;
	zeusp->domaining_enabled = false;
	zeusp->is_next_spconfig = true;
	zeusp->current_spconfig = NULL; /* currently unknown */
	zeusp->current_cookie = NULL;
	zeusp->local_ip = NULL;
	zeusp->remote_ip = NULL;
	zeusp->sdio_enabled = false;
	mutex_init(&zeusp->mlock, USYNC_THREAD|LOCK_ERRORCHECK, NULL);
}

static void
setup_dir_corefile(void)
{
	char buf[MAXPATHLEN];
	const char *execpath;
	char *path, *dir;

	if (option.in_place) {
		zeus.admdir = ".";
		zeus.pkgdir = ".";
	} else {
		/*
		 * Find the package directory by finding where the ldmd
		 * binary resides and going up one directory level.
		 * If any part of path lookup fails default to
		 * current directory ".".
		 */
		if ((execpath = getexecname()) == NULL) {
			zeus.pkgdir = ".";
		} else {
			path = Xstrdup((char *)execpath);
			dir = dirname(path);

			if (dir[0] == '/') {
				/* absolute path */
				zeus.pkgdir = zsprintf("%s/..", dir);
			} else {
				/* relative path, prepend cwd */
				if (getcwd(buf, sizeof (buf)) != NULL)
					zeus.pkgdir = zsprintf("%s/%s/..",
					    buf, dir);
				else
					zeus.pkgdir = ".";
			}
			Xfree(path);
		}

		/*
		 * Set up the ldmd administrative directory. This is the
		 * directory where the database lives and where core and debug
		 * dump files will be put.
		 */
		zeus.admdir = zsprintf("/var/opt/%s", PKGPATH);

		if ((mkdir(zeus.admdir, S_IRWXU) == -1) && (errno != EEXIST))
			efatal("Could not create LDom Manager admin dir \"%s\"",
			    zeus.admdir);
	}

	/*
	 * If the current process core file location is the system default
	 * "core" then change it to be in the admin dir, otherwise leave the
	 * setting as is.
	 */
	if (core_get_process_path(buf, sizeof (buf), getpid()) == 0 &&
	    strcmp(buf, "core") == 0) {
		snprintf(buf, MAXPATHLEN, "%s/core", zeus.admdir);
		(void) core_set_process_path(buf, strlen(buf) + 1, getpid());
	}
}

int
main(int argc, char **argv)
{
	sigset_t	sig_set;

	/*
	 * The main thread runs the sequencer. We need the sequencer to keep
	 * running in the event of a signal. Make it so the the main thread
	 * does NOT handle the SIGUSR1, SIGINT and SIGTEM signals.  We only
	 * want these signals handled in the client loop thread.
	 */
	sigemptyset(&sig_set);
	sigaddset(&sig_set, SIGINT);
	sigaddset(&sig_set, SIGTERM);
	sigaddset(&sig_set, SIGUSR1);
	thr_sigsetmask(SIG_BLOCK, &sig_set, NULL);

	/*
	 * Prevent EPIPE errors from causing Zeus to terminate.
	 */
	signal(SIGPIPE, SIG_IGN);

	/*
	 * initialize i18n
	 */
	(void) setlocale(LC_MESSAGES, "");
	(void) textdomain(TEXT_DOMAIN);

	/*
	 * Only the account running the LDom manager may see working files.
	 */
	umask(077);

	zeus_init(&zeus);

	option.port = getuid();
	if (option.port == 0)
		option.port = DEFAULT_PORT;

	option.be_selected = false;
	option.tgt_selected = false;
	option.flag_daemonise = false;	/* so we get -help output at start */
	option.load_database = true;	/* database is loaded by default */
	option.mgmt_mode = false;
	option.in_place = false;
	option.validate_xml = false;
	option.allow_crypto_dr = true;
	option.allow_pm = true;
	option.allow_drm = true;
	option.secure_migration = true;
	option.incoming_migration_enabled = true;
	option.outgoing_migration_enabled = true;
	option.warm_migration_compression = true;
	option.mig_nworkers = MIG_NWORKERS_USEDEFAULT;
	option.xmpp_enabled = true;
	option.xmpp_port = XMPP_DEFAULT_PORT;
	option.mac_ttl = MAC_MULTI_TTL;
	option.pri_file = NULL;
	option.min_vcpu_per_ldom = MIN_VCPU_PER_LDOM;
	option.max_pm_cpu_adjust = PM_CPU_ADJUST_LIMIT;
	option.autorecovery_policy = AUTORECOVERY_POLICY_LOG;
#if ALLOW_DEFRAG
	option.allow_defrag = true;
#else
	option.allow_defrag = false;
#endif

#if (ZDBG || ALLOW_DASH_P)
	option.allow_dash_p = true;
#else
	option.allow_dash_p = false;
#endif

	process_opts(argc, argv);

	if (option.flag_daemonise) {
		pid_t pid, zeus_pgid;
		int i, fd;

		restart_init();

		/*
		 * Daemonize thyself
		 */

		pid = fork();
		switch (pid) {
		case -1:
			fatal("fork failed");
			/* NOTREACHED */
		case 0:
			/* child continues */
			break;
		default:
			/* parent waits for child to indicate startup status */
			restart_wait();
			/*NOTREACHED*/
		}

		/* Create new session & process group */
		zeus_pgid = setsid();
		if (zeus_pgid == (pid_t)-1)
			/* Zeus already a PGRP leader */
			fatal("setsid failed");

		fd = restart_child();
		ASSERT(fd >= 3);
		for (i = 3; i < fd; i++)
			close(i);
		closefrom(fd+1);
		fd = open("/dev/null", O_RDWR);
		dup2(fd, 0);
		dup2(fd, 1);
#if USE_SYSLOG
		dup2(fd, 2);
		openlog("ldmd", LOG_PID, LOG_DAEMON);
#endif
		close(fd);
	}

	if (!option.be_selected)
		(void) backend_select(NULL);

	/*
	 * Make sure database and core file directory exists.
	 */
	setup_dir_corefile();

	get_plugged_services();

	database_init();

	backend.start();

	if (!option.tgt_selected)
		(void) target_select(NULL);

	/*
	 * check for libds
	 */
	libds_cap.supported = false;
	libds_cap.config = false;
	libds_cap.functional = false;
	libds_cap.convert_libds = false;

	check_libds_support();

	/*
	 * Initialize Physical Resource Inventory (PRI).
	 */
	if (pri_initialize() != 0)
		fatal("Failed to initialize the PRI");

	if (get_pri_method != GET_PRI_DS) {
		/*
		 * The PRI was not obtained via domain services hence the
		 * phase one domain services initialization was not done. Do
		 * the initialization now. The pri domain service will not be
		 * registered since the PRI has already been obtained via
		 * other means.
		 */
		ds_init1();
	}

	/*
	 * Now that we have the PRI, complete the setup of the target and do
	 * the second phase initialization of domain services.
	 */
	target.complete_setup();
	ds_init2();

	/*
	 * Retrieve MDs from HV (i.e. get actual machine state).  This Will
	 * also determine whether Zeus is running in an LDom, and set
	 * zeus.hvctl_ldom_p appropriately if so.
	 */
	backend.get_hv_state();

	/*
	 * Once the channels are open, register the service or the client with
	 * libds. This routine must be invoked *after* zeus has recorded all
	 * LDoms from HV
	 */
	if (libds_cap.functional == true)
		register_libds_plugins();

	/*
	 * Add any LDoms which are defined in the database but not in
	 * the HV MDs.
	 */
	database_add_unbound_ldoms();

	/*
	 * Add UUID for unbound domains that do not have one yet
	 */
	uuid_init();

	/*
	 * Once unbound ldoms are added, now add dependencies
	 */
	database_add_masters();
	/* old_database is now NULL */

	mac_init();
	(void) start_mac_auto_allocate();

	if (ldmd_setup_SSL() == 0) {
		init_event_lists();
		if (thr_create(NULL, NULL, xmpp_loop, NULL, THR_DAEMON,
		    NULL) != 0)
			efatal("Could not create xmpp_loop thread");

		if (option.incoming_migration_enabled == true)
			migration_init();
	} else {
		warning("SSL setup failed. Migration and XMPP server will"
		    " not be usable.");
	}

	if (thr_create(NULL, NULL, client_loop, NULL, THR_DAEMON, NULL) != 0)
		efatal("Could not create client_loop thread");

	start_guest_util_poll();

	/* if get here then startup was successful, notify parent */
	restart_fini();

	/* Let Power Management startup */
	init_pm_events();
	if (option.allow_pm) {
		PRE_SEQUENCE_ENTER();

		pm_start();
		pri_init_cpu_chips();

		PRE_SEQUENCE_EXIT();
	} else {
		zprintf("Power Management disabled\n");
	}

	if (thr_create(NULL, NULL, read_reboot_file, NULL, THR_DAEMON,
	    NULL) != 0) {
		efatal("Could not create read_reboot_file thread");
	}

	/*
	 * FIXME - for now, we call poweron on every startup.  We need to
	 * have VBSC give us an indication of poweron.
	 */
	autosave_poweron();

	autosave_allow_updates();

	/* memory DR recovery */
	PRE_SEQUENCE_ENTER();
	mem_sync_with_domain(NULL);
	PRE_SEQUENCE_EXIT();

	/* enter the sequencer */
	(void) sequence(NULL);

	return (0);
}

static void
error(char *fmt, ...)
{
	va_list vargs;

	va_start(vargs, fmt);
	zvprintf(fmt, vargs);
	va_end(vargs);
	zprintf("\n");

	exit(EXIT_FATAL);	/* don't want SMF to restart us */
}

static void
process_opts(int argc, char **argv)
{
	int	i;
	int	hops;			/* TTL value read from cmd line */
	bool_t	port_set = false;
	bool_t	flag_daemonise = true;

	for (i = 1; i < argc; i++) {
		char *strp;

		if (strcmp(argv[i], "-c") == 0) {
			if (option.mgmt_mode == true)
				error("Can only use one of -c and -m");
			zeus.mode = ZEUS_CONFIG_MODE;
		} else if (strcmp(argv[i], "-m") == 0) {
			if (zeus.mode == ZEUS_CONFIG_MODE)
				error("Can only use one of -c and -m");
			option.mgmt_mode = true;
		} else if (strcmp(argv[i], "-p") == 0) {
			if (port_set)
				error("Port value specified more than once");
			if (i + 1 < argc) {
				strp = argv[++i];
				option.port = atoi(strp);
				if (option.port < MIN_CONSOLE_TCP_PORT ||
				    option.port > MAX_CONSOLE_TCP_PORT)
					error("Illegal port value specified: "
					    "%d", option.port);
			} else {
				error("Port value is missing");
			}
			port_set = true;
		} else if (strcmp(argv[i], "-g") == 0) {
			if (i + 1 < argc) {
				strp = argv[++i];
				debug_level = strtoul(strp, NULL, 0);
			} else {
				error("Debug level is missing");
			}
		} else if (strcmp(argv[i], "-b") == 0) {
			if (option.be_selected)
				error("Backend specified more than once");
			strp = argv[++i];

			if (!backend_select(strp))
				error("Unknown backend specified");
			option.be_selected = true;
		} else if (strcmp(argv[i], "-t") == 0) {
			if (option.tgt_selected)
				error("Target specified more than once");
			strp = argv[++i];

			if (!target_select(strp))
				error("Unknown target specified");
			option.tgt_selected = true;
		} else if (strcmp(argv[i], "-pri") == 0) {
			option.pri_file = argv[++i];
		} else if (strcmp(argv[i], "-bl") == 0) {
			backend_list();
			exit(0);
		} else if (strcmp(argv[i], "-tl") == 0) {
			target_list();
			exit(0);
		} else if (strcmp(argv[i], "-d") == 0) {
			flag_daemonise = false;
		} else if (strcmp(argv[i], "-i") == 0) {
			option.in_place = true;
		} else if (strcmp(argv[i], "-n") == 0) {
			option.load_database = false;
		} else if (strcmp(argv[i], "-x") == 0) {
			option.validate_xml = true;
		} else if (strcmp(argv[i], "-h") == 0 ||
		    strcmp(argv[i], "--help") == 0 ||
		    strcmp(argv[i], "-?") == 0) {
			usage(argv[0]);
		} else if (strcmp(argv[i], "-hops") == 0) {
			/*
			 * Check to see if the number of hops is specified.
			 * If -hops is the last arg or the next one is a new
			 * command line switch, bail out.
			 */
			if ((i == argc - 1) || (argv[++i][0] == '-'))
				error("The number of hops (TTL) needs "
				    "to be specified");
			strp = argv[i];
			hops = atoi(strp);
			if (hops < 0)
				error("Illegal hops value specified: %d", hops);
			option.mac_ttl = hops;
		} else if (strcmp(argv[i], "-autorecovery-policy") == 0) {
			strp = argv[++i];
			if (strp == NULL) {
				warning("No -autorecovery-policy value "
				    "specified, default to LOG");
				option.autorecovery_policy =
				    AUTORECOVERY_POLICY_LOG;
			} else {
				option.autorecovery_policy = atoi(strp);
			}
			switch (option.autorecovery_policy) {
			case AUTORECOVERY_POLICY_LOG:
			case AUTORECOVERY_POLICY_NOTIFY:
			case AUTORECOVERY_POLICY_RECOVER:
				break;
			default:
				warning("Illegal -autorecovery-policy value "
				    "specified: %d, default to LOG",
				    option.autorecovery_policy);
				option.autorecovery_policy =
				    AUTORECOVERY_POLICY_LOG;
				break;
			}
		} else if (strcmp(argv[i], "-nodefrag") == 0) {
			option.allow_defrag = false;
		} else if (strcmp(argv[i], "-minvcpu") == 0) {
			strp = argv[++i];
			option.min_vcpu_per_ldom = atoi(strp);
			if (option.min_vcpu_per_ldom < MIN_VCPU_ALLOWED)
				error("Illegal minimum number of vcpus "
				    "specified: %d", option.min_vcpu_per_ldom);
		} else if (strcmp(argv[i], "-maxpmcpu") == 0) {
			strp = argv[++i];
			option.max_pm_cpu_adjust = atoi(strp);
			if (option.max_pm_cpu_adjust < 1)
				error("Illegal maximum number of PM cpus "
				    "specified: %d", option.max_pm_cpu_adjust);
#if ZDBG
		} else if (strcmp(argv[i], "-nocryptodr") == 0) {
			option.allow_crypto_dr = false;
		} else if (strcmp(argv[i], "-nopm") == 0) {
			option.allow_pm = false;
		} else if (strcmp(argv[i], "--disable-drm") == 0) {
			option.allow_drm = false;
#endif
		} else if (strcmp(argv[i], "--disable-xmpp") == 0) {
			option.xmpp_enabled = false;
		} else if (strcmp(argv[i], "--disable-incoming-migration") ==
		    0) {
			option.incoming_migration_enabled = false;
		} else if (strcmp(argv[i], "--disable-outgoing-migration") ==
		    0) {
			option.outgoing_migration_enabled = false;
#if ZDBG
		} else if (strcmp(argv[i], "--no-migration-compression") == 0) {
			option.warm_migration_compression = false;
		} else if (strcmp(argv[i], "--migration-nworkers") == 0) {
			if (i + 1 < argc) {
				strp = argv[++i];
				option.mig_nworkers = atoi(strp);
				if (option.mig_nworkers < MIG_NWORKERS_MIN ||
				    option.mig_nworkers > MIG_NWORKERS_MAX) {
					error("Illegal number of worker "
					    "threads specified: %d",
					    option.mig_nworkers);
				}
			} else {
				error("No argument supplied to "
				    "--migration-nworkers");
			}
#endif
		} else if (strcmp(argv[i], "--enable-dash-p") == 0) {
			option.allow_dash_p = true;
		} else if (strcmp(argv[i], "--disable-dash-p") == 0) {
			option.allow_dash_p = false;
#if ZDBG
		} else if (strcmp(argv[i], "--disable-migration-ssl") == 0) {
			option.secure_migration = false;
#endif
		/* Process backend & target specific options last */
		} else if (option.be_selected &&
		    backend.process_option(&i, argc, argv)) {
			/* EMPTY */
		} else if (option.tgt_selected &&
		    target.process_option(&i, argc, argv)) {
			/* EMPTY */
		} else {
			error("Unknown command line parameter %s", argv[i]);
		}
	}

	/*
	 * Any options requiring output should be processed by this point,
	 * which means we can now configure daemonise to the desired value.
	 */
	option.flag_daemonise = flag_daemonise;
}

static void
usage(char *progname)
{
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "\t%s [-d] [-b <backend>] [-t <target>] [-p <port>] "
	    "[-h] [--disable-xmpp]\n"
	    "\t\t[--disable-incoming-migration] "
	    "[--disable-outgoing-migration]\n"
	    "\t\t[-autorecovery-policy <policy>]\n",
	    progname);
	fprintf(stderr, "\t%s [-bl]\n", progname);
	fprintf(stderr, "\t%s [-tl]\n", progname);
	fprintf(stderr, "\n");
	fprintf(stderr, "\t\t-d\tDo not run as daemon\n");
	fprintf(stderr, "\t\t-b\tSelect backend of <backend>\n");
	fprintf(stderr, "\t\t-t\tSelect target of <target>\n");
	fprintf(stderr, "\t\t-p\tUse <port> as the CLI interface\n");
	fprintf(stderr, "\t\t-n\tDon't load the database\n");
	fprintf(stderr, "\t\t-bl\tlist available backends\n");
	fprintf(stderr, "\t\t-tl\tlist available targets\n");
#if ZDBG
	fprintf(stderr, "\n");
	fprintf(stderr, "\tDebug options:\n");
	fprintf(stderr, "\t\t-nopm\t\tDisable Power Management\n");
	fprintf(stderr, "\t\t-nocryptodr\tDisable Crypto DR\n");
	/*
	 * these are also available in non-debug but I'm not sure we want to
	 * expose them in the standard help message - best leave them as
	 * undocumented for now
	 */
	fprintf(stderr, "\t\t-nodefrag\tDon't consolidate vcpus\n");
	fprintf(stderr, "\t\t-minvcpu\tmin vcpus powered up per ldom\n");
	fprintf(stderr, "\t\t-maxpmcpu\tmax number of vcpus to adjust power\n");
	fprintf(stderr, "\t\t-pri <file>\tUse <file> as PRI\n");
	fprintf(stderr, "\t\t-g <level>\tSet debug level\n");
	fprintf(stderr, "\t\t-c\t\tConfig mode\n");
	fprintf(stderr, "\t\t-m\t\tManagement mode\n");
	fprintf(stderr, "\t\t-x\t\tValidate XML\n");
#endif

	exit(EXIT_FATAL);	/* don't want SMF to restart us */
}
