/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <stdlib.h>

#include "basics.h"
#include "support.h"
#include "zprintf.h"
#include "fatal.h"
#include "chain.h"
#include "resource_id.h"
#include "allocate.h"
#include "database.h"
#include "database_internal.h"
#include "master.h"

static int count_masters(mlist_t *mlist, ldom_t *mldp);
static mlist_t *master2mlist(ldom_t *ldp);

bool_t
get_failure_policy(char *str, uint64_t *policy)
{
	if (strcmp(str, "ignore") == 0) {
		*policy = (uint64_t)FAILURE_POLICY_IGNORE;
		return (true);
	} else if (strcmp(str, "stop") == 0) {
		*policy = (uint64_t)FAILURE_POLICY_STOP;
		return (true);
	} else if (strcmp(str, "reset") == 0) {
		*policy = (uint64_t)FAILURE_POLICY_RESET;
		return (true);
	} else if (strcmp(str, "panic") == 0) {
		*policy = (uint64_t)FAILURE_POLICY_PANIC;
		return (true);
	}

	return (false);
}

/*
 * Returns a string that represents the failure policy passed in. The
 * string array used locally must be kept in sync with the command line
 * arguments available to the user to set the failure policy mode. If
 * the policy is not valid, the default policy to ignore all failures
 * is assumed. Memory is allocated for the policy string that must be
 * deallocated by the caller.
 */
char *
get_failure_policy_string(uint64_t policy)
{
	char *policy_str[] = { "ignore", "stop", "reset", "panic" };

	/* default is to not take any action */
	if (!VALID_FAILURE_POLICY(policy))
		policy = (uint64_t)FAILURE_POLICY_IGNORE;

	return (Xstrdup(policy_str[(int)policy]));
}

/*
 * Generates a string that contains a comma separated list of all LDoms
 * on which the specified LDom depends. If there are no dependencies, a
 * NULL pointer is returned. Memory is allocated for the dependency string
 * that must be deallocated by the caller.
 */
char *
get_masters_string(ldom_t *ldp)
{
	char		*masters_str = NULL;
	master_t	*master;

	CHAIN_ITER(ldp->props.masters, master, link) {

		if (!is_valid_master(master))
			continue;

		/* allocate memory for the first valid entry */
		if (masters_str == NULL) {
			masters_str = Xmalloc(pri.max_guest_dependencies *
			    (MAXLEN_VALID_NAME + 1));
			sprintf(masters_str, "%s", master->ldp->namep);
		} else {
			strcat(masters_str, ",");
			strcat(masters_str, master->ldp->namep);
		}
	}

	return (masters_str);
}

static bool_t
is_valid_ldp(ldom_t *ldomp)
{
	ldom_t	*ldp = NULL;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp == ldomp)
			return (true);
	}
	return (false);
}

bool_t
is_valid_master(master_t *mp)
{
	if ((mp == NULL) || (mp->ldp == NULL))
		return (false);

	/* scan for the master ldom */
	return (is_valid_ldp(mp->ldp));
}

bool_t
has_bound_slaves(ldom_t *ldp, response_t *resp)
{
	ldom_t		*slave_ldp;
	master_t	*master;
	bool_t		rc = false;

	CHAIN_ITER(zeus.ldoms, slave_ldp, link) {
		if (slave_ldp->state < BOUND)
			continue;
		CHAIN_ITER(slave_ldp->props.masters, master, link) {
			if (master->ldp != ldp)
				continue;

			rprintf(resp, gettext("LDom \"%s\" is bound with a "
			    "dependency on LDom \"%s\"\n"), slave_ldp->namep,
			    ldp->namep);

			rc = true;
		}
	}
	return (rc);
}

/*
 * Before rm-domain of a master completes, shoot-down all
 * slave domains that could hold a stray pointer to the
 * master about to be destroyed.
 */
void
invalidate_slaves(ldom_t *master_ldp)
{
	ldom_t		*ldp = NULL;
	master_t	*master = NULL;
	master_t	*next = NULL;
	bool_t		purged;


	/* There's nothing to do */
	if (!LDOM_DEPENDENCIES_SUPPORTED())
		return;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp == master_ldp)
			continue;

		purged = false;
		CHAIN_ITER_VOL(ldp->props.masters, master, next, link) {
			if (master->ldp == master_ldp) {
				purged = true;
				CHAIN_UNLINK(ldp->props.masters, master, link);
				Xfree(master);
			}
		}
		if (purged == true)
			database_ldom_set_master(ldp);
	}
}

void
print_failure_policy(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	char	*policy_str;

	if ((dctl->oformat && !dctl->odomain) || !LDOM_DEPENDENCIES_SUPPORTED())
		return;

	policy_str = get_failure_policy_string(ldp->props.failure_policy);

	rprintf(resp, "%s", dctl_label("control", dctl));

	if (dctl->parseable) {
		rprintf(resp, "|failure-policy=%s\n", policy_str);
	} else {
		rprintf(resp, "\n    failure-policy=%s\n\n", policy_str);
	}

	Xfree(policy_str);
}

void
print_master(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	char	*masters_str;

	if ((dctl->oformat && !dctl->odomain) || !LDOM_DEPENDENCIES_SUPPORTED())
		return;

	masters_str = get_masters_string(ldp);

	/* handle the case where the domain has no masters */
	if (masters_str == NULL) {
		if (dctl->longf)
			masters_str = zsprintf(" ");
		else
			return;
	}

	rprintf(resp, "%s", dctl_label("dependency", dctl));
	if (dctl->parseable)
		rprintf(resp, "|master=%s\n", masters_str);
	else
		rprintf(resp, "\n    master=%s\n\n", masters_str);

	Xfree(masters_str);
}

bool_t
bind_dependency(ldom_t *ldp, response_t *resp)
{
	master_t	*master = NULL;
	mlist_t		*mlist = NULL;
	bool_t		rc;

	if (validate_masters(ldp) == false) {
		fatal("Internal error: LDom %s has %d masters\n", ldp->namep,
		    NUM_DEPENDENCIES(ldp));
		return (false);
	}

	/* Make sure all masters are bound */
	CHAIN_ITER(ldp->props.masters, master, link) {
		if (master->ldp->state < BOUND) {
			rprintf(resp, gettext("Cannot bind LDom \"%s\", "
			    "because LDom \"%s\" is not bound\n"), ldp->namep,
			    master->ldp->namep);
			return (false);
		}
	}

	/* Prevent cycles of dependency */
	mlist = master2mlist(ldp);
	rc = dependency_cycle(ldp, mlist, resp);
	free_mlist(mlist);
	return ((rc == true) ? false : true);
}

bool_t
dependency_cycle(ldom_t *ldp, mlist_t *mlist, response_t *resp)
{
	mlist_t		*ml = NULL;
	bool_t		rc = false;

	for (ml = mlist; ml != NULL; ml = ml->next) {

		if (!is_valid_ldp(ml->master))
			continue;

		if (ml->master == ldp) {
			rprintf(resp, gettext("Dependency cycle detected: "
			    "LDom %s has a dependency on LDom %s\n"),
			    ml->namep, ldp->namep);
			return (true);
		} else {
			mlist_t	*rlist = master2mlist(ml->master);

			rc = dependency_cycle(ldp, rlist, resp);
			free_mlist(rlist);
			if (rc == true)
				return (true);
		}
	}

	return (false);
}

bool_t
validate_masters(ldom_t *ldp)
{
	master_t	*master;
	master_t	*next;

	if (NUM_DEPENDENCIES(ldp) == 0)
		return (true);

	CHAIN_ITER_VOL(ldp->props.masters, master, next, link) {

		/* skip valid masters */
		if (is_valid_master(master))
			continue;

		DBG(zprintf("%s: ldom=%s contains an invalid master\n",
		    __func__, ldp->namep));

		/* get rid of the invalid master */
		CHAIN_UNLINK(ldp->props.masters, master, link);
		Xfree(master);
	}

	if (DEPENDENCY_LIMIT_OVERFLOW(ldp)) {
		DBG(zprintf("%s: LDom %s has %d masters, min is 0 and "
		    "max is %d\n", __func__, ldp->namep, NUM_DEPENDENCIES(ldp),
		    pri.max_guest_dependencies));
		return (false);
	}

	return (true);
}

bool_t
unique_masters(mlist_t *mlist, response_t *resp)
{
	mlist_t		*ml = NULL;
	int		num = 0;

	if (mlist == NULL)
		return (false);

	for (ml = mlist; ml != NULL; ml = ml->next) {
		if ((num = count_masters(mlist, ml->master)) > 1) {
			rprintf(resp, gettext("LDom \"%s\" was specified %d "
			    "times\n"), ml->master->namep, num);
			return (false);
		}
	}

	return (true);
}

static int
count_masters(mlist_t *mlist, ldom_t *mldp)
{
	mlist_t		*ml = NULL;
	int		found = 0;

	for (ml = mlist; ml != NULL; ml = ml->next) {
		if (ml->master == mldp)
			++found;
	}

	return (found);
}

static mlist_t *
master2mlist(ldom_t *ldp)
{
	master_t	*master = NULL;
	mlist_t		*mlist = NULL;
	mlist_t		*ml = NULL;

	CHAIN_ITER(ldp->props.masters, master, link) {
		if (ml == NULL) {
			ml = Xcalloc(1, mlist_t);
			mlist = ml;
		} else {
			ml->next = Xcalloc(1, mlist_t);
			ml = ml->next;
		}
		ml->namep = ldp->namep;
		ml->master = master->ldp;
		ml->next = NULL;
	}

	return (mlist);
}

void
mlist2master(ldom_t *ldp, mlist_t *mlist)
{
	master_t	*master = NULL;
	mlist_t		*ml = NULL;

	free_masters(ldp);
	for (ml = mlist; ml != NULL; ml = ml->next) {
		master = Xcalloc(1, master_t);
		master->ldp = ml->master;
		CHAIN_ADD(ldp->props.masters, master, link);

		/* truncate list if dependency limit reached */
		if (DEPENDENCY_LIMIT_REACHED(ldp))
			break;
	}
}

void
free_mlist(mlist_t *mlist)
{
	mlist_t	*ml = mlist;

	while (ml != NULL) {
		mlist = mlist->next;
		Xfree(ml);
		ml = mlist;
	}
}

void
free_masters(ldom_t *ldp)
{
	master_t	*master = NULL;
	master_t	*next = NULL;

	CHAIN_ITER_VOL(ldp->props.masters, master, next, link) {
		CHAIN_UNLINK(ldp->props.masters, master, link);
		Xfree(master);
	}
}
