/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>

#include "basics.h"
#include "chain.h"
#include "database.h"
#include "mdesc_impl.h"
#include "ldom.h"
#include "clp.h"
#include "allocate.h"
#include "errors.h"
#include "zprintf.h"
#include "zpri.h"
#include "response.h"
#include "fatal.h"
#include "ldomconfigure.h"
#include "sequencer.h"
#include "vcpu.h"
#include "vmau.h"
#include "dr_crypto.h"
#include "autosave.h"
#include "core_md.h"

/*
 * This file implements crypto resources. There is one crypto unit (mau) per
 * core. A guest have access to an MAU on a core provided at least one strand
 * from the core belongs to the guest.
 */


static vmau_t *
id_to_vmau(ldom_t *ldp, int id)
{
	vmau_t *vmau;

	CHAIN_ITER(ldp->bound.vmaus, vmau, ldom) {
		if (vmau->mau->mau_id == id)
			break;
	}

	return (vmau);
}

static pri_mau_t *
id_to_mau(int id)
{
	pri_mau_t *mau = NULL;

	CHAIN_ITER(pri.mau_list, mau, link) {
		if (mau->mau_id == id)
			break;
	}

	return (mau);
}

/*
 * Number of virtual maus bound to a LDOM.
 */
int
ldom_num_bound_maus(ldom_t *ldp)
{
	return (CHAIN_LENGTH(ldp->bound.vmaus));
}

/*
 * Get the free amount from a "physical" mau
 */
static int
mau_get_free_amt(pri_mau_t *mau)
{
	if (CHAIN_LENGTH(mau->bound.vmaus) > 0)
		return (0);
	else
		return (1);
}

int
check_crypto_dr_services(ldom_t *ldp)
{
	ds_svc_t *mau_svcp;
	ds_svc_t *cwq_svcp;

	/* lookup the crypto DR services */
	mau_svcp = ds_lookup_svc(ldp->ds_chan, DR_CRYPTO_MAU_SVC);
	cwq_svcp = ds_lookup_svc(ldp->ds_chan, DR_CRYPTO_CWQ_SVC);

	/*
	 * If at least one of the services is available, the
	 * domain supports crypto DR. So, only fail if neither
	 * service is available.
	 */
	if ((mau_svcp == NULL) && (cwq_svcp == NULL))
		return (ERR_CRYPTO_DR_NOTSUP);

	return (CRYPTO_DR_OP_SUCC);
}

/* Check to see if a domain is capable of doing Crypto DR operations */
int
check_crypto_dr_capable(ldom_t *ldp)
{
	ASSERT(ldp->state == ACTIVE);

	if (check_crypto_dr_services(ldp) != 0)
		return (ERR_CRYPTO_DR_NOTSUP);

	if (ds_lookup_svc(ldp->ds_chan, MD_UPDATE_SVC) == NULL)
		return (ERR_MD_UPDATE_NOTSUP);

	return (0);
}

/*
 * Check to see if a domain is capable of doing Crypto DR operations
 * This function also results in an error message being displayed.
 */
int
cli_check_crypto_dr_capable(ldom_t *ldp, response_t *resp)
{
	int rc;

	if ((rc = check_crypto_dr_capable(ldp)) != 0) {

		if (ldp->hv_state.soft != SIS_NORMAL)
			rprintf(resp, "Notice: guest operating system "
			    "may not be running.\n");

		if (rc == ERR_CRYPTO_DR_NOTSUP) {
			rprintf(resp, gettext("LDom %s is unable to "
			    "dynamically reconfigure crypto units.\n"),
			    ldp->namep);
		} else if (rc == ERR_MD_UPDATE_NOTSUP) {
			rprintf(resp, gettext("LDom %s does not support "
			    "runtime re-configuration.\n"), ldp->namep);
		} else {
			FATAL_UNEXPECTED_ERROR(rc);
		}
	}

	return (rc);
}

static int
dr_add_cryptos(ldom_t *ldp, uint32_t num, response_t *resp)
{
	int ret;

	if (zeus.mode == ZEUS_CONFIG_MODE)
		return (0);

	if ((ret = seq_add_cryptos(ldp, resp)) == CRYPTO_DR_OP_FAILED) {
		ldp->required.num_vmaus -= num;
		rprintf(resp, gettext("Unable to add the requested crypto "
		    "unit(s) to LDom %s\n"), ldp->namep);

	} else if (ret == CRYPTO_DR_OP_NO_RESPONSE) {
		rprintf(resp, gettext("Crypto unit(s) are considered "
		    "allocated to LDom, but might not be available\n to "
		    "the guest OS\n"));

	} else if ((ret != 0) && (ret != num)) {
		/* reduce required by amount that couldn't be allocated */
		ldp->required.num_vmaus -= (num - ret);
		rprintf(resp, gettext("Only %d crypto unit(s) could be "
		    "added to LDom %s\n"), ret, ldp->namep);
	}

	return (ret);
}

/*
 * Add mau(s) to a guest. If the guest is already bound, must do the necessary
 * check to ensure that a free MAU is available which is accessible through
 * one of the strands bound to the guest.
 */
bool_t
mau_add(ldom_t *ldp, mau_adm_t *mau_adm, response_t *resp)
{
	int ret = 0;

	DBG(zprintf("%s: %d\n", __func__, mau_adm->num));

	if (ldp->state == STARTING || ldp->state == STOPPING) {
		rprintf(resp, gettext("Cannot add crypto units to LDom %s "
		    "while it is currently %s.\n"), ldp->namep,
		    ldom_state_str(ldp->state));
		return (false);
	}

	if (ldp->state >= BOUND) {
		if (cons_bind_vmaus(ldp, mau_adm, resp) == false) {
			return (false);
		}

		ldp->required.num_vmaus += mau_adm->num;

		if (ldp->state == ACTIVE && zeus.mode == ZEUS_MGMT_MODE) {
			/*
			 * If already in delayed reconfig or unable to perform
			 * crypto DR, proceed with delayed reconfig
			 */
			if (in_delayed_reconfig(NULL) ||
			    cli_check_crypto_dr_capable(ldp, resp) != 0) {
				/* handle as a delayed reconfig operation */
				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_ADD, resp);
			} else {
				ret = dr_add_cryptos(ldp, mau_adm->num, resp);
			}
		} else {
			/* we can reconfigure now */
			seq_download_mds_do_config(ldp);
		}
	} else {
		ldp->required.num_vmaus += mau_adm-> num;
	}

	if (ret != CRYPTO_DR_OP_FAILED)
		database_ldom_set_maus(ldp, ldp->required.num_vmaus);

	return ((ret < 0) ? false : true);
}

static void
clear_visited_flags(void)
{
	pri_mau_t *mau;

	CHAIN_ITER(pri.mau_list, mau, link) {
		mau->flag_visited = false;
	}
}

static void
mau_unbind(vmau_t *vmau)
{
	ldom_t *ldp = vmau->ldp;
	vcpu_t *vcpu;

	CHAIN_UNLINK(vmau->mau->bound.vmaus, vmau, pmau);
	CHAIN_UNLINK(ldp->bound.vmaus, vmau, ldom);

	CHAIN_ITER(vmau->vcpus, vcpu, mau) {
		vcpu->vmau = NULL;
	}

	free_res(vmau->vmauid);		/* guest vmau id */

	Xfree(vmau);
}

static bool_t
vmau_in_list(uint64_t vid, int num, struct vmau **vmaus)
{
	int i;

	for (i = 0; i < num; i++) {
		if (vid == vmaus[i]->vid)
			return (true);
	}

	return (false);
}

/*
 * Unbind the specified crypto units from the guest. We go through
 * the bound cryptos in the guest, and unbind the ones in the list.
 */
void
cleanup_partial_crypto_add(ldom_t *ldp, int num, struct vmau **vmaus)
{
	vmau_t *vmau, *nextvmau;

	CHAIN_ITER_VOL(ldp->bound.vmaus, vmau, nextvmau, ldom) {
		if (vmau_in_list(vmau->vid, num, vmaus) == true) {
			DBG(zprintf("%s: unbinding vmau %" PRId64 "\n",
			    __func__, vmau->vid));
			mau_unbind(vmau);
		}
	}
}

static bool_t
get_maus_to_delete(ldom_t *ldp, mau_adm_t *mau_adm, response_t *resp,
    vmau_t **vmaus)
{
	int i;

	if (mau_adm->type == CPU_ADM_NUM) {
		int num_bound = ldom_num_bound_maus(ldp);
		vmau_t *vmau;
		vcpu_t *vcpu;

		if (mau_adm->num > num_bound) {
			rprintf(resp, gettext("LDom %s does not have %d crypto "
			    "unit(s) bound\n"), ldp->namep, mau_adm->num);
			return (false);
		}

		/*
		 * For now, just pull them off the list from
		 * highest to lowest id, making sure that at
		 * least one of the corresponding vcpus is available
		 * (i.e. not being power managed).
		 */
		i = 0;
		vmau = CHAIN_END(ldp->bound.vmaus);
		while ((vmau != NULL) && (i < mau_adm->num)) {
			CHAIN_ITER(vmau->vcpus, vcpu, mau) {
				if (get_core_md_cpu_status(vcpu->cpu->phys_id,
				    CORE_MD_PM_STATUS) == PM_NORMAL)
					break;
			}
			if (vcpu == NULL) {
				DBG(zprintf("Skipping MAU %"PRIu64"; all its"
				    " strands are parked by PM\n", vmau->vid));
				rprintf(resp, gettext("Cannot remove one or "
				    "more crypto units due to PM activity."
				    " Set the PM\n policy to performance "
				    "and try again\n"));
				return (false);
			}
			vmaus[i++] = vmau;
			vmau = CHAIN_PREV(vmau->ldom);
		}

		ASSERT(i == mau_adm->num);
	} else {
		for (i = 0; i < mau_adm->num; i++) {
			int id = mau_adm->ids[i];
			vmaus[i] = id_to_vmau(ldp, id);
			if (vmaus[i] == NULL) {
				rprintf(resp, gettext("crypto unit %d is not "
				    "bound in LDom %s\n"), id, ldp->namep);
				return (false);
			}
		}
	}

	return (true);
}

/*
 * If partial removal is not allowed, this routine would add any crypto unit
 * back if the domain refused to remove it
 */
static int
dr_rm_cryptos(ldom_t *ldp, vmau_t **vmaus, int *num,
    response_t *resp, bool_t restore)
{
	uint32_t	*op_stats;
	int		i, rc;
	int		ret = 0;
	int		num_removed = 0;

	/* create an array for the DR status of each crypto unit */
	op_stats = Xmalloc(*num * (int)sizeof (uint32_t));
	for (i = 0; i < *num; i++)
		op_stats[i] = DR_CRYPTO_RES_OK;

	if (zeus.mode == ZEUS_CONFIG_MODE)
		rc = *num;
	else
		rc = seq_rm_cryptos(ldp, vmaus, *num, op_stats, resp);

	if (rc == CRYPTO_DR_OP_FAILED) {
		rprintf(resp, gettext("Unable to remove the requested "
		    "crypto unit(s) from LDom %s\n"), ldp->namep);
		*num = 0;
		ret = ERR_CRYPTO_DR_OP_FAILED;
	} else if (rc == CRYPTO_DR_OP_NO_RESPONSE) {
		rprintf(resp, gettext("crypto unit(s) will remain allocated "
		    "to LDom, but might not be available\nto the guest OS\n"));
		*num = 0;
		ret = ERR_CRYPTO_DR_OP_NO_RESPONSE;

	} else if (rc < 0) {
		FATAL_UNEXPECTED_ERROR(rc);
	} else {
		int	j;
		vmau_t	**re_vmaus;

		j = 0;
		re_vmaus = Xcalloc(*num, vmau_t *);
		for (i = 0; i < *num; i++) {
			if (op_stats[i] == DR_CRYPTO_RES_OK) {
				re_vmaus[j++] = vmaus[i];
			}
		}
		num_removed = j;

		/*
		 * The return value must match the number of cryptos
		 * successfully removed
		 */
		ASSERT(num_removed == rc);

		/*
		 * Unbind the crypto units for normal crypto removal.
		 * Otherwise for migration (restore == true), unbind only if
		 * all the crypto units were successfully removed. For partial
		 * removal with restore being true, we attempt to add back the
		 * successfully removed units - and if that fails too, we unbind
		 * the crypto units the guest was able to remove.
		 */
		if (num_removed == 0) {
			rprintf(resp, gettext("LDom %s didn't allow "
			    "removal of any crypto unit\n"), ldp->namep);
			ret =  ERR_CRYPTO_DR_OP_FAILED;
		} else if (num_removed == *num || restore == false) {
			if (num_removed < *num)
				rprintf(resp, gettext("LDom %s permitted "
				    "removal of only %d crypto unit(s)\n"),
				    ldp->namep, num_removed);
			/* unbind each crypto unit that was unconfigured */
			for (i = 0; i < num_removed; i++) {
				mau_unbind(re_vmaus[i]);
			}
			seq_rm_cryptos_cleanup(ldp);
			autosave_update_autosave();
		} else {
			/*
			 * Partial removal with restore equals true. We just
			 * instruct the guest to add the just removed cryptos
			 * back
			 */
			rprintf(resp, gettext("LDom %s permitted removal "
			    "of only %d crypto unit(s)\n"), ldp->namep,
			    num_removed);

			/*
			 * seq_reconfig_cryptos() will ask the guest to add the
			 * just removed cryptos. If the guest cannot add a
			 * particular crypto unit back, it's unbound
			 */
			ret = seq_reconfig_cryptos(ldp, re_vmaus, num_removed,
			    resp);
			if (ret == num_removed ||
			    ret == CRYPTO_DR_OP_NO_RESPONSE) {
				/*
				 * If no response from the guest, we assume
				 * the guest may actually have added them and
				 * we keep all the cryptos bound to the guest
				 */
				goto skip;
			} else if (ret < 0) {
				/*
				 * Failed to add any crypto. Remove these from
				 * the guest
				 */
				cleanup_partial_crypto_add(ldp, num_removed,
				    re_vmaus);
			}

			/*
			 * if (ret < num_removed), seq_reconfig_cryptos() would
			 * have already un-bound the crypto units which guest
			 * could not add
			 */
			if (ret < 0 || ret < num_removed) {
				seq_rm_cryptos_cleanup(ldp);
				autosave_update_autosave();
			}
		}

skip:
		Xfree(re_vmaus);

		*num = num_removed;
	}

	Xfree(op_stats);

	return (ret);
}

bool_t
mau_del(ldom_t *ldp, mau_adm_t *mau_adm, response_t *resp)
{
	vmau_t	**vmaus;
	int	num = mau_adm->num;
	int	i;

	DBG(zprintf("%s: request to remove %d crypto unit%s\n", __func__,
	    num, (num == 1) ? "" : "s"));

	if (ldp->state == STARTING || ldp->state == STOPPING) {
		rprintf(resp, gettext("Cannot remove crypto units from LDom %s "
		    "while it is currently %s.\n"), ldp->namep,
		    ldom_state_str(ldp->state));
		return (false);
	}

	switch (ldp->state) {
	case BOUND:
	case ACTIVE:
		vmaus = Xmalloc(num * (int)sizeof (vmau_t *));
		if (!get_maus_to_delete(ldp, mau_adm, resp, vmaus)) {
			Xfree(vmaus);
			return (false);
		}

		if (ldp->state == ACTIVE && zeus.mode == ZEUS_MGMT_MODE) {
			/*
			 * If already in delayed reconfig or unable to perform
			 * crypto DR, proceed with delayed reconfig
			 */
			if (in_delayed_reconfig(NULL) ||
			    cli_check_crypto_dr_capable(ldp, resp) != 0) {
				/*
				 * Handle as a delayed reconfig operation.
				 * First call unbind, then initiate the delayed
				 * reconfig.
				 */
				for (i = 0; i < num; i++)
					mau_unbind(vmaus[i]);

				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_REMOVE_PHYS, resp);
			} else {
				if (dr_rm_cryptos(ldp, vmaus, &num, resp,
				    false) > 0) {
					/* Failed to remove any */
					if (num == 0) {
						Xfree(vmaus);
						return (false);
					}
				}
			}
		} else {
			/* unbind & reconfigure now */
			for (i = 0; i < num; i++)
				mau_unbind(vmaus[i]);

			seq_download_mds_do_config(ldp);
		}

		/* Complete or partial success */
		ldp->required.num_vmaus -= num;

		Xfree(vmaus);
		break;

	case INACTIVE:
		if (num > ldp->required.num_vmaus) {
			rprintf(resp, gettext("You do not have %d crypto "
			    "unit(s) requested\n"), num);
			return (false);
		}

		ldp->required.num_vmaus -= num;

		break;

	default:
		abort();	/* Internal error */
	}

	database_ldom_set_maus(ldp, ldp->required.num_vmaus);

	return (true);
}

/*
 * Bind an MAU to the guest.
 * Here we just ensure that the MAU is available and
 * attach a vmau to it.
 */
bool_t
mau_bind(ldom_t *ldp, pri_mau_t *mau, int vid)
{
	vmau_t *vmau;

	if (mau == NULL)
		return (false);

	/* check whether the physical MAU resource is available */
	if (mau_get_free_amt(mau) < 1)
		return (false);

	DBG(zprintf("Binding guest %s vmau %d physical mau %s\n",
	    ldp->namep, vid, mau->repr));

	vmau = Xmalloc(sizeof (*vmau));

	if (vid == -1)
		vmau->vmauid = alloc_res(&ldp->guest_vmau_ids);
	else
		vmau->vmauid = pull_res_id(&ldp->guest_vmau_ids, vid);

	if (vmau->vmauid == NULL) {
		DBG(zprintf("Unable to obtain vmau %d\n", vid));
		Xfree(vmau);
		return (false);
	}

	vmau->vid = vmau->vmauid->id;
	vmau->mau = mau;
	vmau->ldp = ldp;
	vmau->temp.mau_node = NULL;
	vmau->temp.cwq_node = NULL;
	vmau->flag_isconfigd = false;

	CHAIN_ADD(ldp->bound.vmaus, vmau, ldom);
	CHAIN_ADD(mau->bound.vmaus, vmau, pmau);

	/*
	 * Now, link all the bound vcpus in the chain which share
	 * a vmau (=> physical mau) in a guest.
	 */
	mau_link_vcpus(ldp, vmau);

	return (true);
}

/*
 * Check whether a MAU is accessible via a VCPU bound to a guest.
 */
bool_t
mau_accessible_to_ldom(ldom_t *ldp, pri_mau_t *mau)
{
	vcpu_t *vcpu;

	/*
	 * For each bound vmau, traverse through the cpus bound
	 * to it and check if the id matches up.
	 */
	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		if (vcpu->cpu->maup == mau)
			return (true);
	}

	return (false);
}

/*
 * Given a vid of a cpu bound to a guest, check whether
 * the cpu is linked to a vmau in this guest.
 */
vmau_t *
mau_bound_to_cpu(ldom_t *ldp, uint32_t cpuvid)
{
	vmau_t *vmau;
	vcpu_t *vcpu;

	/*
	 * For each bound vmau, traverse through the cpus bound
	 * to it and check if the id matches up.
	 */
	CHAIN_ITER(ldp->bound.vmaus, vmau, ldom) {
		CHAIN_ITER(vmau->vcpus, vcpu, mau) {
			if (cpuvid == vcpu->vid)
				return (vmau);
		}
	}

	return (NULL);
}

/*ARGSUSED3*/
bool_t
mau_gen_hvmd_nodes(ldom_t *ldp, md_t *mdp, md_node_t *mausp, md_node_t *guestp)
{
	vmau_t *vmau;
	md_node_t *cwqs;

	if (pri.have_cwq) {
		cwqs = md_find_node(mdp, "cwqs");
		ASSERT(cwqs != NULL);
	}

	CHAIN_ITER(ldp->bound.vmaus, vmau, ldom) {
		pri_mau_t *mau = vmau->mau;

		vmau->temp.mau_node = md_link_new_node(mdp, "mau", mausp,
		    "fwd", "back");
		md_add_value_property(mdp, vmau->temp.mau_node,
		    "ino", mau->mau_ino);
		md_add_value_property(mdp, vmau->temp.mau_node,
		    "resource_id", mau->mau_id);

		if (pri.have_cwq) {
			vmau->temp.cwq_node = md_link_new_node(mdp, "cwq", cwqs,
			    "fwd", "back");
			md_add_value_property(mdp, vmau->temp.cwq_node,
			    "ino", mau->cwq_ino);
			md_add_value_property(mdp, vmau->temp.cwq_node,
			    "resource_id", mau->mau_id);
		}
	}

	return (true);
}

/*
 * Returns the LDom a MAU is bound to or NULL if the MAU is not bound.
 */
ldom_t *
get_mau_ldom(pri_mau_t *mau)
{
	vmau_t *vmau = CHAIN_START(mau->bound.vmaus);

	return (vmau == NULL ? NULL : vmau->ldp);
}

static bool_t
get_mau_ino(pri_mau_t *mau, response_t *resp)
{
	mau->mau_ino = pri_alloc_mau_ino();
	if (mau->mau_ino == 0) {
		rprintf(resp, gettext("Bind failed: No interrupt number "
		    "available for MAU\n"));
		return (false);
	}

	if (pri.have_cwq) {
		mau->cwq_ino = pri_alloc_cwq_ino();
		if (mau->cwq_ino == 0) {
			fatal("No ino available for CWQ");
		}
	}

	return (true);
}

static bool_t
bind_num_vmaus(ldom_t *ldp, uint32_t num_to_bind, response_t *resp)
{
	pri_mau_t *mau;
	vcpu_t *vcpu;
	int num_avail;

	if (num_to_bind == 0)
		return (true);

	/*
	 * To find whether there is a free crypto units, we only need to go
	 * through the strands bound to the guest, and check the maus
	 * available to the strands.  Since multiple strands share a mau, we
	 * ensure we count only once. There is a presumption of a whole
	 * physical MAU for each virtual MAU.
	 */
	clear_visited_flags();
	num_avail = 0;
	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		/* Skip power managed vcpus */
		if (get_core_md_cpu_status(vcpu->cpu->phys_id,
		    CORE_MD_PM_STATUS) == PM_MANAGED)
			continue;
		/* the mau accessible to the bound vcpu/pcpu */
		mau = vcpu->cpu->maup;
		if (mau == NULL)
			continue;
		if ((!mau->flag_visited) && (mau_get_free_amt(mau) > 0)) {
			mau->flag_visited = true;
			/* check if MAU/CWQ has an ino and allocate if not */
			if (mau->mau_ino == 0) {
				if (!get_mau_ino(mau, resp))
					return (false);
			}

			if (++num_avail >= num_to_bind)
				break;
		}
	}

	if (num_avail == 0) {
		rprintf(resp, gettext("No available crypto unit resources "
		    "accessible by LDom %s\n"), ldp->namep);
		return (false);
	}

	if (num_avail < num_to_bind) {
		/*
		 * TRANSATION_NOTE
		 * Both %d's are the amount of physical crypto unit
		 * resources available
		 */
		rprintf(resp, gettext(
		    "Only %d physical crypto unit resource(s) available "
		    "to bind to LDom %s,\nproceeding with binding "
		    "%d additional crypto unit(s)\n"), num_avail, ldp->namep,
		    num_avail);

		num_to_bind = num_avail;
	}

	/*
	 * We now know there are sufficient resources, so we
	 * walk the list again, this time binding as we go.
	 */
	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		/* the mau accessible to the bound vcpu/pcpu */
		mau = vcpu->cpu->maup;
		if (mau == NULL)
			continue;
		if ((mau->flag_visited) && (mau_get_free_amt(mau) > 0)) {
			mau->flag_visited = false;
			if (!mau_bind(ldp, mau, NEW_VMAUID))
				fatal("Bind failed after finding available "
				    "crypto unit to bind");
			if (--num_to_bind == 0)
				break;
		} else {
			continue;
		}
	}

	if (num_to_bind != 0)
		fatal("%s failed after finding sufficient crypto units to "
		    "bind", __func__);

	return (true);
}

static bool_t
bind_explict_vmaus(ldom_t *ldp, mau_adm_t *mau_adm, response_t *resp)
{
	pri_mau_t *mau;
	int i, id;

	for (i = 0; i < mau_adm->num; i++) {
		id = mau_adm->ids[i];
		mau = id_to_mau(id);
		if (mau == NULL) {
			rprintf(resp, gettext("No physical crypto unit ID "
			    "%d\n"), id);
			return (false);
		}

		if (mau_get_free_amt(mau) == 0) {
			rprintf(resp,
			    gettext("Physical crypto unit ID %d is not "
			    "free\n"), id);
			return (false);
		}

		if (!mau_accessible_to_ldom(ldp, mau)) {
			rprintf(resp, gettext("Physical crypto unit ID %d "
			    "is not accessible from the bound VCPUs\n"), id);
			return (false);
		}

		/* check if MAU/CWQ has an ino and allocate if not */
		if (mau->mau_ino == 0) {
			if (!get_mau_ino(mau, resp))
				return (false);
		}
	}

	for (i = 0; i < mau_adm->num; i++) {
		id = mau_adm->ids[i];
		mau = id_to_mau(id);
		ASSERT(mau != NULL);
		if (!mau_bind(ldp, mau, NEW_VMAUID)) {
			fatal("Bind failed after finding available crypto "
			    "unit to bind");
		}
	}

	return (true);
}

/*
 * Bind crypto units for an ldom. For now we assume that cpus must be bound
 * ahead of the crypto units. When binding crypto units, all we need to do is
 * to go through the cpus (strands) bound to the guest, and assign a free
 * crypto unit available to the strands.
 */
bool_t
cons_bind_vmaus(ldom_t *ldp, mau_adm_t *mau_adm, response_t *resp)
{
	/*
	 * The CPUs must be bound prior to binding the crypto units.
	 */
	if (ldom_num_bound_vcpus(ldp) == 0)
		fatal("No VCPU resource info associated with LDom %s",
		    ldp->namep);

	if (mau_adm->type == MAU_ADM_NUM)
		return (bind_num_vmaus(ldp, mau_adm->num, resp));
	else
		return (bind_explict_vmaus(ldp, mau_adm, resp));
}

void
mau_unbind_all(ldom_t *ldp)
{
	vmau_t *vmau, *next;

	for (vmau = CHAIN_START(ldp->bound.vmaus); vmau != NULL; vmau = next) {
		next = CHAIN_NEXT(vmau->ldom);
		mau_unbind(vmau);
	}
}

/*
 * Function called by the migration code which shrinks a domain down to
 * have no crypto so that it can be suspended.
 */
bool_t
mau_mig_del(ldom_t *ldp, response_t *resp)
{
	int amt = ldom_num_bound_maus(ldp);
	vmau_t	**vmaus;
	mau_adm_t mau_adm = { .type = MAU_ADM_NUM, .num = amt };
	int	num = amt;

	if (num == 0)
		return (true);

	DBG(zprintf("%s: request to remove %d crypto unit%s\n", __func__,
	    num, (num <= 1) ? "" : "s"));

	vmaus = Xmalloc(num * (int)sizeof (vmau_t *));
	if (!get_maus_to_delete(ldp, &mau_adm, resp, vmaus)) {
		Xfree(vmaus);
		return (false);
	}

	if (dr_rm_cryptos(ldp, vmaus, &num, resp, true) > 0) {
		/* Failed to remove all */
		if (num < amt) {
			DBG(zprintf("Didn't succeed in removing all the "
			    "requested %d crypto units\n", amt));
		}
	}
	Xfree(vmaus);

	return (num < amt ? false : true);
}

/*
 * Attempt to add the maus which were removed from the source LDom back
 * in the target LDom after the migration completed. The target may not be
 * capable of adding the maus back to the LDom, hence failure to add
 * the maus back is not considered a failure of migration. The same routine is
 * also used to add back the maus in case migration failed.
 */
bool_t
mau_mig_add(ldom_t *ldp, int num_req, response_t *resp)
{
	int	ret;
	mau_adm_t mau_adm = { .type = MAU_ADM_NUM, .num = num_req };

	if (num_req == 0)
		return (true);

	if (check_crypto_dr_capable(ldp) != 0) {
		DBG(zprintf("%s: Crypto DR domain service not available\n",
		    __func__));
		return (false);
	}

	DBG(zprintf("%s: need to increase bound cryptos from %d to %d\n",
	    __func__, ldom_num_bound_maus(ldp), num_req));

	/*
	 * First, bind the required number of maus to the domain
	 */
	if (cons_bind_vmaus(ldp, &mau_adm, resp) == false) {
		return (false);
	}

	/* add the MAUs to the domain */
	ret = dr_add_cryptos(ldp, num_req, resp);

	/*
	 * dr_add_cryptos() would set the ldp->required.num_vmaus to
	 * the number actually bound to the domain
	 */

	if (ret == CRYPTO_DR_OP_FAILED)
		return (false);

	return (ret < num_req ? false : true);
}
