/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)mdesc_genbin.c	1.4	09/05/20 SMI"

#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#include <basics.h>
#include <allocate.h>
#include <fatal.h>
#include <mdesc.h>
#include <mdesc_impl.h>

#define	ROUNDUP(_a, _n)   (((_a) + ((_n) - 1)) & ~((_n) - 1))

void create_mde(md_element_t *mdep, int type, md_node_t *nodep,
    md_prop_t *propp);

int
md_gen_bin(md_t *mdp, uint8_t **bufvalp)
{
	uint32_t offset;
	md_node_t *nodep;
	md_data_block_t *mdbp;
	md_string_t *msp;
	md_header_t *mdhp;
	md_element_t *mdep;
	uint32_t strings_size;
	uint32_t data_block_size;
	int total_size;
	uint8_t *bufferp;
	uint8_t *string_bufferp;
	uint8_t *data_block_bufferp;

	mutex_lock(&mdp->md_lock);
	/*
	 * Skip through strings to compute offsets.
	 */
	offset = 0;
	for (msp = CHAIN_START(mdp->string_list); msp != NULL;
	    msp = CHAIN_NEXT(msp->link)) {
		msp->build_offset = offset;
		offset += msp->size;
	}
	strings_size = ROUNDUP(offset, MD_ALIGNMENT_SIZE);

	/*
	 * Skip through data blocks to compute offsets.
	 */

	offset = 0;
	for (mdbp = CHAIN_START(mdp->data_block_list); mdbp != NULL;
	    mdbp = CHAIN_NEXT(mdbp->link)) {
		mdbp->build_offset = offset;
		offset += mdbp->size;
		offset = ROUNDUP(offset, MD_ALIGNMENT_SIZE);
	}
	data_block_size = ROUNDUP(offset, MD_ALIGNMENT_SIZE);

	/*
	 * Compute the MD elements required to build the element list.
	 * For each node there is a node start and end, and one
	 * element for each property.
	 */

	offset = 0;
	for (nodep = CHAIN_START(mdp->node_list); nodep != NULL;
	    nodep = CHAIN_NEXT(nodep->link)) {
		nodep->build_index = offset;
		offset += 2+CHAIN_LENGTH(nodep->prop_list);
		nodep->next_index = offset;
	}
	offset += 1;	/* add the LIST_END element */

	total_size = MD_HEADER_SIZE + offset * MD_ELEMENT_SIZE +
	    strings_size + data_block_size;

	/*
	 * Allocate output buffer.
	 */

	bufferp = Xcalloc(total_size, uint8_t);
	mdhp = (md_header_t *)bufferp;

	string_bufferp = bufferp + MD_HEADER_SIZE + offset * MD_ELEMENT_SIZE;
	data_block_bufferp = string_bufferp + strings_size;

	mdhp->transport_version = htomd32(MD_TRANSPORT_VERSION);
	mdhp->node_blk_sz = htomd32(offset * MD_ELEMENT_SIZE);
	mdhp->name_blk_sz = htomd32(strings_size);
	mdhp->data_blk_sz = htomd32(data_block_size);

	/*
	 * Build the element list.
	 * For each node there is a node start and end, and one
	 * element for each property.
	 */

	offset = 0;
	mdep = (md_element_t *)(bufferp + MD_HEADER_SIZE);
	for (nodep = CHAIN_START(mdp->node_list); nodep != NULL;
	    nodep = CHAIN_NEXT(nodep->link)) {
		md_prop_t *propp;

		create_mde(mdep, MDET_NODE, nodep, NULL);
		mdep++;

		for (propp = CHAIN_START(nodep->prop_list); propp != NULL;
		    propp = CHAIN_NEXT(propp->link)) {
			create_mde(mdep, propp->type, nodep, propp);
			mdep++;
		}

		create_mde(mdep, MDET_NODE_END, NULL, NULL);
		mdep++;
	}

	create_mde(mdep, MDET_LIST_END, NULL, NULL);
	mdep++;

	/*
	 * Quick sanity check.
	 */

	ASSERT(((uint8_t *)mdep) == ((uint8_t *)string_bufferp));

	/*
	 * Skip through strings and stash them..
	 */

	offset = 0;
	for (msp = CHAIN_START(mdp->string_list); msp != NULL;
	    msp = CHAIN_NEXT(msp->link)) {
		memcpy(string_bufferp + msp->build_offset, msp->strp,
		    msp->size);
	}

	/*
	 * Skip through data blocks and stash them.
	 */

	offset = 0;
	for (mdbp = CHAIN_START(mdp->data_block_list); mdbp != NULL;
	    mdbp = CHAIN_NEXT(mdbp->link)) {
		memcpy(data_block_bufferp + mdbp->build_offset,
		    mdbp->datap, mdbp->size);
	}

	*bufvalp = bufferp;
	mutex_unlock(&mdp->md_lock);
	return (total_size);
}

void
create_mde(md_element_t *mdep, int type, md_node_t *nodep, md_prop_t *propp)
{
	memset(mdep, 0, MD_ELEMENT_SIZE);
	mdep->tag = htomd8(type);

	switch (type) {
	case MDET_NODE:
		mdep->d.prop_idx = htomd32(nodep->next_index);
		mdep->name_len = htomd8(nodep->typep->size -1);
		mdep->name_offset = htomd32(nodep->typep->build_offset);
		break;

	case MDET_PROP_ARC:
		ASSERT(propp->d.arc.is_ptr);
		mdep->d.prop_idx = htomd64(propp->d.arc.val.nodep->build_index);
		goto fix_name;

	case MDET_PROP_VAL:
		mdep->d.prop_val = htomd64(propp->d.value);
		goto fix_name;

	case MDET_PROP_STR:
	case MDET_PROP_DAT:
		mdep->d.prop_data.offset = htomd32(propp->d.dbp->build_offset);
		mdep->d.prop_data.len = htomd32(propp->d.dbp->size);
		goto fix_name;

fix_name:
		mdep->name_len = htomd8(propp->sp->size -1);
		mdep->name_offset = htomd32(propp->sp->build_offset);
		break;

	case MDET_NULL:
	case MDET_NODE_END:
	case MDET_LIST_END:
		break;

	default:
		fatal("Internal error: illegal element type");
	}
}
