/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)mdesc_lib.c	1.17	09/05/20 SMI"

#include <strings.h>
#include <stdio.h>
#include <unistd.h>
#include <alloca.h>

#include "basics.h"
#include "allocate.h"
#include "fatal.h"

#include "mdesc.h"
#include "mdesc_impl.h"

static uint32_t
md_byte_hash(uint8_t *bp, int len)
{
	uint32_t hash = 0;
	int i;

	for (i = 0; i < len; i++) {
		/* 5 bit rotation */
		hash = (hash>>27) | (hash<<5) | bp[i];
	}

	return (hash);
}

md_string_t *
md_find_string(md_t *mdp, char *strp, uint32_t *hashp)
{
	md_string_t *msp;
	uint32_t hash;

	hash = md_byte_hash((uint8_t *)strp, strlen(strp));

	if (hashp != NULL)
		*hashp = hash;

	CHAIN_ITER(mdp->string_list, msp, link) {
		if (msp->hash == hash && strcmp(msp->strp, strp) == 0)
			return (msp);
	}

	return (NULL);
}

md_string_t *
md_find_name(md_t *mdp, char *strp)
{
	return (md_find_string(mdp, strp, NULL));
}

static md_string_t *
md_new_string(md_t *mdp, char *strp)
{
	md_string_t *msp;
	uint32_t hash;

	msp = md_find_string(mdp, strp, &hash);
	if (msp == NULL) {
		int space;

		msp = Xcalloc(1, md_string_t);
		space = strlen(strp);
		msp->strp = Xstrdup(strp);
		msp->size = space + 1;
		msp->hash = hash;
		msp->ref_cnt = 0;
		msp->build_offset = -1;
		CHAIN_ADD(mdp->string_list, msp, link);
	}
	msp->ref_cnt++;
	return (msp);
}

static md_data_block_t *
md_find_data_block(md_t *mdp, uint8_t *datap, int len, uint32_t *hashp)
{
	md_data_block_t *dbp;
	uint32_t hash;

	hash = md_byte_hash(datap, len);

	if (hashp != NULL)
		*hashp = hash;

	CHAIN_ITER(mdp->data_block_list, dbp, link) {
		if (dbp->size == len &&
		    dbp->hash == hash && bcmp(dbp->datap, datap, len) == 0)
			return (dbp);
	}

	return (NULL);
}

md_data_block_t *
md_new_data_block(md_t *mdp, uint8_t *bufp, int len)
{
	md_data_block_t *dbp;
	uint32_t	hash;

	dbp = md_find_data_block(mdp, bufp, len, &hash);
	if (dbp == NULL) {
		dbp = Xcalloc(1, md_data_block_t);
		dbp->datap = Xmalloc(len);
		memcpy(dbp->datap, bufp, len);
		dbp->size = len;
		dbp->hash = hash;
		dbp->ref_cnt = 0;
		dbp->build_offset = -1;
		CHAIN_ADD(mdp->data_block_list, dbp, link);
	}
	dbp->ref_cnt++;

	return (dbp);
}

md_node_t *
md_find_node_index(md_t *mdp, int index)
{
	md_node_t *nodep;

	for (nodep = CHAIN_START(mdp->node_list);
	    nodep != NULL; nodep = CHAIN_NEXT(nodep->link)) {

		if (!nodep->deleted && (nodep->build_index == index))
			return (nodep);
	}

	return (NULL);
}

md_node_t *
md_new_node(md_t *mdp, char *sp)
{
	md_node_t *nodep;

	nodep = Xcalloc(1, md_node_t);
	nodep->typep = md_new_string(mdp, sp);
	CHAIN_INIT(nodep->prop_list);
	CHAIN_ADD(mdp->node_list, nodep, link);

	return (nodep);
}

md_prop_t *
md_new_property(md_t *mdp, md_node_t *nodep, uint8_t type, char *sp)
{
	md_prop_t *propp;

	propp = Xcalloc(1, md_prop_t);
	propp->type = type;
	propp->sp = md_new_string(mdp, sp);

	CHAIN_ADD(nodep->prop_list, propp, link);

	return (propp);
}

void
md_add_value_property(md_t *mdp, md_node_t *nodep, char *sp, uint64_t value)
{
	md_prop_t *propp;

	propp = md_new_property(mdp, nodep, MDET_PROP_VAL, sp);
	propp->d.value = value;
}

void
md_add_string_property(md_t *mdp, md_node_t *nodep, char *sp, char *bufp)
{
	md_prop_t *propp;

	propp = md_new_property(mdp, nodep, MDET_PROP_STR, sp);
	propp->d.dbp = md_new_data_block(mdp, (uint8_t *)bufp, strlen(bufp)+1);
}

void
md_add_data_property(md_t *mdp, md_node_t *nodep, char *sp, int len,
    uint8_t *bufp)
{
	md_prop_t *propp;

	propp = md_new_property(mdp, nodep, MDET_PROP_DAT, sp);
	propp->d.dbp = md_new_data_block(mdp, bufp, len);
}

void
md_add_string_as_data_property(md_t *mdp, md_node_t *nodep, char *sp,
    char *bufp)
{
	md_add_data_property(mdp, nodep, sp, strlen(bufp) + 1, (uint8_t *)bufp);
}

void
md_add_arc_property(md_t *mdp, md_node_t *nodep, char *arcnamep,
    md_node_t *tgtnodep)
{
	md_prop_t *propp;

	propp = md_new_property(mdp, nodep, MDET_PROP_ARC, arcnamep);
	propp->d.arc.is_ptr = true;
	propp->d.arc.val.nodep = tgtnodep;
}

/*
 * Helper function: md_link_new_node
 * Frequent operation is to create a node, and link it to a
 * parent with an arc of some name ...
 * .. the back link is optional, as not usually used for hypervisor MDs
 */
md_node_t *
md_link_new_node(md_t *mdp, char *nodenamep, md_node_t *parentnodep,
    char *linktonewp, char *linkbackp)
{
	md_node_t *nodep;

	nodep = md_new_node(mdp, nodenamep);

	ASSERT(linktonewp != NULL);
	ASSERT(parentnodep != NULL && !parentnodep->deleted);

	md_add_arc_property(mdp, parentnodep, linktonewp, nodep);

	if (linkbackp != NULL) {
		md_add_arc_property(mdp, nodep, linkbackp, parentnodep);
	}

	return (nodep);
}

/*
 * Cleanup the MD data structures.
 */
int
md_fini(md_t *mdp)
{
	md_node_t *nodep;

	for (nodep = CHAIN_START(mdp->node_list); nodep != NULL; ) {
		md_node_t *tmp_nodep;

		tmp_nodep = CHAIN_NEXT(nodep->link);
		md_free_node(mdp, nodep);

		nodep = tmp_nodep;
	}

	mutex_destroy(&mdp->md_lock);
	/* should have deleted all the string refs by here */
	ASSERT(CHAIN_LENGTH(mdp->string_list) == 0);
	Xfree(mdp);

	return (0);
}

void
md_free_node(md_t *mdp, md_node_t *nodep)
{
	md_prop_t *propp;

	md_free_string(mdp, nodep->typep);

	for (propp = CHAIN_START(nodep->prop_list); propp != NULL; ) {
		md_prop_t *tmp_propp;

		tmp_propp = CHAIN_NEXT(propp->link);
		md_free_prop(mdp, propp);

		propp = tmp_propp;
	}

	Xfree(nodep);
}

void
md_free_prop(md_t *mdp, md_prop_t *propp)
{
	md_free_string(mdp, propp->sp);

	switch (propp->type) {
	case MDET_PROP_VAL:
		break;

	case MDET_PROP_ARC:
		break;

	case MDET_PROP_STR:
	case MDET_PROP_DAT:
		md_free_data_block(mdp, propp->d.dbp);
		break;

	default:
		fatal("Internal error - unexpected property type");
	}

	Xfree(propp);
}

void
md_free_string(md_t *mdp, md_string_t *msp)
{
	if (msp->ref_cnt <= 0)
		fatal("Internal error: freed string with zero ref count !");

	msp->ref_cnt --;

	if (msp->ref_cnt == 0) {
		Xfree(msp->strp);
		CHAIN_UNLINK(mdp->string_list, msp, link);
		Xfree(msp);
	}
}

void
md_free_data_block(md_t *mdp, md_data_block_t *mdbp)
{
	if (mdbp->ref_cnt <= 0)
		fatal("Internal error: freed data block with zero ref count !");

	mdbp->ref_cnt --;

	if (mdbp->ref_cnt == 0) {
		Xfree(mdbp->datap);
		CHAIN_UNLINK(mdp->data_block_list, mdbp, link);
		Xfree(mdbp);
	}
}

/*
 * Functions equivalent to the Solaris ones, but
 * operating on the internal format not the MD format.
 */

/*
 * Will accept NULL as a name, and match nothing
 */
md_prop_t *
md_find_node_prop(md_t *mdp, md_node_t *nodep,
    md_string_t *prop_namep, int type)
{
	md_prop_t *propp;

	for (propp = CHAIN_START(nodep->prop_list);
	    propp != NULL && !(propp->sp == prop_namep && propp->type == type);
	    propp = CHAIN_NEXT(propp->link)) /* LINTED */
		;

	return (propp);
}

int
md_get_prop_data(md_t *mdp, md_node_t *nodep,
    char *prop_namep, uint8_t **datapp, int *lenp)
{
	md_prop_t *propp;

	propp = md_find_node_prop(mdp, nodep,
	    md_find_name(mdp, prop_namep), MDET_PROP_DAT);

	if (propp == NULL)
		return (-1);	/* no such name */

	*lenp = propp->d.dbp->size;
	*datapp = propp->d.dbp->datap;

	return (0);
}

int
md_get_prop_str(md_t *mdp, md_node_t *nodep, char *prop_namep, char **strpp)
{
	md_prop_t *propp;

	propp = md_find_node_prop(mdp, nodep,
	    md_find_name(mdp, prop_namep), MDET_PROP_STR);

	if (propp == NULL)
		return (-1);	/* no such name */

	*strpp = (char *)propp->d.dbp->datap;

	return (0);
}

int
md_get_prop_val(md_t *mdp, md_node_t *nodep, char *prop_namep, uint64_t *valp)
{
	md_prop_t *propp;

	propp = md_find_node_prop(mdp, nodep,
	    md_find_name(mdp, prop_namep), MDET_PROP_VAL);

	if (propp == NULL)
		return (-1);	/* no such name */

	*valp = propp->d.value;

	return (0);
}

void
md_update_value_property(md_t *mdp, md_node_t *nodep,
    char *prop_namep, uint64_t val)
{
	md_prop_t *propp;

	propp = md_find_node_prop(mdp, nodep,
	    md_find_name(mdp, prop_namep), MDET_PROP_VAL);

	if (propp == NULL) {
		propp = md_new_property(mdp, nodep, MDET_PROP_VAL, prop_namep);
	}
	propp->d.value = val;
}

void
md_update_string_property(md_t *mdp, md_node_t *nodep,
    char *prop_namep, char *bufp)
{
	md_prop_t *propp;

	propp = md_find_node_prop(mdp, nodep,
	    md_find_name(mdp, prop_namep), MDET_PROP_STR);

	if (propp != NULL)
		md_delete_prop(mdp, nodep, propp);

	md_add_string_property(mdp, nodep, prop_namep, bufp);
}

void
md_update_data_property(md_t *mdp, md_node_t *nodep,
    char *prop_namep, int len, uint8_t *bufp)
{
	md_prop_t *propp;

	propp = md_find_node_prop(mdp, nodep,
	    md_find_name(mdp, prop_namep), MDET_PROP_DAT);

	if (propp != NULL)
		md_delete_prop(mdp, nodep, propp);

	md_add_data_property(mdp, nodep, prop_namep, len, bufp);
}

int
md_get_prop_arc(md_t *mdp, md_node_t *nodep, char *prop_namep,
    md_node_t **nodepp)
{
	md_prop_t *propp;

	propp = md_find_node_prop(mdp, nodep,
	    md_find_name(mdp, prop_namep), MDET_PROP_ARC);

	if (propp == NULL)
		return (-1);	/* no such name */

	ASSERT(propp->d.arc.is_ptr);
	*nodepp = propp->d.arc.val.nodep;

	return (0);
}

int
md_node_count(md_t *mdp)
{
	return (CHAIN_LENGTH(mdp->node_list));
}

md_node_t *
md_root_node(md_t *mdp)
{
	md_string_t *strp;
	md_node_t *snp;

	strp = md_find_name(mdp, "root");

	for (snp = CHAIN_START(mdp->node_list);
	    snp != NULL && snp->typep != strp;
	    snp = CHAIN_NEXT(snp->link)) /* LINTED */
		;

	return (snp);
}

static void mdl_scan_dag(md_t *mdp, md_node_t *startnodep,
    md_string_t *node_namep, md_string_t *arc_namep, int *idxp,
    md_node_t **stashpp, int level);

int
md_scan_dag(md_t *mdp, md_node_t *startnodep, md_string_t *node_namep,
    md_string_t *arc_namep, md_node_t **stashpp)
{
	int idx;
	md_node_t *nodep;

	/*
	 * Possible the caller was lazy and didn't check the
	 * validity of either the node name or the arc name
	 * on calling ... in which case fail to find any
	 * nodes ;-)
	 * This is distinct, from a fail (-1) since we return
	 * that nothing was found.
	 */

	if (node_namep == NULL || arc_namep == NULL)
		return (0);

	/*
	 * If we want to start at the top, start at index 0.
	 */

	if (startnodep == NULL) {
		startnodep = md_root_node(mdp);
		if (startnodep == NULL)
			return (0);
	}

	/*
	 * Clean up the seen flags.
	 */
	for (nodep = CHAIN_START(mdp->node_list);
	    nodep != NULL; nodep = CHAIN_NEXT(nodep->link))
		nodep->seen = 0;

	/*
	 * Now build the list of requested nodes.
	 */
	idx = 0;
	mdl_scan_dag(mdp, startnodep, node_namep, arc_namep, &idx, stashpp, 0);

	return (idx);
}

static void
mdl_scan_dag(md_t *mdp, md_node_t *nodep, md_string_t *node_namep,
    md_string_t *arc_namep, int *idxp, md_node_t **stashpp, int level)
{
	md_prop_t *propp;

	/* have we been here before or is this a marked as deleted node ? */
	if (nodep->seen || nodep->deleted)
		return;

	nodep->seen = 1;

	/* is this node of the type we seek ? */
	if (nodep->typep == node_namep) {
		/* record the node in the list and keep searching */
		if (stashpp != NULL) {
			stashpp[*idxp] = nodep;
		}
		(*idxp)++;
	}

	/*
	 * Simply walk the elements in the node.
	 * if we find a matching arc, then recursively call
	 * the subordinate looking for a match
	 */

	for (propp = CHAIN_START(nodep->prop_list); propp != NULL;
	    propp = CHAIN_NEXT(propp->link)) {
		if (propp->type == MDET_PROP_ARC && propp->sp == arc_namep) {
			ASSERT(propp->d.arc.is_ptr);
			mdl_scan_dag(mdp, propp->d.arc.val.nodep,
			    node_namep, arc_namep, idxp, stashpp, level+1);
		}
	}
}

int
md_scan_nodes(md_t *mdp, md_string_t *node_namep, md_node_t **stashpp)
{
	int idx = 0;
	md_node_t *nodep;

	/*
	 * Possible the caller was lazy and didn't check the validity of the
	 * node name on calling, in which case fail to find any nodes.
	 */
	if (node_namep == NULL)
		return (0);

	for (nodep = CHAIN_START(mdp->node_list);
	    nodep != NULL; nodep = CHAIN_NEXT(nodep->link)) {

		if (!nodep->deleted && (nodep->typep == node_namep)) {
			/* record the node in the list and keep searching */
			if (stashpp != NULL)
				stashpp[idx] = nodep;
			idx++;
		}
	}

	return (idx);
}

/*
 * Delete a node, then all arc references to it.
 */
void
md_delete_node(md_t *mdp, md_node_t *nodep)
{
	md_node_t *snp;

	CHAIN_UNLINK(mdp->node_list, nodep, link);

	for (snp = CHAIN_START(mdp->node_list);
	    snp != NULL;
	    snp = CHAIN_NEXT(snp->link)) {
		md_prop_t *propp;
		md_prop_t *next_propp;

		for (propp = CHAIN_START(snp->prop_list);
		    propp != NULL; propp = next_propp) {

			next_propp = CHAIN_NEXT(propp->link);

			if (MDET_PROP_ARC == propp->type) {
				ASSERT(propp->d.arc.is_ptr);
				if (propp->d.arc.val.nodep == nodep) {
					md_delete_prop(mdp, snp, propp);
				}
			}
		}
	}

	md_free_node(mdp, nodep);
}

void
md_mark_node_deleted(md_node_t *nodep)
{
	nodep->deleted = 1;
}

void
md_delete_prop(md_t *mdp, md_node_t *nodep, md_prop_t *propp)
{
	CHAIN_UNLINK(nodep->prop_list, propp, link);
	md_free_prop(mdp, propp);
}

md_node_t *
md_find_subnode(md_t *md, md_node_t *node, char *node_name)
{
	md_node_t **nodes;
	md_node_t *return_node;
	int num;

	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(md));
	num = md_scan_dag(md, node, md_find_name(md, node_name),
	    md_find_name(md, "fwd"), nodes);

	if (num > 0)
		return_node = nodes[0];
	else
		return_node = NULL;

	Xfree(nodes);

	return (return_node);
}

md_node_t *
md_find_node(md_t *md, char *node_name)
{
	return (md_find_subnode(md, NULL, node_name));
}

bool_t
md_find_delete_node(md_t *md, char *node_name)
{
	md_node_t *node;

	if ((node = md_find_node(md, node_name)) == NULL) {
		return (false);
	} else {
		md_delete_node(md, node);
		return (true);
	}
}

/*
 * Very basic scour function. Delete nodes that are
 * not referenced by any arcs.
 */
void
md_scour(md_t *mdp, md_string_t *exclude_arc)
{
	md_node_t *node, *next_node;
	CHAIN(md_node_t, referenced);

	CHAIN_INIT(referenced);

	for (node = CHAIN_START(mdp->node_list); node != NULL;
	    node = CHAIN_NEXT(node->link)) {
		node->seen = 0;
	}

	/*
	 * Mark all nodes referenced from root.
	 */
	node = md_root_node(mdp);
	ASSERT(node != NULL && !node->deleted);
	CHAIN_ADD(referenced, node, ref);
	node->seen = 1;

	do {
		md_prop_t *prop;

		for (prop = CHAIN_START(node->prop_list); prop != NULL;
		    prop = CHAIN_NEXT(prop->link)) {
			if (MDET_PROP_ARC == prop->type &&
			    prop->sp != exclude_arc &&
			    !prop->d.arc.val.nodep->seen &&
			    !prop->d.arc.val.nodep->deleted) {
				CHAIN_ADD(referenced, prop->d.arc.val.nodep,
				    ref);
				prop->d.arc.val.nodep->seen = 1;
			}
		}

		node = CHAIN_NEXT(node->ref);
	} while (node != NULL);

	/*
	 * Delete all arcs from referenced nodes to unreferenced nodes.
	 */
	for (node = CHAIN_START(referenced); node != NULL;
	    node = CHAIN_NEXT(node->ref)) {
		md_prop_t *prop;
		md_prop_t *next_prop;

		for (prop = CHAIN_START(node->prop_list); prop != NULL;
		    prop = next_prop) {

			next_prop = CHAIN_NEXT(prop->link);

			if (MDET_PROP_ARC == prop->type &&
			    prop->d.arc.val.nodep->seen == 0)
				md_delete_prop(mdp, node, prop);
		}
	}

	/*
	 * Free all unreferenced nodes.
	 */
	for (node = CHAIN_START(mdp->node_list); node != NULL;
	    node = next_node) {
		next_node = CHAIN_NEXT(node->link);
		if (node->seen == 0) {
			CHAIN_UNLINK(mdp->node_list, node, link);
			md_free_node(mdp, node);
		}
	}
}

/*
 * Allocate and init an md.
 */
md_t *
md_new_md(void)
{
	md_t *mdp;

	mdp = Xcalloc(1, md_t);
	CHAIN_INIT(mdp->node_list);
	CHAIN_INIT(mdp->string_list);
	CHAIN_INIT(mdp->data_block_list);

	mutex_init(&mdp->md_lock, USYNC_THREAD, NULL);
	return (mdp);
}

/*
 * Copy all value, string and data properties, from MD node 'from', to MD node
 * 'to' in MD 'md'. Properties which already exist in node 'to' are
 * overwritten.
 */
void
md_copy_properties(md_t *md, md_node_t *to, md_node_t *from)
{
	md_prop_t *prop;
	char *name;

	for (prop = CHAIN_START(from->prop_list); prop != NULL;
	    prop = CHAIN_NEXT(prop->link)) {

		name = prop->sp->strp;

		switch (prop->type) {
		case MDET_PROP_VAL:
			md_update_value_property(md, to, name, prop->d.value);
			break;
		case MDET_PROP_STR:
			md_update_string_property(md, to, name,
			    (char *)(prop->d.dbp->datap));
			break;
		case MDET_PROP_DAT:
			md_update_data_property(md, to, name, prop->d.dbp->size,
			    prop->d.dbp->datap);
			break;
		default:
			break;
		}
	}
}

/*
 * Add one MD to another.
 */
void
md_add(md_t *mda, md_t *mdb)
{
	md_node_t	**nodes;
	md_node_t	*node, *newnode;
	md_prop_t	*prop;
	char		*name;
	int		idx;

	mutex_lock(&mda->md_lock);
	mutex_lock(&mdb->md_lock);
	/* clear build index for MD being added to */
	for (node = CHAIN_START(mda->node_list); node != NULL;
	    node = CHAIN_NEXT(node->link))
		node->build_index = -1;

	nodes = Xmalloc(md_node_count(mdb) * sizeof (md_node_t *));

	/* Walk the nodes and clone them */
	for (idx = 0, node = CHAIN_START(mdb->node_list); node != NULL;
	    idx++, node = CHAIN_NEXT(node->link)) {

		newnode = md_new_node(mda, node->typep->strp);
		newnode->deleted = node->deleted;
		newnode->build_index = idx;
		node->build_index = idx;
		nodes[idx] = newnode;
	}

	/*
	 * Now clone the properties, when we find an arc we can use the build
	 * index to find the target for the equivalent clone node.
	 */

	for (idx = 0, node = CHAIN_START(mdb->node_list); node != NULL;
	    idx++, node = CHAIN_NEXT(node->link)) {

		newnode = nodes[idx];

		for (prop = CHAIN_START(node->prop_list); prop != NULL;
		    prop = CHAIN_NEXT(prop->link)) {

			name = prop->sp->strp;

			switch (prop->type) {
			case MDET_PROP_VAL:
				md_add_value_property(mda, newnode, name,
				    prop->d.value);
				break;
			case MDET_PROP_STR:
				md_add_string_property(mda, newnode, name,
				    (char *)(prop->d.dbp->datap));
				break;
			case MDET_PROP_DAT:
				md_add_data_property(mda, newnode, name,
				    prop->d.dbp->size, prop->d.dbp->datap);
				break;
			case MDET_PROP_ARC:
				ASSERT(prop->d.arc.is_ptr);
				md_add_arc_property(mda, newnode, name,
				    nodes[prop->d.arc.val.nodep->build_index]);
				break;
			default:
				fatal("md_add: unknown property type found");
			}
		}
	}

	Xfree(nodes);
	mutex_unlock(&mdb->md_lock);
	mutex_unlock(&mda->md_lock);
}

/*
 * Clone an MD.
 */
md_t *
md_clone(md_t *origmdp)
{
	md_t		*mdp;

	/* new empty core md */
	mdp = md_new_md();

	/* add original md to the new md */
	md_add(mdp, origmdp);

	return (mdp);
}
