/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2008 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)mdesc_scanbin.c	1.9	08/01/03 SMI"

#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>

#include <basics.h>
#include <allocate.h>
#include <fatal.h>
#include <mdesc.h>
#include <mdesc_impl.h>
#include <fcntl.h>

/*
 * For sorting MD nodes by build_index.
 */
static int
node_cmp(const void *a, const void *b)
{
	md_node_t *node_a = *(md_node_t **)a;
	md_node_t *node_b = *(md_node_t **)b;

	if (node_a->build_index < node_b->build_index)
		return (-1);
	else if (node_a->build_index == node_b->build_index)
		return (0);
	else
		return (1);
}

/*
 * Scan in a binary machine description
 *
 * Gets handed a buffer for the a binary format MD, parses its contents, and
 * then builds and returns the internal format version.
 */
md_t *
md_scan_bin(uint8_t *binmdp, int size)
{
	md_t *mdp;
	md_header_t *mdhp;
	md_element_t *mep;
	uint8_t *bufferp;
	char *buffer_strp;
	uint8_t *buffer_datap;
	bool_t flag_in_node;
	bool_t done;
	int total_size;
	md_node_t **nodes;
	int maxnodes, numnodes;
	md_node_t key;
	md_node_t *keyp = &key;

	md_node_t *nodep;
	md_prop_t *propp;

	if (size < sizeof (md_header_t)) {
		warning("Unable to read machine description header");
		return (NULL);
	}

	mdhp = (md_header_t *)binmdp;

	/*
	 * Determine the buffer size required for the remainder.
	 */

	total_size = md_scan_bin_size(mdhp);
	if (total_size < 0)
		return (NULL);

	if (size < total_size) {
		warning("MD is larger than given buffer");
		return (NULL);
	}

	/*
	 * Array of pointers to MD nodes. This will be sorted by build_index
	 * to allow for efficient lookup of nodes by build_index when
	 * stitching together the MD arcs.
	 */
	maxnodes = mdtoh32(mdhp->node_blk_sz) / sizeof (md_element_t);
	nodes = Xmalloc(maxnodes * sizeof (md_node_t *));

	/*
	 * Initialise the md_t.
	 */

	mdp = md_new_md();

	/*
	 * Now just walk the elements filling in the
	 * local data structures as required.
	 */

	bufferp = binmdp + MD_HEADER_SIZE;
	buffer_strp = (char *)(bufferp + mdtoh32(mdhp->node_blk_sz));
	buffer_datap = ((uint8_t *)buffer_strp) + mdtoh32(mdhp->name_blk_sz);

	numnodes = 0;
	nodep = NULL;
	flag_in_node = false;
	done = false;

	for (mep = (md_element_t *)bufferp; !done; mep++) {
		char *namep;

		switch (MDE_TAG(mep)) {
		case MDET_NODE:
		case MDET_PROP_VAL:
		case MDET_PROP_STR:
		case MDET_PROP_DAT:
		case MDET_PROP_ARC:
			namep = buffer_strp + MDE_NAME(mep);
			break;
		default:
			namep = NULL;
		}

		switch (MDE_TAG(mep)) {
		case MDET_LIST_END:
			done = true;
			break;

		case MDET_NULL:
			break;

		case MDET_NODE:
			if (flag_in_node) {
				warning("mal-formed MD - node within a node");
				goto cleanup;
			}
			flag_in_node = true;

			nodep = md_new_node(mdp, namep);
			nodep->build_index = mep - ((md_element_t *)bufferp);

			nodes[numnodes++] = nodep;

			break;

		case MDET_NODE_END:
			if (!flag_in_node) {
				warning("Mal-formed MD - surplus node end");
				goto cleanup;
			}
			flag_in_node = false;
			nodep = NULL;
			break;

		case MDET_PROP_VAL:
			md_add_value_property(mdp, nodep, namep,
			    MDE_PROP_VALUE(mep));
			break;

		case MDET_PROP_STR:
			md_add_string_property(mdp, nodep, namep,
			    (char *)(buffer_datap + MDE_PROP_DATA_OFFSET(mep)));
			break;

		case MDET_PROP_DAT:
			md_add_data_property(mdp, nodep, namep,
			    MDE_PROP_DATA_LEN(mep),
			    buffer_datap + MDE_PROP_DATA_OFFSET(mep));
			break;

		case MDET_PROP_ARC:
			propp = md_new_property(mdp, nodep, MDET_PROP_ARC,
			    namep);
			propp->d.arc.is_ptr = false;
			propp->d.arc.val.index = MDE_PROP_INDEX(mep);
			break;
		default:
			warning("Unknown tag type");
			goto cleanup;
		}
	}

	/* Now stitch all the arcs */

	qsort(nodes, numnodes, sizeof (md_node_t *), node_cmp);

	for (nodep = CHAIN_START(mdp->node_list); nodep != NULL;
	    nodep = CHAIN_NEXT(nodep->link)) {
		for (propp = CHAIN_START(nodep->prop_list); propp != NULL;
		    propp = CHAIN_NEXT(propp->link)) {
			if ((propp->type == MDET_PROP_ARC) &&
			    !propp->d.arc.is_ptr) {
				md_node_t **tnp;
				keyp->build_index = propp->d.arc.val.index;
				tnp = bsearch(&keyp, nodes, numnodes,
				    sizeof (md_node_t *), node_cmp);
				if (tnp == NULL) {
					warning("Arc points to non-existent "
					    "node");
					goto cleanup;
				}
				propp->d.arc.is_ptr = true;
				propp->d.arc.val.index = -1;
				propp->d.arc.val.nodep = *tnp;
			}
		}
	}

	Xfree(nodes);
	return (mdp);

cleanup:
	Xfree(nodes);
	md_fini(mdp);
	return (NULL);
}

int
md_scan_bin_size(md_header_t *mdhp)
{
	if (mdtoh32(mdhp->transport_version) != MD_TRANSPORT_VERSION) {
		warning("Unrecognised MD transport format");
		return (-1);
	}

	return (MD_HEADER_SIZE +
	    mdtoh32(mdhp->node_blk_sz) +
	    mdtoh32(mdhp->name_blk_sz) +
	    mdtoh32(mdhp->data_blk_sz));
}
