/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Module implementing the 'mdstore' DS protocol
 */

#include <sys/types.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "allocate.h"
#include "backend.h"
#include "basics.h"
#include "database.h"
#include "ds.h"
#include "ds_registry.h"
#include "fatal.h"
#include "ldom.h"
#include "mdesc.h"
#include "mdesc_impl.h"
#include "mdstore.h"
#include "zpri.h"
#include "response.h"
#include "sequencer.h"
#include "zprintf.h"
#include "svcs.h"
#include "support.h"
#include "keys.h"
#include "variables.h"
#include "xml_cntrl_common.h"
#include "xml_v3_defs.h"
#include "database_internal.h"
#include "frags.h"
#include "autosave.h"

#define	MDSTORE_DS_TIMEOUT	(60)

static uint64_t mdstore_seq = 0;	/* request sequence number */

static void
setup_guest_md(md_t *md, bool_t ctrl_domain)
{
	md_update_value_property(md, md_root_node(md), "md-generation#", 1);
}

/*
 * Sends the 'ds' domain service request message in 'ds_req' with request
 * number 'reqnum' and receives the response. Returns the data portion of the
 * response, i.e. the domain services data handle followed by the payload,
 * and sets 'resplen' to its length. Returns NULL on error.
 */
static void *
send_recv(ds_svc_t *ds, ds_req_t *ds_req, uint64_t reqnum, uint32_t *resplen)
{
	ds_msg_t msg, resp;
	int rc;

	msg.buf = ds_req;
	msg.len = DS_SIZE(ds_req);
	msg.seq = reqnum;

	DBG(zprintf("sending MDSTORE message %"PRIu64"\n", msg.seq));

	rc = ds_send_data_msg(ds->chan, &msg, MDSTORE_DS_TIMEOUT, &resp);
	if (rc != 0) {
		DBG(zprintf("error sending MDSTORE request: %s\n",
		    strerror(rc)));
		return (NULL);
	}

	/* a response is expected */
	if (resp.buf == NULL) {
		DBG(zprintf("no response received from MDSTORE request\n"));
		return (NULL);
	} else {
		*resplen = resp.len;
		return (resp.buf);
	}
}

static ds_svc_t *
get_mdstore_service(response_t *resp)
{
	ds_svc_t *svcp;

	svcp = ds_lookup_svc(ds_sp_channel(), "mdstore");
	if (svcp == NULL || svcp->hdl == DS_SVC_HDL_INVALID) {
		rprintf(resp, "Error: Operation failed because the mdstore "
		    "service to the system controller\nis not registered\n");
		return (NULL);
	} else {
		return (svcp);
	}
}

int
download_md(void *md, uint32_t type, uint32_t size, uint64_t offset,
    response_t *resp)
{
	ds_svc_t *svcp;
	ds_req_t *ds_req;
	ds_resp_payload_t *ds_resp;
	mds_transfer_req_t *mdstore_req;
	mds_resp_t *mdstore_resp;
	uint32_t result, resplen;
	size_t data_len, msg_len;
	int rv = -1;

	if ((svcp = get_mdstore_service(resp)) == NULL)
		return (-1);

	data_len = sizeof (mds_transfer_req_t) + size;
	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) + data_len;

	ds_req = Xmalloc(msg_len);
	DBG(memset(ds_req, 0, msg_len));

	ds_fill_data_header(ds_req, svcp->hdl, data_len);

	/* fill in the transfer request */
	mdstore_req = (mds_transfer_req_t *)DS_DATA_BUF(ds_req);
	mdstore_req->reqnum = hton64(mdstore_seq++);
	mdstore_req->command = hton16(MD_TRANSFER_REQUEST);
	mdstore_req->type = hton16(type);
	mdstore_req->size = hton32(size);
	mdstore_req->offset = hton64(offset);
	memcpy(mdstore_req->md, md, size);

	if ((ds_resp = send_recv(svcp, ds_req, mdstore_req->reqnum,
	    &resplen)) == NULL) {
		rprintf(resp, "Error: Operation failed because of an error "
		    "communicating\nwith the system controller\n");
		goto done;
	}

	mdstore_resp = (mds_resp_t *)((char *)ds_resp +
	    sizeof (ds_data_handle_t));

	result = ntoh32(mdstore_resp->result);
	switch (result) {
	case MDST_SUCCESS:
		rv = 0;
		break;
	case MDST_ALLOC_MD_ERR:
		rprintf(resp, "Error: Operation failed because the "
		    "system controller\nran out of memory\n");
		break;
	case MDST_MD_SIZE_ERR:
		rprintf(resp, "Error: Operation failed because an MD in "
		    "the configuration\nis too large (%d bytes)\n", size);
		break;
	case MDST_MD_TYPE_ERR:
		rprintf(resp, "Error: Operation failed because of an "
		    "internal error.\nInvalid MD type (%d)\n", type);
		break;
	case MDST_MD_COUNT_ERR:
		rprintf(resp, "Error: Operation failed because of an "
		    "internal error.\nMD count mismatch\n");
		break;
	default:
		warning("%s: unexpected result %d", __func__, result);
		rprintf(resp, "Error: Operation failed with an "
		    "unexpected error (%d)\n", result);
	}

	Xfree(ds_resp);
done:
	Xfree(ds_req);
	return (rv);
}

int
bracket_transfer(int command, char *name, int num_mds, response_t *resp)
{
	ds_svc_t *svcp;
	ds_req_t *ds_req;
	ds_resp_payload_t *ds_resp;
	mds_begin_end_req_t *mdstore_req;
	mds_resp_t *mdstore_resp;
	uint32_t result, namelen, resplen;
	size_t data_len, msg_len;
	int rv = -1;

	if ((svcp = get_mdstore_service(resp)) == NULL)
		return (-1);

	if (name == NULL)
		name = "";

	namelen = strlen(name);
	data_len = sizeof (mds_begin_end_req_t) + namelen;
	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) + data_len;

	ds_req = Xmalloc(msg_len);
	DBG(memset(ds_req, 0, msg_len));

	ds_fill_data_header(ds_req, svcp->hdl, data_len);

	/* fill in the transfer begin or end request */
	mdstore_req = (mds_begin_end_req_t *)DS_DATA_BUF(ds_req);
	mdstore_req->reqnum = hton64(mdstore_seq++);
	mdstore_req->command = hton16(command);
	mdstore_req->nmds = hton16(num_mds);
	mdstore_req->namelen = hton32(namelen);
	memcpy(mdstore_req->name, name, namelen);

	if ((ds_resp = send_recv(svcp, ds_req, mdstore_req->reqnum,
	    &resplen)) == NULL) {
		rprintf(resp, "Error: Operation failed because of an error "
		    "communicating\nwith the system controller\n");
		goto done;
	}

	mdstore_resp = (mds_resp_t *)((char *)ds_resp +
	    sizeof (ds_data_handle_t));

	result = ntoh32(mdstore_resp->result);
	switch (result) {
	case MDST_SUCCESS:
		rv = 0;
		break;
	case MDST_BAD_NAME_ERR:
		rprintf(resp, "Error: Operation failed because an "
		    "invalid configuration\nname was given\n");
		break;
	case MDST_SET_EXISTS_ERR:
		rprintf(resp, "Error: Operation failed because a "
		    "configuration\nnamed \"%s\" already exists on the "
		    "system controller.\nBefore being able to save a new "
		    "configuration with\nthis name the existing one must "
		    "be removed\n", name);
		break;
	case MDST_MAX_MDS_ERR:
		rprintf(resp, "Error: Operation failed because the "
		    "maximum\nsuported number of configurations is already "
		    "stored on\nthe system controller. Before being able to "
		    "save a new\nconfiguration one of the existing ones must "
		    "be deleted\n");
		break;
	case MDST_ALLOC_SET_ERR:
		rprintf(resp, "Error: Operation failed because the "
		    "the system controller\nran out of memory. Before being "
		    "able to save a new configuration,\none or more of the "
		    "existing ones must be deleted\n");
		break;
	case MDST_MD_COUNT_ERR:
		rprintf(resp, "Error: Operation failed because of an "
		    "internal error.\nMD count mismatch\n");
		break;
	default:
		warning("%s: unexpected result %d", __func__, result);
		rprintf(resp, "Error: Operation failed with an "
		    "unexpected error (%d)\n", result);
		break;
	}

	Xfree(ds_resp);
done:
	Xfree(ds_req);
	return (rv);
}

static int
manage_md_set(int command, char *mdset_name, response_t *resp)
{
	ds_svc_t *svcp;
	ds_req_t *ds_req;
	ds_resp_payload_t *ds_resp;
	mds_sel_del_req_t *mdstore_req;
	mds_resp_t *mdstore_resp;
	uint32_t result, namelen, resplen;
	size_t data_len, msg_len;
	int rv = -1;

	if ((svcp = get_mdstore_service(resp)) == NULL)
		return (-1);

	namelen = strlen(mdset_name);
	data_len = sizeof (mds_sel_del_req_t) + namelen;
	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) + data_len;

	ds_req = Xmalloc(msg_len);
	DBG(memset(ds_req, 0, msg_len));

	ds_fill_data_header(ds_req, svcp->hdl, data_len);

	/* fill in the select or delete request */
	mdstore_req = (mds_sel_del_req_t *)DS_DATA_BUF(ds_req);
	mdstore_req->reqnum = hton64(mdstore_seq++);
	mdstore_req->command = hton16(command);
	mdstore_req->_rsvd = 0;
	mdstore_req->namelen = hton32(namelen);
	memcpy(mdstore_req->name, mdset_name, namelen);

	if ((ds_resp = send_recv(svcp, ds_req, mdstore_req->reqnum,
	    &resplen)) == NULL) {
		rprintf(resp, "Error: Operation failed because of an error "
		    "communicating\nwith the system controller\n");
		goto done;
	}

	mdstore_resp = (mds_resp_t *)((char *)ds_resp +
	    sizeof (ds_data_handle_t));

	/* check the result */
	result = ntoh32(mdstore_resp->result);
	switch (result) {
	case MDST_SUCCESS:
		rv = 0;
		if (command == MDSET_DELETE_REQUEST) {
			if (zeus.current_spconfig != NULL &&
			    strcmp(mdset_name, zeus.current_spconfig) == 0) {
				/*
				 * Reset zeus.current_spconfig, since we just
				 * deleted the config by that name on the SP.
				 * Also, clear zeus.is_next_spconfig, as the
				 * currently running config no longer matches
				 * any on the SP.
				 */
				Xfree(zeus.current_spconfig);
				zeus.current_spconfig = NULL;
				zeus.is_next_spconfig = false;
				seq_update_hvmd(); /* persist the changes */
			}

			autosave_remove_spconfig_update(mdset_name);
		}

		break;
	case MDST_BAD_NAME_ERR:
		rprintf(resp, "Error: Operation failed because an "
		    "invalid configuration name was given\n");
		break;
	case MDST_NOT_EXIST_ERR:
		rprintf(resp, "Error: Operation failed because there is "
		    "no configuration\nnamed \"%s\" on the system "
		    "controller\n", mdset_name);
		break;
	default:
		warning("%s: unexpected result %d", __func__, result);
		rprintf(resp, "Error: Operation failed with an "
		    "unexpected error (%d)\n", result);
		break;
	}

	Xfree(ds_resp);
done:
	Xfree(ds_req);
	return (rv);
}

static int
list_md_sets(mds_list_resp_t *mdsets, int len, response_t *resp,
    xmlNodePtr data)
{
	int index, namelen;
	char *set;
	xmlNodePtr set_node, new_env, section, item, prop;
	xmlChar *version;

	/* strip off headers to get length of sets description */
	len -= sizeof (ds_data_handle_t) + sizeof (mds_list_resp_t);
	if (len <= 0)
		return (-1);

	/* check for NUL termination */
	if (mdsets->sets[len-1] != '\0')
		return (-1);

	set = mdsets->sets;
	index = 0;
	while (len > 0) {
		bool_t current, next;

		/*
		 * Determine if the set is "current" (i.e. matches the
		 * actively running config), "next" (i.e. the one that will be
		 * booted into on the next poweron), both (in which case we
		 * report "[current]") or neither.
		 */
		current = next = false;
		if (zeus.current_spconfig) {
			if (zeus.is_next_spconfig &&
			    strcmp(zeus.current_spconfig, set) == 0) {
				current = true;
			}
		} else if (index == mdsets->booted_set) {
			/*
			 * Legacy behavior in case we don't know the current
			 * config.  If this config is the booted set and
			 * hasn't been modified, mark it as current.
			 */
			if (zeus.is_next_spconfig)
				current = true;
		}

		/*
		 * If we haven't already marked this set as current and it's
		 * the boot set, mark it as next.
		 */
		if (current == false && index == mdsets->boot_set) {
			next = true;
		}

		namelen = strlen(set) + 1;
		if (data == NULL) {
			rprintf(resp, "%s", set);
			if (current) {
				rprintf(resp, " [current]");
			} else if (next) {
				rprintf(resp, " [next poweron]");
			}
			rprintf(resp, "\n");
		} else {
			version = xmlGetProp(data, VERSION_ATTR);
			if (strcmp((char *)version, "2.0") == 0) {
				set_node = xmlNewChild(data, NULL,
				    XML_SPCONFIG, NULL);
				xmlNewChild(set_node, NULL, XML_SPCONFIG_NAME,
				    (xmlChar *)set);
				if (current) {
					xmlNewChild(set_node, NULL,
					    XML_SPCONFIG_STATUS,
					    XML_CURRENT_STATUS);
				} else if (next) {
					xmlNewChild(set_node, NULL,
					    XML_SPCONFIG_STATUS,
					    XML_NEXT_STATUS);
				}
			} else if (strcmp((char *)version, "3.0") == 0) {
				new_env = data->xmlChildrenNode;
				if (new_env == NULL) {
					new_env = xmlNewChild(data, NULL,
					    ENVELOPE_NODE, NULL);
					xmlNewChild(new_env, NULL,
					    REFERENCES_NODE, NULL);
					section = xmlNewChild(new_env, NULL,
					    SECTION_NODE, NULL);
				}
				item = xmlNewChild(section, NULL, ITEM_NODE,
				    NULL);
				xmlNewChild(item, NULL,
				    OTHER_RESOURCE_NODE_OUT, XML_SPCONFIG);
				prop = xmlNewChild(item, NULL,
				    PROPERTY_NODE_OUT, (xmlChar *)set);
				xmlNewProp(prop, KEY_ATTR, XML_SPCONFIG_NAME);
				if (current) {
					prop = xmlNewChild(item, NULL,
					    PROPERTY_NODE_OUT,
					    XML_CURRENT_STATUS);
					xmlNewProp(prop, KEY_ATTR,
					    XML_SPCONFIG_STATUS);
				} else if (next) {
					prop = xmlNewChild(item, NULL,
					    PROPERTY_NODE_OUT,
					    XML_NEXT_STATUS);
					xmlNewProp(prop, KEY_ATTR,
					    XML_SPCONFIG_STATUS);
				}
			}
			xmlFree(version);
		}

		len -= namelen;
		set += namelen;
		index++;
	}

	return (0);
}

static int
list_sets(response_t *resp, xmlNodePtr data)
{
	ds_svc_t *svcp;
	ds_req_t *ds_req;
	ds_resp_t *ds_resp;
	mds_list_req_t *mdlist_req;
	mds_list_resp_t *mdlist_resp;
	uint32_t result, paylen, resplen;
	int rv = -1;

	if ((svcp = get_mdstore_service(resp)) == NULL)
		return (-1);

	paylen = sizeof (mds_list_req_t);
	ds_req = Xmalloc(offsetof(ds_req_t, payload.data.bytes) + paylen);

	DBG(memset(ds_req, 0, offsetof(ds_req_t, payload.data.bytes) + paylen));

	ds_fill_data_header(ds_req, svcp->hdl, paylen);

	/* fill in the list request */
	mdlist_req = (mds_list_req_t *)DS_DATA_BUF(ds_req);
	mdlist_req->reqnum = hton64(mdstore_seq++);
	mdlist_req->command = hton16(MDSET_LIST_REQUEST);

	if ((ds_resp = send_recv(svcp, ds_req, mdlist_req->reqnum,
	    &resplen)) == NULL) {
		rprintf(resp, "Error: Operation failed because of an error "
		    "communicating\nwith the system controller\n");
		goto done;
	}

	mdlist_resp = (mds_list_resp_t *)((char *)ds_resp +
	    sizeof (ds_data_handle_t));

	result = ntoh32(mdlist_resp->result);
	switch (result) {
	case MDST_SUCCESS:
		/* format list output */
		if ((rv = list_md_sets(mdlist_resp, resplen, resp, data)) != 0)
			if (data == NULL) {
				rprintf(resp, "Error: Operation failed "
				    "because of an internal error.\nInvalid"
				    " message format\n");
			} else {
				xml_add_status(data, XML_FAILURE, "Error: "
				    "Operation failed because of an interal"
				    " error: Invalid message format\n");
			}
		break;
	default:
		warning("%s: unexpected result %d", __func__, result);
		rprintf(resp, "Error: Operation failed with an "
		    "unexpected error (%d)\n", result);
		break;
	}

	Xfree(ds_resp);
done:
	Xfree(ds_req);
	return (rv);
}

static void
prepare_guest(ldom_t *ldp)
{
	/*
	 * Ensure that HV MD generation will use current md frag pointer.
	 * Save current flag for later restoration.
	 *
	 * Clearing this flag is also a requirement to ensure that the
	 * guest MD generation count will not get incremented. This is
	 * critical for keeping the generation sane when a migration is
	 * performed concurrently with mdstore/autosave operations.
	 */
	ldp->restore.flag_gen_new_md = ldp->flag.gen_new_md;
	ldp->flag.gen_new_md = false;

	ASSERT(ldp->restore.committed_pa == INVALID_FRAG_PA);
	ASSERT(ldp->restore.parsed_pa == INVALID_FRAG_PA);
	ASSERT(ldp->restore.next_pa == INVALID_FRAG_PA);

	/*
	 * Save address (for later restoration) and then free any frag
	 * space used for MDs for this guest.
	 */
	if (ldp->frags.committed_md_frag != NULL) {
		ldp->restore.committed_pa =
		    ldp->frags.committed_md_frag->paddr;
		frag_free(ldp->frags.committed_md_frag);
		ldp->frags.committed_md_frag = NULL;
	}
	if (ldp->frags.parsed_md_frag != NULL) {
		ldp->restore.parsed_pa =
		    ldp->frags.parsed_md_frag->paddr;
		frag_free(ldp->frags.parsed_md_frag);
		ldp->frags.parsed_md_frag = NULL;
	}
	if (ldp->frags.next_md_frag != NULL) {
		ldp->restore.next_pa = ldp->frags.next_md_frag->paddr;
		frag_free(ldp->frags.next_md_frag);
		ldp->frags.next_md_frag = NULL;
	}
}

static void
free_guest_mds(ldom_t *ldp)
{
	if (ldp->frags.committed_md_frag != NULL) {
		frag_free(ldp->frags.committed_md_frag);
		ldp->frags.committed_md_frag = NULL;
	}
	if (ldp->frags.parsed_md_frag != NULL) {
		frag_free(ldp->frags.parsed_md_frag);
		ldp->frags.parsed_md_frag = NULL;
	}
	if (ldp->frags.next_md_frag != NULL) {
		frag_free(ldp->frags.next_md_frag);
		ldp->frags.next_md_frag = NULL;
	}
}

static void
restore_guest(ldom_t *ldp)
{
	ldp->flag.gen_new_md = ldp->restore.flag_gen_new_md;

	if (ldp->restore.committed_pa != INVALID_FRAG_PA) {
		ldp->frags.committed_md_frag =
		    frag_assign(ldp, ldp->restore.committed_pa);
		ldp->restore.committed_pa = INVALID_FRAG_PA;
		ASSERT(ldp->frags.committed_md_frag != NULL);
	}
	if (ldp->restore.parsed_pa != INVALID_FRAG_PA) {
		/*
		 * It's possible that the parsed_md_frag is the same as
		 * the committed_md_frag (i.e. during a delayed reconfig).
		 * In this case, simply clone the reference.
		 */
		if (ldp->frags.committed_md_frag->paddr ==
		    ldp->restore.parsed_pa) {
			ldp->frags.parsed_md_frag =
			    frag_clone(ldp->frags.committed_md_frag);
		} else {
			ldp->frags.parsed_md_frag =
			    frag_assign(ldp, ldp->restore.parsed_pa);
		}
		ldp->restore.parsed_pa = INVALID_FRAG_PA;
		ASSERT(ldp->frags.parsed_md_frag != NULL);
	}
	if (ldp->restore.next_pa != INVALID_FRAG_PA) {
		/*
		 * It's possible that the next_md_frag is the same as
		 * the parsed_md_frag (i.e. during a delayed reconfig).
		 * In this case, simply clone the reference.
		 */
		if (ldp->frags.parsed_md_frag->paddr == ldp->restore.next_pa) {
			ldp->frags.next_md_frag =
			    frag_clone(ldp->frags.parsed_md_frag);
		} else {
			ldp->frags.next_md_frag =
			    frag_assign(ldp, ldp->restore.next_pa);
		}
		ldp->restore.next_pa = INVALID_FRAG_PA;
		ASSERT(ldp->frags.next_md_frag != NULL);
	}
}

extern void conv_vldc_ports_to_ds(ldom_t *ldp);
extern void conv_ds_ports_to_vldc(ldom_t *ldp);

static int
do_download(char *mdset_name, int numguests, response_t *resp, bool_t download,
    bool_t autosave)
{
	ldom_t *primary, *ldp;
	uint64_t mdaddr;
	uint8_t *mdbuf;
	md_t *md;
	int err = 0;
	int i, size, mdtype;
	bool_t ctrl_domain;

	primary = find_ldom_by_name(SOD_CONTROL_DOMAIN);
	if (primary == NULL)
		fatal("Couldn't find %s", SOD_CONTROL_DOMAIN);

	if (download) {
		DBG(zprintf("Beginning transfer of set %s (%d guest MDs, "
		    "HV MD and PRI)\n",
		    mdset_name == NULL ? "<unknown>" : mdset_name, numguests));

		err = bracket_transfer(MDSET_BEGIN_REQUEST, mdset_name,
		    numguests + 2, resp);
		if (err != 0) {
			return (-1);
		}
	}

	/* generate and download guest MDs */
	i = 1;
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state < BOUND)
			continue;

		mdtype = 0;
		if (ldp == primary) {
			ctrl_domain = true;
			mdtype = CTL_DOM_MD_TYPE;
			conv_vldc_ports_to_ds(ldp);
		} else {
			ctrl_domain = false;
		}

		/* Does domain own any I/O? */
		if (CHAIN_LENGTH(ldp->bound.iodevices) > 0)
			mdtype |= SVC_DOM_MD_TYPE;

		/* generate guest MD, including PM'ed resources */
		md = seq_gen_guest_md(ldp, true);

		/*
		 * convert the userland ds ports back to vldc
		 * The conv_vldc_ports_to_ds() call changes internal
		 * zeus state by changing the vldc ports in old config
		 * to userland ds ports, we must revert back to vldc after
		 * the MD is generated to keep zeus's internal state
		 * consistent with the current system runtime state.
		 */
		if (ldp == primary)
			conv_ds_ports_to_vldc(ldp);

		backend.aug_guestmd(ldp, md);
		setup_guest_md(md, ctrl_domain);
		size = md_gen_bin(md, &mdbuf);
		md_fini(md);
		if (size > (FRAG_SIZE - pri.md_elbow_room)) {
			fatal("Configuration required a MD of size 0x%x"
			    "bytes which\nis larger than the maximum "
			    "supported size of 0x%x bytes",
			    size, FRAG_SIZE - pri.md_elbow_room);
		}

		ldp->frags.committed_md_frag = frag_alloc(ldp, size);
		mdaddr = ldp->frags.committed_md_frag->paddr;

		if (download) {
			DBG(zprintf("Downloading guest MD %d of %d to SC, "
			    "size=%d, offset=0x%"PRIx64", type=0x%x\n",
			    i, numguests, size, mdaddr, mdtype));

			err = download_md(mdbuf, mdtype, size, mdaddr, resp);
			DBG(if (err != 0)
			    zprintf("Failed to download guest MD to SC\n"));
		}

		if (autosave) {
			autosave_gen_guest_md(ldp, mdaddr, mdbuf, size);
		} else if (!download || (debug_level & 0x7))
			dump_to_file(mdbuf, size, "dump-sp-%s-md.%"PRIx64".bin",
			    ldp->namep, mdaddr);

		free(mdbuf);
		if (err != 0) {
			return (-1);
		}
		i++;
	}

	/* generate and download hypervisor MD, sans pm_status */
	md = target.gen_hvmd(false);	/* allocates zeus.next_hvmd_fp */
	backend.aug_hvmd(md);

	size = md_gen_bin(md, &mdbuf);
	md_fini(md);
	if (size > (max_hv_md_size - pri.md_elbow_room)) {
		fatal("Configuration required a MD of size 0x%x bytes "
		    "which\nis larger than the maximum supported size of "
		    "0x%x bytes", size, max_hv_md_size - pri.md_elbow_room);
	}

	/* Rotate in new HV MD */
	ASSERT(zeus.current_hvmd_fp == NULL);
	zeus.current_hvmd_fp = zeus.next_hvmd_fp;
	zeus.next_hvmd_fp = NULL;
	mdaddr = zeus.current_hvmd_fp->paddr;

	if (download) {
		DBG(zprintf("Downloading HV MD to SP, size=%d, "
		    "offset=0x%"PRIx64"\n", size, mdaddr));

		err = download_md(mdbuf, HV_MD_TYPE, size, mdaddr, resp);
		DBG(if (err != 0) zprintf("Failed to download HV MD to SC\n"));
	}

	if (autosave) {
		autosave_gen_hv_md(mdaddr, mdbuf, size);
	} else if (!download || (debug_level & 0x7))
		dump_to_file(mdbuf, size, "dump-sp-hv-md.%"PRIx64".bin",
		    mdaddr);

	free(mdbuf);
	if (err != 0)
		return (-1);

	/* download PRI */
	size = md_gen_bin(pri.pri, &mdbuf);
	if (size > (FRAG_SIZE - pri.md_elbow_room)) {
		fatal("Configuration required a MD of size 0x%x bytes "
		    "which\nis larger than the maximum supported size of "
		    "0x%x bytes", size, FRAG_SIZE - pri.md_elbow_room);
	}
	if (download) {
		DBG(zprintf("Downloading PRI to SP, size=%d\n", size));

		err = download_md(mdbuf, PRI_TYPE, size, 0, resp);
		DBG(if (err != 0) zprintf("Failed to download PRI to SC\n"));
	}

	if (autosave) {
		autosave_gen_sp_pri_md(mdbuf, size);
	} else if (!download || (debug_level & 0x7))
		dump_to_file(mdbuf, size, "dump-sp-pri.bin");

	free(mdbuf);
	if (err != 0)
		return (-1);

	if (download) {
		DBG(zprintf("Ending transfer of set %s\n",
		    mdset_name == NULL ? "<unknown>" : mdset_name));

		(void) bracket_transfer(MDSET_END_REQUEST, NULL, numguests + 2,
		    resp);
	}

	return (0);
}

int
mdstore_download(char *mdset_name, response_t *resp, bool_t download,
    bool_t autosave)
{
	ldom_t *ldp;
	int err = 0;
	int numguests;
	char *old_current_spconfig;
	bool_t old_is_next_spconfig;

	if (download == true || autosave == true) {
		/*
		 * Update zeus.current_spconfig before we download.  Save old
		 * values to revert to after autosave, or in case the
		 * download fails.
		 */
		old_current_spconfig  = zeus.current_spconfig;
		zeus.current_spconfig = Xstrdup(mdset_name);
		old_is_next_spconfig = zeus.is_next_spconfig;
		zeus.is_next_spconfig = true;
	}

	/*
	 * A download (or debug dump) operation has been requested.
	 */
	ASSERT(zeus.restore.current_pa == INVALID_FRAG_PA);
	ASSERT(zeus.restore.next_pa == INVALID_FRAG_PA);
	ASSERT(zeus.restore.reconf_pa == INVALID_FRAG_PA);

	/*
	 * Prepare for guest and HV download config MD generation by saving
	 * the current MD addresses and freeing all MD frags.
	 */
	numguests = 0;
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state >= BOUND) {
			numguests++;
			prepare_guest(ldp);
		}
	}
	if (zeus.current_hvmd_fp != NULL) {
		zeus.restore.current_pa = zeus.current_hvmd_fp->paddr;
		hv_md_free(zeus.current_hvmd_fp);
		zeus.current_hvmd_fp = NULL;
	}
	if (zeus.next_hvmd_fp != NULL) {
		zeus.restore.next_pa = zeus.next_hvmd_fp->paddr;
		hv_md_free(zeus.next_hvmd_fp);
		zeus.next_hvmd_fp = NULL;
	}
	if (zeus.reconf_fp != NULL) {
		zeus.restore.reconf_pa = zeus.reconf_fp->paddr;
		hv_md_free(zeus.reconf_fp);
		zeus.reconf_fp = NULL;
	}

	/*
	 * Do the config generation and download to SC.
	 */

	err = do_download(mdset_name, numguests, resp, download, autosave);

	/*
	 * Restore all MDs to state prior to download.
	 */

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state >= BOUND)
			free_guest_mds(ldp);
	}
	if (zeus.current_hvmd_fp != NULL) {
		hv_md_free(zeus.current_hvmd_fp);
		zeus.current_hvmd_fp = NULL;
	}
	if (zeus.next_hvmd_fp != NULL) {
		hv_md_free(zeus.next_hvmd_fp);
		zeus.next_hvmd_fp = NULL;
	}
	if (zeus.reconf_fp != NULL) {
		hv_md_free(zeus.reconf_fp);
		zeus.reconf_fp = NULL;
	}

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state >= BOUND)
			restore_guest(ldp);
	}
	if (zeus.restore.current_pa != INVALID_FRAG_PA) {
		zeus.current_hvmd_fp = hv_md_assign(zeus.restore.current_pa);
		zeus.restore.current_pa = INVALID_FRAG_PA;
		ASSERT(zeus.current_hvmd_fp != NULL);
	}
	if (zeus.restore.next_pa != INVALID_FRAG_PA) {
		zeus.next_hvmd_fp = hv_md_assign(zeus.restore.next_pa);
		zeus.restore.next_pa = INVALID_FRAG_PA;
		ASSERT(zeus.next_hvmd_fp != NULL);
	}
	if (zeus.restore.reconf_pa != INVALID_FRAG_PA) {
		zeus.reconf_fp = hv_md_assign(zeus.restore.reconf_pa);
		zeus.restore.reconf_pa = INVALID_FRAG_PA;
		ASSERT(zeus.reconf_fp != NULL);
	}

	if (download == true && err == 0) {
		if (old_current_spconfig) {
			Xfree(old_current_spconfig);
			(void) autosave_set_spconfig_timestamp(mdset_name);
		} else if (autosave_rename_default(mdset_name, resp) == true) {
			(void) autosave_set_spconfig_timestamp(mdset_name);
		}
		seq_update_hvmd(); /* persist the changes */
	} else if (autosave == true || err != 0) {
		/*
		 * Either we're just doing a host-side autosave, or the download
		 * to the SP failed.  In either case, we revert to the previous
		 * values for the spconfig-related properties.
		 */
		Xfree(zeus.current_spconfig);
		zeus.current_spconfig = old_current_spconfig;
		zeus.is_next_spconfig = old_is_next_spconfig;
	}
	return (err);
}

/*
 * CLI entry point.
 */
int
mdstore(char *op, char *mdset_name, response_t *resp, xmlNodePtr data)
{
	int err = 0;
	bool_t download = false;

	/*
	 * Get op(eration).  If "select" or "remove" call manage_md_set()
	 */
	if (strcmp(op, "select") == 0) {
		return (manage_md_set(MDSET_SELECT_REQUEST, mdset_name, resp));
	} else if (strcmp(op, "remove") == 0) {
		return (manage_md_set(MDSET_DELETE_REQUEST, mdset_name, resp));
	} else if (strcmp(op, "list") == 0) {
		return (list_sets(resp, data));
	} else if (strcmp(op, "download") == 0) {
		download = true;

#if ZEUS_DEBUG_CLI
	} else if (strcmp(op, "dump") == 0) {
		download = false;
#endif
	} else {
		DBG(zprintf("mdstore: unknown operation.\n"));
		return (-1);
	}

	err = mdstore_download(mdset_name, resp, download, false);

	return (err);
}

static void
mdstore_data_handler(ds_svc_t *svc, void *buf, size_t msglen)
{
	ds_msg_tgt_t tgt;
	ds_msg_t resp;
	mds_resp_t *mdstore_resp;

	if (msglen < sizeof (ds_data_handle_t) + sizeof (mds_resp_t)) {
		DBG(zprintf("received invalid MDSTORE response, got %d bytes "
		    "and expected >= %d\n", msglen,
		    sizeof (ds_data_handle_t) + sizeof (mds_resp_t)));
		Xfree(buf);
		return;
	}

	tgt.chan = svc->chan;
	tgt.ds_hdl = svc->hdl;

	mdstore_resp = (mds_resp_t *)((char *)buf + sizeof (ds_data_handle_t));

	resp.buf = buf;
	resp.len = msglen;
	resp.seq = ntoh64(mdstore_resp->reqnum);

	/*
	 * Check the pending queue for a matching message and
	 * signal the waiting thread.
	 */
	if (ds_check_resp(&tgt, &resp) == DS_MSG_CLAIMED) {
		/* match found */
		return;
	}

	warning("<%"PRIx64"> unmatched MDSTORE message %"PRIu64"",
	    svc->hdl, resp.seq);
	Xfree(buf);
}

static void
mdstore_reg_handler(ds_svc_t *svc)
{
	ds_seq_req_t *dsreq;

	dsreq = Xmalloc(sizeof (*dsreq));
	dsreq->svc = svc;
	dsreq->msg = NULL;
	dsreq->len = 0;

	seq_enqueue_ds_req(dsreq);

	autosave_notify_mdstore_ready();
}

#if 0	/* FIXME - funciton is unused */
static void
flush_updates(ds_svc_t *svc)
{
	ds_req_t *ds_req;
	ds_resp_payload_t *ds_resp;
	mds_flush_req_t *mdstore_req;
	mds_resp_t *mdstore_resp;
	uint32_t result, resplen;
	size_t data_len, msg_len;

	DBG(zprintf("%s: flushing variable updates\n", __func__));

	data_len = sizeof (mds_flush_req_t);
	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) + data_len;

	ds_req = Xmalloc(msg_len);
	DBG(memset(ds_req, 0, msg_len));

	ds_fill_data_header(ds_req, svc->hdl, data_len);

	/* fill in the update request */
	mdstore_req = (mds_flush_req_t *)DS_DATA_BUF(ds_req);
	mdstore_req->reqnum = hton64(mdstore_seq++);
	mdstore_req->command = hton16(MDSET_FLUSH_REQUEST);

	if ((ds_resp = send_recv(svc, ds_req, mdstore_req->reqnum,
	    &resplen)) == NULL) {
		warning("%s: error communicating with the system controller",
		    __func__);
		goto done;
	}

	mdstore_resp = (mds_resp_t *)((char *)ds_resp +
	    sizeof (ds_data_handle_t));

	result = ntoh32(mdstore_resp->result);
	switch (result) {
	case MDST_SUCCESS:
		break;
	default:
		warning("%s: result %d", __func__, result);
	}

	Xfree(ds_resp);
done:
	Xfree(ds_req);
}
#endif

static void
apply_updates(ds_svc_t *svc, void *md_buf, uint32_t md_size)
{
	md_t *md;
	md_node_t **nodes;
	bool_t update_md = false;
	int num;

	DBG(zprintf("%s: applying variable updates\n", __func__));
	DBG(dump_to_file(md_buf, md_size, "dump-var-updates.bin"));

	if ((md = md_scan_bin(md_buf, md_size)) == NULL) {
		warning("Invalid variable updates payload");
		return;
	}

	nodes = Xmalloc(md_node_count(md) * sizeof (md_node_t *));
	num = md_scan_dag(md, NULL, md_find_name(md, "variables"),
	    md_find_name(md, "fwd"), nodes);

	if (num > 0) {
		/*
		 * We no longer flush variable updates, as that causes loss
		 * of persistence for variables previously set (e.g. at the
		 * OpenBoot prompt).
		 */
		if (set_vars_from_properties(zeus.hvctl_ldom_p, nodes[0])) {
			/* some variable updates were made */
			database_save();
			update_md = true;
		} else {
			DBG(zprintf("%s: no changes to variables\n", __func__));
		}
	}

	num = md_scan_dag(md, NULL, md_find_name(md, "keystore"),
	    md_find_name(md, "fwd"), nodes);

	if (num > 0) {
		/*
		 * We do not flush key updates, as that causes loss of
		 * persistence for keys previously set (e.g. at the
		 * OpenBoot prompt).
		 */
		if (set_keys_from_properties(zeus.hvctl_ldom_p, nodes[0])) {
			/* some key updates were made */
			update_md = true;
		} else {
			DBG(zprintf("%s: no changes to keys\n", __func__));
		}

	}

	if (update_md)
		/* update machine description */
		seq_update_vars(zeus.hvctl_ldom_p, true);

	md_fini(md);
	Xfree(nodes);
}

/*
 * This is only invoked via a queued up request from the registration handler.
 * Once the protocol is registered we query the system controller for any
 * variable updates.
 */
/*ARGSUSED1*/
void
mdstore_seq_handler(ds_svc_t *svc, void *msg, size_t msglen)
{
	ds_req_t *ds_req;
	ds_resp_payload_t *ds_resp;
	mds_update_req_t *mdstore_req;
	mds_resp_t *mdstore_resp;
	uint32_t result, resplen;
	size_t data_len, msg_len;

	ASSERT(msg == NULL);

	data_len = sizeof (mds_update_req_t);
	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) + data_len;

	ds_req = Xmalloc(msg_len);
	DBG(memset(ds_req, 0, msg_len));

	ds_fill_data_header(ds_req, svc->hdl, data_len);

	/* fill in the update request */
	mdstore_req = (mds_update_req_t *)DS_DATA_BUF(ds_req);
	mdstore_req->reqnum = hton64(mdstore_seq++);
	mdstore_req->command = hton16(MDSET_UPDATES_REQUEST);

	if ((ds_resp = send_recv(svc, ds_req, mdstore_req->reqnum,
	    &resplen)) == NULL) {
		warning("%s: error communicating with the system controller",
		    __func__);
		goto done;
	}

	mdstore_resp = (mds_resp_t *)((char *)ds_resp +
	    sizeof (ds_data_handle_t));

	/* check the result */
	result = ntoh32(mdstore_resp->result);
	switch (result) {
	case MDST_SUCCESS:
		/* strip off headers to get size of updates MD */
		resplen -= sizeof (ds_data_handle_t) + sizeof (mds_resp_t);
		apply_updates(svc, mdstore_resp->data, resplen);
		break;
	default:
		warning("%s: result %d", __func__, result);
		break;
	}

	Xfree(ds_resp);
done:
	Xfree(ds_req);
}

/*
 * The versions supported
 */
static prot_version_t mdstore_versions[] =
{
	{1, 0}
};

static svc_ops_t mdstore_ops = {
	mdstore_reg_handler,		/* reg_handler */
	NULL,				/* unreg_handler */
	mdstore_data_handler,		/* data_handler */
	mdstore_seq_handler,		/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	svc_info_t *info;

	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 1;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(SP_CHAN_NAME);
	info->num_svcs = 1;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(MD_STORE_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = 0;
	info->num_vers = NUM_VERS(mdstore_versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	memcpy(info->versions, mdstore_versions,
	    info->num_vers * sizeof (prot_version_t));

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, MD_STORE_SVC) == 0)
		return (&mdstore_ops);
	else
		return (NULL);
}
