/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <values.h>
#include <sys/sysmacros.h>
#include <sys/errno.h>

#include "basics.h"
#include "chain.h"
#include "errors.h"
#include "mdesc_impl.h"
#include "ldom.h"
#include "clp.h"
#include "allocate.h"
#include "database.h"
#include "zprintf.h"
#include "zpri.h"
#include "response.h"
#include "fatal.h"
#include "support.h"

#include "memory.h"
#include "dr_mem.h"
#include "sequencer.h"
#include "cli_conn.h"
#include "xml_common.h"
#include "xml_cntrl_common.h"
#include "xml_parse.h"
#include "xml_v3_defs.h"
#include "xmpp_server.h"
#include "xml_output.h"
#include "xml_event.h"
#include "database_internal.h"
#include "ds_impl.h"
#include "backend.h"
#include "ldomconfigure.h"

/*
 * The reconfig memory can be done via either a memory DR or a delayed reconfig
 * operation.
 * 1. If the domain is not active, reconfig memory is done immediately. (no
 *	memory DR or delayed reconfig)
 * 2. If the domain is active:
 *    2-1 If the domain is a non-control domain:
 *	  2-1-1 If the control domain is in a delayed reconfig, reconfig memory
 *		will be rejected.
 *	  2-1-2 if the control domain is not in a delayed reconfig:
 *		2-1-2-1 if the domain support memory DR, memory reconfig will
 *		        be done via memory DR.
 *		2-1-2-2 If the domain does not support memory DR, reconfig
 *			memory will be rejected.
 *    2-2 if the domain is a control domain:
 *	  2-2-1 If the domain is in factory-default config, reconfig memory
 *		will be done via delayed reconfig.
 *	  2-2-2 if the domain is not in factory-default config:
 *		2-2-2-1 if the domain support memory DR, reconfig memory will
 *			be done via memory DR.
 *		2-2-2-2 if the domain does not support memory DR, reconfig
 *			memory will be done via delayed reconfig.
 */
extern bool_t hv_is_mem_dr_supported(void);

#define	NO_RA	(-1)			/* the unset RA value */
#define	KB(x)	(x / 1024)


enum {
	PHYSICAL_BINDING,
	VIRTUAL_BINDING
};

static void check_mem_addr(ldom_t *ldp, uint64_t raddr, uint64_t size);
static int check_mem_dr_capable(ldom_t *ldp);

/* must be aligned to the largest possible page size */
#define	GUEST_REAL_BASE		0x0

/* list of free memory blocks */
mem_blocks_t free_memory = { .blocks = DEFAULTCHAIN };

/* list of memory blocks allocated for system use, i.e. frags, HV, etc. */
mem_blocks_t system_memory = { .blocks = DEFAULTCHAIN };

/*
 * Create and return a new memory block for a physical memory range within a
 * physical memory block. This just creates and initializes the block. It does
 * not link it into any lists or update the free list. That is to be done by
 * the caller which must also ensure that the passed in memory range is within
 * the physical memory block.
 */
mblock_t *
mem_new_mblock(pri_mblock_t *physblk, uint64_t paddr, uint64_t size)
{
	mblock_t *mblk;

	ASSERT((physblk->pa_base <= paddr) &&
	    ((physblk->pa_base + physblk->size) >= (paddr + size)));

	mblk = Xcalloc(1, mblock_t);
	mblk->paddr = paddr;
	mblk->size = size;
	mblk->raddr = NO_RA;
	mblk->pmblk = physblk;
	mblk->ldp = NULL;
	mblk->mblk_resid = NULL;
	mblk->flag_isconfigd = false;

	return (mblk);
}

/*
 * Place a memory block on the free list.
 */
void
mem_free_mblock(mblock_t *mbp)
{
	mblock_t *next_mbp, *prev_mbp;

	DBG(zprintf("%s: freeing block [0x%"PRIx64" + 0x%"PRIx64"]\n",
	    __func__, mbp->raddr, mbp->size));

	/* find entries with PAs bracketing the one we're freeing */
	prev_mbp = NULL;
	for (next_mbp = CHAIN_START(free_memory.blocks);
	    next_mbp != NULL && next_mbp->paddr <= mbp->paddr;
	    next_mbp = CHAIN_NEXT(next_mbp->pblk)) {
		prev_mbp = next_mbp;
	}

	/* can we coalesce with the previous mblock? */
	if ((prev_mbp != NULL) && (prev_mbp->pmblk == mbp->pmblk) &&
	    ((prev_mbp->paddr + prev_mbp->size) == mbp->paddr)) {
		prev_mbp->size += mbp->size;
		Xfree(mbp);
		mbp = prev_mbp;
		DBG(zprintf("%s: coalesced with prev mblk [0x%"PRIx64
		    " + 0x%"PRIx64"]\n", __func__, mbp->paddr, mbp->size));
	} else {
		DBG(zprintf("%s: creating new entry after prev mblk\n",
		    __func__));
		CHAIN_INSERT_AFTER(free_memory.blocks, prev_mbp, mbp, pblk);
	}

	/* can we coalesce with the next mblock? */
	if ((next_mbp != NULL) && (next_mbp->pmblk == mbp->pmblk) &&
	    ((mbp->paddr + mbp->size) == next_mbp->paddr)) {
		mbp->size += next_mbp->size;
		CHAIN_UNLINK(free_memory.blocks, next_mbp, pblk);
		Xfree(next_mbp);
		DBG(zprintf("%s: coalesced with next mblk [0x%"PRIx64
		    " + 0x%"PRIx64"]\n", __func__, mbp->paddr, mbp->size));
	}
}

/*
 * Unbind a mblk
 */
void
mem_unbind(mblock_t *mblk)
{
	CHAIN_UNLINK(mblk->ldp->bound.memory, mblk, ldom);
	CHAIN_UNLINK(mblk->pmblk->bound.memory, mblk, pblk);

	free_res(mblk->mblk_resid);

	/*
	 * Mblk may or may not contains actual memory. Return memory to
	 * free memory pool if mblk contains memory otherwise,
	 * just free the mblk.
	 */
	if (mblk->size > 0)
		mem_free_mblock(mblk);
	else
		Xfree(mblk);
}

/*
 * Check if we should do memory dr or delayed reconfig.
 * This function is called by mem_add, mem_del and mem_unbind_all. In all those
 * 3 cases we need to check if we are doing a delayed reconfig or memory DR.
 * 1. If domain is not active, we don't do memory DR.
 * 2. Already in delayed reconfig, allow memory reconfig (not memory DR).
 * 3. If the domain is active and it is not the control domain,
 *    memory DR capability was verified by config_mod_allow. Do memory DR.
 * 4. If it is control domain and it is in factory-default configuration,
 * do delayed reconfig (not memory DR).
 * 5. If the domain is control domain and supports memory DR, do memory DR.
 *    Otherwise, do delayed_reconfig.
 */
static bool_t
can_use_memdr(ldom_t *ldp)
{
	if (ldp->state != ACTIVE)
		return (false);

	/* check if already in delayed reconfig */
	if (in_delayed_reconfig(NULL) == true)
		return (false);

	/* check if domain supports memory DR */
	if (check_mem_dr_capable(ldp) != MEM_DR_OP_SUCC)
		return (false);

	/* check if it is a control domain */
	if (IS_CONTROL_DOMAIN(ldp) == false) {
		return (true);
	}

	/*
	 * check if the control domain has all memory,
	 * includes factory-default config
	 */
	if (CHAIN_START(free_memory.blocks) == NULL) {
		/*
		 * No free memory. Check if there is any other bound
		 * or active domain has bound memory.
		 */
		ldom_t *tmpldp;

		for (tmpldp = CHAIN_START(zeus.ldoms); tmpldp != NULL;
		    tmpldp = CHAIN_NEXT(tmpldp->link)) {
			if (IS_CONTROL_DOMAIN(tmpldp))
				continue;
			if (ldom_size_bound_mem(tmpldp) > 0)
				break;
		}

		if (tmpldp == NULL)
			/*
			 * No other guest domain has memory,
			 * do delayed reconfig.
			 */
			return (false);
	}

	return (true);

}

/*
 * Unbind all of a LDom's current memory bindings.
 */
void
mem_unbind_all(ldom_t *ldp)
{
	mblock_t *mblk, *next;

	/*
	 * For mem dr, only unbind mblks that are allocated but
	 * not configured.
	 */
	if (can_use_memdr(ldp) == true) {
		dr_mblk_t *mblkp;
		int nmblk;

		nmblk = seq_get_unconfigured_mblks(ldp, (void **)&mblkp);
		if (nmblk > 0) {
			mem_unbind_unconfigured(ldp, nmblk, mblkp);
			Xfree(mblkp);
		}
		return;
	}

	for (mblk = CHAIN_START(ldp->bound.memory); mblk != NULL; mblk = next) {
		next = CHAIN_NEXT(mblk->ldom);
		mem_unbind(mblk);
	}
}

static uint64_t
get_raddr(uint64_t paddr, uint64_t segnum)
{
	uint64_t raddr;

	/*
	 * RA address alignment is tricky - we need to pick an RA whose
	 * largest page size alignment is the same as the RA alignment
	 * of the PA.  Also we'd need to ensure that we don't overlap
	 * other allocated RA blocks for this LDom, and that the RAs
	 * chosen are different in one or more of the bits in the range:
	 * zeus.ra2pa_shift through
	 * zeus.ra2pa_shift + Log(2)(zeus.guest_max_mblocks).
	 */

	/*
	 * Make RA congruent to PA for the largest possible page size
	 * get_raddr does not need paddr since currently (paddr &
	 * (pri.max_pagesize - 1)) == 0. paddr is kept here for
	 * potential future use.
	 */
	raddr = (paddr & (pri.max_pagesize-1)) | GUEST_REAL_BASE;

	/*
	 * If raddr != 0, either caller should adjust the max size of
	 * the segment, revise calculation of end segment, or a max size
	 * of the segment will take 2 segments.
	 */
	ASSERT(raddr == 0);

#if OBP_RA_ERRATA
	/*
	 * OBP currently doesn't like RAs of 0
	 * Caller is responsible to make sure the max size of
	 * this segment is (1LL << zeus.ra2pa_shft - pri.max_pagesize)
	 */
	if (raddr == 0 && segnum == 0)
		raddr = pri.max_pagesize;
#endif

	/* Caller is responsible for providing proper RA segment num */
	raddr |= segnum << zeus.ra2pa_shift;
	return (raddr);
}

static uint64_t
granularity_multiple(uint64_t input, uint64_t granularity)
{
	if (granularity == 0)
		return (input);
	return ((input + granularity - 1)/granularity * granularity);
}

/*
 * Allocate a physical memory range from a block of free memory.
 * The free memory list is updated.
 */
static mblock_t *
allocate_mem(mblock_t *freeblk, uint64_t paddr, uint64_t size)
{
	mblock_t *mblk, *newblk;

	ASSERT(freeblk->pmblk->pa_base <= paddr &&
	    (freeblk->pmblk->pa_base + freeblk->pmblk->size) >= (paddr + size));

	mblk = mem_new_mblock(freeblk->pmblk, paddr, size);

	/* associate allocated mblock with physical mblock */
	CHAIN_ADD(freeblk->pmblk->bound.memory, mblk, pblk);

	/* unlink the original free block which has been allocated from */
	CHAIN_UNLINK(free_memory.blocks, freeblk, pblk);

	/* if space available at the beginning create a new free block */
	if (freeblk->paddr < mblk->paddr) {
		newblk = mem_new_mblock(freeblk->pmblk, freeblk->paddr,
		    mblk->paddr - freeblk->paddr);
		mem_free_mblock(newblk);
	}

	/* if space available at the end create a new free block */
	if ((freeblk->paddr + freeblk->size) > (mblk->paddr + mblk->size)) {
		newblk = mem_new_mblock(freeblk->pmblk,
		    mblk->paddr + mblk->size,
		    (freeblk->paddr + freeblk->size) -
		    (mblk->paddr + mblk->size));
		mem_free_mblock(newblk);
	}

	Xfree(freeblk);

	mblk->flag_isconfigd = false;
	return (mblk);
}

/*
 * Find and return the free memory block which contains a given physical
 * memory range. Returns NULL if no such block exists.
 */
static mblock_t *
find_free_mblock(uint64_t basepa, uint64_t size)
{
	mblock_t *freeblk = NULL;

	CHAIN_ITER(free_memory.blocks, freeblk, pblk) {
		if ((freeblk->paddr <= basepa) &&
		    ((freeblk->paddr + freeblk->size) >= (basepa + size)))
			break;
	}

	return (freeblk);
}

/*
 * Bind a given physical memory range. Returns the bound memory block or NULL
 * if the binding could not be satisfied.
 */
mblock_t *
mem_bind_phys_explicit(uint64_t basepa, uint64_t size)
{
	mblock_t *freeblk, *mblk;

	if ((freeblk = find_free_mblock(basepa, size)) == NULL)
		return (NULL);

	mblk = allocate_mem(freeblk, basepa, size);
	CHAIN_ADD(system_memory.blocks, mblk, sys);

	return (mblk);
}

/*
 * Bind to a LDom a list of virtual memory bindings.
 * Returns true if successful otherwise false.
 * When bind to an existing LDom list, if the list is not 256M aligned,
 * remaining memory pool after binding is not aligned to 256M.
 */
bool_t
mem_bind_virt_explicit(ldom_t *ldp, mem_blocks_t *bindings)
{
	mblock_t *freeblk, *mblk, *bound;

	CHAIN_ITER(bindings->blocks, mblk, blk) {
		freeblk = find_free_mblock(mblk->paddr, mblk->size);
		if (freeblk == NULL) {
			/* no free memory which can satisfy the binding */
			mem_unbind_all(ldp);
			return (false);
		}

		bound = allocate_mem(freeblk, mblk->paddr, mblk->size);
		bound->raddr = mblk->raddr;
		bound->mblk_resid = mblk->mblk_resid;
		bound->ldp = ldp;
		bound->flag_isconfigd = mblk->flag_isconfigd;

		CHAIN_ADD(ldp->bound.memory, bound, ldom);
	}

	return (true);
}

/*
 * There are several cases for handling the granularity requirement:
 *
 * 1) Usable part of this mblk (after accounting for PA alignment restriction)
 *    is large enough to hold remainder of total allocation, even if we need to
 *    round up to meet granularity requirement.
 *    ACTION: Round up (this is final mblk)
 *
 * 2) Usable portion of mblk large enough to hold remainder of allocation, but
 *    not after rounding up to meet granularity requirement.
 *    ACTION: Round down (remainder to be satisfied by next mblk to meet
 *            case #1).
 *
 * 3) Usable portion of mblk not large enough to hold remainder of allocation,
 *    but equal to or larger than granularity.
 *    ACTION: Use as much of mblk as possible modulo granularity requirement.
 *
 * 4) Usable portion of mblk smaller than granularity.
 *    ACTION: Skip mblk.
 */
static bool_t
handle_granularity(uint64_t aligned_size, uint64_t *size, uint64_t *alloc_size)
{
	if (aligned_size >= granularity_multiple(*size, MEM_MINSIZE)) {
		/* case 1 */
		*size = granularity_multiple(*size, MEM_MINSIZE);
		*alloc_size = *size;
	} else if (aligned_size >= *size) {
		/* case 2 */
		*alloc_size = *size - *size % MEM_MINSIZE;
	} else if (aligned_size >= MEM_MINSIZE) {
		/* case 3 */
		*alloc_size = aligned_size - aligned_size % MEM_MINSIZE;
	} else {
		/* case 4 */
		return (false);
	}

	if (*alloc_size < MEM_MINSIZE) {
		return (false);
	}

	return (true);
}

/*
 * Warn that allocated size is different from asked size
 */
static void
allocation_warn(uint64_t asked_size, uint64_t alloc_size, response_t *resp)
{
	char *asked, *alloc;

	if (resp == NULL)
		return;

	asked = conv_size(asked_size);
	alloc = conv_size(alloc_size);

	rprintf(resp, MSG_WARN_INC_MEM_ALLOC, asked, alloc);

	Xfree(asked);
	Xfree(alloc);
}

static uint_t
get_end_seg_num(uint64_t raddr, uint64_t size)
{
	return ((raddr + size - 1) >> zeus.ra2pa_shift);
}

/*
 * Find the first unused segment number to be assigned for a mblk.
 * We mark all the segments currently used by the domain. Then we find
 * continuous segments that are not used for the new allocated memory.
 */
static uint64_t
get_seg_num(ldom_t *ldp, uint64_t req_size, uint64_t pa_addr)
{
	uint64_t raddr, size;
	mblock_t *mblk;
	uint_t *seg_list;
	uint_t eseg_num, seg_num, last_used_segment;
	bool_t found;

	last_used_segment = 0;

	seg_list = Xcalloc(zeus.guest_max_mblocks, uint_t);

	/*
	 * Mark all the segments currently used by the domain.
	 */
	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		raddr = mblk->raddr;
		size = mblk->size;

		seg_num = raddr >> zeus.ra2pa_shift;
		eseg_num = get_end_seg_num(raddr, size);

		while (seg_num <= eseg_num) {
			/*
			 * An mblk that spans multiple segment sizes
			 * consumes multiple segment numbers.
			 */
			seg_list[seg_num] = 1;
			seg_num++;
		}
		if (last_used_segment < eseg_num)
			last_used_segment = eseg_num;
	}
	seg_num = 0;
	found = false;
	/*
	 * Check if there are continuous segments within seg list
	 * can be used for this mblk.
	 */
	while (seg_num <= last_used_segment && !found) {

		/* first potential segment */
		if (seg_list[seg_num] == 0) {

			raddr = get_raddr(pa_addr, seg_num);
			eseg_num = get_end_seg_num(raddr, req_size);

			if (eseg_num >= last_used_segment) {
				break;
			}

			/* can rest segments fit? */
			found = true;
			for (int i = seg_num + 1; i <= eseg_num; i++) {

				if (seg_list[i] == 1) {
					/* segment is occupied */
					seg_num =  i + 1;
					found = false;
					break;
				}
			}
		} else {
			seg_num++;
		}
	}

	if (!found) {
		/*
		 * Did not find suitable free segments between 0 and last used
		 * segments. See if there are sutable free segments between
		 * last used segments and max segments.
		 */
		seg_num = last_used_segment + 1;
		raddr = get_raddr(pa_addr, seg_num);
		eseg_num = get_end_seg_num(raddr, req_size);

		/*
		 * Check if we are over the segment limitation.
		 */
		if (eseg_num >= zeus.guest_max_mblocks)
			seg_num = zeus.guest_max_mblocks;
	}

	Xfree(seg_list);
	return (seg_num);
}

static uint_t
get_seg_count(ldom_t *ldp)
{
	mblock_t *mblk;
	uint_t seg_count, eseg_num, seg_num;

	seg_count = 0;

	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {

		seg_num = mblk->raddr >> zeus.ra2pa_shift;
		eseg_num = get_end_seg_num(mblk->raddr, mblk->size);

		seg_count += (eseg_num - seg_num + 1);

	}

	return (seg_count);
}

/*
 * Check if a given raddr and size already allocated in the ldp
 */
static void
check_mem_addr(ldom_t *ldp, uint64_t raddr, uint64_t size)
{
	mblock_t *mblk;
	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		if (mblk->raddr >= raddr + size)
			continue;
		if (mblk->raddr + mblk->size <= raddr)
			continue;
		zprintf("%s: address found in bound mem addr=%"PRIx64
		    " size=%"PRIx64" alloc addr=%"PRIx64" size=%"PRIx64"\n",
		    ldp->namep, mblk->raddr, mblk->size, raddr, size);
		break;
	}
	if (mblk != NULL) {
		CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
			zprintf("%s: raddr = %"PRIx64" size = %"PRIx64"\n",
			    ldp->namep, mblk->raddr, mblk->size);
		}

		fatal("Duplicate addr, size found \n");
	}
}

/*
 * Create memory bindings (returned in bindings) for the given type and
 * size. For virtual bindings ldp is the LDom being bound. For physical
 * bindings ldp should be NULL. Returns true if the requested size
 * successfully bound otherwise false.
 */

static bool_t
mem_bind(ldom_t *ldp, uint64_t size, int binding_type, response_t *resp,
    mem_blocks_t *bindings)
{
	mblock_t *fmp, *next_fmp;
	int binding_num, segment_num, next_segment_num;
	uint_t max_bindings, max_segments;

	if (size == 0)
		return (false);

	CHAIN_INIT(bindings->blocks);

	binding_num = segment_num = next_segment_num = 0;

	if (binding_type == PHYSICAL_BINDING) {
		ASSERT(ldp == NULL);
		max_bindings = 1;
		max_segments = MAXINT; /* N/A to physical bindings */
	} else {
		max_bindings = MAXINT;

		max_segments = zeus.guest_max_mblocks;

		if (ldp->state == ACTIVE && in_delayed_reconfig(ldp) == false) {
			/*
			 * We get here because of memory DR. For an active
			 * domain, max binding number = MAXINT. Don't bind
			 * any more memory if the binding num already reaches
			 * the max.
			 */
			ASSERT(binding_type == VIRTUAL_BINDING);
			binding_num = get_seg_count(ldp);
			if (binding_num >= max_bindings) {
				rprintf(resp, MSG_ERR_BINDING_LIMIT,
				    max_bindings);
				return (false);
			}
		}
	}

	/*
	 * Loop through free mblocks until we find enough memory for the LDom.
	 * There is a minimum size for a block to be considered for allocation
	 * to an LDom. Other than that, this is a simplistic, first-fit
	 * algorithm.
	 */

	/*
	 * Can't use CHAIN_ITER, as bind_explicit() could delete entries
	 * off the free list.
	 */
	fmp = CHAIN_START(free_memory.blocks);

	while (fmp != NULL && binding_num < max_bindings) {
		uint64_t alignment, aligned_pa, aligned_size, raddr;
		uint64_t min_size, alloc_size;
		mblock_t *bound;

		next_fmp = CHAIN_NEXT(fmp->pblk);

		if (fmp->size < MEM_MINSIZE) {
			fmp = next_fmp;
			continue;
		}

		/*
		 * Set minimum block size, alignment and granularity
		 * restrictions. Force 256MB alignment
		 */

		if (binding_type == PHYSICAL_BINDING) {
			min_size = MEM_MINSIZE;
			alignment = 0;
		} else {
			alignment = MEM_MINSIZE;

			/*
			 * The first mblock in a virtual binding has
			 * additional restrictions imposed on it by the chosen
			 * boot image: a (stricter) minimum allocation size
			 * and and a PA alignment restriction. Note that since
			 * we keep free mblocks ordered by increasing base PA,
			 * see free_mblock(), the first mblock will have the
			 * lowest PA.
			 */
			if (binding_num == 0)
				min_size = ldp->boot_imagep->min_allocation;
			else
				min_size = PAGESIZE;
		}


		/* Enforce alignment restriction */
		aligned_pa = granularity_multiple(fmp->paddr, alignment);
		aligned_size = fmp->paddr + fmp->size - aligned_pa;

		/* Enforce minimum size restriction */
		if (min_size > 0) {
			if (size < min_size) {
				/*
				 * Increase allocation to match min_size
				 */
				DBG(zprintf("%s: increasing memory allocation "
				    "to match size restriction of 0x%llx",
				    __func__, min_size));
				if (binding_num == 0)
					allocation_warn(size, min_size, resp);

				size = min_size;
			}

			if (aligned_size < min_size) {
				/*
				 * Alignment restriction can't be met
				 * by this block due to min_size restriction.
				 */
				fmp = next_fmp;
				continue;
			}
		}

		if (!handle_granularity(aligned_size, &size, &alloc_size)) {
			fmp = next_fmp;
			continue;
		}

		/* max allocation size to fit in one segment */
		alloc_size = MIN(alloc_size, 1LL << zeus.ra2pa_shift);

		if (binding_type == VIRTUAL_BINDING) {
			if (ldp->state == ACTIVE)
				segment_num = (uint_t)get_seg_num(ldp,
				    alloc_size, aligned_pa);
			else
				segment_num = next_segment_num;

			if (segment_num >= max_segments) {
				if (resp != NULL)
					rprintf(resp, MSG_ERR_SEGMENT_LIMIT,
					    max_segments);
				mem_unbind_all(ldp);
				return (false);
			}

			raddr = get_raddr(aligned_pa, segment_num);
#if OBP_RA_ERRATA
			if (ldp->state != ACTIVE) {
				/*
				 * Segment 0 is a special case. Because OBP
				 * does not handle a segment with raddr = 0,
				 * the raddr for segment 0 is set to
				 * pri.max_pagesize. In order to keep the
				 * base address adjustment from causing the
				 * allocation to spill into another RA2PA
				 * setment, this segment's max size =
				 * (1LL << zeus.ra2pa_shift) - pri.max_pagesize.
				 */
				if (raddr == pri.max_pagesize &&
				    alloc_size >= (1LL << zeus.ra2pa_shift))
					alloc_size -= pri.max_pagesize;
			}
#endif

			/*
			 * Calculate next available segment number.
			 */
			next_segment_num = ((raddr + alloc_size) >>
			    zeus.ra2pa_shift);
			if (next_segment_num == segment_num)
				next_segment_num++;
		}

		/*
		 * Check if this is the last possible mblock we can bind,
		 * whether it's because we've hit the limit on the number of
		 * segments for virtual (real) bindings, or we've reached the
		 * limit on the number of physical mblock bindings.
		 */
		if ((binding_type == VIRTUAL_BINDING &&
		    next_segment_num >= max_segments) ||
		    (binding_type == PHYSICAL_BINDING &&
		    binding_num + 1 == max_bindings)) {
			if (alloc_size < size) {
				/*
				 * This is the last available segment we can
				 * assign; don't accept an allocation less than
				 * the remaining requirement.
				 */
				fmp = next_fmp;
				continue;
			}
		}


		/* generate new block binding */
		bound = allocate_mem(fmp, aligned_pa, alloc_size);
		if (bound == NULL) {
			fmp = next_fmp;
			continue;
		}

		if (binding_type == VIRTUAL_BINDING) {
			bound->mblk_resid = alloc_res(&zeus.free_mblk_ids);
			if (bound->mblk_resid == NULL)
				fatal("Ran out of mblock slots!");

			bound->raddr = raddr;

			bound->ldp = ldp;
			check_mem_addr(ldp, raddr, alloc_size);
			CHAIN_ADD(ldp->bound.memory, bound, ldom);
			ASSERT(bound->size % MEM_MINSIZE == 0);
		} else {
			/* physical binding */
			CHAIN_ADD(system_memory.blocks, bound, sys);
		}

		CHAIN_ADD(bindings->blocks, bound, blk);

		if ((size -= alloc_size) == 0)
			break;

		/*
		 * Re-fetch fmp from the free memory list
		 * The original free block (fmp) is possibly unlinked and freed
		 * by allocate_mem() and a new block(s) is possibly added in
		 * case of excess memory.
		 */
		fmp = CHAIN_START(free_memory.blocks);
		++binding_num;
	}

	if (size != 0) {
		/* unable to satisfy binding with current constraints */
		if (binding_type == VIRTUAL_BINDING) {
			if (resp != NULL) {
				if (next_segment_num >= max_segments) {
					rprintf(resp, MSG_ERR_SEGMENT_LIMIT,
					    max_segments);
					zprintf(MSG_ERR_SEGMENT_LIMIT,
					    max_segments);
				} else {
					rprintf(resp, MSG_ERR_NO_ENOUGH_MEM);
					zprintf(MSG_ERR_NO_ENOUGH_MEM);
				}

			}
			mem_unbind_all(ldp);
		}

		return (false);
	}

	return (true);
}


/*
 * Bind a given size of physical memory. Returns the bound memory block or
 * NULL if the binding could not be satisfied.
 */
mblock_t *
mem_bind_phys(uint64_t size)
{
	mem_blocks_t bindings;

	if (mem_bind(NULL, size, PHYSICAL_BINDING, NULL, &bindings)) {
		return (CHAIN_START(bindings.blocks));
	} else {
		return (NULL);
	}
}

/*
 * Bind to a LDom a given size of physical memory. Returns true if successful
 * otherwise false.
 */
bool_t
mem_bind_virt(ldom_t *ldp, uint64_t size, response_t *resp)
{
	mem_blocks_t bindings;

	return (mem_bind(ldp, size, VIRTUAL_BINDING, resp, &bindings));
}

/*
 * Size of bound memory in a ldom
 */
uint64_t
ldom_size_bound_mem(ldom_t *ldp)
{
	mblock_t *mblk;
	uint64_t memsize = 0;

	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		memsize += mblk->size;
	}

	return (memsize);
}

bool_t
mem_gen_hvmd_nodes(ldom_t *ldp, md_t *mdp, md_node_t *memsp, md_node_t *guestp)
{
	mblock_t *mbp;

	CHAIN_ITER(ldp->bound.memory, mbp, ldom) {
		md_node_t *memnode;

		memnode = md_link_new_node(mdp, "mblock", memsp, "fwd", "back");
		md_add_arc_property(mdp, guestp, "fwd", memnode);
		md_add_arc_property(mdp, memnode, "back", guestp);

		md_add_value_property(mdp, memnode, "membase", mbp->paddr);
		md_add_value_property(mdp, memnode, "memsize", mbp->size);
		md_add_value_property(mdp, memnode, "realbase", mbp->raddr);
		md_add_value_property(mdp, memnode, "resource_id",
		    mbp->mblk_resid->id);
	}

	return (true);
}

/*
 * Free cloned bindings. This just frees storage. All resource ids must
 * already have been freed or reused.
 */
void
mem_free_cloned_bindings(mem_blocks_t *clones)
{
	mblock_t *mblk, *next;

	for (mblk = CHAIN_START(clones->blocks); mblk != NULL; mblk = next) {
		next = CHAIN_NEXT(mblk->blk);
		CHAIN_UNLINK(clones->blocks, mblk, blk);
		Xfree(mblk);
	}
}

static void
clone_a_binding(mem_blocks_t *clones, mblock_t *mblk)
{
	mblock_t *clone;

	clone = Xmalloc(sizeof (mblock_t));
	clone->raddr = mblk->raddr;
	clone->paddr = mblk->paddr;
	clone->size = mblk->size;
	/*
	 * Save the resource id number. The mblk_resid pointer cannot just be
	 * copied as it will get freed in unbind_existing().
	 */
	clone->id = mblk->mblk_resid->id;
	CHAIN_ADD(clones->blocks, clone, blk);
}

/*
 * Clone LDom 'ldp's current memory bindings. This includes the memory block
 * resource IDs so that upon restoring the bindings the same ID's can be used
 * and no MD updates will be needed.
 */
void
mem_clone_bindings(ldom_t *ldp, mem_blocks_t *clones, mem_blocks_t *committed)
{
	mblock_t *mblk;

	CHAIN_INIT(clones->blocks);

	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		clone_a_binding(clones, mblk);
	}

	if (committed != NULL && ldp->flag.config_committed == true) {
		mem_free_cloned_bindings(committed);
		CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
			clone_a_binding(committed, mblk);
		}
	}
}

/*
 * Pull the memory resource ids of cloned bindings out of the
 * resource pool.
 */
static void
pull_clone_resids(mem_blocks_t *clones)
{
	mblock_t *mblk;

	CHAIN_ITER(clones->blocks, mblk, blk) {

		mblk->mblk_resid = pull_res_id(&zeus.free_mblk_ids, mblk->id);
		ASSERT(mblk->mblk_resid != NULL);
	}
}

/*
 * Return the memory resource ids of cloned bindings to the resource pool.
 */
static void
free_clone_resids(mem_blocks_t *clones)
{
	mblock_t *mblk;

	CHAIN_ITER(clones->blocks, mblk, blk) {
		free_res(mblk->mblk_resid);
		mblk->mblk_resid = NULL;
	}
}

/*
 * Restore LDom 'ldp's memory bindings to the cloned bindings saved in
 * 'clones' and then free 'clones'.
 */
static void
restore_bindings(ldom_t *ldp, mem_blocks_t *clones, mem_blocks_t *committed)
{
	if (committed != NULL) {
		/* Revert to cloned resid_t's */
		free_clone_resids(committed);
		pull_clone_resids(clones);
	}

	if (!mem_bind_virt_explicit(ldp, clones)) {
		/*
		 * Rebinding should never fail. If we get here there is a bug
		 * and the best that can be done is to abort and the memory
		 * bindings will be reestablished upon restart.
		 */
		fatal("internal error restoring memory bindings");
	}

	mem_free_cloned_bindings(clones);
}

/*
 * HV scrub and HV mem_dr_verify can be only done after HV reconfig.
 * Sequence of the operation is:
 * 1. unbind all unconfigured span
 * 2. HV reconfig
 * 3. HV scrub and verify
 * During unbind_span, we need to save paddr, raddr and size for HV scrub
 * and verify.
 */
void
mem_alloc_scrub_mblks(ldom_t *ldp, int nmblks)
{
	if (nmblks == 0)
		fatal("bad mem alloc scrub mblks\n");

	ldp->mem_dr_state.scrub.unbind_mblks = Xcalloc(nmblks, mblock_t);
	ldp->mem_dr_state.scrub.max_nmblks = nmblks;
	ldp->mem_dr_state.scrub.nmblks = 0;
}

static void
save_mblk_for_scrub(ldom_t *ldp, uint64_t paddr, uint64_t raddr, uint64_t size)
{
	mblock_t *mblkp;
	if (ldp->mem_dr_state.scrub.nmblks ==
	    ldp->mem_dr_state.scrub.max_nmblks)
		fatal("internal error max allocation of scrub mblks reached\n");
	mblkp = &((mblock_t *)ldp->mem_dr_state.scrub.unbind_mblks)
	    [ldp->mem_dr_state.scrub.nmblks];
	mblkp->raddr = raddr;
	mblkp->paddr = paddr;
	mblkp->size = size;
	ldp->mem_dr_state.scrub.nmblks++;
}

/*
 * Match addr, size in ldp bound memory list and unbind the span
 * Mem add calls this function if for some reason, not all allocated memory
 * was added to the domain. However, because the memory was scrubbed already
 * during the allocation, no need to scrub again. (do_scrub = false)
 *
 * Mem delete calls this function if there is memory unconfigured by the
 * domain. Unconfigured memory needs to be scrubbed and verified by HV.
 * (do_scrub = true).
 *
 * If do_scrub == true, we need to store paddr, raddr and size since the
 * information of paddr based on raddr won't exist after unbind. So we
 * saved the list of (paddr, raddr, size) to ldp->mem_dr_state.
 */
uint64_t
unbind_span(ldom_t *ldp, uint64_t addr, uint64_t size, bool_t do_scrub)
{
	mblock_t *mblk;
	uint64_t slo;	/* remaining mem at low end */
	uint64_t shi;	/* remaining mem at high end */
	uint64_t paddr;

	DBG(zprintf("%s unbind_span addr=0x%"PRIx64
	    ", size=%"PRIx64"\n", ldp->namep, addr, size));

	/* find & remove the span */
	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		uint64_t maddr = mblk->raddr;
		uint64_t msize = mblk->size;
		pri_mblock_t *pmp = mblk->pmblk;

		/* skip the mblock if it does not intersect with the span */
		if ((addr < maddr) || (addr >=  maddr + msize))
			continue;

		/* verify that the mblock completely contains the span */
		ASSERT(addr >= maddr && addr + size <= maddr + msize);

		/*
		 * There are two cases to be considered:
		 * 1. All memory in a mblk needs to be to be removed -
		 *    unbind the mblock.
		 * 2. Part(s) of the memory in a mblk needs to be removed:
		 *    - resize the mblk (no splitting mblk needed)
		 *    - split the mblk into 3 mblks unbind the middle one and
		 *    resize both lower and upper mblks
		 */
		if (size == msize) {
			ASSERT(addr == maddr);

			DBG(zprintf(
			    "Unbind resid=%"PRId64" ra=0x%"PRIx64" pa=0x%"
			    PRIx64" sz=0x%"PRIx64" \n", mblk->mblk_resid->id,
			    mblk->raddr, mblk->paddr, mblk->size));

			/* unlink mblock from both PRI mblock & LDom chains */
			CHAIN_UNLINK(pmp->bound.memory, mblk, pblk);
			CHAIN_UNLINK(ldp->bound.memory, mblk, ldom);
			SANITY(CHAIN_LINK_INIT(mblk->ldom));

			/* free mblock resource ID */
			free_res(mblk->mblk_resid);

			/* so free_mblock doesn't throw an assertion */
			mblk->ldp = NULL;
			if (do_scrub)
				save_mblk_for_scrub(ldp, mblk->paddr,
				    addr, size);
			mem_free_mblock(mblk);
			return (size);
		}

		ASSERT(size < msize);

		slo = addr - maddr;
		shi =  (maddr + msize) - (addr + size);
		paddr = mblk->paddr;

		ASSERT(slo || shi);

		if (slo && !shi) {
			/* free memory at the higher end */
			mblk->size = slo;

			/* create a new mlock for the free memory */
			ASSERT(slo + size == msize);
			mblk = mem_new_mblock(pmp, paddr + slo, size);
			DBG(zprintf("Free pa=0x%"PRIx64", sz=%"PRId64"MB \n",
			    mblk->paddr, BYTES2MB(mblk->size)));
			if (do_scrub)
				save_mblk_for_scrub(ldp, paddr + slo,
				    addr, size);
			mem_free_mblock(mblk);
			return (size);
		}


		if (shi && !slo) {
			/* free memory at the lower end */
			mblk->paddr = paddr + msize - shi;
			mblk->raddr = maddr + msize - shi;
			mblk->size = shi;

			DBG(zprintf("Free pa=0x%"PRIx64", sz=%"PRId64"MB \n",
			    mblk->paddr, BYTES2MB(mblk->size)));

			/* create a new mlock for the free memory */
			ASSERT(shi + size == msize);
			mblk = mem_new_mblock(pmp, paddr, size);
			if (do_scrub)
				save_mblk_for_scrub(ldp, paddr, addr, size);
			mem_free_mblock(mblk);
			return (size);
		}

		/*
		 * Case of unconfigured memory in the middle.
		 * Although we skip the mblk if a (addr, size) pair has reloc
		 * memory in middle, a large size of  mblk could be split
		 * into smaller mblks, whose upper or lower
		 * end could be unconfigured. Relative to mblk in ldp, those
		 * unconfigured spans are in the middle of mblk.
		 */
		ASSERT(shi && slo);
		ASSERT(addr % MEM_MINSIZE == 0);
		ASSERT(addr == mblk->raddr + slo);
		ASSERT(mblk->raddr + mblk->size == mblk->raddr + slo +
		    size + shi);

		/*
		 * Free memory at both lower and upper ends
		 * The mblk  is split into 3 spans. The middle span is
		 * deallocated and return to the free memory pool while the
		 * lower and upper spans remain in the domain.
		 */

		/* resize the mblk as lower span */
		mblk->size = slo;

		/* add an upper span to the ldom */
		mblk = mem_new_mblock(pmp, paddr + slo + size, shi);
		mblk->raddr = addr +  size;

		mblk->mblk_resid = alloc_res(&zeus.free_mblk_ids);
		if (mblk->mblk_resid == NULL)
			fatal("Ran out of mblock slots!");
		mblk->ldp = ldp;
		mblk->flag_isconfigd = true;
		CHAIN_ADD(ldp->bound.memory, mblk, ldom);
		CHAIN_ADD(pmp->bound.memory, mblk, pblk);


		ASSERT(mblk->raddr+mblk->size == maddr + msize);

		/* add a free mblk */
		mblk = mem_new_mblock(pmp, paddr + slo, size);
		if (do_scrub)
			save_mblk_for_scrub(ldp, mblk->paddr, addr, mblk->size);
		mem_free_mblock(mblk);
	}
	return (size);
}

/*
 * Unbind memory that was previously bound, but was not
 * successfully configured. This could happen when the domain for some
 * reason failed to add allocated memory.
 */
void
mem_unbind_unconfigured(ldom_t *ldp, int nmblks, void *mblks)
{
	int i;
	uint64_t size = 0;
	dr_mblk_t *f_blks = (dr_mblk_t *)mblks;

	DBG(zprintf("mem_unbind_unconfigured %"PRId16"\n", nmblks));

	for (i = 0; i < nmblks; i++) {
		unbind_span(ldp, f_blks[i].addr, f_blks[i].size, false);
		size += f_blks[i].size;
	}

	ldp->required.size_mem -= size;
}

/*
 * Error message(s) for memory DR.
 */
static void
mem_dr_msg(ldom_t *ldp, uint64_t size, int rv, response_t *rsp)
{

	/*
	 * When ldp->mem_dr_state.size != 0, we calculate the size according
	 * to the CLI unit as part of error message.
	 */
	if (ldp->mem_dr_state.size == 0) {
		/*
		 * If the DR failed, the size of memory DR should be part
		 * of error message.
		 */
		if (rv == ERR_MEM_DR_OP_FAILED)
			ldp->mem_dr_state.size = size;
	}

	if (ldp->mem_dr_state.size > 0) {
		switch (ldp->mem_dr_state.unit) {
		case 'G':
		case 'g':
			ldp->mem_dr_state.size =
			    BYTES2MB(ldp->mem_dr_state.size);
			if (ldp->mem_dr_state.size % 1024  == 0)
				ldp->mem_dr_state.size /= 1024;
			else
				ldp->mem_dr_state.unit = 'M';
			break;

		case 'M':
		case 'm':
			ldp->mem_dr_state.size =
			    BYTES2MB(ldp->mem_dr_state.size);
			break;

		case 'K':
		case 'k':
			ldp->mem_dr_state.size =
			    BYTES2KB(ldp->mem_dr_state.size);
			break;

		default:
			break;
		}
	}

	/*
	 * If there is auto_adj_size and the DR is successful,
	 * the user should be notified.
	 */
	if (ldp->mem_dr_state.auto_adj_size > 0 &&
	    (rv == MEM_DR_OP_SUCC || rv == ERR_MEM_DR_OP_PARTIAL)) {
		rprintf(rsp, MSG_WARN_MEMORY_CONSTRAINS, ldp->namep,
		    BYTES2MB(ldp->mem_dr_state.auto_adj_size));
	}

	switch (rv) {
	case MEM_DR_OP_SUCC:
		return;

	case ERR_MEM_DR_OP_NO_RESPONSE:
		rprintf(rsp, MSG_ERR_MEM_DR_NO_RESP,
		    (ldp->mem_dr_state.is_mem_add ? "add" : "remove"));
		break;

	case ERR_MEM_DR_VERIFY_FAILED:
		rprintf(rsp, MSG_ERR_MEM_DR_VERIFY_FAILED, ldp->namep);
		break;

	case ERR_MEM_DR_SCRUB_FAILED:
		rprintf(rsp, MSG_ERR_MEM_DR_SCRUB_FAILED);
		break;

	case ERR_MEM_DR_OP_PARTIAL:
		if (ldp->mem_dr_state.is_mem_add == true) {
			rprintf(rsp, MSG_ERR_MEM_DR_PARTIAL_ADD,
			    ldp->mem_dr_state.size,
			    ldp->mem_dr_state.unit, ldp->namep);
		} else {
			rprintf(rsp, MSG_ERR_MEM_DR_PARTIAL_DEL,
			    ldp->mem_dr_state.size,
			    ldp->mem_dr_state.unit, ldp->namep);
		}
		return;

	case ERR_MEM_DR_OP_CANCELLED:
		rprintf(rsp, MSG_ERR_MEM_DR_CANCELLED);
		if (ldp->mem_dr_state.size > 0) {
			rprintf(rsp, MSG_ERR_MEM_DR_PARTIAL_CANCEL,
			    ldp->mem_dr_state.size, ldp->mem_dr_state.unit,
			    ldp->namep);
		}
		break;

	case ERR_MEM_DR_OP_DOMAIN_FAILED:
		rprintf(rsp, MSG_ERR_MEM_DR_DOMAIN_FAILED,
		    ldp->mem_dr_state.status);
		break;

	case ERR_MEM_DR_OP_FAILED:
		rprintf(rsp, MSG_ERR_MEM_DR_OP_FAILED, ldp->namep,
		    (ldp->mem_dr_state.is_mem_add ? "Addition" : "Deletion"),
		    ldp->mem_dr_state.size,
		    ldp->mem_dr_state.unit);
		break;

	default:
		break;
	}

	if (ldp->mem_dr_state.is_mem_add == true)
		rprintf(rsp, MSG_ERR_MEM_DR_ADD_FAILED);
	else
		rprintf(rsp, MSG_ERR_MEM_DR_REMOVE_FAILED);
}

/*
 * Send delete memory request to the domain
 * This function is used outside of sequencer.
 */
static int
mem_req_del(ldom_t *ldp, uint64_t size, uint_t *nmblk, void *mblkp)
{
	int rv;
	dr_mem_ufg_f dr_mem_unconfigure;

	dr_mem_unconfigure = (dr_mem_ufg_f)ds_service_routine(ldp->ds_chan,
	    DR_MEM_SVC, "dr_mem_unconfigure");

	/* mem_del has pinged memory dr service alive */
	ASSERT(dr_mem_unconfigure != NULL);

	/* set mem dr flag in MD */
	ldp->mem_dr_state.inprogress = true;
	seq_enqueue_work(seq_update_guest_mds, (void *)ldp, true);
	rv = (*dr_mem_unconfigure)(ldp, size, nmblk, mblkp);

	return (rv);
}

/*
 * Send memory add request to an domain
 * the LDom is active and extra memory has been added. Hypervisor have been
 * re-configured, we request domain to configure those memory.
 * returns:
 *  failed_mblkp - addr, size pair of memory did not configure into the domain
 *	the calling routine needs to free failed_mblkp.
 *  n_fail - number of mblk did not configure into the domain
 */
static int
mem_req_add(ldom_t *ldp, uint64_t size, uint_t *n_fail, void **failed_mblkp)
{
	int rv, nmblk;
	dr_mblk_t *mblkp;
	dr_mem_cfg_f dr_mem_configure;
	uint64_t *op_stats;

	*failed_mblkp = NULL;

	/*
	 * Get an array of the unconfigured mblks, which is done before we
	 * release sequencer. The list contains newly allocated memory for
	 * the domain. the memory size of each entry, is less then 16G already.
	 */
	nmblk = ldp->mem_dr_state.scrub.nmblks;
	mblkp = (dr_mblk_t *)ldp->mem_dr_state.scrub.unbind_mblks;

	*n_fail = 0;

	ASSERT(nmblk);

	dr_mem_configure = (dr_mem_cfg_f)ds_service_routine(ldp->ds_chan,
	    DR_MEM_SVC, "dr_mem_configure");

	/* mem_add has pinged memory dr service alive */
	ASSERT(dr_mem_configure != NULL);

	op_stats = Xcalloc(nmblk, uint64_t);

	ldp->mem_dr_state.size = size;

	rv = (*dr_mem_configure)(ldp, nmblk, mblkp, op_stats);
	if (rv == ERR_MEM_DR_OP_FAILED) {
		/* none mblk configured - failed mblk = allocated mblk */
		*n_fail = nmblk;
		*failed_mblkp = mblkp;
	} else  if (rv == ERR_MEM_DR_OP_PARTIAL) {
		/*
		 * Count the number of failed cases and allocate
		 * an array to hold them
		 */
		int i;
		uint64_t sz;
		dr_mblk_t *f_mblkp;

		*failed_mblkp = f_mblkp = Xcalloc(nmblk, dr_mblk_t);

		/*
		 * Select the mblks to rollback
		 */
		sz = 0;
		*n_fail = 0;
		for (i = 0; i < nmblk; i++) {
			if (op_stats[i] == DR_MEM_RES_OK) {
				continue;
			}
			sz += mblkp[i].size;
			/* this mblk has failed to configure */
			f_mblkp[*n_fail] = mblkp[i];
			(*n_fail)++;
		}

		zprintf("n_fail=%"PRId8" size=%"PRIx64"\n", n_fail,
		    sz);

		ASSERT(*n_fail > 0);
		ASSERT(sz > 0);

		if (sz > 0)
			ldp->mem_dr_state.size = sz;
		if (sz == size)
			rv = ERR_MEM_DR_OP_FAILED;
		DBG(zprintf("%s %"PRId64" MB of requested %"PRId64
		    "MB memory has been added\n", ldp->namep,
		    BYTES2MB(size - sz), BYTES2MB(size)));
		Xfree(mblkp);
	} else {
		Xfree(mblkp);
	}

	Xfree(op_stats);
	DBG(zprintf("seq_add_mem returned %"PRId8"\n", rv));

	return (rv);
}

/*
 * Memory DR Thread
 * This is the portion of Memory DR out of sequencer
 * (see comments before mem-add).
 */
static void *
do_mem_dr(void *args)
{
	ldom_t	    *ldp = ((mem_dr_args_t *)args)->ldp;
	uint64_t    size = ((mem_dr_args_t *)args)->size;
	cli_conn_t  *client = ((mem_dr_args_t *)args)->client;
	int	    rv;
	response_t  *resp = client->resp;
	seq_mem_dr_arg_t arg;
	uint_t nmblk;
	dr_mblk_t *mblkp = NULL;

	Xfree(args);
	client->send_resp = false;

	/* send request to domain */
	if (ldp->mem_dr_state.is_mem_add)
		rv = mem_req_add(ldp, size, &nmblk, (void **)&mblkp);
	else
		rv = mem_req_del(ldp, size, &nmblk, (void **)&mblkp);

	arg.rv = rv;
	arg.ldp = ldp;
	arg.nmblk = nmblk;
	arg.mblkp = mblkp;
	arg.size = size;

	/*
	 * Handle result of memory dr from domain. Since we need to update
	 * free memory pool and MD, handling result must be done inside
	 * sequencer.
	 */
	if (ldp->mem_dr_state.is_mem_add) {
		seq_enqueue_work(seq_chk_add_mem, (void *)&arg, true);
	} else {
		seq_enqueue_work(seq_chk_del_mem, (void *)&arg, true);
	}

	if (mblkp != NULL) {
		Xfree(mblkp);
	}

	rv = arg.rv;

	/* update required.size_mem */
	if (ldp->state == ACTIVE) {
		ldp->required.size_mem = ldom_size_bound_mem(ldp);
	}

	mem_dr_msg(ldp, size, rv, resp);

	/* send result back */
	if (rv == MEM_DR_OP_SUCC || rv == ERR_MEM_DR_OP_PARTIAL)
		xml_send_async_cmd_result(client,
		    ldp->mem_dr_state.is_mem_add ? XML_ADD_MEM : XML_REMOVE_MEM,
		    ldp->namep, CLI_RESULT_OK);
	else
		xml_send_async_cmd_result(client,
		    ldp->mem_dr_state.is_mem_add ? XML_ADD_MEM : XML_REMOVE_MEM,
		    ldp->namep, CLI_RESULT_FAIL);

	/*
	 * If rv = DR_MEM_OP_NO_RESPONSE, domain is in a unknown state
	 * we keep the lock. User needs to manually restart the domain,
	 * to clear the lock.
	 */
	if (rv != ERR_MEM_DR_OP_NO_RESPONSE) {
		ldp->flag.memdr_in_progress = false;
		ldp->flag.locked = false;
		database_ldom_set_memory(ldp, ldp->required.size_mem);
		event_send_res_event(ldp->mem_dr_state.event, ldp, NULL,
		    NULL, client);
		event_send_progress_event(MEMDR_PROGRESS, ldp, XML_COMPLETE);
		database_save();
	} else {
		zprintf("%s got NO RESPONSE\n", ldp->namep);
	}
	ldp->mem_dr_state.tid = NULL;

	return (NULL);
}

static bool_t
is_sparse_mem()
{
	return (zeus.ra2pa_shift != pri.ra2pa_shift);
}

static int
check_mem_dr_capable(ldom_t *ldp)
{

	/* check HV version */
	if (hv_is_mem_dr_supported() == false && !IS_CONTROL_DOMAIN(ldp))
		return (ERR_MEM_DR_HV_NOTSUP);

	if (is_sparse_mem())
		return (ERR_MEM_DR_SPARSE_MEM);

	if (in_delayed_reconfig(NULL) == true) {
		/*
		 * In delayed configuration, only if the domain is control
		 * domain, memory change is supported.
		 */
		if (in_delayed_reconfig(ldp) == false ||
		    !IS_CONTROL_DOMAIN(ldp)) {
			return (ERR_MEM_DR_DELAYED_RECONF);
		}
	}

	if (ldp->state  == ACTIVE) {

		if (IS_LDOM_LOCKED(ldp) == true)
			return (ERR_MEM_DR_OP_LOCKED);

		if (ds_lookup_svc(ldp->ds_chan, DR_MEM_SVC) == NULL) {
			return (ERR_MEM_DR_NOTSUP);
		}

		if (ds_lookup_svc(ldp->ds_chan, MD_UPDATE_SVC) == NULL) {
			return (ERR_MEM_DR_NOTSUP);
		}
	}

	return (MEM_DR_OP_SUCC);
}

int
cli_check_mem_dr_capable(ldom_t *ldp, response_t *resp)
{
	int rv = check_mem_dr_capable(ldp);

	switch (rv) {
	case MEM_DR_OP_SUCC:
		break;

	case ERR_MEM_DR_DELAYED_RECONF:
		rprintf(resp, MSG_ERR_MEM_DELAYED_RECONF_IN_PROGRESS,
		    zeus.hvctl_ldom_p->namep, zeus.hvctl_ldom_p->namep);
		break;

	case ERR_MEM_DR_OP_LOCKED:
		rprintf(resp, MSG_ERR_MEM_LDOM_LOCK, ldp->namep);
		break;

	case ERR_MEM_DR_HV_NOTSUP:
		rprintf(resp, MSG_ERR_MEM_DR_HV_DOWNREV, ldp->namep);
		break;

	case ERR_MEM_DR_NOTSUP:
		rprintf(resp, MSG_ERR_MEM_DR_NOTSUP, ldp->namep);
		break;

	case ERR_MEM_DR_UNALIGN_SIZE:
		rprintf(resp, MSG_ERR_MEM_SIZE_UNALIGN, BYTES2MB(MEM_MINSIZE));
		break;

	case ERR_MEM_DR_SPARSE_MEM:
		rprintf(resp, MSG_ERR_MEM_DR_SPARSE_MEM, ldp->namep);
		break;

	default:
		fatal("internal error:unknown mem dr error\n");
		break;
	}
	return (rv);
}

/*
 * Cancel a memory deletion
 * Only memory deletion can be cancelled.
 * Memory add does not support cancel.
 */
bool_t
mem_cancel_dr(ldom_t *ldp, response_t *resp)
{
	int rv;
	dr_mem_cancel_f dr_mem_cancel;

	if (ldp->flag.memdr_in_progress == false) {
		rprintf(resp, MSG_ERR_MEM_DR_NOT_IN_PROGRESS, ldp->namep);
		return (false);
	}

	if (ldp->mem_dr_state.is_mem_add) {
		rprintf(resp, MSG_ERR_MEM_DR_CANNOT_CANCEL_ADD);
		return (false);
	}

	dr_mem_cancel = (dr_mem_cancel_f)ds_service_routine(ldp->ds_chan,
	    DR_MEM_SVC, "dr_mem_cancel");

	if (dr_mem_cancel == NULL) {
		rprintf(resp, MSG_ERR_MEM_DR_NOTSUP, ldp->namep);
		return (false);
	}

	/* send cancel request to the domain */
	rv = (*dr_mem_cancel)(ldp);

	if (rv == MEM_DR_OP_SUCC)
		return (true);

	rprintf(resp, MSG_ERR_MEM_DR_DONE_CANNOT_CANCEL);
	return (false);

}

/*
 * Memory DR has 3 portions:
 * 1. Inside sequencer
 *  A memory DR request starts in the sequencer, which is single threaded.
 *  For memory add, we allocate memory and update MD.
 * 2. Out of sequencer
 *  Since a domain may take a long time to do memory DR, both mem_add and
 *  mem_del create separate threads to send memory DR request to the domain and
 *  release the sequencer.
 * 3. After the domain completes the request, both add and delete operation
 * need to update MD. To allow only one thread to modify MD, we inserted the
 * process of post domain request into the sequencer's working queue.
 * For memory add, the process includes unbind mblks were not added and update
 * MD. For memory deletion, the process includes unbind deleted mblks and
 * update MD.
 */
bool_t
mem_add(ldom_t *ldp, uint64_t  size, void *clientp)
{
	cli_conn_t *client = (cli_conn_t *)clientp;
	response_t *resp = client->resp;

	if ((size % MEM_MINSIZE) != 0) {
		rprintf(resp, MSG_ERR_MEM_SIZE_UNALIGN, BYTES2MB(MEM_MINSIZE));
		return (false);
	}

	ldp->required.size_mem += size;

	if (can_use_memdr(ldp) == true) {
		mem_dr_args_t *args;

		/* bind additional memory */
		if (!mem_bind_virt(ldp, size, resp)) {
			ldp->required.size_mem = ldom_size_bound_mem(ldp);
			return (false);
		}

		/*
		 * The indication of unconfigured mblks is the configd flag in
		 * each mblk structure. This flag can be changed by another
		 * thread via seq_update_guest_mds because seq_update_guest_mds
		 * updates MD of all domains.
		 * To prevent this race condition, we store unconfigured mblks
		 * in mem_dr_state when we are still in the sequencer.
		 */
		ldp->mem_dr_state.scrub.nmblks = seq_get_unconfigured_mblks(ldp,
		    &ldp->mem_dr_state.scrub.unbind_mblks);
		if (ldp->mem_dr_state.scrub.nmblks <= 0)
			return (false);

		/*
		 * Solaris expects the memory to be dynamically added to
		 * be present in a distinct mblock. For example, we can't have
		 * a 2G mblock in the MD and ask Solaris to dynamically add 1G
		 * of that mblock, that's why we don't want adjacent mblocks to
		 * be coalesced in the MD. This shouldn't preclude mblocks to
		 * be coalesced in the free pool but it does mean use of
		 * additional ra2pa segments.
		 */

		seq_dr_mem_md(ldp);

		client->send_resp = false;
		ldp->flag.locked = true;
		ldp->mem_dr_state.is_mem_add = true;
		ldp->flag.memdr_in_progress = true;


		/*
		 * Create a thread to send request and release the
		 * sequencer.
		 */
		args = Xmalloc(sizeof (mem_dr_args_t));
		args->ldp = ldp;
		args->size = size;
		args->client = client;

		if (thr_create(NULL, NULL, do_mem_dr, args, THR_DETACHED,
		    &(ldp->mem_dr_state.tid)) != 0)
			efatal("Could not create thread to add memory");
		return (true);
	}

	if (ldp->state >= BOUND) {
		mem_blocks_t current_bindings;

		/*
		 * Save current bindings so that they can be restored in case
		 * of failure to add memory.
		 */
		mem_clone_bindings(ldp, &current_bindings,
		    &ldp->bound.committed_memory);

		/*
		 * Unbind current bindings and then rebind the newly requested
		 * amount. The current bindings are unbound so that some
		 * defragmentation can occur.
		 */
		size = ldp->required.size_mem;

		mem_unbind_all(ldp);

		/*
		 * The new memory binding resource ids must be distinct from
		 * the committed ones, so they are pulled out of the free pool.
		 */
		pull_clone_resids(&ldp->bound.committed_memory);


		/* rebind memory */
		if (!mem_bind_virt(ldp, size, resp)) {
			restore_bindings(ldp, &current_bindings,
			    &ldp->bound.committed_memory);
			/* binding failed, restore required memory */
			ldp->required.size_mem = ldom_size_bound_mem(ldp);
			return (false);
		}
		free_clone_resids(&ldp->bound.committed_memory);
		mem_free_cloned_bindings(&current_bindings);

		if (zeus.mode == ZEUS_MGMT_MODE) {
			if (ldp->state != BOUND) {
				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_ADD, resp);
			} else {
				/* do reconfig now */
				seq_download_mds_do_config(ldp);
			}
		}
	}

	database_ldom_set_memory(ldp, ldp->required.size_mem);

	return (true);
}

bool_t
mem_del(ldom_t *ldp, uint64_t  size, void *clientp)
{
	cli_conn_t *client = (cli_conn_t *)clientp;
	response_t *resp = client->resp;

	if ((size % MEM_MINSIZE) != 0) {
		rprintf(resp, MSG_ERR_MEM_SIZE_UNALIGN, BYTES2MB(MEM_MINSIZE));
		return (false);
	}

	/*
	 * Check that ldp has min of MEM_MINSIZE of memory.
	 * If ldp->required.size_mem - size <  MEM_MINSIZE return false.
	 * to avoid the negative, use MEM_MINSIZE + size >
	 * ldp->required_size_mem.
	 */
	if (MEM_MINSIZE + size > ldp->required.size_mem) {
		char *s = conv_size(size);
		if (size == ldp->required.size_mem ||
		    ldp->required.size_mem - size < MEM_MINSIZE)
			rprintf(resp, MSG_ERR_MEM_DR_CANNOT_DEL_ALL_MEM, s);
		else {
			char *s1 = conv_size(ldp->required.size_mem);
			rprintf(resp, MSG_ERR_MEM_DR_CANNOT_DEL_MORE, s, s1);
			Xfree(s1);
		}

		Xfree(s);
		return (false);
	}

	if (can_use_memdr(ldp) == true) {
		mem_dr_args_t *args;

		ldp->flag.locked = true;
		ldp->flag.memdr_in_progress = true;
		ldp->mem_dr_state.is_mem_add = false;
		ldp->required.size_mem -= size;

		client->send_resp = false;
		args = Xmalloc(sizeof (mem_dr_args_t));
		args->ldp = ldp;
		args->size = size;
		args->client = client;

		/*
		 * Create a thread to send request and release the
		 * sequencer.
		 */
		if (thr_create(NULL, NULL, do_mem_dr, args, THR_DETACHED,
		    &ldp->mem_dr_state.tid) != 0)
			efatal("Could not create thread to del memory");
		return (true);
	}

	if (ldp->state >= BOUND) {
		mem_blocks_t current_bindings;

		/*
		 * Save current bindings so that they can be restored in case
		 * of failure to delete memory.
		 */
		mem_clone_bindings(ldp, &current_bindings,
		    &ldp->bound.committed_memory);

		ldp->required.size_mem -= size;
		size = ldp->required.size_mem;

		/*
		 * Unbind current bindings and then rebind the newly requested
		 * amount.
		 */
		mem_unbind_all(ldp);
		/*
		 * The new memory binding resource ids must be distinct from
		 * the committed ones, so they are pulled out of the free pool.
		 */
		pull_clone_resids(&ldp->bound.committed_memory);

		/* rebind memory */
		if (!mem_bind_virt(ldp, size, resp)) {
			restore_bindings(ldp, &current_bindings,
			    &ldp->bound.committed_memory);

			/* binding failed, restore required memory */
			ldp->required.size_mem = ldom_size_bound_mem(ldp);
			return (false);
		}

		free_clone_resids(&ldp->bound.committed_memory);
		mem_free_cloned_bindings(&current_bindings);

		if (zeus.mode == ZEUS_MGMT_MODE) {
			if (ldp->state != BOUND) {
				/*
				 * For a control domain there are two cases
				 * a delayed reconfig will be done:
				 * 1. In factory-default config - all
				 * memory allocated to control domain.
				 * 2. control domain does not support
				 * memory dr.
				 */
				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_REMOVE_PHYS, resp);
			} else {
				/* Do reconfig now */
				seq_download_mds_do_config(ldp);
			}
		}
	} else {
		if (size == ldp->required.size_mem) {
			rprintf(resp, "Warning: setting amount of "
			    "memory requested to zero\n");
		}
		ldp->required.size_mem -= size;
	}

	database_ldom_set_memory(ldp, ldp->required.size_mem);

	return (true);
}

/*
 * Finds and binds an area of free physical memory to LDom 'ldp' such that the
 * real address and size matches that of 'mblk'.
 */
static int
bind_real(ldom_t *ldp, mblock_t *mblk)
{
	mblock_t *freeblk, *bound;
	uint64_t pa_offset, ra_offset, pa_base;

	CHAIN_ITER(free_memory.blocks, freeblk, pblk) {
		/*
		 * Determine the physical address within the current free
		 * block which is as close as possible to the block's base
		 * address and which is congruent to the required real
		 * address.
		 */
		pa_offset = freeblk->paddr % pri.max_pagesize;
		ra_offset = mblk->raddr % pri.max_pagesize;

		if (pa_offset == ra_offset)
			pa_base = freeblk->paddr;
		else if (pa_offset < ra_offset)
			pa_base = freeblk->paddr + ra_offset - pa_offset;
		else
			pa_base = freeblk->paddr + pri.max_pagesize -
			    pa_offset + ra_offset;

		/*
		 * If the size of the memory from the physical address
		 * computed above to the top of the free block is large enough
		 * to hold the required block size then bind it to the LDom
		 * and return success.
		 */
		if (pa_base + mblk->size <= freeblk->paddr + freeblk->size) {
			bound = allocate_mem(freeblk, pa_base, mblk->size);
			bound->mblk_resid = alloc_res(&zeus.free_mblk_ids);
			if (bound->mblk_resid == NULL)
				return (ERR_MEM_NO_MBLOCK_IDS);
			bound->raddr = mblk->raddr;
			bound->ldp = ldp;

			CHAIN_ADD(ldp->bound.memory, bound, ldom);

			ASSERT((bound->paddr % pri.max_pagesize) ==
			    (bound->raddr % pri.max_pagesize));

			return (0);
		}
	}

	return (ERR_MEM_BIND_REAL);
}

/*
 * Binds the memory for LDom 'ldp' to match the real addresses and sizes of
 * the blocks described in 'bindings'. The block physical addresses may
 * change.
 */
bool_t
mem_bind_real(ldom_t *ldp, mem_blocks_t *bindings, response_t *resp)
{
	mblock_t	*mblk;

	CHAIN_ITER(bindings->blocks, mblk, blk) {
		if (bind_real(ldp, mblk) != 0) {
			rprintf(resp, MSG_ERR_NO_ENOUGH_MEM);
			return (false);
		}
	}

	return (true);
}

/*
 * Free memory
 */
static uint64_t
available_free_mem(void)
{
	mblock_t *mblk;
	uint64_t size = 0;

	CHAIN_ITER(free_memory.blocks, mblk, pblk) {
		size += mblk->size;
	}

	return (size);
}

/*
 * Only check whether the memory can be bound, for this we actually bind the
 * mblocks and then unbind all the allocated blocks
 */
bool_t
mem_bind_real_check(ldom_t *ldp, mem_blocks_t *bindings, response_t *resp)
{
	bool_t		rv = true;

	/*
	 * First check whether there is enough free memory
	 */
	if (ldp->required.size_mem > available_free_mem()) {
		rprintf(resp, gettext(
		    "Could not bind requested memory for LDom %s\n"),
		    ldp->namep);
		return (false);
	}

	/*
	 * If there is no existing binding, check if the required memory can
	 * actually be bound
	 */
	if (bindings == NULL) {
		/*
		 * Cold migration, no existing memory blocks
		 */
		rv = mem_bind_virt(ldp, ldp->required.size_mem, resp);
	} else {
		/*
		 * Warm migration
		 */
		rv = mem_bind_real(ldp, bindings, resp);
	}

	/*
	 * Now unbind all the allocated mblocks and return the result
	 */
	mem_unbind_all(ldp);

	return (rv);
}

/*
 * Reassign unaligned memory to control domain. For domains that configured
 * before memory DR, 256M alignment were not forced. Free memory pool
 * may contain unaligned memory, which cannot be used by memory DR.
 * Through delayed reconfiguration, we reassign unaligned memory to control
 * domain.
 * The following cases that this can happen:
 * 1. binding memory from an existing configuration that contains unaligned
 *   memory. (mem_bind_virt_explicit)
 * 2. If an existing domain, that contains unaligned memory, unbinds memory,
 *  unaligned memory will release to free memory pool. (mem_add & mem_del)
 * return true if there is unaligned memory assigned to primary domain
 * and return false otherwise.
 */
bool_t
mem_reassign_unaligned(response_t *resp)
{
	uint64_t aligned_pa;
	uint64_t aligned_size;
	uint64_t size;
	uint64_t unaligned_pa;
	uint64_t unaligned_size = 0;
	uint64_t alloc_size;
	int segment_num;
	bool_t	do_reconf = false;
	ldom_t *ldp;
	uint64_t raddr;
	mblock_t *bound;
	mblock_t *fmp, *fmp_next;

	/* check if there is any memory DR in progress */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->flag.memdr_in_progress == true) {
			/* disable assigning unaligned memory */
			rprintf(resp, MSG_ERR_MEM_ASSIGN_UNALIGN, ldp->namep);
			return (false);
		}
	}

	ldp = zeus.hvctl_ldom_p;
	/*
	 * Can't use CHAIN_ITER, as bind_explicit() could delete entries
	 * off the free list.
	 */
	fmp = CHAIN_START(free_memory.blocks);

	while (fmp != NULL) {

		fmp_next = CHAIN_NEXT(fmp->pblk);

		aligned_pa = granularity_multiple(fmp->paddr, MEM_MINSIZE);
		aligned_size = fmp->paddr + fmp->size - aligned_pa;
		size = fmp->size / MEM_MINSIZE * MEM_MINSIZE;
		unaligned_pa = 0;
		unaligned_size = 0;

		if (!handle_granularity(aligned_size, &size, &alloc_size)) {

			/* Entire free blk is not aligned */
			unaligned_pa = fmp->paddr;
			unaligned_size = fmp->size;

		} else if (aligned_pa == fmp->paddr) {

			if (alloc_size == fmp->size) {
				/* Entire free blk is aligned */
				fmp = fmp_next;
				continue;
			}

			ASSERT(alloc_size < fmp->size);
			/* Unaligned block is at the higher end */
			unaligned_pa = fmp->paddr + alloc_size;
			unaligned_size = fmp->size - alloc_size;

		} else {

			ASSERT(fmp->paddr < aligned_pa);
			ASSERT(aligned_size + aligned_pa ==
			    fmp->paddr + fmp->size);
			/*
			 * Unaligned block on the lower end and
			 * possibly at the upper end. However,
			 * deal with lower end only and next iterate
			 * will take care the upper end.
			 */
			unaligned_pa = fmp->paddr;
			unaligned_size = aligned_pa - fmp->paddr;
		}

		ASSERT(unaligned_size <= fmp->size);
		do_reconf = true;

		/* Add unaligned blk to control domain */
		segment_num = get_seg_num(ldp, unaligned_size, unaligned_pa);

		if (segment_num >= zeus.guest_max_mblocks - 1) {
			zprintf("Control domain run out of segments\n");
			return (false);
		}

		raddr = get_raddr(unaligned_pa, segment_num);

		bound = allocate_mem(fmp, unaligned_pa, unaligned_size);
		if (bound == NULL) {
			zprintf("Cannot bind unaligned memory\n");
			return (false);
		}

		bound->mblk_resid = alloc_res(&zeus.free_mblk_ids);
		if (bound->mblk_resid == NULL) {
			zprintf("Ran out of mblock slots!\n");
			return (false);
		}

		bound->raddr = raddr;
		bound->ldp = ldp;

		zprintf("add unaligned memory paddr=0x%"PRIx64" size=%"
		    PRId64" MB radd= x%"PRIx64" to %s \n", bound->paddr,
		    BYTES2MB(bound->size), bound->raddr, ldp->namep);
		CHAIN_ADD(ldp->bound.memory, bound, ldom);
		/* free memory list has changed, start over again */
		fmp = CHAIN_START(free_memory.blocks);
	}

	if (do_reconf == true) {
		initiate_or_update_delayed_reconfig(ldp,
		    DEL_RECONFIG_ADD, resp);
		return (true);
	}

	rprintf(resp, MSG_ERR_MEM_NO_UNALIGN_MEM);
	return (false);
}

/*
 * If there was a memory dr in progress before zeus is up, memory in HV/zeus
 * may not match memory in the domain. We need to sync HV/zeus memory with
 * the domain.
 * steps as the follows:
 * 1. lock all domains with memdr in progress before zeus crashed. This must
 *    be done in sequencer to avoid modify domain memory.
 * 2. create a thread to obtain domain's memory map. This is done outside
 *    sequencer to avoid long waiting on a particular domain.
 * 3. modify zeus memory map according to domain memory map. This is must be
 *    done in sequencer since memory may be removed and returned to the pool.
 */
/*
 * out of sequencer
 */
static void *
do_sync_mem(void *arg)
{
	ldom_t		*ldp;
	dr_get_mem_map_f dr_get_mem_map;
	seq_mem_dr_arg_t args;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->mem_dr_state.inprogress == true) {
			/* Get an array of the configured mblks from domain */
			ASSERT(ldp->ds_chan);
			dr_get_mem_map = (dr_get_mem_map_f)ds_service_routine(
			    ldp->ds_chan, DR_MEM_SVC, "dr_get_mem_map");
			if (dr_get_mem_map == NULL) {
				/* domain is gone */
				zprintf("domain %s does not have dr mem"
				    "service cannot do sync mem\n", ldp->namep);
				if (ldp->state != ACTIVE)
					mem_dr_cleanup(ldp);

				continue;
			}

			ldp->mem_dr_state.tid = thr_self();
			args.ldp = ldp;
			args.rv = (*dr_get_mem_map)(ldp, &args.nmblk,
			    &args.mblkp);
			ldp->mem_dr_state.tid = NULL;
			seq_enqueue_work(seq_sync_mem, (void *)&args, true);
		}
	}
	return (NULL);
}

/*
 * must be in sequencer
 */
void
mem_sync_with_domain(void *arg)
{
	ldom_t		*ldp;
	int		sync_mem = 0;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->mem_dr_state.inprogress == true) {
			/* there was memory dr in progress */
			if (ldp->state != ACTIVE) {
				/* domain went down, clear MD flag */
				ldp->mem_dr_state.inprogress = false;
				ldp->flag.memdr_in_progress = false;
				ldp->flag.locked = false;
				if (ldp->state != INACTIVE)
					seq_update_guest_mds(ldp);
			} else {
				DBG(zprintf("%s: mem dr crash recovery:"
				    "sync mem with domain %s\n",
				    __func__, ldp->namep));
				ldp->flag.memdr_in_progress = true;
				ldp->flag.locked = true;
				sync_mem++;
			}
		}
	}
	if (sync_mem > 0) {
		if (thr_create(NULL, NULL, do_sync_mem, NULL, THR_DETACHED,
		    NULL) != 0)
			efatal("Could not create thread to add memory");
	}
}

/*
 * Must be in the sequencer to call this routine.
 */
void
mem_dr_cleanup(ldom_t *ldp)
{

	if (ldp->flag.memdr_in_progress != true)
		return;

	ASSERT(ldp->state != ACTIVE);

	ldp->flag.memdr_in_progress = false;
	ldp->flag.locked = false;

	if (ldp->mem_dr_state.tid != NULL) {
		/*
		 * terminate the thread waiting for response from
		 * domain
		 */
		thr_kill(ldp->mem_dr_state.tid, SIGTERM);
		ldp->mem_dr_state.tid = NULL;
	}
	if (ldp->mem_dr_state.inprogress == true) {
		ldp->mem_dr_state.inprogress = false;
		if (ldp->state != INACTIVE)
			seq_update_guest_mds(ldp);
	}

}
