/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_MIG_H
#define	_MIG_H

#ifdef __cplusplus
extern "C" {
#endif

#include <sys/types.h>
#include <time.h>
#include <thread.h>
#include <synch.h>

#include "chain.h"
#include "cli_conn.h"
#include "ldmd_types.h"
#include "ldm_protocol.h"
#include "mig_class.h"

/* macros to tell if a migration is in progress for a domain */
#define	MIG_INPROGRESS(_ldp)	(((_ldp) != NULL) &&			\
				(((_ldp)->flag.migration_target) ||	\
				((_ldp)->flag.migration_source)))

#define	MIG_INPROGRESS_SRC(_ldp) (((_ldp) != NULL) &&			\
				((_ldp)->flag.migration_source))

#define	MIG_INPROGRESS_TGT(_ldp) (((_ldp) != NULL) &&			\
				((_ldp)->flag.migration_target))


/* macro to tell if a migration has been cancelled */
#define	MIG_CANCELED(_ldp)	((_ldp)->mig_state.cancel == true)

/*
 * Macros to determine characteristics of the migration in progress.
 */
#define	IS_MIG_COLD(_argsp)	((_argsp)->final_state != ACTIVE)
#define	IS_MIG_DRY_RUN(_argsp)	((_argsp)->dry_run == true)

/* timeout for intial handshake message exchanges (in seconds) */
#define	MIG_HANDSHAKE_TIMEOUT	60
#define	MIG_TRANSFER_TIMEOUT	60

/* timeout for connection to a host's migration port */
#define	MIG_SOCK_CONNECTION_TIMEOUT	1000	/* units: milliseconds (ms) */

/*
 * Min and max number of compression/decompression worker threads. A DEBUG-
 * only option allows for tuning of the number of worker threads used. A
 * a value of MIG_NWORKERS_USEDEFAULT indicates the number of worker threads
 * should be derived from the number of CPUs in the control domain.
 */
#define	MIG_NWORKERS_MIN	2
#define	MIG_NWORKERS_MAX	16
#define	MIG_NWORKERS_USEDEFAULT	-1

typedef enum {
	MIG_CONN_INIT	= 0x1,
	MIG_CONN_UP	= 0x2,
	MIG_CONN_SEND 	= 0x3,
	MIG_CONN_RECV	= 0x4,
	MIG_CONN_ERR	= 0x5,
	MIG_CONN_DOWN	= 0x6,
} mig_conn_state_t;

typedef struct mig_msgbuf mig_msgbuf_t;

typedef struct mig_conn mig_conn_t;

struct io {
	void	*(*cb)(mig_conn_t *);	/* IO callback */
	int	left;			/* # of bytes left to transfer */
	char	*buf;			/* IO buffer */
	char	*bufp;			/* current position in IO buffer */
};

/*
 * State of a migration connection between the LDom Manager and the
 * peer LDom Manager.
 */
struct mig_conn {
	int		fd;		/* connect socket file descriptor */
	SSL		*ssl;		/* SSL descriptor */
	ldm_hdr_t	hdr;		/* space to buffer message header */
	struct io	read;
	struct io	write;
	mig_conn_state_t state;		/* state of the connection */
	int		reqid;
	time_t		conntime;	/* time connection made */
	mutex_t		lock;
	cond_t		cond;
	uint16_t	major_vers;	/* negotiated major version */
	uint16_t	minor_vers;	/* negotiated minor version */
};

/*
 * A cache of the migration operation state used to display the
 * status of the operation.
 */
typedef struct mig_cache {
	uint32_t	progress;	/* % completed */
	char		*peer_name;

	/* source only: used to calculate progress completion percentage */
	uint64_t	total_mem;	/* total amount of mem to be migrated */
	uint64_t	migrated_mem;	/* running count of amount migrated */
} mig_cache_t;

/*
 * Information about the state of a migration used internally while
 * a migration is in progress.
 */
typedef struct mig_state {
	bool_t		cancel;		/* was the operation cancelled */
	bool_t		suspended;	/* has the LDom been suspended */
	md_t		*bind_mdp;	/* MD containing binding information */
	bound_mfrag_t	*bind_fp;	/* frag holding the binding MD */
} mig_state_t;

/*
 * Arguments passed to various functions during the course of a
 * migration.
 */
typedef struct mig_seq_args {
	mig_conn_t	*mconn;
	ldom_t		*ldp;
	response_t	*resp;
	int		status;
	bool_t		dry_run;	/* indicate whether -n was specified */
	bool_t		force;		/* force a migration */
	uint32_t	final_state;	/* the final state for the target */
	char		*tgt_ldom;	/* name of target LDom to be created */
	char		*peer_hostname;	/* hostname of peer machine */

	/* used in target only */
	mem_blocks_t	*mem;		/* memory bindings */
	bool_t		mem_rcvd;	/* some memory was transferred */
	char	*uuid;			/* universally unique id (UUID) */

	/* used in source only */
	struct cli_conn	*client;
} mig_seq_args_t;

void migration_init(void);
mig_conn_t *mig_new_connection(int fd);
bool_t mig_check_version(uint16_t major, uint16_t minor);
void mig_close_conn(mig_conn_t *conn);
int mig_write_msg(mig_conn_t *conn, void *buf, size_t len, int timeout);
int mig_write_buf(mig_conn_t *conn, void *buf, size_t len, int timeout);
int mig_read_msg(mig_conn_t *conn, int timeout);
int mig_read_mig_msg(mig_conn_t *conn, void *dest, size_t max_size,
    int timeout);
bool_t is_migration_in_progress(ldom_t *pend);
ldom_t *mig_inprogress_domain(void);
void flag_migration_for_cancellation(ldom_t *ldp);
void *migration_target_thr(void *arg);
void mig_wait_for_domain_services(ldom_t *ldp);
int mig_compress(uint8_t *in, uint32_t in_size, uint8_t *out,
    uint32_t *out_size);
int mig_decompress(uint8_t *in, uint32_t in_size, uint8_t *out,
    uint32_t *out_size);
int mig_max_compressed_size(uint32_t in_size, uint32_t *out_size);

void	free_mig_seq_args(mig_seq_args_t *args);
int	send_mig_msg_resp(mig_seq_args_t *args, uint32_t type, uint32_t status);
bool_t	recv_mig_msg_resp(mig_seq_args_t *args, uint32_t type);
void	read_error_mig_msg(mig_msg_t *mig_msg, response_t *resp);
int	set_sock_nonblocking(int sock);

#ifdef __cplusplus
}
#endif

#endif	/* _MIG_H */
