/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include "basics.h"
#include "database.h"
#include "errors.h"
#include "ldomconfigure.h"
#include "mig_be.h"
#include "mig_class.h"
#include "sequencer.h"
#include "xml_event.h"
#include "xml_cntrl_common.h"

/* Function Prototypes */
static bool_t	i_src_finish_cold_migration(mig_seq_args_t *args);
static void	i_tgt_finish_cold_migration(void *args);

/*
 * Source ops vector
 */
static bool_t mig_src_cold_init(mig_seq_args_t *args);
static bool_t mig_src_cold_send_config(mig_seq_args_t *args);
static bool_t mig_src_cold_send_bindings(mig_seq_args_t *args);
static bool_t mig_src_cold_feasibility_check(mig_seq_args_t *args);
static bool_t mig_src_cold_bind_check(mig_seq_args_t *args);
static bool_t mig_src_cold_send_state(mig_seq_args_t *args);

mig_src_ops_t mig_src_cold_ops = {
	mig_src_cold_init,			/* mig_src_init */
	mig_src_cold_send_config,		/* mig_src_send_config */
	mig_src_cold_send_bindings,		/* mig_src_send_bindings */
	mig_src_cold_feasibility_check,		/* mig_src_feasibility_check */
	mig_src_cold_bind_check,		/* mig_src_bind_check */
	mig_src_cold_send_state			/* mig_src_send_state */
};

/*
 * Implements mig_src_init()
 */
static bool_t
mig_src_cold_init(mig_seq_args_t *args)
{
	ASSERT(args != NULL);

	/*
	 * Check for any conditions on the source machine that
	 * may block a migration (delayed reconfig, etc)
	 */
	if (mig_src_local_feasibility_check(args) == false)
		return (false);

	if (args->final_state == BOUND) {
		return (mig_src_bound_feasibility_check(args));
	}

	return (true);
}

/*
 * Implements mig_src_send_config()
 */
static bool_t
mig_src_cold_send_config(mig_seq_args_t *args)
{
	mig_conn_t	*mconn = args->mconn;
	ldom_t		*ldp = args->ldp;

	DBG(zprintf("%s\n", __func__));

	/*
	 * Transfer any static configuration information so that
	 * it can be used for the feasibility checks.
	 */
	if (migrate_config(mconn, ldp) != 0) {
		args->status = MIG_STATUS_FAILURE;
		DBG(zprintf("unable to send MIG config information\n"));
		return (false);
	}

	return (true);
}

/*
 * Implements mig_src_send_bindings()
 */
/*ARGSUSED*/
static bool_t
mig_src_cold_send_bindings(mig_seq_args_t *args)
{
	DBG(zprintf("%s\n", __func__));

	/*
	 * No binding info needed for cold migration.
	 * If a domain is to be bound, the bindings
	 * can be generated from scratch on the target.
	 */
	return (true);
}

/*
 * Implements mig_src_feasibility_check()
 */
static bool_t
mig_src_cold_feasibility_check(mig_seq_args_t *args)
{
	DBG(zprintf("%s\n", __func__));

	/*
	 * No feasibility checks needed for a cold
	 * migration of an inactive domain.
	 */
	if (args->final_state == INACTIVE)
		return (true);

	ASSERT(args->final_state == BOUND);

	/* perform the checks needed for a bound domain */
	if (recv_mig_msg_resp(args, MIG_CHECK_CONFIG) == false)
		return (false);

	return (true);
}

/*
 * Implements mig_src_bind_check()
 */
/*ARGSUSED*/
static bool_t
mig_src_cold_bind_check(mig_seq_args_t *args)
{
	DBG(zprintf("%s\n", __func__));

	/*
	 * No checking is needed on a cold migration.
	 *
	 * (1) INACTIVE domain: no bind will be performed on the target.
	 * (2) BOUND domain: the call to bind the domain is made in
	 * i_tgt_finish_cold_migration() and any bind failure messages
	 * are sent back from there.
	 */

	return (true);
}

/*
 * Implements mig_src_send_state()
 */
static bool_t
mig_src_cold_send_state(mig_seq_args_t *args)
{
	DBG(zprintf("%s\n", __func__));

	/*
	 * There is no state to send for a cold migration so
	 * the migration is now complete.
	 */
	return (i_src_finish_cold_migration(args));
}


/*
 * After sending the configuration information, the cold
 * migration is complete. All that is left to do is wait
 * for the final handshake message and do the teardown of
 * the source domain.
 */
static bool_t
i_src_finish_cold_migration(mig_seq_args_t *args)
{
	mig_conn_t	*mconn = args->mconn;
	ldom_t		*ldp = args->ldp;
	bool_t		is_successful = false;

	DBG(zprintf("%s\n", __func__));

	ASSERT(mconn != NULL);
	ASSERT(ldp != NULL);

	/* wait for the final handoff message */
	if (recv_mig_msg_resp(args, MIG_HANDOFF) == false)
		goto done;

	if (IS_MIG_DRY_RUN(args)) {
		DBG(zprintf("Dry run only, operation complete\n"));
		ldp->flag.migration_source = false;
		if (args->status == MIG_STATUS_SUCCESS)
			is_successful = true;
		goto done;
	}

	/* We only proceed with the teardown if the target reported success */
	if (args->status != MIG_STATUS_SUCCESS)
		goto done;

	/* Done with the migration, cleanup source */
	DBG(zprintf("tearing down source LDom %s\n", ldp->namep));

	if (ldp->state == BOUND) {
		/*
		 * Unbind the source domain.
		 */
		ldp->state = UNBINDING;
		CHAIN_ADD(zeus.inprogress, ldp, inprog);

		seq_unbind_ldoms();
		cleanse_progress_chain();
	}

	if (!IS_MIG_DRY_RUN(args)) {
		/* Send a domain event. */
		event_send_dom_event(XML_MIGRATE_DOMAIN, args->ldp,
		    args->client);
	}

	ASSERT(ldp->state == INACTIVE);
	UNLOCK_LDOM(ldp);
	if (destroy_one_ldom(ldp, args->resp, args->client) == true) {
		/* clear ldp entry in mig args struct */
		args->ldp = NULL;
	} else {
		args->status = MIG_STATUS_FAILURE;
		rprintf(args->resp, MSG_ERR_MIG_SOURCE_DESTROY);
		ldp->flag.migration_source = false;
		goto done;
	}

	is_successful = true;

done:
	/* make sure a response is sent */
	args->client->send_resp = true;

	return (is_successful);
}

/*
 * Target ops vector
 */
static bool_t mig_tgt_cold_init(mig_seq_args_t *args);
static bool_t mig_tgt_cold_recv_config(mig_seq_args_t *args);
static bool_t mig_tgt_cold_recv_bindings(mig_seq_args_t *args);
static bool_t mig_tgt_cold_feasibility_check(mig_seq_args_t *args);
static bool_t mig_tgt_cold_bind(mig_seq_args_t *args);
static bool_t mig_tgt_cold_recv_state(mig_seq_args_t *args);

mig_tgt_ops_t mig_tgt_cold_ops = {
	mig_tgt_cold_init,			/* mig_tgt_init */
	mig_tgt_cold_recv_config,		/* mig_tgt_recv_config */
	mig_tgt_cold_recv_bindings,		/* mig_tgt_recv_bindings */
	mig_tgt_cold_feasibility_check,		/* mig_tgt_feasibility_check */
	mig_tgt_cold_bind,			/* mig_tgt_bind */
	mig_tgt_cold_recv_state			/* mig_tgt_recv_state */
};

static void
i_tgt_finish_cold_migration(void *args)
{
	mig_seq_args_t	*wargs = args;
	ldom_t		*ldp = wargs->ldp;
	ldom_state_t	state = wargs->final_state;
	response_t	*resp = wargs->resp;
	bool_t		bound;

	/* nothing to do for a dry run - LDom teardown already done */
	if (IS_MIG_DRY_RUN(wargs)) {
		ASSERT(find_ldom_by_name(wargs->tgt_ldom) == NULL);
		goto done;
	}

	UNLOCK_LDOM(ldp);

	/* We only proceed if the target was successful so far */
	if (wargs->status != MIG_STATUS_SUCCESS)
		goto done;

	switch (state) {
	case INACTIVE:
		DBG(zprintf("Cold migration of inactive LDom '%s' complete\n",
		    ldp->namep));

		/*
		 * XXXMIG - Now that the operation is done, we can remove the
		 * hack LDom variable being used to tell zeus to destroy the
		 * domain during startup.
		 */
		database_ldom_del_variable(ldp, "SUNW,mig-target");
		database_save();
		break;

	case BOUND:
		bound = bind_an_ldom(ldp, resp);
		if (!bound) {
			DBG(zprintf("%s: Unable to bind LDom '%s' for cold "
			    "migration\n", __func__, ldp->namep));
			wargs->status = MIG_STATUS_FAILURE;
			break;
		}

		ldp->flag.gen_new_md = true;
		ldp->state = BINDING;
		CHAIN_ADD(zeus.inprogress, ldp, inprog);

		/* notify HV & effected services of new binding */
		seq_bind_new();

		/*
		 * XXXMIG - Now that the domain is bound, we can remove the
		 * hack LDom variable being used to tell zeus to destroy the
		 * domain during startup.
		 */
		database_ldom_del_variable(ldp, "SUNW,mig-target");
		database_save();

		seq_send_md_update_notifs();

		/* register guest for Domain Services */
		ds_register_domain(ldp, true);

		cleanse_progress_chain();

		DBG(zprintf("Cold migration of bound LDom '%s' complete\n",
		    ldp->namep));
		break;
	default:
		DBG(zprintf("Unexpected state %d, aborting\n", (int)state));
		break;
	}

done:

	/* send the final ok message or failure message if needed */
	if (wargs->status == MIG_STATUS_SUCCESS) {
		(void) send_mig_msg_resp(wargs, MIG_HANDOFF, wargs->status);
	} else {
		ASSERT(wargs->status == MIG_STATUS_FAILURE);
		(void) send_mig_msg_resp(wargs, MIG_ERROR_RESP, wargs->status);
	}
}

/*
 * Implements mig_tgt_init()
 */
static bool_t
mig_tgt_cold_init(mig_seq_args_t *args)
{
	ASSERT(args != NULL);

	/*
	 * Check for any conditions on the target machine that
	 * may block a migration (delayed reconfig, etc)
	 */
	return (mig_tgt_local_feasibility_check(args));
}

/*
 * Implements mig_tgt_recv_config()
 */
static bool_t
mig_tgt_cold_recv_config(mig_seq_args_t *args)
{
	args->ldp = receive_config(args);
	if (args->ldp == NULL) {
		DBG(zprintf("Unable to create target LDom\n"));
		args->status = MIG_STATUS_FAILURE;
		if (send_mig_msg_resp(args, MIG_ERROR_RESP, args->status) != 0)
			warning("Failed to send migration error message to ldmd"
			    " on source machine %s\n", args->peer_hostname);
		return (false);
	}

	if (args->status != MIG_STATUS_SUCCESS) {
		/*
		 * Domain creation failed halfway
		 */
		DBG(zprintf("%s: Error during guest creation, cleaning up...\n",
		    __func__));
		return (false);
	}

	return (true);
}

/*
 * Implements mig_tgt_recv_bindings()
 */
/*ARGSUSED*/
static bool_t
mig_tgt_cold_recv_bindings(mig_seq_args_t *args)
{
	/*
	 * No binding info needed for cold migration.
	 * Bindings are generated from scratch for a
	 * cold migration.
	 */
	return (true);
}

/*
 * Implements mig_tgt_feasibility_check()
 */
static bool_t
mig_tgt_cold_feasibility_check(mig_seq_args_t *args)
{
	ASSERT(args != NULL);
	ASSERT(args->ldp != NULL);

	ldom_t		*ldp = args->ldp;
	response_t	*resp = args->resp;
	uint32_t	mig_msg_type = MIG_CHECK_CONFIG;
	bool_t		can_migrate = true;

	DBG(zprintf("%s\n", __func__));

	if (args->final_state == INACTIVE)
		goto done;		/* nothing to check */

	if (mig_tgt_bound_feasibility_check(args) == false) {
		args->status = MIG_STATUS_FAILURE;
		mig_msg_type = MIG_ERROR_RESP;
		can_migrate = false;
	}

	if (IS_MIG_DRY_RUN(args) && (args->status == MIG_STATUS_FAILURE))
		rprintf(resp, MSG_ERR_MIG_DRY_RUN_FAILURE, ldp->namep);

	send_mig_msg_resp(args, mig_msg_type, args->status);

done:
	/*
	 * If this is a dry run there is nothing more to do so we can
	 * now destroy the domain that was temporarily created.
	 */
	if (IS_MIG_DRY_RUN(args)) {
		ASSERT(ldp->state == INACTIVE);
		UNLOCK_LDOM(ldp);
		if (destroy_one_ldom(ldp, resp, NULL) == true) {
			/* clear ldp entry in mig args struct */
			args->ldp = NULL;
		} else {
			args->status = MIG_STATUS_FAILURE;
			ASSERT(ldp);
			ASSERT(ldp->namep);
			warning("Unable to destroy temporary LDom %s which was "
			    "created during a migration dry run\n", ldp->namep);
		}
		DBG(zprintf("Dry run only, operation complete\n"));
	}

	return (can_migrate);
}

/*
 * Implements mig_tgt_bind()
 */
/*ARGSUSED*/
static bool_t
mig_tgt_cold_bind(mig_seq_args_t *args)
{
	return (true);
}

/*
 * Implements mig_tgt_recv_state()
 */
static bool_t
mig_tgt_cold_recv_state(mig_seq_args_t *args)
{
	DBG(zprintf("%s\n", __func__));

	/* the operation is complete */
	seq_enqueue_work(i_tgt_finish_cold_migration, args, true);

	if (args->status == MIG_STATUS_SUCCESS)
		return (true);
	else
		return (false);
}
