/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#include "backend.h"
#include "database_internal.h"
#include "errors.h"
#include "ldom.h"
#include "option.h"
#include "mig.h"
#include "mig_class.h"
#include "mig_common.h"
#include "sequencer.h"
#include "support.h"
#include "vcpu.h"
#include "vio.h"
#include "xml_common.h"
#include "xml_cntrl_common.h"
#include "xml_output.h"
#include "xml_parse.h"
#include "xml_v3_defs.h"

extern bool_t	hv_is_migration_supported(void);

/*
 * Fill in the common parts of a MIG message. Takes the length of the
 * variable data. The caller is responsible for deallocating the msg.
 * The space needed for the ldm_hdr_t and mig_msg_t structs will be
 * accounted for. 'vlen' must be the size needed for the message struct
 * identified by 'msg_type'.
 */
uint8_t *
init_mig_msg(int msg_type, uint32_t vlen, uint32_t status)
{
	uint8_t		*bufp;
	ldm_hdr_t	*hdr;
	mig_msg_t	*mig_msg;

	bufp = Xmalloc(MIG_MSG_HDR_LEN + vlen);

	hdr = (ldm_hdr_t *)bufp;
	hdr->msg_type = hton32(LDM_MIG_MSG);
	hdr->payload_len = hton32(sizeof (mig_msg_t) + vlen);

	mig_msg = (mig_msg_t *)(bufp + sizeof (ldm_hdr_t));
	mig_msg->type = hton32(msg_type);
	mig_msg->len = hton32(vlen);
	mig_msg->status = hton32(status);
	mig_msg->magic = hton32(MIG_MAGIC);

	return (bufp);
}

int
migrate_config(mig_conn_t *mconn, ldom_t *ldp)
{
	int			rc = ERR_GENERIC;
	char			*argv[1];
	int			argc = 1;
	response_t		*resp;
	ldm_op_reply_t		*reply;
	cli_reply_data_t	*data_info;
	char			*resp_msg;

	/*
	 * Use a local response structure to get the XML config
	 * data. If the global structure was used, the XML data
	 * would be mixed in with the informational messages
	 * intended for the user.
	 */
	init_response(&resp);

	/* create some arguments for getting the constraints */
	argv[0] = Xstrdup(ldp->namep);

	xml_output_constraints(argc, argv, resp);

	/* pull out the constraints from the response structure */
	reply = (ldm_op_reply_t *)(resp->bufp + sizeof (ldm_hdr_t));
	data_info = (cli_reply_data_t *)reply->rp_data;
	resp_msg = xml_parse_resp(data_info->text);

	if (resp_msg != NULL) {
		DBG2(zprintf("%s XML response:\n%s\n", __func__, resp_msg));

		/* send the constraints to the target machine */
		rc = mig_write_buf(mconn, resp_msg, strlen(resp_msg),
		    MIG_TRANSFER_TIMEOUT);
		if (rc == -1) {
			DBG(zprintf("failure sending xml constraints\n"));
		}
	}

	Xfree(argv[0]);
	Xfree(resp_msg);
	fini_response(resp);

	return (rc);
}

/*
 * Common Target Functions
 */
static char *
parse_data_failure(xmlNodePtr data)
{
	xmlNodePtr	resp_node, resp_msg;

	resp_node = xml_find_subnode(data, XML_RESPONSE);
	resp_msg = xml_find_subnode(resp_node, XML_RESP_MSG);
	if (resp_msg == NULL)
		return (NULL);
	else
		return ((char *)xmlNodeGetContent(resp_msg));
}

static ldom_t *
migrate_domain_create(xmlNodePtr data_node, char *tgt_name, response_t *resp)
{
	int		ret, rv = 0;
	cli_conn_t	client;
	xmlNodePtr	subnode, envelope, content, ldom_info_section;
	xmlNodePtr	ldom_info, temp_data;
	char		*ldm_name, *resp_msg;
	xmlChar		*mac_addr, *section_name;
	xmlChar		*hostid;
	xmlChar		*failure_policy;
	xmlChar		*master;
	const int	create_args = 5;
	int		argc = 0;
	void		*argv[create_args];
	ldom_t		*ldp = NULL;

	ASSERT(tgt_name != NULL);

	/* create the dummy client struct */
	init_response(&client.resp);

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	if (envelope == NULL) {
		rprintf(resp, "Constraint XML missing Envelope\n");
		rv = -1;
		goto end;
	}

	content = xml_find_subnode(envelope, CONTENT_NODE);
	if (content == NULL) {
		rprintf(resp, "Constraint XML missing Content\n");
		rv = -1;
		goto end;
	}

	/*
	 * Retrieve the name of the domain on the source.
	 */
	ldm_name = (char *)xmlGetProp(content, ID_ATTR_IN);

	if (ldm_name == NULL) {
		rprintf(resp, "Constraint XML: Content ID attribute\n");
		rv = -1;
		goto end;
	}

	/* XXXMIG - stash the source domain name for use in status output */

	xmlFree((xmlChar *)ldm_name);

	ldom_info_section = xml_find_section(content, XML_LDM_INFO);
	if (ldom_info_section != NULL) {
		ldom_info = xml_find_subnode(ldom_info_section, ITEM_NODE);
		subnode = xml_find_subnode(ldom_info, ADDRESS_NODE_IN);
		if (subnode != NULL) {
			mac_addr = xmlNodeGetContent(subnode);
			if (strcmp((char *)mac_addr,
			    XML_AUTO_ALLOCATED_MACADDR) != 0) {
				argv[argc++] = zsprintf("mac-addr=%s",
				    mac_addr);
			}
			xmlFree(mac_addr);
		}
		subnode =  xml_find_v3_property(ldom_info, XML_HOSTID);
		if (subnode != NULL) {
			hostid = xmlNodeGetContent(subnode);
			if (hostid != NULL) {
				uint64_t h_id = strtoll((char *)hostid, NULL,
				    16);
				if (h_id != UNASSIGNED_HOSTID) {
					argv[argc++] = zsprintf("hostid=%s",
					    (char *)hostid);
				}
			}
			xmlFree(hostid);
		}

		/*
		 * No need to add to the command line args if the FW level
		 * does not support failure policy or domain dependencies
		 */
		if (LDOM_DEPENDENCIES_SUPPORTED()) {
			subnode = xml_find_v3_property(ldom_info,
			    XML_FAILURE_POLICY);
			if (subnode != NULL) {
				failure_policy = xmlNodeGetContent(subnode);
				argv[argc++] = zsprintf("failure-policy=%s",
				    (char *)failure_policy);
				xmlFree(failure_policy);
			}
			subnode = xml_find_v3_property(ldom_info, XML_MASTER);
			if (subnode != NULL) {
				master = xmlNodeGetContent(subnode);
				argv[argc++] = zsprintf("master=%s",
				    (char *)master);
				xmlFree(master);
			}
		}
	}
	temp_data = xmlNewNode(NULL, (xmlChar *)"data");
	argv[argc++] = Xstrdup(tgt_name);

	ASSERT(argc < create_args);

	ret = xml_call_cli_cmd(argc, argv, LDM_CREATE, temp_data, &client);
	free_argv(argv, argc);
	argc = 0;

	if (ret == false) {
		resp_msg = parse_data_failure(temp_data);
		if (resp_msg != NULL) {
			rprintf(resp, "%s\n", resp_msg);
			xmlFree(resp_msg);
		}
		rprintf(resp, "Failed to recreate domain on target machine\n");
		xmlFreeNode(temp_data);
		rv = -1;
		goto end;
	}
	xmlFreeNode(temp_data);

	/*
	 * Get the LDom pointer for the newly created domain - should not
	 * have got this far if the creation had failed
	 */
	ldp = find_ldom_by_name(tgt_name);
	ASSERT(ldp != NULL);

	/* XXXMIG - could cache source name at this point if (ldp != NULL) */

	subnode = content->xmlChildrenNode;
	while (subnode != NULL) {
		if (subnode->type != XML_ELEMENT_NODE) {
			subnode = subnode->next;
			continue;
		}

		section_name = get_section_resource_name(subnode);
		if (section_name == NULL) {
			subnode = subnode->next;
			xmlFree(section_name);
			continue;
		}
		if (xmlStrcmp(section_name,
		    (const xmlChar *)XML_LDM_INFO) == 0) {
			subnode = subnode->next;
			xmlFree(section_name);
			continue;
		}
		if (xmlStrcmp(section_name,
		    (const xmlChar *)XML_LDM_INFO) == 0) {
			subnode = subnode->next;
			xmlFree(section_name);
			continue;
		}
		temp_data = xmlNewNode(NULL, (xmlChar *)"data");
		xmlNewProp(temp_data, VERSION_ATTR, LDOM_V3_DATA_VERSION);
		ret = goto_resource_routine((char *)section_name, temp_data,
		    ldp, XML_ADD_RES, subnode, &client);
		xmlFree(section_name);
		if (ret == -1) {
			rv = -1;
			rprintf(resp, "Constraint XML: Failed to "
			    "create domain\n");
			resp_msg = parse_data_failure(temp_data);
			if (resp_msg != NULL) {
				rprintf(resp, "%s\n", resp_msg);
				xmlFree(resp_msg);
			}
			xmlFreeNode(temp_data);
			goto end;
		}
		xmlFreeNode(temp_data);
		subnode = subnode->next;
	}

	/*
	 * XXXMIG - need some way of tagging this target domain in
	 * the database so that if zeus dies before the bind, the
	 * domain will not be resurrected out of the database. As
	 * a temporary hack, add an LDom variable for this purpose.
	 */
	database_ldom_add_variable(ldp, "SUNW,mig-target", "true");
	database_save();

end:
	fini_response(client.resp);
	if (rv != 0)
		return (NULL);

	return (ldp);
}

ldom_t *
receive_config(mig_seq_args_t *args)
{
	char		*tgt_name = args->tgt_ldom;
	mig_conn_t	*mconn = args->mconn;
	response_t	*resp = args->resp;
	mig_msg_t	*mig_msg;
	xmlDocPtr	xml_cons;
	ldom_t		*ldp;
	xmlNodePtr	root, data;

	if (mig_read_mig_msg(mconn, NULL, MIG_MAX_MSG_LEN,
	    MIG_TRANSFER_TIMEOUT) == -1) {
		args->status = MIG_STATUS_FAILURE;
		DBG(zprintf("%s: read timed out\n", __func__));
		return (NULL);
	}

	mig_msg = (mig_msg_t *)mconn->read.buf;
	ASSERT(mig_msg != NULL);
	ASSERT(mig_msg->data != NULL);

	if (mig_msg->type != MIG_BUFFER) {
		args->status = MIG_STATUS_FAILURE;
		DBG(zprintf("%s: received unexpected message type (%d), "
		    "expected %d\n", __func__, mig_msg->type, (int)MIG_BUFFER));
		return (NULL);
	}

	DBG_BLOCK {
		int	fd;

		fd = open("/export/mig-cons.xml", O_CREAT|O_TRUNC|O_RDWR, 0600);
		if (fd == -1) {
			DBG(zprintf("unable to open file '/export/mig-cons."
			    "xml', skipping save of constraints\n"));
			return (NULL);
		}
		full_write(fd, (uint8_t *)mig_msg->data, mig_msg->len);
		fsync(fd);
		close(fd);
	}

	/* convert the raw stream to an XML document */
	xml_cons = xmlReadDoc((xmlChar *)mig_msg->data, NULL, NULL,
	    XML_PARSE_RECOVER | XML_PARSE_NOBLANKS);
	if (xml_cons == NULL) {
		args->status = MIG_STATUS_FAILURE;
		DBG(zprintf("Invalid XML constraint document\n"));
		return (NULL);
	}

	/* find the data node for this doc */
	root = xmlDocGetRootElement(xml_cons);
	data = xml_find_subnode(root, XML_DATA);

	ldp = migrate_domain_create(data, tgt_name, resp);
	if (ldp != NULL) {
		/* prevent domain modification while being migrated to target */
		LOCK_LDOM(ldp);
	} else {
		args->status = MIG_STATUS_FAILURE;
		/*
		 * A domain created halfway may still exist
		 */
		ldp = find_ldom_by_name(tgt_name);
		if (ldp != NULL) {
			LOCK_LDOM(ldp);
		}
		DBG(zprintf("%s: domain creation failed...\n", __func__));
		DBG_BLOCK {
			char *txt = get_response_text(resp);
			if (txt != NULL) {
				DBG(zprintf("%s", txt));
				Xfree(txt);
			}
		}
	}

	xmlFreeDoc(xml_cons);

	return (ldp);
}


/*
 * This routine takes an LDom structure ptr (created after the constraints
 * were sent across from the source) from the migration args structure passed
 * in and verifies that the VIO services that it uses on the source also
 * exist on the target machine.
 *
 * Note: it does not verify that the virtual disk points to the same backend
 * as the source domain - that will need extra functionality, not yet in ldmd.
 *
 * Return Values:
 *	true	- Success
 *	false	- At least one of the services does not exist
 */
static bool_t
are_vio_services_present(mig_seq_args_t *args)
{
	ASSERT(args != NULL);

	bool_t		found_vol;
	bool_t		found_vcc = false;
	ldom_t		*tmpldp;
	ldom_t		*ldp = args->ldp;
	response_t	*resp = args->resp;
	vds_volume_t	*volume;
	vdisk_t		*vdisk;
	vnet_t		*vnet;
	vds_t		*vds;

	ASSERT(ldp != NULL);

	/* the source already verified that the domain didn't export services */
	ASSERT(CHAIN_LENGTH(ldp->vdss) == 0);
	ASSERT(CHAIN_LENGTH(ldp->vsws) == 0);
	ASSERT(CHAIN_LENGTH(ldp->vccs) == 0);
	ASSERT(CHAIN_LENGTH(ldp->vdpccs) == 0);
	ASSERT(CHAIN_LENGTH(ldp->vdpcss) == 0);

	/* verify vdisk services */
	DBG(zprintf("%s: vdisks: %d\n", __func__, CHAIN_LENGTH(ldp->vdisks)));
	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {
		/* verify that the vds exists */
		vds = vds_find_by_name(vdisk->service);
		if (vds == NULL) {
			DBG(zprintf("%s: failed to find vds %s for LDom %s\n",
			    __func__, vdisk->service, ldp->namep));
			rprintf(resp, MSG_ERR_MIG_NO_TGT_VDS, vdisk->service);
			args->status = MIG_STATUS_FAILURE;
			return (false);
		}

		/* verify that the vdsdev exists */
		found_vol = false;
		CHAIN_ITER(vds->volumes, volume, link) {
			if (strcmp(volume->volname, vdisk->volname) == 0) {
				found_vol = true;
				break;
			}
		}
		if (found_vol == false) {
			DBG(zprintf("%s: failed to find vol %s for LDom %s\n",
			    __func__, vdisk->volname, ldp->namep));
			rprintf(resp, MSG_ERR_MIG_NO_TGT_VOL,
			    vdisk->volname, vdisk->service);
			args->status = MIG_STATUS_FAILURE;
			return (false);
		}
	}

	/* verify vnet services */
	DBG(zprintf("%s: vnets: %d\n", __func__, CHAIN_LENGTH(ldp->vnets)));
	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		if (vsw_find_by_name(vnet->service) == NULL) {
			DBG(zprintf("%s: failed to find vsw %s for LDom %s\n",
			    __func__, vnet->service, ldp->namep));
			rprintf(resp, MSG_ERR_MIG_NO_TGT_VSW, vnet->service);
			args->status = MIG_STATUS_FAILURE;
			return (false);
		}
	}

	/* verify console services */
	CHAIN_ITER(zeus.ldoms, tmpldp, link) {
		DBG(zprintf("%s: vccs: %d on %s\n",
		    __func__, CHAIN_LENGTH(tmpldp->vccs), tmpldp->namep));
		if (CHAIN_LENGTH(tmpldp->vccs) > 0) {
			found_vcc = true;
			break;
		}
	}
	if (found_vcc == false) {
		DBG(zprintf("%s: Failed to find any vcc device\n", __func__));
		rprintf(resp, MSG_ERR_MIG_NO_TGT_VCC);
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	return (true);
}

/*
 * Check for any conditions on the source machine that
 * may block a migration (delayed reconfig, etc)
 */
bool_t
mig_src_local_feasibility_check(mig_seq_args_t *args)
{
	ldom_t		*ldp = args->ldp;
	response_t	*resp = args->resp;
	char		hostname[MAXHOSTNAMELEN];

	DBG(zprintf("%s\n", __func__));

	/*
	 * check to see if the migration of a domain from this machine to
	 * another is not explicitly blocked (either via ldmd's SMF
	 * properties/manifest or passing --disable-outgoing-migration
	 * to ldmd when run from the command line).
	 */
	if (option.outgoing_migration_enabled == false) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, MSG_ERR_MIG_OUTGOING_MIGRATION_DISABLED);
		return (false);
	}

	/* block attempts to migrate the control domain (aka 'primary') */
	if (ldp != NULL && ldp->namep != NULL &&
	    strcmp(ldp->namep, SOD_CONTROL_DOMAIN) == 0) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, MSG_ERR_MIG_NO_MIGRATE_PRIMARY);
		return (false);
	}

	/* verify that no migration is in progress */
	if (is_migration_in_progress(ldp)) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, MSG_ERR_MIG_IN_PROGRESS);
		return (false);
	}

	/* verify that no delayed reconfiguration is pending */
	if (ldp->state >= BOUND && in_delayed_reconfig(NULL)) {
		if (gethostname(hostname, MAXHOSTNAMELEN) != 0)
			strlcpy(hostname, "the source machine", MAXHOSTNAMELEN);

		DBG(zprintf("%s: machine in delayed reconfig\n", __func__));
		rprintf(resp, MSG_ERR_MIG_DELAYED_RECONF_IN_PROGRESS,
		    zeus.delayed_reconfig_ldom->namep, hostname);
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	/* check that the firmware can support migration */
	if (ldp->state == ACTIVE && hv_is_migration_supported() == false) {
		if (gethostname(hostname, MAXHOSTNAMELEN) != 0)
			strlcpy(hostname, "the source machine", MAXHOSTNAMELEN);

		rprintf(args->resp, MSG_ERR_MIG_HV_DOWNREV, hostname);
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	return (true);
}

/*
 * Feasibility checks on source side for domains which are not inactive
 * (i.e. domain is bound or active). Used by both cold and warm migration code
 */
bool_t
mig_src_bound_feasibility_check(mig_seq_args_t *args)
{
	ASSERT(args != NULL);
	ASSERT(args->ldp != NULL);

	response_t	*resp = args->resp;
	ldom_t		*ldp = args->ldp;
	vds_volume_t	*volume;
	vdisk_t		*vdisk;
	vds_t		*vds;

	ASSERT(args->ldp->state >= BOUND);

	/* Can't migrate if physical devices are bound to the domain */
	if (CHAIN_LENGTH(ldp->bound.iodevices) > 0) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, MSG_ERR_MIG_PHYS_DEVS_BOUND, ldp->namep);
		return (false);
	}

	/* Can't migrate if domain is exporting services */
	if (is_providing_services(ldp) == true) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, MSG_ERR_MIG_EXPORTING_SERVICES, ldp->namep);
		return (false);
	}

	/* ensure that volumes aren't part of a multi-path group */
	DBG(zprintf("%s: vdisks: %d\n", __func__, CHAIN_LENGTH(ldp->vdisks)));
	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {
		ASSERT(IS_LDOM_LOCKED(ldp) == true);
		vds = vds_find_by_name(vdisk->service);
		ASSERT(vds != NULL);
		volume = vds_find_volume_by_name(vds, vdisk->volname);
		ASSERT(volume != NULL);
		if (volume->mpgroup != NULL) {
			args->status = MIG_STATUS_FAILURE;
			rprintf(resp, MSG_ERR_MIG_VOL_IN_MPGROUP,
			    vdisk->volname, vdisk->service,
			    volume->mpgroup->name);
			return (false);
		}
	}

	/* block migration if any vdpcc devices are part of the domain */
	DBG(zprintf("%s: vdpccs: %d\n", __func__, CHAIN_LENGTH(ldp->vdpccs)));
	if (CHAIN_LENGTH(ldp->vdpccs) > 0) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, MSG_ERR_MIG_VCPCC_PRESENT);
		return (false);
	}

	/* block migration if any vdpcs devices are part of the domain */
	DBG(zprintf("%s: vdpcss: %d\n", __func__, CHAIN_LENGTH(ldp->vdpcss)));
	if (CHAIN_LENGTH(ldp->vdpcss) > 0) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, MSG_ERR_MIG_VCPCS_PRESENT);
		return (false);
	}

	/* block migration if a bound master has a bound slave */
	if (has_bound_slaves(ldp, resp) == true) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, MSG_ERR_MIG_SLAVE_PRESENT);
		return (false);
	}

	return (true);
}

/*
 * Check for any conditions on the target machine that may block a migration
 * (delayed reconfig, etc)
 *
 * Note: This must be called before the constraints are sent over to
 * create a domain (otherwise the creation will spew a lot of warning
 * messages about a delayed reconfig being in progress)
 */
bool_t
mig_tgt_local_feasibility_check(mig_seq_args_t *args)
{
	ASSERT(args != NULL);
	DBG(zprintf("%s\n", __func__));

	if (recv_mig_msg_resp(args, MIG_CHECK_INIT) == false) {
		warning("Failed to receive initial migration feasibility check "
		    "message from ldmd on source machine\n");
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	/* check that the firmware can support migration */
	if (args->final_state == ACTIVE &&
	    hv_is_migration_supported() == false) {
		char	hostname[MAXHOSTNAMELEN];
		if (gethostname(hostname, MAXHOSTNAMELEN) != 0)
			strlcpy(hostname, "the target machine", MAXHOSTNAMELEN);

		rprintf(args->resp, MSG_ERR_MIG_HV_DOWNREV, hostname);
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	/* verify that no delayed reconfiguration is pending */
	if (in_delayed_reconfig(NULL)) {
		DBG(zprintf("%s: machine in delayed reconfig\n", __func__));
		ASSERT(zeus.delayed_reconfig_ldom->namep != NULL);
		rprintf(args->resp, MSG_ERR_MIG_DELAYED_RECONF_IN_PROGRESS,
		    zeus.delayed_reconfig_ldom->namep, "the target machine");
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	/*
	 * Check if a domain with the same name already exists on target
	 */
	if (find_ldom_by_name(args->tgt_ldom) != NULL) {
		DBG(zprintf("%s: domain %s already exists\n", __func__,
		    args->tgt_ldom));
		args->status = MIG_STATUS_FAILURE;
		rprintf(args->resp, MSG_ERR_MIG_LDOM_EXISTS_ON_TGT,
		    args->tgt_ldom);
		return (false);
	}

	/*
	 * send message back to the source which is waiting before it will
	 * continue with next step (sending the constraint info to create
	 * the domain on the target).
	 */
	if (send_mig_msg_resp(args, MIG_CHECK_INIT, MIG_STATUS_SUCCESS) != 0) {
		warning("Failed to send initial migration feasibility check "
		    "message to ldmd on source machine\n");
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	return (true);
}

static bool_t
is_console_available(mig_seq_args_t *args)
{
	ldom_t		*ldp = args->ldp;
	response_t	*resp = args->resp;
	vcc_t		*vcc;

	DBG(zprintf("%s:\n", __func__));

	ASSERT(args->status == MIG_STATUS_SUCCESS);
	args->status = MIG_STATUS_FAILURE;

	/* XXXMIG - just check the first vcc for now */
	vcc = CHAIN_START(zeus.hvctl_ldom_p->vccs);
	if ((vcc == NULL) || (vcc->binding == NULL)) {
		rprintf(resp, MSG_ERR_MIG_NO_TGT_VCC);
		return (false);
	}

	/* check if there are any available ports */
	if (available_res_ids(&vcc->binding->tcp_pool) <= 0) {
		rprintf(resp, MSG_ERR_MIG_NO_CONS_PORT, ldp->namep);
		return (false);
	}

	/*
	 * XXXMIG - For the initial release of migration, there is no
	 * support for manually assigned console groups. Use the domain
	 * name as the default group name and if it is already in use,
	 * block the migration. In the future, this will need to be
	 * expanded to support migration of console groups.
	 */
	if (console_group_find_by_name(vcc, ldp->namep) != NULL) {
		rprintf(resp, "Unable to bind console for LDom %s. Default "
		    "group %s is already in use\n", ldp->namep, ldp->namep);
		return (false);
	}

	args->status = MIG_STATUS_SUCCESS;
	return (true);
}

bool_t
mig_tgt_bound_feasibility_check(mig_seq_args_t *args)
{
	ASSERT(args != NULL);

	/*
	 * Check that the VIO services on the source are also present on the
	 * target. We need to go through the constraints received earlier
	 * and check to see that the services exist.
	 */
	if (are_vio_services_present(args) == false) {
		DBG(zprintf("%s: Not all VIO services available\n", __func__));
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	if (is_console_available(args) == false) {
		DBG(zprintf("%s: unable to bind console\n", __func__));
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	/*
	 * Check if the required amount of memory is available
	 */
	if (mem_bind_real_check(args->ldp, args->mem, args->resp) == false) {
		DBG(zprintf("%s: unable to bind memory\n", __func__));
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	return (true);
}

/*
 * Called by both backend routines and source/target code (these
 * routines are only used by ldmd - unlike those in mig_common.c)
 */

/* Free the fields allocated as part of the mig_seq_args_t structure */
void
free_mig_seq_args(mig_seq_args_t *args)
{
	ASSERT(args != NULL);

	if (args->mem != NULL) {
		mem_free_cloned_bindings(args->mem);
		Xfree(args->mem);
	}

	Zfree(args->peer_hostname);
	Zfree(args->tgt_ldom);
	Zfree(args);
}

/*
 * If the target encounters an error while it is exchanging migration
 * messages, it returns a msg of type MIG_ERROR_RESP. This routine
 * reads the error string so that it can be returned to ldm(1m).
 */
void
read_error_mig_msg(mig_msg_t *mig_msg, response_t *resp)
{
	char	*msg_str;

	ASSERT(mig_msg != NULL);
	ASSERT(mig_msg->type == MIG_ERROR_RESP);

	/*
	 * If the MIG_ERROR_RESP msg had a payload containing an error
	 * string, we copy it into the 'resp' structure passed in.
	 */
	if (mig_msg->len > 0) {
		msg_str = Xcalloc(mig_msg->len, char);
		strlcpy(msg_str, (char *)mig_msg->data, mig_msg->len);
		rprintf(resp, "%s", msg_str);
		Xfree(msg_str);
	}
}

/*
 * This is a common routine which sends a mig msg with a string payload back
 * to the source machine. It is used for the following migration message types
 *
 *	MIG_CHECK_INIT
 *	MIG_CHECK_API
 *	MIG_CHECK_CONFIG
 *	MIG_CHECK_BIND
 *	MIG_ERROR_RESP
 *	MIG_HANDOFF
 *
 * Return Values:
 *	0		- Success
 *	ERR_GENERIC	- Failed to send message to source
 */
int
send_mig_msg_resp(mig_seq_args_t *wargs, uint32_t type, uint32_t status)
{
	response_t	*resp;
	mig_conn_t	*mconn;
	mig_msg_t	*mig_msg;
	ldm_hdr_t	*hdr;
	int		resp_str_len = 0;
	char		*resp_str = NULL;
	char		*msg = NULL;
	int		rc;

	DBG2(zprintf("%s:\n", __func__));

	ASSERT(wargs != NULL);
	mconn = wargs->mconn;
	resp = wargs->resp;

	/*
	 * If we're sending back a string we need to get the string length
	 * up front so that we know how much to allocate for the message
	 * and the length of the overall msg.
	 */
	switch (type) {
	case MIG_ERROR_RESP:
		ASSERT(wargs->status == MIG_STATUS_FAILURE);
		resp_str = get_response_text(resp);
		if (resp_str != NULL) {
			resp_str_len = strlen(resp_str) + 1;
			reset_response(resp);
		} else {
			/* the error string has previously been sent */
			return (0);
		}
		break;

	case MIG_HANDOFF:
		/* Error paths should send a MIG_ERROR_RESP message */
		ASSERT(status == MIG_STATUS_SUCCESS);
		ASSERT(wargs->status == MIG_STATUS_SUCCESS);
		DBG(zprintf("sending final handoff message\n"));
		/* FALLTHRU */

	case MIG_CHECK_INIT:
	case MIG_CHECK_API:
	case MIG_CHECK_CONFIG:
	case MIG_CHECK_BIND:
		/* all other message types do not have a string payload */
		break;

	default:
		ASSERT(0);		/* programming error */
		wargs->status = MIG_STATUS_FAILURE;
		return (ERR_GENERIC);
	}

	/* populate the MIG msg header and payload */
	msg = Xmalloc(MIG_MSG_HDR_LEN + resp_str_len);
	hdr = (ldm_hdr_t *)msg;
	hdr->msg_type = hton32(LDM_MIG_MSG);
	hdr->payload_len = hton32(sizeof (mig_msg_t) + resp_str_len);

	mig_msg = (mig_msg_t *)(msg + sizeof (ldm_hdr_t));
	mig_msg->type = hton32(type);
	mig_msg->magic = hton32(MIG_MAGIC);
	mig_msg->status = hton32(status);

	switch (type) {
	case MIG_CHECK_INIT:
	case MIG_CHECK_API:
	case MIG_CHECK_CONFIG:
	case MIG_CHECK_BIND:
	case MIG_HANDOFF:
		mig_msg->len = hton32(0);
		break;
	case MIG_ERROR_RESP:
		ASSERT(wargs->status == MIG_STATUS_FAILURE);
		mig_msg->len = hton32(resp_str_len);
		if (resp_str_len > 0) {
			strlcpy(mig_msg->data, resp_str, resp_str_len);
			DBG3(zprintf("%s: [%s]\n", __func__, mig_msg->data));
		}
		break;
	default:
		ASSERT(0);		/* programming error */
		rc = ERR_GENERIC;
		wargs->status = MIG_STATUS_FAILURE;
		goto done;
	}

	DUMP_MIG_MSG(mig_msg, "OUT");

	rc = mig_write_msg(mconn, msg, MIG_MSG_HDR_LEN + resp_str_len,
	    MIG_HANDSHAKE_TIMEOUT);
	if (rc == -1) {
		wargs->status = MIG_STATUS_FAILURE;
		ewarning("Error sending response type (%d) to source "
		    "LDoms Manager", type);
		rc = ERR_GENERIC;
		goto done;
	}

	rc = 0;
	DBG(zprintf("%s: finished sending resp\n", __func__));

done:
	if (resp_str != NULL)
		Xfree(resp_str);
	if (msg != NULL)
		Xfree(msg);

	return (rc);
}

/*
 * This is a common routine which waits for a mig msg to be send back to
 * to the source machine. It is used for the following migration message types
 *
 *	MIG_CHECK_INIT
 *	MIG_CHECK_API
 *	MIG_CHECK_CONFIG
 *	MIG_CHECK_BIND
 *	MIG_HANDOFF
 *
 * Return Values:
 *	true		- Success
 *	false		- Received MIG_ERROR_RESP or invalid message
 */
bool_t
recv_mig_msg_resp(mig_seq_args_t *args, uint32_t type)
{
	ASSERT(args != NULL);

	mig_conn_t	*mconn = args->mconn;
	mig_msg_t	*mig_msg;

	DBG2(zprintf("%s:\n", __func__));

	/* verify that the type is a supported message */
	ASSERT(type == MIG_CHECK_API || type == MIG_CHECK_CONFIG ||
	    type == MIG_CHECK_BIND || type == MIG_CHECK_INIT ||
	    type == MIG_HANDOFF);

	if (mig_read_mig_msg(mconn, NULL, MIG_MAX_MSG_LEN,
	    MIG_HANDSHAKE_TIMEOUT) == -1) {
		args->status = MIG_STATUS_FAILURE;
		ewarning("Failed to read feasibility response type (%d) "
		    "from target LDoms Manager", type);
		return (false);
	}

	mig_msg = (mig_msg_t *)mconn->read.buf;
	DUMP_MIG_MSG(mig_msg, "IN");

	if (ntoh32(mig_msg->type) == MIG_ERROR_RESP) {
		args->status = MIG_STATUS_FAILURE;
		read_error_mig_msg(mig_msg, args->resp);
		return (false);

	} else if (ntoh32(mig_msg->type) != type) {
		args->status = MIG_STATUS_FAILURE;
		ewarning("Received unknown feasibility response type (%d) "
		    "from target LDoms Manager", hton32(mig_msg->type));
		return (false);
	}

	/*
	 * We should only get here if we are successful. If the operation
	 * failed, a MIG_ERROR_RESP msg should have been sent and dealt
	 * with above.
	 */
	ASSERT(ntoh32(mig_msg->status) == MIG_STATUS_SUCCESS);
	ASSERT(ntoh32(mig_msg->len) == 0);
	DBG5(zprintf("%s successful resp for msg type %d\n", __func__, type));

	return (true);
}
