/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#include <synch.h>
#include <thread.h>
#include <unistd.h>

#include "autosave.h"
#include "backend.h"
#include "database_internal.h"
#include "errors.h"
#include "frags.h"
#include "ldomconfigure.h"
#include "mig_be.h"
#include "mig_class.h"
#include "mig_common.h"
#include "option.h"
#include "sequencer.h"
#include "support.h"
#include "uuid.h"
#include "vcpu.h"
#include "vio.h"
#include "vmau.h"
#include "xml_parse.h"
#include "xml_cntrl_common.h"
#include "xml_v3_defs.h"
#include "xmpp_server.h"
#include "xml_common.h"
#include "xml_event.h"

#define	DS_UP_MAX_RETRIES	5

/*
 * Source ops vector
 */
static bool_t mig_src_warm_init(mig_seq_args_t *args);
static bool_t mig_src_warm_send_config(mig_seq_args_t *args);
static bool_t mig_src_warm_send_bindings(mig_seq_args_t *args);
static bool_t mig_src_warm_feasibility_check(mig_seq_args_t *args);
static bool_t mig_src_warm_bind_check(mig_seq_args_t *args);
static bool_t mig_src_warm_send_state(mig_seq_args_t *args);

mig_src_ops_t mig_src_warm_ops = {
	mig_src_warm_init,		/* mig_src_init */
	mig_src_warm_send_config,	/* mig_src_send_config */
	mig_src_warm_send_bindings,	/* mig_src_send_bindings */
	mig_src_warm_feasibility_check,	/* mig_src_feasibility_check */
	mig_src_warm_bind_check,	/* mig_src_bind_check */
	mig_src_warm_send_state		/* mig_src_send_state */
};

static void send_resource_state(mig_seq_args_t *args, state_restype_t restype,
    int resid, uint64_t mdpa, uint64_t size);

/*
 * Parallel Migration
 *
 * The algorithm uses one writer thread, one reader thread, and a
 * variable number of worker threads that perform compression (on the
 * sender side) and decompression (on the receiving side.) On the
 * sending side, the reader thread reads memory chunks from the
 * hypervisor and hands them off to the worker threads to compress.
 * Once compressed, the memory chunks are handed off to the writer
 * thread which writes out mem_block messages to the network. On
 * the receiving side, the reader thread reads mem_block messages
 * from the network and hands them off to the worker threads to
 * decompress. After decompressing the mem_block data, the worker
 * threads provide memory chunks to the writer thread which writes
 * the memory chunks to memory via the hypervisor.
 *
 * Limit the number of worker threads to 16. The number of buffers
 * allocated is a function of the number of worker threads used. We
 * use (nthreads + 3) buffers to allow for the workers, the writer,
 * and the reader threads to all be busy plus one additional buffer.
 */
#define	MIG_MEM_CHUNKLEN	(4 * 1024 * 1024)
#define	MIG_CLAMP_NWORKERS(nthreads)	\
	MAX(MIN((nthreads), MIG_NWORKERS_MAX), MIG_NWORKERS_MIN)
#define	MIG_NBUFFERS(nthreads)	((nthreads) + 3)
#define	MIG_COMPRESS_MIN_MAJOR	1
#define	MIG_COMPRESS_MIN_MINOR	1
#define	MIG_UUID_MIN_MAJOR	1
#define	MIG_UUID_MIN_MINOR	2
#define	MIG_PROGRESS_SLEEP	1
#define	MIG_COMP_RATE(compsize, decompsize)			\
	((compsize) >= (decompsize) ? 0.0f :			\
	(100.0f * ((float)((decompsize) - (compsize))) /	\
	((float)(decompsize))))
#define	MIG_MSG_GET_PROGRESS_NTOH(msghdrp)			\
	ntoh32(((mig_msg_t *)((msghdrp) +			\
	sizeof (ldm_hdr_t)))->progress)

#define	MIG_ERROR_COMPRESS	1
#define	MIG_ERROR_DECOMPRESS	2
#define	MIG_ERROR_COMPRESS_SIZE	3
#define	MIG_ERROR_CONNECTION	4
#define	MIG_ERROR_CANCEL	5
#define	MIG_ERROR_OTHER		6

static int mig_progress_sleep = MIG_PROGRESS_SLEEP;
static size_t mig_mem_chunklen = MIG_MEM_CHUNKLEN;

/* Parallel migration buffer struct */
typedef struct pmig_buf {
	uint8_t		*msgp;			/* pointer to mem block msg */
	uint8_t		*datap;			/* pointer to mem data */
	uint32_t	size_decomp;		/* decompressed size */
	uint32_t	size_comp;		/* compressed size */
	uint64_t	base_ra;		/* mem chunk base addr */
	uint64_t	offset;			/* mem chunk offset */
	CHAIN_LINK(struct pmig_buf, link);
} pmig_buf_t;

/* Parallel migration control struct */
typedef struct {
	ldom_t		*ldp;			/* domain being migrated */
	mig_conn_t	*mconn;			/* connection struct */

	bool_t		workers_eof;		/* signals workers to exit */
	bool_t		writer_eof;		/* signals writer to exit */
	int		error;			/* worker or writer error */

	bool_t		progress_eof;		/* signals progress to exit */
	mutex_t		progress_lock;		/* protects progress_percent */
	uint32_t	progress_percent;	/* overall memory progress % */

	CHAIN(pmig_buf_t, empty_chain);		/* empty buffers */
	CHAIN(pmig_buf_t, unproc_chain);	/* unprocessed buffers */
	CHAIN(pmig_buf_t, proc_chain);		/* processed buffers */
	cond_t		empty_cv;		/* cond: empty bufs avail */
	cond_t		unproc_cv;		/* cond: unproc bufs avail */
	cond_t		proc_cv;		/* cond: proc bufs avail */
	mutex_t		empty_lock;		/* protects empty chain and */
						/* the error field */
	mutex_t		unproc_lock;		/* protects unproc chain */
	mutex_t		proc_lock;		/* protects proc chain */
} pmig_ctrl_t;


/* Parallel migration params struct */
typedef struct {
	bool_t		compression_enabled;	/* is compression enabled? */
	uint32_t	max_block_msg_size;	/* max mblock msg total size */
	uint32_t	max_chunklen;		/* max mblock msg chunklen */
	int		nbuffers;		/* number of buffers to use */
	int		nworkers;		/* number of worker threads */
} pmig_params_t;


/*
 * Returns
 *	true	- target supports API Group versions negotiated
 *	false	- feasibility check failed on the target
 */
static bool_t
i_src_check_api_groups_supported(mig_seq_args_t *args)
{
	ASSERT(args != NULL);

	mig_conn_t			*mconn = args->mconn;
	ldom_t				*ldp = args->ldp;
	hv_api_group_version_t		*versionp = NULL;
	mig_msg_api_group_vers_t	*payload;
	uint32_t		resid = (uint32_t)ldp->res_id->id;
	uint32_t		num_api_groups;
	mig_msg_t		*mig_msg;
	ldm_hdr_t		*hdr;
	char			*buf;
	int			vlen, msg_len;

	/* Check if API Group versions are supported on target machine */
	num_api_groups = call_hv_get_api_groups(resid, &versionp);
	DBG5(zprintf("Got API versions for %d groups\n", num_api_groups));

	if (num_api_groups == 0 || versionp == NULL) {
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	/*
	 * calculate the amount of space needed to store the API group
	 * version count (uint64_t) and the info for each group.
	 */
	vlen = (int)sizeof (uint64_t);		/* num_api_groups */
	vlen += num_api_groups * (int)sizeof (hv_api_group_version_t);

	msg_len = MIG_MSG_HDR_LEN + vlen;
	buf = Xmalloc(msg_len);

	hdr = (ldm_hdr_t *)buf;
	hdr->msg_type = hton32(LDM_MIG_MSG);
	hdr->payload_len = hton32(sizeof (mig_msg_t) + vlen);

	mig_msg = (mig_msg_t *)(buf + sizeof (ldm_hdr_t));
	mig_msg->type = hton32(MIG_CHECK_API);
	mig_msg->len = hton32(vlen);
	mig_msg->magic = hton32(MIG_MAGIC);
	mig_msg->status = 0;
	payload = (mig_msg_api_group_vers_t *)(buf + MIG_MSG_HDR_LEN);
	payload->group_size = num_api_groups;
	/*
	 * copy the version entries (subtracting the size of num_api_groups
	 * from the overall length of the structure).
	 */
	memcpy(payload->entries, versionp, vlen - (int)sizeof (uint64_t));

	DUMP_API_GROUPS(num_api_groups, payload->entries);
	DUMP_MIG_MSG(mig_msg, "OUT");

	if (mig_write_msg(mconn, buf, msg_len, MIG_HANDSHAKE_TIMEOUT) == -1) {
		args->status = MIG_STATUS_FAILURE;
		ewarning("Failed to send API Group feasibility request to "
		    "logical domain manager");
		Xfree(buf);
		Xfree(versionp);
		return (false);
	}

	Xfree(buf);
	Xfree(versionp);

	DBG5(zprintf("MIG_CHECK_API req sent\n"));

	return (recv_mig_msg_resp(args, MIG_CHECK_API));
}

/*
 * Returns
 *	true	- target supports API Group versions negotiated
 *	false	- feasibility check failed on the target
 */
static bool_t
i_src_check_config_supported(mig_seq_args_t *args)
{
	ASSERT(args != NULL);

	mig_conn_t *mconn = args->mconn;
	mig_msg_mach_config_t	*payload;
	mig_msg_t	*mig_msg;
	ldm_hdr_t	*hdr;
	char		*buf;
	int		vlen;		/* length of payload */
	int		msg_len;	/* length of entire mig_msg */
	int		cpuname_len;	/* length of compatible prop */

	/*
	 * calculate the amount of space needed to store the config
	 * struct plus the compatible string (including NUL)
	 */
	vlen = (int)sizeof (mig_msg_mach_config_t);
	cpuname_len = strlen(pri.cpu_compatible) + 1;
	vlen += cpuname_len;

	msg_len = MIG_MSG_HDR_LEN + vlen;
	buf = Xmalloc(msg_len);

	hdr = (ldm_hdr_t *)buf;
	hdr->msg_type = hton32(LDM_MIG_MSG);
	hdr->payload_len = hton32(sizeof (mig_msg_t) + vlen);

	mig_msg = (mig_msg_t *)(buf + sizeof (ldm_hdr_t));
	mig_msg->type = hton32(MIG_CHECK_CONFIG);
	mig_msg->len = hton32(vlen);
	mig_msg->magic = hton32(MIG_MAGIC);
	mig_msg->status = 0;
	payload = (mig_msg_mach_config_t *)(buf + MIG_MSG_HDR_LEN);

	/* copy the CPU's compatible string into the payload */
	payload->stick_freq = pri.stick_frequency;
	memcpy(payload->cpu_compatible, pri.cpu_compatible, cpuname_len);

	DUMP_MIG_MSG(mig_msg, "OUT");

	if (mig_write_msg(mconn, buf, msg_len, MIG_HANDSHAKE_TIMEOUT) == -1) {
		args->status = MIG_STATUS_FAILURE;
		ewarning("Failed to send API Group feasibility request to "
		    "logical domain manager");
		Xfree(buf);
		return (false);
	}

	Xfree(buf);

	DBG5(zprintf("MIG_CHECK_CONFIG req sent\n"));

	return (recv_mig_msg_resp(args, MIG_CHECK_CONFIG));
}

/*
 * Returns
 *	true	- all feasibility checks succeeded on the target
 *	false	- a feasibility check failed on the target (stop upon 1st fail)
 */
static bool_t
i_src_check_migration_feasibility(mig_seq_args_t *args)
{
	ASSERT(args != NULL);

	/*
	 * Check if config details such as CPU Frequency, etc
	 * match those on target machine
	 */
	if (i_src_check_config_supported(args) == false)
		return (false);

	/* Check if the HV API Group versions are supported on the target */
	if (i_src_check_api_groups_supported(args) == false)
		return (false);

	return (true);
}

static void
download_hvmd(md_t *hvmdp)
{
	int	size;
	uint8_t	*bufp;

	/* download the new HV MD */
	size = md_gen_bin(hvmdp, &bufp);
	if (size <= 0) {
		fatal("Unable to generate HV binary MD using internal MD");
	} else if (size > (max_hv_md_size - pri.md_elbow_room)) {
		fatal("Configuration required a MD of size 0x%x bytes which\n"
		    "is larger than the maximum supported size of"
		    "0x%x bytes", size, max_hv_md_size - pri.md_elbow_room);
	}

	if (zeus.next_hvmd_fp == NULL)
		zeus.next_hvmd_fp = hv_md_alloc();

	backend.dl_hvmd(zeus.next_hvmd_fp->paddr, bufp, size, HVMD);
	backend.config_guests(CONFIG);

	hv_md_free(zeus.current_hvmd_fp);
	zeus.current_hvmd_fp = zeus.next_hvmd_fp;
	zeus.next_hvmd_fp = NULL;

	Xfree(bufp);
}

static void
clear_mig_flag_in_hvmd(ldom_t *ldp)
{
	md_node_t	*guest;
	md_prop_t	*prop;
	char		*flag;

	/* check if this is the source or target */
	flag = (ldp->flag.migration_target) ?
	    HVMD_PROP_MIG_TGT : HVMD_PROP_MIG_SRC;

	DBG(zprintf("%s: %s\n", __func__, flag));

	guest = find_guest_md_node(zeus.hv_mdp, ldp);

	/* guest should be in HV MD */
	ASSERT(guest != NULL);

	/* lookup the target flag */
	prop = md_find_node_prop(zeus.hv_mdp, guest,
	    md_find_name(zeus.hv_mdp, flag), MDET_PROP_VAL);

	ASSERT(prop != NULL);

	/* remove the flag */
	md_delete_prop(zeus.hv_mdp, guest, prop);

	download_hvmd(zeus.hv_mdp);
}

static int
save_mig_recovery_md(ldom_t *ldp)
{
	int	size;
	uint8_t	*bufp;

	ASSERT(ldp->mig_state.bind_mdp != NULL);
	ASSERT(ldp->mig_state.bind_fp == NULL);

	size = md_gen_bin(ldp->mig_state.bind_mdp, &bufp);
	if (size <= 0) {
		DBG(zprintf("Unable to generate MD for MIG recovery\n"));
		return (-1);
	}

	/* allocate a frag for this MD */
	ldp->mig_state.bind_fp = frag_alloc(ldp, size);
	if (ldp->mig_state.bind_fp == NULL) {
		DBG(zprintf("Unable to allocate frag for MIG recovery MD\n"));
		return (-1);
	}

	/* XXXMIG - create a new interface to the backend? */
	backend.dl_guestmd(ldp, ldp->mig_state.bind_fp->paddr, bufp, size);
	return (0);
}

/* MIGXXX - move someplace more appropriate */
extern int call_hv_guest_suspend(ldom_t *ldp);
extern int call_hv_guest_resume(ldom_t *ldp);

static int
suspend_domain(ldom_t *ldp, response_t *resp)
{
	int	cc;

	/* XXXMIG - a check for multi-strand suspend would go here */

	/* remove all the cryptos from the domain */
	if (mau_mig_del(ldp, resp) == false) {
		rprintf(resp, "Unable to DR out all the crypto units.\n");
		return (-1);
	}

	/* remove all but one cpu from the domain */
	if (cpu_mig_reduce_to_one(ldp, resp) != 0) {
		rprintf(resp, "Unable to reduce the number of CPUs to one.\n");
		return (-1);
	}

	/*
	 * This is a critical point in the migration. Once the domain
	 * has been isolated, the HV MD and guest MD are out of sync.
	 * In order to restore the HV state if Zeus restarts, set a
	 * persistent flag in the HV MD to force some special recovery
	 * actions in the startup flow.
	 */
	ldp->mig_state.suspended = true;

	/* store the binding MD for error recovery */
	if (save_mig_recovery_md(ldp) != 0) {
		return (-1);
	}

	/* remove all connections from clients to servers */
	/* XXXMIG - add error checking on this call */
	DBG(zprintf("isolating LDom %s...\n", ldp->namep));
	seq_isolate_ldom(ldp, resp);

	/* remove all channels between vnet peers for the domain */

	ldp->state = SUSPENDING;

	/* suspend the domain */
	DBG(zprintf("suspending LDom %s...\n", ldp->namep));
	cc = call_hv_guest_suspend(ldp);
	if (cc != HVctl_st_ok) {
		rprintf(resp, "HVctl suspend command failed\n");
		return (-1);
	}

	ldp->state = SUSPENDED;

	return (0);
}

#if ZDBG
void
store_resource_state(ldom_t *ldp, state_restype_t restype, int resid,
    uint64_t paddr, uint64_t size)
{
	char	fname[MAXPATHLEN];
	char	rnamebuf[32];
	char	*rname;

	switch (restype) {
	case STATE_RES_GUEST:
		rname = "guest";
		break;
	case STATE_RES_VCPU:
		snprintf(rnamebuf, 32, "vcpu%d", resid);
		rname = rnamebuf;
		break;
	case STATE_RES_LDC:
		snprintf(rnamebuf, 32, "ldc%d", resid);
		rname = rnamebuf;
		break;
	default:
		DBG(zprintf("Unknown resource type (%d), unable to store "
		    "resource state\n", (int)restype));
		return;
	}

	snprintf(fname, MAXPATHLEN, "/export/%s-%s-state.bin",
	    ldp->namep, rname);

	do_read_mem_to_file(fname, paddr, size);
}
#endif

static void
send_state_done(mig_seq_args_t *args)
{
	uint8_t			*bufp;
	mig_msg_saved_state_t	*msg_saved;
	mig_msg_t		*mig_msg;
	mig_conn_t		*mconn;
	int			msg_len;

	ASSERT(args != NULL);
	mconn = args->mconn;

	bufp = init_mig_msg(MIG_SAVED_STATE, sizeof (mig_msg_saved_state_t),
	    MIG_STATUS_SUCCESS);
	msg_len = MIG_MSG_HDR_LEN + sizeof (mig_msg_saved_state_t);

	/* fill in the message specific data */
	msg_saved = (mig_msg_saved_state_t *)(bufp + MIG_MSG_HDR_LEN);
	msg_saved->restype = hton32(0x1212); /* XXXMIG - quick EOF hack */

	mig_msg = (mig_msg_t *)(bufp + sizeof (ldm_hdr_t));
	DUMP_MIG_MSG(mig_msg, "OUT");

	if (mig_write_msg(mconn, bufp, msg_len, MIG_HANDSHAKE_TIMEOUT) == -1) {
		args->status = MIG_STATUS_FAILURE;
		ASSERT(args->peer_hostname != NULL);
		ewarning("Failed to send resource state to target host: %s",
		    args->peer_hostname);
	}

	Xfree(bufp);
}

/* XXXMIG - put this in the migration structure */
bound_mfrag_t	*mig_frag = NULL;

/*
 * XXXMIG - Finalize the location of these files and make
 * sure that they are relative to the package install dir,
 * even if an alternate root is selected.
 */
#define	GUEST_STATE_MD	"/opt/SUNWldm/lib/templates/guest-state.bin"
#define	VCPU_STATE_MD	"/opt/SUNWldm/lib/templates/vcpu-state.bin"
#define	LDC_STATE_MD	"/opt/SUNWldm/lib/templates/ldc-state.bin"

extern int call_hv_state_save(ldom_t *ldp, uint64_t paddr,
    state_restype_t restype, uint32_t resid);

/* MIGXXX can't really return -1 in a size_t */
size_t
save_resource_state(ldom_t *ldp, state_restype_t rtype, int resid)
{
	char		*file;
	uint64_t	filesize;
	int		fh;
	struct stat	status;
	int		cc;

	switch (rtype) {
	case STATE_RES_GUEST:
		file = GUEST_STATE_MD;
		break;
	case STATE_RES_VCPU:
		file = VCPU_STATE_MD;
		break;
	case STATE_RES_LDC:
		file = LDC_STATE_MD;
		break;
	default:
		DBG(zprintf("Unknown resource type (%d), unable to save "
		    "resource state\n", (int)rtype));
		return (-1);
	}

	/*
	 * Allocate a frag and write the MD file to that location
	 * in memory.
	 */
	fh = open(file, O_RDONLY);
	if (fh < 0) {
		DBG(zprintf("Opening %s : %s\n", file, strerror(errno)));
		return (-1);
	}

	if (fstat(fh, &status) < 0) {
		DBG(zprintf("Stat'ing %s : %s\n", file, strerror(errno)));
		return (-1);
	}
	filesize = status.st_size;
	close(fh);

	/* free any old frag before we start */
	if (mig_frag != NULL)
		frag_free(mig_frag);

	mig_frag = frag_alloc(ldp, filesize);
	if (mig_frag == NULL) {
		DBG(zprintf("Unable to allocate frag for state MD\n"));
		return (-1);
	}

	if (do_write_file_to_addr(file, mig_frag->paddr) == false)
		return (-1);

	/* issue the HV call to save the state */
	cc = call_hv_state_save(ldp, mig_frag->paddr, rtype, resid);
	if (cc != HVctl_st_ok) {
		DBG(zprintf("HVctl command failed: %d\n", cc));
		return (-1);
	}

	return (filesize);
}

static bool_t
save_ldc_state(mig_seq_args_t *args)
{
	ASSERT(args != NULL);
	ASSERT(args->ldp != NULL);

	ldom_t		*ldp = args->ldp;
	md_t		*mdp;
	md_node_t	**gnodesp;
	md_node_t	**lnodesp;
	int		nldcs;
	int		nguests;
	uint64_t	id;
	uint64_t	chan;
	uint32_t	gid;
	int		i, j;
	size_t		mdsz;

	gid = ldp->res_id->id;

	/*
	 * Scan the current HV MD looking for all endpoints that
	 * belong to the specified guest.
	 */
	mdp = zeus.hv_mdp;
	gnodesp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));
	lnodesp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	nguests = md_scan_dag(mdp, md_root_node(mdp),
	    md_find_name(mdp, "guest"),
	    md_find_name(mdp, "fwd"), gnodesp);

	/* walk the list of guests in the system */
	for (i = 0; i < nguests; i++) {

		if (md_get_prop_val(mdp, gnodesp[i], "gid", &id) != 0)
			fatal("guest node has no 'gid' property");

		/* check if this is the specified guest */
		if (gid != id)
			continue;

		nldcs = md_scan_dag(mdp, gnodesp[i],
		    md_find_name(mdp, "ldc_endpoint"),
		    md_find_name(mdp, "fwd"), lnodesp);

		/* walk the list of ldcs in this guest */
		for (j = 0; j < nldcs; j++) {
			if (md_get_prop_val(mdp, lnodesp[j], "channel", &chan)
			    != 0) {
				fatal("ldc node has no 'channel' property");
			}

			/*
			 * XXXMIG - is it necessary to check the type of
			 * channel? For example, does it make sense to
			 * migrate SP channels? Does it even make sense
			 * to migrate a domain that has SP channels?
			 */

			/* save the state of this ldc */
			DBG(zprintf("Saving state for ldc %"PRId64":\n", chan));

			mdsz = save_resource_state(ldp, STATE_RES_LDC, chan);
			if (mdsz == -1) {
				args->status = MIG_STATUS_FAILURE;
				return (false);
			}

			DBG(store_resource_state(ldp, STATE_RES_LDC, chan,
			    mig_frag->paddr, mdsz));

			send_resource_state(args, STATE_RES_LDC, chan,
			    mig_frag->paddr, mdsz);

			DBG(zprintf("\n"));
		}
	}

	Xfree(gnodesp);
	Xfree(lnodesp);

	return (true);
}

/*
 * Stop the specified LDom and wait until the operation completes.
 * Since this is called from within the sequencer, the asynchronous
 * notification will never be delivered. Consequently, this function
 * manually polls the HV for the state of the domain.
 */
static int
do_stop_ldom(ldom_t *ldp, cli_conn_t *client)
{
	bool_t		force = true;

	ASSERT_IN_SEQUENCER();

	CHAIN_ADD(zeus.inprogress, ldp, inprog);
	seq_stop_bound(force);

	ASSERT(CHAIN_LENGTH(zeus.inprogress) == 1);

	while (ldp->state == STOPPING) {
		backend.get_guest_state(ldp, LDOM_HV_STATE, &ldp->hv_state);

		switch (ldp->hv_state.state) {
		case GUEST_STATE_STOPPED:
			seq_ldom_state_change(ldp, client, GUEST_STATE_STOPPED);
			DBG(zprintf("LDom %s stopped\n", ldp->namep));
			break;

		case GUEST_STATE_RESETTING:
			/* don't think we should ever get here */
			ASSERT(0);
			ldp->state = ACTIVE;
			DBG(zprintf("LDom %s resetting, stop failed\n",
			    ldp->namep));
			return (-1);

		default:
			/* keep trying */
			/* XXXMIG - eventually, this should time out */
			sleep(1);
			DBG(zprintf("checking if domain has stopped...\n"));
			break;
		}
	}

	if (ldp->state != BOUND) {
		DBG(zprintf("LDom %s stop notification failed\n", ldp->namep));
		return (-1);
	}

	return (0);
}

int
resume_domain(mig_seq_args_t *args)
{
	int		cc = -1;
	int		num_req;
	ldom_t		*ldp = args->ldp;
	md_t		*mdp = ldp->mig_state.bind_mdp;
	response_t	*resp = args->resp;
	bool_t		ldom_locked_by_caller;
	ldm_err_t	pm_result = { 0 }; /* Not an array because currently */
					/* only PM_RES_CPU resource type is */
					/* affected by a domain state change. */

	ASSERT_IN_SEQUENCER();

	/* notify PM that the domain is starting */
	if (pm_domain_events_on) {
		send_pm_dom_event(ldp, START_DOMAIN_PRE, &pm_result);
		/*
		 * FIXME: check for errors by checking pm_result.ldme_errs
		 * and traversing pm_result.ldme_private.
		 */
		if (pm_result.ldme_errs != LDME_OP_EOK) {
			rprintf(resp, "Power Management failure when "
			    "resuming LDom %s\n", ldp->namep);
			args->status = MIG_STATUS_FAILURE;
			pm_free_errors(&pm_result);
			return (cc);
		}
		pm_free_errors(&pm_result);
	}

	/* resume the domain */
	DBG(zprintf("resuming LDom %s...\n", ldp->namep));
	cc = call_hv_guest_resume(ldp);
	if (cc != HVctl_st_ok) {
		rprintf(resp, "HVctl resume command failed\n");
		args->status = MIG_STATUS_FAILURE;
		return (cc);
	}

	/* restore connections from clients to servers */
	DBG(zprintf("restoring LDom %s...\n", ldp->namep));

	/*
	 * We need to unlock the LDom in case multiple VIO devices are
	 * specified (adding the second and subsequent devices would
	 * be blocked otherwise).  It is safe to do this since we are
	 * on the sequencer and are protected.
	 */
	ldom_locked_by_caller = IS_LDOM_LOCKED(ldp);
	UNLOCK_LDOM(ldp);
	seq_restore_ldom(ldp, resp);

	/* restore lock if necessary */
	if (ldom_locked_by_caller == true)
		LOCK_LDOM(ldp);

	clear_mig_flag_in_hvmd(ldp);

	if (ldp->mig_state.bind_fp != NULL) {
		frag_free(ldp->mig_state.bind_fp);
		ldp->mig_state.bind_fp = NULL;
	}

	ldp->flag.md_notify = true;
	seq_send_md_update_notifs();

	/* lookup how many CPUs were in the source domain */
	num_req = md_scan_dag(mdp, NULL, md_find_name(mdp, "cpu"),
	    md_find_name(mdp, "fwd"), NULL);

	/* add back any vcpus that were removed */
	if (ldp->state != ACTIVE || !cpu_mig_restore(ldp, num_req, resp)) {
		rprintf(resp, "Unable to restore original CPU configuration\n");
		args->status = MIG_STATUS_FAILURE;
		return (cc);
	}

	/*
	 * Add crypto units back in
	 * lookup how many MAUs were in the source domain
	 */
	num_req = grab_exec_units(mdp, "mau", NULL);

	/* add back any crypto units that were removed */
	if (ldp->state != ACTIVE || !mau_mig_add(ldp, num_req, resp)) {
		rprintf(resp, "Unable to restore original crypto "
		    "configuration\n");
		args->status = MIG_STATUS_FAILURE;
		return (cc);
	}

	return (0);
}

static void
migration_complete(void *args)
{
	cli_result_t	result = CLI_RESULT_FAIL;
	mig_seq_args_t	*mig = args;
	char		*ldom_name;
	ldom_t		*ldp;

	ASSERT((mig != NULL) && (mig->ldp != NULL));

	ldp = mig->ldp;


	/*
	 * On a successful migration, the ldom structure gets
	 * destroyed. Before that happens, save off a copy of
	 * any information needed to send back the response to
	 * the client.
	 */
	ldom_name = Xstrdup(ldp->namep);

	if (mig->status == MIG_STATUS_SUCCESS) {
		/*
		 * Migration succeeded. Perform the various
		 * steps required to teardown the source.
		 */
		DBG(zprintf("tearing down source LDom %s\n", ldp->namep));

		if (ldp->mig_state.bind_fp != NULL) {
			frag_free(ldp->mig_state.bind_fp);
			ldp->mig_state.bind_fp = NULL;
		}

		do_stop_ldom(ldp, mig->client);

		/* unbind the source domain */
		ldp->state = UNBINDING;
		seq_unbind_ldoms();

		cleanse_progress_chain();
		UNLOCK_LDOM(ldp);

		/* Send a domain event. */
		event_send_dom_event(XML_MIGRATE_DOMAIN, ldp, mig->client);
		event_send_progress_event(MIGRATION_PROGRESS, ldp,
		    XML_COMPLETE);

		/* destroy the source domain */
		if (destroy_one_ldom(ldp, mig->client->resp, mig->client) ==
		    true) {
			mig->ldp = NULL;
			result = CLI_RESULT_OK;
		} else {
			mig->status = MIG_STATUS_FAILURE;
			DBG(zprintf("unable to destroy source LDom\n"));
		}
	} else {
		/*
		 * Migration failed or was canceled. Resume the
		 * domain back to its active state.
		 */
		ASSERT(mig->status == MIG_STATUS_FAILURE);

		(void) resume_domain(args);
		UNLOCK_LDOM(ldp);
		ldp->flag.migration_source = false;
		event_send_progress_event(MIGRATION_PROGRESS, ldp,
		    (xmlChar *)XML_FAILURE);

		/* Due to failure ldp stays around. Need to clear mig_info */
		ldp->mig_info.total_mem = 0;
		ldp->mig_info.migrated_mem = 0;
		ldp->mig_info.progress = 0;	/* initialise to 0% */

	}

	xml_send_async_cmd_result(mig->client, XML_MIGRATE_DOMAIN,
	    ldom_name, result);

	Xfree(ldom_name);

	if (mig->mconn != NULL)
		mig_close_conn(mig->mconn);

	fini_response(mig->resp);

	free_mig_seq_args(mig);
}

/*
 * Migrate the guest state.
 */
static void
i_src_do_migrate_state(void *args)
{
	mig_seq_args_t	*mig = args;
	ldom_t		*ldp;
	uint64_t	filesize;
	int		resid;
	state_restype_t	restype;
	vcpu_t		*vcpup;
	response_t	*resp;

	ASSERT_IN_SEQUENCER();

	ldp = mig->ldp;
	resp = mig->resp;
	resid = ldp->res_id->id;
	restype = STATE_RES_GUEST;

	DBG(zprintf("Saving state for guest '%s':\n", ldp->namep));

	/* save the resource state */
	filesize = save_resource_state(ldp, restype, resid);
	if (filesize == -1) {
		mig->status = MIG_STATUS_FAILURE;
		DBG(zprintf("unable to save GUEST resource state\n"));
		rprintf(resp, MSG_ERR_MIG_CANNOT_SAVE_STATE, "GUEST");
		goto done;
	}

	/* XXXMIG - dump the state to a file for now */
	DBG(store_resource_state(ldp, restype, resid, mig_frag->paddr,
	    filesize));
	send_resource_state(mig, restype, resid, mig_frag->paddr, filesize);

	/*
	 * Migrate the vcpu state.
	 */

	/* domain should have been shrunk to one cpu by now */
	ASSERT(CHAIN_LENGTH(ldp->bound.vcpus) == 1);

	vcpup = CHAIN_START(ldp->bound.vcpus);
	if (vcpup == NULL) {
		mig->status = MIG_STATUS_FAILURE;
		DBG(zprintf("LDom %s does not have any bound vcpus\n",
		    ldp->namep));
		rprintf(resp, MSG_ERR_MIG_NO_BOUND_CPUS, ldp->namep);
		goto done;
	}

	resid = vcpup->vid;
	restype = STATE_RES_VCPU;

	DBG(zprintf("Saving state for vcpu %d:\n", resid));

	/* save the resource state */
	filesize = save_resource_state(ldp, restype, resid);
	if (filesize == -1) {
		mig->status = MIG_STATUS_FAILURE;
		DBG(zprintf("unable to save VCPU resource state\n"));
		rprintf(resp, MSG_ERR_MIG_CANNOT_SAVE_STATE, "VCPU");
		goto done;
	}

	/* XXXMIG - dump the state to a file for now */
	DBG(store_resource_state(ldp, restype, resid, mig_frag->paddr,
	    filesize));
	send_resource_state(mig, restype, resid, mig_frag->paddr, filesize);

	/*
	 * Save the LDC state.
	 */
	if (save_ldc_state(args) == false)
		goto done;

	send_state_done(mig);

	/* wait for the final handoff message */
	(void) recv_mig_msg_resp(mig, MIG_HANDOFF);

done:
	/*
	 * The state transfer has now terminated (successfully or not).
	 * Need to call into the sequencer to either complete the
	 * migration work or clean up after the failure.
	 */
	migration_complete(mig);
}

#define	MIG_STATE_CHUNKLEN	(2 * 1024 * 1024)

static void
send_resource_state(mig_seq_args_t *args, state_restype_t restype, int resid,
    uint64_t mdpa, uint64_t size)
{
	int			vlen;
	mig_conn_t		*mconn;
	mig_msg_t		*mig_msg;
	mig_msg_saved_state_t	*msg_saved;
	int			msg_len;
	uint64_t		base;
	int			amt;
	uint8_t			*bufp;
	uint8_t			*curr;

	/* The size of the state MDs are pretty small */
	ASSERT(size <= MIG_STATE_CHUNKLEN);

	ASSERT(args != NULL);
	mconn = args->mconn;

	/* calculate the variable data length */
	vlen = (int) (sizeof (mig_msg_saved_state_t) + size);

	bufp = init_mig_msg(MIG_SAVED_STATE, vlen, MIG_STATUS_SUCCESS);

	/* fill in the message specific data */
	msg_saved = (mig_msg_saved_state_t *)(bufp + MIG_MSG_HDR_LEN);
	msg_saved->restype = hton32(restype);
	msg_saved->resid = hton32(resid);
	msg_saved->size = hton64(size);

	/*
	 * Read in the state MD.
	 */
	base = mdpa;
	curr = msg_saved->statemd;
	while (size > 0) {
		amt = backend.read_mem(base, curr, size);
		if (amt != size)
			DBG(zprintf("Warning: read only %d\n", amt));

		size -= amt;
		base += amt;
		curr += amt;
	}

	mig_msg = (mig_msg_t *)(bufp + sizeof (ldm_hdr_t));
	msg_len = MIG_MSG_HDR_LEN + vlen;

	DUMP_MIG_MSG(mig_msg, "OUT");

	if (mig_write_msg(mconn, bufp, msg_len, MIG_HANDSHAKE_TIMEOUT) == -1) {
		args->status = MIG_STATUS_FAILURE;
		ASSERT(args->peer_hostname != NULL);
		ewarning("Failed to send resource state to target host: %s",
		    args->peer_hostname);
	}

	Xfree(bufp);
}

/*
 * This thread runs while the memory transfer is in progress. It is
 * responsible for updating the mig_info progress field as well as
 * sending progress events. It monitors pmig->progress_percent which
 * is updated by the writer thread. It is used on the source and on
 * the target.
 */
static void *
migrate_progress_thread(void *args)
{
	pmig_ctrl_t	*pmig;
	uint32_t	progress_percent;
	bool_t		is_last_update;

	pmig = (pmig_ctrl_t *)args;

	/*
	 * Loop until pmig->progress_eof is non-zero. Then, check the
	 * progress one last time to ensure the final progress percentage
	 * is sent.
	 */
	while (1) {
		is_last_update = (pmig->progress_eof == true);

		mutex_lock(&pmig->progress_lock);
		progress_percent = pmig->progress_percent;
		mutex_unlock(&pmig->progress_lock);
		ASSERT(progress_percent <= 100);
		ASSERT(progress_percent >= pmig->ldp->mig_info.progress);

		if (progress_percent != pmig->ldp->mig_info.progress) {
			pmig->ldp->mig_info.progress = progress_percent;
			event_send_progress_event(MIGRATION_PROGRESS,
			    pmig->ldp, XML_ONGOING);
		}

		if (is_last_update)
			break;

		sleep(mig_progress_sleep);
	}

	return (NULL);
}

/*
 * This thread runs on the sender side while the memory transfer is in
 * progress. It takes buffers from the processed chain and writes them
 * out to the socket as mblock messages. After each mblock message is
 * sent, the progress_percent is updated.
 */
static void *
src_mblock_write_thread(void *args)
{
	pmig_ctrl_t		*pmig;
	pmig_buf_t		*buf;
	uint32_t		progress_percent;
	uint64_t		progress_bytes;
	uint64_t		total_bytes;
	mblock_t		*mbp;

	pmig = (pmig_ctrl_t *)args;
	progress_percent = progress_bytes = 0;

	/* determine the total number of bytes that will be migrated */
	total_bytes = 0;
	CHAIN_ITER(pmig->ldp->bound.memory, mbp, ldom) {
		total_bytes += mbp->size;
	}
	DBG2(zprintf("total amt mem to migrate=%" PRId64 "\n", total_bytes));

	while (1) {
		uint8_t			*msgp;
		ldm_hdr_t		*hdr;
		mig_msg_t		*mig_msg;
		mig_msg_mem_block_t	*msg_mchunk;

		mutex_lock(&pmig->proc_lock);
		while (CHAIN_LENGTH(pmig->proc_chain) == 0) {
			if (pmig->writer_eof == true) {
				DBG(zprintf("%s: eof\n", __func__));
				mutex_unlock(&pmig->proc_lock);
				return (NULL);
			}
			DBG3(zprintf("%s: sleeping\n", __func__));
			(void) cond_wait(&pmig->proc_cv, &pmig->proc_lock);
		}

		/* get buf from processed chain */
		buf = CHAIN_START(pmig->proc_chain);
		CHAIN_UNLINK(pmig->proc_chain, buf, link);
		mutex_unlock(&pmig->proc_lock);

		msgp = buf->msgp;
		hdr = (ldm_hdr_t *)msgp;
		mig_msg = (mig_msg_t *)(msgp + sizeof (ldm_hdr_t));
		msg_mchunk = (mig_msg_mem_block_t *)(msgp + MIG_MSG_HDR_LEN);

		hdr->payload_len = hton32(sizeof (mig_msg_t) +
		    sizeof (mig_msg_mem_block_t) + buf->size_comp);

		mig_msg->len = hton32(sizeof (mig_msg_mem_block_t) +
		    buf->size_comp);
		mig_msg->progress = hton32(progress_percent);

		msg_mchunk->size = hton64(buf->size_comp);
		msg_mchunk->base_ra = hton64(buf->base_ra);
		msg_mchunk->offset = hton64(buf->offset);
		msg_mchunk->eof = 0;

		/* send the message to the target */
		if (mig_write_msg(pmig->mconn, msgp,
		    hdr->payload_len + sizeof (ldm_hdr_t),
		    MIG_HANDSHAKE_TIMEOUT) == -1) {
			ewarning("Migration aborted: write error, errno: %s\n",
			    strerror(errno));

			mutex_lock(&pmig->empty_lock);
			if (pmig->error == 0) {
				pmig->error = MIG_ERROR_CONNECTION;
				cond_signal(&pmig->empty_cv);
			}
			mutex_unlock(&pmig->empty_lock);
			break;
		}

		/* update progress */
		progress_bytes += buf->size_decomp;
		progress_percent = (int)((progress_bytes * 100) / total_bytes);
		if (progress_bytes < total_bytes) {
			mutex_lock(&pmig->progress_lock);
			pmig->progress_percent = progress_percent;
			mutex_unlock(&pmig->progress_lock);
		}

		DBG(zprintf("base=0x%"PRIx64", "
		    "offset=0x%"PRIx64", size=0x%x, "
		    "compress=%.2f%%, progress=%d%%\n",
		    buf->base_ra, buf->offset, buf->size_comp,
		    MIG_COMP_RATE(buf->size_comp, buf->size_decomp),
		    progress_percent));

		/* add buf to empty chain */
		mutex_lock(&pmig->empty_lock);
		CHAIN_ADD(pmig->empty_chain, buf, link);
		cond_signal(&pmig->empty_cv);
		mutex_unlock(&pmig->empty_lock);
	}

	return (NULL);
}

/*
 * Used by the sender/source only, this thread takes buffers from the
 * unprocessed chain, compresses them, and then places them on the
 * processed chain. After compressing the buffer, buf->size_comp is
 * updated to reflect the compressed size of the memory chunk. The
 * writer thread uses this size to fill in the length fields in the
 * message struct.
 */
static void *
src_mblock_compress_thread(void *args)
{
	pmig_ctrl_t		*pmig;
	pmig_buf_t		*buf;
	mig_msg_mem_block_t	*msg_mchunk;
	int			rv;

	pmig = (pmig_ctrl_t *)args;

	while (1) {
		mutex_lock(&pmig->unproc_lock);
		while (CHAIN_LENGTH(pmig->unproc_chain) == 0) {
			if (pmig->workers_eof == true) {
				DBG(zprintf("%s: eof\n", __func__));
				mutex_unlock(&pmig->unproc_lock);
				return (NULL);
			}
			cond_wait(&pmig->unproc_cv, &pmig->unproc_lock);
		}

		/* get a buffer from the unproc chain */
		buf = CHAIN_START(pmig->unproc_chain);
		CHAIN_UNLINK(pmig->unproc_chain, buf, link);
		mutex_unlock(&pmig->unproc_lock);

		msg_mchunk = (mig_msg_mem_block_t *)
		    (buf->msgp + MIG_MSG_HDR_LEN);

		/* compress the buffer */
		rv = mig_compress(buf->datap, buf->size_decomp,
		    msg_mchunk->chunk, &buf->size_comp);
		if (rv != 0) {
			ewarning("Migration aborted: compression "
			    "error: %d\n", rv);

			mutex_lock(&pmig->empty_lock);
			if (pmig->error == 0) {
				pmig->error = MIG_ERROR_COMPRESS;
				cond_signal(&pmig->empty_cv);
			}
			mutex_unlock(&pmig->empty_lock);
			break;
		}

		/* put the buffer on the processed chain */
		mutex_lock(&pmig->proc_lock);
		CHAIN_ADD(pmig->proc_chain, buf, link);
		cond_signal(&pmig->proc_cv);
		mutex_unlock(&pmig->proc_lock);
	}

	return (NULL);
}

static int
get_nworker_threads(void)
{
	int	nthreads;
#if ZDBG
	/*
	 * Allow the DEBUG-only option to override the number of
	 * workers used. This includes allowing more workers than
	 * there are CPUs in the control domain.
	 */
	if (option.mig_nworkers != MIG_NWORKERS_USEDEFAULT) {
		return (MIG_CLAMP_NWORKERS(option.mig_nworkers));
	}
#endif
	if ((nthreads = sysconf(_SC_NPROCESSORS_ONLN)) == -1) {
		ewarning("Migration warning: sysconf failure");
		return (MIG_NWORKERS_MIN);
	}

	return (MIG_CLAMP_NWORKERS(nthreads));
}

static bool_t
peer_supports_compression(mig_conn_t *mconn)
{
	if (mconn->major_vers > MIG_COMPRESS_MIN_MAJOR ||
	    (mconn->major_vers == MIG_COMPRESS_MIN_MAJOR &&
	    mconn->minor_vers >= MIG_COMPRESS_MIN_MINOR)) {
		return (true);
	}

	return (false);
}

bool_t
is_mig_init_extended(mig_conn_t *mconn)
{
	if (mconn->major_vers > MIG_UUID_MIN_MAJOR ||
	    (mconn->major_vers == MIG_UUID_MIN_MAJOR &&
	    mconn->minor_vers >= MIG_UUID_MIN_MINOR)) {
		return (true);
	}

	return (false);
}

static void
mblock_set_error_resp(mig_seq_args_t *mig, int mig_error)
{
	switch (mig_error) {
	case MIG_ERROR_COMPRESS:
		rprintf(mig->resp, MSG_ERR_MIG_COMPRESSION);
		break;
	case MIG_ERROR_DECOMPRESS:
		rprintf(mig->resp, MSG_ERR_MIG_DECOMPRESSION);
		break;
	case MIG_ERROR_COMPRESS_SIZE:
		rprintf(mig->resp, MSG_ERR_MIG_COMPRESSION_SIZE);
		break;
	case MIG_ERROR_CONNECTION:
		rprintf(mig->resp, MSG_ERR_MIG_CONNECTION_LOST,
		    mig->peer_hostname);
		break;
	case MIG_ERROR_CANCEL:
		rprintf(mig->resp, MSG_ERR_MIG_CANCEL_COMPLETED,
		    mig->ldp->namep);
		break;
	default:
		rprintf(mig->resp, MSG_ERR_MIG_UNEXPECTED_ERROR,
		    mig->ldp->namep, mig_error);
		break;
	}
}

/*
 * Used only on the source/sender side, this function takes empty buffers
 * from the empty chain and fills them with memory data read from the
 * hypervisor. If compression is enabled, the buffers are then placed
 * on the unprocessed chain. If compression is not enabled, the buffers
 * are placed on the processed chain.
 */
static int
migrate_one_mblock(mig_seq_args_t *mig, pmig_ctrl_t *pmig,
    pmig_params_t *params, mblock_t *mbp)
{
	pmig_buf_t	*buf;
	uint64_t	base;
	uint64_t	size;
	uint64_t	chunklen;
	int		amt;

	base = mbp->paddr;
	size = mbp->size;

	while (size > 0) {
		mutex_lock(&pmig->empty_lock);
		if (pmig->error != 0) {
			mblock_set_error_resp(mig, pmig->error);
			mutex_unlock(&pmig->empty_lock);
			return (MIG_STATUS_FAILURE);
		}

		/* wait for an empty buffer to fill */
		while (CHAIN_LENGTH(pmig->empty_chain) == 0) {
			cond_wait(&pmig->empty_cv, &pmig->empty_lock);
			if (pmig->error != 0) {
				mblock_set_error_resp(mig, pmig->error);
				mutex_unlock(&pmig->empty_lock);
				return (MIG_STATUS_FAILURE);
			}
		}

		/* remove the buffer from the empty chain */
		buf = CHAIN_START(pmig->empty_chain);
		CHAIN_UNLINK(pmig->empty_chain, buf, link);
		mutex_unlock(&pmig->empty_lock);

		/*
		 * Check if the user attempted to cancel the migration
		 * request (explicitly via cancel-op or via ^C) and if
		 * so terminate the memory transfer.
		 */
		if (MIG_CANCELED(mig->ldp)) {
			DBG(zprintf("\n*** Migration Canceled ***\n\n"));
			mblock_set_error_resp(mig, MIG_ERROR_CANCEL);
			return (MIG_STATUS_FAILURE);
		}

		chunklen = (size > mig_mem_chunklen) ? mig_mem_chunklen : size;
		amt = backend.read_mem(base, buf->datap, chunklen);
		if (amt != chunklen)
			DBG(zprintf("Warning: read only %d\n", amt));

		buf->size_decomp = amt;
		buf->base_ra = mbp->raddr;
		buf->offset = base - mbp->paddr;

		if (params->compression_enabled) {
			buf->size_comp = params->max_chunklen;
			/* add buffer to unprocessed chain */
			mutex_lock(&pmig->unproc_lock);
			CHAIN_ADD(pmig->unproc_chain, buf, link);
			cond_signal(&pmig->unproc_cv);
			mutex_unlock(&pmig->unproc_lock);
		} else {
			buf->size_comp = buf->size_decomp;
			/* add buffer to processed chain */
			mutex_lock(&pmig->proc_lock);
			CHAIN_ADD(pmig->proc_chain, buf, link);
			cond_signal(&pmig->proc_cv);
			mutex_unlock(&pmig->proc_lock);
		}

		size -= amt;
		base += amt;
	}

	return (MIG_STATUS_SUCCESS);
}

/*
 * Writes a single mblock EOF message to the target. The EOF message
 * does not contain any memory chunk data.
 */
static int
migrate_write_mblock_eof_msg(mig_seq_args_t *mig)
{
	mig_conn_t		*mconn;
	uint8_t			*bufp;
	ldm_hdr_t		*hdr;
	mig_msg_t		*mig_msg;
	mig_msg_mem_block_t	*msg_mchunk;

	mconn = mig->mconn;
	bufp = init_mig_msg(MIG_MEM_BLOCK, sizeof (mig_msg_mem_block_t),
	    MIG_STATUS_SUCCESS);
	hdr = (ldm_hdr_t *)bufp;
	mig_msg = (mig_msg_t *)(bufp + sizeof (ldm_hdr_t));
	msg_mchunk = (mig_msg_mem_block_t *)(bufp + MIG_MSG_HDR_LEN);

	hdr->payload_len = hton32(sizeof (mig_msg_t) +
	    sizeof (mig_msg_mem_block_t));
	mig_msg->len = hton32(sizeof (mig_msg_mem_block_t));
	memset(msg_mchunk, 0, sizeof (mig_msg_mem_block_t));
	msg_mchunk->eof = 1;

	if (mig_write_msg(mconn, bufp, hdr->payload_len +
	    sizeof (ldm_hdr_t), MIG_HANDSHAKE_TIMEOUT) == -1) {
		ewarning("Failed to send EOF mem chunk to target host");
		mblock_set_error_resp(mig, MIG_ERROR_CONNECTION);
		Xfree(bufp);
		return (MIG_STATUS_FAILURE);
	}

	Xfree(bufp);

	return (MIG_STATUS_SUCCESS);
}

static int
init_pmig_params(mig_conn_t *mconn, pmig_params_t *params)
{
	/*
	 * Does the peer support compressed mblock chunk data? Adjust
	 * parameters accordingly. If the peer does not support compression,
	 * compression/decompression worker threads will not be needed. In
	 * this scenario, we do not use worker threads, but we use more than
	 * one pmig_buf_t to allow reading/writing of memory chunks from/to
	 * the hypervisor in one thread while the writer thread performs the
	 * network writes in parallel.
	 */
	params->compression_enabled = peer_supports_compression(mconn);
#if ZDBG
	if (option.warm_migration_compression == false)
		params->compression_enabled = false;
#endif
	if (params->compression_enabled) {
		params->nworkers = get_nworker_threads();
		if (mig_max_compressed_size(mig_mem_chunklen,
		    &params->max_chunklen)) {
			ewarning("Migration aborted: could not determine "
			    "max compressed message size");
			return (MIG_ERROR_COMPRESS_SIZE);
		}
	} else {
		params->nworkers = 0; /* not used when compression disabled */
		params->max_chunklen = mig_mem_chunklen;
	}
	params->nbuffers = MIG_NBUFFERS(params->nworkers);
	params->max_block_msg_size = sizeof (mig_msg_mem_block_t) +
	    params->max_chunklen;

	return (0);
}

static void
init_pmig_buffers(pmig_params_t *params, pmig_buf_t *buffers)
{
	int i;

	for (i = 0; i < params->nbuffers; i++) {
		buffers[i].msgp = init_mig_msg(MIG_MEM_BLOCK,
		    params->max_block_msg_size, MIG_STATUS_SUCCESS);
		if (params->compression_enabled) {
			buffers[i].datap = Xmalloc(mig_mem_chunklen);
		} else {
			buffers[i].datap = ((mig_msg_mem_block_t *)
			    (buffers[i].msgp + MIG_MSG_HDR_LEN))->chunk;
		}
	}
}

/*
 * The main driver for the memory migration on the source/sender side.
 * Starts the progress, writer, and compressor threads and then calls
 * migrate_one_mblock for each mblock in the domain. Depending on the
 * negotiated migration protocol version, decides whether or not to use
 * compression. i.e., compression is not used when communicating with
 * older versions of ldmd, maintaining backwards compatibility.
 */
static int
migrate_mblocks(mig_seq_args_t *mig)
{
	mig_conn_t	*mconn;
	ldom_t		*ldp;
	mblock_t	*mbp;
	pmig_ctrl_t	pmig = {0};
	pmig_params_t	params;
	pmig_buf_t	*buffers;		/* array of buffers */
	int		nworkers_started = 0;	/* number of workers started */
	thread_t	*worker_thread_ids;	/* worker thread IDs */
	thread_t	write_thread_id;	/* writer thread ID */
	thread_t	progress_thread_id;	/* progress thread ID */
	int		mig_result;
	int		rv;
	int		i;

	ASSERT(mig != NULL);

	mig_result = MIG_STATUS_SUCCESS;
	mconn = mig->mconn;
	ldp = mig->ldp;

	pmig.mconn = mconn;
	pmig.ldp = ldp;

	/* initialize parameters dependent on peer version and CPU capacity */
	if ((rv = init_pmig_params(mconn, &params)) != 0) {
		mblock_set_error_resp(mig, rv);
		return (MIG_STATUS_FAILURE);
	}

	DBG(zprintf("Allocating %d buffer(s)\n", params.nbuffers));
	buffers = Xcalloc(params.nbuffers, pmig_buf_t);
	for (i = 0; i < params.nbuffers; i++)
		CHAIN_ADD(pmig.empty_chain, &buffers[i], link);

	/* initialize buffers */
	init_pmig_buffers(&params, buffers);

	/* start the progress thread */
	if (thr_create(NULL, NULL, migrate_progress_thread, &pmig,
	    0, &progress_thread_id) != 0) {
		ewarning("Migration aborted: could not create progress thread");
		mig_result = MIG_STATUS_FAILURE;
		goto cleanup_buffers;
	}

	/* start the write thread */
	if (thr_create(NULL, NULL, src_mblock_write_thread, &pmig,
	    0, &write_thread_id) != 0) {
		ewarning("Migration aborted: could not create writer thread");
		mig_result = MIG_STATUS_FAILURE;
		goto cleanup_progress;
	}

	if (params.compression_enabled) {
		DBG(zprintf("Starting %d comp thread(s)\n", params.nworkers));
		worker_thread_ids = Xcalloc(params.nworkers, thread_t);
		for (i = 0; i < params.nworkers; i++) {
			if (thr_create(NULL, NULL,
			    src_mblock_compress_thread, &pmig, 0,
			    &worker_thread_ids[i]) == 0) {
				nworkers_started++;
			} else {
				ewarning("Migration aborted: could not create "
				    "compress thread");
				mig_result = MIG_STATUS_FAILURE;
				goto cleanup_comp;
			}
		}
	}

	/* call migrate_one_mblock for each mblock */
	CHAIN_ITER(ldp->bound.memory, mbp, ldom) {
		DBG8_BLOCK {
			/* for debug, save the contents of memory to a file */
			char fname[MAXPATHLEN];
			zprintf("Saving mblock @ 0x%"PRIx64" (%"PRId64" "
			    "bytes)\n", mbp->paddr, mbp->size);
			snprintf(fname, MAXPATHLEN, "/export/%s-mem-0x%"PRIx64
			    ".bin", ldp->namep, mbp->paddr);
			do_read_mem_to_file(fname, mbp->paddr, mbp->size);
		}

		mig_result = migrate_one_mblock(mig, &pmig, &params, mbp);
		if (mig_result != MIG_STATUS_SUCCESS) {
			break;
		}
	}

cleanup_comp:
	/* wait for the compression threads to finish */
	if (params.compression_enabled) {
		mutex_lock(&pmig.unproc_lock);
		pmig.workers_eof = true;
		cond_broadcast(&pmig.unproc_cv);
		mutex_unlock(&pmig.unproc_lock);
		for (i = 0; i < nworkers_started; i++) {
			(void) thr_join(worker_thread_ids[i], NULL, NULL);
		}
		Xfree(worker_thread_ids);
	}

	/* wait for the writer thread to finish */
	mutex_lock(&pmig.proc_lock);
	pmig.writer_eof = true;
	cond_signal(&pmig.proc_cv);
	mutex_unlock(&pmig.proc_lock);
	(void) thr_join(write_thread_id, NULL, NULL);

	/* check for any errors that occurred on final buffers  */
	if (mig_result == MIG_STATUS_SUCCESS && pmig.error != 0) {
		mig_result = MIG_STATUS_FAILURE;
		mblock_set_error_resp(mig, pmig.error);
	}

cleanup_progress:
	/* wait for the progress thread to finish */
	pmig.progress_eof = true;
	(void) thr_join(progress_thread_id, NULL, NULL);
	mutex_destroy(&pmig.progress_lock);

cleanup_buffers:
	/* free the message buffers, threads, locks, and cv */
	for (i = 0; i < params.nbuffers; i++) {
		Xfree(buffers[i].msgp);
		if (params.compression_enabled)
			Xfree(buffers[i].datap);
	}
	Xfree(buffers);
	mutex_destroy(&pmig.empty_lock);
	mutex_destroy(&pmig.unproc_lock);
	mutex_destroy(&pmig.proc_lock);
	cond_destroy(&pmig.empty_cv);
	cond_destroy(&pmig.unproc_cv);
	cond_destroy(&pmig.proc_cv);

	/* send the EOF message */
	if (mig_result == MIG_STATUS_SUCCESS)
		mig_result = migrate_write_mblock_eof_msg(mig);

	return (mig_result);
}

static void *
migrate_state(void *args)
{
	mig_seq_args_t	*mig = args;
	ldom_t		*ldp;

	ASSERT(mig != NULL);

	ldp = mig->ldp;

	/*
	 * Migrate the contents of guest memory.
	 */
	mig->status = migrate_mblocks(mig);

	/*
	 * If there were any failures, including an explict cancellation
	 * of the migration, cleanup and report back the failure.
	 */
	if (mig->status == MIG_STATUS_FAILURE) {
		seq_enqueue_work(migration_complete, mig, true);

		/* clear the cancel flag if necessary */
		if (MIG_CANCELED(ldp))
			ldp->mig_state.cancel = false;

		return (NULL);
	}

	/*
	 * XXXMIG - Unfortunately, all the state calls must be done
	 * inside the sequencer. This appears to be necessary because
	 * the hvctl operations must be run through there to ensure
	 * that all responses are paired with the correct request.
	 * In particular, the PM thread polling for utilization info
	 * gets in the way of these hvctl operations outside of the
	 * sequencer. This needs to be resolved in a cleaner way.
	 */
	seq_enqueue_work(i_src_do_migrate_state, mig, false);

	return (NULL);
}

static int
i_src_migrate_bindings(mig_seq_args_t *args)
{
	mig_conn_t	*mconn;
	ldom_t		*ldp;
	int		rc = 0;
	mblock_t	*mblk;
	uint8_t		*bufp = NULL;
	int		size;
	md_t		*mdp;

	/* unpack args */
	mconn = args->mconn;
	ldp = args->ldp;

	DBG(zprintf("%s:\n", __func__));

	DBG_BLOCK {
		/* print out the VIO bindings */
		response_t	*rp;

		init_response(&rp);

		vio_print_bindings(ldp, rp);
		DBG(zprintf("%s", rp->bufp + RESPONSE_CONTENT_OFFSET));

		fini_response(rp);
	}

	int nmblocks = CHAIN_LENGTH(ldp->bound.memory);
	int membindings_len = nmblocks * MBLOCK_BINDING_INFO_SIZE;
	uint64_t *membindings = Xmalloc(membindings_len);
	int i = 0;

	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		membindings[i++] = mblk->raddr;
		membindings[i++] = mblk->paddr;
		membindings[i++] = mblk->size;
	}
	ASSERT(i == (nmblocks * MBLOCK_NUM_BINDING_INFO_FIELDS));
	ASSERT(membindings_len <= MIG_MAX_MSG_LEN);

	if (mig_write_buf(mconn, membindings, membindings_len,
	    MIG_TRANSFER_TIMEOUT) == -1) {
		args->status = MIG_STATUS_FAILURE;
		rc = -1;
		goto done;
	}

	/* read in the current guest MD */
	mdp = backend.get_guest_md(ldp);

	/* patch up any power managed CPUs */
	(void) cpu_mig_add_pm_bindings(ldp, mdp);

	/* augment the MD with extra VIO binding info */
	(void) vio_mig_add_bindings(ldp, mdp);

	/* save the MD for use in recovering from errors */
	if (ldp->mig_state.bind_mdp != NULL)
		md_fini(ldp->mig_state.bind_mdp);
	ldp->mig_state.bind_mdp = mdp;

	/* generate the binary MD */
	size = md_gen_bin(mdp, &bufp);
	if (size <= 0)
		fatal("Unable to generate binary guest MD\n");

	DBG(dump_to_file(bufp, size, "dump-mig-bindings-src.md"));

	/* ship the binary MD to the target */
	if (mig_write_buf(mconn, bufp, size, MIG_TRANSFER_TIMEOUT) == -1) {
		args->status = MIG_STATUS_FAILURE;
		rc = -1;
	}

	/* Free memory allocated by md_gen_bin() */
	Xfree(bufp);

done:
	Xfree(membindings);
	return (rc);
}

/*
 * Implements mig_src_init()
 */
static bool_t
mig_src_warm_init(mig_seq_args_t *args)
{
	int		nbound_vcpus;
	int		nbound_maus;
	ldom_t		*ldp = args->ldp;
	response_t	*resp = args->resp;

	DBG2(zprintf("%s\n", __func__));

	/*
	 * Check for any conditions on the source machine that
	 * may block a migration (delayed reconfig, etc)
	 */
	if (mig_src_local_feasibility_check(args) == false)
		return (false);

	if (mig_src_bound_feasibility_check(args) == false)
		return (false);

	if (!pm_migration_is_possible()) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, MSG_ERR_MIG_PM);
		return (false);
	}

	/*
	 * If the source domain has more than one vcpu, make sure
	 * that the CPU domain service is available so that it can
	 * be reduced to only one vcpu before the migration.
	 */
	nbound_vcpus = ldom_num_bound_vcpus(ldp);
	if ((nbound_vcpus > 1) && (check_cpu_dr_capable(ldp) != 0)) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, gettext("Migrating a domain with more than "
		    "one vcpu requires CPU\nDR, which is not available for "
		    "LDom %s.\n"), ldp->namep);
		return (false);
	}

	/*
	 * If the source domain has any crypto units bound, crypto
	 * DR is required to remove them all before the migration.
	 */
	nbound_maus = ldom_num_bound_maus(ldp);
	if ((nbound_maus > 0) && (check_crypto_dr_capable(ldp) != 0)) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(resp, gettext("Unable to migrate LDom %s while "
		    "there are bound crypto units present.\n"), ldp->namep);
		return (false);
	}

	return (true);
}

/*
 * Implements mig_src_send_config()
 */
static bool_t
mig_src_warm_send_config(mig_seq_args_t *args)
{
	mig_conn_t	*mconn = args->mconn;
	ldom_t		*ldp = args->ldp;

	DBG(zprintf("%s\n", __func__));

	/*
	 * Transfer any static configuration information so that it can
	 * be used to create the domain and for the feasibility checks.
	 */
	if (migrate_config(mconn, ldp) != 0) {
		args->status = MIG_STATUS_FAILURE;
		DBG(zprintf("unable to send MIG config information\n"));
		return (false);
	}

	return (true);
}

static bool_t
mig_src_warm_send_bindings(mig_seq_args_t *args)
{
	DBG(zprintf("%s\n", __func__));

	/*
	 * Transfer the binding information relevant to the guest.
	 *
	 * MIGXXX - Does it make sense to go ahead and do the bind on
	 * the target now if everything looks ok?
	 */
	i_src_migrate_bindings(args);

	return (true);
}

/*
 * Implements mig_src_feasibility_check()
 */
static bool_t
mig_src_warm_feasibility_check(mig_seq_args_t *args)
{
	bool_t	can_migrate;

	DBG(zprintf("%s\n", __func__));

	can_migrate = i_src_check_migration_feasibility(args);

	/* return if we just wanted to check if the migration would succeed */
	if (IS_MIG_DRY_RUN(args)) {
		if (can_migrate == true) {
			/*
			 * This is a dry-run (and thus not a long running
			 * command) so we need to send a response back to
			 * ldm saying it succeeded.  We set the send_resp
			 * to true so that when we return we will do this.
			 */
			args->client->send_resp = true;
			args->ldp->flag.migration_source = false;
			return (true);
		}

		/* report back that the migration would fail */
		ASSERT(args->status == MIG_STATUS_FAILURE);
		rprintf(args->resp, MSG_ERR_MIG_DRY_RUN_FAILURE,
		    args->tgt_ldom);
		return (false);
	}

	return (can_migrate);
}

/*
 * Implements mig_src_bind_check()
 */
static bool_t
mig_src_warm_bind_check(mig_seq_args_t *args)
{
	ASSERT(args != NULL);
	DBG2(zprintf("%s\n", __func__));

	if (IS_MIG_DRY_RUN(args))
		return (true);

	/*
	 * All the bindings have already been sent to the target. The target
	 * will perform the bind. All that has to be done here is wait for
	 * the results of the operation.
	 */
	return (recv_mig_msg_resp(args, MIG_CHECK_BIND));
}

/*
 * Implements mig_src_send_state()
 */
static bool_t
mig_src_warm_send_state(mig_seq_args_t *args)
{
	DBG(zprintf("%s\n", __func__));

	if (args->dry_run)
		return (true);

	if (suspend_domain(args->ldp, args->resp) != 0) {
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	LOCK_LDOM(args->ldp);

	/*
	 * Create a thread to transfer the runtime state. This
	 * process can take a significant amount of time, so it
	 * should not be performed inside the sequencer. That
	 * way, operations that do not involve the migrating
	 * domain can proceed as normal.
	 */
	if (thr_create(NULL, NULL, migrate_state, args, THR_DAEMON, NULL) != 0)
		efatal("Could not create thread to process a migration");

	return (true);
}

/*
 * Target ops vector
 */
static bool_t mig_tgt_warm_init(mig_seq_args_t *args);
static bool_t mig_tgt_warm_recv_config(mig_seq_args_t *args);
static bool_t mig_tgt_warm_recv_bindings(mig_seq_args_t *args);
static bool_t mig_tgt_warm_feasibility_check(mig_seq_args_t *args);
static bool_t mig_tgt_warm_bind(mig_seq_args_t *args);
static bool_t mig_tgt_warm_recv_state(mig_seq_args_t *args);

mig_tgt_ops_t mig_tgt_warm_ops = {
	mig_tgt_warm_init,			/* mig_tgt_init */
	mig_tgt_warm_recv_config,		/* mig_tgt_recv_config */
	mig_tgt_warm_recv_bindings,		/* mig_tgt_recv_bindings */
	mig_tgt_warm_feasibility_check,		/* mig_tgt_feasibility_check */
	mig_tgt_warm_bind,			/* mig_tgt_bind */
	mig_tgt_warm_recv_state			/* mig_tgt_recv_state */
};

extern bool_t cpu_mig_bind(ldom_t *ldp, md_t *mdp, response_t *resp);
extern bool_t cpu_mig_add_remaining(ldom_t *ldp, int num_req, response_t *resp);
extern bool_t check_full_core_allocation(int num);
extern int call_hv_guest_resume(ldom_t *ldp);
extern int call_hv_state_restore(ldom_t *, uint64_t, state_restype_t, uint32_t);

/*
 * Implements mig_tgt_init()
 */
static bool_t
mig_tgt_warm_init(mig_seq_args_t *args)
{
	ASSERT(args != NULL);

	if (!pm_migration_is_possible()) {
		args->status = MIG_STATUS_FAILURE;
		rprintf(args->resp, MSG_ERR_MIG_PM);
		return (false);
	}

	/*
	 * Check for any conditions on the target machine that
	 * may block a migration (delayed reconfig, etc)
	 */
	return (mig_tgt_local_feasibility_check(args));
}

/*
 * Implements mig_tgt_recv_config()
 */
static bool_t
mig_tgt_warm_recv_config(mig_seq_args_t *args)
{
	args->ldp = receive_config(args);
	if (args->ldp == NULL) {
		DBG(zprintf("%s: Unable to create target LDom\n", __func__));
		args->status = MIG_STATUS_FAILURE;
		if (send_mig_msg_resp(args, MIG_ERROR_RESP, args->status) != 0)
			warning("Failed to send migration error message to ldmd"
			    " on source machine %s\n", args->peer_hostname);
		return (false);
	}

	if (args->status != MIG_STATUS_SUCCESS) {
		/*
		 * Domain creation failed halfway
		 */
		DBG(zprintf("%s: Error during guest creation, cleaning up...\n",
		    __func__));
		return (false);
	}

	return (true);
}

static void
mig_clone_a_binding(mem_blocks_t *clones, uint64_t *desc)
{
	mblock_t *clone;

	clone = Xmalloc(sizeof (mblock_t));
	clone->raddr = desc[0];
	clone->paddr = desc[1];
	clone->size = desc[2];
	clone->id = 0;

	CHAIN_ADD(clones->blocks, clone, blk);
}

/*
 * Implements mig_tgt_recv_bindings()
 */
static bool_t
mig_tgt_warm_recv_bindings(mig_seq_args_t *args)
{
	int		nblocks;
	mem_blocks_t	*mem;
	mig_msg_t	*mig_msg;
	uint64_t	*desc;
	ldom_t		*ldp = args->ldp;
	mig_conn_t	*mconn = args->mconn;

	DBG(zprintf("%s\n", __func__));

	args->status = MIG_STATUS_FAILURE;

	mem = Xmalloc(sizeof (mem_blocks_t));
	args->mem = mem;

	/*
	 * Read in the memory bindings.
	 */
	if (mig_read_mig_msg(mconn, NULL, MIG_MAX_MSG_LEN,
	    MIG_TRANSFER_TIMEOUT) == -1) {
		ASSERT(args->peer_hostname != NULL);
		ewarning("Timeout reading memory bindings from source: %s\n",
		    args->peer_hostname);
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	mig_msg = (mig_msg_t *)mconn->read.buf;
	desc = (uint64_t *)mig_msg->data;
	nblocks = mig_msg->len / (3 * (int)sizeof (uint64_t));

	CHAIN_INIT(mem->blocks);

	for (int i = 0; i < nblocks; i++)
		mig_clone_a_binding(mem, desc + (3 * i));

	/*
	 * Read in the guest MD.
	 */
	if (mig_read_mig_msg(mconn, NULL, MIG_MAX_MSG_LEN,
	    MIG_TRANSFER_TIMEOUT) == -1) {
		ASSERT(args->peer_hostname != NULL);
		ewarning("Timeout reading guest MD from source: %s\n",
		    args->peer_hostname);
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	mig_msg = (mig_msg_t *)mconn->read.buf;

	DBG(dump_to_file(mig_msg->data, mig_msg->len, "dump-mig-bindings.md"));

	/* convert the binary MD into a data structure */
	ldp->mig_state.bind_mdp = md_scan_bin((uint8_t *)mig_msg->data,
	    mig_msg->len);
	if (ldp->mig_state.bind_mdp == NULL) {
		rprintf(args->resp, MSG_ERR_MIG_MD_PARSE_FAIL);
		DBG(zprintf("%s: unable to parse binding MD\n", __func__));
		args->status = MIG_STATUS_FAILURE;
		return (false);
	}

	args->status = MIG_STATUS_SUCCESS;
	return (true);
}

static bool_t
i_tgt_check_config_supported(mig_seq_args_t *wargs)
{
	mig_msg_mach_config_t	*config;
	response_t	*resp = wargs->resp;
	mig_conn_t	*mconn = wargs->mconn;
	ldom_t		*ldp = wargs->ldp;
	md_t		*mdp = ldp->mig_state.bind_mdp;
	mig_msg_t	*mig_msg;
	uint32_t	mig_msg_type = MIG_CHECK_CONFIG;
	bool_t		is_supported = true;
	int		rv;
	int		ncpus;

	DBG2(zprintf("%s:\n", __func__));

	wargs->status = MIG_STATUS_SUCCESS;

	rv = mig_read_mig_msg(mconn, NULL, MIG_MAX_MSG_LEN,
	    MIG_TRANSFER_TIMEOUT);
	if (rv == -1) {
		ewarning("Failed to receive Config migration feasibility"
		    " info from the source LDoms Manager");
		goto failure;
	}

	mig_msg = (mig_msg_t *)mconn->read.buf;
	if (mig_msg == NULL || mig_msg->type != MIG_CHECK_CONFIG) {
		ewarning("Received bogus Config migration feasibility"
		    " info from the source LDoms Manager");
		goto failure;
	}

	DBG5(zprintf("%s: read MIG_CHECK_CONFIG req\n", __func__));
	DUMP_MIG_MSG(mig_msg, "IN");
	ASSERT(mig_msg->data != NULL);

	/* Check that the CPU types of both machines are identical */
	config = (mig_msg_mach_config_t *)mig_msg->data;
	if (pri.cpu_compatible == NULL || config->cpu_compatible == NULL) {
		DBG(zprintf("%s: CPU type invalid\n", __func__));
		goto failure;
	}

	if (strcmp(config->cpu_compatible, pri.cpu_compatible) != 0) {
		DBG(zprintf("%s: CPU type mismatch: source %s, target %s\n",
		    __func__, config->cpu_compatible, pri.cpu_compatible));
		rprintf(resp, MSG_ERR_MIG_CPU_TYPE_MISMATCH);
		goto failure;
	}

	/* Check that the CPU frequencies of both machines are identical */
	if (config->stick_freq != pri.stick_frequency) {
		DBG(zprintf("%s: Stick freq mismatch: source %" PRId64
		    " target %" PRId64 "\n", __func__, config->stick_freq,
		    pri.stick_frequency));
		rprintf(resp, MSG_ERR_MIG_CPU_SPEED_MISMATCH);
		goto failure;
	}

	/* Verify that no migration is in progress on the machine */
	if (is_migration_in_progress(ldp) == true) {
		DBG(zprintf("%s: migration occurring on machine\n", __func__));
		rprintf(resp, MSG_ERR_MIG_IN_PROGRESS);
		goto failure;
	}

	/*
	 * Check that full cores are available.
	 *
	 * XXXMIG - Need to break these checks into two categories.
	 * One batch of checks that will always be valid, and one
	 * batch that can go away at some point. May make it easier
	 * in the future to just skip the second batch if both the
	 * source and target support it.
	 */
	ncpus = md_scan_dag(mdp, NULL, md_find_name(mdp, "cpu"),
	    md_find_name(mdp, "fwd"), NULL);

	if (ncpus == 0) {
		warning("Failed to find any cpus in MD\n");
		rprintf(resp, MSG_ERR_MIG_CPU_CORE_UNAVAIL);
		goto failure;
	}

	if (check_full_core_allocation(ncpus) == false) {
		DBG(zprintf(MSG_ERR_MIG_CPU_CORE_UNAVAIL));
		rprintf(resp, MSG_ERR_MIG_CPU_CORE_UNAVAIL);
		goto failure;
	}

	goto done;

failure:
	wargs->status = MIG_STATUS_FAILURE;
	is_supported = false;
	mig_msg_type = MIG_ERROR_RESP;
done:
	/* Let source know that feasibility check has completed */
	rv = send_mig_msg_resp(wargs, mig_msg_type, wargs->status);
	if (rv != 0)
		is_supported = false;

	DBG(zprintf("%s: config check %s\n",
	    __func__, is_supported ? "succeeded" : "failed"));

	return (is_supported);
}

static bool_t
i_tgt_check_api_groups_supported(mig_seq_args_t *wargs)
{
	mig_msg_api_group_vers_t	*api_groups;
	hv_api_group_version_t		*entries;
	mig_conn_t	*mconn = wargs->mconn;
	mig_msg_t	*mig_msg;
	uint32_t	mig_msg_type;
	uint32_t	cc;
	int		rv;

	DBG2(zprintf("%s:\n", __func__));

	rv = mig_read_mig_msg(wargs->mconn, NULL, MIG_MAX_MSG_LEN,
	    MIG_TRANSFER_TIMEOUT);
	if (rv == -1) {
		ewarning("Failed to receive API Group migration feasibility"
		    " info from the source LDoms Manager");
		goto failure;
	}

	DBG5(zprintf("%s: read MIG_CHECK_API req\n", __func__));

	mig_msg = (mig_msg_t *)mconn->read.buf;
	if (mig_msg == NULL || mig_msg->type != MIG_CHECK_API) {
		ewarning("Received bogus API Group migration feasibility"
		    " info from the source LDoms Manager");
		goto failure;
	}
	DUMP_MIG_MSG(mig_msg, "IN");
	ASSERT(mig_msg->data != NULL);

	api_groups = (mig_msg_api_group_vers_t *)mig_msg->data;
	DBG5(DUMP_API_GROUPS(api_groups->group_size, api_groups->entries));
	entries = api_groups->entries;
	cc = call_hv_query_api_groups(0, api_groups->group_size, &entries);
	if (cc == HVctl_st_ok) {
		wargs->status = MIG_STATUS_SUCCESS;
		mig_msg_type = MIG_CHECK_API;
		goto done;
	} else {
		goto failure;
	}

failure:
	wargs->status = MIG_STATUS_FAILURE;
	mig_msg_type = MIG_ERROR_RESP;

done:
	/* Let source know if feasibility check succeeded */
	rv = send_mig_msg_resp(wargs, mig_msg_type, wargs->status);
	if (rv != 0)
		return (false);

	DBG(zprintf("%s: API Group check %s\n", __func__,
	    wargs->status == MIG_STATUS_SUCCESS ? "succeeded" : "failed"));

	return (wargs->status == MIG_STATUS_SUCCESS ? true : false);
}

static void
i_tgt_do_feasibility_check(void *args)
{
	mig_seq_args_t		*wargs = args;
	bool_t			is_supported;
	ldom_t			*ldp = wargs->ldp;

	DBG(zprintf("%s:\n", __func__));

	ASSERT_IN_SEQUENCER();
	ASSERT(wargs != NULL);
	ASSERT(wargs->ldp != NULL);

	is_supported = mig_tgt_bound_feasibility_check(wargs);
	if (is_supported == false)
		goto done;

	is_supported = i_tgt_check_config_supported(wargs);
	if (is_supported == false) {
		DBG(zprintf("%s: config not supported, aborting\n", __func__));
		goto done;
	}

	is_supported = i_tgt_check_api_groups_supported(wargs);
	if (is_supported == false) {
		DBG(zprintf("%s: API groups not supported\n", __func__));
		goto done;
	}

done:
	/*
	 * Completed the feasibility checks. If this was a dry run,
	 * there is nothing left to do. Cleanup the created LDom.
	 *
	 * XXXMIG - In reality, the domain should never have been
	 * created if this was a dry run. Need to patch up the XML
	 * parse to not do any actual work.
	 */
	if (IS_MIG_DRY_RUN(wargs)) {
		UNLOCK_LDOM(ldp);
		if (destroy_one_ldom(ldp, wargs->resp, NULL) == true) {
			wargs->ldp = NULL;
		} else {
			wargs->status = MIG_STATUS_FAILURE;
			DBG(zprintf("unable to destroy target LDom for "
			    "dry run\n"));
		}
	}
}

/*
 * Implements mig_tgt_feasibility_check()
 */
static bool_t
mig_tgt_warm_feasibility_check(mig_seq_args_t *args)
{
	seq_enqueue_work(i_tgt_do_feasibility_check, args, true);

	return (args->status == MIG_STATUS_SUCCESS ? true : false);
}

static void
do_migration_bind(void *args)
{
	mig_seq_args_t	*wargs = args;
	response_t	*resp = wargs->resp;
	ldom_t		*ldp;
	md_t		*mdp;

	ASSERT_IN_SEQUENCER();
	ASSERT(wargs != NULL);
	ASSERT(wargs->ldp != NULL);
	ASSERT(wargs->status == MIG_STATUS_SUCCESS);

	/* we set this back to SUCCESS at the end if we can bind everything */
	wargs->status = MIG_STATUS_FAILURE;

	ldp = wargs->ldp;

	/*
	 * We need to unlock the LDom in case multiple VIO devices are
	 * specified (adding the second and subsequent devices would
	 * be blocked otherwise).  It is safe to do this since we are
	 * on the sequencer and are protected.
	 */
	ASSERT(IS_LDOM_LOCKED(ldp) == true);
	UNLOCK_LDOM(ldp);

	mdp = ldp->mig_state.bind_mdp;
	ldp->flag.migration_target = true;

	ldp->idp = alloc_res(&zeus.free_ldom_ids);
	if (ldp->idp == NULL) {
		rprintf(resp, gettext("No more active LDoms allowed\n"));
		goto done;
	}

	ldp->res_id = alloc_res(&zeus.free_guest_ids);
	if (ldp->res_id == NULL) {
		rprintf(resp, gettext("No more LDom slots available\n"));
		goto done;
	}

	DBG(zprintf("Binding LDom %s, gid %" PRId64 ", res_id %" PRId64 "\n",
	    ldp->namep, ldp->idp->id, ldp->res_id->id));

	/*
	 * Reserve physical CPU resources for the final number of
	 * strands in the target domain. Only the lowest numbered
	 * vcpu in the MD will be bound. The rest have to be added
	 * after the domain is resumed via CPU DR.
	 */
	if (cpu_mig_bind(ldp, mdp, resp) == false) {
		rprintf(resp, gettext("Unable to bind required CPUs\n"));
		goto done;
	}

	/*
	 * Cryptos, if available for the cpus bound to the domain in target
	 * machine would be DR'd in if necessary in i_tgt_receive_state()
	 */

	if (mem_bind_real(ldp, wargs->mem, resp) == false)
		goto done;

	if (bind_boot_options(ldp, resp) == false)
		goto done;

	if (vio_mig_bind(ldp, mdp, resp) != 0)
		goto done;

	DBG_BLOCK {
		/* print out the VIO bindings */
		response_t	*rp;

		init_response(&rp);

		vio_print_bindings(ldp, rp);
		DBG(zprintf("%s", rp->bufp + RESPONSE_CONTENT_OFFSET));

		fini_response(rp);
	}

	ldp->flag.local_tick_sync = true;
	ldp->flag.gen_new_md = true;
	ldp->state = BINDING;
	ldp->mig_state.suspended = true;

	if (wargs->uuid != NULL && is_uuid_valid(wargs->uuid)) {
		/* migrated with a valid UUID */
		memcpy(ldp->props.uuid, wargs->uuid,
		    UUID_PRINTABLE_STRING_LENGTH);
	} else {
		/* downlevel migration; assign UUID now */
		new_ldom_uuid(ldp);
	}

	/*
	 * Perform the bind now, but do not send the MD update
	 * notification to any of the affected domains. This
	 * must be done after the target has been resumed so that
	 * the affected domains become aware of its presence.
	 */
	CHAIN_ADD(zeus.inprogress, ldp, inprog);
	seq_bind_new();
	cleanse_progress_chain();

	/*
	 * XXXMIG - Now that the domain is bound, we can remove the
	 * hack LDom variable being used to tell zeus to destroy the
	 * domain during startup.
	 */
	database_ldom_del_variable(ldp, "SUNW,mig-target");
	database_save();

	/* success */
	wargs->status = MIG_STATUS_SUCCESS;

done:
	/*
	 * lock the LDom again now that the bind is complete - it will
	 * be unlocked for once and for all when the migration is complete.
	 */
	LOCK_LDOM(ldp);
	if (wargs->status == MIG_STATUS_SUCCESS) {
		if (send_mig_msg_resp(args, MIG_CHECK_BIND, wargs->status) != 0)
			warning("Failed to send migration bind message to ldmd"
			    " on source machine %s\n", wargs->peer_hostname);
	} else {
		if (send_mig_msg_resp(args, MIG_ERROR_RESP, wargs->status) != 0)
			warning("Failed to send migration error message to ldmd"
			    " on source machine %s\n", wargs->peer_hostname);
		DBG(zprintf("cleaning up partial binding\n"));
		undo_partial_binding(ldp);
	}
}

/*
 * Implements mig_tgt_bind()
 */
static bool_t
mig_tgt_warm_bind(mig_seq_args_t *args)
{
	if (IS_MIG_DRY_RUN(args))
		return (true);

	seq_enqueue_work(do_migration_bind, args, true);
	if (args->status != MIG_STATUS_SUCCESS)
		return (false);

	return (true);
}

static uint64_t
base_ra2base_pa(ldom_t *ldp, uint64_t base_ra)
{
	mblock_t *mbp;

	CHAIN_ITER(ldp->bound.memory, mbp, ldom) {
		if (base_ra == mbp->raddr)
			return (mbp->paddr);
	}

	return (-1);
}

/*
 * Used only on the target/receiver, this thread is analogous to
 * src_mblock_write_thread. It transitions buffers from the processed
 * chain to the empty chain as it writes mblock chunk data into memory.
 */
static void *
tgt_mblock_write_thread(void *args)
{
	pmig_ctrl_t	*pmig;
	pmig_buf_t	*buf;
	uint64_t	curr_ra = 0;
	uint64_t	base_pa = 0;

	pmig = (pmig_ctrl_t *)args;

	while (1) {
		mutex_lock(&pmig->proc_lock);
		while (CHAIN_LENGTH(pmig->proc_chain) == 0) {
			if (pmig->writer_eof) {
				DBG(zprintf("%s: eof\n", __func__));
				mutex_unlock(&pmig->proc_lock);
				return (NULL);
			}
			(void) cond_wait(&pmig->proc_cv, &pmig->proc_lock);
		}

		buf = CHAIN_START(pmig->proc_chain);
		CHAIN_UNLINK(pmig->proc_chain, buf, link);
		mutex_unlock(&pmig->proc_lock);

		/* calc the pa for the write_mem call */
		if (curr_ra != buf->base_ra) {
			if ((base_pa = base_ra2base_pa(pmig->ldp,
			    buf->base_ra)) == -1) {
				DBG(zprintf("The RA 0x%"PRIx64" does "
				    "not map to a PA bound to ldom %s\n",
				    buf->base_ra, pmig->ldp->namep));

				mutex_lock(&pmig->empty_lock);
				if (pmig->error == 0) {
					pmig->error = MIG_ERROR_OTHER;
					cond_signal(&pmig->empty_cv);
				}
				mutex_unlock(&pmig->empty_lock);
				return (NULL);
			}
			curr_ra = buf->base_ra;
		}

		/* write this memory chunk to the guest's memory */
		backend.write_mem(base_pa + buf->offset,
		    buf->datap, buf->size_decomp);

		/* update progress */
		mutex_lock(&pmig->progress_lock);
		pmig->progress_percent = MIG_MSG_GET_PROGRESS_NTOH(buf->msgp);
		mutex_unlock(&pmig->progress_lock);

		DBG(zprintf("base=0x%"PRIx64", "
		    "offset=0x%"PRIx64", size=0x%x, "
		    "compress=%.2f%%, progress=%d%%\n",
		    buf->base_ra, buf->offset, buf->size_comp,
		    MIG_COMP_RATE(buf->size_comp, buf->size_decomp),
		    MIG_MSG_GET_PROGRESS_NTOH(buf->msgp)));

		mutex_lock(&pmig->empty_lock);
		CHAIN_ADD(pmig->empty_chain, buf, link);
		cond_signal(&pmig->empty_cv);
		mutex_unlock(&pmig->empty_lock);
	}
	/*NOTREACHED*/
}

/*
 * Used only on the target/receiver, this thread is analogous to
 * src_mblock_compress_thread. It decompresses buffers and
 * transitions them from the unprocessed chain to the processed chain.
 * It updates buf->decomp_size to reflect the size of decompressed
 * memory chunk.
 */
static void *
tgt_mblock_decompress_thread(void *args)
{
	pmig_ctrl_t		*pmig;
	pmig_buf_t		*buf;
	mig_msg_mem_block_t	*msg_mchunk;
	int			rv;

	pmig = (pmig_ctrl_t *)args;

	while (1) {
		mutex_lock(&pmig->unproc_lock);
		while (CHAIN_LENGTH(pmig->unproc_chain) == 0) {
			if (pmig->workers_eof) {
				DBG(zprintf("%s: eof\n", __func__));
				mutex_unlock(&pmig->unproc_lock);
				return (NULL);
			}
			cond_wait(&pmig->unproc_cv, &pmig->unproc_lock);
		}

		/* get an unprocessed buffer */
		buf = CHAIN_START(pmig->unproc_chain);
		CHAIN_UNLINK(pmig->unproc_chain, buf, link);
		mutex_unlock(&pmig->unproc_lock);

		msg_mchunk = (mig_msg_mem_block_t *)
		    (buf->msgp + MIG_MSG_HDR_LEN);

		/* decompress the buffer */
		rv = mig_decompress(msg_mchunk->chunk, buf->size_comp,
		    buf->datap, &buf->size_decomp);
		if (rv != 0) {
			ewarning("Migration aborted: decompression "
			    "error: %d\n", rv);

			mutex_lock(&pmig->empty_lock);
			if (pmig->error == 0) {
				pmig->error = MIG_ERROR_COMPRESS;
				cond_signal(&pmig->empty_cv);
			}
			mutex_unlock(&pmig->empty_lock);
			break;
		}

		/* put the buffer on the processed chain */
		mutex_lock(&pmig->proc_lock);
		CHAIN_ADD(pmig->proc_chain, buf, link);
		cond_signal(&pmig->proc_cv);
		mutex_unlock(&pmig->proc_lock);
	}

	return (NULL);
}

static int
receive_mblocks(mig_seq_args_t *args, pmig_ctrl_t *pmig, pmig_params_t *params)
{
	pmig_buf_t		*buf;
	mig_msg_t		*mig_msg;
	mig_msg_mem_block_t	*msg_mchunk;

	/*
	 * If the operation fails or is canceled after this point,
	 * any memory on the target that contains part of the source
	 * memory image must be cleared. For now, all of the target's
	 * memory is scrubbed. In the future, it may be possible to
	 * track which memory has been touched and only clear those
	 * blocks.
	 */
	args->mem_rcvd = true;

	while (1) {

		mutex_lock(&pmig->empty_lock);
		if (pmig->error) {
			mutex_unlock(&pmig->empty_lock);
			return (MIG_STATUS_FAILURE);
		}

		/* wait for an available empty buffer */
		while (CHAIN_LENGTH(pmig->empty_chain) == 0) {
			DBG3(zprintf("%s: sleeping\n", __func__));
			cond_wait(&pmig->empty_cv, &pmig->empty_lock);
			if (pmig->error) {
				mutex_unlock(&pmig->empty_lock);
				return (MIG_STATUS_FAILURE);
			}
		}

		/* remove a buffer from the empty chain */
		buf = CHAIN_START(pmig->empty_chain);
		CHAIN_UNLINK(pmig->empty_chain, buf, link);
		mutex_unlock(&pmig->empty_lock);

		mig_msg = (mig_msg_t *)(buf->msgp + sizeof (ldm_hdr_t));
		msg_mchunk = (mig_msg_mem_block_t *)(buf->msgp +
		    MIG_MSG_HDR_LEN);

		/* read the message */
		if (mig_read_mig_msg(args->mconn, mig_msg,
		    params->max_block_msg_size + sizeof (mig_msg_t),
		    MIG_TRANSFER_TIMEOUT) == -1) {
			ewarning("Timeout reading memory config from source\n");
			return (MIG_STATUS_FAILURE);
		}

		/* check if it is what we are expecting */
		if (ntoh32(mig_msg->type) != MIG_MEM_BLOCK) {
			ewarning("Invalid message type, expected %d, got "
			    "%d\n", (int)MIG_MEM_BLOCK, mig_msg->type);
			return (MIG_STATUS_FAILURE);
		}

		/* eof signals last message, does not include data */
		if (ntoh64(msg_mchunk->eof) != 0)
			break;

		buf->size_comp = ntoh64(msg_mchunk->size);
		buf->base_ra = ntoh64(msg_mchunk->base_ra);
		buf->offset = ntoh64(msg_mchunk->offset);

		if (params->compression_enabled) {
			buf->size_decomp = params->max_chunklen;
			/* add buffer to unprocessed chain */
			mutex_lock(&pmig->unproc_lock);
			CHAIN_ADD(pmig->unproc_chain, buf, link);
			cond_signal(&pmig->unproc_cv);
			mutex_unlock(&pmig->unproc_lock);
		} else {
			buf->size_decomp = buf->size_comp;
			/* add buffer to processed chain */
			mutex_lock(&pmig->proc_lock);
			CHAIN_ADD(pmig->proc_chain, buf, link);
			cond_signal(&pmig->proc_cv);
			mutex_unlock(&pmig->proc_lock);
		}
	}

	return (MIG_STATUS_SUCCESS);
}

/*
 * Used only on the target/receiver, this is the main driver of memory
 * receive. It starts the progress, writer, and decompressor threads.
 * Like on the sender side, the negotiated migration protocol version is
 * used to determine whether or not memory chunk data will need to be
 * compressed. If decompression is not needed, decompression threads will
 * not be started.
 */
static int
receive_mem(mig_seq_args_t *args)
{
	pmig_ctrl_t	pmig = {0};
	pmig_params_t	params;
	pmig_buf_t	*buffers;		/* array of buffers */
	int		nworkers_started = 0;	/* number of workers started */
	thread_t	*worker_thread_ids;	/* worker thread IDs */
	thread_t	write_thread_id;	/* writer thread ID */
	thread_t	progress_thread_id;	/* progress thread ID */
	int		mig_result;
	int		i;

	DBG(zprintf("%s...\n", __func__));

	mig_result = MIG_STATUS_SUCCESS;
	pmig.mconn = args->mconn;
	pmig.ldp = args->ldp;

	/* init parameters dependent on peer version and CPU capacity */
	if (init_pmig_params(args->mconn, &params) != 0)
		return (MIG_STATUS_FAILURE);

	DBG(zprintf("Allocating %d buffer(s)\n", params.nbuffers));
	buffers = Xcalloc(params.nbuffers, pmig_buf_t);
	for (i = 0; i < params.nbuffers; i++)
		CHAIN_ADD(pmig.empty_chain, &buffers[i], link);

	/* initialize buffers */
	init_pmig_buffers(&params, buffers);

	/* start the progress thread */
	if (thr_create(NULL, NULL, migrate_progress_thread, &pmig,
	    0, &progress_thread_id) != 0) {
		ewarning("Migration aborted: could not create progress thread");
		mig_result = MIG_STATUS_FAILURE;
		goto cleanup_buffers;
	}

	/* start the write thread */
	if (thr_create(NULL, NULL, tgt_mblock_write_thread, &pmig, 0,
	    &write_thread_id) != 0) {
		ewarning("Migration aborted: could not create writer thread");
		mig_result = MIG_STATUS_FAILURE;
		goto cleanup_progress;
	}

	/* start the decompression threads */
	if (params.compression_enabled) {
		DBG(zprintf("Starting %d decomp thread(s)\n",
		    params.nworkers));
		worker_thread_ids = Xcalloc(params.nworkers, thread_t);
		for (i = 0; i < params.nworkers; i++) {
			if (thr_create(NULL, NULL,
			    tgt_mblock_decompress_thread, &pmig, 0,
			    &worker_thread_ids[i]) == 0) {
				nworkers_started++;
			} else {
				ewarning("Migration aborted: could not create "
				    "decompression thread");
				mig_result = MIG_STATUS_FAILURE;
				goto cleanup_decomp;
			}
		}
	}

	/* process memory block messages */
	mig_result = receive_mblocks(args, &pmig, &params);

cleanup_decomp:
	/* wait for the decompression threads to finish */
	if (params.compression_enabled) {
		mutex_lock(&pmig.unproc_lock);
		pmig.workers_eof = true;
		cond_broadcast(&pmig.unproc_cv);
		mutex_unlock(&pmig.unproc_lock);
		for (i = 0; i < nworkers_started; i++) {
			(void) thr_join(worker_thread_ids[i], NULL, NULL);
		}
		Xfree(worker_thread_ids);
	}

	/* wait for the writer thread to finish */
	mutex_lock(&pmig.proc_lock);
	pmig.writer_eof = true;
	cond_signal(&pmig.proc_cv);
	mutex_unlock(&pmig.proc_lock);
	(void) thr_join(write_thread_id, NULL, NULL);

	/* check for any errors that occured on final buffers */
	if (mig_result == MIG_STATUS_SUCCESS && pmig.error)
		mig_result = MIG_STATUS_FAILURE;

cleanup_progress:
	/* wait for the progress thread to finish */
	pmig.progress_eof = true;
	(void) thr_join(progress_thread_id, NULL, NULL);
	mutex_destroy(&pmig.progress_lock);

cleanup_buffers:
	/* free the message buffers, locks, and cv */
	for (i = 0; i < params.nbuffers; i++) {
		Xfree(buffers[i].msgp);
		if (params.compression_enabled)
			Xfree(buffers[i].datap);
	}
	Xfree(buffers);
	mutex_destroy(&pmig.empty_lock);
	mutex_destroy(&pmig.unproc_lock);
	mutex_destroy(&pmig.proc_lock);
	cond_destroy(&pmig.empty_cv);
	cond_destroy(&pmig.unproc_cv);
	cond_destroy(&pmig.proc_cv);

	return (mig_result);
}

static int
receive_resource_state(ldom_t *ldp, mig_conn_t *mconn)
{
	mig_msg_t		*mig_msg;
	mig_msg_saved_state_t	*msg_saved;
	uint32_t		restype;
	uint32_t		resid;
	bound_mfrag_t		*state_frag;
	int			amt;
	uint8_t			*mdaddr;
	uint64_t		paddr;
	int			cc;
	uint64_t		mdsz;

	/* read the first message */
	if (mig_read_mig_msg(mconn, NULL, MIG_MAX_MSG_LEN,
	    MIG_TRANSFER_TIMEOUT) == -1) {
		DBG(zprintf("%s: connection terminated...\n", __func__));
		return (-1);
	}

	mig_msg = (mig_msg_t *)mconn->read.buf;
	ASSERT(mig_msg != NULL);
	ASSERT(mig_msg->data != NULL);

	if (ntoh32(mig_msg->type) != MIG_SAVED_STATE) {
		DBG(zprintf("Invalid message type, expected %d, got "
		    "%d\n", (int)MIG_SAVED_STATE, mig_msg->type));
		return (-1);
	}

	msg_saved = (mig_msg_saved_state_t *)mig_msg->data;
	restype = ntoh32(msg_saved->restype);
	resid = ntoh32(msg_saved->resid);
	mdsz = ntoh64(msg_saved->size);

	if ((restype != STATE_RES_GUEST) &&
	    (restype != STATE_RES_VCPU) &&
	    (restype != STATE_RES_LDC)) {
		DBG(zprintf("EOF for resource state migration\n"));
		/* XXXMIG - fix this up somehow >0 means EOF for now */
		return (12);
	}

	DBG(zprintf("MIG Resource State: restype=%s, resid=%d\n",
	    RESTYPE_STR(restype), resid));

	DBG_BLOCK {
		int	fd;
		char	fname[MAXPATHLEN];

		snprintf(fname, MAXPATHLEN, "/export/mig-%s%d-state.bin",
		    RESTYPE_STR(restype), resid);
		fd = open(fname, O_RDWR|O_CREAT|O_TRUNC, 0600);
		if (fd == -1) {
			DBG(zprintf("Failed to open \"%s\": %s\n", fname,
			    strerror(errno)));
			return (-1);
		}

		write(fd, msg_saved->statemd, mdsz);
		close(fd);
	}

	/*
	 * Restore the resource state to the HV.
	 */
	state_frag = frag_alloc(ldp, mdsz);
	if (state_frag == NULL) {
		DBG(zprintf("Unable to allocate frag for state MD\n"));
		return (-1);
	}

	mdaddr = msg_saved->statemd;
	paddr = state_frag->paddr;

	/* write the state MD into the HV memory */
	while (mdsz > 0) {
		amt = backend.write_mem(paddr, mdaddr, mdsz);
		if (amt != mdsz)
			DBG(zprintf("Warning: amount written was %d\n", amt));

		mdsz -= amt;
		paddr += amt;
		mdaddr += amt;
	}

	/* restore the state */
	cc = call_hv_state_restore(ldp, state_frag->paddr, restype, resid);
	if (cc != HVctl_st_ok) {
		DBG(zprintf("HVctl command failed: %d\n", cc));
		return (-1);
	}

	frag_free(state_frag);

	return (0);
}

/*
 * Once the domain services connection has been established, some
 * operations rely on certain services being available. Make an
 * effort to allow all the necessary registrations to occur.
 */
void
mig_wait_for_domain_services(ldom_t *ldp)
{
	int iter;

	/*
	 * Need to stall here to give the domain services a chance
	 * to get registered. A check for the CPU DR domain service
	 * is used because it covers all necessary functionality.
	 * If that becomes insufficient in the future, additional
	 * checks will need to be made here.
	 *
	 * XXXMIG - This is a very arbitrary timeout. Need to tune
	 * this, or figure out a better way to deterministically
	 * tell when all the necessary domain service registrations
	 * happen.
	 */
	for (iter = 0; iter < DS_UP_MAX_RETRIES; iter++) {
		if (check_cpu_dr_capable(ldp) == 0)
			break;
		sleep(1);
	}

	/* check for a timeout */
	if (iter == DS_UP_MAX_RETRIES) {
		DBG(zprintf("Timeout waiting for domain services "
		    "to come online for LDom %s\n", ldp->namep));
		return;
	}

	DBG(zprintf("waited %d second%s for domain services to come "
	    "online for LDom %s\n", iter, (iter == 1) ? "" : "s", ldp->namep));
}

static void
i_tgt_receive_state(void *args)
{
	mig_seq_args_t	*wargs = args;
	int		rv = 0;
	ldom_t		*ldp;
	mig_conn_t	*mconn;
	md_t		*mdp;
	int		cc;
	int		num_req;
	ldm_err_t	pm_result = { 0 }; /* Not an array because currently */
					/* only PM_RES_CPU resource type is */
					/* affected by a domain state change. */

	/* unpack args */
	ldp = wargs->ldp;
	mconn = wargs->mconn;
	mdp = ldp->mig_state.bind_mdp;

	DBG(zprintf("%s...\n", __func__));

	/*
	 * XXXMIG - none of the network receive should be done
	 * inside the sequencer. That defeats the whole intent
	 * of having the ability to query status, or cancel a
	 * long running operation.
	 *
	 * It is only in here now due to the fact that the state
	 * restore to the HV is done inline with the network
	 * transfer of the state itself.
	 *
	 * Should probably be two pass. One pass to collect all
	 * the state, and another pass to restore it to the HV.
	 */

	/* XXXMIG - hack an EOF (rv > 0) */
	while (rv == 0)
		rv = receive_resource_state(ldp, mconn);

	/* check for errors */
	if (rv < 0) {
		wargs->status = MIG_STATUS_FAILURE;
		rv = send_mig_msg_resp(wargs, MIG_ERROR_RESP, wargs->status);
		if (rv != 0)
			warning("Failed to send migration error message to ldmd"
			    " on source machine %s\n", wargs->peer_hostname);
		return;
	}

	DBG(zprintf("*** All state received ***\n"));

	/* check that all the expected state was received */

	/* send the final ok message */
	if (send_mig_msg_resp(wargs, MIG_HANDOFF, wargs->status) != 0) {
		DBG(zprintf("Final handoff failed\n"));
		/* XXXMIG - put in error state, or teardown? */
		return;
	}

	/* notify PM that the domain is starting */
	if (pm_domain_events_on) {
		send_pm_dom_event(ldp, START_DOMAIN_PRE, &pm_result);
		/*
		 * FIXME: check for errors by checking pm_result.ldme_errs
		 * and traversing pm_result.ldme_private.
		 */
		if (pm_result.ldme_errs != LDME_OP_EOK) {
			rprintf(wargs->resp, "Power Management failure when "
			    "resuming LDom %s\n", ldp->namep);
			wargs->status = MIG_STATUS_FAILURE;
			pm_free_errors(&pm_result);
			return;
		}
		pm_free_errors(&pm_result);
	}

	/* resume the guest in its new home */
	DBG(zprintf("resuming target LDom %s\n", ldp->namep));

	cc = call_hv_guest_resume(ldp);
	if (cc != HVctl_st_ok) {
		DBG(zprintf("HVctl to resume target failed\n"));
		/* XXXMIG - put in error state, or teardown? */
		return;
	}

	ldp->state = ACTIVE;

	ldp->mig_state.suspended = false;

	/*
	 * Now that the domain is active, it is the master copy of
	 * the domain. Clear the flag in the HV MD that indicates
	 * that it is in transition. If there is a crash from this
	 * point onward, the domain will not be destroyed on startup.
	 */
	clear_mig_flag_in_hvmd(ldp);

	/*
	 * Create the domain services channel so that when async
	 * registration requests come in, the DS channel will be
	 * ready. These requests can start immediately after the
	 * control domain receives its MD update notification, so
	 * this has to be done before guest reconfig. Otherwise,
	 * registration requests will be dropped.
	 */
	(void) ds_register_domain(ldp, false);

	/*
	 * Perform the MD update notification. If this is done
	 * before the target is resumed, none of the peers will
	 * be aware of its arrival.
	 */
	seq_send_md_update_notifs();

	/* open the DS channel for use */
	(void) ds_open_channel(ldp->ds_chan);

	/* wait for domain services to come online */
	mig_wait_for_domain_services(ldp);

	/* reconfigure peer vnet channels */
	ldp->flag.md_notify = true;
	seq_send_md_update_notifs();

	/* lookup how many CPUs were in the source domain */
	num_req = md_scan_dag(mdp, NULL, md_find_name(mdp, "cpu"),
	    md_find_name(mdp, "fwd"), NULL);

	/*
	 * Attempt to grow the CPU count in the domain back to
	 * the original number that was present in the source.
	 *
	 * XXXMIG - Some error checking? Should at least warn
	 * the user if something goes wrong.
	 */
	if (cpu_mig_add_remaining(ldp, num_req, wargs->resp) == false)
		warning("Failed to grow CPU count in LDom %s\n", ldp->namep);

	/*
	 * Add crypto units back in
	 * lookup how many MAUs were in the source domain
	 */
	num_req = grab_exec_units(mdp, "mau", NULL);

	if (mau_mig_add(ldp, num_req, wargs->resp) == false)
		warning("Failed to grow MAU count in LDom %s\n", ldp->namep);

	/* ensure that the autosave is up to date */
	autosave_update_autosave();

	UNLOCK_LDOM(ldp);
	ldp->flag.migration_target = false;
}

/*
 * Implements mig_tgt_recv_state()
 */
static bool_t
mig_tgt_warm_recv_state(mig_seq_args_t *args)
{
	if (IS_MIG_DRY_RUN(args))
		return (true);

	args->status = receive_mem(args);
	if (args->status != MIG_STATUS_SUCCESS)
		return (false);

	/*
	 * Queue up the state transfer in the sequencer.
	 */
	seq_enqueue_work(i_tgt_receive_state, args, true);
	if (args->status != MIG_STATUS_SUCCESS)
		return (false);

	return (true);
}
