/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_MIG_CLASS_H
#define	_MIG_CLASS_H

#ifdef __cplusplus
extern "C" {
#endif

#include <netdb.h>


/*
 * This describes the order of the message sent between the source and target
 * machines in order to carry out a migration. At any stage, if an error
 * occurs, a MIG_ERROR_RESP message can be returned which the source will
 * handle appropriately and the migration will terminate.
 *
 *	SOURCE					   TARGET
 *	------					   ------
 *	MIG_AUTH_COOKIE_VERIFY ->
 *						<- MIG_AUTH_COOKIE_VERIFY
 *	MIG_INIT_INFO ->
 *						<- MIG_INIT_INFO
 *	MIG_CHECK_INIT ->
 *						<- MIG_CHECK_INIT
 *	MIG_BUFFER (config/constraints) ->
 *	MIG_BUFFER (membindings) ->
 *	MIG_BUFFER (binary MD) ->
 *
 *	MIG_CHECK_CONFIG [*] ->
 *						<- MIG_CHECK_CONFIG
 *	MIG_CHECK_API [*] ->
 *						<- MIG_CHECK_API [*]
 *	MIG_CHECK_BIND ->
 *						<- MIG_CHECK_BIND
 *	MIG_MEM_BLOCK [*] ->
 *		.
 *		.
 *	MIG_MEM_BLOCK [*] ->
 *
 *	MIG_SAVED_STATE [*] ->
 *						<- MIG_HANDOFF
 *
 *
 *	[*] warm migration only
 */

/*
 * Types of migration message.
 */
enum {
	MIG_AUTH_COOKIE_VERIFY,	/* authentication exchange */
	MIG_INIT_INFO,		/* initial negotiation msg starting migration */
	MIG_CHECK_INIT,		/* Initial local check on tgt */
	MIG_CHECK_API,		/* Check for API group versions supported */
	MIG_CHECK_CONFIG,	/* Check CPU type/speed */
	MIG_CHECK_BIND,		/* Indicate bind successful on tgt */
	MIG_BUFFER,		/* send buffer contents to other side */
	MIG_MEM_BLOCK,		/* send contents of a memory block */
	MIG_SAVED_STATE,	/* send guest/vcpu/ldc state info */
	MIG_HANDOFF,		/* sent by tgt -> src when taking control */
	MIG_ERROR_RESP		/* general error message sent by tgt => src */
};

#define	MIG_MAGIC	0x6d696772	/* "migr" */

/*
 * Types of migration message status values
 */
enum {
	MIG_STATUS_SUCCESS,
	MIG_STATUS_FAILURE,
};

typedef struct mig_msg {
	uint32_t	type;
	uint32_t	len;
	uint32_t	magic;
	uint32_t	status;
	uint32_t	progress;	/* % of migration completed [0, 100] */
	char		data[];
} mig_msg_t;

/*
 * mig_msg payload definitions (mig_msg->data)
 */

/* MIG_CHECK_API */
typedef struct mig_msg_api_group_vers {
	uint32_t		group_size;	/* Number of API groups */
	hv_api_group_version_t	entries[];
} mig_msg_api_group_vers_t;

/* MIG_CHECK_CONFIG */
typedef struct mig_msg_mach_config {
	uint64_t	stick_freq;
	char		cpu_compatible[];
} mig_msg_mach_config_t;

/* MIG_MEM_BLOCK */
typedef struct mig_msg_mem_block {
	uint64_t	base_ra;
	uint64_t	offset;
	uint64_t	size;
	uint64_t	eof;	/* XXXMIG - too much padding (is it needed)? */
	uint8_t		chunk[];
} mig_msg_mem_block_t;

/* MIG_SAVED_STATE */
typedef struct mig_msg_saved_state {
	uint32_t	restype;
	uint32_t	resid;
	uint64_t	size;
	uint8_t		statemd[];
} mig_msg_saved_state_t;

/* MIG_INIT_INFO_V1_1 */
typedef struct mig_msg_init_info {
	uint32_t	dryrun;
	uint32_t	final_state;	/* ldom_state_t */
	uint8_t		src_hostname[MAXHOSTNAMELEN];
	uint8_t		tgt_name[];
} mig_msg_init_info_v1_1_t;

/* MIG_INIT_INFO */
typedef struct mig_msg_init_info_ext {
	uint32_t	dryrun;
	uint32_t	final_state;	/* ldom_state_t */
	uint8_t		src_hostname[MAXHOSTNAMELEN];
	uint8_t		tgt_name[MAXLEN_VALID_NAME];
	uint8_t		uuid[UUID_PRINTABLE_STRING_LENGTH];
} mig_msg_init_info_t;

#ifdef __cplusplus
}
#endif

#endif	/* _MIG_CLASS_H */
