/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <thread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <locale.h>

#include "allocate.h"
#include "basics.h"
#include "mig_common.h"
#include "zprintf.h"

/*
 * This file contains migration routines shared between ldm and ldmd
 */

/*
 * Parse the migration target operand.
 *
 * [<user>@]host[:<new ldom>]
 *
 * Returns 0 for success, -1 if memory allocation fails, or -2 for
 * invalid syntax. Modifies 'arg'.
 */
static int
parse_target(char *arg, mig_arg_t *args)
{
	char *s;

	args->user = NULL;
	args->tgt_ldom = NULL;

	if ((s = strchr(arg, '@')) != NULL) {
		if (s == arg)
			return (-2);
		*s = '\0';
		if ((args->user = strdup(arg)) == NULL)
			return (-1);
		arg = s + 1;
	}

	if (*arg == '\0') {
		Xfree(args->user);
		return (-2);
	}

	if ((s = strchr(arg, ':')) != NULL) {
		if (*(s + 1) == '\0')
			return (-2);
		*s = '\0';
		if ((args->tgt_ldom = strdup(s + 1)) == NULL) {
			Xfree(args->user);
			return (-1);
		}
	}

	if ((args->host = strdup(arg)) == NULL) {
		Xfree(args->user);
		Xfree(args->tgt_ldom);
		return (-1);
	}

	return (0);
}

/*
 * Parse the arguments for the migrate-domain CLI.
 *
 *     [-f | --force]
 *     [-n | --dry-run]
 *     [(-p | --password-file) <password_file>]
 *     <source ldom>
 *     [<user>@]<target host>[:<target ldom>]
 *
 * Returns 0 for success, -1 if memory allocation fails, or -2 for
 * invalid syntax.
 */
int
migrate_parse_options(int argc, char **argv, mig_arg_t *args)
{
	char		opt;
	char		*tgt_str;
	int		rc;
	const char	*optstring = "f(force)n(dry-run)p:(password-file)";

	if (argc < 2)
		return (-2);

	ASSERT(args != NULL);

	args->passfile = NULL;
	args->src_ldom = NULL;

	optind = 0;	/* start with the first opt */
	opterr = 0;	/* disable error messages */

	while ((opt = getopt(argc, argv, optstring)) != -1) {
		switch (opt) {
		case 'f':
			args->force = true;
			break;

		case 'n':
			args->dry_run = true;
			break;

		case 'p':
			if ((args->passfile = strdup(optarg)) == NULL)
				return (-1);
			break;

		case '?':
			Zfree(args->passfile);
			return (-2);
		}
	}

	/* there should be 2 operands */
	if ((argc - optind) != 2) {
		Zfree(args->passfile);
		return (-2);
	}

	if ((args->src_ldom = strdup(argv[optind])) == NULL) {
		Zfree(args->passfile);
		return (-1);
	}

	if ((tgt_str = strdup(argv[optind + 1])) == NULL) {
		Zfree(args->passfile);
		Xfree(args->src_ldom);
		return (-1);
	}

	rc = parse_target(tgt_str, args);
	free(tgt_str);
	if (rc != 0) {
		Zfree(args->passfile);
		Xfree(args->src_ldom);
		return (rc);
	}

	return (0);
}

int
set_sock_nonblocking(int sock)
{
	int flags;

	if ((flags = fcntl(sock, F_GETFL, 0)) < 0)
		return (-1);
	if (!(flags & O_NONBLOCK)) {
		flags |= O_NONBLOCK;
		if (fcntl(sock, F_SETFL, flags) == -1)
			return (-1);
	}

	return (0);
}

static int
set_sock_blocking(int sock)
{
	int flags;

	if ((flags = fcntl(sock, F_GETFL, 0)) < 0)
		return (-1);
	if (flags & O_NONBLOCK) {
		flags &= ~O_NONBLOCK;
		if (fcntl(sock, F_SETFL, flags) == -1)
			return (-1);
	}

	return (0);
}


static int
timed_connect(int s, const struct sockaddr *name, int namelen, int timeout)
{
	struct pollfd pfd;
	int rc;

	if (timeout <= 0)
		return (connect(s, name, namelen));

	set_sock_nonblocking(s);

	if (connect(s, name, namelen) == 0) {
		/* immediate success */
		set_sock_blocking(s);
		return (0);
	}

	if (errno != EINPROGRESS) {
		/* something bad is wrong */
		set_sock_blocking(s);
		return (-1);
	}

	set_sock_blocking(s);

	pfd.fd = s;
	pfd.events = POLLOUT;

	do {
		rc = poll(&pfd, 1, timeout);
	} while ((rc == -1) && (errno == EINTR));

	if ((rc == 1) && (pfd.revents & POLLOUT))
		return (0);
	else {
		DBG(zprintf("%s: rc=%d, pfd.revents=%d\n", __func__,
		    rc, pfd.revents));
		return (-1);
	}
}

int
connect_to_host_port(char *host, int port, int timeout)
{
	struct addrinfo *addrs = NULL;
	struct addrinfo hints, *addr;
	char sport[32];
	int sock = -1;

	memset(&hints, 0, sizeof (hints));
	hints.ai_family = PF_INET;
	hints.ai_socktype = SOCK_STREAM;

	snprintf(sport, sizeof (sport), "%d", port);

	if ((getaddrinfo(host, sport, &hints, &addrs)) != 0) {
		(void) fprintf(stderr, gettext("Failed to get address info for "
		    "host \"%s\": %s\n"), host, strerror(errno));
		return (-1);
	}

	for (addr = addrs; addr != NULL; addr = addr->ai_next) {
		if ((addr->ai_family != AF_INET) &&
		    (addr->ai_family != AF_INET6))
			continue;

		sock = socket(addr->ai_family, addr->ai_socktype,
		    addr->ai_protocol);
		if (sock < 0) {
			(void) fprintf(stderr,
			    gettext("Failed to create socket: %s\n"),
			    strerror(errno));
			break;
		}

		if (timed_connect(sock, addr->ai_addr,
		    addr->ai_addrlen, timeout) < 0) {
			DBG(zprintf("Failure to connect errno: %s\n",
			    strerror(errno)));
			close(sock);
			sock = -1;
		} else {
			break;
		}
	}

	freeaddrinfo(addrs);
	return (sock);
}
