/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <fcntl.h>
#include <port.h>
#include <thread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <zlib.h>

#include "allocate.h"
#include "backend.h"
#include "cli_conn.h"
#include "ds_impl.h"
#include "fatal.h"
#include "ldomconfigure.h"
#include "ldm_protocol.h"
#include "memory.h"
#include "mig.h"
#include "mig_common.h"
#include "mig_class.h"
#include "option.h"
#include "sequencer.h"
#include "ssl_connection.h"
#include "vcpu.h"
#include "vmau.h"
#include "vio.h"
#include "zprintf.h"

/*
 * Supported Versions Table
 *
 * Version 1.0 was the first version of the migration protocol, supporting
 * warm and cold migration. The version 1.0 implementation does not accept
 * any migration messages with a payload of more than 5MB. The memory
 * block (MIG_MEM_BLOCK) message chunk size used is 4MB.
 *
 * Version 1.1 introduced zlib compression of MIG_MEM_BLOCK chunk data
 * for warm migration. A peer that attempts to negotiate protocol version
 * 1.1 is required to support zlib compression of the MIG_MEM_BLOCK chunk
 * data. When peers negotiate version 1.1, MIG_MEM_BLOCK chunk data will
 * be sent zlib compressed. Version 1.1 only accepts memory block messages
 * with a compressed chunk length of less than or equal to the result of
 * zlib::deflateBound(4MB). For all other migration messages, those with a
 * payload greater than 5MB are not accepted.
 *
 * Version 1.2 extends mig_msg_init_info_t with UUID. When a domain
 * migrates the UUID must go with it.  The UUID will migrate starting
 * with version 1.2, but will fail when migrating to downlevel versions.
 * A -f flag will be provided to force a migration with subsequent
 * loss of the UUID in the case of a downlevel migration. An upward
 * migration will trigger a new UUID to be assigned when the domain
 * is created.
 */
static prot_version_t mig_protocol_versions[] = {
	{ 1, 2 },
	{ 1, 1 },
	{ 1, 0 }
};

static int migport = -1;	/* event port for migration IO events */

/*
 * Compression level; slower levels typically provide better compression.
 */
static int mig_compression_level = Z_BEST_SPEED;

/*
 * Connection callback to write a message to the client.
 */
static int
mig_timed_write(mig_conn_t *mconn, void *buf, size_t len, int timeout)
{
#if ZDBG
	if (!option.secure_migration)
		return (timed_write(mconn, buf, len, timeout));
#endif
	return (timed_SSL_write(mconn, buf, len, timeout));
}

int
mig_write_msg(mig_conn_t *mconn, void *buf, size_t len, int timeout)
{
	return (mig_timed_write(mconn, buf, len, timeout));
}

/* ARGSUSED0 */
static void
mig_protocol_error(mig_conn_t *mconn, uint32_t errcode)
{
	warning("migration protocol error (%d)", errcode);
}

static int
mig_timed_read(mig_conn_t *mconn, void *buf, size_t size, int timeout)
{
#if ZDBG
	if (!option.secure_migration)
		return (timed_read(mconn, buf, size, timeout));
#endif

	return (timed_SSL_read(mconn, buf, size, timeout));
}

/*
 * Read a message. If the message payload length is not between 'min_size'
 * and 'max_size', an error will be returned and the message payload will
 * not be copied. If the 'dest' pointer argument is non-NULL, the message
 * payload will be copied to 'dest'. Otherwise, the message payload will
 * be copied to the read buffer in the mig_conn_t structure in which case
 * the buffer is freed and reallocated to match the payload size.
 */
static int
read_msg(mig_conn_t *mconn, void *dest, int min_size, size_t max_size,
    int timeout)
{
	time_t	start = time(NULL);
	time_t	tout = timeout;
	int rc;
	void *bufp;

	rc = mig_timed_read(mconn, &mconn->hdr, sizeof (ldm_hdr_t), timeout);
	if (rc == -1)
		return (-1);

	/* all of header has been read */
	mconn->hdr.msg_type = ntoh32(mconn->hdr.msg_type);
	mconn->hdr.payload_len = ntoh32(mconn->hdr.payload_len);

	if ((mconn->hdr.payload_len > max_size) ||
	    (mconn->hdr.payload_len < min_size)) {
		DBG(zprintf("%s: invalid payload length (%d)\n", __func__,
		    mconn->hdr.payload_len));
		mig_protocol_error(mconn, LDM_ERROR_INVALID);
		return (-1);
	}

	tout -= time(NULL) - start;
	if (tout <= 0) {
		errno = ETIME;
		return (-1);
	}

	if (dest == NULL) {
		Zfree(mconn->read.buf);
		bufp = mconn->read.buf = Xmalloc(mconn->hdr.payload_len);
	} else {
		bufp = dest;
	}

	/* now read in the message body */
	rc = mig_timed_read(mconn, bufp, mconn->hdr.payload_len, (int)tout);
	if (rc == -1) {
		ewarning("timed_read failed");
		if (dest == NULL)
			Zfree(mconn->read.buf);
		return (-1);
	}
	return (0);
}

/*
 * Read a message.
 */
int
mig_read_msg(mig_conn_t *mconn, int timeout)
{
	return (read_msg(mconn, NULL, 1, MIG_MAX_MSG_LEN, timeout));
}

/*
 * Read a LDM_MIG_MSG message. If the 'dest' argument is non-NULL, the
 * message is copied to 'dest'. Otherwise, it is copied to the read
 * buffer in the mig_conn_t structure. If the message is larger than
 * 'max_size', the message is not copied and an error is returned.
 */
int
mig_read_mig_msg(mig_conn_t *mconn, void *dest, size_t max_size, int timeout)
{
	mig_msg_t *mig_msg;
	int rc;

	if ((rc = read_msg(mconn, dest, sizeof (mig_msg_t),
	    max_size, timeout)) == -1)
		return (-1);

	if (mconn->hdr.msg_type != LDM_MIG_MSG) {
		DBG(zprintf("%s: invalid message type (%d)\n", __func__,
		    mconn->hdr.msg_type));
		mig_protocol_error(mconn, LDM_ERROR_INVALID);
		goto fail;
	}

	if (dest == NULL)
		mig_msg = (mig_msg_t *)mconn->read.buf;
	else
		mig_msg = (mig_msg_t *)dest;

	if (ntoh32(mig_msg->magic) != MIG_MAGIC) {
		DBG(zprintf("%s: invalid mig magic (0x%x)\n", __func__,
		    ntoh32(mig_msg->magic)));
		mig_protocol_error(mconn, LDM_ERROR_INVALID);
		goto fail;
	}

	mig_msg->type = ntoh32(mig_msg->type);
	mig_msg->len = ntoh32(mig_msg->len);

	return (rc);
fail:
	if (dest == NULL)
		Zfree(mconn->read.buf);

	return (-1);
}

/* ARGSUSED3 */
int
mig_write_buf(mig_conn_t *mconn, void *buf, size_t len, int timeout)
{
	ldm_hdr_t	*hdr;
	mig_msg_t	*mig_msg;
	char		*msg;
	int		msg_len;
	int		rc = 0;

	msg_len = MIG_MSG_HDR_LEN + len;
	msg = Xmalloc(msg_len);

	hdr = (ldm_hdr_t *)msg;
	hdr->msg_type = hton32(LDM_MIG_MSG);
	hdr->payload_len = hton32(sizeof (mig_msg_t) + len);

	mig_msg = (mig_msg_t *)(msg + sizeof (ldm_hdr_t));
	mig_msg->type = hton32(MIG_BUFFER);
	mig_msg->len = hton32(len);
	mig_msg->magic = hton32(MIG_MAGIC);

	memcpy(mig_msg->data, buf, len);

	if (mig_write_msg(mconn, msg, msg_len, MIG_HANDSHAKE_TIMEOUT) == -1) {
		ewarning("Failed to send cookie verification response to "
		    "logical domain manager");
		rc = -1;
	}

	Xfree(msg);

	return (rc);
}

void
mig_close_conn(mig_conn_t *mconn)
{
	/* clean up cached IP addresses */
	Zfree(zeus.local_ip);
	Zfree(zeus.remote_ip);

	if (mconn == NULL)
		return;
	if (mconn->ssl != NULL)
		SSL_free_f(mconn->ssl);
	close(mconn->fd);
	Zfree(mconn->read.buf);
	Zfree(mconn->write.buf);
	mutex_destroy(&mconn->lock);
	cond_destroy(&mconn->cond);
	Xfree(mconn);
}

/*
 * Check that the major and minor number are supported.
 */
bool_t
mig_check_version(uint16_t major, uint16_t minor)
{
	return (supported_version(NUM_VERS(mig_protocol_versions),
	    mig_protocol_versions, major, minor));
}

static int
mig_handle_init_req(mig_conn_t *mconn)
{
	ldm_hdr_t *hdr;
	ldm_init_req_t *init_req;
	ldm_init_ack_t *init_ack;
	ldm_init_nack_t *init_nack;
	uint32_t class;
	uint16_t req_major, req_minor;
	uint16_t major;
	uint16_t minor = 0;
	int len, rc;
	char *buf;

	if ((mconn->hdr.msg_type != LDM_INIT_REQ) ||
	    (mconn->hdr.payload_len != sizeof (ldm_init_req_t))) {
		DBG(zprintf("%s: invalid init req message\n", __func__));
		mig_protocol_error(mconn, LDM_ERROR_INVALID);
		return (-1);
	}

	init_req = (ldm_init_req_t *)mconn->read.buf;
	class = ntoh32(init_req->client_class);
	req_major = ntoh16(init_req->major_vers);
	req_minor = ntoh16(init_req->minor_vers);

	DBG(zprintf("%s: Attempting to negotiate ver (%d.%d)\n",
	    __func__, req_major, req_minor));

	Zfree(mconn->read.buf);

	if (class != LDM_CLASS_MIG) {
		DBG(zprintf("%s: invalid class\n", __func__));
		mig_protocol_error(mconn, LDM_ERROR_INVALID);
		return (-1);
	}

	if (negotiate_version(NUM_VERS(mig_protocol_versions),
	    mig_protocol_versions, req_major, &major, &minor)) {
		mconn->state = MIG_CONN_UP;
		len = sizeof (ldm_hdr_t) + sizeof (ldm_init_ack_t);
		buf = Xmalloc(len);
		hdr = (ldm_hdr_t *)buf;
		hdr->msg_type = hton32(LDM_INIT_ACK);
		hdr->payload_len = hton32(sizeof (ldm_init_ack_t));
		init_ack = (ldm_init_ack_t *)(buf + sizeof (ldm_hdr_t));
		if (minor < req_minor) {
			init_ack->minor_vers = hton16(minor);
			mconn->minor_vers = minor;
		} else {
			init_ack->minor_vers = hton16(req_minor);
			mconn->minor_vers = req_minor;
		}
		mconn->major_vers = major;

		DBG(zprintf("%s: Responding (ACK) with ver (%d.%d)\n",
		    __func__, major, init_ack->minor_vers));
	} else {
		len = sizeof (ldm_hdr_t) + sizeof (ldm_init_nack_t);
		buf = Xmalloc(len);
		hdr = (ldm_hdr_t *)buf;
		hdr->msg_type = hton32(LDM_INIT_NACK);
		hdr->payload_len = hton32(sizeof (ldm_init_nack_t));
		init_nack = (ldm_init_nack_t *)(buf + sizeof (ldm_hdr_t));
		init_nack->major_vers = hton16(major);

		DBG(zprintf("%s: Responding (NACK) with ver (%d.x)\n",
		    __func__, init_nack->major_vers));
	}

	rc = mig_write_msg(mconn, buf, len, MIG_HANDSHAKE_TIMEOUT);
	Xfree(buf);

	return (rc);
}

/*
 * Create a new migration connection on the given file descriptor.
 */
mig_conn_t *
mig_new_connection(int fd)
{
	mig_conn_t *mconn;

	mconn = Xmalloc(sizeof (mig_conn_t));

	mconn->fd = fd;
	mconn->ssl = NULL;
	mconn->conntime = time(NULL);
	mconn->state = MIG_CONN_INIT;

	mutex_init(&mconn->lock, USYNC_THREAD, NULL);
	cond_init(&mconn->cond, USYNC_THREAD, NULL);

	mconn->read.cb = NULL;
	mconn->read.buf = NULL;
	mconn->read.bufp = NULL;
	mconn->read.left = 0;

	mconn->write.cb = NULL;
	mconn->write.buf = NULL;
	mconn->write.bufp = NULL;
	mconn->write.left = 0;

	return (mconn);
}

/*
 * Secure an incoming connection using SSL. If an error occurs during
 * the SSL setup, the caller should close the connection with a call
 * to mig_close_conn().
 */
static bool_t
mig_secure_server_conn(mig_conn_t *mconn)
{
	int	rv;

#if ZDBG
	/* nothing to do if connection is not secure */
	if (!option.secure_migration) {
		DBG(zprintf("Security for incoming migration disabled\n"));
		return (true);
	}
#endif

	mconn->ssl = SSL_new_f(client_ssl_ctx);
	if (mconn->ssl == NULL) {
		warning("Failure in SSL_new for migration");
		return (false);
	}

	rv = SSL_set_fd_f(mconn->ssl, mconn->fd);
	if (rv == 0) {
		warning("Creating new SSL failed");
		return (false);
	}

	rv = SSL_connect_f(mconn->ssl);
	if (rv != 1) {
		warning("SSL connect failed\n");
		return (false);
	}

	DBG(zprintf("Incoming migration request secured\n"));

	return (true);
}

static void *
accept_mig_connection(mig_conn_t *acc_conn)
{
	int		fd;
	int		alen = 0;
	mig_conn_t	*mconn;

	DBG(zprintf("%s: accepting new migration connection\n", __func__));

	if ((fd = accept(acc_conn->fd, NULL, &alen)) < 0) {
		ewarning("accept");
		goto done;
	}

	DBG(zprintf("%s: accepted new migration connection with fd %d\n",
	    __func__, fd));

	mconn = mig_new_connection(fd);

	if (!mig_secure_server_conn(mconn)) {
		mig_close_conn(mconn);
		goto done;
	}

	set_sock_nonblocking(mconn->fd);

	if (mig_read_msg(mconn, MIG_HANDSHAKE_TIMEOUT) == -1) {
		mig_close_conn(mconn);
		goto done;
	}

	if (mig_handle_init_req(mconn) == -1) {
		mig_close_conn(mconn);
		goto done;
	}

	/* start thread to handle migration in the target */
	if (thr_create(NULL, NULL, migration_target_thr, mconn,
	    THR_DAEMON, NULL) != 0) {
		efatal("Could not create thread to process an incoming "
		    "migration request");
	}

done:
	/* accept more connections */
	if (port_associate(migport, PORT_SOURCE_FD, acc_conn->fd, POLLIN,
	    acc_conn) < 0) {
		efatal("%s: port_associate", __func__);
	}

	return (NULL);
}

static void
mig_listen_on_port(int port)
{
	struct sockaddr_in myaddr;
	mig_conn_t	*aconn;
	int		sock;
	int		on = 1;

	/* port should be initialized */
	ASSERT(migport != -1);

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		efatal("%s: socket", __func__);

	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof (on)) < 0)
		efatal("%s: turning on REUSEADDR", __func__);

	myaddr.sin_family = AF_INET;
	myaddr.sin_port = htons((unsigned short)port);
	myaddr.sin_addr.s_addr = htonl(INADDR_ANY);

	if (bind(sock, (struct sockaddr *)&myaddr, sizeof (myaddr)) < 0)
		efatal("%s: bind", __func__);

	(void) listen(sock, 3);

	DBG(zprintf("Listening for migrations on port %d\n", port));

	/* create a connection to handle incoming migrations */
	aconn = mig_new_connection(sock);
	aconn->read.cb = accept_mig_connection;

	/* register interest for incoming events to this port */
	if (port_associate(migport, PORT_SOURCE_FD, sock, POLLIN, aconn) < 0)
		efatal("%s: port_associate", __func__);
}

/*
 * Main loop to handle IO with migration peers.
 * This is run in its own thread.
 */
/* ARGSUSED0 */
static void *
migration_loop(void *unused)
{
	mig_conn_t	*mconn;
	port_event_t	ev;

	/* port should be initialized */
	ASSERT(migport != -1);

	DBG7(memset(&ev, 0, sizeof (ev)));

	/* listen for requests on the standard port */
	mig_listen_on_port(MIG_DEFAULT_PORT);

	/*
	 * It is also necessary to listen for requests on the
	 * legacy port. This allows migrations from machines
	 * running an older version of the LDom manager. This
	 * legacy support should be removed at some point.
	 */
	mig_listen_on_port(MIG_LEGACY_PORT);

	/* loop to receive and handle IO events */
	while (1) {
		if (port_get(migport, &ev, NULL) < 0)
			efatal("%s: port_get", __func__);

		if (ev.portev_source == PORT_SOURCE_FD) {
			mconn = ev.portev_user;
			(void) mconn->read.cb(mconn);
		} else {
			warning("%s: Unexpected port event source", __func__);
		}
	}

	/*NOTREACHED*/
}

extern int resume_domain(mig_seq_args_t *args);

void
migration_init(void)
{
	ldom_t		*ldp;
	ldom_t		*next;
	mig_seq_args_t	args;

	if (migport != -1)
		fatal("Migration already initialized");

	if ((migport = port_create()) < 0)
		efatal("%s: port_create", __func__);

	/*
	 * Cleanup any migrations that were in progress when the LDom
	 * manager went away the last time.
	 */
	CHAIN_ITER_VOL(zeus.ldoms, ldp, next, link) {

		/*
		 * Check for the private variable which indicates
		 * that a migration was interrupted before the
		 * target could be bound. In that case, destroy
		 * the stale target domain.
		 */
		if (get_ldom_variable(ldp, "SUNW,mig-target") != NULL) {
			ASSERT(IS_LDOM_LOCKED(ldp) == false);
			if (!destroy_one_ldom(ldp, NULL, NULL)) {
				DBG(zprintf("%s: unable to destroy stale "
				    "migration target %s (inactive)\n",
				    __func__, ldp->namep));
			}
			continue;
		}

		/*
		 * Check to see if a migration was interrupted before
		 * the final handoff could occur. In that case, it
		 * is necessary to unbind and destroy the stale target.
		 */
		if (ldp->flag.migration_target) {
			DBG(zprintf("migration target crash recovery: "
			    "destroying target domain\n"));

			/* XXXMIG - send something to the log file? */

			ldp->state = UNBINDING;
			CHAIN_ADD(zeus.inprogress, ldp, inprog);

			PRE_SEQUENCE_ENTER();
			seq_unbind_ldoms();
			PRE_SEQUENCE_EXIT();

			cleanse_progress_chain();
			ASSERT(IS_LDOM_LOCKED(ldp) == false);

			if (!destroy_one_ldom(ldp, NULL, NULL)) {
				DBG(zprintf("%s: unable to destroy stale "
				    "migration target %s (bound)\n",
				    __func__, ldp->namep));
			}
		}

		/*
		 * Check to see if a migration was interrupted while
		 * the source was suspended. If so, need to restore the
		 * VIO state and resume the domain.
		 */
		if (ldp->flag.migration_source) {
			DBG(zprintf("migration source crash recovery: "
			    "resuming source domain\n"));

			/* XXXMIG - send something to the log file? */

			args.ldp = ldp;
			args.resp = NULL;

			PRE_SEQUENCE_ENTER();
			resume_domain(&args);
			PRE_SEQUENCE_EXIT();

			ldp->flag.migration_source = false;

			/* cleanup the bindings MD */
			ASSERT(ldp->mig_state.bind_mdp);
			md_fini(ldp->mig_state.bind_mdp);
			ldp->mig_state.bind_mdp = NULL;
		}
	}

	if (thr_create(NULL, NULL, migration_loop, NULL, THR_DAEMON, NULL) != 0)
		efatal("Could not create migration I/O thread");
}

/*
 * Verify that no domain on a machine is the source or target of a
 * migration. This excludes the ldom passed in that is pending the
 * outcome of this check.
 *
 * Return Value:
 *	true	- at least one migration is in progress
 *	false	- no migration in progress.
 */
bool_t
is_migration_in_progress(ldom_t *pend)
{
	ldom_t	*ldp;

	CHAIN_ITER(zeus.ldoms, ldp, link) {

		/* skip the ldom pending an operation */
		if (pend == ldp)
			continue;

		if ((ldp->flag.migration_source != 0) ||
		    (ldp->flag.migration_target != 0))
			return (true);
	}

	return (false);
}

/*
 * Returns a pointer to the ldom structure representing the
 * domain being migrated. Returns NULL if no domain is being
 * migrated.
 */
ldom_t *
mig_inprogress_domain(void)
{
	ldom_t *ldp;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (MIG_INPROGRESS(ldp))
			break;
	}

	return (ldp);
}

/*
 * This routine is called when a migration operation needs to
 * be cancelled. The migration thread watches the cancel flag
 * and terminates if it is set.
 */
void
flag_migration_for_cancellation(ldom_t *ldp)
{
	ldp->mig_state.cancel = true;
}

/*
 * Compress a chunk of memory.
 */
int
mig_compress(uint8_t *in, uint32_t in_size, uint8_t *out, uint32_t *out_size)
{
	z_stream stream = {0};
	int rv;

	stream.opaque = Z_NULL;
	stream.zalloc = Z_NULL;
	stream.zfree = Z_NULL;

	rv = deflateInit(&stream, mig_compression_level);
	if (rv != 0) {
		DBG(zprintf("%s: deflateInit returned %d\n", __func__, rv));
		return (-1);
	}

	stream.next_in = in;
	stream.avail_in = in_size;
	stream.next_out = out;
	stream.avail_out = *out_size;

	rv = deflate(&stream, Z_FINISH);
	if (rv != Z_STREAM_END) {
		DBG(zprintf("%s: deflate returned %d\n", __func__, rv));
		(void) deflateEnd(&stream);
		return (-1);
	}

	*out_size = stream.total_out;

	rv = deflateEnd(&stream);
	if (rv != Z_OK) {
		DBG(zprintf("%s: deflateEnd returned %d\n", __func__, rv));
		return (-1);
	}

	return (0);
}

int
mig_decompress(uint8_t *in, uint32_t in_size, uint8_t *out, uint32_t *out_size)
{
	z_stream stream = {0};
	int rv;

	stream.opaque = Z_NULL;
	stream.zalloc = Z_NULL;
	stream.zfree = Z_NULL;

	rv = inflateInit(&stream);
	if (rv != 0) {
		DBG(zprintf("%s: inflateInit returned %d\n", __func__, rv));
		return (-1);
	}

	stream.next_in = in;
	stream.avail_in = in_size;
	stream.next_out = out;
	stream.avail_out = *out_size;

	rv = inflate(&stream, Z_FINISH);
	if (rv != Z_STREAM_END) {
		DBG(zprintf("%s: inflate returned %d\n", __func__, rv));
		(void) inflateEnd(&stream);
		return (-1);
	}

	*out_size = stream.total_out;

	rv = inflateEnd(&stream);
	if (rv != Z_OK) {
		DBG(zprintf("%s: inflateEnd returned %d\n", __func__, rv));
		return (-1);
	}

	return (0);
}

int
mig_max_compressed_size(uint32_t in_size, uint32_t *out_size)
{
	z_stream stream = {0};
	int rv;

	stream.opaque = NULL;
	stream.zalloc = Z_NULL;
	stream.zfree = Z_NULL;

	rv = deflateInit(&stream, mig_compression_level);
	if (rv != 0) {
		DBG(zprintf("%s: deflateInit returned %d\n", __func__, rv));
		return (-1);
	}

	*out_size = deflateBound(&stream, in_size);

	rv = deflateEnd(&stream);
	if (rv != 0) {
		DBG(zprintf("%s: deflateEnd returned %d\n", __func__, rv));
		return (-1);
	}

	return (0);
}
