/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "allocate.h"
#include "backend.h"
#include "fatal.h"
#include "ldomconfigure.h"
#include "mig.h"
#include "mig_common.h"
#include "zprintf.h"
#include "ldom.h"
#include "ldm_protocol.h"
#include "cli_conn.h"
#include "memory.h"
#include "vcpu.h"
#include "vmau.h"
#include "mig_class.h"
#include "sequencer.h"
#include "vio.h"
#include "database_internal.h"
#include "mig_be.h"
#include "xml_event.h"
#include "xml_cntrl_common.h"
#include "uuid.h"

static bool_t
receive_init(mig_seq_args_t *args)
{
	mig_conn_t		*mconn;
	mig_msg_t		*mig_msg;
	mig_msg_init_info_t	*msg_init;

	mconn = args->mconn;

	/* wait for the init message */
	if (mig_read_mig_msg(mconn, NULL, MIG_MAX_MSG_LEN,
	    MIG_HANDSHAKE_TIMEOUT) == -1) {
		args->status = MIG_STATUS_FAILURE;
		ewarning("Failed to receive init message");
		return (false);
	}

	mig_msg = (mig_msg_t *)mconn->read.buf;
	if (mig_msg->type != MIG_INIT_INFO) {
		args->status = MIG_STATUS_FAILURE;
		warning("Received spurious migration message of type (%d),"
		    "expected MIG_INIT_INFO\n", mig_msg->type);
		return (false);
	}

	msg_init = (mig_msg_init_info_t *)mig_msg->data;

	/* initialize operation args */
	args->dry_run = ntoh32(msg_init->dryrun);
	args->final_state = ntoh32(msg_init->final_state);
	args->tgt_ldom = Xstrdup((char *)msg_init->tgt_name);
	args->peer_hostname = Xstrdup((char *)msg_init->src_hostname);
	if (is_mig_init_extended(mconn) && msg_init->uuid != NULL &&
	    is_uuid_valid((char *)msg_init->uuid))
		args->uuid = Xstrdup((char *)msg_init->uuid);

	return (true);
}

static void
init_target_mig_cache(mig_seq_args_t *args)
{
	ASSERT(args != NULL);
	ASSERT(args->ldp != NULL);
	ASSERT(args->peer_hostname != NULL);

	/* initialise progress to 0% */
	args->ldp->mig_info.progress = 0;

	args->ldp->mig_info.peer_name = Xstrdup(args->peer_hostname);

	/* the following variables are unused on the target */
	args->ldp->mig_info.total_mem = 0;
	args->ldp->mig_info.migrated_mem = 0;
}

static void
teardown_domain(void *args)
{
	mblock_t	*mbp;
	mig_seq_args_t	*wargs = args;
	ldom_t		*ldp = wargs->ldp;

	/* A domain should only be ACTIVE after a handoff succeeded */
	ASSERT(ldp != NULL);
	ASSERT(ldp->state != ACTIVE);

	UNLOCK_LDOM(ldp);

	/*
	 * If the migration failed after any of the memory image
	 * had been received, the target's memory must be cleared.
	 * This is necessary because the next guest to use that
	 * memory will assume that it is a blank slate.
	 */
	if (wargs->mem_rcvd) {
		CHAIN_ITER(ldp->bound.memory, mbp, ldom) {
			(void) scrub_mem(mbp->paddr, mbp->size);
		}
	}

	if (ldp->state == BOUND) {
		/*
		 * Unbind the domain.
		 */
		ldp->state = UNBINDING;
		CHAIN_ADD(zeus.inprogress, ldp, inprog);

		seq_unbind_ldoms();
		cleanse_progress_chain();
	}

	if (destroy_one_ldom(ldp, wargs->resp, NULL) == true)
		wargs->ldp = NULL;
	else
		DBG(zprintf("unable to destroy target LDom\n"));
}

static bool_t
verify_cookie(mig_seq_args_t *args)
{
	ldm_hdr_t	*hdr;
	mig_msg_t	*mig_msg;
	int		msg_len;
	char		*buf = NULL;
	uint32_t	status;
	mig_conn_t	*mconn = args->mconn;

	DBG(zprintf("%s: reading cookie verification request\n", __func__));

	if (mig_read_mig_msg(mconn, NULL, MIG_MAX_MSG_LEN,
	    MIG_HANDSHAKE_TIMEOUT) == -1) {
		args->status = MIG_STATUS_FAILURE;
		ewarning("Failed to receive cookie verification request "
		    "from source logical domain manager");
		return (false);
	}

	mig_msg = (mig_msg_t *)mconn->read.buf;

	if (mig_msg->type != MIG_AUTH_COOKIE_VERIFY) {
		args->status = MIG_STATUS_FAILURE;
		DBG(zprintf("invalid message type, expected %d, got %d\n",
		    (int)MIG_AUTH_COOKIE_VERIFY, mig_msg->type));
		return (false);
	}

	if (((time(NULL) - zeus.cookie_time) <= MIG_TRANSFER_TIMEOUT) &&
	    (strcmp(zeus.current_cookie, mig_msg->data) == 0)) {
		DBG(zprintf("MIGRATION COOKIE AUTH SUCCESS\n"));
		status = hton32(MIG_STATUS_SUCCESS);
	} else {
		args->status = MIG_STATUS_FAILURE;
		DBG(zprintf("MIGRATION COOKIE AUTH FAILURE\n"));
		status = hton32(MIG_STATUS_FAILURE);
		return (false);
	}

	msg_len = sizeof (ldm_hdr_t) + sizeof (mig_msg_t);
	buf = Xmalloc(msg_len);

	hdr = (ldm_hdr_t *)buf;
	hdr->msg_type = hton32(LDM_MIG_MSG);
	hdr->payload_len = hton32(sizeof (mig_msg_t));

	mig_msg = (mig_msg_t *)(buf + sizeof (ldm_hdr_t));
	mig_msg->type = MIG_AUTH_COOKIE_VERIFY;
	mig_msg->len = hton32(0);
	mig_msg->magic = hton32(MIG_MAGIC);
	mig_msg->status = status;

	if (mig_write_msg(mconn, buf, msg_len, MIG_HANDSHAKE_TIMEOUT) == -1) {
		args->status = MIG_STATUS_FAILURE;
		ewarning("Failed to send cookie verification response to "
		    "source logical domain manager");
		Xfree(buf);
		return (false);
	}

	Xfree(buf);

	if (status == hton32(MIG_STATUS_FAILURE))
		return (false);

	DBG(zprintf("handshake complete...\n"));

	return (true);
}

/*
 * Lookup and cache the IP address of the peer initiating the migration.
 * This information is used in filtering out MAC address collisions from
 * the source machine.
 */
static bool_t
get_remote_ip_addr(mig_seq_args_t *args)
{
	char			*ip;
	struct sockaddr_in	*sin;
	struct sockaddr		sock_info;
	socklen_t		sock_info_len = sizeof (struct sockaddr);
	mig_conn_t		*mconn = args->mconn;

	/* lookup the peer initiating the migration */
	if (getpeername(mconn->fd, &sock_info, &sock_info_len) != 0) {
		DBG(zprintf("Failed the getpeername() call\n"));
		return (false);
	}

	if (sock_info_len == 0) {
		ewarning("Failed to get IP address of peer ldmd");
		return (false);
	}

	/* convert the IP address into a string */
	sin = (struct sockaddr_in *)&sock_info;
	ip = Xmalloc(INET_ADDRSTRLEN);
	if (inet_ntop(AF_INET, &sin->sin_addr, ip, INET_ADDRSTRLEN) == NULL) {
		DBG(zprintf("%s: inet_ntop() failed: %s\n", __func__,
		    strerror(errno)));
		Xfree(ip);
		return (false);
	}

	/* save the IP for MAC collision checking */
	zeus.remote_ip = ip;

	DBG(zprintf("Source IP address: %s\n", zeus.remote_ip));

	return (true);
}

static mig_tgt_ops_t *
mig_tgt_ops_init(mig_seq_args_t *args)
{
	/* check for cold migration */
	if (IS_MIG_COLD(args))
		return (&mig_tgt_cold_ops);

	return (&mig_tgt_warm_ops);
}

void *
migration_target_thr(void *arg)
{
	mig_seq_args_t	*args;
	mig_tgt_ops_t	*ops;

	/* initialize args structure */
	args = Xcalloc(1, mig_seq_args_t);
	init_response(&args->resp);
	args->status = MIG_STATUS_SUCCESS;
	args->mconn = arg;

	/* receive and verify the session cookie */
	if (verify_cookie(args) == false)
		goto done;

	if (receive_init(args) == false)
		goto done;

	ops = mig_tgt_ops_init(args);
	ASSERT(ops != NULL);

	DBG(zprintf("Migration started: target LDom: %s, final state: "
	    "%s%s\n", args->tgt_ldom, ldom_state_str(args->final_state),
	    (args->dry_run) ? " (dryrun)" : ""));

	if (get_remote_ip_addr(args) == false)
		goto done;

	/* perform preliminary checks */
	if (ops->mig_tgt_init(args) == false)
		goto done;

	/* receive the configuration information */
	if (ops->mig_tgt_recv_config(args) == false)
		goto done;

	/* fill in UUID info */
	if (args->uuid != NULL && is_uuid_valid(args->uuid))
		memcpy(args->ldp->props.uuid, args->uuid,
		    UUID_PRINTABLE_STRING_LENGTH);

	/* LDom is created, fill in some of the cache info */
	init_target_mig_cache(args);

	/* receive the binding information */
	if (ops->mig_tgt_recv_bindings(args) == false)
		goto done;

	/* perform feasibility checking */
	if (ops->mig_tgt_feasibility_check(args) == false)
		goto done;

	/* bind the new domain */
	if (ops->mig_tgt_bind(args) == false)
		goto done;

	/* receive and restore all domain state */
	if (ops->mig_tgt_recv_state(args) == false)
		goto done;

done:
	if (args->status == MIG_STATUS_SUCCESS) {
		if (!IS_MIG_DRY_RUN(args)) {
			/* Send a domain event. */
			event_send_dom_event(XML_MIGRATE_DOMAIN, args->ldp,
			    NULL);
			if (!IS_MIG_COLD(args)) {
				event_send_progress_event(MIGRATION_PROGRESS,
				    args->ldp, XML_COMPLETE);
			}
			/* Backup UUID to the database */
			if (args->ldp->props.uuid != NULL &&
			    is_uuid_valid(args->ldp->props.uuid)) {
				database_ldom_set_uuid(args->ldp);
				database_save();
			}
		}
	} else {
		/* something went wrong, send error string back */
		if (send_mig_msg_resp(args, MIG_ERROR_RESP, args->status) != 0)
			warning("Failed to send migration error message to ldmd"
			    " on source machine\n");

		/* teardown the LDom if we got as far as creating it */
		if (args->ldp != NULL) {
			if ((!IS_MIG_DRY_RUN(args)) && (!IS_MIG_COLD(args))) {
				event_send_progress_event(MIGRATION_PROGRESS,
				    args->ldp, (xmlChar *)XML_FAILURE);
			}
			seq_enqueue_work(teardown_domain, args, true);
		}
	}

	mig_close_conn(args->mconn);
	fini_response(args->resp);

	free_mig_seq_args(args);

	return (NULL);
}
