/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/param.h>
#include <sys/wait.h>

#include "util.h"
#include "ldmp2v.h"
#include "ldm.h"
#include "disk.h"

#define	FMTHARD_BIN	"/usr/sbin/fmthard"
#define	FMTHARD_CMD	"fmthard"
#define	MKFILE_BIN	"/usr/sbin/mkfile"
#define	MKFILE_CMD	"mkfile"
#define	ZFS_BIN		"/usr/sbin/zfs"
#define	ZFS_CMD		"zfs"
#define	VDISK_DEVPATH	"../../devices/virtual-devices@"

static void setup_vdisks(const config_t *, config_t *);
static int label_vdisk(config_t *, int);
static int fmthard(config_t *, int, const char *);
static int setup_disk_translations(config_t *);
static int setup_fs_translations(config_t *);
static void remove_unused_slices(config_t *);
static int resize_partitions(config_t *);
static void recalc_vtoc(config_t *);
static int create_vdsdev(config_t *, const char *, int);
static int create_zvol(const char *, long, int);
static int create_file(const char *, long, int);
static slice_info_t *get_slice_by_disk_id(config_t *, int, int);
static swap_info_t *get_swap_by_id(config_t *, int);
static ufs_info_t *get_ufs_by_mntpt(config_t *, const char *);
static slice_info_t *get_slice_by_special(config_t *, const char  *);
static disk_info_t *get_disk_by_id(config_t *, int);
static int have_disk(disk_info_t *, int, const char *);
static char *get_disk_dev_name(const char *);

/*
 * Create the virtual disks for the guest domain. The vdisks are initially
 * attached to the control domain so data can be restored on them.
 *
 *   vdisk names: <domain>-disk<n>
 *   vol names:   <domain>-vol<n>
 */
int
create_vdisks(const config_t *src, config_t *tgt, int rm_unused_slices)
{
	int	i;

	msg(0, "Creating vdisks ...\n");

	setup_vdisks(src, tgt);

	if (rm_unused_slices == B_TRUE)
		remove_unused_slices(tgt);

	if (tgt->virt.resize_list != NULL)
		if (resize_partitions(tgt) == -1) {
			warn("Resize failed.\n");
			return (-1);
		}

	recalc_vtoc(tgt);

	for (i = 0; i < tgt->disk_cnt; i++) {
		char	vdisk[32];
		char	vol[128];

		(void) snprintf(vdisk, sizeof (vdisk), "%s-disk%d",
		    tgt->virt.name, i);
		tgt->disks[i].vdisk = Strdup(vdisk);
		(void) snprintf(vol, sizeof (vol), "%s-vol%d@%s",
		    tgt->virt.name, i, tgt->virt.vds);
		msg(1, "Creating vdisk %s ...\n", vdisk);
		if (create_vdsdev(tgt, vol, i) == -1)
			die("Cannot create %s.\n", vol);
		if (ldm_add_vdisk("primary", vdisk, vol) != 0)
			die("Cannot add vdisk %s.\n", vdisk);
	}
	if (setup_disk_translations(tgt) == -1)
		die("Cannot setup disk name translations.\n");
	if (setup_fs_translations(tgt) == -1)
		die("Cannot setup file system name translations.\n");

	for (i = 0; i < tgt->disk_cnt; i++) {
		if (label_vdisk(tgt, i) != 0) {
			warn("Cannot create a VTOC on vdisk disk%d.\n", i);
			return (-1);
		}
	}
	return (0);
}

/*
 * Get information on the user created vdisks backing the file systems mounted
 * at virt.root.  Fill in the vdisk name, the disk name in the primary domain
 * and the disk name in the guest domain.
 */
int
get_vdisks(config_t *tgt)
{
	binding_t *bnd;
	int	disk_cnt;
	int	i;

	msg(0, "Determining vdisks ...\n");

	if ((bnd = ldm_get_bindings("primary", tgt->virt.vds)) == NULL) {
		warn("Cannot get vdisk bindings.\n");
		return (-1);
	}

	/*
	 * Allocate space for 'ufs_cnt' disks to allow for the worst case
	 * of one disk per file system.  Resize when the correct number is
	 * known.
	 */
	tgt->disk_cnt = tgt->ufs_cnt;
	tgt->disks =
	    (disk_info_t *)Malloc(tgt->disk_cnt * sizeof (disk_info_t));

	disk_cnt = 0;
	for (i = 0; i < tgt->ufs_cnt; i++) {
		char	tgt_name[MAXPATHLEN];
		char	*disk;
		char	*slice;

		disk = Strdup(tgt->ufss[i].pri_special);
		if ((slice = strrchr(disk, 's')) == NULL)
			return (-1);
		*slice = '\0';
		if (have_disk(tgt->disks, disk_cnt, disk) == B_FALSE) {
			tgt->disks[disk_cnt].nr = disk_cnt;
			tgt->disks[disk_cnt].blks = -1;
			tgt->disks[disk_cnt].name = NULL;
			tgt->disks[disk_cnt].pri_name = disk;
			(void) snprintf(tgt_name, sizeof (tgt_name),
			    "/dev/dsk/c0d%d", disk_cnt);
			tgt->disks[disk_cnt].virt_name = Strdup(tgt_name);
			disk_cnt++;
		}
	}
	tgt->disk_cnt = disk_cnt;
	if ((tgt->disks = realloc(tgt->disks,
	    disk_cnt * sizeof (disk_info_t))) == NULL)
		die("Out of memory\n");

	/*
	 * Get the names of the vdisks
	 */
	for (i = 0; i < tgt->disk_cnt; i++) {
		char	disk[MAXPATHLEN];
		char	*dev;
		const char *vdisk;

		(void) snprintf(disk, sizeof (disk), "%ss2",
		    tgt->disks[i].pri_name);

		if ((dev = get_disk_dev_name(disk)) == NULL) {
			return (-1);
		}
		if ((vdisk = ldm_get_vdisk_name_by_dev(bnd, dev)) == NULL) {
			warn("Cannot get vdisk name for dev %s.\n", dev);
			free(dev);
			return (-1);
		}
		tgt->disks[i].vdisk = Strdup(vdisk);
		free(dev);
	}
	free(bnd);
	setup_fs_translations(tgt);
	return (0);
}

/*
 * Move the vdisks from the primary domain to the guest domain.
 */
int
attach_vdisks(config_t *tgt)
{
	int	i;

	msg(0, "Attaching vdisks to domain %s ...\n", tgt->virt.name);

	for (i = 0; i < tgt->disk_cnt; i++) {
		char	vdisk[128];
		char	vol[128];

		msg(1, "Attaching vdisk disk%d ...\n", i);

		if (ldm_rm_vdisk("primary", tgt->disks[i].vdisk) != 0) {
			warn("Cannot remove vdisk %s from control domain.\n",
			    tgt->disks[i].vdisk);
			return (-1);
		}

		(void) snprintf(vdisk, sizeof (vdisk), "disk%d", i);
		(void) snprintf(vol, sizeof (vol), "%s-vol%d@%s",
		    tgt->virt.name, i, tgt->virt.vds);
		if (ldm_add_vdisk(tgt->virt.name, vdisk, vol) != 0) {
			warn("Cannot add vdisk %s.\n", vdisk);
			return (-1);
		}
	}
	return (0);
}

/*
 * Put a VTOC on the vdisk. The disk geometry has changed so we need to align
 * all partitions on cylinder boundaries. Rather then reimplement parts of
 * prtvtoc and fmthard, we let them do the work for us.
 */
static int
label_vdisk(config_t *tgt, int disk_nr)
{
	char	raw[MAXPATHLEN];
	const char	*disk;
	slice_info_t	*slicep;
	int	align;
	int	part;
	long	start;
	long	end;

	if ((disk = strchr(tgt->disks[disk_nr].pri_name, 'c')) == NULL)
		return (-1);
	(void) snprintf(raw, sizeof (raw), "/dev/rdsk/%ss2", disk);

	msg(1, "Creating VTOC on %s (disk%d) ...\n", raw, disk_nr);

	if ((align = get_sect_per_cyl(raw)) == -1) {
		warn("Cannot get disk geometry.\n");
		return (-1);
	}

	/*
	 * Calculate new cylinder aligned partition boundaries. Skip slice 2
	 * since fmthard(1M) will create that for us...
	 */
	start = 0;
	for (part = 0; part < 8; part++) {
		long	cyls;

		if ((slicep = get_slice_by_disk_id(tgt,
		    tgt->disks[disk_nr].nr, part)) == NULL ||
		    slicep->slice == 2 ||
		    (slicep->start == 0 && slicep->end == 0))
			continue;

		cyls = (slicep->blks / align) + 1;
		slicep->blks = cyls * align;
		slicep->start = start;
		end = start + slicep->blks - 1;
		slicep->end = end;
		start = end + 1;
	}
	return (fmthard(tgt, disk_nr, raw) == 0 ? 0 : -1);
}

/*
 * Run fmthard(1M) to put the VTOC on the disk.
 */
static int
fmthard(config_t *tgt, int disk_nr, const char *raw)
{
	slice_info_t	*slicep;
	int	part;
	pid_t	pid;
	int	status;
	int	fd[2];
	FILE	*f;

	if (pipe(fd) == -1 || (pid = fork()) == -1) {
		warn("Cannot fork for %s", FMTHARD_BIN);
		return (-1);
	}

	if (pid == 0) {
		(void) dup2(fd[0], STDIN_FILENO);
		(void) dup2(devnull, STDOUT_FILENO);
		close(fd[1]);
		close(fd[0]);
		(void) execl(FMTHARD_BIN, FMTHARD_CMD, "-s", "-", raw,
		    NULL);
		_exit(1);
	}

	close(fd[0]);
	if ((f = fdopen(fd[1], "w")) == NULL)
		return (-1);

	for (part = 0; part < 8; part++) {
		if ((slicep = get_slice_by_disk_id(tgt,
		    tgt->disks[disk_nr].nr, part)) == NULL ||
		    slicep->slice == 2 ||
		    (slicep->start == 0 && slicep->end == 0))
			continue;

		fprintf(f, "%8d%7d    %2d%11ld%10ld\n", part, slicep->tag,
		    slicep->flag, slicep->start, slicep->blks);
	}
	fclose(f);
	close(fd[1]);
	(void) waitpid(pid, &status, 0);
	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0) {
		return (-1);
	}
	return (0);
}

/*
 * Setup the translations for disk and special device names so we can easily
 * translate a disk name on the original physical system to the name in the
 * final guest domain or on the control domain (and vice versa).
 */
static int
setup_disk_translations(config_t *tgt)
{
	binding_t *bnd;
	vdisk_binding_t	*vdisk;

	if ((bnd = ldm_get_bindings("primary", tgt->virt.vds)) == NULL)
		die("Cannot get vdisk bindings.\n");

	for (vdisk = bnd->vdisks; vdisk != NULL; vdisk = vdisk->next) {
		static int	j = 0;

		if (j > tgt->disk_cnt) {
			ldm_free_bindings(bnd);
			return (-1);
		}
		if (strstr(vdisk->name, tgt->virt.name) == vdisk->name) {
			char	buf[MAXPATHLEN];
			char	*pri_name;

			pri_name = get_disk_pri_name(vdisk->dev);
			if (pri_name == NULL) {
				warn("Cannot determine disk name for %s\n",
				    vdisk->dev);
				return (-1);
			}
			(void) snprintf(buf, sizeof (buf), "/dev/dsk/%s",
			    pri_name);
			tgt->disks[j].pri_name = Strdup(buf);
			(void) snprintf(buf, sizeof (buf), "/dev/dks/c0d%d",
			    j);
			tgt->disks[j].virt_name = Strdup(buf);
			j++;
		}
	}
	ldm_free_bindings(bnd);
	return (0);
}

/*
 * Setup translations for disk names used in file system entries so we can
 * easily map device names in the control domain and the guest domain.
 *
 * Translations can be setup from the special device names on the original
 * physical system (phys -> pri & virt), or from the special device names in
 * the control domain (pri -> virt) i.e. when the -R option is used.
 */
static int
setup_fs_translations(config_t *tgt)
{
	int	i;

	for (i = 0; i < tgt->ufs_cnt; i++) {
		char	buf[MAXPATHLEN];
		char	*slice;
		char	*special;
		int	j;

		if (tgt->ufss[i].special != NULL) {
			/*
			 * phys -> pri & virt
			 */
			special = tgt->ufss[i].special;
			for (j = 0; j < tgt->disk_cnt; j++) {
				if (tgt->disks[j].name == NULL ||
				    tgt->disks[j].pri_name == NULL ||
				    tgt->disks[j].virt_name == NULL) {
					warn("Names not set for disk %d\n", j);
					return (-1);
				}
				if (strstr(special, tgt->disks[j].name) ==
				    special) {
					slice = strrchr(special, 's');
					(void) snprintf(buf, sizeof (buf),
					    "%s%s", tgt->disks[j].pri_name,
					    slice);
					tgt->ufss[i].pri_special = Strdup(buf);
					(void) snprintf(buf, sizeof (buf),
					    "%s%s", tgt->disks[j].virt_name,
					    slice);
					tgt->ufss[i].virt_special = Strdup(buf);
					break;
				}
			}
		} else {
			/*
			 * pri -> virt
			 */
			special = tgt->ufss[i].pri_special;
			for (j = 0; j < tgt->disk_cnt; j++) {
				if (tgt->disks[j].pri_name == NULL ||
				    tgt->disks[j].virt_name == NULL) {
					warn("Names not set for disk %d\n", j);
					return (-1);
				}
				if (strstr(special, tgt->disks[j].pri_name) ==
				    special) {

					slice = strrchr(special, 's');
					(void) snprintf(buf, sizeof (buf),
					    "%s%s", tgt->disks[j].virt_name,
					    slice);
					tgt->ufss[i].virt_special = Strdup(buf);
					break;
				}
			}
		}
	}
	return (0);
}

/*
 * Return the Solaris disk name (cXdY) for a vdisk device instance (dev@n) in
 * the current domain.
 */
char *
get_disk_pri_name(const char *dev)
{
	char	link[MAXPATHLEN];
	DIR	*dirp;
	struct dirent	*dp;

	(void) snprintf(link, sizeof (link),
	    "../../devices/virtual-devices@100/channel-devices@200"
	    "/%s:c", dev);
	if ((dirp = opendir("/dev/dsk")) == NULL) {
		return (NULL);
	}
	do {
		if ((dp = readdir(dirp)) != NULL) {
			char	name[MAXPATHLEN];
			char	buf[MAXPATHLEN];
			char	*s;

			(void) snprintf(name, sizeof (name),
			    "/dev/dsk/%s", dp->d_name);
			bzero(buf, sizeof (buf));
			if (readlink(name, buf, sizeof (buf)) != -1 &&
			    strstr(buf, link) == buf) {
				closedir(dirp);
				/*
				 * Remove slice and /dev/dsk/
				 */
				if ((s = strrchr(name, 's')) != NULL)
					*s = '\0';
				return (strdup(name + 9)); /* "/dev/dsk/" */
			}
		}
	} while (dp != NULL);
	closedir(dirp);
	return (NULL);
}

/*
 * Define the disks for the guest domain. Currently we don't do relayout but
 * just copy the disk and file system setup from the physical system.
 */
static void
setup_vdisks(const config_t *src, config_t *tgt)
{
	int	i;
	slice_info_t	*slicep;

	tgt->disk_cnt = src->disk_cnt;
	tgt->disks = (disk_info_t *)
	    Malloc(tgt->disk_cnt * sizeof (disk_info_t));
	for (i = 0; i < tgt->disk_cnt; i++) {
		tgt->disks[i].nr = src->disks[i].nr;
		tgt->disks[i].blks = src->disks[i].blks;
		tgt->disks[i].resized = 0;
		tgt->disks[i].name = Strdup(src->disks[i].name);
	}

	tgt->slice_cnt = src->slice_cnt;
	tgt->slices = (slice_info_t *)
	    Malloc(tgt->slice_cnt * sizeof (slice_info_t));
	for (i = 0; i < tgt->slice_cnt; i++) {
		tgt->slices[i].nr = src->slices[i].nr;
		tgt->slices[i].slice = src->slices[i].slice;
		tgt->slices[i].tag = src->slices[i].tag;
		tgt->slices[i].flag = src->slices[i].flag;
		tgt->slices[i].start = src->slices[i].start;
		tgt->slices[i].end = src->slices[i].end;
		tgt->slices[i].blks = src->slices[i].blks;
		tgt->slices[i].special =
		    Strdup(src->slices[i].special);
		tgt->slices[i].type = SLICE_TYPE_UNUSED;
	}

	tgt->ufs_cnt =  src->ufs_cnt;
	tgt->ufss =  (ufs_info_t *)
	    Malloc(tgt->ufs_cnt * sizeof (ufs_info_t));
	for (i = 0; i < tgt->ufs_cnt; i++) {
		tgt->ufss[i].nr = src->ufss[i].nr;
		tgt->ufss[i].mntpt = Strdup(src->ufss[i].mntpt);
		tgt->ufss[i].special = Strdup(src->ufss[i].special);
		tgt->ufss[i].blks = src->ufss[i].blks;
		tgt->ufss[i].free = src->ufss[i].free;
		tgt->ufss[i].avail = src->ufss[i].avail;

		slicep = get_slice_by_special(tgt, tgt->ufss[i].special);
		if (slicep == NULL)
			die("No slice found for %s\n", tgt->ufss[i].special);
		slicep->type = SLICE_TYPE_UFS;
	}

	tgt->swap_cnt = src->swap_cnt;
	tgt->swaps =  (swap_info_t *)
	    Malloc(tgt->swap_cnt * sizeof (swap_info_t));
	for (i = 0; i < tgt->swap_cnt; i++) {
		tgt->swaps[i].nr = src->swaps[i].nr;
		tgt->swaps[i].blks = src->swaps[i].blks;
		tgt->swaps[i].special = Strdup(src->swaps[i].special);

		slicep = get_slice_by_special(tgt, tgt->swaps[i].special);
		if (slicep == NULL)
			die("No slice found for %s\n", tgt->swaps[i].special);
		slicep->type = SLICE_TYPE_SWAP;
	}
}

/*
 * Remove unused slices to reduce the size of the vdisk.
 */
static void
remove_unused_slices(config_t *tgt)
{
	int	i;

	msg(0, "Removing unused slices ...\n");

	for (i = 0; i < tgt->slice_cnt; i++) {
		if (tgt->slices[i].type == SLICE_TYPE_UNUSED &&
		    tgt->slices[i].slice != 2) {
			disk_info_t	*diskp;

			msg(2, "Removing unused slice %s ...\n",
			    tgt->slices[i].special);

			diskp = get_disk_by_id(tgt, tgt->slices[i].nr);
			if (diskp == NULL)
				die("ERROR: no disk for %s.\n",
				    tgt->slices[i].special);
			diskp->resized = 1;
			diskp->blks -= tgt->slices[i].blks;

			tgt->slices[i].start = 0;
			tgt->slices[i].end = 0;
			tgt->slices[i].blks = 0;
		}
	}
}

/*
 * Change the size of the disk slice for this swap partition. Set the resize
 * flag to force recalculation of this disk's VTOC.
 */
static int
resize_swap(config_t *tgt, resize_info_t *sizep, int nr)
{
	swap_info_t	*swapp;
	slice_info_t	*slicep;
	disk_info_t	*diskp;

	if ((swapp = get_swap_by_id(tgt, nr)) == NULL) {
		warn("ERROR: resize of swap partition %d requested but "
		    "this swap partition is not defined.\n", nr);
		return (-1);
	}
	slicep = get_slice_by_special(tgt, swapp->special);
	if (slicep == NULL)
		die("ERROR: no slice for swap partition %s.\n",
		    swapp->special);
	if ((diskp = get_disk_by_id(tgt, slicep->nr)) == NULL)
		die("ERROR: no disk for %s.\n", swapp->special);

	msg(4, "Resize swap partiton %s: new size = %ld blocks.\n",
	    swapp->special, sizep->blks);

	slicep->blks = sizep->blks;
	diskp->resized = 1;
	diskp->blks = 0;
	return (0);
}

/*
 * Change the size of the disk slice for this file system. Set the resize
 * flag to force recalculation of this disk's VTOC.
 */
static int
resize_filesystem(config_t *tgt, resize_info_t *sizep)
{
	ufs_info_t	*ufsp;
	slice_info_t	*slicep;
	disk_info_t	*diskp;
	long	used_blks;
	long	min_blks;

	if ((ufsp = get_ufs_by_mntpt(tgt, sizep->mntpt)) == NULL) {
		warn("ERROR: resize of %s requested but this is not "
		    "a mountpoint.\n", sizep->mntpt);
		return (-1);
	}
	slicep = get_slice_by_special(tgt, ufsp->special);
	if (slicep == NULL)
		die("ERROR: no slice for mountpoint %s.\n",
		    ufsp->special);
	if ((diskp = get_disk_by_id(tgt, slicep->nr)) == NULL)
		die("ERROR: no disk for %s.\n", ufsp->special);

	used_blks = ufsp->blks - ufsp->free;
	min_blks = (used_blks * 10) / 8;	/* UFS overhead */

	msg(4, "Resize %s: new size = %ld blocks, used = %ld, "
	    "(min size = %ld blocks).\n", ufsp->mntpt,
	    sizep->blks, used_blks, min_blks);

	/*
	 * Verify that the new size is not too small.
	 */
	if (sizep->blks <= used_blks) {
		warn("ERROR: the requested size for file system %s "
		    "is below its current usage.\n", ufsp->mntpt);
		return (-1);
	}
	if (sizep->blks <= min_blks) {
		warn("ERROR: the requested size for file system %s "
		    "is too small.\n", ufsp->mntpt);
		return (-1);
	}
	slicep->blks = sizep->blks;
	diskp->resized = 1;
	diskp->blks = 0;
	return (0);
}

/*
 * Resize file systems or the first swap partition by changing the size of the
 * underlying disk slices. We walk the resize_list, check for valid mountpoints
 * or the special name 'swap' and update the size of the disk slice.
 */
static int
resize_partitions(config_t *tgt)
{
	resize_info_t	*sizep;
	int	rc = 0;

	msg(0, "Resizing partitions ...\n");

	sizep = tgt->virt.resize_list;
	while (sizep != NULL) {

		msg(2, "Resize %s\n", sizep->mntpt);
		if (strcmp(sizep->mntpt, "swap") == 0)
			rc += resize_swap(tgt, sizep, 0);
		else
			rc += resize_filesystem(tgt, sizep);

		sizep = sizep->next;
	}
	return (rc == 0 ? 0 : -1);
}

/*
 * Recalculate the partition start and end for all slices on every disk that
 * has changed size.
 */
static void
recalc_vtoc(config_t *tgt)
{
	int	i;

	for (i = 0; i < tgt->disk_cnt; i++) {
		int	part;
		long	start = 0;
		long	end = 0;
		slice_info_t	*slicep;

		if (tgt->disks[i].resized != 1)
			continue;

		msg(1, "Partition(s) on disk %s were resized, adjusting disk"
		    "size ...\n", tgt->disks[i].name);

		for (part = 0; part < 8; part++) {
			if ((slicep = get_slice_by_disk_id(tgt,
			    tgt->disks[i].nr, part)) == NULL ||
			    slicep->slice == 2 ||
			    (slicep->start == 0 && slicep->end == 0))
				continue;

			slicep->start = start;
			end = start + slicep->blks - 1;
			slicep->end = end;
			start = end + 1;
		}
		if ((slicep = get_slice_by_disk_id(tgt,
		    tgt->disks[i].nr, 2)) == NULL)
			die("No s2 on disk %s!\n", tgt->disks[i].name);

		slicep->start = 0;
		slicep->end = end;
		slicep->blks = start;
		tgt->disks[i].blks = start;
	}
}

/*
 * Create a vds backend device. To allow for geometry change we add 500 MB
 * to the requested size.
 */
static int
create_vdsdev(config_t *tgt, const char *vol, int disk_nr)
{
	char	backend[MAXPATHLEN];
	char	name[MAXPATHLEN];
	long	blks;

	blks = tgt->disks[disk_nr].blks + 1024000;

	msg(1, "Creating volume %s (%ld MB)...\n", vol, blks / 2048);
	switch (tgt->virt.backend_type) {
	case BACKEND_TYPE_ZVOL:
		(void) snprintf(name, sizeof (name), "%s/%s/disk%d",
		    tgt->virt.backend_path, tgt->virt.name, disk_nr);
		if (create_zvol(name, blks, tgt->virt.sparse) != 0) {
			warn("Cannot create ZFS volume %s.\n", name);
			return (-1);
		}
		(void) snprintf(backend, sizeof (backend), "/dev/zvol/dsk/%s",
		    name);
		break;
	case BACKEND_TYPE_FILE:
		(void) snprintf(backend, sizeof (backend), "/%s/%s/disk%d",
		    tgt->virt.backend_path, tgt->virt.name, disk_nr);
		if (create_file(backend, blks, tgt->virt.sparse) != 0) {
			warn("Cannot create backend file %s.\n", backend);
			return (-1);
		}
		break;
	default:
		die("Unknown backend type (%d)\n", tgt->virt.backend_type);
	}
	return (ldm_add_vdsdev(backend, vol) == 0 ? 0 : -1);
}

/*
 * Remove a vds volume and its backend device.
 */
int
remove_vdsdev(const char *vol, const char *device)
{
	const char *zvol_prefix = "/dev/zvol/dsk/";

	msg(2, "Removing volume %s ...\n", vol);
	if (ldm_rm_vdsdev(vol) == -1) {
		warn("Cannot remove volume %s\n", vol);
		return (-1);
	}
	if (strstr(device, zvol_prefix) == device) {
		const char *name = device + strlen(zvol_prefix);

		msg(2, "Removing ZFS volume %s ...\n", name);
		return (exec_cmd(1, ZFS_BIN, ZFS_CMD, "destroy", name,
		    NULL) == 0 ? 0 : -1);
	} else {
		msg(2, "Removing backend file %s ...\n", device);
		return (remove(device) == 0 ? 0 : -1);
	}
}

/*
 * Create a ZFS volume to use as a backend device.  Intermediate datasets will
 * be created if necessary.
 */
static int
create_zvol(const char *name, long blks, int sparse)
{
	char	size[32];
	char	*opt;

	(void) snprintf(size, sizeof (size), "%ldM", blks / 2048);
	opt = (sparse == 1 ? "-psV" : "-pV");

	msg(2, "Creating %sZFS volume %s ...\n", sparse == 1 ? "sparse " : "",
	    name);
	return (exec_cmd(0, ZFS_BIN, ZFS_CMD, "create", opt, size, name,
	    NULL) == 0 ? 0 : -1);
}

/*
 * Create a file to use as a backend device. Intermediate directories will be
 * created if necessary.
 */
static int
create_file(const char *name, long blks, int sparse)
{
	char	size[32];
	char	*dir;
	char	*s;

	(void) snprintf(size, sizeof (size), "%ldM", blks / 2048);

	msg(2, "Creating %sfile %s ...\n", sparse == 1 ? "sparse " : "", name);
	dir = Strdup(name);
	if ((s = strrchr(dir, '/')) != NULL)
		*s = '\0';
	make_dir(dir);
	free(dir);
	if (sparse)
		return (exec_cmd(0, MKFILE_BIN, MKFILE_CMD, "-n",
		    size, name, NULL) == 0 ? 0 : -1);
	else
		return (exec_cmd(0, MKFILE_BIN, MKFILE_CMD,
		    size, name, NULL) == 0 ? 0 : -1);
}

/*
 * Return the mountpoint for a ZFS dataset in buf.
 */
int
get_zfs_mountpoint(const char *dataset, char *buf, int length)
{
	pid_t	pid;
	int	status;
	int	fd[2];
	FILE	*f;
	char	*s;

	if (pipe(fd) == -1 || (pid = fork()) == -1) {
		warn("Cannot fork %s", ZFS_BIN);
		return (-1);
	}
	if (pid == 0) {
		(void) dup2(fd[1], STDOUT_FILENO);
		close(fd[0]);
		(void) dup2(devnull, STDERR_FILENO);
		(void) execl(ZFS_BIN, ZFS_CMD, "list", "-Ho",
		    "mountpoint", dataset, NULL);
		_exit(0);
	}
	close(fd[1]);
	if ((f = fdopen(fd[0], "r")) == NULL) {
		warn("fdopen() failed.\n");
		return (-1);
	}
	if (fgets(buf, length, f) == NULL) {
		fclose(f);
		return (-1);
	}
	fclose(f);
	(void) waitpid(pid, &status, 0);
	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0)
		return (-1);

	if ((s = strrchr(buf, '\n')) != NULL) {
		*s = '\0';
		return (0);
	}
	return (-1);
}

static slice_info_t *
get_slice_by_disk_id(config_t *sys, int disk_nr, int slice_nr)
{
	int	i;

	for (i = 0; i < sys->slice_cnt; i++) {
		if (sys->slices[i].nr == disk_nr &&
		    sys->slices[i].slice == slice_nr)
			return (&sys->slices[i]);
	}
	return (NULL);
}

static swap_info_t *
get_swap_by_id(config_t *sys, int nr)
{
	int	i;

	for (i = 0; i < sys->swap_cnt; i++) {
		if (sys->swaps[i].nr == nr)
			return (&sys->swaps[i]);
	}
	return (NULL);
}

static ufs_info_t *
get_ufs_by_mntpt(config_t *sys, const char *mntpt)
{
	int	i;

	for (i = 0; i < sys->ufs_cnt; i++) {
		if (strcmp(sys->ufss[i].mntpt, mntpt) == 0)
			return (&sys->ufss[i]);
	}
	return (NULL);
}

static slice_info_t *
get_slice_by_special(config_t *sys, const char  *special)
{
	int	i;

	for (i = 0; i < sys->slice_cnt; i++) {
		if (strcmp(sys->slices[i].special, special) == 0)
			return (&sys->slices[i]);
	}
	return (NULL);
}

static disk_info_t *
get_disk_by_id(config_t *sys, int nr)
{
	int	i;

	for (i = 0; i < sys->disk_cnt; i++) {
		if (sys->disks[i].nr == nr)
			return (&sys->disks[i]);
	}
	return (NULL);
}

/*
 * Return B_TRUE if 'name' is already present in 'disks'.
 */
static int
have_disk(disk_info_t *disks, int size, const char *name)
{
	int	i;

	for (i = 0; i < size && disks[i].pri_name != NULL; i++) {
		if (strcmp(disks[i].pri_name, name) == 0)
			return (B_TRUE);
	}
	return (B_FALSE);
}

/*
 * Return the disk@n part of a /dev/dsk/cXdY symlink.
 */
static char *
get_disk_dev_name(const char *disk)
{
	static char	link[MAXPATHLEN];
	char	*dev;
	char	*colon;

	if (readlink(disk, link, sizeof (link)) == -1 ||
	    (dev = strrchr(link, '/')) == NULL ||
	    (colon = strrchr(link, ':')) == NULL) {
		warn("Cannot get device for disk %s.\n", disk);
		return (NULL);
	}

	/*
	 * Make sure it is on a vdisk.
	 */
	if (strncmp(link, VDISK_DEVPATH, strlen(VDISK_DEVPATH)) != 0) {
		warn("Error: disk %s is not a virtual disk.\n", disk);
		return (NULL);
	}
	dev += 1;
	*colon = '\0';
	return (strdup(dev));
}
