/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <unistd.h>
#include <sys/mount.h>
#include <sys/mnttab.h>
#include <sys/param.h>
#include <sys/wait.h>

#include "util.h"
#include "ldmp2v.h"

#define	REQ_ROOT_SIZE	(2 * 1024)
#define	REQ_USR_SIZE	(5 * 1024)
#define	REQ_VAR_SIZE	(3 * 1024)
#define	SIZE_TOO_SMALL	"NOTICE: the size of the %s file system will" \
	" be increased to %ld MB\n"

static int extract_flar(const config_t *, const char *, const char *);
static int flash_id_section(FILE *, char **, char **);
static int exec_archiver(int, const char *, const char *);
static int extract_none(const config_t *);
static int extract_ufsdump(const config_t *, const char *);
static int newfs(const char *);
static int compare(const void *, const void *);

/*
 * Schedule a resize of a file system's slice by adding an entry to the
 * resize list.
 */
static void
resize(config_t *tgt, const char *mntpt, long size)
{
	resize_info_t	*ri;
	resize_info_t	*p;

	ri = (resize_info_t *)Malloc(sizeof (resize_info_t));
	ri->mntpt = Strdup(mntpt);
	ri->blks = size * 2048;
	if ((p = tgt->virt.resize_list) == NULL) {
		tgt->virt.resize_list = ri;
	} else {
		while (p->next != NULL) {
			p = p->next;
		}
		p->next = ri;
	}
}

/*
 * Verify that the disk slices for the OS file systems as specified in the
 * manifest are large enough to accomodate the increased disk space
 * requirements of Solaris 10.  We don't want to duplicate the calculations
 * the installer does here, but we'll make an educated guess that should be
 * good the majority of cases.
 */
void
adjust_filesystem_sizes(const config_t *src, config_t *tgt)
{
	long	root_size = 0;
	long	var_size = 0;
	long	usr_size = 0;
	long	req_root_size;
	long	req_var_size;
	long	req_usr_size;
	int	i;

	for (i = 0; i < src->ufs_cnt; i++) {
		if (strcmp(src->ufss[i].mntpt, "/") == 0) {
			root_size = (src->ufss[i].blks / 2048);
		} else if (strcmp(src->ufss[i].mntpt, "/usr") == 0) {
			usr_size = (src->ufss[i].blks / 2048);
		} else if (strcmp(src->ufss[i].mntpt, "/var") == 0) {
			var_size = (src->ufss[i].blks / 2048);
		}
	}
	/*
	 * Calculate required file system sizes depending on the file system
	 * configuration.
	 */
	if (var_size == 0 && usr_size == 0) {	/* / only */
		req_root_size = REQ_ROOT_SIZE + REQ_USR_SIZE + REQ_VAR_SIZE;
		req_usr_size = 0;
		req_var_size = 0;
	} else if (var_size == 0) {		/* /, /usr */
		req_root_size = REQ_ROOT_SIZE + REQ_VAR_SIZE;
		req_usr_size = REQ_USR_SIZE;
		req_var_size = 0;
	} else if (usr_size == 0) {		/* /, /var */
		req_root_size = REQ_ROOT_SIZE + REQ_USR_SIZE;
		req_usr_size = 0;
		req_var_size = REQ_VAR_SIZE;
	} else {				/* /, /usr, /var */
		req_root_size = REQ_ROOT_SIZE;
		req_usr_size = REQ_USR_SIZE;
		req_var_size = REQ_VAR_SIZE;
	}
	/*
	 * Check the required size against the actual size and resize if
	 * necessary.
	 */
	if (root_size < req_root_size) {
		warn(SIZE_TOO_SMALL, "/", req_root_size);
		resize(tgt, "/", req_root_size);
	}
	if (req_usr_size > 0 && usr_size < req_usr_size) {
		warn(SIZE_TOO_SMALL, "/usr", req_usr_size);
		resize(tgt, "/usr", req_usr_size);
	}
	if (req_var_size > 0 && var_size < req_var_size) {
		warn(SIZE_TOO_SMALL, "/var", req_var_size);
		resize(tgt, "/var", req_var_size);
	}
}

/*
 * Create the file systems required for the guest domain.
 */
int
create_filesystems(const config_t *tgt)
{
	int	i;

	msg(0, "Creating file systems ...\n");

	for (i = 0; i < tgt->ufs_cnt; i++) {
		char	raw[MAXPATHLEN];
		char	*dev;

		if ((dev = strchr(tgt->ufss[i].pri_special, 'c')) == NULL)
			return (-1);
		(void) snprintf(raw, sizeof (raw), "/dev/rdsk/%s", dev);
		msg(1, "Creating UFS file system on %s ...\n", raw);
		if (newfs(raw) != 0)
			return (-1);
	}
	return (0);
}

/*
 * Restore the file system data from whatever type of archive was created
 * during 'collect'.
 */
int
populate_filesystems(const config_t *tgt, const char *method, const char *dir,
    const char *nodename)
{
	char	buf[MAXPATHLEN];
	int	i;

	msg(0, "Populating file systems ...\n");

	for (i = 0; i < tgt->ufs_cnt; i++) {
		if (strcmp(tgt->ufss[i].mntpt, "/") == 0)
			(void) snprintf(buf, sizeof (buf), "%s",
			    tgt->virt.root);
		else
			(void) snprintf(buf, sizeof (buf), "%s%s",
			    tgt->virt.root, tgt->ufss[i].mntpt);

		if (make_dir(buf) != 0) {
			warn("Cannot create mountpoint %s.\n", buf);
			return (-1);
		}
		msg(2, "Mounting %s ...\n", buf);
		if (mount_fs(buf, tgt->ufss[i].pri_special) != 0) {
			warn("Cannot mount %s on %s.\n",
			    tgt->ufss[i].pri_special, buf);
			return (-1);
		}
	}
	if (strcmp(method, "ufsdump") == 0)
		return (extract_ufsdump(tgt, dir));
	else if (strcmp(method, "flash") == 0 || strcmp(method, "flar") == 0)
		return (extract_flar(tgt, dir, nodename));
	else if (strcmp(method, "none") == 0)
		return (extract_none(tgt));
	else {
		warn("Unsupported archive method '%s'.\n", method);
		return (-1);
	}
}

/*
 * Unmount all file systems for the geust domain.
 */
int
unmount_filesystems(const config_t *tgt)
{
	int	i;

	msg(0, "Unmounting file systems ...\n");

	for (i = tgt->ufs_cnt - 1; i >= 0; i--) {
		char	buf[MAXPATHLEN];

		if (i == 0)
			(void) snprintf(buf, sizeof (buf), "%s",
			    tgt->virt.root);
		else
			(void) snprintf(buf, sizeof (buf), "%s%s",
			    tgt->virt.root, tgt->ufss[i].mntpt);

		msg(1, "Unmounting %s ...\n", buf);
		if (umount(buf) != 0) {
			warn("Cannot unmount %s", buf);
			return (-1);
		}
	}
	return (0);
}

/*
 * Return the number of file systems mounted below virt.root. Update the
 * target config with an array of file systems sorted by mountpoint.
 */
int
get_mounted_filesystems(config_t *tgt, int strip_root)
{
	struct mnttab	mntent;
	struct mnttab	filter;
	FILE	*f;
	const char	*root = tgt->virt.root;
	int	cnt;
	int	i;
	int	rootlen;

	if ((f = fopen("/etc/mnttab", "r")) == NULL) {
		warn("Cannot open /etc/mnttab.\n");
		return (-1);
	}

	rootlen = strlen(root);
	cnt = 0;
	bzero(&filter, sizeof (filter));
	filter.mnt_fstype = "ufs";
	while (getmntany(f, &mntent, &filter) == 0) {
		if (strstr(mntent.mnt_mountp, root) == mntent.mnt_mountp)
			cnt++;
	}
	if (cnt == 0) {
		fclose(f);
		return (0);
	}
	tgt->ufss = (ufs_info_t *)Malloc(cnt * sizeof (ufs_info_t));

	/*
	 * Fill in the file system information.
	 */
	resetmnttab(f);
	bzero(&filter, sizeof (filter));
	filter.mnt_fstype = "ufs";
	i = 0;
	while (getmntany(f, &mntent, &filter) == 0) {
		if (strstr(mntent.mnt_mountp, root) == mntent.mnt_mountp) {
			if (strip_root) {
				if (strcmp(mntent.mnt_mountp, root) == 0)
					tgt->ufss[i].mntpt = Strdup("/");
				else
					tgt->ufss[i].mntpt =
					    Strdup(mntent.mnt_mountp + rootlen);
			} else {
				tgt->ufss[i].mntpt = Strdup(mntent.mnt_mountp);
			}
			tgt->ufss[i].special = NULL;
			tgt->ufss[i].pri_special = Strdup(mntent.mnt_special);
			tgt->ufss[i].virt_special = NULL;
			tgt->ufss[i].blks = 0;
			tgt->ufss[i].free = 0;
			tgt->ufss[i].avail = 0;
			i++;
		}
	}
	fclose(f);

	/*
	 * Sort ascending by mountpoint so mounts en unmounts can be done
	 * in the correct order.
	 */
	qsort((void *) tgt->ufss, cnt, sizeof (ufs_info_t), compare);
	for (i = 0; i < cnt; i++) {
		tgt->ufss[i].nr = i;
	}
	tgt->ufs_cnt = cnt;
	return (cnt);
}

static int
compare(const void *p1, const void *p2)
{
	ufs_info_t *m1 = (ufs_info_t *)p1;
	ufs_info_t *m2 = (ufs_info_t *)p2;

	return (strcmp(m1->mntpt, m2->mntpt));
}

/*
 * Extract the UFS dumps that were done during the 'collect' step.
 */
static int
extract_ufsdump(const config_t *tgt, const char *datadir)
{
	int	i;

	for (i = 0; i < tgt->ufs_cnt; i++) {
		char	archive[MAXPATHLEN];
		char	dir[MAXPATHLEN];
		char	old[MAXPATHLEN];

		if (i == 0)
			(void) snprintf(dir, sizeof (dir), "%s",
			    tgt->virt.root);
		else
			(void) snprintf(dir, sizeof (dir), "%s%s",
			    tgt->virt.root, tgt->ufss[i].mntpt);

		if (getcwd(old, sizeof (old)) == NULL) {
			warn("Cannot get current directory");
			return (-1);
		}
		if (chdir(dir) == -1) {
			warn("Cannot change directory to %s", dir);
			return (-1);
		}
		(void) snprintf(archive, sizeof (archive), "%s/ufsdump.%d",
		    datadir, i);
		msg(1, "Restoring %s to %s ...\n", archive, dir);
		if (exec_cmd(1, "/usr/sbin/ufsrestore", "ufsrestore", "rf",
		    archive, NULL) != 0) {
			warn("Restore of archive %s failed.\n", archive);
			return (-1);
		}
		(void) unlink("restoresymtable");
		if (chdir(old) == -1) {
			warn("Cannot change directory to %s", old);
			return (-1);
		}
	}
	return (0);
}

static const char *flash_cookie = "FlAsH-aRcHiVe";
static const char *id_begin = "section_begin=identification\n";
static const char *id_end = "section_end=identification\n";
static const char *archive_begin = "section_begin=archive\n";

/*
 * Extract the Flash archive that was made during the 'collect' step.
 */
static int
extract_flar(const config_t *tgt, const char *datadir, const char *nodename)
{
	char	archive[MAXPATHLEN];
	char	line[MAX_LINE_LENGTH];
	char	*archiver = NULL;
	char	*compression = NULL;
	FILE	*f;
	int	archive_section = B_FALSE;
	int 	result;

	(void) snprintf(archive, sizeof (archive), "%s/%s.flar",
	    datadir, nodename);
	msg(1, "Extracting Flash archive %s to %s ...\n", archive,
	    tgt->virt.root);

	if ((f = fopen(archive, "r")) == NULL) {
		warn("Cannot open Flash archive %s", archive);
		return (-1);
	}

	if (fgets(line, sizeof (line), f) == NULL ||
	    strstr(line, flash_cookie) == NULL) {
		warn("Not a valid Flash archive.\n");
		fclose(f);
		return (-1);
	}
	while (fgets(line, sizeof (line), f) != NULL) {
		if (strcmp(line, archive_begin) == 0) {
			archive_section = B_TRUE;
			break;
		} else if (strcmp(line, id_begin) == 0) {
			if (flash_id_section(f, &archiver,
			    &compression) == -1) {
				fclose(f);
				return (-1);
			}
		}
	}
	if (archive_section != B_TRUE) {
		warn("Not a valid Flash archive.\n");
		result = -1;
	} else {
		char	olddir[MAXPATHLEN];

		/* Reposition at the start of the actual archive */
		lseek(fileno(f), ftell(f), SEEK_SET);
		result = -1;
		if (getcwd(olddir, sizeof (olddir)) == NULL) {
			warn("Cannot get current directory");
			goto out;
		}
		if (chdir(tgt->virt.root) != 0) {
			warn("Cannot cd to %s", tgt->virt.root);
			goto out;
		}
		result = exec_archiver(fileno(f), archiver, compression);
		if (chdir(olddir) != 0) {
			warn("Cannot cd to %s", olddir);
			result = -1;
		}
	}
out:
	fclose(f);
	if (archiver != NULL)
		free(archiver);
	if (compression != NULL)
		free(compression);
	return (result == 0 ? 0 : -1);
}

static const char *archive_method = "files_archived_method";
static const char *compression_method = "files_compressed_method";

/*
 * Process the Flash identification section and extract the archiver and
 * compression type.
 */
static int
flash_id_section(FILE *f, char **archiver, char **compression)
{
	char line[MAX_LINE_LENGTH];

	while (fgets(line, sizeof (line), f) != NULL) {
		char	*name;
		char	*value;

		if (strcmp(line, id_end) == 0)
			return (0);

		if ((name = strtok(line, "=")) == NULL ||
		    (value = strtok(NULL, "\n")) == NULL) {
			warn("Bad Flash identification section.\n");
			return (-1);
		}
		if (strcmp(name, archive_method) == 0) {
			*archiver = Strdup(value);
		} else if (strcmp(name, compression_method) == 0) {
			*compression = Strdup(value);
		} else if (strcmp(name, "section_begin") == 0) {
			warn("Bad Flash identification section.\n");
			return (-1);
		}
	}
	return (-1);
}

/*
 * Run the appropriate archiver to extract the archive. If the archive is
 * compressed we pipe it through zcat first.
 */
static int
exec_archiver(int fd, const char *archiver, const char *compression)
{
	int 	arch[2];
	pid_t	arch_pid;
	pid_t	comp_pid;
	int	status;
	int	result = -1;
	int	compressed = B_TRUE;

	if (pipe(arch) == -1) {
		warn("Cannot create archiver pipe.\n");
		return (-1);
	}
	if (compression == NULL || strcmp(compression, "none") == 0)
		compressed = B_FALSE;

	if (compressed) {
		if (strcmp(compression, "compress") != 0) {
			warn("Unknown compression type '%s'.\n",
			    compression);
			return (-1);
		}
		if ((comp_pid = fork()) == -1) {
			warn("Cannot fork for %s", ZCAT_BIN);
			return (-1);
		} else if (comp_pid == 0) { 	/* compressor child */
			(void) dup2(fd, STDIN_FILENO);
			(void) dup2(arch[1], STDOUT_FILENO);
			close(arch[0]);
			close(arch[1]);
			(void) execl(ZCAT_BIN, ZCAT_CMD, NULL);
			_exit(1);
		}
	}

	if ((arch_pid = fork()) == -1) {
		if (archiver == NULL || strcmp(archiver, CPIO_CMD) == 0)
			warn("Cannot fork for %s", CPIO_BIN);
		else
			warn("Cannot fork for %s", PAX_BIN);
		return (-1);
	} else if (arch_pid == 0) { 	/* archiver child */
		(void) dup2(compressed ? arch[0] : fd, STDIN_FILENO);
		close(arch[0]);
		close(arch[1]);
		if (archiver == NULL || strcmp(archiver, CPIO_CMD) == 0)
			(void) execl(CPIO_BIN, CPIO_CMD, "-icdum", NULL);
		else
			(void) execl(PAX_BIN, PAX_CMD, "-r -p -e", NULL);
		_exit(1);
	}
	/* parent */
	close(arch[0]);
	close(arch[1]);
	result = 0;
	waitpid(arch_pid, &status, 0);
	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0)
		result = -1;
	if (compressed) {
		waitpid(comp_pid, &status, 0);
		if (!WIFEXITED(status) || WEXITSTATUS(status) != 0)
			result = -1;
	}
	return (result);
}

/*
 * Extract file system data from a user defined backup tool.  The file systems
 * have been mounted at their respective mountpoints.  After issuing a prompt
 * that the user can start restoring data, we wait until <ENTER> is typed to
 * signal that the restore has finished.
 */
static int
extract_none(const config_t *tgt)
{
	(void) fprintf(stdout, "Guest file systems have been mounted at %s."
	    "\nPlease start your restore and press <ENTER> when finished.\n",
	    tgt->virt.root);

	while (getchar() != '\n') {
		;
	}
	return (0);
}

static int
newfs(const char *raw)
{
	int	status;
	pid_t	pid;

	if ((pid = fork()) == -1) {
		warn("Cannor fork for %s", NEWFS_BIN);
		return (-1);
	}

	if (pid == 0) {
		/*
		 * newfs(1M) goes interactive if stdin is a terminal, so we
		 * make sure stdin isn't a terminal.
		 */
		(void) dup2(devnull, STDIN_FILENO);
		(void) dup2(devnull, STDOUT_FILENO);
		(void) dup2(devnull, STDERR_FILENO);
		execl(NEWFS_BIN, NEWFS_CMD, raw, NULL);
		_exit(1);

	}
	waitpid(pid, &status, 0);
	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0)
		return (-1);

	return (status == 0 ? 0 : -1);
}
