/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "util.h"
#include "ldmp2v.h"
#include "ldm.h"

#define	LDM_BIN		"/opt/SUNWldm/bin/ldm"
#define	LDM_CMD		"ldm"

static int parse_domain(char *, char **);
static int parse_vdsdev(char *, vdsdev_binding_t **);
static int parse_vdisk(char *, vdisk_binding_t **);

static const char *sep = "|\n";

/*
 * Return available VCPUs
 */
int
ldm_vcpu_avail(void)
{
	char	line[MAX_LINE_LENGTH];
	pid_t	pid;
	int	status;
	int	fd[2];
	FILE	*f;
	int	vcpu_avail = 0;

	if (pipe(fd) == -1 || (pid = fork()) == -1) {
		warn("Cannot fork for %s", LDM_BIN);
		return (NULL);
	}

	if (pid == 0) {
		(void) dup2(fd[1], STDOUT_FILENO);
		close(fd[0]);
		close(fd[1]);
		(void) execl(LDM_BIN, LDM_CMD, "list-devices", "-p",
		    "vcpu", NULL);
		exit(0);
	}

	close(fd[1]);
	if ((f = fdopen(fd[0], "r")) == NULL) {
		warn("fdopen() failed\n");
		return (NULL);
	}
	while (fgets(line, sizeof (line), f) != NULL) {
		if (strstr(line, "|pid=") == line)
			vcpu_avail++;
	}
	fclose(f);
	(void) wait(&status);
	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0) {
		return (0L);
	}
	return (vcpu_avail);
}

/*
 * Return available memory in MB.
 */
long
ldm_mem_avail(void)
{
	char	line[MAX_LINE_LENGTH];
	pid_t	pid;
	int	status;
	int	fd[2];
	FILE	*f;
	long	mem_avail = 0;

	if (pipe(fd) == -1 || (pid = fork()) == -1) {
		warn("Cannot fork for %s", LDM_BIN);
		return (NULL);
	}

	if (pid == 0) {
		(void) dup2(fd[1], STDOUT_FILENO);
		close(fd[0]);
		close(fd[1]);
		(void) execl(LDM_BIN, LDM_CMD, "list-devices", "-p",
		    "memory", NULL);
		exit(0);
	}

	close(fd[1]);
	if ((f = fdopen(fd[0], "r")) == NULL) {
		warn("fdopen() failed\n");
		return (NULL);
	}
	while (fgets(line, sizeof (line), f) != NULL) {
		if (strstr(line, "|pa=0x") == line) {
			char	*s;

			if ((s = strstr(line, "size=")) != NULL) {
				s += strlen("size=");
				mem_avail += (atoll(s) / (1024 * 1024));
			}
			continue;
		}
	}
	fclose(f);
	(void) wait(&status);
	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0) {
		return (0L);
	}
	return (mem_avail);
}

/*
 * Return the domain state and its vdsdev and vdisk bindings.
 */
binding_t *
ldm_get_bindings(const char *domain, const char *vds)
{
	char	line[512];
	char	vds_match[512];
	binding_t	*bnd = NULL;
	pid_t	pid;
	int	status;
	int	fd[2];
	FILE	*f;
	int	is_vds;
	int	err = 0;

	if ((bnd = (binding_t *)malloc(sizeof (binding_t))) == NULL)
		return (NULL);
	bzero(bnd, sizeof (binding_t));

	/*
	 * We only support a single vds for our vdisks, but the primary domain
	 * may have more than one vds configured.  Match only our vds when
	 * parsing the list-bindings output.
	 */
	(void) snprintf(vds_match, sizeof (vds_match), "VDS|name=%s\n", vds);

	if (pipe(fd) == -1 || (pid = fork()) == -1) {
		warn("Cannot fork for %s", LDM_BIN);
		return (NULL);
	}

	if (pid == 0) {
		(void) dup2(fd[1], STDOUT_FILENO);
		close(fd[0]);
		close(fd[1]);
		(void) execl(LDM_BIN, LDM_CMD, "list-bindings", "-p",
		    domain, NULL);
		exit(0);
	}

	close(fd[1]);
	if ((f = fdopen(fd[0], "r")) == NULL) {
		warn("fdopen() failed\n");
		return (NULL);
	}
	is_vds = 0;
	while (fgets(line, sizeof (line), f) != NULL) {
		if (strstr(line, "DOMAIN") == line) {
			if (parse_domain(line, &bnd->domain_state) == -1) {
				err = 1;
				goto out;
			}
			continue;
		}
		if (strstr(line, "VDISK") == line) {
			if (parse_vdisk(line, &bnd->vdisks) == -1) {
				err = 1;
				goto out;
			}
			continue;
		}
		if (strcmp(line, vds_match) == 0) {
			is_vds = 1;		/* VDS section start */
			continue;
		}
		if (is_vds) {
			if (*line == '|') {
				if (parse_vdsdev(line, &bnd->vdsdevs) == -1) {
					err = 1;
					goto out;
				}
				continue;
			} else {		/* != '|': next section */
				is_vds = 0;
			}
		}
	}
out:
	fclose(f);
	(void) wait(&status);
	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0) {
		ldm_free_bindings(bnd);
		return (NULL);
	}
	if (err == 1) {
		ldm_free_bindings(bnd);
		return (NULL);
	}
	return (bnd);
}

/*
 * Return the vdisk device instance (dev@n) given a volume@vsw.
 */
const char *
ldm_get_vdisk_dev_by_vol(binding_t *bnd, const char *vol)
{
	vdisk_binding_t *p;

	p = bnd->vdisks;
	while (p != NULL) {
		if (strcmp(p->vol, vol) == 0)
			return (p->dev);
		p = p->next;
	}
	return (NULL);
}

/*
 * Return the vdisk name given a device instance (dev@n).
 */
const char *
ldm_get_vdisk_name_by_dev(binding_t *bnd, const char *dev)
{
	vdisk_binding_t *p;

	p = bnd->vdisks;
	while (p != NULL) {
		if (strcmp(p->dev, dev) == 0)
			return (p->name);
		p = p->next;
	}
	return (NULL);
}

/*
 * Return the vdisk name given a volume@vsw.
 */
const char *
ldm_get_vdisk_name_by_vol(binding_t *bnd, const char *vol)
{
	vdisk_binding_t *p;

	p = bnd->vdisks;
	while (p != NULL) {
		if (strcmp(p->vol, vol) == 0)
			return (p->name);
		p = p->next;
	}
	return (NULL);
}

void
ldm_free_bindings(binding_t *bnd)
{
	if (bnd != NULL) {
		vdsdev_binding_t *vdsdev;
		vdisk_binding_t	*vdisk;

		if (bnd->domain_state != NULL)
			free(bnd->domain_state);

		vdsdev = bnd->vdsdevs;
		while (vdsdev != NULL) {
			vdsdev_binding_t	*p;

			if (vdsdev->dev != NULL)
				free(vdsdev->dev);
			if (vdsdev->vol != NULL)
				free(vdsdev->vol);
			p = vdsdev->next;
			free(vdsdev);
			vdsdev = p;
		}
		vdisk = bnd->vdisks;
		while (vdisk != NULL) {
			vdisk_binding_t	*p;

			if (vdisk->name != NULL)
				free(vdisk->name);
			if (vdisk->dev != NULL)
				free(vdisk->dev);
			if (vdisk->vol != NULL)
				free(vdisk->vol);
			p = vdisk->next;
			free(vdisk);
			vdisk = p;
		}
		free(bnd);
	}
}

/*
 * Parse the DOMAIN|...|state=aaaa|...| lines.
 */
static int
parse_domain(char *line, char **statep)
{
	char	*nvpair;

	if (strtok(line, sep) == NULL ||
	    strcmp(line, "DOMAIN") != 0)
		return (-1);

	while ((nvpair = strtok(NULL, sep)) != NULL) {
		char	*str;
		char	*eq;

		if ((str = strdup(nvpair)) != NULL &&
		    (eq = strchr(str, '=')) != NULL) {
			char	*name;
			char	*value;

			name = str;
			*eq = '\0';
			if ((value = strdup(eq + 1)) != NULL) {
				if (strcmp(name, "state") == 0) {
					*statep = value;
					free(str);
					return (0);
				} else {
					free(value);
				}
			}
			free(str);
		}
	}
	return (-1);
}

/*
 * Parse the |vol=aaaaa|...|...|dev=bbbbb lines.
 */
static int
parse_vdsdev(char *line, vdsdev_binding_t **listp)
{
	vdsdev_binding_t *vdsdev;
	char	*nvpair;

	if (strstr(line, "vol=") == NULL ||
	    strstr(line, "dev=") == NULL ||
	    (nvpair = strtok(line, sep)) == NULL)
		return (0);

	if ((vdsdev = (vdsdev_binding_t *)
	    malloc(sizeof (vdsdev_binding_t))) == NULL)
		return (-1);
	bzero(vdsdev, sizeof (vdsdev_binding_t));

	do {
		char	*str;
		char	*eq;

		if ((str = strdup(nvpair)) != NULL &&
		    (eq = strchr(str, '=')) != NULL) {
			char	*name;
			char	*value;
			vdsdev_binding_t *p;

			name = str;
			*eq = '\0';
			if ((value = strdup(eq + 1)) != NULL) {
				if (strcmp(name, "vol") == 0)
					vdsdev->vol = value;
				else if (strcmp(name, "dev") == 0)
					vdsdev->dev = value;
				else		/* ignore this nvpair */
					free(value);

				if (vdsdev->vol != NULL &&
				    vdsdev->dev != NULL) {
					free(str);
					if ((p = *listp) == NULL) {
						*listp = vdsdev;
					} else {
						while (p->next != NULL) {
							p = p->next;
						}
						p->next = vdsdev;
					}
					return (0);
				}
			}
			free(str);
		}
	} while ((nvpair = strtok(NULL, sep)) != NULL);
	if (vdsdev->dev != NULL)
		free(vdsdev->dev);
	if (vdsdev->vol != NULL)
		free(vdsdev->vol);
	free(vdsdev);
	return (-1);
}

/*
 * Parse a VDISK|name=....|...|... line.
 */
static int
parse_vdisk(char *line, vdisk_binding_t **listp)
{
	vdisk_binding_t	*vdisk;
	char	*nvpair;

	if (strtok(line, sep) == NULL ||
	    strcmp(line, "VDISK") != 0)
		return (-1);

	if ((vdisk = (vdisk_binding_t *)
	    malloc(sizeof (vdisk_binding_t))) == NULL)
		return (-1);
	bzero(vdisk, sizeof (vdisk_binding_t));

	while ((nvpair = strtok(NULL, sep)) != NULL) {
		char	*str;
		char	*eq;

		if ((str = strdup(nvpair)) != NULL &&
		    (eq = strchr(str, '=')) != NULL) {
			char	*name;
			char	*value;
			vdisk_binding_t	*p;

			name = str;
			*eq = '\0';
			if ((value = strdup(eq + 1)) != NULL) {
				if (strcmp(name, "name") == 0)
					vdisk->name = value;
				else if (strcmp(name, "dev") == 0)
					vdisk->dev = value;
				else if (strcmp(name, "vol") == 0)
					vdisk->vol = value;
				else		/* ignore this nvpair */
					free(value);

				if (vdisk->name != NULL &&
				    vdisk->dev != NULL &&
				    vdisk->vol != NULL) {
					free(str);
					if ((p = *listp) == NULL) {
						*listp = vdisk;
					} else {
						while (p->next != NULL) {
							p = p->next;
						}
						p->next = vdisk;
					}
					return (0);
				}
			}
			free(str);
		}
	}
	if (vdisk->name != NULL)
		free(vdisk->name);
	if (vdisk->dev != NULL)
		free(vdisk->dev);
	if (vdisk->vol != NULL)
		free(vdisk->vol);
	free(vdisk);
	return (-1);

}

/*
 * Return B_TRUE if the domain exists, B_FALSE otherwise.
 */
int
ldm_domain_exists(const char *domain)
{
	return (exec_cmd(1, LDM_BIN, LDM_CMD, "list-domain", domain,
	    NULL) == 0 ?  B_TRUE : B_FALSE);
}

/*
 * Stop a domain.
 */
int
ldm_stop_domain(const char *domain)
{
	return (exec_cmd(0, LDM_BIN, LDM_CMD, "stop-domain", domain, NULL));
}

/*
 * Unbind a domain.
 */
int
ldm_unbind_domain(const char *domain)
{
	return (exec_cmd(0, LDM_BIN, LDM_CMD, "unbind-domain", domain, NULL));
}

/*
 * Set the number of VCPUs for a domain.
 */
int
ldm_set_vcpu(const char *domain, int vcpu)
{
	char	buf[32];

	(void) snprintf(buf, sizeof (buf), "%d", vcpu);
	return (exec_cmd(0, LDM_BIN, LDM_CMD, "set-vcpu", buf, domain, NULL));
}

/*
 * Set the the memory size for a domain.
 */
int
ldm_set_memory(const char *domain, long memsize)
{
	char	buf[32];

	(void) snprintf(buf, sizeof (buf), "%ldM", memsize);
	return (exec_cmd(0, LDM_BIN, LDM_CMD, "set-mem", buf, domain, NULL));
}

/*
 * Add a domain.
 */
int
ldm_add_domain(const char *domain, const char *hostid)
{
	if (hostid == NULL)
		return (exec_cmd(0, LDM_BIN, LDM_CMD, "add-domain", domain,
		    NULL));
	else {
		char	buf[64];

		snprintf(buf, sizeof (buf), "hostid=%s", hostid);
		return (exec_cmd(0, LDM_BIN, LDM_CMD, "add-domain", buf,
		    domain, NULL));
	}
}

/*
 * Remove a domain.
 */
int
ldm_rm_domain(const char *domain)
{
	return (exec_cmd(0, LDM_BIN, LDM_CMD, "rm-domain", domain, NULL));
}

/*
 * Add a virtual network interface.
 */
int
ldm_add_vnet(const char *domain, const char *name, const char *vsw,
    const char *mac)
{
	if (mac == NULL)
		return (exec_cmd(0, LDM_BIN, LDM_CMD, "add-vnet", name, vsw,
		    domain, NULL));
	else {
		char	buf[64];

		snprintf(buf, sizeof (buf), "mac-addr=%s", mac);
		return (exec_cmd(0, LDM_BIN, LDM_CMD, "add-vnet", buf, name,
		    vsw, domain, NULL));
	}
}

/*
 * Remove a virtual network interface.
 */
int
ldm_rm_vnet(const char *domain, const char *name)
{
	return (exec_cmd(0, LDM_BIN, LDM_CMD, "rm-vnet", name, domain, NULL));
}

/*
 * Add a virtual disk server device.
 */
int
ldm_add_vdsdev(const char *backend, const char *vol)
{
	return (exec_cmd(0, LDM_BIN, LDM_CMD, "add-vdsdev", backend, vol,
	    NULL));
}

/*
 * Remove a virtuakl disk server device.
 */
int
ldm_rm_vdsdev(const char *vol)
{
	return (exec_cmd(0, LDM_BIN, LDM_CMD, "rm-vdsdev", vol, NULL));
}

/*
 * Add a virtual disk to a domain.
 */
int
ldm_add_vdisk(const char *domain, const char *disk, const char *vol)
{
	return (exec_cmd(0, LDM_BIN, LDM_CMD, "add-vdisk", disk, vol, domain,
	    NULL));
}

/*
 * Remove a vitual disk from a domain.
 */
int
ldm_rm_vdisk(const char *domain, const char *disk)
{
	return (exec_cmd(0, LDM_BIN, LDM_CMD, "rm-vdisk", disk, domain, NULL));
}
