/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/mount.h>
#include <sys/param.h>

#include "util.h"
#include "ldmp2v.h"
#include "ldm.h"
#include "disk.h"
#include "filesys.h"
#include "mods.h"
#include "manifest.h"

static int allocate_cpu(int, int);
static long allocate_mem(long, long);
static void set_defaults(config_t *);
static int parse_mopt(char *arg, config_t *);
static void set_root(config_t *);

static int create_domain(config_t *);
static int setup_vnets(config_t *, config_t *, int);

static void cleanup(config_t *);

void dump_phys(const config_t *);
void dump_virt(const config_t *);

static const char USAGE[] = "\
ldmp2v prepare [-b zvol | file ] [-c <cpu>] [-o keep-mac] [-o keep-hostid]\n\
               [-m <mountpoint>:<size> [-m ...]] [-M <memsize>] [-p <prefix>]\n\
               [-s] [-x no-auto-adjust-fs] [-x remove-unused-slices]\n\
               [-v] -d <data_dir> <domain>\n\
ldmp2v prepare -R <guest_root>  [-c <cpu>] [-o keep-mac] [-o keep-hostid]\n\
               [-M <memsize>] [-v] -d <data_dir> <domain>\n\
ldmp2v prepare -C <domain>\n\n";

static const char optstr[] = ":b:c:Cd:m:M:o:p:R:svx:";
static const char nullstr[] = "<null>";

static void
usage()
{
	(void) fprintf(stderr, "Usage:\n%s", USAGE);
	exit(1);
}

int
main(int argc, char *argv[])
{
	config_t	src;	/* physical system configuration data */
	config_t	tgt;	/* guest domain configuration data */
	int	is_altroot = B_FALSE;
	int	opt_b = B_FALSE;
	int	opt_c = B_FALSE;
	int	opt_C = B_FALSE;
	int	opt_m = B_FALSE;
	int	opt_M = B_FALSE;
	int	opt_o = B_FALSE;
	int	opt_p = B_FALSE;
	int	opt_s = B_FALSE;
	int	opt_x = B_FALSE;
	int	keep_mac = B_FALSE;
	int	keep_hostid = B_FALSE;
	int	auto_adjust_fs = B_TRUE;
	int	rm_unused_slices = B_FALSE;
	char	*dir;
	int	c;

	/*
	 * Print usage message without a header. Called by ldmp2v to display the
	 * 'prepare' usage message inline with other subcommand usage messages.
	 */
	if (argc == 2 && strcmp(argv[1], "usage") == 0) {
		(void) fprintf(stderr, USAGE);
		exit(1);
	}

	if ((devnull = open("/dev/null", O_RDWR)) == -1)
		die("Cannot open /dev/null\n");
	logfile = devnull;	/* XX */

	if (getuid() != 0)
		die("You must be root to run this program.\n");

	bzero(&src, sizeof (src));
	bzero(&tgt, sizeof (tgt));

	set_defaults(&tgt);

	verbose = 0;
	while ((c = getopt(argc, argv, optstr)) != -1) {
		switch (c) {
		case 'b':
			opt_b = B_TRUE;
			if (strcmp(optarg, "zvol") == 0)
				tgt.virt.backend_type = BACKEND_TYPE_ZVOL;
			else if (strcmp(optarg, "file") == 0)
				tgt.virt.backend_type = BACKEND_TYPE_FILE;
			else {
				warn("Invalid backend type '%s'\n", optarg);
				usage();
			}
			break;
		case 'c':
			opt_c = B_TRUE;
			tgt.cpu = atoi(optarg);
			if (tgt.cpu == 0) {
				warn("Invalid VCPU count '%s'\n", optarg);
				usage();
			}
			break;
		case 'C':
			opt_C = B_TRUE;
			break;
		case 'd':
			dir = optarg;
			break;
		case 'm':
			opt_m = B_TRUE;
			if (parse_mopt(optarg, &tgt) == -1)
				usage();
			break;
		case 'M':
			opt_M = B_TRUE;
			tgt.memsize = atoi(optarg);
			break;
		case 'o':
			opt_o = B_TRUE;
			if (strcmp(optarg, "keep-mac") == 0)
				keep_mac = B_TRUE;
			else if (strcmp(optarg, "keep-hostid") == 0)
				keep_hostid = B_TRUE;
			else {
				warn("Invalid option value '%s'.\n", optarg);
				usage();
			}
			break;
		case 'p':
			opt_p = B_TRUE;
			tgt.virt.backend_path = optarg;
			break;
		case 'R':
			is_altroot = B_TRUE;
			tgt.virt.root = optarg;
			break;
		case 's':
			opt_s = B_TRUE;
			tgt.virt.sparse = B_TRUE;
			break;
		case 'v':
			verbose++;
			break;
		case 'x':
			opt_x = B_TRUE;
			if (strcmp(optarg, "no-auto-adjust-fs") == 0)
				auto_adjust_fs = B_FALSE;
			else if (strcmp(optarg, "remove-unused-slices") == 0)
				rm_unused_slices = B_TRUE;
			else {
				warn("Invalid option value '%s'.\n", optarg);
				usage();
			}
			break;
		case ':':
			warn("Option -%c requires an argument.\n", optopt);
			usage();
			break;
		case '?':
			warn("Invalid option -%c.\n", optopt);
			usage();
			break;
		default:
			usage();
			break;
		}
	}

	if (is_altroot && (opt_b || opt_C || opt_m || opt_p || opt_s || opt_x))
		usage();

	/*
	 * Cleanup only reqires the domain name argument.
	 */
	if (opt_C) {
		if (opt_b || opt_c || opt_m || opt_M || opt_o ||
		    opt_p || opt_s || opt_x)
			usage();

		if (argc - optind != 1) {
			usage();
		}
		tgt.virt.name = Strdup(argv[optind]);
		cleanup(&tgt);
		exit(0);
	}

	/*
	 * We must at least have the domain name and directory arguments.
	 */
	if (argc - optind != 1 || dir == NULL)
		usage();
	if (dir[strlen(dir) - 1] == '/')
		dir[strlen(dir) - 1] = '\0';

	tgt.virt.name = Strdup(argv[optind]);
	if (ldm_domain_exists(tgt.virt.name) == B_TRUE)
		die("Domain %s already exists.\n", tgt.virt.name);

	read_manifest(dir, &src);

	tgt.cpu = allocate_cpu(tgt.cpu, src.cpu);
	tgt.memsize = allocate_mem(tgt.memsize, src.memsize);
	if (keep_hostid)
		tgt.hostid = Strdup(src.hostid);

	if (setup_vnets(&src, &tgt, keep_mac) == -1)
		die("Cannot setup virtual network interfaces.\n");

	if (is_altroot) {
		/*
		 * User-created vdisk(s) (prepare -R <dir>)
		 */
		if (get_mounted_filesystems(&tgt, B_TRUE) < 1)
			die("No file system mounted at %s.\n", tgt.virt.root);
		if (get_vdisks(&tgt) == -1)
			die("Cannot determine vdisk configuration.\n");
	} else {
		/*
		 * Create all vdisks required.  File system sizes will be
		 * adjusted if needed unless -m is used to specify a custom
		 * size or -x no-auto-adjust-fs is present.
		 */
		set_root(&tgt);
		if (!opt_m && auto_adjust_fs)
			adjust_filesystem_sizes(&src, &tgt);
		if (create_vdisks(&src, &tgt, rm_unused_slices) == -1)
			die("Cannot create vdisks.\n");
		if (create_filesystems(&tgt) == -1)
			die("Cannot create file systems.\n");
		if (populate_filesystems(&tgt, src.phys.method, dir,
		    src.nodename) == -1)
			die("Cannot populate file systems.\n");
	}
	if (modify_guest_image(&src, &tgt, is_altroot) == -1)
		die("Cannot modify guest OS image.\n");
	if (unmount_filesystems(&tgt) == -1)
		die("Cannot unmount guest file systems.\n");
	if (create_domain(&tgt) == -1)
		die("Cannot create domain.\n");
	if (attach_vdisks(&tgt) == -1)
		die("Cannot attach vdisks to domain.\n");

	exit(0);
}

/*
 * Read configuration settings from /etc/ldmp2v.conf, applying defaults for
 * parameters that are incomplete or not present in the file.
 */
static void
set_defaults(config_t *tgt)
{
	char	line[MAX_LINE_LENGTH];
	FILE	*f;

	tgt->virt.vds = "primary-vds0";
	tgt->virt.vsw = "primary-vsw0";
	tgt->virt.backend_path = "tank/ldoms";
	tgt->virt.backend_type = BACKEND_TYPE_ZVOL;
	tgt->virt.sparse = 0;

	if ((f = fopen("/etc/ldmp2v.conf", "r")) == NULL)
		return;
	while (fgets(line, sizeof (line), f) != NULL) {
		char	*name;
		char	*value;
		char	*s;

		if ((name = strtok(line, "=")) == NULL ||
		    (value = strtok(NULL, "\n")) == NULL)
			continue;
		/*
		 * Remove leading and trailing whitespace and quotes.
		 */
		while (*value != NULL &&
		    (*value == ' ' || *value == '\t' || *value == '\"')) {
			value++;
		}
		s = value;
		while (*s != NULL &&
		    (*s != ' '&& *s != '\t' && *s != '\"')) {
			s++;
		}
		if (*s != NULL)
			*s = '\0';

		if (strlen(value) == 0)		/* no value, use default */
			continue;

		if (strcasecmp(name, "vds") == 0)
			tgt->virt.vds = Strdup(value);
		else if (strcasecmp(name, "vsw") == 0)
			tgt->virt.vsw = Strdup(value);
		else if (strcasecmp(name, "backend_prefix") == 0)
			tgt->virt.backend_path = Strdup(value);
		else if (strcasecmp(name, "backend_type") == 0) {
			if (strcasecmp(value, "zvol") == 0)
				tgt->virt.backend_type = BACKEND_TYPE_ZVOL;
			else
				tgt->virt.backend_type = BACKEND_TYPE_FILE;
		} else if (strcasecmp(name, "backend_sparse") == 0) {
			if (strcasecmp(value, "yes") == 0)
				tgt->virt.sparse = 1;
			else
				tgt->virt.sparse = 0;
		}
	}
	fclose(f);
}

/*
 * Parse the -m option argument. Since we don't know whether the mountpoint
 * is valid for the system at the time we parse this, all resize requests
 * are added to the resize_list of the guest domain.
 */
static int
parse_mopt(char *arg, config_t *tgt)
{
	char	*mntpt;
	char	*size;
	char	*unit;
	long	multiplier;
	long	blks;
	resize_info_t *ri;
	resize_info_t *p;

	/* arg = <mountpoint>:<num>k|m|g */
	if ((mntpt = strtok(arg, ":")) == NULL ||
	    (size = strtok(NULL, ":\n")) == NULL ||
	    (unit = strpbrk(size, "bBkKmMgG")) == NULL ||
	    unit == size)
		return (-1);

	switch (*unit) {
	case 'b':
	case 'B':
		multiplier = 1;
		break;
	case 'g':
	case 'G':
		multiplier = 2097152;
		break;
	case 'k':
	case 'K':
		multiplier = 2;
		break;
	case 'm':
	case 'M':
	default:
		multiplier = 2048;
		break;
	}
	*unit = '\0';	/* chop unit */
	if ((blks = atol(size)) == 0)
		return (-1);

	ri = (resize_info_t *)Malloc(sizeof (resize_info_t));
	ri->mntpt = Strdup(mntpt);
	ri->blks = blks * multiplier;

	if ((p = tgt->virt.resize_list) == NULL) {
		tgt->virt.resize_list = ri;
	} else {
		while (p->next != NULL) {
			p = p->next;
		}
		p->next = ri;
	}
	return (0);
}

/*
 * Set the guest domain root.
 */
static void
set_root(config_t *tgt)
{
	char	buf[MAXPATHLEN];
	char	mntpt[MAXPATHLEN];
	char	*path;

	if (tgt->virt.backend_type == BACKEND_TYPE_ZVOL) {
		if (get_zfs_mountpoint(tgt->virt.backend_path, mntpt,
		    sizeof (mntpt)) == -1) {
			die("Cannot get mountpoint for dataset %s.\n",
			    tgt->virt.backend_path);
		}
		path = mntpt;
	} else {
		path = tgt->virt.backend_path;
	}
	(void) snprintf(buf, sizeof (buf), "%s%s/%s/a",
	    *path == '/' ? "" : "/", path, tgt->virt.name);
	tgt->virt.root = Strdup(buf);
}

/*
 * Allocate VCPUs to the guest domain. Currently we allocate 1 VCPU for each
 * physical CPU regardless of type and speed.
 */
static int
allocate_cpu(int virt_cpu, int phys_cpu)
{
	int	cpu_wanted;
	int	cpu_avail = ldm_vcpu_avail();

	if (virt_cpu != 0) {	/* overridden by -c option */
		cpu_wanted = virt_cpu;
	} else {
		cpu_wanted = phys_cpu;
	}
	msg(2, "Available VCPUs: %d\n", cpu_avail);

	if (cpu_wanted > cpu_avail) {
		warn("WARNING: there are insufficient free VCPUs available for "
		    "the guest domain (need: %d, available: %d).\n", cpu_wanted,
		    cpu_avail);
	}
	return (cpu_wanted);
}

/*
 * Allocate memory for the guest domain.
 */
static long
allocate_mem(long virt_mem, long phys_mem)
{
	long	mem_wanted;
	long	mem_avail = ldm_mem_avail();

	if (virt_mem != 0) {	/* overridden by -M option */
		mem_wanted = virt_mem;
	} else {
		/* require 1 GB minimum for the guest */
		if (phys_mem < 1024) {
			warn("NOTICE: setting guest memory to 1 GB\n");
			mem_wanted = 1024;
		} else {
			mem_wanted = phys_mem;
		}
	}
	msg(2, "Available memory: %ld MB\n", mem_avail);

	if (mem_wanted > mem_avail) {
		warn("WARNING: there is insufficient memory available for "
		    "the guest domain (need: %ld MB, available: %ld MB)\n",
		    mem_wanted, mem_avail);
	}
	return (mem_wanted);
}

/*
 * Generate the virtual network interface names for the guest domain and copy
 * the mac addresses of the physical interfaces if requested.
 */
static int
setup_vnets(config_t *src, config_t *tgt, int keep_mac)
{
	int	i;

	tgt->nets = (net_info_t *)
	    Malloc(src->net_cnt * sizeof (net_info_t));

	tgt->net_cnt = src->net_cnt;

	for (i = 0; i < src->net_cnt; i++) {
		char	ifname[32];

		(void) snprintf(ifname, sizeof (ifname), "vnet%d", i);
		if (keep_mac)
			tgt->nets[i].mac = Strdup(src->nets[i].mac);
		else
			tgt->nets[i].mac = NULL;
		tgt->nets[i].nr = src->nets[i].nr;
		tgt->nets[i].name = Strdup(ifname);
		tgt->nets[i].ipaddr = Strdup(src->nets[i].ipaddr);
	}
	return (0);
}

/*
 * Create the guest domain.
 */
static int
create_domain(config_t *tgt)
{
	int	i;

	msg(0, "Creating domain ...\n");

	if (ldm_add_domain(tgt->virt.name, tgt->hostid) != 0)
		die("Cannot create domain.\n");
	if (ldm_set_vcpu(tgt->virt.name, tgt->cpu) != 0)
		die("Cannot assign required number of VCPUs.\n");
	if (ldm_set_memory(tgt->virt.name, tgt->memsize) != 0)
		die("Cannot assign required memory.\n");

	for (i = 0; i < tgt->net_cnt; i++) {
		if (ldm_add_vnet(tgt->virt.name, tgt->nets[i].name,
		    tgt->virt.vsw, tgt->nets[i].mac) != 0)
			die("Cannot add virtual network interface %s.\n",
			    tgt->nets[i].name);
	}
	return (0);
}

/*
 * Remove the domain and all its vdisks and backend devices.
 */
static void
cleanup(config_t *tgt)
{
	binding_t *bnd;

	msg(0, "Cleaning up domain %s ...\n", tgt->virt.name);

	if ((bnd = ldm_get_bindings(tgt->virt.name, tgt->virt.vds)) != NULL) {
		if (strcmp(bnd->domain_state, "active") == 0) {
			msg(2, "Stopping domain %s ...\n", tgt->virt.name);
			ldm_stop_domain(tgt->virt.name);
			ldm_unbind_domain(tgt->virt.name);
		} else if (strcmp(bnd->domain_state, "bound") == 0) {
			ldm_unbind_domain(tgt->virt.name);
		}
		ldm_free_bindings(bnd);
		msg(2, "Removing domain %s ...\n", tgt->virt.name);
		if (ldm_rm_domain(tgt->virt.name) == -1)
			die("Cannot delete domain.\n");
	}

	/*
	 * The domain removal above has taken care of deleting the vdisks that
	 * were mapped to the guest domain.  If we failed half-way through a
	 * conversion, we may have vdisks still attached to the control domain
	 * and perhaps also file systems still mounted in the control domain.
	 */
	tgt->virt.root = "";
	if (get_mounted_filesystems(tgt, B_FALSE) < 0)
		die("Cannot determine mounted file systems.\n");

	if ((bnd = ldm_get_bindings("primary", tgt->virt.vds)) != NULL) {
		vdsdev_binding_t *vdsdev;
		char	*filter;
		int	len;
		int	have_mounts;
		int	i;

		len = strlen(tgt->virt.name) + strlen("-vol") + 1;
		filter = Malloc(len);
		(void) snprintf(filter, len, "%s-vol", tgt->virt.name);

		have_mounts = B_FALSE;
		for (vdsdev = bnd->vdsdevs; vdsdev != NULL;
		    vdsdev = vdsdev->next) {
			char	vol[MAXPATHLEN];
			char	disk[MAXPATHLEN];	/* /dev/dsk/cXdYs */
			char	*pri_name;
			const char	*vdisk_name;
			const char	*vdisk_dev;

			if (strstr(vdsdev->vol, filter) != vdsdev->vol)
				continue;

			(void) snprintf(vol, sizeof (vol), "%s@%s",
			    vdsdev->vol, tgt->virt.vds);

			/*
			 * Get the name of the vdisk on this volume (if any).
			 * Delete the volume if there is no vdisk on it.
			 */
			vdisk_name = ldm_get_vdisk_name_by_vol(bnd, vol);
			vdisk_dev = ldm_get_vdisk_dev_by_vol(bnd, vol);
			if (vdisk_name == NULL && vdisk_dev == NULL) {
				if (remove_vdsdev(vol, vdsdev->dev) != 0)
					warn("Cannot remove %s (%s)\n", vol,
					    vdsdev->dev);
				continue;
			}

			/*
			 * The volume has a vdisk on it, get the Solaris disk
			 * name and see if there are any mounted file systems
			 * on the disk.  If so, mark them so we can unmount
			 * them later in the proper order after we have
			 * examined all vdisks.
			 */
			if ((pri_name = get_disk_pri_name(vdisk_dev)) == NULL) {
				continue;
			}
			(void) snprintf(disk, sizeof (disk), "/dev/dsk/%ss",
			    pri_name);
			free(pri_name);

			for (i = tgt->ufs_cnt - 1; i >= 0; i--) {
				if (strstr(tgt->ufss[i].pri_special, disk) ==
				    tgt->ufss[i].pri_special) {
					tgt->ufss[i].free = -1;
					have_mounts = B_TRUE;
				}
			}
		}

		if (have_mounts) {
			for (i = tgt->ufs_cnt - 1; i >= 0; i--) {
				if (tgt->ufss[i].free == -1) {
					msg(2, "Unmounting %s ...\n",
					    tgt->ufss[i].mntpt);
					if (umount(tgt->ufss[i].mntpt) == -1)
						warn("Cannot unmount %s",
						    tgt->ufss[i].mntpt);
				}
			}
		}

		/*
		 * All vdisks can now be deleted since they no longer have
		 * file systems mounted.
		 */
		for (vdsdev = bnd->vdsdevs; vdsdev != NULL;
		    vdsdev = vdsdev->next) {
			char	vol[MAXPATHLEN];
			const char	*vdisk_name;
			const char	*vdisk_dev;

			if (strstr(vdsdev->vol, filter) != vdsdev->vol)
				continue;

			(void) snprintf(vol, sizeof (vol), "%s@%s",
			    vdsdev->vol, tgt->virt.vds);

			vdisk_name = ldm_get_vdisk_name_by_vol(bnd, vol);
			vdisk_dev = ldm_get_vdisk_dev_by_vol(bnd, vol);
			if (vdisk_name == NULL && vdisk_dev == NULL) {
				continue;
			}

			msg(2, "Removing vdisk %s ...\n", vdisk_name);
			if (ldm_rm_vdisk("primary", vdisk_name) == -1)
				warn("Cannot remove vdisk %s\n", vdisk_name);

			if (remove_vdsdev(vol, vdsdev->dev) != 0)
				warn("Cannot remove %s (%s)\n", vol,
				    vdsdev->dev);
		}
		ldm_free_bindings(bnd);
	}
	exit(0);
}

void
dump_phys(const config_t *p)
{
	int	i;

	fprintf(stderr, "cpu: %d\nmemory: %ld MB\n", p->cpu,
	    p->memsize);
	fprintf(stderr, "hostid: %s\narchive_method: %s\n", p->hostid,
	    p->phys.method);
	fprintf(stderr, "OS release: %s\n", p->phys.os_rel);
	fprintf(stderr, "disks: %d\n", p->disk_cnt);
	for (i = 0; i < p->disk_cnt; i++) {
		char	*name = p->disks[i].name;

		fprintf(stderr, "\t%d %s %9ld blks\n", p->disks[i].nr,
		    name ? name : nullstr, p->disks[i].blks);
	}
	fprintf(stderr, "slices: %d\n", p->slice_cnt);
	for (i = 0; i < p->slice_cnt; i++) {
		char	*name = p->slices[i].special;

		fprintf(stderr, "\t%d %s %d %d %02d %9ld %9ld %9ld\n",
		    p->slices[i].nr,  name ? name : nullstr,
		    p->slices[i].slice,
		    p->slices[i].tag, p->slices[i].flag,
		    p->slices[i].start, p->slices[i].blks,
		    p->slices[i].end);
	}
	fprintf(stderr, "UFS file systems: %d\n", p->ufs_cnt);
	for (i = 0; i < p->ufs_cnt; i++) {
		char	*mntpt = p->ufss[i].mntpt;
		char	*special = p->ufss[i].special;

		fprintf(stderr, "\t%d %-20s %s %9ld %9ld %9ld\n",
		    p->ufss[i].nr, mntpt ? mntpt : nullstr,
		    special ? special : nullstr, p->ufss[i].blks,
		    p->ufss[i].free, p->ufss[i].avail);
	}
	fprintf(stderr, "swap partitions: %d\n", p->swap_cnt);
	for (i = 0; i < p->swap_cnt; i++) {
		char	*special = p->swaps[i].special;

		fprintf(stderr, "\t%d %s %9ld\n",  p->swaps[i].nr,
		    special ? special : nullstr,  p->swaps[i].blks);
	}
	fprintf(stderr, "SVM devices: %d\n", p->phys.md_cnt);
	for (i = 0; i < p->phys.md_cnt; i++) {
		char	*md =  p->phys.mds[i].md;
		char	*special =  p->phys.mds[i].special;

		fprintf(stderr, "\t%d %s %s\n", i,
		    md ? md : nullstr, special ? special : nullstr);
	}
	fprintf(stderr, "physical network interfaces: %d\n", p->net_cnt);
	for (i = 0; i < p->net_cnt; i++) {
		char	*name = p->nets[i].name;
		char	*ip =  p->nets[i].ipaddr;
		char	*mac =  p->nets[i].mac;

		fprintf(stderr, "\t%d %-8s %-13s %s\n", p->nets[i].nr,
		    name ? name : nullstr, ip ? ip : nullstr,
		    mac ? mac : nullstr);
	}
}

void
dump_virt(const config_t *v)
{
	resize_info_t	*resize;
	int	i;

	fprintf(stderr, "\ndomain: %s\n",
	    v->virt.name ? v->virt.name : nullstr);
	fprintf(stderr, "\nnodename: %s\n",
	    v->nodename ? v->nodename : nullstr);
	fprintf(stderr, "vsw: %s\nvds: %s\n",
	    v->virt.vsw ? v->virt.vsw : nullstr,
	    v->virt.vds ? v->virt.vds : nullstr);
	fprintf(stderr, "cpu: %d\nmemory: %ld MB\n", v->cpu,
	    v->memsize);
	fprintf(stderr, "hostid: %s\n",
	    v->hostid ?  v->hostid : "<auto>");
	fprintf(stderr, "disks: %d\n", v->disk_cnt);
	for (i = 0; i < v->disk_cnt; i++) {
		char	*name = v->disks[i].name;

		fprintf(stderr, "\t%d %s %9ld blks %s (%s %s, vdisk = %s)\n",
		    v->disks[i].nr, name ? name : nullstr, v->disks[i].blks,
		    v->disks[i].resized == 1 ? "resized" : "",
		    v->disks[i].pri_name ? v->disks[i].pri_name : nullstr,
		    v->disks[i].virt_name ? v->disks[i].virt_name : nullstr,
		    v->disks[i].vdisk ? v->disks[i].vdisk: nullstr);
	}
	fprintf(stderr, "slices: %d\n", v->slice_cnt);
	for (i = 0; i < v->slice_cnt; i++) {
		char	*name = v->slices[i].special;

		fprintf(stderr, "\t%d %s %d %d %02d %9ld %9ld %9ld ",
		    v->slices[i].nr,  name ? name : nullstr,
		    v->slices[i].slice,
		    v->slices[i].tag, v->slices[i].flag,
		    v->slices[i].start, v->slices[i].blks,
		    v->slices[i].end);
		switch (v->slices[i].type) {
			case SLICE_TYPE_UNUSED:
				fprintf(stderr, "(unused)\n");
				break;
			case SLICE_TYPE_UFS:
				fprintf(stderr, "(UFS)\n");
				break;
			case SLICE_TYPE_SWAP:
				fprintf(stderr, "(swap)\n");
				break;
		}
	}
	fprintf(stderr, "Resize list:\n");
	resize = v->virt.resize_list;
	while (resize != NULL) {
		fprintf(stderr, "\t%-20s %9ld\n", resize->mntpt,
		    resize->blks);
		resize = resize->next;
	}
	fprintf(stderr, "UFS file systems: %d\n", v->ufs_cnt);
	for (i = 0; i < v->ufs_cnt; i++) {
		char	*mntpt = v->ufss[i].mntpt;
		char	*special = v->ufss[i].special;

		fprintf(stderr, "\t%d %-20s %s %9ld %9ld %9ld (%s %s)\n",
		    v->ufss[i].nr, mntpt ? mntpt : nullstr,
		    special ? special : nullstr, v->ufss[i].blks,
		    v->ufss[i].free, v->ufss[i].avail,
		    v->ufss[i].pri_special ? v->ufss[i].pri_special : nullstr,
		    v->ufss[i].virt_special ?
		    v->ufss[i].virt_special : nullstr);
	}
	fprintf(stderr, "swap partitions: %d\n", v->swap_cnt);
	for (i = 0; i < v->swap_cnt; i++) {
		char	*special = v->swaps[i].special;

		fprintf(stderr, "\t%d %s %9ld\n",  v->swaps[i].nr,
		    special ? special : nullstr,  v->swaps[i].blks);
	}
	fprintf(stderr, "virtual network interfaces: %d\n", v->net_cnt);
	for (i = 0; i < v->net_cnt; i++) {
		char	*name = v->nets[i].name;
		char	*ip =  v->nets[i].ipaddr;
		char	*mac =  v->nets[i].mac;

		fprintf(stderr, "\t%d %-8s %-13s %s\n", v->nets[i].nr,
		    name ? name : nullstr, ip ? ip : nullstr,
		    mac ? mac : "auto");
	}
}
