/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <errno.h>
#include <sys/param.h>

#include "manifest.h"
#include "ldmp2v.h"
#include "util.h"

static int allocate_items(FILE *, config_t *);
static int parse_type(const char *, char *, config_t *);
static int parse_archive_method(char *, config_t *);
static int parse_cpu(char *, config_t *);
static int parse_disk(char *, config_t *);
static int parse_hostid(char *, config_t *);
static int parse_memory(char *, config_t *);
static int parse_net(char *, config_t *);
static int parse_nodename(char *, config_t *);
static int parse_os_release(char *, config_t *);
static int parse_svm(char *, config_t *);
static int parse_swap(char *, config_t *);
static int parse_ufs(char *, config_t *);
static int parse_vtoc(char *, config_t *);
static int parse_null(char *, config_t *);

static char *fieldsep = " \t\n";

static struct {
	const char *name;
	int (*func) (char *, config_t *);
	int required;
	int seen;
} cmdtab[] = {
	{ "archive_method",	parse_archive_method,	1,	0 },
	{ "cpu",		parse_cpu,		1,	0 },
	{ "disk",		parse_disk,		1,	0 },
	{ "hostid",		parse_hostid,		1,	0 },
	{ "ip",			parse_null,		1,	0 },
	{ "memory",		parse_memory,		1,	0 },
	{ "net",		parse_net,		1,	0 },
	{ "nodename",		parse_nodename,		1,	0 },
	{ "os_release",		parse_os_release,	1,	0 },
	{ "svm",		parse_svm,		0,	0 },
	{ "swap",		parse_swap,		0,	0 },
	{ "ufs",		parse_ufs,		1,	0 },
	{ "vtoc",		parse_vtoc,		1,	0 },
};

#define	NELEM(x)	(sizeof (x) / sizeof (x[0]))

static int
parse_type(const char *type, char *line, config_t *src)
{
	int	i;

	for (i = 0; i < NELEM(cmdtab); i++) {
		if (strcmp(type, cmdtab[i].name) == 0) {
			return (cmdtab[i].func(line, src));
		}
	}
	warn("Ignoring unknown resource type '%s'.\n", type);
	return (0);
}

/*
 * Read the manifest and fill the src_info structure with information
 * about the physical system.
 */
void
read_manifest(const char *dir, config_t *src)
{
	char	name[MAXPATHLEN];
	char	line[MAX_LINE_LENGTH];
	FILE	*f;
	int	err = 0;

	snprintf(name, sizeof (name), "%s/manifest", dir);
	if ((f = fopen(name, "r")) == NULL)
		die("Cannot open %s: %s\n", name, strerror(errno));

	allocate_items(f, src);

	while (fgets(line, sizeof (line), f) != NULL) {
		char	buf[16];
		char	*type;

		(void) strlcpy(buf, line, sizeof (buf));
		if ((type = strtok(buf, fieldsep)) == NULL)
			continue;

		if (parse_type(type, line, src) == -1)
			err++;
	}
	fclose(f);
	if (err != 0)
		die("Could not parse manifest %s\n", name);
}


/*
 * Read though the manifest and count the number of lines per configuration
 * item type and allocate the required amount of records of each type.
 */
static int
allocate_items(FILE *f, config_t *src)
{
	char	line[MAX_LINE_LENGTH];
	int	disk_cnt = 0;
	int	slice_cnt = 0;
	int	ufs_cnt = 0;
	int	net_cnt = 0;
	int	swap_cnt = 0;
	int	md_cnt = 0;
	int	i;

	while (fgets(line, sizeof (line), f) != NULL) {
		char	*type;
		int	found = 0;

		if ((type = strtok(line, fieldsep)) == NULL)
			continue;

		for (i = 0; i < NELEM(cmdtab) && found == 0; i++) {
			if (strcmp(type, cmdtab[i].name) == 0) {
				cmdtab[i].seen++;
				found = 1;
			}
		}
	}
	rewind(f);

	/*
	 * Verify that all required entries are present in the manifest.
	 */
	for (i = 0; i < NELEM(cmdtab); i++) {
		if (cmdtab[i].seen < cmdtab[i].required) {
			die("Invalid manifest: no '%s' entry found.\n",
			    cmdtab[i].name);
		}

		if (strcmp(cmdtab[i].name, "disk") == 0)
			disk_cnt = cmdtab[i].seen;
		else if (strcmp(cmdtab[i].name, "vtoc") == 0)
			slice_cnt = cmdtab[i].seen;
		else if (strcmp(cmdtab[i].name, "ufs") == 0)
			ufs_cnt = cmdtab[i].seen;
		else if (strcmp(cmdtab[i].name, "net") == 0)
			net_cnt = cmdtab[i].seen;
		else if (strcmp(cmdtab[i].name, "swap") == 0)
			swap_cnt = cmdtab[i].seen;
		else if (strcmp(cmdtab[i].name, "svm") == 0)
			md_cnt = cmdtab[i].seen;
	}

	src->disk_cnt = disk_cnt;
	src->disks = (disk_info_t *)
	    Malloc(disk_cnt * sizeof (disk_info_t));
	src->slice_cnt = slice_cnt;
	src->slices = (slice_info_t *)
	    Malloc(slice_cnt * sizeof (slice_info_t));
	src->ufs_cnt = ufs_cnt;
	src->ufss = (ufs_info_t *)
	    Malloc(ufs_cnt * sizeof (ufs_info_t));
	src->net_cnt = net_cnt;
	src->nets = (net_info_t *)
	    Malloc(net_cnt * sizeof (net_info_t));
	src->swap_cnt = swap_cnt;
	src->swaps = (swap_info_t *)
	    Malloc(swap_cnt * sizeof (swap_info_t));
	src->phys.md_cnt = md_cnt;
	src->phys.mds = (svm_info_t *)Malloc(md_cnt * sizeof (svm_info_t));

	return (0);
}

static int
parse_disk(char *line, config_t *src)
{
	static int	i = 0;
	char	*nr;
	char	*special;
	char	*blks;

	if (strtok(line, fieldsep) == NULL ||
	    (nr = strtok(NULL, fieldsep)) == NULL ||
	    (special = strtok(NULL, fieldsep)) == NULL ||
	    (blks = strtok(NULL, fieldsep)) == NULL)
		return (-1);

	bzero(&(src->disks[i]), sizeof (disk_info_t));
	src->disks[i].nr = atoi(nr);
	src->disks[i].name = Strdup(special);
	src->disks[i].blks = atol(blks);
	i++;
	return (0);
}

static int
parse_vtoc(char *line, config_t *src)
{
	static int	i = 0;
	char	*nr;
	char	*special;
	char	*slice;
	char	*tag;
	char	*start;
	char	*end;
	char	*blks;

	if (strtok(line, fieldsep) == NULL ||
	    (nr = strtok(NULL, fieldsep)) == NULL ||
	    (special = strtok(NULL, fieldsep)) == NULL ||
	    (slice = strtok(NULL, fieldsep)) == NULL ||
	    (tag = strtok(NULL, fieldsep)) == NULL ||
	    strtok(NULL, fieldsep) == NULL ||
	    (start = strtok(NULL, fieldsep)) == NULL ||
	    (blks = strtok(NULL, fieldsep)) == NULL ||
	    (end = strtok(NULL, fieldsep)) == NULL)
		return (-1);

	src->slices[i].nr = atoi(nr);
	src->slices[i].special = Strdup(special);
	src->slices[i].slice = atoi(slice);
	src->slices[i].tag = atoi(tag);
	src->slices[i].start = atol(start);
		src->slices[i].blks = atol(blks);
	src->slices[i].end = atol(end);
	i++;
	return (0);
}

static int
parse_swap(char *line, config_t *src)
{
	static int	i = 0;
	char	*nr;
	char	*special;
	char	*blks;

	if (strtok(line, fieldsep) == NULL ||
	    (nr = strtok(NULL, fieldsep)) == NULL ||
	    (special = strtok(NULL, fieldsep)) == NULL ||
	    (blks = strtok(NULL, fieldsep)) == NULL)
		return (-1);

	src->swaps[i].nr = atoi(nr);
	src->swaps[i].special = Strdup(special);
	src->swaps[i].blks = atol(blks);
	i++;
	return (0);
}

static int
parse_ufs(char *line, config_t *src)
{
	static int	i = 0;
	char	*nr;
	char	*mntpt;
	char	*special;
	char	*blks;
	char	*free;
	char	*avail;

	if (strtok(line, fieldsep) == NULL ||
	    (nr = strtok(NULL, fieldsep)) == NULL ||
	    (mntpt = strtok(NULL, fieldsep)) == NULL ||
	    (special = strtok(NULL, fieldsep)) == NULL ||
	    (blks = strtok(NULL, fieldsep)) == NULL ||
	    (free = strtok(NULL, fieldsep)) == NULL ||
	    (avail = strtok(NULL, fieldsep)) == NULL)
		return (-1);

	src->ufss[i].nr = atoi(nr);
	src->ufss[i].mntpt = Strdup(mntpt);
	src->ufss[i].special = Strdup(special);
	src->ufss[i].blks = atol(blks);
	src->ufss[i].free = atol(free);
	src->ufss[i].avail = atol(avail);
	i++;
	return (0);
}

static int
parse_svm(char *line, config_t *src)
{
	static int	i = 0;
	char	*md;
	char	*special;

	if (strtok(line, fieldsep) == NULL ||
	    (md = strtok(NULL, fieldsep)) == NULL ||
	    (special = strtok(NULL, fieldsep)) == NULL)
		return (-1);

	src->phys.mds[i].md = Strdup(md);
	src->phys.mds[i].special = Strdup(special);
	i++;
	return (0);
}

static int
parse_net(char *line, config_t *src)
{
	static int	i = 0;
	char	*nr;
	char	*name;
	char	*ipaddr;
	char	*mac;

	if (strtok(line, fieldsep) == NULL ||
	    (nr = strtok(NULL, fieldsep)) == NULL ||
	    (name = strtok(NULL, fieldsep)) == NULL ||
	    (ipaddr = strtok(NULL, fieldsep)) == NULL ||
	    (mac = strtok(NULL, fieldsep)) == NULL)
		return (-1);

	src->nets[i].nr = atoi(nr);
	src->nets[i].name = Strdup(name);
	src->nets[i].ipaddr = Strdup(ipaddr);
	src->nets[i].mac = Strdup(mac);
	i++;
	return (0);
}

static int
parse_archive_method(char *line, config_t *src)
{
	char	*method;

	if (strtok(line, fieldsep) == NULL ||
	    (method = strtok(NULL, fieldsep)) == NULL)
		return (-1);

	src->phys.method = Strdup(method);

	return (0);
}

static int
parse_cpu(char *line, config_t *src)
{
	char	*num;

	if (strtok(line, fieldsep) == NULL ||
	    (num = strtok(NULL, fieldsep)) == NULL)
		return (-1);

	src->cpu = atoi(num);
	return (0);
}

static int
parse_memory(char *line, config_t *src)
{
	char	*num;

	if (strtok(line, fieldsep) == NULL ||
	    (num = strtok(NULL, fieldsep)) == NULL)
		return (-1);

	src->memsize = atol(num);
	return (0);
}

static int
parse_hostid(char *line, config_t *src)
{
	char	*hostid;

	if (strtok(line, fieldsep) == NULL ||
	    (hostid = strtok(NULL, fieldsep)) == NULL)
		return (-1);

	src->hostid = Strdup(hostid);

	return (0);
}

static int
parse_nodename(char *line, config_t *src)
{
	char	*nodename;

	if (strtok(line, fieldsep) == NULL ||
	    (nodename = strtok(NULL, fieldsep)) == NULL)
		return (-1);

	src->nodename = Strdup(nodename);
	return (0);
}

static int
parse_os_release(char *line, config_t *src)
{
	char	*os_rel;

	if (strtok(line, fieldsep) == NULL ||
	    (os_rel = strtok(NULL, fieldsep)) == NULL)
		return (-1);

	src->phys.os_rel = Strdup(os_rel);

	return (0);
}

/*
 * Skip valid but uninteresting type.
 */
static int
parse_null(char *line, config_t *src)
{
	return (0);
}
