/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Guest OS image modification routines.
 *
 * We clean up the OS image so that it no longer refers to physical hardware
 * and add the proper device links for the virtual disks we created so the
 * standard Solaris Upgrade Install can upgrade the image.
 *
 * - remove all files from /dev and /devices,
 * - create new device links for the virtual disks,
 * - create a new /etc/path_to_inst appropriate for a Logical Domain,
 * - update /etc/vfstab with the new disk device names,
 * - unencapsulate SVM mirrors and remove SVM configuration (if applicable).
 * - remove all system specific Fault Manager data from /var/fm/fmd.
 */

#include <note.h>
#include <ftw.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <sys/param.h>

#include "ldmp2v.h"
#include "util.h"


static int mod_path_to_inst(config_t *);
static int mod_devices(config_t *);
static int clean_dev(const char *, const struct stat *, int, struct FTW *);
static int clean_devices(const char *, const struct stat *, int, struct FTW *);
static int mod_disk_links(config_t *);
static int mod_fmd(config_t *);
static int mod_interfaces(config_t *, config_t *);
static int sanity_check(config_t *, config_t *);
static int mod_svm(config_t *, config_t *, int);
static int mod_vfstab(config_t *i, int);
static int disable_mounts(config_t *);
static int in_swap_list(config_t *, const char *);
static int in_ufs_list(config_t *, const char *);
static int sed(const char *, const char *, const char *);

static const char PATH_TO_INST_BOOTSTRAP[] = "#path_to_inst_bootstrap_1";

static const char DSK_LINK_FMT[] = "\
../../devices/virtual-devices@100/channel-devices@200/disk@%d:%c";

static const char RDSK_LINK_FMT[] = "\
../../devices/virtual-devices@100/channel-devices@200/disk@%d:%c,raw";

static const char disabled[] = "## ldmp2v ##";

/*
 * Perform modifications to the guest OS image so a Solaris Upgrade Install
 * can be done. This entails creating the proper devlinks for the disks,
 * updating /etc/vfstab, /etc/path_to_inst, cleaning out /dev/ and /devices
 * and removing the SVM configuration if present.
 */
int
modify_guest_image(config_t *src, config_t *tgt, int is_altroot)
{
	msg(0, "Modifying guest OS image ...\n");

	if (tgt->virt.root == NULL || strcmp(tgt->virt.root, "/") == 0) {
		warn("Guest root is not set!\n");
		return (-1);
	}
	if (sanity_check(src, tgt) == -1 ||
	    mod_svm(src, tgt, is_altroot) == -1 ||
	    mod_vfstab(tgt, is_altroot) == -1 ||
	    mod_interfaces(src, tgt) == -1 ||
	    mod_devices(tgt) == -1 ||
	    mod_disk_links(tgt) == -1 ||
	    mod_fmd(tgt) == -1 ||
	    mod_path_to_inst(tgt) == -1)
		return (-1);
	return (0);
}

/*
 * Check that a file exists in the guest OS image.
 */
static int
file_exists(config_t *tgt, const char *name)
{
	char	 buf[MAXPATHLEN];
	struct stat st;

	(void) snprintf(buf, sizeof (buf), "%s%s", tgt->virt.root, name);
	if (stat(buf, &st) != 0) {
		warn("File %s does not exist in the guest OS image.\n", name);
		return (B_FALSE);
	}
	return (B_TRUE);
}

/*
 * Perform a sanity check of the unmodified OS image by checking for some
 * key files.
 */
static int
sanity_check(config_t *src, config_t *tgt)
{
	if (!file_exists(tgt, "/etc/system") ||
	    !file_exists(tgt, "/etc/vfstab") ||
	    !file_exists(tgt, "/var/sadm/system/admin/INST_RELEASE"))
		return (-1);

	if (strcmp(src->phys.os_rel, "5.8") != 0)
		if (!file_exists(tgt, "/kernel/drv/md.conf"))
			return (-1);
	return (0);
}

/*
 * Replace the existing /etc/path_to_inst file with a copy appropriate
 * for a guest domain.
 */
static int
mod_path_to_inst(config_t *tgt)
{
	char	buf[MAXPATHLEN];
	FILE	*f;

	msg(1, "Modifying /etc/path_to_inst ...\n");

	(void) snprintf(buf, sizeof (buf), "%s/%s", tgt->virt.root,
	    "etc/path_to_inst");
	if ((f = fopen(buf, "w")) == NULL) {
		warn("Cannot open %s", buf);
		return (-1);
	}
	(void) fprintf(f, PATH_TO_INST_BOOTSTRAP);
	fclose(f);
	return (0);
}

/*
 * Return a pointer to the n-th '/' from the end.
 */
static const char *
tail(const char *name, int n)
{
	const char	*p;
	int	seen = 0;

	if (n > 0) {
		p = name + strlen(name) - 1;
		while (p > name) {
			if (*p == '/' && ++seen == n)
				return (p);
			p--;
		}
	}
	return (NULL);
}

/*
 * Remove everything below /var/fm/fmd, preserving only the ckpt, rsrc and xprt
 * directories and the ckpt/ckpt and rsrc/rsrc files if present.
 */
static int
clean_fmd(const char *name, const struct stat *st, int flags,
    struct FTW *ftw)
{
	const char	*p;

	NOTE(ARGUNUSED(flags))

	switch (ftw->level) {
	case 0:			/* <guest root>/var/fm/fmd */
		return (0);
	case 1:
		if (S_ISDIR(st->st_mode) && (p = tail(name, ftw->level)) &&
		    (strcmp(p, "/ckpt") == 0 ||
		    strcmp(p, "/rsrc") == 0 ||
		    strcmp(p, "/xprt") == 0))
			return (0);
		break;
	case 2:
		if (S_ISREG(st->st_mode) && (p = tail(name, ftw->level)) &&
		    (strcmp(p, "/ckpt/ckpt") == 0 ||
		    strcmp(p, "/rsrc/rsrc") == 0))
			return (0);
		break;
	}

	if (remove(name) != 0)
		warn("Cannot remove %s", name);
	return (0);
}

/*
 * Remove system specific Fault Manager data so we don't propagate events from
 * the original system to the guest.
 */
static int
mod_fmd(config_t *tgt)
{
	struct stat st;
	char	buf[MAXPATHLEN];
	int	depth = 4;
	int	flags = FTW_PHYS | FTW_DEPTH | FTW_MOUNT;

	(void) snprintf(buf, sizeof (buf), "%s/var/fm/fmd", tgt->virt.root);
	if (stat(buf, &st) != 0 || !S_ISDIR(st.st_mode))
		return (0);	/* No FMA on the source system. */

	msg(1, "Cleaning /var/fm/fmd ...\n");
	if (nftw(buf, clean_fmd, depth, flags) != 0) {
		warn("Clean /var/fm/fmd failed.\n");
		return (-1);
	}
	return (0);
}

/*
 * Callback to delete everything below the guest domain's /devices.
 */
static int
clean_devices(const char *name, const struct stat *st, int flags,
    struct FTW *ftw)
{
	NOTE(ARGUNUSED(st, flags))

	if (ftw->level == 0)		/* <guest root>/devices */
		return (0);

	if (remove(name) != 0)
		warn("Cannot remove %s", name);

	return (0);
}

/*
 * Callback to delete everything below the guest domain's /dev. There is no
 * point in preserving dev/dsk and dev/rdsk here: Flash archives do not contain
 * these directories so they need to be recreated anyway in mod_disk_links().
 */
static int
clean_dev(const char *name, const struct stat *st, int flags,
    struct FTW *ftw)
{
	NOTE(ARGUNUSED(st, flags))

	if (ftw->level == 0)		/* <guest root>/dev */
		return (0);

	if (remove(name) != 0)
		warn("Cannot remove %s", name);

	return (0);
}

/*
 * Clean out the /dev and /devices directory of the guest domain
 */
static int
mod_devices(config_t *tgt)
{
	char	buf[MAXPATHLEN];
	int	depth = 4;
	int	flags = FTW_PHYS | FTW_DEPTH | FTW_MOUNT;

	msg(1, "Modifying /devices ...\n");
	(void) snprintf(buf, sizeof (buf), "%s/devices", tgt->virt.root);
	if (nftw(buf, clean_devices, depth, flags) != 0) {
		warn("Clean /devices failed.\n");
		return (-1);
	}

	msg(1, "Modifying /dev ...\n");
	(void) snprintf(buf, sizeof (buf), "%s/dev", tgt->virt.root);
	if (nftw(buf, clean_dev, depth, flags) != 0) {
		warn("Clean /dev failed.\n");
		return (-1);
	}
	return (0);
}

#define	DIRMODE	\
    S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH

/*
 * Recreate new disk device links that point to the created vdisks.
 */
static int
mod_disk_links(config_t *tgt)
{
	char	buf[MAXPATHLEN];
	int	i;

	msg(1, "Creating disk device links ...\n");

	(void) snprintf(buf, sizeof (buf), "%s/dev/dsk", tgt->virt.root);
	if (mkdir(buf, DIRMODE) != 0) {
		warn("Cannot create directory %s", buf);
		return (-1);
	}
	(void) snprintf(buf, sizeof (buf), "%s/dev/rdsk", tgt->virt.root);
	if (mkdir(buf, DIRMODE) != 0) {
		warn("Cannot create directory %s", buf);
		return (-1);
	}
	for (i = 0; i < tgt->disk_cnt; i++) {
		char	disk[MAXPATHLEN];
		char	link[MAXPATHLEN];
		int	slice;

		for (slice = 0; slice < 8; slice++) {
			(void) snprintf(disk, sizeof (disk),
			    "%s/dev/dsk/c0d%ds%d", tgt->virt.root, i, slice);
			(void) snprintf(link, sizeof (link), DSK_LINK_FMT,
			    i, 'a' + slice);
			if (symlink(link, disk) != 0) {
				warn("Cannot create symlink %s\n", disk);
				return (-1);
			}
			(void) snprintf(disk, sizeof (disk),
			    "%s/dev/rdsk/c0d%ds%d", tgt->virt.root, i, slice);
			(void) snprintf(link, sizeof (link), RDSK_LINK_FMT,
			    i, 'a' + slice);
			if (symlink(link, disk) != 0) {
				warn("Cannot create symlink %s\n", disk);
				return (-1);
			}
		}
	}
	return (0);
}

/*
 * Rename /etc/hostname.<ifname> files to /etc/hostname.vnet<n>.
 */
static int
mod_interfaces(config_t *src, config_t *tgt)
{
	int	i;

	msg(1, "Modifying network interfaces ...\n");

	for (i = 0; i < tgt->net_cnt; i++) {
		char	old[MAXPATHLEN];
		char	new[MAXPATHLEN];

		(void) snprintf(old, sizeof (old), "%s/etc/hostname.%s",
		    tgt->virt.root, src->nets[i].name);
		(void) snprintf(new, sizeof (new), "%s/etc/hostname.%s",
		    tgt->virt.root, tgt->nets[i].name);
		if (rename(old, new) != 0) {
			warn("Cannot rename %s to %s\n", old, new);
			return (-1);
		}
	}
	return (0);
}

/*
 * Remove SVM configuration data such as rootdev and the md database location
 * from the system.
 */
static int
mod_svm(config_t *src, config_t *tgt, int is_altroot)
{
	char	infile[MAXPATHLEN];
	char	outfile[MAXPATHLEN];
	char	cmds[] = "/var/run/ldmp2v_XXXXXX";
	FILE	*f;
	int	fd;
	int	i;

	msg(0, "Modifying SVM configuration ...\n");

	if ((fd = mkstemp(cmds)) == -1 ||
	    (f = fdopen(fd, "w")) == NULL) {
		warn("Cannot create temporary file");
		return (-1);
	}
	(void) snprintf(infile, sizeof (infile), "%s/etc/system",
	    tgt->virt.root);
	(void) snprintf(outfile, sizeof (outfile), "%s/etc/system.p2v",
	    tgt->virt.root);

	(void) fprintf(f,
	    "/^* Begin MDD root info /,/^* End MDD root info /d\n");

	/*
	 * Solaris 8 keeps the md database location in /etc/system.
	 */
	if (strcmp(src->phys.os_rel, "5.8") == 0) {
		(void) fprintf(f,
		    "/^* Begin MDD database /,/^* End MDD database /d\n");
	}
	fclose(f);
	if (sed(cmds, infile, outfile) == -1)
		return (-1);

	/*
	 * Solaris 9 and above keep it in /kernel/drv/md.conf.
	 */
	if (strcmp(src->phys.os_rel, "5.8") != 0) {
		if ((fd = mkstemp(cmds)) == -1 ||
		    (f = fdopen(fd, "w")) == NULL) {
			warn("Cannot create temporary file");
			return (-1);
		}
		(void) snprintf(infile, sizeof (infile),
		    "%s/kernel/drv/md.conf", tgt->virt.root);
		(void) snprintf(outfile, sizeof (outfile),
		    "%s/kernel/drv/md.conf.p2v", tgt->virt.root);
		(void) fprintf(f,
		    "/^# Begin MDD database /,/^# End MDD database /d\n");
		fclose(f);
		if (sed(cmds, infile, outfile) == -1)
			return (-1);
	}

	/*
	 * Update /etc/vfstab if we have file systems on SVM devices. The SVM
	 * metadevice names are replaced by the underlying slices as they
	 * were on the original, physical system. mod_vfstab() will translate
	 * these to the new names later on. If the -R option was specified
	 * /etc/vfstab is completely regenerated from the currently mounted
	 * file systems, so there is no need to update vfstab here.
	 */
	if (src->phys.md_cnt == 0)
		return (0);	/* no file systems on SVM, we're done. */
	if (is_altroot == B_TRUE)
		return (0);	/* -R was specified, we're done */

	(void) snprintf(infile, sizeof (infile), "%s/etc/vfstab",
	    tgt->virt.root);
	(void) snprintf(outfile, sizeof (outfile), "%s/etc/vfstab.p2v",
	    tgt->virt.root);
	if ((fd = mkstemp(cmds)) == -1 ||
	    (f = fdopen(fd, "w")) == NULL) {
		warn("Cannot create temporary file");
		return (-1);
	}

	for (i = 0; i < src->phys.md_cnt; i++) {
		char	*md;
		char	*disk;

		md = strrchr(src->phys.mds[i].md, 'd');
		disk = strchr(src->phys.mds[i].special, 'c');
		(void) fprintf(f, "s[/dev/md/dsk/%s[/dev/dsk/%s[\n",
		    md, disk);
		(void) fprintf(f, "s[/dev/md/rdsk/%s[/dev/rdsk/%s[\n",
		    md, disk);
	}
	fclose(f);
	if (sed(cmds, infile, outfile) == -1)
		return (-1);
	return (0);
}

static int
sed(const char *cmds, const char *infile, const char *outfile)
{
	if (exec_sed(cmds, infile, outfile) == -1) {
		(void) remove(cmds);
		return (-1);
	}
	(void) remove(cmds);
	if (rename(outfile, infile) != 0) {
		warn("Cannot rename %s to %s", outfile, infile);
		return (-1);
	}
	return (0);
}

/*
 * Rewrite /etc/vfstab so it references the new disk device names of the guest
 * domain.
 */
static int
mod_vfstab(config_t *tgt, int is_altroot)
{
	char	vfstab[MAXPATHLEN];
	char	new_vfstab[MAXPATHLEN];
	int	i;

	msg(1, "Modifying /etc/vfstab ...\n");
	(void) snprintf(vfstab, sizeof (vfstab), "%s/etc/vfstab",
	    tgt->virt.root);
	(void) snprintf(new_vfstab, sizeof (new_vfstab),
	    "%s/etc/vfstab.p2v", tgt->virt.root);

	if (is_altroot) {
		char	line[MAX_LINE_LENGTH];
		FILE	*oldf;
		FILE	*newf;
		int	i;

		if ((oldf = fopen(vfstab, "r")) == NULL) {
			warn("Cannot open %s", vfstab);
			return (-1);
		}
		if ((newf = fopen(new_vfstab, "w")) == NULL) {
			warn("Cannot open %s", new_vfstab);
			return (-1);
		}
		/*
		 * Comment out all local file systems and add entries for
		 * UFS file systems currently mounted below virt.root.
		 */
		while (fgets(line, sizeof (line), oldf) != NULL) {
			char	*special;
			char	*s;

			s = Strdup(line);
			if ((special = strtok(s, " \t\n")) == NULL) {
				fprintf(newf, "%s", line);
			} else if (strstr(special, "/dev/") == special) {
				fprintf(newf, "%s %s", disabled, line);
			} else {
				fprintf(newf, "%s", line);
			}
			free(s);
		}
		fclose(oldf);

		for (i = 0; i < tgt->ufs_cnt; i++) {
			char	*dsk;
			int	bootfs;

			dsk = strstr(tgt->ufss[i].virt_special, "dsk/");
			if (dsk == NULL) {
				warn("No special device for mountpoint %s\n",
				    tgt->ufss[i].mntpt);
				return (-1);
			}
			if (strcmp(tgt->ufss[i].mntpt, "/") == 0 ||
			    strcmp(tgt->ufss[i].mntpt, "/usr") == 0 ||
			    strcmp(tgt->ufss[i].mntpt, "/var") == 0)
				bootfs = B_TRUE;
			else
				bootfs = B_FALSE;
			(void) fprintf(newf,
			    "%s\t/dev/r%s\t%s\tufs\t%d\t%s\t-\n",
			    tgt->ufss[i].virt_special, dsk, tgt->ufss[i].mntpt,
			    bootfs == B_TRUE ? 1 : 2,
			    bootfs == B_TRUE ? "no" : "yes");
		}
		fclose(newf);
	} else {
		/*
		 * Replace the original disk names by the new names on the
		 * guest domain.
		 */
		char	tmpl[] = "/var/run/ldmp2v_XXXXXX";
		FILE	*f;
		int	fd;
		int	ret;

		if (disable_mounts(tgt) == -1)
			return (-1);
		if ((fd = mkstemp(tmpl)) == -1 ||
		    (f = fdopen(fd, "w")) == NULL) {
			warn("Cannot create temporary file");
			return (-1);
		}
		for (i = 0; i < tgt->disk_cnt; i++) {
			char	*disk;
			char	*virt_disk;

			disk = strchr(tgt->disks[i].name, 'c');
			virt_disk = strchr(tgt->disks[i].virt_name, 'c');
			(void) fprintf(f, "s[/dev/dsk/%ss[/dev/dsk/%ss[\n",
			    disk, virt_disk);
			(void) fprintf(f, "s[/dev/rdsk/%ss[/dev/rdsk/%ss[\n",
			    disk, virt_disk);
		}
		fclose(f);
		ret = exec_sed(tmpl, vfstab, new_vfstab);
		(void) remove(tmpl);
		if (ret != 0)
			return (-1);
	}
	if (rename(new_vfstab, vfstab) != 0) {
		warn("Cannot rename %s to %s", new_vfstab, vfstab);
		return (-1);
	}
	return (0);
}

/*
 * Update /etc/vfstab so that only entries for swap devcies and UFS we
 * migrated are active. Comment out all disk based file systems not on UFS.
 */
static int
disable_mounts(config_t *tgt)
{
	char	vfstab[MAXPATHLEN];
	char	new_vfstab[MAXPATHLEN];
	char	line[MAX_LINE_LENGTH];
	FILE	*oldf;
	FILE	*newf;

	(void) snprintf(vfstab, sizeof (vfstab), "%s/etc/vfstab",
	    tgt->virt.root);
	if ((oldf = fopen(vfstab, "r")) == NULL) {
		warn("Cannot open %s", vfstab);
		return (1);
	}
	(void) snprintf(new_vfstab, sizeof (new_vfstab), "%s/etc/vfstab.p2v",
	    tgt->virt.root);
	if ((newf = fopen(new_vfstab, "w")) == NULL) {
		warn("Cannot open %s", new_vfstab);
		return (1);
	}
	while (fgets(line, sizeof (line), oldf) != NULL) {
		char	*special;
		char	*s;

		s = Strdup(line);
		if ((special = strtok(s, " \t\n")) == NULL) {
			fprintf(newf, "%s", line);
		} else if (*special == '#') {
			fprintf(newf, "%s", line);
		} else if (strstr(special, "/dev/") == special) {
			char	*mntpt;
			char	*fstype;

			if (strtok(NULL, " \t\n") == NULL ||
			    (mntpt = strtok(NULL, " \t\n")) == NULL ||
			    (fstype = strtok(NULL, " \t\n")) == NULL) {
				fprintf(newf, "%s", line);
			} else if (strcmp(fstype, "swap") == 0) {
				if (in_swap_list(tgt, special))
					fprintf(newf, "%s", line);
				else
					fprintf(newf, "%s %s", disabled, line);
			} else if (strcmp(fstype, "ufs") == 0) {
				if (in_ufs_list(tgt, mntpt))
					fprintf(newf, "%s", line);
				else
					fprintf(newf, "%s %s", disabled, line);
			} else {	/* other fstype */
				fprintf(newf, "%s %s", disabled, line);
			}
		} else {
			fprintf(newf, "%s", line);
		}
		free(s);
	}
	fclose(newf);
	fclose(oldf);
	if (rename(new_vfstab, vfstab) != 0) {
		warn("Cannot rename %s to %s", new_vfstab, vfstab);
		return (-1);
	}
	return (0);
}

static int
in_swap_list(config_t *tgt, const char *dev)
{
	int	i;

	for (i = 0; i < tgt->swap_cnt; i++) {
		if (strcmp(dev, tgt->swaps[i].special) == 0)
			return (B_TRUE);
	}
	return (B_FALSE);
}

static int
in_ufs_list(config_t *tgt, const char *mntpt)
{
	int	i;

	for (i = 0; i < tgt->ufs_cnt; i++) {
		if (strcmp(mntpt, tgt->ufss[i].mntpt) == 0)
			return (B_TRUE);
	}
	return (B_FALSE);
}
