/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <strings.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include "util.h"

#define	MOUNT_BIN	"/usr/sbin/mount"
#define	MOUNT_CMD	"mount"
#define	PRTVTOC_BIN	"/usr/sbin/prtvtoc"
#define	PRTVTOC_CMD	"prtvtoc"
#define	SED_BIN		"/usr/bin/sed"
#define	SED_CMD		"sed"

int verbose = 0;
int devnull;
int logfile;

void
msg(int level, const char *format, ...)
{
	va_list	alist;

	if (verbose >= level) {
		va_start(alist, format);
		(void) vfprintf(stderr, format, alist);
	}
}

void
warn(const char *format, ...)
{
	int	err;
	va_list	alist;

	err = errno;
	va_start(alist, format);
	(void) vfprintf(stderr, format, alist);
	va_end(alist);
	if (strchr(format, '\n') == NULL)
		(void) fprintf(stderr, ": %s\n", strerror(err));
}

void
die(const char *format, ...)
{
	int	err;
	va_list	alist;

	err = errno;
	va_start(alist, format);
	(void) vfprintf(stderr, format, alist);
	va_end(alist);
	if (strchr(format, '\n') == NULL)
		(void) fprintf(stderr, ": %s\n", strerror(err));
	exit(1);
}

int
exec_cmd(int quiet, const char *path, ...)
{
	va_list	args;
	pid_t	pid;
	int	status;

	if ((pid = fork()) == -1) {
		warn("Cannot fork for %s", path);
		return (-1);
	}

	if (pid == 0) {
		if (quiet) {
			(void) dup2(devnull, STDOUT_FILENO);
			(void) dup2(devnull, STDERR_FILENO);
		}
		va_start(args, path);
		execv(path, (void *)args);
		va_end(args);
		_exit(1);
	}
	(void) waitpid(pid, &status, 0);
	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0)
		return (-1);
	return (0);
}

void *
Malloc(size_t size)
{
	void	*p;

	if ((p = malloc(size)) == NULL)
		die("Malloc: out of memory.\n");
	return (p);
}

char *
Strdup(const char *str)
{
	char	*p;

	if ((p = strdup(str)) == NULL)
		die("Strdup: out of memory.\n");
	return (p);
}

int
mount_fs(const char *mntpt, const char *special)
{
	return (exec_cmd(0, MOUNT_BIN, MOUNT_CMD, special, mntpt, NULL));
}

/*
 * Create a directory and all intermediate directories.
 */
int
make_dir(const char *path)
{
	return (exec_cmd(1, MKDIR_BIN, MKDIR_CMD, "-p", path, NULL));
}

/*
 * Exec a copy of /usr/sbin/prtvtoc to get the number of sectors per cylinder
 * for a disk.
 */
int
get_sect_per_cyl(const char *raw)
{
	char	line[512];
	pid_t	pid;
	int	status;
	int	fd[2];
	FILE	*f;
	int	num = -1;

	if (pipe(fd) == -1 || (pid = fork()) == -1) {
		warn("Cannot for fork %s", PRTVTOC_BIN);
		return (-1);
	}

	if (pid == 0) {
		(void) dup2(fd[1], STDOUT_FILENO);
		close(fd[0]);
		(void) dup2(devnull, STDERR_FILENO);
		(void) execl(PRTVTOC_BIN, PRTVTOC_CMD, raw, NULL);
		_exit(0);
	}

	close(fd[1]);
	if ((f = fdopen(fd[0], "r")) == NULL) {
		warn("fdopen() failed.\n");
		return (-1);
	}
	while (fgets(line, sizeof (line), f) != NULL) {
		if (strstr(line, "sectors/cylinder") != NULL) {
			char	*s;

			if (strtok(line, " \t\n") == NULL ||
			    (s = strtok(NULL, " \t\n")) == NULL)
				continue;
			num = atoi(s);
			break;
		}
	}
	fclose(f);
	(void) waitpid(pid, &status, 0);
	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0)
		return (-1);
	return (num);
}

/*
 * Exec a copy of sed(1) to run the script 'cmdfile' against 'input', writing
 * the results to 'output'.
 */
int
exec_sed(const char *cmdfile, const char *input, const char *output)
{
	pid_t	pid;
	int	status;
	int	fd;

	if ((pid = fork()) == -1)
		return (-1);

	if (pid == 0) {
		if ((fd = open(output, O_RDWR | O_CREAT,
		    S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) != -1) {
			(void) dup2(fd, STDOUT_FILENO);
			execl(SED_BIN, SED_CMD, "-f", cmdfile, input, NULL);
		}
		_exit(1);
	}
	(void) waitpid(pid, &status, 0);
	if (!WIFEXITED(status) || WEXITSTATUS(status) != 0)
		return (-1);
	return (0);
}
