/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <string.h>

#include "ldom.h"
#include "basics.h"
#include "ds.h"
#include "ds_impl.h"
#include "ds_registry.h"
#include "zprintf.h"
#include "fatal.h"
#include "allocate.h"
#include "zpri.h"
#include "mdesc_impl.h"
#include "ldomconfigure.h"
#include "pcie.h"
#include "database.h"

#define	PCI_BUS_NAME	"pci"

/*
 * MSIs are allocated in blocks with a fixed number of MSIs in each block.
 * The following macros convert from a number of MSIs to the corresponding
 * number of MSI blocks and back.
 */
#define	MSIS_TO_MSIBS(n)	((n) / MSIB_BSIZE)
#define	MSIBS_TO_MSIS(n)	((n) * MSIB_BSIZE)

#define	MSIBS_PER_DEV(rp)	((rp)->req_msi_per_dev / MSIB_BSIZE)

#define	NUM_MSI_DEVS(bio, rp)						\
	(((bio)->virtual_pcie_bus) ?  CHAIN_LENGTH((bio)->dev_list) :	\
	(CHAIN_LENGTH((bio)->dev_list) + (rp)->resv_msi_slots))

#define	NUM_EQ_DEVS(bio, rp)						\
	(((bio)->virtual_pcie_bus) ? CHAIN_LENGTH((bio)->dev_list) :	\
	(CHAIN_LENGTH((bio)->dev_list) + (rp)->resv_eq_slots))

/*
 * If the argument is already a power of 2, just return the argument.
 * Otherwise, return the next higher power of 2.
 */
#define	NEXT_POWER2(i)	(1 << highbit(i - 1))

static void process_devtree_from_guest(ldom_t *ldp, md_t *mdp);
static void process_devnode(md_t *mdp, md_node_t *devnodep,
    pri_pcie_dev_t *pdev, char *path);
static void populate_subdevice_list(ldom_t *ldp, md_t *mdp,
    md_node_t *rcnodep, pri_iodevice_t *pio);
static void pcie_prune_branches(md_t *mdp, md_node_t *nodep,
    bound_iodevice_t *bio);
static void free_msis(bound_iodevice_t *bio,
    pcie_bus_res_mgmt_t *res_mgmtp);
static void free_eqs(bound_iodevice_t *bio,
    pcie_bus_res_mgmt_t *res_mgmtp);

/*
 * Dynamically generated format string for some print/list commands.
 */
static char listio_format[256];

static char *slot_status[] = { "UNK", "EMP", "OCC"};

/*
 * Return true if HV MD format supports Static Direct IO
 */
bool_t
pcie_is_dio_hvmd_supported(void)
{
	return ((zeus.hvmd_vers_major > HVMD_DIO_VERSION_MAJOR_NUMBER) ||
	    ((zeus.hvmd_vers_major == HVMD_DIO_VERSION_MAJOR_NUMBER) &&
	    (zeus.hvmd_vers_minor >= HVMD_DIO_VERSION_MINOR_NUMBER)));
}

static char *
slot_status_string(pri_pcie_dev_t *pdev)
{
	ASSERT(pdev->status < (sizeof (slot_status) / sizeof (char *)));
	return (slot_status[pdev->status]);
}

static void
display_subdevs(char *fmt, pri_pcie_dev_t *pdev, response_t *resp)
{
	pri_pcie_subdev_t *sdev;

	ASSERT(MUTEX_HELD(&pdev->subdev_lock));

	CHAIN_ITER(pdev->sub_devs, sdev, link) {
		rprintf(resp, fmt, sdev->subdev_path);
	}
}

/*
 * Return highest 1-bit set.  (Copied from sparc/os/bitmap_arch.c)
 */
static int
highbit(ulong_t i)
{
	register int h = 1;

	if (i == 0)
		return (0);
#ifdef _LP64
	if (i & 0xffffffff00000000ul) {
		h += 32; i >>= 32;
	}
#endif
	if (i & 0xffff0000) {
		h += 16; i >>= 16;
	}
	if (i & 0xff00) {
		h += 8; i >>= 8;
	}
	if (i & 0xf0) {
		h += 4; i >>= 4;
	}
	if (i & 0xc) {
		h += 2; i >>= 2;
	}
	if (i & 0x2) {
		h += 1;
	}
	return (h);
}


void
pcie_display_io(char *fmt, response_t *resp)
{
	ldom_t *ldp;
	bound_iodevice_t *bio;
	pri_iodevice_t *pio;
	char name[MAXPATHLEN];

	/*
	 * Walk the pri structures because they are static, and we
	 * always want the devices to appear in the same order in
	 * the output.
	 */
	CHAIN_ITER(pri.iodevice_list, pio, link) {
		if (pio->cli_name != NULL)
			(void) strlcpy(name, pio->cli_name, sizeof (name));
		else
			name[0] = '\0';

		if (CHAIN_LENGTH(pio->bound_ios) == 0) {
			rprintf(resp, fmt, pio->name, name, "");
			continue;
		}
		/* Device is bound.  Search for owning ldom. */
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			if (ldp->state < BOUND)
				continue;
			bio = pcie_biodevice_exists(ldp, pio->cfg_handle);
			if ((bio != NULL) && !bio->virtual_pcie_bus)
				break;
		}
		ASSERT(ldp != NULL);
		rprintf(resp, fmt, pio->name, name, ldp->namep);
	}
}

static void
display_one_device(char *fmt,
    char *sub_fmt, pri_pcie_dev_t *pdev, char *name, response_t *resp)
{
	/* Note: token for unoccupied slot is '-' per CLI spec. */
	mutex_lock(&pdev->subdev_lock);
	rprintf(resp, fmt, pdev->name, pdev->cli_name,
	    slot_status_string(pdev), pdev->status == SL_OCCUPIED ? name : "-");
	if (sub_fmt != NULL)
		display_subdevs(sub_fmt, pdev, resp);
	mutex_unlock(&pdev->subdev_lock);
}

static void
refresh_device_status(void)
{
	ldom_t *primary;

	if (!zeus.sdio_enabled)
		return;

	/* We only refresh the device info for the control domain. */
	primary = find_ldom_by_name(SOD_CONTROL_DOMAIN);
	if (primary == NULL)
		fatal("Couldn't find %s", SOD_CONTROL_DOMAIN);

	pcie_get_devinfo(primary);
}

bool_t
pcie_is_device_removable(ldom_t *ldp, pri_pcie_dev_t *pdev)
{
	if (!zeus.sdio_enabled)
		return (false);
#if EMULATE_MDFILES
	return (true);
#else

	/*
	 * Return true without checking the status
	 * for domains other than Control Domain.
	 * This is because the device was assigned originally
	 * based on its status. This may also help unknown
	 * scenarios where a customer has changed the h/w config
	 * and wants to fix the guest domain configs.
	 */
	if (!IS_CONTROL_DOMAIN(ldp))
		return (true);
	/*
	 * Refresh slot status for the control domain,
	 * as the state can change due to hotplug
	 * operations in the Control domain.
	 */
	refresh_device_status();
	return (pdev->status == SL_OCCUPIED);
#endif
}

static ldom_t *
pcie_find_owning_domain(pri_pcie_dev_t *pdev)
{
	ldom_t	*ldp;
	bound_iodevice_t	*bio;
	bound_pcie_dev_t	*bdev;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state < BOUND)
			continue;
		CHAIN_ITER(ldp->bound.iodevices, bio, ldom) {
			CHAIN_ITER(bio->dev_list, bdev, ldom) {
				if (pdev == bdev->phys_pcie_dev)
					return (ldp);
			}
		}
	}

	return (NULL);
}

void
pcie_display_devs(char *fmt, char *sub_fmt, response_t *resp)
{
	ldom_t *ldp;
	pri_iodevice_t *pio;
	pri_pcie_dev_t *pdev;

	refresh_device_status();

	/*
	 * Walk the pri structures because they are static, and we
	 * always want the devices to appear in the same order in
	 * the output.
	 */
	CHAIN_ITER(pri.iodevice_list, pio, link) {
		CHAIN_ITER(pio->dev_list, pdev, link) {
			if (CHAIN_LENGTH(pdev->bound_devs) == 0) {
				display_one_device(fmt,
				    sub_fmt, pdev, "", resp);
				continue;
			}

			/* Device is bound.  Search for owning ldom. */
			ldp = pcie_find_owning_domain(pdev);

			ASSERT(ldp != NULL);
			display_one_device(fmt,
			    sub_fmt, pdev, ldp->namep, resp);
		}
	}
}

/*
 * Initialize the heuristics required for resource allocation.
 * - Root complex owner needs resources (msis, eqs) even when all
 *   slots are removed/loaned out from the Root Domain. To deal
 *   with this case, treat Root Domain as if it has one additional
 *   slot.
 *
 * - MSIs have different restrictions:
 *	- Only one contiguous range can be allocated to a domain.
 *	- Absolute MSI numbers are important as they are provided
 *	  to the PCIe devices for programming their registers.
 * 	- A px driver can allocate msis only in powers of 2 to a device.
 *	  This restriction is only for MSI, but not for MSI-Xs.
 * 	- Allocate in as big chunks as possible so that a px
 *	  driver can allocate in powers of 2 to devices.
 *
 * - EQs have have no restrictions. HV virtualizes them, so
 *        they need to be simply divided for allocation.
 */
void
pcie_init_res_heuristics(void)
{
	pri_iodevice_t *pio;
	pcie_bus_res_mgmt_t *res_mgmtp;
	pcie_bus_res_t *busresp;
	int num_slots;
	int adj_num_slots;
	int num_msi_buckets;

	CHAIN_ITER(pri.iodevice_list, pio, link) {
		if (!pio->is_pciex)
			continue;
		num_slots = CHAIN_LENGTH(pio->dev_list);

		/*
		 * The primary always needs its own MSI resources, so
		 * handle this by fabricating an additional slot.
		 */
		adj_num_slots = num_slots + 1;

		/*
		 * Compute the number of buckets across which the total
		 * number of MSIs will be equally divided.  Round up to
		 * next higher power of 2, if necessary.
		 */
		num_msi_buckets = NEXT_POWER2(adj_num_slots);

		res_mgmtp = &pio->res_mgmt;
		busresp = &pio->busres;

		/*
		 * Extra slots resulting from rounding up to power of 2
		 * belong to primary.
		 */
		res_mgmtp->resv_msi_slots = num_msi_buckets - num_slots;

		/*
		 * Save reserved slot info for EQs, it's assumed
		 * to be assigned to root complex owner for EQ allocation.
		 */
		res_mgmtp->resv_eq_slots = 1;

		res_mgmtp->req_msi_per_dev =
		    busresp->num_msis / num_msi_buckets;
		res_mgmtp->req_eqs_per_dev = busresp->num_eqs / (num_slots + 1);
		DBG(zprintf("%s: MSIs per dev=%d EQs per dev=%d\n",
		    pio->cli_name, res_mgmtp->req_msi_per_dev,
		    res_mgmtp->req_eqs_per_dev));
	}
}

/*
 * Prep/init the MSI bucket structures.
 */
void
pcie_init_res_mgmt(pri_iodevice_t *pio)
{
	int i;
	pcie_bus_res_mgmt_t *res_mgmtp = &pio->res_mgmt;
	pcie_bus_res_t *busresp = &pio->busres;

	res_mgmtp->num_msib = MSIS_TO_MSIBS(busresp->num_msis);
	res_mgmtp->msib = Xcalloc(res_mgmtp->num_msib, msib_info_t);
	for (i = 0; i < res_mgmtp->num_msib; i++) {
		res_mgmtp->msib[i].start = MSIBS_TO_MSIS(i);
		res_mgmtp->msib[i].size = MSIB_BSIZE;
		res_mgmtp->msib[i].free = true;
	}
	res_mgmtp->num_eqs = busresp->num_eqs;
	res_mgmtp->free_eqs = busresp->num_eqs;

	res_mgmtp->num_dvma_ranges = busresp->num_dvma_ranges;
	res_mgmtp->dvma_ranges =
	    Xmalloc(sizeof (pcie_range_t));
	memcpy(res_mgmtp->dvma_ranges, busresp->dvma_ranges,
	    sizeof (pcie_range_t) * busresp->num_dvma_ranges);
}

/*
 * Destroy the data structures related to PCIe resources.
 */
void
pcie_destroy_bound_res(bound_iodevice_t *bio)
{
	/* simply free resources */
	pcie_free_res(bio);
}

/*
 * Allocate all MSI resources for one PCIe device.
 */
static void
alloc_all_msis(bound_iodevice_t *bio, pcie_bus_res_mgmt_t *res_mgmtp)
{
	int i;
	pcie_range_t *rp = &bio->busres.msi_range;

	/* Mark all MSIs as allocated */
	for (i = 0; i < res_mgmtp->num_msib; i++) {
		if (res_mgmtp->msib[i].free != true) {
			fatal("Invalid allocation: "
			    "MSIs already assigned for bus %s\n", bio->name);
		}
		res_mgmtp->msib[i].free = false;
	}
	rp->start = 0;
	rp->size = MSIBS_TO_MSIS(res_mgmtp->num_msib);
}

/*
 * Allocate pre-determined MSI resources for one PCIe device.
 */
static void
alloc_fixed_msis(bound_iodevice_t *bio, pcie_bus_res_mgmt_t *res_mgmtp,
    int num_msis, pcie_range_t *range)
{
	int j;
	int sb;
	int nb;
	pcie_range_t *rp = &bio->busres.msi_range;

	sb = MSIS_TO_MSIBS(range->start);
	nb = MSIS_TO_MSIBS(range->size);

	ASSERT(sb + nb <= res_mgmtp->num_msib);

	for (j = 0; j < nb; j++) {
		ASSERT(res_mgmtp->msib[sb].free == true);
		res_mgmtp->msib[sb++].free = false;
	}
	rp->start = range->start;
	rp->size = range->size;
}

/*
 * Allocate all Eq resources.
 */
static void
alloc_all_eqs(bound_iodevice_t *bio, pcie_bus_res_mgmt_t *res_mgmtp)
{
	if (res_mgmtp->free_eqs != res_mgmtp->num_eqs) {
		fatal("Some Eqs are assigned already");
	}
	bio->busres.num_eqs = res_mgmtp->free_eqs;
	res_mgmtp->free_eqs = 0;
}

/*
 * Allocate all PCIe resources to a pcie_bus.
 * Typically done when pre-DirectIO HVMD is processed or
 * assignment of an entire bus.
 */
void
pcie_alloc_all_res(bound_iodevice_t *bio)
{
	pcie_bus_res_mgmt_t *res_mgmtp = &bio->phys_iodev->res_mgmt;

	alloc_all_msis(bio, res_mgmtp);
	alloc_all_eqs(bio, res_mgmtp);
	bio->busres.num_dvma_ranges = res_mgmtp->num_dvma_ranges;
	bio->busres.dvma_ranges =
	    Xmalloc(res_mgmtp->num_dvma_ranges * sizeof (pcie_range_t));
	memcpy(bio->busres.dvma_ranges, res_mgmtp->dvma_ranges,
	    sizeof (pcie_range_t) * res_mgmtp->num_dvma_ranges);
}

/*
 * Free all PCIe resources assigned to a pcie bus (root complex),
 * Typically done during removal of an entire bus.
 */
void
pcie_free_res(bound_iodevice_t *bio)
{
	pcie_bus_res_mgmt_t *res_mgmtp = &bio->phys_iodev->res_mgmt;

	free_msis(bio, res_mgmtp);
	free_eqs(bio, res_mgmtp);
	Xfree(bio->busres.dvma_ranges);
	bio->busres.num_dvma_ranges = 0;
	bio->busres.dvma_ranges = NULL;
}

/*
 * Free the MSI resources associated with one root complex.
 */
static void
free_msis(bound_iodevice_t *bio, pcie_bus_res_mgmt_t *res_mgmtp)
{
	int i;
	int nb;
	int sb;
	pcie_range_t *rp = &bio->busres.msi_range;

	sb = MSIS_TO_MSIBS(rp->start);
	nb = MSIS_TO_MSIBS(rp->size);
	for (i = 0; i < nb; i++) {
		ASSERT(res_mgmtp->msib[sb].free == false);
		res_mgmtp->msib[sb++].free = true;
	}
	rp->start = 0;
	rp->size = 0;
}

/*
 * Allocate MSI resources for one PCIe device.
 */
static void
alloc_msis(bound_iodevice_t *bio, pcie_bus_res_mgmt_t *res_mgmtp)
{
	int i;
	int j;
	int des_msibs;
	int nb = 0;
	pcie_range_t *rp = &bio->busres.msi_range;

	rp->start = 0;
	rp->size = 0;

	des_msibs = MSIBS_PER_DEV(res_mgmtp) * NUM_MSI_DEVS(bio, res_mgmtp);
	if (des_msibs == 0)
		return;

	/* Find first unassigned bucket */
	for (i = 0; i < res_mgmtp->num_msib; i++) {
		if (res_mgmtp->msib[i].free == true)
			break;
	}

	if (i == res_mgmtp->num_msib) {
		/* No MSIS available */
		return;
	}

	/*
	 * Start from the first free MSI bucket, scan for
	 * contiguous free buckets upto a max of des_msibs
	 * but not go beyond the total buckets(num_msib).
	 */
	for (j = i; (j < (i + des_msibs)) && (j < res_mgmtp->num_msib); j++) {
		if (!res_mgmtp->msib[j].free)
			break;
		nb++;
		res_mgmtp->msib[j].free = false;

	}
	rp->start = MSIBS_TO_MSIS(i);
	rp->size = MSIBS_TO_MSIS(nb);
}

/*
 * Free all Eq resources.
 */
static void
free_eqs(bound_iodevice_t *bio, pcie_bus_res_mgmt_t *res_mgmtp)
{
	res_mgmtp->free_eqs += bio->busres.num_eqs;
	ASSERT(res_mgmtp->free_eqs <= res_mgmtp->num_eqs);
	bio->busres.num_eqs =  0;
}

/*
 * Allocate Eqs for a pcie device.
 */
static void
pcie_alloc_eqs(bound_iodevice_t *bio, pcie_bus_res_mgmt_t *res_mgmtp)
{
	int neqs = NUM_EQ_DEVS(bio, res_mgmtp) * res_mgmtp->req_eqs_per_dev;

	ASSERT(res_mgmtp->free_eqs >= neqs);
	bio->busres.num_eqs = neqs;
	res_mgmtp->free_eqs -= neqs;
}

/*
 * Allocate fixed Eq resources.
 */
static void
pcie_alloc_fixed_eqs(bound_iodevice_t *bio,
    pcie_bus_res_mgmt_t *res_mgmtp, int neqs)
{
	if (res_mgmtp->free_eqs < neqs) {
		/* Allocate available eqs */
		neqs = res_mgmtp->free_eqs;
	}
	bio->busres.num_eqs = neqs;
	res_mgmtp->free_eqs -= neqs;
}

/*
 * Alloc PCIe bus resources during a PCIe device assignment.
 * Assumes resources are being assigned for one PCIe device.
 * Note: the new amount allocated can be 0.
 */
void
pcie_realloc_res(bound_iodevice_t *bio)
{
	pcie_bus_res_mgmt_t *res_mgmtp = &bio->phys_iodev->res_mgmt;

	/* Free current resources and reallocate */
	pcie_free_res(bio);

	/* If there are no devices, no allocation needed */
	if (NUM_MSI_DEVS(bio, res_mgmtp) == 0)
		return;

	alloc_msis(bio, res_mgmtp);
	/* sanity check */
	if (bio->busres.msi_range.size == 0)
		fatal("No MSIs assigned to %s", bio->name);

	pcie_alloc_eqs(bio, res_mgmtp);
	/* sanity check */
	if (bio->busres.num_eqs == 0)
		fatal("No EQs assigned to %s", bio->name);

	bio->busres.num_dvma_ranges = res_mgmtp->num_dvma_ranges;
	bio->busres.dvma_ranges =
	    Xmalloc(res_mgmtp->num_dvma_ranges * sizeof (pcie_range_t));
	memcpy(bio->busres.dvma_ranges, res_mgmtp->dvma_ranges,
	    sizeof (pcie_range_t) * res_mgmtp->num_dvma_ranges);
}

/*
 * Assign pre-determined resources to a bound_iodevice.
 * This is typically done when the properties are being populated
 * with resources already found in the HVMD.
 */
void
pcie_alloc_fixed(bound_iodevice_t *bio, pcie_bus_res_t *busresp)
{
	pcie_bus_res_mgmt_t *res_mgmtp = &bio->phys_iodev->res_mgmt;

	alloc_fixed_msis(bio, res_mgmtp, busresp->num_msis,
	    &busresp->msi_range);
	pcie_alloc_fixed_eqs(bio, res_mgmtp, busresp->num_eqs);
	bio->busres.num_dvma_ranges = busresp->num_dvma_ranges;
	bio->busres.dvma_ranges =
	    Xmalloc(res_mgmtp->num_dvma_ranges * sizeof (pcie_range_t));
	memcpy(bio->busres.dvma_ranges, busresp->dvma_ranges,
	    sizeof (pcie_range_t) * busresp->num_dvma_ranges);

}

/*
 * Create and link the requirement structures for all pcie devices
 * under a specific root complex.  This is used in cases when the
 * binding for the root complex (bound_iodevice_t) itself is built
 * from implied requirements (such as HVMD), and we derive the
 * requirements backwards from the binding info.
 */
void
pcie_update_reqs(ldom_t *ldp, bound_iodevice_t *biop)
{
	bound_pcie_dev_t *bdev;
	req_pcie_dev_t *rdev;

	CHAIN_ITER(biop->dev_list, bdev, ldom) {
		/* Create the requirement structure and link it. */
		rdev =
		    pcie_build_req_pciedev(bdev->phys_pcie_dev->cli_name, true);
		CHAIN_ADD(ldp->required.io_pcie_devs, rdev, link);
		bdev->req_pcie_dev = rdev;
		database_ldom_add_pcie_device(ldp, rdev->cli_name);
	}
	database_save();
}

/*
 * Assign all assignable devices to the pcie bus owner.
 * Typically used only when pre-DirectIO HVMD is being parsed.
 */
void
pcie_assign_all_devs(bound_iodevice_t *bio)
{
	pri_pcie_dev_t *pdev;
	bound_pcie_dev_t *bdev;

	pcie_alloc_all_res(bio);

	CHAIN_ITER(bio->phys_iodev->dev_list, pdev, link) {
		/* Create the binding structure and link it. */
		bdev = pcie_build_bound_pciedev(pdev);
		CHAIN_ADD(bio->dev_list, bdev, ldom);
		CHAIN_ADD(pdev->bound_devs, bdev, pio);
		pdev->free = false;
	}
}

void
pcie_destroy_bound_pciedev(bound_pcie_dev_t *bdev)
{
	Xfree(bdev->cli_name);
	Xfree(bdev->skel_path);
	if (bdev->id != NULL)
		free_res(bdev->id);
	Xfree(bdev);
}

/*
 * Constructor for pciedev requirement structure.
 */
req_pcie_dev_t *
pcie_build_req_pciedev(char *name, bool_t bound)
{
	req_pcie_dev_t *rdev;

	rdev = Xcalloc(1, req_pcie_dev_t);
	rdev->cli_name = Xstrdup(name);
	rdev->bound = bound;
	return (rdev);
}

/*
 * Constructor for pciedev binding structure.
 */
bound_pcie_dev_t *
pcie_build_bound_pciedev(pri_pcie_dev_t *pdev)
{
	bound_pcie_dev_t *bdev;

	bdev = Xcalloc(1, bound_pcie_dev_t);
	bdev->cli_name = Xstrdup(pdev->cli_name);
	bdev->skel_path = Xstrdup(pdev->skel_path);
	bdev->phys_pcie_dev = pdev;
	return (bdev);
}

void
pcie_destroy_req_pciedev(req_pcie_dev_t *rdev)
{
	Xfree(rdev->cli_name);
	Xfree(rdev);
}

/*
 * Unassign all devices.
 * Typically used when an entire bus is being un-assigned.
 */
void
pcie_unassign_all_devs(ldom_t *ldp, bound_iodevice_t *bio)
{
	pri_pcie_dev_t *pdev;
	bound_pcie_dev_t *bdev;
	req_pcie_dev_t *rdev;

	while ((bdev = CHAIN_START(bio->dev_list)) != NULL) {
		/* Destroy the constraint. */
		rdev = bdev->req_pcie_dev;
		database_ldom_del_pcie_device(ldp, rdev->cli_name);
		CHAIN_UNLINK(ldp->required.io_pcie_devs, rdev, link);
		pcie_destroy_req_pciedev(rdev);

		/* Destroy the binding. */
		pdev = bdev->phys_pcie_dev;
		CHAIN_UNLINK(pdev->bound_devs, bdev, pio);
		CHAIN_UNLINK(bio->dev_list, bdev, ldom);
		pdev->free = true;

		pcie_destroy_bound_pciedev(bdev);
	}
	database_save();
}

/*
 * Update the Root complex related MD nodes in the Guest MD
 * to reflect the actual Root Complex resources assigned to this
 * guest.
 */
static void
fix_iodevice_node(ldom_t *ldp, md_t *mdp, md_node_t *nodep)
{
	bound_iodevice_t	*bio;
	uint64_t	cfg_handle;
	int		rval;
	pcie_range_t	*ranges;
	pcie_eqino_t	*eqino;
	pcie_eqino_t	neqino;


	if (md_get_prop_val(mdp, nodep, "cfg-handle", &cfg_handle) < 0)
		fatal("Missing cfg-handle property in guest template "
		    "iodevice node");

	CHAIN_ITER(ldp->bound.iodevices, bio, ldom) {
		if (bio->phys_iodev->cfg_handle == cfg_handle)
			break;
	}
	if (bio == NULL) {
		fatal("No bound iodevice found");
	}

	DBG_BLOCK {
		uint64_t	val;

		if (md_get_prop_val(mdp, nodep, "#msi", &val) != 0)
			fatal("Missing #msi property.");

		if (md_get_prop_data(mdp, nodep, "msi-ranges",
		    (uint8_t **)&ranges, &rval) != 0)
			fatal("Missing msi-ranges property.");

		if (md_get_prop_data(mdp, nodep, "virtual-dma",
		    (uint8_t **)&ranges, &rval) != 0)
			fatal("Missing virtual-dma property.");

		if (md_get_prop_val(mdp, nodep, "#msi-eqs", &val) != 0)
			fatal("Missing #msi-eqs property.");
	}

	/* Update the number of MSIs */
	md_update_value_property(mdp, nodep, "#msi",
	    bio->busres.msi_range.size);

	/* Update the MSI ranges to reflect the new values */
	md_update_data_property(mdp, nodep, "msi-ranges",
	    sizeof (pcie_range_t), (uint8_t *)&bio->busres.msi_range);

	/* Update the number of EQs */
	md_update_value_property(mdp, nodep, "#msi-eqs", bio->busres.num_eqs);

	/* Fix the actual count of Eqs in the msi-eq-to-devino prop */
	if (md_get_prop_data(mdp, nodep, "msi-eq-to-devino",
	    (uint8_t **)&eqino, &rval) != 0) {
		fatal("Missing msi-eq-to-devino property.");
	}
	ASSERT(rval == sizeof (pcie_eqino_t));
	/* update the count only, keep the rest as the same */
	neqino = *eqino;
	neqino.count = bio->busres.num_eqs;
	md_update_data_property(mdp, nodep, "msi-eq-to-devino",
	    sizeof (pcie_eqino_t), (uint8_t *)&neqino);

	md_update_data_property(mdp, nodep, "virtual-dma",
	    bio->busres.num_dvma_ranges * sizeof (pcie_range_t),
	    (uint8_t *)bio->busres.dvma_ranges);

	if (bio->virtual_pcie_bus) {
		md_add_value_property(mdp, nodep, "virtual-root-complex", 0x1);
		pcie_prune_branches(mdp, nodep, bio);
	}
}

/*
 * Determine if the partial device path in 'prefix' matches the initial
 * part of the 'path' element of any of the devices list entries for
 * a given root complex.
 */
static bool_t
is_dev_needed(bound_iodevice_t *bio, char *prefix)
{
	char path[MAXPATHLEN];
	bound_pcie_dev_t *bdev;
	size_t len;
	size_t len_prefix;
	size_t len_path;

	len_prefix = strlen(prefix);

	/*
	 * Walk the list of devices under the root complex 'bio'.
	 * Compare the partial path in 'prefix' against the path of
	 * each device.  If the prefix matches the first part of the
	 * current device's path, then we can't yet prune devices
	 * whose path's begin with 'prefix' from the guest MD, so
	 * we return TRUE.  Otherwise, return FALSE.
	 *
	 * Implementation note: our caller is walking the device
	 * tree in a breadth first manner.  If this function is
	 * presented with a prefix of "@1" and none of the devices
	 * in the dev_list for this root complex begin with that
	 * string, we will return FALSE, allowing the caller to
	 * prune that branch of the tree at a high level.  Otherwise,
	 * we return FALSE and the caller will walk further down
	 * the tree and present successively longer paths for us
	 * to check.
	 */
	CHAIN_ITER(bio->dev_list, bdev, ldom) {
		/* Terminate the string for use with strncmp. */
		(void) snprintf(path, sizeof (path),
		    "%s/", bdev->phys_pcie_dev->rel_path);
		/*
		 * length for strncmp:
		 *
		 * If the length of 'prefix' exceeds the length of 'path'
		 * we want to use the length of 'path' so we keep all
		 * subdevices which fall under the borrowed device.  But,
		 * if the lengths are reversed, we use length of 'prefix'
		 * for the compare which allows us to prune device tree
		 * branches we don't need.
		 */
		len_path = strlen(path);
		len = len_path > len_prefix ? len_prefix : len_path;
		if (strncmp(prefix, path, len) == 0)
			break;
	}

	return (bdev != NULL);
}

/*
 * Examine the MD device nodes at/below 'parent'.  Prune the 'parent'
 * node if the corresponding device is not used in the devices list
 * (for the root complex represented by 'bio').  If the 'parent' node
 * is in use, look further down the tree.
 *
 * Note: This routine assumes that a device tree parent will be encountered
 * in the list returned by md_scan_dag before the nodes for any of its
 * children.
 */
static void
prune_subtree(md_t *mdp, md_node_t *parent, bound_iodevice_t *bio,
    md_node_t **nlist, int n, int nmax, char *path)
{
	uint64_t dev_num;
	md_node_t *np;
	md_prop_t *prop;
	char new_path[MAXPATHLEN];

	for (/* nothing */; n < nmax; n++) {
		np = nlist[n];

		if (md_get_prop_val(mdp, np, "device-number", &dev_num) < 0)
			continue;

		CHAIN_ITER(np->prop_list, prop, link) {
			if (strcmp(prop->sp->strp, "back") != 0 ||
			    prop->d.arc.val.nodep != parent)
				continue;

			(void) snprintf(new_path, sizeof (new_path),
			    "%s@%"PRIx64"/", path, dev_num);
			if (!is_dev_needed(bio, new_path))
				md_mark_node_deleted(np);
			else
				prune_subtree(mdp,
				    np, bio, nlist, n + 1, nmax, new_path);
			break;
		}
	}
}

/*
 * Prune all the branches in the MD except for those in the devices
 * list of the root complex represented by 'bio'.  The devices in
 * the list are the ones being shared with the domain by the domain
 * which owns the respective hardware.
 */
static void
pcie_prune_branches(md_t *mdp, md_node_t *nodep, bound_iodevice_t *bio)
{
	md_node_t **nodelist;
	int num;
	int i;

	/*
	 * Handle special case of 'wart' nodes.  These appear at same
	 * level as root complex, but are actually under a root complex
	 * from a hardware point of view.  We need to get rid of them.
	 */
	nodelist = Xmalloc(md_node_count(mdp) * sizeof (md_node_t *));
	num = md_scan_dag(mdp, nodep,
	    md_find_name(mdp, "wart"),
	    md_find_name(mdp, "fwd"), nodelist);

	for (i = 0; i < num; i++)
		md_mark_node_deleted(nodelist[i]);

	/* Now do the standard pruning. */
	num = md_scan_dag(mdp, nodep,
	    md_find_name(mdp, "iodevice"),
	    md_find_name(mdp, "fwd"), nodelist);

	if (num <= 0)
		return;

	prune_subtree(mdp, nodep, bio, nodelist, 0, num, "");

	Xfree(nodelist);
}

/*
 * Fix the Guest MD to reflect the resources that are actually assigned
 * to this domain. Also prune branches that are not in the path of
 * slots/devices assigned to this domain.
 */
void
pcie_fix_guestmd(ldom_t *ldp, md_t *mdp, md_node_t **ionodelist, int num)
{
	int i;
	md_node_t *nodep;
	char *devtype_p;

	if (!pcie_gen_dio_hvmd()) {
		/* Nothing to fix */
		return;
	}

	for (i = 0; i < num; i++) {
		nodep = ionodelist[i];
		if (md_get_prop_str(mdp, nodep, "device-type",
		    &devtype_p) < 0) {
			/* Ignore it, instead of a fatal() */
			continue;
		}
		if (strcmp(devtype_p, "pciex") == 0) {
			fix_iodevice_node(ldp, mdp, nodep);
		}
	}
}

/*
 * Return the name of the leaf subdevice component given the path to its
 * parent device and its full path.
 */
static char *
get_subdev_name(char *dev_path, char *subdev_path)
{
	char *sbp;
	size_t dev_len = strlen(dev_path);

	if (strncmp(dev_path, subdev_path, dev_len) != 0)
		return (NULL);

	sbp = &subdev_path[dev_len + 1];
	return (Xstrdup(sbp));
}

/*
 * Destroy, and free the space for, the subdevice info of one removeable
 * pcie device.
 */
static void
destroy_subdevs(pri_pcie_dev_t *pdev)
{
	pri_pcie_subdev_t *sdev;

	ASSERT(MUTEX_HELD(&pdev->subdev_lock));

	while (CHAIN_LENGTH(pdev->sub_devs) > 0) {
		sdev = CHAIN_START(pdev->sub_devs);
		CHAIN_UNLINK(pdev->sub_devs, sdev, link);
		Xfree(sdev->subdev_path);
		Xfree(sdev);
	}
}

/*
 * Process one device node of the root domain's device tree read from
 * agent-dio service.  The node provides information not available in
 * the PRI, such as whether the slot is occupied, and if so, the
 * names of the subdevices.  Record this data in the zeus internal
 * PRI structure.
 */
static void
process_devnode(md_t *mdp,
    md_node_t *devnodep, pri_pcie_dev_t *pdev, char *path)
{
	pri_pcie_subdev_t *sdev;
	md_node_t **nodes;
	md_node_t *nodep;
	char	*sbpath;
	char	*sbp;
	int num;
	int i;

	mutex_lock(&pdev->subdev_lock);

	destroy_subdevs(pdev);

	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));
	num = md_scan_dag(mdp, devnodep, md_find_name(mdp, "iodevice"),
	    md_find_name(mdp, "fwd"), nodes);

	for (i = 0; i < num; i++) {
		nodep = nodes[i];
		int nfwd;

		if (nodep == devnodep)
			continue;

		nfwd = md_scan_dag(mdp, nodep, md_find_name(mdp, "iodevice"),
		    md_find_name(mdp, "fwd"), NULL);

		if (nfwd != 1) {
			/* not a leaf device, ignore */
			continue;
		}

		if (md_get_prop_str(mdp, nodep, "dev_path", &sbpath) == -1) {
			continue;
		}
		DBG(zprintf("%s: Subdevice ---> %s\n", __func__, sbpath));
		if ((sbp = get_subdev_name(path, sbpath)) == NULL)
			continue;
		DBG(zprintf("%s: Subdevice(short) ---> %s\n", __func__, sbp));

		sdev = Xcalloc(1, pri_pcie_subdev_t);
		CHAIN_LINK_INIT(sdev->link);
		sdev->subdev_path = sbp;
		CHAIN_ADD(pdev->sub_devs, sdev, link);
	}

	if (CHAIN_LENGTH(pdev->sub_devs) > 0) {
		pdev->status = SL_OCCUPIED;
	} else {
		pdev->status = SL_EMPTY;
	}
	mutex_unlock(&pdev->subdev_lock);
}

/*
 * Create the subdevice list for a single root complex.  The list contains
 * a node for each subdevice.
 */
static void
populate_subdevice_list(ldom_t *ldp, md_t *mdp,
    md_node_t *rcnodep, pri_iodevice_t *pio)
{
	md_node_t **nodes;
	md_node_t *nodep;
	char	*path;
	int num;
	int i;
	pri_pcie_dev_t *pdev;

	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));
	num = md_scan_dag(mdp, rcnodep, md_find_name(mdp, "iodevice"),
	    md_find_name(mdp, "fwd"), nodes);

	for (i = 0; i < num; i++) {
		nodep = nodes[i];
		if (md_get_prop_str(mdp, nodep, "dev_path", &path) == -1) {
			continue;
		}
		/* Skip entries which don't start with / */
		if (path[0] != '/')
			continue;
		CHAIN_ITER(pio->dev_list, pdev, link) {
			/* skip the first '/' */
			if (strcmp(&path[1], pdev->name) == 0) {
				/* Match */
				process_devnode(mdp, nodep, pdev, path);
				break;
			}
		}
	}

	Xfree(nodes);
}

/*
 * Process PCI device tree from Root Domain.  Here we collect information
 * about the device configuration which is not available from the PRI.
 *
 * Call graph:
 *	process_devtree_from_guest - examine all root complexes
 *	    populate_subdevice_list - process a single root complex
 *		process_devnode - process one subdevice node of current RC.
 */
static void
process_devtree_from_guest(ldom_t *ldp, md_t *mdp)
{
	md_node_t	**nodelistp;
	md_node_t	*rootnode;
	md_node_t	*parent;
	md_node_t	*nodep;
	int		num_nodes;
	int		i;
	char		*nexus;
	bound_iodevice_t *bio;

	rootnode = md_root_node(mdp);

	num_nodes = md_node_count(mdp);
	nodelistp = Xmalloc(num_nodes * sizeof (md_node_t *));

	num_nodes = md_scan_dag(mdp, rootnode,
	    md_find_name(mdp, "iodevice"),
	    md_find_name(mdp, "fwd"), nodelistp);

	/* Identify the Root Complex nodes. */
	for (i = 0; i < num_nodes; i++) {
		nodep = nodelistp[i];

		if (md_get_prop_arc(mdp, nodelistp[i], "back", &parent) < 0) {
			warning("Missing back arc in nexus node.\n");
			continue;
		}

		if (parent != rootnode) {
			continue;
		}
		if (md_get_prop_str(mdp, nodep, "dev_path", &nexus) != 0) {
			warning("Missing dev_path for device node.");
			continue;
		}

		if (strlen(nexus) == 0) {
			continue;
		}

		/*
		 * Walk the binding structures for pcie buses looking
		 * for a match with the path of the current MD node.
		 * These MD nodes contain pcie bus data from the DIO
		 * agent and include additonal hardware information.
		 * The string is like "pci@400".
		 */
		CHAIN_ITER(ldp->bound.iodevices, bio, ldom) {
			char *bus_name = &nexus[1]; /* Ignore the leading / */

			if (strcmp(bus_name, bio->name) == 0) {
				/*
				 * Identify the subdevices and create the
				 * subdevice list for this device under
				 * the PRI structure for the root complex.
				 */
				populate_subdevice_list(ldp,
				    mdp, nodep, bio->phys_iodev);
				break;
			}
		}
	}
	Xfree(nodelistp);
}

/*
 * Get the device info from the LDoms dio agent and build the corresponding
 * internal structures.
 */
void
pcie_get_devinfo(ldom_t *ldp)
{
	typedef int (*dsdio_get_devinfo_f)(ldom_t *ldp,
	    void **bufpp, uint32_t *size);
	dsdio_get_devinfo_f dsdio_get_devinfo;
	void		*buf;
	uint32_t	size;
	int		rv;
	md_t		*devtree_md = NULL;

	ASSERT(ldp != NULL);

	/* Use the ds_dio library to communicate with the DIO agent. */
	dsdio_get_devinfo = (dsdio_get_devinfo_f)
	    ds_service_routine(ldp->ds_chan, DIO_SVC, "dsdio_get_devinfo");
	if (dsdio_get_devinfo == NULL) {
		warning("%s: service is not registered(%s)",
		    ldp->namep, DIO_SVC);
		return;
	}

	DBG(zprintf("%s: calling get dsdio_get_devinfo for %s\n",
	    __func__, ldp->namep));
	rv = dsdio_get_devinfo(ldp, &buf, &size);

	if (rv == 0) {
		DBG(zprintf("%s: dsdio_get_devinfo succeeded %d\n",
		    __func__, size));
		devtree_md = md_scan_bin((uint8_t *)buf, size);
		Xfree(buf);
	} else {
		DBG(zprintf("%s: dsdio_get_devinfo failed %d\n",
		    __func__, rv));
		return;
	}

	process_devtree_from_guest(ldp, devtree_md);
	md_fini(devtree_md);
}

bound_iodevice_t *
pcie_biodevice_exists(ldom_t *ldp, uint64_t cfg_handle)
{
	bound_iodevice_t *bio;
	pri_iodevice_t *pio;

	CHAIN_ITER(ldp->bound.iodevices, bio, ldom) {
		pio = bio->phys_iodev;
		if (pio->is_pciex && (pio->cfg_handle == cfg_handle)) {
			return (bio);
		}
	}
	return (NULL);
}

/*
 * Create the resource properties for the node (phase 1).  For physical
 * root complex nodes, also create the node for assignable devices and
 * its child nodes for the individual devices under this root complex.
 */
void
add_pcie_props1(md_t *mdp, ldom_t *ldp, md_node_t *node, bound_iodevice_t *bio)
{
	pri_iodevice_t *pio;
	pri_pcie_dev_t *pdev;
	md_node_t *assignable;
	md_node_t *dev;

	if (bio->busres.msi_range.size == 0)
		fatal("No MSIs assigned to %s:%s", ldp->namep, bio->name);
	if (bio->busres.num_eqs == 0)
		fatal("No EQs assigned to %s:%s", ldp->namep, bio->name);
	if (bio->busres.num_dvma_ranges == 0)
		fatal("No DVMA ranges set for %s:%s", ldp->namep, bio->name);

	md_add_value_property(mdp, node, "#msi",
	    bio->busres.msi_range.size);
	md_add_data_property(mdp, node, "msi-ranges",
	    sizeof (pcie_range_t), (uint8_t *)&bio->busres.msi_range);

	md_add_value_property(mdp, node, "#msi-eqs", bio->busres.num_eqs);
	md_add_data_property(mdp, node, "virtual-dma",
	    bio->busres.num_dvma_ranges * sizeof (pcie_range_t),
	    (uint8_t *)bio->busres.dvma_ranges);

	if (!bio->virtual_pcie_bus) {
		assignable = md_new_node(mdp, "pcie_assignable_devices");
		md_add_arc_property(mdp, node, "fwd", assignable);
		md_add_arc_property(mdp, assignable, "back", node);

		pio = bio->phys_iodev;

		CHAIN_ITER(pio->dev_list, pdev, link) {
			dev = md_new_node(mdp, "pcie_device");
			md_add_string_property(mdp, dev, "path",
			    pdev->rel_path);
			md_add_arc_property(mdp, assignable, "fwd", dev);
			md_add_arc_property(mdp, dev, "back", assignable);
			pdev->hvmd_pciedev = dev;
		}
	}
}

/*
 * Add pcie properties (phase 2).  Create and link the "assigned devices"
 * node for this root complex.  Link the child "pcie device"
 * nodes for this root complex to the new "assigned devices" node.
 */
void
add_pcie_props2(md_t *mdp,
    ldom_t *ldp, md_node_t *node, bound_iodevice_t *bio)
{
	md_node_t *assigned;
	bound_pcie_dev_t *bdev;

	assigned = md_new_node(mdp, "pcie_assigned_devices");
	md_add_arc_property(mdp, node, "fwd", assigned);
	md_add_arc_property(mdp, assigned, "back", node);

	CHAIN_ITER(bio->dev_list, bdev, ldom) {
		md_node_t *pnode = bdev->phys_pcie_dev->hvmd_pciedev;

		md_add_arc_property(mdp, assigned, "fwd", pnode);
		md_add_arc_property(mdp, pnode, "back", assigned);
	}
}

/* Generate Direct I/O modifications to the HVMD? */
bool_t
pcie_gen_dio_hvmd(void)
{
	/* For now just return true if HV is DIO capable */
	return (zeus.sdio_enabled);
}

/*
 * Verify that this root complex has no devices removed.
 */
bool_t
pcie_is_rc_owner(ldom_t *ldp, bound_iodevice_t *bio)
{
	pri_iodevice_t *pio = bio->phys_iodev;
	pri_pcie_dev_t *pdev;

	/* There should be only one bound iodevice */
	if (CHAIN_LENGTH(pio->bound_ios) != 1) {
		return (false);
	}
	/* All devices must be in the assigned state */
	CHAIN_ITER(pio->dev_list, pdev, link) {
		if (pdev->free == true)
			return (false);
	}
	return (true);
}

/* Root complex removable? */
bool_t
pcie_is_rc_removable(ldom_t *oldp, bound_iodevice_t *bio)
{
	ldom_t *ldp;
	pri_iodevice_t *pio = bio->phys_iodev;
	pri_pcie_dev_t *pdev;

	/*
	 * If the Root complex is shared with more than
	 * one domain, then it's not removable.
	 */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (pcie_biodevice_exists(ldp, pio->cfg_handle)) {
			if (oldp != ldp) {
				return (false);
			}
		}
	}

	/* All devices must be in the assigned state */
	CHAIN_ITER(pio->dev_list, pdev, link) {
		if (pdev->free == true)
			return (false);
	}

	return (true);
}

/*
 * Check whether 'name' is one of the valid names for the loanable I/O
 * device represented by the PRI record 'pdev'.
 */
bool_t
pcie_match_pciedev(char *name, pri_pcie_dev_t *pdev)
{
	return ((strcmp(pdev->name, name) == 0) ||
	    (pdev->cli_name && (strcmp(pdev->cli_name, name)) == 0));
}

pri_pcie_dev_t *
pcie_find_matching_dev(char *name)
{
	pri_iodevice_t *pio;
	pri_pcie_dev_t *pdev;

	CHAIN_ITER(pri.iodevice_list, pio, link) {
		CHAIN_ITER(pio->dev_list, pdev, link) {
			if (pcie_match_pciedev(name, pdev))
				return (pdev);
		}
	}

	return (NULL);
}

/*
 * Given 'name' of a valid, loanable pcie device return the corresponding
 * PRI record for it.  The input name may be either of the (possibly) two
 * names associated with the device.  Return NULL if not found or not
 * assignable.  Return pointer to owning root complex via 'piop'.
 */
pri_pcie_dev_t *
pcie_find_pri_pciedev(char *name, pri_iodevice_t **piop)
{
	pri_iodevice_t *pio;
	pri_pcie_dev_t *pdev;

	if (!zeus.sdio_enabled)
		return (false);

	CHAIN_ITER(pri.iodevice_list, pio, link) {
		if (!pio->is_pciex)
			continue;
		CHAIN_ITER(pio->dev_list, pdev, link) {
			if (strcmp(pdev->name, name) == 0 ||
			    (pdev->cli_name != NULL &&
			    strcmp(pdev->cli_name, name) == 0)) {
				*piop = pio;
				return (pdev);
			}
		}
	}

	return (NULL);
}

/*
 * Return pointer to the bound_device struct (or NULL) for a given
 * shareable device.  Also, return a pointer to the device's bus
 * parent via 'biop'.
 */
bound_pcie_dev_t *
pcie_find_bound_pciedev(ldom_t *ldp, char *name, bound_iodevice_t **biop)
{
	pri_iodevice_t *pio;
	pri_pcie_dev_t *pdev;
	bound_iodevice_t *bio;
	bound_pcie_dev_t *bdev;

	if ((pdev = pcie_find_pri_pciedev(name, &pio)) == NULL)
		return (NULL);

	CHAIN_ITER(ldp->bound.iodevices, bio, ldom) {
		if (bio->phys_iodev != pio)
			continue;
		CHAIN_ITER(bio->dev_list, bdev, ldom) {
			if (strcmp(bdev->skel_path, pdev->skel_path) == 0) {
				*biop = bio;
				return (bdev);
			}
		}
	}

	return (NULL);
}

/*
 * Determine if the domain has any devices removed from a phys. root complex.
 */
bool_t
has_io_dependencies(ldom_t *ldp)
{
	bound_iodevice_t *bio;

	CHAIN_ITER(ldp->bound.iodevices, bio, ldom) {
		if (bio->virtual_pcie_bus)
			continue;

		if (!pcie_is_rc_owner(ldp, bio))
			return (true);
	}

	return (false);
}

/*
 * Convert a path like /@400/@0/@d into a path like
 * pci@400/pci@0/pci@d.
 */
static char *
pcie_skelpath_to_fullpath(char *path)
{
	char fpath[MAXPATHLEN];
	char *s;
	int i;

	ASSERT(path != NULL && *path == '/');

	s = path;
	i = 0;
	while ((*s != '\0') &&
	    (i < (sizeof (fpath)) - (sizeof (PCI_BUS_NAME) - 1))) {
		fpath[i++] = *s;
		if (*s == '/') {
			(void) strcpy(fpath + i, PCI_BUS_NAME);
			i += (sizeof (PCI_BUS_NAME) - 1);
		}
		s++;
	}
	fpath[i] = '\0';

	/* skip the leading '/' */
	return (Xstrdup(&fpath[1]));
}

/*
 * Produce a relative path used only for HVMD.
 */
static char *
pcie_skelpath_to_relpath(char *skel_path)
{
	char *s;

	/* skip the leading slash */
	skel_path++;

	/* skip chars until next slash and also the slash */
	s = strchr(skel_path, '/');
	ASSERT(s != NULL);
	s++;

	return (Xstrdup(s));
}

/*
 * Build a skeleton path from cfghdl and relative path.
 */
char *
pcie_build_skel_path(uint64_t cfghdl, char *rel_path)
{
	char skelp[128];

	snprintf(skelp, sizeof (skelp), "/@%"PRIx64"/%s", cfghdl, rel_path);
	return (Xstrdup(skelp));
}

/*
 * Generate the format string used by the list-io command so as to
 * accommodate long column entries but still have the columns line up.
 * Always allow a minimum of two spaces between adjacent column entries.
 */
static void
pcie_gen_listio_format(void)
{
	pri_iodevice_t *pio;
	pri_pcie_dev_t *pdev;
	int name_max = sizeof ("PCIE") - 1;
	int name_len;
	int cli_name_max = sizeof ("PSEUDONYM") - 1;
	int cli_name_len;

	CHAIN_ITER(pri.iodevice_list, pio, link) {
		CHAIN_ITER(pio->dev_list, pdev, link) {
			name_len = strlen(pdev->name);
			cli_name_len = strlen(pdev->cli_name);

			if (name_len > name_max)
				name_max = name_len;
			if (cli_name_len > cli_name_max)
				cli_name_max = cli_name_len;
		}
	}

	(void) sprintf(listio_format, "%%-%ds%%-%ds%%-8s%%s\n",
	    name_max + 2, cli_name_max + 2);
}

char *
pcie_get_listio_format(void)
{
	return (listio_format);
}

/*
 * Find all IO slots that can be assigned to another domain for
 * physical access.
 */
void
pcie_pri_get_io_slots(md_t *mdp)
{
	md_node_t **nodelistp;
	int num;
	int i;

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "component"),
	    md_find_name(mdp, "fwd"), nodelistp);

	for (i = 0; i < num; i++) {
		char *path;
		char *slot;
		md_node_t *nodep;
		pri_iodevice_t *iodev;
		pri_pcie_dev_t *pdev;
		char *slash;
		char bus[64];
		size_t bus_len;
		char pathbuf[MAXPATHLEN];

		nodep = nodelistp[i];

		if (md_get_prop_str(mdp, nodep, "assignable-path",
		    &path) < 0)
			continue;

		if (md_get_prop_str(mdp, nodep, "nac", &slot) < 0) {
			warning("Missing \"nac\" field in PRI component node");
			continue;
		}

		/*
		 * Break the device's path into a 'bus' component and a 'device'
		 * component.  The 'bus' component is saved without the leading
		 * slash for compatibility with other code.
		 *
		 * The 'device' component is the remainder of the path.  It
		 * is saved in a skeleton format required for use in sending
		 * to the HV.  We must drop the leading '/' also the final
		 * component, starting with the last '/'.
		 */
		slash = strchr(path + 1, '/');
		ASSERT(slash != NULL);
		(void) strcpy(bus, PCI_BUS_NAME);
		bus_len = strlen(PCI_BUS_NAME) + slash - path - 1;
		if (bus_len > sizeof (bus) - 1)
			bus_len = sizeof (bus) - 1;
		(void) strncat(bus, path + 1, bus_len - 3);
		bus[bus_len] = '\0';

		(void) strlcpy(pathbuf, path, sizeof (pathbuf));

		/* Find the root complex for this device. */
		CHAIN_ITER(pri.iodevice_list, iodev, link) {
			if (strcmp(iodev->name, bus) == 0)
				break;
		}

		if (iodev == NULL) {
			warning("No root complex "
			    "found for device /%s/%s", bus, pathbuf);
			continue;
		}

		pdev = Xcalloc(1, pri_pcie_dev_t);
		pdev->skel_path = Xstrdup(pathbuf);
		pdev->rel_path = pcie_skelpath_to_relpath(pathbuf);
		pdev->name = pcie_skelpath_to_fullpath(pdev->skel_path);
		pdev->cli_name = Xstrdup(slot);
		pdev->status = SL_UNKNOWN;
		pdev->free = true;

		CHAIN_INIT(pdev->bound_devs);
		CHAIN_INIT(pdev->sub_devs);
		CHAIN_LINK_INIT(pdev->link);
		CHAIN_ADD(iodev->dev_list, pdev, link);
		mutex_init(&pdev->subdev_lock, USYNC_THREAD, NULL);

		DBG(zprintf("shareable device: slot = %s, path = %s\n",
		    pdev->cli_name, pdev->skel_path));
	}

	pcie_gen_listio_format();

	Xfree(nodelistp);
}

/*
 * Get the MSIs, EQs and DVMA ranges for this root complex from the
 * corresponding MD node.
 */
void
pcie_pri_get_resources(md_t *mdp, md_node_t *nodep, pri_iodevice_t *pio)
{
	pcie_range_t *msi_range;
	int num_ranges;
	pcie_range_t *dvma_ranges;
	uint64_t num_msis;
	uint64_t num_eqs;
	int	len;
	int	i;

	if (md_get_prop_val(mdp, nodep, "#msi-eqs", &num_eqs))
		fatal("missing #msi-eqs property in PRI iodevice node");
	pio->busres.num_eqs = num_eqs;

	if (md_get_prop_val(mdp, nodep, "#msi", &num_msis))
		fatal("missing #msi property in PRI iodevice node");
	pio->busres.num_msis = num_msis;
	if (md_get_prop_data(mdp,
	    nodep, "msi-ranges", (uint8_t **)&msi_range, &len))
		fatal("missing msi-ranges property in PRI device node");
	num_ranges = len / sizeof (pcie_range_t);
	ASSERT(num_ranges == 1);
	pio->busres.msi_range.start = msi_range->start;
	pio->busres.msi_range.size = msi_range->size;

	if (md_get_prop_data(mdp,
	    nodep, "virtual-dma", (uint8_t **)&dvma_ranges, &len))
		fatal("missing virtual-dma prop in PRI device node");
	num_ranges = len / sizeof (pcie_range_t);
	pio->busres.dvma_ranges = Xcalloc(num_ranges, pcie_range_t);
	for (i = 0; i < num_ranges; i++) {
		pio->busres.dvma_ranges[i].start = dvma_ranges[i].start;
		pio->busres.dvma_ranges[i].size = dvma_ranges[i].size;
	}
	pio->busres.num_dvma_ranges = num_ranges;
	pcie_init_res_mgmt(pio);
}
