/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_PCIE_H
#define	_PCIE_H

#include "chain.h"
#include "ldom.h"
#include "zpri.h"

#ifdef	__cplusplus
extern "C" {
#endif

/* A common structure to represent a range for both MSIs and DVMA */
typedef struct pcie_range {
	uint64_t	start;		/* start of the range */
	uint64_t	size;		/* size of the range */
} pcie_range_t;

/* Structure for msi-eq-to-devino property */
typedef struct pcie_eqino {
	uint64_t	start;		/* starting EQ number */
	uint64_t	count;		/* number of EQs */
	uint64_t	devino;		/* starting devino */
} pcie_eqino_t;

/* PCIe Root complex resources */
typedef struct pcie_bus_res {
	uint64_t	num_msis;	/* Number of MSIs */
	uint64_t	num_eqs;	/* Number of EQs */
	int		num_dvma_ranges; /* Number of DVMA ranges */
	pcie_range_t	*dvma_ranges;	/* Pointer to the DVMA ranges */
	pcie_range_t	msi_range;	/* MSI range */
} pcie_bus_res_t;


/* MSI Bucket related info */
typedef struct msib_info {
	int		start;		/* starting MSI number */
	int		size;		/* Number of MSIs in this bucket. */
	bool_t		free;		/* free/not-free for allocation */
} msib_info_t;

/* MSI bucket size */
#define	MSIB_BSIZE	8

/* PCIe Root Complex resource management info */
typedef struct pcie_bus_res_mgmt {
	int		resv_msi_slots;	/* No. of slots reserved for RC owner */
	int		req_msi_per_dev; /* No. of MSIs required per device */
	msib_info_t	*msib;		/* MSI Buckets */
	int		num_msib;	/* Number of MSI buckets */

	int		resv_eq_slots;	/* No. of slots reserved for RC owner */
	int		req_eqs_per_dev; /* No. of Eqs required per device */
	uint64_t	num_eqs;	/* Total number of EQs */
	uint64_t	free_eqs;	/* Free Eqs available for allocation. */

	/*
	 * Currently, we don't sub-divide DVMA ranges.
	 * That is, the same range is given to all
	 * pcie_bus/virtual_pcie_bus nodes.
	 */
	int		num_dvma_ranges; /* Number of DVMA ranges */
	pcie_range_t	*dvma_ranges;	/* DVMA ranges */
} pcie_bus_res_mgmt_t;


bool_t pcie_is_dio_hvmd_supported(void);
bool_t pcie_gen_dio_hvmd(void);
bool_t pcie_is_rc_removable(ldom_t *oldp, bound_iodevice_t *bio);
bool_t pcie_is_rc_owner(ldom_t *ldp, bound_iodevice_t *bio);

void add_pcie_props1(md_t *mdp, ldom_t *ldp,
    md_node_t *node, bound_iodevice_t *bio);
void add_pcie_props2(md_t *mdp, ldom_t *ldp,
    md_node_t *node, bound_iodevice_t *bio);

void pcie_fix_guestmd(ldom_t *ldp, md_t *mdp, md_node_t **ionodelist, int num);
void pcie_display_devs(char *fmt, char *sub_fmt, response_t *resp);
void pcie_get_devinfo(ldom_t *ldp);
void pcie_realloc_res(bound_iodevice_t *bio);
void pcie_init_res_mgmt(pri_iodevice_t *iop);
void pcie_init_res_heuristics(void);
void pcie_destroy_bound_res(bound_iodevice_t *bio);
void pcie_alloc_all_res(bound_iodevice_t *bio);
void pcie_assign_all_devs(bound_iodevice_t *bio);
void pcie_alloc_fixed(bound_iodevice_t *bio, pcie_bus_res_t *busresp);
void pcie_free_res(bound_iodevice_t *bio);
void pcie_unassign_all_devs(ldom_t *ldp, bound_iodevice_t *bio);
void pcie_destroy_bound_pciedev(bound_pcie_dev_t *bdev);
void pcie_destroy_req_pciedev(req_pcie_dev_t *rdev);
req_pcie_dev_t *pcie_build_req_pciedev(char *name, bool_t bound);
bound_pcie_dev_t *pcie_build_bound_pciedev(pri_pcie_dev_t *pdev);
pri_pcie_dev_t *pcie_find_pri_pciedev(char *name, pri_iodevice_t **piop);
bound_pcie_dev_t *pcie_find_bound_pciedev(ldom_t *ldp,
    char *name, bound_iodevice_t **bio);
pri_pcie_dev_t *pcie_find_matching_dev(char *name);
bound_iodevice_t *pcie_biodevice_exists(ldom_t *ldp, uint64_t cfg_handle);
bool_t has_io_dependencies(ldom_t *ldp);
bool_t pcie_is_device_removable(ldom_t *ldp, pri_pcie_dev_t *pdev);
void pcie_display_io(char *fmt, response_t *resp);
bool_t pcie_match_pciedev(char *name, pri_pcie_dev_t *pdev);
char *pcie_build_skel_path(uint64_t cfghdl, char *rel_path);
void pcie_pri_get_io_slots(md_t *mdp);
void pcie_pri_get_resources(md_t *mdp, md_node_t *nodep, pri_iodevice_t *iop);
char *pcie_get_listio_format(void);
void pcie_update_reqs(ldom_t *ldp, bound_iodevice_t *biop);

#ifdef	__cplusplus
}
#endif

#endif	/* _PCIE_H */
