/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)platsvc.c	1.19	09/02/09 SMI"

/*
 * Module implementing the 'platsvc' DS protocol
 */

#include <strings.h>
#include <sys/types.h>
#include <unistd.h>

#include "allocate.h"
#include "fatal.h"
#include "ldom.h"
#include "ds.h"
#include "ds_impl.h"
#include "ds_registry.h"
#include "sequencer.h"
#include "zprintf.h"

#define	MD_UPDATE_DS_TIMEOUT		(60)
#define	DOMAIN_SHUTDOWN_DS_TIMEOUT	(10)
#define	DOMAIN_PANIC_DS_TIMEOUT		(10)

static uint64_t md_update_seq = 0;
static uint64_t domain_shutdown_seq = 0;
static uint64_t domain_panic_seq = 0;

/* MD update request */
typedef struct {
	uint64_t	seq_no;
} md_update_req_t;

/* MD update response */
typedef struct {
	uint64_t	seq_no;
	uint32_t	status;
} md_update_resp_t;

/*
 * MD update status codes
 */
#define	MD_UPDATE_SUCCESS	0x0
#define	MD_UPDATE_FAILURE	0x1
#define	MD_UPDATE_INVALID_MSG	0x2

/*
 * Domain shutdown request
 */
typedef struct {
	uint64_t	seq_no;		/* sequence number */
	uint32_t	ms_delay;	/* delay in milliseconds */
} domain_shutdown_req_t;

/*
 * Domain shutdown response
 */
typedef struct {
	uint64_t	seq_no;		/* sequence number */
	uint32_t	status;
	char		reason[1];	/* Optional */
} domain_shutdown_resp_t;

/*
 * Domain shutdown status
 */
#define	DOMAIN_SHUTDOWN_SUCCESS		0x0
#define	DOMAIN_SHUTDOWN_FAILURE		0x1
#define	DOMAIN_SHUTDOWN_INVALID_MSG	0x2

/*
 * Domain panic request
 */
typedef struct {
	uint64_t	seq_no;		/* sequence number */
} domain_panic_req_t;

/*
 * Domain panic response
 */
typedef struct {
	uint64_t	seq_no;		/* sequence number */
	uint32_t	status;
	char		reason[1];	/* Optional */
} domain_panic_resp_t;

/*
 * Domain panic status
 */
#define	DOMAIN_PANIC_SUCCESS		0x0
#define	DOMAIN_PANIC_FAILURE		0x1
#define	DOMAIN_PANIC_INVALID_MSG	0x2

static ds_ver_t platsvc_vers[] = { {1, 0} };
#define	PLATSVC_NVERS  (sizeof (platsvc_vers) / sizeof (platsvc_vers[0]))

static ds_capability_t md_update_cap = {
	MD_UPDATE_SVC,	/* svc_id */
	platsvc_vers,	/* vers */
	PLATSVC_NVERS	/* nvers */
};

static ds_ops_t md_update_libops = {
	dslib_register_handler,		/* ds_reg_cb */
	dslib_unregister_handler,	/* ds_unreg_cb */
	dslib_data_handler,		/* ds_data_cb */
	MD_UPDATE_SVC		/* cb_arg */
};

static ds_capability_t domain_shutdown_cap = {
	DOMAIN_SHUTDOWN_SVC,	/* svc_id */
	platsvc_vers,	/* vers */
	PLATSVC_NVERS	/* nvers */
};

static ds_ops_t domain_shutdown_libops = {
	dslib_register_handler,		/* ds_reg_cb */
	dslib_unregister_handler,	/* ds_unreg_cb */
	dslib_data_handler,		/* ds_data_cb */
	DOMAIN_SHUTDOWN_SVC		/* cb_arg */
};

static ds_capability_t domain_panic_cap = {
	DOMAIN_PANIC_SVC,	/* svc_id */
	platsvc_vers,	/* vers */
	PLATSVC_NVERS	/* nvers */
};

static ds_ops_t domain_panic_libops = {
	dslib_register_handler,		/* ds_reg_cb */
	dslib_unregister_handler,	/* ds_unreg_cb */
	dslib_data_handler,		/* ds_data_cb */
	DOMAIN_PANIC_SVC		/* cb_arg */
};

/*
 * Info needed to invoke ds_clnt_reg/ds_svc_reg when loading plugins
 * There could be multiple services in the same info structure
 */
static libds_info_t platsvc_info[] = {
	{
		DS_CLIENT,
		&md_update_cap,
		&md_update_libops
	},
	{
		DS_CLIENT,
		&domain_shutdown_cap,
		&domain_shutdown_libops
	},
	{
		DS_CLIENT,
		&domain_panic_cap,
		&domain_panic_libops
	}
};

/*
 * Notify a guest to re-read the new MD
 */
int
platsvc_update_md(ldom_t *ldp)
{
	ds_msg_t msg, resp;
	ds_svc_t *svcp;
	md_update_req_t *md_reqp;
	md_update_resp_t *md_resp;
	ds_req_t *ds_reqp;
	size_t msg_len;
	int rv;

	svcp = ds_lookup_svc(ldp->ds_chan, MD_UPDATE_SVC);
	if ((svcp == NULL) || (svcp->hdl == DS_SVC_HDL_INVALID)) {
		DBG(zprintf("%s %s not yet registered\n",
		    ldp->namep, MD_UPDATE_SVC));
		return (-1);
	}

	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) +
	    sizeof (md_update_req_t);
	ds_reqp = Xmalloc(msg_len);
	DBG7(memset(ds_reqp, 0, msg_len));

	ds_fill_data_header(ds_reqp, svcp->hdl, sizeof (md_update_req_t));

	md_reqp = (md_update_req_t *)DS_DATA_BUF(ds_reqp);
	md_reqp->seq_no = hton64(md_update_seq++);

	msg.buf = ds_reqp;
	msg.len = msg_len;
	msg.seq = md_reqp->seq_no;

	DBG(zprintf("Sending MD update message to guest %s, hdl %llx, "
	    "seq %llu\n", ldp->namep, svcp->hdl, ntoh64(msg.seq)));

	rv = ds_send_data_msg(ldp->ds_chan, &msg, MD_UPDATE_DS_TIMEOUT, &resp);
	Xfree(ds_reqp);

	if (rv == 0) {
		md_resp = (md_update_resp_t *)((char *)resp.buf +
		    sizeof (ds_data_handle_t));

		if (ntoh32(md_resp->status) != MD_UPDATE_SUCCESS) {
			DBG(zprintf("MD update failed, status %d\n",
			    ntoh32(md_resp->status)));
			rv = -1;
		}

		Xfree(resp.buf);
	}

	return (rv);
}

/*
 * Send LDom a shutdown message.  Returns -1 if there was an error
 * communicating with the LDom.  Returns 0 if the LDom acked the shutdwon and
 * >0 if the LDom nacked the shutdown.
 */
int
platsvc_send_shutdown(ldom_t *ldp)
{
	ds_msg_t msg, resp;
	ds_svc_t *svcp;
	domain_shutdown_req_t *shutdown_reqp;
	domain_shutdown_resp_t *shutdown_resp;
	ds_req_t *ds_reqp;
	size_t msg_len;
	int rv;

	svcp = ds_lookup_svc(ldp->ds_chan, DOMAIN_SHUTDOWN_SVC);
	if ((svcp == NULL) || (svcp->hdl == DS_SVC_HDL_INVALID)) {
		DBG(zprintf("%s %s not yet registered\n",
		    ldp->namep, DOMAIN_SHUTDOWN_SVC));
		return (-1);
	}

	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) +
	    sizeof (domain_shutdown_req_t);
	ds_reqp = Xmalloc(msg_len);
	DBG7(memset(ds_reqp, 0, msg_len));

	ds_fill_data_header(ds_reqp, svcp->hdl, sizeof (domain_shutdown_req_t));

	shutdown_reqp = (domain_shutdown_req_t *)DS_DATA_BUF(ds_reqp);
	shutdown_reqp->seq_no = hton64(domain_shutdown_seq++);
	shutdown_reqp->ms_delay = hton32(0); /* FIXME once support in Solaris */

	msg.buf = ds_reqp;
	msg.len = msg_len;
	msg.seq = shutdown_reqp->seq_no;

	DBG(zprintf("Sending shutdown request to guest %s, hdl %llx, "
	    "seq %llu\n", ldp->namep, svcp->hdl, ntoh64(msg.seq)));

	/*
	 * Send request and wait for the 'initiation of shutdown'
	 * response. The shutdown request could be denied by the guest.
	 */
	rv = ds_send_data_msg(ldp->ds_chan, &msg, DOMAIN_SHUTDOWN_DS_TIMEOUT,
	    &resp);
	Xfree(ds_reqp);

	if (rv == 0) {
		shutdown_resp = (domain_shutdown_resp_t *)((char *)resp.buf +
		    sizeof (ds_data_handle_t));

		rv = ntoh32(shutdown_resp->status);
		if (rv != DOMAIN_SHUTDOWN_SUCCESS) {
			warning("Shutdown of guest \"%s\" failed: %d with "
			    "reason %s", ldp->namep, rv, shutdown_resp->reason);
		}

		Xfree(resp.buf);
	}

	return (rv);
}

/*
 * Send ldom a panic message.
 */
int
platsvc_send_panic(ldom_t *ldp)
{
	ds_msg_t msg, resp;
	ds_svc_t *svcp;
	domain_panic_req_t *panic_reqp;
	domain_panic_resp_t *panic_resp;
	ds_req_t *ds_reqp;
	size_t msg_len;
	int rv = 0;

	svcp = ds_lookup_svc(ldp->ds_chan, DOMAIN_PANIC_SVC);
	if ((svcp == NULL) || (svcp->hdl == DS_SVC_HDL_INVALID)) {
		DBG(zprintf("%s %s not yet registered\n",
		    ldp->namep, DOMAIN_PANIC_SVC));
		return (-1);
	}

	msg_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) +
	    sizeof (domain_panic_req_t);
	ds_reqp = Xmalloc(msg_len);
	DBG7(memset(ds_reqp, 0, msg_len));

	ds_fill_data_header(ds_reqp, svcp->hdl, sizeof (domain_panic_req_t));

	panic_reqp = (domain_panic_req_t *)DS_DATA_BUF(ds_reqp);
	panic_reqp->seq_no = hton64(domain_panic_seq++);

	msg.buf = ds_reqp;
	msg.len = msg_len;
	msg.seq = panic_reqp->seq_no;

	DBG(zprintf("Sending panic request to guest %s, hdl %llx, "
	    "seq %llu\\n", ldp->namep, svcp->hdl, ntoh64(msg.seq)));

	/*
	 * Send request and wait for the 'initiation of panic' response. The
	 * panic request could be denied by the guest.
	 */
	rv = ds_send_data_msg(ldp->ds_chan, &msg, DOMAIN_PANIC_DS_TIMEOUT,
	    &resp);
	Xfree(ds_reqp);

	if (rv == 0) {
		panic_resp = (domain_panic_resp_t *)((char *)resp.buf +
		    sizeof (ds_data_handle_t));

		rv = ntoh32(panic_resp->status);
		if (rv != DOMAIN_PANIC_SUCCESS) {
			warning("Panic of guest \"%s\" failed: %d with "
			    "reason %s", ldp->namep, rv, panic_resp->reason);
		}

		Xfree(resp.buf);
	}

	return (rv);
}

static void
md_update_reg_handler(ds_svc_t *svc)
{
	if (svc->chan->ldp->flag.md_notify)
		seq_enqueue_md_notify(svc);
}

static void
md_update_data_handler(ds_svc_t *svc, void *buf, size_t msglen)
{
	ds_msg_tgt_t		tgt;
	ds_msg_t		resp;
	md_update_resp_t	*psmsgp;

	if (msglen != sizeof (ds_data_handle_t) + sizeof (md_update_resp_t)) {
		DBG(zprintf("Invalid length %s response message, %d bytes, "
		    "expected %d\n", MD_UPDATE_SVC, msglen,
		    sizeof (ds_data_handle_t) + sizeof (md_update_resp_t)));
		Xfree(buf);
		return;
	}

	tgt.chan = svc->chan;
	tgt.ds_hdl = svc->hdl;

	psmsgp = (md_update_resp_t *)((char *)buf + sizeof (ds_data_handle_t));

	resp.buf = buf;
	resp.len = msglen;
	resp.seq = ntoh64(psmsgp->seq_no);

	if (ds_check_resp(&tgt, &resp) == DS_MSG_CLAIMED) {
		/* message already handled */
		return;
	}

	DBG(zprintf("Discarding unclaimed %s message, seq %llu\n",
	    MD_UPDATE_SVC, resp.seq));
	Xfree(buf);
}

static void
domain_shutdown_data_handler(ds_svc_t *svc, void *buf, size_t msglen)
{
	ds_msg_tgt_t		tgt;
	ds_msg_t		resp;
	domain_shutdown_resp_t	*psmsgp;

	if (msglen != sizeof (ds_data_handle_t) +
	    sizeof (domain_shutdown_resp_t)) {
		DBG(zprintf("Invalid length %s response message, "
		    "%d bytes, expected %d\n", DOMAIN_SHUTDOWN_SVC, msglen,
		    sizeof (ds_data_handle_t) +
		    sizeof (domain_shutdown_resp_t)));
		Xfree(buf);
		return;
	}

	tgt.chan = svc->chan;
	tgt.ds_hdl = svc->hdl;

	psmsgp = (domain_shutdown_resp_t *)((char *)buf +
	    sizeof (ds_data_handle_t));

	resp.buf = buf;
	resp.len = msglen;
	resp.seq = ntoh64(psmsgp->seq_no);

	if (ds_check_resp(&tgt, &resp) == DS_MSG_CLAIMED) {
		/* message already handled */
		return;
	}

	DBG(zprintf("Discarding unclaimed %s message, seq %llu\n",
	    DOMAIN_SHUTDOWN_SVC, resp.seq));
	Xfree(buf);
}

static void
domain_panic_data_handler(ds_svc_t *svc, void *buf, size_t msglen)
{
	ds_msg_tgt_t		tgt;
	ds_msg_t		resp;
	domain_panic_resp_t	*psmsgp;

	if (msglen != sizeof (ds_data_handle_t) +
	    sizeof (domain_panic_resp_t)) {
		DBG(zprintf("Invalid length %s response message, "
		    "%d bytes, expected %d\n", DOMAIN_PANIC_SVC, msglen,
		    sizeof (ds_data_handle_t) + sizeof (domain_panic_resp_t)));
		Xfree(buf);
		return;
	}

	tgt.chan = svc->chan;
	tgt.ds_hdl = svc->hdl;

	psmsgp = (domain_panic_resp_t *)((char *)buf +
	    sizeof (ds_data_handle_t));

	resp.buf = buf;
	resp.len = msglen;
	resp.seq = ntoh64(psmsgp->seq_no);

	if (ds_check_resp(&tgt, &resp) == DS_MSG_CLAIMED) {
		/* message already handled */
		return;
	}

	DBG(zprintf("Discarding unclaimed %s message, seq %llu\n",
	    DOMAIN_PANIC_SVC, resp.seq));
	Xfree(buf);
}

/*
 * The versions supported
 */
static prot_version_t platsvc_versions[] =
{
	{1, 0}
};

static svc_ops_t md_update_ops = {
	md_update_reg_handler,		/* reg_handler */
	NULL,				/* unreg_handler */
	md_update_data_handler,		/* data_handler */
	NULL,				/* seq_handler */
	NULL				/* data */
};

static svc_ops_t domain_shutdown_ops = {
	NULL,				/* reg_handler */
	NULL,				/* unreg_handler */
	domain_shutdown_data_handler,	/* data_handler */
	NULL,				/* seq_handler */
	NULL				/* data */
};

static svc_ops_t domain_panic_ops = {
	NULL,				/* reg_handler */
	NULL,				/* unreg_handler */
	domain_panic_data_handler,	/* data_handler */
	NULL,				/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	int i;
	svc_info_t *info;

	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 2;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(DS_CHAN_NAME);
	info->chan_names[1] = Xstrdup(LIBDS_CHAN_NAME);
	info->num_svcs = 3;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(MD_UPDATE_SVC);
	info->svcs[1] = Xstrdup(DOMAIN_SHUTDOWN_SVC);
	info->svcs[2] = Xstrdup(DOMAIN_PANIC_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = 0;
	info->props[1] = 0;
	info->props[2] = 0;
	info->num_vers = NUM_VERS(platsvc_versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	memcpy(info->versions, platsvc_versions,
	    info->num_vers * sizeof (prot_version_t));

	/*
	 * copy the info structure for each service
	 */
	info->libds_svcs = Xmalloc(info->num_svcs * sizeof (libds_info_t *));
	for (i = 0; i < info->num_svcs; i++)
		info->libds_svcs[i] = cp_dsinfo(&platsvc_info[i],
		    info->svcs[i]);

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, MD_UPDATE_SVC) == 0)
		return (&md_update_ops);
	else if (strcmp(svc_name, DOMAIN_SHUTDOWN_SVC) == 0)
		return (&domain_shutdown_ops);
	else if (strcmp(svc_name, DOMAIN_PANIC_SVC) == 0)
		return (&domain_panic_ops);
	else
		return (NULL);
}
