/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include "ldmd_types.h"
#include "mdesc_impl.h"
#include "pm_shared.h"
#include "zpri.h"
#include "zprintf.h"
#include "vcpu.h"
#include "sequencer.h"
#include "pm_events.h"
#include "pm/pm_caps.h"

#define	DEFRAG_MOVE_STRAND_SUCCESS	(0)
#define	DEFRAG_MOVE_STRAND_ERR_SOURCE	(1)
#define	DEFRAG_MOVE_STRAND_ERR_DEST	(2)
#define	DEFRAG_MOVE_STRAND_ERR_SPARE	(3)
#define	DEFRAG_MOVE_STRAND_ERR_CONFIG	(4)
#define	RES_NUM				1
#define	MOVE_NUM			2
#define	INVALID_CPUID			-1

#define	DEFRAG_CPU_CORE_LIMIT		1

typedef struct defrag_core_stat {
	int		dcs_cpu_on;	/* # of powered-on strands */
	int		dcs_cpu_off;	/* # of powered-off strands */
	int		dcs_cpu_fault;	/* # of strands that are FMA faulted */
	bool_t		dcs_fail_on;	/* has strand power-on failed? */
	bool_t		dcs_fail_off;	/* has strand power-off failed? */
	pri_core_t	*dcs_corep;	/* ptr to associated core */
	ldom_t		*dcs_ldp;	/* ldom with exclusive core use */
} defrag_core_stat_t;

static bool_t	defrag_running = false;

/* A core is considered powered-on if it has at least one powered-on strand */
#define	IS_CPU_CORE_ON(cstat)	(cstat.dcs_cpu_on > 0) ? true : false


/*
 * _find_pm_handle - Finds the PM event handle.
 *
 * Returns the event handle or NULL.
 */
static void *
_find_pm_handle(void)
{
	pm_domain_event_info_t	*event = NULL;

	CHAIN_ITER(pm_domain_list.domain_event_list, event, link) {
		if ((event->event_type == PM_RES_EVENT) ||
		    (event->res_type == PM_RES_CPU)) {
			return ((void *)event->handle);
		}
	}

	return (NULL);
}

/*
 * Set function pointer globally so that it can be overridden by the
 * test harness.
 */
/* LINTED E_GLOBAL_COULD_BE_STATIC */
void *(*find_pm_handle)(void) = _find_pm_handle;


/*
 * pmi_defrag_move_strand - Called on the sequencer.
 * Changes the domain containing powered up physical strand 'from'
 * to contain physical strand 'to.'  If strand 'spare' is > cpu id 0
 * then the 2nd of the three use cases will be applied, providing that
 * DEFRAG_USE_SPARE is defined and 'spare' is powered up.
 *
 * In all cases 'from' must initially be powered up and 'to' & 'spare',
 * if not INVALID_CPUID (-1), must initially be powered down.  After the
 * operation, 'from' will be powered down and 'to' will be powered up.
 *
 * Use cases:
 * 'spare' is always unallocated.
 * 1) 'from' is allocated to a domain but 'to' is unallocated and 'spare'
 *         is unused.
 *    'to' is DR'd into the domain containing 'from' and powered up
 *    'from' is DR'd out of domain and powered down.
 * 2) 'from' is allocated to domain A, 'to' is allocated to domain B
 *    'to' is already powered up but needs to move
 *    'spare' is powered up and DR'd into domain B
 *    'to' is DR'd out of domain B
 *    'to' is DR'd into domain A
 *    'from' is DR'd out of domain A
 *    'from' is powered down
 * 3) 'from' and 'to' exist in the same domain but 'spare' is unused.
 *    'to' is powered up
 *    'from' is powered down
 *
 * Returns distinct error codes for failures or PS_SUCCESS.
 *
 * NOTE: This function is also called by pm_testlib (not part of ldmd)
 */
/* LINTED E_GLOBAL_COULD_BE_STATIC */
int
pmi_defrag_move_strand(uint64_t from, uint64_t to, uint64_t spare)
{
	uint16_t res_state[DEFRAG_CPU_TOTAL];
	cpu_adm_t *from_cpu = NULL, *to_cpu = NULL;
	pm_resid_t res_id[DEFRAG_CPU_TOTAL];
	int ret = DEFRAG_MOVE_STRAND_SUCCESS;
	bool_t from_off, to_off;
	ldom_t *ldp = NULL, *ldp_from = NULL, *ldp_to = NULL;
	ldm_err_t chg_status;
	srcid_t *dom_ids = NULL;
	vcpu_t *vcpu;
	void *handle = NULL;
	int res_count = MOVE_NUM;
	response_t *resp = NULL;
#ifdef DEFRAG_USE_SPARE
	cpu_adm_t *spare_cpu = NULL;
	bool_t spare_off;
	ldom_t *ldp_spare = NULL;
#endif

	/*
	 * Make sure cpu ids are valid and that we're not in management
	 * or delayed reconfig modes.
	 */
	ASSERT(from >= 0);
	ASSERT(to >= 0);
	if (zeus.mode != ZEUS_MGMT_MODE)
		return (DEFRAG_MOVE_STRAND_ERR_CONFIG);

	if (in_delayed_reconfig(NULL)) {
		DBG(zprintf("%s: No defragment while delayed "
		    "reconfiguration in progress.\n", __func__));
		ret = DEFRAG_MOVE_STRAND_ERR_CONFIG;
		goto fail;
	}

#ifndef DEFRAG_USE_SPARE
	/*
	 * If DEFRAG_USE_SPARE is not defined, spare should be ==
	 * INVALID_CPUID.
	 */
	if (spare != INVALID_CPUID)
		return (DEFRAG_MOVE_STRAND_ERR_SPARE);
#endif

	DBG(zprintf("%s: from cpu (%lld), to cpu (%lld), spare cpu (%lld)\n",
	    __func__, from, to, spare));

	/* Find current power states */
#ifdef DEFRAG_USE_SPARE
	res_id[DEFRAG_CPU_SPARE].u.cpu.pi_id_val = spare;
	res_id[DEFRAG_CPU_SPARE].pi_res_type = PM_RES_CPU;
	res_id[DEFRAG_CPU_SPARE].pi_res_name = CPU_RES_NAME;
	res_count = MOVE_NUM + 1;
#endif
	res_id[DEFRAG_CPU_FROM].u.cpu.pi_id_val = from;
	res_id[DEFRAG_CPU_FROM].pi_res_type = PM_RES_CPU;
	res_id[DEFRAG_CPU_FROM].pi_res_name = CPU_RES_NAME;
	res_id[DEFRAG_CPU_TO].u.cpu.pi_id_val = to;
	res_id[DEFRAG_CPU_TO].pi_res_type = PM_RES_CPU;
	res_id[DEFRAG_CPU_TO].pi_res_name = CPU_RES_NAME;
	if (pm_get_defrag_power_states(res_id, res_state,
	    res_count) != PS_SUCCESS) {
		return (DEFRAG_MOVE_STRAND_ERR_CONFIG);
	}

#ifdef DEFRAG_USE_SPARE
	spare_cpu = Xmalloc(sizeof (*spare_cpu) + sizeof (int));
	spare_cpu->type = CPU_ADM_IDS;
	spare_cpu->num = RES_NUM;
	spare_cpu->ids[0] = spare;
	spare_off = (res_state[DEFRAG_CPU_SPARE] == PPS_LOWEST_POWER);
	ldp_spare = find_ldom_by_resid(spare);
#endif
	/* Which resources are PM'd? */
	from_off = (res_state[DEFRAG_CPU_FROM] == PPS_LOWEST_POWER);
	to_off = (res_state[DEFRAG_CPU_TO] == PPS_LOWEST_POWER);
	DBG6(zprintf("%s: \"from\" cpu off? (%d), \"to\" cpu off? (%d)\n",
	    __func__, (int)from_off, (int)to_off));

	/* Find out which domain(s) resources belong to. */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
			if (vcpu->cpu->phys_id == from) {
				ldp_from = ldp;
			} else if (vcpu->cpu->phys_id == to) {
				ldp_to = ldp;
#ifdef DEFRAG_USE_SPARE
			} else if (vcpu->cpu->phys_id == spare) {
				DBG(zprintf("%s: Spare cpu should not be "
				    "allocated to a domain yet.\n", __func__));
				ret = DEFRAG_MOVE_STRAND_ERR_SPARE;
				goto fail;
#endif
			}
		}

#ifndef DEFRAG_USE_SPARE
		/* Fail if we run into use case 2 but it isn't supported. */
		if (ldp_from && ldp_to && (ldp_from != ldp_to)) {
			DBG(zprintf("%s: \"Spare\" cpu usage not supported.\n",
			    __func__));
			ret = DEFRAG_MOVE_STRAND_ERR_SPARE;
			goto fail;
		}
#endif
		/* Found both so get out now */
		if (ldp_from && ldp_to)
			break;
	}

	if (ldp_from == NULL) {
		DBG(zprintf("%s: Search for domain for \"from\" cpu"
		    " failed.\n", __func__));
		ret = DEFRAG_MOVE_STRAND_ERR_CONFIG;
		goto fail;
	}

	/* From resource must already be powered up */
	if (from_off) {
		DBG(zprintf("%s: \"from\" cpu should not already be "
		    "powered down.\n", __func__));
		ret = DEFRAG_MOVE_STRAND_ERR_CONFIG;
		goto fail;
	}

	/* "To" resource must already be powered down */
	if (!to_off) {
		DBG(zprintf("%s: \"To\" resource must already be "
		    "powered down.\n", __func__));
		ret = DEFRAG_MOVE_STRAND_ERR_CONFIG;
		goto fail;
	}

	/* Find event handle. */
	handle = find_pm_handle();
	if (handle == NULL) {
		ret = DEFRAG_MOVE_STRAND_ERR_CONFIG;
		goto fail;
	}

	/*
	 * If the domains for "from" and "to" cpus are the same, follow
	 * use case 3.  Otherwise, follow use cases 1 & 2.
	 */
	if (ldp_from == ldp_to) {
		/* Power up target, allocated resource */
		DBG6(zprintf("%s: power up cpu %"PRIu64"\n", __func__,
		    res_id[DEFRAG_CPU_TO].u.cpu.pi_id_val));

		pm_domain_config_change_notify(handle, (srcid_t)ldp_to,
		    ADD_RESOURCE_DEFRAG, PM_RES_CPU, &res_id[DEFRAG_CPU_TO],
		    RES_NUM, &chg_status);

		if (chg_status.ldme_errs != LDME_OP_EOK) {
			DBG(zprintf("%s: Power up of cpu %"PRIu64" failed\n",
			    __func__, res_id[DEFRAG_CPU_TO].u.cpu.pi_id_val));
			ret = DEFRAG_MOVE_STRAND_ERR_SOURCE;
			pm_free_errors(&chg_status);
			goto fail;
		}

		pm_free_errors(&chg_status);

		/* Power down source CPU */
		DBG6(zprintf("%s: Power down cpu %"PRIu64"\n", __func__,
		    res_id[DEFRAG_CPU_FROM].u.cpu.pi_id_val));

		pm_domain_config_change_notify(handle,
		    (srcid_t)ldp_from, REM_RESOURCE_DEFRAG, PM_RES_CPU,
		    &res_id[DEFRAG_CPU_FROM], RES_NUM, &chg_status);

		if (chg_status.ldme_errs != LDME_OP_EOK) {
			DBG(zprintf("%s: Power down of cpu %"PRIu64" failed\n",
			    __func__, res_id[DEFRAG_CPU_FROM].u.cpu.pi_id_val));
			ret = DEFRAG_MOVE_STRAND_ERR_SOURCE;
			pm_free_errors(&chg_status);
			goto fail;
		}

		pm_free_errors(&chg_status);
	} else {
		/* Create and fill in DR structures */
		from_cpu = Xmalloc(sizeof (*from_cpu) + sizeof (int));
		to_cpu = Xmalloc(sizeof (*to_cpu) + sizeof (int));

		from_cpu->ids[0] = (uint32_t)from;
		from_cpu->type = CPU_ADM_IDS;
		from_cpu->num = RES_NUM;

		to_cpu->ids[0] = (uint32_t)to;
		to_cpu->type = CPU_ADM_IDS;
		to_cpu->num = RES_NUM;

#ifndef DEFRAG_USE_SPARE
		/* Use case 1 only. Make sure "to" cpu is unallocated. */
		if (ldp_to) {
			ret = DEFRAG_MOVE_STRAND_ERR_CONFIG;
			goto fail;
		}
#else
		/* DR spare CPU into source domain */
		DBG6(zprintf("%s: Dynamically reconfiguring in the "
		    "spare CPU.\n", __func__));

		if (!cpu_add(ldp_from, spare_cpu, resp)) {
			DBG(zprintf("%s: Unable to dynamically reconfigure "
			    "in the spare cpu\n", __func__));
			ret = DEFRAG_MOVE_STRAND_ERR_SPARE;
			goto fail;
		}

		/* DR target CPU out of source domain */
		DBG6(zprintf("%s: Dynamically reconfiguring target CPU "
		    "out of source domain.\n", __func__));

		if (!cpu_del(ldp_from, to_cpu, resp, false)) {
			DBG(zprintf("%s: Unable to dynamically reconfigure "
			    "out the target cpu.\n", __func__));
			ret = DEFRAG_MOVE_STRAND_ERR_DEST;
			goto fail;
		}
#endif

		/* Use cases 1 and 2 */
		/* DR "to" cpu into target domain. */
		DBG6(zprintf("%s: Dynamically reconfiguring CPU %d "
		    "into target domain.\n", __func__, to_cpu->ids[0]));

		if (!cpu_add(ldp_from, to_cpu, resp)) {
			DBG(zprintf("%s: Unable to dynamically reconfigure "
			    "in cpu %d.\n", __func__, to_cpu->ids[0]));
			ret = DEFRAG_MOVE_STRAND_ERR_DEST;
			goto fail;
		}

		/*
		 * DR "from" cpu out of domain. If unsuccessful,
		 * DR "to" cpu back out of target domain.
		 */
		DBG6(zprintf("%s: Dynamically reconfiguring CPU %"PRIu64" "
		    "out of domain %s.\n", __func__, from, ldp_from->namep));

		if (!cpu_del(ldp_from, from_cpu, resp, false)) {
			DBG(zprintf("%s: Unable to dynamically reconfigure "
			    "out cpu %"PRIu64".\n", __func__, from));
			if (!cpu_del(ldp_from, to_cpu, resp, false)) {
				DBG(zprintf("%s: Unable to dynamically "
				    "reconfigure back out cpu %d.\n",
				    __func__, to_cpu->ids[0]));
			}
			ret = DEFRAG_MOVE_STRAND_ERR_SOURCE;
			goto fail;
		}
	}

	ret = PS_SUCCESS;
fail:
#ifdef DEFRAG_USE_SPARE
	Xfree(spare_cpu);
#endif
	Xfree(dom_ids);
	Xfree(from_cpu);
	Xfree(to_cpu);
	DBG(zprintf("%s: return == %d\n", __func__, ret));
	return (ret);
}

/*
 * Fill in a given defrag_core_stat_t struct based on a given pri_core_t struct.
 * Routine now does a core worth of cpus in a single HV call instead of 1 HV
 * call per cpu.
 */
static void
defrag_fill_core_stat(pri_core_t *corep, defrag_core_stat_t *cstat)
{
	pri_cpu_t *cpup;
	uint16_t *res_state = NULL;
	pm_resid_t *res_id = NULL;
	bool_t shared_core = false;
	ldom_t  *ldp = NULL;
	int	vid;
	int	count = 0;

	cstat->dcs_cpu_on = 0;
	cstat->dcs_cpu_off = 0;
	cstat->dcs_cpu_fault = 0;
	cstat->dcs_fail_on = false;
	cstat->dcs_fail_off = false;
	cstat->dcs_corep = corep;
	cstat->dcs_ldp = NULL;

	/* Figure out how many cpus there are in core */
	CHAIN_ITER(corep->cpus, cpup, core_link) {
		count++;
	}

	res_state = Xcalloc(count, uint16_t);
	res_id = Xcalloc(count, pm_resid_t);

	count = 0;

	/* Find all pcpus related to this core */
	CHAIN_ITER(corep->cpus, cpup, core_link) {
		res_id[count].u.cpu.pi_id_val = cpup->phys_id;
		res_id[count].pi_res_type = PM_RES_CPU;

		/* Find LDom associated with physical cpu id */
		if (!pid_to_vid(cpup->phys_id, &ldp, &vid))
			continue;

		if (ldp) {
			/* If LDoms are different then this core is shared */
			if (cstat->dcs_ldp && (cstat->dcs_ldp != ldp)) {
				/* Ignore shared cores */
				cstat->dcs_ldp = NULL;
				shared_core = true;
			}
			/* Not shared so save LDoms which has cpu on core */
			if (!shared_core)
				cstat->dcs_ldp = ldp;
		}

		/* Found one */
		count++;
	}

	/* Do core worth of calls to ldm_get_power_states */
	if ((count > 0) && (pm_get_defrag_power_states(res_id, res_state,
	    count)) == PS_SUCCESS) {
		int i;

		/* Post process */
		for (i = 0; i < count; i++) {
			/* Faulted */
			if (res_id[i].pi_fma_status != FMA_CPU_STATUS_ONLINE)
				cstat->dcs_cpu_fault++;

			/* Powered OFF */
			if (res_state[i] == PPS_LOWEST_POWER)
				cstat->dcs_cpu_off++;
			/* Has Performance */
			else
				cstat->dcs_cpu_on++;
		}
	}

	DBG6(zprintf("%s: core%d  on=%d  off=%d  fault=%d fail_on=%d  "
	    "fail_off=%d\n", __func__, corep->idx, cstat->dcs_cpu_on,
	    cstat->dcs_cpu_off, cstat->dcs_cpu_fault, (int)cstat->dcs_fail_on,
	    (int)cstat->dcs_fail_off));

	Xfree(res_state);
	Xfree(res_id);
}

/*
 * Figure out which core is the lowest-numbered core in the list.
 */
static int
defrag_get_lowest_core(defrag_core_stat_t *cores, int num_cores)
{
	int i;
	int low_idx = 0;

	for (i = 0; i < num_cores; i++) {
		if (cores[i].dcs_corep->idx < cores[low_idx].dcs_corep->idx)
			low_idx = i;
	}

	DBG6(zprintf("%s: lowest core is core%d (idx=%d)\n", __func__,
	    cores[low_idx].dcs_corep->idx, low_idx));

	return (low_idx);
}

/*
 * Calculate the total number of powered-off strands on the cores that
 * are powered-on.
 */
static int
defrag_get_total_off_strands(defrag_core_stat_t *cores, int num_cores)
{
	int i;
	int total_off = 0;

	for (i = 0; i < num_cores; i++) {
		if ((IS_CPU_CORE_ON(cores[i])) &&
		    !(cores[i].dcs_fail_on))
			total_off += cores[i].dcs_cpu_off;
	}

	DBG6(zprintf("%s: total number of powered-off strands on the cores"
	    " that are powered-on is %d\n", __func__, total_off));

	return (total_off);
}

/*
 * Determine if it is possible to defragment the CPU cores.
 */
static bool_t
is_defrag_cpu_possible(defrag_core_stat_t *cores, int num_cores, int low_idx)
{
	int i;
	int total_off;

	total_off = defrag_get_total_off_strands(cores, num_cores);

	/*
	 * Defragmentation is possible if the number of powered-on strands
	 * of a core is less than or equal to the total number of powered-off
	 * strands on the other powered-on cores.  If a core has failed to
	 * power-off a strand, skip it.  If a core has at least one faulted
	 * strand, skip it.
	 *
	 * Also, there is a constraint that each chip must have at least one
	 * core powered-on.  To keep things simple, we'll keep the
	 * lowest-numbered core powered-on.
	 */
	for (i = 0; i < num_cores; i++) {
		if ((IS_CPU_CORE_ON(cores[i])) &&
		    (i != low_idx) &&
		    !(cores[i].dcs_fail_off) &&
		    (cores[i].dcs_cpu_fault == 0) &&
		    (cores[i].dcs_cpu_on <= (total_off - cores[i].dcs_cpu_off)))
			return (true);
	}

	DBG6(zprintf("%s: FALSE\n", __func__));

	return (false);
}

/*
 * For defragmentation, find the index of the core to move strands/cpus from.
 *
 * Return the index of the core to move strands from, or
 * -1 if no core is available.
 */
static int
defrag_find_source_core(defrag_core_stat_t *cores, int num_cores, int dest_idx,
    int low_idx)
{
	int i;
	int source_idx = -1;

	/*
	 * Find the core that :
	 * 	a) is powered-on
	 * 	b) is NOT the core to move strands to (dest_idx)
	 * 	c) is NOT the lowest-numbered core (low_idx)
	 * 	d) has NOT failed to power-off a strand
	 *	e) has NO faulted strands
	 * 	f) has the least number of powered-on strands
	 *
	 * We'll start from the highest index and work backwards due to the
	 * constraint that one core needs to be powered-on, and we've selected
	 * this to be the lowest-numbered core (for simplicity).  More than
	 * likely, the lowest-numbered core in the array will be at index zero.
	 */
	for (i = num_cores - 1; i >= 0; i--) {
		if ((IS_CPU_CORE_ON(cores[i])) &&
		    (i != dest_idx) &&
		    (i != low_idx) &&
		    !(cores[i].dcs_fail_off) &&
		    (cores[i].dcs_cpu_fault == 0)) {
			if (source_idx == -1) {
				source_idx = i;	/* first one */
			} else if (cores[i].dcs_cpu_on <
			    cores[source_idx].dcs_cpu_on) {
				source_idx = i;	/* least powered-on strands */
			}
		}
	}

	DBG6(zprintf("%s: source_idx=%d\n", __func__, source_idx));

	return (source_idx);
}

/*
 * For defragmentation, find the index of the core to move strands/cpus to.
 *
 * Return the index of the core to move strands to, or
 * -1 if no core is available.
 */
static int
defrag_find_dest_core(defrag_core_stat_t *cores, int num_cores, int src_idx)
{
	int i;
	int dest_idx = -1;

	/*
	 * Find the core that :
	 * 	a) is powered-on
	 * 	b) is NOT the core to move strands from (src_idx)
	 * 	c) has not failed to power-on a strand
	 * 	d) has at least one powered-off strand
	 * 	e) has the least number of powered-off strands
	 *	f) source and dest core are from the same ldom
	 */
	for (i = 0; i < num_cores; i++) {

		if ((IS_CPU_CORE_ON(cores[i])) &&
		    (i != src_idx) &&
		    !(cores[i].dcs_fail_on) &&
		    (cores[i].dcs_cpu_off > 0) &&
		    (cores[i].dcs_ldp == cores[src_idx].dcs_ldp)) {
			if (dest_idx == -1) {
				dest_idx = i;	/* first one */
			} else if (cores[i].dcs_cpu_off <
			    cores[dest_idx].dcs_cpu_off) {
				dest_idx = i;	/* least powered-off strands */
			}
		}
	}

	DBG6(zprintf("%s: dest_idx=%d\n", __func__, dest_idx));

	return (dest_idx);
}

/*
 * Move a strand from src_core to a strand on dest_core.
 * Return zero for success, and non-zero for failure.
 */
static int
defrag_move_strand(defrag_core_stat_t *src_core, defrag_core_stat_t *dest_core)
{
	pri_cpu_t *src_cpup, *dest_cpup;
	int rv;
	uint16_t res_state[RES_NUM];
	pm_resid_t res_id[RES_NUM];

	/* find a strand on the source core to move from */
	CHAIN_ITER(src_core->dcs_corep->cpus, src_cpup, core_link) {
		res_id[0].u.cpu.pi_id_val = src_cpup->phys_id;
		res_id[0].pi_res_type = PM_RES_CPU;
		if (pm_get_defrag_power_states(res_id, res_state,
		    RES_NUM) == PS_SUCCESS) {
			if (res_id[0].pi_fma_status != FMA_CPU_STATUS_ONLINE) {
				DBG(zprintf("%s: cpu%d on source core is "
				    "faulted\n", __func__, src_cpup->phys_id));
				return (DEFRAG_MOVE_STRAND_ERR_SOURCE);
			} else if (res_state[0] > PPS_LOWEST_POWER) {
				break;
			}
		}
	}

	if (src_cpup == NULL)
		return (DEFRAG_MOVE_STRAND_ERR_SOURCE);

	/* find a strand on the dest core to move to */
	CHAIN_ITER(dest_core->dcs_corep->cpus, dest_cpup, core_link) {
		res_id[0].u.cpu.pi_id_val = dest_cpup->phys_id;
		res_id[0].pi_res_type = PM_RES_CPU;
		if (pm_get_defrag_power_states(res_id, res_state,
		    RES_NUM) == PS_SUCCESS) {
			if (res_state[0] == PPS_LOWEST_POWER)
				break;
		}
	}

	if (dest_cpup == NULL)
		return (DEFRAG_MOVE_STRAND_ERR_DEST);

	DBG6(zprintf("\tdefrag source : core%d strand%d\n",
	    src_core->dcs_corep->idx, src_cpup->phys_id));
	DBG6(zprintf("\tdefrag dest : core%d strand%d\n",
	    dest_core->dcs_corep->idx, dest_cpup->phys_id));

	rv = pmi_defrag_move_strand(src_cpup->phys_id, dest_cpup->phys_id, -1);

#if 0
	if (rv == 0) {
		set_core_md_cpu_status(src_cpup->phys_id, CORE_MD_PM_STATUS,
		    PM_MANAGED);

		set_core_md_cpu_status(dest_cpup->phys_id, CORE_MD_PM_STATUS,
		    PM_NORMAL);
	}
#endif

	return (rv);
}

/*
 * Move all the strands from src_idx core to other cores.
 *
 * Return zero for success, and non-zero for failure.
 */
static int
defrag_free_core(defrag_core_stat_t *cores, int num_cores, int src_idx)
{
	int rv;
	int dest_idx;
	bool_t try_again = false;

	/*
	 * Loop until all powered-on strands on the source core have been
	 * powered-off.  This may require several destination cores.
	 */
	while (cores[src_idx].dcs_cpu_on > 0) {

		dest_idx = defrag_find_dest_core(cores, num_cores, src_idx);

		if (dest_idx < 0)
			return (0);	/* no destination core available */

		while ((cores[src_idx].dcs_cpu_on > 0) &&
		    (cores[dest_idx].dcs_cpu_off > 0)) {
			try_again = false;

			rv = defrag_move_strand(&cores[src_idx],
			    &cores[dest_idx]);

			if (rv == 0) {
				cores[src_idx].dcs_cpu_on--;
				cores[src_idx].dcs_cpu_off++;

				cores[dest_idx].dcs_cpu_on++;
				cores[dest_idx].dcs_cpu_off--;
			} else {
				switch (rv) {
				case DEFRAG_MOVE_STRAND_ERR_SOURCE:
					DBG6(zprintf("%s: failed to move strand"
					    " -- error with src\n", __func__));

					/* blacklist the source core */
					cores[src_idx].dcs_fail_off = true;
					rv = 0;

					break;

				case DEFRAG_MOVE_STRAND_ERR_DEST:
					DBG6(zprintf("%s: failed to move strand"
					    " -- error with dest\n", __func__));

					/* blacklist the dest core */
					cores[dest_idx].dcs_fail_on = true;
					rv = 0;

					break;

				case DEFRAG_MOVE_STRAND_ERR_SPARE:
					DBG6(zprintf("%s: failed to move strand"
					    " -- err with spare\n", __func__));

					/* blacklist the dest core */
					cores[dest_idx].dcs_fail_on = true;
					rv = 0;

					/*
					 * Unsupported scenario so find
					 * another core
					 */
					try_again = true;

					break;

				case DEFRAG_MOVE_STRAND_ERR_CONFIG:
					DBG6(zprintf("%s: failed to move strand"
					    " -- error with cfg\n", __func__));

					/* no blacklist, return error */
					rv = -1;

					break;

				default:
					DBG6(zprintf("%s: failed to move "
					    "strand (rv=%d)\n", __func__, rv));

					/* blacklist both source and dest */
					cores[src_idx].dcs_fail_off = true;
					cores[dest_idx].dcs_fail_on = true;
					rv = 0;
				}
				if (!try_again)
					return (rv);
				break;
			}
		}
	}

	return (0);
}

/*
 * Attempt to defragment cpus/strands and cores.  We're trying to consolidate
 * strands onto cores, so that the cores may be power-managed.
 */
static void
defrag_cpu(ldm_err_t *results)
{
	int i;
	int rv;
	int src_idx;
	int low_idx;
	int num_cores;
	pri_chip_t *chipp;
	pri_core_t *corep;
	defrag_core_stat_t *cores;
	int core_count;

	CHAIN_ITER(pri.chip_list, chipp, link) {
		num_cores = CHAIN_LENGTH(chipp->cores);

		DBG6(zprintf("%s: processing chip%d  num_cores=%d\n", __func__,
		    chipp->idx, num_cores));

		cores = Xcalloc(num_cores, defrag_core_stat_t);

		i = 0;
		CHAIN_ITER(chipp->cores, corep, chip_link) {
			defrag_fill_core_stat(corep, &cores[i]);
			i++;
		}

		core_count = 0;

		low_idx = defrag_get_lowest_core(cores, num_cores);

		while (is_defrag_cpu_possible(cores, num_cores, low_idx)) {

			DBG6(zprintf("%s: defrag is possible\n", __func__));

			src_idx = defrag_find_source_core(cores, num_cores, -1,
			    low_idx);

			if (src_idx < 0)
				break;	/* no source core available */

			rv = defrag_free_core(cores, num_cores, src_idx);

			if (rv != 0) {
				results->ldme_errs = LDME_OP_FAILURE;
				Xfree(cores);
				return;
			}

			DBG6(zprintf("%s: defrag is complete\n\n", __func__));

			if (++core_count >= DEFRAG_CPU_CORE_LIMIT) {
				DBG6(zprintf("defrag cpu core at limit\n"));
				break;
			}
		}

		Xfree(cores);
	}

	results->ldme_errs = LDME_OP_EOK;
}

/*
 * ldm_defragment() requests that Zeus apply the resource allocation strategy
 * indicated by the "policy" parameter to the resource indicated by the
 * "res_type" parameter. If policy is PACK, then ldm_defragment() requests that
 * allocation and/or assignment of physical resources of type res_type be
 * arranged so that one or more physical resources of type res_type may be
 * taken out of service. Any change in allocation of physical resources must
 * result in a call to pm_domain_config_change_notify() if domain config
 * notification for res_type is in effect.  ldm_defragment() is a request for a
 * best effort from Zeus.  Changes in resource allocation may or may not result,
 * and if they do, they  may happen asynchronously with this call.
 *
 * This must be called from inside the sequencer.
 *
 * NOTE: This function is also called by pm_testlib (not part of ldmd)
 */
void
ldm_defragment(defrag_policy_t policy, pm_res_type_t res_type,
    ldm_err_t *results)
{
	ASSERT_IN_SEQUENCER();

	if (zeus.mode != ZEUS_MGMT_MODE) {
		DBG(zprintf("%s: no defragment -- zeus not in config mode\n",
		    __func__));
		results->ldme_errs = LDME_OP_FAILURE;
		return;
	}

	if (policy == PERFORMANCE) {
		DBG(zprintf("%s: no defragment -- policy is PERFORMANCE\n",
		    __func__));
		results->ldme_errs = LDME_OP_EOK;
		return;
	}

	if (defrag_running) {
		results->ldme_errs = LDME_OP_EOK;
		return;
	}

	defrag_running = true;

	switch (res_type) {
	case PM_RES_CPU:
		defrag_cpu(results);
		break;

	case PM_RES_MEM:
		break;

	default:
		DBG(zprintf("%s: illegal resource type (%d)\n", __func__,
		    (int)res_type));
	}

	defrag_running = false;
}
