/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * LDom/PM Interface Functions
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "zpri.h"
#include "hvctl.h"
#include "vcpu.h"
#include "sequencer.h"
#include "core_md.h"
#include "pm_shared.h"

#define	PM_GET_POWER_RETRIES	3
#define	PM_SET_POWER_RETRIES	1
#define	PM_WAIT_BEFORE_RETRY	1	/* wait for 1 second before retry op */

static pwr_req_status_t pm_hv2ps_status(hvctl_status_t cc);

extern double avg_guest_util(ldom_t *ldp, hrtime_t current_time);
extern double avg_cpu_util(vcpu_t *vcpu, hrtime_t current_time);
extern double avg_mem_util(pm_mem_t *pmp, hrtime_t current_time);
extern int do_hv_pm_state(hvctl_pm_state_dir_t dir, int count, char **cookie,
    uint16_t *states, hvctl_status_t *status);
extern int do_hv_pm_version(uint16_t *major, uint16_t *minor);
extern int ldom_num_bound_vcpus(ldom_t *ldp);
extern zeus_pri_t pri;

/*
 * ldm_get_pri - set caller's MD pointer to point to a copy of the
 * Domain Manager's master vBSC pri.
 *
 * Return:
 *	0 - upon success
 *	-1 - upon failure
 */

/*ARGSUSED*/
int
ldm_get_pri(md_t **prip, ldm_err_t *errp)
{
	if (!pri.pri) {
		errp->ldme_errs = LDME_OP_EUNAVAIL;
		return (-1);
	}
	*prip = md_clone(pri.pri);
	errp->ldme_errs = LDME_OP_EOK;
	return (0);
}

/*
 * ldm_release_pri - free up the PRI copy and set caller's MD pointer to NULL
 */
void
ldm_release_pri(md_t **prip)
{
	ASSERT(*prip);
	md_fini(*prip);
	*prip = NULL;
}

/*
 * ldm_get_hv_pm_version - Read HV supported pm-version
 * Return:
 *      0 - upon success
 *      -1 - upon failure
 */
int
ldm_get_hv_pm_version(uint16_t *major, uint16_t *minor)
{
	return (pm_hv2ps_status(do_hv_pm_version(major, minor)));
}

/*
 * ldm_enqueue_work - arranges for "func" to be called with argument arg
 * from within the sequencer.
 */
/*ARGSUSED*/
void
ldm_enqueue_work(void (*func)(void *), void *data)
{
	seq_enqueue_work(func, data, false);
}

/*
 * Takes specified resource off line or online.  Currently only works
 * for cpu strands.
 */
int
ldm_reconfigure_resources(pm_config_op_t op, int res_count, pm_resid_t *ids,
	int *status, ldm_err_t *results, uint32_t flags)
{
	int		i, rv, retval;
	uint32_t	*pids;

	ASSERT_IN_SEQUENCER();

	retval = 0;

	pids = Xcalloc(res_count, uint32_t);

	for (i = 0; i < res_count; ++i) {
		ASSERT(ids[i].pi_res_type == PM_RES_CPU);
		pids[i] = (uint32_t)ids[i].u.cpu.pi_id_val;
	}

	switch (op) {
	case PCO_ADD_RESOURCE:
	case PCO_REMOVE_RESOURCE:
		rv = seq_reconf_pcpus(op, res_count, pids, status, flags);
		break;
	default:
		DBG(zprintf("Illegal op to %s: %d\n", __func__, op));
		Xfree(pids);
		results->ldme_errs = LDME_OP_EINVAL;
		return (-1);
	}

	if (rv != LDME_OP_EOK) {
		results->ldme_errs = rv;
		retval = -1;
	}

	Xfree(pids);
	return (retval);
}

/*
 * ldm_set_power_states - Set the power states of the resources identified by
 * the pm_cookie array to the new_states.
 *
 * Return:
 *     PS_SUCCESS - upon success
 *     PS_TIMEDOUT or PS_EINVAL or PS_FAILURE - upon failure
 */
pwr_req_status_t
ldm_set_power_states(int res_count, char **pm_cookie, uint16_t *new_states,
	pwr_req_status_t *status)
{
	hvctl_status_t	cc, *hv_status;
	int		i, again = 0;

	ASSERT_IN_SEQUENCER();

	hv_status = Xmalloc(res_count * sizeof (hvctl_status_t));
set_again:
	cc = do_hv_pm_state(HVctl_pm_state_set, res_count, pm_cookie,
	    new_states, hv_status);

	switch (cc) {
	case HVctl_st_ok:
		for (i = 0; i < res_count; i++) {
			status[i] = pm_hv2ps_status(hv_status[i]);

			DBG_BLOCK {
				if ((status[i] != PS_SUCCESS) &&
				    (hv_status[i] != HVctl_st_async)) {
					DBG6(zprintf("%s: FAIL STATUS: HVctl_st"
					    "_ok: status[%d]=%d, hv_status[%d]="
					    "%d\n", __func__, i, status[i], i,
					    hv_status[i]));
				}
			}
		}
		break;

	case HVctl_st_einval:
	case HVctl_st_eillegal:
		for (i = 0; i < res_count; i++)
			status[i] = PS_EINVAL;
		DBG6(zprintf("%s: PS_EINVAL: HVctl_cc=%d", __func__, cc));
		break;

	case HVctl_st_timedout:
		/*
		 * Operation timedout, retry once.
		 */
		if (again++ < PM_SET_POWER_RETRIES) {
			DBG6(zprintf("%s: Retry %d times\n", __func__, again));
			sleep(PM_WAIT_BEFORE_RETRY);
			goto set_again;
		} else {
			for (i = 0; i < res_count; i++)
				status[i] = PS_TIMEDOUT;
			DBG6(zprintf("%s: PS_TIMEDOUT: HVctl_cc=%d\n",
			    __func__, cc));
			break;
		}
	default:
		DBG6(zprintf("%s: PS_FAILURE: op failed with unexpected "
		    "HVctl_status %d\n", __func__, cc));
		for (i = 0; i < res_count; i++)
			status[i] = PS_FAILURE;
		break;
	}

	DBG_BLOCK {
		if (cc != HVctl_st_ok) {
			DBG6(zprintf("%s: NON SUCCESS STATUS: "
			    "do_hv_pm_state(hvctl_status)=%d\n",
			    __func__, cc));
			for (i = 0; i < res_count; i++) {
				DBG6(zprintf("status[i]=%d, hv_status[i]=%d, "
				    "pm_cookie[i]=%s\n",
				    status[i], hv_status[i], pm_cookie[i]));
			}
		}
	}

	Xfree(hv_status);
	return (pm_hv2ps_status(cc));
}

/*
 * ldm_get_power_states - get the power states for resources identified by an
 * array of cookies "pm_cookie".
 *
 * Return:
 *     PS_SUCCESS - upon success
 *     PS_TIMEDOUT or PS_EINVAL or PS_FAILURE - upon failure
 */
pwr_req_status_t
ldm_get_power_states(int res_count, char **pm_cookie, uint16_t *current_states,
	pwr_req_status_t *status)
{
	hvctl_status_t	cc;
	hvctl_status_t	*hv_status;
	int		i;
	int		again = 0;

	ASSERT_IN_SEQUENCER();

	hv_status = Xmalloc(res_count * sizeof (hvctl_status_t));
get_again:
	cc = do_hv_pm_state(HVctl_pm_state_get, res_count, pm_cookie,
	    current_states, hv_status);
	DBG(zprintf("%s: do_hv_pm_state returns cc=0x%X\n", __func__, cc));

	switch (cc) {
	case HVctl_st_ok:
		for (i = 0; i < res_count; i++)
			status[i] = pm_hv2ps_status(hv_status[i]);
		break;

	case HVctl_st_einval:
	case HVctl_st_eillegal:
		for (i = 0; i < res_count; i++)
			status[i] = pm_hv2ps_status(cc);
		break;

	case HVctl_st_timedout:
		if (again++ < PM_GET_POWER_RETRIES) {
			DBG(zprintf("%s: retry, #%d\n", __func__, again));
			sleep(PM_WAIT_BEFORE_RETRY);
			goto get_again;
		} else {
			for (i = 0; i < res_count; i++)
				status[i] = PS_TIMEDOUT;
			break;
		}
	default:
		DBG(zprintf("%s: unrecognized hvctl_status %d\n",
		    __func__, cc));
		for (i = 0; i < res_count; i++)
			status[i] = PS_FAILURE;
		break;
	}

	Xfree(hv_status);
	return (pm_hv2ps_status(cc));
}

/*
 * ldm_get_all_domain_ids - Fill up domlist with an array of domain ids of
 * all currently configured domains regardless the domain state.
 * count will be set to the number of domain ids returned.
 *
 * Must be called from inside the sequencer.
 */
/*ARGSUSED*/
int
ldm_get_all_domain_ids(srcid_t **domlist, uint_t *count, ldm_err_t *errp,
    int flag)
{
	ldom_t	*ldp;
	srcid_t	*domlp;
	int	j;

	ASSERT_IN_SEQUENCER();

	domlp = Xmalloc((CHAIN_LENGTH(zeus.ldoms)) *
	    sizeof (srcid_t));
	*domlist = domlp;

	j = 0;
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		switch (flag) {
		case ACTIVE_DOMAINS:
			if (ldp->state == ACTIVE)
				domlp[j++] = (srcid_t)ldp;
			break;

		default:
			domlp[j++] = (srcid_t)ldp;
			break;
		}
	}
	*count = j;

	errp->ldme_errs = LDME_OP_EOK;
	return (0);
}

/*
 * ldm_get_domain_utils - return domain or resource utilization data.
 * Caller is responsible to free "utils".
 */
/*ARGSUSED*/
int
ldm_get_domain_utils(srcid_t src_id, pm_res_type_t res_type,
    ldm_resutils_t **utils, ldm_err_t *errp)
{
	hrtime_t	cur_time;
	ldm_resutils_t	*res_utils, *ldmu;
	ldm_util_t	*lup;
	ldom_t		*ldp = NULL;
	pm_mem_t	*pmp = NULL;
	size_t		sz;
	vcpu_t		*vcpu;
	int		i;

	ASSERT_IN_SEQUENCER();

	/*
	 * Get util info for resource types that don't bind to domains and
	 * return. Currently, memory is the only power managable resource
	 * type that does not specifically bind to domains.
	 */
	if (res_type == PM_RES_MEM) {
		CHAIN_ITER(pm_mem_list.mem_list, pmp, link) {
			if ((pm_mem_t *)src_id == pmp)
				break;
		}

		if (pmp == NULL) {
			DBG(zprintf("%s: memory grouping 0x%p was not "
			    "found in pm_mem_list.\n", __func__, src_id));
			errp->ldme_errs = LDME_OP_EUNAVAIL;
			return (-1);
		}

		sz = sizeof (ldm_resutils_t);
		res_utils = Xmalloc(sz);
		ldmu = &pmp->pm_mem_resutils;
		bcopy((void *)ldmu, (void *)res_utils, sz);

		/* retrieve resource util data */
		res_utils->ldmr_rescnt = 1;
		res_utils->ldmr_util_p = Xcalloc(res_utils->ldmr_rescnt,
		    ldm_util_t);
		cur_time = gethrtime();
		res_utils->ldmr_avg_util = avg_mem_util(
		    (pm_mem_t *)src_id, cur_time);
		res_utils->ldmr_util_p->ldmu_util = res_utils->ldmr_avg_util;
		res_utils->ldmr_util_p->ldmu_resid.pi_res_type = PM_RES_MEM;
		res_utils->ldmr_util_p->ldmu_resid.pi_res_name =
		    Xstrdup(MEM_GROUPING);
		res_utils->ldmr_util_p->ldmu_resid.u.mem.pi_id_val =
		    ((pm_mem_t *)src_id)->pm_mem_resid.id;

		*utils = res_utils;
		errp->ldme_errs = LDME_OP_EOK;
		return (0);
	}

	/*
	 * Get util info for resource types that do bind to domains.
	 * Currently, CPUs are the only power managable resource type that
	 * binds to domains.
	 */
	ASSERT(res_type == PM_RES_CPU);

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if ((ldom_t *)src_id == ldp)
			break;
	}

	if (ldp == NULL) {
		DBG(zprintf("%s: no such domain by domid 0x%p\n",
		    __func__, src_id));
		errp->ldme_errs = LDME_OP_EUNAVAIL;
		return (-1);
	}

	/* copy out per domain res util data */
	sz = sizeof (ldm_resutils_t);
	res_utils = Xmalloc(sz);
	ldmu = &ldp->pm_cpu_resutils;
	bcopy((void *)ldmu, (void *)res_utils, sz);

	cur_time = gethrtime();
	res_utils->ldmr_avg_util = avg_guest_util(ldp, cur_time);

	/* retrieve per resource util data */
	res_utils->ldmr_rescnt = CHAIN_LENGTH(ldp->bound.vcpus);
	res_utils->ldmr_util_p = Xcalloc(res_utils->ldmr_rescnt, ldm_util_t);
	i = 0;
	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		lup = &res_utils->ldmr_util_p[i];
		lup->ldmu_util = avg_cpu_util(vcpu, cur_time);
		lup->ldmu_resid.pi_res_type = PM_RES_CPU;
		lup->ldmu_resid.pi_res_name = Xstrdup(CPU_RES_NAME);
		lup->ldmu_resid.u.cpu.pi_id_val = vcpu->cpu->phys_id;
		i++;
	}

	*utils = res_utils;
	errp->ldme_errs = LDME_OP_EOK;
	return (0);
}

/*
 * ldm_res_power_fault_report - update the Domain Manager about its resources'
 * current status.
 *
 * Input:
 *   res_status - an array of resource ids and status
 *   ev         - PM event that caused the resource power status change
 */
void
ldm_res_power_fault_report(ldm_err_t *res_err_status, ldme_event_t ev)
{
	int  			i;
	ldme_res_sane_t		*lrs;
	ldme_res_status_t	*res_status = NULL;

	switch (ev) {
	case LDME_PE_ENTER_PM:
		DBG(zprintf("%s: LDME_PE_ENTER_PM event: Ignore resource power"
		    "status\n", __func__));
		/*
		 * FIXME: PM has started. All previously deemed unusable
		 * resources now have a chance to be recovered.  Mark all
		 * resources good.
		 */
		break;

	case LDME_PE_UTIL:
		DBG(zprintf("%s: LDME_PE_UTIL event: Ignore resource power"
		    "status\n", __func__));
		/*
		 * Domain Manager don't care this case
		 */
		break;

	case LDME_PE_REPORT_RES:
		DBG(zprintf("%s: LDME_PE_REPORT_RES event: Reporting which "
		    "resources are not power managable to the Domain "
		    "Manager.\n", __func__));

		if (res_err_status)
			res_status = res_err_status->ldme_private;

		if (res_status) {
			lrs = res_status->lrs;

			/*
			 * FIXME: In this case, PM is no longer
			 * involved in powering up or down resources' power,
			 * hence there is no window to recover a resource at
			 * unknown power, or stuck at low power.  The Domain
			 * Manager needs to prevent users from allocating,
			 * using their resources.
			 */
			for (i = 0; i < res_status->rescnt; i++) {
				if (lrs[i].sanity != LDME_RES_GOOD) {
					if (lrs[i].resid->pi_res_type ==
					    PM_RES_CPU) {
						DBG(zprintf("%s: don't use CPU"
						    " resid %"PRIu64"\n",
						    __func__, lrs[i].resid->\
						    u.cpu.pi_id_val));
					} else if (lrs[i].resid->pi_res_type ==
					    PM_RES_MEM) {
						DBG(zprintf("%s: memory "
						    "grouping %"PRIu64" is not "
						    "power managable\n",
						    __func__, lrs[i].resid->\
						    u.mem.pi_id_val));
					}
				}
				/* FIXME: More Domain Manager work goes here */
			}
			Xfree(res_status->lrs);
			Xfree(res_status);
		} else {
			DBG(zprintf("%s: No resource power adjustment status"
			    " available.\n", __func__));
			break;
		}

		break;

	case LDME_PE_DOMAIN_BOOTING:
		DBG(zprintf("%s: LDME_PE_DOMAIN_BOOTING event: Handle "
		    "resource power status.\n", __func__));
		break;

	default:
		zprintf("%s: Warning: unrecognized LDME_PE_ event %d\n",
		    __func__, ev);
		break;
	}
}

/*
 * utility functions
 */

/*
 * pm_hv2ps_status - convert hvctl notion of status code to the LDoms/PM
 * shared status code.
 */
static pwr_req_status_t
pm_hv2ps_status(hvctl_status_t cc)
{
	switch (cc) {
	case HVctl_st_ok:
		return (PS_SUCCESS);

	case HVctl_st_einval:
		return (PS_EINVAL);

	case HVctl_st_eillegal:
		return (PS_EILLEGAL);

	case HVctl_st_timedout:
		return (PS_TIMEDOUT);

	default:
		return (PS_FAILURE);

	}
}

int
ldm_get_domain_config(srcid_t dom_id, pm_res_type_t res_type,
	pm_resid_t **config, int *count, ldm_err_t *results)
{
	ldom_t		*ldp = (ldom_t *)dom_id;
	pm_resid_t	*resids;
	vcpu_t		*vcpu;
	int		i = 0;

	if (ldp == NULL || ldp->state < BOUND) {
		if (results) {
			results->ldme_errs = LDME_OP_EUNAVAIL;
			results->ldme_private = NULL;
		}
		return (-1);
	}

	if (res_type == PM_RES_CPU) {
		*count = ldom_num_bound_vcpus(ldp);
		*config = resids = Xcalloc(*count, pm_resid_t);
		CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
			resids[i].pi_res_type = PM_RES_CPU;
			resids[i].pi_res_name = CPU_RES_NAME;
			resids[i].u.cpu.pi_id_val = vcpu->cpu->phys_id;
			i++;
		}

	} else {
		if (results) {
			results->ldme_errs = LDME_OP_EUNAVAIL;
			results->ldme_private = NULL;
		}

		return (-1);
	}

	return (0);
}


/*
 * get_pm_fma_cpu_status - retrieve the FMA CPU status for PM.
 */
static int
get_pm_fma_cpu_status(uint32_t cpu_id, int *status)
{
	fma_cpu_resp_t		cpu_resp = { 0 };
	fma_cpu_service_req_t	cpu_msg;
	int			rv, stat;

	cpu_msg.req_num = hton64(0);
	cpu_msg.msg_type = hton32(FMA_CPU_REQ_STATUS);
	cpu_msg.cpu_id = hton32(cpu_id);

	rv = fma_cpu_svc_get_p_status(&cpu_msg, &cpu_resp);
	if (rv != 0) {
		DBG3(zprintf("%s: Failed FMA CPU status call for cpu%d\n",
		    __func__, cpu_id));
		return (-1);
	}

	if (ntoh32(cpu_resp.result) != FMA_CPU_RESP_FAILURE) {
		stat = ntoh32(cpu_resp.status);
		DBG(zprintf("%s: FMA CPU status for cpu%d is %d\n", __func__,
		    cpu_id, stat));
		switch (stat) {
		case FMA_CPU_STAT_ONLINE:
			*status = FMA_CPU_STATUS_ONLINE;
			break;
		case FMA_CPU_STAT_OFFLINE:
			*status = FMA_CPU_STATUS_OFFLINE;
			break;
		case FMA_CPU_STAT_ILLEGAL:
		default:
			*status = FMA_CPU_STATUS_ILLEGAL;
		}

		rv = 0;
	} else {
		DBG(zprintf("%s: FMA CPU status returned failure\n",
		    __func__));
		rv = -1;
	}

	return (rv);
}

/*
 * ldm_get_fma_resource_status - retrieve the FMA status of the requested
 * resource.  If success is returned, the "status" parameter contains the FMA
 * status of the requested resource.  This is called from inside the sequencer.
 */
int
ldm_get_fma_resource_status(pm_resid_t *resid, int *status)
{
	int rv;

	switch (resid->pi_res_type) {
	case PM_RES_CPU:
		rv = get_pm_fma_cpu_status((uint32_t)resid->u.cpu.pi_id_val,
		    status);
		break;

	case PM_RES_MEM:
		/* No FMA status change is associated with memory resource */
		*status = FMA_STATUS_UNKNOWN;
		rv = 0;
		break;

	default:
		DBG(ldm_log_msg("%s: Bad res_type (%d)\n", __func__,
		    resid->pi_res_type));
		rv = -1;
	}

	return (rv);
}

/*
 * ldm_migration_in_process() returns "true" if a domain migration is in
 * process, and "false" if there is not a domain migration in process.
 */
bool_t
ldm_migration_in_process(void)
{
	return (is_migration_in_progress(NULL));
}

/*
 * True if Dynamic Resoure Management (DRM) is in process,
 * and false otherwise.
 */
bool_t
ldm_dynamic_resmgmt_in_process(void)
{
	return (is_drm_in_progress());
}

/*
 * ldm_get_resource_pm_status - retrieve the PM status (currently DR) of
 * the requested resource.  If success is returned, the "pm_status_set"
 * parameter contains true if the requested resource is DR'd out of a
 * domain; otherwise it returns false.  This is called from inside the
 * sequencer.
 */
int
ldm_get_resource_pm_status(pm_resid_t *resid, uint32_t type,
    bool_t *pm_status_set)
{
	int	rv = -1;

	switch (resid->pi_res_type) {
	case PM_RES_CPU:
		if (type == PM_DR_STATUS) {
			/* PM_MANAGED == DR'd out; PM_NORMAL == DR'd in */
			*pm_status_set = (get_core_md_cpu_status(
			    resid->u.cpu.pi_id_val,
			    CORE_MD_PM_STATUS) == PM_MANAGED);
			rv = 0;
		}
		break;

	default:
		DBG(ldm_log_msg("%s: Bad res_type (%d)\n", __func__,
		    resid->pi_res_type));
		break;
	}

	return (rv);
}

/*
 * ldm_delayed_reconfig_in_process() returns "true" if any domain
 * is currently in a delayed reconfiguration state if ldp == NULL.  This
 * means we can't do PM when in that state.  This function returns the
 * delayed reconfiguration of the domain pointed to by ldp, if ldp != NULL.
 */
bool_t
ldm_delayed_reconfig_in_process(void *ldp)
{
	return (in_delayed_reconfig(ldp));
}

/*
 * ldm_get_domain_state
 *
 * Gets the Ldom state from the input ldom_t structure.
 *
 * Input: domid: Should be an ldom_t structure
 * Output: A numeric value for the Ldom state.
 */
int
ldm_get_domain_state(void *domid)
{
	ldom_t *ldp = (ldom_t *)domid;
	return ((int)ldp->state);
}

/*
 * ldm_get_domain_name
 *
 * Gets the Ldom name from the input ldom_t structure.
 *
 * Input: domid: Should be an ldom_t structure
 * Output: A pointer to the Ldom name, inside an ldom_t struct
 */
char *
ldm_get_domain_name(void *domid)
{
	ldom_t *ldp = (ldom_t *)domid;
	return (ldp->namep);
}
