/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Establish Platform Power Manageable resource capability,
 * handle resource power level transitions and dependency,
 * coordination requirement among resources.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <macros.h>
#include "vcpu.h"
#include "response.h"
#include "pm_shared.h"
#include "pm_caps.h"
#include "pm_fma.h"

/* Defines */
#define	COMPONENT	"component"
#define	CORE		"core"
#define	CPU		"CPU"
#define	FWD		"fwd"
#define	INVALID_ID	-1
#define	MEMORY		"memory"
#define	MEM_GROUPING	"memory-grouping"
#define	MEM_REGION	"memory-region"
#define	NO_LEAF		-1
#define	PM_MAPPING	"pm_mapping"
#define	PM_RES_PSTATE	 PM_RES_TYPES_TOTAL + 1
#define	PM_RESOURCE	"pm_resource"
#define	POWER_MGMT	"power-management"
#define	PROCESSOR	"processor"
#define	STRAND		"strand"
#define	SYSTEM_BOARD	"System Board"
#define	TYPE		"type"

#ifndef min
#define	min(a, b)	((a) < (b) ? (a) : (b))
#endif	/* min */

/*
 * PM_HV_WORKAROUND is enabled for platform Hypervisor that assumes the lowest
 * numbered strand being its anchor and unable to deal with another strand
 * migrated to be the anchor strand. This workaround keeps the lowest strand
 * always ON until HV comes up with its solution.
 */
#define	PM_HV_WORKAROUND
#ifdef	PM_HV_WORKAROUND
static void epm_find_lowest_cpu(pm_node_t *node, pm_node_t **lowest);
#endif	/* PM_HV_WORKAROUND */

/* HPS variables */
static bool_t hps_enabled = false;	/* Is HPS supported for this system */
static int num_hps_nodes = 0;		/* Number of comps that support HPS */
static int hps_num_lvls = 0;		/* # of HPS levels */
/* List of pm_nodes for HPS nodes and hi/lo strands */
static pm_hps_strand_map_t *hps_strand_map = NULL;

/*
 * Power Management Private variables
 */
static pm_node_t *pm_root_node = NULL;
static bool_t pm_pri_capable = false;
static int dbg_line = 0;
static int pm_version_minor = 0;
static int pm_version_major = 0;
static bool_t pm_mem_supported = false;

#ifdef  PM_HV_WORKAROUND
/* FIXME: CR 6818078 needs PM_HV_WORKAROUND defined */
#define	PM_MAX_CHIPS	4
pm_node_t *epm_find_proc_node(pm_node_t *node, int chip_num);
#endif	/* PM_HV_WORKAROUND */

/* Local Power Management functions */

static bool_t pmi_is_mem_version_ok(void);
static int epm_adjust_pwr(pm_res_type_t res_type, int count,
    pm_config_op_t req_rcfg_op, pmi_pwr_state_t pmi_pwr,  pm_node_t **nodepp,
    int *res_status, uint32_t flags);
static int epm_post_pwr(pm_res_type_t res_type, int count, uint16_t new_state,
    pwrctl_method_t method, pm_node_t **nodepp, uint32_t flags);
static int epm_pre_pwr(pm_res_type_t res_type, int count, uint16_t new_state,
    pwrctl_method_t method, pm_node_t **nodepp, int *needidx, uint32_t flags);
static int epm_ps2pwr_status(pwr_req_status_t cc);

static void epm_debug_print_op(char *fnm, int count, uint16_t new_state,
    pm_node_t **npp);
static void epm_record_range(pm_node_t *node);
static void epm_update_range(pm_node_t *node);

/*
 * pmi_is_pri_pm_capable returns a boolean value to indicate whether or not the
 * platform's pri supports PM.
 */
bool_t
pmi_is_pri_pm_capable(void)
{
	return (pm_pri_capable);
}

/*
 * epm_is_sysfw_supported() checks to see
 *   o if platform PRI "pm-version" is supported by LDoms/PM,
 *   o if platform PRI "pm-version" is supported by Hypervisor,
 *   o if HVctl version supports power management
 * Return:  true if sysfw properly supports PM; false if not.
 */
bool_t
epm_is_sysfw_supported(md_t *pm_md)
{
	md_node_t	*nodep;
	char		*pm_vers;
	uint16_t	hv_major, hv_minor;
	uint32_t	pri_major, pri_minor;

	pm_pri_capable = false;

	if (!(nodep = md_find_node(pm_md, "components"))) {
		ldm_log_msg("Warning: No PRI 'components' node. Power"
		    " Management is not supported on this platform.\n");
		return (false);
	}
	if (md_get_prop_str(pm_md, nodep, "pm-version", &pm_vers) == -1) {
		ldm_log_msg("Warning: No PRI 'pm-version' property. Power"
		    " Management is not supported on this platform.\n");
		return (false);
	}

	DBG6(zprintf("%s: PRI pm-version(%s)\n", __func__, pm_vers));

	if (sscanf(pm_vers, "%d.%d", &pri_major, &pri_minor) != 2) {
		DBG(zprintf("%s: unrecognized pm-version(%s) in PRI\n",
		    __func__, pm_vers));
		return (false);
	}

	/* check HV on pm-version, and HVctl version */
	if (GET_HV_PM_VERSION(&hv_major, &hv_minor) != PS_SUCCESS) {
		ldm_log_msg("Warning: HV does not support PM\n");
		return (false);
	} else {
		DBG6(zprintf("%s: HV pm-version(%d.%d)\n", __func__,
		    hv_major, hv_minor));
		if (pri_major == hv_major) {
			pm_version_major = pri_major;
			pm_version_minor = min(pri_minor, hv_minor);
		} else if (pri_major > hv_major) {
			pm_version_major = hv_major;
			pm_version_minor = hv_minor;
		} else {
			pm_version_major = pri_major;
			pm_version_minor = pri_minor;
		}

		DBG6(zprintf("%s: negotiated pm-version: %d.%d\n", __func__,
		    pm_version_major, pm_version_minor));
		if (((pm_version_major == PRI_PM_VERSION_MAJOR) &&
		    (pm_version_minor >= PRI_PM_VERSION_MINOR)) ||
		    (pm_version_major > PRI_PM_VERSION_MAJOR)) {
			pm_pri_capable = true;
			return (true);
		} else {
			ldm_log_msg("Warning: PM is not supported on this "\
			    "Firmware version %d.%d, expecting %d.%d\n",
			    pm_version_major, pm_version_minor,
			    PRI_PM_VERSION_MAJOR, PRI_PM_VERSION_MINOR);
			return (false);
		}
	}
}

/*
 * epm_cmp_res_id - Used by qsort to compare resources by id.
 *
 * Returns: 1 if p1's id is greater than p2's id
 *          0 if p1's id is equal to p2's id
 *          -1 if p1's id is less than p2's id
 *
 * Note: FMA faulted resources are put at the beginning of the list.
 */
static int
epm_cmp_res_id(const void *p1, const void *p2)
{
	pm_resid_t	*res1 = (pm_resid_t *)p1;
	pm_resid_t	*res2 = (pm_resid_t *)p2;
	bool_t		res1_faulted = pmi_is_resource_faulted(res1);
	bool_t		res2_faulted = pmi_is_resource_faulted(res2);
	int		res1_id_val = 0;
	int		res2_id_val = 0;
	static int	dbg_cnt = 0;

	switch (((pm_resid_t *)p1)->pi_res_type) {
	case PM_RES_CPU:
		res1_id_val = res1->u.cpu.pi_id_val;
		res2_id_val = res2->u.cpu.pi_id_val;
		break;
	case PM_RES_MEM:
		res1_id_val = res1->u.mem.pi_id_val;
		res2_id_val = res2->u.mem.pi_id_val;
		break;
	default:
		/*
		 * If we get here, something went horribly wrong.  We
		 * don't want to flood the logs, so log a one time non
		 * debug message.
		 */
		if (dbg_cnt == 0) {
			zprintf("%s: unknown resource type (%d)\n", __func__,
			    ((pm_resid_t *)p1)->pi_res_type);
			dbg_cnt++;
		}
	}

	if (res1_faulted && res2_faulted) {
		return (0);
	} else if (res1_faulted) {
		return (-1);
	} else if (!res1_faulted && res2_faulted) {
		return (1);
	} else if (res1_id_val > res2_id_val) {
		return (1);
	} else if (res1_id_val < res2_id_val) {
		return (-1);
	} else {
		return (0);
	}
}

/*
 * epm_scan_pm_resources - Scan for "cpu" and "memory" resources.
 *
 * return: 0 - if found any,  -1 - if none.
 */
int
epm_scan_pm_resources(md_t *pri)
{
	md_t		*pm_md = NULL;
	md_node_t	**cpulistp = NULL;
	md_node_t	**regionlistp = NULL;
	md_node_t	**reslistp = NULL;
	pm_mem_t	*new_entry = NULL;
	pm_node_t	*last_enq_under_mem_group;
	pm_node_t	*last_enq_under_processor;
	pm_node_t	*new_node, *last_enqueued;
	char		*last_enq_type = NULL;
	char		*str;
	int		cpu_idx, ncpus, node_count;
	int		pm_node_count = 0;
	int		region_cnt = 0;
	int		res_idx, reg_idx;
	int		total_num_res = 0;
	int		ret;

	pm_md = epm_prune_md(pri);
	node_count = md_node_count(pm_md);
	reslistp = Xcalloc(node_count, md_node_t *);

	/* check SysFW has PM support */
	if (!epm_is_sysfw_supported(pm_md)) {
		ret = -1;
		goto scan_done;
	}

	/*
	 * collect the "component" and "power-management" PRI nodes
	 */
	total_num_res = md_scan_dag(pm_md, md_root_node(pm_md),
	    md_find_name(pm_md, COMPONENT), md_find_name(pm_md, FWD),
	    reslistp);
	DBG(zprintf("%s: md_scan_dag num component resources is %d\n", __func__,
	    total_num_res));

	/* If right versions, add number of memory grouping nodes in PRI */
	if (pmi_is_mem_version_ok())
		total_num_res += md_scan_dag(pm_md, md_root_node(pm_md),
		    md_find_name(pm_md, MEM_GROUPING), md_find_name(pm_md, FWD),
		    &reslistp[total_num_res]);

	DBG(zprintf("%s: md_scan_dag num power mgmt resources is %d\n",
	    __func__, total_num_res));

	/*
	 * create PM device tree root node
	 */
	if (!(pm_root_node = epm_create_root())) {
		DBG(zprintf("Fatal: %s: failed to create pm_root_node\n",
		    __func__));
		ret = -1;
		goto scan_done;
	} else {
		last_enqueued = pm_root_node;
	}

	/*
	 * build up processor, cpu, memory-grouping, and memory-region nodes
	 */
	for (res_idx = 0; res_idx < total_num_res; res_idx++) {
		md_node_t	*top_pm_node;

		/*
		 * If this is a component node, get the type from the 'type'
		 * field. Otherwise, get the type from the actual node type.
		 */
		if (strcmp(reslistp[res_idx]->typep->strp, COMPONENT) == 0) {
			if (md_get_prop_str(pm_md, reslistp[res_idx], TYPE,
			    &str) == -1)
				continue;
		} else {
			str = reslistp[res_idx]->typep->strp;
		}

		/* enqueue "processor" or "memory-grouping" node */
		if ((strcmp(str, PROCESSOR) != 0) &&
		    (strcmp(str, MEM_GROUPING) != 0))
			continue;

		top_pm_node = reslistp[res_idx];

		if (!(new_node = epm_create_node(pm_md, top_pm_node)))
			continue;

		/*
		 * Need to pass in the parent's type as a comparison unless
		 * the PM root node, which has no parent, is being enqueued.
		 */
		if (strcmp(last_enqueued->pn_type, POWER_MGMT) == 0) {
			last_enq_type = last_enqueued->pn_type;
		} else {
			last_enq_type = last_enqueued->pn_parent->pn_type;
		}

		epm_enqueue_node(last_enqueued, new_node, last_enq_type);

		if (strcmp(str, PROCESSOR) == 0)
			last_enq_under_processor = new_node;
		else if (strcmp(str, MEM_GROUPING) == 0) {
			last_enq_under_mem_group = new_node;

			/*
			 * Adding the grouping information for
			 * utilization gathering.
			 */
			new_entry = Xmalloc(sizeof (pm_mem_t));
			new_entry->pm_mem_resid.id = new_node->pn_nodeid;
			(void) memset(&new_entry->stats, 0,
			    sizeof (new_entry->stats));
			new_entry->pm_mem_resutils.ldmr_type = PM_RES_MEM;
			CHAIN_ADD(pm_mem_list.mem_list, new_entry, link);
		}

		last_enqueued = new_node;
		pm_node_count++;

		/*
		 * Enqueue all nodes under the "processor" or
		 * "memory-grouping" node.
		 */
		if (strcmp(str, PROCESSOR) == 0) {
			cpulistp = Xcalloc(node_count, md_node_t *);
			ncpus = md_scan_dag(pm_md, reslistp[res_idx],
			    md_find_name(pm_md, COMPONENT),
			    md_find_name(pm_md, FWD), cpulistp);
			DBG(zprintf("%s: md_scan_dag ncpus is %d\n",
			    __func__, ncpus));

			for (cpu_idx = 0; cpu_idx < ncpus; cpu_idx++) {
				/* Only create nodes for cores and cpus */
				md_get_prop_str(pm_md, cpulistp[cpu_idx],
				    TYPE, &str);

				if ((strcmp(CORE, str) != 0) &&
				    (strcmp(STRAND, str) != 0))
					continue;

				if (!(new_node = epm_create_node(pm_md,
				    cpulistp[cpu_idx])))
					continue;

				epm_enqueue_node(last_enq_under_processor,
				    new_node, NULL);

				last_enq_under_processor = new_node;
				pm_node_count++;
			}
			Xfree(cpulistp);

		} else if (strcmp(str, MEM_GROUPING) == 0) {
			pm_mem_supported = true;
			regionlistp = Xcalloc(node_count, md_node_t *);
			region_cnt = md_scan_dag(pm_md, reslistp[res_idx],
			    md_find_name(pm_md, MEM_REGION),
			    md_find_name(pm_md, FWD), regionlistp);
			DBG(zprintf("%s: md_scan_dag region_cnt is %d\n",
			    __func__, region_cnt));

			for (reg_idx = 0; reg_idx < region_cnt; reg_idx++) {
				if (!(new_node = epm_create_node(pm_md,
				    regionlistp[reg_idx])))
					continue;

				epm_enqueue_node(last_enq_under_mem_group,
				    new_node, NULL);

				last_enq_under_mem_group = new_node;
				pm_node_count++;
			}
			Xfree(regionlistp);
		}
	}

	/* mark strand node id range in ancestor nodes */
	epm_update_range(pm_root_node);
	epm_debug_print_tree(pm_root_node);

	if (pm_node_count == 0) {
		ldm_log_msg("Warning: Current Platform or SysFW does not "
		    "support power management.\n");
		ret = -1;
	} else {

		/*
		 * For HPS, we need to easily find out what chip a strand
		 * is on.  Actually, we need to find out what chip a domain is
		 * on, but this translates into finding out what chip a domain's
		 * strands are on.  So we will build a static global structure
		 * that contains chip IDs and the lo/hi range of the strands for
		 * those chips.  This data will not change since it comes from
		 * the PRI, so it will stay valid until the next reboot.
		 */
		if (pmi_is_hps_supported()) {
			pm_hps_strand_map_t	*hsmp;
			int			err = 0;

			pmi_hps_create_strand_mapping(pm_root_node, -1, &err);

			/* Any errors will disable PState */
			if (err < 0) {
				DBG(zprintf("%s: Failure building "
				    "hps_strand_map list\n", __func__));
				hps_enabled = false;
				Xfree(hps_strand_map);
				hps_strand_map = NULL;
			} else {

				/* Global var for # of HPS levels */
				for (hsmp = hps_strand_map; hsmp;
				    hsmp = hsmp->next) {
					hps_num_lvls =
					    max(hps_num_lvls, hsmp->level);
				}
			}

			DBG_BLOCK {
				DBG6(zprintf("%s: hps_num_lvls=%d\n",
				    __func__, hps_num_lvls));
				hsmp = hps_strand_map;
				while (hsmp != NULL) {
					DBG6(zprintf("%s: Chip %d(%s) Level: %d"
					    ", Hi %d, Lo %d\n", __func__,
					    hsmp->chip_id,
					    hsmp->pmnode->pn_pmname,
					    hsmp->level,
					    hsmp->strand_hi, hsmp->strand_lo));
					hsmp = hsmp->next;
				}
			}

		}

		ret = 0;
#ifdef	PM_HV_WORKAROUND
		/* Find lowest strand for each CPU chip */
		pm_node_t 	*lowest = NULL;
		pm_node_t 	*proc_node;
		int 		chip;
		int 		got_chip = 0;

		/* FIXME: CR 6818078 needs PM_HV_WORKAROUND defined */
		for (chip = 0; chip < PM_MAX_CHIPS; chip++) {
			lowest = NULL;
			proc_node = epm_find_proc_node(pm_root_node, chip);
			if (proc_node) {
				epm_find_lowest_cpu(proc_node, &lowest);
				if (lowest) {
					lowest->pn_flag |= PN_FLAG_KEEPMEUP;
					/* Mark the parent (Core) too */
					lowest->pn_parent->pn_flag |=
					    PN_FLAG_KEEPMEUP;
					got_chip++;

					DBG6(zprintf("%s: lowest_cpu_node "
					    "is %s, lowest_core_node is %s\n",
					    __func__, lowest->pn_pmname,
					    lowest->pn_parent->pn_pmname));

				}
			}
		}

		/* Better have found the lowest strand for one chip */
		ASSERT(got_chip > 0);
#endif	/* PM_HV_WORKAROUND */
	}

scan_done:
	Xfree(reslistp);
	md_fini(pm_md);

	return (ret);
}

md_t *
epm_prune_md(md_t *pri)
{
	md_t *pm_md;

	/*
	 * delete all, except the "components" nodes in template pm_md
	 */
	pm_md = md_clone(pri);


	md_find_delete_node(pm_md, "domain-services");
	md_find_delete_node(pm_md, "virtual-devices");
	md_find_delete_node(pm_md, "channel-endpoints");
	md_find_delete_node(pm_md, "ldc_endpoints");
	md_find_delete_node(pm_md, "firmware");
	md_find_delete_node(pm_md, "devalias");
	md_find_delete_node(pm_md, "cache");	/* can't del */
	md_find_delete_node(pm_md, "tlb");	/* can't del */
	md_find_delete_node(pm_md, "exec-unit");  /* can't del */
	md_find_delete_node(pm_md, "cpu");	/* can't del */
	md_find_delete_node(pm_md, "mblock");
	md_find_delete_node(pm_md, "memory");
	md_find_delete_node(pm_md, "platform");
	md_find_delete_node(pm_md, "memory-segments");
	md_find_delete_node(pm_md, "iodevice");
	md_find_delete_node(pm_md, "memory-bank");
	md_find_delete_node(pm_md, "interrupt-map-entry");
	md_find_delete_node(pm_md, "disk_nac");
	md_find_delete_node(pm_md, "wart");

	/*  augment_core_md(pm_md); */
	md_scour(pm_md, md_find_name(pm_md, "back"));

	return (pm_md);
}

/*
 * create the root node of PM device tree
 */
pm_node_t *
epm_create_root(void)
{
	pm_node_t *np;

	np = Xcalloc(1, pm_node_t);
	np->pn_rank = 0;	/* root node at highest rank */
	np->pn_type = Xmalloc(strlen(POWER_MGMT) + 1);
	strcpy(np->pn_type, POWER_MGMT);
	np->pn_pmname = Xmalloc(strlen(SYSTEM_BOARD) + 1);
	strcpy(np->pn_pmname, SYSTEM_BOARD);

	return (np);
}

int
epm_get_toknum(char *s, char *ss)
{
	char	*s1, *tok;
	int	i;

	if (s == NULL)
		return (0);

	s1 = Xmalloc(strlen(s) + 1);
	strcpy(s1, s);
	tok = strtok(s1, ss);
	for (i = 0; tok; tok = strtok(NULL, ss))
		i++;
	Xfree(s1);
	return (i);
}


#ifdef	PM_HV_WORKAROUND
pm_node_t *
epm_find_proc_node(pm_node_t *node, int chip_num)
{
	while (node) {
		/* Found node case */
		if (strcmp(node->pn_type, PROCESSOR) == 0) {
			pm_node_t *sib;

			for (sib = node; sib; sib = sib->pn_sibling) {
				if (chip_num == sib->pn_nodeid)
					return (sib);
			}
			return (NULL);
		}
		node = node->pn_child;
	}

	return (NULL);
}

void
epm_find_lowest_cpu(pm_node_t *node, pm_node_t **lowest)
{
	pm_node_t	*child;

	/* Leaf node */
	if (node->pn_rcfg) {
		/* CPUs only */
		if (node->pn_rcfg->pi_res_type == PM_RES_CPU) {
			/* First strand is initially lowest */
			if (!(*lowest)) {
				*lowest = node;
				return;
			}
			/* Is this a lower strand id? */
			if ((*lowest)->pn_rcfg->u.cpu.pi_id_val >
			    node->pn_rcfg->u.cpu.pi_id_val)
				*lowest = node;
		}
		return;
	}

	/* Recursive case */
	for (child = node->pn_child; child; child = child->pn_sibling) {
		epm_find_lowest_cpu(child, lowest);
	}
}
#endif	/* PM_HV_WORKAROUND */


int
epm_substr(const char *s1, const char *s2)
{
	return (strncmp(s1, s2, strlen(s2)));
}


/*
 * epm_create_node - create a pm node for "component" or "power-management"
 * nodes that properly export PM capability.
 */
pm_node_t *
epm_create_node(md_t *md, md_node_t *nodep)
{
	pm_node_t	*pm_node = NULL;
	pm_pwr_cap_t	*pmcap = NULL;
	pm_pwr_scheme_t *pwrs = NULL;
	pm_res_type_t	res_type;
	pm_resid_t	*rcfg = NULL;
	char		*propstr;
	char		**statesp;
	char 		strtmp[64];
	char		*tok;
	double		*pmratio = NULL;
	int		dlen, idx, toknum;
	int		pm_states;
	uint8_t		*propdata = NULL;
	uint64_t	propint;

	/* Only component nodes have a pm_resource property. */
	if (strcmp(nodep->typep->strp, COMPONENT) == 0) {
		/* PROP_STR "pm_resource". */
		if (md_get_prop_str(md, nodep, PM_RESOURCE, &propstr) ==
		    -1) {
			DBG(zprintf("%s: lack of pm_resource property\n",
			    __func__));
			return (NULL);
		}

		if (strcmp(propstr, CPU) == 0) {
			res_type = PM_RES_CPU;
		} else {
			DBG(zprintf("%s: Unrecognized pm_resource %s\n",
			    __func__, propstr));
			return (NULL);
		}
	} else if ((strcmp(nodep->typep->strp, MEM_GROUPING) == 0) ||
	    (strcmp(nodep->typep->strp, MEM_REGION) == 0)) {
		res_type = PM_RES_MEM;
	} else {
		DBG(zprintf("%s: Unrecognized node type %s\n", __func__,
		    nodep->typep->strp));
		return (NULL);
	}

	/* allocate a new PM node */
	pm_node = Xcalloc(1, pm_node_t);

	/* PROP_VAL "id" */
	if (md_get_prop_val(md, nodep, "id", &propint) == -1)
		propint = -1;
	pm_node->pn_nodeid = (int)propint;

	/*
	 * If the "type" field exists in the node, use it. Otherwise, use
	 * the actual node type.
	 */
	if (md_get_prop_str(md, nodep, TYPE, &propstr) == -1) {
		pm_node->pn_type = Xmalloc(strlen(nodep->typep->strp) + 1);
		strcpy(pm_node->pn_type, nodep->typep->strp);
	} else {
		pm_node->pn_type = Xmalloc(strlen(propstr) + 1);
		strcpy(pm_node->pn_type, propstr);
	}

	/* PROP_STR "name" */
	if (md_get_prop_str(md, nodep, "name", &propstr) < 0) {
		DBG(zprintf("%s: lack of name property\n", __func__));
		goto pcn_err;
	} else {
		pm_node->pn_pmname = Xmalloc(strlen(propstr) + 1);
		strcpy(pm_node->pn_pmname, propstr);
	}

	/* PROP_DATA "pm_states" */
	md_get_prop_data(md, nodep, "pm_states", &propdata, &dlen);

	/* Not all PM nodes have power states (pm_states) */
	if (propdata) {
		/* Delete trailing NULL at end of pm_states string */
		dlen--;
		ASSERT((dlen > 0) && (propdata[0] != 0));

		/*
		 * Count the null terminated <performace, power state> string
		 * pairs. Sample PRI excerpt for CPUs:
		 * pm_states = {
		 *	"500 HalfSpeed",
		 *	"625 FiveEighthsSpeed",
		 *	"750 ThreeFourthsSpeed",
		 *	"875 SevenEighthsSpeed",
		 *	"1000 FullSpeed"
		 * };
		 *
		 * [i]==0(below) signifies the NULL which terminates a
		 * pm_state string. pm_states count the number of pm_state
		 * strings found. The above example has 5 pm_states.
		 */
		for (idx = 1, pm_states = 1; idx < dlen; idx++) {
			if (propdata[idx] == 0)
				pm_states++;
		}

		/*
		 * Malloc a set of pointers that will point to the pm_states
		 * strings
		 */
		statesp = Xmalloc(pm_states * sizeof (char *));

		/* strdup the 1st pm_state string */
		statesp[0] = Xstrdup((char *)&propdata[0]);

		/* strdup the rest of the pm_state strings */
		for (idx = 1, pm_states = 1; idx < dlen; idx++) {
			if (propdata[idx] == 0)
				statesp[pm_states++] =
				    Xstrdup((char *)&propdata[idx + 1]);
		}

		/* Populate the pm_pwr_scheme structure in the pm_node */
		pwrs = Xcalloc(1, pm_pwr_scheme_t);
		pwrs->pps_count = pm_states;
		pwrs->pps_curlvl = 0;
		pmcap = Xcalloc(pm_states, pm_pwr_cap_t);
		pmratio = Xcalloc(pm_states, double);
		for (idx = 0; idx < pm_states; idx++) {
			pmcap[idx].ppc_lvl = idx;
			tok = strtok(statesp[idx], " ");   /* performance */
			pmcap[idx].ppc_perf = atoi(tok);
			tok = strtok(NULL, " ");	/* power state name */
			pmcap[idx].ppc_lname = Xmalloc(strlen(tok) + 1);
			strcpy(pmcap[idx].ppc_lname, tok);
			/*
			 * Assume highest pwr state is 1000.  If not, then
			 * we will have to change how pmratio[] is calcualted
			 * because we need to know the highest pwr state first
			 */
			pmratio[idx] = pmcap[idx].ppc_perf / 1000.0;
			Xfree(statesp[idx]);
		}
		pwrs->pps_pmcap = pmcap;
		Xfree(statesp);
		pwrs->pps_ratio = pmratio;

		/* PROP_STR "pm_cookie" */
		if (md_get_prop_str(md, nodep, "pm_cookie",
		    &propstr) == -1) {
			DBG(zprintf("%s: lack of pm_cookie property, "
			    "assume auto\n", __func__));
			pwrs->pps_pwr_ctl.pp_method = PWRCTL_METHOD_AUTO;
		} else {
			if (strcmp(pm_node->pn_type, MEM_REGION) == 0) {
				pwrs->pps_pwr_ctl.pp_method =
				    PWRCTL_METHOD_AUTO;
			} else {
				pwrs->pps_pwr_ctl.pp_method =
				    PWRCTL_METHOD_HVCTL;
			}

			pwrs->pps_pwr_ctl.pp_cookie =
			    Xmalloc(strlen(propstr) + 1);
			strcpy(pwrs->pps_pwr_ctl.pp_cookie, propstr);
		}

		/* PROP_STR "pm_dependency" */
		if (md_get_prop_str(md, nodep, "pm_dependency",
		    &propstr) == -1)
			pwrs->pps_deps = PM_DEP_NONE;
		else {
			strcpy(strtmp, propstr);
			tok = strtok(strtmp, " ");
			for (; tok; tok = strtok(NULL, " ")) {
				if (epm_substr(tok, "StrictParental") == 0)
					pwrs->pps_deps |= PM_DEP_STRICTPARENTAL;
				else if (epm_substr(tok, "Performance") == 0)
					pwrs->pps_deps |= PM_DEP_PERFORMANCE;
			}
		}

		/* PROP_STR "pm_coordination" */
		if (md_get_prop_str(md, nodep, "pm_coordination",
		    &propstr) == -1)
			pwrs->pps_coord = PM_COORD_NONE;
		else {
			strcpy(strtmp, propstr);
			tok = strtok(strtmp, " ");
			for (; tok; tok = strtok(NULL, " ")) {
				if (epm_substr(tok, "NeedOne") == 0)
					pwrs->pps_coord |= PM_COORD_NEEDONE;
				else if (epm_substr(tok, "Defrag") == 0)
					pwrs->pps_coord |= PM_COORD_DEFRAG;
				else if (epm_substr(tok,
				    "HigherPowerSaving") == 0) {
					pwrs->pps_coord |= PM_COORD_HPS;
					num_hps_nodes++;
					hps_enabled = true;
				}
			}
		}
		pm_node->pn_pwrs = pwrs;
	}

	/*
	 * Populate the pm_resid_t structure. Obtain pi_res_name and
	 * pi_id_val from pm_mapping data. Currently only "CPU"
	 * nodes have pm_mapping fields.
	 */
	if (md_get_prop_str(md, nodep, PM_MAPPING, &propstr) == 0) {
		toknum = epm_get_toknum(propstr, " ");
		if (toknum == 2) {
			rcfg = Xcalloc(1, pm_resid_t);
			rcfg->pi_res_type = res_type;
			strcpy(strtmp, propstr);
			tok = strtok(strtmp, " ");
			rcfg->pi_res_name = Xmalloc(strlen(tok) + 1);
			strcpy(rcfg->pi_res_name, tok);
			tok = strtok(NULL, " ");
			rcfg->u.cpu.pi_id_val = atoi(tok);
		} else {
			DBG(zprintf("%s: unsupported pm_mapping: %s\n",
			    __func__, propstr));
			goto pcn_err;
		}
	/* Populate fields for memory nodes */
	} else if ((strcmp(pm_node->pn_type, MEM_REGION) == 0) ||
	    (strcmp(pm_node->pn_type, MEM_GROUPING) == 0)) {
		rcfg = Xcalloc(1, pm_resid_t);
		rcfg->pi_res_type = res_type;
		if (md_get_prop_val(md, nodep, "id", &propint) == 0) {
			rcfg->u.mem.pi_id_val = propint;
		} else {
			DBG(zprintf("%s: Unable to obtain memory node id.\n",
			    __func__));
		}

		if (strcmp(pm_node->pn_type, MEM_REGION) == 0) {
			rcfg->pi_res_name = Xmalloc(strlen(MEM_REGION) + 1);
			strcpy(rcfg->pi_res_name, MEM_REGION);
		} else if (strcmp(pm_node->pn_type, MEM_GROUPING) == 0) {
			rcfg->pi_res_name = Xmalloc(strlen(MEM_GROUPING) + 1);
			strcpy(rcfg->pi_res_name, MEM_GROUPING);
		}
	}

	if (strcmp(pm_node->pn_type, MEM_REGION) == 0) {
		/*
		 * Since regions are not power managed, we don't care about
		 * their fma status.
		 */
		rcfg->pi_fma_status = 0;

		if (md_get_prop_val(md, nodep, "base", &propint) != -1) {
			rcfg->u.mem.mem_reg.pi_reg_base = propint;
		} else {
			DBG(zprintf("%s: Lack of memory base "
			    "property.\n", __func__));
			goto pcn_err;
		}
		if (md_get_prop_val(md, nodep, "size", &propint) != -1) {
			rcfg->u.mem.mem_reg.pi_reg_size = propint;
		} else {
			DBG(zprintf("%s: Lack of memory size "
			    "property.\n", __func__));
			goto pcn_err;
		}
	} else if (rcfg) {
		/* Get FMA status for CPUs and memory groupings */
		if (GET_FMA_RESOURCE_STATUS(rcfg,
		    &rcfg->pi_fma_status) != 0) {
			/*
			 * During testing, it's been observed that
			 * this call sometimes fails with a TIMEOUT.
			 * Upon retry, it succeeds.  We can't tell from
			 * here the reason why this fails, we just know
			 * it has failed.  So just to be sure, we'll
			 * retry 1 time.
			 */
			if (GET_FMA_RESOURCE_STATUS(rcfg,
			    &rcfg->pi_fma_status) != 0)
				rcfg->pi_fma_status = FMA_STATUS_UNKNOWN;
		}
	}

	pm_node->pn_rcfg = rcfg;

	return (pm_node);

pcn_err:
	if (pmcap) {
		for (idx = 0; idx < pm_states; idx++) {
			Xfree(pmcap[idx].ppc_lname);
		}
		Xfree(pmcap);
	}

	if (pwrs) {
		Xfree(pwrs->pps_pwr_ctl.pp_cookie);
		Xfree(pwrs);
	}

	if (rcfg) {
		Xfree(rcfg->pi_res_name);
		Xfree(rcfg);
	}

	if (pm_node) {
		Xfree(pm_node->pn_type);
		Xfree(pm_node->pn_pmname);
		Xfree(pm_node);
	}

	return (NULL);
}

/*
 * epm_enqueue_node - builds up platform power manageable device tree.
 * This function is not meant to be used to randomly add a node to the PM
 * device tree.  It is meant to be called by epm_scan_pm_resources() to add
 * nodes while observing the md_scan_dag traversing order which is, from
 * the PRI root node, following the "fwd" arc collecting nodes.
 * The processor and memory grouping nodes are the same rank even though they
 * are of different types. The parent type is passed in as a comparison to
 * show that these nodes should be enqueued at the same rank since they have
 * the same parent.
 */
void
epm_enqueue_node(pm_node_t *last_enq, pm_node_t *newnode, char *parenttype)
{
	pm_node_t	*parent_nodep;
	char		*last_enq_type;
	char		*newnode_type;
	char		*pn_type;

	last_enq_type = last_enq->pn_type;
	newnode_type = newnode->pn_type;

	parent_nodep = last_enq->pn_parent;
	if (parent_nodep)
		pn_type = parent_nodep->pn_type;

	/*
	 * If the parent exists and the types are the same or if the type
	 * of the last enqueued node is the same as the new node, the new
	 * node is the same rank as the last added node.
	 */
	if ((parent_nodep && parenttype &&
	    (strcmp(pn_type, parenttype) == 0)) ||
	    (strcmp(last_enq_type, newnode_type) == 0)) {
		newnode->pn_rank = last_enq->pn_rank;
		last_enq->pn_sibling = newnode;
		newnode->pn_parent = last_enq->pn_parent;
		return;
	}

	/* check type/rank above the last added node */
	while (parent_nodep) {
		if (strcmp(pn_type, newnode_type) == 0)
			break;
		else
			parent_nodep = parent_nodep->pn_parent;
	}

	if (parent_nodep) {
		/* newnode is at parent/grandparent rank */
		newnode->pn_rank = parent_nodep->pn_rank;
		newnode->pn_parent = parent_nodep->pn_parent;
		parent_nodep->pn_sibling = newnode;
	} else {
		/* found none above, so child rank */
		newnode->pn_rank = last_enq->pn_rank + 1;
		last_enq->pn_child = newnode;
		newnode->pn_parent = last_enq;
	}
}

/*
 * Update components' power states and status
 */
void
epm_update_pwr(int count, pm_node_t **nodepp, uint16_t *new_states,
    pwrctl_method_t method, pwr_req_status_t *status)
{
	pm_pwr_scheme_t	*pwrs;
	int 		i;

	for (i = 0; i < count; i++) {
		pwrs = epm_get_pwr_scheme(nodepp[i], method);
		DBG(zprintf("%s: %s: ", __func__, nodepp[i]->pn_pmname));
		ASSERT(pwrs);
		switch (status[i]) {
		case PS_SUCCESS:
			pwrs->pps_curlvl = new_states[i];
			DBG(zprintf("new_state = %d, ", pwrs->pps_curlvl));
			/*FALLTHROUGH*/
		case PS_EINVAL:
		case PS_EILLEGAL:
			pwrs->pps_status = PM_STAT_KNOWN;
			DBG(zprintf("pps_status = %d, \n", pwrs->pps_status));
			break;

		case PS_TIMEDOUT:
		case PS_FAILURE:
		default:
			pwrs->pps_status = PM_STAT_UNKNOWN;
			DBG(zprintf("%s: status = %d\n",
			    __func__, (int)status[i]));
			break;
		}
	}
}

/*
 * epm_pwr_recovery - attempt to recover from ldm_set_power_states failure.
 *
 * Time outs and unknown failures leave resource states unknown. At minimum,
 * error handling attempts to rediscover the resources' states.
 */
/* ARGSUSED3 */
int
epm_pwr_recover(int setcc, int count, char **ckp, pwr_req_status_t *old_st,
    uint16_t new_state, pwrctl_method_t method, pm_node_t **npp,
    int *res_status)
{
	pm_pwr_scheme_t		*pwrs;
	pwr_req_status_t	cc, *new_st;
	int			i;
	int			ret = setcc;
	uint16_t		*cur_states;

	/*
	 * Retrieve resources' power states
	 */
	if (setcc & (PWR_OP_ETIMEDOUT | PWR_OP_EFAILURE)) {
		cur_states = Xcalloc(count, uint16_t);
		new_st = Xcalloc(count, pwr_req_status_t);
		cc = GET_POWER_STATES(count, ckp, cur_states, new_st);
		if (cc != PS_SUCCESS)
			DBG(zprintf("Warning: %s: Unable to recover\n",
			    __func__));
		epm_update_pwr(count, npp, cur_states, method, new_st);
		Xfree(cur_states);
		Xfree(new_st);
	}

	/*
	 * Update resource status.
	 */
	for (i = 0; i < count; i++) {
		pwrs = epm_get_pwr_scheme(npp[i], method);
		ASSERT(pwrs);
		if (pwrs->pps_status == PM_STAT_KNOWN) {
			if (pwrs->pps_curlvl > new_state)
				res_status[i] = LDME_RES_STUCK_HIGH;
			else if (pwrs->pps_curlvl < new_state)
				res_status[i] = LDME_RES_STUCK_LOW;
			else
				res_status[i] = LDME_RES_GOOD;
		} else	/* PM_STAT_UNKNOWN */
			res_status[i] = LDME_RES_UNKNOWN_POWER;
	}

	return (ret);
}

bool_t
epm_match_res(pm_resid_t *res, pm_node_t *node)
{
	pm_resid_t	*rcfg;
	bool_t		ret = false;

	if (!node->pn_rcfg)
		return (false);

	rcfg = node->pn_rcfg;
	switch (res->pi_res_type) {
	case PM_RES_CPU:
		if ((res->pi_res_type == rcfg->pi_res_type) &&
		    (res->u.cpu.pi_id_val == rcfg->u.cpu.pi_id_val))
			ret = true;
		break;
	case PM_RES_MEM:
		if ((res->pi_res_type == rcfg->pi_res_type) &&
		    (strcmp(res->pi_res_name, rcfg->pi_res_name) == 0) &&
		    (res->u.mem.pi_id_val == rcfg->u.mem.pi_id_val))
			ret = true;
		break;
	default:
		DBG(zprintf("%s: unsupported resource type %d\n",
		    __func__, (int)res->pi_res_type));
		break;
	}

	return (ret);
}

/*
 * pmi_find_node_by_type - Find a node in the PM tree given a node
 * type and ID.
 *
 * Input: node - Pointer to a PM node to start/continue search from
 *        node_type - Character string for node type. This type can be
 *	      defined by the "type" property if it exists. If not,
 *	      it is defined by the actual node type (e.g. memory-grouping).
 *        child_id - A PRI ID to match against the pn_nodeid of the child node
 *        parent_id - A PRI ID to match against the pn_nodeid of the parent
 *            node. An id of -1 (INVALID_ID) means that child ids are unique
 *            so it is not necessary to look at the parent id while searching.
 * Output: pm_node_t * - A pointer to the PM node that matches the
 *                       the input, or NULL if no match
 *
 * Recursively walk the PM tree, depth (child) first, comparing the node
 * type (pn_type) and node ID (pn_nodeid) to the input.  If a match is
 * found, return a pointer to the PM node, else return NULL.
 */
pm_node_t *
pmi_find_node_by_type(pm_node_t *node, char *node_type, int child_id,
    int parent_id)
{
	pm_node_t	*child, *ntmp;
	char		*typep = NULL;

	/*
	 * If a caller doesn't know where to start, it can indicate to start
	 * from the root node by passing NULL to the second argument
	 */
	if (node == NULL) {
		if (pm_root_node != NULL)
			node = pm_root_node;
		else
			return (NULL);
	}

	typep = node->pn_type;

	DBG_BLOCK {
		if ((strcmp(typep, node_type) == 0) && node->pn_pmname &&
		    node->pn_nodeid) {
			DBG6(zprintf("%s: At node %s, node_type(%s), "
			    "node_id(%d)\n",
			    __func__, node->pn_pmname, typep,
			    node->pn_nodeid));
		}
	}

	/*
	 * Compare the current node type and parent and child IDs to the
	 * input. If INVALID_ID is passed in, the parent id does not have to
	 * be matched.
	 */
	if ((strcmp(typep, node_type) == 0) &&
	    (node->pn_nodeid == child_id) &&
	    ((parent_id == INVALID_ID) ||
	    (node->pn_parent->pn_nodeid == parent_id))) {
		DBG6(zprintf("%s: Matched node %s\n", __func__,
		    node->pn_pmname));
		return (node);
	}

	/* Recurse the PM tree */
	for (child = node->pn_child; child; child = child->pn_sibling) {
		ntmp = pmi_find_node_by_type(child, node_type, child_id,
		    parent_id);
		if (ntmp)
			return (ntmp);
	}
	return (NULL);
}

/*
 * epm_find_node - return a pointer to the node that matches the given
 * resource id.
 */
pm_node_t *
epm_find_node(pm_resid_t *res, pm_node_t *node, pm_node_t *found_node)
{
	pm_node_t	*child, *ntmp;
	pm_resid_t	*rcfg;
	uint64_t	id;

	if (node)
		DBG5(zprintf("%s: visiting node_id(%d), rank(%d)\n", __func__,
		    node->pn_nodeid, node->pn_rank));

	if (found_node)
		return (found_node);

	/*
	 * If a caller doesn't know where to start, it can indicate to start
	 * from the root node by passing NULL to the second argument
	 */
	if (node == NULL) {
		if (pm_root_node != NULL)
			node = pm_root_node;
		else
			return (NULL);
	}

	if (epm_match_res(res, node) == true) {
		DBG5(zprintf("%s: Found node_id(%d), rank(%d)\n",
		    __func__, node->pn_nodeid, node->pn_rank));
		return (node);
	}

	switch (res->pi_res_type) {
	case PM_RES_CPU:
		id = res->u.cpu.pi_id_val;
		break;

	case PM_RES_MEM:
		id = res->u.mem.pi_id_val;
		break;

	default:
		DBG(zprintf("%s: unsupported resource type %d\n",
		    __func__, (int)res->pi_res_type));
		return (NULL);
	}

	for (child = node->pn_child; child; child = child->pn_sibling) {
		rcfg = child->pn_rcfg;
		if (!rcfg && ((id > child->pn_leafhi) ||
		    (id < child->pn_leaflo))) {
			DBG5(zprintf("%s: skip node_id(%d), rank(%d)\n",
			    __func__, child->pn_nodeid, child->pn_rank));
			continue;
		}
		ntmp = epm_find_node(res, child, NULL);
		if (ntmp)
			return (ntmp);
	}

	return (NULL);
}

void
epm_record_range(pm_node_t *node)
{
	pm_node_t	*rl, *rr;
	pm_node_t	*visited;
	int		kc;

	/* record range on leaf nodes */
	if (!node->pn_child) {
		if (node->pn_rcfg) {
			node->pn_leaflo = node->pn_nodeid;
			node->pn_leafhi = node->pn_nodeid;
		} else {
			node->pn_leaflo = NO_LEAF;
			node->pn_leafhi = NO_LEAF;
		}
		node->kids_count = 0;
		return;
	}

	/*
	 * count kids, find out leftmost and rightmost node that are
	 * either leaf node annotated with "pm_mapping" property or
	 * parent that contains such leaf(leaves).
	 */
	rr = rl = NULL;
	for (kc = 0, visited = node->pn_child; visited;
	    visited = visited->pn_sibling) {
		kc++;
		if (visited->pn_leaflo > -1) {
			if (!rl)
				rl = visited;
			/* move up non-leaf righmost child */
			rr = visited;
		}
	}

	node->kids_count = kc;
	if (rl) {
		node->pn_leaflo = rl->pn_leaflo;
		node->pn_leafhi = rr->pn_leafhi;
	} else {
		node->pn_leaflo = NO_LEAF;
		node->pn_leafhi = NO_LEAF;
	}
}

/*
 * epm_update_range - updates leaf node range in its ancestor nodes
 */
void
epm_update_range(pm_node_t *node)
{
	pm_node_t	*child;

	for (child = node->pn_child; child; child = child->pn_sibling)
		epm_update_range(child);

	epm_record_range(node);
}

/*
 * pmi_get_power_all - update power states for all power manageable resources
 * of the specified type.
 */
/*ARGSUSED*/
int
pmi_get_power_all(pm_res_type_t res_type, pm_node_t *topnode)
{
	pm_node_t	*fnp, *leaf;
	pm_node_t	**nodepp;
	int		i, count;
	int		ret = PWR_OP_OK;

	if (topnode == NULL)
		topnode = pm_root_node;

	/*
	 * Depth first top down traverse to update leaf children's
	 * power states, as well as parents that supports multiple
	 * power states.
	 */
	for (fnp = topnode; fnp; fnp = fnp->pn_sibling) {
		/* get power states updated once reach leaf's parent node */
		leaf = fnp->pn_child;
		if (leaf && leaf->pn_rcfg) {
			DBG(zprintf("%s: get power all for range %d -> %d\n",
			    __func__, fnp->pn_leaflo, fnp->pn_leafhi));
			count = fnp->kids_count;
			nodepp = Xmalloc(count * sizeof (pm_node_t *));
			for (i = 0; (i < count); i++) {
				nodepp[i] = leaf;
				leaf = leaf->pn_sibling;
			}

			ret = epm_get_pwr(res_type, count, nodepp);
			ret |= epm_get_pwr(res_type, 1, &fnp);

			Xfree(nodepp);

		} else {
			/* traverse down to child */
			if (leaf)
				ret |= pmi_get_power_all(res_type, leaf);
		}
	}

	return (ret);
}

/*
 * pmi_update_fma_status - update the fma status of the given res_ids with
 * PM's fma status.
 */
void
pmi_update_fma_status(pm_resid_t *ids, int count)
{
	pm_node_t	*nodep;
	int		i;

	for (i = 0; i < count; i++) {
		nodep = epm_find_node(&ids[i], pm_root_node, NULL);
		if (nodep != NULL)
			ids[i].pi_fma_status = nodep->pn_rcfg->pi_fma_status;
		else
			ids[i].pi_fma_status = FMA_STATUS_UNKNOWN;
	}
}

/*
 * pmi_adjust_power - called by the PM engine to change a set of resources'
 * power level.
 *
 * Input:
 *    res_type -  PM'able resource type
 *    req_rcfg_op - flag that indicates the kind of reconfiguration upon the
 *	the resources if reconfigure is required
 *    pmi_pwr - new power level
 *    count - number of resources in request
 *    ids - an array of PM resources that are capable of reconfiguration
 * Output:
 *    status - an array of resources' LDME_RES_* status.
 *    Caller is responsible to free the status.
 *    flags - override flags
 *
 * Return:
 *     PWR_OP_OK - upon success
 *     PWR_OP_FAILURE - upon failure
 */
int
pmi_adjust_power(pm_res_type_t res_type, pm_config_op_t req_rcfg_op,
	pmi_pwr_state_t pmi_pwr, int res_count, pm_resid_t *ids,
	int **status, uint32_t flags)
{
	pm_node_t	**nodepp, *np;
	int		fault_idx, res_idx, batch_num, batch_head;
	int		*res_status;
	int		ret = PWR_OP_OK;

	if (res_count <= 0)
		return (PWR_OP_EINVAL);

	nodepp = Xmalloc(res_count * sizeof (pm_node_t *));
	res_status = Xcalloc(res_count, int);

	/* Sort the list of resources by resource id. */
	if (res_count > 1)
		qsort(ids, res_count, sizeof (*ids), epm_cmp_res_id);

	/*
	 * Fault loop:
	 * Mark the return status for the faulted resources.  During sort,
	 * faulted resources are placed at the beginning of the list
	 * so that no attempts are made to adjust their power.
	 */
	for (fault_idx = 0; fault_idx < res_count; fault_idx++) {
		if (!pmi_is_resource_faulted(&ids[fault_idx]))
			break;

		res_status[fault_idx] = LDME_RES_FAULTED;
	}

	/*
	 * FIXME: The following code is to handle watchdog timeouts by
	 * checking solaris status of domains then applying specific rules
	 * known to prevent watchdogs during reboot looping tests.
	 * This needs to be updated to use proper Zeus<->PM interfaces as
	 * it touches HV directly here.
	 */

	/* pmi_change_to_performance will use pmi_pwr == PPS_FULL_POWER_FORCE */

	/* Power resources that are part of a domain.  Ignore unallocated. */
	if (req_rcfg_op != PCO_NONE &&
	    (pmi_pwr == PPS_LOWEST_POWER || pmi_pwr == PPS_FULL_POWER)) {
		ldom_t *ldp;

		/*
		 * Keep VCPU 0 up for each domain
		 */
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			vcpu_t *vcpu;
			uint64_t pmcpu = 0;
			/*
			 * These values are set to PM_BIG_NUM so they
			 * will be initialized when the first valid
			 * data is encountered.  They're used to find
			 * the lowest valid data.
			 */
			uint64_t low_pcpu = PM_BIG_NUM;
			int low_idx = PM_BIG_NUM;
			vcpu_t *low_vcpu = NULL;

			/* INACTIVE domains are ignored. */
			if (ldp->state != ACTIVE)
				continue;

			/* Now let's keep VCPU 0 up for each LDom */
			CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
				/*
				 * Take each ldom vcpu and find if it's pcpu
				 * is in resources
				 */
				for (res_idx = fault_idx; res_idx < res_count;
				    res_idx++) {
					pmcpu = ids[res_idx].u.cpu.pi_id_val;
					/* Skip ids set to -1 as not valid */
					if (pmcpu == PM_IGNORE_PID)
						continue;
					/* Check if VCPU is the PCPU in res */
					if (vcpu->cpu->phys_id == pmcpu) {
						DBG(zprintf("%s: trans cpu "
						    "%"PRIu64" for ldom %s\n",
						    __func__, pmcpu,
						    ldp->namep));
						/* Save lowest pcpu found */
						if (pmcpu < low_pcpu) {
							low_idx = res_idx;
							low_pcpu = pmcpu;
							low_vcpu = vcpu;
						}
					}
				}
			}
			/* Found VCPU 0? */
			/*  low_vcpu == NULL for unallocated resources */
			if (low_vcpu && low_vcpu->vid == 0) {
				ids[low_idx].u.cpu.pi_id_val = PM_IGNORE_PID;
				DBG(zprintf("%s ldom low cpu is %"PRIu64"\n",
				    ldp->namep, low_pcpu));
			}
		}
	}
	/* End FIXME */

	/*
	 * Restore PPS_FULL_POWER_FORCE to PPS_FULL_POWER to all rest
	 * of routine to contnue as normal.  This only impacts calls
	 * from pmi_change_to_performance which must get through.
	 */
	if (pmi_pwr == PPS_FULL_POWER_FORCE)
		pmi_pwr = PPS_FULL_POWER;

	/*
	 * Continue from where the fault loop left off.  These are all of
	 * the non-faulted resources.  Find the pm_node_t for each given
	 * pm_resid_t.  Also, figure out which resources we want to actually
	 * adjust power to.
	 */
	for (res_idx = fault_idx; res_idx < res_count; res_idx++) {
		/* Skip VCPU 0 of each domain  and other ignored ids */
		if ((res_type == PM_RES_CPU) &&
		    (ids[res_idx].u.cpu.pi_id_val == PM_IGNORE_PID)) {
			nodepp[res_idx] = NULL;
			continue;
		}
		if ((np = epm_find_node(&ids[res_idx], pm_root_node,
		    NULL)) == NULL) {
			*status = res_status;
			Xfree(nodepp);
			return (PWR_OP_EINVAL);
		}

		nodepp[res_idx] = np;
	}

	/*
	 * Skip all the NULL nodepp until we find a batch start so that
	 * a batch doesn't start with a NULL causing a dereference problem
	 * in epm_set_pwr.
	 */
	batch_head = fault_idx;
	while (batch_head < res_count && nodepp[batch_head] == NULL) {
		batch_head++;
	}
	/* If there are any resources left then PM them */
	if (batch_head < res_count) {
		batch_num = res_count - batch_head;
		ret |= epm_adjust_pwr(res_type, batch_num, req_rcfg_op,
		    pmi_pwr, &nodepp[batch_head],
		    &res_status[batch_head], flags);
	}

	*status = res_status;
	Xfree(nodepp);
	return (ret);
}

/*
 * epm_adjust_pwr - internal function called by pmi_adjust_power. It
 * calculates the power change scheme and new power state to transition
 * the resources to.
 */
int
epm_adjust_pwr(pm_res_type_t res_type, int count, pm_config_op_t rcfg_op,
    pmi_pwr_state_t pmi_pwr,  pm_node_t **nodepp, int *res_status,
    uint32_t flags)
{
	pm_config_op_t	rcfg;
	pm_pwr_scheme_t	*pwrs;
	pm_pwrctl_t	*pwrctl;
	int		ret = PWR_OP_OK;
	uint16_t	new_state;

	pwrs = nodepp[0]->pn_pwrs;
	if (pwrs->pps_count <= 1) {
		/* only one power state is supported, done */
		return (ret);
	}

	/*
	 * Compute a power scheme to use. Simple for now because there is
	 * only one software method, enrich the algorithm to support more
	 * sophisticated power schemes.
	 */
	pwrctl = &pwrs->pps_pwr_ctl;
	switch (pwrctl->pp_method) {
	case PWRCTL_METHOD_HVCTL:
		break;
	case PWRCTL_METHOD_AUTO:
		/*
		 * Hardware initiated change of power states.
		 * Potential support in future go here.
		 */
		return (ret);
	default:
		DBG(zprintf("%s: unrecognized power scheme %d\n",
		    __func__, (int)pwrctl->pp_method));
		return (PWR_OP_EINVAL);
	}

	/*
	 * Rest applies to PWRCTL_METHOD_HVTL method.
	 * Support for new method in future branch out here.
	 */

	/*
	 * Calculate new power state.
	 * expand code here to support sophisticated power adjustment
	 * in future.
	 */
	if (pmi_pwr == PPS_LOWEST_POWER)
		new_state = 0;		/* lowest power index */
	else if (pmi_pwr == PPS_FULL_POWER)
		new_state = pwrs->pps_count - 1; /* highest power index */

	/*
	 * Confirm reconfigure operation. If reconfiguration is requested,
	 * unconfigure prior to transitioning from non-zero performance
	 * state to zero performanc state, and configure after the reverse
	 * transitioning. Note, a resource in unknown power state is a
	 * resource unconfigured.
	 */
	rcfg = PCO_NONE;
	if ((rcfg_op == PCO_ADD_RESOURCE) &&
	    (pwrs->pps_pmcap[new_state].ppc_perf > 0))
			rcfg = PCO_ADD_RESOURCE;
	else if ((rcfg_op == PCO_REMOVE_RESOURCE) &&
	    (pwrs->pps_pmcap[new_state].ppc_perf == 0))
			rcfg = PCO_REMOVE_RESOURCE;

	DBG(zprintf("%s: new_state=%d, rcfg=%d\n", __func__,
	    new_state, (int)rcfg));

	/*
	 * epm_set_pwr is called given specific power scheme and new power
	 * state under that scheme.
	 */
	ret |= epm_set_pwr(res_type, count, rcfg, new_state,
	    PWRCTL_METHOD_HVCTL, nodepp, res_status, flags);

	return (ret);
}

/*
 * epm_create_batches
 * Loops over resources and creates "batches."  A batch refers to all
 * resources which have a common parent.  For core disable, this means cpus,
 * with a batch being all the cpus associated with the same core.
 */
static int
epm_create_batches(int *res_count, pm_node_t **nodepp,
    batch_array_t *batch_array, int *res_status)
{
	int	batch_head, batch_num;
	int	count, skipped;
	int	n, nodep_idx;

	count = *res_count;
	batch_num = 0;
	batch_head = 0;
	n = 0;
	skipped = 0;
	for (nodep_idx = 0; nodep_idx < count; nodep_idx++) {
		if (nodepp[nodep_idx] == NULL) {
			skipped++;
			continue;
		}

		if (batch_num++ == 0)
			batch_head = nodep_idx;

		/*
		 * A batch is "complete" if we've reached the last
		 * resource OR if the next resource is being
		 * skipped (i.e. is NULL) because it is already at the
		 * preferred power level OR if the parent of the next
		 * resource is different from the parent of the current
		 * resource. Do not change the order of this if stmt!
		 */
		if ((nodep_idx == count - 1) ||
		    ((nodepp[nodep_idx + 1] == NULL) ||
		    (nodepp[nodep_idx]->pn_parent !=
		    nodepp[nodep_idx + 1]->pn_parent))) {
			batch_array[n].head = batch_head;
			batch_array[n].num = batch_num;
			batch_array[n].nodepp = &nodepp[batch_head];
			batch_array[n].res_status = &res_status[batch_head];
			batch_array[n].status = EPM_SET_PWR_OK;
			n++;
			batch_num = 0;
		}
	}

	/* Resource count adjusted for skipped resources */
	*res_count -= skipped;

	/* Number of batches */
	return (n);
}

/*
 * epm_find_needidx
 * Searches each batch to see if a resource is a NeedOne candidate which cannot
 * be powered down.  For core disable, this is a resource which must always be
 * kept powered on (a single strand with its parent core) due to PM restrictions
 * of the N2/VF chip.  epm_pre_pwer figures out if a NeedOne cases exists for
 * the batch being processed.
 * Status for the epm_pre_pwr op for this batch is stored in the batch.
 */
static void
epm_find_needidx(batch_array_t *batch_array, int total_batches,
    uint16_t new_state, pm_res_type_t res_type, pwrctl_method_t method,
    int count, uint32_t flags)
{
	pm_pwr_scheme_t	*pwrs;
	int	i, n;
	int	needidx = -1;

	for (i = 0; i < total_batches; i++) {
		batch_array[i].needidx = -1;
		batch_array[i].precc = epm_pre_pwr(res_type, batch_array[i].num,
		    new_state, method, batch_array[i].nodepp,
		    &batch_array[i].needidx, flags);
		if (batch_array[i].precc & PWR_OP_ENEEDONE) {
			ASSERT(batch_array[i].needidx >= 0);
			DBG(zprintf("%s: preserving needidx%d, res nodeid %d\n",
			    __func__, needidx, batch_array[i].nodepp[ \
			    batch_array[i].needidx]->pn_nodeid));
			if (batch_array[i].num == 1) {
				batch_array[i].status = EPM_SET_PWR_DONE;
				continue;
			}
		} else if (batch_array[i].precc & (~PWR_OP_ENEEDONE)) {
			batch_array[i].status = EPM_SET_PWR_RECOVER;
			continue;
		}

		/* Process set power data for this batch */
		batch_array[i].pwr_status = Xmalloc(count *
		    sizeof (pwr_req_status_t));
		batch_array[i].states = Xmalloc(count * sizeof (uint16_t));
		batch_array[i].ckp = Xmalloc(count * sizeof (char *));
		for (n = 0; n < batch_array[i].num; n++) {
			pwrs = epm_get_pwr_scheme(batch_array[i].nodepp[n],
			    method);
			ASSERT(pwrs);
			batch_array[i].ckp[n] = pwrs->pps_pwr_ctl.pp_cookie;
			batch_array[i].states[n] = new_state;
		}

		/* Protect Needone node which will be processed later */
		if (batch_array[i].precc & PWR_OP_ENEEDONE) {
			needidx = batch_array[i].needidx;
			pwrs = epm_get_pwr_scheme(
			    batch_array[i].nodepp[needidx], method);
			batch_array[i].states[needidx] = pwrs->pps_curlvl;
			batch_array[i].pwr_status[needidx] = PS_SUCCESS;
		}
	}
}

/*
 * epm_set_pwr_batches
 * Calls epm_set_pwr for each batch to set the power levels as appropriate
 * for the resources in the batch.  The batch may contain a NeedOne case so
 * that is masked out by doing more than one PM operation if the NeedOne
 * resource is in the middle of the batch.  E.g. say it's resource 3 of 0-7
 * then only 0-2 and 4-7 are PM'd via 2 PM ops.  But if it were resource 0 or
 * resource 7 in this case then only one PM op would be needed (1-7 or 0-6).
 * Status for the epm_set_pwr op for this batch is stored in the batch.
 */
static void
epm_set_pwr_batches(batch_array_t *batch_array, int total_batches,
    pwrctl_method_t method)
{
	pwr_req_status_t	cc;
	int			fst, lst, tot, tott;
	int			i;
	int			needidx = -1;

	for (i = 0; i < total_batches; i++) {
		if (batch_array[i].status != EPM_SET_PWR_OK)
			continue;
		needidx = batch_array[i].needidx;
		if (needidx >= 0) {	/* protect resource nodepp[needidx] */
			fst = (needidx == 0) ? (needidx + 1) : 0;
			lst = (needidx == 0) ? (batch_array[i].num - 1):
			    (needidx - 1);
			tot = lst - fst + 1;
			ASSERT(tot > 0);
			tott = tot;
		} else {		/* nothing to be protected */
			fst = 0;
			lst = batch_array[i].num - 1;
			tot = batch_array[i].num;
		}

/* FIXME: get rid of goto by replacing with a while loop */
rem_res:
		/*
		 * Set resources to new power level
		 * This is where all the action happens in powering
		 * cpus and cores up/down for cpus via calls to the HV
		 * or another power management * mechanism.
		 */
		cc = SET_POWER_STATES(tot, &batch_array[i].ckp[fst],
		    &batch_array[i].states[fst],
		    &batch_array[i].pwr_status[fst]);
		if ((batch_array[i].setcc = epm_ps2pwr_status(cc)) !=
		    PWR_OP_OK) {
			batch_array[i].status = EPM_SET_PWR_RECOVER;
			DBG6(zprintf("%s: SET_POWER_STATES FAILED, PS_cc=%d, "
			    " PWR_OP_cc=%d\n",
			    __func__, cc, batch_array[i].setcc));
			continue;
		} else {
			/* Update power states for the batch */
			epm_update_pwr(tot, &batch_array[i].nodepp[fst],
			    &batch_array[i].states[fst], method,
			    &batch_array[i].pwr_status[fst]);
		}

		/* Observe the order of the expressions so don't change 'em */
		if ((needidx >= 0) && (tott < (batch_array[i].num - 1))) {
			/*
			 * We're here only if the protected res id is located
			 * in the middle of the res id list, meaning
			 * 'batch_array[i].num' must be at least 3.
			 */
			fst = needidx + 1;
			lst = batch_array[i].num - 1;
			tot = lst - fst + 1;
			ASSERT(tot > 0);
			tott += tot;
			ASSERT(tott == (batch_array[i].num - 1));
			goto rem_res;
		}
	}
}

/*
 * epm_dr_resources
 * Looks at every batch and pulls out the resources to be DR'd in to their
 * domains, and performs the DR operation.  The error recovery is done in
 * epm_set_pwr.
 * OUT variables:
 * dr_list is the list of resource that were dynamically reconfigured.  It
 * is used as input to process_dr_errors.
 * dr_tot is the size of dr_list
 * rcfg_status is the error list for the dr_list of resources
 */
static int
epm_dr_resources(batch_array_t *batch_array, int total_batches, pm_resid_t *ids,
    int *dr_list, int *rcfg_status, pm_config_op_t rcfg,
    ldm_err_t *results, int *dr_tot, int flags)
{
	int count;
	int i, n;
	int rcfgcc = PWR_OP_OK;

	*dr_tot = 0;
	/* count is a running total of ids that have been accounted for */
	count = 0;

	/*
	 * Traverse the batch array to pull out all resources to DR,
	 * except the Needone candidates and the error batches because
	 * the DR routine expects contiguous resources.  We then map
	 * failures from the DR routine back to the original list of
	 * resources passed into this function at error recovery.
	 */
	for (i = 0; i < total_batches; i++) {
		/* Ignore batches with bad status */
		if (batch_array[i].status != EPM_SET_PWR_OK) {
			count += batch_array[i].num;
			continue;
		}

		/*
		 * Populate ids array with cpus to be half DR'd.
		 * Half DR means a CPU is DR'd out of a domain in
		 * Zeus but not DR'd out of the HV.  The DR out of
		 * the domain means Solaris a Solaris DR via the
		 * drd service.
		 */
		for (n = 0; n < batch_array[i].num; n++) {
			/* Skip the need resource id */
			if (batch_array[i].needidx >= 0 &&
			    batch_array[i].needidx == n)
				continue;

			/*
			 * We use a separate array as we need to
			 * close up the gaps in the DR list which
			 * are bad status resources or needed
			 * resources.
			 * Say, we need to unconfigure cpus 1-5 but
			 * cpu 2 and cpu 5 are an FMA error and a
			 * NeedOne, respectively.  Then ids will look
			 * like [0]->1,[1]->3,[2]->4, with dr_tot of 3.
			 * ids is passed into the DR commands and dr_list
			 * is set to map back to the original list of 1-5,
			 * for error processing.  This is because
			 * pmi_adjust_power expects to return errors in that
			 * format.
			 */
			ids[*dr_tot] = *(batch_array[i].nodepp[n])->pn_rcfg;
			dr_list[(*dr_tot)++] = count + n;
		}
		count += batch_array[i].num;
	}

	if (*dr_tot > 0) {
		/* Perform DR in as a SINGLE operation using ids list */
		rcfgcc = RECONFIGURE_RESOURCES(rcfg, *dr_tot, ids,
		    rcfg_status, results, flags);
		if (rcfgcc != 0) {
			rcfgcc = PWR_OP_ERECONFIGURE;
		} else
			rcfgcc = PWR_OP_OK;
	}

	return (rcfgcc);
}

/*
 * process_epm_pre_pwr_errors
 * Check each batch to see if the status was an error for the epm_pre_pwr
 * call for this batch.  If epm_pre_pwr failed to power up the parent for
 * this batch (the parent core) then the children will all be stuck low
 * (in other words, the strands will all still be parked).
 * Mark the appropriate bits for the pmi_adjust_power error code.
 */
static int
process_epm_pre_pwr_errors(int total_batches, batch_array_t *batch_array)
{
	int i, n;
	int ret = PWR_OP_OK;

	/* Look at each batch to see if it has an epm_pre_pwr failure */
	for (i = 0; i < total_batches; i++) {
		if (batch_array[i].status == EPM_SET_PWR_RECOVER &&
		    batch_array[i].precc != PWR_OP_OK) {
			/*
			 * Couldn't power up parent so each individual resource
			 * child is marked as being stuck low
			 */
			if (batch_array[i].precc & PWR_OP_EPARENTUPFAILED) {
				for (n = 0; n < batch_array[i].num; n++) {
					batch_array[i].res_status[n] =
					    LDME_RES_STUCK_LOW;
				}
			}
			/* Update the return status to include this error */
			ret |= batch_array[i].precc;
			batch_array[i].status = EPM_SET_PWR_DONE;
		}
	}

	return (ret);
}

/*
 * process_epm_set_pwr_errors
 * Check each batch to see if the status was an error for the epm_set_pwr
 * call.  If epm_set_pwr  failed then attempt recovery for the batch.
 * Mark the appropriate bits for the pmi_adjust_power error code.
 */
static int
process_epm_set_pwr_errors(int total_batches, batch_array_t *batch_array,
    uint16_t new_state, pwrctl_method_t method)
{
	int i;
	int ret = PWR_OP_OK;

	/* Check each batch to find epm_set_pwr failure */
	for (i = 0; i < total_batches; i++) {
		if (batch_array[i].status == EPM_SET_PWR_RECOVER &&
		    batch_array[i].setcc != PWR_OP_OK) {
			/* Update the return status after recovery try */
			ret |= epm_pwr_recover(batch_array[i].setcc,
			    batch_array[i].num, batch_array[i].ckp,
			    batch_array[i].pwr_status, new_state, method,
			    batch_array[i].nodepp, batch_array[i].res_status);
			batch_array[i].status = EPM_SET_PWR_DONE;
		}
	}

	return (ret);
}

/*
 * process_dr_errors
 * Check each resource for which DR occurred to see if an error occurred.
 * If an error did occur then set the resource failure status to the
 * appropriate DR error.  Note the mapping from the consolidated DR list
 * which is different from the list of resources passed to epm_set_pwr
 * as the list of resources passed to epm_set_pwr in not contiguous due
 * to NULL references, and other resources which have been masked out.
 * Mark the appropriate bits for the pmi_adjust_power error code.
 */
static int
process_dr_errors(pm_config_op_t rcfg, int rcfgcc, int dr_tot, int *dr_list,
    int *rcfg_status, int *res_status)
{
	int i;
	int ret = PWR_OP_OK;

	/*
	 * Look for DR failures and gather them together, noting
	 * the resources that failed appropriately.  We're mapping
	 * from DR failure status to caller function status indexes.
	 */
	for (i = 0; i < dr_tot; i++) {
		/* For DR failures, set the appropriate resource status */
		if (rcfg_status[i] != LDME_RES_GOOD)
			res_status[dr_list[i]] =
			    (rcfg == PCO_REMOVE_RESOURCE)
			    ? LDME_RES_UNCFG_FAILURE :
			    LDME_RES_CFG_FAILURE;
	}
	/* Update the return status to include this error */
	ret |= rcfgcc;

	return (ret);
}

/*
 * process_needone_status
 * Check each batch to see if any of the resources are the NeedOne
 * category.  If so, then we mark them so they are not treated as
 * being errors by the policy engine when the policy engine returns
 * the status to ldmd.
 * Mark the appropriate bits for the pmi_adjust_power to show a
 * NeedOne special case.
 */
static int
process_needone_status(int total_batches, batch_array_t *batcharray,
    pm_config_op_t rcfg)
{
	int i;
	int ret = PWR_OP_OK;

	for (i = 0; i < total_batches; i++) {
		if (batcharray[i].precc & PWR_OP_ENEEDONE) {

			/*
			 * NeedOne cases which are not allowed to power
			 * down need to be handled as a special case as
			 * they are not errors.
			 */
			batcharray[i].res_status[batcharray[i].needidx] =
			    (rcfg == PCO_REMOVE_RESOURCE) ?
			    LDME_RES_STUCK_HIGH_REQD : LDME_RES_STUCK_HIGH;
			ret |= PWR_OP_ENEEDONE;
		}
	}

	return (ret);
}

/*
 * process_epm_post_pwr
 * Check each batch for any epm_pre_pwr or epm_set_pwr errors.  If there
 * are none then we can throw away the status of epm_post_pwr as ldmd
 * doesn't care about such post PM status errors.
 */
static void
process_epm_post_pwr(int total_batches, batch_array_t *batcharray, int *ret)
{
	int i;

	/*
	 * Make sure there are no preceding failures as their absense means
	 * we don't care about any postcc errors in a batch.
	 */
	for (i = 0; i < total_batches; i++) {
		if (batcharray[i].precc != 0 || batcharray[i].setcc != 0)
			break;
	}

	/*
	 * post_pwr does not affect the first level resources' status
	 * so merely adjust the return code to PWR_OP_OK.
	 */
	if (i == total_batches)
		*ret = PWR_OP_OK;
}

/*
 * epm_post_pwr_batches
 * For each batch not in error, make a call to epm_post_pwr and record
 * the status for processing by process_epm_post_pwr.
 */
static void
epm_post_pwr_batches(int total_batches, batch_array_t *batcharray,
    uint16_t new_state, pwrctl_method_t method, pm_res_type_t res_type,
    uint32_t flags)
{
	int i;

	for (i = 0; i < total_batches; i++) {
		if (batcharray[i].status != EPM_SET_PWR_OK) {
			DBG6(zprintf("%s: batch(%d) status != "
			    "EPM_SET_PWR_OK\n", __func__, i));
			continue;
		}
		batcharray[i].postcc = epm_post_pwr(res_type, batcharray[i].num,
		    new_state, method, batcharray[i].nodepp, flags);
		if (batcharray[i].postcc != PWR_OP_OK)
			batcharray[i].status = EPM_SET_PWR_RECOVER;
		else
			batcharray[i].status = EPM_SET_PWR_DONE;
	}
}

/*
 * epm_set_pwr - set resources to "new_state" according to the given power
 * scheme method.  This routine can handle 3 different cases which causes
 * differing sections of code to be executed.  These are rcfg = PCO_NONE,
 * rcfg = PCO_ADD_RESOURCE meaning DR the resource into the domains, and
 * rcfg = PCO_REM_RESOURCE meaning DR the resource out of the domains.  All
 * of these have common calls to epm_pre_pwr, epm_set_pwr, and epm_post_pwr.
 */
int
epm_set_pwr(pm_res_type_t res_type, int count, pm_config_op_t rcfg,
    uint16_t new_state,  pwrctl_method_t method, pm_node_t **nodepp,
    int *res_status, uint32_t flags)
{
	batch_array_t	*batch_array = NULL;
	ldm_err_t	results;
	pm_resid_t	*ids = NULL;
	int		*before_dr = NULL, *after_dr = NULL;
	int		*rcfg_status = NULL;
	int		dr_tot = 0;
	int		i, rcfgcc = PWR_OP_OK;
	int		ret = PWR_OP_OK;
	int		total_batches = 0;

	epm_debug_print_op((char *)__func__, count, new_state, nodepp);

	/* Only PWRCTL_METHOD_HVCTL method is supported now. */
	if (method != PWRCTL_METHOD_HVCTL)
		return (PWR_OP_EINVAL);

	/*
	 * This section splits all the sorted resources into batches.
	 * A batch is a group of resources with a common parent whereby
	 * we have an ability to bundle siblings with a common parent
	 * into a single operational call to the HV.
	 *
	 * For CPUs, a batch is all strands on the same core for which
	 * we try to do the PM operation for.  As we can be processing
	 * different CPUs from different cores at once, then we can
	 * have more than once batch.
	 *
	 * We execute operation on strands and cores at the batch level
	 * due to limitations of the HV support.  But we need to execute
	 * CPU DR calls as a single operation due to the expense of this
	 * operation on certain systems.
	 */
	batch_array = Xcalloc(count, batch_array_t);
	total_batches = epm_create_batches(&count, nodepp,
	    batch_array, res_status);

	/*
	 * Loop over batches and find needidx for each batch.  The needidx
	 * being set means a particular batch been identified as having
	 * a NeedOne requirement.  This means a resource in that batch
	 * must always be powered up.  For example, a strand in a core
	 * always must be powered up per physical CPU chip for the likes
	 * of N2 & VF chips.
	 */
	epm_find_needidx(batch_array, total_batches, new_state, res_type,
	    method, count, flags);

	/* PCO_NONE means no DR ops to be done */
	if (rcfg != PCO_NONE) {
		ids = Xmalloc(count * sizeof (pm_resid_t));
		rcfg_status = Xcalloc(count, int);
	}

	/* DR out op occurs first if this is a power down/remove res op */
	if (rcfg == PCO_REMOVE_RESOURCE) {
		before_dr = Xcalloc(count, int);
		rcfgcc = epm_dr_resources(batch_array, total_batches, ids,
		    before_dr, rcfg_status, rcfg, &results, &dr_tot, flags);
	}

	if (rcfgcc == PWR_OP_OK) {
		/* Power resources up or down. */
		epm_set_pwr_batches(batch_array, total_batches, method);
	} else
		DBG6(zprintf("%s: epm_dr_resources(PCO_REMOVE_RESOURCE) FAILED,"
		    " rcfgcc=%d\n", __func__, rcfgcc));

	/* DR in operation occurs after power up/add resource operation. */
	if (rcfg == PCO_ADD_RESOURCE) {
		after_dr = Xcalloc(count, int);
		rcfgcc = epm_dr_resources(batch_array, total_batches, ids,
		    after_dr, rcfg_status, rcfg, &results, &dr_tot, flags);
	}

	/*
	 * Handle power dependency after the requested power transition
	 * This might be powering down a core after all strands are disabled.
	 */
	if (rcfgcc == PWR_OP_OK) {
		epm_post_pwr_batches(total_batches, batch_array,
		    new_state, method, res_type, flags);
	}
	else
		DBG6(zprintf("%s: epm_dr_resources(PCO_ADD_RESOURCE) FAILED,"
		    " rcfgcc=%d\n", __func__, rcfgcc));


	/*
	 * Error handling
	 * Each batch may have had an error calling epm_pre_pwr, epm_set_pwr,
	 * epm_post_pwr, or either a DR in or out.  Those errors are stored
	 * on a per batch basis (except the DR cases) and processed here.
	 */

	ret |= process_epm_pre_pwr_errors(total_batches, batch_array);

	if (rcfg == PCO_REMOVE_RESOURCE)
		ret |= process_dr_errors(rcfg, rcfgcc, dr_tot, before_dr,
		    rcfg_status, res_status);

	ret |= process_epm_set_pwr_errors(total_batches, batch_array,
	    new_state, method);

	if (rcfg == PCO_ADD_RESOURCE)
		ret |= process_dr_errors(rcfg, rcfgcc, dr_tot, after_dr,
		    rcfg_status, res_status);

	/* Process batch epm_post_pwr failures to return updated status */
	process_epm_post_pwr(total_batches, batch_array, &ret);

	/* Update res_status in Needone case */
	ret |= process_needone_status(total_batches, batch_array, rcfg);

	/* Free batch array mallocs */
	for (i = 0; i < total_batches; i++) {
		Xfree(batch_array[i].pwr_status);
		Xfree(batch_array[i].states);
		Xfree(batch_array[i].ckp);
	}
	Xfree(batch_array);
	Xfree(before_dr);
	Xfree(after_dr);
	Xfree(ids);
	Xfree(rcfg_status);

	return (ret);
}

/*
 * epm_get_pwr_scheme - get power scheme(pwrs) by method
 */
pm_pwr_scheme_t *
epm_get_pwr_scheme(pm_node_t *nodep, pwrctl_method_t method)
{
	pm_pwr_scheme_t *pwrs;

	/* There is only one pwrs created in epm_create_node() */
	for (pwrs = nodep->pn_pwrs; pwrs; pwrs = pwrs->pps_next)
		if (pwrs->pps_pwr_ctl.pp_method == method)
			break;
	return (pwrs);
}

/*
 * epm_has_perf - return true if node is at a power state that delivers
 * performance, false if zero performance could be delivered.
 */
bool_t
epm_has_perf(pm_node_t *pn, int inc)
{
	pm_pwr_scheme_t	*pwrs;

	for (pwrs = pn->pn_pwrs; pwrs; pwrs = pwrs->pps_next) {
		if (STATE_TO_PERF(GET_CUR_LEVEL(pwrs, inc), pwrs) == 0)
			return (false);
	}
	return (true);
}

/*
 * epm_pre_pwr - process dependency and coordination prior to set
 * resources to the new power state.
 */
int
epm_pre_pwr(pm_res_type_t res_type, int count, uint16_t new_state,
    pwrctl_method_t method, pm_node_t **npp, int *needidx, uint32_t flags)
{
	bool_t		all_kids_at_lowest = false;
	pm_node_t	*parentp, *gpp, *pn;
	pm_pwr_scheme_t	*pa_pwrs;
	pwrctl_method_t	pa_method;
	int		i, total_kids, lowest_kids;
	int		pa_res_status;
	int		protect_id = -1;
	int		ret = PWR_OP_OK;
	uint16_t	pa_new_lvl;

	epm_debug_print_op((char *)__func__, count, new_state, npp);

	/*
	 * NeedOne is an optional property attached to PM'able node,
	 * it means that at all time at least one node out of the node
	 * and its sibling nodes must be kept up to delivery performance.
	 *
	 * If a node has NeedOne attached, then checking the NeedOne at the
	 * node's level is sufficient, no need to check its parent's level.
	 *
	 * If a node does not has NeedOne and is under StrictParental
	 * control, meaning the parent's power level will be affected
	 * by the sum of its children's power level, then must check
	 * parent node's NeedOne.
	 *
	 * In either case if NeedOne check failed, subtract a node from
	 * the original node list.
	 */

	if (STATE_TO_PERF(new_state, epm_get_pwr_scheme(npp[0], method)) == 0) {
#ifdef  PM_HV_WORKAROUND
		/*
		 * Keep platform's lowest strands ON per CPU chip
		 */
		for (i = 0; i < count; i++) {
			/* FIXME: CR 6818078 needs PM_HV_WORKAROUND */
			if (npp[i]->pn_flag & PN_FLAG_KEEPMEUP) {
				*needidx = i;
				DBG(zprintf("%s: protect needidx %d \n",
				    __func__, *needidx));
				return (ret |= PWR_OP_ENEEDONE);
			}
		}
#else
		/* FIXME: One CR 6818078 is fixed then this code can be used */

		/* First, check NeedOne at current nodes' level  */

		if (PM_COORD_IS_NEEDONE(npp[0]) ||
		    (PM_DEP_IS_STRICTPARENTAL(npp[0]->pn_parent) &&
		    PM_COORD_IS_NEEDONE(npp[0]->pn_parent))) {

			parentp = npp[0]->pn_parent;
			total_kids = parentp->kids_count;

			/* count kids already at lowest, including unknown */
			lowest_kids = 0;
			for (cn = parentp->pn_child; cn; cn = cn->pn_sibling)
				if (!epm_has_perf(cn, PM_ASSUME_LOW))
					lowest_kids++;

			/* add would-be lowest kids, excluding unknown */
			for (i = 0; i < count; i++) {
				pwrs = epm_get_pwr_scheme(npp[i], method);
				if ((epm_has_perf(npp[i], PM_ASSUME_LOW)) &&
				    (STATE_TO_PERF(new_state, pwrs) == 0)) {
					lowest_kids++;
					if (protect_id < 0)
						protect_id = i;
				}
			}
			DBG(zprintf("%s: if Needone, preserve protect_id=%d\n",
			    __func__, protect_id));

			if (lowest_kids == total_kids) {
				all_kids_at_lowest = true;
				if (PM_COORD_IS_NEEDONE(npp[0])) {
					*needidx = protect_id;
					return (ret |= PWR_OP_ENEEDONE);
				}
			}
		}

#endif  /* PM_HV_WORKAROUND */
		/*
		 * if no need to keep a parent (or its sibling) up, or,
		 * not all kids are going to the lowest, return.
		 */
		if (!IS_PM_CAPABLE(npp[0]->pn_parent) ||
		    !(PM_DEP_IS_STRICTPARENTAL(npp[0]->pn_parent) &&
		    PM_COORD_IS_NEEDONE(npp[0]->pn_parent)) ||
		    (all_kids_at_lowest == false)) {
			return (ret);
		}

		/*
		 * StrictParental and NeedOne parent whose kids are about to
		 * go off.  Check to ensure at least one sibling at the
		 * parent's level stay up.
		 */
		parentp = npp[0]->pn_parent;
		gpp = parentp->pn_parent;
		lowest_kids = 0;
		total_kids = gpp->kids_count;

		/* count existing lowest parent siblings, including unknown */
		for (pn = gpp->pn_child; pn; pn = pn->pn_sibling) {
			epm_debug_print_op((char *)__func__, 1, 0, &pn);
			if (!epm_has_perf(pn, PM_ASSUME_LOW))
				lowest_kids++;
		}

		/* add would-be lowest new parent, excluding unknown */
		if ((epm_has_perf(parentp, PM_ASSUME_LOW)) &&
		    (STATE_TO_PERF(0, parentp->pn_pwrs) == 0))
			lowest_kids++;
		epm_debug_print_op((char *)__func__, 1, 0, &parentp);

		if (lowest_kids == total_kids) {
			DBG(zprintf("%s: parent(%s) level NeedOne UP\n",
			    __func__, parentp->pn_pmname));
			*needidx = protect_id;
			return (ret |= PWR_OP_ENEEDONE);
		}

		return (ret);
	}	/* end of checking on going to zero-perf-state */

	/*
	 * Transitioning to high power state, must ensure that the
	 * StrictParental parent is up first.
	 */
	parentp = npp[0]->pn_parent;
	if (IS_PM_CAPABLE(parentp) && PM_DEP_IS_STRICTPARENTAL(parentp) &&
	    !epm_has_perf(parentp, PM_ASSUME_LOW)) {
		pa_pwrs = parentp->pn_pwrs;
		pa_method = pa_pwrs->pps_pwr_ctl.pp_method;
		switch (pa_method) {
		case PWRCTL_METHOD_HVCTL:
			break;
		case PWRCTL_METHOD_AUTO:
			/* done automatically */
			return (ret);
		default:
			return (PWR_OP_EINVAL);
		}

		pa_new_lvl = pa_pwrs->pps_count - 1;
		ASSERT(STATE_TO_PERF(pa_new_lvl, pa_pwrs) > 0);

		ret |= epm_set_pwr(res_type, 1, PCO_NONE, pa_new_lvl,
		    pa_method,  &parentp, &pa_res_status, flags);
		if (ret != PWR_OP_OK)
			ret |= PWR_OP_EPARENTUPFAILED;

		return (ret);
	}	/* end of transtiontion high power checking */

	return (ret);
}

/*
 * epm_post_pwr - process dependency and coordination after setting
 * resources to the new power state.
 */
/* ARGSUSED1 */
int
epm_post_pwr(pm_res_type_t res_type, int count, uint16_t new_state,
	pwrctl_method_t method, pm_node_t **nodepp, uint32_t flags)
{
	pm_node_t	*cn;
	pm_node_t	*parentp = nodepp[0]->pn_parent;
	pm_pwr_scheme_t	*pa_pwrs;
	pwrctl_method_t	pa_method;
	bool_t		kids_at_lowest;
	int		kid_not_at_lowest_has_keepmeup = 0;
	int		kids_not_at_lowest_cnt = 0;
	int		pa_status = 0;
	int		ret = PWR_OP_OK;
	uint16_t	pa_lvl;

	/*
	 * Done, if (parent isn't power managable OR if not under
	 * StrictParental control) AND (parent is not a Higer
	 * Power Savings node).
	 */
	if ((!IS_PM_CAPABLE(parentp) || !PM_DEP_IS_STRICTPARENTAL(parentp)) &&
	    (!PM_COORD_IS_HPS(parentp))) {

		DBG_BLOCK {
			if (PM_COORD_IS_HPS(parentp)) {
				DBG6(zprintf("%s: Parent %s has HIGHERPOWERSAV"
				    "INGS, but returning, which is wrong\n",
				    __func__, parentp->pn_pmname));
			}
		}

		return (ret);
	}

	/*
	 * If the new state for the current node (and possibly some or
	 * all of its siblings) is not the lowest power state, then
	 * there is no action to perform on the parent.
	 */
	if (STATE_TO_PERF(new_state, epm_get_pwr_scheme(nodepp[0], method)) > 0)
		return (ret);

	/* Attempt to power down parent if all kids are at lowest */
	kids_at_lowest = true;
	for (cn = parentp->pn_child; cn; cn = cn->pn_sibling) {
		if (epm_has_perf(cn, PM_ASSUME_HIGH)) {

			if (cn->pn_flag & PN_FLAG_KEEPMEUP) {
				kid_not_at_lowest_has_keepmeup = 1;
			}
			kids_at_lowest = false;
			kids_not_at_lowest_cnt++;
		}
	}

	DBG(zprintf("%s: parent %s %s, kids_not_at_lowest: %d, "
	    " kid_not_at_lowest_has_keepmeup: %d\n",
	    __func__, parentp->pn_pmname,
	    ((kids_at_lowest) ? "KidsAtLowest" : "KidsNotAtLowest"),
	    kids_not_at_lowest_cnt, kid_not_at_lowest_has_keepmeup));

	/*
	 * If all the kids are at their lowest except for one, and that one has
	 * the PN_FLAG_KEEPMEUP bit set, and the parent node is an HPS node,
	 * then see if the HPS node can be powered down to its lowest pwr state.
	 */
	if ((kids_not_at_lowest_cnt == 1) &&
	    (kid_not_at_lowest_has_keepmeup == 1) &&
	    (PM_COORD_IS_HPS(parentp))) {
		pmi_hps_adjust_pwr(HPS_SOURCE_NO_PWR, NULL,
		    HPS_LOW_POWER, HPS_CALC_NUM_NODES);
		/*
		 * There is nothing that the existing post processing can
		 * do for a clock state set failure.  Msgs have been logged
		 * indicating the failure.
		 */
		return (PWR_OP_OK);
	}

	if (kids_at_lowest == false) {
		return (ret);
	}

	pa_pwrs = parentp->pn_pwrs;
	if (pa_pwrs->pps_count <= 1) {
		return (ret);
	}

	if ((pa_method = pa_pwrs->pps_pwr_ctl.pp_method) == PWRCTL_METHOD_AUTO)
		return (ret);

	DBG6(zprintf("%s: powering down %s\n",
	    __func__, parentp->pn_pmname));

	/* power down parent */
	pa_lvl = 0;	/* lowest power index */
	ASSERT(pa_method == PWRCTL_METHOD_HVCTL);
	ret |= epm_set_pwr(res_type, 1, PCO_NONE, pa_lvl, pa_method, &parentp,
	    &pa_status, flags);
	DBG(zprintf("%s: powered down parent %s of %s returns cc=%d\n",
	    __func__, parentp->pn_pmname, nodepp[0]->pn_pmname, ret));

	return (ret);
}

/*
 * epm_build_parent_list - build up unique list of parent pm nodes given
 * a list of kid pm nodes.
 */
int
epm_build_parent_list(int count, pm_node_t **nodepp, pm_node_t ***pnlist)
{
	pm_node_t	**parentnp;
	pnc_t		*headp, *tmpp, *savp, *pnew;
	int		i, parent_count;

	/*
	 * chain up unique parents
	 * add the 1st parent to the chain
	 */
	headp = Xcalloc(1, pnc_t);
	headp->pnc_ndp = nodepp[0]->pn_parent;
	parent_count = 1;

	/* for each kids node, try to match kid's parent */
	for (i = 1; i < count; i++) {
		tmpp = headp;
		while (tmpp) {
			savp = tmpp;
			if (nodepp[i]->pn_parent == tmpp->pnc_ndp)
				break;
			tmpp = tmpp->pnc_next;
		}

		/* if no matche, add kid's parent in chain */
		if (tmpp == NULL) {
			pnew = Xcalloc(1, pnc_t);
			pnew->pnc_ndp = nodepp[i]->pn_parent;
			savp->pnc_next = pnew;
			parent_count++;
			DBG5(zprintf("%s: Add node_id %d 's parent (id %d)\n",
			    __func__, nodepp[i]->pn_nodeid,
			    pnew->pnc_ndp->pn_nodeid));
		}
	}

	/* put parent nodes in an array */
	parentnp = Xmalloc(parent_count * sizeof (pm_node_t *));
	for (i = 0, tmpp = headp; tmpp; i++, tmpp = tmpp->pnc_next)
		parentnp[i] = tmpp->pnc_ndp;

	/* free the chain */
	do {
		tmpp = headp->pnc_next;
		Xfree(headp);
		headp = tmpp;
	} while (tmpp);

	*pnlist = parentnp;
	return (parent_count);
}

/*
 * epm_ps2pwr_status - convert LDoms/PM shared status code to power module
 * internal status code.
 */
int
epm_ps2pwr_status(pwr_req_status_t cc)
{
	switch (cc) {
	case PS_SUCCESS:
		return (PWR_OP_OK);
	case PS_EINVAL:
		return (PWR_OP_EINVAL);
	case PS_EILLEGAL:
		return (PWR_OP_EUNSUPPORTED);
	case PS_TIMEDOUT:
		return (PWR_OP_ETIMEDOUT);
	case PS_FAILURE:
	default:
		return (PWR_OP_EFAILURE);
	}
}

void
epm_debug_print_node(pm_node_t *node)
{
	pm_pwr_scheme_t	*pwrs;
	int		dep, coord, pwrcount;
	int 		i;

	pwrs = node->pn_pwrs;
	dep = (pwrs) ? pwrs->pps_deps : -1;
	coord = (pwrs) ? pwrs->pps_coord : -1;
	pwrcount = (pwrs) ? pwrs->pps_count : -1;

	if ((dbg_line++ % 10) == 0) {
		DBG6(zprintf("\nRank :    Type     :  ID : PID : SID : CID : "
		    "       name       : _Lo : _Hi :  KC : Dep : Coord :"
		    " state/desc/perf/ratio       : "
		    " pm_mapping?\n"));
	}

	DBG6(zprintf("%3d  : %11s : %3d : %3d : %3d : %3d :",
	    node->pn_rank,
	    node->pn_type,
	    node->pn_nodeid,
	    (node->pn_parent ? node->pn_parent->pn_nodeid : -1),
	    (node->pn_sibling ? node->pn_sibling->pn_nodeid : -1),
	    (node->pn_child ? node->pn_child->pn_nodeid : -1)));

	DBG6(zprintf("%18s : %3d : %3d : %3d : %3d : %5d : ",
	    (node->pn_pmname) ? node->pn_pmname : "NULL",
	    node->pn_leaflo,
	    node->pn_leafhi,
	    node->kids_count,
	    dep,
	    coord));

	for (i = 0; i < pwrcount; i++)
		DBG6(zprintf(" %d/%s/%d/%4.3f, ",
		    i,
		    pwrs->pps_pmcap[i].ppc_lname,
		    pwrs->pps_pmcap[i].ppc_perf,
		    pwrs->pps_ratio[i]));
	if (pwrcount <= 0)
		DBG6(zprintf("  ---------------------"));

	DBG6(zprintf("%s\n",
	    ((node->pn_rcfg) ? " Yes " : " No ")));
}

/*
 * print PM device tree nodes
 */
void
epm_debug_print_tree(pm_node_t *tree)
{
	pm_node_t	*cn;

	if (tree)
		epm_debug_print_node(tree);

	for (cn = tree->pn_child; cn; cn = cn->pn_sibling)
		epm_debug_print_tree(cn);
}

void
epm_debug_print_op(char *fnm, int count, uint16_t new_state, pm_node_t **npp)
{
	int i;

	DBG5(zprintf("%s: new_state = %d, pmname(pps_count, pps_lvl):\n",
	    fnm, new_state));
	for (i = 0; i < count; i++) {
		DBG5(zprintf("%s(%d, %d), ", npp[i]->pn_pmname,
		    npp[i]->pn_pwrs->pps_count, npp[i]->pn_pwrs->pps_curlvl));
	}
	DBG5(zprintf("\n"));
}

/*
 * epm_get_pwr - get resources' power state from hardware and update the
 * device tree with current power state and status.
 *
 * Parameters:
 *    res_type - must be PM_RES_CPU
 *    count - number of resources
 *    nodepp -  array resource nodes
 *    states - return in an array of resource states allocated by caller.
 */

/*ARGSUSED*/
int
epm_get_pwr(pm_res_type_t res_type, int count, pm_node_t **nodepp)
{
	pm_pwr_scheme_t		*pwrs;
	pwr_req_status_t	cc;
	pwr_req_status_t	*status = NULL;
	char			**ckp = NULL;
	int 			i;
	int			ret = PWR_OP_OK;
	uint16_t		*states = NULL;

	ckp = Xmalloc(count * sizeof (char *));
	states = Xcalloc(count, uint16_t);
	status = Xcalloc(count, pwr_req_status_t);

	for (i = 0;  i < count; i++) {
		pwrs = epm_get_pwr_scheme(nodepp[i], PWRCTL_METHOD_HVCTL);
		if (!pwrs)
			goto get_pwr_done;
		ckp[i] = pwrs->pps_pwr_ctl.pp_cookie;
	}

	cc = GET_POWER_STATES(count, ckp, states, status);
	ret = epm_ps2pwr_status(cc);

	epm_update_pwr(count, nodepp, states, PWRCTL_METHOD_HVCTL, status);

get_pwr_done:
	Xfree(ckp);
	Xfree(states);
	Xfree(status);
	return (ret);
}

/*
 * pmi_clear_all_fma_status - Recursively traverses the device tree to clear
 * the fma status (i.e. set to FMA_STATUS_UNKNOWN) of all resource nodes.
 */
void
pmi_clear_all_fma_status(pm_node_t *top_node)
{
	pm_node_t	*cp;

	if (top_node == NULL) {
		if (pm_root_node != NULL)
			top_node = pm_root_node;
		else
			return;
	}

	if (top_node->pn_rcfg != NULL)
		top_node->pn_rcfg->pi_fma_status = FMA_STATUS_UNKNOWN;

	for (cp = top_node->pn_child; cp != NULL; cp = cp->pn_sibling) {
		pmi_clear_all_fma_status(cp);
	}
}

/*
 * pmi_find_res_nodeps - Recursively traverses the device tree to find all
 * resources of a particular type.  The list is a pointer to a list of
 * pointers to all the nodes.  This is different from pmi_find_res_nodes
 * which does a struct copy instead of using a pointer to the object.
 */
void
pmi_find_res_nodeps(pm_node_t *top_node, pm_res_type_t res_type,
    pm_node_t ***res_nodes, int *n)
{
	pm_node_t	*child;

	if (top_node == NULL) {
		if (pm_root_node != NULL)
			top_node = pm_root_node;
		else
			return;
	}

	if ((top_node->pn_rcfg != NULL) &&
	    (top_node->pn_rcfg->pi_res_type == res_type)) {
		*res_nodes = (pm_node_t **)Xrealloc(*res_nodes,
		    (*n+1) * sizeof (pm_node_t **));
		ASSERT(*res_nodes != NULL);
		*((*res_nodes) + *n) = top_node;
		(*n)++;
	}

	for (child = top_node->pn_child; child != NULL;
	    child = child->pn_sibling) {
		pmi_find_res_nodeps(child, res_type, res_nodes, n);
	}
}

/*
 * pmi_find_res_nodes - Recursively traverses the pm_node tree to find all
 * resources of a particular type.
 */
void
pmi_find_res_nodes(pm_node_t *top_node, pm_res_type_t res_type,
    pm_node_t **res_nodes, int *n)
{
	pm_node_t	*child;

	if (top_node == NULL) {
		if (pm_root_node != NULL)
			top_node = pm_root_node;
		else
			return;
	}

	if ((top_node->pn_rcfg != NULL) &&
	    (top_node->pn_rcfg->pi_res_type == res_type) &&
	    !(pmi_is_resource_faulted(top_node->pn_rcfg))) {
		/*
		 * Add node to res_nodes array.  If the res type is
		 * PM_RES_MEM, we only grab the memory grouping nodes since
		 * they are power managable and memory region nodes aren't.
		 */
		if ((res_type == PM_RES_CPU) || ((res_type == PM_RES_MEM) &&
		    (strcmp(top_node->pn_rcfg->pi_res_name, MEM_GROUPING)
		    == 0))) {
			*res_nodes = (pm_node_t *)Xrealloc(*res_nodes,
			    (*n+1) * sizeof (pm_node_t));
			ASSERT(*res_nodes != NULL);
			memcpy((*res_nodes) + *n, top_node, sizeof (pm_node_t));
			(*n)++;
		}
	}

	for (child = top_node->pn_child; child != NULL;
	    child = child->pn_sibling) {
		pmi_find_res_nodes(child, res_type, res_nodes, n);
	}
}

/*
 * pmi_get_container_id - return the container ID in the "container_id"
 * parameter for the resource in the "res_id" parameter.
 * (e.g. The CPU chip ID for a cpu)  A container is the next level PM'able
 * component.
 *
 * This is intended to be a short-lived interface.  It is expected that the
 * Domain Manager  will have a replacement implementation for ldm_defragment()
 * and this interface will be removed.
 *
 * Returns PS_SUCCESS, PS_FAILURE
 */
int
pmi_get_container_id(pm_resid_t *res_id, int *container_id)
{
	pm_node_t	*node, *pa;

	node = epm_find_node(res_id, NULL, NULL);
	if (!node) {
		DBG(zprintf("%s: can't find pm node for %s\n",
		    __func__, res_id->pi_res_name));
		return (-1);
	}

	pa = node;
	while (pa && (pa->pn_parent != pm_root_node))
		pa = pa->pn_parent;
	ASSERT(pa);
	DBG(zprintf("%s: Found container %s for %s\n",
	    __func__, pa->pn_pmname, res_id->pi_res_name));
	*container_id = pa->pn_nodeid;

	return (0);
}

/*
 * pmi_is_mem_version_ok - are the FW components in-sync to support
 * memory pm?
 */
bool_t
pmi_is_mem_version_ok(void)
{
	/*
	 * Both pm-version and hvctl version for memory utilization
	 * should be satisfied.
	 */
	if (hv_is_mem_util_supported() &&
	    ((pm_version_major == PM_MEM_VERSION_MAJOR) &&
	    (pm_version_minor >= PM_MEM_VERSION_MINOR)) ||
	    (pm_version_major > PM_MEM_VERSION_MAJOR))
		return (true);
	else
		return (false);
}

/*
 * pmi_is_mem_supported - is memory pm supported?
 */
bool_t
pmi_is_mem_supported(void)
{
	if (pmi_is_mem_version_ok())
		return (pm_mem_supported);
	else
		return (false);
}

/*
 * pmi_is_hps_supported
 *
 * Function that determines if HPS is supported and if so, if
 * it can be used.
 *
 */
bool_t
pmi_is_hps_supported(void)
{
	/* System does not support HPS */
	if ((!hps_enabled) || (num_hps_nodes <= 0)) {
		DBG6(zprintf("%s: HPS not supported\n", __func__));
		return (false);
	}

	return (true);
}

/*
 * pmi_hps_create_strand_mapping
 *
 * This recursive function creates a list of pm_hps_strand_t which contains the
 * node ID and Hi and Lo strand ID for each node in the system that
 * supports HPS.  This list will be used later in the algorithm
 * to efficiently determine which node a domain strand is on.
 * This function is called from epm_scan_pm_resources() only once.
 *
 * Input:
 *   nodep - Pointer to a PM node
 *   level - Level of tree we are on. 0 is the top node level.
 *   *err  - Current error value: 0 -> OK, -1 -> Failure
 *
 * Output:
 *   hps_strand_map - Global pointer is set to a list of pm_hps_strand_t
 *   *err  0 -> Success in building Map
 *        -1 -> Failure in building Map
 *
 */
void
pmi_hps_create_strand_mapping(pm_node_t *nodep, int level, int *err)
{
	pm_node_t		*child;
	pm_pwr_scheme_t		*pwrs;
	static pm_hps_strand_map_t *new_ptr;
	static pm_hps_strand_map_t *hsmp;

	/*
	 * One time allocate of the memory needed for the list.
	 * The number of HPS nodes was found when the PRI was scanned.
	 */
	if (hps_strand_map == NULL) {
		hps_strand_map = Xcalloc(num_hps_nodes, pm_hps_strand_map_t);
		if (hps_strand_map == NULL) {
			*err = -1;
			return;
		}

		/*
		 * hps_strand_map is a global that always points to the list
		 * head. hsmp is a static ptr to the node just filled in with
		 * data.  new_ptr is a ptr to the next node to fill in.
		 * We need the two walking ptrs in order to fill in the 'next'
		 * link of a previous node (hsmp) when a new node
		 * (new_ptr) is being filled in.
		 */
		hsmp = hps_strand_map;
		new_ptr = hsmp;
	}

	/* Recursive search for PM nodes that indicate HPS supported. */
	level++;
	for (child = nodep->pn_child; child; child = child->pn_sibling) {
		pwrs = child->pn_pwrs;

		while (pwrs != NULL) {
			if (PM_COORD_IS_HPS(child)) {

				/* Fill in data for current node */
				new_ptr->pmnode = child;
				new_ptr->level = level;
				new_ptr->chip_id = child->pn_nodeid;
				new_ptr->strand_lo = child->pn_leaflo;
				new_ptr->strand_hi = child->pn_leafhi;

				/*
				 * If this is not the first Map node, then link
				 * the previous node to this new one, then
				 * advance the new Ptr to the next node to be
				 * filled in.
				 */
				if (new_ptr != hps_strand_map) {
					hsmp->next = new_ptr;
					hsmp = new_ptr++;

				/*
				 * This is the first Map node, so there is no
				 * linking of nodes to do yet.  This else case
				 * will be hit once.
				 */
				} else {
					hsmp->next = NULL;
					new_ptr++;
				}
			}
			pwrs = pwrs->pps_next;
		}

		/* Keep looking for HPS nodes */
		pmi_hps_create_strand_mapping(child, level, err);
	}
}

/*
 * pmi_create_hps_node_vector
 *
 * Creates vectors denoting which HPS nodes the domain has strands on. This
 * function creates linked lists of PM nodes that are HPS nodes for strands in
 * the Domain.  The lists are multi-leveled to accomodate nodes at different
 * levels in the PM tree that are HPS.  For instance, a HW processor chip can
 * be HPS, and a core within a chip can also be HPS.
 *
 * View the vectors being created as an array of pointers to linked lists of PM
 * nodes.  Each array element is a different level in the PM tree with level 0
 * being the level highest in the tree.  The entries on each linked list, are
 * those PM nodes that are HPS components that the Domain is dependent on.
 *
 * Input:
 *   domp - Pointer to an active domain
 *
 * Output:
 *   domp->dom_utilsp - Pointer to the domain Utils
 *   domp->node_vec[] - HPS vector lists
 *   Function return:
 *     PS_SUCCESS: Vectors created
 *     PS_FAILURE: Failed to create vectors
 *
 */
int
pmi_create_hps_node_vector(pm_hps_domain_t *domp)
{
	ldm_err_t		err = {0, NULL};
	ldm_resutils_t  	*dom_utilsp;
	pm_hps_strand_map_t	*hsmp;
	pm_node_vec_t		**node_vecpp;
	int 			util_idx;
	uint32_t		res_id;

	/* Get the CPU Utils for this domain */
	if (GET_DOMAIN_UTILS(
	    domp->domain, PM_RES_CPU, &dom_utilsp, &err) !=
	    PS_SUCCESS) {
		DBG(zprintf("%s: pmi_create_hps_node_vector() Can't "
		    "GET_DOMAIN_UTILS for domain %s in state %d\n",
		    __func__, GET_DOMAIN_NAME(domp->domain),
		    GET_DOMAIN_STATE(domp->domain)));
		return (PS_FAILURE);
	}

	domp->dom_utilsp = dom_utilsp;

	/*
	 * Find out how many levels of HPS nodes there are in the PM tree, in
	 * order to create the initial memory to hold the pointers to the
	 * pm_node_vec_t structs.  These structs are used to tell what HPS nodes
	 * the Domain is dependent on.
	 */

	/* hps_num_lvls is 0 based, so add 1 */
	domp->num_lvls = hps_num_lvls + 1;

	node_vecpp = Xcalloc(domp->num_lvls, pm_node_vec_t *);

	/*
	 * Walk thru the Utils and find out which node each strand is on,
	 * then update the HPS vector.
	 */

	/* For each CPU resource in the Domain... */
	for (util_idx = 0; util_idx < dom_utilsp->ldmr_rescnt; util_idx++) {

		/* Get the strand ID */
		res_id =
		    dom_utilsp->ldmr_util_p[util_idx].\
		    ldmu_resid.u.cpu.pi_id_val;

		/*
		 * Compare the strand ID to the strand ranges for each HPS node
		 * and set a node vector when a range is found.  The vector
		 * indicates which node the strand is on.
		 */
		for (hsmp = hps_strand_map; hsmp; hsmp = hsmp->next) {
			int		lvl;
			pm_node_vec_t	*node_vecp, *nxt_vecp;

			if ((res_id < hsmp->strand_lo) ||
			    (res_id > hsmp->strand_hi))
				continue;

			/*
			 * Add an entry to the node_vec array for the
			 * correct HPS level.  This array holds ptrs to
			 * the PM nodes that this Domain is depenednt
			 * on wrt HPS adjustments.
			 */
			lvl = hsmp->level;
			if (node_vecpp[lvl] == NULL) {
				/* 1st entty on list for this level */
				nxt_vecp =
				    Xmalloc(sizeof (pm_node_vec_t));
				node_vecpp[lvl] = nxt_vecp;
				nxt_vecp->pmnode = hsmp->pmnode;
				nxt_vecp->chip_id = hsmp->chip_id;
				nxt_vecp->new_ratio = 1.0;
				nxt_vecp->next = NULL;
				node_vecp = nxt_vecp;
				continue;
			}

			/*
			 * There are entires for the level
			 * already, so start at the head of the
			 * vector list and walk towards the end
			 * checking to see if we already have
			 * the PM HPS node.  If not, then add
			 * it to the end.
			 */
			for (nxt_vecp = node_vecpp[lvl]; nxt_vecp;
			    nxt_vecp = nxt_vecp->next) {
				/* On list already */
				if (hsmp->pmnode == nxt_vecp->pmnode)
					break;
				node_vecp = nxt_vecp;
			}

			/* Add new entry to vector */
			if (nxt_vecp == NULL) {
				nxt_vecp = Xmalloc(sizeof (pm_node_vec_t));
				node_vecp->next = nxt_vecp;
				nxt_vecp->pmnode = hsmp->pmnode;
				nxt_vecp->chip_id = hsmp->chip_id;
				nxt_vecp->new_ratio = 1.0;
				nxt_vecp->next = NULL;
			}
		}
	}

	/* This is the array of pointers to Node vectors */
	domp->node_vec = node_vecpp;

	return (PS_SUCCESS);
}

/*
 * pmi_hps_node_match
 *
 * This function compares two hps_domain_t structures to see if their
 * respective Domains have any HPS nodes in common. Either all node
 * levels are checked for a match or just a particular level is checked.
 *
 * Input:  dom1 - Pointer to a HDL entry
 *         dom2 - Pointer to an Active List entry
 *         lvl  - Node level to check.
 *                -1 -> Check all levels
 * Output:
 *   Function Return:
 *     PS_SUCESSS - Match occurred
 *     PS_FAILURE - No Match
 *
 */
pwr_req_status_t
pmi_hps_node_match(pm_hps_domain_t *dom1, pm_hps_domain_t *dom2, int lvl)
{
	int		i, i_strt, i_end;
	int		j, j_strt, j_end;
	pm_node_vec_t	*vec1p, *vec2p;

	if (lvl == -1) {
		i_strt = 0;
		i_end  = dom1->num_lvls;
		j_strt = 0;
		j_end  = dom2->num_lvls;
	} else {
		i_strt = lvl;
		i_end  = lvl + 1;
		j_strt = lvl;
		j_end  = lvl + 1;
	}

	for (i = i_strt; i < i_end; i++) {

		for (vec1p = dom1->node_vec[i]; vec1p;
		    vec1p = vec1p->next) {

			for (j = j_strt; j < j_end; j++) {

				for (vec2p = dom2->node_vec[j]; vec2p;
				    vec2p = vec2p->next) {

					if (vec1p->pmnode ==
					    vec2p->pmnode) {
						return (PS_SUCCESS);
					}
				}
			}
		}
	}
	return (PS_FAILURE);
}

/*
 * pmi_hps_set_node_dependencies
 *
 * This function compares two Domains HPS vectors for any matches.
 * For a match, the different HPS vector entries from the 2nd
 * Domain are created in the 1st Domain.
 *
 * Input:  dom1 - Pointer to a HDL entry to compare
 *         dom2 - Pointer to a HDL entry to compare
 * Output:
 *   HDL->node_vec[i] - Node vector list possibly updated
 *
 */
void
pmi_hps_set_node_dependencies(pm_hps_domain_t *dom1, pm_hps_domain_t *dom2)
{

	bool_t		done;
	int		lvl;
	pm_node_vec_t	*vec1p, *vec2p;

	/*
	 * Look at the HPS vector list for each level in dom1 to see if
	 * any entry from the HPS vector list in dom2 at the same level
	 * matches one of those in dom1.  If so, add the entries from dom2
	 * to the dom1 list.  This just means that if a dom1 HPS vector list
	 * entry is power adjusted, then all the dom2 entires would have to
	 * be adjusted also.
	 */
	for (lvl = 0; lvl < dom1->num_lvls; lvl++) {

		/* Walk the dom1 list for a certain level */
		done = false;
		for (vec1p = dom1->node_vec[lvl]; vec1p;
		    vec1p = vec1p->next) {

			vec2p = dom2->node_vec[lvl];
			while (vec2p != NULL) {

				/* Same entry on both lists */
				if (vec1p->pmnode == vec2p->pmnode) {

					/*
					 * Add dom2 entires onto end of dom1. Do
					 * not duplicate entries.
					 */
					pmi_hps_copy_vectors(lvl, dom1, dom2);
					done = true;
					break;
				}
				vec2p = vec2p->next;
			}
			if (done)
				break;
		}
	}
}
/*
 * pmi_hps_copy_vectors
 *
 * This function copies HPS vectors (pm_node_vec_t) from one Domain to another,
 * for a particular level. HPS vectors will not be duplicated in a Domain.
 *
 * Input:  dom_to - Pointer to a HDL entry to copy to
 *         dom_from - Pointer to a HDL entry to copy from
 * Output:
 *   dom_to->node_vec[i] - Node vector list possibly updated
 *
 */
void
pmi_hps_copy_vectors(int lvl, pm_hps_domain_t *dom_to,
    pm_hps_domain_t *dom_from)
{
	bool_t		on_list;
	pm_node_vec_t	*end_vec_to, *new_vecp, *vec_to, *vec_from;

	/* Find the end of dom_to list first */
	vec_to = dom_to->node_vec[lvl];
	while (vec_to->next != NULL)
		vec_to = vec_to->next;
	end_vec_to = vec_to;

	/*
	 * Loop thru the dom_from HPS vector and copy(create) a new entry on the
	 * dom_to vector for each dom_from entry that is not in dom_to.
	 */
	for (vec_from = dom_from->node_vec[lvl]; vec_from;
	    vec_from = vec_from->next) {

		on_list = false;
		for (vec_to = dom_to->node_vec[lvl]; vec_to;
		    vec_to = vec_to->next) {
			if (vec_from->pmnode == vec_to->pmnode) {
				on_list = true;
				break;
			}
		}

		/* Add a new vector entry to the end of the dom_to HPS vector */
		if (!on_list) {
			new_vecp = Xmalloc(sizeof (pm_node_vec_t));
			end_vec_to->next = new_vecp;
			new_vecp->next = NULL;
			new_vecp->pmnode = vec_from->pmnode;
			new_vecp->chip_id = vec_from->chip_id;
			new_vecp->new_ratio = 1.0;
			end_vec_to = new_vecp;
		}
	}
}

/*
 * pmi_hps_pre_set_pwr
 *
 * Pre processing to set the power states for HPS nodes.
 * Sets arrays for the cookies and new power states.
 *
 * Input:
 *   cookies: Array of char ptrs for the cookies
 *   states:  Array of uint16 ptrs for the new pwr states
 *   hps_cnt: # of HPS nodes
 *   hpsp:    Ptr to a structure list of HPS Domains
 *   source:  Flag to tell where to HPS nodes came from
 *            HPS_SOURCE_ALL: PM node tree, all HPS nodes
 *            HPS_SOURCE_HDLP: HPS Domain List
 *            HPS_SOURCE_NO_PWR: PM node tree, unused nodes
 *   pstate:  New Power State for HPS nodes
 *            HPS_FULL_POWER: Set level to it highest state
 *            HPS_CALC_POWER: State has already been determined
 *            HPS_LOW_POWER: Set level to it lowest state
 *
 * Output:
 *   Function return: # of nodes to set pwr for
 */
int
pmi_hps_pre_set_pwr(char **cookies, uint16_t *states,
    int hps_cnt, void *hpsp, int source, int pstate)
{
	pm_hps_domain_t	*hdlp = NULL;
	pm_node_t	**pmnpp = NULL;
	pm_node_vec_t	*nvp;
	int 		i, lvl;
	int 		node_idx = 0;

	switch (source) {
	case HPS_SOURCE_HDLP:
		/*
		 * Use hdlp as the source of HPS nodes. HPS_CALC_POWER
		 * uses this case.
		 */
		hdlp = (pm_hps_domain_t *)hpsp;
		for (lvl = 0; lvl < hdlp->num_lvls; lvl++) {
			for (nvp = hdlp->node_vec[lvl]; nvp;
			    nvp = nvp->next, node_idx++) {
				states[node_idx] = nvp->new_state;
				cookies[node_idx] =
				    nvp->pmnode->pn_pwrs->pps_pwr_ctl.pp_cookie;
			}
		}
		break;

	case HPS_SOURCE_NO_PWR:
	case HPS_SOURCE_ALL:
		/* Use pmnpp as the source of HPS nodes */
		pmnpp = (pm_node_t **)hpsp;
		for (node_idx = 0; node_idx < hps_cnt; node_idx++) {
			/*
			 * All HPS nodes in the system do not have to have
			 * the same PM state index for their highest
			 * pwr level, so get that from the PM node.
			 */

			/* Full power */
			if (pstate == HPS_FULL_POWER)
				states[node_idx] =
				    pmnpp[node_idx]->pn_pwrs->pps_count - 1;

			/* Other than full power */
			else if (pstate == HPS_LOW_POWER)
				states[node_idx] = 0;

			/* Set to same as in PM node */
			else if (pstate == HPS_SAME_POWER)
				states[node_idx] =
				    pmnpp[node_idx]->pn_pwrs->pps_curlvl;

			/* Use input state */
			else
				states[node_idx] = (uint16_t)pstate;

			cookies[node_idx] =
			    pmnpp[node_idx]->pn_pwrs->pps_pwr_ctl.pp_cookie;
		}
		break;
	}

	/* Debug the cookies for the HPS nodes */
	DBG_BLOCK {
		for (i = 0; i < node_idx; i++)
			DBG6(zprintf("%s:%d; ", cookies[i], states[i]));
		DBG6(zprintf("\n"));
	}

	return (node_idx);
}

/*
 * pmi_hps_set_pwr
 *
 * Call the LDoms interface to set power states for the input.
 *
 * Input:
 *   cookies: Array of char ptrs for the cookies
 *   states:  Array of uint16 pwr states for the cookies
 *   num_nodes: Number of cookies
 *
 * Output:
 *   status: Array of status values for each cookie
 *   Function return: Return from SET_POWER_STATES()
 */
pwr_req_status_t
pmi_hps_set_pwr(char **cookies, uint16_t *states, pwr_req_status_t *status,
    int num_nodes)
{

	pwr_req_status_t	rc;

	rc = SET_POWER_STATES(num_nodes, &cookies[0], &states[0], &status[0]);

	return (rc);
}

/*
 * pmi_hps_adjust_pwr
 *
 * Pre processing to set the power states for HPS nodes.
 * There can be multiple sources for the HPS nodes, the
 * PM tree and the HPS Domain List (HDL).
 *
 * Input:
 *   source: Flag to tell where to get HPS nodes from
 *   hpsp:   Pointer to a structure of HPS nodes
 *   pstate: New power state for selected HPS nodes
 *   count:  Number of HPS nodes, if known, -1 if not
 *
 * Output:
 *   Function return: Return code from HV mapped to PM values
 */
pwr_req_status_t
pmi_hps_adjust_pwr(int source, void *hpsp, int pstate, int count)
{
	pm_node_t		**head_pmnpp = NULL;
	pm_node_t		**pmnpp = NULL;
	pwr_req_status_t	*status = NULL;
	char			**cookies = NULL;
	int			busy = 0;
	int			cnt_hps_nodes = 0;
	int			num_hps;
	int			rtn = PS_SUCCESS;
	uint16_t		*states = NULL;

	/*
	 * Get the number of HPS nodes to power adjust depending on
	 * the input source of the HPS nodes.
	 */
	switch (source) {
	case HPS_SOURCE_HDLP:
		/* Get only the HPS nodes for the current Event domain */
		cnt_hps_nodes = count;
		DBG6(zprintf("%s: HPS_SOURCE_HDLP: count=%d\n",
		    __func__, cnt_hps_nodes));
		break;

	case HPS_SOURCE_ALL:
		/* Get all the HPS nodes in the system */
		head_pmnpp =
		    (pm_node_t **)Xcalloc(num_hps_nodes, pm_node_t *);
		pmnpp = head_pmnpp;
		pmi_hps_get_pmnodes(pm_root_node, &pmnpp, &cnt_hps_nodes);
		if (pmnpp != head_pmnpp)
			hpsp = (void *)head_pmnpp;
		else {
			/* No HPS nodes found?? Should not happen */
			DBG6(zprintf("%s: No HPS nodes found in PM tree for "
			    "HPS_SOURCE_ALL\n", __func__));
		}
		DBG6(zprintf("%s: HPS_SOURCE_ALL: count=%d\n",
		    __func__, cnt_hps_nodes));
		break;

	case HPS_SOURCE_NO_PWR:

		/*
		 * Need ptrs to all the HPS nodes to power down. Allocate
		 * space for pointers to the max number of HPS nodes in
		 * the system.  Not of all of this space will be used
		 * of course.
		 */
		head_pmnpp = (pm_node_t **)
		    Xcalloc(num_hps_nodes, pm_node_t *);
		/* pmnpp will change in value, so need to keep list head ptr */
		pmnpp = head_pmnpp;

		/*
		 * Find all the HPS nodes in the PM tree that can be powered
		 * down to their lowest state.
		 */
		pmi_hps_get_nodes_to_pwr_down(pm_root_node, &pmnpp, &busy,
		    &cnt_hps_nodes);
		if (pmnpp != head_pmnpp)
			hpsp = (void *)head_pmnpp;

		DBG6(zprintf("%s: HPS_SOURCE_NO_PWR: count=%d\n",
		    __func__, cnt_hps_nodes));
		break;

	default:
		break;
	}

	if (cnt_hps_nodes <= 0) {
		DBG6(zprintf("%s: No HPS nodes found to adjust "
		    "power on for source %d\n", __func__, source));
		Xfree(head_pmnpp);
		return (rtn);
	}

	states = Xcalloc(cnt_hps_nodes, uint16_t);
	status = Xcalloc(cnt_hps_nodes, pwr_req_status_t);
	cookies = Xcalloc(cnt_hps_nodes, char *);
	num_hps = pmi_hps_pre_set_pwr(cookies, states, cnt_hps_nodes,
	    hpsp, source, pstate);
	if (num_hps != 0)
		rtn = pmi_hps_set_pwr(cookies, states, status, num_hps);
	else
		DBG6(zprintf("%s: No HPS nodes found to adjust pwr\n",
		    __func__));

	pmi_hps_post_set_pwr();

	Xfree(states);
	Xfree(cookies);
	Xfree(status);
	Xfree(head_pmnpp);

	return (rtn);
}

/*
 * pmi_hps_post_set_pwr
 *
 * Post processes the setting of power states for HPS nodes.
 * Refreshes the Pwr levl (pps_curlvl) in the PM node for
 * all HPS nodes in the system.
 *
 * Currently, we refresh all system HPS nodes since there
 * are only a few of them.  This routine can be modified
 * to set the node Pwr level for a select set of nodes if
 * so desired.
 */
pwr_req_status_t
pmi_hps_post_set_pwr(void)
{
	return (pmi_hps_get_update_pwr());
}

/*
 * pmi_hps_get_update_pwr
 *
 * Get the current HV power setting for all HPS nodes in the PM tree and
 * then update their PM nodes with this data.
 *
 * Input: None
 * Output:
 *   PS_SUCCESS: If return from epm_get_pwr() was PWR_OP_OK
 *   PS_FAILURE: Otherwise
 */
pwr_req_status_t
pmi_hps_get_update_pwr(void)
{
	pm_node_t	**pmnpp;
	pm_node_t	**head_pmnpp;
	int		cnt_hps_nodes;
	int		rtn = PS_SUCCESS;

	/* Need ptrs to all the HPS nodes in the PM tree */
	head_pmnpp = (pm_node_t **)Xcalloc(num_hps_nodes, pm_node_t *);
	pmnpp = head_pmnpp;

	/* Find all the HPS nodes in the PM tree */
	pmi_hps_get_pmnodes(pm_root_node, &pmnpp, &cnt_hps_nodes);

	if (pmnpp != head_pmnpp) {
		rtn = epm_get_pwr(PM_RES_PSTATE, num_hps_nodes, head_pmnpp);
		if (rtn != PWR_OP_OK)
			rtn = PS_FAILURE;
		else
			rtn = PS_SUCCESS;
	}

	Xfree(head_pmnpp);
	return (rtn);
}

/*
 * pmi_hps_get_pmnodes
 *
 * Recursive function to create an array of pointers to all the HPS
 * nodes (pm_node_t) in the system.
 *
 * Input
 *   pmnp: Pointer to the PM node to start the search at. Usually pm_root_node
 * ***pmnppp: A ptr to an array of ptrs to pm_node_t structs
 * ^^^
 * |||--> A pointer to
 * ||---> An array of pointers to
 * |----> A pm_node_t struct
 *  cnt_hps_nodes: Address of var to return count of HPS nodes
 *
 * Output:
 *   **pmnpp: The array of pm_node_t pointers
 *   *cnt_hps: # of HPS nodes found in the PM node tree
 */
void
pmi_hps_get_pmnodes(pm_node_t *pmnp, pm_node_t ***pmnppp, int *cnt_hps)
{
	pm_node_t  	*nodep;

	/*
	 * Depth first (child) top down traverse of PM tree to
	 * find all HPS nodes and save a pointer to each one.
	 */
	if (IS_PM_CAPABLE(pmnp) && PM_COORD_IS_HPS(pmnp)) {
		/* Same as the array index pmnpp[x] */
		**pmnppp = pmnp;
		/* Advance ptr to next array element */
		(*pmnppp)++;
		(*cnt_hps)++;
	}

	for (nodep = pmnp->pn_child; nodep; nodep = nodep->pn_sibling) {
		pmi_hps_get_pmnodes(nodep, pmnppp, cnt_hps);
	}
}

/*
 * pmi_hps_get_nodes_to_pwr_down
 *
 * Recursive function to walk the PM tree to determine which
 * HPS nodes have all CPU resources below them, powered off.
 * Creates a list of PM nodes that can then be powered down
 * by the caller.
 *
 * Input:
 *   node:   The current PM node to look at
 *   pmnppp: A pointer to an array of pointers to pm_node_t structs
 *
 * Output:
 *   pmnpp: Elements in this array are updated to point to PM nodes in the
 *          PM tree for nodes that can be powered down
 *   num_nodes: A ptr to an int for the total # of nodes to be pwr'd down
 *   busy: A ptr to an int which is a flag to denote if the current node,
 *         strand, core or HPS, is in a non powered off state or not.
 */
void
pmi_hps_get_nodes_to_pwr_down(pm_node_t *node, pm_node_t ***pmnppp,
    int *busy, int *num_nodes)
{
	pm_node_t	*child;
	int 		in_use = 0;

	if ((node->pn_rcfg != NULL) &&
	    (node->pn_rcfg->pi_res_type == PM_RES_CPU)) {

		/*
		 * We don't want to count the lowest strand on a chip
		 * if it is powered up.  Even though it might belong
		 * to a Domain, the chances of it being the only
		 * strand for that Domain on this node are low, so
		 * powering down the chip should not make much of an
		 * impact on the performance of the Domain.
		 */

		*busy = 0;
		if ((node->pn_pwrs->pps_curlvl != PPS_LOWEST_POWER) &&
		    (!(node->pn_flag & PN_FLAG_KEEPMEUP))) {
			in_use = 1;
			*busy = 1;
		}

	} else {
		for (child = node->pn_child; child; child = child->pn_sibling) {
			in_use = 0;

			/* recurse */
			pmi_hps_get_nodes_to_pwr_down(child, pmnppp, &in_use,
			    num_nodes);

			/*
			 * The current node is not in use (powered on), so add
			 * it, if an HPS node, to the node list if its current
			 * pwr state is not at lowest
			 */
			if (in_use == 0) {
				if ((PM_COORD_IS_HPS(child)) &&
				    (child->pn_pwrs->pps_curlvl != 0)) {
					/* Same as array idx pmnpp[x] */
					**pmnppp = child;
					(*num_nodes)++;
					/* Advance ptr to nxt element */
					(*pmnppp)++;
				}
			} else {
				*busy = 1;
			}
		}
	}
}
