/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_PM_CAPS_H_
#define	_PM_CAPS_H_

#ifdef __cplusplus
extern "C" {
#endif

#include "basics.h"
#include "pm_internal.h"
#include "zprintf.h"
#include "mdesc.h"
#include "mdesc_impl.h"

/*
 * Platform power manageable components' data structure and hierarchical
 * interrelationships.
 */

/*
 * the PRI "pm-version" that PM supports
 */
#define	PRI_PM_VERSION_MAJOR	2
#define	PRI_PM_VERSION_MINOR	0

/* Starting version that Memory PM is supported */
#define	PM_MEM_VERSION_MAJOR	2
#define	PM_MEM_VERSION_MINOR	1

typedef enum {
	TYPE_PROP = 0,
	PM_TYPE_PROP,
	PM_NAME_PROP,
	PM_COOKIE_PROP,
	PM_STATES_PROP,
	PM_PERFORMANCE_PROP,
	PM_DEPENDENCY_PROP,
	PM_COORDINATION_PROP,
	PM_MAPPING_PROP,
	PM_PROP_MAX
} pm_prop_name_t;

typedef struct {
	pm_prop_name_t	j;
	char 		*propstr;
} pm_prop_data_t;

typedef uint32_t pm_pwr_dep_t;
typedef uint32_t pm_pwr_coord_t;

/* pm_pwr_dep_t values */
#define	PM_DEP_NONE		0
#define	PM_DEP_STRICTPARENTAL	1
#define	PM_DEP_PERFORMANCE	2

/*
 * pm_pwr_coord_t values. These need to be powers of 2
 * since they are used in OP '|' operations.
 */
#define	PM_COORD_NONE		0
#define	PM_COORD_NEEDONE	1
#define	PM_COORD_DEFRAG		2
#define	PM_COORD_HPS		8

typedef struct {
	int	ppc_lvl;	/* index to "pm_states" */
	char	*ppc_lname; 	/* string name of a "pm_states" token */
	int	ppc_perf;	/* performance in 1000th */
} pm_pwr_cap_t;

typedef enum {
	PWRCTL_METHOD_HVCTL = 1,	/* via HVctl call */
	PWRCTL_METHOD_AUTO		/* Hardware automatic set */
} pwrctl_method_t;

typedef struct {
	pwrctl_method_t	pp_method;
	char	*pp_cookie;
} pm_pwrctl_t;

/* Power levels and other flags for setting HPS PStates */
#define	HPS_SOURCE_NO_PWR	0	/* Get all HPS nodes not being used */
#define	HPS_SOURCE_ALL		1	/* Get all HPS node in PM tree */
#define	HPS_SOURCE_HDLP		2	/* HPS nodes are in HPS Domain List */
#define	HPS_LOW_POWER		0	/* Set HPS nodes to lowest pstate */
#define	HPS_FULL_POWER		1	/* Set HPS nodes to highest pstate */
#define	HPS_CALC_POWER		2	/* PState already calculated */
#define	HPS_SAME_POWER		2	/* Use PM node PState pwr level */
#define	HPS_CALC_NUM_NODES	-1	/* Calculate # of nodes */


typedef struct pm_pwr_scheme {
	int		pps_count;	/* number of power levels */
	int		pps_curlvl;	/* current power level/index */
	int		pps_status;	/* PM_STAT_KNOWN,PM_STAT_UNKNOWN */
	pm_pwr_cap_t	*pps_pmcap;	/* an array of power levels */
	double		*pps_ratio;	/* Pwr lvl ratios */
	pm_pwrctl_t	pps_pwr_ctl;	/* get and set power level ops */
	pm_pwr_dep_t	pps_deps;	/* vertical power dependency */
	pm_pwr_coord_t	pps_coord;	/* horizontal power dependency */
	struct pm_pwr_scheme	*pps_next;	/* next power scheme */
} pm_pwr_scheme_t;

typedef struct pm_hps_strand {
	struct pm_node *pmnode;
	int	level;			/* Level in PM node tree chip is */
	int	chip_id;		/* Chip/Node ID that supports HPS */
	int	strand_hi;		/* Highest strand on chip */
	int	strand_lo;		/* Lowest strand on chip */
	struct pm_hps_strand   *next;	/* Next chip in list */
} pm_hps_strand_map_t;

typedef struct pm_node_vec {
	struct pm_node		*pmnode;	/* PM node of HPS node */
	struct pm_node_vec	*next;		/* Link to next node */
	int			chip_id;	/* Chip/Node ID */
	int			new_state;	/* Calcualted PM state idx */
	double			new_ratio;	/* Pwr ratio set for level */
} pm_node_vec_t;

typedef struct pm_hps_domain {
	srcid_t		*domain;	/* LDom in question */
	ldm_resutils_t	*dom_utilsp;	/* Utils for domain CPUs */
	pm_node_vec_t	**node_vec;	/* Array of node_vec_t ptrs */
	pm_util_dir_t	direction;	/* Util indicator */
	int		abv_hiwm_pcent;	/* # Strands at  hi_wm < x <100 % */
	int		blo_lowm_pcent;	/* # Strands at  0< x <low_wm % */
	int		hundred_pcent;	/* # Strands at 100% */
	int		id;		/* unique ID used for matching */
	int		faulted;	/* # of faulted strands */
	int		new_state;	/* New PM state index */
	int		num_nodes;	/* # of nodes (hps) affected */
	int		num_lvls;	/* # of clock levels */
	int		strands_adjust;	/* # Strands to add/remove */
	int		strands_new;	/* calculated new # of strands */
	int		strands_off;	/* # strands powered off */
	int		strands_on;	/* # strands powered on */
	int		within_wm_pcent; /* # Strands in-band */
	int		zero_pcent;	/* # Strands at 0 % */
	bool_t		event_dom;	/* Dom is from an event */
	bool_t		processed;	/* Dom processed */
	struct pm_hps_domain   *next;	/* Next LDom in question */
} pm_hps_domain_t;

/* pps_status */
#define	PM_STAT_KNOWN	0
#define	PM_STAT_UNKNOWN	1

#define	STATUS_SUCCESS	0
#define	STATUS_FAILURE	1

typedef struct pm_node {
	int	pn_rank;	/* node rank or seniority, 0 is the highest */
	char	*pn_type;	/* PRI type */
	char	*pn_pmname;	/* human readable "pm_name" */
	int	pn_nodeid;		/* PRI node  "id" value */
	pm_resid_t	*pn_rcfg;	/* res reconfiguration info */
	pm_pwr_scheme_t	*pn_pwrs;	/* linked list of pwr cntl methods */
	struct pm_node  *pn_parent;	/* parent node pointer */
	struct pm_node  *pn_sibling;	/* sibling node pointer */
	struct pm_node  *pn_child;	/* child node pointer */
	int	pn_leaflo;		/* lowest leaf node_id */
	int	pn_leafhi;		/* highest leaf node_id */
	int	kids_count;		/* count of child nodes */
	uint32_t	pn_flag;	/* special node flag */
} pm_node_t;

enum {
	EPM_SET_PWR_OK,		/* Individual batch processing ok to continue */
	EPM_SET_PWR_DONE,	/* Individual batch processing complete */
	EPM_SET_PWR_RECOVER	/* Individual batch processing error recovery */
};

typedef struct {
	int head;			/* Array index where batch starts */
	int num;			/* Number of entites in this batch */
	pm_node_t **nodepp;		/* Pointers to nodes of cpu ids */
	int *res_status;		/* Individual resource statuses */
	int status;			/* Status for batch.  See enum above */
	int needidx;			/* Index of res which must stay up */
	int precc;			/* Batch ret code for epm_pre_pwr */
	int setcc;			/* Batch ret code for epm_set_pwr */
	int postcc;			/* Batch ret code for epm_post_pwr */
	pwr_req_status_t *pwr_status;	/* Power status for each resource */
	uint16_t *states;		/* State change for this batch */
	char **ckp;			/* Cookies for this batch */
} batch_array_t;


#define	IS_PM_CAPABLE(nodep)	((nodep->pn_pwrs) ? true : false)
#define	IS_RCFG_CAPABLE(nodep)	((nodep->pn_rcfg) ? true : false)

#define	PM_DEP_TYPE(nodep)	(nodep->pn_pwrs->pps_deps)
#define	PM_DEP_IS_NONE(nodep)	((PM_DEP_TYPE(nodep)) & PM_DEP_NONE)
#define	PM_DEP_IS_STRICTPARENTAL(nodep)	\
	(PM_DEP_TYPE(nodep) & PM_DEP_STRICTPARENTAL)
#define	PM_DEP_IS_PERFORMANCE(nodep)	\
	(PM_DEP_TYPE(nodep) & PM_DEP_PERFORMANCE)

#define	PM_COORD_TYPE(nodep)	(nodep->pn_pwrs->pps_coord)
#define	PM_COORD_IS_NONE(nodep)	((PM_COORD_TYPE(nodep)) & PM_COORD_NONE)
#define	PM_COORD_IS_NEEDONE(nodep)	\
	(PM_COORD_TYPE(nodep) & PM_COORD_NEEDONE)
#define	PM_COORD_IS_DEFRAG(nodep)	\
	(PM_COORD_TYPE(nodep) & PM_COORD_DEFRAG)
#define	PM_COORD_IS_HPS(nodep)	\
	(PM_COORD_TYPE(nodep) & PM_COORD_HPS)

/* possible error codes */
#define	PWR_OP_OK		0	/* op successful */
#define	PWR_OP_ERECONFIGURE	0x1	/* reconfiguration failed */
#define	PWR_OP_EINVAL		0x2	/* invalid argument */
#define	PWR_OP_EUNSUPPORTED	0x4	/* Unsupported operation */
#define	PWR_OP_ETIMEDOUT	0x8	/* op timedout */
#define	PWR_OP_ENEEDONE		0x10	/* cannot power off all */
#define	PWR_OP_EKIDSNOTATLOWEST	0x20	/* not all kids are at lowest */
#define	PWR_OP_EPARENTUPFAILED	0x40	/* failed to power up parents */
#define	PWR_OP_EFAILURE		0x80	/* op failed */

/* Set transitioning domains pids to this value */
#define	PM_IGNORE_PID	-1
#define	PM_BIG_NUM	65535

/*
 * Special pn_flag types
 */
#define	PN_FLAG_KEEPMEUP	0x1	/* Resource must never be PM'd */

#define	STATE_TO_PERF(state, pwrs)	\
	(pwrs->pps_pmcap[state].ppc_perf)

#define	PM_ASSUME_LOW	0	/* assume unknown power state low */
#define	PM_ASSUME_HIGH	1	/* assume unknown power state high */

/* retrieve current power state per power scheme */
#define	GET_CUR_LEVEL(pwrs, inc) \
	(pwrs->pps_status == PM_STAT_KNOWN) ? pwrs->pps_curlvl : \
	((inc == PM_ASSUME_LOW) ? 0 : (pwrs->pps_count - 1))

/* defragmentation buffer index location */
#define	DEFRAG_CPU_FROM		0
#define	DEFRAG_CPU_TO		1
#define	DEFRAG_CPU_SPARE	2
#define	DEFRAG_CPU_TOTAL	(DEFRAG_CPU_SPARE + 1)

/* Functions called by policy engine */
typedef enum {
	PPS_LOWEST_POWER,
	PPS_FULL_POWER,
	PPS_FULL_POWER_FORCE
} pmi_pwr_state_t;

/* utility functions and related data structures */
/* opaque function argument buffer */
typedef struct pack_f {
	int		argc;
	uint64_t	*argv;
} pack_f_t;

typedef struct pnc {
	pm_node_t *pnc_ndp;
	struct pnc *pnc_next;
} pnc_t;


/* shared internal functions */

extern bool_t epm_has_perf(pm_node_t *, int);
extern bool_t pmi_is_pri_pm_capable(void);
extern int epm_scan_pm_resources(md_t *);
extern int pmi_get_container_id(pm_resid_t *res_id, int *container_id);
extern int pmi_get_power_all(pm_res_type_t res_type, pm_node_t *topnode);
extern int pmi_get_power_all(pm_res_type_t res_type, pm_node_t *topnode);
extern pm_node_t *epm_find_node(pm_resid_t *res, pm_node_t *node,
    pm_node_t *found_node);
extern pm_node_t *pmi_find_node_by_type(pm_node_t *node, char *node_type,
    int child_id, int parent_id);
extern pm_pwr_scheme_t *epm_get_pwr_scheme(pm_node_t *, pwrctl_method_t);
extern void pmi_clear_all_fma_status(pm_node_t *);
extern void pmi_find_res_nodes(pm_node_t *, pm_res_type_t, pm_node_t **, int *);
extern void pmi_find_res_nodeps(pm_node_t *, pm_res_type_t,
    pm_node_t ***, int *);
extern void pmi_update_fma_status(pm_resid_t *ids, int count);

/* Function prototypes moved from pm_caps.c */

int epm_build_parent_list(int count, pm_node_t **nodepp, pm_node_t ***parentnp);
int epm_get_pwr(pm_res_type_t res_type, int count, pm_node_t **nodepp);
int epm_get_toknum(char *s, char *ss);
int epm_pwr_recover(int setcc, int count, char **ckp, pwr_req_status_t *status,
    uint16_t new_state, pwrctl_method_t method, pm_node_t **nodepp,
    int *res_status);
int epm_set_pwr(pm_res_type_t res_type, int count, pm_config_op_t rcfg,
    uint16_t new_state,  pwrctl_method_t method, pm_node_t **nodepp,
    int *res_status, uint32_t flags);
int pmi_adjust_power(pm_res_type_t res_type, pm_config_op_t req_rcfg_op,
    pmi_pwr_state_t pmi_pwr, int count, pm_resid_t *ids, int **res_status,
    uint32_t);
int pmi_create_hps_node_vector(pm_hps_domain_t *);
int pmi_hps_pre_set_pwr(char **, uint16_t *, int, void *, int, int);

md_t *epm_prune_md(md_t *pri);

pm_node_t *epm_create_root(void);
pm_node_t *epm_create_node(md_t *md,  md_node_t *nodep);
/* pwr_req_status_t pmi_hps_adjust_pwr(pm_hps_domain_t *, uint16_t); TODO */
pwr_req_status_t pmi_hps_adjust_pwr(int, void *, int, int);
pwr_req_status_t pmi_hps_get_update_pwr(void);
pwr_req_status_t pmi_hps_node_match(pm_hps_domain_t *, pm_hps_domain_t *,
    int);
pwr_req_status_t pmi_hps_post_set_pwr(void);
pwr_req_status_t pmi_hps_set_pwr(char **, uint16_t *, pwr_req_status_t *,
    int);

void epm_debug_dump_props(int id, char **propp);
void epm_debug_print_node(pm_node_t *node);
void epm_debug_print_tree(pm_node_t *tree);
void epm_enqueue_node(pm_node_t *last_enq, pm_node_t *newnode,
    char *parenttype);
void epm_update_pwr(int res_count, pm_node_t **nodepp, uint16_t *new_states,
    pwrctl_method_t method, pwr_req_status_t *status);

void pmi_get_chip_ids(pm_resid_t *, int *);
void pmi_hps_copy_vectors(int, pm_hps_domain_t *, pm_hps_domain_t *);
void pmi_hps_create_strand_mapping(pm_node_t *, int, int *);
void pmi_hps_get_nodes_to_pwr_down(pm_node_t *, pm_node_t ***,
    int *, int *);
void pmi_hps_get_pmnodes(pm_node_t *, pm_node_t ***, int *);
void pmi_hps_power_down_nodes(void);
void pmi_hps_set_node_dependencies(pm_hps_domain_t *, pm_hps_domain_t *);

#ifdef __cplusplus
}
#endif

#endif /* _PM_CAPS_H_ */
