/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <thread.h>
#include <synch.h>
#include <sys/ddi.h>
#include "allocate.h"
#include "pm_internal.h"
#include "pm_caps.h"
#include "mdesc_impl.h"
#include "pm_fma.h"
#include "pm_shared.h"
#include "option.h"

#define	ALREADY_CHOSEN		-1
#define	BOOT_DELAY		20
#define	EXTENDED_SLEEP_TIME	15
#define	HUNDRED_PERCENT		100.0
#define	MATCH_ALL_LEVELS	-1
#define	MAX_ATTEMPTS		20
#define	MAX_PERCENTAGE		100
#define	MEM_RES_SEQ_LIMIT	1
#define	PWR_OFF			0
#define	SLEEP_TIME		5
#define	UTIL_CALC_ADJUST	1
#define	PM_UTIL_IS_VALID(dom_utilsp, low_wm, hi_wm) \
	((((dom_utilsp)->ldmr_avg_util < low_wm) && \
	((dom_utilsp)->ldmr_util_dir == PM_UTIL_LOW)) || \
	(((dom_utilsp)->ldmr_avg_util > hi_wm) && \
	((dom_utilsp)->ldmr_util_dir == PM_UTIL_HIGH))) ? true : false


/*
 * ldm_register_util_change_event() flags
 */
#define	DEFAULT_CPU_LOW_UTIL_WM		50
#define	DEFAULT_CPU_HI_UTIL_WM		70
#define	DEFAULT_CPU_TIME_LOW_UTIL_MS	2000
#define	DEFAULT_CPU_TIME_HI_UTIL_MS	2000
#define	DEFAULT_MEM_LOW_UTIL_WM		50
#define	DEFAULT_MEM_HI_UTIL_WM		70
#define	DEFAULT_MEM_TIME_LOW_UTIL_MS	2000
#define	DEFAULT_MEM_TIME_HI_UTIL_MS	2000
#define	MEDIAN_CPU (DEFAULT_CPU_LOW_UTIL_WM +\
	((DEFAULT_CPU_HI_UTIL_WM - DEFAULT_CPU_LOW_UTIL_WM)\
	/ 2))

/* HPS defines */
#define	PM_ACTIVE_BOUND	1	/* Collect active & bound domains */
#define	PM_ACTIVE_ONLY	0	/* Collect only active domains */

/* Event registration handles. */
static void *pm_dom_cfg_hdl;
static void *pm_fma_hdl[PM_RES_TYPES_TOTAL];
static void *pm_res_cfg_hdl[PM_RES_TYPES_TOTAL];
static void *pm_util_hdl[PM_RES_TYPES_TOTAL];

/* Globals needed for utilization sorting */
static double		*valid_utils = NULL;
static pm_node_t	**pm_nodes = NULL;

/* Globals needed for managing policy polling */
/* Used to alert pmi_poll_policy() that pmi_process_policy_change() is done */
static bool_t	policy_poll_seq = false;
static bool_t	policy_poll_stop;
static cond_t	policy_poll_cv = DEFAULTCV;
static mutex_t	boot_list_lock = DEFAULTMUTEX;
static mutex_t	policy_change_lock = DEFAULTMUTEX;

/* Globals for HPS */
static pm_hps_domain_t	*active_dom_list = NULL;
static pm_hps_domain_t	**hps_dom_list = NULL;
static int		hps_dom_idx = 0;

/* Struct for holding policy registration data. */
typedef struct {
	pmi_policy_t	prp_policy;
	int		prp_rv;
} pmi_reg_policy_t;

/* Struct for holding data for handling guest domain boots. */
typedef struct pmi_domain_boot pmi_domain_boot_t;
struct pmi_domain_boot {
	srcid_t			domid;
	pmi_domain_boot_t	*next;
	int			num_attempts;
	bool_t			boot_ready;
	bool_t			in_boot_delay;
};

/* Global needed for keeping track of guest domains booting */
static pmi_domain_boot_t	*booting_domain_list = NULL;

/* Internal functions */
static int pmi_calc_hps_strands(int, int, double, double,
    pm_node_t *);
static int pmi_change_to_elastic(void);
static int pmi_register_for_notifications(void);
static int pmi_same_res(pm_resid_t *, pm_resid_t *);
static int pmi_unregister_for_notifications(void);
static pwr_req_status_t pmi_create_active_domain_list(int *);
static pwr_req_status_t pmi_create_hps_domain_list(pmi_notify_t *,
    int);
static pwr_req_status_t pmi_process_hps(pmi_notify_t *);
static pwr_req_status_t pmi_process_hpsdl(pmi_notify_t *);
static void pmi_adjust_util_res_pwr(ldm_resutils_t *, pmi_notify_t *,
    int, int);
static void pmi_bundle_status(int *, char *, int, ldm_err_t *,
    pm_resid_t *, int *);
static void pmi_calc_power_change(ldm_resutils_t *, int *);
static void pmi_collect_stats(void);
static void pmi_free_active_list_utils(pm_hps_domain_t *);
static void pmi_free_domain_utils(ldm_resutils_t *);
static void pmi_free_hps(void);
static void pmi_gather_allocated_resources(int, srcid_t *, int *,
    pm_res_type_t, pm_node_t *, pm_resid_t *, pm_node_t *, int);
static void pmi_gather_unallocated_resources(pm_resid_t *, int, int,
    pm_node_t *);
static void pmi_process_policy_change(void *);
static void pmi_recurse_calc_hps_strands(pm_hps_domain_t *, int);
static void pmi_unlink_domain(srcid_t);
static void pmi_util_adjust_power(pm_res_type_t, pm_config_op_t,
    pmi_pwr_state_t, int, int, int *, int, int);
static void *pmi_poll_policy(void *);

/*
 * pmi_get_adjust_limit - return the max number of resources to adjust power to
 * so that we don't spend too much time on the sequencer.
 */
static int
pmi_get_adjust_limit(pm_res_type_t res_type, int count)
{
	int limit;

	switch (res_type) {
	case PM_RES_CPU:
		limit = min(count, option.max_pm_cpu_adjust);
		break;

	case PM_RES_MEM:
		limit = MEM_RES_SEQ_LIMIT;
		break;

	default:
		DBG(zprintf("%s: unknown resource type (%d)\n", __func__,
		    res_type));
		limit = count;
	}

	return (limit);
}


/*
 * pmi_process_config_change - Processes a platform configuration change.
 * Implements resource power management to accommodate the configuration change.
 *
 * Return: PS_SUCCESS, PS_FAILURE
 */
int
pmi_process_config_change(pmi_notify_t *config_change, ldm_err_t *err)
{
	pm_resid_t		*residsp = NULL;
	pm_resid_t		*residsp2 = NULL;
	pmi_domain_boot_t	*booting_domain_cur = NULL;
	pmi_domain_boot_t	*guest_domain = NULL;
	pwr_req_status_t 	rtn;
	void			*srcidp;
	char			*log_msg = "";
	int			cc;
	int			num_doms;
	int			pwr_flag = PCO_NONE;
	int			res_count = 0;
	int			*res_statusp = NULL;
	int			rv = PS_SUCCESS;

	srcidp = (void *)config_change->pn_src_ids[0];

	switch (config_change->pn_flag) {
	case START_DOMAIN_PRE:
		if (GET_DOMAIN_CONFIG(config_change->pn_src_ids[0], PM_RES_CPU,
		    &residsp, &res_count, err) != PS_SUCCESS) {
			DBG(ldm_log_msg("%s: Unable to obtain domain config "
			    "(err = %d)\n", __func__, err->ldme_errs));
			rv = PS_FAILURE;
			break;
		}

		DBG_BLOCK {
			DBG6(zprintf("%s: START_DOMAIN_PRE: %s\n",
			    __func__, GET_DOMAIN_NAME(srcidp)));
			for (cc = 0; cc < res_count; cc++) {
				DBG6(zprintf("%s: Start: %"PRIu64" \n",
				    __func__,
				    (&residsp[cc])->u.cpu.pi_id_val));
			}
		}

		cc = pmi_adjust_power(PM_RES_CPU, PCO_ADD_RESOURCE,
		    PPS_FULL_POWER, res_count, residsp, &res_statusp,
		    PM_NO_FLAGS);
		cc &= ~(PWR_OP_EPARENTUPFAILED | PWR_OP_ETIMEDOUT);
		if (cc != PWR_OP_OK) {
			log_msg = "Unable to power up any";
			rv = PS_FAILURE;
		}

		pmi_bundle_status(res_statusp, log_msg, res_count,
		    err, residsp, NULL);

		break;

	/*
	 * FIXME A no-op at this time due to not being able to power down
	 * resources in a bound state.  This may change in the future but
	 * for now it will cause watchdog errors in the HV.
	 */
	case STOP_DOMAIN_POST:

		DBG6(zprintf("%s: STOP_DOMAIN_POST: %s pn_count=%d\n",
		    __func__, GET_DOMAIN_NAME(config_change->pn_src_ids[0]),
		    config_change->pn_count));

		/*
		 * Since this Domain is no longer Active the HPS nodes
		 * associated with it could have been power adjusted due to
		 * conditions in this Domain.  So, it makes sense that the
		 * HPS nodes could possibly be set to a lower state.  In order
		 * to determine this, we have to run a fake Util adjustment
		 * on all Active domains.
		 */
		if (pmi_is_hps_supported()) {
			rtn = pmi_create_active_domain_list(&num_doms);
			if (rtn == PS_SUCCESS) {
				rtn = pmi_create_hps_domain_list(
				    NULL, num_doms);
				if (rtn == PS_SUCCESS) {
					pmi_collect_stats();
					pmi_process_hpsdl(config_change);
				}
			}
			pmi_free_hps();
		}

		break;

	case ADD_RESOURCE_DEFRAG:
	case ADD_RESOURCE_PRE:
		/*
		 * If a defrag is in process, resource is allocated to a
		 * domain so the right flag needs to be passed to
		 * pmi_adjust_power.
		 */
		if (config_change->pn_flag == ADD_RESOURCE_DEFRAG)
			pwr_flag = PCO_ADD_RESOURCE;
		else
			pwr_flag = PCO_NONE;

		DBG_BLOCK {
			residsp2 = config_change->pn_resids;
			DBG6(zprintf("%s: ADD_RESOURCE_*: res_cnt=%d ldom=%s\n",
			    __func__, config_change->pn_count,
			    GET_DOMAIN_NAME(srcidp)));
			if (config_change->pn_flag == ADD_RESOURCE_DEFRAG) {
				for (cc = 0; cc < config_change->pn_count; cc++)
					DBG6(zprintf("%s: ADD_RESOURCE_DEFRAG: "
					    "%"PRIu64" \n", __func__,
					    (&residsp2[cc])->u.cpu.pi_id_val));
			} else {
				for (cc = 0; cc < config_change->pn_count; cc++)
					DBG6(zprintf("%s: ADD_RESOURCE_PRE: "
					    "%"PRIu64" \n", __func__,
					    (&residsp2[cc])->u.cpu.pi_id_val));
			}
		}

		cc = pmi_adjust_power(config_change->pn_res_type,
		    pwr_flag, PPS_FULL_POWER,
		    config_change->pn_count, config_change->pn_resids,
		    &res_statusp, PM_NO_FLAGS);
		cc &= ~(PWR_OP_EPARENTUPFAILED | PWR_OP_ETIMEDOUT);
		if (cc != PWR_OP_OK) {
			log_msg = "During add operation, unable to power up";
			rv = PS_FAILURE;
		}

		pmi_bundle_status(res_statusp, log_msg,
		    config_change->pn_count, err,
		    config_change->pn_resids, NULL);

		break;

	case REM_RESOURCE_DEFRAG:
	case REM_RESOURCE_POST:

		/*
		 * If a defrag is in process, resource is allocated to a
		 * domain so the right flag needs to be passed to
		 * pmi_adjust_power.
		 */
		if (config_change->pn_flag == REM_RESOURCE_DEFRAG)
			pwr_flag = PCO_REMOVE_RESOURCE;
		else
			pwr_flag = PCO_NONE;

		DBG_BLOCK {
			residsp2 = config_change->pn_resids;
			DBG6(zprintf("%s: REM_RESOURCE_*: res_cnt=%d ldom=%s\n",
			    __func__, config_change->pn_count,
			    GET_DOMAIN_NAME(srcidp)));

			if (config_change->pn_flag == REM_RESOURCE_DEFRAG) {
				for (cc = 0; cc < config_change->pn_count; cc++)
					DBG6(zprintf("%s: REM_RESOURCE_DEFRAG: "
					    "%"PRIu64" \n", __func__,
					    (&residsp2[cc])->u.cpu.pi_id_val));
			} else {
				for (cc = 0; cc < config_change->pn_count; cc++)
					DBG6(zprintf("%s: REM_RESOURCE_POST: "
					    "%"PRIu64" \n", __func__,
					    (&residsp2[cc])->u.cpu.pi_id_val));
			}
		}

		cc = pmi_adjust_power(config_change->pn_res_type,
		    pwr_flag, PPS_LOWEST_POWER, config_change->pn_count,
		    config_change->pn_resids, &res_statusp, PM_NO_FLAGS);
		cc &= ~(PWR_OP_ENEEDONE | PWR_OP_ETIMEDOUT |
		    PWR_OP_ERECONFIGURE);
		if (cc != PWR_OP_OK) {
			log_msg = "During remove op, unable to power down";
			rv = PS_FAILURE;
		}

		pmi_bundle_status(res_statusp, log_msg,
		    config_change->pn_count, err,
		    config_change->pn_resids, NULL);

		/*
		 * Check to see if there are no more strands powered on
		 * for any HPS node in the system.  This is an umbrella
		 * check instead of trying to specifically find only the
		 * HPS nodes affected by the Domain containing this/these
		 * CPU resources.
		 * IOW, when resources are removed, such as a 'ldm unbind'
		 * then those resources could be the last ones that were
		 * active on a node.  If so, that node can be set to its
		 * lowest state.
		 */
		if (pmi_is_hps_supported())
			pmi_hps_adjust_pwr(HPS_SOURCE_NO_PWR, NULL,
			    HPS_LOW_POWER, HPS_CALC_NUM_NODES);
		break;

	case GUEST_DOMAIN_DR_REG:
		/*
		 * When a guest domain boots, a notification gets sent
		 * so that PM will know to power up its resources.
		 */
		guest_domain = Xmalloc(sizeof (pmi_domain_boot_t));
		guest_domain->domid = config_change->pn_src_ids[0];
		guest_domain->num_attempts = 0;
		guest_domain->boot_ready = true;
		guest_domain->in_boot_delay = false;
		guest_domain->next = NULL;

		/* Add domain to a linked list of booting domains */
		(void) mutex_lock(&boot_list_lock);
		booting_domain_cur = booting_domain_list;
		if (!booting_domain_cur) {
			booting_domain_list = guest_domain;
		} else {
			while (booting_domain_cur->next != NULL) {
				booting_domain_cur =
				    booting_domain_cur->next;
			}
			booting_domain_cur->next = guest_domain;
		}
		(void) mutex_unlock(&boot_list_lock);

		DBG6(zprintf("%s: GUEST_DOMAIN_DR_REG: ldom=%s\n",
		    __func__, GET_DOMAIN_NAME(srcidp)));

		break;

	case GUEST_DOMAIN_DOWN:
		/*
		 * The guest is halting, resetting, or suspended.
		 * If the guest is queued to be powered up, remove it.
		 */
		pmi_unlink_domain(config_change->pn_src_ids[0]);

		DBG6(zprintf("%s: GUEST_DOMAIN_DOWN: ldom=%s\n",
		    __func__, GET_DOMAIN_NAME(srcidp)));

		break;

	default:
		DBG(ldm_log_msg("%s: Invalid configuration change "
		    "flag.\n", __func__));
		rv = PS_FAILURE;
	}

	pmi_free_hps();
	Xfree(residsp);

	return (rv);
}


/*
 * pmi_unlink_domain - Removes a domain from the list of guest domains
 * that are rebooting.
 */
void
pmi_unlink_domain(srcid_t domid)
{
	pmi_domain_boot_t	*temp_domain = NULL;
	pmi_domain_boot_t	*cur_domain = NULL;

	(void) mutex_lock(&boot_list_lock);
	cur_domain = booting_domain_list;
	while (cur_domain != NULL) {
		if (domid == cur_domain->domid) {
			/*
			 * If domain is in a boot delay, remove it from
			 * the list, but don't free it. pmi_boot_delay
			 * will free the domain.
			 */
			if ((cur_domain->next == NULL) &&
			    (temp_domain == NULL)) {
				/* domid is only element */
				if (!cur_domain->in_boot_delay)
					Xfree(cur_domain);
				booting_domain_list = NULL;
				break;
			} else {
				if (temp_domain == NULL) {
					/* domid is first element */
					booting_domain_list =
					    cur_domain->next;
					if (!cur_domain->in_boot_delay)
						Xfree(cur_domain);
					cur_domain = booting_domain_list;
				} else {
					/* domid found inside list */
					temp_domain->next = cur_domain->next;
					if (!cur_domain->in_boot_delay)
						Xfree(cur_domain);
					cur_domain = temp_domain->next;
				}
				continue;
			}
		}
		temp_domain = cur_domain;
		cur_domain = temp_domain->next;
	}
	(void) mutex_unlock(&boot_list_lock);
}


/*
 * pmi_process_util_change - Implements resource power management when
 * utilization for one or more groups of resources is out of band for a
 * duration exceeding the PM threshold. If utilization is too low, resource
 * power will be lowered. If utilization is too high, resource power will be
 * increased.
 *
 * Return: PS_SUCCESS, PS_FAILURE
 */
int
pmi_process_util_change(pmi_notify_t *util_change_notify)
{
	bool_t		run_defrag = false;
	ldm_err_t	err = { 0 };
	ldm_resutils_t	*res_utilsp = NULL;
	pmi_policy_t	pmi_current_policy;
	void		*srcidp;
	int		cnt_idx;
	int		num_res_pwr_chg = 0;
	int		rv = PS_SUCCESS;
	int		type_idx = PM_RES_CPU;
	double		pm_low_util_wm, pm_hi_util_wm;


	/* Make sure policy is still elastic. */
	if ((*pmi_get_policy)(&pmi_current_policy) != PS_SUCCESS) {
		DBG(ldm_log_msg("%s: Unable to retrieve PM policy.\n",
		    __func__));
		rv = PS_FAILURE;
		goto bail;
	}

	if (pmi_current_policy != PM_POLICY_ELASTIC) {
		DBG(ldm_log_msg("%s: Current power management policy is not "
		    "elastic. Unable to process utilization "
		    "notifications.\n", __func__));
		rv = PS_FAILURE;
		goto bail;
	}

	/*
	 * For systems that have HPS nodes that can have their Pwr state
	 * changed, we want to power adjust the HPS node and the
	 * number of strands per domain in order to obtain the best power
	 * savings and remain within utilization bounds for each domain.  There
	 * will be exceptions and those are explained in the following function.
	 */

	/*
	 * If the system has HPS nodes, then attempt the HPS node and strand
	 * power adjustment here and skip the original CoreDisable code.
	 */

	if (pmi_is_hps_supported()) {

		if (pmi_process_hps(util_change_notify) == PS_SUCCESS) {
			/* HPS and/or strands adjusted, we are done */
			run_defrag = true;
			type_idx = util_change_notify->pn_res_type;
			if (run_defrag && option.allow_defrag)
				DEFRAGMENT(PACK,
				    util_change_notify->pn_res_type, &err);
			goto bail;
		} else {
			DBG6(zprintf("%s: HPS failure, fallthrough to "
			    "CoreDisable\n", __func__));
		}

		/* Fall through */
		/* Need to do orig CoreDisable; No adjustement made yet */
	}

	/*
	 * Cycle through each domain or resource that we received a
	 * notification for. For PM'able memory, since it does not bind to
	 * domains, pn_count == the number of groupings that triggered the
	 * notification. For resources bound to domains, such as CPUs,
	 * pn_count == the number of domains that generated a util notification.
	 */
	for (cnt_idx = 0; cnt_idx < util_change_notify->pn_count;
	    cnt_idx++) {

		srcidp = (void *)util_change_notify->pn_src_ids[cnt_idx];

		/* Get utilization data. */
		if (GET_DOMAIN_UTILS(util_change_notify->pn_src_ids[cnt_idx],
		    util_change_notify->pn_res_type, &res_utilsp,
		    &err) != PS_SUCCESS) {
			DBG(ldm_log_msg("%s: Unable to obtain resource"
			    " utilization data. (err = %d)\n",
			    __func__, err.ldme_errs));
			continue;
		}

		if (util_change_notify->pn_res_type == PM_RES_CPU) {
			pm_low_util_wm = DEFAULT_CPU_LOW_UTIL_WM;
			pm_hi_util_wm = DEFAULT_CPU_HI_UTIL_WM;
		} else if (util_change_notify->pn_res_type == PM_RES_MEM) {
			pm_low_util_wm = DEFAULT_MEM_LOW_UTIL_WM;
			pm_hi_util_wm = DEFAULT_MEM_HI_UTIL_WM;
		} else {
			DBG(ldm_log_msg("%s: Invalid resource type passed in"
			    " via a utilization notification.\n", __func__));
			rv = PS_FAILURE;
			goto bail;
		}

		if (!PM_UTIL_IS_VALID(res_utilsp, pm_low_util_wm,
		    pm_hi_util_wm)) {
			DBG6(ldm_log_msg("%s: Resource utilization is no longer"
			    " out of band for %s.\n",
			    __func__, GET_DOMAIN_NAME(srcidp)));
			pmi_free_domain_utils(res_utilsp);
			continue;
		}

		pmi_calc_power_change(res_utilsp, &num_res_pwr_chg);

		/*
		 * Adjust resource power based on utilization
		 * calculations and power data.
		 */
		if (num_res_pwr_chg > 0) {

			DBG6(zprintf("%s: GET_DOMAIN_UTILS returns %d cpus"
			    " in %s\n", __func__, res_utilsp->ldmr_rescnt,
			    GET_DOMAIN_NAME(srcidp)));

			DBG6(zprintf("%s: %s %d cpus from domain %s\n",
			    __func__,
			    (res_utilsp->ldmr_util_dir == PM_UTIL_LOW) ?
			    "Remove" : "Add", num_res_pwr_chg,
			    GET_DOMAIN_NAME(srcidp)));

			pmi_adjust_util_res_pwr(res_utilsp,
			    util_change_notify, num_res_pwr_chg,
			    res_utilsp->ldmr_util_dir);

			/* Only run defrag on CPUs */
			if (util_change_notify->pn_res_type == PM_RES_CPU)
				run_defrag = true;
		} else {
			DBG6(zprintf("%s: No resource change needed for "
			    " %s\n", __func__, GET_DOMAIN_NAME(srcidp)));
		}

		pmi_free_domain_utils(res_utilsp);
	}

	type_idx = util_change_notify->pn_res_type;

	if (run_defrag && option.allow_defrag)
		DEFRAGMENT(PACK, util_change_notify->pn_res_type, &err);

bail:
	if (RESUME_UTIL_CHANGE_NOTIFY(pm_util_hdl[type_idx],
	    &err) != PS_SUCCESS) {
		DBG(ldm_log_msg("%s: Unable to resume processing of "
		    "utilization notifications. (err = %d)\n", __func__,
		    err.ldme_errs));
		rv = PS_FAILURE;
	}

	return (rv);
}


/*
 * pmi_cmp - Used by qsort to compare utilization levels for resources.
 *
 * Returns: 1 if p1 > p2 util or if only p1 power status is unknown
 *	    0 if p1 == p2 util or if p1 and p2 power status is unknown
 *	   -1 if p1 < p2 util or if only p2 power status is unknown
 */
static int
pmi_cmp(const void *p1, const void *p2)
{
	int 		n1 = *(int *)p1;
	int 		n2 = *(int *)p2;
	pm_node_t	*pm_nodep1 = pm_nodes[n1];
	pm_node_t	*pm_nodep2 = pm_nodes[n2];

	if (pm_nodep1->pn_pwrs->pps_status == PM_STAT_UNKNOWN) {
		if (pm_nodep2->pn_pwrs->pps_status == PM_STAT_UNKNOWN) {
			/* PM_STATE_UNKNOWN == PM_STATE_UNKNOWN */
			return (0);
		} else {
			/* PM_STATE_UNKNOWN > any util */
			return (1);
		}
	} else if (pm_nodep2->pn_pwrs->pps_status == PM_STAT_UNKNOWN) {
		/* any_util < PM_STATE_UNKNOWN */
		return (-1);
	} else if (valid_utils[n1] < valid_utils[n2]) {
		/* util n1 < util n2 */
		return (-1);
	} else if (valid_utils[n1] > valid_utils[n2]) {
		/* util n1 > util n2 */
		return (1);
	}

	/* util n1 == util n2 */
	return (0);
}


/*
 * pmi_adjust_util_res_pwr - Searches for the the correct resources, sorts
 * them by utilization, and adjusts power on those resources.
 */
void
pmi_adjust_util_res_pwr(ldm_resutils_t *res_utilsp,
    pmi_notify_t *util_change_notify, int num_res_pwr_chg, int direction)
{
	pm_node_t	**pm_node = NULL;
	int		res_idx;
	int		power_level = PPS_LOWEST_POWER;
	int		reconfig_action = PCO_REMOVE_RESOURCE;
	int		*ids = NULL;
	int		id_cnt = 0;
	int		start_id = 0;
	bool_t		is_drd_out = false;

	/* Array to hold reference to all nodes */
	pm_nodes = (pm_node_t **)Xcalloc(res_utilsp->ldmr_rescnt,
	    pm_node_t *);
	pm_node = pm_nodes;

	/* Array to hold sorted node indexes by utilization */
	ids = Xcalloc(res_utilsp->ldmr_rescnt, int);

	valid_utils = Xcalloc(res_utilsp->ldmr_rescnt, double);

	/* Find resources with appropriate power levels. */
	for (res_idx = 0;
	    res_idx < res_utilsp->ldmr_rescnt; res_idx++) {
		*pm_node = epm_find_node(
		    &(res_utilsp->ldmr_util_p[res_idx].ldmu_resid),
		    NULL, NULL);

		if (*pm_node == NULL) {
			DBG(ldm_log_msg("%s: Resource not found in "
			    "device tree.\n", __func__));
			continue;
		}

		if (pmi_is_resource_faulted((*pm_node)->pn_rcfg)) {
			DBG(ldm_log_msg("%s: Resource's fma status is "
			    "faulted: type=%d, id=%d, fma_status=%d\n",
			    __func__, (*pm_node)->pn_rcfg->pi_res_type,
			    (*pm_node)->pn_nodeid,
			    (*pm_node)->pn_rcfg->pi_fma_status));
			continue;
		}

		if (direction == PM_UTIL_LOW) {
			/*
			 * Ignore powered off resources. Resources with an
			 * unknown power status are considered off so ignore
			 * those too.
			 */
			if (!epm_has_perf(*pm_node, PM_ASSUME_LOW)) {
				continue;
			}
		} else if (direction == PM_UTIL_HIGH) {
			if (res_utilsp->ldmr_type == PM_RES_CPU) {
				/* Skip over CPUs that are already DRd in. */
				if (GET_RESOURCE_PM_STATUS((*pm_node)->pn_rcfg,
				    PM_DR_STATUS, &is_drd_out) == 0 &&
				    is_drd_out == false)
					continue;
			} else if (res_utilsp->ldmr_type == PM_RES_MEM) {
				/*
				 * Skip over memory groupings that are already
				 * powered on.
				 */
				if (epm_has_perf(*pm_node, PM_ASSUME_HIGH))
					continue;
			} else {
				DBG(ldm_log_msg("%s: Invalid resource type: "
				    "%d.\n", __func__, res_utilsp->ldmr_type));
				goto bail;
			}
		}

		valid_utils[id_cnt] =
		    res_utilsp->ldmr_util_p[res_idx].ldmu_util;
		pm_node++;
		ids[id_cnt] = id_cnt;
		id_cnt++;
	}

	/* Nothing to do if id_cnt is 0 */
	if (id_cnt == 0)
		goto bail;

	/* Sort indexes of nodes by util. */
	qsort((void *)(&ids[0]), id_cnt, sizeof (int), pmi_cmp);

	switch (res_utilsp->ldmr_type) {
	case PM_RES_CPU:
		if (direction == PM_UTIL_LOW) {
			power_level = PPS_LOWEST_POWER;
			reconfig_action = PCO_REMOVE_RESOURCE;
		} else if (direction == PM_UTIL_HIGH) {
			power_level = PPS_FULL_POWER;
			reconfig_action = PCO_ADD_RESOURCE;
		}
		break;
	case PM_RES_MEM:
		if (direction == PM_UTIL_LOW) {
			power_level = PPS_LOWEST_POWER;
			reconfig_action = PCO_NONE;
		} else if (direction == PM_UTIL_HIGH) {
			power_level = PPS_FULL_POWER;
			reconfig_action = PCO_NONE;
		}
		break;
	default:
		break;
	}

	/* Actual final id is one less */
	id_cnt--;

	/* Adjust power on resources. */
	pmi_util_adjust_power(util_change_notify->pn_res_type,
	    reconfig_action, power_level, num_res_pwr_chg,
	    id_cnt, ids, start_id, num_res_pwr_chg);
bail:
	Xfree(ids);
	Xfree(valid_utils);
	Xfree(pm_nodes);
}


/*
 * pmi_util_adjust_power - Grabs the first num_res_pwr_chg resources in
 * the pm_nodes array and adjusts power on them. It then takes retry_cnt,
 * returned by pmi_bundle_status(), and if > 0, grabs the next retry_cnt
 * number of resources and adjusts power on them. This is repeated
 * until either num_res_pwr_chg number of resources have had their power
 * adjusted, or all resources in pm_nodes have been exhausted.
 */
void
pmi_util_adjust_power(pm_res_type_t res_type, pm_config_op_t reconfig_action,
    pmi_pwr_state_t power_level, int num_res_pwr_chg, int last_id,
    int *all_ids, int start_id, int end_id)
{
	pm_resid_t		*residsp = NULL;
	int			retry_cnt = 0;
	int			res_idx, node_idx;
	int			id_idx = 0;
	int			*res_statusp = NULL;
	ldm_err_t		tot_res_status = { 0 };
	char			*log_msg = "";

	/*
	 * If power adjustment fails on any # of
	 * resources, try other resources.
	 */
	while (num_res_pwr_chg > 0) {
		/* Construct resource array. */
		residsp = Xcalloc(num_res_pwr_chg, pm_resid_t);
		for (res_idx = 0, id_idx = start_id; res_idx < num_res_pwr_chg;
		    res_idx++, id_idx++) {
			node_idx = all_ids[id_idx];
			residsp[res_idx] = *(pm_nodes[node_idx]->pn_rcfg);
		}

		/*
		 * retry_cnt keeps track of retries needed per call to
		 * pmi_adjust_power() so need to reset with every call.
		 */
		retry_cnt = 0;

		if (pmi_adjust_power(res_type, reconfig_action, power_level,
		    num_res_pwr_chg, residsp, &res_statusp, PM_NO_FLAGS) !=
		    PWR_OP_OK) {
			log_msg = "Unable to adjust power on";
		}

		pmi_bundle_status(res_statusp, log_msg, num_res_pwr_chg,
		    &tot_res_status, residsp, &retry_cnt);

		/* Need to report to domain mgr. */
		RES_POWER_FAULT_REPORT(&tot_res_status, LDME_PE_UTIL);

		start_id = end_id;
		end_id = min(last_id, (end_id + retry_cnt));
		num_res_pwr_chg = end_id - start_id;

		Xfree(residsp);
	}
}


/*
 * pmi_power_up_stuck_high_resources tries to DR in any stuck
 * high cpus it finds.
 */
static void
pmi_power_up_stuck_high_resources(void *ignore)
{
	pm_node_t		*pm_nodep = NULL;
	int			res_idx;
	bool_t			is_drd_out = false;
	int			*status = NULL;
	int			stuck_cnt = 0;
	static int		tot_res_count = 0;
	static pm_node_t	**pm_nodepp = NULL;
	static pm_resid_t	*stuck_res = NULL;

	/*
	 * Find all nodes in the PRI for this resource type.
	 * This assumes a static PRI.
	 */
	if (pm_nodepp == NULL) {
		pmi_find_res_nodeps(NULL, PM_RES_CPU, &pm_nodepp,
		    &tot_res_count);
		if (tot_res_count > 0) {
			stuck_res = Xcalloc(tot_res_count, pm_resid_t);
		}
	}

	/* Scan every resource and find the stuck high cpus */
	for (res_idx = 0; res_idx < tot_res_count; res_idx++) {
		pm_nodep = pm_nodepp[res_idx];

		if (pmi_is_resource_faulted(pm_nodep->pn_rcfg)) {
			continue;
		}

		/* Skip powered off resources */
		if (!epm_has_perf(pm_nodep, PM_ASSUME_LOW))
			continue;

		/*
		 * Check if resource is DR'd out of a domain.  If so
		 * then add it to the list of CPUs to be DR'd in.
		 */
		if (GET_RESOURCE_PM_STATUS(pm_nodep->pn_rcfg, PM_DR_STATUS,
		    &is_drd_out) == 0 && is_drd_out == true) {
			DBG(zprintf("%s: CPU %"PRIu64"  is stuck high so "
			    "attempting to DR in\n", __func__,
			    pm_nodep->pn_rcfg->u.cpu.pi_id_val));
			stuck_res[stuck_cnt] = *(pm_nodep->pn_rcfg);
			stuck_cnt++;
		}
	}

	/* DR in stuck high CPUs */
	if (stuck_cnt > 0)
		(void) pmi_adjust_power(PM_RES_CPU, PCO_ADD_RESOURCE,
		    PPS_FULL_POWER, stuck_cnt, stuck_res, &status, PM_NO_FLAGS);

	Xfree(status);
}

/*
 * pmi_calc_power_change - Calculates the number of resources that
 * need their power adjusted to accommodate domain utilization.
 */
void
pmi_calc_power_change(ldm_resutils_t *res_utilsp, int *final_res_pwr_chg)
{
	pm_node_t	*pm_nodep = NULL;
	int		res_idx;
	double		num_res_pwr_chg;
	double		num_on = 0;
	double		num_off = 0;
	double		amt_over_target = 0;
	double		amt_under_target = 0;
	double		median_wm = 0;
	double		pm_low_util_wm, pm_hi_util_wm;

	/*
	 * Find number of resources that are powered on with a
	 * known status
	 */
	for (res_idx = 0; res_idx < res_utilsp->ldmr_rescnt; res_idx++) {
		pm_nodep = epm_find_node(
		    &(res_utilsp->ldmr_util_p[res_idx].ldmu_resid),
		    NULL, NULL);

		if (pm_nodep == NULL) {
			DBG(ldm_log_msg("%s: Resource not found in"
			" device tree.\n", __func__));
			continue;
		}

		if (pmi_is_resource_faulted(pm_nodep->pn_rcfg)) {
			continue;
		}

		/*
		 * Resources with a known power status that consume power
		 * are considered on.  All others are considered off.
		 */
		if (epm_has_perf(pm_nodep, PM_ASSUME_LOW)) {
			num_on++;
		} else {
			num_off++;
		}
	}

	if (res_utilsp->ldmr_type == PM_RES_CPU) {
		pm_low_util_wm = DEFAULT_CPU_LOW_UTIL_WM;
		pm_hi_util_wm = DEFAULT_CPU_HI_UTIL_WM;
	} else if (res_utilsp->ldmr_type == PM_RES_MEM) {
		pm_low_util_wm = DEFAULT_MEM_LOW_UTIL_WM;
		pm_hi_util_wm = DEFAULT_MEM_HI_UTIL_WM;
	} else {
		DBG(ldm_log_msg("%s: Attempt to use invalid resource type.\n",
		    __func__));
		return;
	}

	median_wm = pm_low_util_wm + ((pm_hi_util_wm - pm_low_util_wm) / 2);
	ASSERT(median_wm > 0);

	/* Calculate the number of resources needing power changed. */
	if (res_utilsp->ldmr_util_dir == PM_UTIL_HIGH) {
		/* Calculate amount over threshold */
		amt_over_target = res_utilsp->ldmr_avg_util -
		    median_wm;

		/*
		 * Calculate number of resources to turn on.
		 * Always round up so we don't undershoot
		 * the target utilization.
		 */
		num_res_pwr_chg =
		    ((amt_over_target * num_on)/median_wm) +
		    UTIL_CALC_ADJUST;

		/*
		 * Don't want to power on more resources than
		 * available.
		 */
		if (num_res_pwr_chg > num_off) {
			*final_res_pwr_chg = (int)num_off;
		} else {
			*final_res_pwr_chg =
			    (int)(max(1, num_res_pwr_chg));
		}
	} else if (res_utilsp->ldmr_util_dir == PM_UTIL_LOW) {
		/* Calculate amount under threshold */
		amt_under_target = median_wm - res_utilsp->ldmr_avg_util;

		/*
		 * Calculate number of resources to turn off.
		 * If min powered on resource limit in domain
		 * has been reached, can't power off any more
		 * so return.
		 */
		num_res_pwr_chg = (amt_under_target * num_on)/median_wm;

		if ((num_on - num_res_pwr_chg) > option.min_vcpu_per_ldom) {
			*final_res_pwr_chg = (int)(num_res_pwr_chg);
		} else {
			*final_res_pwr_chg = max(0,
			    (int)(num_on - option.min_vcpu_per_ldom));
		}
	}

	*final_res_pwr_chg = pmi_get_adjust_limit(res_utilsp->ldmr_type,
	    *final_res_pwr_chg);

	DBG5(ldm_log_msg("%s: Number of resources able to be power adjusted"
	    " = %d\n", __func__, *final_res_pwr_chg));
}


/*
 * pmi_process_policy_change - Implements resource power management when
 * the power management policy has been changed. The two policies currently
 * handled are PERFORMANCE and ELASTIC.
 *
 * Return: PS_SUCCESS, PS_FAILURE
 */
void
pmi_process_policy_change(void *pp)
{
	pmi_policy_t	pmi_new_policy;
	pmi_policy_t	pmi_current_policy;
	bool_t		pm_stop = false;

	/*
	 * Make sure policy is still different now that
	 * we're on the sequencer.
	 */

	if ((*pmi_get_policy)(&pmi_new_policy) != PS_SUCCESS) {
		DBG(ldm_log_msg("%s: Unable to retrieve PM policy.\n",
		    __func__));
		((pmi_reg_policy_t *)pp)->prp_rv = PS_FAILURE;
		goto cleanup;
	}

	/*
	 * If the current and the new policies are the same, no
	 * need to continue.
	 */
	pmi_current_policy = ((pmi_reg_policy_t *)pp)->prp_policy;

	if (pmi_current_policy == pmi_new_policy) {
		goto cleanup;
	}

	switch (pmi_new_policy) {
	case PM_POLICY_ELASTIC:
		/*
		 * FIXME
		 * When in delayed reconfig mode for a domain then we can't
		 * do PM any more when switching to elastic mode until the
		 * delayed reconfig domain has rebooted.  This is because
		 * the delayed reconfig domain is still running with the
		 * prior cpu config, but the config PM will have access to
		 * is the config after the domain reboots.  This will panic
		 * the domain if we allow PM while in this state.
		 */
		if (DELAYED_RECONFIG_IN_PROCESS(NULL)) {
			DBG(ldm_log_msg("%s: In delayed reconfiguration mode, "
			    "cannot change to elastic mode.\n", __func__));
			goto cleanup;
		}

		if (MIGRATION_IN_PROCESS()) {
			DBG(ldm_log_msg("%s: Migration in process, cannot "
			    "change to elastic mode.\n", __func__));
			goto cleanup;
		}

		if (DYNAMIC_RESMGMT_IN_PROCESS()) {
			DBG(ldm_log_msg("%s: Dynamic Resource Management "
			    "in process, cannot change to elastic mode.\n",
			    __func__));
			goto cleanup;
		}

		if (pmi_change_to_elastic() == PS_SUCCESS) {
			ldm_log_msg("Power Management policy == "
			    "elastic mode.\n");
		} else {
			DBG(ldm_log_msg("%s: Error while setting Power "
			    "Management policy to elastic mode.\n",
			    __func__));
			((pmi_reg_policy_t *)pp)->prp_rv = PS_FAILURE;
		}

		break;
	case PM_POLICY_PERFORMANCE:
		if (pmi_change_to_performance(pm_stop) == PS_SUCCESS) {
			ldm_log_msg("Power Management policy == "
			    "performance mode.\n");
		} else {
			DBG(ldm_log_msg("%s: Error while setting Power "
			    "Management policy to performance mode.\n",
			    __func__));
			((pmi_reg_policy_t *)pp)->prp_rv = PS_FAILURE;
		}

		break;
	default:
		DBG(ldm_log_msg("%s: Invalid policy. No action "
		    "taken.\n", __func__));
		((pmi_reg_policy_t *)pp)->prp_rv = PS_FAILURE;
		goto cleanup;
	}

	((pmi_reg_policy_t *)pp)->prp_policy = pmi_new_policy;

cleanup:
	(void) mutex_lock(&policy_change_lock);
	policy_poll_seq = false;
	(void) cond_signal(&policy_poll_cv);
	(void) mutex_unlock(&policy_change_lock);
}


/*
 * pmi_register_policy_change_event - Kicks off the thread that polls
 * the SP for policy changes.
 *
 * Return: PS_SUCCESS, PS_FAILURE
 */
int
pmi_register_policy_change_event(pmi_policy_t pmi_init_policy)
{
	pmi_reg_policy_t	*pmi_policyp;

	pmi_policyp = (pmi_reg_policy_t *)Xmalloc(sizeof (pmi_reg_policy_t));
	pmi_policyp->prp_policy = pmi_init_policy;
	pmi_policyp->prp_rv = PS_SUCCESS;

	pmi_process_policy_change(pmi_policyp);

	/* Ignore the error but log a warning */
	if (pmi_policyp->prp_rv != PS_SUCCESS) {
		ldm_log_msg("Warning: Unable to bring resources to initial "
		    "power states.\n");
	}

	/*
	 * The policy poller thread must always start as SP could be
	 * booting.
	 */
	policy_poll_stop = false;
	if (thr_create(NULL, NULL, pmi_poll_policy, pmi_policyp,
	    THR_DAEMON, NULL) != PS_SUCCESS) {
		ldm_log_msg("Fatal: Unable to launch PM Policy polling and "
		    "processing thread.\n");
		pm_stop(NULL);	/* Bring resources back to full power. */
		Xfree(pmi_policyp);
		return (PS_FAILURE);
	}

	return (PS_SUCCESS);
}


/*
 * pmi_poll_policy - Started on a separate thread by
 * pmi_register_policy_change_event() to poll the SP for policy changes.
 *
 * Return: PS_SUCCESS, PS_FAILURE
 */
void *
pmi_poll_policy(void *policyp)
{
	pmi_policy_t	new_policy;
	time_t		time_now = 0;
	static time_t	time_before = 0;

	/* Wait for a bit and then start the policy polling thread. */
	(void) sleep(SLEEP_TIME);

	while (!policy_poll_stop) {
		if ((*pmi_get_policy)(&new_policy) != PS_SUCCESS) {
			DBG(ldm_log_msg("%s: Unable to retrieve PM policy.\n",
			    __func__));

			/*
			 * Something may be wrong with the system.
			 * Wait for a while longer before trying again.
			 */
			(void) sleep(EXTENDED_SLEEP_TIME);
			continue;
		}

		/*
		 * Only enqueue on Zeus sequencer if policy has changed.
		 * Don't pass in the new policy as it may change by the time
		 * this is processed on the sequencer.
		 */
		if (((pmi_reg_policy_t *)policyp)->prp_policy != new_policy) {
			(void) mutex_lock(&policy_change_lock);
			policy_poll_seq = true;
			ENQUEUE_WORK(pmi_process_policy_change, policyp);
			while (policy_poll_seq) {
				cond_wait(&policy_poll_cv, &policy_change_lock);
			}

			(void) mutex_unlock(&policy_change_lock);
		}

		/*
		 * Power up stuck high resources every so often.
		 * A stuck high resource is a
		 * resource that has power but isn't DR'd into a domain.  The
		 * way to handle these is to DR them into the domain then allow
		 * PM to manage them next time.  We can't power them down as
		 * they could be boot cpus in a bound domain which would panic
		 * the guest and/or the primary domain.
		 */
		if (time_before == 0) {
			/* Intialize timeout for stuck high */
			time(&time_before);
		}

		time(&time_now);
		if (time_now - time_before > PM_STUCK_HIGH_TIMEOUT) {
			/* Time out has expired so call stuck high routine */
			int ignore;

			ENQUEUE_WORK(pmi_power_up_stuck_high_resources,
			    &ignore);
			time(&time_before);
		}

		(void) sleep(SLEEP_TIME);
	}

	DBG2(ldm_log_msg("%s thread exiting\n", __func__));

	Xfree(policyp);
	return (NULL);
}


/*
 * pmi_change_to_elastic - Powers down power manageable resources not allocated
 * to any domains and registers to receive configuration change and utilization
 * out of band notifications.
 *
 * Return: PS_SUCCESS, PS_FAILURE
 */
int
pmi_change_to_elastic(void)
{
	ldm_err_t		get_domid_err = { 0 };
	ldm_err_t		tot_res_status = { 0 };
	pm_node_t		*all_res_nodesp = NULL;
	pm_node_t		*alloc_nodesp = NULL;
	pm_res_type_t		res_type = PM_RES_CPU;
	pm_resid_t		*unalloc_residsp = NULL;
	srcid_t			*tot_domid_listp = NULL;
	char			*log_msg = "";
	int			alloc_count = 0;
	int			cc, type_idx;
	int			*res_statusp = NULL;
	int			rv = PS_SUCCESS;
	int			tot_res_count = 0;
	int			unalloc_count = 0;
	uint_t			num_domains = 0;


	/* Get domain ids for all domains. */
	if (GET_ALL_DOMAIN_IDS(&tot_domid_listp, &num_domains,
	    &get_domid_err, ALL_DOMAINS) != PS_SUCCESS) {
		DBG(ldm_log_msg("%s: Unable to obtain domain IDs. "
		    "(err = %d)\n", __func__, get_domid_err.ldme_errs));
		rv = PS_FAILURE;
		return (rv);
	}

	/*
	 * Clear the FMA status of all resources.  The FMA status will be
	 * updated/reset when we call pmi_find_res_nodes() a bit later.
	 */
	pmi_clear_all_fma_status(NULL);

	/*
	 * Go through all resource types and lower power levels
	 * for unallocated resources.
	 */
	for (type_idx = 0; type_idx < PM_RES_TYPES_TOTAL;
	    type_idx++, res_type++) {

		/* Reset the pointers for each resource type */
		unalloc_residsp = NULL;
		alloc_nodesp = NULL;
		all_res_nodesp = NULL;

		/* Update power state internal bookkeeping. */
		if (pmi_get_power_all(res_type, NULL) != PS_SUCCESS) {
			DBG(ldm_log_msg("%s: Error while obtaining power "
			    "states.\n", __func__));
		}

		/*
		 * Memory does not bind to domains so no power adjustments
		 * need to be made when going to elastic mode.
		 */
		if (type_idx == PM_RES_MEM) {
			continue;
		}

		/* Reset the total number of resources of the current type. */
		tot_res_count = 0;

		/* Find all nodes in the pm_node tree for this resource type. */
		pmi_find_res_nodes(NULL, res_type, &all_res_nodesp,
		    &tot_res_count);

		/* No nodes found for this resource type so try next type.  */
		if (tot_res_count == 0) {
			continue;
		}

		/*
		 * Cycle through domains and put allocated resource nodes
		 * into an array.
		 */
		alloc_nodesp = Xcalloc(tot_res_count, pm_node_t);

		alloc_count = 0;
		pmi_gather_allocated_resources(num_domains, tot_domid_listp,
		    &alloc_count, res_type, all_res_nodesp, NULL,
		    alloc_nodesp, tot_res_count);

		/*
		 * If there are no unallocated resources, go to next
		 * resource type.
		 */
		unalloc_count = tot_res_count - alloc_count;

		DBG6(zprintf("%s: Total %d; Alloced %d; Unalloced; %d\n",
		    __func__, tot_res_count, alloc_count, unalloc_count));

		if (unalloc_count <= 0) {
			goto cleanup;
		}

		/*
		 * Check for unallocated resources and put those into
		 * an array.
		 */
		unalloc_residsp = Xcalloc(unalloc_count, pm_resid_t);

		pmi_gather_unallocated_resources(unalloc_residsp,
		    tot_res_count, unalloc_count, all_res_nodesp);

		/* Power down unallocated resources. */
		cc = pmi_adjust_power(res_type, PCO_NONE, PPS_LOWEST_POWER,
		    unalloc_count, unalloc_residsp, &res_statusp, PM_NO_FLAGS);
		cc &= ~(PWR_OP_ENEEDONE | PWR_OP_ETIMEDOUT |
		    PWR_OP_ERECONFIGURE);
		if (cc != PWR_OP_OK) {
			log_msg = "Unable to adjust power on any unallocated";
			rv = PS_FAILURE;
		}

		pmi_bundle_status(res_statusp, log_msg, unalloc_count,
		    &tot_res_status, unalloc_residsp, NULL);

		/*
		 * Have to report to domain mgr about failed
		 * resources so that domain mgr will not allow
		 * user to add them via CLI.
		 */
		RES_POWER_FAULT_REPORT(&tot_res_status, LDME_PE_ENTER_PM);

cleanup:
		Xfree(unalloc_residsp);
		Xfree(alloc_nodesp);
		Xfree(all_res_nodesp);

	}

	/*
	 * Now that power adjustments have been made, register to receive
	 * notifications.
	 */
	if ((rv = pmi_register_for_notifications()) != PS_SUCCESS) {
		DBG(zprintf("%s: pmi_register_for_notifications(FAILED) "
		    "rc=%d\n", __func__, rv));
		rv = PS_FAILURE;
	}

	Xfree(tot_domid_listp);
	return (rv);
}


/*
 * pmi_change_to_performance - Brings all power manageable resources to full
 * power. Unregisters to receive configuration change and utilization out of
 * band notifications.
 *
 * Return: PS_SUCCESS, PS_FAILURE
 */
int
pmi_change_to_performance(bool_t pm_stop)
{
	ldm_err_t		get_domid_err = { 0 };
	ldm_err_t		tot_res_status = { 0 };
	pm_node_t		*all_res_nodesp = NULL;
	pm_res_type_t		res_type = PM_RES_CPU;
	pm_resid_t		*alloc_residsp = NULL;
	pm_resid_t		*unalloc_residsp = NULL;
	char			*log_msg = "";
	int			alloc_count = 0;
	int			tot_res_count = 0;
	int			*res_statusp = NULL;
	int			rv = PS_SUCCESS;
	int			type_idx;
	int			unalloc_count = 0;
	srcid_t			*tot_domid_listp = NULL;
	uint_t			num_domains = 0;
	uint32_t		flags = PM_NO_FLAGS;


	if (pm_stop) {
		policy_poll_stop = true;
	}

	/*
	 * Since we're going to PERFORMANCE mode, we unregister for
	 * all notifications except guest domain boot notifications.
	 */
	if (pmi_unregister_for_notifications() != PS_SUCCESS) {
		rv = PS_FAILURE;
		DBG6(ldm_log_msg("%s: Error while unregistering for"
		    " notifications.\n", __func__));
	}

	/* Set all HPS nodes in the system to highest Pwr level */
	if (pmi_is_hps_supported()) {
		DBG6(zprintf("%s: Increase PState to highest for all "
		    "HPS nodes\n", __func__));
		pmi_hps_adjust_pwr(HPS_SOURCE_ALL, NULL, HPS_FULL_POWER,
		    HPS_CALC_NUM_NODES);
	}

	/* Get domain ids for all domains. */
	if (GET_ALL_DOMAIN_IDS(&tot_domid_listp, &num_domains,
	    &get_domid_err, ALL_DOMAINS) != PS_SUCCESS) {
		DBG6(ldm_log_msg("%s: Unable to obtain domain IDs. "
		    "(err = %d)\n", __func__, get_domid_err.ldme_errs));
		rv = PS_FAILURE;
		return (rv);
	}

	/* Go through all resource types and bring them all to full power. */
	for (type_idx = 0; type_idx < PM_RES_TYPES_TOTAL;
	    type_idx++, res_type++) {

		/* Reset the pointers for each resource type */
		alloc_residsp = NULL;
		unalloc_residsp = NULL;
		all_res_nodesp = NULL;

		if (pmi_get_power_all(res_type, NULL) != PS_SUCCESS) {
			DBG(ldm_log_msg("%s: Error while obtaining power "
			    "states.\n", __func__));
		}

		/* Reset the total number of resources of the current type. */
		tot_res_count = 0;

		pmi_find_res_nodes(NULL, res_type, &all_res_nodesp,
		    &tot_res_count);

		/* No nodes found for this resource type so try next type. */
		if (tot_res_count == 0) {
			continue;
		}

		/*
		 * Cycle through domains and put allocated resources
		 * into an array. Since power manageable memory is not bound
		 * to domains, it is considered to be unallocated.
		 */
		alloc_count = 0;
		if (type_idx == PM_RES_MEM) {
			unalloc_count = tot_res_count;
		} else {
			alloc_residsp = Xcalloc(tot_res_count, pm_resid_t);

			pmi_gather_allocated_resources(num_domains,
			    tot_domid_listp, &alloc_count, res_type,
			    all_res_nodesp, alloc_residsp, NULL, tot_res_count);

			unalloc_count = tot_res_count - alloc_count;
		}

		DBG6(zprintf("%s: Total %d; Alloced %d; Unalloced %d\n",
		    __func__, tot_res_count, alloc_count, unalloc_count));

		/*
		 * Check for unallocated resources, put those into
		 * an array and power them up.
		 */
		if (unalloc_count > 0) {
			unalloc_residsp = Xcalloc(unalloc_count, pm_resid_t);

			pmi_gather_unallocated_resources(unalloc_residsp,
			    tot_res_count, unalloc_count,
			    all_res_nodesp);

			if (pmi_adjust_power(res_type, PCO_NONE,
			    PPS_FULL_POWER_FORCE, unalloc_count,
			    unalloc_residsp, &res_statusp, PM_NO_FLAGS)
			    != PWR_OP_OK) {
				log_msg = "Error powering up unallocated";
				DBG6(zprintf("%s: Error powering up "
				    "unallocated\n", __func__));
				rv = PS_FAILURE;
			}

			pmi_bundle_status(res_statusp, log_msg,
			    unalloc_count, &tot_res_status,
			    unalloc_residsp, NULL);

			/*
			 * Have to report to domain mgr about failed
			 * resources so that domain mgr will not
			 * allow user to add them via CLI.
			 */
			RES_POWER_FAULT_REPORT(&tot_res_status,
			    LDME_PE_REPORT_RES);
		}

		/* Power up allocated resources. */
		if (alloc_count > 0) {
			/*
			 * If ldmd is stopping then the drd service may already
			 * be down.  If so, then the DR attempt will fail and
			 * ldm_reconfigure_resources will try to back out the
			 * change in the domain MD.  This flag prevents that
			 * back out from occuring as the cpus have already been
			 * powered up.  If a reboot occurs after the back out
			 * then the cpus have already been powered up, and
			 * this can sometimes lead to a watchdog if the
			 * rebooting domain is the control domain and the
			 * chosen boot cpu is one of those which was backed
			 * out of the guest MD.  The watchdog occurs because
			 * the chosen boot cpu is not part of any MD whereas
			 * HV expects all powered up CPUs to be part of an MD.
			 * The handling is safe as ldmd is coming down also so
			 * that after the reboot the cpus will be DR'd back into
			 * the domain.
			 */
			if (pm_stop) {
				flags |= IGNORE_DR_ERRORS;
			}
			if (pmi_adjust_power(res_type, PCO_ADD_RESOURCE,
			    PPS_FULL_POWER_FORCE, alloc_count,
			    alloc_residsp, &res_statusp, flags) != PWR_OP_OK) {
				log_msg = "Error powering up allocated";
				DBG6(zprintf("%s: Error powering up "
				    "allocated\n", __func__));
				rv = PS_FAILURE;
			}

			pmi_bundle_status(res_statusp, log_msg,
			    alloc_count, &tot_res_status,
			    alloc_residsp, NULL);

			RES_POWER_FAULT_REPORT(&tot_res_status,
			    LDME_PE_REPORT_RES);
		}

		Xfree(alloc_residsp);
		Xfree(unalloc_residsp);
		Xfree(all_res_nodesp);
	}

	Xfree(tot_domid_listp);
	return (rv);
}


/*
 * pmi_register_for_notifications - Registers the PM Engine to receive
 * configuration change events and utilization threshold crossing events.
 *
 * Returns: PS_SUCCESS, PS_FAILURE
 */
int
pmi_register_for_notifications(void)
{
	int		rv = PS_SUCCESS;
	pm_res_type_t	res_type = PM_RES_CPU;
	int		type_idx;
	ldm_err_t	err = { 0 };
	double		pm_low_util_wm, pm_hi_util_wm;
	uint32_t	pm_time_low_util_ms, pm_time_hi_util_ms;

	/* Register to receive domain configuration change events. */
	pm_dom_cfg_hdl = REGISTER_DOMAIN_CONFIG_CHANGE_EVENT(&err);

	if (pm_dom_cfg_hdl == NULL) {
		DBG(ldm_log_msg("%s: Error registering for domain "
		    "configuration change events. (err = %d)\n", __func__,
		    err.ldme_errs));
		rv = PS_FAILURE;
	}

	/*
	 * Go through all resource types and register to receive resource
	 * configuration and utilization events.
	 */
	for (type_idx = 0; type_idx < PM_RES_TYPES_TOTAL;
	    type_idx++, res_type++) {
		if (res_type == PM_RES_CPU) {
			pm_low_util_wm = DEFAULT_CPU_LOW_UTIL_WM;
			pm_hi_util_wm = DEFAULT_CPU_HI_UTIL_WM;
			pm_time_low_util_ms = DEFAULT_CPU_TIME_LOW_UTIL_MS;
			pm_time_hi_util_ms = DEFAULT_CPU_TIME_HI_UTIL_MS;
		} else if (res_type == PM_RES_MEM) {
			if (!pmi_is_mem_supported()) {
				DBG(ldm_log_msg("%s: Memory PM is not "
				    "supported on this platform.  No "
				    "utilization event for memory is "
				    "registered.\n", __func__));
				continue;
			}
			pm_low_util_wm = DEFAULT_MEM_LOW_UTIL_WM;
			pm_hi_util_wm = DEFAULT_MEM_HI_UTIL_WM;
			pm_time_low_util_ms = DEFAULT_MEM_TIME_LOW_UTIL_MS;
			pm_time_hi_util_ms = DEFAULT_MEM_TIME_HI_UTIL_MS;
		} else {
			DBG(ldm_log_msg("%s: Attempt to register for "
			    "notifications using an invalid resource type.\n",
			    __func__));
			rv = PS_FAILURE;
		}

		pm_util_hdl[type_idx] =
		    REGISTER_UTIL_CHANGE_EVENT(res_type,
		    pm_low_util_wm, pm_hi_util_wm, pm_time_low_util_ms,
		    pm_time_hi_util_ms, &err);

		if (pm_util_hdl[type_idx] == NULL) {
			DBG(ldm_log_msg("%s: Error registering for utilization "
			    "change events. (err = %d) (type = %d)\n",
			    __func__, err.ldme_errs, res_type));
			rv = PS_FAILURE;
		}

		/* If resource is memory, no cfg & FMA events are registered. */
		if (res_type == PM_RES_MEM)
			continue;

		pm_res_cfg_hdl[type_idx] =
		    REGISTER_RESOURCE_CONFIG_CHANGE_EVENT(res_type, &err);

		if (pm_res_cfg_hdl[type_idx] == NULL) {
			DBG(ldm_log_msg("%s: Error registering for resource "
			    "configuration change events. (err = %d) "
			    "(type = %d)\n", __func__, err.ldme_errs,
			    res_type));
			rv = PS_FAILURE;
		}

		pm_fma_hdl[type_idx] = REGISTER_FMA_EVENT(res_type, &err);

		if (pm_fma_hdl[type_idx] == NULL) {
			DBG(ldm_log_msg("%s: Error registering for resource "
			    "retire and unretire events. (err = %d) "
			    "(type = %d)\n", __func__, err.ldme_errs,
			    res_type));
			rv = PS_FAILURE;
		}
	}

	return (rv);
}


/*
 * pmi_unregister_for_notifications - Unregisters the PM Engine from
 * receiving configuration change events and utilization threshold
 * crossing events. Does not unregister for guest domain DR
 * registration events or events sent when a guest domain stops running.
 *
 * Returns: PS_SUCCESS, PS_FAILURE
 */
int
pmi_unregister_for_notifications(void)
{
	int		rv = PS_SUCCESS;
	int		type_idx;
	ldm_err_t	err = { 0 };
	pm_res_type_t	res_type = PM_RES_CPU;

	if (pm_dom_cfg_hdl != NULL) {
		if (UNREGISTER_NON_UTIL_EVENTS(pm_dom_cfg_hdl,
		    &err) != PS_SUCCESS) {
			DBG(ldm_log_msg("%s: Unable to unregister from "
			    "receiving domain configuration notifications. "
			    "(err = %d)\n", __func__, err.ldme_errs));
			rv = PS_FAILURE;
		}
	}

	for (type_idx = 0; type_idx < PM_RES_TYPES_TOTAL;
	    type_idx++, res_type++) {
		/*
		 * Unregister to receive notifications for this resource
		 * type.
		 */
		if (pm_util_hdl[type_idx] != NULL) {
			if (UNREGISTER_UTIL_CHANGE_EVENT(pm_util_hdl[type_idx],
			    &err) != PS_SUCCESS) {
				DBG(ldm_log_msg("%s: Unable to unregister from "
				    "receiving utilization notifications. "
				    "(err = %d) (type = %d)\n", __func__,
				    err.ldme_errs, res_type));
				rv = PS_FAILURE;
			}
		}

		if (pm_res_cfg_hdl[type_idx] != NULL) {
			if (UNREGISTER_NON_UTIL_EVENTS(
			    pm_res_cfg_hdl[type_idx], &err) != PS_SUCCESS) {
				DBG(ldm_log_msg("%s: Unable to unregister from "
				    "receiving resource configuration "
				    "notifications. (err = %d) "
				    "(type = %d)\n", __func__,
				    err.ldme_errs, res_type));
				rv = PS_FAILURE;
			}
		}

		if (pm_fma_hdl[type_idx] != NULL) {
			if (UNREGISTER_NON_UTIL_EVENTS(
			    pm_fma_hdl[type_idx], &err) != PS_SUCCESS) {
				DBG(ldm_log_msg("%s: Unable to unregister from "
				    "receiving resource retire and unretire "
				    "notifications. (err = %d) "
				    "(type = %d)\n", __func__,
				    err.ldme_errs, res_type));
				rv = PS_FAILURE;
			}
		}
	}

	return (rv);
}


/*
 * pmi_same_res - Compares two resource arrays and finds the
 * resources that are the same.
 *
 * Returns: true, false
 */
int
pmi_same_res(pm_resid_t *resap, pm_resid_t *resbp)
{
	if ((resap == NULL) || (resbp == NULL)) {
		return (false);
	} else {
		/*
		 * Make sure that we are comparing resources of the same
		 * type.
		 */
		ASSERT(resap->pi_res_type == resbp->pi_res_type);
		switch (resap->pi_res_type) {
		case PM_RES_CPU:
			return (resap->u.cpu.pi_id_val ==
			    resbp->u.cpu.pi_id_val);

		default:
			DBG(ldm_log_msg("%s: Invalid resource type "
			    "encountered.\n", __func__));
			ASSERT(0);
		}
	}

	return (false);
}


/*
 * pmi_gather_allocated_resources - Cycles through all domains and
 * marks the allocated resources in the all_res_nodesp array by
 * setting the resource pointers to NULL.  If the caller sends in
 * an allocated alloc_residsp, an array will be created that contains
 * resource IDs for all allocated resources.  If the caller sends in
 * an allocated alloc_nodesp, an array will be created that contains
 * PM nodes for all allocated resources.
 */
void
pmi_gather_allocated_resources(int num_domains, srcid_t *tot_domid_listp,
    int *alloc_count, pm_res_type_t res_type, pm_node_t *all_res_nodesp,
    pm_resid_t *alloc_residsp, pm_node_t *alloc_nodesp, int tot_res_count)
{
	int		dom_idx;
	int		dom_res_idx;
	int		all_idx;
	pm_resid_t	*res_ptr;
	pm_node_t	*node_ptr;
	ldm_err_t	err = { 0 };
	int		dom_res_count = 0;
	pm_resid_t	*dom_residsp = NULL;

	res_ptr = alloc_residsp;
	node_ptr = alloc_nodesp;

	for (dom_idx = 0; dom_idx < num_domains; dom_idx++) {
		if (GET_DOMAIN_CONFIG(tot_domid_listp[dom_idx],
		    res_type, &dom_residsp, &dom_res_count,
		    &err) != PS_SUCCESS) {
			DBG(ldm_log_msg("%s: Unable to obtain domain "
			    "%s config. (err = %d)\n", __func__,
			    GET_DOMAIN_NAME(tot_domid_listp[dom_idx]),
			    err.ldme_errs));
			continue;
		}

		for (dom_res_idx = 0; dom_res_idx < dom_res_count;
		    dom_res_idx++) {
			for (all_idx = 0; all_idx < tot_res_count;
			    all_idx++) {
				if (pmi_same_res(&dom_residsp[dom_res_idx],
				    all_res_nodesp[all_idx].pn_rcfg)) {
					/* Only create arrays if needed */
					if (alloc_nodesp != NULL) {
						*node_ptr++ =
						    all_res_nodesp[all_idx];
					}
					if (alloc_residsp != NULL) {
						*res_ptr++ =
						    *all_res_nodesp[all_idx].\
						    pn_rcfg;
					}
					all_res_nodesp[all_idx].pn_rcfg
					    = NULL;
					(*alloc_count)++;
					break;
				}
			}
		}

		Xfree(dom_residsp);
	}
}


/*
 * pmi_gather_unallocated_resources - Checks the all_res_nodesp array
 * for pointers that were set to NULL by pmi_gather_allocated_resources().
 * Takes the resources != NULL and creates an array that contains
 * resource IDs for all unallocated resources.
 */
void
pmi_gather_unallocated_resources(pm_resid_t *unalloc_residsp, int tot_res_count,
    int unalloc_count, pm_node_t *all_res_nodesp)
{
	pm_resid_t	*res_ptr;
	int		all_idx;
	int		unalloc_idx;

	res_ptr = unalloc_residsp;

	for (all_idx = 0, unalloc_idx = 0; (all_idx < tot_res_count) &&
	    (unalloc_idx < unalloc_count); all_idx++) {
		if (all_res_nodesp[all_idx].pn_rcfg != NULL) {
			*res_ptr++ =
			    *all_res_nodesp[all_idx].pn_rcfg;
			unalloc_idx++;
		}
	}
}


/*
 * pmi_bundle_status - Gathers up status info per resource info to report
 * to the domain mgr.
 */
void
pmi_bundle_status(int *res_statusp, char *log_msg, int res_array_count,
    ldm_err_t *bundle_res_statusp, pm_resid_t *residsp, int *err_cnt)
{
	int			res_idx;
	uint64_t		id_val;
	ldme_res_status_t	*temp_bundle_statusp;

	if ((res_array_count <= 0) || (res_statusp == NULL))
		return;

	bundle_res_statusp->ldme_private = Xcalloc(1, ldme_res_status_t);
	temp_bundle_statusp =
	    (ldme_res_status_t *)(bundle_res_statusp->ldme_private);
	temp_bundle_statusp->lrs = Xcalloc(res_array_count, ldme_res_sane_t);

	bundle_res_statusp->ldme_errs = LDME_OP_EOK;
	temp_bundle_statusp->rescnt = res_array_count;

	for (res_idx = 0; res_idx < res_array_count; res_idx++) {
		if (res_statusp[res_idx] != LDME_RES_GOOD) {
			bundle_res_statusp->ldme_errs = LDME_OP_FAILURE;
			if (residsp[res_idx].pi_res_type == PM_RES_CPU)
				id_val = residsp[res_idx].u.cpu.pi_id_val;
			else if (residsp[res_idx].pi_res_type == PM_RES_MEM)
				id_val = residsp[res_idx].u.mem.pi_id_val;
			else
				id_val = -1;

			DBG(ldm_log_msg("%s: %s %s %"PRIu64" (err = %d).\n",
			    __func__, log_msg, residsp[res_idx].pi_res_name,
			    id_val, res_statusp[res_idx]));

			/*
			 * NeedOne cpus are not errors so must be special cased
			 * A NeedOne cpu is a cpu which must always be powered
			 * up due to requiring one cpu up per CPU chip to handle
			 * interrupts in the HV.
			 */
			if (res_statusp[res_idx] == LDME_RES_STUCK_HIGH_REQD)
				continue;

			if (err_cnt != NULL) {
				(*err_cnt)++;
			}
		}

		temp_bundle_statusp->lrs[res_idx].sanity = res_statusp[res_idx];
		temp_bundle_statusp->lrs[res_idx].resid = &residsp[res_idx];
	}

	Xfree(res_statusp);
}

/*
 * pmi_free_domain_utils
 *
 * Frees all data associated with domain utilization.
 */
void
pmi_free_domain_utils(ldm_resutils_t *res_utilsp)
{
	int	res_idx;

	for (res_idx = 0; res_idx < res_utilsp->ldmr_rescnt; res_idx++) {
		Xfree(res_utilsp->ldmr_util_p[res_idx].\
		    ldmu_resid.pi_res_name);
	}
	Xfree(res_utilsp->ldmr_util_p);
	Xfree(res_utilsp);
}

/*
 * pmi_free_active_list_utils
 *
 * This function frees all data associated with domain utilization
 * in the Active Domain list.
 *
 * Input:
 *	listp - Pointer to the active domain list
 */
void
pmi_free_active_list_utils(pm_hps_domain_t *listp)
{
	pm_hps_domain_t	*ptr;

	/* Walk the active domain list */
	for (ptr = listp; ptr; ptr = ptr->next) {
		pmi_free_domain_utils(ptr->dom_utilsp);
	}
}

/*
 * pmi_process_hps
 *
 * Called by the PM engine for utilization events, to adjust the HPS
 * nodes Pwr states and Domain strands to maintain utilization within band.
 * This is the main processing function from which all other supporting
 * functions are called.
 *
 * Input:
 *   notifyp: A pointer to a pm_notify_t structure
 * Output:
 *   PS_FAILURE: HPS not applicable for this system
 *   PS_SUCCESS: HPS is applicable for this system
 */
pwr_req_status_t
pmi_process_hps(pmi_notify_t *notifyp)
{
	int		num_active;
	int		rtn = PS_SUCCESS;	/* Return value for function */

	/*
	 * Get a list of all the Active Domains.  The pointer to the list
	 * is kept in global active_dom_list.
	 */
	rtn = pmi_create_active_domain_list(&num_active);
	if (rtn != PS_SUCCESS) {
		goto dc_done;
	}

	/*
	 * Create the HPS Domain list.  The pointer to the list is kept
	 * in global hps_dom_list.
	 */
	rtn = pmi_create_hps_domain_list(notifyp, num_active);
	if (rtn != PS_SUCCESS) {
		goto dc_done;
	}

	/* Gather Domain strand stats for use in HPS/Strand calculations */
	pmi_collect_stats();

	/* Process the HPS Domain list */
	rtn = pmi_process_hpsdl(notifyp);

dc_done:

	/* Free HPS node vector structs and pointer array */
	pmi_free_hps();
	return (rtn);
}

/*
 * pmi_free_hps
 *
 * Free memory for the global data structures used for HPS.
 *
 * Input: None
 * Output: None
 */
void
pmi_free_hps()
{
	/*
	 * Free the Active List along with Domain Utils
	 * and the HPS node vectors
	 */
	if (active_dom_list != NULL) {
		int i;
		pm_hps_domain_t	*nxt_node;
		pm_node_vec_t	*nxt_vec, *tmp_vec;

		/* Free the Domain Utils */
		pmi_free_active_list_utils(active_dom_list);

		/* Walk Active List to free node vectors */
		for (nxt_node = active_dom_list; nxt_node;
		    nxt_node = nxt_node->next) {

			/* Walk node_vec[] array */
			for (i = 0; i < nxt_node->num_lvls; i++) {

				/* Walk node_vec[x] list */
				nxt_vec = nxt_node->node_vec[i];
				while (nxt_vec != NULL) {
					tmp_vec = nxt_vec->next;
					/* Free pm_node_vec_t */
					Xfree(nxt_vec);
					nxt_vec = tmp_vec;
				}
			}

			/* Free array of node vector pointers */
			Xfree(nxt_node->node_vec);
		}

		/* Free the Active Domain list */
		Xfree(active_dom_list);
		active_dom_list = NULL;
	}

	/* Free the HDL - HPS Domain List */
	if (hps_dom_list != NULL) {
		Xfree(hps_dom_list);
		hps_dom_list = NULL;
		hps_dom_idx = 0;
	}
}

/*
 * pmi_calc_hps_strands
 *
 * This function performs the HPS node Pwr state and strand calculation to bring
 * a dependent Domain's utilization inband. No Pwr state or strand adjustments
 * are made, only the calculation is done.
 *
 * Input:
 *   hdl_idx: HPS Dom list index for domain to process
 *   pms_idx: Next PM state (Pwr lvl) to try calculation with
 *   ances_cur_ratio: Current multiplicative power ratio of ancestor HPS nodes
 *   ances_best_ratio: Next multaplicative power ratio of ancestor HPS nodes
 *   pmnp: Ptr to HPS PM node for the event domain for current level
 *
 * Output:
 *   hps_dom_list[nxt_idx]->strands_new: New # of strands needed
 *   hps_dom_list[nxt_idx]->new_state: New PM state for HPS node
 *   hps_dom_list[nxt_idx]->direction: Indicates to add or remove strands
 *   hps_dom_list[nxt_idx]->strands_adjust: # of strands to add/remove
 *   Function return:
 *       PS_SUCCESS: Calculation results valid
 *       PS_FAILURE: # new strands > # strands avail in Domain
 */
int
pmi_calc_hps_strands(int hdl_idx, int pms_idx,
    double ances_cur_ratio, double ances_best_ratio,
    pm_node_t *pmnp)
{
	ldm_resutils_t	*utilsp;
	int		adjust_strands;
	int		max_strands, new_strands;
	int		off_strands, on_strands;
	int		rtn = PS_SUCCESS;
	double 		load, median;
	double 		cur_ratio, best_ratio;

	/* Data needed for the strand calculation */
	utilsp = hps_dom_list[hdl_idx]->dom_utilsp;
	on_strands = hps_dom_list[hdl_idx]->strands_on;
	off_strands = hps_dom_list[hdl_idx]->strands_off;
	max_strands = (int)utilsp->ldmr_rescnt;
	load = utilsp->ldmr_avg_util;
	median = MEDIAN_CPU;
	cur_ratio = pmnp->pn_pwrs->pps_ratio[pmnp->pn_pwrs->pps_curlvl];
	cur_ratio *= ances_cur_ratio;
	best_ratio = pmnp->pn_pwrs->pps_ratio[pms_idx];
	best_ratio *= ances_best_ratio;

	/* ********  Here's the Calculation!!!! ******* */
	new_strands = (int)((cur_ratio * on_strands * load) /
	    (best_ratio * median));

	/* Total # of strands needed */
	new_strands = new_strands + UTIL_CALC_ADJUST;

	hps_dom_list[hdl_idx]->strands_new = new_strands;

	/* Calculate the # of strands to add/remove */

	/* Strands will stay the same */
	if (on_strands == new_strands) {
		/*
		 * The Domain can be above the Hi WM if
		 * all its strands are powered on and the
		 * Node is at the highest Pwr state, which
		 * this algorithm will ensure.  Domain
		 * has no strands to power on.
		 */
		adjust_strands = on_strands - new_strands;
		hps_dom_list[hdl_idx]->direction = PM_UTIL_WITHIN;

	/* Too many strands, remove some */
	} else if (on_strands > new_strands) {
		adjust_strands = on_strands - new_strands;

		/* Reset strands so the minimum strands per domain is met */
		if (new_strands < option.min_vcpu_per_ldom) {
			adjust_strands = (on_strands -
			    option.min_vcpu_per_ldom);
		}
		hps_dom_list[hdl_idx]->direction = PM_UTIL_LOW;

	/* Not enough strands, add some */
	} else if (on_strands < new_strands) {

		adjust_strands = new_strands - on_strands;
		if (adjust_strands > off_strands)
			adjust_strands = off_strands;

		hps_dom_list[hdl_idx]->direction = PM_UTIL_HIGH;

		/*
		 * If a Domain's Avg Util is above the
		 * Hi WM, and has a strand at 0%, then
		 * we don't add any strands.
		 *
		 * A Domain may have no more than one 0% strand.
		 * They just consume power and don't do any work.
		 *
		 * A Domain may not have any 0% strands if there
		 * are strands with less than 50% utilization.
		 */
		if (((load > DEFAULT_CPU_HI_UTIL_WM) &&
		    (hps_dom_list[hdl_idx]->zero_pcent > 0)) ||
		    (hps_dom_list[hdl_idx]->zero_pcent > 1) ||
		    ((hps_dom_list[hdl_idx]->blo_lowm_pcent > 0) &&
		    (hps_dom_list[hdl_idx]->zero_pcent > 0))) {

			adjust_strands = 0;
			hps_dom_list[hdl_idx]->direction = PM_UTIL_WITHIN;

			/*
			 * Have to adjust new_strands because of this
			 * exception.  IOW, make this Domain appear
			 * to be OK for now wrt # of strands it has
			 * powered on.  There could be any number of issues
			 * why one of these exceptions occurred, but it
			 * is not this codes duty to fix that.
			 */

			/* Don't want to return a retry signal */
			new_strands = max_strands;
			DBG6(zprintf("%s: %s: Exception Adding strands: "
			    "zero_pcent=%d, blo_lowm_pcent=%d, "
			    "load=%5.2f%%, hiwm=%d%%\n", __func__,
			    GET_DOMAIN_NAME(hps_dom_list[hdl_idx]->domain),
			    hps_dom_list[hdl_idx]->zero_pcent,
			    hps_dom_list[hdl_idx]->blo_lowm_pcent,
			    load, DEFAULT_CPU_HI_UTIL_WM));
		}
	}
	hps_dom_list[hdl_idx]->strands_adjust = adjust_strands;

	/*
	 * More strands are needed than what the Domain has available, so
	 * increase the node Pwr state again.  Once the max Pwr state is reached
	 * and there still is not enough strands a Domain has available, then
	 * there is nothing that can be done for that Domain short of
	 * adding more vcpus.
	 */
	if (new_strands > max_strands)
		rtn = PS_FAILURE;

	return (rtn);
}

/*
 * pmi_recurse_calc_hps_strands
 *
 * Recursively calculate the best PM power state for Higher Power
 * Saving nodes along with the number of strands needed in each
 * affected Domain wrt to the HPS nodes.  The calculation tries to
 * keep each Domain Util in-band while lowering power on the nodes
 * that give the greatest power savings.
 *
 * The strategy is to start at HPS node level 0 and come up with
 * a Pwr state, assuming all higher HPS levels will be at their
 * highest Pwr state, to have all affected Domain Util remain
 * in-band.  This is done because the lower the level, the more
 * Power savings realized.  After level 0 has been calculated,
 * then do the same for level 1..n.  This is done recursively.
 *
 * When we percolate back up, the Pwr state will have been
 * recorded for each level of HPS nodes and the strands for
 * each domain calculated, to keep all affected domains in-band.
 *
 * Input:
 *   hdlp: Pointer to the current event domain on the HDL to process
 *   lvl:  The current level of the node vector being processed.
 *
 * Output: hdlp elements
 */
void
pmi_recurse_calc_hps_strands(pm_hps_domain_t *hdlp, int lvl)
{
	pm_node_vec_t	*hvp;			/* HPS vec prt */
	int		end_pms_idx, pms_idx;
	int		nxt_idx, start_idx;
	int		retry;
	double		ances_cur_ratio;
	double		ances_best_ratio;

	/* num_lvls is 1 based, lvl is 0 based */
	if (lvl >= hdlp->num_lvls)
		return;

	/* Max PM state index for HPS nodes at this level */
	end_pms_idx = hdlp->node_vec[lvl]->pmnode->pn_pwrs->pps_count - 1;

	/* Level 0 values; Non level 0 get calcualted from these */
	ances_cur_ratio = 1.0;
	ances_best_ratio = 1.0;

	/*
	 * If not at level 0, then get the multaplicative power ratio of
	 * all the ancestors for their current states and for their
	 * calculated (best) new states.  These ratios are needed in
	 * the strand calculation.
	 */
	if (lvl > 0) {
		int tmp_lvl = lvl - 1;
		while (tmp_lvl >= 0) {
			int idx;

			idx =
			    hdlp->node_vec[tmp_lvl]->pmnode->pn_pwrs->\
			    pps_curlvl;
			ances_cur_ratio *= hdlp->node_vec[tmp_lvl]->pmnode->\
			    pn_pwrs->pps_ratio[idx];
			ances_best_ratio *= hdlp->node_vec[tmp_lvl]->new_ratio;
			tmp_lvl--;
		}
	}

	/*
	 * Start at the lowest Pwr state for the current HPS level.  All
	 * higher levels had their best ratio initialized to 1.0.
	 * retry==1 means that the current Pwr state (pms_idx) was too
	 * low for at least 1 Domain wrt to the number of strands
	 * needed in the Domain to stay in-band.  So we try the next
	 * Pwr state (pms_idx)
	 */

	/*
	 * Exception: If the event domain has one or more strands at 100%,
	 * and we are processing level 0 (which has the greast power savings
	 * capability), then start the Pwr state at 1 more than the current
	 * Pwr state for the HPS node.  This is because adding more strands
	 * will not help this case.  A thread executing on a strand at 100%
	 * cannot be scheduled on more than 1 strand, thus increasing the
	 * Node power state is the only way to try an get less than 100%
	 * util for that thread.
	 */
	if ((lvl == 0) && (hdlp->hundred_pcent > 0))
		start_idx = hdlp->node_vec[0]->pmnode->pn_pwrs->pps_curlvl + 1;
	else
		start_idx = 0;

	retry = 1;
	for (pms_idx = start_idx; ((pms_idx <= end_pms_idx) && retry);
	    pms_idx++) {

		/*
		 * Walk the HDL to find Domains that are affected by
		 * changing the Pwr state of the HPS node that the event
		 * Domain is dependent on.
		 */
		for (nxt_idx = 0; nxt_idx < hps_dom_idx; nxt_idx++) {

			/* Next Domain dependent on event Domain? */
			if (pmi_hps_node_match(hdlp, hps_dom_list[nxt_idx],
			    lvl) != PS_SUCCESS)
				continue;

			hps_dom_list[nxt_idx]->processed = true;

			/* Calcualte the # of strands for the current domain */
			retry = pmi_calc_hps_strands(
			    nxt_idx, pms_idx, ances_cur_ratio,
			    ances_best_ratio, hdlp->node_vec[lvl]->pmnode);

			/*
			 * The Domain just checked does not have enough
			 * strands for the current Pwr state, so try
			 * the next higher Pwr state
			 */
			if (retry) {
				break;
			}
		}
	}

	/*
	 * Could not find a Pwr state which satisfies all domains wrt the
	 * the number of available strands they have to power on. Thus
	 * we can only set the highest Pwr state and turn on whatever
	 * strands were calculated for each domain.
	 */
	DBG_BLOCK {
		if (retry) {
			DBG6(zprintf("%s: %s: Cannot get a PState so that"
			    " all domains use <= their Max strands\n",
			    __func__, GET_DOMAIN_NAME(hdlp->domain)));
		}
	}

	/* Need to reduce pms_idx by 1 because of the for loop */
	pms_idx--;

	/*
	 * Update the new PM state and best Pwr ratio for all HPS vectors
	 * at this level.
	 */
	for (nxt_idx = 0; nxt_idx < hps_dom_idx; nxt_idx++) {
		if (pmi_hps_node_match(hdlp, hps_dom_list[nxt_idx],
		    lvl) != PS_SUCCESS)
			continue;

		for (hvp = hps_dom_list[nxt_idx]->node_vec[lvl]; hvp;
		    hvp = hvp->next) {
			hvp->new_state = pms_idx;
			hvp->new_ratio =
			    hvp->pmnode->pn_pwrs->pps_ratio[pms_idx];
		}
	}

	/* Recurse next level and calcualte the best power state/strands */
	pmi_recurse_calc_hps_strands(hdlp, lvl + 1);
}

/*
 * pmi_create_active_domain_list
 *
 * This function creates a list of pm_hps_domain_t structures representing all
 * the ACTIVE domains in the system. This list is the Active Domain List.
 *
 * Input:
 *   num_active - Pointer to the count of Active Domains
 *
 * Output:
 *   *num_active: The count of Active Domains
 *   active_dom_list: Global pointer for the list
 *     NULL - No list created
 *     Non Null - The head of the list
 *   Function return
 *     PS_SUCCESS: List created
 *     PS_FAILURE: List not created
 *
 * Note: The caller must ensure all allocated memory is freed.
 */
pwr_req_status_t
pmi_create_active_domain_list(int *num_domains)
{
	ldm_err_t		ldm_err;
	pm_hps_domain_t		*active_domp = NULL;
	pm_hps_domain_t		*next_active_domp = NULL;
	srcid_t			*hps_act_domid_listp = NULL;
	int			active_list_id = 0;
	int			i, rc;		/* Counters, rtn code */
	uint_t			num_doms = 0;

	*num_domains = 0;
	/*
	 * Get all the Active domains on this system.  This is the starting
	 * point in order to determine how many HPS nodes are involved for
	 * a power adjustment.
	 */
	GET_ALL_DOMAIN_IDS(&hps_act_domid_listp, &num_doms,
	    &ldm_err, ACTIVE_DOMAINS);

	/*
	 * Allocate the memory needed for the Active Domain list.
	 * The list is used later in building an overall HPS vector
	 * that indicates which nodes are affected by a HPS node Pwr
	 * change for the current event Domain.
	 */
	if (active_dom_list == NULL) {
		active_dom_list = Xcalloc(num_doms, pm_hps_domain_t);
		active_domp = active_dom_list;
		next_active_domp = active_domp;
	}

	/*
	 * Create a linked list of all the Active domains.  Also create a HPS
	 * node vector for each domain, that indicates which HPS nodes the
	 * domain's strands are on. This list will be used later in building an
	 * overall vector that will indicate which HPS nodes are affected by
	 * an HPS node Pwr change.
	 */
	for (i = 0; i < num_doms; i++) {
		(*num_domains)++;

		/* Save the domain pointer */
		next_active_domp->domain = hps_act_domid_listp[i];

		/* Create the HPS node vec for nodes the dom has strands on */
		rc = pmi_create_hps_node_vector(next_active_domp);

		/* Can't do HPS Pwr change, so do only CoreDisable */
		if (rc != PS_SUCCESS) {
			DBG(zprintf("%s: Failure from "
			    "pmi_create_hps_node_vector()\n", __func__));
			Xfree(hps_act_domid_listp);
			return (PS_FAILURE);
		}

		/*
		 * If this is not the first Active List node, then link the
		 * previous node to this new one, then advnace the new ptr to
		 * the next node to be filled in.  Also set the id. This is used
		 * later when trying to match a node between the Active and HPS
		 * Domain (explained later) lists.
		 */
		next_active_domp->id = active_list_id++;
		if (next_active_domp != active_dom_list) {
			active_domp->next = next_active_domp;
			active_domp = next_active_domp++;
			active_domp->next = NULL;

		/*
		 * This is the first Active List node, so there is no linking
		 * of nodes to do yet.  This else case will be hit once.
		 */
		} else {
			active_domp->next = NULL;
			next_active_domp++;
		}
	}

	Xfree(hps_act_domid_listp);
	return (PS_SUCCESS);
}

/*
 * pmi_create_hps_domain_list
 *
 * This function creates an array of pm_hps_domain_t pointers called the HPS
 * Domain List, HDL, whose entries are pointers into the Active Domain List.
 * This list will contain only the Domains to be considered for strand/HPS node
 * adjustments. The HDL list is created wrt to the input event Domain, and
 * will include all dependent domains.  A NULL is also allowed for the input
 * event domain, and in this case, all Active domains are considered event
 * domains.  This latter case is used to recalcualte HPS node pwr and domain
 * strands for all domains.
 *
 * Input: None
 *   num_domains: The number of Active Domains
 *   notifyp: Pointer to the event notification structure
 *            NULL - Use all Active Domains
 *
 * Output:
 *   Function return
 *     PS_SUCCESS: HDL created
 *     PS_FAILURE: HDL not created
 */
pwr_req_status_t
pmi_create_hps_domain_list(pmi_notify_t *notifyp, int num_domains)
{

	bool_t		found = false;
	pm_hps_domain_t	*next_active_domp = NULL;
	int		base_idx, nxt_idx;	/* Array indicies */
	int		i, j;			/* Counters */
	int		dom_count;
	int		want_all_active = 0;	/* Which Doms to use */

	/*
	 * We now have a list of Active domains with HPS vector(s) indicating
	 * which HPS nodes each domain has strands on.  We have to build another
	 * list, a HPS Domain list, that initially has the domains from the
	 * event, then add any domain from the Active list in which the
	 * HPS vectors match at least one HPS node.  The end result is a list
	 * which contains domains that will be affected by changing the
	 * Pwr state on one or more HPS nodes for one of the event Domains.
	 *
	 * First put the domains from the event on the HPS domain list or
	 * put all the Active domains on the HPS domain list.  The latter
	 * is done when we want to re-calculate the strands for all Domains.
	 * The HPS Domain list is just an array of pointers into the Active
	 * list.
	 */
	hps_dom_idx = 0;
	hps_dom_list = (pm_hps_domain_t **)
	    Xcalloc(num_domains, pm_hps_domain_t *);

	if (notifyp == NULL) {
		want_all_active = 1;
		dom_count = num_domains;
	} else
		dom_count = notifyp->pn_count;

	/*
	 * Construct the initial HDL with the event domains or all the
	 * Active domains.
	 */
	for (i = 0; i < dom_count; i++) {

		/*
		 * Link the next index in the HDL to the corresponding node in
		 * the Active Domain list. That way we don't have to copy data.
		 * The HDL list is the list to be used later on in determining
		 * what resources get adjusted.
		 */
		found = false;
		for (next_active_domp = active_dom_list; next_active_domp;
		    next_active_domp = next_active_domp->next) {

			if (!want_all_active) {
				/*
				 * Compare the LDom name for the current event
				 * domain to the LDom names on the Active List.
				 * When matched then link the HDL entry to the
				 * Active List node.
				 */
				if (strcmp(
				    GET_DOMAIN_NAME(notifyp->pn_src_ids[i]),
				    GET_DOMAIN_NAME(next_active_domp->domain))
				    == 0) {

					/* See if Utils are still OOB */
					if (!PM_UTIL_IS_VALID(
					    next_active_domp->dom_utilsp,
					    DEFAULT_CPU_LOW_UTIL_WM,
					    DEFAULT_CPU_HI_UTIL_WM)) {
						continue;
					}
				} else
					continue;
			}

			/*
			 * This code only get executed when we add a new entry
			 * to the HDL. The above code controls if we get here.
			 * If we want all Active Domains, then we will always
			 * get here.
			 */
			hps_dom_list[hps_dom_idx++] =
			    next_active_domp;

			/* Indicate this is an event domain */
			next_active_domp->event_dom = true;
			found = true;

			/*
			 * If we don't want all the Active domains, then break
			 * out back to the loop to get the next event domain.
			 */
			if (!want_all_active)
				break;
		}
		if (!found) {
			DBG6(zprintf("%s: Did not find Event Domain"
			    " on Active list\n", __func__));
		}

		/*
		 * All the Active domains have been added to the HDL, so set
		 * the outer loop to terminate.
		 */
		if (want_all_active)
			i = dom_count;
	}

	/*
	 * The HDL now only contains the domains from the event or all active
	 * domains.  The next procedure is to walk this list, and for each
	 * node, scan the Active List to see if there is a HPS node match. Any
	 * Active List node that has a HPS node match will be added to the
	 * HDL if not already on it.  This is done only if the whole Active
	 * list is not asked for.
	 *
	 * Example:
	 * Domains A & B are in a Util event.
	 * Domains A,B,C,D,E are active.
	 * Domain A has strands on Node0, B on Node1, C on Nodes1,3 D on Node2
	 * E on Node1
	 * A & B are put on the HDL
	 *
	 * Start with A, the active list is walked and the HPS vector from A is
	 * compared to the vector for all the active domains. If there is a
	 * match of at least 1 Node, then the active domain is placed on the
	 * end of the HDL if not already on it.
	 *
	 * So the progression is:
	 * Start: A & B on list
	 * Process A: No matches
	 * Process B: matches C:1, E:1; Add C,E to list: A,B,C,E
	 * Process C: mathces B,C,E - All on list
	 * Process E: mathces B,C,E - All on list
	 *
	 * So Nodes 0,1,3 and domains A,B,C,E are affected for a HPS node
	 * Pwr change.
	 */

	/* Walk the HPS Domain list one time */
	for (base_idx = 0; base_idx < hps_dom_idx; base_idx++) {
		bool_t			on_list;
		pm_hps_domain_t 	*activep;

		/* All Active domains are already on the HDL */
		if (want_all_active)
			break;

		/* Have to look at all nodes on the active list */
		for (activep = active_dom_list; activep;
		    activep = activep->next) {

			/*
			 * We just need to match one node from each list
			 * to declare a dependency.  The node match can
			 * be at any level.
			 */
			if (pmi_hps_node_match(hps_dom_list[base_idx],
			    activep, -1) == PS_SUCCESS) {

				/*
				 * Have to check if active node is already on
				 * the HDL.  If not, append it.
				 */
				on_list = false;
				for (j = 0; j < hps_dom_idx; j++) {
					if (hps_dom_list[j]->id ==
					    activep->id) {
						on_list = true;
						break;
					}
				}

				if (on_list) {
					continue;
				}

				/* Add active node to chip domain list */
				hps_dom_list[hps_dom_idx++] = activep;
			}
		}
	}

	/*
	 * Right now, each HPS Domain node has a HPS node vector that tells us
	 * which HPS nodes this Domain has strand on.  What is needed, is to
	 * know all the nodes that would be affected if the nodes for this
	 * Domain have their Pwr state changed.
	 *
	 * For example: Domain A has strands on nodes 0,2.  If we
	 * need to change Pwr on nodes 0,2, then we may have to also
	 * change the Pwr on other nodes because Domain B could
	 * have strands on nodes 1,2, thus node 1 needs the same
	 * Pwr state as nodes 0,2.
	 */
	/* We need two passes to get the dependent vectors set correctly */
	for (i = 0; i < 2; i++) {
		for (base_idx = 0; base_idx < hps_dom_idx; base_idx++) {
			for (nxt_idx = 0; nxt_idx < hps_dom_idx; nxt_idx++) {
				pmi_hps_set_node_dependencies(
				    hps_dom_list[base_idx],
				    hps_dom_list[nxt_idx]);
			}
		}
	}

	/* Get the number of HPS nodes for each HDL entry. */
	for (i = 0; i < hps_dom_idx; i++) {
		pm_node_vec_t *vecp;
		int num_nodes = 0;
		for (j = 0; j < hps_dom_list[i]->num_lvls; j++) {
			for (vecp = hps_dom_list[i]->node_vec[j]; vecp;
			    vecp = vecp->next) {
				num_nodes++;
			}
		}
		hps_dom_list[i]->num_nodes = num_nodes;
	}

	return (PS_SUCCESS);
}

/*
 * pmi_collect_stats
 *
 * This function collects several different stats about the cpu resources for
 * each Domain that is on the HDL. These stats will be used in determining if
 * strands can be adjusted for that Domain.
 *
 * Input: None
 *   hps_dom_list: Global HPS Domain List
 *   hps_dom_idx: Global max number of Domains on the HDL
 *
 * Output:
 *   Each CDL entry is updated with strand statistic info
 *
 */
void
pmi_collect_stats()
{
	ldm_resutils_t		*utilsp;
	pm_node_t		*pmnp;
	bool_t			is_drd_out = false;
	int			base_idx;
	int			res;

	/*
	 * Now the HDL contains pointers into the Active list, for all domains
	 * that could be affected by an HPS node Pwr change. Gather a bunch of
	 * data about the domain strands.  This will be used when trying to
	 * determine Pwr state and strand adjustment to keep a domain within
	 * a utilization band.
	 */

	/* Loop thru each Domain on the HDL */
	for (base_idx = 0; base_idx < hps_dom_idx; base_idx++) {

		/* Look at each cpu resource in the Domain */
		utilsp = hps_dom_list[base_idx]->dom_utilsp;
		for (res = 0; res < utilsp->ldmr_rescnt; res++) {

			/* Need the PM node to determine fault status */
			pmnp = epm_find_node(
			    &(utilsp->ldmr_util_p[res].ldmu_resid),
			    NULL, NULL);

			/* This case should not happen */
			if (pmnp == NULL) {
				DBG6(zprintf("%s: pm node NULL\n",
				    __func__));
				continue;
			}

			/* Faulted strands. Increment a count. */
			if (pmi_is_resource_faulted(pmnp->pn_rcfg)) {
				DBG6(zprintf("%s: strand %"PRIu64"  Faulted\n",
				    __func__,
				    pmnp->pn_rcfg->u.cpu.pi_id_val));
				hps_dom_list[base_idx]->faulted++;
				continue;
			}

			/*
			 * Don't count strands that don't have PERF or are
			 * dr'd out
			 */
			if ((!epm_has_perf(pmnp, PM_ASSUME_LOW)) ||
			    (!((GET_RESOURCE_PM_STATUS(pmnp->pn_rcfg,
			    PM_DR_STATUS, &is_drd_out) == 0) &&
			    is_drd_out == false))) {
				hps_dom_list[base_idx]->strands_off++;
				continue;
			}

			/* Strand has PERF & not dr'd out,  Increment a count */
			hps_dom_list[base_idx]->strands_on++;

			/* Strand at 100 %.  Increment a count */
			if (utilsp->ldmr_util_p[res].ldmu_util ==
			    (double)HUNDRED_PERCENT) {
				hps_dom_list[base_idx]->hundred_pcent++;
			}

			/* Strand at 0 %.  Increment a count */
			else if (utilsp->ldmr_util_p[res].ldmu_util ==
			    (double)0) {
				hps_dom_list[base_idx]->zero_pcent++;
			}

			/* Strand < pm_low_util_wm %.  Increment a count */
			else if ((utilsp->ldmr_util_p[res].ldmu_util <
			    DEFAULT_CPU_LOW_UTIL_WM) &&
			    (utilsp->ldmr_util_p[res].ldmu_util > 0)) {
				hps_dom_list[base_idx]->blo_lowm_pcent++;
			}

			/* Strand > pm_hi_util_wm %.  Increment a count */
			else if ((utilsp->ldmr_util_p[res].ldmu_util >
			    DEFAULT_CPU_HI_UTIL_WM) &&
			    (utilsp->ldmr_util_p[res].ldmu_util <
			    (double)HUNDRED_PERCENT)) {
				hps_dom_list[base_idx]->abv_hiwm_pcent++;
			}

			/* Strand is within the hi/lo wm band %.  Inc a count */
			else if ((utilsp->ldmr_util_p[res].ldmu_util <=
			    (double)DEFAULT_CPU_HI_UTIL_WM) &&
			    (utilsp->ldmr_util_p[res].ldmu_util >=
			    DEFAULT_CPU_LOW_UTIL_WM)) {
				hps_dom_list[base_idx]->within_wm_pcent++;
			}
		}
	}
}

/*
 * pmi_process_hpsdl
 *
 * This function process the entries on the HPS Domain List (HDL).
 *
 * So now we have a list of Domains, headed by the Domains that
 * were in the event, and followed by other ACTIVE Domains that
 * could be affected if a HPS node Pwr state was changed for one of
 * the event Domains.
 *
 * We care only about the event Domains, since they reported
 * the event.  We will process each event Domain to completion
 * before moving on to the next event Domain.  This means that
 * all HPS nodes and strand adjustments are made for the event
 * Domain and all dependent domains.  Note that other event
 * Domains may be dependent on the one being processed, so
 * when we come around to processing subsequent event Domains,
 * there will be nothing to do since they have already been
 * processed.  Examples:
 *
 * System has Domains A,B,C,D,E,F; Dependencies are simply
 * Domains that share the same HPS nodes, so these are for this Ex:
 * A,C,E;  B,D,F;  Event Domains are A,B.
 * A,C,E can be on the same node, or A on 0, C on 0,1 E on 1.
 *
 * Thus A is processed first, and if a node Pwr state change is
 * needed, then all nodes that contains A,C,E are affected and
 * strands for A,C,E and also adjusted.  Nodes for B,D,F and
 * their strands are not affected.
 *
 * We determine if a HPS node state change is needed for
 * an event Domain, and if not, only strands will be adjusted
 * for the event Domains.  If a HPS node state change is needed,
 * then other Domains on the list are affected and strand
 * adjustments will be calculated for all Domains wrt the new
 * node Pwr state.
 *
 * Guidelines are also used to determine any
 * special cases, as outlined in the Design Doc.  This will
 * determine which nodes to change the Pwr state and the new number
 * of strands for each domain on those nodes.
 *
 * Input: None
 *   hps_dom_list: Global HPS Domain List
 *   hps_dom_idx: Global max number of Domains on the HDL
 *   notifyp: Needed for input to pmi_adjust_util_res_pwr()
 *
 * Output:
 *   Function return:
 *     PS_SUCCESS: The CDL list was processed successfully
 *     PS_FAILURE: An error occurred in processing the CDL list
 */
pwr_req_status_t
pmi_process_hpsdl(pmi_notify_t *notifyp)
{
	pm_hps_domain_t		*hdlp;
	pwr_req_status_t	rtn = PS_SUCCESS;
	int			base_idx, nxt_idx;
	int 			lvl;
	int			new_state, cur_state;

	/*
	 * Go thru the HDL and process the Event Domains. All non
	 * event domains are processed in another loop farther down.
	 */
	for (base_idx = 0; base_idx < hps_dom_idx; base_idx++) {

		hdlp = hps_dom_list[base_idx];

		/* Don't care about non event Domains or processed ones */
		if ((hdlp->event_dom == false) ||
		    (hdlp->processed == true))
			continue;

		/*
		 * Also don't want to process a Domain that has all its
		 * strands faulted or if strands_on == 0.
		 */
		if ((hdlp->faulted >= hdlp->strands_on) ||
		    (hdlp->strands_on == 0))
			continue;

		/*
		 * Calculate the HPS node power states and strands needed for
		 * each Domain dependent on the Event Domain being processed.
		 */
		pmi_recurse_calc_hps_strands(hdlp, 0);

		/*
		 * If the HPS node Pwr state is to be increased, that
		 * usually means that strands will be removed, so do the
		 * Pwr increase first. A decrease in Pwr state usually means
		 * strands will be added, so add them first.
		 * This order is done to help keep performance at a
		 * higher level when making the power state and strand
		 * adjustments.
		 *
		 * Power states for all nodes at the same level will be
		 * at the same state.
		 *
		 * In order to determine if a HPS nodes Power State
		 * is to be changed first or last, we have to compare the
		 * current state to the new calculated state starting at
		 * level 0, until a determination can be made.  The first
		 * level where the current and new states are not equal
		 * will give us the greatest power change, so we don't
		 * need to check any levels passed this one.  Other levels
		 * can have Pwr state changes in the opposite direction,
		 * but they will be at a lesser power change.
		 */
		for (lvl = 0; lvl < hdlp->num_lvls; lvl++) {
			cur_state = hdlp->node_vec[lvl]->pmnode->pn_pwrs->\
			    pps_curlvl;
			new_state = hdlp->node_vec[lvl]->new_state;
			if (new_state != cur_state)
				break;
		}

		/* Node Pwr increase first, then adjust strands */
		if (new_state > cur_state) {

			/* Increase Pwr state first */
			DBG6(zprintf("%s: Increase PState from %d to %d\n",
			    __func__, cur_state, new_state));

			pmi_hps_adjust_pwr(HPS_SOURCE_HDLP, (void *)hdlp,
			    HPS_CALC_POWER, hdlp->num_nodes);

			/* Adjust strands for each dependent domain */
			for (nxt_idx = 0; nxt_idx < hps_dom_idx; nxt_idx++) {
				pm_hps_domain_t *hdnp;
				hdnp = hps_dom_list[nxt_idx];

				/* Any node vector match? */
				if (pmi_hps_node_match(hdlp, hdnp,
				    MATCH_ALL_LEVELS) !=
				    PS_SUCCESS)
					continue;

				/* No strands to adjust */
				if (hdnp->strands_adjust == 0)
					continue;

				DBG6(zprintf("%s: %s: %s %d strands\n",
				    __func__,
				    GET_DOMAIN_NAME(hdnp->domain),
				    ((hdnp->direction == PM_UTIL_HIGH) ?
				    "Add":"Remove"),
				    hdnp->strands_adjust));

				/* Existing Legacy code to adjust strands */
				pmi_adjust_util_res_pwr(hdnp->dom_utilsp,
				    notifyp, hdnp->strands_adjust,
				    hdnp->direction);
			}

		/* Adjust strands first, then decrease Pwr state */
		} else if (new_state < cur_state) {

			/* Adjust strands first */
			for (nxt_idx = 0; nxt_idx < hps_dom_idx; nxt_idx++) {
				pm_hps_domain_t *hdnp;
				hdnp = hps_dom_list[nxt_idx];

				/* Any node vector match? */
				if (pmi_hps_node_match(hdlp, hdnp, -1) !=
				    PS_SUCCESS)
					continue;

				/* No strands to adjust */
				if (hdnp->strands_adjust == 0)
					continue;

				DBG6(zprintf("%s: %s: %s %d strands\n",
				    __func__,
				    GET_DOMAIN_NAME(hdnp->domain),
				    ((hdnp->direction  == PM_UTIL_HIGH) ?
				    "Add":"Remove"),
				    hdnp->strands_adjust));

				/* Existing Legacy code to adjust strands */
				pmi_adjust_util_res_pwr(hdnp->dom_utilsp,
				    notifyp, hdnp->strands_adjust,
				    hdnp->direction);
			}

			/* Decrease Pwr state last */
			DBG6(zprintf("%s: Decrease PState from %d to %d\n",
			    __func__, cur_state, new_state));

			pmi_hps_adjust_pwr(HPS_SOURCE_HDLP, (void *)hdlp,
			    HPS_CALC_POWER, hdlp->num_nodes);

		/*
		 * No Pwr state change, only strand adjustment.  This applies
		 * only to the Event domains.  Non event domains don't need
		 * strands adjustments.
		 */
		} else {

			for (nxt_idx = 0; nxt_idx < hps_dom_idx; nxt_idx++) {
				pm_hps_domain_t *hdnp;
				hdnp = hps_dom_list[nxt_idx];

				/* Only want Event Domains at this point */
				if (hdnp->event_dom != 1)
					continue;

				/*
				 * See if other Event Domains are dependent on
				 * the current one.  If so, their strands have
				 * also been calculated, so adjust them too.
				 */
				if (pmi_hps_node_match(hdlp, hdnp, -1) !=
				    PS_SUCCESS) {
					continue;
				}

				/*
				 * A dependent Event Domain has no strands
				 * to adjust
				 */
				if (hdnp->strands_adjust == 0) {
					continue;
				}

				/* A Dep Event Dom has only a strand change */
				DBG6(zprintf("%s: %s: No PState change, %s "
				    "%d strands\n", __func__,
				    GET_DOMAIN_NAME(hdnp->domain),
				    ((hdnp->direction == PM_UTIL_HIGH) ?
				    "Add":"Remove"),
				    hdnp->strands_adjust));

				/* Done with this domain */
				if (hdnp->strands_adjust == 0)
					continue;

				/* Existing Legacy code to adjust strands */
				pmi_adjust_util_res_pwr(hdnp->dom_utilsp,
				    notifyp, hdnp->strands_adjust,
				    hdnp->direction);
			}
		}
	}

	return (rtn);
}
