/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include "pm_shared.h"
#include "pm_caps.h"

/*
 * Check if the fma status of the given node is NOT "online" or
 * "unretired" (or something equivalent).
 * If so, return true.  If not, return false.
 */
bool_t
pmi_is_resource_faulted(pm_resid_t *rcfg)
{
	bool_t rv;

	/* If the fma status is "unknown", attempt to get it */
	if (rcfg->pi_fma_status == FMA_STATUS_UNKNOWN) {
		if (GET_FMA_RESOURCE_STATUS(rcfg, &rcfg->pi_fma_status) != 0)
			rcfg->pi_fma_status = FMA_STATUS_UNKNOWN;
	}

	switch (rcfg->pi_res_type) {
	case PM_RES_CPU:
		if (rcfg->pi_fma_status != FMA_CPU_STATUS_ONLINE)
			rv = true;
		else
			rv = false;
		break;

	case PM_RES_MEM:
		/* No FMA status change is associated with memory resource */
		rv = false;
		break;

	default:
		rv = true;
	}

	return (rv);
}

/*
 * Process a CPU retire event.
 */
static void
pmi_fma_process_cpu_retire(srcid_t dom_id, pm_resid_t *resource_id)
{
	pm_node_t *pm_nodep;

	pm_nodep = epm_find_node(resource_id, NULL, NULL);

	if (pm_nodep == NULL) {
		DBG(ldm_log_msg("%s: resource not found in device tree.\n",
		    __func__));
		return;
	}

	/* Set FMA CPU status code */
	pm_nodep->pn_rcfg->pi_fma_status = FMA_CPU_STATUS_OFFLINE;
}

/*
 * Process a CPU unretire event.
 */
static void
pmi_fma_process_cpu_unretire(srcid_t dom_id, pm_resid_t *resource_id)
{
	pm_node_t *pm_nodep;

	pm_nodep = epm_find_node(resource_id, NULL, NULL);

	if (pm_nodep == NULL) {
		DBG(ldm_log_msg("%s: resource not found in device tree.\n",
		    __func__));
		return;
	}

	/* Set FMA CPU status code */
	pm_nodep->pn_rcfg->pi_fma_status = FMA_CPU_STATUS_ONLINE;
}

/*
 * Process a RESOURCE_RETIRE_EVENT.
 */
void
pmi_fma_process_retire_event(srcid_t dom_id, pm_resid_t *resource_id)
{
	switch (resource_id->pi_res_type) {
	case PM_RES_CPU:
		pmi_fma_process_cpu_retire(dom_id, resource_id);
		break;

	default:
		DBG(ldm_log_msg("%s: Bad res_type (%d)\n", __func__,
		    resource_id->pi_res_type));
	}
}

/*
 * Process a RESOURCE_UNRETIRE_EVENT.
 */
void
pmi_fma_process_unretire_event(srcid_t dom_id, pm_resid_t *resource_id)
{
	switch (resource_id->pi_res_type) {
	case PM_RES_CPU:
		pmi_fma_process_cpu_unretire(dom_id, resource_id);
		break;

	default:
		DBG(ldm_log_msg("%s: Bad res_type (%d)\n", __func__,
		    resource_id->pi_res_type));
	}
}
