/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include "pm_internal.h"
#include "pm_policy.h"
#include "pm_caps.h"
#include "pm_fma.h"


/*
 * Set function pointers to virtualization SW functions.
 */
static ldm_ops_vector_t ldm_ops = {
	ldm_register_util_change_event,
	ldm_unregister_util_change_event,
	ldm_resume_util_change_notify,
	ldm_get_all_domain_ids,
	ldm_get_domain_utils,
	ldm_get_domain_config,
	ldm_register_domain_config_change_event,
	ldm_register_domain_dr_reg_event,
	ldm_register_domain_down_event,
	ldm_register_resource_config_change_event,
	ldm_register_fma_event,
	ldm_unregister_non_util_events,
	ldm_get_fma_resource_status,
	ldm_get_resource_pm_status,
	ldm_enqueue_work,
	ldm_get_pri,
	ldm_release_pri,
	ldm_get_hv_pm_version,
	ldm_defragment,
	ldm_set_power_states,
	ldm_get_power_states,
	ldm_reconfigure_resources,
	ldm_res_power_fault_report,
	ldm_migration_in_process,
	ldm_delayed_reconfig_in_process,
	ldm_dynamic_resmgmt_in_process,
	ldm_get_domain_state,
	ldm_get_domain_name
};

static md_t	*pmi_pri;

static bool_t	is_pm_running = false;
static void	*pm_dom_dr_reg_hdl;
static void	*pm_dom_down_hdl;

/*
 * pm_start - Called by the LDom Manager during initialization. Starts
 * power management.
 */
void
pm_start(void)
{
	ldm_err_t	results;

	ldm_log_msg("Power Management starting.\n");

	is_pm_running = true;

	/* Set a pointer to the LDOMs ops function vector. */
	virt_ops = &ldm_ops;

	if (GET_PRI(&pmi_pri, &results) == PS_FAILURE) {
		ldm_log_msg("Fatal: %s: Unable to obtain the PRI. Exiting"
		    " power management.\n", __func__);
		is_pm_running = false;
		return;
	}

	/* Build PM Capabilities Table. */
	if (epm_scan_pm_resources(pmi_pri) != PS_SUCCESS) {
		ldm_log_msg("Warning: %s: Unable to build PM Capabilities "
		    "table. Exiting power management.\n", __func__);
		/* Error already flagged */
		is_pm_running = false;
		RELEASE_PRI(&pmi_pri);
		return;
	}

	/* Initialize mechanism for retreiving policy from the SP. */
	if (pmi_init_policy_comm() != PS_SUCCESS) {
		ldm_log_msg("Warning: %s: Unable to init SP Policy "
		    "communications. Exiting power management.\n", __func__);
		is_pm_running = false;
		RELEASE_PRI(&pmi_pri);
		return;
	}

	/* Register for policy change events. */
	if (pmi_register_policy_change_event(PM_POLICY_UNKNOWN) != PS_SUCCESS) {
		ldm_log_msg("Warning: %s: Unable to register for Policy change "
		    "events. Exiting power management.\n", __func__);
		pm_stop(NULL);
		return;
	}
/*
 * This pmi_boot_delay() code is temporarily disabled via non-registration
 * of the event types as this code still needs further stress testing.
 * PM will revert to the stuck high polling loop from LDoms 1.1 patch.
 */
#if 0
	pm_dom_dr_reg_hdl = REGISTER_DOMAIN_DR_REG_EVENT(&results);
	if (pm_dom_dr_reg_hdl == NULL) {
		ldm_log_msg("%s: Error registering for guest domain "
		    "DR registration events. (err = %d)\n", __func__,
		    results.ldme_errs);
		pm_stop(NULL);
	}

	pm_dom_down_hdl = REGISTER_DOMAIN_DOWN_EVENT(&results);
	if (pm_dom_down_hdl == NULL) {
		ldm_log_msg("%s: Error registering for events when a "
		    "guest domain stops running. (err = %d)\n", __func__,
		    results.ldme_errs);
		pm_stop(NULL);
	}
#endif
}


/*
 * pm_util_change_notify - Called by Zeus when a resource's utilization
 * is out of band for a duration exceeding the PM threshold.
 */
void
pm_util_change_notify(void *handlep, srcid_t src_ids, int count,
    pm_res_type_t res_type)
{
	pmi_notify_t	*util_change_notify;

	/*
	 * If the PM Engine is registered to process utilization
	 * notifications, process them. Otherwise, ignore them.
	 */
	if (handlep) {
		util_change_notify = Xmalloc(sizeof (pmi_notify_t));

		util_change_notify->pn_event_type = PE_UTIL_CHANGE_EVENT;
		util_change_notify->pn_src_ids = src_ids;
		util_change_notify->pn_res_type = res_type;
		util_change_notify->pn_resids = NULL;
		util_change_notify->pn_count = count;
		util_change_notify->pn_flag = NULL;

		if (pmi_process_util_change(util_change_notify) != PS_SUCCESS) {
			DBG(ldm_log_msg("%s: Error processing utilization "
			    "notification.\n", __func__));
		}

		Xfree(util_change_notify);
	}
}


/*
 * pm_domain_config_change_notify - Called by Zeus when a domain configuration
 * change occurs.
 */
void
pm_domain_config_change_notify(void *handlep, srcid_t dom_id, uint_t chg_flag,
	pm_res_type_t res_type, pm_resid_t *resource_idsp,  int count,
	ldm_err_t *resultp)
{
	pmi_notify_t	*config_change_notify;

	/*
	 * If the PM Engine is registered to process domain config change
	 * notifications, process them. Otherwise, ignore them.
	 */
	if (handlep) {
		if (resultp)
			resultp->ldme_errs = LDME_OP_EOK;

		config_change_notify = Xmalloc(sizeof (pmi_notify_t));

		pmi_update_fma_status(resource_idsp, count);

		config_change_notify->pn_src_ids = Xmalloc(sizeof (srcid_t));
		config_change_notify->pn_src_ids[0] = dom_id;

		config_change_notify->pn_resids = resource_idsp;
		config_change_notify->pn_event_type = PE_CONFIG_CHANGE_EVENT;
		config_change_notify->pn_res_type = res_type;
		config_change_notify->pn_count = count;
		config_change_notify->pn_flag = chg_flag;

		if (pmi_process_config_change(config_change_notify,
		    resultp) != PS_SUCCESS) {
			DBG(ldm_log_msg("%s: Power management error occurred "
			    "during a configuration change.\n", __func__));
		}

		Xfree(config_change_notify->pn_src_ids);
		Xfree(config_change_notify);
	}
}


/*
 * pm_fma_event_notify - called by zeus, inside the sequencer, when the
 * conditions specified by an uncanceled call to ldm_register_fma_event()are
 * met. The event triggering the notification is described by the parameters of
 * the call.
 * 	dom_id		: indicates the domain which the event is associated
 * 	resource_id	: indicates the specific resource
 *	fma_flag	: identifies the specific event.
 *
 * If fma_flag indicates a retire event, then pm_fma_event_notify() is called
 * after zeus has processed the retire event, and if fma_flag indicates an
 * unretire event, pm_fma_event_notify() is called before zeus processes the
 * unretire event.
 */
void
pm_fma_event_notify(srcid_t dom_id, pm_resid_t resource_id, uint_t fma_flag)
{
	DBG(ldm_log_msg("pm_fma_event_notify called\n"));

	switch (fma_flag) {
	case RESOURCE_RETIRE_EVENT:
		DBG(ldm_log_msg("%s: RETIRE_EVENT received\n", __func__));
		pmi_fma_process_retire_event(dom_id, &resource_id);
		break;

	case RESOURCE_UNRETIRE_EVENT:
		DBG(ldm_log_msg("%s: UNRETIRE_EVENT received\n", __func__));
		pmi_fma_process_unretire_event(dom_id, &resource_id);
		break;

	default:
		DBG(ldm_log_msg("%s: Received unknown fma_flag (%d)\n",
		    __func__, fma_flag));
	}
}


/*
 * pm_stop - Called when Zeus is performing graceful shutdown. Effectively
 * disables power management by putting system into performance mode.
 * Ensures that all power manageable resources are at their full power
 * level and unregisters to receive notifications.
 */
void
pm_stop(void *empty)
{
	bool_t		pm_stop = true;
	ldm_err_t	err;

	if (!is_pm_running)
		return;

	if (pm_dom_dr_reg_hdl != NULL) {
		if (UNREGISTER_NON_UTIL_EVENTS(pm_dom_dr_reg_hdl,
		    &err) != PS_SUCCESS) {
			DBG(ldm_log_msg("%s: Unable to unregister from "
			    "receiving guest domain DR registration "
			    "notifications. (err = %d)\n", __func__,
			    err.ldme_errs));
		}
	}

	if (pm_dom_down_hdl != NULL) {
		if (UNREGISTER_NON_UTIL_EVENTS(pm_dom_down_hdl,
		    &err) != PS_SUCCESS) {
			DBG(ldm_log_msg("%s: Unable to unregister from "
			    "receiving notifications when a guest domain "
			    "stops running. (err = %d)\n", __func__,
			    err.ldme_errs));
		}
	}

	if (pmi_change_to_performance(pm_stop) != PS_SUCCESS) {
		ldm_log_msg("Unable to shut down power management "
		    "gracefully.\n");
	}

	pmi_policy_comm_fini();
	RELEASE_PRI(&pmi_pri);

	is_pm_running = false;
}


/*
 * pm_free errors - Called by LDoms after a notification call is made,
 * to free per resource status data.
 */
void
pm_free_errors(ldm_err_t *err)
{
	ldme_res_status_t	*pm_lrs;

	if (err->ldme_private != NULL) {
		pm_lrs = err->ldme_private;
		Xfree(pm_lrs->lrs);
		Xfree(err->ldme_private);
	}
}


/*
 * Wrapper function to obtain a resource's container ID.
 * (e.g. A cpu's chip ID)
 *
 * Returns results of call to pmi_get_container_id().
 */
int
pm_get_container_id(pm_resid_t *res_id, int *container_id)
{
	return (pmi_get_container_id(res_id, container_id));
}


/*
 * pm_get_defrag_power_states determines the current power states of all
 * resources being moved for purposes of defragmentation.
 *
 * Returns PS_SUCCESS or PS_FAILURE.
 */
pwr_req_status_t
pm_get_defrag_power_states(pm_resid_t *res_id, uint16_t *pwr_states,
    int res_count)
{
	pm_node_t		*nodep;
	char			**ckp = NULL;
	pwr_req_status_t	*req_status = NULL;
	pm_pwr_scheme_t		*pwrs = NULL;
	int			res_idx;
	pwr_req_status_t	ret = PS_SUCCESS;

	ckp = Xmalloc(res_count * sizeof (char *));
	req_status = Xcalloc(res_count, pwr_req_status_t);

	for (res_idx = 0; res_idx < res_count; res_idx++) {
		nodep = epm_find_node(&res_id[res_idx], NULL, NULL);
		if (nodep == NULL) {
			DBG(ldm_log_msg("%s: Resource not found in "
			    "device tree.\n", __func__));
			ret = PS_FAILURE;
			goto cleanup;
		}

		res_id[res_idx].pi_fma_status = nodep->pn_rcfg->pi_fma_status;

		pwrs = epm_get_pwr_scheme(nodep,
		    nodep->pn_pwrs->pps_pwr_ctl.pp_method);

		if (!pwrs) {
			ret = PS_FAILURE;
			goto cleanup;
		}

		ckp[res_idx] = pwrs->pps_pwr_ctl.pp_cookie;
	}

	if (GET_POWER_STATES(res_count, ckp, pwr_states, req_status) !=
	    PS_SUCCESS) {
		ret = PS_FAILURE;
		goto cleanup;
	}

cleanup:
	Xfree(ckp);
	Xfree(req_status);
	return (ret);
}

/*
 * Helper function to return policy state of whether the PM mode is
 * currently Performance, or not.  Returns true if in Performance mode,
 * otherwise returns false.
 */
static bool_t
pm_in_performance_mode(void)
{
	pmi_policy_t	policy;

	if (!is_pm_running)
		return (true);

	if ((*pmi_get_policy)(&policy) != PS_SUCCESS) {
		DBG(ldm_log_msg("%s: Unable to get PM policy.\n", __func__));
		return (false);
	}

	return (policy == PM_POLICY_PERFORMANCE);
}

/*
 * pm_migration_is_possible() returns "true" if a domain migration is possible,
 * and "false" if a domain migration is not possible.  Migration is not possible
 * when in elastic mode, so it's blocked.
 */
bool_t
pm_migration_is_possible(void)
{
	return (pm_in_performance_mode());
}

/*
 * pm_add_crypto_units_is_possible() returns "true" if it is okay to add
 * crypto units to a domain.  This is determined by whether the PM mode is
 * Performance or not.  The function returns true for being in Performance
 * mode, otherwise it returns false.
 */
bool_t
pm_add_crypto_units_is_possible(void)
{
	return (pm_in_performance_mode());
}
