/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)pm_policy_via_snmp.c	1.7	09/05/01 SMI"

#include <stdio.h>
#include <errno.h>
#include <thread.h>
#include <synch.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include "pm_internal.h"
#include "basics.h"
#include "zprintf.h"
#include "pm_policy.h"

#define	POLICY_INIT_TIMEOUT	15

typedef struct {
	pmi_policy_t	policy;
	int		error;
} pmi_thread_info_t;

/* Globals needed for policy initialization */
static bool_t		snmp_thread_running = false;
static bool_t		snmp_thread_waiting = false;
static cond_t		thread_policy_cv = DEFAULTCV;
static mutex_t		snmp_thread_lock = DEFAULTMUTEX;
static mutex_t		handle_lock = DEFAULTMUTEX;
static pmi_thread_info_t	thread_info;
static sigset_t			new_sig;
static struct sigaction		act;
static thread_t		snmp_tid = { 0 };
static void		*snmp_hdlp = NULL;

static void *pmi_snmp_get_policy_loop(void *p);

/*
 * pmi_get_policy_by_snmp
 * This function spawns a thread to retrieve the PM policy via an SNMP
 * call.  Multiple callers may execute this function so the critical
 * region is protected by mutex handle_lock.  The function uses a
 * separate thread to make the SNMP call to avoid blocking.  If the
 * thread does not exist then it is spawned.  The functiona waits for
 * the SNMP thread to complete via a condition variable.  If the SNMP
 * thread does not return in time then the SNMP thread is signaled via
 * SIGUSR1.  Once killed, the SNMP thread will be spawned the next
 * time this function is called.  The SNMP thread is also signaled if
 * bad error status is returned in thread_info.
 *
 * SNMP thread is function pmi_snmp_get_policy_loop
 *
 * Return: PS_SUCCESS or appropriate error such as PS_FAILURE.
 */
int
pmi_get_policy_by_snmp(pmi_policy_t *pm_policy)
{
	int		err;
	timestruc_t	to;

	/* Make sure only one instance is running */
	(void) mutex_lock(&handle_lock);

	/*
	 * Grab the mutex so pmi_snmp_get_policy_loop can't set
	 * snmp_thread_waiting
	 */
	(void) mutex_lock(&snmp_thread_lock);

	/*
	 * This struct is used to pass data between pmi_snmp_get_policy_loop
	 * and this function.
	 */
	thread_info.policy = PM_POLICY_UNKNOWN;
	thread_info.error = PS_FAILURE;

	/* Create SNMP thread if it doesn't exist */
	if (!snmp_thread_running) {

		/*
		 * thread_policy is used to pass data between this
		 * function and the SNMP thread.
		 */
		err = thr_create(NULL, 0, pmi_snmp_get_policy_loop,
		    NULL, THR_DAEMON, &snmp_tid);
		if (err != 0) {
			thread_info.error = PS_FAILURE;
			goto done;
		}
		snmp_thread_running = true;
	}

	/* The SNMP thread can now run */
	snmp_thread_waiting = false;

	/* Wait for the SNMP thread to finish */
	while (snmp_thread_waiting == false) {
		to.tv_sec = POLICY_INIT_TIMEOUT;
		to.tv_nsec = 0;
		err = cond_reltimedwait(&thread_policy_cv, &snmp_thread_lock,
		    &to);

		if ((err == ETIME) || (thread_info.error == PS_FAILURE)) {
			/*
			 * The SNMP thread didn't return in time or
			 * returned failure, so it
			 * is signaled here and failure is returned.
			 */
			thr_kill(snmp_tid, SIGUSR1);
			thr_join(snmp_tid, NULL, NULL);
			snmp_thread_running = false;

			if (err == ETIME) {
				ldm_log_msg("%s: Unable to get the PM "
				    "Policy - timeout.\n", __func__);
				thread_info.error = PS_TIMEDOUT;
			}

			goto done;
		}
	}

	/* Can never happen */
	ASSERT(thread_info.error == PS_SUCCESS);

	/* Validate Policy */
	if (thread_info.policy > PM_POLICY_UNKNOWN) {
		*pm_policy = thread_info.policy;
	} else {
		thread_info.error = PS_FAILURE;
	}
done:
	(void) mutex_unlock(&snmp_thread_lock);
	(void) mutex_unlock(&handle_lock);

	return (thread_info.error);
}

/* SNMP thread signal handler */
static void
pmi_snmp_thread_exit_handler()
{
	DBG2(ldm_log_msg("%s SIG handler called\n", __func__));

	if (snmp_hdlp) {
		snmp_fini(snmp_hdlp);
		snmp_hdlp = NULL;
	}
	thr_exit(NULL);
}

/*
 * pmi_snmp_get_policy_loop is the SNMP thread loop
 * This function constantly polls for the snmp policy and stashes that
 * information in thread_info.  When an error occurs then the thread
 * returns bad status in thread_info or the operation will block.  A
 * signal handler is set up to intercept SIGUSR1 in such cases so the
 * thread can exist gracefully.
 * The thread only executes when pmi_get_policy_by_snmp resets the
 * run variable snmp_thread_waiting.  Synchronization between this
 * thread and pmi_get_policy_by_snmp occurs via the conditional variable
 * set.  When this thread runs then pmi_get_policy_by_snmp is blocked
 * and vice versa, except at startup of the thread where it immediately
 * executes the code to get the initial policy.
 */
static void *
pmi_snmp_get_policy_loop(void *p)
{
	bool_t		show_err_msg = true;
	int		snmp_syserr = 0;
	int		rv, row = 0;

	/* Set signal mask for SNMP thread */
	sigemptyset(&new_sig);
	sigaddset(&new_sig, SIGUSR1);

	/* Set thread SIGUSR1 handler in case thread is signaled on error */
	act.sa_handler = pmi_snmp_thread_exit_handler;
	sigaction(SIGUSR1, &act, NULL);
	thr_sigsetmask(SIG_UNBLOCK, &new_sig, NULL);

	/* Initialize SNMP handle */
	ASSERT(snmp_hdlp == NULL);
	do {
		snmp_hdlp = snmp_init();
		if (!snmp_hdlp) {
			/* Don't flood log */
			if (show_err_msg) {
				ldm_log_msg("%s: snmp_init()\n", __func__);
				show_err_msg = false;
			}
			sleep(1);
		}
	} while (!snmp_hdlp);

	show_err_msg = true;

	/* Loop forever until a signal received */
	do {
		rv = snmp_get_int(snmp_hdlp, OID_ilomPMPolicy, row,
		    (int *)&thread_info.policy, &snmp_syserr);
		thread_info.error = PS_SUCCESS;
		if ((snmp_syserr == ECANCELED) || (rv < 0)) {
			/*
			 * We observed after an error that we sometimes need
			 * a new handle generated.
			 */
			snmp_fini(snmp_hdlp);
			snmp_hdlp = NULL;

			ldm_log_msg("%s: snmp_get_int failed (rv=%d) "
			    "(err=%d)\n", __func__, rv, snmp_syserr);
			thread_info.error = PS_FAILURE;
			thread_info.policy = PM_POLICY_UNKNOWN;
		}

		/*
		 * All done so tell pmi_get_policy_by_snmp function
		 * this thread is waiting
		 */
		(void) mutex_lock(&snmp_thread_lock);
		snmp_thread_waiting = true;
		(void) cond_signal(&thread_policy_cv);
		(void) mutex_unlock(&snmp_thread_lock);

		/* Sleep until told to do more work */
		while (snmp_thread_waiting == true) {
			sleep(1);
		}
	} while (true);

	/* LINTED */
}



/*
 * Manage the initialization of the PM policy communication mechanism.
 */
int
pmi_init_policy_comm(void)
{
	pmi_policy_t policy;

	/* Assign function pointer.  This is overridden by test library. */
	pmi_get_policy = &pmi_get_policy_by_snmp;

	/* Log if there is a startup error.  Maybe SP is rebooting. */
	if (pmi_get_policy(&policy) != PS_SUCCESS) {
		ldm_log_msg("%s: pmi_get_policy() failed.\n", __func__);
	}

	/* Always succeed */
	return (PS_SUCCESS);
}


void
pmi_policy_comm_fini(void)
{
	/*
	 * FIXME Intentionally a no-op as the policy poller needs to be
	 * brought down first before the snmp thread can be gracefully
	 * killed.  To do this, set policy_poll_stop to false in
	 * pm_engine.c so that pmi_poll_policy() will exit gracefully.
	 * A check needs to be added to pmi_get_policy_by_snmp() in
	 * this file immediately after the function is entered to return
	 * PS_FAILURE.  This is to prevent pmi_get_policy_by_snmp() from
	 * respawning the snmp thread.  With these in place then it
	 * should be safe to call thr_kill() on the snmp thread.
	 * Note that no manipulation of snmp_hdlp is needed here now as
	 * that is handled by the snmp thread's signal handler.
	 */
}
