/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_PM_SHARED_H_
#define	_PM_SHARED_H_

#ifdef __cplusplus
extern "C" {
#endif

#include "basics.h"
#include <strings.h>
#include "allocate.h"
#include "zprintf.h"
#include "mdesc.h"
#include "mdesc_impl.h"
#include "resource_id.h"

/* LDoms and Power Management interface functions and shared data structures */

/*
 * FIXME: There are several asserts in the Zeus code requiring that
 * PM_RES_TYPES_TOTAL == 1. This will be fixed when the error handling code
 * within Zeus gets updated to handle > 1 resource type.
 */
typedef enum {
	PM_RES_CPU,
	PM_RES_MEM,
	PM_RES_TYPES_TOTAL
} pm_res_type_t;

#define	CPU_RES_NAME	"cpu"
#define	MEM_GROUPING	"memory-grouping"

typedef struct {
	pm_res_type_t	pi_res_type;	/* pm_type enum value */
	int	pi_fma_status;		/* FMA status */
	char	*pi_res_name;	/* resource name: "cpu" or "memory" */
	union pi_res {
		struct pi_cpu_res {
			uint64_t pi_id_val;	/* "id" value */
		} cpu;
		struct pi_mem_res {
			uint64_t pi_id_val;	/* group or region "id" value */
			struct pi_mem_reg {
				uint64_t pi_reg_base;
				uint64_t pi_reg_size;
			} mem_reg;
		} mem;
	} u;
} pm_resid_t;

typedef void * srcid_t;	/* source id (e.g. domain or memory grouping) */

typedef struct {
	pm_resid_t ldmu_resid;	/* resource id */
	double	ldmu_util;	/* resource utilization percentage */
} ldm_util_t;

typedef enum {
	PM_UTIL_LOW = 1,
	PM_UTIL_WITHIN,
	PM_UTIL_HIGH
} pm_util_dir_t;

typedef struct {
	pm_res_type_t	ldmr_type;	/* type of resource */
	uint_t	ldmr_rescnt;		/* number of resources in the group */
	double	ldmr_avg_util;		/* avg. resource util percentage */
	hrtime_t	ldmr_time;	/* timestamp of threshold crossing */
	pm_util_dir_t	ldmr_util_dir;	/* direction of threshold crossing */
	ldm_util_t	*ldmr_util_p;	/* per resource utilization */
} ldm_resutils_t;

/* ldm_register_domain_config_change_event() flags */
#define	START_DOMAIN_PRE	0x01
#define	STOP_DOMAIN_POST	0x02
#define	ADD_RESOURCE_PRE	0x04
#define	REM_RESOURCE_POST	0x08
#define	REM_RESOURCE_DEFRAG	0x10
#define	ADD_RESOURCE_DEFRAG	0x20

#define	GUEST_DOMAIN_DR_REG	0x40 /* ldm_register_domain_dr_reg_event flag */
#define	GUEST_DOMAIN_DOWN	0x80 /* ldm_register_domain_down_event flag */

/*
 * Flag to ignore DR errors when doing a change to Performance mode so there
 * is no backout of the operation if the drd service has already stopped on
 * system shutdown.
 */
#define	IGNORE_DR_ERRORS	0x100
#define	PM_NO_FLAGS		0x0

/* FMA event type */
#define	RESOURCE_RETIRE_EVENT	1
#define	RESOURCE_UNRETIRE_EVENT	2

/* FMA CPU status codes */
#define	FMA_CPU_STATUS_ONLINE	0
#define	FMA_CPU_STATUS_OFFLINE	1
#define	FMA_CPU_STATUS_ILLEGAL	2

#define	FMA_STATUS_UNKNOWN	-1

/* Default min number of vcpus that must be powered on per LDom. */
#if TWO_MIN_VCPU_PER_LDOM
#define	MIN_VCPU_PER_LDOM	2
#else
#define	MIN_VCPU_PER_LDOM	1
#endif

#define	MIN_VCPU_ALLOWED	1

/* Types supported by ldm_get_resource_pm_status */
#define	PM_DR_STATUS	1

/* Flags for ldm_get_all_domain_ids */
#define	ALL_DOMAINS	0
#define	ACTIVE_DOMAINS	1

/* Default limits/caps when adjusting power */
#define	PM_CPU_ADJUST_LIMIT	512	/* for cpu resources */

/* ldm_defragment policy */
typedef enum {
	PERFORMANCE,		/* power management is turned off */
	PACK			/* pack parked strands within core boundary */
} defrag_policy_t;

/* ldm_err_t data structure */
typedef struct {
	int	sanity;		/* resource sanity condition */
	pm_resid_t *resid;	/* resource id */
} ldme_res_sane_t;

typedef struct {
	int	rescnt;		/* resource count */
	ldme_res_sane_t	*lrs;	/* array of res sanity conditions */
} ldme_res_status_t;

/* ldme_errs code */
typedef enum {
	LDME_OP_EOK = 0,	/* op succeeded */
	LDME_OP_EINVAL,		/* invalid arg */
	LDME_OP_EUNAVAIL,	/* requested object unavailable */
	LDME_OP_ETIMEDOUT,	/* op timed out */
	LDME_OP_EAGAIN,		/* try op later */
	LDME_OP_EUNSUPPORTED,	/* unsupported op */
	LDME_OP_EEXISTS,	/* Registration already exists */
	LDME_OP_FAILURE		/* Op failed for other reasons */
} ldme_code_t;

typedef struct {
	ldme_code_t	ldme_errs;	/* function error code */
	void	*ldme_private;	/* function private error information */
} ldm_err_t;

/* possible 'sanity' value */
#define	LDME_RES_GOOD		0	/* res good */
#define	LDME_RES_UNCFG_FAILURE	0x1	/* res unconfigure failed */
#define	LDME_RES_CFG_FAILURE	0x2	/* res configure failed */
#define	LDME_RES_STUCK_LOW	0x4	/* res stuck at low power */
#define	LDME_RES_STUCK_HIGH	0x8	/* res stuck at high power */
#define	LDME_RES_UNKNOWN_POWER	0x10	/* res power state unknown */
#define	LDME_RES_FAULTED	0x20	/* res is FMA faulted */
#define	LDME_RES_STUCK_HIGH_REQD	0x40	/* res is NeedOne */

/* PM events that resulted in resource status change */
typedef enum {
	LDME_PE_ENTER_PM,	/* policy changed to enable PM */
	LDME_PE_EXIT_PM,	/* stopped PM */
	LDME_PE_UTIL,		/* utilization change notification event */
	LDME_PE_DOMAIN_BOOTING, /* guest domain is booting */
	LDME_PE_REPORT_RES	/* report a resource as not PM'able */
} ldme_event_t;

typedef enum {
	PS_SUCCESS,		/* request was successful */
	PS_EINVAL,		/* invalid argument */
	PS_EILLEGAL,		/* request illegal */
	PS_TIMEDOUT,		/* request timed out */
	PS_FAILURE		/* request failed */
} pwr_req_status_t;

typedef enum {
	PCO_NONE,		/* no reconfiguration operation needed */
	PCO_ADD_RESOURCE,	/* add resource to guest domain */
	PCO_REMOVE_RESOURCE,	/* remove resource from guest domain */
	PCO_FORCE_ADD		/* force a resource to be powered on */
} pm_config_op_t;

typedef enum {
	PE_CONFIG_CHANGE_EVENT,
	PE_UTIL_CHANGE_EVENT,
	PE_POLICY_CHANGE_EVENT,
	PE_FMA_EVENT
} pm_event_t;


/*
 * Interface functions implemented as part of core Zeus functionality
 * called by Power Management daemon.
 */
extern void *ldm_register_util_change_event(pm_res_type_t res_type,
    double low_wm, double high_wm, uint32_t time_low_ms, uint32_t time_high_ms,
    ldm_err_t *results);
extern int ldm_unregister_util_change_event(void *res_handle,
    ldm_err_t *results);
extern int ldm_resume_util_change_notify(void *res_handle, ldm_err_t *results);
extern int ldm_get_all_domain_ids(srcid_t **domlist, uint_t *count,
    ldm_err_t *results, int flag);
extern int ldm_get_domain_utils(srcid_t src_id, pm_res_type_t res_type,
    ldm_resutils_t **utils, ldm_err_t *results);
extern int ldm_get_domain_config(srcid_t dom_id, pm_res_type_t res_types,
    pm_resid_t **config, int *count, ldm_err_t *results);
extern void *ldm_register_domain_config_change_event(ldm_err_t *results);
extern void *ldm_register_domain_dr_reg_event(ldm_err_t *results);
extern void *ldm_register_domain_down_event(ldm_err_t *results);
extern void *ldm_register_resource_config_change_event(pm_res_type_t res_type,
	ldm_err_t *results);
extern void *ldm_register_fma_event(pm_res_type_t res_type, ldm_err_t *results);
extern int ldm_unregister_non_util_events(void *handle, ldm_err_t *results);
extern int ldm_get_fma_resource_status(pm_resid_t *resid, int *status);
extern int ldm_get_resource_pm_status(pm_resid_t *resid,
    uint32_t, bool_t *status);
extern void ldm_enqueue_work(void (*func)(void *), void *arg);
extern int ldm_get_pri(md_t **pri, ldm_err_t *results);
extern void ldm_release_pri(md_t **pri);
extern int ldm_get_hv_pm_version(uint16_t *major, uint16_t *minor);
extern void ldm_defragment(defrag_policy_t policy, pm_res_type_t res_type,
    ldm_err_t *results);
extern pwr_req_status_t ldm_set_power_states(int res_count, char **pm_cookie,
    uint16_t *new_states, pwr_req_status_t *status);
extern pwr_req_status_t ldm_get_power_states(int res_count, char **pm_cookie,
    uint16_t *cur_state, pwr_req_status_t *status);
extern int ldm_reconfigure_resources(pm_config_op_t op, int res_count,
    pm_resid_t *ids, int *status, ldm_err_t *results, uint32_t flags);
extern void ldm_res_power_fault_report(ldm_err_t *res_status,
    ldme_event_t event);
extern bool_t ldm_migration_in_process(void);
extern bool_t ldm_delayed_reconfig_in_process(void *ldp);
extern bool_t ldm_dynamic_resmgmt_in_process(void);
extern bool_t hv_is_mem_util_supported(void);
extern int ldm_get_domain_state(void *ldom);
extern char *ldm_get_domain_name(void *ldom);

/* LDOM ops vector for PM engine's use */
typedef struct ldm_ops_vector {
	void *(*register_util_change_event)(pm_res_type_t res_type,
	    double low_wm, double high_wm, uint32_t time_low_ms,
	    uint32_t time_high_ms, ldm_err_t *results);
	int (*unregister_util_change_event)(void *res_handle,
	    ldm_err_t *results);
	int (*resume_util_change_notify)(void *res_handle, ldm_err_t *results);
	int (*get_all_domain_ids)(srcid_t **domlist, uint_t *count,
	    ldm_err_t *results, int flag);
	int (*get_domain_utils)(srcid_t src_id, pm_res_type_t res_type,
	    ldm_resutils_t **utils, ldm_err_t *results);
	int (*get_domain_config)(srcid_t dom_id, pm_res_type_t res_types,
	    pm_resid_t **config, int *count, ldm_err_t *results);
	void *(*register_domain_config_change_event)(ldm_err_t *results);
	void *(*register_domain_dr_reg_event)(ldm_err_t *results);
	void *(*register_domain_down_event)(ldm_err_t *results);
	void *(*register_resource_config_change_event)(pm_res_type_t res_type,
	    ldm_err_t *results);
	void *(*register_fma_event)(pm_res_type_t res_type, ldm_err_t *results);
	int (*unregister_non_util_events)(void *handle, ldm_err_t *results);
	int (*get_fma_resource_status)(pm_resid_t *resid, int *status);
	int (*get_resource_pm_status)(pm_resid_t *resid, uint32_t,
	    bool_t *status);
	void (*enqueue_work)(void (*func)(void *), void *arg);
	int (*get_pri)(md_t **pri, ldm_err_t *results);
	void (*release_pri)(md_t **pri);
	int (*get_hv_pm_version)(uint16_t *major, uint16_t *minor);
	void (*defragment)(defrag_policy_t policy, pm_res_type_t res_type,
	    ldm_err_t *results);
	pwr_req_status_t (*set_power_states)(int res_count, char **pm_cookie,
	    uint16_t *new_state, pwr_req_status_t *status);
	pwr_req_status_t (*get_power_states)(int res_count, char **pm_cookie,
	    uint16_t *cur_state, pwr_req_status_t *status);
	int (*reconfigure_resources)(pm_config_op_t op, int res_count,
	    pm_resid_t *ids, int *status, ldm_err_t *results, uint32_t flags);
	void (*res_power_fault_report)(ldm_err_t *res_status,
	    ldme_event_t event);
	bool_t (*migration_in_process)(void);
	bool_t (*delayed_reconfig_in_process)(void *ldp);
	bool_t (*drm_in_process)(void);
	int (*get_domain_state)(void *ldom);
	char *(*get_domain_name)(void *ldom);
} ldm_ops_vector_t;

ldm_ops_vector_t	*virt_ops;

#define	REGISTER_UTIL_CHANGE_EVENT(_rt, _lw, _hw, _tlm, _thm, _res) \
    virt_ops->register_util_change_event(_rt, _lw, _hw, _tlm, _thm, _res)

#define	UNREGISTER_UTIL_CHANGE_EVENT(_rh, _res)	\
    virt_ops->unregister_util_change_event(_rh, _res)

#define	RESUME_UTIL_CHANGE_NOTIFY(_rh, _res) \
    virt_ops->resume_util_change_notify(_rh, _res)

#define	GET_ALL_DOMAIN_IDS(_dl, _c, _res, _flag) \
    virt_ops->get_all_domain_ids(_dl, _c, _res, _flag)

#define	GET_DOMAIN_UTILS(_si, _rt, _utils, _res) \
    virt_ops->get_domain_utils(_si, _rt, _utils, _res)

#define	GET_DOMAIN_CONFIG(_di, _rt, _conf, _c, _res) \
    virt_ops->get_domain_config(_di, _rt, _conf, _c, _res)

#define	REGISTER_DOMAIN_CONFIG_CHANGE_EVENT(_res) \
    virt_ops->register_domain_config_change_event(_res)

#define	REGISTER_DOMAIN_DR_REG_EVENT(_res) \
    virt_ops->register_domain_dr_reg_event(_res)

#define	REGISTER_DOMAIN_DOWN_EVENT(_res) \
    virt_ops->register_domain_down_event(_res)

#define	REGISTER_RESOURCE_CONFIG_CHANGE_EVENT(_rt, _res) \
    virt_ops->register_resource_config_change_event(_rt, _res)

#define	REGISTER_FMA_EVENT(_rt, _res) \
    virt_ops->register_fma_event(_rt, _res)

#define	UNREGISTER_NON_UTIL_EVENTS(_hdl, _res) \
    virt_ops->unregister_non_util_events(_hdl, _res)

#define	GET_FMA_RESOURCE_STATUS(_i, _s) \
    virt_ops->get_fma_resource_status(_i, _s)

#define	GET_RESOURCE_PM_STATUS(_i, _t, _s) \
    virt_ops->get_resource_pm_status(_i, _t, _s)

#define	ENQUEUE_WORK(_func, _arg) \
    virt_ops->enqueue_work(_func, _arg)

#define	GET_PRI(_pri, _res) \
    virt_ops->get_pri(_pri, _res)

#define	RELEASE_PRI(_pri) \
    virt_ops->release_pri(_pri)

#define	GET_HV_PM_VERSION(_major, _minor) \
    virt_ops->get_hv_pm_version(_major, _minor)

#define	DEFRAGMENT(_pol, _rt, _res) \
    virt_ops->defragment(_pol, _rt, _res)

#define	SET_POWER_STATES(_rc, _pm_cookie, _new_st, _stat)	\
    virt_ops->set_power_states(_rc, _pm_cookie, _new_st, _stat)

#define	GET_POWER_STATES(_rc, _pm_cookie, _cur_st, _stat) \
    virt_ops->get_power_states(_rc, _pm_cookie, _cur_st, _stat)

#define	RECONFIGURE_RESOURCES(_op, _rc, _ids, _stat, _res, _fl) \
    virt_ops->reconfigure_resources(_op, _rc, _ids, _stat, _res, _fl)

#define	RES_POWER_FAULT_REPORT(_st, _ev) \
    virt_ops->res_power_fault_report(_st, _ev)

#define	MIGRATION_IN_PROCESS() \
    virt_ops->migration_in_process()

#define	DELAYED_RECONFIG_IN_PROCESS(_dm) \
    virt_ops->delayed_reconfig_in_process(_dm)

#define	DYNAMIC_RESMGMT_IN_PROCESS() \
    virt_ops->drm_in_process()

#define	GET_DOMAIN_STATE(_ldom) \
    virt_ops->get_domain_state(_ldom)

#define	GET_DOMAIN_NAME(_ldom) \
    virt_ops->get_domain_name(_ldom)

/*
 * Interface functions implemented as part of core Power Management
 * functionality called by Zeus.
 */
extern void pm_start(void);
extern void pm_stop(void *);
extern void pm_util_change_notify(void *handle, srcid_t src_ids, int count,
	pm_res_type_t res_type);
extern void pm_domain_config_change_notify(void *handle, srcid_t dom_id,
	uint_t flags, pm_res_type_t res_type, pm_resid_t *resource_ids,
	int count, ldm_err_t *result);
extern void pm_fma_event_notify(srcid_t dom_id, pm_resid_t resource_id,
	uint_t fma_flag);
extern void pm_free_errors(ldm_err_t *err);
extern int pm_get_container_id(pm_resid_t *resource_id, int *container_id);
extern pwr_req_status_t pm_get_defrag_power_states(pm_resid_t *resource_id,
    uint16_t *pwr_states, int res_count);
extern bool_t pm_migration_is_possible(void);
extern bool_t pm_add_crypto_units_is_possible(void);

/*
 * Function pointers so that Zeus can call Power Management functions via
 * an ops vector.
 */
typedef struct pm_ops_vector {
	int revision;
	void (*pm_start)(void);
	void (*pm_stop)(void *);
	void (*pm_util_change_notify)(void *handle, srcid_t src_id,
		int count, pm_res_type_t res_type);
	void (*pm_domain_config_change_notify)(void *handle, srcid_t dom_id,
		uint_t flags, pm_res_type_t res_type, pm_resid_t *resource_ids,
		int count, ldm_err_t *result);
	void (*pm_fma_event_notify)(srcid_t dom_id, pm_resid_t resource_id,
		uint_t fma_flag);
	void (*pm_free_errors)(ldm_err_t *err);
	int (*pm_get_container_id)(pm_resid_t *resource_id, int *container_id);
	pwr_req_status_t (*pm_get_defrag_power_states)(pm_resid_t *res_ids,
	    uint16_t *pwr_states, int res_count);
	bool_t (*pm_migration_is_possible)(void);
	bool_t (*pm_add_crypto_units_is_possible)(void);
} pm_ops_vector_t;

pm_ops_vector_t	*pm_ops;  /* PM ops vector for LDOM's use */


void *lp_get_domain_handle(void);
void *lp_get_util_handle(int);

#define	ldm_log_msg zprintf

#define	MEM_UTIL_HIST_TIME	(20000) /* utilization history - in ms */
#define	MEM_UTIL_POLL_INTV	(2000)  /* utilization poll interval - in ms */
#define	MEM_UTIL_TAB_NROWS	(MEM_UTIL_HIST_TIME / MEM_UTIL_POLL_INTV)

/*
 * Increment the index of the mem's utilization history table and add the
 * utilization value and time to the table.
 */
#define	ADD_MEM_UTIL(pmp, val, t) { \
	pmp->stats.hist_idx = (pmp->stats.hist_idx + 1) % MEM_UTIL_TAB_NROWS; \
	pmp->stats.hist_table[pmp->stats.hist_idx].value = val; \
	pmp->stats.hist_table[pmp->stats.hist_idx].time = t; \
}

/*
 * Get the index of the last most recent entry in the utilization history table
 * from the given index.
 */
#define	GET_PREV_MEM_UTIL_IDX(i) \
	(i == 0) ? (MEM_UTIL_TAB_NROWS - 1) : i - 1

/* Entry in the utilization history table */
typedef struct pm_mem_util {
	double		value;	   /* utilization value */
	hrtime_t	time;	   /* time util value was obtained */
} pm_mem_util_t;

/* Memory utilization states */
typedef struct mem_states {
	int		hist_idx;	/* most recent entry in util hist */
	pm_mem_util_t	hist_table[MEM_UTIL_TAB_NROWS]; /* util hist table */
	uint64_t	last_perf_cntr;	/* last perf counter (absolute) */
	uint64_t	highest;	/* highest so far (per msec) */
	uint64_t	lowest;		/* lowest so far (per msec) */
} mem_states_t;

/* Memory resource info */
typedef struct pm_mem {
	resid_t		pm_mem_resid;		/* resource id */
	mem_states_t	stats;			/* resource-spec. util info */
	ldm_resutils_t	pm_mem_resutils;	/* resource-generic util info */
	CHAIN_LINK(struct pm_mem, link);
} pm_mem_t;

/* List of memory resources. */
typedef struct pm_mem_list {
	CHAIN(pm_mem_t, mem_list);
} pm_mem_list_t;

pm_mem_list_t	pm_mem_list;

#ifdef __cplusplus
}
#endif

#endif /* _PM_SHARED_H_ */
