/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)pm_events.c	1.6	09/04/10 SMI"

#include <unistd.h>

#include "ldom.h"
#include "allocate.h"
#include "chain.h"
#include "pm_events.h"

mutex_t pm_util_list_lock = DEFAULTMUTEX;

void
init_pm_events(void)
{
	CHAIN_INIT(pm_util_list.util_event_list);
	CHAIN_INIT(pm_domain_list.domain_event_list);

	pm_domain_events_on = false;
	pm_cpu_events_on = false;
	pm_fma_events_on = false;
}

void *
ldm_register_util_change_event(pm_res_type_t res_type, double low_wm,
    double high_wm, uint32_t time_low_ms, uint32_t time_high_ms,
    ldm_err_t *results)
{
	pm_util_event_info_t	*new_event, *event;

	mutex_lock(&pm_util_list_lock);

	/* Check to see if the event for this res_type already exists */
	CHAIN_ITER(pm_util_list.util_event_list, event, link) {
		if (event->res_type == res_type)
			break;
	}

	/* If we do then we remove the old registration and create a new one */
	if (event != NULL) {
		CHAIN_UNLINK(pm_util_list.util_event_list, event, link);
		Xfree(event);
	}

	new_event = Xmalloc(sizeof (pm_util_event_info_t));
	new_event->res_type = res_type;
	new_event->low_wm = low_wm;
	new_event->high_wm = high_wm;
	new_event->time_low_ms = time_low_ms;
	new_event->time_high_ms = time_high_ms;
	new_event->active = true;
	new_event->handle = new_event;

	CHAIN_ADD(pm_util_list.util_event_list, new_event, link);

	if (res_type == PM_RES_CPU)
		reset_cpu_util_data(new_event);

	mutex_unlock(&pm_util_list_lock);

	results->ldme_errs = LDME_OP_EOK;

	return ((void *)new_event);
}

int
ldm_unregister_util_change_event(void *handle, ldm_err_t *results)
{
	pm_util_event_info_t	*event;

	mutex_lock(&pm_util_list_lock);

	CHAIN_ITER(pm_util_list.util_event_list, event, link) {
		if (event->handle == handle)
			break;
	}

	if (event != NULL) {
		CHAIN_UNLINK(pm_util_list.util_event_list, event, link);
		Xfree(event);
	} else {
		mutex_unlock(&pm_util_list_lock);
		results->ldme_errs = LDME_OP_EUNAVAIL;
		return (-1);
	}

	mutex_unlock(&pm_util_list_lock);

	results->ldme_errs = LDME_OP_EOK;
	return (0);
}

int
ldm_resume_util_change_notify(void *handle, ldm_err_t *results)
{
	pm_util_event_info_t	*event;

	CHAIN_ITER(pm_util_list.util_event_list, event, link) {
		if (event->handle == handle)
			break;
	}

	if (event != NULL) {
		if (event->res_type == PM_RES_CPU) {
			reset_cpu_util_data(event);
		}
		event->active = true;
	} else {
		results->ldme_errs = LDME_OP_EUNAVAIL;
		return (-1);
	}

	results->ldme_errs = LDME_OP_EOK;
	return (0);
}

void *
ldm_register_domain_config_change_event(ldm_err_t *results)
{
	pm_domain_event_info_t	*new_event, *event;

	CHAIN_ITER(pm_domain_list.domain_event_list, event, link) {
		if (event->event_type == PM_DOMAIN_CONFIG_EVENT) {
			results->ldme_errs = LDME_OP_EEXISTS;
			return (NULL);
		}
	}

	new_event = Xmalloc(sizeof (pm_domain_event_info_t));
	new_event->event_type = PM_DOMAIN_CONFIG_EVENT;
	new_event->handle = new_event;

	CHAIN_ADD(pm_domain_list.domain_event_list, new_event, link);

	pm_domain_events_on = true;
	results->ldme_errs = LDME_OP_EOK;

	return ((void *)new_event);
}

/*
 * ldm_register_domain_dr_reg_event - Registers the PM Engine to receive
 * events when a guest domain's DR service registers with LDoms.
 */
void *
ldm_register_domain_dr_reg_event(ldm_err_t *results)
{
	pm_domain_event_info_t	*new_event, *event;

	CHAIN_ITER(pm_domain_list.domain_event_list, event, link) {
		if (event->event_type == PM_DOMAIN_DR_REG_EVENT) {
			results->ldme_errs = LDME_OP_EEXISTS;
			return (NULL);
		}
	}

	new_event = Xmalloc(sizeof (pm_domain_event_info_t));
	new_event->event_type = PM_DOMAIN_DR_REG_EVENT;
	new_event->handle = new_event;

	CHAIN_ADD(pm_domain_list.domain_event_list, new_event, link);
	pm_domain_dr_reg_event_on = true;
	results->ldme_errs = LDME_OP_EOK;
	return ((void *)new_event);
}

/*
 * ldm_register_domain_down_event - Registers the PM Engine to receive
 * events when a guest domain is no longer running.
 */
void *
ldm_register_domain_down_event(ldm_err_t *results)
{
	pm_domain_event_info_t	*new_event, *event;

	CHAIN_ITER(pm_domain_list.domain_event_list, event, link) {
		if (event->event_type == PM_DOMAIN_DOWN_EVENT) {
			results->ldme_errs = LDME_OP_EEXISTS;
			return (NULL);
		}
	}

	new_event = Xmalloc(sizeof (pm_domain_event_info_t));
	new_event->event_type = PM_DOMAIN_DOWN_EVENT;
	new_event->handle = new_event;

	CHAIN_ADD(pm_domain_list.domain_event_list, new_event, link);
	results->ldme_errs = LDME_OP_EOK;
	return ((void *)new_event);
}

void *
ldm_register_resource_config_change_event(pm_res_type_t res_type,
    ldm_err_t *results)
{
	pm_domain_event_info_t	*new_event, *event;

	/* This check is in until other resources are supported. */
	if (res_type != PM_RES_CPU) {
		results->ldme_errs = LDME_OP_EUNSUPPORTED;
		return (NULL);
	}

	CHAIN_ITER(pm_domain_list.domain_event_list, event, link) {
		if ((event->event_type == PM_RES_EVENT) &&
		    (event->res_type == res_type)) {
			results->ldme_errs = LDME_OP_EEXISTS;
			return (NULL);
		}
	}

	new_event = Xmalloc(sizeof (pm_domain_event_info_t));
	new_event->event_type = PM_RES_EVENT;
	new_event->res_type = res_type;
	new_event->handle = new_event;

	CHAIN_ADD(pm_domain_list.domain_event_list, new_event, link);

	if (res_type == PM_RES_CPU)
		pm_cpu_events_on = true;

	results->ldme_errs = LDME_OP_EOK;
	return ((void *)new_event);
}

int
ldm_unregister_non_util_events(void *handle, ldm_err_t *results)
{
	pm_domain_event_info_t	*event;

	CHAIN_ITER(pm_domain_list.domain_event_list, event, link) {
		if (event->handle == handle)
			break;
	}

	if (event != NULL) {
		switch (event->event_type) {
		case PM_DOMAIN_CONFIG_EVENT:
			pm_domain_events_on = false;
			break;
		case PM_RES_EVENT:
			if (event->res_type == PM_RES_CPU)
				pm_cpu_events_on = false;
			break;
		case PM_FMA_EVENT:
			pm_fma_events_on = false;
			break;
		case PM_DOMAIN_DR_REG_EVENT:
			pm_domain_dr_reg_event_on = false;
			break;
		}
		CHAIN_UNLINK(pm_domain_list.domain_event_list, event, link);
		Xfree(event);
	} else {
		results->ldme_errs = LDME_OP_EUNAVAIL;
		return (-1);
	}

	results->ldme_errs = LDME_OP_EOK;
	return (0);
}

void *
ldm_register_fma_event(pm_res_type_t res_type, ldm_err_t *results)
{
	pm_domain_event_info_t	*new_event, *event;

	/* This check is in until other resources are supported. */
	if (res_type != PM_RES_CPU) {
		results->ldme_errs = LDME_OP_EUNSUPPORTED;
		return (NULL);
	}

	CHAIN_ITER(pm_domain_list.domain_event_list, event, link) {
		if ((event->event_type == PM_FMA_EVENT) &&
		    (event->res_type == res_type)) {
			results->ldme_errs = LDME_OP_EEXISTS;
			return (NULL);
		}
	}

	new_event = Xmalloc(sizeof (pm_domain_event_info_t));
	new_event->event_type = PM_FMA_EVENT;
	new_event->res_type = res_type;
	new_event->handle = new_event;

	CHAIN_ADD(pm_domain_list.domain_event_list, new_event, link);

	pm_fma_events_on = true;

	results->ldme_errs = LDME_OP_EOK;
	return ((void *)new_event);
}

void
send_pm_dom_event(ldom_t *ldp, uint_t flags, ldm_err_t *result)
{
	pm_domain_event_info_t	*event;

	CHAIN_ITER(pm_domain_list.domain_event_list, event, link) {
		if (flags == GUEST_DOMAIN_DOWN) {
			if (event->event_type == PM_DOMAIN_DOWN_EVENT)
				break;
		} else {
			if (event->event_type == PM_DOMAIN_CONFIG_EVENT)
				break;
		}
	}

	if (event != NULL) {
		pm_domain_config_change_notify(event->handle, (srcid_t)ldp,
		    flags, 0, NULL, 0, result);
	}
}

void
send_pm_cpu_event(ldom_t *ldp, uint_t flags, pm_resid_t *resids,
    int nresids, ldm_err_t *result)
{
	pm_domain_event_info_t	*event;

	ASSERT(pm_cpu_events_on);

	CHAIN_ITER(pm_domain_list.domain_event_list, event, link) {
		if ((event->event_type == PM_RES_EVENT) ||
		    (event->res_type == PM_RES_CPU)) {
			break;
		}
	}

	pm_domain_config_change_notify(event->handle, (srcid_t)ldp,
	    flags, PM_RES_CPU, resids, nresids, result);
}
