/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <thread.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <dlfcn.h>
#include <ctype.h>
#include "fatal.h"

#include "zprintf.h"
#include "ds_impl.h"
#include "support.h"
#include "svcs.h"
#include "allocate.h"
#include "libmgr.h"
#include "ldom.h"
#include "ldomconfigure.h"
#include "response.h"
#include "vcpu.h"
#include "core_md.h"
#include "pm/pm_shared.h"
#include "pm/pm_caps.h"
#include "pm/pm_policy.h"
#include "pm/pm_internal.h"


/* Commands */
#define	PMTEST_CMD_CLOCK		"clock "
#define	PMTEST_CMD_CORE			"core "
#define	PMTEST_CMD_DEFRAG		"defrag "
#define	PMTEST_CMD_DIR			"dir "
#define	PMTEST_CMD_DRCPU		"drcpu "
#define	PMTEST_CMD_PMCPU		"pmcpu "
#define	PMTEST_CMD_FMA			"fma "
#define	PMTEST_CMD_MEMORY		"mem "
#define	PMTEST_CMD_MOVESTRAND		"movestrand "
#define	PMTEST_CMD_PMSTAT		"pmstat "
#define	PMTEST_CMD_POLICY		"policy "
#define	PMTEST_CMD_PRI			"pri "
#define	PMTEST_CMD_QUIT			"quit"
#define	PMTEST_CMD_STRAND		"strand "
#define	PMTEST_CMD_UTIL			"util "
#define	PMTEST_CMD_WM			"wm "

/* Sub commands */
#define	PMTEST_SUBCMD_ADD		"add "
#define	PMTEST_SUBCMD_CLEAR		"clear"
#define	PMTEST_SUBCMD_CLEARSP		"clear "
#define	PMTEST_SUBCMD_CPU		"cpu "
#define	PMTEST_SUBCMD_DISABLE		"disable "
#define	PMTEST_SUBCMD_ENABLE		"enable "
#define	PMTEST_SUBCMD_GET		"get"
#define	PMTEST_SUBCMD_GETGROUPING	"getgrouping"
#define	PMTEST_SUBCMD_GETSP		"get "
#define	PMTEST_SUBCMD_GETRANGES		"getranges "
#define	PMTEST_SUBCMD_GETSTATES		"getstates"
#define	PMTEST_SUBCMD_GETSTEPS		"getsteps "
#define	PMTEST_SUBCMD_HIGH		"high "
#define	PMTEST_SUBCMD_IN		"in "
#define	PMTEST_SUBCMD_LOW		"low "
#define	PMTEST_SUBCMD_MEM		"mem "
#define	PMTEST_SUBCMD_MOVE		"move "
#define	PMTEST_SUBCMD_RM		"rm "
#define	PMTEST_SUBCMD_RUN		"run"
#define	PMTEST_SUBCMD_SET		"set"
#define	PMTEST_SUBCMD_SETAVG		"setavg "
#define	PMTEST_SUBCMD_SETSP		"set "
#define	PMTEST_SUBCMD_STATESP		"state "
#define	PMTEST_SUBCMD_RETIRE		"retire"
#define	PMTEST_SUBCMD_UNRETIRE		"unretire"

/* Responses */
#define	PMTEST_RESP_ACK			"ACK"
#define	PMTEST_RESP_NAK			"NAK"
#define	PMTEST_RESP_PASS		"PASS\n"
#define	PMTEST_RESP_PASS_NOCR		"PASS"
#define	PMTEST_RESP_FAIL		"FAIL\n"

/* Misc */
#define	BUF_SZ		1000
#define	INVALID_ID	-1
#define	NUM_SOURCES	10
#define	MEM_GROUPING	"memory-grouping"
#define	MEM_REGION	"memory-region"
#define	RESP_BUF_SZ	5000

#define	PM_CHIPS	16
#define	PM_DEBUG_PREFIX	"pm_testlib"
#define	PM_DEBUG 	zprintf
#define	PM_RES_MAX	512

#define	PMTEST_CLOCK	'D'
#define	PMTEST_CORE	'C'
#define	PMTEST_MEMORY	'M'
#define	PMTEST_PORT	2222
#define	PMTEST_STRAND	'S'

#define	PMTEST_DISABLE	(uint16_t)0
#define	PMTEST_ENABLE	(uint16_t)1

/* Variables */
static srcid_t 		src_ids[NUM_SOURCES] = { 0 };
static ldm_resutils_t 	src_resutils[NUM_SOURCES] = { 0 };
static mutex_t 		pm_seq_mutex;
static pmi_policy_t 	pm_local_policy = PM_POLICY_PERFORMANCE;

/*
 * FIXME: The index multiplier should be removed at some point and logic
 * cleaned up.
 */
static char *test_cookies[PM_RES_MAX * 3];

static int pm_policy_override = 0;

/*
 * FIXME: The index multiplier should be removed at some point and logic
 * cleaned up.
 */
static uint32_t g_res[PM_RES_MAX * 6];
static uint32_t *g_comp = &g_res[PM_RES_MAX * 3];

/* Debug interface entry points */
bool_t do_pm_caps(response_t *resp);
int cli_get_cpu_states(int count, uint32_t *cpuids);
int cli_set_cpu_states(int count, uint32_t *cpuids, int dir, int do_rcfg);
int pmtest_parse_all(char *p, char type, int *c);
void pm_reg(void);
void pm_unreg(void);


/*
 * *************************************************************************
 * Miscellaneous helper routines
 * *************************************************************************
 */

char *
pmi_err_str(int err)
{
	switch (err) {
		case PS_SUCCESS:
			return ("success");
		case PS_EINVAL:
			return ("invalid argument");
		case PS_EILLEGAL:
			return ("illegal request");
		case PS_TIMEDOUT:
			return ("request timed out");
		case PS_FAILURE:
			return ("request failed");
		default:
			return ("invalid error");
	}
}

void
pm_f(void *data)
{
	pack_f_t	*bufp = data;
	void (*fn)(char *, int);

	fn = (void (*)(char *, int))bufp->argv[0];
	fn(*((char **)bufp->argv[1]), *((int *)bufp->argv[2]));
	mutex_unlock(& pm_seq_mutex);
}

static void
pmtest_q(void (*f)(char *, int), char *cmd, int fd)
{
	pack_f_t	*bufp;

	bufp = Xmalloc(sizeof (pack_f_t));
	bufp->argc = 3;
	bufp->argv = Xmalloc(bufp->argc * sizeof (uint64_t));
	bufp->argv[0] = (uint64_t)f;
	bufp->argv[1] = (uint64_t)&cmd;
	bufp->argv[2] = (uint64_t)&fd;
	mutex_lock(& pm_seq_mutex);
	ldm_enqueue_work(pm_f, (void *)bufp);
}

static void
set_source_util(srcid_t src_id, double avg_util, pm_res_type_t res_type)
{
	int src_idx;

	for (src_idx = 0; src_idx < NUM_SOURCES; src_idx++)
		if (src_ids[src_idx] == NULL || src_ids[src_idx] == src_id) {
			src_ids[src_idx] = src_id;
			src_resutils[src_idx].ldmr_type = res_type;
			src_resutils[src_idx].ldmr_rescnt = 0;
			src_resutils[src_idx].ldmr_avg_util = avg_util;
			break;
		}
}

static void
clear_source_util(srcid_t src_id)
{
	int	src_idx;

	for (src_idx = 0; src_idx < NUM_SOURCES; src_idx++)
		if (src_ids[src_idx] == src_id)
			src_ids[src_idx] = 0;
}

static double
get_source_util(srcid_t src_id)
{
	int src_idx;

	for (src_idx = 0; src_idx < NUM_SOURCES; src_idx++) {
		if (src_ids[src_idx] == src_id) {
			return (src_resutils[src_idx].ldmr_avg_util);
		}
	}

	PM_DEBUG("%s: Util value not set yet for this source\n",
	    PM_DEBUG_PREFIX);
	return (-1);
}

pm_mem_t *
find_grouping_by_id(char *memstr) {
	uint64_t	mem_id;
	pm_mem_t	*memp = NULL;

	mem_id = strtoll(memstr, NULL, 10);
	CHAIN_ITER(pm_mem_list.mem_list, memp, link) {
		if (memp->pm_mem_resid.id == mem_id) {
			return (memp);
		}
	}

	PM_DEBUG("%s: Memory grouping %"PRIu64" not found\n", PM_DEBUG_PREFIX,
	    mem_id);
	return (NULL);
}

static int
get_util_dir(srcid_t src_id)
{
	int	src_idx;

	for (src_idx = 0; src_idx < NUM_SOURCES; src_idx++) {
		if (src_ids[src_idx] == src_id) {
			return (src_resutils[src_idx].ldmr_util_dir);
		}
	}

	PM_DEBUG("%s: Util data not set in test harness.\n", PM_DEBUG_PREFIX);
	return (-1);
}

int
get_domain_utils(srcid_t src_id, pm_res_type_t res_type,
    ldm_resutils_t **utils, ldm_err_t *results)
{
	ldom_t		*ldp = NULL;
	pm_mem_t	*pmp = NULL;
	ldm_resutils_t  *res_utils, *ldmu;
	ldm_util_t	*lup;
	vcpu_t		*vcpu;
	size_t		sz;
	int		i;

	/*
	 * Get util info for resource types that don't bind to domains and
	 * return. Currently, memory is the only power managable resource
	 * type that does not specifically bind to domains.
	 */
	if (res_type == PM_RES_MEM) {
		CHAIN_ITER(pm_mem_list.mem_list, pmp, link) {
			if ((pm_mem_t *)src_id == pmp)
				break;
		}

		if (pmp == NULL) {
			DBG(zprintf("%s: memory grouping 0x%p was not found "
			    "in pm_mem_list.\n", PM_DEBUG_PREFIX, src_id));
			return (-1);
		}

		sz = sizeof (ldm_resutils_t);
		res_utils = Xmalloc(sz);
		ldmu = &pmp->pm_mem_resutils;
		bcopy((void *)ldmu, (void *)res_utils, sz);

		/* retrieve resource util data */
		res_utils->ldmr_rescnt = 1;
		res_utils->ldmr_util_p = Xcalloc(res_utils->ldmr_rescnt,
		    ldm_util_t);
		res_utils->ldmr_avg_util = get_source_util(src_id);
		res_utils->ldmr_util_dir = get_util_dir(src_id);
		res_utils->ldmr_util_p->ldmu_util = get_source_util(src_id);
		res_utils->ldmr_util_p->ldmu_resid.pi_res_type = PM_RES_MEM;
		res_utils->ldmr_util_p->ldmu_resid.pi_res_name =
		    Xstrdup(MEM_GROUPING);
		res_utils->ldmr_util_p->ldmu_resid.u.mem.pi_id_val =
		    ((pm_mem_t *)src_id)->pm_mem_resid.id;

		*utils = res_utils;
		return (PS_SUCCESS);
	}

	/*
	 * Get util info for resource types that bind to domains.
	 * Currently, CPUs are the only power managable resource type that
	 * binds to domains.
	 */
	ASSERT(res_type == PM_RES_CPU);

	CHAIN_ITER(zeus.ldoms, ldp, link) {
	if ((ldom_t *)src_id == ldp)
		break;
	}

	if (ldp == NULL) {
		DBG(zprintf("%s: no such domain by domid 0x%p\n",
		    PM_DEBUG_PREFIX, src_id));
		return (-1);
	}

	/* copy out per domain res util data */
	sz = sizeof (ldm_resutils_t);
	res_utils = Xmalloc(sz);
	ldmu = &ldp->pm_cpu_resutils;
	bcopy((void *)ldmu, (void *)res_utils, sz);

	/* t = gethrtime();  - lint fix */
	res_utils->ldmr_avg_util = get_source_util(src_id);
	res_utils->ldmr_util_dir = get_util_dir(src_id);

	/* retrieve per resource util data */
	res_utils->ldmr_rescnt = CHAIN_LENGTH(ldp->bound.vcpus);
	res_utils->ldmr_util_p = Xcalloc(res_utils->ldmr_rescnt, ldm_util_t);
	i = 0;
	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		lup = &res_utils->ldmr_util_p[i];
		if (get_core_md_cpu_status(vcpu->cpu->phys_id,
		    CORE_MD_PM_STATUS) == PM_MANAGED)
			lup->ldmu_util = 0.0;
		else
			lup->ldmu_util = get_source_util(src_id);
		lup->ldmu_resid.pi_res_type = PM_RES_CPU;
		lup->ldmu_resid.pi_res_name = Xstrdup(CPU_RES_NAME);
		lup->ldmu_resid.u.cpu.pi_id_val = vcpu->cpu->phys_id;
		i++;
	}

	*utils = res_utils;
	return (PS_SUCCESS);
}

static void
set_util_dir(srcid_t src_id, pm_util_dir_t dir)
{
	int src_idx;

	for (src_idx = 0; src_idx < NUM_SOURCES; src_idx++) {
		if (src_ids[src_idx] == src_id)
			src_resutils[src_idx].ldmr_util_dir = dir;
	}
}

int
resume_util_change_notify(void *handle, ldm_err_t *results)
{
	PM_DEBUG("resume_util_change_notify callback called\n");
	return (0);
}

int
get_local_policy(pmi_policy_t *pm)
{
	*pm = pm_local_policy;
	return (PS_SUCCESS);
}

/*
 * find_region_in_mem_grouping - Searches memory groupings to fine the parent
 * grouping that contains the memory region that was passed in.
 *
 * Input parms:
 *   phys_addr - Physical address of the memory region in question.
 *   grp_cnt - Total number of memory groupings.
 * Output parms:
 *   grp_idx - The memory grouping id for the grouping containing the memory
 *     region.
 *   pmnp - PM node for the memory grouping containing the memory region.
 *
 *   Returns true if the region exists within the defined memory groupings.
 *   Otherwise, false.
 */
bool_t
find_region_in_mem_grouping(int *grp_idx, pm_node_t **pmnp, uint64_t phys_addr,
    int group_cnt)
{
	char		*cookie;
	char		*id_str = NULL;
	int		mem_grp_id = 0;
	pm_node_t	*child_np = NULL;
	pm_node_t	*region_np = NULL;
	uint64_t	mem_base = 0;

	/*
	 * Search thru the pm_nodes to find a matching memory
	 * grouping node for each entry in test_cookies.
	 * Any invalid entry will make the command fail.
	 */
	for (*grp_idx = 0; *grp_idx < group_cnt; (*grp_idx)++) {
		PM_DEBUG("%s: mem get cookie: %s\n", PM_DEBUG_PREFIX,
		    test_cookies[*grp_idx]);

		/* strdup so test_cookies[] is not altered via strtok */
		cookie = strdup(test_cookies[*grp_idx]);
		id_str = strtok(cookie, " ");
		id_str = strtok(NULL, " ");
		mem_grp_id = atoi(id_str);
		PM_DEBUG("%s: Searching memory grouping id %d\n",
		    PM_DEBUG_PREFIX, mem_grp_id);
		free(cookie);

		*pmnp = pmi_find_node_by_type(NULL, MEM_GROUPING, mem_grp_id,
		    INVALID_ID);
		if (*pmnp != NULL) {
			for (child_np = (*pmnp)->pn_child; child_np;
			    child_np = child_np->pn_sibling) {
				region_np = pmi_find_node_by_type(NULL,
				    MEM_REGION, child_np->pn_nodeid,
				    mem_grp_id);

				mem_base =
				    region_np->pn_rcfg->\
				    u.mem.mem_reg.pi_reg_base;

				if (phys_addr == mem_base) {
					PM_DEBUG("%s: memory grouping %d "
					    "contains address\n",
					    PM_DEBUG_PREFIX, mem_grp_id);
					return (true);
				}
			}
		}
	}
	return (false);
}

static void
pmtest_make_cookies(uint32_t *g, char type, int *res_cnt)
{
	int		res_idx;
	char		cookie[10];
	pm_node_t	*pmnp = NULL;

	*res_cnt = 0;

	if (type != PMTEST_MEMORY) {
		for (res_idx = 0; res_idx < PM_RES_MAX*3; res_idx++) {
			if (g[res_idx] != 1)
				continue;
			sprintf(cookie, "%c %d", type, res_idx-PM_RES_MAX);
			if (test_cookies[*res_cnt])
				free(test_cookies[*res_cnt]);
			test_cookies[*res_cnt] = strdup(cookie);
			PM_DEBUG("CPU cookie is %s; g_res idx=%d\n",
			    test_cookies[*res_cnt], res_idx);
			(*res_cnt)++;
		}
	} else {
		for (res_idx = 0; res_idx < PM_RES_MAX; res_idx++) {
			pmnp = pmi_find_node_by_type(NULL, MEM_GROUPING,
			    *res_cnt, INVALID_ID);
			if (pmnp) {
				sprintf(cookie, "%c %d", type, *res_cnt);
				if (test_cookies[*res_cnt])
					free(test_cookies[*res_cnt]);
				test_cookies[*res_cnt] = strdup(cookie);
				PM_DEBUG("Memory Grouping cookie is %s; g_res"
				    " idx=%d\n", test_cookies[*res_cnt],
				    res_idx);
				(*res_cnt)++;
			} else {
				break;
			}
		}
	}
	if (test_cookies[*res_cnt])
		free(test_cookies[*res_cnt]);
	test_cookies[*res_cnt] = NULL;
}

/* Handle "a..b", "..b", "a.." and "a" syntax */
static int
pmtest_parse_sub(char *in_strp)
{
	int start = -1;
	int end = -1;
	char *dd;
	int i;
	int tmp;

	dd = strstr(in_strp, "..");

	/* Only one value */
	if (!dd) {
		start = atoi(in_strp);
		end = atoi(in_strp);
	} else {

		/* Handle ".." at beginning of string */
		if (dd == in_strp)
			start = 0;

		/* Handle ".." at end of string */
		dd += 2;
		if (strlen(dd) == 0 || *dd == ' ' || *dd == 13)
			end = PM_RES_MAX - 1;

		/* Handle "a" in "a.." */
		if (start < 0)
			start = atoi(in_strp);

		/* Handle "b" in "..b" */
		if (end < 0)
			end = atoi(dd);

		/* Swap if s > e */
		if (start > end) {
			tmp = end;
			end = start;
			start = tmp;
		}

		/* Check if out of bounds */
		if (((start + PM_RES_MAX) < 0) || (end > (2 * PM_RES_MAX)))
			return (1);
	}

	/* Fill in relevant resources */
	PM_DEBUG("pmtest_parse_sub(): start=%d end=%d\n", start, end);
	for (i = start; i <= end; i++)
		/* Real resources offset 0 is g_res[PM_RES_MAX] */
		g_res[i + PM_RES_MAX] = 1;

	return (0);
}

static void
pmtest_translate_array(uint32_t *go, uint32_t *gn, int c)
{
	int goidx, gnidx = 0;

	/* Traverse array */
	for (goidx = 0; goidx < c; goidx++) {
		/* If cookie index is flagged */
		if (go[goidx] == 1) {
			/* Transfer from index entry to entry set to index */
			PM_DEBUG("g[%d] -> g[%d]\n", goidx, gnidx);
			gn[gnidx] = goidx - PM_RES_MAX;
			gnidx++;
		}
	}
}

/*
 * Input: strp - pointer to a string to parse
 *        type - Type of PM resource 'D', 'C', 'S' or 'M'
 * Output: num - Number of resources found from input string
 * Function return: 0 - OK; 1 - parse failure; 2 - Invalid parameter
 *
 *
 * handle "a,b,c",  "..b", "a..", and "a" and syntax
 */
int
pmtest_parse_all(char *in_strp, char type, int *num)
{
	char *tmp_strp = in_strp;
	int ok = 0;
	int i, rc;

	memset(&g_res, 0x00, sizeof (g_res));
	PM_DEBUG("pmtest_parse_all(ENTER): string=%s\n", in_strp);

	/*
	 * strcspn() in the usage below, only makes sure the string begins with
	 * one of the characters in s2, so it really does not detect all illegal
	 * strings.  Thus, 'core state ,)2'  gets interpreted as 'core state 0'.
	 * So the user can enter an invalid parameter and the command might
	 * still work, but giving results not expected.  This code will not be
	 * changed to make it more defensive, since it is only a test interface,
	 * and there is a test interface spec that defines the proper usage.
	 *
	 * Note: The '-' char was removed from the s2 string since it is not
	 * accounted for in this parsing code.
	 */

	rc = strcspn(in_strp, "0123456789,.");
	PM_DEBUG("pmtest_parse_all(): strcspn=%d cmd=%s\n", rc, in_strp);
	if (strcspn(in_strp, "0123456789,.")) {

		PM_DEBUG("Invalid chars at start of: %s\n", in_strp);
		return (2);
	}

	/* Reject a string with "...", ",," or "-" */
	if (strstr(in_strp, "...") || strstr(in_strp, ",,") ||
	    strstr(in_strp, "-")) {
		PM_DEBUG("Illegal range characters in: %s\n", in_strp);
		return (2);
	}

	/*
	 * This while loop just makes sure that if there is a '.' in the
	 * string, it is followed by another '.'.  There can be more than
	 * 1 '..' pairs in the string.  This is just a sanity check to
	 * enforce no single '.' appears without a 2nd '.'.
	 * Ex: "foo..bar..blah.." is a valid string that passes this while.
	 */
	while (tmp_strp = strstr(tmp_strp, ".")) {
		tmp_strp++;
		/* Process next char after '.' to find another '.' */
		if (!tmp_strp) {
			/* EOL */
			PM_DEBUG("Illegal EOL; cmd=%s\n", in_strp);
			return (2);
		}
		if (*tmp_strp != '.') {
			/* Not a '.' */
			PM_DEBUG("Missing a extra '.'; cmd=%s\n", in_strp);
			return (2);
		}
		/* Done with this char so next */
		tmp_strp++;
	}

	/*
	 * Break up parameter into tokens that are separated by ",".
	 * If there are no ',' in the string, then there
	 * will be only 1 token and in_strp points to it.  Note
	 * that the token could be '..3' or '0..2", and that
	 * case is handled in parse_sub().
	 */
	in_strp = strtok(in_strp, ",");
	while (in_strp) {
		ok = 1;
		if (pmtest_parse_sub(in_strp) != 0) {
			PM_DEBUG("Failed pmtest_parse_sub\n");
			return (1);
		}
		in_strp = strtok(NULL, ",");
	}

	if (ok) {
		pmtest_make_cookies(g_res, type, num);
		PM_DEBUG("pmtest_make_cookies ok\n");
		pmtest_translate_array(&g_res[0], g_comp, PM_RES_MAX * 3);
		memcpy(&g_res[0], g_comp, PM_RES_MAX * 3);

		/*
		 * Don't need to check for a chip CLOCK out of bounds
		 * here because that will be done back in the handle_clock
		 * function.
		 */
		for (i = 0; i < *num; i++)
			if ((g_res[i] < 0) ||
			    ((type == PMTEST_CORE) &&
			    (g_res[i] >= PM_RES_MAX / 8)) ||
			    ((type == PMTEST_STRAND) &&
			    (g_res[i] >= PM_RES_MAX))) {
				PM_DEBUG("Resource %d is out of bounds\n",
				    g_res[i]);
				return (1);
			}
	}

	return (0);
}

/*
 * *************************************************************************
 * Command handler routines
 * *************************************************************************
 */

/*
 * Handle the PMTEST_CMD_UTIL command.
 *
 * Sub commands are :
 * 	PMTEST_SUBCMD_GET <ldom name|memory grouping id> : Get the current ldom
 *				or memory grouping utilization settings.
 * 	PMTEST_SUBCMD_CLEAR <ldom name|memory grouping id> : Clear user
 *				defined utilization data.
 * 	PMTEST_SUBCMD_SETAVG <ldom name|memory grouping id> <avg_util> :  Set
 *				utilization values. "avg_util" is <0..100>
 *
 * Usage: util [cpu|mem] <clear|get|setavg> <ldom name|memory grouping id>
 * 	If [cpu|mem] is not specified, cpu is the default.
 */
static void
pmtest_handle_util(char *cmd, int fd)
{
	double		util;
	ldom_t		*ldp = NULL;
	char		buffer[100];
	char		*subcmd;
	int		res_type;
	uint64_t	mem_id = 0;
	pm_mem_t	*mem_grp = NULL;

	cmd += strlen(PMTEST_CMD_UTIL);
	subcmd = cmd;

	/* Check if "cpu" or "mem" was entered. Advance cmd pointer if so. */
	if (strncmp(subcmd, PMTEST_SUBCMD_MEM,
	    strlen(PMTEST_SUBCMD_MEM)) == 0) {
		res_type = PM_RES_MEM;
		cmd += strlen(PMTEST_SUBCMD_MEM);
	} else if (strncmp(subcmd, PMTEST_SUBCMD_CPU,
	    strlen(PMTEST_SUBCMD_CPU)) == 0) {
		res_type = PM_RES_CPU;
		cmd += strlen(PMTEST_SUBCMD_CPU);
	} else {
		res_type = PM_RES_CPU;
	}

	/* util [cpu|mem] get <ldom name|memory grouping id> */
	if (strncmp(cmd, PMTEST_SUBCMD_GET,
	    strlen(PMTEST_SUBCMD_GET)) == 0) {

		cmd += strlen(PMTEST_SUBCMD_GETSP);
		subcmd = strstr(cmd, "\r");
		if (subcmd)
			*subcmd = 0;

		if (res_type == PM_RES_CPU) {
			ldp = find_ldom_by_name(cmd);
			if (!ldp)
				goto fail;

			util = get_source_util((srcid_t)ldp);
		} else if (res_type == PM_RES_MEM) {
			mem_grp = find_grouping_by_id(cmd);
			if (!mem_grp) {
				mem_id = strtoll(cmd, NULL, 10);
				goto fail;
			}

			util = get_source_util(mem_grp);
		}

		PM_DEBUG("Get Util=%f\n", util);
		sprintf(buffer, "%s %.0f\n", PMTEST_RESP_PASS_NOCR, util);
		(void) write(fd, buffer, strlen(buffer));

	/* util [cpu|mem] clear <ldom|memory grouping> */
	} else if (strncmp(cmd, PMTEST_SUBCMD_CLEAR,
	    strlen(PMTEST_SUBCMD_CLEAR)) == 0) {

		cmd += strlen(PMTEST_SUBCMD_CLEARSP);
		subcmd = strstr(cmd, "\r");
		if (subcmd)
			*subcmd = 0;

		if (res_type == PM_RES_CPU) {
			ldp = find_ldom_by_name(cmd);
			if (!ldp)
				goto fail;

			clear_source_util((srcid_t)ldp);
		} else if (res_type == PM_RES_MEM) {
			mem_grp = find_grouping_by_id(cmd);
			if (!mem_grp) {
				mem_id = strtoll(cmd, NULL, 10);
				goto fail;
			}

			/* Clear out grouping id data */
			mem_id = mem_grp->pm_mem_resid.id;
			src_ids[mem_id] = 0;
		}

	/* util [cpu|mem] setavg <ldom|memory grouping> <avg_util> */
	} else if (strncmp(cmd, PMTEST_SUBCMD_SETAVG,
	    strlen(PMTEST_SUBCMD_SETAVG)) == 0) {

		cmd += strlen(PMTEST_SUBCMD_SETAVG);
		subcmd = strstr(cmd, " ");
		if (subcmd)
			*subcmd = 0;

		if (res_type == PM_RES_CPU) {
			ldp = find_ldom_by_name(cmd);
			if (!ldp)
				goto fail;
		} else if (res_type == PM_RES_MEM) {
			mem_grp = find_grouping_by_id(cmd);
			if (!mem_grp) {
				mem_id = strtoll(cmd, NULL, 10);
				goto fail;
			}
		}

		/* util setavg <ldom> <avg>. Ex: 'util setavg ldg1 55' */
		cmd = subcmd + 1;
		util = -1;
		if (cmd)
			util = atof(cmd);

		if (res_type == PM_RES_CPU) {
			PM_DEBUG("Set Avg CPU util=%f\n", util);
			set_source_util((srcid_t)ldp, util, res_type);
		} else if (res_type == PM_RES_MEM) {
			PM_DEBUG("Set Avg memory grouping util=%f\n", util);
			set_source_util((srcid_t)mem_grp, util, res_type);
		}
	} else {
		PM_DEBUG("%s: Unexpected util sub cmd: %s\n",
		    PM_DEBUG_PREFIX, cmd);
		sprintf(buffer, "Unexpected util sub cmd: %s\n%s", cmd,
		    PMTEST_RESP_FAIL);
		write(fd, buffer, strlen(buffer));
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;

	}

	(void) write(fd, PMTEST_RESP_PASS, sizeof (PMTEST_RESP_PASS));
	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
	return;
fail:
	if (res_type == PM_RES_CPU) {
		sprintf(buffer, "Unknown LDOM: %s\n%s", cmd, PMTEST_RESP_FAIL);
		write(fd, buffer, strlen(buffer));
		PM_DEBUG("Unknown LDOM: %s\n", cmd);
	} else if (res_type == PM_RES_MEM) {
		sprintf(buffer, "Memory Grouping ID %"PRIu64" does not "
		    "exist\n%s", mem_id, PMTEST_RESP_FAIL);
		write(fd, buffer, strlen(buffer));
		PM_DEBUG("Memory Grouping ID %"PRIu64" does not "
		    "exist\n", mem_id);
	}
	(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
}

/*
 * Handle the PMTEST_CMD_WM command.
 *
 * Sub commands are :
 * 	PMTEST_SUBCMD_GET		: Return current watermark values.
 */
static void
pmtest_handle_wm(char *cmd, int fd)
{
	char *str = "50 70\n";

	cmd += strlen(PMTEST_CMD_WM);

	if (strncmp(cmd, PMTEST_SUBCMD_GET,
	    strlen(PMTEST_SUBCMD_GET)) == 0) {
		PM_DEBUG("%s: recv'd wm get cmd\n", PM_DEBUG_PREFIX);
		(void) write(fd, str, strlen(str));
	} else {
		PM_DEBUG("%s: Unexpected wm sub cmd : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
}

/*
 * Handle the PMTEST_CMD_PMCPU command allowing cpus to be PM'd like the
 * policy engine does via pmi_adjust_power.  This is useful if somebody
 * is trying to get specific physical cpus offline.  It doesn't protect
 * against offlining the last cpu so be careful.
 *
 * Sub commands are :
 * 	PMTEST_SUBCMD_ADD		: PM vcpu(s).
 * 	PMTEST_SUBCMD_RM		: unPM vcpu(s).
 */
static void
pmtest_handle_pmcpu(char *cmd, int fd)
{
	cli_result_t	r;
	char 		*str;
	pm_resid_t	ids[PM_RES_MAX] = { 0 };
	int		i, c = 0, rc;
	int		*s = NULL;

	cmd += strlen(PMTEST_CMD_PMCPU);

	/* UnPM an already power managed cpu (does unpark/DR in step */
	if (strncmp(cmd, PMTEST_SUBCMD_ADD,
	    strlen(PMTEST_SUBCMD_ADD)) == 0) {
		PM_DEBUG("%s: recv'd drcpu add cmd\n", PM_DEBUG_PREFIX);
		cmd += strlen(PMTEST_SUBCMD_ADD);

		/* Parses the cpu list */
		rc = pmtest_parse_all(cmd, PMTEST_STRAND, & c);
		if (rc == 0) {
			zprintf("pmtest_parse_all rc = %d\n", rc);
		}

		/* Builds the list of PM objects for the cpu list */
		for (i = 0; i < c; i++) {
			ids[i].pi_res_type = PM_RES_CPU;
			ids[i].pi_res_name = "cpu";
			ids[i].u.cpu.pi_id_val = g_res[i];
			zprintf("DR CPU %"PRIu64"\n", ids[i].u.cpu.pi_id_val);
		}

		/* Power up the cpus */
		r = pmi_adjust_power(PM_RES_CPU, PCO_ADD_RESOURCE,
		    PPS_FULL_POWER, c, &ids[0], &s, 0);

		/* Caller must free s */
		Xfree(s);
		if (r == 0)
			str = "PASS\n";
		else {
			PM_DEBUG("cli_vcpu_add rc %d\n", r);
			str = "FAIL\n";
		}
		(void) write(fd, str, strlen(str));

	/* PM an unpower managed cpu */
	} else if (strncmp(cmd, PMTEST_SUBCMD_RM,
	    strlen(PMTEST_SUBCMD_RM)) == 0) {
		PM_DEBUG("%s: recv'd pmcpu rm cmd\n", PM_DEBUG_PREFIX);
		cmd += strlen(PMTEST_SUBCMD_RM);

		/* Parses the cpu list */
		rc = pmtest_parse_all(cmd, PMTEST_STRAND, & c);
		if (rc == 0) {
			zprintf("pmtest_parse_all rc = %d\n", rc);
		}

		/* Builds the list of PM objects for the cpu list */
		for (i = 0; i < c; i++) {
			ids[i].pi_res_type = PM_RES_CPU;
			ids[i].pi_res_name = "cpu";
			ids[i].u.cpu.pi_id_val = g_res[i];
			zprintf("PM CPU %"PRIu64"\n", ids[i].u.cpu.pi_id_val);
		}

		/* Power up the cpus */
		r = pmi_adjust_power(PM_RES_CPU, PCO_REMOVE_RESOURCE,
		    PPS_LOWEST_POWER, c, &ids[0], &s, 0);

		/* Caller must free s */
		Xfree(s);
		if (r == 0)
			str = "PASS\n";
		else
			str = "FAIL\n";
		(void) write(fd, str, strlen(str));
	} else {
		zprintf("%s: Unexpected wm sub cmd : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
}


/*
 * Handle the PMTEST_CMD_DRCPU command.
 *
 * Sub commands are :
 * 	PMTEST_SUBCMD_ADD		: Add vcpu(s).
 * 	PMTEST_SUBCMD_RM		: Remove vcpu(s).
 */
static void
pmtest_handle_drcpu(char *cmd, int fd)
{
	response_t *resp;
	cli_result_t r;
	char *str;
	pm_resid_t ids[PM_RES_MAX] = { 0 };
	int status[PM_RES_MAX] =  { 0 };
	ldm_err_t res = { 0 };
	int i, c = 0, rc;

	init_response(& resp);

	cmd += strlen(PMTEST_CMD_DRCPU);

	if (strncmp(cmd, PMTEST_SUBCMD_ADD,
	    strlen(PMTEST_SUBCMD_ADD)) == 0) {
		PM_DEBUG("%s: recv'd drcpu add cmd\n", PM_DEBUG_PREFIX);
		cmd += strlen(PMTEST_SUBCMD_ADD);
		rc = pmtest_parse_all(cmd, PMTEST_STRAND, & c);
		if (rc == 0) {
			PM_DEBUG("cli_set_cpu_states rc = %d\n", rc);
		}

		for (i = 0; i < c; i++) {
			ids[i].pi_res_type = PM_RES_CPU;
			ids[i].pi_res_name = "cpu";
			ids[i].u.cpu.pi_id_val = g_res[i];
			PM_DEBUG("DR CPU %"PRIu64"\n", ids[i].u.cpu.pi_id_val);
		}

		r = ldm_reconfigure_resources(PCO_ADD_RESOURCE, c, &ids[0],
		    &status[0], &res, 0);

		if (r == 0)
			str = "PASS\n";
		else {
			PM_DEBUG("cli_vcpu_add rc %d\n", r);
			str = "FAIL\n";
		}
		(void) write(fd, str, strlen(str));
	} else if (strncmp(cmd, PMTEST_SUBCMD_RM,
	    strlen(PMTEST_SUBCMD_RM)) == 0) {
		PM_DEBUG("%s: recv'd drcpu rm cmd\n", PM_DEBUG_PREFIX);
		cmd += strlen(PMTEST_SUBCMD_RM);
		PM_DEBUG("DR CPU %"PRIu64"\n", ids[0].u.cpu.pi_id_val);
		rc = pmtest_parse_all(cmd, PMTEST_STRAND, & c);
		if (rc == 0) {
			PM_DEBUG("cli_set_cpu_states rc = %d\n", rc);
		}

		for (i = 0; i < c; i++) {
			ids[i].pi_res_type = PM_RES_CPU;
			ids[i].pi_res_name = "cpu";
			ids[i].u.cpu.pi_id_val = g_res[i];
			PM_DEBUG("DR CPU %"PRIu64"\n", ids[i].u.cpu.pi_id_val);
		}

		r = ldm_reconfigure_resources(PCO_REMOVE_RESOURCE, c, &ids[0],
		    &status[0], &res, 0);
		if (r == 0)
			str = "PASS\n";
		else
			str = "FAIL\n";
		(void) write(fd, str, strlen(str));
	} else {
		PM_DEBUG("%s: Unexpected wm sub cmd : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
}

static int pmtest_i = -1;
extern void *(*find_pm_handle)(void);
void *
pmtest_find_pm_handle(void)
{
	return (&pmtest_i);
}

/*
 * Handle the PMTEST_CMD_DIR command.
 *
 * Sub commands are :
 * 	PMTEST_SUBCMD_HIGH <source>	: Generate a PM_UTIL_HIGH notification
 * 	PMTEST_SUBCMD_LOW <source>	: Generate a PM_UTIL_LOW notification
 * 	PMTEST_SUBCMD_IN <source>	: Generate a PM_UTIL_WITHIN notification
 * <source> is either 'ldom name' or 'memory grouping ID'.
 *
 * Usage: dir [cpu|mem] <high|low|in> <ldom name|memory grouping id>
 *	If [cpu|mem] is not specified, cpu is the default.
 */
static void
pmtest_handle_dir(char *cmd, int fd)
{
	pm_util_dir_t util_dir;
	ldom_t *ldp = NULL;
	int src_cnt = 0;
	int src_idx;
	int save_pm_local_policy;
	int (*save_pmi_get_policy)(pmi_policy_t *pm);
	void *(*hdl_fn)(void) = NULL;
	char *subcmd;
	int res_type;
	pm_mem_t *mem_grp = NULL;
	srcid_t	*src_id[NUM_SOURCES] = { 0 };
	char buf[BUF_SZ];

	cmd += strlen(PMTEST_CMD_DIR);
	subcmd = cmd;

	/* Check if "cpu" or "mem" was entered. */
	if (strncmp(subcmd, PMTEST_SUBCMD_MEM,
	    strlen(PMTEST_SUBCMD_MEM)) == 0) {
		res_type = PM_RES_MEM;
		cmd += strlen(PMTEST_SUBCMD_MEM);
	} else if (strncmp(subcmd, PMTEST_SUBCMD_CPU,
	    strlen(PMTEST_SUBCMD_CPU)) == 0) {
		res_type = PM_RES_CPU;
		cmd += strlen(PMTEST_SUBCMD_CPU);
	} else {
		res_type = PM_RES_CPU;
	}

	if (strncmp(cmd, PMTEST_SUBCMD_HIGH,
	    strlen(PMTEST_SUBCMD_HIGH)) == 0) {
		PM_DEBUG("%s: recv'd dir high command\n", PM_DEBUG_PREFIX);
		util_dir = PM_UTIL_HIGH;
		cmd += strlen(PMTEST_SUBCMD_HIGH);
	} else if (strncmp(cmd, PMTEST_SUBCMD_LOW,
	    strlen(PMTEST_SUBCMD_LOW)) == 0) {
		PM_DEBUG("%s: recv'd dir low command\n", PM_DEBUG_PREFIX);
		util_dir = PM_UTIL_LOW;
		cmd += strlen(PMTEST_SUBCMD_LOW);
	} else if (strncmp(cmd, PMTEST_SUBCMD_IN,
	    strlen(PMTEST_SUBCMD_IN)) == 0) {
		PM_DEBUG("%s: recv'd dir in command\n", PM_DEBUG_PREFIX);
		util_dir = PM_UTIL_WITHIN;
		cmd += strlen(PMTEST_SUBCMD_IN);
	} else {
		sprintf(buf, "Unexpected dir sub cmd : %s\n", cmd);
		(void) write(fd, buf, strlen(buf));
		PM_DEBUG("%s: Unexpected dir sub cmd : %s\n", PM_DEBUG_PREFIX,
		    cmd);
		goto fail;

	}

	/* Set direction for every ldom or grouping entered on CLI */
	cmd = strtok(cmd, ",\r");
	while (cmd) {
		if (res_type == PM_RES_CPU) {
			ldp = find_ldom_by_name(cmd);
			if (ldp) {
				set_util_dir((srcid_t)ldp, util_dir);

				/* get domain utilization */
				for (src_idx = 0; src_idx < NUM_SOURCES;
				    src_idx++) {
					if (src_ids[src_idx] != ldp)
						continue;
					if (&src_id[src_cnt])
						src_id[src_cnt] = (srcid_t)
						    &src_resutils[src_idx];
				}

				src_id[src_cnt] = (srcid_t)ldp;
				src_cnt++;
			} else {
				PM_DEBUG("%s: Domain %s not found\n",
				    PM_DEBUG_PREFIX, cmd);
				sprintf(buf, "Domain %s not found\n", cmd);
				(void) write(fd, buf, strlen(buf));
				goto fail;
			}
		} else if (res_type == PM_RES_MEM) {
			mem_grp = find_grouping_by_id(cmd);
			if (mem_grp) {
				set_util_dir((srcid_t)mem_grp, util_dir);
				src_id[src_cnt] = (srcid_t)mem_grp;
				src_cnt++;
			} else {
				sprintf(buf, "Memory Grouping ID %s not "
				    "found\n", cmd);
				(void) write(fd, buf, strlen(buf));
				goto fail;
			}
		}
		cmd = strtok(NULL, ",\r");
	}
	/*
	 * User has set at least one util direction for the specified
	 * resources.
	 */
	if (src_cnt > 0) {
		/* Set up override functions to mimic the Policy Engine */
		pmtest_i = src_cnt;
		hdl_fn = find_pm_handle;
		find_pm_handle = pmtest_find_pm_handle;
		virt_ops->get_domain_utils = get_domain_utils;
		virt_ops->resume_util_change_notify = resume_util_change_notify;

		/*
		 * Save and override the actual PM policy. Pretend we're in
		 * Elastic mode but the test harness can only be executed in
		 * Performance mode. Otherwise, the Policy Engine will get
		 * actual utilization values.  We could be overriding the
		 * policy set via the test interface, so that is saved too.
		 */
		save_pm_local_policy = pm_local_policy;
		pm_local_policy = PM_POLICY_ELASTIC;
		save_pmi_get_policy = pmi_get_policy;
		pmi_get_policy = get_local_policy;

		/* Fire the appropriate notification */
		pm_util_change_notify(&src_cnt, src_id, src_cnt, res_type);

		/* Restore saved function pointers */
		pm_local_policy = save_pm_local_policy;
		pmi_get_policy = save_pmi_get_policy;

		/*
		 * Restore global function pointers.  We can do this as we
		 * have the sequencer.
		 */
		virt_ops->get_domain_utils = ldm_get_domain_utils;
		virt_ops->resume_util_change_notify =
		    ldm_resume_util_change_notify;

		find_pm_handle = hdl_fn;
		(void) write(fd, PMTEST_RESP_PASS, sizeof (PMTEST_RESP_PASS));
		(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
		return;
	}
fail:
	(void) write(fd, PMTEST_RESP_FAIL, sizeof (PMTEST_RESP_FAIL));
	(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
}


/*
 * Handle the PMTEST_CMD_STRAND command.
 *
 * Sub commands are :
 * 	PMTEST_SUBCMD_ENABLE		: Enable strand(s).
 * 	PMTEST_SUBCMD_DISABLE		: Disable strand(s).
 *	PMTEST_SUBCMD_STATESP		: Return the state(s) of the strand(s).
 */
static void
pmtest_handle_strand(char *cmd, int fd)
{

	int c = 0;
	uint16_t current_states[PM_RES_MAX];
	pwr_req_status_t status[PM_RES_MAX];
	int n;
	int rc;
	char *pass = "PASS\n";
	char *fail = "FAIL\n";
	char *response;
	char respbuf[RESP_BUF_SZ];
	char buf[BUF_SZ];

	cmd += strlen(PMTEST_CMD_STRAND);

	if (strncmp(cmd, PMTEST_SUBCMD_ENABLE,
	    strlen(PMTEST_SUBCMD_ENABLE)) == 0) {
		PM_DEBUG("%s: recv'd strand enable command\n",
		    PM_DEBUG_PREFIX);
		cmd += strlen(PMTEST_SUBCMD_ENABLE);
		rc = pmtest_parse_all(cmd, PMTEST_STRAND, & c);
		if (rc == 0) {
			rc = cli_set_cpu_states(c, &g_res[0],
			    PPS_FULL_POWER, PCO_NONE);
			PM_DEBUG("cli_set_cpu_states rc = %d\n", rc);
		}
		response = fail;
		if (rc == 0)
			response = pass;
		(void) write(fd, response, strlen(response));
	} else if (strncmp(cmd, PMTEST_SUBCMD_DISABLE,
	    strlen(PMTEST_SUBCMD_DISABLE)) == 0) {
		PM_DEBUG("%s: recv'd strand disable command\n",
		    PM_DEBUG_PREFIX);
		cmd += strlen(PMTEST_SUBCMD_DISABLE);
		rc = pmtest_parse_all(cmd, PMTEST_STRAND, & c);
		if (rc == 0) {
			rc = cli_set_cpu_states(c, &g_res[0],
			    PPS_LOWEST_POWER, PCO_NONE);
			PM_DEBUG("cli_set_cpu_states rc = %d\n", rc);
		}
		response = fail;
		if (rc == 0)
			response = pass;
		(void) write(fd, response, strlen(response));
	} else if (strncmp(cmd, PMTEST_SUBCMD_STATESP,
	    strlen(PMTEST_SUBCMD_STATESP)) == 0) {
		PM_DEBUG("%s: recv'd strand state command\n", PM_DEBUG_PREFIX);
		cmd += strlen(PMTEST_SUBCMD_STATESP);
		rc = pmtest_parse_all(cmd, PMTEST_STRAND, & c);
		if (rc == 0) {
			rc = ldm_get_power_states(c,
			    &test_cookies[0], &current_states[0],
			    &status[0]);
			PM_DEBUG("ldm_get_power_states rc = %d\n", rc);
		}
		response = fail;
		if (rc == PS_SUCCESS)
			response = pass;
		(void) write(fd, response, strlen(response));
		strcpy(respbuf, "");
		if (rc == 0) {
			for (n = 0; n < c; n++) {
				sprintf(buf, "%s = %d\n", test_cookies[n],
				    current_states[n]);
				strcat(respbuf, buf);
			}
		}
		(void) write(fd, respbuf, strlen(respbuf));
	} else {
		PM_DEBUG("%s: Unexpected strand sub cmd : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
}

/*
 * pmtest_handle_memory - Handles the PMTEST_CMD_MEMORY command.
 *
 * Input parms:
 *   fd - socket connection
 *   cmd - Sub command of PMTEST_CMD_MEMORY :
 *      PMTEST_SUBCMD_GETSP     : Get current PM state for physical addr
 *      PMTEST_SUBCMD_SETSP     : Set PM state of grouping that physical
 *                                address belongs to.
 *      PMTEST_SUBCMD_GETSTATES : Get supported PM states
 *      PMTEST_SUBCMD_GETRANGES : Get power manageable physical addr ranges
 *      PMTEST_SUBCMD_GETGROUPING : Get total number of available groupings
 * Output parms: None
 */
static void
pmtest_handle_memory(char *cmd, int fd)
{
	char 		buf[BUF_SZ];
	char		*set_str = NULL;
	int 		grp_cnt = 0;
	int		grp_idx = 0;
	int		mem_grp_id = 0;
	int		rc = 0;
	int		st_idx = 0;
	pm_node_t	*child_np = NULL;
	pm_node_t	*pmnp = NULL;
	pwr_req_status_t status;
	uint16_t 	grp_pm_state = 0;
	uint64_t	mem_base = 0;
	uint64_t	mem_size = 0;
	uint16_t 	new_state = 0;
	uint64_t	phys_addr = 0;

	cmd += strlen(PMTEST_CMD_MEMORY);

	/* mem get <physical address> */
	if (strncmp(cmd, PMTEST_SUBCMD_GETSP,
	    strlen(PMTEST_SUBCMD_GETSP)) == 0) {
		cmd += strlen(PMTEST_SUBCMD_GETSP);
		rc = pmtest_parse_all(cmd, PMTEST_MEMORY, &grp_cnt);

		/* The physical address entered on the CLI is returned */
		phys_addr = strtoll(cmd, NULL, 16);

		/* Invalid parameters/format */
		if (rc == 2)
			goto mem_failnak;

		if (!find_region_in_mem_grouping(&grp_idx, &pmnp, phys_addr,
		    grp_cnt)) {
			sprintf(buf, "Physical address 0x%llx not found within "
			    "defined PRI memory groupings.\n", phys_addr);
			PM_DEBUG("%s: Physical address 0x%llx not found within "
			    "defined PRI memory groupings.\n", PM_DEBUG_PREFIX,
			    phys_addr);
			goto mem_passnak;
		}

		/* Get the memory grouping's current PM state */
		rc = ldm_get_power_states(1, &test_cookies[grp_idx],
		    &grp_pm_state, &status);
		PM_DEBUG("%s: ldm_get_power_states rc=%d\n",
		    PM_DEBUG_PREFIX, rc);

		if (rc == 0) {
			/* Print the debug and interface output */
			PM_DEBUG("%s: mem get: state=%d:%s status=%d\n",
			    PM_DEBUG_PREFIX, grp_pm_state,
			    pmnp->pn_pwrs->pps_pmcap[grp_pm_state].ppc_lname,
			    status);
			(void) write(fd, PMTEST_RESP_PASS,
			    strlen(PMTEST_RESP_PASS));
			sprintf(buf, "%d\t%s\n", grp_pm_state,
			    pmnp->pn_pwrs->pps_pmcap[grp_pm_state].ppc_lname);
			(void) write(fd, buf, strlen(buf));
			goto mem_passack;

		/* Print out to debug and interface, the HV fail output */
		} else {
			sprintf(buf, "get failure - %s\n", pmi_err_str(rc));
			goto mem_passnak;
		}

	/* mem set <physical address> <state> */
	} else if (strncmp(cmd, PMTEST_SUBCMD_SETSP,
	    strlen(PMTEST_SUBCMD_SETSP)) == 0) {
		cmd += strlen(PMTEST_SUBCMD_SETSP);
		rc = pmtest_parse_all(cmd, PMTEST_MEMORY, &grp_cnt);

		/* Invalid parameters/format */
		if (rc == 2)
			goto mem_failnak;

		/* The physical address entered on the CLI is returned */
		strcpy(buf, cmd);
		set_str = strtok(buf, "x");
		set_str = strtok(NULL, " ");
		if (!set_str || !isdigit(*set_str))
			goto mem_failnak;

		phys_addr = strtoll(set_str, NULL, 16);
		set_str = strtok(NULL, " ");
		if (!set_str || !isdigit(*set_str))
			goto mem_failnak;

		new_state = (uint16_t)atoi(set_str);

		if (!find_region_in_mem_grouping(&grp_idx, &pmnp, phys_addr,
		    grp_cnt)) {
			sprintf(buf, "Physical address 0x%llx not found within"
			    " defined PRI memory groupings.\n", phys_addr);
			PM_DEBUG("%s: Physical address 0x%llx not found within"
			    " defined PRI memory groupings.\n", PM_DEBUG_PREFIX,
			    phys_addr);
			goto mem_passnak;
		}

		/* Set the memory grouping's PM state */
		rc = ldm_set_power_states(1, &test_cookies[grp_idx],
		    &new_state, &status);
		PM_DEBUG("%s: ldm_set_power_states rc=%d\n",
		    PM_DEBUG_PREFIX, rc);

		if (rc == 0) {
			/* Print the debug and interface output */
			PM_DEBUG("%s: mem set: state=%d(%s) status=%d\n",
			    PM_DEBUG_PREFIX, new_state,
			    pmnp->pn_pwrs->pps_pmcap[new_state].ppc_lname,
			    status);
			(void) write(fd, PMTEST_RESP_PASS,
			    strlen(PMTEST_RESP_PASS));
			sprintf(buf, "%d\t%s\n", new_state,
			    pmnp->pn_pwrs->pps_pmcap[st_idx].ppc_lname);
			(void) write(fd, buf, strlen(buf));
			goto mem_passack;

		/* Print out to debug and interface, the HV fail output */
		} else {
			sprintf(buf, "set failure - %s\n", pmi_err_str(rc));
			goto mem_passnak;
		}

	/* mem getranges <grouping id> */
	} else if (strncmp(cmd, PMTEST_SUBCMD_GETRANGES,
	    strlen(PMTEST_SUBCMD_GETRANGES)) == 0) {
		cmd += strlen(PMTEST_SUBCMD_GETRANGES);

		/*
		 * This command only accepts a single integer for the memory
		 * groupings value.
		 */
		if (!cmd || !isdigit(*cmd))
			goto mem_failnak;

		/*
		 * Search thru the pm_nodes to find a matching memory
		 * grouping node
		 */
		mem_grp_id = atoi(cmd);
		pmnp = pmi_find_node_by_type(NULL, MEM_GROUPING, mem_grp_id,
		    INVALID_ID);

		if (pmnp) {
			/*
			 * Print out all memory region address ranges
			 * for the memory grouping.
			 */
			(void) write(fd, PMTEST_RESP_PASS,
			    strlen(PMTEST_RESP_PASS));
			for (child_np = pmnp->pn_child; child_np;
			    child_np = child_np->pn_sibling) {
				child_np = pmi_find_node_by_type(NULL,
				    MEM_REGION, child_np->pn_nodeid,
				    mem_grp_id);

				mem_base =
				    child_np->pn_rcfg->\
				    u.mem.mem_reg.pi_reg_base;
				mem_size =
				    child_np->pn_rcfg->\
				    u.mem.mem_reg.pi_reg_size;

				PM_DEBUG("0x%llx\t0x%llx\n", mem_base,
				    mem_size);
				sprintf(buf, "0x%llx\t0x%llx\n", mem_base,
				    mem_size);
				(void) write(fd, buf, strlen(buf));
			}
			goto mem_passack;
		} else {
			/* Memory grouping does not exist */
			sprintf(buf, "memory grouping %d not present\n",
			    mem_grp_id);
			goto mem_passnak;
		}

	/* mem getstates */
	} else if (strncmp(cmd, PMTEST_SUBCMD_GETSTATES,
	    strlen(PMTEST_SUBCMD_GETSTATES)) == 0) {
		/*
		 * We only need to look at the states defined in the first
		 * memory grouping since all groupings will have the same
		 * states defined.
		 */
		pmnp = pmi_find_node_by_type(NULL, MEM_GROUPING, 0, INVALID_ID);

		(void) write(fd, PMTEST_RESP_PASS, sizeof (PMTEST_RESP_PASS));
		if (pmnp != NULL) {
			for (st_idx = 0; st_idx < pmnp->pn_pwrs->pps_count;
			    st_idx++) {
				PM_DEBUG("%d\t%s\n",
				    pmnp->pn_pwrs->pps_pmcap[st_idx].ppc_lvl,
				    pmnp->pn_pwrs->pps_pmcap[st_idx].ppc_lname);
				sprintf(buf, "%d\t%s\n",
				    pmnp->pn_pwrs->pps_pmcap[st_idx].ppc_lvl,
				    pmnp->pn_pwrs->pps_pmcap[st_idx].ppc_lname);
				(void) write(fd, buf, strlen(buf));
			}
			goto mem_passack;
		} else {
			/* Memory grouping does not exist */
			sprintf(buf, "memory grouping %d not present\n",
			    mem_grp_id);
		}
	/* mem getgrouping */
	} else if (strncmp(cmd, PMTEST_SUBCMD_GETGROUPING,
	    strlen(PMTEST_SUBCMD_GETGROUPING)) == 0) {
		do {
			pmnp = pmi_find_node_by_type(NULL, MEM_GROUPING,
			    grp_idx, INVALID_ID);
			if (pmnp)
				grp_idx++;
		} while (pmnp);

		(void) write(fd, PMTEST_RESP_PASS, sizeof (PMTEST_RESP_PASS));
		PM_DEBUG("%d\n", grp_idx);
		sprintf(buf, "%d\n", grp_idx);
		(void) write(fd, buf, strlen(buf));
		goto mem_passack;
	} else {
		goto mem_failnak;
	}

mem_passnak:
	(void) write(fd, PMTEST_RESP_PASS, sizeof (PMTEST_RESP_PASS));
	(void) write(fd, buf, strlen(buf));
	(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
	return;

mem_passack:
	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
	return;

mem_failnak:
	PM_DEBUG("%s: Invalid parameter format: %s\n", PM_DEBUG_PREFIX, cmd);
	(void) write(fd, PMTEST_RESP_FAIL, sizeof (PMTEST_RESP_FAIL));
	sprintf(buf, "Invalid parameter format\n");
	(void) write(fd, buf, strlen(buf));
	(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));

} /* pmtest_handle_memory() */

/*
 * Handle the PMTEST_CMD_CLOCK command.
 *
 * Sub commands are :
 * 	PMTEST_SUBCMD_GETSP	: Get chip clock step
 * 	PMTEST_SUBCMD_GETSTEPS	: Gets the number of clock steps & descriptions
 *	PMTEST_SUBCMD_SETSP  	: Set a chip clock step
 */
static void
pmtest_handle_clock(char *cmd, int fd)
{
	pm_node_t		*pmnp;
	pm_pwr_scheme_t		*pwrs;
	pwr_req_status_t 	status[PM_RES_MAX] = { 0 };
	uint16_t 		new_states[PM_RES_MAX] = { 0 };

	int 	cnt = 0;
	int	i, rc;
	int	chip_num;
	int	steps_idx = 0;
	int	step_str_idx = 0;
	char 	buf[BUF_SZ];
	char	chip_str[128];	/* Just need enough space for parameters */
	char	step_str[9];	/* Just need enough space for chip number */
	char 	*chip_tok, *step_tok;

	cmd += strlen(PMTEST_CMD_CLOCK);

	/* clock get <chip(s)> */
	if (strncmp(cmd, PMTEST_SUBCMD_GETSP,
	    strlen(PMTEST_SUBCMD_GETSP)) == 0) {
		cmd += strlen(PMTEST_SUBCMD_GETSP);
		rc = pmtest_parse_all(cmd, PMTEST_CLOCK, &cnt);

		/* Invalid parameters/format */
		if (rc == 2) {
			goto clk_inv_param;
		}

		/*
		 * Search thru the pm_nodes to find a matching processor node
		 * for each entry in test_cookies.  Any invalid entry
		 * will make the command fail
		 */
		for (i = 0; i < cnt; i++) {
			chip_num = atoi(test_cookies[i]+2);
			PM_DEBUG("clock get cookie: %s, chip_num=%d\n",
			    test_cookies[i], chip_num);

			pmnp = pmi_find_node_by_type(NULL, "processor",
			    chip_num, INVALID_ID);
			if (pmnp == NULL) {
				sprintf(buf, "chip %d not present\n", chip_num);
				PM_DEBUG("chip %d not present\n", chip_num);
				goto clk_passnak;
			}
		}

		/* Get the chip(s) clock steps */
		if (rc == 0) {
			rc = ldm_get_power_states(cnt, &test_cookies[0],
			    &new_states[0], &status[0]);
			PM_DEBUG("ldm_get_power_states rc=%d\n", rc);
		}

		/* Success in getting the clock power states */
		if (rc == 0) {
			(void) write(fd, PMTEST_RESP_PASS,
			    strlen(PMTEST_RESP_PASS));

			/* Print the debug and interface output */
			for (i = 0; i < cnt; i++) {
				PM_DEBUG("clock get: %s = state:%d status:%d\n",
				    test_cookies[i], new_states[i],
				    status[i]);
				sprintf(buf, "%d\n", new_states[i]);
				(void) write(fd, buf, strlen(buf));
			}
			goto clk_ack;

		/* Print out to debug and interface, the HV fail output */
		} else {
			sprintf(buf, "get failure - %s\n", pmi_err_str(rc));
			goto clk_passnak;
		}


	/* clock getsteps <chip> */
	} else if (strncmp(cmd, PMTEST_SUBCMD_GETSTEPS,
	    strlen(PMTEST_SUBCMD_GETSTEPS)) == 0) {
		cmd += strlen(PMTEST_SUBCMD_GETSTEPS);

		/*
		 * This command only accepts a single integer for the <chip>
		 * value. The parameter must begin with a digit, but then can
		 * have any other character after it because aoti() is used to
		 * get the integer value.We aren't going to be completely
		 * defensive in parsing becasue the Test Harness spec defines
		 * what the legal syntax is.  We will defend against ',',  '.'
		 * and '-' in the string, but '1foo' will get through.
		 */
		if ((isdigit(*cmd) == 0) || (strstr(cmd, ".") != NULL) ||
		    (strstr(cmd, ",") != NULL) || (strstr(cmd, "-") != NULL)) {
			goto clk_inv_param;
		}

		/* Search thru the pm_nodes to find a matching processor node */
		chip_num = atoi(cmd);
		pmnp = pmi_find_node_by_type(NULL, "processor", chip_num,
		    INVALID_ID);

		if (pmnp != NULL) {
			/* Print out the # of pm_states and descriptions */
			(void) write(fd, PMTEST_RESP_PASS,
			    strlen(PMTEST_RESP_PASS));

			/*
			 * Print out the total number of pm_states, then
			 * walk the pm_node->pn_pwrs->pps_pmcap array
			 * and print out the ppc_perf and ppc_lname values.
			 * These last pairs are the pm_state descriptions.
			 */
			pwrs = pmnp->pn_pwrs;
			cnt = pwrs->pps_count;

			PM_DEBUG("clock getsteps: Total pm_states=%d\n", cnt);
			sprintf(buf, "%d\n", cnt);
			(void) write(fd, buf, strlen(buf));

			for (i = 0; i < cnt; i++) {
				PM_DEBUG("clock getsteps:pm_states[%d]=%d %s\n",
				    i, pwrs->pps_pmcap[i].ppc_perf,
				    pwrs->pps_pmcap[i].ppc_lname);
				sprintf(buf, "%d %s\n",
				    pwrs->pps_pmcap[i].ppc_perf,
				    pwrs->pps_pmcap[i].ppc_lname);
				(void) write(fd, buf, strlen(buf));
			}
			goto clk_ack;
		} else {
			/* chip number does not exist */
			sprintf(buf, "chip %d not present\n", chip_num);
			goto clk_passnak;
		}


	/* clock set <chip(s)> <step(s)> */
	} else if (strncmp(cmd, PMTEST_SUBCMD_SETSP,
	    strlen(PMTEST_SUBCMD_SETSP)) == 0) {
		cmd += strlen(PMTEST_SUBCMD_SETSP);

		/*
		 * cmd now should point to '<chip(s)> <step(s)>' string.  We
		 * want to only pass in the '<chip(s)>' string to parse_all(),
		 * so do the string manipulation to get pointers to the chip(s)
		 * and step(s) sub strings. We also want to screen for '.' and
		 * '-' in cmd.  These are illegal chars for this sub command.
		 */
		strcpy(chip_str, cmd);

		if ((strstr(chip_str, " ") == NULL) ||
		    (strstr(chip_str, ".") != NULL) ||
		    (strstr(chip_str, "-") != NULL)) {
			goto clk_inv_param;
		}

		/*
		 * Now we want to send the '<chip(s)>' sub string to parse_all()
		 * and we will parse the '<step(s)>' sub string here.
		 * parse_all() will build the cookie strings from '<chip(s)>'
		 * and the new_states[] will be populated here from '<step(s)>'.
		 * An example of the string is: '1,2  0,3'  or '1 3'.
		 *
		 * Point to '<chip(s)>' sub string NULL terminated by strtok().
		 * Need to get the '<step(s)>' substring also since parse_all()
		 * calls strtok(), which will change the chip_tok pointer.
		 */
		chip_tok = strtok(chip_str, " ");
		step_tok = strtok(NULL, " ");
		rc = pmtest_parse_all(chip_tok, PMTEST_CLOCK, &cnt);

		if (rc == 0) {
			/*
			 * Validate the '<step(s)>' part of the string. There
			 * should be a digit char after the space char.
			 */
			if ((step_tok == NULL) || ((isdigit(*step_tok) == 0) &&
			    (*step_tok != '0'))) {
				goto clk_inv_param;
			}

			/*
			 * Have to have more than 1 step since we have more than
			 * 1 chip (cnt>1) More than 1 step is indicated by
			 * having ',' in the string.
			 */
			if ((cnt > 1) && (strstr(step_tok, ",") == NULL)) {
				goto clk_inv_param;
			}

			/*
			 * Can have only 1 step since have only 1 chip (cnt==1)
			 * Can't have any ',' in the string for only 1 step.
			 */
			if ((cnt == 1) && (strstr(step_tok, ",") != NULL)) {
				goto clk_inv_param;
			}

			/*
			 * test_cookies[] has been populated from parse_all(),
			 * so now populate new_states[] from the '<step(s)>' sub
			 * string.  Walk thru '<step(s)>' char by char.  Once we
			 * get a set of chars that represent an integer, which
			 * represents a pm_states index, put that integer into
			 * new_states[]. <step(s)>: 1,2 | 0 | 1,2,3,...,n.
			 * The char digits are accumulated in step_str[].
			 */
			while (*step_tok != NULL) {
				/* Accumulate digits into a string */
				if (isdigit(*step_tok) || (*step_tok == '0')) {
					step_str[step_str_idx++] = *step_tok;

				/* Not a digit, see if digits have been seen */
				} else if (step_str_idx > 0) {
					step_str[step_str_idx] = '\0';
					new_states[steps_idx++] =
					    (uint16_t)atoi(step_str);
					step_str_idx = 0;

				/* Step string not of form: x or x,y,z,... */
				} else {
					goto clk_inv_param;
				}
				step_tok++;
			}

			/* Catch the last digit of '<step(s)>' */
			if (step_str_idx > 0) {
				step_str[step_str_idx] = '\0';
				new_states[steps_idx++] =
				    (uint16_t)atoi(step_str);
				step_str_idx = 0;
			}

			/* Same # of steps as chips? */
			if (cnt != steps_idx) {
				goto clk_inv_param;
			}

			/*
			 * Search for chips that are not present or if the step
			 * index is greater than the number of pm_states
			 * supported. Example: A chip's PRI pm_states has 5
			 * entries, meaning 5 (0..4) step indexes.  If the
			 * command is to set the clock step to a step index >
			 * 4, then this is invalid.
			 */
			for (i = 0; i < cnt; i++) {
				chip_num = atoi(test_cookies[i]+2);
				PM_DEBUG("chip_num=%d\n", chip_num);

				pmnp = pmi_find_node_by_type(NULL, "processor",
				    chip_num, INVALID_ID);
				/* Check if chip is present */
				if (pmnp == NULL) {
					sprintf(buf, "chip %d not present\n",
					    chip_num);
					PM_DEBUG("chip %d not present\n",
					    chip_num);
					goto clk_passnak;
				}

				/* Chip present, step level is supported? */
				if (new_states[i] >
				    (pmnp->pn_pwrs->pps_count - 1)) {
					sprintf(buf,
					    "chip %d step %d not supported\n",
					    chip_num, new_states[i]);
					PM_DEBUG("chip %d step %d "
					    "not supported\n",
					    chip_num, new_states[i]);
					goto clk_passnak;
				}

			}

			/* Set the new Clock steps */
			for (i = 0; i < cnt; i++) {
				PM_DEBUG("clock set: cookie=%s, new_state=%d\n",
				    test_cookies[i], new_states[i]);
			}
			rc = ldm_set_power_states(cnt, &test_cookies[0],
			    &new_states[0], &status[0]);
			PM_DEBUG("ldm_set_power_states rc = %d:%s\n", rc,
			    pmi_err_str(rc));

			if (rc == PS_SUCCESS) {

				/* Cmd was valid, so print a PASS */
				(void) write(fd, PMTEST_RESP_PASS,
				    strlen(PMTEST_RESP_PASS));

				/* print detailed info for state and status */
				for (i = 0; i < cnt; i++) {
					PM_DEBUG("%s %s = %d %d\n",
					    "clock set output:",
					    test_cookies[i],
					    new_states[i],
					    status[i]);

					sprintf(buf, "%s = %d %d\n",
					    test_cookies[i], new_states[i],
					    status[i]);
					(void) write(fd, buf, strlen(buf));
				}
				goto clk_ack;
			}

			/* Failure to set power states */
			else {
				/* Debug output */
				for (i = 0; i < cnt; i++) {
					PM_DEBUG("%s %s = %d %d\n",
					    "clock set output:",
					    test_cookies[i],
					    new_states[i],
					    status[i]);
				}
				sprintf(buf, "set failure - %s\n",
				    pmi_err_str(rc));
				goto clk_passnak;
			}

		/* Invalid parameter */
		} else if (rc == 2) {
			goto clk_inv_param;

		} else {
			sprintf(buf, "resource out of bounds\n");
			goto clk_failnak;
		}

	} else {
		PM_DEBUG("%s: Unknown clock sub cmd: %s\n", PM_DEBUG_PREFIX,
		    cmd);
		sprintf(buf, "unknown clock sub cmd: %s\n", cmd);
		goto clk_failnak;
	}

clk_passnak:
	(void) write(fd, PMTEST_RESP_PASS, sizeof (PMTEST_RESP_PASS));
	(void) write(fd, buf, strlen(buf));
	(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
	return;

clk_ack:
	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
	return;

clk_inv_param:
	PM_DEBUG("%s: invalid parameter format: %s\n", PM_DEBUG_PREFIX, cmd);
	sprintf(buf, "invalid parameter format\n");

clk_failnak:
	(void) write(fd, PMTEST_RESP_FAIL, sizeof (PMTEST_RESP_FAIL));
	(void) write(fd, buf, strlen(buf));
	(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
	return;

}


/*
 * Handle the PMTEST_CMD_CORE command.
 *
 * Sub commands are :
 * 	PMTEST_SUBCMD_ENABLE		: Enable core(s).
 * 	PMTEST_SUBCMD_DISABLE		: Disable core(s).
 *	PMTEST_SUBCMD_STATESP		: Return the state(s) of the core(s).
 */
static void
pmtest_handle_core(char *cmd, int fd)
{
	int c = 0;
	pwr_req_status_t status[PM_RES_MAX] = { 0 };
	uint16_t new_states[PM_RES_MAX] = { 0 };
	int n;
	int rc;
	char *pass = "PASS\n";
	char *fail = "FAIL\n";
	char *resp;
	char respbuf[RESP_BUF_SZ];
	char buf[BUF_SZ];

	cmd += strlen(PMTEST_CMD_CORE);

	if (strncmp(cmd, PMTEST_SUBCMD_ENABLE,
	    strlen(PMTEST_SUBCMD_ENABLE)) == 0) {
		PM_DEBUG("%s: recv'd core enable command\n", PM_DEBUG_PREFIX);
		cmd += strlen(PMTEST_SUBCMD_ENABLE);
		rc = pmtest_parse_all(cmd, PMTEST_CORE, & c);

		if (rc == 0) {
			for (n = 0; n < c; n++) {
				new_states[n] = PMTEST_ENABLE;
			}

			rc = ldm_set_power_states(c, &test_cookies[0],
			    &new_states[0], &status[0]);
			PM_DEBUG("ldm_set_power_states rc = %d\n", rc);

			if (rc == PS_SUCCESS) {
				for (n = 0; n < c; n++) {
					PM_DEBUG("%s %s = %d %d\n",
					    "core enable output:",
					    test_cookies[n],
					    new_states[n],
					    status[n]);

					if (status[n] != 0) {
						sprintf(buf, "%s = %d\n",
						    test_cookies[n],
						    status[n]);
						(void) write(fd, buf,
						    strlen(buf));
						rc++;
					}
				}
			}
			/* Debug output for failure to set power states */
			else {
				for (n = 0; n < c; n++) {
					PM_DEBUG("%s %s = %d %d\n",
					    "core enable output:",
					    test_cookies[n],
					    new_states[n],
					    status[n]);
				}
			}
		}
		resp = fail;
		if (rc == 0)
			resp = pass;
		(void) write(fd, resp, strlen(resp));

	} else if (strncmp(cmd, PMTEST_SUBCMD_DISABLE,
	    strlen(PMTEST_SUBCMD_DISABLE)) == 0) {
		PM_DEBUG("%s: recv'd core disable command\n", PM_DEBUG_PREFIX);
		cmd += strlen(PMTEST_SUBCMD_DISABLE);
		rc = pmtest_parse_all(cmd, PMTEST_CORE, & c);

		if (rc == 0) {
			for (n = 0; n < c; n++) {
				new_states[n] = PMTEST_DISABLE;
			}

			rc = ldm_set_power_states(c, & test_cookies[0],
			    &new_states[0], &status[0]);
			PM_DEBUG("ldm_set_power_states rc = %d\n", rc);

			if (rc == PS_SUCCESS) {
				for (n = 0; n < c; n++) {
					PM_DEBUG("%s %s = %d %d\n",
					    "core disble output:",
					    test_cookies[n],
					    new_states[n],
					    status[n]);

					if (status[n] != 0) {
						sprintf(buf, "%s = %d\n",
						    test_cookies[n],
						    status[n]);
						(void) write(fd, buf,
						    strlen(buf));
						rc++;
					}
				}
			}
			/* Debug output for failure to set power states */
			else {
				for (n = 0; n < c; n++) {
					PM_DEBUG("%s %s = %d %d\n",
					    "core disable output:",
					    test_cookies[n],
					    new_states[n],
					    status[n]);
				}
			}
		}

		resp = fail;
		if (rc == 0)
			resp = pass;
		(void) write(fd, resp, strlen(resp));

	} else if (strncmp(cmd, PMTEST_SUBCMD_STATESP,
	    strlen(PMTEST_SUBCMD_STATESP)) == 0) {
		PM_DEBUG("%s: recv'd core state command\n", PM_DEBUG_PREFIX);
		cmd += strlen(PMTEST_SUBCMD_STATESP);
		rc = pmtest_parse_all(cmd, PMTEST_CORE, & c);

		if (rc == 0) {
			rc = ldm_get_power_states(c, &test_cookies[0],
			    &new_states[0], & status[0]);
			PM_DEBUG("ldm_get_power_states rc = %d\n", rc);
		}

		resp = fail;
		if (rc == PS_SUCCESS)
			resp = pass;

		(void) write(fd, resp, strlen(resp));
		strcpy(respbuf, "");

		if (rc == 0) {
			for (n = 0; n < c; n++) {
				PM_DEBUG("core state output: %s = %d %d\n",
				    test_cookies[n], new_states[n],
				    status[n]);
				sprintf(buf, "%s = %d %d\n",
				    test_cookies[n], new_states[n], status[n]);
				strcat(respbuf, buf);
			}
		}

		(void) write(fd, respbuf, strlen(respbuf));

	} else {
		PM_DEBUG("%s: Unexpected core sub cmd : %s\n", PM_DEBUG_PREFIX,
		    cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
}

/*
 * Handle the PMTEST_CMD_PRI command.
 *
 * Sub commands are :
 * 	PMTEST_SUBCMD_GET		: Get current PRI.
 */
static void
pmtest_handle_pri(char *cmd, int fd)
{
	char buf[BUF_SZ];

	cmd += strlen(PMTEST_CMD_PRI);

	if (strncmp(cmd, PMTEST_SUBCMD_GET,
	    strlen(PMTEST_SUBCMD_GET)) == 0) {
		PM_DEBUG("%s: recv'd pri get cmd\n", PM_DEBUG_PREFIX);

		sprintf(buf, "PRI check returned %s",
		    pmi_is_pri_pm_capable() ? PMTEST_RESP_PASS :
		    PMTEST_RESP_FAIL);

		(void) write(fd, buf, strlen(buf));
	} else {
		PM_DEBUG("%s: Unexpected pri sub cmd : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
}

/*
 * Handle the PMTEST_CMD_POLICY command.
 *
 * Sub commands are :
 * 	PMTEST_SUBCMD_GET		: Get current policy setting.
 * 	PMTEST_SUBCMD_CLEAR		: Resume/start getting the policy from
 *					  the SP.
 *	PMTEST_SUBCMD_SET <value>	: Stop the PM engine from getting the
 *					  policy from the SP (if necessary) and
 *					  set the policy to the given value.
 *
 *	Where "value" is a string like "elastic" or "performance".
 */
static void
pmtest_handle_policy(char *cmd, int fd)
{
	pmi_policy_t pol = { 0 };
	int r;
	char str[BUF_SZ];
	char *tmp = NULL;

	cmd += strlen(PMTEST_CMD_POLICY);

	if (strncmp(cmd, PMTEST_SUBCMD_GET,
	    strlen(PMTEST_SUBCMD_GET)) == 0) {
		PM_DEBUG("%s: recv'd policy get cmd\n", PM_DEBUG_PREFIX);

		/* Get policy */
		r = pmi_get_policy(&pol);

		/*
		 * If local override then get global policy
		 * from the SP via snmp
		 */
		if (pm_policy_override == 1) {
			r = pmi_get_policy_by_snmp(&pol);
		}

		if (r == 0) {
			switch (pol) {
			case PM_POLICY_PERFORMANCE:
				strcpy(str, "Performance Policy\nPASS\n");
				break;
			case PM_POLICY_ELASTIC:
				strcpy(str, "Elastic Policy\nPASS\n");
				break;
			default:
				strcpy(str, "Invalid Policy\nFAIL\n");
				break;
			}

			if (pm_policy_override == 1) {
				tmp = strdup(str);
				if (!tmp)
					strcpy(str, "strdup FAIL\n");
				else {
					strcpy(str, "Local-");
					strcat(str, (pm_local_policy ==
					    PM_POLICY_ELASTIC) ? "Elastic"
					    : "Performance");
					strcat(str, "; ");
					strcat(str, tmp);
					free(tmp);
				}
			}
			(void) write(fd, str, strlen(str));

		/*
		 * Could not get policy from the SP via snmp.
		 * Doesn't matter if local override is on or not
		 */
		} else {
			strcpy(str, "failure to get policy via snmp\n");
			(void) write(fd, str, strlen(str));
			(void) write(fd, PMTEST_RESP_FAIL,
			    strlen(PMTEST_RESP_FAIL));
			(void) write(fd, PMTEST_RESP_ACK,
			    sizeof (PMTEST_RESP_ACK));
			return;

		}

	} else if (strncmp(cmd, PMTEST_SUBCMD_CLEAR,
	    strlen(PMTEST_SUBCMD_CLEAR)) == 0) {
		PM_DEBUG("%s: recv'd policy clear cmd\n", PM_DEBUG_PREFIX);
		pmi_get_policy = &pmi_get_policy_by_snmp;
		pm_policy_override = 0;
	} else if (strncmp(cmd, PMTEST_SUBCMD_SETSP,
	    strlen(PMTEST_SUBCMD_SETSP)) == 0) {
		PM_DEBUG("%s: recv'd policy set cmd\n", PM_DEBUG_PREFIX);

		cmd += strlen(PMTEST_SUBCMD_SETSP);

		pmi_get_policy = &get_local_policy;
		pm_policy_override = 1;
		if (strncmp(cmd, "elastic", strlen("elastic")) == 0)
			pm_local_policy = PM_POLICY_ELASTIC;
		else
			pm_local_policy = PM_POLICY_PERFORMANCE;

	} else {
		PM_DEBUG("%s: Unexpected policy sub cmd : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	(void) write(fd, PMTEST_RESP_PASS, strlen(PMTEST_RESP_PASS));
	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
}

/*
 * Handle the PMTEST_CMD_FMA command.
 *
 * Sub commands are :
 * 	PMTEST_SUBCMD_RETIRE		: retire/offline a resource.
 * 	PMTEST_SUBCMD_UNRETIRE		: unretire/online a resource.
 *
 *	PMTEST_CMD_FMA [RETIRE | UNRETIRE] <domid> <res_type> <res_id>
 *
 *	Where "domid" is the name of the domain.
 *	Where "res_type" is a string like "cpu" or "memory".
 *	Where "res_id" is the ID value of the resource.
 */
static void
pmtest_handle_fma(char *cmd, int fd)
{
	char *dname;
	pm_resid_t res;
	uint32_t res_id;
	uint_t fma_flag;
	ldom_t *ldp;

	cmd += strlen(PMTEST_CMD_FMA);

	if (strncmp(cmd, PMTEST_SUBCMD_RETIRE,
	    strlen(PMTEST_SUBCMD_RETIRE)) == 0) {
		PM_DEBUG("%s: recv'd fma retire cmd\n", PM_DEBUG_PREFIX);

		cmd += strlen(PMTEST_SUBCMD_RETIRE);
		fma_flag = RESOURCE_RETIRE_EVENT;

	} else if (strncmp(cmd, PMTEST_SUBCMD_UNRETIRE,
	    strlen(PMTEST_SUBCMD_UNRETIRE)) == 0) {
		PM_DEBUG("%s: recv'd fma unretire cmd\n", PM_DEBUG_PREFIX);

		cmd += strlen(PMTEST_SUBCMD_UNRETIRE);
		fma_flag = RESOURCE_UNRETIRE_EVENT;

	} else {
		PM_DEBUG("%s: Unexpected fma sub cmd : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	/* Parameter : domain name */
	if ((cmd = strtok(cmd, " ")) == NULL)
		goto fma_fail;
	dname = cmd;
	if ((ldp = find_ldom_by_name(dname)) == NULL) {
		PM_DEBUG("%s: unable to find domain name : %s\n",
		    PM_DEBUG_PREFIX, dname);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	/* Parameter : resource type */
	if ((cmd = strtok(NULL, " ")) == NULL)
		goto fma_fail;
	if (strncmp(cmd, CPU_RES_NAME, strlen(CPU_RES_NAME)) == 0) {
		res.pi_res_type = PM_RES_CPU;
		res.pi_res_name = CPU_RES_NAME;
	} else {
		PM_DEBUG("%s: illegal resource type : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	/* Parameter : resource id */
	if ((cmd = strtok(NULL, " ")) == NULL)
		goto fma_fail;
	res_id = atoi(cmd);
	if (res.pi_res_type == PM_RES_CPU)
		res.u.cpu.pi_id_val = res_id;

	pm_fma_event_notify((srcid_t)ldp, res, fma_flag);

	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));

	return;

fma_fail:
	PM_DEBUG("%s: not enough args for fma command\n", PM_DEBUG_PREFIX);
	PM_DEBUG("%s: usage : fma [retire|unretire] <domain_name> "
	    "<resource_type> <resource_id>\n", PM_DEBUG_PREFIX);
	(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
}

/*
 * Handle the PMTEST_CMD_DEFRAG command.
 *
 * Sub commands are :
 * 	PMTEST_SUBCMD_RUN		: call ldm_defragment().
 *
 * 	PMTEST_CMD_DEFRAG RUN <res_type>
 *
 *	Where "res_type" is a string like "cpu" or "memory".
 */
static void
pmtest_handle_defrag(char *cmd, int fd)
{
	pm_res_type_t res_type;
	ldm_err_t results;

	cmd += strlen(PMTEST_CMD_DEFRAG);

	if (strncmp(cmd, PMTEST_SUBCMD_RUN, strlen(PMTEST_SUBCMD_RUN)) == 0) {
		cmd += (strlen(PMTEST_SUBCMD_RUN) + 1);
	} else {
		PM_DEBUG("%s: Unexpected defrag sub cmd : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	/* Parameter : resource type */
	if (strncmp(cmd, CPU_RES_NAME, strlen(CPU_RES_NAME)) == 0) {
		res_type = PM_RES_CPU;
	} else {
		PM_DEBUG("%s: illegal resource type : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	ldm_defragment(PACK, res_type, &results);

	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
}

/*
 * Handle the PMTEST_CMD_PMSTAT command.
 *
 * Sub commands are :
 * 	PMTEST_SUBCMD_SET		: set PM_MANAGED status.
 * 	PMTEST_SUBCMD_CLEAR		: clear PM_MANAGED status.
 *
 * 	PMTEST_CMD_PMSTAT [SET|CLEAR] <res_type> <res_id>
 *
 *	Where "res_type" is a string like "cpu" or "memory".
 *	Where "res_id" is the ID value of the resource.
 */
static void
pmtest_handle_pmstat(char *cmd, int fd)
{
	pm_res_type_t res_type;
	pm_status_t pm_status;
	int i, c, rc;

	cmd += strlen(PMTEST_CMD_PMSTAT);

	if (strncmp(cmd, PMTEST_SUBCMD_SET, strlen(PMTEST_SUBCMD_SET)) == 0) {
		cmd += strlen(PMTEST_SUBCMD_SETSP);
		pm_status = PM_MANAGED;
	} else if (strncmp(cmd, PMTEST_SUBCMD_CLEAR,
	    strlen(PMTEST_SUBCMD_CLEAR)) == 0) {
		cmd += strlen(PMTEST_SUBCMD_CLEARSP);
		pm_status = PM_NORMAL;
	} else {
		PM_DEBUG("%s: Unexpected manage sub cmd : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	/* Parameter : resource type */
	if ((cmd = strtok(cmd, " ")) == NULL)
		goto pmstat_fail;
	if (strncmp(cmd, CPU_RES_NAME, strlen(CPU_RES_NAME)) == 0) {
		res_type = PM_RES_CPU;
	} else {
		PM_DEBUG("%s: illegal resource type : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	/* Parameter : resource id(s) */
	if ((cmd = strtok(NULL, " ")) == NULL)
		goto pmstat_fail;
	rc = pmtest_parse_all(cmd, PMTEST_STRAND, &c);
	if (rc != 0) {
		PM_DEBUG("%s: illegal resource id syntax : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	if (res_type == PM_RES_CPU) {
		for (i = 0; i < c; i++) {
			set_core_md_cpu_status(g_res[i], CORE_MD_PM_STATUS,
			    pm_status);
		}
	}

	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));

	return;

pmstat_fail:
	PM_DEBUG("%s: not enough args for manage command\n", PM_DEBUG_PREFIX);
	PM_DEBUG("%s: usage : manage [set|clear] <resource_type> "
	    "<resource_id>\n", PM_DEBUG_PREFIX);
	(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
}

int pmi_defrag_move_strand(int from, int to, int spare);

/*
 * Handle the PMTEST_CMD_MOVESTRAND command.
 *
 * Sub command:
 * 	PMTEST_SUBCMD_MOVE		: move a strand
 *
 * 	PMTEST_CMD_MOVESTRAND <from> <to> <spare>
 *
 *	Where "from" is the strand to be powered down so that
 *	    its core can be powered down.
 *	Where "to" is the strand on a different core to be powered
 *	    up in from's place
 *	If a strand needs to move from domain A to B, "spare"
 *	    gets powered up to replace the stand taken from domain A.
 */
static void
pmtest_handle_movestrand(char *cmd, int fd)
{
	char *end;
	int from, to, spare;

	PM_DEBUG("cmd(%s)\n", cmd);
	cmd += strlen(PMTEST_CMD_MOVESTRAND);

	if (strncmp(cmd, PMTEST_SUBCMD_MOVE,
	    strlen(PMTEST_SUBCMD_MOVE)) == 0) {
		cmd += strlen(PMTEST_SUBCMD_MOVE);

	if ((cmd = strtok(NULL, " ")) == NULL)
		if ((end = strstr(cmd, " ")) == NULL) {
			PM_DEBUG("%s: Missing cli parameters.\n", __func__);
			return;
		}
		from = atoi(cmd);

		cmd = end + 1;
		if ((end = strstr(cmd, " ")) == NULL) {
			PM_DEBUG("%s: Missing cli parameters.\n", __func__);
			return;
		}
		to = atoi(cmd);

		cmd = end + 1;
		if ((end = strstr(cmd, " ")) == NULL) {
			PM_DEBUG("%s: Missing cli parameters.\n", __func__);
			return;
		}
		spare = atoi(cmd);

		PM_DEBUG("move: from(%d) to(%d) spare(%d)\n",
		    from, to, spare);
		(void) pmi_defrag_move_strand(from, to, spare);
	} else {
		PM_DEBUG("%s: Unexpected movestrand sub cmd : %s\n",
		    PM_DEBUG_PREFIX, cmd);
		(void) write(fd, PMTEST_RESP_NAK, sizeof (PMTEST_RESP_NAK));
		return;
	}

	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
}



/*
 * Handle the PMTEST_CMD_QUIT command.
 */
static void
pmtest_handle_quit(char *cmd, int fd)
{
	PM_DEBUG("%s: recv'd quit command\n", PM_DEBUG_PREFIX);

	(void) write(fd, PMTEST_RESP_ACK, sizeof (PMTEST_RESP_ACK));
}

/*
 * *************************************************************************
 * Main socket and command processor routine
 * *************************************************************************
 */

/*
 * Set up a socket and listen for incoming connections.  When a connection
 * is established (usually telnet is the client) then various commands can
 * be executed to set utilization for cpus, trigger utilization events, set
 * policy values, enable/disable cores, enable/disable strands, add/remove
 * vcpu(s), trigger fma events, etc.
 */
void *
pmtest_thread(void *rubbish)
{
	int fd, conn_fd;
	struct sockaddr_in s = { 0 };
	struct sockaddr in = { 0 };
	int len;
	char buffer[201];
	int b;
	char *cmd, *end;

	PM_DEBUG("%s: %s socket listen thread started\n", __func__,
	    PM_DEBUG_PREFIX);

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (fd < 0) {
		perror("powermgmt socket() failed");
		goto exit;
	}

	PM_DEBUG("%s: socket ok\n", PM_DEBUG_PREFIX);

	s.sin_port = htons(PMTEST_PORT);
	s.sin_family = AF_INET;
	s.sin_addr.s_addr = htons(INADDR_ANY);
	if (bind(fd, (struct sockaddr *)& s, sizeof (s)) != 0) {
		perror("pm_testlib bind failed.  Control-C ldmd then wait "
		    "2 mins and restart.");
		return (NULL);
	}

	PM_DEBUG("%s: socket bind ok\n", PM_DEBUG_PREFIX);

	/*
	 * Loop forever as this is the socket listener
	 */
	while (1) {
		if (listen(fd, 1) != 0) {
			perror("powermgmt listen() failed");
			goto exit;
		}

		PM_DEBUG("%s: listen ok\n", PM_DEBUG_PREFIX);

		len = sizeof (in);
		if ((conn_fd = accept(fd, (struct sockaddr *)&in,
		    &len)) == -1) {
			perror("powermgmt accept() failed");
			goto exit;
		}

		/* Process this connection until reset by client */
		while (1) {
			/* Read from socket */
			b = read(conn_fd, buffer, 200);
			if (b <= 0) {
				/* Connection reset */
				close(conn_fd);
				break;
			}

			cmd = buffer;
			if (!(end = strstr(cmd, "\n"))) {
				close(conn_fd);
				break;
			}

			*end = 0;

			PM_DEBUG("%s: command: %s\n", PM_DEBUG_PREFIX, cmd);

			/* Process commands */
			if (strncmp(cmd, PMTEST_CMD_UTIL,
			    strlen(PMTEST_CMD_UTIL)) == 0) {
				pmtest_q(pmtest_handle_util, cmd, conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_DIR,
			    strlen(PMTEST_CMD_DIR)) == 0) {
				pmtest_q(pmtest_handle_dir, cmd, conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_WM,
			    strlen(PMTEST_CMD_WM)) == 0) {
				pmtest_q(pmtest_handle_wm, cmd, conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_POLICY,
			    strlen(PMTEST_CMD_POLICY)) == 0) {
				pmtest_q(pmtest_handle_policy, cmd, conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_CLOCK,
			    strlen(PMTEST_CMD_CLOCK)) == 0) {
				pmtest_q(pmtest_handle_clock, cmd, conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_CORE,
			    strlen(PMTEST_CMD_CORE)) == 0) {
				pmtest_q(pmtest_handle_core, cmd, conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_STRAND,
			    strlen(PMTEST_CMD_STRAND)) == 0) {
				pmtest_q(pmtest_handle_strand, cmd, conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_DRCPU,
			    strlen(PMTEST_CMD_DRCPU)) == 0) {
				pmtest_q(pmtest_handle_drcpu, cmd, conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_PMCPU,
			    strlen(PMTEST_CMD_PMCPU)) == 0) {
				pmtest_q(pmtest_handle_pmcpu, cmd, conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_FMA,
			    strlen(PMTEST_CMD_FMA)) == 0) {
				pmtest_q(pmtest_handle_fma, cmd, conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_DEFRAG,
			    strlen(PMTEST_CMD_DEFRAG)) == 0) {
				pmtest_q(pmtest_handle_defrag, cmd, conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_PMSTAT,
			    strlen(PMTEST_CMD_PMSTAT)) == 0) {
				pmtest_q(pmtest_handle_pmstat, cmd, conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_MOVESTRAND,
			    strlen(PMTEST_CMD_PMSTAT)) == 0) {
				pmtest_q(pmtest_handle_movestrand, cmd,
				    conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_QUIT,
			    strlen(PMTEST_CMD_QUIT)) == 0) {
				pmtest_handle_quit(cmd, conn_fd);
				close(conn_fd);
				break;
			} else if (strncmp(cmd, PMTEST_CMD_PRI,
			    strlen(PMTEST_CMD_PRI)) == 0) {
				pmtest_q(pmtest_handle_pri, cmd, conn_fd);
			} else if (strncmp(cmd, PMTEST_CMD_MEMORY,
			    strlen(PMTEST_CMD_MEMORY)) == 0) {
				pmtest_q(pmtest_handle_memory, cmd, conn_fd);
			} else {
				PM_DEBUG("%s: Unexpected command : %s\n",
				    PM_DEBUG_PREFIX, cmd);
				(void) write(conn_fd, PMTEST_RESP_NAK,
				    sizeof (PMTEST_RESP_NAK));
				continue;
			}

		}
	}
exit:
	close(conn_fd);
	close(fd);
	PM_DEBUG("ldmd exiting\n");
	return (NULL);
}

/*
 * *************************************************************************
 * Stuff to fake our presence as a domain service plugin
 * *************************************************************************
 */

/* Need to have a version */
static prot_version_t pmtestlib_versions[] =
{
	{1, 0}
};

/* We're required to provide this function */
svc_ops_t *
ops(char *svc_name)
{
	return (NULL);
}

/*
 * We're required to provide this function.
 *
 * This is the function that gets called when we first get loaded.
 * We rely on this to happen.  If this were to change, it may break us.
 * See get_plugged_services() in loadsvcs.c.
 */
svc_info_t *
info(void)
{
	svc_info_t	*info;
	void		*dl_hdl;

	/*
	 * We need to fake this struct since Zeus references this
	 * and some of it's members without checking for NULL values.
	 */
	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 1;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(SP_CHAN_NAME);
	info->num_svcs = 1;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup("PM_TESTLIB");
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = 0;
	info->num_vers = NUM_VERS(pmtestlib_versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));

	memcpy(info->versions, pmtestlib_versions,
	    info->num_vers * sizeof (prot_version_t));

	/*
	 * Load ourselves to bump our library's reference count within Zeus.
	 * This way we don't get unloaded after the initial load attempt.
	 */
	dl_hdl = load_library("lib/ds/pm_test-1.0_lib.so");
	if (dl_hdl == NULL) {
		PM_DEBUG("Unable to open powermgmt test plugin: %s", dlerror());
		return (info);
	}

	/* Start the socket listening thread */
	if (0 != thr_create(NULL, NULL, pmtest_thread, NULL, THR_DAEMON, NULL))
		efatal("Could not create pmtest_thread");

	return (info);
}
