/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <stdlib.h>
#include <macros.h>
#include <alloca.h>
#include <time.h>
#include <tzfile.h>

#include "basics.h"
#include "support.h"
#include "zprintf.h"
#include "fatal.h"
#include "chain.h"
#include "tod.h"
#include "resource_id.h"
#include "ldom.h"
#include "allocate.h"
#include "database.h"
#include "database_internal.h"
#include "xml_common.h"
#include "ldomconfigure.h"
#include "mdesc_impl.h"
#include "vio.h"
#include "vcpu.h"
#include "sequencer.h"
#include "errors.h"
#include "policy.h"

/* policy.c is for Dynamic Resource Management policies */

/*
 * Unlink and free a policy tied to a given ldp. And clean
 * out the policy instance in the database.
 */
bool_t
destroy_one_policy(ldom_t *ldp, char *policy_name)
{
	policy_t	*polp;
	policy_t	*next;

	/* caller is responsible for locking */
	CHAIN_ITER_VOL(ldp->pchain.policy, polp, next, link) {
		if (strcmp(policy_name, polp->namep) == 0) {
			CHAIN_UNLINK(ldp->pchain.policy, polp, link);
			Xfree(polp->namep);
			Xfree(polp);
			database_ldom_destroy_policy_inst(ldp, policy_name);
			return (true);
		}
	}

	return (false);
}

/*
 * Unlink and free all policies tied to a given ldp.  And clean
 * out the policy instance in the database.
 */
void
free_policies(ldom_t *ldp)
{
	policy_t	*polp;
	policy_t	*next;

	mutex_lock(&ldp->pchain.lock);
	if (CHAIN_LENGTH(ldp->pchain.policy) > 0) {
		CHAIN_ITER_VOL(ldp->pchain.policy, polp, next, link) {
			CHAIN_UNLINK(ldp->pchain.policy, polp, link);
			Xfree(polp->namep);
			Xfree(polp);
		}
	}
	mutex_unlock(&ldp->pchain.lock);
	database_save();
}

void
ldp_set_policy(ldom_t *ldp, policy_t *polp)
{
	mutex_lock(&ldp->pchain.lock);

	if (find_policy_by_name(ldp, polp->namep) != NULL)
		destroy_one_policy(ldp, polp->namep);

	polp->ldp = ldp;
	CHAIN_LINK_INIT(polp->link);
	CHAIN_ADD(ldp->pchain.policy, polp, link);

	mutex_unlock(&ldp->pchain.lock);
}

static void
print_drm_policies(ldom_t *ldp, bool_t disabled, response_t *resp,
    dispctl_t *dctl)
{
	policy_t	*polp;
	policy_t	*epolp = NULL;	/* effective policy */
	char		*bgn;
	char		*end;
	char		stat[9];	/* 9 == strlen("disabled") + 1 */
	int		tod;
	int		priority;

	tod = current_tod();
	priority = INVALID_POLICY_PRIORITY;

	/*
	 * TRANSLATION_NOTE
	 * The following are column headers in a table with
	 * fixed column widths.
	 * WIDTH  HEADER	MEANING
	 * 8	  "STATUS"	enabled 'on' or 'off'
	 * 4	  "PRI"		priority
	 * 3	  "MIN"		mimumum cpus
	 * 3	  "MAX"		maximum cpus
	 * 3	  "LO"		lower bound of cpus
	 * 3	  "UP"		upper bound of cpus
	 * 8	  "BEGIN"	time of day start
	 * 8	  "END"		time of day stop
	 * 4	  "RATE"	sample rate
	 * 4	  "EM"		elastic margin
	 * 3	  "ATK"		attack rate
	 * 3	  "DK"		decay rate
	 *	  "NAME"	policy name
	 */
	rprintf(resp, "%s\n", dctl_label("policy", dctl));
	if (!dctl->parseable) {
		rprintf(resp, "    %-8s %-4s %-3s %-3s %-3s %-3s %-8s %-8s "
		    "%-4s %-3s %-3s %-3s %s\n",
		    gettext("STATUS"), gettext("PRI"), gettext("MIN"),
		    gettext("MAX"), gettext("LO"), gettext("UP"),
		    gettext("BEGIN"), gettext("END"), gettext("RATE"),
		    gettext("EM"), gettext("ATK"), gettext("DK"),
		    gettext("NAME"));
	}

	mutex_lock(&ldp->pchain.lock);
	/* Find the effective policy for active domains */
	if (ldp->state == ACTIVE) {
		CHAIN_ITER(ldp->pchain.policy, polp, link) {
			if (ldp->state != ACTIVE)
				break;
			if (tod < polp->tod_begin || tod > polp->tod_end)
				continue;
			if (polp->enable == false)
				continue;
			/* Look for best/oldest priority */
			if (priority == INVALID_POLICY_PRIORITY ||
			    polp->priority < priority) {
				priority = polp->priority;
				epolp = polp;
			}
		}
	}

	CHAIN_ITER(ldp->pchain.policy, polp, link) {
		char fmt_max[MAXLEN_VALID_NAME];
		char fmt_attack[MAXLEN_VALID_NAME];

		if (disabled)
			strcpy(stat, "disabled");
		else if (epolp == polp)
			strcpy(stat, "running");
		else if (polp->enable == true)
			strcpy(stat, "on");
		else
			strcpy(stat, "off");
		if (polp->vcpu_max == POLICY_UNLIMITED_CPU)
			sprintf(fmt_max, "U");
		else
			sprintf(fmt_max, "%d", polp->vcpu_max);
		if (polp->attack == POLICY_UNLIMITED_ATTACK)
			sprintf(fmt_attack, "U");
		else
			sprintf(fmt_attack, "%d", polp->attack);
		bgn = Xstrdup(tod2str(polp->tod_begin));
		end = Xstrdup(tod2str(polp->tod_end));

		if (dctl->parseable) {
			rprintf(resp, "|status=%s|priority=%d|vcpu-min=%d"
			    "|vcpu-max=%s|util-lower=%d|util-upper=%d"
			    "|tod-begin=%s|tod-end=%s|sample-rate=%d"
			    "|elastic-margin=%d|attack=%s|decay=%d|name=%s\n",
			    stat, polp->priority, polp->vcpu_min,
			    fmt_max, polp->util_lower, polp->util_upper,
			    bgn, end, polp->sample_rate, polp->elastic_margin,
			    fmt_attack, polp->decay, polp->namep);
		} else {
			rprintf(resp, "    %-8s %-4d %-3d %-3s %-3d %-3d %-8s "
			    "%-8s %-4d %-3d %-3s %-3d %s\n", stat,
			    polp->priority, polp->vcpu_min, fmt_max,
			    polp->util_lower, polp->util_upper, bgn, end,
			    polp->sample_rate, polp->elastic_margin,
			    fmt_attack, polp->decay, polp->namep);
		}
		Xfree(bgn);
		Xfree(end);
	}
	mutex_unlock(&ldp->pchain.lock);

	if (!dctl->parseable) {
		if (disabled) {
			rprintf(resp, gettext("    "));
			rprintf(resp, MSG_ERR_DRM_DISABLED_BY_PM);
		}
		rprintf(resp, gettext("\n"));
	}
}

static void
print_drm_stats(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	policy_t	*polp = NULL;
	bool_t		running = false;
	int		tod;

	tod = current_tod();

	mutex_lock(&ldp->pchain.lock);
	CHAIN_ITER(ldp->pchain.policy, polp, link) {
		if (polp->enable == true && tod >= polp->tod_begin &&
		    tod <= polp->tod_end) {
			running = true;
			break;
		}
	}
	mutex_unlock(&ldp->pchain.lock);
	if (!running)
		return;

	rprintf(resp, "%s\n", dctl_label("weighted mean utilization", dctl));
	if (dctl->parseable) {
		rprintf(resp, gettext("|mean-util=%.1f%%"),
		    ldp->pchain.mean_util[0]);
	} else if (ldp->pchain.mean_util[0] < 0.0 ||
	    ldp->pchain.mean_util[0] > 100.0) {
		rprintf(resp, gettext("    ---\n"));
	} else if (ldp->pchain.mean_util[0] >= 10.0) {
		rprintf(resp, gettext("    %.0f%%\n"),
		    ldp->pchain.mean_util[0]);
	} else {
		rprintf(resp, gettext("    %.1f%%\n"),
		    ldp->pchain.mean_util[0]);
	}

	rprintf(resp, gettext("\n"));
}

void
print_resmgmt(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	bool_t	pm_elastic = false;

	if (dctl->oformat && !dctl->oresmgmt)
		return;

	/* PM must be in performance mode */
	if (!pm_migration_is_possible())
		pm_elastic = true;

	if (CHAIN_LENGTH(ldp->pchain.policy) > 0) {
		print_drm_policies(ldp, pm_elastic, resp, dctl);
		if (ldp->state >= ACTIVE && pm_elastic == false)
			print_drm_stats(ldp, resp, dctl);
	}
}

policy_t *
find_policy_by_name(ldom_t *ldp, char *policy_name)
{
	policy_t	*polp;

	/* caller is responsible for locking */
	CHAIN_ITER(ldp->pchain.policy, polp, link) {
		if (strcmp(policy_name, polp->namep) == 0)
			return (polp);
	}

	return (NULL);
}

void
calc_mean(ldom_t *ldp, policy_t *polp)
{
	double	weight = MAX_POLICY_SAMPLES;
	double	weight_vcpuUtil = 0.0;
	double	curr_weight_vcpuUtil = 0.0;
	double	total_weight = 0.0;
	int	i;

	/* Rotate in the new mean calculation */
	for (i = MAX_POLICY_SAMPLES - 1; i >= 0; --i)
		ldp->pchain.mean_util[i] = ldp->pchain.mean_util[i - 1];

	ldp->pchain.mean_util[0] =
	    ldp->stats.hist_table[ldp->stats.hist_idx].value;
	/*
	 * Weighted mean utilization
	 */
	for (i = 0; i < MAX_POLICY_SAMPLES; ++i, --weight) {
		curr_weight_vcpuUtil = ldp->pchain.mean_util[i] * weight;
		weight_vcpuUtil += curr_weight_vcpuUtil;
		total_weight += weight;
	}
	ldp->pchain.mean_util[0] = weight_vcpuUtil / total_weight;
}

void
check_vcpu_limits(ldom_t *ldp, policy_t *polp)
{
	double	util_required = 0;
	double	vcpu_required = 0;
	double	mean_util = 0;
	int	delta_vcpus = 0;
	int	nfree_vcpus = 0;
	int	num_vcpus = 0;

	mean_util = (int)(ldp->pchain.mean_util[0] + 0.5);
	nfree_vcpus = num_free_vcpus();
	num_vcpus = CHAIN_LENGTH(ldp->bound.vcpus);

	/* Utilization below the low water mark */
	if (mean_util < polp->util_lower) {
		if (num_vcpus < polp->vcpu_min) {
			delta_vcpus = polp->vcpu_min - num_vcpus;
			delta_vcpus = min(delta_vcpus, nfree_vcpus);
			if (polp->attack != POLICY_UNLIMITED_ATTACK)
				delta_vcpus = min(delta_vcpus, polp->attack);
			drm_vcpus(ldp, delta_vcpus, ACT_ADD);
		} else {
			delta_vcpus = num_vcpus - polp->vcpu_min;
			delta_vcpus = min(delta_vcpus, polp->decay);
			util_required = polp->util_lower + polp->elastic_margin;
			vcpu_required = num_vcpus * ((mean_util +
			    polp->elastic_margin) / 100.0);
			vcpu_required = (vcpu_required * 100.0) / util_required;
			DBG9(zprintf("vLWM: vcpu_required=%f\n",
			    vcpu_required));
			/* round up to next whole number */
			delta_vcpus = min(delta_vcpus,
			    (int)(vcpu_required + 0.5));
			drm_vcpus(ldp, delta_vcpus, ACT_REMOVE);
		}
		return;
	}

	/* Utilization above the high water mark */
	if (mean_util > polp->util_upper) {
		if (polp->vcpu_max == POLICY_UNLIMITED_CPU) {
			if (polp->attack == POLICY_UNLIMITED_ATTACK)
				delta_vcpus = nfree_vcpus;
			else
				delta_vcpus = min(nfree_vcpus, polp->attack);
			drm_vcpus(ldp, delta_vcpus, ACT_ADD);
		} else if (num_vcpus > polp->vcpu_max) {
			delta_vcpus = num_vcpus - polp->vcpu_max;
			delta_vcpus = min(delta_vcpus, polp->decay);
			drm_vcpus(ldp, delta_vcpus, ACT_REMOVE);
		} else {
			delta_vcpus = polp->vcpu_max - num_vcpus;
			if (polp->attack != POLICY_UNLIMITED_ATTACK)
				delta_vcpus = min(delta_vcpus, polp->attack);
			util_required = polp->util_upper - polp->elastic_margin;
			vcpu_required = num_vcpus * ((mean_util -
			    polp->elastic_margin) / 100.0);
			vcpu_required = (vcpu_required * 100.0) / util_required;
			DBG9(zprintf("^HWM: vcpu_required=%f\n",
			    vcpu_required));
			/* round up to next whole number */
			delta_vcpus = min(delta_vcpus,
			    (int)(vcpu_required + 0.5));
			drm_vcpus(ldp, delta_vcpus, ACT_ADD);
		}
		return;
	}

	/* Number of vcpus above maximum */
	if (polp->vcpu_max != POLICY_UNLIMITED_CPU &&
	    num_vcpus > polp->vcpu_max) {
		delta_vcpus = num_vcpus - polp->vcpu_max;
		delta_vcpus = min(delta_vcpus, polp->decay);
		drm_vcpus(ldp, delta_vcpus, ACT_REMOVE);
		return;
	}

	/* Number of vcpus below minimum */
	if (num_vcpus < polp->vcpu_min) {
		delta_vcpus = polp->vcpu_min - num_vcpus;
		delta_vcpus = min(delta_vcpus, nfree_vcpus);
		if (polp->attack != POLICY_UNLIMITED_ATTACK)
			delta_vcpus = min(delta_vcpus, polp->attack);
		drm_vcpus(ldp, delta_vcpus, ACT_ADD);
		return;
	}
}

void
process_stats(ldom_t *ldp, time_t now)
{
	policy_t	*polp;
	char		winner[MAXLEN_VALID_NAME] = { NULL };
	int		tod;
	int		priority;

	ASSERT_IN_SEQUENCER();

	/* No policies to check */
	if (CHAIN_LENGTH(ldp->pchain.policy) < 1)
		return;

	/* Make sure PM is in performance mode */
	if (pm_migration_is_possible() == false)
		return;

	/*
	 * LDom is being modified or migrated, but no need to
	 * grab the lock now.  Instead check again and lock
	 * when an actual change is ready to be made.
	 */
	if (IS_LDOM_LOCKED(ldp) == true)
		return;

	tod = current_tod();
	priority = INVALID_POLICY_PRIORITY;

	mutex_lock(&ldp->pchain.lock);

	/*
	 * Policies exist and the lock is now held
	 */
	CHAIN_ITER(ldp->pchain.policy, polp, link) {
		/* Check effectivity */
		if (tod < polp->tod_begin || tod > polp->tod_end)
			continue;
		if (polp->enable == false)
			continue;
		/* Look for best/oldest priority */
		if (priority == INVALID_POLICY_PRIORITY ||
		    polp->priority < priority) {
			priority = polp->priority;
			strcpy(winner, polp->namep);
		}
	}

	/* No effective policy could be found */
	if (priority == INVALID_POLICY_PRIORITY) {
		mutex_unlock(&ldp->pchain.lock);
		return;
	}

	/*
	 * Cannot run any faster than the sample rate, and check for
	 * a change in clock or a new day.
	 */
	polp = find_policy_by_name(ldp, winner);
	if (tod >= polp->tm_stamp + polp->sample_rate || tod < polp->tm_stamp) {
		/*
		 * See if DRM is just starting or has been locked out for
		 * some time. This is just a simple first-order attempt to
		 * dampen a potentially vigorous calculation.
		 */
		if (polp->tm_stamp == 0 || tod > polp->tm_stamp +
		    (STATS_STALE_FACTOR *  polp->sample_rate))
			seed_mean_util_stats(ldp);
		polp->tm_stamp = tod;
		DBG9(zprintf("%s: winner=%s\n", __func__, winner));
		calc_mean(ldp, polp);
		check_vcpu_limits(ldp, polp);
	}
	mutex_unlock(&ldp->pchain.lock);
}

/*
 * FUTURE WORK: Once drm runs in its own thread, more locking
 * will be needed to prevent collisions, say adding a cpu to a
 * domain that is being migrated at the same time.
 */
void
drm_vcpus(ldom_t *ldp, int num_vcpus, int action)
{
	response_t	*resp;
	cpu_adm_t	*cpu_adm = NULL;

	ASSERT_IN_SEQUENCER();

	/* Nothing to do */
	if (num_vcpus < 1)
		return;

	/* PM must be in performance mode */
	if (pm_migration_is_possible() == false)
		return;

	/* bit bucket output */
	init_response(&resp);
	cpu_adm = Xmalloc(sizeof (*cpu_adm));
	cpu_adm->type = CPU_ADM_NUM;
	cpu_adm->num =  num_vcpus;

	/* Is LDom being modified or migrated */
	if (!config_mod_allowed(ldp, DEL_RECONFIG_OTHER, DR_OP_TYPE_DOMAIN,
	    resp))
		return;

	if (action == ACT_ADD)
		cpu_add(ldp, cpu_adm, resp);
	else if (action == ACT_REMOVE)
		cpu_del(ldp, cpu_adm, resp, false);

	Xfree(cpu_adm);
	fini_response(resp);
}

void
seed_mean_util_stats(ldom_t *ldp)
{
	double		seed;
	int		i;

	seed = ldp->stats.hist_table[ldp->stats.hist_idx].value;
	DBG9(zprintf("%s: seed=%.1f%%\n", __func__, seed));
	if (seed < 0.0 || seed > 100.0)
		seed = 0.0;

	for (i = 0; i < MAX_POLICY_SAMPLES; ++i)
		ldp->pchain.mean_util[i] = seed;
}

/*
 * Only have to find one policy enabled.
 */
bool_t
is_drm_in_progress(void)
{
	ldom_t		*ldp = NULL;
	policy_t	*polp = NULL;

	/* Called from PM via the sequencer */
	ASSERT_IN_SEQUENCER();

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		CHAIN_ITER(ldp->pchain.policy, polp, link) {
			if (polp->enable == true)
				return (true);
		}
	}

	return (false);
}
