/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/* policy.h is for Dyamic Resource Management policies */

#ifndef	_POLICY_H_
#define	_POLICY_H_

#ifdef __cplusplus
extern "C" {
#endif

#include <thread.h>
#include <synch.h>

#include "allocate.h"
#include "basics.h"
#include "chain.h"
#include "dispctl.h"
#include "ldmd_types.h"
#include "response.h"


#define	INVALID_POLICY_PROPERTY		(-1)
#define	INVALID_POLICY_PRIORITY		INVALID_POLICY_PROPERTY
#define	INVALID_POLICY_CPU		INVALID_POLICY_PROPERTY
#define	INVALID_POLICY_UTIL		INVALID_POLICY_PROPERTY
#define	INVALID_POLICY_TOD		0xFFFFFFFFFFFFFFFF
#define	INVALID_POLICY_SAMPLE_RATE	INVALID_POLICY_PROPERTY
#define	INVALID_POLICY_ELASTIC_MARGIN	INVALID_POLICY_PROPERTY
#define	INVALID_POLICY_ATTACK		INVALID_POLICY_PROPERTY
#define	INVALID_POLICY_DECAY		INVALID_POLICY_PROPERTY

#define	POLICY_DEFAULT_PRIORITY		99
#define	POLICY_MIN_CPU			1
#define	POLICY_DEFAULT_VCPU_MIN		POLICY_MIN_CPU
#define	POLICY_UNLIMITED_CPU		0
#define	POLICY_DEFAULT_UTIL_LOWER	60
#define	POLICY_DEFAULT_UTIL_UPPER	85
#define	POLICY_START_OF_DAY		0
#define	POLICY_END_OF_DAY		((23 * 3600) + (59 * 60) + 59)
#define	POLICY_DEFAULT_SAMPLE_RATE	10
#define	POLICY_DEFAULT_ELASTIC_MARGIN	5
#define	POLICY_UNLIMITED_ATTACK		0
#define	POLICY_DEFAULT_DECAY		1

/*
 * These unlimited default values were once reserved only for the control
 * and service domains, but it was deemed appropriate (if not cavalier)
 * to make them the default for all domains in order to keep the Zeus
 * syntax consistent.
 */
#define	POLICY_DEFAULT_VCPU_MAX		POLICY_UNLIMITED_CPU
#define	POLICY_DEFAULT_ATTACK		POLICY_UNLIMITED_ATTACK

#define	MAX_NUM_POLICIES		1<<8	/* 256 policy max */
#define	MAX_POLICY_SAMPLES		4	/* per original code */
#define	STATS_STALE_FACTOR		10

#define	BLOCK_ERR_BAD_TOD \
{	char *bgn; \
	char *end; \
	bgn = Xstrdup(tod2str(tod_begin)); \
	end = Xstrdup(tod2str(tod_end)); \
	rprintf(resp, gettext("tod_begin=%s cannot be greater " \
	    "than or equal to tod_end=%s\n"), bgn, end); \
	Xfree(bgn); \
	Xfree(end); \
}


typedef struct policy {
	ldom_t		*ldp;			/* domain owner */
	char		*namep;
	bool_t		enable;
	int		priority;
	int		vcpu_min;
	int		vcpu_max;
	int		util_lower;
	int		util_upper;
	int		tod_begin;
	int		tod_end;
	int		sample_rate;		/* in seconds */
	int		elastic_margin;
	int		attack;
	int		decay;
	int		tm_stamp;
	CHAIN_LINK(struct policy, link);	/* next policy */
} policy_t;

typedef struct policy_chain {
	mutex_t		lock;
	CHAIN(policy_t, policy);
	double		mean_util[MAX_POLICY_SAMPLES];
} policy_chain_t;

bool_t destroy_one_policy(ldom_t *ldp, char *policy_name);
void free_policies(ldom_t *ldp);
void polp_add_val(policy_t *polp, char *name, int val);
void ldp_set_policy(ldom_t *ldp, policy_t *polp);
void print_resmgmt(ldom_t *ldp, response_t *resp, dispctl_t *dctl);
policy_t *find_policy_by_name(ldom_t *ldp, char *policy_name);
void process_stats(ldom_t *ldp, time_t now);
void drm_vcpus(ldom_t *ldp, int num_vcpus, int action);
void seed_mean_util_stats(ldom_t *ldp);
bool_t is_drm_in_progress(void);


#ifdef __cplusplus
}
#endif

#endif /* _POLICY_H_ */
