/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <basics.h>
#include <fatal.h>
#include <macros.h>

#include "ldmd_types.h"
#include "allocate.h"
#include "chain.h"
#include "mdesc.h"
#include "mdesc_impl.h"
#include "ldom.h"
#include "zpri.h"
#include "backend.h"
#include "zprintf.h"
#include "support.h"
#include "core_md.h"
#include "memory.h"
#include "console.h"

#define	MEM_LG_NODE "memory-latency-group"
#define	DMA_LG_NODE "dma-latency-group"
#define	CACHE_NODE "cache"

#define	ADDR_MASK_PROP "address-mask"
#define	INDEX_MASK_PROP "index-mask"

static void pri_init_firmware(md_t *mdp);
static void pri_init_memory(md_t *mdp);
static void pri_init_cpus(md_t *mdp);
static void pri_init_cores(md_t *mdp);
static void pri_init_maus(md_t *mdp);
static void pri_init_phys_io(md_t *mdp);

extern bool_t is_hvctl_ver_supported(uint64_t maj, uint64_t min);
zeus_pri_t pri;

/*
 * Find the most significant bit, returning its bit position + 1.
 */
static int
find_msb(uint64_t n)
{
	int i;

	for (i = 0; n > 0; i++)
		n >>= 1;

	return (i);
}

/*
 * Search the MD for nodes named 'node_name' and find the maximum
 * value of the property named 'prop_name' across all these nodes
 */
static uint64_t
pri_max_prop(md_t *mdp, char *node_name, char *prop_name, bool_t warn)
{
	md_node_t **nodelistp;
	int i, num;
	uint64_t max_val = 0;

	nodelistp = Xmalloc(sizeof (md_node_t) * md_node_count(mdp));
	num = md_scan_dag(mdp, NULL, md_find_name(mdp, node_name),
	    md_find_name(mdp, "fwd"), nodelistp);

	/*
	 * For each matching node, read the value of prop_name and find the
	 * maximum
	 */
	for (i = 0; i < num; i++) {
		md_node_t *nodep;
		uint64_t prop_val;

		nodep = nodelistp[i];
		if (md_get_prop_val(mdp, nodep, prop_name, &prop_val) < 0) {
			if (warn == true)
				warning("No %s property in %s node", prop_name,
				    node_name);
			continue;
		}

		if (prop_val > max_val)
			max_val = prop_val;
	}

	Xfree(nodelistp);
	DBG(zprintf("Found %d %s nodes, max %s = %llx\n", num, node_name,
	    prop_name, max_val));

	return (max_val);
}

static void
add_console_device(md_t *md)
{
	md_node_t *vdevs, *console;
	char *compatible = "SUNW,sun4v-console";

	vdevs = md_find_node(md, "virtual-devices");
	console = md_link_new_node(md, "virtual-device", vdevs, "fwd", "back");
	md_add_string_property(md, console, "name", "console");
	md_add_string_property(md, console, "device-type", "serial");
	md_add_value_property(md, console, "intr", 0x1);
	md_add_value_property(md, console, "ino", LDC_CONSOLE_INO);
	md_add_value_property(md, console, "channel#", 0x0);
	md_add_value_property(md, console, "cfg-handle", 0x1);
	md_add_data_property(md, console, "compatible", strlen(compatible) + 1,
	    (uint8_t *)compatible);
}

/*
 * Add a devaliases node suitable for logical domains other than the
 * control domain.
 */
static void
add_devaliases(md_t *md)
{
	md_node_t *openboot, *devalias;

	openboot = md_find_node(md, "openboot");
	devalias = md_link_new_node(md, "devalias", openboot,
	    "fwd", "back");

	/*
	 * for a guest md template,  add virtual-console alias only.
	 * "disk" and "net" aliases will be added when guest md actually
	 * created.
	 */
	md_add_string_property(md, devalias, "virtual-console",
	    "/virtual-devices/console@1");
}

static md_t *
create_guest_template(md_t *pri_md)
{
	md_t *template_md;

	template_md = md_clone(pri_md);
	md_find_delete_node(template_md, "domain-services");
	md_find_delete_node(template_md, "channel-devices");
	md_find_delete_node(template_md, "channel-endpoints");
	md_find_delete_node(template_md, "ldc_endpoints");
	md_find_delete_node(template_md, "firmware");
	md_find_delete_node(template_md, "devalias");
	md_find_delete_node(template_md, "components");
	md_find_delete_node(template_md, "memory-segments");

	/* FIXME PRI temporary until have it in PRI */
	add_console_device(template_md);
	add_devaliases(template_md);

	md_scour(template_md, md_find_name(template_md, "back"));
	return (template_md);
}

static md_t *
create_core_md(md_t *pri_md)
{
	md_t *core_md;

	core_md = md_clone(pri_md);
	md_find_delete_node(core_md, "domain-services");
	md_find_delete_node(core_md, "virtual-devices");
	md_find_delete_node(core_md, "channel-endpoints");
	md_find_delete_node(core_md, "ldc_endpoints");
	md_find_delete_node(core_md, "firmware");

	augment_core_md(core_md);
	md_scour(core_md, md_find_name(core_md, "back"));
	return (core_md);
}

int
pri_initialize(void)
{
	uint64_t cache_imask, memlg_amask, dma_amask, max_mask;

	pri.have_cwq = false;

	if ((pri.pri = backend.get_pri()) == NULL)
		return (-1);

	pri.guest_template = create_guest_template(pri.pri);
	pri.core_mdp = create_core_md(pri.pri);

	DBG(dump_md_to_file(pri.guest_template, "guest_template"));
	DBG(dump_md_to_file(pri.core_mdp, "core_md"));
	DBG(dump_md_to_file(pri.pri, "pri"));

	/*
	 * Scan the given MD for the info we really need ... namely cpu
	 * strands, physical memory, and physical devices to assign.
	 */

	CHAIN_INIT(pri.image_list);	/* bootable images */
	CHAIN_INIT(pri.container_list);	/* PROM list */

	DBG(zprintf("Physical Resource Inventory configuration:\n"));

	pri_init_memory(pri.pri);
	pri_init_firmware(pri.pri);
	pri_init_cpus(pri.pri);
	pri_init_cores(pri.pri);
	pri_init_maus(pri.pri);
	pri_init_phys_io(pri.pri);

	/*
	 * Read the max value of index-mask from cache nodes, and the
	 * addresss-mask values from the relevant latency group nodes
	 * and find the maximum
	 */
	cache_imask = pri_max_prop(pri.pri, CACHE_NODE, INDEX_MASK_PROP, false);
	memlg_amask = pri_max_prop(pri.pri, MEM_LG_NODE, ADDR_MASK_PROP, true);
	dma_amask = pri_max_prop(pri.pri, DMA_LG_NODE, ADDR_MASK_PROP, false);

	max_mask = max(cache_imask, max(memlg_amask, dma_amask));

	pri.addr_congr_mask = ((uint64_t)1 << find_msb(max_mask)) - 1;

	DBG(zprintf("Max mask = 0x%llx, congr_mask = 0x%llx\n",
	    max_mask, pri.addr_congr_mask));

	/* do any backend specific initialization */
	backend.init_pri(pri.pri);

	return (0);
}

/*
 * Get bootable images from PRI.
 */
static void
get_boot_images(md_t *pri_md, md_node_t *fwnode)
{
	md_node_t **nodelistp;
	md_node_t **imglistp;
	int num, i;

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(pri_md));
	imglistp = Xmalloc(sizeof (md_node_t *) * md_node_count(pri_md));

	DBG(zprintf("ROM images:\n"));

	num = md_scan_dag(pri_md, fwnode,
	    md_find_name(pri_md, "read_only_memory"),
	    md_find_name(pri_md, "fwd"), nodelistp);

	for (i = 0; i < num; i++) {
		md_node_t *nodep;
		pri_image_t *imgp;
		pri_img_cont_t *icp;
		int imgnum;
		char *strp;
		uint64_t pa_base, size, alignment, min_allocation;
		int j;

		nodep = nodelistp[i];

		if (md_get_prop_val(pri_md, nodep, "base", &pa_base) < 0)
			fatal("Missing ROM base address");
		if (md_get_prop_val(pri_md, nodep, "size", &size) < 0)
			fatal("Missing ROM base address");
		if (md_get_prop_str(pri_md, nodep, "name", &strp) < 0)
			fatal("Missing ROM name field");

		DBG(zprintf("\tROM: %s\t[ 0x%llx : 0x%llx ]\n", strp,
		    pa_base, pa_base+size));

		/* scan for images, if none skip this prom */

		imgnum = md_scan_dag(pri_md, nodep,
		    md_find_name(pri_md, "rom_img"),
		    md_find_name(pri_md, "fwd"), imglistp);
		if (imgnum == 0)
			continue;

		/* allocate the image container */
		icp = Xcalloc(1, pri_img_cont_t);
		icp->namep = Xstrdup(strp);
		icp->pa_base = pa_base;
		icp->size = size;
		CHAIN_LINK_INIT(icp->link);

		CHAIN_ADD(pri.container_list, icp, link);

		/* fill in the images */
		for (j = 0; j < imgnum; j++) {
			uint64_t gu, offset, size;
			char *namep, *versionp, *notep;

			nodep = imglistp[j];

			if (md_get_prop_val(pri_md, nodep, "offset",
			    &offset) < 0)
				fatal("Missing image offset");
			if (md_get_prop_val(pri_md, nodep, "size", &size) < 0)
				fatal("Missing image size");
			if (md_get_prop_str(pri_md, nodep, "name", &strp) < 0)
				fatal("Missing image name");
			namep = Xstrdup(strp);
			if (md_get_prop_val(pri_md, nodep, "guest_use",
			    &gu) < 0) {
				gu = 0;
				alignment = 0;
				min_allocation = 0;
			} else if (gu != 0) {
				if (md_get_prop_val(pri_md, nodep, "alignment",
				    &alignment) < 0)
					fatal("Missing image alignment");
				if (md_get_prop_val(pri_md, nodep,
				    "min_allocation", &min_allocation) < 0)
					fatal("Missing image min_allocation");
			}

			versionp = NULL;
			if (md_get_prop_str(pri_md, nodep, "version",
			    &strp) == 0)
				versionp = Xstrdup(strp);

			notep = NULL;
			if (md_get_prop_str(pri_md, nodep, "note", &strp) == 0)
				notep = Xstrdup(strp);

			DBG(zprintf("\t\tImage: [%s] %s [+0x%llx "
			    "size = 0x%llx]\n",
			    gu == 0 ? " " : "*", strp, offset, size));

			imgp = Xcalloc(1, pri_image_t);
			imgp->namep = namep;
			imgp->offset = offset;
			imgp->size = size;
			imgp->alignment = alignment;
			imgp->min_allocation = min_allocation;
			imgp->guest_bootable = (gu == 0) ? false : true;
			imgp->versionp = versionp;
			imgp->notep = notep;
			imgp->imgcp = icp;
			CHAIN_LINK_INIT(imgp->link);

			CHAIN_ADD(pri.image_list, imgp, link);
		}
	}

	Xfree(nodelistp);
	Xfree(imglistp);
}

/*
 * Get hypervisor version from ROM images.
 * FIXME: this should come from a firmware version property in the PRI.
 */
static char *
get_hv_version(void)
{
	pri_img_cont_t	*icp;

	CHAIN_ITER(pri.container_list, icp, link) {
		pri_image_t	*romp;

		CHAIN_ITER(pri.image_list, romp, link) {
			if (icp != romp->imgcp)
				continue;
			if (strcmp(romp->namep, "Hypervisor") == 0)
				return (romp->versionp);
		}
	}

	return (NULL);
}

static void
fatal_missing_prop(char *propname)
{
	fatal("Missing %s property in firmware node of PRI.\nPlease check that "
	    "the LDom Manager version is compatible\nwith the installed "
	    "firmware.", propname);

}

/*
 * Hardcoded values for "0.0.0" fw only
 */
#define	GUEST_RA_HASH_BITS 34LL
#define	NUM_RA_BITS 40LL
#define	MAX_MBLOCKS_PER_GUEST 32LL
#define	HV_MAX_MBLOCKS 2048

/*
 * Retrieve firmware/platform configuration parameters
 */
static void
pri_init_firmware(md_t *mdp)
{
	md_node_t **nodes, **platform_nodes;
	char *hv_version;
	uint64_t val;
	int num;
	pri_mblock_t *membp;

	/*
	 * Verify that memory has already been initialized
	 * by pri_limit_memory.
	 */
	ASSERT(CHAIN_LENGTH(pri.memory_list) > 0);

	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "firmware"),
	    md_find_name(mdp, "fwd"), nodes);

	if (num != 1)
		fatal("PRI does not contain exactly one firmware node");

	platform_nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "platform"),
	    md_find_name(mdp, "fwd"), platform_nodes);

	if (num != 1)
		fatal("PRI does not contain exactly one platform node");

	if (md_get_prop_val(mdp, platform_nodes[0], "stick-frequency",
	    &val) != 0)
		fatal("Missing stick-frequency prop in platform node of PRI");
	pri.stick_frequency = val;

	get_boot_images(mdp, nodes[0]);

	if (md_get_prop_val(mdp, nodes[0], "max_guests", &val) != 0)
		fatal("Missing max_guests property in firmware node of PRI");
	pri.max_guests = (int)val;

	if (md_get_prop_val(mdp, nodes[0], "max_hv_ldcs", &val) != 0)
		fatal("Missing max_hv_ldcs property in firmware node of PRI");
	pri.max_hv_ldcs = (int)val;

	if (md_get_prop_val(mdp, nodes[0], "max_guest_ldcs", &val) != 0)
		fatal("Missing max_guest_ldcs property in firmware "
		    "node of PRI");
	pri.max_guest_ldcs = (int)val;

	if (md_get_prop_val(mdp, nodes[0], "max_guest_vids", &val) != 0) {
		/* If not available, fall back on max-cpus in platform node */
		if (md_get_prop_val(mdp, platform_nodes[0], "max-cpus",
		    &val) != 0)
			fatal("Missing max_guest_vids property in firmware "
			    "node of PRI");
	}
	pri.max_guest_vids = (int)val;

	if (md_get_prop_val(mdp, nodes[0], "max_vcpus", &val) != 0) {
		/* If not available, fall back on max-cpus in platform node */
		if (md_get_prop_val(mdp, platform_nodes[0], "max-cpus",
		    &val) != 0)
			fatal("Missing max_vcpus property in firmware "
			    "node of PRI");
	}
	pri.max_vcpus = (int)val;

	if (md_get_prop_val(mdp, nodes[0], "max_guest_dependencies", &val) == 0)
		pri.max_guest_dependencies = (int)val;
	else {
		DBG(zprintf("Notice: Firmware does not support domain "
		    "dependency operations\n"));
		pri.max_guest_dependencies = 0;
	}

	Xfree(platform_nodes);

	if (md_get_prop_val(mdp, nodes[0], "md_elbow_room", &val) == 0)
		pri.md_elbow_room = (int)val;
	else
		pri.md_elbow_room = 0;

	if (md_get_prop_val(mdp, nodes[0], "no_pci_reset", &val) == 0 &&
	    val == 1)
		pri.no_pcie_reset = true;
	else
		pri.no_pcie_reset = false;

	pri.fw_dio_capability = 0;
	if (md_get_prop_val(mdp, nodes[0], "directio_capability", &val) == 0) {
		pri.fw_dio_capability = val;
		if (val == DIO_STATIC_CAPAB) {
			zeus.sdio_enabled =  true;
		}
	}


	/*
	 * HV version compatibility checks.
	 */
	hv_version = get_hv_version();

	if (md_get_prop_val(mdp, nodes[0], "max_mblocks", &val) == 0) {
		pri.max_mblocks = (int)val;
	} else {
		if ((hv_version == NULL) || (strcmp(hv_version, "0.0.0") != 0))
			fatal_missing_prop("max_mblocks");
		else
			/* Use hardcoded value for version 0.0.0 */
			pri.max_mblocks = HV_MAX_MBLOCKS;
	}

	if (md_get_prop_val(mdp, nodes[0], "ra2pa_shift", &val) == 0) {
		pri.ra2pa_shift = val;
	} else {
		if ((hv_version == NULL) || (strcmp(hv_version, "0.0.0") != 0))
			fatal_missing_prop("ra2pa_shift");
		else
			/* Use hardcoded value for version 0.0.0 */
			pri.ra2pa_shift = GUEST_RA_HASH_BITS;
	}

	if (md_get_prop_val(mdp, nodes[0], "max_ra_bits", &val) == 0) {
		zeus.max_ra_bits = val;
		pri.guest_max_mblocks = 1 << (zeus.max_ra_bits -
		    pri.ra2pa_shift - 1);
	} else {
		if ((hv_version == NULL) || (strcmp(hv_version, "0.0.0") != 0))
			fatal_missing_prop("max_ra_bits");
		else {
			/* Use hardcoded value fors version 0.0.0 */
			zeus.max_ra_bits = NUM_RA_BITS;
			pri.guest_max_mblocks = MAX_MBLOCKS_PER_GUEST;
		}
	}

	Xfree(nodes);

	/*
	 * See if we need to generate multiple mblocks per RA2PA segment
	 */
	num = 0;
	CHAIN_ITER(pri.memory_list, membp, link) {
		/*
		 * count the segments consumed by each mblock. A mblock
		 * containing less than 16G may consume 2 segments.
		 */
		num += (membp->size >> pri.ra2pa_shift) + 2;
	}

	if (num > pri.guest_max_mblocks) {
		/*
		 * Fake Zeus into thinking the RA2PA segments are smaller
		 * and more numerous than they actually are. This assumes
		 * hypervisor supports multiple mblocks per segment.
		 * This is a bit of hack, but less painful than
		 * rewriting the memory allocation logic.
		 */
		if (is_hvctl_ver_supported(HVCTL_MULT_SEG_MAJOR_VERSION,
		    HVCTL_MULT_SEG_MINOR_VERSION) == true) {

			warning("Sparse memory detected. Allocation of "
			    "multiple memory blocks per real memory segment "
			    "will be utilized to support this hardware "
			    "configuration.\n");

			zeus.ra2pa_shift = pri.ra2pa_shift -4;

			/*
			 * Reserve last true segment for LDC shared
			 * memory
			 */
			zeus.guest_max_mblocks =
			    (pri.guest_max_mblocks -1) << 4;

			DBG(zprintf("zeus.ra2pa_shift=%"PRId64
			    ",zeus.max_ra_bits=%"PRId64
			    ", zeus.guest_max_mblocks=%"PRId64"\n",
			    zeus.ra2pa_shift, zeus.max_ra_bits,
			    zeus.guest_max_mblocks));
			return;
		}

		warning("Sparse memory config found but HV version "
		    "does not support the workaround!");
	}

	zeus.ra2pa_shift = pri.ra2pa_shift;

	/* Reserve last true segment for LDC shared memory */
	zeus.guest_max_mblocks = pri.guest_max_mblocks -1;

	DBG(zprintf("zeus.ra2pa_shift=%"PRId64",zeus.max_ra_bits=%"PRId64","
	    "zeus.guest_max_mblocks=%"PRId64"\n",
	    zeus.ra2pa_shift, zeus.max_ra_bits, zeus.guest_max_mblocks));
}

void
pri_display_firmware(response_t *resp)
{
	pri_img_cont_t	*icp;

	CHAIN_ITER(pri.container_list, icp, link) {
		pri_image_t	*romp;

		rprintf(resp, "%s:\n", icp->namep);

		CHAIN_ITER(pri.image_list, romp, link) {
			if (icp != romp->imgcp) continue;

			rprintf(resp, "\t%s", romp->namep);
			if (romp->versionp != NULL) {
				rprintf(resp, "\tv. %s", romp->versionp);
				if (romp->notep != NULL) {
					rprintf(resp, "\t%s", romp->notep);
				}
			}
			rprintf(resp, "\n");
		}
	}
}

static void
pri_init_memory(md_t *mdp)
{
	int num, i;
	md_node_t ** nodelistp;

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	CHAIN_INIT(pri.memory_list);	/* physical memory */

	/*
	 * For each memory block - extract the info and
	 * insert into the resource chain.
	 */

	DBG(zprintf("\tMemory blocks:\n"));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "mblock"),
	    md_find_name(mdp, "fwd"), nodelistp);

	if (num == 0)
		fatal("No available physical memory");

	/*
	 * FIXME: should build list of pri mblocks and add them all
	 * to bind_info.bmblk_list to create free memory list in one call.
	 */
	for (i = 0; i < num; i++) {
		md_node_t *nodep;
		pri_mblock_t *membp;
		uint64_t base, size;
		mblock_t *mblk;

		nodep = nodelistp[i];
		if (md_get_prop_val(mdp, nodep, "base", &base) < 0)
			fatal("Missing base address");
		if (md_get_prop_val(mdp, nodep, "size", &size) < 0)
			fatal("Missing mblock size");

		DBG(zprintf("\t\tbase= 0x%llx, size=0x%llx : "
		    "[0x%llx -> 0x%llx]\n", base, size, base, base + size));

		/* Allocate the description of the physical memory block */
		membp = Xcalloc(1, pri_mblock_t);
		membp->pa_base = base;
		membp->size = size;

		CHAIN_ADD(pri.memory_list, membp, link);

		/*
		 * Add to memory free list.
		 */
		mblk = mem_new_mblock(membp, base, size);
		mem_free_mblock(mblk);
	}

	Xfree(nodelistp);
}

/*
 * Given a particular cpu node in the PRI, lookup the physical ID of
 * the strand. This hides the platform specific usage of the 'id' and
 * 'pid' properties.
 */
uint64_t
pri_cpu_phys_id(md_t *pri, md_node_t *cpu_node)
{
	uint64_t id;
	uint64_t pid;

	/*
	 * First check for the 'pid' property. This is only used on N1
	 * platforms, but if it is present, it is the physical ID of the
	 * strand. This check must be done first since the 'id' prop has
	 * a slightly different meaning on N1.
	 */
	if (md_get_prop_val(pri, cpu_node, "pid", &pid) < 0) {
		/*
		 * All other platforms use the 'id' property as the
		 * physical ID of the strand.
		 */
		if (md_get_prop_val(pri, cpu_node, "id", &id) < 0)
			fatal("PRI CPU node has no id or pid property");

		return (id);
	}

	return (pid);
}

/*
 * For each cpu extract the info and insert into the resource train.
 */

/* constants to calculate max_pagesize from mmu-page-size-list bitmask */
#define	BASE_PAGESIZE 8192
#define	PAGESIZE_SHIFT_MULTIPLIER 8

static void
pri_init_cpus(md_t *mdp)
{
	md_node_t	**nodelistp;
	uint8_t		*compat = NULL;
	int		num;
	int		len;
	int		rv;
	int		i;
	uint64_t	max_pagesize = 0;

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	CHAIN_INIT(pri.cpu_list); 	/* physical strands */

	DBG(zprintf("\tCPUs:"));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "cpu"),
	    md_find_name(mdp, "fwd"), nodelistp);

	if (num == 0)
		fatal("No available CPUs");

	rv = md_get_prop_data(mdp, nodelistp[0], "compatible", &compat, &len);
	if (rv < 0)
		fatal("Missing compatible field in PRI cpu node");

	pri.cpu_compatible = Xstrdup((char *)compat);

	for (i = 0; i < num; i++) {
		md_node_t *nodep;
		pri_cpu_t *cpup;
		uint64_t id, pgsz_list;

		cpup = Xmalloc(sizeof (*cpup));
		cpup->res_id = NULL;
		cpup->core = NULL;
		CHAIN_LINK_INIT(cpup->link);
		CHAIN_LINK_INIT(cpup->mau_link);
		CHAIN_LINK_INIT(cpup->core_link);
		cpup->bound.fraction_free = PRI_CPU_FRAC_DENOM;
		CHAIN_INIT(cpup->bound.vcpus);

		nodep = nodelistp[i];
		if (md_get_prop_val(mdp, nodep, "id", &id) < 0)
			fatal("Missing CPU id in PRI");
		cpup->id = id;
		cpup->phys_id = pri_cpu_phys_id(mdp, nodep);

		DBG(zprintf("%s0x%02"PRIx32"",
		    (i & 7) == 0 ? "\n\t\t" : " ", cpup->phys_id));

		/* add strand to PRI */
		CHAIN_ADD(pri.cpu_list, cpup, link);

		/* Calculate largest available pagesize across all cpus */
		if (md_get_prop_val(mdp, nodep, "mmu-page-size-list",
		    &pgsz_list) < 0)
			fatal("Missing mmu-page-size-list in cpu node of pri");
		max_pagesize |= pgsz_list;
	}

	pri.max_pagesize = BASE_PAGESIZE;

	while ((max_pagesize >>= 1) != 0)
		pri.max_pagesize *= PAGESIZE_SHIFT_MULTIPLIER;

	pri.num_cpus = num;

	Xfree(nodelistp);
	DBG(zprintf("\n"));
}

int
grab_exec_units(md_t *mdp, const char *type_name, md_node_t **resnodesp)
{
	int num, i, j = 0;
	md_node_t **nodelistp;

	/*
	 * First grab all exec-unit nodes
	 */
	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "exec-unit"),
	    md_find_name(mdp, "fwd"), nodelistp);

	if (num == 0)
		warning("No available exec-unit nodes");

	for (i = 0; i < num; i++) {
		md_node_t *nodep;
		uint8_t *type;
		int type_len;

		nodep = nodelistp[i];
		/*
		 * For each node, check if the type property matches type_name
		 */
		if (md_get_prop_data(mdp, nodep, "type", &type, &type_len) != 0)
			fatal("pri's exec-unit has no type property");

			if (type_name == NULL || (strncmp((const char *)type,
			    type_name, strlen(type_name)) == 0)) {
				if (resnodesp != NULL)
					resnodesp[j] = nodep;
				j++;
			}
	}

	Xfree(nodelistp);
	DBG(zprintf("Found %d exec-unit nodes of type %s\n", j,
	    type_name == NULL ? "any" : type_name));

	return (j);
}

void
pri_init_cores(md_t *mdp)
{
	int i, j, numtlbs, numcpus, coreidx;
	md_node_t **tlbnodes, **cpunodes;
	uint8_t *types;
	int types_len;

	/*
	 * First determine mapping of CPUs to cores.  There's no direct
	 * way to determine this, so we make the assumption that there's
	 * a 1-1 mapping between TLBs and cores.
	 */
	tlbnodes = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	numtlbs = md_scan_dag(mdp, NULL, md_find_name(mdp, "tlb"),
	    md_find_name(mdp, "fwd"), tlbnodes);

	cpunodes = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	CHAIN_INIT(pri.core_list); 	/* physical cores */

	coreidx = 0;
	for (i = 0; i < numtlbs; ++i) {
		pri_core_t *core;

		if (md_get_prop_data(mdp, tlbnodes[i], "type", &types,
		    &types_len) != 0)
			fatal("No type property in TLB node");
		/* Let's just look at dTLBs */
		if (strcmp((char *)types, "data") != 0)
			continue;

		core = Xcalloc(1, pri_core_t);
		core->idx = coreidx++;
		CHAIN_ADD(pri.core_list, core, link);

		numcpus = md_scan_dag(mdp, tlbnodes[i],
		    md_find_name(mdp, "cpu"),
		    md_find_name(mdp, "back"), cpunodes);

		create_res_pool(&core->mmu_free_partids, 1, numcpus,
		    Res_sorted);

		for (j = 0; j < numcpus; ++j) {
			uint64_t pid;
			pri_cpu_t *pcpu;

			pid = pri_cpu_phys_id(mdp, cpunodes[j]);

			pcpu = pri_find_pstrand(pid);
			if (pcpu == NULL)
				fatal("No pri_cpu_t for CPU %"PRId64"", pid);

			pcpu->core = core;
			CHAIN_ADD(core->cpus, pcpu, core_link);
		}
	}

	Xfree(cpunodes);
	Xfree(tlbnodes);
}

/*
 * Initialize and populate the list of CPU chips.
 */
void
pri_init_cpu_chips()
{
	pri_cpu_t *pcpu;
	pri_core_t *corep;
	pri_chip_t *chipp;
	int chip_id;
	pm_resid_t res_id;

	res_id.pi_res_type = PM_RES_CPU;
	res_id.pi_res_name = Xstrdup(CPU_RES_NAME);

	CHAIN_INIT(pri.chip_list);	/* physical chips */

	CHAIN_ITER(pri.core_list, corep, link) {
		pcpu = CHAIN_START(corep->cpus);
		res_id.u.cpu.pi_id_val = pcpu->phys_id;
		if (pm_get_container_id(&res_id, &chip_id) != 0)
			continue;

		/* Find the chip in the list of chips */
		CHAIN_ITER(pri.chip_list, chipp, link) {
			if (chipp->idx == chip_id)
				break;
		}

		if (chipp == NULL) {
			chipp = Xcalloc(1, pri_chip_t);
			CHAIN_ADD(pri.chip_list, chipp, link);
			chipp->idx = chip_id;
		}

		CHAIN_ADD(chipp->cores, corep, chip_link);
	}

	Xfree(res_id.pi_res_name);
}

/*
 * Returns the virtual-device node in MD 'mdp' for the device with name
 * 'devname', or NULL if there is no such node.
 */
static md_node_t *
get_virtual_device(md_t *mdp, char *devname)
{
	md_node_t **vdevs;
	int num, i;

	vdevs = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));
	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "virtual-device"),
	    md_find_name(mdp, "fwd"), vdevs);

	for (i = 0; i < num; i++) {
		char *name;
		md_node_t *vdev = vdevs[i];

		if (md_get_prop_str(mdp, vdev, "name", &name) < 0)
			fatal("PRI virtual-device node (%d) has no "
			    "name property", i);

		if (strcmp(name, devname) == 0) {
			Xfree(vdevs);
			return (vdev);
		}
	}

	Xfree(vdevs);

	return (NULL);
}

/*
 * Mark an ino from an array of num_inos inos as in use.  Inos in the array
 * which are in use are negative and free inos are positive. Returns true if
 * the ino is in the array and was free, otherwise returns false.
 */
static bool_t
pull_ino(int *inos, int num_inos, int ino)
{
	int i;

	for (i = 0; i < num_inos; i++) {
		if (inos[i] == ino) {
			inos[i] = -inos[i];
			return (true);
		}
	}

	return (false);
}

/*
 * Allocate and return a free ino from an array of num_inos inos.  Inos in the
 * array which are in use are negative and free inos are positive.
 */
static int
alloc_ino(int *inos, int num_inos)
{
	int i, ino;

	for (i = 0; i < num_inos; i++) {
		if (inos[i] > 0) {
			ino = inos[i];
			inos[i] = -inos[i];
			return (ino);
		}
	}

	return (0);
}

bool_t
pri_pull_mau_ino(int ino)
{
	return (pull_ino(pri.mau_inos, pri.num_mau_inos, ino));
}

bool_t
pri_pull_cwq_ino(int ino)
{
	return (pull_ino(pri.cwq_inos, pri.num_cwq_inos, ino));
}

int
pri_alloc_mau_ino(void)
{
	return (alloc_ino(pri.mau_inos, pri.num_mau_inos));
}

int
pri_alloc_cwq_ino(void)
{
	return (alloc_ino(pri.cwq_inos, pri.num_cwq_inos));
}

static void
set_mau_representation(pri_mau_t *maup)
{
	pri_cpu_t *cpup;
	int len;

	/* space for () plus num of cpus ints separated by ", " */
	len = 2 + (INT_STR_LEN + 2) * CHAIN_LENGTH(maup->cpus);

	maup->repr = Xmalloc(len);
	maup->parse_repr = Xmalloc(len);

	cpup = CHAIN_START(maup->cpus);
	sprintf(maup->parse_repr, "%d", cpup->phys_id);
	sprintf(maup->repr, "(%d", cpup->phys_id);

	for (cpup = CHAIN_NEXT(cpup->mau_link); cpup != NULL;
	    cpup = CHAIN_NEXT(cpup->mau_link)) {
		sprintf(maup->parse_repr, "%s,%d", maup->parse_repr,
		    cpup->phys_id);
		sprintf(maup->repr, "%s, %d", maup->repr, cpup->phys_id);
	}

	sprintf(maup->repr, "%s)", maup->repr);
}

void
pri_init_maus(md_t *mdp)
{
	int nummaus, i, inos_len;
	md_node_t **maunodesp;
	md_node_t *ncp, *cwq;
	uint64_t *inos = NULL;
	md_node_t **stranslistp;

	maunodesp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	/*
	 * Get the maus (crypto units). An LDom can access an
	 * mau only through one of its strands belonging
	 * to the same core as the mau.
	 * To find the maus, get a list of all "exec-unit"
	 * nodes, and the maus are those nodes with a "type"
	 * property named "mau"
	 */
	nummaus = grab_exec_units(mdp, "mau", maunodesp);
	if (nummaus == 0) {
		Xfree(maunodesp);
		warning("No available MAUs");
		return;
	}

	/*
	 * Get the "virtual-device" node with name "ncp".
	 */
	ncp = get_virtual_device(mdp, "ncp");
	if (ncp == NULL)
		fatal("No ncp virtual-device node in the PRI");

	/*
	 * Get and save the array of MAU inos.
	 */
	if (md_get_prop_data(mdp, ncp, "ino", (uint8_t **)&inos,
	    &inos_len) != 0)
		fatal("No ino property in ncp node of PRI");
	pri.num_mau_inos = inos_len/sizeof (uint64_t);
	pri.mau_inos = Xmalloc(pri.num_mau_inos * sizeof (int));
	for (i = 0; i < pri.num_mau_inos; i++)
		pri.mau_inos[i] = (int)inos[i];

	cwq = get_virtual_device(mdp, "n2cp");
	if (cwq != NULL) {
		/*
		 * The host has CWQs. Get and save the array of CWQ inos.
		 */
		if (md_get_prop_data(mdp, cwq, "ino", (uint8_t **)&inos,
		    &inos_len) != 0)
			fatal("No ino property in n2cp node of PRI");
		pri.num_cwq_inos = inos_len/sizeof (uint64_t);
		pri.cwq_inos = Xmalloc(pri.num_cwq_inos * sizeof (int));
		for (i = 0; i < pri.num_cwq_inos; i++)
			pri.cwq_inos[i] = (int)inos[i];

		pri.have_cwq = true;
	}

	/*
	 * Correlate the maus with physical strands.
	 */
	stranslistp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));
	for (i = 0; i < nummaus; i++) {
		md_node_t *nodep;
		pri_mau_t *maup;
		int j, numstrands;

		nodep = maunodesp[i];

		maup = Xcalloc(1, pri_mau_t);
		maup->mau_id = i;
		CHAIN_ADD(pri.mau_list, maup, link);

		/* assign a fake id to track maus */
		CHAIN_INIT(maup->cpus);
		maup->flag_visited = false;

		/*
		 * For each MAU, trace the back arc to figure the strands
		 * sharing this.
		 */
		numstrands = md_scan_dag(mdp,  nodep, md_find_name(mdp, "cpu"),
		    md_find_name(mdp, "back"), stranslistp);

		if (numstrands == 0)
			fatal("No CPUs share this MAU");

		/*
		 * For each strand sharing this MAU, find the appropriate
		 * pri_cpu_t structure and add to the list.
		 */
		for (j = 0; j < numstrands; j++) {
			uint64_t strand_id;
			pri_cpu_t *cpup;

			strand_id = pri_cpu_phys_id(mdp, stranslistp[j]);

			cpup = pri_find_pstrand(strand_id);
			if (cpup == NULL)
				fatal("No matching CPU for strand %"PRId64"\n",
				    strand_id);

			cpup->maup = maup;
			CHAIN_ADD(maup->cpus, cpup, mau_link);
		}

		set_mau_representation(maup);
	}

	pri.num_maus = nummaus;

	Xfree(stranslistp);
	Xfree(maunodesp);
}

/*
 * Add all assignable IO devices to the resource train.  Only the top-level
 * iodevice nodes in the PRI (i.e. direct descendants of the phys_io node) are
 * considered to represent assignable bus-level io devices, i.e. buses
 * which users can add to or remove from domains. Assignable slots/devices
 * are determined via "assignable-path" in the component tree of the PRI.
 */
static void
pri_init_phys_io(md_t *mdp)
{
	int num, i;
	md_node_t *phys_io;
	md_node_t **nodelistp;
	uint64_t rcid;

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	CHAIN_INIT(pri.iodevice_list);	/* io devices */

	DBG(zprintf("\tIO:\n"));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "phys_io"),
	    md_find_name(mdp, "fwd"), nodelistp);
	if (num != 1)
		fatal("The PRI does not contain a singleton phys_io node");
	phys_io = nodelistp[0];

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "iodevice"),
	    md_find_name(mdp, "fwd"), nodelistp);
	if (num == 0)
		fatal("No available physical devices");

	for (i = 0; i < num; i++) {
		md_node_t *nodep, *parent;
		nodep = nodelistp[i];
		pri_iodevice_t *iop;
		char *strp, *devtype_p;
		char *cli_name = NULL;
		uint8_t *compat;
		int len;
		uint64_t cfg_handle, disabled, no_reset;

		if (md_get_prop_arc(mdp, nodep, "back", &parent) < 0)
			fatal("Missing back arc in PRI iodevice node");

		if (parent != phys_io)
			continue;	/* not at top level */

		if (md_get_prop_str(mdp, nodep, "device-type", &devtype_p) < 0)
			fatal("Missing device-type field in PRI iodevice node");

		if (md_get_prop_str(mdp, nodep, "name", &strp) < 0) {
			warning("Missing name field in PRI iodevice node, "
			    "assuming pci");
			strp = "pci";
		}

		if (md_get_prop_data(mdp, nodep, "compatible",
		    &compat, &len) < 0)
			fatal("Missing compatible field in PRI iodevice node");

		if (md_get_prop_str(mdp, nodep, "pseudonym", &cli_name) < 0)
			warning("Missing pseudonym field in PRI iodevice node");

		if (md_get_prop_val(mdp, nodep, "cfg-handle", &cfg_handle) < 0)
			fatal("Missing cfg-handle field in PRI iodevice node");

		/* Get rcid if it exists */
		if (md_get_prop_val(mdp, nodep, "rcid", &rcid) != 0)
			rcid = ILLEGAL_RCID;

		iop = Xcalloc(1, pri_iodevice_t);
		iop->free = true;
		iop->rcid = rcid;
		iop->device_type = Xstrdup(devtype_p);
		CHAIN_INIT(iop->dev_list);
		if (strcmp(devtype_p, "pciex") == 0) {
			iop->is_pciex = true;
			iop->res_id_pool = &zeus.free_pcidev_ids;
		} else {
			/*
			 * FIXME: for now everything which is not of type
			 * "pciex" is assumed to be of type "network_device".
			 */
			iop->res_id_pool = &zeus.free_netdev_ids;
		}
		iop->name = zsprintf("%s@%"PRIx64"", strp, cfg_handle);
		iop->compatible = Xstrdup((char *)compat);
		if (cli_name != NULL)
			iop->cli_name = Xstrdup(cli_name);
		iop->cfg_handle = cfg_handle;

		if (md_get_prop_val(mdp, nodep, "no_pci_reset", &no_reset)
		    == 0 && no_reset == 1)
			iop->no_pcie_reset = true;
		else
			iop->no_pcie_reset = false;

		if (md_get_prop_val(mdp, nodep, "disabled", &disabled) == 0)
			iop->disabled = disabled;

		if (iop->is_pciex) {
			pcie_pri_get_resources(mdp, nodep, iop);
		}
		CHAIN_INIT(iop->bound_ios);
		CHAIN_LINK_INIT(iop->link);
		CHAIN_ADD(pri.iodevice_list, iop, link);

		DBG(zprintf("iodevice: pseudonym = %s, name = %s\n",
		    iop->cli_name != NULL ? iop->cli_name : "unspecified",
		    iop->name));
	}

	Xfree(nodelistp);

	/* Get slot info */
	pcie_pri_get_io_slots(mdp);

	pcie_init_res_heuristics();
	DBG(zprintf("\n"));
}

/*
 * Misc support functions.
 */

pri_cpu_t *
pri_find_pstrand(int pid)
{
	pri_cpu_t *pcpup;

	for (pcpup = CHAIN_START(pri.cpu_list);
	    pcpup != NULL && pcpup->phys_id != pid;
	    pcpup = CHAIN_NEXT(pcpup->link)) /* LINTED */
		;

	return (pcpup);
}

void
pri_delete_pstrand(int pid)
{
	pri_cpu_t *pcpup;

	pcpup = pri_find_pstrand(pid);
	if (pcpup != NULL) {
		CHAIN_UNLINK(pri.cpu_list, pcpup, link);
		Xfree(pcpup);
	}
}

/*
 * Given a cpu pid, find which physical mau
 * is accessible to it
 */
pri_mau_t *
pri_find_pmau_from_cpuid(int pid)
{
	pri_cpu_t *cpup;

	CHAIN_ITER(pri.cpu_list, cpup, link) {
		if (pid == cpup->phys_id)
			return (cpup->maup);
	}
	return (NULL);
}

/*
 * Find boot image in PRI, make sure it matches expected PA
 */
pri_image_t *
pri_get_boot_image(uint64_t basepa)
{
	pri_image_t *imgp;

	CHAIN_ITER(pri.image_list, imgp, link) {
		if (imgp->guest_bootable)
			break;
	}
	if (imgp == NULL || basepa != (imgp->offset + imgp->imgcp->pa_base))
		return (NULL);

	return (imgp);
}
