/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)pri_svc.c	1.11	09/05/20 SMI"

/*
 * Module implementing the 'pri' DS protocol
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <alloca.h>
#include <errno.h>

#include "allocate.h"
#include "backend.h"
#include "basics.h"
#include "ds.h"
#include "fatal.h"
#include "ldom.h"
#include "mdesc.h"
#include "mdesc_impl.h"
#include "pri_svc.h"
#include "option.h"
#include "zpri.h"
#include "response.h"
#include "sequencer.h"
#include "zprintf.h"
#include "mac_addrs.h"

static uint64_t pri_seq = 0;		/* sequence number */


static ds_ver_t pri_svc_vers[] = { {1, 0} };
#define	PRI_SVC_NVERS  (sizeof (pri_svc_vers) / sizeof (pri_svc_vers[0]))

static ds_capability_t pri_svc_cap = {
	PRI_SVC,	/* svc_id */
	pri_svc_vers,	/* vers */
	PRI_SVC_NVERS	/* nvers */
};

static ds_ops_t pri_svc_libops = {
	dslib_register_handler,		/* ds_reg_cb */
	dslib_unregister_handler,	/* ds_unreg_cb */
	dslib_data_handler,		/* ds_data_cb */
	PRI_SVC				/* cb_arg */
};

/*
 * Info needed to invoke ds_clnt_reg/ds_svc_reg when loading plugins
 * There could be multiple services in the same info structure
 */
static libds_info_t pri_svc_info[] = {
	{
		DS_CLIENT,
		&pri_svc_cap,
		&pri_svc_libops
	}
};

static void
mark_pruned_pri_node(md_t *mdp, md_node_t *node, bool_t *is_pruned_pri_node)
{
	md_node_t *parent;
	md_node_t *n;
	int	i;

	for (n = CHAIN_START(mdp->node_list), i = 0; n != NULL;
	    n = CHAIN_NEXT(n->link), i++) {
		if (n == node) {
			is_pruned_pri_node[i] = true;
			break;
		}
	}

	ASSERT(n != NULL);
	if (node  == md_root_node(mdp)) {
		return;
	}

	/* find and mark the parent as well */
	if (md_get_prop_arc(mdp, node, "back", &parent) < 0)
		fatal("Missing parent of pciexrc/niu node");

	mark_pruned_pri_node(mdp, parent, is_pruned_pri_node);

}

static md_t *
gen_pruned_pri(ldom_t *ldp)
{

	md_t		*mdp;
	md_node_t	*node;
	md_node_t	**nodelist;
	bool_t		*is_pruned_pri_node;
	int		num, i;
	bound_iodevice_t *bip;

	mdp = md_clone(pri.pri);

	num = md_node_count(mdp);

	is_pruned_pri_node = Xmalloc(num * sizeof (bool_t));
	for (i = 0; i < num; i++)
		is_pruned_pri_node[i] = false;

	nodelist = Xmalloc(num * sizeof (md_node_t *));

	/*
	 * Find pciexrc complex owned by ldp.
	 */

	num  = md_scan_dag(mdp, NULL, md_find_name(mdp, "component"),
	    md_find_name(mdp, "fwd"), nodelist);

	/*
	 * a prined pri always has a root node regardless whether
	 * a guest domain owns any pciexrc or niu complex.
	 */
	mark_pruned_pri_node(mdp, md_root_node(mdp), is_pruned_pri_node);


	for (i = 0; i < num; i++) {
		uint64_t	cfg_handle;
		char		*path, *cfgpath, *end;
		char		*name;

		node = nodelist[i];
		path = NULL;

		if (md_get_prop_str(mdp, node, "topo-hc-name", &name) != 0)
			continue;

		if (strcmp(name, "pciexrc") != 0 && strcmp(name, "niu") != 0)
			continue;

		if (md_get_prop_str(mdp, node, "path", &path) < 0) {
			fatal("Missing path property in pciexrc or niu node");
		}
		ASSERT(path != NULL && strlen(path) >= 3);

		/*
		 * The "path" property of the pciexrc nodes is of the
		 * form "/@400/@0" (since it's actually a root port node, not
		 * a root complex node), whereas for niu nodes it's of the form
		 * "/@600".
		 * Since we just want the numerical value, we skip the leading
		 * "/@".
		 */
		cfgpath = Xstrdup(&path[2]);

		/* strip any extraneous path components that might exist */
		cfgpath = strtok(cfgpath, "/");

		cfg_handle = strtoll(cfgpath, &end, 16);

		if (end == cfgpath)
			fatal("Malformed path property in pciexrc or niu node");

		Xfree(cfgpath);

		DBG(zprintf("PRI \"path\" property \"%s\" corresponds to "
		    "cfg-handle %"PRIu64"\n", path, cfg_handle));

		/*
		 * if guest domain owns a iodevice and its cfg-handle property
		 * equals a pciexrc or niu complex node, the guest domain
		 * owns that node and all its ancestors.
		 */
		CHAIN_ITER(ldp->bound.iodevices, bip, ldom) {
			if (bip->phys_iodev->cfg_handle == cfg_handle) {
				break;
			}
		}

		if (bip == NULL) {
			/* not bound */
			continue;
		}
		/* mark node and its parents */
		mark_pruned_pri_node(mdp, node, is_pruned_pri_node);
	}

	Xfree(nodelist);

	/* delete nodes are not in pruned pri */
	for (node = CHAIN_START(mdp->node_list), i = 0; node != NULL;
	    node = CHAIN_NEXT(node->link), i++) {
		if (is_pruned_pri_node[i] == true) {
			continue;
		}
		md_mark_node_deleted(node);
	}

	Xfree(is_pruned_pri_node);

	md_scour(mdp, md_find_name(mdp, "back"));

	return (mdp);
}

static int
send_pruned_pri(ds_svc_t *svc, uint64_t req_num)
{
	ds_msg_t msg;
	ds_req_t *ds_req;
	pri_data_t *pri_data;
	int msglen, paylen;
	int md_size;
	uint8_t	*mdbuf;
	int rv;
	md_t		*mdp;

	mdp = gen_pruned_pri(svc->chan->ldp);
	md_size = md_gen_bin(mdp, &mdbuf);

	paylen = sizeof (pri_req_t) + md_size;
	msglen = offsetof(ds_resp_t, payload.data.bytes) + paylen;

	ds_req = Xmalloc(msglen);

	DBG7(memset(ds_req, 0, msglen));

	ds_fill_data_header(ds_req, svc->hdl, paylen);

	/* fill in the request */
	pri_data = (pri_data_t *)DS_DATA_BUF(ds_req);
	pri_data->hdr.type = hton64(PRI_DATA);
	pri_data->hdr.reqnum = hton64(req_num);
	memcpy(pri_data->data, mdbuf, md_size);

	msg.buf = ds_req;
	msg.len = msglen;
	msg.seq = pri_data->hdr.reqnum;

	rv = ds_send_data_msg(svc->chan, &msg, DS_NO_BLOCK, NULL);

	Xfree(mdbuf);
	md_fini(mdp);
	Xfree(ds_req);

	return (rv);
}

static int
request_pri(ds_svc_t *svc)
{
	ds_msg_t msg;
	ds_req_t *ds_req;
	pri_req_t *pri_req;
	int msglen, paylen;
	int rv;

	paylen = sizeof (pri_req_t);
	msglen = offsetof(ds_resp_t, payload.data.bytes) + paylen;

	ds_req = Xmalloc(msglen);

	DBG7(memset(ds_req, 0, msglen));

	ds_fill_data_header(ds_req, svc->hdl, paylen);

	/* fill in the request */
	pri_req = (pri_req_t *)DS_DATA_BUF(ds_req);
	pri_req->hdr.type = hton64(PRI_REQUEST);
	pri_req->hdr.reqnum = hton64(pri_seq++);

	msg.buf = ds_req;
	msg.len = msglen;
	msg.seq = pri_req->hdr.reqnum;

	rv = ds_send_data_msg(svc->chan, &msg, DS_NO_BLOCK, NULL);

	Xfree(ds_req);

	return (rv);
}

extern mutex_t pri_lock;
extern cond_t pri_cond;
extern md_t *initial_pri;

static void
pri_data_handler(ds_svc_t *svc, void *buf, size_t msglen)
{
	pri_hdr_t *hdr;
	md_t *pri;
	int pri_size;

	if (msglen < sizeof (ds_data_handle_t) + sizeof (pri_hdr_t)) {
		DBG(zprintf("received invalid PRI UPDATE message, got %"PRId16
		    " bytes and expected >= %"PRId16"\n", msglen,
		    sizeof (ds_data_handle_t) + sizeof (pri_hdr_t)));
		Xfree(buf);
		return;
	}

	hdr = (pri_hdr_t *)((char *)buf + sizeof (ds_data_handle_t));

	hdr->type = ntoh64(hdr->type);
	hdr->reqnum = ntoh64(hdr->reqnum);

	DBG(zprintf("received %s message, type %"PRIu64", reqnum %"PRIu64"\n",
	    PRI_SVC, hdr->type, hdr->reqnum));

	switch (hdr->type) {
	case PRI_REQUEST:
		if (strcmp(svc->chan->name, SP_CHAN_NAME) == 0) {
			warning("Unknown %s service message type %"PRId16,
			    PRI_SVC, hdr->type);
			break;
		}

		send_pruned_pri(svc, hdr->reqnum);
		break;

	case PRI_UPDATE:
		/* for now just ignore updates */
		break;

	case PRI_DATA: {
		pri_data_t *pri_data = (pri_data_t *)hdr;
		pri_size = msglen - sizeof (ds_data_handle_t) -
		    sizeof (pri_hdr_t);

		dump_to_file(pri_data->data, pri_size, "dump-vbsc-pri.bin");

		pri = md_scan_bin((uint8_t *)pri_data->data, pri_size);
		if (pri == NULL)
			fatal("Invalid PRI from VBSC");

		mutex_lock(&pri_lock);
		/* for now all PRIs except the first are discarded */
		if (initial_pri == NULL) {
			initial_pri = pri;
			cond_signal(&pri_cond);
		} else {
			md_fini(pri);
		}
		mutex_unlock(&pri_lock);
		break;
	}
	default:
		warning("Unknown %s service message type %"PRId16, PRI_SVC,
		    hdr->type);
	}

	Xfree(buf);
}

static void
pri_reg_handler(ds_svc_t *svc)
{
	if (strcmp(svc->chan->name, SP_CHAN_NAME) == 0)
		request_pri(svc);
}

/*
 * The versions supported
 */
static prot_version_t pri_svc_versions[] =
{
	{1, 0}
};

static svc_ops_t pri_svc_ops = {
	pri_reg_handler,		/* reg_handler */
	NULL,				/* unreg_handler */
	pri_data_handler,		/* data_handler */
	NULL,				/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	int i;
	svc_info_t *info;

	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 3;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(SP_CHAN_NAME);
	info->chan_names[1] = Xstrdup(DS_CHAN_NAME);
	info->chan_names[2] = Xstrdup(LIBDS_CHAN_NAME);
	info->num_svcs = 1;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(PRI_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = 0;
	info->num_vers = NUM_VERS(pri_svc_versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	memcpy(info->versions, pri_svc_versions,
	    info->num_vers * sizeof (prot_version_t));

	/*
	 * copy the info structure for each service
	 */
	info->libds_svcs = Xmalloc(info->num_svcs * sizeof (libds_info_t *));
	for (i = 0; i < info->num_svcs; i++) {
		info->libds_svcs[i] = cp_dsinfo(&pri_svc_info[i],
		    info->svcs[i]);
	}

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, PRI_SVC) == 0)
		return (&pri_svc_ops);
	else
		return (NULL);
}
