/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2006 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)restart.c	1.1	06/11/14 SMI"

#include <errno.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "basics.h"
#include "fatal.h"
#include "restart.h"
#include "zprintf.h"

static int socks[2] = { -1, -1 };

/*
 * Initialize parent/child communication sockets. This is called once before
 * forking and only when daemonizing.
 */
void
restart_init(void)
{
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, socks) < 0) {
		zprintf("Failed to create socket pair: %s", strerror(errno));
		exit(EXIT_FATAL);
	}
}

/*
 * Called by child immediately after fork.  Cleans up communication sockets
 * and returns descriptor for the socket which will be used to write the
 * startup status to the parent.
 */
int
restart_child(void)
{
	close(socks[0]);
	socks[0] = -1;

	return (socks[1]);
}

/*
 * Called by parent to wait for startup status notification from child.
 * When not daemonizing this is a NOP.
 */
void
restart_wait(void)
{
	uint8_t status;

	if (socks[0] != -1) {
		close(socks[1]);

		/* read socket for startup status from child */
		if (read(socks[0], &status, sizeof (status)) ==
		    sizeof (status)) {
			exit(status);
		} else {
			/* child aborted during startup, don't restart */
			exit(EXIT_FATAL);
		}
	}
}

/*
 * Write startup 'status' to parent.
 * When not daemonizing this is a NOP.
 */
static void
write_status(uint8_t status)
{
	if (socks[1] != -1) {
		if (write(socks[1], &status, sizeof (status)) !=
		    sizeof (status))
			efatal("writing start status");
		close(socks[1]);
		socks[1] = -1;
	}
}

/*
 * Notify "parent" that startup has successfully concluded.
 * When not daemonizing this is a NOP.
 */
void
restart_fini(void)
{
	write_status(EXIT_OK);
}
