/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <thread.h>
#include <synch.h>

#include "fatal.h"
#include "basics.h"
#include "allocate.h"
#include "clp.h"
#include "cli_conn.h"
#include "errors.h"
#include "keys.h"
#include "variables.h"
#include "mdesc.h"
#include "mdesc_impl.h"
#include "ldom.h"
#include "zpri.h"
#include "backend.h"
#include "ds.h"
#include "ds_impl.h"
#include "zprintf.h"
#include "dr_cpu.h"
#include "dr_vio.h"
#include "dr_crypto.h"
#include "database.h"
#include "hvctl.h"
#include "sequencer.h"
#include "xml_parse.h"
#include "xml_cntrl_common.h"
#include "control_class.h"
#include "restart.h"
#include "mac_addrs.h"
#include "frags.h"
#include "utilization.h"
#include "vcpu.h"
#include "vmau.h"
#include "vio.h"
#include "xmpp_server.h"
#include "xml_event.h"
#include "pm_events.h"
#include "core_md.h"
#include "mig.h"
#include "nvlist.h"
#include "autosave.h"
#include "dr_mem.h"
#include "hvctl.h"

/*
 * address-congruence-offset property provides hints to the
 * guest operating system for hardware page coloring and
 * memory placement optimizations. N1 based platforms have
 * no hardware page coloring and this property would be set
 * to 0. On N2 based platforms, L2 cache nodes have index-mask
 * properties and address-congruence-offset property can be used
 * by guest OS to do hardware page coloring. On VF based
 * platforms, there would be latency group nodes and this
 * property is used for memory placement optimizations in
 * addition to hardware page coloring
 */
#define	ADDR_CONGR_OFF_PROP "address-congruence-offset"

/*
 * Queue of tasks for sequencer to perform.
 */
static struct sequencer sequencer = {
	DEFAULTMUTEX,
	DEFAULTCV,
	DEFAULTCV,
	DEFAULTCHAIN,
	DEFAULTCHAIN,
	1,	/* start at 1 because the processed ID will 0 initially */
	0	/* initially no tasks have been processed, ID=1 will be 1st */
};

/* forward declarations */
static bool_t seq_download_mds(void);
static bool_t seq_config_guests(void);

#if ZDBG

mutex_t	seq_monitor = DEFAULTMUTEX;

#endif

char *
hv_state_str(uint64_t hv_state)
{
	switch (hv_state) {
	case GUEST_STATE_STOPPED:	return ("stopped");
	case GUEST_STATE_RESETTING:	return ("resetting");
	case GUEST_STATE_NORMAL: 	return ("normal");
	case GUEST_STATE_SUSPENDED:	return ("suspended");
	case GUEST_STATE_EXITING:	return ("exiting");
	case GUEST_STATE_UNCONFIGURED:	return ("unconfigured");
	default:			break;
	}

	return ("unknown");
}

static void
commit_delayed_reconfig(ldom_t *ldp)
{
	ldm_err_t		pm_result = { 0 };
	pm_domain_event_info_t	*event;

	/*
	 * Clean up any pending delayed reconfig operation
	 * on this LDom.
	 */
	if (in_delayed_reconfig(ldp) == true) {
		seq_delayed_reconfig_finish(ldp);
		if (ldp->pm_vcpu_count > 0) {
			CHAIN_ITER(pm_domain_list.domain_event_list, event,
			    link) {
				if ((event->event_type == PM_RES_EVENT) ||
				    (event->res_type == PM_RES_CPU)) {
					break;
				}
			}
			pm_domain_config_change_notify(event->handle,
			    (srcid_t)ldp, REM_RESOURCE_POST, PM_RES_CPU,
			    ldp->pm_vcpu_list, ldp->pm_vcpu_count, &pm_result);

			/* Free per resource status data */
			pm_free_errors(&pm_result);

			ldp->pm_vcpu_count = 0;
			Xfree(ldp->pm_vcpu_list);
		}

		zeus.delayed_reconfig_ldom = NULL;
	}
}

/*
 * Handle LDom state change received via asynchronous hvctl notification or
 * polling the LDom state.
 */
void
seq_ldom_state_change(ldom_t *ldp, cli_conn_t *client, uint64_t hv_state)
{
	cli_conn_t	*event_client = NULL;
	ldm_err_t	result = { 0 };	/* Not an array because currently */
					/* only PM_RES_CPU resource type is */
					/* affected by a domain state change. */

	DBG(zprintf("%s: %s, zeus state %s, hv state %s\n", __func__,
	    ldp->namep, ldom_state_str(ldp->state), hv_state_str(hv_state)));

	ASSERT_IN_SEQUENCER();

	if (hv_state != GUEST_STATE_NORMAL && ldp->flag.memdr_in_progress) {
		/* mem dr aborted, clear flag in MD */
		mem_dr_cleanup(ldp);
	}

	/*
	 * Either ldm stop/ldm start or a reboot will trigger
	 * the Hypervisor to commit a delayed reconfig, so
	 * we check in both GUEST_STATE_STOPPED & GUEST_STATE_RESETTING.
	 */
	switch (hv_state) {
	case GUEST_STATE_STOPPED:		/* dead pending restart */
		if (ldp->state != BOUND) {
			ldp->state = BOUND;

			/* Send notification to remove domain from boot list */
			if (pm_domain_dr_reg_event_on) {
				send_pm_dom_event(ldp, GUEST_DOMAIN_DOWN,
				    NULL);
			}

			/* Send notification to power down resources */
			if (pm_domain_events_on) {
				send_pm_dom_event(ldp, STOP_DOMAIN_POST,
				    &result);
				/*
				 * FIXME: check for errors by checking
				 * result.ldme_errs and traversing
				 * result.ldme_private.
				 */
				pm_free_errors(&result);
			}
			/*
			 * If client == NULL this call has come from an
			 * asynchronous HV call. We assume now that there are
			 * never multiple waiters. If this changes we need to
			 * revisit this code.
			 */
			if (client == NULL) {
				waiter_t *w;
				ASSERT(CHAIN_LENGTH(ldp->waiters) <= 1);
				w = CHAIN_START(ldp->waiters);
				if (w != NULL) {
					event_client = w->client;
				}
			}
			release_waiters(ldp, client, BOUND);
			event_send_dom_event(XML_STOP, ldp, event_client);
			cpu_release_reserved(ldp);
			commit_delayed_reconfig(ldp);
		}
		break;

	case GUEST_STATE_RESETTING:
		event_send_dom_event(XML_DOMAIN_RESET, ldp, client);
		/* Send event to notify PM that guest is no longer running. */
		if (pm_domain_dr_reg_event_on) {
			send_pm_dom_event(ldp, GUEST_DOMAIN_DOWN, NULL);
		}

		if (ldp->state != BOUND) {
			commit_delayed_reconfig(ldp);
			cpu_release_reserved(ldp);
		}
		/* FALLTHROUGH */
	case GUEST_STATE_NORMAL:
		if (ldp->state == BOUND || ldp->state == STARTING) {
			/*
			 * The guest is bound or starting. Mark it now as
			 * active and release any clients waiting for it to
			 * become active.
			 */
			ldp->state = ACTIVE;
			release_waiters(ldp, client, ACTIVE);
			event_send_dom_event(XML_START_DOMAIN, ldp, client);
			/*
			 * If opening the domain services channel when binding
			 * the guest failed, try opening it again. If the
			 * channel open fails we take no further action. Later
			 * attempts to use services which may have been
			 * provided over this channel will fail and action may
			 * be taken there.
			 */
			if (ldp->ds_chan->state == CHAN_DOWN)
				(void) ds_open_channel(ldp->ds_chan);
		}
		break;

	case GUEST_STATE_SUSPENDED:		/* suspended */
		/* Send event to notify PM that guest is no longer running. */
		if (pm_domain_dr_reg_event_on) {
			send_pm_dom_event(ldp, GUEST_DOMAIN_DOWN, NULL);
		}

		break;
	case GUEST_STATE_EXITING:
		/* in process of exiting, will soon be stopped */

		break;
	case GUEST_STATE_UNCONFIGURED:		/* not in HV */
		break;
	default:
		break;
	}
}

static void
dec_chan_usecnt(ds_channel_t *chan, bool_t locking_reqd)
{
	if (locking_reqd == true)
		(void) mutex_lock(&chan->lock);
	chan->usecnt--;
	ASSERT(chan->usecnt >= 0);
	if (chan->usecnt == 0)
		(void) cond_broadcast(&chan->cond);
	if (locking_reqd == true)
		(void) mutex_unlock(&chan->lock);
}

/*
 * Send an MD update notification to LDom 'ldp'.
 */
static int
md_update_notify(ldom_t *ldp)
{
	update_md_f update_md;
	int rv = 0;

	ASSERT_IN_SEQUENCER();

	/* make sure the domain is up */
	if (ldp->state <= BOUND)
		return (-1);

	/* make sure the DS channel exists */
	if (ldp->ds_chan == NULL)
		return (-1);

	mutex_lock(&ldp->ds_chan->lock);
	if (ldp->ds_chan->state != CHAN_UP) {
		mutex_unlock(&ldp->ds_chan->lock);
		DBG(zprintf("Unable to notify %s of updated MD, "
		    "channel is down\n", ldp->namep));
		return (-1);
	} else {
		ldp->ds_chan->usecnt++;
		mutex_unlock(&ldp->ds_chan->lock);
	}

	update_md = (update_md_f)ds_service_routine(ldp->ds_chan,
	    MD_UPDATE_SVC, "platsvc_update_md");

	if (update_md != NULL) {
		if (update_md(ldp) == 0) {
			ldp->flag.md_notify = false;
			DBG(zprintf("%s: notified guest %s of new MD\n",
			    __func__, ldp->namep));
		} else {
			rv = -1;
			DBG(zprintf("Unable to notify %s of updated MD\n",
			    ldp->namep));
		}
	} else {
		rv = -1;
		DBG(zprintf("%s service not registered for guest %s\n",
		    MD_UPDATE_SVC, ldp->namep));
	}

	dec_chan_usecnt(ldp->ds_chan, true);
	return (rv);
}

/*
 * Send an MD update notification to each guest which needs one.
 */
void
seq_send_md_update_notifs(void)
{
	ldom_t *ldp;

	if (zeus.mode == ZEUS_CONFIG_MODE)
		return;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if ((ldp->state > BOUND) && ldp->flag.md_notify)
			md_update_notify(ldp);
	}
}

/*
 * Reset the MD update and notify flags in case something fails
 */
void
seq_clear_ldom_flags(void)
{
	ldom_t *ldp;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		ldp->flag.gen_new_md = false;
		ldp->flag.md_notify = false;
	}
}

static void
reconfigure_guest_node(md_t *reconf_md, md_node_t *reconf_node,
    md_t *new_md, md_node_t *new_node)
{
	uint64_t mdpa;

	if (md_get_prop_val(new_md, new_node, "mdpa", &mdpa) != 0)
		fatal("Guest node in new HV MD is missing the mdpa property");

	md_update_value_property(reconf_md, reconf_node, "mdpa", mdpa);
}

/*
 * Reconfigure the 'reconf_md' hypervisor MD to point to guests in the
 * 'new_md' hypervisor MD and to point to the new MD.
 */
static void
md_reconfiguration(md_t *reconf_md, md_t *new_md, uint64_t new_mdpa)
{
	md_node_t **reconf_nodes, **new_nodes;
	uint64_t reconf_gid, new_gid;
	int num_reconf, num_new;
	int i, j;

	reconf_nodes = Xmalloc(md_node_count(reconf_md) * sizeof (md_node_t *));
	new_nodes = Xmalloc(md_node_count(new_md) * sizeof (md_node_t *));

	num_reconf = md_scan_dag(reconf_md, NULL,
	    md_find_name(reconf_md, "guest"), md_find_name(reconf_md, "fwd"),
	    reconf_nodes);

	num_new = md_scan_dag(new_md, NULL,
	    md_find_name(new_md, "guest"), md_find_name(new_md, "fwd"),
	    new_nodes);

	for (i = 0; i < num_reconf; i++) {
		if (md_get_prop_val(reconf_md, reconf_nodes[i], "gid",
		    &reconf_gid) != 0)
			fatal("Guest %d in current MD is missing gid property",
			    i);

		for (j = 0; j < num_new; j++) {
			if (md_get_prop_val(new_md, new_nodes[j], "gid",
			    &new_gid) != 0)
				fatal("Guest %d in new MD is missing gid "
				    "property", j);

			if (reconf_gid == new_gid) {
				reconfigure_guest_node(reconf_md,
				    reconf_nodes[i], new_md, new_nodes[j]);
				break;
			}
		}
	}

	/* add pointer to new HV MD */
	md_add_value_property(reconf_md, md_root_node(reconf_md),
	    "new_mdpa", new_mdpa);

	Xfree(reconf_nodes);
	Xfree(new_nodes);
}

/*
 * Check to see if a delayed reconfiguration operation is in progress.
 *
 * If the 'ldp' argument is NULL, then we check to see if any domain is
 * delayed config mode otherwise if a pointer to a domain is specified
 * we just check to see if that domain is in delayed reconfiguration mode.
 *
 * The function returns 'true' if in delayed reconfiguration mode
 */
bool_t
in_delayed_reconfig(ldom_t *ldp)
{
	if (ldp == NULL) {
		/* check if *any* domain is in delayed reconfig */
		if (zeus.delayed_reconfig_ldom == NULL)
			return (false);
		else
			return (true);
	} else {
		/* check if 'ldp' domain is in delayed reconfig */
		if (zeus.delayed_reconfig_ldom == ldp)
			return (true);
		else
			return (false);
	}
}

static void
get_next_md_frag(ldom_t *ldp, int size, bool_t commit)
{
	if (!in_delayed_reconfig(NULL)) {
		ASSERT(ldp->frags.next_md_frag == NULL);
		ldp->frags.next_md_frag = frag_alloc(ldp, size);
	} else {
		if (ldp->frags.parsed_md_frag != NULL) {
			frag_free(ldp->frags.parsed_md_frag);
		}
		ldp->frags.next_md_frag = frag_alloc(ldp, size);

		/* Need to preserve in case of HV parse failure & rollback */
		ldp->frags.parsed_md_frag = frag_clone(ldp->frags.next_md_frag);

		if (commit) {
			ldp->frags.committed_md_frag =
			    frag_clone(ldp->frags.next_md_frag);
		}
	}
}

/*
 * Take the current HV MD and modify it to point to the current version
 * of each guest's MD. Other than that, keep everything the same.
 *
 * XXX - This is something of a hack because it introduces inconsistencies
 * between the HV MD and the guest MDs in terms of the LDC configuration.
 * This is ok for suspend-resume testing as long as the resume puts things
 * back the way they should be before any real action is performed.
 */
static void
seq_hvmd_reconfig(void)
{
	md_t		*hvmd;
	uint64_t	mdpa;
	ldom_t		*ldp;
	md_node_t	**nodes;
	int		num_nodes;
	int		i;
	uint64_t	gid;

	hvmd = zeus.hv_mdp;

	nodes = Xmalloc(md_node_count(hvmd) * sizeof (md_node_t *));
	num_nodes = md_scan_dag(hvmd, NULL, md_find_name(hvmd, "guest"),
	    md_find_name(hvmd, "fwd"), nodes);

	CHAIN_ITER(zeus.ldoms, ldp, link) {

		if (!ldp->flag.gen_new_md)
			continue;

		/* get the PA for the guest MD */
		mdpa = ldp->frags.next_md_frag->paddr;

		for (i = 0; i < num_nodes; i++) {
			if (md_get_prop_val(hvmd, nodes[i], "gid", &gid) != 0)
				fatal("Guest %d is missing gid property", i);

			if (ldp->idp->id != gid)
				continue;

			/* update the HV MD with the new guest MD PA */
			md_update_value_property(hvmd, nodes[i], "mdpa", mdpa);
			break;
		}
	}

	Xfree(nodes);
}

/*
 * Walk all the LDoms and update the MDs of those requiring new MDs.
 */
static void
seq_update_mds_if_needed(void)
{
	uint8_t	*bufp;
	ldom_t	*ldp;
	md_t	*mdp;
	int	size;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (!ldp->flag.gen_new_md)
			continue;
		if (ldp->state == INACTIVE)
			fatal("%s: Internal error, inactive guest %s flagged "
			    "for MD generation", __func__, ldp->namep);

		DBG(zprintf("\n*** Downloading new guest MD for %s ***\n\n",
		    ldp->namep));

		mdp = seq_gen_guest_md(ldp, false);
		backend.aug_guestmd(ldp, mdp);

		/*
		 * Allocate location of new guest MD. It's double buffered, so
		 * can't clobber the old one yet.
		 */
		size = md_gen_bin(mdp, &bufp);
		if (size <= 0)
			fatal("Unable to generate binary MD for LDom %s "
			    "using internal MD", ldp->namep);
		else if (size > (FRAG_SIZE - pri.md_elbow_room))
			fatal("Configuration required a MD of size 0x%x "
			    "bytes which\nis larger than the maximum supported "
			    "size of 0x%x bytes", size,
			    FRAG_SIZE - pri.md_elbow_room);

		get_next_md_frag(ldp, size, true);

		backend.dl_guestmd(ldp, ldp->frags.next_md_frag->paddr, bufp,
		    size);

		Xfree(bufp);
		md_fini(mdp);
	}
}

static int
seq_reconfigure_guests_migrate(void)
{
	int	size;
	uint8_t	*bufp;
	ldom_t	*ldp;

	if (zeus.mode == ZEUS_CONFIG_MODE)
		return (0);

	DBG2(zprintf("%s\n", __func__));

	/*
	 * Walk all the LDoms and update the MDs of those requiring new MDs.
	 */
	seq_update_mds_if_needed();

	seq_hvmd_reconfig();

	size = md_gen_bin(zeus.hv_mdp, &bufp);
	if (size <= 0) {
		fatal("Unable to generate HV binary MD using internal MD");
	} else if (size > (max_hv_md_size - pri.md_elbow_room)) {
		fatal("Configuration required a MD of size 0x%x"
		    " bytes which\nis larger than the maximum supported"
		    " size of 0x%x bytes",
		    size, max_hv_md_size - pri.md_elbow_room);
	}

	/*
	 * Find a frag for the "next" HV MD. Reuse a previously allocated
	 * one if available.  This assumes the HV isn't going to try to
	 * read/parse the MD we're overwriting before we send the next
	 * HVctl_op_reconfigure().
	 */
	if (zeus.next_hvmd_fp == NULL)
		zeus.next_hvmd_fp = hv_md_alloc(); /* reconfig MD frag */
	ASSERT(zeus.next_hvmd_fp != NULL);

	backend.dl_hvmd(zeus.next_hvmd_fp->paddr, bufp, size, RECONF_HVMD);
	Xfree(bufp);

	backend.config_guests(GUEST_CONFIG);

	seq_send_md_update_notifs();

	/* rotate the new HV MD into the current slot */
	hv_md_free(zeus.current_hvmd_fp);
	zeus.current_hvmd_fp = zeus.next_hvmd_fp;
	zeus.next_hvmd_fp = NULL;

	/*
	 * Cleanup the MD for any domain that was modified.
	 */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (!ldp->flag.gen_new_md)
			continue;

		if (ldp->state == INACTIVE) {
			fatal("%s: Internal error, inactive guest %s flagged "
			    "for MD generation", __func__, ldp->namep);
		}

		/* rotate the next MD into the current MD */
		frag_free(ldp->frags.committed_md_frag);
		ldp->frags.committed_md_frag = ldp->frags.next_md_frag;
		ldp->frags.next_md_frag = NULL;

		ldp->flag.gen_new_md = false;

		free_from_stage(&ldp->guest_free_ldcs);
	}

	return (0);
}

/*
 * Reconfigure guests.
 */
static int
seq_reconfigure_guests(bool_t del_reconf)
{
	md_t *mdp;
	int size;
	uint8_t *bufp;
	bool_t old_next_spconfig;

	if (zeus.mode == ZEUS_CONFIG_MODE)
		return (0);

	DBG2(zprintf("%s\n", __func__));

	/*
	 * Walk all the LDoms and update the MDs of those requiring new MDs.
	 */
	seq_update_mds_if_needed();

	/*
	 * Generate and download the reconfiguration and HV MDs.
	 */

	DBG(zprintf("\n*** Downloading and initiating the reconfiguration and "
	    "new hypervisor MDs ***\n\n"));

	/*
	 * In the case of a delayed reconfig, we only set zeus.is_next_spconfig
	 * false temporarily, because until the delayed reconfig commits, we
	 * still need to track the state of the running config.
	 */
	if (del_reconf)
		old_next_spconfig = zeus.is_next_spconfig;

	/* New HV MD does not match SP's boot set */
	zeus.is_next_spconfig = false;

	mdp = target.gen_hvmd(true);	/* Allocates zeus.next_hvmd_fp */
	backend.aug_hvmd(mdp);

	if (del_reconf) {
		/*
		 * Revert zeus.is_next_spconfig here.  It'll get updated when
		 * the delayed reconfig commits.
		 */
		zeus.is_next_spconfig = old_next_spconfig;
	}

	/* download new hvmd */
	size = md_gen_bin(mdp, &bufp);
	if (size <= 0)
		fatal("Unable to generate HV binary MD using internal MD");
	else if (size > (max_hv_md_size - pri.md_elbow_room))
		fatal("Configuration required a MD of size 0x%x bytes which\n"
		    "is larger than the maximum supported size of"
		    "0x%x bytes", size, max_hv_md_size - pri.md_elbow_room);

	ASSERT(zeus.next_hvmd_fp != NULL);
	backend.dl_hvmd(zeus.next_hvmd_fp->paddr, bufp, size, HVMD);

	Xfree(bufp);

	/*
	 * Take the currently active HV MD (in zeus.hv_mdp) and update it to
	 * point at the new guest MDs, by looking through the new HV MD (mdp),
	 * matching & copying pointers.  We reconfigure the system with this
	 * so-called "reconf" HV MD first.  This allows all the guests to
	 * reconfigure themselves to handle the imminent removal of HV
	 * resources, before we actually remove those resources.
	 *
	 * A second backend.config_guests() call is subsequently invoked with
	 * the final, new HV MD, which causes the actual resources (e.g. vcpus
	 * or LDC channels) to be removed.
	 *
	 * In addition, a special property, "new_mdpa" gets added to the root
	 * of the reconf HV MD.  This allows the LDom Manager to detect and
	 * recover from the case that it dies between the two
	 * backend.config_guest operations.
	 */
	md_reconfiguration(zeus.hv_mdp, mdp, zeus.next_hvmd_fp->paddr);

	size = md_gen_bin(zeus.hv_mdp, &bufp);
	if (size <= 0)
		fatal("Unable to generate HV binary MD using internal MD");
	else if (size > (max_hv_md_size - pri.md_elbow_room))
		fatal("Configuration required a MD of size 0x%x bytes which\n"
		    "is larger than the maximum supported size of 0x%x bytes",
		    size, max_hv_md_size - pri.md_elbow_room);

	/*
	 * Find a frag for the "reconf" HV MD.  Reuse a previously allocated
	 * one if available.  This assumes the HV isn't going to try to
	 * read/parse the MD we're overwriting before we send the next
	 * HVctl_op_reconfigure().
	 */
	if (zeus.reconf_fp == NULL)
		zeus.reconf_fp = hv_md_alloc(); /* reconfig MD frag */
	ASSERT(zeus.reconf_fp != NULL);

	backend.dl_hvmd(zeus.reconf_fp->paddr, bufp, size, RECONF_HVMD);
	Xfree(bufp);

	/* replace current HV MD with new one */
	/*
	 * Since we can call seq_reconfigure_guests more than
	 * once per delayed reconfiguration, the HV MD replacement takes
	 * place in seq_delayed_reconfig_finish().
	 */
	if (!in_delayed_reconfig(NULL)) {
		md_fini(zeus.hv_mdp);
		zeus.hv_mdp = mdp;
	} else {
		/* Save mdp in shadow area until we can commit to zeus.hv_mdp */
		zeus.delayed_hvmd = mdp;
	}

	if (in_delayed_reconfig(NULL)) {
		backend.config_guests(DELAYED_RECONFIG);
	} else {
		backend.config_guests(RECONFIG);
	}

	seq_send_md_update_notifs();

	/*
	 * Now reconfigure with the new HV MD.  For delayed reconfig,
	 * cleanup is accomplished in seq_delayed_reconfig_finish(), which
	 * is triggered by the guest exiting.
	 */
	if (in_delayed_reconfig(NULL)) {
		backend.config_guests(DELAYED_CONFIG);
	} else {
		backend.config_guests(CONFIG);

		hv_md_free(zeus.reconf_fp);
		zeus.reconf_fp = NULL;
		hv_md_free(zeus.current_hvmd_fp);
		zeus.current_hvmd_fp = zeus.next_hvmd_fp;
		zeus.next_hvmd_fp = NULL;
	}

	return (0);
}

/*
 * For each LDom "in progress", look for the ones in the BINDING state, take
 * their binding info, update the hypervisor and their machine descriptions,
 * then generate the new hypervisor description, so that these in progress
 * LDoms can move into the BOUND state.
 *
 * FIXME: Need to handle failures ... either leave in BINDING state, or clean
 * up to the INACTIVE state (freeing binding resources).
 */
void
seq_bind_new(void)
{
	ldom_t	*ldp;

	/*
	 * Download all new and updated MDs and then if successful
	 * configure the guests being bound.
	 */
	zeus.is_next_spconfig = false; /* mark config as different from SP */
	if (seq_download_mds() == false)
		fatal("Failed to download new MDs");
	if (seq_config_guests() == false)
		fatal("Failed to configure guests");

	/*
	 * Only get here when all guests in the progress list were
	 * successfully bound in the HV. Now update their state.
	 */
	CHAIN_ITER(zeus.inprogress, ldp, inprog) {
		if (ldp->state == BINDING) {
			DBG2(zprintf("Changing LDom %s state from BINDING "
			    "to BOUND\n", ldp->namep));
			ldp->state = BOUND;
		} else {
			fatal("%s: Internal error in progress "
			    "list has LDom %s in state (%d) %s", __func__,
			    ldp->namep, (int)ldp->state,
			    ldom_state_str(ldp->state));
		}
	}

	autosave_update_autosave();
}

static int
get_unconfigured_vcpus(ldom_t *ldp, uint32_t **vcpus)
{
	vcpu_t *vcpu;
	int count = 0;

	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		if (!vcpu->flag_isconfigd)
			count++;
	}

	*vcpus = Xmalloc(count * sizeof (*vcpus));
	count = 0;

	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		if (!vcpu->flag_isconfigd) {
			(*vcpus)[count] = vcpu->vid;
			count++;
		}
	}

	return (count);
}

static struct vcpu_ldom_list {
	ldom_t *ldp;
	int nvcpus;
	uint32_t *vcpus;
	int *cpus_index;
	int *pm_status;
};

/* From cpu.c */
int check_cpu_dr_capable(ldom_t *ldp);

void
seq_update_guest_mds(void *arg)
{
	ldom_t *ldp = (ldom_t *)arg;

	/*
	 * Mark the domain for an MD update.
	 */
	ldp->flag.gen_new_md = true;
	ldp->flag.md_notify = true;

	if (seq_download_mds() == false)
		fatal("Failed to download new MDs for PM");

	/* Tell HV & guests about new MDs */
	backend.config_guests(GUEST_CONFIG);
	seq_send_md_update_notifs();

	/*
	 * Do most of the same cleanup that seq_config_guests() performs,
	 * except modifying the configd flags of the vcpus.
	 */
	hv_md_free(zeus.current_hvmd_fp);
	zeus.current_hvmd_fp = zeus.next_hvmd_fp;
	zeus.next_hvmd_fp = NULL;

	/*
	 * Rotate the next MD into the current MD
	 */
	frag_free(ldp->frags.committed_md_frag);
	ldp->frags.committed_md_frag = ldp->frags.next_md_frag;
	ldp->frags.next_md_frag = NULL;
	ldp->flag.gen_new_md = false;

	/* No Autosave */
}

/*
 * Takes specified cpus on or off line.  Utilizes the cpu DR infrastructure to
 * (un)configure strands, but does not actually change the resource's
 * machine configuration (i.e. no reconfiguration occurs at the hypervisor
 * level).
 * FIXME: Using the DR infrastructure to effect reconfiguring of cpu resources
 * for the purposes of power management should be considered a temporary
 * solution, until true support for the P_POWEROFF state is added to Solaris.
 */
int
seq_reconf_pcpus(pm_config_op_t op, int ncpus, uint32_t *cpus, int *status,
    uint32_t flags)
{
	dr_cpu_cfg_f dr_cpu_op;
	int i, j, rv, rc, retval, nldps;
	uint32_t *op_stats;
	struct vcpu_ldom_list *vcpu_info;
	ldom_t *ldp;

	nldps = 0;
	retval = LDME_OP_EOK;

	/* can't perform PM reconf while migration is in progress */
	if ((ldp = mig_inprogress_domain()) != NULL) {
		DBG(zprintf("%s: Not allowed while migration of %s "
		    "in progress\n", __func__, ldp->namep));
		return (LDME_OP_EAGAIN);
	}

	/* Can't perform PM reconf while delayed reconf in progress */
	if (in_delayed_reconfig(NULL) == true) {
		DBG(zprintf("%s: Not allowed while delayed reconfig "
		    "in progress\n", __func__));
		return (LDME_OP_EAGAIN);
	}

	vcpu_info = Xcalloc(CHAIN_LENGTH(zeus.ldoms), struct vcpu_ldom_list);
	op_stats = Xcalloc(ncpus, uint32_t);

	/*
	 * For each physical cpu specified, get the corresponding ldom
	 * & vcpu info. Loop through the list, building up a list of ldoms
	 * & vcpus to reconfigure (either configure or unconfigure).
	 */
	for (i = 0; i < ncpus; ++i) {
		int vid;
		int ldom_nvcpus;

		op_stats[i] = DR_CPU_RES_OK;

		/*
		 * Default status is set to failure; it gets changed
		 * to DR_CPU_RES_OK later if the operation succeeds.
		 */
		if (status != NULL)
			status[i] = DR_CPU_RES_FAILURE;

		if (pid_to_vid(cpus[i], &ldp, &vid) == false) {
			warning("Attempt to reconfigure pcpu %"PRIu32" which "
			    "is not bound to an ldom.\n", cpus[i]);

			if (status != NULL) {
				for (j = i + 1; j < ncpus; j++) {
					status[j] = DR_CPU_RES_FAILURE;
				}
			}

			retval = LDME_OP_EINVAL;
			goto offline_ret;
		}

		/* Find existing ldom index for vcpu_info */
		for (j = 0; j < nldps; ++j) {
			if (ldp == vcpu_info[j].ldp)
				break;
		}

		ASSERT(j < CHAIN_LENGTH(zeus.ldoms));

		/* Add vcpu to list */
		ldom_nvcpus = ++vcpu_info[j].nvcpus;
		vcpu_info[j].vcpus = Xrealloc(vcpu_info[j].vcpus,
		    ldom_nvcpus * sizeof (*vcpu_info[j].vcpus));
		vcpu_info[j].cpus_index = Xrealloc(vcpu_info[j].cpus_index,
		    ldom_nvcpus * sizeof (*vcpu_info[j].cpus_index));
		vcpu_info[j].pm_status = Xrealloc(vcpu_info[j].pm_status,
		    ldom_nvcpus * sizeof (*vcpu_info[j].pm_status));
		vcpu_info[j].vcpus[ldom_nvcpus - 1] = vid;
		/* Save global index for returning status */
		vcpu_info[j].cpus_index[ldom_nvcpus - 1] = i;

		vcpu_info[j].ldp = ldp;

		if (j == nldps)
			nldps++;
	}

	/*
	 * Now walk through the ldoms, issuing either a configure or
	 * unconfigure request to each.
	 */
	for (i = 0; i < nldps; ++i) {
		ldom_t *ldp = vcpu_info[i].ldp;
		bool_t need_another_md_update = false;

		if (ldp->state != ACTIVE && ldp->state != BOUND) {
			warning("Cannot reconfigure CPUs in guest %s "
			    "while it is currently %s.\n", ldp->namep,
			    ldom_state_str(ldp->state));
			retval = LDME_OP_EAGAIN;
			continue;
		}

		/*
		 * For bound domains simply update the core MD status
		 * and generate new MDs.
		 */
		if (ldp->state == BOUND) {
			/* Update core MD */
			DBG(zprintf("%s: handling BOUND domain %s; op %d\n",
			    __func__, ldp->namep, op));
			for (j = 0; j < vcpu_info[i].nvcpus; ++j) {
				int cpus_index = vcpu_info[i].cpus_index[j];

				if (set_core_md_cpu_status(cpus[cpus_index],
				    CORE_MD_PM_STATUS, (op == PCO_ADD_RESOURCE ?
				    PM_NORMAL : PM_MANAGED)) != 0)
					fatal("set_core_md_cpu_status of cpu %d"
					    " failed\n", cpus[cpus_index]);

				status[cpus_index] = DR_CPU_RES_OK;
			}

			/* Update MDs */
			seq_update_guest_mds(ldp);
			continue;
		}

		/*
		 * Verify that CPU DR is available for the domain. Ignore
		 * errors due to the fact that a domain has been migrated.
		 * For the purposes of power management, CPU DR operations
		 * are still safe on a migrated domain.
		 */
		rc = check_cpu_dr_capable(ldp);
		if ((rc != 0) && (rc != ERR_CPU_DR_MIG_DISABLED)) {
			switch (rc) {
			case ERR_CPU_DR_BOUND_MAUS:
			case ERR_CPU_DR_NOTSUP:
			case ERR_MD_UPDATE_NOTSUP:
				/*
				 * Silently fail to avoid flooding the
				 * LDoms log if the operation is not supported
				 */
				break;

			default:
				FATAL_UNEXPECTED_ERROR(rc);
				break;
			}
			retval = LDME_OP_EUNAVAIL;
			continue;
		}

		/*
		 * We put a check here to disallow CPU removal via PM if there
		 * are any MAUs bound on the domain.  This is a temporary fix
		 * until PM fixes the fact that they orphan MAUs by removing
		 * all CPUs associated with an MAU.
		 */
		if ((ldom_num_bound_maus(ldp) > 0) &&
		    (op == PCO_REMOVE_RESOURCE)) {
			retval = LDME_OP_EUNAVAIL;
			continue;
		}
		dr_cpu_op = (dr_cpu_cfg_f)ds_service_routine
		    (ldp->ds_chan, DR_CPU_SVC, (op == PCO_ADD_RESOURCE ?
		    "dr_cpu_configure" : "dr_cpu_unconfigure"));
		if (dr_cpu_op == NULL) {
			warning("Cannot reconfigure CPUs in guest %s, "
			    "service %s not registered", ldp->namep,
			    DR_CPU_SVC);
			retval = LDME_OP_EUNAVAIL;
			continue;
		}

		/*
		 * Update core_md status now so new MD is properly generated.
		 * If there are any failures we revert the core_md status &
		 * send another MD update.
		 */
		for (j = 0; j < vcpu_info[i].nvcpus; ++j) {
			int cpus_index = vcpu_info[i].cpus_index[j];

			vcpu_info[i].pm_status[j] = get_core_md_cpu_status(
			    cpus[cpus_index], CORE_MD_PM_STATUS);

			if (set_core_md_cpu_status(cpus[cpus_index],
			    CORE_MD_PM_STATUS, (op == PCO_ADD_RESOURCE ?
			    PM_NORMAL : PM_MANAGED)) != 0)
				fatal("set_core_md_cpu_status of cpu %d "
				    "failed\n", cpus[cpus_index]);
		}

		/*
		 * Update guest MDs before reconfiguring cpus.  Note that
		 * this is different then typical DR, where the MD update
		 * happens _after_ the unconfigure in the remove case.
		 * The concern is that opens a race condition where the
		 * unconfigure completes but before the MD gets updated,
		 * the domain reboots, leading to a "cpu X failed to start"
		 * panic due to the attempt to start a parked strand.
		 * FIXME: might want to make this change for normal DR as well
		 */
		seq_update_guest_mds(ldp);

		rv = (*dr_cpu_op)(ldp, vcpu_info[i].nvcpus, vcpu_info[i].vcpus,
		    0, op_stats, NULL);

		/*
		 * If there's any error, we need to revert the core_md status &
		 * mark the domain for another MD update.
		 */
		for (j = 0; j < vcpu_info[i].nvcpus; ++j) {
			int cpus_index = vcpu_info[i].cpus_index[j];

			if (rv == CPU_DR_OP_FAILED ||
			    rv == CPU_DR_OP_NO_RESPONSE ||
			    op_stats[j] != DR_CPU_RES_OK) {
				/*
				 * IGNORE_DR_ERRORS is set when
				 * pmi_change_to_performance is called from
				 * pm_stop.  If this isn't done, and the DR
				 * op fails, then there is the potential that
				 * the control domain could hit a watchdog
				 * when it is rebooted.  The watchdog is hit
				 * because HV expects all powered up cpus to
				 * be allocated to a guest MD if they're in the
				 * HV active set.
				 * For this bug, the watchdog is hit because a
				 * boot cpu is chosen which is powered up but
				 * not part of a guest MD, even though it's in
				 * the HV active set.
				 * The situation that causes this is when the
				 * control domain is shutting down.  A race
				 * condition between ldmd and drd can see the
				 * drd service come down first causing the DR
				 * ops to fail.
				 * Since the cpus have already been powered up,
				 * and guest MD downloaded with those cpus, the
				 * recovery step is to ignore DR errors instead
				 * of downloading the original guest MD, as the
				 * original guest MD doesn't contain the newly
				 * powered up cpus.  We don't want to simply
				 * power off the cpus as the ldmd shutdown
				 * needs all cpus powered up and added to the
				 * guest MD.
				 * As drd normally only comes down on a reboot
				 * then this ignore step is safe as the cpus
				 * are DR'd back into the domain after the
				 * reboot.
				 */
				if (flags & IGNORE_DR_ERRORS) {
					/* Set status to ignore DR failure */
					status[cpus_index] = DR_CPU_RES_OK;
					DBG(zprintf("%s:DR failure - ignoring"
					    " status of strand %d\n",
					    __func__, cpus[cpus_index]));
					continue;
				}
				if (set_core_md_cpu_status(cpus[cpus_index],
				    CORE_MD_PM_STATUS,
				    vcpu_info[i].pm_status[j]) != 0)
					fatal("set_core_md_cpu_status "
					    "of cpu %d failed\n",
					    cpus[cpus_index]);
				DBG(zprintf("%s:DR failure - backing out PM "
				    "status of strand %d to %d\n",
				    __func__, cpus[cpus_index],
				    vcpu_info[i].pm_status[j]));
				need_another_md_update = true;
			} else {
				if (status != NULL)
					status[cpus_index] = DR_CPU_RES_OK;
			}
		}

		if (need_another_md_update)
			seq_update_guest_mds(ldp);

		/* Update return value.  We return last error found */
		switch (rv) {
		case CPU_DR_OP_FAILED:
			warning("Unable to reconfigure CPUs in guest %s",
			    ldp->namep);
			retval = LDME_OP_EUNSUPPORTED;
			break;
		case CPU_DR_OP_NO_RESPONSE:
			warning("No response to reconfigure CPUs in guest %s",
			    ldp->namep);
			retval = LDME_OP_ETIMEDOUT;
			break;
		case CPU_DR_OP_PARTIAL:
			retval = LDME_OP_EAGAIN;
			break;
		case CPU_DR_OP_SUCC:
			break;
		default:
			FATAL_UNEXPECTED_ERROR(rv);
			/*NOTREACHED*/
		}
	}

offline_ret:
	/* Free all allocated structures */
	for (i = 0; i < nldps; ++i) {
		Xfree(vcpu_info[i].vcpus);
		Xfree(vcpu_info[i].cpus_index);
		Xfree(vcpu_info[i].pm_status);
	}
	Xfree(vcpu_info);
	Xfree(op_stats);

	return (retval);
}

/*
 * Get called when an LDom is active (or bound) and extra
 * cpus have been added. Here we simply re-configure the
 * guest and hypervisor and then prod the guests domain service
 * to bring the extra cpus online. Returns the number of cpus
 * actually added (when the guest is in bound state) in case
 * of full or partial success, otherwise CPU_DR_OP_FAILED or
 * CPU_DR_OP_NO_RESPONSE.
 */
int
seq_add_vcpus(ldom_t *ldp, response_t *resp)
{
	int nvcpu;
	uint32_t *vcpup;
	int rv;
	dr_cpu_cfg_f dr_cpu_configure;
	int ret;
	uint32_t *op_stats;

	if (zeus.mode == ZEUS_CONFIG_MODE)
		return (0);

	/*
	 * Get an array of the unconfigured cpus.  This has to be done before
	 * seq_config_guests because the configd flag gets set in that
	 * function.
	 */
	rv = nvcpu = get_unconfigured_vcpus(ldp, &vcpup);

	dr_cpu_configure = (dr_cpu_cfg_f)ds_service_routine(ldp->ds_chan,
	    DR_CPU_SVC, "dr_cpu_configure");

	if (dr_cpu_configure == NULL) {
		warning("%s service not registered for guest %s\n",
		    DR_CPU_SVC, ldp->namep);
		cleanup_partial_vcpu_add(ldp, nvcpu, vcpup);
		rv = CPU_DR_OP_FAILED;
		goto done;
	}

	ldp->flag.gen_new_md = true;

	zeus.is_next_spconfig = false; /* mark config as different from SP */
	if (seq_download_mds() == false)
		fatal("Failed to download new MDs");

	if (seq_config_guests() == false) {
		warning("Failed to reconfigure LDom %s with new cpus",
		    ldp->namep);
		rv = CPU_DR_OP_FAILED;
		goto done;
	}

	md_update_notify(ldp);

	op_stats = Xcalloc(nvcpu, uint64_t);
	ret = (*dr_cpu_configure)(ldp, nvcpu, vcpup, false, op_stats, resp);
	if (ret == CPU_DR_OP_FAILED) {
		/*
		 * Rollback
		 */
		cleanup_partial_vcpu_add(ldp, nvcpu, vcpup);
		seq_rm_vcpus_cleanup(ldp);
		rv = CPU_DR_OP_FAILED;
	} else if (ret == CPU_DR_OP_PARTIAL) {
		/*
		 * Count the number of failed cases and allocate
		 * an array to hold them
		 */
		int i, n_fail;
		uint32_t *failed_vcpus;

		failed_vcpus = Xcalloc(nvcpu, uint32_t);

		/*
		 * Select the vcpus to rollback
		 */
		n_fail = 0;
		for (i = 0; i < nvcpu; i++)
			if (op_stats[i] != DR_CPU_RES_OK)
				failed_vcpus[n_fail++] = vcpup[i];

		if (n_fail == 0)
			warning("seq_add_vcpus: illegal DS response\n");

		cleanup_partial_vcpu_add(ldp, n_fail, failed_vcpus);
		seq_rm_vcpus_cleanup(ldp);
		Xfree(failed_vcpus);
		rv = nvcpu - n_fail;

		if (rv == 0)
			rv = CPU_DR_OP_FAILED;
	} else if (ret == CPU_DR_OP_NO_RESPONSE) {
		/* No response; not safe to rollback */
		rv = CPU_DR_OP_NO_RESPONSE;
	}

	Xfree(op_stats);

	if (rv > 0)
		autosave_update_autosave();

done:
	Xfree(vcpup);
	DBG(zprintf("seq_add_vcpus returned %d\n", rv));

	return (rv);
}

/*
 * Get called when an LDom is active (or bound) and CPUs are removed. We
 * simply prod the guest's domain service to unconfigure the CPUs and
 * reconfigure the hypervisor.  Returns the number of CPUs successfully
 * removed or an error < 0 if no CPUs were removed.
 */
int
seq_rm_vcpus(ldom_t *ldp, vcpu_t **vcpus, int num,
    bool_t force, uint32_t *op_stats, response_t *resp)
{
	uint32_t	*vcpuids;
	int		idx;
	dr_cpu_cfg_f 	dr_cpu_unconfigure;
	int			i, rv, r = 0;
	int		n_succ;
	pm_resid_t		*resids = Xmalloc(sizeof (pm_resid_t) * num);
	char			*res_name = "cpu";
	ldm_err_t		pm_result = { 0 };
	pm_domain_event_info_t	*event;

	if (zeus.mode == ZEUS_CONFIG_MODE) {
		Xfree(resids);
		return (0);
	}

	dr_cpu_unconfigure = (dr_cpu_cfg_f)ds_service_routine(ldp->ds_chan,
	    DR_CPU_SVC, "dr_cpu_unconfigure");
	if (dr_cpu_unconfigure == NULL) {
		warning("Unable to unconfigure CPUs out of guest %s, "
		    "service %s not registered", ldp->namep, DR_CPU_SVC);
		Xfree(resids);
		return (CPU_DR_OP_FAILED);
	}

	vcpuids = Xcalloc(num, uint64_t);

	for (idx = 0; idx < num; idx++) {
		vcpuids[idx] = vcpus[idx]->vid;
		resids[idx].pi_res_type = PM_RES_CPU;
		resids[idx].pi_res_name = res_name;
		resids[idx].u.cpu.pi_id_val = vcpus[idx]->cpu->phys_id;
	}

	rv = (*dr_cpu_unconfigure)(ldp, num, vcpuids, force, op_stats, resp);

	Xfree(vcpuids);

	if (rv == CPU_DR_OP_FAILED) {
		warning("Unable to unconfigure CPUs out of guest %s",
		    ldp->namep);
		Xfree(resids);
		return (CPU_DR_OP_FAILED);
	} else if (rv == CPU_DR_OP_NO_RESPONSE) {
		warning("No response to unconfigure CPUs out of guest %s",
		    ldp->namep);
		Xfree(resids);
		return (CPU_DR_OP_NO_RESPONSE);
	} else if (rv == CPU_DR_OP_SUCC) {
		if (pm_cpu_events_on) {
			CHAIN_ITER(pm_domain_list.domain_event_list, event,
			    link) {
				if ((event->event_type == PM_RES_EVENT) ||
				    (event->res_type == PM_RES_CPU)) {
					break;
				}
			}
			pm_domain_config_change_notify(event->handle,
			    (srcid_t)ldp, REM_RESOURCE_POST, PM_RES_CPU,
			    resids, num, &pm_result);

			/* Free per resource status data */
			pm_free_errors(&pm_result);
		}
		Xfree(resids);
		return (num);
	}

	/*
	 * We had a partial success with unconfiguring the listed cpus.
	 * Count the cpus for which the operation was successful.
	 */
	for (n_succ = 0, i = 0; i < num; i++) {
		if (op_stats[i] == DR_CPU_RES_OK) {
			n_succ++;
			if (pm_cpu_events_on) {
				resids[r].pi_res_type = PM_RES_CPU;
				resids[r].pi_res_name = res_name;
				resids[r].u.cpu.pi_id_val =
				    vcpus[i]->cpu->phys_id;
				r++;
			}
		}
	}

	if ((pm_cpu_events_on) && (n_succ > 0)) {
		CHAIN_ITER(pm_domain_list.domain_event_list, event, link) {
			if ((event->event_type == PM_RES_EVENT) ||
			    (event->res_type == PM_RES_CPU)) {
				break;
			}
		}
		pm_domain_config_change_notify(event->handle, (srcid_t)ldp,
		    REM_RESOURCE_POST, PM_RES_CPU, resids, n_succ, &pm_result);

		/* Free per resource status data */
		pm_free_errors(&pm_result);
	}
	Xfree(resids);

	/*
	 * autosave_update_autosave cannot be called from here
	 * because the MD's are updated later in cpu.c:dr_rm_cpus.
	 */

	return (n_succ);
}

/*
 * Cleanup after the CPUs have been unbound.
 */
void
seq_rm_vcpus_cleanup(ldom_t *ldp)
{
	if (zeus.mode == ZEUS_CONFIG_MODE)
		return;

	ldp->flag.gen_new_md = true;
	ldp->flag.md_notify = true;

	seq_reconfigure_guests(false);

	ldp->flag.gen_new_md = false;
	frag_free(ldp->frags.committed_md_frag);
	ldp->frags.committed_md_frag = ldp->frags.next_md_frag;
	ldp->frags.next_md_frag = NULL;
}
/*
 * Coalesce memory blks in a ldp
 * Before zeus sync memory with a domain, zeus coalesce memory in order
 * to match with domain's memory table.
 * Within a mblock, memory can only grow from one end and one segment at one
 * timed due to HV limitation. This may be a temperary limitation.
 * As a result, we need to do multiple coalesce and update MD.
 * return true if no further coalesce can be done.
 * return false if further coalesce can be done.
 */
bool_t
seq_coalesce_mem(ldom_t *ldp)
{
	mblock_t *mblk, *next_mblk;
	bool_t rtn = true;

	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		mblk->flag_isconfigd = true;
	}

	/*
	 * Can't use CHAIN_ITER, as unbind could delete entries
	 * off the ldp->bound list.
	 */
	CHAIN_ITER_VOL(ldp->bound.memory, mblk, next_mblk, ldom) {
		mblock_t *mblk1;
		CHAIN_ITER(ldp->bound.memory, mblk1, ldom) {
			if (mblk1 == mblk)
				continue;
			if (mblk->raddr + mblk->size == mblk1->raddr &&
			    mblk->paddr + mblk->size == mblk1->paddr) {
				if (mblk1->flag_isconfigd == false) {
					/* mblk1 modified already */
					rtn = false;
				} else {
					mblk1->size += mblk->size;
					mblk1->raddr = mblk->raddr;
					mblk1->paddr = mblk->paddr;
					mblk->size = 0;
					mem_unbind(mblk);
					mblk1->flag_isconfigd = false;
				}
				break;
			}
		}
	}
	return (rtn);
}

static int
seq_scrub_verify_mem(ldom_t *ldp)
{
	mblock_t *mblk;

	for (int i = 0; i < ldp->mem_dr_state.scrub.nmblks; i++) {
		mblk = &((mblock_t *)
		    ldp->mem_dr_state.scrub.unbind_mblks)[i];

		if (scrub_mem(mblk->paddr, mblk->size) != mblk->size) {
			return (ERR_MEM_DR_SCRUB_FAILED);
		}

		if (backend.mem_dr_verify(ldp, mblk->paddr,
		    mblk->raddr, mblk->size) != 0) {
			Xfree(ldp->mem_dr_state.scrub.unbind_mblks);
			ldp->mem_dr_state.scrub.nmblks = 0;
			ldp->mem_dr_state.scrub.max_nmblks = 0;

			return (ERR_MEM_DR_VERIFY_FAILED);
		}
	}
	if (ldp->mem_dr_state.scrub.nmblks > 0) {
		Xfree(ldp->mem_dr_state.scrub.unbind_mblks);
		ldp->mem_dr_state.scrub.nmblks = 0;
		ldp->mem_dr_state.scrub.max_nmblks = 0;
	}

	return (MEM_DR_OP_SUCC);
}

/*
 * Cleanup after memory has been unbound
 * This is called from both memory del and memory add. In memory add case,
 * memory is allocated before sending request to the domain. If memory add
 * failed or partial failed, we need to remove allocated memory and cleanup.
 * In memory delete case, after memory unbound, we need to cleanup and update
 * MD.
 */
void
seq_rm_mem_cleanup(void *arg)
{
	ldom_t *ldp = (ldom_t *)arg;

	ldp->flag.gen_new_md = true;
	ldp->flag.md_notify = true;

	/* reset mem dr flag in MD */
	ldp->mem_dr_state.inprogress = false;

	ASSERT(ldp->frags.next_md_frag == NULL);
	seq_reconfigure_guests(false);

	if (ldp->mem_dr_state.is_mem_add == false)
		seq_scrub_verify_mem(ldp);

	ldp->flag.gen_new_md = false;
	frag_free(ldp->frags.committed_md_frag);
	ldp->frags.committed_md_frag = ldp->frags.next_md_frag;
	ldp->frags.next_md_frag = NULL;
}

void
seq_dr_mem_md(void *arg)
{
	ldom_t *ldp = (ldom_t *)arg;

	ldp->flag.gen_new_md = true;

	/* set mem dr flag in MD */
	ldp->mem_dr_state.inprogress = true;

	ASSERT(ldp->frags.next_md_frag == NULL);

	if (seq_download_mds() == false)
		fatal("Failed to download new MDs");

	if (seq_config_guests() == false) {
		ldp->mem_dr_state.status = ERR_MEM_DR_OP_FAILED;
		return;
	}

	ASSERT(ldp->frags.next_md_frag == NULL);
	md_update_notify(ldp);
	ldp->mem_dr_state.status = MEM_DR_OP_SUCC;
}

/*
 * get mem blks that are allocated but have not been configured.
 */
int
seq_get_unconfigured_mblks(ldom_t *ldp, void **mblkp)
{
	int count = 0;
	mblock_t *mblk;
	dr_mblk_t **mblkpp = (dr_mblk_t **)mblkp;

	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		if (!mblk->flag_isconfigd) {
			if (mblk->size > DR_MEM_MAXSIZE) {
				count += mblk->size / DR_MEM_MAXSIZE;
				if (mblk->size % DR_MEM_MAXSIZE > 0)
					count++;
			} else {
				count++;
			}
		}
	}

	if (count == 0) {
		return (0);
	}

	*mblkpp = Xcalloc(count, dr_mblk_t);
	count = 0;

	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		if (!mblk->flag_isconfigd) {
			uint64_t size = mblk->size;
			uint64_t raddr = mblk->raddr;

			while (size >= DR_MEM_MAXSIZE) {
				(*mblkpp)[count].addr = raddr;
				(*mblkpp)[count++].size = DR_MEM_MAXSIZE;
				raddr += DR_MEM_MAXSIZE;
				size  -= DR_MEM_MAXSIZE;
			}
			if (size > 0) {
				(*mblkpp)[count].addr = raddr;
				(*mblkpp)[count++].size = size;
			}

			/* clear memory */
			if (scrub_mem(mblk->paddr, mblk->size) != mblk->size) {
				return (-1);
			}
		}
	}

	return (count);
}

/*
 * Handle the result of memory configure from domain
 * Since we need to update the free memory pool and guest md, this routine
 * must be done in the sequencer.
 */
void
seq_chk_add_mem(void *arg)
{
	seq_mem_dr_arg_t *argp = (seq_mem_dr_arg_t *)arg;
	ldom_t *ldp = argp->ldp;

	switch (argp->rv) {
	case MEM_DR_OP_SUCC:
		/* reset mem dr flag in MD */
		ldp->mem_dr_state.inprogress = false;
		if (ldp->state == ACTIVE) {
			ldp->required.size_mem = ldom_size_bound_mem(ldp);
			seq_update_guest_mds(ldp);
		}
		break;

	case ERR_MEM_DR_OP_NO_RESPONSE:
		zprintf("got no response at add %s\n", ldp->namep);
		break;

	case ERR_MEM_DR_OP_FAILED:
	case ERR_MEM_DR_OP_PARTIAL:
	default:
		if (ldp->state == ACTIVE) {
			mem_unbind_unconfigured(ldp, argp->nmblk, argp->mblkp);
			seq_rm_mem_cleanup(ldp);
		}
		break;
	}
}

extern void print_mem(ldom_t *ldp);
/*
 * Handle the result of memory unconfigure from domain
 * Since we need to update the free memory pool and guest md, this routine
 * must be done in the sequencer.
 */
void
seq_chk_del_mem(void *arg)
{
	seq_mem_dr_arg_t *argp = (seq_mem_dr_arg_t *)arg;
	bool_t	    do_unbind = false;
	int	    i;
	uint64_t    partial_size = 0;
	ldom_t *ldp = argp->ldp;

	switch (argp->rv) {
	case MEM_DR_OP_SUCC:
		do_unbind = true;
		if (argp->nmblk == 0)
			fatal(" Memory delete succeeded but nmblk = 0\n");
		break;

	case ERR_MEM_DR_OP_CANCELLED:
	case ERR_MEM_DR_OP_PARTIAL:

		if (argp->nmblk == 0) {
			ASSERT(argp->rv == ERR_MEM_DR_OP_CANCELLED);
			ldp->required.size_mem += argp->size;
			break;
		}

		for (i = 0; i < argp->nmblk; i++)
			partial_size += argp->mblkp[i].size;

		ldp->required.size_mem += (argp->size - partial_size);
		do_unbind = true;
		break;

	case ERR_MEM_DR_OP_NO_RESPONSE:
		zprintf("got no response at rem %s\n", ldp->namep);
		break;

	case ERR_MEM_DR_OP_FAILED:
	case ERR_MEM_DR_OP_DOMAIN_FAILED:
		ldp->required.size_mem += argp->size;
		break;


	default:
		zprintf("Unknown response from domain %s rv %d\n",
		    ldp->namep, argp->rv);
		argp->rv = ERR_MEM_DR_OP_NO_RESPONSE;
		break;
	}

	SANITY(CHAIN_INIT(zeus.inprogress));
	SANITY(CHAIN_LINK_INIT(ldp->inprog));

	if (ldp->state != ACTIVE)
		return;
	if (do_unbind) {
		if (argp->nmblk > 0)
			mem_alloc_scrub_mblks(argp->ldp, argp->nmblk);
		for (i = 0; i < argp->nmblk; i++) {
			unbind_span(ldp, argp->mblkp[i].addr,
			    argp->mblkp[i].size, true);
		}
		seq_rm_mem_cleanup(ldp);
	} else if (argp->rv != ERR_MEM_DR_OP_NO_RESPONSE) {
		/* reset mem dr flag in MD */
		ldp->mem_dr_state.inprogress = false;
		seq_update_guest_mds(ldp);
	}
}

/*
 * Sync memory (raddr, size) with domain.
 * For a given raddr, size from domain, this routine handles the following case:
 * 1. If mblk->raddr < raddr and mblk->raddr + size > raddr, the mblk is split
 * into 2 mblks so that one of the mblk->raddr == raddr. call sync_mblk again.
 * 2. if mblk->raddr == raddr
 *	if (mblk->size == size) done
 *	if (mblk->size > size ) the mblk is split into two mblk one of them
 *	    mblk->size == size  done
 *	if (mblk->size < size) recursive sync_mblk with raddr + mblk->size,
 *	    size - mblk->size
 */
static void
sync_mblk(ldom_t *ldp, uint64_t raddr, uint64_t size)
{
	mblock_t *mblk;
	mblock_t *new_mblk;


	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		uint64_t hv_addr = mblk->raddr;
		uint64_t hv_size = mblk->size;
		pri_mblock_t *pmp = mblk->pmblk;

		if (hv_addr >= raddr + size)
			continue;

		if (hv_addr + hv_size  <= raddr)
			continue;

		DBG(zprintf("zeus:mblk found raddr=%"PRIx64"paddr=%"PRIx64
		    " size=%"PRId64" (MB)  domain raddr=%"PRIx64" size=%"
		    PRId64" (MB)\n", hv_addr, hv_addr, BYTES2MB(hv_size), raddr,
		    BYTES2MB(size)));

		if (hv_addr < raddr) {
			/*
			 * hv_addr < raddr && hv_addr + hv_size > raddr
			 * split mblk into 2 parts and call sync_mblk again
			 */
			mblk->size = raddr - hv_addr;

			new_mblk = mem_new_mblock(pmp,
			    mblk->paddr + raddr - hv_addr,
			    hv_size - (raddr - hv_addr));

			new_mblk->raddr = raddr;
			new_mblk->mblk_resid = alloc_res(&zeus.free_mblk_ids);
			if (new_mblk->mblk_resid == NULL)
				fatal("Ran out of mblock slots!");
			new_mblk->flag_isconfigd = false;
			new_mblk->ldp = ldp;
			CHAIN_ADD(ldp->bound.memory, new_mblk, ldom);
			CHAIN_ADD(pmp->bound.memory, new_mblk, pblk);
			sync_mblk(ldp, raddr, size);
			break;
		}

		/* hv_addr == paddr match found */
		mblk->flag_isconfigd = true;
		if (hv_size == size) {
			/* domain mblk match with hv mblk */
			break;
		}
		if (hv_size > size) {
			/* split this mblk into 2 */
			mblk->size = size;
			new_mblk = mem_new_mblock(pmp,
			    mblk->paddr + size, hv_size - size);
			new_mblk->raddr = raddr +  size;
			new_mblk->mblk_resid = alloc_res(&zeus.free_mblk_ids);
			if (new_mblk->mblk_resid == NULL)
				fatal("Ran out of mblock slots!");
			new_mblk->flag_isconfigd = false;
			new_mblk->ldp = ldp;
			CHAIN_ADD(ldp->bound.memory, new_mblk, ldom);
			CHAIN_ADD(pmp->bound.memory, new_mblk, pblk);
			break;
		}

		/* hv_size < size, continue */
		sync_mblk(ldp, raddr + hv_size, size - hv_size);
		break;
	}
	if (mblk != NULL)
		return;
	zprintf("%s: does not have %llx %llx\n", ldp->namep, raddr, size);
	fatal("Domain has more memory than HV !\n");
}

/*
 *  Last part of sync with domain memory map. This part must be executed
 *  inside sequencer. See comments do_sync_mem() and mem_sync()_with_domain
 *  for an overview of how memory sync works.
 */
void
seq_sync_mem(void *args)
{

	seq_mem_dr_arg_t *argp = (seq_mem_dr_arg_t *)args;
	int rv;
	int i;
	mblock_t *mblk, *mblk_next;
	int unbound_nmblk;
	uint_t nmblk;
	dr_mblk_t *domain_mblk;
	ldom_t *ldp;
	bool_t done;

	nmblk = argp->nmblk;
	domain_mblk = argp->mblkp;
	ldp = argp->ldp;
	rv = argp->rv;

	switch (rv) {
	case ERR_MEM_DR_OP_NO_RESPONSE:
	case ERR_MEM_DR_OP_FAILED:
	default:
		/*
		 * domain is in a unknown state
		 * we keep the lock user needs to manually restart the domain,
		 * to clear the lock.
		 */
		zprintf("domain %s does not respond to mem list requests",
		    ldp->namep);
		return;

	case MEM_DR_OP_SUCC:
		break;
	}

	unbound_nmblk = 0;

	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		mblk->flag_isconfigd = false;
	}

	for (i = 0; i < nmblk; i++) {
		sync_mblk(ldp, domain_mblk[i].addr, domain_mblk[i].size);

	}

	Xfree(domain_mblk);

	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		if (mblk->flag_isconfigd == false) {
			unbound_nmblk++;
		}
	}

	if (unbound_nmblk > 0)
		mem_alloc_scrub_mblks(ldp, unbound_nmblk);

	/*
	 * can't use CHAIN_ITER, as unbind_span() could delete entries
	 * off the ldp->bound list.
	 */
	CHAIN_ITER_VOL(ldp->bound.memory, mblk, mblk_next, ldom) {

		if (mblk->flag_isconfigd == false) {
			unbind_span(ldp, mblk->raddr, mblk->size, true);
		}
	}

	if (unbound_nmblk > 0) {
		seq_rm_mem_cleanup(ldp);
	}


	ldp->mem_dr_state.inprogress = false;

	do {
		done = seq_coalesce_mem(ldp);
		/*
		 * Since HV only allow us to grow one end of mblk not
		 * both ends, we need to update MD multiple times.
		 */
		seq_update_guest_mds(ldp);
	} while (done == false);

	ldp->flag.memdr_in_progress = false;
	ldp->flag.locked = false;
	ldp->required.size_mem = ldom_size_bound_mem(ldp);
}

/*
 * Gets called when an LDom is active and an I/O device has been added.
 * It sends a domain services message to the domain which attempts the
 * operation.
 *
 * Returns Values:
 *	(0)	=> Success (DR_OP_SUCC)
 *	(>0)	=> Failure, request failed on Solaris side
 *	(<0)	=> Failure, request failed in Zeus (ERR_DR_VIO_yyy)
 */
int
seq_vio_dr_add(ldom_t *ldp, char *type, uint64_t dev_id, response_t *resp)
{
	int rv;
	dr_vio_cfg_f dr_vio_configure;

	ASSERT(zeus.mode == ZEUS_MGMT_MODE);

	dr_vio_configure = (dr_vio_cfg_f)ds_service_routine(ldp->ds_chan,
	    DR_VIO_SVC, "dr_vio_configure");

	if (dr_vio_configure == NULL) {
		warning("%s service not registered for guest %s\n",
		    DR_VIO_SVC, ldp->namep);
		return (ERR_DR_VIO_NOTSUP);
	}

	if ((dev_id == INVALID_DEVID) || (dev_id > IDENT_MAX)) {
		warning("Attempt to reconfigure VIO device with invalid "
		    " device ID in guest %s\n", ldp->namep);
		return (ERR_DR_VIO_INVALID_ARG);
	}

	if (type == NULL) {
		warning("Attempt to reconfigure VIO device with invalid device"
		    " type in guest %s\n", ldp->namep);
		return (ERR_DR_VIO_INVALID_ARG);
	}

	rv = (*dr_vio_configure)(ldp, dev_id, type, false, resp);
	DBG(zprintf("%s returned %d\n", __func__, rv));

	if (rv == 0)
		autosave_update_autosave();

	return (rv);
}

/*
 * Gets called when an LDom is active and an I/O device is to be removed.
 * It sends a domain services message to the domain which attempts the
 * operation.
 *
 * Returns Values:
 *	(0)	=> Success (DR_OP_SUCC)
 *	(>0)	=> Failure, request failed on Solaris side
 *	(<0)	=> Failure, request failed in Zeus (ERR_DR_VIO_yyy)
 */
int
seq_vio_dr_rm(ldom_t *ldp, char *type, ident_val_t dev_id, response_t *resp)
{
	int rv;
	dr_vio_cfg_f dr_vio_unconfigure;

	ASSERT(zeus.mode == ZEUS_MGMT_MODE);

	dr_vio_unconfigure = (dr_vio_cfg_f)ds_service_routine(ldp->ds_chan,
	    DR_VIO_SVC, "dr_vio_unconfigure");

	if (dr_vio_unconfigure == NULL) {
		warning("%s service not registered for guest %s\n",
		    DR_VIO_SVC, ldp->namep);
		return (ERR_DR_VIO_NOTSUP);
	}

	if ((dev_id == INVALID_DEVID) || (dev_id > IDENT_MAX)) {
		warning("Attempt to reconfigure VIO device with invalid "
		    " device ID in guest %s\n", ldp->namep);
		return (ERR_DR_VIO_INVALID_ARG);
	}

	if (type == NULL) {
		warning("Attempt to reconfigure VIO device with invalid device"
		    " type in guest %s\n", ldp->namep);
		return (ERR_DR_VIO_INVALID_ARG);
	}

	rv = (*dr_vio_unconfigure)(ldp, dev_id, type, false, resp);
	DBG(zprintf("%s returned %d\n", __func__, rv));

	/*
	 * autosave_update_autosave cannot be called from here
	 * because the MD's are updated later in vio.c:vio_dr_rm_phase2
	 */

	return (rv);
}

/*
 * Find all the unconfigured crypto units associated with a particular
 * LDom. Allocates and populates an array of vmau pointers which must be
 * deallocated by the caller. The number of elements in the array is
 * returned. If there are no unconfigured crypto units in the LDom, zero is
 * returned and the pointer to the array is set to NULL.
 */
static int
get_unconfigured_cryptos(ldom_t *ldp, vmau_t ***vmaus)
{
	vmau_t	*vmau;
	int	count = 0;

	/*
	 * count the number of unconfigured crypto units
	 */
	CHAIN_ITER(ldp->bound.vmaus, vmau, ldom) {
		if (!vmau->flag_isconfigd) {
			count++;
		}
	}

	/* check if all units are already configured */
	if (count == 0) {
		*vmaus = NULL;
		return (count);
	}

	*vmaus = Xmalloc(count * sizeof (**vmaus));
	count = 0;

	/*
	 * Walk the list of crypto units again and save off
	 * pointers to the ones that are unconfigured.
	 */
	CHAIN_ITER(ldp->bound.vmaus, vmau, ldom) {
		if (!vmau->flag_isconfigd) {
			(*vmaus)[count] = vmau;
			count++;
		}
	}

	return (count);
}

/*
 * Returns a function pointer for the specified crypto DR operation.
 * Since a single configure or unconfigure call handles both MAU and
 * CWQ devices, it does not matter which of the services are actually
 * registered.
 */
static dr_crypto_cfg_f
lookup_dr_crypto_routine(ldom_t *ldp, char *routine)
{
	dr_crypto_cfg_f dr_crypto_cfg;

	/* try the lookup using the MAU service first */
	dr_crypto_cfg = (dr_crypto_cfg_f)ds_service_routine(ldp->ds_chan,
	    DR_CRYPTO_MAU_SVC, routine);

	if (dr_crypto_cfg != NULL)
		return (dr_crypto_cfg);

	/* no MAU service, try the lookup using the CWQ service */
	dr_crypto_cfg = (dr_crypto_cfg_f)ds_service_routine(ldp->ds_chan,
	    DR_CRYPTO_CWQ_SVC, routine);

	return (dr_crypto_cfg);
}

/*
 * This function gets called when an LDom is active (or bound) and new
 * crypto units are added. The domain and hypervisor are reconfigured
 * and then the crypto DR domain service is used to notify the guest
 * that the new resources are available. On full or partial success,
 * the number of crypto units successfully added is returned.
 */
int
seq_add_cryptos(ldom_t *ldp, response_t *resp)
{
	int		idx;
	int		rv;
	int		ncrypto;
	vmau_t		**cryptop;
	dr_crypto_cfg_f	dr_crypto_configure;
	int		ret;
	uint32_t	*op_stats;
	struct vcpu	*vcpu;
	uint32_t	*vcpuids;

	if (zeus.mode == ZEUS_CONFIG_MODE)
		return (0);

	/*
	 * Get an array of the unconfigured cryptos. This has to be done
	 * before configuring the domain because the config'd flag gets
	 * set as a side effect of that operation.
	 */
	rv = ncrypto = get_unconfigured_cryptos(ldp, &cryptop);

	/* lookup the function to perform the configure operation */
	dr_crypto_configure = lookup_dr_crypto_routine(ldp,
	    "dr_crypto_configure");

	if (dr_crypto_configure == NULL) {
		warning("Crypto DR service not registered for domain %s\n",
		    ldp->namep);
		cleanup_partial_crypto_add(ldp, ncrypto, cryptop);
		rv = CRYPTO_DR_OP_FAILED;
		goto done;
	}

	ldp->flag.gen_new_md = true;

	zeus.is_next_spconfig = false; /* mark config as different from SP */
	if (seq_download_mds() == false)
		fatal("Failed to download new MDs");

	if (seq_config_guests() == false) {
		warning("Failed to reconfigure LDom %s with new crypto "
		    "units", ldp->namep);
		rv = CRYPTO_DR_OP_FAILED;
		goto done;
	}

	md_update_notify(ldp);

	op_stats = Xcalloc(ncrypto, uint32_t);
	vcpuids = Xmalloc(ncrypto * sizeof (uint32_t));

	/*
	 * The domain can only identify a particular crypto unit by
	 * the vcpus that can access it. For each specified crypto
	 * unit, find a corresponding vcpu to identify it to the
	 * domain.
	 */
	for (idx = 0; idx < ncrypto; idx++) {
		/*
		 * We need to pick a vcpu whose strand is not parked
		 * by PM. We've already verified there is at least
		 * one such strand for each crypto unit on this list.
		 */
		CHAIN_ITER(cryptop[idx]->vcpus, vcpu, mau) {
			if (get_core_md_cpu_status(vcpu->cpu->phys_id,
			    CORE_MD_PM_STATUS) == PM_NORMAL)
				break;
		}
		ASSERT(vcpu != NULL);

		vcpuids[idx] = vcpu->vid;
	}

	/* contact the domain to complete the configure operation */
	ret = (*dr_crypto_configure)(ldp, ncrypto, vcpuids, false,
	    op_stats, resp);

	Xfree(vcpuids);

	if (ret == CRYPTO_DR_OP_FAILED) {
		/* backout the configure operation */
		cleanup_partial_crypto_add(ldp, ncrypto, cryptop);
		seq_rm_cryptos_cleanup(ldp);
		rv = CRYPTO_DR_OP_FAILED;

	} else if (ret == CRYPTO_DR_OP_PARTIAL) {
		int		i;
		int		n_fail = 0;
		vmau_t		**failed_cryptos;

		failed_cryptos = Xmalloc(ncrypto * sizeof (vmau_t *));

		/* select the crypto units to rollback */
		for (i = 0; i < ncrypto; i++) {
			if (op_stats[i] != DR_CRYPTO_RES_OK)
				failed_cryptos[n_fail++] = cryptop[i];
		}

		if (n_fail == 0) {
			warning("Invalid response to add request from the "
			    "domain\n");
		}

		cleanup_partial_crypto_add(ldp, n_fail, failed_cryptos);
		seq_rm_cryptos_cleanup(ldp);
		Xfree(failed_cryptos);
		rv = ncrypto - n_fail;

		/*
		 * If not even one crypto unit could be configured,
		 * mark the entire operation a failure so that fact
		 * will propagate back to the user.
		 */
		if (rv == 0)
			rv = CRYPTO_DR_OP_FAILED;

	} else if (ret == CRYPTO_DR_OP_NO_RESPONSE) {
		/* no response so it is not safe to rollback */
		rv = CRYPTO_DR_OP_NO_RESPONSE;
	}

	Xfree(op_stats);

	if (rv > 0)
		autosave_update_autosave();

done:
	Zfree(cryptop);
	DBG(zprintf("%s returned %d\n", __func__, rv));

	return (rv);
}

/*
 * This function gets called when an LDom is active (or bound) and
 * crypto units are removed. First the crypto DR domain service is
 * used to instruct the guest to unconfigure the crypto units. This
 * is followed by reconfiguring the hypervisor. On full or partial
 * success, the number of crypto units successfully removed is
 * returned.
 */
int
seq_rm_cryptos(ldom_t *ldp, struct vmau **vmaus, int num, uint32_t *op_stats,
    response_t *resp)
{
	int		idx;
	int		rv;
	struct vcpu	*vcpu;
	uint32_t	*vcpuids;
	dr_crypto_cfg_f	dr_crypto_unconfig;
	int		n_succ;

	if (zeus.mode == ZEUS_CONFIG_MODE)
		return (0);

	/* lookup the function to perform the unconfigure operation */
	dr_crypto_unconfig = lookup_dr_crypto_routine(ldp,
	    "dr_crypto_unconfigure");

	if (dr_crypto_unconfig == NULL) {
		warning("Unable to unconfigure crypto units out of domain "
		    "%s, crypto DR service is not registered", ldp->namep);
		return (CRYPTO_DR_OP_FAILED);
	}

	vcpuids = Xmalloc(num * sizeof (uint32_t));

	/*
	 * The domain can only identify a particular crypto unit by
	 * the vcpus that can access it. For each specified crypto
	 * unit, find a corresponding vcpu to identify it to the
	 * domain.
	 */
	for (idx = 0; idx < num; idx++) {
		/*
		 * We need to pick a vcpu whose strand is not parked
		 * by PM. We've already verified there is at least
		 * one such strand for each crypto unit on this list.
		 */
		CHAIN_ITER(vmaus[idx]->vcpus, vcpu, mau) {
			if (get_core_md_cpu_status(vcpu->cpu->phys_id,
			    CORE_MD_PM_STATUS) == PM_NORMAL)
				break;
		}
		ASSERT(vcpu != NULL);

		vcpuids[idx] = vcpu->vid;
	}

	/* contact the domain to initiate the unconfigure operation */
	rv =  (*dr_crypto_unconfig)(ldp, num, vcpuids, false, op_stats, resp);

	Xfree(vcpuids);

	if (rv == CRYPTO_DR_OP_FAILED) {
		warning("Unable to unconfigure crypto units out of domain %s",
		    ldp->namep);
		return (CRYPTO_DR_OP_FAILED);

	} else if (rv == CRYPTO_DR_OP_NO_RESPONSE) {
		warning("No response to request to remove crypto units out "
		    "of domain %s", ldp->namep);
		return (CRYPTO_DR_OP_NO_RESPONSE);

	} else if (rv == CRYPTO_DR_OP_SUCC) {
		return (num);
	}

	/*
	 * Some of the requested crypto units were unconfigured
	 * successfully. Count the crypto units for which the
	 * operation was successful.
	 */
	for (n_succ = 0, idx = 0; idx < num; idx++)
		if (op_stats[idx] == DR_CRYPTO_RES_OK)
			n_succ++;

	/*
	 * autosave_update_autosave cannot be called from here
	 * because the MD's are updated later in mau.c:dr_rm_cryptos
	 */

	return (n_succ);
}

/*
 * Just ask the guest to configure the list of crypto units back which it
 * had successfully removed. For partial or complete success, return the
 * number of crypto units successfully added, in all other cases return a
 * failure code.
 */
int
seq_reconfig_cryptos(ldom_t *ldp, vmau_t **cryptop, int num, response_t *resp)
{
	int		idx;
	int		rv;
	int		ncrypto;
	dr_crypto_cfg_f	dr_crypto_configure;
	int		ret;
	uint32_t	*op_stats;
	uint32_t	*vcpuids;

	rv = ncrypto = num;

	/* lookup the function to perform the configure operation */
	dr_crypto_configure = lookup_dr_crypto_routine(ldp,
	    "dr_crypto_configure");

	if (dr_crypto_configure == NULL) {
		warning("Crypto DR service not registered for domain %s\n",
		    ldp->namep);
		cleanup_partial_crypto_add(ldp, ncrypto, cryptop);
		rv = CRYPTO_DR_OP_FAILED;
		return (rv);
	}

	op_stats = Xcalloc(ncrypto, uint32_t);
	vcpuids = Xmalloc(ncrypto * sizeof (uint32_t));

	/*
	 * The domain can only identify a particular crypto unit by
	 * the vcpus that can access it. For each specified crypto
	 * unit, find a corresponding vcpu to identify it to the
	 * domain.
	 */
	for (idx = 0; idx < ncrypto; idx++) {
		/*
		 * Pick the first strand (vcpu) from each core.
		 */
		vcpuids[idx] = CHAIN_START(cryptop[idx]->vcpus)->vid;
	}

	/* contact the domain to complete the configure operation */
	ret = (*dr_crypto_configure)(ldp, ncrypto, vcpuids, false,
	    op_stats, resp);

	Xfree(vcpuids);

	if (ret == CRYPTO_DR_OP_SUCC) {
		/*
		 * If the cryptos recently removed were added back
		 * successfully. No need to download a new MD and notify the
		 * guests
		 */
		goto done;
	} else if (ret == CRYPTO_DR_OP_FAILED) {
		/*
		 * Failed to add back the cryptos recently removed by guest.
		 * The caller would need to unbind all the crypto units from
		 * the guest and then download new MDs.
		 */
		warning("Unable to configure crypto units back in to the "
		    "domain %s", ldp->namep);
		rv = CRYPTO_DR_OP_FAILED;
		goto done;
	} else if (rv == CRYPTO_DR_OP_NO_RESPONSE) {
		/*
		 * No response received, the crypto units must be left
		 * configured in the guest
		 */
		warning("No response to request to add crypto units back "
		    "in to the domain %s", ldp->namep);
		rv = CRYPTO_DR_OP_NO_RESPONSE;
		goto done;
	} else if (ret == CRYPTO_DR_OP_PARTIAL) {
		int		i;
		int		n_fail = 0;
		vmau_t		**failed_cryptos;

		failed_cryptos = Xmalloc(ncrypto * sizeof (vmau_t *));

		/* select the crypto units to rollback */
		for (i = 0; i < ncrypto; i++) {
			if (op_stats[i] != DR_CRYPTO_RES_OK)
				failed_cryptos[n_fail++] = cryptop[i];
		}

		/*
		 * There must be at least one crypto unit that failed to
		 * unconfigure
		 */
		if (n_fail == 0) {
			warning("Invalid response to add request from the "
			    "domain\n");
		}

		/*
		 * For partial success, we just unbind the units that could
		 * not be added successfully. The caller would need to also
		 * unbind the crypto units that were previously removed before
		 * downloading the MDs.
		 */
		cleanup_partial_crypto_add(ldp, n_fail, failed_cryptos);
		Xfree(failed_cryptos);
		rv = ncrypto - n_fail;

		/*
		 * If not even one crypto unit could be configured,
		 * mark the entire operation a failure so that fact
		 * will propagate back to the user.
		 */
		if (rv == 0)
			rv = CRYPTO_DR_OP_FAILED;

	}

done:
	Xfree(op_stats);

	DBG(zprintf("%s returned %d\n", __func__, rv));

	return (rv);
}

void
seq_rm_cryptos_cleanup(ldom_t *ldp)
{
	if (zeus.mode == ZEUS_CONFIG_MODE)
		return;

	ldp->flag.gen_new_md = true;
	ldp->flag.md_notify = true;

	seq_reconfigure_guests(false);

	ldp->flag.gen_new_md = false;
	frag_free(ldp->frags.committed_md_frag);
	ldp->frags.committed_md_frag = ldp->frags.next_md_frag;
	ldp->frags.next_md_frag = NULL;
}

/*
 * Download HV MD solely for purposes of persisting information for Zeus.  This
 * routine must have no other side effects!
 */
void
seq_update_hvmd(void)
{
	ASSERT_IN_SEQUENCER();

	if (zeus.mode == ZEUS_CONFIG_MODE)
		return;

	if (seq_download_mds() == false)
		fatal("%s failed to download new MDs", __func__);

	/*
	 * Specify guest only if available to prevent unnecessary parsing of HV
	 * MD.
	 */
	if (in_delayed_reconfig(NULL)) {
		DBG(zprintf("%s called with delayed reconfig in progress\n",
		    __func__));
		backend.config_guests(DELAYED_CONFIG);
	} else {
		if (hvctl_state.minor_ver > 0) {
			backend.config_guests(GUEST_CONFIG);
		} else {
			backend.config_guests(CONFIG);
		}

		hv_md_free(zeus.current_hvmd_fp);
		zeus.current_hvmd_fp = zeus.next_hvmd_fp;
		zeus.next_hvmd_fp = NULL;
	}
}

/*
 * Generic routine to download the updated MDs, send the reconfig request
 * to the HV, and then tell the guests about the MD updates.
 */
void
seq_download_mds_do_config(ldom_t *ldp)
{
	if (zeus.mode == ZEUS_CONFIG_MODE)
		return;

	ldp->flag.gen_new_md = true;
	ldp->flag.md_notify = true;

	zeus.is_next_spconfig = false; /* mark config as different from SP */
	if (seq_download_mds() == false)
		fatal("Failed to download new MDs for VIO DR");

	if (seq_config_guests() == false) {
		warning("Failed to reconfigure with new MDs for VIO DR");
		return;
	}

	/* Notify all affected guests' DS clients of MD update */
	seq_send_md_update_notifs();

	autosave_update_autosave();
}

/*
 * Handle any delayed reconfigure sequencing.
 */
void
seq_delayed_reconfig(ldom_t *ldp, del_reconfig_type_t type)
{
	/*
	 * Need to generate new guest MD, but not trigger any MD updates
	 * until the new configuration is committed.
	 */
	ldp->flag.gen_new_md = true;

	/*
	 * Delayed config type DEL_RECONFIG_REMOVE_2PHASE handled via similar
	 * two step process used for removing CPUs or unbinding LDoms.  This
	 * is needed to give the guests a chance to gracefully close any LDC
	 * channels that are about to be unconfigured by the Hypervisor.
	 */
	if (type == DEL_RECONFIG_REMOVE_2PHASE) {
		/*
		 * If HVctl_op_guest_reconf() not available, use old broken
		 * methodology of cancelling and rebuilding the delayed reconfig
		 */
		if (hvctl_state.minor_ver < 1) {
			/* Cancel any pending delayed reconfig */
			backend.config_guests(CANCEL_DELAYED_CONFIG);
		}

		if (seq_reconfigure_guests(true) != 0)
			fatal("Failed to download new MDs for delayed "
			    "reconfig");
	} else {
		bool_t old_next_spconfig;

		/*
		 * Everything else is handled as a simple delayed reconfig
		 * request.
		 */

		/*
		 * We only set zeus.is_next_spconfig false temporarily, because
		 * until the delayed reconfig commits, we still need to track
		 * the state of the running config.
		 */
		old_next_spconfig = zeus.is_next_spconfig;

		/* New HV MD does not match SP's boot set */
		zeus.is_next_spconfig = false;

		if (seq_download_mds() == false)
			fatal("Failed to download new MDs for delayed config");

		/*
		 * Revert zeus.is_next_spconfig here.  It'll get updated when
		 * the delayed reconfig commits.
		 */
		zeus.is_next_spconfig = old_next_spconfig;

		backend.config_guests(DELAYED_CONFIG);
	}

	autosave_update_autosave();
}

/*
 * Free all staged out ldcs from all dependent ldoms having ldc connection to
 * this ldom. When a partial bind fails, the ldc channels which were allocated
 * need to be freed. The partial unbind puts all the ldc channels which need
 * to be freed in the staging pool, these staged out ldc resources must be
 * de-allocated here to complete the process.
 */
void
free_all_staged_ldcs(ldom_t *ldom)
{
	ldom_t *ldp;
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state >= UNBINDING || ldp == ldom)
			free_from_stage(&ldp->guest_free_ldcs);
	}
}

void
seq_delayed_reconfig_finish(ldom_t *ldom)
{
	ldom_t *ldp;

	/*
	 * We have two things we might need to do before we finish up:
	 * 1) Do the cleanup from the completion of the two-step
	 *    reconfiguration for DEL_RECONFIG_REMOVE_2PHASE
	 * 2) Issue an MD update to notify guests of the just committed
	 *    HVctl_op_guest_delayed_reconf
	 */

	if (zeus.reconf_fp) {
		/*
		 * Now discard reconfiguration HV MD (& frag space) & save
		 * new HV MD.
		 */
		md_fini(zeus.hv_mdp);
		zeus.hv_mdp = zeus.delayed_hvmd;
		hv_md_free(zeus.reconf_fp);
		zeus.reconf_fp = NULL;
	}

	/* In case there are outstanding MD update requests to be sent */
	seq_send_md_update_notifs();

	/* Rotate in new HVMD */
	hv_md_free(zeus.current_hvmd_fp);
	zeus.current_hvmd_fp = zeus.next_hvmd_fp;
	zeus.next_hvmd_fp = NULL;

	/* Free rollback frag */
	if (zeus.prev_del_reconf_hvmd_fp) {
		hv_md_free(zeus.prev_del_reconf_hvmd_fp);
		zeus.prev_del_reconf_hvmd_fp = NULL;
	}

	/*
	 * Rotate in new guest MDs
	 */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (!ldp->flag.gen_new_md && ldp != ldom)
			continue;

		if (ldp->state == INACTIVE)
			fatal("%s: Internal error, inactive guest %s flagged "
			    "for MD generation", __func__, ldp->namep);

		frag_free(ldp->frags.committed_md_frag);
		ldp->frags.committed_md_frag = ldp->frags.next_md_frag;
		ldp->frags.next_md_frag = NULL;

		if (ldp->frags.parsed_md_frag != NULL) {
			frag_free(ldp->frags.parsed_md_frag);
			ldp->frags.parsed_md_frag = NULL;
		}

		ldp->flag.gen_new_md = false;

		free_from_stage(&ldp->guest_free_ldcs);
	}

	/* cleanup */

	/* Mark the LDom as having its new config committed */
	ldom->flag.config_committed = true;
	zeus.delayed_hvmd = NULL;
	zeus.delayed_reconfig_type = DEL_RECONFIG_UNKNOWN;

	/* Mark current config as not matching the SP's selected bootset */
	zeus.is_next_spconfig = false;

	autosave_finish_delayed_reconfig();
}

/*
 * Initiate a delayed reconfig operation.
 */
void
initiate_or_update_delayed_reconfig(ldom_t *ldp, del_reconfig_type_t type,
    response_t *resp)
{
	/* Must pass in a valid type */
	ASSERT(type != DEL_RECONFIG_UNKNOWN);
	ASSERT(IS_CONTROL_DOMAIN(ldp));

	if (in_delayed_reconfig(NULL) == false) {
		zeus.delayed_reconfig_ldom = ldp;
		zeus.delayed_reconfig_type = 0; /* Real value gets set below */

		/* Save current autosave state in case of cancel */
		autosave_initiate_delayed_reconfig();

		rprintf(resp,
"Initiating delayed reconfigure operation on LDom %s.  All configuration\n"
"changes for other LDoms are disabled until the LDom reboots, at which time\n"
"the new configuration for LDom %s will also take effect.\n",
		    ldp->namep, ldp->namep);
	} else {
		ASSERT(zeus.delayed_reconfig_type != DEL_RECONFIG_UNKNOWN);
	}


	if (hvctl_state.minor_ver < 1 && type == DEL_RECONFIG_REMOVE_PHYS)
		rprintf(resp,
"\nNotice: this remove operation will prevent any future VIO device\n"
"removal operation from being accepted for the duration of this delayed\n"
"reconfiguration, i.e. until the domain reboots or the delayed reconfig is\n"
"cancelled.\n");

	zeus.delayed_reconfig_type |= type;
	seq_delayed_reconfig(ldp, type);
}

/*
 * Some operations don't initiate delayed reconfig operations, but must
 * be handled as such if there's one already in progress.  This function
 * handles these cases.
 */
void
seq_config_now_or_later(ldom_t *ldp, del_reconfig_type_t type, response_t *resp)
{
	/* No delayed reconfig in progress */
	if (ldp->state == BOUND || !in_delayed_reconfig(NULL)) {
		seq_download_mds_do_config(ldp);
		free_all_staged_ldcs(ldp);
	} else {
		initiate_or_update_delayed_reconfig(ldp, type, resp);
	}
}

void
cancel_delayed_reconfig(ldom_t *ldp, cli_conn_t *client, response_t *resp)
{
	/* Send cancel request to Hypervisor */
	backend.config_guests(CANCEL_DELAYED_CONFIG);

	/* Restore autosave to state before delayed reconfig */
	autosave_cancel_delayed_reconfig();

	/*
	 * For now, we abort, relying on SMF to restart us.  This is the best
	 * way to get zeus back in sync with the state of the machine.
	 */
	warning("Exiting to re-sync after cancelling a delayed reconfig "
	    "operation");
	if (client->class == LDM_CLASS_CLI) {
		resp->ret = LDOM_OK;
		client_write_response(client);
	}

	exit(EXIT_RESTART);
	/* NOTREACHED */

	/* Free MDs associated with delayed reconfig */
	if (ldp->frags.next_md_frag != NULL) {
		frag_free(ldp->frags.next_md_frag);
		ldp->frags.next_md_frag = NULL;
	}
	if (ldp->frags.parsed_md_frag != NULL) {
		frag_free(ldp->frags.parsed_md_frag);
		ldp->frags.parsed_md_frag = NULL;
	}

	free_all_staged_ldcs(ldp);
}

void
seq_rm_reconfigure(ldom_t *ldom)
{
	ldom_t *ldp;

	ldom->flag.gen_new_md = true;

	if (seq_reconfigure_guests(false) != 0)
		fatal("seq_rm_reconfigure: failed to download new MDs");

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (!ldp->flag.gen_new_md)
			continue;

		if (ldp->state == INACTIVE)
			fatal("%s: Internal error, inactive guest %s flagged "
			    "for MD generation", __func__, ldp->namep);

		frag_free(ldp->frags.committed_md_frag);
		ldp->frags.committed_md_frag = ldp->frags.next_md_frag;
		ldp->frags.next_md_frag = NULL;
		ldp->flag.gen_new_md = false;
	}
}

/*
 * Update a guest's variable store.
 */
void
seq_update_vars(ldom_t *ldp, bool_t notify_guest)
{
	ldom_t	*mig_ldp;

	ASSERT_IN_SEQUENCER();

	if (zeus.mode == ZEUS_CONFIG_MODE)
		return;

	/* check if a delayed reconfig is in progress */
	if (ldp->state != BOUND && in_delayed_reconfig(NULL)) {
		initiate_or_update_delayed_reconfig(ldp,
		    DEL_RECONFIG_OTHER, NULL);
		return;
	}

	ASSERT(ldp->state >= BOUND);

	ldp->flag.gen_new_md = true;

	/*
	 * If a domain is in the process of being migrated
	 * off the machine, care must be taken in generating
	 * the guest MD and HV MD. They must not disturb the
	 * the MDs crafted during the migration.
	 */
	mig_ldp = mig_inprogress_domain();
	if (MIG_INPROGRESS_SRC(mig_ldp)) {
		ldp->flag.md_notify = notify_guest;
		(void) seq_reconfigure_guests_migrate();
		return;
	}

	/*
	 * If a domain is in the process of being migrated
	 * onto the machine, the control domain must not be
	 * sent an MD update notification until the migration
	 * operation is ready.
	 */
	if (MIG_INPROGRESS_TGT(mig_ldp) && IS_CONTROL_DOMAIN(ldp))
		notify_guest = false;

	if (seq_download_mds() == false)
		fatal("Failed to download new MDs for variable update");

	if (seq_config_guests() == false) {
		warning("Failed to reconfigure with new MDs for "
		    "variable update");
		return;
	}

	/* nothing more to do */
	if (!notify_guest)
		return;

	/* notify the affected guest of MD update */
	ldp->flag.md_notify = true;
	(void) md_update_notify(ldp);

	/*
	 * It does not make much sense to automatically save
	 * the configuration while a migration is in progress
	 * since the migration is in the middle of changing
	 * the configuration. After the migration completes,
	 * it will force an autosave.
	 */
	if (!MIG_INPROGRESS(mig_ldp))
		autosave_update_autosave();
}

/*
 * Updates the HV MD and the guest MD for each LDom with
 * a new MD pending.
 */
static bool_t
seq_download_mds(void)
{
	ldom_t		*ldp;
	md_t		*mdp;
	int		size;
	uint8_t		*bufp;

	if (zeus.mode == ZEUS_CONFIG_MODE)
		return (true);

	DBG(zprintf("%s\n", __func__));

	/*
	 * Walk all the LDoms and update the MDs of those requiring new MDs.
	 */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (!ldp->flag.gen_new_md)
			continue;
		if (ldp->state == INACTIVE)
			fatal("%s: Internal error, inactive guest %s flagged "
			    "for MD generation", __func__, ldp->namep);

		DBG(zprintf("\n*** Downloading new guest MD for %s ***\n\n",
		    ldp->namep));

		mdp = seq_gen_guest_md(ldp, false);

		/* any backend specific additions? */
		backend.aug_guestmd(ldp, mdp);

		/*
		 * Allocate location of new guest MD.
		 * It's double buffered, so can't clobber the old one yet.
		 */
		size = md_gen_bin(mdp, &bufp);
		if (size <= 0)
			fatal("Unable to generate binary MD for LDom %s "
			    "using internal MD", ldp->namep);
		else if (size > (FRAG_SIZE - pri.md_elbow_room))
			fatal("Configuration required a MD of size 0x%x "
			    "bytes which\nis larger than the maximum supported "
			    "size of 0x%x bytes", size,
			    FRAG_SIZE - pri.md_elbow_room);

		get_next_md_frag(ldp, size, false);

		backend.dl_guestmd(ldp, ldp->frags.next_md_frag->paddr, bufp,
		    size);

		Xfree(bufp);
		md_fini(mdp);
	}

	/*
	 * Download the new HV MD.
	 */

	DBG(zprintf("\n*** Downloading and initiating the new hypervisor "
	    "MD ***\n\n"));

	mdp = target.gen_hvmd(true);	/* Allocates zeus.next_hvmd_fp */

	backend.aug_hvmd(mdp);

	size = md_gen_bin(mdp, &bufp);
	if (size <= 0)
		fatal("Unable to generate HV binary MD using internal MD");
	else if (size > (max_hv_md_size - pri.md_elbow_room))
		fatal("Configuration required a MD of size 0x%x bytes which\n"
		    "is larger than the maximum supported size of 0x%x bytes",
		    size, max_hv_md_size - pri.md_elbow_room);

	ASSERT(zeus.next_hvmd_fp->paddr != NULL);

	backend.dl_hvmd(zeus.next_hvmd_fp->paddr, bufp, size, HVMD);

	Xfree(bufp);

	/* Replace saved HV MD */
	if (zeus.hv_mdp != NULL)
		md_fini(zeus.hv_mdp);
	zeus.hv_mdp = mdp;

	return (true);
}

/*
 * Updates every LDom that has a new MD pending.  Also configures any
 * unconfigured CPUs and other resources that are new to the guests.
 *
 * Returns false if there was a problem.
 */
static bool_t
seq_config_guests(void)
{
	ldom_t *ldp;

	if (zeus.mode == ZEUS_CONFIG_MODE)
		return (true);

	DBG(zprintf("seq_config_guests()\n"));

	backend.config_guests(CONFIG);

	/*
	 * If successful, release the old HVMD
	 */
	hv_md_free(zeus.current_hvmd_fp);
	zeus.current_hvmd_fp = zeus.next_hvmd_fp;
	zeus.next_hvmd_fp = NULL;

	/*
	 * Rotate the next MD into the current MD of all those LDoms
	 * whose MDs need updating.
	 */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (!ldp->flag.gen_new_md)
			continue;

		if (ldp->state == INACTIVE)
			fatal("%s: Internal error, inactive guest %s flagged "
			    "for MD generation", __func__, ldp->namep);

		/*
		 * Rotate the next MD into the current MD
		 */
		frag_free(ldp->frags.committed_md_frag);
		ldp->frags.committed_md_frag = ldp->frags.next_md_frag;
		ldp->frags.next_md_frag = NULL;
		ldp->flag.gen_new_md = false;
	}

	/*
	 * Any previously unconfigured CPUs or crypto devices or memory blocks
	 * are now configured.
	 */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		vcpu_t	*vcpu;
		vmau_t	*vmau;
		mblock_t    *mblk;

		if (ldp->state == INACTIVE)
			continue;

		CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
			vcpu->flag_isconfigd = true;
		}

		CHAIN_ITER(ldp->bound.vmaus, vmau, ldom) {
			vmau->flag_isconfigd = true;
		}

		CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
			mblk->flag_isconfigd = true;
		}
	}

	return (true);
}

/*
 * Sequencer operations to start LDoms.
 */
void
seq_start_bound(void)
{
	backend.start_new_guests();
}

/*
 * Sequencer operations to stop LDoms.
 */
void
seq_stop_bound(bool_t force)
{
	int	rc = 0;
	ldom_t	*ldp;
	int	(*send_shutdown)(ldom_t *) = NULL;
	bool_t	channel_up;

	/*
	 * Notify all domains that they should shutdown.
	 */
	CHAIN_ITER(zeus.inprogress, ldp, inprog) {
		ASSERT(ldp->state > BOUND);

		DBG(zprintf("%s: stopping %s\n", __func__, ldp->namep));

		/* reset for this LDom */
		channel_up = false;

		/*
		 * If the domain has been suspended, the domain
		 * services channel has already been removed. So,
		 * skip the attempt to gracefully shut the domain
		 * down.
		 */
		if (ldp->state != SUSPENDED) {
			mutex_lock(&ldp->ds_chan->lock);
			if (ldp->ds_chan->state == CHAN_UP) {
				ldp->ds_chan->usecnt++;
				channel_up = true;
			}
			mutex_unlock(&ldp->ds_chan->lock);
		}

		if (channel_up) {
			send_shutdown = (int (*)(ldom_t *))
			    ds_service_routine(ldp->ds_chan,
			    DOMAIN_SHUTDOWN_SVC, "platsvc_send_shutdown");

			/*
			 * Send a shutdown request to the LDom if a forced
			 * stop has not been requested and the LDom has
			 * registered the domain shutdown service.
			 */
			if ((!force) && (send_shutdown != NULL)) {
				rc = send_shutdown(ldp);
			} else {
				rc = 0;
			}

			dec_chan_usecnt(ldp->ds_chan, true);

			if (rc > 0)
				continue; /* LDom nacked shutdown, move on */
		}

		/*
		 * Either the LDom acked a shutdown request or a stop request
		 * is going to be sent to the hypervisor. Hence we mark it as
		 * stopping. If the HV stop request fails (because the guest
		 * is beyond the point of no return in the HV during a reset)
		 * the guest will be marked as active again.
		 */
		ldp->state = STOPPING;

		/*
		 * Forcibly stop the LDom if a forced stop was requested or if
		 * a domain shutdown request could not be delivered to the
		 * LDom (either because of error or because the LDom has not
		 * registered the domain shutdown service.)
		 */
		if (force || (send_shutdown == NULL) || (rc < 0)) {
			/*
			 * If the LDom owns any I/O, and the PRI has flagged
			 * to us that we can't perform a reset of any
			 * underlying PCIe root complex, then we can't perform
			 * the HV stop request and we just leave the domain in
			 * ACTIVE state.
			 */
			if (!ldom_io_reset_allowed(ldp)) {
				warning("seq_stop_bound: force stop disabled "
				    "on I/O domain %s", ldp->namep);
				ldp->state = ACTIVE;
			} else {
				mem_dr_cleanup(ldp);
				backend.stop_guest(ldp);
			}
		}
	}
}

/*
 * Free all the queued up asynchronous HV messages related to LDom 'ldp'.
 */
static void
free_asynchronous_messages(ldom_t *ldp)
{
	seq_t *seq, *next;
	hvctl_msg_t *msg;
	uint32_t res, resid;

	mutex_lock(&sequencer.lock);

	for (seq = CHAIN_START(sequencer.seq); seq != NULL; seq = next) {
		next = CHAIN_NEXT(seq->link);

		if (seq->type == SEQ_ASYNC_HV_MSG) {
			msg = seq->d.async_hv.msg;
			res = ntoh32(msg->msg.resstat.res);
			if (res == HVctl_res_guest) {
				resid = ntoh32(msg->msg.resstat.resid);
				if (resid == ldp->res_id->id) {
					CHAIN_UNLINK(sequencer.seq, seq, link);
					Xfree(msg);
					Xfree(seq);
				}
			}
		}
	}

	mutex_unlock(&sequencer.lock);
}

extern void teardown_vnet_peer_bindings(vnet_t *vnet, bool_t unbinding);
extern int vnet_mig_link(ldom_t *ldp, md_t *hvmdp, response_t *resp);

void
seq_isolate_ldom(ldom_t *ldp, response_t *resp)
{
	vnet_t *vnet;

	/*
	 * Need to unlink all peer vnet connections. These have
	 * no meaning after the migration. Do not touch any of
	 * the connections to a virtual switch.
	 */
	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		if (vnet->binding == NULL)
			continue;

		if (CHAIN_LENGTH(vnet->binding->peers) != 0) {
			teardown_vnet_peer_bindings(vnet, true);
			ldp->flag.gen_new_md = true;
			ldp->flag.md_notify = true;
		}
	}

	seq_rm_reconfigure(ldp);

	ds_unregister_domain(ldp);

	free_asynchronous_messages(ldp);

	/*
	 * Before we unbind the clients in an LDom, we need to mark
	 * all guests hosting either servers or clients tied somehow
	 * to this LDom with a flag that they need to be informed
	 * about an MD update.
	 */
	(void) vio_unbind(ldp, resp);

	free_all_staged_ldcs(NULL);

	/*
	 * Need to reconfigure the guests, but do not want to touch
	 * the actual HV configuration yet. This is a special case
	 * for migration.
	 */
	seq_reconfigure_guests_migrate();
}

void
seq_restore_ldom(ldom_t *ldp, response_t *resp)
{
	md_t	*mdp = ldp->mig_state.bind_mdp;

	ldp->state = BINDING;

	if (vio_mig_restore(ldp, mdp, zeus.hv_mdp, resp) != 0) {
		DBG2(zprintf("%s: vio_bind failed for %s\n",
		    __func__, ldp->namep));
		return;
	}

	DBG(zprintf("%s: vio_mig_restore complete...\n", __func__));

	DBG(zprintf("%s: %s %s a new MD\n", __func__, ldp->namep,
	    (ldp->flag.gen_new_md) ? "*needs*" : "*does not need*"));

	ldp->state = ACTIVE;

	/*
	 * Create the domain services channel so that when async
	 * registration requests come in, the DS channel will be
	 * ready. These requests can start immediately after the
	 * control domain receives its MD update notification, so
	 * this has to be done before guest reconfig. Otherwise,
	 * registration requests will be dropped.
	 */
	(void) ds_register_domain(ldp, false);

	seq_reconfigure_guests_migrate();

	/* open the DS channel for use */
	(void) ds_open_channel(ldp->ds_chan);

	/* wait for domain services to come online */
	mig_wait_for_domain_services(ldp);

	/* bind the vnet to any peers */
	if (vnet_mig_link(ldp, zeus.hv_mdp, resp) != 0) {
		DBG(zprintf("unable to link peer vnets\n"));
		return;
	}

	/* activate the new vnet peer bindings */
	seq_download_mds_do_config(ldp);
}

/*
 * This unbinds the now stopped resources that have been given to a guest, and
 * it's servers.  After the logical reclaim we download new MDs where relevant
 * to the HV, and get it to close down LDC channels etc.
 */
void
seq_unbind_ldoms(void)
{
	ldom_t *ldp;

	/*
	 * Unregister domain services, free any asynchronous messages queued
	 * for LDoms being unbound and unbind VIO before new MDs are generated
	 * and downloaded. VIO is unbound now so that any connections to the
	 * LDoms being unbound will be removed in the MDs of the LDoms which
	 * remain bound.
	 */
	CHAIN_ITER(zeus.inprogress, ldp, inprog) {

		ASSERT(ldp->state == UNBINDING);

		ds_unregister_domain(ldp);

		free_asynchronous_messages(ldp);

		/*
		 * Before we unbind the clients in an LDom, we need to mark
		 * all guests hosting either servers or clients tied somehow
		 * to this LDom with a flag that they need to be informed
		 * about an MD update.
		 */
		vio_unbind(ldp, NULL);
	}

	free_all_staged_ldcs(NULL);

	/*
	 * Reconfigure guests.
	 */
	seq_reconfigure_guests(false);

	/*
	 * Unbind the (remaining) resources bound to the LDoms being unbound.
	 * Rotate the next MD into the current MD of all those LDoms whose MDs
	 * were updated.
	 */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state == UNBINDING) {
			int i;
			vcpu_t *vcpu;
			ldm_err_t pm_result;
			pm_resid_t *resids;

			resids = Xcalloc(ldom_num_bound_vcpus(ldp) +
			    ldom_num_rsvd_vcpus(ldp), pm_resid_t);

			/*
			 * Grab list of resources for PM post-unbind
			 * notification.
			 */
			i = 0;
			CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
				resids[i].pi_res_type = PM_RES_CPU;
				resids[i].pi_res_name = CPU_RES_NAME;
				resids[i].u.cpu.pi_id_val =
				    vcpu->cpu->phys_id;
				i++;
			}
			ASSERT(i == ldom_num_bound_vcpus(ldp));

			/*
			 * Also include the reserved cpus (by migration)
			 * in the list we send to PM.
			 */
			CHAIN_ITER(ldp->mig_rsvd.vcpus, vcpu, ldom) {
				resids[i].pi_res_type = PM_RES_CPU;
				resids[i].pi_res_name = CPU_RES_NAME;
				resids[i].u.cpu.pi_id_val =
				    vcpu->cpu->phys_id;
				i++;
			}
			ASSERT(i == ldom_num_bound_vcpus(ldp) +
			    ldom_num_rsvd_vcpus(ldp));

			unbind_an_ldom(ldp);
			ldp->state = INACTIVE;

			/* Send post-unbind PM notification here */
			if (pm_cpu_events_on) {
				send_pm_cpu_event(ldp, REM_RESOURCE_POST,
				    resids, i, &pm_result);
				/* FIXME: add error handling */
				pm_free_errors(&pm_result);
			}
			Xfree(resids);

			DBG2(zprintf("Guest %s set to INACTIVE\n", ldp->namep));
		}

		if (!ldp->flag.gen_new_md)
			continue;

		if (ldp->state == INACTIVE)
			fatal("%s: Internal error, inactive guest %s flagged "
			    "for MD generation", __func__, ldp->namep);

		/* rotate the next MD into the current MD */
		frag_free(ldp->frags.committed_md_frag);
		ldp->frags.committed_md_frag = ldp->frags.next_md_frag;
		ldp->frags.next_md_frag = NULL;

		ldp->flag.gen_new_md = false;

		free_from_stage(&ldp->guest_free_ldcs);
	}

	autosave_update_autosave();
}

void
seq_rm_ldoms(void)
{
	ldom_t *ldp;

	/* Upon success - INACTIVE -> *POOF* */
	CHAIN_ITER(zeus.inprogress, ldp, inprog) {
		switch (ldp->state) {
		case INACTIVE:
			break;
		default:
			fatal("%s: Internal error in progress list has LDom %s "
			    "in state (%d) %s\n", __func__, ldp->namep,
			    (int)ldp->state, ldom_state_str(ldp->state));
		}

		DBG(zprintf("%s: removing %s\n", __func__, ldp->namep));
	}
}

static void
set_mem_generation_id(md_t *mdp, ldom_t *ldp)
{
	md_node_t **nodelistp;
	int num;

	if (ldp->props.mem_generation_id != 0) {
		nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));
		num = md_scan_dag(mdp, NULL, md_find_name(mdp, "memory"),
		    md_find_name(mdp, "fwd"), nodelistp);
		if (num > 0)
			md_update_value_property(mdp, nodelistp[0],
			    "memory-generation-id#",
			    ldp->props.mem_generation_id);
		Xfree(nodelistp);
	}
}

md_node_t *
md_clone_node(md_t *mdp, md_node_t *node)
{
	md_node_t *newnode;
	md_prop_t *prop;
	char *name;

	/* Create new node of same type */
	newnode = md_new_node(mdp, node->typep->strp);

	/*
	 * Now clone the properties.  An arc is treated like any other
	 * property; we copy its value.  This means it points to the same
	 * node as the original's.  We then traverse the new arc, and create
	 * a new inverse arc back to the clone.
	 */
	for (prop = CHAIN_START(node->prop_list); prop != NULL;
	    prop = CHAIN_NEXT(prop->link)) {

		name = prop->sp->strp;

		switch (prop->type) {
		case MDET_PROP_VAL:
			md_add_value_property(mdp, newnode, name,
			    prop->d.value);
			break;
		case MDET_PROP_STR:
			md_add_string_property(mdp, newnode, name,
			    (char *)(prop->d.dbp->datap));
			break;
		case MDET_PROP_DAT:
			md_add_data_property(mdp, newnode, name,
			    prop->d.dbp->size, prop->d.dbp->datap);
			break;
		case MDET_PROP_ARC:
			ASSERT(prop->d.arc.is_ptr);
			md_add_arc_property(mdp, newnode, name,
			    prop->d.arc.val.nodep);
			/* Add new arc back to clone */
			md_add_arc_property(mdp, prop->d.arc.val.nodep,
			    strcmp(name, "fwd") ? "fwd" : "back", newnode);
			break;
		default:
			fatal("md_clone_node: unknown property type found");
		}
	}
	return (newnode);
}

/*
 * Generate memory and mblock nodes.
 */
static void
add_guest_memory(ldom_t *ldp, md_t *mdp)
{
	md_node_t **nodelist;
	int idx, num;

	nodelist = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	/*
	 * Find and match up all the mblock nodes, delete the unused ones,
	 * and clone any that have multiple bound regions on them.
	 */
	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "mblock"),
	    md_find_name(mdp, "fwd"), nodelist);
	if (num <= 0)
		fatal("Internal error cloned PRI MD has no mblock nodes");

	for (idx = 0; idx < num; ++idx) {
		uint64_t paddr, size;
		md_node_t *nodep;
		mblock_t *mblk;
		int num_in_pblock = 0;
		uint64_t congr_off;

		nodep = nodelist[idx];

		if (md_get_prop_val(mdp, nodep, "base", &paddr) != 0)
			fatal("Internal error mblock node has no "
			    "base property");

		if (md_get_prop_val(mdp, nodep, "size", &size) != 0)
			fatal("Internal error mblock node has no "
			    "size property");
		/*
		 * See which bound mblocks for this LDom fall within
		 * this physical block.  If there are none, delete the
		 * node; if there is one, rewrite its base & size props;
		 * any more requires cloning the physical block node.
		 */
		CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
			if (mblk->paddr >= paddr &&
			    mblk->paddr < (paddr + size)) {
				md_node_t *newnode;
				if (++num_in_pblock > 1) {
					newnode = md_clone_node(mdp, nodep);
				} else {
					newnode = nodep;
				}

				md_update_value_property(mdp, newnode, "base",
				    mblk->raddr);
				md_update_value_property(mdp, newnode, "size",
				    mblk->size);
				congr_off = mblk->paddr - mblk->raddr;
				DBG(zprintf("CONGR: PA_base = 0x%"PRIx64", "
				    "RA_base = 0x%"PRIx64", "
				    "PA - RA = 0x%"PRIx64"\n",
				    mblk->paddr, mblk->raddr, congr_off));
				congr_off &= pri.addr_congr_mask;
				md_update_value_property(mdp, newnode,
				    ADDR_CONGR_OFF_PROP, congr_off);
			}
		}

		if (num_in_pblock == 0)
			md_delete_node(mdp, nodep);
	}

	Xfree(nodelist);
}

static void
delete_maus_cwq(md_t *md, md_node_t *mau)
{
	md_node_t **nodes;
	int i, num;

	nodes = Xmalloc(md_node_count(md) * sizeof (md_node_t *));

	/*
	 * Find cwq corresponding to mau by going back to an owning cpu and
	 * then forward to the cwq exec unit.
	 */

	num = md_scan_dag(md, mau, md_find_name(md, "cpu"),
	    md_find_name(md, "back"), nodes);
	if (num < 1)
		fatal("No cpu associated with a bound mau");

	num = md_scan_dag(md, nodes[0], md_find_name(md, "exec-unit"),
	    md_find_name(md, "fwd"), nodes);

	for (i = 0; i < num; i++) {
		char *type;
		int len;

		if (md_get_prop_data(md, nodes[i], "type", (uint8_t **)&type,
		    &len) != 0)
			fatal("exec-unit has no type property");

		if (strcmp(type, "cwq") == 0)
			break;
	}

	if (i == num)
		fatal("No cwq associated with mau");

	md_delete_node(md, nodes[i]);

	Xfree(nodes);
}

/*
 * Add devaliases "disk" and "net" for the guest domain if needed.
 * "disk" default to the first virtual disk unless it is already configured.
 * "net" default to the first virtual net unless it is already configured.
 */
static void
add_ldm_devaliases(md_t *md, ldom_t *ldp)
{
	md_node_t	*devalias;
	char		*alias;
	bool_t		have_alias;
	int		dev_id;
	vdisk_t		*vdisk;
	vnet_t		*vnet;

	devalias = md_find_node(md, "devalias");
	ASSERT(devalias != NULL);

	have_alias = false;
	dev_id = -1;
	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {
		if (strcmp(vdisk->name, "disk") == 0)
			have_alias = true;
		else if (dev_id == -1)
			dev_id = vdisk_devid(vdisk);
	}

	if (!have_alias && dev_id != -1) {
		/* add disk alias to devalias node */
		alias = zsprintf("/%s@%x/%s@%x/disk@%d", VDEVS_NAME,
		    VDEVS_CFG_HANDLE, CHANNELS_NAME, CHANNELS_CFG_HANDLE,
		    dev_id);
		md_add_string_property(md, devalias, "disk", alias);
		Xfree(alias);
	}

	have_alias = false;
	dev_id = -1;
	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		if (strcmp(vnet->name, "net") == 0)
			have_alias = true;
		else if (dev_id == -1)
			dev_id = vnet_devid(vnet);
	}

	if (!have_alias && dev_id != -1) {
		/* add net alias to devalias node */
		alias = zsprintf("/%s@%x/%s@%x/network@%d", VDEVS_NAME,
		    VDEVS_CFG_HANDLE, CHANNELS_NAME, CHANNELS_CFG_HANDLE,
		    dev_id);
		md_add_string_property(md, devalias, "net", alias);
		Xfree(alias);
	}
}

md_t *
seq_gen_guest_md(ldom_t *ldp, bool_t ignore_pm)
{
	bound_iodevice_t *bip;
	md_t		*mdp;
	md_node_t	**nodelist, **core_md_nodelist;
	md_node_t	**ionodelist;
	md_node_t	*phys_io;
	md_node_t	*platnode;
	int		num, core_md_num, x;
	int		idx;
	int		delidx;
	int		incidx;
	md_string_t	*mdstrp;
	uint64_t	status;

	/* start with the template created from the PRI */
	mdp = md_clone(pri.guest_template);

	nodelist = Xmalloc(md_node_count(mdp) * sizeof (md_node_t *));

	if (ldp != zeus.hvctl_ldom_p) {
		int i, num_rng_nodes;
		/*
		 * This is not a control domain hence prune off all nodes
		 * which are only of interest in a control domain.
		 */
		md_find_delete_node(mdp, "platform-services");
		md_find_delete_node(mdp, "dimm_configuration");

		/* Remove all RNG nodes from non-primary domains */
		num_rng_nodes = grab_exec_units(mdp, "rng", nodelist);
		for (i = 0; i < num_rng_nodes; ++i)
			md_mark_node_deleted(nodelist[i]);
	}

	/* FIXME PRI where does reset reason come from */
	md_update_value_property(mdp, md_root_node(mdp), "reset-reason", 0);

	/*
	 * Assign the generation number for the new MD that is created
	 * for this LDom. Only increment the generation number if the
	 * MD is being actively modified.
	 */
	if (ldp->flag.gen_new_md)
		ldp->props.md_generation_num++;

	md_update_value_property(mdp, md_root_node(mdp), "md-generation#",
	    ldp->props.md_generation_num);

	set_mem_generation_id(mdp, ldp);

	/*
	 * Find and prune off all the unused CPU nodes.
	 * Give each virtual CPU its real virtual ID.
	 *
	 * 2.0 FIXME: Clone the cpu nodes if there is more than one vcpu
	 * sharing a physical one.
	 */
	DBG_BLOCK {
		vcpu_t *vcpu;
		CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
			zprintf("Guest %s: Vcpu %d bound to strand %d\n",
			    ldp->namep, vcpu->vid, vcpu->cpu->phys_id);
		}
	}

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "cpu"),
	    md_find_name(mdp, "fwd"), nodelist);
	if (num <= 0)
		fatal("Guest template has no cpu nodes");

	if (!ignore_pm) {
		core_md_nodelist = Xcalloc(md_node_count(pri.core_mdp),
		    md_node_t *);

		core_md_num = md_scan_dag(pri.core_mdp, NULL,
		    md_find_name(pri.core_mdp, "cpu"),
		    md_find_name(pri.core_mdp, "fwd"), core_md_nodelist);
	}

	/*
	 * For each node, find its binding and re-id it,
	 * and delete any unfound nodes.
	 */
	/*
	 * FIXME: this is a really dumb implementation,
	 * need to change from an n^2 algorithm
	 */

	for (idx = 0, delidx = 0; idx < num; idx++) {
		md_node_t *nodep;
		uint64_t pid;
		uint64_t core_md_status;
		vcpu_t *vcpu = NULL;

		nodep = nodelist[idx];

		pid = pri_cpu_phys_id(mdp, nodep);

		/*
		 * For each physical cpu in the guest template portion of the
		 * PRI, attempt to match it against a bound vcpu.  If a
		 * match is found, convert the "id" property to the virtual
		 * value.  If a match is not found, the cpu node is marked for
		 * removal from the MD.
		 *
		 * If ignore_pm is true, include the power-managed cpus in the
		 * match attempt, otherwise do not (causing them to be removed
		 * from the MD).
		 */
		if (ignore_pm) {
			core_md_status = PM_NORMAL;
		} else {
			for (x = 0; x < core_md_num; ++x) {
				if (pid == pri_cpu_phys_id(pri.core_mdp,
				    core_md_nodelist[x])) {
					status = md_get_prop_val(pri.core_mdp,
					    core_md_nodelist[x],
					    CORE_MD_PM_STATUS,
					    &core_md_status);
					break;
				}
			}
			ASSERT(x < core_md_num && status == 0);
		}

		if (core_md_status != PM_MANAGED) {
			CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
				if (vcpu->cpu->phys_id == pid) {
					md_update_value_property(mdp, nodep,
					    "id", vcpu->vid);
					break;
				}
			}
		}

		if (vcpu == NULL)
			/* mark this physical CPU node for removal */
			nodelist[delidx++] = nodep;
	}

	/* delete the unused nodes */
	for (idx = 0; idx < delidx; idx++) {
		md_mark_node_deleted(nodelist[idx]);
	}

	/*
	 * Find and prune off all the unused MAU nodes.
	 * give each virtual MAU its real virtual ID.
	 */
	num = grab_exec_units(mdp, "mau", nodelist);

	if (ldom_num_bound_maus(ldp) == 0) {
		delidx = num; 	/* mark all MAU nodes for removal */
	} else {
		md_node_t **stranslistp;

		DBG_BLOCK {
			vmau_t *vmau;
			CHAIN_ITER(ldp->bound.vmaus, vmau, ldom) {
				zprintf("Guest %s: vmau %"PRIu64" bound to "
				    "mau %s\n", ldp->namep, vmau->vid,
				    vmau->mau->repr);
			}
		}

		/*
		 * In nodelist, we have the mau nodes which are accessible to
		 * the guest, based on the guest's cpu binding. However, the
		 * guest may have fewer maus bound, so we need to prune off
		 * the ones not bound. For this, we look at each MAU node and
		 * see if the strands are in the guest's bound mau list.
		 */

		stranslistp = Xmalloc(sizeof (md_node_t) * md_node_count(mdp));

		for (idx = 0, delidx = 0; idx < num; idx++) {
			int		numstrands;
			md_node_t	*nodep;
			int		j;

			nodep = nodelist[idx];

			/*
			 * Find the cpu nodes sharing this mau
			 */
			numstrands = md_scan_dag(mdp,  nodep,
			    md_find_name(mdp, "cpu"),
			    md_find_name(mdp, "back"), stranslistp);

			if (numstrands == 0)
				fatal("Guest template contains mau node with "
				    "no associated cpu nodes");

			/*
			 * For each strand sharing this MAU, check
			 * if the strand is bound to the guest
			 */
			for (j = 0; j < numstrands; j++) {
				uint64_t strand_id;
				if (md_get_prop_val(mdp, stranslistp[j],
				    "id", &strand_id) < 0)
					fatal("Missing cpu id");
				if (mau_bound_to_cpu(ldp, strand_id) != NULL)
					break;
			}

			if (j == numstrands)
				/* mark this physical MAU node for removal */
				nodelist[delidx++] = nodep;
		}

		Xfree(stranslistp);
	}

	/* now delete the unused nodes */
	for (idx = 0; idx < delidx; idx++) {
		if (pri.have_cwq)
			delete_maus_cwq(mdp, nodelist[idx]);
		md_mark_node_deleted(nodelist[idx]);
	}

	add_guest_memory(ldp, mdp);

	/*
	 * Find and prune off all the unused IO nodes.
	 *
	 * 2.0 FIXME: Clone the IO nodes if there is more than one
	 * VIO sharing a physical one.
	 */
	DBG_BLOCK {
		for (bip = CHAIN_START(ldp->bound.iodevices);
		    bip != NULL; bip = CHAIN_NEXT(bip->ldom)) {
			zprintf("IO %s bound to phys IO %s\n",
			    bip->name, bip->phys_iodev->name);
		}
	}

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "phys_io"),
	    md_find_name(mdp, "fwd"), nodelist);
	if (num != 1)
		fatal("Guest template does not have a singleton phys_io node");
	phys_io = nodelist[0];

	num = md_scan_dag(mdp, phys_io, md_find_name(mdp, "iodevice"),
	    md_find_name(mdp, "fwd"), nodelist);
	if (num <= 0)
		fatal("Guest template has no iodevice nodes");

	ionodelist = Xcalloc(num, sizeof (md_node_t *));

	/*
	 * Walk top level IO MD nodes and delete any not bound to this LDom.
	 */

	for (idx = 0, delidx = 0, incidx = 0; idx < num; idx++) {
		md_node_t	*nodep, *parent;
		uint64_t	cfg_handle;

		nodep = nodelist[idx];

		if (md_get_prop_arc(mdp, nodep, "back", &parent) < 0)
			fatal("Missing back arc in guest template "
			    "iodevice node");

		if (parent != phys_io)
			continue;	/* not at top level */

		if (md_get_prop_val(mdp, nodep, "cfg-handle", &cfg_handle) < 0)
			fatal("Missing cfg-handle property in guest template "
			    "iodevice node");

		CHAIN_ITER(ldp->bound.iodevices, bip, ldom) {
			if (bip->phys_iodev->cfg_handle == cfg_handle) {
				break;
			}
		}

		if (bip == NULL)
			/* not bound - mark this physical IO node for removal */
			nodelist[delidx++] = nodep;
		else
			ionodelist[incidx++] = nodep;
	}

	/* delete the unused nodes */
	for (idx = 0; idx < delidx; idx++) {
		md_mark_node_deleted(nodelist[idx]);
	}

	pcie_fix_guestmd(ldp, mdp, ionodelist, incidx);
	Xfree(ionodelist);

	/*
	 * Zeus is running hence domaining is enabled.
	 */
	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "platform"),
	    md_find_name(mdp, "fwd"), nodelist);
	if (num != 1)
		fatal("Guest template does not have a singleton platform node");
	platnode = nodelist[0];

	md_update_value_property(mdp, platnode, "domaining-enabled", 1);

	/*
	 * Set MAC address & hostid by updating template values previously
	 * cloned from PRI into guest MD's platform node. Do this only if this
	 * is a non-primary domain.
	 */
	if ((ldp != zeus.hvctl_ldom_p) &&
	    (ldp->props.mac_address != ILLEGAL_MAC_ADDR)) {
		/* Override MAC address */
		md_update_value_property(mdp, platnode, "mac-address",
		    ldp->props.mac_address);

	}

	if (ldp != zeus.hvctl_ldom_p) {
		uint64_t	hostid;
		md_string_t	*mdstrp;
		md_prop_t	*propp;
		if (ldp->props.hostid == UNASSIGNED_HOSTID) {
			/* Derive hostid */
			hostid = DERIVED_HOSTID(ldp->props.mac_address);
		} else {
			hostid = ldp->props.hostid;
		}

		mdstrp = md_find_name(mdp, "hostid");
		propp = md_find_node_prop(mdp, platnode, mdstrp, MDET_PROP_VAL);
		if (NULL == propp) {
			warning("platform node missing hostid property");
			md_add_value_property(mdp, platnode, "hostid", hostid);
		} else {
			/* Override hostid */
			propp->d.value = hostid;
		}
	}

	/*
	 * Generate guest keys and variables.
	 */
	gen_guest_md_keys(ldp, mdp);
	gen_guest_md_variables(ldp, mdp);

	/* Add universally unique id (UUID) */
	md_add_string_property(mdp, platnode, "uuid", ldp->props.uuid);

	/*
	 * Now add virtual-only nodes (i.e. nodes not already in the PRI-based
	 * cloned MD) to guest MD.
	 */
	vio_gen_guest_md(ldp, mdp);

	/*
	 * Set saved devaliases if any.
	 */
	if (ldp->devalias_md != NULL) {
		md_node_t *devalias;

		/* remove current devalias node */
		mdstrp = md_find_name(mdp, "devalias");
		num = md_scan_dag(mdp, NULL, mdstrp, md_find_name(mdp, "fwd"),
		    nodelist);
		if (num > 0)
			md_delete_node(mdp, nodelist[0]);

		/* add saved devaliases */
		md_add(mdp, ldp->devalias_md);

		/* link up devalias node */
		mdstrp = md_find_name(mdp, "devalias");
		num = md_scan_nodes(mdp, mdstrp, nodelist);
		devalias = nodelist[0];
		mdstrp = md_find_name(mdp, "openboot");
		num = md_scan_dag(mdp, NULL, mdstrp, md_find_name(mdp, "fwd"),
		    nodelist);
		md_add_arc_property(mdp, nodelist[0], "fwd", devalias);
	}

	if (ldp != zeus.hvctl_ldom_p) {
		/* add "disk" and "net" aliases to devalias node if needed */
		add_ldm_devaliases(mdp, ldp);
	}

	/*
	 * Clean up all un-referenced or marked as deleted nodes.
	 */
	md_scour(mdp, md_find_name(mdp, "back"));

	Xfree(nodelist);
	if (!ignore_pm)
		Xfree(core_md_nodelist);

	return (mdp);
}

/*
 * Add a sequencer callout to check client's dependencies.
 */
static void
add_client_callout(cli_conn_t *client)
{
	callout_t *callout;

	callout = Xmalloc(sizeof (*callout));

	client->callout = callout;
	callout->client = client;
	callout->time = time(NULL) + CLIENT_CALLOUT_TIMEOUT;

	CHAIN_ADD(sequencer.callouts, callout, link);
}

/*
 * Delete the callout for client 'client'.
 */
void
seq_delete_client_callout(cli_conn_t *client)
{
	CHAIN_UNLINK(sequencer.callouts, client->callout, link);
	Xfree(client->callout);
	client->callout = NULL;
}

/*
 * Process a callout. The associated client has timed out waiting on
 * events and a failure response is sent.
 */
static void
process_client_callout(callout_t *callout)
{
	cli_conn_t *client = callout->client;
	waiter_t *w;

	CHAIN_ITER(client->waiters, w, next_ldom) {
		if (w->event & ACTIVE) {
			if (client->class == LDM_CLASS_CONTROL) {
				xml_update_status(client, w->cmd,
				    w->ldom->namep, XML_FAILURE,
				    "%s start timed out", w->ldom->namep);
			} else {
				rprintf(client->resp, "%s start timed out\n",
				    w->ldom->namep);
			}
		} else {
			if (client->class == LDM_CLASS_CONTROL) {
				xml_update_status(client, w->cmd,
				    w->ldom->namep, XML_FAILURE,
				    "%s stop timed out. The domain might "
				    "still be in the process of shutting "
				    "down. Either let it continue, or "
				    "specify -f to force it to stop.",
				    w->ldom->namep);
			} else {
				rprintf(client->resp, "%s stop timed out. "
				    "The domain might still be in the process "
				    "of shutting down.\nEither let it "
				    "continue, or specify -f to force it to "
				    "stop.\n", w->ldom->namep);
			}
		}
	}

	free_clients_waiters(client);
	seq_delete_client_callout(client);

	client->resp->ret = LDOM_GENERAL_FAILURE;

	if (client->class == LDM_CLASS_CONTROL) {
		if (!client->xmpp_msg) {
			client_write_buffer(client, client->bufp);
		} else {
			xmpp_write_buffer(client, false);
		}
	} else {
		client_write_response(client);
	}
}

/*
 * Check the queue of callouts and process those which have timed out.
 */
static void
process_client_callouts(void)
{
	callout_t *callout;

	while (CHAIN_LENGTH(sequencer.callouts) > 0) {
		callout = CHAIN_START(sequencer.callouts);

		if (callout->time < time(NULL)) {
			process_client_callout(callout);
		} else {
			break;
		}
	}
}

/*
 * The sequencer service routine. This routine runs in a thread and loops
 * forever consuming and handling sequencer tasks.
 */
/*ARGSUSED*/
void *
sequence(void *unused)
{
	seq_t *seq;
	callout_t *callout;
	cli_conn_t *conn;
	ds_seq_req_t *dsreq;
	ds_svc_t *ds;
	timestruc_t to;
	int err = 0;
	bool_t do_vcpus;

#if ZDBG
	mutex_lock(&seq_monitor);
#endif

	while (1) {
		/* wait for something to do */
		mutex_lock(&sequencer.lock);
		do {
			seq = CHAIN_START(sequencer.seq);
			if (seq == NULL) {
				callout = CHAIN_START(sequencer.callouts);
				if (callout == NULL) {
					err = cond_wait(&sequencer.cond,
					    &sequencer.lock);
				} else {
					to.tv_sec = callout->time;
					to.tv_nsec = 0;
					err = cond_timedwait(&sequencer.cond,
					    &sequencer.lock, &to);
				}
			} else {
				CHAIN_UNLINK(sequencer.seq, seq, link);
			}
		} while ((seq == NULL) && (err != ETIME));
		mutex_unlock(&sequencer.lock);

		if (seq == NULL) {
			process_client_callouts();
			continue;
		}

		switch (seq->type) {
		case SEQ_ASYNC_HV_MSG:
			/* call callback to handle it */
			err = seq->d.async_hv.cb(seq->d.async_hv.msg);
			if (err != 0)
				fatal("(%d) Processing asynchronous HV message",
				    err);
			break;

		case SEQ_CLI_CMD:
			conn = seq->d.cli_cmd;

			(void) ldm_cli_command(conn);
			free(conn->buf);
			conn->buf = 0;
			if (CHAIN_LENGTH(conn->waiters) == 0)
				client_write_response(conn);
			else
				add_client_callout(conn);
			break;

		case SEQ_XML_CMD:
			conn = seq->d.cli_cmd;

			zeus_ldom_xml_command(conn);
			if (CHAIN_LENGTH(conn->waiters) == 0) {
				if (!conn->xmpp_msg) {
					client_write_buffer(conn, conn->bufp);
				} else {
					xmpp_write_buffer(conn, false);
				}
			} else {
				add_client_callout(conn);
			}
			break;

		case SEQ_MD_UPDATE:
			/*FALLTHROUGH*/
		case SEQ_DS_REQ:
			if (seq->type == SEQ_MD_UPDATE) {
				ds = seq->d.ds;
				if (ds->chan->ldp->flag.md_notify)
					md_update_notify(ds->chan->ldp);
			} else {
				dsreq = seq->d.ds_req;
				ds = dsreq->svc;
				ds->ops->seq_handler(ds, dsreq->msg,
				    dsreq->len);
				Xfree(dsreq->msg);
				Xfree(dsreq);
			}

			dec_chan_usecnt(ds->chan, true);

			ds_svc_decref(ds, 1);
			break;

		case SEQ_PRI_REQ:
			conn = seq->d.cli_cmd;

			cntrl_pri_req(conn);
			break;

		case SEQ_UTIL_READ:
			do_vcpus = seq->d.do_vcpus;

			handle_utils(do_vcpus);
			break;

		case SEQ_DO_WORK:
			seq->d.work.func(seq->d.work.args);
			(void) mutex_lock(&sequencer.lock);
			sequencer.last_processed_id = seq->id;
			DBG2(zprintf("%s: seqID 0x%" PRIx64 " processed (of 0x%"
			    PRIx64 " enqueued)\n", __func__, seq->id,
			    sequencer.next_enqueued_id - 1));
			(void) cond_broadcast(&sequencer.task_cond);
			(void) mutex_unlock(&sequencer.lock);
			break;

		default:
			fatal("unknown sequence type %d", seq->type);
			break;
		}

		Xfree(seq);

		process_client_callouts();
	}

	/*NOTREACHED*/
}

/*
 * Queue up an asynchronous HV message for handling within the sequencer.
 */
void
seq_enqueue_async_hv_msg(seq_async_hv_cb_t *cb, void *msg)
{
	seq_t *seq;

	seq = Xmalloc(sizeof (*seq));
	seq->type = SEQ_ASYNC_HV_MSG;
	seq->d.async_hv.cb = cb;
	seq->d.async_hv.msg = msg;

	mutex_lock(&sequencer.lock);
	CHAIN_ADD(sequencer.seq, seq, link);
	(void) cond_signal(&sequencer.cond);
	mutex_unlock(&sequencer.lock);
}

/*
 * Queue up a utilization read for handling within the sequencer.
 */
void
seq_enqueue_util(bool_t do_vcpus)
{
	seq_t *seq;

	seq = Xmalloc(sizeof (*seq));
	seq->type = SEQ_UTIL_READ;
	seq->d.do_vcpus = do_vcpus;

	mutex_lock(&sequencer.lock);
	CHAIN_ADD(sequencer.seq, seq, link);
	(void) cond_signal(&sequencer.cond);
	mutex_unlock(&sequencer.lock);
}

/*
 * Queue up a command for handling within the sequencer.
 */
void
seq_enqueue_cmd(cli_conn_t *conn, int type)
{
	seq_t *seq;

	seq = Xmalloc(sizeof (*seq));
	seq->type = type;
	seq->d.cli_cmd = conn;

	mutex_lock(&sequencer.lock);
	CHAIN_ADD(sequencer.seq, seq, link);
	(void) cond_signal(&sequencer.cond);
	mutex_unlock(&sequencer.lock);
}

/*
 * Queue up a domain service request for handling within the sequencer.  This
 * is only called from within service data handlers, hence the channel
 * reference count will be >= 1 when this is called and there is no danger of
 * the channel being freed during execution.
 */
void
seq_enqueue_ds_req(ds_seq_req_t *dsreq)
{
	ds_channel_t *chan;
	seq_t *seq;

	chan = dsreq->svc->chan;

	seq = Xmalloc(sizeof (*seq));
	seq->type = SEQ_DS_REQ;
	seq->d.ds_req = dsreq;

	mutex_lock(&sequencer.lock);

	mutex_lock(&chan->lock);
	if (chan->state != CHAN_UP) {
		mutex_unlock(&chan->lock);
		mutex_unlock(&sequencer.lock);
		Xfree(dsreq);
		Xfree(seq);
		return;
	}
	chan->usecnt++;
	mutex_unlock(&chan->lock);

	ds_svc_incref(dsreq->svc);

	CHAIN_ADD(sequencer.seq, seq, link);
	(void) cond_signal(&sequencer.cond);

	mutex_unlock(&sequencer.lock);
}

/*
 * Queue up a MD update notification to be delivered from within the
 * sequencer.
 */
void
seq_enqueue_md_notify(ds_svc_t *ds)
{
	seq_t *seq;

	seq = Xmalloc(sizeof (*seq));
	seq->type = SEQ_MD_UPDATE;
	seq->d.ds = ds;

	mutex_lock(&sequencer.lock);

	mutex_lock(&ds->chan->lock);
	if (ds->chan->state != CHAN_UP) {
		mutex_unlock(&ds->chan->lock);
		mutex_unlock(&sequencer.lock);
		Xfree(seq);
		return;
	}
	ds->chan->usecnt++;
	mutex_unlock(&ds->chan->lock);

	ds_svc_incref(ds);

	CHAIN_ADD(sequencer.seq, seq, link);
	(void) cond_signal(&sequencer.cond);

	mutex_unlock(&sequencer.lock);
}

/*
 * Remove from the sequencer queue all requests for domain service 'svc'.
 * the lock argument specifies whether to acquire channel mutex or not.
 */
void
seq_dequeue_ds_svc(ds_svc_t *svc, bool_t locking_reqd)
{
	seq_t *seq, *next;
	int count = 0;

	mutex_lock(&sequencer.lock);

	for (seq = CHAIN_START(sequencer.seq); seq != NULL; seq = next) {
		next = CHAIN_NEXT(seq->link);

		if (seq->type == SEQ_DS_REQ && seq->d.ds_req->svc == svc) {
			CHAIN_UNLINK(sequencer.seq, seq, link);

			dec_chan_usecnt(svc->chan, locking_reqd);
			count++;

			Xfree(seq->d.ds_req->msg);
			Xfree(seq->d.ds_req);
			Xfree(seq);

		} else if (seq->type == SEQ_MD_UPDATE && seq->d.ds == svc) {
			CHAIN_UNLINK(sequencer.seq, seq, link);

			dec_chan_usecnt(svc->chan, locking_reqd);
			count++;

			Xfree(seq);
		}
	}

	mutex_unlock(&sequencer.lock);
	if (count > 0)
		ds_svc_decref(svc, count);
}

/*
 * Remove from the sequencer queue all domain service requests for
 * domain service channel 'chan'.
 */
void
seq_dequeue_ds_chan(ds_channel_t *chan)
{
	seq_t *seq, *next;

	mutex_lock(&sequencer.lock);

	for (seq = CHAIN_START(sequencer.seq); seq != NULL; seq = next) {
		next = CHAIN_NEXT(seq->link);

		if (seq->type == SEQ_DS_REQ &&
		    seq->d.ds_req->svc->chan == chan) {
			CHAIN_UNLINK(sequencer.seq, seq, link);

			dec_chan_usecnt(chan, true);

			Xfree(seq->d.ds_req->msg);
			Xfree(seq->d.ds_req);
			Xfree(seq);

		} else if (seq->type == SEQ_MD_UPDATE &&
		    seq->d.ds->chan == chan) {
			CHAIN_UNLINK(sequencer.seq, seq, link);

			dec_chan_usecnt(chan, true);

			Xfree(seq);
		}
	}

	mutex_unlock(&sequencer.lock);
}

/*
 * This routine allows for a task to be enqueued on the sequencer.
 *
 * Arguments:
 *	work_fn		- a pointer to the function to be executed
 *	args		- the arguments used by 'work_fn'
 *	synchronous	- allows the caller to specify if the routine should
 *			  wait for the enqueued function (synchronous == true)
 *			  or to return immediately (synchronous == false)
 */
void
seq_enqueue_work(seq_work_fn_t *work_fn, void *args, bool_t synchronous)
{
	seq_t 			*seq;

	seq = Xmalloc(sizeof (*seq));
	seq->type = SEQ_DO_WORK;
	seq->d.work.func = work_fn;
	seq->d.work.args = args;

	mutex_lock(&sequencer.lock);
	seq->id = sequencer.next_enqueued_id++;
	CHAIN_ADD(sequencer.seq, seq, link);
	DBG2(zprintf("%s: seqID 0x%" PRIx64 " enqueued\n", __func__, seq->id));
	(void) cond_signal(&sequencer.cond);
	mutex_unlock(&sequencer.lock);

	/* if the caller doesn't want to wait, we can bail out now */
	if (synchronous == false)
		return;

	/* wait until the sequence service routine has executed work_fn */
	mutex_lock(&sequencer.lock);
	while (seq->id > sequencer.last_processed_id)
		cond_wait(&sequencer.task_cond, &sequencer.lock);
	DBG2(zprintf("%s: seqID 0x%" PRIx64 " task done\n", __func__, seq->id));
	mutex_unlock(&sequencer.lock);
}
