/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_SEQUENCER_H
#define	_SEQUENCER_H

#ifdef __cplusplus
extern "C" {
#endif

#include <sys/types.h>
#include <time.h>
#include <thread.h>
#include <synch.h>

#include "basics.h"
#include "chain.h"
#include "cli_conn.h"
#include "ds_impl.h"
#include "hvctl.h"
#include "response.h"
#include "dr_mem.h"

#if ZDBG

extern mutex_t seq_monitor;

#define	ASSERT_IN_SEQUENCER()	ASSERT(MUTEX_HELD(&seq_monitor))
#define	PRE_SEQUENCE_ENTER()	(void) mutex_lock(&seq_monitor);
#define	PRE_SEQUENCE_EXIT()	(void) mutex_unlock(&seq_monitor);

#else

#define	ASSERT_IN_SEQUENCER()
#define	PRE_SEQUENCE_ENTER()
#define	PRE_SEQUENCE_EXIT()

#endif

typedef int seq_async_hv_cb_t(void *);
typedef void seq_work_fn_t(void *);

typedef struct seq_async_hv {
	seq_async_hv_cb_t	*cb;
	void			*msg;
} seq_async_hv_t;

typedef struct seq_work {
	seq_work_fn_t	*func;			/* ptr to routine to be run */
	void		*args;			/* args used by func */
} seq_work_t;

/*
 * Args passed between memory dr and sequencer.
 */
typedef struct {
	uint_t nmblk;		    /* nmblks to be unbound */
	dr_mblk_t *mblkp;	    /* list of mblks to be unbound */
	ldom_t *ldp;		    /* ldom */
	int rv;			    /* memory dr Results from domain */
	uint64_t size;		    /* size of requested memory dr in bytes */
} seq_mem_dr_arg_t;

/*
 * Element of queue of tasks to be handled by the sequencer.
 */
typedef struct seq {
	CHAIN_LINK(struct seq, link);		/* next in queue */
	int		type;			/* type of task */
	uint64_t	id;			/* unique ID */

	union {					/* task specific data */
		seq_async_hv_t	async_hv;
		cli_conn_t	*cli_cmd;
		ds_seq_req_t	*ds_req;
		ds_svc_t	*ds;
		bool_t		do_vcpus;
		seq_work_t	work;
	} d;
} seq_t;

/*
 * Types of sequencer task.
 */
enum {
	SEQ_ASYNC_HV_MSG = 1,		/* asynchronous HV message */
	SEQ_CLI_CMD,			/* CLI command */
	SEQ_DS_REQ,			/* domain services request */
	SEQ_XML_CMD,			/* XML command (document) */
	SEQ_PRI_REQ,			/* control channel PRI request */
	SEQ_MD_UPDATE,			/* send an MD update notification */
	SEQ_UTIL_READ,			/* Get ldom and vcpu utilization */
	SEQ_DO_WORK,			/* general purpose work */
};

typedef struct callout {
	CHAIN_LINK(struct callout, link);
	cli_conn_t	*client;
	time_t		time;
} callout_t;

#define	CLIENT_CALLOUT_TIMEOUT	30	/* in seconds */

/*
 * Sequencer task and callout queues.
 */
struct sequencer {
	mutex_t	lock;			/* protects add/remove from queue */
	cond_t	cond;			/* used to wait on queue not empty */
	cond_t	task_cond;		/* used to wait on specific tasks */
	CHAIN(seq_t, seq);		/* queue of tasks */
	CHAIN(callout_t, callouts);	/* queue of callouts */

	/*
	 * for now, these IDs are only used by seq_work_t which has support to
	 * wait until the enqueued task is completed but they could be used
	 * by the other structures if they requred that type of waiter.
	 */
	uint64_t	next_enqueued_id;	/* ID of next task enqueued */
	uint64_t	last_processed_id;	/* ID of last task processed */
};

void seq_enqueue_cmd(cli_conn_t *conn, int type);
void seq_enqueue_ds_req(ds_seq_req_t *dsreq);
void seq_enqueue_async_hv_msg(seq_async_hv_cb_t *cb, void *msg);
void seq_enqueue_md_notify(ds_svc_t *ds);
void seq_enqueue_util(bool_t do_vcpus);
void seq_enqueue_pm_req(void (*svcf)(void *), void *data);
void seq_enqueue_work(seq_work_fn_t *, void *args, bool_t synchronous);
void seq_ldom_state_change(ldom_t *ldp, cli_conn_t *client, uint64_t new_state);
void seq_delete_client_callout(cli_conn_t *client);
void seq_isolate_ldom(ldom_t *ldp, response_t *resp);
void seq_restore_ldom(ldom_t *ldp, response_t *resp);
int seq_reconf_pcpus(pm_config_op_t op, int ncpus, uint32_t *cpus,
    int *status, uint32_t flags);
void seq_update_hvmd(void);
int seq_get_unconfigured_mblks(ldom_t *ldp, void **mblkp);
void seq_update_guest_mds(void *ldp);
void seq_rm_mem_cleanup(void *ldp);
void seq_delayed_reconfig(ldom_t *ldp, del_reconfig_type_t type);
void seq_sync_mem(void *arg);
void seq_chk_add_mem(void *);
void seq_chk_del_mem(void *);
void seq_syn_mem(void *);
int seq_cancel_mem_dr(ldom_t *ldp);
void seq_dr_mem_md(void *arg);
bool_t seq_coalesce_mem(ldom_t *ldp);

#ifdef __cplusplus
}
#endif

#endif	/* _SEQUENCER_H */
