/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#include "allocate.h"
#include "basics.h"
#include "fatal.h"
#include "ldom.h"
#include "support.h"
#include "zprintf.h"
#include "backend.h"
#include "ssl_connection.h"

extern bool_t hv_is_scrub_mem_supported(void);

int
timed_read(mig_conn_t *mconn, void *buf, size_t size, int timeout)
{
	struct pollfd pfd;
	int rc, left, tout;
	int start = time(NULL);

	pfd.fd = mconn->fd;
	pfd.events = POLLIN;

	tout = timeout;
	left = size;

	while (left > 0) {
		rc = read(mconn->fd, buf, left);
		if (rc < 0) {
			if (errno == EAGAIN) {
				rc = poll(&pfd, 1, tout * MILLISEC);
				if (rc <= 0) {
					if (rc == 0) {
						DBG(zprintf("Read timeout "
						    "reached \n"));
						errno = ETIME;
					}
					return (-1);
				}
			} else {
				DBG(zprintf("read(2) error = %d\n", errno));
				return (-1);
			}
		} else {
			buf = (char *)buf + rc;
			left -= rc;
			tout = timeout - (time(NULL) - start);
			if (tout <= 0) {
				errno = ETIME;
				return (-1);
			}
		}
	}

	return (size);
}

/*
 * timeout is in seconds
 */
int
timed_SSL_read(mig_conn_t *mconn, void *buf, size_t size, int timeout)
{
	int fd = mconn->fd;
	SSL *ssl = mconn->ssl;
	struct pollfd pfd;
	int rc, rv, left, tout;
	int start = time(NULL);

	ASSERT(ssl != NULL);

	pfd.fd = fd;
	pfd.events = POLLIN;

	tout = timeout;
	left = size;

	while (left > 0) {
		rc = SSL_read_f(ssl, buf, left);
		if (rc < 0) {
			rv = SSL_get_error_f(ssl, rc);
			if ((rv == SSL_ERROR_WANT_READ) ||
			    (rv == SSL_ERROR_WANT_WRITE)) {
				rc = poll(&pfd, 1, tout * MILLISEC);
				if (rc <= 0) {
					if (rc == 0) {
						DBG(zprintf("Read timeout "
						    "reached \n"));
						errno = ETIME;
					}
					return (-1);
				}
			} else {
				DBG(zprintf("SSL_read error = %d\n",
				    SSL_get_error_f(ssl, rc)));
				return (-1);
			}
		} else if (rc == 0) {
			DBG(zprintf("READ FAILED DUE TO SHUTDOWN\n"));
			return (-1);
		} else {
			buf = (char *)buf + rc;
			left -= rc;
			tout = timeout - (time(NULL) - start);
			if (tout <= 0) {
				errno = ETIME;
				return (-1);
			}
		}
	}

	return (size);
}

int
timed_write(mig_conn_t *mconn, void *buf, size_t size, int timeout)
{
	struct pollfd pfd;
	int rc, left, tout;
	int start = time(NULL);

	pfd.fd = mconn->fd;
	pfd.events = POLLOUT;

	tout = timeout;
	left = size;

	while (left > 0) {
		rc = write(mconn->fd, buf, left);
		if (rc < 0) {
			if (errno == EAGAIN) {
				rc = poll(&pfd, 1, tout * MILLISEC);
				if (rc <= 0) {
					if (rc == 0) {
						DBG(zprintf("Write timeout "
						    "reached \n"));
						errno = ETIME;
					}
					return (-1);
				}
			} else {
				DBG(zprintf("write(2) error = %d\n", errno));
				return (-1);
			}
		} else {
			buf = (char *)buf + rc;
			left -= rc;
			tout = timeout - (time(NULL) - start);
			if (tout <= 0) {
				errno = ETIME;
				return (-1);
			}
		}
	}

	return (size);
}

/*
 * timeout is in seconds
 */
int
timed_SSL_write(mig_conn_t *mconn, void *buf, size_t size, int timeout)
{
	int fd = mconn->fd;
	SSL *ssl = mconn->ssl;
	struct pollfd pfd;
	int rc, rv, left, tout;
	int start = time(NULL);

	ASSERT(ssl != NULL);

	pfd.fd = fd;
	pfd.events = POLLOUT;

	tout = timeout;
	left = size;

	while (left > 0) {
		rc = SSL_write_f(ssl, buf, left);
		if (rc < 0) {
			rv = SSL_get_error_f(ssl, rc);
			if ((rv == SSL_ERROR_WANT_READ) ||
			    (rv == SSL_ERROR_WANT_WRITE)) {
				rc = poll(&pfd, 1, tout * MILLISEC);
				if (rc <= 0) {
					if (rc == 0) {
						DBG(zprintf("Read timeout "
						    "reached \n"));
						errno = ETIME;
					}
					return (-1);
				}
			} else {
				DBG(zprintf("SSL_write error = %d\n",
				    SSL_get_error_f(ssl, rc)));
				return (-1);
			}
		} else if (rc == 0) {
			DBG(zprintf("READ FAILED DUE TO SHUTDOWN\n"));
			return (-1);
		} else {
			buf = (char *)buf + rc;
			left -= rc;
			tout = timeout - (time(NULL) - start);
			if (tout <= 0) {
				errno = ETIME;
				return (-1);
			}
		}
	}

	return (size);
}

int
full_read(int fd, uint8_t *bufp, int size)
{
	int res;
	int left;

	for (left = size; left > 0; left -= res) {
		res = read(fd, bufp, left);
		/* return on error */
		if (res < 0)
			return (res);
		/* if zero, return what we've already read */
		if (res == 0)
			return (size - left);

		bufp += res;
	}

	return (size);
}

int
full_write(int fd, uint8_t *bufp, int size)
{
	int res;
	int left;

	for (left = size; left > 0; left -= res) {
		res = write(fd, bufp, left);
		/* return on error */
		if (res < 0)
			return (res);
		/* if zero, return what we've already written */
		if (res == 0)
			return (size - left);

		bufp += res;
	}

	return (size);
}

/*
 * Find the max and min version supported.
 */
static void
min_max_versions(int num_versions, prot_version_t *sup_versionsp,
    uint16_t *min_major, uint16_t *max_major)
{
	int i;

	*min_major = sup_versionsp[0].major_number;
	*max_major = *min_major;

	for (i = 1; i < num_versions; i++) {
		if (sup_versionsp[i].major_number < *min_major)
			*min_major = sup_versionsp[i].major_number;

		if (sup_versionsp[i].major_number > *max_major)
			*max_major = sup_versionsp[i].major_number;
	}
}

/*
 * Check if the supplied (major, minor) combination is supported.
 */
bool_t
supported_version(int num_versions, prot_version_t *sup_versionsp,
    uint16_t major, uint16_t minor)
{
	int i;

	for (i = 0; i < num_versions; i++) {
		if (sup_versionsp[i].major_number == major &&
		    sup_versionsp[i].minor_number == minor) {
			return (true);
		}
	}

	return (false);
}

/*
 * Check whether the major and minor numbers requested by the peer can be
 * satisfied. If the requested major is supported, true is returned, and the
 * agreed minor is returned in new_minor. If the requested major is not
 * supported, the routine returns false, and the closest major is returned in
 * *new_major, upon which the peer should re-negotiate. The closest major is
 * the just lower that the requested major number.
 */
bool_t
negotiate_version(int num_versions, prot_version_t *sup_versionsp,
    uint16_t req_major, uint16_t *new_majorp, uint16_t *new_minorp)
{
	int i;
	uint16_t major, lower_major;
	uint16_t min_major = 0, max_major;
	bool_t found_match = false;

	min_max_versions(num_versions, sup_versionsp, &min_major, &max_major);

	/*
	 * If the minimum version supported is greater than
	 * the version requested, return the lowest version
	 * supported
	 */
	if (min_major > req_major) {
		*new_majorp = min_major;
		return (false);
	}

	/*
	 * If the largest version supported is lower than
	 * the version requested, return the largest version
	 * supported
	 */
	if (max_major < req_major) {
		*new_majorp = max_major;
		return (false);
	}

	/*
	 * Now we know that the requested version lies between the
	 * min and max versions supported. Check if the requested
	 * major can be found in supported versions.
	 */
	lower_major = min_major;
	for (i = 0; i < num_versions; i++) {
		major = sup_versionsp[i].major_number;
		if (major == req_major) {
			found_match = true;
			*new_majorp = req_major;
			*new_minorp = sup_versionsp[i].minor_number;
			break;
		} else {
			if ((major < req_major) && (major > lower_major))
				lower_major = major;
		}
	}

	/*
	 * If no match is found, return the closest available number
	 */
	if (!found_match)
		*new_majorp = lower_major;

	return (found_match);
}

#define	BYTESPERLINE	8
#define	LINEWIDTH	((BYTESPERLINE * 3) + (BYTESPERLINE + 2) + 1)
#define	ASCIIOFFSET	((BYTESPERLINE * 3) + 2)

/*
 * Output a buffer formatted with a set number of bytes on
 * each line. Append each line with the ASCII equivalent of
 * each byte if it falls within the printable ASCII range,
 * and '.' otherwise.
 */
void
dump_msg(uint8_t *buf, int len)
{
	int	i, j;
	char	*curr;
	char	*aoff;
	char	line[LINEWIDTH];

	/* walk the buffer one line at a time */
	for (i = 0; i < len; i += BYTESPERLINE) {

		memset(line, 0, LINEWIDTH);

		curr = line;
		aoff = line + ASCIIOFFSET;

		/*
		 * Walk the bytes in the current line, storing
		 * the hex value of the byte as well as the
		 * ASCII representation in a temporary buffer.
		 * All ASCII values are placed at the end of
		 * the line.
		 */
		for (j = 0; (j < BYTESPERLINE) && ((i + j) < len); j++) {
			(void) sprintf(curr, " %02x", buf[i + j]);
			*aoff = (isprint(buf[i + j])) ? buf[i + j] : '.';
			curr += 3;
			aoff++;
		}

		/*
		 * Fill in to the start of the ASCII translation
		 * with spaces. This will only be necessary if
		 * this is the last line and there are not enough
		 * bytes to fill the whole line.
		 */
		while (curr != (line + ASCIIOFFSET))
			*curr++ = ' ';

		zprintf("%s\n", line);
	}
}

bool_t
dump_to_file(void *buf, int size, char *ftemplate, ...)
{
	char fname[256];
	char *path;
	int fd;
	va_list vargs;

	va_start(vargs, ftemplate);
	vsnprintf(fname, sizeof (fname), ftemplate, vargs);
	va_end(vargs);
	path = zsprintf("%s/%s", zeus.admdir, fname);
	DBG(zprintf("%s: writing %s\n", __func__, path));
	fd = open(path, O_CREAT|O_TRUNC|O_RDWR, 0600);
	Xfree(path);
	if (fd == -1)
		return (false);
	full_write(fd, buf, size);
	fsync(fd);
	close(fd);
	return (true);
}

md_t *
read_md_from_file(char *filename)
{
	int fh, size;
	md_t *md;
	uint8_t *buf;
	struct stat status;

	fh = open(filename, O_RDONLY);
	if (fh < 0)
		efatal("Opening %s", filename);
	if (fstat(fh, &status) < 0)
		efatal("Stat'ing %s", filename);

	buf = Xmalloc(status.st_size);
	size = full_read(fh, buf, status.st_size);
	close(fh);

	md = md_scan_bin(buf, size);
	Xfree(buf);

	if (md == NULL)
		fatal("Failed reading MD from %s", filename);

	return (md);
}

/*
 * Attempt to adjust unit strings to nearest viable value.  However care is
 * required so we do not truncate.  The returned buffer must be Xfreed after
 * use.
 */
char *
conv_size(uint64_t size)
{
	static const char factor[] = "!KMGT";
	int idx = 0;

	if (size == 0ull)
		return (Xstrdup("0"));

	while (idx < 4 && (size & 0x3ffull) == 0) {
		idx ++;
		size >>= 10;
	}

	if (idx == 0)
		return (zsprintf("%"PRId64"", size));
	else
		return (zsprintf("%"PRId64"%c", size, factor[idx]));
}

#define	KB	(1024)
#define	MB	(1024 * 1024)
#define	GB	(1024 * 1024 * 1024)

static void
generate_size_string(uint64_t bytes, char **str, size_t len)
{
	/* handle bytes */
	if (bytes < KB) {
		(void) snprintf(*str, len, "%"PRId64" Bytes", bytes);
		return;
	}

	/* handle kilobytes */
	if (bytes < MB) {
		(void) snprintf(*str, len, "%.2f KB", (double)bytes / KB);
		return;
	}

	/* handle megabytes */
	if (bytes < GB) {
		(void) snprintf(*str, len, "%.2f MB", (double)bytes / MB);
		return;
	}

	/* default is gigabytes */
	(void) snprintf(*str, len, "%.2f GB", (double)bytes / GB);
}

static void
generate_time_string(uint64_t ns, char **str, size_t len)
{
	uint64_t hrs;
	uint64_t min;
	uint64_t sec = ns / NANOSEC;
	uint64_t msec = ns / MICROSEC;
	uint64_t usec = ns / MILLISEC;

	min = sec / 60;
	hrs = min / 60;

	/* handle very small time increments */
	if (msec == 0) {
		snprintf(*str, len, "%"PRId64"us", usec);
		return;
	}

	if (sec == 0) {
		snprintf(*str, len, "%"PRId64"ms", msec);
		return;
	}

	/* handle very large time increments */
	if (min > 60) {
		min = min % 60;
		sec = sec % 60;
		snprintf(*str, len, "%"PRId64":%02"PRId64":%02"PRId64,
		    hrs, min, sec);
		return;
	}

	if (sec > 60) {
		sec = sec % 60;
		snprintf(*str, len, "%"PRId64":%02"PRId64, min, sec);
		return;
	}

	/* default is seconds */
	snprintf(*str, len, "%"PRId64"s", sec);
}

/*
 * Very basic profiling support that can be used to time various
 * sections of code. Only enabled in debug versions of Zeus.
 */
typedef struct prof {
	hrtime_t	start;
	hrtime_t	end;
} prof_t;

#if ZDBG

static void
profile_on(prof_t *prof)
{
	if (prof == NULL)
		return;

	prof->start = gethrtime();
}

static void
profile_off(prof_t *prof)
{
	hrtime_t	tmdiff;
	char		*tmstr;

	if (prof == NULL)
		return;

	prof->end = gethrtime();

	/* calculate the time difference */
	tmdiff = prof->end - prof->start;
	if (tmdiff < 0)
		tmdiff = 0;

	/* create a human readable string for the difference */
	tmstr = Xmalloc(UINT64_STR_LEN);
	generate_time_string(tmdiff, (char **)&tmstr, UINT64_STR_LEN);

	zprintf("Total elapsed time: %s\n", tmstr);

	Xfree(tmstr);
}

#define	PROFILE_ON(p)		profile_on(p)
#define	PROFILE_OFF(p)		profile_off(p)

#else /* ZDBG */

#define	PROFILE_ON(p)
#define	PROFILE_OFF(p)

#endif /* ZDBG */


#define	CHUNKLEN	(64 * 1024 * 1024)

bool_t
do_read_mem_to_file(char *file, uint64_t addr, uint64_t len)
{
	int		fd;
	int		amt;
	uint8_t		*bufp;
	uint64_t	base;
	uint64_t	chunklen;
	char		*szstr;
	prof_t		prof = { 0 };

	fd = open(file, O_RDWR|O_CREAT|O_TRUNC, 0600);
	if (fd == -1) {
		/* XXXMIG - non-debug? */
		DBG(zprintf("Failed to open \"%s\": %s\n", file,
		    strerror(errno)));
		return (false);
	}

	/*
	 * Print out the details of the requested operation.
	 */
	szstr = Xmalloc(256);
	generate_size_string(len, &szstr, 256);

	DBG(zprintf("Reading %s from address 0x%"PRIx64" to file %s\n",
	    szstr, addr, file));

	Xfree(szstr);

	bufp = (len > CHUNKLEN) ? Xmalloc(CHUNKLEN) : Xmalloc(len);
	base = addr;

	PROFILE_ON(&prof);

	while (len > 0) {
		chunklen = (len > CHUNKLEN) ? CHUNKLEN : len;

		amt = backend.read_mem(base, bufp, chunklen);
		if (amt != chunklen) {
			DBG(zprintf("Warning: read only %d\n", amt));
		}
		write(fd, bufp, amt);
		len -= amt;
		base += amt;
	}

	PROFILE_OFF(&prof);

	Xfree(bufp);

	/* cleanup */
	close(fd);

	return (true);
}

bool_t
do_write_file_to_addr(char *file, uint64_t addr)
{
	int		fh;
	struct stat	status;
	uint8_t		*bufp;
	uint64_t	len;
	int		amt;
	uint64_t	chunklen;
	char		*szstr;
	prof_t		prof = { 0 };

	/* open the input file for reading */
	fh = open(file, O_RDONLY);
	if (fh < 0) {
		DBG(zprintf("Opening %s : %s\n", file, strerror(errno)));
		return (false);
	}

	if (fstat(fh, &status) < 0) {
		DBG(zprintf("Stat'ing %s : %s\n", file, strerror(errno)));
		return (false);
	}

	len = status.st_size;

	/*
	 * Print out the details of the requested operation.
	 */
	szstr = Xmalloc(256);
	generate_size_string(len, &szstr, 256);

	DBG(zprintf("Writing file %s (%s) to address 0x%"PRIx64"\n",
	    file, szstr, addr));

	Xfree(szstr);

	bufp = (len > CHUNKLEN) ? Xmalloc(CHUNKLEN) : Xmalloc(len);

	PROFILE_ON(&prof);

	while (len > 0) {
		/* read in the next chunk */
		chunklen = (len > CHUNKLEN) ? CHUNKLEN : len;
		full_read(fh, bufp, chunklen);

		amt = backend.write_mem(addr, bufp, chunklen);
		if (amt != chunklen) {
			DBG(zprintf("Warning: amount written was %d\n", amt));
		}
		len -= amt;
		addr += amt;
	}

	PROFILE_OFF(&prof);

	close(fh);
	Xfree(bufp);

	return (true);
}

static int
int_cmp(const void *a, const void *b)
{
	if (*(int *)a < *(int *)b)
		return (-1);
	else if (*(int *)a == *(int *)b)
		return (0);
	else
		return (1);
}

void
sort_int_array(int *array, int nelem)
{
	qsort(array, nelem, sizeof (int), int_cmp);
}

/*
 * Search the HV MD for a particular guest node as specified by the
 * name string passed in as an argument. Returns NULL if the node
 * cannot be found.
 */
md_node_t *
find_guest_md_node(md_t *hvmdp, ldom_t *ldp)
{
	md_node_t	**guests;
	int		nguests;
	char		*name_prop;

	guests = Xmalloc((int)sizeof (*guests) * md_node_count(hvmdp));
	nguests = md_scan_dag(hvmdp, NULL, md_find_name(hvmdp, "guest"),
	    md_find_name(hvmdp, "fwd"), guests);

	for (int i = 0; i < nguests; i++) {

		/* check if this is the domain of interest */
		if (md_get_prop_str(hvmdp, guests[i], "name", &name_prop)) {
			DBG(zprintf("%s: guest node has no 'name' property\n",
			    __func__));
			continue;
		}

		if (strcmp(ldp->namep, name_prop) == 0)
			return (guests[i]);
	}

	Xfree(guests);
	return (NULL);
}

static int64_t
scrub_mem_write(uint64_t pa, uint64_t len)
{
	prof_t		prof = { 0 };
	uint8_t		*bufp;
	uint64_t	chunklen;
	int		amt;
	int64_t		total;
	uint64_t	base;

	DBG(zprintf("%s: clearing 0x%"PRIx64" bytes @ 0x%"PRIx64"\n",
	    __func__, len, pa));

	/* allocate a zero-filled buffer */
	bufp = Xcalloc(CHUNKLEN, uint8_t);

	total = 0;
	base = pa;

	PROFILE_ON(&prof);

	while (len > 0) {
		chunklen = (len > CHUNKLEN) ? CHUNKLEN : len;

		amt = backend.write_mem(base, bufp, chunklen);

		ASSERT(amt >= 0);

		len -= amt;
		base += amt;
		total += amt;
	}

	PROFILE_OFF(&prof);

	Xfree(bufp);

	return (total);
}

static int64_t
scrub_mem_firmware(uint64_t pa, uint64_t len)
{
	prof_t		prof = { 0 };
	int64_t		amt;
	int64_t		total;
	uint64_t	base;

	DBG(zprintf("%s: clearing 0x%"PRIx64" bytes @ 0x%"PRIx64"\n",
	    __func__, len, pa));

	total = 0;
	base = pa;

	PROFILE_ON(&prof);

	while (len > 0) {

		amt = backend.scrub_mem(base, len);

		ASSERT(amt >= 0);

		len -= amt;
		base += amt;
		total += amt;
	}

	PROFILE_OFF(&prof);

	return (total);
}

/*
 * Scrub a range of physical memory. If the firmware has native support
 * for scrubbing memory, that is the preferred method. Otherwise, fall
 * back to the less efficient method of writing zeros to the specified
 * memory range. Returns the number of bytes scrubbed on success and -1
 * on failure.
 */
int64_t
scrub_mem(uint64_t pa, uint64_t len)
{
	/*
	 * Use firmware memory scrub if it is available. Support
	 * was introduced in HVCTL version 1.4.
	 */
	if (hv_is_scrub_mem_supported())
		return (scrub_mem_firmware(pa, len));

	return (scrub_mem_write(pa, len));
}

void
dump_md_to_file(md_t *md, char *filename)
{
	int size;
	uint8_t *mdbuf;

	size = md_gen_bin(md, &mdbuf);
	dump_to_file(mdbuf, size, "dump-%s.bin", filename);
	free(mdbuf);
}

md_t *
get_md_from_file(char *mdname)
{
	md_t *md = NULL;
	int	fd;
	char path[256];

	snprintf(path, sizeof (path), "%s/%s", zeus.admdir, mdname);
	fd = open(path, O_RDONLY);
	if (fd != -1) {
		close(fd);
		md = read_md_from_file(path);
		DBG(zprintf("Reading %s MD from file %s\n", mdname, path));
	}
	return (md);
}
