/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <stdio.h>
#include <alloca.h>
#include <sys/types.h>

#include "basics.h"
#include "fatal.h"
#include "allocate.h"
#include "database.h"
#include "keys.h"
#include "variables.h"
#include "ldom.h"
#include "mdesc.h"
#include "mdesc_impl.h"
#include "zpri.h"
#include "backend.h"
#include "option.h"
#include "ds.h"
#include "ds_impl.h"
#include "zprintf.h"
#include "mac_addrs.h"
#include "frags.h"
#include "vcpu.h"
#include "vmau.h"
#include "ldcs.h"
#include "vio.h"
#include "core_md.h"
#include "master.h"
#include "uuid.h"

static void	niagara1_init(void);
static bool_t	niagara1_process_option(int *ip, int argc, char **argv);
static void	niagara1_complete_setup(void);
static md_t	*niagara1_gen_hvmd(bool_t include_pm_status);
static void	niagara1_parse_hvmd(md_t *);
static void	niagara1_fini(void);
static ldom_t	*grab_basic_guest_state(md_t *mdp, int, md_node_t *guestp);
static void	grab_frag_mblock(md_t *mdp, md_node_t *cpup);
static void	grab_pio_state(ldom_t *ldp, md_t *guest_mdp,
    md_node_t *hv_guestp, md_t *hvmdp);
static void	grab_platform_sp_channels(md_t *hvmdp);
static void	gen_hvmd_sp_channels(md_t *mdp, md_node_t *ldc_endsp);
static void	grab_guest_dependencies(md_t *mdp, md_node_t *guestp,
    ldom_t *ldp);
static void	get_pcie_props(md_t *hvmdp, md_node_t **hvnodes, int num,
    uint64_t cfg_hdl, uint64_t *resid, bound_iodevice_t *bio);

target_t niagara1_tgt = {
	"n1",
	"Niagara1 hypervisor",
	niagara1_init,
	niagara1_process_option,
	niagara1_complete_setup,
	niagara1_gen_hvmd,
	niagara1_parse_hvmd,
	niagara1_fini,
	TARGET_MAGIC
};

/*
 * Holds SP->guest|HV LDC endpoints read out of HV MD.  We save this info
 * (configured by the SP) separately, since we have no corresponding VIO
 * server class, and use it to regenerate these endpoints when creating a new
 * HV MD.
 */
CHAIN(sp_guest_ldc_t, sp_ldc_chan_list);

CHAIN(ldc_epdb_t, ldc_epdb);

typedef struct hio_ent hio_ent_t;
struct hio_ent {
    md_node_t *hio;
    md_node_t *guest;
    CHAIN_LINK(hio_ent_t, next);
};

static CHAIN(hio_ent_t, hio_entries);


static void initialize_ldcdb(md_t *hv_mdp, md_node_t **guestlistp,
    int numguests);
static void cleanup_ldcdb(void);

static void
niagara1_init(void)
{
	DBG(zprintf("%s\n", __func__));
	CHAIN_INIT(sp_ldc_chan_list);
}

/*
 * Target specific command line options.
 */
/*ARGSUSED0*/
static bool_t
niagara1_process_option(int *ip, int argc, char **argv)
{
	return (false);
}


static void
niagara1_complete_setup(void)
{
	DBG(zprintf("%s\n", __func__));

	/*
	 * Allocate the res_ids for all the HV top-level resource structures
	 */
	create_res_pool(&zeus.free_ldom_ids, 0, pri.max_guests-1, Res_sorted);
	create_res_pool(&zeus.free_vcpu_ids, 0, pri.max_vcpus-1, Res_sorted);
	create_res_pool(&zeus.free_mau_ids, 0, pri.num_maus-1, Res_sorted);
	create_res_pool(&zeus.free_guest_ids, 0, pri.max_guests-1, Res_sorted);
	create_res_pool(&zeus.free_console_ids, 0, pri.max_guests-1,
	    Res_sorted);
	create_res_pool(&zeus.free_ldc_ids, 0,
	    (pri.max_guests * pri.max_guest_ldcs) - 1, Res_sorted);
	create_res_pool(&zeus.free_mblk_ids, 0, pri.max_mblocks-1, Res_none);
	/* FIXME: these values should come from the PRI */
	create_res_pool(&zeus.free_pcidev_ids, 0, 1023, Res_sorted);
	create_res_pool(&zeus.free_netdev_ids, 0, 1023, Res_sorted);
	create_res_pool(&zeus.free_sp_ldcs, 0, 255, Res_sorted);
	/*
	 * This pool currently contains a single id for the hvctl HV endpoint.
	 * The pool will be eliminated once "guest0" support is implemented.
	 */
	create_res_pool(&zeus.free_hv_ldc_ids, 0, 0, Res_sorted);

	/* debug aids */
	name_res_pool(&zeus.free_ldom_ids, "free_ldom_ids");
	name_res_pool(&zeus.free_vcpu_ids, "free_vcpu_ids");
	name_res_pool(&zeus.free_mau_ids, "free_mau_ids");
	name_res_pool(&zeus.free_guest_ids, "free_guest_ids");
	name_res_pool(&zeus.free_console_ids, "free_console_ids");
	name_res_pool(&zeus.free_ldc_ids, "free_ldc_ids");
	name_res_pool(&zeus.free_mblk_ids, "free_mblk_ids");
	name_res_pool(&zeus.free_sp_ldcs, "free_sp_ldcs");
	name_res_pool(&zeus.free_hv_ldc_ids, "free_hv_ldc_ids");
}

static void
niagara1_fini(void)
{
	DBG(zprintf("%s\n", __func__));
}

static md_t *
niagara1_gen_hvmd(bool_t include_pm_status)
{
	md_t			*mdp;
	md_node_t		*rootp;
	md_node_t		*guestsp;
	md_node_t		*cpusp;
	md_node_t		*memsp;
	md_node_t		*mausp;
	md_node_t		*devices;
	md_node_t		*ldc_endpointsp;
	md_node_t		*spacep;
	md_node_t		*node;
	ldom_t			*ldp;
	mblock_t		*bmbp;
	bound_iodevice_t	*bio;
	uint64_t		hvmd_version;
	md_node_t		**gnodes;
	master_t		*master;
	hio_ent_t		*hiop;
	hio_ent_t		*hp;
	hio_ent_t		*tmp_hp;
	int			nbldp;
	int			i = 0;

	DBG(zprintf("%s\n", __func__));

	mdp = md_new_md();
	rootp = md_new_node(mdp, "root");

	/*
	 * Set content-version
	 */
	hvmd_version = (zeus.hvmd_vers_major << 32) | zeus.hvmd_vers_minor;
	/*
	 * If we need to generate DIO version of HVMD and if it is
	 * an old format HVMD, then we need update the version
	 * number accordingly.
	 */
	if (pcie_gen_dio_hvmd() && !pcie_is_dio_hvmd_supported()) {
		hvmd_version = (HVMD_VERSION_MAJOR_NUMBER << 32) |
		    HVMD_VERSION_MINOR_NUMBER;
	}
	md_add_value_property(mdp, rootp, "content-version", hvmd_version);

	/*
	 * Set prop that says if this is the same as the
	 * selected boot set on SP.
	 */
	md_add_value_property(mdp, rootp, "is_next_spconfig",
	    zeus.is_next_spconfig);

	/* set name of current spconfig, if known */
	if (zeus.current_spconfig != NULL)
		md_add_string_property(mdp, rootp, "current_spconfig",
		    zeus.current_spconfig);

	if (zeus.hvmd_vers_major == 0 ||
	    (zeus.hvmd_vers_major == 1 && zeus.hvmd_vers_minor < 1)) {
		/*
		 * Legacy support for HV properties that have since been moved
		 * to the mini MD delivered from the SP to HV.
		 */
		md_add_value_property(mdp, rootp, "stick-frequency",
		    pri.stick_frequency);

		if (pri.tod_frequency > 0) {
			md_add_value_property(mdp, rootp, "tod-frequency",
			    pri.tod_frequency);
		}
		if (pri.tod > 0) {
			md_add_value_property(mdp, rootp, "tod", pri.tod);
		}
		if (pri.erpt_pa > 0) {
			md_add_value_property(mdp, rootp, "erpt-pa",
			    pri.erpt_pa);
		}
		if (pri.erpt_size > 0) {
			md_add_value_property(mdp, rootp, "erpt-size",
			    pri.erpt_size);
		}

		/* add platform node */
		node = md_link_new_node(mdp, "platform", rootp, "fwd", "back");
		md_add_value_property(mdp, node, "stick-frequency",
		    pri.stick_frequency);
	}

	/* Create a node for the guests */
	guestsp = md_link_new_node(mdp, "guests", rootp, "fwd", "back");

	/* Create a node for the cpus */
	cpusp = md_link_new_node(mdp, "cpus", rootp, "fwd", "back");

	/* Create a node for the mem info */
	memsp = md_link_new_node(mdp, "memory", rootp, "fwd", "back");

	/* Create new nodes for the maus and cwqs (if any) */
	mausp = md_link_new_node(mdp, "maus", rootp, "fwd", "back");
	if (pri.have_cwq)
		(void) md_link_new_node(mdp, "cwqs", rootp, "fwd", "back");

	/* Create a node for IO devices */
	devices = md_link_new_node(mdp, "devices", rootp, "fwd", "back");

	/* create the LDC endpoint aggregator node */
	ldc_endpointsp = md_link_new_node(mdp, "ldc_endpoints",
	    rootp, "fwd", "back");

	/* create the console aggregator node */
	md_link_new_node(mdp, "consoles", rootp, "fwd", "back");

	gnodes = Xcalloc(CHAIN_LENGTH(zeus.ldoms), md_node_t *);

	CHAIN_INIT(hio_entries);

	/*
	 * For each bound LDom create a guest instance.
	 */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		md_node_t	*guestp;
		md_node_t	*dependency;
		bound_mfrag_t	*fragp;

		DBG2(zprintf("LDom %s state is %s\n", ldp->namep,
		    ldom_state_str(ldp->state)));

		if (ldp->state == INACTIVE || ldp->state == UNBINDING)
			continue;

		gnodes[i++] = guestp =
		    md_link_new_node(mdp, "guest", guestsp, "fwd", "back");

		md_add_string_property(mdp, guestp, "name", ldp->namep);

		/* global ID */
		md_add_value_property(mdp, guestp, "gid", ldp->idp->id);

		/* HV Index */
		md_add_value_property(mdp, guestp, "resource_id",
		    ldp->res_id->id);

		/* partition ID ( 0 is reserved for HV ) */
		md_add_value_property(mdp, guestp, "pid", 1 + ldp->idp->id);

		md_add_value_property(mdp, guestp, "tod-offset",
		    ldp->props.tod_offset);

		md_add_value_property(mdp, guestp, "reset-reason", 0);

		md_add_value_property(mdp, guestp, "rngctlaccessible",
		    ldp->props.rngctlaccessible);

#if ZDBG
		/* Always enable access for debug builds */
		md_add_value_property(mdp, guestp, "diagpriv", 1);
		md_add_value_property(mdp, guestp, "perfctraccess", 1);
		md_add_value_property(mdp, guestp, "perfctrhtaccess", 1);
#else
		/*
		 * All guests get the same value for diagpriv as set in the
		 * control domain.
		 */
		md_add_value_property(mdp, guestp, "diagpriv",
		    zeus.hvctl_ldom_p->props.diagpriv);

		md_add_value_property(mdp, guestp, "perfctraccess",
		    ldp->props.perfctraccess);

		md_add_value_property(mdp, guestp, "perfctrhtaccess",
		    ldp->props.perfctrhtaccess);

#endif
		md_add_value_property(mdp, guestp, "ldc_mapin_base",
		    ldp->props.ldc_mapin_base);

		md_add_value_property(mdp, guestp, "ldc_mapin_size",
		    ldp->props.ldc_mapin_size);

		/*
		 * Add virtual and channel devices nodes.
		 */
		node = md_link_new_node(mdp, "virtual_devices", guestp,
		    "fwd", "back");
		md_add_value_property(mdp, node, "cfghandle",
		    VDEVS_CFG_HANDLE);

		node = md_link_new_node(mdp, "channel_devices", guestp,
		    "fwd", "back");
		md_add_value_property(mdp, node, "cfghandle",
		    CHANNELS_CFG_HANDLE);

		/*
		 * Add properties to denote owned IO bridges.  This is done
		 * in two passes since some targets of "back" links may
		 * not have been created yet in the first pass (see
		 * add_pcie_props2 call below).
		 */
		CHAIN_ITER(ldp->bound.iodevices, bio, ldom) {
			md_node_t *iodev_node;
			md_node_t *hydev_node = NULL;
			char *pcie_node_name = (bio->virtual_pcie_bus) ?
			    "virtual_pcie_bus" : "pcie_bus";

			if (bio->phys_iodev->is_pciex) {
				iodev_node = md_link_new_node(mdp,
				    pcie_node_name, guestp, "fwd", "back");
				bio->hvmd_iodev = iodev_node;

				if (pcie_gen_dio_hvmd())
					add_pcie_props1(mdp,
					    ldp, iodev_node, bio);
			} else  {
				hiop = Xcalloc(1, hio_ent_t);
				/*
				 * FIXME: for now everything which is not of
				 * type "pciex" is assumed to be of type
				 * "network_device".
				 */
				iodev_node = md_link_new_node(mdp,
				    "network_device", guestp, "fwd", "back");
				hydev_node = md_link_new_node(mdp,
				    "hybrid_network_device", devices, "fwd",
				    "back");
				/*
				 * Add the properties to the hybrid IO nodes
				 */
				md_add_value_property(mdp, hydev_node,
				    "resource_id", bio->id->id);
				md_add_value_property(mdp, hydev_node,
				    "cfghandle", bio->phys_iodev->cfg_handle);

				hiop->hio = hydev_node;
				hiop->guest = guestp;
				CHAIN_ADD(hio_entries, hiop, next);
			}

			/* Link in to top-level devices node */
			md_add_arc_property(mdp, devices, "fwd", iodev_node);
			md_add_arc_property(mdp, iodev_node, "back", devices);

			md_add_value_property(mdp, iodev_node, "resource_id",
			    bio->id->id);

			md_add_value_property(mdp, iodev_node, "cfghandle",
			    bio->phys_iodev->cfg_handle);

			if (bio->phys_iodev->rcid != ILLEGAL_RCID) {
				md_add_value_property(mdp, iodev_node, "rcid",
				    bio->phys_iodev->rcid);
			}

			/* FIXME: temporary */
			md_add_value_property(mdp, iodev_node, "gid",
			    ldp->idp->id);

			/*
			 * If domaining_enabled is false, base bypass
			 * mode on requirement, not binding.  This allows
			 * us to enforce our own default behavior of no
			 * bypass mode.
			 */
			if ((zeus.domaining_enabled == false &&
			    bio->req_iop->io_mmu_bypass) ||
			    (zeus.domaining_enabled == true &&
			    bio->io_mmu_bypass)) {
				md_add_value_property(mdp, iodev_node,
				    "allow_bypass", 1);
			}

		}

		/*
		 * Create nodes for all the bound maus. The links from the cpu
		 * nodes will be created when creating nodes for the bound
		 * cpus.
		 */
		mau_gen_hvmd_nodes(ldp, mdp, mausp, guestp);

		cpu_gen_hvmd_nodes(ldp, mdp, cpusp, guestp, include_pm_status);

		mem_gen_hvmd_nodes(ldp, mdp, memsp, guestp);

		/*
		 * For each LDom the location of the MD depends whether a new
		 * MD is being generated and downloaded - relevant to this HV
		 * MD. For this we look to the gen_new_md flag.
		 */
		fragp = ldp->flag.gen_new_md ?
		    ldp->frags.next_md_frag : ldp->frags.committed_md_frag;
		md_add_value_property(mdp, guestp, "mdpa", fragp->paddr);

		/* add ldc_endpoints, subordinate nodes and other props */
		vio_gen_ldc_endpoint_nodes(ldp, mdp, guestp, ldc_endpointsp);
		backend.aug_hv_guest(ldp, guestp, mdp);

		/* Add guestcontrol */
		md_add_value_property(mdp, guestp, HVMD_NODE_GUESTCONTROL,
		    ldp->props.failure_policy);

		/* Add guestdependency */
		CHAIN_ITER(ldp->props.masters, master, link) {

			if (!is_valid_master(master))
				continue;

			/* create the new guestdependency node */
			dependency = md_link_new_node(mdp,
			    HVMD_NODE_GUESTDEPENDENCY, guestp,
			    "fwd", "back");

			/* add the guest ID property to the dependency */
			md_add_value_property(mdp, dependency,
			    HVMD_PROP_DEPENDENCY_GID,
			    master->ldp->res_id->id);
		}
	}
	nbldp = i; /* Number of bound ldoms */

	/*
	 * Walk through all the bound iodevices (pass 2), this time creating
	 * and linking the "assigned devices" node and creating links to
	 * its child "pcie device" nodes.
	 */
	if (pcie_gen_dio_hvmd()) {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			if (ldp->state == INACTIVE || ldp->state == UNBINDING)
				continue;
			CHAIN_ITER(ldp->bound.iodevices, bio, ldom) {
				if (bio->phys_iodev->is_pciex)
					add_pcie_props2(mdp,
					    ldp, bio->hvmd_iodev, bio);
			}
		}
	}

	/*
	 * For each "hybrid_network_device" in "devices" node, create
	 * a 'fwd' link from each guest node which does not have the
	 * "network_device" node
	 */
	CHAIN_ITER(hio_entries, hiop, next) {
		for (i = 0; i < nbldp; i++) {
			if (gnodes[i] != hiop->guest) {
				md_add_arc_property(mdp, gnodes[i], "fwd",
				    hiop->hio);
				md_add_arc_property(mdp, hiop->hio, "back",
				    gnodes[i]);
			}
		}
	}

	/*
	 * Remove hybrid io list
	 */
	CHAIN_ITER_VOL(hio_entries, hp, tmp_hp, next) {
		CHAIN_UNLINK(hio_entries, hp, next);
		Xfree(hp);
	}

	Xfree(gnodes);

	/*
	 * The frag for the HV MD must be allocated before the HV MD is
	 * generated because the HV MD must include the frag mblock containing
	 * the frag. We really should do away with having the frag mblocks in
	 * the HV MD.
	 */
	if (!in_delayed_reconfig(NULL)) {
		ASSERT(zeus.next_hvmd_fp == NULL);
		zeus.next_hvmd_fp = hv_md_alloc();
	} else {
		if (zeus.next_hvmd_fp == NULL) {
			zeus.next_hvmd_fp = hv_md_alloc();
		} else {
			/*
			 * We must prevent the previous HV MD
			 * from being overwritten, in case HV performs
			 * a rollback to the previous pending state
			 * (i.e. on a parse failure during delayed reconfig).
			 */
			if (zeus.prev_del_reconf_hvmd_fp)
				hv_md_free(zeus.prev_del_reconf_hvmd_fp);
			zeus.prev_del_reconf_hvmd_fp = zeus.next_hvmd_fp;
			zeus.next_hvmd_fp = hv_md_alloc();
		}
	}

	/*
	 * FIXME: Isn't there a better way ?
	 * Stash in the HV MD the list of areas we use for
	 * fragment allocation. The HV doesn't use this, but
	 * it will help Zeus figure out which memory is free
	 * and which is in used.
	 * FIXME: Maybe we should list all the free frags ?
	 */
	spacep = md_link_new_node(mdp, "frag_space", rootp, "fwd", "back");

	md_add_value_property(mdp, spacep, "fragsize", FRAG_SIZE);

	CHAIN_ITER(zeus.frag_mblocks, bmbp, ldom) {
		md_node_t	*fragp;
		fragp = md_link_new_node(mdp, "frag_mblock", spacep,
		    "fwd", "back");

		md_add_value_property(mdp, fragp, "base", bmbp->paddr);
		md_add_value_property(mdp, fragp, "size", bmbp->size);
	}

	/* generate any needed rsvd_mblock nodes */
	if (CHAIN_LENGTH(zeus.rsvd_mblocks) > 0) {
		spacep = md_link_new_node(mdp, "rsvd_space", rootp, "fwd",
		    "back");

		CHAIN_ITER(zeus.rsvd_mblocks, bmbp, ldom) {
			md_node_t	*rsvdp;
			rsvdp = md_link_new_node(mdp, "rsvd_mblock", spacep,
			    "fwd", "back");

			md_add_value_property(mdp, rsvdp, "base", bmbp->paddr);
			md_add_value_property(mdp, rsvdp, "size", bmbp->size);
		}
	}

	hv_md_mblock_gen_node(mdp);

	/*
	 * Generate any lingering SP channel nodes.
	 */
	if (CHAIN_LENGTH(sp_ldc_chan_list) > 0)
		gen_hvmd_sp_channels(mdp, ldc_endpointsp);

	return (mdp);
}

static void
gen_hvmd_sp_channels(md_t *mdp, md_node_t *ldc_endsp)
{
	sp_guest_ldc_t	*sp_chan_p;
	md_node_t	*ldc_endpoint;

	CHAIN_ITER(sp_ldc_chan_list, sp_chan_p, link) {
		ldc_endpoint = md_link_new_node(mdp, "ldc_endpoint",
		    ldc_endsp, "fwd", "back");

		md_add_value_property(mdp, ldc_endpoint, "target_type",
		    sp_chan_p->target_type);
		md_add_value_property(mdp, ldc_endpoint, "channel",
		    sp_chan_p->channel->chan->id);
		md_add_value_property(mdp, ldc_endpoint, "resource_id",
		    sp_chan_p->channel->id->id);
		/* SP->HV channels have no target_guest */
		if (sp_chan_p->target_type == TYPE_GUEST)
			md_add_value_property(mdp, ldc_endpoint, "target_guest",
			    sp_chan_p->target_guest);
		md_add_value_property(mdp, ldc_endpoint, "target_channel",
		    sp_chan_p->target_channel);
	}
}

/*
 * By the time this function gets called we have looked at all the guest MDs
 * and instantiated all the client nodes, and the server nodes with their
 * ports.
 *
 * However all they are not connected to each other, and for that we must go
 * back to the HV MD and using the endpoint info there, discover both client
 * and server and stitch the client_binding data structures together.
 *
 * We do all this in this function.
 *
 * FIXME: The algorithm is simple and brute force.
 * 	We can do better, but this is functionally correct at least
 */

ldc_epdb_t *
stitch_ldc_epdb_pair(uint64_t gid, uint64_t ldc,
    uint64_t peer_gid, uint64_t peer_ldc)
{
	ldc_epdb_t *ep, *peer_ep;

	/*
	 * Check if local endpoint is already present.  If so just increment
	 * the counts. It is assumed that the HV MD is correct so an endpoint
	 * will only ever be part of a single pair.
	 */
	CHAIN_ITER(ldc_epdb, ep, link) {
		if (ep->gid == gid && ep->ldc == ldc) {
			ep->count++;
			ep->peer->count++;

			ASSERT(ep->peer->gid == peer_gid);
			ASSERT(ep->peer->ldc == peer_ldc);

			return (ep);
		}
	}

	/*
	 * Not found, create the pair and cross-link.
	 */

	ep = Xcalloc(1, ldc_epdb_t);
	ep->gid = gid;
	ep->ldc = ldc;
	ep->ldp = NULL;
	ep->id = NULL;
	ep->count = 1;

	peer_ep = Xcalloc(1, ldc_epdb_t);
	peer_ep->gid = peer_gid;
	peer_ep->ldc = peer_ldc;
	peer_ep->ldp = NULL;
	peer_ep->id = NULL;
	peer_ep->count = 1;

	ep->peer = peer_ep;
	peer_ep->peer = ep;

	CHAIN_ADD(ldc_epdb, ep, link);
	CHAIN_ADD(ldc_epdb, peer_ep, link);

	return (ep);
}

static void
new_sp_endpoint(uint64_t channel, uint64_t target_type, uint64_t target_guest,
    uint64_t target_channel, uint64_t res_id)
{
	sp_guest_ldc_t *sp_chan_p;

	sp_chan_p = Xcalloc(1, sp_guest_ldc_t);
	CHAIN_LINK_INIT(sp_chan_p->link);

	sp_chan_p->channel = Xcalloc(1, ldc_chan_t);
	CHAIN_LINK_INIT(sp_chan_p->channel->link);
	sp_chan_p->channel->chan = pull_res_id(&zeus.free_sp_ldcs, channel);
	if (sp_chan_p->channel->chan == NULL)
		fatal("SP channel # 0x%"PRIx64" not available", channel);

	if (res_id != -1) {
		sp_chan_p->channel->id =
		    pull_res_id(&zeus.free_ldc_ids, res_id);
		if (sp_chan_p->channel->id == NULL)
			fatal("SP resource ID %"PRId64" not available", res_id);
	} else {
		sp_chan_p->channel->id = alloc_res(&zeus.free_ldc_ids);
		ASSERT(sp_chan_p->channel->id != NULL);
	}

	sp_chan_p->target_type = target_type;
	sp_chan_p->target_guest = target_guest;
	sp_chan_p->target_channel = target_channel;

	DBG(zprintf(" %"PRId64"->%"PRId64"@%"PRId64,
	    channel, target_channel, target_guest));

	CHAIN_ADD(sp_ldc_chan_list, sp_chan_p, link);
}

sp_guest_ldc_t *
sp_guest_ldc_find_by_id(uint64_t ldcid)
{
	sp_guest_ldc_t *sp_chan;

	CHAIN_ITER(sp_ldc_chan_list, sp_chan, link) {
		if (sp_chan->channel->chan->id == ldcid)
			break;
	}

	return (sp_chan);
}

sp_guest_ldc_t *
sp_guest_ldc_find_by_target(uint64_t target_guest, uint64_t target_channel)
{
	sp_guest_ldc_t *sp_chan;

	CHAIN_ITER(sp_ldc_chan_list, sp_chan, link) {
		if ((sp_chan->target_guest == target_guest) &&
		    (sp_chan->target_channel == target_channel))
			break;
	}

	return (sp_chan);
}

static void
add_pri_ldc_endpoints(void)
{
	md_node_t **nodes, *node;
	ldc_epdb_t *ep;
	uint64_t channel, target_channel, target_type, target_guest;
	int i, num;

	DBG(zprintf_lock());
	DBG(zprintf("%s:", __func__));

	nodes = Xmalloc(md_node_count(pri.pri) * sizeof (md_node_t *));

	num = md_scan_dag(pri.pri, NULL, md_find_name(pri.pri, "ldc_endpoint"),
	    md_find_name(pri.pri, "fwd"), nodes);

	for (i = 0; i < num; i++) {
		node = nodes[i];

		if (md_get_prop_val(pri.pri, node, "channel", &channel) < 0)
			fatal("No channel in endpoint");

		if (md_get_prop_val(pri.pri, node, "target_channel",
		    &target_channel) < 0)
			fatal("No target_channel in endpoint");

		if (md_get_prop_val(pri.pri, node, "target_type",
		    &target_type) < 0)
			fatal("No target_type in endpoint");

		if (target_type == TYPE_SP) {

			/* add the connection to the DB */
			ep = stitch_ldc_epdb_pair(zeus.hvctl_ldom_p->idp->id,
			    channel, LDCDB_SPGID, target_channel);

			ep->ldp = zeus.hvctl_ldom_p;
			ep->id = alloc_res(&zeus.free_ldc_ids);

		} else if (target_type == TYPE_GUEST) {

			if (md_get_prop_val(pri.pri, node, "target_guest",
			    &target_guest) < 0) {
				fatal("No target_guest in endpoint");
			}
			ASSERT(target_guest == zeus.hvctl_ldom_p->idp->id);

			new_sp_endpoint(channel, target_type, target_guest,
			    target_channel, -1);

			(void) stitch_ldc_epdb_pair(LDCDB_SPGID, channel,
			    target_guest, target_channel);
		}
	}

	Xfree(nodes);
	DBG(zprintf("\n"));
	DBG(zprintf_unlock());
}

static void
initialize_ldcdb(md_t *hv_mdp, md_node_t **guestlistp, int numguests)
{
	int idx;
	md_node_t **eplistp;
	ldc_epdb_t *ep;

	CHAIN_INIT(ldc_epdb);
	eplistp = Xmalloc(md_node_count(hv_mdp) * sizeof (eplistp[0]));

	for (idx = 0; idx < numguests; idx++) {
		md_node_t *guest_np;
		ldom_t *ldp;
		int i, num;
		uint64_t local_gid;

		/* Find the LDom relevant to this HV MD guest */
		guest_np = guestlistp[idx];
		if (md_get_prop_val(hv_mdp, guest_np, "gid", &local_gid) < 0) {
			warning("Guest node #%d in HV MD has no gid", idx);
			continue;
		}

		/* Find the LDom then */
		ldp = find_ldom_by_gid(local_gid);
		ASSERT(ldp != NULL);	/* should have an LDom by this point */

		num = md_scan_dag(hv_mdp, guest_np,
		    md_find_name(hv_mdp, "ldc_endpoint"),
		    md_find_name(hv_mdp, "fwd"), eplistp);

		for (i = 0; i < num; i++) {
			uint64_t local_ldc;
			uint64_t res_id;
			uint64_t remote_ldc;
			uint64_t remote_gid;
			ldc_epdb_t *ep;
			uint64_t type;
			md_node_t *mdep;

			mdep = eplistp[i];

			if (md_get_prop_val(hv_mdp, mdep, "channel",
			    &local_ldc) < 0) {
				fatal("No channel in endpoint");
			}

			if (md_get_prop_val(hv_mdp, mdep, "resource_id",
			    &res_id) < 0) {
				fatal("No res_id in endpoint");
			}

			/* OK what's it connected to ? */
			if (md_get_prop_val(hv_mdp, mdep, "target_type",
			    &type) < 0) {
				fatal("No target_type in endpoint");
			}
			if (md_get_prop_val(hv_mdp, mdep, "target_channel",
			    &remote_ldc) < 0) {
				fatal("No target_channel in endpoint");
			}

			switch (type) {
			case TYPE_GUEST:
				if (md_get_prop_val(hv_mdp, mdep,
				    "target_guest", &remote_gid) < 0) {
					fatal("No target_guest in endpoint");
				}
				break;
			case TYPE_HV:
				remote_gid = LDCDB_HVGID;
				break;
			case TYPE_SP:
				remote_gid = LDCDB_SPGID;
				break;
			default:
				fatal("Illegal type 0x%"PRIx64" in LDC "
				    "endpoint", type);
			}

			/* Add the connection to the DB */
			ep = stitch_ldc_epdb_pair(local_gid, local_ldc,
			    remote_gid, remote_ldc);
			if (ep->ldp == NULL) {
				ep->ldp = ldp;
				ep->id = pull_res_id(&zeus.free_ldc_ids,
				    res_id);
			} else {
				ASSERT(ep->ldp == ldp);
			}
		}
	}

	/*
	 * Quick sanity check. Check that each pair has a legitimate domain,
	 * HV or SP endpoint.
	 */

	CHAIN_ITER(ldc_epdb, ep, link) {
		ASSERT((ep->gid == LDCDB_HVGID && ep->ldp == NULL) ||
		    (ep->gid == LDCDB_SPGID && ep->ldp == NULL) ||
		    (ep->ldp != NULL));
	}

	Xfree(eplistp);
}

static void
dump_ldcdb(void)
{
	ldc_epdb_t *ep, *peer_ep;
	char gid[8], peer_gid[8];

	if (CHAIN_LENGTH(ldc_epdb) > 0)
		zprintf("GID    NAME   LDCID        GID    NAME   LDCID\n");

	CHAIN_ITER(ldc_epdb, ep, link) {
		peer_ep = ep->peer;

		if (ep->gid == LDCDB_HVGID)
			(void) strcpy(gid, "HV");
		else if (ep->gid == LDCDB_SPGID)
			(void) strcpy(gid, "SP");
		else
			(void) snprintf(gid, sizeof (gid), "%"PRId64, ep->gid);

		if (peer_ep->gid == LDCDB_HVGID)
			(void) strcpy(peer_gid, "HV");
		else if (peer_ep->gid == LDCDB_SPGID)
			(void) strcpy(peer_gid, "SP");
		else
			(void) snprintf(peer_gid, sizeof (gid), "%"PRId64,
			    peer_ep->gid);

		zprintf("(%2s, %9s, %2"PRId64") [%d] -> "
		    "(%2s, %9s, %2"PRId64") [%d]\n",
		    gid, (ep->ldp != NULL) ? ep->ldp->namep: "",
		    ep->ldc, ep->count, peer_gid,
		    (peer_ep->ldp != NULL) ? peer_ep->ldp->namep: "",
		    peer_ep->ldc, peer_ep->count);
	}
}

static void
cleanup_ldcdb(void)
{
	ldc_epdb_t *ep;
	ldc_epdb_t *nextep;

	for (ep = CHAIN_START(ldc_epdb); ep != NULL; ep = nextep) {
		nextep = CHAIN_NEXT(ep->link);
		/*
		 * Release all unclaimed guest endpoints from the database.
		 * Unclaimed guest endpoints should only be present in a
		 * migration recovery scenario. The recovery process does
		 * not use the database but rather grabs LDC IDs directly.
		 */
		if (LDCDB_IS_GUESTGID(ep->peer->gid) && (ep->id != NULL)) {
			DBG5(zprintf("releasing resid from ldcdb: gid=%"PRId64
			    ", ldc=%"PRId64", id=%"PRId64"\n", ep->gid,
			    ep->ldc, ep->id->id));
			free_res(ep->id);
		}
		Xfree(ep);
	}
}

ldc_epdb_t *
ldc_epdb_find(int gid, int ldc)
{
	ldc_epdb_t *ep;

	CHAIN_ITER(ldc_epdb, ep, link) {
		if (ep->gid == gid && ep->ldc == ldc)
			break;
	}

	return (ep);
}

static uint64_t
grab_guest_mem_generation_id(md_t *mdp)
{
	md_prop_t *propp;
	md_string_t *gnamep;
	md_node_t **nodelistp;
	int num;

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));
	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "memory"),
	    md_find_name(mdp, "fwd"), nodelistp);
	if (num > 0) {
		gnamep = md_find_name(mdp, "memory-generation-id#");
		if (gnamep != NULL) {
			propp = md_find_node_prop(mdp, nodelistp[0],
			    gnamep, MDET_PROP_VAL);
			if (propp != NULL) {
				Xfree(nodelistp);
				return (propp->d.value);
			}
		}
	}

	Xfree(nodelistp);

	return (0);
}

static void
check_if_domaining_enabled(md_t *mdp)
{
	uint64_t domaining_enabled = 0;
	md_node_t **nodelistp;
	int num;

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "platform"),
	    md_find_name(mdp, "fwd"), nodelistp);
	if (num > 0)
		md_get_prop_val(mdp, nodelistp[0], "domaining-enabled",
		    &domaining_enabled);

	if (domaining_enabled != 0)
		zeus.domaining_enabled = true;

#if CONFIG_MODE
	if (domaining_enabled == 0 && option.mgmt_mode == false)
		zeus.mode = ZEUS_CONFIG_MODE;
#endif

	if (zeus.mode == ZEUS_CONFIG_MODE)
		zprintf("Notice: LDom Manager is running in configuration "
		    "mode\n");

	Xfree(nodelistp);
}

static void
grab_devaliases(ldom_t *ldp, md_t *md)
{
	md_t *devmd;
	md_node_t *devalias, **nodelist;
	md_string_t *mdstr;
	int num;

	nodelist = Xmalloc(md_node_count(md) * sizeof (md_node_t *));
	mdstr = md_find_name(md, "devalias");
	num = md_scan_dag(md, NULL, mdstr, md_find_name(md, "fwd"), nodelist);
	if (num > 0) {
		devmd = md_new_md();
		devalias = md_new_node(devmd, "devalias");
		md_copy_properties(devmd, devalias, nodelist[0]);

		ldp->devalias_md = devmd;
	}

	Xfree(nodelist);
}

static void
bind_cpus_and_maus(md_t *hvmdp,	int numguests, md_node_t **guestlistp,
    ldom_t **ldomlistp)
{
	md_node_t	**cpulistp;
	md_node_t	**maulistp;
	md_node_t	**scratchlistp;
	md_node_t	**cpuguestp;
	md_node_t	*guestp;
	int		numcpus;
	int		nummaus;
	md_string_t	*md_strp, *md_fwd, *md_back;
	int		i;
	ldom_t		*ldp;
	bool_t		*flag_visited;

	cpulistp = Xmalloc(sizeof (md_node_t *) * md_node_count(hvmdp));
	maulistp = Xmalloc(sizeof (md_node_t *) * md_node_count(hvmdp));
	scratchlistp = Xmalloc(sizeof (md_node_t *) * md_node_count(hvmdp));
	cpuguestp = Xmalloc(sizeof (md_node_t *)*md_node_count(hvmdp));

	/*
	 * Scan all the supplied CPUs and bind them appropriately to their
	 * guests.
	 */
	md_fwd = md_find_name(hvmdp, "fwd");
	md_back = md_find_name(hvmdp, "back");

	md_strp = md_find_name(hvmdp, "cpu");
	numcpus = md_scan_dag(hvmdp, NULL, md_strp, md_fwd, cpulistp);

	/*
	 * This is somewhat redundant given the guest parsing we do and the
	 * structure of the HV MD, but we'll have to fix this later.
	 */

	md_strp = md_find_name(hvmdp, "guest");
	for (i = 0; i < numcpus; i++) {
		uint64_t	pid;
		uint64_t	vid;
		uint64_t	partid;
		uint64_t	pm_status, fma_status;
		pri_cpu_t	*pcpup;
		md_node_t	*cpup;
		uint64_t	res_id;
		resid_t		*idx;
		int		j, num;

		cpup = cpulistp[i];

		if (md_get_prop_val(hvmdp, cpup, "pid", &pid))
			fatal("CPU has no pid");
		if (md_get_prop_val(hvmdp, cpup, "vid", &vid))
			fatal("CPU has no vid");
		if (md_get_prop_val(hvmdp, cpup, "partid", &partid))
			fatal("CPU has no partid");
		if (md_get_prop_val(hvmdp, cpup, "resource_id", &res_id))
			fatal("CPU has no res_id");
		idx = pull_res_id(&zeus.free_vcpu_ids, res_id);
		if (idx == NULL)
			fatal("Could not assign CPU's res_id of %"PRId64"",
			    res_id);

		/*
		 * Allocate out the VCPU from the PRI.  Bind it to the matched
		 * LDom.
		 */

		DBG(zprintf("\tCPU %d : vid %d\n", (int)pid, (int)vid));

		num = md_scan_dag(hvmdp, cpup, md_strp, md_back, cpuguestp);

		if (num != 1)
			fatal("Cpu has %d guest(s)!", num);

		guestp = cpuguestp[0];

		for (j = 0; j < numguests; j++)
			if (guestp == guestlistp[j])
				break;
		if (j >= numguests)
			fatal("CPU allocated to missing guest");

		ldp = ldomlistp[j];
		DBG(zprintf("\t\t-> %s\n", ldp->namep));

		pcpup = pri_find_pstrand((int)pid);
		if (pcpup == NULL)
			fatal("No CPU with pid 0x%"PRIx64" in PRI to match "
			    "with HV MD", pid);

		if (pcpup->res_id != NULL)
			fatal("Multiple resource IDs for CPU pid 0x%"PRIx64"",
			    pid);
		pcpup->res_id = idx;

		cpu_bind(pcpup, ldp, (int)vid, (int)partid, PRI_CPU_FRAC_DENOM);


		/* Update core_md status info (i.e. FMA & PM state) */
		if (md_get_prop_val(hvmdp, cpup, "fma_status", &fma_status)) {
			DBG(zprintf("CPU %"PRIu64" has no FMA status info\n",
			    pid));
		} else {
			if (set_core_md_cpu_status(pid, CORE_MD_OP_STATUS,
			    fma_status) != 0)
				warning("Unable to update FMA status of "
				    "cpu %"PRIu64" to %"PRIu64, pid,
				    fma_status);
		}
		if (md_get_prop_val(hvmdp, cpup, "pm_status", &pm_status)) {
			DBG(zprintf("CPU %"PRIu64" has no PM status info\n",
			    pid));
		} else {
			if (set_core_md_cpu_status(pid, CORE_MD_PM_STATUS,
			    pm_status) != 0)
				warning("Unable to update PM status of "
				    "cpu %"PRIu64" to %"PRIu64, pid, pm_status);
		}
	}
	Xfree(cpuguestp);

	/*
	 * If the guest is using its own value of tick, all the strands
	 * on the core must be allocated to the guest. Any strands that
	 * are not already bound must be reserved to ensure that they are
	 * not used by any other domain.
	 */
	for (i = 0; i < numguests; i++) {
		int		j;
		uint64_t	*pids = NULL;
		int		npids = 0;
		pri_cpu_t	*pcpu;

		ldp = ldomlistp[i];
		guestp = guestlistp[i];

		/* check if the guest is using its own tick */
		if (!ldp->flag.local_tick_sync)
			continue;

		/*
		 * To handle a possible crash during a migration, check
		 * for any cores that were active before the migration
		 * was started. These also need to be reserved.
		 */
		md_get_prop_data(hvmdp, guestp, HVMD_PROP_MIG_RSVD,
		    (uint8_t **)&pids, &npids);

		npids /= sizeof (uint64_t);

		ASSERT(npids != 0);

		for (j = 0; j < npids; j++) {
			pcpu = pid_to_pcpu((int)pids[j]);

			/* reserve all strands on the core */
			if (!cpu_reserve_core(ldp, pcpu->core, NULL)) {
				fatal("unable to reserve unbound strands on "
				    "core %d\n", pcpu->core->idx);
			}
		}
	}

	/*
	 * Get all the mau nodes currently in use
	 */
	md_strp = md_find_name(hvmdp, "mau");
	nummaus = md_scan_dag(hvmdp, NULL, md_strp, md_fwd, maulistp);

	if (nummaus > 0) {
		flag_visited = Xmalloc(nummaus * sizeof (int));
		for (i = 0; i < nummaus; i++)
			flag_visited[i] = false;
	}

	/*
	 * Now that we have bound all the cpus, now re-iterate over the
	 * vcpu nodes and bind the maus
	 */
	for (i = 0; i < numcpus; i++) {
		uint64_t	pid, ino;
		md_node_t	*cpup, *maup, *cwqp;
		md_node_t	*guestp;
		pri_mau_t	*pmaup;
		int		j, k, num;

		cpup = cpulistp[i];

		if (md_get_prop_val(hvmdp, cpup, "pid", &pid))
			fatal("CPU has no pid");

		num = md_scan_dag(hvmdp, cpup, md_find_name(hvmdp, "mau"),
		    md_fwd, scratchlistp);
		if (num == 0)
			continue;
		if (num != 1)
			fatal("CPU %"PRIu64" has %d maus", pid, num);

		maup = scratchlistp[0];

		if (pri.have_cwq) {
			num = md_scan_dag(hvmdp, cpup,
			    md_find_name(hvmdp, "cwq"), md_fwd, scratchlistp);
			if (num != 1)
				fatal("CPU %"PRIu64" has mau and %d cwqs",
				    pid, num);
			cwqp = scratchlistp[0];
		}

		num = md_scan_dag(hvmdp, cpup, md_find_name(hvmdp, "guest"),
		    md_back, scratchlistp);
		if (num != 1)
			fatal("CPU %"PRIu64" has %d guests", pid, num);

		guestp =  scratchlistp[0];

		for (j = 0; j < numguests; j++)
			if (guestp == guestlistp[j])
				break;
		if (j >= numguests)
			fatal("CPU %"PRIu64" allocated to missing guest", pid);

		ldp = ldomlistp[j];

		/*
		 * Get index of mau under consideration
		 */
		for (k = 0; k < nummaus; k++)
			if (maup == maulistp[k])
				break;
		if (k >= nummaus)
			fatal("mau node referenced in cpu node "
			    "could not be found in HV MD");

		/*
		 * At this point, we can associate an MAU with a guest,
		 * and also associate the mau node in HV to a PRI entry.
		 */
		pmaup = pri_find_pmau_from_cpuid(pid);

		/*
		 * Bind the mau to the guest
		 */
		if (!flag_visited[k]) {
			if (md_get_prop_val(hvmdp, maup, "ino", &ino))
				fatal("MAU has no ino");
			if (!pri_pull_mau_ino(ino))
				fatal("MAU ino %"PRIu64" not in PRI", ino);
			pmaup->mau_ino = ino;

			if (pri.have_cwq) {
				if (md_get_prop_val(hvmdp, cwqp, "ino", &ino))
					fatal("CWQ has no ino");
				if (!pri_pull_cwq_ino(ino))
					fatal("CWQ ino %"PRIu64" not in PRI",
					    ino);
				pmaup->cwq_ino = ino;
			}

			if (mau_bind(ldp, pmaup, k))
				flag_visited[k] = true;
		}
	}

	if (nummaus > 0)
		Xfree(flag_visited);

	/*
	 * Now we've counted the cpus & bound them forcibly update the CPU
	 * requirements.
	 */

	for (i = 0; i < numguests; i++) {
		ldom_t *ldp;
		vcpu_t *vcpu;
		vmau_t *vmau;
		int ncpu, nmau;

		ldp = ldomlistp[i];

		/*
		 * Mark all CPUs and crypto units in the guest configured.
		 * FIXME: We really need to get this information from the
		 * hypervisor.
		 */
		CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
			vcpu->flag_isconfigd = true;
		}

		CHAIN_ITER(ldp->bound.vmaus, vmau, ldom) {
			vmau->flag_isconfigd = true;
		}

		ncpu = ldom_num_bound_vcpus(ldp);
		if (database_ldom_get_cpu(ldp) < 0)
			ldp->required.num_vcpus = ncpu;
		database_ldom_set_cpus(ldp, ncpu);

		nmau = ldom_num_bound_maus(ldp);
		if (database_ldom_get_mau(ldp) < 0)
			ldp->required.num_vmaus = nmau;
		database_ldom_set_maus(ldp, nmau);
	}

	Xfree(cpulistp);
	Xfree(maulistp);
	Xfree(scratchlistp);
}

static void
sanity_check_memory(ldom_t *ldp)
{
	uint64_t bound_mem, reqd_mem;

	bound_mem = ldom_size_bound_mem(ldp);
	reqd_mem = ldp->required.size_mem;

	if (bound_mem != reqd_mem) {
		warning("Bound memory (0x%"PRIx64") for guest %s does not "
		    "match requested memory (0x%"PRIx64")", bound_mem,
		    ldp->namep, reqd_mem);
		ldp->required.size_mem = bound_mem;
	}
}

static void
sanity_check_cpus(ldom_t *ldp, md_t *md)
{
	md_node_t **cpulist;
	int bound_cpus, reqd_cpus, num_vcpus;

	cpulist = Xmalloc(sizeof (md_node_t *) * md_node_count(md));
	num_vcpus = md_scan_dag(md, NULL, md_find_name(md, "cpu"),
	    md_find_name(md, "fwd"), cpulist);
	Xfree(cpulist);

	bound_cpus = ldom_num_bound_vcpus(ldp);
	reqd_cpus = ldp->required.num_vcpus;

	if (bound_cpus != num_vcpus)
		warning("CPU count mismatch for guest %s, between HV MD "
		    "and guest MD", ldp->namep);
	if (bound_cpus != reqd_cpus) {
		warning("Bound CPUs (%u) for guest %s does not "
		    "match requested CPUs (%u)", bound_cpus,
		    ldp->namep, reqd_cpus);
		ldp->required.num_vcpus = bound_cpus;
	}
}

static void
sanity_check_maus(ldom_t *ldp)
{
	uint64_t bound_maus, reqd_maus;

	bound_maus = ldom_num_bound_maus(ldp);
	reqd_maus = ldp->required.num_vmaus;

	if (bound_maus != reqd_maus) {
		warning("Bound maus (%"PRIu64") for guest %s does not "
		    "match requested maus (%"PRIu64")", bound_maus,
		    ldp->namep, reqd_maus);
		ldp->required.num_vmaus = bound_maus;
	}
}

/*
 * Parse the current HV MD.
 */
static void
niagara1_parse_hvmd(md_t *hvmdp)
{
	md_node_t	**nodelistp;
	md_node_t	**guestlistp;
	int		num;
	int		numguests;
	md_string_t	*md_strp, *md_fwd;
	uint64_t	fragsize;
	uint64_t	hvmd_version;
	uint64_t	next_spconfig;
	char		*current_spconfig;
	int		i;
	ldom_t 		**ldomlistp = NULL;
	ldom_t		*ldp;

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(hvmdp));

	/*
	 * Get version # of HV MD.
	 */
	if (md_get_prop_val(hvmdp, md_root_node(hvmdp), "content-version",
	    &hvmd_version) != 0) {
		zeus.hvmd_vers_major = 0;
		zeus.hvmd_vers_minor = 0;
	} else {
		zeus.hvmd_vers_major = hvmd_version >> 32;
		zeus.hvmd_vers_minor = hvmd_version & 0xffffffffLL;
	}

	DBG(zprintf("HV MD version from HV is %"PRIu64".%"PRIu64"\n",
	    zeus.hvmd_vers_major, zeus.hvmd_vers_minor));

	/*
	 * This logic assumes we support any major version of MD up to and
	 * including HVMD_VERSION_MAJOR_NUMBER.
	 *
	 * For major version 1 the meaning of the minor versions are:
	 * 0: baseline
	 * 1: removal of properties that are provided by the mini MD from
	 *    SP to HV
	 */
	if (zeus.hvmd_vers_major > HVMD_VERSION_MAJOR_NUMBER)
		fatal("HV MD major version mismatch: found version %"PRIu64",\n"
		    "can only support version %"PRIu64"", zeus.hvmd_vers_major,
		    HVMD_VERSION_MAJOR_NUMBER);

	if (HVMD_VERSION_MINOR_NUMBER < zeus.hvmd_vers_minor)
		zeus.hvmd_vers_minor = HVMD_VERSION_MINOR_NUMBER;

	DBG(zprintf("Negotiated HV MD version is %"PRIu64".%"PRIu64"\n",
	    zeus.hvmd_vers_major, zeus.hvmd_vers_minor));

	/*
	 * Is the current HV config the same as the one selected as the
	 * boot set on the SP?
	 */
	if (md_get_prop_val(hvmdp, md_root_node(hvmdp), "is_next_spconfig",
	    &next_spconfig) == 0)
		zeus.is_next_spconfig = next_spconfig;

	/* get the current spconfig name if one was saved */
	if (md_get_prop_str(hvmdp, md_root_node(hvmdp), "current_spconfig",
	    &current_spconfig) == 0)
		zeus.current_spconfig = Xstrdup(current_spconfig);

	/*
	 * Pick up any memory frag blocks provided by the HV.
	 */
	md_fwd = md_find_name(hvmdp, "fwd");

	md_strp = md_find_string(hvmdp, "frag_space", NULL);
	num = md_scan_dag(hvmdp, NULL, md_strp, md_fwd, nodelistp);
	if (num != 1)
		fatal("%d frag_space nodes in the HV MD", num);

	if (md_get_prop_val(hvmdp, nodelistp[0], "fragsize", &fragsize))
		fatal("HV MD does not specify fragsize");

	/* FIXME: need to adjust to specified frag size */
	if (fragsize != FRAG_SIZE)
		fatal("Inconsistent FRAG_SIZE between HV MD (%"PRIu64") "
		    "and LDom Mgr (%d)", fragsize, FRAG_SIZE);

	/*
	 * Attempt to reserve frag at physical address 0 as a red-zone.
	 * Ignore any failure, as it most likely means there's no memory
	 * at physical location 0 anyway.
	 */
	(void) pull_bmblock(0, fragsize);

	/* grab pre-allocated frag space */
	md_strp = md_find_string(hvmdp, "frag_mblock", NULL);
	num = md_scan_dag(hvmdp, nodelistp[0], md_strp, md_fwd, nodelistp);
	if (num < 1)
		warning("No frag_mblock nodes in the HV MD");

	for (i = 0; i < num; i++) {
		grab_frag_mblock(hvmdp, nodelistp[i]);
	}

	/*
	 * Grab pre-allocated HV MD space if available (else use frag space
	 * for HV MDs).
	 */
	md_strp = md_find_string(hvmdp, "hvmd_mblock", NULL);
	num = md_scan_dag(hvmdp, NULL, md_strp, md_fwd, nodelistp);
	if (num > 0) {
		uint64_t base, size, md_maxsize;

		if (md_get_prop_val(hvmdp, nodelistp[0], "base", &base))
			fatal("Missing base property in hvmd_mblock node of "
			    "HV MD");
		if (md_get_prop_val(hvmdp, nodelistp[0], "size", &size))
			fatal("Missing size property in hvmd_mblock node of "
			    "HV MD");
		if (md_get_prop_val(hvmdp, nodelistp[0], "md_maxsize",
		    &md_maxsize))
			fatal("Missing md_maxsize property in hvmd_mblock node "
			    "of HV MD");

		if (pull_bmblock(base, size) == NULL)
			fatal("Unable to assign HV MD area");

		hv_md_alloc_init(base, size, md_maxsize);
	}

	/*
	 * find other reserved memory areas and remove from free pool
	 */
	md_strp = md_find_string(hvmdp, "rsvd_space", NULL);
	num = md_scan_dag(hvmdp, NULL, md_strp, md_fwd, nodelistp);
	if (num > 0) {
		if (num != 1)
			fatal("%d rsvd_space nodes in the HV MD", num);

		md_strp = md_find_string(hvmdp, "rsvd_mblock", NULL);
		num = md_scan_dag(hvmdp, nodelistp[0], md_strp, md_fwd,
		    nodelistp);
		for (i = 0; i < num; ++i) {
			uint64_t base, size;
			mblock_t *mblk;

			if (md_get_prop_val(hvmdp, nodelistp[i], "base", &base))
				fatal("Missing base prop in rsvd_mblock node");
			if (md_get_prop_val(hvmdp, nodelistp[i], "size", &size))
				fatal("Missing size prop in rsvd_mblock node");

			DBG(zprintf("Reserving mem: 0x%"PRIx64":0x%"PRIx64"\n",
			    base, size));

			/* remove from free pool, never to be seen again */
			mblk = pull_bmblock(base, size);
			if (mblk == NULL)
				fatal("Unable to assign reserved area: "
				    "0x%"PRIx64":0x%"PRIx64"\n", base, size);
			CHAIN_ADD(zeus.rsvd_mblocks, mblk, ldom);

		}
	}

	/*
	 * Scan all the currently running guests and match them to their LDOM
	 * requirements.
	 */
	guestlistp = Xmalloc(sizeof (md_node_t *) * md_node_count(hvmdp));
	md_strp = md_find_name(hvmdp, "guest");
	numguests = md_scan_dag(hvmdp, NULL, md_strp, md_fwd, guestlistp);

	if (numguests > 0) {
		/* create the bound or active guests found in the MD */

		ldomlistp = Xmalloc(numguests * sizeof (ldom_t *));
		for (i = 0; i < numguests; i++) {
			ldomlistp[i] = grab_basic_guest_state(hvmdp, i,
			    guestlistp[i]);
		}
	}

	bind_cpus_and_maus(hvmdp, numguests, guestlistp, ldomlistp);

	/*
	 * Build a temporary database of the LDC cross-reference
	 * map. Simple Linear list of tuples:
	 * (guestid, endpoint, ldom) -> (guestid, endpoint, ldom)
	 */
	initialize_ldcdb(hvmdp, guestlistp, numguests);

	DBG(dump_ldcdb());

	/*
	 * Now cross check with the guest MDs to complete
	 * the VIO state information.
	 */

	for (i = 0; i < numguests; i++) {
		md_t *guest_mdp;
		md_node_t *hv_guestp;
		ldom_t	*ldp;
		req_iodevice_t *riop;
		bound_iodevice_t *biop;

		ldp = ldomlistp[i];
		hv_guestp = guestlistp[i];

		/*
		 * Need to reconstruct Virtual I/O client & servers,
		 * and the bindings between them.  This requires us
		 * to walk through the guest MDs as stored in the HV.
		 */
		guest_mdp = backend.get_guest_md(ldp);

		/*
		 * Pick up the current guest md generation number and memory
		 * generation ID, if any.
		 */
		md_get_prop_val(guest_mdp, md_root_node(guest_mdp),
		    "md-generation#", &ldp->props.md_generation_num);
		ldp->props.mem_generation_id =
		    grab_guest_mem_generation_id(guest_mdp);

		/* Restore UUID from the MD or the database */
		grab_guest_uuid(ldp, guest_mdp);

		/*
		 * If the guest is active and has a MD generation number
		 * greater than zero, then mark it as requiring an MD update
		 * notification, in case the LDom manager failed earlier to
		 * send a notification for the current MD. The notification is
		 * sent when the MD update domain service is registered by the
		 * guest.
		 */
		if (ldp->state == ACTIVE && ldp->props.md_generation_num > 0)
			ldp->flag.md_notify = true;

		/*
		 * Find the allocated physical I/O devices
		 * and force bind them.
		 */
		grab_pio_state(ldp, guest_mdp, hv_guestp, hvmdp);

		/* Initialize internal VIO state */
		grab_guest_vio_state(ldp, guest_mdp, hv_guestp, hvmdp);

		/*
		 * Override requirements from Zeus local persistence
		 * with info based on any bound active resources
		 */
		sanity_check_memory(ldp);
		sanity_check_cpus(ldp, guest_mdp);
		sanity_check_maus(ldp);

		/* Restore any guest keys and variables */
		restore_guest_keys(ldp, guest_mdp);
		restore_guest_variables(ldp, guest_mdp);

		/*
		 * If this is the control domain check if domaining enabled is
		 * set and grab the devaliases.
		 */
		if (ldp == zeus.hvctl_ldom_p) {
			grab_devaliases(ldp, guest_mdp);
			check_if_domaining_enabled(guest_mdp);
		}

		/* Update physio requirements */
		CHAIN_ITER(ldp->bound.iodevices, biop, ldom) {
			/*
			 * If virtual_pcie_bus is false, means its
			 * either physical pcie bus iodevice or non-pcie
			 * bus device (niu).
			 */
			if (biop->virtual_pcie_bus == false)  {
				riop = Xcalloc(1, req_iodevice_t);
				riop->name = Xstrdup(biop->name);
				/*
				 * Don't propagate start-of-day default
				 * (which is bypass enabled); our default
				 * is bypass disabled.
				 */
				if (zeus.domaining_enabled == false) {
					riop->io_mmu_bypass = false;
				} else {
					riop->io_mmu_bypass =
					    biop->io_mmu_bypass;
				}
				riop->bound = true;
				CHAIN_ADD(ldp->required.io_devices, riop, link);
				biop->req_iop = riop;
				database_ldom_add_iodevice(ldp, riop->name,
				    riop->io_mmu_bypass);
			}

			/* Update the pcie device requirements. */
			pcie_update_reqs(ldp, biop);
		}

		backend.parse_hv_guest(ldp, hv_guestp, hvmdp);
		md_fini(guest_mdp);
	}

	/*
	 * Parse HVCTL channel data from the HV MD. At this point the LDom
	 * owning this channel is known.
	 */
	add_hvctl_from_md(hvmdp, zeus.hvctl_ldom_p);

	grab_platform_sp_channels(hvmdp);

	if (!zeus.domaining_enabled) {
		add_pri_ldc_endpoints();
		add_sp_vio(zeus.hvctl_ldom_p);
	}

	/*
	 * Now do VIO binding between clients and server ports.
	 */
	finalize_vio_bindings();

	cleanup_ldcdb();

	/*
	 * Once the guest MDs are parsed, check if libds is functional
	 */
	if (libds_cap.supported == true)
		initialize_libds();

	/*
	 * Register bound domains for Domain Service communication.
	 * If the domain is recovering from an interrupted migration,
	 * the DS setup will be done later, so it can be skipped here.
	 */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if ((ldp->state >= BOUND) && !ldp->flag.migration_source) {
			if (ds_register_domain(ldp, true) != 0)
				warning("Failed to register LDom %s for "
				    "Domain Services", ldp->namep);
		}
	}

	/* database may have been updated, save it */
	database_save();

	Xfree(nodelistp);
	Xfree(guestlistp);
	Xfree(ldomlistp);
}

static void
grab_platform_sp_channels(md_t *hvmdp)
{
	md_node_t	**nodelistp;
	md_string_t	*md_strp, *md_fwd;
	int		num;
	uint64_t	dummy;
	int		i;

	nodelistp = Xmalloc(md_node_count(hvmdp) * sizeof (md_node_t *));

	md_fwd = md_find_name(hvmdp, "fwd");

	md_strp = md_find_name(hvmdp, "ldc_endpoint");
	num = md_scan_dag(hvmdp, NULL, md_strp, md_fwd, nodelistp);

	DBG(zprintf_lock());
	DBG(zprintf("SP->guest endpoints:"));

	for (i = 0; i < num; i++) {
		md_node_t		*sp_ldc_p;
		uint64_t		target_type;
		uint64_t		channel;
		uint64_t		target_guest;
		uint64_t		target_channel;
		uint64_t		res_id;

		sp_ldc_p = nodelistp[i];

		if (!md_get_prop_val(hvmdp, sp_ldc_p, "svc_id", &dummy)) {
			/* Not a SP endpoint - skip */
			continue;
		}

		if (!md_get_prop_val(hvmdp, sp_ldc_p, "tx-ino", &dummy)) {
			/* Not a SP endpoint - skip */
			continue;
		}

		if (md_get_prop_val(hvmdp, sp_ldc_p, "target_type",
		    &target_type)) {
			warning("SP ldc_endpoint %d has no target_type", i);
			continue;
		}

		if (md_get_prop_val(hvmdp, sp_ldc_p, "channel", &channel)) {
			warning("SP ldc_endpoint %d has no channel", i);
			continue;
		}

		if (md_get_prop_val(hvmdp, sp_ldc_p, "resource_id",
		    &res_id)) {
			warning("SP ldc_endpoint %d has no resource_id", i);
			continue;
		}

		/* SP->HV channels have no target_guest field */
		target_guest = -1;
		if (target_type == TYPE_GUEST) {
			if (md_get_prop_val(hvmdp, sp_ldc_p, "target_guest",
			    &target_guest)) {
				warning("SP ldc_endpoint %d has no "
				    "target_guest", i);
				continue;
			}
		}

		if (md_get_prop_val(hvmdp, sp_ldc_p, "target_channel",
		    &target_channel)) {
			warning("SP ldc_endpoint %d has no target_channel", i);
			continue;
		}

		new_sp_endpoint(channel, target_type, target_guest,
		    target_channel, res_id);
	}

	Xfree(nodelistp);
	DBG(zprintf("\n"));
	DBG(zprintf_unlock());
}

extern int call_hv_get_tick_sync_mode(ldom_t *ldp, uint32_t *mode);
extern md_t *read_md(uint64_t md_paddr);

/*
 * Initialises running/bound LDom state based off the hypervisor's MD state at
 * startup.
 */
static ldom_t *
grab_basic_guest_state(md_t *mdp, int idx, md_node_t *guestp)
{
	mblock_t	*bmp, *next_bmp;
	char		*sp;
	uint64_t	gid, guest_res_id, mblock_res_id;
	uint64_t	membase;
	uint64_t	memsize;
	uint64_t	realbase;
	uint64_t	guestmdpa;
	uint64_t	mig_flag;
	uint64_t	mig_mdpa;
	uint64_t	mem_dr_flag;
	ldom_t		*ldp;
	md_node_t	**memlist;
	int		m, num_memblocks;
	mem_blocks_t	mi;
	uint32_t	mode;
	int		rv;

	if (md_get_prop_str(mdp, guestp, "name", &sp)) {
		warning("Guest %d has no name - naming it domain%d", idx, idx);
		sp = alloca(16);
		(void) sprintf(sp, "domain%d", idx);
	}

	DBG(zprintf("Guest %s:\n", sp));

	if (md_get_prop_val(mdp, guestp, "gid", &gid))
		fatal("Guest has no gid");

	memlist = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	num_memblocks = md_scan_dag(mdp, guestp, md_find_name(mdp, "mblock"),
	    md_find_name(mdp, "fwd"), memlist);

	CHAIN_INIT(mi.blocks);

	for (m = 0; m < num_memblocks; ++m) {
		mblock_t *lm;
		md_node_t *mblock = memlist[m];

		if (md_get_prop_val(mdp, mblock, "membase", &membase))
			fatal("Guest mblock has no membase");
		if (md_get_prop_val(mdp, mblock, "memsize", &memsize))
			fatal("Guest mblock has no memsize");
		if (md_get_prop_val(mdp, mblock, "realbase", &realbase))
			fatal("Guest mblock has no realbase");
		if (md_get_prop_val(mdp, mblock, "resource_id", &mblock_res_id))
			fatal("Guest mblock has no res_id");

		lm = Xcalloc(1, mblock_t);
		lm->paddr = membase;
		lm->raddr = realbase;
		lm->size = memsize;
		lm->flag_isconfigd = true;

		lm->mblk_resid = pull_res_id(&zeus.free_mblk_ids,
		    mblock_res_id);
		if (lm->mblk_resid == NULL)
			fatal("Couldn't assign mblock res_id %"PRIu64"",
			    mblock_res_id);

		CHAIN_ADD(mi.blocks, lm, blk);

		DBG(zprintf("\tID:%"PRIu64"\tmblock %d "
		    "[0x%"PRIx64" + 0x%"PRIx64"]\trealbase = 0x%"PRIx64"\n",
		    gid, m, membase, memsize, realbase));
	}

	Xfree(memlist);

	ldp = find_ldom_by_name(sp);

	if (ldp == NULL)
		ldp = new_ldom(sp);

	ldp->idp = pull_res_id(&zeus.free_ldom_ids, gid);
	if (ldp->idp == NULL)
		fatal("ID 0x%"PRIx64" is not available", gid);

	/*
	 * Check if this LDom is in a delayed reconfig.
	 */
	if (zeus.delayed_reconfig_gid == gid)
		zeus.delayed_reconfig_ldom = ldp;

	/*
	 * Check if this LDom was part of a migration.
	 */
	if (md_get_prop_val(mdp, guestp, HVMD_PROP_MIG_TGT, &mig_flag) == 0)
		ldp->flag.migration_target = true;

	if (md_get_prop_val(mdp, guestp, HVMD_PROP_MIG_SRC, &mig_flag) == 0) {
		if (md_get_prop_val(mdp, guestp, HVMD_PROP_MIG_MDPA,
		    &mig_mdpa) == 0) {
			ldp->flag.migration_source = true;
			ldp->mig_state.bind_mdp = read_md(mig_mdpa);
			ldp->mig_state.bind_fp = frag_assign(ldp, mig_mdpa);

			ASSERT(ldp->mig_state.bind_mdp != NULL);
			ASSERT(ldp->mig_state.bind_fp != NULL);
		}
	}

	/*
	 * Check if this LDom was part of memory DR.
	 * Memory DR has the following sequences:
	 * For memory add:
	 *  1. Allocate request memory from memory pool and update MD with
	 *	HVMD_PROP_MEMDR_IN_PROGRESS flag set in MD.
	 *  2. Send memory add request to the domain.
	 *  3. If the request succeeded, reset flag in MD.
	 *  If the request failed or partial succeeded, update MD reset
	 *  flag in MD.
	 * For memory del:
	 *  1. update MD with HVMD_PROP_MEMDR_IN_PROGRESS flag set in MD.
	 *  2. send memory del request to the domain.
	 *  3. update MD based on the result from the domain and reset the
	 *  the flag in the MD.
	 * If zeus craches or disabled for whatever reason while memory dr
	 * is in progress, memory list in MD and in the domain may be
	 * inconsistant. When zeus is intialized, if the
	 * HVMD_PROP_MEMDR_IN_PROGRESS flag is set, zeus requests memory list
	 * from the domain and update MD accordingly.
	 */
	if (md_get_prop_val(mdp, guestp, HVMD_PROP_MEMDR_IN_PROGRESS,
	    &mem_dr_flag) == 0) {
		ldp->mem_dr_state.inprogress = true;
	} else {
		ldp->mem_dr_state.inprogress = false;
	}

	/* lookup any domain dependencies for this domain */
	grab_guest_dependencies(mdp, guestp, ldp);

	if (md_get_prop_val(mdp, guestp, "resource_id", &guest_res_id))
		fatal("Guest has no resource_id");

	ldp->res_id = pull_res_id(&zeus.free_guest_ids, guest_res_id);
	if (ldp->res_id == NULL)
		fatal("Guest res_id 0x%"PRIx64" is not available",
		    guest_res_id);

	(void) md_get_prop_val(mdp, guestp, "tod-offset",
	    &ldp->props.tod_offset);

	(void) md_get_prop_val(mdp, guestp, "perfctraccess",
	    &ldp->props.perfctraccess);

	(void) md_get_prop_val(mdp, guestp, "perfctrhtaccess",
	    &ldp->props.perfctrhtaccess);

	(void) md_get_prop_val(mdp, guestp, "rngctlaccessible",
	    &ldp->props.rngctlaccessible);

	(void) md_get_prop_val(mdp, guestp, "diagpriv", &ldp->props.diagpriv);

	/* explicit bind based on info in MD */
	if (!mem_bind_virt_explicit(ldp, &mi))
		fatal("Cannot bind guest's memory");

	ldp->required.size_mem = ldom_size_bound_mem(ldp);
	database_ldom_set_memory(ldp, ldp->required.size_mem);

	/* Free mi chain */
	CHAIN_ITER_VOL(mi.blocks, bmp, next_bmp, blk) {
		Xfree(bmp);
	}

	/*
	 * Locate and reserve the guest MD
	 */
	if (md_get_prop_val(mdp, guestp, "mdpa", &guestmdpa) != 0)
		fatal("Guest node of HV MD missing mdpa property");

	ldp->flag.gen_new_md = false;	/* currently using existing one */
	if (zeus.delayed_reconfig_ldom == ldp) {
		/* We're still in a delayed reconfig - use next_md_frag */
		ldp->frags.next_md_frag = frag_assign(ldp, guestmdpa);
		ASSERT(ldp->frags.next_md_frag != NULL);
		/* set parsed the same as next */
		ldp->frags.parsed_md_frag = frag_clone(ldp->frags.next_md_frag);
		/* committed_md_frag will be set after HV MD is parsed */
		ldp->frags.committed_md_frag = NULL;
		ldp->flag.gen_new_md = true;
	} else {
		ldp->frags.committed_md_frag = frag_assign(ldp, guestmdpa);
		ASSERT(ldp->frags.committed_md_frag != NULL);
		ldp->frags.next_md_frag = NULL;
	}

	DBG(zprintf("\tGuest MD is located with frag at 0x%"PRIx64", "
	    "gen %"PRIx64"\n", guestmdpa, ldp->props.md_generation_num));

	/* get guest's state from the hypervisor */
	backend.get_guest_state(ldp, LDOM_HV_STATE|LDOM_SOFT_STATE,
	    &ldp->hv_state);
	if (ldp->hv_state.state == GUEST_STATE_NORMAL ||
	    ldp->hv_state.state == GUEST_STATE_RESETTING)
		ldp->state = ACTIVE;
	else
		ldp->state = BOUND;

	/* get guest's tick sync mode from the hypervisor */
	/* XXXMIG - make this a proper backend function */
	rv = call_hv_get_tick_sync_mode(ldp, &mode);
	if ((rv == 0) && (mode == GUEST_TICK_SYNC_LOCAL))
		ldp->flag.local_tick_sync = true;

	return (ldp);
}

static void
get_iodev_props(md_t *hvmdp, md_node_t **hvnodes, int num, uint64_t cfg_hdl,
    uint64_t *resid, bool_t *io_mmu_bypass)
{
	uint64_t dev_cfg_hdl, bypass;
	md_node_t *dev;
	int i;

	for (i = 0; i < num; ++i) {
		dev = hvnodes[i];
		if (md_get_prop_val(hvmdp, dev, "cfghandle", &dev_cfg_hdl) != 0)
			fatal("HV MD physical device node has no "
			    "cfghandle property");

		if (cfg_hdl != dev_cfg_hdl)
			continue;

		/* found match - get resource_id & bypass mode */
		if (md_get_prop_val(hvmdp, dev, "resource_id", resid) != 0)
			fatal("HV MD physical device node has no "
			    "resource_id property");

		if (md_get_prop_val(hvmdp, dev, "allow_bypass", &bypass) != 0 ||
		    bypass == 0)
			*io_mmu_bypass = false;
		else
			*io_mmu_bypass = true;
		break;
	}

	if (i == num)
		fatal("No HV MD node found for iodevice");
}

/*
 * Locate & bind the physical iodevices.
 */
static void
grab_pio_state(ldom_t *ldp, md_t *guest_mdp, md_node_t *hv_guestp, md_t *hvmdp)
{
	bound_iodevice_t	*biop;
	bound_iodevice_t	*tbiop;
	md_node_t		**nodelistp, **hvnodelistp;
	md_node_t		*phys_io, *parent;
	int			i, num, num_iodevices;
	char			*device_type;
	uint64_t		cfghdl, res_id;
	bool_t			io_mmu_bypass;
	uint64_t		hostid;
	bool_t			virtual = false;
	bool_t			is_dio_hvmd = pcie_is_dio_hvmd_supported();

	nodelistp = Xmalloc(sizeof (md_node_t *)*md_node_count(guest_mdp));
	hvnodelistp = Xmalloc(sizeof (md_node_t *)*md_node_count(hvmdp));

	num = md_scan_dag(guest_mdp, NULL, md_find_name(guest_mdp, "phys_io"),
	    md_find_name(guest_mdp, "fwd"), nodelistp);
	if (num != 1)
		fatal("Guest MD does not contain a singleton phys_io node");
	phys_io = nodelistp[0];

	num_iodevices = md_scan_dag(guest_mdp, NULL,
	    md_find_name(guest_mdp, "iodevice"),
	    md_find_name(guest_mdp, "fwd"), nodelistp);

	for (i = 0; i < num_iodevices; i++) {
		/* we're only interested top-level devices */
		if (md_get_prop_arc(guest_mdp, nodelistp[i], "back",
		    &parent) < 0)
			fatal("Missing back arc in guest MD iodevice node");
		if (parent != phys_io)
			continue;	/* not at top level */

		if (md_get_prop_str(guest_mdp, nodelistp[i],
		    "device-type", &device_type) != 0)
			fatal("Guest MD iodevice node has no device-type "
			    "property");
		if (md_get_prop_val(guest_mdp, nodelistp[i],
		    "cfg-handle", &cfghdl) != 0)
			fatal("Guest MD iodevice node has no cfg-handle "
			    "property");

		if (strcmp(device_type, "pciex") == 0) {
			num = md_scan_dag(hvmdp, hv_guestp,
			    md_find_name(hvmdp, "pcie_bus"),
			    md_find_name(hvmdp, "fwd"), hvnodelistp);
			/*
			 * If it is not Pre-DirectIO HVMD, then
			 * we need to look for virtual_pcie_bus node.
			 */
			if ((num == 0) && is_dio_hvmd) {
				num = md_scan_dag(hvmdp, hv_guestp,
				    md_find_name(hvmdp, "virtual_pcie_bus"),
				    md_find_name(hvmdp, "fwd"), hvnodelistp);
				virtual = true;
			}

			if (num <= 0) {
				fatal("Corresponding pcie_bus/virtual_pcie_bus"
				    " not found in HVMD");
			}

			/*
			 * Allocate a temporary bound_iodevice structure to
			 * accumulate the info needed to create the final
			 * version.
			 */
			tbiop = Xcalloc(1, bound_iodevice_t);
			tbiop->virtual_pcie_bus = virtual;

			get_pcie_props(hvmdp, hvnodelistp, num, cfghdl,
			    &res_id, tbiop);

			biop = pull_pcie_biodevice(device_type,
			    cfghdl, res_id, tbiop, !is_dio_hvmd);
			/* dev_list directly copied, so don't need to free */
			Xfree(tbiop);
		} else {
			/*
			 * FIXME: for now everything which is not of type
			 * "pciex" is assumed to be of type "network_device".
			 */
			num = md_scan_dag(hvmdp, hv_guestp,
			    md_find_name(hvmdp, "network_device"),
			    md_find_name(hvmdp, "fwd"), hvnodelistp);

			get_iodev_props(hvmdp, hvnodelistp, num, cfghdl,
			    &res_id, &io_mmu_bypass);
			biop = pull_biodevice(device_type, cfghdl, res_id,
			    io_mmu_bypass);

		}

		if (biop == NULL) {
			DBG(zprintf("WARNING: iodevice %s (cfg-handle "
			    "%"PRIu64") not in PRI\n", device_type, cfghdl));
			continue;
		}

		CHAIN_ADD(ldp->bound.iodevices, biop, ldom);
	}

	/* get MAC address and hostid */
	num = md_scan_dag(guest_mdp, NULL, md_find_name(guest_mdp, "platform"),
	    md_find_name(guest_mdp, "fwd"), nodelistp);
	if (num > 0) {
		if (md_get_prop_val(guest_mdp, nodelistp[0], "mac-address",
		    &ldp->props.mac_address) == 0) {
			database_ldom_set_mac_addr(ldp, ldp->props.mac_address);
		}

		if (md_get_prop_val(guest_mdp, nodelistp[0], "hostid",
		    &hostid) == 0) {
			if (!IS_DERIVED_HOSTID_ASSIGNED(ldp->props.mac_address,
			    hostid))
				ldp->props.hostid = hostid;
			database_ldom_set_hostid(ldp, ldp->props.hostid);
		}
	}

	Xfree(nodelistp);
	Xfree(hvnodelistp);
}

static void
grab_frag_mblock(md_t *mdp, md_node_t *fragp)
{
	uint64_t	base;
	uint64_t	size;
	mblock_t	*bmbp;

	if (md_get_prop_val(mdp, fragp, "base", &base))
		fatal("Frag mblock has no base");
	if (md_get_prop_val(mdp, fragp, "size", &size))
		fatal("Frag mblock has no size");

	DBG(zprintf("\tFrag mblock @ 0x%"PRIx64" + 0x%"PRIx64"\n", base, size));

	bmbp = pull_bmblock(base, size);
	if (bmbp == NULL)
		fatal("Unable to assign frag mblock from free PRI list");

	frag_free_mblock(bmbp);
}

static void
grab_guest_dependencies(md_t *mdp, md_node_t *guestp, ldom_t *ldp)
{
	int		i;
	md_node_t	**masters;
	int		num_masters;
	master_t	*master;
	ldom_t		*mldp;
	uint64_t	resid = 0;
	uint64_t	gctrl;

	/* dependencies should not be setup yet */
	ASSERT(NUM_DEPENDENCIES(ldp) == 0);

	/* lookup the domain behavior for dependent guests */
	if (md_get_prop_val(mdp, guestp, HVMD_NODE_GUESTCONTROL, &gctrl) == 0) {
		ldp->props.failure_policy = gctrl;
		database_ldom_set_failure_policy(ldp, gctrl);
	}

	masters = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));
	num_masters = md_scan_dag(mdp, guestp,
	    md_find_name(mdp, HVMD_NODE_GUESTDEPENDENCY),
	    md_find_name(mdp, "fwd"), masters);

	ASSERT(num_masters <= pri.max_guest_dependencies);

	if (num_masters > 0)
		free_masters(ldp);

	for (i = 0; i < num_masters; i++) {
		if (md_get_prop_val(mdp, masters[i],
		    HVMD_PROP_DEPENDENCY_GID, &resid) != 0) {
			DBG(zprintf("%s: guest dependency has no resource "
			    "id\n", __func__));
			continue;
		}

		if ((mldp = find_ldom_by_resid(resid)) == NULL) {
			DBG(zprintf("%s: specified guest dependency LDom "
			    "%"PRId64" does not exist\n", __func__, resid));
			continue;
		}

		DBG(zprintf("%s: LDom %s has dependency on LDom %s\n",
		    __func__, ldp->namep, mldp->namep));

		if (DEPENDENCY_LIMIT_REACHED(ldp)) {
			DBG(zprintf("%s: LDom %s has reached the limit for "
			    "guest dependencies (%d)\n", __func__, ldp->namep,
			    pri.max_guest_dependencies));
			goto done;
		}

		master = Xcalloc(1, master_t);
		master->ldp = mldp;

		CHAIN_ADD(ldp->props.masters, master, link);
		database_ldom_set_master(ldp);
	}

done:
	Xfree(masters);
}

static void
get_assigned_pcie_devs(uint64_t cfg_hdl, bound_iodevice_t *bio,
    md_t *hvmdp, md_node_t *dev)
{
	md_node_t **nodes;
	int num;
	int i;
	md_node_t *nodep;

	nodes = Xmalloc(md_node_count(hvmdp) * sizeof (md_node_t *));

	num = md_scan_dag(hvmdp, dev,
	    md_find_name(hvmdp, "pcie_assigned_devices"),
	    md_find_name(hvmdp, "fwd"), nodes);

	if (num == 0) {
		Xfree(nodes);
		return;
	}

	ASSERT(num == 1);

	nodep = nodes[0];

	num = md_scan_dag(hvmdp, nodep,
	    md_find_name(hvmdp, "pcie_device"),
	    md_find_name(hvmdp, "fwd"), nodes);

	if (num <= 0) {
		Xfree(nodes);
		return;
	}

	for (i = 0; i < num; i++) {
		char *path;
		bound_pcie_dev_t *bdev;

		nodep = nodes[i];
		if (md_get_prop_str(hvmdp, nodep, "path", &path) != 0) {
			warning("pcie_device node has no path prop");
			continue;
		}

		bdev = Xcalloc(1, bound_pcie_dev_t);
		bdev->skel_path = pcie_build_skel_path(cfg_hdl, path);
		CHAIN_ADD(bio->dev_list, bdev, ldom);
	}

	Xfree(nodes);
}

/*
 * Create a bound_iodevice_t and populate it with data from HV MD
 * and guest MD.  Unpopulated elements will be added later.
 */
static void
get_pcie_props(md_t *hvmdp, md_node_t **hvnodes, int num,
    uint64_t cfg_hdl, uint64_t *resid, bound_iodevice_t *bio)
{
	uint64_t dev_cfg_hdl, bypass;
	md_node_t *dev;
	int i;
	int rlen;
	int vlen;
	pcie_range_t *msi_range;
	pcie_bus_res_t *busresp;

	busresp = &bio->busres;
	CHAIN_INIT(bio->dev_list);

	for (i = 0; i < num; ++i) {
		dev = hvnodes[i];
		if (md_get_prop_val(hvmdp, dev, "cfghandle",
		    &dev_cfg_hdl) != 0) {
			warning("HV MD physical device node has no "
			    "cfghandle property for 0x%"PRIx64, cfg_hdl);
			continue;
		}

		if (cfg_hdl != dev_cfg_hdl)
			continue;

		if (md_get_prop_val(hvmdp, dev, "resource_id", resid) != 0) {
			warning("HV MD physical device node has no "
			    "resource_id property for 0x%"PRIx64, cfg_hdl);
			continue;
		}

		bio->io_mmu_bypass = false;
		if (md_get_prop_val(hvmdp, dev, "allow_bypass", &bypass) == 0) {
			bio->io_mmu_bypass = (bypass == 1);
		}

		if (!pcie_is_dio_hvmd_supported()) {
			return;
		}

		get_assigned_pcie_devs(cfg_hdl, bio, hvmdp, dev);

		if (md_get_prop_val(hvmdp, dev, "#msi",
		    &busresp->num_msis) != 0) {
			warning("Missing #msi property for 0x%"PRIx64, cfg_hdl);
			continue;
		}

		if (md_get_prop_data(hvmdp, dev, "msi-ranges",
		    (uint8_t **)&msi_range, &rlen) != 0) {
			warning("Missing msi-ranges property "
			    "for 0x%"PRIx64, cfg_hdl);
			continue;
		}

		ASSERT((rlen / sizeof (pcie_range_t)) == 1);

		busresp->msi_range = *msi_range;

		if (md_get_prop_data(hvmdp, dev, "virtual-dma",
		    (uint8_t **)&busresp->dvma_ranges, &vlen) != 0) {
			warning("Missing virtual-dma property"
			    "for 0x%"PRIx64, cfg_hdl);
			continue;
		}

		busresp->num_dvma_ranges = vlen / sizeof (pcie_range_t);

		if (md_get_prop_val(hvmdp,
		    dev, "#msi-eqs", &busresp->num_eqs) != 0) {
			warning("Missing #msi-eqs property"
			    "for 0x%"PRIx64, cfg_hdl);
			continue;
		}
		break;
	}
	if (i == num)
		fatal("No HV MD node found for iodevice");
}
