/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2008 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)allocate.c	1.6	08/10/17 SMI"

#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include <stdlib.h>

#include "fatal.h"
#include "allocate.h"

/*
 * Simple allocation routines
 */

void *
xmalloc(int size, int linen, char *filen)
{
	void *p;

	if (size <= 0) {
		if (size < 0) {
			fatal("xmalloc of negative size (%d) at %d in %s",
			    size, linen, filen);
		}
		warning("xmalloc of zero size at %d in %s", linen, filen);
		return (NULL);
	}

	p = malloc(size);
	if (p == NULL)
		fatal("malloc of %d at %d in %s", size, linen, filen);

	return (p);
}

void *
xcalloc(int num, int size, int linen, char *filen)
{
	void *p;

	if (size <= 0 || num <= 0) {
		fatal("xcalloc(%d,%d) : one of number or size is <= 0 "
		    "at line %d of %s", num, size, linen, filen);
	}

	p = calloc(num, size);
	if (p == NULL) {
		fatal("calloc of %d of size %d at %d in %s",
		    num, size, linen, filen);
	}

	return (p);
}

void *
xrealloc(void *oldp, int size, int linen, char *filen)
{
	void *p;

	if (size <= 0) {
		if (size == 0) {
			free(oldp);
			warning("xrealloc to zero size at %d in %s",
			    linen, filen);
			return (NULL);
		}

		fatal("xrealloc to negative size %d at %d in %s",
		    size, linen, filen);
	}

	if (oldp == NULL) {
		p = malloc(size);
	} else {
		p = realloc(oldp, size);
	}
	if (p == NULL)
		fatal("xrealloc failed @ %d in %s", linen, filen);

	return (p);
}

char *
xstrdup(char *strp, int linen, char *filen)
{
	char *p;

	p = strdup(strp);
	if (p == NULL)
		fatal("xstrdup @ %d in %s failed", linen, filen);

	return (p);
}

/*
 * Duplicate a message while guaranteeing that it fits within
 * a given size and is properly terminated.  This function is
 * intended for use in displaying messages received from
 * untrusted sources.
 */
char *
xstrldup(char *strp, int maxlen, int linen, char *filen)
{
	char *p;
	size_t len;

	p = xmalloc(maxlen, linen, filen);

	strlcpy(p, strp, maxlen);

	len = strlen(p);

	/*
	 * If necessary, force final character of msg to be a newline.
	 * If there is room, extend the message to do this, otherwise
	 * overwrite the final character of the message provided.
	 */
	if (p[len - 1] != '\n')
		if (len == maxlen - 1)
			p[len - 1] = '\n';
		else {
			p[len++] = '\n';
			p[len] = '\0';
		}

	return (p);
}
