/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2008 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)nvlist.c	1.2	08/12/09 SMI"

#include <stdlib.h>
#include <string.h>

#include "allocate.h"
#include "basics.h"
#include "nvlist.h"
#include "response.h"

/*
 * Module implementing lists of name-value pairs
 */

void
nvlist_init(nvlist_t *nvl)
{
	nvl->size = 0;
	nvl->len = 0;
	nvl->pairs = NULL;
	nvl->ordered = true;
}

static void
free_nvpair(nvpair_t *nvp)
{
	if (nvp) {
		Xfree(nvp->name);
		if (nvp->value != NULL)
			Xfree(nvp->value);
	}
}

void
nvlist_clear(nvlist_t *nvl)
{
	int i;

	for (i = 0; i < nvl->len; i++)
		free_nvpair(nvl->pairs + i);
	free(nvl->pairs);

	nvl->pairs = NULL;
	nvl->len = 0;
	nvl->size = 0;
}

static int
nvpair_cmp(const void *a, const void *b)
{
	return (strcmp(((nvpair_t *)a)->name,
	    ((nvpair_t *)b)->name));
}

void
nvlist_sort(nvlist_t *nvl)
{
	qsort(nvl->pairs, nvl->len, sizeof (nvpair_t), nvpair_cmp);
	nvl->ordered = true;
}

nvpair_t *
nvlist_find(nvlist_t *nvl, char *name)
{
	int i;

	for (i = 0; i < nvl->len; i++)
		if (strcmp(nvl->pairs[i].name, name) == 0)
			return (nvl->pairs + i);
	return (NULL);
}

/*
 * Add a <'name', 'value'> pair to a name-value list'. If there is already a
 * pair in the store with the same name its value is replaced with the new
 * value. Returns true iff a change occurred.
 */
bool_t
nvlist_add_pair(nvlist_t *nvl, char *name, char *value)
{
	nvpair_t *nvp;

	if (value == NULL)
		value = "";	/* no value is treated as an empty string */

	nvp = nvlist_find(nvl, name);
	if (nvp == NULL) {
		/*
		 * The nvpair doesn't exist. Extend the nvpair array and
		 * add the nvpair at the end of the array.
		 */
		nvl->pairs = Xrealloc(nvl->pairs,
		    (nvl->len + 1) * sizeof (nvpair_t));
		nvp = nvl->pairs + nvl->len;
		nvl->len++;
		nvp->name = Xstrdup(name);
		/*
		 * Add the size of the new nvpair name. The nvpair's value
		 * and size are added below.
		 */
		nvl->size += strlen(name);

		nvl->ordered = false;
	} else {
		/*
		 * The nvpair already exists. Check if the value
		 * is the same in which case no change is made.
		 */
		if ((value == NULL && nvp->value == NULL) ||
		    (value != NULL && nvp->value != NULL) &&
		    strcmp(value, nvp->value) == 0) {
			return (false);
		}

		/*
		 * Value differs. Subtract the size of the
		 * current value which is being replaced and free it. The new
		 * (replacement) value and its size are added below.  No need
		 * to subtract and free if the value was NULL before.
		 */
		if (nvp->value != NULL) {
			nvl->size -= strlen(nvp->value);
			Xfree(nvp->value);
		}
	}

	if (value == NULL) {
		nvp->value = NULL;
	} else {
		nvp->value = Xstrdup(value);
		nvl->size += strlen(value);
	}

	return (true);
}

/*
 * Delete a nvpair from a nvlist.
 * Returns true iff a change occurred.
 */
bool_t
nvlist_delete_pair(nvlist_t *nvl, char *name)
{
	nvpair_t *nvp, *newpairs;
	int i, newlen = 0;

	if ((nvp = nvlist_find(nvl, name)) == NULL) {
		/* nothing to do */
		return (false);
	}

	/*
	 * The nvpair to delete is present in the array, so the array will
	 * need to shrink. First create a resized array.
	 */
	if (nvl->len == 1)
		newpairs = NULL;
	else
		newpairs = Xmalloc((nvl->len - 1) * sizeof (nvpair_t));

	/*
	 * The nvpair could be anywhere in the array. Copy
	 * the old array to the new array one element at a
	 * time, skipping the nvpair to delete.
	 */
	for (i = 0; i < nvl->len; i++) {
		if (nvp == &nvl->pairs[i]) {
			/* skip the nvpair to delete */
			continue;
		}

		newpairs[newlen++] = nvl->pairs[i];
	}

	ASSERT(newlen == (nvl->len - 1));

	/* update var store parameters */
	nvl->len--;
	nvl->size -= strlen(nvp->name);
	if (nvp->value != NULL)
		nvl->size -= strlen(nvp->value);

	/* clean up the deleted nvpair */
	free_nvpair(nvp);

	/* install the new array */
	Xfree(nvl->pairs);
	nvl->pairs = newpairs;

	return (true);
}

void
nvlist_print(nvlist_t *nvl, char *fmt, response_t *resp)
{
	if (!nvl->ordered)
		nvlist_sort(nvl);

	for (int i = 0; i < nvl->len; i++) {
		char *s;
		nvpair_t *nvp = &nvl->pairs[i];

		if (nvp->value != NULL &&
		    strcmp(nvp->name, "security-password") != 0)
			s = nvp->value;
		else
			s = "";

		rprintf(resp, fmt, nvp->name, s);
	}
}
