/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2008 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)resource_id.c	1.11	08/09/24 SMI"

#include <strings.h>

#include "basics.h"
#include "allocate.h"
#include "chain.h"
#include "fatal.h"
#include "ldom.h"
#include "resource_id.h"
#include "zprintf.h"

#define	MIN_FILL 8

/*
 * Identifier resource pool.
 */

void
create_res_pool(res_pool_t *resp, ident_val_t minid, ident_val_t maxid,
    res_flags_t flags)
{
	ASSERT(minid <= maxid);

	resp->minid = minid;
	resp->maxid = maxid;
	resp->flags = flags;
	CHAIN_INIT(resp->free_list);
	CHAIN_INIT(resp->free_staging);
	resp->nextid = minid;
	memset(resp->name, 0, sizeof (resp->name));
}

void
name_res_pool(res_pool_t *resp, char *name)
{
	(void) strlcpy(resp->name, name, sizeof (resp->name));
}

void
name_ldom_res_pool(res_pool_t *resp, ldom_t *ldp, char *name)
{
	char *poolname;

	poolname = zsprintf("%s %s", ldp->namep, name);
	name_res_pool(resp, poolname);
	Xfree(poolname);
}

void
name_ldom_vdev_res_pool(res_pool_t *resp, ldom_t *ldp, char *vdev,
    char *vdev_name, char *name)
{
	char *poolname;

	poolname = zsprintf("%s %s %s %s", ldp->namep, vdev, vdev_name, name);
	name_res_pool(resp, poolname);
	Xfree(poolname);
}

static void
free_res_internal(resid_t *rp)
{
	res_pool_t *resp = rp->pool;

	/* Dumb check for possible sorting */
	/* Work on the assumption that the resource is unique ! */
	if (CHAIN_LENGTH(resp->free_list) > 0 &&
	    (resp->flags & Res_sorted) != 0) {
		resid_t *ep, *ip;
		ep = CHAIN_END(resp->free_list);
		if (rp->id > ep->id)
			goto at_end;
		ep = NULL;

		CHAIN_ITER(resp->free_list, ip, link) {
			if (rp->id < ip->id)
				break;
			ep = ip;
		}
		if (ep == NULL) {
			CHAIN_ADD_FIRST(resp->free_list, rp, link);
		} else {
			CHAIN_INSERT_AFTER(resp->free_list, ep, rp, link);
		}
		return;
	}
at_end:;
	CHAIN_ADD(resp->free_list, rp, link);
}

void
free_res(resid_t *rp)
{
	DBG5(zprintf("Freeing res %"PRIu64" %s\n", rp->id, rp->pool->name));
	free_res_internal(rp);
}

static void
fill_res(res_pool_t *resp)
{
	int count;

	for (count = MIN_FILL; count > 0 && resp->nextid <= resp->maxid;
	    count--) {
		resid_t *rp;

		rp = Xcalloc(1, resid_t);
		rp->id = resp->nextid++;
		rp->pool = resp;
		free_res_internal(rp);
	}
}

resid_t *
alloc_res(res_pool_t *resp)
{
	resid_t *rp;

again:;
	rp = CHAIN_START(resp->free_list);
	if (rp == NULL) {
		if (resp->nextid > resp->maxid) {
			DBG5(zprintf("Resource alloc failed %s\n", resp->name));
			return (NULL);
		}
		fill_res(resp);
		goto again;
	}

	CHAIN_UNLINK(resp->free_list, rp, link);
	DBG5(zprintf("Allocated res %"PRIu64" %s\n", rp->id, resp->name));
	return (rp);
}

/*
 * Stage resource for later freeing
 */
void
stage_res(resid_t *rp)
{
	DBG(zprintf("Staging res %"PRIu64" %s\n", rp->id, rp->pool->name));
	CHAIN_ADD(rp->pool->free_staging, rp, stage_link);
}

/*
 * Free everything on staging list
 */
void
free_from_stage(res_pool_t *resp)
{
	resid_t *rp, *rp_next;

	CHAIN_ITER_VOL(resp->free_staging, rp, rp_next, stage_link) {
		DBG(zprintf("Freeing staged res %"PRIu64" %s\n",
		    rp->id, resp->name));
		CHAIN_UNLINK(resp->free_staging, rp, stage_link);
		free_res_internal(rp);
	}
}

static resid_t *
pull_res_id_internal(res_pool_t *resp, ident_val_t id)
{
	resid_t *rp;

	/* early termination if requested id is out of range */
	if (id > resp->maxid) {
		return (NULL);
	}

	/* Check to see if it's been allocated yet */
	while (resp->nextid <= id)
		fill_res(resp);

	CHAIN_ITER(resp->free_list, rp, link) {
		if (rp->id == id) {
			CHAIN_UNLINK(resp->free_list, rp, link);
			return (rp);
		}
	}

	return (NULL);
}

resid_t *
pull_res_id(res_pool_t *resp, ident_val_t id)
{
	resid_t *rp;

	rp = pull_res_id_internal(resp, id);

	if (rp == NULL) {
		DBG5(zprintf("Pull failed res %"PRIu64" %s\n", id, resp->name));
	} else {
		DBG5(zprintf("Pulled res %"PRIu64" %s\n", id,
		    resp->name));
	}

	return (rp);
}

ident_val_t
available_res_ids(res_pool_t *resp)
{
	return (resp->maxid - resp->nextid + 1 + CHAIN_LENGTH(resp->free_list));
}

ident_val_t
used_res_ids(res_pool_t *resp)
{
	return (resp->nextid - resp->minid - CHAIN_LENGTH(resp->free_list));
}

/*
 * Check if a particular resource is available for allocation
 * without actually allocating it. Returns 1 (true) for success,
 * 0 (false) for failure.
 */
int
is_res_available(res_pool_t *resp, ident_val_t id)
{
	resid_t *rp;

	/*
	 * Check if out of range
	 */
	if ((id > resp->maxid) || (id < resp->minid))
		return (0);

	/*
	 * Check if the resource is in the not yet allocated
	 * list
	 */
	if ((id <= resp->maxid) && (id >= resp->nextid)) {
		return (1);
	}

	/*
	 * Else it might be in the list of free resource ids
	 */
	CHAIN_ITER(resp->free_list, rp, link) {
		if (rp->id == id)
			return (1);
	}

	/*
	 * Not found anywhere
	 */
	return (0);
}

void
destroy_res_pool(res_pool_t *resp)
{
	ident_val_t i;

	/*
	 * Quick sanity check that everything allocated is back on the free
	 * list ready for release.
	 */

	for (i = resp->minid; i < resp->nextid; i++) {
		resid_t *rp;

		rp = pull_res_id_internal(resp, i);
		ASSERT(rp != NULL);
		Xfree(rp);
	}
}
