/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)response.c	1.14	08/10/29 SMI"

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>	/* isprint() */

#include "allocate.h"
#include "basics.h"
#include "cli_conn.h"
#include "ldm_protocol.h"
#include "response.h"

#define	RESPONSE_BLOCK	1024	/* always a power of 2 */

/*
 * This routine returns the amount of additional space to expand any
 * control characters found.  It returns zero if no control chars found.
 */
static int
count_ctl(char *buf)
{
	int cnt = 0;

	for (; *buf; buf++) {
		if (!isprint(*buf) && (*buf != '\n') && (*buf != '\t')) {
			cnt += 3;	/* we need to add 3 more spaces */
		}
	}
	return (cnt);
}


/*
 * This routine expands any control characters found into the form:
 * \xxx where xxx is the zero filled octal expansion of the control character.
 */
static void
expand_ctl(char *data_end, char *buf_end)
{
	char respchar;

	/*
	 * Start at the end of the buffer and the end of the message and
	 * copy last characater of the message to last characater in buffer
	 * until either a control character is found or until the two pointers
	 * are equal (no more control characters left).  When a control
	 * character is found, expand it.
	 */
	while (data_end-- < buf_end--) {
		respchar = *data_end;
		if (isprint(respchar) || respchar == '\n' || respchar == '\t') {
			*buf_end = respchar;
		} else {
			/*
			 * expand the numeric portion into zero filled octal
			 */
			for (int j = 3; j > 0; j--) {
				*buf_end-- = respchar % 8 + '0';
				respchar /= 8;
			}
			/*
			 * set the \ char, but don't move the ptr as it will
			 * be done in the while.
			 */
			*buf_end = '\\';
		}
	}
}


int
rprintf(response_t *resp, char *fmt, ...)
{
	int gap, len;
	int olen, cnt;

	va_list args;

	if (resp == NULL)
		return (0);

	va_start(args, fmt);

again:;
	gap = resp->space - resp->length;

	len = vsnprintf(resp->bufp + resp->length, gap, fmt, args);
	olen = len;

	cnt = count_ctl(resp->bufp + resp->length);
	len += cnt;

	if (len > gap-1) {
		int amt;
		amt = len - (gap-1);
		amt = (amt + RESPONSE_BLOCK -1) & ~(RESPONSE_BLOCK-1);

		resp->space += amt;
		resp->bufp = Xrealloc(resp->bufp, resp->space);
		goto again;
	}

	if (cnt != 0) {
		/*
		 * We found some control characters.  Expand them inline and
		 * then add a NUL terminator as the previously written one is
		 * overwritten by the expansion.
		 */
		expand_ctl(resp->bufp + resp->length + olen,
		    resp->bufp + resp->length + len);
		*(resp->bufp + resp->length + len) = '\0';
	}

	va_end(args);

	resp->length += len;

	return (len);
}

/*
 * "Print" the input string into the response, guaranteeing that it
 * is null-terminated within the maximum length (maxlen).  Note:
 * Xstrldup guarantees that the destination ends with '\n', '\0'.
 *
 * Returns: length of string copied into the response buf.  Similar
 * to puts(3c).
 */
int
rputs_safe(response_t *resp, char *s, size_t maxlen)
{
	char	*buf;
	int	len;

	if (s == NULL)
		return (0);

	buf = Xstrldup(s, maxlen);
	rprintf(resp, buf);

	len = strlen(buf);
	len = (len > maxlen) ? maxlen : len;

	Xfree(buf);

	return (len);
}

void
init_response(response_t **resp)
{
	int offset;

	*resp = Xmalloc(sizeof (response_t));
	(*resp)->ret = 0;
	(*resp)->space = RESPONSE_BLOCK;
	(*resp)->bufp = Xmalloc((*resp)->space);

	/*
	 * Leave space at beginning of the response for the LDM protocol
	 * headers and the response length, condition code and text block
	 * length. These will be known and filled in when the response is
	 * sent.
	 */
	offset = RESPONSE_CONTENT_OFFSET;
	ASSERT(RESPONSE_BLOCK > offset);
	*((*resp)->bufp + offset) = '\0';
	(*resp)->length = offset;
	(*resp)->marker = offset;	/* initial mark at beginning of text */
}

void
fini_response(response_t *resp)
{
	if (resp == NULL)
		return;

	Xfree(resp->bufp);
	Xfree(resp);
}

void
mark_response(response_t *resp)
{
	resp->marker = resp->length;
}

void
reset_response(response_t *resp)
{
	resp->length = resp->marker;
	*(resp->bufp + resp->length) = '\0';
}

/*
 * This routine returns the string stored in a response_t structure.
 * It allows for the dumping of error message to other formats (e.g. XML)
 * or to be transferred to other machines (e.g. via a mig_msg)
 *
 * NOTE: The caller is responsible for freeing the memory allocated to
 * store the string returned.
 */
char *
get_response_text(response_t *resp)
{
	ldm_op_reply_t		*reply;
	cli_reply_data_t	*data_info;

	if (resp == NULL)
		return (NULL);

	ASSERT(resp->bufp != NULL);
	reply = (ldm_op_reply_t *)(resp->bufp + sizeof (ldm_hdr_t));

	ASSERT(reply != NULL);
	data_info = (cli_reply_data_t *)reply->rp_data;

	if (data_info == NULL || data_info->text[0] == '\0')
		return (NULL);

	return (Xstrdup(data_info->text));
}
