/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)ssl_connection.c	1.3	09/05/20 SMI"

#include <dlfcn.h>
#include <thread.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>

#include "allocate.h"
#include "basics.h"
#include "fatal.h"
#include "ssl_connection.h"
#include "zprintf.h"

mutex_t	*ssl_lock_array;
long	*ssl_lock_count;
SSL_CTX *server_ssl_ctx = NULL;
SSL_CTX	*client_ssl_ctx = NULL;

/*
 * pointers to functions needed from libssl.
 */
SSL_load_error_strings_pt	SSL_load_error_strings_f = NULL;
SSLv23_server_method_pt		SSLv23_server_method_f = NULL;
SSL_CTX_use_PrivateKey_file_pt	SSL_CTX_use_PrivateKey_file_f = NULL;
SSL_get_error_pt		SSL_get_error_f = NULL;
ERR_get_error_pt		ERR_get_error_f = NULL;
ERR_error_string_pt		ERR_error_string_f = NULL;
SSL_accept_pt			SSL_accept_f = NULL;
SSL_CTX_use_certificate_chain_file_pt
SSL_CTX_use_certificate_chain_file_f = NULL;

SSL_library_init_pt		SSL_library_init_f = NULL;
SSL_CTX_new_pt			SSL_CTX_new_f = NULL;
SSLv23_client_method_pt		SSLv23_client_method_f = NULL;
SSL_write_pt			SSL_write_f = NULL;
RAND_seed_pt			RAND_seed_f = NULL;
SSL_connect_pt			SSL_connect_f = NULL;
SSL_set_fd_pt			SSL_set_fd_f = NULL;
SSL_free_pt			SSL_free_f = NULL;
SSL_read_pt			SSL_read_f = NULL;
SSL_new_pt			SSL_new_f = NULL;
SSL_get_SSL_CTX_pt		SSL_get_SSL_CTX_f = NULL;
SSL_CTX_free_pt			SSL_CTX_free_f = NULL;
CRYPTO_set_locking_callback_pt	CRYPTO_set_locking_callback_f = NULL;
CRYPTO_num_locks_pt		CRYPTO_num_locks_f = NULL;
CRYPTO_set_id_callback_pt	CRYPTO_set_id_callback_f = NULL;
ENGINE_by_id_pt			ENGINE_by_id_f = NULL;
ENGINE_set_default_ciphers_pt	ENGINE_set_default_ciphers_f = NULL;
ENGINE_load_pk11_pt		ENGINE_load_pk11_f = NULL;
ENGINE_init_pt			ENGINE_init_f = NULL;

void *ssl_dl = NULL;

/*
 * This routine will run through the first time we get a remote XMPP
 * connection. After that we will not need to do this again. It cannot be run
 * from main at start as we need to alert remote users if the TLS handshake
 * failed.
 */
int
load_SSL_lib(void)
{
	int	ret = 0;

	/* If we have already opened the library no need to do it again. */
	if (ssl_dl != NULL)
		return (0);

	if ((ssl_dl = dlopen("libssl.so", RTLD_NOW)) == NULL) {
		zprintf("libssl.so could not be loaded: %s", dlerror());
		return (-1);
	}

	FUNCTION_ADD(SSL_load_error_strings_f, SSL_load_error_strings_pt,
	    ssl_dl, "SSL_load_error_strings", ret);
	FUNCTION_ADD(SSL_library_init_f, SSL_library_init_pt, ssl_dl,
	    "SSL_library_init", ret);
	FUNCTION_ADD(SSL_CTX_new_f, SSL_CTX_new_pt, ssl_dl,
	    "SSL_CTX_new", ret);
	FUNCTION_ADD(SSLv23_server_method_f, SSLv23_server_method_pt, ssl_dl,
	    "SSLv23_server_method", ret);
	FUNCTION_ADD(SSLv23_client_method_f, SSLv23_client_method_pt, ssl_dl,
	    "SSLv23_client_method", ret);
	FUNCTION_ADD(SSL_write_f, SSL_write_pt, ssl_dl, "SSL_write", ret);
	FUNCTION_ADD(SSL_CTX_use_PrivateKey_file_f,
	    SSL_CTX_use_PrivateKey_file_pt, ssl_dl,
	    "SSL_CTX_use_PrivateKey_file", ret);
	FUNCTION_ADD(RAND_seed_f, RAND_seed_pt, ssl_dl, "RAND_seed", ret);
	FUNCTION_ADD(SSL_get_error_f, SSL_get_error_pt, ssl_dl,
	    "SSL_get_error", ret);
	FUNCTION_ADD(ERR_get_error_f, ERR_get_error_pt, ssl_dl,
	    "ERR_get_error", ret);
	FUNCTION_ADD(ERR_error_string_f, ERR_error_string_pt, ssl_dl,
	    "ERR_error_string", ret);
	FUNCTION_ADD(SSL_accept_f, SSL_accept_pt, ssl_dl, "SSL_accept", ret);
	FUNCTION_ADD(SSL_connect_f, SSL_connect_pt, ssl_dl, "SSL_connect",
	    ret);
	FUNCTION_ADD(SSL_CTX_use_certificate_chain_file_f,
	    SSL_CTX_use_certificate_chain_file_pt, ssl_dl,
	    "SSL_CTX_use_certificate_chain_file", ret);
	FUNCTION_ADD(SSL_set_fd_f, SSL_set_fd_pt, ssl_dl, "SSL_set_fd", ret);
	FUNCTION_ADD(SSL_get_fd_f, SSL_get_fd_pt, ssl_dl, "SSL_get_fd", ret);
	FUNCTION_ADD(SSL_free_f, SSL_free_pt, ssl_dl, "SSL_free", ret);
	FUNCTION_ADD(SSL_read_f, SSL_read_pt, ssl_dl, "SSL_read", ret);
	FUNCTION_ADD(SSL_new_f, SSL_new_pt, ssl_dl, "SSL_new", ret);
	FUNCTION_ADD(SSL_get_SSL_CTX_f, SSL_get_SSL_CTX_pt, ssl_dl,
	    "SSL_get_SSL_CTX", ret);
	FUNCTION_ADD(SSL_CTX_free_f, SSL_CTX_free_pt, ssl_dl,
	    "SSL_CTX_free", ret);
	FUNCTION_ADD(CRYPTO_set_locking_callback_f,
	    CRYPTO_set_locking_callback_pt, ssl_dl,
	    "CRYPTO_set_locking_callback", ret);
	FUNCTION_ADD(CRYPTO_num_locks_f, CRYPTO_num_locks_pt, ssl_dl,
	    "CRYPTO_num_locks", ret);
	FUNCTION_ADD(CRYPTO_set_id_callback_f, CRYPTO_set_id_callback_pt,
	    ssl_dl, "CRYPTO_set_id_callback", ret);

	FUNCTION_ADD(ENGINE_load_pk11_f, ENGINE_load_pk11_pt, ssl_dl,
	    "ENGINE_load_pk11", ret);
	FUNCTION_ADD(ENGINE_by_id_f, ENGINE_by_id_pt, ssl_dl,
	    "ENGINE_by_id", ret);
	FUNCTION_ADD(ENGINE_init_f, ENGINE_init_pt, ssl_dl, "ENGINE_init",
	    ret);
	FUNCTION_ADD(ENGINE_set_default_ciphers_f,
	    ENGINE_set_default_ciphers_pt, ssl_dl,
	    "ENGINE_set_default_ciphers", ret);

	if (ret < 0)
		return (-1);
	else
		return (0);
}

/*ARGSUSED2*/
static void
ssl_locking_cb(int mode, int type, char *file, int line)
{
	if (mode & CRYPTO_LOCK) {
		(void) mutex_lock(&(ssl_lock_array[type]));
		ssl_lock_count[type]++;
	} else {
		(void) mutex_unlock(&(ssl_lock_array[type]));
	}
}

static unsigned long
ssl_thread_id(void)
{
	return ((unsigned long)thr_self());
}

static int
ldmd_setup_server_ctx(void)
{
	int		rv;
	unsigned long	err_err;


	server_ssl_ctx = SSL_CTX_new_f(SSLv23_server_method_f());
	if (server_ssl_ctx == NULL) {
		DBG(zprintf("SSL_CTX_new FAILED\n"));
		warning("Failed to create a new SSL CTX");
		return (-1);
	}

	rv = SSL_CTX_use_certificate_chain_file_f(server_ssl_ctx, SERVER_CRT);
	if (rv == 1) {
		DBG(zprintf("CERTIFICATE SET\n"));
	} else {
		SSL_CTX_free_f(server_ssl_ctx);
		DBG(zprintf("use_certificate_chain_file failed\n"));
		err_err = ERR_get_error_f();
		DBG(zprintf("%s\n", ERR_error_string_f(err_err, NULL)));
		warning("Failed to load certificate chain from file");
		return (-1);
	}

	rv = SSL_CTX_use_PrivateKey_file_f(server_ssl_ctx, SERVER_KEY,
	    SSL_FILETYPE_PEM);
	if (rv == 1) {
		DBG(zprintf("PRIVATE KEY SET\n"));
	} else {
		SSL_CTX_free_f(server_ssl_ctx);
		DBG(zprintf("use_PrivateKey_file failed\n"));
		err_err = ERR_get_error_f();
		DBG(zprintf("%s\n", ERR_error_string_f(err_err, NULL)));
		warning("Failed to add the first private key found in file");
		return (-1);
	}

	return (0);
}

static int
ldmd_setup_client_ctx(void)
{
	client_ssl_ctx = SSL_CTX_new_f(SSLv23_client_method_f());
	if (client_ssl_ctx == NULL) {
		DBG(zprintf("CTX_new called failed\n"));
		return (-1);
	}
	return (0);
}

static void
ssl_setup_engine(void)
{
	ENGINE	*e;

	/* try and initialize the hw */
	ENGINE_load_pk11_f();

	/* get structural reference */
	if ((e = ENGINE_by_id_f(PKCS11_ENGINE)) == NULL) {
		DBG(zprintf("%s engine does not exist. Using software "
		    "engine.\n", PKCS11_ENGINE));
		return;
	}

	/* get functional reference */
	if ((ENGINE_init_f(e) == 0) || (!ENGINE_set_default_ciphers_f(e))) {
		DBG(zprintf("No hardware crypto resources available. Using "
		    "software engine.\n"));
		return;
	}

	DBG(zprintf("Hardware crypto resources initialized.\n"));
}

int
ldmd_setup_SSL(void)
{
	int		rv, urand_fd, i;
	char		rand_buf[RAND_BUF_SIZE];

	rv = load_SSL_lib();
	if (rv == -1) {
		warning("load_SSL_lib failed\n");
		return (-1);
	}

	ssl_setup_engine();

	/*
	 * Set up SSL for thread-safe operation.
	 */
	ssl_lock_array = Xmalloc(CRYPTO_num_locks_f() * sizeof (mutex_t));
	if (ssl_lock_array == NULL) {
		warning("SSL_lock_array malloc failed");
		return (-1);
	}

	ssl_lock_count = Xmalloc(CRYPTO_num_locks_f() * sizeof (long));
	if (ssl_lock_count == NULL) {
		warning("SSL_lock_count malloc failed");
		return (-1);
	}

	for (i = 0; i < CRYPTO_num_locks_f(); i++) {
		ssl_lock_count[i] = 0;
		(void) mutex_init(&ssl_lock_array[i], USYNC_THREAD, NULL);
	}

	CRYPTO_set_id_callback_f((unsigned long (*) ())ssl_thread_id);
	CRYPTO_set_locking_callback_f((void (*) ())ssl_locking_cb);

	urand_fd = open("/dev/urandom", O_RDONLY);
	if (urand_fd == -1) {
		warning("Opening /dev/urandom failed\n");
		return (-1);
	}
	read(urand_fd, rand_buf, RAND_BUF_SIZE);
	close(urand_fd);

	SSL_load_error_strings_f();
	SSL_library_init_f();

	RAND_seed_f(rand_buf, RAND_BUF_SIZE);

	if (ldmd_setup_server_ctx() != 0) {
		warning("server CTX setup failed");
		return (-1);
	}

	if (ldmd_setup_client_ctx() != 0) {
		warning("client CTX setup failed");
		return (-1);
	}

	return (0);
}
