/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <thread.h>
#include <signal.h>
#include "fatal.h"
#include "ldom.h"
#include "backend.h"
#include "sequencer.h"
#include "vcpu.h"
#include "zprintf.h"
#include "pm_events.h"
#include "pm/pm_shared.h"
#include "option.h"
#include "policy.h"


/* Used for time conversion */
#define	NSEC_PER_MSEC	1000000L

/* Thread ID of the guest utilization poller thread */
static thread_t guest_util_poll_tid;

/* Should the guest utilization poller thread be stopped? */
static bool_t guest_util_poll_stop;

/* Is the guest utilization poller thread running? */
static bool_t guest_util_poll_running = false;

/* To wait for the sequencer to finish a utilization read */
static cond_t guest_util_poll_cv = DEFAULTCV;

/* Predicate for guest_util_poll_cv */
static bool_t guest_util_poll_seq;

/* Lock to protect guest_util_poll* variables */
static mutex_t guest_util_poll_lock = DEFAULTMUTEX;

/*
 * Calculate and return the utilization for the given guest util struct.
 */
static double
calc_guest_util(rs_guest_util_t *gutil, int ncpu, int fcpu)
{
	uint64_t used_cycles, total_cycles, yielded_cycles;

	/* Don't include the cycles of faulted cpus */
	total_cycles = (gutil->active_delta * (ncpu - fcpu)) -
	    gutil->stopped_cycles;

	/*
	 * Don't include the cycles consumed by faulted cpus.  This logic
	 * assumes that faulted cpus accrue as yielded cycles.
	 * Also, a simplfying assumption is being made that the cpu has been
	 * faulted for the entire active_delta time interval.
	 */
	if (gutil->yielded_cycles > (gutil->active_delta * fcpu))
		yielded_cycles = gutil->yielded_cycles -
		    (gutil->active_delta * fcpu);
	else
		yielded_cycles = 0;

	/*
	 * There is a small chance that yielded_cycles > total_cycles
	 * in which case clamp to the min limit.
	 */
	if (yielded_cycles >= total_cycles)
		return (0.0);

	used_cycles = total_cycles - yielded_cycles;

	return ((100.0 * (double)used_cycles) / (double)(total_cycles));
}

/*
 * Get the guest utilization from HV for the given guest and
 * return the utilization.
 */
static double
get_guest_util(ldom_t *ldp)
{
	rs_guest_util_t gutil;
	double util;
	int ncpu, fcpu;

	if (backend.get_guest_utilization(ldp, &gutil) == 0) {
		ncpu = ldom_num_bound_vcpus(ldp);
		fcpu = ldom_num_faulted_vcpus(ldp);
		util = calc_guest_util(&gutil, ncpu, fcpu);
		ldp->stats.uptime = gutil.lifespan / pri.stick_frequency;
	} else {
		util = 0.0;
	}

	return (util);
}

/*
 * Calculate and return the utilization for the given vcpu state.
 */
static double
calc_vcpu_util(rs_vcpu_state_t *vstate)
{
	uint64_t used_cycles;

	used_cycles = vstate->active_delta - vstate->yielded_cycles;

	return ((100.0 * (double)used_cycles) / (double)(vstate->active_delta));
}

/*
 * Get the vcpu state from HV for the given vcpu and return the utilization.
 */
static double
get_vcpu_util(vcpu_t *vcpu)
{
	rs_vcpu_state_t vstate;
	double util;

	/* Need to pass the resource ID, not the virtual ID */
	if (backend.get_vcpu_state(vcpu->cpu->res_id, &vstate) == 0)
		util = calc_vcpu_util(&vstate);
	else
		util = 0.0;

	return (util);
}

/*
 * To be called when PM begins or resumes cpu utilization notifications.
 * Reset the guest utilization monitor data for every guest.
 */
void
reset_cpu_util_data(pm_util_event_info_t *ep)
{
	ldom_t *ldp;

	mutex_lock(&zeus.mlock);

	/* Loop for every guest */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		ldp->pm_cpu_resutils.ldmr_util_dir = PM_UTIL_WITHIN;
		ldp->pm_cpu_resutils.ldmr_time = 0;
	}

	mutex_unlock(&zeus.mlock);
}

/*
 * Monitor utilization for the given high/low thresholds.
 */
static pm_util_dir_t
guest_util_monitor(void *srcp, pm_res_type_t res_type, double util,
    hrtime_t current_time, double lo_wm, double hi_wm, uint32_t lo_ms,
    uint32_t hi_ms)
{
	int prev_idx;
	uint32_t rd_diff, wm_diff;
	pm_util_dir_t util_dir = PM_UTIL_WITHIN;
	ldm_resutils_t *resutils;
	hrtime_t hist_time;

	switch (res_type) {
	case PM_RES_CPU:
		resutils = &((ldom_t *)srcp)->pm_cpu_resutils;
		prev_idx =
		    GET_PREV_GUEST_UTIL_IDX(((ldom_t *)srcp)->stats.hist_idx);
		hist_time = ((ldom_t *)srcp)->stats.hist_table[prev_idx].time;
		break;

	case PM_RES_MEM:
		resutils = &((pm_mem_t *)srcp)->pm_mem_resutils;
		prev_idx =
		    GET_PREV_MEM_UTIL_IDX(((pm_mem_t *)srcp)->stats.hist_idx);
		hist_time = ((pm_mem_t *)srcp)->stats.hist_table[prev_idx].time;
		break;

	default:
		fatal("Unsupported resource!");
		break;
	}

	DBG5(zprintf("(%s %"PRIx64"): time= 0x%"PRIx64", util= %.2f, lo_wm= "
	    "%.2f, hi_wm= %.2f, lo_ms= %x, hi_ms= %x\n", ((res_type == 0) ?
	    "Domain" : "Mem"), ((res_type == 0) ? ((ldom_t *)srcp)->idp->id :
	    ((pm_mem_t *)srcp)->pm_mem_resid.id), current_time, util, lo_wm,
	    hi_wm, lo_ms, hi_ms));

	if (util > hi_wm) {
		if (resutils->ldmr_util_dir != PM_UTIL_HIGH) {
			resutils->ldmr_util_dir = PM_UTIL_HIGH;
			resutils->ldmr_time = current_time;
		}

		/* calculate the time diff since the last util was read */
		if (hist_time != 0)
			rd_diff = (current_time - hist_time) / NSEC_PER_MSEC;
		else
			rd_diff = 0;	/* no prev utilization data */

		/* calculate the time diff since util first crossed high_wm */
		wm_diff = (current_time - resutils->ldmr_time) / NSEC_PER_MSEC;
		if ((rd_diff > hi_ms) || (wm_diff > hi_ms)) {
			DBG5(zprintf("utilization above HIGH threshold: "
			    "low=%.2f, high=%.2f, util=%.2f\n", lo_wm, hi_wm,
			    util));
			util_dir = PM_UTIL_HIGH;
		}

	} else if (util < lo_wm) {
		if (resutils->ldmr_util_dir != PM_UTIL_LOW) {
			resutils->ldmr_util_dir = PM_UTIL_LOW;
			resutils->ldmr_time = current_time;
		}

		/* calculate the time diff since the last util was read */
		if (hist_time != 0)
			rd_diff = (current_time - hist_time) / NSEC_PER_MSEC;
		else
			rd_diff = 0;	/* no prev utilization data */


		/* calculate the time diff since util first crossed low_wm */
		wm_diff = (current_time - resutils->ldmr_time) / NSEC_PER_MSEC;
		if ((rd_diff > lo_ms) || (wm_diff > lo_ms)) {
			DBG5(zprintf("utilization below LOW threshold: "
			    "util=%.2f, low=%.2f, high=%.2f\n", util,
			    lo_wm, hi_wm));
			util_dir = PM_UTIL_LOW;
		}

	} else {
		/* utilization is within high & low watermarks */
		resutils->ldmr_util_dir = PM_UTIL_WITHIN;
		resutils->ldmr_time = 0;
	}

	return (util_dir);
}

/*
 * Handle memory utilization.
 */
static void
handle_memory_util()
{
	pm_util_event_info_t *eventp = NULL;
	rs_mem_state_t state;
	uint64_t diff, delta, within;
	double diff_percent;
	uint64_t delta_time;
	pm_mem_t *pmp;
	pm_mem_t **pmpp = NULL;
	int prev_idx;
	int pmp_cnt = 0;

	mutex_lock(&pm_util_list_lock);

	/* Find the PM_RES_MEM utilization event registration */
	CHAIN_ITER(pm_util_list.util_event_list, eventp, link) {
		if (eventp->res_type == PM_RES_MEM)
			break;
	}

	/* Nothing to do if there is no entry */
	if (!eventp) {
		mutex_unlock(&pm_util_list_lock);
		return;
	}

	/*
	 * Walk through memory resources and determine their utilization state.
	 */
	CHAIN_ITER(pm_mem_list.mem_list, pmp, link) {

		/*
		 * Get performnace counter and timestamp values from
		 * Hypervisor.
		 */
		if (backend.get_mem_state(&pmp->pm_mem_resid, &state) != 0) {
			DBG5(zprintf("failed to get mem status!\n"));
			continue;
		}

		/* First read establishes the baseline. */
		if (pmp->stats.last_perf_cntr == 0) {
			pmp->stats.last_perf_cntr = state.perf_cntr;
			ADD_MEM_UTIL(pmp, 0.0, state.sample_time);
			continue;
		}

		/* Delta time in msec since last sampling. */
		prev_idx = GET_PREV_MEM_UTIL_IDX(pmp->stats.hist_idx);
		delta_time = ((state.sample_time -
		    pmp->stats.hist_table[prev_idx].time) / NSEC_PER_MSEC);

		/* Delta in performance counter readings per msec */
		diff = state.perf_cntr - pmp->stats.last_perf_cntr;
		diff /= delta_time;

		/*
		 * Adjust highest & lowest values that we have recorded so far.
		 */
		if (pmp->stats.highest == 0)  {
			ASSERT(pmp->stats.lowest == 0);
			pmp->stats.highest = diff;
			pmp->stats.lowest = diff;
			ADD_MEM_UTIL(pmp, 0.0, state.sample_time);
			continue;
		} else if (pmp->stats.highest < diff) {
			pmp->stats.highest = diff;
		} else if (pmp->stats.lowest > diff) {
			pmp->stats.lowest = diff;
		}

		/*
		 * Based on current highest & lowest which have been
		 * recorded to this point, determine where this reading
		 * sits percentage-wise.
		 */
		delta = pmp->stats.highest - pmp->stats.lowest;
		within = diff - pmp->stats.lowest;
		diff_percent = ((double)within * 100.0) / (double)delta;

		DBG5(zprintf("%s: lowest= 0x%"PRIx64", highest= 0x%"PRIx64", "
		    "diff= 0x%"PRIx64", percentage=%.4f\n", __func__,
		    pmp->stats.lowest, pmp->stats.highest, diff, diff_percent));

		/* Record the reading. */
		pmp->stats.last_perf_cntr = state.perf_cntr;
		ADD_MEM_UTIL(pmp, diff_percent, state.sample_time);

		/* Determine whether we are within bound or out of bound. */
		if (guest_util_monitor((void *)pmp, eventp->res_type,
		    diff_percent, (hrtime_t)state.sample_time, eventp->low_wm,
		    eventp->high_wm, eventp->time_low_ms,
		    eventp->time_high_ms) != PM_UTIL_WITHIN) {
			pmp_cnt++;
			pmpp = Xrealloc(pmpp, (sizeof (*pmpp) * pmp_cnt));
			pmpp[pmp_cnt - 1] = pmp;
		}
	}

	/* If any of memory resources are out of bound, notify PM. */
	if (pmp_cnt > 0) {
		eventp->active = false;
		pm_util_change_notify(eventp->handle, (srcid_t)pmpp, pmp_cnt,
		    PM_RES_MEM);
		Xfree(pmpp);
	}

	mutex_unlock(&pm_util_list_lock);
}

/*
 * Calculate the average memory utilization in the history table.
 */
double
avg_mem_util(pm_mem_t *pmp, hrtime_t current_time)
{
	int i = pmp->stats.hist_idx;
	int num_values = 0;
	double total_util = 0;
	hrtime_t util_time;

	/*
	 * Loop through the utilization history table to total the values.
	 * Only include the values that are within MEM_UTIL_HIST_TIME.
	 * It's possible, though very unlikely, that no values are within
	 * MEM_UTIL_HIST_TIME.  In this case, just use the most recent value
	 * for total_util.
	 */
	do {
		ASSERT(i <= MEM_UTIL_TAB_NROWS);
		total_util += pmp->stats.hist_table[i].value;
		num_values++;

		/* Move to the next most recent entry in the history table */
		if (i == 0)
			i = MEM_UTIL_TAB_NROWS - 1;
		else
			i--;

		util_time = pmp->stats.hist_table[i].time;
	} while (((current_time - util_time) < MEM_UTIL_HIST_TIME) &&
	    (num_values < MEM_UTIL_TAB_NROWS));

	return (total_util / num_values);
}

/*
 * Handle getting guest and vcpu utilization values for the poller thread.
 * Note : this routine runs on the sequencer.
 */
static void
handle_guest_util_read(bool_t do_vcpus)
{
	pm_util_event_info_t *ep = NULL;
	ldom_t *ldp;
	ldom_t **lpp = NULL;
	uint_t lpp_count = 0;
	vcpu_t *vcpu;
	double vcpu_util, guest_util;
	double lo_wm, hi_wm;
	uint32_t lo_ms, hi_ms;
	hrtime_t t;
	time_t now;
	static time_t last_sec = 0;

	mutex_lock(&pm_util_list_lock);

	/* Find the PM_RES_CPU utilization event registration */
	CHAIN_ITER(pm_util_list.util_event_list, ep, link) {
		if (ep->res_type == PM_RES_CPU) {
			/* Found it.  Set util monitor parameters. */
			lo_wm = ep->low_wm;
			hi_wm = ep->high_wm;
			lo_ms = ep->time_low_ms;
			hi_ms = ep->time_high_ms;

			break;
		}
	}

	time(&now);

	/* Loop for every guest */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		t = gethrtime();

		if (ldp->state == ACTIVE) {
			/*
			 * Get the utilization for this guest and add it
			 * to this guest's utilization history table.
			 */
			guest_util = get_guest_util(ldp);
			ADD_GUEST_UTIL(ldp, guest_util, t);

			if (do_vcpus) {
				/*
				 * For all "active" guests, get the current
				 * utilization value for each vcpu and add
				 * the value to the vcpu's utilization
				 * history table.
				 */
				CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
					vcpu_util = get_vcpu_util(vcpu);
					ADD_VCPU_UTIL(vcpu, vcpu_util, t);
				}
			}

			/* monitor this guest's utilization, if necessary */
			if ((ep != NULL) && (ep->active) &&
			    (guest_util_monitor((void *)ldp, ep->res_type,
			    guest_util, t, lo_wm, hi_wm, lo_ms, hi_ms) !=
			    PM_UTIL_WITHIN)) {
				/*
				 * This guest's utilization crossed a
				 * threshold.  Add this ldp to the array of
				 * ldp's for PM notification later.
				 */
				lpp_count++;
				lpp = Xrealloc(lpp,
				    (sizeof (*lpp) * lpp_count));
				lpp[lpp_count - 1] = ldp;
			}
			/*
			 * DRM does not run faster than 1 sec
			 */
			if (option.allow_drm == true && last_sec != now)
				process_stats(ldp, now);
		} else {
			/*
			 * There's no need to get the utilization for not
			 * active guests.  Just add a zero value to the
			 * guest's utilization history table.
			 */
			ADD_GUEST_UTIL(ldp, 0.0, t);

			if (do_vcpus) {
				/*
				 * For all not active guests, there's no
				 * need to get the utilization value for
				 * each vcpu.  Just add a zero value to
				 * the vcpu's utilization history table.
				 */
				CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
					ADD_VCPU_UTIL(vcpu, 0.0, t);
				}
			}

			/* reset guest utilization monitor data */
			ldp->pm_cpu_resutils.ldmr_util_dir = PM_UTIL_WITHIN;
			ldp->pm_cpu_resutils.ldmr_time = 0;
		}
	}

	last_sec = now;

	if (lpp_count != 0) {
		/*
		 * Suspend this utilization event registration.  It will
		 * become active again when PM resumes notifications (via
		 * ldm_resume_util_change_notify).
		 */
		ep->active = false;

		pm_util_change_notify(ep->handle, (srcid_t)lpp, lpp_count,
		    PM_RES_CPU);

		Xfree(lpp);
	}

	mutex_unlock(&pm_util_list_lock);
}

/*
 * Handle resource utilization tasks.
 */
void
handle_utils(bool_t do_vcpus)
{
	handle_guest_util_read(do_vcpus);
	handle_memory_util();

	(void) mutex_lock(&guest_util_poll_lock);
	guest_util_poll_seq = false;
	(void) cond_signal(&guest_util_poll_cv);
	(void) mutex_unlock(&guest_util_poll_lock);
}

/*
 * Poll HV for guest and vcpu utilization.
 */
static void *
guest_util_poller(void *arg)
{
	struct timespec delay;
	int sec_fract;
	int vcpu_poll_count, poll_count;

	DBG(zprintf("guest utilization poller running.\n"));

	if (GUEST_UTIL_POLL_INTV < 1000) {
		delay.tv_sec = 0;
		/* Convert milliseconds to nanoseconds */
		delay.tv_nsec = (long)GUEST_UTIL_POLL_INTV * NSEC_PER_MSEC;
	} else {
		/* Convert milliseconds to seconds */
		delay.tv_sec = GUEST_UTIL_POLL_INTV / 1000;

		/* Set fraction of a second, if necessary */
		sec_fract = GUEST_UTIL_POLL_INTV % 1000;
		if (sec_fract == 0)
			delay.tv_nsec = 0L;
		else
			delay.tv_nsec = (long)sec_fract * NSEC_PER_MSEC;

	}

	vcpu_poll_count = VCPU_UTIL_POLL_INTV / GUEST_UTIL_POLL_INTV;
	if ((VCPU_UTIL_POLL_INTV % GUEST_UTIL_POLL_INTV) != 0)
		vcpu_poll_count++;

	poll_count = 0;

	/* Loop until told to stop */
	while (!guest_util_poll_stop) {

		(void) mutex_lock(&guest_util_poll_lock);

		guest_util_poll_seq = true;

		seq_enqueue_util((poll_count == vcpu_poll_count));

		/*
		 * Wait for the sequencer to finish this utilization read.
		 * We don't want to queue another utilization read on the
		 * sequencer until this one has completed.
		 */
		while (guest_util_poll_seq)
			cond_wait(&guest_util_poll_cv, &guest_util_poll_lock);

		(void) mutex_unlock(&guest_util_poll_lock);

		if (poll_count < vcpu_poll_count)
			poll_count++;
		else
			poll_count = 0;

		(void) nanosleep(&delay, NULL);
	}

	DBG(zprintf("guest utilization poller stopped.\n"));

	return (NULL);
}

/*
 * Start the guest utilization poller thread.
 */
void
start_guest_util_poll(void)
{
	mutex_lock(&guest_util_poll_lock);

	if (!guest_util_poll_running) {
		guest_util_poll_stop = false;

		if (thr_create(NULL, NULL, guest_util_poller, NULL, THR_DAEMON,
		    &guest_util_poll_tid) == 0)
			guest_util_poll_running = true;
		else
			efatal("Couldn't create guest utilization poll thread");
	}

	mutex_unlock(&guest_util_poll_lock);
}

/*
 * Stop the guest utilization poller thread.
 */
void
stop_guest_util_poll(void)
{
	mutex_lock(&guest_util_poll_lock);

	if (guest_util_poll_running) {
		guest_util_poll_stop = true;
		(void) thr_join(guest_util_poll_tid, NULL, NULL);
		guest_util_poll_running = false;
	}

	mutex_unlock(&guest_util_poll_lock);
}
