/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <stdlib.h>
#include <uuid/uuid.h>

#include "basics.h"
#include "support.h"
#include "zprintf.h"
#include "fatal.h"
#include "chain.h"
#include "allocate.h"
#include "database.h"
#include "sequencer.h"
#include "uuid.h"

/*
 * Only unbound domains are handled here when ldmd initializes,
 * see grab_guest_uuid() for handling bound domains.
 */
void
uuid_init()
{
	ldom_t	*ldp = NULL;
	bool_t	uuid_set = false;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state >= BOUND)
			continue;

		/* Already has a UUID */
		if (is_uuid_valid(ldp->props.uuid))
			continue;

		/* Otherwise, check for UUID in the old database */
		if (database_ldom_get_uuid(ldp))
			continue;

		/* If not, then give it one now */
		new_ldom_uuid(ldp);
		DBG(zprintf("%s: LDom %s new uuid=%s\n", __func__,
		    ldp->namep, ldp->props.uuid));
		uuid_set = true;
	}

	if (uuid_set)
		database_save();
}

/*
 * Newly created LDoms receive a UUID
 */
void
new_ldom_uuid(ldom_t *ldp)
{
	uuid_t	uuid;

	uuid_generate(uuid);
	uuid_unparse(uuid, ldp->props.uuid);
	database_ldom_set_uuid(ldp);
}

/*
 * Previously bound domains have a UUID or receive one now.
 */
void
grab_guest_uuid(ldom_t *ldp, md_t *guest_mdp)
{
	md_node_t *pnode = NULL;
	char *uuidp = NULL;

	/* Restore/override UUID from the MD */
	pnode = md_find_node(guest_mdp, "platform");
	if (md_get_prop_str(guest_mdp, pnode, "uuid", &uuidp) == 0) {
		memcpy(ldp->props.uuid, uuidp, UUID_PRINTABLE_STRING_LENGTH);
		DBG(zprintf("%s: LDom %s restored uuid=%s from MD\n",
		    __func__, ldp->namep, ldp->props.uuid));
		database_ldom_set_uuid(ldp);
		database_save();
		return;
	}

	/* Or restore UUID from the database */
	if (database_ldom_get_uuid(ldp))
		return;

	/* Otherwise assign the UUID now */
	new_ldom_uuid(ldp);
	DBG(zprintf("%s: LDom %s new uuid=%s\n", __func__, ldp->namep,
	    ldp->props.uuid));
	database_save();
}

bool_t
is_uuid_valid(char *uuidp)
{
	int	len = 0;

	if (uuidp == NULL)
		return (false);

	len = strlen(uuidp);
	return (len == (UUID_PRINTABLE_STRING_LENGTH - 1) ? true : false);
}
