/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)varconfig.c	1.24	09/02/09 SMI"

/*
 * Module implementing the 'varconfig' DS protocol
 */

#include <strings.h>
#include <sys/types.h>

#include "ldom.h"
#include "allocate.h"
#include "zprintf.h"
#include "ds.h"
#include "ds_registry.h"
#include "sequencer.h"
#include "variables.h"
#include "database.h"
#include "xml_event.h"
#include "xml_cntrl_common.h"

/*
 * The Variable Configuration Domain Service
 *
 * This DS provides the ability to manipulate the LDom variable store.
 * Only two operations are supported, setting a variable and deleting
 * a variable. Setting a variable that is not present in the store
 * creates it.
 */

static ds_ver_t var_config_vers[] = { {1, 0} };
#define	VAR_CONFIG_NVERS  (sizeof (var_config_vers) \
	/ sizeof (var_config_vers[0]))

static ds_capability_t var_config_cap = {
	VAR_CONFIG_SVC,	/* svc_id */
	var_config_vers,	/* vers */
	VAR_CONFIG_NVERS	/* nvers */
};

static ds_ops_t var_config_libops = {
	dslib_register_handler,		/* ds_reg_cb */
	dslib_unregister_handler,	/* ds_unreg_cb */
	dslib_data_handler,		/* ds_data_cb */
	VAR_CONFIG_SVC			/* cb_arg */
};

/*
 * Info needed to invoke ds_clnt_reg/ds_svc_reg when loading plugins
 * There could be multiple services in the same info structure
 */
static libds_info_t var_config_info[] = {
	{
		DS_CLIENT,
		&var_config_cap,
		&var_config_libops
	}
};

typedef enum {
	VAR_CONFIG_SET_REQ	= 0x0,
	VAR_CONFIG_DELETE_REQ	= 0x1,
	VAR_CONFIG_SET_RESP	= 0x2,
	VAR_CONFIG_DELETE_RESP	= 0x3
} var_config_cmd_t;

typedef struct {
	var_config_cmd_t cmd;
} var_config_hdr_t;

typedef struct {
	char	name_and_value[1];
} var_config_set_req_t;

typedef struct {
	char	name[1];
} var_config_delete_req_t;

typedef enum {
	VAR_CONFIG_SUCCESS,
	VAR_CONFIG_NO_SPACE,
	VAR_CONFIG_INVALID_VAR,
	VAR_CONFIG_INVALID_VAL,
	VAR_CONFIG_VAR_NOT_PRESENT
} var_config_status_t;

typedef struct {
	var_config_status_t status;
} var_config_resp_t;

typedef struct {
	var_config_hdr_t vc_hdr;
	union {
		var_config_set_req_t	vc_set;
		var_config_delete_req_t	vc_delete;
		var_config_resp_t	vc_resp;
	} u;
} var_config_msg_t;

#define	var_config_cmd		vc_hdr.cmd
#define	var_config_set		u.vc_set
#define	var_config_delete	u.vc_delete
#define	var_config_status	u.vc_resp.status

#define	VAR_CONFIG_MAX_STR_LEN		1024

#define	VAR_CONFIG_MAX_SET_REQ_LEN	((VAR_CONFIG_MAX_STR_LEN * 2) +	\
					sizeof (var_config_hdr_t))
#define	VAR_CONFIG_MAX_DEL_REQ_LEN	(VAR_CONFIG_MAX_STR_LEN +	\
					sizeof (var_config_hdr_t))

static bool_t
var_config_is_str_valid(char *str)
{
	int	len;

	if (str == NULL) {
		DBG(zprintf("%s: invalid string: NULL\n", __func__));
		return (false);
	}

	len = strlen(str);

	if (len <= 0) {
		DBG(zprintf("%s: invalid string: too short\n", __func__));
		return (false);
	}

	if (len > VAR_CONFIG_MAX_STR_LEN) {
		DBG(zprintf("%s: invalid string: too long\n", __func__));
		return (false);
	}

	return (true);
}

static var_config_status_t
var_config_set_var(ldom_t *ldp, var_config_msg_t *vc_reqp, uint16_t len)
{
	char		*name;
	char		*value;

	/* sanity check message */
	if (len > VAR_CONFIG_MAX_SET_REQ_LEN) {
		DBG(zprintf("%s: message too long: %d bytes, %d bytes max\n",
		    __func__, len, VAR_CONFIG_MAX_SET_REQ_LEN));
		return (VAR_CONFIG_INVALID_VAR);
	}

	name = (char *)vc_reqp + sizeof (var_config_hdr_t);

	if (!var_config_is_str_valid(name)) {
		DBG(zprintf("%s: invalid 'name' parameter\n", __func__));
		return (VAR_CONFIG_INVALID_VAR);
	}

	value = name + strlen(name) + 1;

	/* No validity checking needed on value */

	if ((name != NULL) && (strcmp(name, "security-password") == 0)) {
		DBG(zprintf("%s: name=%s, value=\n", __func__, name));
	} else {
		DBG(zprintf("%s: name=%s, value=%s\n", __func__, name, value));
	}

	/* update variable store */
	if (add_ldom_variable(ldp, name, value)) {

		/* update database */
		database_ldom_add_variable(ldp, name, value);
		database_save();
		event_send_res_event(XML_SET_VAR, ldp, NULL, NULL, NULL);

		/*
		 * Update machine description.  Special case for
		 * "reboot-command" variable update: we don't want to send an
		 * MD update, as only the Solaris CIF or OBP would send this
		 * request.
		 */
		if (ldp->state != INACTIVE) {
			bool_t notify_guest;

			if (strcmp(name, "reboot-command") == 0) {
				notify_guest = false;
			} else {
				/* mark config as different from SP */
				zeus.is_next_spconfig = false;
				notify_guest = true;
			}

			seq_update_vars(ldp, notify_guest);
		}
	}

	return (VAR_CONFIG_SUCCESS);
}

static var_config_status_t
var_config_del_var(ldom_t *ldp, var_config_msg_t *vc_reqp, uint16_t len)
{
	char *name;

	/* sanity check message */
	if (len > VAR_CONFIG_MAX_DEL_REQ_LEN) {
		DBG(zprintf("%s: message too long: %d bytes, %d bytes max\n",
		    __func__, len, VAR_CONFIG_MAX_DEL_REQ_LEN));
		return (VAR_CONFIG_INVALID_VAR);
	}

	name = (char *)vc_reqp + sizeof (var_config_hdr_t);

	if (!var_config_is_str_valid(name)) {
		DBG(zprintf("%s: invalid 'name' parameter\n", __func__));
		return (VAR_CONFIG_INVALID_VAR);
	}

	DBG(zprintf("%s: name=%s\n", __func__, name));

	/* check that the variable is present */
	if (get_ldom_variable(ldp, name) == NULL) {
		DBG(zprintf("%s: variable not present\n", __func__));
		return (VAR_CONFIG_VAR_NOT_PRESENT);
	}

	/* update variable store */
	(void) del_ldom_variable(ldp, name);

	/* update database */
	database_ldom_del_variable(ldp, name);
	database_save();
	event_send_res_event(XML_REMOVE_VAR, ldp, NULL, NULL, NULL);

	/*
	 * Update machine description.  Special case for "reboot-command"
	 * variable update: we don't want to send an MD update, as only the
	 * Solaris CIF or OBP would send this request.
	 */
	if (ldp->state != INACTIVE) {
		bool_t notify_guest;

		if (strcmp(name, "reboot-command") == 0) {
			notify_guest = false;
		} else {
			/* mark config as different from SP */
			zeus.is_next_spconfig = false;
			notify_guest = true;
		}

		seq_update_vars(ldp, notify_guest);
	}

	return (VAR_CONFIG_SUCCESS);
}

void
varconfig_data_handler(ds_svc_t *svc, void *msg, size_t msglen)
{
	ds_seq_req_t *dsreq;

	dsreq = Xmalloc(sizeof (*dsreq));
	dsreq->svc = svc;
	dsreq->msg = msg;
	dsreq->len = msglen;

	seq_enqueue_ds_req(dsreq);
}

void
varconfig_seq_handler(ds_svc_t *svc, void *msg, size_t msglen)
{
	int			cmd;
	int			len;
	ds_msg_t		ds_msg;
	var_config_msg_t	*vc_req;
	var_config_msg_t	*vc_resp;
	var_config_status_t	status;
	ds_resp_t		*ds_resp;
	size_t			resp_len, resp_data_len;

	vc_req = (var_config_msg_t *)((char *)msg + sizeof (ds_data_handle_t));
	cmd = ntoh32(vc_req->var_config_cmd);

	/*
	 * Prime a response.
	 */
	resp_data_len = sizeof (var_config_hdr_t) + sizeof (var_config_resp_t);
	resp_len = sizeof (ds_hdr_t) + sizeof (ds_data_handle_t) +
	    resp_data_len;
	ds_resp = Xmalloc(resp_len);
	DBG7(memset(ds_resp, 0, resp_len));

	ds_fill_data_header(ds_resp, svc->hdl, resp_data_len);

	vc_resp = (var_config_msg_t *)((char *)ds_resp + sizeof (ds_hdr_t) +
	    sizeof (ds_data_handle_t));

	ds_msg.buf = ds_resp;
	ds_msg.len = resp_len;
	ds_msg.seq = 0;

	len = msglen - sizeof (ds_data_handle_t);
	switch (cmd) {
	case VAR_CONFIG_SET_REQ:

		status = var_config_set_var(svc->chan->ldp, vc_req, len);

		vc_resp->var_config_cmd = hton32(VAR_CONFIG_SET_RESP);
		vc_resp->var_config_status = hton32(status);
		break;

	case VAR_CONFIG_DELETE_REQ:

		status = var_config_del_var(svc->chan->ldp, vc_req, len);

		vc_resp->var_config_cmd = hton32(VAR_CONFIG_DELETE_RESP);
		vc_resp->var_config_status = hton32(status);
		break;

	case VAR_CONFIG_SET_RESP:
	case VAR_CONFIG_DELETE_RESP:

		zprintf("Unsupported var_config command: %d\n", cmd);
		vc_resp->var_config_status = hton32(VAR_CONFIG_INVALID_VAR);
		break;

	default:
		zprintf("Unknown var_config command: %d\n", cmd);
		vc_resp->var_config_status = hton32(VAR_CONFIG_INVALID_VAR);
		break;
	}

	/* send reply, don't wait for a response */
	(void) ds_send_data_msg(svc->chan, &ds_msg, DS_NO_BLOCK, NULL);
	Xfree(ds_resp);
}

/*
 * The versions supported
 */
static prot_version_t varconfig_versions[] =
{
	{1, 0}
};

static svc_ops_t varconfig_ops = {
	NULL,				/* reg_handler */
	NULL,				/* unreg_handler */
	varconfig_data_handler,		/* data_handler */
	varconfig_seq_handler,		/* seq_handler */
	NULL				/* data */
};

svc_info_t *
info(void)
{
	int i;
	svc_info_t *info;

	info = Xcalloc(1, svc_info_t);
	info->svcs_ver.major_number = SVCS_VER_MAJOR;
	info->svcs_ver.minor_number = SVCS_VER_MINOR;
	info->num_chans = 2;
	info->chan_names = Xmalloc(info->num_chans * sizeof (char *));
	info->chan_names[0] = Xstrdup(DS_CHAN_NAME);
	info->chan_names[1] = Xstrdup(LIBDS_CHAN_NAME);
	info->num_svcs = 1;
	info->svcs = Xmalloc(info->num_svcs * sizeof (char *));
	info->svcs[0] = Xstrdup(VAR_CONFIG_SVC);
	info->props = Xmalloc(info->num_svcs * sizeof (int));
	info->props[0] = 0;
	info->num_vers = NUM_VERS(varconfig_versions);
	info->versions = Xmalloc(info->num_vers * sizeof (prot_version_t));
	memcpy(info->versions, varconfig_versions,
	    info->num_vers * sizeof (prot_version_t));

	/*
	 * copy the info structure for each service
	 */
	info->libds_svcs = Xmalloc(info->num_svcs * sizeof (libds_info_t *));
	for (i = 0; i < info->num_svcs; i++)
		info->libds_svcs[i] = cp_dsinfo(&var_config_info[i],
		    info->svcs[i]);

	return (info);
}

svc_ops_t *
ops(char *svc_name)
{
	if (strcmp(svc_name, VAR_CONFIG_SVC) == 0)
		return (&varconfig_ops);
	else
		return (NULL);
}
