/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/types.h>

#include "basics.h"
#include "chain.h"
#include "variables.h"
#include "mdesc_impl.h"
#include "nvlist.h"
#include "ldom.h"
#include "allocate.h"
#include "database.h"
#include "zprintf.h"

void
init_variables(variables_t *vars)
{
	nvlist_init(vars);
}

void
free_variables(variables_t *vars)
{
	nvlist_clear(vars);
}

/*
 * Generate the variable store description for LDom 'ldp' in the machine
 * description 'mdp'.
 */
void
gen_guest_md_variables(ldom_t *ldp, md_t *mdp)
{
	variable_t *var;
	md_node_t *variables, **nodelistp;
	int i, num;

	if (ldp->variables.len == 0)
		return;			/* no variables defined */

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "variables"),
	    md_find_name(mdp, "fwd"), nodelistp);

	Xfree(nodelistp);

	assert(num == 0);

	variables = md_link_new_node(mdp, "variables",
	    md_root_node(mdp), "fwd", "back");

	var = ldp->variables.pairs;
	for (i = 0; i < ldp->variables.len; i++, var++)
		md_add_string_property(mdp, variables, var->name, var->value);
}

bool_t
set_vars_from_properties(ldom_t *ldp, md_node_t *node)
{
	md_prop_t *propp;
	char *val, *name;
	bool_t update = false;

	for (propp = CHAIN_START(node->prop_list); propp != NULL;
	    propp = CHAIN_NEXT(propp->link)) {
		if (propp->type == MDET_PROP_STR) {
			name = propp->sp->strp;
			val = (char *)propp->d.dbp->datap;
			if (strcmp(name, "_delete") == 0) {
				if (nvlist_delete_pair(&ldp->variables, val)) {
					update = true;
					database_ldom_del_variable(ldp, val);
				}
			} else {
				if (nvlist_add_pair(&ldp->variables, name,
				    val)) {
					update = true;
					database_ldom_add_variable(ldp, name,
					    val);
				}
			}
		}
	}

	return (update);
}

/*
 * Restore the variable store for LDom 'ldp' from the machine description
 * 'mdp'.
 */
void
restore_guest_variables(ldom_t *ldp, md_t *mdp)
{
	md_node_t **nodelistp;
	int num;

	DBG(zprintf("Restoring variables for guest %s\n", ldp->namep));

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, "variables"),
	    md_find_name(mdp, "fwd"), nodelistp);

	if (num == 1)
		set_vars_from_properties(ldp, nodelistp[0]);

	Xfree(nodelistp);
}

/*
 * Add a variable <'name','value'> pair to the variable store represented by
 * resource 'handle'.
 */
bool_t
add_ldom_variable(ldom_t *ldp, char *name, char *value)
{
	return (nvlist_add_pair(&ldp->variables, name, value));
}

/*
 * Delete a variable 'name' from the variable store represented by
 * resource 'handle'.
 */
bool_t
del_ldom_variable(ldom_t *ldp, char *name)
{
	return (nvlist_delete_pair(&ldp->variables, name));
}

/*
 * Lookup a variable 'name' from the variable store represented by
 * resource 'handle'.
 */
char *
get_ldom_variable(ldom_t *ldp, char *name)
{
	variable_t *var = nvlist_find(&ldp->variables, name);

	return ((var != NULL) ? var->value : NULL);
}
