/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <string.h>

#include "ldmd_types.h"
#include "allocate.h"
#include "database.h"
#include "dr.h"
#include "errors.h"
#include "fatal.h"
#include "ldom.h"
#include "ldcs.h"
#include "mdesc_impl.h"
#include "response.h"
#include "vio.h"
#include "zpri.h"
#include "zprintf.h"

/* list of all vcc services */
CHAIN(vcc_t, vcc_services) = DEFAULTCHAIN;

struct vcc_attributes vcc_attrs = {
	.name =		"vcc",
	.compatible =	"SUNW,sun4v-console-concentrator",
	.md_name =	"virtual-console-concentrator",
	.cli_name =	"vcc",
	.device_type =	"vcc",
	.port_name =	"vcc-port"
};

ident_val_t
vcc_devid(vcc_t *vcc)
{
	if (vcc->binding == NULL)
		return (INVALID_DEVID);
	else
		return (vcc->binding->cfg_handle->id);
}

static res_pool_t *
vcc_cfg_handles(ldom_t *ldp)
{
#if VDEV_CFG_HANDLE
	return (&ldp->free_cfg_handles);
#else
	return (&ldp->bound.vcc_cfg_handles);
#endif
}

vcc_t *
vcc_find_by_name(char *service_name)
{
	vcc_t *vcc;

	CHAIN_ITER(vcc_services, vcc, link) {
		if (strcmp(service_name, vcc->service) == 0)
			break;
	}

	return (vcc);
}

console_group_t *
console_group_find_by_name(vcc_t *vcc, char *name)
{
	console_group_t *group;

	if (name == NULL || vcc->binding == NULL)
		return (NULL);

	CHAIN_ITER(vcc->binding->groups, group, link) {
		if (strcmp(group->group_name, name) == 0)
			break;
	}

	return (group);
}

console_group_t *
console_group_find_by_port(vcc_t *vcc, int tcp_port)
{
	console_group_t *group;

	if (vcc->binding == NULL)
		return (NULL);

	if (!PORT_IN_RANGE(tcp_port, vcc))
		return (NULL);

	/*
	 * check if this port is in use by a group
	 */
	CHAIN_ITER(vcc->binding->groups, group, link)
		if (tcp_port == group->tcp_port->id)
			return (group);

	return (NULL);
}

/*
 * Find the vcc console port using a given LDC channel.
 */
vcc_console_t *
vcc_console_find_by_ldcid(ldom_t *ldp, uint64_t ldcid)
{
	vcc_t *vcc;
	vcc_console_t *cons;

	CHAIN_ITER(ldp->vccs, vcc, ldom) {
		if (vcc->binding == NULL)
			continue;
		CHAIN_ITER(vcc->binding->consoles, cons, link) {
			if (ldcs_search(&cons->ldcs, ldcid) != NULL)
				return (cons);
		}
	}

	return (NULL);
}

/*
 * Create a new console group.
 */
static int
new_console_group(vcc_t *vcc, char *group_name, int tcp_port,
    console_group_t **group)
{
	console_group_t *g;
	resid_t *gport;

	if (IS_VAL_AUTO(tcp_port)) {
		gport = alloc_res(&vcc->binding->tcp_pool);
		if (gport == NULL)
			return (ERR_CONSOLE_NO_PORTS);
	} else {
		gport = pull_res_id(&vcc->binding->tcp_pool, tcp_port);
		if (gport == NULL)
			return (ERR_CONSOLE_PORT_EXISTS);
	}

	g = Xmalloc(sizeof (console_group_t));

	g->tcp_port = gport;
	g->group_name = Xstrdup(group_name);
	g->refcnt = 0;
	g->vcc = vcc;

	CHAIN_ADD(vcc->binding->groups, g, link);

	*group = g;
	return (0);
}

vcc_console_t *
vcc_new_console(vcc_t *vcc, console_group_t *group, int cons_port)
{
	vcc_console_t *vcons;

	vcons = Xmalloc(sizeof (vcc_console_t));

	if (cons_port == NEW_INSTANCE)
		vcons->port = alloc_res(&vcc->binding->free_ports);
	else
		vcons->port = pull_res_id(&vcc->binding->free_ports, cons_port);

	if (vcons->port == NULL)
		fatal("Unable to allocate port %d on vcc %s",
		    cons_port, vcc->service);

	vcons->console = NULL;		/* no client side yet */
	vcons->vcc = vcc;
	vcons->group = group;
	CHAIN_INIT(vcons->ldcs.ldcs);

	CHAIN_ADD(vcc->binding->consoles, vcons, link);

	return (vcons);
}

vcc_t *
vcc_new(ldom_t *ldp, char *name, int min_tcp_port, int max_tcp_port)
{
	vcc_t *vcc;

	if ((vcc = vcc_find_by_name(name)) != NULL)
		return (NULL);

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(ldp->vccs) == 0) {
		create_res_pool(&ldp->bound.vcc_cfg_handles,
		    IDENT_MIN, IDENT_MAX, Res_sorted);
		name_ldom_res_pool(&ldp->bound.vcc_cfg_handles, ldp,
		    "vcc cfg handles");
	}
#endif
	vcc = Xmalloc(sizeof (vcc_t));

	vcc->ldp = ldp;
	vcc->service = Xstrdup(name);
	vcc->min_tcp_port = min_tcp_port;
	vcc->max_tcp_port = max_tcp_port;
	vcc->binding = NULL;

	CHAIN_ADD(ldp->vccs, vcc, ldom);
	CHAIN_ADD(vcc_services, vcc, link);

	return (vcc);
}

static void
bind_vcc(vcc_t *vcc, int cfg_handle)
{
	ldom_t *ldp = vcc->ldp;
	vcc_binding_t *binding;

	binding = Xmalloc(sizeof (vcc_binding_t));

	binding->cfg_handle = vio_get_cfg_handle(cfg_handle,
	    vcc_cfg_handles(ldp));
	if (binding->cfg_handle == NULL)
		fatal("Could not obtain vcc cfg_handle %d on LDom %s\n",
		    cfg_handle, ldp->namep);

	CHAIN_INIT(binding->consoles);
	CHAIN_INIT(binding->groups);
	create_res_pool(&binding->free_ports, 0, IDENT_MAX, Res_sorted);
	name_ldom_vdev_res_pool(&binding->free_ports, ldp, "vcc",
	    vcc->service, "ports");

	/* create a resource pool to hold the TCP port range */
	create_res_pool(&binding->tcp_pool, vcc->min_tcp_port,
	    vcc->max_tcp_port, Res_sorted);

	ldp->bound.num_services++;
	vcc->binding = binding;
}

static void
unbind_vcc(vcc_t *vcc)
{
	if (vcc->binding == NULL)
		return;

	ASSERT(CHAIN_LENGTH(vcc->binding->consoles) == 0);
	ASSERT(CHAIN_LENGTH(vcc->binding->groups) == 0);

	vcc->ldp->bound.num_services--;
	destroy_res_pool(&vcc->binding->free_ports);
	destroy_res_pool(&vcc->binding->tcp_pool);
	free_res(vcc->binding->cfg_handle);
	Xfree(vcc->binding);
	vcc->binding = NULL;
}

/*ARGSUSED1*/
int
vcc_bind_all(ldom_t *ldp, response_t *resp)
{
	vcc_t *vcc;

	CHAIN_ITER(ldp->vccs, vcc, ldom) {
		bind_vcc(vcc, NEW_INSTANCE);
	}

	return (0);
}

/*ARGSUSED1*/
int
vcc_unbind_all(ldom_t *ldp, response_t *resp)
{
	vcc_t *vcc;

	CHAIN_ITER(ldp->vccs, vcc, ldom) {
		unbind_vcc(vcc);
	}

	return (0);
}

/*
 * Ensure that the min_port and max_port has no overlap with other vcc servers
 * in a guest.
 */
static bool_t
check_port_range(ldom_t *ldp, vcc_t *excl_vcc, int min_port,
    int max_port, response_t *resp)
{
	vcc_t *vcc;

	/*
	 * Check all vcc servers bound to the guest excluding the one
	 * specified.
	 */
	CHAIN_ITER(ldp->vccs, vcc, ldom) {
		if (vcc == excl_vcc)
			continue;

		if ((min_port >= vcc->min_tcp_port ||
		    max_port >= vcc->min_tcp_port) &&
		    (min_port <= vcc->max_tcp_port ||
		    max_port <= vcc->max_tcp_port)) {
			if (resp != NULL) {
				rprintf(resp, gettext("Error: port range %d-%d "
				    "overlaps with existing port range "
				    "%d-%d\n"), min_port, max_port,
				    vcc->min_tcp_port, vcc->max_tcp_port);
			}
			return (false);
		}
	}

	return (true);
}

static void
destroy_console_group(console_group_t *group)
{
	ASSERT(group->refcnt == 0);

	free_res(group->tcp_port);
	Xfree(group->group_name);
	Xfree(group);
}

void
release_console_group(console_group_t *group)
{
	ASSERT(group->refcnt > 0);

	if (--group->refcnt == 0) {
		CHAIN_UNLINK(group->vcc->binding->groups, group, link);
		destroy_console_group(group);
	}
}

void
vcc_destory_console(vcc_console_t *vcons)
{
	if (vcons->group != NULL)
		release_console_group(vcons->group);
	if (vcons->port != NULL)
		free_res(vcons->port);
	ldcs_free(&vcons->ldcs);
	Xfree(vcons);
}

int
vcc_unbind_console(vcc_console_t *vcons)
{
	CHAIN_UNLINK(vcons->vcc->binding->consoles, vcons, link);
	ldcs_stage(&vcons->ldcs);
	vcc_destory_console(vcons);

	return (0);
}

static void
destroy_vcc(vcc_t *vcc)
{
	ldom_t *ldp = vcc->ldp;

	ASSERT(vcc->binding == NULL);

	Xfree(vcc->service);
	CHAIN_UNLINK(ldp->vccs, vcc, ldom);
	CHAIN_UNLINK(vcc_services, vcc, link);
	Xfree(vcc);

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(ldp->vccs) == 0)
		destroy_res_pool(&ldp->bound.vcc_cfg_handles);
#endif
}

void
vcc_delete_all(ldom_t *ldp)
{
	vcc_t *vcc, *vcc_next;

	for (vcc = CHAIN_START(ldp->vccs); vcc != NULL; vcc = vcc_next) {
		vcc_next = CHAIN_NEXT(vcc->ldom);
		destroy_vcc(vcc);
	}
}

/*
 * Called from the add-vcc CLI.
 */
int
vcc_add(ldom_t *ldp, char *service_name, int min_port, int max_port,
    response_t *resp)
{
	vcc_t	*vcc;
	int	rv;

	if (vcc_find_by_name(service_name) != NULL) {
		rprintf(resp, gettext("%s service named %s already exists\n"),
		    vcc_attrs.name, service_name);
		return (ERR_SERVICE_EXISTS);
	}

	if (check_port_range(ldp, NULL, min_port, max_port, resp) == false)
		return (ERR_VCC_PORT_RANGE_INVALID);

	vcc = vcc_new(ldp, service_name, min_port, max_port);

	if (ldp->state >= BOUND) {
		bind_vcc(vcc, NEW_INSTANCE);

		if (zeus.mode == ZEUS_MGMT_MODE) {
			if (ldp->state == BOUND) {
				/* do reconfig now */
				seq_download_mds_do_config(ldp);
				free_all_staged_ldcs(ldp);

			} else if (vio_check_dr_capable(ldp, resp) == true) {
				/* attempt DR operation */
				rv = vio_dr_add(ldp, vcc_attrs.md_name,
				    vcc_devid(vcc), resp);
				/*
				 * If the DR operation failed and the dr-vio
				 * domain service determines that it is OK to
				 * rollback we remove the vnet again.
				 */
				if (rv != 0) {
					if (rv == ERR_DR_VIO_ROLLBACK) {
						unbind_vcc(vcc);
						destroy_vcc(vcc);
						seq_rm_reconfigure(ldp);
						free_all_staged_ldcs(ldp);
					}
					return (rv);
				}
			} else {
				/* handle as a delayed reconfig operation */
				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_ADD, resp);
			}
		}
	}

	database_ldom_add_console_service(ldp, service_name,
	    min_port, max_port);

	return (0);
}

/*
 * Check if all the groups have ports within the specified
 * range. Return the port used by a client if outside the
 * port range specified, otherwise return 0.
 */
static int
ports_in_range(vcc_t *vcc, int min_port, int max_port)
{
	console_group_t *group;

	CHAIN_ITER(vcc->binding->groups, group, link) {
		int port = group->tcp_port->id;
		if (port < min_port || port > max_port)
			return (port);
	}

	return (0);
}

static void
save_and_free_vcc_ports(vcc_t *vcc)
{
	console_group_t *group;

	CHAIN_ITER(vcc->binding->groups, group, link) {
		group->work.tcp_port = group->tcp_port->id;
		free_res(group->tcp_port);
	}
}

static void
restore_vcc_ports(vcc_t *vcc)
{
	console_group_t *group;

	CHAIN_ITER(vcc->binding->groups, group, link) {
		group->tcp_port = pull_res_id(&vcc->binding->tcp_pool,
		    group->work.tcp_port);
	}
}

static int
update_port_range(vcc_t *vcc, int min_port, int max_port, response_t *resp)
{
	ASSERT(min_port >= MIN_CONSOLE_TCP_PORT &&
	    min_port <= MAX_CONSOLE_TCP_PORT && min_port <= max_port);
	ASSERT(max_port >= MIN_CONSOLE_TCP_PORT &&
	    max_port <= MAX_CONSOLE_TCP_PORT);

	/*
	 * If vcc is already bound, we need to destroy the current pool and
	 * recreate the pool with new range. In addition, if there were bound
	 * consoles, we also need to save the current ports prior to
	 * destroying and re-allocate afterwards
	 */
	if (vcc->binding != NULL) {
		int nconsoles;
		if ((nconsoles = CHAIN_LENGTH(vcc->binding->consoles)) > 0) {
			int port;
			/*
			 * There are bound clients, do not allow the port range
			 * to be modified if a group is allotted a port outside
			 * of this range.
			 */
			if ((port = ports_in_range(vcc, min_port,
			    max_port)) != 0) {
				if (resp != NULL)
					rprintf(resp, gettext("A console group "
					    "is using a port %d outside this "
					    "range\n"), port);
				return (ERR_VCC_PORT_OUT_OF_RANGE);
			}

			/*
			 * Before we destroy the resource pool, we must first
			 * release all the ports being used by the group, and
			 * then reallocate them from the new pool.
			 */
			save_and_free_vcc_ports(vcc);
		}
		destroy_res_pool(&vcc->binding->tcp_pool);
		create_res_pool(&vcc->binding->tcp_pool,
		    min_port, max_port, Res_sorted);
		if (nconsoles > 0)
			restore_vcc_ports(vcc);
	}

	vcc->min_tcp_port = min_port;
	vcc->max_tcp_port = max_port;

	return (0);
}

/*
 * Called from the set-vcc CLI. Modifying vcc port range only changes the zeus
 * vcc state and the constraint DB, no changes in MD are needed.
 */
int
vcc_modify(vcc_t *vcc, int min_port, int max_port, response_t *resp)
{
	ldom_t *ldp = vcc->ldp;
	int rtn = 0;

	/*
	 * When checking port range, we need to check port-ranges
	 * for all vcc services in this guest **EXCLUDING** the
	 * service being modified.
	 */
	if (check_port_range(ldp, vcc, min_port, max_port, resp) == false)
		return (ERR_VCC_PORT_RANGE_INVALID);

	rtn = update_port_range(vcc, min_port, max_port, resp);
	if (rtn != 0)
		return (rtn);

	database_ldom_modify_console_service(ldp, vcc->service,
	    min_port, max_port);

	if (zeus.mode == ZEUS_CONFIG_MODE)
		free_all_staged_ldcs(ldp);

	return (0);
}

void
vcc_remove_do_teardown(vcc_t *vcc)
{
	database_ldom_del_console_service(vcc->ldp, vcc->service);
	unbind_vcc(vcc);
	destroy_vcc(vcc);
}

/*
 * Called from the remove-vcc CLI.
 */
/*ARGSUSED1*/
int
vcc_remove(vcc_t *vcc, bool_t force, response_t *resp)
{
	ldom_t *ldp = vcc->ldp;
	int	rv;

	if ((vcc->binding != NULL) &&
	    (CHAIN_LENGTH(vcc->binding->consoles) > 0)) {
		/* clients are bound to the vcc service */
		rprintf(resp, MSG_ERR_SERVICE_BOUND_CLIENTS);
		return (ERR_SERVICE_BOUND_CLIENTS);
	}

	switch (zeus.mode) {
	case ZEUS_MGMT_MODE:
		if (ldp->state < BOUND) {
			vcc_remove_do_teardown(vcc);

		} else if (ldp->state == BOUND) {
			/* do reconfig now */
			vcc_remove_do_teardown(vcc);
			seq_download_mds_do_config(ldp);
			free_all_staged_ldcs(ldp);

		} else if (vio_check_dr_capable(ldp, resp)) {
			/* Attempt DR operation */
			rv = vio_dr_rm_phase1(ldp, vcc_attrs.md_name,
			    vcc_devid(vcc), resp);
			if (rv != 0)
				return (rv);
			vcc_remove_do_teardown(vcc);
			vio_dr_rm_phase2(ldp);

		} else {
			/* handle as a delayed reconfig operation */
			vcc_remove_do_teardown(vcc);
			initiate_or_update_delayed_reconfig(ldp,
			    DEL_RECONFIG_REMOVE, resp);
		}
		break;

	case ZEUS_CONFIG_MODE:
		vcc_remove_do_teardown(vcc);
		free_all_staged_ldcs(ldp);
		break;

	default:
		ASSERT(0);
		return (ERR_GENERIC);
	}

	return (0);
}

static void
deduce_port_range(ldom_t *ldp, md_t *guest_mdp, md_node_t *vdev_p,
    uint64_t *min_port, uint64_t *max_port)
{
	md_node_t **nodelistp;
	uint64_t tcp_port;
	int i, num_client_bindings;

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(guest_mdp));

	num_client_bindings = md_scan_dag(guest_mdp, vdev_p,
	    md_find_name(guest_mdp, "virtual-device-port"),
	    md_find_name(guest_mdp, "fwd"), nodelistp);

	if (num_client_bindings == 0) {
		/*
		 * There are no console clients so use some reasonable
		 * default.
		 */
		*min_port = DEFAULT_MIN_VCC_PORT;
		*max_port = *min_port + pri.max_guests - 1;

		/*
		 * See if range is available. If not, keep walking through
		 * successive ranges until a suitable one is found.
		 */
		while (check_port_range(ldp, NULL, *min_port, *max_port,
		    NULL) == false) {
			*min_port += pri.max_guests;
			*max_port += pri.max_guests;

			if (*max_port > DEFAULT_MAX_VCC_PORT) {
				/* no ports available */
				*min_port = *max_port = 0;
				Xfree(nodelistp);
				return;
			}
		}

		Xfree(nodelistp);
	} else {
		uint64_t try_max_port;

		for (i = 0; i < num_client_bindings; i++) {
			if (md_get_prop_val(guest_mdp, nodelistp[i],
			    VCONS_TCP_PORT, &tcp_port) != 0)
				fatal("vport has no %s property",
				    VCONS_TCP_PORT);
			if (i == 0)
				*min_port = *max_port = tcp_port;
			else if (tcp_port < *min_port)
				*min_port = tcp_port;
			else if (tcp_port > *max_port)
				*max_port = tcp_port;
		}

		Xfree(nodelistp);

		/*
		 * Now try to add up to max_guests available ports
		 */
		try_max_port = *max_port + pri.max_guests - 1;

		while (check_port_range(ldp, NULL, *min_port, try_max_port,
		    NULL) == false) {
			if (try_max_port-- == *max_port) {
				/*
				 * This should never happen, as it means the
				 * existing allocated ports do not define a
				 * single valid range.
				 */
				*min_port = *max_port = 0;
				return;
			}
		}
		*max_port = try_max_port;
	}
}

/*
 * Get a console group. If there already one exists, return that one, provided
 * the port of that group matches. If there is no matching group with the
 * specified name, allocate a new group with the given group name and port.
 */
console_group_t *
vcc_get_console_group(vcc_t *vcc, char *gname, int tcp_port)
{
	console_group_t *group = NULL;

	if (gname == NULL)
		return (NULL);

	group = console_group_find_by_name(vcc, gname);

	/*
	 * If group is found, and the TCP port does not match the one
	 * specified, return failure.
	 */
	if (group != NULL) {
		if ((tcp_port > _AUTO_VAL_) &&
		    (group->tcp_port->id != tcp_port))
			return (NULL);
	} else {
		if (new_console_group(vcc, gname, tcp_port, &group) != 0)
			return (NULL);
	}

	group->refcnt++;
	return (group);
}

void
vcc_add_from_mdnode(ldom_t *ldp, md_t *guest_md, md_node_t *vdev)
{
	vcc_t *vcc;
	char *service_name;
	uint64_t cfg_handle;
	uint64_t min_port = 0, max_port = 0;
	int i, num_client_bindings;
	md_node_t **nodes;

	if (md_get_prop_val(guest_md, vdev, "cfg-handle", &cfg_handle) != 0)
		fatal("vdevice has no cfg-handle property");

	if (md_get_prop_str(guest_md, vdev, DEVALIAS, &service_name) == 0) {
		service_name = Xstrdup(service_name);
	} else if (md_get_prop_str(guest_md, vdev, SERVICE_NAME,
	    &service_name) == 0) {
		service_name = Xstrdup(service_name);
	} else {
		service_name = get_service_name_from_db(ldp, vcc_attrs.name,
		    cfg_handle);
		if (service_name == NULL)
			/* use canonical name */
			service_name = vio_canonical_service_name(ldp,
			    vcc_attrs.name, cfg_handle);
	}

	if (md_get_prop_val(guest_md, vdev, VCC_MIN_PORT, &min_port) != 0)
		warning("vdevice has no %s property", VCC_MIN_PORT);
	if (md_get_prop_val(guest_md, vdev, VCC_MAX_PORT, &max_port) != 0)
		warning("vdevice has no %s property", VCC_MAX_PORT);
	if (max_port == 0 && (database_get_vcc_properties(ldp,
	    service_name, &min_port, &max_port) == false ||
	    (min_port >= max_port))) {
		/*
		 * If no port constraint could be recovered either from vcc
		 * MD node or the DB, deduce the range from the vcc ports in MD
		 */
		deduce_port_range(ldp, guest_md, vdev, &min_port, &max_port);
		if (min_port == 0 && max_port == 0) {
			warning("Unable to deduce TCP port range for\n"
			    "console server instance %"PRIu64" on LDom %s",
			    cfg_handle, ldp->namep);
			return;
		}
		warning("Applying deduced TCP port range of %"PRIu64"-%"PRIu64""
		    " for\nconsole server instance %"PRIu64" on LDom %s",
		    min_port, max_port, cfg_handle, ldp->namep);
	}

	if ((vcc = vcc_new(ldp, service_name, min_port, max_port)) == NULL)
		fatal("Duplicate vcc service %s", service_name);

	bind_vcc(vcc, cfg_handle);

	database_ldom_add_console_service(ldp, service_name,
	    min_port, max_port);

	Xfree(service_name);

	DBG(zprintf("Added new vcc server to guest %s, min_port %"PRIu64", "
	    "max_port %"PRIu64"\n", ldp->namep, min_port, max_port));

	/*
	 * Scan through the vport list and add a client binding for
	 * each one.
	 */
	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(guest_md));

	num_client_bindings = md_scan_dag(guest_md, vdev,
	    md_find_name(guest_md, "virtual-device-port"),
	    md_find_name(guest_md, "fwd"), nodes);

	for (i = 0; i < num_client_bindings; ++i) {
		console_group_t *group;
		vcc_console_t *vcons;
		char *group_name, *name;
		uint64_t cons_port, tcp_port;

		md_node_t *vport = nodes[i];

		if (md_get_prop_str(guest_md, vport, "name", &name) != 0)
			fatal("vport has no name property");
		if (md_get_prop_val(guest_md, vport, "id", &cons_port) != 0)
			fatal("vport has no id property");

		/*
		 * Get the vcc-group-name and vcc-tcp-port properties.
		 */
		if (md_get_prop_str(guest_md, vport, VCONS_GROUP,
		    &group_name) != 0)
			fatal("vport has no %s property", VCONS_GROUP);
		if (md_get_prop_val(guest_md, vport,
		    VCONS_TCP_PORT, &tcp_port) != 0)
			fatal("vport has no %s property", VCONS_TCP_PORT);

		/*
		 * Get a console group for the binding's name and port.
		 */
		group = vcc_get_console_group(vcc, group_name, tcp_port);
		if (group == NULL)
			fatal("Could not reserve VCC group %s port %"PRIu64,
			    group_name, tcp_port);

		vcons = vcc_new_console(vcc, group, cons_port);

		ldc_get_and_pull_ids(ldp, &ldp->guest_free_ldcs,
		    guest_md, vport, &vcons->ldcs);

		DBG(zprintf("Added new vcc-port server to guest %s "
		    "group %s, tcp port %"PRIu64"\n", ldp->namep,
		    group_name, tcp_port));
	}

	Xfree(nodes);
}

void
vcc_gen_ldc_endpoints(ldom_t *ldp, md_t *hv_md, md_node_t *hv_guest,
    md_node_t *ldc_endpoints)
{
	vcc_t *vcc;
	vcc_console_t *cons;

	CHAIN_ITER(ldp->vccs, vcc, ldom) {
		CHAIN_ITER(vcc->binding->consoles, cons, link) {
			vio_gen_ldcs_endpoint_nodes(hv_md, hv_guest,
			    ldc_endpoints, TYPE_GUEST, TYPE_GUEST,
			    &cons->ldcs, cons->console->ldp->idp->id,
			    &cons->console->binding->ldcs, NULL);
		}
	}
}

static void
gen_vcc_guest_md_one(vcc_t *vcc, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vcc_console_t *cons;
	md_node_t *vdev;

	vdev = md_link_new_node(md, VDEV_NAME, cdevs, "fwd", "back");

	/*
	 * Compatible property is an array of strings, though we only
	 * currently have one.
	 */
	md_add_string_as_data_property(md, vdev, "compatible",
	    vcc_attrs.compatible);
	md_add_string_property(md, vdev, "device-type", vcc_attrs.device_type);
	md_add_string_property(md, vdev, "name", vcc_attrs.md_name);
	md_add_value_property(md, vdev, "cfg-handle", vcc_devid(vcc));
	md_add_string_property(md, vdev, DEVALIAS, vcc->service);
	md_add_value_property(md, vdev, VCC_MIN_PORT, vcc->min_tcp_port);
	md_add_value_property(md, vdev, VCC_MAX_PORT, vcc->max_tcp_port);

	CHAIN_ITER(vcc->binding->consoles, cons, link) {
		md_node_t *port_node;

		port_node = vio_gen_vdev_port_node(md, vdev, chan_endpoints,
		    vcc_attrs.port_name, cons->port->id, &cons->ldcs);
		md_add_string_property(md, port_node, VCONS_DOMAIN,
		    cons->console->ldp->namep);
		md_add_string_property(md, port_node, VCONS_GROUP,
		    cons->group->group_name);
		md_add_value_property(md, port_node, VCONS_TCP_PORT,
		    cons->group->tcp_port->id);
	}
}

void
vcc_gen_guest_md(ldom_t *ldp, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vcc_t *vcc;

	CHAIN_ITER(ldp->vccs, vcc, ldom) {
		gen_vcc_guest_md_one(vcc, md, cdevs, chan_endpoints);
	}
}

static void
vcc_console_list(vcc_console_t *vcons, response_t *resp, dispctl_t *dctl)
{
	char *client;

	client = zsprintf("%s@%s", vcons->group->group_name,
	    vcons->vcc->service);

	if (dctl->parseable)
		rprintf(resp, "|client=%s|port=%"PRIu64"", client,
		    vcons->group->tcp_port->id);
	else
		rprintf(resp, "        "FBNDNAME" "FPORTd, client,
		    vcons->group->tcp_port->id);
	Xfree(client);

	if (dctl->extended) {
		if (dctl->parseable)
			rprintf(resp, "|ldc=");
		else
			rprintf(resp, " ");
		ldcs_print(&vcons->ldcs, resp);
	}

	rprintf(resp, "\n");
}

static void
vcc_list_one(vcc_t *vcc, response_t *resp, dispctl_t *dctl)
{
	vcc_binding_t *binding = vcc->binding;
	int nclients;

	/*
	 * TRANSLATION_NOTE
	 * The following are column headers in tables with fixed column widths.
	 * WIDTH  HEADER	MEANING
	 * 16	  "NAME"	VIO server name
	 * 16	  "LDOM"	the LDom name iff list-services
	 *	  "PORT-RANGE"	console port range
	 */
	if (dctl->header && !dctl->parseable) {
		dctl->header = 0;
		rprintf(resp, "    "FNAME" ", gettext("NAME"));
		if (dctl->services)
			rprintf(resp, ""FNAME" %s\n", gettext("LDOM"),
			    gettext("PORT-RANGE"));
		else
			rprintf(resp, "%s\n", gettext("PORT-RANGE"));
	}

	if (binding != NULL)
		nclients = CHAIN_LENGTH(binding->consoles);
	else
		nclients = 0;

	print_server_header(vcc->service, vcc_attrs.name, nclients, resp, dctl);

	if (dctl->parseable) {
		rprintf(resp, "|port-range=%d-%d", vcc->min_tcp_port,
		    vcc->max_tcp_port);
		if (dctl->services)
			rprintf(resp, "|ldom=%s", vcc->ldp->namep);
		rprintf(resp, "\n");
	} else {
		if (dctl->services)
			rprintf(resp, " "FNAME, vcc->ldp->namep);
		rprintf(resp, " %d-%d\n", vcc->min_tcp_port, vcc->max_tcp_port);
	}

	if (dctl->bindings && (nclients > 0)) {
		vcc_console_t *vcons;

		dctl->header = 1;
		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in tables with fixed
		 * column widths.
		 * WIDTH  HEADER	MEANING
		 * 27	  "CLIENT"	name of client of service
		 * 6	  "PORT"	virtual disk volume name
		 * 3	  "LDC"		Logical Domain Channel number
		 */
		if (!dctl->parseable) {
			rprintf(resp, "        "FBNDNAME" "FPORT,
			    gettext("CLIENT"), gettext("PORT"));
			if (dctl->extended)
				rprintf(resp, " %s", gettext("LDC"));
			rprintf(resp, "\n");
		}

		CHAIN_ITER(binding->consoles, vcons, link) {
			vcc_console_list(vcons, resp, dctl);
		}
	}
}

void
vcc_list(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	vcc_t *vcc;

	if (dctl->oformat && !dctl->oconsole)
		return;

	if (CHAIN_LENGTH(ldp->vccs) == 0)
		return;

	dctl->header = 1;

	if (!dctl->parseable)
		rprintf(resp, "%s\n", dctl_label(vcc_attrs.name, dctl));

	CHAIN_ITER(ldp->vccs, vcc, ldom) {
		vcc_list_one(vcc, resp, dctl);
	}

	if (!dctl->parseable)
		rprintf(resp, "\n");
}
