/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <string.h>

#include "ldmd_types.h"
#include "allocate.h"
#include "chain.h"
#include "database.h"
#include "database_internal.h"
#include "errors.h"
#include "fatal.h"
#include "ldom.h"
#include "resource_id.h"
#include "vio.h"
#if BIND_CHECK
#include "zprintf.h"
#endif

struct vdisk_attributes vdisk_attrs = {
	.name =		"disk",
	.compatible =	"SUNW,sun4v-disk",
	.md_name =	"disk",
	.cli_name =	"vdisk",
	.device_type =	"block",
	.port_name =	"vdc-port"
};

ident_val_t
vdisk_devid(vdisk_t *vdisk)
{
	if (vdisk->binding == NULL)
		return (INVALID_DEVID);
	else
		return (vdisk->binding->cfg_handle->id);
}

static ident_val_t
vdisk_port_id(vdisk_port_t *vdisk_port)
{
	if (vdisk_port->port == NULL)
		return (INVALID_DEVID);
	else
		return (vdisk_port->port->id);
}

static void
vdisk_flag_svc_domains(vdisk_t *vdisk)
{
	vdisk_port_t *vdisk_port;

	if (vdisk->binding == NULL)
		return;

	CHAIN_ITER(vdisk->binding->vdisk_ports, vdisk_port, link) {
		ASSERT(vdisk_port->vds_port != NULL);
		ASSERT(vdisk_port->vds_port->vds != NULL);
		ASSERT(vdisk_port->vds_port->vds->ldp != NULL);
		vdisk_port->vds_port->vds->ldp->flag.md_notify = true;
	}
}

static res_pool_t *
vdisk_cfg_handles(ldom_t *ldp)
{
#if VDEV_CFG_HANDLE
	return (&ldp->free_cfg_handles);
#else
	return (&ldp->bound.vdisk_cfg_handles);
#endif
}

vdisk_t *
vdisk_find_by_name(ldom_t *ldp, char *vdisk_name)
{
	vdisk_t *vdisk;

	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {
		if (strcmp(vdisk_name, vdisk->name) == 0)
			break;
	}

	return (vdisk);
}

vdisk_t *
vdisk_new(ldom_t *ldp, char *name, char *service, char *volname, int timeout,
    int diskid)
{
	vdisk_t *vdisk;

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(ldp->vdisks) == 0) {
		create_res_pool(&ldp->bound.vdisk_cfg_handles,
		    IDENT_MIN, IDENT_MAX, Res_sorted);
		name_ldom_res_pool(&ldp->bound.vdisk_cfg_handles, ldp,
		    "vdisk cfg handles");
	}
#endif

	vdisk = Xcalloc(1, vdisk_t);

	vdisk->ldp = ldp;
	vdisk->name = Xstrdup(name);
	vdisk->service = (service != NULL) ? Xstrdup(service) : NULL;
	vdisk->volname = (volname != NULL) ? Xstrdup(volname) : NULL;
	vdisk->timeout = timeout;
	vdisk->diskid = diskid;

	vdisk->binding = NULL;
	CHAIN_LINK_INIT(vdisk->ldom);

	CHAIN_ADD(ldp->vdisks, vdisk, ldom);

	return (vdisk);
}

/*
 * Initiate a vdisk binding
 */
static int
vdisk_init_binding(vdisk_t *vdisk, ident_val_t cfg_handle)
{
	ldom_t *ldp = vdisk->ldp;
	vdisk_binding_t *binding;

	vdisk->binding = binding = Xcalloc(1, vdisk_binding_t);

	if ((binding->cfg_handle = vio_get_cfg_handle(cfg_handle,
	    vdisk_cfg_handles(ldp))) == NULL) {
			DBG(zprintf("Could not obtain vdisk cfg_handle %" PRId64
			    " on LDom %s\n", cfg_handle, ldp->namep));
			return (-1);
	}

	create_res_pool(&binding->free_ports, VDISK_MIN_PORT, IDENT_MAX,
	    Res_sorted);
	name_ldom_vdev_res_pool(&binding->free_ports, ldp, "vdisk",
	    vdisk->name, "ports");
	binding->vdisk = vdisk;

	CHAIN_INIT(binding->vdisk_ports);

	/*
	 * if the value for diskid is NO_NEW_ID, update the diskid with
	 * the cfg_handle id just pulled from the cfg_handle pool
	 */
	if (vdisk->diskid == NO_NEW_ID)
		vdisk->diskid = vdisk->binding->cfg_handle->id;

	return (0);
}

/*
 * Add a vdisk port
 */
static vdisk_port_t *
vdisk_add_port(vdisk_t *vdisk, uint64_t port_id)
{
	vdisk_binding_t *vdisk_binding = vdisk->binding;
	vdisk_port_t *vdisk_port = Xcalloc(1, vdisk_port_t);

	CHAIN_INIT(vdisk_port->ldcs.ldcs);
	CHAIN_LINK_INIT(vdisk_port->link);
	vdisk_port->vdisk = vdisk;

	if (port_id == NEW_INSTANCE)
		vdisk_port->port = alloc_res(&vdisk_binding->free_ports);
	else
		vdisk_port->port = pull_res_id(&vdisk_binding->free_ports,
		    port_id);

	if (vdisk_port->port == NULL)
		fatal("Unable to allocate port %"PRId64" on vdisk %s",
		    port_id, vdisk->name);

	CHAIN_ADD(vdisk_binding->vdisk_ports, vdisk_port, link);

	return (vdisk_port);
}

static int can_bind_vdisk(response_t *resp, vdisk_t *vdisk);
static void destroy_vdisk_binding(vdisk_t *vdisk);

int
vdisk_mig_bind(ldom_t *ldp, md_t *mdp, md_t *hvmdp, md_node_t *vdev)
{
	int		i, rc;
	vdisk_t		*vdisk;
	uint64_t	cfg_handle;
	char		*client_name;
	uint64_t	md_timeout;
	int		timeout;
	md_node_t	**nodes;
	int		nbound;
	uint64_t	port_id;
	vdisk_port_t	*vdisk_port;
	response_t	*resp;
	vds_t		*vds;
	vds_volume_t	*vol;

	/* the name of the device is found in the devalias */
	if (md_get_prop_str(mdp, vdev, "devalias", &client_name) != 0) {
		DBG(zprintf("vdevice has no devalias property\n"));
		return (-1);
	}

	/* lookup the vdisk, it should already exist */
	if ((vdisk = vdisk_find_by_name(ldp, client_name)) == NULL) {
		DBG(zprintf("LDom '%s' has no vdisk named '%s'\n", ldp->namep,
		    client_name));
		return (-1);
	}

	nodes = Xcalloc(md_node_count(mdp), md_node_t *);
	nbound = md_scan_dag(mdp, vdev, md_find_name(mdp, VDEV_PORT),
	    md_find_name(mdp, "fwd"), nodes);

	if (md_get_prop_val(mdp, nodes[0], VDISK_TIMEOUT_PROP,
	    &md_timeout) != 0) {
		timeout = VDISK_DEFAULT_TIMEOUT;
	} else {
		timeout = (int)md_timeout;
	}

	vdisk->timeout = timeout;

	/* get the config handle for this vdisk */
	if (md_get_prop_val(mdp, vdev, "cfg-handle", &cfg_handle) != 0) {
		DBG(zprintf("vdevice has no cfg-handle property\n"));
		Xfree(nodes);
		return (-1);
	}

	/* initialize the binding */
	(void) vdisk_init_binding(vdisk, cfg_handle);

	init_response(&resp);

	/* check if the volume can be bound */
	if ((rc = can_bind_vdisk(resp, vdisk)) != 0) {
		DBG(zprintf("unable to bind vdisk\n"));
		fini_response(resp);
		Xfree(nodes);
		return (rc);
	}

	fini_response(resp);

	if ((vds = vds_find_by_name(vdisk->service)) == NULL) {
		DBG(zprintf("no vds named %s\n", vdisk->service));
		Xfree(nodes);
		return (-1);
	}

	if ((vol = vds_find_volume_by_name(vds, vdisk->volname)) == NULL) {
		DBG(zprintf("no volume named %s\n", vdisk->volname));
		Xfree(nodes);
		return (-1);
	}

	for (i = 0; i < nbound; i++) {
		if (md_get_prop_val(mdp, nodes[i], "id", &port_id) != 0) {
			DBG(zprintf("vport has no id property\n"));
			Xfree(nodes);
			return (-1);
		}
		DBG(zprintf("binding port %"PRId64"\n", port_id));

		vdisk_port = vdisk_add_port(vdisk, port_id);

		/* get the LDC associated with this port */
		rc = ldc_mig_get_and_pull_ids(ldp, hvmdp, &ldp->guest_free_ldcs,
		    mdp, nodes[i], &vdisk_port->ldcs);
		if (rc != 0) {
			DBG(zprintf("Unable to get necessary LDC ID\n"));
			Xfree(nodes);
			return (rc);
		}

		/* create the vds port */
		vds_port_t *vds_port = vds_add_port(vds, vol, NEW_INSTANCE);
		vdisk_port->vds_port = vds_port;
		vds_port->vdisk_port = vdisk_port;

		/* assign the channel for the server */
		rc = ldc_mig_get_and_pull_peer_ids(ldp, hvmdp,
		    &vdisk_port->ldcs, &vds_port->ldcs,
		    &vds->ldp->guest_free_ldcs);

		if (rc != 0) {
			destroy_vdisk_binding(vdisk);
			DBG(zprintf("Insufficient free LDC channels on "
			    "LDom %s to bind vdisk %s\n", vds->ldp->namep,
			    vdisk->name));
			Xfree(nodes);
			return (rc);
		}
	}

	if (vds->ldp != vdisk->ldp) {
		vds->ldp->bound.num_clients++;
		vds->ldp->flag.gen_new_md = true;
		vds->ldp->flag.md_notify = true;
	}

	Xfree(nodes);

	DBG(zprintf("vdisk bound successfully\n"));

	/*
	 * Future: If the disk backend is not accessible from the
	 * target machine, but the vdsdev is setup, the domain after
	 * a migration hangs. I suppose the only way to guard against
	 * that if possible is do some type of backend device check.
	 * However, that is currently only possible if the vds is on
	 * the same machine as zeus. At a minimum, this will require
	 * the integration of the support for device checking (6447740).
	 */

	return (0);
}

/*
 * Add a vdisk from MD Node.
 */
void
vdisk_add_from_mdnode(ldom_t *ldp, md_t *guest_md, md_node_t *vdev)
{
	ident_val_t cfg_handle;
	uint64_t md_timeout_val;
	int num_vports, timeout;
	char *client_name, *devalias;
	md_node_t **nodes;
	vdisk_t *vdisk;
	int i;
	vdisk_binding_t *vdisk_binding;
	char *vds_name = NULL;
	char *vdsdev_name = NULL;
	int diskid = NO_NEW_ID;
	bool_t db_check = false;

	if (md_get_prop_val(guest_md, vdev, "cfg-handle", &cfg_handle) != 0)
		fatal("vdevice has no cfg-handle property");

	/*
	 * Get the disk name from the devalias property if it exists,
	 * otherwise get it from the database.
	 */
	if (md_get_prop_str(guest_md, vdev, DEVALIAS, &devalias) == 0)
		client_name = Xstrdup(devalias);
	else {
		client_name = get_vdisk_name_from_db(ldp, cfg_handle);
		if (client_name == NULL)
			client_name = vio_canonical_client_name(ldp,
			    vdisk_attrs.name, cfg_handle);
	}

	nodes = Xcalloc(md_node_count(guest_md), md_node_t *);
	num_vports = md_scan_dag(guest_md, vdev,
	    md_find_name(guest_md, VDEV_PORT),
	    md_find_name(guest_md, "fwd"), nodes);

	if (md_get_prop_val(guest_md, nodes[0], VDISK_TIMEOUT_PROP,
	    &md_timeout_val) != 0) {
		timeout = VDISK_DEFAULT_TIMEOUT;
	} else {
		timeout = (int)md_timeout_val;
	}

	/*
	 * Retrieve the user constraints from the DB
	 */
	db_check = get_vdisk_props_from_db(ldp, client_name, &vds_name,
	    &vdsdev_name, &diskid);

	/*
	 * Use the cfg_handle value as a diskid if failed to get the
	 * diskid from the DB
	 */
	if (diskid == NO_NEW_ID)
		diskid = cfg_handle;

	vdisk = vdisk_new(ldp, client_name, NULL, NULL, timeout, diskid);

	/*
	 * If the domain was in the process of being migrated, the
	 * migration bindings MD must be consulted to recreate the
	 * information about the service the vdisk was connected to.
	 * However, the actual binding of the vdisk does not happen
	 * until the migration recovery later in startup.
	 */
	if (ldp->flag.migration_source) {
		int		i;
		md_node_t	**vdevs;
		int		nvdevs;
		char		*name = NULL;
		char		*service = NULL;
		char		*vol = NULL;
		md_t		*bind_mdp = ldp->mig_state.bind_mdp;

		vdevs = Xcalloc(md_node_count(bind_mdp), md_node_t *);
		nvdevs = md_scan_dag(bind_mdp, NULL,
		    md_find_name(bind_mdp, "virtual-device"),
		    md_find_name(bind_mdp, "fwd"), vdevs);

		for (i = 0; i < nvdevs; i++) {
			md_get_prop_str(bind_mdp, vdevs[i], "name", &name);
			ASSERT(name != NULL);
			if (strcmp(name, "disk") != 0)
				continue;

			md_get_prop_str(bind_mdp, vdevs[i], "devalias", &name);
			if (strcmp(name, client_name) == 0)
				break;
		}

		ASSERT(i != nvdevs);

		md_get_prop_str(bind_mdp, vdevs[i], "service-name", &service);
		ASSERT(service);
		vdisk->service = Xstrdup(service);

		md_get_prop_str(bind_mdp, vdevs[i], "volume-name", &vol);
		ASSERT(vol);
		vdisk->volname = Xstrdup(vol);

		Xfree(vdevs);
		goto done;
	}

	(void) vdisk_init_binding(vdisk, cfg_handle);
	vdisk_binding = vdisk->binding;

	if (db_check == true) {
		vdisk_binding->work.db_service = vds_name;
		vdisk_binding->work.db_volume = vdsdev_name;
	}

	for (i = 0; i < num_vports; ++i) {
		uint64_t port_id;
		vdisk_port_t *vdisk_port;

		md_node_t *vport = nodes[i];

		if (md_get_prop_val(guest_md, vport, "id", &port_id) != 0)
			fatal("vport has no id property");
		vdisk_port = vdisk_add_port(vdisk, port_id);

		ldc_get_and_pull_ids(ldp, &ldp->guest_free_ldcs,
		    guest_md, vport, &vdisk_port->ldcs);
	}

done:
	Xfree(client_name);
	Xfree(nodes);
}

void
vdisk_gen_ldc_endpoints(ldom_t *ldp, md_t *hv_md, md_node_t *hv_guest,
    md_node_t *ldc_endpoints)
{
	vdisk_t *vdisk;
	vdisk_port_t *vdisk_port;
	vds_port_t *vds_port;

	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {
		if (vdisk->binding == NULL)
			continue;
		CHAIN_ITER(vdisk->binding->vdisk_ports, vdisk_port, link) {
			vds_port = vdisk_port->vds_port;
			vio_gen_ldcs_endpoint_nodes(hv_md, hv_guest,
			    ldc_endpoints, TYPE_GUEST, TYPE_GUEST,
			    &vdisk_port->ldcs, vds_port->vds->ldp->idp->id,
			    &vds_port->ldcs, NULL);
		}
	}
}

static void
gen_vdisk_guest_md_one(vdisk_t *vdisk, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	md_node_t *vdev;
	md_node_t *vport;
	vdisk_port_t *vdisk_port;

	vdev = md_link_new_node(md, VDEV_NAME, cdevs, "fwd", "back");

	/*
	 * Compatible property is an array of strings, though we only
	 * currently have one.
	 */
	md_add_string_as_data_property(md, vdev, "compatible",
	    vdisk_attrs.compatible);

	md_add_string_property(md, vdev, "device-type",
	    vdisk_attrs.device_type);
	md_add_string_property(md, vdev, "name", vdisk_attrs.md_name);
	md_add_value_property(md, vdev, "cfg-handle", vdisk_devid(vdisk));
	md_add_string_property(md, vdev, DEVALIAS, vdisk->name);

	/* create vport node(s) for the connection to the disk server(s) */
	CHAIN_ITER(vdisk->binding->vdisk_ports, vdisk_port, link) {
		vport = vio_gen_vdev_port_node(md, vdev, chan_endpoints,
		    vdisk_attrs.port_name, vdisk_port->port->id,
		    &vdisk_port->ldcs);
		if (vdisk->timeout > 0)
			md_add_value_property(md, vport, VDISK_TIMEOUT_PROP,
			    vdisk->timeout);
	}
}

void
vdisk_gen_guest_md(ldom_t *ldp, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vdisk_t *vdisk;

	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {
		gen_vdisk_guest_md_one(vdisk, md, cdevs, chan_endpoints);
	}
}

static void
vdisk_list_one(vdisk_t *vdisk, response_t *resp, dispctl_t *dctl)
{
	vdisk_binding_t *binding = vdisk->binding;
	vdisk_port_t *vdisk_port = NULL;
	vdisk_port_t *vport;
	bool_t mpg = false; /* multipathing enabled on bound vdisk ? */
	char *diskname, *volname, *devname, *timeout, *id, *server_ldom;
	char *mpgname = NULL;
	vds_port_t *vds_port;

	if (binding != NULL) {
		mpgroup_t *mpgrp;
		vdisk_port = CHAIN_START(binding->vdisk_ports);
		if ((mpgrp = vdisk_port->vds_port->vol->mpgroup) != NULL) {
			mpg = true;
			mpgname = mpgrp->name;
		}
	}

	/*
	 * TRANSLATION_NOTE
	 * The following are column headers in tables with fixed column widths.
	 * WIDTH  HEADER	MEANING
	 * 16	  "NAME"	name of disk client
	 * 27	  "VOLUME"	virtual disk volume name
	 * 4	  "TOUT"	virtual disk timeout value
	 * 4	  "ID"		disk id
	 * 10	  "DEVICE"	disk device name (e.g. disk@0)
	 * 14	  "SERVER"	name of LDom providing disk service
	 * 14	  "MPGROUP"	mpgroup name
	 * 3	  "LDC"		Logical Domain Channel number
	 * 6	  "PORT"	vdisk port# to each vdsdev in mpgroup
	 */

	if (!dctl->parseable && dctl->header) {
		dctl->header = 0;
		rprintf(resp, "%s\n", dctl_label(vdisk_attrs.name, dctl));
		rprintf(resp,
		    "    "FNAME" "FBNDNAME" "FTOUT" "FID, gettext("NAME"),
		    gettext("VOLUME"), gettext("TOUT"), gettext("ID"));
		if (!dctl->constraints) {
			rprintf(resp,
			    " "FDDEVICE" "FSERVER" "FVOLMGP, gettext("DEVICE"),
			    gettext("SERVER"), gettext("MPGROUP"));
			if (dctl->bindings && dctl->extended && binding != NULL)
				rprintf(resp, " %s", gettext("LDC"));
		}
		rprintf(resp, "\n");
	}

	diskname = vdisk->name;

	volname = zsprintf("%s@%s", vdisk->volname, vdisk->service);
	if (!dctl->constraints) {
		if (binding == NULL) {
			server_ldom = "";
			devname = zsprintf("");
		} else {
			vds_t *vds = vds_find_by_name(vdisk->service);
			vds_volume_t *vol = vds_find_volume_by_name(vds,
			    vdisk->volname);
			vds_port = vol->vds_port;
			server_ldom = vds_port->vds->ldp->namep;
			devname = zsprintf("%s@%"PRIu64"", vdisk_attrs.name,
			    vdisk_devid(vdisk));
		}
	}

	if (vdisk->timeout >= 0)
		timeout = zsprintf("%d", vdisk->timeout);
	else
		timeout = zsprintf("");

	if (vdisk->diskid != NO_NEW_ID)
		id = zsprintf("%d", vdisk->diskid);
	else
		id = zsprintf("");

	if (mpgname == NULL)
		mpgname = "";

	if (dctl->parseable) {
		rprintf(resp, "%s|name=%s",
		    dctl_label(vdisk_attrs.cli_name, dctl), diskname);
		rprintf(resp, "|vol=%s", volname);
		rprintf(resp, "|timeout=%s", timeout);
		if (!dctl->constraints) {
			rprintf(resp, "|dev=%s", devname);
			rprintf(resp, "|server=%s", server_ldom);
			rprintf(resp, "|mpgroup=%s", mpgname);
			if (dctl->bindings && (binding != NULL) &&
			    dctl->extended) {
				rprintf(resp, "|ldc=");
				ldcs_print(&vdisk_port->ldcs, resp);
			}
		}
		rprintf(resp, "|id=%s", id);
	} else {
		rprintf(resp,
		    "    "FNAME" "FBNDNAME" "FTOUT" "FID, diskname,
		    volname, timeout, id);
		if (!dctl->constraints) {
				rprintf(resp,
				    " "FDDEVICE" "FSERVER" "FVOLMGP,
				    devname, server_ldom, mpgname);

			if (dctl->bindings && dctl->extended &&
			    binding != NULL) {
				rprintf(resp, " ");
				ldcs_print(&vdisk_port->ldcs, resp);
			}
		}
	}

	rprintf(resp, "\n");

	Xfree(volname);
	if (!dctl->constraints) {
		Xfree(devname);
	}
	Xfree(timeout);

	if (dctl->bindings && (mpg == true)) {
		if (!dctl->parseable) {
			rprintf(resp, "        "FPORT" "FBNDNAME" "FSERVER,
			    gettext("PORT"),
			    gettext("MPGROUP VOLUME"),
			    gettext("MPGROUP SERVER"));
			if (dctl->extended)
				rprintf(resp, " %s", gettext("LDC"));
			rprintf(resp, "\n");
			dctl->header = 1;
		}

		CHAIN_ITER(binding->vdisk_ports, vport, link) {
			char *fmt;
			vds_port = vport->vds_port;
			volname = zsprintf("%s@%s", vds_port->vol->volname,
			    vds_port->vds->service);
			server_ldom = vds_port->vds->ldp->namep;
			if (dctl->parseable)
				fmt = "|port=%"PRIu64"|mpg-vol=%s"
				    "|mpg-server=%s";
			else
				fmt = "        "FPORTd" "FBNDNAME" "FSERVER;
			rprintf(resp, fmt, vdisk_port_id(vport), volname,
			    server_ldom);

			if (dctl->extended) {
				if (dctl->parseable)
					rprintf(resp, "|ldc=");
				else
					rprintf(resp, " ");
				ldcs_print(&vport->ldcs, resp);
			}
			rprintf(resp, "\n");
			Xfree(volname);
		}
	}
}

void
vdisk_list(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	vdisk_t *vdisk;

	if (dctl->oformat && !dctl->odisk)
		return;

	if (CHAIN_LENGTH(ldp->vdisks) == 0)
		return;

	dctl->header = 1;

	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {
		vdisk_list_one(vdisk, resp, dctl);
	}

	if (!dctl->parseable)
		rprintf(resp, "\n");
}

/*
 * The vdisk client is linked with the corresponding vds-port, i.e the volume
 * in the vds server.
 */
extern ident_val_t vds_devid(vds_t *);
extern void vds_mpgroup_add_vol(char *, vds_volume_t *);
static int
link_vdisk_client(ldom_t *ldp, vdisk_t *vdisk)
{
	vdisk_port_t *vdisk_port;
	ldc_epdb_t *ep;
	ldc_chan_t *ldc;
	vdisk_binding_t *vdisk_binding = vdisk->binding;
	int num_ports = CHAIN_LENGTH(vdisk_binding->vdisk_ports);

	/*
	 * RULES:
	 * With new MDs containing vol-name in vds-port nodes, the vdsdev
	 * should get the volume name during link time here. Otherwise,
	 * if the DB could be read, the names should come from the DB. If
	 * neither holds true, the names need to be derived appropriately,
	 * mpgroup-name = <guest-name>-<vdisk>-<instance>
	 * vol-name = <guest-name>-<vdsdev>-<instance>
	 */

	CHAIN_ITER(vdisk_binding->vdisk_ports, vdisk_port, link) {
		char *mpgroup = NULL;
		vds_port_t *vds_port;
		vds_volume_t *vol;

		ldc = CHAIN_START(vdisk_port->ldcs.ldcs);
		if ((ep = ldc_epdb_find(ldp->idp->id, ldc->chan->id)) != NULL)
			vds_port = vds_port_find_by_ldcid(ep->peer->ldp,
			    ep->peer->ldc);
		if ((ep == NULL) || (vds_port == NULL))
			fatal("Could not find matching vds port for vdisk %s "
			    "on LDom %s", vdisk->name, ldp->namep);

		/*
		 * Cross link the client & server port binding
		 */
		vdisk_port->vds_port = vds_port;
		vds_port->vdisk_port = vdisk_port;

		if (vds_port->vds->ldp != ldp)
			vds_port->vds->ldp->bound.num_clients++;

		if (vdisk_port == CHAIN_START(vdisk_binding->vdisk_ports)) {
			/*
			 * Fill up the vdisk constraints only once
			 */
			ASSERT(vdisk->service == NULL &&
			    vdisk->volname == NULL);
			if (vdisk_binding->work.db_service != NULL &&
			    vdisk_binding->work.db_volume != NULL) {
				vdisk->service = vdisk_binding->work.db_service;
				vdisk->volname = vdisk_binding->work.db_volume;
				vdisk_binding->work.db_service = NULL;
				vdisk_binding->work.db_volume = NULL;
			} else {
				/*
				 * With newer MD, the port node contains the
				 * volume and service information.
				 */
				vdisk->service =
				    Xstrdup(vds_port->vds->service);

				if (vds_port->work.md_volume != NULL) {
					vdisk->volname =
					    Xstrdup(vds_port->work.md_volume);
				} else {
					/*
					 * Old MD, no DB, Need to generate
					 * canonical name for volume
					 */
					vdisk->volname =
					    zsprintf("%s-%s%"PRIu64"-vol%d",
					    ldp->namep, vds_attrs.name,
					    vds_devid(vds_port->vds),
					    vds_port->port->id);
				}
			}

			database_ldom_add_disk_client(ldp, vdisk->name,
			    vdisk->service, vdisk->volname, vdisk->timeout,
			    vdisk->diskid);

		}


		/*
		 * Fill up the vds volume name and mpgroup for each
		 * vdisk port bound to them
		 */
		vol = vds_port->vol;
		if (vds_port->work.md_volume != NULL) {
			vol->volname = vds_port->work.md_volume;
			vds_port->work.md_volume = NULL;
		}

		/*
		 * Still no vdedev name (older MD) ? Get the vdisk one based
		 * on the vdisk constraint from the DB
		 */
		if (vol->volname == NULL)
			vol->volname = Xstrdup(vdisk->volname);

		/*
		 * Since the mpgroup is part of the MD, no need to get this
		 * from DB
		 */

		/*
		 * Check if there is an mpgroup associated with the vdsdev
		 * (vds-port).
		 */
		if (vds_port->work.md_mpgroup != NULL) {
			mpgroup = vds_port->work.md_mpgroup;
		} else {
			/*
			 * There must be exactly one vdisk port if the
			 * volume link to isn't part of an mpgroup
			 */
			ASSERT(num_ports == 1);
		}
		vds_port->work.md_mpgroup = NULL;

		/*
		 * If an mpgroup is specified, add this volume to the mpgroup
		 */
		if (mpgroup != NULL)
			vds_mpgroup_add_vol(mpgroup, vol);

		database_ldom_add_disk_service_vol(vds_port->vds->ldp,
		    vds_port->vds->service, vol->volname, vol->blkdev,
		    vol->options, mpgroup);

		Xfree(mpgroup);
	}

	return (0);
}

int
link_vdisk_clients(ldom_t *ldp)
{
	vdisk_t *vdisk;

	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {
		link_vdisk_client(ldp, vdisk);
	}

	return (0);
}

static void
teardown_vdisk_binding(vdisk_t *vdisk, bool_t unbinding)
{
	vdisk_binding_t *vdisk_binding = vdisk->binding;
	vds_port_t *vds_port;
	vdisk_port_t *vdisk_port, *temp;

	if (vdisk_binding == NULL)
		return;

	CHAIN_ITER_VOL(vdisk_binding->vdisk_ports, vdisk_port, temp, link) {
		vds_port = vdisk_port->vds_port;
		if (unbinding) {
			if (vds_port->vds->ldp != vdisk->ldp) {
				vds_port->vds->ldp->bound.num_clients--;
				vds_port->vds->ldp->flag.gen_new_md = true;
				vds_port->vds->ldp->flag.md_notify = true;
			}

			vds_unbind_port(vds_port);
			ldcs_stage(&vdisk_port->ldcs);
		} else {
			if (vds_port != NULL)
				vds_destroy_port(vds_port);
			ldcs_free(&vdisk_port->ldcs);
		}
		CHAIN_UNLINK(vdisk_binding->vdisk_ports, vdisk_port, link);
		free_res(vdisk_port->port);
		Xfree(vdisk_port);
	}

	free_res(vdisk_binding->cfg_handle);
	destroy_res_pool(&vdisk_binding->free_ports);
	Xfree(vdisk_binding);
	vdisk->binding = NULL;
}

static void
destroy_vdisk_binding(vdisk_t *vdisk)
{
	teardown_vdisk_binding(vdisk, false);
}

static void
unbind_vdisk(vdisk_t *vdisk)
{
	teardown_vdisk_binding(vdisk, true);
}

static void
destroy_vdisk(vdisk_t *vdisk)
{
	ASSERT(vdisk->binding == NULL);

	CHAIN_UNLINK(vdisk->ldp->vdisks, vdisk, ldom);

	Xfree(vdisk->name);
	Xfree(vdisk->service);
	Xfree(vdisk->volname);

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(vdisk->ldp->vdisks) == 0)
		destroy_res_pool(&vdisk->ldp->bound.vdisk_cfg_handles);
#endif

	Xfree(vdisk);
}

void
vdisk_delete_all(ldom_t *ldp)
{
	vdisk_t *vdisk, *vdisk_next;

	for (vdisk = CHAIN_START(ldp->vdisks); vdisk != NULL;
	    vdisk = vdisk_next) {
		vdisk_next = CHAIN_NEXT(vdisk->ldom);
		destroy_vdisk(vdisk);
	}
}

/*
 * ensure that the vds server and volume exists
 */
static int
can_bind_vdisk(response_t *resp, vdisk_t *vdisk)
{
	vds_t *vds;

	/*
	 * We must be able to bind to the vds and volume specified
	 */
	vds = vds_find_by_name(vdisk->service);
	if (vds == NULL || vds->binding == NULL) {
		rprintf(resp, gettext("The virtual disk service \"%s\" for "
		    "virtual disk \"%s\",\ndoes not exist or is not bound.\n"),
		    vdisk->service, vdisk->name);
		return (ERR_NO_SERVICE);
	}

	if (vds_find_volume_by_name(vds, vdisk->volname) == NULL) {
		rprintf(resp, gettext("The volume \"%s\" does not exist on "
		    "virtual disk server \"%s\".\n"),
		    vdisk->volname, vdisk->service);
		return (ERR_VDS_VOLUME_NOEXISTS);
	}

	return (0);
}

/*
 * check if a volume is available to be bound to a vdisk
 */
static int
can_bind_vol(response_t *resp, vdisk_t *vdisk, vds_volume_t *vol)
{
	vds_t *vds = vol->vds;

	if (vds->ldp->flag.locked) {
		rprintf(resp, gettext("Cannot add virtual disk %s because that "
		    "would modify locked LDom %s\n"), vdisk->name,
		    vds->ldp->namep);
		return (ERR_LDOM_LOCKED);
	}

	if (vol->vds_port != NULL) {
		rprintf(resp, gettext("The volume \"%s\" is already in use.\n"),
		    vdisk->volname);
		return (ERR_VDS_VOLUME_BUSY);
	}

	return (0);
}

static int
check_vdisk(vdisk_t *vdisk, response_t *resp)
{
	vds_t *vds;
	vds_volume_t *vol;
	int err = 0;

	if ((err = can_bind_vdisk(resp, vdisk)) != 0)
		return (err);

	vds = vds_find_by_name(vdisk->service);
	vol = vds_find_volume_by_name(vds, vdisk->volname);

	if (vol->mpgroup == NULL) {
		if (vds->ldp->state < BOUND) {
			rprintf(resp, gettext("Cannot add virtual disk "
			    "\"%s\" because the LDom %s hosting the service %s "
			    "is not bound\n"), vdisk->name, vds->ldp->namep,
			    vdisk->service);

		}

		if ((err = can_bind_vol(resp, vdisk, vol)) != 0)
			return (err);

		/*
		 * One more LDC channel is needed locally and on the
		 * disk server.
		 */
		vdisk->ldp->work.num_ldcs++;
		vds->ldp->work.num_ldcs++;

	} else {
		vds_volume_t *volp;
		/*
		 * Cycle through all the volumes in the same mpgroup chain
		 * and make sure all the volumes can be bound to this vdisk
		 * volume
		 */
		CHAIN_ITER(vol->mpgroup->volumes, volp, mpgroup_link) {
			if (volp->vds->ldp->state < BOUND) {
				char *volname = zsprintf("%s@%s",
				    volp->volname, volp->vds->service);
				rprintf(resp, gettext("The vdsdev \"%s\" for "
				    "virtual disk \"%s\" which is part of\n"
				    "mpgroup %s does not exist or is not "
				    "bound\n"), volname, vdisk->name,
				    vol->mpgroup->name);
				Xfree(volname);
				return (ERR_VDS_VOLUME_NOEXISTS);
			} else {
				if ((err = can_bind_vol(resp, vdisk,
				    volp)) == 0) {
					vdisk->ldp->work.num_ldcs++;
					volp->vds->ldp->work.num_ldcs++;
				} else {
					return (err);
				}
			}
		}
	}

	return (0);
}

#if BIND_CHECK
static int
add_vdisk_check(vdisk_t *vdisk, response_t *resp)
{
	ldom_t *ldp;
	int rc;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		ldp->work.num_ldcs = 0;
	}

	if ((rc = check_vdisk(vdisk, resp)) != 0)
		return (rc);

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		DBG(zprintf("%s: %s needs %d new channels\n", __func__,
		    ldp->namep, ldp->work.num_ldcs));

		if ((ldp->work.num_ldcs > 0) && ldp->flag.locked) {
			rprintf(resp, gettext("Cannot add virtual disk "
			    "\"%s\" because that would modify locked LDom "
			    "%s\n"), vdisk->name, ldp->namep);
			return (ERR_LDOM_LOCKED);
		}

		if (available_res_ids(&ldp->guest_free_ldcs) <
		    ldp->work.num_ldcs) {
			rprintf(resp, gettext("Insufficient free LDC "
			    "channels to bind virtual disk %s\n"),
			    vdisk->name);
			return (ERR_LDC_NO_IDS);
		}
	}

	return (0);
}
#endif

int
vdisk_bind_check(ldom_t *ldp, response_t *resp)
{
	vdisk_t *vdisk, *local_vdisk;
	int rc;

	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {
		if ((rc = check_vdisk(vdisk, resp)) != 0)
			return (rc);
		/*
		 * Check that volume is not also being used by any other
		 * vdisks which are still to be bound.
		 */
		for (local_vdisk = CHAIN_NEXT(vdisk->ldom); local_vdisk != NULL;
		    local_vdisk = CHAIN_NEXT(local_vdisk->ldom)) {
			if ((strcmp(vdisk->service,
			    local_vdisk->service) == 0) &&
			    (strcmp(vdisk->volname,
			    local_vdisk->volname) == 0)) {
				rprintf(resp, gettext("The volume \"%s\" is "
				    "already in use\n"), vdisk->volname);
				return (ERR_VDS_VOLUME_BUSY);
			}
		}
	}

	return (0);
}

/*
 * Bind a vdisk to a volume (or one of the volumes in a failover group)
 */
static int
bind_vdisk_to_vol(response_t *resp, vdisk_t *vdisk, vds_volume_t *vol)
{
	vds_t *vds = vol->vds;
	vds_port_t *vds_port = NULL;
	vdisk_port_t *vdisk_port = NULL;
	int rc = 0;

	if ((rc = can_bind_vol(resp, vdisk, vol)) != 0)
		return (rc);

	if ((vdisk_port = vdisk_add_port(vdisk, NEW_INSTANCE)) == NULL) {
		destroy_vdisk_binding(vdisk);
		rprintf(resp, gettext("Failed to add a vdisk port on "
		    "LDom %s to bind vdisk %s\n"), vdisk->ldp->namep,
		    vdisk->name);
		return (ERR_GENERIC);
	}

	rc = ldcs_chan_alloc(&vdisk_port->ldcs, &vdisk->ldp->guest_free_ldcs);
	if (rc != 0) {
		destroy_vdisk_binding(vdisk);
		rprintf(resp, gettext("Insufficient free LDC channels on "
		    "LDom %s to bind vdisk %s\n"), vdisk->ldp->namep,
		    vdisk->name);
		return (rc);
	}

	vds_port = vds_add_port(vds, vol, NEW_INSTANCE);
	vdisk_port->vds_port = vds_port;
	vds_port->vdisk_port = vdisk_port;

	rc = ldcs_chan_alloc(&vds_port->ldcs, &vds->ldp->guest_free_ldcs);
	if (rc != 0) {
		destroy_vdisk_binding(vdisk);
		rprintf(resp, gettext("Insufficient free LDC channels on "
		    "LDom %s to bind vdisk %s\n"), vds->ldp->namep,
		    vdisk->name);
		return (rc);
	}

	if (vds->ldp != vdisk->ldp) {
		vds->ldp->bound.num_clients++;
		vds->ldp->flag.gen_new_md = true;
		vds->ldp->flag.md_notify = true;
	}

	return (0);
}

static int
bind_vdisk(vdisk_t *vdisk, response_t *resp)
{
	vds_t *vds;
	vds_volume_t *vol, *volp;
	int rc = 0;

	/*
	 * We must be able to bind to the vds and volume specified
	 */
	if ((rc = can_bind_vdisk(resp, vdisk)) != 0)
		return (rc);

	vds = vds_find_by_name(vdisk->service);
	vol = vds_find_volume_by_name(vds, vdisk->volname);

	/*
	 * If the target volume is part of an mpgroup, Cycle through all the
	 * volumes in the same mpgroup chain and make sure all the volumes can
	 * be bound to this vdisk volume
	 */
	if (vol->mpgroup != NULL) {
		CHAIN_ITER(vol->mpgroup->volumes, volp, mpgroup_link) {
			if (volp->vds->ldp->state < BOUND) {
				char *volname = zsprintf("%s@%s",
				    volp->volname, volp->vds->service);
				rprintf(resp, gettext("The vdsdev \"%s\" for "
				    "virtual disk \"%s\" which is part of\n"
				    "mpgroup %s does not exist or is not "
				    "bound\n"), volname, vdisk->name,
				    vol->mpgroup->name);
				Xfree(volname);
				return (ERR_VDS_VOLUME_NOEXISTS);
			}
		}
	}

	/*
	 * Initiate the bind process. For volume in an mpgroup, the vdisk need
	 * to be bound to each *bound* volume in the same mpgroup
	 *
	 * if the diskid is specified with the CLI, the diskid value
	 * will be used to pull from the cfg_handle pool. If it fails
	 * to get the cfg_handle, generate the duplicate id error message.
	 */
	if (vdisk->diskid != NO_NEW_ID) {
		if ((rc = vdisk_init_binding(vdisk, vdisk->diskid)) != 0) {
			rprintf(resp, MSG_ERR_ID_EXISTS);
			return (rc);
		}
	} else {
		rc = vdisk_init_binding(vdisk, NEW_INSTANCE);
		ASSERT(rc == 0);
	}

	if (vol->mpgroup == NULL) {
		rc = bind_vdisk_to_vol(resp, vdisk, vol);
	} else {
		CHAIN_ITER(vol->mpgroup->volumes, volp, mpgroup_link)
			if ((rc = bind_vdisk_to_vol(resp, vdisk, volp)) != 0)
				break;
	}

	/*
	 * Cleanup in case of bind failure
	 */
	if (rc != 0)
		destroy_vdisk_binding(vdisk);

	return (rc);
}

int
vdisk_bind_all(ldom_t *ldp, response_t *resp)
{
	vdisk_t *vdisk;
	int rc;

	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {
		if ((rc = bind_vdisk(vdisk, resp)) != 0)
			return (rc);
	}

	return (0);
}

/*ARGSUSED1*/
int
vdisk_unbind_all(ldom_t *ldp, response_t *resp)
{
	vdisk_t *vdisk;

	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {
		unbind_vdisk(vdisk);
	}

	return (0);
}

/*
 * Called from the add-vdisk CLI.
 */
int
vdisk_add(ldom_t *ldp, char *diskname, char *service, char *volname,
    int timeout, int diskid, response_t *resp)
{
	vdisk_t *vdisk;
	int	rc;

	if (vdisk_find_by_name(ldp, diskname) != NULL) {
		rprintf(resp, gettext("%s client named %s already exists "
		    "in guest %s\n"), vdisk_attrs.cli_name,
		    diskname, ldp->namep);
		return (ERR_CLIENT_EXISTS);
	}

	/*
	 * if id is not specified in the CLI or the domain is still inactive,
	 * get the lowest free id value from the DB to use as a diskid
	 */
	if (diskid == NO_NEW_ID) {
		diskid = database_ldom_get_lowest_id(ldp, DISK_NODE);
	} else if (database_ldom_check_dup_id(ldp, DISK_NODE, diskid)
	    == true) {
		/*
		 * make sure that the same id doesn't already exist in the DB
		 */
		rprintf(resp, MSG_ERR_ID_EXISTS);
		return (ERR_DUP_ID);
	}

	vdisk = vdisk_new(ldp, diskname, service, volname, timeout, diskid);

	if (ldp->state >= BOUND) {
#if BIND_CHECK
		if ((rc = add_vdisk_check(vdisk, resp)) != 0) {
			destroy_vdisk(vdisk);
			return (rc);
		}
#endif
		if ((rc = bind_vdisk(vdisk, resp)) != 0) {
			destroy_vdisk(vdisk);
			return (rc);
		}

		if (zeus.mode == ZEUS_MGMT_MODE) {
			/*
			 * The service domain(s) need to be notified to
			 * load the vdsdev.
			 */
			vdisk_flag_svc_domains(vdisk);

			if (ldp->state == BOUND) {
				/* do reconfig now */
				seq_download_mds_do_config(ldp);
				free_all_staged_ldcs(ldp);

			} else if (vio_check_dr_capable(ldp, resp)) {
				/* DR in the vdisk */
				rc = vio_dr_add(ldp, vdisk_attrs.md_name,
				    vdisk_devid(vdisk), resp);
				/*
				 * If the DR operation failed and the dr-vio
				 * domain service determines that it is OK to
				 * rollback we remove the disk again.
				 */
				if (rc != 0) {
					if (rc == ERR_DR_VIO_ROLLBACK) {
						/*
						 * remove the vio device info
						 * from the guest MD
						 */
						(void) vio_dr_rm_phase1(ldp,
						    vdisk_attrs.md_name,
						    vdisk_devid(vdisk), resp);

						unbind_vdisk(vdisk);
						destroy_vdisk(vdisk);

						(void) vio_dr_rm_phase2(ldp);
					}
					return (rc);
				}
			} else {
				/* handle as a delayed reconfig */
				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_ADD, resp);
			}
		}
	}


	database_ldom_add_disk_client(ldp, vdisk->name,
	    vdisk->service, vdisk->volname, vdisk->timeout, vdisk->diskid);

	return (0);
}

/*
 * Called from the set-vdisk CLI.
 */
int
vdisk_modify(ldom_t *ldp, char *diskname, char *service, char *volname,
    int timeout, response_t *resp)
{
	vdisk_t *vdisk;

	vdisk = vdisk_find_by_name(ldp, diskname);
	if (vdisk == NULL) {
		rprintf(resp, gettext("%s client named %s does not exist "
		    "in guest %s\n"), vdisk_attrs.cli_name,
		    diskname, ldp->namep);
		return (ERR_NO_CLIENT);
	}

	/*
	 *  check if anything changed
	 */

	if (service != NULL && strcmp(service, vdisk->service) == 0)
		service = NULL;

	if (volname != NULL && strcmp(volname, vdisk->volname) == 0)
		volname = NULL;

	if (timeout != VDISK_NO_NEW_TIMEOUT && timeout == vdisk->timeout)
		timeout = VDISK_NO_NEW_TIMEOUT;

	if (timeout == VDISK_NO_NEW_TIMEOUT && service == NULL &&
	    volname == NULL) {
		return (0);
	}

	/* Live client rebinding is not supported */
	if ((ldp->state >= BOUND) && (zeus.mode == ZEUS_MGMT_MODE) &&
	    volname != NULL) {
		rprintf(resp,
		    gettext("Modifying a bound vdisk is not supported\n"));
		return (ERR_VDISK_NO_MODIFY_BOUND);
	}

	/*
	 * Update the client from DB
	 */
	database_ldom_modify_disk_client(ldp, diskname, service, volname,
	    timeout);

	if (service != NULL) {
		if (vdisk->service != NULL)
			Xfree(vdisk->service);
		vdisk->service = Xstrdup(service);
	}

	if (volname != NULL) {
		if (vdisk->volname != NULL)
			Xfree(vdisk->volname);
		vdisk->volname = Xstrdup(volname);
	}

	if (timeout != VDISK_NO_NEW_TIMEOUT) {
		vdisk->timeout = timeout;
	}

	/* reconfig vdisk timeout property */
	if (ldp->state >= BOUND) {
		if ((ldp->state > BOUND) && (zeus.mode == ZEUS_MGMT_MODE)) {
			/* handle as a delayed reconfig */
			initiate_or_update_delayed_reconfig(ldp,
			    DEL_RECONFIG_OTHER, resp);
		} else if (zeus.mode == ZEUS_MGMT_MODE) {
			/* do reconfig now */
			seq_download_mds_do_config(ldp);
			free_all_staged_ldcs(ldp);
		}
	}

	return (0);
}

/*
 * This function removes the disk from the DB and destroys the in-memory copy
 */
static void
vdisk_remove_do_teardown(vdisk_t *vdisk)
{
	ASSERT(vdisk != NULL);

	database_ldom_del_disk_client(vdisk->ldp, vdisk->name);
	unbind_vdisk(vdisk);
	destroy_vdisk(vdisk);
}

/*
 * Called from the remove-vdisk CLI.
 */
/*ARGSUSED2*/
int
vdisk_remove(ldom_t *ldp, char *diskname, bool_t force, response_t *resp)
{
	vdisk_t *vdisk;
	int	rc;

	if ((vdisk = vdisk_find_by_name(ldp, diskname)) == NULL) {
		if (resp)
			rprintf(resp, gettext("%s named %s not found\n"),
			    vdisk_attrs.cli_name, diskname);
		return (ERR_NO_CLIENT);
	}

	switch (zeus.mode) {
	case ZEUS_MGMT_MODE:
		if (ldp->state < BOUND) {
			vdisk_remove_do_teardown(vdisk);

		} else if (ldp->state == BOUND) {
			/* do reconfig now */
			vdisk_remove_do_teardown(vdisk);
			seq_rm_reconfigure(ldp);
			free_all_staged_ldcs(ldp);

		} else if (vio_check_dr_capable(ldp, resp)) {
			/* Attempt DR operation */
			rc = vio_dr_rm_phase1(ldp, vdisk_attrs.md_name,
			    vdisk_devid(vdisk), resp);
			if (rc != 0)
				return (rc);

			vdisk_flag_svc_domains(vdisk);

			vdisk_remove_do_teardown(vdisk);
			vio_dr_rm_phase2(ldp);

		} else {
			/* handle as a delayed reconfig operation */
			vdisk_remove_do_teardown(vdisk);
			initiate_or_update_delayed_reconfig(ldp,
			    DEL_RECONFIG_REMOVE_2PHASE, resp);
		}
		break;

	case ZEUS_CONFIG_MODE:
		vdisk_remove_do_teardown(vdisk);
		free_all_staged_ldcs(ldp);
		break;

	default:
		ASSERT(0);
		return (ERR_GENERIC);
	}

	return (0);
}

void
vdisk_print_bindings(ldom_t *ldp, response_t *resp)
{
	char		idstr[UINT64_STR_LEN];
	vdisk_t		*vdisk;
	vdisk_port_t	*port;
	vdisk_binding_t	*bind;
	ldc_chan_t	*ldc;

	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {

		bind = vdisk->binding;

		(void) snprintf(idstr, UINT64_STR_LEN, "vdisk%"PRId64,
		    bind->cfg_handle->id);
		rprintf(resp, "  %-11s", idstr);

		CHAIN_ITER(bind->vdisk_ports, port, link) {
			CHAIN_ITER(port->ldcs.ldcs, ldc, link) {
				rprintf(resp, "%2"PRId64" ", ldc->chan->id);
			}
		}

		rprintf(resp, "  %s@%s\n", vdisk->volname, vdisk->service);
	}
}

int
vdisk_mig_add_bindings(ldom_t *ldp, md_t *mdp)
{
	int		i;
	int		rc = -1;
	md_node_t	**vdevs;
	int		nvdevs;
	char		*name;
	vdisk_t		*vdisk;

	/*
	 * Scan through the MD for all virtual device nodes.
	 */
	vdevs = Xmalloc(md_node_count(mdp) * sizeof (md_node_t *));
	nvdevs = md_scan_dag(mdp, NULL, md_find_name(mdp, VDEV_NAME),
	    md_find_name(mdp, "fwd"), vdevs);

	ASSERT(nvdevs > 0);

	for (i = 0; i < nvdevs; i++) {
		if (md_get_prop_str(mdp, vdevs[i], "name", &name) < 0) {
			DBG(zprintf("virtual-device node has no name prop\n"));
			goto done;
		}

		if (strcmp(name, "disk") != 0)
			continue;

		if (md_get_prop_str(mdp, vdevs[i], "devalias", &name) < 0) {
			DBG(zprintf("vdisk node has no devalias prop\n"));
			goto done;
		}

		if ((vdisk = vdisk_find_by_name(ldp, name)) == NULL) {
			DBG(zprintf("%s: Unknown vdisk %s\n", __func__, name));
			goto done;
		}

		/* add the service name */
		if (vdisk->service != NULL) {
			md_add_string_property(mdp, vdevs[i], "service-name",
			    vdisk->service);
		}

		/* add the volume name */
		if (vdisk->volname != NULL) {
			md_add_string_property(mdp, vdevs[i], "volume-name",
			    vdisk->volname);
		}
	}

	rc = 0;

done:
	Xfree(vdevs);
	return (rc);
}
