/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)vdpcc.c	1.6	09/04/22 SMI"

#include <string.h>

#include "ldmd_types.h"
#include "allocate.h"
#include "chain.h"
#include "database.h"
#include "errors.h"
#include "fatal.h"
#include "ldom.h"
#include "resource_id.h"
#include "vio.h"
#include "zprintf.h"

struct vdpcc_attributes vdpcc_attrs = {
	.name =		"vdpcc",
	.compatible =	"SUNW,sun4v-data-plane-channel",
	.md_name =	"virtual-data-plane-channel-client",
	.cli_name =	"vdpcc",
	.device_type =	"serial",
	.port_name =	"vdpc-port"
};

int
vdpcc_devid(vdpcc_t *vdpcc)
{
	if (vdpcc->binding == NULL)
		return (INVALID_DEVID);
	else
		return (vdpcc->binding->cfg_handle->id);
}

static res_pool_t *
vdpcc_cfg_handles(ldom_t *ldp)
{
#if VDEV_CFG_HANDLE
	return (&ldp->free_cfg_handles);
#else
	return (&ldp->bound.vdpcc_cfg_handles);
#endif
}

vdpcc_t *
vdpcc_find_by_name(ldom_t *ldp, char *vdpcc_name)
{
	vdpcc_t *vdpcc;

	CHAIN_ITER(ldp->vdpccs, vdpcc, ldom) {
		if (strcmp(vdpcc_name, vdpcc->name) == 0)
			break;
	}

	return (vdpcc);
}

vdpcc_t *
vdpcc_new(ldom_t *ldp, char *name, char *service)
{
	vdpcc_t *vdpcc;

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(ldp->vdpccs) == 0) {
		create_res_pool(&ldp->bound.vdpcc_cfg_handles,
		    IDENT_MIN, IDENT_MAX, Res_sorted);
		name_ldom_res_pool(&ldp->bound.vdpcc_cfg_handles, ldp,
		    "vdpcc cfg handles");
	}
#endif

	vdpcc = Xmalloc(sizeof (vdpcc_t));

	vdpcc->ldp = ldp;
	vdpcc->name = Xstrdup(name);
	vdpcc->service = (service != NULL) ? Xstrdup(service) : NULL;
	vdpcc->binding = NULL;

	CHAIN_ADD(ldp->vdpccs, vdpcc, ldom);

	return (vdpcc);
}

static void
init_vdpcc_binding(vdpcc_t *vdpcc, int cfg_handle)
{
	ldom_t *ldp = vdpcc->ldp;
	vdpcc_binding_t *binding;

	binding = Xmalloc(sizeof (vdpcc_binding_t));

	binding->cfg_handle = vio_get_cfg_handle(cfg_handle,
	    vdpcc_cfg_handles(ldp));
	if (binding->cfg_handle == NULL)
		fatal("Could not obtain vdpcc cfg_handle %d on LDom %s\n",
		    cfg_handle, ldp->namep);

	binding->vdpcs_port = NULL;
	CHAIN_INIT(binding->ldcs.ldcs);

	vdpcc->binding = binding;
}

/*
 * Add a vdpcc from MD Node.
 */
void
vdpcc_add_from_mdnode(ldom_t *ldp, md_t *guest_md, md_node_t *vdev)
{
	uint64_t cfg_handle;
	int num_vports;
	char *client_name;
	md_node_t **nodelistp;
	vdpcc_t *vdpcc;

	if (md_get_prop_val(guest_md, vdev, "cfg-handle", &cfg_handle) != 0)
		fatal("vdevice has no cfg-handle property");

	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(guest_md));
	num_vports = md_scan_dag(guest_md, vdev,
	    md_find_name(guest_md, "virtual-device-port"),
	    md_find_name(guest_md, "fwd"), nodelistp);

	ASSERT(num_vports == 1);

	/*
	 * Get the client name from the virtual-device-port node vdpc-svc-name
	 * property.
	 */
	if (md_get_prop_str(guest_md, nodelistp[0], "vdpc-svc-name",
	    &client_name) != 0)
		fatal("vdpc client virtual-device-port has no vdpc-svc-name "
		    "property");

	vdpcc = vdpcc_new(ldp, client_name, NULL);
	init_vdpcc_binding(vdpcc, cfg_handle);

	ldc_get_and_pull_ids(ldp, &ldp->guest_free_ldcs,
	    guest_md, nodelistp[0], &vdpcc->binding->ldcs);

	Xfree(nodelistp);
}

static void
teardown_vdpcc_binding(vdpcc_t *vdpcc, bool_t unbinding)
{
	if (vdpcc->binding == NULL)
		return;

	vdpcs_port_t *vdpcs_port = vdpcc->binding->vdpcs_port;

	if (unbinding) {
		if (vdpcs_port->vdpcs->ldp != vdpcc->ldp) {
			vdpcs_port->vdpcs->ldp->bound.num_clients--;
			vdpcs_port->vdpcs->ldp->flag.gen_new_md = true;
			vdpcs_port->vdpcs->ldp->flag.md_notify = true;
		}

		vdpcs_unbind_port(vdpcs_port);
		ldcs_stage(&vdpcc->binding->ldcs);
	} else {
		if (vdpcs_port != NULL)
			vdpcs_destroy_port(vdpcs_port);
		ldcs_free(&vdpcc->binding->ldcs);
	}

	free_res(vdpcc->binding->cfg_handle);
	Xfree(vdpcc->binding);
	vdpcc->binding = NULL;
}

static void
destroy_vdpcc_binding(vdpcc_t *vdpcc)
{
	teardown_vdpcc_binding(vdpcc, false);
}

static void
unbind_vdpcc(vdpcc_t *vdpcc)
{
	teardown_vdpcc_binding(vdpcc, true);
}

static void
destroy_vdpcc(vdpcc_t *vdpcc)
{
	ASSERT(vdpcc->binding == NULL);

	CHAIN_UNLINK(vdpcc->ldp->vdpccs, vdpcc, ldom);

	Xfree(vdpcc->name);
	Xfree(vdpcc->service);

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(vdpcc->ldp->vdpccs) == 0)
		destroy_res_pool(&vdpcc->ldp->bound.vdpcc_cfg_handles);
#endif

	Xfree(vdpcc);
}

#if BIND_CHECK
static int
check_vdpcc(vdpcc_t *vdpcc, response_t *resp)
{
	vdpcs_t *vdpcs;

	vdpcs = vdpcs_find_by_name(vdpcc->service);
	if (vdpcs == NULL ||
	    ((vdpcs->ldp != vdpcc->ldp) && (vdpcs->binding == NULL))) {
		rprintf(resp, gettext("The virtual data plane service \"%s\" "
		    "for client \"%s\",\ndoes not exist or is "
		    "not bound\n"), vdpcc->service, vdpcc->name);
		return (ERR_NO_SERVICE);
	}

	/*
	 * One more LDC channel is needed locally and on the server.
	 */
	vdpcc->ldp->work.num_ldcs++;
	vdpcs->ldp->work.num_ldcs++;

	return (0);
}

static int
add_vdpcc_check(vdpcc_t *vdpcc, response_t *resp)
{
	ldom_t *ldp;
	int rc;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		ldp->work.num_ldcs = 0;
	}

	if ((rc = check_vdpcc(vdpcc, resp)) != 0)
		return (rc);

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		DBG(zprintf("%s: %s needs %d new channels\n", __func__,
		    ldp->namep, ldp->work.num_ldcs));

		if ((ldp->work.num_ldcs > 0) && ldp->flag.locked) {
			rprintf(resp, gettext("Cannot add virtual data plane"
			    "client %s because that would modify locked LDom "
			    "%s\n"), vdpcc->name, ldp->namep);
			return (ERR_LDOM_LOCKED);
		}

		if (available_res_ids(&ldp->guest_free_ldcs) <
		    ldp->work.num_ldcs) {
			rprintf(resp, gettext("Insufficient free LDC channels "
			    "to bind virtual data plane client %s\n"),
			    vdpcc->name);
			return (ERR_LDC_NO_IDS);
		}
	}

	return (0);
}

int
vdpcc_bind_check(ldom_t *ldp, response_t *resp)
{
	vdpcc_t *vdpcc;
	int rc;

	CHAIN_ITER(ldp->vdpccs, vdpcc, ldom) {
		if ((rc = check_vdpcc(vdpcc, resp)) != 0)
			return (rc);
	}

	return (0);
}
#endif

void
vdpcc_delete_all(ldom_t *ldp)
{
	vdpcc_t *vdpcc, *vdpcc_next;

	for (vdpcc = CHAIN_START(ldp->vdpccs); vdpcc != NULL;
	    vdpcc = vdpcc_next) {
		vdpcc_next = CHAIN_NEXT(vdpcc->ldom);
		destroy_vdpcc(vdpcc);
	}
}

static int
bind_vdpcc(vdpcc_t *vdpcc, response_t *resp)
{
	vdpcs_t *vdpcs;
	vdpcs_port_t *vdpcs_port;
	int rc;

	vdpcs = vdpcs_find_by_name(vdpcc->service);
	if (vdpcs == NULL || vdpcs->binding == NULL) {
		rprintf(resp, gettext("The virtual data plane service \"%s\" "
		    "for virtual data plane client \"%s\",\ndoes not exist or "
		    "is not bound.\n"), vdpcc->service, vdpcc->name);
		return (ERR_NO_SERVICE);
	}

	if (vdpcs->ldp->flag.locked) {
		rprintf(resp, gettext("Cannot add virtual data plane client %s "
		    "because that would modify locked LDom %s\n"),
		    vdpcc->name, vdpcs->ldp->namep);
		return (ERR_LDOM_LOCKED);
	}

	init_vdpcc_binding(vdpcc, NEW_INSTANCE);
	rc = ldcs_chan_alloc(&vdpcc->binding->ldcs,
	    &vdpcc->ldp->guest_free_ldcs);
	if (rc != 0) {
		destroy_vdpcc_binding(vdpcc);
		rprintf(resp, gettext("Insufficient free LDC channels on "
		    "LDom %s to bind vdpcc %s\n"), vdpcc->ldp->namep,
		    vdpcc->name);
		return (rc);
	}

	vdpcs_port = vdpcs_add_port(vdpcs, vdpcc, NEW_INSTANCE);
	vdpcc->binding->vdpcs_port = vdpcs_port;

	rc = ldcs_chan_alloc(&vdpcs_port->ldcs, &vdpcs->ldp->guest_free_ldcs);
	if (rc != 0) {
		destroy_vdpcc_binding(vdpcc);
		rprintf(resp, gettext("Insufficient free LDC channels on "
		    "LDom %s to bind vdpcc %s\n"), vdpcs->ldp->namep,
		    vdpcc->name);
		return (rc);
	}

	if (vdpcs->ldp != vdpcc->ldp) {
		vdpcs->ldp->bound.num_clients++;
		vdpcs->ldp->flag.gen_new_md = true;
		vdpcs->ldp->flag.md_notify = true;
	}

	return (0);
}

int
vdpcc_bind_all(ldom_t *ldp, response_t *resp)
{
	vdpcc_t *vdpcc;
	int rc;

	CHAIN_ITER(ldp->vdpccs, vdpcc, ldom) {
		if ((rc = bind_vdpcc(vdpcc, resp)) != 0)
			return (rc);
	}

	return (0);
}

int
vdpcc_unbind_all(ldom_t *ldp, response_t *resp)
{
	vdpcc_t *vdpcc;

	CHAIN_ITER(ldp->vdpccs, vdpcc, ldom) {
		unbind_vdpcc(vdpcc);
	}

	return (0);
}

/*
 * Called from the add-vdpcc CLI.
 */
int
vdpcc_add(ldom_t *ldp, char *vdpcc_name, char *service, response_t *resp)
{
	vdpcc_t *vdpcc;
	int rc;

	if (vdpcc_find_by_name(ldp, vdpcc_name) != NULL) {
		rprintf(resp, gettext("%s client named %s already exists "
		    "in guest %s\n"), vdpcc_attrs.cli_name,
		    vdpcc_name, ldp->namep);
		return (ERR_CLIENT_EXISTS);
	}

	vdpcc = vdpcc_new(ldp, vdpcc_name, service);

	if (ldp->state >= BOUND) {
#if BIND_CHECK
		if ((rc = add_vdpcc_check(vdpcc, resp)) != 0) {
			destroy_vdpcc(vdpcc);
			return (rc);
		}
#endif
		if ((rc = bind_vdpcc(vdpcc, resp)) != 0) {
			destroy_vdpcc(vdpcc);
			return (rc);
		}

		if (zeus.mode == ZEUS_MGMT_MODE)
			seq_config_now_or_later(ldp, DEL_RECONFIG_ADD, resp);
	}

	database_ldom_add_vdpc_client(ldp, vdpcc_name, service);

	return (0);
}

/*
 * Called from the remove-vdpcc CLI.
 */
int
vdpcc_remove(ldom_t *ldp, char *vdpcc_name, bool_t force, response_t *resp)
{
	vdpcc_t *vdpcc;

	if ((vdpcc = vdpcc_find_by_name(ldp, vdpcc_name)) == NULL) {
		if (resp)
			rprintf(resp, gettext("%s named %s not found\n"),
			    vdpcc_attrs.cli_name, vdpcc_name);
		return (ERR_NO_CLIENT);
	}

	/*
	 * Delete the client from DB irrespective of the runtime state.
	 */
	database_ldom_del_vdpc_client(ldp, vdpcc_name);

	unbind_vdpcc(vdpcc);
	destroy_vdpcc(vdpcc);

	if (ldp->state >= BOUND) {
		if ((ldp->state > BOUND) && (zeus.mode == ZEUS_MGMT_MODE)) {
			/* handle as a delayed reconfig operation */
			initiate_or_update_delayed_reconfig(ldp,
			    DEL_RECONFIG_REMOVE_2PHASE, resp);
		} else if (zeus.mode == ZEUS_MGMT_MODE) {
			/* do reconfig now */
			seq_rm_reconfigure(ldp);
			free_all_staged_ldcs(ldp);
		}

		if (zeus.mode == ZEUS_CONFIG_MODE)
			free_all_staged_ldcs(ldp);
	}

	return (0);
}

void
vdpcc_gen_ldc_endpoints(ldom_t *ldp, md_t *hv_md, md_node_t *hv_guest,
    md_node_t *ldc_endpoints)
{
	vdpcc_t *vdpcc;

	CHAIN_ITER(ldp->vdpccs, vdpcc, ldom) {
		vdpcs_port_t *vdpcs_port = vdpcc->binding->vdpcs_port;
		vio_gen_ldcs_endpoint_nodes(hv_md, hv_guest, ldc_endpoints,
		    TYPE_GUEST, TYPE_GUEST, &vdpcc->binding->ldcs,
		    vdpcs_port->vdpcs->ldp->idp->id, &vdpcs_port->ldcs, NULL);
	}
}

static void
gen_vdpcc_guest_md_one(vdpcc_t *vdpcc, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	md_node_t *vdev, *vport;

	vdev = md_link_new_node(md, VDEV_NAME, cdevs, "fwd", "back");

	/*
	 * Compatible property is an array of strings, though we only
	 * currently have one.
	 */
	md_add_string_as_data_property(md, vdev, "compatible",
	    vdpcc_attrs.compatible);

	md_add_string_property(md, vdev, "device-type",
	    vdpcc_attrs.device_type);
	md_add_string_property(md, vdev, "name", vdpcc_attrs.md_name);
	md_add_value_property(md, vdev, "cfg-handle", vdpcc_devid(vdpcc));

	/* create vport node for connection to the server */
	vport = vio_gen_vdev_port_node(md, vdev, chan_endpoints,
	    vdpcc_attrs.port_name, 0, &vdpcc->binding->ldcs);
	md_add_string_property(md, vport, "vdpc-svc-name", vdpcc->name);
}

void
vdpcc_gen_guest_md(ldom_t *ldp, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vdpcc_t *vdpcc;

	CHAIN_ITER(ldp->vdpccs, vdpcc, ldom) {
		gen_vdpcc_guest_md_one(vdpcc, md, cdevs, chan_endpoints);
	}
}

static void
vdpcc_list_one(vdpcc_t *vdpcc, response_t *resp, dispctl_t *dctl)
{
	vdpcc_binding_t *binding = vdpcc->binding;

	/*
	 * TRANSLATION_NOTE
	 * The following are column headers in tables with fixed column widths.
	 * WIDTH  HEADER	MEANING
	 * 16	  "NAME"	name of client
	 * 27	  "SERVICE"	name of the service the client is using
	 * 3	  "LDC"		Logical Domain Channel number
	 */
	if (!dctl->parseable && dctl->header) {
		dctl->header = 0;
		rprintf(resp, "%s\n", dctl_label(vdpcc_attrs.name, dctl));
		if (dctl->bindings && dctl->extended)
			rprintf(resp, "    "FNAME" "FBNDNAME" %s\n",
			    gettext("NAME"), gettext("SERVICE"),
			    gettext("LDC"));
		else
			rprintf(resp, "    "FNAME" "FBNDNAME"\n",
			    gettext("NAME"), gettext("SERVICE"));
	}

	if (dctl->parseable)
		rprintf(resp, "%s|name=%s",
		    dctl_label(vdpcc_attrs.cli_name, dctl), vdpcc->name);
	else
		rprintf(resp, "    "FNAME,  vdpcc->name);

	char *service;

	if (binding != NULL) {
		vdpcs_t *vdpcs = binding->vdpcs_port->vdpcs;
		service = zsprintf("%s@%s", vdpcs->service, vdpcs->ldp->namep);
	} else {
		service = vdpcc->service;
	}

	if (dctl->parseable)
		rprintf(resp, "|service=%s", service);
	else
		rprintf(resp, " "FBNDNAME, service);

	if (binding != NULL)
		Xfree(service);

	if (dctl->bindings && (binding != NULL) && dctl->extended) {
		if (dctl->parseable)
			rprintf(resp, "|ldc=");
		else
			rprintf(resp, " ");
		ldcs_print(&vdpcc->binding->ldcs, resp);
	}

	rprintf(resp, "\n");
}

void
vdpcc_list(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	vdpcc_t *vdpcc;

	if (dctl->oformat && !dctl->oserial)
		return;

	if (CHAIN_LENGTH(ldp->vdpccs) == 0)
		return;

	dctl->header = 1;

	CHAIN_ITER(ldp->vdpccs, vdpcc, ldom) {
		vdpcc_list_one(vdpcc, resp, dctl);
	}

	if (!dctl->parseable)
		rprintf(resp, "\n");
}

static int
link_vdpcc_client(ldom_t *ldp, vdpcc_t *vdpcc)
{
	ldc_epdb_t *ep;
	ldc_chan_t *ldc;
	vdpcs_port_t *vdpcs_port;

	if (vdpcc->binding == NULL)
		fatal("Vdpc client %s has no binding to a vdpc service",
		    vdpcc->name);

	/* bind to service */
	ldc = CHAIN_START(vdpcc->binding->ldcs.ldcs);
	if ((ep = ldc_epdb_find(ldp->idp->id, ldc->chan->id)) != NULL)
		vdpcs_port = vdpcs_port_find_by_ldcid(ep->peer->ldp,
		    ep->peer->ldc);
	if ((ep == NULL) || (vdpcs_port == NULL))
		fatal("Could not find matching vdpcs port for vdpcc %s on "
		    "LDom %s", vdpcc->name, ldp->namep);

	vdpcs_port->vdpcc = vdpcc;
	vdpcc->binding->vdpcs_port = vdpcs_port;

	if (vdpcs_port->vdpcs->ldp != ldp)
		vdpcs_port->vdpcs->ldp->bound.num_clients++;

	ASSERT(vdpcc->service == NULL);
	vdpcc->service = Xstrdup(vdpcs_port->vdpcs->service);

	database_ldom_add_vdpc_client(ldp, vdpcc->name, vdpcc->service);

	return (0);
}

int
link_vdpcc_clients(ldom_t *ldp)
{
	vdpcc_t *vdpcc;

	CHAIN_ITER(ldp->vdpccs, vdpcc, ldom) {
		link_vdpcc_client(ldp, vdpcc);
	}

	return (0);
}
