/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)vdpcs.c	1.8	09/04/22 SMI"

#include <strings.h>

#include "ldmd_types.h"
#include "allocate.h"
#include "chain.h"
#include "database.h"
#include "errors.h"
#include "fatal.h"
#include "ldom.h"
#include "resource_id.h"
#include "vio.h"

/* list of all vdpcs services */
static CHAIN(vdpcs_t, vdpcs_services) = DEFAULTCHAIN;

struct vdpcs_attributes vdpcs_attrs = {
	.name =		"vdpcs",
	.compatible =	"SUNW,sun4v-data-plane-channel",
	.md_name =	"virtual-data-plane-channel",
	.cli_name =	"vdpcs",
	.device_type =	"serial",
	.port_name =	"vdpc-port"
};

int
vdpcs_devid(vdpcs_t *vdpcs)
{
	if (vdpcs->binding == NULL)
		return (INVALID_DEVID);
	else
		return (vdpcs->binding->cfg_handle->id);
}

static res_pool_t *
vdpcs_cfg_handles(ldom_t *ldp)
{
#if VDEV_CFG_HANDLE
	return (&ldp->free_cfg_handles);
#else
	return (&ldp->bound.vdpcs_cfg_handles);
#endif
}

vdpcs_t *
vdpcs_find_by_name(char *service_name)
{
	vdpcs_t *vdpcs;

	CHAIN_ITER(vdpcs_services, vdpcs, link) {
		if (strcmp(service_name, vdpcs->service) == 0)
			break;
	}

	return (vdpcs);
}

vdpcs_t *
vdpcs_new(ldom_t *ldp, char *name)
{
	vdpcs_t *vdpcs;

	if ((vdpcs = vdpcs_find_by_name(name)) != NULL)
		return (NULL);

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(ldp->vdpcss) == 0) {
		create_res_pool(&ldp->bound.vdpcs_cfg_handles,
		    IDENT_MIN, IDENT_MAX, Res_sorted);
		name_ldom_res_pool(&ldp->bound.vdpcs_cfg_handles, ldp,
		    "vdpcs cfg handles");
	}
#endif
	vdpcs = Xmalloc(sizeof (vdpcs_t));

	vdpcs->ldp = ldp;
	vdpcs->service = Xstrdup(name);
	vdpcs->binding = NULL;

	CHAIN_ADD(ldp->vdpcss, vdpcs, ldom);
	CHAIN_ADD(vdpcs_services, vdpcs, link);

	return (vdpcs);
}

static void
bind_vdpcs(vdpcs_t *vdpcs, int cfg_handle)
{
	ldom_t *ldp = vdpcs->ldp;
	vdpcs_binding_t *binding;

	binding = Xmalloc(sizeof (vdpcs_binding_t));

	binding->cfg_handle = vio_get_cfg_handle(cfg_handle,
	    vdpcs_cfg_handles(ldp));
	if (binding->cfg_handle == NULL)
		fatal("Could not obtain vdpcs cfg_handle %d on LDom %s\n",
		    cfg_handle, ldp->namep);

	create_res_pool(&binding->free_ports, 0, IDENT_MAX, Res_sorted);
	name_ldom_vdev_res_pool(&binding->free_ports, ldp, "vdpcs",
	    vdpcs->service, "ports");
	CHAIN_INIT(binding->vports);

	ldp->bound.num_services++;
	vdpcs->binding = binding;
}

static void
unbind_vdpcs(vdpcs_t *vdpcs)
{
	if (vdpcs->binding == NULL)
		return;

	ASSERT(CHAIN_LENGTH(vdpcs->binding->vports) == 0);

	vdpcs->ldp->bound.num_services--;
	destroy_res_pool(&vdpcs->binding->free_ports);
	free_res(vdpcs->binding->cfg_handle);
	Xfree(vdpcs->binding);
	vdpcs->binding = NULL;
}

/*ARGSUSED1*/
int
vdpcs_bind_all(ldom_t *ldp, response_t *resp)
{
	vdpcs_t *vdpcs;

	CHAIN_ITER(ldp->vdpcss, vdpcs, ldom) {
		bind_vdpcs(vdpcs, NEW_INSTANCE);
	}

	return (0);
}

/*ARGSUSED1*/
int
vdpcs_unbind_all(ldom_t *ldp, response_t *resp)
{
	vdpcs_t *vdpcs;

	CHAIN_ITER(ldp->vdpcss, vdpcs, ldom) {
		unbind_vdpcs(vdpcs);
	}

	return (0);
}

/*
 * Called from vdpcc code to create a new port on the vdpcs when binding a
 * new vdpcc to it. Also called when initializing vio bindings from MDs.
 */
vdpcs_port_t *
vdpcs_add_port(vdpcs_t *vdpcs, vdpcc_t *vdpcc, uint64_t port)
{
	vdpcs_port_t *vport;

	vport = Xmalloc(sizeof (vdpcs_port_t));

	CHAIN_INIT(vport->ldcs.ldcs);
	if (port == NEW_INSTANCE)
		vport->port = alloc_res(&vdpcs->binding->free_ports);
	else
		vport->port = pull_res_id(&vdpcs->binding->free_ports, port);

	if (vport->port == NULL)
		fatal("Unable to allocate port %"PRIu64" on vdpcs %s",
		    port, vdpcs->service);

	vport->vdpcc = vdpcc;
	vport->vdpcs = vdpcs;

	CHAIN_ADD(vdpcs->binding->vports, vport, link);

	return (vport);
}

void
vdpcs_add_from_mdnode(ldom_t *ldp, md_t *guest_md, md_node_t *vdev)
{
	char *service_name;
	uint64_t cfg_handle;
	int i, num_client_bindings;
	md_node_t **nodelistp;
	vdpcs_t *vdpcs;

	if (md_get_prop_val(guest_md, vdev, "cfg-handle", &cfg_handle) != 0)
		fatal("vdevice has no cfg-handle property");

	/*
	 * Get the service name from the svc-name property if it exists,
	 * otherwise get it from the database.
	 */
	if (md_get_prop_str(guest_md, vdev, DEVALIAS, &service_name) == 0) {
		service_name = Xstrdup(service_name);
	} else if (md_get_prop_str(guest_md, vdev, SERVICE_NAME,
	    &service_name) == 0) {
		service_name = Xstrdup(service_name);
	} else {
		service_name = get_service_name_from_db(ldp, vdpcs_attrs.name,
		    cfg_handle);
		if (service_name == NULL)
			/* use canonical name */
			service_name = vio_canonical_service_name(ldp,
			    vdpcs_attrs.name, cfg_handle);
	}

	vdpcs = vdpcs_new(ldp, service_name);
	bind_vdpcs(vdpcs, cfg_handle);

	database_ldom_add_vdpc_service(ldp, service_name);

	/*
	 * Now, scan through the vport list and add a client binding for
	 * each one
	 */
	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(guest_md));

	num_client_bindings = md_scan_dag(guest_md, vdev,
	    md_find_name(guest_md, "virtual-device-port"),
	    md_find_name(guest_md, "fwd"), nodelistp);

	for (i = 0; i < num_client_bindings; ++i) {
		uint64_t port;
		char *name;

		md_node_t *vport = nodelistp[i];

		if (md_get_prop_str(guest_md, vport, "name", &name) != 0)
			fatal("vport has no name property");
		if (md_get_prop_val(guest_md, vport, "id", &port) != 0)
			fatal("vport has no id property");

		vdpcs_port_t *vdpcs_port;

		vdpcs_port = vdpcs_add_port(vdpcs, NULL, port);

		ldc_get_and_pull_ids(ldp, &ldp->guest_free_ldcs,
		    guest_md, vport, &vdpcs_port->ldcs);
	}

	Xfree(service_name);
	Xfree(nodelistp);
}

void
vdpcs_destroy_port(vdpcs_port_t *vdpcs_port)
{
	CHAIN_UNLINK(vdpcs_port->vdpcs->binding->vports, vdpcs_port, link);

	if (vdpcs_port->port != NULL)
		free_res(vdpcs_port->port);

	ldcs_free(&vdpcs_port->ldcs);
	Xfree(vdpcs_port);
}

static void
destroy_vdpcs(vdpcs_t *vdpcs)
{
	ldom_t *ldp = vdpcs->ldp;

	ASSERT(vdpcs->binding == NULL);

	Xfree(vdpcs->service);
	CHAIN_UNLINK(ldp->vdpcss, vdpcs, ldom);
	CHAIN_UNLINK(vdpcs_services, vdpcs, link);
	Xfree(vdpcs);

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(ldp->vdpcss) == 0)
		destroy_res_pool(&ldp->bound.vdpcs_cfg_handles);
#endif
}

void
vdpcs_delete_all(ldom_t *ldp)
{
	vdpcs_t *vdpcs, *vdpcs_next;

	for (vdpcs = CHAIN_START(ldp->vdpcss); vdpcs != NULL;
	    vdpcs = vdpcs_next) {
		vdpcs_next = CHAIN_NEXT(vdpcs->ldom);
		destroy_vdpcs(vdpcs);
	}
}

void
vdpcs_unbind_port(vdpcs_port_t *vdpcs_port)
{
	ldcs_stage(&vdpcs_port->ldcs);
	vdpcs_destroy_port(vdpcs_port);
}

/*
 * Called from the add-vdpcs CLI.
 */
int
vdpcs_add(ldom_t *ldp, char *service_name, response_t *resp)
{
	vdpcs_t *vdpcs;

	if (vdpcs_find_by_name(service_name) != NULL) {
		rprintf(resp, "%s service named %s already exists\n",
		    vdpcs_attrs.name, service_name);
		return (ERR_SERVICE_EXISTS);
	}

	vdpcs = vdpcs_new(ldp, service_name);
	ASSERT(vdpcs != NULL);

	if (ldp->state >= BOUND) {
		bind_vdpcs(vdpcs, NEW_INSTANCE);

		if (zeus.mode == ZEUS_MGMT_MODE)
			seq_config_now_or_later(ldp, DEL_RECONFIG_ADD, resp);
	}

	database_ldom_add_vdpc_service(ldp, service_name);

	return (0);
}

/*
 * Called from the remove-vdpcs CLI.
 */
/*ARGSUSED1*/
int
vdpcs_remove(vdpcs_t *vdpcs, bool_t force, response_t *resp)
{
	ldom_t *ldp = vdpcs->ldp;

	if ((vdpcs->binding != NULL) &&
	    (CHAIN_LENGTH(vdpcs->binding->vports) > 0)) {
		/* clients are bound to the vdpcs service */
		rprintf(resp, MSG_ERR_SERVICE_BOUND_CLIENTS);
		return (ERR_SERVICE_BOUND_CLIENTS);
	}

	database_ldom_del_vdpc_service(ldp, vdpcs->service);
	unbind_vdpcs(vdpcs);
	destroy_vdpcs(vdpcs);

	if ((ldp->state >= BOUND) && (zeus.mode == ZEUS_MGMT_MODE)) {
		if (ldp->state != BOUND) {
			/* handle as a delayed reconfig operation */
			initiate_or_update_delayed_reconfig(ldp,
			    DEL_RECONFIG_REMOVE_2PHASE, resp);
		} else {
			/* do reconfig now */
			seq_download_mds_do_config(ldp);
			free_all_staged_ldcs(ldp);
		}
	}

	if (zeus.mode == ZEUS_CONFIG_MODE)
		free_all_staged_ldcs(ldp);

	return (0);
}

void
vdpcs_gen_ldc_endpoints(ldom_t *ldp, md_t *hv_md, md_node_t *hv_guest,
    md_node_t *ldc_endpoints)
{
	vdpcs_t *vdpcs;
	vdpcs_port_t *vdpcs_port;

	CHAIN_ITER(ldp->vdpcss, vdpcs, ldom) {
		CHAIN_ITER(vdpcs->binding->vports, vdpcs_port, link) {
			vio_gen_ldcs_endpoint_nodes(hv_md, hv_guest,
			    ldc_endpoints, TYPE_GUEST, TYPE_GUEST,
			    &vdpcs_port->ldcs, vdpcs_port->vdpcc->ldp->idp->id,
			    &vdpcs_port->vdpcc->binding->ldcs, NULL);
		}
	}
}

static void
gen_vdpcs_guest_md_one(vdpcs_t *vdpcs, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vdpcs_port_t *vdpcs_port;
	md_node_t *vdev;

	vdev = md_link_new_node(md, VDEV_NAME, cdevs, "fwd", "back");

	/*
	 * Compatible property is an array of strings, though we only
	 * currently have one.
	 */
	md_add_string_as_data_property(md, vdev, "compatible",
	    vdpcs_attrs.compatible);
	md_add_string_property(md, vdev, "device-type",
	    vdpcs_attrs.device_type);
	md_add_string_property(md, vdev, "name", vdpcs_attrs.md_name);
	md_add_value_property(md, vdev, "cfg-handle", vdpcs_devid(vdpcs));
	md_add_string_property(md, vdev, DEVALIAS, vdpcs->service);

	CHAIN_ITER(vdpcs->binding->vports, vdpcs_port, link) {
		md_node_t *port_node;

		port_node = vio_gen_vdev_port_node(md, vdev, chan_endpoints,
		    vdpcs_attrs.port_name, vdpcs_port->port->id,
		    &vdpcs_port->ldcs);
		md_add_string_property(md, port_node, "vdpcs-svc-name",
		    vdpcs->service);
	}
}

void
vdpcs_gen_guest_md(ldom_t *ldp, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vdpcs_t *vdpcs;

	CHAIN_ITER(ldp->vdpcss, vdpcs, ldom) {
		gen_vdpcs_guest_md_one(vdpcs, md, cdevs, chan_endpoints);
	}
}

static void
print_vdpcs_header(vdpcs_t *vdpcs, char *servertype, int nclients,
    response_t *resp, dispctl_t *dctl)
{
	if (dctl->header && !dctl->parseable) {
		dctl->header = 0;
		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in tables with fixed
		 * column widths.
		 * WIDTH  HEADER	MEANING
		 * 16	  "NAME"	VIO server name
		 * 16	  "LDOM"	the LDom name iff list-services
		 */
		rprintf(resp, "    "FNAME, gettext("NAME"));
		if (dctl->services)
			rprintf(resp, " "FNAME, gettext("LDOM"));
		rprintf(resp, "\n");
	}

	print_server_header(vdpcs->service, servertype, nclients, resp, dctl);
	if (dctl->services) {
		if (dctl->parseable)
			rprintf(resp, "|ldom=%s", vdpcs->ldp->namep);
		else
			rprintf(resp, " "FNAME, vdpcs->ldp->namep);
	}
}

void
vdpcs_port_list(vdpcs_port_t *vdpcs_port, response_t *resp, dispctl_t *dctl)
{
	vdpcc_t *vdpcc;
	char *clientname;

	vdpcc = vdpcs_port->vdpcc;

	clientname = zsprintf("%s@%s", vdpcc->name, vdpcc->ldp->namep);

	if (dctl->parseable) {
		rprintf(resp, "|client=%s", clientname);
	} else {
		rprintf(resp, "        "FBNDNAME, clientname);
	}
	Xfree(clientname);

	if (dctl->extended) {
		if (dctl->parseable)
			rprintf(resp, "|ldc=");
		else
			rprintf(resp, " ");
		ldcs_print(&vdpcs_port->ldcs, resp);
	}

	rprintf(resp, "\n");
}

static void
vdpcs_list_one(vdpcs_t *vdpcs, response_t *resp, dispctl_t *dctl)
{
	int nclients;

	if (vdpcs->binding != NULL)
		nclients = CHAIN_LENGTH(vdpcs->binding->vports);
	else
		nclients = 0;

	print_vdpcs_header(vdpcs, vdpcs_attrs.name, nclients, resp, dctl);

	rprintf(resp, "\n");

	if (dctl->bindings && nclients > 0) {
		vdpcs_port_t *vdpcs_port;

		dctl->header = 1;

		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in tables with fixed
		 * column widths.
		 * WIDTH  HEADER	MEANING
		 * 27	  "CLIENT"	name of client of service
		 * 3	  "LDC"		Logical Domain Channel number
		 */
		rprintf(resp, "        "FBNDNAME, gettext("CLIENT"));
		if (dctl->extended)
			rprintf(resp, " %s", gettext("LDC"));
		rprintf(resp, "\n");

		CHAIN_ITER(vdpcs->binding->vports, vdpcs_port, link) {
			vdpcs_port_list(vdpcs_port, resp, dctl);
		}
	}
}

void
vdpcs_list(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	vdpcs_t *vdpcs;

	if (dctl->oformat && !dctl->oserial)
		return;

	if (CHAIN_LENGTH(ldp->vdpcss) == 0)
		return;

	dctl->header = 1;

	if (!dctl->parseable)
		rprintf(resp, "%s\n", dctl_label(vdpcs_attrs.name, dctl));

	CHAIN_ITER(ldp->vdpcss, vdpcs, ldom) {
		vdpcs_list_one(vdpcs, resp, dctl);
	}

	if (!dctl->parseable)
		rprintf(resp, "\n");
}

vdpcs_port_t *
vdpcs_port_find_by_ldcid(ldom_t *ldp, uint64_t ldcid)
{
	vdpcs_t *vdpcs;
	vdpcs_port_t *vport;

	CHAIN_ITER(ldp->vdpcss, vdpcs, ldom) {
		if (vdpcs->binding == NULL)
			continue;
		CHAIN_ITER(vdpcs->binding->vports, vport, link) {
			if (ldcs_search(&vport->ldcs, ldcid) != NULL)
				return (vport);
		}
	}

	return (NULL);
}
