/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <alloca.h>
#include <string.h>

#include "ldmd_types.h"
#include "allocate.h"
#include "chain.h"
#include "database.h"
#include "errors.h"
#include "fatal.h"
#include "ldom.h"
#include "resource_id.h"
#include "vio.h"

struct vds_attributes vds_attrs = {
	.name =		"vds",
	.compatible =	"SUNW,sun4v-disk-server",
	.md_name =	"virtual-disk-server",
	.cli_name =	"vds",
	.device_type =	"vds",
	.port_name =	"vds-port"
};

/*
 * The following two arrays must exactly be in the same order
 */
char *vdsdev_opts[] = {
	VDSDEV_OPTS_RONLY,
	VDSDEV_OPTS_SLICE,
	VDSDEV_OPTS_EXCL,
	NULL
};

int vdsdev_optval[] = {
	VDEV_RONLY,
	VDEV_SLICE,
	VDEV_EXCL,
	0
};

static int check_modify(vds_volume_t *vol, char *options, bool_t force,
    char *mpgname, response_t *resp, bool_t *modify_mpg, bool_t *modify_opt);

ident_val_t
vds_devid(vds_t *vds)
{
	if (vds->binding == NULL)
		return (INVALID_DEVID);
	else
		return (vds->binding->cfg_handle->id);
}

static res_pool_t *
vds_cfg_handles(ldom_t *ldp)
{
#if VDEV_CFG_HANDLE
	return (&ldp->free_cfg_handles);
#else
	return (&ldp->bound.vds_cfg_handles);
#endif
}

vds_t *
vds_find_by_name(char *service_name)
{
	vds_t *vds;

	CHAIN_ITER(zeus.vds_services, vds, link) {
		if (strcmp(service_name, vds->service) == 0)
			break;
	}

	return (vds);
}

vds_t *
vds_new(ldom_t *ldp, char *name)
{
	vds_t *vds;

	if ((vds = vds_find_by_name(name)) != NULL)
		return (NULL);

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(ldp->vdss) == 0) {
		create_res_pool(&ldp->bound.vds_cfg_handles,
		    IDENT_MIN, IDENT_MAX, Res_sorted);
		name_ldom_res_pool(&ldp->bound.vds_cfg_handles, ldp,
		    "vds cfg handles");
	}
#endif
	vds = Xcalloc(1, vds_t);

	vds->ldp = ldp;
	vds->service = Xstrdup(name);
	vds->binding = NULL;

	CHAIN_INIT(vds->volumes);

	CHAIN_ADD(ldp->vdss, vds, ldom);
	CHAIN_ADD(zeus.vds_services, vds, link);

	return (vds);
}

static void
bind_vds(vds_t *vds, int cfg_handle)
{
	ldom_t *ldp = vds->ldp;
	vds_binding_t *binding;

	binding = Xcalloc(1, vds_binding_t);

	binding->cfg_handle = vio_get_cfg_handle(cfg_handle,
	    vds_cfg_handles(ldp));
	if (binding->cfg_handle == NULL)
		fatal("Could not obtain vds cfg_handle %d on LDom %s\n",
		    cfg_handle, ldp->namep);

	create_res_pool(&binding->free_ports, 0, IDENT_MAX, Res_sorted);
	name_ldom_vdev_res_pool(&binding->free_ports, ldp, "vds",
	    vds->service, "ports");
	CHAIN_INIT(binding->vports);

	ldp->bound.num_services++;
	vds->binding = binding;
}

static void
unbind_vds(vds_t *vds)
{
	if (vds->binding == NULL)
		return;

	ASSERT(CHAIN_LENGTH(vds->binding->vports) == 0);

	vds->ldp->bound.num_services--;
	destroy_res_pool(&vds->binding->free_ports);
	free_res(vds->binding->cfg_handle);
	Xfree(vds->binding);
	vds->binding = NULL;
}

/*ARGSUSED1*/
int
vds_bind_all(ldom_t *ldp, response_t *resp)
{
	vds_t *vds;

	CHAIN_ITER(ldp->vdss, vds, ldom) {
		bind_vds(vds, NEW_INSTANCE);
	}

	return (0);
}

/*ARGSUSED1*/
int
vds_unbind_all(ldom_t *ldp, response_t *resp)
{
	vds_t *vds;

	CHAIN_ITER(ldp->vdss, vds, ldom) {
		unbind_vds(vds);
	}

	return (0);
}

static mpgroup_t *
vds_find_mpgroup_by_name(char *mpgname)
{
	mpgroup_t *mpgroup = NULL;

	CHAIN_ITER(zeus.mpgroups, mpgroup, link) {
		if (strcmp(mpgroup->name, mpgname) == 0)
			break;
	}

	return (mpgroup);
}

static mpgroup_t *
vds_new_mpgroup(char *mpgname)
{
	mpgroup_t *mpgroup = Xcalloc(1, mpgroup_t);

	mpgroup->name = Xstrdup(mpgname);
	CHAIN_INIT(mpgroup->volumes);
	CHAIN_LINK_INIT(mpgroup->link);

	CHAIN_ADD(zeus.mpgroups, mpgroup, link);

	return (mpgroup);
}

static void
vds_remove_mpgroup(mpgroup_t *mpgroup)
{
	ASSERT(mpgroup != NULL);
	ASSERT(CHAIN_LENGTH(mpgroup->volumes) == 0);
	CHAIN_UNLINK(zeus.mpgroups, mpgroup, link);

	Xfree(mpgroup->name);
	Xfree(mpgroup);
}

/*
 * A new mpgroup is created only if no other mpgroup with the
 * same name exists
 */
void
vds_mpgroup_add_vol(char *mpgname, vds_volume_t *vol)
{
	mpgroup_t *mpgroup = vds_find_mpgroup_by_name(mpgname);
	if (mpgroup == NULL)
		mpgroup = vds_new_mpgroup(mpgname);
	vol->mpgroup = mpgroup;
	CHAIN_ADD(mpgroup->volumes, vol, mpgroup_link);
}

/*
 * An mpgroup is removed when the last volume ceases to be a member
 * of the mpgroup. There is no other way to remove an mpgroup in
 * that we don't have to walk the mpgroup chain removing volumes.
 */
static bool_t
vds_mpgroup_rem_vol(vds_volume_t *vol)
{
	if (vol->mpgroup != NULL) {
		mpgroup_t *mpgroup = vol->mpgroup;
		/*
		 * Disallow if there is a disk bound to this volume
		 */
		if (vol->vds_port != NULL)
			return (false);

		CHAIN_UNLINK(mpgroup->volumes, vol, mpgroup_link);
		vol->mpgroup = NULL;

		/*
		 * If this was the last volume to be removed, free the mpgroup
		 */
		if (CHAIN_LENGTH(mpgroup->volumes) == 0)
			vds_remove_mpgroup(mpgroup);
	}

	return (true);
}

/*
 * Make sure a vdsdev being added has its option specified identically
 * to the existing vdsdev in the same mpgroup
 */
static bool_t
add_mpg_check(char *mpgname, char *options, vds_volume_t *vol, response_t *resp)
{
	vds_volume_t *avol;
	mpgroup_t *mpgroup = NULL;

	if (mpgname != NULL && mpgname[0] != '\0')
		mpgroup = vds_find_mpgroup_by_name(mpgname);

	if (mpgroup == NULL)
		return (true);

	/*
	 * If there are volumes already in the mpgroup, make sure the
	 * options match
	 */
	CHAIN_ITER(mpgroup->volumes, avol, mpgroup_link) {
		/*
		 * Ensure the new volume has the same option as the existing
		 * ones
		 */
		if (avol == vol)
			continue;
		if (vds_decode_opts(options) !=
		    vds_decode_opts(avol->options)) {
			if (resp)
				rprintf(resp, gettext("The option must be "
				    "the same on all vdsdev in the same "
				    "mpgroup\n"));
			return (false);
		}
	}

	return (true);
}

static vds_volume_t *
add_vds_volume(vds_t *vds, char *volname, char *blkdev, char *options,
	char *mpgname, response_t *resp)
{
	vds_volume_t *vol;

	if (add_mpg_check(mpgname, options, NULL, resp) == false)
		return (NULL);

	vol = Xcalloc(1, vds_volume_t);

	vol->volname = (volname != NULL) ? Xstrdup(volname) : NULL;
	vol->blkdev = (blkdev != NULL) ? Xstrdup(blkdev) : NULL;
	vol->options = (options != NULL) ? Xstrdup(options) : NULL;
	vol->vds_port = NULL;
	vol->vds = vds;

	CHAIN_LINK_INIT(vol->link);
	CHAIN_LINK_INIT(vol->mpgroup_link);
	CHAIN_ADD(vds->volumes, vol, link);

	/*
	 * If an mpgroup is specified, add this volume to the mpgroup
	 */
	if (mpgname != NULL)
		vds_mpgroup_add_vol(mpgname, vol);

	return (vol);
}

/*
 * Called from vdisk code to create a new port on the vds when binding a
 * new vdisk to it. Also called when initializing vio bindings from MDs.
 */
vds_port_t *
vds_add_port(vds_t *vds, vds_volume_t *vol, uint64_t port)
{
	vds_port_t *vport;

	vport = Xcalloc(1, vds_port_t);

	CHAIN_INIT(vport->ldcs.ldcs);
	if (port == NEW_INSTANCE)
		vport->port = alloc_res(&vds->binding->free_ports);
	else
		vport->port = pull_res_id(&vds->binding->free_ports, port);

	if (vport->port == NULL)
		fatal("Unable to allocate port %"PRIu64" on vds %s",
		    port, vds->service);

	vport->vds = vds;
	vport->vol = vol;

	vol->vds_port = vport;

	CHAIN_ADD(vds->binding->vports, vport, link);

	return (vport);
}

vds_volume_t *
vds_find_volume_by_name(vds_t *vds, char *volname)
{
	vds_volume_t *vol;

	for (vol = CHAIN_START(vds->volumes); vol != NULL;
	    vol = CHAIN_NEXT(vol->link)) {
		if (strcmp(vol->volname, volname) == 0)
			break;
	}

	return (vol);
}

/*
 * Called from database code to add a volume in the database to a disk server.
 */
void
vds_add_volume(vds_t *vds, char *volname, char *blkdev, char *options,
	char *mpgname)
{
	if (vds_find_volume_by_name(vds, volname) == NULL)
		(void) add_vds_volume(vds, volname, blkdev, options, mpgname,
		    NULL);
}

void
vds_add_from_mdnode(ldom_t *ldp, md_t *guest_md, md_node_t *vdev)
{
	char *service_name;
	uint64_t cfg_handle;
	int i, num_client_bindings;
	md_node_t **nodes;
	vds_t *vds;

	if (md_get_prop_val(guest_md, vdev, "cfg-handle", &cfg_handle) != 0)
		fatal("vdevice has no cfg-handle property");

	/*
	 * Get the service name from the svc-name property if it exists,
	 * otherwise get it from the database.
	 */
	if (md_get_prop_str(guest_md, vdev, DEVALIAS, &service_name) == 0) {
		service_name = Xstrdup(service_name);
	} else if (md_get_prop_str(guest_md, vdev, SERVICE_NAME,
	    &service_name) == 0) {
		service_name = Xstrdup(service_name);
	} else {
		service_name = get_service_name_from_db(ldp, vds_attrs.name,
		    cfg_handle);
		if (service_name == NULL)
			/* use canonical name */
			service_name = vio_canonical_service_name(ldp,
			    vds_attrs.name, cfg_handle);
	}

	if ((vds = vds_new(ldp, service_name)) == NULL)
		fatal("Duplicate vds service \"%s\"", service_name);

	bind_vds(vds, cfg_handle);

	database_ldom_add_disk_service(ldp, service_name);

	/*
	 * Now, scan through the vport list and add a client binding for
	 * each one
	 */
	nodes = Xcalloc(md_node_count(guest_md), md_node_t *);

	num_client_bindings = md_scan_dag(guest_md, vdev,
	    md_find_name(guest_md, "virtual-device-port"),
	    md_find_name(guest_md, "fwd"), nodes);

	for (i = 0; i < num_client_bindings; ++i) {
		uint64_t port;
		char *name, *blkdev, *options_prop, *options;
		char *md_volname = NULL;
		char *md_mpgname = NULL;
		int len;
		vds_port_t *vds_port;
		vds_volume_t *vds_vol;

		md_node_t *vport = nodes[i];

		if (md_get_prop_str(guest_md, vport, "name", &name) != 0)
			fatal("vport has no name property");
		if (md_get_prop_val(guest_md, vport, "id", &port) != 0)
			fatal("vport has no id property");
		if (md_get_prop_str(guest_md, vport, VDS_BLOCK_DEV,
		    &blkdev) != 0)
			fatal("vport missing %s property", VDS_BLOCK_DEV);

		if (md_get_prop_data(guest_md, vport, VDS_BLOCK_DEV_OPTS,
		    (uint8_t **)&options_prop, &len) != 0) {
			options = NULL;
		} else {
			options = alloca(len);
			(void) memcpy(options, options_prop, len);

			int j = strlen(options);
			while (j + 1 < len) {
				options[j] = ',';	/* replace \0 with , */
				j = strlen(options);
			}
		}

		/*
		 * Add the volume with no name. The name of the vdsdev
		 * would be known when linked with the vdisk
		 */
		vds_vol = add_vds_volume(vds, NULL, blkdev, options,
		    NULL, NULL);

		vds_port = vds_add_port(vds, vds_vol, port);

		if (md_get_prop_str(guest_md, vport, VOLNAME_PROP,
		    &md_volname) == 0) {
			vds_port->work.md_volume = Xstrdup(md_volname);
		}

		if (md_get_prop_str(guest_md, vport, MPGROUP_PROP,
		    &md_mpgname) == 0) {
				vds_port->work.md_mpgroup = Xstrdup(md_mpgname);
		}

		ldc_get_and_pull_ids(ldp, &ldp->guest_free_ldcs,
		    guest_md, vport, &vds_port->ldcs);
	}

	Xfree(service_name);
	Xfree(nodes);
}

void
vds_gen_ldc_endpoints(ldom_t *ldp, md_t *hv_md, md_node_t *hv_guest,
    md_node_t *ldc_endpoints)
{
	vds_t *vds;
	vds_port_t *vds_port;

	CHAIN_ITER(ldp->vdss, vds, ldom) {
		CHAIN_ITER(vds->binding->vports, vds_port, link) {
			vio_gen_ldcs_endpoint_nodes(hv_md, hv_guest,
			    ldc_endpoints, TYPE_GUEST, TYPE_GUEST,
			    &vds_port->ldcs,
			    vds_port->vdisk_port->vdisk->ldp->idp->id,
			    &vds_port->vdisk_port->ldcs, NULL);
		}
	}
}

static void
gen_vds_guest_md_one(vds_t *vds, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vds_port_t *vds_port;
	md_node_t *vdev;

	vdev = md_link_new_node(md, VDEV_NAME, cdevs, "fwd", "back");

	/*
	 * Compatible property is an array of strings, though we only
	 * currently have one.
	 */
	md_add_string_as_data_property(md, vdev, "compatible",
	    vds_attrs.compatible);
	md_add_string_property(md, vdev, "device-type", vds_attrs.device_type);
	md_add_string_property(md, vdev, "name", vds_attrs.md_name);
	md_add_value_property(md, vdev, "cfg-handle", vds_devid(vds));
	md_add_string_property(md, vdev, DEVALIAS, vds->service);

	CHAIN_ITER(vds->binding->vports, vds_port, link) {
		md_node_t *port_node;

		port_node = vio_gen_vdev_port_node(md, vdev, chan_endpoints,
		    vds_attrs.port_name, vds_port->port->id, &vds_port->ldcs);
		md_add_string_property(md, port_node, VDS_BLOCK_DEV,
		    vds_port->vol->blkdev);
		md_add_string_property(md, port_node, VOLNAME_PROP,
		    vds_port->vol->volname);
		if (vds_port->vol->mpgroup != NULL)
			md_add_string_property(md, port_node, MPGROUP_PROP,
			    vds_port->vol->mpgroup->name);
		if (vds_port->vol->options != NULL) {
			int i, space;
			char *opt, *opt_copy;
			uint8_t *vdsdev_opts;

			space = strlen(vds_port->vol->options) + 1;
			vdsdev_opts = alloca(space);
			opt_copy = Xstrdup(vds_port->vol->options);
			opt = strtok(opt_copy, ",");
			i = 0;
			while (opt != NULL) {
				strcpy((char *)(vdsdev_opts + i), opt);
				i += strlen(opt) + 1;
				opt = strtok(NULL, ",");
			}

			md_add_data_property(md, port_node,
			    VDS_BLOCK_DEV_OPTS, i, vdsdev_opts);

			Xfree(opt_copy);
		}
	}
}

void
vds_gen_guest_md(ldom_t *ldp, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vds_t *vds;

	CHAIN_ITER(ldp->vdss, vds, ldom) {
		gen_vds_guest_md_one(vds, md, cdevs, chan_endpoints);
	}
}

static void
print_vds_header(char *server, char *servertype, int nclients, response_t *resp,
    dispctl_t *dctl)
{
	if (dctl->header && !dctl->parseable) {
		dctl->header = 0;
		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in tables with fixed
		 * column widths.
		 * WIDTH  HEADER	MEANING
		 * 16	  "NAME"	VIO server name
		 * 16	  "LDOM"	the LDom name iff list-services
		 * 14	  "VOLUME"	virtual disk volume name
		 *	  "OPTIONS"	volume options
		 *	  "MPGROUP"	mpgroup of this volume
		 *	  "DEVICE"	disk or network device name
		 */
		rprintf(resp, "    "FNAME, gettext("NAME"));
		if (dctl->services)
			rprintf(resp, " "FNAME, gettext("LDOM"));
		rprintf(resp, " "FVOL" "FVOLOPTS" "FVOLMGP" %s\n",
		    gettext("VOLUME"), gettext("OPTIONS"), gettext("MPGROUP"),
		    gettext("DEVICE"));
	}

	print_server_header(server, servertype, nclients, resp, dctl);
}

static void
vds_port_list(vds_port_t *vds_port, response_t *resp, dispctl_t *dctl)
{
	vdisk_t *vdisk;
	char *clientname;

	vdisk = vds_port->vdisk_port->vdisk;

	clientname = zsprintf("%s@%s", vdisk->name, vdisk->ldp->namep);

	if (dctl->parseable) {
		rprintf(resp, "|client=%s|vol=%s", clientname,
		    vds_port->vol->volname);
	} else {
		rprintf(resp, "        "FBNDNAME" "FVOL, clientname,
		    vds_port->vol->volname);
	}
	Xfree(clientname);

	if (dctl->extended) {
		if (dctl->parseable)
			rprintf(resp, "|ldc=");
		else
			rprintf(resp, " ");
		ldcs_print(&vds_port->ldcs, resp);
	}

	rprintf(resp, "\n");
}

static void
vds_volume_list(vds_volume_t *vol, response_t *resp, dispctl_t *dctl)
{
	char *dev = (vol->blkdev != NULL) ? vol->blkdev : "";
	char *opts = (vol->options != NULL) ? vol->options : "";
	char *mpg = (vol->mpgroup != NULL) ? vol->mpgroup->name : "";

	if (dctl->parseable) {
		rprintf(resp, "\n|vol=%s|opts=%s|dev=%s|mpgroup=%s",
		    vol->volname, opts, dev, mpg);
	} else {
		if (!dctl->physfirst) {
			rprintf(resp, "\n    "FNAME, "");
			if (dctl->services)
				rprintf(resp, " "FNAME, "");
		}
		rprintf(resp, " "FVOL" "FVOLOPTS" "FVOLMGP" %s",
		    vol->volname, opts, mpg, dev);
	}

	dctl->physfirst = 0;
}

static void
vds_list_one(vds_t *vds, response_t *resp, dispctl_t *dctl)
{
	int nclients, nvols;

	if (vds->binding != NULL) {
		nclients = CHAIN_LENGTH(vds->binding->vports);
	} else {
		nclients = 0;
	}

	nvols = CHAIN_LENGTH(vds->volumes);

	print_vds_header(vds->service, vds_attrs.name, nclients, resp, dctl);

	if (dctl->services) {
		if (dctl->parseable)
			rprintf(resp, "|ldom=%s", vds->ldp->namep);
		else
			rprintf(resp, " "FNAME, vds->ldp->namep);
	}

	if (nvols == 0) {
		rprintf(resp, "\n");
		return;
	}

	vds_volume_t *vds_vol;

	dctl->physfirst = 1;
	CHAIN_ITER(vds->volumes, vds_vol, link) {
		vds_volume_list(vds_vol, resp, dctl);
	}
	rprintf(resp, "\n");

	if (dctl->bindings && nclients > 0) {
		vds_port_t *vds_port;

		dctl->header = 1;

		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in tables with fixed
		 * column widths.
		 * WIDTH  HEADER	MEANING
		 * 27	  "CLIENT"	name of client of service
		 * 14	  "VOLUME"	virtual disk volume name
		 * 3	  "LDC"		Logical Domain Channel number
		 */
		if (!dctl->parseable) {
			rprintf(resp, "        "FBNDNAME" "FVOL,
			    gettext("CLIENT"), gettext("VOLUME"));
			if (dctl->extended)
				rprintf(resp, " %s", gettext("LDC"));
			rprintf(resp, "\n");
		}

		CHAIN_ITER(vds->binding->vports, vds_port, link) {
			vds_port_list(vds_port, resp, dctl);
		}
	}
}

void
vds_list(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	vds_t *vds;

	if (dctl->oformat && !dctl->odisk)
		return;

	if (CHAIN_LENGTH(ldp->vdss) == 0)
		return;

	dctl->header = 1;

	if (!dctl->parseable)
		rprintf(resp, "%s\n", dctl_label(vds_attrs.name, dctl));

	CHAIN_ITER(ldp->vdss, vds, ldom) {
		vds_list_one(vds, resp, dctl);
	}

	if (!dctl->parseable)
		rprintf(resp, "\n");
}

vds_port_t *
vds_port_find_by_ldcid(ldom_t *ldp, uint64_t ldcid)
{
	vds_t *vds;
	vds_port_t *vport;

	CHAIN_ITER(ldp->vdss, vds, ldom) {
		if (vds->binding == NULL)
			continue;
		CHAIN_ITER(vds->binding->vports, vport, link) {
			if (ldcs_search(&vport->ldcs, ldcid) != NULL)
				return (vport);
		}
	}

	return (NULL);
}

/*
 * Before freeing a volume, make sure to free from any mpgroup if member
 */
static void
destroy_volume(vds_volume_t *vol)
{
	ASSERT(vol->vds_port == NULL);	/* not in use */

	/*
	 * Remove volume if part of an mpgroup and from the vds's chain
	 */
	(void) vds_mpgroup_rem_vol(vol);

	CHAIN_UNLINK(vol->vds->volumes, vol, link);

	Xfree(vol->volname);
	Xfree(vol->blkdev);
	Xfree(vol->options);
	Xfree(vol);
}

static void
destroy_vds(vds_t *vds)
{
	ldom_t *ldp = vds->ldp;
	vds_volume_t *vol, *next_vol;

	ASSERT(vds->binding == NULL);

	/*
	 * Walk the volumes in the vds, unlink from the chain and free
	 */
	CHAIN_ITER_VOL(vds->volumes, vol, next_vol, link) {
		destroy_volume(vol);
	}

	CHAIN_UNLINK(ldp->vdss, vds, ldom);
	CHAIN_UNLINK(zeus.vds_services, vds, link);

	Xfree(vds->service);
	Xfree(vds);

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(ldp->vdss) == 0)
		destroy_res_pool(&ldp->bound.vds_cfg_handles);
#endif
}

void
vds_delete_all(ldom_t *ldp)
{
	vds_t *vds, *vds_next;

	CHAIN_ITER_VOL(ldp->vdss, vds, vds_next, ldom) {
		destroy_vds(vds);
	}
}

void
vds_destroy_port(vds_port_t *vds_port)
{
	CHAIN_UNLINK(vds_port->vds->binding->vports, vds_port, link);

	vds_port->vol->vds_port = NULL;	/* unlink from the volume */
	free_res(vds_port->port);
	ldcs_free(&vds_port->ldcs);
	Xfree(vds_port);
}

void
vds_unbind_port(vds_port_t *vds_port)
{
	ldcs_stage(&vds_port->ldcs);
	vds_destroy_port(vds_port);
}

/*
 * Called from the add-vds CLI.
 */
int
vds_add(ldom_t *ldp, char *service_name, response_t *resp)
{
	vds_t *vds;
	int	rc;

	if (vds_find_by_name(service_name) != NULL) {
		rprintf(resp, gettext("A %s service named %s already exists\n"),
		    vds_attrs.name, service_name);
		return (ERR_SERVICE_EXISTS);
	}

	vds = vds_new(ldp, service_name);

	if (ldp->state >= BOUND) {
		bind_vds(vds, NEW_INSTANCE);

		if (zeus.mode == ZEUS_MGMT_MODE) {
			if (ldp->state == BOUND) {
				/* do reconfig now */
				seq_download_mds_do_config(ldp);
				free_all_staged_ldcs(ldp);

			} else if (vio_check_dr_capable(ldp, resp) == true) {
				rc = vio_dr_add(ldp, vds_attrs.md_name,
				    vds_devid(vds), resp);
				/*
				 * If the DR operation failed and the dr-vio
				 * domain service determines that it is OK to
				 * rollback we remove the vnet again.
				 */
				if (rc != 0) {
					if (rc == ERR_DR_VIO_ROLLBACK) {
						unbind_vds(vds);
						destroy_vds(vds);
						seq_rm_reconfigure(ldp);
						free_all_staged_ldcs(ldp);
					}
					return (rc);
				}
			} else {
				/* handle as a delayed reconfig operation */
				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_ADD, resp);
			}
		}
	}

	database_ldom_add_disk_service(ldp, service_name);

	return (0);
}

static void
vds_remove_do_teardown(vds_t *vds)
{
	ASSERT(vds != NULL);
	database_ldom_del_disk_service(vds->ldp, vds->service);
	unbind_vds(vds);
	destroy_vds(vds);
}

/*
 * Called from the remove-vds CLI.
 */
/*ARGSUSED1*/
int
vds_remove(vds_t *vds, bool_t force, response_t *resp)
{
	ldom_t *ldp = vds->ldp;
	int	rv;

	if ((vds->binding != NULL) &&
	    (CHAIN_LENGTH(vds->binding->vports) > 0)) {
		/* clients are bound to the vds service */
		rprintf(resp, MSG_ERR_SERVICE_BOUND_CLIENTS);
		return (ERR_SERVICE_BOUND_CLIENTS);
	}

	switch (zeus.mode) {
	case ZEUS_MGMT_MODE:
		if (ldp->state < BOUND) {
			vds_remove_do_teardown(vds);

		} else if (ldp->state == BOUND) {
			/* do reconfig now */
			vds_remove_do_teardown(vds);
			seq_download_mds_do_config(ldp);
			free_all_staged_ldcs(ldp);

		} else if (vio_check_dr_capable(ldp, resp)) {
			/* Attempt DR operation */
			rv = vio_dr_rm_phase1(ldp, vds_attrs.md_name,
			    vds_devid(vds), resp);
			if (rv != 0)
				return (rv);
			vds_remove_do_teardown(vds);
			vio_dr_rm_phase2(ldp);

		} else {
			/* handle as a delayed reconfig operation */
			if (CHAIN_LENGTH(vds->volumes) > 0) {
				/* volumes are bound to the vds */
				rprintf(resp, gettext("Removal is not allowed; "
				    "the vds has one or more associated "
				    "volumes\n"));
				return (ERR_SERVICE_BOUND_CLIENTS);
			}
			vds_remove_do_teardown(vds);
			initiate_or_update_delayed_reconfig(ldp,
			    DEL_RECONFIG_REMOVE, resp);
		}
		break;

	case ZEUS_CONFIG_MODE:
		vds_remove_do_teardown(vds);
		free_all_staged_ldcs(ldp);
		break;

	default:
		ASSERT(0);
		return (ERR_GENERIC);
	}

	return (0);
}

/*
 * Called from the set-vdsdev CLI.
 */
int
vds_modify_dev(vds_t *vds, char *volname, char *options,
    char *mpgname, bool_t force, response_t *resp)
{
	vds_volume_t *vol, *match_vol;
	mpgroup_t *mpg;
	uint_t opt = 0;
	int err = 0;
	bool_t modify_opts = false;
	bool_t modify_mpg = false;

	vol = vds_find_volume_by_name(vds, volname);
	if (vol == NULL) {
		rprintf(resp, gettext("Volume %s does not exist on disk "
		    "server %s\n"), volname, vds->service);
		return (ERR_VDS_VOLUME_NOEXISTS);
	}

	/*
	 * Volume can't be modified if it's in use by a client.
	 */
	if (vol->vds_port != NULL) {
		rprintf(resp, gettext("Cannot modify this vdsdev, vdisk(s) "
		    "are bound\n"));
		return (ERR_VDS_VOLUME_BUSY);
	}

	if ((err = check_modify(vol, options, force, mpgname, resp, &modify_mpg,
	    &modify_opts)) != 0)
		return (err);

	if (modify_opts == false && modify_mpg == false) {
		rprintf(resp, gettext("No modifications made\n"));
		return (0);
	}

	/*
	 * Allocating a vdsdev to an mpgroup or to a different mpgroup is not
	 * allowed if there is already a vdisk bound to the new mpgroup
	 */
	if (modify_mpg == true && mpgname[0] != '\0' &&
	    (mpg = vds_find_mpgroup_by_name(mpgname)) != NULL) {
		CHAIN_ITER(mpg->volumes, match_vol, mpgroup_link) {
			if (match_vol->vds_port != NULL) {
				rprintf(resp, gettext("Cannot assign vdsdev to "
				    "mpgroup %s on vds %s on service\ndomain "
				    "\"%s\" because there is a vdisk bound to "
				    "this mpgroup\n"), mpgname, vds->service,
				    vds->ldp->namep);
				return (ERR_VDS_ARGS_INVALID);
			}
		}
	}

	if (modify_mpg == true) {
		if (mpgname[0] == '\0') {
			if (vol->mpgroup != NULL) {
				(void) vds_mpgroup_rem_vol(vol);
			}
		} else if (vol->mpgroup == NULL) {
			vds_mpgroup_add_vol(mpgname, vol);
		} else if (strcmp(mpgname, vol->mpgroup->name) != 0) {
			(void) vds_mpgroup_rem_vol(vol);
			vds_mpgroup_add_vol(mpgname, vol);
		}
	}

	if (modify_opts == true) {
		if (options[0] == '\0') {
			/* Set default using 'option=' , i.e. no option */
			if (vol->options != NULL) {
				Xfree(vol->options);
				vol->options = NULL;
			}
		} else if (vol->options == NULL) {
			vol->options = Xstrdup(options);
		} else if (opt != vds_decode_opts(vol->options)) {
			Xfree(vol->options);
			vol->options = Xstrdup(options);
		}
	}

	database_ldom_modify_vdsdev_props(vds->ldp, vds->service,
	    volname, (modify_opts == true) ? options : NULL,
	    (modify_mpg == true) ? mpgname : NULL);

	return (0);
}

/*
 * Called from the add-vdsdev CLI.
 */
int
vds_add_dev(vds_t *vds, char *volname, char *blkdev, char *options,
    char *mpgname, char *servicep, bool_t force, response_t *resp)
{
	vds_t *check_vds;
	vds_volume_t *match_vol;
	uint_t opt_val = vds_decode_opts(options);
	mpgroup_t *mpg;

	if (opt_val == VDS_DUP_OPTS) {
		rprintf(resp, gettext("Failed to add device %s\nas %s@%s\n"),
		    blkdev, volname, servicep);
		rprintf(resp, gettext("Duplicates found in options\n"));
		return (ERR_VDS_ARGS_INVALID);
	}

	/*
	 * Check whether the volume name is already used on this vds.
	 */
	if (vds_find_volume_by_name(vds, volname) != NULL) {
		rprintf(resp, MSG_ERR_ADD_DEV_FAILED, blkdev, volname,
		    servicep, vds->ldp->namep);
		rprintf(resp, gettext("Volume %s already exists in vds %s\n"),
		    volname, vds->service);
		return (ERR_VDS_VOLUME_EXISTS);
	}

	/*
	 * Allocating a vdsdev to an mpgroup is not allowed if there is already
	 * a vdisk bound to it
	 */
	if (mpgname != NULL && (mpg = vds_find_mpgroup_by_name(mpgname)) !=
	    NULL) {
		CHAIN_ITER(mpg->volumes, match_vol, mpgroup_link) {
			if (match_vol->vds_port != NULL) {
				rprintf(resp, gettext("Failed to add device "
				    "%s\nas %s@%s\n"), blkdev, volname,
				    servicep);
				rprintf(resp, gettext("Cannot add vdsdev to "
				    "mpgroup %s on vds %s on service\ndomain "
				    "\"%s\" because there is a vdisk bound to "
				    "this mpgroup\n"), mpgname, vds->service,
				    vds->ldp->namep);
				return (ERR_VDS_ARGS_INVALID);
			}
		}
	}

	/*
	 * Adding volumes pointing to the same backing store in the same
	 * service domain is allowed only if all the following conditions are
	 * met:
	 * 1) the existing volume and the one being added must be read only
	 * 2) If mpgroup is speicified on both, they must be different, i.e.
	 * all vdsdevs in the same mpgroup must be bound to vds servers in
	 * distinct domains
	 * 3) If excl option is set in either of them, they must be in same vds
	 * 4) The option must be the same on both
	 *
	 * In addition, all volumes in an mpgroup must have the same option
	 *
	 * FIXME: we currently use strcmp() to compare two blkdev entries to
	 * determine whether they point to the same backing store.  This is
	 * insufficient; it limits us to checking only within a guest, and
	 * even then, this check is not foolproof (i.e. symlinks cause
	 * problems).
	 */
	CHAIN_DOUBLE_ITER(vds->ldp->vdss, check_vds, ldom,
	    volumes, match_vol, link) {
		uint_t t_opt, mod;

		if (strcmp(match_vol->blkdev, blkdev) != 0)
			continue;

		t_opt = vds_decode_opts(match_vol->options);

		mod = (opt_val & VDEV_RONLY) & (t_opt & VDEV_RONLY);
		if (mod == 0) {
			if (force == false) {
				/*
				 * All vdsdev must be read only if force
				 * options isn't specified.
				 */
				rprintf(resp, MSG_ERR_ADD_DEV_FAILED, blkdev,
				    volname, servicep, vds->ldp->namep);
				rprintf(resp, MSG_ERR_VDSDEV_ONLY_PERMITTED,
				    "the device is exported read-only");
				return (ERR_VDS_ARGS_INVALID);
			}
		}

		/*
		 * 'excl' option can be set for only one backend
		 */
		if ((opt_val & VDEV_EXCL) != 0 ||
		    (t_opt & VDEV_EXCL) != 0) {
			rprintf(resp, MSG_ERR_ADD_DEV_FAILED, blkdev,
			    volname, servicep, vds->ldp->namep);
			rprintf(resp, gettext("You cannot add "
			    "multiple instances of the same device "
			    "with the excl option set.\n"));
			return (ERR_VDS_ARGS_INVALID);
		}

		/*
		 * Now check the rest of the options after removing the excl
		 * and read-only bits. The option must be the same on this
		 * and the matching volume
		 */
		if ((opt_val & ~(VDEV_RONLY | VDEV_EXCL)) != (t_opt &
		    ~(VDEV_RONLY | VDEV_EXCL))) {
			rprintf(resp, MSG_ERR_ADD_DEV_FAILED, blkdev,
			    volname, servicep, vds->ldp->namep);
			rprintf(resp, MSG_ERR_VDSDEV_ONLY_PERMITTED,
			    "the device is exported with the same set "
			    "of options");
			return (ERR_VDS_ARGS_INVALID);
		}

		/*
		 * If mpgroup is specified on this and the matching
		 * vdsdev, both must be different
		 */
		if (mpgname != NULL && match_vol->mpgroup != NULL &&
		    strcmp(mpgname, match_vol->mpgroup->name) == 0) {
			rprintf(resp, gettext("Failed to add device "
			    "%s as %s@%s because this device is "
			    "already exported on domain %s in mgroup "
			    "%s.\n"), blkdev, volname, servicep,
			    vds->ldp->namep, match_vol->mpgroup->name);
			rprintf(resp, MSG_ERR_VDSDEV_ONLY_PERMITTED,
			    "all devices are in different mpgroups");
			return (ERR_VDS_ARGS_INVALID);
		}
	}

	/*
	 * Check if a volume is being added to an existing mpgroup,
	 * the option is the same
	 */
	if (add_vds_volume(vds, volname, blkdev, options, mpgname,
	    resp) == NULL) {
		rprintf(resp, gettext("Failed to add device %s\nas %s@%s\n"),
		    blkdev, volname, servicep);
		return (ERR_VDS_ARGS_INVALID);
	}

	/*
	 * Add a record for the disk volume in the DB.
	 */
	database_ldom_add_disk_service_vol(vds->ldp, vds->service,
	    volname, blkdev, options, mpgname);

	/*
	 * A disk device with no clients is not represented in the guest MD,
	 * hence no MD update is needed.
	 */

	return (0);
}

/*
 * Called from the remove-vdsdev CLI.
 */
/*ARGSUSED2*/
int
vds_remove_dev(vds_t *vds, char *volname, bool_t force, response_t *resp)
{
	vds_volume_t *vol;

	vol = vds_find_volume_by_name(vds, volname);
	if (vol == NULL) {
		rprintf(resp, gettext("Volume %s does not exist on disk "
		    "server %s\n"), volname, vds->service);
		return (ERR_VDS_VOLUME_NOEXISTS);
	}

	/*
	 * Volume can't be deleted if it's in use by a client.
	 */
	if (vol->vds_port != NULL) {
		vdisk_t *client = vol->vds_port->vdisk_port->vdisk;
		rprintf(resp, gettext("Cannot delete this vdsdev, vdisk "
		    "client %s@%s is attached\n"), client->name,
		    client->ldp->namep);
		return (ERR_VDS_VOLUME_BUSY);
	}

	database_ldom_del_disk_service_block(vds->ldp, vds->service, volname);

	destroy_volume(vol);

	/*
	 * A disk device with no clients is not represented in the guest MD,
	 * hence no MD update is needed.
	 */

	return (0);
}

/*
 * Decode an options string to an integer value
 */
uint_t
vds_decode_opts(char *options)
{
	uint_t val = 0;
	char *opt, *opt_copy;

	if (options == NULL || options[0] == '\0')
		return (0);

	opt_copy = Xstrdup(options);
	opt = strtok(opt_copy, ",");

	while (opt != NULL) {
		int j = 0;

		while (vdsdev_opts[j] != NULL) {
			if (strcmp(opt, vdsdev_opts[j]) == 0)
				break;
			j++;
		}
		ASSERT(vdsdev_opts[j] != NULL);
		if (val & vdsdev_optval[j])
			return (VDS_DUP_OPTS);	/* duplicate */
		else
			val |= vdsdev_optval[j];
		opt = strtok(NULL, ",");
	}
	Xfree(opt_copy);

	return (val);
}

/*
 * Check that a modify operation on a vdsdev does not conflict with existing
 * vdsdev(s) on the same guest
 */
static int
check_modify(vds_volume_t *vol, char *options, bool_t force, char *mpgname,
	response_t *resp, bool_t *modify_mpg, bool_t *modify_opts)
{
	vds_t *vds = vol->vds;
	vds_t *check_vds;
	vds_volume_t *match_vol;
	uint_t opt_val = 0;
	int err = 0;
	char *volname;
	char *mpg;
	char *opt;

	*modify_mpg = *modify_opts = false;

	/*
	 * First ensure whether any modifications are being made
	 */
	if (mpgname != NULL) {
		if (mpgname[0] == '\0') {
			if (vol->mpgroup != NULL) {
				*modify_mpg = true;
			}
		} else if (vol->mpgroup == NULL) {
			*modify_mpg = true;
		} else if (strcmp(mpgname, vol->mpgroup->name) != 0) {
			*modify_mpg = true;
		}
	}

	if (options != NULL) {
		opt_val = vds_decode_opts(options);
		if (opt_val == VDS_DUP_OPTS) {
			rprintf(resp, "Duplicates found in options\n");
			return (ERR_VDS_ARGS_INVALID);
		}

		if (options[0] == '\0') {
			/* Set default using 'option=' , i.e. no option */
			if (vol->options != NULL) {
				*modify_opts = true;
			}
		} else if (vol->options == NULL) {
			*modify_opts = true;
		} else if (opt_val != vds_decode_opts(vol->options)) {
			*modify_opts = true;
		}
	}

	if (*modify_mpg == false && *modify_opts == false)
		return (0);

	/*
	 * ensure that the (new) option matches the options of the
	 * (new) mpgroup
	 */
	mpg = (*modify_mpg == true) ? mpgname :
	    ((vol->mpgroup == NULL) ? NULL : vol->mpgroup->name);
	opt = (*modify_opts == true) ? options : vol->options;

	if (add_mpg_check(mpg, opt, vol, resp) == false)
		return (ERR_VDS_ARGS_INVALID);

	volname = zsprintf("%s@%s", vol->volname, vds->service);
	/*
	 * Cycle through all vdsdev in the guest and see there is a conflict
	 * with another vdsdev
	 */
	CHAIN_DOUBLE_ITER(vds->ldp->vdss, check_vds, ldom,
	    volumes, match_vol, link) {
		uint_t t_opt, mod;

		if (strcmp(match_vol->blkdev, vol->blkdev) != 0 ||
		    match_vol == vol)
			continue;

		if (*modify_opts == true) {
			t_opt = vds_decode_opts(match_vol->options);

			mod = (opt_val & VDEV_RONLY) & (t_opt & VDEV_RONLY);
			if (mod == 0) {
				if (force == false) {
					/*
					 * All vdsdev must be read only if force
					 * options isn't specified.
					 */
					rprintf(resp, gettext("Cannot modify "
					    "\"%s\" because its device matches "
					    "an existing vdsdev\non service "
					    "domain \"%s\". All the vdsdev on "
					    "the same guest with the same "
					    "\nblock-dev must all be read-only"
					    "\n"), volname, vds->ldp->namep);
					err = ERR_VDS_ARGS_INVALID;
					goto ret;
				}
			}

			/*
			 * Only one backend in an LDom can have the 'excl'
			 * option set.
			 */
			if ((opt_val & VDEV_EXCL) != 0 ||
			    (t_opt & VDEV_EXCL) != 0) {
				rprintf(resp, gettext("Cannot modify \"%s\" "
				    "because its device matches an existing "
				    "vdsdev\non service domain \"%s\".\nA "
				    "backing store with excl option set cannot "
				    "be exported multiple times\n"), volname,
				    vds->ldp->namep);
				err = ERR_VDS_ARGS_INVALID;
				goto ret;
			}

			/*
			 * The option must be the same on this and the
			 * matching volume
			 */
			if ((opt_val & ~(VDEV_RONLY | VDEV_EXCL)) != (t_opt &
			    ~(VDEV_RONLY | VDEV_EXCL))) {
				rprintf(resp, gettext("Cannot modify \"%s\" "
				    "because its device matches an existing "
				    "vdsdev\non service domain \"%s\". All the "
				    "vdsdev with the same block-dev on the same"
				    "\nservice domain must all have the same "
				    "option\n"), volname, vds->ldp->namep);
				err = ERR_VDS_ARGS_INVALID;
				goto ret;
			}
		}

		if (*modify_mpg == true) {
			/*
			 * If mpgroup is specified on this and the matching
			 * vdsdev, both must be different
			 */
			if (mpgname != NULL && match_vol->mpgroup != NULL &&
			    strcmp(mpgname, match_vol->mpgroup->name) == 0) {
				rprintf(resp, gettext("Cannot modify \"%s\" "
				    "because its device matches an existing "
				    "vdsdev\non service domain \"%s\" in "
				    "mpgroup \"%s\". All the vdsdev with the "
				    "same\nblock-dev on the same guest must "
				    "all be in different mpgroups\n"), volname,
				    vds->ldp->namep, match_vol->mpgroup->name);
				err = ERR_VDS_ARGS_INVALID;
				goto ret;
			}
		}

	}

ret:
	Xfree(volname);
	return (err);
}
