/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <string.h>
#include <alloca.h>

#include "allocate.h"
#include "dr.h"
#include "ds_impl.h"
#include "ds_registry.h"
#include "errors.h"
#include "fatal.h"
#include "ldom.h"
#include "mdesc_impl.h"
#include "response.h"
#include "console.h"
#include "vio.h"
#include "zpri.h"
#include "zprintf.h"
#include "autosave.h"

/*
 * Save properties of a guest to SP domain services channel.  All the
 * necessary information is saved for later re-generation of the channel in HV
 * and guest MDs.
 */
static void
save_ds_sp_port(ldom_t *ldp, ldc_epdb_t *ep, md_t *guestmd, md_node_t *port)
{
	md_t *dsmd;
	md_node_t *new_port, *endpoint, **ldc_nodes;
	guest_sp_ldc_t *ds_sp_ldc;
	uint64_t id;
	int num;

	/* first allocate ds port id */
	if (md_get_prop_val(guestmd, port, "id", &id) != 0)
		fatal("DS port for LDom %s has no id property", ldp->namep);
	if (pull_res_id(&ldp->free_ds_ports, id) == NULL)
		fatal("Could not allocate DS port id (0x%"PRIx64") for LDom %s",
		    id, ldp->namep);

	/* the guest MD information is saved in a small MD */
	dsmd = md_new_md();

	new_port = md_new_node(dsmd, DS_PORT_NODE_NAME);
	md_copy_properties(dsmd, new_port, port);

	/* find ldc channel endpoint node under the port node */
	ldc_nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(guestmd));
	num = md_scan_dag(guestmd, port, md_find_name(guestmd,
	    CHANNEL_ENDPOINT), md_find_name(guestmd, "fwd"), ldc_nodes);
	ASSERT(num == 1);

	endpoint = md_link_new_node(dsmd, CHANNEL_ENDPOINT, new_port,
	    "fwd", "back");
	md_copy_properties(dsmd, endpoint, ldc_nodes[0]);

	ldp->ds_sp_md = dsmd;

	ds_sp_ldc = Xcalloc(1, guest_sp_ldc_t);

	ds_sp_ldc->channel = Xcalloc(1, ldc_chan_t);
	ds_sp_ldc->channel->chan = pull_res_id(&ldp->guest_free_ldcs, ep->ldc);
	if (ds_sp_ldc->channel->chan == NULL)
		fatal("LDom %s channel 0x%"PRIx64" not available",
		    ldp->namep, ep->ldc);

	ds_sp_ldc->channel->id = ep->id;
	ds_sp_ldc->target_channel = ep->peer->ldc;

	CHAIN_ADD(ldp->sp_ldcs, ds_sp_ldc, link);

	Xfree(ldc_nodes);
}

/*
 * Add the Domain Services clients.
 */
bool_t
add_domain_services_from_md(ldom_t *ldp, md_t *guest_md)
{
	int i, num, num_ports, num_ldc_ids;
	md_node_t **nodes;
	ds_port_t *ds_port;
	uint64_t reg;
	uint64_t *ldc_ids;
	ldc_epdb_t *ep;

	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(guest_md));

	num = md_scan_dag(guest_md, NULL,
	    md_find_name(guest_md, DS_NODE_NAME),
	    md_find_name(guest_md, "fwd"), nodes);
	if (num != 1)
		fatal("LDom %s MD does not contain one %s node",
		    ldp->namep, DS_NODE_NAME);

	num_ports = md_scan_dag(guest_md, nodes[0],
	    md_find_name(guest_md, DS_PORT_NODE_NAME),
	    md_find_name(guest_md, "fwd"), nodes);

	if (num_ports == 0) {
		warning("LDom %s MD has no %s node(s)", ldp->namep,
		    DS_PORT_NODE_NAME);
	}

	for (i = 0; i < num_ports; i++) {
		md_node_t *ds_port_node = nodes[i];

		/* Find out instance # */
		if (md_get_prop_val(guest_md, ds_port_node, "id", &reg) != 0)
			fatal("%s has no id property", DS_PORT_NODE_NAME);

		get_ldc_ids(guest_md, ds_port_node, &num_ldc_ids, &ldc_ids);

		if (num_ldc_ids != 1)
			fatal("Domain services port has multiple LDC IDs");

		ep = ldc_epdb_find(ldp->idp->id, ldc_ids[0]);
		if (ep == NULL)
			fatal("No endpoint in HV MD for LDC 0x%"PRIx64""
			    " in LDom %s", ldc_ids[0], ldp->namep);

		if (ep->peer->gid == LDCDB_SPGID) {
			/* domain<->SP domain services channel */
			save_ds_sp_port(ldp, ep, guest_md, ds_port_node);
		} else {
			ds_port = ds_port_new(ldp, reg);
			CHAIN_ADD(ldp->ds_ports, ds_port, ldom);

			/*
			 * If the domain was in the process of being migrated,
			 * do not bind the DS port here. Rather, the migration
			 * recovery will handle this later in the startup.
			 */
			if (ldp->flag.migration_source == false) {
				/* allocate the LDC IDs */
				ldc_pull_ids(ldp, &ldp->guest_free_ldcs,
				    num_ldc_ids, ldc_ids, &ds_port->ldcs);
			} else {
				DBG(zprintf("skipping DS port LDC ID "
				    "allocation for LDom %s\n", ldp->namep));
			}
		}

		Xfree(ldc_ids);
	}

	Xfree(nodes);

	return (true);
}

static void
add_vdevices(ldom_t *ldp, md_t *md, int n_vdevices, md_node_t **vdevices)
{
	for (int i = 0; i < n_vdevices; ++i) {
		char *md_name;
		md_node_t *vdevice = vdevices[i];

		if (md_get_prop_str(md, vdevice, "name", &md_name) != 0)
			fatal("vdevice has no name property");

		if (strcmp(md_name, vldc_attrs.md_name) == 0)
			vldc_add_from_mdnode(ldp, md, vdevice);
		else if (strcmp(md_name, vldcc_attrs.md_name) == 0)
			vldcc_add_from_mdnode(ldp, md, vdevice);
		else if (strcmp(md_name, vcc_attrs.md_name) == 0)
			vcc_add_from_mdnode(ldp, md, vdevice);
		else if (strcmp(md_name, vsw_attrs.md_name) == 0)
			vsw_add_from_mdnode(ldp, md, vdevice);
		else if (strcmp(md_name, vnet_attrs.md_name) == 0)
			vnet_add_from_mdnode(ldp, md, vdevice);
		else if (strcmp(md_name, vds_attrs.md_name) == 0)
			vds_add_from_mdnode(ldp, md, vdevice);
		else if (strcmp(md_name, vdisk_attrs.md_name) == 0)
			vdisk_add_from_mdnode(ldp, md, vdevice);
		else if (strcmp(md_name, vdpcs_attrs.md_name) == 0)
			vdpcs_add_from_mdnode(ldp, md, vdevice);
		else if (strcmp(md_name, vdpcc_attrs.md_name) == 0)
			vdpcc_add_from_mdnode(ldp, md, vdevice);
		else if (strcmp(md_name, "virtual-domain-service") == 0) {
			DBG(zprintf("Found virtual-domain-service vdev in MD;\n"
			    "handled as part of special-case domain services "
			    "code.\n"));
		} else {
			warning("Skipping unknown virtual device %s", md_name);
		}
	}
}

/*
 * We need to resurrect the guest VIO state by looking at the MD tree. There
 * are two nodes of interest, the "vdevices" node under which all the clients
 * and the services are present as individual "vdevice" nodes. The second one
 * is the "domain-services" node under root, which has the details of the
 * domain-service service for the guest.
 */
void
grab_guest_vio_state(ldom_t *ldp, md_t *guest_md, md_node_t *hv_guest,
    md_t *hv_md)
{
	int		n_vdevices;
	md_node_t	**vdevnodelistp;

	vdevnodelistp = Xmalloc(sizeof (md_node_t *) *
	    md_node_count(guest_md));

	/*
	 * We only check LDC based virtual-devices here; other
	 * virtual-devices are scanned in grab_guest_platform_services().
	 */
	n_vdevices = md_scan_dag(guest_md, NULL,
	    md_find_name(guest_md, "channel-devices"),
	    md_find_name(guest_md, "fwd"), vdevnodelistp);

	if (n_vdevices != 1)
		fatal("%d channel-devices node(s) found", n_vdevices);

	n_vdevices = md_scan_dag(guest_md, vdevnodelistp[0],
	    md_find_name(guest_md, VDEV_NAME),
	    md_find_name(guest_md, "fwd"), vdevnodelistp);

	add_vdevices(ldp, guest_md, n_vdevices, vdevnodelistp);

	Xfree(vdevnodelistp);

	/*
	 * Console and domain services are special cases as they have no
	 * corresponding vdevice node.
	 */
	console_add_from_md(ldp, hv_md, hv_guest);

	if (add_domain_services_from_md(ldp, guest_md))
		DBG(zprintf("Added domain-service for guest %s\n", ldp->namep));
}

void
add_sp_vio(ldom_t *ldp)
{
	md_node_t **nodes;
	md_node_t *chandevs;
	int n_vdevices;

	if ((chandevs = md_find_node(pri.pri, "channel-devices")) == NULL)
		return;

	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(pri.pri));

	n_vdevices = md_scan_dag(pri.pri, chandevs,
	    md_find_name(pri.pri, VDEV_NAME),
	    md_find_name(pri.pri, "fwd"), nodes);

	add_vdevices(ldp, pri.pri, n_vdevices, nodes);

	Xfree(nodes);
}

void
finalize_vio_bindings(void)
{
	ldom_t *ldp;

	CHAIN_ITER(zeus.ldoms, ldp, link) {

		if (ldp->flag.migration_source)
			continue;

		link_vldc_sp_ports(ldp);
		link_vldcc_clients(ldp);
		link_vnet_clients(ldp);
		link_vdisk_clients(ldp);
		link_vdpcc_clients(ldp);
		link_console(ldp);
		link_domain_services(ldp);
	}
}

/*
 * Routine to create an "ldc-endpoint" node under a specified node.
 */
md_node_t *
vio_gen_ldc_endpoint_node(md_t *md, md_node_t *hv_guest, md_node_t *parent,
    uint64_t type, uint64_t target_type, ldc_chan_t *channel,
    uint64_t target_guest, ldc_chan_t *target_channel)
{
	md_node_t *ldc_endpoint;

	ldc_endpoint = md_link_new_node(md, "ldc_endpoint", parent,
	    "fwd", "back");

	if (hv_guest != NULL) {
		md_add_arc_property(md, hv_guest, "fwd", ldc_endpoint);
		md_add_arc_property(md, ldc_endpoint, "back", hv_guest);
	}

	/* required properties */
	md_add_value_property(md, ldc_endpoint, "target_type", target_type);
	md_add_value_property(md, ldc_endpoint, "channel", channel->chan->id);
	md_add_value_property(md, ldc_endpoint, "resource_id", channel->id->id);

	/* guest endpoint needs tx and rx inos */
	if (type == TYPE_GUEST) {
		md_add_value_property(md, ldc_endpoint, "tx-ino",
		    TX_INO(channel->chan->id));
		md_add_value_property(md, ldc_endpoint, "rx-ino",
		    RX_INO(channel->chan->id));
	}

	/* only required when target is guest */
	if (target_type == TYPE_GUEST)
		md_add_value_property(md, ldc_endpoint,
		    "target_guest", target_guest);

	md_add_value_property(md, ldc_endpoint, "target_channel",
	    target_channel->chan->id);

	return (ldc_endpoint);
}

void
vio_gen_ldcs_endpoint_nodes(md_t *md, md_node_t *hv_guest, md_node_t *parent,
    uint64_t type, uint64_t target_type, ldcs_t *ldcs,
    uint64_t target_guest, ldcs_t *target_ldcs, md_node_t ***endpoints)
{
	ldc_chan_t *ldc, *target_ldc;
	int i, num_endpoints;

	num_endpoints = CHAIN_LENGTH(ldcs->ldcs);
	if (num_endpoints != CHAIN_LENGTH(target_ldcs->ldcs))
		fatal("Mismatch in number of local and target LDCs");

	if (endpoints != NULL)
		*endpoints = Xmalloc(num_endpoints * sizeof (md_node_t *));

	target_ldc = CHAIN_START(target_ldcs->ldcs);
	i = 0;
	CHAIN_ITER(ldcs->ldcs, ldc, link) {
		md_node_t *endpoint;
		endpoint = vio_gen_ldc_endpoint_node(md, hv_guest, parent, type,
		    target_type, ldc, target_guest, target_ldc);
		if (endpoints != NULL)
			(*endpoints)[i++] = endpoint;

		target_ldc = CHAIN_NEXT(target_ldc->link);
	}
}

/*
 * Generate non VIO related guest LDC endpoints for channels to the SP.
 */
static void
guest_sp_gen_ldc_endpoints(ldom_t *ldp, md_t *md, md_node_t *hv_guest,
    md_node_t *parent)
{
	guest_sp_ldc_t	*guest_sp_ldc;
	md_node_t	*ldc_endpoint;
	uint64_t	ldc_id;

	CHAIN_ITER(ldp->sp_ldcs, guest_sp_ldc, link) {
		ldc_endpoint = md_link_new_node(md, "ldc_endpoint",
		    parent, "fwd", "back");

		md_add_arc_property(md, hv_guest, "fwd", ldc_endpoint);
		md_add_arc_property(md, ldc_endpoint, "back", hv_guest);

		ldc_id = guest_sp_ldc->channel->chan->id;

		md_add_value_property(md, ldc_endpoint, "target_type", TYPE_SP);
		md_add_value_property(md, ldc_endpoint, "channel", ldc_id);
		md_add_value_property(md, ldc_endpoint, "resource_id",
		    guest_sp_ldc->channel->id->id);
		md_add_value_property(md, ldc_endpoint, "target_channel",
		    guest_sp_ldc->target_channel);
		md_add_value_property(md, ldc_endpoint, "tx-ino",
		    TX_INO(ldc_id));
		md_add_value_property(md, ldc_endpoint, "rx-ino",
		    RX_INO(ldc_id));
	}
}

static void
ds_gen_ldc_endpoints(ldom_t *ldp, md_t *md, md_node_t *hv_guest,
    md_node_t *parent)
{
	ds_port_t *ds_port;

	CHAIN_ITER(ldp->ds_ports, ds_port, ldom) {
		if (DS_PORT_IS_VLDC(ds_port)) {
			vldc_port_t *vldc_port = ds_port->vldc_port;
			DBG(zprintf("vldc:%s: guest %s, port %"PRId64"\n",
			    __func__, ldp->namep, ds_port->port->id));

			vio_gen_ldcs_endpoint_nodes(md, hv_guest, parent,
			    TYPE_GUEST, TYPE_GUEST, &ds_port->ldcs,
			    vldc_port->vldc->ldp->idp->id, &vldc_port->ldcs,
			    NULL);
		} else if (DS_PORT_IS_VLDS(ds_port)) {
			ds_port_t *peer_ds_port = ds_port->peer_ds_port;
			DBG(zprintf("ds:%s: guest %s, port %"PRId64"\n",
			    __func__, ldp->namep, ds_port->port->id));

			vio_gen_ldcs_endpoint_nodes(md, hv_guest, parent,
			    TYPE_GUEST, TYPE_GUEST, &ds_port->ldcs,
			    peer_ds_port->ldp->idp->id, &peer_ds_port->ldcs,
			    NULL);
		}
	}
}

/*
 * Generate all the LDC endpoint nodes for a LDom.
 */
void
vio_gen_ldc_endpoint_nodes(ldom_t *ldp, md_t *hv_md, md_node_t *hv_guest,
    md_node_t *ldc_endpoints)
{
	vldc_gen_ldc_endpoints(ldp, hv_md, hv_guest, ldc_endpoints);
	vcc_gen_ldc_endpoints(ldp, hv_md, hv_guest, ldc_endpoints);
	vsw_gen_ldc_endpoints(ldp, hv_md, hv_guest, ldc_endpoints);
	vds_gen_ldc_endpoints(ldp, hv_md, hv_guest, ldc_endpoints);
	vdpcs_gen_ldc_endpoints(ldp, hv_md, hv_guest, ldc_endpoints);

	vldcc_gen_ldc_endpoints(ldp, hv_md, hv_guest, ldc_endpoints);
	console_gen_ldc_endpoints(ldp, hv_md, hv_guest, ldc_endpoints);
	vnet_gen_ldc_endpoints(ldp, hv_md, hv_guest, ldc_endpoints);
	vdisk_gen_ldc_endpoints(ldp, hv_md, hv_guest, ldc_endpoints);
	vdpcc_gen_ldc_endpoints(ldp, hv_md, hv_guest, ldc_endpoints);

	ds_gen_ldc_endpoints(ldp, hv_md, hv_guest, ldc_endpoints);
	guest_sp_gen_ldc_endpoints(ldp, hv_md, hv_guest, ldc_endpoints);
}

void
gen_guest_ldc_nodes(md_t *md, md_node_t *vport, md_node_t *chan_endpoints,
    ldcs_t *ldcs)
{
	ldc_chan_t *ldc;

	CHAIN_ITER(ldcs->ldcs, ldc, link) {
		md_node_t *ldc_node;

		/* add new endpoint under channel endpoints aggregator node */
		ldc_node = md_link_new_node(md, CHANNEL_ENDPOINT,
		    chan_endpoints, "fwd", "back");

		md_add_value_property(md, ldc_node, "id", ldc->chan->id);
		md_add_value_property(md, ldc_node, "tx-ino",
		    TX_INO(ldc->chan->id));
		md_add_value_property(md, ldc_node, "rx-ino",
		    RX_INO(ldc->chan->id));

		/* link the endpoint node with the vport node node */
		md_add_arc_property(md, vport, "fwd", ldc_node);
		md_add_arc_property(md, ldc_node, "back", vport);
	}
}

md_node_t *
vio_gen_vdev_port_node(md_t *md, md_node_t *vdev, md_node_t *chan_endpoints,
    char *name, uint64_t port_id, ldcs_t *ldcs)
{
	md_node_t *vport;

	vport = md_link_new_node(md, VDEV_PORT, vdev, "fwd", "back");

	md_add_string_property(md, vport, "name", name);
	md_add_value_property(md, vport, "id", port_id);

	/* add the channel endpoints */
	gen_guest_ldc_nodes(md, vport, chan_endpoints, ldcs);

	return (vport);
}

void
vio_gen_guest_md(ldom_t *ldp, md_t *md)
{
	md_node_t *vdevs, *cdevs, *chan_endpoints;

	/*
	 * The top-level virtual node is "virtual-devices". One might already
	 * be in the MD, coming from the PRI template. If it doesn't exist
	 * then create it.
	 */
	if ((vdevs = md_find_node(md, VDEVS_NAME)) == NULL) {
		vdevs = md_link_new_node(md, VDEVS_NAME,
		    md_root_node(md), "fwd", "back");
		md_add_string_property(md, vdevs, "name", VDEVS_NAME);
		md_add_string_property(md, vdevs, "device-type",
		    VDEVS_DEV_TYPE);
		md_add_value_property(md, vdevs, "cfg-handle",
		    VDEVS_CFG_HANDLE);
		md_add_data_property(md, vdevs, "compatible",
		    strlen(VDEVS_COMPAT) + 1, (uint8_t *)VDEVS_COMPAT);
	}

	/*
	 * LDC based VIO devices live under the "channel-devices" node. As in
	 * the case with virtual-devices, one might already exist in the MD,
	 * coming from the PRI template.
	 */
	if ((cdevs = md_find_subnode(md, vdevs, CHANNELS_NAME)) == NULL) {
		cdevs = md_link_new_node(md, CHANNELS_NAME, vdevs,
		    "fwd", "back");
		md_add_string_property(md, cdevs, "name", CHANNELS_NAME);
		md_add_string_property(md, cdevs, "device-type",
		    CHANNELS_DEV_TYPE);
		md_add_value_property(md, cdevs, "cfg-handle",
		    CHANNELS_CFG_HANDLE);
		md_add_data_property(md, cdevs, "compatible",
		    strlen(CHANNELS_COMPAT) + 1, (uint8_t *)CHANNELS_COMPAT);
	}

	/*
	 * Check for the channel endpoints node and create it if needed.
	 */
	if ((chan_endpoints = md_find_node(md, CHANNEL_ENDPOINTS)) == NULL)
		chan_endpoints = md_link_new_node(md, CHANNEL_ENDPOINTS,
		    md_root_node(md), "fwd", "back");

	/*
	 * Now add MD nodes for all the VIO servers and clients. Note that
	 * there is no channel device MD node under channel devices.
	 */
	vldc_gen_guest_md(ldp, md, cdevs, chan_endpoints);
	vcc_gen_guest_md(ldp, md, cdevs, chan_endpoints);
	vsw_gen_guest_md(ldp, md, cdevs, chan_endpoints);
	vds_gen_guest_md(ldp, md, cdevs, chan_endpoints);
	vdpcs_gen_guest_md(ldp, md, cdevs, chan_endpoints);

	vldcc_gen_guest_md(ldp, md, cdevs, chan_endpoints);
	vnet_gen_guest_md(ldp, md, cdevs, chan_endpoints);
	vdisk_gen_guest_md(ldp, md, cdevs, chan_endpoints);
	vdpcc_gen_guest_md(ldp, md, cdevs, chan_endpoints);

	/*
	 * Add nodes for guest domain services.
	 */
	ds_gen_guest_md(ldp, md, chan_endpoints);
}

resid_t *
vio_get_cfg_handle(int cfg_handle, res_pool_t *free_cfg_handles)
{
	if (cfg_handle == NEW_INSTANCE)
		return (alloc_res(free_cfg_handles));
	else
		return (pull_res_id(free_cfg_handles, cfg_handle));
}

void
print_server_header(char *server, char *servertype, int nclients,
    response_t *resp, dispctl_t *dctl)
{
	if (dctl->parseable) {
		rprintf(resp, "%s", dctl_label(servertype, dctl));
		rprintf(resp, "|name=%s", server);

		if (dctl->list) {
			if (dctl->parseable)
				rprintf(resp, "|nclients=%d", nclients);
			else if (nclients > 1)
				rprintf(resp, " (%d clients)", nclients);
			else if (nclients == 1)
				rprintf(resp, " (%d client)", nclients);
		}
	} else {
		rprintf(resp, "    "FNAME, server);
	}
}

void
print_vio(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	vcc_list(ldp, resp, dctl);
	vsw_list(ldp, resp, dctl);
	vds_list(ldp, resp, dctl);
	vdpcs_list(ldp, resp, dctl);
	vldc_list(ldp, resp, dctl);

	if (!dctl->services || dctl->oformat) {
		vnet_list(ldp, resp, dctl);
		vdisk_list(ldp, resp, dctl);
		vdpcc_list(ldp, resp, dctl);
		vldcc_list(ldp, resp, dctl);
		console_list(ldp, resp, dctl);
	}
}

char *
vio_canonical_client_name(ldom_t *ldp, char *client_class, int instance)
{
	return (zsprintf("%s-%s%d", ldp->namep, client_class, instance));
}

char *
vio_canonical_service_name(ldom_t *ldp, char *service_type, int instance)
{
	return (zsprintf("%s-%s%d", ldp->namep, service_type, instance));
}

#if BIND_CHECK

static int
bind_check(ldom_t *ldp, response_t *resp)
{
	ldom_t *lp;
	int rc;

	CHAIN_ITER(zeus.ldoms, lp, link) {
		lp->work.num_ldcs = 0;
	}

	if ((rc = vnet_bind_check(ldp, resp)) != 0)
		return (rc);
	if ((rc = vdisk_bind_check(ldp, resp)) != 0)
		return (rc);
	if ((rc = vdpcc_bind_check(ldp, resp)) != 0)
		return (rc);
	if ((rc = console_bind_check(ldp, resp)) != 0)
		return (rc);
	ds_bind_check(ldp);

	CHAIN_ITER(zeus.ldoms, lp, link) {
		DBG(zprintf("%s: %s needs %d new channels\n", __func__,
		    lp->namep, lp->work.num_ldcs));

		if ((lp->work.num_ldcs > 0) && ldp->flag.locked) {
			rprintf(resp, gettext("Cannot bind Ldom %s because "
			    "that would modify locked LDom %s\n"),
			    ldp->namep, lp->namep);
			return (ERR_LDOM_LOCKED);
		}
		if (available_res_ids(&lp->guest_free_ldcs) <
		    lp->work.num_ldcs) {
			rprintf(resp, gettext("Insufficient free LDC channels "
			    "on LDom %s to bind LDom %s\n"),
			    lp->namep, ldp->namep);
			return (ERR_LDC_NO_IDS);
		}
	}

	return (0);
}

#endif

int
vio_bind(ldom_t *ldp, response_t *resp)
{
	int rc;

#if BIND_CHECK
	if ((rc = bind_check(ldp, resp)) != 0)
		return (rc);
#endif
	/*
	 * Bind all the servers first as there may be a client binding to
	 * one of them.
	 */
	if ((rc = vcc_bind_all(ldp, resp)) != 0)
		return (rc);
	if ((rc = vsw_bind_all(ldp, resp)) != 0)
		return (rc);
	if ((rc = vds_bind_all(ldp, resp)) != 0)
		return (rc);
	if ((rc = vdpcs_bind_all(ldp, resp)) != 0)
		return (rc);

	if ((rc = vnet_bind_all(ldp, resp)) != 0)
		return (rc);
	if ((rc = vdisk_bind_all(ldp, resp)) != 0)
		return (rc);
	if ((rc = vdpcc_bind_all(ldp, resp)) != 0)
		return (rc);

	if ((rc = ds_bind(ldp, resp)) != 0)
		return (rc);

	rc = console_bind(ldp, resp);

	return (rc);
}

int
vio_unbind(ldom_t *ldp, response_t *resp)
{
	console_unbind(ldp);
	ds_unbind(ldp);

	vdpcc_unbind_all(ldp, resp);
	vdisk_unbind_all(ldp, resp);
	vnet_unbind_all(ldp, resp);

	vdpcs_unbind_all(ldp, resp);
	vds_unbind_all(ldp, resp);
	vsw_unbind_all(ldp, resp);
	vcc_unbind_all(ldp, resp);

	return (0);
}

void
vio_delete(ldom_t *ldp)
{
	ds_port_t *ds_port, *temp_ds_port;

	console_delete(ldp);

	CHAIN_ITER_VOL(ldp->ds_ports, ds_port, temp_ds_port, ldom) {
		ds_delete(ldp, ds_port);
	}

	vdpcc_delete_all(ldp);
	vdisk_delete_all(ldp);
	vnet_delete_all(ldp);

	vdpcs_delete_all(ldp);
	vds_delete_all(ldp);
	vsw_delete_all(ldp);
	vcc_delete_all(ldp);
}

/*
 * Returns true iff LDom 'ldp' is providing services.
 */
bool_t
is_providing_services(ldom_t *ldp)
{
	return (ldp->bound.num_services > 0);
}

int *
parse_num_list(char *vlans, int *num)
{
	int i = 0;
	char *str, *vid;
	int *values = NULL;

	if (vlans == NULL) {
		*num = 0;
		return (NULL);
	} else if (IS_EMPTY_STR(vlans)) {
		*num = 1;
		return (NULL);
	}

	/*
	 * Make a copy of the argument to find out number of vlans
	 */
	str = alloca(strlen(vlans) + 1);
	strcpy(str, vlans);

	if ((vid = strtok(str, ",")) != NULL) {
		i++;
		while ((vid = strtok(NULL, ",")) != NULL) {
			i++;
		}
	}

	*num = i;
	if (*num == 0)
		return (NULL);

	strcpy(str, vlans);

	values = Xcalloc(i, int);
	i = 0;
	if ((vid = strtok(str, ",")) != NULL) {
		char *esp;
		values[i] = strtoll(vid, &esp, 10);
		if ((esp == vid) || (esp[0] != '\0'))
			goto err;
		i++;
		while ((vid = strtok(NULL, ",")) != NULL) {
			values[i] = strtoll(vid, &esp, 10);
			if ((esp == vid) || ((esp[0] != ',') &&
			    (esp[0] != '\0')))
				goto err;
			i++;
		}
	}

	return (values);

err:
	*num = 0;
	Xfree(values);
	values = NULL;

	return (values);
}

int
are_vio_dr_services_online(ldom_t *ldp)
{
	ds_svc_t *svc;

	ASSERT(ldp->state == ACTIVE);

	svc = ds_lookup_svc(ldp->ds_chan, DR_VIO_SVC);
	if (svc == NULL || svc == DS_SVC_HDL_INVALID)
		return (ERR_DR_VIO_NOTSUP);

	svc = ds_lookup_svc(ldp->ds_chan, MD_UPDATE_SVC);
	if (svc == NULL || svc == DS_SVC_HDL_INVALID)
		return (ERR_MD_UPDATE_NOTSUP);

	return (0);
}

/*
 * This function checks if the domain 'ldp' is capable of doing
 * a VIO DR operation. Conditions that need to be met in order
 * for a VIO DR operation to succeed.
 *	- Domain must be active
 *	- No delayed reconfig operation in progress on domain
 *	- VIO DR services must be running in domain
 *
 * Return:
 *	'true' if domain can do VIO DR operation
 *	'false' otherwise.
 */
bool_t
vio_check_dr_capable(ldom_t *ldp, response_t *resp)
{
	int rv;

	if (ldp->state == STARTING || ldp->state == STOPPING) {
		rprintf(resp, gettext("Cannot dynamically add/remove VIO device"
		    " as LDom %s is currently %s\n"), ldp->namep,
		    ldom_state_str(ldp->state));
		return (false);
	}

	if (ldp->state != ACTIVE)
		return (false);

	/* just in case: this should have been caught earlier in the code */
	if (in_delayed_reconfig(NULL) == true)
		return (false);

	rv = are_vio_dr_services_online(ldp);

	switch (rv) {
	case 0: /* success */
		break;

	case ERR_DR_VIO_NOTSUP:
		rprintf(resp, MSG_ERR_DR_VIO_NOTSUP, ldp->namep);
		break;

	case ERR_MD_UPDATE_NOTSUP:
		rprintf(resp, MSG_ERR_MD_NOTSUP, ldp->namep);
		break;

	default:
		FATAL_UNEXPECTED_ERROR(rv);
	}

	/* if rv is zero then VIO DR operations are possible */
	return (rv ? false : true);
}

/*
 * This function is called by the VIO devices (vdisk, vnet, vsw, etc)
 * once the new device has been created. It downloads the new MD and
 * reconfigures the domains and then notifies the 'dr-vio' domain
 * service in the guest and tells it to configure the device.
 */
int
vio_dr_add(ldom_t *ldp, char *type, ident_val_t devid, response_t *resp)
{
	int	rv;

	if (type == NULL)
		return (ERR_GENERIC);

	seq_download_mds_do_config(ldp);
	free_all_staged_ldcs(ldp);

	rv = seq_vio_dr_add(ldp, type, devid, resp);

	switch (rv) {
	case DR_OP_SUCC:
		/* Success */
		break;
	case ERR_DR_VIO_NOTSUP:
		rprintf(resp, MSG_ERR_DR_VIO_NOTSUP, ldp->namep);
		break;
	case ERR_DR_VIO_ROLLBACK:
		rprintf(resp, MSG_ERR_VIO_DR_CONFIG_OP_NO_RESP, ldp->namep);
		break;
	case ERR_DR_VIO_NO_RESPONSE:
	case DR_OP_NO_RESPONSE:
		rprintf(resp, MSG_ERR_VIO_DR_CONFIG_OP_NO_RESP, ldp->namep);
		rprintf(resp, MSG_ERR_VIO_DR_CONFIG_UNK_STATE);
		break;
	case ERR_DR_VIO_INVALID_ARG:
		rprintf(resp, MSG_ERR_DR_VIO_INVALID_ARG, rv, ldp->namep);
		break;
	case ERR_DR_VIO_IN_USE:
		rprintf(resp, MSG_ERR_DR_VIO_IN_USE, ldp->namep);
		break;
	case ERR_DR_VIO_FAILED:
		/* Message displayed from dr_vio code. */
		break;
	case DR_OP_FAILED:
		rprintf(resp, MSG_ERR_VIO_DR_CONFIG_OP_FAILED, ldp->namep);
		break;
	default:
		rprintf(resp, MSG_ERR_VIO_UNKNOWN_FAILURE, rv, ldp->namep);
		break;
	}

	return (rv);
}

/*
 * Removal of VIO devices is a 2 phase process
 * Phase 1)
 *	notification is sent to the guest domain to unconfigure the
 *	device if it is not in use.
 * Phase 2)
 *	If Phase 1 was successful, the device is removed from the MD(s)
 */
int
vio_dr_rm_phase1(ldom_t *ldp, char *type, ident_val_t devid, response_t *resp)
{
	int	rv;

	if (type == NULL)
		return (ERR_GENERIC);

	rv = seq_vio_dr_rm(ldp, type, devid, resp);

	switch (rv) {
	case DR_OP_SUCC:
		/* Success */
		break;
	case ERR_DR_VIO_NOTSUP:
		rprintf(resp, MSG_ERR_DR_VIO_NOTSUP, ldp->namep);
		break;
	case ERR_DR_VIO_ROLLBACK:
		rprintf(resp, MSG_ERR_VIO_DR_CONFIG_OP_NO_RESP, ldp->namep);
		break;
	case ERR_DR_VIO_NO_RESPONSE:
	case DR_OP_NO_RESPONSE:
		rprintf(resp, MSG_ERR_VIO_DR_CONFIG_OP_NO_RESP, ldp->namep);
		rprintf(resp, MSG_ERR_VIO_DR_CONFIG_UNK_STATE);
		break;
	case ERR_DR_VIO_INVALID_ARG:
		rprintf(resp, MSG_ERR_DR_VIO_INVALID_ARG, rv, ldp->namep);
		break;
	case ERR_DR_VIO_IN_USE:
		rprintf(resp, MSG_ERR_DR_VIO_IN_USE, ldp->namep);
		break;
	case ERR_DR_VIO_FAILED:
		/* Message displayed from dr_vio code. */
		break;
	case DR_OP_FAILED:
		rprintf(resp, MSG_ERR_VIO_DR_CONFIG_OP_FAILED, ldp->namep);
		break;
	default:
		rprintf(resp, MSG_ERR_VIO_UNKNOWN_FAILURE, rv, ldp->namep);
		break;
	}

	return (rv);
}

int
vio_dr_rm_phase2(ldom_t *ldp)
{
	ldp->flag.md_notify = true;

	seq_rm_reconfigure(ldp);
	free_all_staged_ldcs(ldp);

	autosave_update_autosave();

	return (0);
}

int
vid_check_one(response_t *resp, int vid)
{

	if (!VID_IN_RANGE(vid)) {
		rprintf(resp, gettext("vid %d out of range, value must "
		    "be within %d and %d\n"), vid, VID_MIN, VID_MAX);
		return (ERR_NET_VID_INVALID);
	}

	return (0);

}

int
vid_check(response_t *resp, int num_vlans, int *vlans)
{
	int i;

	/*
	 * Check whether the vid values are in range
	 */
	if ((num_vlans < 1) || (vlans == NULL))
		return (0);

	for (i = 0; i < num_vlans; i++) {
		if (vid_check_one(resp, vlans[i]) != 0)
			return (ERR_NET_VID_INVALID);
	}

	return (0);
}

/* XXXMIG - move someplace more general */

extern int vnet_mig_bind(ldom_t *ldp, md_t *mdp, md_t *hvmdp, md_node_t *vdev);
extern int vdisk_mig_bind(ldom_t *ldp, md_t *mdp, md_t *hvmdp, md_node_t *vdev);
extern int ds_mig_bind(ldom_t *ldp, md_t *mdp, md_t *hvmdp, response_t *resp);
extern int vnet_mig_link(ldom_t *ldp, md_t *hvmdp, response_t *resp);

/*
 * XXXMIG:
 * A stubbed version of add_vdevices() that will slowly be expanded
 * as versions of *mig_bind() are implemented. Until that is done
 * for all vio resource types, a *bind_all() will be called as a stub
 * in an effort to keep things sane.
 */
static int
mig_add_vdevices(ldom_t *ldp, md_t *mdp, md_t *hvmdp, int nvdevs,
    md_node_t **vdevs, response_t *resp)
{
	char		*md_name;
	md_node_t	*vdevice;

	for (int i = 0; i < nvdevs; ++i) {

		vdevice = vdevs[i];

		if (md_get_prop_str(mdp, vdevice, "name", &md_name) != 0) {
			DBG(zprintf("vdevice has no name property\n"));
			return (-1);
		}

		if (strcmp(md_name, vdisk_attrs.md_name) == 0) {
			if (vdisk_mig_bind(ldp, mdp, hvmdp, vdevice) != 0) {
				return (-1);
			}
			DBG(zprintf("%s: vdisks bound successfully\n",
			    __func__));
		} else if (strcmp(md_name, vnet_attrs.md_name) == 0) {
			if (vnet_mig_bind(ldp, mdp, hvmdp, vdevice) != 0) {
				return (-1);
			}
			DBG(zprintf("%s: vnets bound successfully\n",
			    __func__));
		}

		/* XXXMIG - currently does not handle VIO servers */
	}


	if (ds_mig_bind(ldp, mdp, hvmdp, resp) != 0)
		return (-1);

	DBG(zprintf("%s: ds channel bound successfully\n", __func__));

	if (console_mig_bind(ldp, mdp, hvmdp, resp) != 0)
		return (-1);

	DBG(zprintf("%s: console bound successfully\n", __func__));

	return (0);
}

static int
do_vio_mig_bind(ldom_t *ldp, md_t *mdp, md_t *hvmd, response_t *resp)
{
	uint64_t	gen = 0;
	md_node_t	**vdevsp;
	int		n_vdevs;

	/* get the MD generation number */
	md_get_prop_val(mdp, md_root_node(mdp), "md-generation#", &gen);
	DBG(zprintf("guest MD generation: %"PRId64"\n", gen));

	/* XXXMIG - does this belong someplace generic? */
	ldp->props.md_generation_num = (gen + 1);

	/* scratch space for searching */
	vdevsp = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	/* navigate to the channel-devices node */
	n_vdevs = md_scan_dag(mdp, NULL, md_find_name(mdp, "channel-devices"),
	    md_find_name(mdp, "fwd"), vdevsp);

	if (n_vdevs != 1) {
		DBG(zprintf("invalid guest MD for migration, more than "
		    "one channel-devices node\n"));
		Xfree(vdevsp);
		return (-1);
	}

	/* read in all the channel devices */
	n_vdevs = md_scan_dag(mdp, vdevsp[0], md_find_name(mdp, VDEV_NAME),
	    md_find_name(mdp, "fwd"), vdevsp);

	if (mig_add_vdevices(ldp, mdp, hvmd, n_vdevs, vdevsp, resp) != 0) {
		DBG(zprintf("unable to add all vdevs specified in the "
		    "source MD\n"));
		Xfree(vdevsp);
		return (-1);
	}

	Xfree(vdevsp);

	return (0);
}

int
vio_mig_restore(ldom_t *ldp, md_t *mdp, md_t *hvmdp, response_t *resp)
{
	return (do_vio_mig_bind(ldp, mdp, hvmdp, resp));
}

int
vio_mig_bind(ldom_t *ldp, md_t *mdp, response_t *resp)
{
	if (do_vio_mig_bind(ldp, mdp, NULL, resp) != 0)
		return (-1);

	return (vnet_mig_link(ldp, NULL, resp));
}

int
vio_mig_add_bindings(ldom_t *ldp, md_t *mdp)
{
	if (vdisk_mig_add_bindings(ldp, mdp) != 0) {
		DBG(zprintf("%s: failed to add vdisk bindings\n", __func__));
		return (-1);
	}

	if (vnet_mig_add_bindings(ldp, mdp) != 0) {
		DBG(zprintf("%s: failed to add vnet bindings\n", __func__));
		return (-1);
	}

	if (console_mig_add_bindings(ldp, mdp) != 0) {
		DBG(zprintf("%s: failed to add console bindings\n", __func__));
		return (-1);
	}

	return (0);
}

void
vio_print_bindings(ldom_t *ldp, response_t *resp)
{
	/* XXXMIG - only ones supported for now */
	vnet_print_bindings(ldp, resp);
	vdisk_print_bindings(ldp, resp);
	ds_print_bindings(ldp, resp);
	console_print_bindings(ldp, resp);
}

/*
 * Ouput vio services (VCC, VSW, VDS, VDPCS, VLDC)
 * by first printing active domains then inactive.
 */
void
print_group_vio(response_t *resp, dispctl_t *dctl)
{
	ldom_t	*ldp;

	/* VCC */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state >= BOUND)
			vcc_list(ldp, resp, dctl);
	}
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state < BOUND)
			vcc_list(ldp, resp, dctl);
	}

	/* VSW */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state >= BOUND)
			vsw_list(ldp, resp, dctl);
	}
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state < BOUND)
			vsw_list(ldp, resp, dctl);
	}

	/* VDS */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state >= BOUND)
			vds_list(ldp, resp, dctl);
	}
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state < BOUND)
			vds_list(ldp, resp, dctl);
	}

	/* VDPCS */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state >= BOUND)
			vdpcs_list(ldp, resp, dctl);
	}
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state < BOUND)
			vdpcs_list(ldp, resp, dctl);
	}

	/* VLDC */
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state >= BOUND)
			vldc_list(ldp, resp, dctl);
	}
	CHAIN_ITER(zeus.ldoms, ldp, link) {
		if (ldp->state < BOUND)
			vldc_list(ldp, resp, dctl);
	}
}
