/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)violdc.c	1.9	09/06/21 SMI"

#include <string.h>

#include "ldmd_types.h"
#include "allocate.h"
#include "fatal.h"
#include "ldcs.h"
#include "vio.h"
#include "ds.h"

static CHAIN(vldc_t, vldc_services) = DEFAULTCHAIN;

struct vldc_attributes vldc_attrs = {
	.name =		"vldc",
	.compatible =	"SUNW,sun4v-channel",
	.md_name =	"virtual-channel",
	.cli_name =	"vldc",
	.device_type =	"serial",
	.port_name =	"vldc-port"
};

struct vldcc_attributes vldcc_attrs = {
	.name =		"vldcc",
	.compatible =	"SUNW,sun4v-channel",
	.md_name =	"virtual-channel-client",
	.cli_name =	"vldcc",
	.device_type =	"serial",
	.port_name =	"vldc-port"
};

struct hvctl_attributes hvctl_attrs = {
	.name =		"hvctl",
	.compatible =	"SUNW,sun4v-hvctl",
	.md_name =	"hvctl",
	.device_type =	"serial",
	.port_name =	"vldc-port"
};

static int
vldc_devid(vldc_t *vldc)
{
	return (vldc->cfg_handle->id);
}

static res_pool_t *
vldc_cfg_handles(ldom_t *ldp)
{
#if VDEV_CFG_HANDLE
	return (&ldp->free_cfg_handles);
#else
	return (&ldp->bound.vldc_cfg_handles);
#endif
}

/*
 * Find the vldc server with a given instance.
 */
static vldc_t *
vldc_find_by_instance(ldom_t *ldp, int instance)
{
	vldc_t *vldc;

	CHAIN_ITER(ldp->vldcs, vldc, ldom) {
		if (vldc_devid(vldc) == instance)
			break;
	}

	return (vldc);
}

/*
 * Find the vldc port using a given guest LDC channel.
 */
static vldc_port_t *
vldc_port_find_by_guest_ldcid(ldom_t *ldp, uint64_t ldcid)
{
	vldc_t *vldc;
	vldc_port_t *vport;

	CHAIN_ITER(ldp->vldcs, vldc, ldom) {
		CHAIN_ITER(vldc->vports, vport, link) {
			if ((vport->type != VLDC_TYPE_HVCTL) &&
			    (ldcs_search(&vport->ldcs, ldcid) != NULL))
				return (vport);
		}
	}

	return (NULL);
}

/*
 * Find the ds port using a given guest LDC channel.
 */
static ds_port_t *
ds_port_find_by_guest_ldcid(ldom_t *ldp, uint64_t ldcid)
{
	ds_port_t *ds_port;

	CHAIN_ITER(ldp->ds_ports, ds_port, ldom) {
		if (ldcs_search(&ds_port->ldcs, ldcid) != NULL)
			return (ds_port);
	}

	return (NULL);
}

static vldc_port_t *
vldc_add_port(vldc_t *vldc, uint64_t port, char *svc_name, vldc_type_t type)
{
	vldc_port_t *vport;

	vport = Xmalloc(sizeof (vldc_port_t));

	vport->type = type;
	/* the peer gets filled in when finalizing vio */
	if (type == VLDC_TYPE_DS)
		vport->peer.ds = NULL;
	else if (type == VLDC_TYPE_HVCTL)
		vport->peer.hvctl = NULL;
	else if (type == VLDC_TYPE_SP)
		vport->peer.sp_ldc = NULL;
	else
		vport->peer.vldcc = NULL;

	if (port == NEW_INSTANCE)
		vport->port = alloc_res(&vldc->free_ports);
	else
		vport->port = pull_res_id(&vldc->free_ports, port);

	if (vport->port == NULL)
		fatal("Unable to allocate port %s on vldc %s",
		    svc_name, vldc->service);

	vport->svc_name = Xstrdup(svc_name);
	vport->vldc = vldc;
	CHAIN_INIT(vport->ldcs.ldcs);

	CHAIN_ADD(vldc->vports, vport, link);

	return (vport);
}

/*
 * Copy the ldcs from one ldcs_t to another
 */
static void
cp_ldcs(ldcs_t *src, ldcs_t *dest)
{
	ldc_chan_t *ldc;

	CHAIN_ITER(src->ldcs, ldc, link) {
		CHAIN_ADD(dest->ldcs, ldc, link);
	}
}

static void
destroy_vldc_port(vldc_port_t *vldc_port, bool_t keep_ldcs)
{
	ASSERT(vldc_port->type == VLDC_TYPE_DS);

	CHAIN_UNLINK(vldc_port->vldc->vports, vldc_port, link);

	/*
	 * When converting existing vldc ports to libds port, the
	 * ldcs are not destroyed
	 */
	if (keep_ldcs == false)
		ldcs_free(&vldc_port->ldcs);

	free_res(vldc_port->port);
	Xfree(vldc_port->svc_name);
	Xfree(vldc_port);
}

static void
unbind_vldc_port(vldc_port_t *vldc_port)
{
	ASSERT(vldc_port->type == VLDC_TYPE_DS);

	ldcs_stage(&vldc_port->ldcs);
	destroy_vldc_port(vldc_port, false);
}

static vldc_t *
vldc_new(ldom_t *ldp, char *name, int cfg_handle)
{
	vldc_t *vldc;

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(ldp->vldcs) == 0)
		create_res_pool(&ldp->bound.vldc_cfg_handles,
		    IDENT_MIN, IDENT_MAX, Res_sorted);
#endif
	vldc = Xmalloc(sizeof (vldc_t));

	vldc->ldp = ldp;
	vldc->service = Xstrdup(name);
	vldc->cfg_handle = pull_res_id(vldc_cfg_handles(ldp), cfg_handle);
	if (vldc->cfg_handle == NULL)
		fatal("Could not get vldc cfg_handle %d", cfg_handle);

	create_res_pool(&vldc->free_ports, 0, IDENT_MAX, Res_sorted);
	CHAIN_INIT(vldc->vports);

	CHAIN_ADD(ldp->vldcs, vldc, ldom);
	CHAIN_ADD(vldc_services, vldc, link);

	return (vldc);
}

/*
 * Find the control domain vlds node if present
 */
static md_node_t *
find_vlds_node(md_t *md)
{
	int i, num;
	md_node_t **nodes, *ch_devs;
	md_node_t *vlds_node = NULL;

	ASSERT(md_node_count(md) > 0);
	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(md));

	num = md_scan_dag(md, NULL, md_find_name(md,
	    CHANNELS_NAME), md_find_name(md, "fwd"), nodes);
	ASSERT(num == 1);
	ch_devs = nodes[0];

	/*
	 * See if there is the vlds node underneath CHANNELS_NAME
	 */
	num = md_scan_dag(md, ch_devs, md_find_name(md,
	    VDEV_NAME), md_find_name(md, "fwd"), nodes);

	for (i = 0; i < num; i++) {
		char *md_name;
		if (md_get_prop_str(md, nodes[i], "name", &md_name) != 0)
			fatal("vdevice has no name property");

		if (strcmp(md_name, VLDS_MD_NAME) == 0) {
			vlds_node = nodes[i];
			break;
		}
	}
	Xfree(nodes);

	return (vlds_node);
}

/*
 * We are given a pointer to the MD node of the "vdevice".
 * We have to parse the node and add a vldc server to zeus.
 */
void
vldc_add_from_mdnode(ldom_t *ldp, md_t *guest_md, md_node_t *vdev)
{
	uint64_t cfg_handle;
	int i, num_client_bindings;
	md_node_t **nodelistp;
	char *vldc_name;
	vldc_t *vldc, *hvctl_vldc;
	int num_dsports = 0;

	/* find out instance # */
	if (md_get_prop_val(guest_md, vdev, "cfg-handle", &cfg_handle) != 0)
		fatal("vdevice has no cfg-handle property");

	/*
	 * Check if server already created. This may be the case when adding
	 * from a MD node in the PRI.
	 */
	vldc = vldc_find_by_instance(ldp, cfg_handle);
	if (vldc == NULL) {
		/*
		 * Add a vldc server with the default name.
		 */
		vldc_name = zsprintf("%s-%s%" PRId64,
		    ldp->namep, vldc_attrs.name, cfg_handle);
		vldc = vldc_new(ldp, vldc_name, cfg_handle);
		Xfree(vldc_name);
	}

	/*
	 * Scan through the vport list and add a client binding for
	 * each one.
	 */
	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(guest_md));

	num_client_bindings = md_scan_dag(guest_md, vdev,
	    md_find_name(guest_md, VDEV_PORT),
	    md_find_name(guest_md, "fwd"), nodelistp);

	for (i = 0; i < num_client_bindings; ++i) {
		uint64_t port;
		char *name;
		char *svc_name;
		vldc_type_t vldc_type;

		md_node_t *vport = nodelistp[i];

		if (md_get_prop_str(guest_md, vport, "name", &name) != 0)
			fatal("vport has no name property");
		if (md_get_prop_val(guest_md, vport, "id", &port) != 0)
			fatal("vport has no id property");
		if (md_get_prop_str(guest_md, vport, VLDC_SVC_NAME_PROP,
		    &svc_name) != 0)
			fatal("vport has no %s property", VLDC_SVC_NAME_PROP);
		/*
		 * Determine type of service from vldc-svc-name property.
		 */
		if (strcmp(svc_name, VLDC_HVCTL_SVC_NAME) == 0) {
			vldc_type = VLDC_TYPE_HVCTL;
		} else if (strcmp(svc_name, VLDC_FMACTL_SVC_NAME) == 0) {
			vldc_type = VLDC_TYPE_FMA;
		} else if (strncmp(svc_name, VLDC_DS_SVC_PREFIX,
		    strlen(VLDC_DS_SVC_PREFIX)) == 0 ||
		    strcmp(svc_name, CD_VLDC_DEV_NAME) == 0) {
			vldc_type = VLDC_TYPE_DS;
			num_dsports++;
		} else {
			vldc_type = VLDC_TYPE_SP;
		}

		int num_ldc_ids;
		uint64_t *ldc_ids;
		vldc_port_t *vldc_port;

		vldc_port = vldc_add_port(vldc, port, svc_name, vldc_type);
		get_ldc_ids(guest_md, vport, &num_ldc_ids, &ldc_ids);
		ldc_pull_ids(ldp, &ldp->guest_free_ldcs, num_ldc_ids,
		    ldc_ids, &vldc_port->ldcs);
		Xfree(ldc_ids);

		if (vldc_type == VLDC_TYPE_HVCTL) {
			/* this is the control domain */
			CONTROL_DOMAIN = ldp;
			hvctl_vldc = vldc;
		} else if (vldc_type == VLDC_TYPE_DS) {
			/*
			 * This is the vldc server which provides domain
			 * services.
			 */
			zeus.ds_server = vldc;
		}
	}

	Xfree(nodelistp);

	if (zeus.ds_server == NULL)
		zeus.ds_server = hvctl_vldc;

	/*
	 * To determine whether libds can be used, only the control domain
	 * vldc node is relevant
	 */
	if (IS_CONTROL_DOMAIN(ldp) == false) {
		return;
	}

	if (find_vlds_node(guest_md) == NULL) {
		libds_cap.config = false; /* no vlds node, no libds */
	} else {
		/*
		 * vlds node is in present config, but this does not gurantee
		 * that libds is operational (if the boot config didn't have
		 * vlds node in primary).
		 */
		libds_cap.config = true;
	}

	/*
	 * If multiple vldc ports servicing domain service to guest LDoms is
	 * detected, we disable libds and use vldc instead.
	 */
	if (num_dsports > 1) {
		DBG(zprintf("%s: %d vldc-ports connected to DS are detected, "
		    "libds will not be used for DS communication\n",
		    __func__, num_dsports));

		/*
		 * libds isn't enabled but if support is available, flag
		 * this for conversion to libds ports
		 */
		libds_cap.config = false; /* not a libds port config */
		libds_cap.convert_libds = libds_cap.supported;
	} else {
		libds_cap.convert_libds = (libds_cap.supported == true &&
		    libds_cap.config == false) ? true : false;
	}
}

static void
gen_vldc_guest_endpoint(vldc_port_t *vldc_port, md_t *hv_md,
    md_node_t *hv_guest, md_node_t *ldc_endpoints)
{
	ldcs_t *target_ldcs;
	uint64_t target_guest;

	if (vldc_port->type == VLDC_TYPE_DS) {
		ds_port_t *ds = vldc_port->peer.ds;

		target_ldcs = &ds->ldcs;
		target_guest = ds->ldp->idp->id;
	} else {
		vldcc_t *vldcc = vldc_port->peer.vldcc;

		target_ldcs = &vldcc->vport->ldcs;
		target_guest = vldcc->ldp->idp->id;
	}

	vio_gen_ldcs_endpoint_nodes(hv_md, hv_guest, ldc_endpoints, TYPE_GUEST,
	    TYPE_GUEST, &vldc_port->ldcs, target_guest, target_ldcs, NULL);
}

static void
gen_vldc_hvctl_endpoints(vldc_port_t *vldc_port, md_t *hv_md,
    md_node_t *hv_guest, md_node_t *ldc_endpoints)
{
	md_node_t *ldc_endpoint;

	vio_gen_ldc_endpoint_node(hv_md, hv_guest, ldc_endpoints, TYPE_GUEST,
	    TYPE_HV, CHAIN_START(vldc_port->ldcs.ldcs), -1,
	    &vldc_port->peer.hvctl->ldc);

	ldc_endpoint = vio_gen_ldc_endpoint_node(hv_md, NULL, ldc_endpoints,
	    TYPE_HV, TYPE_GUEST, &vldc_port->peer.hvctl->ldc,
	    vldc_port->vldc->ldp->idp->id, CHAIN_START(vldc_port->ldcs.ldcs));

	md_add_value_property(hv_md, ldc_endpoint, "svc_id", LDC_HVCTL_SVC);
}

void
vldc_gen_ldc_endpoints(ldom_t *ldp, md_t *hv_md, md_node_t *hv_guest,
    md_node_t *ldc_endpoints)
{
	vldc_t *vldc;
	vldc_port_t *vldc_port;

	CHAIN_ITER(ldp->vldcs, vldc, ldom) {
		CHAIN_ITER(vldc->vports, vldc_port, link) {
			/*
			 * Generate LDC endpoint according to port type.
			 */
			if (vldc_port->type == VLDC_TYPE_SP) {
				vio_gen_ldc_endpoint_node(hv_md, hv_guest,
				    ldc_endpoints, TYPE_GUEST, TYPE_SP,
				    CHAIN_START(vldc_port->ldcs.ldcs), -1,
				    vldc_port->peer.sp_ldc->channel);
			} else if (vldc_port->type == VLDC_TYPE_HVCTL) {
				gen_vldc_hvctl_endpoints(vldc_port, hv_md,
				    hv_guest, ldc_endpoints);
			} else {
				gen_vldc_guest_endpoint(vldc_port, hv_md,
				    hv_guest, ldc_endpoints);
			}
		}
	}
}

static void
gen_vldc_guest_md_one(vldc_t *vldc, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vldc_port_t *vldc_port;
	md_node_t *vdev;

	vdev = md_link_new_node(md, VDEV_NAME, cdevs, "fwd", "back");

	/*
	 * Compatible property is an array of strings, though we only
	 * currently have one.
	 */
	md_add_string_as_data_property(md, vdev, "compatible",
	    vldc_attrs.compatible);
	md_add_string_property(md, vdev, "device-type", vldc_attrs.device_type);
	md_add_string_property(md, vdev, "name", vldc_attrs.md_name);
	md_add_value_property(md, vdev, "cfg-handle", vldc_devid(vldc));

	CHAIN_ITER(vldc->vports, vldc_port, link) {
		md_node_t *port_node;

		port_node = vio_gen_vdev_port_node(md, vdev, chan_endpoints,
		    vldc_attrs.port_name, vldc_port->port->id,
		    &vldc_port->ldcs);
		md_add_string_property(md, port_node, "vldc-svc-name",
		    vldc_port->svc_name);
	}
}

void
vldc_gen_guest_md(ldom_t *ldp, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vldc_t *vldc;

	CHAIN_ITER(ldp->vldcs, vldc, ldom) {
		gen_vldc_guest_md_one(vldc, md, cdevs, chan_endpoints);
	}
}

/*
 * Called when reconstructing state from MDs.
 * Links all vldc ports to the SP with the corresponding SP LDC.
 */
int
link_vldc_sp_ports(ldom_t *ldp)
{
	vldc_t *vldc;
	vldc_port_t *vldc_port;
	ldc_epdb_t *ep;
	ldc_chan_t *ldc;
	sp_guest_ldc_t *sp_ldc;

	CHAIN_ITER(ldp->vldcs, vldc, ldom) {
		CHAIN_ITER(vldc->vports, vldc_port, link) {
			if (vldc_port->type != VLDC_TYPE_SP)
				continue;

			ldc = CHAIN_START(vldc_port->ldcs.ldcs);
			ep = ldc_epdb_find(ldp->idp->id, ldc->chan->id);
			if (ep != NULL)
				sp_ldc = sp_guest_ldc_find_by_id(ep->peer->ldc);
			if ((ep == NULL) || (sp_ldc == NULL))
				fatal("Could not find peer SP LDC for vldc "
				    "LDC %d", (int)ldc->chan->id);
			vldc_port->peer.sp_ldc = sp_ldc;
		}
	}

	return (0);
}

static void
print_vldc_header(vldc_t *vldc, char *servertype, int nclients,
    response_t *resp, dispctl_t *dctl)
{
	if (dctl->header && !dctl->parseable) {
		dctl->header = 0;
		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in tables with fixed
		 * column widths.
		 * WIDTH  HEADER	MEANING
		 * 16	  "NAME"	vldc server name
		 * 16	  "LDOM"	the LDom name iff list-services
		 */
		rprintf(resp, "    "FNAME, gettext("NAME"));
		if (dctl->services)
			rprintf(resp, " "FNAME, gettext("LDOM"));
		rprintf(resp, "\n");
	}

	print_server_header(vldc->service, servertype, nclients, resp, dctl);
	if (dctl->services) {
		if (dctl->parseable)
			rprintf(resp, "|ldom=%s",  vldc->ldp->namep);
		else
			rprintf(resp, " "FNAME, vldc->ldp->namep);
	}
	rprintf(resp, "\n");
}

static void
vldc_list_one(vldc_t *vldc, response_t *resp, dispctl_t *dctl)
{
	vldc_port_t *vldc_port;
	char *client, *client_name, *client_ldom, *desc;

	int nclients = CHAIN_LENGTH(vldc->vports);
	print_vldc_header(vldc, vldc_attrs.name, nclients, resp, dctl);

	if (!dctl->bindings || nclients == 0)
		return;

	dctl->header = 1;
	if (!dctl->parseable)
		rprintf(resp, "        "FBNDNAME" "FDESC" %s\n",
		    gettext("CLIENT"), gettext("DESC"), gettext("LDC"));

	CHAIN_ITER(vldc->vports, vldc_port, link) {
		switch (vldc_port->type) {

		case VLDC_TYPE_HVCTL:
			client = zsprintf("%s@%s", vldc_port->svc_name,
			    vldc->ldp->namep);
			if (dctl->parseable)
				rprintf(resp, "|client=%s|desc=hcvtl|ldc=",
				    client);
			else
				rprintf(resp, "        "FBNDNAME" "FDESC" ",
				    client, "hvctl");
			ldcs_print(&vldc_port->ldcs, resp);
			rprintf(resp, "\n");
			Xfree(client);
			break;

		case VLDC_TYPE_SP:
			if (dctl->parseable)
				rprintf(resp, "|client=SP|desc=%s",
				    vldc_port->svc_name);
			else
				rprintf(resp, "        "FBNDNAME" "FDESC, "SP",
				    vldc_port->svc_name);
			if (dctl->parseable)
				rprintf(resp, "|ldc=");
			else
				rprintf(resp, " ");
			ldcs_print(&vldc_port->ldcs, resp);
			rprintf(resp, "\n");
			break;

		case VLDC_TYPE_DS:
		case VLDC_TYPE_FMA:
			if (vldc_port->type == VLDC_TYPE_DS) {
				client_name = "ds";
				client_ldom = vldc_port->peer.ds->ldp->namep;
				desc = "domain-services";
			} else {
				vldcc_t *vldcc = vldc_port->peer.vldcc;
				client_name = vldcc->name;
				client_ldom = vldcc->ldp->namep;
				desc = "fma";
			}

			client = zsprintf("%s@%s", client_name, client_ldom);

			if (dctl->parseable)
				rprintf(resp, "|client=%s|desc=%s",
				    client, desc);
			else
				rprintf(resp, "        "FBNDNAME" "FDESC,
				    client, desc);
			if (dctl->parseable)
				rprintf(resp, "|ldc=");
			else
				rprintf(resp, " ");
			ldcs_print(&vldc_port->ldcs, resp);
			rprintf(resp, "\n");
			Xfree(client);
			break;
		default:
			break;
		}
	}
}

void
vldc_list(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	vldc_t *vldc;

	if (dctl->oformat && !dctl->oserial)
		return;

	if (!dctl->extended || (CHAIN_LENGTH(ldp->vldcs) == 0))
		return;

	dctl->header = 1;

	if (!dctl->parseable)
		rprintf(resp, "%s\n", dctl_label(vldc_attrs.name, dctl));

	CHAIN_ITER(ldp->vldcs, vldc, ldom) {
		vldc_list_one(vldc, resp, dctl);
	}

	if (!dctl->parseable)
		rprintf(resp, "\n");
}

/*
 * VLDC clients.
 */

int
vldcc_devid(vldcc_t *vldcc)
{
	return (vldcc->cfg_handle->id);
}

static res_pool_t *
vldcc_cfg_handles(ldom_t *ldp)
{
#if VDEV_CFG_HANDLE
	return (&ldp->free_cfg_handles);
#else
	return (&ldp->bound.vldcc_cfg_handles);
#endif
}

vldcc_t *
vldcc_new(ldom_t *ldp, int cfg_handle)
{
	vldcc_t *vldcc;

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(ldp->vldccs) == 0)
		create_res_pool(&ldp->bound.vldcc_cfg_handles,
		    IDENT_MIN, IDENT_MAX, Res_sorted);
#endif
	vldcc = Xmalloc(sizeof (vldcc_t));

	vldcc->cfg_handle = pull_res_id(vldcc_cfg_handles(ldp), cfg_handle);
	if (vldcc->cfg_handle == NULL)
		fatal("Could not get vldcc cfg_handle %d", cfg_handle);
	vldcc->name = zsprintf("vldcc%d", cfg_handle);

	vldcc->ldp = ldp;
	vldcc->vport = NULL;

	CHAIN_ADD(ldp->vldccs, vldcc, ldom);

	return (vldcc);
}

ds_port_t *
ds_port_new(ldom_t *ldp, int id)
{
	ds_port_t *dsport = Xcalloc(1, ds_port_t);

	if (id == NEW_DS_PORT)
		dsport->port = alloc_res(&ldp->free_ds_ports);
	else
		dsport->port = pull_res_id(&ldp->free_ds_ports, id);

	if (dsport->port == NULL)
		fatal("Could not allocate LDom %s domain service port id %d",
		    ldp->namep, id);

	dsport->ldp = ldp;
	dsport->work.port = NEW_DS_PORT;
	CHAIN_INIT(dsport->ldcs.ldcs);

	return (dsport);
}

/* ARGSUSED1 */
vldcc_port_t *
add_vldcc_port(vldcc_t *vldcc, uint64_t port, char *svc_name, vldc_type_t type)
{
	vldcc_port_t *vport;

	vport = Xmalloc(sizeof (vldcc_port_t));

	vport->vldcc = vldcc;
	vport->type = type;
	vport->svc_name = Xstrdup(svc_name);
	CHAIN_INIT(vport->ldcs.ldcs);
	if (type == VLDC_TYPE_SP)
		vport->peer.vldc_port = NULL;
	else
		vport->peer.sp_ldc = NULL;

	vldcc->vport = vport;

	return (vport);
}

/*
 * We are given a pointer to the MD node of the "vdevice".  We have to parse
 * the node and add a vldc client to zeus.  Domain Service port and hvctl
 * ports are not parsed here since they belong at different places in the MD.
 * Here we only parse two types of clients: SP and FMA.
 */
void
vldcc_add_from_mdnode(ldom_t *ldp, md_t *guest_md, md_node_t *vdev)
{
	uint64_t cfg_handle, port;
	md_node_t **nodelistp;
	md_node_t *vldc_portnode_p;
	char *name, *svc_name;
	vldc_type_t client_type;
	int num_vports, num_ldc_ids;
	uint64_t *ldc_ids;
	vldcc_t *vldcc;
	vldcc_port_t *vldcc_port;

	/* find out instance # */
	if (md_get_prop_val(guest_md, vdev, "cfg-handle", &cfg_handle) != 0)
		fatal("vdevice has no cfg-handle property");

	/* get port info */
	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(guest_md));
	num_vports = md_scan_dag(guest_md, vdev,
	    md_find_name(guest_md, VDEV_PORT),
	    md_find_name(guest_md, "fwd"), nodelistp);

	if (num_vports != 1)
		fatal("vldc client virtual device has %d ports and should have "
		    "exactly one", num_vports);

	vldc_portnode_p = nodelistp[0];

	if (md_get_prop_str(guest_md, vldc_portnode_p, "name", &name) != 0)
		fatal("vldcc vport has no name property");
	if (md_get_prop_val(guest_md, vldc_portnode_p, "id", &port) != 0)
		fatal("vldcc vport has no id property");

	/*
	 * Get client port name, the service port name in the requirement
	 * structure will be filled up when the server ports are stitched with
	 * the vldcc clients.
	 */
	if (md_get_prop_str(guest_md, vldc_portnode_p, VLDC_SVC_NAME_PROP,
	    &svc_name) != 0)
		fatal("vldc client port has no %s property",
		    VLDC_SVC_NAME_PROP);
	/*
	 * It's either the special case FMA loopback channel "ldmfma",
	 * or a channel to the SP.
	 */
	if (strcmp(svc_name, VLDCC_FMA_SVC_NAME) == 0)
		client_type = VLDC_TYPE_FMA;
	else
		client_type = VLDC_TYPE_SP;

	vldcc = vldcc_new(ldp, cfg_handle);
	vldcc_port = add_vldcc_port(vldcc, port, svc_name, client_type);

	get_ldc_ids(guest_md, vldc_portnode_p, &num_ldc_ids, &ldc_ids);
	ldc_pull_ids(ldp, &ldp->guest_free_ldcs, num_ldc_ids, ldc_ids,
	    &vldcc_port->ldcs);
	Xfree(ldc_ids);
	Xfree(nodelistp);
}

static vldc_port_t *
find_hvctl_vldc_port(ldom_t *ldp)
{
	vldc_t *vldc;
	vldc_port_t *vport;

	CHAIN_ITER(ldp->vldcs, vldc, ldom) {
		CHAIN_ITER(vldc->vports, vport, link) {
			if (vport->type == VLDC_TYPE_HVCTL)
				return (vport);
		}
	}

	return (NULL);
}

/*
 * Add hvctl endpoint information to the hvctl vldc port.
 */
void
add_hvctl_from_md(md_t *hv_md, ldom_t *ldp)
{
	vldc_port_t *vport;
	md_node_t **nodes;
	md_node_t *hvldc_node = NULL;
	int num;

	if (ldp == NULL) {
		/*
		 * The control domain may be unknown when running
		 * a backend other than the HV backend.
		 */
		if (zeus.simulating) {
			warning("No control domain is defined");
			return;
		} else {
			fatal("No control domain is defined");
		}
	}

	vport = find_hvctl_vldc_port(ldp);
	if (vport == NULL)
		fatal("Could not find the HVCTL vldc port");

	/* find the HV LDC endpoint for the HVCTL channel */
	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(hv_md));
	num = md_scan_dag(hv_md, NULL, md_find_name(hv_md, "ldc_endpoint"),
	    md_find_name(hv_md, "fwd"), nodes);

	for (int i = 0; i < num; i++) {
		uint64_t svc_id;

		if (md_get_prop_val(hv_md, nodes[i], "svc_id",
		    &svc_id) != 0)
			continue;

		if (svc_id == LDC_HVCTL_SVC) {
			/* this is the HV LDC endpoint for the HVCTL channel */
			hvldc_node = nodes[i];
			break;
		}
	}

	Xfree(nodes);

	if (hvldc_node == NULL)
		fatal("Could not find the HVCTL channel HV LDC endpoint");

	uint64_t ldc_id, res_id;

	if (md_get_prop_val(hv_md, hvldc_node, "channel", &ldc_id) != 0)
		fatal("ldc_endpoint missing channel property");
	if (md_get_prop_val(hv_md, hvldc_node, "resource_id", &res_id) != 0)
		fatal("ldc_endpoint missing resource_id property");

	vport->peer.hvctl = Xmalloc(sizeof (hvctl_port_t));
	vport->peer.hvctl->ldc.chan = pull_res_id(&zeus.free_hv_ldc_ids,
	    ldc_id);
	vport->peer.hvctl->ldc.id = pull_res_id(&zeus.free_ldc_ids, res_id);
	if ((vport->peer.hvctl->ldc.chan == NULL) ||
	    (vport->peer.hvctl->ldc.id == NULL))
		fatal("HVCTL channel HV LDC endpoint resources unavailable");
}

void
vldcc_gen_ldc_endpoints(ldom_t *ldp, md_t *hv_md, md_node_t *hv_guest,
    md_node_t *ldc_endpoints)
{
	vldcc_t *vldcc;

	CHAIN_ITER(ldp->vldccs, vldcc, ldom) {
		vldcc_port_t *vldcc_port = vldcc->vport;

		if (vldcc_port->type == VLDC_TYPE_SP) {
			vio_gen_ldc_endpoint_node(hv_md, hv_guest,
			    ldc_endpoints, TYPE_GUEST, TYPE_SP,
			    CHAIN_START(vldcc_port->ldcs.ldcs), -1,
			    vldcc_port->peer.sp_ldc->channel);
		} else {
			vldc_port_t *vldc_port = vldcc_port->peer.vldc_port;

			vio_gen_ldcs_endpoint_nodes(hv_md, hv_guest,
			    ldc_endpoints, TYPE_GUEST, TYPE_GUEST,
			    &vldcc_port->ldcs, vldc_port->vldc->ldp->idp->id,
			    &vldc_port->ldcs, NULL);
		}
	}
}

static void
gen_vldcc_guest_md_one(vldcc_t *vldcc, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vldcc_port_t *vport = vldcc->vport;

	if ((vport->type == VLDC_TYPE_SP) || (vport->type == VLDC_TYPE_FMA)) {
		md_node_t *vdev, *port;

		vdev = md_link_new_node(md, VDEV_NAME, cdevs, "fwd", "back");

		/*
		 * Compatible property is an array of strings, though we only
		 * currently have one.
		 */
		md_add_string_as_data_property(md, vdev, "compatible",
		    vldcc_attrs.compatible);

		md_add_string_property(md, vdev, "device-type",
		    vldcc_attrs.device_type);
		md_add_string_property(md, vdev, "name", vldcc_attrs.md_name);
		md_add_value_property(md, vdev, "cfg-handle",
		    vldcc_devid(vldcc));

		port = vio_gen_vdev_port_node(md, vdev, chan_endpoints,
		    vldcc_attrs.port_name, 0, &vport->ldcs);

		md_add_string_property(md, port, VLDC_SVC_NAME_PROP,
		    vport->svc_name);
	}
}

void
vldcc_gen_guest_md(ldom_t *ldp, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vldcc_t *vldcc;

	CHAIN_ITER(ldp->vldccs, vldcc, ldom) {
		gen_vldcc_guest_md_one(vldcc, md, cdevs, chan_endpoints);
	}
}

/*
 * Called when reconstructing state from MDs.
 * Using the LDC endpoint DB links up all vldcc clients with their
 * vldc port or SP LDC.
 */
int
link_vldcc_clients(ldom_t *ldp)
{
	vldcc_t *vldcc;
	ldc_epdb_t *ep;
	ldc_chan_t *ldc;
	vldc_port_t *vldc_port;
	sp_guest_ldc_t *sp_ldc;

	CHAIN_ITER(ldp->vldccs, vldcc, ldom) {
		ldc = CHAIN_START(vldcc->vport->ldcs.ldcs);
		if ((ep = ldc_epdb_find(ldp->idp->id, ldc->chan->id)) == NULL)
			fatal("Could not find vldcc LDC %d",
			    (int)ldc->chan->id);
		if (vldcc->vport->type == VLDC_TYPE_SP) {
			/* SP type hence must link to a SP LDC endpoint */
			sp_ldc = sp_guest_ldc_find_by_id(ep->peer->ldc);
			if (sp_ldc == NULL)
				fatal("Could not find peer SP LDC for vldcc "
				    "LDC %d", (int)ldc->chan->id);
			vldcc->vport->peer.sp_ldc = sp_ldc;
		} else {
			/* not SP type hence must link to guest LDC endpoint */
			vldc_port = vldc_port_find_by_guest_ldcid(ep->peer->ldp,
			    ep->peer->ldc);
			if (vldc_port == NULL)
				fatal("Could not find vldc port for vldcc "
				    "LDC %d", (int)ep->peer->ldc);
			vldc_port->peer.vldcc = vldcc;
			vldcc->vport->peer.vldc_port = vldc_port;
		}
	}

	return (0);
}

static void
vldcc_ds_header(response_t *resp, dispctl_t *dctl)
{
	/*
	 * TRANSLATION_NOTE
	 * The following are column headers in tables with fixed column widths.
	 * WIDTH  HEADER	MEANING
	 * 16	  "NAME"	name of client
	 * 27	  "SERVICE"	name of the service the client is using
	 * 17	  "DESC"	description of VLDC channel service
	 * 3	  "LDC"		Logical Domain Channel number
	 */

	if (dctl->header) {
		dctl->header = 0;
		if (dctl->bindings)
			rprintf(resp, "    "FNAME" "FBNDNAME" "FDESC" %s\n",
			    gettext("NAME"), gettext("SERVICE"),
			    gettext("DESC"), gettext("LDC"));
		else
			rprintf(resp, "    "FNAME" "FBNDNAME" %s\n",
			    gettext("NAME"), gettext("SERVICE"),
			    gettext("DESC"));
	}
}

static void
ds_list_one(ds_port_t *dsport, response_t *resp, dispctl_t *dctl)
{
	if (dctl->parseable) {
		rprintf(resp, "%s", dctl_label(vldcc_attrs.cli_name, dctl));
		rprintf(resp, "|name=ds");
	} else {
		vldcc_ds_header(resp, dctl);
		rprintf(resp, "    "FNAME, "ds");
	}

	vldc_port_t *vldc_port = dsport->vldc_port;
	char *desc = "domain-services";
	char *service;

	if (vldc_port == NULL) {
		if (zeus.ds_server != NULL)
			service = zsprintf("%s@%s", zeus.ds_server->service,
			    zeus.ds_server->ldp->namep);
		else
			service = Xstrdup("");
	} else {
		service = zsprintf("%s@%s", vldc_port->vldc->service,
		    vldc_port->vldc->ldp->namep);
	}

	if (dctl->parseable)
		rprintf(resp, "|service=%s|desc=%s", service, desc);
	else
		rprintf(resp, " "FBNDNAME" "FDESC, service, desc);
	Xfree(service);

	if (dctl->bindings) {
		if (dctl->parseable)
			rprintf(resp, "|ldc=");
		else
			rprintf(resp, " ");
		ldcs_print(&dsport->ldcs, resp);
	}

	rprintf(resp, "\n");
}

static void
vldcc_list_one(vldcc_t *vldcc, response_t *resp, dispctl_t *dctl)
{
	if (dctl->parseable) {
		rprintf(resp, "%s", dctl_label(vldcc_attrs.cli_name, dctl));
		rprintf(resp, "|name=%s", vldcc->name);
	} else {
		vldcc_ds_header(resp, dctl);
		rprintf(resp, "    "FNAME, vldcc->name);
	}

	if (vldcc->vport->type == VLDC_TYPE_SP) {
		if (dctl->parseable)
			rprintf(resp, "|service=SP|desc=%s",
			    vldcc->vport->svc_name);
		else
			rprintf(resp, " "FBNDNAME" "FDESC, "SP",
			    vldcc->vport->svc_name);
	} else {
		vldc_port_t *vldc_port = vldcc->vport->peer.vldc_port;

		if (dctl->parseable) {
			rprintf(resp, "|service=%s@%s|desc=%s",
			    vldc_port->vldc->service,
			    vldc_port->vldc->ldp->namep,
			    vldcc->vport->svc_name);
		} else {
			char *service = zsprintf("%s@%s",
			    vldc_port->vldc->service,
			    vldc_port->vldc->ldp->namep);
			rprintf(resp, " "FBNDNAME" "FDESC, service,
			    vldcc->vport->svc_name);
			Xfree(service);
		}
	}

	if (dctl->bindings) {
		if (dctl->parseable)
			rprintf(resp, "|ldc=");
		else
			rprintf(resp, " ");
		ldcs_print(&vldcc->vport->ldcs, resp);
	}

	rprintf(resp, "\n");
}

void
vldcc_list(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	vldcc_t *vldcc;
	ds_port_t *ds_port;

	if (dctl->oformat && !dctl->oserial)
		return;

	if (!dctl->extended ||
	    ((CHAIN_LENGTH(ldp->vldccs) == 0) &&
	    (CHAIN_LENGTH(ldp->ds_ports) == 0)))
		return;

	dctl->header = 1;
	if (!dctl->parseable)
		rprintf(resp, "%s\n", dctl_label(vldcc_attrs.name, dctl));

	CHAIN_ITER(ldp->ds_ports, ds_port, ldom) {
		ds_list_one(ds_port, resp, dctl);
	}

	CHAIN_ITER(ldp->vldccs, vldcc, ldom) {
		vldcc_list_one(vldcc, resp, dctl);
	}

	if (!dctl->parseable)
		rprintf(resp, "\n");
}

/*
 * Miscellaneous.
 */

/*
 * Add a hvctl client for testing purposes.
 */
static void
add_hvctl(ldom_t *ldp)
{
	vldc_t *vldc;
	vldc_port_t *vldc_port;

	ASSERT(zeus.simulating);

	/*
	 * Add a vldc server for the hvctl client. This will also serve as
	 * the domain services server.
	 */
	vldc = vldc_new(ldp, ldp->namep, VLDC_INSTANCE);
	zeus.ds_server = vldc;

	/* add the hvctl client and bind it */
	vldc_port = vldc_add_port(vldc, NEW_INSTANCE, VLDC_HVCTL_SVC_NAME,
	    VLDC_TYPE_HVCTL);
	if (ldcs_chan_alloc(&vldc_port->ldcs, &ldp->guest_free_ldcs) != 0)
		fatal("Could not allocate a LDC for hvctl");

	vldc_port->peer.hvctl = Xmalloc(sizeof (hvctl_port_t));
	vldc_port->peer.hvctl->ldc.chan = alloc_res(&zeus.free_hv_ldc_ids);
	vldc_port->peer.hvctl->ldc.id = alloc_res(&zeus.free_ldc_ids);

	/* mark this LDom as the control domain */
	CONTROL_DOMAIN = ldp;
}

static void
add_domain_services(ldom_t *ldp)
{
	ds_port_t *ds_port;

	ds_port = ds_port_new(ldp, NEW_DS_PORT);
	CHAIN_ADD(ldp->ds_ports, ds_port, ldom);
}

/*
 * Set up vldc clients for a new LDom.
 *
 * If there is currently no control domain (only the case when not running on
 * hardware) the new LDom is set up to be the control domain by adding a hvctl
 * channel.
 */
void
new_ldom_vldc(ldom_t *ldp, response_t *resp)
{
	if (CONTROL_DOMAIN == NULL) {
		add_hvctl(ldp);
		if (resp != NULL) {
			rprintf(resp, "Set LDom %s to be the control domain\n",
			    ldp->namep);
		}
	}

	add_domain_services(ldp);
}

/*
 * After reading the state from the HV, if the control domain is set up to
 * use vldc at the other end of it's domain service, migrate this connection
 * to a loopback ds <--> ds connection, and also convert the connection
 * to other domains to exclusive ds to ds connection.
 * 'ldp' is zeus.hvctl_ldom_p
 */
void
conv_vldc_ports_to_ds(ldom_t *ldp)
{
	vldc_t		*vldc = zeus.ds_server;
	vldc_port_t	*vldc_port, *nport;
	ds_port_t *ds_port, *peer_ds;

	/*
	 * Check if libds is supported in control domain
	 */
	if (libds_cap.convert_libds == false) {
		return;
	}

	DBG(zprintf("%s: Converting vldc-port to vlds-port config\n",
	    __func__));

	ASSERT(IS_CONTROL_DOMAIN(ldp) == true);

	/*
	 * Go thru each vldc port and check if it's of type DS.
	 * Inherit the ldcs from the vldc port to a new DS port.
	 * Store the ds port at the remote end of the ds port
	 * For loopback channel to control domain, leave the
	 * vldc port untouched.
	 */
	CHAIN_ITER_VOL(vldc->vports, vldc_port, nport, link) {
		if (vldc_port->type != VLDC_TYPE_DS ||
		    vldc_port->peer.ds == NULL ||
		    vldc_port->peer.ds->ldp == ldp)
			continue;
		peer_ds = vldc_port->peer.ds;
		ds_port = ds_port_new(ldp, NEW_DS_PORT);
		CHAIN_ADD(ldp->ds_ports, ds_port, ldom);
		ds_port->work.port = vldc_port->port->id;
		cp_ldcs(&vldc_port->ldcs, &ds_port->ldcs);
		destroy_vldc_port(vldc_port, true);
		peer_ds->vldc_port = NULL;
		peer_ds->peer_ds_port = ds_port;
		ds_port->peer_ds_port = peer_ds;
	}
}

/*
 * Once the vldc ports are converted to ds and an MD is generated
 * revert back to vldc
 */
void
conv_ds_ports_to_vldc(ldom_t *ldp)
{
	ldom_t		*rldp;
	vldc_port_t	*vldc_port;
	ds_port_t *local_ds_port, *peer_ds_port, *ndsp;
	char *svc_name;
	int64_t port = NEW_DS_PORT;

	/*
	 * If the reverse operation didn't happen, ignore
	 */
	if (libds_cap.convert_libds == false) {
		return;
	}

	DBG(zprintf("%s: Converting vlds-port to vldc-port config\n",
	    __func__));

	ASSERT(IS_CONTROL_DOMAIN(ldp) == true);

	/*
	 * Traverse through the contol domain ds ports and convert
	 * each to a vldc port
	 */
	CHAIN_ITER_VOL(ldp->ds_ports, local_ds_port, ndsp, ldom) {
		if (local_ds_port->vldc_port != NULL ||
		    local_ds_port->peer_ds_port == NULL)
			continue;
		peer_ds_port = local_ds_port->peer_ds_port;

		ASSERT(peer_ds_port != NULL);

		rldp = peer_ds_port->ldp; /* remote ldom */
		svc_name = zsprintf("%s%s", VLDC_DS_SVC_PREFIX, rldp->namep);
		port = local_ds_port->work.port;
		ASSERT(port != NEW_DS_PORT); /* Must have been saved earlier */
		vldc_port = vldc_add_port(zeus.ds_server, port, svc_name,
		    VLDC_TYPE_DS);
		Xfree(svc_name);
		cp_ldcs(&local_ds_port->ldcs, &vldc_port->ldcs);

		/*
		 * destroy the local_ds_port after linking
		 */
		peer_ds_port->vldc_port = vldc_port;
		peer_ds_port->peer_ds_port = NULL;
		vldc_port->peer.ds = peer_ds_port;
		CHAIN_UNLINK(ldp->ds_ports, local_ds_port, ldom);
		free_res(local_ds_port->port);
		Xfree(local_ds_port);
	}
}

/*
 * Called when reconstructing state from MDs.
 * Using the LDC endpoint DB links up the domain service client in a LDom
 * to the vldc port on the domain services vldc server.
 */
void
link_domain_services(ldom_t *ldp)
{
	ldc_epdb_t *ep;
	ldc_chan_t *ldc;
	vldc_port_t *vldc_port;
	ds_port_t *ds_port, *peer_ds_port;

	CHAIN_ITER(ldp->ds_ports, ds_port, ldom) {
		ldc = CHAIN_START(ds_port->ldcs.ldcs);
		if ((ep = ldc_epdb_find(ldp->idp->id, ldc->chan->id)) != NULL) {
			vldc_port = vldc_port_find_by_guest_ldcid(ep->peer->ldp,
			    ep->peer->ldc);
			if (vldc_port) {
				/*
				 * Linked to control domain vldc port
				 */
				vldc_port->peer.ds = ds_port;
				ds_port->vldc_port = vldc_port;
				ASSERT(vldc_port->vldc == zeus.ds_server);
				continue;
			}

			/*
			 * Check if linked to a DS port instead
			 */
			peer_ds_port = ds_port_find_by_guest_ldcid
			    (ep->peer->ldp, ep->peer->ldc);
			if (peer_ds_port) {
				peer_ds_port->peer_ds_port = ds_port;
				ds_port->peer_ds_port = peer_ds_port;
				if (ldp != CONTROL_DOMAIN)
					ASSERT(peer_ds_port->ldp ==
					    CONTROL_DOMAIN);
				continue;
			}
		}
		fatal("Could not find matching port for domain "
		    "services on LDom %s", ldp->namep);
	}
}

void
ds_gen_guest_md(ldom_t *ldp, md_t *md, md_node_t *chan_endpoints)
{
	md_node_t **nodes;
	md_node_t *dsnode;
	md_node_t *vlds_node = NULL;
	int num;
	ds_port_t *ds_port, *peer_ds_port;

	if (CHAIN_LENGTH(ldp->ds_ports) == 0)
		return;

	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(md));

	/*
	 * Check if a "domain-services" node exists. If not, create
	 * one directly under the root node.
	 */
	num = md_scan_dag(md, NULL, md_find_name(md, DS_NODE_NAME),
	    md_find_name(md, "fwd"), nodes);
	if (num == 0) {
		dsnode = md_link_new_node(md, DS_NODE_NAME,
		    md_root_node(md), "fwd", "back");
	} else {
		dsnode = nodes[0];
	}

	/*
	 * Add SP domain service port if any.
	 */
	if (ldp->ds_sp_md != NULL) {
		md_t		*dsmd = ldp->ds_sp_md;
		md_node_t	*port;
		md_node_t	*endpoint;
		md_string_t	*mdstrp;

		/* merge in the SP domain services MD */
		md_add(md, dsmd);

		/* lookup port info in the SP domain services MD */
		mdstrp = md_find_name(dsmd, DS_PORT_NODE_NAME);
		num = md_scan_nodes(dsmd, mdstrp, nodes);
		port = nodes[0];
		port = md_find_node_index(md, port->build_index);

		/* lookup endpoint info in the SP domain services MD */
		mdstrp = md_find_name(dsmd, CHANNEL_ENDPOINT);
		num = md_scan_nodes(dsmd, mdstrp, nodes);
		endpoint = nodes[0];
		endpoint = md_find_node_index(md, endpoint->build_index);

		/* link port node to the domain services node */
		md_add_arc_property(md, dsnode, "fwd", port);
		md_add_arc_property(md, port, "back", dsnode);

		/* link endpoint node to the channel endpoints node */
		mdstrp = md_find_name(md, CHANNEL_ENDPOINTS);
		num = md_scan_dag(md, NULL, mdstrp, md_find_name(md, "fwd"),
		    nodes);
		md_add_arc_property(md, nodes[0], "fwd", endpoint);
	}

	/*
	 * For each domain-services port, create the port nodes
	 */
	CHAIN_ITER(ldp->ds_ports, ds_port, ldom) {
		md_node_t	*dsport_node;
		vldc_port_t	*vldc_port = ds_port->vldc_port;

		dsport_node = md_link_new_node(md, DS_PORT_NODE_NAME, dsnode,
		    "fwd", "back");

		md_add_value_property(md, dsport_node, "id", ds_port->port->id);

		/* add the channel endpoints to the port node */
		gen_guest_ldc_nodes(md, dsport_node, chan_endpoints,
		    &ds_port->ldcs);

		/*
		 * Create the vlds properties in each DS port responsible for
		 * guest domain services. Also, create ONE instance of vlds
		 * node in every guest and create one vlds-port node
		 * underneath the vlds node for each domain service port.
		 * The control domain would always have a vldc port connected
		 * to its domain services port and a vlds node is generated
		 * for this as well
		 */
		if ((vldc_port && vldc_port->type == VLDC_TYPE_DS &&
		    (peer_ds_port = vldc_port->peer.ds) != NULL) ||
		    ((peer_ds_port = ds_port->peer_ds_port) != NULL)) {
			md_node_t *ch_devs, *vlds_port, *chan;

			md_add_string_property(md, dsport_node,
			    VLDS_REM_DOM_NAME_PROP, peer_ds_port->ldp->namep);
			md_add_value_property(md, dsport_node,
			    VLDS_REM_DOM_HDL_PROP, peer_ds_port->ldp->idp->id);

			/*
			 * Create the virtual device & port info for the vlds
			 * driver.  Connect the port to the same channel just
			 * created above.
			 */
			num = md_scan_dag(md, NULL, md_find_name(md,
			    CHANNELS_NAME), md_find_name(md, "fwd"), nodes);

			ASSERT(num == 1);
			ch_devs = nodes[0];

			/*
			 * Create one vlds node
			 */
			if (vlds_node == NULL) {
				/* virtual device node  */
				vlds_node = md_link_new_node(md,
				    VDEV_NAME, ch_devs, "fwd", "back");
				md_add_string_property(md, vlds_node, "name",
				    VLDS_MD_NAME);
				md_add_value_property(md, vlds_node,
				    VLDS_DOM_HANDLE,
				    ds_port->ldp->idp->id);
				md_add_string_property(md, vlds_node,
				    VLDS_DOM_NAME_PROP, ldp->namep);
				md_add_string_property(md, vlds_node,
				    "device-type", "serial");
				md_add_string_property(md, vlds_node,
				    "compatible", VLDS_COMPAT);
				md_add_value_property(md, vlds_node,
				    "cfg-handle", 0);
			}

			/* virtual device port node */
			vlds_port = md_link_new_node(md, VDEV_PORT,
			    vlds_node, "fwd", "back");
			md_add_string_property(md, vlds_port, "name",
			    VLDS_PORT_NAME);
			md_add_string_property(md, vlds_port,
			    VLDS_REM_DOM_NAME_PROP, peer_ds_port->ldp->namep);
			md_add_value_property(md, vlds_port,
			    VLDS_REM_DOM_HDL_PROP, peer_ds_port->ldp->idp->id);

			md_add_value_property(md, vlds_port, "id",
			    ds_port->port->id);

			/* Link to channel endpoint */
			if (md_get_prop_arc(md, dsport_node, "fwd", &chan) < 0)
				fatal("dsport_node missing channel-endpoint");

			/* link the endpoint node with the port node */
			md_add_arc_property(md, vlds_port, "fwd", chan);
			md_add_arc_property(md, chan, "back", vlds_port);

		}
	}
	Xfree(nodes);
}

/*
 * Link the guest DS port to a vldc port in control domain
 */
static int
ds_link_vldc(ldom_t *ldp, response_t *resp)
{
	int		rc;
	vldc_port_t	*vldc_port;
	char		*svc_name;
	ds_port_t *ds_port = CHAIN_START(ldp->ds_ports);

	ASSERT(CHAIN_LENGTH(ldp->ds_ports) <= 1);

	DBG(zprintf("%s: LDom: %s\n", __func__, ldp->namep));

	svc_name = zsprintf("%s%s", VLDC_DS_SVC_PREFIX, ldp->namep);
	vldc_port = vldc_add_port(zeus.ds_server, NEW_INSTANCE,
	    svc_name, VLDC_TYPE_DS);
	Xfree(svc_name);

	rc = ldcs_chan_alloc(&vldc_port->ldcs,
	    &vldc_port->vldc->ldp->guest_free_ldcs);
	if (rc != 0) {
		rprintf(resp, gettext("Insufficient free LDC channels on "
		    "LDom %s to bind the domain services\n"),
		    vldc_port->vldc->ldp->namep);
		ldcs_free(&ds_port->ldcs);
		destroy_vldc_port(vldc_port, false);

		return (rc);
	}

	ds_port->vldc_port = vldc_port;
	vldc_port->peer.ds = ds_port;

	if (vldc_port->vldc->ldp != ldp) {
		vldc_port->vldc->ldp->flag.gen_new_md = true;
		vldc_port->vldc->ldp->flag.md_notify = true;
	}

	return (0);
}

/*
 * Link the domain service client on the guest domain to a DS port
 * on the control domain
 */
static int
ds_link_libds(ldom_t *ldp, response_t *resp)
{
	int		rc;
	ds_port_t *ds_port = CHAIN_START(ldp->ds_ports);
	ds_port_t *peer_ds_port;

	ASSERT(CHAIN_LENGTH(ldp->ds_ports) <= 1);

	peer_ds_port = ds_port_new(CONTROL_DOMAIN, NEW_DS_PORT);

	DBG(zprintf("%s: LDom: %s\n", __func__, ldp->namep));

	/*
	 * allocate the channels in the control domain
	 */
	rc = ldcs_chan_alloc(&peer_ds_port->ldcs,
	    &(CONTROL_DOMAIN)->guest_free_ldcs);

	if (rc != 0) {
		rprintf(resp, gettext("Insufficient free LDC channels "
		    "on LDom %s to bind userland domain services\n"),
		    (CONTROL_DOMAIN)->namep);
		ldcs_free(&ds_port->ldcs);
		ds_delete(CONTROL_DOMAIN, peer_ds_port);
		return (rc);
	}

	CHAIN_ADD((CONTROL_DOMAIN)->ds_ports, peer_ds_port, ldom);
	ds_port->peer_ds_port = peer_ds_port;
	peer_ds_port->peer_ds_port = ds_port;

	if (!IS_CONTROL_DOMAIN(ldp)) {
		(CONTROL_DOMAIN)->flag.gen_new_md = true;
		(CONTROL_DOMAIN)->flag.md_notify = true;
	}

	return (0);
}

static int
ds_mig_link_vldc(ldom_t *ldp, md_t *hvmdp, response_t *resp)
{
	int		rc;
	vldc_port_t	*vldc_port;
	char		*svc_name;
	ds_port_t *ds_port;

	svc_name = zsprintf("%s%s", VLDC_DS_SVC_PREFIX, ldp->namep);
	vldc_port = vldc_add_port(zeus.ds_server, NEW_INSTANCE,
	    svc_name, VLDC_TYPE_DS);
	Xfree(svc_name);

	ds_port = CHAIN_START(ldp->ds_ports);

	ASSERT(CHAIN_LENGTH(ldp->ds_ports) <= 1);

	rc = ldc_mig_get_and_pull_peer_ids(ldp, hvmdp, &ds_port->ldcs,
	    &vldc_port->ldcs, &vldc_port->vldc->ldp->guest_free_ldcs);

	if (rc != 0) {
		rprintf(resp, gettext("Insufficient free LDC channels on "
		    "LDom %s to bind the domain services\n"),
		    vldc_port->vldc->ldp->namep);
		ldcs_free(&ds_port->ldcs);
		destroy_vldc_port(vldc_port, false);

		return (rc);
	}

	ds_port->vldc_port = vldc_port;
	vldc_port->peer.ds = ds_port;

	if (vldc_port->vldc->ldp != ldp) {
		vldc_port->vldc->ldp->flag.gen_new_md = true;
		vldc_port->vldc->ldp->flag.md_notify = true;
	}

	return (0);
}

static int
ds_mig_link_libds(ldom_t *ldp, md_t *hvmdp, response_t *resp)
{
	int		rc;
	ds_port_t *ds_port, *peer_ds_port;

	ds_port = CHAIN_START(ldp->ds_ports);
	peer_ds_port = ds_port_new(CONTROL_DOMAIN, NEW_DS_PORT);

	ASSERT(CHAIN_LENGTH(ldp->ds_ports) <= 1);

	rc = ldc_mig_get_and_pull_peer_ids(ldp, hvmdp, &ds_port->ldcs,
	    &peer_ds_port->ldcs, &peer_ds_port->ldp->guest_free_ldcs);

	if (rc != 0) {
		rprintf(resp, gettext("Insufficient free LDC channels on "
		    "LDom %s to bind the domain services\n"),
		    peer_ds_port->ldp->namep);
		ldcs_free(&ds_port->ldcs);
		ds_delete(CONTROL_DOMAIN, peer_ds_port);

		return (rc);
	}

	CHAIN_ADD((CONTROL_DOMAIN)->ds_ports, peer_ds_port, ldom);
	ds_port->peer_ds_port = peer_ds_port;
	peer_ds_port->peer_ds_port = ds_port;

	if (peer_ds_port->ldp != ldp) {
		peer_ds_port->ldp->flag.gen_new_md = true;
		peer_ds_port->ldp->flag.md_notify = true;
	}

	return (0);
}

int
ds_mig_bind(ldom_t *ldp, md_t *mdp, md_t *hvmdp, response_t *resp)
{
	int		i, num;
	md_node_t	**nodes;
	int		nports;

	/* domain should be created by now */
	ASSERT(CHAIN_LENGTH(ldp->ds_ports) != 0);

	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	num = md_scan_dag(mdp, NULL, md_find_name(mdp, DS_NODE_NAME),
	    md_find_name(mdp, "fwd"), nodes);

	if (num != 1) {
		DBG(zprintf("LDom %s MD does not have exactly one %s "
		    "node\n", ldp->namep, DS_NODE_NAME));
		Xfree(nodes);
		return (-1);
	}

	nports = md_scan_dag(mdp, nodes[0],
	    md_find_name(mdp, DS_PORT_NODE_NAME),
	    md_find_name(mdp, "fwd"), nodes);

	if (nports == 0) {
		DBG(zprintf("LDom %s MD does not have any %s "
		    "node\n", ldp->namep, DS_PORT_NODE_NAME));
		Xfree(nodes);
		return (-1);
	}

	/*
	 * Right now, there is a pretty big assumption that all domains
	 * will have only one domain service port, with the exception
	 * of the control domain. For example, ds_bind() makes this
	 * assumption. Continue with this assumption here, but this may
	 * need to change in a new domain service view of the world where
	 * domains can talk to each other.
	 */

	for (i = 0; i < nports; i++) {
		int rc;
		md_node_t *ds_port_node = nodes[i];
		ds_port_t *ds_port = CHAIN_START(ldp->ds_ports);

		ldc_mig_get_and_pull_ids(ldp, hvmdp, &ldp->guest_free_ldcs,
		    mdp, ds_port_node, &ds_port->ldcs);

		rc = (libds_cap.functional) ?
		    ds_mig_link_libds(ldp, hvmdp, resp) :
		    ds_mig_link_vldc(ldp, hvmdp, resp);

		if (rc != 0) {
			DBG(zprintf("unable to link DS port to its peer\n"));
			fini_response(resp);
			Xfree(nodes);
			return (-1);
		}
	}

	Xfree(nodes);
	return (0);
}

/*
 * Called when binding a domain.
 * Creates a vldc port on the domain services server for the domain
 * and links the domains' domain service with it.
 */
int
ds_bind(ldom_t *ldp, response_t *resp)
{
	int rc;
	ds_port_t *ds_port;

	if (CHAIN_LENGTH(ldp->ds_ports) == 0)
		return (0);

	ds_port = CHAIN_START(ldp->ds_ports);

	rc = ldcs_chan_alloc(&ds_port->ldcs, &ldp->guest_free_ldcs);

	if (rc != 0) {
		rprintf(resp, gettext("Insufficient free LDC channels on "
		    "LDom %s to bind the domain services\n"), ldp->namep);
		return (rc);
	}

	/*
	 * Depending on whether libds exists or not, link to a DS port or a
	 * vldc port
	 */
	/*
	 * If vldc is being used, link to a vldc port, otherwise link
	 * to a ds port instead
	 */
	rc = (libds_cap.functional) ?
	    ds_link_libds(ldp, resp) : ds_link_vldc(ldp, resp);

	return (rc);
}

/*
 * Unbind a domain's domain services from the domain services server.
 * Called when unbinding a domain.
 */
void
ds_unbind(ldom_t *ldp)
{
	ds_port_t *ds_port;

	CHAIN_ITER(ldp->ds_ports, ds_port, ldom) {
		if (ds_port->vldc_port != NULL) {
			vldc_port_t *vldc_port = ds_port->vldc_port;

			if (vldc_port->vldc->ldp != ldp) {
				vldc_port->vldc->ldp->flag.gen_new_md = true;
				vldc_port->vldc->ldp->flag.md_notify = true;
			}

			ldcs_stage(&ds_port->ldcs);
			unbind_vldc_port(vldc_port);
			ds_port->vldc_port = NULL;
		} else if (ds_port->peer_ds_port != NULL) {
			ds_port_t *peer_ds_port = ds_port->peer_ds_port;

			/*
			 * Assume MD update flags already set by virtue
			 * of vldc port handled just above.
			 */
			ldcs_stage(&ds_port->ldcs);
			ldcs_free(&peer_ds_port->ldcs);
			ds_delete(CONTROL_DOMAIN, peer_ds_port);
			ds_port->peer_ds_port = NULL;
		}
	}
}

/*
 * Free up resources used by domain's domain service.
 * Called when deleting a domain.
 */
void
ds_delete(ldom_t *ldp, ds_port_t *ds_port)
{
	ASSERT(ds_port->vldc_port == NULL); /* not bound */

	free_res(ds_port->port);
	ldcs_free(&ds_port->ldcs);
	CHAIN_UNLINK(ldp->ds_ports, ds_port, ldom);
	Xfree(ds_port);
}

void
ds_bind_check(ldom_t *ldp)
{
	ldp->work.num_ldcs++;
	zeus.ds_server->ldp->work.num_ldcs++;
}

void
ds_print_bindings(ldom_t *ldp, response_t *resp)
{
	char		idstr[UINT64_STR_LEN];
	ldc_chan_t	*ldcp;
	ds_port_t	*ds_port = CHAIN_START(ldp->ds_ports);

	(void) snprintf(idstr, UINT64_STR_LEN, "ds%d", ds_port->port->id);
	rprintf(resp, "  %-11s", idstr);

	CHAIN_ITER(ds_port->ldcs.ldcs, ldcp, link) {
		rprintf(resp, "%2"PRId64" ", ldcp->chan->id);
	}

	rprintf(resp, "\n");
}
