/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <string.h>
#include <alloca.h>

#include "ldmd_types.h"
#include "allocate.h"
#include "chain.h"
#include "database.h"
#include "database_internal.h"
#include "errors.h"
#include "fatal.h"
#include "ldom.h"
#include "resource_id.h"
#include "vio.h"
#include "zprintf.h"

struct vnet_attributes vnet_attrs = {
	.name =		"network",
	.compatible =	"SUNW,sun4v-network",
	.md_name =	"network",
	.cli_name =	"vnet",
	.device_type =	"network",
	.port_name =	"vnet-port"
};

ident_val_t
vnet_devid(vnet_t *vnet)
{
	if (vnet->binding == NULL)
		return (INVALID_DEVID);
	else
		return (vnet->binding->cfg_handle->id);
}

static res_pool_t *
vnet_cfg_handles(ldom_t *ldp)
{
#if VDEV_CFG_HANDLE
	return (&ldp->free_cfg_handles);
#else
	return (&ldp->bound.vnet_cfg_handles);
#endif
}

vnet_t *
vnet_find_by_name(ldom_t *ldp, char *vnet_name)
{
	vnet_t *vnet;

	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		if (strcmp(vnet_name, vnet->name) == 0)
			break;
	}

	return (vnet);
}

vnet_t *
vnet_new(ldom_t *ldp, char *name, char *service, mac_addr_t mac, int mtu,
    uint64_t phys_state, char *vnet_mode, int pvid, int num_vlans, int *vlans,
    int networkid)
{
	vnet_t *vnet;

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(ldp->vnets) == 0) {
		create_res_pool(&ldp->bound.vnet_cfg_handles,
		    IDENT_MIN, IDENT_MAX, Res_sorted);
		name_ldom_res_pool(&ldp->bound.vnet_cfg_handles, ldp,
		    "vnet cfg handles");
	}
#endif

	vnet = Xcalloc(1, vnet_t);

	vnet->ldp = ldp;
	vnet->name = Xstrdup(name);
	vnet->service = (service != NULL) ? Xstrdup(service) : NULL;
	vnet->mac = mac;
	vnet->mtu = mtu;
	vnet->phys_state = phys_state;
	vnet->vnet_mode = (vnet_mode != NULL) ? Xstrdup(vnet_mode) : NULL;
	vnet->pvid = pvid;
	vnet->num_vlans = (vlans == NULL) ? 0 : num_vlans;
	vnet->vlans = vlans;
	vnet->networkid = networkid;
	vnet->binding = NULL;

	CHAIN_ADD(ldp->vnets, vnet, ldom);
	return (vnet);
}

static bool_t
check_vnet_peer(vnet_t *local_vnet, vnet_t *remote_vnet)
{
	vnet_peer_t	*local_peer;
	vnet_peer_t	*remote_peer;

	if ((local_vnet == NULL) || (local_vnet->binding == NULL))
		return (false);

	CHAIN_ITER(local_vnet->binding->peers, local_peer, link) {
		remote_peer = local_peer->peer;

		if (remote_peer->vnet == remote_vnet)
			return (true);
	}

	return (false);
}

static vnet_peer_t *
add_vnet_peer(vnet_t *vnet, uint64_t port)
{
	vnet_peer_t *peer;

	peer = Xcalloc(1, vnet_peer_t);

	if (port == NEW_INSTANCE)
		peer->port = alloc_res(&vnet->binding->free_ports);
	else
		peer->port = pull_res_id(&vnet->binding->free_ports, port);

	if (peer->port == NULL)
		fatal("Unable to allocate port %"PRIu64" on vnet %s",
		    port, vnet->name);

	peer->vnet = vnet;
	peer->peer = NULL;

	CHAIN_INIT(peer->ldcs.ldcs);

	CHAIN_ADD(vnet->binding->peers, peer, link);

	return (peer);
}

static int
init_vnet_binding(vnet_t *vnet, int cfg_handle)
{
	ldom_t *ldp = vnet->ldp;
	vnet_binding_t *binding;

	binding = Xcalloc(1, vnet_binding_t);

	binding->cfg_handle = vio_get_cfg_handle(cfg_handle,
	    vnet_cfg_handles(ldp));
	if (binding->cfg_handle == NULL) {
		DBG(zprintf("Could not obtain vnet cfg_handle %d on LDom %s\n",
		    cfg_handle, ldp->namep));
		return (-1);
	}

	binding->vsw_port = NULL;

	create_res_pool(&binding->free_ports, VNET_MIN_PORT, IDENT_MAX,
	    Res_sorted);
	name_ldom_vdev_res_pool(&binding->free_ports, ldp, "vnet",
	    vnet->name, "ports");

	CHAIN_INIT(binding->peers);
	CHAIN_INIT(binding->ldcs.ldcs);

	vnet->binding = binding;

	/*
	 * if the value for networkid is NO_NEW_ID, update the networkid with
	 * the cfg_handle id just pulled from the cfg_handle pool
	 */
	if (vnet->networkid == NO_NEW_ID)
		vnet->networkid = vnet->binding->cfg_handle->id;

	return (0);
}

static void
destroy_vnet_peer(vnet_peer_t *vnet_peer)
{
	CHAIN_UNLINK(vnet_peer->vnet->binding->peers, vnet_peer, link);

	ldcs_free(&vnet_peer->ldcs);
	free_res(vnet_peer->port);
	Xfree(vnet_peer);
}

void
teardown_vnet_peer_bindings(vnet_t *vnet, bool_t unbinding)
{
	vnet_peer_t	*local_peer;
	vnet_peer_t	*remote_peer;
	ldom_t		*ldp = vnet->ldp;

	if (vnet->binding == NULL)
		return;

	while ((local_peer = CHAIN_START(vnet->binding->peers)) != NULL) {
		remote_peer = local_peer->peer;

		if (unbinding) {
			ldcs_stage(&local_peer->ldcs);
			ldcs_stage(&remote_peer->ldcs);

			if (remote_peer->vnet->ldp != ldp) {
				remote_peer->vnet->ldp->flag.gen_new_md = true;
				remote_peer->vnet->ldp->flag.md_notify = true;
			}
		}

		destroy_vnet_peer(local_peer);
		if (remote_peer)
			destroy_vnet_peer(remote_peer);
	}
}

static void
teardown_vnet_binding(vnet_t *vnet, bool_t unbinding)
{
	if (vnet->binding == NULL)
		return;

	ldom_t *ldp = vnet->ldp;
	vsw_port_t *vsw_port = vnet->binding->vsw_port;

	if (unbinding) {
		if (vsw_port) {
			vsw_t *vsw = vsw_port->vsw;

			if (vsw->ldp != ldp) {
				vsw->ldp->bound.num_clients--;
				vsw->ldp->flag.gen_new_md = true;
				vsw->ldp->flag.md_notify = true;
			}

			vsw_unbind_port(vsw_port);
		}
		ldcs_stage(&vnet->binding->ldcs);
	} else {
		if (vsw_port)
			vsw_destroy_port(vsw_port);
		ldcs_free(&vnet->binding->ldcs);
	}

	teardown_vnet_peer_bindings(vnet, unbinding);

	free_res(vnet->binding->cfg_handle);
	destroy_res_pool(&vnet->binding->free_ports);
	Xfree(vnet->binding);
	vnet->binding = NULL;
}

static void
destroy_vnet_binding(vnet_t *vnet)
{
	teardown_vnet_binding(vnet, false);
}

static int
vnet_link_peer(vnet_t *vnet, vsw_port_t *vsw_port, response_t *resp)
{
	int		rc;
	ldom_t		*remote_ldp;
	vnet_peer_t	*local_peer;
	vnet_peer_t	*remote_peer;

	remote_ldp = vsw_port->vnet->ldp;

	local_peer = add_vnet_peer(vnet, NEW_INSTANCE);
	rc = ldcs_chan_alloc(&local_peer->ldcs, &vnet->ldp->guest_free_ldcs);
	if (rc != 0) {
		destroy_vnet_binding(vnet);
		rprintf(resp, gettext("Insufficient free LDC channels on LDom "
		    "%s to bind vnet %s\n"), vnet->ldp->namep, vnet->name);
		return (rc);
	}

	remote_peer = add_vnet_peer(vsw_port->vnet, NEW_INSTANCE);
	local_peer->peer = remote_peer;
	remote_peer->peer = local_peer;

	rc = ldcs_chan_alloc(&remote_peer->ldcs, &remote_ldp->guest_free_ldcs);
	if (rc != 0) {
		destroy_vnet_binding(vnet);
		rprintf(resp, gettext("Insufficient free LDC channels on LDom "
		    "%s to bind vnet %s\n"), remote_ldp->namep, vnet->name);
		return (rc);
	}

	if (remote_ldp != vnet->ldp) {
		remote_ldp->flag.gen_new_md = true;
		remote_ldp->flag.md_notify = true;
	}

	return (0);
}

static int
vnet_link_peers(vnet_t *vnet, vsw_t *vsw, response_t *resp)
{
	int		rc;
	ldom_t		*remote_ldp;
	vsw_port_t	*vsw_port;

	/*
	 * For each vnet connected to the vswitch add a port to that vnet
	 * and connect it to a port on the vnet being bound.
	 */
	CHAIN_ITER(vsw->binding->vports, vsw_port, link) {

		remote_ldp = vsw_port->vnet->ldp;

		/* skip the vswitch port to this vnet */
		if (vsw_port == vnet->binding->vsw_port)
			continue;

		/* skip if peers are already linked */
		if (check_vnet_peer(vnet, vsw_port->vnet) == true)
			continue;

		if (remote_ldp->flag.locked) {
			destroy_vnet_binding(vnet);
			rprintf(resp, gettext("Cannot add virtual network %s "
			    "because that would modify locked LDom %s\n"),
			    vnet->name, remote_ldp->namep);
			return (ERR_LDOM_LOCKED);
		}

		if ((rc = vnet_link_peer(vnet, vsw_port, resp)) != 0) {
			destroy_vnet_binding(vnet);
			return (rc);
		}
	}

	return (0);
}

/* ARGSUSED */
static int
mig_link_bound_vnet(ldom_t *ldp, md_t *hvmdp, vnet_t *vnet, vsw_t *vsw,
    response_t *resp)
{
	int	rc;

	DBG(zprintf("%s\n", __func__));

	if ((rc = vnet_link_peers(vnet, vsw, resp)) != 0)
		return (rc);

	if (vsw->ldp != vnet->ldp) {
		vsw->ldp->bound.num_clients++;
		vsw->ldp->flag.gen_new_md = true;
		vsw->ldp->flag.md_notify = true;
	}

	return (0);
}

static int
link_bound_vnet(vnet_t *vnet, vsw_t *vsw, response_t *resp)
{
	int		rc;

	/* add a port on the vsw for this vnet and allocate LDCs */
	vnet->binding->vsw_port = vsw_add_port(vsw, vnet, NEW_INSTANCE,
	    vnet->vnet_mode);

	rc = ldcs_chan_alloc(&vnet->binding->vsw_port->ldcs,
	    &vsw->ldp->guest_free_ldcs);
	if (rc != 0) {
		destroy_vnet_binding(vnet);
		rprintf(resp, gettext("Insufficient free LDC channels on "
		    "LDom %s to bind vnet %s\n"), vsw->ldp->namep,
		    vnet->name);
		return (rc);
	}

	if ((rc = vnet_link_peers(vnet, vsw, resp)) != 0)
		return (rc);

	if (vsw->ldp != vnet->ldp) {
		vsw->ldp->bound.num_clients++;
		vsw->ldp->flag.gen_new_md = true;
		vsw->ldp->flag.md_notify = true;
	}

	return (0);
}

int
vnet_mig_link(ldom_t *ldp, md_t *hvmdp, response_t *resp)
{
	vnet_t	*vnet;
	vsw_t	*vsw;
	int	rc;

	DBG(zprintf("%s\n", __func__));

	CHAIN_ITER(ldp->vnets, vnet, ldom) {

		vsw = vsw_find_by_name(vnet->service);
		if ((vsw == NULL) || (vsw->binding == NULL)) {
			DBG(zprintf("no bound virtual switch service %s\n",
			    vnet->service));
			return (-1);
		}

		/* link this vnet with the vsw and peer vnets */
		if ((rc = mig_link_bound_vnet(ldp, hvmdp, vnet, vsw, resp))
		    != 0) {
			destroy_vnet_binding(vnet);
			return (rc);
		}
	}

	return (0);
}

int
vnet_mig_bind(ldom_t *ldp, md_t *mdp, md_t *hvmdp, md_node_t *vdev)
{
	int		i, rv;
	vnet_t		*vnet;
	vsw_t		*vsw;
	uint64_t	cfg_handle;
	uint64_t	mtu;
	uint64_t	phys_state;
	uint64_t	local_mac;
	uint64_t	pvid = DEFAULT_VLAN_ID;
	uint64_t	*vids = NULL;
	int		*vlans = NULL;
	int		nvlans = 0;
	char		*client_name;
	md_node_t	**nodes;
	int		nbound;
	md_node_t	*vport;
	uint64_t	swport;

	DBG(zprintf("%s\n", __func__));

	/* the name of the device is found in the devalias */
	if (md_get_prop_str(mdp, vdev, "devalias", &client_name) != 0) {
		DBG(zprintf("vdevice has no devalias property\n"));
		return (-1);
	}

	/* lookup the vnet, it should already exist */
	if ((vnet = vnet_find_by_name(ldp, client_name)) == NULL) {
		DBG(zprintf("LDom '%s' has no vnet named '%s'\n", ldp->namep,
		    client_name));
		return (-1);
	}

	/*
	 * Phase 1: Patch up vnet props based on the MD.
	 */

	if (md_get_prop_val(mdp, vdev, "mtu", &mtu) != 0) {
		DBG(zprintf("vdevice has no mtu property\n"));
		return (-1);
	}

	if (md_get_prop_val(mdp, vdev, "linkprop", &phys_state) != 0) {
		DBG(zprintf("vdevice has no linkprop property\n"));
	}

	if (md_get_prop_val(mdp, vdev, "local-mac-address", &local_mac) != 0) {
		DBG(zprintf("vdevice has no local-mac-address property\n"));
		return (-1);
	}

	if (md_get_prop_val(mdp, vdev, PVID_PROP, &pvid) != 0) {
		DBG(zprintf("vdevice has no %s property\n", PVID_PROP));
	}

	if (md_get_prop_data(mdp, vdev, VID_PROP, (uint8_t **)&vids,
	    &nvlans) != 0) {
		DBG(zprintf("vdevice has no %s property\n", VID_PROP));
	}

	nvlans /= sizeof (uint64_t);
	if (nvlans) {
		vlans = Xmalloc(nvlans * sizeof (int));
		for (i = 0; i < nvlans; i++)
			vlans[i] = (int)vids[i];
	}

	vnet->mac = local_mac;	/* XXXMIG - need to check mac here */
	vnet->mtu = mtu;
	vnet->phys_state = phys_state;
	vnet->pvid = pvid;
	vnet->num_vlans = nvlans;
	vnet->vlans = vlans;

	/*
	 * Phase 2: Setup the bindings for this vnet.
	 */

	/* get the config handle for this vnet */
	if (md_get_prop_val(mdp, vdev, "cfg-handle", &cfg_handle) != 0) {
		DBG(zprintf("vdevice has no cfg-handle property\n"));
		Xfree(vlans);
		return (-1);
	}

	/* initialize the bindings */
	(void) init_vnet_binding(vnet, cfg_handle);

	nodes = Xmalloc(sizeof (md_node_t *) * md_node_count(mdp));

	nbound = md_scan_dag(mdp, vdev, md_find_name(mdp, VDEV_PORT),
	    md_find_name(mdp, "fwd"), nodes);

	ASSERT(nbound > 0);

	for (i = 0; i < nbound; i++) {

		vport = nodes[i];

		/* XXMIG - should prune out peer vnet ports */
		rv = md_get_prop_val(mdp, vport, VNET_SW_PORT_PROP, &swport);
		if (rv != 0) {
			uint64_t port;
			(void) md_get_prop_val(mdp, vport, "id", &port);
			DBG(zprintf("ignoring peer vnet port %"PRId64"\n",
			    port));
			continue;
		}

		/* get the LDC associated with this port */
		ldc_mig_get_and_pull_ids(ldp, hvmdp, &ldp->guest_free_ldcs,
		    mdp, vport, &vnet->binding->ldcs);
	}

	Xfree(nodes);

	/* add a port on the vsw for this vnet and allocate LDCs */
	vsw = vsw_find_by_name(vnet->service);
	if ((vsw == NULL) || (vsw->binding == NULL)) {
		DBG(zprintf("no bound virtual switch service %s\n",
		    vnet->service));
		return (-1);
	}

	vnet->binding->vsw_port = vsw_add_port(vsw, vnet, NEW_INSTANCE,
	    vnet->vnet_mode);

	rv = ldc_mig_get_and_pull_peer_ids(ldp, hvmdp, &vnet->binding->ldcs,
	    &vnet->binding->vsw_port->ldcs, &vsw->ldp->guest_free_ldcs);

	if (rv != 0) {
		destroy_vnet_binding(vnet);
		DBG(zprintf("Insufficient free LDC channels on LDom %s "
		    "to bind vnet %s\n", vsw->ldp->namep, vnet->name));
		return (rv);
	}

	return (0);
}

/*
 * Add a vnet client from MD Node.
 */
void
vnet_add_from_mdnode(ldom_t *ldp, md_t *guest_md, md_node_t *vdev)
{
	uint64_t cfg_handle;
	uint64_t mtu;
	uint64_t phys_state = 0;
	uint64_t local_mac;
	char *client_name, *devalias;
	int db_mtu = -1;
	uint64_t db_physstate = ILLEGAL_LNKP;
	int i, num_vnet_bindings;
	md_node_t **nodelistp = NULL;
	vnet_t *vnet;
	uint64_t default_vid = DEFAULT_VLAN_ID;
	uint64_t pvid = default_vid;
	uint64_t *vids = NULL;
	int num_vlans = 0;
	int *vlans = NULL;
	int networkid = NO_NEW_ID;

	if (md_get_prop_val(guest_md, vdev, "cfg-handle", &cfg_handle) != 0)
		fatal("vdevice has no cfg-handle property");
	if (md_get_prop_val(guest_md, vdev, MTU_PROP, &mtu) != 0)
		fatal("vnet node missing mtu property");
	md_get_prop_val(guest_md, vdev, LINK_PROP, &phys_state);
	if (md_get_prop_val(guest_md, vdev, "local-mac-address",
	    &local_mac) != 0)
		fatal("vnet node missing mac-address property");
	if (md_get_prop_val(guest_md, vdev, PVID_PROP,
	    &pvid) != 0)
		warning("vnet node missing %s property", PVID_PROP);
	if (md_get_prop_data(guest_md, vdev, VID_PROP, (uint8_t **)&vids,
	    &num_vlans) != 0)
		warning("vnet node missing %s property", VID_PROP);
	else {
		num_vlans /= sizeof (uint64_t);
		vlans = Xcalloc(num_vlans, int);
		for (i = 0; i < num_vlans; i++)
			vlans[i] = (int)vids[i];
	}

	/*
	 * Get the network name from the devalias property if it exists,
	 * otherwise try get it from the database.
	 */
	if (md_get_prop_str(guest_md, vdev, DEVALIAS, &devalias) == 0) {
		client_name = Xstrdup(devalias);
	} else {
		client_name = get_vnet_name_from_db(ldp, cfg_handle);
		if (client_name == NULL)
			client_name = vio_canonical_client_name(ldp,
			    vnet_attrs.name, cfg_handle);
	}


	/*
	 * Create the client. The service name will be determined when the
	 * client is linked with the vswitch. We need to get the mtu value
	 * from the DB because the mtu in the MD may be the vsw's mtu. If
	 * no mtu constraint was given to the vnet, the mtu value in MD comes
	 * from the vsw's mtu, hence we mustn't set the vnet's mtu constraint
	 * to equal vsw's mtu value.
	 */
	if (get_vnet_props_from_db(ldp, client_name, &db_mtu) == true) {
		if (db_mtu != -1)
			mtu = db_mtu;
	}

	if (get_vnet_linkprop_from_db(ldp, client_name, &db_physstate) &&
	    db_physstate != ILLEGAL_LNKP) {
		phys_state = db_physstate;
	}

	/*
	 * Use the cfg_handle value as a networkid if failed to get the
	 * networkid from the DB
	 */
	get_vnet_id_from_db(ldp, client_name, &networkid);
	if (networkid == NO_NEW_ID)
		networkid = cfg_handle;

	vnet = vnet_new(ldp, client_name, NULL, local_mac, mtu, phys_state,
	    NULL, pvid, num_vlans, vlans, networkid);

	/*
	 * If the domain was in the process of being migrated, the
	 * migration bindings MD must be consulted to recreate the
	 * information about the service the vnet was connected to.
	 * However, the actual binding of the vnet does not happen
	 * until the migration recovery later in startup.
	 */
	if (ldp->flag.migration_source) {
		int		i;
		md_node_t	**vdevs;
		int		nvdevs;
		char		*name = NULL;
		char		*service = NULL;
		md_t		*bind_mdp = ldp->mig_state.bind_mdp;

		vdevs = Xcalloc(md_node_count(bind_mdp), md_node_t *);
		nvdevs = md_scan_dag(bind_mdp, NULL,
		    md_find_name(bind_mdp, "virtual-device"),
		    md_find_name(bind_mdp, "fwd"), vdevs);

		for (i = 0; i < nvdevs; i++) {
			md_get_prop_str(bind_mdp, vdevs[i], "name", &name);
			ASSERT(name != NULL);
			if (strcmp(name, "network") != 0)
				continue;

			md_get_prop_str(bind_mdp, vdevs[i], "devalias", &name);
			if (strcmp(name, client_name) == 0)
				break;
		}

		ASSERT(i != nvdevs);

		md_get_prop_str(bind_mdp, vdevs[i], "service-name", &service);
		ASSERT(service);
		vnet->service = Xstrdup(service);

		Xfree(vdevs);
		goto done;
	}

	/*
	 * Now, scan through the vport list and add a vnet peer for
	 * each one.
	 */
	nodelistp = Xcalloc(md_node_count(guest_md), md_node_t *);

	num_vnet_bindings = md_scan_dag(guest_md, vdev,
	    md_find_name(guest_md, VDEV_PORT),
	    md_find_name(guest_md, "fwd"), nodelistp);

	ASSERT(num_vnet_bindings > 0);

	(void) init_vnet_binding(vnet, cfg_handle);

	for (i = 0; i < num_vnet_bindings; ++i) {
		uint64_t port;
		uint64_t switch_port;
		uint64_t *mac_addr;
		int mac_addr_len;

		md_node_t *vport = nodelistp[i];

		if (md_get_prop_val(guest_md, vport, "id", &port) != 0)
			fatal("vnet vport has no id property");
		if (md_get_prop_data(guest_md, vport, "remote-mac-address",
		    (uint8_t **)&mac_addr, &mac_addr_len) != 0)
			fatal("vnet vport has no remote-mac-address property");

		if (md_get_prop_val(guest_md, vport, VNET_SW_PORT_PROP,
		    &switch_port) == 0) {
			/* this is the port to the switch */
			ldc_get_and_pull_ids(ldp, &ldp->guest_free_ldcs,
			    guest_md, vport, &vnet->binding->ldcs);
		} else {
			vnet_peer_t *peer;

			peer = add_vnet_peer(vnet, port);
			ldc_get_and_pull_ids(ldp, &ldp->guest_free_ldcs,
			    guest_md, vport, &peer->ldcs);
		}
	}

done:
	Xfree(client_name);
	Xfree(nodelistp);
}

void
vnet_gen_ldc_endpoints(ldom_t *ldp, md_t *hv_md, md_node_t *hv_guest,
    md_node_t *ldc_endpoints)
{
	vnet_t		*vnet;
	vnet_peer_t	*vnet_peer;
	vsw_port_t	*vsw_port;

	CHAIN_ITER(ldp->vnets, vnet, ldom) {

		if (vnet->binding == NULL)
			continue;

		vsw_port = vnet->binding->vsw_port;

		vio_gen_ldcs_endpoint_nodes(hv_md, hv_guest, ldc_endpoints,
		    TYPE_GUEST, TYPE_GUEST, &vnet->binding->ldcs,
		    vsw_port->vsw->ldp->idp->id, &vsw_port->ldcs, NULL);

		CHAIN_ITER(vnet->binding->peers, vnet_peer, link) {
			vio_gen_ldcs_endpoint_nodes(hv_md, hv_guest,
			    ldc_endpoints, TYPE_GUEST, TYPE_GUEST,
			    &vnet_peer->ldcs,
			    vnet_peer->peer->vnet->ldp->idp->id,
			    &vnet_peer->peer->ldcs, NULL);
		}
	}
}
/*
 * Add pvid and vlan-ids to the vswitch port node
 */
void
add_vlan_md_props(md_t *md, md_node_t *node, int pvid, int num_vlans,
	int *vlans, bool_t is_port)
{
	int i;
	uint64_t *md_vids;
	char *pvid_prop = (is_port == true) ? PORT_PVID_PROP : PVID_PROP;
	char *vid_prop = (is_port == true) ? PORT_VID_PROP : VID_PROP;

	md_add_value_property(md, node, pvid_prop, pvid);
	if (num_vlans > 0) {
		md_vids = alloca(num_vlans * sizeof (uint64_t));
		for (i = 0; i < num_vlans; i++)
			md_vids[i] = vlans[i];
		md_add_data_property(md, node, vid_prop,
		    num_vlans * sizeof (uint64_t), (uint8_t *)md_vids);
	}
}

static void
gen_vnet_guest_md_one(vnet_t *vnet, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	md_node_t *vdev, *vport;
	vnet_peer_t *vnet_peer;
	vsw_t *vsw = vnet->binding->vsw_port->vsw;
	uint64_t mac;
	uint64_t vsw_mtu = (vsw->mtu != VIO_NONEMTU) ? vsw->mtu : VIO_ETHERMTU;
	int mtu = (vnet->mtu != VIO_NONEMTU) ? vnet->mtu : vsw_mtu;

	vdev = md_link_new_node(md, VDEV_NAME, cdevs, "fwd", "back");

	/*
	 * Compatible property is an array of strings, though we only
	 * currently have one.
	 */
	md_add_string_as_data_property(md, vdev, "compatible",
	    vnet_attrs.compatible);

	md_add_string_property(md, vdev, "device-type", vnet_attrs.device_type);
	md_add_string_property(md, vdev, "name", vnet_attrs.md_name);
	md_add_value_property(md, vdev, "cfg-handle", vnet_devid(vnet));
	md_add_value_property(md, vdev, "local-mac-address", vnet->mac);
	md_add_value_property(md, vdev, MTU_PROP, mtu);
	md_add_value_property(md, vdev, LINK_PROP, vnet->phys_state);
	md_add_string_property(md, vdev, DEVALIAS, vnet->name);
	if (vsw->num_ether_types > 0) {
		md_add_data_property(md, vdev, VSW_ETHER_TYPE_PROP,
		    vsw->num_ether_types * sizeof (uint64_t),
		    (uint8_t *)vsw->ether_types);
	}

	/* create vport node for connection to the vswitch */
	vport = vio_gen_vdev_port_node(md, vdev, chan_endpoints,
	    vnet_attrs.port_name, VNET_VSW_PORT, &vnet->binding->ldcs);
	md_add_value_property(md, vport, VNET_SW_PORT_PROP, VNET_VSW_PORT);

	/*
	 * Add remote mac address property. MD spec calls for an array
	 * of mac addresses but we only have one.
	 */
	mac = hton64(vsw->mac);
	md_add_data_property(md, vport, "remote-mac-address",
	    sizeof (mac), (uint8_t *)&mac);

	/*
	 * Add pvid and vlan-ids to the vswitch port node
	 */
	add_vlan_md_props(md, vport, vsw->pvid, vsw->num_vlans,
	    vsw->vlans, true);

	md_add_value_property(md, vdev, DEFAULT_VID_PROP,
	    vsw->default_vid);
	add_vlan_md_props(md, vdev, vnet->pvid, vnet->num_vlans,
	    vnet->vlans, false);

	/* create vport node for all peer vnets */
	CHAIN_ITER(vnet->binding->peers, vnet_peer, link) {
		vnet_t *rvnet = vnet_peer->peer->vnet;
		vport = vio_gen_vdev_port_node(md, vdev, chan_endpoints,
		    vnet_attrs.port_name, vnet_peer->port->id,
		    &vnet_peer->ldcs);
		/*
		 * Add remote mac address property. MD spec calls for an array
		 * of mac addresses but we only have one.
		 */
		mac = hton64(vnet_peer->peer->vnet->mac);
		md_add_data_property(md, vport, "remote-mac-address",
		    sizeof (mac), (uint8_t *)&mac);
		add_vlan_md_props(md, vport, rvnet->pvid, rvnet->num_vlans,
		    rvnet->vlans, true);
	}
}

void
vnet_gen_guest_md(ldom_t *ldp, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vnet_t *vnet;

	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		gen_vnet_guest_md_one(vnet, md, cdevs, chan_endpoints);
	}
}

static void
destroy_vnet(vnet_t *vnet)
{
	ASSERT(vnet->binding == NULL);

	CHAIN_UNLINK(vnet->ldp->vnets, vnet, ldom);

	if (vnet->num_vlans > 0)
		Xfree(vnet->vlans);
	Xfree(vnet->name);
	Xfree(vnet->service);
	Xfree(vnet->vnet_mode);

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(vnet->ldp->vnets) == 0)
		destroy_res_pool(&vnet->ldp->bound.vnet_cfg_handles);
#endif

	Xfree(vnet);
}

void
vnet_delete_all(ldom_t *ldp)
{
	vnet_t *vnet, *vnet_next;

	for (vnet = CHAIN_START(ldp->vnets); vnet != NULL; vnet = vnet_next) {
		vnet_next = CHAIN_NEXT(vnet->ldom);
		destroy_vnet(vnet);
	}
}

static void
unbind_vnet(vnet_t *vnet)
{
	teardown_vnet_binding(vnet, true);
}


#if BIND_CHECK
static int
check_vnet(vnet_t *vnet, response_t *resp)
{
	vsw_t *vsw;
	vsw_port_t *vsw_port;

	vsw = vsw_find_by_name(vnet->service);
	if ((vsw == NULL) ||
	    ((vsw->ldp != vnet->ldp) && (vsw->binding == NULL))) {
		rprintf(resp, gettext("The virtual switch service \"%s\" for "
		    "virtual network \"%s\"\ndoes not exist or is not "
		    "bound.\n"), vnet->service, vnet->name);
		return (ERR_NO_SERVICE);
	}

	/*
	 * One more LDC channel is needed on the switch and on each of the
	 * vnets attached to the switch ports.
	 */
	vsw->ldp->work.num_ldcs++;
	if (vsw->binding != NULL) {
		CHAIN_ITER(vsw->binding->vports, vsw_port, link) {
			vsw_port->vnet->ldp->work.num_ldcs++;
		}
	}

	/*
	 * Locally add a LDC channel for the connection to the switch and
	 * for each of the current switch ports.
	 */
	vnet->ldp->work.num_ldcs++;
	if (vsw->binding != NULL)
		vnet->ldp->work.num_ldcs += CHAIN_LENGTH(vsw->binding->vports);

	return (0);
}

static int
add_vnet_check(vnet_t *vnet, response_t *resp)
{
	ldom_t *ldp;
	int rc;

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		ldp->work.num_ldcs = 0;
	}

	if ((rc = check_vnet(vnet, resp)) != 0)
		return (rc);

	CHAIN_ITER(zeus.ldoms, ldp, link) {
		DBG(zprintf("%s: %s needs %d new channels\n", __func__,
		    ldp->namep, ldp->work.num_ldcs));

		if ((ldp->work.num_ldcs > 0) && ldp->flag.locked) {
			rprintf(resp, gettext("Cannot add virtual network "
			    "\"%s\" because that would modify locked LDom "
			    "%s\n"), vnet->name, ldp->namep);
			return (ERR_LDOM_LOCKED);
		}

		if (available_res_ids(&ldp->guest_free_ldcs) <
		    ldp->work.num_ldcs) {
			rprintf(resp, gettext("Insufficient free LDC "
			    "channels to bind virtual network %s\n"),
			    vnet->name);
			return (ERR_LDC_NO_IDS);
		}
	}

	return (0);
}

int
vnet_bind_check(ldom_t *ldp, response_t *resp)
{
	vnet_t *vnet, *local_vnet;
	int rc;

	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		if ((rc = check_vnet(vnet, resp)) != 0)
			return (rc);
		/*
		 * Add channels for local connections to other vnets which are
		 * still to be bound to the same switch.
		 */
		for (local_vnet = CHAIN_NEXT(vnet->ldom); local_vnet != NULL;
		    local_vnet = CHAIN_NEXT(local_vnet->ldom)) {
			if (strcmp(vnet->service, local_vnet->service) == 0)
				ldp->work.num_ldcs += 2;
		}
	}

	return (0);
}
#endif

static int
bind_vnet(vnet_t *vnet, response_t *resp)
{
	int	rc;
	vsw_t	*vsw;

	vsw = vsw_find_by_name(vnet->service);
	if (vsw == NULL || vsw->binding == NULL) {
		rprintf(resp, gettext("The virtual switch service %s for "
		    "virtual network %s\ndoes not exist or is not "
		    "bound\n"), vnet->service, vnet->name);
		return (ERR_NO_SERVICE);
	}

	if (vsw->ldp->flag.locked) {
		rprintf(resp, gettext("Cannot add virtual network %s "
		    "because that would modify locked LDom %s\n"),
		    vnet->name, vsw->ldp->namep);
		return (ERR_LDOM_LOCKED);
	}

	if (vsw->phys_dev == NULL && vnet->phys_state == 0x1) {
		rprintf(resp, MSG_ERR_BIND_VNET_NO_NET_DEV, vsw->service);
		return (ERR_VSW_LINKPROP_VNET);
	}

	/*
	 * if the networkid is specified with the CLI, the networkid value
	 * will be used to pull from the cfg_handle pool. If it fails
	 * to get the cfg_handle, generate the duplicate id error message.
	 */
	if (vnet->networkid != NO_NEW_ID) {
		if ((rc = init_vnet_binding(vnet, vnet->networkid)) != 0) {
			rprintf(resp, MSG_ERR_ID_EXISTS);
			return (rc);
		}
	} else {
		rc = init_vnet_binding(vnet, NEW_INSTANCE);
		ASSERT(rc == 0);
	}

	rc = ldcs_chan_alloc(&vnet->binding->ldcs, &vnet->ldp->guest_free_ldcs);
	if (rc != 0) {
		destroy_vnet_binding(vnet);
		rprintf(resp, gettext("Insufficient free LDC channels on "
		    "LDom %s to bind vnet %s\n"), vnet->ldp->namep,
		    vnet->name);
		return (rc);
	}

	/* link this vnet with the vsw and peer vnets */
	if ((rc = link_bound_vnet(vnet, vsw, resp)) != 0) {
		destroy_vnet_binding(vnet);
		return (rc);
	}

	return (0);
}

int
vnet_bind_all(ldom_t *ldp, response_t *resp)
{
	vnet_t *vnet;
	int rc;

	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		if ((rc = bind_vnet(vnet, resp)) != 0)
			return (rc);
	}

	return (0);
}

/*ARGSUSED1*/
void
vnet_unbind_all(ldom_t *ldp, response_t *resp)
{
	vnet_t *vnet;

	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		unbind_vnet(vnet);
	}
}

/* Future: vio.c might be a better home for this */
char *
comma_sepa_str(int num_vlans, int *vlans)
{
	int i, len;
	char *vlan_str;

	if ((num_vlans == 0) || (vlans == NULL))
		return (NULL);

	/*
	 * allocate space to store the comma separated list
	 * of vlans
	 */
	len = 0;
	for (i = 0; i < num_vlans; i++) {
		char vid_str[INT_STR_LEN];
		(void) snprintf(vid_str, sizeof (vid_str), "%d", vlans[i]);
		len += strlen(vid_str) + 1;
	}

	vlan_str = Xcalloc(len, char);

	/*
	 * Copy individual vlans to create a comma delimited
	 * string
	 */
	(void) snprintf(vlan_str, len, "%d", vlans[0]);
	for (i = 1; i < num_vlans; i++) {
		(void) snprintf(vlan_str, len, "%s,%d", vlan_str, vlans[i]);
	}

	return (vlan_str);
}

/*
 * Called from the add-vnet CLI.
 */
int
vnet_add(ldom_t *ldp, char *vnet_name, char *service, mac_addr_t mac,
    char *vnet_mode, int pvid, char *vids, int mtu, uint64_t phys_state,
    int networkid, response_t *resp)
{
	vnet_t *vnet;
	int rc = 0;
	int num_vlans;
	bool_t is_vid = (vids != NULL) ? true : false;
	int *vlans = parse_num_list(vids, &num_vlans);
	int i;

	if (vnet_find_by_name(ldp, vnet_name) != NULL) {
		rprintf(resp, "A %s client named %s already exists "
		    "in guest %s\n", vnet_attrs.name, vnet_name, ldp->namep);
		rc = ERR_CLIENT_EXISTS;
		goto ret;
	}

	if (is_vid == true) {
		if ((num_vlans == 1) && (vlans == NULL)) {
			rprintf(resp, gettext("Must specify vlan ids\n"));
			rc = ERR_NET_VID_INVALID;
			goto ret;
		} else if (vlans == NULL) {
			rprintf(resp, gettext("vid incorrectly specified\n"));
			rc = ERR_NET_VID_INVALID;
			goto ret;
		}
	}

	/*
	 * pvid value is already checked in cli front end
	 */
	if (vid_check(resp, num_vlans, vlans) != 0) {
		rc = ERR_NET_VID_INVALID;
		goto ret;
	}

	if (vlans != NULL) {
		for (i = 0; i < num_vlans; i++) {
			if (vlans[i] == pvid) {
				rprintf(resp, gettext("vid must not match "
				    "the port-vlan-id\n"));
				rc = ERR_NET_VID_INVALID;
				goto ret;
			}
		}
	}

	if (mtu == VIO_NOARGSUP)
		mtu = VIO_NONEMTU;

	/*
	 * if id is not specified in the CLI or the domain is still inactive,
	 * get the lowest free id value from the DB to use as a networkid
	 */
	if (networkid == NO_NEW_ID) {
		networkid = database_ldom_get_lowest_id(ldp, VNET_NODE);
	} else if (database_ldom_check_dup_id(ldp, VNET_NODE, networkid)
	    == true) {
		/*
		 * make sure that the same id doesn't already exist in the DB
		 */
		rprintf(resp, MSG_ERR_ID_EXISTS);
		return (ERR_DUP_ID);
	}

	vnet = vnet_new(ldp, vnet_name, service, mac, mtu, phys_state,
	    vnet_mode, pvid, num_vlans, vlans, networkid);

	if (ldp->state >= BOUND) {
#if BIND_CHECK
		if ((rc = add_vnet_check(vnet, resp)) != 0) {
			destroy_vnet(vnet);
			return (rc);
		}
#endif
		if ((rc = bind_vnet(vnet, resp)) != 0) {
			destroy_vnet(vnet);
			return (rc);
		}

		if (zeus.mode == ZEUS_MGMT_MODE) {
			if (vio_check_dr_capable(ldp, resp) == true) {
				/* DR in the vnet */
				rc = vio_dr_add(ldp, vnet_attrs.md_name,
				    vnet_devid(vnet), resp);
				/*
				 * If the DR operation failed and the dr-vio
				 * domain service determines that it is OK to
				 * rollback we remove the vnet again.
				 */
				if (rc != 0) {
					if (rc == ERR_DR_VIO_ROLLBACK) {
						unbind_vnet(vnet);
						destroy_vnet(vnet);
						seq_rm_reconfigure(ldp);
						free_all_staged_ldcs(ldp);
						return (rc);
					}
					/*
					 * Alert ldomconfigure.c that even
					 * though there was an error we kept
					 * the vnet in LDMD memory.
					 */
					return (ERR_DR_VIO_STILL_ALLOCATED);
				}

			} else if (ldp->state == BOUND) {
				/* do reconfig now */
				seq_download_mds_do_config(ldp);
				free_all_staged_ldcs(ldp);
			} else {
				/* handle as a delayed reconfig */
				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_ADD, resp);
			}
		}
	}

	database_ldom_add_net_client(ldp, vnet->name, vnet->service,
	    vnet->mac, vnet->mtu, vnet->phys_state, vnet_mode, pvid,
	    vids, networkid);

ret:
	if (rc != 0)
		Xfree(vlans);

	return (rc);
}

/*
 * Called from the set-vnet CLI.
 * Use set-vnet mode= <vnet-name> to remove hybrid io mode
 */
int
vnet_modify(vnet_t *vnet, char *service, mac_addr_t mac, char *vnet_mode,
	int pvid, char *vids, int mtu, uint64_t phys_state, response_t *resp)
{
	ldom_t *ldp = vnet->ldp;
	ldom_t *md_ldp = vnet->ldp;
	int upd_type = NO_UPD;
	int md_upd = MD_NONE;
	int num_vlans = 0;
	int *vlans = parse_num_list(vids, &num_vlans);
	vnet_binding_t *binding = vnet->binding;
	int n_vlans = (vlans != NULL) ? num_vlans : vnet->num_vlans;
	int *t_vlans = (vlans != NULL) ? vlans : vnet->vlans;
	int rc = 0;
	int i;
	int mod = 0;

	if ((mac != ILLEGAL_MAC_ADDR) && (mac != vnet->mac)) {
		upd_type |= OTHER_UPD; /* Must be unbound */
		mod |= VNET_MAC;
	}

	if ((service != NULL) && (strcmp(service, vnet->service) != 0)) {
		upd_type |= OTHER_UPD; /* Must be unbound */
		mod |= VNET_SERVICE;
	}

	if (pvid == -1)
		pvid = vnet->pvid;
	else if (pvid == 0)
		pvid = DEFAULT_VLAN_ID;

	if (mtu == VIO_NOARGSUP)
		mtu = vnet->mtu;

	if (phys_state == ILLEGAL_LNKP)
		phys_state = vnet->phys_state;

	/*
	 * pvid value is already checked in cli front end
	 */

	if ((vids != NULL) && (num_vlans < 1)) {
		rprintf(resp, gettext("vid incorrectly specified\n"));
		rc = ERR_NET_VID_INVALID;
		goto ret;
	}

	if (vid_check(resp, num_vlans, vlans) != 0) {
		rc = ERR_NET_VID_INVALID;
		goto ret;
	} else if (vids != NULL) {
		upd_type |= DYNAMIC_UPD;
		md_upd |= (MD_SELF|MD_PEERS|MD_SERVER);
		mod |= VNET_VID;
	}

	/*
	 * pvid must not match any vid
	 */
	for (i = 0; i < n_vlans; i++) {
		if (t_vlans[i] == pvid) {
			rprintf(resp, gettext("port-vlan-id must "
			    "not match any vid\n"));
			rc = ERR_NET_VID_INVALID;
			goto ret;
		}
	}

	if (((pvid != 0) && (vnet->pvid != pvid)) || (num_vlans > 0)) {
		upd_type |= DYNAMIC_UPD;
		md_upd |= (MD_SELF|MD_PEERS|MD_SERVER);
		mod |= VNET_PVID;
	}

	if (vnet_mode != NULL) {
		if (vnet_mode[0] == '\0') {
			if (vnet->vnet_mode != NULL) {
				md_upd |= MD_SERVER;
				mod |= VNET_MODE;
#if LIVE_HIO
				upd_type |= DYNAMIC_UPD;
#else
				upd_type |= DELAYED_UPD;
#endif
			}
		} else if (vnet->vnet_mode == NULL) {
			md_upd |= MD_SERVER;
			mod |= VNET_MODE;
#if LIVE_HIO
			upd_type |= DYNAMIC_UPD;
#else
			upd_type |= DELAYED_UPD;
#endif
		} else if (strcmp(vnet_mode, vnet->vnet_mode) != 0) {
			md_upd |= MD_SERVER;
			mod |= VNET_MODE;
#if LIVE_HIO
			upd_type |= DYNAMIC_UPD;
#else
			upd_type |= DELAYED_UPD;
#endif
		}
	}

	if (vnet->mtu != mtu) {
		mod |= VNET_MTU;
		md_upd |= MD_SELF;
#if LIVE_SET_MTU
		upd_type |= DYNAMIC_UPD;
#else
		upd_type |= DELAYED_UPD;
#endif
	}

	if (vnet->phys_state != phys_state) {
		mod |= VNET_PSTATE;
		md_upd |= MD_SELF;
		upd_type |= DYNAMIC_UPD;
	}

	/*
	 * No changes to the config need to be made - any options
	 * specified match the existing value(s).
	 */
	if (upd_type == NO_UPD) {
		rc = 0;
		goto ret;
	}

	if ((upd_type & DELAYED_UPD) != 0 && config_mod_allowed(ldp,
	    DEL_RECONFIG_OTHER, DR_OP_TYPE_VIO_SET, resp) == false) {
		rc = ERR_NO_DELRECONF_OTHER;
		goto ret;
	}

	/*
	 * set-vnet is allowed on a bound ldom only if the changes in
	 * parameters result in the flags being set (vlan, hybrid mode,
	 * mtu). Some changes like MAC/service name result in changes
	 * to multiple MDs and are classified as OTHER_UPD and can be
	 * done on unbound LDom only.
	 */
	if ((ldp->state >= BOUND) && (zeus.mode == ZEUS_MGMT_MODE)) {
		if ((upd_type & OTHER_UPD) != 0) {
			rprintf(resp, gettext("This modification to "));
			if (ldp->state == ACTIVE)
				rprintf(resp, gettext("an "));
			else
				rprintf(resp, gettext("a "));
			rprintf(resp, gettext("%s vnet is not supported\n"),
			    ldom_state_str(ldp->state));
			rc = ERR_VNET_NO_MODIFY_BOUND;
			goto ret;
		}
	}

	/*
	 * Actual updates can proceed now
	 */
	if ((service != NULL) && (strcmp(service, vnet->service) != 0)) {
		Xfree(vnet->service);
		vnet->service = Xstrdup(service);
	}

	if (vnet_mode != NULL) {
		if (vnet_mode[0] == '\0') {
			if (vnet->vnet_mode != NULL) {
				Xfree(vnet->vnet_mode);
				vnet->vnet_mode = NULL;
			}
		} else if (vnet->vnet_mode == NULL) {
			vnet->vnet_mode = Xstrdup(vnet_mode);
		} else if (strcmp(vnet_mode, vnet->vnet_mode) != 0) {
			Xfree(vnet->vnet_mode);
			vnet->vnet_mode = Xstrdup(vnet_mode);
		}
	}

	if ((mac != ILLEGAL_MAC_ADDR) && (mac != vnet->mac))
		vnet->mac = mac;

	if (vnet->pvid != pvid)
		vnet->pvid = pvid;

	if (num_vlans > 0) {
		Xfree(vnet->vlans);
		vnet->vlans = vlans;
		vnet->num_vlans = (vlans == NULL) ? 0 : num_vlans;
	}

	if (vnet->mtu != mtu)
		vnet->mtu = mtu;

	if (vnet->phys_state != phys_state)
		vnet->phys_state = phys_state;

	database_ldom_modify_net_client(ldp, mod, vnet->name, service, mac,
	    vnet->mtu, vnet->phys_state, vnet_mode, pvid, num_vlans,
	    (vlans == NULL) ? NULL : vids);

	/*
	 * For changes in hybrid IO mode, guest owning vsw needs new MD
	 * for changes in vlan or mtu, the switch, peer vnets and this vnet
	 * all need to be updated.
	 */
	if (binding == NULL) {
		goto ret;
	} else {
		if ((md_upd & MD_SELF) != 0) {
			md_ldp = vnet->ldp;
			md_ldp->flag.gen_new_md = true;
			md_ldp->flag.md_notify = true;
		}
		if ((md_upd & MD_SERVER) != 0) {
			md_ldp = vnet->binding->vsw_port->vsw->ldp;
			md_ldp->flag.gen_new_md = true;
			md_ldp->flag.md_notify = true;
		}
		if ((md_upd & MD_PEERS) != 0) {
			vnet_peer_t *vnet_peer;
			CHAIN_ITER(binding->peers, vnet_peer, link) {
				md_ldp = vnet_peer->peer->vnet->ldp;
				md_ldp->flag.gen_new_md = true;
				md_ldp->flag.md_notify = true;
			}
		}
	}

	if (ldp->state >= BOUND && zeus.mode == ZEUS_MGMT_MODE) {
		if (ldp->state != BOUND && ((upd_type & DELAYED_UPD) != 0)) {
			ldom_t *dldp;
			if ((md_upd & MD_SELF) != 0)
				dldp = vnet->ldp;
			else if ((md_upd & MD_SERVER) != 0)
				dldp = vnet->binding->vsw_port->vsw->ldp;
			else {
				rprintf(resp, gettext("Modifying a bound "
				    "vnet is not supported\n"));
				rc = ERR_VNET_NO_MODIFY_BOUND;
				goto ret;
			}

			/* handle as a delayed reconfig */
			initiate_or_update_delayed_reconfig(dldp,
			    DEL_RECONFIG_OTHER, resp);
		} else {
			DBG(zprintf("%s: sending new md and notify guests\n",
			    __func__));
			seq_download_mds_do_config(md_ldp);
			/*
			 * No setup/teardown of ldc channels needed
			 */
		}
	}

ret:
	if (rc != 0)
		Xfree(vlans);

	return (rc);
}

/*
 * Called from vnet_remove() which is triggerd by the remove-vnet CLI.
 *
 * This function removes the vnet from the DB and destroys the in-memory copy
 */
void
vnet_remove_do_teardown(vnet_t *vnet)
{
	database_ldom_del_net_client(vnet->ldp, vnet->name);
	unbind_vnet(vnet);
	destroy_vnet(vnet);

}

/*
 * Called from the remove-vnet CLI.
 */
/*ARGSUSED1*/
int
vnet_remove(vnet_t *vnet, bool_t force, response_t *resp)
{
	ldom_t *ldp = vnet->ldp;
	int	rc;

	if ((ldp->state >= BOUND) && (zeus.mode == ZEUS_MGMT_MODE)) {
		if (ldp->state == BOUND) {
			/* do reconfig now */
			vnet_remove_do_teardown(vnet);
			seq_rm_reconfigure(ldp);
			free_all_staged_ldcs(ldp);

		} else if (vio_check_dr_capable(ldp, resp)) {
			/* handle as a DR operation */
			rc = vio_dr_rm_phase1(ldp, vnet_attrs.md_name,
			    vnet_devid(vnet), resp);
			if (rc != 0)
				return (rc);
			vnet_remove_do_teardown(vnet);
			vio_dr_rm_phase2(ldp);

		} else {
			/* handle as a delayed reconfig operation */
			vnet_remove_do_teardown(vnet);
			initiate_or_update_delayed_reconfig(ldp,
			    DEL_RECONFIG_REMOVE_2PHASE, resp);
		}
	} else {
		/* remove the vnet from the configuration */
		vnet_remove_do_teardown(vnet);
	}

	return (0);
}

static void
vnet_peer_list(vnet_peer_t *vnet_peer, response_t *resp, dispctl_t *dctl)
{
	char mac[MAC_STR_LEN];
	char *peer, *fmt;
	vnet_peer_t *remote_peer = vnet_peer->peer;
	vnet_t *vnet = remote_peer->vnet;
	char *vlan_str = NULL;
	char *p_vlan_str;
	int pvid = vnet->pvid;
	int mtu = vnet->mtu;
	char *mode_str = (vnet->vnet_mode != NULL) ? vnet->vnet_mode : "";

	if (mtu == VIO_NONEMTU) {
		mtu = vnet->binding->vsw_port->vsw->mtu;
		if (mtu == VIO_NONEMTU)
			mtu = VIO_ETHERMTU;
	}

	if (vnet->num_vlans > 0)
		vlan_str = comma_sepa_str(vnet->num_vlans, vnet->vlans);
	p_vlan_str = (vlan_str == NULL) ? "" : vlan_str;
	mactoa(remote_peer->vnet->mac, mac);
	peer = zsprintf("%s@%s", remote_peer->vnet->name,
	    remote_peer->vnet->ldp->namep);

	if (dctl->parseable)
		fmt = "|peer=%s|mac-addr=%s|mode=%s|pvid=%d|vid=%s|mtu=%d|"
		    "linkprop=%s";
	else
		fmt = "        "FBNDNAME" "FMAC" "FMODE" "FPVIDd" "FVID" "
		    FMTUd" "FLNKP;
	rprintf(resp, fmt, peer, mac, mode_str, pvid, p_vlan_str, mtu,
	    ((vnet->phys_state == 0x1) ? "phys-state" : ""));

	Xfree(peer);
	Xfree(vlan_str);

	if (dctl->extended) {
		if (dctl->parseable)
			rprintf(resp, "|ldc=");
		else
			rprintf(resp, " ");
		ldcs_print(&vnet_peer->ldcs, resp);
	}

	rprintf(resp, "\n");
}

static void
vnet_list_one(vnet_t *vnet, response_t *resp, dispctl_t *dctl)
{
	vnet_binding_t *binding = vnet->binding;
	char mac[MAC_STR_LEN];
	char *service_ldom, *devid;
	char *mode_str = (vnet->vnet_mode != NULL) ? vnet->vnet_mode : "";
	char *vlan_str = NULL;
	char *p_vlan_str;
	char *id_str;
	int pvid = vnet->pvid;
	int mtu = vnet->mtu;
	char *mtus;
	char *pstate;

	/*
	 * TRANSLATION_NOTE
	 * The following are column headers in tables with fixed column widths.
	 * WIDTH  HEADER	MEANING
	 * 16	  "NAME"	name of client
	 * 27	  "SERVICE"	name of the service the client is using
	 * 4	  "ID"		network id
	 * 10	  "DEVICE"	network device name
	 * 17	  "MAC"		MAC address
	 * 6	  "MODE"	MODE
	 * 6	  "PVID"	port vlan id
	 * 20	  "VID"		vlan id
	 * 5	  "MTU"		mtu
	 * 10	  "LINKPROP"	"phys-state" or ""
	 * 27	  "PEER"	name of peer network endpoint
	 * 3	  "LDC"		Logical Domain Channel number
	 */

	if (dctl->header && !dctl->parseable) {
		rprintf(resp, "    "FNAME" "FBNDNAME" "FID" "FDEVICE" "FMAC" "
		    FMODE" "FPVID" "FVID" "FMTU" "FLNKP"\n", gettext("NAME"),
		    gettext("SERVICE"), gettext("ID"), gettext("DEVICE"),
		    gettext("MAC"), gettext("MODE"), gettext("PVID"),
		    gettext("VID"), gettext("MTU"), gettext("LINKPROP"));
		dctl->header = 0;
	}

	if (dctl->constraints && is_mac_allocated(vnet->mac))
		(void) strcpy(mac, "");
	else
		mactoa(vnet->mac, mac);

	if (binding != NULL) {
		int vsw_mtu = binding->vsw_port->vsw->mtu;

		if (vsw_mtu == VIO_NONEMTU)
			vsw_mtu = VIO_ETHERMTU;
		service_ldom = binding->vsw_port->vsw->ldp->namep;
		devid =
		    zsprintf("%s@%"PRIu64"", vnet_attrs.name, vnet_devid(vnet));
		if (vnet->mtu == VIO_NONEMTU)
			mtu = vsw_mtu;
	} else {
		service_ldom = NULL;
		devid = "";
	}

	if (vnet->mtu != VIO_NONEMTU) {
		mtus = zsprintf("%d", mtu);
	} else {
		if (dctl->constraints || binding == NULL)
			mtus = zsprintf("");
		else
			mtus = zsprintf("%d", mtu);
	}

	pstate = zsprintf("%s", (vnet->phys_state == 0x1) ? "phys-state" : "");

	if (vnet->num_vlans > 0)
		vlan_str = comma_sepa_str(vnet->num_vlans, vnet->vlans);
	p_vlan_str = (vlan_str == NULL) ? "" : vlan_str;

	if (vnet->networkid != NO_NEW_ID)
		id_str =  zsprintf("%d", vnet->networkid);
	else
		id_str = zsprintf("");

	if (dctl->parseable) {
		rprintf(resp, "%s", dctl_label(vnet_attrs.cli_name, dctl));
		rprintf(resp, "|name=%s", vnet->name);
		rprintf(resp, "|dev=%s", devid);
		rprintf(resp, "|service=%s", vnet->service);

		/* print LDom name where service is running */
		if (service_ldom != NULL && !dctl->constraints)
			rprintf(resp, "@%s", service_ldom);

		rprintf(resp, "|mac-addr=%s", mac);
		rprintf(resp, "|mode=%s", mode_str);
		rprintf(resp, "|pvid=%d", pvid);
		rprintf(resp, "|vid=%s", p_vlan_str);
		rprintf(resp, "|mtu=%s", mtus);
		rprintf(resp, "|linkprop=%s", pstate);
		rprintf(resp, "|id=%s", id_str);
	} else {
		rprintf(resp, "    "FNAME, vnet->name);
		if (service_ldom != NULL && !dctl->constraints) {
			char *s;
			s = zsprintf("%s@%s", vnet->service, service_ldom);
			rprintf(resp, " "FBNDNAME, s);
			Xfree(s);
		} else {
			rprintf(resp, " "FBNDNAME, vnet->service);
		}
		rprintf(resp, " "FID, id_str);
		rprintf(resp, " "FDEVICE, devid);
		rprintf(resp, " "FMAC, mac);
		rprintf(resp, " "FMODE, mode_str);
		rprintf(resp, " "FPVIDd, pvid);
		rprintf(resp, " "FVID, p_vlan_str);
		rprintf(resp, " "FMTU, mtus);
		rprintf(resp, " "FLNKP, pstate);
	}
	Xfree(mtus);
	rprintf(resp, "\n");

	if (binding != NULL)
		Xfree(devid);
	Xfree(vlan_str);

	if (dctl->bindings && (binding != NULL)) {
		char *peer, *fmt;
		vnet_peer_t *vnet_peer;
		vsw_t *vsw = binding->vsw_port->vsw;
		int pvid = vsw->pvid;
		int vsw_mtu = (vsw->mtu != VIO_NONEMTU) ? vsw->mtu :
		    VIO_ETHERMTU;
		char *mtus;
		char *vlan_str = NULL;
		char *mode = (vsw->mode != NULL) ? vsw->mode : "";

		if (!dctl->parseable) {
			rprintf(resp, "        "FBNDNAME" "FMAC " "FMODE
			    " "FPVID" "FVID" "FMTU" "FLNKP, gettext("PEER"),
			    gettext("MAC"), gettext("MODE"), gettext("PVID"),
			    gettext("VID"), gettext("MTU"),
			    gettext("LINKPROP"));
			if (dctl->extended)
				rprintf(resp, " %s", gettext("LDC"));
			rprintf(resp, "\n");
			dctl->header = 1;
		}

		/* display the switch binding */
		mactoa(binding->vsw_port->vsw->mac, mac);
		peer = zsprintf("%s@%s", vnet->service, service_ldom);

		if (vsw->num_vlans > 0)
			vlan_str = comma_sepa_str(vsw->num_vlans, vsw->vlans);
		p_vlan_str = (vlan_str == NULL) ? "" : vlan_str;

		mtus = zsprintf("%d", vsw_mtu);

		if (dctl->parseable)
			fmt = "|peer=%s|mac-addr=%s|mode=%s|pvid=%d|vid=%s"
			    "|mtu=%s";
		else
			fmt = "        "FBNDNAME" "FMAC" "FMODE" "FPVIDd
			    " "FVID" "FMTU;
		rprintf(resp, fmt, peer, mac, mode, pvid, p_vlan_str, mtus);

		Xfree(peer);
		Xfree(vlan_str);
		Xfree(mtus);

		if (dctl->extended) {
			if (dctl->parseable)
				rprintf(resp, "|ldc=");
			else
				rprintf(resp, " ");
			ldcs_print(&binding->ldcs, resp);
		}
		rprintf(resp, "\n");

		CHAIN_ITER(binding->peers, vnet_peer, link) {
			vnet_peer_list(vnet_peer, resp, dctl);
		}
	}


}

void
vnet_list(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	vnet_t *vnet;

	if (dctl->oformat && !dctl->onetwork)
		return;

	if (CHAIN_LENGTH(ldp->vnets) == 0)
		return;

	dctl->header = 1;

	if (!dctl->parseable)
		rprintf(resp, "%s\n", dctl_label(vnet_attrs.name, dctl));

	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		vnet_list_one(vnet, resp, dctl);
	}

	if (!dctl->parseable)
		rprintf(resp, "\n");
}

static vnet_peer_t *
vnet_peer_find_by_ldcid(ldom_t *ldp, uint64_t ldcid)
{
	vnet_t *vnet;
	vnet_peer_t *peer;

	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		if (vnet->binding == NULL)
			continue;
		CHAIN_ITER(vnet->binding->peers, peer, link) {
			if (ldcs_search(&peer->ldcs, ldcid) != NULL)
				return (peer);
		}
	}

	return (NULL);
}

static int
link_vnet_client(ldom_t *ldp, vnet_t *vnet)
{
	ldc_epdb_t *ep;
	ldc_chan_t *ldc;
	vsw_port_t *vsw_port;
	vnet_peer_t *peer, *remote;
	char *vnet_mode = NULL;
	char *vlan_str = NULL;

	if (vnet->binding == NULL)
		fatal("Vnet %s has no binding to a vswitch", vnet->name);

	/* bind to switch */
	ldc = CHAIN_START(vnet->binding->ldcs.ldcs);
	if ((ep = ldc_epdb_find(ldp->idp->id, ldc->chan->id)) != NULL)
		vsw_port = vsw_port_find_by_ldcid(ep->peer->ldp, ep->peer->ldc);
	if ((ep == NULL) || (vsw_port == NULL))
		fatal("Could not find matching vsw port for vnet %s on "
		    "LDom %s", vnet->name, ldp->namep);

	vsw_port->vnet = vnet;
	vnet->binding->vsw_port = vsw_port;

	/*
	 * Get vnet's niu params from its switch-port
	 */
	vnet_mode = vsw_port->work.vnet_mode;
	vnet->vnet_mode = (vnet_mode != NULL) ? Xstrdup(vnet_mode) : NULL;

	if (vsw_port->vsw->ldp != ldp)
		vsw_port->vsw->ldp->bound.num_clients++;

	ASSERT(vnet->service == NULL);
	vnet->service = Xstrdup(vsw_port->vsw->service);

	/* bind to peers */
	CHAIN_ITER(vnet->binding->peers, peer, link) {
		ldc = CHAIN_START(peer->ldcs.ldcs);
		if ((ep = ldc_epdb_find(ldp->idp->id, ldc->chan->id)) != NULL)
			remote = vnet_peer_find_by_ldcid(ep->peer->ldp,
			    ep->peer->ldc);
		if ((ep == NULL) || (remote == NULL))
			fatal("Could not find matching vnet peer for vnet %s "
			    "on LDom %s", vnet->name, ldp->namep);
		remote->peer = peer;
		peer->peer = remote;
	}

	vlan_str = comma_sepa_str(vnet->num_vlans, vnet->vlans);

	database_ldom_add_net_client(ldp, vnet->name, vnet->service,
	    vnet->mac, vnet->mtu, vnet->phys_state, vnet_mode, vnet->pvid,
	    vlan_str, vnet->networkid);
	Xfree(vlan_str);

	return (0);
}

int
link_vnet_clients(ldom_t *ldp)
{
	vnet_t *vnet;

	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		(void) link_vnet_client(ldp, vnet);
	}

	return (0);
}

void
vnet_print_bindings(ldom_t *ldp, response_t *resp)
{
	char		idstr[UINT64_STR_LEN];
	vnet_t		*vnet;
	vnet_binding_t	*bind;
	vnet_peer_t	*peer;
	ldc_chan_t	*ldc;

	CHAIN_ITER(ldp->vnets, vnet, ldom) {

		bind = vnet->binding;

		(void) snprintf(idstr, UINT64_STR_LEN, "vnet%"PRId64,
		    bind->cfg_handle->id);
		rprintf(resp, "  %-11s", idstr);

		CHAIN_ITER(bind->ldcs.ldcs, ldc, link) {
			rprintf(resp, "%2"PRId64" ", ldc->chan->id);
		}

		rprintf(resp, "  %s@%s\n", vnet->name, vnet->service);

		CHAIN_ITER(bind->peers, peer, link) {
			if ((ldc = CHAIN_START(peer->ldcs.ldcs)) == NULL)
				continue;

			rprintf(resp, "  %-11s%2"PRId64"   %s@%s\n", "",
			    ldc->chan->id, peer->peer->vnet->name,
			    peer->peer->vnet->ldp->namep);
		}
	}
}

int
vnet_mig_add_bindings(ldom_t *ldp, md_t *mdp)
{
	int		i;
	int		rc = -1;
	md_node_t	**vdevs;
	int		nvdevs;
	char		*name;
	vnet_t		*vnet;

	/*
	 * Scan through the MD for all virtual device nodes.
	 */
	vdevs = Xmalloc(md_node_count(mdp) * sizeof (md_node_t *));
	nvdevs = md_scan_dag(mdp, NULL, md_find_name(mdp, VDEV_NAME),
	    md_find_name(mdp, "fwd"), vdevs);

	ASSERT(nvdevs > 0);

	for (i = 0; i < nvdevs; i++) {
		if (md_get_prop_str(mdp, vdevs[i], "name", &name) < 0) {
			DBG(zprintf("virtual-device node has no name prop\n"));
			goto done;
		}

		if (strcmp(name, "network") != 0)
			continue;

		if (md_get_prop_str(mdp, vdevs[i], "devalias", &name) < 0) {
			DBG(zprintf("virtual-device node has no name prop\n"));
			goto done;
		}

		if ((vnet = vnet_find_by_name(ldp, name)) == NULL) {
			DBG(zprintf("%s: Unknown vnet %s\n", __func__, name));
			goto done;
		}

		/* add the service name */
		if (vnet->service != NULL) {
			md_add_string_property(mdp, vdevs[i], "service-name",
			    vnet->service);
		}
	}

	rc = 0;

done:
	Xfree(vdevs);
	return (rc);
}

/*
 * Count how many bound vnet(s) have phys_state set
 * per a specific vsw
 */
int
vnet_num_linkprop_set(vsw_t *vsw)
{
	vsw_port_t	*vsw_port;
	int		count = 0;

	CHAIN_ITER(vsw->binding->vports, vsw_port, link) {
		if (vsw_port->vnet->phys_state == 0x1) {
			++count;
		}
	}

	return (count);
}
