/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_VNET_H
#define	_VNET_H

#ifdef __cplusplus
extern "C" {
#endif

#include "ldmd_types.h"
#include "chain.h"
#include "ptrvec.h"
#include "ldcs.h"
#include "mac_addrs.h"
#include "resource_id.h"
#include "response.h"
#include "dispctl.h"

/*
 * Port 0 is reserved for the vnet to vswitch binding.
 */
#define	VNET_VSW_PORT	0
#define	VNET_MIN_PORT	1

#define	VNET_DEFAULT_MTU 1500

#define	VNET_SW_PORT_PROP	"switch-port"

struct vnet_attributes {
	char		*name;
	char		*compatible;
	char		*md_name;
	char		*cli_name;
	char		*device_type;
	char		*port_name;
};

/*
 * Represent each argument to add/set_vnet used
 * to identy which has valid values
 */
typedef enum {
	VNET_SERVICE = 0x1,
	VNET_MAC = 0x2,
	VNET_MODE = 0x4,
	VNET_PVID = 0x8,
	VNET_VID = 0x10,
	VNET_MTU = 0x20,
	VNET_PSTATE = 0x40
} vnet_args_t;

struct vnet {
	ldom_t 		*ldp;		/* owner ldom */
	char		*name;
	char		*service;	/* name of virtual switch service */
	mac_addr_t	mac;
	uint32_t	mtu;
	uint64_t	phys_state;	/* IPMP network link status */
	char		*vnet_mode;
	int		*vlans;
	int		num_vlans;
	int		pvid;
	int		networkid;	/* network id */
	vnet_binding_t	*binding;
	CHAIN_LINK(vnet_t, ldom);	/* for list of vnets owned by a ldom */
};

struct vnet_binding {
	ldcs_t		ldcs;		/* LDC channels to the vswitch */
	resid_t 	*cfg_handle;
	vsw_port_t	*vsw_port;	/* vswitch port this vnet is bound to */
	res_pool_t	free_ports;	/* free port ids */
	CHAIN(vnet_peer_t, peers); 	/* list of peer vnets */
};

/*
 * Records a binding to another vnet connected to the same switch, i.e. a peer
 * vnet.
 */
struct vnet_peer {
	ldcs_t		ldcs;		/* LDC channels to peer vnet */
	resid_t		*port;		/* port id */
	vnet_t		*vnet;		/* parent vnet */
	vnet_peer_t	*peer;		/* peer vnet on remote end of channel */
	CHAIN_LINK(vnet_peer_t, link);
};

extern struct vnet_attributes vnet_attrs;

ident_val_t vnet_devid(vnet_t *vnet);
vnet_t *vnet_find_by_name(ldom_t *ldp, char *vnet_name);
void vnet_add_from_mdnode(ldom_t *ldp, md_t *guest_md, md_node_t *vdev);
int link_vnet_clients(ldom_t *ldp);
void vnet_list(ldom_t *ldp, response_t *resp, dispctl_t *dctl);
vnet_t *vnet_new(ldom_t *ldp, char *name, char *service, mac_addr_t mac,
    int mtu, uint64_t phys_state, char *vnet_mode, int pvid, int num_vlans,
    int *vlans, int networkid);
int vnet_add(ldom_t *ldp, char *vnet_name, char *service, mac_addr_t mac,
    char *mode, int pvid, char *vids, int mtu, uint64_t phys_state,
    int networkid, response_t *resp);
int vnet_modify(vnet_t *vnet, char *service, mac_addr_t mac, char *mode,
    int pvid, char *vids, int mtu, uint64_t phys_state, response_t *resp);
int vnet_remove(vnet_t *vnet, bool_t force, response_t *resp);
int vnet_rm_dr(vnet_t *vnet, bool_t force, response_t *resp);
void vnet_rm_reconfig(vnet_t *vnet, response_t *resp);
void vnet_gen_ldc_endpoints(ldom_t *ldp, md_t *hv_md, md_node_t *hv_guest,
    md_node_t *ldc_endpoints);
void vnet_gen_guest_md(ldom_t *ldp, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints);
int vnet_bind_all(ldom_t *ldp, response_t *resp);
void vnet_unbind_all(ldom_t *ldp, response_t *resp);
void vnet_delete_all(ldom_t *ldp);
int vnet_bind_check(ldom_t *ldp, response_t *resp);
void vnet_print_bindings(ldom_t *ldp, response_t *resp);
int vnet_mig_add_bindings(ldom_t *ldp, md_t *mdp);
int vnet_num_linkprop_set(vsw_t *vsw);

#ifdef __cplusplus
}
#endif

#endif	/* _VNET_H */
