/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <string.h>

#include "ldmd_types.h"
#include "allocate.h"
#include "chain.h"
#include "database.h"
#include "database_internal.h"
#include "errors.h"
#include "fatal.h"
#include "ldom.h"
#include "resource_id.h"
#include "vio.h"

/* list of all vsw services */
static CHAIN(vsw_t, vsw_services) = DEFAULTCHAIN;

struct vsw_attributes vsw_attrs = {
	.name =		"vsw",
	.compatible =	"SUNW,sun4v-network-switch",
	.md_name =	"virtual-network-switch",
	.cli_name =	"vsw",
	.device_type =	"vsw",
	.port_name =	"vsw-port"
};

/*
 * A simple list of ldoms to prevent duplicate listing
 */
typedef struct doml doml_t;
struct doml {
	ldom_t *ldp;
	CHAIN_LINK(doml_t, link);
};

static CHAIN(doml_t, domains) = DEFAULTCHAIN;

/* vsw mode definitions */
vsw_mode_options_t vsw_mode_options [] = {
	{"sc", 0x833LL},
	{NULL, 0LL}
};

/*
 * Returns the switch mode as a function of the network device.
 */
void
vsw_get_sw_mode(char *net_dev, int *sw_mode_len, uint8_t **sw_mode,
    char **sw_mode_xml)
{
	static uint8_t *dev_sw_mode = NULL;
	static int dev_sw_mode_len;
	static uint8_t *nodev_sw_mode;
	static int nodev_sw_mode_len;

	if (dev_sw_mode == NULL) {
		/* first time called, init the MD forms of the switch mode */
		dev_sw_mode_len = strlen(VSW_SWITCHED) +
		    strlen(VSW_PROMISC) + 2;
		dev_sw_mode = Xcalloc(dev_sw_mode_len, uint8_t);
		strcpy((char *)dev_sw_mode, VSW_SWITCHED);
		strcpy((char *)(dev_sw_mode + strlen(VSW_SWITCHED) + 1),
		    VSW_PROMISC);

		nodev_sw_mode = (uint8_t *)VSW_ROUTED;
		nodev_sw_mode_len = strlen(VSW_ROUTED) + 1;
	}

	if ((net_dev == NULL) || (net_dev[0] == '\0')) {
		if (sw_mode != NULL) {
			*sw_mode = nodev_sw_mode;
			*sw_mode_len = nodev_sw_mode_len;
		}
		if (sw_mode_xml != NULL)
			*sw_mode_xml = VSW_ROUTED;
	} else {
		if (sw_mode != NULL) {
			*sw_mode = dev_sw_mode;
			*sw_mode_len = dev_sw_mode_len;
		}
		if (sw_mode_xml != NULL)
			*sw_mode_xml = VSW_SWITCHED "," VSW_PROMISC;
	}
}

ident_val_t
vsw_devid(vsw_t *vsw)
{
	if (vsw->binding == NULL)
		return (INVALID_DEVID);
	else
		return (vsw->binding->cfg_handle->id);
}

static res_pool_t *
vsw_cfg_handles(ldom_t *ldp)
{
#if VDEV_CFG_HANDLE
	return (&ldp->free_cfg_handles);
#else
	return (&ldp->bound.vsw_cfg_handles);
#endif
}

vsw_t *
vsw_find_by_name(char *service_name)
{
	vsw_t *vsw;

	CHAIN_ITER(vsw_services, vsw, link) {
		if (strcmp(service_name, vsw->service) == 0)
			break;
	}

	return (vsw);
}

static vsw_t *
find_by_phys_dev(ldom_t *ldp, char *phys_dev)
{
	vsw_t *vsw;

	CHAIN_ITER(ldp->vsws, vsw, ldom) {
		if ((vsw->phys_dev != NULL) &&
		    (strcmp(phys_dev, vsw->phys_dev) == 0))
			break;
	}

	return (vsw);
}

/*
 * Map vsw mode from md value to cli.
 */
static char *
vsw_mode_md_to_cli(uint64_t md)
{
	int i = 0;

	while (vsw_mode_options[i].cli != NULL) {
		if (md == vsw_mode_options[i].md)
			return (vsw_mode_options[i].cli);
		i++;
	}

	return (NULL);
}


/*
 * Map vsw mode from cli to md value
 */
static uint64_t
vsw_mode_cli_to_md(char *cli)
{
	int i = 0;

	ASSERT(cli != NULL && *cli != '\0');
	while (vsw_mode_options[i].cli != NULL) {
		if (strcmp(cli, vsw_mode_options[i].cli) == 0) {
			return (vsw_mode_options[i].md);
		}
		i++;
	}

	return (0);
}

static void
set_ether_types_from_mode(vsw_t *vsw, char *mode)
{
	char *m;
	char *s = NULL;
	int count = 0;

	if (mode == NULL) {
		vsw->num_ether_types = 0;
		vsw->ether_types = NULL;
		return;
	}

	/* look for mode(s) */
	m = Xstrdup(mode);
	s = strtok(m, ",");
	while (s != NULL) {
		count++;
		s = strtok(NULL, ",");
	}

	Xfree(m);

	if (count == 0) {
		vsw->num_ether_types = 0;
		vsw->ether_types = NULL;
		return;
	}

	vsw->num_ether_types = count;
	vsw->ether_types = Xmalloc(sizeof (uint64_t) * count);

	count = 0;

	m = Xstrdup(mode);
	s = strtok(m, ",");
	while (s != NULL) {
		vsw->ether_types[count++] =  vsw_mode_cli_to_md(s);
		s = strtok(NULL, ",");
	}

	Xfree(m);
}

vsw_t *
vsw_new(ldom_t *ldp, char *service_name, char *phys_dev, mac_addr_t mac,
    char *mode, int default_vid, int pvid, int num_vlans, int *vlans, int mtu,
    uint64_t phys_state, int switchid)
{
	vsw_t *vsw;

	if ((vsw = vsw_find_by_name(service_name)) != NULL)
		return (NULL);

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(ldp->vsws) == 0) {
		create_res_pool(&ldp->bound.vsw_cfg_handles,
		    IDENT_MIN, IDENT_MAX, Res_sorted);
		name_ldom_res_pool(&ldp->bound.vsw_cfg_handles, ldp,
		    "vsw cfg handles");
	}
#endif

	vsw = Xmalloc(sizeof (vsw_t));

	vsw->ldp = ldp;
	vsw->service = Xstrdup(service_name);
	vsw->phys_dev = (phys_dev != NULL) ? Xstrdup(phys_dev) : NULL;
	vsw->mode = (mode != NULL) ? Xstrdup(mode) : NULL;
	set_ether_types_from_mode(vsw, mode);
	vsw->mac = mac;
	vsw->default_vid = default_vid;
	vsw->pvid = pvid;
	vsw->num_vlans = (vlans == NULL) ? 0 : num_vlans;
	vsw->vlans = vlans;
	vsw->mtu = (mtu > 0) ? mtu : VIO_NONEMTU;
	vsw->phys_state = phys_state;
	vsw->switchid = switchid;
	vsw->binding = NULL;

	CHAIN_ADD(vsw_services, vsw, link);
	CHAIN_ADD(ldp->vsws, vsw, ldom);

	return (vsw);
}

static int
bind_vsw(vsw_t *vsw, int cfg_handle)
{
	ldom_t *ldp = vsw->ldp;
	vsw_binding_t *binding;

	/*
	 * Do not bind if the vsw or any vnet(s) in the self-same domain
	 * are depending upon a physical net-dev that has still not yet
	 * been assigned
	 */
	if (vsw->phys_dev == NULL) {
		if (vsw->phys_state == 0x1) {
			return (ERR_VSW_LINKPROP_VSW);
		} else {
			vnet_t	*vnet;

			CHAIN_ITER(vsw->ldp->vnets, vnet, ldom) {
				if (vnet->service == NULL ||
				    vsw->service == NULL)
					continue;
				if (strcmp(vnet->service, vsw->service) == 0) {
					if (vnet->phys_state == 0x1)
						return (ERR_VSW_LINKPROP_VNET);
				}
			}
		}
	}

	/*
	 * if the given cfg_handle value is same as NO_NEW_ID,
	 * use the vsw's switch id which is already retrieved from the DB
	 * or set with the lowest available number from the DB
	 */
	if (cfg_handle == NO_NEW_ID)
		cfg_handle = vsw->switchid;

	binding = Xmalloc(sizeof (vsw_binding_t));

	binding->cfg_handle = vio_get_cfg_handle(cfg_handle,
	    vsw_cfg_handles(ldp));
	if (binding->cfg_handle == NULL) {
		DBG(zprintf("Could not obtain vsw cfg_handle %d on LDom %s\n",
		    cfg_handle, ldp->namep));
		Xfree(binding);
		return (-1);
	}

	create_res_pool(&binding->free_ports, 0, IDENT_MAX, Res_sorted);
	name_ldom_vdev_res_pool(&binding->free_ports, ldp, "vsw",
	    vsw->service, "ports");

	CHAIN_INIT(binding->vports);

	ldp->bound.num_services++;
	vsw->binding = binding;

	/*
	 * if the value for switchid is NO_NEW_ID, update the switchid with
	 * the cfg_handle id just pulled from the cfg_handle pool
	 */
	if (vsw->switchid == NO_NEW_ID)
		vsw->switchid = vsw->binding->cfg_handle->id;

	return (0);
}

static void
unbind_vsw(vsw_t *vsw)
{
	if (vsw->binding == NULL)
		return;

	ASSERT(CHAIN_LENGTH(vsw->binding->vports) == 0); /* no active clients */

	vsw->ldp->bound.num_services--;
	destroy_res_pool(&vsw->binding->free_ports);
	free_res(vsw->binding->cfg_handle);
	Xfree(vsw->binding);
	vsw->binding = NULL;
}

static void
destroy_vsw(vsw_t *vsw)
{
	ldom_t *ldp = vsw->ldp;

	ASSERT(vsw->binding == NULL);

	CHAIN_UNLINK(ldp->vsws, vsw, ldom);
	CHAIN_UNLINK(vsw_services, vsw, link);

	if (vsw->num_vlans > 0)
		Xfree(vsw->vlans);
	Xfree(vsw->phys_dev);
	Xfree(vsw->mode);
	Xfree(vsw->ether_types);
	Xfree(vsw->service);
	Xfree(vsw);

#if !VDEV_CFG_HANDLE
	if (CHAIN_LENGTH(ldp->vsws) == 0)
		destroy_res_pool(&ldp->bound.vsw_cfg_handles);
#endif
}

/*ARGSUSED1*/
int
vsw_bind_all(ldom_t *ldp, response_t *resp)
{
	vsw_t	*vsw;
	int	rc;

	CHAIN_ITER(ldp->vsws, vsw, ldom) {
		rc = bind_vsw(vsw, NEW_INSTANCE);
		if (rc == ERR_VSW_LINKPROP_VSW) {
			rprintf(resp, MSG_ERR_BIND_VSW_NO_NET_DEV);
			return (rc);
		} else if (rc == ERR_VSW_LINKPROP_VNET) {
			rprintf(resp, MSG_ERR_BIND_VSW_NO_NET_DEV_VNET);
			return (rc);
		}
	}

	return (0);
}

/*ARGSUSED1*/
int
vsw_unbind_all(ldom_t *ldp, response_t *resp)
{
	vsw_t *vsw;

	CHAIN_ITER(ldp->vsws, vsw, ldom) {
		unbind_vsw(vsw);
	}

	return (0);
}

/*
 * Called from vnet code to create a new port on the vswitch when binding a
 * new vnet to it. Also called when initializing vio bindings from MDs.
 */
vsw_port_t *
vsw_add_port(vsw_t *vsw, vnet_t *vnet, uint64_t port, char *vnet_mode)
{
	vsw_port_t *vport;

	vport = Xmalloc(sizeof (vsw_port_t));

	if (port == NEW_INSTANCE)
		vport->port = alloc_res(&vsw->binding->free_ports);
	else
		vport->port = pull_res_id(&vsw->binding->free_ports, port);

	vport->vnet = vnet;
	vport->vsw = vsw;

	vport->work.vnet_mode = (vnet_mode != NULL) ?
	    Xstrdup(vnet_mode) : NULL;

	CHAIN_INIT(vport->ldcs.ldcs);

	CHAIN_ADD(vsw->binding->vports, vport, link);

	return (vport);
}

static char *
gen_vsw_mode_cli(uint8_t *md_prop_str, int md_prop_len)
{
	int  i = 0;
	uint8_t	*md_mode = md_prop_str;
	uint64_t md_prop_val;
	char *vsw_mode_cli = NULL;
	char *cli;
	int  space = 0;

	ASSERT(md_prop_len >= sizeof (uint64_t));

	while (md_mode < (md_prop_str + md_prop_len)) {
		md_prop_val = *((uint64_t *)md_mode);

		cli = vsw_mode_md_to_cli(md_prop_val);
		ASSERT(cli != NULL);
		if (cli == NULL)
			return (NULL);

		space += strlen(cli)+1;
		md_mode += sizeof (uint64_t);
	}

	vsw_mode_cli = Xcalloc(space, char);

	md_mode = md_prop_str;
	while (md_mode < (md_prop_str + md_prop_len)) {
		md_prop_val = *((uint64_t *)md_mode);

		cli = vsw_mode_md_to_cli(md_prop_val);
		if (i == 0) {
			strcpy(vsw_mode_cli, cli);
			i++;
		} else {
			strcat(vsw_mode_cli, ",");
			strcat(vsw_mode_cli, cli);
		}

		md_mode += sizeof (uint64_t);
	}

	return (vsw_mode_cli);
}


void
vsw_add_from_mdnode(ldom_t *ldp, md_t *guest_md, md_node_t *vdev)
{
	char *service_name, *mode;
	uint8_t *phys_dev;
	uint64_t cfg_handle, mac_addr;
	uint8_t *ether_types;
	int i, num_client_bindings, len;
	md_node_t **nodelistp;
	vsw_t *vsw;
	char *vlan_str = NULL;
	uint64_t default_vid = DEFAULT_VLAN_ID;
	uint64_t pvid = default_vid;
	uint64_t *vids = NULL;
	int num_vlans = 0;
	int *vlans = NULL;
	int db_mtu = -1;
	uint64_t db_physstate = ILLEGAL_LNKP;
	uint64_t mtu = VIO_ETHERMTU;
	uint64_t phys_state = 0;
	int switchid = NO_NEW_ID;

	if (md_get_prop_val(guest_md, vdev, "cfg-handle", &cfg_handle) != 0)
		fatal("vdevice has no cfg-handle property");

	/*
	 * The vsw-phys-dev property is an array of strings, we just retrieve
	 * the first value (2.0 FIXME).
	 */
	if (md_get_prop_data(guest_md, vdev, VSW_PHYS_DEV_PROP, &phys_dev,
	    &len) != 0)
		fatal("vsw-vport %"PRIx64" missing %s property", cfg_handle,
		    VSW_PHYS_DEV_PROP);
	if (phys_dev[0] == '\0')
		phys_dev = NULL;

	if (md_get_prop_data(guest_md, vdev, VSW_ETHER_TYPE_PROP,
	    &ether_types, &len) == 0)
		mode = gen_vsw_mode_cli(ether_types, len);
	else
		mode = NULL;

	if (md_get_prop_val(guest_md, vdev, VSW_LOCAL_MAC_PROP, &mac_addr) != 0)
		mac_addr = 0ull;	/* default to 0 */

	/*
	 * Get the service name from the devalias (check first) or
	 * svc-name property if it exists, otherwise get it from the database.
	 */
	if (md_get_prop_str(guest_md, vdev, DEVALIAS, &service_name) == 0)
		service_name = Xstrdup(service_name);
	else if (md_get_prop_str(guest_md, vdev, SERVICE_NAME,
	    &service_name) == 0)
		service_name = Xstrdup(service_name);
	else {
		service_name = get_service_name_from_db(ldp, vsw_attrs.name,
		    cfg_handle);
		if (service_name == NULL)
			/* use canonical name */
			service_name = vio_canonical_service_name(ldp,
			    vsw_attrs.name, cfg_handle);
	}

	if (md_get_prop_val(guest_md, vdev, DEFAULT_VID_PROP,
	    &default_vid) != 0)
		warning("vswitch node missing %s property", DEFAULT_VID_PROP);
	if (md_get_prop_val(guest_md, vdev, PVID_PROP,
	    &pvid) != 0)
		warning("vswitch node missing %s property", PVID_PROP);
	if (md_get_prop_data(guest_md, vdev, VID_PROP, (uint8_t **)&vids,
	    &num_vlans) != 0)
		warning("vswitch node missing %s property", VID_PROP);
	else {
		num_vlans /= sizeof (uint64_t);
		vlans = Xmalloc(num_vlans * sizeof (int));
		for (i = 0; i < num_vlans; i++)
			vlans[i] = (int)vids[i];
	}

	if (md_get_prop_val(guest_md, vdev, MTU_PROP, &mtu) != 0)
		warning("vdevice has no mtu property");

	if (get_vsw_props_from_db(ldp, service_name, &db_mtu) == true) {
		if (db_mtu != -1)
			mtu = db_mtu;
	}

	md_get_prop_val(guest_md, vdev, LINK_PROP, &phys_state);
	if (get_vsw_linkprop_from_db(ldp, service_name, &db_physstate) &&
	    db_physstate != ILLEGAL_LNKP) {
		phys_state = db_physstate;
	}

	if (phys_dev != NULL)
		if (find_by_phys_dev(ldp, (char *)phys_dev) != 0)
			fatal("Multiple vswitches on LDom %s have the same "
			    "physical device %s", ldp->namep, (char *)phys_dev);

	/*
	 * Use the cfg_handle value as a switchid if failed to get the
	 * switchid from the DB
	 */
	get_vsw_id_from_db(ldp, service_name, &switchid);
	if (switchid == NO_NEW_ID)
		switchid = cfg_handle;

	if ((vsw = vsw_new(ldp, service_name, (char *)phys_dev,
	    (mac_addr_t)mac_addr, mode, default_vid, pvid, num_vlans,
	    vlans, mtu, phys_state, switchid)) == NULL)
		fatal("Duplicate virtual switch %s in MDs", service_name);

	(void) bind_vsw(vsw, cfg_handle);

	if (num_vlans > 0)
		vlan_str = comma_sepa_str(num_vlans, vlans);

	database_ldom_add_net_service(ldp, service_name, (mac_addr_t)mac_addr,
	    (char *)phys_dev, phys_state, mode, default_vid, pvid, vlan_str,
	    mtu, switchid);
	Xfree(vlan_str);

	Xfree(service_name);

	/*
	 * Now, scan through the vport list and add a client binding for
	 * each one.
	 */
	nodelistp = Xmalloc(sizeof (md_node_t *) * md_node_count(guest_md));

	num_client_bindings = md_scan_dag(guest_md, vdev,
	    md_find_name(guest_md, "virtual-device-port"),
	    md_find_name(guest_md, "fwd"), nodelistp);

	for (i = 0; i < num_client_bindings; ++i) {
		uint64_t port;
		uint64_t *mac_addr;
		int mac_addr_len;
		char *name;
		vsw_port_t *vsw_port;
		uint64_t hybrid_mode;
		char *vnet_mode = NULL;

		md_node_t *vport = nodelistp[i];

		if (md_get_prop_str(guest_md, vport, "name", &name) != 0)
			fatal("vport has no name property");
		if (md_get_prop_val(guest_md, vport, "id", &port) != 0)
			fatal("vport has no id property");
		if (md_get_prop_data(guest_md, vport, VSW_REMOTE_MAC_PROP,
		    (uint8_t **)&mac_addr, &mac_addr_len) != 0)
			fatal("vport has no %s property", VSW_REMOTE_MAC_PROP);

		if (md_get_prop_val(guest_md, vport, VSW_HYBRID_MODE_PROP,
		    &hybrid_mode) == 0)
			vnet_mode = VIO_MODE_HYBRID;

		vsw_port = vsw_add_port(vsw, NULL, port, vnet_mode);

		ldc_get_and_pull_ids(ldp, &ldp->guest_free_ldcs,
		    guest_md, vport, &vsw_port->ldcs);
	}

	Xfree(nodelistp);
}

void
vsw_gen_ldc_endpoints(ldom_t *ldp, md_t *hv_md, md_node_t *hv_guest,
    md_node_t *ldc_endpoints)
{
	vsw_t *vsw;
	vsw_port_t *vsw_port;

	CHAIN_ITER(ldp->vsws, vsw, ldom) {
		CHAIN_ITER(vsw->binding->vports, vsw_port, link) {
			vio_gen_ldcs_endpoint_nodes(hv_md, hv_guest,
			    ldc_endpoints, TYPE_GUEST, TYPE_GUEST,
			    &vsw_port->ldcs, vsw_port->vnet->ldp->idp->id,
			    &vsw_port->vnet->binding->ldcs, NULL);
		}
	}
}

static void
gen_vsw_guest_md_one(vsw_t *vsw, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vsw_port_t *vsw_port;
	md_node_t *vdev;
	uint8_t *sw_mode;
	int sw_mode_len;
	char *phys_dev;
	uint64_t vsw_mtu = (vsw->mtu != VIO_NONEMTU) ? vsw->mtu : VIO_ETHERMTU;

	vdev = md_link_new_node(md, VDEV_NAME, cdevs, "fwd", "back");

	/*
	 * Compatible property is an array of strings, though we only
	 * currently have one.
	 */
	md_add_string_as_data_property(md, vdev, "compatible",
	    vsw_attrs.compatible);
	md_add_string_property(md, vdev, "device-type", vsw_attrs.device_type);
	md_add_string_property(md, vdev, "name", vsw_attrs.md_name);
	md_add_value_property(md, vdev, "cfg-handle", vsw_devid(vsw));
	md_add_value_property(md, vdev, VSW_LOCAL_MAC_PROP, vsw->mac);
	md_add_string_property(md, vdev, DEVALIAS, vsw->service);

	/*
	 * Vsw-phys-dev property is an array of strings, though we only
	 * currently have one.
	 */
	phys_dev = vsw->phys_dev != NULL ? vsw->phys_dev : "";
	md_add_string_as_data_property(md, vdev, VSW_PHYS_DEV_PROP, phys_dev);

	md_add_value_property(md, vdev, LINK_PROP, vsw->phys_state);

	vsw_get_sw_mode(vsw->phys_dev, &sw_mode_len, &sw_mode, NULL);
	md_add_data_property(md, vdev, VSW_SW_MODE_PROP, sw_mode_len, sw_mode);
	md_add_value_property(md, vdev, DEFAULT_VID_PROP, vsw->default_vid);
	add_vlan_md_props(md, vdev, vsw->pvid, vsw->num_vlans,
	    vsw->vlans, false);

	if (vsw->num_ether_types > 0)
		md_add_data_property(md, vdev, VSW_ETHER_TYPE_PROP,
		    vsw->num_ether_types * sizeof (uint64_t),
		    (uint8_t *)vsw->ether_types);

	md_add_value_property(md, vdev, MTU_PROP, vsw_mtu);

	CHAIN_ITER(vsw->binding->vports, vsw_port, link) {
		md_node_t *port_node;
		uint64_t mac;
		vnet_t *vnet = vsw_port->vnet;

		port_node = vio_gen_vdev_port_node(md, vdev, chan_endpoints,
		    vsw_attrs.port_name, vsw_port->port->id, &vsw_port->ldcs);
		/*
		 * Add remote mac address property. MD spec calls for an array
		 * of mac addresses but we only have one.
		 */
		mac = hton64(vnet->mac);
		md_add_data_property(md, port_node, VSW_REMOTE_MAC_PROP,
		    sizeof (mac), (uint8_t *)&mac);

		/*
		 * Add mode property for hybrid I/O
		 */
		if (vnet->vnet_mode != NULL)
			md_add_value_property(md, port_node,
			    VSW_HYBRID_MODE_PROP, VIO_HYBRID_VAL);
		add_vlan_md_props(md, port_node, vnet->pvid,
		    vnet->num_vlans,
		    vnet->vlans, true);
	}
}

void
vsw_gen_guest_md(ldom_t *ldp, md_t *md, md_node_t *cdevs,
    md_node_t *chan_endpoints)
{
	vsw_t *vsw;

	CHAIN_ITER(ldp->vsws, vsw, ldom) {
		gen_vsw_guest_md_one(vsw, md, cdevs, chan_endpoints);
	}
}

static void
print_vsw_header(vsw_t *vsw, char *servertype, int nclients,
    response_t *resp, dispctl_t *dctl)
{
	char *server = vsw->service;

	if (dctl->header && !dctl->parseable) {
		dctl->header = 0;
		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in tables with fixed
		 * column widths.
		 * WIDTH  HEADER	MEANING
		 * 16	  "NAME"	VIO server name
		 * 16	  "LDOM"	the LDom name iff list-services
		 * 17	  "MAC"		MAC address
		 * 9	  "NET-DEV"	network device name
		 * 4	  "ID"		network id
		 * 9	  "DEVICE"	disk or network device name
		 * 10	  "LINKPROP"	"phys-state" or ""
		 * 15	  "DEFAULT-VLAN-ID" default vlan id
		 * 4	  "PVID"	port vlan id
		 * 20	  "VID"		vlan id list
		 * 20	  "MTU"		MTU
		 * 6	  "MODE"	virtual switch mode
		 */
		rprintf(resp, "    "FNAME, gettext("NAME"));
		if (dctl->services)
			rprintf(resp, " "FNAME, gettext("LDOM"));
		rprintf(resp, " "FMAC" "FNETDEV" "FID" "FDEVICE" "FLNKP" "
		    FDEFVID" "FPVID" "FVID" "FMTU" "FMODE"\n", gettext("MAC"),
		    gettext("NET-DEV"), gettext("ID"), gettext("DEVICE"),
		    gettext("LINKPROP"), gettext("DEFAULT-VLAN-ID"),
		    gettext("PVID"), gettext("VID"), gettext("MTU"),
		    gettext("MODE"));
	}

	print_server_header(server, servertype, nclients, resp, dctl);
}

static void
vsw_port_list(vsw_port_t *vsw_port, response_t *resp, dispctl_t *dctl)
{
	char mac[MAC_STR_LEN];
	char *peer, *fmt;
	vnet_t *vnet = vsw_port->vnet;
	char *vlan_str = NULL;
	char *p_vlan_str;
	int vsw_mtu = (vsw_port->vsw->mtu != VIO_NONEMTU) ? vsw_port->vsw->mtu :
	    VIO_ETHERMTU;
	int mtu = (vnet->mtu != VIO_NONEMTU) ? vnet->mtu : vsw_mtu;

	mactoa(vsw_port->vnet->mac, mac);

	peer = zsprintf("%s@%s", vsw_port->vnet->name,
	    vsw_port->vnet->ldp->namep);

	if (dctl->parseable)
		fmt = "|peer=%s|mac-addr=%s|pvid=%d|vid=%s|mtu=%d";
	else
		fmt = "        "FBNDNAME" "FMAC" "FPVIDd" "FVID" "FMTUd;

	if (vnet->num_vlans > 0)
		vlan_str = comma_sepa_str(vnet->num_vlans, vnet->vlans);
	p_vlan_str = (vlan_str == NULL) ? "" : vlan_str;

	rprintf(resp, fmt, peer, mac, vnet->pvid, p_vlan_str, mtu);
	Xfree(peer);
	Xfree(vlan_str);

	if (dctl->extended) {
		if (dctl->parseable)
			rprintf(resp, "|ldc=");
		else
			rprintf(resp, " ");
		ldcs_print(&vsw_port->ldcs, resp);
	}

	rprintf(resp, "\n");
}

static void
vsw_list_one(vsw_t *vsw, response_t *resp, dispctl_t *dctl)
{
	char macstr[MAC_STR_LEN];
	char *devinst, *devpath;
	int nclients;
	vsw_port_t *vsw_port;
	char *vlan_str = NULL;
	char *p_vlan_str;
	char *mtu_str;
	char *id_str;
	char *pstate_str = (vsw->phys_state == 0x1) ? "phys-state" : "";
	char *mode = (vsw->mode != NULL) ? vsw->mode : "";
	int vsw_mtu = (vsw->mtu != VIO_NONEMTU) ? vsw->mtu : VIO_ETHERMTU;

	if (vsw->binding != NULL) {
		nclients = CHAIN_LENGTH(vsw->binding->vports);
		devinst = zsprintf("switch@%"PRIu64"", vsw_devid(vsw));
	} else {
		nclients = 0;
		devinst = Xstrdup("");
	}

	devpath = (vsw->phys_dev != NULL) ? vsw->phys_dev : "";

	print_vsw_header(vsw, vsw_attrs.name, nclients, resp, dctl);

	if (dctl->constraints && is_mac_allocated(vsw->mac))
		strcpy(macstr, "");
	else
		mactoa(vsw->mac, macstr);

	if (dctl->constraints && vsw->mtu == VIO_NONEMTU)
		mtu_str = zsprintf("");
	else
		mtu_str = zsprintf("%d", vsw_mtu);

	if (vsw->num_vlans > 0)
		vlan_str = comma_sepa_str(vsw->num_vlans, vsw->vlans);
	p_vlan_str = (vlan_str == NULL) ? "" : vlan_str;

	if (vsw->switchid != NO_NEW_ID)
		id_str = zsprintf("%d", vsw->switchid);
	else
		id_str = zsprintf("");

	if (dctl->parseable) {
		rprintf(resp, "|mac-addr=%s|net-dev=%s|dev=%s|default-vlan-id="
		    "%d|pvid=%d|vid=%s|mode=%s|mtu=%s|linkprop=%s", macstr,
		    devpath, devinst, vsw->default_vid, vsw->pvid, p_vlan_str,
		    mode, mtu_str, pstate_str);
		if (dctl->services)
			rprintf(resp, "|ldom=%s", vsw->ldp->namep);
		rprintf(resp, "|id=%s", id_str);
	} else {
		if (dctl->services)
			rprintf(resp, " "FNAME, vsw->ldp->namep);
		rprintf(resp, " "FMAC" "FNETDEV" "FID" "FDEVICE" "FLNKP" "
		    FDEFVIDd" " FPVIDd" " FVID" "FMTU" "FMODE, macstr, devpath,
		    id_str, devinst, pstate_str, vsw->default_vid, vsw->pvid,
		    p_vlan_str, mtu_str, mode);
	}

	Xfree(devinst);
	Xfree(vlan_str);
	Xfree(mtu_str);

	rprintf(resp, "\n");

	if (!dctl->bindings || (nclients == 0))
		return;

	if (!dctl->parseable) {
		dctl->header = 1;
		/*
		 * TRANSLATION_NOTE
		 * The following are column headers in tables with fixed
		 * column widths.
		 * WIDTH  HEADER	MEANING
		 * 27	  "PEER"	name of peer network endpoint
		 * 17	  "MAC"		MAC address
		 * 3	  "LDC"		Logical Domain Channel number
		 */
		rprintf(resp, "        "FBNDNAME" "FMAC" "FPVID" "FVID
		    " "FMTU" "FLNKP, gettext("PEER"), gettext("MAC"),
		    gettext("PVID"), gettext("VID"), gettext("MTU"),
		    gettext("LINKPROP"));
		if (dctl->extended)
			rprintf(resp, " %s", gettext("LDC"));
		rprintf(resp, "\n");
	}

	CHAIN_ITER(vsw->binding->vports, vsw_port, link) {
		vsw_port_list(vsw_port, resp, dctl);
	}
}

void
vsw_list(ldom_t *ldp, response_t *resp, dispctl_t *dctl)
{
	vsw_t *vsw;

	if (dctl->oformat && !dctl->onetwork)
		return;

	if (CHAIN_LENGTH(ldp->vsws) == 0)
		return;

	dctl->header = 1;

	if (!dctl->parseable)
		rprintf(resp, "%s\n", dctl_label(vsw_attrs.name, dctl));

	CHAIN_ITER(ldp->vsws, vsw, ldom) {
		vsw_list_one(vsw, resp, dctl);
	}

	if (!dctl->parseable)
		rprintf(resp, "\n");
}

vsw_port_t *
vsw_port_find_by_ldcid(ldom_t *ldp, uint64_t ldcid)
{
	vsw_t *vsw;
	vsw_port_t *vport;

	CHAIN_ITER(ldp->vsws, vsw, ldom) {
		if (vsw->binding == NULL)
			continue;
		CHAIN_ITER(vsw->binding->vports, vport, link) {
			if (ldcs_search(&vport->ldcs, ldcid) != NULL)
				return (vport);
		}
	}

	return (NULL);
}

void
vsw_destroy_port(vsw_port_t *vsw_port)
{
	CHAIN_UNLINK(vsw_port->vsw->binding->vports, vsw_port, link);

	if (vsw_port->port != NULL)
		free_res(vsw_port->port);
	Xfree(vsw_port->work.vnet_mode);

	ldcs_free(&vsw_port->ldcs);
	Xfree(vsw_port);
}

void
vsw_delete_all(ldom_t *ldp)
{
	vsw_t *vsw, *vsw_next;

	for (vsw = CHAIN_START(ldp->vsws); vsw != NULL; vsw = vsw_next) {
		vsw_next = CHAIN_NEXT(vsw->ldom);
		destroy_vsw(vsw);
	}
}

void
vsw_unbind_port(vsw_port_t *vsw_port)
{
	ldcs_stage(&vsw_port->ldcs);
	vsw_destroy_port(vsw_port);
}

/*
 * Called from the add-vswitch CLI.
 */
int
vsw_add(ldom_t *ldp, char *service_name, char *dev_path, mac_addr_t mac_address,
    char *mode, int default_vid, int pvid, char *vids, int mtu,
    uint64_t phys_state, int switchid, response_t *resp)
{
	vsw_t *vsw;
	int num_vlans;
	int rc = 0;
	bool_t is_vid = (vids != NULL) ? true : false;
	int *vlans = parse_num_list(vids, &num_vlans);
	int i;

	if (is_vid == true) {
		if ((num_vlans == 1) && (vlans == NULL)) {
			rprintf(resp, gettext("Must specify vlan ids\n"));
			rc = ERR_NET_VID_INVALID;
			goto ret;
		} else if (vlans == NULL) {
			rprintf(resp, gettext("vid incorrectly specified\n"));
			rc = ERR_NET_VID_INVALID;
			goto ret;
		}
	}

	if (default_vid == -1)
		default_vid = DEFAULT_VLAN_ID;
	/*
	 * default-vlan-id value is already checked in cli front end
	 */

	if (pvid == -1)
		pvid = default_vid;

	if (mtu == VIO_NOARGSUP)
		mtu = VIO_NONEMTU;

	/*
	 * pvid value is already checked in cli front end
	 */

	if (vid_check(resp, num_vlans, vlans) != 0) {
		rc = ERR_NET_VID_INVALID;
		goto ret;
	}

	if (vsw_find_by_name(service_name) != NULL) {
		rprintf(resp, gettext("A %s service named %s already exists\n"),
		    vsw_attrs.name, service_name);
		rc = ERR_SERVICE_EXISTS;
		goto ret;
	}

	if (dev_path != NULL) {
		if ((vsw = find_by_phys_dev(ldp, dev_path)) != NULL) {
			/* already in use */
			ASSERT(vsw->service != NULL);
			rprintf(resp,
			    gettext("The device path %s for virtual switch "
			    "%s\nis already in use by virtual switch %s\n"),
			    dev_path, service_name, vsw->service);
			rc = ERR_VSW_DEV_PATH_EXISTS;
			goto ret;
		}
	}

	/*
	 * pvid or default-vlan-id must not match any of the vid specified
	 */
	if (vlans != NULL) {
		for (i = 0; i < num_vlans; i++) {
			if ((vlans[i] == pvid) || (vlans[i] == default_vid)) {
				rprintf(resp, gettext("vid must not match "
				    "the pvid or the default-vlan-id\n"));
				rc = ERR_NET_VID_INVALID;
				goto ret;
			}
		}
	}

	/*
	 * if id is not specified in the CLI or the domain is still inactive,
	 * get the lowest free id value from the DB to use as a switchid
	 */
	if (switchid == NO_NEW_ID) {
		switchid = database_ldom_get_lowest_id(ldp, VSW_NODE);
	} else if (database_ldom_check_dup_id(ldp, VSW_NODE, switchid)
	    == true) {
		/*
		 * make sure that the same id doesn't already exist in the DB
		 */
		rprintf(resp, MSG_ERR_ID_EXISTS);
		return (ERR_DUP_ID);
	}

	vsw = vsw_new(ldp, service_name, dev_path, mac_address, mode,
	    default_vid, pvid, num_vlans, vlans, mtu, phys_state, switchid);

	if (ldp->state >= BOUND) {
		/*
		 * if the switchid is specified with the CLI, the switchid value
		 * will be used to pull from the cfg_handle pool. If it fails to
		 * get the cfg_handle, generate the duplicate id error message.
		 */
		if (vsw->switchid != NO_NEW_ID) {
			rc = bind_vsw(vsw, vsw->switchid);
			if (rc != 0) {
				if (rc == ERR_VSW_LINKPROP_VSW) {
					rprintf(resp,
					    MSG_ERR_BIND_VSW_NO_NET_DEV);
				} else if (rc == ERR_VSW_LINKPROP_VNET) {
					rprintf(resp,
					    MSG_ERR_BIND_VSW_NO_NET_DEV_VNET);
				} else {
					rprintf(resp, MSG_ERR_ID_EXISTS);
					rc = ERR_DUP_ID;
				}
				unbind_vsw(vsw);
				destroy_vsw(vsw);
				goto ret;
			}
		} else {
			rc = bind_vsw(vsw, NEW_INSTANCE);
			if (rc != 0) {
				if (rc == ERR_VSW_LINKPROP_VSW) {
					rprintf(resp,
					    MSG_ERR_BIND_VSW_NO_NET_DEV);
				} else if (rc == ERR_VSW_LINKPROP_VNET) {
					rprintf(resp,
					    MSG_ERR_BIND_VSW_NO_NET_DEV_VNET);
				}
				unbind_vsw(vsw);
				destroy_vsw(vsw);
				goto ret;
			}
		}

		if (zeus.mode == ZEUS_MGMT_MODE) {
			if (ldp->state == BOUND) {
				/* do reconfig now */
				seq_download_mds_do_config(ldp);
				free_all_staged_ldcs(ldp);

			} else if (vio_check_dr_capable(ldp, resp) == true) {
				rc = vio_dr_add(ldp, vsw_attrs.md_name,
				    vsw_devid(vsw), resp);
				/*
				 * If the DR operation failed and the dr-vio
				 * domain service determines that it is OK to
				 * rollback we remove the vsw again.
				 */
				if (rc != 0) {
					if (rc == ERR_DR_VIO_ROLLBACK) {
						unbind_vsw(vsw);
						destroy_vsw(vsw);
						seq_rm_reconfigure(ldp);
						free_all_staged_ldcs(ldp);
						return (rc);
					}
					/*
					 * Alert ldomconfigure.c that even
					 * though there was an error we kept
					 * the vsw in LDMD memory.
					 */
					return (ERR_DR_VIO_STILL_ALLOCATED);
				}
			} else {
				/* handle as a delayed reconfig operation */
				initiate_or_update_delayed_reconfig(ldp,
				    DEL_RECONFIG_ADD, resp);
			}
		}
	}

	database_ldom_add_net_service(ldp, service_name, mac_address,
	    dev_path, phys_state, mode, default_vid, pvid, vids, mtu,
	    switchid);

ret:
	if (rc != 0)
		Xfree(vlans);

	return (rc);
}

static bool_t
dom_in_list(ldom_t *ldp)
{
	doml_t *domp;

	CHAIN_ITER(domains, domp, link) {
		if (domp->ldp == ldp)
			break;
	}

	return (domp == NULL ? false : true);
}

static void
free_dom_list()
{
	doml_t *domp, *ndomp;

	CHAIN_ITER_VOL(domains, domp, ndomp, link)
		CHAIN_UNLINK(domains, domp, link);
}

/*
 * Print a warning when vsw's mtu is updated
 */
int
print_mtu_upd_warning(vsw_t *vsw, response_t *resp)
{
	char *glist = NULL;
	vsw_port_t *vsw_port;
	int n_vnet = 0;

	CHAIN_ITER(vsw->binding->vports, vsw_port, link) {
		vnet_t *vnet = vsw_port->vnet;
		bool_t own_mtu = (vnet->mtu != VIO_NONEMTU) ? true : false;

		if (vnet->ldp->state == ACTIVE && own_mtu == false) {
			n_vnet++;
			if (dom_in_list(vnet->ldp) == false) {
				doml_t *domp = Xcalloc(1, doml_t);
				domp->ldp = vnet->ldp;

				CHAIN_ADD(domains, domp, link);

				glist = (glist == NULL) ? zsprintf("%s",
				    vnet->ldp->namep) :
				    zsprintf("%s, %s", glist, vnet->ldp->namep);
			}
		}
	}

	free_dom_list();

	if (glist != NULL) {
		rprintf(resp, "Please reboot the guest %s for the changes to "
		    "take effect.\nOne or more vnet on the following guests "
		    "are still using the old MTU and may experience network "
		    "connectivity issues until reboot\n%s\n", vsw->ldp->namep,
		    glist);

		Xfree(glist);
	}

	return (n_vnet);
}

/*
 * Called from the set-vswitch and add-vswitch CLIs.
 */
int
vsw_set(vsw_t *vsw, char *phys_dev, mac_addr_t mac, char *mode, int pvid,
    char *vids, int mtu, uint64_t phys_state, response_t *resp)
{
	ldom_t *ldp = vsw->ldp;
	ldom_t *md_ldp;
	int upd_type = NO_UPD;
	int md_upd = MD_NONE;
	int num_vlans = 0;
	bool_t is_vid = (vids != NULL) ? true : false;
	int *vlans = parse_num_list(vids, &num_vlans);
	int rc = 0;
	int i = 0;
	int n_vlans = (vlans != NULL) ? num_vlans : vsw->num_vlans;
	int *t_vlans = (vlans != NULL) ? vlans : vsw->vlans;
	int mod = 0;

	/*
	 * The front end checks mac_address against ILLEGAL_MAC_ADDR but from
	 * here on the back end code checks against 0.
	 */
	if (mac == ILLEGAL_MAC_ADDR)
		mac = 0;

	if ((mac != 0) && (mac != vsw->mac)) {
		upd_type |= DELAYED_UPD;
		mod |= VSW_MAC;
	}

	if (pvid == -1)
		pvid = vsw->pvid;
	else if (pvid == 0)
		pvid = vsw->default_vid;

	if (mtu == VIO_NOARGSUP)
		mtu = vsw->mtu;

	/*
	 * pvid value is already checked in cli front end
	 */
	if (is_vid == true) {
		if (num_vlans == 0) {
			rprintf(resp, gettext("vid incorrectly specified\n"));
			rc = ERR_NET_VID_INVALID;
			goto ret;
		}
	}

	if (vid_check(resp, num_vlans, vlans) != 0) {
		rc = ERR_NET_VID_INVALID;
		goto ret;
	}

	/*
	 * pvid and default-vlan-id must not match any vid
	 */
	for (i = 0; i < n_vlans; i++) {
		if ((t_vlans[i] == pvid) || (t_vlans[i] == vsw->default_vid)) {
			rprintf(resp, gettext("pvid or default-vlan-id must "
			    "not match any vid\n"));
			rc = ERR_NET_VID_INVALID;
			goto ret;
		}
	}

	/*
	 * Verify that the phys-dev is acceptable and set flags
	 */
	if (phys_dev != NULL) {
		if (strcmp(phys_dev, "") == 0) {
			/* empty string means no device */
			if (vsw->phys_dev != NULL) {
				upd_type |= DELAYED_UPD;
				md_upd |= MD_SELF;
				mod |= VSW_NET_DEV;
			}
		} else if ((vsw->phys_dev == NULL) ||
		    (strcmp(vsw->phys_dev, phys_dev) != 0)) {
			/* differs from existing device */
			vsw_t *v;
			if ((v = find_by_phys_dev(ldp, phys_dev)) != NULL) {
				rprintf(resp, gettext("The device path %s is "
				    "already in use by virtual switch %s\n"),
				    phys_dev, v->service);
				rc = ERR_VSW_DEV_PATH_EXISTS;
				goto ret;
			} else {
				upd_type |= DELAYED_UPD;
				md_upd |= MD_SELF;
				mod |= VSW_NET_DEV;
			}
		}
	}

	/*
	 * Set appropriate flags if mode is changed
	 */
	if ((mode != NULL) &&
	    ((vsw->mode == NULL) || (strcmp(mode, vsw->mode) != 0))) {
		upd_type |= DELAYED_UPD;
		md_upd |= MD_SELF;
		mod |= VSW_MODE;
	}

	if ((mac != 0) && (vsw->mac != mac)) {
		upd_type |= DYNAMIC_UPD;
		mod |= VSW_MAC;

		/*
		 * Mark all bound clients of the switch as needing an updated
		 * MD. This is because the switch MAC address is stored in the
		 * client MD.
		 */
		md_upd |= (MD_SELF|MD_PEERS);
	}

	/*
	 * The dynamic update can take place only if the attributes that need
	 * delayed reconfiguration are not being modified. For example, if the
	 * mac address is being modified in addition to the pvid, this must be
	 * a delayed reconfiguration and not a runtime update
	 */

	if (((pvid != 0) && (vsw->pvid != pvid)) ||
	    (num_vlans > 0)) {
		upd_type |= DYNAMIC_UPD;
		md_upd |= (MD_SELF|MD_PEERS);
		mod |= VSW_PVID;
	}

	if (vsw->mtu != mtu) {
		if ((mtu == VIO_NONEMTU && vsw->mtu == VIO_ETHERMTU) ||
		    (mtu == VIO_ETHERMTU && vsw->mtu == VIO_NONEMTU)) {
			mod |= VSW_MTU;
			/*
			 * No change is needed in MD
			 */
		} else {
			mod |= VSW_MTU;
			md_upd |= (MD_SELF|MD_PEERS);
#if LIVE_SET_MTU
			upd_type |= DYNAMIC_UPD;
#else
			/*
			 * Allow delayed reconfiguration only if no vnet is
			 * bound
			 */
			if (ldp->state >= BOUND && vsw->binding != NULL) {
				/*
				 * Print a warning about guests with vnet
				 * having derived mtu from vsw. We cannot
				 * use delayed reconfig in the case where
				 * there are dependent guests.
				 */
				if (print_mtu_upd_warning(vsw, resp) > 0)
					upd_type |= DYNAMIC_UPD;
				else
					upd_type |= DELAYED_UPD;
			}
#endif
		}
	}

	if (phys_state == ILLEGAL_LNKP) {
		phys_state = vsw->phys_state;
	}
	if (vsw->phys_state != phys_state) {
		mod |= VSW_PSTATE;
		md_upd |= MD_SELF;
		upd_type |= DYNAMIC_UPD;
	}

	/*
	 * No changes to the config need to be made - any options
	 * specified match the existing value(s).
	 */
	if (upd_type == NO_UPD) {
		rc = 0;
		goto ret;
	}

	if ((upd_type & DELAYED_UPD) != 0 && config_mod_allowed(ldp,
	    DEL_RECONFIG_OTHER, DR_OP_TYPE_VIO_SET, resp) == false) {
		rc = ERR_NO_DELRECONF_OTHER;
		goto ret;
	}

	/*
	 * If the OTHER_UPD flag is set, there is a request to modify an
	 * argument which isn't supported when the LDom is bound (e.g. mtu)
	 */
	if ((ldp->state >= BOUND) && (zeus.mode == ZEUS_MGMT_MODE)) {
		if ((upd_type & OTHER_UPD) != 0) {
			rprintf(resp, gettext("These vsw modifications are not "
			    "supported while LDom %s is bound\n"), ldp->namep);
			rc = ERR_LDOM_IS_BOUND;
			goto ret;
		}
	}

	/*
	 * Update Mac
	 */
	if ((mac != 0) && (vsw->mac != mac)) {
		vsw->mac = mac;
	}

	/*
	 * Update phys-dev
	 */
	if (phys_dev != NULL) {
		if (strcmp(phys_dev, "") == 0) {
			/* Make sure vnet phy_state is not left dangling */
			if (vsw->ldp->state >= BOUND &&
			    vnet_num_linkprop_set(vsw) > 0) {
				rprintf(resp, MSG_ERR_VNET_LINKPROP);
				rc = ERR_VSW_LINKPROP_VNET;
				goto ret;
			}
			/* empty string means no device */
			if (vsw->phys_dev != NULL) {
				Xfree(vsw->phys_dev);
				mod |= VSW_NET_DEV;
				/* impacts phys_state */
				vsw->phys_dev = NULL;
			}
		} else if ((vsw->phys_dev == NULL) ||
		    (strcmp(vsw->phys_dev, phys_dev) != 0)) {
			Xfree(vsw->phys_dev);
			mod |= VSW_NET_DEV;
			vsw->phys_dev = Xstrdup(phys_dev);
		}
	}

	/*
	 * Make sure vsw phys_state is not left dangling
	 */
	if (phys_state == 0x1) {
		/* Linkprop will remain or will become set */
		if (vsw->ldp->state >= BOUND && vsw->phys_dev == NULL) {
			rprintf(resp, MSG_ERR_VSW_LINKPROP);
			rc = ERR_VSW_LINKPROP_VSW;
			goto ret;
		}
	}

	/*
	 * Update mode
	 */
	if ((mode != NULL) &&
	    ((vsw->mode == NULL) || (strcmp(mode, vsw->mode) != 0))) {
		/* mode has changed, update it and the priority ether types */
		Xfree(vsw->mode);
		Xfree(vsw->ether_types);
		vsw->mode = (strcmp(mode, "") != 0) ? Xstrdup(mode) : NULL;
		set_ether_types_from_mode(vsw, mode);
	}

	/*
	 * Update pvid, vid & mtu
	 */
	if ((pvid != 0) && (vsw->pvid != pvid))
		vsw->pvid = pvid;

	if (num_vlans > 0) {
		Xfree(vsw->vlans);
		vsw->vlans = vlans;
		vsw->num_vlans = (vlans == NULL) ? 0 : num_vlans;
	}

	if (mod & VSW_MTU)
		vsw->mtu = mtu;

	if (mod & VSW_PSTATE)
		vsw->phys_state = phys_state;

	/* update the DB with the new properties */
	database_ldom_modify_net_service(ldp, mod, vsw->service, mac,
	    phys_dev, vsw->phys_state, mode, pvid, num_vlans,
	    (vlans == NULL) ? NULL : vids, vsw->mtu);

	if (vsw->binding == NULL)
		/*
		 * If the vsw is not bound, there is nothing more to do
		 */
		goto ret;

	if ((md_upd & MD_SELF) != 0) {
		md_ldp = vsw->ldp;
		md_ldp->flag.gen_new_md = true;
		md_ldp->flag.md_notify = true;
	}

	if ((md_upd & MD_PEERS) != 0) {
		vsw_port_t *vsw_port;
		CHAIN_ITER(vsw->binding->vports, vsw_port, link) {
			vnet_t *vnet = vsw_port->vnet;
			/*
			 * special case: in case of MTU only
			 * modification, the vnet peer does not need
			 * updated MD if it has it's own mtu
			 */
			if ((mod & VSW_MTU) && ((mod ^ VSW_MTU) == 0) &&
			    (vnet->mtu != VIO_NONEMTU))
				continue;
			vnet->ldp->flag.gen_new_md = true;
			vnet->ldp->flag.md_notify = true;
		}
	}

	if (ldp->state >= BOUND && zeus.mode == ZEUS_MGMT_MODE) {
		if (ldp->state != BOUND && ((upd_type & DELAYED_UPD) != 0)) {
			/*
			 * Handle as a delayed reconfig, even if there
			 * isn't one already in progress.  FIXME: this is
			 * a workaround to prevent a potential Solaris
			 * panic in the vsw module.
			 */
			initiate_or_update_delayed_reconfig(ldp,
			    DEL_RECONFIG_OTHER, resp);
		} else {
			/* do reconfig now */
			seq_download_mds_do_config(ldp);
			free_all_staged_ldcs(ldp);
		}
	}

	if (zeus.mode == ZEUS_CONFIG_MODE)
		free_all_staged_ldcs(ldp);

ret:
	if (rc != 0)
		Xfree(vlans);

	return (rc);
}

static void
vsw_remove_do_teardown(vsw_t *vsw)
{
	database_ldom_del_net_service(vsw->ldp, vsw->service);
	unbind_vsw(vsw);
	destroy_vsw(vsw);
}

/*
 * Called from the remove-vswitch CLI.
 */
/*ARGSUSED1*/
int
vsw_remove(vsw_t *vsw, bool_t force, response_t *resp)
{
	ldom_t *ldp = vsw->ldp;
	int	rv;

	if (vsw->binding != NULL) {
		if (CHAIN_LENGTH(vsw->binding->vports) > 0) {
			/* clients are bound to the switch */
			rprintf(resp, MSG_ERR_SERVICE_BOUND_CLIENTS);
			return (ERR_SERVICE_BOUND_CLIENTS);
		}
	}

	switch (zeus.mode) {
	case ZEUS_MGMT_MODE:
		if (ldp->state < BOUND) {
			vsw_remove_do_teardown(vsw);

		} else if (ldp->state == BOUND) {
			/* do reconfig now */
			vsw_remove_do_teardown(vsw);
			seq_download_mds_do_config(ldp);
			free_all_staged_ldcs(ldp);

		} else if (vio_check_dr_capable(ldp, resp)) {
			/* Attempt DR operation */
			rv = vio_dr_rm_phase1(ldp, vsw_attrs.md_name,
			    vsw_devid(vsw), resp);
			if (rv != 0)
				return (rv);
			vsw_remove_do_teardown(vsw);
			vio_dr_rm_phase2(ldp);

		} else {
			/* handle as a delayed reconfig operation */
			vsw_remove_do_teardown(vsw);
			initiate_or_update_delayed_reconfig(ldp,
			    DEL_RECONFIG_REMOVE, resp);
		}
		break;

	case ZEUS_CONFIG_MODE:
		vsw_remove_do_teardown(vsw);
		free_all_staged_ldcs(ldp);

		break;
	default:
		ASSERT(0);
		return (ERR_GENERIC);
	}

	return (0);
}
