/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

/*
 * Utility functions for managing CLI clients waiting on
 * LDom state changes.
 */

#include "allocate.h"
#include "basics.h"
#include "chain.h"
#include "cli_conn.h"
#include "clp.h"
#include "fatal.h"
#include "ldom.h"
#include "sequencer.h"
#include "waiter.h"
#include "xml_cntrl_common.h"
#include "xmpp_server.h"

waiter_t *
new_waiter(ldom_t *ldom, cli_conn_t *client, int event, ldm_cmd_num_t cmd)
{
	waiter_t *w;

	w = Xmalloc(sizeof (*w));

	w->ldom = ldom;
	w->client = client;
	w->event = event;
	w->cmd = cmd;
	CHAIN_LINK_INIT(w->next_ldom);
	CHAIN_LINK_INIT(w->next_client);

	return (w);
}

/*
 * Add a dependency that CLI client 'client' is waiting for 'event' to occur
 * on LDom 'ldom'.
 *
 * No locking is required as this is to be called from within the sequencer
 * only.
 */
void
add_waiter(ldom_t *ldom, cli_conn_t *client, int event, ldm_cmd_num_t cmd)
{
	waiter_t *w;

	ASSERT_IN_SEQUENCER();

	w = new_waiter(ldom, client, event, cmd);

	CHAIN_ADD(ldom->waiters, w, next_client);
	CHAIN_ADD(client->waiters, w, next_ldom);
}

/*
 * Release all client dependencies on 'event' occuring on LDom 'ldom'.  If
 * this results in a client now having no events left to wait on then send the
 * client's response so long as it not client 'client'.
 *
 * No locking is required as this is to be called from within the sequencer
 * only.
 */
void
release_waiters(ldom_t *ldom, cli_conn_t *client, int event)
{
	waiter_t *w, *nextw;

	ASSERT_IN_SEQUENCER();

	for (w = CHAIN_START(ldom->waiters); w != NULL; w = nextw) {
		nextw = CHAIN_NEXT(w->next_client);
		if ((w->event & event) != 0) {
			if (w->cmd == LDM_STOP) {
				rprintf(w->client->resp, "LDom %s stopped\n",
				    w->ldom->namep);
			} else if (w->cmd == LDM_START) {
				if (event == BOUND) {
					if (client->class ==
					    LDM_CLASS_CONTROL) {
						xml_update_status(client,
						    w->cmd, w->ldom->namep,
						    XML_FAILURE, "LDom %s "
						    "start failed, retry "
						    "the operation",
						    w->ldom->namep);
					} else {
						rprintf(w->client->resp,
						    "LDom %s start failed, "
						    "retry the operation\n",
						    w->ldom->namep);
					}
				} else {
					rprintf(w->client->resp,
					    "LDom %s started\n",
					    w->ldom->namep);
				}
			} else {
				fatal("Unexpected waiter command %d", w->cmd);
			}
		}

		CHAIN_UNLINK(w->ldom->waiters, w, next_client);
		CHAIN_UNLINK(w->client->waiters, w, next_ldom);

		if (CHAIN_LENGTH(w->client->waiters) == 0) {
			if (w->client->callout)
				seq_delete_client_callout(w->client);
			if (w->client != client) {
				if (w->client->class == LDM_CLASS_CONTROL) {
					if (!w->client->xmpp_msg) {
						client_write_buffer(w->client,
						    w->client->bufp);
					} else {
						xmpp_write_buffer(w->client,
						    false);
					}
				} else {
					client_write_response(w->client);
				}
			}
		}

		Xfree(w);
	}
}

/*
 * Free all all of client 'client's dependencies.
 *
 * No locking is required as this is to be called from within the sequencer
 * only.
 */
void
free_clients_waiters(cli_conn_t *client)
{
	waiter_t *w, *nextw;

	ASSERT_IN_SEQUENCER();

	for (w = CHAIN_START(client->waiters); w != NULL; w = nextw) {
		nextw = CHAIN_NEXT(w->next_ldom);
		CHAIN_UNLINK(w->client->waiters, w, next_ldom);
		CHAIN_UNLINK(w->ldom->waiters, w, next_client);
		Xfree(w);
	}
}
