/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <string.h>
#include <sys/varargs.h>

#include "control_class.h"
#include "allocate.h"
#include "support.h"
#include "zprintf.h"
#include "fatal.h"
#include "ldomconfigure.h"
#include "xml_common.h"
#include "xml_cntrl_common.h"
#include "database_internal.h"
#include "xml_parse.h"
#include "xml_v3_defs.h"

static prot_version_t xml_data_versions[] = {
	{ 1, 0 },
	{ 2, 0 },
	{ 3, 0 }
};

static xmlNodePtr find_cmd_section(char *cmd_name, xmlNodePtr root);

xmlNodePtr
xml_find_data_section(xmlNodePtr cmd, char *name)
{
	xmlNodePtr	subnode, envelope, content;
	char		*ldm_name;

	subnode = xml_find_subnode(cmd, XML_DATA);
	while ((subnode != NULL) &&
	    (xmlStrcmp(subnode->name, XML_DATA) == 0)) {
		envelope = xml_find_subnode(subnode, ENVELOPE_NODE);
		content = xml_find_subnode(envelope, CONTENT_NODE);
		ldm_name = (char *)xmlGetProp(content, ID_ATTR);
		if (ldm_name == NULL) {
			ldm_name = (char *)xmlGetProp(content, ID_ATTR_IN);
		}

		if (strcmp(ldm_name, name) == 0) {
			xmlFree(ldm_name);
			return (subnode);
		}
		xmlFree(ldm_name);
		subnode =  xml_get_next_ele_node(subnode);
	}
	return (NULL);
}

xmlNodePtr
xml_find_section_subnode(xmlNodePtr section, const xmlChar *name)
{
	xmlNodePtr subnode, item;

	item = xml_find_subnode(section, ITEM_NODE);
	if (item == NULL)
		return (NULL);
	subnode = item->xmlChildrenNode;
	while (subnode != NULL) {
		if (xmlStrcmp(subnode->name, name) == 0)
			break;
		subnode = subnode->next;
	}

	return (subnode);
}

xmlChar *
get_section_resource_name(xmlNodePtr section)
{
	xmlNodePtr	item, resource_name;

	item = xml_find_subnode(section, ITEM_NODE);
	if (item == NULL)
		return (NULL);
	resource_name = xml_find_subnode(item, OTHER_RESOURCE_NODE_IN);
	if (resource_name == NULL)
		return (NULL);
	return (xmlNodeGetContent(resource_name));
}

xmlNodePtr
xml_find_v3_property(xmlNodePtr item, xmlChar *prop_name)
{
	xmlNodePtr	subnode;
	xmlChar		*prop_key;

	subnode = item->xmlChildrenNode;
	while (subnode != NULL) {
		prop_key = xmlGetProp(subnode, KEY_ATTR);
		if (prop_key == NULL) {
			subnode = subnode->next;
			continue;
		}
		if (xmlStrcmp(prop_key, prop_name) == 0) {
			xmlFree(prop_key);
			break;
		}
		xmlFree(prop_key);
		subnode = subnode->next;
	}
	return (subnode);
}

/*
 * Find and return the first <Section> which has the resource type  'name'.
 */
xmlNodePtr
xml_find_section(xmlNodePtr node, const xmlChar *name)
{
	xmlNodePtr	section;
	xmlChar		*section_name;

	section = node->xmlChildrenNode;
	while (section != NULL) {
		section_name = get_section_resource_name(section);
		if ((section_name != NULL) &&
		    (xmlStrcmp(section_name, name) == 0)) {
			free(section_name);
			break;
		}
		if (section_name != NULL)
			free(section_name);
		section = section->next;
	}

	return (section);
}

static xmlNodePtr
find_cmd_section(char *cmd_name, xmlNodePtr root)
{
	xmlNodePtr	cmd, action_node;
	char		*action;

	cmd = xml_find_subnode(root, XML_CMD);
	while (cmd != NULL) {
		action_node = xml_find_subnode(cmd, XML_ACTION);
		action = (char *)xmlNodeGetContent(action_node);
		if (strcmp(action, cmd_name) == 0) {
			xmlFree((xmlChar *)action);
			break;
		}
		cmd = xml_get_next_ele_node(cmd);
		xmlFree((xmlChar *)action);
	}

	return (cmd);
}

xmlNodePtr
find_data_section(char *ldm_name, xmlNodePtr cmd)
{
	xmlNodePtr	data;
	char		*name;
	int		version, ret;

	data = xml_find_subnode(cmd, XML_DATA);
	ret = version_check(data, cmd, &version);
	if (ret == -1) {
		xml_add_status(cmd, XML_FAILURE, NULL);
		return (NULL);
	}
	if (version >= 3) {
		return (xml_find_data_section(cmd, ldm_name));
	}
	while (data != NULL) {
		name = get_name_from_data(data);
		if (strcmp(name, ldm_name) == 0) {
			xmlFree((xmlChar *)name);
			break;
		}
		data = xml_get_next_ele_node(data);
		xmlFree((xmlChar *)name);
	}

	return (data);
}

/*
 * This routine updates the status for both a data section and the XML
 * document and adds a resp_msg for the data section of a given ldom. It will
 * only get called if a callout fails. Thus the only proper value for
 * status_str will be XML_FAILURE.
 */
void
xml_update_status(cli_conn_t *client, int cmd, char *ldm_name,
    char *status_str, char *resp_msg_str, ...)
{
	ldm_op_req_t 	*req;
	xmlDocPtr	ret_doc, new_return;
	xmlNodePtr	root, data, resp, status;
	xmlNodePtr	cmd_node;
	va_list		args;
	char		*buf, *content;
	char		*xml_buf, *pkt_buf;
	int		xml_buf_len;

	if (!client->xmpp_msg) {
		req = (ldm_op_req_t *)(client->bufp + sizeof (ldm_hdr_t));
		/* Convert the req->rq_data to XML format */
		ret_doc = xmlParseMemory((const char *)req->rq_data,
		    req->rq_len);
	} else {
		ret_doc = xmlParseMemory(client->buf, client->buf_size);
		(void) memset(client->buf, 0, client->buf_size);
		client->buf_size = 0;
	}
	/*
	 * Now copy the XML document so that we don't overwrite memory
	 * allocated originally for the client_bufp.
	 */
	new_return = xmlCopyDoc(ret_doc, 1);
	/* Free up the packet buffer with the old XML document */
	Xfree(client->bufp);

	root = xmlDocGetRootElement(new_return);
	if (cmd == LDM_START) {
		cmd_node = find_cmd_section((char *)XML_START_DOMAIN, root);
	} else if (cmd == LDM_STOP) {
		cmd_node = find_cmd_section((char *)XML_STOP, root);
	} else {
		DBG(zprintf("Cmd sent in was %d\n", cmd));
		return;
	}
	ASSERT(cmd_node != NULL);
	data = find_data_section(ldm_name, cmd_node);

	if (data == NULL)
		DBG(zprintf("Did not find <data> section for %s\n", ldm_name));

	resp = xml_find_subnode(data, XML_RESPONSE);
	status = xml_find_subnode(resp, XML_STATUS);
	xmlNodeSetContent(status, (xmlChar *)status_str);

	if (resp_msg_str != NULL) {
		va_start(args, resp_msg_str);
		buf = zvsprintf(resp_msg_str, args);
		va_end(args);
		xmlNewChild(resp, NULL, XML_RESP_MSG, (xmlChar *)buf);
		Xfree(buf);
	}

	/* Update the cmd response section */
	resp = xml_find_subnode(cmd_node, XML_RESPONSE);
	status = xml_find_subnode(resp, XML_STATUS);

	content = (char *)xmlNodeGetContent(status);
	if (strcmp(content, (char *)XML_FAILURE) != 0)
		xmlNodeSetContent(status, (xmlChar *)status_str);
	xmlFree(content);

	/* Update the root document response section */
	resp = xml_find_subnode(root, XML_RESPONSE);
	status = xml_find_subnode(resp, XML_STATUS);

	content = (char *)xmlNodeGetContent(status);
	if (strcmp(content, (char *)XML_FAILURE) != 0)
		xmlNodeSetContent(status, (xmlChar *)status_str);
	xmlFree(content);

	(void) xmlDocDumpMemory(new_return, (xmlChar **)&xml_buf,
	    &xml_buf_len);

	if (!client->xmpp_msg) {
		create_pkt_buf(&pkt_buf, xml_buf, xml_buf_len,
		    LDM_CNTRL_XML_RESP);

		client->buf = client->bufp = (void *)pkt_buf;
	} else {
		client->buf = client->bufp = Xmalloc(xml_buf_len);
		(void) memcpy(client->buf, xml_buf, xml_buf_len);
		client->buf_size = xml_buf_len;
	}

	xmlFree(xml_buf);
	xmlFreeDoc(new_return);
}

/*
 * Adds a <Response> node with the <Status> subnode filled in and <Resp_Msg>
 * subnodes filled in if it is supplied.
 */
void
xml_add_status(xmlNodePtr parent, char *status_str, char *resp_msg_str, ...)
{
	va_list		args;
	char		*buf, *old_str;
	xmlNodePtr	resp, status;

	resp = xml_find_subnode(parent, XML_RESPONSE);
	if (resp == NULL) {
		resp = xmlNewChild(parent, NULL, XML_RESPONSE, NULL);
		xmlNewChild(resp, NULL, XML_STATUS, (xmlChar *)status_str);
	} else {
		/*
		 * If there is already a status then we only need to change the
		 * status if it is going from success to failure.
		 */
		status = xml_find_subnode(resp, XML_STATUS);
		old_str = (char *)xmlNodeGetContent(status);
		if ((strcmp(status_str, XML_FAILURE) == 0) &&
		    (strcmp(old_str, XML_SUCCESS) == 0)) {
			xmlNodeSetContent(status, (xmlChar *)status_str);
		}
		xmlFree(old_str);
	}

	if (resp_msg_str != NULL) {
		va_start(args, resp_msg_str);
		buf = zvsprintf(resp_msg_str, args);
		va_end(args);
		xmlNewChild(resp, NULL, XML_RESP_MSG, (xmlChar *)buf);
		Xfree(buf);
	}
}

char *
xml_parse_resp(char *text)
{
	char	*curstr;
	char	*usage_start;
	char	*resp_msg;
	int	err_len;

	curstr = text;

	if (curstr == NULL) {
		DBG(zprintf("%s resp is NULL", __func__));
		return (NULL);
	}

	if (strncmp(text, "Notice:", 7) == 0) {
		curstr = strstr(text, "-\n");
		curstr = curstr + 2;
	}

	usage_start = strstr(text, "Usage:");
	if (usage_start == NULL) {
		DBG(zprintf("%s usage_start is NULL\n", __func__));
		err_len = strlen(curstr);
	} else {
		DBG(zprintf("%s usage_start is %s\n", __func__, usage_start));
		err_len = usage_start - curstr;
	}

	DBG(zprintf("%s err_len=%d\n", __func__, err_len));

	if (err_len == 0)
		return (NULL);

	resp_msg = Xmalloc(err_len);
	(void) strlcpy(resp_msg, curstr, err_len);
	return (resp_msg);
}

int
xml_call_cli_cmd(int argc, void *argv, int cmd_num, xmlNodePtr data,
    cli_conn_t *client)
{
	cli_result_t		result = CLI_RESULT_OK;
	response_t		*resp;
	char			*resp_msg;
	char			*resp_str;

	resp = client->resp;

	switch (cmd_num) {
	case LDM_ADD_VCPU:
		result = cli_vcpu_add(argc, argv, client);
		break;
	case LDM_MODIFY_VCPU:
		result = cli_vcpu_set(argc, argv, client);
		break;
	case LDM_REMOVE_VCPU:
		result = cli_vcpu_remove(argc, argv, client);
		break;
	case LDM_ADD_CRYPTO:
		result = cli_mau_add(argc, argv, client);
		break;
	case LDM_MODIFY_CRYPTO:
		result = cli_mau_set(argc, argv, client);
		break;
	case LDM_REMOVE_CRYPTO:
		result = cli_mau_remove(argc, argv, client);
		break;
	case LDM_ADD_MEM:
		result = cli_mem_add(argc, argv, client);
		break;
	case LDM_MODIFY_MEM:
		result = cli_mem_set(argc, argv, client);
		break;
	case LDM_REMOVE_MEM:
		result = cli_mem_remove(argc, argv, client);
		break;
	case LDM_ADD_VDS:
		result = cli_vds_add(argc, argv, client);
		break;
	case LDM_REMOVE_VDS:
		result = cli_vds_remove(argc, argv, client);
		break;
	case LDM_ADD_VDSDEV:
		result = cli_vdsdev_add(argc, argv, client);
		break;
	case LDM_MODIFY_VDSDEV:
		result = cli_vdsdev_set(argc, argv, client);
		break;
	case LDM_REMOVE_VDSDEV:
		result = cli_vdsdev_remove(argc, argv, client);
		break;
	case LDM_ADD_VDISK:
		result = cli_vdisk_add(argc, argv, client);
		break;
	case LDM_MODIFY_VDISK:
		result = cli_vdisk_set(argc, argv, client);
		break;
	case LDM_REMOVE_VDISK:
		result = cli_vdisk_remove(argc, argv, client);
		break;
	case LDM_ADD_VSW:
		result = cli_vswitch_add(argc, argv, client);
		break;
	case LDM_MODIFY_VSW:
		result = cli_vswitch_set(argc, argv, client);
		break;
	case LDM_REMOVE_VSW:
		result = cli_vswitch_remove(argc, argv, client);
		break;
	case LDM_ADD_VNET:
		result = cli_vnet_add(argc, argv, client);
		break;
	case LDM_MODIFY_VNET:
		result = cli_vnet_set(argc, argv, client);
		break;
	case LDM_REMOVE_VNET:
		result = cli_vnet_remove(argc, argv, client);
		break;
	case LDM_ADD_VCC:
		result = cli_vcc_add(argc, argv, client);
		break;
	case LDM_MODIFY_VCC:
		result = cli_vcc_set(argc, argv, client);
		break;
	case LDM_REMOVE_VCC:
		result = cli_vcc_remove(argc, argv, client);
		break;
	case LDM_SET_VAR:
		result = cli_var_set(argc, argv, client);
		break;
	case LDM_UNSET_VAR:
		result = cli_var_remove(argc, argv, client);
		break;
	case LDM_ADD_IO:
		result = cli_io_add(argc, argv, client);
		break;
	case LDM_REMOVE_IO:
		result = cli_io_remove(argc, argv, client);
		break;
	case LDM_CREATE:
		result = cli_ldom_create(argc, argv, client);
		break;
	case LDM_SET:
		result = cli_ldom_set(argc, argv, client);
		break;
	case LDM_DELETE:
		result = cli_ldom_destroy(argc, argv, client);
		break;
	case LDM_BIND:
		result = cli_ldom_bind(argc, argv, client);
		break;
	case LDM_UNBIND:
		result = cli_ldom_unbind(argc, argv, client);
		break;
	case LDM_PANIC:
		result = cli_ldom_panic(argc, argv, client);
		break;
	case LDM_START:
		result = cli_ldom_start(argc, argv, client);
		break;
	case LDM_STOP:
		result = cli_ldom_stop(argc, argv, client);
		break;
	case LDM_ADD_SET:
		result = cli_spconfig_add(argc, argv, client);
		break;
	case LDM_USE_SET:
		result = cli_spconfig_set(argc, argv, client);
		break;
	case LDM_REMOVE_SET:
		result = cli_spconfig_remove(argc, argv, client);
		break;
	case LDM_CANCEL_DEL_RECONF:
		result = cli_cancel_reconf(argc, argv, client);
		break;
	case LDM_MODIFY_VCONS:
		result = cli_vcons_set(argc, argv, client);
		break;
	case LDM_ADD_VDPCS:
		result = cli_vdpcs_add(argc, argv, client);
		break;
	case LDM_REMOVE_VDPCS:
		result = cli_vdpcs_remove(argc, argv, client);
		break;
	case LDM_ADD_VDPCC:
		result = cli_vdpcc_add(argc, argv, client);
		break;
	case LDM_REMOVE_VDPCC:
		result = cli_vdpcc_remove(argc, argv, client);
		break;
	case LDM_MIGRATE_DOMAIN:
		result = cli_ldom_migrate(argc, argv, client);
		break;
	case LDM_CANCEL_OPERATION:
		result = cli_cancel_operation(argc, argv, client);
		break;
	case LDM_ADD_POLICY:
		result = cli_policy_add(argc, argv, client);
		break;
	case LDM_MODIFY_POLICY:
		result = cli_policy_set(argc, argv, client);
		break;
	case LDM_REMOVE_POLICY:
		result = cli_policy_remove(argc, argv, client);
		break;
	}

	resp_str = get_response_text(resp);
	if (resp_str != NULL) {
		resp_msg = xml_parse_resp(resp_str);
		if (result != CLI_RESULT_OK)
			xml_add_status(data, XML_FAILURE, resp_msg);
		else
			xml_add_status(data, XML_SUCCESS, resp_msg);
		Xfree(resp_msg);
		Xfree(resp_str);
	}

	/*
	 * Here we need to reset the response pointer to make sure that all
	 * future responses do not show cruft from past responses.
	 */
	fini_response(client->resp);
	init_response(&client->resp);

	return (result == CLI_RESULT_OK ? true : false);
}

void
free_argv(void **argv, int argc)
{
	int	i;

	for (i = 0; i < argc; i++) {
		Xfree(argv[i]);
	}
}

int
version_check(xmlNodePtr data_node, xmlNodePtr ret_cmd, int *maj_ver)
{
	xmlChar		*version;
	uint16_t 	major, minor, sup_major, sup_minor;
	xmlNodePtr	ret_data;

	version = xmlGetProp(data_node, VERSION_ATTR);
	if (version == NULL) {
		ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
		xml_add_status(ret_data, XML_FAILURE,
		    "data tag does not have a version property");
		return (-1);
	}

	if (sscanf((char *)version, "%hu.%hu", &major, &minor) != 2) {
		ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
		xml_add_status(ret_data, XML_FAILURE,
		    "Failure converting version number");
		xmlFree(version);
		return (-1);
	}

	if (!negotiate_version(NUM_VERS(xml_data_versions), xml_data_versions,
	    major, &sup_major, &sup_minor)) {
		ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
		xmlNewProp(ret_data, VERSION_ATTR, version);
		xml_add_status(ret_data, XML_FAILURE,
		    "XML data version %s not supported", (char *)version);
		xmlFree(version);
		return (-1);
	}

	xmlFree(version);

	*maj_ver = major;

	return (0);
}

void
set_missing_prop_msg(xmlNodePtr ret_data, xmlChar *resource, xmlChar *property)
{
	ASSERT(ret_data != NULL);
	ASSERT(resource != NULL);
	ASSERT(property != NULL);

	xml_add_status(ret_data, XML_FAILURE, "Malformed XML Document:"
	    " '%s' tag missing '%s' tag", (char *)resource, (char *)property);
}

void
xml_create_return_doc(xmlDocPtr *ret_docp)
{
	xmlNodePtr	root;
	xmlDocPtr	ret_doc;

	ret_doc = xmlNewDoc(XML_VERSION);
	root =  xmlNewDocNode(ret_doc, NULL, XML_LDM_INTERFACE, NULL);
	xmlDocSetRootElement(ret_doc, root);

	xmlNewProp(root, VERSION_ATTR, LDOM_INTERFACE_VERSION);

	*ret_docp = ret_doc;
}
