/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <alloca.h>

#include "allocate.h"
#include "database.h"
#include "response.h"
#include "ldom.h"
#include "basics.h"
#include "zprintf.h"
#include "clp.h"
#include "ds_impl.h"
#include "database_internal.h"
#include "xml_parse.h"
#include "xml_common.h"
#include "xml_cntrl_common.h"
#include "xml_output.h"
#include "xml_commands.h"
#include "mac_addrs.h"
#include "fatal.h"

#define	SINGLE_CLI_ARG	1

typedef int (*mdstore_f)(char *, char *, response_t *, xmlNodePtr);

int
xml_list_variable(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{
	int		ret = 0, rv = 0;
	xmlNodePtr	ret_data, ret_ldom, info_node, ldom;
	xmlNodePtr	var_node, name;
	char		*ldm_name, *var_name;
	ldom_t		*ldp;

	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);

	ldom = xml_find_subnode(data_node, LDOM_NODE);
	if (ldom != NULL) {
		ldm_name = get_name_from_data(data_node);
		if (ldm_name == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, XML_LDM_NAME);
			return (-1);
		}

		ldp = find_ldom_by_name(ldm_name);
		if (ldp == NULL) {
			xml_add_status(ret_data, XML_FAILURE, "LDom %s not "
			    "found", ldm_name);
			xmlFree((xmlChar *)ldm_name);
			return (-1);
		}

		ret_ldom = xmlNewChild(ret_data, NULL, LDOM_NODE, NULL);

		info_node = xmlNewChild(ret_ldom, NULL, XML_LDM_INFO, NULL);
		xmlNewChild(info_node, NULL, XML_LDM_NAME,
		    (xmlChar *)ldm_name);
		xmlFree((xmlChar *)ldm_name);

		var_node = xml_find_subnode(ldom, VAR_NODE);
		if (var_node == NULL) {
			add_var_output(ret_ldom, ldp, XML_CONSTRAINTS_OUTPUT,
			    NULL);
		} else {
			var_node = ldom->xmlChildrenNode;
			while (var_node != NULL) {
				if (xmlStrcmp(var_node->name, VAR_NODE)
				    != 0) {
					var_node =
					    xml_get_next_ele_node(var_node);
					continue;
				}
				name = xml_find_subnode(var_node, NAME_NODE);
				var_name = (char *)xmlNodeGetContent(name);
				ret = add_var_output(ret_ldom, ldp, -1,
				    var_name);
				if (ret == -1) {
					xml_add_status(ret_data, XML_FAILURE,
					    "variable %s not found", var_name);
					rv = ret;
				}
				var_node = xml_get_next_ele_node(var_node);
				xmlFree((xmlChar *)var_name);
			}
		}
	} else {
		set_missing_prop_msg(ret_data, XML_DATA, LDOM_NODE);
		return (-1);
	}

	if (rv == 0)
		xml_add_status(ret_data, XML_SUCCESS, NULL);

	return (rv);
}

int
xml_list_sets(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{
	xmlNodePtr	ret_data;
	int		ret = 0;
	mdstore_f	mdstore;

	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);

	mdstore = (mdstore_f)ds_channel_service_routine(ds_sp_channel(),
	    "mdstore", "mdstore");
	if (mdstore == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    "mdstore routine failed to load");
		return (-1);
	}

	ret = (*mdstore)("list", NULL, NULL, ret_data);
	if (ret != 0) {
		xml_add_status(ret_data, XML_FAILURE,
		    "list-set operation failed");
		return (-1);
	}

	xml_add_status(ret_data, XML_SUCCESS, NULL);
	return (ret);
}

int
xml_list_services(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{
	int		ret = 0;
	xmlNodePtr	ret_data, info_node, ldom;
	xmlNodePtr	next_data;
	char		*ldm_name;
	ldom_t		*ldp;

	ldom = xml_find_subnode(data_node, LDOM_NODE);
	if (ldom != NULL) {
		ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
		xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);

		ldm_name = get_name_from_data(data_node);
		if (ldm_name == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, XML_LDM_NAME);
			return (-1);
		}

		ldp = find_ldom_by_name(ldm_name);
		if (ldp == NULL) {
			xml_add_status(ret_data, XML_FAILURE,
			    "LDom %s not found", ldm_name);
			xmlFree((xmlChar *)ldm_name);
			return (-1);
		}

		ldom = xmlNewChild(ret_data, NULL, LDOM_NODE, NULL);

		info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO, NULL);
		xmlNewChild(info_node, NULL, XML_LDM_NAME,
		    (xmlChar *)ldm_name);
		xmlFree((xmlChar *)ldm_name);

		add_vsw_output(ldom, ldp, XML_SERVICES_OUTPUT);
		add_vcc_output(ldom, ldp, XML_SERVICES_OUTPUT);
		add_vds_output(ldom, ldp, XML_SERVICES_OUTPUT);
		add_vdpcs_output(ldom, ldp, XML_SERVICES_OUTPUT);
		xml_add_status(ret_data, XML_SUCCESS, NULL);
	} else {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			next_data = xmlNewNode(NULL, (xmlChar *)"data");
			ldom = xmlNewChild(next_data, NULL, LDOM_NODE, NULL);
			info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO,
			    NULL);
			xmlNewChild(info_node, NULL, XML_LDM_NAME,
			    (xmlChar *)ldp->namep);
			ret = xml_list_services(next_data, ret_cmd, client,
			    cmd_num);
			xmlFreeNode(next_data);
			if (ret == -1)
				return (-1);
		}
	}
	return (ret);
}

int
xml_list_devices(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{
	xmlNodePtr	ret_data;

	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);

	if (xml_find_subnode(data_node, CPU_NODE) != NULL) {
		add_free_cpu(ret_data);
		xml_add_status(ret_data, XML_SUCCESS, NULL);
		return (0);
	} else if (xml_find_subnode(data_node, MAU_NODE) != NULL) {
		add_free_mau(ret_data);
		xml_add_status(ret_data, XML_SUCCESS, NULL);
		return (0);
	} else if (xml_find_subnode(data_node, MEMORY_NODE) != NULL) {
		add_free_mem(ret_data);
		xml_add_status(ret_data, XML_SUCCESS, NULL);
		return (0);
	} else if (xml_find_subnode(data_node, PHYSIO_NODE) != NULL) {
		add_free_io(ret_data);
		xml_add_status(ret_data, XML_SUCCESS, NULL);
		return (0);
	}

	add_free_cpu(ret_data);
	add_free_mau(ret_data);
	add_free_mem(ret_data);
	add_free_io(ret_data);
	xml_add_status(ret_data, XML_SUCCESS, NULL);

	return (0);
}

int
xml_list(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
    int cmd_num)
{
	xmlNodePtr	ret_data, info_node, ldom;
	xmlNodePtr	next_data;
	ldom_t		*ldp;
	char		*ldm_name;
	int		ret;

	ldom = xml_find_subnode(data_node, LDOM_NODE);
	if (ldom != NULL) {
		ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
		xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);

		ldm_name = get_name_from_data(data_node);
		if (ldm_name == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, XML_LDM_NAME);
			return (-1);
		}

		ldom = xmlNewChild(ret_data, NULL, LDOM_NODE, NULL);
		info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO, NULL);
		xmlNewChild(info_node, NULL, XML_LDM_NAME,
		    (xmlChar *)ldm_name);
		ldp = find_ldom_by_name(ldm_name);
		if (ldp == NULL) {
			xml_add_status(ret_data, XML_FAILURE,
			    "LDom %s not found", ldm_name);
			xmlFree((char *)ldm_name);
			return (-1);
		}
		xmlFree((char *)ldm_name);

		xmlNewChild(info_node, NULL, STATE_NODE,
		    (xmlChar *)ldom_state_str(ldp->state));
		xml_add_status(ret_data, XML_SUCCESS, NULL);
		return (0);
	} else {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			next_data = xmlNewNode(NULL, (xmlChar *)"data");
			ldom = xmlNewChild(next_data, NULL, LDOM_NODE, NULL);
			info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO,
			    NULL);
			xmlNewChild(info_node, NULL, XML_LDM_NAME,
			    (xmlChar *)ldp->namep);
			ret = xml_list(next_data, ret_cmd, client, cmd_num);
			xmlFreeNode(next_data);
			if (ret == -1)
				return (-1);
		}
	}

	return (0);
}

int
xml_list_bindings(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{
	int		ret = 0;
	char		*ldm_name;
	ldom_t		*ldp;
	xmlNodePtr	ret_data, info_node, ldom;
	xmlNodePtr	next_data;
	ldom_state_t	state;
	xmlChar		*mac;
	xmlChar		*hostid;

	ldom = xml_find_subnode(data_node, LDOM_NODE);
	if (ldom != NULL) {
		ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
		xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);

		ldm_name = get_name_from_data(data_node);
		if (ldm_name == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, XML_LDM_NAME);
			return (-1);
		}

		ldom = xmlNewChild(ret_data, NULL, LDOM_NODE, NULL);

		info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO, NULL);
		xmlNewChild(info_node, NULL, XML_LDM_NAME,
		    (xmlChar *)ldm_name);

		ldp = find_ldom_by_name(ldm_name);
		if (ldp == NULL) {
			xml_add_status(ret_data, XML_FAILURE,
			    "LDom %s not found", ldm_name);
			xmlFree((xmlChar *)ldm_name);
			return (-1);
		}

		state = ldp->state;
		if ((state != ACTIVE) && (state != BOUND)) {
			xml_add_status(ret_data, XML_FAILURE,
			    "LDom %s not Active or Bound", ldm_name);
			xmlFree((xmlChar *)ldm_name);
			return (-1);
		}
		xmlFree((xmlChar *)ldm_name);
		mac = alloca(sizeof (char) * UINT64_STR_LEN);
		mactoa(ldp->props.mac_address, (char *)mac);
		xmlNewChild(info_node, NULL, MAC_ADDRESS_NODE, mac);

		hostid = alloca(sizeof (char) * UINT64_STR_LEN);
		sprintf((char *)hostid, "0x%"PRIx64, ldp->props.hostid);
		xmlNewChild(info_node, NULL, HOSTID_NODE, hostid);

		xmlNewChild(info_node, NULL, UUID_NODE,
		    (xmlChar *)ldp->props.uuid);

		add_cpu_output(ldom, ldp, XML_BINDING_OUTPUT);
		add_mau_output(ldom, ldp, XML_BINDING_OUTPUT);
		add_mem_output(ldom, ldp, XML_BINDING_OUTPUT);
		add_physio_output(ldom, ldp, XML_BINDING_OUTPUT);
		add_io_output(ldom, ldp->namep, ldp, XML_BINDING_OUTPUT);
		add_var_output(ldom, ldp, XML_CONSTRAINTS_OUTPUT, NULL);
		xml_add_status(ret_data, XML_SUCCESS, NULL);
	} else {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			/*
			 * Do not give ANY information for ldoms which are
			 * not active or bound.
			 */
			state = ldp->state;
			if ((state != ACTIVE) && (state != BOUND)) {
				continue;
			}

			next_data = xmlNewNode(NULL, (xmlChar *)"data");
			ldom = xmlNewChild(next_data, NULL, LDOM_NODE, NULL);
			info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO,
			    NULL);
			xmlNewChild(info_node, NULL, XML_LDM_NAME,
			    (xmlChar *)ldp->namep);
			ret = xml_list_bindings(next_data, ret_cmd, client,
			    cmd_num);
			xmlFreeNode(next_data);
			if (ret == -1)
				return (-1);
		}
	}

	return (0);
}

int
xml_list_constraints(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client, int cmd_num)
{
	int		ret = 0;
	char		*ldm_name;
	ldom_t		*ldp;
	xmlNodePtr	ret_data, info_node, ldom;
	xmlNodePtr	next_data;
	xmlChar		*mac;

	ldom = xml_find_subnode(data_node, LDOM_NODE);
	if (ldom != NULL) {
		ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
		xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);

		ldm_name = get_name_from_data(data_node);
		if (ldm_name == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, XML_LDM_NAME);
			return (-1);
		}

		ldom = xmlNewChild(ret_data, NULL, LDOM_NODE, NULL);

		info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO, NULL);
		xmlNewChild(info_node, NULL, XML_LDM_NAME,
		    (xmlChar *)ldm_name);

		ldp = find_ldom_by_name(ldm_name);
		if (ldp == NULL) {
			xml_add_status(ret_data, XML_FAILURE,
			    "LDom %s not found", ldm_name);
			xmlFree((xmlChar *)ldm_name);
			return (-1);
		}
		xmlFree((xmlChar *)ldm_name);
		if (!is_mac_allocated(ldp->props.mac_address)) {
			mac = alloca(sizeof (char) * UINT64_STR_LEN);
			mactoa(ldp->props.mac_address, (char *)mac);
			xmlNewChild(info_node, NULL, MAC_ADDRESS_NODE, mac);
		}
		if (!IS_DERIVED_HOSTID_ASSIGNED(ldp->props.mac_address,
		    ldp->props.hostid)) {
			xmlChar *hostid =
			    alloca(sizeof (char) * UINT64_STR_LEN);
			sprintf((char *)hostid, "0x%"PRIx64,
			    ldp->props.hostid);
			xmlNewChild(info_node, NULL, HOSTID_NODE, hostid);
		}
		add_cpu_output(ldom, ldp, XML_CONSTRAINTS_OUTPUT);
		add_mau_output(ldom, ldp, XML_CONSTRAINTS_OUTPUT);
		add_mem_output(ldom, ldp, XML_CONSTRAINTS_OUTPUT);
		add_physio_output(ldom, ldp, XML_CONSTRAINTS_OUTPUT);
		add_io_output(ldom, ldp->namep, ldp, XML_CONSTRAINTS_OUTPUT);
		add_var_output(ldom, ldp, XML_CONSTRAINTS_OUTPUT, NULL);
		xml_add_status(ret_data, XML_SUCCESS, NULL);
	} else {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			next_data = xmlNewNode(NULL, XML_DATA);
			ldom = xmlNewChild(next_data, NULL, LDOM_NODE, NULL);
			info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO,
			    NULL);
			xmlNewChild(info_node, NULL, XML_LDM_NAME,
			    (xmlChar *)ldp->namep);
			ret = xml_list_constraints(next_data, ret_cmd, client,
			    cmd_num);
			xmlFreeNode(next_data);
			if (ret == -1)
				return (-1);
		}
	}
	return (0);
}

/*
 * This routine is because in LDOMS 1.0 the data section did NOT have a <LDOM>
 * tag.  What we are doing here is adding in the <LDOM> tag and putting what
 * was in the <DATA> tag under the <LDOM> tag.
 */
static void update_v1(xmlNodePtr data_node)
{
	xmlNodePtr	ldom, child;

	ldom = xmlNewNode(NULL, LDOM_NODE);
	child = data_node->xmlChildrenNode;
	while (child != NULL) {
		xmlUnlinkNode(child);
		xmlAddChild(ldom, child);
		child = data_node->xmlChildrenNode;
	}
	xmlAddChild(data_node, ldom);
	xmlSetProp(data_node, VERSION_ATTR, LDOM_DATA_VERSION);
}

int
xml_start(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
    int cmd_num)
{
	int		ret, rv = 0, version;
	xmlNodePtr	ret_data, ldom, next_data, info_node;
	char		*ldm_name;
	ldom_t		*ldp;
	int		argc = SINGLE_CLI_ARG;
	void		*argv[argc];

	ret = version_check(data_node, ret_cmd, &version);
	if (ret == -1)
		return (ret);
	if (version == 1)
		update_v1(data_node);

	ldom = xml_find_subnode(data_node, LDOM_NODE);
	if (ldom != NULL) {
		ldm_name = get_name_from_data(data_node);
		if (ldm_name == NULL) {
			ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
			xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);
			set_missing_prop_msg(ret_data, LDOM_NODE,
			    XML_LDM_NAME);
			return (-1);
		}

		ldp = find_ldom_by_name(ldm_name);
		if (ldp == NULL) {
			ret = xml_create(data_node, ret_cmd, client,
			    LDM_CREATE);
			if (ret == -1) {
				xmlFree((xmlChar *)ldm_name);
				return (-1);
			}
			ret = xml_bind(data_node, ret_cmd, client, LDM_BIND);
			if (ret == -1) {
				xmlFree((xmlChar *)ldm_name);
				return (-1);
			}
		} else if (ldp->state != BOUND) {
			ret = xml_bind(data_node, ret_cmd, client, LDM_BIND);
			if (ret == -1) {
				xmlFree((xmlChar *)ldm_name);
				return (-1);
			}
		}

		ret_data = find_data_section(ldm_name, ret_cmd);
		if (ret_data == NULL) {
			ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
			xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);
			ldom = xmlNewChild(ret_data, NULL, LDOM_NODE, NULL);

			info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO, NULL);
			xmlNewChild(info_node, NULL, XML_LDM_NAME,
			    (xmlChar *)ldm_name);
		}

		argv[0] = Xstrdup((char *)ldm_name);
		xmlFree((xmlChar *)ldm_name);

		ret = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
		free_argv(argv, argc);

		if (ret == false)
			return (-1);
		else
			xml_add_status(ret_data, XML_SUCCESS, NULL);
	} else {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			next_data = xmlNewNode(NULL, XML_DATA);
			xmlNewProp(next_data, VERSION_ATTR, LDOM_DATA_VERSION);
			ldom = xmlNewChild(next_data, NULL, LDOM_NODE, NULL);
			info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO,
			    NULL);
			xmlNewChild(info_node, NULL, XML_LDM_NAME,
			    (xmlChar *)ldp->namep);
			ret = xml_start(next_data, ret_cmd, client, cmd_num);
			xmlFreeNode(next_data);
			if (ret == -1)
				rv = -1;
		}
	}

	return (rv);
}

int
xml_bind(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
    int cmd_num)
{
	int		ret, rv = 0, version;
	xmlNodePtr	ret_data, ldom, next_data, info_node;
	char		*ldm_name;
	ldom_t		*ldp;
	int		argc = SINGLE_CLI_ARG;
	void		*argv[argc];

	ret = version_check(data_node, ret_cmd, &version);
	if (ret == -1)
		return (ret);
	if (version == 1)
		update_v1(data_node);

	ldom = xml_find_subnode(data_node, LDOM_NODE);
	if (ldom != NULL) {
		ldm_name = get_name_from_data(data_node);
		if (ldm_name == NULL) {
			ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
			xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);
			set_missing_prop_msg(ret_data, LDOM_NODE,
			    XML_LDM_NAME);
			return (-1);
		}

		ldp = find_ldom_by_name(ldm_name);
		if (ldp == NULL) {
			ret = xml_create(data_node, ret_cmd, client,
			    LDM_CREATE);
		}

		ret_data = find_data_section(ldm_name, ret_cmd);
		if (ret_data == NULL) {
			ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
			xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);
			ldom = xmlNewChild(ret_data, NULL, LDOM_NODE, NULL);

			info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO, NULL);
			xmlNewChild(info_node, NULL, XML_LDM_NAME,
			    (xmlChar *)ldm_name);
		}

		argv[0] = Xstrdup(ldm_name);
		xmlFree((xmlChar *)ldm_name);

		ret = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
		free_argv(argv, argc);

		if (ret == false)
			return (-1);
		else
			xml_add_status(ret_data, XML_SUCCESS, NULL);
	} else {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			next_data = xmlNewNode(NULL, XML_DATA);
			xmlNewProp(next_data, VERSION_ATTR, LDOM_DATA_VERSION);
			ldom = xmlNewChild(next_data, NULL, LDOM_NODE, NULL);
			info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO,
			    NULL);
			xmlNewChild(info_node, NULL, XML_LDM_NAME,
			    (xmlChar *)ldp->namep);
			ret = xml_bind(next_data, ret_cmd, client, cmd_num);
			xmlFreeNode(next_data);
			if (ret == -1)
				rv = -1;
		}
	}

	return (rv);
}

int
xml_create(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
    int cmd_num)
{
	int		ret = 0, rv = 0, version;
	xmlNodePtr	subnode, ret_data, ldom_info;
	xmlNodePtr	ldom, info_node, sent_ldom;
	char		*ldm_name;
	ldom_t		*ldp;
	const int	create_args = 3;
	int		argc = 0;
	void		*argv[create_args];
	xmlChar		*mac_addr;

	ret = version_check(data_node, ret_cmd, &version);
	if (ret == -1)
		return (ret);
	if (version == 1)
		update_v1(data_node);

	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);

	ldm_name = get_name_from_data(data_node);
	if (ldm_name == NULL) {
		set_missing_prop_msg(ret_data, LDOM_NODE, XML_LDM_NAME);
		return (-1);
	}
	sent_ldom = xml_find_subnode(data_node, LDOM_NODE);

	ldom = xmlNewChild(ret_data, NULL, LDOM_NODE, NULL);

	info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO, NULL);
	xmlNewChild(info_node, NULL, XML_LDM_NAME, (xmlChar *)ldm_name);

	ldom = xml_find_subnode(data_node, LDOM_NODE);
	ldom_info = xml_find_subnode(ldom, XML_LDM_INFO);
	subnode = xml_find_subnode(ldom_info, MAC_ADDRESS_NODE);
	if (subnode != NULL) {
		mac_addr = xmlNodeGetContent(subnode);
		if (strcmp((char *)mac_addr, XML_AUTO_ALLOCATED_MACADDR) != 0) {
			argv[argc++] = zsprintf("mac-addr=%s", mac_addr);
		}
		xmlFree(mac_addr);
	}

	subnode = xml_find_subnode(ldom_info, HOSTID_NODE);
	if (subnode != NULL) {
		xmlChar		*hostid;
		hostid = xmlNodeGetContent(subnode);
		if (hostid != NULL) {
			uint64_t h_id = strtoll((char *)hostid, NULL, 16);
			if (h_id != UNASSIGNED_HOSTID) {
				argv[argc++] = zsprintf("hostid=%s", hostid);
			}
		}
		xmlFree(hostid);
	}
	argv[argc++] = Xstrdup((char *)ldm_name);

	ret = xml_call_cli_cmd(argc, argv, LDM_CREATE, ret_data, client);
	free_argv(argv, argc);

	if (ret == false) {
		xmlFree((xmlChar *)ldm_name);
		return (-1);
	}

	ldp = find_ldom_by_name(ldm_name);
	xmlFree((xmlChar *)ldm_name);
	subnode = sent_ldom->xmlChildrenNode;
	while (subnode != NULL) {
		if (subnode->type != XML_ELEMENT_NODE) {
			subnode = subnode->next;
			continue;
		}

		if (xmlStrcmp(subnode->name,
		    (const xmlChar *)XML_LDM_INFO) == 0) {
			subnode = subnode->next;
			continue;
		}
		ret = goto_resource_routine((char *)subnode->name, ret_data,
		    ldp, XML_ADD_RES, subnode, client);
		if (ret == -1) {
			rv = -1;
			DBG(zprintf("Failed working with node %s",
			    (char *)subnode->name));
		}
		subnode = subnode->next;
	}

	if (rv == 0)
		xml_add_status(ret_data, XML_SUCCESS, NULL);

	return (rv);
}

int
xml_common_cmd(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
    int cmd_num)
{
	int		ret = 0, rv = 0;
	char		*ldm_name;
	ldom_t		*ldp, *nextldp;
	xmlNodePtr	next_data, info_node;
	xmlNodePtr	ret_data, ldom;
	int		argc = SINGLE_CLI_ARG;
	void		*argv[argc];

	ldom = xml_find_subnode(data_node, LDOM_NODE);
	if (ldom != NULL) {
		ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
		xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);

		ldm_name = get_name_from_data(data_node);
		if (ldm_name == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE,
			    XML_LDM_NAME);
			return (-1);
		}

		ldom = xmlNewChild(ret_data, NULL, LDOM_NODE, NULL);
		info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO, NULL);
		xmlNewChild(info_node, NULL, XML_LDM_NAME,
		    (xmlChar *)ldm_name);

		argv[0] = Xstrdup((char *)ldm_name);
		xmlFree((xmlChar *)ldm_name);

		ret = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
		free_argv(argv, argc);

		if (ret == false)
			rv = -1;
		else
			xml_add_status(ret_data, XML_SUCCESS, NULL);

	} else {
		CHAIN_ITER_VOL(zeus.ldoms, ldp, nextldp, link) {
			next_data = xmlNewNode(NULL, XML_DATA);
			ldom = xmlNewChild(next_data, NULL, LDOM_NODE, NULL);
			info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO,
			    NULL);
			xmlNewChild(info_node, NULL, XML_LDM_NAME,
			    (xmlChar *)ldp->namep);
			ret = xml_common_cmd(next_data, ret_cmd, client,
			    cmd_num);
			if (ret == -1)
				rv = -1;
			xmlFreeNode(next_data);
		}
	}

	return (rv);
}

static xmlNodePtr
get_resource_node(xmlNodePtr data_node, xmlNodePtr ret_data, int cmd_num)
{
	xmlNodePtr	res_node = NULL;
	xmlNodePtr	ldom_node;

	switch (cmd_num) {
	case LDM_ADD_VCPU:
	case LDM_MODIFY_VCPU:
	case LDM_REMOVE_VCPU:
		ldom_node = xml_find_subnode(data_node, LDOM_NODE);
		res_node = xml_find_subnode(ldom_node, CPU_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE, CPU_NODE);
		}
		break;
	case LDM_ADD_CRYPTO:
	case LDM_MODIFY_CRYPTO:
	case LDM_REMOVE_CRYPTO:
		ldom_node = xml_find_subnode(data_node, LDOM_NODE);
		res_node = xml_find_subnode(ldom_node, MAU_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE, MAU_NODE);
		}
		break;
	case LDM_ADD_MEM:
	case LDM_MODIFY_MEM:
	case LDM_REMOVE_MEM:
		ldom_node = xml_find_subnode(data_node, LDOM_NODE);
		res_node = xml_find_subnode(ldom_node, MEMORY_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE, MEMORY_NODE);
		}
		break;
	case LDM_ADD_IO:
	case LDM_REMOVE_IO:
		ldom_node = xml_find_subnode(data_node, LDOM_NODE);
		res_node = xml_find_subnode(ldom_node, XML_PHYSIO_DEVICE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE,
			    XML_PHYSIO_DEVICE);
		}
		break;
	case LDM_SET_VAR:
	case LDM_UNSET_VAR:
		ldom_node = xml_find_subnode(data_node, LDOM_NODE);
		res_node = xml_find_subnode(ldom_node, VAR_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE, VAR_NODE);
		}
		break;
	case LDM_ADD_VCC:
		ldom_node = xml_find_subnode(data_node, LDOM_NODE);
		res_node = xml_find_subnode(ldom_node, VCC_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE, VCC_NODE);
		}
		break;
	case LDM_MODIFY_VCC:
	case LDM_REMOVE_VCC:
		res_node = xml_find_subnode(data_node, VCC_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, VCC_NODE);
		}
		break;
	case LDM_ADD_VDISK:
	case LDM_MODIFY_VDISK:
	case LDM_REMOVE_VDISK:
		ldom_node = xml_find_subnode(data_node, LDOM_NODE);
		res_node = xml_find_subnode(ldom_node, DISK_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE, DISK_NODE);
		}
		break;
	case LDM_ADD_VDS:
		ldom_node = xml_find_subnode(data_node, LDOM_NODE);
		res_node = xml_find_subnode(ldom_node, VDS_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE, VDS_NODE);
		}
		break;
	case LDM_REMOVE_VDS:
		res_node = xml_find_subnode(data_node, VDS_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, VDS_NODE);
		}
		break;
	case LDM_ADD_VDSDEV:
	case LDM_MODIFY_VDSDEV:
	case LDM_REMOVE_VDSDEV:
		res_node = xml_find_subnode(data_node, VDS_VOLUME_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA,
			    VDS_VOLUME_NODE);
		}
		break;
	case LDM_ADD_VNET:
	case LDM_MODIFY_VNET:
	case LDM_REMOVE_VNET:
		ldom_node = xml_find_subnode(data_node, LDOM_NODE);
		res_node = xml_find_subnode(ldom_node, VNET_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE, VNET_NODE);
		}
		break;
	case LDM_ADD_VSW:
		ldom_node = xml_find_subnode(data_node, LDOM_NODE);
		res_node = xml_find_subnode(ldom_node, VSW_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE, VSW_NODE);
		}
		break;
	case LDM_MODIFY_VSW:
	case LDM_REMOVE_VSW:
		res_node = xml_find_subnode(data_node, VSW_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, VSW_NODE);
		}
		break;
	case LDM_ADD_SET:
	case LDM_USE_SET:
	case LDM_REMOVE_SET:
		res_node = xml_find_subnode(data_node, XML_SPCONFIG);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, XML_SPCONFIG);
		}
		break;
	case LDM_MODIFY_VCONS:
		ldom_node = xml_find_subnode(data_node, LDOM_NODE);
		res_node = xml_find_subnode(ldom_node, CONSOLE_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE,
			    CONSOLE_NODE);
		}
		break;
	case LDM_ADD_VDPCS:
		ldom_node = xml_find_subnode(data_node, LDOM_NODE);
		res_node = xml_find_subnode(ldom_node, VDPCS_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE, VDPCS_NODE);
		}
		break;
	case LDM_REMOVE_VDPCS:
		res_node = xml_find_subnode(data_node, VDPCS_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, VSW_NODE);
		}
		break;
	case LDM_ADD_VDPCC:
	case LDM_REMOVE_VDPCC:
		ldom_node = xml_find_subnode(data_node, LDOM_NODE);
		res_node = xml_find_subnode(ldom_node, VDPCC_NODE);
		if (res_node == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE, VDPCC_NODE);
		}
		break;
	}

	return (res_node);
}

int
xml_add_cmd(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{
	int		ret = 0;
	xmlNodePtr	res_node, ret_data;
	xmlNodePtr	ldom, info_node;
	char		*ldm_name;
	ldom_t		*ldp = NULL;

	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);

	/* If device is not a vdsdev then we will need the ldom name */
	if ((cmd_num != LDM_ADD_VDSDEV) && (cmd_num != LDM_ADD_SET)) {
		ldm_name = get_name_from_data(data_node);
		if (ldm_name == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE,
			    XML_LDM_NAME);
			return (-1);
		}

		ldom = xmlNewChild(ret_data, NULL, LDOM_NODE, NULL);

		info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO, NULL);
		xmlNewChild(info_node, NULL, XML_LDM_NAME,
		    (xmlChar *)ldm_name);

		ldp = find_ldom_by_name(ldm_name);
		if (ldp == NULL) {
			xml_add_status(ret_data, XML_FAILURE,
			    "LDom %s not found", ldm_name);
			xmlFree((xmlChar *)ldm_name);
			return (-1);
		}
		xmlFree((xmlChar *)ldm_name);
	}

	res_node = get_resource_node(data_node, ret_data, cmd_num);
	if (res_node == NULL)
		return (-1);

	ret = goto_resource_routine((char *)res_node->name, ret_data,
	    ldp, XML_ADD_RES, res_node, client);
	if (ret == 0)
		xml_add_status(ret_data, XML_SUCCESS, NULL);
	return (ret);
}

int
xml_set_cmd(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{
	int		ret = 0;
	xmlNodePtr	res_node, ret_data;
	xmlNodePtr	ldom, info_node;
	char		*ldm_name;
	ldom_t		*ldp = NULL;

	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);

	switch (cmd_num) {
	case LDM_SET:
	case LDM_MODIFY_CRYPTO:
	case LDM_MODIFY_MEM:
	case LDM_SET_VAR:
	case LDM_MODIFY_VCPU:
	case LDM_MODIFY_VNET:
	case LDM_MODIFY_VCONS:
	case LDM_MODIFY_VDISK:
		ldm_name = get_name_from_data(data_node);
		if (ldm_name == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE,
			    XML_LDM_NAME);
			return (-1);
		}

		ldom = xmlNewChild(ret_data, NULL, LDOM_NODE, NULL);

		info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO, NULL);
		xmlNewChild(info_node, NULL, XML_LDM_NAME,
		    (xmlChar *)ldm_name);

		ldp = find_ldom_by_name(ldm_name);
		if (ldp == NULL) {
			xml_add_status(ret_data, XML_FAILURE,
			    "LDom %s not found", ldm_name);
			xmlFree((xmlChar *)ldm_name);
			return (-1);
		}
		xmlFree((xmlChar *)ldm_name);
	}

	res_node = get_resource_node(data_node, ret_data, cmd_num);
	if (res_node == NULL)
		return (-1);

	ret = goto_resource_routine((char *)res_node->name, ret_data,
	    ldp, XML_MODIFY_RES, res_node, client);
	if (ret == 0)
		xml_add_status(ret_data, XML_SUCCESS, NULL);
	return (ret);
}

int
xml_remove_cmd(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{
	int		ret = 0;
	xmlNodePtr	res_node, ret_data;
	xmlNodePtr	ldom, info_node;
	char		*ldm_name;
	ldom_t		*ldp = NULL;

	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_DATA_VERSION);

	switch (cmd_num) {
	case LDM_REMOVE_IO:
	case LDM_REMOVE_CRYPTO:
	case LDM_REMOVE_MEM:
	case LDM_UNSET_VAR:
	case LDM_REMOVE_VCPU:
	case LDM_REMOVE_VDISK:
	case LDM_REMOVE_VNET:
	case LDM_REMOVE_VDPCC:
		ldm_name = get_name_from_data(data_node);
		if (ldm_name == NULL) {
			set_missing_prop_msg(ret_data, LDOM_NODE,
			    XML_LDM_NAME);
			return (-1);
		}

		ldom = xmlNewChild(ret_data, NULL, LDOM_NODE, NULL);

		info_node = xmlNewChild(ldom, NULL, XML_LDM_INFO, NULL);
		xmlNewChild(info_node, NULL, XML_LDM_NAME,
		    (xmlChar *)ldm_name);

		ldp = find_ldom_by_name(ldm_name);
		if (ldp == NULL) {
			xml_add_status(ret_data, XML_FAILURE,
			    "LDom %s not found", ldm_name);
			xmlFree((xmlChar *)ldm_name);
			return (-1);
		}
		xmlFree((xmlChar *)ldm_name);
	}

	res_node = get_resource_node(data_node, ret_data, cmd_num);
	if (res_node == NULL)
		return (-1);

	ret = goto_resource_routine((char *)res_node->name, ret_data,
	    ldp, XML_REMOVE_RES, res_node, client);
	if (ret == 0)
		xml_add_status(ret_data, XML_SUCCESS, NULL);
	return (ret);
}

int
xml_unsupported(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{
	int ret = 0;

	return (ret);
}
