/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#pragma ident	"@(#)xml_common.c	1.8	09/02/13 SMI"

#include <stdlib.h>
#include <string.h>

#include "database_internal.h"
#include "xml_common.h"
#include "xml_cntrl_common.h"
#include "xml_v3_defs.h"

void
xml_add_ldm_interface_properties(xmlNodePtr root)
{
	xmlNewProp(root, XSI, XML_SCHEMA);
	xmlNewProp(root, NONAMESPACE, LDMD_SCHEMA);
	xmlNewProp(root, OVF, ENVELOPE_SCHEMA);
	xmlNewProp(root, RASD, RASD_SCHEMA);
	xmlNewProp(root, VSSD, VSSD_SCHEMA);
	xmlNewProp(root, GPROP, GPROP_SCHEMA);
	xmlNewProp(root, BIND, BIND_SCHEMA);
}

/*
 * Returns the uint64 value stored as the content of 'node'.
 */
uint64_t
xml_get_uint64_content(xmlNodePtr node)
{
	xmlChar *content;
	uint64_t value;

	content = xmlNodeGetContent(node);
	value = strtoull((char *)content, NULL, 0);
	xmlFree(content);

	return (value);
}

/*
 * Returns the long value stored as the content of 'node'.
 */
long
xml_get_long_content(xmlNodePtr node)
{
	xmlChar *content;
	long value;

	content = xmlNodeGetContent(node);
	value = strtol((char *)content, NULL, 0);
	xmlFree(content);

	return (value);
}

/*
 * Returns the int value stored as the content of 'node'.
 */
int
xml_get_int_content(xmlNodePtr node)
{
	return ((int)xml_get_long_content(node));
}

/*
 * Sey the content of a node to a given integer value.
 */
void
xml_set_int_content(xmlNodePtr node, int value)
{
	char int_str[INT_STR_LEN];

	snprintf(int_str, sizeof (int_str), "%d", value);
	xmlNodeSetContent(node, (xmlChar *)int_str);
}

/*
 * Returns the next elemental sibling node of the node you pass in. If there
 * are no more nodes it returns NULL.
 */
xmlNodePtr
xml_get_next_ele_node(xmlNodePtr node)
{
	xmlNodePtr sib_node;

	sib_node = node->next;

	while (sib_node != NULL) {
		if (sib_node->type != XML_ELEMENT_NODE)
			sib_node = sib_node->next;
		else
			break;
	}

	return (sib_node);
}

/*
 * Find and return the first-level subnode (if any) of 'node' which has name
 * 'name'.
 */
xmlNodePtr
xml_find_subnode(xmlNodePtr node, const xmlChar *name)
{
	xmlNodePtr subnode;

	subnode = node->xmlChildrenNode;
	while (subnode != NULL) {
		if (xmlStrcmp(subnode->name, name) == 0)
			break;
		subnode = subnode->next;
	}

	return (subnode);
}

/*
 * Used for LDMD XML calls. Used if the caller already has the data node
 * parsed out. Caller must free the returned string.
 */
char *
get_name_from_data(xmlNodePtr data_node)
{
	xmlNodePtr	info_node = NULL;
	xmlNodePtr	name_node = NULL;
	xmlNodePtr	ldom = NULL;
	xmlChar		*content = NULL;

	ldom = xml_find_subnode(data_node, LDOM_NODE);
	if (ldom == NULL) {
		return (NULL);
	}

	info_node = xml_find_subnode(ldom, XML_LDM_INFO);
	if (info_node == NULL) {
		return (NULL);
	}

	name_node = xml_find_subnode(info_node, XML_LDM_NAME);

	if (name_node == NULL) {
		return (NULL);
	}

	content = xmlNodeGetContent(name_node);

	return ((char *)content);
}

/*
 * Used for LDMD XML calls. Used if the caller already has the data node
 * parsed out. Caller must free the returned string.
 */
char *
v3_get_name_from_data(xmlNodePtr data_node)
{
	xmlNodePtr	envelope = NULL;
	xmlNodePtr	content = NULL;
	xmlChar		*name = NULL;

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	if (content == NULL) {
		return (NULL);
	}

	content = xml_find_subnode(envelope, CONTENT_NODE);
	if (content == NULL) {
		return (NULL);
	}

	name = xmlGetProp(content, ID_ATTR);

	return ((char *)name);
}

/*
 * create_pkt_buf creates a ldom operation request message with an XML
 * charater buffer array.  The user must also send in if the message buffer.
 * will be of type LDM_CNTRL_XML_REQ or LDM_CNTRL_XML_RESP.  The user must
 * free the message buffer once done transmitting it.
 */
int
create_pkt_buf(char **pkt_buf, char *xml_buf, int xml_buf_len, int msg_type)
{
	int		pkt_buf_len;
	ldm_hdr_t	*hdr;
	ldm_op_req_t	*req;

	pkt_buf_len = sizeof (ldm_hdr_t) + sizeof (ldm_op_req_t) + xml_buf_len;

	*pkt_buf = malloc(pkt_buf_len);

	/* set LDM protocol headers */
	hdr = (ldm_hdr_t *)*pkt_buf;
	hdr->msg_type = hton32(msg_type);
	hdr->payload_len = hton32(sizeof (ldm_op_req_t) + xml_buf_len);

	/* set LDM op_req fields */
	req = (ldm_op_req_t *)(*pkt_buf + sizeof (ldm_hdr_t));
	req->rq_len = hton32(xml_buf_len);
	req->rq_id = 1;

	/* Copy XML buffer into the rq_data section */
	memcpy(req->rq_data, xml_buf, xml_buf_len);

	return (pkt_buf_len);
}

xmlNodePtr
create_v3_section(xmlNodePtr parent, xmlChar *type, xmlChar *obj_type)
{
	xmlNodePtr	section, item;

	section = xmlNewChild(parent, NULL, SECTION_NODE, NULL);
	xmlNewProp(section, TYPE_ATTR, type);
	item = xmlNewChild(section, NULL, ITEM_NODE, NULL);
	xmlNewChild(item, NULL, OTHER_RESOURCE_NODE_OUT, obj_type);

	return (item);
}

void
create_v3_property(xmlNodePtr parent, xmlChar *key, xmlChar *value)
{
	xmlNodePtr	property;

	property = xmlNewChild(parent, NULL, PROPERTY_NODE_OUT, value);
	xmlNewProp(property, KEY_ATTR, key);
}

int
parse_options_node(xmlNodePtr ret_cmd, char ***options)
{
	xmlNodePtr	options_node;
	char		*option_str, *option_str_count, *p;
	int		n_options = 0, i = 0;
	char		**temp = NULL;

	options_node = xml_find_subnode(ret_cmd, XML_OPTIONS);
	if (options_node == NULL)
		return (n_options);

	option_str = (char *)xmlNodeGetContent(options_node);

	/*
	 * Count the number of arguments in the string passed in.
	 * We limit the length of the string to prevent an overflow.
	 */
	if ((strlen(option_str) + 1) > MAX_OPTIONS_LEN)
		return (-1);
	option_str_count = Xmalloc(strlen(option_str) + 1);
	strlcpy(option_str_count, option_str, MAX_OPTIONS_LEN);
	p = strtok(option_str_count, " ");
	if (p == NULL)
		return (n_options);

	/*
	 * Count the remaining arguments in the string passed in.
	 * Initialize count to 1 to include the first token read above.
	 */
	n_options = 1;
	while ((p = strtok(NULL, " ")) != NULL) {
		n_options++;
	}
	Xfree(option_str_count);

	temp = Xmalloc(sizeof (char *) * n_options);

	/* Copy the first argument into the array */
	p = strtok(option_str, " ");
	temp[i] = Xmalloc(strlen(p) + 1);
	strcpy(temp[i++], p);
	p = strtok(NULL, " ");
	while ((p != NULL) && (i < n_options)) {
		temp[i] = Xmalloc(strlen(p) + 1);
		strcpy(temp[i++], p);
		p = strtok(NULL, " ");
	}
	xmlFree(option_str);
	*options = temp;
	return (n_options);
}
