/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <string.h>
#include <unistd.h>

#include "basics.h"
#include "allocate.h"
#include "chain.h"
#include "database_internal.h"
#include "xml_parse.h"
#include "xml_cntrl_common.h"
#include "xml_v3_defs.h"
#include "xml_event.h"
#include "xml_common.h"
#include "xmpp_server.h"

event_client_list_t	domain_event_list;
event_client_list_t	resource_event_list;
event_client_list_t	hardware_event_list;
event_client_list_t	bind_resource_event_list;
event_client_list_t	progress_event_list;

void
init_event_lists()
{
	CHAIN_INIT(domain_event_list.client_list);
	CHAIN_INIT(resource_event_list.client_list);
	CHAIN_INIT(hardware_event_list.client_list);
	CHAIN_INIT(bind_resource_event_list.client_list);
	CHAIN_INIT(progress_event_list.client_list);
}

/*
 * This routine creates the return event document and return back to the
 * caller the Envelope XML Node pointer.
 */
static xmlNodePtr
event_create_event_doc(xmlChar *event, xmlDocPtr *ret_docp)
{
	xmlNodePtr	root, cmd, data, env;
	xmlDocPtr	ret_doc;

	ret_doc = xmlNewDoc(XML_VERSION);
	root =  xmlNewDocNode(ret_doc, NULL, XML_LDM_EVENT, NULL);
	xmlDocSetRootElement(ret_doc, root);

	xmlNewProp(root, VERSION_ATTR, LDOM_EVENT_VERSION);
	xml_add_ldm_interface_properties(root);
	cmd = xmlNewChild(root, NULL, XML_CMD, NULL);
	xmlNewChild(cmd, NULL, XML_ACTION, event);
	data = xmlNewChild(cmd, NULL, XML_DATA, NULL);
	xmlNewProp(data, VERSION_ATTR, LDOM_V3_DATA_VERSION);
	env = xmlNewChild(data, NULL, ENVELOPE_NODE, NULL);
	xmlNewChild(env, NULL, REFERENCES_NODE, NULL);

	*ret_docp = ret_doc;

	return (env);
}

/*ARGSUSED3*/
int
event_rm_all_event(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client,	int cmd_num)
{
	(void) event_rm_event(data_node, ret_cmd, client, LDM_UNREG_DOM_EVENT);
	(void) event_rm_event(data_node, ret_cmd, client, LDM_UNREG_RES_EVENT);
	(void) event_rm_event(data_node, ret_cmd, client, LDM_UNREG_HW_EVENT);
	(void) event_rm_event(data_node, ret_cmd, client,
	    LDM_UNREG_PROGRESS_EVENT);

	return (0);
}

int
event_rm_event(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client,	int cmd_num)
{
	event_client_t		*list_client;
	event_client_list_t	*event_list;

	switch (cmd_num) {
	case LDM_UNREG_DOM_EVENT:
		event_list = &domain_event_list;
		break;
	case LDM_UNREG_RES_EVENT:
		event_list = &resource_event_list;
		break;
	case LDM_UNREG_HW_EVENT:
		event_list = &hardware_event_list;
		break;
	case LDM_UNREG_PROGRESS_EVENT:
		event_list = &progress_event_list;
		break;
	default:
		xml_add_status(ret_cmd, XML_FAILURE, "Unknown event type");
		return (-1);
	}

	CHAIN_ITER(event_list->client_list, list_client, link) {
		if (list_client->client == client)
			break;
	}
	if (list_client != NULL) {
		CHAIN_UNLINK(event_list->client_list, list_client, link);
		Xfree(list_client);
	}
	return (0);
}

int
event_add_event(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client,	int cmd_num)
{
	event_client_t		*new_client, *send_client;
	event_client_list_t	*event_list;

	switch (cmd_num) {
	case LDM_REG_DOM_EVENT:
		event_list = &domain_event_list;
		break;
	case LDM_REG_RES_EVENT:
		event_list = &resource_event_list;
		break;
	case LDM_REG_HW_EVENT:
		event_list = &hardware_event_list;
		break;
	case LDM_REG_PROGRESS_EVENT:
		event_list = &progress_event_list;
		break;
	default:
		xml_add_status(ret_cmd, XML_FAILURE, "Unknown event type");
		return (-1);
	}

	CHAIN_ITER(event_list->client_list, send_client, link) {
		if (send_client->client == client) {
			xml_add_status(ret_cmd, XML_FAILURE,
			    "Client already registered for hardware events\n");
			return (-1);
		}
	}
	new_client = Xmalloc(sizeof (event_client_t));
	new_client->client = client;
	CHAIN_ADD(event_list->client_list, new_client, link);

	return (0);
}

void
event_send_hardware_event(xmlChar *event, ldom_t *ldp, char *name,
	cli_conn_t *client)
{
	xmlDocPtr	ret_doc;
	xmlNodePtr	env, content, item;
	char		*xml_buf;
	int		xml_buf_len;
	event_client_t	*send_client;

	env = event_create_event_doc(event, &ret_doc);
	if (ldp != NULL) {
		content = xmlNewChild(env, NULL, CONTENT_NODE, NULL);
		xmlNewProp(content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
		xmlNewProp(content, ID_ATTR, (xmlChar *)ldp->namep);
	} else {
		item = create_v3_section(env, VIRTUAL_HARDWARE_TYPE,
		    XML_SPCONFIG);
		create_v3_property(item, XML_SPCONFIG_NAME, (xmlChar *)name);
	}

	xmlDocDumpMemory(ret_doc, (xmlChar **)&xml_buf, &xml_buf_len);

	CHAIN_ITER(hardware_event_list.client_list, send_client, link) {
		/*
		 * We make this check so that we don't send events back to
		 * users who have triggered the event.
		 */
		if ((client != NULL) && (client == send_client->client)) {
			continue;
		}
		send_client->client->event_buf = Xmalloc(xml_buf_len);
		memcpy(send_client->client->event_buf, xml_buf, xml_buf_len);
		send_client->client->event_buf_size = xml_buf_len;
		xmpp_write_buffer(send_client->client, true);
	}

	xmlFree(xml_buf);
	xmlFreeDoc(ret_doc);
}

void
event_send_res_event(xmlChar *event, ldom_t *ldp, char *name,
    xmlChar *res_type, cli_conn_t *client)
{
	xmlDocPtr	ret_doc;
	xmlNodePtr	env, content, item;
	char		*xml_buf;
	int		xml_buf_len;
	event_client_t	*send_client;

	env = event_create_event_doc(event, &ret_doc);
	if (ldp != NULL) {
		content = xmlNewChild(env, NULL, CONTENT_NODE, NULL);
		xmlNewProp(content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
		xmlNewProp(content, ID_ATTR, (xmlChar *)ldp->namep);
	} else {
		item = create_v3_section(env, VIRTUAL_HARDWARE_TYPE, res_type);
		create_v3_property(item, SERVICE_NAME_NODE, (xmlChar *)name);
	}

	xmlDocDumpMemory(ret_doc, (xmlChar **)&xml_buf, &xml_buf_len);

	CHAIN_ITER(resource_event_list.client_list, send_client, link) {
		/*
		 * We make this check so that we don't send events back to
		 * users who have triggered the event.
		 */
		if ((client != NULL) && (client == send_client->client)) {
			continue;
		}
		send_client->client->event_buf = Xmalloc(xml_buf_len);
		memcpy(send_client->client->event_buf, xml_buf, xml_buf_len);
		send_client->client->event_buf_size = xml_buf_len;
		xmpp_write_buffer(send_client->client, true);
	}

	xmlFree(xml_buf);
	xmlFreeDoc(ret_doc);
}

int
event_add_all_event(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client,	int cmd_num)
{
	int		ret;

	ret = event_add_event(data_node, ret_cmd, client, LDM_REG_RES_EVENT);
	if (ret == -1)
		return (ret);

	ret = event_add_event(data_node, ret_cmd, client, LDM_REG_DOM_EVENT);
	if (ret == -1) {
		event_rm_all_event(data_node, ret_cmd, client, -1);
		return (ret);
	}

	ret = event_add_event(data_node, ret_cmd, client, LDM_REG_HW_EVENT);
	if (ret == -1) {
		event_rm_all_event(data_node, ret_cmd, client, -1);
		return (ret);
	}

	ret = event_add_event(data_node, ret_cmd, client,
	    LDM_REG_PROGRESS_EVENT);
	if (ret == -1) {
		event_rm_all_event(data_node, ret_cmd, client, -1);
		return (ret);
	}
	return (ret);
}

void
event_send_dom_event(xmlChar *event, ldom_t *ldp, cli_conn_t *client)
{
	xmlDocPtr	ret_doc;
	xmlNodePtr	env, content;
	char		*xml_buf;
	int		xml_buf_len;
	event_client_t	*send_client;

	ASSERT(ldp != NULL);

	env = event_create_event_doc(event, &ret_doc);
	content = xmlNewChild(env, NULL, CONTENT_NODE, NULL);
	xmlNewProp(content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
	xmlNewProp(content, ID_ATTR, (xmlChar *)ldp->namep);

	xmlDocDumpMemory(ret_doc, (xmlChar **)&xml_buf, &xml_buf_len);

	CHAIN_ITER(domain_event_list.client_list, send_client, link) {
		/*
		 * We make this check so that we don't send events back to
		 * users who have triggered the event.
		 */
		if ((client != NULL) && (client == send_client->client)) {
			continue;
		}
		send_client->client->event_buf = Xmalloc(xml_buf_len);
		memcpy(send_client->client->event_buf, xml_buf, xml_buf_len);
		send_client->client->event_buf_size = xml_buf_len;
		xmpp_write_buffer(send_client->client, true);
	}

	xmlFree(xml_buf);
	xmlFreeDoc(ret_doc);
}

void
event_send_progress_event(xmlChar *event, ldom_t *ldp, xmlChar *status)
{
	xmlDocPtr	ret_doc;
	xmlNodePtr	env, content, info_node;
	char		*xml_buf, progress_str[INT_STR_LEN + 1];
	int		xml_buf_len;
	event_client_t	*send_client;
	char		hostname[MAXHOSTNAMELEN];

	ASSERT(ldp != NULL);
	ASSERT(status != NULL);

	(void) gethostname(hostname, MAXHOSTNAMELEN);

	env = event_create_event_doc(event, &ret_doc);
	content = xmlNewChild(env, NULL, CONTENT_NODE, NULL);
	xmlNewProp(content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
	xmlNewProp(content, ID_ATTR, (xmlChar *)ldp->namep);
	info_node = create_v3_section(content, RESOURCE_SECTION_TYPE,
	    XML_LDM_INFO);

	if (strcmp((char *)event, (char *)MIGRATION_PROGRESS) == 0)
		sprintf((char *)progress_str, "%d%%", ldp->mig_info.progress);
	else if (strcmp((char *)event, (char *)MEMDR_PROGRESS) == 0)
		sprintf((char *)progress_str, "%d%%",
		    ldp->mem_dr_state.progress);
	else
		fatal("Unsupported progress event");

	create_v3_property(info_node, XML_PROGRESS, (xmlChar *)progress_str);
	create_v3_property(info_node, XML_STATUS, status);
	create_v3_property(info_node, XML_SOURCE, (xmlChar *)hostname);

	xmlDocDumpMemory(ret_doc, (xmlChar **)&xml_buf, &xml_buf_len);

	CHAIN_ITER(progress_event_list.client_list, send_client, link) {
		send_client->client->event_buf = Xmalloc(xml_buf_len);
		memcpy(send_client->client->event_buf, xml_buf, xml_buf_len);
		send_client->client->event_buf_size = xml_buf_len;
		xmpp_write_buffer(send_client->client, true);
	}

	xmlFree(xml_buf);
	xmlFreeDoc(ret_doc);
}
