/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <alloca.h>
#include <libxml/tree.h>
#include <libxml/parser.h>

#include "basics.h"
#include "zprintf.h"
#include "support.h"
#include "database.h"
#include "allocate.h"
#include "ldom.h"
#include "zpri.h"
#include "response.h"
#include "ldomconfigure.h"
#include "xml_parse.h"
#include "xml_common.h"
#include "xml_cntrl_common.h"
#include "xml_output.h"
#include "database_internal.h"
#include "mac_addrs.h"
#include "variables.h"
#include "fatal.h"
#include "ptrvec.h"
#include "vcpu.h"
#include "vmau.h"
#include "memory.h"
#include "vio.h"

#define	XML_VERSION	((xmlChar *)"1.0") /* version of XML used */
#define	MBYTE		(1024 * 1024)

int
add_var_output(xmlNodePtr ret_ldom, ldom_t *ldp, int type, char *var_name)
{
	variables_t		*vars;
	variable_t		*var;
	int			i;
	xmlNodePtr		new_var;
	char			*s;
	bool_t			var_found = false;

	if (ldp->variables.len == 0) {
		if (type == XML_CONSTRAINTS_OUTPUT)
			return (0);
		else
			return (-1);
	}

	vars = &ldp->variables;

	for (i = 0; i < vars->len; i++) {
		var = &vars->pairs[i];

		if ((var->name != NULL) &&
		    (strcmp(var->name, "security-password") == 0)) {
			var_found = true;
			continue;
		}

		if (type == XML_CONSTRAINTS_OUTPUT) {
			s = var->value ? var->value : "";
			new_var = xmlNewChild(ret_ldom, NULL, VAR_NODE, NULL);

			xmlNewChild(new_var, NULL, NAME_NODE,
			    (xmlChar *)var->name);
			xmlNewChild(new_var, NULL, VALUE_NODE, (xmlChar *)s);
			var_found = true;
		} else {
			if ((var->name != NULL) &&
			    (strcmp(var->name, var_name) == 0)) {
				s = var->value ? var->value : "";
				new_var = xmlNewChild(ret_ldom, NULL,
				    VAR_NODE, NULL);

				xmlNewChild(new_var, NULL, NAME_NODE,
				    (xmlChar *)var->name);
				xmlNewChild(new_var, NULL, VALUE_NODE,
				    (xmlChar *)s);
				var_found = true;
			}
		}
	}

	if (var_found == true)
		return (0);
	else
		return (-1);
}

void
add_io_output(xmlNodePtr data, char *ldm_name, ldom_t *ldp, int type)
{
	add_vsw_output(data, ldp, type);
	add_vcc_output(data, ldp, type);
	add_vds_output(data, ldp, type);
	add_vdsdev_output(data, ldp, type);
	add_network_output(data, ldp, type);
	add_vdpcs_output(data, ldp, type);
	add_disk_output(data, ldp, type);
	add_vdpcc_output(data, ldp, type);
	if (type == XML_BINDING_OUTPUT)
		add_vcons_output(data, ldp, type);
}

static void
add_vcc_binding(xmlNodePtr vcc_node, vcc_t *vcc)
{
	xmlNodePtr binding, new_cons;
	xmlChar port[INT_STR_LEN];
	vcc_console_t *vcons;

	if (vcc->binding == NULL)
		return;

	CHAIN_ITER(vcc->binding->consoles, vcons, link) {
		binding = xmlNewChild(vcc_node, NULL, XML_BINDING, NULL);
		new_cons = xmlNewChild(binding, NULL, CONSOLE_NODE, NULL);
		xmlNewChild(new_cons, NULL, XML_LDM_NAME,
		    (xmlChar *)vcons->console->ldp->namep);
		snprintf((char *)port, sizeof (port), "%d",
		    (int)vcons->group->tcp_port->id);
		xmlNewChild(new_cons, NULL, XML_PORT, port);
	}
}

void
add_vcc_output(xmlNodePtr data, ldom_t *ldp, int type)
{
	xmlNodePtr new_vcc;
	xmlChar port[INT_STR_LEN];
	vcc_t *vcc;

	CHAIN_ITER(ldp->vccs, vcc, ldom) {
		new_vcc = xmlNewChild(data, NULL, VCC_NODE, NULL);

		xmlNewChild(new_vcc, NULL, SERVICE_NAME_NODE,
		    (xmlChar *)vcc->service);

		snprintf((char *)port, sizeof (port), "%d", vcc->min_tcp_port);
		xmlNewChild(new_vcc, NULL, MIN_PORT_NODE, port);

		snprintf((char *)port, sizeof (port), "%d", vcc->max_tcp_port);
		xmlNewChild(new_vcc, NULL, MAX_PORT_NODE, port);

		if (type == XML_BINDING_OUTPUT) {
			add_vcc_binding(new_vcc, vcc);
		}
	}
}

static void
add_vdsdev_binding(xmlNodePtr vdsdev, vdisk_t *vdisk)
{
	xmlNodePtr binding = NULL;
	xmlNodePtr new_disk;

	binding = xmlNewChild(vdsdev, NULL, XML_BINDING, NULL);

	new_disk = xmlNewChild(binding, NULL, DISK_NODE, NULL);

	xmlNewChild(new_disk, NULL, VDISK_NAME_NODE,
	    (xmlChar *)vdisk->name);

	xmlNewChild(new_disk, NULL, XML_LDM_NAME, (xmlChar *)vdisk->ldp->namep);
}

void
add_vdsdev_output(xmlNodePtr data, ldom_t *ldp, int type)
{
	xmlNodePtr new_vdsdev;
	vds_t *vds;
	vds_volume_t *vol;

	CHAIN_ITER(ldp->vdss, vds, ldom) {
		CHAIN_ITER(vds->volumes, vol, link) {
			new_vdsdev = xmlNewChild(data, NULL, VDS_VOLUME_NODE,
			    NULL);

			xmlNewChild(new_vdsdev, NULL, VOLUME_NAME_NODE,
			    (xmlChar *)vol->volname);
			if (vol->blkdev != NULL) {
				xmlNewChild(new_vdsdev, NULL, BLOCKDEV_NODE,
				    (xmlChar *)vol->blkdev);
			}

			xmlNewChild(new_vdsdev, NULL, SERVICE_NAME_NODE,
			    (xmlChar *)vds->service);

			if (vol->options != NULL) {
				xmlNewChild(new_vdsdev, NULL, VOLUME_OPTS_NODE,
				    (xmlChar *)vol->options);
			}

			if ((type == XML_BINDING_OUTPUT) &&
			    (vol->vds_port != NULL)) {
				add_vdsdev_binding(new_vdsdev,
				    vol->vds_port->vdisk_port->vdisk);
			}
		}
	}
}

static void
add_vds_binding(xmlNodePtr vds_node, vds_t *vds)
{
	xmlNodePtr binding, new_vdsdev;
	vds_volume_t *vol;

	CHAIN_ITER(vds->volumes, vol, link) {
		binding = xmlNewChild(vds_node, NULL, XML_BINDING, NULL);

		new_vdsdev = xmlNewChild(binding, NULL, VDS_VOLUME_NODE, NULL);

		xmlNewChild(new_vdsdev, NULL, VOLUME_NAME_NODE,
		    (xmlChar *)vol->volname);

		if (vol->blkdev != NULL) {
			xmlNewChild(new_vdsdev, NULL, BLOCKDEV_NODE,
			    (xmlChar *)vol->blkdev);
		}
	}
}

void
add_vds_output(xmlNodePtr data, ldom_t *ldp, int type)
{
	xmlNodePtr new_vds;
	vds_t *vds;

	CHAIN_ITER(ldp->vdss, vds, ldom) {
		new_vds = xmlNewChild(data, NULL, VDS_NODE, NULL);

		xmlNewChild(new_vds, NULL, SERVICE_NAME_NODE,
		    (xmlChar *)vds->service);

		if ((type == XML_BINDING_OUTPUT) ||
		    (type == XML_SERVICES_OUTPUT)) {
			add_vds_binding(new_vds, vds);
		}
	}
}

static void
add_vdpcs_binding(xmlNodePtr vdpcs_node, vdpcs_t *vdpcs)
{
	xmlNodePtr binding, new_vdpcc;
	vdpcs_port_t *vdpcs_port;

	if (vdpcs->binding == NULL)
		return;

	CHAIN_ITER(vdpcs->binding->vports, vdpcs_port, link) {
		binding = xmlNewChild(vdpcs_node, NULL, XML_BINDING, NULL);
		new_vdpcc = xmlNewChild(binding, NULL, VDPCC_NODE, NULL);
		xmlNewChild(new_vdpcc, NULL, VDPCC_NAME_NODE,
		    (xmlChar *)vdpcs_port->vdpcc->name);

		xmlNewChild(new_vdpcc, NULL, XML_LDM_NAME,
		    (xmlChar *)vdpcs_port->vdpcc->ldp->namep);
	}
}

void
add_vdpcs_output(xmlNodePtr data, ldom_t *ldp, int type)
{
	xmlNodePtr new_vdpcs;
	vdpcs_t *vdpcs;

	CHAIN_ITER(ldp->vdpcss, vdpcs, ldom) {
		new_vdpcs = xmlNewChild(data, NULL, VDPCS_NODE, NULL);

		xmlNewChild(new_vdpcs, NULL, SERVICE_NAME_NODE,
		    (xmlChar *)vdpcs->service);

		if ((type == XML_BINDING_OUTPUT) ||
		    (type == XML_SERVICES_OUTPUT)) {
			add_vdpcs_binding(new_vdpcs, vdpcs);
		}
	}
}

void
add_vdpcc_output(xmlNodePtr data, ldom_t *ldp, int type)
{
	xmlNodePtr new_vdpcc;
	vdpcc_t *vdpcc;

	CHAIN_ITER(ldp->vdpccs, vdpcc, ldom) {
		new_vdpcc = xmlNewChild(data, NULL, VDPCC_NODE, NULL);

		xmlNewChild(new_vdpcc, NULL, VDPCC_NAME_NODE,
		    (xmlChar *)vdpcc->name);

		xmlNewChild(new_vdpcc, NULL, SERVICE_NAME_NODE,
		    (xmlChar *)vdpcc->service);
	}
}

static void
add_vsw_binding(xmlNodePtr vsw_node, vsw_t *vsw)
{
	vsw_port_t *vsw_port;
	xmlNodePtr binding, new_net;
	xmlChar mac[MAC_STR_LEN] = { 0 };

	if (vsw->binding == NULL)
		return;

	CHAIN_ITER(vsw->binding->vports, vsw_port, link) {
		vnet_t *vnet = vsw_port->vnet;
		binding = xmlNewChild(vsw_node, NULL, XML_BINDING, NULL);
		new_net = xmlNewChild(binding, NULL, VNET_NODE, NULL);
		xmlNewChild(new_net, NULL, VNET_NAME_NODE,
		    (xmlChar *)vnet->name);

		mactoa(vnet->mac, (char *)mac);
		xmlNewChild(new_net, NULL, MAC_ADDRESS_NODE, mac);

		xmlNewChild(new_net, NULL, XML_LDM_NAME,
		    (xmlChar *)vnet->ldp->namep);
	}
}

void
add_vsw_output(xmlNodePtr data, ldom_t *ldp, int type)
{
	xmlNodePtr new_vsw;
	xmlChar mac[MAC_STR_LEN] = { 0 };
	vsw_t *vsw;

	CHAIN_ITER(ldp->vsws, vsw, ldom) {
		new_vsw = xmlNewChild(data, NULL, VSW_NODE, NULL);

		xmlNewChild(new_vsw, NULL, SERVICE_NAME_NODE,
		    (xmlChar *)vsw->service);

		if (type == XML_CONSTRAINTS_OUTPUT) {
			if (!is_mac_allocated(vsw->mac)) {
				mactoa(vsw->mac, (char *)mac);
				xmlNewChild(new_vsw, NULL, MAC_ADDRESS_NODE,
				    mac);
			} else {
				xmlNewChild(new_vsw, NULL, MAC_ADDRESS_NODE,
				    (xmlChar *)XML_AUTO_ALLOCATED_MACADDR);
			}
		} else if ((type == XML_BINDING_OUTPUT) ||
		    (type == XML_SERVICES_OUTPUT)) {
			char *sw_mode;
			mactoa(vsw->mac, (char *)mac);
			xmlNewChild(new_vsw, NULL, MAC_ADDRESS_NODE, mac);
			vsw_get_sw_mode(vsw->phys_dev, NULL, NULL, &sw_mode);
			xmlNewChild(new_vsw, NULL, VSW_POLICIES_NODE,
			    (xmlChar *)sw_mode);
		}

		if (vsw->mode != NULL)
			xmlNewChild(new_vsw, NULL, VSW_MODE_NODE,
			    (xmlChar *)vsw->mode);

		if (vsw->phys_dev != NULL)
			xmlNewChild(new_vsw, NULL, DEV_PATH_NODE,
			    (xmlChar *)vsw->phys_dev);

		if (type == XML_BINDING_OUTPUT) {
			add_vsw_binding(new_vsw, vsw);
		}
	}
}

void
add_disk_output(xmlNodePtr data, ldom_t *ldp, int type)
{
	xmlNodePtr new_disk;
	char *timeout_str;
	vdisk_t *vdisk;

	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {
		new_disk = xmlNewChild(data, NULL, DISK_NODE, NULL);

		xmlNewChild(new_disk, NULL, VDISK_NAME_NODE,
		    (xmlChar *)vdisk->name);

		xmlNewChild(new_disk, NULL, SERVICE_NAME_NODE,
		    (xmlChar *)vdisk->service);

		xmlNewChild(new_disk, NULL, VOLUME_NAME_NODE,
		    (xmlChar *)vdisk->volname);

		if (vdisk->timeout > 0) {
			timeout_str = zsprintf("%d", vdisk->timeout);
			xmlNewChild(new_disk, NULL, TIMEOUT_NODE,
			    (xmlChar *)timeout_str);
			free(timeout_str);
		}
	}
}

static void
network_binding_one(xmlNodePtr binding, char *name, mac_addr_t mac_addr,
    char *ldom)
{
	xmlNodePtr new_net;
	xmlChar mac[MAC_STR_LEN] = { 0 };

	new_net = xmlNewChild(binding, NULL, VNET_NODE, NULL);

	xmlNewChild(new_net, NULL, VNET_NAME_NODE, (xmlChar *)name);

	mactoa(mac_addr, (char *)mac);
	xmlNewChild(new_net, NULL, MAC_ADDRESS_NODE, mac);

	xmlNewChild(new_net, NULL, XML_LDM_NAME, (xmlChar *)ldom);
}

static void
add_network_binding(xmlNodePtr vnet_node, vnet_t *vnet)
{
	xmlNodePtr binding;
	vnet_peer_t *local_peer;

	if (vnet->binding == NULL)
		return;

	/* add bindings to peer vnets */
	CHAIN_ITER(vnet->binding->peers, local_peer, link) {
		vnet_peer_t *remote_peer = local_peer->peer;

		binding = xmlNewChild(vnet_node, NULL, XML_BINDING, NULL);
		network_binding_one(binding, remote_peer->vnet->name,
		    remote_peer->vnet->mac, remote_peer->vnet->ldp->namep);
	}
}

void
add_network_output(xmlNodePtr data, ldom_t *ldp, int type)
{
	xmlNodePtr new_net;
	xmlChar mac[MAC_STR_LEN] = { 0 };
	vnet_t *vnet;

	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		new_net = xmlNewChild(data, NULL, VNET_NODE, NULL);

		xmlNewChild(new_net, NULL, VNET_NAME_NODE,
		    (xmlChar *)vnet->name);

		xmlNewChild(new_net, NULL, SERVICE_NAME_NODE,
		    (xmlChar *)vnet->service);

		if (type == XML_BINDING_OUTPUT &&
		    vnet->binding != NULL) {
			vsw_t *vsw = vnet->binding->vsw_port->vsw;
			xmlNewChild(new_net, NULL, XML_SERVICE_DOMAIN,
			    (xmlChar *)vsw->ldp->namep);
		}

		if (type == XML_CONSTRAINTS_OUTPUT) {
			/*
			 * Check to see if the MAC address for the VNET is an
			 * allocated one. If not add the <mac_address> node and
			 * address as is. If so then don't add the <mode> tag
			 * and if the VNET should have a public or private
			 * MAC address.
			 */
			if (!is_mac_allocated(vnet->mac)) {
				mactoa(vnet->mac, (char *)mac);
				xmlNewChild(new_net, NULL, MAC_ADDRESS_NODE,
				    mac);
			} else {
				xmlNewChild(new_net, NULL, MAC_ADDRESS_NODE,
				    (xmlChar *)XML_AUTO_ALLOCATED_MACADDR);
			}
		} else if (type == XML_BINDING_OUTPUT) {
			mactoa(vnet->mac, (char *)mac);
			xmlNewChild(new_net, NULL, MAC_ADDRESS_NODE, mac);
		}

		if (type == XML_BINDING_OUTPUT) {
			add_network_binding(new_net, vnet);
		}
	}
}

void
add_vcons_output(xmlNodePtr data, ldom_t *ldp, int type)
{
	xmlNodePtr new_cons;
	xmlChar port[INT_STR_LEN];
	console_t *cons = &ldp->console;
	vcc_console_t *vcc_cons;

	if (cons->binding != NULL) {
		new_cons = xmlNewChild(data, NULL, CONSOLE_NODE, NULL);

		if (cons->binding->sp_cons) {
			xmlNewChild(new_cons, NULL, SERVICE_NAME_NODE,
			    (xmlChar *)"SP");
		} else if (cons->binding->uartbase != 0) {
			xmlNewChild(new_cons, NULL, SERVICE_NAME_NODE,
			    (xmlChar *)"UART");
		} else {
			vcc_cons = cons->binding->vcc_cons;

			xmlNewChild(new_cons, NULL, SERVICE_NAME_NODE,
			    (xmlChar *)vcc_cons->vcc->service);

			snprintf((char *)port, sizeof (port), "%d",
			    (int)vcc_cons->group->tcp_port->id);
			xmlNewChild(new_cons, NULL, XML_PORT, port);
		}
	}
}

void
add_free_cpu(xmlNodePtr data)
{
	pri_cpu_t	*pcpu_p;
	int		frac_free;
	xmlNodePtr	cpu, free;
	char		cpu_no[INT_STR_LEN];

	cpu = xmlNewChild(data, NULL, CPU_NODE, NULL);
	CHAIN_ITER(pri.cpu_list, pcpu_p, link) {
		frac_free = pcpu_p->bound.fraction_free;
		if (frac_free > 0) {
			free = xmlNewChild(cpu, NULL, XML_FREE, NULL);
			snprintf(cpu_no, sizeof (cpu_no), "%d",
			    pcpu_p->phys_id);
			xmlNewChild(free, NULL, XML_PID, (xmlChar *)cpu_no);

			snprintf(cpu_no, sizeof (cpu_no), "%d",
			    frac_free * 100 / PRI_CPU_FRAC_DENOM);
			xmlNewChild(free, NULL, XML_STRAND_PERCENT,
			    (xmlChar *)cpu_no);
		}
	}
}

static void
add_cpu_binding(xmlNodePtr cpu, ldom_t *ldp)
{
	xmlNodePtr	binding;
	vcpu_t		*vcpu;
	char		cpu_no[INT_STR_LEN];

	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		binding = xmlNewChild(cpu, NULL, XML_BINDING, NULL);

		snprintf(cpu_no, sizeof (cpu_no), "%d", vcpu->vid);
		xmlNewChild(binding, NULL, XML_VID, (xmlChar *)cpu_no);

		snprintf(cpu_no, sizeof (cpu_no), "%d", vcpu->cpu->phys_id);
		xmlNewChild(binding, NULL, XML_PID, (xmlChar *)cpu_no);

		snprintf(cpu_no, sizeof (cpu_no), "%d",
		    vcpu->fraction*100 / PRI_CPU_FRAC_DENOM);
		xmlNewChild(binding, NULL, XML_STRAND_PERCENT,
		    (xmlChar *)cpu_no);
	}
}

void
add_cpu_output(xmlNodePtr data, ldom_t *ldp, int type)
{
	char		cpu_no[INT_STR_LEN];
	xmlNodePtr	cpu;

	if (ldp->required.num_vcpus > 0) {
		cpu = xmlNewChild(data, NULL, CPU_NODE, NULL);
		snprintf(cpu_no, sizeof (cpu_no), "%d",
		    ldp->required.num_vcpus);
		xmlNewChild(cpu, NULL, NUMBER_NODE, (xmlChar *)cpu_no);
		if (type == XML_BINDING_OUTPUT) {
			add_cpu_binding(cpu, ldp);
		}
	}
}

void
add_free_mau(xmlNodePtr data)
{
	pri_cpu_t	*cpup;
	pri_mau_t 	*maup;
	xmlNodePtr	mau, free, cpu, cpuset;
	char		cpu_no[INT_STR_LEN];

	mau = xmlNewChild(data, NULL, MAU_NODE, NULL);
	CHAIN_ITER(pri.mau_list, maup, link) {
		if (get_mau_ldom(maup) == NULL) {
			/* this mau is unbound */
			free = xmlNewChild(mau, NULL, XML_FREE, NULL);
			cpuset = xmlNewChild(free, NULL, XML_CPUSET, NULL);
			CHAIN_ITER(maup->cpus, cpup, mau_link) {
				cpu = xmlNewChild(cpuset, NULL, CPU_NODE, NULL);
				snprintf(cpu_no, sizeof (cpu_no), "%d",
				    cpup->phys_id);
				xmlNewChild(cpu, NULL, NUMBER_NODE,
				    (xmlChar *)cpu_no);
			}
		}
	}
}

static void
add_mau_binding(xmlNodePtr mau, ldom_t *ldp)
{
	xmlNodePtr	binding, cpuset, cpu;
	vmau_t		*vmau;
	pri_cpu_t	*pcpu;
	char		cpu_no[INT_STR_LEN];

	CHAIN_ITER(ldp->bound.vmaus, vmau, ldom) {
		binding = xmlNewChild(mau, NULL, XML_BINDING, NULL);
		cpuset = xmlNewChild(binding, NULL, XML_CPUSET, NULL);

		CHAIN_ITER(vmau->mau->cpus, pcpu, mau_link) {
			cpu = xmlNewChild(cpuset, NULL, CPU_NODE, NULL);
			snprintf(cpu_no, sizeof (cpu_no), "%d", pcpu->phys_id);
			xmlNewChild(cpu, NULL, NUMBER_NODE, (xmlChar *)cpu_no);
		}
	}
}

void
add_mau_output(xmlNodePtr data, ldom_t *ldp, int type)
{
	char		mau_no[INT_STR_LEN];
	xmlNodePtr	mau;

	mau = xmlNewChild(data, NULL, MAU_NODE, NULL);

	snprintf(mau_no, sizeof (mau_no), "%d", ldp->required.num_vmaus);
	xmlNewChild(mau, NULL, NUMBER_NODE, (xmlChar *)mau_no);

	if ((ldom_num_bound_maus(ldp) > 0) && (type == XML_BINDING_OUTPUT))
		add_mau_binding(mau, ldp);
}

void
add_free_mem(xmlNodePtr data)
{
	xmlNodePtr	mem, free_mem;
	char		phys_mem[UINT64_STR_LEN];
	char		*mem_amt;
	mblock_t	*mblk;

	mem = xmlNewChild(data, NULL, MEMORY_NODE, NULL);

	CHAIN_ITER(free_memory.blocks, mblk, pblk) {
		free_mem = xmlNewChild(mem, NULL, XML_FREE, NULL);

		snprintf(phys_mem, sizeof (phys_mem), "0x%"PRIx64"",
		    mblk->paddr);
		xmlNewChild(free_mem, NULL, XML_PHYS_ADDR, (xmlChar *)phys_mem);

		mem_amt = conv_size(mblk->size);
		xmlNewChild(free_mem, NULL, SIZE_NODE, (xmlChar *)mem_amt);
		Xfree(mem_amt);
	}
}

static void
add_mem_binding(xmlNodePtr mem, ldom_t *ldp)
{
	char		*mem_amt;
	char		phys_mem[UINT64_STR_LEN];
	xmlNodePtr	binding;
	mblock_t	*mblk;

	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		binding = xmlNewChild(mem, NULL, XML_BINDING, NULL);

		snprintf(phys_mem, sizeof (phys_mem), "0x%"PRIx64"",
		    mblk->raddr);
		xmlNewChild(binding, NULL, XML_REAL_ADDR, (xmlChar *)phys_mem);

		snprintf(phys_mem, sizeof (phys_mem), "0x%"PRIx64"",
		    mblk->paddr);
		xmlNewChild(binding, NULL, XML_PHYS_ADDR, (xmlChar *)phys_mem);

		mem_amt = conv_size(mblk->size);
		xmlNewChild(binding, NULL, SIZE_NODE, (xmlChar *)mem_amt);
		Xfree(mem_amt);
	}
}

void
add_mem_output(xmlNodePtr data, ldom_t *ldp, int type)
{
	xmlNodePtr	mem;
	char		*mem_amt;

	if (ldp->required.size_mem > 0) {
		mem = xmlNewChild(data, NULL, MEMORY_NODE, NULL);
		mem_amt = conv_size(ldp->required.size_mem);
		xmlNewChild(mem, NULL, SIZE_NODE, (xmlChar *)mem_amt);
		Xfree(mem_amt);
		if (type == XML_BINDING_OUTPUT) {
			add_mem_binding(mem, ldp);
		}
	}
}

void
add_free_io(xmlNodePtr data)
{
	pri_iodevice_t	*pio_p;
	xmlNodePtr	io, free;

	io = xmlNewChild(data, NULL, PHYSIO_NODE, NULL);

	CHAIN_ITER(pri.iodevice_list, pio_p, link) {
		if (pio_p->free) {
			free = xmlNewChild(io, NULL, XML_FREE, NULL);
			xmlNewChild(free, NULL, XML_DEV_TYPE,
			    (xmlChar *)pio_p->device_type);
			xmlNewChild(free, NULL, NAME_NODE,
			    (xmlChar *)pio_p->name);
		}
	}
}

void
add_physio_output(xmlNodePtr data, ldom_t *ldp, int type)
{
	req_iodevice_t		*iodevs_p;
	bound_iodevice_t	*biodevs_p;
	xmlNodePtr		io_dev;

	if (type == XML_CONSTRAINTS_OUTPUT) {
		CHAIN_ITER(ldp->required.io_devices, iodevs_p, link) {
			io_dev = xmlNewChild(data, NULL, XML_PHYSIO_DEVICE,
			    NULL);
			xmlNewChild(io_dev, NULL, NAME_NODE,
			    (xmlChar *)iodevs_p->name);
		}
	}

	if (type == XML_BINDING_OUTPUT) {
		CHAIN_ITER(ldp->bound.iodevices, biodevs_p, ldom) {
			io_dev = xmlNewChild(data, NULL, XML_PHYSIO_DEVICE,
			    NULL);
			xmlNewChild(io_dev, NULL, NAME_NODE,
			    (xmlChar *)biodevs_p->name);
			xmlNewChild(io_dev, NULL, XML_DEV_TYPE,
			    (xmlChar *)biodevs_p->phys_iodev->device_type);
		}
	}
}
