/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ddi.h>
#include <errno.h>
#include <libxml/tree.h>
#include <libxml/parser.h>

#include "basics.h"
#include "errors.h"
#include "support.h"
#include "database.h"
#include "sequencer.h"
#include "ldomconfigure.h"
#include "ldom.h"
#include "response.h"
#include "clp.h"
#include "allocate.h"
#include "fatal.h"
#include "zprintf.h"
#include "auth_ldom.h"
#include "audit_ldom.h"
#include "control_class.h"
#include "xml_common.h"
#include "xml_cntrl_common.h"
#include "database_internal.h"
#include "xml_parse.h"
#include "xml_commands.h"
#include "xml_resources.h"
#include "xml_event.h"
#include "xml_v3_defs.h"

static char *xml_get_action(xmlNodePtr cmd_node);
static char *xml_get_ldm_name(xmlNodePtr cmd);

/*
 * Command table
 */
static xml_cmd_t xml_tab[] = {
	{ XML_ADD_DOMAIN, xml_create, LDM_CREATE, 1, 0 },
	{ XML_LIST_CONST, xml_list_constraints, LDM_LIST_CONSTRAINTS, 1, 0 },
	{ XML_LIST_BIND, xml_list_bindings, LDM_LIST_BINDINGS, 1, 0 },
	{ XML_LIST, xml_list, LDM_LIST, 1, 0 },
	{ XML_LIST_DEVICES, xml_list_devices, LDM_LIST_DEVICES, 1, 0 },
	{ XML_LIST_SERVICES, xml_list_services, LDM_LIST_SERVICES, 1, 0 },
	{ XML_LIST_SETS, xml_list_sets, LDM_LIST_SETS, 1, 0 },
	{ XML_LIST_VARS, xml_list_variable, LDM_GET, 1, 0 },
	{ XML_DELETE, xml_common_cmd, LDM_DELETE, 1, 0 },
	{ XML_STOP, xml_common_cmd, LDM_STOP, 1, 0 },
	{ XML_START_DOMAIN, xml_start, LDM_START, 1, 0 },
	{ XML_BIND_DOMAIN, xml_bind, LDM_BIND, 1, 0 },
	{ XML_UNBIND, xml_common_cmd, LDM_UNBIND, 1, 0 },
	{ XML_ADD_VCPU, xml_add_cmd, LDM_ADD_VCPU, 1, 0 },
	{ XML_ADD_MAU, xml_add_cmd, LDM_ADD_CRYPTO, 1, 0 },
	{ XML_ADD_MEM, xml_add_cmd, LDM_ADD_MEM, 1, 0 },
	{ XML_ADD_IO, xml_add_cmd, LDM_ADD_IO, 1, 0 },
	{ XML_ADD_VAR, xml_add_cmd, LDM_SET_VAR, 1, 0 },
	{ XML_ADD_VCONSCON, xml_add_cmd, LDM_ADD_VCC, 1, 0 },
	{ XML_ADD_VDISK, xml_add_cmd, LDM_ADD_VDISK, 1, 0 },
	{ XML_ADD_VDS, xml_add_cmd, LDM_ADD_VDS, 1, 0 },
	{ XML_ADD_VDSDEV, xml_add_cmd, LDM_ADD_VDSDEV, 1, 0 },
	{ XML_ADD_VNET, xml_add_cmd, LDM_ADD_VNET, 1, 0 },
	{ XML_ADD_VSW, xml_add_cmd, LDM_ADD_VSW, 1, 0 },
	{ XML_ADD_SPCONFIG, xml_add_cmd, LDM_ADD_SET, 1, 0 },
	{ XML_ADD_VDPCS, xml_add_cmd, LDM_ADD_VDPCS, 1, 0 },
	{ XML_ADD_VDPCC, xml_add_cmd, LDM_ADD_VDPCC, 1, 0 },
	{ XML_SET_VCPU, xml_set_cmd, LDM_MODIFY_VCPU, 1, 0 },
	{ XML_SET_MAU, xml_set_cmd, LDM_MODIFY_CRYPTO, 1, 0 },
	{ XML_SET_MEM, xml_set_cmd, LDM_MODIFY_MEM, 1, 0 },
	{ XML_SET_VAR, xml_add_cmd, LDM_SET_VAR, 1, 0 },
	{ XML_SET_VCONSCON, xml_set_cmd, LDM_MODIFY_VCC, 1, 0 },
	{ XML_SET_VDISK, xml_set_cmd, LDM_MODIFY_VDISK, 1, 1 },
	{ XML_SET_VDSDEV, xml_set_cmd, LDM_MODIFY_VDSDEV, 1, 1 },
	{ XML_SET_VNET, xml_set_cmd, LDM_MODIFY_VNET, 1, 0 },
	{ XML_SET_VSW, xml_set_cmd, LDM_MODIFY_VSW, 1, 0 },
	{ XML_SET_SPCONFIG, xml_set_cmd, LDM_USE_SET, 1, 0  },
	{ XML_SET_VCONSOLE, xml_set_cmd, LDM_MODIFY_VCONS, 1, 0 },
	{ XML_REMOVE_VCPU, xml_remove_cmd, LDM_REMOVE_VCPU, 1, 0 },
	{ XML_REMOVE_MAU, xml_remove_cmd, LDM_REMOVE_CRYPTO, 1, 0 },
	{ XML_REMOVE_MEM, xml_remove_cmd, LDM_REMOVE_MEM, 1, 0 },
	{ XML_REMOVE_IO, xml_remove_cmd, LDM_REMOVE_IO, 1, 0 },
	{ XML_REMOVE_VAR, xml_remove_cmd, LDM_UNSET_VAR, 1, 0 },
	{ XML_REMOVE_VCONSCON, xml_remove_cmd, LDM_REMOVE_VCC, 1, 0 },
	{ XML_REMOVE_VDISK, xml_remove_cmd, LDM_REMOVE_VDISK, 1, 0 },
	{ XML_REMOVE_VDS, xml_remove_cmd, LDM_REMOVE_VDS, 1, 0 },
	{ XML_REMOVE_VDSDEV, xml_remove_cmd, LDM_REMOVE_VDSDEV, 1, 0 },
	{ XML_REMOVE_VNET, xml_remove_cmd, LDM_REMOVE_VNET, 1, 0 },
	{ XML_REMOVE_VSW, xml_remove_cmd, LDM_REMOVE_VSW, 1, 0 },
	{ XML_REMOVE_SPCONFIG, xml_remove_cmd, LDM_REMOVE_SET, 1, 0 },
	{ XML_REMOVE_VDPCS, xml_remove_cmd, LDM_REMOVE_VDPCS, 1, 0 },
	{ XML_REMOVE_VDPCC, xml_remove_cmd, LDM_REMOVE_VDPCC, 1, 0 },
	{ XML_REMOVE_RECONF, xml_common_cmd, LDM_CANCEL_DEL_RECONF, 1, 0 },
};

#define	XML_CMD_MAX	(sizeof (xml_tab) / sizeof (xml_tab[0]))

/*
 * Command table
 */
static xml_cmd_t xml_v3_tab[] = {
	{ XML_ADD_DOMAIN, xml_v3_create, LDM_CREATE, 1, 0 },
	{ XML_SET_DOMAIN, xml_v3_set_cmd, LDM_SET, 1, 2 },
	{ XML_LIST_BIND, xml_v3_list_bindings, LDM_LIST_BINDINGS, 1, 0 },
	{ XML_LIST_VARS, xml_v3_list_variable, LDM_GET, 1, 0 },
	{ XML_LIST_DEVICES, xml_v3_list_devices, LDM_LIST_DEVICES, 1, 0 },
	{ XML_LIST_CONST, xml_v3_list_constraints, LDM_LIST_CONSTRAINTS, 1, 0 },
	{ XML_LIST, xml_v3_list, LDM_LIST, 1, 0 },
	{ XML_LIST_SERVICES, xml_v3_list_services, LDM_LIST_SERVICES, 1, 0 },
	{ XML_LIST_SETS, xml_v3_list_sets, LDM_LIST_SETS, 1, 0 },
	{ XML_START_DOMAIN, xml_v3_start, LDM_START, 1, 0 },
	{ XML_BIND_DOMAIN, xml_v3_bind, LDM_BIND, 1, 0 },
	{ XML_DELETE, xml_v3_common_cmd, LDM_DELETE, 1, 0 },
	{ XML_STOP, xml_v3_common_cmd, LDM_STOP, 1, 0 },
	{ XML_UNBIND, xml_v3_common_cmd, LDM_UNBIND, 1, 0 },
	{ XML_PANIC_DOMAIN, xml_v3_common_cmd, LDM_PANIC, 1, 1 },
	{ XML_REMOVE_RECONF, xml_v3_common_cmd, LDM_CANCEL_DEL_RECONF, 1, 0 },
	{ XML_ADD_VCPU, xml_v3_add_cmd, LDM_ADD_VCPU, 1, 0 },
	{ XML_ADD_MAU, xml_v3_add_cmd, LDM_ADD_CRYPTO, 1, 0 },
	{ XML_ADD_CRYPTO, xml_v3_add_cmd, LDM_ADD_CRYPTO, 1, 0 },
	{ XML_ADD_MEM, xml_v3_add_cmd, LDM_ADD_MEM, 1, 0 },
	{ XML_ADD_IO, xml_v3_add_cmd, LDM_ADD_IO, 1, 0 },
	{ XML_ADD_VAR, xml_v3_add_cmd, LDM_SET_VAR, 1, 0 },
	{ XML_ADD_VCONSCON, xml_v3_add_cmd, LDM_ADD_VCC, 1, 0 },
	{ XML_ADD_VDISK, xml_v3_add_cmd, LDM_ADD_VDISK, 1, 0 },
	{ XML_ADD_VDS, xml_v3_add_cmd, LDM_ADD_VDS, 1, 0 },
	{ XML_ADD_VDSDEV, xml_v3_add_cmd, LDM_ADD_VDSDEV, 1, 0 },
	{ XML_ADD_VNET, xml_v3_add_cmd, LDM_ADD_VNET, 1, 0 },
	{ XML_ADD_VSW, xml_v3_add_cmd, LDM_ADD_VSW, 1, 0 },
	{ XML_ADD_SPCONFIG, xml_v3_add_cmd, LDM_ADD_SET, 1, 0 },
	{ XML_ADD_VDPCS, xml_v3_add_cmd, LDM_ADD_VDPCS, 1, 0 },
	{ XML_ADD_VDPCC, xml_v3_add_cmd, LDM_ADD_VDPCC, 1, 0 },
	{ XML_ADD_POLICY, xml_v3_add_cmd, LDM_ADD_POLICY, 1, 0 },
	{ XML_SET_VAR, xml_v3_add_cmd, LDM_SET_VAR, 1, 0 },
	{ XML_SET_VCPU, xml_v3_set_cmd, LDM_MODIFY_VCPU, 1, 0 },
	{ XML_SET_MAU, xml_v3_set_cmd, LDM_MODIFY_CRYPTO, 1, 0 },
	{ XML_SET_CRYPTO, xml_v3_set_cmd, LDM_MODIFY_CRYPTO, 1, 0 },
	{ XML_SET_MEM, xml_v3_set_cmd, LDM_MODIFY_MEM, 1, 0 },
	{ XML_SET_VCONSCON, xml_v3_set_cmd, LDM_MODIFY_VCC, 1, 0 },
	{ XML_SET_VDISK, xml_v3_set_cmd, LDM_MODIFY_VDISK, 1, 1 },
	{ XML_SET_VDSDEV, xml_v3_set_cmd, LDM_MODIFY_VDSDEV, 1, 1 },
	{ XML_SET_VNET, xml_v3_set_cmd, LDM_MODIFY_VNET, 1, 0 },
	{ XML_SET_VSW, xml_v3_set_cmd, LDM_MODIFY_VSW, 1, 0 },
	{ XML_SET_SPCONFIG, xml_v3_set_cmd, LDM_USE_SET, 1, 0 },
	{ XML_SET_VCONSOLE, xml_v3_set_cmd, LDM_MODIFY_VCONS, 1, 0 },
	{ XML_SET_POLICY, xml_v3_set_cmd, LDM_MODIFY_POLICY, 1, 0 },
	{ XML_REMOVE_VCPU, xml_v3_remove_cmd, LDM_REMOVE_VCPU, 1, 0 },
	{ XML_REMOVE_MAU, xml_v3_remove_cmd, LDM_REMOVE_CRYPTO, 1, 0 },
	{ XML_REMOVE_CRYPTO, xml_v3_remove_cmd, LDM_REMOVE_CRYPTO, 1, 0 },
	{ XML_REMOVE_MEM, xml_v3_remove_cmd, LDM_REMOVE_MEM, 1, 0 },
	{ XML_REMOVE_IO, xml_v3_remove_cmd, LDM_REMOVE_IO, 1, 0 },
	{ XML_REMOVE_VAR, xml_v3_remove_cmd, LDM_UNSET_VAR, 1, 0 },
	{ XML_REMOVE_VCONSCON, xml_v3_remove_cmd, LDM_REMOVE_VCC, 1, 0 },
	{ XML_REMOVE_VDISK, xml_v3_remove_cmd, LDM_REMOVE_VDISK, 1, 0 },
	{ XML_REMOVE_VDS, xml_v3_remove_cmd, LDM_REMOVE_VDS, 1, 0 },
	{ XML_REMOVE_VDSDEV, xml_v3_remove_cmd, LDM_REMOVE_VDSDEV, 1, 0 },
	{ XML_REMOVE_VNET, xml_v3_remove_cmd, LDM_REMOVE_VNET, 1, 0 },
	{ XML_REMOVE_VSW, xml_v3_remove_cmd, LDM_REMOVE_VSW, 1, 0 },
	{ XML_REMOVE_SPCONFIG, xml_v3_remove_cmd, LDM_REMOVE_SET, 1, 0 },
	{ XML_REMOVE_VDPCS, xml_v3_remove_cmd, LDM_REMOVE_VDPCS, 1, 0 },
	{ XML_REMOVE_VDPCC, xml_v3_remove_cmd, LDM_REMOVE_VDPCC, 1, 0 },
	{ XML_REMOVE_POLICY, xml_v3_remove_cmd, LDM_REMOVE_POLICY, 1, 0 },
	{ EVENT_REG_ALL, event_add_all_event, -1, 1, 1 },
	{ EVENT_REG_DOM, event_add_event, LDM_REG_DOM_EVENT, 1, 1 },
	{ EVENT_REG_RES, event_add_event, LDM_REG_RES_EVENT, 1, 1 },
	{ EVENT_REG_HW, event_add_event, LDM_REG_HW_EVENT, 1, 1 },
	{ EVENT_REG_PROGRESS, event_add_event,  LDM_REG_PROGRESS_EVENT, 1, 1 },
	{ EVENT_UNREG_ALL, event_rm_all_event, -1, 1, 1 },
	{ EVENT_UNREG_DOM, event_rm_event, LDM_UNREG_DOM_EVENT, 1, 1 },
	{ EVENT_UNREG_RES, event_rm_event, LDM_UNREG_RES_EVENT, 1, 1 },
	{ EVENT_UNREG_HW, event_rm_event, LDM_UNREG_HW_EVENT, 1, 1 },
	{ EVENT_UNREG_PROGRESS, event_rm_event,
		LDM_UNREG_PROGRESS_EVENT, 1, 1 },
	{ XML_MIGRATE_DOMAIN, xml_v3_domain_migrate_cmd, LDM_MIGRATE_DOMAIN,
	1, 1 },
	{ XML_GET_COOKIE, xml_v3_get_cookie_cmd, 0, 1, 1 },
	{ XML_CANCEL_OP, xml_v3_common_cmd, LDM_CANCEL_OPERATION, 1, 2 },
	{ XML_REBOOT_PRIMARY, xml_v3_reboot_primary,
		LDM_REBOOT_PRIMARY, 1, 2 },
};

#define	XML_V3_CMD_MAX	(sizeof (xml_v3_tab) / sizeof (xml_v3_tab[0]))

/* Ldom interface version of XML. */
static uint16_t xml_major;
static uint16_t xml_minor;

/*
 * Resource table
 */
static xml_resource_t xml_res[] = {
	{ CPU_NODE, xml_vcpu },
	{ MAU_NODE, xml_vcrypto },
	{ MEMORY_NODE, xml_vmem },
	{ VDS_NODE, xml_vds },
	{ VDS_VOLUME_NODE, xml_vdsdev },
	{ DISK_NODE, xml_vdisk },
	{ VSW_NODE, xml_vsw },
	{ VNET_NODE, xml_vnet },
	{ VCC_NODE, xml_vcc },
	{ VAR_NODE, xml_var },
	{ XML_PHYSIO_DEVICE, xml_physio },
	{ XML_SPCONFIG, xml_spconfig },
	{ CONSOLE_NODE, xml_console },
	{ VDPCS_NODE, xml_vdpcs },
	{ VDPCC_NODE, xml_vdpcc },
};

#define	XML_RES_MAX	(sizeof (xml_res) / sizeof (xml_res[0]))

static xml_resource_t xml_v3_res[] = {
	{ CPU_NODE, xml_v3_vcpu },
	{ MAU_NODE, xml_v3_vcrypto },
	{ MEMORY_NODE, xml_v3_vmem },
	{ VDS_NODE, xml_v3_vds },
	{ VDS_VOLUME_NODE, xml_v3_vdsdev },
	{ DISK_NODE, xml_v3_vdisk },
	{ VSW_NODE, xml_v3_vsw },
	{ VNET_NODE, xml_v3_vnet },
	{ VCC_NODE, xml_v3_vcc },
	{ VAR_NODE, xml_v3_var },
	{ XML_PHYSIO_DEVICE, xml_v3_physio },
	{ XML_SPCONFIG, xml_v3_spconfig },
	{ CONSOLE_NODE, xml_v3_console },
	{ VDPCS_NODE, xml_v3_vdpcs },
	{ VDPCC_NODE, xml_v3_vdpcc },
	{ XML_LDM_INFO, xml_v3_ldm_info },
	{ POLICY_NODE, xml_v3_policy },
};

#define	XML_V3_RES_MAX	(sizeof (xml_v3_res) / sizeof (xml_v3_res[0]))

static void
print_element_names(xmlNode * a_node)
{
	xmlNode *cur_node = NULL;

	for (cur_node = a_node; cur_node; cur_node = cur_node->next) {
		if (cur_node->type == XML_ELEMENT_NODE) {
			DBG(zprintf("node type: Element, name: %s\n",
			    cur_node->name));
		} else
			DBG(zprintf("node type = %d name:%s\n", cur_node->type,
			    cur_node->name));

		print_element_names(cur_node->children);
	}
}

int
goto_resource_routine(char *name, xmlNodePtr data, ldom_t *ldp, int action,
    xmlNodePtr res_node, cli_conn_t *client)
{
	int		ret = 0;
	int		cur_cmd = -1;
	int		i, no_of_resources;
	xmlChar		*version;
	xml_resource_t	*res;

	version = xmlGetProp(data, VERSION_ATTR);
	if (strcmp((char *)version, "3.0") == 0) {
		res = xml_v3_res;
		no_of_resources = XML_V3_RES_MAX;
	} else {
		res = xml_res;
		no_of_resources = XML_RES_MAX;
	}
	xmlFree(version);

	for (i = 0; i < no_of_resources; i++) {
		if (strcmp(name, (char *)res[i].resource_name) == 0) {
			cur_cmd = i;
			break;
		}
	}

	if (cur_cmd == -1) {
		xml_add_status(data, XML_FAILURE, "%s not a known resource",
		    name);
		return (-1);
	}

	ret = res[cur_cmd].funcp(res_node, ldp, action, data, client);

	return (ret);
}

int
goto_action_routine(char *name, xmlNodePtr ret_cmd, xmlNodePtr data_node,
    cli_conn_t *client)
{
	int		ret = 0, rv = 0;
	int		cur_cmd = -1;
	int		i, version;
	xmlNodePtr	next_data = NULL;
	int		no_of_cmds;
	xml_cmd_t	*xml_cmds;

	ret = version_check(data_node, ret_cmd, &version);
	if (ret == -1)
		return (-1);

	if (version < 3) {
		no_of_cmds = XML_CMD_MAX;
		xml_cmds = xml_tab;
	} else {
		no_of_cmds = XML_V3_CMD_MAX;
		xml_cmds = xml_v3_tab;
	}

	for (i = 0; i < no_of_cmds; i++) {
		if (strcmp(name, (char *)xml_cmds[i].act_name) == 0) {
			cur_cmd = i;
			break;
		}
	}

	if (cur_cmd == -1) {
		xml_add_status(ret_cmd, XML_FAILURE, "%s not a known action",
		    name);
		return (-1);
	}

	if (xml_cmds[cur_cmd].major > xml_major ||
	    (xml_cmds[cur_cmd].major == xml_major &&
	    xml_cmds[cur_cmd].minor > xml_minor)) {
		xml_add_status(ret_cmd, XML_FAILURE, "%s is not a valid command"
		    " for version %d.%d", name, xml_major, xml_minor);
		return (-1);
	}

	ret = xml_cmds[cur_cmd].funcp(data_node, ret_cmd, client,
	    xml_cmds[cur_cmd].cmd_num);

	if (ret == -1)
		rv = -1;

	next_data = xml_get_next_ele_node(data_node);
	if ((next_data != NULL) && (xmlStrcmp(next_data->name,
	    (xmlChar *)"data") != 0)) {
		xml_add_status(ret_cmd, XML_FAILURE,
		    "%s should not be a sibling of a data tag\n",
		    next_data->name);
		return (-1);
	}
	while (next_data != NULL) {
		ret = xml_cmds[cur_cmd].funcp(next_data, ret_cmd, client,
		    xml_cmds[cur_cmd].cmd_num);
		if (ret == -1)
			rv = -1;
		next_data = xml_get_next_ele_node(next_data);
		if ((next_data != NULL) && (xmlStrcmp(next_data->name,
		    (xmlChar *)"data") != 0)) {
			xml_add_status(ret_cmd, XML_FAILURE,
			    "%s should not be a sibling of a data tag\n",
			    next_data->name);
			return (-1);
		}
	}

	return (rv);
}

/*
 * Write an audit record for this event, except for SNMP traps.
 * Get and pass command and, if present, the LDoms domain name
 * and SNMP v3 user name.
 *
 * Return the audit handle for use with close_audit_stream(), which
 * must be called to close the audit record.
 */
static audit_handle_t
audit_xml_request(xmlNodePtr cmd_node, int fd, char *action)
{
	xmlNodePtr	snmp_node;
	char		*ldm_name = NULL, *snmp_user = NULL;
	audit_handle_t	audit_handle = NULL_AUDIT_HANDLE;

	/* Sanity check */
	if (action == NULL) {
		return (NULL_AUDIT_HANDLE);
	}

	snmp_node = xml_find_subnode(cmd_node, XML_SNMP_USER);
	snmp_user = (snmp_node != NULL) ?
	    (char *)xmlNodeGetContent(snmp_node) : NULL;

	if ((snmp_user == NULL) || (strcmp(snmp_user, "snmpTrap") != 0)) {
		ldm_name = xml_get_ldm_name(cmd_node);

		/* Create a audit record depending on what we know */
		if (ldm_name != NULL) { /* pass command and domain name */
			if (snmp_user != NULL) {
				audit_handle = audit_ldm_request(fd, 5, "xml",
				    "snmp_user", snmp_user, action, ldm_name);
			} else {
				audit_handle = audit_ldm_request(fd, 3, "xml",
				    action, ldm_name);
			}

		} else { /* no domain name--just pass command */
			if (snmp_user != NULL) {
				audit_handle = audit_ldm_request(fd, 4, "xml",
				    "snmp_user", snmp_user, action);
			} else {
				audit_handle = audit_ldm_request(fd, 2, "xml",
				    action);
			}
		}
		xmlFree((xmlChar *)ldm_name);
		return (audit_handle);

	} else { /* don't audit SNMP traps, which are too frequent */
		return (NULL_AUDIT_HANDLE);
	}
}

static int
check_ldom_interface_version(xmlNodePtr root_node, xmlNodePtr ret_cmd)
{
	xmlChar *version = NULL;
	int major, minor;

	version = xmlGetProp(root_node, VERSION_ATTR);

	if (version == NULL) {
		xml_add_status(ret_cmd, XML_FAILURE,
		    "Missing ldom interface version property");
		return (-1);
	}

	if (sscanf((char *)version, "%d.%d", &major, &minor) != 2) {
		xml_add_status(ret_cmd, XML_FAILURE,
		    "Failure converting ldom interface version number");
		xmlFree(version);
		return (-1);
	}

	xmlFree(version);

	if (major > LDOM_INTERFACE_VERSION_MAJOR) {
		xml_add_status(ret_cmd, XML_FAILURE,
		    "ldom interface version number %d.%d not supported; major"
		    "version needs to be %d or lower", major, minor,
		    LDOM_INTERFACE_VERSION_MAJOR);
		return (-1);
	}

	xml_major = major;
	xml_minor = min(minor, LDOM_INTERFACE_VERSION_MINOR);

	return (0);
}

int
parse_xml_file(xmlDocPtr doc, xmlDocPtr ret_doc, cli_conn_t *client)
{
	int		ret = 0, rv = 0, version;
	int		fd = client->fd;
	int		auth_ret = 0;
	int		authorized;
	xmlNodePtr	root_node, data_node, action_node, cmd_node;
	xmlNodePtr	ret_root, ret_cmd, option_node;
	char		*action, *options;
	audit_handle_t	audit_handle = NULL_AUDIT_HANDLE;
	uid_t		zeus_euid = geteuid();
	bool_t		hdr_done = false;
	root_node = xmlDocGetRootElement(doc);
	ret_root = xmlDocGetRootElement(ret_doc);

	print_element_names(root_node);

	ret = check_ldom_interface_version(root_node, ret_root);
	if (ret == -1) {
		return (-1);
	}

	cmd_node = xml_find_subnode(root_node, XML_CMD);

	while (cmd_node != NULL) {

		ret_cmd = xmlNewChild(ret_root, NULL, XML_CMD, NULL);
		action = xml_get_action(cmd_node);
		if (action == NULL) {
			set_missing_prop_msg(ret_cmd, XML_CMD, XML_ACTION);
			rv = -1;
			cmd_node = xml_get_next_ele_node(cmd_node);
			continue;
		}

		xmlNewChild(ret_cmd, NULL, XML_ACTION, (xmlChar *)action);

		/*
		 * Add the <options> tag on the return document. This can be
		 * check by routines that might use flags such as the
		 * "cancel-operation" routine.
		 */
		option_node = xml_find_subnode(cmd_node, XML_OPTIONS);
		if (option_node != NULL) {
			options = (char *)xmlNodeGetContent(option_node);
			xmlNewChild(ret_cmd, NULL, XML_OPTIONS,
			    (xmlChar *)options);
			xmlFree(options);
		}

		/* Authorization check and audit record write */
		if (zeus_euid == 0) {
			/* Authorization check if not simulating */
			if (!client->xmpp_msg) {
				authorized = auth_check_xml_euid(
				    socket_peer_euid(fd), action);
			} else {
				if (client->local_socket == true) {
					authorized = auth_check_xml_euid(
					    socket_peer_euid(fd), action);
				} else {
					authorized = auth_check_xml_username(
					    client->uname, action);
				}
			}
			if (authorized != 0) {
				xml_add_status(ret_cmd, XML_FAILURE,
				    "Authorization failed");
				rv = -1;
				cmd_node = xml_get_next_ele_node(cmd_node);
				xmlFree(action);
				continue;
			}
		}
		audit_handle = audit_xml_request(cmd_node, fd, action);

		data_node = xml_find_subnode(cmd_node, XML_DATA);
		/*
		 * We did not find a data section. No work can be done
		 * without this
		 */
		if (data_node == NULL) {
			set_missing_prop_msg(ret_cmd, XML_CMD, XML_DATA);
			xmlFree(action);
			if (audit_handle != NULL_AUDIT_HANDLE) {
				close_audit_stream(audit_handle, auth_ret);
			}
			return (-1);
		}

		ret = version_check(data_node, ret_cmd, &version);
		if (ret == -1) {
			xml_add_status(ret_cmd, XML_FAILURE, NULL);
			if (audit_handle != NULL_AUDIT_HANDLE) {
				close_audit_stream(audit_handle, auth_ret);
			}
			xmlFree(action);
			return (-1);
		}

		if ((version == 3) && (!hdr_done)) {
			xml_add_ldm_interface_properties(ret_root);
			hdr_done = true;
		}

		action_node = xml_find_subnode(cmd_node, XML_ACTION);
		xmlUnlinkNode(action_node);
		xmlFreeNode(action_node);

		ret = goto_action_routine(action, ret_cmd, data_node, client);

		if (ret == -1) {
			xml_add_status(ret_cmd, XML_FAILURE, NULL);
			rv = -1;
		} else
			xml_add_status(ret_cmd, XML_SUCCESS, NULL);

		xmlFree((xmlChar *)action);

		if (audit_handle != NULL_AUDIT_HANDLE) {
			close_audit_stream(audit_handle, auth_ret);
		}

		cmd_node = xml_get_next_ele_node(cmd_node);
	}

	return (rv);
}

int
zeus_ldom_xml_command(cli_conn_t *client)
{
	int		ret = 0;
	xmlDocPtr	doc, ret_doc;
	char		*xml_buf, *pkt_buf;
	int		xml_buf_len;
	xmlNodePtr	root;

	xml_create_return_doc(&ret_doc);
	root = xmlDocGetRootElement(ret_doc);

	if (!client->xmpp_msg) {
		ldm_op_req_t	*req;

		req = (ldm_op_req_t *)client->buf;

		doc = xmlParseMemory((const char *)req->rq_data, req->rq_len);

		free(client->buf);
	} else {
		doc = xmlParseMemory((const char *)client->buf,
		    client->buf_size);
		free(client->buf);
		client->buf_size = 0;
	}

	if (doc == NULL) {
		xml_add_status(root, XML_FAILURE, "Malformed XML document");
		xmlDocDumpMemory(ret_doc, (xmlChar **)&xml_buf, &xml_buf_len);
		create_pkt_buf(&pkt_buf, xml_buf, xml_buf_len,
		    LDM_CNTRL_XML_RESP);
		client->buf = client->bufp = (void *)pkt_buf;
		xmlFree(xml_buf);
		xmlFreeDoc(ret_doc);
		return (-1);
	}

	ret = parse_xml_file(doc, ret_doc, client);
	if (ret != -1)
		xml_add_status(root, XML_SUCCESS, NULL);
	else
		xml_add_status(root, XML_FAILURE, NULL);
	xmlFreeDoc(doc);

	root = xmlDocGetRootElement(ret_doc);
	print_element_names(root);

	xmlDocDumpMemory(ret_doc, (xmlChar **)&xml_buf, &xml_buf_len);

	if (!client->xmpp_msg) {
		create_pkt_buf(&pkt_buf, xml_buf, xml_buf_len,
		    LDM_CNTRL_XML_RESP);

		client->buf = client->bufp = (void *)pkt_buf;
	} else {
		client->buf = client->bufp = Xmalloc(xml_buf_len);
		memcpy(client->buf, xml_buf, xml_buf_len);
		client->buf_size = xml_buf_len;
	}

	xmlFree(xml_buf);
	xmlFreeDoc(ret_doc);

	return (ret);
}

static char *
xml_get_action(xmlNodePtr cmd)
{
	xmlNodePtr	action_node;
	char		*action, *ret_action;

	action_node = xml_find_subnode(cmd, XML_ACTION);
	/* We did not find a data section. No work can be done without this */
	if (action_node == NULL) {
		DBG(zprintf("XML file does not have <action> tag\n"));
		return (NULL);
	}

	action = (char *)xmlNodeGetContent(action_node);
	ret_action = Xstrdup(action);
	xmlFree(action);
	return (ret_action);
}

/* Used if the caller does NOT already have the data node parsed out */
static char *
xml_get_ldm_name(xmlNodePtr cmd)
{
	xmlNodePtr	data_node = NULL;

	data_node = xml_find_subnode(cmd, XML_DATA);
	/* We did not find a data section. No work can be done without this */
	if (data_node == NULL) {
		DBG(zprintf("XML file does not have <data> tag\n"));
		return (NULL);
	}

	return (get_name_from_data(data_node));
}

static void
sequenced_cli_cmds(void *args)
{
	cli_result_t		result = CLI_RESULT_OK;
	xml_sequence_args_t	*arg  = (xml_sequence_args_t *)args;
	char			*resp_str;

	switch (arg->cmd) {
	case LDM_BIND:
		result = cli_ldom_bind(arg->argc, arg->argv, arg->client);
		if (result != CLI_RESULT_OK) {
			resp_str = get_response_text(arg->client->resp);
			zprintf(MSG_ERR_REBOOT_BIND, arg->argv[0], resp_str);
			Xfree(resp_str);
		}
		break;
	case LDM_START:
		result = cli_ldom_start(arg->argc, arg->argv, arg->client);
		if (result != CLI_RESULT_OK) {
			resp_str = get_response_text(arg->client->resp);
			if (strcmp(arg->argv[0], "-a") == 0) {
				zprintf(MSG_ERR_REBOOT_START_ALL, resp_str);
			} else {
				zprintf(MSG_ERR_REBOOT_START, arg->argv[0],
				    resp_str);
			}
			Xfree(resp_str);
		}
		break;
	case LDM_ADD_SET:
		result = cli_spconfig_add(arg->argc, arg->argv, arg->client);
		if (result != CLI_RESULT_OK) {
			resp_str = get_response_text(arg->client->resp);
			zprintf(MSG_ERR_REBOOT_SPCONFIG, arg->argv[0],
			    resp_str);
			Xfree(resp_str);
		}
		break;
	}

	Xfree(arg->argv[0]);
	Xfree(arg);
}

static void
reboot_setup_cmds(xmlNodePtr cmd, cli_conn_t *client, int command)
{
	xmlNodePtr		data = NULL;
	xmlNodePtr		envelope, content;
	ldom_t			*ldp;
	char			*argv[1];
	xml_sequence_args_t	*args;
	char			*ldm_name;

	data = xml_find_subnode(cmd, XML_DATA);
	if (data != NULL) {
		envelope = xml_find_subnode(data, ENVELOPE_NODE);
		if (envelope == NULL) {
			if (command == LDM_BIND) {
				CHAIN_ITER(zeus.ldoms, ldp, link) {
					if (strcmp(ldp->namep, "primary") == 0)
						continue;
					/*
					 * Will be freed in
					 * sequenced_cli_cmds
					 */
					argv[0] = Xstrdup(ldp->namep);
					/*
					 * Will be freed in
					 * sequenced_cli_cmds
					 */
					args = Xmalloc(
					    sizeof (xml_sequence_args_t));
					args->argc = 1;
					args->argv = argv;
					args->cmd = LDM_BIND;
					args->client = client;
					seq_enqueue_work(sequenced_cli_cmds,
					    args, true);
				}
			} else {
				/* Will be freed in sequenced_cli_cmds */
				argv[0] = zsprintf("-a");
				/* Will be freed in sequenced_cli_cmds */
				args = Xmalloc(sizeof (xml_sequence_args_t));
				args->argc = 1;
				args->argv = argv;
				args->cmd = LDM_START;
				args->client = client;
				seq_enqueue_work(sequenced_cli_cmds, args,
				    true);
			}
		} else {
			content = xml_find_subnode(envelope, CONTENT_NODE);

			while (content != NULL) {
				ldm_name = (char *)xmlGetProp(content,
				    ID_ATTR_IN);
				/* Will be freed in sequenced_cli_cmds */
				argv[0] = Xstrdup(ldm_name);
				xmlFree(ldm_name);
				/* Will be freed in sequenced_cli_cmds */
				args = Xmalloc(sizeof (xml_sequence_args_t));
				args->argc = 1;
				args->argv = argv;
				args->cmd = command;
				args->client = client;
				seq_enqueue_work(sequenced_cli_cmds,
				    args, true);

				xmlUnlinkNode(content);
				xmlFreeNode(content);
				content = xml_find_subnode(envelope,
				    CONTENT_NODE);
			}
		}
	}
}

/* ARGSUSED0 */
void *
read_reboot_file(void *unused)
{
	char			*filename, *spconfig = NULL;
	xmlNodePtr		envelope, root, cmd, data, section, item;
	xmlNodePtr		subnode, action;
	xmlDocPtr		reboot_doc = NULL;
	char			*argv[1], *command;
	cli_conn_t		client;
	xml_sequence_args_t	*args;

	filename = zsprintf("%s/%s", zeus.admdir, REBOOT_FILE);
	if (open(filename, O_RDONLY) < 0) {
		Xfree(filename);
		return (NULL);
	}
	reboot_doc = xmlParseFile(filename);
	if (reboot_doc == NULL) {
		Xfree(filename);
		return (NULL);
	}

	init_response(&client.resp);

	root = xmlDocGetRootElement(reboot_doc);
	cmd = xml_find_subnode(root, XML_CMD);
	data = xml_find_subnode(cmd, XML_DATA);
	/*
	 * The 1st <cmd> section will ALWAYS be an SPCONFIG section if the
	 * user wants to re-save the configuration after auto-start of the
	 * guest domains.  If they don't want this the <data> section must be
	 * empty.
	 */
	envelope = xml_find_subnode(data, ENVELOPE_NODE);
	if (envelope != NULL) {
		section = xml_find_section(envelope, XML_SPCONFIG);
		if (section != NULL) {
			item = xml_find_subnode(section, ITEM_NODE);
			if (item != NULL) {
				subnode = xml_find_v3_property(item,
				    XML_SPCONFIG_NAME);
				if (subnode != NULL) {
					spconfig =
					    (char *)xmlNodeGetContent(subnode);
				}
			}
		}
	}

	do {
		xmlUnlinkNode(cmd);
		xmlFreeNode(cmd);
		cmd = xml_find_subnode(root, XML_CMD);
		if (cmd == NULL) {
			break;
		}
		action = xml_find_subnode(cmd, XML_ACTION);
		command = (char *)xmlNodeGetContent(action);
		if (command == NULL) {
			continue;
		}
		if (strcmp(command, (char *)XML_BIND_DOMAIN) == 0) {
			/*
			 * If there is a bind <cmd> section this means the
			 * user wishes to bind some or all of the guest
			 * domains.  If the <data> section is empty then they
			 * want ALL of the guest domains bound. If it has an
			 * envelope node in it then we will bind those
			 * domains only.
			 */
			reboot_setup_cmds(cmd, &client, LDM_BIND);
		} else if (strcmp(command, (char *)XML_START_DOMAIN) == 0) {
			/*
			 * If there is a start <cmd> section this means the
			 * user wishes to start some or all of the guest
			 * domains.  If the <data> section is empty then they
			 * want ALL of the guest domains started. If it has an
			 * envelope node in it then we will start those
			 * domains only.
			 */
			reboot_setup_cmds(cmd, &client, LDM_START);
		}
		xmlFree(command);
	} while (cmd != NULL);

	if (spconfig != NULL) {
		argv[0] = Xstrdup(spconfig);
		args = Xmalloc(sizeof (xml_sequence_args_t));
		args->argc = 1;
		args->argv = argv;
		args->cmd = LDM_ADD_SET;
		args->client = &client;
		seq_enqueue_work(sequenced_cli_cmds, args, true);
		xmlFree(spconfig);
	}

	if (remove(filename) < 0) {
		warning(MSG_ERR_REBOOT_FILE_REMOVAL, filename);
	}
	Xfree(filename);
	fini_response(client.resp);
	return (NULL);
}
