/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <fcntl.h>
#include <alloca.h>
#include <sys/reboot.h>

#include "allocate.h"
#include "database.h"
#include "errors.h"
#include "response.h"
#include "ldom.h"
#include "basics.h"
#include "zprintf.h"
#include "clp.h"
#include "ds_impl.h"
#include "vcpu.h"
#include "fatal.h"
#include "database_internal.h"
#include "xml_parse.h"
#include "xml_common.h"
#include "xml_cntrl_common.h"
#include "xml_output.h"
#include "xml_commands.h"
#include "mac_addrs.h"
#include "xml_v3_defs.h"
#include "xmpp_server.h"
#include "xmpp_client.h"
#include "ldomconfigure.h"
#include "autosave.h"

#define	SINGLE_CLI_ARG	1

typedef int (*mdstore_f)(char *, char *, response_t *, xmlNodePtr);

int
xml_v3_domain_migrate_cmd(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client,	int cmd_num)
{
	xmlNodePtr	envelope, content;
	xmlNodePtr	ret_data, subnode, section, item;
	int		encode_user_len, encode_pass_len;
	char		*encode_user = NULL, *encode_pass = NULL;
	char		*user = NULL, *pass = NULL;
	char		*ldom = NULL, *new_ldom = NULL;
	char		*target_name = NULL;
	int		rv = -1;	/* overall return value from func */
	int		status;		/* return value from callee funcs */
	uint_t		outlen;
	const int	dm_args = 3;
	void		*argv[dm_args];
	int		argc = 0;
	bool_t		result;

	ASSERT(cmd_num == LDM_MIGRATE_DOMAIN);
	ASSERT(data_node != NULL);
	ASSERT(ret_cmd != NULL);

	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_V3_DATA_VERSION);
	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	if (envelope == NULL) {
		set_missing_prop_msg(ret_data, XML_DATA, ENVELOPE_NODE);
		goto cleanup;
	}

	/* Get 1st content node in envelope */
	content = xml_find_subnode(envelope, CONTENT_NODE);
	if (content == NULL) {
		set_missing_prop_msg(ret_data, XML_DATA, CONTENT_NODE);
		goto cleanup;
	}
	section = xml_find_section(content, XML_LDM_INFO);

	/*
	 * If the content node does not have a section, it is the source LDom.
	 * If it DOES have a section, it is the target LDom information.
	 */
	if (section == NULL) {
		/*
		 * content node contains source ldom info. save ldom name
		 * and delete the content node.
		 */
		ldom = (char *)xmlGetProp(content, ID_ATTR_IN);
		xmlUnlinkNode(content);
		xmlFreeNode(content);

		/* get the next (target LDom) content node */
		content = xml_find_subnode(envelope, CONTENT_NODE);
		if (content == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, CONTENT_NODE);
			DBG(zprintf("Target LDom node not specified in XML"));
			goto cleanup;
		}
		section = xml_find_section(content, XML_LDM_INFO);
		if (section == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, XML_LDM_INFO);
			DBG(zprintf("Target LDom section unspecified in XML"));
			goto cleanup;
		}
	}
	new_ldom = (char *)xmlGetProp(content, ID_ATTR_IN);
	if (new_ldom == NULL) {
		set_missing_prop_msg(ret_data, XML_DATA, ID_ATTR_IN);
		goto cleanup;
	}

	/* get the target machine's hostname */
	item = xml_find_subnode(section, ITEM_NODE);
	if (item == NULL) {
		set_missing_prop_msg(ret_data, XML_DATA, ITEM_NODE);
		goto cleanup;
	}
	subnode = xml_find_v3_property(item, XML_TARGET);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, XML_DATA, XML_TARGET);
		goto cleanup;
	}
	target_name = (char *)xmlNodeGetContent(subnode);
	if (target_name == NULL) {
		xml_add_status(ret_data, XML_FAILURE, "Target LDom node empty");
		goto cleanup;
	}

	/* get username property and decode it */
	subnode = xml_find_v3_property(item, XML_USERNAME);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, XML_DATA, XML_USERNAME);
		goto cleanup;
	}
	encode_user = (char *)xmlNodeGetContent(subnode);
	if (encode_user == NULL) {
		xml_add_status(ret_data, XML_FAILURE, "Username node empty");
		goto cleanup;
	}
	encode_user_len = strlen(encode_user);
	status = do_sasl_decode64(encode_user, encode_user_len, &user, &outlen);
	if (status == -1 || user == NULL) {
		xml_add_status(ret_data, XML_FAILURE, "Username decode failed");
		goto cleanup;
	}

	/* get password property and decode it */
	subnode = xml_find_v3_property(item, XML_PASSWORD);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, XML_DATA, XML_PASSWORD);
		goto cleanup;
	}
	encode_pass = (char *)xmlNodeGetContent(subnode);
	if (encode_pass == NULL) {
		xml_add_status(ret_data, XML_FAILURE, "Password node empty");
		goto cleanup;
	}
	encode_pass_len = strlen(encode_pass);
	status = do_sasl_decode64(encode_pass, encode_pass_len, &pass, &outlen);
	if (status == -1 || pass == NULL) {
		xml_add_status(ret_data, XML_FAILURE, "Password decode failed");
		goto cleanup;
	}

	/* initiate connection with target */
	client->cookie = send_get_cookie_command(target_name, user, pass);
	if (client->cookie == NULL) {
		xml_add_status(ret_data, XML_FAILURE, MSG_ERR_MIG_CONN_FAILED,
		    target_name);
		goto cleanup;
	}

	/*
	 * Check for the optional "dry_run" tag which indicates that only the
	 * feasibility checks should be run
	 */
	subnode = xml_find_v3_property(item, XML_DRY_RUN);
	if (subnode != NULL) {
		ASSERT(argc == 0); /* must be 1st entry in cmd line args */
		argv[argc++] = zsprintf("-n");
	}

	/*
	 * Check the force flag , aka -f, to insure that any subsequent
	 * loss of exented data is intentional/acceptable.
	 */
	subnode = xml_find_v3_property(item, XML_FORCE);
	if (subnode != NULL)
		argv[argc++] = zsprintf("-f");

	/*
	 * If 'ldom' is still NULL, the first content node in the envelope
	 * must have been the target LDom info so we must now get the
	 * source LDom info
	 */
	if (ldom == NULL) {
		/*
		 * The first content we found was the new_ldom content.  We
		 * now unlink and delete it and then get the source ldom
		 * content.
		 */
		xmlUnlinkNode(content);
		xmlFreeNode(content);
		content = xml_find_subnode(envelope, CONTENT_NODE);
		if (content == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, CONTENT_NODE);
			DBG(zprintf("Source LDom name not specified in XML"));
			goto cleanup;
		}
		ldom = (char *)xmlGetProp(content, ID_ATTR_IN);
	}

	argv[argc++] = Xstrdup((char *)ldom);
	argv[argc++] = zsprintf("%s@%s:%s", user, target_name, new_ldom);
	ASSERT(argc <= dm_args);

	/*
	 * Set the 'send_resp' variable to false so that the sequencer
	 * does not send a response now to this "long-running" command
	 * (response will be sent when the migration thread completes)
	 */
	client->send_resp = false;

	result =
	    xml_call_cli_cmd(argc, argv, LDM_MIGRATE_DOMAIN, ret_data, client);
	if (result == false) {
		/*
		 * CLI call failed, set send_resp back to true to send an
		 * XML response back to the client.
		 */
		client->send_resp = true;
	} else {
		/* set the return value to indicate everything succeeded */
		rv = 0;
	}

cleanup:
	free_argv(argv, argc);
	if (ldom != NULL)
		xmlFree(ldom);
	if (new_ldom != NULL)
		xmlFree(new_ldom);
	if (target_name != NULL)
		xmlFree(target_name);
	if (encode_user != NULL)
		xmlFree(encode_user);
	if (encode_pass != NULL)
		xmlFree(encode_pass);
	Zfree(pass);
	Zfree(user);

	return (rv);
}

/* ARGSUSED0 */
int
xml_v3_get_cookie_cmd(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client,	int cmd_num)
{
	xmlNodePtr	ret_data;
	uint64_t	random_data;
	char		cookie_data[UINT64_STR_LEN];

	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_V3_DATA_VERSION);

	int fd = open("/dev/urandom", O_RDONLY);
	if (fd == -1) {
		xml_add_status(ret_data, XML_FAILURE,
		    "/dev/urandom open failed");
		return (-1);
	}

	if (read(fd, &random_data, sizeof (random_data)) !=
	    sizeof (random_data)) {
		xml_add_status(ret_data, XML_FAILURE,
		    "failed reading random number");
		close(fd);
		return (-1);
	}

	snprintf(cookie_data, sizeof (cookie_data), "%"PRIu64, random_data);

	/* Clean out last cookie if there was one. */
	if (zeus.current_cookie != NULL) {
		Xfree(zeus.current_cookie);
	}
	zeus.current_cookie = strdup(cookie_data);
	zeus.cookie_time = time(NULL);
	xmlNewChild(ret_data, NULL, XML_COOKIE, (xmlChar *)cookie_data);

	xml_add_status(ret_data, XML_SUCCESS, NULL);
	close(fd);
	return (0);
}

/* ARGSUSED2 */
int
xml_v3_list_variable(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client,	int cmd_num)
{
	int		ret = 0, rv = 0;
	xmlNodePtr	envelope, content;
	xmlNodePtr	new_env, new_content;
	xmlChar		*section, *key;
	xmlNodePtr	ret_data, var_node, name;
	char		*ldm_name, *var_name;
	ldom_t		*ldp;

	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_V3_DATA_VERSION);

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	if (envelope != NULL) {
		content = xml_find_subnode(envelope, CONTENT_NODE);
		ldm_name = (char *)xmlGetProp(content, ID_ATTR_IN);

		if (ldm_name == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, XML_LDM_NAME);
			return (-1);
		}

		ldp = find_ldom_by_name(ldm_name);
		if (ldp == NULL) {
			xml_add_status(ret_data, XML_FAILURE, "LDom %s not "
			    "found", ldm_name);
			xmlFree((xmlChar *)ldm_name);
			return (-1);
		}

		new_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE, NULL);
		xmlNewChild(new_env, NULL, REFERENCES_NODE, NULL);
		new_content = xmlNewChild(new_env, NULL, CONTENT_NODE, NULL);
		xmlNewProp(new_content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
		xmlNewProp(new_content, ID_ATTR, (xmlChar *)ldm_name);
		xmlFree(ldm_name);

		var_node = xml_find_section(content, VAR_NODE);
		if (var_node == NULL) {
			add_v3_var_output(new_content, ldp,
			    XML_CONSTRAINTS_OUTPUT, NULL);
		} else {
			var_node = content->xmlChildrenNode;
			while (var_node != NULL) {
				section = get_section_resource_name(var_node);
				if (xmlStrcmp(section, VAR_NODE) != 0) {
					var_node =
					    xml_get_next_ele_node(var_node);
					xmlFree(section);
					continue;
				}
				xmlFree(section);
				name = xml_find_section_subnode(var_node,
				    PROPERTY_NODE_IN);
				key = xmlGetProp(name, KEY_ATTR);
				if (xmlStrcmp(key, NAME_NODE) != 0) {
					var_node =
					    xml_get_next_ele_node(var_node);
					xmlFree(key);
					continue;
				}
				xmlFree(key);

				var_name = (char *)xmlNodeGetContent(name);
				ret = add_v3_var_output(new_content, ldp, -1,
				    var_name);
				if (ret == -1) {
					xml_add_status(ret_data, XML_FAILURE,
					    "variable %s not found", var_name);
					rv = ret;
				}
				var_node = xml_get_next_ele_node(var_node);
				xmlFree((xmlChar *)var_name);
			}
		}
	} else {
		set_missing_prop_msg(ret_data, XML_DATA, ENVELOPE_NODE);
		return (-1);
	}

	if (rv == 0)
		xml_add_status(ret_data, XML_SUCCESS, NULL);

	return (rv);
}

/* ARGSUSED2 */
int
xml_v3_list_devices(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client,	int cmd_num)
{
	xmlNodePtr	ret_data, ret_env, envelope;

	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_V3_DATA_VERSION);

	ret_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE, NULL);
	xmlNewChild(ret_env, NULL, REFERENCES_NODE, NULL);

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	if (envelope == NULL) {
		add_v3_free_cpu(ret_env);
		add_v3_free_mau(ret_env);
		add_v3_free_mem(ret_env);
		add_v3_free_io(ret_env);
		xml_add_status(ret_data, XML_SUCCESS, NULL);

		return (0);
	}

	if (xml_find_section(envelope, CPU_NODE) != NULL) {
		add_v3_free_cpu(ret_env);
		xml_add_status(ret_data, XML_SUCCESS, NULL);
		return (0);
	} else if (xml_find_section(envelope, MAU_NODE) != NULL) {
		add_v3_free_mau(ret_env);
		xml_add_status(ret_data, XML_SUCCESS, NULL);
		return (0);
	} else if (xml_find_section(envelope, MEMORY_NODE) != NULL) {
		add_v3_free_mem(ret_env);
		xml_add_status(ret_data, XML_SUCCESS, NULL);
		return (0);
	} else if (xml_find_section(envelope, PHYSIO_NODE) != NULL) {
		add_v3_free_io(ret_env);
		xml_add_status(ret_data, XML_SUCCESS, NULL);
		return (0);
	}

	xml_add_status(ret_data, XML_FAILURE, "Unknown or no device type in"
	    "Envelope section\n");

	return (-1);
}

int
xml_v3_list_bindings(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client, int cmd_num)
{
	int		ret = 0;
	char		*ldm_name;
	ldom_t		*ldp;
	xmlNodePtr	ret_data, envelope, content, info_node;
	xmlNodePtr	next_data, new_env, new_content;
	xmlChar		*mac;
	xmlChar		*hostid;
	xmlChar		*failure_policy = NULL;
	xmlChar		*masters_str = NULL;

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	if (envelope != NULL) {
		content = xml_find_subnode(envelope, CONTENT_NODE);

		while (content != NULL) {
			ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
			xmlNewProp(ret_data, VERSION_ATTR,
			    LDOM_V3_DATA_VERSION);

			ldm_name = (char *)xmlGetProp(content, ID_ATTR_IN);
			if (ldm_name == NULL) {
				set_missing_prop_msg(ret_data, XML_DATA,
				    XML_LDM_NAME);
				return (-1);
			}

			new_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE,
			    NULL);
			xmlNewChild(new_env, NULL, REFERENCES_NODE, NULL);
			new_content = xmlNewChild(new_env, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(new_content, TYPE_ATTR,
			    VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(new_content, ID_ATTR, (xmlChar *)ldm_name);

			ldp = find_ldom_by_name(ldm_name);
			if (ldp == NULL) {
				xml_add_status(ret_data, XML_FAILURE,
				    "LDom %s not found", ldm_name);
				xmlFree((xmlChar *)ldm_name);
				ret = -1;
				xmlUnlinkNode(content);
				xmlFreeNode(content);
				content = xml_find_subnode(envelope,
				    CONTENT_NODE);
				continue;
			}

			xmlFree((xmlChar *)ldm_name);

			/* section==LDOM_INFO */
			info_node = create_v3_section(new_content,
			    RESOURCE_SECTION_TYPE, XML_LDM_INFO);

			/* section==LDOM_INFO  property==MAC address */
			mac = alloca(sizeof (char) * UINT64_STR_LEN);
			mactoa(ldp->props.mac_address, (char *)mac);
			xmlNewChild(info_node, NULL, ADDRESS_NODE_OUT, mac);

			/* section==LDOM_INFO  property==Hostid */
			hostid = alloca(sizeof (char) * UINT64_STR_LEN);
			if (ldp->props.hostid != UNASSIGNED_HOSTID)
				sprintf((char *)hostid, "0x%"PRIx64,
				    ldp->props.hostid);
			else
				sprintf((char *)hostid, "0x%"PRIx64,
				    DERIVED_HOSTID(ldp->props.mac_address));
			create_v3_property(info_node, XML_HOSTID, hostid);

			/* section==LDOM_INFO  property==Failure-policy */
			failure_policy = alloca(sizeof (char) * UINT64_STR_LEN);
			sprintf((char *)failure_policy, "%"PRIx64,
			    ldp->props.failure_policy);
			xmlNewChild(info_node, NULL, XML_FAILURE_POLICY,
			    failure_policy);

			/* section==LDOM_INFO  property==Master */
			masters_str = (xmlChar *)get_masters_string(ldp);
			if (masters_str != NULL) {
				xmlNewChild(info_node, NULL, XML_MASTER,
				    masters_str);
				xmlFree(masters_str);
			}

			/* section==LDOM_INFO  property==LDOM State */
			create_v3_property(info_node, STATE_NODE,
			    (xmlChar *)ldom_state_str(ldp->state));

			add_v3_cpu_output(new_content, ldp,
			    XML_BINDING_OUTPUT);
			add_v3_mau_output(new_content, ldp,
			    XML_BINDING_OUTPUT);
			add_v3_mem_output(new_content, ldp,
			    XML_BINDING_OUTPUT);
			add_v3_physio_output(new_content, ldp,
			    XML_BINDING_OUTPUT);
			add_v3_io_output(new_content, ldp->namep, ldp,
			    XML_BINDING_OUTPUT);
			add_v3_var_output(new_content, ldp,
			    XML_CONSTRAINTS_OUTPUT, NULL);
			add_v3_policy_output(new_content, ldp,
			    XML_BINDING_OUTPUT);

			xmlUnlinkNode(content);
			xmlFreeNode(content);
			content = xml_find_subnode(envelope, CONTENT_NODE);

			xml_add_status(ret_data, XML_SUCCESS, NULL);
		}
		return (ret);
	} else {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			next_data = xmlNewNode(NULL, (xmlChar *)"data");
			envelope = xmlNewChild(next_data, NULL, ENVELOPE_NODE,
			    NULL);
			content = xmlNewChild(envelope, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(content, ID_ATTR_IN, (xmlChar *)ldp->namep);

			ret = xml_v3_list_bindings(next_data, ret_cmd, client,
			    cmd_num);
			xmlFreeNode(next_data);
			if (ret == -1)
				return (-1);
		}
	}

	return (0);
}

int
xml_v3_list_constraints(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client, int cmd_num)
{
	int		ret = 0;
	char		*ldm_name;
	ldom_t		*ldp;
	xmlNodePtr	ret_data, envelope, content, info_node;
	xmlNodePtr	next_data, new_env, new_content;
	xmlChar		mac[MAC_STR_LEN] = { 0 };
	xmlChar		hostid[UINT64_STR_LEN] = { 0 };
	xmlChar		failure_policy[UINT64_STR_LEN] = { 0 };
	xmlChar		*masters_str = NULL;

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	if (envelope != NULL) {
		content = xml_find_subnode(envelope, CONTENT_NODE);

		while (content != NULL) {
			ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
			xmlNewProp(ret_data, VERSION_ATTR,
			    LDOM_V3_DATA_VERSION);

			ldm_name = (char *)xmlGetProp(content, ID_ATTR_IN);
			if (ldm_name == NULL) {
				set_missing_prop_msg(ret_data, XML_DATA,
				    XML_LDM_NAME);
				return (-1);
			}

			new_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE,
			    NULL);
			xmlNewChild(new_env, NULL, REFERENCES_NODE, NULL);
			new_content = xmlNewChild(new_env, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(new_content, TYPE_ATTR,
			    VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(new_content, ID_ATTR, (xmlChar *)ldm_name);

			ldp = find_ldom_by_name(ldm_name);
			if (ldp == NULL) {
				xml_add_status(ret_data, XML_FAILURE,
				    "LDom %s not found", ldm_name);
				xmlFree((xmlChar *)ldm_name);
				ret = -1;
				xmlUnlinkNode(content);
				xmlFreeNode(content);
				content = xml_find_subnode(envelope,
				    CONTENT_NODE);
				continue;
			}
			xmlFree((xmlChar *)ldm_name);

			/* section==LDOM_INFO */
			info_node = create_v3_section(new_content,
			    RESOURCE_SECTION_TYPE, XML_LDM_INFO);

			/* section==LDOM_INFO  property==MAC address */
			if (!is_mac_allocated(ldp->props.mac_address)) {
				mactoa(ldp->props.mac_address, (char *)mac);
				xmlNewChild(info_node, NULL, ADDRESS_NODE_OUT,
				    mac);
			}

			/* section==LDOM_INFO  property==Hostid */
			if (!IS_DERIVED_HOSTID_ASSIGNED(ldp->props.mac_address,
			    ldp->props.hostid)) {
				sprintf((char *)hostid, "0x%"PRIx64,
				    ldp->props.hostid);
				create_v3_property(info_node, XML_HOSTID,
				    hostid);
			}

			/* section==LDOM_INFO  property==Failure-policy */
			sprintf((char *)failure_policy, "%"PRIx64,
			    ldp->props.failure_policy);
			xmlNewChild(info_node, NULL, XML_FAILURE_POLICY,
			    failure_policy);

			/* section==LDOM_INFO  property==Master */
			masters_str = (xmlChar *)get_masters_string(ldp);
			if (masters_str != NULL) {
				xmlNewChild(info_node, NULL, XML_MASTER,
				    masters_str);
				xmlFree(masters_str);
			}

			/* section==LDOM_INFO  property==LDOM State */
			create_v3_property(info_node, STATE_NODE,
			    (xmlChar *)ldom_state_str(ldp->state));

			add_v3_cpu_output(new_content, ldp,
			    XML_CONSTRAINTS_OUTPUT);
			add_v3_mau_output(new_content, ldp,
			    XML_CONSTRAINTS_OUTPUT);
			add_v3_mem_output(new_content, ldp,
			    XML_CONSTRAINTS_OUTPUT);
			add_v3_physio_output(new_content, ldp,
			    XML_CONSTRAINTS_OUTPUT);
			add_v3_io_output(new_content, ldp->namep, ldp,
			    XML_CONSTRAINTS_OUTPUT);
			add_v3_var_output(new_content, ldp,
			    XML_CONSTRAINTS_OUTPUT, NULL);
			add_v3_policy_output(new_content, ldp,
			    XML_CONSTRAINTS_OUTPUT);

			xmlUnlinkNode(content);
			xmlFreeNode(content);
			content = xml_find_subnode(envelope, CONTENT_NODE);

			xml_add_status(ret_data, XML_SUCCESS, NULL);
		}
		return (ret);
	} else {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			next_data = xmlNewNode(NULL, (xmlChar *)"data");
			envelope = xmlNewChild(next_data, NULL, ENVELOPE_NODE,
			    NULL);
			content = xmlNewChild(envelope, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(content, ID_ATTR_IN, (xmlChar *)ldp->namep);

			ret = xml_v3_list_constraints(next_data, ret_cmd,
			    client, cmd_num);
			xmlFreeNode(next_data);
			if (ret == -1)
				return (-1);
		}
	}
	return (0);
}

int
xml_v3_list(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
    int cmd_num)
{
	int		ret = 0;
	char		*ldm_name;
	ldom_t		*ldp;
	xmlNodePtr	ret_data, envelope, content, info_node;
	xmlNodePtr	next_data, new_env, new_content, new_cons;

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	if (envelope != NULL) {
		content = xml_find_subnode(envelope, CONTENT_NODE);

		while (content != NULL) {
			ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
			xmlNewProp(ret_data, VERSION_ATTR,
			    LDOM_V3_DATA_VERSION);

			ldm_name = (char *)xmlGetProp(content, ID_ATTR_IN);
			if (ldm_name == NULL) {
				set_missing_prop_msg(ret_data, XML_DATA,
				    XML_LDM_NAME);
				return (-1);
			}

			new_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE,
			    NULL);
			xmlNewChild(new_env, NULL, REFERENCES_NODE, NULL);
			new_content = xmlNewChild(new_env, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(new_content, TYPE_ATTR,
			    VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(new_content, ID_ATTR, (xmlChar *)ldm_name);

			ldp = find_ldom_by_name(ldm_name);
			if (ldp == NULL) {
				xml_add_status(ret_data, XML_FAILURE,
				    "LDom %s not found", ldm_name);
				xmlFree((char *)ldm_name);
				ret = -1;
				xmlUnlinkNode(content);
				xmlFreeNode(content);
				content = xml_find_subnode(envelope,
				    CONTENT_NODE);
				continue;
			}
			xmlFree((char *)ldm_name);

			info_node = create_v3_section(new_content,
			    RESOURCE_SECTION_TYPE, XML_LDM_INFO);
			update_statistics(ldp, true);
			/* Domain State */
			create_v3_property(info_node, STATE_NODE,
			    (xmlChar *)ldom_state_str(ldp->state));

			/* Domain flags */
			create_v3_property(info_node, XML_FLAGS,
			    (xmlChar *)ldom_flags(ldp));

			add_v3_cpu_output(new_content, ldp,
			    XML_CONSTRAINTS_OUTPUT);
			add_v3_mem_output(new_content, ldp,
			    XML_CONSTRAINTS_OUTPUT);
			if (ldp->console.binding != NULL) {
				new_cons =  create_v3_section(new_content,
				    VIRTUAL_HARDWARE_TYPE, CONSOLE_NODE);
				create_v3_property(new_cons, XML_PORT,
				    (xmlChar *)console_print(ldp));
			}

			/* CPU utilization and uptime if applicable */
			if (ldp->state > BOUND) {
				uint64_t uptime, days, hours, mins;
				char	utilstr[8], uptimestr[64];
				if (ldp->stats.cpu >= 10.0)
					snprintf(utilstr, sizeof (utilstr),
					    "%.0f%%", ldp->stats.cpu);
				else
					snprintf(utilstr, sizeof (utilstr),
					    "%.1f%%", ldp->stats.cpu);
				create_v3_property(info_node, XML_UTIL,
				    (xmlChar *)utilstr);

				uptime = ldp->stats.uptime + 30;
				days = uptime / (60 * 60 * 24);
				uptime %= (60 * 60 * 24);
				hours = uptime / (60 * 60);
				uptime %= (60 * 60);
				mins = uptime / 60;
				strcpy(uptimestr, "");
				if (days > 0)
					sprintf(uptimestr, "%"PRIu64"d", days);
				if (hours > 0)
					sprintf(uptimestr + strlen(uptimestr),
					    "%"PRIu64"h", hours);
				if ((days < 100) && (mins > 0))
					sprintf(uptimestr + strlen(uptimestr),
					    "%"PRIu64"m", mins);

				if (strlen(uptimestr) == 0)
					sprintf(uptimestr, "%"PRIu64"s",
					    ldp->stats.uptime);
				create_v3_property(info_node, XML_UPTIME,
				    (xmlChar *)uptimestr);
			}

			xmlUnlinkNode(content);
			xmlFreeNode(content);
			content = xml_find_subnode(envelope, CONTENT_NODE);
			xml_add_status(ret_data, XML_SUCCESS, NULL);
		}
		return (ret);
	} else {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			next_data = xmlNewNode(NULL, (xmlChar *)"data");
			envelope = xmlNewChild(next_data, NULL, ENVELOPE_NODE,
			    NULL);
			content = xmlNewChild(envelope, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(content, ID_ATTR_IN, (xmlChar *)ldp->namep);

			ret = xml_v3_list(next_data, ret_cmd,
			    client, cmd_num);
			xmlFreeNode(next_data);
			if (ret == -1)
				return (-1);
		}
	}

	return (0);
}

int
xml_v3_list_services(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client,	int cmd_num)
{
	int		ret = 0;
	char		*ldm_name;
	ldom_t		*ldp;
	xmlNodePtr	ret_data, envelope, content;
	xmlNodePtr	next_data, new_env, new_content;

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	if (envelope != NULL) {
		content = xml_find_subnode(envelope, CONTENT_NODE);

		while (content != NULL) {
			ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
			xmlNewProp(ret_data, VERSION_ATTR,
			    LDOM_V3_DATA_VERSION);

			ldm_name = (char *)xmlGetProp(content, ID_ATTR_IN);
			if (ldm_name == NULL) {
				set_missing_prop_msg(ret_data, XML_DATA,
				    XML_LDM_NAME);
				return (-1);
			}

			new_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE,
			    NULL);
			xmlNewChild(new_env, NULL, REFERENCES_NODE, NULL);
			new_content = xmlNewChild(new_env, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(new_content, TYPE_ATTR,
			    VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(new_content, ID_ATTR, (xmlChar *)ldm_name);

			ldp = find_ldom_by_name(ldm_name);
			if (ldp == NULL) {
				xml_add_status(ret_data, XML_FAILURE,
				    "LDom %s not found", ldm_name);
				xmlFree((xmlChar *)ldm_name);
				ret = -1;
				xmlUnlinkNode(content);
				xmlFreeNode(content);
				content = xml_find_subnode(envelope,
				    CONTENT_NODE);
				continue;
			}
			xmlFree((xmlChar *)ldm_name);

			add_v3_vsw_output(new_content, ldp,
			    XML_SERVICES_OUTPUT);
			add_v3_vcc_output(new_content, ldp,
			    XML_SERVICES_OUTPUT);
			add_v3_vds_output(new_content, ldp,
			    XML_SERVICES_OUTPUT);
			add_v3_vdpcs_output(new_content, ldp,
			    XML_SERVICES_OUTPUT);

			xmlUnlinkNode(content);
			xmlFreeNode(content);
			content = xml_find_subnode(envelope, CONTENT_NODE);

			xml_add_status(ret_data, XML_SUCCESS, NULL);
		}
		return (ret);
	} else {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			next_data = xmlNewNode(NULL, (xmlChar *)"data");
			envelope = xmlNewChild(next_data, NULL, ENVELOPE_NODE,
			    NULL);
			content = xmlNewChild(envelope, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(content, ID_ATTR_IN, (xmlChar *)ldp->namep);

			ret = xml_v3_list_services(next_data, ret_cmd,
			    client, cmd_num);
			xmlFreeNode(next_data);
			if (ret == -1)
				return (-1);
		}
	}
	return (ret);
}

int
xml_v3_list_autosave(xmlNodePtr data_node, xmlNodePtr ret_data, int n_options,
	char **options, cli_conn_t *client)
{
	int		i;
	char		**mdset_names = NULL;
	int		nmdsets = 0;
	int		ret = 0;
	char		*mdset_name = NULL;
	xmlNodePtr	new_env, section, item, prop;

	ASSERT(ret_data != NULL);
	ASSERT(n_options > 0);
	ASSERT(options[0] != NULL);

	if ((strcmp(options[0], "-r") != 0) || (n_options > 2)) {
		xml_add_status(ret_data, XML_FAILURE,
		    "Options for list-spconfig: [-r [<autosave>]]");
		return (-1);
	}
	if ((n_options == 2) && valid_name(client->resp, options[1]) == false) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_AUTOSAVE_CONFIG_NAME_INVALID, options[1]);
		return (-1);
	}

	if (n_options == 2) {
		mdset_name = options[1];
	} else {
		mdset_name = NULL;
	}

	if (autosave_get_mdsets(mdset_name, &nmdsets, &mdset_names) != 0) {
		xml_add_status(ret_data, XML_FAILURE, "Failed to get mdsets");
		return (-1);
	}

	if (nmdsets == 0) {
		if (mdset_name != NULL) {
			xml_add_status(ret_data, XML_FAILURE,
			    MSG_ERR_AUTOSAVE_NO_SUCH_CONFIG, mdset_name);
		} else {
			xml_add_status(ret_data, XML_FAILURE,
			    MSG_ERR_AUTOSAVE_NO_CONFIGS);
		}
		return (-1);
	}
	new_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE, NULL);
	xmlNewChild(new_env, NULL, REFERENCES_NODE, NULL);
	section = xmlNewChild(new_env, NULL, SECTION_NODE, NULL);
	for (i = 0; i < nmdsets; i++) {
		item = xmlNewChild(section, NULL, ITEM_NODE, NULL);
		xmlNewChild(item, NULL, OTHER_RESOURCE_NODE_OUT,
		    XML_SPCONFIG);
		prop = xmlNewChild(item, NULL, PROPERTY_NODE_OUT,
		    (xmlChar *)mdset_names[i]);
		xmlNewProp(prop, KEY_ATTR, XML_SPCONFIG_NAME);
		if (autosave_autosave_is_newer(mdset_names[i])) {
			prop = xmlNewChild(item, NULL,
			    PROPERTY_NODE_OUT, (xmlChar *)"newer");
			xmlNewProp(prop, KEY_ATTR, XML_SPCONFIG_STATUS);
		}
	}
	autosave_free_list(mdset_names, nmdsets);
	return (ret);
}

/* ARGSUSED0 */
int
xml_v3_list_sets(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{
	xmlNodePtr	ret_data;
	int		ret = 0;
	mdstore_f	mdstore;
	int		n_options = 0;
	char		**options = NULL;

	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	ASSERT(ret_data != NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_V3_DATA_VERSION);

	n_options = parse_options_node(ret_cmd, &options);
	if (n_options == -1) {
		xml_add_status(ret_data->parent, XML_FAILURE,
		    "<%s> tag contents too long", XML_OPTIONS);
		return (-1);
	}
	if (n_options != 0) {
		ret = xml_v3_list_autosave(data_node, ret_data, n_options,
		    options, client);
		return (ret);
	}
	mdstore = (mdstore_f)ds_channel_service_routine(ds_sp_channel(),
	    "mdstore", "mdstore");
	if (mdstore == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    "mdstore routine failed to load");
		return (-1);
	}

	ret = (*mdstore)("list", NULL, NULL, ret_data);
	if (ret != 0) {
		xml_add_status(ret_data, XML_FAILURE,
		    "list-set operation failed");
		return (-1);
	}

	xml_add_status(ret_data, XML_SUCCESS, NULL);
	return (ret);
}

/*
 * xml_v3_create:  Note calling functions should not use the "data_node"
 * argument passed in after this function returns as the <content> nodes are
 * destroyed as they are detected. This is because there can be multiple
 * <content> nodes in an envelope.
 */
/* ARGSUSED3 */
int
xml_v3_create(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
    int cmd_num)
{
	int		ret = 0, rv = 0;
	xmlNodePtr	subnode, ret_data, ldom_info, envelope, content;
	xmlNodePtr	new_env, new_content, ldom_info_section;
	char		*ldm_name;
	ldom_t		*ldp;
	const int	create_args = 5;
	int		argc = 0;
	void		*argv[create_args];
	xmlChar		*mac_addr, *section_name;
	xmlChar		*hostid;
	xmlChar		*master;
	xmlChar		*failure_policy;

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	if (envelope == NULL) {
		set_missing_prop_msg(ret_cmd, XML_DATA, ENVELOPE_NODE);
		return (-1);
	}

	content = xml_find_subnode(envelope, CONTENT_NODE);
	if (content == NULL) {
		set_missing_prop_msg(ret_cmd, ENVELOPE_NODE, CONTENT_NODE);
		return (-1);
	}

	while (content != NULL) {
		ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
		xmlNewProp(ret_data, VERSION_ATTR, LDOM_V3_DATA_VERSION);

		ldm_name = (char *)xmlGetProp(content, ID_ATTR_IN);
		if (ldm_name == NULL) {
			set_missing_prop_msg(ret_data, XML_DATA, XML_LDM_NAME);
			rv = -1;
			xmlUnlinkNode(content);
			xmlFreeNode(content);
			content = xml_find_subnode(envelope, CONTENT_NODE);
			continue;
		}

		new_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE, NULL);
		xmlNewChild(new_env, NULL, REFERENCES_NODE, NULL);
		new_content = xmlNewChild(new_env, NULL, CONTENT_NODE, NULL);
		xmlNewProp(new_content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
		xmlNewProp(new_content, ID_ATTR, (xmlChar *)ldm_name);

		ldom_info_section = xml_find_section(content, XML_LDM_INFO);
		if (ldom_info_section != NULL) {
			ldom_info = xml_find_subnode(ldom_info_section,
			    ITEM_NODE);
			subnode = xml_find_subnode(ldom_info, ADDRESS_NODE_IN);
			if (subnode != NULL) {
				mac_addr = xmlNodeGetContent(subnode);
				if (strcmp((char *)mac_addr,
				    XML_AUTO_ALLOCATED_MACADDR) != 0) {
					argv[argc++] = zsprintf("mac-addr=%s",
					    mac_addr);
				}
				xmlFree(mac_addr);
			}
			subnode = xml_find_v3_property(ldom_info, XML_HOSTID);
			if (subnode != NULL) {
				hostid = xmlNodeGetContent(subnode);
				if (hostid != NULL) {
					uint64_t h_id =
					    strtoll((char *)hostid, NULL, 16);
					if (h_id != UNASSIGNED_HOSTID) {
						argv[argc++] =
						    zsprintf("hostid=%s",
						    hostid);
					}
				}
				xmlFree(hostid);
			}
			subnode = xml_find_v3_property(ldom_info,
			    XML_FAILURE_POLICY);
			if (subnode != NULL) {
				failure_policy = xmlNodeGetContent(subnode);
				argv[argc++] = zsprintf("failure-policy=%s",
				    failure_policy);
				xmlFree(failure_policy);
			}
			subnode = xml_find_v3_property(ldom_info, XML_MASTER);
			if (subnode != NULL) {
				master = xmlNodeGetContent(subnode);
				argv[argc++] = zsprintf("master=%s", master);
				xmlFree(master);
			}
		}
		argv[argc++] = Xstrdup((char *)ldm_name);
		ret = xml_call_cli_cmd(argc, argv, LDM_CREATE, ret_data,
		    client);
		free_argv(argv, argc);
		argc = 0;

		/* If creation of ldom failed return error */
		if (ret == false) {
			rv = -1;
			xmlUnlinkNode(content);
			xmlFreeNode(content);
			content = xml_find_subnode(envelope, CONTENT_NODE);
			xmlFree((xmlChar *)ldm_name);
			continue;
		}

		/* Find LDOM pointer and get first Resorce node */
		ldp = find_ldom_by_name(ldm_name);
		xmlFree((xmlChar *)ldm_name);
		subnode = content->xmlChildrenNode;
		while (subnode != NULL) {
			if (subnode->type != XML_ELEMENT_NODE) {
				subnode = subnode->next;
				continue;
			}

			section_name = get_section_resource_name(subnode);
			if (section_name == NULL) {
				subnode = subnode->next;
				xmlFree(section_name);
				continue;
			}
			if (xmlStrcmp(section_name,
			    (const xmlChar *)XML_LDM_INFO) == 0) {
				subnode = subnode->next;
				xmlFree(section_name);
				continue;
			}
			ret = goto_resource_routine((char *)section_name,
			    ret_data, ldp, XML_ADD_RES, subnode, client);
			xmlFree(section_name);
			if (ret == -1) {
				rv = -1;
				DBG(zprintf("Failed working with node %s",
				    (char *)subnode->name));
			}
			subnode = subnode->next;
		}

		if (rv == 0)
			xml_add_status(ret_data, XML_SUCCESS, NULL);
		xmlUnlinkNode(content);
		xmlFreeNode(content);
		content = xml_find_subnode(envelope, CONTENT_NODE);
	}

	return (rv);
}

/*
 * xml_v3_start:  Note calling functions should not use the "data_node"
 * arguement passed in after this function returns as the <content> nodes are
 * destroyed as they are detected. This is because there can be multiple
 * <content> nodes in an envelope.
 */
int
xml_v3_start(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
    int cmd_num)
{
	int		ret, rv = 0;
	xmlNodePtr	envelope, content, new_env, new_content;
	xmlNodePtr	ret_data, next_data, data_copy;
	char		*ldm_name;
	ldom_t		*ldp;
	int		argc = SINGLE_CLI_ARG;
	void		*argv[argc];

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	if (envelope != NULL) {
		content = xml_find_subnode(envelope, CONTENT_NODE);

		while (content != NULL) {
			ret_data = xmlNewNode(NULL, XML_DATA);
			xmlNewProp(ret_data, VERSION_ATTR,
			    LDOM_V3_DATA_VERSION);

			ldm_name = (char *)xmlGetProp(content, ID_ATTR_IN);
			if (ldm_name == NULL) {
				set_missing_prop_msg(ret_data, LDOM_NODE,
				    XML_LDM_NAME);
				rv = -1;
				xmlAddChild(ret_cmd, ret_data);
				xmlUnlinkNode(content);
				xmlFreeNode(content);
				content = xml_find_subnode(envelope,
				    CONTENT_NODE);
				continue;
			}

			ldp = find_ldom_by_name(ldm_name);
			if (ldp == NULL) {
				data_copy = xmlCopyNode(data_node, 1);
				ret = xml_v3_create(data_copy, ret_cmd, client,
				    LDM_CREATE);
				xmlFreeNode(data_copy);
				if (ret == -1) {
					xmlFree((xmlChar *)ldm_name);
					rv = -1;
					xmlUnlinkNode(content);
					xmlFreeNode(content);
					content = xml_find_subnode(envelope,
					    CONTENT_NODE);
					continue;
				}
				/*
				 * Remove the <data> section created by
				 * xml_v3_create() as we will replace it later
				 */
				data_copy = xml_find_data_section(ret_cmd,
				    ldm_name);
				xmlUnlinkNode(data_copy);
				xmlFreeNode(data_copy);
				ldp = find_ldom_by_name(ldm_name);
			}
			if (ldp->state == INACTIVE) {
				data_copy = xmlCopyNode(data_node, 1);
				ret = xml_v3_bind(data_copy, ret_cmd, client,
				    LDM_BIND);
				xmlFreeNode(data_copy);
				if (ret == -1) {
					xmlFree((xmlChar *)ldm_name);
					rv = -1;
					xmlUnlinkNode(content);
					xmlFreeNode(content);
					content = xml_find_subnode(envelope,
					    CONTENT_NODE);
					continue;
				}
				/*
				 * Remove the <data> section created by
				 * xml_v3_bind() as we will replace it later
				 */
				data_copy = xml_find_data_section(ret_cmd,
				    ldm_name);
				xmlUnlinkNode(data_copy);
				xmlFreeNode(data_copy);
			}

			xmlAddChild(ret_cmd, ret_data);
			new_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE,
			    NULL);
			new_content = xmlNewChild(new_env, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(new_content, TYPE_ATTR,
			    VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(new_content, ID_ATTR, (xmlChar *)ldm_name);

			argv[0] = Xstrdup((char *)ldm_name);
			xmlFree((xmlChar *)ldm_name);

			ret = xml_call_cli_cmd(argc, argv, cmd_num, ret_data,
			    client);
			free_argv(argv, argc);

			if (ret == true) {
				xml_add_status(ret_data, XML_SUCCESS, NULL);
			} else {
				rv = -1;
			}
			xmlUnlinkNode(content);
			xmlFreeNode(content);
			content = xml_find_subnode(envelope, CONTENT_NODE);
		}
	} else {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			next_data = xmlNewNode(NULL, (xmlChar *)"data");
			envelope = xmlNewChild(next_data, NULL, ENVELOPE_NODE,
			    NULL);
			content = xmlNewChild(envelope, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(content, ID_ATTR_IN, (xmlChar *)ldp->namep);

			ret = xml_v3_start(next_data, ret_cmd,
			    client, cmd_num);
			xmlFreeNode(next_data);
			if (ret == -1)
				rv = -1;
		}
	}

	return (rv);
}

/*
 * xml_v3_bind:  Note calling functions should not use the "data_node"
 * arguement passed in after this function returns as the <content> nodes are
 * destroyed as they are detected. This is because there can be multiple
 * <content> nodes in an envelope.
 */
int
xml_v3_bind(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
    int cmd_num)
{
	int		ret, rv = 0;
	xmlNodePtr	envelope, content, new_env, new_content;
	xmlNodePtr	ret_data, next_data, data_copy;
	char		*ldm_name;
	ldom_t		*ldp;
	int		argc = SINGLE_CLI_ARG;
	void		*argv[argc];

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	if (envelope != NULL) {
		content = xml_find_subnode(envelope, CONTENT_NODE);

		while (content != NULL) {
			ret_data = xmlNewNode(NULL, XML_DATA);
			xmlNewProp(ret_data, VERSION_ATTR,
			    LDOM_V3_DATA_VERSION);

			ldm_name = (char *)xmlGetProp(content, ID_ATTR_IN);
			if (ldm_name == NULL) {
				set_missing_prop_msg(ret_data, LDOM_NODE,
				    XML_LDM_NAME);
				rv = -1;
				xmlAddChild(ret_cmd, ret_data);
				xmlUnlinkNode(content);
				xmlFreeNode(content);
				content = xml_find_subnode(envelope,
				    CONTENT_NODE);
				continue;
			}

			ldp = find_ldom_by_name(ldm_name);
			if (ldp == NULL) {
				data_copy = xmlCopyNode(data_node, 1);
				ret = xml_v3_create(data_copy, ret_cmd, client,
				    LDM_CREATE);
				xmlFreeNode(data_copy);
				if (ret == -1) {
					xmlFree((xmlChar *)ldm_name);
					rv = -1;
					xmlUnlinkNode(content);
					xmlFreeNode(content);
					content = xml_find_subnode(envelope,
					    CONTENT_NODE);
					continue;
				}
				/*
				 * Remove the <data> section created by
				 * xml_v3_create() as we will replace it later
				 */
				data_copy = xml_find_data_section(ret_cmd,
				    ldm_name);
				xmlUnlinkNode(data_copy);
				xmlFreeNode(data_copy);
			}

			xmlAddChild(ret_cmd, ret_data);
			new_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE,
			    NULL);
			new_content = xmlNewChild(new_env, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(new_content, TYPE_ATTR,
			    VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(new_content, ID_ATTR, (xmlChar *)ldm_name);

			argv[0] = Xstrdup((char *)ldm_name);
			xmlFree((xmlChar *)ldm_name);

			ret = xml_call_cli_cmd(argc, argv, cmd_num, ret_data,
			    client);
			free_argv(argv, argc);

			if (ret == true) {
				xml_add_status(ret_data, XML_SUCCESS, NULL);
			} else {
				rv = -1;
			}
			xmlUnlinkNode(content);
			xmlFreeNode(content);
			content = xml_find_subnode(envelope, CONTENT_NODE);
		}
	} else {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			next_data = xmlNewNode(NULL, (xmlChar *)"data");
			envelope = xmlNewChild(next_data, NULL, ENVELOPE_NODE,
			    NULL);
			content = xmlNewChild(envelope, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(content, ID_ATTR_IN, (xmlChar *)ldp->namep);

			ret = xml_v3_bind(next_data, ret_cmd,
			    client, cmd_num);
			xmlFreeNode(next_data);
			if (ret == -1)
				rv = -1;
		}
	}

	return (rv);
}

int
xml_v3_common_cmd(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{
	int		ret = 0, rv = 0, i;
	char		*ldm_name;
	ldom_t		*ldp, *nextldp;
	xmlNodePtr	envelope, content, new_env, new_content;
	xmlNodePtr	next_data, ret_data;
	int		n_options = 0;
	char		**options = NULL;
	int		cmd_args = SINGLE_CLI_ARG;
	int		argc = 0;

	n_options = parse_options_node(ret_cmd, &options);
	if (n_options == -1) {
		xml_add_status(ret_cmd, XML_FAILURE, "<%s> tag contents too"
		    "long", XML_OPTIONS);
		return (-1);
	}
	cmd_args += n_options;

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	if (envelope != NULL) {
		void		*argv[cmd_args];

		content = xml_find_subnode(envelope, CONTENT_NODE);

		while (content != NULL) {
			/* Add in any options to the argv */
			if (n_options != 0) {
				for (i = 0; i < n_options; i++) {
					argv[argc++] = Xstrdup(options[i]);
				}
			}
			ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
			xmlNewProp(ret_data, VERSION_ATTR,
			    LDOM_V3_DATA_VERSION);

			ldm_name = (char *)xmlGetProp(content, ID_ATTR_IN);
			if (ldm_name == NULL) {
				set_missing_prop_msg(ret_data, LDOM_NODE,
				    XML_LDM_NAME);
				rv = -1;
				xmlUnlinkNode(content);
				xmlFreeNode(content);
				content = xml_find_subnode(envelope,
				    CONTENT_NODE);
				free_argv(argv, argc);
				argc = 0;
				continue;
			}

			new_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE,
			    NULL);
			new_content = xmlNewChild(new_env, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(new_content, TYPE_ATTR,
			    VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(new_content, ID_ATTR, (xmlChar *)ldm_name);

			argv[argc++] = Xstrdup((char *)ldm_name);
			xmlFree((xmlChar *)ldm_name);

			ret = xml_call_cli_cmd(argc, argv, cmd_num, ret_data,
			    client);
			free_argv(argv, argc);
			argc = 0;

			if (ret == true)
				xml_add_status(ret_data, XML_SUCCESS, NULL);
			else
				rv = -1;

			xmlUnlinkNode(content);
			xmlFreeNode(content);
			content = xml_find_subnode(envelope, CONTENT_NODE);
		}
	} else {
		CHAIN_ITER_VOL(zeus.ldoms, ldp, nextldp, link) {
			next_data = xmlNewNode(NULL, (xmlChar *)"data");
			envelope = xmlNewChild(next_data, NULL, ENVELOPE_NODE,
			    NULL);
			content = xmlNewChild(envelope, NULL, CONTENT_NODE,
			    NULL);
			xmlNewProp(content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
			xmlNewProp(content, ID_ATTR_IN, (xmlChar *)ldp->namep);
			ret = xml_v3_common_cmd(next_data, ret_cmd, client,
			    cmd_num);
			xmlFreeNode(next_data);
			if (ret == -1)
				rv = -1;
		}
	}
	if (options != NULL) {
		free_argv((void **)options, n_options);
		Xfree(options);
	}
	return (rv);
}

/*
 * The again bool_t is when we wish to look for a 2nd resource in the same
 * content section.
 *  True == Do NOT set missing property error.
 *  False == SET the missing property error.
 */
static xmlNodePtr
get_v3_resource_node(xmlNodePtr data_node, xmlNodePtr ret_data, int cmd_num,
	bool_t report_err)
{
	xmlNodePtr	res_node = NULL;
	xmlNodePtr	envelope, content;
	xmlChar		*parent = NULL, *node = NULL;

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);

	switch (cmd_num) {
	case LDM_SET:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, XML_LDM_INFO);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = CPU_NODE;
		}
		break;
	case LDM_ADD_VCPU:
	case LDM_MODIFY_VCPU:
	case LDM_REMOVE_VCPU:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, CPU_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = CPU_NODE;
		}
		break;
	case LDM_ADD_CRYPTO:
	case LDM_MODIFY_CRYPTO:
	case LDM_REMOVE_CRYPTO:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, MAU_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = MAU_NODE;
		}
		break;
	case LDM_ADD_MEM:
	case LDM_MODIFY_MEM:
	case LDM_REMOVE_MEM:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, MEMORY_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = MEMORY_NODE;
		}
		break;
	case LDM_ADD_IO:
	case LDM_REMOVE_IO:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, XML_PHYSIO_DEVICE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = XML_PHYSIO_DEVICE;
		}
		break;
	case LDM_SET_VAR:
	case LDM_UNSET_VAR:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, VAR_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = VAR_NODE;
		}
		break;
	case LDM_ADD_VCC:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, VCC_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = VCC_NODE;
		}
		break;
	case LDM_MODIFY_VCC:
	case LDM_REMOVE_VCC:
		res_node = xml_find_section(envelope, VCC_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = ENVELOPE_NODE;
			node = VCC_NODE;
		}
		break;
	case LDM_ADD_VDISK:
	case LDM_MODIFY_VDISK:
	case LDM_REMOVE_VDISK:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, DISK_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = DISK_NODE;
		}
		break;
	case LDM_ADD_VDS:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, VDS_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = VDS_NODE;
		}
		break;
	case LDM_REMOVE_VDS:
		res_node = xml_find_section(envelope, VDS_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = ENVELOPE_NODE;
			node = VDS_NODE;
		}
		break;
	case LDM_ADD_VDSDEV:
	case LDM_MODIFY_VDSDEV:
	case LDM_REMOVE_VDSDEV:
		res_node = xml_find_section(envelope, VDS_VOLUME_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = ENVELOPE_NODE;
			node = VDS_VOLUME_NODE;
		}
		break;
	case LDM_ADD_VNET:
	case LDM_MODIFY_VNET:
	case LDM_REMOVE_VNET:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, VNET_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = VNET_NODE;
		}
		break;
	case LDM_ADD_VSW:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, VSW_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = VSW_NODE;
		}
		break;
	case LDM_MODIFY_VSW:
	case LDM_REMOVE_VSW:
		res_node = xml_find_section(envelope, VSW_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = ENVELOPE_NODE;
			node = VSW_NODE;
		}
		break;
	case LDM_ADD_SET:
	case LDM_USE_SET:
	case LDM_REMOVE_SET:
		res_node = xml_find_section(envelope, XML_SPCONFIG);
		if ((res_node == NULL) && (report_err == true)) {
			parent = ENVELOPE_NODE;
			node = XML_SPCONFIG;
		}
		break;
	case LDM_MODIFY_VCONS:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, CONSOLE_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = CONSOLE_NODE;
		}
		break;
	case LDM_ADD_VDPCS:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, VDPCS_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = VDPCS_NODE;
		}
		break;
	case LDM_REMOVE_VDPCS:
		res_node = xml_find_section(envelope, VDPCS_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = ENVELOPE_NODE;
			node = VDPCS_NODE;
		}
		break;
	case LDM_ADD_VDPCC:
	case LDM_REMOVE_VDPCC:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, VDPCC_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = VDPCC_NODE;
		}
		break;
	case LDM_ADD_POLICY:
	case LDM_MODIFY_POLICY:
	case LDM_REMOVE_POLICY:
		content = xml_find_subnode(envelope, CONTENT_NODE);
		res_node = xml_find_section(content, POLICY_NODE);
		if ((res_node == NULL) && (report_err == true)) {
			parent = CONTENT_NODE;
			node = POLICY_NODE;
		}
		break;
	}

	if (parent != NULL) {
		set_missing_prop_msg(ret_data, parent, node);
	}

	return (res_node);
}

int
xml_handle_resources(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client, int action, int cmd_num, char *ldm_name)
{
	int		rv = 0;
	xmlNodePtr	ret_data, new_env, new_content;
	xmlNodePtr	res_node;
	xmlChar		*section_name;
	ldom_t		*ldp = NULL;

	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_V3_DATA_VERSION);
	new_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE, NULL);
	xmlNewChild(new_env, NULL, REFERENCES_NODE, NULL);

	if (ldm_name != NULL) {
		new_content = xmlNewChild(new_env, NULL, CONTENT_NODE, NULL);
		xmlNewProp(new_content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
		xmlNewProp(new_content, ID_ATTR, (xmlChar *)ldm_name);

		ldp = find_ldom_by_name(ldm_name);
		if (ldp == NULL) {
			xml_add_status(ret_data, XML_FAILURE,
			    "LDom %s not found", ldm_name);
			return (-1);
		}
	}

	res_node = get_v3_resource_node(data_node, ret_data, cmd_num, true);
	if (res_node == NULL)
		return (-1);

	while (res_node != NULL) {
		int	ret = 0;

		section_name = get_section_resource_name(res_node);
		if (section_name == NULL) {
			xml_add_status(ret_data, XML_FAILURE,
			    "Section name missing");
			rv = -1;
			xmlUnlinkNode(res_node);
			xmlFreeNode(res_node);
			res_node = get_v3_resource_node(data_node, ret_data,
			    cmd_num, false);
			continue;
		}

		ret = goto_resource_routine((char *)section_name, ret_data,
		    ldp, action, res_node, client);
		xmlFree(section_name);
		if (ret != 0)
			rv = -1;
		xmlUnlinkNode(res_node);
		xmlFreeNode(res_node);
		res_node = get_v3_resource_node(data_node, ret_data, cmd_num,
		    false);
	}
	if (rv == 0)
		xml_add_status(ret_data, XML_SUCCESS, NULL);
	return (rv);
}

/*
 * This routine handles ALL commands which must have a <content> section to
 * run.  I have taken it out of the add, set, remove commands because of the
 * fact that we can have multiple content sections in an envelope and it has
 * to be handled differently than if we did not have a content section.
 */
int xml_v3_content_cmd(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client, int cmd_num, int action)
{
	int		rv = 0;
	xmlNodePtr	ret_data;
	xmlNodePtr	envelope, content, new_env;
	char		*ldm_name;

	envelope = xml_find_subnode(data_node, ENVELOPE_NODE);
	content = xml_find_subnode(envelope, CONTENT_NODE);

	while (content != NULL) {
		int	res_ret = 0;

		ldm_name = (char *)xmlGetProp(content, ID_ATTR_IN);
		if (ldm_name == NULL) {
			ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
			xmlNewProp(ret_data, VERSION_ATTR,
			    LDOM_V3_DATA_VERSION);
			new_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE,
			    NULL);
			xmlNewChild(new_env, NULL, REFERENCES_NODE, NULL);
			set_missing_prop_msg(ret_data, LDOM_NODE,
			    XML_LDM_NAME);
			rv = -1;
			xmlUnlinkNode(content);
			xmlFreeNode(content);
			content = xml_find_subnode(envelope, CONTENT_NODE);
			continue;
		}

		res_ret = xml_handle_resources(data_node, ret_cmd, client,
		    action, cmd_num, ldm_name);
		if (res_ret == -1) {
			rv = -1;
		}

		xmlFree((xmlChar *)ldm_name);
		xmlUnlinkNode(content);
		xmlFreeNode(content);
		content = xml_find_subnode(envelope, CONTENT_NODE);
	}

	return (rv);
}

int
xml_v3_add_cmd(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{
	/*
	 * If device is not a vdsdev then we will need to parse the one or
	 * more content tags in the envelope.
	 */
	if ((cmd_num != LDM_ADD_VDSDEV) && (cmd_num != LDM_ADD_SET)) {
		return (xml_v3_content_cmd(data_node, ret_cmd, client, cmd_num,
		    XML_ADD_RES));
	}

	return (xml_handle_resources(data_node, ret_cmd, client,
	    XML_ADD_RES, cmd_num, NULL));
}

int
xml_v3_set_cmd(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{
	switch (cmd_num) {
	case LDM_SET:
	case LDM_MODIFY_CRYPTO:
	case LDM_MODIFY_MEM:
	case LDM_SET_VAR:
	case LDM_MODIFY_VCPU:
	case LDM_MODIFY_VNET:
	case LDM_MODIFY_VCONS:
	case LDM_MODIFY_VDISK:
	case LDM_MODIFY_POLICY:
		return (xml_v3_content_cmd(data_node, ret_cmd, client, cmd_num,
		    XML_MODIFY_RES));
	}

	return (xml_handle_resources(data_node, ret_cmd, client,
	    XML_MODIFY_RES, cmd_num, NULL));
}

int
xml_v3_remove_cmd(xmlNodePtr data_node, xmlNodePtr ret_cmd, cli_conn_t *client,
	int cmd_num)
{

	switch (cmd_num) {
	case LDM_REMOVE_IO:
	case LDM_REMOVE_CRYPTO:
	case LDM_REMOVE_MEM:
	case LDM_UNSET_VAR:
	case LDM_REMOVE_VCPU:
	case LDM_REMOVE_VDISK:
	case LDM_REMOVE_VNET:
	case LDM_REMOVE_VDPCC:
	case LDM_REMOVE_POLICY:
		return (xml_v3_content_cmd(data_node, ret_cmd, client, cmd_num,
		    XML_REMOVE_RES));
	}

	return (xml_handle_resources(data_node, ret_cmd, client,
	    XML_REMOVE_RES, cmd_num, NULL));
}

int
xml_v3_reboot_primary(xmlNodePtr data_node, xmlNodePtr ret_cmd,
    cli_conn_t *client,	int cmd_num)
{
	xmlDocPtr	reboot_doc = data_node->doc;
	char		*filename;

	filename = zsprintf("%s/%s", zeus.admdir, REBOOT_FILE);
	if (xmlSaveFormatFile(filename, reboot_doc, true) == -1) {
		xml_add_status(ret_cmd, XML_FAILURE,
		    MSG_ERR_REBOOT_FILE_CREATION);
		Xfree(filename);
		return (-1);
	}
	if (reboot(RB_AUTOBOOT, NULL) < 0) {
		xml_add_status(ret_cmd, XML_FAILURE, MSG_ERR_REBOOT_SYSTEM,
		    strerror(errno));
		Xfree(filename);
		return (-1);
	}
	return (0);
}
