/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <libxml/tree.h>
#include <libxml/parser.h>

#include "basics.h"
#include "zprintf.h"
#include "support.h"
#include "database.h"
#include "allocate.h"
#include "tod.h"
#include "ldom.h"
#include "zpri.h"
#include "response.h"
#include "xml_parse.h"
#include "xml_common.h"
#include "xml_output.h"
#include "xml_cntrl_common.h"
#include "database_internal.h"
#include "mac_addrs.h"
#include "variables.h"
#include "policy.h"
#include "xml_v3_defs.h"
#include "vcpu.h"
#include "vmau.h"
#include "memory.h"
#include "vio.h"
#include "core_md.h"
#include "vsw.h"

#define	XML_VERSION	((xmlChar *)"1.0") /* version of XML used */
#define	MBYTE		(1024 * 1024)

static void
constraints_data(ldom_t *ldp, xmlNodePtr envelope)
{
	xmlNodePtr	content, info_node;
	xmlChar		mac[MAC_STR_LEN] = { 0 };
	xmlChar		hostid[UINT64_STR_LEN] = { 0 };
	xmlChar		*failure_policy_str = NULL;
	xmlChar		*masters_str = NULL;
	uint64_t	failure_policy;

	content = xmlNewChild(envelope, NULL, CONTENT_NODE, NULL);
	xmlNewProp(content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
	xmlNewProp(content, ID_ATTR, (xmlChar *)ldp->namep);
	info_node = create_v3_section(content, RESOURCE_SECTION_TYPE,
	    XML_LDM_INFO);

	/*
	 * Add the domain MAC address.
	 */
	if ((!is_mac_allocated(ldp->props.mac_address)) ||
	    (ldp->flag.migration_source == true)) {
		mactoa(ldp->props.mac_address, (char *)mac);
		xmlNewChild(info_node, NULL, ADDRESS_NODE_OUT, mac);
	} else {
		xmlNewChild(info_node, NULL, ADDRESS_NODE_OUT,
		    (xmlChar *)XML_AUTO_ALLOCATED_MACADDR);
	}

	/*
	 * Add the domain hostid.
	 */
	if (!IS_DERIVED_HOSTID_ASSIGNED(ldp->props.mac_address,
	    ldp->props.hostid)) {
		sprintf((char *)hostid, "0x%"PRIx64, ldp->props.hostid);
		create_v3_property(info_node, XML_HOSTID, hostid);
	}

	/*
	 * No need to output failure policy and domain dependencies
	 * unless the firmware level supports it.
	 */
	if (LDOM_DEPENDENCIES_SUPPORTED()) {
		/* Add failure policy */
		failure_policy = ldp->props.failure_policy;
		failure_policy_str =
		    (xmlChar *)get_failure_policy_string(failure_policy);
		if (failure_policy_str != NULL) {
			create_v3_property(info_node, XML_FAILURE_POLICY,
			    failure_policy_str);
			xmlFree(failure_policy_str);
		}

		/* Add domain dependencies */
		masters_str = (xmlChar *)get_masters_string(ldp);
		if (masters_str != NULL) {
			create_v3_property(info_node, XML_MASTER, masters_str);
			xmlFree(masters_str);
		}
	}

	add_v3_stats_ldom_info(info_node, ldp, XML_CONSTRAINTS_OUTPUT);

	add_v3_cpu_output(content, ldp, XML_CONSTRAINTS_OUTPUT);
	add_v3_mau_output(content, ldp, XML_CONSTRAINTS_OUTPUT);
	add_v3_mem_output(content, ldp, XML_CONSTRAINTS_OUTPUT);
	add_v3_physio_output(content, ldp, XML_CONSTRAINTS_OUTPUT);
	add_v3_io_output(content, ldp->namep, ldp, XML_CONSTRAINTS_OUTPUT);
	add_v3_var_output(content, ldp, XML_CONSTRAINTS_OUTPUT, NULL);
	add_v3_policy_output(content, ldp, XML_CONSTRAINTS_OUTPUT);
}

void
xml_output_constraints(int argc, char **argv, response_t *resp)
{
	xmlDocPtr	xml_output;
	xmlNodePtr	root, data, envelope;
	char		*xml_buf;
	int		xml_buf_len, i;
	ldom_t		*ldp;

	xml_output = xmlNewDoc(XML_VERSION);
	root =  xmlNewDocNode(xml_output, NULL, XML_LDM_INTERFACE, NULL);
	xmlDocSetRootElement(xml_output, root);
	xmlNewProp(root, VERSION_ATTR, LDOM_INTERFACE_VERSION);

	xml_add_ldm_interface_properties(root);

	data = xmlNewChild(root, NULL, XML_DATA, NULL);
	xmlNewProp(data, VERSION_ATTR, LDOM_V3_DATA_VERSION);
	envelope = xmlNewChild(data, NULL, ENVELOPE_NODE, NULL);
	xmlNewChild(envelope, NULL, REFERENCES_NODE, NULL);

	if (argc == 0) {
		CHAIN_ITER(zeus.ldoms, ldp, link) {
			constraints_data(ldp, envelope);
		}
	} else {
		/* walk the list of LDoms */
		for (i = 0; i < argc; i++) {
			ldp = find_ldom_by_name((char *)argv[i]);
			if (ldp == NULL) {
				/*
				 * If we only have 1 domain requested and it
				 * doesn't exist we do not want to send back
				 * an empty XML Document. If there are more
				 * then 1 domain the non-existant domain
				 * will not show up in the document but will
				 * NOT report as not being there.
				 */
				if (argc == 1) {
					xmlFreeDoc(xml_output);
					rprintf(resp, "LDom \"%s\" was not "
					    "found\n", argv[i]);
					return;
				}
				continue;
			}
			constraints_data(ldp, envelope);
		}
	}

	(void) xmlDocDumpFormatMemory(xml_output, (xmlChar **)&xml_buf,
	    &xml_buf_len, 1);
	rprintf(resp, "%s", xml_buf);
	xmlFree(xml_buf);
	xmlFreeDoc(xml_output);
}

int
add_v3_var_output(xmlNodePtr content, ldom_t *ldp, int type, char *var_name)
{
	variables_t		*vars;
	variable_t		*var;
	int			i;
	xmlNodePtr		new_var;
	char			*s;
	bool_t			var_found = false;

	if (ldp->variables.len == 0) {
		if (type == XML_CONSTRAINTS_OUTPUT)
			return (0);
		else
			return (-1);
	}

	vars = &ldp->variables;

	for (i = 0; i < vars->len; i++) {
		var = &vars->pairs[i];

		if ((var->name != NULL) &&
		    (strcmp(var->name, "security-password") == 0)) {
			var_found = true;
			continue;
		}

		if (type == XML_CONSTRAINTS_OUTPUT) {
			s = var->value ? var->value : "";
			/* Resource==VAR parent==LDOM */
			new_var = create_v3_section(content,
			    VIRTUAL_HARDWARE_TYPE, VAR_NODE);

			/* property==NAME parent==VAR */
			create_v3_property(new_var, NAME_NODE,
			    (xmlChar *)var->name);

			/* property==VALUE parent==VAR */
			create_v3_property(new_var, XML_VALUE, (xmlChar *)s);

			var_found = true;
		} else {
			if ((var->name != NULL) &&
			    (strcmp(var->name, var_name) == 0)) {
				s = var->value ? var->value : "";
				/* Resource==VAR parent==LDOM */
				new_var = create_v3_section(content,
				    VIRTUAL_HARDWARE_TYPE, VAR_NODE);

				/* property==NAME parent==VAR */
				create_v3_property(new_var, NAME_NODE,
				    (xmlChar *)var->name);

				/* property==VALUE parent==VAR */
				create_v3_property(new_var, XML_VALUE,
				    (xmlChar *)s);
				var_found = true;
			}
		}
	}

	if (var_found == true)
		return (0);
	else
		return (-1);
}

void
add_v3_io_output(xmlNodePtr data, char *ldm_name, ldom_t *ldp, int type)
{
	add_v3_vsw_output(data, ldp, type);
	add_v3_vcc_output(data, ldp, type);
	add_v3_vds_output(data, ldp, type);
	add_v3_vdsdev_output(data, ldp, type);
	add_v3_network_output(data, ldp, type);
	add_v3_vdpcs_output(data, ldp, type);
	add_v3_disk_output(data, ldp, type);
	add_v3_vdpcc_output(data, ldp, type);
	add_v3_vcons_output(data, ldp, type);
}

static void
add_v3_vcc_binding(xmlNodePtr vcc_node, vcc_t *vcc)
{
	xmlNodePtr binding, new_cons;
	xmlChar port[INT_STR_LEN];
	vcc_console_t *vcons;

	if (vcc->binding == NULL)
		return;

	binding = xmlNewChild(vcc_node, NULL, BINDING_NODE_OUT, NULL);

	CHAIN_ITER(vcc->binding->consoles, vcons, link) {
		new_cons = xmlNewChild(binding, NULL, ITEM_NODE, NULL);
		xmlNewChild(new_cons, NULL, OTHER_RESOURCE_NODE_OUT,
		    CONSOLE_NODE);

		/* property==LDM_NAME parent==CONSOLE */
		create_v3_property(new_cons, XML_LDM_NAME,
		    (xmlChar *)vcons->console->ldp->namep);

		/* property==PORT parent==CONSOLE */
		snprintf((char *)port, sizeof (port), "%d",
		    (int)vcons->group->tcp_port->id);
		create_v3_property(new_cons, XML_PORT, port);
	}
}

void
add_v3_vcc_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	xmlNodePtr new_vcc;
	xmlChar port[INT_STR_LEN];
	vcc_t *vcc;

	CHAIN_ITER(ldp->vccs, vcc, ldom) {
		/* Resource==VCC parent==LDOM */
		new_vcc = create_v3_section(content, VIRTUAL_HARDWARE_TYPE,
		    VCC_NODE);

		/* property==SERVICE_NAME parent==VCC */
		create_v3_property(new_vcc, SERVICE_NAME_NODE,
		    (xmlChar *)vcc->service);

		/* property==MIN_PORT parent==VCC */
		snprintf((char *)port, sizeof (port), "%d", vcc->min_tcp_port);
		create_v3_property(new_vcc, MIN_PORT_NODE, port);

		/* property==MAX_PORT parent==VCC */
		snprintf((char *)port, sizeof (port), "%d", vcc->max_tcp_port);
		create_v3_property(new_vcc, MAX_PORT_NODE, port);

		if (type == XML_BINDING_OUTPUT) {
			add_v3_vcc_binding(new_vcc, vcc);
		}
	}
}

static void
add_v3_vdsdev_binding(xmlNodePtr vdsdev, vds_port_t *vds_port)
{
	xmlNodePtr binding = NULL;
	xmlNodePtr new_disk;
	vdisk_t *vdisk;

	binding = xmlNewChild(vdsdev, NULL, BINDING_NODE_OUT, NULL);

	if (vds_port == NULL)
		return;

	vdisk = vds_port->vdisk_port->vdisk;

	new_disk = xmlNewChild(binding, NULL, ITEM_NODE, NULL);
	xmlNewChild(new_disk, NULL, OTHER_RESOURCE_NODE_OUT, DISK_NODE);

	/* property==VDISK_NAME parent==DISK */
	create_v3_property(new_disk, VDISK_NAME_NODE, (xmlChar *)vdisk->name);

	/* property==LDM_NAME parent==DISK */
	create_v3_property(new_disk, XML_LDM_NAME,
	    (xmlChar *)vdisk->ldp->namep);
}

void
add_v3_vdsdev_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	xmlNodePtr new_vdsdev;
	vds_t *vds;
	vds_volume_t *vol;

	CHAIN_ITER(ldp->vdss, vds, ldom) {
		CHAIN_ITER(vds->volumes, vol, link) {
			/* Resource==VDSDEV parent==LDOM */
			new_vdsdev = create_v3_section(content,
			    VIRTUAL_HARDWARE_TYPE, VDS_VOLUME_NODE);

			/* property==VOLUME_NAME parent==VDSDEV */
			create_v3_property(new_vdsdev, VOLUME_NAME_NODE,
			    (xmlChar *)vol->volname);

			if (vol->blkdev != NULL) {
				/* property==BLOCKDEV parent==VDSDEV */
				create_v3_property(new_vdsdev, BLOCKDEV_NODE,
				    (xmlChar *)vol->blkdev);
			}

			/* property==SERVICE_NAME parent==VDSDEV */
			create_v3_property(new_vdsdev, SERVICE_NAME_NODE,
			    (xmlChar *)vds->service);

			/* property==VOLUME_OPTS parent==VDSDEV */
			if (vol->options != NULL) {
				create_v3_property(new_vdsdev, VOLUME_OPTS_NODE,
				    (xmlChar *)vol->options);
			}

			if (vol->mpgroup != NULL) {
				create_v3_property(new_vdsdev,
				    VOLUME_MPGROUP_NODE,
				    (xmlChar *)vol->mpgroup->name);
			}

			if (type == XML_BINDING_OUTPUT) {
				add_v3_vdsdev_binding(new_vdsdev,
				    vol->vds_port);
			}
		}
	}
}

static void
add_v3_vds_binding(xmlNodePtr vds_node, vds_t *vds)
{
	xmlNodePtr binding, new_vdsdev;
	vds_volume_t *vol;

	binding = xmlNewChild(vds_node, NULL, BINDING_NODE_OUT, NULL);

	CHAIN_ITER(vds->volumes, vol, link) {
		new_vdsdev = xmlNewChild(binding, NULL, ITEM_NODE, NULL);
		xmlNewChild(new_vdsdev, NULL, OTHER_RESOURCE_NODE_OUT,
		    VDS_VOLUME_NODE);

		/* property==VOLUME_NAME parent==VDS_VOLUME */
		create_v3_property(new_vdsdev, VOLUME_NAME_NODE,
		    (xmlChar *)vol->volname);

		if (vol->blkdev != NULL) {
			/* property==BLOCKDEV parent==VDS_VOLUME */
			create_v3_property(new_vdsdev, BLOCKDEV_NODE,
			    (xmlChar *)vol->blkdev);
		}
	}
}

void
add_v3_vds_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	xmlNodePtr new_vds;
	vds_t *vds;

	CHAIN_ITER(ldp->vdss, vds, ldom) {
		/* Resource==VDS parent==LDOM */
		new_vds = create_v3_section(content, VIRTUAL_HARDWARE_TYPE,
		    VDS_NODE);

		/* property==SERVICE_NAME parent==VDS */
		create_v3_property(new_vds, SERVICE_NAME_NODE,
		    (xmlChar *)vds->service);

		if ((type == XML_BINDING_OUTPUT) ||
		    (type == XML_SERVICES_OUTPUT)) {
			add_v3_vds_binding(new_vds, vds);
		}
	}
}

static void
add_v3_vdpcs_binding(xmlNodePtr vdpcs_node, vdpcs_t *vdpcs)
{
	xmlNodePtr binding, new_vdpcc;
	vdpcs_port_t *vdpcs_port;

	binding = xmlNewChild(vdpcs_node, NULL, BINDING_NODE_OUT, NULL);

	if (vdpcs->binding == NULL)
		return;

	CHAIN_ITER(vdpcs->binding->vports, vdpcs_port, link) {
		new_vdpcc = xmlNewChild(binding, NULL, ITEM_NODE, NULL);
		xmlNewChild(new_vdpcc, NULL, OTHER_RESOURCE_NODE_OUT,
		    VDPCC_NODE);

		/* property==VDPCC_NAME parent==VDPCC */
		create_v3_property(new_vdpcc, VDPCC_NAME_NODE,
		    (xmlChar *)vdpcs_port->vdpcc->name);

		/* property==LDM_NAME parent==VDPCC */
		create_v3_property(new_vdpcc, XML_LDM_NAME,
		    (xmlChar *)vdpcs_port->vdpcc->ldp->namep);
	}
}

void
add_v3_vdpcs_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	xmlNodePtr new_vdpcs;
	vdpcs_t *vdpcs;

	CHAIN_ITER(ldp->vdpcss, vdpcs, ldom) {
		/* Resource==VDPCS parent==LDOM */
		new_vdpcs = create_v3_section(content, VIRTUAL_HARDWARE_TYPE,
		    VDPCS_NODE);

		/* property==SERVICE_NAME parent==VDPCS */
		create_v3_property(new_vdpcs, SERVICE_NAME_NODE,
		    (xmlChar *)vdpcs->service);

		if ((type == XML_BINDING_OUTPUT) ||
		    (type == XML_SERVICES_OUTPUT)) {
			add_v3_vdpcs_binding(new_vdpcs, vdpcs);
		}
	}
}

void
add_v3_vdpcc_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	xmlNodePtr new_vdpcc;
	vdpcc_t *vdpcc;

	CHAIN_ITER(ldp->vdpccs, vdpcc, ldom) {
		/* Resource==VDPCC parent==LDOM */
		new_vdpcc = create_v3_section(content, VIRTUAL_HARDWARE_TYPE,
		    VDPCC_NODE);

		/* property==VDPCC_NAME parent==VDPCC */
		create_v3_property(new_vdpcc, VDPCC_NAME_NODE,
		    (xmlChar *)vdpcc->name);

		/* property==SERVICE_NAME parent==VDPCC */
		create_v3_property(new_vdpcc, SERVICE_NAME_NODE,
		    (xmlChar *)vdpcc->service);
	}
}

static void
add_v3_vsw_binding(xmlNodePtr vsw_node, vsw_t *vsw)
{
	vsw_port_t *vsw_port;
	xmlNodePtr binding, new_net;
	xmlChar mac[MAC_STR_LEN] = { 0 };
	char *vlan_str = NULL;
	char int_string[INT_STR_LEN];

	if ((vsw->binding == NULL) || (CHAIN_LENGTH(vsw->binding->vports) == 0))
		return;

	binding = xmlNewChild(vsw_node, NULL, BINDING_NODE_OUT, NULL);

	CHAIN_ITER(vsw->binding->vports, vsw_port, link) {
		vnet_t *vnet = vsw_port->vnet;
		int vsw_mtu = (vsw_port->vsw->mtu != VIO_NONEMTU) ?
		    vsw_port->vsw->mtu : VIO_ETHERMTU;
		int mtu = (vnet->mtu != VIO_NONEMTU) ? vnet->mtu : vsw_mtu;
		new_net = xmlNewChild(binding, NULL, ITEM_NODE, NULL);
		xmlNewChild(new_net, NULL, OTHER_RESOURCE_NODE_OUT, VNET_NODE);

		/* property==MAC_ADDRESS parent==VNET */
		mactoa(vnet->mac, (char *)mac);
		xmlNewChild(new_net, NULL, ADDRESS_NODE_OUT, mac);

		/* property==VNET_NAME parent==VNET */
		create_v3_property(new_net, VNET_NAME_NODE,
		    (xmlChar *)vnet->name);

		/* property==LDM_NAMES parent==VNET */
		create_v3_property(new_net, XML_LDM_NAME,
		    (xmlChar *)vnet->ldp->namep);

		snprintf(int_string, sizeof (int_string), "%d", vnet->pvid);
		/* property==PVID parent==VNET */
		create_v3_property(new_net, VNET_PORT_VLAN_NODE,
		    (xmlChar *)int_string);

		if (vnet->num_vlans > 0) {
			vlan_str = comma_sepa_str(vnet->num_vlans,
			    vnet->vlans);
		}
		if (vlan_str != NULL) {
			/* property==VID parent==VNET */
			create_v3_property(new_net, VNET_VLANS_NODE,
			    (xmlChar *)vlan_str);
			Xfree(vlan_str);
			vlan_str = NULL;
		}

		snprintf(int_string, sizeof (int_string), "%d", mtu);
		/* property==MTU parent==VNET */
		create_v3_property(new_net, MTU_NODE,
		    (xmlChar *)int_string);

	}
}

void
add_v3_vsw_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	xmlNodePtr new_vsw;
	xmlChar mac[MAC_STR_LEN] = { 0 };
	vsw_t *vsw;
	char *vlan_str = NULL;
	char int_string[INT_STR_LEN];
	char *devinst;
	char *id_str = NULL;

	CHAIN_ITER(ldp->vsws, vsw, ldom) {
		/* Resource==VSW parent==LDOM */
		new_vsw = create_v3_section(content, VIRTUAL_HARDWARE_TYPE,
		    VSW_NODE);

		if (type == XML_CONSTRAINTS_OUTPUT) {
			if ((!is_mac_allocated(vsw->mac)) ||
			    (ldp->flag.migration_source == true)) {
				mactoa(vsw->mac, (char *)mac);
				/* property==rasd:ADDRESS parent==VSW */
				xmlNewChild(new_vsw, NULL, ADDRESS_NODE_OUT,
				    mac);
			} else {
				xmlNewChild(new_vsw, NULL, ADDRESS_NODE_OUT,
				    (xmlChar *)XML_AUTO_ALLOCATED_MACADDR);
			}
		} else if ((type == XML_BINDING_OUTPUT) ||
		    (type == XML_SERVICES_OUTPUT)) {
			mactoa(vsw->mac, (char *)mac);
			/* property==rasd:ADDRESS parent==VSW */
			xmlNewChild(new_vsw, NULL, ADDRESS_NODE_OUT, mac);
		}

		/* property==SERVICE_NAME parent==VSW */
		create_v3_property(new_vsw, SERVICE_NAME_NODE,
		    (xmlChar *)vsw->service);

		if (vsw->mode != NULL) {
			/* property==MODE  parent==VSW */
			create_v3_property(new_vsw, VSW_MODE_NODE,
			    (xmlChar *)vsw->mode);
		}

		if (vsw->phys_dev != NULL) {
			/* property==DEV_PATH parent==VSW */
			create_v3_property(new_vsw, DEV_PATH_NODE,
			    (xmlChar *)vsw->phys_dev);
		}

		/* property==LINKPROP parent==VSW */
		if (vsw->phys_state == 0x1) {
			create_v3_property(new_vsw, LINKPROP_NODE,
			    (xmlChar *)SET_PHYS_STATE);
		}

		snprintf(int_string, sizeof (int_string), "%d",
		    vsw->default_vid);
		create_v3_property(new_vsw, VSW_DEF_VLAN_NODE,
		    (xmlChar *)int_string);

		snprintf(int_string, sizeof (int_string), "%d", vsw->pvid);
		/* property==PVID parent==VSW */
		create_v3_property(new_vsw, VNET_PORT_VLAN_NODE,
		    (xmlChar *)int_string);

		if (vsw->num_vlans > 0) {
			vlan_str = comma_sepa_str(vsw->num_vlans, vsw->vlans);
		}
		if (vlan_str != NULL) {
			/* property==VID parent==VSW */
			create_v3_property(new_vsw, VSW_VLANS_NODE,
			    (xmlChar *)vlan_str);
			Xfree(vlan_str);
			vlan_str = NULL;
		}

		if (vsw->mtu != VIO_NONEMTU) {
			snprintf(int_string, sizeof (int_string), "%d",
			    vsw->mtu);
			/* property==MTU parent==VSW */
			create_v3_property(new_vsw, MTU_NODE,
			    (xmlChar *)int_string);
		}

		if (vsw->binding != NULL) {
			devinst = zsprintf("switch@%"PRIu64, vsw_devid(vsw));
			/* property==device parent==VSW */
			create_v3_property(new_vsw, XML_DEVICE,
			    (xmlChar *)devinst);
			Xfree(devinst);
		}

		if (vsw->switchid != NO_NEW_ID) {
			id_str = zsprintf("%d", vsw->switchid);
			/* property==ID  parent==VSW */
			create_v3_property(new_vsw, ID_NODE,
			    (xmlChar *)id_str);
			Xfree(id_str);
		}

		if (type == XML_BINDING_OUTPUT) {
			add_v3_vsw_binding(new_vsw, vsw);
		}
	}
}

static void
add_v3_disk_binding(xmlNodePtr disk_node, vdisk_t *vdisk)
{

	vdisk_port_t *vport;
	xmlNodePtr binding, new_disk;

	if (vdisk->binding == NULL)
		return;

	binding = xmlNewChild(disk_node, NULL, BINDING_NODE_OUT, NULL);
	CHAIN_ITER(vdisk->binding->vdisk_ports, vport, link) {
		vds_port_t *vds_port = vport->vds_port;
		new_disk = xmlNewChild(binding, NULL, ITEM_NODE, NULL);
		xmlNewChild(new_disk, NULL, OTHER_RESOURCE_NODE_OUT,
		    DISK_NODE);

		/* property==VOLUME_NAME parent==VDISK */
		create_v3_property(new_disk, VOLUME_NAME_NODE,
		    (xmlChar *)vds_port->vol->volname);

		/* property==SERVICE_NAME parent==DISK */
		create_v3_property(new_disk, SERVICE_NAME_NODE,
		    (xmlChar *)vds_port->vds->service);

		/* property==LDM_NAMES parent==DISK */
		create_v3_property(new_disk, XML_LDM_NAME,
		    (xmlChar *)vds_port->vds->ldp->namep);
	}
}

void
add_v3_disk_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	xmlNodePtr new_disk;
	vdisk_t *vdisk;
	char *devinst;
	char *id_str = NULL;

	CHAIN_ITER(ldp->vdisks, vdisk, ldom) {
		/* Resource==DISK parent==LDOM */
		new_disk = create_v3_section(content, VIRTUAL_HARDWARE_TYPE,
		    DISK_NODE);

		/* property==VDISK_NAME parent==DISK */
		create_v3_property(new_disk, VDISK_NAME_NODE,
		    (xmlChar *)vdisk->name);

		/* property==SERVICE_NAME parent==DISK */
		create_v3_property(new_disk, SERVICE_NAME_NODE,
		    (xmlChar *)vdisk->service);

		/* property==VOLUME_NAME parent==DISK */
		create_v3_property(new_disk, VOLUME_NAME_NODE,
		    (xmlChar *)vdisk->volname);

		/* property=TIMEOUT_OUT  parent=DISK */
		if (vdisk->timeout > 0) {
			char *timeout_str = zsprintf("%d", vdisk->timeout);
			create_v3_property(new_disk, TIMEOUT_NODE,
			    (xmlChar *)timeout_str);
			free(timeout_str);
		}

		if (vdisk->binding != NULL) {
			devinst = zsprintf("disk@%"PRIu64, vdisk_devid(vdisk));
			create_v3_property(new_disk, XML_DEVICE,
			    (xmlChar *)devinst);
			Xfree(devinst);
		}

		/* property==ID  parent==DISK */
		if (vdisk->diskid != NO_NEW_ID) {
			id_str = zsprintf("%d", vdisk->diskid);
			create_v3_property(new_disk, ID_NODE,
			    (xmlChar *)id_str);
			Xfree(id_str);
		}

		if (type == XML_BINDING_OUTPUT)
			add_v3_disk_binding(new_disk, vdisk);
	}
}

static void
add_v3_network_binding(xmlNodePtr vnet_node, vnet_t *vnet)
{
	xmlNodePtr binding, new_net, new_vsw;
	xmlChar mac[MAC_STR_LEN] = { 0 };
	vnet_peer_t *local_peer;
	char *vlan_str = NULL;
	char int_string[INT_STR_LEN];
	vsw_t *vsw;
	int mtu;
	int vsw_mtu;

	if (vnet->binding == NULL)
		return;

	binding = xmlNewChild(vnet_node, NULL, BINDING_NODE_OUT, NULL);

	/* Display the switch binding */
	/* section==VSW_NODE */
	new_vsw = create_v3_section(binding, VIRTUAL_HARDWARE_TYPE, VSW_NODE);

	vsw = vnet->binding->vsw_port->vsw;
	vsw_mtu = (vsw->mtu != VIO_NONEMTU) ? vsw->mtu : VIO_ETHERMTU;

	/* section==VSW_NODE property==MAC address */
	mactoa(vsw->mac, (char *)mac);
	xmlNewChild(new_vsw, NULL, ADDRESS_NODE_OUT, mac);

	/* section==VSW_NODE property==SERVICE_NAME */
	create_v3_property(new_vsw, SERVICE_NAME_NODE,
	    (xmlChar *)vnet->service);

	/* section==VSW_NODE property==LDM_NAME */
	create_v3_property(new_vsw, XML_LDM_NAME,
	    (xmlChar *)vsw->ldp->namep);

	/* section==VSW_NODE property==MODE */
	if (vsw->mode != NULL) {
		create_v3_property(new_vsw, VSW_MODE_NODE,
		    (xmlChar *)vsw->mode);
	}

	/* section==VSW_NODE property==PVID */
	snprintf(int_string, sizeof (int_string), "%d", vsw->pvid);
	create_v3_property(new_vsw, VNET_PORT_VLAN_NODE,
	    (xmlChar *)int_string);

	/* section==VSW_NODE property==VID */
	if (vsw->num_vlans > 0) {
		vlan_str = comma_sepa_str(vsw->num_vlans, vsw->vlans);
	}
	if (vlan_str != NULL) {
		create_v3_property(new_vsw, VSW_VLANS_NODE,
		    (xmlChar *)vlan_str);
		Xfree(vlan_str);
		vlan_str = NULL;
	}

	snprintf(int_string, sizeof (int_string), "%d", vsw_mtu);
	create_v3_property(new_vsw, MTU_NODE, (xmlChar *)int_string);

	CHAIN_ITER(vnet->binding->peers, local_peer, link) {
		vnet_peer_t *remote_peer = local_peer->peer;

		new_net = xmlNewChild(binding, NULL, ITEM_NODE, NULL);
		xmlNewChild(new_net, NULL, OTHER_RESOURCE_NODE_OUT, VNET_NODE);

		/* property==rasd:ADDRESS parent==VNET */
		mactoa(remote_peer->vnet->mac, (char *)mac);
		xmlNewChild(new_net, NULL, ADDRESS_NODE_OUT, mac);

		/* property==VNET_NAME parent==VNET */
		create_v3_property(new_net, VNET_NAME_NODE,
		    (xmlChar *)remote_peer->vnet->name);

		/* property==LDM_NAME parent==VNET */
		create_v3_property(new_net, XML_LDM_NAME,
		    (xmlChar *)remote_peer->vnet->ldp->namep);

		snprintf(int_string, sizeof (int_string), "%d",
		    remote_peer->vnet->pvid);
		/* property==PVID parent==VNET */
		create_v3_property(new_net, VNET_PORT_VLAN_NODE,
		    (xmlChar *)int_string);

		if (remote_peer->vnet->num_vlans > 0) {
			vlan_str = comma_sepa_str(remote_peer->vnet->num_vlans,
			    remote_peer->vnet->vlans);
		}
		if (vlan_str != NULL) {
			/* property==VID parent==VNET */
			create_v3_property(new_net, VNET_VLANS_NODE,
			    (xmlChar *)vlan_str);
			Xfree(vlan_str);
			vlan_str = NULL;
		}

		mtu = remote_peer->vnet->mtu;
		if (mtu == VIO_NONEMTU)
			mtu = vsw_mtu;
		snprintf(int_string, sizeof (int_string), "%d", mtu);
		create_v3_property(new_net, MTU_NODE, (xmlChar *)int_string);
	}
}

void
add_v3_network_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	xmlNodePtr new_net;
	xmlChar mac[MAC_STR_LEN] = { 0 };
	vnet_t *vnet;
	char *vlan_str = NULL;
	char *devinst;
	char int_string[INT_STR_LEN];
	char *id_str = NULL;

	CHAIN_ITER(ldp->vnets, vnet, ldom) {
		int mtu = vnet->mtu;
		uint64_t phys_state = vnet->phys_state;

		/* Resource==VNET parent==LDOM */
		new_net = create_v3_section(content, VIRTUAL_HARDWARE_TYPE,
		    VNET_NODE);

		if (type == XML_CONSTRAINTS_OUTPUT) {
			/*
			 * Check to see if the MAC address for the VNET is an
			 * allocated one. If not add the <mac_address> node and
			 * address as is. If so then don't add the <mode> tag
			 * and if the VNET should have a public or private
			 * MAC address.
			 */
			if ((!is_mac_allocated(vnet->mac))||
			    (ldp->flag.migration_source == true)) {
				/* property==rasd:ADDRESS parent==VNET */
				mactoa(vnet->mac, (char *)mac);
				xmlNewChild(new_net, NULL, ADDRESS_NODE_OUT,
				    mac);
			} else {
				xmlNewChild(new_net, NULL, ADDRESS_NODE_OUT,
				    (xmlChar *)XML_AUTO_ALLOCATED_MACADDR);
			}
		} else if (type == XML_BINDING_OUTPUT) {
			/* property==rasd:ADDRESS parent==VNET */
			mactoa(vnet->mac, (char *)mac);
			xmlNewChild(new_net, NULL, ADDRESS_NODE_OUT, mac);
		}

		if (type != XML_CONSTRAINTS_OUTPUT && vnet->binding != NULL &&
		    mtu == VIO_NONEMTU) {
			mtu = vnet->binding->vsw_port->vsw->mtu;
			if (mtu == VIO_NONEMTU)
				mtu = VIO_ETHERMTU;
		}

		/* property==VNET_NAME parent==VNET */
		create_v3_property(new_net, VNET_NAME_NODE,
		    (xmlChar *)vnet->name);

		/* property==SERVICE_NAME parent==VNET */
		create_v3_property(new_net, SERVICE_NAME_NODE,
		    (xmlChar *)vnet->service);

		/* property==HYBRID_MODE parent==VNET */
		if (vnet->vnet_mode != NULL)
			create_v3_property(new_net, VNET_MODE_NODE,
			    (xmlChar *)vnet->vnet_mode);

		snprintf(int_string, sizeof (int_string), "%d", vnet->pvid);
		/* property==PVID parent==VNET */
		create_v3_property(new_net, VNET_PORT_VLAN_NODE,
		    (xmlChar *)int_string);

		if (vnet->num_vlans > 0) {
			vlan_str = comma_sepa_str(vnet->num_vlans,
			    vnet->vlans);
		}
		if (vlan_str != NULL) {
			/* property==VID parent==VNET */
			create_v3_property(new_net, VNET_VLANS_NODE,
			    (xmlChar *)vlan_str);
			Xfree(vlan_str);
			vlan_str = NULL;
		}

		/* property==MTU parent==VNET */
		if (mtu != VIO_NONEMTU) {
			snprintf(int_string, sizeof (int_string), "%d", mtu);
			create_v3_property(new_net, MTU_NODE,
			    (xmlChar *)int_string);
		}

		/* property==LINKPROP parent==VNET */
		if (phys_state == 0x1) {
			create_v3_property(new_net, LINKPROP_NODE,
			    (xmlChar *)SET_PHYS_STATE);
		}

		if ((type == XML_BINDING_OUTPUT) && (vnet->binding != NULL)) {
			vsw_t *vsw = vnet->binding->vsw_port->vsw;

			/* property==SERVICE_DOMAIN parent==VNET */
			create_v3_property(new_net, XML_SERVICE_DOMAIN,
			    (xmlChar *)vsw->ldp->namep);
		}

		if (vnet->binding != NULL) {
			devinst = zsprintf("network@%"PRIu64,
			    vnet_devid(vnet));
			create_v3_property(new_net, XML_DEVICE,
			    (xmlChar *)devinst);
			Xfree(devinst);
		}

		/* property==ID parent==VNET */
		if (vnet->networkid != NO_NEW_ID) {
			id_str = zsprintf("%d", vnet->networkid);
			create_v3_property(new_net, ID_NODE,
			    (xmlChar *)id_str);
			Xfree(id_str);
		}

		if (type == XML_BINDING_OUTPUT) {
			add_v3_network_binding(new_net, vnet);
		}
	}
}

static void
vcons_constraints_output(xmlNodePtr content, console_t *cons)
{
	xmlNodePtr	new_cons;
	xmlChar		port[INT_STR_LEN];

	/* Resource==CONSOLE parent==LDOM */
	new_cons = create_v3_section(content, VIRTUAL_HARDWARE_TYPE,
	    CONSOLE_NODE);

	/* property==GROUP_NAME parent==CONSOLE */
	if (cons->group_name != NULL) {
		create_v3_property(new_cons, GROUP_NODE,
		    (xmlChar *)cons->group_name);
	}

	/* property==SERVICE_NAME parent==CONSOLE */
	if (cons->service_name != NULL) {
		create_v3_property(new_cons, SERVICE_NAME_NODE,
		    (xmlChar *)cons->service_name);
	}

	/* property==PORT parent==CONSOLE */
	if (cons->tcp_port != AUTO_CONSOLE_TCP_PORT) {
		snprintf((char *)port, sizeof (port), "%d",
		    (int)cons->tcp_port);
		create_v3_property(new_cons, XML_PORT, port);
	}
}

void
add_v3_vcons_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	xmlNodePtr new_cons;
	xmlChar port[INT_STR_LEN];
	console_t *cons = &ldp->console;
	vcc_console_t *vcc_cons;

	if (type == XML_BINDING_OUTPUT) {
		if (cons->binding != NULL)  {
			/* Resource==CONSOLE parent==LDOM */
			new_cons = create_v3_section(content,
			    VIRTUAL_HARDWARE_TYPE, CONSOLE_NODE);

			if (cons->binding->sp_cons) {
				/* property==SERVICE_NAME parent==CONSOLE */
				create_v3_property(new_cons, SERVICE_NAME_NODE,
				    (xmlChar *)"SP");
			} else if (cons->binding->uartbase != 0) {
				/* property==SERVICE_NAME parent==CONSOLE */
				create_v3_property(new_cons, SERVICE_NAME_NODE,
				    (xmlChar *)"UART");
			} else {
				vcc_cons = cons->binding->vcc_cons;

				/* property==GROUP parent==CONSOLE */
				create_v3_property(new_cons, GROUP_NODE,
				    (xmlChar *)vcc_cons->group->group_name);

				/* property==SERVICE_NAME parent==CONSOLE */
				create_v3_property(new_cons, SERVICE_NAME_NODE,
				    (xmlChar *)vcc_cons->vcc->service);

				/* property==PORT parent==CONSOLE */
				snprintf((char *)port, sizeof (port), "%d",
				    (int)vcc_cons->group->tcp_port->id);
				create_v3_property(new_cons, XML_PORT, port);
			}
		} else if (is_console_constraints_set(ldp)) {
			vcons_constraints_output(content, cons);
		}
	} else if (type == XML_CONSTRAINTS_OUTPUT) {
		if (is_console_constraints_set(ldp)) {
			vcons_constraints_output(content, cons);
		}
	}
}

void
add_v3_free_cpu(xmlNodePtr envelope)
{
	pri_cpu_t	*pcpu_p;
	int		frac_free;
	xmlNodePtr	cpu, section;
	char		cpu_no[INT_STR_LEN];

	section = xmlNewChild(envelope, NULL, SECTION_NODE, NULL);
	xmlNewProp(section, TYPE_ATTR, VIRTUAL_HARDWARE_TYPE);

	CHAIN_ITER(pri.cpu_list, pcpu_p, link) {
		frac_free = pcpu_p->bound.fraction_free;
		if (frac_free > 0) {
			cpu = xmlNewChild(section, NULL, ITEM_NODE, NULL);
			xmlNewChild(cpu, NULL, OTHER_RESOURCE_NODE_OUT,
			    CPU_NODE);

			/* property==PID parent==CPU */
			snprintf(cpu_no, sizeof (cpu_no), "%d",
			    pcpu_p->phys_id);
			create_v3_property(cpu, XML_PID, (xmlChar *)cpu_no);

			/* property==STRAND_PERCENT parent==FREE */
			snprintf(cpu_no, sizeof (cpu_no), "%d",
			    frac_free * 100 / PRI_CPU_FRAC_DENOM);
			create_v3_property(cpu, XML_STRAND_PERCENT,
			    (xmlChar *)cpu_no);
		}
	}
}

static void
add_v3_cpu_binding(xmlNodePtr cpu, ldom_t *ldp)
{
	xmlNodePtr	binding, bound_cpu;
	vcpu_t		*vcpu;
	char		cpu_no[INT_STR_LEN];
	pm_status_t	pm_status;
	char		util[8];

	binding = xmlNewChild(cpu, NULL, BINDING_NODE_OUT, NULL);

	CHAIN_ITER(ldp->bound.vcpus, vcpu, ldom) {
		bound_cpu = xmlNewChild(binding, NULL, ITEM_NODE, NULL);
		xmlNewChild(bound_cpu, NULL, OTHER_RESOURCE_NODE_OUT, CPU_NODE);

		/* property==VID parent==BINDING */
		snprintf(cpu_no, sizeof (cpu_no), "%d", vcpu->vid);
		create_v3_property(bound_cpu, XML_VID, (xmlChar *)cpu_no);

		/* property==PID parent==BINDING */
		snprintf(cpu_no, sizeof (cpu_no), "%d", vcpu->cpu->phys_id);
		create_v3_property(bound_cpu, XML_PID, (xmlChar *)cpu_no);

		/* property==STRAND_PERCENT parent==BINDING */
		snprintf(cpu_no, sizeof (cpu_no), "%d",
		    vcpu->fraction*100 / PRI_CPU_FRAC_DENOM);
		create_v3_property(bound_cpu, XML_STRAND_PERCENT,
		    (xmlChar *)cpu_no);

		/* property==UTIL_PERCENT parent==BINDING */
		pm_status = get_core_md_cpu_status(vcpu->cpu->phys_id,
		    CORE_MD_PM_STATUS);

		if (pm_status != PM_MANAGED) {
			if (ldp->state > BOUND) {
				if (vcpu->util >= 10.0)
					snprintf(util, sizeof (util),
					    "%3.0f%%", vcpu->util);
				else
					snprintf(util, sizeof (util),
					    "%3.1f%%", vcpu->util);
			} else {
				strcpy(util, "0");
			}
			create_v3_property(bound_cpu, XML_UTIL_PERCENT,
			    (xmlChar *)util);
		}
	}
}

void
add_v3_cpu_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	char		cpu_no[INT_STR_LEN];
	xmlNodePtr	cpu;

	if (ldp->required.num_vcpus > 0) {
		cpu = create_v3_section(content, VIRTUAL_HARDWARE_TYPE,
		    CPU_NODE);

		/* property==NUMBER parent==CPU */
		snprintf(cpu_no, sizeof (cpu_no), "%d",
		    ldp->required.num_vcpus);
		xmlNewChild(cpu, NULL, ALLOCATION_NODE_OUT, (xmlChar *)cpu_no);

		if (type == XML_BINDING_OUTPUT) {
			add_v3_cpu_binding(cpu, ldp);
		}
	}
}

void
add_v3_free_mau(xmlNodePtr envelope)
{
	pri_cpu_t	*cpup;
	pri_mau_t 	*maup;
	xmlNodePtr	mau, section;
	char		cpu_no[INT_STR_LEN];

	section = xmlNewChild(envelope, NULL, SECTION_NODE, NULL);
	xmlNewProp(section, TYPE_ATTR, VIRTUAL_HARDWARE_TYPE);

	CHAIN_ITER(pri.mau_list, maup, link) {
		if (get_mau_ldom(maup) == NULL) {
			/* this mau is unbound */
			mau = xmlNewChild(section, NULL, ITEM_NODE, NULL);
			xmlNewChild(mau, NULL, OTHER_RESOURCE_NODE_OUT,
			    MAU_NODE);

			CHAIN_ITER(maup->cpus, cpup, mau_link) {
				/* property==NUMBER parent==CPU */
				snprintf(cpu_no, sizeof (cpu_no), "%d",
				    cpup->phys_id);
				create_v3_property(mau, XML_PID,
				    (xmlChar *)cpu_no);
			}
		}
	}
}

static void
add_v3_mau_binding(xmlNodePtr mau, ldom_t *ldp)
{
	xmlNodePtr	binding, cpuset;
	vmau_t		*vmau;
	pri_cpu_t	*pcpu;
	char		cpu_no[INT_STR_LEN];

	binding = xmlNewChild(mau, NULL, BINDING_NODE_OUT, NULL);

	CHAIN_ITER(ldp->bound.vmaus, vmau, ldom) {
		cpuset = create_v3_section(binding, VIRTUAL_HARDWARE_TYPE,
		    XML_CPUSET);

		CHAIN_ITER(vmau->mau->cpus, pcpu, mau_link) {
			/* property==NUMBER parent==CPU */
			snprintf(cpu_no, sizeof (cpu_no), "%d", pcpu->phys_id);
			create_v3_property(cpuset,  NUMBER_NODE,
			    (xmlChar *)cpu_no);
		}
	}
}

void
add_v3_mau_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	char		mau_no[INT_STR_LEN];
	xmlNodePtr	mau;

	mau = create_v3_section(content, VIRTUAL_HARDWARE_TYPE, MAU_NODE);

	snprintf(mau_no, sizeof (mau_no), "%d", ldp->required.num_vmaus);

	/* property==NUMBER parent==MAU */
	xmlNewChild(mau, NULL, ALLOCATION_NODE_OUT, (xmlChar *)mau_no);

	if ((ldom_num_bound_maus(ldp) > 0) && (type == XML_BINDING_OUTPUT)) {
		add_v3_mau_binding(mau, ldp);
	}
}

void
add_v3_free_mem(xmlNodePtr envelope)
{
	xmlNodePtr	mem, section;
	char		phys_mem[UINT64_STR_LEN];
	char		*mem_amt;
	mblock_t	*mblk;

	section = xmlNewChild(envelope, NULL, SECTION_NODE, NULL);
	xmlNewProp(section, TYPE_ATTR, VIRTUAL_HARDWARE_TYPE);

	CHAIN_ITER(free_memory.blocks, mblk, pblk) {
		mem = xmlNewChild(section, NULL, ITEM_NODE, NULL);
		xmlNewChild(mem, NULL, OTHER_RESOURCE_NODE_OUT, MEMORY_NODE);

		/* property==PHYS_ADDR parent==FREE */
		snprintf(phys_mem, sizeof (phys_mem), "0x%"PRIx64"",
		    mblk->paddr);
		create_v3_property(mem, XML_PHYS_ADDR, (xmlChar *)phys_mem);

		/* property==SIZE parent==FREE */
		mem_amt = conv_size(mblk->size);
		create_v3_property(mem, SIZE_NODE, (xmlChar *)mem_amt);
		Xfree(mem_amt);
	}
}

static void
add_v3_mem_binding(xmlNodePtr mem, ldom_t *ldp)
{
	char		*mem_amt;
	char		phys_mem[UINT64_STR_LEN];
	xmlNodePtr	binding;
	mblock_t	*mblk;

	binding = xmlNewChild(mem, NULL, BINDING_NODE_OUT, NULL);

	CHAIN_ITER(ldp->bound.memory, mblk, ldom) {
		/* property==REAL_ADDR parent==BINDING */
		snprintf(phys_mem, sizeof (phys_mem), "0x%"PRIx64"",
		    mblk->raddr);
		create_v3_property(binding, XML_REAL_ADDR, (xmlChar *)phys_mem);

		/* property==PHYS_ADDR parent==BINDING */
		snprintf(phys_mem, sizeof (phys_mem), "0x%"PRIx64"",
		    mblk->paddr);
		create_v3_property(binding, XML_PHYS_ADDR, (xmlChar *)phys_mem);

		/* property==SIZE parent==BINDING */
		mem_amt = conv_size(mblk->size);
		create_v3_property(binding, SIZE_NODE, (xmlChar *)mem_amt);
		Xfree(mem_amt);
	}
}

void
add_v3_mem_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	char		*mem_amt;
	xmlNodePtr	mem;

	if (ldp->required.size_mem > 0) {
		mem = create_v3_section(content, VIRTUAL_HARDWARE_TYPE,
		    MEMORY_NODE);

		/* property==SIZE parent==MEM */
		mem_amt = conv_size(ldp->required.size_mem);
		xmlNewChild(mem, NULL, ALLOCATION_NODE_OUT, (xmlChar *)mem_amt);
		Xfree(mem_amt);

		if (type == XML_BINDING_OUTPUT) {
			add_v3_mem_binding(mem, ldp);
		}
	}
}

void
add_v3_free_io(xmlNodePtr content)
{
	pri_iodevice_t	*pio_p;
	xmlNodePtr	io, section;

	section = xmlNewChild(content, NULL, SECTION_NODE, NULL);
	xmlNewProp(section, TYPE_ATTR, VIRTUAL_HARDWARE_TYPE);

	CHAIN_ITER(pri.iodevice_list, pio_p, link) {
		if (pio_p->free) {
			io = xmlNewChild(section, NULL, ITEM_NODE, NULL);
			xmlNewChild(io, NULL, OTHER_RESOURCE_NODE_OUT,
			    PHYSIO_NODE);

			/* property==DEV_TYPE parent==IO */
			create_v3_property(io, XML_DEV_TYPE,
			    (xmlChar *)pio_p->device_type);

			/* property==NAME parent==IO */
			create_v3_property(io, NAME_NODE,
			    (xmlChar *)pio_p->name);
		}
	}
}

void
add_v3_physio_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	req_iodevice_t		*iodevs_p;
	req_pcie_dev_t		*pciedev_p;
	bound_iodevice_t	*biodevs_p;
	xmlNodePtr		io_dev;

	if (type == XML_CONSTRAINTS_OUTPUT) {
		CHAIN_ITER(ldp->required.io_devices, iodevs_p, link) {

			io_dev = create_v3_section(content,
			    VIRTUAL_HARDWARE_TYPE, XML_PHYSIO_DEVICE);

			/* property==NAME parent==PHYSIO_DEVICE */
			create_v3_property(io_dev, NAME_NODE,
			    (xmlChar *)iodevs_p->name);

		}

		CHAIN_ITER(ldp->required.io_pcie_devs, pciedev_p, link) {

			io_dev = create_v3_section(content,
			    VIRTUAL_HARDWARE_TYPE, XML_PHYSIO_DEVICE);

			/* property==NAME parent==PHYSIO_DEVICE */
			create_v3_property(io_dev, NAME_NODE,
			    (xmlChar *)pciedev_p->cli_name);
		}
	}

	if (type == XML_BINDING_OUTPUT) {
		CHAIN_ITER(ldp->bound.iodevices, biodevs_p, ldom) {
			io_dev = create_v3_section(content,
			    VIRTUAL_HARDWARE_TYPE, XML_PHYSIO_DEVICE);

			/* property==NAME parent==PHYSIO_DEVICE */
			create_v3_property(io_dev, NAME_NODE,
			    (xmlChar *)biodevs_p->name);

			/* property==PSEUDONYM parent==PHYSIO_DEVICE */
			create_v3_property(io_dev, XML_PSEUDONYM,
			    (xmlChar *)biodevs_p->phys_iodev->cli_name);

			/* property==DEV_TYPE parent==PHYSIO_DEVICE */
			create_v3_property(io_dev, XML_DEV_TYPE,
			    (xmlChar *)biodevs_p->phys_iodev->device_type);
		}
	}
}

void
add_v3_policy_output(xmlNodePtr content, ldom_t *ldp, int type)
{
	xmlNodePtr	pnode;
	policy_t	*polp;
	char		int_str[INT_STR_LEN];

	if (CHAIN_LENGTH(ldp->pchain.policy) < 1)
		return;

	mutex_lock(&ldp->pchain.lock);
	CHAIN_ITER(ldp->pchain.policy, polp, link) {
		/* Resource==POLICY parent==LDOM */
		pnode = create_v3_section(content, VIRTUAL_HARDWARE_TYPE,
		    POLICY_NODE);

		/* property==NAME parent==POLICY */
		create_v3_property(pnode, POLICY_NAME_NODE,
		    (xmlChar *)polp->namep);

		/* property==ENABLE parent==POLICY */
		if (polp->enable == true)
			create_v3_property(pnode, POLICY_ENABLE_NODE,
			    (xmlChar *)"on");
		else
			create_v3_property(pnode, POLICY_ENABLE_NODE,
			    (xmlChar *)"off");

		/* property==PRIORITY parent==POLICY */
		snprintf(int_str, sizeof (int_str), "%d", polp->priority);
		create_v3_property(pnode, POLICY_PRIORITY_NODE,
		    (xmlChar *)int_str);

		/* property==VCPU_MIN parent==POLICY */
		snprintf(int_str, sizeof (int_str), "%d", polp->vcpu_min);
		create_v3_property(pnode, POLICY_VCPU_MIN_NODE,
		    (xmlChar *)int_str);

		/* property==VCPU_MAX parent==POLICY */
		if (polp->vcpu_max != POLICY_UNLIMITED_CPU) {
			snprintf(int_str, sizeof (int_str), "%d",
			    polp->vcpu_max);
			create_v3_property(pnode, POLICY_VCPU_MAX_NODE,
			    (xmlChar *)int_str);
		}

		/* property==UTIL_LOWER parent==POLICY */
		snprintf(int_str, sizeof (int_str), "%d", polp->util_lower);
		create_v3_property(pnode, POLICY_UTIL_LOWER_NODE,
		    (xmlChar *)int_str);

		/* property==UTIL_UPPER parent==POLICY */
		snprintf(int_str, sizeof (int_str), "%d", polp->util_upper);
		create_v3_property(pnode, POLICY_UTIL_UPPER_NODE,
		    (xmlChar *)int_str);

		/* property==TOD_BEGIN parent==POLICY */
		create_v3_property(pnode, POLICY_TOD_BEGIN_NODE,
		    (xmlChar *)tod2str(polp->tod_begin));

		/* property==TOD_END parent==POLICY */
		create_v3_property(pnode, POLICY_TOD_END_NODE,
		    (xmlChar *)tod2str(polp->tod_end));

		/* property==SAMPLE_RATE parent==POLICY */
		snprintf(int_str, sizeof (int_str), "%d", polp->sample_rate);
		create_v3_property(pnode, POLICY_SAMPLE_RATE_NODE,
		    (xmlChar *)int_str);

		/* property==ELASTIC_MARGIN parent==POLICY */
		snprintf(int_str, sizeof (int_str), "%d", polp->elastic_margin);
		create_v3_property(pnode, POLICY_ELASTIC_MARGIN_NODE,
		    (xmlChar *)int_str);

		/* property==ATTACK parent==POLICY */
		if (polp->attack != POLICY_UNLIMITED_ATTACK) {
			snprintf(int_str, sizeof (int_str), "%d", polp->attack);
			create_v3_property(pnode, POLICY_ATTACK_NODE,
			    (xmlChar *)int_str);
		}

		/* property==DECAY parent==POLICY */
		snprintf(int_str, sizeof (int_str), "%d", polp->decay);
		create_v3_property(pnode, POLICY_DECAY_NODE,
		    (xmlChar *)int_str);
	}
	mutex_unlock(&ldp->pchain.lock);
}

void
add_v3_stats_ldom_info(xmlNodePtr ldom_info, ldom_t *ldp, int type)
{
	policy_t	*polp;
	char		util_str[INT_STR_LEN];

	if (CHAIN_LENGTH(ldp->pchain.policy) < 1)
		return;
	if (!pm_migration_is_possible())
		return;
	if (ldp->state < ACTIVE)
		return;

	mutex_lock(&ldp->pchain.lock);
	CHAIN_ITER(ldp->pchain.policy, polp, link) {
		if (polp->enable == true) {
			/* property==WEIGHTED-MEAN-UTIL parent==LDOM_INFO */
			if (ldp->pchain.mean_util[0] >= 10.0) {
				snprintf(util_str, sizeof (util_str), "%.0f%%",
				    ldp->pchain.mean_util[0]);
			} else {
				snprintf(util_str, sizeof (util_str), "%.1f%%",
				    ldp->pchain.mean_util[0]);
			}
			create_v3_property(ldom_info, XML_WEIGHTED_MEAN_UTIL,
			    (xmlChar *)util_str);

			mutex_unlock(&ldp->pchain.lock);
			return;
		}
	}
	mutex_unlock(&ldp->pchain.lock);
}
