/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "basics.h"
#include "errors.h"
#include "support.h"
#include "ldomconfigure.h"
#include "allocate.h"
#include "database.h"
#include "response.h"
#include "ldom.h"
#include "fatal.h"
#include "clp.h"
#include "tod.h"
#include "zprintf.h"
#include "database_internal.h"
#include "xml_output.h"
#include "xml_parse.h"
#include "xml_cntrl_common.h"
#include "xml_common.h"
#include "xml_v3_defs.h"

#define	FILE_MODE	(S_ISVTX + S_IRUSR + S_IWUSR)
#define	ZVOL_DSK	"/dev/zvol/dsk/"
#define	ZVOL_RDSK	"/dev/zvol/rdsk/"

static bool_t
empty_tag_check(xmlChar *tag_data, xmlNodePtr ret_data, xmlChar *node_name)
{
	if ((tag_data == NULL) || (strcmp((char *)tag_data, "") == 0)) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_TAG_MISSING_DATA, (char *)node_name);
		/*
		 * If the char * is filled we need to free it to avoid memory
		 * leaks.
		 */
		if (tag_data != NULL) {
			xmlFree(tag_data);
		}
		return (true);
	}
	return (false);
}

/*
 * Parse the options tag of the resourse commands and return the number of
 * options
 */
int
xml_v3_res_opts(xmlNodePtr ret_data, char ***options)
{
	xmlNodePtr ret_cmd;
	int n_options;

	ret_cmd = ret_data->parent;
	ASSERT(ret_cmd != NULL);

	n_options = parse_options_node(ret_cmd, options);

	if (n_options == -1) {
		xml_add_status(ret_cmd, XML_FAILURE, "<%s> tag contents too"
		    "long", XML_OPTIONS);
		return (-1);
	}

	return (n_options);
}

int
xml_v3_vcpu(xmlNodePtr res_node, ldom_t *ldp, int action, xmlNodePtr ret_data,
	cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	vcpu_args = 2;
	void		*argv[vcpu_args];
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*number;
	int		argc = 0, i;
	bool_t		result;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "CPU");
		return (-1);
	}

	subnode = xml_find_subnode(item, ALLOCATION_NODE_IN);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, CPU_NODE, ALLOCATION_NODE_IN);
		return (-1);
	}

	number = xmlNodeGetContent(subnode);
	if (empty_tag_check(number, ret_data, ALLOCATION_NODE_IN)) {
		return (-1);
	}
	argv[argc++] = Xstrdup((char *)number);
	xmlFree(number);

	argv[argc++] = Xstrdup(ldp->namep);

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	switch (action) {
	case XML_ADD_RES:
		cmd_num = LDM_ADD_VCPU;
		break;
	case XML_MODIFY_RES:
		cmd_num = LDM_MODIFY_VCPU;
		break;
	case XML_REMOVE_RES:
		cmd_num = LDM_REMOVE_VCPU;
		break;
	}
	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_vcrypto(xmlNodePtr res_node, ldom_t *ldp, int action,
    xmlNodePtr ret_data, cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	vcrypto_args = 2;
	void		*argv[vcrypto_args];
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*number;
	int		argc = 0, i;
	bool_t		result;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "MAU");
		return (-1);
	}

	subnode = xml_find_subnode(item, ALLOCATION_NODE_IN);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, MAU_NODE, ALLOCATION_NODE_IN);
		return (-1);
	}

	number = xmlNodeGetContent(subnode);
	if (empty_tag_check(number, ret_data, ALLOCATION_NODE_IN)) {
		return (-1);
	}
	argv[argc++] = Xstrdup((char *)number);
	xmlFree(number);

	argv[argc++] = Xstrdup(ldp->namep);

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	switch (action) {
	case XML_ADD_RES:
		cmd_num = LDM_ADD_CRYPTO;
		break;
	case XML_MODIFY_RES:
		cmd_num = LDM_MODIFY_CRYPTO;
		break;
	case XML_REMOVE_RES:
		cmd_num = LDM_REMOVE_CRYPTO;
		break;
	}
	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_vmem(xmlNodePtr res_node, ldom_t *ldp, int action, xmlNodePtr ret_data,
    cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	vmem_args = 4;
	void		*argv[vmem_args];
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*mem_size;
	int		argc = 0, i;
	bool_t		result;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "Memory");
		return (-1);
	}

	if (action == XML_MODIFY_RES) {
		/* option --auto-adj? */
		subnode = xml_find_subnode(item, XML_AUTO_ADJ);
		if (subnode != NULL)
			argv[argc++] = zsprintf("--auto-adj");
	}

	subnode = xml_find_subnode(item, ALLOCATION_NODE_IN);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, MEMORY_NODE,
		    ALLOCATION_NODE_OUT);
		return (-1);
	}

	mem_size = xmlNodeGetContent(subnode);
	if (empty_tag_check(mem_size, ret_data, ALLOCATION_NODE_IN)) {
		return (-1);
	}
	argv[argc++] = Xstrdup((char *)mem_size);
	xmlFree(mem_size);

	argv[argc++] = Xstrdup(ldp->namep);

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	switch (action) {
	case XML_ADD_RES:
		cmd_num = LDM_ADD_MEM;
		break;
	case XML_MODIFY_RES:
		cmd_num = LDM_MODIFY_MEM;
		break;
	case XML_REMOVE_RES:
		cmd_num = LDM_REMOVE_MEM;
		break;
	}
	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_vds(xmlNodePtr res_node, ldom_t *ldp, int action, xmlNodePtr ret_data,
    cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	vds_args = 2;
	void		*argv[vds_args];
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*vds_name;
	int		argc = 0, i;
	bool_t		result;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "VDS Service");
		return (-1);
	}

	subnode = xml_find_v3_property(item, SERVICE_NAME_NODE);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, VDS_NODE, SERVICE_NAME_NODE);
		return (-1);
	}

	vds_name = xmlNodeGetContent(subnode);
	if (empty_tag_check(vds_name, ret_data, SERVICE_NAME_NODE)) {
		return (-1);
	}
	argv[argc++] = Xstrdup((char *)vds_name);
	xmlFree(vds_name);

	if (action == XML_ADD_RES) {
		argv[argc++] = Xstrdup(ldp->namep);
		cmd_num = LDM_ADD_VDS;
	} else {
		cmd_num = LDM_REMOVE_VDS;
	}

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
vdsdev_mkfile(xmlNodePtr ret_data, xmlNodePtr mkfile_node, xmlChar *vds_device,
	cli_conn_t *client)
{
	int		fd;
	xmlChar		*file_size;
	uint64_t	size;
	char		unit;

	file_size = xmlNodeGetContent(mkfile_node);
	if (empty_tag_check(file_size, ret_data, XML_MKFILE)) {
		return (-1);
	}

	if (!parse_mem_size(NULL, &size, (char *)file_size, &unit)) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_INVALID_FILE_SIZE, vds_device);
		xmlFree(file_size);
		return (-1);
	}

	fd = open64((char *)vds_device, O_CREAT | O_RDWR | O_EXCL |
	    O_LARGEFILE, FILE_MODE);
	if (fd < 0) {
		xml_add_status(ret_data, XML_FAILURE, MSG_ERR_UNABLE_TO_OPEN,
		    vds_device, strerror(errno));
		xmlFree(file_size);
		return (-1);
	}
	if (lseek64(fd, (size - 1), SEEK_SET) < 0) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_UNABLE_TO_CREATE, vds_device);
		DBG(zprintf("Failed lseek(): %s\n", strerror(errno)));
		xmlFree(file_size);
		close(fd);
		remove((char *)vds_device);
		return (-1);
	}
	if (write(fd, "", 1) != 1) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_UNABLE_TO_CREATE, vds_device);
		xmlFree(file_size);
		close(fd);
		remove((char *)vds_device);
		return (-1);
	}
	close(fd);
	xmlFree(file_size);
	return (0);
}

int
vdsdev_zfs(xmlNodePtr ret_data, xmlNodePtr zfs_node, char *vds_device)
{
	int		ret;
	xmlChar		*file_size;
	char		*system_cmd;
	char		*zvol;

	file_size = xmlNodeGetContent(zfs_node);
	if (empty_tag_check((xmlChar *)vds_device, ret_data, XML_ZFS)) {
		return (-1);
	}

	if (strncmp(vds_device, ZVOL_DSK, (sizeof (ZVOL_DSK) - 1)) == 0) {
		zvol = vds_device + (sizeof (ZVOL_DSK) - 1);
	} else if (strncmp(vds_device, ZVOL_RDSK,
	    (sizeof (ZVOL_RDSK) - 1)) == 0) {
		zvol = vds_device + (sizeof (ZVOL_RDSK) - 1);
	} else {
		xml_add_status(ret_data, XML_FAILURE, MSG_ERR_ZVOL_CREATE);
		xmlFree(file_size);
		return (-1);
	}
	system_cmd = zsprintf("zfs create -V %s %s", file_size, zvol);
	xmlFree(file_size);
	ret = system(system_cmd);
	Xfree(system_cmd);
	if (ret != 0) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_UNABLE_TO_CREATE, vds_device);
		return (-1);
	}
	return (0);
}

int
xml_v3_vdsdev(xmlNodePtr res_node, ldom_t *ldp, int action,
    xmlNodePtr ret_data, cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	vdsdev_args = 5;
	void		*argv[vdsdev_args];
	xmlNodePtr	subnode = NULL;
	xmlNodePtr	item, mkfile, zfs;
	xmlNodePtr	optnode = NULL;
	xmlNodePtr	mgpnode = NULL;
	xmlChar		*vds_name, *vdsdev_name, *vds_device;
	xmlChar		*options, *mgpname;
	int		argc = 0, i;
	bool_t		result;
	int		n_options;
	char		**opts = NULL;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "vds_volume");
		return (-1);
	}

	if (action != XML_REMOVE_RES) {
		n_options = xml_v3_res_opts(ret_data, &opts);
		if (n_options < 0) {
			return (-1);
		} else if (n_options > 1) {
			xml_add_status(ret_data, XML_FAILURE, "Malformed XML "
			    "Document: vds_volume Section have more than one "
			    "options, only -f is allowed");
			free_argv((void **)opts, n_options);
			Xfree(opts);
			return (-1);
		} else if (n_options > 0) {
			for (i = 0; i < n_options; i++) {
				argv[argc++] = Xstrdup(opts[i]);
			}
			free_argv((void **)opts, n_options);
			Xfree(opts);
		}

		optnode = xml_find_v3_property(item, VOLUME_OPTS_NODE);
		if (optnode != NULL) {
			options = xmlNodeGetContent(optnode);
			if (options == NULL)
				argv[argc++] = zsprintf("options=");
			else {
				argv[argc++] = zsprintf("options=%s", options);
				xmlFree(options);
			}
		}

		mgpnode = xml_find_v3_property(item, VOLUME_MPGROUP_NODE);
		if (mgpnode != NULL) {
			mgpname = xmlNodeGetContent(mgpnode);
			if (mgpnode == NULL)
				argv[argc++] = zsprintf("mpgroup=");
			else {
				argv[argc++] = zsprintf("mpgroup=%s", mgpname);
				xmlFree(mgpname);
			}
		}

		if (action == XML_MODIFY_RES && optnode == NULL &&
		    mgpnode == NULL) {
			xml_add_status(ret_data, XML_FAILURE,
			    MSG_ERR_VDSDEV_MISSING_TAG,
			    (char *)VDS_VOLUME_NODE, (char *)VOLUME_OPTS_NODE,
			    (char *)VOLUME_MPGROUP_NODE);
			return (-1);
		}

		if (action == XML_ADD_RES) {
			subnode = xml_find_v3_property(item, BLOCKDEV_NODE);
			if (subnode == NULL) {
				set_missing_prop_msg(ret_data, VDS_VOLUME_NODE,
				    BLOCKDEV_NODE);
				free_argv(argv, argc);
				return (-1);
			}
			vds_device = xmlNodeGetContent(subnode);
			if (empty_tag_check(vds_device, ret_data,
			    BLOCKDEV_NODE)) {
				free_argv(argv, argc);
				return (-1);
			}
			argv[argc++] = Xstrdup((char *)vds_device);
			mkfile = xml_find_v3_property(item, XML_MKFILE);
			zfs = xml_find_v3_property(item, XML_ZFS);
			if ((mkfile != NULL) && (zfs != NULL)) {
				xml_add_status(ret_data, XML_FAILURE,
				    MSG_ERR_MKFILE_AND_ZVOL);
				return (-1);
			}
			if (mkfile != NULL) {
				ret = vdsdev_mkfile(ret_data, mkfile,
				    vds_device, client);
				if (ret == -1) {
					free_argv(argv, argc);
					xmlFree(vds_device);
					return (-1);
				}
			}
			if (zfs != NULL) {
				ret = vdsdev_zfs(ret_data, zfs,
				    (char *)vds_device);
				if (ret == -1) {
					free_argv(argv, argc);
					xmlFree(vds_device);
					return (-1);
				}
			}
			xmlFree(vds_device);
			cmd_num = LDM_ADD_VDSDEV;
		} else if (action == XML_MODIFY_RES) {
			cmd_num = LDM_MODIFY_VDSDEV;
		}
	} else {
		cmd_num = LDM_REMOVE_VDSDEV;
	}

	/* Get the volume name and the service name */
	subnode = xml_find_v3_property(item, VOLUME_NAME_NODE);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, VDS_VOLUME_NODE,
		    VOLUME_NAME_NODE);
		free_argv(argv, argc);
		return (-1);
	}
	vdsdev_name = xmlNodeGetContent(subnode);
	if (empty_tag_check(vdsdev_name, ret_data, VOLUME_NAME_NODE)) {
		free_argv(argv, argc);
		return (-1);
	}

	subnode = xml_find_v3_property(item, SERVICE_NAME_NODE);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, VDS_VOLUME_NODE,
		    SERVICE_NAME_NODE);
		xmlFree(vdsdev_name);
		free_argv(argv, argc);
		return (-1);
	}
	vds_name = xmlNodeGetContent(subnode);
	if (empty_tag_check(vds_name, ret_data, SERVICE_NAME_NODE)) {
		xmlFree(vdsdev_name);
		free_argv(argv, argc);
		return (-1);
	}

	/*
	 * Put the volume and service name in the
	 * <volume_name>@<service_name> format and make it the 2nd arguement
	 */
	argv[argc++] = zsprintf("%s@%s", vdsdev_name, vds_name);
	xmlFree(vdsdev_name);
	xmlFree(vds_name);

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_vdisk(xmlNodePtr res_node, ldom_t *ldp, int action, xmlNodePtr ret_data,
    cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	vdisk_args = 5;
	void		*argv[vdisk_args];
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*vds_name, *vdsdev_name, *vdisk_name, *timeout, *id;
	int		argc = 0, i;
	bool_t		result;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "disk");
		return (-1);
	}

	/*
	 * parse id, timeout, vds server and volume nodes for add-vdisk
	 * and set-vdisk
	 */
	if (action != XML_REMOVE_RES) {
		xmlNodePtr volume_node = NULL;
		xmlNodePtr service_node = NULL;
		xmlNodePtr timeout_node = NULL;
		xmlNodePtr id_node = NULL;

		/* option timeout */
		timeout_node = xml_find_v3_property(item, TIMEOUT_NODE);
		if (timeout_node != NULL) {
			timeout = xmlNodeGetContent(timeout_node);
			if (timeout == NULL)
				argv[argc++] = zsprintf("timeout=");
			else {
				argv[argc++] = zsprintf("timeout=%s", timeout);
				xmlFree(timeout);
			}
		}

		/* option id */
		id_node = xml_find_v3_property(item, ID_NODE);
		if (id_node != NULL) {
			id = xmlNodeGetContent(id_node);
			if (id != NULL) {
				if (action == XML_ADD_RES) {
					argv[argc++] = zsprintf("id=%s", id);
					xmlFree(id);
				}
			}
		}

		service_node = xml_find_v3_property(item, SERVICE_NAME_NODE);
		volume_node = xml_find_v3_property(item, VOLUME_NAME_NODE);

		if (action == XML_ADD_RES) {
			/*
			 * for add-vdisk, both vds server and volume are
			 * non NULL strings.
			 */
			if (volume_node == NULL) {
				set_missing_prop_msg(ret_data, DISK_NODE,
				    VOLUME_NAME_NODE);
				free_argv(argv, argc);
				return (-1);
			}

			if (service_node == NULL) {
				set_missing_prop_msg(ret_data, DISK_NODE,
				    SERVICE_NAME_NODE);
				free_argv(argv, argc);
				return (-1);
			}

			vds_name = xmlNodeGetContent(service_node);
			if (empty_tag_check(vds_name, ret_data,
			    SERVICE_NAME_NODE)) {
				free_argv(argv, argc);
				return (-1);
			}
			vdsdev_name = xmlNodeGetContent(volume_node);
			if (empty_tag_check(vdsdev_name, ret_data,
			    VOLUME_NAME_NODE)) {
				xmlFree(vds_name);
				free_argv(argv, argc);
				return (-1);
			}
		} else {
			/*
			 * for set-vdisk, either both vds server and volume are
			 * non NULL strings or both are NULL strings.
			 */
			if (service_node == NULL && volume_node != NULL) {
				set_missing_prop_msg(ret_data, DISK_NODE,
				    SERVICE_NAME_NODE);
				free_argv(argv, argc);
				return (-1);
			}

			if (volume_node == NULL && service_node != NULL) {
				set_missing_prop_msg(ret_data, DISK_NODE,
				    VOLUME_NAME_NODE);
				free_argv(argv, argc);
				return (-1);
			}

			if (volume_node == NULL && service_node == NULL &&
			    timeout_node == NULL) {
				xml_add_status(ret_data, XML_FAILURE,
				    MSG_ERR_VDISK_MISSING_ALL);
				free_argv(argv, argc);
				return (-1);
			}

			if (service_node != NULL) {
				vds_name = xmlNodeGetContent(service_node);
				if (empty_tag_check(vds_name, ret_data,
				    SERVICE_NAME_NODE)) {
					free_argv(argv, argc);
					return (-1);
				}
				vdsdev_name = xmlNodeGetContent(volume_node);
				if (empty_tag_check(vdsdev_name, ret_data,
				    VOLUME_NAME_NODE)) {
					xmlFree(vds_name);
					free_argv(argv, argc);
					return (-1);
				}
				argv[argc++] = zsprintf("volume=%s@%s",
				    vdsdev_name, vds_name);
				xmlFree(vds_name);
				xmlFree(vdsdev_name);
			}

		}

	}

	/* parse vdisk name */
	subnode = xml_find_v3_property(item, VDISK_NAME_NODE);
	if (subnode == NULL) {
		set_missing_prop_msg(item, DISK_NODE, VDISK_NAME_NODE);
		free_argv(argv, argc);
		return (-1);
	}
	vdisk_name = xmlNodeGetContent(subnode);
	if (empty_tag_check(vdisk_name, ret_data, VDISK_NAME_NODE)) {
		free_argv(argv, argc);
		return (-1);
	}
	argv[argc++] = Xstrdup((char *)vdisk_name);
	xmlFree(vdisk_name);

	if (action == XML_ADD_RES) {
		argv[argc++] = zsprintf("%s@%s", vdsdev_name, vds_name);
		xmlFree(vdsdev_name);
		xmlFree(vds_name);
		cmd_num = LDM_ADD_VDISK;
	} else if (action == XML_MODIFY_RES) {
		cmd_num = LDM_MODIFY_VDISK;
	} else {
		cmd_num = LDM_REMOVE_VDISK;
	}

	argv[argc++] = Xstrdup(ldp->namep);

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_vsw(xmlNodePtr res_node, ldom_t *ldp, int action, xmlNodePtr ret_data,
    cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	vsw_args = 11;
	void		*argv[vsw_args];
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*nws_name, *mac_addr;
	xmlChar		*content, *net_dev, *phys_state;
	xmlChar		*mode, *id;
	int		argc = 0, i;
	bool_t		result;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "vsw");
		return (-1);
	}

	subnode = xml_find_v3_property(item, VSW_DEF_VLAN_NODE);
	if (subnode != NULL) {
		content = xmlNodeGetContent(subnode);
		if (empty_tag_check(content, ret_data, VSW_DEF_VLAN_NODE)) {
			return (-1);
		}
		argv[argc++] = zsprintf("default-vlan-id=%s", content);
		xmlFree(content);
	}

	subnode = xml_find_v3_property(item, VSW_PORT_VLAN_NODE);
	if (subnode != NULL) {
		content = xmlNodeGetContent(subnode);
		if (content == NULL) {
			if (action == LDM_MODIFY_VSW) {
				argv[argc++] = zsprintf("pvid=");
			} else if (action == LDM_ADD_VSW) {
				xml_add_status(ret_data, XML_FAILURE,
				    MSG_ERR_VSW_MUST_HAVE_DATA, "pvid");
				free_argv(argv, argc);
				return (-1);
			}
		} else {
			argv[argc++] = zsprintf("pvid=%s", content);
			xmlFree(content);
		}
	}

	subnode = xml_find_v3_property(item, VSW_VLANS_NODE);
	if (subnode != NULL) {
		content = xmlNodeGetContent(subnode);
		if (content == NULL) {
			if (action == LDM_MODIFY_VSW) {
				argv[argc++] = zsprintf("vid=");
			} else if (action == LDM_ADD_VSW) {
				xml_add_status(ret_data, XML_FAILURE,
				    MSG_ERR_VSW_MUST_HAVE_DATA, "vid");
				free_argv(argv, argc);
				return (-1);
			}
		} else {
			argv[argc++] = zsprintf("vid=%s", content);
			xmlFree(content);
		}
	}

	subnode = xml_find_v3_property(item, VSW_MTU_NODE);
	if (subnode != NULL) {
		content = xmlNodeGetContent(subnode);
		if (content == NULL) {
			if (action == LDM_MODIFY_VNET) {
				argv[argc++] = zsprintf("mtu=");
			} else if (action == LDM_ADD_VSW) {
				xml_add_status(ret_data, XML_FAILURE,
				    MSG_ERR_VSW_MUST_HAVE_DATA, "mtu");
				free_argv(argv, argc);
				return (-1);
			}
		} else {
			argv[argc++] = zsprintf("mtu=%s", content);
			xmlFree(content);
		}
	}

	subnode = xml_find_subnode(item, ADDRESS_NODE_IN);
	if (subnode != NULL) {
		mac_addr = xmlNodeGetContent(subnode);
		if (empty_tag_check(mac_addr, ret_data, ADDRESS_NODE_IN)) {
			free_argv(argv, argc);
			return (-1);
		}
		if (strcmp((char *)mac_addr,
		    XML_AUTO_ALLOCATED_MACADDR) != 0) {
			argv[argc++] = zsprintf("mac-addr=%s", mac_addr);
		}
		xmlFree(mac_addr);
	}

	subnode = xml_find_v3_property(item, DEV_PATH_NODE);
	if (subnode != NULL) {
		net_dev = xmlNodeGetContent(subnode);
		if (empty_tag_check(net_dev, ret_data, DEV_PATH_NODE)) {
			free_argv(argv, argc);
			return (-1);
		}
		argv[argc++] = zsprintf("net-dev=%s", net_dev);
		xmlFree(net_dev);

		/*
		 * If we got this far, a non-NULL net-dev just got copied
		 * and now time to copy linkprop
		 */
		subnode = xml_find_v3_property(item, LINKPROP_NODE);
		if (subnode != NULL) {
			phys_state = xmlNodeGetContent(subnode);
			if (phys_state == NULL) {
				argv[argc++] = zsprintf("linkprop=");
			} else {
				argv[argc++] = zsprintf("linkprop=%s",
				    phys_state);
				xmlFree(phys_state);
			}
		}
	}

	subnode = xml_find_v3_property(item, VSW_MODE_NODE);
	if (subnode != NULL) {
		mode = xmlNodeGetContent(subnode);
		if (mode == NULL)
			argv[argc++] = zsprintf("mode=");
		else {
			argv[argc++] = zsprintf("mode=%s", mode);
			xmlFree(mode);
		}
	}

	subnode = xml_find_v3_property(item, ID_NODE);
	if (subnode != NULL) {
		id = xmlNodeGetContent(subnode);
		if (id != NULL) {
			if (action == XML_ADD_RES) {
				argv[argc++] = zsprintf("id=%s", id);
				xmlFree(id);
			}
		}
	}

	subnode = xml_find_v3_property(item, SERVICE_NAME_NODE);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, VSW_NODE, SERVICE_NAME_NODE);
		free_argv(argv, argc);
		return (-1);
	}
	nws_name = xmlNodeGetContent(subnode);
	if (empty_tag_check(nws_name, ret_data, SERVICE_NAME_NODE)) {
		free_argv(argv, argc);
		return (-1);
	}
	argv[argc++] = Xstrdup((char *)nws_name);
	xmlFree(nws_name);

	if (action == XML_ADD_RES) {
		argv[argc++] = Xstrdup(ldp->namep);
	}

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	switch (action) {
	case XML_ADD_RES:
		cmd_num = LDM_ADD_VSW;
		break;
	case XML_MODIFY_RES:
		cmd_num = LDM_MODIFY_VSW;
		break;
	case XML_REMOVE_RES:
		cmd_num = LDM_REMOVE_VSW;
		break;
	}
	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

static int
fill_v3_vsw(xmlNodePtr item, xmlNodePtr ret_data, int action, int *argc,
    void **argv)
{
	xmlNodePtr	subnode;
	xmlChar		*nws_name;

	subnode = xml_find_v3_property(item, SERVICE_NAME_NODE);
	if ((action == XML_ADD_RES) && (subnode == NULL)) {
		set_missing_prop_msg(ret_data, VNET_NODE, SERVICE_NAME_NODE);
		return (-1);
	} else if (subnode == NULL) {
		return (0);
	}

	nws_name = xmlNodeGetContent(subnode);
	if (empty_tag_check(nws_name, ret_data, SERVICE_NAME_NODE)) {
		return (-1);
	}
	if (action == XML_ADD_RES) {
		argv[(*argc)++] = Xstrdup((char *)nws_name);
		xmlFree(nws_name);
	} else {
		argv[(*argc)++] = zsprintf("vswitch=%s", nws_name);
		xmlFree(nws_name);
	}
	return (0);
}

int
xml_v3_vnet(xmlNodePtr res_node, ldom_t *ldp, int action, xmlNodePtr ret_data,
    cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	vnet_args = 10;
	void		*argv[vnet_args];
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*mac_addr;
	xmlChar		*content, *vnet_name;
	xmlChar		*vnet_mode;
	int		argc = 0, i;
	bool_t		result;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "network");
		return (-1);
	}

	if (action != XML_REMOVE_RES) {
		subnode = xml_find_subnode(item, ADDRESS_NODE_IN);
		if (subnode != NULL) {
			mac_addr = xmlNodeGetContent(subnode);
			if (empty_tag_check(mac_addr, ret_data,
			    ADDRESS_NODE_IN)) {
				return (-1);
			}
			if (strcmp((char *)mac_addr,
			    XML_AUTO_ALLOCATED_MACADDR) != 0) {
				argv[argc++] = zsprintf("mac-addr=%s",
				    mac_addr);
			}
			xmlFree(mac_addr);
		}
		subnode = xml_find_v3_property(item, VNET_MODE_NODE);
		if (subnode != NULL) {
			vnet_mode = xmlNodeGetContent(subnode);
			if (vnet_mode == NULL && action == XML_ADD_RES) {
				xml_add_status(ret_data, XML_FAILURE,
				    MSG_ERR_MISSING_PROPERTY,
				    (char *)VNET_NODE, (char *)VNET_MODE_NODE);
				return (-1);
			} else if (vnet_mode == NULL) {
				argv[argc++] = zsprintf("mode=");
			} else {
				argv[argc++] = zsprintf("mode=%s",
				    (char *)vnet_mode);
				xmlFree(vnet_mode);
			}
		}
		subnode = xml_find_v3_property(item, VNET_PORT_VLAN_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (content == NULL) {
				if (action == LDM_MODIFY_VNET) {
					argv[argc++] = zsprintf("pvid=");
				} else if (action == LDM_ADD_VNET) {
					xml_add_status(ret_data, XML_FAILURE,
					    MSG_ERR_VNET_MUST_HAVE_DATA,
					    "pvid");
					return (-1);
				}
			} else {
				argv[argc++] = zsprintf("pvid=%s", content);
				xmlFree(content);
			}
		}
		subnode = xml_find_v3_property(item, VNET_VLANS_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (content == NULL) {
				if (action == LDM_MODIFY_VNET) {
					argv[argc++] = zsprintf("vid=");
				} else if (action == LDM_ADD_VNET) {
					xml_add_status(ret_data, XML_FAILURE,
					    MSG_ERR_VNET_MUST_HAVE_DATA,
					    "vid");
					return (-1);
				}
			} else {
				argv[argc++] = zsprintf("vid=%s", content);
				xmlFree(content);
			}
		}
		subnode = xml_find_v3_property(item, MTU_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (content == NULL) {
				if (action == LDM_MODIFY_VNET) {
					argv[argc++] = zsprintf("mtu=");
				} else if (action == LDM_ADD_VNET) {
					xml_add_status(ret_data, XML_FAILURE,
					    MSG_ERR_VNET_MUST_HAVE_DATA,
					    "mtu");
					return (-1);
				}
			} else {
				argv[argc++] = zsprintf("mtu=%s", content);
				xmlFree(content);
			}
		}
		subnode = xml_find_v3_property(item, LINKPROP_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (content == NULL) {
				if (action == LDM_MODIFY_VNET) {
					argv[argc++] = zsprintf("linkprop=");
				} else if (action == LDM_ADD_VNET) {
					xml_add_status(ret_data, XML_FAILURE,
					    MSG_ERR_VNET_MUST_HAVE_DATA,
					    "linkprop");
					return (-1);
				}
			} else {
				argv[argc++] = zsprintf("linkprop=%s", content);
				xmlFree(content);
			}
		}
		subnode = xml_find_v3_property(item, ID_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (content != NULL) {
				if (action == XML_ADD_RES) {
					argv[argc++] = zsprintf("id=%s",
					    content);
					xmlFree(content);
				}
			}
		}
	}

	if (action == XML_MODIFY_RES) {
		ret = fill_v3_vsw(item, ret_data, action, &argc, argv);
		if (ret == -1) {
			free_argv(argv, argc);
			return (-1);
		}
	}

	subnode = xml_find_v3_property(item, VNET_NAME_NODE);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, VNET_NODE, VNET_NAME_NODE);
		free_argv(argv, argc);
		return (-1);
	}
	vnet_name = xmlNodeGetContent(subnode);
	if (empty_tag_check(vnet_name, ret_data, VNET_NAME_NODE)) {
		free_argv(argv, argc);
		return (-1);
	}
	argv[argc++] = Xstrdup((char *)vnet_name);
	xmlFree(vnet_name);

	if (action == XML_ADD_RES) {
		ret = fill_v3_vsw(item, ret_data, action, &argc, argv);
		if (ret == -1) {
			free_argv(argv, argc);
			return (-1);
		}
	}

	argv[argc++] = Xstrdup(ldp->namep);

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	switch (action) {
	case XML_ADD_RES:
		cmd_num = LDM_ADD_VNET;
		break;
	case XML_MODIFY_RES:
		cmd_num = LDM_MODIFY_VNET;
		break;
	case XML_REMOVE_RES:
		cmd_num = LDM_REMOVE_VNET;
		break;
	}
	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_vcc(xmlNodePtr res_node, ldom_t *ldp, int action, xmlNodePtr ret_data,
    cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	vcc_args = 3;
	void		*argv[vcc_args];
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*vcc_name, *port_low, *port_high;
	int		argc = 0, i;
	bool_t		result;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "vcc");
		return (-1);
	}

	if (action != XML_REMOVE_RES) {
		subnode = xml_find_v3_property(item, MIN_PORT_NODE);
		if (subnode == NULL) {
			set_missing_prop_msg(ret_data, VCC_NODE,
			    MIN_PORT_NODE);
			return (-1);
		}
		port_low = xmlNodeGetContent(subnode);
		if (empty_tag_check(port_low, ret_data, MIN_PORT_NODE)) {
			return (-1);
		}

		subnode = xml_find_v3_property(item, MAX_PORT_NODE);
		if (subnode == NULL) {
			set_missing_prop_msg(ret_data, VCC_NODE,
			    MAX_PORT_NODE);
			xmlFree(port_low);
			return (-1);
		}
		port_high = xmlNodeGetContent(subnode);
		if (empty_tag_check(port_high, ret_data, MAX_PORT_NODE)) {
			xmlFree(port_low);
			return (-1);
		}
		argv[argc++] = zsprintf("port-range=%s-%s", port_low,
		    port_high);
		xmlFree(port_low);
		xmlFree(port_high);
	}

	subnode = xml_find_v3_property(item, SERVICE_NAME_NODE);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, VCC_NODE, SERVICE_NAME_NODE);
		free_argv(argv, argc);
		return (-1);
	}
	vcc_name = xmlNodeGetContent(subnode);
	if (empty_tag_check(vcc_name, ret_data, SERVICE_NAME_NODE)) {
		free_argv(argv, argc);
		return (-1);
	}
	argv[argc++] = Xstrdup((char *)vcc_name);
	xmlFree(vcc_name);

	if (action == XML_ADD_RES) {
		argv[argc++] = Xstrdup(ldp->namep);
	}

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	switch (action) {
	case XML_ADD_RES:
		cmd_num = LDM_ADD_VCC;
		break;
	case XML_MODIFY_RES:
		cmd_num = LDM_MODIFY_VCC;
		break;
	case XML_REMOVE_RES:
		cmd_num = LDM_REMOVE_VCC;
		break;
	}
	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data,
	    client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_physio(xmlNodePtr res_node, ldom_t *ldp, int action,
    xmlNodePtr ret_data, cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	physio_args = 2;
	void		*argv[physio_args];
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*bus;
	int		argc = 0, i;
	bool_t		result;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "physio_device");
		return (-1);
	}

	subnode = xml_find_v3_property(item, NAME_NODE);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, XML_PHYSIO_DEVICE, NAME_NODE);
		return (-1);
	}
	bus = xmlNodeGetContent(subnode);
	if (empty_tag_check(bus, ret_data, NAME_NODE)) {
		return (-1);
	}
	argv[argc++] = Xstrdup((char *)bus);
	xmlFree(bus);

	argv[argc++] = Xstrdup(ldp->namep);

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	switch (action) {
	case XML_ADD_RES:
		cmd_num = LDM_ADD_IO;
		break;
	case XML_REMOVE_RES:
		cmd_num = LDM_REMOVE_IO;
		break;
	}
	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_var(xmlNodePtr res_node, ldom_t *ldp, int action, xmlNodePtr ret_data,
    cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	var_args = 2;
	void		*argv[var_args];
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*name, *value;
	int		argc = 0, i;
	bool_t		result;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "var");
		return (-1);
	}

	subnode = xml_find_v3_property(item, NAME_NODE);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, VAR_NODE, NAME_NODE);
		return (-1);
	}
	name = xmlNodeGetContent(subnode);
	if (empty_tag_check(name, ret_data, NAME_NODE)) {
		return (-1);
	}

	if (action == XML_ADD_RES) {
		subnode = xml_find_v3_property(item, VALUE_NODE);
		if (subnode == NULL) {
			set_missing_prop_msg(ret_data, VAR_NODE, VALUE_NODE);
			xmlFree(name);
			return (-1);
		}
		value = xmlNodeGetContent(subnode);
		if (value == NULL) {
			argv[argc++] = zsprintf("%s=", name);
		} else {
			argv[argc++] = zsprintf("%s=%s", name, value);
		}
		xmlFree(name);
		xmlFree(value);
		cmd_num = LDM_SET_VAR;
	} else {
		argv[argc++] = Xstrdup((char *)name);
		xmlFree(name);
		cmd_num = LDM_UNSET_VAR;
	}

	argv[argc++] = Xstrdup(ldp->namep);

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_spconfig(xmlNodePtr res_node, ldom_t *ldp, int action,
    xmlNodePtr ret_data, cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	int		var_args = 1;
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*name;
	int		argc = 0, i;
	bool_t		result;
	int		n_options = 0;
	char		**options = NULL;

	n_options = parse_options_node(ret_data->parent, &options);
	if (n_options == -1) {
		xml_add_status(ret_data->parent, XML_FAILURE,
		    "<%s> tag contents too long", XML_OPTIONS);
		return (-1);
	}
	var_args += n_options;

	void *argv[var_args];
	/* Add in any options to the argv */
	if (n_options != 0) {
		for (i = 0; i < n_options; i++) {
			argv[argc++] = Xstrdup(options[i]);
		}
	}

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "spconfig");
		return (-1);
	}

	subnode = xml_find_v3_property(item, XML_SPCONFIG_NAME);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, XML_SPCONFIG,
		    XML_SPCONFIG_NAME);
		return (-1);
	}
	name = xmlNodeGetContent(subnode);
	if (empty_tag_check(name, ret_data, XML_SPCONFIG_NAME)) {
		return (-1);
	}
	argv[argc++] = Xstrdup((char *)name);
	xmlFree(name);

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	switch (action) {
	case XML_ADD_RES:
		cmd_num = LDM_ADD_SET;
		break;
	case XML_MODIFY_RES:
		cmd_num = LDM_USE_SET;
		break;
	case XML_REMOVE_RES:
		cmd_num = LDM_REMOVE_SET;
		break;
	}
	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_console(xmlNodePtr res_node, ldom_t *ldp, int action,
    xmlNodePtr ret_data, cli_conn_t *client)
{
	int		ret = 0;
	const int	var_args = 4;
	void		*argv[var_args];
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*port, *group, *vcc_name;
	int		argc = 0, i;
	bool_t		result;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "console");
		return (-1);
	}

	subnode = xml_find_v3_property(item, VCONS_PORT_NODE);
	if (subnode != NULL) {
		port = xmlNodeGetContent(subnode);
		if (empty_tag_check(port, ret_data, VCONS_PORT_NODE)) {
			return (-1);
		}
		argv[argc++] = zsprintf("port=%s", port);
		xmlFree(port);
	}

	subnode = xml_find_v3_property(item, GROUP_NODE);
	if (subnode != NULL) {
		group = xmlNodeGetContent(subnode);
		if (empty_tag_check(group, ret_data, GROUP_NODE)) {
			free_argv(argv, argc);
			return (-1);
		}
		argv[argc++] = zsprintf("group=%s", group);
		xmlFree(group);
	}

	subnode = xml_find_v3_property(item, SERVICE_NAME_NODE);
	if (subnode != NULL) {
		vcc_name = xmlNodeGetContent(subnode);
		if (empty_tag_check(vcc_name, ret_data, SERVICE_NAME_NODE)) {
			free_argv(argv, argc);
			return (-1);
		}
		argv[argc++] = zsprintf("service=%s", vcc_name);
		xmlFree(vcc_name);
	}

	if (argc == 0) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "console");
		return (-1);
	}

	argv[argc++] = Xstrdup(ldp->namep);

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	result = xml_call_cli_cmd(argc, argv, LDM_MODIFY_VCONS, ret_data,
	    client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_vdpcs(xmlNodePtr res_node, ldom_t *ldp, int action, xmlNodePtr ret_data,
    cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	vdpcs_args = 2;
	void		*argv[vdpcs_args];
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*vdpcs_name;
	int		argc = 0, i;
	bool_t		result;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "vdpcs");
		return (-1);
	}

	subnode = xml_find_v3_property(item, SERVICE_NAME_NODE);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, VDPCS_NODE, SERVICE_NAME_NODE);
		return (-1);
	}

	vdpcs_name = xmlNodeGetContent(subnode);
	if (empty_tag_check(vdpcs_name, ret_data, SERVICE_NAME_NODE)) {
		return (-1);
	}
	argv[argc++] = Xstrdup((char *)vdpcs_name);
	xmlFree(vdpcs_name);

	if (action == XML_ADD_RES) {
		argv[argc++] = Xstrdup(ldp->namep);
		cmd_num = LDM_ADD_VDPCS;
	} else {
		cmd_num = LDM_REMOVE_VDPCS;
	}

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_vdpcc(xmlNodePtr res_node, ldom_t *ldp, int action, xmlNodePtr ret_data,
    cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	vdpcc_args = 3;
	void		*argv[vdpcc_args];
	xmlNodePtr	subnode = NULL, item;
	xmlChar		*vdpcs_name, *vdpcc_name;
	int		argc = 0, i;
	bool_t		result;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "vdpcc");
		return (-1);
	}

	subnode = xml_find_v3_property(item, VDPCC_NAME_NODE);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, VDPCC_NODE, VDPCC_NAME_NODE);
		return (-1);
	}
	vdpcc_name = xmlNodeGetContent(subnode);
	if (empty_tag_check(vdpcc_name, ret_data, VDPCC_NAME_NODE)) {
		return (-1);
	}
	argv[argc++] = Xstrdup((char *)vdpcc_name);
	xmlFree(vdpcc_name);

	if (action == XML_ADD_RES) {
		subnode = xml_find_v3_property(item, SERVICE_NAME_NODE);
		if (subnode == NULL) {
			set_missing_prop_msg(ret_data, VDPCC_NODE,
			    SERVICE_NAME_NODE);
			free_argv(argv, argc);
			return (-1);
		}
		vdpcs_name = xmlNodeGetContent(subnode);
		if (empty_tag_check(vdpcs_name, ret_data, SERVICE_NAME_NODE)) {
			free_argv(argv, argc);
			return (-1);
		}
		argv[argc++] = Xstrdup((char *)vdpcs_name);
		xmlFree(vdpcs_name);
		cmd_num = LDM_ADD_VDPCC;
	} else {
		cmd_num = LDM_REMOVE_VDPCC;
	}

	argv[argc++] = Xstrdup(ldp->namep);

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_ldm_info(xmlNodePtr res_node, ldom_t *ldp, int action,
    xmlNodePtr ret_data, cli_conn_t *client)
{
	int		ret = 0, cmd_num;
	const int	ldm_info_args = 5;
	void		*argv[ldm_info_args];
	xmlNodePtr	item;
	xmlNodePtr	sub_mac_address = NULL;
	xmlNodePtr	sub_hostid = NULL;
	xmlChar		*mac_address, *hostid;
	xmlNodePtr	sub_failure_policy = NULL;
	xmlNodePtr	sub_master = NULL;
	xmlChar		*failure_policy, *master;
	int		argc = 0, i;
	bool_t		result;


	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE,
		    MSG_ERR_SECTION_MISSING_CHILDREN, "ldom_info");
		return (-1);
	}

	sub_mac_address = xml_find_subnode(item, ADDRESS_NODE_IN);
	sub_hostid = xml_find_v3_property(item, XML_HOSTID);
	sub_failure_policy = xml_find_v3_property(item, FAILURE_POLICY_NODE);
	sub_master = xml_find_v3_property(item, MASTER_NODE);

	if (sub_mac_address == NULL && sub_hostid == NULL &&
	    sub_failure_policy == NULL && sub_master == NULL) {
		xml_add_status(ret_data, MSG_ERR_SECTION_HAS_NO_PROPERTY,
		    "ldom_info");
		return (-1);
	}

	if (sub_mac_address != NULL) {
		mac_address = xmlNodeGetContent(sub_mac_address);
		ASSERT(mac_address != NULL);
		if (strcmp((char *)mac_address,
		    XML_AUTO_ALLOCATED_MACADDR) != 0) {
			argv[argc++] = zsprintf("mac-addr=%s", mac_address);
		}
		xmlFree(mac_address);
	}

	if (sub_hostid != NULL) {
		hostid = xmlNodeGetContent(sub_hostid);
		if (hostid != NULL) {
			uint64_t h_id = strtoll((char *)hostid, NULL, 16);
			if (h_id != UNASSIGNED_HOSTID) {
				argv[argc++] = zsprintf("hostid=%s", hostid);
			}
		}
		xmlFree(hostid);
	}

	if (sub_failure_policy != NULL) {
		failure_policy = xmlNodeGetContent(sub_failure_policy);
		ASSERT(failure_policy != NULL);
		argv[argc++] = zsprintf("failure-policy=%s", failure_policy);
		xmlFree(failure_policy);
	}

	if (sub_master != NULL) {
		master = xmlNodeGetContent(sub_master);
		ASSERT(master != NULL);
		argv[argc++] = zsprintf("master=%s", master);
		xmlFree(master);
	}

	argv[argc++] = Xstrdup(ldp->namep);

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	cmd_num = LDM_SET;
	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}

int
xml_v3_policy(xmlNodePtr res_node, ldom_t *ldp, int action, xmlNodePtr ret_data,
    cli_conn_t *client)
{
	xmlNodePtr	item = NULL;
	xmlNodePtr	subnode = NULL;
	const int	policy_args = 13;
	void		*argv[policy_args];
	xmlChar		*content = NULL;
	bool_t		result;
	int		argc = 0, i;
	int		ret = 0;
	int		cmd_num;

	item = res_node->xmlChildrenNode;
	if (item == NULL) {
		xml_add_status(ret_data, XML_FAILURE, "Malformed XML Document:"
		    " policy tag contains no children tags");
		return (-1);
	}

	subnode = xml_find_v3_property(item, POLICY_NAME_NODE);
	if (subnode == NULL) {
		set_missing_prop_msg(ret_data, POLICY_NODE,
		    POLICY_NAME_NODE);
		return (-1);
	}
	content = xmlNodeGetContent(subnode);
	if (empty_tag_check(content, ret_data, POLICY_NAME_NODE))
		return (-1);
	argv[argc++] = zsprintf("name=%s", content);
	xmlFree(content);

	if (action != XML_REMOVE_RES) {
		subnode = xml_find_v3_property(item, POLICY_ENABLE_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (action == XML_ADD_RES &&
			    empty_tag_check(content, ret_data,
			    POLICY_ENABLE_NODE))
				return (-1);
			if (strcmp((char *)content, "on") == 0)
				argv[argc++] = zsprintf("enable=yes");
			else
				argv[argc++] = zsprintf("enable=no");
			xmlFree(content);
		}

		subnode = xml_find_v3_property(item, POLICY_PRIORITY_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (action == XML_ADD_RES &&
			    empty_tag_check(content, ret_data,
			    POLICY_PRIORITY_NODE))
				return (-1);
			argv[argc++] = zsprintf("priority=%s", content);
			xmlFree(content);
		}

		subnode = xml_find_v3_property(item, POLICY_VCPU_MIN_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (action == XML_ADD_RES &&
			    empty_tag_check(content, ret_data,
			    POLICY_VCPU_MIN_NODE))
				return (-1);
			argv[argc++] = zsprintf("vcpu-min=%s", content);
			xmlFree(content);
		}

		subnode = xml_find_v3_property(item, POLICY_VCPU_MAX_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			argv[argc++] = zsprintf("vcpu-max=%s", content);
			xmlFree(content);
		}

		subnode = xml_find_v3_property(item, POLICY_UTIL_LOWER_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (action == XML_ADD_RES &&
			    empty_tag_check(content, ret_data,
			    POLICY_UTIL_LOWER_NODE))
				return (-1);
			argv[argc++] = zsprintf("util-lower=%s", content);
			xmlFree(content);
		}

		subnode = xml_find_v3_property(item, POLICY_UTIL_UPPER_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (action == XML_ADD_RES &&
			    empty_tag_check(content, ret_data,
			    POLICY_UTIL_UPPER_NODE))
				return (-1);
			argv[argc++] = zsprintf("util-upper=%s", content);
			xmlFree(content);
		}

		subnode = xml_find_v3_property(item, POLICY_TOD_BEGIN_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (action == XML_ADD_RES &&
			    empty_tag_check(content, ret_data,
			    POLICY_TOD_BEGIN_NODE))
				return (-1);
			argv[argc++] = zsprintf("tod-begin=%s", content);
			xmlFree(content);
		}

		subnode = xml_find_v3_property(item, POLICY_TOD_END_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (action == XML_ADD_RES &&
			    empty_tag_check(content, ret_data,
			    POLICY_TOD_END_NODE))
				return (-1);
			argv[argc++] = zsprintf("tod-end=%s", content);
			xmlFree(content);
		}

		subnode = xml_find_v3_property(item, POLICY_SAMPLE_RATE_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (action == XML_ADD_RES &&
			    empty_tag_check(content, ret_data,
			    POLICY_SAMPLE_RATE_NODE))
				return (-1);
			argv[argc++] = zsprintf("sample-rate=%s", content);
			xmlFree(content);
		}

		subnode = xml_find_v3_property(item,
		    POLICY_ELASTIC_MARGIN_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (action == XML_ADD_RES &&
			    empty_tag_check(content, ret_data,
			    POLICY_ELASTIC_MARGIN_NODE))
				return (-1);
			argv[argc++] = zsprintf("elastic-margin=%s", content);
			xmlFree(content);
		}

		subnode = xml_find_v3_property(item, POLICY_ATTACK_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			argv[argc++] = zsprintf("attack=%s", content);
			xmlFree(content);
		}

		subnode = xml_find_v3_property(item, POLICY_DECAY_NODE);
		if (subnode != NULL) {
			content = xmlNodeGetContent(subnode);
			if (action == XML_ADD_RES &&
			    empty_tag_check(content, ret_data,
			    POLICY_DECAY_NODE))
				return (-1);
			argv[argc++] = zsprintf("decay=%s", content);
			xmlFree(content);
		}
	}

	argv[argc++] = Xstrdup(ldp->namep);

	DBG_BLOCK {
		for (i = 0; i < argc; i++)
			zprintf("Argv[%d] = %s\n", i, argv[i]);
	}

	switch (action) {
	case XML_ADD_RES:
		cmd_num = LDM_ADD_POLICY;
		break;
	case XML_MODIFY_RES:
		cmd_num = LDM_MODIFY_POLICY;
		break;
	case XML_REMOVE_RES:
		cmd_num = LDM_REMOVE_POLICY;
		break;
	}

	result = xml_call_cli_cmd(argc, argv, cmd_num, ret_data, client);
	free_argv(argv, argc);
	if (result == false)
		return (-1);
	else
		return (ret);
}
