/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <stdio.h>
#include <dlfcn.h>
#include <sys/socket.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <netdb.h>

#include "auth_ldom.h"
#include "cli_conn.h"
#include "database_internal.h"
#include "ssl_connection.h"
#include "xml_cntrl_common.h"
#include "xml_parse.h"
#include "xml_common.h"
#include "xmpp_common.h"
#include "xmpp_client.h"
#include "xmpp_server.h"
#include "zprintf.h"
#include "mig_common.h"


/* function prototypes */
static void	start_element(void *state, const xmlChar *name,
    const xmlChar **attrs);
static void	end_element(void *state, const xmlChar *name);
static int	xmpp_client_reader(void *xmpp_conn);
static char	*handle_cookie_resp(cli_conn_t *conn);
static xmlNodePtr create_get_cookie_XML(void);

static xmlSAXHandler xmpp_client_hdlr = {
	NULL,		/* internalSubsetSAXFunc */
	NULL,		/* isStandaloneSAXFunc */
	NULL,		/* hasInternalSubsetSAXFunc */
	NULL,		/* hasExternalSubsetSAXFunc */
	NULL,		/* resolveEntitySAXFunc */
	NULL,		/* getEntitySAXFunc */
	NULL,		/* entityDeclSAXFunc */
	NULL,		/* notationDeclSAXFunc */
	NULL,		/* attributeDeclSAXFunc */
	NULL,		/* elementDeclSAXFunc */
	NULL,		/* unparsedEntityDeclSAXFunc */
	NULL,		/* setDocumentLocatorSAXFunc */
	NULL,		/* startDocumentSAXFunc */
	NULL,		/* endDocumentSAXFunc */
	start_element,	/* startElementSAXFunc */
	end_element,	/* endElementSAXFunc */
	NULL,		/* referenceSAXFunc */
	NULL,		/* charactersSAXFunc */
	NULL,		/* ignorableWhitespaceSAXFunc */
	NULL,		/* processingInstructionSAXFunc */
	NULL,		/* commentSAXFunc */
	NULL,		/* warningSAXFunc */
	error_func,	/* errorSAXFunc */
	NULL,		/* fatalErrorSAXFunc */
	NULL,		/* getParameterEntitySAXFunc */
	NULL,		/* cdataBlockSAXFunc */
	NULL,		/* externalSubsetSAXFunc */
	0,		/* unsigned int */
	NULL,		/* void * _private */
	NULL,		/* startElementNsSAX2Func */
	NULL,		/* endElementNsSAX2Func */
	NULL		/* xmlStructuredErrorFunc */
};

char *
ldm_send_async_command(int xmpp_sock, xmlNodePtr root)
{
	cli_conn_t	*conn = malloc(sizeof (cli_conn_t));
	int		ret, count;

	priv_on();
	if (ldmd_setup_SSL() != 0) {
		zprintf("SSL setup failure\n");
		priv_off();
		return (NULL);
	}
	priv_off();

	conn->state = CONN_INIT;
	conn->fd = xmpp_sock;
	conn->buf = NULL;
	conn->buf_size = 0;
	conn->tls_started = false;
	conn->parser = xmlCreatePushParserCtxt(&xmpp_client_hdlr, conn, NULL,
	    NULL, NULL);
	if (conn->parser == NULL) {
		DBG(zprintf("Failed in xmlCreatePushParserCtxt\n"));
		return (NULL);
	}
	conn->uname = NULL;
	conn->password = NULL;
	conn->resp = NULL;
	conn->xml_command = root;
	conn->local_socket = true;
	conn->cookie = NULL;

	/* Client initiates stream to server */
	count = iowrite(conn, STREAM_RETURN, strlen(STREAM_RETURN));
	if (count <= 0)
		return (NULL);

	ret = xmpp_client_reader(conn);
	if (ret == -1)
		return (NULL);

	close(conn->fd);

	if (conn->state == LDM_INTERFACE_FINISHED)
		return (strdup(conn->buf));
	else
		return (NULL);
}

char *
send_get_cookie_command(char *target, char *uname, char *password)
{
	cli_conn_t	*conn = malloc(sizeof (cli_conn_t));
	int		count, ret;
	char		*response;

	ASSERT(uname != NULL);
	ASSERT(target != NULL);
	ASSERT(password != NULL);

	conn->fd = connect_to_host_port(target, XMPP_DEFAULT_PORT,
	    XMPP_CONNECT_TIMEOUT);
	if (conn->fd == -1) {
		ewarning("Failed to connect to host '%s'\n", target);
		return (NULL);
	}
	conn->state = CONN_INIT;
	conn->buf = NULL;
	conn->buf_size = 0;
	conn->parse_buf = NULL;
	conn->tls_started = false;
	conn->parser = xmlCreatePushParserCtxt(&xmpp_client_hdlr, conn, NULL,
	    NULL, NULL);
	conn->uname = Xstrdup(uname);
	conn->password = Xstrdup(password);
	conn->resp = NULL;
	conn->local_socket = false;
	conn->xmpp_server = false;
	conn->cookie = NULL;
	conn->xml_command = create_get_cookie_XML();

	/* Client initiates stream to server */
	count = iowrite(conn, STREAM_RETURN, strlen(STREAM_RETURN));
	if (count <= 0)
		return (NULL);

	ret = xmpp_client_reader(conn);
	if (ret == -1)
		return (NULL);
	response = handle_cookie_resp(conn);
	xmpp_close_conn(conn);
	return (response);
}

int
connect_local(int port)
{
	struct sockaddr_in myaddr;
	int error, sock;

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock == -1) {
		fprintf(stderr, "%s: %s\n", gettext("Failed to create socket"),
		    strerror(errno));
		return (-1);
	}

	myaddr.sin_family = AF_INET;
	myaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	myaddr.sin_port = htons(port);

	error = connect(sock, (struct sockaddr *)(&myaddr),
	    sizeof (struct sockaddr_in));

	if (error == -1) {
		uid_t	zeus_euid = geteuid();

		/* try simulator port if not root */
		if (zeus_euid != 0) {
			close(sock);
			sock = socket(AF_INET, SOCK_STREAM, 0);
			if (sock == -1) {
				fprintf(stderr, "%s: %s\n",
				    gettext("Failed to create socket"),
				    strerror(errno));
				return (-1);
			}

			myaddr.sin_port = htons(zeus_euid);

			error = connect(sock, (struct sockaddr *)(&myaddr),
			    sizeof (struct sockaddr_in));
		}

		if (error == -1) {
			fprintf(stderr, "%s: %s\n", gettext(
			    "Failed to connect to logical domain manager"),
			    strerror(errno));
			return (-1);
		}
	}

	return (sock);
}

/*
 * Module private functions
 */
static int
start_client_tls(cli_conn_t *conn)
{
	int		rv;

	ASSERT(client_ssl_ctx != NULL);

	conn->ssl = SSL_new_f(client_ssl_ctx);
	if (conn->ssl == NULL) {
		DBG(zprintf("Failure in SSL_new\n"));
		return (-1);
	}
	rv = SSL_set_fd_f(conn->ssl, conn->fd);
	if (rv == 0) {
		DBG(zprintf("Creating new SSL failed\n"));
		SSL_free_f(conn->ssl);
		return (-1);
	}
	rv = SSL_connect_f(conn->ssl);
	if (rv != 1) {
		DBG(zprintf("SSL connect failed\n"));
		SSL_free_f(conn->ssl);
		return (-1);
	} else {
		DBG2(zprintf("%s: SSL_connect succeeded\n", __func__));
		conn->tls_started = true;
		conn->state = TLS_STARTED;
	}

	iowrite(conn, STREAM_RETURN, strlen(STREAM_RETURN));

	return (0);
}

static void
start_client_sasl(cli_conn_t *conn)
{
	char		*in;
	int		inlen;
	char		*out = NULL;
	xmlNodePtr	auth;

	ASSERT(conn->uname != NULL);
	ASSERT(conn->password != NULL);

	inlen = strlen(conn->uname) + strlen(conn->password) + 2;
	in = Xmalloc(inlen + 1);
	sprintf(in, "%c%s%c%s", 0, conn->uname, 0, conn->password);

	if (!encode_string_base64(in, inlen, &out)) {
		DBG(zprintf("%s: base64 encode failed\n", __func__));
		goto done;
	}

	auth = xmlNewNode(NULL, AUTH_NODE);
	xmlNewProp(auth, XMLNS_PROP, NS_XMPP_SASL);
	xmlNodeAddContent(auth, (xmlChar *)out);

	send_xml_node(auth, conn);
	xmlFreeNode(auth);

done:
	Xfree(in);
	Zfree(out);
}

static void
end_element(void *state, const xmlChar *name)
{
	cli_conn_t	*conn;
	int		count;

	conn = (cli_conn_t *)state;

	if (xmlStrcmp(name, STREAM_NODE) == 0) {
		DBG(zprintf("Setting return_val to STREAM_DONE\n"));
		conn->state = STREAM_DONE;
	} else if (xmlStrcmp(name, STARTTLS_NODE) == 0) {
		DBG2(zprintf("END tag name = %s\n", (char *)name));
		count = iowrite(conn, START_TLS, strlen(START_TLS));
		if (count == -1) {
			DBG(zprintf("failed to send tlsstart\n"));
			return;
		}
	} else if (xmlStrcmp(name, SUCCESS_NODE) == 0) {
		conn->state = SASL_STARTED;
		iowrite(conn, STREAM_RETURN, strlen(STREAM_RETURN));
	} else if (xmlStrcmp(name, PROCEED_NODE) == 0) {
		DBG2(zprintf("END tag name = %s\n", (char *)name));
		start_client_tls(conn);
	} else if (xmlStrcmp(name, FEATURE_NODE) == 0) {
		if (conn->state == TLS_STARTED)
			if (!conn->local_socket) {
				start_client_sasl(conn);
			} else {
				send_xml_node(conn->xml_command, conn);
			}
		if (conn->state == SASL_STARTED)
			send_xml_node(conn->xml_command, conn);
	} else if (xmlStrcmp(name, XML_LDM_INTERFACE) == 0) {
		conn->state = LDM_INTERFACE_FINISHED;
	} else {
		DBG2(zprintf("END tag name = %s\n", (char *)name));
	}
}

/*ARGSUSED2*/
static void
start_element(void *state, const xmlChar *name, const xmlChar **attrs)
{
	cli_conn_t	*conn;

	conn = (cli_conn_t *)state;

	if (xmlStrcmp(name, STREAM_NODE) == 0) {
		DBG2(zprintf("tag name = %s\n", (char *)name));
	} else if (xmlStrcmp(name, XML_LDM_INTERFACE) == 0) {
		conn->state = LDM_INTERFACE_STARTED;
	} else if (xmlStrcmp(name, FAILURE_NODE) == 0) {
		conn->state = XMPP_FAILURE;
	} else {
		DBG2(zprintf("START tag name = %s\n", (char *)name));
	}
}

static int
xmpp_client_reader(void *xmpp_conn)
{
	cli_conn_t	*conn = (cli_conn_t *)xmpp_conn;
	char		*buf;
	int		count, rv, current_position = 0;

	buf = malloc(READ_BUF_SIZE);

	while (conn->state != STREAM_DONE) {
		memset(buf, 0, READ_BUF_SIZE);
		count = ioread(conn, buf);
		if ((count == -1) || (count == 0)) {
			xmpp_close_conn(conn);
			return (-1);
		}

		conn->parse_buf = buf;

		rv = xmlParseChunk(conn->parser, buf, count, 0);
		if (rv != 0) {
			DBG(zprintf("Parse Error = %d\n", rv));
			return (-1);
		}

		DBG5(zprintf("%s: conn->state=0x%x\n",
		    __func__, (int)conn->state));

		switch (conn->state) {
		case TLS_STARTED:
		case SASL_STARTED:
			xmlCtxtResetPush(conn->parser, NULL, NULL, NULL,
			    NULL);
			break;
		case LDM_INTERFACE_STARTED:
			conn->buf_size += count;
			conn->buf = realloc(conn->buf, conn->buf_size + 1);
			memset((conn->buf + current_position), 0, count + 1);
			memcpy((conn->buf + current_position), buf, count);
			current_position += count;
			break;
		case LDM_INTERFACE_FINISHED:
			conn->buf_size += count;
			conn->buf = realloc(conn->buf, conn->buf_size + 1);
			memset((conn->buf + current_position), 0, count + 1);
			memcpy((conn->buf + current_position), buf, count);
			return (0);
		}
	}
	return (0);
}

static xmlNodePtr
create_get_cookie_XML(void)
{
	xmlNodePtr	root, cmd, data;

	root = xmlNewNode(NULL, XML_LDM_INTERFACE);
	xmlNewProp(root, VERSION_ATTR, LDOM_INTERFACE_VERSION);
	cmd = xmlNewChild(root, NULL, XML_CMD, NULL);
	xmlNewChild(cmd, NULL, XML_ACTION, XML_GET_COOKIE);
	data = xmlNewChild(cmd, NULL, XML_DATA, NULL);
	xmlNewProp(data, VERSION_ATTR, LDOM_V3_DATA_VERSION);

	return (root);
}

static char *
handle_cookie_resp(cli_conn_t *conn)
{
	xmlDocPtr	xml_output;
	xmlNodePtr	root, resp, status, cmd, data, cookie;
	char		*status_str, *cookie_str, *return_str;

	ASSERT(conn != NULL);

	if (conn->buf == NULL || conn->buf_size == 0)
		return (NULL);

	xml_output = xmlParseMemory((const char *)conn->buf, conn->buf_size);

	root = xmlDocGetRootElement(xml_output);
	cmd = xml_find_subnode(root, XML_CMD);
	data = xml_find_subnode(cmd, XML_DATA);
	resp = xml_find_subnode(cmd, XML_RESPONSE);
	status = xml_find_subnode(resp, XML_STATUS);
	status_str = (char *)xmlNodeGetContent(status);

	if (strcmp(status_str, XML_FAILURE) == 0) {
		return_str = NULL;
	} else {
		cookie = xml_find_subnode(data, XML_COOKIE);
		cookie_str = (char *)xmlNodeGetContent(cookie);
		return_str = strdup(cookie_str);
		xmlFree(cookie_str);
	}

	xmlFree(status_str);
	xmlFreeDoc(xml_output);
	return (return_str);
}
