/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <sys/types.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <libxml/tree.h>
#include <libxml/parser.h>
#include <sasl/saslutil.h>
#include <unistd.h>
#include <port.h>

#include "allocate.h"
#include "cli_conn.h"
#include "ssl_connection.h"
#include "xmpp_common.h"
#include "xml_event.h"
#include "zprintf.h"

mutex_t	xmpp_thr_mutex;
int	xmpp_thr_count;
int	evport;

void
xmpp_close_conn(cli_conn_t *conn)
{
	cli_conn_t	*sock_conn;
	bool_t		xmpp_server;

	if (conn == NULL) {
		DBG(zprintf("%s: error, conn == NULL\n", __func__));
		return;
	}

	xmpp_server = conn->xmpp_server;
	if (xmpp_server) {
		event_rm_all_event(NULL, NULL, conn, 0);
	}

	(void) close(conn->fd);

	if (conn->parse_buf != NULL)
		Xfree(conn->parse_buf);

	if (conn->buf != NULL)
		Xfree(conn->buf);

	if (conn->parser != NULL)
		xmlFreeParserCtxt(conn->parser);

	if (conn->tls_started)
		SSL_free_f(conn->ssl);

	if (conn->resp != NULL) {
		Xfree(conn->resp->bufp);
		Xfree(conn->resp);
	}

	if (conn->uname != NULL)
		Xfree(conn->uname);

	if (conn->password != NULL)
		Xfree(conn->password);

	if (conn->cookie != NULL)
		Xfree(conn->cookie);

	if (conn->xml_command != NULL)
		xmlFreeNode(conn->xml_command);

	if (xmpp_server) {
		(void) mutex_lock(&xmpp_thr_mutex);
		if (xmpp_thr_count == MAX_THREADS) {
			DBG(zprintf("MAX_THREADS (%d) reached\n", MAX_THREADS));
			sock_conn = (cli_conn_t *)conn->sock_conn;
			if (port_associate(evport, PORT_SOURCE_FD,
			    sock_conn->fd, POLLIN, sock_conn) < 0)
				efatal("port_associate");
		}
		xmpp_thr_count--;
		(void) mutex_unlock(&xmpp_thr_mutex);
		Xfree(conn);
		thr_exit(0);
		/* NOTREACHED */
	}

	Xfree(conn);
}


int
iowrite(cli_conn_t *conn, char *buf, int size)
{
	int	count;

	if (conn == NULL)
		return (-1);

	if (buf == NULL)
		return (0);

	if (conn->tls_started) {
		count = SSL_write_f(conn->ssl, (char *)buf, size);
	} else {
		count = send(conn->fd, (char *)buf, size, 0);
		if (count == -1) {
			DBG(zprintf("send failed. handle_msg: error = %d: %s\n",
			    errno, strerror(errno)));
		}
	}

	return (count);
}

int
ioread(void *state, char *buf)
{
	int		count;
	cli_conn_t	*conn = (cli_conn_t *)state;
	pollfd_t	pollfd;

	if (conn == NULL)
		return (-1);

	if (buf == NULL)
		return (0);

	(void) memset(buf, 0, READ_BUF_SIZE);

	if (conn->tls_started == true) {
		ASSERT(conn->ssl != NULL);
		count = SSL_read_f(conn->ssl, buf, READ_BUF_SIZE);
	} else {
		if (conn->local_socket == false) {
			pollfd.fd = conn->fd;
			pollfd.events = POLLIN;
			count = poll(&pollfd, 1, XMPP_TIMEOUT);
			if (count == 0) {
				DBG(zprintf("%s:poll() timed out\n", __func__));
				/*
				 * If buf has already been filled it was set to
				 * conn->parse_buf which gets free'd in
				 * xmpp_close_conn(). BUT to be safe free it
				 * here and set to NULL so it won't be free'd
				 * again.
				 */
				if (conn->parse_buf == NULL) {
					Xfree(buf);
					buf = NULL;
				}
				return (0);
			}
		}
		count = read(conn->fd, buf, READ_BUF_SIZE);
	}

	DBG(zprintf("%s:BUFFER size = %d\n BUFFER: %s\n", __func__,
	    count, buf));

	return (count);
}

void
send_xml_node(xmlNodePtr node, cli_conn_t *conn)
{
	xmlBufferPtr	xmlbuf;

	xmlbuf = xmlBufferCreateSize(READ_BUF_SIZE);
	xmlbuf->alloc = XML_BUFFER_ALLOC_EXACT;
	xmlNodeDump(xmlbuf, NULL, node, 0, 0);

	DBG5(zprintf("OUTPUT: %s\n", (char *)xmlbuf->content));
	DBG5(zprintf("strlen = %d\n", strlen((char *)xmlbuf->content)));
	(void) iowrite(conn, (char *)xmlbuf->content, xmlbuf->use);

	xmlBufferFree(xmlbuf);
}

/* ARGSUSED1 */
void
error_func(void *state, const char *msg, ...)
{
	cli_conn_t	*conn;

	DBG(zprintf("IN ERROR FUNC\n"));
	conn = (cli_conn_t *)state;
	DBG(zprintf("Buffer: %s\n", conn->parse_buf));
}

void
xmpp_set_conn_state(cli_conn_t *conn, int state)
{
	DBG(print_state_change(conn, state));
	conn->state = state;
}

/*
 * Convert the string passed in as an argument into a base64, NULL
 * terminated string. On success, memory is allocated for the output
 * string which must be deallocated by the caller.
 */
bool_t
encode_string_base64(const char *in, uint_t inlen, char **out)
{
	char	*buf;
	int	buflen;
	uint_t	outlen;
	int	rv;

	ASSERT((in != NULL) && (out != NULL));

	/*
	 * Calculate the required size of the output buffer. With
	 * base64 encoding, every three bytes of the input becomes
	 * four bytes in the output. If the input length is not a
	 * multiple of three, the output string is padded. Adding
	 * two to the input length when calculating the number of
	 * three byte groups in the input takes this into account.
	 * The output buffer is also padded for NULL termination.
	 */
	buflen = (((inlen + 2) / 3) * 4) + 1;
	buf = Xmalloc(buflen);

	rv = sasl_encode64(in, inlen, buf, buflen, &outlen);
	if (rv == SASL_BUFOVER) {
		DBG(zprintf("%s: resizing output buffer to %d bytes\n",
		    __func__, outlen + 1));
		buflen = outlen + 1;
		buf = Xrealloc(buf, buflen);
		rv = sasl_encode64(in, inlen, buf, buflen, &outlen);
	}

	ASSERT(rv == SASL_OK);

	if (rv != SASL_OK) {
		Xfree(buf);
		return (false);
	}

	*out = buf;
	return (true);
}

#if ZDBG
char *
get_state_string(int state)
{
	switch (state) {
	case SOCKET_WAITING:
		return (Xstrdup("SOCKET_WAITING"));
	case STREAM_TAG_ONE_SENT:
		return (Xstrdup("STREAM_TAG_ONE_SENT"));
	case STREAM_TLS_SENT:
		return (Xstrdup("STREAM_TLS_SENT"));
	case TLS_START:
		return (Xstrdup("TLS_START"));
	case STREAM_TAG_TWO_SENT:
		return (Xstrdup("STREAM_TAG_TWO_SENT"));
	case STREAM_MECH_SENT:
		return (Xstrdup("STREAM_MECH_SENT"));
	case CHALLENGE_ONE_SENT:
		return (Xstrdup("CHALLENGE_ONE_SENT"));
	case CHALLENGE_TWO_SENT:
		return (Xstrdup("CHALLENGE_TWO_SENT"));
	case SUCCESS_SENT:
		return (Xstrdup("SUCCESS_SENT"));
	case STREAM_TAG_THREE_SENT:
		return (Xstrdup("STREAM_TAG_THREE_SENT"));
	case CONNECTION_COMPLETE:
		return (Xstrdup("CONNECTION_COMPLETE"));
	case LDM_INTERFACE_STARTED:
		return (Xstrdup("LDM_INTERFACE_STARTED"));
	case STREAM_DONE:
		return (Xstrdup("STREAM_DONE"));
	}

	return (Xstrdup("UNKNOWN"));
}

void
print_state_change(cli_conn_t *conn, int new_state)
{
	char *old_state_str, *new_state_str;

	old_state_str = get_state_string(conn->state);
	new_state_str = get_state_string(new_state);

	zprintf("Changing state from %s to %s\n", old_state_str,
	    new_state_str);

	Xfree(old_state_str);
	Xfree(new_state_str);
}
#endif	/* ZDBG */
