/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <fcntl.h>
#include <unistd.h>
#include <netdb.h>
#include <ctype.h>
#include <port.h>
#include <string.h>
#include <errno.h>

#include <sasl/sasl.h>
#include <sasl/saslutil.h>
#include <security/pam_appl.h>

#include "allocate.h"
#include "basics.h"
#include "option.h"
#include "sequencer.h"
#include "fatal.h"
#include "cli_conn.h"
#include "auth_ldom.h"
#include "zprintf.h"
#include "ssl_connection.h"
#include "xmpp_server.h"
#include "xmpp_common.h"
#include "xml_cntrl_common.h"
#include "xml_v3_defs.h"
#include "xml_parse.h"
#include "xml_common.h"
#include "ldomconfigure.h"
#include "database_internal.h"

/* defined in xmpp_common.c */
extern int xmpp_thr_count;
extern mutex_t xmpp_thr_mutex;
extern int evport;

/*
 * The header size number is the size of the header tacked onto the front
 * of the XML document for the OLD communication mechanism.  For XMPP we don't
 * need this and thus add this number to the buffer sent to us from
 * zeus_ldom_xml_command() in xml_parse.c
 */
#define	HEADER_SIZE 22


static int sasl_log_func(void *context, int level, const char *msg);
static int check_password(sasl_conn_t *conn, void *context, const char *user,
    const char *pass, uint_t passlen, struct propctx *ctx);
static int canon_user(sasl_conn_t *conn, void *context, const char *user,
    unsigned ulen, unsigned flags, const char *user_realm, char *out_user,
    unsigned out_umax, unsigned *out_ulen);

static sasl_callback_t	srvcallbacks[] =
{
	{ SASL_CB_CANON_USER, &canon_user, NULL },
	{ SASL_CB_SERVER_USERDB_CHECKPASS, &check_password, NULL },
	{ SASL_CB_LOG, &sasl_log_func, NULL },
	{ SASL_CB_LIST_END, NULL, NULL }
};

static void start_element(void *, const xmlChar *, const xmlChar **attrs);
static void end_element(void *state, const xmlChar *name);
static void *new_xmpp_client(cli_conn_t *conn);

static xmlSAXHandler my_handler = {
	NULL,		/* internalSubsetSAXFunc */
	NULL,		/* isStandaloneSAXFunc */
	NULL,		/* hasInternalSubsetSAXFunc */
	NULL,		/* hasExternalSubsetSAXFunc */
	NULL,		/* resolveEntitySAXFunc */
	NULL,		/* getEntitySAXFunc */
	NULL,		/* entityDeclSAXFunc */
	NULL,		/* notationDeclSAXFunc */
	NULL,		/* attributeDeclSAXFunc */
	NULL,		/* elementDeclSAXFunc */
	NULL,		/* unparsedEntityDeclSAXFunc */
	NULL,		/* setDocumentLocatorSAXFunc */
	NULL,		/* startDocumentSAXFunc */
	NULL,		/* endDocumentSAXFunc */
	start_element,	/* startElementSAXFunc */
	end_element,	/* endElementSAXFunc */
	NULL,		/* referenceSAXFunc */
	NULL,		/* charactersSAXFunc */
	NULL,		/* ignorableWhitespaceSAXFunc */
	NULL,		/* processingInstructionSAXFunc */
	NULL,		/* commentSAXFunc */
	NULL,		/* warningSAXFunc */
	error_func,	/* errorSAXFunc */
	NULL,		/* fatalErrorSAXFunc */
	NULL,		/* getParameterEntitySAXFunc */
	NULL,		/* cdataBlockSAXFunc */
	NULL,		/* externalSubsetSAXFunc */
	0,		/* unsigned int */
	NULL,		/* void * _private */
	NULL,		/* startElementNsSAX2Func */
	NULL,		/* endElementNsSAX2Func */
	NULL		/* xmlStructuredErrorFunc */
};


void
xmpp_write_buffer(cli_conn_t *conn, bool_t event)
{
	char	*buf;
	int	buf_size;

	if (event) {
		buf = conn->event_buf;
		buf_size = conn->event_buf_size;
	} else {
		buf = conn->buf;
		buf_size = conn->buf_size;
	}
	ASSERT(buf != NULL);

	/*
	 * If this is a long running cmd send_resp should be false. Do not send
	 * the buffer back to the client.  It will be sent later from the
	 * long running command.
	 */
	if (conn->send_resp == true) {
		DBG(zprintf("write buffer send:\n%s\n", buf + HEADER_SIZE));
		iowrite(conn, buf + HEADER_SIZE, buf_size-HEADER_SIZE);
	}
	memset(buf, 0, buf_size);
	Xfree(buf);
	if (event) {
		conn->event_buf = NULL;
		conn->event_buf_size = 0;
	} else {
		conn->buf = NULL;
		conn->buf_size = 0;
	}
	if (conn->send_resp == true) {
		conn->seq_finished = true;
		cond_signal(&conn->cvp);
	}
}

int
do_sasl_decode64(const char *in, unsigned inlen, char **out, unsigned *outlen)
{
	unsigned len;
	int r;

	/*
	 * Upper bound on length of base64 decoding plus 1 for the NUL
	 * terminator which this routine adds.
	 */
	len = ((inlen / 4) + 1) * 3 + 1;

	*out = Xmalloc(len);
	if (*out == NULL)
		return (-1);

	r = sasl_decode64(in, inlen, *out, len, outlen);
	ASSERT(r == SASL_OK);

	return (0);
}

/*
 * Main loop to handle IO with XMPP clients.
 * This is run in its own thread.
 */
/* ARGSUSED0 */
void *
xmpp_loop(void *unused)
{
	struct sockaddr_in myaddr;
	cli_conn_t *acc_conn, *conn;
	port_event_t ev;
	int sock;
	int on = 1;

	DBG7(memset(&ev, 0, sizeof (ev)));

	xmpp_thr_count = 0;
	mutex_init(&xmpp_thr_mutex, USYNC_PROCESS, NULL);

	if ((evport = port_create()) < 0) {
		SSL_CTX_free_f(server_ssl_ctx);
		warning("XMPP port_create failed. XMPP server not available");
		return (NULL);
	}

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		SSL_CTX_free_f(server_ssl_ctx);
		warning("XMPP socket creation failed. XMPP server not "
		    "available");
		return (NULL);
	}

	if (setsockopt(sock, SOL_SOCKET,
	    SO_REUSEADDR, (char *)&on, sizeof (on)) < 0) {
		SSL_CTX_free_f(server_ssl_ctx);
		warning("setsockopt failed. XMPP server not available");
		return (NULL);
	}

	myaddr.sin_family = AF_INET;
	myaddr.sin_port = htons(option.xmpp_port);
	if (option.xmpp_enabled) {
		myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	} else {
		myaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	}

	if (bind(sock, (struct sockaddr *)&myaddr, sizeof (myaddr)) < 0) {
		SSL_CTX_free_f(server_ssl_ctx);
		warning("XMPP bind failed. XMPP server not available");
		return (NULL);
	}

	(void) listen(sock, 3);

	DBG(zprintf("Listening on port %d\n", option.xmpp_port));

	/* create a connection which handles accepting new clients */
	acc_conn = Xmalloc(sizeof (cli_conn_t));
	acc_conn->fd = sock;
	acc_conn->callback = new_xmpp_client;

	if (port_associate(evport, PORT_SOURCE_FD, sock, POLLIN, acc_conn) < 0)
		efatal("port_associate");

	/* loop - waiting for and handling IO events */
	while (1) {
		if (port_get(evport, &ev, NULL) < 0)
			efatal("port_get");

		if (ev.portev_source == PORT_SOURCE_FD) {
			conn = ev.portev_user;
			(void) conn->callback(conn);
		} else {
			fatal("Unexpected port event source");
		}
	}

	/*NOTREACHED*/
}

/*
 * Module private functions
 */

static void
send_failure_node(cli_conn_t *conn, xmlChar *namespace, char *extra_info)
{
	xmlNodePtr	failure;

	failure = xmlNewNode(NULL, FAILURE_NODE);
	if (namespace != NULL) {
		xmlNewProp(failure, XMLNS_PROP, namespace);
	}
	if (extra_info != NULL) {
		xmlNewChild(failure, NULL, (xmlChar *)extra_info, NULL);
	}

	send_xml_node(failure, conn);
	xmlFreeNode(failure);
	iowrite(conn, END_STREAM_NODE, strlen(END_STREAM_NODE));
	xmpp_close_conn(conn);
}

static void
printesc(char *s, int len)
{
	char	c;

	while (len > 0) {
		c = *s++;
		len --;

		if (isprint(c)) {
			putchar(c);
		} else {
			char exp[3];
			for (int i = 2; i >= 0; i--) {
				exp[i] = c % 8 + '0';
				c /= 8;
			}
			putchar('\\');
			putchar(exp[0]);
			putchar(exp[1]);
			putchar(exp[2]);
		}
	}
}

/* ARGSUSED0 */
static int
sasl_log_func(void *context, int level, const char *msg)
{
	int	rv = 0;

	DBG(zprintf("LOG MESSAGE: %s\n", msg));

	return (rv);
}

/* ARGSUSED0 */
static int
canon_user(sasl_conn_t *conn, void *context, const char *user,
    unsigned ulen, unsigned flags, const char *user_realm, char *out_user,
    unsigned out_umax, unsigned *out_ulen)
{
	char *p;

	if ((p = strchr(user, '@')) != NULL) {
		if (p - user + 1 < out_umax)
			out_umax = p - user + 1;
	}

	strlcpy(out_user, user, out_umax);
	*out_ulen = strlen(out_user);

	DBG(zprintf("%s: %s -> %s\n", __func__, user, out_user));

	return (SASL_OK);
}

/* ARGSUSED0 */
static int
check_password(sasl_conn_t *conn, void *context, const char *user,
    const char *pass, uint_t passlen, struct propctx *ctx)
{
	int		rv;
	pam_handle_t	*pamh = NULL;
	cli_conn_t	*xmpp_conn;

	xmpp_conn = (cli_conn_t *)context;

	DBG(zprintf("USER: %s\n", user));
	DBG(zprintf("PASSWORD: %s\n", pass));

	rv = pam_start("xmpp_server", user, NULL, &pamh);
	if (rv != PAM_SUCCESS) {
		DBG(zprintf("PAM START FAILED: %s\n", pam_strerror(pamh, rv)));
		if (pamh != NULL)
			pam_end(pamh, 0);
		return (SASL_BADAUTH);
	}
	rv = pam_set_item(pamh, PAM_AUTHTOK, pass);
	if (rv != PAM_SUCCESS) {
		DBG(zprintf("PAM SET ITEM FAILED: %s\n",
		    pam_strerror(pamh, rv)));
		if (pamh != NULL)
			pam_end(pamh, 0);
		return (SASL_BADAUTH);
	}

	rv = pam_authenticate(pamh, PAM_SILENT);
	if (rv != PAM_SUCCESS) {
		DBG(zprintf("PAM AUTHENTICATE FAILED: %s\n",
		    pam_strerror(pamh, rv)));
		if (pamh != NULL)
			pam_end(pamh, 0);
		return (SASL_BADAUTH);
	}

	xmpp_conn->uname = Xstrdup((char *)user);

	if (pamh != NULL)
		pam_end(pamh, 0);

	return (SASL_OK);
}

static void
init_stream(cli_conn_t *conn)
{
	int		count, str_len;
	char		*return_str;
	xmlNodePtr	feature, mechs, starttls;

	return_str = STREAM_RETURN;
	str_len = strlen(return_str);
	/* Server responds by sending stream tag to client */
	count = iowrite(conn, return_str, str_len);
	if (count == -1) {
		DBG(zprintf("init_stream: send failed. handle_msg: error = "
		    "%d: %s\n", errno, strerror(errno)));
		xmpp_close_conn(conn);
		/* NOTREACHED - thread killed */
	} else {
		if (conn->state == SOCKET_WAITING) {
			xmpp_set_conn_state(conn, STREAM_TAG_ONE_SENT);
		} else if (conn->state == TLS_START) {
			xmpp_set_conn_state(conn, STREAM_TAG_TWO_SENT);
		} else if (conn->state == SUCCESS_SENT) {
			xmpp_set_conn_state(conn, STREAM_TAG_THREE_SENT);
		} else {
			send_failure_node(conn, NULL, NULL);
			/* NOTREACHED - thread killed */
		}
	}

	feature = xmlNewNode(NULL, FEATURE_NODE);
	if (conn->state == STREAM_TAG_ONE_SENT) {
		/*
		 * The first stream:stream tag was sent to the client. Send
		 * now the TLS feature tag.
		 */
		starttls = xmlNewChild(feature, NULL, STARTTLS_NODE, NULL);
		xmlNewProp(starttls, XMLNS_PROP, (xmlChar *)NS_XMPP_TLS);
		xmlNewChild(starttls, NULL, REQUIRED_NODE, NULL);

		mechs = xmlNewChild(feature, NULL, MECHS_NODE, NULL);
		xmlNewProp(mechs, XMLNS_PROP, NS_XMPP_SASL);
		xmlNewChild(mechs, NULL, MECH_NODE, (xmlChar *)"PLAIN");
		xmpp_set_conn_state(conn, STREAM_TLS_SENT);
	} else if (conn->state == STREAM_TAG_TWO_SENT) {
		/*
		 * The second stream:stream tag was sent to the client. If not
		 * a local socket send the SASL feature tag.
		 */
		if (conn->local_socket == false) {
			mechs = xmlNewChild(feature, NULL, MECHS_NODE, NULL);
			xmlNewProp(mechs, XMLNS_PROP, NS_XMPP_SASL);
			xmlNewChild(mechs, NULL, MECH_NODE,
			    (xmlChar *)"PLAIN");
			xmpp_set_conn_state(conn, STREAM_MECH_SENT);
		} else {
			xmpp_set_conn_state(conn, CONNECTION_COMPLETE);
		}
	} else if (conn->state == STREAM_TAG_THREE_SENT) {
		xmpp_set_conn_state(conn, CONNECTION_COMPLETE);
	}
	send_xml_node(feature, conn);
	xmlFreeNode(feature);
}

static void
start_tls(cli_conn_t *conn)
{
	int		rv, ssl_err;
	xmlNodePtr	proceed_node;
	unsigned long	err_err;

	ASSERT(server_ssl_ctx != NULL);

	conn->ssl = SSL_new_f(server_ssl_ctx);
	rv = SSL_set_fd_f(conn->ssl, conn->fd);
	if (rv == 0) {
		DBG(zprintf("SSL_set_fd failed\n"));
		send_failure_node(conn, NS_XMPP_TLS, NULL);
		/* NOTREACHED - thread killed */
	}

	proceed_node = xmlNewNode(NULL, PROCEED_NODE);
	xmlNewProp(proceed_node, XMLNS_PROP, (xmlChar *)NS_XMPP_TLS);
	send_xml_node(proceed_node, conn);
	xmlFreeNode(proceed_node);
	xmpp_set_conn_state(conn, TLS_START);

	rv = SSL_accept_f(conn->ssl);
	if (rv == 1) {
		conn->tls_started = true;
		DBG(zprintf("TLS handshake PASSED\n"));
	} else {
		ssl_err = SSL_get_error_f(conn->ssl, rv);
		DBG(zprintf("ACCEPT FAILED ssl_err = %d\n", ssl_err));
		err_err = ERR_get_error_f();
		DBG(zprintf("%s\n", ERR_error_string_f(err_err, NULL)));
		send_failure_node(conn, NS_XMPP_TLS, NULL);
		/* NOTREACHED - thread killed */
	}
}

static void
auth_stream(cli_conn_t *conn)
{
	int		rv;
	char		hostname[MAXHOSTNAMELEN];
	sasl_conn_t	*sconn = NULL;
	xmlDocPtr	auth_input;
	xmlNodePtr	auth_node, success;
	char		*auth, *out;
	uint_t		outlen = 0;
	uint_t		serveroutlen;
	const char	*serverout = NULL;

	srvcallbacks[1].context = (void *)conn;

	rv = sasl_server_init(srvcallbacks, "xmpp_server");
	if (rv != SASL_OK) {
		DBG(zprintf("sasl_server_init failure: %s\n",
		    sasl_errstring(rv, NULL, NULL)));
	}
	gethostname(hostname, MAXHOSTNAMELEN);
	rv = sasl_server_new("XMPP", hostname, NULL, NULL, NULL,
	    srvcallbacks, 0, &sconn);
	if (rv != SASL_OK) {
		DBG(zprintf("sasl_server_new failure: %s\n",
		    sasl_errstring(rv, NULL, NULL)));
		send_failure_node(conn, NS_XMPP_SASL, "incorrect-encoding");
		/* NOTREACHED - thread killed */
	}

	auth_input = xmlParseMemory(conn->parse_buf, READ_BUF_SIZE);
	if (auth_input == NULL) {
		DBG(zprintf("Failed to parse auth\n"));
		send_failure_node(conn, NS_XMPP_SASL, "incorrect-encoding");
		/* NOTREACHED - thread killed */
	}

	auth_node = xmlDocGetRootElement(auth_input);
	if (auth_node == NULL) {
		DBG(zprintf("Failed to get root node\n"));
		send_failure_node(conn, NS_XMPP_SASL, "incorrect-encoding");
		/* NOTREACHED - thread killed */
	}

	auth = (char *)xmlNodeGetContent(auth_node);
	rv = do_sasl_decode64(auth, strlen(auth), &out, &outlen);
	xmlFreeDoc(auth_input);
	xmlFree(auth);
	if (rv != 0) {
		DBG(zprintf("do_sasl_decode64 failure\n"));
		send_failure_node(conn, NS_XMPP_SASL, "incorrect-encoding");
		/* NOTREACHED - thread killed */
	}

	DBG(zprintf("DECODED RESP: "));
	DBG(printesc(out, outlen));
	DBG(zprintf("\n"));
	rv = sasl_server_start(sconn, "PLAIN", out, outlen, &serverout,
	    &serveroutlen);

	if (serverout != NULL)
		Xfree((void *)serverout);
	if (sconn != NULL)
		sasl_dispose(&sconn);
	sasl_done();
	Xfree(out);

	if ((rv != SASL_OK) && (rv != SASL_CONTINUE)) {
		DBG(zprintf("FAILED SERVER_START: %s\n",
		    sasl_errstring(rv, NULL, NULL)));
		send_failure_node(conn, NS_XMPP_SASL,
		    "temporary-auth-failure");
		/* NOTREACHED - thread killed */
	} else {
		success = xmlNewNode(NULL, SUCCESS_NODE);
		xmlNewProp(success, XMLNS_PROP, NS_XMPP_SASL);
		send_xml_node(success, conn);
		xmlFreeNode(success);
		xmpp_set_conn_state(conn, SUCCESS_SENT);
	}
}

static void
end_element(void *state, const xmlChar *name)
{
	cli_conn_t	*conn;

	conn = (cli_conn_t *)state;

	if (xmlStrcmp(name, STREAM_NODE) == 0) {
		DBG(zprintf("STREAM_DONE closing the connection\n"));
		xmpp_close_conn(conn);
		/* NOTREACHED - thread killed */
	} else if (xmlStrcmp(name, AUTH_NODE) == 0) {
		if (conn->state != STREAM_MECH_SENT) {
			send_failure_node(conn, NS_XMPP_SASL, NULL);
			/* NOTREACHED - thread killed */
		}
		DBG2(zprintf("END tag name = %s\n", (char *)name));
		auth_stream(conn);
	} else if (xmlStrcmp(name, STARTTLS_NODE) == 0) {
		if (conn->state != STREAM_TLS_SENT) {
			send_failure_node(conn, NS_XMPP_TLS, NULL);
			/* NOTREACHED - thread killed */
		}
		DBG2(zprintf("END tag name = %s\n", (char *)name));
		start_tls(conn);
	} else if (xmlStrcmp(name, XML_LDM_INTERFACE) == 0) {
		char	*next_buf;

		conn->buf_size += conn->parse_len;
		conn->buf = Xrealloc(conn->buf, conn->buf_size + 1);
		memset((conn->buf + conn->position), 0, conn->parse_len + 1);
		memcpy((conn->buf + conn->position), conn->parse_buf,
		    conn->parse_len);
		DBG(zprintf("BEFORE STRSTR check:\n %s\n", conn->buf));
		next_buf = strstr(conn->buf, "/LDM_interface>");
		next_buf = strchr(next_buf, '<');
		if ((next_buf == NULL) ||
		    (next_buf > (conn->buf + conn->buf_size))) {
			DBG(zprintf("SINGLE LDM_INTERFACE OUTPUT:\n%s\n",
			    conn->buf));
			conn->seq_finished = false;
			seq_enqueue_cmd(conn, SEQ_XML_CMD);
			mutex_lock(&conn->mp);
			while (conn->seq_finished == false) {
				cond_wait(&conn->cvp, &conn->mp);
			}
			mutex_unlock(&conn->mp);
			conn->position = 0;
			xmpp_set_conn_state(conn, CONNECTION_COMPLETE);
		} else {
			uint64_t	extra_size, iface_size;

			memset(conn->parse_buf, 0, READ_BUF_SIZE);
			iface_size = next_buf - conn->buf;
			extra_size = conn->buf_size - iface_size;
			conn->parse_len = extra_size;
			memcpy(conn->parse_buf, next_buf, extra_size);
			memset(next_buf, 0, extra_size);
			DBG(zprintf("MULTI LDM_INTERFACE OUTPUT:\n%s\n",
			    conn->buf));
			conn->seq_finished = false;
			seq_enqueue_cmd(conn, SEQ_XML_CMD);
			mutex_lock(&conn->mp);
			while (conn->seq_finished == false) {
				cond_wait(&conn->cvp, &conn->mp);
			}
			mutex_unlock(&conn->mp);
			conn->position = 0;
			/*
			 * We will be setting the extra data into the
			 * conn->buf.  We will be parsing it still though
			 * so don't want to set the state to
			 * LDM_INTERFACE_START or that will mean that the data
			 * will be set in conn->buf again.  We have created
			 * LDM_INTERFACE_RESET for this condition.
			 */
			xmpp_set_conn_state(conn, LDM_INTERFACE_RESET);
			/*
			 * check to see if we have a complete interface still
			 * in the parse buffer. If we do NOT then we have to
			 * save the data we already have in the conn->buf
			 * pointer. If we DO then we do not have to save it
			 * as the parser will catch the whole interface.
			 */
			next_buf = strstr(conn->parse_buf, "/LDM_interface>");
			if (next_buf == NULL) {
				conn->buf_size += extra_size;
				conn->buf = Xrealloc(conn->buf,
				    conn->buf_size);
				memcpy((conn->buf + conn->position),
				    conn->parse_buf, extra_size);
				conn->position += extra_size;
				DBG(zprintf("NEXT INTERFACE START:\n%s\n",
				    conn->buf));
			}
		}
	} else {
		DBG2(zprintf("END tag name = %s\n", (char *)name));
	}
}

/* ARGSUSED2 */
static void
start_element(void *state, const xmlChar *name, const xmlChar **attrs)
{
	cli_conn_t	*conn;

	conn = (cli_conn_t *)state;

	if (xmlStrcmp(name, STREAM_NODE) == 0) {
		DBG2(zprintf("tag name = %s\n", (char *)name));
		init_stream(conn);
	} else if (xmlStrcmp(name, XML_LDM_INTERFACE) == 0) {
		/*
		 * We don't want to save the parsed data to the conn->buf
		 * again if it is the next LDM_interface before completing
		 * the 1st one. Simply return without changing the state.
		 */
		if (conn->state == LDM_INTERFACE_RESET)
			return;
		if (conn->state != CONNECTION_COMPLETE) {
			send_failure_node(conn, NULL, NULL);
			/* NOTREACHED - thread killed */;
		}
		xmpp_set_conn_state(conn, LDM_INTERFACE_STARTED);
	} else {
		DBG2(zprintf("START tag name = %s\n", (char *)name));
	}
}

static void *
xmpp_reader(void *xmpp_conn)
{
	cli_conn_t	*conn = (cli_conn_t *)xmpp_conn;
	int		rv, count;
	char		*buf;

	buf = Xmalloc(READ_BUF_SIZE);

	while (conn->state != STREAM_DONE) {
		memset(buf, 0, READ_BUF_SIZE);
		count = ioread(conn, buf);
		if ((count == -1) || (count == 0)) {
			DBG(zprintf("%s: count=%d, calling xmpp_close_conn()\n",
			    __func__, count));
			/*
			 * If buf has already been filled it was set to
			 * conn->parse_buf which gets free'd in
			 * xmpp_close_conn(). BUT to be safe free it here and
			 * set to NULL so it won't be free'd again.
			 */
			if (conn->parse_buf == NULL) {
				ASSERT(buf != NULL);
				Xfree(buf);
				buf = NULL;
			}
			xmpp_close_conn(conn);
			/* NOTREACHED - thread killed */
		}

		conn->parse_buf = buf;
		conn->parse_len = count;

		rv = xmlParseChunk(conn->parser, buf, count, 0);
		if (rv != 0)
			DBG(zprintf("Parse Error = %d\n", rv));

		switch (conn->state) {
		case SUCCESS_SENT:
		case TLS_START:
			xmlCtxtResetPush(conn->parser, NULL, NULL, NULL,
			    NULL);
			break;
		case STREAM_DONE:
			DBG(zprintf("%s: state == STREAM_DONE\n", __func__));
			xmpp_close_conn(conn);
			/* NOTREACHED - thread killed */
			break;
		case LDM_INTERFACE_STARTED:
			conn->buf_size += count;
			conn->buf = Xrealloc(conn->buf, conn->buf_size);
			memcpy((conn->buf + conn->position), buf, count);
			conn->position += count;
			break;
		case LDM_INTERFACE_RESET:
			/*
			 * We have another LDM_interface which is already in
			 * conn->buf.  Simply change the state and get the
			 * rest of the data off the stream.
			 */
			xmpp_set_conn_state(conn, LDM_INTERFACE_STARTED);
			break;
		}
	}
	return (NULL);
}

/*
 * Callback to accept a new client connection.
 */
static void *
new_xmpp_client(cli_conn_t *conn)
{
	int alen = 0;
	cli_conn_t *xmpp_conn;
	int client_fd;

	client_fd = accept(conn->fd, NULL, &alen);
	if (client_fd < 0) {
		ewarning("accept");
	} else {
		xmpp_conn = Xmalloc(sizeof (cli_conn_t));
		xmpp_conn->class = LDM_CLASS_CONTROL;
		xmpp_conn->sock_conn = conn;
		xmpp_conn->fd = client_fd;
		xmpp_conn->conntime = time(NULL);
		xmpp_conn->buf = NULL;
		xmpp_conn->buf_size = 0;
		xmpp_conn->position = 0;
		xmpp_set_conn_state(xmpp_conn, SOCKET_WAITING);
		xmpp_conn->callout = NULL;
		xmpp_conn->tls_started = false;
		xmpp_conn->uname = NULL;
		xmpp_conn->password = NULL;
		xmpp_conn->xmpp_msg = true;
		xmpp_conn->xmpp_server = true;
		xmpp_conn->send_resp = true;
		/* cookie is used for domain migration */
		xmpp_conn->cookie = NULL;
		/* xml_command is used in the XMPP client code */
		xmpp_conn->xml_command = NULL;
		if (socket_peer_euid(client_fd) == -1) {
			xmpp_conn->local_socket = false;
		} else {
			xmpp_conn->local_socket = true;
		}
		mutex_init(&xmpp_conn->mp, USYNC_PROCESS, NULL);
		cond_init(&xmpp_conn->cvp, NULL, NULL);
		xmpp_conn->parser = xmlCreatePushParserCtxt(&my_handler,
		    xmpp_conn, NULL, NULL, NULL);
		if (xmpp_conn->parser == NULL) {
			DBG(zprintf("%s: parser == NULL", __func__));
			close(xmpp_conn->fd);
			Xfree(xmpp_conn);
			return (NULL);
		}
		init_response(&xmpp_conn->resp);
		CHAIN_INIT(xmpp_conn->waiters);
		xmpp_conn->parse_len = 0;
		xmpp_conn->parse_buf = NULL;
		xmpp_conn->left = 0;

		if (thr_create(NULL, NULL, xmpp_reader, (void *)xmpp_conn,
		    THR_DAEMON, NULL) != 0)
			efatal("Could not create client_loop thread");
		else {
			mutex_lock(&xmpp_thr_mutex);
			xmpp_thr_count++;
			mutex_unlock(&xmpp_thr_mutex);
		}
	}

	/* reset to poll for new clients */
	mutex_lock(&xmpp_thr_mutex);
	if (xmpp_thr_count < MAX_THREADS) {
		if (port_associate(evport, PORT_SOURCE_FD, conn->fd, POLLIN,
		    conn) < 0)
			efatal("port_associate");
	} else {
		DBG(zprintf("MAX CONNECTIONS(%d) REACHED\n", MAX_THREADS));
	}
	mutex_unlock(&xmpp_thr_mutex);

	return (NULL);
}

void
xml_send_async_cmd_result(cli_conn_t *client, xmlChar *cmd,
    char *ldom_name, unsigned int status)
{
	response_t	*resp;
	xmlDocPtr	ret_doc;
	xmlNodePtr	root;
	xmlNodePtr	ret_cmd;
	xmlNodePtr	ret_data;
	xmlNodePtr	new_env;
	xmlNodePtr	new_content;
	char		*resp_msg;
	char		*resp_str;
	char		*xml_buf;
	int		xml_buf_len;
	char		*xml_status;

	DBG(zprintf("\n*** %s Complete, Sending Result ***\n\n", (char *)cmd));
	resp = client->resp;

	xml_status = (status == CLI_RESULT_OK ? XML_SUCCESS : XML_FAILURE);

	resp->ret = LDOM_OK;
	close_audit_stream(client->audit_handle, resp->ret);
	free(client->buf);
	client->buf = NULL;

	/* Create return document */
	xml_create_return_doc(&ret_doc);
	root = xmlDocGetRootElement(ret_doc);
	xml_add_ldm_interface_properties(root);

	/* Create <cmd> section */
	ret_cmd = xmlNewChild(root, NULL, XML_CMD, NULL);
	xmlNewChild(ret_cmd, NULL, XML_ACTION, cmd);

	/* Create <data> section and fill it in with return data info */
	ret_data = xmlNewChild(ret_cmd, NULL, XML_DATA, NULL);
	xmlNewProp(ret_data, VERSION_ATTR, LDOM_V3_DATA_VERSION);
	new_env = xmlNewChild(ret_data, NULL, ENVELOPE_NODE, NULL);
	new_content = xmlNewChild(new_env, NULL, CONTENT_NODE, NULL);
	xmlNewProp(new_content, TYPE_ATTR, VIRTUAL_SYSTEM_TYPE);
	xmlNewProp(new_content, ID_ATTR, (xmlChar*)ldom_name);

	/* msg for <data> section */
	resp_str = get_response_text(resp);
	if (resp_str == NULL) {
		if (status != CLI_RESULT_OK) {
			rprintf(resp, "Unknown %s failure\n", (char *)cmd);
			resp_str = get_response_text(resp);
			ASSERT(resp_str != NULL);
		}
	}

	if (resp_str != NULL)
		resp_msg = xml_parse_resp(resp_str);
	else
		resp_msg = NULL;

	/* <data> show status and msg */
	xml_add_status(ret_data, xml_status, resp_msg);

	/* <cmd> and root simply show failure status */
	xml_add_status(ret_cmd, xml_status, NULL);
	xml_add_status(root, xml_status, NULL);


	if (resp_msg)
		Xfree(resp_msg);
	if (resp_str)
		Xfree(resp_str);

	xmlDocDumpMemory(ret_doc, (xmlChar **)&xml_buf, &xml_buf_len);

	/*
	 * The buf should be null because xmpp_write_buffer has
	 * already cleaned it out.
	 */
	ASSERT(client->buf == NULL);
	client->buf = Xmalloc(xml_buf_len);
	(void *)memcpy(client->buf, xml_buf, xml_buf_len);
	client->buf_size = xml_buf_len;
	client->send_resp = true;

	xmpp_write_buffer(client, false);

	xmlFree(xml_buf);
	xmlFreeDoc(ret_doc);
}
