/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */

/*
 * Copyright 2010 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#ifndef	_ZPRI_H
#define	_ZPRI_H

#ifdef __cplusplus
extern "C" {
#endif

#include "basics.h"
#include "chain.h"
#include "dispctl.h"
#include "ldom.h"
#include "resource_id.h"
#include "pcie.h"

#define	PRI_CPU_FRAC_DENOM 1

struct pri_cpu {
	int id;
	int phys_id;
	/*
	 * For now the resource id is associated with the physical CPU.
	 * This will need to be rethought once sub CPU allocation is
	 * supported.
	 */
	resid_t *res_id;
	pri_mau_t *maup; 			/* strand's MAU */
	pri_core_t *core;			/* strand's core */
	CHAIN_LINK(pri_cpu_t, link);		/* next CPU in PRI */
	CHAIN_LINK(pri_cpu_t, mau_link); 	/* next strand sharing MAU */
	CHAIN_LINK(pri_cpu_t, core_link);	/* next CPU on same core */
	struct {
		int fraction_free;
		CHAIN(struct vcpu, vcpus);
	} bound;
};

struct pri_core {
	CHAIN(pri_cpu_t, cpus);			/* strands sharing the core */
	/*
	 * For setting MMU partition ID to handle strands on same core
	 * being assigned to different ldoms.
	 */
	res_pool_t mmu_free_partids;
	int idx;				/* core index */
	CHAIN_LINK(pri_core_t, link);		/* next in list of cores */
	CHAIN_LINK(pri_core_t, chip_link);	/* cores sharing the chip */
};

struct pri_chip {
	CHAIN(pri_core_t, cores);	/* the cores sharing the chip */
	int idx;			/* chip ID */
	CHAIN_LINK(pri_chip_t, link);	/* next in list of chips */
};

struct pri_mau {
	int mau_id;
	uint64_t mau_ino;
	uint64_t cwq_ino;		/* for N2 */
	CHAIN_LINK(pri_mau_t, link);	/* next in list of all MAUs */
	/*
	 * One mau per core, each associated with a number of strands.
	 */
	CHAIN(pri_cpu_t, cpus); 	/* the strands sharing the MAU */
	/*
	 * To ensure an mau shared by multiple
	 * cpus is accounted for only once.
	 */
	bool_t	flag_visited;
	struct {
		CHAIN(struct vmau, vmaus);
	} bound;
	/*
	 * Printable representations.
	 */
	char *repr;
	char *parse_repr;
};

struct pri_mblock {
	uint64_t pa_base;
	uint64_t size;
	CHAIN_LINK(pri_mblock_t, link);		/* all physical memory */
	struct {
		CHAIN(struct mblock, memory);
	} bound;
};

#define	ILLEGAL_RCID (0xffffffffffffffffULL)

struct pri_iodevice {
	bool_t		free;
	/*
	 * no_pcie_reset means the firmware doesn't support warm
	 * reset of this PCIe bus.  In this case, we do not send
	 * the guest_stop HVCTL operation for an domain with this IO device.
	 */
	bool_t		no_pcie_reset;
	uint64_t	disabled;
	char		*name;
	char		*cli_name;
	char		*device_type;
	char		*compatible;
	uint64_t	cfg_handle;
	uint64_t	rcid;		/* root complex ID */
	bool_t		is_pciex;	/* Is it a PCIe Root complex? */
	res_pool_t	*res_id_pool;	/* resource id pool for device */
	pcie_bus_res_t	busres;
	pcie_bus_res_mgmt_t	res_mgmt;
	CHAIN(pri_pcie_dev_t, dev_list);
	CHAIN(bound_iodevice_t, bound_ios);
	CHAIN_LINK(pri_iodevice_t, link);
};

typedef enum {
	SL_UNKNOWN		= 0x0,	/* slot status unknown */
	SL_EMPTY		= 0x1,	/* slot empty  */
	SL_OCCUPIED		= 0x2	/* slot has devices */
} pcie_slot_status_t;

struct pri_pcie_dev {
	bool_t		free;		/* true if dev not assigned to domain */
	uint64_t	disabled;	/* true if marked 'disabled' in MD */
	char		*skel_path;	/* like "/@400/@0/@d" */
	char		*rel_path;	/* relative to bus, like "@0/@d" */
	char		*name;		/* like "pci@400/pci@0/pci@c" */
	char		*cli_name;	/* NAC name for slot or MB location */
	md_node_t	*hvmd_pciedev;	/* used for HV MD generation */

	CHAIN(bound_pcie_dev_t, bound_devs); /* domains in which dev bound */
	CHAIN_LINK(pri_pcie_dev_t, link);    /* next dev under root complex */

	/* lock to protect slot status and subdevs */
	mutex_t		subdev_lock;		/* protect sub_devs chain */
	pcie_slot_status_t	status;		/* slot status */
	CHAIN(pri_pcie_subdev_t, sub_devs);	/* subdevices */
};

struct pri_pcie_subdev {
	char		*subdev_path;		/* like "network@1" */
	CHAIN_LINK(pri_pcie_subdev_t, link);	/* next/prev subdevice */
};

/*
 * Prom location, or file.
 */
struct pri_image_cont {
	char		*namep;
	uint64_t	pa_base;	/* for system prom only */
	uint64_t	size;
	CHAIN_LINK(pri_img_cont_t, link);
};

struct pri_image {
	char		*namep;
	bool_t		guest_bootable;
	pri_img_cont_t	*imgcp;
	uint64_t	offset;
	uint64_t	size;
	uint64_t	alignment;
	uint64_t	min_allocation;
	char		*versionp;
	char		*notep;
	CHAIN_LINK(pri_image_t, link);
};

typedef struct zeus_pri {
	md_t		*pri;
	md_t		*guest_template;
	md_t		*core_mdp;		/* system MD */
	int		max_guests;		/* number of guests */
	int		max_hv_ldcs;		/* number of HV LDCs */
	int		max_guest_ldcs;		/* number of LDCs per guest */
	int		max_guest_vids;		/* number of vids per guest */
	int		max_vcpus;		/* max. vcpus supported by HV */
	int		max_mblocks;		/* number of mblocks */
	int		max_guest_dependencies;	/* max. num. master domains */
	int		max_pagesize;		/* largest supported pagesize */
	int		md_elbow_room;		/* elbow room for MD growth */
	/*
	 * no_pcie_reset means the firmware doesn't support warm
	 * reset of the PCIe busses.  In this case, we do not send
	 * the guest_stop HVCTL operation for an I/O domain.
	 */
	bool_t		no_pcie_reset;
	uint64_t	stick_frequency;
	uint64_t	tod_frequency;		/* For legacy HV MDs */
	uint64_t	tod;			/* For legacy HV MDs */
	uint64_t	erpt_pa;		/* For legacy HV MDs */
	uint64_t	erpt_size;		/* For legacy HV MDs */
	CHAIN(pri_mblock_t, memory_list);
	CHAIN(pri_chip_t, chip_list);
	CHAIN(pri_core_t, core_list);
	CHAIN(pri_cpu_t, cpu_list);
	CHAIN(pri_mau_t, mau_list);
	CHAIN(pri_iodevice_t, iodevice_list);
	CHAIN(pri_image_t, image_list);		/* possible boot images */
	CHAIN(pri_img_cont_t, container_list);	/* container list */
	int		num_cpus;
	int		num_maus;
	int		*mau_inos;
	int		*cwq_inos;
	int		num_mau_inos;
	int		num_cwq_inos;
	bool_t		have_cwq;
	uint64_t	addr_congr_mask;
	char		*cpu_compatible;	/* supported cpu type(s) */
	/* RA bit definitions (as implemented within hypervisor) */
	uint64_t	ra2pa_shift;	/* 1st bit of RA denoting segment # */
	uint64_t	guest_max_mblocks;	/* number supported mblocks */
	bool_t		fw_dio_capability;	/* FW DIO capability */
} zeus_pri_t;

enum {
	GET_PRI_INVAL,
	GET_PRI_LIBPRI,
	GET_PRI_DS
};

/* Direct I/O Capability values */
#define	DIO_STATIC_CAPAB	0x1

extern int get_pri_method;
extern zeus_pri_t pri;

int pri_initialize(void);
pri_cpu_t *pri_find_pstrand(int pid);
pri_mau_t *pri_find_pmau_from_cpuid(int pid);
int grab_exec_units(md_t *mdp, const char *type_name,
    md_node_t **maunodesp);
ldom_t *get_mau_ldom(pri_mau_t *mau);
pri_image_t *pri_get_boot_image(uint64_t basepa);
void pri_delete_pstrand(int pid);
void pri_display_firmware(response_t *);
bool_t pri_pull_mau_ino(int ino);
int pri_alloc_mau_ino(void);
bool_t pri_pull_cwq_ino(int ino);
int pri_alloc_cwq_ino(void);
void pri_init_cpu_chips(void);
uint64_t pri_cpu_phys_id(md_t *pri, md_node_t *cpu_node);

#ifdef __cplusplus
}
#endif

#endif /* _ZPRI_H */
